export type Size = 'xs' | 'sm' | 'md' | 'lg' | 'xl' | '2xl'

export type Color =
  | 'vortico'
  | 'bosque'
  | 'bruma'
  | 'ciclon'
  | 'flama'
  | 'primary'
  | 'vortico-light'
  | 'bosque-light'
  | 'bruma-light'
  | 'ciclon-light'
  | 'flama-light'
  | 'primary-light'
  | 'vortico-dark'
  | 'bosque-dark'
  | 'bruma-dark'
  | 'ciclon-dark'
  | 'flama-dark'
  | 'primary-dark'
