<!doctype html><html lang="en"><head><meta charset="UTF-8"/><meta name="viewport" content="width=device-width,initial-scale=1"/><link rel="preconnect" href="https://fonts.googleapis.com"/><link rel="preconnect" href="https://fonts.gstatic.com" crossorigin/><link href="https://fonts.googleapis.com/css2?family=Lato:ital,wght@0,100;0,300;0,400;0,700;0,900;1,100;1,300;1,400;1,700;1,900&family=Montserrat:ital,wght@0,100..900;1,100..900&display=swap" rel="stylesheet"/><title>Page Not Found - Flama</title><style>:root {
  --scalar-border-width: 1px;

  --scalar-radius: 0px;
  --scalar-radius-lg: 0px;
  --scalar-radius-xl: 0px;

  --scalar-text-decoration: none;
  --scalar-text-decoration-hover: none;

  --scalar-font:
    'Lato', ui-sans-serif, system-ui, sans-serif, 'Apple Color Emoji', 'Segoe UI Emoji', 'Segoe UI Symbol',
    'Noto Color Emoji';
  --scalar-font-code:
    'Fira Mono', ui-monospace, SFMono-Regular, Menlo, Monaco, Consolas, 'Liberation Mono', 'Courier New', monospace;
}

.scalar-app:not(#headlessui-portal-root) {
  margin: calc(3.5rem + 1px) 0 0 0;
  height: calc(100vh - 6.5rem - 2px);
  min-height: calc(100vh - 6.5rem - 2px);
  max-height: calc(100vh - 6.5rem - 2px);
  overflow: auto;
}

:where(.t-doc__sidebar, .references-navigation-list) {
  height: calc(100vh - 6.5rem - 2px) !important;
  width: 18rem;
  min-height: calc(100vh - 6.5rem - 2px);
  max-height: calc(100vh - 6.5rem - 2px);
  overflow: auto !important;
  position: fixed !important;
  left: 0 !important;
  top: calc(3.5rem + 1px) !important;
  z-index: 10;
}

.dark-mode {
  color-scheme: dark;
}

.light-mode {
  color-scheme: light;
}

.light-mode,
.dark-mode {
  --scalar-scrollbar-color: #808290;
  --scalar-scrollbar-color-active: #808290;

  --scalar-shadow-1: 0 1px 3px 0 #aaabb6;
  --scalar-shadow-2: #aaabb6 0px 13px 20px 0px, #aaabb6 0px 3px 8px 0px, #aaabb6 0px 0 0 0.5px;

  /* Basic theme */
  --scalar-background-1: #e5e6e8;
  --scalar-background-2: #f5f5f6;
  --scalar-background-3: #cdced4;
  --scalar-background-accent: #f0ac81;

  --scalar-color-1: #575863;
  --scalar-color-2: #656775;
  --scalar-color-3: #4a4a54;

  --scalar-color-accent: #e25822;
  --scalar-border-color: #aaabb6;

  /* Advanced theme */
  --scalar-color-green: #00976e;
  --scalar-color-red: #9e744f;
  --scalar-color-yellow: #f0ac81;
  --scalar-color-blue: #00bbd5;
  --scalar-color-orange: #e25822;
  --scalar-color-purple: #008080;

  --scalar-button-1: var(--scalar-background-2);
  --scalar-button-1-hover: var(--scalar-background-accent);
  --scalar-button-1-color: var(--scalar-color-1);

  /* Document Sidebar */
  .t-doc__sidebar {
    --scalar-sidebar-background-1: var(--scalar-background-1);
    --scalar-sidebar-color-1: var(--scalar-color-1);
    --scalar-sidebar-color-2: var(--scalar-color-3);
    --scalar-sidebar-border-color: var(--scalar-border-color);

    --scalar-sidebar-item-hover-background: var(--scalar-background-accent);
    --scalar-sidebar-item-hover-color: var(--scalar-color-3);

    --scalar-sidebar-item-active-background: var(--scalar-background-3);
    --scalar-sidebar-color-active: var(--scalar-color-3);

    --scalar-sidebar-indent-border: var(--scalar-sidebar-border-color);
    --scalar-sidebar-indent-border-hover: var(--scalar-color-accent);
    --scalar-sidebar-indent-border-active: var(--scalar-color-3);

    --scalar-sidebar-search-background: var(--scalar-background-2);
    --scalar-sidebar-search-color: var(--scalar-color-2);
    --scalar-sidebar-search-border-color: var(--scalar-border-color);
  }
}

*::selection {
  background-color: color-mix(in srgb, var(--scalar-color-accent), transparent 75%);
}

* {
  border-radius: 0 !important;
}
.start[data-v-391b4e37]{padding:24px;display:flex;flex-flow:wrap;justify-content:space-between;position:relative;z-index:0}.swagger-editor .start[data-v-391b4e37]{padding-top:24px}.start-h1[data-v-391b4e37]{font-size:var(--scalar-heading-2);margin-top:0;line-height:1.45;margin-bottom:0;font-weight:var(--scalar-bold);color:var(--scalar-color-1);width:100%;position:relative}.start-h3[data-v-391b4e37]{font-size:var(--scalar-paragraph);margin-top:0;margin-bottom:6px;display:block;line-height:1.45;font-weight:var(--scalar-bold);color:var(--scalar-color-1);width:100%}.start-h1[data-v-391b4e37]:not(:first-of-type){margin-top:24px}.start-p[data-v-391b4e37]{font-size:var(--scalar-paragraph);color:var(--scalar-color-2);line-height:1.5;width:100%;margin-top:12px}.start-ul[data-v-391b4e37]{margin-top:12px;font-size:var(--scalar-paragraph);line-height:1.5;padding-left:0;list-style:initial;display:flex;flex-flow:wrap;gap:24px}.start-ul li[data-v-391b4e37]{margin:0;padding:0;list-style:none;width:calc(50% - 24px);color:var(--scalar-color-2)}.start-ul li[data-v-391b4e37]:first-of-type{margin-top:0}.start-section[data-v-391b4e37]{width:100%;margin-bottom:12px;display:flex;flex-flow:wrap}.start-section[data-v-391b4e37]:last-of-type{margin-bottom:48px}.start-h2[data-v-391b4e37]{background:var(--scalar-background-2);border-top-left-radius:var(--scalar-radius-lg);border-top-right-radius:var(--scalar-radius-lg);border:1px solid var(--scalar-border-color);color:var(--scalar-color-3);font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);padding:9px;width:100%}.start-item[data-v-391b4e37]{align-items:center;background:var(--scalar-background-2);border-right:1px solid var(--scalar-border-color);border-bottom:1px solid var(--scalar-border-color);color:var(--scalar-color-1);cursor:pointer;display:flex;flex:1;font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);padding:9px;text-transform:capitalize;-webkit-user-select:none;-moz-user-select:none;user-select:none}.start-section-integrations .start-item[data-v-391b4e37]:first-of-type{border-bottom-left-radius:var(--scalar-radius-lg);border-left:1px solid var(--scalar-border-color)}.start-section-integrations .start-item[data-v-391b4e37]:last-of-type{border-bottom-right-radius:var(--scalar-radius-lg)}.start-section-colors .start-item[data-v-391b4e37]{min-width:33.33%}.start-section-colors .start-item[data-v-391b4e37]:nth-child(3n+2){border-left:1px solid var(--scalar-border-color)}.start-section-colors .start-item[data-v-391b4e37]:last-of-type,.start-section-colors .start-item-active[data-v-391b4e37]:last-of-type:before{border-radius:0 0 var(--scalar-radius-lg) var(--scalar-radius-lg)}.start-item[data-v-391b4e37]:empty{pointer-events:none}.start-item svg[data-v-391b4e37]{width:14px;height:14px;margin-right:6px}.start-item[data-v-391b4e37]:hover{background:var(--scalar-background-3)}.start-item-active[data-v-391b4e37]{z-index:10;position:relative;color:var(--scalar-color-1)}.start-item-active[data-v-391b4e37]:before{border:1px solid var(--scalar-color-1);content:"";top:-1px;right:-1px;bottom:-1px;left:-1px;pointer-events:none;position:absolute}.start-section-color .start-item[data-v-391b4e37]{text-transform:capitalize}.start-cta[data-v-391b4e37]{display:flex;gap:12px;width:100%;margin-top:24px;margin-bottom:0}.start-row[data-v-391b4e37]{width:100%;margin-top:12px;overflow:hidden}.start-hero-copy[data-v-391b4e37]{background:var(--scalar-background-2);padding:12px;border-radius:var(--scalar-radius-lg)}.start-p-small[data-v-391b4e37]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-2);margin-bottom:12px;line-height:1.4}.start-cta[data-v-391b4e37]{margin-bottom:12px;width:-moz-fit-content;width:fit-content;white-space:nowrap}.start-copy[data-v-391b4e37]{padding:76px 48px 48px;display:flex;justify-content:center;align-items:center;flex-direction:column;text-align:center}.start-logo[data-v-391b4e37]{color:var(--scalar-color-1);margin-bottom:24px;width:72px;aspect-ratio:1;position:relative;box-shadow:var(--scalar-shadow-2);border-radius:50%}.start-logo[data-v-391b4e37]:before{content:"";width:300%;aspect-ratio:1;left:-100%;top:-100%;position:absolute;z-index:-1;border-radius:50%;background-size:24px 24px;box-shadow:inset 0 0 50px var(--scalar-background-1),inset 0 0 50px var(--scalar-background-1);background-image:linear-gradient(to right,var(--scalar-border-color) 1px,transparent 1px),linear-gradient(to bottom,var(--scalar-border-color) 1px,transparent 1px)}.start-logo svg[data-v-391b4e37]{width:100%;height:auto;background:var(--scalar-background-1);padding:3px;border-radius:50%;position:relative}@media screen and (max-width: 600px){.start-section-colors .start-item[data-v-391b4e37],.start-item[data-v-391b4e37]{width:100%;border-radius:0;border-right:none;border-top:1px solid var(--scalar-border-color)}.start-item[data-v-391b4e37]:empty{display:none}.start-h2[data-v-391b4e37]{border-bottom:none}.start li[data-v-391b4e37]{width:100%}.start-copy[data-v-391b4e37]{padding:48px 0 24px}}@media screen and (max-width: 1000px){.start[data-v-391b4e37]{padding:0;overflow:auto}}.sidebar-heading-type[data-v-f51e2805]{display:block;min-width:3.9em;overflow:hidden;line-height:14px;flex-shrink:0;color:#fff;color:color-mix(in srgb,var(--method-color, var(--scalar-color-1)),transparent 0%);text-transform:uppercase;font-size:10px;font-weight:var(--scalar-bold);text-align:right;position:relative;font-family:var(--scalar-font-code);white-space:nowrap;margin-left:3px}.sidebar-heading[data-v-18e6dc3d]{display:flex;gap:6px;color:var(--scalar-sidebar-color-2, var(--scalar-color-2));font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);word-break:break-word;line-height:1.385;max-width:100%;position:relative;cursor:pointer;border-radius:var(--scalar-radius);flex:1;padding-right:9px;-webkit-user-select:none;-moz-user-select:none;user-select:none}.sidebar-heading-link-method[data-v-18e6dc3d]{margin:0}.sidebar-heading.deprecated .sidebar-heading-link-title[data-v-18e6dc3d]{text-decoration:line-through}.sidebar-heading-link-title[data-v-18e6dc3d]{margin:0}.sidebar-heading[data-v-18e6dc3d]:hover{background:var( --scalar-sidebar-item-hover-background, var(--scalar-background-2) )}.sidebar-heading:hover .sidebar-heading-link-title[data-v-18e6dc3d]{color:var(--scalar-sidebar-item-hover-color)}.sidebar-heading-link[data-v-18e6dc3d]:focus-visible{outline:none}.sidebar-heading[data-v-18e6dc3d]:has(>.sidebar-heading-link:focus-visible){z-index:1;outline:1px solid var(--scalar-color-accent)}.active_page.sidebar-heading[data-v-18e6dc3d]:hover,.active_page.sidebar-heading[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-active, var(--scalar-color-accent));background:var( --scalar-sidebar-item-active-background, var(--scalar-background-accent) )}.active_page.sidebar-heading p[data-v-18e6dc3d]{font-weight:var(--scalar-sidebar-font-weight-active, var(--scalar-semibold))}.active_page.sidebar-heading:hover .sidebar-heading-link-title[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-active, var(--scalar-color-accent))}.sidebar-indent-nested .sidebar-indent-nested .sidebar-heading[data-v-18e6dc3d]:before{content:"";position:absolute;top:0;left:calc((var(--scalar-sidebar-level) * 12px));width:1px;height:100%;background:var(--scalar-sidebar-indent-border)}.sidebar-indent-nested .sidebar-indent-nested .sidebar-heading[data-v-18e6dc3d]:hover:before{background:var(--scalar-sidebar-indent-border-hover)}.sidebar-indent-nested .sidebar-indent-nested .active_page.sidebar-heading[data-v-18e6dc3d]:before{background:var(--scalar-sidebar-indent-border-active)}.sidebar-heading-link[data-v-18e6dc3d]{text-decoration:none;color:inherit;padding:6px 0;display:flex;flex:1;justify-content:space-between;gap:2px}.sidebar-heading p[data-v-18e6dc3d]{height:-moz-fit-content;height:fit-content;display:flex;align-items:center;font-weight:var(--scalar-sidebar-font-weight, var(--scalar-semibold))}.sidebar-heading p[data-v-18e6dc3d]:empty{display:none}.link-icon[data-v-18e6dc3d]{position:relative;left:4px}.sidebar-icon[data-v-18e6dc3d]{display:flex;align-items:center;justify-content:center;margin-right:6px;width:13px;height:13px}.sidebar-icon>svg[data-v-18e6dc3d]{width:13px;height:13px}.sidebar-group-item[data-v-18e6dc3d]{position:relative}.sidebar-heading-chevron[data-v-18e6dc3d]{margin:5px -5.5px 5px -9px}.sidebar-heading-chevron .toggle-nested-icon[data-v-18e6dc3d]:focus-visible{outline:none}.sidebar-heading[data-v-18e6dc3d]:has(.sidebar-heading-chevron .toggle-nested-icon:focus-visible){outline:none;box-shadow:inset 0 0 0 1px var(--scalar-color-accent)}.toggle-nested-icon[data-v-18e6dc3d]{color:var(--scalar-color-3);width:20px;height:20px;display:flex;align-items:center;justify-content:center}.active_page .toggle-nested-icon[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-active, var(--scalar-color-accent))}.toggle-nested-icon[data-v-18e6dc3d]:hover,.toggle-nested-icon[data-v-18e6dc3d]:focus-visible{color:currentColor}.action-menu[data-v-18e6dc3d]{position:absolute;top:5px;right:5px;display:flex;gap:6px}.action-menu[data-v-18e6dc3d] .button-wrapper button{opacity:0;width:20px;height:20px;padding:4px}.action-menu[data-v-18e6dc3d]:hover .button-wrapper button,.action-menu[data-v-18e6dc3d] .button-wrapper button:hover,.sidebar-heading:hover~.action-menu[data-v-18e6dc3d] .button-wrapper button,.action-menu[data-v-18e6dc3d] .button-wrapper button[aria-expanded=true]{opacity:1}.sidebar-heading[data-v-18e6dc3d]:has(~.action-menu:hover){color:var(--scalar-sidebar-color-1, var(--scalar-color-1));background:var( --scalar-sidebar-item-hover-background, var(--scalar-background-2) )}.sidebar-group-item__folder[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-1, var(--scalar-color-1));text-transform:var(--scalar-tag-text-transform, initial)}.sidebar-group[data-v-39c84840]{list-style:none;width:100%;margin:0;padding:0}.sidebar-indent-nested[data-v-39c84840] .sidebar-heading{padding-left:calc((var(--scalar-sidebar-level) * var(--scalar-sidebar-indent-base)) + 12px)!important}.sidebar-indent-nested[data-v-39c84840] .sidebar-heading .toggle-nested-icon{left:calc((var(--scalar-sidebar-level) * var(--scalar-sidebar-indent-base)) + 2px)!important}:where(.sidebar-indent-nested[data-v-39c84840]) .sidebar-heading{color:var(--scalar-sidebar-color-1, var(--scalar-color-1))}:where(.sidebar-indent-nested[data-v-39c84840]) :where(.sidebar-indent-nested) .sidebar-heading{color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}.sidebar[data-v-b863366c]{--scalar-sidebar-indent-base: 12px;--scalar-sidebar-font-weight-active: var(--scalar-semibold);--scalar-sidebar-font-weight: var(--scalar-semibold);flex:1;height:100%;display:flex;flex-direction:column;border-right:var(--scalar-border-width) solid var(--scalar-sidebar-border-color, var(--scalar-border-color));background:var(--scalar-sidebar-background-1, var(--scalar-background-1));--scalar-sidebar-level: 0}.sidebar-pages[data-v-b863366c]{flex:1;padding:9px 12px}@media (max-width: 1000px){.sidebar[data-v-b863366c]{min-height:0;border-right:none}.sidebar-pages[data-v-b863366c]{padding-top:12px}}.sidebar-group-title[data-v-b863366c]{color:var(--scalar-sidebar-color-1);font-size:var(--scalar-mini);padding:12px 6px 6px;font-weight:var(--scalar-semibold);text-transform:uppercase;word-break:break-word;line-height:1.385}.sidebar-group-item+.sidebar-group-title[data-v-b863366c]{border-top:var(--scalar-border-width) solid var(--scalar-sidebar-border-color);margin-top:9px}.show-more[data-v-252cd03d]{background:var(--scalar-background-1);-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none;border:var(--scalar-border-width) solid var(--scalar-border-color);margin:auto;padding:8px 12px;border-radius:30px;color:var(--scalar-color-1);font-weight:var(--scalar-semibold);font-size:var(--scalar-small);display:flex;align-items:center;justify-content:center;position:relative;top:-48px}.show-more[data-v-252cd03d]:hover{color:var(--scalar-color-2);cursor:pointer}.show-more-icon[data-v-252cd03d]{width:16px!important;height:16px!important;margin-left:3px}.show-more[data-v-252cd03d]:active{box-shadow:0 0 0 1px var(--scalar-border-color)}@container narrow-references-container (max-width: 900px){.show-more[data-v-252cd03d]{top:-24px}}.screenreader-only[data-v-df2e1026]{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);border:0}.section[data-v-64ccd510]{position:relative;display:flex;flex-direction:column;max-width:var(--refs-content-max-width);margin:auto;padding:90px 0;scroll-margin-top:var(--refs-header-height)}.section[data-v-64ccd510]:has(~div.contents){border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.references-classic .section[data-v-64ccd510]{padding:48px 0;gap:24px}@container narrow-references-container (max-width: 900px){.references-classic .section[data-v-64ccd510],.section[data-v-64ccd510]{padding:48px 24px}}.section[data-v-64ccd510]:not(:last-of-type){border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.loading[data-v-c90b2c46]{background:var(--scalar-background-3);animation:loading-skeleton-c90b2c46 1.5s infinite alternate;border-radius:var(--scalar-radius-lg);min-height:1.6em;margin:.6em 0;max-width:100%}.loading[data-v-c90b2c46]:first-of-type{min-height:3em;margin-bottom:24px;margin-top:0}.loading[data-v-c90b2c46]:last-of-type{width:60%}.loading.single-line[data-v-c90b2c46]{min-height:3em;margin:.6em 0;max-width:80%}@keyframes loading-skeleton-c90b2c46{0%{opacity:1}to{opacity:.33}}.section-header-wrapper[data-v-b6aef53d]{display:grid;grid-template-columns:1fr}@media (min-width: 1200px){.section-header-wrapper[data-v-b6aef53d]{grid-template-columns:repeat(2,1fr)}}.section-header[data-v-b6aef53d]{font-size:var(--font-size, var(--scalar-heading-2));font-weight:var(--font-weight, var(--scalar-bold));color:var(--scalar-color-1);word-wrap:break-word;line-height:1.45;margin-top:0;margin-bottom:12px}.section-header.tight[data-v-b6aef53d]{margin-bottom:6px}.section-header.loading[data-v-b6aef53d]{width:80%}.label[data-v-3cb2dd80]{position:relative;display:inline-block;word-break:break-all}.anchor[data-v-3cb2dd80]{position:relative;display:inline-block;opacity:0}.anchor-copy[data-v-3cb2dd80]{position:absolute;left:0;top:50%;transform:translateY(-50%);cursor:pointer;padding:0 6px;color:var(--scalar-color-3);font-weight:var(--scalar-semibold);font-size:.8em}.anchor-copy[data-v-3cb2dd80]:hover,.anchor-copy[data-v-3cb2dd80]:focus-visible{color:var(--scalar-color-2)}.label:hover .anchor[data-v-3cb2dd80],.label:has(:focus-visible) .anchor[data-v-3cb2dd80]{opacity:1}@container narrow-references-container (max-width: 900px){.section-content--with-columns[data-v-9735459e]{flex-direction:column;gap:24px}}.section-columns[data-v-8b9602bf]{display:flex;gap:48px}@container narrow-references-container (max-width: 900px){.section-columns[data-v-8b9602bf]{flex-direction:column;gap:24px}}.section-column[data-v-699c28e3]{flex:1;min-width:0}@container narrow-references-container (max-width: 900px){.section-column[data-v-699c28e3]:nth-of-type(2){padding-top:0}}.scalar-card[data-v-38801681]{all:unset;font-family:var(--scalar-font);border-radius:var(--scalar-radius-lg);overflow:hidden;border:var(--scalar-border-width) solid var(--scalar-border-color);background:var(--scalar-background-2);display:flex;flex-direction:column;max-height:calc(((var(--full-height) - var(--refs-header-height)) - 60px) / 2);position:relative}.scalar-card-sticky[data-v-38801681]{position:sticky;top:calc(var(--refs-header-height) + 24px)}.scalar-card-content[data-v-dd83e46e]{overflow:auto;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color);display:grid}.scalar-card-content[data-v-dd83e46e] .simple-table .simple-header{display:none}.scalar-card-content[data-v-dd83e46e]:last-of-type,.scalar-card-content.scalar-card--borderless[data-v-dd83e46e]{border-bottom:none}.scalar-card--muted[data-v-dd83e46e]{background:var(--scalar-background-2)}.scalar-card--contrast[data-v-dd83e46e]{background:var(--scalar-background-3)}.scalar-card--frameless[data-v-dd83e46e]{padding:0}.scalar-card--transparent[data-v-dd83e46e]{background:var(--scalar-background-1)}.scalar-card-header[data-v-838dc259]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-3);padding:9px 3px 9px 12px;flex-shrink:0}.scalar-card-header.scalar-card--borderless[data-v-838dc259]+.scalar-card-content{margin-top:-9px}.scalar-card-header-slots[data-v-838dc259]{display:flex;justify-content:space-between;line-height:1.35}.scalar-card-header-title[data-v-838dc259]{text-transform:uppercase;flex:1;min-width:0;text-overflow:ellipsis;overflow:hidden}.scalar-card-header-actions[data-v-838dc259]{display:flex}.endpoints[data-v-cfef28a1]{overflow:auto;background:var(--scalar-background-2);padding:10px 12px}@media (max-width: 580px){.endpoints[data-v-cfef28a1]{max-height:calc(100vh - 150px)}}.endpoints span+span[data-v-cfef28a1]{text-align:left;margin-left:12px;text-transform:initial}.endpoint[data-v-cfef28a1]{display:flex;white-space:nowrap;cursor:pointer;text-decoration:none}.endpoint:hover .endpoint-path[data-v-cfef28a1],.endpoint:focus-visible .endpoint-path[data-v-cfef28a1]{text-decoration:underline}.endpoint span[data-v-cfef28a1]:first-of-type{text-transform:uppercase}.endpoint .post[data-v-cfef28a1],.endpoint .get[data-v-cfef28a1],.endpoint .delete[data-v-cfef28a1],.endpoint .put[data-v-cfef28a1]{white-space:nowrap}.endpoint span[data-v-cfef28a1]{color:var(--scalar-color-1);min-width:62px;display:inline-block;text-align:right;line-height:1.55;font-family:var(--scalar-font-code);font-size:var(--scalar-small);cursor:pointer}.deprecated[data-v-cfef28a1]{text-decoration:line-through}.section-container[data-v-1913755a]{position:relative;padding:0 60px;width:100%;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.section-container[data-v-1913755a]:has(.introduction-section){border-top:none}@container narrow-references-container (max-width: 900px){.section-container[data-v-1913755a]{padding:0}}.section-container[data-v-d47a803a]{border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.section-container[data-v-d47a803a]:has(.show-more){background-color:color-mix(in srgb,var(--scalar-background-2),transparent)}.operation-path[data-v-ec6c8861]{overflow:hidden;word-wrap:break-word;font-weight:var(--scalar-semibold);line-break:anywhere}.deprecated[data-v-ec6c8861]{text-decoration:line-through}.content-type[data-v-4b712394]{display:flex;align-items:center;font-size:var(--scalar-heading-4);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);line-height:1.45;margin-top:24px;padding-bottom:12px;border-bottom:1px solid var(--scalar-border-color);flex-flow:wrap}.content-type-select[data-v-4b712394]{position:relative;padding-left:9px;height:-moz-fit-content;height:fit-content;color:var(--scalar-color-2);font-size:var(--scalar-font-size-3);display:flex;align-items:center}.content-type-no-select.content-type-select[data-v-4b712394]{pointer-events:none}.content-type-no-select.content-type-select[data-v-4b712394]:after{display:none}.content-type-select span[data-v-4b712394]{display:flex;align-items:center}.content-type-select[data-v-4b712394]:after{content:"";width:7px;height:7px;transform:rotate(45deg) translate3d(-2px,-4px,0);display:block;margin-left:7px;box-shadow:1px 1px 0 currentColor}.content-type-select select[data-v-4b712394]{border:none;outline:none;cursor:pointer;background:var(--scalar-background-3);box-shadow:-2px 0 0 0 var(--scalar-background-3);position:absolute;top:0;left:0;width:100%;height:100%;opacity:0;-webkit-appearance:none;-moz-appearance:none;appearance:none}.content-type-select[data-v-4b712394]:hover{color:var(--scalar-color-1)}@media (max-width: 460px){.content-type-select[data-v-4b712394]{margin-left:auto;padding-right:3px}}.schema-type-icon[data-v-2a52eb0b]{color:var(--scalar-color-1);display:none}.schema-type[data-v-2a52eb0b]{font-family:var(--scalar-font-code);color:var(--scalar-color-1)}.error[data-v-f48d0197]{background-color:var(--scalar-color-red)}.schema-card[data-v-f48d0197]{z-index:0;position:relative;font-size:var(--scalar-font-size-4);color:var(--scalar-color-1)}.schema-card-title[data-v-f48d0197]{height:var(--schema-title-height);padding:6px 10px;display:flex;align-items:center;gap:4px;color:var(--scalar-color-2);font-weight:var(--scalar-semibold);font-size:var(--scalar-micro);border-bottom:var(--scalar-border-width) solid transparent}button.schema-card-title[data-v-f48d0197]{cursor:pointer}button.schema-card-title[data-v-f48d0197]:hover{color:var(--scalar-color-1)}.schema-card-title-icon--open[data-v-f48d0197]{transform:rotate(45deg)}.schema-properties-open>.schema-card-title[data-v-f48d0197]{border-bottom-left-radius:0;border-bottom-right-radius:0;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.schema-properties-open>.schema-properties[data-v-f48d0197]{width:-moz-fit-content;width:fit-content}.schema-card-description+.schema-properties[data-v-f48d0197]{width:-moz-fit-content;width:fit-content;margin-top:12px}.schema-properties-open.schema-properties[data-v-f48d0197],.schema-properties-open>.schema-card--open[data-v-f48d0197]{width:100%}.schema-card .property[data-v-f48d0197]:last-of-type{padding-bottom:10px}.schema-properties[data-v-f48d0197]{display:flex;flex-direction:column;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);width:-moz-fit-content;width:fit-content}.schema-properties .schema-properties[data-v-f48d0197]{border-radius:13.5px}.schema-properties .schema-properties.schema-properties-open[data-v-f48d0197]{border-radius:13.5px 13.5px 9px 9px}.schema-properties-open[data-v-f48d0197]{width:100%}.schema-card--compact[data-v-f48d0197]{align-self:flex-start}.schema-card--compact.schema-card--open[data-v-f48d0197]{align-self:initial}.schema-card-title--compact[data-v-f48d0197]{color:var(--scalar-color-2);padding:6px 8px;height:auto;border-bottom:none}.schema-card--compact>.schema-properties[data-v-f48d0197],.schema-card-title--compact[data-v-f48d0197]{border-radius:13.5px}.schema-card-title--compact>.schema-card-title-icon[data-v-f48d0197]{margin:0}.schema-card-title--compact>.schema-card-title-icon--open[data-v-f48d0197]{transform:rotate(45deg)}.schema-properties-open>.schema-card-title--compact[data-v-f48d0197]{position:static}.schema-card--compact.schema-card--level-0>.schema-properties[data-v-f48d0197]{border:none}[data-v-f48d0197] .schema-card-description p{font-size:var(--scalar-mini, var(--scalar-paragraph));color:var(--scalar-color-2);line-height:1.5;display:block;margin-bottom:6px}.schema-card-description[data-v-f48d0197]:first-of-type{padding-top:10px}.children .schema-card-description[data-v-f48d0197]:first-of-type{padding-top:0}.models-list-item .schema-properties[data-v-f48d0197]{margin-bottom:10px}.property-detail[data-v-3ae98a1b]{display:inline-flex}.property-detail+.property-detail[data-v-3ae98a1b]:before{display:block;content:"·";margin:0 .5ch}.property-detail-truncate[data-v-3ae98a1b]{overflow:hidden}.property-detail-truncate>.property-detail-value[data-v-3ae98a1b]{overflow:hidden;text-overflow:ellipsis}.property-detail-prefix[data-v-3ae98a1b]{color:var(--scalar-color-2)}code.property-detail-value[data-v-3ae98a1b]{font-family:var(--scalar-font-code);font-size:var(--scalar-font-size-3);color:var(--scalar-color-2);background:var(--scalar-background-3);padding:0 4px;border:.5px solid var(--scalar-border-color);border-radius:var(--scalar-radius)}.badge[data-v-190574ac]{color:var(--scalar-color-2);font-size:var(--scalar-micro);background:var(--scalar-background-2);padding:2px 6px;border-radius:12px;font-weight:var(--scalar-semibold);display:inline-block;text-transform:uppercase}.property-heading[data-v-dd95f4d7]{display:flex;flex-wrap:wrap;align-items:baseline;row-gap:9px;white-space:nowrap}.property-heading[data-v-dd95f4d7]>*{margin-right:9px}.property-heading[data-v-dd95f4d7]:last-child{margin-right:0}.property-heading>.property-detail[data-v-dd95f4d7]:not(:last-of-type){margin-right:0}.property-name[data-v-dd95f4d7]{font-family:var(--scalar-font-code);font-weight:var(--scalar-semibold);font-size:var(--scalar-font-size-3);display:flex}.property-additional[data-v-dd95f4d7]{font-family:var(--scalar-font-code)}.property-required[data-v-dd95f4d7],.property-optional[data-v-dd95f4d7]{color:var(--scalar-color-2)}.property-required[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-orange)}.property-read-only[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-blue)}.property-write-only[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-green)}.property-detail[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-2);display:flex;align-items:center;min-width:0}.property-const[data-v-dd95f4d7]{color:var(--scalar-color-1)}.property[data-v-f5c211dc]{color:var(--scalar-color-1);padding:10px;font-size:var(--scalar-mini)}.property[data-v-f5c211dc]:last-of-type{padding-bottom:0}.property--compact.property--level-0[data-v-f5c211dc]{padding:12px 0}.property--deprecated[data-v-f5c211dc]{background:repeating-linear-gradient(-45deg,var(--scalar-background-2) 0,var(--scalar-background-2) 2px,transparent 2px,transparent 5px);background-size:100%}.property--deprecated[data-v-f5c211dc]>*{opacity:.75}.property-description[data-v-f5c211dc]{margin-top:6px;line-height:1.4;font-size:var(--scalar-small)}.property-heading:empty+.property-description[data-v-f5c211dc]:last-of-type,.property-description[data-v-f5c211dc]:first-of-type:last-of-type{margin-top:0}[data-v-f5c211dc] .property-description *{color:var(--scalar-color-2)!important}.property[data-v-f5c211dc]:not(:last-of-type){border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.children[data-v-f5c211dc]{display:flex;flex-direction:column;padding-top:8px}.property-example[data-v-f5c211dc]{display:flex;flex-direction:column;margin-top:6px;max-height:calc(((var(--full-height) - var(--refs-header-height))) / 2);font-size:var(--scalar-micro);border:var(--scalar-border-width) solid var(--scalar-border-color);background:var(--scalar-background-2);border-radius:var(--scalar-radius-lg)}.property-example-label[data-v-f5c211dc]{font-weight:var(--scalar-semibold);color:var(--scalar-color-3);padding:6px}.property-example-value[data-v-f5c211dc]{all:unset;font-family:var(--scalar-font-code);padding:6px;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.property-rule[data-v-f5c211dc]{display:flex;flex-direction:column;gap:6px;margin-top:12px;border-radius:var(--scalar-radius-lg)}.property-enum-value[data-v-f5c211dc]{padding:3px 0;color:var(--scalar-color-2);line-height:1.5;word-break:break-word}.property-enum-value[data-v-f5c211dc]:before{content:"⊢";margin-right:6px;color:var(--scalar-color-3)}.property-enum-values[data-v-f5c211dc]{margin-top:8px;list-style:none}.property--compact .property-example[data-v-f5c211dc]{background:transparent;border:none;display:flex;flex-direction:row;gap:8px}.property--compact .property-example-label[data-v-f5c211dc],.property--compact .property-example-value[data-v-f5c211dc]{padding:3px 0 0}.property--compact .property-example-value[data-v-f5c211dc]{background:var(--scalar-background-2);border-top:0;border-radius:var(--scalar-radius);padding:3px 4px}.property-list[data-v-f5c211dc]{border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:var(--scalar-radius);margin-top:10px}.property-list .property[data-v-f5c211dc]:last-of-type{padding-bottom:10px}.property-name[data-v-f5c211dc]{font-family:var(--scalar-font-code);font-weight:var(--scalar-semibold)}.enum-toggle-button[data-v-f5c211dc]{align-items:center;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:13.5px;cursor:pointer;color:var(--scalar-color-2);display:flex;font-weight:var(--scalar-semibold);gap:4px;margin-top:8px;padding:6px 10px;-webkit-user-select:none;-moz-user-select:none;user-select:none;white-space:nowrap}.enum-toggle-button[data-v-f5c211dc]:hover{color:var(--scalar-color-1)}.enum-toggle-button-icon--open[data-v-f5c211dc]{transform:rotate(45deg)}.headers-card[data-v-2f99eae6]{z-index:0;margin-top:12px;margin-bottom:6px;position:relative;font-size:var(--scalar-font-size-4);color:var(--scalar-color-1);align-self:flex-start}.headers-card.headers-card--open[data-v-2f99eae6]{align-self:initial}.headers-card-title[data-v-2f99eae6]{padding:6px 10px;display:flex;align-items:center;gap:4px;color:var(--scalar-color-3);font-weight:var(--scalar-semibold);font-size:var(--scalar-micro);border-radius:13.5px}button.headers-card-title[data-v-2f99eae6]{cursor:pointer}button.headers-card-title[data-v-2f99eae6]:hover{color:var(--scalar-color-1)}.headers-card-title-icon--open[data-v-2f99eae6]{transform:rotate(45deg)}.headers-properties-open>.headers-card-title[data-v-2f99eae6]{border-bottom-left-radius:0;border-bottom-right-radius:0;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.headers-properties-open[data-v-2f99eae6]{width:100%}.headers-properties[data-v-2f99eae6]{display:flex;flex-direction:column;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:13.5px;width:-moz-fit-content;width:fit-content}.headers-card .property[data-v-2f99eae6]:last-of-type{padding-bottom:10px}.headers-card-title>.headers-card-title-icon[data-v-2f99eae6]{width:10px;height:10px;margin:0}.headers-card-title>.headers-card-title-icon--open[data-v-2f99eae6]{transform:rotate(45deg)}.parameter-item[data-v-e1efed5d]{display:flex;flex-direction:column;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.parameter-item:last-of-type .parameter-schema[data-v-e1efed5d]{padding-bottom:0}.parameter-item-container[data-v-e1efed5d]{padding:0}.parameter-item-headers[data-v-e1efed5d]{border:var(--scalar-border-width) solid var(--scalar-border-color)}.parameter-item-name[data-v-e1efed5d]{margin-right:6px;font-weight:var(--scalar-semibold);font-size:var(--scalar-font-size-3);font-family:var(--scalar-font-code);color:var(--scalar-color-1)}.parameter-item-type[data-v-e1efed5d]{font-size:var(--scalar-micro);color:var(--scalar-color-2);margin-right:6px;line-height:1.4;white-space:nowrap;text-overflow:ellipsis;width:100%;overflow:hidden}.parameter-item-trigger-open .parameter-item-type[data-v-e1efed5d]{white-space:normal}.parameter-item-trigger+.parameter-item-container[data-v-e1efed5d] .property--level-0>.property-heading .property-detail-value{font-size:var(--scalar-font-size-3)}.parameter-item-required-optional[data-v-e1efed5d]{color:var(--scalar-color-2);font-weight:var(--scalar-semibold);margin-right:6px;position:relative}.parameter-item--required[data-v-e1efed5d]{text-transform:uppercase;font-size:var(--scalar-micro);font-weight:var(--scalar-semibold);color:var(--scalar-color-orange)}.parameter-item-description[data-v-e1efed5d]{margin-top:3px!important;font-size:var(--scalar-small);color:var(--scalar-color-2);line-height:1.4}.parameter-item-description[data-v-e1efed5d] p{margin-top:4px;font-size:var(--scalar-small);color:var(--scalar-color-2);line-height:1.4}.parameter-schema[data-v-e1efed5d]{padding-bottom:9px;margin-top:3px}.parameter-item-trigger[data-v-e1efed5d]{padding:12px 0;cursor:pointer;outline:none;text-align:left;position:relative;align-items:baseline}.parameter-item-trigger-open[data-v-e1efed5d]{padding-bottom:0}.parameter-item-trigger[data-v-e1efed5d]:after{content:"";position:absolute;height:10px;width:100%;bottom:0}.parameter-item-icon[data-v-e1efed5d]{color:var(--scalar-color-3);height:18px;left:-19px;position:absolute;top:11px;width:18px}.parameter-item-trigger:hover .parameter-item-icon[data-v-e1efed5d],.parameter-item-trigger:focus-visible .parameter-item-icon[data-v-e1efed5d]{color:var(--scalar-color-1)}.parameter-item-trigger:focus-visible .parameter-item-icon[data-v-e1efed5d]{outline:1px solid var(--scalar-color-accent);outline-offset:2px;border-radius:var(--scalar-radius)}.parameter-item-content-type[data-v-e1efed5d]{margin-left:auto;opacity:0;transition:opacity .1s ease-in-out;color:var(--scalar-color-3);font-size:var(--scalar-micro);background:var(--scalar-background-2);padding:2px 6px;border-radius:12px}.parameter-item-trigger:hover .parameter-item-content-type[data-v-e1efed5d]{opacity:1}.parameter-list[data-v-b05cf5ff]{margin-top:24px}.parameter-list-title[data-v-b05cf5ff]{font-size:var(--scalar-font-size-2);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);line-height:1.45;margin-top:12px;margin-bottom:12px}.parameter-list-items[data-v-b05cf5ff]{list-style:none;padding:0;margin:0 0 12px;font-size:var(--scalar-small)}.request-body-title[data-v-dc8f9923]{display:flex;align-items:center;font-size:var(--scalar-font-size-2);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);line-height:1.45;margin-top:24px;padding-bottom:12px;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color);flex-flow:wrap}.request-body-title-select[data-v-dc8f9923]{position:relative;height:-moz-fit-content;height:fit-content;margin-left:auto;font-weight:var(--scalar-regular);display:flex;align-items:center;color:var(--scalar-color-3);font-size:var(--scalar-micro);background:var(--scalar-background-2);padding:2px 6px;border-radius:12px;border:var(--scalar-border-width) solid var(--scalar-border-color)}.request-body-title-no-select.request-body-title-select[data-v-dc8f9923]{pointer-events:none}.request-body-title-no-select[data-v-dc8f9923]{border:none}.request-body-title-no-select.request-body-title-select[data-v-dc8f9923]:after{display:none}.request-body-title-select span[data-v-dc8f9923]{display:flex;align-items:center}.request-body-title-select[data-v-dc8f9923]:after{content:"";width:6px;height:6px;transform:rotate(45deg) translate3d(0,-3px,0);display:block;margin-left:6px;box-shadow:1px 1px 0 currentColor;margin-right:5px}.request-body-title-select select[data-v-dc8f9923]{border:none;outline:none;cursor:pointer;background:var(--scalar-background-3);position:absolute;top:0;left:0;width:100%;height:100%;opacity:0;-webkit-appearance:none;-moz-appearance:none;appearance:none}.request-body-title-select[data-v-dc8f9923]:hover{color:var(--scalar-color-1)}.request-body-description[data-v-dc8f9923]{margin-top:6px;font-size:var(--scalar-small);width:100%}.request-body-description[data-v-dc8f9923] .markdown *{color:var(--scalar-color-2)!important}@media (max-width: 460px){.request-body-title-select[data-v-dc8f9923]{margin-left:auto;padding-right:3px}}.section-wrapper[data-v-07cabcb0]{color:var(--scalar-color-1);padding-top:12px;margin-top:-12px}.section-accordion[data-v-07cabcb0]{display:flex;flex-direction:column;border-radius:var(--scalar-radius-lg);background:var(--scalar-background-2)}.section-accordion-transparent[data-v-07cabcb0]{background:transparent;border:1px solid var(--scalar-border-color)}.section-accordion-button[data-v-07cabcb0]{padding:6px;display:flex;align-items:center;gap:6px;cursor:pointer}.section-accordion-button-content[data-v-07cabcb0]{flex:1;min-width:0}.section-accordion-button-actions[data-v-07cabcb0]{display:flex;align-items:center;gap:6px;color:var(--scalar-color-3)}.section-accordion-chevron[data-v-07cabcb0]{margin-right:4px;width:20px;cursor:pointer;opacity:1;color:var(--scalar-color-3)}.section-accordion-button:hover .section-accordion-chevron[data-v-07cabcb0]{color:var(--scalar-color-1)}.section-accordion-content[data-v-07cabcb0]{border-top:1px solid var(--scalar-border-color);display:flex;flex-direction:column}.section-accordion-description[data-v-07cabcb0]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color--1);padding:10px 12px 0}.section-accordion-content-card[data-v-07cabcb0] .property:last-of-type{padding-bottom:9px}.show-api-client-button[data-v-ea8189a2]{-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none;padding:4px 6px;white-space:nowrap;border-radius:var(--scalar-radius);display:flex;justify-content:center;align-items:center;font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-background-2);font-family:var(--scalar-font);background:var(--scalar-button-1);position:relative;cursor:pointer;box-sizing:border-box;box-shadow:inset 0 0 0 1px #0000001a;outline-offset:2px}.show-api-client-button span[data-v-ea8189a2],.show-api-client-button svg[data-v-ea8189a2]{fill:currentColor;color:var(--scalar-button-1-color);z-index:1}.show-api-client-button[data-v-ea8189a2]:hover{background:var(--scalar-button-1-hover)}.show-api-client-button svg[data-v-ea8189a2]{margin-right:4px}.empty-state[data-v-4b51e8ca]{margin:10px 0 10px 12px;text-align:center;font-size:var(--scalar-micro);min-height:56px;display:flex;align-items:center;justify-content:center;border-radius:var(--scalar-radius-lg);color:var(--scalar-color-2)}.rule-title[data-v-4b51e8ca]{font-family:var(--scalar-font-code);color:var(--scalar-color-1);display:inline-block;margin:12px 0 6px;border-radius:var(--scalar-radius)}.rule[data-v-4b51e8ca]{margin:0 12px;border-radius:var(--scalar-radius-lg)}.rule-items[data-v-4b51e8ca]{counter-reset:list-number;display:flex;flex-direction:column;gap:12px;border-left:1px solid var(--scalar-border-color);padding:12px 0}.rule-item[data-v-4b51e8ca]{counter-increment:list-number;border:1px solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);overflow:hidden;margin-left:24px}.rule-item[data-v-4b51e8ca]:before{border:1px solid var(--scalar-border-color);border-top:0;border-right:0;content:" ";display:block;width:24px;height:6px;border-radius:0 0 0 var(--scalar-radius-lg);margin-top:6px;color:var(--scalar-color-2);transform:translate(-25px);color:var(--scalar-color-1);position:absolute}.tab-list[data-v-03fdb072]{display:flex;gap:6px;position:relative;flex:1;padding:9px 12px;overflow:auto}.scalar-card-header.scalar-card-header-tabs[data-v-03fdb072]{padding:0}.tab[data-v-d0e07e65]{background:none;border:none;font-size:var(--scalar-mini);font-family:var(--scalar-font);color:var(--scalar-color-2);font-weight:var(--scalar-semibold);line-height:calc(var(--scalar-mini) + 2px);white-space:nowrap;cursor:pointer;padding:0;margin-right:3px;text-transform:uppercase;position:relative;line-height:1.35}.tab[data-v-d0e07e65]:before{content:"";position:absolute;z-index:0;left:-6px;top:-6px;width:calc(100% + 12px);height:calc(100% + 12px);border-radius:var(--scalar-radius);background:var(--scalar-background-3);opacity:0}.tab[data-v-d0e07e65]:hover:before,.tab[data-v-d0e07e65]:focus-visible:before{opacity:1}.tab[data-v-d0e07e65]:focus-visible:before{outline:1px solid var(--scalar-color-accent)}.tab span[data-v-d0e07e65]{z-index:1;position:relative}.tab-selected[data-v-d0e07e65]{color:var(--scalar-color-1);text-decoration:underline;text-underline-offset:var(--scalar-micro)}.scalar-card-footer[data-v-7621d1be]{flex-shrink:0}.text-select{position:relative;height:-moz-fit-content;height:fit-content}.text-select:has(:focus-visible) .text-select-label{outline:1px solid var(--scalar-color-accent);outline-offset:4px;border-radius:calc(var(--scalar-radius) / 2)}.text-select--single-option{pointer-events:none}.text-select select{border:none;outline:none;cursor:pointer;background:var(--scalar-background-3);box-shadow:-2px 0 0 0 var(--scalar-background-3);position:absolute;top:0;left:0;width:100%;height:100%;opacity:0;-moz-appearance:none;-webkit-appearance:none;appearance:none}.text-select span{font-size:var(--scalar-mini);color:var(--scalar-color-2);font-weight:var(--scalar-semibold);white-space:nowrap;display:flex;align-items:center;justify-content:center}.text-select:hover span{color:var(--scalar-color-1)}.text-select:not(.text-select--single-option) span:after{content:"";width:7px;height:7px;transform:rotate(45deg) translate3d(-2px,-2px,0);display:block;margin-left:6px;box-shadow:1px 1px 0 currentColor}.text-select span:hover{background:var(--scalar-background-2)}.example-selector[data-v-a7e22e14]{padding:4px}.markdown[data-v-d84b9039] *{margin:0}.code-copy[data-v-d84b9039]{display:flex;align-items:center;justify-content:center;-moz-appearance:none;appearance:none;-webkit-appearance:none;outline:none;background:transparent;cursor:pointer;color:var(--scalar-color-3);border:none;padding:0;margin-right:12px}.code-copy[data-v-d84b9039]:hover{color:var(--scalar-color-1)}.code-copy svg[data-v-d84b9039]{width:13px;height:13px}.response-card-footer[data-v-d84b9039]{display:flex;flex-direction:column;flex-shrink:0;padding:10px 12px;gap:8px;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.response-example-selector[data-v-d84b9039]{align-self:flex-start;margin:-4px}.response-description[data-v-d84b9039]{font-weight:var(--scalar-semibold);font-size:var(--scalar-micro);color:var(--scalar-color--1);display:flex;align-items:center;box-sizing:border-box}.schema-type[data-v-d84b9039]{font-size:var(--scalar-micro);color:var(--scalar-color-2);font-weight:var(--scalar-semibold);background:var(--scalar-background-3);padding:2px 4px;border-radius:4px;margin-right:4px}.schema-example[data-v-d84b9039]{font-size:var(--scalar-micro);color:var(--scalar-color-2);font-weight:var(--scalar-semibold)}.example-response-tab[data-v-d84b9039]{display:block;margin:6px}.scalar-card-container[data-v-d84b9039]{flex:1;background:var(--scalar-background-2);display:grid}.scalar-card-container[data-v-d84b9039] .cm-scroller{overflow-y:hidden}.scalar-card-checkbox[data-v-d84b9039]{display:flex;align-items:center;justify-content:center;position:relative;min-height:17px;cursor:pointer;-webkit-user-select:none;-moz-user-select:none;user-select:none;font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-2);width:-moz-fit-content;width:fit-content;white-space:nowrap;margin-right:9px;gap:6px}.scalar-card-checkbox:has(.scalar-card-checkbox-input:focus-visible) .scalar-card-checkbox-checkmark[data-v-d84b9039]{outline:1px solid var(--scalar-color-accent)}.scalar-card-checkbox[data-v-d84b9039]:hover{color:var(--scalar-color--1)}.scalar-card-checkbox .scalar-card-checkbox-input[data-v-d84b9039]{position:absolute;opacity:0;cursor:pointer;height:0;width:0}.scalar-card-checkbox-checkmark[data-v-d84b9039]{height:17px;width:17px;border-radius:var(--scalar-radius);background-color:transparent;background-color:var(--scalar-background-3);box-shadow:inset 0 0 0 1px var(--scalar-border-color)}.scalar-card-checkbox[data-v-d84b9039]:has(.scalar-card-checkbox-input:checked){color:var(--scalar-color-1)}.scalar-card-checkbox .scalar-card-checkbox-input:checked~.scalar-card-checkbox-checkmark[data-v-d84b9039]{background-color:var(--scalar-button-1);box-shadow:none}.scalar-card-checkbox-checkmark[data-v-d84b9039]:after{content:"";position:absolute;display:none}.scalar-card-checkbox .scalar-card-checkbox-input:checked~.scalar-card-checkbox-checkmark[data-v-d84b9039]:after{display:block}.scalar-card-checkbox .scalar-card-checkbox-checkmark[data-v-d84b9039]:after{right:6px;top:36.5%;width:5px;height:9px;border:solid 1px var(--scalar-button-1-color);border-width:0 1.5px 1.5px 0;transform:rotate(45deg)}.request[data-v-3ff33ef0]{display:flex;flex-wrap:nowrap}.request-header[data-v-3ff33ef0]{display:flex;gap:6px;text-transform:initial}.request-method[data-v-3ff33ef0]{font-family:var(--scalar-font-code);text-transform:uppercase}.request-client-picker[data-v-3ff33ef0]{padding-left:12px;padding-right:9px}.request-card-footer[data-v-3ff33ef0]{display:flex;justify-content:flex-end;padding:6px;flex-shrink:0}.request-card-footer-addon[data-v-3ff33ef0]{display:flex;align-items:center;flex:1;min-width:0}.request-editor-section[data-v-3ff33ef0]{display:flex;flex:1}.request-card-simple[data-v-3ff33ef0]{display:flex;align-items:center;justify-content:space-between;padding:8px 8px 8px 12px;font-size:var(--scalar-small)}.code-snippet[data-v-3ff33ef0]{display:flex;flex-direction:column;width:100%}.operation-title[data-v-e104650f]{display:flex;justify-content:space-between}.operation-details[data-v-e104650f]{display:flex;align-items:center;margin-top:0;gap:9px;min-width:0;flex-shrink:1}.endpoint-type[data-v-e104650f]{display:flex;align-items:center;justify-content:center;gap:6px;position:relative;z-index:0;width:60px;padding:6px;flex-shrink:0;font-size:var(--scalar-small);text-transform:uppercase;font-weight:var(--scalar-bold);font-family:var(--scalar-font)}.endpoint-type[data-v-e104650f]:after{content:"";position:absolute;top:0;right:0;bottom:0;left:0;z-index:-1;background:currentColor;opacity:.15;border-radius:var(--scalar-radius-lg)}.endpoint-anchor[data-v-e104650f]{display:flex;align-items:center;min-width:0;flex-shrink:1;font-size:20px}.endpoint-anchor.label[data-v-e104650f]{display:flex}.endpoint-label[data-v-e104650f]{display:flex;align-items:baseline;gap:9px;min-width:0;flex-shrink:1;color:var(--scalar-color-1)}.endpoint-label-path[data-v-e104650f]{font-family:var(--scalar-font-code);font-size:var(--scalar-mini);text-overflow:ellipsis;white-space:nowrap;overflow:hidden}.endpoint-label-path[data-v-e104650f] em{color:var(--scalar-color-2)}.endpoint-label-name[data-v-e104650f]{color:var(--scalar-color-2);font-size:var(--scalar-small);flex-shrink:1000000000;text-overflow:ellipsis;white-space:nowrap;overflow:hidden}.endpoint-try-hint[data-v-e104650f]{padding:2px;height:24px;width:24px;flex-shrink:0}.endpoint-copy[data-v-e104650f]{color:currentColor;padding:2px}.endpoint-copy[data-v-e104650f] svg{stroke-width:2px}.endpoint-content[data-v-e104650f]{display:grid;grid-auto-columns:1fr;grid-auto-flow:row;gap:9px;padding:9px}@media (min-width: 1000px){.endpoint-content[data-v-e104650f]{grid-auto-flow:column}}@container (max-width: 900px){.endpoint-content[data-v-e104650f]{grid-template-columns:1fr}}.endpoint-content[data-v-e104650f]>*{max-height:unset}.operation-details-card[data-v-e104650f]{display:flex;flex-direction:column;gap:12px}.operation-details-card-item[data-v-e104650f] .parameter-list{border:1px solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);margin-top:0}.operation-details-card-item[data-v-e104650f]{display:flex;flex-direction:column;gap:12px}.operation-details-card-item[data-v-e104650f] .parameter-list-items{margin-bottom:0}.operation-details-card[data-v-e104650f] .parameter-item:last-of-type .parameter-schema{padding-bottom:12px}.operation-details-card[data-v-e104650f] .parameter-list .parameter-list{margin-bottom:12px}.operation-details-card[data-v-e104650f] .parameter-item{margin:0;padding:0 9px}.operation-details-card[data-v-e104650f] .property{padding:9px;margin:0}.operation-details-card[data-v-e104650f] .parameter-list-title,.operation-details-card[data-v-e104650f] .request-body-title{text-transform:uppercase;font-weight:var(--scalar-bold);font-size:var(--scalar-mini);color:var(--scalar-color-2);line-height:1.33;padding:9px;margin:0}.operation-details-card[data-v-e104650f] .request-body-title-select{text-transform:initial;font-weight:initial;margin-left:auto}.examples[data-v-0916dd9f]{position:sticky;top:calc(var(--refs-header-height) + 24px)}.deprecated[data-v-0916dd9f] *{text-decoration:line-through}.example-path[data-v-0916dd9f]{color:var(--scalar-color-2);font-family:var(--scalar-font-code)}.example-path[data-v-0916dd9f] em{color:var(--scalar-color-1);font-style:normal}.references-loading{position:absolute;top:0;left:0;right:0;z-index:1;grid-area:rendered;background:var(--scalar-background-1)}.references-loading-top-spacer{top:-1px}.references-loading-hidden-tag .section-container>.section:first-child{display:none}.download[data-v-1fc94525]{margin-bottom:24px}.download-button[data-v-1fc94525]{color:var(--scalar-link-color, var(--scalar-color-accent));font-weight:var(--scalar-link-font-weight, inherit);-webkit-text-decoration:var(--scalar-text-decoration)!important;text-decoration:var(--scalar-text-decoration)!important;text-decoration-color:var(--scalar-text-decoration-color);font-size:var(--scalar-paragraph);cursor:pointer}.download-button[data-v-1fc94525]:hover{color:var(--scalar-link-color-hover, var(--scalar-color-accent));-webkit-text-decoration:var(--scalar-text-decoration-hover)!important;text-decoration:var(--scalar-text-decoration-hover)!important;text-decoration-color:var(--scalar-text-decoration-color-hover)}.introduction-description-heading[data-v-fa3d93c5]{scroll-margin-top:64px;margin-top:1em;margin-bottom:.5em}.markdown+.markdown[data-v-fa3d93c5]{margin-top:1em}.introduction-description[data-v-fa3d93c5]{display:flex;flex-direction:column}.references-classic .introduction-description[data-v-fa3d93c5] img{max-width:720px}.sticky-cards[data-v-0336cd4f]{display:flex;flex-direction:column;position:sticky;top:calc(var(--refs-header-height) + 24px)}.client-libraries-content[data-v-b29373e8]{container:client-libraries-content / inline-size;display:flex;justify-content:center;overflow:hidden;padding:0 12px;background-color:var(--scalar-background-1);border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.client-libraries[data-v-b29373e8]{display:flex;align-items:center;justify-content:center;width:100%;position:relative;cursor:pointer;white-space:nowrap;padding:8px 2px;gap:6px;color:var(--scalar-color-3);border-bottom:1px solid transparent;-webkit-user-select:none;-moz-user-select:none;user-select:none}.client-libraries[data-v-b29373e8]:first-child{border-radius:var(--scalar-radius) 0 0 0}.client-libraries[data-v-b29373e8]:not(.client-libraries__active):hover:before{content:"";position:absolute;width:calc(100% - 4px);height:calc(100% - 4px);background:var(--scalar-background-2);left:2px;top:2px;z-index:0;border-radius:var(--scalar-radius)}.client-libraries[data-v-b29373e8]:active{color:var(--scalar-color-1)}.client-libraries[data-v-b29373e8]:focus-visible{outline:none;box-shadow:inset 0 0 0 1px var(--scalar-color-accent)}@media screen and (max-width: 450px){.client-libraries[data-v-b29373e8]:nth-of-type(4),.client-libraries[data-v-b29373e8]:nth-of-type(5){display:none}}.client-libraries-icon[data-v-b29373e8]{max-width:14px;max-height:14px;min-width:14px;width:100%;aspect-ratio:1;display:flex;align-items:center;justify-content:center;position:relative;box-sizing:border-box;color:currentColor}.client-libraries-icon__more svg[data-v-b29373e8]{height:initial}@container client-libraries-content (width < 400px){.client-libraries__select[data-v-b29373e8]{width:-moz-fit-content;width:fit-content}.client-libraries__select .client-libraries-icon__more+span[data-v-b29373e8]{display:none}}@container client-libraries-content (width < 380px){.client-libraries[data-v-b29373e8]{width:100%}.client-libraries span[data-v-b29373e8]{display:none}}.client-libraries__active[data-v-b29373e8]{color:var(--scalar-color-1);border-bottom:1px solid var(--scalar-color-1)}@keyframes codeloader-b29373e8{0%{transform:rotate(0)}to{transform:rotate(1turn)}}.client-libraries .client-libraries-text[data-v-b29373e8]:last-of-type{font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);position:relative;display:flex;align-items:center}.client-libraries__active .client-libraries-text[data-v-b29373e8]{color:var(--scalar-color-1)}.client-libraries__select select[data-v-b29373e8]{background:var(--scalar-background-3);color:var(--scalar-color-2);opacity:0;height:100%;width:100%;aspect-ratio:1;position:absolute;top:0;left:0;cursor:pointer;z-index:1;-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none}.client-libraries__select[data-v-b29373e8]:has(select:focus-visible){border-radius:var(--scalar-radius);box-shadow:inset 0 0 0 1px var(--scalar-color-accent)}@media screen and (max-width: 600px){.references-classic .client-libraries[data-v-b29373e8]{flex-direction:column}}.selected-client[data-v-35fd541a]{color:var(--scalar-color-1);font-size:var(--scalar-mini);font-family:var(--scalar-font-code);padding:9px 12px;border-top:none;white-space:nowrap;overflow:hidden;text-overflow:ellipsis;background:var(--scalar-background-1);border-top:var(--scalar-border-width) solid var(--scalar-border-color);min-height:-moz-fit-content;min-height:fit-content}.client-libraries-heading[data-v-35fd541a]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-1);padding:9px 12px;background-color:var(--scalar-background-2);display:flex;align-items:center;max-height:32px}.section-accordion-wrapper[data-v-1a30b4ff]{padding:0 60px}.section-accordion[data-v-1a30b4ff]{position:relative;width:100%;max-width:var(--refs-content-max-width);margin:auto}.section-accordion-content[data-v-1a30b4ff]{display:flex;flex-direction:column;gap:12px;padding-top:12px}.section-accordion-button[data-v-1a30b4ff]{width:100%;display:flex;cursor:pointer;padding:6px 0;margin:-6px 0;border-radius:var(--scalar-radius)}.section-accordion-chevron[data-v-1a30b4ff]{width:24px;position:absolute;left:-24px;height:auto;top:10px;color:var(--scalar-color-3)}.section-accordion-button:hover .section-accordion-chevron[data-v-1a30b4ff]{color:var(--scalar-color-1)}.section-accordion-title[data-v-1a30b4ff]{display:flex;flex-direction:column;align-items:flex-start;flex:1;padding:0 6px}.section-accordion-title[data-v-1a30b4ff] .section-header-wrapper{grid-template-columns:1fr}.section-accordion-title[data-v-1a30b4ff] .section-header{margin-bottom:0}@container narrow-references-container (max-width: 900px){.section-accordion-chevron[data-v-1a30b4ff]{width:16px;left:-16px;top:14px}.section-accordion-wrapper[data-v-1a30b4ff]{padding:calc(var(--refs-header-height)) 24px 0 24px}}.tag-section[data-v-43e1c813]{margin-bottom:48px}.tag-name[data-v-43e1c813]{text-transform:capitalize}.tag-description[data-v-43e1c813]{padding-bottom:4px;text-align:left}.webhook-request-body[data-v-cbaa1001]{margin-top:-18px}.collapsible-section[data-v-b32ceff4]{border-top:var(--scalar-border-width) solid var(--scalar-border-color);position:relative}.collapsible-section-header[data-v-b32ceff4]{color:var(--scalar-color-1)}.collapsible-section .collapsible-section-trigger[data-v-b32ceff4]{display:flex;align-items:center;cursor:pointer;padding:10px 0;font-size:var(--scalar-font-size-3);z-index:1;position:relative}.collapsible-section .collapsible-section-trigger[data-v-b32ceff4]:after{content:"";height:10px;width:100%;position:absolute;bottom:0}.collapsible-section-trigger svg[data-v-b32ceff4]{color:var(--scalar-color-3);position:absolute;left:-19px}.collapsible-section:hover .collapsible-section-trigger svg[data-v-b32ceff4]{color:var(--scalar-color-1)}.collapsible-section .collapsible-section-trigger[data-v-b32ceff4] .anchor-copy{line-height:18.5px}.collapsible-section-content[data-v-b32ceff4]{padding:0;margin:0;scroll-margin-top:140px}.webhooks-list[data-v-8cee6fce]{display:contents}.webhooks-list-truncated .webhooks-list-item[data-v-8cee6fce]:last-child{border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.webhooks-list-item[aria-expanded=true][data-v-8cee6fce]{padding-bottom:48px}.webhook-section-description.webhook-section-description[data-v-8cee6fce]{color:var(--scalar-color-2);font-size:var(--scalar-font-size-3);padding-bottom:12px}.show-more-webhooks[data-v-8cee6fce]{margin-top:32px;top:0}.reference-models[data-v-4396a256]{margin-bottom:48px}.reference-models-anchor[data-v-4396a256]{display:flex;align-items:center;font-size:20px;padding-left:6px;color:var(--scalar-color-1)}.reference-models-label[data-v-4396a256]{font-size:var(--scalar-mini)}.reference-models-label[data-v-4396a256] em{font-weight:var(--scalar-bold)}.models-list[data-v-7c1cd4b4]{display:contents}.models-list-truncated .models-list-item[data-v-7c1cd4b4]:last-child{border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.show-more-models[data-v-7c1cd4b4]{margin-top:32px;top:0}.narrow-references-container{container-name:narrow-references-container;container-type:inline-size}.render-loading[data-v-42cfaf8d]{height:calc(var(--full-height) - var(--refs-header-height));display:flex;align-items:center;justify-content:center}.introduction-card[data-v-42cfaf8d]{display:flex;flex-direction:column;background:var(--scalar-background-1)}.introduction-card-item[data-v-42cfaf8d]{display:flex;overflow:hidden;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);margin-bottom:12px;flex-direction:column;justify-content:start}@container narrow-references-container (max-width: 900px){.introduction-card-item[data-v-42cfaf8d]{border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}}.introduction-card-item[data-v-42cfaf8d]:has(.description) .server-form-container{border-bottom-left-radius:0;border-bottom-right-radius:0}.introduction-card-item[data-v-42cfaf8d] .request-item{border-bottom:0}.introduction-card-title[data-v-42cfaf8d]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-3)}.introduction-card-row[data-v-42cfaf8d]{gap:24px}@media (min-width: 600px){.introduction-card-row[data-v-42cfaf8d]{flex-flow:row wrap}}.introduction-card-row[data-v-42cfaf8d]>*{flex:1}@media (min-width: 600px){.introduction-card-row[data-v-42cfaf8d]>*{min-width:-moz-min-content;min-width:min-content}}@media (max-width: 600px){.introduction-card-row[data-v-42cfaf8d]>*{max-width:100%}}@container (max-width: 900px){.introduction-card-row[data-v-42cfaf8d]{flex-direction:column;align-items:stretch;gap:0px}}.introduction-card[data-v-42cfaf8d] .security-scheme-label{text-transform:uppercase;font-weight:var(--scalar-semibold)}.references-classic .introduction-card-row[data-v-42cfaf8d] .scalar-card:nth-of-type(2) .scalar-card-header{display:none}.references-classic .introduction-card-row[data-v-42cfaf8d] .scalar-card:nth-of-type(2) .scalar-card-header.scalar-card--borderless+.scalar-card-content{margin-top:0}.section-flare[data-v-42cfaf8d]{top:0;right:0;pointer-events:none}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}}@layer scalar-base{:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>");mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>")}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}}@layer scalar-base{:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-api-reference{color:var(--scalar-color-1)}.scalar-api-reference .container{width:100%}@media (min-width: 400px){.scalar-api-reference .container{max-width:400px}}@media (min-width: 600px){.scalar-api-reference .container{max-width:600px}}@media (min-width: 800px){.scalar-api-reference .container{max-width:800px}}@media (min-width: 1000px){.scalar-api-reference .container{max-width:1000px}}@media (min-width: 1200px){.scalar-api-reference .container{max-width:1200px}}.scalar-api-reference .row{display:flex;flex-direction:row}.scalar-api-reference .sr-only{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);white-space:nowrap;border-width:0}.scalar-api-reference .visible{visibility:visible}.scalar-api-reference .collapse{visibility:collapse}.scalar-api-reference .static{position:static}.scalar-api-reference .fixed{position:fixed}.scalar-api-reference .absolute{position:absolute}.scalar-api-reference .relative{position:relative}.scalar-api-reference .sticky{position:sticky}.scalar-api-reference .right-0{right:0}.scalar-api-reference .block{display:block}.scalar-api-reference .flex{display:flex}.scalar-api-reference .grid{display:grid}.scalar-api-reference .contents{display:contents}.scalar-api-reference .hidden{display:none}.scalar-api-reference .h-8{height:32px}.scalar-api-reference .flex-shrink,.scalar-api-reference .shrink{flex-shrink:1}.scalar-api-reference .transform{transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-api-reference .resize{resize:both}.scalar-api-reference .flex-wrap{flex-wrap:wrap}.scalar-api-reference .content-start{align-content:flex-start}.scalar-api-reference .content-end{align-content:flex-end}.scalar-api-reference .items-center{align-items:center}.scalar-api-reference .gap-12{gap:48px}.scalar-api-reference .divide-y>:not([hidden])~:not([hidden]){--tw-divide-y-reverse:0;border-top-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-y-reverse)));border-bottom-width:calc(var(--scalar-border-width) * var(--tw-divide-y-reverse))}.scalar-api-reference .truncate{overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.scalar-api-reference .border{border-width:var(--scalar-border-width)}.scalar-api-reference .bg-b-2{background-color:var(--scalar-background-2)}.scalar-api-reference .px-3{padding-left:12px;padding-right:12px}.scalar-api-reference .py-1{padding-top:4px;padding-bottom:4px}.scalar-api-reference .py-1\.5{padding-top:6px;padding-bottom:6px}.scalar-api-reference .py-2{padding-top:8px;padding-bottom:8px}.scalar-api-reference .py-2\.5{padding-top:10px;padding-bottom:10px}.scalar-api-reference .font{font-family:var(--scalar-font)}.scalar-api-reference .text-sm{font-size:var(--scalar-font-size-4)}.scalar-api-reference .font{font-weight:var(--scalar-regular)}.scalar-api-reference .font-medium{font-weight:var(--scalar-semibold)}.scalar-api-reference .uppercase{text-transform:uppercase}.scalar-api-reference .lowercase{text-transform:lowercase}.scalar-api-reference .capitalize{text-transform:capitalize}.scalar-api-reference .leading{line-height:1.5}.scalar-api-reference .text-c-2{color:var(--scalar-color-2)}.scalar-api-reference .text-c-3{color:var(--scalar-color-3)}.scalar-api-reference .text-green{color:var(--scalar-color-green)}.scalar-api-reference .text-orange{color:var(--scalar-color-orange)}.scalar-api-reference .text-red{color:var(--scalar-color-red)}.scalar-api-reference .underline{text-decoration-line:underline}.scalar-api-reference .outline{outline-style:solid}.scalar-api-reference .filter{filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-api-reference .transition{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,-webkit-backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter,-webkit-backdrop-filter;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-api-reference .\[--scalar-address-bar-height\:0px\]{--scalar-address-bar-height:0px}.scalar-api-reference .hover\:text-c-1:hover{color:var(--scalar-color-1)}.nav-item[data-v-d56f12d9]{padding:0 1rem;cursor:pointer;flex:1;justify-content:center;align-items:center;display:flex;border-radius:var(--scalar-radius-lg);background:var(--scalar-background-3);border:var(--scalar-border-width) solid var(--scalar-background-2);color:var(--scalar-color-3);padding:4.5px;min-width:0;overflow:hidden;position:relative}.dark-mode .nav-item[data-v-d56f12d9]{background:color-mix(in srgb,var(--scalar-background-2),transparent)}.nav-item-icon-copy[data-v-d56f12d9]{max-width:100%;white-space:nowrap;overflow:hidden;-webkit-mask-image:linear-gradient(to left,transparent 0,var(--scalar-background-2) 20px);mask-image:linear-gradient(to left,transparent 0,var(--scalar-background-2) 20px)}.nav-item:hover .nav-item-icon-copy[data-v-d56f12d9]{-webkit-mask-image:linear-gradient(to left,transparent 20px,var(--scalar-background-2) 40px);mask-image:linear-gradient(to left,transparent 20px,var(--scalar-background-2) 40px)}.nav-item-copy[data-v-d56f12d9]{max-width:calc(100% - 20px)}.nav-item[data-v-d56f12d9]:hover{color:var(--scalar-color-1)}.nav-item__active[data-v-d56f12d9]{background-color:var(--scalar-background-1);color:var(--scalar-color-1);border-color:var(--scalar-border-color)}.dark-mode .nav-item__active[data-v-d56f12d9]{background-color:var(--scalar-background-2)}.nav-item-close[data-v-d56f12d9]{position:absolute;right:3px;padding:2px;border-radius:var(--scalar-radius);background:transparent;max-width:20px;stroke-width:1.5px;color:var(--scalar-color-3);margin-left:-20px;opacity:0}.nav-item:hover .nav-item-close[data-v-d56f12d9]{opacity:1}.nav-item-close[data-v-d56f12d9]:hover{background-color:var(--scalar-background-4)}.nav-item__active .nav-item-close[data-v-d56f12d9]:hover{background-color:var(--scalar-background-2)}.download-app-button[data-v-c896207d]{box-shadow:0 0 0 .5px var(--scalar-border-color);background:linear-gradient(#ffffffbf,#00000009)}.dark-mode .download-app-button[data-v-c896207d]{background:linear-gradient(#ffffff1a,#00000026)}.download-app-button[data-v-c896207d]:hover{background:linear-gradient(#00000009,#ffffffbf)}.dark-mode .download-app-button[data-v-c896207d]:hover{background:linear-gradient(#00000026,#ffffff1a)}.http-bg-gradient[data-v-46723816]{background:linear-gradient(#ffffffbf,#00000009)}.http-bg-gradient[data-v-46723816]:hover{background:linear-gradient(#00000009,#ffffffbf)}.dark-mode .http-bg-gradient[data-v-46723816]{background:linear-gradient(#ffffff09,#00000026)}.dark-mode .http-bg-gradient[data-v-46723816]:hover{background:linear-gradient(#00000026,#ffffff09)}.commandmenu[data-v-e9fd9234]{box-shadow:var(--scalar-shadow-2);border-radius:var(--scalar-radius-lg);background-color:var(--scalar-background-1);max-height:60dvh;width:100%;max-width:580px;margin:12px;position:fixed;left:50%;top:150px;opacity:0;transform:translate3d(-50%,10px,0);animation:fadeincommandmenu-e9fd9234 ease-in-out .3s forwards;animation-delay:.1s}.commandmenu-overlay[data-v-e9fd9234]{position:fixed;top:0;right:0;bottom:0;left:0;background:#0003;animation:fadeincommand-e9fd9234 ease-in-out .3s forwards;cursor:pointer}@keyframes fadeincommand-e9fd9234{0%{opacity:0}to{opacity:1}}@keyframes fadeincommandmenu-e9fd9234{0%{opacity:0;transform:translate3d(-50%,10px,0)}to{opacity:1;transform:translate3d(-50%,0,0)}}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>");mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>")}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>");mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>")}@layer scalar-base{*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar .scalar-app-layout[data-v-cc5e3e81]{background:var(--scalar-background-1);height:calc(100% - 120px);max-width:1390px;width:100%;margin:auto;opacity:0;animation:scalarapiclientfadein-cc5e3e81 .35s forwards;position:relative;overflow:hidden;border-radius:8px;border:var(--scalar-border-width) solid var(--scalar-border-color)}@keyframes scalarapiclientfadein-cc5e3e81{0%{opacity:0}to{opacity:1}}.scalar .scalar-app-exit[data-v-cc5e3e81]{position:fixed;top:0;left:0;width:100vw;height:100vh;background:#00000038;transition:all .3s ease-in-out;cursor:pointer;animation:scalardrawerexitfadein-cc5e3e81 .35s forwards}.dark-mode .scalar .scalar-app-exit[data-v-cc5e3e81]{background:#00000073}.scalar .scalar-app-exit[data-v-cc5e3e81]:before{font-family:sans-serif;position:absolute;top:0;font-size:30px;font-weight:100;line-height:50px;right:12px;text-align:center;color:#fff;opacity:.6}.scalar .scalar-app-exit[data-v-cc5e3e81]:hover:before{opacity:1}@keyframes scalardrawerexitfadein-cc5e3e81{0%{opacity:0}to{opacity:1}}.scalar-container[data-v-cc5e3e81]{overflow:hidden;visibility:visible;position:fixed;bottom:0;top:0;left:0;width:100%;height:100%;display:flex;align-items:center;justify-content:center;z-index:10000}.scalar .url-form-input[data-v-cc5e3e81]{min-height:auto!important}.scalar .scalar-container[data-v-cc5e3e81]{line-height:normal}.scalar .scalar-app-header span[data-v-cc5e3e81]{color:var(--scalar-color-3)}.scalar .scalar-app-header a[data-v-cc5e3e81]{color:var(--scalar-color-1)}.scalar .scalar-app-header a[data-v-cc5e3e81]:hover{text-decoration:underline}.scalar-activate[data-v-cc5e3e81]{width:-moz-fit-content;width:fit-content;margin:0px .75rem .75rem auto;line-height:24px;cursor:pointer;font-size:.875rem;font-weight:600;display:flex;align-items:center;gap:6px}.scalar-activate-button[data-v-cc5e3e81]{display:flex;gap:6px;align-items:center;color:var(--scalar-color-blue);-webkit-appearance:none;-moz-appearance:none;appearance:none;outline:none;border:none;background:transparent;padding:0 .5rem}.scalar-activate:hover .scalar-activate-button[data-v-cc5e3e81]{background:var(--scalar-background-3);border-radius:3px}[data-v-4d46a9ea] .cm-editor{height:100%;outline:none;padding:0;background:transparent}[data-v-4d46a9ea] .cm-placeholder{color:var(--scalar-color-3)}[data-v-4d46a9ea] .cm-content{font-family:var(--scalar-font-code);font-size:var(--scalar-small);max-height:20px;padding:8px 0}[data-v-4d46a9ea] .cm-tooltip{background:transparent!important;filter:brightness(var(--scalar-lifted-brightness));border-radius:var(--scalar-radius);box-shadow:var(--scalar-shadow-2);border:none!important;outline:none!important;overflow:hidden!important}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul li{padding:3px 6px!important}[data-v-4d46a9ea] .cm-completionIcon-type:after{color:var(--scalar-color-3)!important}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul li[aria-selected]{background:var(--scalar-background-2)!important;color:var(--scalar-color-1)!important}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul{padding:6px!important;position:relative}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul li:hover{border-radius:3px;color:var(--scalar-color-1)!important;background:var(--scalar-background-3)!important}[data-v-4d46a9ea] .cm-activeLine,[data-v-4d46a9ea] .cm-activeLineGutter{background-color:transparent}[data-v-4d46a9ea] .cm-selectionMatch,[data-v-4d46a9ea] .cm-matchingBracket{border-radius:var(--scalar-radius);background:var(--scalar-background-4)!important}[data-v-4d46a9ea] .cm-css-color-picker-wrapper{display:inline-flex;outline:1px solid var(--scalar-background-3);border-radius:3px;overflow:hidden}[data-v-4d46a9ea] .cm-gutters{background-color:transparent;border-right:none;color:var(--scalar-color-3);font-size:var(--scalar-mini);line-height:1.44;border-radius:0 0 0 3px}[data-v-4d46a9ea] .cm-gutters:before{content:"";position:absolute;top:2px;left:2px;width:calc(100% - 2px);height:calc(100% - 4px);border-radius:var(--scalar-radius) 0 0 var(--scalar-radius);background-color:var(--scalar-background-1)}[data-v-4d46a9ea] .cm-gutterElement{font-family:var(--scalar-font-code)!important;padding:0 6px 0 8px!important;display:flex;align-items:center;justify-content:flex-end;position:relative}[data-v-4d46a9ea] .cm-gutter+.cm-gutter :not(.cm-foldGutter) .cm-gutterElement{padding-left:0!important}[data-v-4d46a9ea] .cm-scroller{overflow:auto}.peer:hover .copy-button[data-v-4d46a9ea],.copy-button[data-v-4d46a9ea]:focus-visible{opacity:100}.scalar-code-copy[data-v-4d46a9ea]{align-items:flex-start;display:flex;top:0;right:0;bottom:0;left:0;justify-content:flex-end;position:sticky}.copy-button[data-v-4d46a9ea]{align-items:center;display:flex;position:relative;background-color:var(--scalar-background-2);border:1px solid var(--scalar-border-color);border-radius:3px;color:var(--scalar-color-3);cursor:pointer;height:30px;margin-bottom:-30px;opacity:0;padding:6px;transition:opacity .15s ease-in-out,color .15s ease-in-out;top:0;right:0}.scalar-code-copy[data-v-4d46a9ea],.copy-button[data-v-4d46a9ea]{background:inherit}.copy-button[data-v-4d46a9ea]:hover{color:var(--scalar-color-1)}.copy-button svg[data-v-4d46a9ea]{stroke-width:1.5}.cm-pill{--tw-bg-base: var(--scalar-color-1);color:var(--tw-bg-base);padding:0 9px;display:inline-block;border-radius:30px;font-size:var(--scalar-mini);background:color-mix(in srgb,var(--tw-bg-base),transparent 94%)!important}.cm-pill.bg-grey{background:var(--scalar-background-3)!important}.dark-mode .cm-pill{background:color-mix(in srgb,var(--tw-bg-base),transparent 80%)!important}.cm-pill:first-of-type{margin-left:0}.cm-editor .cm-widgetBuffer{display:none}.cm-foldPlaceholder:hover{color:var(--scalar-color-1)}.cm-foldGutter .cm-gutterElement{font-size:var(--scalar-heading-4);padding:2px!important}.cm-foldGutter .cm-gutterElement:first-of-type{display:none}.cm-foldGutter .cm-gutterElement .cm-foldMarker{padding:2px}.cm-foldGutter .cm-gutterElement:hover .cm-foldMarker{background:var(--scalar-background-2);border-radius:var(--scalar-radius);color:var(--scalar-color-1)}[data-v-cb0ff5b1] .cm-editor{padding:0}[data-v-cb0ff5b1] .cm-content{align-items:center;background-color:transparent;display:flex;font-family:var(--scalar-font);font-size:var(--scalar-mini);padding:6px 8px}[data-v-cb0ff5b1] .cm-content:has(.cm-pill){padding:4px 3px}[data-v-cb0ff5b1] .cm-content .cm-pill:not(:last-of-type){margin-right:.5px}[data-v-cb0ff5b1] .cm-content .cm-pill:not(:first-of-type){margin-left:.5px}[data-v-cb0ff5b1] .cm-line{padding:0}.required[data-v-cb0ff5b1]:after{content:"Required"}input[data-v-cb0ff5b1]::-moz-placeholder{color:var(--scalar-color-3)}input[data-v-cb0ff5b1]::placeholder{color:var(--scalar-color-3)}.scalar-modal-layout .scalar-button-danger[data-v-b6a60dbc]{background:color-mix(in srgb,var(--scalar-color-red),transparent 95%);color:var(--scalar-color-red)}.scalar-modal-layout .scalar-button-danger[data-v-b6a60dbc]:hover,.scalar-modal-layout .scalar-button-danger[data-v-b6a60dbc]:focus{background:color-mix(in srgb,var(--scalar-color-red),transparent 90%)}.fade-request-section-content[data-v-7906217f]{background:linear-gradient(to left,var(--scalar-background-1) 64%,transparent)}.filter-hover[data-v-7906217f]{height:100%;padding-right:42px;padding-left:24px;position:absolute;right:0;transition:width 0s ease-in-out .2s;overflow:hidden}.filter-hover[data-v-7906217f]:hover,.filter-hover[data-v-7906217f]:has(:focus-visible){width:100%;z-index:10}.filter-hover[data-v-7906217f]:before{content:"";position:absolute;top:0;left:0;width:100%;height:100%;background-color:var(--scalar-background-1);opacity:0;transition:all .3s ease-in-out;pointer-events:none}.filter-hover-item[data-v-7906217f]{opacity:0}.filter-hover-item[data-v-7906217f]:not(:last-of-type){transform:translate3d(0,3px,0)}.filter-hover:hover .filter-hover-item[data-v-7906217f]{transition:opacity .2s ease-in-out,transform .2s ease-in-out}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(1){transition-delay:.05s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(2){transition-delay:.1s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(3){transition-delay:.15s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(4){transition-delay:.2s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(5){transition-delay:.25s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(6){transition-delay:.3s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(7){transition-delay:.35s}.filter-hover:hover .filter-hover-item[data-v-7906217f],.filter-hover:has(:focus-visible) .filter-hover-item[data-v-7906217f]{opacity:1;transform:translateZ(0)}.filter-hover[data-v-7906217f]:hover:before,.filter-hover[data-v-7906217f]:has(:focus-visible):before{opacity:.9;-webkit-backdrop-filter:blur(10px);backdrop-filter:blur(10px)}.context-bar-group:hover .context-bar-group-hover\:text-c-1[data-v-7906217f],.context-bar-group:has(:focus-visible) .context-bar-group-hover\:text-c-1[data-v-7906217f]{--tw-text-opacity: 1;color:rgb(var(--scalar-color-1) / var(--tw-text-opacity))}.context-bar-group:hover .context-bar-group-hover\:hidden[data-v-7906217f],.context-bar-group:has(:focus-visible) .context-bar-group-hover\:hidden[data-v-7906217f]{display:none}.light-mode .bg-preview[data-v-5f138d26]{background-image:url("data:image/svg+xml,%3Csvg xmlns=%27http://www.w3.org/2000/svg%27 width=%2716%27 height=%2716%27 fill=%27%23000%27 fill-opacity=%2710%25%27%3E%3Crect width=%278%27 height=%278%27 /%3E%3Crect x=%278%27 y=%278%27 width=%278%27 height=%278%27 /%3E%3C/svg%3E")}.dark-mode .bg-preview[data-v-5f138d26]{background-image:url("data:image/svg+xml,%3Csvg xmlns=%27http://www.w3.org/2000/svg%27 width=%2716%27 height=%2716%27 fill=%27%23FFF%27 fill-opacity=%2710%25%27%3E%3Crect width=%278%27 height=%278%27 /%3E%3Crect x=%278%27 y=%278%27 width=%278%27 height=%278%27 /%3E%3C/svg%3E")}[data-v-99361e9b] .cm-editor{background-color:transparent;font-size:var(--scalar-mini);outline:none}[data-v-99361e9b] .cm-gutters{background-color:var(--scalar-background-1);border-radius:var(--scalar-radius) 0 0 var(--scalar-radius)}.scalar-code-copy[data-v-99361e9b]{align-items:flex-start;display:flex;position:absolute;top:6px;right:6px;z-index:10;pointer-events:none;position:sticky;transform:translate(-6px)}.copy-button[data-v-99361e9b]{align-items:center;display:flex;background-color:var(--scalar-background-1);border:1px solid var(--scalar-border-color);border-radius:3px;color:var(--scalar-color-3);cursor:pointer;height:30px;opacity:0;padding:6px;pointer-events:auto;transition:opacity .15s ease-in-out,color .15s ease-in-out}.body-raw:hover .copy-button[data-v-99361e9b],.copy-button[data-v-99361e9b]:focus-visible{opacity:1}.copy-button[data-v-99361e9b]:hover{color:var(--scalar-color-1)}.scalar-code-block[data-v-cf80aa59] .hljs *{font-size:var(--scalar-mini)}.ascii-art-animate .ascii-art-line[data-v-0018c948]{border-right:1ch solid transparent;animation:typewriter-0018c948 4s steps(1) 1s 1 normal both,blinkTextCursor-0018c948 .5s steps(1) infinite normal}@keyframes typewriter-0018c948{0%{width:0}to{width:100%}}@keyframes blinkTextCursor-0018c948{0%{border-right-color:currentColor}50%{border-right-color:transparent}}.keycap-n[data-v-6c848a44]{background:-webkit-linear-gradient(5deg,transparent 30%,var(--scalar-color-3) 50%);-webkit-background-clip:text;-webkit-text-fill-color:transparent}.keycap-hotkey[data-v-6c848a44]{position:absolute;top:32px;line-height:26px}.scalar-version-number[data-v-04a8f835]{transform:skew(0,13deg);width:76px;height:76px;position:absolute;margin-left:-36px;font-size:8px;font-family:var(--scalar-font-code);line-height:11px;margin-top:-113px;border-radius:9px 9px 16px 12px;box-shadow:inset 2px 0 0 2px var(--scalar-background-2);text-align:center;display:flex;flex-direction:column;align-items:center;justify-content:center;text-transform:initial;text-decoration-color:var(--scalar-color-3)}.scalar-version-number a[data-v-04a8f835]{font-weight:700;background:var(--scalar-background-2);padding:2px 4px;border-radius:3px;text-decoration:none;border:.5px solid var(--scalar-border-color)}.gitbook-show[data-v-04a8f835]{display:none}.v-enter-active[data-v-ef5f83d8]{transition:opacity .5s ease}.v-enter-from[data-v-ef5f83d8]{opacity:0}.response-body-virtual[data-headlessui-state=open],.response-body-virtual[data-headlessui-state=open] .diclosure-panel{display:flex;flex-direction:column;flex-grow:1}.animate-response-heading .response-heading[data-v-ec732740]{animation:push-response-ec732740 .2s ease-in-out forwards;opacity:1}@keyframes push-response-ec732740{0%{opacity:1;transform:translateY(0)}to{opacity:0;transform:translateY(-4px)}}.animate-response-heading .animate-response-children[data-v-ec732740]{animation:response-spans-ec732740 .2s ease-in-out forwards .05s;opacity:0}@keyframes response-spans-ec732740{0%{opacity:0;transform:translateY(4px)}to{opacity:1;transform:translateY(0)}}.schema>span[data-v-30c955c6]:not(:first-child):before{content:"·";display:block;margin:0 .5ch}.schema>span[data-v-30c955c6]{display:flex;white-space:nowrap}[data-v-ccef6266] .cm-editor{padding:0}[data-v-ccef6266] .cm-content{align-items:center;background-color:transparent;display:flex;font-family:var(--scalar-font);font-size:var(--scalar-mini);padding:6px 8px}[data-v-ccef6266] .cm-content:has(.cm-pill){padding:4px 3px}[data-v-ccef6266] .cm-content .cm-pill:not(:last-of-type){margin-right:.5px}[data-v-ccef6266] .cm-content .cm-pill:not(:first-of-type){margin-left:.5px}[data-v-ccef6266] .cm-line{padding:0}.filemask[data-v-ccef6266]{-webkit-mask-image:linear-gradient(to right,transparent 0,var(--scalar-background-2) 20px);mask-image:linear-gradient(to right,transparent 0,var(--scalar-background-2) 20px)}[data-v-3f1cd888] .cm-content{font-size:var(--scalar-mini)}.auth-blend-required[data-v-2b678dcb] .scalar-input-required,.auth-blend-required[data-v-2b678dcb] .required{background:var(--scalar-background-2);--tw-bg-base: var(--scalar-background-2);--tw-shadow: -8px 0 4px var(--scalar-background-2)}.request-example-references-header[data-v-2b678dcb]+tr>td{border-top:0;border-top-left-radius:0;border-top-right-radius:0}.auth-combobox-position[data-v-3569c0b2]{margin-left:120px}.scroll-timeline-x[data-v-3569c0b2]{overflow:auto;scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;scrollbar-width:none;-ms-overflow-style:none}.fade-left[data-v-3569c0b2],.fade-right[data-v-3569c0b2]{position:sticky;content:"";height:100%;animation-name:fadein-3569c0b2;animation-duration:1ms;animation-direction:reverse;animation-timeline:--scroll-timeline;min-height:24px;pointer-events:none}.fade-left[data-v-3569c0b2]{background:linear-gradient(-90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);min-width:3px;left:-1px;animation-direction:normal}.fade-right[data-v-3569c0b2]{background:linear-gradient(90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);margin-left:-20px;min-width:24px;right:-1px;top:0}@keyframes fadein-3569c0b2{0%{opacity:0}15%{opacity:1}}.auth-combobox-position[data-v-3dc33a43]{margin-left:120px}.scroll-timeline-x[data-v-3dc33a43]{overflow:auto;scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;scrollbar-width:none;-ms-overflow-style:none}.fade-left[data-v-3dc33a43],.fade-right[data-v-3dc33a43]{position:sticky;content:"";height:100%;animation-name:fadein-3dc33a43;animation-duration:1ms;animation-direction:reverse;animation-timeline:--scroll-timeline;min-height:24px;pointer-events:none}.fade-left[data-v-3dc33a43]{background:linear-gradient(-90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);min-width:3px;left:-1px;animation-direction:normal}.fade-right[data-v-3dc33a43]{background:linear-gradient(90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);margin-left:-20px;min-width:24px;right:-1px;top:0}@keyframes fadein-3dc33a43{0%{opacity:0}15%{opacity:1}}[data-v-c2f2c62b] code.hljs *{font-size:var(--scalar-mini)}.request-section-content[data-v-5ab558f8]{--scalar-border-width: .5px}.request-section-content-filter[data-v-5ab558f8]{box-shadow:0 -10px 0 10px var(--scalar-background-1)}.request-item:focus-within .request-meta-buttons[data-v-5ab558f8]{opacity:1}.group-hover-input[data-v-5ab558f8]{border-width:var(--scalar-border-width);border-color:transparent}.group:hover .group-hover-input[data-v-5ab558f8]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2));border-color:var(--scalar-border-color)}.group-hover-input[data-v-5ab558f8]:focus{background:transparent!important;border-color:var(--scalar-border-color)!important}@media (min-width: 800px){.has-no-import-url,.has-import-url{max-width:100dvw;overflow-x:hidden;contain:paint}.has-no-import-url{opacity:1;background:var(--scalar-background-1);animation:transform-restore-layout ease-in-out .3s forwards}.has-import-url .scalar-client>main{opacity:0;transform:scale(.85) translate3d(calc(50dvw + 80px),0,0);animation:transform-fade-layout ease-in-out .3s forwards;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:12px;overflow:hidden;z-index:10000}.has-import-url .scalar-client .sidenav{display:none}.has-no-import-url .scalar-app,.has-import-url .scalar-app{background:var(--scalar-background-1)!important}}@keyframes transform-fade-layout{0%{opacity:0;transform:scale(.85) translate3d(calc(50dvw + 80px),10px,0)}to{opacity:1;transform:scale(.85) translate3d(calc(50dvw + 80px),0,0)}}@keyframes transform-restore-layout{0%{opacity:1;transform:scale(.85) translate3d(calc(50dvw + 80px),0,0)}to{opacity:1;transform:scale(1) translateZ(0)}}.openapi-color{color:var(--scalar-color-green)}.section-flare{position:fixed;top:0;right:-50dvw}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}:where(.scalar-app) .container{width:100%}@media (min-width: 400px){:where(.scalar-app) .container{max-width:400px}}@media (min-width: 600px){:where(.scalar-app) .container{max-width:600px}}@media (min-width: 800px){:where(.scalar-app) .container{max-width:800px}}@media (min-width: 1000px){:where(.scalar-app) .container{max-width:1000px}}@media (min-width: 1200px){:where(.scalar-app) .container{max-width:1200px}}:where(.scalar-app) .sr-only{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);white-space:nowrap;border-width:0}:where(.scalar-app) .pointer-events-none{pointer-events:none}:where(.scalar-app) .visible{visibility:visible}:where(.scalar-app) .invisible{visibility:hidden}:where(.scalar-app) .fixed{position:fixed}:where(.scalar-app) .absolute{position:absolute}:where(.scalar-app) .relative{position:relative}:where(.scalar-app) .sticky{position:sticky}:where(.scalar-app) .inset-0{top:0;right:0;bottom:0;left:0}:where(.scalar-app) .left-0{left:0}:where(.scalar-app) .left-2{left:8px}:where(.scalar-app) .left-2\.5{left:10px}:where(.scalar-app) .left-4{left:16px}:where(.scalar-app) .left-border{left:var(--scalar-border-width)}:where(.scalar-app) .left-px{left:1px}:where(.scalar-app) .right-2{right:8px}:where(.scalar-app) .top-0{top:0}:where(.scalar-app) .top-1\/2{top:50%}:where(.scalar-app) .top-12{top:48px}:where(.scalar-app) .top-2{top:8px}:where(.scalar-app) .top-22{top:88px}:where(.scalar-app) .top-\[0\.5px\]{top:.5px}:where(.scalar-app) .top-px{top:1px}:where(.scalar-app) .-z-1{z-index:-1}:where(.scalar-app) .z-1{z-index:1}:where(.scalar-app) .z-\[1001\]{z-index:1001}:where(.scalar-app) .z-context{z-index:1000}:where(.scalar-app) .\!m-0{margin:0!important}:where(.scalar-app) .-m-px{margin:-1px}:where(.scalar-app) .m-0{margin:0}:where(.scalar-app) .m-auto{margin:auto}:where(.scalar-app) .-mx-0{margin-left:-0px;margin-right:-0px}:where(.scalar-app) .-mx-0\.75{margin-left:-3px;margin-right:-3px}:where(.scalar-app) .-mx-px{margin-left:-1px;margin-right:-1px}:where(.scalar-app) .-my-1{margin-top:-4px;margin-bottom:-4px}:where(.scalar-app) .-my-1\.5{margin-top:-6px;margin-bottom:-6px}:where(.scalar-app) .mx-2{margin-left:8px;margin-right:8px}:where(.scalar-app) .mx-auto{margin-left:auto;margin-right:auto}:where(.scalar-app) .mx-px{margin-left:1px;margin-right:1px}:where(.scalar-app) .my-0{margin-top:0;margin-bottom:0}:where(.scalar-app) .my-0\.75{margin-top:3px;margin-bottom:3px}:where(.scalar-app) .my-3{margin-top:12px;margin-bottom:12px}:where(.scalar-app) .-mb-1{margin-bottom:-4px}:where(.scalar-app) .-ml-0{margin-left:-0px}:where(.scalar-app) .-ml-0\.75{margin-left:-3px}:where(.scalar-app) .-mr-0{margin-right:-0px}:where(.scalar-app) .-mr-0\.25{margin-right:-1px}:where(.scalar-app) .mb-0{margin-bottom:0}:where(.scalar-app) .ml-1{margin-left:4px}:where(.scalar-app) .ml-auto{margin-left:auto}:where(.scalar-app) .mr-2{margin-right:8px}:where(.scalar-app) .mr-3{margin-right:12px}:where(.scalar-app) .mt-0{margin-top:0}:where(.scalar-app) .mt-1{margin-top:4px}:where(.scalar-app) .mt-1\.5{margin-top:6px}:where(.scalar-app) .mt-2{margin-top:8px}:where(.scalar-app) .mt-20{margin-top:80px}:where(.scalar-app) .line-clamp-4{overflow:hidden;display:-webkit-box;-webkit-box-orient:vertical;-webkit-line-clamp:4}:where(.scalar-app) .block{display:block}:where(.scalar-app) .inline-block{display:inline-block}:where(.scalar-app) .flex{display:flex}:where(.scalar-app) .table{display:table}:where(.scalar-app) .grid{display:grid}:where(.scalar-app) .contents{display:contents}:where(.scalar-app) .hidden{display:none}:where(.scalar-app) .aspect-square{aspect-ratio:1 / 1}:where(.scalar-app) .size-10{width:40px;height:40px}:where(.scalar-app) .size-2{width:8px;height:8px}:where(.scalar-app) .size-3{width:12px;height:12px}:where(.scalar-app) .size-3\.5{width:14px;height:14px}:where(.scalar-app) .size-4{width:16px;height:16px}:where(.scalar-app) .size-5{width:20px;height:20px}:where(.scalar-app) .size-6{width:24px;height:24px}:where(.scalar-app) .size-8{width:32px;height:32px}:where(.scalar-app) .size-\[23px\]{width:23px;height:23px}:where(.scalar-app) .size-full{width:100%;height:100%}:where(.scalar-app) .h-10{height:40px}:where(.scalar-app) .h-3{height:12px}:where(.scalar-app) .h-3\.5{height:14px}:where(.scalar-app) .h-32{height:128px}:where(.scalar-app) .h-4{height:16px}:where(.scalar-app) .h-5{height:20px}:where(.scalar-app) .h-6{height:24px}:where(.scalar-app) .h-8{height:32px}:where(.scalar-app) .h-\[100dvh\]{height:100dvh}:where(.scalar-app) .h-border{height:var(--scalar-border-width)}:where(.scalar-app) .h-fit{height:-moz-fit-content;height:fit-content}:where(.scalar-app) .h-full{height:100%}:where(.scalar-app) .h-screen{height:100vh}:where(.scalar-app) .max-h-\[440px\]{max-height:440px}:where(.scalar-app) .max-h-\[calc\(100dvh-120px\)\]{max-height:calc(100dvh - 120px)}:where(.scalar-app) .max-h-\[calc\(100dvh-180px\)\]{max-height:calc(100dvh - 180px)}:where(.scalar-app) .max-h-\[calc\(100dvh-240px\)\]{max-height:calc(100dvh - 240px)}:where(.scalar-app) .max-h-\[inherit\]{max-height:inherit}:where(.scalar-app) .max-h-dvh{max-height:100dvh}:where(.scalar-app) .max-h-radix-popper{max-height:calc(var(--radix-popper-available-height) - 8px)}:where(.scalar-app) .min-h-0{min-height:0px}:where(.scalar-app) .min-h-12{min-height:48px}:where(.scalar-app) .min-h-5{min-height:20px}:where(.scalar-app) .min-h-6{min-height:24px}:where(.scalar-app) .min-h-8{min-height:32px}:where(.scalar-app) .min-h-96{min-height:384px}:where(.scalar-app) .min-h-\[77px\]{min-height:77px}:where(.scalar-app) .min-h-header{min-height:48px}:where(.scalar-app) .w-12{width:48px}:where(.scalar-app) .w-16{width:64px}:where(.scalar-app) .w-20{width:80px}:where(.scalar-app) .w-24{width:96px}:where(.scalar-app) .w-3{width:12px}:where(.scalar-app) .w-32{width:128px}:where(.scalar-app) .w-4{width:16px}:where(.scalar-app) .w-40{width:160px}:where(.scalar-app) .w-48{width:192px}:where(.scalar-app) .w-5{width:20px}:where(.scalar-app) .w-56{width:224px}:where(.scalar-app) .w-6{width:24px}:where(.scalar-app) .w-64{width:256px}:where(.scalar-app) .w-72{width:288px}:where(.scalar-app) .w-8{width:32px}:where(.scalar-app) .w-\[100dvw\]{width:100dvw}:where(.scalar-app) .w-\[38px\]{width:38px}:where(.scalar-app) .w-auto{width:auto}:where(.scalar-app) .w-full{width:100%}:where(.scalar-app) .w-min{width:-moz-min-content;width:min-content}:where(.scalar-app) .min-w-0{min-width:0px}:where(.scalar-app) .min-w-5{min-width:20px}:where(.scalar-app) .min-w-6{min-width:24px}:where(.scalar-app) .min-w-8{min-width:32px}:where(.scalar-app) .min-w-max{min-width:-moz-max-content;min-width:max-content}:where(.scalar-app) .min-w-min{min-width:-moz-min-content;min-width:min-content}:where(.scalar-app) .max-w-screen-lg{max-width:800px}:where(.scalar-app) .max-w-screen-md{max-width:640px}:where(.scalar-app) .max-w-screen-sm{max-width:540px}:where(.scalar-app) .max-w-screen-xl{max-width:1000px}:where(.scalar-app) .max-w-screen-xs{max-width:480px}:where(.scalar-app) .max-w-screen-xxs{max-width:360px}:where(.scalar-app) .max-w-xs{max-width:320px}:where(.scalar-app) .flex-1{flex:1 1 0%}:where(.scalar-app) .shrink-0{flex-shrink:0}:where(.scalar-app) .origin-top-left{transform-origin:top left}:where(.scalar-app) .-translate-y-1\/2{--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .translate-x-2{--tw-translate-x: 8px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .translate-x-2\.5{--tw-translate-x: 10px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .translate-x-\[14px\]{--tw-translate-x: 14px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .rotate-90{--tw-rotate: 90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .transform{transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .cursor-auto{cursor:auto}:where(.scalar-app) .cursor-default{cursor:default}:where(.scalar-app) .cursor-not-allowed{cursor:not-allowed}:where(.scalar-app) .cursor-pointer{cursor:pointer}:where(.scalar-app) .resize-none{resize:none}:where(.scalar-app) .resize{resize:both}:where(.scalar-app) .appearance-none{-webkit-appearance:none;-moz-appearance:none;appearance:none}:where(.scalar-app) .grid-cols-\[auto_1fr\]{grid-template-columns:auto 1fr}:where(.scalar-app) .flex-col{flex-direction:column}:where(.scalar-app) .flex-wrap{flex-wrap:wrap}:where(.scalar-app) .items-start{align-items:flex-start}:where(.scalar-app) .items-center{align-items:center}:where(.scalar-app) .justify-start{justify-content:flex-start}:where(.scalar-app) .justify-end{justify-content:flex-end}:where(.scalar-app) .justify-center{justify-content:center}:where(.scalar-app) .justify-between{justify-content:space-between}:where(.scalar-app) .gap-0{gap:0px}:where(.scalar-app) .gap-0\.5{gap:2px}:where(.scalar-app) .gap-0\.75{gap:3px}:where(.scalar-app) .gap-1{gap:4px}:where(.scalar-app) .gap-1\.5{gap:6px}:where(.scalar-app) .gap-16{gap:64px}:where(.scalar-app) .gap-2{gap:8px}:where(.scalar-app) .gap-2\.5{gap:10px}:where(.scalar-app) .gap-3{gap:12px}:where(.scalar-app) .gap-4{gap:16px}:where(.scalar-app) .gap-x-4{-moz-column-gap:16px;column-gap:16px}:where(.scalar-app) .gap-y-8{row-gap:32px}:where(.scalar-app) .divide-y>:not([hidden])~:not([hidden]){--tw-divide-y-reverse: 0;border-top-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-y-reverse)));border-bottom-width:calc(var(--scalar-border-width) * var(--tw-divide-y-reverse))}:where(.scalar-app) .self-center{align-self:center}:where(.scalar-app) .overflow-auto{overflow:auto}:where(.scalar-app) .overflow-hidden{overflow:hidden}:where(.scalar-app) .overflow-visible{overflow:visible}:where(.scalar-app) .overflow-y-scroll{overflow-y:scroll}:where(.scalar-app) .overscroll-contain{overscroll-behavior:contain}:where(.scalar-app) .truncate{overflow:hidden;text-overflow:ellipsis;white-space:nowrap}:where(.scalar-app) .text-ellipsis{text-overflow:ellipsis}:where(.scalar-app) .whitespace-nowrap{white-space:nowrap}:where(.scalar-app) .text-wrap{text-wrap:wrap}:where(.scalar-app) .text-nowrap{text-wrap:nowrap}:where(.scalar-app) .rounded{border-radius:var(--scalar-radius)}:where(.scalar-app) .rounded-full{border-radius:9999px}:where(.scalar-app) .rounded-lg{border-radius:var(--scalar-radius-lg)}:where(.scalar-app) .rounded-md{border-radius:var(--scalar-radius)}:where(.scalar-app) .rounded-none{border-radius:0}:where(.scalar-app) .rounded-xl{border-radius:var(--scalar-radius-xl)}:where(.scalar-app) .rounded-t-none{border-top-left-radius:0;border-top-right-radius:0}:where(.scalar-app) .border{border-width:var(--scalar-border-width)}:where(.scalar-app) .border-0{border-width:0}:where(.scalar-app) .border-2{border-width:calc(var(--scalar-border-width) * 2)}:where(.scalar-app) .border-\[1px\]{border-width:1px}:where(.scalar-app) .border-b{border-bottom-width:var(--scalar-border-width)}:where(.scalar-app) .border-r{border-right-width:var(--scalar-border-width)}:where(.scalar-app) .border-t{border-top-width:var(--scalar-border-width)}:where(.scalar-app) .border-t-1\/2{border-top-width:calc(var(--scalar-border-width) / 2)}:where(.scalar-app) .border-solid{border-style:solid}:where(.scalar-app) .border-dashed{border-style:dashed}:where(.scalar-app) .border-none{border-style:none}:where(.scalar-app) .border-border{border-color:var(--scalar-border-color)}:where(.scalar-app) .border-c-1{border-color:var(--scalar-color-1)}:where(.scalar-app) .border-c-3{border-color:var(--scalar-color-3)}:where(.scalar-app) .border-red{border-color:var(--scalar-color-red)}:where(.scalar-app) .border-sidebar-border{border-color:var(--scalar-sidebar-border-color, var(--scalar-border-color))}:where(.scalar-app) .bg-b-1{background-color:var(--scalar-background-1)}:where(.scalar-app) .bg-b-2{background-color:var(--scalar-background-2)}:where(.scalar-app) .bg-b-3{background-color:var(--scalar-background-3)}:where(.scalar-app) .bg-b-btn{background-color:var(--scalar-button-1)}:where(.scalar-app) .bg-backdrop{background-color:#00000038}:where(.scalar-app) .bg-blue{background-color:var(--scalar-color-blue)}:where(.scalar-app) .bg-border{background-color:var(--scalar-border-color)}:where(.scalar-app) .bg-c-accent{background-color:var(--scalar-color-accent)}:where(.scalar-app) .bg-c-danger{background-color:var(--scalar-color-danger)}:where(.scalar-app) .bg-red{background-color:var(--scalar-color-red)}:where(.scalar-app) .bg-sidebar-b-1{background-color:var(--scalar-sidebar-background-1, var(--scalar-background-1))}:where(.scalar-app) .bg-transparent{background-color:transparent}:where(.scalar-app) .bg-white{--tw-bg-opacity: 1;background-color:rgb(255 255 255 / var(--tw-bg-opacity))}:where(.scalar-app) .p-0{padding:0}:where(.scalar-app) .p-0\.5{padding:2px}:where(.scalar-app) .p-0\.75{padding:3px}:where(.scalar-app) .p-1{padding:4px}:where(.scalar-app) .p-1\.5{padding:6px}:where(.scalar-app) .p-2{padding:8px}:where(.scalar-app) .p-3{padding:12px}:where(.scalar-app) .p-4{padding:16px}:where(.scalar-app) .p-6{padding:24px}:where(.scalar-app) .px-1{padding-left:4px;padding-right:4px}:where(.scalar-app) .px-1\.5{padding-left:6px;padding-right:6px}:where(.scalar-app) .px-2{padding-left:8px;padding-right:8px}:where(.scalar-app) .px-2\.5{padding-left:10px;padding-right:10px}:where(.scalar-app) .px-3{padding-left:12px;padding-right:12px}:where(.scalar-app) .px-4{padding-left:16px;padding-right:16px}:where(.scalar-app) .px-6{padding-left:24px;padding-right:24px}:where(.scalar-app) .px-9{padding-left:36px;padding-right:36px}:where(.scalar-app) .py-1{padding-top:4px;padding-bottom:4px}:where(.scalar-app) .py-1\.5{padding-top:6px;padding-bottom:6px}:where(.scalar-app) .py-2{padding-top:8px;padding-bottom:8px}:where(.scalar-app) .py-2\.5{padding-top:10px;padding-bottom:10px}:where(.scalar-app) .py-3{padding-top:12px;padding-bottom:12px}:where(.scalar-app) .py-4{padding-top:16px;padding-bottom:16px}:where(.scalar-app) .pb-0{padding-bottom:0}:where(.scalar-app) .pl-8{padding-left:32px}:where(.scalar-app) .pl-\[114px\]{padding-left:114px}:where(.scalar-app) .pl-\[24px\]{padding-left:24px}:where(.scalar-app) .pl-\[42px\]{padding-left:42px}:where(.scalar-app) .pl-\[60px\]{padding-left:60px}:where(.scalar-app) .pl-\[6px\]{padding-left:6px}:where(.scalar-app) .pl-\[78px\]{padding-left:78px}:where(.scalar-app) .pl-\[96px\]{padding-left:96px}:where(.scalar-app) .pr-3{padding-right:12px}:where(.scalar-app) .pt-3{padding-top:12px}:where(.scalar-app) .pt-6{padding-top:24px}:where(.scalar-app) .text-left{text-align:left}:where(.scalar-app) .font-code{font-family:var(--scalar-font-code)}:where(.scalar-app) .text-3xs{font-size:var(--scalar-font-size-7)}:where(.scalar-app) .text-lg{font-size:var(--scalar-font-size-2)}:where(.scalar-app) .text-sm{font-size:var(--scalar-font-size-4)}:where(.scalar-app) .text-xl{font-size:var(--scalar-font-size-1)}:where(.scalar-app) .text-xs{font-size:var(--scalar-font-size-5)}:where(.scalar-app) .text-xxs{font-size:var(--scalar-font-size-6)}:where(.scalar-app) .font-bold{font-weight:var(--scalar-bold)}:where(.scalar-app) .font-medium{font-weight:var(--scalar-semibold)}:where(.scalar-app) .font-normal{font-weight:var(--scalar-regular)}:where(.scalar-app) .leading-3{line-height:var(--scalar-line-height-3)}:where(.scalar-app) .leading-none{line-height:1}:where(.scalar-app) .leading-snug{line-height:1.375}:where(.scalar-app) .text-b-1{color:var(--scalar-background-1)}:where(.scalar-app) .text-blue{color:var(--scalar-color-blue)}:where(.scalar-app) .text-c-1{color:var(--scalar-color-1)}:where(.scalar-app) .text-c-2{color:var(--scalar-color-2)}:where(.scalar-app) .text-c-3{color:var(--scalar-color-3)}:where(.scalar-app) .text-c-accent{color:var(--scalar-color-accent)}:where(.scalar-app) .text-c-btn{color:var(--scalar-button-1-color)}:where(.scalar-app) .text-c-danger{color:var(--scalar-color-danger)}:where(.scalar-app) .text-c-ghost{color:var(--scalar-color-ghost)}:where(.scalar-app) .text-red{color:var(--scalar-color-red)}:where(.scalar-app) .text-sidebar-c-2{color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}:where(.scalar-app) .text-transparent{color:transparent}:where(.scalar-app) .text-white{--tw-text-opacity: 1;color:rgb(255 255 255 / var(--tw-text-opacity))}:where(.scalar-app) .no-underline{text-decoration-line:none}:where(.scalar-app) .opacity-0{opacity:0}:where(.scalar-app) .opacity-40{opacity:.4}:where(.scalar-app) .opacity-50{opacity:.5}:where(.scalar-app) .shadow{--tw-shadow: var(--scalar-shadow-1);--tw-shadow-colored: var(--scalar-shadow-1);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-border{--tw-shadow: inset 0 0 0 1px var(--scalar-border-color);--tw-shadow-colored: inset 0 0 0 1px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-lg{--tw-shadow: var(--scalar-shadow-2);--tw-shadow-colored: var(--scalar-shadow-2);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-none{--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-sm{--tw-shadow: rgba(0, 0, 0, .09) 0px 1px 4px;--tw-shadow-colored: 0px 1px 4px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-border{--tw-shadow-color: var(--scalar-border-color);--tw-shadow: var(--tw-shadow-colored)}:where(.scalar-app) .shadow-current{--tw-shadow-color: currentColor;--tw-shadow: var(--tw-shadow-colored)}:where(.scalar-app) .outline-none{outline:2px solid transparent;outline-offset:2px}:where(.scalar-app) .outline{outline-style:solid}:where(.scalar-app) .-outline-offset-1{outline-offset:-1px}:where(.scalar-app) .outline-offset-1{outline-offset:1px}:where(.scalar-app) .blur{--tw-blur: blur(8px);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .brightness-lifted{--tw-brightness: brightness(var(--scalar-lifted-brightness));filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .filter{filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .backdrop-blur{--tw-backdrop-blur: blur(8px);-webkit-backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia);backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia)}:where(.scalar-app) .transition{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,-webkit-backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter,-webkit-backdrop-filter;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .transition-colors{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .transition-opacity{transition-property:opacity;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .transition-transform{transition-property:transform;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .duration-100{transition-duration:.1s}:where(.scalar-app) .duration-300{transition-duration:.3s}:where(.scalar-app) .ease-in-out{transition-timing-function:cubic-bezier(.4,0,.2,1)}:where(.scalar-app) .bg-mix-transparent{background-color:color-mix(var(--tw-bg-mix-method, in srgb),transparent calc(var(--tw-bg-mix-amount, 0) * 1%),var(--tw-bg-base))}:where(.scalar-app) .bg-mix-amount-25{--tw-bg-mix-amount: 25}:where(.scalar-app) .bg-b-1{--tw-bg-base: var(--scalar-background-1)}:where(.scalar-app) .bg-b-2{--tw-bg-base: var(--scalar-background-2)}:where(.scalar-app) .bg-b-3{--tw-bg-base: var(--scalar-background-3)}:where(.scalar-app) .bg-b-btn{--tw-bg-base: var(--scalar-button-1)}:where(.scalar-app) .bg-backdrop{--tw-bg-base: rgba(0, 0, 0, .22)}:where(.scalar-app) .bg-blue{--tw-bg-base: var(--scalar-color-blue)}:where(.scalar-app) .bg-border{--tw-bg-base: var(--scalar-border-color)}:where(.scalar-app) .bg-c-accent{--tw-bg-base: var(--scalar-color-accent)}:where(.scalar-app) .bg-c-danger{--tw-bg-base: var(--scalar-color-danger)}:where(.scalar-app) .bg-red{--tw-bg-base: var(--scalar-color-red)}:where(.scalar-app) .bg-sidebar-b-1{--tw-bg-base: var(--scalar-sidebar-background-1, var(--scalar-background-1))}:where(.scalar-app) .bg-transparent{--tw-bg-base: transparent}:where(.scalar-app) .bg-white{--tw-bg-opacity: 1;--tw-bg-base: rgb(255 255 255 / var(--tw-bg-opacity))}:where(.scalar-app) .row,:where(.scalar-app) .scalar-row{display:flex;flex-direction:row}:where(.scalar-app) .col{display:flex;flex-direction:column}:where(.scalar-app) .centered-x{left:50%;--tw-translate-x: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .centered{left:50%;--tw-translate-x: -50%;top:50%;--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .\*\:flex>*{display:flex}:where(.scalar-app) .\*\:flex-1>*{flex:1 1 0%}:where(.scalar-app) .\*\:items-center>*{align-items:center}:where(.scalar-app) .\*\:gap-px>*{gap:1px}:where(.scalar-app) .hover\:bg-b-2:hover{background-color:var(--scalar-background-2)}:where(.scalar-app) .hover\:bg-b-3:hover{background-color:var(--scalar-background-3)}:where(.scalar-app) .hover\:bg-h-btn:hover{background-color:var(--scalar-button-1-hover)}:where(.scalar-app) .hover\:text-c-1:hover{color:var(--scalar-color-1)}:where(.scalar-app) .hover\:underline:hover{text-decoration-line:underline}:where(.scalar-app) .hover\:bg-b-2:hover{--tw-bg-base: var(--scalar-background-2)}:where(.scalar-app) .hover\:bg-b-3:hover{--tw-bg-base: var(--scalar-background-3)}:where(.scalar-app) .hover\:bg-h-btn:hover{--tw-bg-base: var(--scalar-button-1-hover)}:where(.scalar-app) .focus\:outline-none:focus{outline:2px solid transparent;outline-offset:2px}:where(.scalar-app) .focus-visible\:border-c-btn:focus-visible{border-color:var(--scalar-button-1-color)}:where(.scalar-app) .focus-visible\:ring-1:focus-visible{--tw-ring-offset-shadow: var(--tw-ring-inset) 0 0 0 var(--tw-ring-offset-width) var(--tw-ring-offset-color);--tw-ring-shadow: var(--tw-ring-inset) 0 0 0 calc(1px + var(--tw-ring-offset-width)) var(--tw-ring-color);box-shadow:var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow, 0 0 #0000)}:where(.scalar-app) .focus-visible\:ring-c-accent:focus-visible{--tw-ring-color: var(--scalar-color-accent)}:where(.scalar-app) .active\:bg-b-btn:active{background-color:var(--scalar-button-1)}:where(.scalar-app) .active\:text-c-1:active{color:var(--scalar-color-1)}:where(.scalar-app) .active\:shadow-none:active{--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .active\:brightness-90:active{--tw-brightness: brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .active\:bg-b-btn:active{--tw-bg-base: var(--scalar-button-1)}:where(.scalar-app) .group\/button:hover .group-hover\/button\:text-c-1,:where(.scalar-app) .group:hover .group-hover\:text-c-1{color:var(--scalar-color-1)}:where(.scalar-app) .group\/item:hover .group-hover\/item\:opacity-100{opacity:1}:where(.scalar-app) .group\/toggle:focus-visible .group-focus-visible\/toggle\:outline{outline-style:solid}:where(.scalar-app) .peer:checked~.peer-checked\:text-c-2{color:var(--scalar-color-2)}:where(.scalar-app) .peer:checked~.peer-checked\:opacity-100{opacity:1}:where(.scalar-app) .has-\[\:focus-visible\]\:bg-b-1:has(:focus-visible){background-color:var(--scalar-background-1)}:where(.scalar-app) .has-\[\:focus-visible\]\:outline:has(:focus-visible){outline-style:solid}:where(.scalar-app) .has-\[input\:focus-visible\]\:outline:has(input:focus-visible){outline-style:solid}:where(.scalar-app) .has-\[\:focus-visible\]\:bg-b-1:has(:focus-visible){--tw-bg-base: var(--scalar-background-1)}:where(.scalar-app) .ui-open\:rotate-180[data-headlessui-state~=open]{--tw-rotate: 180deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) :where([data-headlessui-state~=open]) .ui-open\:rotate-180{--tw-rotate: 180deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .has-actv-btn\:border:has(button:active){border-width:var(--scalar-border-width)}:where(.scalar-app) .has-actv-btn\:border-border:has(button:active){border-color:var(--scalar-border-color)}:where(.scalar-app) .\!empty\:flex:not(:empty){display:flex}:where(.scalar-app) .\!empty\:w-7:not(:empty){width:28px}:where(.scalar-app) .\!empty\:items-center:not(:empty){align-items:center}:where(.scalar-app) .\!empty\:pr-3:not(:empty){padding-right:12px}:where(.scalar-app) .hocus\:text-c-1:hover{color:var(--scalar-color-1)}:where(.scalar-app) .hocus\:brightness-90:hover{--tw-brightness: brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .hocus\:text-c-1:focus-visible{color:var(--scalar-color-1)}:where(.scalar-app) .hocus\:brightness-90:focus-visible{--tw-brightness: brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}@media (min-width: 800px){:where(.scalar-app) .md\:w-\[calc\(100vw-16px\)\]{width:calc(100vw - 16px)}:where(.scalar-app) .md\:row{display:flex;flex-direction:row}}@media (min-width: 1000px){:where(.scalar-app) .lg\:w-\[calc\(100vw-32px\)\]{width:calc(100vw - 32px)}:where(.scalar-app) .lg\:w-full{width:100%}}:where(.scalar-app) .dark\:bg-b-3:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-3)}:where(.scalar-app) .dark\:bg-backdropdark:where(.dark-mode,.dark-mode *){background-color:#00000073}:where(.scalar-app) .dark\:bg-b-3:where(.dark-mode,.dark-mode *){--tw-bg-base: var(--scalar-background-3)}:where(.scalar-app) .dark\:bg-backdropdark:where(.dark-mode,.dark-mode *){--tw-bg-base: rgba(0, 0, 0, .45)}:where(.scalar-app) .dark\:hover\:bg-b-3:hover:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-3);--tw-bg-base: var(--scalar-background-3)}.loader-wrapper[data-v-65265e87]{position:relative;display:flex;align-items:center;justify-content:center;--loader-size: 50%}.svg-loader[data-v-65265e87]{width:var(--loader-size);height:var(--loader-size);top:1rem;right:.9rem;overflow:visible;fill:none;background-color:transparent;stroke:currentColor}.svg-path[data-v-65265e87]{stroke-width:14;fill:none;transition:.3s}.svg-x-mark[data-v-65265e87]{stroke-dasharray:57;stroke-dashoffset:57;transition-delay:0s}.svg-check-mark[data-v-65265e87]{stroke-dasharray:149;stroke-dashoffset:149;transition-delay:0s}.icon-is-invalid .svg-x-mark[data-v-65265e87],.icon-is-valid .svg-check-mark[data-v-65265e87]{stroke-dashoffset:0;transition-delay:.3s}.circular-loader[data-v-65265e87]{animation:rotate-65265e87 .7s linear infinite,fade-in-65265e87 .4s;transform-origin:center center;transform:scale(5);background:transparent}.loader-path[data-v-65265e87]{stroke-dasharray:50,200;stroke-dashoffset:-100;stroke-linecap:round}.loader-path-off[data-v-65265e87]{stroke-dasharray:50,200;stroke-dashoffset:-100;transition:opacity .3s;opacity:0}@keyframes fade-in-65265e87{0%{opacity:0}70%{opacity:0}to{opacity:1}}@keyframes rotate-65265e87{0%{transform:scale(5) rotate(0)}to{transform:scale(5) rotate(360deg)}}.scalar-icon[data-v-f4f53d32],.scalar-icon[data-v-f4f53d32] *{stroke-width:var(--ce5c5c8a)}.scalar-code-block{background:inherit;position:relative;overflow:auto;padding:.75rem}.scalar-code-block:hover .copy-button,.copy-button:focus-visible{opacity:100}.scalar-codeblock-pre{all:unset;margin:0;background:transparent;text-wrap:nowrap;white-space-collapse:preserve;border-radius:0;width:-moz-fit-content;width:fit-content}.scalar-code-copy{display:flex;align-items:flex-start;justify-content:flex-end;position:sticky;top:0;right:0;bottom:0;left:0}.copy-button{position:relative;top:0;right:0;display:flex;align-items:center;background-color:var(--scalar-background-2);border:1px solid var(--scalar-border-color);border-radius:3px;color:var(--scalar-color-3);cursor:pointer;height:30px;margin-bottom:-30px;opacity:0;padding:6px;transition:opacity .15s ease-in-out,color .15s ease-in-out}.scalar-code-copy,.copy-button{background:inherit}.copy-button:hover{color:var(--scalar-color-1)}.copy-button svg{stroke-width:1.5}.toggle-icon-ellipse[data-v-8bd002b1]{width:7px;height:7px;border-radius:50%;box-shadow:inset 0 0 0 1px currentColor;overflow:hidden;position:relative;background:var(--scalar-background-1);display:inline-block;transition:width .3s ease-in-out,height .3s ease-in-out}.toggle-icon-moon-mask[data-v-8bd002b1]{width:100%;height:100%;border:1px solid currentColor;display:block;left:2.5px;bottom:2.5px;position:absolute;border-radius:50%;background:var(--scalar-background-1);transition:transform .3s ease-in-out;transform:translate3d(4px,-4px,0)}.toggle-icon-sun-ray[data-v-8bd002b1]{width:12px;height:1px;border-radius:8px;background:currentColor;position:absolute;transition:transform .3s ease-in-out}.toggle-icon-sun-ray[data-v-8bd002b1]:nth-of-type(2){transform:rotate(90deg)}.toggle-icon-sun-ray[data-v-8bd002b1]:nth-of-type(3){transform:rotate(45deg)}.toggle-icon-sun-ray[data-v-8bd002b1]:nth-of-type(4){transform:rotate(-45deg)}.toggle-icon-dark .toggle-icon-ellipse[data-v-8bd002b1]{width:10px;height:10px;-webkit-mask-image:radial-gradient(circle at bottom left,pink 10px,transparent 12px);mask-image:radial-gradient(circle at bottom left,pink 10px,transparent 12px)}.toggle-icon-dark .toggle-icon-sun-ray[data-v-8bd002b1]{transform:scale(0)}.toggle-icon-dark .toggle-icon-moon-mask[data-v-8bd002b1]{transform:translateZ(0)}.scalar-app code.hljs *{font-size:var(--scalar-small);font-family:var(--scalar-font-code);text-align:left;white-space:pre;word-spacing:normal;word-break:normal;word-wrap:normal;line-height:1.4;-moz-tab-size:4;-o-tab-size:4;tab-size:4}.scalar-app code.hljs{all:unset;font-size:var(--scalar-small);color:var(--scalar-color-2);font-family:var(--scalar-font-code);display:inline-block;counter-reset:linenumber}.scalar-app .hljs{background:transparent;color:var(--scalar-color-2)}.scalar-app .hljs .line:before{color:var(--scalar-color-3);display:inline-block;counter-increment:linenumber;content:counter(linenumber);margin-right:.7em;min-width:calc(var(--line-digits) * 1ch);text-align:right}.scalar-app .hljs-comment,.scalar-app .hljs-quote{color:var(--scalar-color-3);font-style:italic}.scalar-app .hljs-number{color:var(--scalar-color-orange)}.scalar-app .hljs-regexp,.scalar-app .hljs-string,.scalar-app .hljs-built_in{color:var(--scalar-color-blue)}.scalar-app .hljs-title.class_{color:var(--scalar-color-1)}.scalar-app .hljs-keyword{color:var(--scalar-color-purple)}.scalar-app .hljs-title.function_{color:var(--scalar-color-orange)}.scalar-app .hljs-subst,.scalar-app .hljs-name{color:var(--scalar-color-blue)}.scalar-app .hljs-attr,.scalar-app .hljs-attribute{color:var(--scalar-color-1)}.scalar-app .hljs-addition,.scalar-app .hljs-literal,.scalar-app .hljs-selector-tag,.scalar-app .hljs-type{color:var(--scalar-color-green)}.scalar-app .hljs-selector-attr,.scalar-app .hljs-selector-pseudo{color:var(--scalar-color-orange)}.scalar-app .hljs-doctag,.scalar-app .hljs-section,.scalar-app .hljs-title{color:var(--scalar-color-blue)}.scalar-app .hljs-selector-id,.scalar-app .hljs-template-variable,.scalar-app .hljs-variable{color:var(--scalar-color-1)}.scalar-app .hljs-name,.scalar-app .hljs-section,.scalar-app .hljs-strong{font-weight:var(--scalar-semibold)}.scalar-app .hljs-bullet,.scalar-app .hljs-link,.scalar-app .hljs-meta,.scalar-app .hljs-symbol{color:var(--scalar-color-blue)}.scalar-app .hljs-deletion{color:var(--scalar-color-red)}.scalar-app .hljs-formula{background:var(--scalar-color-1)}.scalar-app .hljs-emphasis{font-style:italic}.scalar-app .credentials{font-size:0!important;color:transparent}.scalar-app .credentials:after{content:"·····";font-size:var(--scalar-small);color:var(--scalar-color-3);-webkit-user-select:none;-moz-user-select:none;user-select:none}.hljs.language-html{color:var(--scalar-color-1)}.hljs.language-html .hljs-attr{color:var(--scalar-color-2)}.hljs.language-curl .hljs-keyword{color:var(--scalar-color-orange)}.hljs.language-curl .hljs-string{color:var(--scalar-color-1)}.hljs.language-curl .hljs-literal,.hljs.language-php .hljs-variable{color:var(--scalar-color-blue)}.hljs.language-objectivec .hljs-meta{color:var(--scalar-color-1)}.hljs.language-objectivec .hljs-built_in,.hljs-built_in{color:var(--scalar-color-orange)}.scalar-app .markdown{font-family:var(--scalar-font);color:var(--scalar-color-1);word-break:break-word}.scalar-app .markdown h1,.scalar-app .markdown h2,.scalar-app .markdown h3,.scalar-app .markdown h4,.scalar-app .markdown h5,.scalar-app .markdown h6,.scalar-app .markdown p,.scalar-app .markdown div,.scalar-app .markdown img,.scalar-app .markdown details,.scalar-app .markdown summary,.scalar-app .markdown ul,.scalar-app .markdown ol,.scalar-app .markdown table,.scalar-app .markdown blockquote,.scalar-app .markdown code{margin:12px 0}.scalar-app .markdown details{margin:12px 0;color:var(--scalar-color-1)}.scalar-app .markdown summary{display:block;margin:1em 0;padding-left:20px;position:relative;font-weight:var(--scalar-semibold);cursor:pointer;-webkit-user-select:none;-moz-user-select:none;user-select:none}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>");mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>")}.scalar-app .markdown summary:hover:after{background-color:var(--scalar-color-1)}.scalar-app .markdown details[open] summary:after{transform:rotate(90deg)}.scalar-app .markdown summary::-webkit-details-marker{display:none}.scalar-app .markdown img{overflow:hidden;border-radius:var(--scalar-radius);max-width:100%}.scalar-app .markdown>:first-child{margin-top:0;margin-bottom:0}.scalar-app .markdown h1{--font-size: 1.4em}.scalar-app .markdown h2{--font-size: 1.25em}.scalar-app .markdown h3{--font-size: 1.1em}.scalar-app .markdown h4,.scalar-app .markdown h6{--font-size: 1em}.scalar-app .markdown h1,.scalar-app .markdown h2,.scalar-app .markdown h3,.scalar-app .markdown h4,.scalar-app .markdown h5,.scalar-app .markdown h6{font-size:var(--font-size);margin:18px 0 6px;font-weight:var(--scalar-bold);display:block;line-height:1.45}.scalar-app .markdown b,.scalar-app .markdown strong{font-weight:var(--scalar-bold)}.scalar-app .markdown p{color:inherit;font-weight:var(--font-weight, var(--scalar-regular));line-height:1.5;margin-bottom:0;display:block}.scalar-app .markdown ul,.scalar-app .markdown ol{padding-left:24px;line-height:1.5;margin:12px 0;display:block}.scalar-app .markdown ul{list-style:disc}.scalar-app .markdown ol{list-style:decimal}.scalar-app .markdown ul.contains-task-list{list-style:none;padding-left:0}.scalar-app .markdown li{margin:6px 0;display:list-item}.scalar-app .markdown ul:first-of-type li:first-of-type{margin-top:0}.scalar-app .markdown a{color:var(--scalar-link-color, var(--scalar-color-accent));font-weight:var(--scalar-link-font-weight, inherit);-webkit-text-decoration:var(--scalar-text-decoration);text-decoration:var(--scalar-text-decoration);cursor:pointer}.scalar-app .markdown a:hover{color:var(--scalar-link-color-hover, var(--scalar-color-accent));-webkit-text-decoration:var(--scalar-text-decoration-hover);text-decoration:var(--scalar-text-decoration-hover)}.scalar-app .markdown em{font-style:italic}.scalar-app .markdown sup{font-size:var(--scalar-micro);vertical-align:super;font-weight:450}.scalar-app .markdown sub{font-size:var(--scalar-micro);vertical-align:sub;font-weight:450}.scalar-app .markdown del{text-decoration:line-through}.scalar-app .markdown code{font-family:var(--scalar-font-code);background-color:var(--scalar-background-2);box-shadow:0 0 0 var(--scalar-border-width) var(--scalar-border-color);font-size:var(--scalar-micro);border-radius:2px;padding:0 3px}.scalar-app .markdown .hljs{font-size:var(--scalar-small)}.scalar-app .markdown pre code{display:block;white-space:pre;padding:12px;line-height:1.5;margin:12px 0;-webkit-overflow-scrolling:touch;overflow-x:auto;max-width:100%;min-width:100px}.scalar-app .markdown hr{border:none;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.scalar-app .markdown blockquote{border-left:3px solid var(--scalar-border-color);padding-left:12px;margin:0;display:block}.scalar-app .markdown table{display:block;overflow-x:auto;position:relative;border-collapse:collapse;width:-moz-max-content;width:max-content;max-width:100%;margin:1em 0;box-shadow:0 0 0 var(--scalar-border-width) var(--scalar-border-color);border-radius:var(--scalar-radius-lg)}.scalar-app .markdown tbody{display:table-row-group;vertical-align:middle}.scalar-app .markdown thead{display:table-header-group;vertical-align:middle}.scalar-app .markdown tr{display:table-row;border-color:inherit;vertical-align:inherit}.scalar-app .markdown td,.scalar-app .markdown th{display:table-cell;vertical-align:inherit;min-width:1em;padding:6px 9px;vertical-align:top;line-height:1.5;position:relative;word-break:initial;font-size:var(--scalar-small);color:var(--scalar-color-1);font-weight:var(--font-weight, var(--scalar-regular));border-right:var(--scalar-border-width) solid var(--scalar-border-color);border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.scalar-app .markdown td>*,.scalar-app .markdown th>*{margin-bottom:0}.scalar-app .markdown th:empty{display:none}.scalar-app .markdown td:first-of-type,.scalar-app .markdown th:first-of-type{border-left:none}.scalar-app .markdown td:last-of-type,.scalar-app .markdown th:last-of-type{border-right:none}.scalar-app .markdown tr:last-of-type td{border-bottom:none}.scalar-app .markdown th{font-weight:var(--scalar-semibold)!important;text-align:left;border-left-color:transparent;background:var(--scalar-background-2)}.scalar-app .markdown tr>[align=left]{text-align:left}.scalar-app .markdown tr>[align=right]{text-align:right}.scalar-app .markdown tr>[align=center]{text-align:center}.dark-mode .scalar-dropdown-item[data-v-942bce7e]:hover,.dark-mode .scalar-dropdown-item[data-v-457c1cd3]:hover{filter:brightness(1.1)}.scalar-modal-layout[data-v-0f8e0211]{animation:fadein-layout-0f8e0211 ease-in-out .3s forwards}.scalar-modal[data-v-0f8e0211]{animation:fadein-modal-0f8e0211 ease-in-out .3s forwards;animation-delay:.1s;position:fixed;left:0;top:0;right:0;box-shadow:var(--scalar-shadow-2);transform:translate3d(0,10px,0)}.scalar-modal-layout-full[data-v-0f8e0211]{opacity:1!important;background:transparent!important}.dark-mode .scalar-modal[data-v-0f8e0211]{background-color:color-mix(in srgb,var(--scalar-background-1),black)}.scalar-modal.scalar-modal-search[data-v-0f8e0211]{max-width:540px;max-height:440px;background-color:transparent}.modal-content-search .modal-body[data-v-0f8e0211]{padding:0;overflow:hidden;display:flex;flex-direction:column;max-height:440px}@media (max-width: 1280px){.scalar-modal[data-v-0f8e0211]{max-height:calc(100% - 56px);top:28px}}@keyframes fadein-layout-0f8e0211{0%{opacity:0}to{opacity:1}}@keyframes fadein-modal-0f8e0211{0%{opacity:0;transform:translate3d(0,10px,0)}to{opacity:1;transform:translateZ(0)}}.full-size-styles[data-v-0f8e0211]{transform:translateZ(0);animation:fadein-layout-0f8e0211 ease-in-out .3s forwards;max-height:100%!important;top:0!important;left:0;position:absolute!important;margin:initial;border-radius:0!important;background-color:var(--scalar-background-1)!important;box-shadow:none!important;border-right:var(--scalar-border-width) solid var(--scalar-border-color)}@media (min-width: 800px){.full-size-styles[data-v-0f8e0211]{width:50dvw!important}}.full-size-styles[data-v-0f8e0211]:after{content:"";width:50dvw;height:100dvh;position:absolute;right:-50dvw;top:0}.search-background[data-v-2f2316c6]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2))}.search-background[data-v-2f2316c6]:focus-within{background:transparent}.scalar-input-wrapper-focus .scalar-input-label[data-v-4de7a24e]{color:var(--scalar-color-1)}.scalar-input-wrapper-error .scalar-input-label[data-v-4de7a24e]{color:var(--scalar-color-red)}.scalar-input[data-v-4de7a24e]::-moz-selection{color:var(--scalar-color-1);background:#ffa55859}.scalar-input[data-v-4de7a24e]::selection{color:var(--scalar-color-1);background:#ffa55859}.scalar-input[data-v-4de7a24e]:has(+.scalar-input-label){opacity:0}.scalar-input[data-v-4de7a24e]:not(:-moz-placeholder-shown){opacity:1;-moz-transition:opacity .2s ease-in-out .15s;transition:opacity .2s ease-in-out .15s}.scalar-input[data-v-4de7a24e]:not(:placeholder-shown),.scalar-input-wrapper-focus .scalar-input[data-v-4de7a24e]{opacity:1;transition:opacity .2s ease-in-out .15s}.scalar-input[data-v-4de7a24e]:-webkit-autofill,.scalar-input[data-v-4de7a24e]:-webkit-autofill:hover,.scalar-input[data-v-4de7a24e]:-webkit-autofill:focus,.scalar-input[data-v-4de7a24e]:-webkit-autofill:active,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill:hover,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill:focus,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill:active{-webkit-box-shadow:0 0 0px 1000px var(--scalar-background-1) inset!important;-webkit-text-fill-color:var(--scalar-color-1);color:var(--scalar-color-1);border-radius:var(--scalar-radius)}.scalar-input:not(:-moz-placeholder-shown)+.scalar-input-label[data-v-4de7a24e]{transform:translate3d(0,-20px,0) scale(.8);transform-origin:top left}.scalar-input-wrapper-focus .scalar-input-label[data-v-4de7a24e],.scalar-input:not(:placeholder-shown)+.scalar-input-label[data-v-4de7a24e]{transform:translate3d(0,-20px,0) scale(.8);transform-origin:top left}.scalar-input-wrapper-focus:has(button:active) .scalar-input-label[data-v-4de7a24e]{color:var(--scalar-color-3)!important}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}}:where(.scalar-app){font-family:var(--scalar-font);line-height:1.15;-webkit-text-size-adjust:100%;-moz-tab-size:4;-o-tab-size:4;tab-size:4}:where(.scalar-app) *,:where(.scalar-app) *:before,:where(.scalar-app) *:after{box-sizing:border-box;border-width:0;border-style:solid;border-color:var(--scalar-border-color);outline-width:1px;outline-style:none;outline-color:var(--scalar-color-accent);font-family:inherit;font-feature-settings:inherit;font-variation-settings:inherit;font-size:inherit;font-weight:inherit;line-height:inherit;color:inherit;margin:unset;padding:unset;text-rendering:optimizeLegibility;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale}:where(.scalar-app) :before,:where(.scalar-app) :after{--tw-content: ""}:where(.scalar-app) button,:where(.scalar-app) input,:where(.scalar-app) optgroup,:where(.scalar-app) select,:where(.scalar-app) textarea,:where(.scalar-app) ::file-selector-button{background:transparent}:where(.scalar-app) ol,:where(.scalar-app) ul,:where(.scalar-app) menu{list-style:none}:where(.scalar-app) input:where(:not([type=button],[type=reset],[type=submit])),:where(.scalar-app) select,:where(.scalar-app) textarea{border-radius:var(--scalar-radius);border-width:1px}:where(.scalar-app) input::-moz-placeholder{color:var(--scalar-color-3);font-family:var(--scalar-font)}:where(.scalar-app) input::placeholder{color:var(--scalar-color-3);font-family:var(--scalar-font)}:where(.scalar-app) input[type=search]::-webkit-search-cancel-button,:where(.scalar-app) input[type=search]::-webkit-search-decoration{-webkit-appearance:none;-moz-appearance:none;appearance:none}:where(.scalar-app) input:-webkit-autofill{-webkit-background-clip:text!important;background-clip:text!important}:where(.scalar-app) :focus-visible{outline-style:solid;border-radius:var(--scalar-radius)}:where(.scalar-app) button:focus-visible,:where(.scalar-app) [role=button]:focus-visible{outline-offset:-1px}:where(.scalar-app) button,:where(.scalar-app) [role=button]{cursor:pointer}:where(.scalar-app) :disabled{cursor:default}:where(.scalar-app) img,:where(.scalar-app) svg,:where(.scalar-app) video,:where(.scalar-app) canvas,:where(.scalar-app) audio,:where(.scalar-app) iframe,:where(.scalar-app) embed,:where(.scalar-app) object{display:block;vertical-align:middle}:where(.scalar-app) [hidden]{display:none}.scalar-app .cm-scroller,.scalar-app .custom-scroll{overflow-y:auto;scrollbar-color:transparent transparent;scrollbar-width:thin;-webkit-overflow-scrolling:touch}.scalar-app .custom-scroll-self-contain-overflow{overscroll-behavior:contain}.scalar-app .cm-scroller:hover,.scalar-app .custom-scroll:hover,.scalar-app.scalar-scrollbars-obtrusive .cm-scroller,.scalar-app.scalar-scrollbars-obtrusive .custom-scroll{scrollbar-color:var(--scalar-scrollbar-color, transparent) transparent}.scalar-app .cm-scroller:hover::-webkit-scrollbar-thumb,.scalar-app .custom-scroll:hover::-webkit-scrollbar-thumb{background:var(--scalar-scrollbar-color);background-clip:content-box;border:3px solid transparent}.scalar-app .cm-scroller::-webkit-scrollbar-thumb:active,.scalar-app .custom-scroll::-webkit-scrollbar-thumb:active{background:var(--scalar-scrollbar-color-active);background-clip:content-box;border:3px solid transparent}.scalar-app .cm-scroller::-webkit-scrollbar-corner,.scalar-app .custom-scroll::-webkit-scrollbar-corner{background:transparent}.scalar-app .cm-scroller::-webkit-scrollbar,.scalar-app .custom-scroll::-webkit-scrollbar{height:12px;width:12px}.scalar-app .cm-scroller::-webkit-scrollbar-track,.scalar-app .custom-scroll::-webkit-scrollbar-track{background:transparent}.scalar-app .cm-scroller::-webkit-scrollbar-thumb,.scalar-app .custom-scroll::-webkit-scrollbar-thumb{border-radius:20px;background:transparent;background-clip:content-box;border:3px solid transparent}@media (pointer: coarse){.scalar-app .cm-scroller,.scalar-app .custom-scroll{padding-right:12px}}@layer scalar-base{:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-client{line-height:1.5;color:var(--scalar-color-1)}.scalar-client .container{width:100%}@media (min-width: 400px){.scalar-client .container{max-width:400px}}@media (min-width: 600px){.scalar-client .container{max-width:600px}}@media (min-width: 800px){.scalar-client .container{max-width:800px}}@media (min-width: 1000px){.scalar-client .container{max-width:1000px}}@media (min-width: 1200px){.scalar-client .container{max-width:1200px}}.scalar-client .row{display:flex;flex-direction:row}.scalar-client .row-1{display:flex;flex-direction:row;flex:1 1 0%}.scalar-client .col{display:flex;flex-direction:column}.scalar-client .col-1{display:flex;flex-direction:column;flex:1 1 0%}.scalar-client .flex-center{display:flex;justify-content:center;align-items:center}.scalar-client .centered-x{left:50%;--tw-translate-x: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .centered-y{top:50%;--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .centered{left:50%;--tw-translate-x: -50%;top:50%;--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .floating-bg:before{background-color:var(--scalar-background-2);border-radius:var(--scalar-radius);content:"";height:calc(100% - 4px);left:-4px;opacity:0;position:absolute;top:2.5px;transition:opacity .2s ease-in-out;width:calc(100% + 8px);z-index:1}.scalar-client .floating-bg:hover:before{opacity:1}.scalar-client .sr-only{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);white-space:nowrap;border-width:0}.scalar-client .not-sr-only{position:static;width:auto;height:auto;padding:0;margin:0;overflow:visible;clip:auto;white-space:normal}.scalar-client .pointer-events-none{pointer-events:none}.scalar-client .pointer-events-auto{pointer-events:auto}.scalar-client .visible{visibility:visible}.scalar-client .static{position:static}.scalar-client .fixed{position:fixed}.scalar-client .absolute{position:absolute}.scalar-client .relative{position:relative}.scalar-client .sticky{position:sticky}.scalar-client .inset-0{top:0;right:0;bottom:0;left:0}.scalar-client .-inset-x-96{left:-384px;right:-384px}.scalar-client .inset-x-1{left:4px;right:4px}.scalar-client .-right-\[30px\]{right:-30px}.scalar-client .-top-\[--scalar-address-bar-height\]{top:calc(var(--scalar-address-bar-height) * -1)}.scalar-client .bottom-0{bottom:0}.scalar-client .bottom-1\/2{bottom:50%}.scalar-client .bottom-\[var\(--scalar-border-width\)\]{bottom:var(--scalar-border-width)}.scalar-client .left-0{left:0}.scalar-client .left-1\/2{left:50%}.scalar-client .right-0{right:0}.scalar-client .right-1{right:4px}.scalar-client .right-1\/2{right:50%}.scalar-client .right-14{right:56px}.scalar-client .right-16{right:64px}.scalar-client .right-2{right:8px}.scalar-client .right-4{right:16px}.scalar-client .right-7{right:28px}.scalar-client .top-0{top:0}.scalar-client .top-1\/2{top:50%}.scalar-client .top-12{top:48px}.scalar-client .top-2{top:8px}.scalar-client .-z-1{z-index:-1}.scalar-client .z-0{z-index:0}.scalar-client .z-1{z-index:1}.scalar-client .z-10{z-index:10}.scalar-client .z-20{z-index:20}.scalar-client .z-50{z-index:50}.scalar-client .z-\[10001\]{z-index:10001}.scalar-client .z-\[1\]{z-index:1}.scalar-client .z-context{z-index:1000}.scalar-client .z-context-plus{z-index:1001}.scalar-client .z-overlay{z-index:10000}.scalar-client .order-last{order:9999}.scalar-client .col-span-full{grid-column:1 / -1}.scalar-client .\!m-0{margin:0!important}.scalar-client .m-0{margin:0}.scalar-client .m-4{margin:16px}.scalar-client .m-auto{margin:auto}.scalar-client .\!mx-0{margin-left:0!important;margin-right:0!important}.scalar-client .-mx-1{margin-left:-4px;margin-right:-4px}.scalar-client .-my-1{margin-top:-4px;margin-bottom:-4px}.scalar-client .my-1{margin-top:4px;margin-bottom:4px}.scalar-client .my-1\.25{margin-top:5px;margin-bottom:5px}.scalar-client .my-12{margin-top:48px;margin-bottom:48px}.scalar-client .\!mb-0{margin-bottom:0!important}.scalar-client .\!mr-0{margin-right:0!important}.scalar-client .-mb-0{margin-bottom:-0px}.scalar-client .-mb-0\.25{margin-bottom:-1px}.scalar-client .-mb-\[var\(--scalar-border-width\)\]{margin-bottom:calc(var(--scalar-border-width) * -1)}.scalar-client .-ml-0{margin-left:-0px}.scalar-client .-ml-0\.5{margin-left:-2px}.scalar-client .-ml-12{margin-left:-48px}.scalar-client .-mr-1{margin-right:-4px}.scalar-client .-mr-1\.5{margin-right:-6px}.scalar-client .-mr-3{margin-right:-12px}.scalar-client .-mt-0{margin-top:-0px}.scalar-client .-mt-0\.25{margin-top:-1px}.scalar-client .-mt-\[\.5px\]{margin-top:-.5px}.scalar-client .mb-0{margin-bottom:0}.scalar-client .mb-1{margin-bottom:4px}.scalar-client .mb-1\.5{margin-bottom:6px}.scalar-client .mb-2{margin-bottom:8px}.scalar-client .mb-4{margin-bottom:16px}.scalar-client .mb-\[\.5px\]{margin-bottom:.5px}.scalar-client .ml-0{margin-left:0}.scalar-client .ml-0\.5{margin-left:2px}.scalar-client .ml-0\.75{margin-left:3px}.scalar-client .ml-1{margin-left:4px}.scalar-client .ml-2{margin-left:8px}.scalar-client .ml-auto{margin-left:auto}.scalar-client .mr-0{margin-right:0}.scalar-client .mr-0\.25{margin-right:1px}.scalar-client .mr-0\.75{margin-right:3px}.scalar-client .mr-1{margin-right:4px}.scalar-client .mr-1\.5{margin-right:6px}.scalar-client .mr-2{margin-right:8px}.scalar-client .mr-2\.5{margin-right:10px}.scalar-client .mr-3{margin-right:12px}.scalar-client .mr-auto{margin-right:auto}.scalar-client .mt-1{margin-top:4px}.scalar-client .mt-10{margin-top:40px}.scalar-client .mt-2{margin-top:8px}.scalar-client .mt-3{margin-top:12px}.scalar-client .mt-5{margin-top:20px}.scalar-client .mt-auto{margin-top:auto}.scalar-client .line-clamp-1{overflow:hidden;display:-webkit-box;-webkit-box-orient:vertical;-webkit-line-clamp:1}.scalar-client .block{display:block}.scalar-client .inline-block{display:inline-block}.scalar-client .\!flex{display:flex!important}.scalar-client .flex{display:flex}.scalar-client .inline-flex{display:inline-flex}.scalar-client .table{display:table}.scalar-client .grid{display:grid}.scalar-client .contents{display:contents}.scalar-client .\!hidden{display:none!important}.scalar-client .hidden{display:none}.scalar-client .aspect-\[4\/3\]{aspect-ratio:4/3}.scalar-client .aspect-square{aspect-ratio:1 / 1}.scalar-client .size-10{width:40px;height:40px}.scalar-client .size-2{width:8px;height:8px}.scalar-client .size-2\.5{width:10px;height:10px}.scalar-client .size-3{width:12px;height:12px}.scalar-client .size-3\.5{width:14px;height:14px}.scalar-client .size-3\/4{width:75%;height:75%}.scalar-client .size-4{width:16px;height:16px}.scalar-client .size-7{width:28px;height:28px}.scalar-client .h-1{height:4px}.scalar-client .h-1\.5{height:6px}.scalar-client .h-10{height:40px}.scalar-client .h-12{height:48px}.scalar-client .h-2{height:8px}.scalar-client .h-2\.5{height:10px}.scalar-client .h-24{height:96px}.scalar-client .h-3{height:12px}.scalar-client .h-3\.5{height:14px}.scalar-client .h-4{height:16px}.scalar-client .h-5{height:20px}.scalar-client .h-6{height:24px}.scalar-client .h-64{height:256px}.scalar-client .h-7{height:28px}.scalar-client .h-8{height:32px}.scalar-client .h-9{height:36px}.scalar-client .h-\[--scalar-address-bar-height\]{height:var(--scalar-address-bar-height)}.scalar-client .h-\[68px\]{height:68px}.scalar-client .h-\[calc\(100\%-273\.5px\)\]{height:calc(100% - 273.5px)}.scalar-client .h-\[calc\(100\%_-_50px\)\]{height:calc(100% - 50px)}.scalar-client .h-\[var\(--scalar-border-width\)\]{height:var(--scalar-border-width)}.scalar-client .h-auto{height:auto}.scalar-client .h-fit{height:-moz-fit-content;height:fit-content}.scalar-client .h-full{height:100%}.scalar-client .h-header{height:48px}.scalar-client .h-px{height:1px}.scalar-client .h-screen{height:100vh}.scalar-client .\!max-h-\[initial\]{max-height:initial!important}.scalar-client .max-h-40{max-height:160px}.scalar-client .max-h-8{max-height:32px}.scalar-client .max-h-\[40dvh\]{max-height:40dvh}.scalar-client .max-h-\[50dvh\]{max-height:50dvh}.scalar-client .max-h-\[60svh\]{max-height:60svh}.scalar-client .max-h-\[calc\(100\%-32px\)\]{max-height:calc(100% - 32px)}.scalar-client .max-h-\[inherit\]{max-height:inherit}.scalar-client .max-h-fit{max-height:-moz-fit-content;max-height:fit-content}.scalar-client .max-h-screen{max-height:100vh}.scalar-client .min-h-0{min-height:0px}.scalar-client .min-h-10{min-height:40px}.scalar-client .min-h-11{min-height:44px}.scalar-client .min-h-12{min-height:48px}.scalar-client .min-h-16{min-height:64px}.scalar-client .min-h-20{min-height:80px}.scalar-client .min-h-8{min-height:32px}.scalar-client .min-h-fit{min-height:-moz-fit-content;min-height:fit-content}.scalar-client .\!w-fit{width:-moz-fit-content!important;width:fit-content!important}.scalar-client .w-0{width:0px}.scalar-client .w-0\.5{width:2px}.scalar-client .w-1{width:4px}.scalar-client .w-1\.5{width:6px}.scalar-client .w-1\/2{width:50%}.scalar-client .w-10{width:40px}.scalar-client .w-2{width:8px}.scalar-client .w-2\.5{width:10px}.scalar-client .w-20{width:80px}.scalar-client .w-3{width:12px}.scalar-client .w-3\.5{width:14px}.scalar-client .w-4{width:16px}.scalar-client .w-5{width:20px}.scalar-client .w-56{width:224px}.scalar-client .w-6{width:24px}.scalar-client .w-64{width:256px}.scalar-client .w-72{width:288px}.scalar-client .w-8{width:32px}.scalar-client .w-\[60px\]{width:60px}.scalar-client .w-\[calc\(100\%-10px\)\]{width:calc(100% - 10px)}.scalar-client .w-\[calc\(100\%_-_8px\)\]{width:calc(100% - 8px)}.scalar-client .w-\[inherit\]{width:inherit}.scalar-client .w-auto{width:auto}.scalar-client .w-dvw{width:100dvw}.scalar-client .w-fit{width:-moz-fit-content;width:fit-content}.scalar-client .w-full{width:100%}.scalar-client .w-max{width:-moz-max-content;width:max-content}.scalar-client .min-w-0{min-width:0px}.scalar-client .min-w-2{min-width:8px}.scalar-client .min-w-2\.5{min-width:10px}.scalar-client .min-w-3{min-width:12px}.scalar-client .min-w-3\.5{min-width:14px}.scalar-client .min-w-48{min-width:192px}.scalar-client .min-w-8{min-width:32px}.scalar-client .min-w-\[296px\]{min-width:296px}.scalar-client .min-w-\[37px\]{min-width:37px}.scalar-client .min-w-fit{min-width:-moz-fit-content;min-width:fit-content}.scalar-client .min-w-full{min-width:100%}.scalar-client .max-w-\[100\%\]{max-width:100%}.scalar-client .max-w-\[14px\]{max-width:14px}.scalar-client .max-w-\[150px\]{max-width:150px}.scalar-client .max-w-\[320px\]{max-width:320px}.scalar-client .max-w-\[37px\]{max-width:37px}.scalar-client .max-w-\[380px\]{max-width:380px}.scalar-client .max-w-\[420px\]{max-width:420px}.scalar-client .max-w-\[720px\]{max-width:720px}.scalar-client .max-w-\[calc\(100dvw-24px\)\]{max-width:calc(100dvw - 24px)}.scalar-client .max-w-full{max-width:100%}.scalar-client .flex-1{flex:1 1 0%}.scalar-client .shrink-0{flex-shrink:0}.scalar-client .flex-grow{flex-grow:1}.scalar-client .-translate-x-1\/2{--tw-translate-x:-50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .-translate-y-1\/2{--tw-translate-y:-50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .translate-x-0{--tw-translate-x:0px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .translate-x-1\/2{--tw-translate-x:50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .translate-y-1\/2{--tw-translate-y:50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .rotate-90{--tw-rotate:90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .scale-75{--tw-scale-x:.75;--tw-scale-y:.75;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .transform{transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .cursor-auto{cursor:auto}.scalar-client .cursor-default{cursor:default}.scalar-client .cursor-help{cursor:help}.scalar-client .cursor-pointer{cursor:pointer}.scalar-client .cursor-text{cursor:text}.scalar-client .select-none{-webkit-user-select:none;-moz-user-select:none;user-select:none}.scalar-client .resize-none{resize:none}.scalar-client .resize{resize:both}.scalar-client .auto-rows-\[32px\]{grid-auto-rows:32px}.scalar-client .auto-rows-auto{grid-auto-rows:auto}.scalar-client .grid-cols-2{grid-template-columns:repeat(2,minmax(0,1fr))}.scalar-client .grid-cols-\[44px\,1fr\,repeat\(3\,auto\)\]{grid-template-columns:44px 1fr repeat(3,auto)}.scalar-client .grid-cols-\[auto\,1fr\]{grid-template-columns:auto 1fr}.scalar-client .grid-cols-\[repeat\(auto-fill\,minmax\(32px\,1fr\)\)\]{grid-template-columns:repeat(auto-fill,minmax(32px,1fr))}.scalar-client .flex-row{flex-direction:row}.scalar-client .flex-col{flex-direction:column}.scalar-client .flex-wrap{flex-wrap:wrap}.scalar-client .content-start{align-content:flex-start}.scalar-client .content-between{align-content:space-between}.scalar-client .items-start{align-items:flex-start}.scalar-client .items-end{align-items:flex-end}.scalar-client .items-center{align-items:center}.scalar-client .items-stretch{align-items:stretch}.scalar-client .justify-start{justify-content:flex-start}.scalar-client .justify-end{justify-content:flex-end}.scalar-client .justify-center{justify-content:center}.scalar-client .justify-between{justify-content:space-between}.scalar-client .justify-stretch{justify-content:stretch}.scalar-client .\!gap-2{gap:8px!important}.scalar-client .gap-0{gap:0px}.scalar-client .gap-0\.5{gap:2px}.scalar-client .gap-0\.75{gap:3px}.scalar-client .gap-1{gap:4px}.scalar-client .gap-1\.5{gap:6px}.scalar-client .gap-1\/2{gap:.5px}.scalar-client .gap-10{gap:40px}.scalar-client .gap-2{gap:8px}.scalar-client .gap-2\.5{gap:10px}.scalar-client .gap-3{gap:12px}.scalar-client .gap-4{gap:16px}.scalar-client .gap-6{gap:24px}.scalar-client .gap-8{gap:32px}.scalar-client .gap-\[1\.5px\]{gap:1.5px}.scalar-client .gap-px{gap:1px}.scalar-client .gap-x-2{-moz-column-gap:8px;column-gap:8px}.scalar-client .gap-x-2\.5{-moz-column-gap:10px;column-gap:10px}.scalar-client .space-x-1>:not([hidden])~:not([hidden]){--tw-space-x-reverse:0;margin-right:calc(4px * var(--tw-space-x-reverse));margin-left:calc(4px * calc(1 - var(--tw-space-x-reverse)))}.scalar-client .divide-x>:not([hidden])~:not([hidden]){--tw-divide-x-reverse:0;border-right-width:calc(var(--scalar-border-width) * var(--tw-divide-x-reverse));border-left-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-x-reverse)))}.scalar-client .divide-y>:not([hidden])~:not([hidden]){--tw-divide-y-reverse:0;border-top-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-y-reverse)));border-bottom-width:calc(var(--scalar-border-width) * var(--tw-divide-y-reverse))}.scalar-client .self-center{align-self:center}.scalar-client .justify-self-end{justify-self:end}.scalar-client .overflow-auto{overflow:auto}.scalar-client .overflow-hidden{overflow:hidden}.scalar-client .overflow-visible{overflow:visible}.scalar-client .overflow-x-auto{overflow-x:auto}.scalar-client .overflow-y-auto{overflow-y:auto}.scalar-client .truncate{overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.scalar-client .text-ellipsis{text-overflow:ellipsis}.scalar-client .whitespace-nowrap{white-space:nowrap}.scalar-client .whitespace-pre{white-space:pre}.scalar-client .text-balance{text-wrap:balance}.scalar-client .text-pretty{text-wrap:pretty}.scalar-client .break-words{overflow-wrap:break-word}.scalar-client .break-all{word-break:break-all}.scalar-client .\!rounded-none{border-radius:0!important}.scalar-client .rounded{border-radius:var(--scalar-radius)}.scalar-client .rounded-\[10px\]{border-radius:10px}.scalar-client .rounded-full{border-radius:9999px}.scalar-client .rounded-lg{border-radius:var(--scalar-radius-lg)}.scalar-client .rounded-md{border-radius:var(--scalar-radius)}.scalar-client .rounded-px{border-radius:1px}.scalar-client .rounded-xl{border-radius:var(--scalar-radius-xl)}.scalar-client .rounded-b{border-bottom-right-radius:var(--scalar-radius);border-bottom-left-radius:var(--scalar-radius)}.scalar-client .rounded-b-lg{border-bottom-right-radius:var(--scalar-radius-lg);border-bottom-left-radius:var(--scalar-radius-lg)}.scalar-client .rounded-t{border-top-left-radius:var(--scalar-radius);border-top-right-radius:var(--scalar-radius)}.scalar-client .rounded-t-lg{border-top-left-radius:var(--scalar-radius-lg);border-top-right-radius:var(--scalar-radius-lg)}.scalar-client .rounded-t-none{border-top-left-radius:0;border-top-right-radius:0}.scalar-client .\!border-0{border-width:0!important}.scalar-client .border{border-width:var(--scalar-border-width)}.scalar-client .border-0{border-width:0}.scalar-client .border-1\/2{border-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-\[1\.5px\]{border-width:1.5px}.scalar-client .border-\[1px\]{border-width:1px}.scalar-client .\!border-r-1\/2{border-right-width:calc(var(--scalar-border-width) / 2)!important}.scalar-client .border-b{border-bottom-width:var(--scalar-border-width)}.scalar-client .border-b-0{border-bottom-width:0}.scalar-client .border-b-1\/2{border-bottom-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-b-\[1px\]{border-bottom-width:1px}.scalar-client .border-l{border-left-width:var(--scalar-border-width)}.scalar-client .border-l-0{border-left-width:0}.scalar-client .border-r{border-right-width:var(--scalar-border-width)}.scalar-client .border-r-0{border-right-width:0}.scalar-client .border-r-1\/2{border-right-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-t{border-top-width:var(--scalar-border-width)}.scalar-client .border-t-0{border-top-width:0}.scalar-client .border-t-1\/2{border-top-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-solid{border-style:solid}.scalar-client .border-dashed{border-style:dashed}.scalar-client .border-none{border-style:none}.scalar-client .\!border-current{border-color:currentColor!important}.scalar-client .border-b-3{border-color:var(--scalar-background-3)}.scalar-client .border-c-3{border-color:var(--scalar-color-3)}.scalar-client .border-transparent{border-color:transparent}.scalar-client .border-r-transparent{border-right-color:transparent}.scalar-client .bg-b-1{background-color:var(--scalar-background-1)}.scalar-client .bg-b-2{background-color:var(--scalar-background-2)}.scalar-client .bg-b-3{background-color:var(--scalar-background-3)}.scalar-client .bg-blue{background-color:var(--scalar-color-blue)}.scalar-client .bg-border{background-color:var(--scalar-border-color)}.scalar-client .bg-c-2{background-color:var(--scalar-color-2)}.scalar-client .bg-c-3{background-color:var(--scalar-color-3)}.scalar-client .bg-c-accent{background-color:var(--scalar-color-accent)}.scalar-client .bg-current{background-color:currentColor}.scalar-client .bg-green{background-color:var(--scalar-color-green)}.scalar-client .bg-grey{background-color:var(--scalar-color-3)}.scalar-client .bg-orange{background-color:var(--scalar-color-orange)}.scalar-client .bg-purple{background-color:var(--scalar-color-purple)}.scalar-client .bg-red{background-color:var(--scalar-color-red)}.scalar-client .bg-sidebar-active-b{background-color:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .bg-yellow{background-color:var(--scalar-color-yellow)}.scalar-client .bg-none{background-image:none}.scalar-client .fill-current{fill:currentColor}.scalar-client .stroke-2{stroke-width:2}.scalar-client .stroke-\[1\.5\]{stroke-width:1.5}.scalar-client .stroke-\[1\.75\]{stroke-width:1.75}.scalar-client .stroke-\[2\.25\]{stroke-width:2.25}.scalar-client .object-contain{-o-object-fit:contain;object-fit:contain}.scalar-client .\!p-0{padding:0!important}.scalar-client .p-0{padding:0}.scalar-client .p-0\.5{padding:2px}.scalar-client .p-0\.75{padding:3px}.scalar-client .p-1{padding:4px}.scalar-client .p-1\.5{padding:6px}.scalar-client .p-1\.75{padding:7px}.scalar-client .p-2{padding:8px}.scalar-client .p-3{padding:12px}.scalar-client .p-4{padding:16px}.scalar-client .p-\[3px\]{padding:3px}.scalar-client .p-\[5px\]{padding:5px}.scalar-client .\!px-3{padding-left:12px!important;padding-right:12px!important}.scalar-client .\!py-1{padding-top:4px!important;padding-bottom:4px!important}.scalar-client .\!py-1\.5{padding-top:6px!important;padding-bottom:6px!important}.scalar-client .px-0{padding-left:0;padding-right:0}.scalar-client .px-0\.5{padding-left:2px;padding-right:2px}.scalar-client .px-1{padding-left:4px;padding-right:4px}.scalar-client .px-1\.5{padding-left:6px;padding-right:6px}.scalar-client .px-2{padding-left:8px;padding-right:8px}.scalar-client .px-2\.5{padding-left:10px;padding-right:10px}.scalar-client .px-3{padding-left:12px;padding-right:12px}.scalar-client .px-4{padding-left:16px;padding-right:16px}.scalar-client .px-5{padding-left:20px;padding-right:20px}.scalar-client .px-6{padding-left:24px;padding-right:24px}.scalar-client .px-8{padding-left:32px;padding-right:32px}.scalar-client .py-0{padding-top:0;padding-bottom:0}.scalar-client .py-0\.5{padding-top:2px;padding-bottom:2px}.scalar-client .py-0\.75{padding-top:3px;padding-bottom:3px}.scalar-client .py-1{padding-top:4px;padding-bottom:4px}.scalar-client .py-1\.25{padding-top:5px;padding-bottom:5px}.scalar-client .py-1\.5{padding-top:6px;padding-bottom:6px}.scalar-client .py-2{padding-top:8px;padding-bottom:8px}.scalar-client .py-2\.5{padding-top:10px;padding-bottom:10px}.scalar-client .py-3{padding-top:12px;padding-bottom:12px}.scalar-client .py-5{padding-top:20px;padding-bottom:20px}.scalar-client .py-8{padding-top:32px;padding-bottom:32px}.scalar-client .py-px{padding-top:1px;padding-bottom:1px}.scalar-client .\!pl-3{padding-left:12px!important}.scalar-client .\!pt-0{padding-top:0!important}.scalar-client .pb-0{padding-bottom:0}.scalar-client .pb-14{padding-bottom:56px}.scalar-client .pb-2{padding-bottom:8px}.scalar-client .pb-3{padding-bottom:12px}.scalar-client .pb-5{padding-bottom:20px}.scalar-client .pb-8{padding-bottom:32px}.scalar-client .pb-\[75px\]{padding-bottom:75px}.scalar-client .pl-1{padding-left:4px}.scalar-client .pl-1\.25{padding-left:5px}.scalar-client .pl-1\.5{padding-left:6px}.scalar-client .pl-2{padding-left:8px}.scalar-client .pl-3{padding-left:12px}.scalar-client .pl-6{padding-left:24px}.scalar-client .pl-8{padding-left:32px}.scalar-client .pl-9{padding-left:36px}.scalar-client .pl-px{padding-left:1px}.scalar-client .pr-0{padding-right:0}.scalar-client .pr-1{padding-right:4px}.scalar-client .pr-1\.5{padding-right:6px}.scalar-client .pr-2{padding-right:8px}.scalar-client .pr-2\.5{padding-right:10px}.scalar-client .pr-4{padding-right:16px}.scalar-client .pr-6{padding-right:24px}.scalar-client .pr-\[26px\]{padding-right:26px}.scalar-client .pr-px{padding-right:1px}.scalar-client .pt-0{padding-top:0}.scalar-client .pt-2{padding-top:8px}.scalar-client .pt-3{padding-top:12px}.scalar-client .pt-4{padding-top:16px}.scalar-client .pt-px{padding-top:1px}.scalar-client .text-left{text-align:left}.scalar-client .text-center{text-align:center}.scalar-client .text-right{text-align:right}.scalar-client .font-code{font-family:var(--scalar-font-code)}.scalar-client .text-3xs{font-size:var(--scalar-font-size-7)}.scalar-client .text-\[11px\]{font-size:11px}.scalar-client .text-\[21px\]{font-size:21px}.scalar-client .text-\[6px\]{font-size:6px}.scalar-client .text-base{font-size:var(--scalar-font-size-3)}.scalar-client .text-sm{font-size:var(--scalar-font-size-4)}.scalar-client .text-xl{font-size:var(--scalar-font-size-1)}.scalar-client .text-xs{font-size:var(--scalar-font-size-5)}.scalar-client .text-xxs{font-size:var(--scalar-font-size-6)}.scalar-client .\!font-normal{font-weight:var(--scalar-regular)!important}.scalar-client .font-bold{font-weight:var(--scalar-bold)}.scalar-client .font-medium{font-weight:var(--scalar-semibold)}.scalar-client .font-normal{font-weight:var(--scalar-regular)}.scalar-client .uppercase{text-transform:uppercase}.scalar-client .lowercase{text-transform:lowercase}.scalar-client .capitalize{text-transform:capitalize}.scalar-client .\!leading-\[6px\]{line-height:6px!important}.scalar-client .leading{line-height:1.5}.scalar-client .leading-3{line-height:var(--scalar-line-height-3)}.scalar-client .leading-5{line-height:var(--scalar-line-height-5)}.scalar-client .leading-\[1\.44\]{line-height:1.44}.scalar-client .leading-\[21px\]{line-height:21px}.scalar-client .leading-\[7px\]{line-height:7px}.scalar-client .leading-\[normal\]{line-height:normal}.scalar-client .leading-none{line-height:1}.scalar-client .leading-normal{line-height:1.5}.scalar-client .leading-snug{line-height:1.375}.scalar-client .\!text-c-1{color:var(--scalar-color-1)!important}.scalar-client .text-b-1{color:var(--scalar-background-1)}.scalar-client .text-blue{color:var(--scalar-color-blue)}.scalar-client .text-border{color:var(--scalar-border-color)}.scalar-client .text-c-1{color:var(--scalar-color-1)}.scalar-client .text-c-2{color:var(--scalar-color-2)}.scalar-client .text-c-3{color:var(--scalar-color-3)}.scalar-client .text-c-btn{color:var(--scalar-button-1-color)}.scalar-client .text-green{color:var(--scalar-color-green)}.scalar-client .text-grey{color:var(--scalar-color-3)}.scalar-client .text-orange{color:var(--scalar-color-orange)}.scalar-client .text-purple{color:var(--scalar-color-purple)}.scalar-client .text-red{color:var(--scalar-color-red)}.scalar-client .text-sidebar-active-c{color:var(--scalar-sidebar-color-active, currentColor)}.scalar-client .text-sidebar-c-2{color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}.scalar-client .text-transparent{color:transparent}.scalar-client .text-yellow{color:var(--scalar-color-yellow)}.scalar-client .underline{text-decoration-line:underline}.scalar-client .no-underline{text-decoration-line:none}.scalar-client .decoration-c-3{text-decoration-color:var(--scalar-color-3)}.scalar-client .underline-offset-2{text-underline-offset:2px}.scalar-client .opacity-0{opacity:0}.scalar-client .opacity-100{opacity:1}.scalar-client .opacity-25{opacity:.25}.scalar-client .shadow{--tw-shadow:var(--scalar-shadow-1);--tw-shadow-colored:var(--scalar-shadow-1);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-\[-8px_0_4px_var\(--scalar-background-1\)\]{--tw-shadow:-8px 0 4px var(--scalar-background-1);--tw-shadow-colored:-8px 0 4px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-\[0_-8px_0_8px_var\(--scalar-background-1\)\,0_0_8px_8px_var\(--scalar-background-1\)\]{--tw-shadow:0 -8px 0 8px var(--scalar-background-1),0 0 8px 8px var(--scalar-background-1);--tw-shadow-colored:0 -8px 0 8px var(--tw-shadow-color), 0 0 8px 8px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-border{--tw-shadow:inset 0 0 0 1px var(--scalar-border-color);--tw-shadow-colored:inset 0 0 0 1px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-lg{--tw-shadow:var(--scalar-shadow-2);--tw-shadow-colored:var(--scalar-shadow-2);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-none{--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-border{--tw-shadow-color:var(--scalar-border-color);--tw-shadow:var(--tw-shadow-colored)}.scalar-client .outline-none{outline:2px solid transparent;outline-offset:2px}.scalar-client .outline{outline-style:solid}.scalar-client .-outline-offset-1{outline-offset:-1px}.scalar-client .-outline-offset-2{outline-offset:-2px}.scalar-client .outline-offset-2{outline-offset:2px}.scalar-client .blur{--tw-blur:blur(8px);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .brightness-\[\.9\]{--tw-brightness:brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .brightness-lifted{--tw-brightness:brightness(var(--scalar-lifted-brightness));filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .filter{filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .backdrop-filter{-webkit-backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia);backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia)}.scalar-client .transition{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,-webkit-backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter,-webkit-backdrop-filter;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .transition-colors{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .transition-none{transition-property:none}.scalar-client .transition-opacity{transition-property:opacity;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .transition-transform{transition-property:transform;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .duration-150{transition-duration:.15s}.scalar-client .duration-200{transition-duration:.2s}.scalar-client .duration-300{transition-duration:.3s}.scalar-client .ease-in-out{transition-timing-function:cubic-bezier(.4,0,.2,1)}.scalar-client .bg-mix-transparent{background-color:color-mix(var(--tw-bg-mix-method, in srgb),transparent calc(var(--tw-bg-mix-amount, 0) * 1%),var(--tw-bg-base))}.scalar-client .bg-mix-amount-90{--tw-bg-mix-amount:90}.scalar-client .bg-mix-amount-95{--tw-bg-mix-amount:95}.scalar-client .bg-b-1{--tw-bg-base:var(--scalar-background-1)}.scalar-client .bg-b-2{--tw-bg-base:var(--scalar-background-2)}.scalar-client .bg-b-3{--tw-bg-base:var(--scalar-background-3)}.scalar-client .bg-blue{--tw-bg-base:var(--scalar-color-blue)}.scalar-client .bg-border{--tw-bg-base:var(--scalar-border-color)}.scalar-client .bg-c-2{--tw-bg-base:var(--scalar-color-2)}.scalar-client .bg-c-3{--tw-bg-base:var(--scalar-color-3)}.scalar-client .bg-c-accent{--tw-bg-base:var(--scalar-color-accent)}.scalar-client .bg-current{--tw-bg-base:currentColor}.scalar-client .bg-green{--tw-bg-base:var(--scalar-color-green)}.scalar-client .bg-grey{--tw-bg-base:var(--scalar-color-3)}.scalar-client .bg-orange{--tw-bg-base:var(--scalar-color-orange)}.scalar-client .bg-purple{--tw-bg-base:var(--scalar-color-purple)}.scalar-client .bg-red{--tw-bg-base:var(--scalar-color-red)}.scalar-client .bg-sidebar-active-b{--tw-bg-base:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .bg-yellow{--tw-bg-base:var(--scalar-color-yellow)}.scalar-client .app-drag-region{-webkit-app-region:drag}.scalar-client .app-no-drag-region{-webkit-app-region:no-drag}.scalar-client .\[--scalar-address-bar-height\:32px\]{--scalar-address-bar-height:32px}.scalar-client .\[--scalar-small\:--scalar-font-size-4\]{--scalar-small:var(--scalar-font-size-4)}.scalar-client .\*\:flex>*{display:flex}.scalar-client .\*\:h-8>*{height:32px}.scalar-client .\*\:cursor-pointer>*{cursor:pointer}.scalar-client .\*\:items-center>*{align-items:center}.scalar-client .\*\:rounded-none>*{border-radius:0}.scalar-client .\*\:border-t-1\/2>*{border-top-width:calc(var(--scalar-border-width) / 2)}.scalar-client .\*\:px-1>*{padding-left:4px;padding-right:4px}.scalar-client .\*\:px-1\.5>*{padding-left:6px;padding-right:6px}.scalar-client .\*\:pl-4>*{padding-left:16px}.scalar-client .before\:pointer-events-none:before{content:var(--tw-content);pointer-events:none}.scalar-client .before\:absolute:before{content:var(--tw-content);position:absolute}.scalar-client .before\:left-3:before{content:var(--tw-content);left:12px}.scalar-client .before\:left-\[calc\(\.75rem_\+_\.5px\)\]:before{content:var(--tw-content);left:calc(.75rem + .5px)}.scalar-client .before\:right-\[23px\]:before{content:var(--tw-content);right:23px}.scalar-client .before\:top-0:before{content:var(--tw-content);top:0}.scalar-client .before\:top-0\.5:before{content:var(--tw-content);top:2px}.scalar-client .before\:z-1:before{content:var(--tw-content);z-index:1}.scalar-client .before\:h-\[calc\(100\%_\+_\.5px\)\]:before{content:var(--tw-content);height:calc(100% + .5px)}.scalar-client .before\:min-h-\[calc\(100\%-4px\)\]:before{content:var(--tw-content);min-height:calc(100% - 4px)}.scalar-client .before\:w-3:before{content:var(--tw-content);width:12px}.scalar-client .before\:w-\[\.5px\]:before{content:var(--tw-content);width:.5px}.scalar-client .before\:bg-border:before{content:var(--tw-content);background-color:var(--scalar-border-color)}.scalar-client .before\:bg-gradient-to-r:before{content:var(--tw-content);background-image:linear-gradient(to right,var(--tw-gradient-stops))}.scalar-client .before\:from-transparent:before{content:var(--tw-content);--tw-gradient-from:transparent var(--tw-gradient-from-position);--tw-gradient-to:rgb(0 0 0 / 0) var(--tw-gradient-to-position);--tw-gradient-stops:var(--tw-gradient-from), var(--tw-gradient-to)}.scalar-client .before\:to-b-1:before{content:var(--tw-content);--tw-gradient-to:var(--scalar-background-1) var(--tw-gradient-to-position)}.scalar-client .before\:content-\[\'\'\]:before{--tw-content:"";content:var(--tw-content)}.scalar-client .before\:bg-border:before{content:var(--tw-content);--tw-bg-base:var(--scalar-border-color)}.scalar-client .after\:content-\[\'\:\'\]:after{--tw-content:":";content:var(--tw-content)}.scalar-client .first\:border-t:first-child{border-top-width:var(--scalar-border-width)}.scalar-client .first\:\*\:rounded-l>*:first-child{border-top-left-radius:var(--scalar-radius);border-bottom-left-radius:var(--scalar-radius)}.scalar-client .first\:\*\:border-t-0>*:first-child{border-top-width:0}.scalar-client .last\:mb-0:last-child{margin-bottom:0}.scalar-client .last\:border-r-0:last-child{border-right-width:0}.scalar-client .last\:\*\:rounded-r>*:last-child{border-top-right-radius:var(--scalar-radius);border-bottom-right-radius:var(--scalar-radius)}.scalar-client .last\:before\:h-full:last-child:before{content:var(--tw-content);height:100%}.scalar-client .focus-within\:z-20:focus-within{z-index:20}.scalar-client .focus-within\:border-b-3:focus-within{border-color:var(--scalar-background-3)}.scalar-client .focus-within\:bg-b-1:focus-within{background-color:var(--scalar-background-1)}.scalar-client .focus-within\:text-c-1:focus-within{color:var(--scalar-color-1)}.scalar-client .focus-within\:bg-b-1:focus-within{--tw-bg-base:var(--scalar-background-1)}.scalar-client .hover\:cursor-default:hover{cursor:default}.scalar-client .hover\:bg-b-2:hover{background-color:var(--scalar-background-2)}.scalar-client .hover\:bg-b-3:hover{background-color:var(--scalar-background-3)}.scalar-client .hover\:bg-inherit:hover{background-color:inherit}.scalar-client .hover\:bg-sidebar-active-b:hover{background-color:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .hover\:text-c-1:hover{color:var(--scalar-color-1)}.scalar-client .hover\:text-c-2:hover{color:var(--scalar-color-2)}.scalar-client .hover\:bg-b-2:hover{--tw-bg-base:var(--scalar-background-2)}.scalar-client .hover\:bg-b-3:hover{--tw-bg-base:var(--scalar-background-3)}.scalar-client .hover\:bg-inherit:hover{--tw-bg-base:inherit}.scalar-client .hover\:bg-sidebar-active-b:hover{--tw-bg-base:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .focus\:text-c-1:focus{color:var(--scalar-color-1)}.scalar-client .focus\:outline-none:focus{outline:2px solid transparent;outline-offset:2px}.scalar-client .focus-visible\:z-10:focus-visible{z-index:10}.scalar-client .active\:text-c-1:active{color:var(--scalar-color-1)}.scalar-client .disabled\:cursor-default:disabled{cursor:default}.scalar-client .disabled\:text-c-2:disabled{color:var(--scalar-color-2)}.scalar-client .group\/label:last-child .group-last\/label\:rounded-br-lg{border-bottom-right-radius:var(--scalar-radius-lg)}.scalar-client .group:last-child .group-last\:border-b-transparent{border-bottom-color:transparent}.scalar-client .group\/upload:hover .group-hover\/upload\:block,.scalar-client .group:hover .group-hover\:block{display:block}.scalar-client .group\/item:hover .group-hover\/item\:flex,.scalar-client .group:hover .group-hover\:flex{display:flex}.scalar-client .group:hover .group-hover\:hidden{display:none}.scalar-client .group:hover .group-hover\:pr-5{padding-right:20px}.scalar-client .group\/info:hover .group-hover\/info\:text-c-1{color:var(--scalar-color-1)}.scalar-client .group\/scopes-accordion:hover .group-hover\/scopes-accordion\:text-c-2{color:var(--scalar-color-2)}.scalar-client .group:hover .group-hover\:text-c-1{color:var(--scalar-color-1)}.scalar-client .group\/cell:hover .group-hover\/cell\:opacity-100,.scalar-client .group\/params:hover .group-hover\/params\:opacity-100,.scalar-client .group:hover .group-hover\:opacity-100{opacity:1}.scalar-client .group:hover .group-hover\:opacity-80{opacity:.8}.scalar-client .group:focus-visible .group-focus-visible\:opacity-100{opacity:1}.scalar-client .group:focus-visible .group-focus-visible\:outline{outline-style:solid}.scalar-client .group.alert .group-\[\.alert\]\:bg-b-alert{background-color:var(--scalar-background-alert)}.scalar-client .group.alert .group-\[\.alert\]\:bg-transparent{background-color:transparent}.scalar-client .group.alert .group-\[\.alert\]\:shadow-none{--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .group.alert .group-\[\.alert\]\:outline-orange{outline-color:var(--scalar-color-orange)}.scalar-client .group.alert .group-\[\.alert\]\:bg-b-alert{--tw-bg-base:var(--scalar-background-alert)}.scalar-client .group.alert .group-\[\.alert\]\:bg-transparent{--tw-bg-base:transparent}.scalar-client .group.alert .group-\[\.alert\]\:before\:to-b-alert:before{content:var(--tw-content);--tw-gradient-to:var(--scalar-background-alert) var(--tw-gradient-to-position)}.scalar-client .peer:checked~.peer-checked\:text-c-1{color:var(--scalar-color-1)}.scalar-client .has-\[\:focus-visible\]\:rounded-\[4px\]:has(:focus-visible){border-radius:4px}.scalar-client .has-\[\.empty-sidebar-item\]\:border-t-1\/2:has(.empty-sidebar-item){border-top-width:calc(var(--scalar-border-width) / 2)}.scalar-client .has-\[\:focus-visible\]\:opacity-100:has(:focus-visible){opacity:1}.scalar-client .has-\[\:focus-visible\]\:outline:has(:focus-visible){outline-style:solid}.scalar-client .group\/label:has(input) .group-has-\[input\]\/label\:mr-0{margin-right:0}.scalar-client .group\/cell:has(:focus-visible) .group-has-\[\:focus-visible\]\/cell\:border-c-accent{border-color:var(--scalar-color-accent)}.scalar-client .group\/cell:has(:focus-visible) .group-has-\[\:focus-visible\]\/cell\:opacity-100{opacity:1}.scalar-client .group:has(:focus-visible) .group-has-\[\:focus-visible\]\:opacity-100{opacity:1}.scalar-client .peer:has(.color-selector)~.peer-has-\[\.color-selector\]\:hidden{display:none}.scalar-client .peer:has(.cm-focused)~.peer-has-\[\.cm-focused\]\:opacity-0{opacity:0}.scalar-client .peer:has(:focus-visible)~.peer-has-\[\:focus-visible\]\:opacity-0{opacity:0}.scalar-client .ui-open\:rotate-90[data-headlessui-state~=open]{--tw-rotate:90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client :where([data-headlessui-state~=open]) .ui-open\:rotate-90{--tw-rotate:90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .ui-not-open\:invisible[data-headlessui-state]:not([data-headlessui-state~=open]){visibility:hidden}.scalar-client .ui-not-open\:rotate-0[data-headlessui-state]:not([data-headlessui-state~=open]){--tw-rotate:0deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client :where([data-headlessui-state]:not([data-headlessui-state~=open])) .ui-not-open\:invisible:not([data-headlessui-state]){visibility:hidden}.scalar-client :where([data-headlessui-state]:not([data-headlessui-state~=open])) .ui-not-open\:rotate-0:not([data-headlessui-state]){--tw-rotate:0deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .ui-checked\:bg-b-3[data-headlessui-state~=checked]{background-color:var(--scalar-background-3);--tw-bg-base:var(--scalar-background-3)}.scalar-client :where([data-headlessui-state~=checked]) .ui-checked\:bg-b-3{background-color:var(--scalar-background-3);--tw-bg-base:var(--scalar-background-3)}.scalar-client .ui-active\:bg-b-2[data-headlessui-state~=active]{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .\*\:ui-active\:bg-b-2[data-headlessui-state~=active]>*{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client :where([data-headlessui-state~=active]) .ui-active\:bg-b-2{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client :where([data-headlessui-state~=active]) .\*\:ui-active\:bg-b-2>*{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .app-platform-desktop .desktop\:rounded-lg{border-radius:var(--scalar-radius-lg)}.scalar-client .app-platform-mac .mac\:hidden{display:none}.scalar-client .app-platform-mac .mac\:pl-\[72px\]{padding-left:72px}@media (min-width: 600px){.scalar-client .sm\:not-sr-only{position:static;width:auto;height:auto;padding:0;margin:0;overflow:visible;clip:auto;white-space:normal}.scalar-client .sm\:order-none{order:0}.scalar-client .sm\:mb-1{margin-bottom:4px}.scalar-client .sm\:mb-1\.5{margin-bottom:6px}.scalar-client .sm\:ml-1{margin-left:4px}.scalar-client .sm\:ml-1\.5{margin-left:6px}.scalar-client .sm\:mr-1{margin-right:4px}.scalar-client .sm\:mr-1\.5{margin-right:6px}.scalar-client .sm\:flex{display:flex}.scalar-client .sm\:hidden{display:none}.scalar-client .sm\:min-w-max{min-width:-moz-max-content;min-width:max-content}.scalar-client .sm\:max-w-max{max-width:-moz-max-content;max-width:max-content}.scalar-client .sm\:flex-row{flex-direction:row}.scalar-client .sm\:flex-col{flex-direction:column}.scalar-client .sm\:justify-between{justify-content:space-between}.scalar-client .sm\:gap-px{gap:1px}.scalar-client .sm\:rounded{border-radius:var(--scalar-radius)}.scalar-client .sm\:rounded-lg{border-radius:var(--scalar-radius-lg)}.scalar-client .sm\:px-2{padding-left:8px;padding-right:8px}.scalar-client .sm\:px-3{padding-left:12px;padding-right:12px}.scalar-client .sm\:py-1{padding-top:4px;padding-bottom:4px}.scalar-client .sm\:py-1\.5{padding-top:6px;padding-bottom:6px}.scalar-client .sm\:\*\:rounded-lg>*{border-radius:var(--scalar-radius-lg)}}@media (min-width: 800px){.scalar-client .md\:bottom-10{bottom:40px}.scalar-client .md\:right-10{right:40px}.scalar-client .md\:-ml-1{margin-left:-4px}.scalar-client .md\:-ml-1\.25{margin-left:-5px}.scalar-client .md\:block{display:block}.scalar-client .md\:w-full{width:100%}.scalar-client .md\:min-w-fit{min-width:-moz-fit-content;min-width:fit-content}.scalar-client .md\:flex-none{flex:none}.scalar-client .md\:translate-x-0{--tw-translate-x:0px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .md\:translate-y-0{--tw-translate-y:0px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .md\:flex-row{flex-direction:row}.scalar-client .md\:border-b-0{border-bottom-width:0}.scalar-client .md\:border-r-1\/2{border-right-width:calc(var(--scalar-border-width) / 2)}.scalar-client .md\:p-1{padding:4px}.scalar-client .md\:p-1\.5{padding:6px}.scalar-client .md\:px-0{padding-left:0;padding-right:0}.scalar-client .md\:px-1{padding-left:4px;padding-right:4px}.scalar-client .md\:px-1\.5{padding-left:6px;padding-right:6px}.scalar-client .md\:px-2{padding-left:8px;padding-right:8px}.scalar-client .md\:px-2\.5{padding-left:10px;padding-right:10px}.scalar-client .md\:px-4{padding-left:16px;padding-right:16px}.scalar-client .md\:px-\[18px\]{padding-left:18px;padding-right:18px}.scalar-client .md\:py-2{padding-top:8px;padding-bottom:8px}.scalar-client .md\:py-2\.5{padding-top:10px;padding-bottom:10px}.scalar-client .md\:pb-2{padding-bottom:8px}.scalar-client .md\:pb-2\.5{padding-bottom:10px}.scalar-client .md\:pb-\[37px\]{padding-bottom:37px}.scalar-client .md\:\*\:border-t-0>*{border-top-width:0}}@media (min-width: 1000px){.scalar-client .lg\:order-none{order:0}.scalar-client .lg\:-mr-1{margin-right:-4px}.scalar-client .lg\:mb-0{margin-bottom:0}.scalar-client .lg\:flex{display:flex}.scalar-client .lg\:min-h-client-header{min-height:48px}.scalar-client .lg\:w-\[120px\]{width:120px}.scalar-client .lg\:w-auto{width:auto}.scalar-client .lg\:min-w-\[580px\]{min-width:580px}.scalar-client .lg\:max-w-\[580px\]{max-width:580px}.scalar-client .lg\:flex-1{flex:1 1 0%}.scalar-client .lg\:p-1{padding:4px}.scalar-client .lg\:px-1{padding-left:4px;padding-right:4px}.scalar-client .lg\:px-2{padding-left:8px;padding-right:8px}.scalar-client .lg\:px-2\.5{padding-left:10px;padding-right:10px}.scalar-client .lg\:pr-24{padding-right:96px}.scalar-client .lg\:pt-1{padding-top:4px}.scalar-client .lg\:text-sm{font-size:var(--scalar-font-size-4)}}@media (min-width: 1200px){.scalar-client .xl\:\!flex{display:flex!important}.scalar-client .xl\:flex{display:flex}.scalar-client .xl\:hidden{display:none}.scalar-client .xl\:h-full{height:100%}.scalar-client .xl\:min-h-client-header{min-height:48px}.scalar-client .xl\:min-w-0{min-width:0px}.scalar-client .xl\:min-w-\[720px\]{min-width:720px}.scalar-client .xl\:max-w-\[720px\]{max-width:720px}.scalar-client .xl\:flex-row{flex-direction:row}.scalar-client .xl\:overflow-hidden{overflow:hidden}.scalar-client .xl\:rounded-none{border-radius:0}.scalar-client .xl\:pl-2{padding-left:8px}.scalar-client .xl\:pr-0{padding-right:0}.scalar-client .xl\:pr-0\.5{padding-right:2px}.scalar-client .xl\:\*\:border-l-1\/2>*{border-left-width:calc(var(--scalar-border-width) / 2)}.scalar-client .xl\:\*\:border-t-0>*{border-top-width:0}.scalar-client .xl\:first\:\*\:border-l-0>*:first-child{border-left-width:0}}.scalar-client .dark\:bg-b-2:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .dark\:hover\:bg-b-2:hover:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .\[\&\>a\]\:pl-5>a{padding-left:20px}.scalar-client .\[\&\>a\]\:pl-\[1\.625rem\]>a{padding-left:1.625rem}html:has(.dark-mode){--scalar-background-1: #0f0f0f}html,body{background-color:var(--scalar-background-1);overscroll-behavior:none}#scalar-client{display:flex;flex-direction:column;height:100dvh;width:100dvw;position:relative;background-color:var(--scalar-background-2)}.dark-mode #scalar-client{background-color:color-mix(in srgb,var(--scalar-background-1) 65%,black)}.open-api-client-button[data-v-2e9f75f1]{cursor:pointer;width:100%;padding:9px 12px;height:31px;display:block;text-align:center;display:flex;align-items:center;justify-content:center;font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);line-height:1.385;text-decoration:none;border-radius:var(--scalar-radius);box-shadow:0 0 0 .5px var(--scalar-border-color);gap:6px;color:var(--scalar-sidebar-color-1)}.open-api-client-button[data-v-2e9f75f1]:hover{background:var( --scalar-sidebar-item-hover-background, var(--scalar-background-2) )}.addressbar-history-button[data-v-a323ca24]:hover{background:var(--scalar-background-3)}.addressbar-history-button[data-v-a323ca24]:focus-within{background:var(--scalar-background-2)}.addressbar-bg-states:has(.cm-focused) .codemirror-bg-switcher[data-v-a323ca24]{--scalar-background-1: var(--scalar-background-1)}.addressbar-bg-states[data-v-a323ca24]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2))}.addressbar-bg-states[data-v-a323ca24]:has(.cm-focused){background:var(--scalar-background-1);border-color:var(--scalar-border-color);outline:1px solid var(--scalar-color-accent)}.description[data-v-8eba7808] .markdown{font-weight:var(--scalar-semibold);color:var(--scalar-color--1);padding:0;display:block}.description[data-v-8eba7808] .markdown>*:first-child{margin-top:0}[data-v-b4d63e6c] .cm-editor{background-color:var(--scalar-background-1);height:100%;outline:none;width:100%}[data-v-b4d63e6c] .cm-content{padding:0;display:flex;align-items:center;font-size:var(--scalar-mini)}.scroll-timeline-x[data-v-b4d63e6c]{scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;-ms-overflow-style:none}.scroll-timeline-x-hidden[data-v-b4d63e6c]{overflow-x:auto}.scroll-timeline-x-hidden[data-v-b4d63e6c] .cm-scroller{scrollbar-width:none;-ms-overflow-style:none;padding-right:20px;overflow:auto}.scroll-timeline-x-hidden[data-v-b4d63e6c]::-webkit-scrollbar{width:0;height:0;display:none}.scroll-timeline-x-hidden[data-v-b4d63e6c] .cm-scroller::-webkit-scrollbar{width:0;height:0;display:none}.scroll-timeline-x-address[data-v-b4d63e6c]{line-height:27px;scrollbar-width:none}.scroll-timeline-x-address[data-v-b4d63e6c]:after{content:"";position:absolute;height:100%;width:24px;right:0;cursor:text}.scroll-timeline-x-address[data-v-b4d63e6c]:empty:before{content:"Enter URL or cURL request";color:var(--scalar-color-3);pointer-events:none}.fade-left[data-v-b4d63e6c],.fade-right[data-v-b4d63e6c]{content:"";position:sticky;height:100%;animation-name:fadein-b4d63e6c;animation-duration:1ms;animation-direction:reverse;animation-timeline:--scroll-timeline;pointer-events:none}.fade-left[data-v-b4d63e6c]{background:linear-gradient(-90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 30%,var(--scalar-background-1) 100%);left:0;min-width:3px;animation-direction:normal}.fade-right[data-v-b4d63e6c]{background:linear-gradient(90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 30%,var(--scalar-background-1) 100%);right:0;min-width:24px}@keyframes fadein-b4d63e6c{0%{opacity:0}1%{opacity:1}}.codemirror-bg-switcher[data-v-b4d63e6c]{--scalar-background-1: color-mix( in srgb, var(--scalar-background-1), var(--scalar-background-2) )}.addressbar-bg-states:has(.cm-focused) .codemirror-bg-switcher[data-v-b4d63e6c]{--scalar-background-1: var(--scalar-background-1)}.addressbar-bg-states[data-v-b4d63e6c]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2))}.addressbar-bg-states[data-v-b4d63e6c]:has(.cm-focused){background:var(--scalar-background-1);border-color:var(--scalar-border-color);outline:1px solid var(--scalar-color-accent)}.sidebar-height[data-v-5397f2a7]{min-height:100%}@media (min-width: 800px){.sidebar-mask[data-v-5397f2a7]{-webkit-mask-image:linear-gradient(0,transparent 0,transparent 0,var(--scalar-background-2) 30px);mask-image:linear-gradient(0,transparent 0,transparent 0,var(--scalar-background-2) 30px)}}.resizer[data-v-5397f2a7]{width:5px;cursor:col-resize;position:absolute;top:0;right:0;bottom:0;border-right:2px solid transparent;transition:border-right-color .3s}.resizer[data-v-5397f2a7]:hover,.dragging .resizer[data-v-5397f2a7]{border-right-color:var(--scalar-background-3)}.dragging[data-v-5397f2a7]{cursor:col-resize}.dragging[data-v-5397f2a7]:before{content:"";display:block;position:absolute;width:100%;height:100%}.scroll-timeline-x[data-v-b77d5bd3]{overflow:auto;scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;-ms-overflow-style:none;scrollbar-width:none}.gitbook-show[data-v-cac03082]{display:none}.app-exit-button[data-v-cac03082]{color:#fff;background:#0000001a}.app-exit-button[data-v-cac03082]:hover{background:#ffffff1a}.ellipsis-position[data-v-daf8ea6e]{transform:translate3d(calc(-100% - 4.5px),0,0)}.dragover-asChild[data-v-3449bf19],.dragover-above[data-v-3449bf19],.dragover-below[data-v-3449bf19]{position:relative}.dragover-above[data-v-3449bf19]:after,.dragover-below[data-v-3449bf19]:after{content:"";position:absolute;top:-1.5px;height:3px;width:100%;display:block;background:color-mix(in srgb,var(--scalar-color-blue),transparent 85%);pointer-events:none;border-radius:var(--scalar-radius)}.dragover-below[data-v-3449bf19]:after{top:initial;bottom:-1.5px}.dragover-asChild[data-v-3449bf19]:after{content:"";position:absolute;top:0;left:0;width:100%;height:100%;display:block;background:color-mix(in srgb,var(--scalar-color-blue),transparent 85%);pointer-events:none;border-radius:var(--scalar-radius)}.indent-border-line-offset[data-v-682d438e]:before{left:var(--0ff55690)}.indent-padding-left[data-v-682d438e]{padding-left:calc(var(--ecb155a4) + 6px)}.sidebar-folderitem[data-v-682d438e] .ellipsis-position{right:6px;transform:none}.search-button-fade[data-v-b7bf7c07]{background:linear-gradient(var(--scalar-background-1) 32px,color-mix(in srgb,var(--scalar-background-1),transparent) 38px,transparent)}.empty-sidebar-item-content[data-v-b7bf7c07]{display:none}.empty-sidebar-item .empty-sidebar-item-content[data-v-b7bf7c07]{display:block}.rabbitjump[data-v-b7bf7c07]{opacity:0}.empty-sidebar-item:hover .rabbitjump[data-v-b7bf7c07]{opacity:1;animation:rabbitAnimation-b7bf7c07 .5s steps(1) infinite}.empty-sidebar-item:hover .rabbitsit[data-v-b7bf7c07]{opacity:0;animation:rabbitAnimation2-b7bf7c07 .5s steps(1) infinite}.empty-sidebar-item:hover .rabbit-ascii[data-v-b7bf7c07]{animation:rabbitRun-b7bf7c07 8s infinite linear}@keyframes rabbitRun-b7bf7c07{0%{transform:translateZ(0)}25%{transform:translate3d(250px,0,0)}25.01%{transform:translate3d(-250px,0,0)}75%{transform:translate3d(250px,0,0)}75.01%{transform:translate3d(-250px,0,0)}to{transform:translateZ(0)}}@keyframes rabbitAnimation-b7bf7c07{0%,to{opacity:1}50%{opacity:0}}@keyframes rabbitAnimation2-b7bf7c07{0%,to{opacity:0}50%{opacity:1;transform:translate3d(0,-8px,0)}}.request-text-color-text[data-v-02ba9b2d]{color:var(--scalar-color-1);background:linear-gradient(var(--scalar-background-1),var(--scalar-background-3));box-shadow:0 0 0 1px var(--scalar-border-color)}@media screen and (max-width: 800px){.sidebar-active-hide-layout[data-v-02ba9b2d]{display:none}.sidebar-active-width[data-v-02ba9b2d]{width:100%}}.active-link[data-v-a6c1ff1c]{background-color:var(--scalar-background-2);color:var(--scalar-color-1);--tw-bg-base: var(--scalar-background-2)}.empty-variable-name[data-v-a6c1ff1c]:empty:before{content:"Untitled";color:var(--scalar-color-3)}.form-group[data-v-c25da223]{margin-bottom:1rem}.modal-actions[data-v-c25da223]{display:flex;justify-content:flex-end;gap:1rem}:root{--scalar-loaded-api-reference: true}@layer scalar-config{.scalar-api-reference[data-v-10065932]{--refs-sidebar-width: var(--scalar-sidebar-width, 0px);--refs-header-height: calc( var(--scalar-y-offset) + var(--scalar-header-height, 0px) );--refs-content-max-width: var(--scalar-content-max-width, 1540px)}.scalar-api-reference.references-classic[data-v-10065932]{--refs-content-max-width: var(--scalar-content-max-width, 1420px);min-height:100dvh;--refs-sidebar-width: 0}}.references-layout[data-v-10065932]{min-height:100dvh;min-width:100%;max-width:100%;flex:1;--full-height: 100dvh;display:grid;grid-template-rows:var(--scalar-header-height, 0px) repeat(2,auto);grid-template-columns:var(--refs-sidebar-width) 1fr;grid-template-areas:"header header" "navigation rendered" "footer footer";background:var(--scalar-background-1)}.references-header[data-v-10065932]{grid-area:header;position:sticky;top:var(--scalar-custom-header-height, 0px);z-index:10;height:var(--scalar-header-height, 0px)}.references-editor[data-v-10065932]{grid-area:editor;display:flex;min-width:0;background:var(--scalar-background-1);z-index:1}.references-navigation[data-v-10065932]{grid-area:navigation}.references-rendered[data-v-10065932]{position:relative;grid-area:rendered;min-width:0;background:var(--scalar-background-1)}.scalar-api-reference.references-classic[data-v-10065932],.references-classic .references-rendered[data-v-10065932]{--full-height: fit-content !important;height:initial!important;max-height:initial!important}.references-navigation-list[data-v-10065932]{position:sticky;top:var(--refs-header-height);height:calc(100dvh - var(--refs-header-height));background:var(--scalar-sidebar-background-1, var(--scalar-background-1));overflow-y:auto;display:flex;flex-direction:column}.references-editor-textarea[data-v-10065932]{position:sticky;top:var(--refs-header-height);height:calc(var(--full-height) - var(--refs-header-height));display:flex;min-width:0;flex:1}.references-editable[data-v-10065932]{grid-template-columns:var(--refs-sidebar-width) 1fr 1fr;grid-template-areas:"header header header" "navigation editor rendered" "footer footer footer"}@layer scalar-config{.references-sidebar[data-v-10065932]{--refs-sidebar-width: var(--scalar-sidebar-width, 280px)}}.references-footer[data-v-10065932]{grid-area:footer}@media (max-width: 1150px){.references-layout[data-v-10065932]{grid-template-columns:var(--refs-sidebar-width) 1fr 0px}}@media (max-width: 1000px){.references-layout[data-v-10065932]{grid-template-columns:auto;grid-template-rows:var(--scalar-header-height, 0px) 0px auto auto;grid-template-areas:"header" "navigation" "rendered" "footer"}.references-editable[data-v-10065932]{grid-template-areas:"header" "navigation" "editor"}.references-navigation[data-v-10065932],.references-rendered[data-v-10065932]{max-height:unset}.references-rendered[data-v-10065932]{position:static}.references-navigation[data-v-10065932]{display:none;z-index:10}.references-sidebar-mobile-open .references-navigation[data-v-10065932]{display:block;top:var(--refs-header-height);height:calc(100dvh - var(--refs-header-height));width:100%;position:sticky}.references-navigation-list[data-v-10065932]{position:absolute;top:-1px;height:calc(var(--full-height) - var(--refs-header-height) + 1px);width:100%;border-top:1px solid var(--scalar-border-color);display:flex;flex-direction:column}}.references-classic-header[data-v-a860f78f]{display:flex;align-items:center;justify-content:space-between;gap:12px;max-width:var(--refs-content-max-width);margin:auto;padding:12px 0}.references-classic-header-container[data-v-a860f78f]{padding:0 60px}@container narrow-references-container (max-width: 900px){.references-classic-header[data-v-a860f78f]{padding:12px 24px}.references-classic-header-container[data-v-a860f78f]{padding:0}}.references-classic-header-icon[data-v-a860f78f]{height:24px;color:var(--scalar-color-1)}a[data-v-829b4762]{text-decoration:none}.ref-search-container[data-v-829b4762]{display:flex;flex-direction:column;padding:12px 12px 0}.ref-search-results[data-v-829b4762]{padding:12px}.ref-search-meta[data-v-829b4762]{background:var(--scalar-background-3);padding:6px 12px;font-size:var(--scalar-font-size-4);color:var(--scalar-color-3);font-weight:var(--scalar-semibold);display:flex;gap:12px}.deprecated[data-v-829b4762]{text-decoration:line-through}.sidebar-search[data-v-9f0bd6b4]{display:flex;align-items:center;position:relative;padding:0 3px 0 9px;min-width:254px;max-width:100%;font-family:var(--scalar-font);background:var( --scalar-sidebar-search-background, var(--scalar-background-1) );color:var(--scalar-sidebar-color-2, var(--scalar-color-2));border-radius:var(--scalar-radius);box-shadow:0 0 0 .5px var(--scalar-sidebar-search-border-color, var(--scalar-border-color));cursor:pointer;-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none}.sidebar-search-input[data-v-9f0bd6b4]{font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);height:31px;-webkit-user-select:none;-moz-user-select:none;user-select:none;z-index:10;position:relative;display:flex;width:100%;justify-content:space-between;align-items:center}.sidebar-search-key[data-v-9f0bd6b4]{text-transform:uppercase;background-color:var(--scalar-background-2);padding:3px 5px;margin:2px;border-radius:var(--scalar-radius);color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}.scalar-search-icon[data-v-9f0bd6b4]{padding:0;margin-right:6px;width:12px}.references-mobile-header[data-v-b5fb4b8b]{display:none;align-items:center;height:100%;width:100%;padding:0 8px;background:var(--scalar-background-1);border-bottom:1px solid var(--scalar-border-color)}.references-mobile-breadcrumbs[data-v-b5fb4b8b]{flex:1;min-width:0;font-size:var(--scalar-small);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);overflow:hidden;white-space:nowrap;text-overflow:ellipsis}.references-mobile-header-actions[data-v-b5fb4b8b]{display:flex;flex-direction:row;gap:4px;height:24px;align-items:center;padding-left:4px}@media (max-width: 1000px){.references-mobile-header[data-v-b5fb4b8b]{display:flex}}@media (max-width: 1000px){.scalar-api-references-standalone-mobile{--scalar-header-height: 50px}}.scalar-api-references-standalone-search[data-v-3ca2ed96]{display:flex;flex-direction:column;padding:12px 12px 6px}.darklight-reference[data-v-3ca2ed96]{width:100%;margin-top:auto}@layer scalar-base{body{margin:0;background-color:var(--scalar-background-1)}}
/*! tailwindcss v4.0.7 | MIT License | https://tailwindcss.com */
@layer theme{:root,:host{--font-sans:"Lato",ui-sans-serif,system-ui,sans-serif,"Apple Color Emoji","Segoe UI Emoji","Segoe UI Symbol","Noto Color Emoji";--font-serif:"Lato",ui-serif,Georgia,Cambria,"Times New Roman",Times,serif;--font-mono:"Fira Mono",ui-monospace,SFMono-Regular,Menlo,Monaco,Consolas,"Liberation Mono","Courier New",monospace;--spacing:.25rem;--breakpoint-sm:40rem;--breakpoint-md:48rem;--breakpoint-lg:64rem;--breakpoint-xl:80rem;--breakpoint-2xl:96rem;--container-3xs:16rem;--container-2xs:18rem;--container-xs:20rem;--container-sm:24rem;--container-md:28rem;--container-lg:32rem;--container-xl:36rem;--container-2xl:42rem;--container-3xl:48rem;--container-4xl:56rem;--container-5xl:64rem;--container-6xl:72rem;--container-7xl:80rem;--text-xs:.75rem;--text-xs--line-height:calc(1/.75);--text-sm:.875rem;--text-sm--line-height:calc(1.25/.875);--text-base:1rem;--text-base--line-height:calc(1.5/1);--text-lg:1.125rem;--text-lg--line-height:calc(1.75/1.125);--text-xl:1.25rem;--text-xl--line-height:calc(1.75/1.25);--text-2xl:1.5rem;--text-2xl--line-height:calc(2/1.5);--text-3xl:1.875rem;--text-3xl--line-height:calc(2.25/1.875);--text-4xl:2.25rem;--text-4xl--line-height:calc(2.5/2.25);--text-5xl:3rem;--text-5xl--line-height:1;--text-6xl:3.75rem;--text-6xl--line-height:1;--text-7xl:4.5rem;--text-7xl--line-height:1;--text-8xl:6rem;--text-8xl--line-height:1;--text-9xl:8rem;--text-9xl--line-height:1;--font-weight-thin:100;--font-weight-extralight:200;--font-weight-light:300;--font-weight-normal:400;--font-weight-medium:500;--font-weight-semibold:600;--font-weight-bold:700;--font-weight-extrabold:800;--font-weight-black:900;--tracking-tighter:-.05em;--tracking-tight:-.025em;--tracking-normal:0em;--tracking-wide:.025em;--tracking-wider:.05em;--tracking-widest:.1em;--leading-tight:1.25;--leading-snug:1.375;--leading-normal:1.5;--leading-relaxed:1.625;--leading-loose:2;--radius-xs:.125rem;--radius-sm:.25rem;--radius-md:.375rem;--radius-lg:.5rem;--radius-xl:.75rem;--radius-2xl:1rem;--radius-3xl:1.5rem;--radius-4xl:2rem;--shadow-2xs:0 1px #0000000d;--shadow-xs:0 1px 2px 0 #0000000d;--shadow-sm:0 1px 3px 0 #0000001a,0 1px 2px -1px #0000001a;--shadow-md:0 4px 6px -1px #0000001a,0 2px 4px -2px #0000001a;--shadow-lg:0 10px 15px -3px #0000001a,0 4px 6px -4px #0000001a;--shadow-xl:0 20px 25px -5px #0000001a,0 8px 10px -6px #0000001a;--shadow-2xl:0 25px 50px -12px #00000040;--inset-shadow-2xs:inset 0 1px #0000000d;--inset-shadow-xs:inset 0 1px 1px #0000000d;--inset-shadow-sm:inset 0 2px 4px #0000000d;--drop-shadow-xs:0 1px 1px #0000000d;--drop-shadow-sm:0 1px 2px #00000026;--drop-shadow-md:0 3px 3px #0000001f;--drop-shadow-lg:0 4px 4px #00000026;--drop-shadow-xl:0 9px 7px #0000001a;--drop-shadow-2xl:0 25px 25px #00000026;--ease-in:cubic-bezier(.4,0,1,1);--ease-out:cubic-bezier(0,0,.2,1);--ease-in-out:cubic-bezier(.4,0,.2,1);--animate-spin:spin 1s linear infinite;--animate-ping:ping 1s cubic-bezier(0,0,.2,1)infinite;--animate-pulse:pulse 2s cubic-bezier(.4,0,.6,1)infinite;--animate-bounce:bounce 1s infinite;--blur-xs:4px;--blur-sm:8px;--blur-md:12px;--blur-lg:16px;--blur-xl:24px;--blur-2xl:40px;--blur-3xl:64px;--perspective-dramatic:100px;--perspective-near:300px;--perspective-normal:500px;--perspective-midrange:800px;--perspective-distant:1200px;--aspect-video:16/9;--default-transition-duration:.15s;--default-transition-timing-function:cubic-bezier(.4,0,.2,1);--default-font-family:var(--font-sans);--default-font-feature-settings:var(--font-sans--font-feature-settings);--default-font-variation-settings:var(--font-sans--font-variation-settings);--default-mono-font-family:var(--font-mono);--default-mono-font-feature-settings:var(--font-mono--font-feature-settings);--default-mono-font-variation-settings:var(--font-mono--font-variation-settings);--font-alternative:"Montserrat",ui-sans-serif,system-ui,sans-serif,"Apple Color Emoji","Segoe UI Emoji","Segoe UI Symbol","Noto Color Emoji";--color-vortico:teal;--color-vortico-50:#00dcdc;--color-vortico-100:#00d2d2;--color-vortico-200:#00bdbd;--color-vortico-300:#00a9a9;--color-vortico-400:#009494;--color-vortico-500:teal;--color-vortico-600:#007171;--color-vortico-700:#006161;--color-vortico-800:#005252;--color-vortico-900:#004343;--color-vortico-950:#003b3b;--color-bruma:#00bbd5;--color-bruma-50:#ebfeff;--color-bruma-100:#cefcff;--color-bruma-200:#a2f7ff;--color-bruma-300:#63effd;--color-bruma-400:#1cdcf4;--color-bruma-500:#00bbd5;--color-bruma-600:#0398b7;--color-bruma-700:#0a7994;--color-bruma-800:#126178;--color-bruma-900:#145165;--color-bruma-950:#063646;--color-bosque:#9e744f;--color-bosque-50:#f7f4ef;--color-bosque-100:#ebe3d6;--color-bosque-200:#d9c8af;--color-bosque-300:#c2a682;--color-bosque-400:#b0895f;--color-bosque-500:#9e744f;--color-bosque-600:#8a5f44;--color-bosque-700:#6f4939;--color-bosque-800:#5f3f34;--color-bosque-900:#523731;--color-bosque-950:#2f1d19;--color-ciclon:#00976e;--color-ciclon-50:#ebfef6;--color-ciclon-100:#d0fbe7;--color-ciclon-200:#a4f6d3;--color-ciclon-300:#6aebbd;--color-ciclon-400:#2fd8a1;--color-ciclon-500:#0abf8a;--color-ciclon-600:#00976e;--color-ciclon-700:#007c5e;--color-ciclon-800:#03624b;--color-ciclon-900:#04503f;--color-ciclon-950:#012d25;--color-flama:#e25822;--color-flama-50:#fdf5ef;--color-flama-100:#fbe8d9;--color-flama-200:#f6cdb2;--color-flama-300:#f0ac81;--color-flama-400:#e9804e;--color-flama-500:#e25822;--color-flama-600:#d54821;--color-flama-700:#b1351d;--color-flama-800:#8d2d1f;--color-flama-900:#72271c;--color-flama-950:#3d110d;--color-primary:#656775;--color-primary-50:#f5f5f6;--color-primary-100:#e5e6e8;--color-primary-200:#cdced4;--color-primary-300:#aaabb6;--color-primary-400:#808290;--color-primary-500:#656775;--color-primary-600:#575863;--color-primary-700:#4a4a54;--color-primary-800:#3f3f46;--color-primary-900:#3a3a3f;--color-primary-950:#242428;--breakpoint-3xl:120rem;--container-8xl:96rem;--container-9xl:120rem}}@layer base{*,:after,:before,::backdrop{box-sizing:border-box;border:0 solid;margin:0;padding:0}::file-selector-button{box-sizing:border-box;border:0 solid;margin:0;padding:0}html,:host{-webkit-text-size-adjust:100%;tab-size:4;line-height:1.5;font-family:var(--default-font-family,ui-sans-serif,system-ui,sans-serif,"Apple Color Emoji","Segoe UI Emoji","Segoe UI Symbol","Noto Color Emoji");font-feature-settings:var(--default-font-feature-settings,normal);font-variation-settings:var(--default-font-variation-settings,normal);-webkit-tap-highlight-color:transparent}body{line-height:inherit}hr{height:0;color:inherit;border-top-width:1px}abbr:where([title]){-webkit-text-decoration:underline dotted;text-decoration:underline dotted}h1,h2,h3,h4,h5,h6{font-size:inherit;font-weight:inherit}a{color:inherit;-webkit-text-decoration:inherit;-webkit-text-decoration:inherit;-webkit-text-decoration:inherit;text-decoration:inherit}b,strong{font-weight:bolder}code,kbd,samp,pre{font-family:var(--default-mono-font-family,ui-monospace,SFMono-Regular,Menlo,Monaco,Consolas,"Liberation Mono","Courier New",monospace);font-feature-settings:var(--default-mono-font-feature-settings,normal);font-variation-settings:var(--default-mono-font-variation-settings,normal);font-size:1em}small{font-size:80%}sub,sup{vertical-align:baseline;font-size:75%;line-height:0;position:relative}sub{bottom:-.25em}sup{top:-.5em}table{text-indent:0;border-color:inherit;border-collapse:collapse}:-moz-focusring{outline:auto}progress{vertical-align:baseline}summary{display:list-item}ol,ul,menu{list-style:none}img,svg,video,canvas,audio,iframe,embed,object{vertical-align:middle;display:block}img,video{max-width:100%;height:auto}button,input,select,optgroup,textarea{font:inherit;font-feature-settings:inherit;font-variation-settings:inherit;letter-spacing:inherit;color:inherit;opacity:1;background-color:#0000;border-radius:0}::file-selector-button{font:inherit;font-feature-settings:inherit;font-variation-settings:inherit;letter-spacing:inherit;color:inherit;opacity:1;background-color:#0000;border-radius:0}:where(select:is([multiple],[size])) optgroup{font-weight:bolder}:where(select:is([multiple],[size])) optgroup option{padding-inline-start:20px}::file-selector-button{margin-inline-end:4px}::placeholder{opacity:1;color:color-mix(in oklab,currentColor 50%,transparent)}textarea{resize:vertical}::-webkit-search-decoration{-webkit-appearance:none}::-webkit-date-and-time-value{min-height:1lh;text-align:inherit}::-webkit-datetime-edit{display:inline-flex}::-webkit-datetime-edit-fields-wrapper{padding:0}::-webkit-datetime-edit{padding-block:0}::-webkit-datetime-edit-year-field{padding-block:0}::-webkit-datetime-edit-month-field{padding-block:0}::-webkit-datetime-edit-day-field{padding-block:0}::-webkit-datetime-edit-hour-field{padding-block:0}::-webkit-datetime-edit-minute-field{padding-block:0}::-webkit-datetime-edit-second-field{padding-block:0}::-webkit-datetime-edit-millisecond-field{padding-block:0}::-webkit-datetime-edit-meridiem-field{padding-block:0}:-moz-ui-invalid{box-shadow:none}button,input:where([type=button],[type=reset],[type=submit]){appearance:button}::file-selector-button{appearance:button}::-webkit-inner-spin-button{height:auto}::-webkit-outer-spin-button{height:auto}[hidden]:where(:not([hidden=until-found])){display:none!important}}@layer components;@layer utilities{.absolute{position:absolute}.fixed{position:fixed}.relative{position:relative}.inset-0{inset:calc(var(--spacing)*0)}.inset-x-0{inset-inline:calc(var(--spacing)*0)}.inset-x-4{inset-inline:calc(var(--spacing)*4)}.inset-y-\[5vh\]{inset-block:5vh}.top-0{top:calc(var(--spacing)*0)}.top-4{top:calc(var(--spacing)*4)}.right-4{right:calc(var(--spacing)*4)}.-bottom-2{bottom:calc(var(--spacing)*-2)}.left-1\/2{left:50%}.z-10{z-index:10}.z-20{z-index:20}.z-50{z-index:50}.mx-auto{margin-inline:auto}.my-14{margin-block:calc(var(--spacing)*14)}.-mt-px{margin-top:-1px}.mt-\[calc\(3\.5rem\+1px\)\]{margin-top:calc(3.5rem + 1px)}.mt-px{margin-top:1px}.mb-px{margin-bottom:1px}.-ml-px{margin-left:-1px}.ml-1{margin-left:calc(var(--spacing)*1)}.block{display:block}.flex{display:flex}.hidden{display:none}.inline{display:inline}.table{display:table}.h-2{height:calc(var(--spacing)*2)}.h-4{height:calc(var(--spacing)*4)}.h-5{height:calc(var(--spacing)*5)}.h-6{height:calc(var(--spacing)*6)}.h-7{height:calc(var(--spacing)*7)}.h-8{height:calc(var(--spacing)*8)}.h-9{height:calc(var(--spacing)*9)}.h-12{height:calc(var(--spacing)*12)}.h-14{height:calc(var(--spacing)*14)}.h-\[672px\]{height:672px}.h-\[calc\(100vh-6\.5rem-2px\)\]{height:calc(100vh - 6.5rem - 2px)}.h-fit{height:fit-content}.h-full{height:100%}.h-screen{height:100vh}.max-h-0{max-height:calc(var(--spacing)*0)}.max-h-\[32px\]{max-height:32px}.max-h-\[calc\(85vh\)\]{max-height:85vh}.max-h-\[calc\(85vh\+2rem\+3px\)\]{max-height:calc(85vh + 2rem + 3px)}.max-h-\[calc\(100\%-2rem-1px\)\]{max-height:calc(100% - 2rem - 1px)}.max-h-full{max-height:100%}.min-h-full{min-height:100%}.w-2{width:calc(var(--spacing)*2)}.w-4{width:calc(var(--spacing)*4)}.w-5{width:calc(var(--spacing)*5)}.w-6{width:calc(var(--spacing)*6)}.w-7{width:calc(var(--spacing)*7)}.w-8{width:calc(var(--spacing)*8)}.w-9{width:calc(var(--spacing)*9)}.w-40{width:calc(var(--spacing)*40)}.w-fit{width:fit-content}.w-full{width:100%}.w-screen{width:100vw}.max-w-7xl{max-width:var(--container-7xl)}.max-w-8xl{max-width:var(--container-8xl)}.flex-auto{flex:auto}.flex-none{flex:none}.table-fixed{table-layout:fixed}.border-collapse{border-collapse:collapse}.-translate-x-1\/2{--tw-translate-x:calc(calc(1/2*100%)*-1);translate:var(--tw-translate-x)var(--tw-translate-y)}.translate-y-full{--tw-translate-y:100%;translate:var(--tw-translate-x)var(--tw-translate-y)}.rotate-90{rotate:90deg}.cursor-pointer{cursor:pointer}.flex-col{flex-direction:column}.items-center{align-items:center}.items-start{align-items:flex-start}.justify-between{justify-content:space-between}.justify-center{justify-content:center}.justify-end{justify-content:flex-end}.justify-start{justify-content:flex-start}.gap-1{gap:calc(var(--spacing)*1)}.gap-2{gap:calc(var(--spacing)*2)}.gap-8{gap:calc(var(--spacing)*8)}.gap-x-2{column-gap:calc(var(--spacing)*2)}:where(.divide-x>:not(:last-child)){--tw-divide-x-reverse:0;border-inline-style:var(--tw-border-style);border-inline-start-width:calc(1px*var(--tw-divide-x-reverse));border-inline-end-width:calc(1px*calc(1 - var(--tw-divide-x-reverse)))}:where(.divide-y>:not(:last-child)){--tw-divide-y-reverse:0;border-bottom-style:var(--tw-border-style);border-top-style:var(--tw-border-style);border-top-width:calc(1px*var(--tw-divide-y-reverse));border-bottom-width:calc(1px*calc(1 - var(--tw-divide-y-reverse)))}:where(.divide-flama-500\/50>:not(:last-child)){border-color:color-mix(in oklab,var(--color-flama-500)50%,transparent)}:where(.divide-vortico-500\/50>:not(:last-child)){border-color:color-mix(in oklab,var(--color-vortico-500)50%,transparent)}.truncate{text-overflow:ellipsis;white-space:nowrap;overflow:hidden}.overflow-auto{overflow:auto}.overflow-hidden{overflow:hidden}.overflow-visible{overflow:visible}.overflow-y-scroll{overflow-y:scroll}.rounded{border-radius:.25rem}.rounded-lg{border-radius:var(--radius-lg)}.border{border-style:var(--tw-border-style);border-width:1px}.border-t{border-top-style:var(--tw-border-style);border-top-width:1px}.border-r{border-right-style:var(--tw-border-style);border-right-width:1px}.border-b{border-bottom-style:var(--tw-border-style);border-bottom-width:1px}.border-l-4{border-left-style:var(--tw-border-style);border-left-width:4px}.border-current{border-color:currentColor}.border-flama-500{border-color:var(--color-flama-500)}.border-flama-500\/50{border-color:color-mix(in oklab,var(--color-flama-500)50%,transparent)}.border-primary-100{border-color:var(--color-primary-100)}.border-primary-300{border-color:var(--color-primary-300)}.border-primary-500{border-color:var(--color-primary-500)}.border-transparent{border-color:#0000}.border-vortico-500\/50{border-color:color-mix(in oklab,var(--color-vortico-500)50%,transparent)}.bg-flama-500\/25{background-color:color-mix(in oklab,var(--color-flama-500)25%,transparent)}.bg-flama-700{background-color:var(--color-flama-700)}.bg-primary-100{background-color:var(--color-primary-100)}.bg-primary-200{background-color:var(--color-primary-200)}.bg-primary-500\/30{background-color:color-mix(in oklab,var(--color-primary-500)30%,transparent)}.bg-primary-800{background-color:var(--color-primary-800)}.bg-primary-900{background-color:var(--color-primary-900)}.bg-primary-950\/30{background-color:color-mix(in oklab,var(--color-primary-950)30%,transparent)}.bg-linear-to-b{--tw-gradient-position:to bottom in oklab;background-image:linear-gradient(var(--tw-gradient-stops))}.from-flama-500\/10{--tw-gradient-from:color-mix(in oklab,var(--color-flama-500)10%,transparent);--tw-gradient-stops:var(--tw-gradient-via-stops,var(--tw-gradient-position),var(--tw-gradient-from)var(--tw-gradient-from-position),var(--tw-gradient-to)var(--tw-gradient-to-position))}.p-2{padding:calc(var(--spacing)*2)}.p-3{padding:calc(var(--spacing)*3)}.px-2{padding-inline:calc(var(--spacing)*2)}.px-4{padding-inline:calc(var(--spacing)*4)}.py-1{padding-block:calc(var(--spacing)*1)}.py-2{padding-block:calc(var(--spacing)*2)}.py-6{padding-block:calc(var(--spacing)*6)}.py-8{padding-block:calc(var(--spacing)*8)}.pr-3{padding-right:calc(var(--spacing)*3)}.pl-1{padding-left:calc(var(--spacing)*1)}.pl-2{padding-left:calc(var(--spacing)*2)}.pl-4{padding-left:calc(var(--spacing)*4)}.text-left{text-align:left}.text-right{text-align:right}.font-alternative{font-family:var(--font-alternative)}.font-mono{font-family:var(--font-mono)}.text-2xl{font-size:var(--text-2xl);line-height:var(--tw-leading,var(--text-2xl--line-height))}.text-base{font-size:var(--text-base);line-height:var(--tw-leading,var(--text-base--line-height))}.text-lg{font-size:var(--text-lg);line-height:var(--tw-leading,var(--text-lg--line-height))}.text-sm{font-size:var(--text-sm);line-height:var(--tw-leading,var(--text-sm--line-height))}.text-xl{font-size:var(--text-xl);line-height:var(--tw-leading,var(--text-xl--line-height))}.text-xs{font-size:var(--text-xs);line-height:var(--tw-leading,var(--text-xs--line-height))}.leading-4{--tw-leading:calc(var(--spacing)*4);line-height:calc(var(--spacing)*4)}.leading-5{--tw-leading:calc(var(--spacing)*5);line-height:calc(var(--spacing)*5)}.leading-6{--tw-leading:calc(var(--spacing)*6);line-height:calc(var(--spacing)*6)}.leading-7{--tw-leading:calc(var(--spacing)*7);line-height:calc(var(--spacing)*7)}.leading-8{--tw-leading:calc(var(--spacing)*8);line-height:calc(var(--spacing)*8)}.leading-9{--tw-leading:calc(var(--spacing)*9);line-height:calc(var(--spacing)*9)}.font-bold{--tw-font-weight:var(--font-weight-bold);font-weight:var(--font-weight-bold)}.font-medium{--tw-font-weight:var(--font-weight-medium);font-weight:var(--font-weight-medium)}.font-semibold{--tw-font-weight:var(--font-weight-semibold);font-weight:var(--font-weight-semibold)}.text-wrap{text-wrap:wrap}.whitespace-pre{white-space:pre}.text-bosque-200{color:var(--color-bosque-200)}.text-bosque-500{color:var(--color-bosque-500)}.text-bosque-800{color:var(--color-bosque-800)}.text-bruma-200{color:var(--color-bruma-200)}.text-bruma-500{color:var(--color-bruma-500)}.text-bruma-800{color:var(--color-bruma-800)}.text-ciclon-200{color:var(--color-ciclon-200)}.text-ciclon-500{color:var(--color-ciclon-500)}.text-ciclon-800{color:var(--color-ciclon-800)}.text-flama{color:var(--color-flama)}.text-flama-200{color:var(--color-flama-200)}.text-flama-400{color:var(--color-flama-400)}.text-flama-500{color:var(--color-flama-500)}.text-flama-800{color:var(--color-flama-800)}.text-primary-200{color:var(--color-primary-200)}.text-primary-300{color:var(--color-primary-300)}.text-primary-400{color:var(--color-primary-400)}.text-primary-500{color:var(--color-primary-500)}.text-primary-600{color:var(--color-primary-600)}.text-primary-700{color:var(--color-primary-700)}.text-primary-800{color:var(--color-primary-800)}.text-vortico-200{color:var(--color-vortico-200)}.text-vortico-500{color:var(--color-vortico-500)}.text-vortico-800{color:var(--color-vortico-800)}.capitalize{text-transform:capitalize}.italic{font-style:italic}.opacity-0{opacity:0}.opacity-100{opacity:1}.ring-1{--tw-ring-shadow:var(--tw-ring-inset,)0 0 0 calc(1px + var(--tw-ring-offset-width))var(--tw-ring-color,currentColor);box-shadow:var(--tw-inset-shadow),var(--tw-inset-ring-shadow),var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow)}.shadow-md{--tw-shadow:0 4px 6px -1px var(--tw-shadow-color,#0000001a),0 2px 4px -2px var(--tw-shadow-color,#0000001a);box-shadow:var(--tw-inset-shadow),var(--tw-inset-ring-shadow),var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow)}.shadow-primary-300{--tw-shadow-color:var(--color-primary-300)}.ring-primary-500\/50{--tw-ring-color:color-mix(in oklab,var(--color-primary-500)50%,transparent)}.filter{filter:var(--tw-blur,)var(--tw-brightness,)var(--tw-contrast,)var(--tw-grayscale,)var(--tw-hue-rotate,)var(--tw-invert,)var(--tw-saturate,)var(--tw-sepia,)var(--tw-drop-shadow,)}.backdrop-blur{--tw-backdrop-blur:blur(8px);-webkit-backdrop-filter:var(--tw-backdrop-blur,)var(--tw-backdrop-brightness,)var(--tw-backdrop-contrast,)var(--tw-backdrop-grayscale,)var(--tw-backdrop-hue-rotate,)var(--tw-backdrop-invert,)var(--tw-backdrop-opacity,)var(--tw-backdrop-saturate,)var(--tw-backdrop-sepia,);backdrop-filter:var(--tw-backdrop-blur,)var(--tw-backdrop-brightness,)var(--tw-backdrop-contrast,)var(--tw-backdrop-grayscale,)var(--tw-backdrop-hue-rotate,)var(--tw-backdrop-invert,)var(--tw-backdrop-opacity,)var(--tw-backdrop-saturate,)var(--tw-backdrop-sepia,)}.backdrop-blur-sm{--tw-backdrop-blur:blur(var(--blur-sm));-webkit-backdrop-filter:var(--tw-backdrop-blur,)var(--tw-backdrop-brightness,)var(--tw-backdrop-contrast,)var(--tw-backdrop-grayscale,)var(--tw-backdrop-hue-rotate,)var(--tw-backdrop-invert,)var(--tw-backdrop-opacity,)var(--tw-backdrop-saturate,)var(--tw-backdrop-sepia,);backdrop-filter:var(--tw-backdrop-blur,)var(--tw-backdrop-brightness,)var(--tw-backdrop-contrast,)var(--tw-backdrop-grayscale,)var(--tw-backdrop-hue-rotate,)var(--tw-backdrop-invert,)var(--tw-backdrop-opacity,)var(--tw-backdrop-saturate,)var(--tw-backdrop-sepia,)}.transition{transition-property:color,background-color,border-color,outline-color,text-decoration-color,fill,stroke,--tw-gradient-from,--tw-gradient-via,--tw-gradient-to,opacity,box-shadow,transform,translate,scale,rotate,filter,-webkit-backdrop-filter,backdrop-filter;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.transition-all{transition-property:all;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.transition-colors{transition-property:color,background-color,border-color,outline-color,text-decoration-color,fill,stroke,--tw-gradient-from,--tw-gradient-via,--tw-gradient-to;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.transition-opacity{transition-property:opacity;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.transition-transform{transition-property:transform,translate,scale,rotate;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.duration-200{--tw-duration:.2s;transition-duration:.2s}.duration-500{--tw-duration:.5s;transition-duration:.5s}.ring-inset{--tw-ring-inset:inset}.select-none{-webkit-user-select:none;user-select:none}@media (hover:hover){.group-hover\:block:is(:where(.group):hover *){display:block}.group-hover\:opacity-100:is(:where(.group):hover *){opacity:1}}.after\:absolute:after{content:var(--tw-content);position:absolute}.after\:bottom-\[100\%\]:after{content:var(--tw-content);bottom:100%}.after\:left-1\/2:after{content:var(--tw-content);left:50%}.after\:-translate-x-1\/2:after{content:var(--tw-content);--tw-translate-x:calc(calc(1/2*100%)*-1);translate:var(--tw-translate-x)var(--tw-translate-y)}.after\:border-8:after{content:var(--tw-content);border-style:var(--tw-border-style);border-width:8px}.after\:border-x-transparent:after{content:var(--tw-content);border-inline-color:#0000}.after\:border-t-transparent:after{content:var(--tw-content);border-top-color:#0000}.after\:border-b-primary-800:after{content:var(--tw-content);border-bottom-color:var(--color-primary-800)}.after\:content-\[\'\'\]:after{content:var(--tw-content);--tw-content:"";content:var(--tw-content)}.first\:pl-0:first-child{padding-left:calc(var(--spacing)*0)}.last\:pr-0:last-child{padding-right:calc(var(--spacing)*0)}@media (hover:hover){.hover\:border-l-4:hover{border-left-style:var(--tw-border-style);border-left-width:4px}.hover\:border-flama-500:hover{border-color:var(--color-flama-500)}.hover\:bg-flama-300:hover{background-color:var(--color-flama-300)}.hover\:pl-0:hover{padding-left:calc(var(--spacing)*0)}.hover\:text-bosque:hover{color:var(--color-bosque)}.hover\:text-ciclon:hover{color:var(--color-ciclon)}.hover\:text-flama:hover{color:var(--color-flama)}.hover\:text-flama-500:hover{color:var(--color-flama-500)}.hover\:text-primary-800:hover{color:var(--color-primary-800)}.hover\:text-vortico-500:hover{color:var(--color-vortico-500)}.hover\:shadow-lg:hover{--tw-shadow:0 10px 15px -3px var(--tw-shadow-color,#0000001a),0 4px 6px -4px var(--tw-shadow-color,#0000001a);box-shadow:var(--tw-inset-shadow),var(--tw-inset-ring-shadow),var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow)}.hover\:shadow-flama-300:hover{--tw-shadow-color:var(--color-flama-300)}}@media (width>=40rem){.sm\:inset-x-6{inset-inline:calc(var(--spacing)*6)}.sm\:px-6{padding-inline:calc(var(--spacing)*6)}.sm\:py-10{padding-block:calc(var(--spacing)*10)}}@media (width>=48rem){.md\:sticky{position:sticky}.md\:inset-x-8{inset-inline:calc(var(--spacing)*8)}.md\:left-0{left:calc(var(--spacing)*0)}.md\:block{display:block}.md\:flex{display:flex}.md\:hidden{display:none}.md\:h-5{height:calc(var(--spacing)*5)}.md\:h-6{height:calc(var(--spacing)*6)}.md\:h-7{height:calc(var(--spacing)*7)}.md\:h-8{height:calc(var(--spacing)*8)}.md\:h-9{height:calc(var(--spacing)*9)}.md\:h-10{height:calc(var(--spacing)*10)}.md\:w-5{width:calc(var(--spacing)*5)}.md\:w-6{width:calc(var(--spacing)*6)}.md\:w-7{width:calc(var(--spacing)*7)}.md\:w-8{width:calc(var(--spacing)*8)}.md\:w-9{width:calc(var(--spacing)*9)}.md\:w-10{width:calc(var(--spacing)*10)}.md\:px-8{padding-inline:calc(var(--spacing)*8)}.md\:py-12{padding-block:calc(var(--spacing)*12)}.md\:text-2xl{font-size:var(--text-2xl);line-height:var(--tw-leading,var(--text-2xl--line-height))}.md\:text-3xl{font-size:var(--text-3xl);line-height:var(--tw-leading,var(--text-3xl--line-height))}.md\:text-base{font-size:var(--text-base);line-height:var(--tw-leading,var(--text-base--line-height))}.md\:text-lg{font-size:var(--text-lg);line-height:var(--tw-leading,var(--text-lg--line-height))}.md\:text-sm{font-size:var(--text-sm);line-height:var(--tw-leading,var(--text-sm--line-height))}.md\:text-xl{font-size:var(--text-xl);line-height:var(--tw-leading,var(--text-xl--line-height))}.md\:leading-5{--tw-leading:calc(var(--spacing)*5);line-height:calc(var(--spacing)*5)}.md\:leading-6{--tw-leading:calc(var(--spacing)*6);line-height:calc(var(--spacing)*6)}.md\:leading-7{--tw-leading:calc(var(--spacing)*7);line-height:calc(var(--spacing)*7)}.md\:leading-8{--tw-leading:calc(var(--spacing)*8);line-height:calc(var(--spacing)*8)}.md\:leading-9{--tw-leading:calc(var(--spacing)*9);line-height:calc(var(--spacing)*9)}.md\:leading-10{--tw-leading:calc(var(--spacing)*10);line-height:calc(var(--spacing)*10)}}@media (width>=64rem){.lg\:w-xs{width:var(--container-xs)}.lg\:flex-row{flex-direction:row}}}:root{background-color:var(--color-primary-100);color:var(--color-primary-600);scrollbar-color:var(--color-primary-400)transparent;z-index:0}body{height:100%}@keyframes spin{to{transform:rotate(360deg)}}@keyframes ping{75%,to{opacity:0;transform:scale(2)}}@keyframes pulse{50%{opacity:.5}}@keyframes bounce{0%,to{animation-timing-function:cubic-bezier(.8,0,1,1);transform:translateY(-25%)}50%{animation-timing-function:cubic-bezier(0,0,.2,1);transform:none}}@property --tw-translate-x{syntax:"*";inherits:false;initial-value:0}@property --tw-translate-y{syntax:"*";inherits:false;initial-value:0}@property --tw-translate-z{syntax:"*";inherits:false;initial-value:0}@property --tw-divide-x-reverse{syntax:"*";inherits:false;initial-value:0}@property --tw-border-style{syntax:"*";inherits:false;initial-value:solid}@property --tw-divide-y-reverse{syntax:"*";inherits:false;initial-value:0}@property --tw-gradient-position{syntax:"*";inherits:false}@property --tw-gradient-from{syntax:"<color>";inherits:false;initial-value:#0000}@property --tw-gradient-via{syntax:"<color>";inherits:false;initial-value:#0000}@property --tw-gradient-to{syntax:"<color>";inherits:false;initial-value:#0000}@property --tw-gradient-stops{syntax:"*";inherits:false}@property --tw-gradient-via-stops{syntax:"*";inherits:false}@property --tw-gradient-from-position{syntax:"<length-percentage>";inherits:false;initial-value:0%}@property --tw-gradient-via-position{syntax:"<length-percentage>";inherits:false;initial-value:50%}@property --tw-gradient-to-position{syntax:"<length-percentage>";inherits:false;initial-value:100%}@property --tw-leading{syntax:"*";inherits:false}@property --tw-font-weight{syntax:"*";inherits:false}@property --tw-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-shadow-color{syntax:"*";inherits:false}@property --tw-inset-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-inset-shadow-color{syntax:"*";inherits:false}@property --tw-ring-color{syntax:"*";inherits:false}@property --tw-ring-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-inset-ring-color{syntax:"*";inherits:false}@property --tw-inset-ring-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-ring-inset{syntax:"*";inherits:false}@property --tw-ring-offset-width{syntax:"<length>";inherits:false;initial-value:0}@property --tw-ring-offset-color{syntax:"*";inherits:false;initial-value:#fff}@property --tw-ring-offset-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-blur{syntax:"*";inherits:false}@property --tw-brightness{syntax:"*";inherits:false}@property --tw-contrast{syntax:"*";inherits:false}@property --tw-grayscale{syntax:"*";inherits:false}@property --tw-hue-rotate{syntax:"*";inherits:false}@property --tw-invert{syntax:"*";inherits:false}@property --tw-opacity{syntax:"*";inherits:false}@property --tw-saturate{syntax:"*";inherits:false}@property --tw-sepia{syntax:"*";inherits:false}@property --tw-drop-shadow{syntax:"*";inherits:false}@property --tw-backdrop-blur{syntax:"*";inherits:false}@property --tw-backdrop-brightness{syntax:"*";inherits:false}@property --tw-backdrop-contrast{syntax:"*";inherits:false}@property --tw-backdrop-grayscale{syntax:"*";inherits:false}@property --tw-backdrop-hue-rotate{syntax:"*";inherits:false}@property --tw-backdrop-invert{syntax:"*";inherits:false}@property --tw-backdrop-opacity{syntax:"*";inherits:false}@property --tw-backdrop-saturate{syntax:"*";inherits:false}@property --tw-backdrop-sepia{syntax:"*";inherits:false}@property --tw-duration{syntax:"*";inherits:false}@property --tw-content{syntax:"*";inherits:false;initial-value:""}
</style></head><body><div id="app"/><script>(()=>{"use strict";var e={34:(e,t,n)=>{var r=n(4901);e.exports=function(e){return"object"==typeof e?null!==e:r(e)}},81:(e,t,n)=>{var r=n(9565),a=n(9306),l=n(8551),o=n(6823),i=n(851),u=TypeError;e.exports=function(e,t){var n=arguments.length<2?i(e):t;if(a(n))return l(r(n,e));throw new u(o(e)+" is not iterable")}},230:(e,t,n)=>{var r=n(6518),a=n(9306),l=n(6194),o=n(6223),i=TypeError;r({target:"Map",proto:!0,real:!0,forced:!0},{reduce:function(e){var t=l(this),n=arguments.length<2,r=n?void 0:arguments[1];if(a(e),o(t,(function(a,l){n?(n=!1,r=a):r=e(r,a,l,t)})),n)throw new i("Reduce of empty map with no initial value");return r}})},283:(e,t,n)=>{var r=n(9504),a=n(9039),l=n(4901),o=n(9297),i=n(3724),u=n(350).CONFIGURABLE,s=n(3706),c=n(1181),f=c.enforce,d=c.get,p=String,h=Object.defineProperty,m=r("".slice),v=r("".replace),g=r([].join),y=i&&!a((function(){return 8!==h((function(){}),"length",{value:8}).length})),b=String(String).split("String"),k=e.exports=function(e,t,n){"Symbol("===m(p(t),0,7)&&(t="["+v(p(t),/^Symbol\(([^)]*)\).*$/,"$1")+"]"),n&&n.getter&&(t="get "+t),n&&n.setter&&(t="set "+t),(!o(e,"name")||u&&e.name!==t)&&(i?h(e,"name",{value:t,configurable:!0}):e.name=t),y&&n&&o(n,"arity")&&e.length!==n.arity&&h(e,"length",{value:n.arity});try{n&&o(n,"constructor")&&n.constructor?i&&h(e,"prototype",{writable:!1}):e.prototype&&(e.prototype=void 0)}catch(e){}var r=f(e);return o(r,"source")||(r.source=g(b,"string"==typeof t?t:"")),e};Function.prototype.toString=k((function(){return l(this)&&d(this).source||s(this)}),"toString")},350:(e,t,n)=>{var r=n(3724),a=n(9297),l=Function.prototype,o=r&&Object.getOwnPropertyDescriptor,i=a(l,"name"),u=i&&"something"===function(){}.name,s=i&&(!r||r&&o(l,"name").configurable);e.exports={EXISTS:i,PROPER:u,CONFIGURABLE:s}},421:e=>{e.exports={}},507:(e,t,n)=>{var r=n(9565);e.exports=function(e,t,n){for(var a,l,o=n?e:e.iterator,i=e.next;!(a=r(i,o)).done;)if(void 0!==(l=t(a.value)))return l}},616:(e,t,n)=>{var r=n(9039);e.exports=!r((function(){var e=function(){}.bind();return"function"!=typeof e||e.hasOwnProperty("prototype")}))},741:e=>{var t=Math.ceil,n=Math.floor;e.exports=Math.trunc||function(e){var r=+e;return(r>0?n:t)(r)}},757:(e,t,n)=>{var r=n(7751),a=n(4901),l=n(1625),o=n(7040),i=Object;e.exports=o?function(e){return"symbol"==typeof e}:function(e){var t=r("Symbol");return a(t)&&l(t.prototype,i(e))}},851:(e,t,n)=>{var r=n(6955),a=n(5966),l=n(4117),o=n(6269),i=n(8227)("iterator");e.exports=function(e){if(!l(e))return a(e,i)||a(e,"@@iterator")||o[r(e)]}},961:(e,t,n)=>{!function e(){if("undefined"!=typeof __REACT_DEVTOOLS_GLOBAL_HOOK__&&"function"==typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE)try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(e){console.error(e)}}(),e.exports=n(6221)},1181:(e,t,n)=>{var r,a,l,o=n(8622),i=n(4576),u=n(34),s=n(6699),c=n(9297),f=n(7629),d=n(6119),p=n(421),h="Object already initialized",m=i.TypeError,v=i.WeakMap;if(o||f.state){var g=f.state||(f.state=new v);g.get=g.get,g.has=g.has,g.set=g.set,r=function(e,t){if(g.has(e))throw new m(h);return t.facade=e,g.set(e,t),t},a=function(e){return g.get(e)||{}},l=function(e){return g.has(e)}}else{var y=d("state");p[y]=!0,r=function(e,t){if(c(e,y))throw new m(h);return t.facade=e,s(e,y,t),t},a=function(e){return c(e,y)?e[y]:{}},l=function(e){return c(e,y)}}e.exports={set:r,get:a,has:l,enforce:function(e){return l(e)?a(e):r(e,{})},getterFor:function(e){return function(t){var n;if(!u(t)||(n=a(t)).type!==e)throw new m("Incompatible receiver, "+e+" required");return n}}}},1247:(e,t,n)=>{var r=n(9982),a=n(6540),l=n(961);function o(e){var t="https://react.dev/errors/"+e;if(1<arguments.length){t+="?args[]="+encodeURIComponent(arguments[1]);for(var n=2;n<arguments.length;n++)t+="&args[]="+encodeURIComponent(arguments[n])}return"Minified React error #"+e+"; visit "+t+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function i(e){return!(!e||1!==e.nodeType&&9!==e.nodeType&&11!==e.nodeType)}var u=Symbol.for("react.element"),s=Symbol.for("react.transitional.element"),c=Symbol.for("react.portal"),f=Symbol.for("react.fragment"),d=Symbol.for("react.strict_mode"),p=Symbol.for("react.profiler"),h=Symbol.for("react.provider"),m=Symbol.for("react.consumer"),v=Symbol.for("react.context"),g=Symbol.for("react.forward_ref"),y=Symbol.for("react.suspense"),b=Symbol.for("react.suspense_list"),k=Symbol.for("react.memo"),w=Symbol.for("react.lazy");Symbol.for("react.scope"),Symbol.for("react.debug_trace_mode");var x=Symbol.for("react.offscreen");Symbol.for("react.legacy_hidden"),Symbol.for("react.tracing_marker");var S=Symbol.for("react.memo_cache_sentinel"),E=Symbol.iterator;function C(e){return null===e||"object"!=typeof e?null:"function"==typeof(e=E&&e[E]||e["@@iterator"])?e:null}var N=Symbol.for("react.client.reference");function _(e){if(null==e)return null;if("function"==typeof e)return e.$$typeof===N?null:e.displayName||e.name||null;if("string"==typeof e)return e;switch(e){case f:return"Fragment";case c:return"Portal";case p:return"Profiler";case d:return"StrictMode";case y:return"Suspense";case b:return"SuspenseList"}if("object"==typeof e)switch(e.$$typeof){case v:return(e.displayName||"Context")+".Provider";case m:return(e._context.displayName||"Context")+".Consumer";case g:var t=e.render;return(e=e.displayName)||(e=""!==(e=t.displayName||t.name||"")?"ForwardRef("+e+")":"ForwardRef"),e;case k:return null!==(t=e.displayName||null)?t:_(e.type)||"Memo";case w:t=e._payload,e=e._init;try{return _(e(t))}catch(e){}}return null}var P,z,j=a.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,T=Object.assign;function O(e){if(void 0===P)try{throw Error()}catch(e){var t=e.stack.trim().match(/\n( *(at )?)/);P=t&&t[1]||"",z=-1<e.stack.indexOf("\n    at")?" (<anonymous>)":-1<e.stack.indexOf("@")?"@unknown:0:0":""}return"\n"+P+e+z}var L=!1;function M(e,t){if(!e||L)return"";L=!0;var n=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{var r={DetermineComponentFrameRoot:function(){try{if(t){var n=function(){throw Error()};if(Object.defineProperty(n.prototype,"props",{set:function(){throw Error()}}),"object"==typeof Reflect&&Reflect.construct){try{Reflect.construct(n,[])}catch(e){var r=e}Reflect.construct(e,[],n)}else{try{n.call()}catch(e){r=e}e.call(n.prototype)}}else{try{throw Error()}catch(e){r=e}(n=e())&&"function"==typeof n.catch&&n.catch((function(){}))}}catch(e){if(e&&r&&"string"==typeof e.stack)return[e.stack,r.stack]}return[null,null]}};r.DetermineComponentFrameRoot.displayName="DetermineComponentFrameRoot";var a=Object.getOwnPropertyDescriptor(r.DetermineComponentFrameRoot,"name");a&&a.configurable&&Object.defineProperty(r.DetermineComponentFrameRoot,"name",{value:"DetermineComponentFrameRoot"});var l=r.DetermineComponentFrameRoot(),o=l[0],i=l[1];if(o&&i){var u=o.split("\n"),s=i.split("\n");for(a=r=0;r<u.length&&!u[r].includes("DetermineComponentFrameRoot");)r++;for(;a<s.length&&!s[a].includes("DetermineComponentFrameRoot");)a++;if(r===u.length||a===s.length)for(r=u.length-1,a=s.length-1;1<=r&&0<=a&&u[r]!==s[a];)a--;for(;1<=r&&0<=a;r--,a--)if(u[r]!==s[a]){if(1!==r||1!==a)do{if(r--,0>--a||u[r]!==s[a]){var c="\n"+u[r].replace(" at new "," at ");return e.displayName&&c.includes("<anonymous>")&&(c=c.replace("<anonymous>",e.displayName)),c}}while(1<=r&&0<=a);break}}}finally{L=!1,Error.prepareStackTrace=n}return(n=e?e.displayName||e.name:"")?O(n):""}function A(e){switch(e.tag){case 26:case 27:case 5:return O(e.type);case 16:return O("Lazy");case 13:return O("Suspense");case 19:return O("SuspenseList");case 0:case 15:return M(e.type,!1);case 11:return M(e.type.render,!1);case 1:return M(e.type,!0);default:return""}}function R(e){try{var t="";do{t+=A(e),e=e.return}while(e);return t}catch(e){return"\nError generating stack: "+e.message+"\n"+e.stack}}function F(e){var t=e,n=e;if(e.alternate)for(;t.return;)t=t.return;else{e=t;do{!!(4098&(t=e).flags)&&(n=t.return),e=t.return}while(e)}return 3===t.tag?n:null}function D(e){if(13===e.tag){var t=e.memoizedState;if(null===t&&null!==(e=e.alternate)&&(t=e.memoizedState),null!==t)return t.dehydrated}return null}function I(e){if(F(e)!==e)throw Error(o(188))}function U(e){var t=e.tag;if(5===t||26===t||27===t||6===t)return e;for(e=e.child;null!==e;){if(null!==(t=U(e)))return t;e=e.sibling}return null}var $=Array.isArray,H=l.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,V={pending:!1,data:null,method:null,action:null},B=[],q=-1;function W(e){return{current:e}}function Q(e){0>q||(e.current=B[q],B[q]=null,q--)}function K(e,t){q++,B[q]=e.current,e.current=t}var G=W(null),Y=W(null),X=W(null),Z=W(null);function J(e,t){switch(K(X,t),K(Y,e),K(G,null),e=t.nodeType){case 9:case 11:t=(t=t.documentElement)&&(t=t.namespaceURI)?Yc(t):0;break;default:if(t=(e=8===e?t.parentNode:t).tagName,e=e.namespaceURI)t=Xc(e=Yc(e),t);else switch(t){case"svg":t=1;break;case"math":t=2;break;default:t=0}}Q(G),K(G,t)}function ee(){Q(G),Q(Y),Q(X)}function te(e){null!==e.memoizedState&&K(Z,e);var t=G.current,n=Xc(t,e.type);t!==n&&(K(Y,e),K(G,n))}function ne(e){Y.current===e&&(Q(G),Q(Y)),Z.current===e&&(Q(Z),Ff._currentValue=V)}var re=Object.prototype.hasOwnProperty,ae=r.unstable_scheduleCallback,le=r.unstable_cancelCallback,oe=r.unstable_shouldYield,ie=r.unstable_requestPaint,ue=r.unstable_now,se=r.unstable_getCurrentPriorityLevel,ce=r.unstable_ImmediatePriority,fe=r.unstable_UserBlockingPriority,de=r.unstable_NormalPriority,pe=r.unstable_LowPriority,he=r.unstable_IdlePriority,me=r.log,ve=r.unstable_setDisableYieldValue,ge=null,ye=null;function be(e){if("function"==typeof me&&ve(e),ye&&"function"==typeof ye.setStrictMode)try{ye.setStrictMode(ge,e)}catch(e){}}var ke=Math.clz32?Math.clz32:function(e){return 0==(e>>>=0)?32:31-(we(e)/xe|0)|0},we=Math.log,xe=Math.LN2,Se=128,Ee=4194304;function Ce(e){var t=42&e;if(0!==t)return t;switch(e&-e){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:return 64;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return 4194176&e;case 4194304:case 8388608:case 16777216:case 33554432:return 62914560&e;case 67108864:return 67108864;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 0;default:return e}}function Ne(e,t){var n=e.pendingLanes;if(0===n)return 0;var r=0,a=e.suspendedLanes,l=e.pingedLanes,o=e.warmLanes;e=0!==e.finishedLanes;var i=134217727&n;return 0!==i?0!=(n=i&~a)?r=Ce(n):0!=(l&=i)?r=Ce(l):e||0!=(o=i&~o)&&(r=Ce(o)):0!=(i=n&~a)?r=Ce(i):0!==l?r=Ce(l):e||0!=(o=n&~o)&&(r=Ce(o)),0===r?0:0===t||t===r||t&a||!((a=r&-r)>=(o=t&-t)||32===a&&4194176&o)?r:t}function _e(e,t){return!(e.pendingLanes&~(e.suspendedLanes&~e.pingedLanes)&t)}function Pe(e,t){switch(e){case 1:case 2:case 4:case 8:return t+250;case 16:case 32:case 64:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return t+5e3;default:return-1}}function ze(){var e=Se;return!(4194176&(Se<<=1))&&(Se=128),e}function je(){var e=Ee;return!(62914560&(Ee<<=1))&&(Ee=4194304),e}function Te(e){for(var t=[],n=0;31>n;n++)t.push(e);return t}function Oe(e,t){e.pendingLanes|=t,268435456!==t&&(e.suspendedLanes=0,e.pingedLanes=0,e.warmLanes=0)}function Le(e,t,n){e.pendingLanes|=t,e.suspendedLanes&=~t;var r=31-ke(t);e.entangledLanes|=t,e.entanglements[r]=1073741824|e.entanglements[r]|4194218&n}function Me(e,t){var n=e.entangledLanes|=t;for(e=e.entanglements;n;){var r=31-ke(n),a=1<<r;a&t|e[r]&t&&(e[r]|=t),n&=~a}}function Ae(e){return 2<(e&=-e)?8<e?134217727&e?32:268435456:8:2}function Re(){var e=H.p;return 0!==e?e:void 0===(e=window.event)?32:Zf(e.type)}var Fe=Math.random().toString(36).slice(2),De="__reactFiber$"+Fe,Ie="__reactProps$"+Fe,Ue="__reactContainer$"+Fe,$e="__reactEvents$"+Fe,He="__reactListeners$"+Fe,Ve="__reactHandles$"+Fe,Be="__reactResources$"+Fe,qe="__reactMarker$"+Fe;function We(e){delete e[De],delete e[Ie],delete e[$e],delete e[He],delete e[Ve]}function Qe(e){var t=e[De];if(t)return t;for(var n=e.parentNode;n;){if(t=n[Ue]||n[De]){if(n=t.alternate,null!==t.child||null!==n&&null!==n.child)for(e=sf(e);null!==e;){if(n=e[De])return n;e=sf(e)}return t}n=(e=n).parentNode}return null}function Ke(e){if(e=e[De]||e[Ue]){var t=e.tag;if(5===t||6===t||13===t||26===t||27===t||3===t)return e}return null}function Ge(e){var t=e.tag;if(5===t||26===t||27===t||6===t)return e.stateNode;throw Error(o(33))}function Ye(e){var t=e[Be];return t||(t=e[Be]={hoistableStyles:new Map,hoistableScripts:new Map}),t}function Xe(e){e[qe]=!0}var Ze=new Set,Je={};function et(e,t){tt(e,t),tt(e+"Capture",t)}function tt(e,t){for(Je[e]=t,e=0;e<t.length;e++)Ze.add(t[e])}var nt=!("undefined"==typeof window||void 0===window.document||void 0===window.document.createElement),rt=RegExp("^[:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD][:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD\\-.0-9\\u00B7\\u0300-\\u036F\\u203F-\\u2040]*$"),at={},lt={};function ot(e,t,n){if(a=t,re.call(lt,a)||!re.call(at,a)&&(rt.test(a)?lt[a]=!0:(at[a]=!0,0)))if(null===n)e.removeAttribute(t);else{switch(typeof n){case"undefined":case"function":case"symbol":return void e.removeAttribute(t);case"boolean":var r=t.toLowerCase().slice(0,5);if("data-"!==r&&"aria-"!==r)return void e.removeAttribute(t)}e.setAttribute(t,""+n)}var a}function it(e,t,n){if(null===n)e.removeAttribute(t);else{switch(typeof n){case"undefined":case"function":case"symbol":case"boolean":return void e.removeAttribute(t)}e.setAttribute(t,""+n)}}function ut(e,t,n,r){if(null===r)e.removeAttribute(n);else{switch(typeof r){case"undefined":case"function":case"symbol":case"boolean":return void e.removeAttribute(n)}e.setAttributeNS(t,n,""+r)}}function st(e){switch(typeof e){case"bigint":case"boolean":case"number":case"string":case"undefined":case"object":return e;default:return""}}function ct(e){var t=e.type;return(e=e.nodeName)&&"input"===e.toLowerCase()&&("checkbox"===t||"radio"===t)}function ft(e){e._valueTracker||(e._valueTracker=function(e){var t=ct(e)?"checked":"value",n=Object.getOwnPropertyDescriptor(e.constructor.prototype,t),r=""+e[t];if(!e.hasOwnProperty(t)&&void 0!==n&&"function"==typeof n.get&&"function"==typeof n.set){var a=n.get,l=n.set;return Object.defineProperty(e,t,{configurable:!0,get:function(){return a.call(this)},set:function(e){r=""+e,l.call(this,e)}}),Object.defineProperty(e,t,{enumerable:n.enumerable}),{getValue:function(){return r},setValue:function(e){r=""+e},stopTracking:function(){e._valueTracker=null,delete e[t]}}}}(e))}function dt(e){if(!e)return!1;var t=e._valueTracker;if(!t)return!0;var n=t.getValue(),r="";return e&&(r=ct(e)?e.checked?"true":"false":e.value),(e=r)!==n&&(t.setValue(e),!0)}function pt(e){if(void 0===(e=e||("undefined"!=typeof document?document:void 0)))return null;try{return e.activeElement||e.body}catch(t){return e.body}}var ht=/[\n"\\]/g;function mt(e){return e.replace(ht,(function(e){return"\\"+e.charCodeAt(0).toString(16)+" "}))}function vt(e,t,n,r,a,l,o,i){e.name="",null!=o&&"function"!=typeof o&&"symbol"!=typeof o&&"boolean"!=typeof o?e.type=o:e.removeAttribute("type"),null!=t?"number"===o?(0===t&&""===e.value||e.value!=t)&&(e.value=""+st(t)):e.value!==""+st(t)&&(e.value=""+st(t)):"submit"!==o&&"reset"!==o||e.removeAttribute("value"),null!=t?yt(e,o,st(t)):null!=n?yt(e,o,st(n)):null!=r&&e.removeAttribute("value"),null==a&&null!=l&&(e.defaultChecked=!!l),null!=a&&(e.checked=a&&"function"!=typeof a&&"symbol"!=typeof a),null!=i&&"function"!=typeof i&&"symbol"!=typeof i&&"boolean"!=typeof i?e.name=""+st(i):e.removeAttribute("name")}function gt(e,t,n,r,a,l,o,i){if(null!=l&&"function"!=typeof l&&"symbol"!=typeof l&&"boolean"!=typeof l&&(e.type=l),null!=t||null!=n){if(("submit"===l||"reset"===l)&&null==t)return;n=null!=n?""+st(n):"",t=null!=t?""+st(t):n,i||t===e.value||(e.value=t),e.defaultValue=t}r="function"!=typeof(r=null!=r?r:a)&&"symbol"!=typeof r&&!!r,e.checked=i?e.checked:!!r,e.defaultChecked=!!r,null!=o&&"function"!=typeof o&&"symbol"!=typeof o&&"boolean"!=typeof o&&(e.name=o)}function yt(e,t,n){"number"===t&&pt(e.ownerDocument)===e||e.defaultValue===""+n||(e.defaultValue=""+n)}function bt(e,t,n,r){if(e=e.options,t){t={};for(var a=0;a<n.length;a++)t["$"+n[a]]=!0;for(n=0;n<e.length;n++)a=t.hasOwnProperty("$"+e[n].value),e[n].selected!==a&&(e[n].selected=a),a&&r&&(e[n].defaultSelected=!0)}else{for(n=""+st(n),t=null,a=0;a<e.length;a++){if(e[a].value===n)return e[a].selected=!0,void(r&&(e[a].defaultSelected=!0));null!==t||e[a].disabled||(t=e[a])}null!==t&&(t.selected=!0)}}function kt(e,t,n){null==t||((t=""+st(t))!==e.value&&(e.value=t),null!=n)?e.defaultValue=null!=n?""+st(n):"":e.defaultValue!==t&&(e.defaultValue=t)}function wt(e,t,n,r){if(null==t){if(null!=r){if(null!=n)throw Error(o(92));if($(r)){if(1<r.length)throw Error(o(93));r=r[0]}n=r}null==n&&(n=""),t=n}n=st(t),e.defaultValue=n,(r=e.textContent)===n&&""!==r&&null!==r&&(e.value=r)}function xt(e,t){if(t){var n=e.firstChild;if(n&&n===e.lastChild&&3===n.nodeType)return void(n.nodeValue=t)}e.textContent=t}var St=new Set("animationIterationCount aspectRatio borderImageOutset borderImageSlice borderImageWidth boxFlex boxFlexGroup boxOrdinalGroup columnCount columns flex flexGrow flexPositive flexShrink flexNegative flexOrder gridArea gridRow gridRowEnd gridRowSpan gridRowStart gridColumn gridColumnEnd gridColumnSpan gridColumnStart fontWeight lineClamp lineHeight opacity order orphans scale tabSize widows zIndex zoom fillOpacity floodOpacity stopOpacity strokeDasharray strokeDashoffset strokeMiterlimit strokeOpacity strokeWidth MozAnimationIterationCount MozBoxFlex MozBoxFlexGroup MozLineClamp msAnimationIterationCount msFlex msZoom msFlexGrow msFlexNegative msFlexOrder msFlexPositive msFlexShrink msGridColumn msGridColumnSpan msGridRow msGridRowSpan WebkitAnimationIterationCount WebkitBoxFlex WebKitBoxFlexGroup WebkitBoxOrdinalGroup WebkitColumnCount WebkitColumns WebkitFlex WebkitFlexGrow WebkitFlexPositive WebkitFlexShrink WebkitLineClamp".split(" "));function Et(e,t,n){var r=0===t.indexOf("--");null==n||"boolean"==typeof n||""===n?r?e.setProperty(t,""):"float"===t?e.cssFloat="":e[t]="":r?e.setProperty(t,n):"number"!=typeof n||0===n||St.has(t)?"float"===t?e.cssFloat=n:e[t]=(""+n).trim():e[t]=n+"px"}function Ct(e,t,n){if(null!=t&&"object"!=typeof t)throw Error(o(62));if(e=e.style,null!=n){for(var r in n)!n.hasOwnProperty(r)||null!=t&&t.hasOwnProperty(r)||(0===r.indexOf("--")?e.setProperty(r,""):"float"===r?e.cssFloat="":e[r]="");for(var a in t)r=t[a],t.hasOwnProperty(a)&&n[a]!==r&&Et(e,a,r)}else for(var l in t)t.hasOwnProperty(l)&&Et(e,l,t[l])}function Nt(e){if(-1===e.indexOf("-"))return!1;switch(e){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}var _t=new Map([["acceptCharset","accept-charset"],["htmlFor","for"],["httpEquiv","http-equiv"],["crossOrigin","crossorigin"],["accentHeight","accent-height"],["alignmentBaseline","alignment-baseline"],["arabicForm","arabic-form"],["baselineShift","baseline-shift"],["capHeight","cap-height"],["clipPath","clip-path"],["clipRule","clip-rule"],["colorInterpolation","color-interpolation"],["colorInterpolationFilters","color-interpolation-filters"],["colorProfile","color-profile"],["colorRendering","color-rendering"],["dominantBaseline","dominant-baseline"],["enableBackground","enable-background"],["fillOpacity","fill-opacity"],["fillRule","fill-rule"],["floodColor","flood-color"],["floodOpacity","flood-opacity"],["fontFamily","font-family"],["fontSize","font-size"],["fontSizeAdjust","font-size-adjust"],["fontStretch","font-stretch"],["fontStyle","font-style"],["fontVariant","font-variant"],["fontWeight","font-weight"],["glyphName","glyph-name"],["glyphOrientationHorizontal","glyph-orientation-horizontal"],["glyphOrientationVertical","glyph-orientation-vertical"],["horizAdvX","horiz-adv-x"],["horizOriginX","horiz-origin-x"],["imageRendering","image-rendering"],["letterSpacing","letter-spacing"],["lightingColor","lighting-color"],["markerEnd","marker-end"],["markerMid","marker-mid"],["markerStart","marker-start"],["overlinePosition","overline-position"],["overlineThickness","overline-thickness"],["paintOrder","paint-order"],["panose-1","panose-1"],["pointerEvents","pointer-events"],["renderingIntent","rendering-intent"],["shapeRendering","shape-rendering"],["stopColor","stop-color"],["stopOpacity","stop-opacity"],["strikethroughPosition","strikethrough-position"],["strikethroughThickness","strikethrough-thickness"],["strokeDasharray","stroke-dasharray"],["strokeDashoffset","stroke-dashoffset"],["strokeLinecap","stroke-linecap"],["strokeLinejoin","stroke-linejoin"],["strokeMiterlimit","stroke-miterlimit"],["strokeOpacity","stroke-opacity"],["strokeWidth","stroke-width"],["textAnchor","text-anchor"],["textDecoration","text-decoration"],["textRendering","text-rendering"],["transformOrigin","transform-origin"],["underlinePosition","underline-position"],["underlineThickness","underline-thickness"],["unicodeBidi","unicode-bidi"],["unicodeRange","unicode-range"],["unitsPerEm","units-per-em"],["vAlphabetic","v-alphabetic"],["vHanging","v-hanging"],["vIdeographic","v-ideographic"],["vMathematical","v-mathematical"],["vectorEffect","vector-effect"],["vertAdvY","vert-adv-y"],["vertOriginX","vert-origin-x"],["vertOriginY","vert-origin-y"],["wordSpacing","word-spacing"],["writingMode","writing-mode"],["xmlnsXlink","xmlns:xlink"],["xHeight","x-height"]]),Pt=/^[\u0000-\u001F ]*j[\r\n\t]*a[\r\n\t]*v[\r\n\t]*a[\r\n\t]*s[\r\n\t]*c[\r\n\t]*r[\r\n\t]*i[\r\n\t]*p[\r\n\t]*t[\r\n\t]*:/i;function zt(e){return Pt.test(""+e)?"javascript:throw new Error('React has blocked a javascript: URL as a security precaution.')":e}var jt=null;function Tt(e){return(e=e.target||e.srcElement||window).correspondingUseElement&&(e=e.correspondingUseElement),3===e.nodeType?e.parentNode:e}var Ot=null,Lt=null;function Mt(e){var t=Ke(e);if(t&&(e=t.stateNode)){var n=e[Ie]||null;e:switch(e=t.stateNode,t.type){case"input":if(vt(e,n.value,n.defaultValue,n.defaultValue,n.checked,n.defaultChecked,n.type,n.name),t=n.name,"radio"===n.type&&null!=t){for(n=e;n.parentNode;)n=n.parentNode;for(n=n.querySelectorAll('input[name="'+mt(""+t)+'"][type="radio"]'),t=0;t<n.length;t++){var r=n[t];if(r!==e&&r.form===e.form){var a=r[Ie]||null;if(!a)throw Error(o(90));vt(r,a.value,a.defaultValue,a.defaultValue,a.checked,a.defaultChecked,a.type,a.name)}}for(t=0;t<n.length;t++)(r=n[t]).form===e.form&&dt(r)}break e;case"textarea":kt(e,n.value,n.defaultValue);break e;case"select":null!=(t=n.value)&&bt(e,!!n.multiple,t,!1)}}}var At=!1;function Rt(e,t,n){if(At)return e(t,n);At=!0;try{return e(t)}finally{if(At=!1,(null!==Ot||null!==Lt)&&(Ds(),Ot&&(t=Ot,e=Lt,Lt=Ot=null,Mt(t),e)))for(t=0;t<e.length;t++)Mt(e[t])}}function Ft(e,t){var n=e.stateNode;if(null===n)return null;var r=n[Ie]||null;if(null===r)return null;n=r[t];e:switch(t){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":case"onMouseEnter":(r=!r.disabled)||(r=!("button"===(e=e.type)||"input"===e||"select"===e||"textarea"===e)),e=!r;break e;default:e=!1}if(e)return null;if(n&&"function"!=typeof n)throw Error(o(231,t,typeof n));return n}var Dt=!1;if(nt)try{var It={};Object.defineProperty(It,"passive",{get:function(){Dt=!0}}),window.addEventListener("test",It,It),window.removeEventListener("test",It,It)}catch(e){Dt=!1}var Ut=null,$t=null,Ht=null;function Vt(){if(Ht)return Ht;var e,t,n=$t,r=n.length,a="value"in Ut?Ut.value:Ut.textContent,l=a.length;for(e=0;e<r&&n[e]===a[e];e++);var o=r-e;for(t=1;t<=o&&n[r-t]===a[l-t];t++);return Ht=a.slice(e,1<t?1-t:void 0)}function Bt(e){var t=e.keyCode;return"charCode"in e?0===(e=e.charCode)&&13===t&&(e=13):e=t,10===e&&(e=13),32<=e||13===e?e:0}function qt(){return!0}function Wt(){return!1}function Qt(e){function t(t,n,r,a,l){for(var o in this._reactName=t,this._targetInst=r,this.type=n,this.nativeEvent=a,this.target=l,this.currentTarget=null,e)e.hasOwnProperty(o)&&(t=e[o],this[o]=t?t(a):a[o]);return this.isDefaultPrevented=(null!=a.defaultPrevented?a.defaultPrevented:!1===a.returnValue)?qt:Wt,this.isPropagationStopped=Wt,this}return T(t.prototype,{preventDefault:function(){this.defaultPrevented=!0;var e=this.nativeEvent;e&&(e.preventDefault?e.preventDefault():"unknown"!=typeof e.returnValue&&(e.returnValue=!1),this.isDefaultPrevented=qt)},stopPropagation:function(){var e=this.nativeEvent;e&&(e.stopPropagation?e.stopPropagation():"unknown"!=typeof e.cancelBubble&&(e.cancelBubble=!0),this.isPropagationStopped=qt)},persist:function(){},isPersistent:qt}),t}var Kt,Gt,Yt,Xt={eventPhase:0,bubbles:0,cancelable:0,timeStamp:function(e){return e.timeStamp||Date.now()},defaultPrevented:0,isTrusted:0},Zt=Qt(Xt),Jt=T({},Xt,{view:0,detail:0}),en=Qt(Jt),tn=T({},Jt,{screenX:0,screenY:0,clientX:0,clientY:0,pageX:0,pageY:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,getModifierState:pn,button:0,buttons:0,relatedTarget:function(e){return void 0===e.relatedTarget?e.fromElement===e.srcElement?e.toElement:e.fromElement:e.relatedTarget},movementX:function(e){return"movementX"in e?e.movementX:(e!==Yt&&(Yt&&"mousemove"===e.type?(Kt=e.screenX-Yt.screenX,Gt=e.screenY-Yt.screenY):Gt=Kt=0,Yt=e),Kt)},movementY:function(e){return"movementY"in e?e.movementY:Gt}}),nn=Qt(tn),rn=Qt(T({},tn,{dataTransfer:0})),an=Qt(T({},Jt,{relatedTarget:0})),ln=Qt(T({},Xt,{animationName:0,elapsedTime:0,pseudoElement:0})),on=Qt(T({},Xt,{clipboardData:function(e){return"clipboardData"in e?e.clipboardData:window.clipboardData}})),un=Qt(T({},Xt,{data:0})),sn={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},cn={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},fn={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function dn(e){var t=this.nativeEvent;return t.getModifierState?t.getModifierState(e):!!(e=fn[e])&&!!t[e]}function pn(){return dn}var hn=Qt(T({},Jt,{key:function(e){if(e.key){var t=sn[e.key]||e.key;if("Unidentified"!==t)return t}return"keypress"===e.type?13===(e=Bt(e))?"Enter":String.fromCharCode(e):"keydown"===e.type||"keyup"===e.type?cn[e.keyCode]||"Unidentified":""},code:0,location:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,repeat:0,locale:0,getModifierState:pn,charCode:function(e){return"keypress"===e.type?Bt(e):0},keyCode:function(e){return"keydown"===e.type||"keyup"===e.type?e.keyCode:0},which:function(e){return"keypress"===e.type?Bt(e):"keydown"===e.type||"keyup"===e.type?e.keyCode:0}})),mn=Qt(T({},tn,{pointerId:0,width:0,height:0,pressure:0,tangentialPressure:0,tiltX:0,tiltY:0,twist:0,pointerType:0,isPrimary:0})),vn=Qt(T({},Jt,{touches:0,targetTouches:0,changedTouches:0,altKey:0,metaKey:0,ctrlKey:0,shiftKey:0,getModifierState:pn})),gn=Qt(T({},Xt,{propertyName:0,elapsedTime:0,pseudoElement:0})),yn=Qt(T({},tn,{deltaX:function(e){return"deltaX"in e?e.deltaX:"wheelDeltaX"in e?-e.wheelDeltaX:0},deltaY:function(e){return"deltaY"in e?e.deltaY:"wheelDeltaY"in e?-e.wheelDeltaY:"wheelDelta"in e?-e.wheelDelta:0},deltaZ:0,deltaMode:0})),bn=Qt(T({},Xt,{newState:0,oldState:0})),kn=[9,13,27,32],wn=nt&&"CompositionEvent"in window,xn=null;nt&&"documentMode"in document&&(xn=document.documentMode);var Sn=nt&&"TextEvent"in window&&!xn,En=nt&&(!wn||xn&&8<xn&&11>=xn),Cn=String.fromCharCode(32),Nn=!1;function _n(e,t){switch(e){case"keyup":return-1!==kn.indexOf(t.keyCode);case"keydown":return 229!==t.keyCode;case"keypress":case"mousedown":case"focusout":return!0;default:return!1}}function Pn(e){return"object"==typeof(e=e.detail)&&"data"in e?e.data:null}var zn=!1,jn={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function Tn(e){var t=e&&e.nodeName&&e.nodeName.toLowerCase();return"input"===t?!!jn[e.type]:"textarea"===t}function On(e,t,n,r){Ot?Lt?Lt.push(r):Lt=[r]:Ot=r,0<(t=Rc(t,"onChange")).length&&(n=new Zt("onChange","change",null,n,r),e.push({event:n,listeners:t}))}var Ln=null,Mn=null;function An(e){Pc(e,0)}function Rn(e){if(dt(Ge(e)))return e}function Fn(e,t){if("change"===e)return t}var Dn=!1;if(nt){var In;if(nt){var Un="oninput"in document;if(!Un){var $n=document.createElement("div");$n.setAttribute("oninput","return;"),Un="function"==typeof $n.oninput}In=Un}else In=!1;Dn=In&&(!document.documentMode||9<document.documentMode)}function Hn(){Ln&&(Ln.detachEvent("onpropertychange",Vn),Mn=Ln=null)}function Vn(e){if("value"===e.propertyName&&Rn(Mn)){var t=[];On(t,Mn,e,Tt(e)),Rt(An,t)}}function Bn(e,t,n){"focusin"===e?(Hn(),Mn=n,(Ln=t).attachEvent("onpropertychange",Vn)):"focusout"===e&&Hn()}function qn(e){if("selectionchange"===e||"keyup"===e||"keydown"===e)return Rn(Mn)}function Wn(e,t){if("click"===e)return Rn(t)}function Qn(e,t){if("input"===e||"change"===e)return Rn(t)}var Kn="function"==typeof Object.is?Object.is:function(e,t){return e===t&&(0!==e||1/e==1/t)||e!=e&&t!=t};function Gn(e,t){if(Kn(e,t))return!0;if("object"!=typeof e||null===e||"object"!=typeof t||null===t)return!1;var n=Object.keys(e),r=Object.keys(t);if(n.length!==r.length)return!1;for(r=0;r<n.length;r++){var a=n[r];if(!re.call(t,a)||!Kn(e[a],t[a]))return!1}return!0}function Yn(e){for(;e&&e.firstChild;)e=e.firstChild;return e}function Xn(e,t){var n,r=Yn(e);for(e=0;r;){if(3===r.nodeType){if(n=e+r.textContent.length,e<=t&&n>=t)return{node:r,offset:t-e};e=n}e:{for(;r;){if(r.nextSibling){r=r.nextSibling;break e}r=r.parentNode}r=void 0}r=Yn(r)}}function Zn(e,t){return!(!e||!t)&&(e===t||(!e||3!==e.nodeType)&&(t&&3===t.nodeType?Zn(e,t.parentNode):"contains"in e?e.contains(t):!!e.compareDocumentPosition&&!!(16&e.compareDocumentPosition(t))))}function Jn(e){for(var t=pt((e=null!=e&&null!=e.ownerDocument&&null!=e.ownerDocument.defaultView?e.ownerDocument.defaultView:window).document);t instanceof e.HTMLIFrameElement;){try{var n="string"==typeof t.contentWindow.location.href}catch(e){n=!1}if(!n)break;t=pt((e=t.contentWindow).document)}return t}function er(e){var t=e&&e.nodeName&&e.nodeName.toLowerCase();return t&&("input"===t&&("text"===e.type||"search"===e.type||"tel"===e.type||"url"===e.type||"password"===e.type)||"textarea"===t||"true"===e.contentEditable)}function tr(e,t){var n=Jn(t);t=e.focusedElem;var r=e.selectionRange;if(n!==t&&t&&t.ownerDocument&&Zn(t.ownerDocument.documentElement,t)){if(null!==r&&er(t))if(e=r.start,void 0===(n=r.end)&&(n=e),"selectionStart"in t)t.selectionStart=e,t.selectionEnd=Math.min(n,t.value.length);else if((n=(e=t.ownerDocument||document)&&e.defaultView||window).getSelection){n=n.getSelection();var a=t.textContent.length,l=Math.min(r.start,a);r=void 0===r.end?l:Math.min(r.end,a),!n.extend&&l>r&&(a=r,r=l,l=a),a=Xn(t,l);var o=Xn(t,r);a&&o&&(1!==n.rangeCount||n.anchorNode!==a.node||n.anchorOffset!==a.offset||n.focusNode!==o.node||n.focusOffset!==o.offset)&&((e=e.createRange()).setStart(a.node,a.offset),n.removeAllRanges(),l>r?(n.addRange(e),n.extend(o.node,o.offset)):(e.setEnd(o.node,o.offset),n.addRange(e)))}for(e=[],n=t;n=n.parentNode;)1===n.nodeType&&e.push({element:n,left:n.scrollLeft,top:n.scrollTop});for("function"==typeof t.focus&&t.focus(),t=0;t<e.length;t++)(n=e[t]).element.scrollLeft=n.left,n.element.scrollTop=n.top}}var nr=nt&&"documentMode"in document&&11>=document.documentMode,rr=null,ar=null,lr=null,or=!1;function ir(e,t,n){var r=n.window===n?n.document:9===n.nodeType?n:n.ownerDocument;or||null==rr||rr!==pt(r)||(r="selectionStart"in(r=rr)&&er(r)?{start:r.selectionStart,end:r.selectionEnd}:{anchorNode:(r=(r.ownerDocument&&r.ownerDocument.defaultView||window).getSelection()).anchorNode,anchorOffset:r.anchorOffset,focusNode:r.focusNode,focusOffset:r.focusOffset},lr&&Gn(lr,r)||(lr=r,0<(r=Rc(ar,"onSelect")).length&&(t=new Zt("onSelect","select",null,t,n),e.push({event:t,listeners:r}),t.target=rr)))}function ur(e,t){var n={};return n[e.toLowerCase()]=t.toLowerCase(),n["Webkit"+e]="webkit"+t,n["Moz"+e]="moz"+t,n}var sr={animationend:ur("Animation","AnimationEnd"),animationiteration:ur("Animation","AnimationIteration"),animationstart:ur("Animation","AnimationStart"),transitionrun:ur("Transition","TransitionRun"),transitionstart:ur("Transition","TransitionStart"),transitioncancel:ur("Transition","TransitionCancel"),transitionend:ur("Transition","TransitionEnd")},cr={},fr={};function dr(e){if(cr[e])return cr[e];if(!sr[e])return e;var t,n=sr[e];for(t in n)if(n.hasOwnProperty(t)&&t in fr)return cr[e]=n[t];return e}nt&&(fr=document.createElement("div").style,"AnimationEvent"in window||(delete sr.animationend.animation,delete sr.animationiteration.animation,delete sr.animationstart.animation),"TransitionEvent"in window||delete sr.transitionend.transition);var pr=dr("animationend"),hr=dr("animationiteration"),mr=dr("animationstart"),vr=dr("transitionrun"),gr=dr("transitionstart"),yr=dr("transitioncancel"),br=dr("transitionend"),kr=new Map,wr="abort auxClick beforeToggle cancel canPlay canPlayThrough click close contextMenu copy cut drag dragEnd dragEnter dragExit dragLeave dragOver dragStart drop durationChange emptied encrypted ended error gotPointerCapture input invalid keyDown keyPress keyUp load loadedData loadedMetadata loadStart lostPointerCapture mouseDown mouseMove mouseOut mouseOver mouseUp paste pause play playing pointerCancel pointerDown pointerMove pointerOut pointerOver pointerUp progress rateChange reset resize seeked seeking stalled submit suspend timeUpdate touchCancel touchEnd touchStart volumeChange scroll scrollEnd toggle touchMove waiting wheel".split(" ");function xr(e,t){kr.set(e,t),et(t,[e])}var Sr=[],Er=0,Cr=0;function Nr(){for(var e=Er,t=Cr=Er=0;t<e;){var n=Sr[t];Sr[t++]=null;var r=Sr[t];Sr[t++]=null;var a=Sr[t];Sr[t++]=null;var l=Sr[t];if(Sr[t++]=null,null!==r&&null!==a){var o=r.pending;null===o?a.next=a:(a.next=o.next,o.next=a),r.pending=a}0!==l&&jr(n,a,l)}}function _r(e,t,n,r){Sr[Er++]=e,Sr[Er++]=t,Sr[Er++]=n,Sr[Er++]=r,Cr|=r,e.lanes|=r,null!==(e=e.alternate)&&(e.lanes|=r)}function Pr(e,t,n,r){return _r(e,t,n,r),Tr(e)}function zr(e,t){return _r(e,null,null,t),Tr(e)}function jr(e,t,n){e.lanes|=n;var r=e.alternate;null!==r&&(r.lanes|=n);for(var a=!1,l=e.return;null!==l;)l.childLanes|=n,null!==(r=l.alternate)&&(r.childLanes|=n),22===l.tag&&(null===(e=l.stateNode)||1&e._visibility||(a=!0)),e=l,l=l.return;a&&null!==t&&3===e.tag&&(l=e.stateNode,a=31-ke(n),null===(e=(l=l.hiddenUpdates)[a])?l[a]=[t]:e.push(t),t.lane=536870912|n)}function Tr(e){if(50<Ps)throw Ps=0,zs=null,Error(o(185));for(var t=e.return;null!==t;)t=(e=t).return;return 3===e.tag?e.stateNode:null}var Or={},Lr=new WeakMap;function Mr(e,t){if("object"==typeof e&&null!==e){var n=Lr.get(e);return void 0!==n?n:(t={value:e,source:t,stack:R(t)},Lr.set(e,t),t)}return{value:e,source:t,stack:R(t)}}var Ar=[],Rr=0,Fr=null,Dr=0,Ir=[],Ur=0,$r=null,Hr=1,Vr="";function Br(e,t){Ar[Rr++]=Dr,Ar[Rr++]=Fr,Fr=e,Dr=t}function qr(e,t,n){Ir[Ur++]=Hr,Ir[Ur++]=Vr,Ir[Ur++]=$r,$r=e;var r=Hr;e=Vr;var a=32-ke(r)-1;r&=~(1<<a),n+=1;var l=32-ke(t)+a;if(30<l){var o=a-a%5;l=(r&(1<<o)-1).toString(32),r>>=o,a-=o,Hr=1<<32-ke(t)+a|n<<a|r,Vr=l+e}else Hr=1<<l|n<<a|r,Vr=e}function Wr(e){null!==e.return&&(Br(e,1),qr(e,1,0))}function Qr(e){for(;e===Fr;)Fr=Ar[--Rr],Ar[Rr]=null,Dr=Ar[--Rr],Ar[Rr]=null;for(;e===$r;)$r=Ir[--Ur],Ir[Ur]=null,Vr=Ir[--Ur],Ir[Ur]=null,Hr=Ir[--Ur],Ir[Ur]=null}var Kr=null,Gr=null,Yr=!1,Xr=null,Zr=!1,Jr=Error(o(519));function ea(e){throw la(Mr(Error(o(418,"")),e)),Jr}function ta(e){var t=e.stateNode,n=e.type,r=e.memoizedProps;switch(t[De]=e,t[Ie]=r,n){case"dialog":zc("cancel",t),zc("close",t);break;case"iframe":case"object":case"embed":zc("load",t);break;case"video":case"audio":for(n=0;n<Nc.length;n++)zc(Nc[n],t);break;case"source":zc("error",t);break;case"img":case"image":case"link":zc("error",t),zc("load",t);break;case"details":zc("toggle",t);break;case"input":zc("invalid",t),gt(t,r.value,r.defaultValue,r.checked,r.defaultChecked,r.type,r.name,!0),ft(t);break;case"select":zc("invalid",t);break;case"textarea":zc("invalid",t),wt(t,r.value,r.defaultValue,r.children),ft(t)}"string"!=typeof(n=r.children)&&"number"!=typeof n&&"bigint"!=typeof n||t.textContent===""+n||!0===r.suppressHydrationWarning||Hc(t.textContent,n)?(null!=r.popover&&(zc("beforetoggle",t),zc("toggle",t)),null!=r.onScroll&&zc("scroll",t),null!=r.onScrollEnd&&zc("scrollend",t),null!=r.onClick&&(t.onclick=Vc),t=!0):t=!1,t||ea(e)}function na(e){for(Kr=e.return;Kr;)switch(Kr.tag){case 3:case 27:return void(Zr=!0);case 5:case 13:return void(Zr=!1);default:Kr=Kr.return}}function ra(e){if(e!==Kr)return!1;if(!Yr)return na(e),Yr=!0,!1;var t,n=!1;if((t=3!==e.tag&&27!==e.tag)&&((t=5===e.tag)&&(t=!("form"!==(t=e.type)&&"button"!==t)||Zc(e.type,e.memoizedProps)),t=!t),t&&(n=!0),n&&Gr&&ea(e),na(e),13===e.tag){if(!(e=null!==(e=e.memoizedState)?e.dehydrated:null))throw Error(o(317));e:{for(e=e.nextSibling,n=0;e;){if(8===e.nodeType)if("/$"===(t=e.data)){if(0===n){Gr=uf(e.nextSibling);break e}n--}else"$"!==t&&"$!"!==t&&"$?"!==t||n++;e=e.nextSibling}Gr=null}}else Gr=Kr?uf(e.stateNode.nextSibling):null;return!0}function aa(){Gr=Kr=null,Yr=!1}function la(e){null===Xr?Xr=[e]:Xr.push(e)}var oa=Error(o(460)),ia=Error(o(474)),ua={then:function(){}};function sa(e){return"fulfilled"===(e=e.status)||"rejected"===e}function ca(){}function fa(e,t,n){switch(void 0===(n=e[n])?e.push(t):n!==t&&(t.then(ca,ca),t=n),t.status){case"fulfilled":return t.value;case"rejected":if((e=t.reason)===oa)throw Error(o(483));throw e;default:if("string"==typeof t.status)t.then(ca,ca);else{if(null!==(e=ts)&&100<e.shellSuspendCounter)throw Error(o(482));(e=t).status="pending",e.then((function(e){if("pending"===t.status){var n=t;n.status="fulfilled",n.value=e}}),(function(e){if("pending"===t.status){var n=t;n.status="rejected",n.reason=e}}))}switch(t.status){case"fulfilled":return t.value;case"rejected":if((e=t.reason)===oa)throw Error(o(483));throw e}throw da=t,oa}}var da=null;function pa(){if(null===da)throw Error(o(459));var e=da;return da=null,e}var ha=null,ma=0;function va(e){var t=ma;return ma+=1,null===ha&&(ha=[]),fa(ha,e,t)}function ga(e,t){t=t.props.ref,e.ref=void 0!==t?t:null}function ya(e,t){if(t.$$typeof===u)throw Error(o(525));throw e=Object.prototype.toString.call(t),Error(o(31,"[object Object]"===e?"object with keys {"+Object.keys(t).join(", ")+"}":e))}function ba(e){return(0,e._init)(e._payload)}function ka(e){function t(t,n){if(e){var r=t.deletions;null===r?(t.deletions=[n],t.flags|=16):r.push(n)}}function n(n,r){if(!e)return null;for(;null!==r;)t(n,r),r=r.sibling;return null}function r(e){for(var t=new Map;null!==e;)null!==e.key?t.set(e.key,e):t.set(e.index,e),e=e.sibling;return t}function a(e,t){return(e=Fu(e,t)).index=0,e.sibling=null,e}function l(t,n,r){return t.index=r,e?null!==(r=t.alternate)?(r=r.index)<n?(t.flags|=33554434,n):r:(t.flags|=33554434,n):(t.flags|=1048576,n)}function i(t){return e&&null===t.alternate&&(t.flags|=33554434),t}function u(e,t,n,r){return null===t||6!==t.tag?((t=Hu(n,e.mode,r)).return=e,t):((t=a(t,n)).return=e,t)}function d(e,t,n,r){var l=n.type;return l===f?h(e,t,n.props.children,r,n.key):null!==t&&(t.elementType===l||"object"==typeof l&&null!==l&&l.$$typeof===w&&ba(l)===t.type)?(ga(t=a(t,n.props),n),t.return=e,t):(ga(t=Iu(n.type,n.key,n.props,null,e.mode,r),n),t.return=e,t)}function p(e,t,n,r){return null===t||4!==t.tag||t.stateNode.containerInfo!==n.containerInfo||t.stateNode.implementation!==n.implementation?((t=Vu(n,e.mode,r)).return=e,t):((t=a(t,n.children||[])).return=e,t)}function h(e,t,n,r,l){return null===t||7!==t.tag?((t=Uu(n,e.mode,r,l)).return=e,t):((t=a(t,n)).return=e,t)}function m(e,t,n){if("string"==typeof t&&""!==t||"number"==typeof t||"bigint"==typeof t)return(t=Hu(""+t,e.mode,n)).return=e,t;if("object"==typeof t&&null!==t){switch(t.$$typeof){case s:return ga(n=Iu(t.type,t.key,t.props,null,e.mode,n),t),n.return=e,n;case c:return(t=Vu(t,e.mode,n)).return=e,t;case w:return m(e,t=(0,t._init)(t._payload),n)}if($(t)||C(t))return(t=Uu(t,e.mode,n,null)).return=e,t;if("function"==typeof t.then)return m(e,va(t),n);if(t.$$typeof===v)return m(e,Ni(e,t),n);ya(e,t)}return null}function g(e,t,n,r){var a=null!==t?t.key:null;if("string"==typeof n&&""!==n||"number"==typeof n||"bigint"==typeof n)return null!==a?null:u(e,t,""+n,r);if("object"==typeof n&&null!==n){switch(n.$$typeof){case s:return n.key===a?d(e,t,n,r):null;case c:return n.key===a?p(e,t,n,r):null;case w:return g(e,t,n=(a=n._init)(n._payload),r)}if($(n)||C(n))return null!==a?null:h(e,t,n,r,null);if("function"==typeof n.then)return g(e,t,va(n),r);if(n.$$typeof===v)return g(e,t,Ni(e,n),r);ya(e,n)}return null}function y(e,t,n,r,a){if("string"==typeof r&&""!==r||"number"==typeof r||"bigint"==typeof r)return u(t,e=e.get(n)||null,""+r,a);if("object"==typeof r&&null!==r){switch(r.$$typeof){case s:return d(t,e=e.get(null===r.key?n:r.key)||null,r,a);case c:return p(t,e=e.get(null===r.key?n:r.key)||null,r,a);case w:return y(e,t,n,r=(0,r._init)(r._payload),a)}if($(r)||C(r))return h(t,e=e.get(n)||null,r,a,null);if("function"==typeof r.then)return y(e,t,n,va(r),a);if(r.$$typeof===v)return y(e,t,n,Ni(t,r),a);ya(t,r)}return null}function b(u,d,p,h){if("object"==typeof p&&null!==p&&p.type===f&&null===p.key&&(p=p.props.children),"object"==typeof p&&null!==p){switch(p.$$typeof){case s:e:{for(var k=p.key;null!==d;){if(d.key===k){if((k=p.type)===f){if(7===d.tag){n(u,d.sibling),(h=a(d,p.props.children)).return=u,u=h;break e}}else if(d.elementType===k||"object"==typeof k&&null!==k&&k.$$typeof===w&&ba(k)===d.type){n(u,d.sibling),ga(h=a(d,p.props),p),h.return=u,u=h;break e}n(u,d);break}t(u,d),d=d.sibling}p.type===f?((h=Uu(p.props.children,u.mode,h,p.key)).return=u,u=h):(ga(h=Iu(p.type,p.key,p.props,null,u.mode,h),p),h.return=u,u=h)}return i(u);case c:e:{for(k=p.key;null!==d;){if(d.key===k){if(4===d.tag&&d.stateNode.containerInfo===p.containerInfo&&d.stateNode.implementation===p.implementation){n(u,d.sibling),(h=a(d,p.children||[])).return=u,u=h;break e}n(u,d);break}t(u,d),d=d.sibling}(h=Vu(p,u.mode,h)).return=u,u=h}return i(u);case w:return b(u,d,p=(k=p._init)(p._payload),h)}if($(p))return function(a,o,i,u){for(var s=null,c=null,f=o,d=o=0,p=null;null!==f&&d<i.length;d++){f.index>d?(p=f,f=null):p=f.sibling;var h=g(a,f,i[d],u);if(null===h){null===f&&(f=p);break}e&&f&&null===h.alternate&&t(a,f),o=l(h,o,d),null===c?s=h:c.sibling=h,c=h,f=p}if(d===i.length)return n(a,f),Yr&&Br(a,d),s;if(null===f){for(;d<i.length;d++)null!==(f=m(a,i[d],u))&&(o=l(f,o,d),null===c?s=f:c.sibling=f,c=f);return Yr&&Br(a,d),s}for(f=r(f);d<i.length;d++)null!==(p=y(f,a,d,i[d],u))&&(e&&null!==p.alternate&&f.delete(null===p.key?d:p.key),o=l(p,o,d),null===c?s=p:c.sibling=p,c=p);return e&&f.forEach((function(e){return t(a,e)})),Yr&&Br(a,d),s}(u,d,p,h);if(C(p)){if("function"!=typeof(k=C(p)))throw Error(o(150));return function(a,i,u,s){if(null==u)throw Error(o(151));for(var c=null,f=null,d=i,p=i=0,h=null,v=u.next();null!==d&&!v.done;p++,v=u.next()){d.index>p?(h=d,d=null):h=d.sibling;var b=g(a,d,v.value,s);if(null===b){null===d&&(d=h);break}e&&d&&null===b.alternate&&t(a,d),i=l(b,i,p),null===f?c=b:f.sibling=b,f=b,d=h}if(v.done)return n(a,d),Yr&&Br(a,p),c;if(null===d){for(;!v.done;p++,v=u.next())null!==(v=m(a,v.value,s))&&(i=l(v,i,p),null===f?c=v:f.sibling=v,f=v);return Yr&&Br(a,p),c}for(d=r(d);!v.done;p++,v=u.next())null!==(v=y(d,a,p,v.value,s))&&(e&&null!==v.alternate&&d.delete(null===v.key?p:v.key),i=l(v,i,p),null===f?c=v:f.sibling=v,f=v);return e&&d.forEach((function(e){return t(a,e)})),Yr&&Br(a,p),c}(u,d,p=k.call(p),h)}if("function"==typeof p.then)return b(u,d,va(p),h);if(p.$$typeof===v)return b(u,d,Ni(u,p),h);ya(u,p)}return"string"==typeof p&&""!==p||"number"==typeof p||"bigint"==typeof p?(p=""+p,null!==d&&6===d.tag?(n(u,d.sibling),(h=a(d,p)).return=u,u=h):(n(u,d),(h=Hu(p,u.mode,h)).return=u,u=h),i(u)):n(u,d)}return function(e,t,n,r){try{ma=0;var a=b(e,t,n,r);return ha=null,a}catch(t){if(t===oa)throw t;var l=Au(29,t,null,e.mode);return l.lanes=r,l.return=e,l}}}var wa=ka(!0),xa=ka(!1),Sa=W(null),Ea=W(0);function Ca(e,t){K(Ea,e=ss),K(Sa,t),ss=e|t.baseLanes}function Na(){K(Ea,ss),K(Sa,Sa.current)}function _a(){ss=Ea.current,Q(Sa),Q(Ea)}var Pa=W(null),za=null;function ja(e){var t=e.alternate;K(Ma,1&Ma.current),K(Pa,e),null===za&&(null===t||null!==Sa.current||null!==t.memoizedState)&&(za=e)}function Ta(e){if(22===e.tag){if(K(Ma,Ma.current),K(Pa,e),null===za){var t=e.alternate;null!==t&&null!==t.memoizedState&&(za=e)}}else Oa()}function Oa(){K(Ma,Ma.current),K(Pa,Pa.current)}function La(e){Q(Pa),za===e&&(za=null),Q(Ma)}var Ma=W(0);function Aa(e){for(var t=e;null!==t;){if(13===t.tag){var n=t.memoizedState;if(null!==n&&(null===(n=n.dehydrated)||"$?"===n.data||"$!"===n.data))return t}else if(19===t.tag&&void 0!==t.memoizedProps.revealOrder){if(128&t.flags)return t}else if(null!==t.child){t.child.return=t,t=t.child;continue}if(t===e)break;for(;null===t.sibling;){if(null===t.return||t.return===e)return null;t=t.return}t.sibling.return=t.return,t=t.sibling}return null}var Ra="undefined"!=typeof AbortController?AbortController:function(){var e=[],t=this.signal={aborted:!1,addEventListener:function(t,n){e.push(n)}};this.abort=function(){t.aborted=!0,e.forEach((function(e){return e()}))}},Fa=r.unstable_scheduleCallback,Da=r.unstable_NormalPriority,Ia={$$typeof:v,Consumer:null,Provider:null,_currentValue:null,_currentValue2:null,_threadCount:0};function Ua(){return{controller:new Ra,data:new Map,refCount:0}}function $a(e){e.refCount--,0===e.refCount&&Fa(Da,(function(){e.controller.abort()}))}var Ha=null,Va=0,Ba=0,qa=null;function Wa(){if(0==--Va&&null!==Ha){null!==qa&&(qa.status="fulfilled");var e=Ha;Ha=null,Ba=0,qa=null;for(var t=0;t<e.length;t++)(0,e[t])()}}var Qa=j.S;j.S=function(e,t){"object"==typeof t&&null!==t&&"function"==typeof t.then&&function(e,t){if(null===Ha){var n=Ha=[];Va=0,Ba=wc(),qa={status:"pending",value:void 0,then:function(e){n.push(e)}}}Va++,t.then(Wa,Wa)}(0,t),null!==Qa&&Qa(e,t)};var Ka=W(null);function Ga(){var e=Ka.current;return null!==e?e:ts.pooledCache}function Ya(e,t){K(Ka,null===t?Ka.current:t.pool)}function Xa(){var e=Ga();return null===e?null:{parent:Ia._currentValue,pool:e}}var Za=0,Ja=null,el=null,tl=null,nl=!1,rl=!1,al=!1,ll=0,ol=0,il=null,ul=0;function sl(){throw Error(o(321))}function cl(e,t){if(null===t)return!1;for(var n=0;n<t.length&&n<e.length;n++)if(!Kn(e[n],t[n]))return!1;return!0}function fl(e,t,n,r,a,l){return Za=l,Ja=t,t.memoizedState=null,t.updateQueue=null,t.lanes=0,j.H=null===e||null===e.memoizedState?No:_o,al=!1,l=n(r,a),al=!1,rl&&(l=pl(t,n,r,a)),dl(e),l}function dl(e){j.H=Co;var t=null!==el&&null!==el.next;if(Za=0,tl=el=Ja=null,nl=!1,ol=0,il=null,t)throw Error(o(300));null===e||Bo||null!==(e=e.dependencies)&&Si(e)&&(Bo=!0)}function pl(e,t,n,r){Ja=e;var a=0;do{if(rl&&(il=null),ol=0,rl=!1,25<=a)throw Error(o(301));if(a+=1,tl=el=null,null!=e.updateQueue){var l=e.updateQueue;l.lastEffect=null,l.events=null,l.stores=null,null!=l.memoCache&&(l.memoCache.index=0)}j.H=Po,l=t(n,r)}while(rl);return l}function hl(){var e=j.H,t=e.useState()[0];return t="function"==typeof t.then?kl(t):t,e=e.useState()[0],(null!==el?el.memoizedState:null)!==e&&(Ja.flags|=1024),t}function ml(){var e=0!==ll;return ll=0,e}function vl(e,t,n){t.updateQueue=e.updateQueue,t.flags&=-2053,e.lanes&=~n}function gl(e){if(nl){for(e=e.memoizedState;null!==e;){var t=e.queue;null!==t&&(t.pending=null),e=e.next}nl=!1}Za=0,tl=el=Ja=null,rl=!1,ol=ll=0,il=null}function yl(){var e={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return null===tl?Ja.memoizedState=tl=e:tl=tl.next=e,tl}function bl(){if(null===el){var e=Ja.alternate;e=null!==e?e.memoizedState:null}else e=el.next;var t=null===tl?Ja.memoizedState:tl.next;if(null!==t)tl=t,el=e;else{if(null===e){if(null===Ja.alternate)throw Error(o(467));throw Error(o(310))}e={memoizedState:(el=e).memoizedState,baseState:el.baseState,baseQueue:el.baseQueue,queue:el.queue,next:null},null===tl?Ja.memoizedState=tl=e:tl=tl.next=e}return tl}function kl(e){var t=ol;return ol+=1,null===il&&(il=[]),e=fa(il,e,t),t=Ja,null===(null===tl?t.memoizedState:tl.next)&&(t=t.alternate,j.H=null===t||null===t.memoizedState?No:_o),e}function wl(e){if(null!==e&&"object"==typeof e){if("function"==typeof e.then)return kl(e);if(e.$$typeof===v)return Ci(e)}throw Error(o(438,String(e)))}function xl(e){var t=null,n=Ja.updateQueue;if(null!==n&&(t=n.memoCache),null==t){var r=Ja.alternate;null!==r&&null!==(r=r.updateQueue)&&null!=(r=r.memoCache)&&(t={data:r.data.map((function(e){return e.slice()})),index:0})}if(null==t&&(t={data:[],index:0}),null===n&&(n={lastEffect:null,events:null,stores:null,memoCache:null},Ja.updateQueue=n),n.memoCache=t,void 0===(n=t.data[t.index]))for(n=t.data[t.index]=Array(e),r=0;r<e;r++)n[r]=S;return t.index++,n}function Sl(e,t){return"function"==typeof t?t(e):t}function El(e){return Cl(bl(),el,e)}function Cl(e,t,n){var r=e.queue;if(null===r)throw Error(o(311));r.lastRenderedReducer=n;var a=e.baseQueue,l=r.pending;if(null!==l){if(null!==a){var i=a.next;a.next=l.next,l.next=i}t.baseQueue=a=l,r.pending=null}if(l=e.baseState,null===a)e.memoizedState=l;else{var u=i=null,s=null,c=t=a.next,f=!1;do{var d=-536870913&c.lane;if(d!==c.lane?(rs&d)===d:(Za&d)===d){var p=c.revertLane;if(0===p)null!==s&&(s=s.next={lane:0,revertLane:0,action:c.action,hasEagerState:c.hasEagerState,eagerState:c.eagerState,next:null}),d===Ba&&(f=!0);else{if((Za&p)===p){c=c.next,p===Ba&&(f=!0);continue}d={lane:0,revertLane:c.revertLane,action:c.action,hasEagerState:c.hasEagerState,eagerState:c.eagerState,next:null},null===s?(u=s=d,i=l):s=s.next=d,Ja.lanes|=p,fs|=p}d=c.action,al&&n(l,d),l=c.hasEagerState?c.eagerState:n(l,d)}else p={lane:d,revertLane:c.revertLane,action:c.action,hasEagerState:c.hasEagerState,eagerState:c.eagerState,next:null},null===s?(u=s=p,i=l):s=s.next=p,Ja.lanes|=d,fs|=d;c=c.next}while(null!==c&&c!==t);if(null===s?i=l:s.next=u,!Kn(l,e.memoizedState)&&(Bo=!0,f&&null!==(n=qa)))throw n;e.memoizedState=l,e.baseState=i,e.baseQueue=s,r.lastRenderedState=l}return null===a&&(r.lanes=0),[e.memoizedState,r.dispatch]}function Nl(e){var t=bl(),n=t.queue;if(null===n)throw Error(o(311));n.lastRenderedReducer=e;var r=n.dispatch,a=n.pending,l=t.memoizedState;if(null!==a){n.pending=null;var i=a=a.next;do{l=e(l,i.action),i=i.next}while(i!==a);Kn(l,t.memoizedState)||(Bo=!0),t.memoizedState=l,null===t.baseQueue&&(t.baseState=l),n.lastRenderedState=l}return[l,r]}function _l(e,t,n){var r=Ja,a=bl(),l=Yr;if(l){if(void 0===n)throw Error(o(407));n=n()}else n=t();var i=!Kn((el||a).memoizedState,n);if(i&&(a.memoizedState=n,Bo=!0),a=a.queue,Zl(jl.bind(null,r,a,e),[e]),a.getSnapshot!==t||i||null!==tl&&1&tl.memoizedState.tag){if(r.flags|=2048,Ql(9,zl.bind(null,r,a,n,t),{destroy:void 0},null),null===ts)throw Error(o(349));l||60&Za||Pl(r,t,n)}return n}function Pl(e,t,n){e.flags|=16384,e={getSnapshot:t,value:n},null===(t=Ja.updateQueue)?(t={lastEffect:null,events:null,stores:null,memoCache:null},Ja.updateQueue=t,t.stores=[e]):null===(n=t.stores)?t.stores=[e]:n.push(e)}function zl(e,t,n,r){t.value=n,t.getSnapshot=r,Tl(t)&&Ol(e)}function jl(e,t,n){return n((function(){Tl(t)&&Ol(e)}))}function Tl(e){var t=e.getSnapshot;e=e.value;try{var n=t();return!Kn(e,n)}catch(e){return!0}}function Ol(e){var t=zr(e,2);null!==t&&Os(t,0,2)}function Ll(e){var t=yl();if("function"==typeof e){var n=e;if(e=n(),al){be(!0);try{n()}finally{be(!1)}}}return t.memoizedState=t.baseState=e,t.queue={pending:null,lanes:0,dispatch:null,lastRenderedReducer:Sl,lastRenderedState:e},t}function Ml(e,t,n,r){return e.baseState=n,Cl(e,el,"function"==typeof r?r:Sl)}function Al(e,t,n,r,a){if(xo(e))throw Error(o(485));if(null!==(e=t.action)){var l={payload:a,action:e,next:null,isTransition:!0,status:"pending",value:null,reason:null,listeners:[],then:function(e){l.listeners.push(e)}};null!==j.T?n(!0):l.isTransition=!1,r(l),null===(n=t.pending)?(l.next=t.pending=l,Rl(t,l)):(l.next=n.next,t.pending=n.next=l)}}function Rl(e,t){var n=t.action,r=t.payload,a=e.state;if(t.isTransition){var l=j.T,o={};j.T=o;try{var i=n(a,r),u=j.S;null!==u&&u(o,i),Fl(e,t,i)}catch(n){Il(e,t,n)}finally{j.T=l}}else try{Fl(e,t,l=n(a,r))}catch(n){Il(e,t,n)}}function Fl(e,t,n){null!==n&&"object"==typeof n&&"function"==typeof n.then?n.then((function(n){Dl(e,t,n)}),(function(n){return Il(e,t,n)})):Dl(e,t,n)}function Dl(e,t,n){t.status="fulfilled",t.value=n,Ul(t),e.state=n,null!==(t=e.pending)&&((n=t.next)===t?e.pending=null:(n=n.next,t.next=n,Rl(e,n)))}function Il(e,t,n){var r=e.pending;if(e.pending=null,null!==r){r=r.next;do{t.status="rejected",t.reason=n,Ul(t),t=t.next}while(t!==r)}e.action=null}function Ul(e){e=e.listeners;for(var t=0;t<e.length;t++)(0,e[t])()}function $l(e,t){return t}function Hl(e,t){if(Yr){var n=ts.formState;if(null!==n){e:{var r=Ja;if(Yr){if(Gr){t:{for(var a=Gr,l=Zr;8!==a.nodeType;){if(!l){a=null;break t}if(null===(a=uf(a.nextSibling))){a=null;break t}}a="F!"===(l=a.data)||"F"===l?a:null}if(a){Gr=uf(a.nextSibling),r="F!"===a.data;break e}}ea(r)}r=!1}r&&(t=n[0])}}return(n=yl()).memoizedState=n.baseState=t,r={pending:null,lanes:0,dispatch:null,lastRenderedReducer:$l,lastRenderedState:t},n.queue=r,n=bo.bind(null,Ja,r),r.dispatch=n,r=Ll(!1),l=wo.bind(null,Ja,!1,r.queue),a={state:t,dispatch:null,action:e,pending:null},(r=yl()).queue=a,n=Al.bind(null,Ja,a,l,n),a.dispatch=n,r.memoizedState=e,[t,n,!1]}function Vl(e){return Bl(bl(),el,e)}function Bl(e,t,n){t=Cl(e,t,$l)[0],e=El(Sl)[0],t="object"==typeof t&&null!==t&&"function"==typeof t.then?kl(t):t;var r=bl(),a=r.queue,l=a.dispatch;return n!==r.memoizedState&&(Ja.flags|=2048,Ql(9,ql.bind(null,a,n),{destroy:void 0},null)),[t,l,e]}function ql(e,t){e.action=t}function Wl(e){var t=bl(),n=el;if(null!==n)return Bl(t,n,e);bl(),t=t.memoizedState;var r=(n=bl()).queue.dispatch;return n.memoizedState=e,[t,r,!1]}function Ql(e,t,n,r){return e={tag:e,create:t,inst:n,deps:r,next:null},null===(t=Ja.updateQueue)&&(t={lastEffect:null,events:null,stores:null,memoCache:null},Ja.updateQueue=t),null===(n=t.lastEffect)?t.lastEffect=e.next=e:(r=n.next,n.next=e,e.next=r,t.lastEffect=e),e}function Kl(){return bl().memoizedState}function Gl(e,t,n,r){var a=yl();Ja.flags|=e,a.memoizedState=Ql(1|t,n,{destroy:void 0},void 0===r?null:r)}function Yl(e,t,n,r){var a=bl();r=void 0===r?null:r;var l=a.memoizedState.inst;null!==el&&null!==r&&cl(r,el.memoizedState.deps)?a.memoizedState=Ql(t,n,l,r):(Ja.flags|=e,a.memoizedState=Ql(1|t,n,l,r))}function Xl(e,t){Gl(8390656,8,e,t)}function Zl(e,t){Yl(2048,8,e,t)}function Jl(e,t){return Yl(4,2,e,t)}function eo(e,t){return Yl(4,4,e,t)}function to(e,t){if("function"==typeof t){e=e();var n=t(e);return function(){"function"==typeof n?n():t(null)}}if(null!=t)return e=e(),t.current=e,function(){t.current=null}}function no(e,t,n){n=null!=n?n.concat([e]):null,Yl(4,4,to.bind(null,t,e),n)}function ro(){}function ao(e,t){var n=bl();t=void 0===t?null:t;var r=n.memoizedState;return null!==t&&cl(t,r[1])?r[0]:(n.memoizedState=[e,t],e)}function lo(e,t){var n=bl();t=void 0===t?null:t;var r=n.memoizedState;if(null!==t&&cl(t,r[1]))return r[0];if(r=e(),al){be(!0);try{e()}finally{be(!1)}}return n.memoizedState=[r,t],r}function oo(e,t,n){return void 0===n||1073741824&Za?e.memoizedState=t:(e.memoizedState=n,e=Ts(),Ja.lanes|=e,fs|=e,n)}function io(e,t,n,r){return Kn(n,t)?n:null!==Sa.current?(e=oo(e,n,r),Kn(e,t)||(Bo=!0),e):42&Za?(e=Ts(),Ja.lanes|=e,fs|=e,t):(Bo=!0,e.memoizedState=n)}function uo(e,t,n,r,a){var l=H.p;H.p=0!==l&&8>l?l:8;var o,i,u,s=j.T,c={};j.T=c,wo(e,!1,t,n);try{var f=a(),d=j.S;null!==d&&d(c,f),null!==f&&"object"==typeof f&&"function"==typeof f.then?ko(e,t,(o=r,i=[],u={status:"pending",value:null,reason:null,then:function(e){i.push(e)}},f.then((function(){u.status="fulfilled",u.value=o;for(var e=0;e<i.length;e++)(0,i[e])(o)}),(function(e){for(u.status="rejected",u.reason=e,e=0;e<i.length;e++)(0,i[e])(void 0)})),u),js()):ko(e,t,r,js())}catch(n){ko(e,t,{then:function(){},status:"rejected",reason:n},js())}finally{H.p=l,j.T=s}}function so(){}function co(e,t,n,r){if(5!==e.tag)throw Error(o(476));var a=fo(e).queue;uo(e,a,t,V,null===n?so:function(){return po(e),n(r)})}function fo(e){var t=e.memoizedState;if(null!==t)return t;var n={};return(t={memoizedState:V,baseState:V,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:Sl,lastRenderedState:V},next:null}).next={memoizedState:n,baseState:n,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:Sl,lastRenderedState:n},next:null},e.memoizedState=t,null!==(e=e.alternate)&&(e.memoizedState=t),t}function po(e){ko(e,fo(e).next.queue,{},js())}function ho(){return Ci(Ff)}function mo(){return bl().memoizedState}function vo(){return bl().memoizedState}function go(e){for(var t=e.return;null!==t;){switch(t.tag){case 24:case 3:var n=js(),r=Oi(t,e=Ti(n),n);return null!==r&&(Os(r,0,n),Li(r,t,n)),t={cache:Ua()},void(e.payload=t)}t=t.return}}function yo(e,t,n){var r=js();n={lane:r,revertLane:0,action:n,hasEagerState:!1,eagerState:null,next:null},xo(e)?So(t,n):null!==(n=Pr(e,t,n,r))&&(Os(n,0,r),Eo(n,t,r))}function bo(e,t,n){ko(e,t,n,js())}function ko(e,t,n,r){var a={lane:r,revertLane:0,action:n,hasEagerState:!1,eagerState:null,next:null};if(xo(e))So(t,a);else{var l=e.alternate;if(0===e.lanes&&(null===l||0===l.lanes)&&null!==(l=t.lastRenderedReducer))try{var o=t.lastRenderedState,i=l(o,n);if(a.hasEagerState=!0,a.eagerState=i,Kn(i,o))return _r(e,t,a,0),null===ts&&Nr(),!1}catch(e){}if(null!==(n=Pr(e,t,a,r)))return Os(n,0,r),Eo(n,t,r),!0}return!1}function wo(e,t,n,r){if(r={lane:2,revertLane:wc(),action:r,hasEagerState:!1,eagerState:null,next:null},xo(e)){if(t)throw Error(o(479))}else null!==(t=Pr(e,n,r,2))&&Os(t,0,2)}function xo(e){var t=e.alternate;return e===Ja||null!==t&&t===Ja}function So(e,t){rl=nl=!0;var n=e.pending;null===n?t.next=t:(t.next=n.next,n.next=t),e.pending=t}function Eo(e,t,n){if(4194176&n){var r=t.lanes;n|=r&=e.pendingLanes,t.lanes=n,Me(e,n)}}var Co={readContext:Ci,use:wl,useCallback:sl,useContext:sl,useEffect:sl,useImperativeHandle:sl,useLayoutEffect:sl,useInsertionEffect:sl,useMemo:sl,useReducer:sl,useRef:sl,useState:sl,useDebugValue:sl,useDeferredValue:sl,useTransition:sl,useSyncExternalStore:sl,useId:sl};Co.useCacheRefresh=sl,Co.useMemoCache=sl,Co.useHostTransitionStatus=sl,Co.useFormState=sl,Co.useActionState=sl,Co.useOptimistic=sl;var No={readContext:Ci,use:wl,useCallback:function(e,t){return yl().memoizedState=[e,void 0===t?null:t],e},useContext:Ci,useEffect:Xl,useImperativeHandle:function(e,t,n){n=null!=n?n.concat([e]):null,Gl(4194308,4,to.bind(null,t,e),n)},useLayoutEffect:function(e,t){return Gl(4194308,4,e,t)},useInsertionEffect:function(e,t){Gl(4,2,e,t)},useMemo:function(e,t){var n=yl();t=void 0===t?null:t;var r=e();if(al){be(!0);try{e()}finally{be(!1)}}return n.memoizedState=[r,t],r},useReducer:function(e,t,n){var r=yl();if(void 0!==n){var a=n(t);if(al){be(!0);try{n(t)}finally{be(!1)}}}else a=t;return r.memoizedState=r.baseState=a,e={pending:null,lanes:0,dispatch:null,lastRenderedReducer:e,lastRenderedState:a},r.queue=e,e=e.dispatch=yo.bind(null,Ja,e),[r.memoizedState,e]},useRef:function(e){return e={current:e},yl().memoizedState=e},useState:function(e){var t=(e=Ll(e)).queue,n=bo.bind(null,Ja,t);return t.dispatch=n,[e.memoizedState,n]},useDebugValue:ro,useDeferredValue:function(e,t){return oo(yl(),e,t)},useTransition:function(){var e=Ll(!1);return e=uo.bind(null,Ja,e.queue,!0,!1),yl().memoizedState=e,[!1,e]},useSyncExternalStore:function(e,t,n){var r=Ja,a=yl();if(Yr){if(void 0===n)throw Error(o(407));n=n()}else{if(n=t(),null===ts)throw Error(o(349));60&rs||Pl(r,t,n)}a.memoizedState=n;var l={value:n,getSnapshot:t};return a.queue=l,Xl(jl.bind(null,r,l,e),[e]),r.flags|=2048,Ql(9,zl.bind(null,r,l,n,t),{destroy:void 0},null),n},useId:function(){var e=yl(),t=ts.identifierPrefix;if(Yr){var n=Vr;t=":"+t+"R"+(n=(Hr&~(1<<32-ke(Hr)-1)).toString(32)+n),0<(n=ll++)&&(t+="H"+n.toString(32)),t+=":"}else t=":"+t+"r"+(n=ul++).toString(32)+":";return e.memoizedState=t},useCacheRefresh:function(){return yl().memoizedState=go.bind(null,Ja)}};No.useMemoCache=xl,No.useHostTransitionStatus=ho,No.useFormState=Hl,No.useActionState=Hl,No.useOptimistic=function(e){var t=yl();t.memoizedState=t.baseState=e;var n={pending:null,lanes:0,dispatch:null,lastRenderedReducer:null,lastRenderedState:null};return t.queue=n,t=wo.bind(null,Ja,!0,n),n.dispatch=t,[e,t]};var _o={readContext:Ci,use:wl,useCallback:ao,useContext:Ci,useEffect:Zl,useImperativeHandle:no,useInsertionEffect:Jl,useLayoutEffect:eo,useMemo:lo,useReducer:El,useRef:Kl,useState:function(){return El(Sl)},useDebugValue:ro,useDeferredValue:function(e,t){return io(bl(),el.memoizedState,e,t)},useTransition:function(){var e=El(Sl)[0],t=bl().memoizedState;return["boolean"==typeof e?e:kl(e),t]},useSyncExternalStore:_l,useId:mo};_o.useCacheRefresh=vo,_o.useMemoCache=xl,_o.useHostTransitionStatus=ho,_o.useFormState=Vl,_o.useActionState=Vl,_o.useOptimistic=function(e,t){return Ml(bl(),0,e,t)};var Po={readContext:Ci,use:wl,useCallback:ao,useContext:Ci,useEffect:Zl,useImperativeHandle:no,useInsertionEffect:Jl,useLayoutEffect:eo,useMemo:lo,useReducer:Nl,useRef:Kl,useState:function(){return Nl(Sl)},useDebugValue:ro,useDeferredValue:function(e,t){var n=bl();return null===el?oo(n,e,t):io(n,el.memoizedState,e,t)},useTransition:function(){var e=Nl(Sl)[0],t=bl().memoizedState;return["boolean"==typeof e?e:kl(e),t]},useSyncExternalStore:_l,useId:mo};function zo(e,t,n,r){n=null==(n=n(r,t=e.memoizedState))?t:T({},t,n),e.memoizedState=n,0===e.lanes&&(e.updateQueue.baseState=n)}Po.useCacheRefresh=vo,Po.useMemoCache=xl,Po.useHostTransitionStatus=ho,Po.useFormState=Wl,Po.useActionState=Wl,Po.useOptimistic=function(e,t){var n=bl();return null!==el?Ml(n,0,e,t):(n.baseState=e,[e,n.queue.dispatch])};var jo={isMounted:function(e){return!!(e=e._reactInternals)&&F(e)===e},enqueueSetState:function(e,t,n){e=e._reactInternals;var r=js(),a=Ti(r);a.payload=t,null!=n&&(a.callback=n),null!==(t=Oi(e,a,r))&&(Os(t,0,r),Li(t,e,r))},enqueueReplaceState:function(e,t,n){e=e._reactInternals;var r=js(),a=Ti(r);a.tag=1,a.payload=t,null!=n&&(a.callback=n),null!==(t=Oi(e,a,r))&&(Os(t,0,r),Li(t,e,r))},enqueueForceUpdate:function(e,t){e=e._reactInternals;var n=js(),r=Ti(n);r.tag=2,null!=t&&(r.callback=t),null!==(t=Oi(e,r,n))&&(Os(t,0,n),Li(t,e,n))}};function To(e,t,n,r,a,l,o){return"function"==typeof(e=e.stateNode).shouldComponentUpdate?e.shouldComponentUpdate(r,l,o):!(t.prototype&&t.prototype.isPureReactComponent&&Gn(n,r)&&Gn(a,l))}function Oo(e,t,n,r){e=t.state,"function"==typeof t.componentWillReceiveProps&&t.componentWillReceiveProps(n,r),"function"==typeof t.UNSAFE_componentWillReceiveProps&&t.UNSAFE_componentWillReceiveProps(n,r),t.state!==e&&jo.enqueueReplaceState(t,t.state,null)}function Lo(e,t){var n=t;if("ref"in t)for(var r in n={},t)"ref"!==r&&(n[r]=t[r]);if(e=e.defaultProps)for(var a in n===t&&(n=T({},n)),e)void 0===n[a]&&(n[a]=e[a]);return n}var Mo="function"==typeof reportError?reportError:function(e){if("object"==typeof window&&"function"==typeof window.ErrorEvent){var t=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:"object"==typeof e&&null!==e&&"string"==typeof e.message?String(e.message):String(e),error:e});if(!window.dispatchEvent(t))return}else if("object"==typeof process&&"function"==typeof process.emit)return void process.emit("uncaughtException",e);console.error(e)};function Ao(e){Mo(e)}function Ro(e){console.error(e)}function Fo(e){Mo(e)}function Do(e,t){try{(0,e.onUncaughtError)(t.value,{componentStack:t.stack})}catch(e){setTimeout((function(){throw e}))}}function Io(e,t,n){try{(0,e.onCaughtError)(n.value,{componentStack:n.stack,errorBoundary:1===t.tag?t.stateNode:null})}catch(e){setTimeout((function(){throw e}))}}function Uo(e,t,n){return(n=Ti(n)).tag=3,n.payload={element:null},n.callback=function(){Do(e,t)},n}function $o(e){return(e=Ti(e)).tag=3,e}function Ho(e,t,n,r){var a=n.type.getDerivedStateFromError;if("function"==typeof a){var l=r.value;e.payload=function(){return a(l)},e.callback=function(){Io(t,n,r)}}var o=n.stateNode;null!==o&&"function"==typeof o.componentDidCatch&&(e.callback=function(){Io(t,n,r),"function"!=typeof a&&(null===xs?xs=new Set([this]):xs.add(this));var e=r.stack;this.componentDidCatch(r.value,{componentStack:null!==e?e:""})})}var Vo=Error(o(461)),Bo=!1;function qo(e,t,n,r){t.child=null===e?xa(t,null,n,r):wa(t,e.child,n,r)}function Wo(e,t,n,r,a){n=n.render;var l=t.ref;if("ref"in r){var o={};for(var i in r)"ref"!==i&&(o[i]=r[i])}else o=r;return Ei(t),r=fl(e,t,n,o,l,a),i=ml(),null===e||Bo?(Yr&&i&&Wr(t),t.flags|=1,qo(e,t,r,a),t.child):(vl(e,t,a),di(e,t,a))}function Qo(e,t,n,r,a){if(null===e){var l=n.type;return"function"!=typeof l||Ru(l)||void 0!==l.defaultProps||null!==n.compare?((e=Iu(n.type,null,r,t,t.mode,a)).ref=t.ref,e.return=t,t.child=e):(t.tag=15,t.type=l,Ko(e,t,l,r,a))}if(l=e.child,!pi(e,a)){var o=l.memoizedProps;if((n=null!==(n=n.compare)?n:Gn)(o,r)&&e.ref===t.ref)return di(e,t,a)}return t.flags|=1,(e=Fu(l,r)).ref=t.ref,e.return=t,t.child=e}function Ko(e,t,n,r,a){if(null!==e){var l=e.memoizedProps;if(Gn(l,r)&&e.ref===t.ref){if(Bo=!1,t.pendingProps=r=l,!pi(e,a))return t.lanes=e.lanes,di(e,t,a);131072&e.flags&&(Bo=!0)}}return Zo(e,t,n,r,a)}function Go(e,t,n){var r=t.pendingProps,a=r.children,l=!!(2&t.stateNode._pendingVisibility),o=null!==e?e.memoizedState:null;if(Xo(e,t),"hidden"===r.mode||l){if(128&t.flags){if(r=null!==o?o.baseLanes|n:n,null!==e){for(a=t.child=e.child,l=0;null!==a;)l=l|a.lanes|a.childLanes,a=a.sibling;t.childLanes=l&~r}else t.childLanes=0,t.child=null;return Yo(e,t,r,n)}if(!(536870912&n))return t.lanes=t.childLanes=536870912,Yo(e,t,null!==o?o.baseLanes|n:n,n);t.memoizedState={baseLanes:0,cachePool:null},null!==e&&Ya(0,null!==o?o.cachePool:null),null!==o?Ca(t,o):Na(),Ta(t)}else null!==o?(Ya(0,o.cachePool),Ca(t,o),Oa(),t.memoizedState=null):(null!==e&&Ya(0,null),Na(),Oa());return qo(e,t,a,n),t.child}function Yo(e,t,n,r){var a=Ga();return a=null===a?null:{parent:Ia._currentValue,pool:a},t.memoizedState={baseLanes:n,cachePool:a},null!==e&&Ya(0,null),Na(),Ta(t),null!==e&&xi(e,t,r,!0),null}function Xo(e,t){var n=t.ref;if(null===n)null!==e&&null!==e.ref&&(t.flags|=2097664);else{if("function"!=typeof n&&"object"!=typeof n)throw Error(o(284));null!==e&&e.ref===n||(t.flags|=2097664)}}function Zo(e,t,n,r,a){return Ei(t),n=fl(e,t,n,r,void 0,a),r=ml(),null===e||Bo?(Yr&&r&&Wr(t),t.flags|=1,qo(e,t,n,a),t.child):(vl(e,t,a),di(e,t,a))}function Jo(e,t,n,r,a,l){return Ei(t),t.updateQueue=null,n=pl(t,r,n,a),dl(e),r=ml(),null===e||Bo?(Yr&&r&&Wr(t),t.flags|=1,qo(e,t,n,l),t.child):(vl(e,t,l),di(e,t,l))}function ei(e,t,n,r,a){if(Ei(t),null===t.stateNode){var l=Or,o=n.contextType;"object"==typeof o&&null!==o&&(l=Ci(o)),l=new n(r,l),t.memoizedState=null!==l.state&&void 0!==l.state?l.state:null,l.updater=jo,t.stateNode=l,l._reactInternals=t,(l=t.stateNode).props=r,l.state=t.memoizedState,l.refs={},zi(t),o=n.contextType,l.context="object"==typeof o&&null!==o?Ci(o):Or,l.state=t.memoizedState,"function"==typeof(o=n.getDerivedStateFromProps)&&(zo(t,n,o,r),l.state=t.memoizedState),"function"==typeof n.getDerivedStateFromProps||"function"==typeof l.getSnapshotBeforeUpdate||"function"!=typeof l.UNSAFE_componentWillMount&&"function"!=typeof l.componentWillMount||(o=l.state,"function"==typeof l.componentWillMount&&l.componentWillMount(),"function"==typeof l.UNSAFE_componentWillMount&&l.UNSAFE_componentWillMount(),o!==l.state&&jo.enqueueReplaceState(l,l.state,null),Fi(t,r,l,a),Ri(),l.state=t.memoizedState),"function"==typeof l.componentDidMount&&(t.flags|=4194308),r=!0}else if(null===e){l=t.stateNode;var i=t.memoizedProps,u=Lo(n,i);l.props=u;var s=l.context,c=n.contextType;o=Or,"object"==typeof c&&null!==c&&(o=Ci(c));var f=n.getDerivedStateFromProps;c="function"==typeof f||"function"==typeof l.getSnapshotBeforeUpdate,i=t.pendingProps!==i,c||"function"!=typeof l.UNSAFE_componentWillReceiveProps&&"function"!=typeof l.componentWillReceiveProps||(i||s!==o)&&Oo(t,l,r,o),Pi=!1;var d=t.memoizedState;l.state=d,Fi(t,r,l,a),Ri(),s=t.memoizedState,i||d!==s||Pi?("function"==typeof f&&(zo(t,n,f,r),s=t.memoizedState),(u=Pi||To(t,n,u,r,d,s,o))?(c||"function"!=typeof l.UNSAFE_componentWillMount&&"function"!=typeof l.componentWillMount||("function"==typeof l.componentWillMount&&l.componentWillMount(),"function"==typeof l.UNSAFE_componentWillMount&&l.UNSAFE_componentWillMount()),"function"==typeof l.componentDidMount&&(t.flags|=4194308)):("function"==typeof l.componentDidMount&&(t.flags|=4194308),t.memoizedProps=r,t.memoizedState=s),l.props=r,l.state=s,l.context=o,r=u):("function"==typeof l.componentDidMount&&(t.flags|=4194308),r=!1)}else{l=t.stateNode,ji(e,t),c=Lo(n,o=t.memoizedProps),l.props=c,f=t.pendingProps,d=l.context,s=n.contextType,u=Or,"object"==typeof s&&null!==s&&(u=Ci(s)),(s="function"==typeof(i=n.getDerivedStateFromProps)||"function"==typeof l.getSnapshotBeforeUpdate)||"function"!=typeof l.UNSAFE_componentWillReceiveProps&&"function"!=typeof l.componentWillReceiveProps||(o!==f||d!==u)&&Oo(t,l,r,u),Pi=!1,d=t.memoizedState,l.state=d,Fi(t,r,l,a),Ri();var p=t.memoizedState;o!==f||d!==p||Pi||null!==e&&null!==e.dependencies&&Si(e.dependencies)?("function"==typeof i&&(zo(t,n,i,r),p=t.memoizedState),(c=Pi||To(t,n,c,r,d,p,u)||null!==e&&null!==e.dependencies&&Si(e.dependencies))?(s||"function"!=typeof l.UNSAFE_componentWillUpdate&&"function"!=typeof l.componentWillUpdate||("function"==typeof l.componentWillUpdate&&l.componentWillUpdate(r,p,u),"function"==typeof l.UNSAFE_componentWillUpdate&&l.UNSAFE_componentWillUpdate(r,p,u)),"function"==typeof l.componentDidUpdate&&(t.flags|=4),"function"==typeof l.getSnapshotBeforeUpdate&&(t.flags|=1024)):("function"!=typeof l.componentDidUpdate||o===e.memoizedProps&&d===e.memoizedState||(t.flags|=4),"function"!=typeof l.getSnapshotBeforeUpdate||o===e.memoizedProps&&d===e.memoizedState||(t.flags|=1024),t.memoizedProps=r,t.memoizedState=p),l.props=r,l.state=p,l.context=u,r=c):("function"!=typeof l.componentDidUpdate||o===e.memoizedProps&&d===e.memoizedState||(t.flags|=4),"function"!=typeof l.getSnapshotBeforeUpdate||o===e.memoizedProps&&d===e.memoizedState||(t.flags|=1024),r=!1)}return l=r,Xo(e,t),r=!!(128&t.flags),l||r?(l=t.stateNode,n=r&&"function"!=typeof n.getDerivedStateFromError?null:l.render(),t.flags|=1,null!==e&&r?(t.child=wa(t,e.child,null,a),t.child=wa(t,null,n,a)):qo(e,t,n,a),t.memoizedState=l.state,e=t.child):e=di(e,t,a),e}function ti(e,t,n,r){return aa(),t.flags|=256,qo(e,t,n,r),t.child}var ni={dehydrated:null,treeContext:null,retryLane:0};function ri(e){return{baseLanes:e,cachePool:Xa()}}function ai(e,t,n){return e=null!==e?e.childLanes&~n:0,t&&(e|=hs),e}function li(e,t,n){var r,a=t.pendingProps,l=!1,i=!!(128&t.flags);if((r=i)||(r=(null===e||null!==e.memoizedState)&&!!(2&Ma.current)),r&&(l=!0,t.flags&=-129),r=!!(32&t.flags),t.flags&=-33,null===e){if(Yr){if(l?ja(t):Oa(),Yr){var u,s=Gr;if(u=s){e:{for(u=s,s=Zr;8!==u.nodeType;){if(!s){s=null;break e}if(null===(u=uf(u.nextSibling))){s=null;break e}}s=u}null!==s?(t.memoizedState={dehydrated:s,treeContext:null!==$r?{id:Hr,overflow:Vr}:null,retryLane:536870912},(u=Au(18,null,null,0)).stateNode=s,u.return=t,t.child=u,Kr=t,Gr=null,u=!0):u=!1}u||ea(t)}if(null!==(s=t.memoizedState)&&null!==(s=s.dehydrated))return"$!"===s.data?t.lanes=16:t.lanes=536870912,null;La(t)}return s=a.children,a=a.fallback,l?(Oa(),s=ii({mode:"hidden",children:s},l=t.mode),a=Uu(a,l,n,null),s.return=t,a.return=t,s.sibling=a,t.child=s,(l=t.child).memoizedState=ri(n),l.childLanes=ai(e,r,n),t.memoizedState=ni,a):(ja(t),oi(t,s))}if(null!==(u=e.memoizedState)&&null!==(s=u.dehydrated)){if(i)256&t.flags?(ja(t),t.flags&=-257,t=ui(e,t,n)):null!==t.memoizedState?(Oa(),t.child=e.child,t.flags|=128,t=null):(Oa(),l=a.fallback,s=t.mode,a=ii({mode:"visible",children:a.children},s),(l=Uu(l,s,n,null)).flags|=2,a.return=t,l.return=t,a.sibling=l,t.child=a,wa(t,e.child,null,n),(a=t.child).memoizedState=ri(n),a.childLanes=ai(e,r,n),t.memoizedState=ni,t=l);else if(ja(t),"$!"===s.data){if(r=s.nextSibling&&s.nextSibling.dataset)var c=r.dgst;r=c,(a=Error(o(419))).stack="",a.digest=r,la({value:a,source:null,stack:null}),t=ui(e,t,n)}else if(Bo||xi(e,t,n,!1),r=!!(n&e.childLanes),Bo||r){if(null!==(r=ts)){if(42&(a=n&-n))a=1;else switch(a){case 2:a=1;break;case 8:a=4;break;case 32:a=16;break;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:a=64;break;case 268435456:a=134217728;break;default:a=0}if(0!==(a=a&(r.suspendedLanes|n)?0:a)&&a!==u.retryLane)throw u.retryLane=a,zr(e,a),Os(r,0,a),Vo}"$?"===s.data||Bs(),t=ui(e,t,n)}else"$?"===s.data?(t.flags|=128,t.child=e.child,t=ic.bind(null,e),s._reactRetry=t,t=null):(e=u.treeContext,Gr=uf(s.nextSibling),Kr=t,Yr=!0,Xr=null,Zr=!1,null!==e&&(Ir[Ur++]=Hr,Ir[Ur++]=Vr,Ir[Ur++]=$r,Hr=e.id,Vr=e.overflow,$r=t),(t=oi(t,a.children)).flags|=4096);return t}return l?(Oa(),l=a.fallback,s=t.mode,c=(u=e.child).sibling,(a=Fu(u,{mode:"hidden",children:a.children})).subtreeFlags=31457280&u.subtreeFlags,null!==c?l=Fu(c,l):(l=Uu(l,s,n,null)).flags|=2,l.return=t,a.return=t,a.sibling=l,t.child=a,a=l,l=t.child,null===(s=e.child.memoizedState)?s=ri(n):(null!==(u=s.cachePool)?(c=Ia._currentValue,u=u.parent!==c?{parent:c,pool:c}:u):u=Xa(),s={baseLanes:s.baseLanes|n,cachePool:u}),l.memoizedState=s,l.childLanes=ai(e,r,n),t.memoizedState=ni,a):(ja(t),e=(n=e.child).sibling,(n=Fu(n,{mode:"visible",children:a.children})).return=t,n.sibling=null,null!==e&&(null===(r=t.deletions)?(t.deletions=[e],t.flags|=16):r.push(e)),t.child=n,t.memoizedState=null,n)}function oi(e,t){return(t=ii({mode:"visible",children:t},e.mode)).return=e,e.child=t}function ii(e,t){return $u(e,t,0,null)}function ui(e,t,n){return wa(t,e.child,null,n),(e=oi(t,t.pendingProps.children)).flags|=2,t.memoizedState=null,e}function si(e,t,n){e.lanes|=t;var r=e.alternate;null!==r&&(r.lanes|=t),ki(e.return,t,n)}function ci(e,t,n,r,a){var l=e.memoizedState;null===l?e.memoizedState={isBackwards:t,rendering:null,renderingStartTime:0,last:r,tail:n,tailMode:a}:(l.isBackwards=t,l.rendering=null,l.renderingStartTime=0,l.last=r,l.tail=n,l.tailMode=a)}function fi(e,t,n){var r=t.pendingProps,a=r.revealOrder,l=r.tail;if(qo(e,t,r.children,n),2&(r=Ma.current))r=1&r|2,t.flags|=128;else{if(null!==e&&128&e.flags)e:for(e=t.child;null!==e;){if(13===e.tag)null!==e.memoizedState&&si(e,n,t);else if(19===e.tag)si(e,n,t);else if(null!==e.child){e.child.return=e,e=e.child;continue}if(e===t)break e;for(;null===e.sibling;){if(null===e.return||e.return===t)break e;e=e.return}e.sibling.return=e.return,e=e.sibling}r&=1}switch(K(Ma,r),a){case"forwards":for(n=t.child,a=null;null!==n;)null!==(e=n.alternate)&&null===Aa(e)&&(a=n),n=n.sibling;null===(n=a)?(a=t.child,t.child=null):(a=n.sibling,n.sibling=null),ci(t,!1,a,n,l);break;case"backwards":for(n=null,a=t.child,t.child=null;null!==a;){if(null!==(e=a.alternate)&&null===Aa(e)){t.child=a;break}e=a.sibling,a.sibling=n,n=a,a=e}ci(t,!0,n,null,l);break;case"together":ci(t,!1,null,null,void 0);break;default:t.memoizedState=null}return t.child}function di(e,t,n){if(null!==e&&(t.dependencies=e.dependencies),fs|=t.lanes,!(n&t.childLanes)){if(null===e)return null;if(xi(e,t,n,!1),!(n&t.childLanes))return null}if(null!==e&&t.child!==e.child)throw Error(o(153));if(null!==t.child){for(n=Fu(e=t.child,e.pendingProps),t.child=n,n.return=t;null!==e.sibling;)e=e.sibling,(n=n.sibling=Fu(e,e.pendingProps)).return=t;n.sibling=null}return t.child}function pi(e,t){return!!(e.lanes&t)||!(null===(e=e.dependencies)||!Si(e))}function hi(e,t,n){if(null!==e)if(e.memoizedProps!==t.pendingProps)Bo=!0;else{if(!(pi(e,n)||128&t.flags))return Bo=!1,function(e,t,n){switch(t.tag){case 3:J(t,t.stateNode.containerInfo),yi(0,Ia,e.memoizedState.cache),aa();break;case 27:case 5:te(t);break;case 4:J(t,t.stateNode.containerInfo);break;case 10:yi(0,t.type,t.memoizedProps.value);break;case 13:var r=t.memoizedState;if(null!==r)return null!==r.dehydrated?(ja(t),t.flags|=128,null):n&t.child.childLanes?li(e,t,n):(ja(t),null!==(e=di(e,t,n))?e.sibling:null);ja(t);break;case 19:var a=!!(128&e.flags);if((r=!!(n&t.childLanes))||(xi(e,t,n,!1),r=!!(n&t.childLanes)),a){if(r)return fi(e,t,n);t.flags|=128}if(null!==(a=t.memoizedState)&&(a.rendering=null,a.tail=null,a.lastEffect=null),K(Ma,Ma.current),r)break;return null;case 22:case 23:return t.lanes=0,Go(e,t,n);case 24:yi(0,Ia,e.memoizedState.cache)}return di(e,t,n)}(e,t,n);Bo=!!(131072&e.flags)}else Bo=!1,Yr&&1048576&t.flags&&qr(t,Dr,t.index);switch(t.lanes=0,t.tag){case 16:e:{e=t.pendingProps;var r=t.elementType,a=r._init;if(r=a(r._payload),t.type=r,"function"!=typeof r){if(null!=r){if((a=r.$$typeof)===g){t.tag=11,t=Wo(null,t,r,e,n);break e}if(a===k){t.tag=14,t=Qo(null,t,r,e,n);break e}}throw t=_(r)||r,Error(o(306,t,""))}Ru(r)?(e=Lo(r,e),t.tag=1,t=ei(null,t,r,e,n)):(t.tag=0,t=Zo(null,t,r,e,n))}return t;case 0:return Zo(e,t,t.type,t.pendingProps,n);case 1:return ei(e,t,r=t.type,a=Lo(r,t.pendingProps),n);case 3:e:{if(J(t,t.stateNode.containerInfo),null===e)throw Error(o(387));var l=t.pendingProps;r=(a=t.memoizedState).element,ji(e,t),Fi(t,l,null,n);var i=t.memoizedState;if(l=i.cache,yi(0,Ia,l),l!==a.cache&&wi(t,[Ia],n,!0),Ri(),l=i.element,a.isDehydrated){if(a={element:l,isDehydrated:!1,cache:i.cache},t.updateQueue.baseState=a,t.memoizedState=a,256&t.flags){t=ti(e,t,l,n);break e}if(l!==r){la(r=Mr(Error(o(424)),t)),t=ti(e,t,l,n);break e}for(Gr=uf(t.stateNode.containerInfo.firstChild),Kr=t,Yr=!0,Xr=null,Zr=!0,n=xa(t,null,l,n),t.child=n;n;)n.flags=-3&n.flags|4096,n=n.sibling}else{if(aa(),l===r){t=di(e,t,n);break e}qo(e,t,l,n)}t=t.child}return t;case 26:return Xo(e,t),null===e?(n=gf(t.type,null,t.pendingProps,null))?t.memoizedState=n:Yr||(n=t.type,e=t.pendingProps,(r=Gc(X.current).createElement(n))[De]=t,r[Ie]=e,Wc(r,n,e),Xe(r),t.stateNode=r):t.memoizedState=gf(t.type,e.memoizedProps,t.pendingProps,e.memoizedState),null;case 27:return te(t),null===e&&Yr&&(r=t.stateNode=cf(t.type,t.pendingProps,X.current),Kr=t,Zr=!0,Gr=uf(r.firstChild)),r=t.pendingProps.children,null!==e||Yr?qo(e,t,r,n):t.child=wa(t,null,r,n),Xo(e,t),t.child;case 5:return null===e&&Yr&&((a=r=Gr)&&(null!==(r=function(e,t,n,r){for(;1===e.nodeType;){var a=n;if(e.nodeName.toLowerCase()!==t.toLowerCase()){if(!r&&("INPUT"!==e.nodeName||"hidden"!==e.type))break}else if(r){if(!e[qe])switch(t){case"meta":if(!e.hasAttribute("itemprop"))break;return e;case"link":if("stylesheet"===(l=e.getAttribute("rel"))&&e.hasAttribute("data-precedence"))break;if(l!==a.rel||e.getAttribute("href")!==(null==a.href?null:a.href)||e.getAttribute("crossorigin")!==(null==a.crossOrigin?null:a.crossOrigin)||e.getAttribute("title")!==(null==a.title?null:a.title))break;return e;case"style":if(e.hasAttribute("data-precedence"))break;return e;case"script":if(((l=e.getAttribute("src"))!==(null==a.src?null:a.src)||e.getAttribute("type")!==(null==a.type?null:a.type)||e.getAttribute("crossorigin")!==(null==a.crossOrigin?null:a.crossOrigin))&&l&&e.hasAttribute("async")&&!e.hasAttribute("itemprop"))break;return e;default:return e}}else{if("input"!==t||"hidden"!==e.type)return e;var l=null==a.name?null:""+a.name;if("hidden"===a.type&&e.getAttribute("name")===l)return e}if(null===(e=uf(e.nextSibling)))break}return null}(r,t.type,t.pendingProps,Zr))?(t.stateNode=r,Kr=t,Gr=uf(r.firstChild),Zr=!1,a=!0):a=!1),a||ea(t)),te(t),a=t.type,l=t.pendingProps,i=null!==e?e.memoizedProps:null,r=l.children,Zc(a,l)?r=null:null!==i&&Zc(a,i)&&(t.flags|=32),null!==t.memoizedState&&(a=fl(e,t,hl,null,null,n),Ff._currentValue=a),Xo(e,t),qo(e,t,r,n),t.child;case 6:return null===e&&Yr&&((e=n=Gr)&&(null!==(n=function(e,t,n){if(""===t)return null;for(;3!==e.nodeType;){if((1!==e.nodeType||"INPUT"!==e.nodeName||"hidden"!==e.type)&&!n)return null;if(null===(e=uf(e.nextSibling)))return null}return e}(n,t.pendingProps,Zr))?(t.stateNode=n,Kr=t,Gr=null,e=!0):e=!1),e||ea(t)),null;case 13:return li(e,t,n);case 4:return J(t,t.stateNode.containerInfo),r=t.pendingProps,null===e?t.child=wa(t,null,r,n):qo(e,t,r,n),t.child;case 11:return Wo(e,t,t.type,t.pendingProps,n);case 7:return qo(e,t,t.pendingProps,n),t.child;case 8:case 12:return qo(e,t,t.pendingProps.children,n),t.child;case 10:return r=t.pendingProps,yi(0,t.type,r.value),qo(e,t,r.children,n),t.child;case 9:return a=t.type._context,r=t.pendingProps.children,Ei(t),r=r(a=Ci(a)),t.flags|=1,qo(e,t,r,n),t.child;case 14:return Qo(e,t,t.type,t.pendingProps,n);case 15:return Ko(e,t,t.type,t.pendingProps,n);case 19:return fi(e,t,n);case 22:return Go(e,t,n);case 24:return Ei(t),r=Ci(Ia),null===e?(null===(a=Ga())&&(a=ts,l=Ua(),a.pooledCache=l,l.refCount++,null!==l&&(a.pooledCacheLanes|=n),a=l),t.memoizedState={parent:r,cache:a},zi(t),yi(0,Ia,a)):(!!(e.lanes&n)&&(ji(e,t),Fi(t,null,null,n),Ri()),a=e.memoizedState,l=t.memoizedState,a.parent!==r?(a={parent:r,cache:r},t.memoizedState=a,0===t.lanes&&(t.memoizedState=t.updateQueue.baseState=a),yi(0,Ia,r)):(r=l.cache,yi(0,Ia,r),r!==a.cache&&wi(t,[Ia],n,!0))),qo(e,t,t.pendingProps.children,n),t.child;case 29:throw t.pendingProps}throw Error(o(156,t.tag))}var mi=W(null),vi=null,gi=null;function yi(e,t,n){K(mi,t._currentValue),t._currentValue=n}function bi(e){e._currentValue=mi.current,Q(mi)}function ki(e,t,n){for(;null!==e;){var r=e.alternate;if((e.childLanes&t)!==t?(e.childLanes|=t,null!==r&&(r.childLanes|=t)):null!==r&&(r.childLanes&t)!==t&&(r.childLanes|=t),e===n)break;e=e.return}}function wi(e,t,n,r){var a=e.child;for(null!==a&&(a.return=e);null!==a;){var l=a.dependencies;if(null!==l){var i=a.child;l=l.firstContext;e:for(;null!==l;){var u=l;l=a;for(var s=0;s<t.length;s++)if(u.context===t[s]){l.lanes|=n,null!==(u=l.alternate)&&(u.lanes|=n),ki(l.return,n,e),r||(i=null);break e}l=u.next}}else if(18===a.tag){if(null===(i=a.return))throw Error(o(341));i.lanes|=n,null!==(l=i.alternate)&&(l.lanes|=n),ki(i,n,e),i=null}else i=a.child;if(null!==i)i.return=a;else for(i=a;null!==i;){if(i===e){i=null;break}if(null!==(a=i.sibling)){a.return=i.return,i=a;break}i=i.return}a=i}}function xi(e,t,n,r){e=null;for(var a=t,l=!1;null!==a;){if(!l)if(524288&a.flags)l=!0;else if(262144&a.flags)break;if(10===a.tag){var i=a.alternate;if(null===i)throw Error(o(387));if(null!==(i=i.memoizedProps)){var u=a.type;Kn(a.pendingProps.value,i.value)||(null!==e?e.push(u):e=[u])}}else if(a===Z.current){if(null===(i=a.alternate))throw Error(o(387));i.memoizedState.memoizedState!==a.memoizedState.memoizedState&&(null!==e?e.push(Ff):e=[Ff])}a=a.return}null!==e&&wi(t,e,n,r),t.flags|=262144}function Si(e){for(e=e.firstContext;null!==e;){if(!Kn(e.context._currentValue,e.memoizedValue))return!0;e=e.next}return!1}function Ei(e){vi=e,gi=null,null!==(e=e.dependencies)&&(e.firstContext=null)}function Ci(e){return _i(vi,e)}function Ni(e,t){return null===vi&&Ei(e),_i(e,t)}function _i(e,t){var n=t._currentValue;if(t={context:t,memoizedValue:n,next:null},null===gi){if(null===e)throw Error(o(308));gi=t,e.dependencies={lanes:0,firstContext:t},e.flags|=524288}else gi=gi.next=t;return n}var Pi=!1;function zi(e){e.updateQueue={baseState:e.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,lanes:0,hiddenCallbacks:null},callbacks:null}}function ji(e,t){e=e.updateQueue,t.updateQueue===e&&(t.updateQueue={baseState:e.baseState,firstBaseUpdate:e.firstBaseUpdate,lastBaseUpdate:e.lastBaseUpdate,shared:e.shared,callbacks:null})}function Ti(e){return{lane:e,tag:0,payload:null,callback:null,next:null}}function Oi(e,t,n){var r=e.updateQueue;if(null===r)return null;if(r=r.shared,2&es){var a=r.pending;return null===a?t.next=t:(t.next=a.next,a.next=t),r.pending=t,t=Tr(e),jr(e,null,n),t}return _r(e,r,t,n),Tr(e)}function Li(e,t,n){if(null!==(t=t.updateQueue)&&(t=t.shared,4194176&n)){var r=t.lanes;n|=r&=e.pendingLanes,t.lanes=n,Me(e,n)}}function Mi(e,t){var n=e.updateQueue,r=e.alternate;if(null!==r&&n===(r=r.updateQueue)){var a=null,l=null;if(null!==(n=n.firstBaseUpdate)){do{var o={lane:n.lane,tag:n.tag,payload:n.payload,callback:null,next:null};null===l?a=l=o:l=l.next=o,n=n.next}while(null!==n);null===l?a=l=t:l=l.next=t}else a=l=t;return n={baseState:r.baseState,firstBaseUpdate:a,lastBaseUpdate:l,shared:r.shared,callbacks:r.callbacks},void(e.updateQueue=n)}null===(e=n.lastBaseUpdate)?n.firstBaseUpdate=t:e.next=t,n.lastBaseUpdate=t}var Ai=!1;function Ri(){if(Ai&&null!==qa)throw qa}function Fi(e,t,n,r){Ai=!1;var a=e.updateQueue;Pi=!1;var l=a.firstBaseUpdate,o=a.lastBaseUpdate,i=a.shared.pending;if(null!==i){a.shared.pending=null;var u=i,s=u.next;u.next=null,null===o?l=s:o.next=s,o=u;var c=e.alternate;null!==c&&(i=(c=c.updateQueue).lastBaseUpdate)!==o&&(null===i?c.firstBaseUpdate=s:i.next=s,c.lastBaseUpdate=u)}if(null!==l){var f=a.baseState;for(o=0,c=s=u=null,i=l;;){var d=-536870913&i.lane,p=d!==i.lane;if(p?(rs&d)===d:(r&d)===d){0!==d&&d===Ba&&(Ai=!0),null!==c&&(c=c.next={lane:0,tag:i.tag,payload:i.payload,callback:null,next:null});e:{var h=e,m=i;d=t;var v=n;switch(m.tag){case 1:if("function"==typeof(h=m.payload)){f=h.call(v,f,d);break e}f=h;break e;case 3:h.flags=-65537&h.flags|128;case 0:if(null==(d="function"==typeof(h=m.payload)?h.call(v,f,d):h))break e;f=T({},f,d);break e;case 2:Pi=!0}}null!==(d=i.callback)&&(e.flags|=64,p&&(e.flags|=8192),null===(p=a.callbacks)?a.callbacks=[d]:p.push(d))}else p={lane:d,tag:i.tag,payload:i.payload,callback:i.callback,next:null},null===c?(s=c=p,u=f):c=c.next=p,o|=d;if(null===(i=i.next)){if(null===(i=a.shared.pending))break;i=(p=i).next,p.next=null,a.lastBaseUpdate=p,a.shared.pending=null}}null===c&&(u=f),a.baseState=u,a.firstBaseUpdate=s,a.lastBaseUpdate=c,null===l&&(a.shared.lanes=0),fs|=o,e.lanes=o,e.memoizedState=f}}function Di(e,t){if("function"!=typeof e)throw Error(o(191,e));e.call(t)}function Ii(e,t){var n=e.callbacks;if(null!==n)for(e.callbacks=null,e=0;e<n.length;e++)Di(n[e],t)}function Ui(e,t){try{var n=t.updateQueue,r=null!==n?n.lastEffect:null;if(null!==r){var a=r.next;n=a;do{if((n.tag&e)===e){r=void 0;var l=n.create,o=n.inst;r=l(),o.destroy=r}n=n.next}while(n!==a)}}catch(e){rc(t,t.return,e)}}function $i(e,t,n){try{var r=t.updateQueue,a=null!==r?r.lastEffect:null;if(null!==a){var l=a.next;r=l;do{if((r.tag&e)===e){var o=r.inst,i=o.destroy;if(void 0!==i){o.destroy=void 0,a=t;var u=n;try{i()}catch(e){rc(a,u,e)}}}r=r.next}while(r!==l)}}catch(e){rc(t,t.return,e)}}function Hi(e){var t=e.updateQueue;if(null!==t){var n=e.stateNode;try{Ii(t,n)}catch(t){rc(e,e.return,t)}}}function Vi(e,t,n){n.props=Lo(e.type,e.memoizedProps),n.state=e.memoizedState;try{n.componentWillUnmount()}catch(n){rc(e,t,n)}}function Bi(e,t){try{var n=e.ref;if(null!==n){var r=e.stateNode;switch(e.tag){case 26:case 27:case 5:var a=r;break;default:a=r}"function"==typeof n?e.refCleanup=n(a):n.current=a}}catch(n){rc(e,t,n)}}function qi(e,t){var n=e.ref,r=e.refCleanup;if(null!==n)if("function"==typeof r)try{r()}catch(n){rc(e,t,n)}finally{e.refCleanup=null,null!=(e=e.alternate)&&(e.refCleanup=null)}else if("function"==typeof n)try{n(null)}catch(n){rc(e,t,n)}else n.current=null}function Wi(e){var t=e.type,n=e.memoizedProps,r=e.stateNode;try{e:switch(t){case"button":case"input":case"select":case"textarea":n.autoFocus&&r.focus();break e;case"img":n.src?r.src=n.src:n.srcSet&&(r.srcset=n.srcSet)}}catch(t){rc(e,e.return,t)}}function Qi(e,t,n){try{var r=e.stateNode;!function(e,t,n,r){switch(t){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"input":var a=null,l=null,i=null,u=null,s=null,c=null,f=null;for(h in n){var d=n[h];if(n.hasOwnProperty(h)&&null!=d)switch(h){case"checked":case"value":break;case"defaultValue":s=d;default:r.hasOwnProperty(h)||Bc(e,t,h,null,r,d)}}for(var p in r){var h=r[p];if(d=n[p],r.hasOwnProperty(p)&&(null!=h||null!=d))switch(p){case"type":l=h;break;case"name":a=h;break;case"checked":c=h;break;case"defaultChecked":f=h;break;case"value":i=h;break;case"defaultValue":u=h;break;case"children":case"dangerouslySetInnerHTML":if(null!=h)throw Error(o(137,t));break;default:h!==d&&Bc(e,t,p,h,r,d)}}return void vt(e,i,u,s,c,f,l,a);case"select":for(l in h=i=u=p=null,n)if(s=n[l],n.hasOwnProperty(l)&&null!=s)switch(l){case"value":break;case"multiple":h=s;default:r.hasOwnProperty(l)||Bc(e,t,l,null,r,s)}for(a in r)if(l=r[a],s=n[a],r.hasOwnProperty(a)&&(null!=l||null!=s))switch(a){case"value":p=l;break;case"defaultValue":u=l;break;case"multiple":i=l;default:l!==s&&Bc(e,t,a,l,r,s)}return t=u,n=i,r=h,void(null!=p?bt(e,!!n,p,!1):!!r!=!!n&&(null!=t?bt(e,!!n,t,!0):bt(e,!!n,n?[]:"",!1)));case"textarea":for(u in h=p=null,n)if(a=n[u],n.hasOwnProperty(u)&&null!=a&&!r.hasOwnProperty(u))switch(u){case"value":case"children":break;default:Bc(e,t,u,null,r,a)}for(i in r)if(a=r[i],l=n[i],r.hasOwnProperty(i)&&(null!=a||null!=l))switch(i){case"value":p=a;break;case"defaultValue":h=a;break;case"children":break;case"dangerouslySetInnerHTML":if(null!=a)throw Error(o(91));break;default:a!==l&&Bc(e,t,i,a,r,l)}return void kt(e,p,h);case"option":for(var m in n)p=n[m],n.hasOwnProperty(m)&&null!=p&&!r.hasOwnProperty(m)&&("selected"===m?e.selected=!1:Bc(e,t,m,null,r,p));for(s in r)p=r[s],h=n[s],!r.hasOwnProperty(s)||p===h||null==p&&null==h||("selected"===s?e.selected=p&&"function"!=typeof p&&"symbol"!=typeof p:Bc(e,t,s,p,r,h));return;case"img":case"link":case"area":case"base":case"br":case"col":case"embed":case"hr":case"keygen":case"meta":case"param":case"source":case"track":case"wbr":case"menuitem":for(var v in n)p=n[v],n.hasOwnProperty(v)&&null!=p&&!r.hasOwnProperty(v)&&Bc(e,t,v,null,r,p);for(c in r)if(p=r[c],h=n[c],r.hasOwnProperty(c)&&p!==h&&(null!=p||null!=h))switch(c){case"children":case"dangerouslySetInnerHTML":if(null!=p)throw Error(o(137,t));break;default:Bc(e,t,c,p,r,h)}return;default:if(Nt(t)){for(var g in n)p=n[g],n.hasOwnProperty(g)&&void 0!==p&&!r.hasOwnProperty(g)&&qc(e,t,g,void 0,r,p);for(f in r)p=r[f],h=n[f],!r.hasOwnProperty(f)||p===h||void 0===p&&void 0===h||qc(e,t,f,p,r,h);return}}for(var y in n)p=n[y],n.hasOwnProperty(y)&&null!=p&&!r.hasOwnProperty(y)&&Bc(e,t,y,null,r,p);for(d in r)p=r[d],h=n[d],!r.hasOwnProperty(d)||p===h||null==p&&null==h||Bc(e,t,d,p,r,h)}(r,e.type,n,t),r[Ie]=t}catch(t){rc(e,e.return,t)}}function Ki(e){return 5===e.tag||3===e.tag||26===e.tag||27===e.tag||4===e.tag}function Gi(e){e:for(;;){for(;null===e.sibling;){if(null===e.return||Ki(e.return))return null;e=e.return}for(e.sibling.return=e.return,e=e.sibling;5!==e.tag&&6!==e.tag&&27!==e.tag&&18!==e.tag;){if(2&e.flags)continue e;if(null===e.child||4===e.tag)continue e;e.child.return=e,e=e.child}if(!(2&e.flags))return e.stateNode}}function Yi(e,t,n){var r=e.tag;if(5===r||6===r)e=e.stateNode,t?8===n.nodeType?n.parentNode.insertBefore(e,t):n.insertBefore(e,t):(8===n.nodeType?(t=n.parentNode).insertBefore(e,n):(t=n).appendChild(e),null!=(n=n._reactRootContainer)||null!==t.onclick||(t.onclick=Vc));else if(4!==r&&27!==r&&null!==(e=e.child))for(Yi(e,t,n),e=e.sibling;null!==e;)Yi(e,t,n),e=e.sibling}function Xi(e,t,n){var r=e.tag;if(5===r||6===r)e=e.stateNode,t?n.insertBefore(e,t):n.appendChild(e);else if(4!==r&&27!==r&&null!==(e=e.child))for(Xi(e,t,n),e=e.sibling;null!==e;)Xi(e,t,n),e=e.sibling}var Zi=!1,Ji=!1,eu=!1,tu="function"==typeof WeakSet?WeakSet:Set,nu=null,ru=!1;function au(e,t,n){var r=n.flags;switch(n.tag){case 0:case 11:case 15:gu(e,n),4&r&&Ui(5,n);break;case 1:if(gu(e,n),4&r)if(e=n.stateNode,null===t)try{e.componentDidMount()}catch(e){rc(n,n.return,e)}else{var a=Lo(n.type,t.memoizedProps);t=t.memoizedState;try{e.componentDidUpdate(a,t,e.__reactInternalSnapshotBeforeUpdate)}catch(e){rc(n,n.return,e)}}64&r&&Hi(n),512&r&&Bi(n,n.return);break;case 3:if(gu(e,n),64&r&&null!==(r=n.updateQueue)){if(e=null,null!==n.child)switch(n.child.tag){case 27:case 5:case 1:e=n.child.stateNode}try{Ii(r,e)}catch(e){rc(n,n.return,e)}}break;case 26:gu(e,n),512&r&&Bi(n,n.return);break;case 27:case 5:gu(e,n),null===t&&4&r&&Wi(n),512&r&&Bi(n,n.return);break;case 12:default:gu(e,n);break;case 13:gu(e,n),4&r&&cu(e,n);break;case 22:if(!(a=null!==n.memoizedState||Zi)){t=null!==t&&null!==t.memoizedState||Ji;var l=Zi,o=Ji;Zi=a,(Ji=t)&&!o?bu(e,n,!!(8772&n.subtreeFlags)):gu(e,n),Zi=l,Ji=o}512&r&&("manual"===n.memoizedProps.mode?Bi(n,n.return):qi(n,n.return))}}function lu(e){var t=e.alternate;null!==t&&(e.alternate=null,lu(t)),e.child=null,e.deletions=null,e.sibling=null,5===e.tag&&null!==(t=e.stateNode)&&We(t),e.stateNode=null,e.return=null,e.dependencies=null,e.memoizedProps=null,e.memoizedState=null,e.pendingProps=null,e.stateNode=null,e.updateQueue=null}var ou=null,iu=!1;function uu(e,t,n){for(n=n.child;null!==n;)su(e,t,n),n=n.sibling}function su(e,t,n){if(ye&&"function"==typeof ye.onCommitFiberUnmount)try{ye.onCommitFiberUnmount(ge,n)}catch(e){}switch(n.tag){case 26:Ji||qi(n,t),uu(e,t,n),n.memoizedState?n.memoizedState.count--:n.stateNode&&(n=n.stateNode).parentNode.removeChild(n);break;case 27:Ji||qi(n,t);var r=ou,a=iu;for(ou=n.stateNode,uu(e,t,n),t=(n=n.stateNode).attributes;t.length;)n.removeAttributeNode(t[0]);We(n),ou=r,iu=a;break;case 5:Ji||qi(n,t);case 6:a=ou;var l=iu;if(ou=null,uu(e,t,n),iu=l,null!==(ou=a))if(iu)try{e=ou,r=n.stateNode,8===e.nodeType?e.parentNode.removeChild(r):e.removeChild(r)}catch(e){rc(n,t,e)}else try{ou.removeChild(n.stateNode)}catch(e){rc(n,t,e)}break;case 18:null!==ou&&(iu?(t=ou,n=n.stateNode,8===t.nodeType?lf(t.parentNode,n):1===t.nodeType&&lf(t,n),vd(t)):lf(ou,n.stateNode));break;case 4:r=ou,a=iu,ou=n.stateNode.containerInfo,iu=!0,uu(e,t,n),ou=r,iu=a;break;case 0:case 11:case 14:case 15:Ji||$i(2,n,t),Ji||$i(4,n,t),uu(e,t,n);break;case 1:Ji||(qi(n,t),"function"==typeof(r=n.stateNode).componentWillUnmount&&Vi(n,t,r)),uu(e,t,n);break;case 21:uu(e,t,n);break;case 22:Ji||qi(n,t),Ji=(r=Ji)||null!==n.memoizedState,uu(e,t,n),Ji=r;break;default:uu(e,t,n)}}function cu(e,t){if(null===t.memoizedState&&null!==(e=t.alternate)&&null!==(e=e.memoizedState)&&null!==(e=e.dehydrated))try{vd(e)}catch(e){rc(t,t.return,e)}}function fu(e,t){var n=function(e){switch(e.tag){case 13:case 19:var t=e.stateNode;return null===t&&(t=e.stateNode=new tu),t;case 22:return null===(t=(e=e.stateNode)._retryCache)&&(t=e._retryCache=new tu),t;default:throw Error(o(435,e.tag))}}(e);t.forEach((function(t){var r=uc.bind(null,e,t);n.has(t)||(n.add(t),t.then(r,r))}))}function du(e,t){var n=t.deletions;if(null!==n)for(var r=0;r<n.length;r++){var a=n[r],l=e,i=t,u=i;e:for(;null!==u;){switch(u.tag){case 27:case 5:ou=u.stateNode,iu=!1;break e;case 3:case 4:ou=u.stateNode.containerInfo,iu=!0;break e}u=u.return}if(null===ou)throw Error(o(160));su(l,i,a),ou=null,iu=!1,null!==(l=a.alternate)&&(l.return=null),a.return=null}if(13878&t.subtreeFlags)for(t=t.child;null!==t;)hu(t,e),t=t.sibling}var pu=null;function hu(e,t){var n=e.alternate,r=e.flags;switch(e.tag){case 0:case 11:case 14:case 15:du(t,e),mu(e),4&r&&($i(3,e,e.return),Ui(3,e),$i(5,e,e.return));break;case 1:du(t,e),mu(e),512&r&&(Ji||null===n||qi(n,n.return)),64&r&&Zi&&null!==(e=e.updateQueue)&&null!==(r=e.callbacks)&&(n=e.shared.hiddenCallbacks,e.shared.hiddenCallbacks=null===n?r:n.concat(r));break;case 26:var a=pu;if(du(t,e),mu(e),512&r&&(Ji||null===n||qi(n,n.return)),4&r){var l=null!==n?n.memoizedState:null;if(r=e.memoizedState,null===n)if(null===r)if(null===e.stateNode){e:{r=e.type,n=e.memoizedProps,a=a.ownerDocument||a;t:switch(r){case"title":(!(l=a.getElementsByTagName("title")[0])||l[qe]||l[De]||"http://www.w3.org/2000/svg"===l.namespaceURI||l.hasAttribute("itemprop"))&&(l=a.createElement(r),a.head.insertBefore(l,a.querySelector("head > title"))),Wc(l,r,n),l[De]=e,Xe(l),r=l;break e;case"link":var i=Pf("link","href",a).get(r+(n.href||""));if(i)for(var u=0;u<i.length;u++)if((l=i[u]).getAttribute("href")===(null==n.href?null:n.href)&&l.getAttribute("rel")===(null==n.rel?null:n.rel)&&l.getAttribute("title")===(null==n.title?null:n.title)&&l.getAttribute("crossorigin")===(null==n.crossOrigin?null:n.crossOrigin)){i.splice(u,1);break t}Wc(l=a.createElement(r),r,n),a.head.appendChild(l);break;case"meta":if(i=Pf("meta","content",a).get(r+(n.content||"")))for(u=0;u<i.length;u++)if((l=i[u]).getAttribute("content")===(null==n.content?null:""+n.content)&&l.getAttribute("name")===(null==n.name?null:n.name)&&l.getAttribute("property")===(null==n.property?null:n.property)&&l.getAttribute("http-equiv")===(null==n.httpEquiv?null:n.httpEquiv)&&l.getAttribute("charset")===(null==n.charSet?null:n.charSet)){i.splice(u,1);break t}Wc(l=a.createElement(r),r,n),a.head.appendChild(l);break;default:throw Error(o(468,r))}l[De]=e,Xe(l),r=l}e.stateNode=r}else zf(a,e.type,e.stateNode);else e.stateNode=Sf(a,r,e.memoizedProps);else l!==r?(null===l?null!==n.stateNode&&(n=n.stateNode).parentNode.removeChild(n):l.count--,null===r?zf(a,e.type,e.stateNode):Sf(a,r,e.memoizedProps)):null===r&&null!==e.stateNode&&Qi(e,e.memoizedProps,n.memoizedProps)}break;case 27:if(4&r&&null===e.alternate){a=e.stateNode,l=e.memoizedProps;try{for(var s=a.firstChild;s;){var c=s.nextSibling,f=s.nodeName;s[qe]||"HEAD"===f||"BODY"===f||"SCRIPT"===f||"STYLE"===f||"LINK"===f&&"stylesheet"===s.rel.toLowerCase()||a.removeChild(s),s=c}for(var d=e.type,p=a.attributes;p.length;)a.removeAttributeNode(p[0]);Wc(a,d,l),a[De]=e,a[Ie]=l}catch(t){rc(e,e.return,t)}}case 5:if(du(t,e),mu(e),512&r&&(Ji||null===n||qi(n,n.return)),32&e.flags){a=e.stateNode;try{xt(a,"")}catch(t){rc(e,e.return,t)}}4&r&&null!=e.stateNode&&Qi(e,a=e.memoizedProps,null!==n?n.memoizedProps:a),1024&r&&(eu=!0);break;case 6:if(du(t,e),mu(e),4&r){if(null===e.stateNode)throw Error(o(162));r=e.memoizedProps,n=e.stateNode;try{n.nodeValue=r}catch(t){rc(e,e.return,t)}}break;case 3:if(_f=null,a=pu,pu=pf(t.containerInfo),du(t,e),pu=a,mu(e),4&r&&null!==n&&n.memoizedState.isDehydrated)try{vd(t.containerInfo)}catch(t){rc(e,e.return,t)}eu&&(eu=!1,vu(e));break;case 4:r=pu,pu=pf(e.stateNode.containerInfo),du(t,e),mu(e),pu=r;break;case 12:du(t,e),mu(e);break;case 13:du(t,e),mu(e),8192&e.child.flags&&null!==e.memoizedState!=(null!==n&&null!==n.memoizedState)&&(bs=ue()),4&r&&null!==(r=e.updateQueue)&&(e.updateQueue=null,fu(e,r));break;case 22:if(512&r&&(Ji||null===n||qi(n,n.return)),s=null!==e.memoizedState,c=null!==n&&null!==n.memoizedState,Zi=(f=Zi)||s,Ji=(d=Ji)||c,du(t,e),Ji=d,Zi=f,mu(e),(t=e.stateNode)._current=e,t._visibility&=-3,t._visibility|=2&t._pendingVisibility,8192&r&&(t._visibility=s?-2&t._visibility:1|t._visibility,s&&(t=Zi||Ji,null===n||c||t||yu(e)),null===e.memoizedProps||"manual"!==e.memoizedProps.mode))e:for(n=null,t=e;;){if(5===t.tag||26===t.tag||27===t.tag){if(null===n){c=n=t;try{if(a=c.stateNode,s)"function"==typeof(l=a.style).setProperty?l.setProperty("display","none","important"):l.display="none";else{i=c.stateNode;var h=null!=(u=c.memoizedProps.style)&&u.hasOwnProperty("display")?u.display:null;i.style.display=null==h||"boolean"==typeof h?"":(""+h).trim()}}catch(e){rc(c,c.return,e)}}}else if(6===t.tag){if(null===n){c=t;try{c.stateNode.nodeValue=s?"":c.memoizedProps}catch(e){rc(c,c.return,e)}}}else if((22!==t.tag&&23!==t.tag||null===t.memoizedState||t===e)&&null!==t.child){t.child.return=t,t=t.child;continue}if(t===e)break e;for(;null===t.sibling;){if(null===t.return||t.return===e)break e;n===t&&(n=null),t=t.return}n===t&&(n=null),t.sibling.return=t.return,t=t.sibling}4&r&&null!==(r=e.updateQueue)&&null!==(n=r.retryQueue)&&(r.retryQueue=null,fu(e,n));break;case 19:du(t,e),mu(e),4&r&&null!==(r=e.updateQueue)&&(e.updateQueue=null,fu(e,r));break;case 21:break;default:du(t,e),mu(e)}}function mu(e){var t=e.flags;if(2&t){try{if(27!==e.tag){e:{for(var n=e.return;null!==n;){if(Ki(n)){var r=n;break e}n=n.return}throw Error(o(160))}switch(r.tag){case 27:var a=r.stateNode;Xi(e,Gi(e),a);break;case 5:var l=r.stateNode;32&r.flags&&(xt(l,""),r.flags&=-33),Xi(e,Gi(e),l);break;case 3:case 4:var i=r.stateNode.containerInfo;Yi(e,Gi(e),i);break;default:throw Error(o(161))}}}catch(t){rc(e,e.return,t)}e.flags&=-3}4096&t&&(e.flags&=-4097)}function vu(e){if(1024&e.subtreeFlags)for(e=e.child;null!==e;){var t=e;vu(t),5===t.tag&&1024&t.flags&&t.stateNode.reset(),e=e.sibling}}function gu(e,t){if(8772&t.subtreeFlags)for(t=t.child;null!==t;)au(e,t.alternate,t),t=t.sibling}function yu(e){for(e=e.child;null!==e;){var t=e;switch(t.tag){case 0:case 11:case 14:case 15:$i(4,t,t.return),yu(t);break;case 1:qi(t,t.return);var n=t.stateNode;"function"==typeof n.componentWillUnmount&&Vi(t,t.return,n),yu(t);break;case 26:case 27:case 5:qi(t,t.return),yu(t);break;case 22:qi(t,t.return),null===t.memoizedState&&yu(t);break;default:yu(t)}e=e.sibling}}function bu(e,t,n){for(n=n&&!!(8772&t.subtreeFlags),t=t.child;null!==t;){var r=t.alternate,a=e,l=t,o=l.flags;switch(l.tag){case 0:case 11:case 15:bu(a,l,n),Ui(4,l);break;case 1:if(bu(a,l,n),"function"==typeof(a=(r=l).stateNode).componentDidMount)try{a.componentDidMount()}catch(e){rc(r,r.return,e)}if(null!==(a=(r=l).updateQueue)){var i=r.stateNode;try{var u=a.shared.hiddenCallbacks;if(null!==u)for(a.shared.hiddenCallbacks=null,a=0;a<u.length;a++)Di(u[a],i)}catch(e){rc(r,r.return,e)}}n&&64&o&&Hi(l),Bi(l,l.return);break;case 26:case 27:case 5:bu(a,l,n),n&&null===r&&4&o&&Wi(l),Bi(l,l.return);break;case 12:default:bu(a,l,n);break;case 13:bu(a,l,n),n&&4&o&&cu(a,l);break;case 22:null===l.memoizedState&&bu(a,l,n),Bi(l,l.return)}t=t.sibling}}function ku(e,t){var n=null;null!==e&&null!==e.memoizedState&&null!==e.memoizedState.cachePool&&(n=e.memoizedState.cachePool.pool),e=null,null!==t.memoizedState&&null!==t.memoizedState.cachePool&&(e=t.memoizedState.cachePool.pool),e!==n&&(null!=e&&e.refCount++,null!=n&&$a(n))}function wu(e,t){e=null,null!==t.alternate&&(e=t.alternate.memoizedState.cache),(t=t.memoizedState.cache)!==e&&(t.refCount++,null!=e&&$a(e))}function xu(e,t,n,r){if(10256&t.subtreeFlags)for(t=t.child;null!==t;)Su(e,t,n,r),t=t.sibling}function Su(e,t,n,r){var a=t.flags;switch(t.tag){case 0:case 11:case 15:xu(e,t,n,r),2048&a&&Ui(9,t);break;case 3:xu(e,t,n,r),2048&a&&(e=null,null!==t.alternate&&(e=t.alternate.memoizedState.cache),(t=t.memoizedState.cache)!==e&&(t.refCount++,null!=e&&$a(e)));break;case 12:if(2048&a){xu(e,t,n,r),e=t.stateNode;try{var l=t.memoizedProps,o=l.id,i=l.onPostCommit;"function"==typeof i&&i(o,null===t.alternate?"mount":"update",e.passiveEffectDuration,-0)}catch(e){rc(t,t.return,e)}}else xu(e,t,n,r);break;case 23:break;case 22:l=t.stateNode,null!==t.memoizedState?4&l._visibility?xu(e,t,n,r):Cu(e,t):4&l._visibility?xu(e,t,n,r):(l._visibility|=4,Eu(e,t,n,r,!!(10256&t.subtreeFlags))),2048&a&&ku(t.alternate,t);break;case 24:xu(e,t,n,r),2048&a&&wu(t.alternate,t);break;default:xu(e,t,n,r)}}function Eu(e,t,n,r,a){for(a=a&&!!(10256&t.subtreeFlags),t=t.child;null!==t;){var l=e,o=t,i=n,u=r,s=o.flags;switch(o.tag){case 0:case 11:case 15:Eu(l,o,i,u,a),Ui(8,o);break;case 23:break;case 22:var c=o.stateNode;null!==o.memoizedState?4&c._visibility?Eu(l,o,i,u,a):Cu(l,o):(c._visibility|=4,Eu(l,o,i,u,a)),a&&2048&s&&ku(o.alternate,o);break;case 24:Eu(l,o,i,u,a),a&&2048&s&&wu(o.alternate,o);break;default:Eu(l,o,i,u,a)}t=t.sibling}}function Cu(e,t){if(10256&t.subtreeFlags)for(t=t.child;null!==t;){var n=e,r=t,a=r.flags;switch(r.tag){case 22:Cu(n,r),2048&a&&ku(r.alternate,r);break;case 24:Cu(n,r),2048&a&&wu(r.alternate,r);break;default:Cu(n,r)}t=t.sibling}}var Nu=8192;function _u(e){if(e.subtreeFlags&Nu)for(e=e.child;null!==e;)Pu(e),e=e.sibling}function Pu(e){switch(e.tag){case 26:_u(e),e.flags&Nu&&null!==e.memoizedState&&function(e,t,n){if(null===Tf)throw Error(o(475));var r=Tf;if(!("stylesheet"!==t.type||"string"==typeof n.media&&!1===matchMedia(n.media).matches||4&t.state.loading)){if(null===t.instance){var a=yf(n.href),l=e.querySelector(bf(a));if(l)return null!==(e=l._p)&&"object"==typeof e&&"function"==typeof e.then&&(r.count++,r=Lf.bind(r),e.then(r,r)),t.state.loading|=4,t.instance=l,void Xe(l);l=e.ownerDocument||e,n=kf(n),(a=ff.get(a))&&Cf(n,a),Xe(l=l.createElement("link"));var i=l;i._p=new Promise((function(e,t){i.onload=e,i.onerror=t})),Wc(l,"link",n),t.instance=l}null===r.stylesheets&&(r.stylesheets=new Map),r.stylesheets.set(t,e),(e=t.state.preload)&&!(3&t.state.loading)&&(r.count++,t=Lf.bind(r),e.addEventListener("load",t),e.addEventListener("error",t))}}(pu,e.memoizedState,e.memoizedProps);break;case 5:default:_u(e);break;case 3:case 4:var t=pu;pu=pf(e.stateNode.containerInfo),_u(e),pu=t;break;case 22:null===e.memoizedState&&(null!==(t=e.alternate)&&null!==t.memoizedState?(t=Nu,Nu=16777216,_u(e),Nu=t):_u(e))}}function zu(e){var t=e.alternate;if(null!==t&&null!==(e=t.child)){t.child=null;do{t=e.sibling,e.sibling=null,e=t}while(null!==e)}}function ju(e){var t=e.deletions;if(16&e.flags){if(null!==t)for(var n=0;n<t.length;n++){var r=t[n];nu=r,Lu(r,e)}zu(e)}if(10256&e.subtreeFlags)for(e=e.child;null!==e;)Tu(e),e=e.sibling}function Tu(e){switch(e.tag){case 0:case 11:case 15:ju(e),2048&e.flags&&$i(9,e,e.return);break;case 3:case 12:default:ju(e);break;case 22:var t=e.stateNode;null!==e.memoizedState&&4&t._visibility&&(null===e.return||13!==e.return.tag)?(t._visibility&=-5,Ou(e)):ju(e)}}function Ou(e){var t=e.deletions;if(16&e.flags){if(null!==t)for(var n=0;n<t.length;n++){var r=t[n];nu=r,Lu(r,e)}zu(e)}for(e=e.child;null!==e;){switch((t=e).tag){case 0:case 11:case 15:$i(8,t,t.return),Ou(t);break;case 22:4&(n=t.stateNode)._visibility&&(n._visibility&=-5,Ou(t));break;default:Ou(t)}e=e.sibling}}function Lu(e,t){for(;null!==nu;){var n=nu;switch(n.tag){case 0:case 11:case 15:$i(8,n,t);break;case 23:case 22:if(null!==n.memoizedState&&null!==n.memoizedState.cachePool){var r=n.memoizedState.cachePool.pool;null!=r&&r.refCount++}break;case 24:$a(n.memoizedState.cache)}if(null!==(r=n.child))r.return=n,nu=r;else e:for(n=e;null!==nu;){var a=(r=nu).sibling,l=r.return;if(lu(r),r===n){nu=null;break e}if(null!==a){a.return=l,nu=a;break e}nu=l}}}function Mu(e,t,n,r){this.tag=e,this.key=n,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.refCleanup=this.ref=null,this.pendingProps=t,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=r,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function Au(e,t,n,r){return new Mu(e,t,n,r)}function Ru(e){return!(!(e=e.prototype)||!e.isReactComponent)}function Fu(e,t){var n=e.alternate;return null===n?((n=Au(e.tag,t,e.key,e.mode)).elementType=e.elementType,n.type=e.type,n.stateNode=e.stateNode,n.alternate=e,e.alternate=n):(n.pendingProps=t,n.type=e.type,n.flags=0,n.subtreeFlags=0,n.deletions=null),n.flags=31457280&e.flags,n.childLanes=e.childLanes,n.lanes=e.lanes,n.child=e.child,n.memoizedProps=e.memoizedProps,n.memoizedState=e.memoizedState,n.updateQueue=e.updateQueue,t=e.dependencies,n.dependencies=null===t?null:{lanes:t.lanes,firstContext:t.firstContext},n.sibling=e.sibling,n.index=e.index,n.ref=e.ref,n.refCleanup=e.refCleanup,n}function Du(e,t){e.flags&=31457282;var n=e.alternate;return null===n?(e.childLanes=0,e.lanes=t,e.child=null,e.subtreeFlags=0,e.memoizedProps=null,e.memoizedState=null,e.updateQueue=null,e.dependencies=null,e.stateNode=null):(e.childLanes=n.childLanes,e.lanes=n.lanes,e.child=n.child,e.subtreeFlags=0,e.deletions=null,e.memoizedProps=n.memoizedProps,e.memoizedState=n.memoizedState,e.updateQueue=n.updateQueue,e.type=n.type,t=n.dependencies,e.dependencies=null===t?null:{lanes:t.lanes,firstContext:t.firstContext}),e}function Iu(e,t,n,r,a,l){var i=0;if(r=e,"function"==typeof e)Ru(e)&&(i=1);else if("string"==typeof e)i=function(e,t,n){if(1===n||null!=t.itemProp)return!1;switch(e){case"meta":case"title":return!0;case"style":if("string"!=typeof t.precedence||"string"!=typeof t.href||""===t.href)break;return!0;case"link":if("string"!=typeof t.rel||"string"!=typeof t.href||""===t.href||t.onLoad||t.onError)break;return"stylesheet"!==t.rel||(e=t.disabled,"string"==typeof t.precedence&&null==e);case"script":if(t.async&&"function"!=typeof t.async&&"symbol"!=typeof t.async&&!t.onLoad&&!t.onError&&t.src&&"string"==typeof t.src)return!0}return!1}(e,n,G.current)?26:"html"===e||"head"===e||"body"===e?27:5;else e:switch(e){case f:return Uu(n.children,a,l,t);case d:i=8,a|=24;break;case p:return(e=Au(12,n,t,2|a)).elementType=p,e.lanes=l,e;case y:return(e=Au(13,n,t,a)).elementType=y,e.lanes=l,e;case b:return(e=Au(19,n,t,a)).elementType=b,e.lanes=l,e;case x:return $u(n,a,l,t);default:if("object"==typeof e&&null!==e)switch(e.$$typeof){case h:case v:i=10;break e;case m:i=9;break e;case g:i=11;break e;case k:i=14;break e;case w:i=16,r=null;break e}i=29,n=Error(o(130,null===e?"null":typeof e,"")),r=null}return(t=Au(i,n,t,a)).elementType=e,t.type=r,t.lanes=l,t}function Uu(e,t,n,r){return(e=Au(7,e,r,t)).lanes=n,e}function $u(e,t,n,r){(e=Au(22,e,r,t)).elementType=x,e.lanes=n;var a={_visibility:1,_pendingVisibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null,_current:null,detach:function(){var e=a._current;if(null===e)throw Error(o(456));if(!(2&a._pendingVisibility)){var t=zr(e,2);null!==t&&(a._pendingVisibility|=2,Os(t,0,2))}},attach:function(){var e=a._current;if(null===e)throw Error(o(456));if(2&a._pendingVisibility){var t=zr(e,2);null!==t&&(a._pendingVisibility&=-3,Os(t,0,2))}}};return e.stateNode=a,e}function Hu(e,t,n){return(e=Au(6,e,null,t)).lanes=n,e}function Vu(e,t,n){return(t=Au(4,null!==e.children?e.children:[],e.key,t)).lanes=n,t.stateNode={containerInfo:e.containerInfo,pendingChildren:null,implementation:e.implementation},t}function Bu(e){e.flags|=4}function qu(e,t){if("stylesheet"!==t.type||4&t.state.loading)e.flags&=-16777217;else if(e.flags|=16777216,!jf(t)){if(null!==(t=Pa.current)&&((4194176&rs)===rs?null!==za:(62914560&rs)!==rs&&!(536870912&rs)||t!==za))throw da=ua,ia;e.flags|=8192}}function Wu(e,t){null!==t&&(e.flags|=4),16384&e.flags&&(t=22!==e.tag?je():536870912,e.lanes|=t,ms|=t)}function Qu(e,t){if(!Yr)switch(e.tailMode){case"hidden":t=e.tail;for(var n=null;null!==t;)null!==t.alternate&&(n=t),t=t.sibling;null===n?e.tail=null:n.sibling=null;break;case"collapsed":n=e.tail;for(var r=null;null!==n;)null!==n.alternate&&(r=n),n=n.sibling;null===r?t||null===e.tail?e.tail=null:e.tail.sibling=null:r.sibling=null}}function Ku(e){var t=null!==e.alternate&&e.alternate.child===e.child,n=0,r=0;if(t)for(var a=e.child;null!==a;)n|=a.lanes|a.childLanes,r|=31457280&a.subtreeFlags,r|=31457280&a.flags,a.return=e,a=a.sibling;else for(a=e.child;null!==a;)n|=a.lanes|a.childLanes,r|=a.subtreeFlags,r|=a.flags,a.return=e,a=a.sibling;return e.subtreeFlags|=r,e.childLanes=n,t}function Gu(e,t,n){var r=t.pendingProps;switch(Qr(t),t.tag){case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:case 1:return Ku(t),null;case 3:return n=t.stateNode,r=null,null!==e&&(r=e.memoizedState.cache),t.memoizedState.cache!==r&&(t.flags|=2048),bi(Ia),ee(),n.pendingContext&&(n.context=n.pendingContext,n.pendingContext=null),null!==e&&null!==e.child||(ra(t)?Bu(t):null===e||e.memoizedState.isDehydrated&&!(256&t.flags)||(t.flags|=1024,null!==Xr&&(Ms(Xr),Xr=null))),Ku(t),null;case 26:return n=t.memoizedState,null===e?(Bu(t),null!==n?(Ku(t),qu(t,n)):(Ku(t),t.flags&=-16777217)):n?n!==e.memoizedState?(Bu(t),Ku(t),qu(t,n)):(Ku(t),t.flags&=-16777217):(e.memoizedProps!==r&&Bu(t),Ku(t),t.flags&=-16777217),null;case 27:ne(t),n=X.current;var a=t.type;if(null!==e&&null!=t.stateNode)e.memoizedProps!==r&&Bu(t);else{if(!r){if(null===t.stateNode)throw Error(o(166));return Ku(t),null}e=G.current,ra(t)?ta(t):(e=cf(a,r,n),t.stateNode=e,Bu(t))}return Ku(t),null;case 5:if(ne(t),n=t.type,null!==e&&null!=t.stateNode)e.memoizedProps!==r&&Bu(t);else{if(!r){if(null===t.stateNode)throw Error(o(166));return Ku(t),null}if(e=G.current,ra(t))ta(t);else{switch(a=Gc(X.current),e){case 1:e=a.createElementNS("http://www.w3.org/2000/svg",n);break;case 2:e=a.createElementNS("http://www.w3.org/1998/Math/MathML",n);break;default:switch(n){case"svg":e=a.createElementNS("http://www.w3.org/2000/svg",n);break;case"math":e=a.createElementNS("http://www.w3.org/1998/Math/MathML",n);break;case"script":(e=a.createElement("div")).innerHTML="<script><\/script>",e=e.removeChild(e.firstChild);break;case"select":e="string"==typeof r.is?a.createElement("select",{is:r.is}):a.createElement("select"),r.multiple?e.multiple=!0:r.size&&(e.size=r.size);break;default:e="string"==typeof r.is?a.createElement(n,{is:r.is}):a.createElement(n)}}e[De]=t,e[Ie]=r;e:for(a=t.child;null!==a;){if(5===a.tag||6===a.tag)e.appendChild(a.stateNode);else if(4!==a.tag&&27!==a.tag&&null!==a.child){a.child.return=a,a=a.child;continue}if(a===t)break e;for(;null===a.sibling;){if(null===a.return||a.return===t)break e;a=a.return}a.sibling.return=a.return,a=a.sibling}t.stateNode=e;e:switch(Wc(e,n,r),n){case"button":case"input":case"select":case"textarea":e=!!r.autoFocus;break e;case"img":e=!0;break e;default:e=!1}e&&Bu(t)}}return Ku(t),t.flags&=-16777217,null;case 6:if(e&&null!=t.stateNode)e.memoizedProps!==r&&Bu(t);else{if("string"!=typeof r&&null===t.stateNode)throw Error(o(166));if(e=X.current,ra(t)){if(e=t.stateNode,n=t.memoizedProps,r=null,null!==(a=Kr))switch(a.tag){case 27:case 5:r=a.memoizedProps}e[De]=t,(e=!!(e.nodeValue===n||null!==r&&!0===r.suppressHydrationWarning||Hc(e.nodeValue,n)))||ea(t)}else(e=Gc(e).createTextNode(r))[De]=t,t.stateNode=e}return Ku(t),null;case 13:if(r=t.memoizedState,null===e||null!==e.memoizedState&&null!==e.memoizedState.dehydrated){if(a=ra(t),null!==r&&null!==r.dehydrated){if(null===e){if(!a)throw Error(o(318));if(!(a=null!==(a=t.memoizedState)?a.dehydrated:null))throw Error(o(317));a[De]=t}else aa(),!(128&t.flags)&&(t.memoizedState=null),t.flags|=4;Ku(t),a=!1}else null!==Xr&&(Ms(Xr),Xr=null),a=!0;if(!a)return 256&t.flags?(La(t),t):(La(t),null)}if(La(t),128&t.flags)return t.lanes=n,t;if(n=null!==r,e=null!==e&&null!==e.memoizedState,n){a=null,null!==(r=t.child).alternate&&null!==r.alternate.memoizedState&&null!==r.alternate.memoizedState.cachePool&&(a=r.alternate.memoizedState.cachePool.pool);var l=null;null!==r.memoizedState&&null!==r.memoizedState.cachePool&&(l=r.memoizedState.cachePool.pool),l!==a&&(r.flags|=2048)}return n!==e&&n&&(t.child.flags|=8192),Wu(t,t.updateQueue),Ku(t),null;case 4:return ee(),null===e&&Oc(t.stateNode.containerInfo),Ku(t),null;case 10:return bi(t.type),Ku(t),null;case 19:if(Q(Ma),null===(a=t.memoizedState))return Ku(t),null;if(r=!!(128&t.flags),null===(l=a.rendering))if(r)Qu(a,!1);else{if(0!==cs||null!==e&&128&e.flags)for(e=t.child;null!==e;){if(null!==(l=Aa(e))){for(t.flags|=128,Qu(a,!1),e=l.updateQueue,t.updateQueue=e,Wu(t,e),t.subtreeFlags=0,e=n,n=t.child;null!==n;)Du(n,e),n=n.sibling;return K(Ma,1&Ma.current|2),t.child}e=e.sibling}null!==a.tail&&ue()>ks&&(t.flags|=128,r=!0,Qu(a,!1),t.lanes=4194304)}else{if(!r)if(null!==(e=Aa(l))){if(t.flags|=128,r=!0,e=e.updateQueue,t.updateQueue=e,Wu(t,e),Qu(a,!0),null===a.tail&&"hidden"===a.tailMode&&!l.alternate&&!Yr)return Ku(t),null}else 2*ue()-a.renderingStartTime>ks&&536870912!==n&&(t.flags|=128,r=!0,Qu(a,!1),t.lanes=4194304);a.isBackwards?(l.sibling=t.child,t.child=l):(null!==(e=a.last)?e.sibling=l:t.child=l,a.last=l)}return null!==a.tail?(t=a.tail,a.rendering=t,a.tail=t.sibling,a.renderingStartTime=ue(),t.sibling=null,e=Ma.current,K(Ma,r?1&e|2:1&e),t):(Ku(t),null);case 22:case 23:return La(t),_a(),r=null!==t.memoizedState,null!==e?null!==e.memoizedState!==r&&(t.flags|=8192):r&&(t.flags|=8192),r?!!(536870912&n)&&!(128&t.flags)&&(Ku(t),6&t.subtreeFlags&&(t.flags|=8192)):Ku(t),null!==(n=t.updateQueue)&&Wu(t,n.retryQueue),n=null,null!==e&&null!==e.memoizedState&&null!==e.memoizedState.cachePool&&(n=e.memoizedState.cachePool.pool),r=null,null!==t.memoizedState&&null!==t.memoizedState.cachePool&&(r=t.memoizedState.cachePool.pool),r!==n&&(t.flags|=2048),null!==e&&Q(Ka),null;case 24:return n=null,null!==e&&(n=e.memoizedState.cache),t.memoizedState.cache!==n&&(t.flags|=2048),bi(Ia),Ku(t),null;case 25:return null}throw Error(o(156,t.tag))}function Yu(e,t){switch(Qr(t),t.tag){case 1:return 65536&(e=t.flags)?(t.flags=-65537&e|128,t):null;case 3:return bi(Ia),ee(),65536&(e=t.flags)&&!(128&e)?(t.flags=-65537&e|128,t):null;case 26:case 27:case 5:return ne(t),null;case 13:if(La(t),null!==(e=t.memoizedState)&&null!==e.dehydrated){if(null===t.alternate)throw Error(o(340));aa()}return 65536&(e=t.flags)?(t.flags=-65537&e|128,t):null;case 19:return Q(Ma),null;case 4:return ee(),null;case 10:return bi(t.type),null;case 22:case 23:return La(t),_a(),null!==e&&Q(Ka),65536&(e=t.flags)?(t.flags=-65537&e|128,t):null;case 24:return bi(Ia),null;default:return null}}function Xu(e,t){switch(Qr(t),t.tag){case 3:bi(Ia),ee();break;case 26:case 27:case 5:ne(t);break;case 4:ee();break;case 13:La(t);break;case 19:Q(Ma);break;case 10:bi(t.type);break;case 22:case 23:La(t),_a(),null!==e&&Q(Ka);break;case 24:bi(Ia)}}var Zu={getCacheForType:function(e){var t=Ci(Ia),n=t.data.get(e);return void 0===n&&(n=e(),t.data.set(e,n)),n}},Ju="function"==typeof WeakMap?WeakMap:Map,es=0,ts=null,ns=null,rs=0,as=0,ls=null,os=!1,is=!1,us=!1,ss=0,cs=0,fs=0,ds=0,ps=0,hs=0,ms=0,vs=null,gs=null,ys=!1,bs=0,ks=1/0,ws=null,xs=null,Ss=!1,Es=null,Cs=0,Ns=0,_s=null,Ps=0,zs=null;function js(){return 2&es&&0!==rs?rs&-rs:null!==j.T?0!==Ba?Ba:wc():Re()}function Ts(){0===hs&&(hs=536870912&rs&&!Yr?536870912:ze());var e=Pa.current;return null!==e&&(e.flags|=32),hs}function Os(e,t,n){(e===ts&&2===as||null!==e.cancelPendingCommit)&&(Us(e,0),Fs(e,rs,hs,!1)),Oe(e,n),2&es&&e===ts||(e===ts&&(!(2&es)&&(ds|=n),4===cs&&Fs(e,rs,hs,!1)),mc(e))}function Ls(e,t,n){if(6&es)throw Error(o(327));for(var r=!n&&!(60&t)&&!(t&e.expiredLanes)||_e(e,t),a=r?function(e,t){var n=es;es|=2;var r=Hs(),a=Vs();ts!==e||rs!==t?(ws=null,ks=ue()+500,Us(e,t)):is=_e(e,t);e:for(;;)try{if(0!==as&&null!==ns){t=ns;var l=ls;t:switch(as){case 1:as=0,ls=null,Ys(e,t,l,1);break;case 2:if(sa(l)){as=0,ls=null,Gs(t);break}t=function(){2===as&&ts===e&&(as=7),mc(e)},l.then(t,t);break e;case 3:as=7;break e;case 4:as=5;break e;case 7:sa(l)?(as=0,ls=null,Gs(t)):(as=0,ls=null,Ys(e,t,l,7));break;case 5:var i=null;switch(ns.tag){case 26:i=ns.memoizedState;case 5:case 27:var u=ns;if(!i||jf(i)){as=0,ls=null;var s=u.sibling;if(null!==s)ns=s;else{var c=u.return;null!==c?(ns=c,Xs(c)):ns=null}break t}}as=0,ls=null,Ys(e,t,l,5);break;case 6:as=0,ls=null,Ys(e,t,l,6);break;case 8:Is(),cs=6;break e;default:throw Error(o(462))}}Qs();break}catch(t){$s(e,t)}return gi=vi=null,j.H=r,j.A=a,es=n,null!==ns?0:(ts=null,rs=0,Nr(),cs)}(e,t):qs(e,t,!0),l=r;;){if(0===a){is&&!r&&Fs(e,t,0,!1);break}if(6===a)Fs(e,t,0,!os);else{if(n=e.current.alternate,l&&!Rs(n)){a=qs(e,t,!1),l=!1;continue}if(2===a){if(l=t,e.errorRecoveryDisabledLanes&l)var i=0;else i=0!=(i=-536870913&e.pendingLanes)?i:536870912&i?536870912:0;if(0!==i){t=i;e:{var u=e;a=vs;var s=u.current.memoizedState.isDehydrated;if(s&&(Us(u,i).flags|=256),2!==(i=qs(u,i,!1))){if(us&&!s){u.errorRecoveryDisabledLanes|=l,ds|=l,a=4;break e}l=gs,gs=a,null!==l&&Ms(l)}a=i}if(l=!1,2!==a)continue}}if(1===a){Us(e,0),Fs(e,t,0,!0);break}e:{switch(r=e,a){case 0:case 1:throw Error(o(345));case 4:if((4194176&t)===t){Fs(r,t,hs,!os);break e}break;case 2:gs=null;break;case 3:case 5:break;default:throw Error(o(329))}if(r.finishedWork=n,r.finishedLanes=t,(62914560&t)===t&&10<(l=bs+300-ue())){if(Fs(r,t,hs,!os),0!==Ne(r,0))break e;r.timeoutHandle=ef(As.bind(null,r,n,gs,ws,ys,t,hs,ds,ms,os,2,-0,0),l)}else As(r,n,gs,ws,ys,t,hs,ds,ms,os,0,-0,0)}}break}mc(e)}function Ms(e){null===gs?gs=e:gs.push.apply(gs,e)}function As(e,t,n,r,a,l,i,u,s,c,f,d,p){var h=t.subtreeFlags;if((8192&h||!(16785408&~h))&&(Tf={stylesheets:null,count:0,unsuspend:Of},Pu(t),null!==(t=function(){if(null===Tf)throw Error(o(475));var e=Tf;return e.stylesheets&&0===e.count&&Af(e,e.stylesheets),0<e.count?function(t){var n=setTimeout((function(){if(e.stylesheets&&Af(e,e.stylesheets),e.unsuspend){var t=e.unsuspend;e.unsuspend=null,t()}}),6e4);return e.unsuspend=t,function(){e.unsuspend=null,clearTimeout(n)}}:null}())))return e.cancelPendingCommit=t(Js.bind(null,e,n,r,a,i,u,s,1,d,p)),void Fs(e,l,i,!c);Js(e,n,r,a,i,u,s)}function Rs(e){for(var t=e;;){var n=t.tag;if((0===n||11===n||15===n)&&16384&t.flags&&null!==(n=t.updateQueue)&&null!==(n=n.stores))for(var r=0;r<n.length;r++){var a=n[r],l=a.getSnapshot;a=a.value;try{if(!Kn(l(),a))return!1}catch(e){return!1}}if(n=t.child,16384&t.subtreeFlags&&null!==n)n.return=t,t=n;else{if(t===e)break;for(;null===t.sibling;){if(null===t.return||t.return===e)return!0;t=t.return}t.sibling.return=t.return,t=t.sibling}}return!0}function Fs(e,t,n,r){t&=~ps,t&=~ds,e.suspendedLanes|=t,e.pingedLanes&=~t,r&&(e.warmLanes|=t),r=e.expirationTimes;for(var a=t;0<a;){var l=31-ke(a),o=1<<l;r[l]=-1,a&=~o}0!==n&&Le(e,n,t)}function Ds(){return!!(6&es)||(vc(0,!1),!1)}function Is(){if(null!==ns){if(0===as)var e=ns.return;else gi=vi=null,gl(e=ns),ha=null,ma=0,e=ns;for(;null!==e;)Xu(e.alternate,e),e=e.return;ns=null}}function Us(e,t){e.finishedWork=null,e.finishedLanes=0;var n=e.timeoutHandle;-1!==n&&(e.timeoutHandle=-1,tf(n)),null!==(n=e.cancelPendingCommit)&&(e.cancelPendingCommit=null,n()),Is(),ts=e,ns=n=Fu(e.current,null),rs=t,as=0,ls=null,os=!1,is=_e(e,t),us=!1,ms=hs=ps=ds=fs=cs=0,gs=vs=null,ys=!1,8&t&&(t|=32&t);var r=e.entangledLanes;if(0!==r)for(e=e.entanglements,r&=t;0<r;){var a=31-ke(r),l=1<<a;t|=e[a],r&=~l}return ss=t,Nr(),n}function $s(e,t){Ja=null,j.H=Co,t===oa?(t=pa(),as=3):t===ia?(t=pa(),as=4):as=t===Vo?8:null!==t&&"object"==typeof t&&"function"==typeof t.then?6:1,ls=t,null===ns&&(cs=1,Do(e,Mr(t,e.current)))}function Hs(){var e=j.H;return j.H=Co,null===e?Co:e}function Vs(){var e=j.A;return j.A=Zu,e}function Bs(){cs=4,os||(4194176&rs)!==rs&&null!==Pa.current||(is=!0),!(134217727&fs)&&!(134217727&ds)||null===ts||Fs(ts,rs,hs,!1)}function qs(e,t,n){var r=es;es|=2;var a=Hs(),l=Vs();ts===e&&rs===t||(ws=null,Us(e,t)),t=!1;var o=cs;e:for(;;)try{if(0!==as&&null!==ns){var i=ns,u=ls;switch(as){case 8:Is(),o=6;break e;case 3:case 2:case 6:null===Pa.current&&(t=!0);var s=as;if(as=0,ls=null,Ys(e,i,u,s),n&&is){o=0;break e}break;default:s=as,as=0,ls=null,Ys(e,i,u,s)}}Ws(),o=cs;break}catch(t){$s(e,t)}return t&&e.shellSuspendCounter++,gi=vi=null,es=r,j.H=a,j.A=l,null===ns&&(ts=null,rs=0,Nr()),o}function Ws(){for(;null!==ns;)Ks(ns)}function Qs(){for(;null!==ns&&!oe();)Ks(ns)}function Ks(e){var t=hi(e.alternate,e,ss);e.memoizedProps=e.pendingProps,null===t?Xs(e):ns=t}function Gs(e){var t=e,n=t.alternate;switch(t.tag){case 15:case 0:t=Jo(n,t,t.pendingProps,t.type,void 0,rs);break;case 11:t=Jo(n,t,t.pendingProps,t.type.render,t.ref,rs);break;case 5:gl(t);default:Xu(n,t),t=hi(n,t=ns=Du(t,ss),ss)}e.memoizedProps=e.pendingProps,null===t?Xs(e):ns=t}function Ys(e,t,n,r){gi=vi=null,gl(t),ha=null,ma=0;var a=t.return;try{if(function(e,t,n,r,a){if(n.flags|=32768,null!==r&&"object"==typeof r&&"function"==typeof r.then){if(null!==(t=n.alternate)&&xi(t,n,a,!0),null!==(n=Pa.current)){switch(n.tag){case 13:return null===za?Bs():null===n.alternate&&0===cs&&(cs=3),n.flags&=-257,n.flags|=65536,n.lanes=a,r===ua?n.flags|=16384:(null===(t=n.updateQueue)?n.updateQueue=new Set([r]):t.add(r),ac(e,r,a)),!1;case 22:return n.flags|=65536,r===ua?n.flags|=16384:(null===(t=n.updateQueue)?(t={transitions:null,markerInstances:null,retryQueue:new Set([r])},n.updateQueue=t):null===(n=t.retryQueue)?t.retryQueue=new Set([r]):n.add(r),ac(e,r,a)),!1}throw Error(o(435,n.tag))}return ac(e,r,a),Bs(),!1}if(Yr)return null!==(t=Pa.current)?(!(65536&t.flags)&&(t.flags|=256),t.flags|=65536,t.lanes=a,r!==Jr&&la(Mr(e=Error(o(422),{cause:r}),n))):(r!==Jr&&la(Mr(t=Error(o(423),{cause:r}),n)),(e=e.current.alternate).flags|=65536,a&=-a,e.lanes|=a,r=Mr(r,n),Mi(e,a=Uo(e.stateNode,r,a)),4!==cs&&(cs=2)),!1;var l=Error(o(520),{cause:r});if(l=Mr(l,n),null===vs?vs=[l]:vs.push(l),4!==cs&&(cs=2),null===t)return!0;r=Mr(r,n),n=t;do{switch(n.tag){case 3:return n.flags|=65536,e=a&-a,n.lanes|=e,Mi(n,e=Uo(n.stateNode,r,e)),!1;case 1:if(t=n.type,l=n.stateNode,!(128&n.flags||"function"!=typeof t.getDerivedStateFromError&&(null===l||"function"!=typeof l.componentDidCatch||null!==xs&&xs.has(l))))return n.flags|=65536,a&=-a,n.lanes|=a,Ho(a=$o(a),e,n,r),Mi(n,a),!1}n=n.return}while(null!==n);return!1}(e,a,t,n,rs))return cs=1,Do(e,Mr(n,e.current)),void(ns=null)}catch(t){if(null!==a)throw ns=a,t;return cs=1,Do(e,Mr(n,e.current)),void(ns=null)}32768&t.flags?(Yr||1===r?e=!0:is||536870912&rs?e=!1:(os=e=!0,(2===r||3===r||6===r)&&null!==(r=Pa.current)&&13===r.tag&&(r.flags|=16384)),Zs(t,e)):Xs(t)}function Xs(e){var t=e;do{if(32768&t.flags)return void Zs(t,os);e=t.return;var n=Gu(t.alternate,t,ss);if(null!==n)return void(ns=n);if(null!==(t=t.sibling))return void(ns=t);ns=t=e}while(null!==t);0===cs&&(cs=5)}function Zs(e,t){do{var n=Yu(e.alternate,e);if(null!==n)return n.flags&=32767,void(ns=n);if(null!==(n=e.return)&&(n.flags|=32768,n.subtreeFlags=0,n.deletions=null),!t&&null!==(e=e.sibling))return void(ns=e);ns=e=n}while(null!==e);cs=6,ns=null}function Js(e,t,n,r,a,l,i,u,s,c){var f=j.T,d=H.p;try{H.p=2,j.T=null,function(e,t,n,r,a,l,i,u){do{tc()}while(null!==Es);if(6&es)throw Error(o(327));var s=e.finishedWork;if(r=e.finishedLanes,null===s)return null;if(e.finishedWork=null,e.finishedLanes=0,s===e.current)throw Error(o(177));e.callbackNode=null,e.callbackPriority=0,e.cancelPendingCommit=null;var c=s.lanes|s.childLanes;if(function(e,t,n,r,a,l){var o=e.pendingLanes;e.pendingLanes=n,e.suspendedLanes=0,e.pingedLanes=0,e.warmLanes=0,e.expiredLanes&=n,e.entangledLanes&=n,e.errorRecoveryDisabledLanes&=n,e.shellSuspendCounter=0;var i=e.entanglements,u=e.expirationTimes,s=e.hiddenUpdates;for(n=o&~n;0<n;){var c=31-ke(n),f=1<<c;i[c]=0,u[c]=-1;var d=s[c];if(null!==d)for(s[c]=null,c=0;c<d.length;c++){var p=d[c];null!==p&&(p.lane&=-536870913)}n&=~f}0!==r&&Le(e,r,0),0!==l&&0===a&&0!==e.tag&&(e.suspendedLanes|=l&~(o&~t))}(e,r,c|=Cr,l,i,u),e===ts&&(ns=ts=null,rs=0),!(10256&s.subtreeFlags)&&!(10256&s.flags)||Ss||(Ss=!0,Ns=c,_s=n,ae(de,(function(){return tc(),null}))),n=!!(15990&s.flags),15990&s.subtreeFlags||n?(n=j.T,j.T=null,l=H.p,H.p=2,i=es,es|=4,function(e,t){if(e=e.containerInfo,Qc=qf,er(e=Jn(e))){if("selectionStart"in e)var n={start:e.selectionStart,end:e.selectionEnd};else e:{var r=(n=(n=e.ownerDocument)&&n.defaultView||window).getSelection&&n.getSelection();if(r&&0!==r.rangeCount){n=r.anchorNode;var a=r.anchorOffset,l=r.focusNode;r=r.focusOffset;try{n.nodeType,l.nodeType}catch(e){n=null;break e}var i=0,u=-1,s=-1,c=0,f=0,d=e,p=null;t:for(;;){for(var h;d!==n||0!==a&&3!==d.nodeType||(u=i+a),d!==l||0!==r&&3!==d.nodeType||(s=i+r),3===d.nodeType&&(i+=d.nodeValue.length),null!==(h=d.firstChild);)p=d,d=h;for(;;){if(d===e)break t;if(p===n&&++c===a&&(u=i),p===l&&++f===r&&(s=i),null!==(h=d.nextSibling))break;p=(d=p).parentNode}d=h}n=-1===u||-1===s?null:{start:u,end:s}}else n=null}n=n||{start:0,end:0}}else n=null;for(Kc={focusedElem:e,selectionRange:n},qf=!1,nu=t;null!==nu;)if(e=(t=nu).child,1028&t.subtreeFlags&&null!==e)e.return=t,nu=e;else for(;null!==nu;){switch(l=(t=nu).alternate,e=t.flags,t.tag){case 0:case 11:case 15:case 5:case 26:case 27:case 6:case 4:case 17:break;case 1:if(1024&e&&null!==l){e=void 0,n=t,a=l.memoizedProps,l=l.memoizedState,r=n.stateNode;try{var m=Lo(n.type,a,(n.elementType,n.type));e=r.getSnapshotBeforeUpdate(m,l),r.__reactInternalSnapshotBeforeUpdate=e}catch(e){rc(n,n.return,e)}}break;case 3:if(1024&e)if(9===(n=(e=t.stateNode.containerInfo).nodeType))of(e);else if(1===n)switch(e.nodeName){case"HEAD":case"HTML":case"BODY":of(e);break;default:e.textContent=""}break;default:if(1024&e)throw Error(o(163))}if(null!==(e=t.sibling)){e.return=t.return,nu=e;break}nu=t.return}m=ru,ru=!1}(e,s),hu(s,e),tr(Kc,e.containerInfo),qf=!!Qc,Kc=Qc=null,e.current=s,au(e,s.alternate,s),ie(),es=i,H.p=l,j.T=n):e.current=s,Ss?(Ss=!1,Es=e,Cs=r):ec(e,c),0===(c=e.pendingLanes)&&(xs=null),function(e){if(ye&&"function"==typeof ye.onCommitFiberRoot)try{ye.onCommitFiberRoot(ge,e,void 0,!(128&~e.current.flags))}catch(e){}}(s.stateNode),mc(e),null!==t)for(a=e.onRecoverableError,s=0;s<t.length;s++)a((c=t[s]).value,{componentStack:c.stack});!!(3&Cs)&&tc(),c=e.pendingLanes,4194218&r&&42&c?e===zs?Ps++:(Ps=0,zs=e):Ps=0,vc(0,!1)}(e,t,n,r,d,a,l,i)}finally{j.T=f,H.p=d}}function ec(e,t){0==(e.pooledCacheLanes&=t)&&null!=(t=e.pooledCache)&&(e.pooledCache=null,$a(t))}function tc(){if(null!==Es){var e=Es,t=Ns;Ns=0;var n=Ae(Cs),r=j.T,a=H.p;try{if(H.p=32>n?32:n,j.T=null,null===Es)var l=!1;else{n=_s,_s=null;var i=Es,u=Cs;if(Es=null,Cs=0,6&es)throw Error(o(331));var s=es;if(es|=4,Tu(i.current),Su(i,i.current,u,n),es=s,vc(0,!1),ye&&"function"==typeof ye.onPostCommitFiberRoot)try{ye.onPostCommitFiberRoot(ge,i)}catch(e){}l=!0}return l}finally{H.p=a,j.T=r,ec(e,t)}}return!1}function nc(e,t,n){t=Mr(n,t),null!==(e=Oi(e,t=Uo(e.stateNode,t,2),2))&&(Oe(e,2),mc(e))}function rc(e,t,n){if(3===e.tag)nc(e,e,n);else for(;null!==t;){if(3===t.tag){nc(t,e,n);break}if(1===t.tag){var r=t.stateNode;if("function"==typeof t.type.getDerivedStateFromError||"function"==typeof r.componentDidCatch&&(null===xs||!xs.has(r))){e=Mr(n,e),null!==(r=Oi(t,n=$o(2),2))&&(Ho(n,r,t,e),Oe(r,2),mc(r));break}}t=t.return}}function ac(e,t,n){var r=e.pingCache;if(null===r){r=e.pingCache=new Ju;var a=new Set;r.set(t,a)}else void 0===(a=r.get(t))&&(a=new Set,r.set(t,a));a.has(n)||(us=!0,a.add(n),e=lc.bind(null,e,t,n),t.then(e,e))}function lc(e,t,n){var r=e.pingCache;null!==r&&r.delete(t),e.pingedLanes|=e.suspendedLanes&n,e.warmLanes&=~n,ts===e&&(rs&n)===n&&(4===cs||3===cs&&(62914560&rs)===rs&&300>ue()-bs?!(2&es)&&Us(e,0):ps|=n,ms===rs&&(ms=0)),mc(e)}function oc(e,t){0===t&&(t=je()),null!==(e=zr(e,t))&&(Oe(e,t),mc(e))}function ic(e){var t=e.memoizedState,n=0;null!==t&&(n=t.retryLane),oc(e,n)}function uc(e,t){var n=0;switch(e.tag){case 13:var r=e.stateNode,a=e.memoizedState;null!==a&&(n=a.retryLane);break;case 19:r=e.stateNode;break;case 22:r=e.stateNode._retryCache;break;default:throw Error(o(314))}null!==r&&r.delete(t),oc(e,n)}var sc=null,cc=null,fc=!1,dc=!1,pc=!1,hc=0;function mc(e){var t;e!==cc&&null===e.next&&(null===cc?sc=cc=e:cc=cc.next=e),dc=!0,fc||(fc=!0,t=gc,rf((function(){6&es?ae(ce,t):t()})))}function vc(e,t){if(!pc&&dc){pc=!0;do{for(var n=!1,r=sc;null!==r;){if(!t)if(0!==e){var a=r.pendingLanes;if(0===a)var l=0;else{var o=r.suspendedLanes,i=r.pingedLanes;l=(1<<31-ke(42|e)+1)-1,l=201326677&(l&=a&~(o&~i))?201326677&l|1:l?2|l:0}0!==l&&(n=!0,kc(r,l))}else l=rs,!(3&(l=Ne(r,r===ts?l:0)))||_e(r,l)||(n=!0,kc(r,l));r=r.next}}while(n);pc=!1}}function gc(){dc=fc=!1;var e,t=0;0!==hc&&(((e=window.event)&&"popstate"===e.type?e!==Jc&&(Jc=e,!0):(Jc=null,!1))&&(t=hc),hc=0);for(var n=ue(),r=null,a=sc;null!==a;){var l=a.next,o=yc(a,n);0===o?(a.next=null,null===r?sc=l:r.next=l,null===l&&(cc=r)):(r=a,(0!==t||3&o)&&(dc=!0)),a=l}vc(t,!1)}function yc(e,t){for(var n=e.suspendedLanes,r=e.pingedLanes,a=e.expirationTimes,l=-62914561&e.pendingLanes;0<l;){var o=31-ke(l),i=1<<o,u=a[o];-1===u?i&n&&!(i&r)||(a[o]=Pe(i,t)):u<=t&&(e.expiredLanes|=i),l&=~i}if(n=rs,n=Ne(e,e===(t=ts)?n:0),r=e.callbackNode,0===n||e===t&&2===as||null!==e.cancelPendingCommit)return null!==r&&null!==r&&le(r),e.callbackNode=null,e.callbackPriority=0;if(!(3&n)||_e(e,n)){if((t=n&-n)===e.callbackPriority)return t;switch(null!==r&&le(r),Ae(n)){case 2:case 8:n=fe;break;case 32:default:n=de;break;case 268435456:n=he}return r=bc.bind(null,e),n=ae(n,r),e.callbackPriority=t,e.callbackNode=n,t}return null!==r&&null!==r&&le(r),e.callbackPriority=2,e.callbackNode=null,2}function bc(e,t){var n=e.callbackNode;if(tc()&&e.callbackNode!==n)return null;var r=rs;return 0===(r=Ne(e,e===ts?r:0))?null:(Ls(e,r,t),yc(e,ue()),null!=e.callbackNode&&e.callbackNode===n?bc.bind(null,e):null)}function kc(e,t){if(tc())return null;Ls(e,t,!0)}function wc(){return 0===hc&&(hc=ze()),hc}function xc(e){return null==e||"symbol"==typeof e||"boolean"==typeof e?null:"function"==typeof e?e:zt(""+e)}function Sc(e,t){var n=t.ownerDocument.createElement("input");return n.name=t.name,n.value=t.value,e.id&&n.setAttribute("form",e.id),t.parentNode.insertBefore(n,t),e=new FormData(e),n.parentNode.removeChild(n),e}for(var Ec=0;Ec<wr.length;Ec++){var Cc=wr[Ec];xr(Cc.toLowerCase(),"on"+(Cc[0].toUpperCase()+Cc.slice(1)))}xr(pr,"onAnimationEnd"),xr(hr,"onAnimationIteration"),xr(mr,"onAnimationStart"),xr("dblclick","onDoubleClick"),xr("focusin","onFocus"),xr("focusout","onBlur"),xr(vr,"onTransitionRun"),xr(gr,"onTransitionStart"),xr(yr,"onTransitionCancel"),xr(br,"onTransitionEnd"),tt("onMouseEnter",["mouseout","mouseover"]),tt("onMouseLeave",["mouseout","mouseover"]),tt("onPointerEnter",["pointerout","pointerover"]),tt("onPointerLeave",["pointerout","pointerover"]),et("onChange","change click focusin focusout input keydown keyup selectionchange".split(" ")),et("onSelect","focusout contextmenu dragend focusin keydown keyup mousedown mouseup selectionchange".split(" ")),et("onBeforeInput",["compositionend","keypress","textInput","paste"]),et("onCompositionEnd","compositionend focusout keydown keypress keyup mousedown".split(" ")),et("onCompositionStart","compositionstart focusout keydown keypress keyup mousedown".split(" ")),et("onCompositionUpdate","compositionupdate focusout keydown keypress keyup mousedown".split(" "));var Nc="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange resize seeked seeking stalled suspend timeupdate volumechange waiting".split(" "),_c=new Set("beforetoggle cancel close invalid load scroll scrollend toggle".split(" ").concat(Nc));function Pc(e,t){t=!!(4&t);for(var n=0;n<e.length;n++){var r=e[n],a=r.event;r=r.listeners;e:{var l=void 0;if(t)for(var o=r.length-1;0<=o;o--){var i=r[o],u=i.instance,s=i.currentTarget;if(i=i.listener,u!==l&&a.isPropagationStopped())break e;l=i,a.currentTarget=s;try{l(a)}catch(e){Mo(e)}a.currentTarget=null,l=u}else for(o=0;o<r.length;o++){if(u=(i=r[o]).instance,s=i.currentTarget,i=i.listener,u!==l&&a.isPropagationStopped())break e;l=i,a.currentTarget=s;try{l(a)}catch(e){Mo(e)}a.currentTarget=null,l=u}}}}function zc(e,t){var n=t[$e];void 0===n&&(n=t[$e]=new Set);var r=e+"__bubble";n.has(r)||(Lc(t,e,2,!1),n.add(r))}function jc(e,t,n){var r=0;t&&(r|=4),Lc(n,e,r,t)}var Tc="_reactListening"+Math.random().toString(36).slice(2);function Oc(e){if(!e[Tc]){e[Tc]=!0,Ze.forEach((function(t){"selectionchange"!==t&&(_c.has(t)||jc(t,!1,e),jc(t,!0,e))}));var t=9===e.nodeType?e:e.ownerDocument;null===t||t[Tc]||(t[Tc]=!0,jc("selectionchange",!1,t))}}function Lc(e,t,n,r){switch(Zf(t)){case 2:var a=Wf;break;case 8:a=Qf;break;default:a=Kf}n=a.bind(null,t,n,e),a=void 0,!Dt||"touchstart"!==t&&"touchmove"!==t&&"wheel"!==t||(a=!0),r?void 0!==a?e.addEventListener(t,n,{capture:!0,passive:a}):e.addEventListener(t,n,!0):void 0!==a?e.addEventListener(t,n,{passive:a}):e.addEventListener(t,n,!1)}function Mc(e,t,n,r,a){var l=r;if(!(1&t||2&t||null===r))e:for(;;){if(null===r)return;var o=r.tag;if(3===o||4===o){var i=r.stateNode.containerInfo;if(i===a||8===i.nodeType&&i.parentNode===a)break;if(4===o)for(o=r.return;null!==o;){var u=o.tag;if((3===u||4===u)&&((u=o.stateNode.containerInfo)===a||8===u.nodeType&&u.parentNode===a))return;o=o.return}for(;null!==i;){if(null===(o=Qe(i)))return;if(5===(u=o.tag)||6===u||26===u||27===u){r=l=o;continue e}i=i.parentNode}}r=r.return}Rt((function(){var r=l,a=Tt(n),o=[];e:{var i=kr.get(e);if(void 0!==i){var u=Zt,s=e;switch(e){case"keypress":if(0===Bt(n))break e;case"keydown":case"keyup":u=hn;break;case"focusin":s="focus",u=an;break;case"focusout":s="blur",u=an;break;case"beforeblur":case"afterblur":u=an;break;case"click":if(2===n.button)break e;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":u=nn;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":u=rn;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":u=vn;break;case pr:case hr:case mr:u=ln;break;case br:u=gn;break;case"scroll":case"scrollend":u=en;break;case"wheel":u=yn;break;case"copy":case"cut":case"paste":u=on;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":u=mn;break;case"toggle":case"beforetoggle":u=bn}var c=!!(4&t),f=!c&&("scroll"===e||"scrollend"===e),d=c?null!==i?i+"Capture":null:i;c=[];for(var p,h=r;null!==h;){var m=h;if(p=m.stateNode,5!==(m=m.tag)&&26!==m&&27!==m||null===p||null===d||null!=(m=Ft(h,d))&&c.push(Ac(h,m,p)),f)break;h=h.return}0<c.length&&(i=new u(i,s,null,n,a),o.push({event:i,listeners:c}))}}if(!(7&t)){if(u="mouseout"===e||"pointerout"===e,(!(i="mouseover"===e||"pointerover"===e)||n===jt||!(s=n.relatedTarget||n.fromElement)||!Qe(s)&&!s[Ue])&&(u||i)&&(i=a.window===a?a:(i=a.ownerDocument)?i.defaultView||i.parentWindow:window,u?(u=r,null!==(s=(s=n.relatedTarget||n.toElement)?Qe(s):null)&&(f=F(s),c=s.tag,s!==f||5!==c&&27!==c&&6!==c)&&(s=null)):(u=null,s=r),u!==s)){if(c=nn,m="onMouseLeave",d="onMouseEnter",h="mouse","pointerout"!==e&&"pointerover"!==e||(c=mn,m="onPointerLeave",d="onPointerEnter",h="pointer"),f=null==u?i:Ge(u),p=null==s?i:Ge(s),(i=new c(m,h+"leave",u,n,a)).target=f,i.relatedTarget=p,m=null,Qe(a)===r&&((c=new c(d,h+"enter",s,n,a)).target=p,c.relatedTarget=f,m=c),f=m,u&&s)e:{for(d=s,h=0,p=c=u;p;p=Fc(p))h++;for(p=0,m=d;m;m=Fc(m))p++;for(;0<h-p;)c=Fc(c),h--;for(;0<p-h;)d=Fc(d),p--;for(;h--;){if(c===d||null!==d&&c===d.alternate)break e;c=Fc(c),d=Fc(d)}c=null}else c=null;null!==u&&Dc(o,i,u,c,!1),null!==s&&null!==f&&Dc(o,f,s,c,!0)}if("select"===(u=(i=r?Ge(r):window).nodeName&&i.nodeName.toLowerCase())||"input"===u&&"file"===i.type)var v=Fn;else if(Tn(i))if(Dn)v=Qn;else{v=qn;var g=Bn}else!(u=i.nodeName)||"input"!==u.toLowerCase()||"checkbox"!==i.type&&"radio"!==i.type?r&&Nt(r.elementType)&&(v=Fn):v=Wn;switch(v&&(v=v(e,r))?On(o,v,n,a):(g&&g(e,i,r),"focusout"===e&&r&&"number"===i.type&&null!=r.memoizedProps.value&&yt(i,"number",i.value)),g=r?Ge(r):window,e){case"focusin":(Tn(g)||"true"===g.contentEditable)&&(rr=g,ar=r,lr=null);break;case"focusout":lr=ar=rr=null;break;case"mousedown":or=!0;break;case"contextmenu":case"mouseup":case"dragend":or=!1,ir(o,n,a);break;case"selectionchange":if(nr)break;case"keydown":case"keyup":ir(o,n,a)}var y;if(wn)e:{switch(e){case"compositionstart":var b="onCompositionStart";break e;case"compositionend":b="onCompositionEnd";break e;case"compositionupdate":b="onCompositionUpdate";break e}b=void 0}else zn?_n(e,n)&&(b="onCompositionEnd"):"keydown"===e&&229===n.keyCode&&(b="onCompositionStart");b&&(En&&"ko"!==n.locale&&(zn||"onCompositionStart"!==b?"onCompositionEnd"===b&&zn&&(y=Vt()):($t="value"in(Ut=a)?Ut.value:Ut.textContent,zn=!0)),0<(g=Rc(r,b)).length&&(b=new un(b,e,null,n,a),o.push({event:b,listeners:g}),(y||null!==(y=Pn(n)))&&(b.data=y))),(y=Sn?function(e,t){switch(e){case"compositionend":return Pn(t);case"keypress":return 32!==t.which?null:(Nn=!0,Cn);case"textInput":return(e=t.data)===Cn&&Nn?null:e;default:return null}}(e,n):function(e,t){if(zn)return"compositionend"===e||!wn&&_n(e,t)?(e=Vt(),Ht=$t=Ut=null,zn=!1,e):null;switch(e){case"paste":default:return null;case"keypress":if(!(t.ctrlKey||t.altKey||t.metaKey)||t.ctrlKey&&t.altKey){if(t.char&&1<t.char.length)return t.char;if(t.which)return String.fromCharCode(t.which)}return null;case"compositionend":return En&&"ko"!==t.locale?null:t.data}}(e,n))&&0<(b=Rc(r,"onBeforeInput")).length&&(g=new un("onBeforeInput","beforeinput",null,n,a),o.push({event:g,listeners:b}),g.data=y),function(e,t,n,r,a){if("submit"===t&&n&&n.stateNode===a){var l=xc((a[Ie]||null).action),o=r.submitter;o&&null!==(t=(t=o[Ie]||null)?xc(t.formAction):o.getAttribute("formAction"))&&(l=t,o=null);var i=new Zt("action","action",null,r,a);e.push({event:i,listeners:[{instance:null,listener:function(){if(r.defaultPrevented){if(0!==hc){var e=o?Sc(a,o):new FormData(a);co(n,{pending:!0,data:e,method:a.method,action:l},null,e)}}else"function"==typeof l&&(i.preventDefault(),e=o?Sc(a,o):new FormData(a),co(n,{pending:!0,data:e,method:a.method,action:l},l,e))},currentTarget:a}]})}}(o,e,r,n,a)}Pc(o,t)}))}function Ac(e,t,n){return{instance:e,listener:t,currentTarget:n}}function Rc(e,t){for(var n=t+"Capture",r=[];null!==e;){var a=e,l=a.stateNode;5!==(a=a.tag)&&26!==a&&27!==a||null===l||(null!=(a=Ft(e,n))&&r.unshift(Ac(e,a,l)),null!=(a=Ft(e,t))&&r.push(Ac(e,a,l))),e=e.return}return r}function Fc(e){if(null===e)return null;do{e=e.return}while(e&&5!==e.tag&&27!==e.tag);return e||null}function Dc(e,t,n,r,a){for(var l=t._reactName,o=[];null!==n&&n!==r;){var i=n,u=i.alternate,s=i.stateNode;if(i=i.tag,null!==u&&u===r)break;5!==i&&26!==i&&27!==i||null===s||(u=s,a?null!=(s=Ft(n,l))&&o.unshift(Ac(n,s,u)):a||null!=(s=Ft(n,l))&&o.push(Ac(n,s,u))),n=n.return}0!==o.length&&e.push({event:t,listeners:o})}var Ic=/\r\n?/g,Uc=/\u0000|\uFFFD/g;function $c(e){return("string"==typeof e?e:""+e).replace(Ic,"\n").replace(Uc,"")}function Hc(e,t){return t=$c(t),$c(e)===t}function Vc(){}function Bc(e,t,n,r,a,l){switch(n){case"children":"string"==typeof r?"body"===t||"textarea"===t&&""===r||xt(e,r):("number"==typeof r||"bigint"==typeof r)&&"body"!==t&&xt(e,""+r);break;case"className":it(e,"class",r);break;case"tabIndex":it(e,"tabindex",r);break;case"dir":case"role":case"viewBox":case"width":case"height":it(e,n,r);break;case"style":Ct(e,r,l);break;case"data":if("object"!==t){it(e,"data",r);break}case"src":case"href":if(""===r&&("a"!==t||"href"!==n)){e.removeAttribute(n);break}if(null==r||"function"==typeof r||"symbol"==typeof r||"boolean"==typeof r){e.removeAttribute(n);break}r=zt(""+r),e.setAttribute(n,r);break;case"action":case"formAction":if("function"==typeof r){e.setAttribute(n,"javascript:throw new Error('A React form was unexpectedly submitted. If you called form.submit() manually, consider using form.requestSubmit() instead. If you\\'re trying to use event.stopPropagation() in a submit event handler, consider also calling event.preventDefault().')");break}if("function"==typeof l&&("formAction"===n?("input"!==t&&Bc(e,t,"name",a.name,a,null),Bc(e,t,"formEncType",a.formEncType,a,null),Bc(e,t,"formMethod",a.formMethod,a,null),Bc(e,t,"formTarget",a.formTarget,a,null)):(Bc(e,t,"encType",a.encType,a,null),Bc(e,t,"method",a.method,a,null),Bc(e,t,"target",a.target,a,null))),null==r||"symbol"==typeof r||"boolean"==typeof r){e.removeAttribute(n);break}r=zt(""+r),e.setAttribute(n,r);break;case"onClick":null!=r&&(e.onclick=Vc);break;case"onScroll":null!=r&&zc("scroll",e);break;case"onScrollEnd":null!=r&&zc("scrollend",e);break;case"dangerouslySetInnerHTML":if(null!=r){if("object"!=typeof r||!("__html"in r))throw Error(o(61));if(null!=(n=r.__html)){if(null!=a.children)throw Error(o(60));e.innerHTML=n}}break;case"multiple":e.multiple=r&&"function"!=typeof r&&"symbol"!=typeof r;break;case"muted":e.muted=r&&"function"!=typeof r&&"symbol"!=typeof r;break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"defaultValue":case"defaultChecked":case"innerHTML":case"ref":case"autoFocus":break;case"xlinkHref":if(null==r||"function"==typeof r||"boolean"==typeof r||"symbol"==typeof r){e.removeAttribute("xlink:href");break}n=zt(""+r),e.setAttributeNS("http://www.w3.org/1999/xlink","xlink:href",n);break;case"contentEditable":case"spellCheck":case"draggable":case"value":case"autoReverse":case"externalResourcesRequired":case"focusable":case"preserveAlpha":null!=r&&"function"!=typeof r&&"symbol"!=typeof r?e.setAttribute(n,""+r):e.removeAttribute(n);break;case"inert":case"allowFullScreen":case"async":case"autoPlay":case"controls":case"default":case"defer":case"disabled":case"disablePictureInPicture":case"disableRemotePlayback":case"formNoValidate":case"hidden":case"loop":case"noModule":case"noValidate":case"open":case"playsInline":case"readOnly":case"required":case"reversed":case"scoped":case"seamless":case"itemScope":r&&"function"!=typeof r&&"symbol"!=typeof r?e.setAttribute(n,""):e.removeAttribute(n);break;case"capture":case"download":!0===r?e.setAttribute(n,""):!1!==r&&null!=r&&"function"!=typeof r&&"symbol"!=typeof r?e.setAttribute(n,r):e.removeAttribute(n);break;case"cols":case"rows":case"size":case"span":null!=r&&"function"!=typeof r&&"symbol"!=typeof r&&!isNaN(r)&&1<=r?e.setAttribute(n,r):e.removeAttribute(n);break;case"rowSpan":case"start":null==r||"function"==typeof r||"symbol"==typeof r||isNaN(r)?e.removeAttribute(n):e.setAttribute(n,r);break;case"popover":zc("beforetoggle",e),zc("toggle",e),ot(e,"popover",r);break;case"xlinkActuate":ut(e,"http://www.w3.org/1999/xlink","xlink:actuate",r);break;case"xlinkArcrole":ut(e,"http://www.w3.org/1999/xlink","xlink:arcrole",r);break;case"xlinkRole":ut(e,"http://www.w3.org/1999/xlink","xlink:role",r);break;case"xlinkShow":ut(e,"http://www.w3.org/1999/xlink","xlink:show",r);break;case"xlinkTitle":ut(e,"http://www.w3.org/1999/xlink","xlink:title",r);break;case"xlinkType":ut(e,"http://www.w3.org/1999/xlink","xlink:type",r);break;case"xmlBase":ut(e,"http://www.w3.org/XML/1998/namespace","xml:base",r);break;case"xmlLang":ut(e,"http://www.w3.org/XML/1998/namespace","xml:lang",r);break;case"xmlSpace":ut(e,"http://www.w3.org/XML/1998/namespace","xml:space",r);break;case"is":ot(e,"is",r);break;case"innerText":case"textContent":break;default:(!(2<n.length)||"o"!==n[0]&&"O"!==n[0]||"n"!==n[1]&&"N"!==n[1])&&ot(e,n=_t.get(n)||n,r)}}function qc(e,t,n,r,a,l){switch(n){case"style":Ct(e,r,l);break;case"dangerouslySetInnerHTML":if(null!=r){if("object"!=typeof r||!("__html"in r))throw Error(o(61));if(null!=(n=r.__html)){if(null!=a.children)throw Error(o(60));e.innerHTML=n}}break;case"children":"string"==typeof r?xt(e,r):("number"==typeof r||"bigint"==typeof r)&&xt(e,""+r);break;case"onScroll":null!=r&&zc("scroll",e);break;case"onScrollEnd":null!=r&&zc("scrollend",e);break;case"onClick":null!=r&&(e.onclick=Vc);break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"innerHTML":case"ref":case"innerText":case"textContent":break;default:Je.hasOwnProperty(n)||("o"!==n[0]||"n"!==n[1]||(a=n.endsWith("Capture"),t=n.slice(2,a?n.length-7:void 0),"function"==typeof(l=null!=(l=e[Ie]||null)?l[n]:null)&&e.removeEventListener(t,l,a),"function"!=typeof r)?n in e?e[n]=r:!0===r?e.setAttribute(n,""):ot(e,n,r):("function"!=typeof l&&null!==l&&(n in e?e[n]=null:e.hasAttribute(n)&&e.removeAttribute(n)),e.addEventListener(t,r,a)))}}function Wc(e,t,n){switch(t){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"img":zc("error",e),zc("load",e);var r,a=!1,l=!1;for(r in n)if(n.hasOwnProperty(r)){var i=n[r];if(null!=i)switch(r){case"src":a=!0;break;case"srcSet":l=!0;break;case"children":case"dangerouslySetInnerHTML":throw Error(o(137,t));default:Bc(e,t,r,i,n,null)}}return l&&Bc(e,t,"srcSet",n.srcSet,n,null),void(a&&Bc(e,t,"src",n.src,n,null));case"input":zc("invalid",e);var u=r=i=l=null,s=null,c=null;for(a in n)if(n.hasOwnProperty(a)){var f=n[a];if(null!=f)switch(a){case"name":l=f;break;case"type":i=f;break;case"checked":s=f;break;case"defaultChecked":c=f;break;case"value":r=f;break;case"defaultValue":u=f;break;case"children":case"dangerouslySetInnerHTML":if(null!=f)throw Error(o(137,t));break;default:Bc(e,t,a,f,n,null)}}return gt(e,r,u,s,c,i,l,!1),void ft(e);case"select":for(l in zc("invalid",e),a=i=r=null,n)if(n.hasOwnProperty(l)&&null!=(u=n[l]))switch(l){case"value":r=u;break;case"defaultValue":i=u;break;case"multiple":a=u;default:Bc(e,t,l,u,n,null)}return t=r,n=i,e.multiple=!!a,void(null!=t?bt(e,!!a,t,!1):null!=n&&bt(e,!!a,n,!0));case"textarea":for(i in zc("invalid",e),r=l=a=null,n)if(n.hasOwnProperty(i)&&null!=(u=n[i]))switch(i){case"value":a=u;break;case"defaultValue":l=u;break;case"children":r=u;break;case"dangerouslySetInnerHTML":if(null!=u)throw Error(o(91));break;default:Bc(e,t,i,u,n,null)}return wt(e,a,l,r),void ft(e);case"option":for(s in n)n.hasOwnProperty(s)&&null!=(a=n[s])&&("selected"===s?e.selected=a&&"function"!=typeof a&&"symbol"!=typeof a:Bc(e,t,s,a,n,null));return;case"dialog":zc("cancel",e),zc("close",e);break;case"iframe":case"object":zc("load",e);break;case"video":case"audio":for(a=0;a<Nc.length;a++)zc(Nc[a],e);break;case"image":zc("error",e),zc("load",e);break;case"details":zc("toggle",e);break;case"embed":case"source":case"link":zc("error",e),zc("load",e);case"area":case"base":case"br":case"col":case"hr":case"keygen":case"meta":case"param":case"track":case"wbr":case"menuitem":for(c in n)if(n.hasOwnProperty(c)&&null!=(a=n[c]))switch(c){case"children":case"dangerouslySetInnerHTML":throw Error(o(137,t));default:Bc(e,t,c,a,n,null)}return;default:if(Nt(t)){for(f in n)n.hasOwnProperty(f)&&void 0!==(a=n[f])&&qc(e,t,f,a,n,void 0);return}}for(u in n)n.hasOwnProperty(u)&&null!=(a=n[u])&&Bc(e,t,u,a,n,null)}var Qc=null,Kc=null;function Gc(e){return 9===e.nodeType?e:e.ownerDocument}function Yc(e){switch(e){case"http://www.w3.org/2000/svg":return 1;case"http://www.w3.org/1998/Math/MathML":return 2;default:return 0}}function Xc(e,t){if(0===e)switch(t){case"svg":return 1;case"math":return 2;default:return 0}return 1===e&&"foreignObject"===t?0:e}function Zc(e,t){return"textarea"===e||"noscript"===e||"string"==typeof t.children||"number"==typeof t.children||"bigint"==typeof t.children||"object"==typeof t.dangerouslySetInnerHTML&&null!==t.dangerouslySetInnerHTML&&null!=t.dangerouslySetInnerHTML.__html}var Jc=null,ef="function"==typeof setTimeout?setTimeout:void 0,tf="function"==typeof clearTimeout?clearTimeout:void 0,nf="function"==typeof Promise?Promise:void 0,rf="function"==typeof queueMicrotask?queueMicrotask:void 0!==nf?function(e){return nf.resolve(null).then(e).catch(af)}:ef;function af(e){setTimeout((function(){throw e}))}function lf(e,t){var n=t,r=0;do{var a=n.nextSibling;if(e.removeChild(n),a&&8===a.nodeType)if("/$"===(n=a.data)){if(0===r)return e.removeChild(a),void vd(t);r--}else"$"!==n&&"$?"!==n&&"$!"!==n||r++;n=a}while(n);vd(t)}function of(e){var t=e.firstChild;for(t&&10===t.nodeType&&(t=t.nextSibling);t;){var n=t;switch(t=t.nextSibling,n.nodeName){case"HTML":case"HEAD":case"BODY":of(n),We(n);continue;case"SCRIPT":case"STYLE":continue;case"LINK":if("stylesheet"===n.rel.toLowerCase())continue}e.removeChild(n)}}function uf(e){for(;null!=e;e=e.nextSibling){var t=e.nodeType;if(1===t||3===t)break;if(8===t){if("$"===(t=e.data)||"$!"===t||"$?"===t||"F!"===t||"F"===t)break;if("/$"===t)return null}}return e}function sf(e){e=e.previousSibling;for(var t=0;e;){if(8===e.nodeType){var n=e.data;if("$"===n||"$!"===n||"$?"===n){if(0===t)return e;t--}else"/$"===n&&t++}e=e.previousSibling}return null}function cf(e,t,n){switch(t=Gc(n),e){case"html":if(!(e=t.documentElement))throw Error(o(452));return e;case"head":if(!(e=t.head))throw Error(o(453));return e;case"body":if(!(e=t.body))throw Error(o(454));return e;default:throw Error(o(451))}}var ff=new Map,df=new Set;function pf(e){return"function"==typeof e.getRootNode?e.getRootNode():e.ownerDocument}var hf=H.d;H.d={f:function(){var e=hf.f(),t=Ds();return e||t},r:function(e){var t=Ke(e);null!==t&&5===t.tag&&"form"===t.type?po(t):hf.r(e)},D:function(e){hf.D(e),vf("dns-prefetch",e,null)},C:function(e,t){hf.C(e,t),vf("preconnect",e,t)},L:function(e,t,n){hf.L(e,t,n);var r=mf;if(r&&e&&t){var a='link[rel="preload"][as="'+mt(t)+'"]';"image"===t&&n&&n.imageSrcSet?(a+='[imagesrcset="'+mt(n.imageSrcSet)+'"]',"string"==typeof n.imageSizes&&(a+='[imagesizes="'+mt(n.imageSizes)+'"]')):a+='[href="'+mt(e)+'"]';var l=a;switch(t){case"style":l=yf(e);break;case"script":l=wf(e)}ff.has(l)||(e=T({rel:"preload",href:"image"===t&&n&&n.imageSrcSet?void 0:e,as:t},n),ff.set(l,e),null!==r.querySelector(a)||"style"===t&&r.querySelector(bf(l))||"script"===t&&r.querySelector(xf(l))||(Wc(t=r.createElement("link"),"link",e),Xe(t),r.head.appendChild(t)))}},m:function(e,t){hf.m(e,t);var n=mf;if(n&&e){var r=t&&"string"==typeof t.as?t.as:"script",a='link[rel="modulepreload"][as="'+mt(r)+'"][href="'+mt(e)+'"]',l=a;switch(r){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":l=wf(e)}if(!ff.has(l)&&(e=T({rel:"modulepreload",href:e},t),ff.set(l,e),null===n.querySelector(a))){switch(r){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":if(n.querySelector(xf(l)))return}Wc(r=n.createElement("link"),"link",e),Xe(r),n.head.appendChild(r)}}},X:function(e,t){hf.X(e,t);var n=mf;if(n&&e){var r=Ye(n).hoistableScripts,a=wf(e),l=r.get(a);l||((l=n.querySelector(xf(a)))||(e=T({src:e,async:!0},t),(t=ff.get(a))&&Nf(e,t),Xe(l=n.createElement("script")),Wc(l,"link",e),n.head.appendChild(l)),l={type:"script",instance:l,count:1,state:null},r.set(a,l))}},S:function(e,t,n){hf.S(e,t,n);var r=mf;if(r&&e){var a=Ye(r).hoistableStyles,l=yf(e);t=t||"default";var o=a.get(l);if(!o){var i={loading:0,preload:null};if(o=r.querySelector(bf(l)))i.loading=5;else{e=T({rel:"stylesheet",href:e,"data-precedence":t},n),(n=ff.get(l))&&Cf(e,n);var u=o=r.createElement("link");Xe(u),Wc(u,"link",e),u._p=new Promise((function(e,t){u.onload=e,u.onerror=t})),u.addEventListener("load",(function(){i.loading|=1})),u.addEventListener("error",(function(){i.loading|=2})),i.loading|=4,Ef(o,t,r)}o={type:"stylesheet",instance:o,count:1,state:i},a.set(l,o)}}},M:function(e,t){hf.M(e,t);var n=mf;if(n&&e){var r=Ye(n).hoistableScripts,a=wf(e),l=r.get(a);l||((l=n.querySelector(xf(a)))||(e=T({src:e,async:!0,type:"module"},t),(t=ff.get(a))&&Nf(e,t),Xe(l=n.createElement("script")),Wc(l,"link",e),n.head.appendChild(l)),l={type:"script",instance:l,count:1,state:null},r.set(a,l))}}};var mf="undefined"==typeof document?null:document;function vf(e,t,n){var r=mf;if(r&&"string"==typeof t&&t){var a=mt(t);a='link[rel="'+e+'"][href="'+a+'"]',"string"==typeof n&&(a+='[crossorigin="'+n+'"]'),df.has(a)||(df.add(a),e={rel:e,crossOrigin:n,href:t},null===r.querySelector(a)&&(Wc(t=r.createElement("link"),"link",e),Xe(t),r.head.appendChild(t)))}}function gf(e,t,n,r){var a,l,i,u,s=(s=X.current)?pf(s):null;if(!s)throw Error(o(446));switch(e){case"meta":case"title":return null;case"style":return"string"==typeof n.precedence&&"string"==typeof n.href?(t=yf(n.href),(r=(n=Ye(s).hoistableStyles).get(t))||(r={type:"style",instance:null,count:0,state:null},n.set(t,r)),r):{type:"void",instance:null,count:0,state:null};case"link":if("stylesheet"===n.rel&&"string"==typeof n.href&&"string"==typeof n.precedence){e=yf(n.href);var c=Ye(s).hoistableStyles,f=c.get(e);if(f||(s=s.ownerDocument||s,f={type:"stylesheet",instance:null,count:0,state:{loading:0,preload:null}},c.set(e,f),(c=s.querySelector(bf(e)))&&!c._p&&(f.instance=c,f.state.loading=5),ff.has(e)||(n={rel:"preload",as:"style",href:n.href,crossOrigin:n.crossOrigin,integrity:n.integrity,media:n.media,hrefLang:n.hrefLang,referrerPolicy:n.referrerPolicy},ff.set(e,n),c||(a=s,l=e,i=n,u=f.state,a.querySelector('link[rel="preload"][as="style"]['+l+"]")?u.loading=1:(l=a.createElement("link"),u.preload=l,l.addEventListener("load",(function(){return u.loading|=1})),l.addEventListener("error",(function(){return u.loading|=2})),Wc(l,"link",i),Xe(l),a.head.appendChild(l))))),t&&null===r)throw Error(o(528,""));return f}if(t&&null!==r)throw Error(o(529,""));return null;case"script":return t=n.async,"string"==typeof(n=n.src)&&t&&"function"!=typeof t&&"symbol"!=typeof t?(t=wf(n),(r=(n=Ye(s).hoistableScripts).get(t))||(r={type:"script",instance:null,count:0,state:null},n.set(t,r)),r):{type:"void",instance:null,count:0,state:null};default:throw Error(o(444,e))}}function yf(e){return'href="'+mt(e)+'"'}function bf(e){return'link[rel="stylesheet"]['+e+"]"}function kf(e){return T({},e,{"data-precedence":e.precedence,precedence:null})}function wf(e){return'[src="'+mt(e)+'"]'}function xf(e){return"script[async]"+e}function Sf(e,t,n){if(t.count++,null===t.instance)switch(t.type){case"style":var r=e.querySelector('style[data-href~="'+mt(n.href)+'"]');if(r)return t.instance=r,Xe(r),r;var a=T({},n,{"data-href":n.href,"data-precedence":n.precedence,href:null,precedence:null});return Xe(r=(e.ownerDocument||e).createElement("style")),Wc(r,"style",a),Ef(r,n.precedence,e),t.instance=r;case"stylesheet":a=yf(n.href);var l=e.querySelector(bf(a));if(l)return t.state.loading|=4,t.instance=l,Xe(l),l;r=kf(n),(a=ff.get(a))&&Cf(r,a),Xe(l=(e.ownerDocument||e).createElement("link"));var i=l;return i._p=new Promise((function(e,t){i.onload=e,i.onerror=t})),Wc(l,"link",r),t.state.loading|=4,Ef(l,n.precedence,e),t.instance=l;case"script":return l=wf(n.src),(a=e.querySelector(xf(l)))?(t.instance=a,Xe(a),a):(r=n,(a=ff.get(l))&&Nf(r=T({},n),a),Xe(a=(e=e.ownerDocument||e).createElement("script")),Wc(a,"link",r),e.head.appendChild(a),t.instance=a);case"void":return null;default:throw Error(o(443,t.type))}else"stylesheet"===t.type&&!(4&t.state.loading)&&(r=t.instance,t.state.loading|=4,Ef(r,n.precedence,e));return t.instance}function Ef(e,t,n){for(var r=n.querySelectorAll('link[rel="stylesheet"][data-precedence],style[data-precedence]'),a=r.length?r[r.length-1]:null,l=a,o=0;o<r.length;o++){var i=r[o];if(i.dataset.precedence===t)l=i;else if(l!==a)break}l?l.parentNode.insertBefore(e,l.nextSibling):(t=9===n.nodeType?n.head:n).insertBefore(e,t.firstChild)}function Cf(e,t){null==e.crossOrigin&&(e.crossOrigin=t.crossOrigin),null==e.referrerPolicy&&(e.referrerPolicy=t.referrerPolicy),null==e.title&&(e.title=t.title)}function Nf(e,t){null==e.crossOrigin&&(e.crossOrigin=t.crossOrigin),null==e.referrerPolicy&&(e.referrerPolicy=t.referrerPolicy),null==e.integrity&&(e.integrity=t.integrity)}var _f=null;function Pf(e,t,n){if(null===_f){var r=new Map,a=_f=new Map;a.set(n,r)}else(r=(a=_f).get(n))||(r=new Map,a.set(n,r));if(r.has(e))return r;for(r.set(e,null),n=n.getElementsByTagName(e),a=0;a<n.length;a++){var l=n[a];if(!(l[qe]||l[De]||"link"===e&&"stylesheet"===l.getAttribute("rel"))&&"http://www.w3.org/2000/svg"!==l.namespaceURI){var o=l.getAttribute(t)||"";o=e+o;var i=r.get(o);i?i.push(l):r.set(o,[l])}}return r}function zf(e,t,n){(e=e.ownerDocument||e).head.insertBefore(n,"title"===t?e.querySelector("head > title"):null)}function jf(e){return!!("stylesheet"!==e.type||3&e.state.loading)}var Tf=null;function Of(){}function Lf(){if(this.count--,0===this.count)if(this.stylesheets)Af(this,this.stylesheets);else if(this.unsuspend){var e=this.unsuspend;this.unsuspend=null,e()}}var Mf=null;function Af(e,t){e.stylesheets=null,null!==e.unsuspend&&(e.count++,Mf=new Map,t.forEach(Rf,e),Mf=null,Lf.call(e))}function Rf(e,t){if(!(4&t.state.loading)){var n=Mf.get(e);if(n)var r=n.get(null);else{n=new Map,Mf.set(e,n);for(var a=e.querySelectorAll("link[data-precedence],style[data-precedence]"),l=0;l<a.length;l++){var o=a[l];"LINK"!==o.nodeName&&"not all"===o.getAttribute("media")||(n.set(o.dataset.precedence,o),r=o)}r&&n.set(null,r)}o=(a=t.instance).getAttribute("data-precedence"),(l=n.get(o)||r)===r&&n.set(null,a),n.set(o,a),this.count++,r=Lf.bind(this),a.addEventListener("load",r),a.addEventListener("error",r),l?l.parentNode.insertBefore(a,l.nextSibling):(e=9===e.nodeType?e.head:e).insertBefore(a,e.firstChild),t.state.loading|=4}}var Ff={$$typeof:v,Provider:null,Consumer:null,_currentValue:V,_currentValue2:V,_threadCount:0};function Df(e,t,n,r,a,l,o,i){this.tag=1,this.containerInfo=e,this.finishedWork=this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=-1,this.callbackNode=this.next=this.pendingContext=this.context=this.cancelPendingCommit=null,this.callbackPriority=0,this.expirationTimes=Te(-1),this.entangledLanes=this.shellSuspendCounter=this.errorRecoveryDisabledLanes=this.finishedLanes=this.expiredLanes=this.warmLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=Te(0),this.hiddenUpdates=Te(null),this.identifierPrefix=r,this.onUncaughtError=a,this.onCaughtError=l,this.onRecoverableError=o,this.pooledCache=null,this.pooledCacheLanes=0,this.formState=i,this.incompleteTransitions=new Map}function If(e,t,n,r,a,l,o,i,u,s,c,f){return e=new Df(e,t,n,o,i,u,s,f),t=1,!0===l&&(t|=24),l=Au(3,null,null,t),e.current=l,l.stateNode=e,(t=Ua()).refCount++,e.pooledCache=t,t.refCount++,l.memoizedState={element:r,isDehydrated:n,cache:t},zi(l),e}function Uf(e){return e?e=Or:Or}function $f(e,t,n,r,a,l){a=Uf(a),null===r.context?r.context=a:r.pendingContext=a,(r=Ti(t)).payload={element:n},null!==(l=void 0===l?null:l)&&(r.callback=l),null!==(n=Oi(e,r,t))&&(Os(n,0,t),Li(n,e,t))}function Hf(e,t){if(null!==(e=e.memoizedState)&&null!==e.dehydrated){var n=e.retryLane;e.retryLane=0!==n&&n<t?n:t}}function Vf(e,t){Hf(e,t),(e=e.alternate)&&Hf(e,t)}function Bf(e){if(13===e.tag){var t=zr(e,67108864);null!==t&&Os(t,0,67108864),Vf(e,67108864)}}var qf=!0;function Wf(e,t,n,r){var a=j.T;j.T=null;var l=H.p;try{H.p=2,Kf(e,t,n,r)}finally{H.p=l,j.T=a}}function Qf(e,t,n,r){var a=j.T;j.T=null;var l=H.p;try{H.p=8,Kf(e,t,n,r)}finally{H.p=l,j.T=a}}function Kf(e,t,n,r){if(qf){var a=Gf(r);if(null===a)Mc(e,t,r,Yf,n),id(e,r);else if(function(e,t,n,r,a){switch(t){case"focusin":return ed=ud(ed,e,t,n,r,a),!0;case"dragenter":return td=ud(td,e,t,n,r,a),!0;case"mouseover":return nd=ud(nd,e,t,n,r,a),!0;case"pointerover":var l=a.pointerId;return rd.set(l,ud(rd.get(l)||null,e,t,n,r,a)),!0;case"gotpointercapture":return l=a.pointerId,ad.set(l,ud(ad.get(l)||null,e,t,n,r,a)),!0}return!1}(a,e,t,n,r))r.stopPropagation();else if(id(e,r),4&t&&-1<od.indexOf(e)){for(;null!==a;){var l=Ke(a);if(null!==l)switch(l.tag){case 3:if((l=l.stateNode).current.memoizedState.isDehydrated){var o=Ce(l.pendingLanes);if(0!==o){var i=l;for(i.pendingLanes|=2,i.entangledLanes|=2;o;){var u=1<<31-ke(o);i.entanglements[1]|=u,o&=~u}mc(l),!(6&es)&&(ks=ue()+500,vc(0,!1))}}break;case 13:null!==(i=zr(l,2))&&Os(i,0,2),Ds(),Vf(l,2)}if(null===(l=Gf(r))&&Mc(e,t,r,Yf,n),l===a)break;a=l}null!==a&&r.stopPropagation()}else Mc(e,t,r,null,n)}}function Gf(e){return Xf(e=Tt(e))}var Yf=null;function Xf(e){if(Yf=null,null!==(e=Qe(e))){var t=F(e);if(null===t)e=null;else{var n=t.tag;if(13===n){if(null!==(e=D(t)))return e;e=null}else if(3===n){if(t.stateNode.current.memoizedState.isDehydrated)return 3===t.tag?t.stateNode.containerInfo:null;e=null}else t!==e&&(e=null)}}return Yf=e,null}function Zf(e){switch(e){case"beforetoggle":case"cancel":case"click":case"close":case"contextmenu":case"copy":case"cut":case"auxclick":case"dblclick":case"dragend":case"dragstart":case"drop":case"focusin":case"focusout":case"input":case"invalid":case"keydown":case"keypress":case"keyup":case"mousedown":case"mouseup":case"paste":case"pause":case"play":case"pointercancel":case"pointerdown":case"pointerup":case"ratechange":case"reset":case"resize":case"seeked":case"submit":case"toggle":case"touchcancel":case"touchend":case"touchstart":case"volumechange":case"change":case"selectionchange":case"textInput":case"compositionstart":case"compositionend":case"compositionupdate":case"beforeblur":case"afterblur":case"beforeinput":case"blur":case"fullscreenchange":case"focus":case"hashchange":case"popstate":case"select":case"selectstart":return 2;case"drag":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"mousemove":case"mouseout":case"mouseover":case"pointermove":case"pointerout":case"pointerover":case"scroll":case"touchmove":case"wheel":case"mouseenter":case"mouseleave":case"pointerenter":case"pointerleave":return 8;case"message":switch(se()){case ce:return 2;case fe:return 8;case de:case pe:return 32;case he:return 268435456;default:return 32}default:return 32}}var Jf=!1,ed=null,td=null,nd=null,rd=new Map,ad=new Map,ld=[],od="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput copy cut paste click change contextmenu reset".split(" ");function id(e,t){switch(e){case"focusin":case"focusout":ed=null;break;case"dragenter":case"dragleave":td=null;break;case"mouseover":case"mouseout":nd=null;break;case"pointerover":case"pointerout":rd.delete(t.pointerId);break;case"gotpointercapture":case"lostpointercapture":ad.delete(t.pointerId)}}function ud(e,t,n,r,a,l){return null===e||e.nativeEvent!==l?(e={blockedOn:t,domEventName:n,eventSystemFlags:r,nativeEvent:l,targetContainers:[a]},null!==t&&null!==(t=Ke(t))&&Bf(t),e):(e.eventSystemFlags|=r,t=e.targetContainers,null!==a&&-1===t.indexOf(a)&&t.push(a),e)}function sd(e){var t=Qe(e.target);if(null!==t){var n=F(t);if(null!==n)if(13===(t=n.tag)){if(null!==(t=D(n)))return e.blockedOn=t,void function(e){var t=H.p;try{return H.p=e,function(){if(13===n.tag){var e=js(),t=zr(n,e);null!==t&&Os(t,0,e),Vf(n,e)}}()}finally{H.p=t}}(e.priority)}else if(3===t&&n.stateNode.current.memoizedState.isDehydrated)return void(e.blockedOn=3===n.tag?n.stateNode.containerInfo:null)}e.blockedOn=null}function cd(e){if(null!==e.blockedOn)return!1;for(var t=e.targetContainers;0<t.length;){var n=Gf(e.nativeEvent);if(null!==n)return null!==(t=Ke(n))&&Bf(t),e.blockedOn=n,!1;var r=new(n=e.nativeEvent).constructor(n.type,n);jt=r,n.target.dispatchEvent(r),jt=null,t.shift()}return!0}function fd(e,t,n){cd(e)&&n.delete(t)}function dd(){Jf=!1,null!==ed&&cd(ed)&&(ed=null),null!==td&&cd(td)&&(td=null),null!==nd&&cd(nd)&&(nd=null),rd.forEach(fd),ad.forEach(fd)}function pd(e,t){e.blockedOn===t&&(e.blockedOn=null,Jf||(Jf=!0,r.unstable_scheduleCallback(r.unstable_NormalPriority,dd)))}var hd=null;function md(e){hd!==e&&(hd=e,r.unstable_scheduleCallback(r.unstable_NormalPriority,(function(){hd===e&&(hd=null);for(var t=0;t<e.length;t+=3){var n=e[t],r=e[t+1],a=e[t+2];if("function"!=typeof r){if(null===Xf(r||n))continue;break}var l=Ke(n);null!==l&&(e.splice(t,3),t-=3,co(l,{pending:!0,data:a,method:n.method,action:r},r,a))}})))}function vd(e){function t(t){return pd(t,e)}null!==ed&&pd(ed,e),null!==td&&pd(td,e),null!==nd&&pd(nd,e),rd.forEach(t),ad.forEach(t);for(var n=0;n<ld.length;n++){var r=ld[n];r.blockedOn===e&&(r.blockedOn=null)}for(;0<ld.length&&null===(n=ld[0]).blockedOn;)sd(n),null===n.blockedOn&&ld.shift();if(null!=(n=(e.ownerDocument||e).$$reactFormReplay))for(r=0;r<n.length;r+=3){var a=n[r],l=n[r+1],o=a[Ie]||null;if("function"==typeof l)o||md(n);else if(o){var i=null;if(l&&l.hasAttribute("formAction")){if(a=l,o=l[Ie]||null)i=o.formAction;else if(null!==Xf(a))continue}else i=o.action;"function"==typeof i?n[r+1]=i:(n.splice(r,3),r-=3),md(n)}}}function gd(e){this._internalRoot=e}function yd(e){this._internalRoot=e}yd.prototype.render=gd.prototype.render=function(e){var t=this._internalRoot;if(null===t)throw Error(o(409));$f(t.current,js(),e,t,null,null)},yd.prototype.unmount=gd.prototype.unmount=function(){var e=this._internalRoot;if(null!==e){this._internalRoot=null;var t=e.containerInfo;0===e.tag&&tc(),$f(e.current,2,null,e,null,null),Ds(),t[Ue]=null}},yd.prototype.unstable_scheduleHydration=function(e){if(e){var t=Re();e={blockedOn:null,target:e,priority:t};for(var n=0;n<ld.length&&0!==t&&t<ld[n].priority;n++);ld.splice(n,0,e),0===n&&sd(e)}};var bd=a.version;if("19.0.0"!==bd)throw Error(o(527,bd,"19.0.0"));H.findDOMNode=function(e){var t=e._reactInternals;if(void 0===t){if("function"==typeof e.render)throw Error(o(188));throw e=Object.keys(e).join(","),Error(o(268,e))}return e=function(e){var t=e.alternate;if(!t){if(null===(t=F(e)))throw Error(o(188));return t!==e?null:e}for(var n=e,r=t;;){var a=n.return;if(null===a)break;var l=a.alternate;if(null===l){if(null!==(r=a.return)){n=r;continue}break}if(a.child===l.child){for(l=a.child;l;){if(l===n)return I(a),e;if(l===r)return I(a),t;l=l.sibling}throw Error(o(188))}if(n.return!==r.return)n=a,r=l;else{for(var i=!1,u=a.child;u;){if(u===n){i=!0,n=a,r=l;break}if(u===r){i=!0,r=a,n=l;break}u=u.sibling}if(!i){for(u=l.child;u;){if(u===n){i=!0,n=l,r=a;break}if(u===r){i=!0,r=l,n=a;break}u=u.sibling}if(!i)throw Error(o(189))}}if(n.alternate!==r)throw Error(o(190))}if(3!==n.tag)throw Error(o(188));return n.stateNode.current===n?e:t}(t),null===(e=null!==e?U(e):null)?null:e.stateNode};var kd={bundleType:0,version:"19.0.0",rendererPackageName:"react-dom",currentDispatcherRef:j,findFiberByHostInstance:Qe,reconcilerVersion:"19.0.0"};if("undefined"!=typeof __REACT_DEVTOOLS_GLOBAL_HOOK__){var wd=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(!wd.isDisabled&&wd.supportsFiber)try{ge=wd.inject(kd),ye=wd}catch(e){}}t.createRoot=function(e,t){if(!i(e))throw Error(o(299));var n=!1,r="",a=Ao,l=Ro,u=Fo;return null!=t&&(!0===t.unstable_strictMode&&(n=!0),void 0!==t.identifierPrefix&&(r=t.identifierPrefix),void 0!==t.onUncaughtError&&(a=t.onUncaughtError),void 0!==t.onCaughtError&&(l=t.onCaughtError),void 0!==t.onRecoverableError&&(u=t.onRecoverableError),void 0!==t.unstable_transitionCallbacks&&t.unstable_transitionCallbacks),t=If(e,1,!1,null,0,n,r,a,l,u,0,null),e[Ue]=t.current,Oc(8===e.nodeType?e.parentNode:e),new gd(t)},t.hydrateRoot=function(e,t,n){if(!i(e))throw Error(o(299));var r=!1,a="",l=Ao,u=Ro,s=Fo,c=null;return null!=n&&(!0===n.unstable_strictMode&&(r=!0),void 0!==n.identifierPrefix&&(a=n.identifierPrefix),void 0!==n.onUncaughtError&&(l=n.onUncaughtError),void 0!==n.onCaughtError&&(u=n.onCaughtError),void 0!==n.onRecoverableError&&(s=n.onRecoverableError),void 0!==n.unstable_transitionCallbacks&&n.unstable_transitionCallbacks,void 0!==n.formState&&(c=n.formState)),(t=If(e,1,!0,t,0,r,a,l,u,s,0,c)).context=Uf(null),n=t.current,(a=Ti(r=js())).callback=null,Oi(n,a,r),t.current.lanes=r,Oe(t,r),mc(t),e[Ue]=t.current,Oc(e),new yd(t)},t.version="19.0.0"},1291:(e,t,n)=>{var r=n(741);e.exports=function(e){var t=+e;return t!=t||0===t?0:r(t)}},1517:(e,t,n)=>{var r=n(6518),a=n(6194),l=n(2248).remove;r({target:"Map",proto:!0,real:!0,forced:!0},{deleteAll:function(){for(var e,t=a(this),n=!0,r=0,o=arguments.length;r<o;r++)e=l(t,arguments[r]),n=n&&e;return!!n}})},1625:(e,t,n)=>{var r=n(9504);e.exports=r({}.isPrototypeOf)},1828:(e,t,n)=>{var r=n(9504),a=n(9297),l=n(5397),o=n(9617).indexOf,i=n(421),u=r([].push);e.exports=function(e,t){var n,r=l(e),s=0,c=[];for(n in r)!a(i,n)&&a(r,n)&&u(c,n);for(;t.length>s;)a(r,n=t[s++])&&(~o(c,n)||u(c,n));return c}},2140:(e,t,n)=>{var r={};r[n(8227)("toStringTag")]="z",e.exports="[object z]"===String(r)},2195:(e,t,n)=>{var r=n(9504),a=r({}.toString),l=r("".slice);e.exports=function(e){return l(a(e),8,-1)}},2248:(e,t,n)=>{var r=n(9504),a=Map.prototype;e.exports={Map,set:r(a.set),get:r(a.get),has:r(a.has),remove:r(a.delete),proto:a}},2359:(e,t,n)=>{var r=n(6518),a=n(6080),l=n(6194),o=n(6223);r({target:"Map",proto:!0,real:!0,forced:!0},{find:function(e){var t=l(this),n=a(e,arguments.length>1?arguments[1]:void 0),r=o(t,(function(e,r){if(n(e,r,t))return{value:e}}),!0);return r&&r.value}})},2652:(e,t,n)=>{var r=n(6080),a=n(9565),l=n(8551),o=n(6823),i=n(4209),u=n(6198),s=n(1625),c=n(81),f=n(851),d=n(9539),p=TypeError,h=function(e,t){this.stopped=e,this.result=t},m=h.prototype;e.exports=function(e,t,n){var v,g,y,b,k,w,x,S=n&&n.that,E=!(!n||!n.AS_ENTRIES),C=!(!n||!n.IS_RECORD),N=!(!n||!n.IS_ITERATOR),_=!(!n||!n.INTERRUPTED),P=r(t,S),z=function(e){return v&&d(v,"normal",e),new h(!0,e)},j=function(e){return E?(l(e),_?P(e[0],e[1],z):P(e[0],e[1])):_?P(e,z):P(e)};if(C)v=e.iterator;else if(N)v=e;else{if(!(g=f(e)))throw new p(o(e)+" is not iterable");if(i(g)){for(y=0,b=u(e);b>y;y++)if((k=j(e[y]))&&s(m,k))return k;return new h(!1)}v=c(e,g)}for(w=C?e.next:v.next;!(x=a(w,v)).done;){try{k=j(x.value)}catch(e){d(v,"throw",e)}if("object"==typeof k&&k&&s(m,k))return k}return new h(!1)}},2777:(e,t,n)=>{var r=n(9565),a=n(34),l=n(757),o=n(5966),i=n(4270),u=n(8227),s=TypeError,c=u("toPrimitive");e.exports=function(e,t){if(!a(e)||l(e))return e;var n,u=o(e,c);if(u){if(void 0===t&&(t="default"),n=r(u,e,t),!a(n)||l(n))return n;throw new s("Can't convert object to primitive value")}return void 0===t&&(t="number"),i(e,t)}},2796:(e,t,n)=>{var r=n(9039),a=n(4901),l=/#|\.prototype\./,o=function(e,t){var n=u[i(e)];return n===c||n!==s&&(a(t)?r(t):!!t)},i=o.normalize=function(e){return String(e).replace(l,".").toLowerCase()},u=o.data={},s=o.NATIVE="N",c=o.POLYFILL="P";e.exports=o},2839:(e,t,n)=>{var r=n(4576).navigator,a=r&&r.userAgent;e.exports=a?String(a):""},3317:e=>{e.exports=function(e,t){return e===t||e!=e&&t!=t}},3392:(e,t,n)=>{var r=n(9504),a=0,l=Math.random(),o=r(1..toString);e.exports=function(e){return"Symbol("+(void 0===e?"":e)+")_"+o(++a+l,36)}},3706:(e,t,n)=>{var r=n(9504),a=n(4901),l=n(7629),o=r(Function.toString);a(l.inspectSource)||(l.inspectSource=function(e){return o(e)}),e.exports=l.inspectSource},3717:(e,t)=>{t.f=Object.getOwnPropertySymbols},3724:(e,t,n)=>{var r=n(9039);e.exports=!r((function(){return 7!==Object.defineProperty({},1,{get:function(){return 7}})[1]}))},3777:(e,t,n)=>{var r=n(6518),a=n(6080),l=n(6194),o=n(6223);r({target:"Map",proto:!0,real:!0,forced:!0},{every:function(e){var t=l(this),n=a(e,arguments.length>1?arguments[1]:void 0);return!1!==o(t,(function(e,r){if(!n(e,r,t))return!1}),!0)}})},4055:(e,t,n)=>{var r=n(4576),a=n(34),l=r.document,o=a(l)&&a(l.createElement);e.exports=function(e){return o?l.createElement(e):{}}},4117:e=>{e.exports=function(e){return null==e}},4190:(e,t,n)=>{var r=n(6518),a=n(6080),l=n(6194),o=n(2248),i=n(6223),u=o.Map,s=o.set;r({target:"Map",proto:!0,real:!0,forced:!0},{filter:function(e){var t=l(this),n=a(e,arguments.length>1?arguments[1]:void 0),r=new u;return i(t,(function(e,a){n(e,a,t)&&s(r,a,e)})),r}})},4209:(e,t,n)=>{var r=n(8227),a=n(6269),l=r("iterator"),o=Array.prototype;e.exports=function(e){return void 0!==e&&(a.Array===e||o[l]===e)}},4270:(e,t,n)=>{var r=n(9565),a=n(4901),l=n(34),o=TypeError;e.exports=function(e,t){var n,i;if("string"===t&&a(n=e.toString)&&!l(i=r(n,e)))return i;if(a(n=e.valueOf)&&!l(i=r(n,e)))return i;if("string"!==t&&a(n=e.toString)&&!l(i=r(n,e)))return i;throw new o("Can't convert object to primitive value")}},4477:(e,t)=>{function n(e,t){var n=e.length;e.push(t);e:for(;0<n;){var r=n-1>>>1,a=e[r];if(!(0<l(a,t)))break e;e[r]=t,e[n]=a,n=r}}function r(e){return 0===e.length?null:e[0]}function a(e){if(0===e.length)return null;var t=e[0],n=e.pop();if(n!==t){e[0]=n;e:for(var r=0,a=e.length,o=a>>>1;r<o;){var i=2*(r+1)-1,u=e[i],s=i+1,c=e[s];if(0>l(u,n))s<a&&0>l(c,u)?(e[r]=c,e[s]=n,r=s):(e[r]=u,e[i]=n,r=i);else{if(!(s<a&&0>l(c,n)))break e;e[r]=c,e[s]=n,r=s}}}return t}function l(e,t){var n=e.sortIndex-t.sortIndex;return 0!==n?n:e.id-t.id}if(t.unstable_now=void 0,"object"==typeof performance&&"function"==typeof performance.now){var o=performance;t.unstable_now=function(){return o.now()}}else{var i=Date,u=i.now();t.unstable_now=function(){return i.now()-u}}var s=[],c=[],f=1,d=null,p=3,h=!1,m=!1,v=!1,g="function"==typeof setTimeout?setTimeout:null,y="function"==typeof clearTimeout?clearTimeout:null,b="undefined"!=typeof setImmediate?setImmediate:null;function k(e){for(var t=r(c);null!==t;){if(null===t.callback)a(c);else{if(!(t.startTime<=e))break;a(c),t.sortIndex=t.expirationTime,n(s,t)}t=r(c)}}function w(e){if(v=!1,k(e),!m)if(null!==r(s))m=!0,T();else{var t=r(c);null!==t&&O(w,t.startTime-e)}}var x,S=!1,E=-1,C=5,N=-1;function _(){return!(t.unstable_now()-N<C)}function P(){if(S){var e=t.unstable_now();N=e;var n=!0;try{e:{m=!1,v&&(v=!1,y(E),E=-1),h=!0;var l=p;try{t:{for(k(e),d=r(s);null!==d&&!(d.expirationTime>e&&_());){var o=d.callback;if("function"==typeof o){d.callback=null,p=d.priorityLevel;var i=o(d.expirationTime<=e);if(e=t.unstable_now(),"function"==typeof i){d.callback=i,k(e),n=!0;break t}d===r(s)&&a(s),k(e)}else a(s);d=r(s)}if(null!==d)n=!0;else{var u=r(c);null!==u&&O(w,u.startTime-e),n=!1}}break e}finally{d=null,p=l,h=!1}n=void 0}}finally{n?x():S=!1}}}if("function"==typeof b)x=function(){b(P)};else if("undefined"!=typeof MessageChannel){var z=new MessageChannel,j=z.port2;z.port1.onmessage=P,x=function(){j.postMessage(null)}}else x=function(){g(P,0)};function T(){S||(S=!0,x())}function O(e,n){E=g((function(){e(t.unstable_now())}),n)}t.unstable_IdlePriority=5,t.unstable_ImmediatePriority=1,t.unstable_LowPriority=4,t.unstable_NormalPriority=3,t.unstable_Profiling=null,t.unstable_UserBlockingPriority=2,t.unstable_cancelCallback=function(e){e.callback=null},t.unstable_continueExecution=function(){m||h||(m=!0,T())},t.unstable_forceFrameRate=function(e){0>e||125<e?console.error("forceFrameRate takes a positive int between 0 and 125, forcing frame rates higher than 125 fps is not supported"):C=0<e?Math.floor(1e3/e):5},t.unstable_getCurrentPriorityLevel=function(){return p},t.unstable_getFirstCallbackNode=function(){return r(s)},t.unstable_next=function(e){switch(p){case 1:case 2:case 3:var t=3;break;default:t=p}var n=p;p=t;try{return e()}finally{p=n}},t.unstable_pauseExecution=function(){},t.unstable_requestPaint=function(){},t.unstable_runWithPriority=function(e,t){switch(e){case 1:case 2:case 3:case 4:case 5:break;default:e=3}var n=p;p=e;try{return t()}finally{p=n}},t.unstable_scheduleCallback=function(e,a,l){var o=t.unstable_now();switch(l="object"==typeof l&&null!==l&&"number"==typeof(l=l.delay)&&0<l?o+l:o,e){case 1:var i=-1;break;case 2:i=250;break;case 5:i=1073741823;break;case 4:i=1e4;break;default:i=5e3}return e={id:f++,callback:a,priorityLevel:e,startTime:l,expirationTime:i=l+i,sortIndex:-1},l>o?(e.sortIndex=l,n(c,e),null===r(s)&&e===r(c)&&(v?(y(E),E=-1):v=!0,O(w,l-o))):(e.sortIndex=i,n(s,e),m||h||(m=!0,T())),e},t.unstable_shouldYield=_,t.unstable_wrapCallback=function(e){var t=p;return function(){var n=p;p=t;try{return e.apply(this,arguments)}finally{p=n}}}},4495:(e,t,n)=>{var r=n(9519),a=n(9039),l=n(4576).String;e.exports=!!Object.getOwnPropertySymbols&&!a((function(){var e=Symbol("symbol detection");return!l(e)||!(Object(e)instanceof Symbol)||!Symbol.sham&&r&&r<41}))},4576:function(e,t,n){var r=function(e){return e&&e.Math===Math&&e};e.exports=r("object"==typeof globalThis&&globalThis)||r("object"==typeof window&&window)||r("object"==typeof self&&self)||r("object"==typeof n.g&&n.g)||r("object"==typeof this&&this)||function(){return this}()||Function("return this")()},4848:(e,t,n)=>{e.exports=n(9698)},4901:e=>{var t="object"==typeof document&&document.all;e.exports=void 0===t&&void 0!==t?function(e){return"function"==typeof e||e===t}:function(e){return"function"==typeof e}},4913:(e,t,n)=>{var r=n(3724),a=n(5917),l=n(8686),o=n(8551),i=n(6969),u=TypeError,s=Object.defineProperty,c=Object.getOwnPropertyDescriptor,f="enumerable",d="configurable",p="writable";t.f=r?l?function(e,t,n){if(o(e),t=i(t),o(n),"function"==typeof e&&"prototype"===t&&"value"in n&&p in n&&!n[p]){var r=c(e,t);r&&r[p]&&(e[t]=n.value,n={configurable:d in n?n[d]:r[d],enumerable:f in n?n[f]:r[f],writable:!1})}return s(e,t,n)}:s:function(e,t,n){if(o(e),t=i(t),o(n),a)try{return s(e,t,n)}catch(e){}if("get"in n||"set"in n)throw new u("Accessors not supported");return"value"in n&&(e[t]=n.value),e}},5031:(e,t,n)=>{var r=n(7751),a=n(9504),l=n(8480),o=n(3717),i=n(8551),u=a([].concat);e.exports=r("Reflect","ownKeys")||function(e){var t=l.f(i(e)),n=o.f;return n?u(t,n(e)):t}},5122:(e,t,n)=>{var r=n(6518),a=n(6194),l=n(2652),o=n(2248).set;r({target:"Map",proto:!0,real:!0,arity:1,forced:!0},{merge:function(e){for(var t=a(this),n=arguments.length,r=0;r<n;)l(arguments[r++],(function(e,n){o(t,e,n)}),{AS_ENTRIES:!0});return t}})},5338:(e,t,n)=>{!function e(){if("undefined"!=typeof __REACT_DEVTOOLS_GLOBAL_HOOK__&&"function"==typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE)try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(e){console.error(e)}}(),e.exports=n(1247)},5397:(e,t,n)=>{var r=n(7055),a=n(7750);e.exports=function(e){return r(a(e))}},5610:(e,t,n)=>{var r=n(1291),a=Math.max,l=Math.min;e.exports=function(e,t){var n=r(e);return n<0?a(n+t,0):l(n,t)}},5745:(e,t,n)=>{var r=n(7629);e.exports=function(e,t){return r[e]||(r[e]=t||{})}},5917:(e,t,n)=>{var r=n(3724),a=n(9039),l=n(4055);e.exports=!r&&!a((function(){return 7!==Object.defineProperty(l("div"),"a",{get:function(){return 7}}).a}))},5966:(e,t,n)=>{var r=n(9306),a=n(4117);e.exports=function(e,t){var n=e[t];return a(n)?void 0:r(n)}},6080:(e,t,n)=>{var r=n(7476),a=n(9306),l=n(616),o=r(r.bind);e.exports=function(e,t){return a(e),void 0===t?e:l?o(e,t):function(){return e.apply(t,arguments)}}},6097:(e,t,n)=>{var r=n(6518),a=n(6080),l=n(6194),o=n(6223);r({target:"Map",proto:!0,real:!0,forced:!0},{findKey:function(e){var t=l(this),n=a(e,arguments.length>1?arguments[1]:void 0),r=o(t,(function(e,r){if(n(e,r,t))return{key:r}}),!0);return r&&r.key}})},6119:(e,t,n)=>{var r=n(5745),a=n(3392),l=r("keys");e.exports=function(e){return l[e]||(l[e]=a(e))}},6194:(e,t,n)=>{var r=n(2248).has;e.exports=function(e){return r(e),e}},6198:(e,t,n)=>{var r=n(8014);e.exports=function(e){return r(e.length)}},6221:(e,t,n)=>{var r=n(6540);function a(e){var t="https://react.dev/errors/"+e;if(1<arguments.length){t+="?args[]="+encodeURIComponent(arguments[1]);for(var n=2;n<arguments.length;n++)t+="&args[]="+encodeURIComponent(arguments[n])}return"Minified React error #"+e+"; visit "+t+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function l(){}var o={d:{f:l,r:function(){throw Error(a(522))},D:l,C:l,L:l,m:l,X:l,S:l,M:l},p:0,findDOMNode:null},i=Symbol.for("react.portal"),u=r.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE;function s(e,t){return"font"===e?"":"string"==typeof t?"use-credentials"===t?t:"":void 0}t.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=o,t.createPortal=function(e,t){var n=2<arguments.length&&void 0!==arguments[2]?arguments[2]:null;if(!t||1!==t.nodeType&&9!==t.nodeType&&11!==t.nodeType)throw Error(a(299));return function(e,t,n){var r=3<arguments.length&&void 0!==arguments[3]?arguments[3]:null;return{$$typeof:i,key:null==r?null:""+r,children:e,containerInfo:t,implementation:n}}(e,t,null,n)},t.flushSync=function(e){var t=u.T,n=o.p;try{if(u.T=null,o.p=2,e)return e()}finally{u.T=t,o.p=n,o.d.f()}},t.preconnect=function(e,t){"string"==typeof e&&(t=t?"string"==typeof(t=t.crossOrigin)?"use-credentials"===t?t:"":void 0:null,o.d.C(e,t))},t.prefetchDNS=function(e){"string"==typeof e&&o.d.D(e)},t.preinit=function(e,t){if("string"==typeof e&&t&&"string"==typeof t.as){var n=t.as,r=s(n,t.crossOrigin),a="string"==typeof t.integrity?t.integrity:void 0,l="string"==typeof t.fetchPriority?t.fetchPriority:void 0;"style"===n?o.d.S(e,"string"==typeof t.precedence?t.precedence:void 0,{crossOrigin:r,integrity:a,fetchPriority:l}):"script"===n&&o.d.X(e,{crossOrigin:r,integrity:a,fetchPriority:l,nonce:"string"==typeof t.nonce?t.nonce:void 0})}},t.preinitModule=function(e,t){if("string"==typeof e)if("object"==typeof t&&null!==t){if(null==t.as||"script"===t.as){var n=s(t.as,t.crossOrigin);o.d.M(e,{crossOrigin:n,integrity:"string"==typeof t.integrity?t.integrity:void 0,nonce:"string"==typeof t.nonce?t.nonce:void 0})}}else null==t&&o.d.M(e)},t.preload=function(e,t){if("string"==typeof e&&"object"==typeof t&&null!==t&&"string"==typeof t.as){var n=t.as,r=s(n,t.crossOrigin);o.d.L(e,n,{crossOrigin:r,integrity:"string"==typeof t.integrity?t.integrity:void 0,nonce:"string"==typeof t.nonce?t.nonce:void 0,type:"string"==typeof t.type?t.type:void 0,fetchPriority:"string"==typeof t.fetchPriority?t.fetchPriority:void 0,referrerPolicy:"string"==typeof t.referrerPolicy?t.referrerPolicy:void 0,imageSrcSet:"string"==typeof t.imageSrcSet?t.imageSrcSet:void 0,imageSizes:"string"==typeof t.imageSizes?t.imageSizes:void 0,media:"string"==typeof t.media?t.media:void 0})}},t.preloadModule=function(e,t){if("string"==typeof e)if(t){var n=s(t.as,t.crossOrigin);o.d.m(e,{as:"string"==typeof t.as&&"script"!==t.as?t.as:void 0,crossOrigin:n,integrity:"string"==typeof t.integrity?t.integrity:void 0})}else o.d.m(e)},t.requestFormReset=function(e){o.d.r(e)},t.unstable_batchedUpdates=function(e,t){return e(t)},t.useFormState=function(e,t,n){return u.H.useFormState(e,t,n)},t.useFormStatus=function(){return u.H.useHostTransitionStatus()},t.version="19.0.0"},6223:(e,t,n)=>{var r=n(9504),a=n(507),l=n(2248),o=l.Map,i=l.proto,u=r(i.forEach),s=r(i.entries),c=s(new o).next;e.exports=function(e,t,n){return n?a({iterator:s(e),next:c},(function(e){return t(e[1],e[0])})):u(e,t)}},6269:e=>{e.exports={}},6395:e=>{e.exports=!1},6518:(e,t,n)=>{var r=n(4576),a=n(7347).f,l=n(6699),o=n(6840),i=n(9433),u=n(7740),s=n(2796);e.exports=function(e,t){var n,c,f,d,p,h=e.target,m=e.global,v=e.stat;if(n=m?r:v?r[h]||i(h,{}):r[h]&&r[h].prototype)for(c in t){if(d=t[c],f=e.dontCallGetSet?(p=a(n,c))&&p.value:n[c],!s(m?c:h+(v?".":"#")+c,e.forced)&&void 0!==f){if(typeof d==typeof f)continue;u(d,f)}(e.sham||f&&f.sham)&&l(d,"sham",!0),o(n,c,d,e)}}},6540:(e,t,n)=>{e.exports=n(9869)},6699:(e,t,n)=>{var r=n(3724),a=n(4913),l=n(6980);e.exports=r?function(e,t,n){return a.f(e,t,l(1,n))}:function(e,t,n){return e[t]=n,e}},6823:e=>{var t=String;e.exports=function(e){try{return t(e)}catch(e){return"Object"}}},6840:(e,t,n)=>{var r=n(4901),a=n(4913),l=n(283),o=n(9433);e.exports=function(e,t,n,i){i||(i={});var u=i.enumerable,s=void 0!==i.name?i.name:t;if(r(n)&&l(n,s,i),i.global)u?e[t]=n:o(t,n);else{try{i.unsafe?e[t]&&(u=!0):delete e[t]}catch(e){}u?e[t]=n:a.f(e,t,{value:n,enumerable:!1,configurable:!i.nonConfigurable,writable:!i.nonWritable})}return e}},6955:(e,t,n)=>{var r=n(2140),a=n(4901),l=n(2195),o=n(8227)("toStringTag"),i=Object,u="Arguments"===l(function(){return arguments}());e.exports=r?l:function(e){var t,n,r;return void 0===e?"Undefined":null===e?"Null":"string"==typeof(n=function(e,t){try{return e[t]}catch(e){}}(t=i(e),o))?n:u?l(t):"Object"===(r=l(t))&&a(t.callee)?"Arguments":r}},6969:(e,t,n)=>{var r=n(2777),a=n(757);e.exports=function(e){var t=r(e,"string");return a(t)?t:t+""}},6980:e=>{e.exports=function(e,t){return{enumerable:!(1&e),configurable:!(2&e),writable:!(4&e),value:t}}},7040:(e,t,n)=>{var r=n(4495);e.exports=r&&!Symbol.sham&&"symbol"==typeof Symbol.iterator},7055:(e,t,n)=>{var r=n(9504),a=n(9039),l=n(2195),o=Object,i=r("".split);e.exports=a((function(){return!o("z").propertyIsEnumerable(0)}))?function(e){return"String"===l(e)?i(e,""):o(e)}:o},7268:(e,t,n)=>{var r=n(6518),a=n(6080),l=n(6194),o=n(6223);r({target:"Map",proto:!0,real:!0,forced:!0},{some:function(e){var t=l(this),n=a(e,arguments.length>1?arguments[1]:void 0);return!0===o(t,(function(e,r){if(n(e,r,t))return!0}),!0)}})},7273:(e,t,n)=>{var r=n(6518),a=n(3317),l=n(6194),o=n(6223);r({target:"Map",proto:!0,real:!0,forced:!0},{includes:function(e){return!0===o(l(this),(function(t){if(a(t,e))return!0}),!0)}})},7347:(e,t,n)=>{var r=n(3724),a=n(9565),l=n(8773),o=n(6980),i=n(5397),u=n(6969),s=n(9297),c=n(5917),f=Object.getOwnPropertyDescriptor;t.f=r?f:function(e,t){if(e=i(e),t=u(t),c)try{return f(e,t)}catch(e){}if(s(e,t))return o(!a(l.f,e,t),e[t])}},7415:(e,t,n)=>{var r=n(6518),a=n(6194),l=n(6223);r({target:"Map",proto:!0,real:!0,forced:!0},{keyOf:function(e){var t=l(a(this),(function(t,n){if(t===e)return{key:n}}),!0);return t&&t.key}})},7476:(e,t,n)=>{var r=n(2195),a=n(9504);e.exports=function(e){if("Function"===r(e))return a(e)}},7583:(e,t,n)=>{var r=n(6518),a=n(6080),l=n(6194),o=n(2248),i=n(6223),u=o.Map,s=o.set;r({target:"Map",proto:!0,real:!0,forced:!0},{mapValues:function(e){var t=l(this),n=a(e,arguments.length>1?arguments[1]:void 0),r=new u;return i(t,(function(e,a){s(r,a,n(e,a,t))})),r}})},7629:(e,t,n)=>{var r=n(6395),a=n(4576),l=n(9433),o="__core-js_shared__",i=e.exports=a[o]||l(o,{});(i.versions||(i.versions=[])).push({version:"3.40.0",mode:r?"pure":"global",copyright:"© 2014-2025 Denis Pushkarev (zloirock.ru)",license:"https://github.com/zloirock/core-js/blob/v3.40.0/LICENSE",source:"https://github.com/zloirock/core-js"})},7740:(e,t,n)=>{var r=n(9297),a=n(5031),l=n(7347),o=n(4913);e.exports=function(e,t,n){for(var i=a(t),u=o.f,s=l.f,c=0;c<i.length;c++){var f=i[c];r(e,f)||n&&r(n,f)||u(e,f,s(t,f))}}},7750:(e,t,n)=>{var r=n(4117),a=TypeError;e.exports=function(e){if(r(e))throw new a("Can't call method on "+e);return e}},7751:(e,t,n)=>{var r=n(4576),a=n(4901);e.exports=function(e,t){return arguments.length<2?(n=r[e],a(n)?n:void 0):r[e]&&r[e][t];var n}},8014:(e,t,n)=>{var r=n(1291),a=Math.min;e.exports=function(e){var t=r(e);return t>0?a(t,9007199254740991):0}},8227:(e,t,n)=>{var r=n(4576),a=n(5745),l=n(9297),o=n(3392),i=n(4495),u=n(7040),s=r.Symbol,c=a("wks"),f=u?s.for||s:s&&s.withoutSetter||o;e.exports=function(e){return l(c,e)||(c[e]=i&&l(s,e)?s[e]:f("Symbol."+e)),c[e]}},8480:(e,t,n)=>{var r=n(1828),a=n(8727).concat("length","prototype");t.f=Object.getOwnPropertyNames||function(e){return r(e,a)}},8551:(e,t,n)=>{var r=n(34),a=String,l=TypeError;e.exports=function(e){if(r(e))return e;throw new l(a(e)+" is not an object")}},8622:(e,t,n)=>{var r=n(4576),a=n(4901),l=r.WeakMap;e.exports=a(l)&&/native code/.test(String(l))},8686:(e,t,n)=>{var r=n(3724),a=n(9039);e.exports=r&&a((function(){return 42!==Object.defineProperty((function(){}),"prototype",{value:42,writable:!1}).prototype}))},8727:e=>{e.exports=["constructor","hasOwnProperty","isPrototypeOf","propertyIsEnumerable","toLocaleString","toString","valueOf"]},8773:(e,t)=>{var n={}.propertyIsEnumerable,r=Object.getOwnPropertyDescriptor,a=r&&!n.call({1:2},1);t.f=a?function(e){var t=r(this,e);return!!t&&t.enumerable}:n},8981:(e,t,n)=>{var r=n(7750),a=Object;e.exports=function(e){return a(r(e))}},9039:e=>{e.exports=function(e){try{return!!e()}catch(e){return!0}}},9297:(e,t,n)=>{var r=n(9504),a=n(8981),l=r({}.hasOwnProperty);e.exports=Object.hasOwn||function(e,t){return l(a(e),t)}},9306:(e,t,n)=>{var r=n(4901),a=n(6823),l=TypeError;e.exports=function(e){if(r(e))return e;throw new l(a(e)+" is not a function")}},9433:(e,t,n)=>{var r=n(4576),a=Object.defineProperty;e.exports=function(e,t){try{a(r,e,{value:t,configurable:!0,writable:!0})}catch(n){r[e]=t}return t}},9504:(e,t,n)=>{var r=n(616),a=Function.prototype,l=a.call,o=r&&a.bind.bind(l,l);e.exports=r?o:function(e){return function(){return l.apply(e,arguments)}}},9519:(e,t,n)=>{var r,a,l=n(4576),o=n(2839),i=l.process,u=l.Deno,s=i&&i.versions||u&&u.version,c=s&&s.v8;c&&(a=(r=c.split("."))[0]>0&&r[0]<4?1:+(r[0]+r[1])),!a&&o&&(!(r=o.match(/Edge\/(\d+)/))||r[1]>=74)&&(r=o.match(/Chrome\/(\d+)/))&&(a=+r[1]),e.exports=a},9539:(e,t,n)=>{var r=n(9565),a=n(8551),l=n(5966);e.exports=function(e,t,n){var o,i;a(e);try{if(!(o=l(e,"return"))){if("throw"===t)throw n;return n}o=r(o,e)}catch(e){i=!0,o=e}if("throw"===t)throw n;if(i)throw o;return a(o),n}},9565:(e,t,n)=>{var r=n(616),a=Function.prototype.call;e.exports=r?a.bind(a):function(){return a.apply(a,arguments)}},9617:(e,t,n)=>{var r=n(5397),a=n(5610),l=n(6198),o=function(e){return function(t,n,o){var i=r(t),u=l(i);if(0===u)return!e&&-1;var s,c=a(o,u);if(e&&n!=n){for(;u>c;)if((s=i[c++])!=s)return!0}else for(;u>c;c++)if((e||c in i)&&i[c]===n)return e||c||0;return!e&&-1}};e.exports={includes:o(!0),indexOf:o(!1)}},9698:(e,t)=>{var n=Symbol.for("react.transitional.element"),r=Symbol.for("react.fragment");function a(e,t,r){var a=null;if(void 0!==r&&(a=""+r),void 0!==t.key&&(a=""+t.key),"key"in t)for(var l in r={},t)"key"!==l&&(r[l]=t[l]);else r=t;return t=r.ref,{$$typeof:n,type:e,key:a,ref:void 0!==t?t:null,props:r}}t.Fragment=r,t.jsx=a,t.jsxs=a},9733:(e,t,n)=>{var r=n(6518),a=n(9306),l=n(6194),o=n(2248),i=TypeError,u=o.get,s=o.has,c=o.set;r({target:"Map",proto:!0,real:!0,forced:!0},{update:function(e,t){var n=l(this),r=arguments.length;a(t);var o=s(n,e);if(!o&&r<3)throw new i("Updating absent value");var f=o?u(n,e):a(r>2?arguments[2]:void 0)(e,n);return c(n,e,t(f,e,n)),n}})},9869:(e,t)=>{var n=Symbol.for("react.transitional.element"),r=Symbol.for("react.portal"),a=Symbol.for("react.fragment"),l=Symbol.for("react.strict_mode"),o=Symbol.for("react.profiler"),i=Symbol.for("react.consumer"),u=Symbol.for("react.context"),s=Symbol.for("react.forward_ref"),c=Symbol.for("react.suspense"),f=Symbol.for("react.memo"),d=Symbol.for("react.lazy"),p=Symbol.iterator,h={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},m=Object.assign,v={};function g(e,t,n){this.props=e,this.context=t,this.refs=v,this.updater=n||h}function y(){}function b(e,t,n){this.props=e,this.context=t,this.refs=v,this.updater=n||h}g.prototype.isReactComponent={},g.prototype.setState=function(e,t){if("object"!=typeof e&&"function"!=typeof e&&null!=e)throw Error("takes an object of state variables to update or a function which returns an object of state variables.");this.updater.enqueueSetState(this,e,t,"setState")},g.prototype.forceUpdate=function(e){this.updater.enqueueForceUpdate(this,e,"forceUpdate")},y.prototype=g.prototype;var k=b.prototype=new y;k.constructor=b,m(k,g.prototype),k.isPureReactComponent=!0;var w=Array.isArray,x={H:null,A:null,T:null,S:null},S=Object.prototype.hasOwnProperty;function E(e,t,r,a,l,o){return r=o.ref,{$$typeof:n,type:e,key:t,ref:void 0!==r?r:null,props:o}}function C(e){return"object"==typeof e&&null!==e&&e.$$typeof===n}var N=/\/+/g;function _(e,t){return"object"==typeof e&&null!==e&&null!=e.key?(n=""+e.key,r={"=":"=0",":":"=2"},"$"+n.replace(/[=:]/g,(function(e){return r[e]}))):t.toString(36);var n,r}function P(){}function z(e,t,a,l,o){var i=typeof e;"undefined"!==i&&"boolean"!==i||(e=null);var u,s,c=!1;if(null===e)c=!0;else switch(i){case"bigint":case"string":case"number":c=!0;break;case"object":switch(e.$$typeof){case n:case r:c=!0;break;case d:return z((c=e._init)(e._payload),t,a,l,o)}}if(c)return o=o(e),c=""===l?"."+_(e,0):l,w(o)?(a="",null!=c&&(a=c.replace(N,"$&/")+"/"),z(o,t,a,"",(function(e){return e}))):null!=o&&(C(o)&&(u=o,s=a+(null==o.key||e&&e.key===o.key?"":(""+o.key).replace(N,"$&/")+"/")+c,o=E(u.type,s,void 0,0,0,u.props)),t.push(o)),1;c=0;var f,h=""===l?".":l+":";if(w(e))for(var m=0;m<e.length;m++)c+=z(l=e[m],t,a,i=h+_(l,m),o);else if("function"==typeof(m=null===(f=e)||"object"!=typeof f?null:"function"==typeof(f=p&&f[p]||f["@@iterator"])?f:null))for(e=m.call(e),m=0;!(l=e.next()).done;)c+=z(l=l.value,t,a,i=h+_(l,m++),o);else if("object"===i){if("function"==typeof e.then)return z(function(e){switch(e.status){case"fulfilled":return e.value;case"rejected":throw e.reason;default:switch("string"==typeof e.status?e.then(P,P):(e.status="pending",e.then((function(t){"pending"===e.status&&(e.status="fulfilled",e.value=t)}),(function(t){"pending"===e.status&&(e.status="rejected",e.reason=t)}))),e.status){case"fulfilled":return e.value;case"rejected":throw e.reason}}throw e}(e),t,a,l,o);throw t=String(e),Error("Objects are not valid as a React child (found: "+("[object Object]"===t?"object with keys {"+Object.keys(e).join(", ")+"}":t)+"). If you meant to render a collection of children, use an array instead.")}return c}function j(e,t,n){if(null==e)return e;var r=[],a=0;return z(e,r,"","",(function(e){return t.call(n,e,a++)})),r}function T(e){if(-1===e._status){var t=e._result;(t=t()).then((function(t){0!==e._status&&-1!==e._status||(e._status=1,e._result=t)}),(function(t){0!==e._status&&-1!==e._status||(e._status=2,e._result=t)})),-1===e._status&&(e._status=0,e._result=t)}if(1===e._status)return e._result.default;throw e._result}var O="function"==typeof reportError?reportError:function(e){if("object"==typeof window&&"function"==typeof window.ErrorEvent){var t=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:"object"==typeof e&&null!==e&&"string"==typeof e.message?String(e.message):String(e),error:e});if(!window.dispatchEvent(t))return}else if("object"==typeof process&&"function"==typeof process.emit)return void process.emit("uncaughtException",e);console.error(e)};function L(){}t.Children={map:j,forEach:function(e,t,n){j(e,(function(){t.apply(this,arguments)}),n)},count:function(e){var t=0;return j(e,(function(){t++})),t},toArray:function(e){return j(e,(function(e){return e}))||[]},only:function(e){if(!C(e))throw Error("React.Children.only expected to receive a single React element child.");return e}},t.Component=g,t.Fragment=a,t.Profiler=o,t.PureComponent=b,t.StrictMode=l,t.Suspense=c,t.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=x,t.act=function(){throw Error("act(...) is not supported in production builds of React.")},t.cache=function(e){return function(){return e.apply(null,arguments)}},t.cloneElement=function(e,t,n){if(null==e)throw Error("The argument must be a React element, but you passed "+e+".");var r=m({},e.props),a=e.key;if(null!=t)for(l in t.ref,void 0!==t.key&&(a=""+t.key),t)!S.call(t,l)||"key"===l||"__self"===l||"__source"===l||"ref"===l&&void 0===t.ref||(r[l]=t[l]);var l=arguments.length-2;if(1===l)r.children=n;else if(1<l){for(var o=Array(l),i=0;i<l;i++)o[i]=arguments[i+2];r.children=o}return E(e.type,a,void 0,0,0,r)},t.createContext=function(e){return(e={$$typeof:u,_currentValue:e,_currentValue2:e,_threadCount:0,Provider:null,Consumer:null}).Provider=e,e.Consumer={$$typeof:i,_context:e},e},t.createElement=function(e,t,n){var r,a={},l=null;if(null!=t)for(r in void 0!==t.key&&(l=""+t.key),t)S.call(t,r)&&"key"!==r&&"__self"!==r&&"__source"!==r&&(a[r]=t[r]);var o=arguments.length-2;if(1===o)a.children=n;else if(1<o){for(var i=Array(o),u=0;u<o;u++)i[u]=arguments[u+2];a.children=i}if(e&&e.defaultProps)for(r in o=e.defaultProps)void 0===a[r]&&(a[r]=o[r]);return E(e,l,void 0,0,0,a)},t.createRef=function(){return{current:null}},t.forwardRef=function(e){return{$$typeof:s,render:e}},t.isValidElement=C,t.lazy=function(e){return{$$typeof:d,_payload:{_status:-1,_result:e},_init:T}},t.memo=function(e,t){return{$$typeof:f,type:e,compare:void 0===t?null:t}},t.startTransition=function(e){var t=x.T,n={};x.T=n;try{var r=e(),a=x.S;null!==a&&a(n,r),"object"==typeof r&&null!==r&&"function"==typeof r.then&&r.then(L,O)}catch(e){O(e)}finally{x.T=t}},t.unstable_useCacheRefresh=function(){return x.H.useCacheRefresh()},t.use=function(e){return x.H.use(e)},t.useActionState=function(e,t,n){return x.H.useActionState(e,t,n)},t.useCallback=function(e,t){return x.H.useCallback(e,t)},t.useContext=function(e){return x.H.useContext(e)},t.useDebugValue=function(){},t.useDeferredValue=function(e,t){return x.H.useDeferredValue(e,t)},t.useEffect=function(e,t){return x.H.useEffect(e,t)},t.useId=function(){return x.H.useId()},t.useImperativeHandle=function(e,t,n){return x.H.useImperativeHandle(e,t,n)},t.useInsertionEffect=function(e,t){return x.H.useInsertionEffect(e,t)},t.useLayoutEffect=function(e,t){return x.H.useLayoutEffect(e,t)},t.useMemo=function(e,t){return x.H.useMemo(e,t)},t.useOptimistic=function(e,t){return x.H.useOptimistic(e,t)},t.useReducer=function(e,t,n){return x.H.useReducer(e,t,n)},t.useRef=function(e){return x.H.useRef(e)},t.useState=function(e){return x.H.useState(e)},t.useSyncExternalStore=function(e,t,n){return x.H.useSyncExternalStore(e,t,n)},t.useTransition=function(){return x.H.useTransition()},t.version="19.0.0"},9929:(e,t,n)=>{var r=n(6518),a=n(6080),l=n(6194),o=n(2248),i=n(6223),u=o.Map,s=o.set;r({target:"Map",proto:!0,real:!0,forced:!0},{mapKeys:function(e){var t=l(this),n=a(e,arguments.length>1?arguments[1]:void 0),r=new u;return i(t,(function(e,a){s(r,n(e,a,t),e)})),r}})},9982:(e,t,n)=>{e.exports=n(4477)}},t={};function n(r){var a=t[r];if(void 0!==a)return a.exports;var l=t[r]={exports:{}};return e[r].call(l.exports,l,l.exports,n),l.exports}n.g=function(){if("object"==typeof globalThis)return globalThis;try{return this||new Function("return this")()}catch(e){if("object"==typeof window)return window}}();var r=n(5338);n(1517),n(3777),n(4190),n(2359),n(6097),n(7273),n(7415),n(9929),n(7583),n(5122),n(230),n(7268),n(9733);class a{constructor(){this.path="||@ request.path @||",this.method="||@ request.method @||",this.clientHost="||@ request.client.host @||",this.clientPort=parseInt("||@ request.client.port @||"),this.pathParams=new Map(Object.entries(JSON.parse("||@ request.params.path|safe_json @||"))),this.queryParams=new Map(Object.entries(JSON.parse("||@ request.params.query|safe_json @||"))),this.headers=new Map(Object.entries(JSON.parse("||@ request.headers|safe_json @||"))),this.cookies=new Map(Object.entries(JSON.parse("||@ request.cookies|safe_json @||")))}}var l=n(6540),o={outline:{xmlns:"http://www.w3.org/2000/svg",width:24,height:24,viewBox:"0 0 24 24",fill:"none",stroke:"currentColor",strokeWidth:2,strokeLinecap:"round",strokeLinejoin:"round"},filled:{xmlns:"http://www.w3.org/2000/svg",width:24,height:24,viewBox:"0 0 24 24",fill:"currentColor",stroke:"none"}};const i=(e,t,n,r)=>{const a=(0,l.forwardRef)((({color:n="currentColor",size:a=24,stroke:i=2,title:u,className:s,children:c,...f},d)=>(0,l.createElement)("svg",{ref:d,...o[e],width:a,height:a,className:["tabler-icon",`tabler-icon-${t}`,s].join(" "),..."filled"===e?{fill:n}:{strokeWidth:i,stroke:n},...f},[u&&(0,l.createElement)("title",{key:"svg-title"},u),...r.map((([e,t])=>(0,l.createElement)(e,t))),...Array.isArray(c)?c:[c]])));return a.displayName=`${n}`,a};var u=i("outline","book","IconBook",[["path",{d:"M3 19a9 9 0 0 1 9 0a9 9 0 0 1 9 0",key:"svg-0"}],["path",{d:"M3 6a9 9 0 0 1 9 0a9 9 0 0 1 9 0",key:"svg-1"}],["path",{d:"M3 6l0 13",key:"svg-2"}],["path",{d:"M12 6l0 13",key:"svg-3"}],["path",{d:"M21 6l0 13",key:"svg-4"}]]),s=i("outline","script","IconScript",[["path",{d:"M17 20h-11a3 3 0 0 1 0 -6h11a3 3 0 0 0 0 6h1a3 3 0 0 0 3 -3v-11a2 2 0 0 0 -2 -2h-10a2 2 0 0 0 -2 2v8",key:"svg-0"}]]),c=i("outline","brand-github","IconBrandGithub",[["path",{d:"M9 19c-4.3 1.4 -4.3 -2.5 -6 -3m12 5v-3.5c0 -1 .1 -1.4 -.5 -2c2.8 -.3 5.5 -1.4 5.5 -6a4.6 4.6 0 0 0 -1.3 -3.2a4.2 4.2 0 0 0 -.1 -3.2s-1.1 -.3 -3.5 1.3a12.3 12.3 0 0 0 -6.2 0c-2.4 -1.6 -3.5 -1.3 -3.5 -1.3a4.2 4.2 0 0 0 -.1 3.2a4.6 4.6 0 0 0 -1.3 3.2c0 4.6 2.7 5.7 5.5 6c-.6 .6 -.6 1.2 -.5 2v3.5",key:"svg-0"}]]),f=i("outline","x","IconX",[["path",{d:"M18 6l-12 12",key:"svg-0"}],["path",{d:"M6 6l12 12",key:"svg-1"}]]),d=i("outline","menu-2","IconMenu2",[["path",{d:"M4 6l16 0",key:"svg-0"}],["path",{d:"M4 12l16 0",key:"svg-1"}],["path",{d:"M4 18l16 0",key:"svg-2"}]]),p=n(961),h=n(4848);function m(e){let{...t}=e;return(0,h.jsx)("svg",{viewBox:"0 0 1024 1024",fill:"currentColor",xmlns:"http://www.w3.org/2000/svg",...t,children:(0,h.jsx)("path",{d:"M734.61,441.99c-7.05-15.75-9.96-32.93-8.26-50.83,1.72-18.08,13.06-35.59,26.3-52.84,.66-1.73,8.25-11.78-7.07-5.84-3.79,.74-16.01,5.48-19.74,6.61-32.26,9.73-43.32,29.76-52.49,66.59-3.39,13.6-3.19,28.47-4.11,42.8-1.09,17.02-1.47,34.1-2.49,51.13-.4,6.65-.96,13.51-2.72,19.8-9.42,33.62-43.08,55.79-72.92,48.45-24.22-5.95-39.46-26.85-40.25-55.39-.36-13.17,1.9-25.66,8.56-36.6,4.8-7.89,9.78-15.64,14.95-23.2,14.1-20.64,26-42.69,33.76-67.61,4.88-15.67,5.79-31.99,6.1-48.32,.25-12.84,.78-25.97-1.13-38.5-5.01-32.85-16.97-62.77-33.87-89.52-19.85-31.43-45.02-56.86-72.02-79.83-40.98-32.52-5.03,50.15-4.46,81.96,.35,19.63,.31,39.17-3.85,58.36-6.6,30.4-20.91,55.67-39.07,78.36-20.7,25.85-44.14,48.14-67.7,70.28-27.96,26.28-56.18,52.2-81.3,82.27-27.42,32.82-48.89,69.9-61.12,113.69-5.24,18.77-10.29,37.67-11.06,57.49-.77,19.99-.76,40.03-.77,60.04,0,7.1,.32,14.35,1.55,21.29,10.52,59.38,34.5,109.72,75.2,148.62,29.12,27.83,52.9,41.17,85.72,53.58-10.19-10.79-16.94-20.99-26.23-35.86-35.68-57.1-48.33-121.18-40.92-190.73,3.28-30.8,11.37-60.07,22.68-88.23,14.75-36.72,19.81-50.3,45.56-81.01,21.79-31.69,23.03-10.04,22.49-6.18-3.86,27.43-6.75,54.98-4.67,82.8,2.13,28.45,12.22,52.83,29.27,72.9,21.31,25.08,47.84,30.81,76.54,23.2,19.23-5.1,38.46-8.62,58.18-7.81,8.4,.35,16.84,.07,24.9,4.1,27.05,13.52,45.64,36.41,55.25,68.43,8.76,29.2,7.75,59.16,1.55,88.8-4.91,23.47-19.9,51.68-36.69,69.87,9.99-7.51,18.59-17.66,29.05-26,11.97-8.06,46.48-42.15,72.77-72.74,117.87-120.71,14.8-391.43,4.53-414.37Z",fill:"#e05922"})})}function v(e){let{...t}=e;return(0,h.jsxs)("svg",{viewBox:"0 0 1024 1024",xmlns:"http://www.w3.org/2000/svg",...t,children:[(0,h.jsx)("path",{d:"M341.49,413.75c-4.13,3.29-7.85,7.08-11.69,10.63-3.67,3.77-7.31,7.52-10.71,11.48-13.54,15.88-25.22,33.01-34.21,51.46-18.33,36.64-27.95,76.95-29.36,116.91-.64,19.99,.85,39.91,4.29,59.3,3.76,19.34,9.4,38.15,17.34,55.83,7.83,17.71,18.02,34.17,30.16,49.5,12.02,15.37,25.79,29.57,40.97,42.3,15.05,12.88,31.44,24.38,48.76,34.48,17.22,10.29,35.46,19.01,54.23,26.65,18.76,7.69,38.19,13.94,57.97,19.23,19.85,5.05,40.06,9.06,60.53,12.06,20.48,3,41.23,4.61,62.11,5.66,20.89,.87,41.91,.7,63.02-.11,21.08-1.09,42.29-2.61,63.49-5.17,21.26-2.26,125.44-27.98,33.77,3.27-10.2,3.64-20.42,7.38-30.83,10.52-20.71,6.81-41.99,11.73-63.46,16.19-21.55,3.99-43.37,7.12-65.41,8.72-5.5,.51-11.03,.67-16.56,.93-5.53,.27-11.07,.48-16.62,.43-5.55,0-11.1,.12-16.66,.01l-16.68-.66c-44.48-2.39-89.13-10.67-131.81-26.02-21.35-7.61-42.13-17.1-62.13-28.2-19.9-11.28-38.93-24.35-56.6-39.25-17.54-15.05-33.78-31.87-47.88-50.57-13.9-18.74-26.19-39.3-34.7-61.32-8.61-21.93-14.47-44.79-17.21-67.89-2.73-23.1-2.51-46.35,.21-69.14,5.74-45.52,21.81-89.49,47.76-126.49,19.27-27.67,45.15-50.46,73.59-67.06,15.58-9.09,12.19-.88,4.33,6.3Z",fill:"#e05922"}),(0,h.jsx)("path",{d:"M459.32,516.03c4.44-3.84,10.57,1.79,7.52,6.81-4.99,8.18-7.54,17.19-6.95,25.72,.51,8.41,3.67,15.98,8.7,22.02,4.93,5.69,11.76,10.38,19.26,13.32,15.06,6.17,31.86,5.47,44.44-.91,12.66-6.28,23.08-18.15,28.39-31.95,5.51-13.78,5.92-28.56,1.92-41.61-1.95-6.6-5.4-12.76-9.53-18.72l-1.51-2.25-1.65-1.92-1.6-1.95-1.88-1.88c-1.19-1.28-2.47-2.53-3.91-3.69-1.37-1.2-2.7-2.43-4.26-3.48-11.78-9.12-26.26-15.18-40.74-17.29-7.29-.85-14.59-1.1-21.9-.32-1.83,.23-3.64,.65-5.47,.93l-2.74,.44-2.71,.76-5.19,1.38c-.17,.05-.34,.1-.51,.16l-5.12,1.95c-14.45,5.37-27.44,13.84-37.15,25.22-10.01,11.1-17.29,24.81-21.02,39.89-3.75,15.11-3.76,31.41-.7,47.65,1.48,8.14,4.12,16.18,7.32,24.12,2.98,7.83,6.67,14.57,11.45,21.26,18.93,26.56,53.25,45,90.6,51.63,18.72,3.52,38.35,4.07,57.9,1.5,9.88-1.38,19.8-3.54,29.53-6.75,5.52-1.82,9.38,5.57,4.65,8.96-8.94,6.4-18.81,11.76-29.12,16-20.39,8.5-42.63,12.95-65.24,13.53-22.6,.48-45.76-2.69-68.01-10.85-11.16-3.97-22-9.31-32.32-15.87-10.29-6.58-20.16-14.38-28.69-23.73-8.53-9.18-16.21-19.97-21.97-30.92-5.74-10.74-10.56-22.15-14.02-34.12-7.19-23.87-8.8-49.99-4.32-75.53,4.59-25.55,16.38-50.09,33.24-70.47,17-20.42,39.62-35.63,63.87-44.89l9.02-3.2c.14-.05,.28-.09,.42-.13l9.38-2.44,4.81-1.18,4.93-.81c3.3-.5,6.57-1.09,9.89-1.45,13.33-1.25,26.85-.88,40.07,1.34,26.3,5.06,51,16.33,70.94,33.99,2.56,2.1,4.94,4.49,7.31,6.88,2.41,2.31,4.69,4.83,6.88,7.48l3.18,3.77c.08,.09,.15,.19,.22,.29l2.99,4.1,2.92,4.11c.08,.11,.15,.22,.22,.34l2.5,4.17c6.99,11.47,12.28,24.45,14.99,38.08,2.75,13.59,3.16,27.69,.94,41.21-2.22,13.53-6.8,26.43-13.32,38.13-6.47,11.72-15.09,22.19-25.47,30.81-10.17,8.62-22.47,15.42-35.6,19.14-13.12,3.76-26.95,4.55-39.89,2.27-12.95-2.27-25.1-7.13-35.66-14.31-10.47-7.25-19.72-16.64-25.64-28.65-5.59-11.7-7.51-25.35-4.7-37.54,1.25-6.13,3.74-11.79,6.69-16.94,3.23-5.02,6.85-9.61,11.16-13.36,.09-.07,.17-.15,.26-.22Z",fill:"#018080"}),(0,h.jsx)("path",{d:"M98.42,646.77l.07-16.47,.52-16.51c1.03-22.04,3.8-44.08,8.11-65.99,8.71-43.73,24.57-87.01,49.25-125.97,24.47-38.99,58.12-73.12,97.84-98.05,19.84-12.47,41.14-22.6,63.18-30.52,22.13-7.67,44.99-13.18,68.18-16.14,46.35-5.78,94-2.55,139.16,12.63,11.27,3.81,22.39,8.57,33.11,14.01,10.64,5.57,20.7,11.96,30.36,18.8,9.61,6.92,18.64,14.52,27.17,22.58,8.41,8.19,16.39,16.76,23.76,25.83,29.1,36.36,48.92,80.72,52.58,126.94,1.85,23.01-.5,46.21-7.07,67.84-6.51,21.64-22.14,38.76-19.52,28.34,3.07-10.23,4.93-20.71,6.18-31.1,1.03-10.45,1.32-20.85,.77-31.14-.86-10.29-1.86-20.48-4.35-30.41-9.05-39.83-30.33-75.81-58.1-104.69-27.76-29.04-62.14-50.98-99.26-61.61-37.74-10.7-78.7-12.91-118.28-7.67-39.65,5.17-78.28,17.85-112.81,37.81-34.75,19.7-64.98,47.12-89.77,79.46-24.82,32.41-43.74,69.97-58.02,109.77-7.22,19.9-13.15,40.44-18.35,61.32l-3.68,15.75-3.28,15.89-3.14,15.99c-2.34,13.37-4.27-5.72-4.62-16.7Z",fill:"#1ab7d1"}),(0,h.jsx)("path",{d:"M922.46,330.55c.86,10.35,1.65,20.72,2.14,31.12,1.23,20.81,1.19,41.7,.44,62.66-1.75,41.88-7.37,84.19-19.75,125.55-12.21,41.26-31.57,81.7-59.3,116.49-13.72,17.46-29.51,33.28-46.44,47.55-17.03,14.18-35.39,26.53-54.36,37.43-19.06,10.8-38.88,19.76-58.94,27.65-10.03,4-20.23,7.31-30.39,10.74l-15.32,4.71c-5.03,1.57-10.71,3.08-16.05,4.29-21.66,5.07-43.48,8.01-65.46,8.76-21.96,.74-44.01-.54-65.72-4.41-43.33-7.44-85.59-25.6-118.6-54.31-16.49-14.26-30.84-30.66-42.15-48.7-2.71-4.6-5.55-9.06-8.08-13.7-2.4-4.72-4.98-9.3-7.18-14.06-2.05-4.85-4.2-9.62-6.12-14.46-1.81-4.9,.67-6.16,3.5-1.87,2.98,4.2,6.16,8.26,9.2,12.35,3.1,4.05,6.42,7.92,9.59,11.86,3.28,3.86,6.83,7.48,10.19,11.2,13.95,14.48,29.35,27.1,45.95,37.33,33.09,20.71,70.7,31.76,108.92,34.74,38.17,3.02,77.34-1.07,113.89-11.4,4.64-1.31,8.86-2.66,13.62-4.25l14.06-4.78c9.24-3.45,18.56-6.75,27.54-10.66,18.15-7.41,35.87-15.62,52.57-25.27,33.62-18.89,63.84-42.56,88.67-71.32,12.47-14.33,23.53-29.95,33.31-46.54,9.68-16.64,18.18-34.2,25.49-52.44,14.8-36.45,24.81-75.52,32.69-115.35,7.99-39.88,21.11-101.25,22.1-90.9Z",fill:"#9d744f"}),(0,h.jsx)("path",{d:"M733.6,626.97c-2.93,5.29-10.97,2.03-9.45-3.82,14.58-55.99,21.84-112.28,17.4-167.2-4.75-66.45-28.39-129.68-68.89-179.51-40.28-50.26-98.17-84.83-162.02-104.09-31.99-9.66-65.52-16.11-99.82-19.24-8.56-1.08-17.19-1.22-25.85-1.97-8.67-.88-17.35-.84-26.1-1.18-8.74-.25-17.55-.77-26.32-.41-8.81-.04-12.77-4.21-5.81-6.22,4.67-1.35,9.39-2.55,14.14-3.62,14.27-3.21,28.63-6.77,43.25-8.46,8.91-1.08,17.84-2.5,26.84-3.13,35.99-3.21,72.78-1.35,109.11,5.38,36.3,6.8,72.1,19.32,105.08,37.84,4.24,2.1,8.22,4.67,12.17,7.29l11.92,7.76c7.78,5.31,15.07,11.35,22.61,17.04,.11,.08,.22,.17,.32,.26,7.01,6.34,14.07,12.66,20.79,19.32l9.73,10.5c3.25,3.5,6.54,6.96,9.37,10.81,5.76,7.61,11.94,14.92,16.9,23.09l7.81,11.98c2.49,4.06,4.67,8.3,6.94,12.49,2.19,4.23,4.6,8.35,6.58,12.66l5.73,13.05c4.09,8.59,6.86,17.68,9.95,26.62,11.69,36,16.61,73.75,15.42,110.81-1.86,61.23-19.21,120.32-47.78,171.95Z",fill:"#00966f"})]})}const g={xs:"h-4 w-4 md:h-5 md:w-5",sm:"h-5 w-5 md:h-6 md:w-6",md:"h-6 w-6 md:h-7 md:w-7",lg:"h-7 w-7 md:h-8 md:w-8",xl:"h-8 w-8 md:h-9 md:w-9","2xl":"h-9 w-9 md:h-10 md:w-10"},y={xs:"text-xs leading-4 md:text-sm md:leading-5",sm:"text-sm leading-5 md:text-base md:leading-6",md:"text-base leading-6 md:text-lg md:leading-7",lg:"text-lg leading-7 md:text-xl md:leading-8",xl:"text-xl leading-8 md:text-2xl md:leading-9","2xl":"text-2xl leading-9 md:text-3xl md:leading-10"},b={vortico:"text-vortico-500",bosque:"text-bosque-500",bruma:"text-bruma-500",ciclon:"text-ciclon-500",flama:"text-flama-500",primary:"text-primary-500","vortico-light":"text-vortico-200","bosque-light":"text-bosque-200","bruma-light":"text-bruma-200","ciclon-light":"text-ciclon-200","flama-light":"text-flama-200","primary-light":"text-primary-200","vortico-dark":"text-vortico-800","bosque-dark":"text-bosque-800","bruma-dark":"text-bruma-800","ciclon-dark":"text-ciclon-800","flama-dark":"text-flama-800","primary-dark":"text-primary-800"},k={flama:{text:"Flama",icon:(0,h.jsx)(m,{})},vortico:{text:"Vortico",icon:(0,h.jsx)(v,{})}};function w(e){let{logo:t,color:n,size:r="md"}=e;const{text:a,icon:l}=k[t],o=n?b[n]:"";return(0,h.jsxs)("div",{className:"flex items-center justify-center gap-1","aria-label":`${a} logo`,children:[(0,h.jsx)("div",{className:`${g[r]} ${o}`,children:l}),(0,h.jsx)("span",{className:`text-brand ${y[r]} ${o}`,children:a})]})}function x(e){let{onClick:t,...n}=e;const r=[{href:"https://flama.dev/docs/",title:"Docs",icon:(0,h.jsx)(u,{}),label:"Flama docs"},{href:"https://flama.dev/blog/",title:"Blog",icon:(0,h.jsx)(s,{}),label:"Flama blog"}];return(0,h.jsx)("nav",{children:(0,h.jsx)("ul",{...n,children:r.map(((e,n)=>{let{href:r,title:a,icon:l,label:o}=e;return(0,h.jsx)("li",{onClick:t,children:(0,h.jsx)("a",{href:r,"aria-label":o,children:(0,h.jsxs)("div",{className:"flex h-12 cursor-pointer items-center justify-start gap-2",children:[(0,h.jsx)("div",{className:"text-primary-400 hover:text-flama-500 h-6 w-6 transition-colors duration-200 md:hidden",children:l}),(0,h.jsx)("div",{className:"text-primary-400 hover:text-flama-500 text-md font-medium transition-colors duration-200",children:a})]})})},n)}))})})}function S(e){let{...t}=e;const n=[{href:"https://github.com/vortico/flama",title:"GitHub",icon:(0,h.jsx)(c,{}),label:"Flama on GitHub"}];return(0,h.jsx)("div",{children:(0,h.jsx)("ul",{...t,children:n.map(((e,t)=>{let{href:n,title:r,icon:a,label:l}=e;return(0,h.jsx)("li",{children:(0,h.jsx)("a",{href:n,"aria-label":l,children:(0,h.jsxs)("div",{className:"flex h-12 cursor-pointer items-center justify-start gap-2",children:[(0,h.jsx)("div",{className:"text-primary-400 hover:text-flama-500 h-6 w-6 transition-colors duration-200",children:a}),(0,h.jsx)("div",{className:"text-primary-400 hover:text-flama-500 text-md font-medium transition-colors duration-200 md:hidden",children:r})]})})},t)}))})})}function E(e){let{onClose:t}=e;return(0,h.jsxs)("div",{className:"bg-primary-100 fixed inset-x-0 top-0 z-50 h-screen w-screen overflow-auto","aria-modal":"true",role:"dialog",children:[(0,h.jsxs)("div",{className:"border-flama-500 flex h-14 w-full items-center justify-between border-b px-4 sm:px-6 md:px-8",children:[(0,h.jsx)("a",{href:"https://flama.dev",className:"block cursor-pointer","aria-label":"Flama website",children:(0,h.jsx)(w,{logo:"flama",color:"flama",size:"lg"})}),(0,h.jsx)("button",{className:"h-6 w-6",onClick:t,"aria-label":"Close menu",children:(0,h.jsx)(f,{className:"text-primary-400 hover:text-flama-500 h-full w-full transition-colors duration-200"})})]}),(0,h.jsxs)("div",{className:"w-full px-4 sm:px-6 md:px-8",children:[(0,h.jsx)(x,{className:"flex flex-col items-start justify-start",onClick:t}),(0,h.jsx)(S,{className:"flex flex-col items-start justify-start gap-8"})]})]})}function C(e){let{onOpen:t}=e;return(0,h.jsx)(h.Fragment,{children:(0,h.jsxs)("div",{className:"flex h-14 w-full items-center justify-between",children:[(0,h.jsx)("a",{href:"https://flama.dev",className:"block cursor-pointer","aria-label":"Flama website",children:(0,h.jsx)(w,{logo:"flama",color:"flama",size:"lg"})}),(0,h.jsxs)("div",{className:"hidden items-center justify-between gap-8 md:flex",children:[(0,h.jsx)(x,{className:"flex items-center justify-start gap-8"}),(0,h.jsx)(S,{className:"flex items-center justify-start gap-8"})]}),(0,h.jsx)("button",{className:"block h-6 w-6 cursor-pointer md:hidden",onClick:t,"aria-label":"Open menu",children:(0,h.jsx)(d,{className:"h-full w-full"})})]})})}function N(){const[e,t]=(0,l.useState)(!1),n=(0,l.useCallback)((()=>t(!0)),[t]),r=(0,l.useCallback)((()=>t(!1)),[t]);return(0,h.jsxs)("div",{className:"border-flama-500 bg-primary-100 fixed inset-x-0 top-0 z-10 mx-auto border-b",children:[(0,h.jsx)("div",{className:"divide-flama-500/50 mx-auto w-full divide-y px-4 sm:px-6 md:px-8",children:(0,h.jsx)(C,{onOpen:n})}),e&&(0,p.createPortal)((0,h.jsx)(E,{onClose:r}),document.body)]})}function _(e){let{title:t,border:n=!0,gradient:r=!0,className:a="",children:l,...o}=e;return(0,h.jsxs)("section",{className:`my-14 ${n?"border-flama-500/50 border-t":""} ${a}`,...o,children:[(0,h.jsx)("div",{className:r?"from-flama-500/10 bg-linear-to-b":"",children:t&&(0,h.jsx)("div",{className:"max-w-8xl mx-auto px-4 py-8 sm:px-6 sm:py-10 md:px-8 md:py-12",children:(0,h.jsx)("h2",{className:"text-primary-700 text-2xl font-semibold",children:t})})}),(0,h.jsx)("div",{className:"max-w-8xl mx-auto px-4 sm:px-6 md:px-8",children:l})]})}function P(e){let{error:t,method:n,path:r,description:a}=e;return(0,h.jsxs)(h.Fragment,{children:[(0,h.jsxs)("h1",{className:"text-2xl",children:[(0,h.jsx)("span",{className:"text-flama-500 font-bold",children:t}),(0,h.jsx)("span",{className:"pl-2",children:"raised at"}),(0,h.jsx)("span",{className:"pl-2 font-bold",children:n}),(0,h.jsx)("span",{className:"pl-2 font-mono",children:r})]}),a&&(0,h.jsx)("h2",{className:"text-xl font-medium",children:a})]})}function z(e,t){(null==t||t>e.length)&&(t=e.length);for(var n=0,r=Array(t);n<t;n++)r[n]=e[n];return r}function j(e){return function(e){if(Array.isArray(e))return z(e)}(e)||function(e){if("undefined"!=typeof Symbol&&null!=e[Symbol.iterator]||null!=e["@@iterator"])return Array.from(e)}(e)||function(e,t){if(e){if("string"==typeof e)return z(e,t);var n={}.toString.call(e).slice(8,-1);return"Object"===n&&e.constructor&&(n=e.constructor.name),"Map"===n||"Set"===n?Array.from(e):"Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)?z(e,t):void 0}}(e)||function(){throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}()}var T=i("filled","circle-filled","IconCircleFilled",[["path",{d:"M7 3.34a10 10 0 1 1 -4.995 8.984l-.005 -.324l.005 -.324a10 10 0 0 1 4.995 -8.336z",key:"svg-0"}]]),O=i("outline","chevron-right","IconChevronRight",[["path",{d:"M9 6l6 6l-6 6",key:"svg-0"}]]);class L{constructor(){this.name="||@ app.name @||",this.path="||@ app.path @||",this.apps=JSON.parse("||@ app.apps|safe_json @||"),this.endpoints=JSON.parse("||@ app.endpoints|safe_json @||")}}function M(e){let{data:t}=e;return(0,h.jsx)("div",{className:"bg-primary-800 text-primary-200 after:border-b-primary-800 absolute -bottom-2 left-1/2 z-20 hidden -translate-x-1/2 translate-y-full rounded-lg p-3 text-left text-sm group-hover:block after:absolute after:bottom-[100%] after:left-1/2 after:-translate-x-1/2 after:border-8 after:border-x-transparent after:border-t-transparent after:content-['']",children:(0,h.jsx)("ul",{children:Array.from(t.entries()).sort(((e,t)=>e[0].localeCompare(t[0]))).filter((e=>e[1])).map(((e,t)=>{let[n,r]=e;return(0,h.jsxs)("li",{children:[(0,h.jsxs)("span",{className:"text-flama-400 capitalize",children:[n,":"]})," ",(0,h.jsx)("span",{children:r})]},t)}))})})}function A(e){let{endpoint:t}=e;const{endpoint:n,file:r,line:a,module:l,name:o,path:i}=t;return(0,h.jsxs)("div",{className:"group border-primary-300 bg-primary-100 shadow-primary-300 relative -mt-px border leading-8 shadow-md",children:[(0,h.jsxs)("div",{className:"hover:border-flama-500 hover:bg-flama-300 -ml-px flex cursor-pointer border-l-4 border-transparent transition-colors duration-200",children:[(0,h.jsx)("div",{className:"text-primary-500 flex h-8 w-8 items-center justify-center",children:(0,h.jsx)(T,{className:"h-2 w-2"})}),(0,h.jsx)("div",{children:i})]}),(0,h.jsx)(M,{data:new Map(Object.entries({endpoint:n,file:r,line:a,module:l,name:o,path:i}))})]})}function R(e){let{app:t}=e;const{name:n,path:r,apps:a,endpoints:o}=t,[i,u]=(0,l.useState)(!1),s=(0,l.useRef)(null),c=(0,l.useCallback)(((e,t)=>t.length+e.reduce(((e,t)=>e+1+c(t.apps,t.endpoints)),0)),[]),f=(0,l.useMemo)((()=>c(a,o)),[c,a,o]);return(0,l.useEffect)((()=>{s.current&&(s.current.style.maxHeight=(i?34*f:0)+"px")}),[i,s,f]),(0,h.jsxs)("div",{children:[(0,h.jsxs)("div",{className:"group border-primary-300 bg-primary-200 shadow-primary-300 relative -mt-px border leading-8 shadow-md",children:[(0,h.jsxs)("div",{className:"hover:border-flama-500 hover:bg-flama-300 -ml-px flex cursor-pointer items-center border-l-4 border-transparent transition-colors duration-200",onClick:()=>u(!i),children:[(0,h.jsx)("div",{className:"flex h-8 w-8 items-center justify-center transition-transform "+(i?"rotate-90":""),children:(0,h.jsx)(O,{className:"h-4 w-4"})}),(0,h.jsx)("div",{children:r})]}),(0,h.jsx)(M,{data:new Map(Object.entries({name:n,path:r}))})]}),(0,h.jsx)("div",{ref:s,className:"ml-1 max-h-0 pl-4 transition-all "+(i?"overflow-visible opacity-100":"overflow-hidden opacity-0"),children:(0,h.jsx)(F,{apps:a,endpoints:o})})]})}function F(e){let{apps:t,endpoints:n}=e;const r=[].concat(j(t.map((e=>({key:e.path,node:(0,h.jsx)("li",{children:(0,h.jsx)(R,{app:e},e.path)},e.path)})))),j(n.map((e=>({key:e.path,node:(0,h.jsx)("li",{children:(0,h.jsx)(A,{endpoint:e},e.path)},e.path)}))))).sort(((e,t)=>e.key.localeCompare(t.key)));return(0,h.jsx)("ul",{children:r.map((e=>{let{node:t}=e;return t}))})}function D(){const{apps:e,endpoints:t}=new L;return(0,h.jsx)("div",{children:(0,h.jsx)(F,{apps:e,endpoints:t})})}function I(e){let{children:t}=e;return(0,h.jsx)("table",{className:"border-primary-300 w-full table-fixed border-collapse overflow-hidden border",children:(0,h.jsx)("tbody",{className:"text-left text-wrap",children:t})})}function U(e){let{header:t,children:n,...r}=e;return(0,h.jsxs)("tr",{...r,children:[(0,h.jsx)("th",{className:"border-primary-300 w-40 border px-4 py-1",children:(0,h.jsx)("span",{className:"text-flama",children:t})}),(0,h.jsx)("td",{className:"border-primary-300 border px-4 py-1",children:n})]})}const $=function(e){return(new DOMParser).parseFromString(e,"text/html").body.textContent||""};function H(e){let{value:t}=e;return(0,h.jsx)("span",{children:$(String(t))})}function V(e){let{map:t}=e;return(0,h.jsx)("ul",{children:Array.from(t).map(((e,t)=>{let[n,r]=e;return(0,h.jsxs)("li",{children:[(0,h.jsx)("span",{className:"font-semibold",children:`${n}:`}),(0,h.jsx)("span",{className:"ml-1",children:(0,h.jsx)(H,{value:r})})]},t)}))})}function B(){const{path:e,method:t,pathParams:n,queryParams:r,headers:l,cookies:o,clientHost:i,clientPort:u}=new a;return(0,h.jsxs)(I,{children:[(0,h.jsx)(U,{header:"Path",children:(0,h.jsx)(H,{value:e})}),(0,h.jsx)(U,{header:"Method",children:(0,h.jsx)(H,{value:t})}),(0,h.jsx)(U,{header:"Query parameters",children:(0,h.jsx)(V,{map:r})}),(0,h.jsx)(U,{header:"Path parameters",children:(0,h.jsx)(V,{map:n})}),(0,h.jsx)(U,{header:"Headers",children:(0,h.jsx)(V,{map:l})}),(0,h.jsx)(U,{header:"Cookies",children:(0,h.jsx)(V,{map:o})}),i&&(0,h.jsx)(U,{header:"Client host",children:(0,h.jsx)(H,{value:i})}),u&&(0,h.jsx)(U,{header:"Client port",children:(0,h.jsx)(H,{value:String(u)})})]})}class q{constructor(){this.pythonVersion="||@ environment.python_version|safe @||",this.python="||@ environment.python|safe @||",this.platform="||@ environment.platform|safe @||",this.path=JSON.parse("||@ environment.path|safe_json @||")}}function W(e){let{array:t}=e;return(0,h.jsx)("ul",{children:t.map(((e,t)=>(0,h.jsx)("li",{children:(0,h.jsx)(H,{value:e})},t)))})}function Q(){const{pythonVersion:e,python:t,platform:n,path:r}=new q;return(0,h.jsxs)(I,{children:[(0,h.jsx)(U,{header:"Python version",children:(0,h.jsx)(H,{value:e})}),(0,h.jsx)(U,{header:"Python",children:(0,h.jsx)(H,{value:t})}),(0,h.jsx)(U,{header:"Platform",children:(0,h.jsx)(H,{value:n})}),(0,h.jsx)(U,{header:"Path",children:(0,h.jsx)(W,{array:r})})]})}var K=i("outline","brand-linkedin","IconBrandLinkedin",[["path",{d:"M8 11v5",key:"svg-0"}],["path",{d:"M8 8v.01",key:"svg-1"}],["path",{d:"M12 16v-5",key:"svg-2"}],["path",{d:"M16 16v-3a2 2 0 1 0 -4 0",key:"svg-3"}],["path",{d:"M3 7a4 4 0 0 1 4 -4h10a4 4 0 0 1 4 4v10a4 4 0 0 1 -4 4h-10a4 4 0 0 1 -4 -4z",key:"svg-4"}]]),G=i("outline","brand-x","IconBrandX",[["path",{d:"M4 4l11.733 16h4.267l-11.733 -16z",key:"svg-0"}],["path",{d:"M4 20l6.768 -6.768m2.46 -2.46l6.772 -6.772",key:"svg-1"}]]),Y=i("outline","brand-medium","IconBrandMedium",[["path",{d:"M4 4m0 2a2 2 0 0 1 2 -2h12a2 2 0 0 1 2 2v12a2 2 0 0 1 -2 2h-12a2 2 0 0 1 -2 -2z",key:"svg-0"}],["path",{d:"M8 9h1l3 3l3 -3h1",key:"svg-1"}],["path",{d:"M8 15l2 0",key:"svg-2"}],["path",{d:"M14 15l2 0",key:"svg-3"}],["path",{d:"M9 9l0 6",key:"svg-4"}],["path",{d:"M15 9l0 6",key:"svg-5"}]]);const X=[{url:"https://github.com/vortico/",icon:(0,h.jsx)(c,{})},{url:"https://www.linkedin.com/company/vortico-tech/",icon:(0,h.jsx)(K,{})},{url:"https://twitter.com/vorticotech",icon:(0,h.jsx)(G,{})},{url:"https://dev.to/vortico",icon:(0,h.jsx)(Y,{})}];function Z(){return(0,h.jsx)("footer",{children:(0,h.jsx)("div",{className:"border-vortico-500/50 border-t px-4 py-2 sm:px-6 md:px-8",children:(0,h.jsxs)("div",{className:"mx-auto flex max-w-7xl items-center justify-between",children:[(0,h.jsx)("a",{href:"https://vortico.tech",children:(0,h.jsx)(w,{logo:"vortico",color:"vortico",size:"lg"})}),(0,h.jsx)("div",{className:"divide-vortico-500/50 flex h-8 items-center justify-end divide-x",children:X.map(((e,t)=>(0,h.jsx)("a",{href:e.url,className:"px-2 first:pl-0 last:pr-0",children:(0,h.jsx)("div",{className:"text-primary-400 hover:text-vortico-500 h-6 w-6 transition-colors duration-200",children:e.icon})},t)))})]})})})}function J(){const e=new a;return(0,h.jsxs)("div",{children:[(0,h.jsx)(N,{}),(0,h.jsxs)("div",{className:"mt-[calc(3.5rem+1px)] h-[calc(100vh-6.5rem-2px)] overflow-auto",children:[(0,h.jsx)("header",{children:(0,h.jsx)(_,{id:"error",border:!1,children:(0,h.jsx)(P,{error:"Not Found",path:e.path,method:e.method})})}),(0,h.jsxs)("main",{children:[(0,h.jsx)(_,{id:"url-tree",title:"Application URLs",children:(0,h.jsx)(D,{})}),(0,h.jsx)(_,{id:"request",title:"Request",children:(0,h.jsx)(B,{})}),(0,h.jsx)(_,{id:"environment",title:"Environment",children:(0,h.jsx)(Q,{})})]})]}),(0,h.jsx)(Z,{})]})}r.createRoot(document.getElementById("app")).render((0,h.jsx)(J,{}))})();</script></body></html>