<!doctype html><html lang="en"><head><meta charset="UTF-8"/><meta name="viewport" content="width=device-width,initial-scale=1"/><link rel="preconnect" href="https://fonts.googleapis.com"/><link rel="preconnect" href="https://fonts.gstatic.com" crossorigin/><link href="https://fonts.googleapis.com/css2?family=Lato:ital,wght@0,100;0,300;0,400;0,700;0,900;1,100;1,300;1,400;1,700;1,900&family=Montserrat:ital,wght@0,100..900;1,100..900&display=swap" rel="stylesheet"/><title>Docs - Flama</title><style>:root {
  --scalar-border-width: 1px;

  --scalar-radius: 0px;
  --scalar-radius-lg: 0px;
  --scalar-radius-xl: 0px;

  --scalar-text-decoration: none;
  --scalar-text-decoration-hover: none;

  --scalar-font:
    'Lato', ui-sans-serif, system-ui, sans-serif, 'Apple Color Emoji', 'Segoe UI Emoji', 'Segoe UI Symbol',
    'Noto Color Emoji';
  --scalar-font-code:
    'Fira Mono', ui-monospace, SFMono-Regular, Menlo, Monaco, Consolas, 'Liberation Mono', 'Courier New', monospace;
}

.scalar-app:not(#headlessui-portal-root) {
  margin: calc(3.5rem + 1px) 0 0 0;
  height: calc(100vh - 6.5rem - 2px);
  min-height: calc(100vh - 6.5rem - 2px);
  max-height: calc(100vh - 6.5rem - 2px);
  overflow: auto;
}

:where(.t-doc__sidebar, .references-navigation-list) {
  height: calc(100vh - 6.5rem - 2px) !important;
  width: 18rem;
  min-height: calc(100vh - 6.5rem - 2px);
  max-height: calc(100vh - 6.5rem - 2px);
  overflow: auto !important;
  position: fixed !important;
  left: 0 !important;
  top: calc(3.5rem + 1px) !important;
  z-index: 10;
}

.dark-mode {
  color-scheme: dark;
}

.light-mode {
  color-scheme: light;
}

.light-mode,
.dark-mode {
  --scalar-scrollbar-color: #808290;
  --scalar-scrollbar-color-active: #808290;

  --scalar-shadow-1: 0 1px 3px 0 #aaabb6;
  --scalar-shadow-2: #aaabb6 0px 13px 20px 0px, #aaabb6 0px 3px 8px 0px, #aaabb6 0px 0 0 0.5px;

  /* Basic theme */
  --scalar-background-1: #e5e6e8;
  --scalar-background-2: #f5f5f6;
  --scalar-background-3: #cdced4;
  --scalar-background-accent: #f0ac81;

  --scalar-color-1: #575863;
  --scalar-color-2: #656775;
  --scalar-color-3: #4a4a54;

  --scalar-color-accent: #e25822;
  --scalar-border-color: #aaabb6;

  /* Advanced theme */
  --scalar-color-green: #00976e;
  --scalar-color-red: #9e744f;
  --scalar-color-yellow: #f0ac81;
  --scalar-color-blue: #00bbd5;
  --scalar-color-orange: #e25822;
  --scalar-color-purple: #008080;

  --scalar-button-1: var(--scalar-background-2);
  --scalar-button-1-hover: var(--scalar-background-accent);
  --scalar-button-1-color: var(--scalar-color-1);

  /* Document Sidebar */
  .t-doc__sidebar {
    --scalar-sidebar-background-1: var(--scalar-background-1);
    --scalar-sidebar-color-1: var(--scalar-color-1);
    --scalar-sidebar-color-2: var(--scalar-color-3);
    --scalar-sidebar-border-color: var(--scalar-border-color);

    --scalar-sidebar-item-hover-background: var(--scalar-background-accent);
    --scalar-sidebar-item-hover-color: var(--scalar-color-3);

    --scalar-sidebar-item-active-background: var(--scalar-background-3);
    --scalar-sidebar-color-active: var(--scalar-color-3);

    --scalar-sidebar-indent-border: var(--scalar-sidebar-border-color);
    --scalar-sidebar-indent-border-hover: var(--scalar-color-accent);
    --scalar-sidebar-indent-border-active: var(--scalar-color-3);

    --scalar-sidebar-search-background: var(--scalar-background-2);
    --scalar-sidebar-search-color: var(--scalar-color-2);
    --scalar-sidebar-search-border-color: var(--scalar-border-color);
  }
}

*::selection {
  background-color: color-mix(in srgb, var(--scalar-color-accent), transparent 75%);
}

* {
  border-radius: 0 !important;
}
.start[data-v-391b4e37]{padding:24px;display:flex;flex-flow:wrap;justify-content:space-between;position:relative;z-index:0}.swagger-editor .start[data-v-391b4e37]{padding-top:24px}.start-h1[data-v-391b4e37]{font-size:var(--scalar-heading-2);margin-top:0;line-height:1.45;margin-bottom:0;font-weight:var(--scalar-bold);color:var(--scalar-color-1);width:100%;position:relative}.start-h3[data-v-391b4e37]{font-size:var(--scalar-paragraph);margin-top:0;margin-bottom:6px;display:block;line-height:1.45;font-weight:var(--scalar-bold);color:var(--scalar-color-1);width:100%}.start-h1[data-v-391b4e37]:not(:first-of-type){margin-top:24px}.start-p[data-v-391b4e37]{font-size:var(--scalar-paragraph);color:var(--scalar-color-2);line-height:1.5;width:100%;margin-top:12px}.start-ul[data-v-391b4e37]{margin-top:12px;font-size:var(--scalar-paragraph);line-height:1.5;padding-left:0;list-style:initial;display:flex;flex-flow:wrap;gap:24px}.start-ul li[data-v-391b4e37]{margin:0;padding:0;list-style:none;width:calc(50% - 24px);color:var(--scalar-color-2)}.start-ul li[data-v-391b4e37]:first-of-type{margin-top:0}.start-section[data-v-391b4e37]{width:100%;margin-bottom:12px;display:flex;flex-flow:wrap}.start-section[data-v-391b4e37]:last-of-type{margin-bottom:48px}.start-h2[data-v-391b4e37]{background:var(--scalar-background-2);border-top-left-radius:var(--scalar-radius-lg);border-top-right-radius:var(--scalar-radius-lg);border:1px solid var(--scalar-border-color);color:var(--scalar-color-3);font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);padding:9px;width:100%}.start-item[data-v-391b4e37]{align-items:center;background:var(--scalar-background-2);border-right:1px solid var(--scalar-border-color);border-bottom:1px solid var(--scalar-border-color);color:var(--scalar-color-1);cursor:pointer;display:flex;flex:1;font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);padding:9px;text-transform:capitalize;-webkit-user-select:none;-moz-user-select:none;user-select:none}.start-section-integrations .start-item[data-v-391b4e37]:first-of-type{border-bottom-left-radius:var(--scalar-radius-lg);border-left:1px solid var(--scalar-border-color)}.start-section-integrations .start-item[data-v-391b4e37]:last-of-type{border-bottom-right-radius:var(--scalar-radius-lg)}.start-section-colors .start-item[data-v-391b4e37]{min-width:33.33%}.start-section-colors .start-item[data-v-391b4e37]:nth-child(3n+2){border-left:1px solid var(--scalar-border-color)}.start-section-colors .start-item[data-v-391b4e37]:last-of-type,.start-section-colors .start-item-active[data-v-391b4e37]:last-of-type:before{border-radius:0 0 var(--scalar-radius-lg) var(--scalar-radius-lg)}.start-item[data-v-391b4e37]:empty{pointer-events:none}.start-item svg[data-v-391b4e37]{width:14px;height:14px;margin-right:6px}.start-item[data-v-391b4e37]:hover{background:var(--scalar-background-3)}.start-item-active[data-v-391b4e37]{z-index:10;position:relative;color:var(--scalar-color-1)}.start-item-active[data-v-391b4e37]:before{border:1px solid var(--scalar-color-1);content:"";top:-1px;right:-1px;bottom:-1px;left:-1px;pointer-events:none;position:absolute}.start-section-color .start-item[data-v-391b4e37]{text-transform:capitalize}.start-cta[data-v-391b4e37]{display:flex;gap:12px;width:100%;margin-top:24px;margin-bottom:0}.start-row[data-v-391b4e37]{width:100%;margin-top:12px;overflow:hidden}.start-hero-copy[data-v-391b4e37]{background:var(--scalar-background-2);padding:12px;border-radius:var(--scalar-radius-lg)}.start-p-small[data-v-391b4e37]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-2);margin-bottom:12px;line-height:1.4}.start-cta[data-v-391b4e37]{margin-bottom:12px;width:-moz-fit-content;width:fit-content;white-space:nowrap}.start-copy[data-v-391b4e37]{padding:76px 48px 48px;display:flex;justify-content:center;align-items:center;flex-direction:column;text-align:center}.start-logo[data-v-391b4e37]{color:var(--scalar-color-1);margin-bottom:24px;width:72px;aspect-ratio:1;position:relative;box-shadow:var(--scalar-shadow-2);border-radius:50%}.start-logo[data-v-391b4e37]:before{content:"";width:300%;aspect-ratio:1;left:-100%;top:-100%;position:absolute;z-index:-1;border-radius:50%;background-size:24px 24px;box-shadow:inset 0 0 50px var(--scalar-background-1),inset 0 0 50px var(--scalar-background-1);background-image:linear-gradient(to right,var(--scalar-border-color) 1px,transparent 1px),linear-gradient(to bottom,var(--scalar-border-color) 1px,transparent 1px)}.start-logo svg[data-v-391b4e37]{width:100%;height:auto;background:var(--scalar-background-1);padding:3px;border-radius:50%;position:relative}@media screen and (max-width: 600px){.start-section-colors .start-item[data-v-391b4e37],.start-item[data-v-391b4e37]{width:100%;border-radius:0;border-right:none;border-top:1px solid var(--scalar-border-color)}.start-item[data-v-391b4e37]:empty{display:none}.start-h2[data-v-391b4e37]{border-bottom:none}.start li[data-v-391b4e37]{width:100%}.start-copy[data-v-391b4e37]{padding:48px 0 24px}}@media screen and (max-width: 1000px){.start[data-v-391b4e37]{padding:0;overflow:auto}}.sidebar-heading-type[data-v-f51e2805]{display:block;min-width:3.9em;overflow:hidden;line-height:14px;flex-shrink:0;color:#fff;color:color-mix(in srgb,var(--method-color, var(--scalar-color-1)),transparent 0%);text-transform:uppercase;font-size:10px;font-weight:var(--scalar-bold);text-align:right;position:relative;font-family:var(--scalar-font-code);white-space:nowrap;margin-left:3px}.sidebar-heading[data-v-18e6dc3d]{display:flex;gap:6px;color:var(--scalar-sidebar-color-2, var(--scalar-color-2));font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);word-break:break-word;line-height:1.385;max-width:100%;position:relative;cursor:pointer;border-radius:var(--scalar-radius);flex:1;padding-right:9px;-webkit-user-select:none;-moz-user-select:none;user-select:none}.sidebar-heading-link-method[data-v-18e6dc3d]{margin:0}.sidebar-heading.deprecated .sidebar-heading-link-title[data-v-18e6dc3d]{text-decoration:line-through}.sidebar-heading-link-title[data-v-18e6dc3d]{margin:0}.sidebar-heading[data-v-18e6dc3d]:hover{background:var( --scalar-sidebar-item-hover-background, var(--scalar-background-2) )}.sidebar-heading:hover .sidebar-heading-link-title[data-v-18e6dc3d]{color:var(--scalar-sidebar-item-hover-color)}.sidebar-heading-link[data-v-18e6dc3d]:focus-visible{outline:none}.sidebar-heading[data-v-18e6dc3d]:has(>.sidebar-heading-link:focus-visible){z-index:1;outline:1px solid var(--scalar-color-accent)}.active_page.sidebar-heading[data-v-18e6dc3d]:hover,.active_page.sidebar-heading[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-active, var(--scalar-color-accent));background:var( --scalar-sidebar-item-active-background, var(--scalar-background-accent) )}.active_page.sidebar-heading p[data-v-18e6dc3d]{font-weight:var(--scalar-sidebar-font-weight-active, var(--scalar-semibold))}.active_page.sidebar-heading:hover .sidebar-heading-link-title[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-active, var(--scalar-color-accent))}.sidebar-indent-nested .sidebar-indent-nested .sidebar-heading[data-v-18e6dc3d]:before{content:"";position:absolute;top:0;left:calc((var(--scalar-sidebar-level) * 12px));width:1px;height:100%;background:var(--scalar-sidebar-indent-border)}.sidebar-indent-nested .sidebar-indent-nested .sidebar-heading[data-v-18e6dc3d]:hover:before{background:var(--scalar-sidebar-indent-border-hover)}.sidebar-indent-nested .sidebar-indent-nested .active_page.sidebar-heading[data-v-18e6dc3d]:before{background:var(--scalar-sidebar-indent-border-active)}.sidebar-heading-link[data-v-18e6dc3d]{text-decoration:none;color:inherit;padding:6px 0;display:flex;flex:1;justify-content:space-between;gap:2px}.sidebar-heading p[data-v-18e6dc3d]{height:-moz-fit-content;height:fit-content;display:flex;align-items:center;font-weight:var(--scalar-sidebar-font-weight, var(--scalar-semibold))}.sidebar-heading p[data-v-18e6dc3d]:empty{display:none}.link-icon[data-v-18e6dc3d]{position:relative;left:4px}.sidebar-icon[data-v-18e6dc3d]{display:flex;align-items:center;justify-content:center;margin-right:6px;width:13px;height:13px}.sidebar-icon>svg[data-v-18e6dc3d]{width:13px;height:13px}.sidebar-group-item[data-v-18e6dc3d]{position:relative}.sidebar-heading-chevron[data-v-18e6dc3d]{margin:5px -5.5px 5px -9px}.sidebar-heading-chevron .toggle-nested-icon[data-v-18e6dc3d]:focus-visible{outline:none}.sidebar-heading[data-v-18e6dc3d]:has(.sidebar-heading-chevron .toggle-nested-icon:focus-visible){outline:none;box-shadow:inset 0 0 0 1px var(--scalar-color-accent)}.toggle-nested-icon[data-v-18e6dc3d]{color:var(--scalar-color-3);width:20px;height:20px;display:flex;align-items:center;justify-content:center}.active_page .toggle-nested-icon[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-active, var(--scalar-color-accent))}.toggle-nested-icon[data-v-18e6dc3d]:hover,.toggle-nested-icon[data-v-18e6dc3d]:focus-visible{color:currentColor}.action-menu[data-v-18e6dc3d]{position:absolute;top:5px;right:5px;display:flex;gap:6px}.action-menu[data-v-18e6dc3d] .button-wrapper button{opacity:0;width:20px;height:20px;padding:4px}.action-menu[data-v-18e6dc3d]:hover .button-wrapper button,.action-menu[data-v-18e6dc3d] .button-wrapper button:hover,.sidebar-heading:hover~.action-menu[data-v-18e6dc3d] .button-wrapper button,.action-menu[data-v-18e6dc3d] .button-wrapper button[aria-expanded=true]{opacity:1}.sidebar-heading[data-v-18e6dc3d]:has(~.action-menu:hover){color:var(--scalar-sidebar-color-1, var(--scalar-color-1));background:var( --scalar-sidebar-item-hover-background, var(--scalar-background-2) )}.sidebar-group-item__folder[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-1, var(--scalar-color-1));text-transform:var(--scalar-tag-text-transform, initial)}.sidebar-group[data-v-39c84840]{list-style:none;width:100%;margin:0;padding:0}.sidebar-indent-nested[data-v-39c84840] .sidebar-heading{padding-left:calc((var(--scalar-sidebar-level) * var(--scalar-sidebar-indent-base)) + 12px)!important}.sidebar-indent-nested[data-v-39c84840] .sidebar-heading .toggle-nested-icon{left:calc((var(--scalar-sidebar-level) * var(--scalar-sidebar-indent-base)) + 2px)!important}:where(.sidebar-indent-nested[data-v-39c84840]) .sidebar-heading{color:var(--scalar-sidebar-color-1, var(--scalar-color-1))}:where(.sidebar-indent-nested[data-v-39c84840]) :where(.sidebar-indent-nested) .sidebar-heading{color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}.sidebar[data-v-b863366c]{--scalar-sidebar-indent-base: 12px;--scalar-sidebar-font-weight-active: var(--scalar-semibold);--scalar-sidebar-font-weight: var(--scalar-semibold);flex:1;height:100%;display:flex;flex-direction:column;border-right:var(--scalar-border-width) solid var(--scalar-sidebar-border-color, var(--scalar-border-color));background:var(--scalar-sidebar-background-1, var(--scalar-background-1));--scalar-sidebar-level: 0}.sidebar-pages[data-v-b863366c]{flex:1;padding:9px 12px}@media (max-width: 1000px){.sidebar[data-v-b863366c]{min-height:0;border-right:none}.sidebar-pages[data-v-b863366c]{padding-top:12px}}.sidebar-group-title[data-v-b863366c]{color:var(--scalar-sidebar-color-1);font-size:var(--scalar-mini);padding:12px 6px 6px;font-weight:var(--scalar-semibold);text-transform:uppercase;word-break:break-word;line-height:1.385}.sidebar-group-item+.sidebar-group-title[data-v-b863366c]{border-top:var(--scalar-border-width) solid var(--scalar-sidebar-border-color);margin-top:9px}.show-more[data-v-252cd03d]{background:var(--scalar-background-1);-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none;border:var(--scalar-border-width) solid var(--scalar-border-color);margin:auto;padding:8px 12px;border-radius:30px;color:var(--scalar-color-1);font-weight:var(--scalar-semibold);font-size:var(--scalar-small);display:flex;align-items:center;justify-content:center;position:relative;top:-48px}.show-more[data-v-252cd03d]:hover{color:var(--scalar-color-2);cursor:pointer}.show-more-icon[data-v-252cd03d]{width:16px!important;height:16px!important;margin-left:3px}.show-more[data-v-252cd03d]:active{box-shadow:0 0 0 1px var(--scalar-border-color)}@container narrow-references-container (max-width: 900px){.show-more[data-v-252cd03d]{top:-24px}}.screenreader-only[data-v-df2e1026]{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);border:0}.section[data-v-64ccd510]{position:relative;display:flex;flex-direction:column;max-width:var(--refs-content-max-width);margin:auto;padding:90px 0;scroll-margin-top:var(--refs-header-height)}.section[data-v-64ccd510]:has(~div.contents){border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.references-classic .section[data-v-64ccd510]{padding:48px 0;gap:24px}@container narrow-references-container (max-width: 900px){.references-classic .section[data-v-64ccd510],.section[data-v-64ccd510]{padding:48px 24px}}.section[data-v-64ccd510]:not(:last-of-type){border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.loading[data-v-c90b2c46]{background:var(--scalar-background-3);animation:loading-skeleton-c90b2c46 1.5s infinite alternate;border-radius:var(--scalar-radius-lg);min-height:1.6em;margin:.6em 0;max-width:100%}.loading[data-v-c90b2c46]:first-of-type{min-height:3em;margin-bottom:24px;margin-top:0}.loading[data-v-c90b2c46]:last-of-type{width:60%}.loading.single-line[data-v-c90b2c46]{min-height:3em;margin:.6em 0;max-width:80%}@keyframes loading-skeleton-c90b2c46{0%{opacity:1}to{opacity:.33}}.section-header-wrapper[data-v-b6aef53d]{display:grid;grid-template-columns:1fr}@media (min-width: 1200px){.section-header-wrapper[data-v-b6aef53d]{grid-template-columns:repeat(2,1fr)}}.section-header[data-v-b6aef53d]{font-size:var(--font-size, var(--scalar-heading-2));font-weight:var(--font-weight, var(--scalar-bold));color:var(--scalar-color-1);word-wrap:break-word;line-height:1.45;margin-top:0;margin-bottom:12px}.section-header.tight[data-v-b6aef53d]{margin-bottom:6px}.section-header.loading[data-v-b6aef53d]{width:80%}.label[data-v-3cb2dd80]{position:relative;display:inline-block;word-break:break-all}.anchor[data-v-3cb2dd80]{position:relative;display:inline-block;opacity:0}.anchor-copy[data-v-3cb2dd80]{position:absolute;left:0;top:50%;transform:translateY(-50%);cursor:pointer;padding:0 6px;color:var(--scalar-color-3);font-weight:var(--scalar-semibold);font-size:.8em}.anchor-copy[data-v-3cb2dd80]:hover,.anchor-copy[data-v-3cb2dd80]:focus-visible{color:var(--scalar-color-2)}.label:hover .anchor[data-v-3cb2dd80],.label:has(:focus-visible) .anchor[data-v-3cb2dd80]{opacity:1}@container narrow-references-container (max-width: 900px){.section-content--with-columns[data-v-9735459e]{flex-direction:column;gap:24px}}.section-columns[data-v-8b9602bf]{display:flex;gap:48px}@container narrow-references-container (max-width: 900px){.section-columns[data-v-8b9602bf]{flex-direction:column;gap:24px}}.section-column[data-v-699c28e3]{flex:1;min-width:0}@container narrow-references-container (max-width: 900px){.section-column[data-v-699c28e3]:nth-of-type(2){padding-top:0}}.scalar-card[data-v-38801681]{all:unset;font-family:var(--scalar-font);border-radius:var(--scalar-radius-lg);overflow:hidden;border:var(--scalar-border-width) solid var(--scalar-border-color);background:var(--scalar-background-2);display:flex;flex-direction:column;max-height:calc(((var(--full-height) - var(--refs-header-height)) - 60px) / 2);position:relative}.scalar-card-sticky[data-v-38801681]{position:sticky;top:calc(var(--refs-header-height) + 24px)}.scalar-card-content[data-v-dd83e46e]{overflow:auto;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color);display:grid}.scalar-card-content[data-v-dd83e46e] .simple-table .simple-header{display:none}.scalar-card-content[data-v-dd83e46e]:last-of-type,.scalar-card-content.scalar-card--borderless[data-v-dd83e46e]{border-bottom:none}.scalar-card--muted[data-v-dd83e46e]{background:var(--scalar-background-2)}.scalar-card--contrast[data-v-dd83e46e]{background:var(--scalar-background-3)}.scalar-card--frameless[data-v-dd83e46e]{padding:0}.scalar-card--transparent[data-v-dd83e46e]{background:var(--scalar-background-1)}.scalar-card-header[data-v-838dc259]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-3);padding:9px 3px 9px 12px;flex-shrink:0}.scalar-card-header.scalar-card--borderless[data-v-838dc259]+.scalar-card-content{margin-top:-9px}.scalar-card-header-slots[data-v-838dc259]{display:flex;justify-content:space-between;line-height:1.35}.scalar-card-header-title[data-v-838dc259]{text-transform:uppercase;flex:1;min-width:0;text-overflow:ellipsis;overflow:hidden}.scalar-card-header-actions[data-v-838dc259]{display:flex}.endpoints[data-v-cfef28a1]{overflow:auto;background:var(--scalar-background-2);padding:10px 12px}@media (max-width: 580px){.endpoints[data-v-cfef28a1]{max-height:calc(100vh - 150px)}}.endpoints span+span[data-v-cfef28a1]{text-align:left;margin-left:12px;text-transform:initial}.endpoint[data-v-cfef28a1]{display:flex;white-space:nowrap;cursor:pointer;text-decoration:none}.endpoint:hover .endpoint-path[data-v-cfef28a1],.endpoint:focus-visible .endpoint-path[data-v-cfef28a1]{text-decoration:underline}.endpoint span[data-v-cfef28a1]:first-of-type{text-transform:uppercase}.endpoint .post[data-v-cfef28a1],.endpoint .get[data-v-cfef28a1],.endpoint .delete[data-v-cfef28a1],.endpoint .put[data-v-cfef28a1]{white-space:nowrap}.endpoint span[data-v-cfef28a1]{color:var(--scalar-color-1);min-width:62px;display:inline-block;text-align:right;line-height:1.55;font-family:var(--scalar-font-code);font-size:var(--scalar-small);cursor:pointer}.deprecated[data-v-cfef28a1]{text-decoration:line-through}.section-container[data-v-1913755a]{position:relative;padding:0 60px;width:100%;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.section-container[data-v-1913755a]:has(.introduction-section){border-top:none}@container narrow-references-container (max-width: 900px){.section-container[data-v-1913755a]{padding:0}}.section-container[data-v-d47a803a]{border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.section-container[data-v-d47a803a]:has(.show-more){background-color:color-mix(in srgb,var(--scalar-background-2),transparent)}.operation-path[data-v-ec6c8861]{overflow:hidden;word-wrap:break-word;font-weight:var(--scalar-semibold);line-break:anywhere}.deprecated[data-v-ec6c8861]{text-decoration:line-through}.content-type[data-v-4b712394]{display:flex;align-items:center;font-size:var(--scalar-heading-4);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);line-height:1.45;margin-top:24px;padding-bottom:12px;border-bottom:1px solid var(--scalar-border-color);flex-flow:wrap}.content-type-select[data-v-4b712394]{position:relative;padding-left:9px;height:-moz-fit-content;height:fit-content;color:var(--scalar-color-2);font-size:var(--scalar-font-size-3);display:flex;align-items:center}.content-type-no-select.content-type-select[data-v-4b712394]{pointer-events:none}.content-type-no-select.content-type-select[data-v-4b712394]:after{display:none}.content-type-select span[data-v-4b712394]{display:flex;align-items:center}.content-type-select[data-v-4b712394]:after{content:"";width:7px;height:7px;transform:rotate(45deg) translate3d(-2px,-4px,0);display:block;margin-left:7px;box-shadow:1px 1px 0 currentColor}.content-type-select select[data-v-4b712394]{border:none;outline:none;cursor:pointer;background:var(--scalar-background-3);box-shadow:-2px 0 0 0 var(--scalar-background-3);position:absolute;top:0;left:0;width:100%;height:100%;opacity:0;-webkit-appearance:none;-moz-appearance:none;appearance:none}.content-type-select[data-v-4b712394]:hover{color:var(--scalar-color-1)}@media (max-width: 460px){.content-type-select[data-v-4b712394]{margin-left:auto;padding-right:3px}}.schema-type-icon[data-v-2a52eb0b]{color:var(--scalar-color-1);display:none}.schema-type[data-v-2a52eb0b]{font-family:var(--scalar-font-code);color:var(--scalar-color-1)}.error[data-v-f48d0197]{background-color:var(--scalar-color-red)}.schema-card[data-v-f48d0197]{z-index:0;position:relative;font-size:var(--scalar-font-size-4);color:var(--scalar-color-1)}.schema-card-title[data-v-f48d0197]{height:var(--schema-title-height);padding:6px 10px;display:flex;align-items:center;gap:4px;color:var(--scalar-color-2);font-weight:var(--scalar-semibold);font-size:var(--scalar-micro);border-bottom:var(--scalar-border-width) solid transparent}button.schema-card-title[data-v-f48d0197]{cursor:pointer}button.schema-card-title[data-v-f48d0197]:hover{color:var(--scalar-color-1)}.schema-card-title-icon--open[data-v-f48d0197]{transform:rotate(45deg)}.schema-properties-open>.schema-card-title[data-v-f48d0197]{border-bottom-left-radius:0;border-bottom-right-radius:0;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.schema-properties-open>.schema-properties[data-v-f48d0197]{width:-moz-fit-content;width:fit-content}.schema-card-description+.schema-properties[data-v-f48d0197]{width:-moz-fit-content;width:fit-content;margin-top:12px}.schema-properties-open.schema-properties[data-v-f48d0197],.schema-properties-open>.schema-card--open[data-v-f48d0197]{width:100%}.schema-card .property[data-v-f48d0197]:last-of-type{padding-bottom:10px}.schema-properties[data-v-f48d0197]{display:flex;flex-direction:column;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);width:-moz-fit-content;width:fit-content}.schema-properties .schema-properties[data-v-f48d0197]{border-radius:13.5px}.schema-properties .schema-properties.schema-properties-open[data-v-f48d0197]{border-radius:13.5px 13.5px 9px 9px}.schema-properties-open[data-v-f48d0197]{width:100%}.schema-card--compact[data-v-f48d0197]{align-self:flex-start}.schema-card--compact.schema-card--open[data-v-f48d0197]{align-self:initial}.schema-card-title--compact[data-v-f48d0197]{color:var(--scalar-color-2);padding:6px 8px;height:auto;border-bottom:none}.schema-card--compact>.schema-properties[data-v-f48d0197],.schema-card-title--compact[data-v-f48d0197]{border-radius:13.5px}.schema-card-title--compact>.schema-card-title-icon[data-v-f48d0197]{margin:0}.schema-card-title--compact>.schema-card-title-icon--open[data-v-f48d0197]{transform:rotate(45deg)}.schema-properties-open>.schema-card-title--compact[data-v-f48d0197]{position:static}.schema-card--compact.schema-card--level-0>.schema-properties[data-v-f48d0197]{border:none}[data-v-f48d0197] .schema-card-description p{font-size:var(--scalar-mini, var(--scalar-paragraph));color:var(--scalar-color-2);line-height:1.5;display:block;margin-bottom:6px}.schema-card-description[data-v-f48d0197]:first-of-type{padding-top:10px}.children .schema-card-description[data-v-f48d0197]:first-of-type{padding-top:0}.models-list-item .schema-properties[data-v-f48d0197]{margin-bottom:10px}.property-detail[data-v-3ae98a1b]{display:inline-flex}.property-detail+.property-detail[data-v-3ae98a1b]:before{display:block;content:"·";margin:0 .5ch}.property-detail-truncate[data-v-3ae98a1b]{overflow:hidden}.property-detail-truncate>.property-detail-value[data-v-3ae98a1b]{overflow:hidden;text-overflow:ellipsis}.property-detail-prefix[data-v-3ae98a1b]{color:var(--scalar-color-2)}code.property-detail-value[data-v-3ae98a1b]{font-family:var(--scalar-font-code);font-size:var(--scalar-font-size-3);color:var(--scalar-color-2);background:var(--scalar-background-3);padding:0 4px;border:.5px solid var(--scalar-border-color);border-radius:var(--scalar-radius)}.badge[data-v-190574ac]{color:var(--scalar-color-2);font-size:var(--scalar-micro);background:var(--scalar-background-2);padding:2px 6px;border-radius:12px;font-weight:var(--scalar-semibold);display:inline-block;text-transform:uppercase}.property-heading[data-v-dd95f4d7]{display:flex;flex-wrap:wrap;align-items:baseline;row-gap:9px;white-space:nowrap}.property-heading[data-v-dd95f4d7]>*{margin-right:9px}.property-heading[data-v-dd95f4d7]:last-child{margin-right:0}.property-heading>.property-detail[data-v-dd95f4d7]:not(:last-of-type){margin-right:0}.property-name[data-v-dd95f4d7]{font-family:var(--scalar-font-code);font-weight:var(--scalar-semibold);font-size:var(--scalar-font-size-3);display:flex}.property-additional[data-v-dd95f4d7]{font-family:var(--scalar-font-code)}.property-required[data-v-dd95f4d7],.property-optional[data-v-dd95f4d7]{color:var(--scalar-color-2)}.property-required[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-orange)}.property-read-only[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-blue)}.property-write-only[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-green)}.property-detail[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-2);display:flex;align-items:center;min-width:0}.property-const[data-v-dd95f4d7]{color:var(--scalar-color-1)}.property[data-v-f5c211dc]{color:var(--scalar-color-1);padding:10px;font-size:var(--scalar-mini)}.property[data-v-f5c211dc]:last-of-type{padding-bottom:0}.property--compact.property--level-0[data-v-f5c211dc]{padding:12px 0}.property--deprecated[data-v-f5c211dc]{background:repeating-linear-gradient(-45deg,var(--scalar-background-2) 0,var(--scalar-background-2) 2px,transparent 2px,transparent 5px);background-size:100%}.property--deprecated[data-v-f5c211dc]>*{opacity:.75}.property-description[data-v-f5c211dc]{margin-top:6px;line-height:1.4;font-size:var(--scalar-small)}.property-heading:empty+.property-description[data-v-f5c211dc]:last-of-type,.property-description[data-v-f5c211dc]:first-of-type:last-of-type{margin-top:0}[data-v-f5c211dc] .property-description *{color:var(--scalar-color-2)!important}.property[data-v-f5c211dc]:not(:last-of-type){border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.children[data-v-f5c211dc]{display:flex;flex-direction:column;padding-top:8px}.property-example[data-v-f5c211dc]{display:flex;flex-direction:column;margin-top:6px;max-height:calc(((var(--full-height) - var(--refs-header-height))) / 2);font-size:var(--scalar-micro);border:var(--scalar-border-width) solid var(--scalar-border-color);background:var(--scalar-background-2);border-radius:var(--scalar-radius-lg)}.property-example-label[data-v-f5c211dc]{font-weight:var(--scalar-semibold);color:var(--scalar-color-3);padding:6px}.property-example-value[data-v-f5c211dc]{all:unset;font-family:var(--scalar-font-code);padding:6px;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.property-rule[data-v-f5c211dc]{display:flex;flex-direction:column;gap:6px;margin-top:12px;border-radius:var(--scalar-radius-lg)}.property-enum-value[data-v-f5c211dc]{padding:3px 0;color:var(--scalar-color-2);line-height:1.5;word-break:break-word}.property-enum-value[data-v-f5c211dc]:before{content:"⊢";margin-right:6px;color:var(--scalar-color-3)}.property-enum-values[data-v-f5c211dc]{margin-top:8px;list-style:none}.property--compact .property-example[data-v-f5c211dc]{background:transparent;border:none;display:flex;flex-direction:row;gap:8px}.property--compact .property-example-label[data-v-f5c211dc],.property--compact .property-example-value[data-v-f5c211dc]{padding:3px 0 0}.property--compact .property-example-value[data-v-f5c211dc]{background:var(--scalar-background-2);border-top:0;border-radius:var(--scalar-radius);padding:3px 4px}.property-list[data-v-f5c211dc]{border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:var(--scalar-radius);margin-top:10px}.property-list .property[data-v-f5c211dc]:last-of-type{padding-bottom:10px}.property-name[data-v-f5c211dc]{font-family:var(--scalar-font-code);font-weight:var(--scalar-semibold)}.enum-toggle-button[data-v-f5c211dc]{align-items:center;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:13.5px;cursor:pointer;color:var(--scalar-color-2);display:flex;font-weight:var(--scalar-semibold);gap:4px;margin-top:8px;padding:6px 10px;-webkit-user-select:none;-moz-user-select:none;user-select:none;white-space:nowrap}.enum-toggle-button[data-v-f5c211dc]:hover{color:var(--scalar-color-1)}.enum-toggle-button-icon--open[data-v-f5c211dc]{transform:rotate(45deg)}.headers-card[data-v-2f99eae6]{z-index:0;margin-top:12px;margin-bottom:6px;position:relative;font-size:var(--scalar-font-size-4);color:var(--scalar-color-1);align-self:flex-start}.headers-card.headers-card--open[data-v-2f99eae6]{align-self:initial}.headers-card-title[data-v-2f99eae6]{padding:6px 10px;display:flex;align-items:center;gap:4px;color:var(--scalar-color-3);font-weight:var(--scalar-semibold);font-size:var(--scalar-micro);border-radius:13.5px}button.headers-card-title[data-v-2f99eae6]{cursor:pointer}button.headers-card-title[data-v-2f99eae6]:hover{color:var(--scalar-color-1)}.headers-card-title-icon--open[data-v-2f99eae6]{transform:rotate(45deg)}.headers-properties-open>.headers-card-title[data-v-2f99eae6]{border-bottom-left-radius:0;border-bottom-right-radius:0;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.headers-properties-open[data-v-2f99eae6]{width:100%}.headers-properties[data-v-2f99eae6]{display:flex;flex-direction:column;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:13.5px;width:-moz-fit-content;width:fit-content}.headers-card .property[data-v-2f99eae6]:last-of-type{padding-bottom:10px}.headers-card-title>.headers-card-title-icon[data-v-2f99eae6]{width:10px;height:10px;margin:0}.headers-card-title>.headers-card-title-icon--open[data-v-2f99eae6]{transform:rotate(45deg)}.parameter-item[data-v-e1efed5d]{display:flex;flex-direction:column;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.parameter-item:last-of-type .parameter-schema[data-v-e1efed5d]{padding-bottom:0}.parameter-item-container[data-v-e1efed5d]{padding:0}.parameter-item-headers[data-v-e1efed5d]{border:var(--scalar-border-width) solid var(--scalar-border-color)}.parameter-item-name[data-v-e1efed5d]{margin-right:6px;font-weight:var(--scalar-semibold);font-size:var(--scalar-font-size-3);font-family:var(--scalar-font-code);color:var(--scalar-color-1)}.parameter-item-type[data-v-e1efed5d]{font-size:var(--scalar-micro);color:var(--scalar-color-2);margin-right:6px;line-height:1.4;white-space:nowrap;text-overflow:ellipsis;width:100%;overflow:hidden}.parameter-item-trigger-open .parameter-item-type[data-v-e1efed5d]{white-space:normal}.parameter-item-trigger+.parameter-item-container[data-v-e1efed5d] .property--level-0>.property-heading .property-detail-value{font-size:var(--scalar-font-size-3)}.parameter-item-required-optional[data-v-e1efed5d]{color:var(--scalar-color-2);font-weight:var(--scalar-semibold);margin-right:6px;position:relative}.parameter-item--required[data-v-e1efed5d]{text-transform:uppercase;font-size:var(--scalar-micro);font-weight:var(--scalar-semibold);color:var(--scalar-color-orange)}.parameter-item-description[data-v-e1efed5d]{margin-top:3px!important;font-size:var(--scalar-small);color:var(--scalar-color-2);line-height:1.4}.parameter-item-description[data-v-e1efed5d] p{margin-top:4px;font-size:var(--scalar-small);color:var(--scalar-color-2);line-height:1.4}.parameter-schema[data-v-e1efed5d]{padding-bottom:9px;margin-top:3px}.parameter-item-trigger[data-v-e1efed5d]{padding:12px 0;cursor:pointer;outline:none;text-align:left;position:relative;align-items:baseline}.parameter-item-trigger-open[data-v-e1efed5d]{padding-bottom:0}.parameter-item-trigger[data-v-e1efed5d]:after{content:"";position:absolute;height:10px;width:100%;bottom:0}.parameter-item-icon[data-v-e1efed5d]{color:var(--scalar-color-3);height:18px;left:-19px;position:absolute;top:11px;width:18px}.parameter-item-trigger:hover .parameter-item-icon[data-v-e1efed5d],.parameter-item-trigger:focus-visible .parameter-item-icon[data-v-e1efed5d]{color:var(--scalar-color-1)}.parameter-item-trigger:focus-visible .parameter-item-icon[data-v-e1efed5d]{outline:1px solid var(--scalar-color-accent);outline-offset:2px;border-radius:var(--scalar-radius)}.parameter-item-content-type[data-v-e1efed5d]{margin-left:auto;opacity:0;transition:opacity .1s ease-in-out;color:var(--scalar-color-3);font-size:var(--scalar-micro);background:var(--scalar-background-2);padding:2px 6px;border-radius:12px}.parameter-item-trigger:hover .parameter-item-content-type[data-v-e1efed5d]{opacity:1}.parameter-list[data-v-b05cf5ff]{margin-top:24px}.parameter-list-title[data-v-b05cf5ff]{font-size:var(--scalar-font-size-2);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);line-height:1.45;margin-top:12px;margin-bottom:12px}.parameter-list-items[data-v-b05cf5ff]{list-style:none;padding:0;margin:0 0 12px;font-size:var(--scalar-small)}.request-body-title[data-v-dc8f9923]{display:flex;align-items:center;font-size:var(--scalar-font-size-2);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);line-height:1.45;margin-top:24px;padding-bottom:12px;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color);flex-flow:wrap}.request-body-title-select[data-v-dc8f9923]{position:relative;height:-moz-fit-content;height:fit-content;margin-left:auto;font-weight:var(--scalar-regular);display:flex;align-items:center;color:var(--scalar-color-3);font-size:var(--scalar-micro);background:var(--scalar-background-2);padding:2px 6px;border-radius:12px;border:var(--scalar-border-width) solid var(--scalar-border-color)}.request-body-title-no-select.request-body-title-select[data-v-dc8f9923]{pointer-events:none}.request-body-title-no-select[data-v-dc8f9923]{border:none}.request-body-title-no-select.request-body-title-select[data-v-dc8f9923]:after{display:none}.request-body-title-select span[data-v-dc8f9923]{display:flex;align-items:center}.request-body-title-select[data-v-dc8f9923]:after{content:"";width:6px;height:6px;transform:rotate(45deg) translate3d(0,-3px,0);display:block;margin-left:6px;box-shadow:1px 1px 0 currentColor;margin-right:5px}.request-body-title-select select[data-v-dc8f9923]{border:none;outline:none;cursor:pointer;background:var(--scalar-background-3);position:absolute;top:0;left:0;width:100%;height:100%;opacity:0;-webkit-appearance:none;-moz-appearance:none;appearance:none}.request-body-title-select[data-v-dc8f9923]:hover{color:var(--scalar-color-1)}.request-body-description[data-v-dc8f9923]{margin-top:6px;font-size:var(--scalar-small);width:100%}.request-body-description[data-v-dc8f9923] .markdown *{color:var(--scalar-color-2)!important}@media (max-width: 460px){.request-body-title-select[data-v-dc8f9923]{margin-left:auto;padding-right:3px}}.section-wrapper[data-v-07cabcb0]{color:var(--scalar-color-1);padding-top:12px;margin-top:-12px}.section-accordion[data-v-07cabcb0]{display:flex;flex-direction:column;border-radius:var(--scalar-radius-lg);background:var(--scalar-background-2)}.section-accordion-transparent[data-v-07cabcb0]{background:transparent;border:1px solid var(--scalar-border-color)}.section-accordion-button[data-v-07cabcb0]{padding:6px;display:flex;align-items:center;gap:6px;cursor:pointer}.section-accordion-button-content[data-v-07cabcb0]{flex:1;min-width:0}.section-accordion-button-actions[data-v-07cabcb0]{display:flex;align-items:center;gap:6px;color:var(--scalar-color-3)}.section-accordion-chevron[data-v-07cabcb0]{margin-right:4px;width:20px;cursor:pointer;opacity:1;color:var(--scalar-color-3)}.section-accordion-button:hover .section-accordion-chevron[data-v-07cabcb0]{color:var(--scalar-color-1)}.section-accordion-content[data-v-07cabcb0]{border-top:1px solid var(--scalar-border-color);display:flex;flex-direction:column}.section-accordion-description[data-v-07cabcb0]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color--1);padding:10px 12px 0}.section-accordion-content-card[data-v-07cabcb0] .property:last-of-type{padding-bottom:9px}.show-api-client-button[data-v-ea8189a2]{-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none;padding:4px 6px;white-space:nowrap;border-radius:var(--scalar-radius);display:flex;justify-content:center;align-items:center;font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-background-2);font-family:var(--scalar-font);background:var(--scalar-button-1);position:relative;cursor:pointer;box-sizing:border-box;box-shadow:inset 0 0 0 1px #0000001a;outline-offset:2px}.show-api-client-button span[data-v-ea8189a2],.show-api-client-button svg[data-v-ea8189a2]{fill:currentColor;color:var(--scalar-button-1-color);z-index:1}.show-api-client-button[data-v-ea8189a2]:hover{background:var(--scalar-button-1-hover)}.show-api-client-button svg[data-v-ea8189a2]{margin-right:4px}.empty-state[data-v-4b51e8ca]{margin:10px 0 10px 12px;text-align:center;font-size:var(--scalar-micro);min-height:56px;display:flex;align-items:center;justify-content:center;border-radius:var(--scalar-radius-lg);color:var(--scalar-color-2)}.rule-title[data-v-4b51e8ca]{font-family:var(--scalar-font-code);color:var(--scalar-color-1);display:inline-block;margin:12px 0 6px;border-radius:var(--scalar-radius)}.rule[data-v-4b51e8ca]{margin:0 12px;border-radius:var(--scalar-radius-lg)}.rule-items[data-v-4b51e8ca]{counter-reset:list-number;display:flex;flex-direction:column;gap:12px;border-left:1px solid var(--scalar-border-color);padding:12px 0}.rule-item[data-v-4b51e8ca]{counter-increment:list-number;border:1px solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);overflow:hidden;margin-left:24px}.rule-item[data-v-4b51e8ca]:before{border:1px solid var(--scalar-border-color);border-top:0;border-right:0;content:" ";display:block;width:24px;height:6px;border-radius:0 0 0 var(--scalar-radius-lg);margin-top:6px;color:var(--scalar-color-2);transform:translate(-25px);color:var(--scalar-color-1);position:absolute}.tab-list[data-v-03fdb072]{display:flex;gap:6px;position:relative;flex:1;padding:9px 12px;overflow:auto}.scalar-card-header.scalar-card-header-tabs[data-v-03fdb072]{padding:0}.tab[data-v-d0e07e65]{background:none;border:none;font-size:var(--scalar-mini);font-family:var(--scalar-font);color:var(--scalar-color-2);font-weight:var(--scalar-semibold);line-height:calc(var(--scalar-mini) + 2px);white-space:nowrap;cursor:pointer;padding:0;margin-right:3px;text-transform:uppercase;position:relative;line-height:1.35}.tab[data-v-d0e07e65]:before{content:"";position:absolute;z-index:0;left:-6px;top:-6px;width:calc(100% + 12px);height:calc(100% + 12px);border-radius:var(--scalar-radius);background:var(--scalar-background-3);opacity:0}.tab[data-v-d0e07e65]:hover:before,.tab[data-v-d0e07e65]:focus-visible:before{opacity:1}.tab[data-v-d0e07e65]:focus-visible:before{outline:1px solid var(--scalar-color-accent)}.tab span[data-v-d0e07e65]{z-index:1;position:relative}.tab-selected[data-v-d0e07e65]{color:var(--scalar-color-1);text-decoration:underline;text-underline-offset:var(--scalar-micro)}.scalar-card-footer[data-v-7621d1be]{flex-shrink:0}.text-select{position:relative;height:-moz-fit-content;height:fit-content}.text-select:has(:focus-visible) .text-select-label{outline:1px solid var(--scalar-color-accent);outline-offset:4px;border-radius:calc(var(--scalar-radius) / 2)}.text-select--single-option{pointer-events:none}.text-select select{border:none;outline:none;cursor:pointer;background:var(--scalar-background-3);box-shadow:-2px 0 0 0 var(--scalar-background-3);position:absolute;top:0;left:0;width:100%;height:100%;opacity:0;-moz-appearance:none;-webkit-appearance:none;appearance:none}.text-select span{font-size:var(--scalar-mini);color:var(--scalar-color-2);font-weight:var(--scalar-semibold);white-space:nowrap;display:flex;align-items:center;justify-content:center}.text-select:hover span{color:var(--scalar-color-1)}.text-select:not(.text-select--single-option) span:after{content:"";width:7px;height:7px;transform:rotate(45deg) translate3d(-2px,-2px,0);display:block;margin-left:6px;box-shadow:1px 1px 0 currentColor}.text-select span:hover{background:var(--scalar-background-2)}.example-selector[data-v-a7e22e14]{padding:4px}.markdown[data-v-d84b9039] *{margin:0}.code-copy[data-v-d84b9039]{display:flex;align-items:center;justify-content:center;-moz-appearance:none;appearance:none;-webkit-appearance:none;outline:none;background:transparent;cursor:pointer;color:var(--scalar-color-3);border:none;padding:0;margin-right:12px}.code-copy[data-v-d84b9039]:hover{color:var(--scalar-color-1)}.code-copy svg[data-v-d84b9039]{width:13px;height:13px}.response-card-footer[data-v-d84b9039]{display:flex;flex-direction:column;flex-shrink:0;padding:10px 12px;gap:8px;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.response-example-selector[data-v-d84b9039]{align-self:flex-start;margin:-4px}.response-description[data-v-d84b9039]{font-weight:var(--scalar-semibold);font-size:var(--scalar-micro);color:var(--scalar-color--1);display:flex;align-items:center;box-sizing:border-box}.schema-type[data-v-d84b9039]{font-size:var(--scalar-micro);color:var(--scalar-color-2);font-weight:var(--scalar-semibold);background:var(--scalar-background-3);padding:2px 4px;border-radius:4px;margin-right:4px}.schema-example[data-v-d84b9039]{font-size:var(--scalar-micro);color:var(--scalar-color-2);font-weight:var(--scalar-semibold)}.example-response-tab[data-v-d84b9039]{display:block;margin:6px}.scalar-card-container[data-v-d84b9039]{flex:1;background:var(--scalar-background-2);display:grid}.scalar-card-container[data-v-d84b9039] .cm-scroller{overflow-y:hidden}.scalar-card-checkbox[data-v-d84b9039]{display:flex;align-items:center;justify-content:center;position:relative;min-height:17px;cursor:pointer;-webkit-user-select:none;-moz-user-select:none;user-select:none;font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-2);width:-moz-fit-content;width:fit-content;white-space:nowrap;margin-right:9px;gap:6px}.scalar-card-checkbox:has(.scalar-card-checkbox-input:focus-visible) .scalar-card-checkbox-checkmark[data-v-d84b9039]{outline:1px solid var(--scalar-color-accent)}.scalar-card-checkbox[data-v-d84b9039]:hover{color:var(--scalar-color--1)}.scalar-card-checkbox .scalar-card-checkbox-input[data-v-d84b9039]{position:absolute;opacity:0;cursor:pointer;height:0;width:0}.scalar-card-checkbox-checkmark[data-v-d84b9039]{height:17px;width:17px;border-radius:var(--scalar-radius);background-color:transparent;background-color:var(--scalar-background-3);box-shadow:inset 0 0 0 1px var(--scalar-border-color)}.scalar-card-checkbox[data-v-d84b9039]:has(.scalar-card-checkbox-input:checked){color:var(--scalar-color-1)}.scalar-card-checkbox .scalar-card-checkbox-input:checked~.scalar-card-checkbox-checkmark[data-v-d84b9039]{background-color:var(--scalar-button-1);box-shadow:none}.scalar-card-checkbox-checkmark[data-v-d84b9039]:after{content:"";position:absolute;display:none}.scalar-card-checkbox .scalar-card-checkbox-input:checked~.scalar-card-checkbox-checkmark[data-v-d84b9039]:after{display:block}.scalar-card-checkbox .scalar-card-checkbox-checkmark[data-v-d84b9039]:after{right:6px;top:36.5%;width:5px;height:9px;border:solid 1px var(--scalar-button-1-color);border-width:0 1.5px 1.5px 0;transform:rotate(45deg)}.request[data-v-3ff33ef0]{display:flex;flex-wrap:nowrap}.request-header[data-v-3ff33ef0]{display:flex;gap:6px;text-transform:initial}.request-method[data-v-3ff33ef0]{font-family:var(--scalar-font-code);text-transform:uppercase}.request-client-picker[data-v-3ff33ef0]{padding-left:12px;padding-right:9px}.request-card-footer[data-v-3ff33ef0]{display:flex;justify-content:flex-end;padding:6px;flex-shrink:0}.request-card-footer-addon[data-v-3ff33ef0]{display:flex;align-items:center;flex:1;min-width:0}.request-editor-section[data-v-3ff33ef0]{display:flex;flex:1}.request-card-simple[data-v-3ff33ef0]{display:flex;align-items:center;justify-content:space-between;padding:8px 8px 8px 12px;font-size:var(--scalar-small)}.code-snippet[data-v-3ff33ef0]{display:flex;flex-direction:column;width:100%}.operation-title[data-v-e104650f]{display:flex;justify-content:space-between}.operation-details[data-v-e104650f]{display:flex;align-items:center;margin-top:0;gap:9px;min-width:0;flex-shrink:1}.endpoint-type[data-v-e104650f]{display:flex;align-items:center;justify-content:center;gap:6px;position:relative;z-index:0;width:60px;padding:6px;flex-shrink:0;font-size:var(--scalar-small);text-transform:uppercase;font-weight:var(--scalar-bold);font-family:var(--scalar-font)}.endpoint-type[data-v-e104650f]:after{content:"";position:absolute;top:0;right:0;bottom:0;left:0;z-index:-1;background:currentColor;opacity:.15;border-radius:var(--scalar-radius-lg)}.endpoint-anchor[data-v-e104650f]{display:flex;align-items:center;min-width:0;flex-shrink:1;font-size:20px}.endpoint-anchor.label[data-v-e104650f]{display:flex}.endpoint-label[data-v-e104650f]{display:flex;align-items:baseline;gap:9px;min-width:0;flex-shrink:1;color:var(--scalar-color-1)}.endpoint-label-path[data-v-e104650f]{font-family:var(--scalar-font-code);font-size:var(--scalar-mini);text-overflow:ellipsis;white-space:nowrap;overflow:hidden}.endpoint-label-path[data-v-e104650f] em{color:var(--scalar-color-2)}.endpoint-label-name[data-v-e104650f]{color:var(--scalar-color-2);font-size:var(--scalar-small);flex-shrink:1000000000;text-overflow:ellipsis;white-space:nowrap;overflow:hidden}.endpoint-try-hint[data-v-e104650f]{padding:2px;height:24px;width:24px;flex-shrink:0}.endpoint-copy[data-v-e104650f]{color:currentColor;padding:2px}.endpoint-copy[data-v-e104650f] svg{stroke-width:2px}.endpoint-content[data-v-e104650f]{display:grid;grid-auto-columns:1fr;grid-auto-flow:row;gap:9px;padding:9px}@media (min-width: 1000px){.endpoint-content[data-v-e104650f]{grid-auto-flow:column}}@container (max-width: 900px){.endpoint-content[data-v-e104650f]{grid-template-columns:1fr}}.endpoint-content[data-v-e104650f]>*{max-height:unset}.operation-details-card[data-v-e104650f]{display:flex;flex-direction:column;gap:12px}.operation-details-card-item[data-v-e104650f] .parameter-list{border:1px solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);margin-top:0}.operation-details-card-item[data-v-e104650f]{display:flex;flex-direction:column;gap:12px}.operation-details-card-item[data-v-e104650f] .parameter-list-items{margin-bottom:0}.operation-details-card[data-v-e104650f] .parameter-item:last-of-type .parameter-schema{padding-bottom:12px}.operation-details-card[data-v-e104650f] .parameter-list .parameter-list{margin-bottom:12px}.operation-details-card[data-v-e104650f] .parameter-item{margin:0;padding:0 9px}.operation-details-card[data-v-e104650f] .property{padding:9px;margin:0}.operation-details-card[data-v-e104650f] .parameter-list-title,.operation-details-card[data-v-e104650f] .request-body-title{text-transform:uppercase;font-weight:var(--scalar-bold);font-size:var(--scalar-mini);color:var(--scalar-color-2);line-height:1.33;padding:9px;margin:0}.operation-details-card[data-v-e104650f] .request-body-title-select{text-transform:initial;font-weight:initial;margin-left:auto}.examples[data-v-0916dd9f]{position:sticky;top:calc(var(--refs-header-height) + 24px)}.deprecated[data-v-0916dd9f] *{text-decoration:line-through}.example-path[data-v-0916dd9f]{color:var(--scalar-color-2);font-family:var(--scalar-font-code)}.example-path[data-v-0916dd9f] em{color:var(--scalar-color-1);font-style:normal}.references-loading{position:absolute;top:0;left:0;right:0;z-index:1;grid-area:rendered;background:var(--scalar-background-1)}.references-loading-top-spacer{top:-1px}.references-loading-hidden-tag .section-container>.section:first-child{display:none}.download[data-v-1fc94525]{margin-bottom:24px}.download-button[data-v-1fc94525]{color:var(--scalar-link-color, var(--scalar-color-accent));font-weight:var(--scalar-link-font-weight, inherit);-webkit-text-decoration:var(--scalar-text-decoration)!important;text-decoration:var(--scalar-text-decoration)!important;text-decoration-color:var(--scalar-text-decoration-color);font-size:var(--scalar-paragraph);cursor:pointer}.download-button[data-v-1fc94525]:hover{color:var(--scalar-link-color-hover, var(--scalar-color-accent));-webkit-text-decoration:var(--scalar-text-decoration-hover)!important;text-decoration:var(--scalar-text-decoration-hover)!important;text-decoration-color:var(--scalar-text-decoration-color-hover)}.introduction-description-heading[data-v-fa3d93c5]{scroll-margin-top:64px;margin-top:1em;margin-bottom:.5em}.markdown+.markdown[data-v-fa3d93c5]{margin-top:1em}.introduction-description[data-v-fa3d93c5]{display:flex;flex-direction:column}.references-classic .introduction-description[data-v-fa3d93c5] img{max-width:720px}.sticky-cards[data-v-0336cd4f]{display:flex;flex-direction:column;position:sticky;top:calc(var(--refs-header-height) + 24px)}.client-libraries-content[data-v-b29373e8]{container:client-libraries-content / inline-size;display:flex;justify-content:center;overflow:hidden;padding:0 12px;background-color:var(--scalar-background-1);border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.client-libraries[data-v-b29373e8]{display:flex;align-items:center;justify-content:center;width:100%;position:relative;cursor:pointer;white-space:nowrap;padding:8px 2px;gap:6px;color:var(--scalar-color-3);border-bottom:1px solid transparent;-webkit-user-select:none;-moz-user-select:none;user-select:none}.client-libraries[data-v-b29373e8]:first-child{border-radius:var(--scalar-radius) 0 0 0}.client-libraries[data-v-b29373e8]:not(.client-libraries__active):hover:before{content:"";position:absolute;width:calc(100% - 4px);height:calc(100% - 4px);background:var(--scalar-background-2);left:2px;top:2px;z-index:0;border-radius:var(--scalar-radius)}.client-libraries[data-v-b29373e8]:active{color:var(--scalar-color-1)}.client-libraries[data-v-b29373e8]:focus-visible{outline:none;box-shadow:inset 0 0 0 1px var(--scalar-color-accent)}@media screen and (max-width: 450px){.client-libraries[data-v-b29373e8]:nth-of-type(4),.client-libraries[data-v-b29373e8]:nth-of-type(5){display:none}}.client-libraries-icon[data-v-b29373e8]{max-width:14px;max-height:14px;min-width:14px;width:100%;aspect-ratio:1;display:flex;align-items:center;justify-content:center;position:relative;box-sizing:border-box;color:currentColor}.client-libraries-icon__more svg[data-v-b29373e8]{height:initial}@container client-libraries-content (width < 400px){.client-libraries__select[data-v-b29373e8]{width:-moz-fit-content;width:fit-content}.client-libraries__select .client-libraries-icon__more+span[data-v-b29373e8]{display:none}}@container client-libraries-content (width < 380px){.client-libraries[data-v-b29373e8]{width:100%}.client-libraries span[data-v-b29373e8]{display:none}}.client-libraries__active[data-v-b29373e8]{color:var(--scalar-color-1);border-bottom:1px solid var(--scalar-color-1)}@keyframes codeloader-b29373e8{0%{transform:rotate(0)}to{transform:rotate(1turn)}}.client-libraries .client-libraries-text[data-v-b29373e8]:last-of-type{font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);position:relative;display:flex;align-items:center}.client-libraries__active .client-libraries-text[data-v-b29373e8]{color:var(--scalar-color-1)}.client-libraries__select select[data-v-b29373e8]{background:var(--scalar-background-3);color:var(--scalar-color-2);opacity:0;height:100%;width:100%;aspect-ratio:1;position:absolute;top:0;left:0;cursor:pointer;z-index:1;-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none}.client-libraries__select[data-v-b29373e8]:has(select:focus-visible){border-radius:var(--scalar-radius);box-shadow:inset 0 0 0 1px var(--scalar-color-accent)}@media screen and (max-width: 600px){.references-classic .client-libraries[data-v-b29373e8]{flex-direction:column}}.selected-client[data-v-35fd541a]{color:var(--scalar-color-1);font-size:var(--scalar-mini);font-family:var(--scalar-font-code);padding:9px 12px;border-top:none;white-space:nowrap;overflow:hidden;text-overflow:ellipsis;background:var(--scalar-background-1);border-top:var(--scalar-border-width) solid var(--scalar-border-color);min-height:-moz-fit-content;min-height:fit-content}.client-libraries-heading[data-v-35fd541a]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-1);padding:9px 12px;background-color:var(--scalar-background-2);display:flex;align-items:center;max-height:32px}.section-accordion-wrapper[data-v-1a30b4ff]{padding:0 60px}.section-accordion[data-v-1a30b4ff]{position:relative;width:100%;max-width:var(--refs-content-max-width);margin:auto}.section-accordion-content[data-v-1a30b4ff]{display:flex;flex-direction:column;gap:12px;padding-top:12px}.section-accordion-button[data-v-1a30b4ff]{width:100%;display:flex;cursor:pointer;padding:6px 0;margin:-6px 0;border-radius:var(--scalar-radius)}.section-accordion-chevron[data-v-1a30b4ff]{width:24px;position:absolute;left:-24px;height:auto;top:10px;color:var(--scalar-color-3)}.section-accordion-button:hover .section-accordion-chevron[data-v-1a30b4ff]{color:var(--scalar-color-1)}.section-accordion-title[data-v-1a30b4ff]{display:flex;flex-direction:column;align-items:flex-start;flex:1;padding:0 6px}.section-accordion-title[data-v-1a30b4ff] .section-header-wrapper{grid-template-columns:1fr}.section-accordion-title[data-v-1a30b4ff] .section-header{margin-bottom:0}@container narrow-references-container (max-width: 900px){.section-accordion-chevron[data-v-1a30b4ff]{width:16px;left:-16px;top:14px}.section-accordion-wrapper[data-v-1a30b4ff]{padding:calc(var(--refs-header-height)) 24px 0 24px}}.tag-section[data-v-43e1c813]{margin-bottom:48px}.tag-name[data-v-43e1c813]{text-transform:capitalize}.tag-description[data-v-43e1c813]{padding-bottom:4px;text-align:left}.webhook-request-body[data-v-cbaa1001]{margin-top:-18px}.collapsible-section[data-v-b32ceff4]{border-top:var(--scalar-border-width) solid var(--scalar-border-color);position:relative}.collapsible-section-header[data-v-b32ceff4]{color:var(--scalar-color-1)}.collapsible-section .collapsible-section-trigger[data-v-b32ceff4]{display:flex;align-items:center;cursor:pointer;padding:10px 0;font-size:var(--scalar-font-size-3);z-index:1;position:relative}.collapsible-section .collapsible-section-trigger[data-v-b32ceff4]:after{content:"";height:10px;width:100%;position:absolute;bottom:0}.collapsible-section-trigger svg[data-v-b32ceff4]{color:var(--scalar-color-3);position:absolute;left:-19px}.collapsible-section:hover .collapsible-section-trigger svg[data-v-b32ceff4]{color:var(--scalar-color-1)}.collapsible-section .collapsible-section-trigger[data-v-b32ceff4] .anchor-copy{line-height:18.5px}.collapsible-section-content[data-v-b32ceff4]{padding:0;margin:0;scroll-margin-top:140px}.webhooks-list[data-v-8cee6fce]{display:contents}.webhooks-list-truncated .webhooks-list-item[data-v-8cee6fce]:last-child{border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.webhooks-list-item[aria-expanded=true][data-v-8cee6fce]{padding-bottom:48px}.webhook-section-description.webhook-section-description[data-v-8cee6fce]{color:var(--scalar-color-2);font-size:var(--scalar-font-size-3);padding-bottom:12px}.show-more-webhooks[data-v-8cee6fce]{margin-top:32px;top:0}.reference-models[data-v-4396a256]{margin-bottom:48px}.reference-models-anchor[data-v-4396a256]{display:flex;align-items:center;font-size:20px;padding-left:6px;color:var(--scalar-color-1)}.reference-models-label[data-v-4396a256]{font-size:var(--scalar-mini)}.reference-models-label[data-v-4396a256] em{font-weight:var(--scalar-bold)}.models-list[data-v-7c1cd4b4]{display:contents}.models-list-truncated .models-list-item[data-v-7c1cd4b4]:last-child{border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.show-more-models[data-v-7c1cd4b4]{margin-top:32px;top:0}.narrow-references-container{container-name:narrow-references-container;container-type:inline-size}.render-loading[data-v-42cfaf8d]{height:calc(var(--full-height) - var(--refs-header-height));display:flex;align-items:center;justify-content:center}.introduction-card[data-v-42cfaf8d]{display:flex;flex-direction:column;background:var(--scalar-background-1)}.introduction-card-item[data-v-42cfaf8d]{display:flex;overflow:hidden;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);margin-bottom:12px;flex-direction:column;justify-content:start}@container narrow-references-container (max-width: 900px){.introduction-card-item[data-v-42cfaf8d]{border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}}.introduction-card-item[data-v-42cfaf8d]:has(.description) .server-form-container{border-bottom-left-radius:0;border-bottom-right-radius:0}.introduction-card-item[data-v-42cfaf8d] .request-item{border-bottom:0}.introduction-card-title[data-v-42cfaf8d]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-3)}.introduction-card-row[data-v-42cfaf8d]{gap:24px}@media (min-width: 600px){.introduction-card-row[data-v-42cfaf8d]{flex-flow:row wrap}}.introduction-card-row[data-v-42cfaf8d]>*{flex:1}@media (min-width: 600px){.introduction-card-row[data-v-42cfaf8d]>*{min-width:-moz-min-content;min-width:min-content}}@media (max-width: 600px){.introduction-card-row[data-v-42cfaf8d]>*{max-width:100%}}@container (max-width: 900px){.introduction-card-row[data-v-42cfaf8d]{flex-direction:column;align-items:stretch;gap:0px}}.introduction-card[data-v-42cfaf8d] .security-scheme-label{text-transform:uppercase;font-weight:var(--scalar-semibold)}.references-classic .introduction-card-row[data-v-42cfaf8d] .scalar-card:nth-of-type(2) .scalar-card-header{display:none}.references-classic .introduction-card-row[data-v-42cfaf8d] .scalar-card:nth-of-type(2) .scalar-card-header.scalar-card--borderless+.scalar-card-content{margin-top:0}.section-flare[data-v-42cfaf8d]{top:0;right:0;pointer-events:none}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}}@layer scalar-base{:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>");mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>")}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}}@layer scalar-base{:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-api-reference{color:var(--scalar-color-1)}.scalar-api-reference .container{width:100%}@media (min-width: 400px){.scalar-api-reference .container{max-width:400px}}@media (min-width: 600px){.scalar-api-reference .container{max-width:600px}}@media (min-width: 800px){.scalar-api-reference .container{max-width:800px}}@media (min-width: 1000px){.scalar-api-reference .container{max-width:1000px}}@media (min-width: 1200px){.scalar-api-reference .container{max-width:1200px}}.scalar-api-reference .row{display:flex;flex-direction:row}.scalar-api-reference .sr-only{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);white-space:nowrap;border-width:0}.scalar-api-reference .visible{visibility:visible}.scalar-api-reference .collapse{visibility:collapse}.scalar-api-reference .static{position:static}.scalar-api-reference .fixed{position:fixed}.scalar-api-reference .absolute{position:absolute}.scalar-api-reference .relative{position:relative}.scalar-api-reference .sticky{position:sticky}.scalar-api-reference .right-0{right:0}.scalar-api-reference .block{display:block}.scalar-api-reference .flex{display:flex}.scalar-api-reference .grid{display:grid}.scalar-api-reference .contents{display:contents}.scalar-api-reference .hidden{display:none}.scalar-api-reference .h-8{height:32px}.scalar-api-reference .flex-shrink,.scalar-api-reference .shrink{flex-shrink:1}.scalar-api-reference .transform{transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-api-reference .resize{resize:both}.scalar-api-reference .flex-wrap{flex-wrap:wrap}.scalar-api-reference .content-start{align-content:flex-start}.scalar-api-reference .content-end{align-content:flex-end}.scalar-api-reference .items-center{align-items:center}.scalar-api-reference .gap-12{gap:48px}.scalar-api-reference .divide-y>:not([hidden])~:not([hidden]){--tw-divide-y-reverse:0;border-top-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-y-reverse)));border-bottom-width:calc(var(--scalar-border-width) * var(--tw-divide-y-reverse))}.scalar-api-reference .truncate{overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.scalar-api-reference .border{border-width:var(--scalar-border-width)}.scalar-api-reference .bg-b-2{background-color:var(--scalar-background-2)}.scalar-api-reference .px-3{padding-left:12px;padding-right:12px}.scalar-api-reference .py-1{padding-top:4px;padding-bottom:4px}.scalar-api-reference .py-1\.5{padding-top:6px;padding-bottom:6px}.scalar-api-reference .py-2{padding-top:8px;padding-bottom:8px}.scalar-api-reference .py-2\.5{padding-top:10px;padding-bottom:10px}.scalar-api-reference .font{font-family:var(--scalar-font)}.scalar-api-reference .text-sm{font-size:var(--scalar-font-size-4)}.scalar-api-reference .font{font-weight:var(--scalar-regular)}.scalar-api-reference .font-medium{font-weight:var(--scalar-semibold)}.scalar-api-reference .uppercase{text-transform:uppercase}.scalar-api-reference .lowercase{text-transform:lowercase}.scalar-api-reference .capitalize{text-transform:capitalize}.scalar-api-reference .leading{line-height:1.5}.scalar-api-reference .text-c-2{color:var(--scalar-color-2)}.scalar-api-reference .text-c-3{color:var(--scalar-color-3)}.scalar-api-reference .text-green{color:var(--scalar-color-green)}.scalar-api-reference .text-orange{color:var(--scalar-color-orange)}.scalar-api-reference .text-red{color:var(--scalar-color-red)}.scalar-api-reference .underline{text-decoration-line:underline}.scalar-api-reference .outline{outline-style:solid}.scalar-api-reference .filter{filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-api-reference .transition{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,-webkit-backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter,-webkit-backdrop-filter;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-api-reference .\[--scalar-address-bar-height\:0px\]{--scalar-address-bar-height:0px}.scalar-api-reference .hover\:text-c-1:hover{color:var(--scalar-color-1)}.nav-item[data-v-d56f12d9]{padding:0 1rem;cursor:pointer;flex:1;justify-content:center;align-items:center;display:flex;border-radius:var(--scalar-radius-lg);background:var(--scalar-background-3);border:var(--scalar-border-width) solid var(--scalar-background-2);color:var(--scalar-color-3);padding:4.5px;min-width:0;overflow:hidden;position:relative}.dark-mode .nav-item[data-v-d56f12d9]{background:color-mix(in srgb,var(--scalar-background-2),transparent)}.nav-item-icon-copy[data-v-d56f12d9]{max-width:100%;white-space:nowrap;overflow:hidden;-webkit-mask-image:linear-gradient(to left,transparent 0,var(--scalar-background-2) 20px);mask-image:linear-gradient(to left,transparent 0,var(--scalar-background-2) 20px)}.nav-item:hover .nav-item-icon-copy[data-v-d56f12d9]{-webkit-mask-image:linear-gradient(to left,transparent 20px,var(--scalar-background-2) 40px);mask-image:linear-gradient(to left,transparent 20px,var(--scalar-background-2) 40px)}.nav-item-copy[data-v-d56f12d9]{max-width:calc(100% - 20px)}.nav-item[data-v-d56f12d9]:hover{color:var(--scalar-color-1)}.nav-item__active[data-v-d56f12d9]{background-color:var(--scalar-background-1);color:var(--scalar-color-1);border-color:var(--scalar-border-color)}.dark-mode .nav-item__active[data-v-d56f12d9]{background-color:var(--scalar-background-2)}.nav-item-close[data-v-d56f12d9]{position:absolute;right:3px;padding:2px;border-radius:var(--scalar-radius);background:transparent;max-width:20px;stroke-width:1.5px;color:var(--scalar-color-3);margin-left:-20px;opacity:0}.nav-item:hover .nav-item-close[data-v-d56f12d9]{opacity:1}.nav-item-close[data-v-d56f12d9]:hover{background-color:var(--scalar-background-4)}.nav-item__active .nav-item-close[data-v-d56f12d9]:hover{background-color:var(--scalar-background-2)}.download-app-button[data-v-c896207d]{box-shadow:0 0 0 .5px var(--scalar-border-color);background:linear-gradient(#ffffffbf,#00000009)}.dark-mode .download-app-button[data-v-c896207d]{background:linear-gradient(#ffffff1a,#00000026)}.download-app-button[data-v-c896207d]:hover{background:linear-gradient(#00000009,#ffffffbf)}.dark-mode .download-app-button[data-v-c896207d]:hover{background:linear-gradient(#00000026,#ffffff1a)}.http-bg-gradient[data-v-46723816]{background:linear-gradient(#ffffffbf,#00000009)}.http-bg-gradient[data-v-46723816]:hover{background:linear-gradient(#00000009,#ffffffbf)}.dark-mode .http-bg-gradient[data-v-46723816]{background:linear-gradient(#ffffff09,#00000026)}.dark-mode .http-bg-gradient[data-v-46723816]:hover{background:linear-gradient(#00000026,#ffffff09)}.commandmenu[data-v-e9fd9234]{box-shadow:var(--scalar-shadow-2);border-radius:var(--scalar-radius-lg);background-color:var(--scalar-background-1);max-height:60dvh;width:100%;max-width:580px;margin:12px;position:fixed;left:50%;top:150px;opacity:0;transform:translate3d(-50%,10px,0);animation:fadeincommandmenu-e9fd9234 ease-in-out .3s forwards;animation-delay:.1s}.commandmenu-overlay[data-v-e9fd9234]{position:fixed;top:0;right:0;bottom:0;left:0;background:#0003;animation:fadeincommand-e9fd9234 ease-in-out .3s forwards;cursor:pointer}@keyframes fadeincommand-e9fd9234{0%{opacity:0}to{opacity:1}}@keyframes fadeincommandmenu-e9fd9234{0%{opacity:0;transform:translate3d(-50%,10px,0)}to{opacity:1;transform:translate3d(-50%,0,0)}}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>");mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>")}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>");mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>")}@layer scalar-base{*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar .scalar-app-layout[data-v-cc5e3e81]{background:var(--scalar-background-1);height:calc(100% - 120px);max-width:1390px;width:100%;margin:auto;opacity:0;animation:scalarapiclientfadein-cc5e3e81 .35s forwards;position:relative;overflow:hidden;border-radius:8px;border:var(--scalar-border-width) solid var(--scalar-border-color)}@keyframes scalarapiclientfadein-cc5e3e81{0%{opacity:0}to{opacity:1}}.scalar .scalar-app-exit[data-v-cc5e3e81]{position:fixed;top:0;left:0;width:100vw;height:100vh;background:#00000038;transition:all .3s ease-in-out;cursor:pointer;animation:scalardrawerexitfadein-cc5e3e81 .35s forwards}.dark-mode .scalar .scalar-app-exit[data-v-cc5e3e81]{background:#00000073}.scalar .scalar-app-exit[data-v-cc5e3e81]:before{font-family:sans-serif;position:absolute;top:0;font-size:30px;font-weight:100;line-height:50px;right:12px;text-align:center;color:#fff;opacity:.6}.scalar .scalar-app-exit[data-v-cc5e3e81]:hover:before{opacity:1}@keyframes scalardrawerexitfadein-cc5e3e81{0%{opacity:0}to{opacity:1}}.scalar-container[data-v-cc5e3e81]{overflow:hidden;visibility:visible;position:fixed;bottom:0;top:0;left:0;width:100%;height:100%;display:flex;align-items:center;justify-content:center;z-index:10000}.scalar .url-form-input[data-v-cc5e3e81]{min-height:auto!important}.scalar .scalar-container[data-v-cc5e3e81]{line-height:normal}.scalar .scalar-app-header span[data-v-cc5e3e81]{color:var(--scalar-color-3)}.scalar .scalar-app-header a[data-v-cc5e3e81]{color:var(--scalar-color-1)}.scalar .scalar-app-header a[data-v-cc5e3e81]:hover{text-decoration:underline}.scalar-activate[data-v-cc5e3e81]{width:-moz-fit-content;width:fit-content;margin:0px .75rem .75rem auto;line-height:24px;cursor:pointer;font-size:.875rem;font-weight:600;display:flex;align-items:center;gap:6px}.scalar-activate-button[data-v-cc5e3e81]{display:flex;gap:6px;align-items:center;color:var(--scalar-color-blue);-webkit-appearance:none;-moz-appearance:none;appearance:none;outline:none;border:none;background:transparent;padding:0 .5rem}.scalar-activate:hover .scalar-activate-button[data-v-cc5e3e81]{background:var(--scalar-background-3);border-radius:3px}[data-v-4d46a9ea] .cm-editor{height:100%;outline:none;padding:0;background:transparent}[data-v-4d46a9ea] .cm-placeholder{color:var(--scalar-color-3)}[data-v-4d46a9ea] .cm-content{font-family:var(--scalar-font-code);font-size:var(--scalar-small);max-height:20px;padding:8px 0}[data-v-4d46a9ea] .cm-tooltip{background:transparent!important;filter:brightness(var(--scalar-lifted-brightness));border-radius:var(--scalar-radius);box-shadow:var(--scalar-shadow-2);border:none!important;outline:none!important;overflow:hidden!important}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul li{padding:3px 6px!important}[data-v-4d46a9ea] .cm-completionIcon-type:after{color:var(--scalar-color-3)!important}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul li[aria-selected]{background:var(--scalar-background-2)!important;color:var(--scalar-color-1)!important}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul{padding:6px!important;position:relative}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul li:hover{border-radius:3px;color:var(--scalar-color-1)!important;background:var(--scalar-background-3)!important}[data-v-4d46a9ea] .cm-activeLine,[data-v-4d46a9ea] .cm-activeLineGutter{background-color:transparent}[data-v-4d46a9ea] .cm-selectionMatch,[data-v-4d46a9ea] .cm-matchingBracket{border-radius:var(--scalar-radius);background:var(--scalar-background-4)!important}[data-v-4d46a9ea] .cm-css-color-picker-wrapper{display:inline-flex;outline:1px solid var(--scalar-background-3);border-radius:3px;overflow:hidden}[data-v-4d46a9ea] .cm-gutters{background-color:transparent;border-right:none;color:var(--scalar-color-3);font-size:var(--scalar-mini);line-height:1.44;border-radius:0 0 0 3px}[data-v-4d46a9ea] .cm-gutters:before{content:"";position:absolute;top:2px;left:2px;width:calc(100% - 2px);height:calc(100% - 4px);border-radius:var(--scalar-radius) 0 0 var(--scalar-radius);background-color:var(--scalar-background-1)}[data-v-4d46a9ea] .cm-gutterElement{font-family:var(--scalar-font-code)!important;padding:0 6px 0 8px!important;display:flex;align-items:center;justify-content:flex-end;position:relative}[data-v-4d46a9ea] .cm-gutter+.cm-gutter :not(.cm-foldGutter) .cm-gutterElement{padding-left:0!important}[data-v-4d46a9ea] .cm-scroller{overflow:auto}.peer:hover .copy-button[data-v-4d46a9ea],.copy-button[data-v-4d46a9ea]:focus-visible{opacity:100}.scalar-code-copy[data-v-4d46a9ea]{align-items:flex-start;display:flex;top:0;right:0;bottom:0;left:0;justify-content:flex-end;position:sticky}.copy-button[data-v-4d46a9ea]{align-items:center;display:flex;position:relative;background-color:var(--scalar-background-2);border:1px solid var(--scalar-border-color);border-radius:3px;color:var(--scalar-color-3);cursor:pointer;height:30px;margin-bottom:-30px;opacity:0;padding:6px;transition:opacity .15s ease-in-out,color .15s ease-in-out;top:0;right:0}.scalar-code-copy[data-v-4d46a9ea],.copy-button[data-v-4d46a9ea]{background:inherit}.copy-button[data-v-4d46a9ea]:hover{color:var(--scalar-color-1)}.copy-button svg[data-v-4d46a9ea]{stroke-width:1.5}.cm-pill{--tw-bg-base: var(--scalar-color-1);color:var(--tw-bg-base);padding:0 9px;display:inline-block;border-radius:30px;font-size:var(--scalar-mini);background:color-mix(in srgb,var(--tw-bg-base),transparent 94%)!important}.cm-pill.bg-grey{background:var(--scalar-background-3)!important}.dark-mode .cm-pill{background:color-mix(in srgb,var(--tw-bg-base),transparent 80%)!important}.cm-pill:first-of-type{margin-left:0}.cm-editor .cm-widgetBuffer{display:none}.cm-foldPlaceholder:hover{color:var(--scalar-color-1)}.cm-foldGutter .cm-gutterElement{font-size:var(--scalar-heading-4);padding:2px!important}.cm-foldGutter .cm-gutterElement:first-of-type{display:none}.cm-foldGutter .cm-gutterElement .cm-foldMarker{padding:2px}.cm-foldGutter .cm-gutterElement:hover .cm-foldMarker{background:var(--scalar-background-2);border-radius:var(--scalar-radius);color:var(--scalar-color-1)}[data-v-cb0ff5b1] .cm-editor{padding:0}[data-v-cb0ff5b1] .cm-content{align-items:center;background-color:transparent;display:flex;font-family:var(--scalar-font);font-size:var(--scalar-mini);padding:6px 8px}[data-v-cb0ff5b1] .cm-content:has(.cm-pill){padding:4px 3px}[data-v-cb0ff5b1] .cm-content .cm-pill:not(:last-of-type){margin-right:.5px}[data-v-cb0ff5b1] .cm-content .cm-pill:not(:first-of-type){margin-left:.5px}[data-v-cb0ff5b1] .cm-line{padding:0}.required[data-v-cb0ff5b1]:after{content:"Required"}input[data-v-cb0ff5b1]::-moz-placeholder{color:var(--scalar-color-3)}input[data-v-cb0ff5b1]::placeholder{color:var(--scalar-color-3)}.scalar-modal-layout .scalar-button-danger[data-v-b6a60dbc]{background:color-mix(in srgb,var(--scalar-color-red),transparent 95%);color:var(--scalar-color-red)}.scalar-modal-layout .scalar-button-danger[data-v-b6a60dbc]:hover,.scalar-modal-layout .scalar-button-danger[data-v-b6a60dbc]:focus{background:color-mix(in srgb,var(--scalar-color-red),transparent 90%)}.fade-request-section-content[data-v-7906217f]{background:linear-gradient(to left,var(--scalar-background-1) 64%,transparent)}.filter-hover[data-v-7906217f]{height:100%;padding-right:42px;padding-left:24px;position:absolute;right:0;transition:width 0s ease-in-out .2s;overflow:hidden}.filter-hover[data-v-7906217f]:hover,.filter-hover[data-v-7906217f]:has(:focus-visible){width:100%;z-index:10}.filter-hover[data-v-7906217f]:before{content:"";position:absolute;top:0;left:0;width:100%;height:100%;background-color:var(--scalar-background-1);opacity:0;transition:all .3s ease-in-out;pointer-events:none}.filter-hover-item[data-v-7906217f]{opacity:0}.filter-hover-item[data-v-7906217f]:not(:last-of-type){transform:translate3d(0,3px,0)}.filter-hover:hover .filter-hover-item[data-v-7906217f]{transition:opacity .2s ease-in-out,transform .2s ease-in-out}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(1){transition-delay:.05s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(2){transition-delay:.1s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(3){transition-delay:.15s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(4){transition-delay:.2s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(5){transition-delay:.25s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(6){transition-delay:.3s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(7){transition-delay:.35s}.filter-hover:hover .filter-hover-item[data-v-7906217f],.filter-hover:has(:focus-visible) .filter-hover-item[data-v-7906217f]{opacity:1;transform:translateZ(0)}.filter-hover[data-v-7906217f]:hover:before,.filter-hover[data-v-7906217f]:has(:focus-visible):before{opacity:.9;-webkit-backdrop-filter:blur(10px);backdrop-filter:blur(10px)}.context-bar-group:hover .context-bar-group-hover\:text-c-1[data-v-7906217f],.context-bar-group:has(:focus-visible) .context-bar-group-hover\:text-c-1[data-v-7906217f]{--tw-text-opacity: 1;color:rgb(var(--scalar-color-1) / var(--tw-text-opacity))}.context-bar-group:hover .context-bar-group-hover\:hidden[data-v-7906217f],.context-bar-group:has(:focus-visible) .context-bar-group-hover\:hidden[data-v-7906217f]{display:none}.light-mode .bg-preview[data-v-5f138d26]{background-image:url("data:image/svg+xml,%3Csvg xmlns=%27http://www.w3.org/2000/svg%27 width=%2716%27 height=%2716%27 fill=%27%23000%27 fill-opacity=%2710%25%27%3E%3Crect width=%278%27 height=%278%27 /%3E%3Crect x=%278%27 y=%278%27 width=%278%27 height=%278%27 /%3E%3C/svg%3E")}.dark-mode .bg-preview[data-v-5f138d26]{background-image:url("data:image/svg+xml,%3Csvg xmlns=%27http://www.w3.org/2000/svg%27 width=%2716%27 height=%2716%27 fill=%27%23FFF%27 fill-opacity=%2710%25%27%3E%3Crect width=%278%27 height=%278%27 /%3E%3Crect x=%278%27 y=%278%27 width=%278%27 height=%278%27 /%3E%3C/svg%3E")}[data-v-99361e9b] .cm-editor{background-color:transparent;font-size:var(--scalar-mini);outline:none}[data-v-99361e9b] .cm-gutters{background-color:var(--scalar-background-1);border-radius:var(--scalar-radius) 0 0 var(--scalar-radius)}.scalar-code-copy[data-v-99361e9b]{align-items:flex-start;display:flex;position:absolute;top:6px;right:6px;z-index:10;pointer-events:none;position:sticky;transform:translate(-6px)}.copy-button[data-v-99361e9b]{align-items:center;display:flex;background-color:var(--scalar-background-1);border:1px solid var(--scalar-border-color);border-radius:3px;color:var(--scalar-color-3);cursor:pointer;height:30px;opacity:0;padding:6px;pointer-events:auto;transition:opacity .15s ease-in-out,color .15s ease-in-out}.body-raw:hover .copy-button[data-v-99361e9b],.copy-button[data-v-99361e9b]:focus-visible{opacity:1}.copy-button[data-v-99361e9b]:hover{color:var(--scalar-color-1)}.scalar-code-block[data-v-cf80aa59] .hljs *{font-size:var(--scalar-mini)}.ascii-art-animate .ascii-art-line[data-v-0018c948]{border-right:1ch solid transparent;animation:typewriter-0018c948 4s steps(1) 1s 1 normal both,blinkTextCursor-0018c948 .5s steps(1) infinite normal}@keyframes typewriter-0018c948{0%{width:0}to{width:100%}}@keyframes blinkTextCursor-0018c948{0%{border-right-color:currentColor}50%{border-right-color:transparent}}.keycap-n[data-v-6c848a44]{background:-webkit-linear-gradient(5deg,transparent 30%,var(--scalar-color-3) 50%);-webkit-background-clip:text;-webkit-text-fill-color:transparent}.keycap-hotkey[data-v-6c848a44]{position:absolute;top:32px;line-height:26px}.scalar-version-number[data-v-04a8f835]{transform:skew(0,13deg);width:76px;height:76px;position:absolute;margin-left:-36px;font-size:8px;font-family:var(--scalar-font-code);line-height:11px;margin-top:-113px;border-radius:9px 9px 16px 12px;box-shadow:inset 2px 0 0 2px var(--scalar-background-2);text-align:center;display:flex;flex-direction:column;align-items:center;justify-content:center;text-transform:initial;text-decoration-color:var(--scalar-color-3)}.scalar-version-number a[data-v-04a8f835]{font-weight:700;background:var(--scalar-background-2);padding:2px 4px;border-radius:3px;text-decoration:none;border:.5px solid var(--scalar-border-color)}.gitbook-show[data-v-04a8f835]{display:none}.v-enter-active[data-v-ef5f83d8]{transition:opacity .5s ease}.v-enter-from[data-v-ef5f83d8]{opacity:0}.response-body-virtual[data-headlessui-state=open],.response-body-virtual[data-headlessui-state=open] .diclosure-panel{display:flex;flex-direction:column;flex-grow:1}.animate-response-heading .response-heading[data-v-ec732740]{animation:push-response-ec732740 .2s ease-in-out forwards;opacity:1}@keyframes push-response-ec732740{0%{opacity:1;transform:translateY(0)}to{opacity:0;transform:translateY(-4px)}}.animate-response-heading .animate-response-children[data-v-ec732740]{animation:response-spans-ec732740 .2s ease-in-out forwards .05s;opacity:0}@keyframes response-spans-ec732740{0%{opacity:0;transform:translateY(4px)}to{opacity:1;transform:translateY(0)}}.schema>span[data-v-30c955c6]:not(:first-child):before{content:"·";display:block;margin:0 .5ch}.schema>span[data-v-30c955c6]{display:flex;white-space:nowrap}[data-v-ccef6266] .cm-editor{padding:0}[data-v-ccef6266] .cm-content{align-items:center;background-color:transparent;display:flex;font-family:var(--scalar-font);font-size:var(--scalar-mini);padding:6px 8px}[data-v-ccef6266] .cm-content:has(.cm-pill){padding:4px 3px}[data-v-ccef6266] .cm-content .cm-pill:not(:last-of-type){margin-right:.5px}[data-v-ccef6266] .cm-content .cm-pill:not(:first-of-type){margin-left:.5px}[data-v-ccef6266] .cm-line{padding:0}.filemask[data-v-ccef6266]{-webkit-mask-image:linear-gradient(to right,transparent 0,var(--scalar-background-2) 20px);mask-image:linear-gradient(to right,transparent 0,var(--scalar-background-2) 20px)}[data-v-3f1cd888] .cm-content{font-size:var(--scalar-mini)}.auth-blend-required[data-v-2b678dcb] .scalar-input-required,.auth-blend-required[data-v-2b678dcb] .required{background:var(--scalar-background-2);--tw-bg-base: var(--scalar-background-2);--tw-shadow: -8px 0 4px var(--scalar-background-2)}.request-example-references-header[data-v-2b678dcb]+tr>td{border-top:0;border-top-left-radius:0;border-top-right-radius:0}.auth-combobox-position[data-v-3569c0b2]{margin-left:120px}.scroll-timeline-x[data-v-3569c0b2]{overflow:auto;scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;scrollbar-width:none;-ms-overflow-style:none}.fade-left[data-v-3569c0b2],.fade-right[data-v-3569c0b2]{position:sticky;content:"";height:100%;animation-name:fadein-3569c0b2;animation-duration:1ms;animation-direction:reverse;animation-timeline:--scroll-timeline;min-height:24px;pointer-events:none}.fade-left[data-v-3569c0b2]{background:linear-gradient(-90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);min-width:3px;left:-1px;animation-direction:normal}.fade-right[data-v-3569c0b2]{background:linear-gradient(90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);margin-left:-20px;min-width:24px;right:-1px;top:0}@keyframes fadein-3569c0b2{0%{opacity:0}15%{opacity:1}}.auth-combobox-position[data-v-3dc33a43]{margin-left:120px}.scroll-timeline-x[data-v-3dc33a43]{overflow:auto;scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;scrollbar-width:none;-ms-overflow-style:none}.fade-left[data-v-3dc33a43],.fade-right[data-v-3dc33a43]{position:sticky;content:"";height:100%;animation-name:fadein-3dc33a43;animation-duration:1ms;animation-direction:reverse;animation-timeline:--scroll-timeline;min-height:24px;pointer-events:none}.fade-left[data-v-3dc33a43]{background:linear-gradient(-90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);min-width:3px;left:-1px;animation-direction:normal}.fade-right[data-v-3dc33a43]{background:linear-gradient(90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);margin-left:-20px;min-width:24px;right:-1px;top:0}@keyframes fadein-3dc33a43{0%{opacity:0}15%{opacity:1}}[data-v-c2f2c62b] code.hljs *{font-size:var(--scalar-mini)}.request-section-content[data-v-5ab558f8]{--scalar-border-width: .5px}.request-section-content-filter[data-v-5ab558f8]{box-shadow:0 -10px 0 10px var(--scalar-background-1)}.request-item:focus-within .request-meta-buttons[data-v-5ab558f8]{opacity:1}.group-hover-input[data-v-5ab558f8]{border-width:var(--scalar-border-width);border-color:transparent}.group:hover .group-hover-input[data-v-5ab558f8]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2));border-color:var(--scalar-border-color)}.group-hover-input[data-v-5ab558f8]:focus{background:transparent!important;border-color:var(--scalar-border-color)!important}@media (min-width: 800px){.has-no-import-url,.has-import-url{max-width:100dvw;overflow-x:hidden;contain:paint}.has-no-import-url{opacity:1;background:var(--scalar-background-1);animation:transform-restore-layout ease-in-out .3s forwards}.has-import-url .scalar-client>main{opacity:0;transform:scale(.85) translate3d(calc(50dvw + 80px),0,0);animation:transform-fade-layout ease-in-out .3s forwards;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:12px;overflow:hidden;z-index:10000}.has-import-url .scalar-client .sidenav{display:none}.has-no-import-url .scalar-app,.has-import-url .scalar-app{background:var(--scalar-background-1)!important}}@keyframes transform-fade-layout{0%{opacity:0;transform:scale(.85) translate3d(calc(50dvw + 80px),10px,0)}to{opacity:1;transform:scale(.85) translate3d(calc(50dvw + 80px),0,0)}}@keyframes transform-restore-layout{0%{opacity:1;transform:scale(.85) translate3d(calc(50dvw + 80px),0,0)}to{opacity:1;transform:scale(1) translateZ(0)}}.openapi-color{color:var(--scalar-color-green)}.section-flare{position:fixed;top:0;right:-50dvw}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}:where(.scalar-app) .container{width:100%}@media (min-width: 400px){:where(.scalar-app) .container{max-width:400px}}@media (min-width: 600px){:where(.scalar-app) .container{max-width:600px}}@media (min-width: 800px){:where(.scalar-app) .container{max-width:800px}}@media (min-width: 1000px){:where(.scalar-app) .container{max-width:1000px}}@media (min-width: 1200px){:where(.scalar-app) .container{max-width:1200px}}:where(.scalar-app) .sr-only{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);white-space:nowrap;border-width:0}:where(.scalar-app) .pointer-events-none{pointer-events:none}:where(.scalar-app) .visible{visibility:visible}:where(.scalar-app) .invisible{visibility:hidden}:where(.scalar-app) .fixed{position:fixed}:where(.scalar-app) .absolute{position:absolute}:where(.scalar-app) .relative{position:relative}:where(.scalar-app) .sticky{position:sticky}:where(.scalar-app) .inset-0{top:0;right:0;bottom:0;left:0}:where(.scalar-app) .left-0{left:0}:where(.scalar-app) .left-2{left:8px}:where(.scalar-app) .left-2\.5{left:10px}:where(.scalar-app) .left-4{left:16px}:where(.scalar-app) .left-border{left:var(--scalar-border-width)}:where(.scalar-app) .left-px{left:1px}:where(.scalar-app) .right-2{right:8px}:where(.scalar-app) .top-0{top:0}:where(.scalar-app) .top-1\/2{top:50%}:where(.scalar-app) .top-12{top:48px}:where(.scalar-app) .top-2{top:8px}:where(.scalar-app) .top-22{top:88px}:where(.scalar-app) .top-\[0\.5px\]{top:.5px}:where(.scalar-app) .top-px{top:1px}:where(.scalar-app) .-z-1{z-index:-1}:where(.scalar-app) .z-1{z-index:1}:where(.scalar-app) .z-\[1001\]{z-index:1001}:where(.scalar-app) .z-context{z-index:1000}:where(.scalar-app) .\!m-0{margin:0!important}:where(.scalar-app) .-m-px{margin:-1px}:where(.scalar-app) .m-0{margin:0}:where(.scalar-app) .m-auto{margin:auto}:where(.scalar-app) .-mx-0{margin-left:-0px;margin-right:-0px}:where(.scalar-app) .-mx-0\.75{margin-left:-3px;margin-right:-3px}:where(.scalar-app) .-mx-px{margin-left:-1px;margin-right:-1px}:where(.scalar-app) .-my-1{margin-top:-4px;margin-bottom:-4px}:where(.scalar-app) .-my-1\.5{margin-top:-6px;margin-bottom:-6px}:where(.scalar-app) .mx-2{margin-left:8px;margin-right:8px}:where(.scalar-app) .mx-auto{margin-left:auto;margin-right:auto}:where(.scalar-app) .mx-px{margin-left:1px;margin-right:1px}:where(.scalar-app) .my-0{margin-top:0;margin-bottom:0}:where(.scalar-app) .my-0\.75{margin-top:3px;margin-bottom:3px}:where(.scalar-app) .my-3{margin-top:12px;margin-bottom:12px}:where(.scalar-app) .-mb-1{margin-bottom:-4px}:where(.scalar-app) .-ml-0{margin-left:-0px}:where(.scalar-app) .-ml-0\.75{margin-left:-3px}:where(.scalar-app) .-mr-0{margin-right:-0px}:where(.scalar-app) .-mr-0\.25{margin-right:-1px}:where(.scalar-app) .mb-0{margin-bottom:0}:where(.scalar-app) .ml-1{margin-left:4px}:where(.scalar-app) .ml-auto{margin-left:auto}:where(.scalar-app) .mr-2{margin-right:8px}:where(.scalar-app) .mr-3{margin-right:12px}:where(.scalar-app) .mt-0{margin-top:0}:where(.scalar-app) .mt-1{margin-top:4px}:where(.scalar-app) .mt-1\.5{margin-top:6px}:where(.scalar-app) .mt-2{margin-top:8px}:where(.scalar-app) .mt-20{margin-top:80px}:where(.scalar-app) .line-clamp-4{overflow:hidden;display:-webkit-box;-webkit-box-orient:vertical;-webkit-line-clamp:4}:where(.scalar-app) .block{display:block}:where(.scalar-app) .inline-block{display:inline-block}:where(.scalar-app) .flex{display:flex}:where(.scalar-app) .table{display:table}:where(.scalar-app) .grid{display:grid}:where(.scalar-app) .contents{display:contents}:where(.scalar-app) .hidden{display:none}:where(.scalar-app) .aspect-square{aspect-ratio:1 / 1}:where(.scalar-app) .size-10{width:40px;height:40px}:where(.scalar-app) .size-2{width:8px;height:8px}:where(.scalar-app) .size-3{width:12px;height:12px}:where(.scalar-app) .size-3\.5{width:14px;height:14px}:where(.scalar-app) .size-4{width:16px;height:16px}:where(.scalar-app) .size-5{width:20px;height:20px}:where(.scalar-app) .size-6{width:24px;height:24px}:where(.scalar-app) .size-8{width:32px;height:32px}:where(.scalar-app) .size-\[23px\]{width:23px;height:23px}:where(.scalar-app) .size-full{width:100%;height:100%}:where(.scalar-app) .h-10{height:40px}:where(.scalar-app) .h-3{height:12px}:where(.scalar-app) .h-3\.5{height:14px}:where(.scalar-app) .h-32{height:128px}:where(.scalar-app) .h-4{height:16px}:where(.scalar-app) .h-5{height:20px}:where(.scalar-app) .h-6{height:24px}:where(.scalar-app) .h-8{height:32px}:where(.scalar-app) .h-\[100dvh\]{height:100dvh}:where(.scalar-app) .h-border{height:var(--scalar-border-width)}:where(.scalar-app) .h-fit{height:-moz-fit-content;height:fit-content}:where(.scalar-app) .h-full{height:100%}:where(.scalar-app) .h-screen{height:100vh}:where(.scalar-app) .max-h-\[440px\]{max-height:440px}:where(.scalar-app) .max-h-\[calc\(100dvh-120px\)\]{max-height:calc(100dvh - 120px)}:where(.scalar-app) .max-h-\[calc\(100dvh-180px\)\]{max-height:calc(100dvh - 180px)}:where(.scalar-app) .max-h-\[calc\(100dvh-240px\)\]{max-height:calc(100dvh - 240px)}:where(.scalar-app) .max-h-\[inherit\]{max-height:inherit}:where(.scalar-app) .max-h-dvh{max-height:100dvh}:where(.scalar-app) .max-h-radix-popper{max-height:calc(var(--radix-popper-available-height) - 8px)}:where(.scalar-app) .min-h-0{min-height:0px}:where(.scalar-app) .min-h-12{min-height:48px}:where(.scalar-app) .min-h-5{min-height:20px}:where(.scalar-app) .min-h-6{min-height:24px}:where(.scalar-app) .min-h-8{min-height:32px}:where(.scalar-app) .min-h-96{min-height:384px}:where(.scalar-app) .min-h-\[77px\]{min-height:77px}:where(.scalar-app) .min-h-header{min-height:48px}:where(.scalar-app) .w-12{width:48px}:where(.scalar-app) .w-16{width:64px}:where(.scalar-app) .w-20{width:80px}:where(.scalar-app) .w-24{width:96px}:where(.scalar-app) .w-3{width:12px}:where(.scalar-app) .w-32{width:128px}:where(.scalar-app) .w-4{width:16px}:where(.scalar-app) .w-40{width:160px}:where(.scalar-app) .w-48{width:192px}:where(.scalar-app) .w-5{width:20px}:where(.scalar-app) .w-56{width:224px}:where(.scalar-app) .w-6{width:24px}:where(.scalar-app) .w-64{width:256px}:where(.scalar-app) .w-72{width:288px}:where(.scalar-app) .w-8{width:32px}:where(.scalar-app) .w-\[100dvw\]{width:100dvw}:where(.scalar-app) .w-\[38px\]{width:38px}:where(.scalar-app) .w-auto{width:auto}:where(.scalar-app) .w-full{width:100%}:where(.scalar-app) .w-min{width:-moz-min-content;width:min-content}:where(.scalar-app) .min-w-0{min-width:0px}:where(.scalar-app) .min-w-5{min-width:20px}:where(.scalar-app) .min-w-6{min-width:24px}:where(.scalar-app) .min-w-8{min-width:32px}:where(.scalar-app) .min-w-max{min-width:-moz-max-content;min-width:max-content}:where(.scalar-app) .min-w-min{min-width:-moz-min-content;min-width:min-content}:where(.scalar-app) .max-w-screen-lg{max-width:800px}:where(.scalar-app) .max-w-screen-md{max-width:640px}:where(.scalar-app) .max-w-screen-sm{max-width:540px}:where(.scalar-app) .max-w-screen-xl{max-width:1000px}:where(.scalar-app) .max-w-screen-xs{max-width:480px}:where(.scalar-app) .max-w-screen-xxs{max-width:360px}:where(.scalar-app) .max-w-xs{max-width:320px}:where(.scalar-app) .flex-1{flex:1 1 0%}:where(.scalar-app) .shrink-0{flex-shrink:0}:where(.scalar-app) .origin-top-left{transform-origin:top left}:where(.scalar-app) .-translate-y-1\/2{--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .translate-x-2{--tw-translate-x: 8px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .translate-x-2\.5{--tw-translate-x: 10px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .translate-x-\[14px\]{--tw-translate-x: 14px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .rotate-90{--tw-rotate: 90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .transform{transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .cursor-auto{cursor:auto}:where(.scalar-app) .cursor-default{cursor:default}:where(.scalar-app) .cursor-not-allowed{cursor:not-allowed}:where(.scalar-app) .cursor-pointer{cursor:pointer}:where(.scalar-app) .resize-none{resize:none}:where(.scalar-app) .resize{resize:both}:where(.scalar-app) .appearance-none{-webkit-appearance:none;-moz-appearance:none;appearance:none}:where(.scalar-app) .grid-cols-\[auto_1fr\]{grid-template-columns:auto 1fr}:where(.scalar-app) .flex-col{flex-direction:column}:where(.scalar-app) .flex-wrap{flex-wrap:wrap}:where(.scalar-app) .items-start{align-items:flex-start}:where(.scalar-app) .items-center{align-items:center}:where(.scalar-app) .justify-start{justify-content:flex-start}:where(.scalar-app) .justify-end{justify-content:flex-end}:where(.scalar-app) .justify-center{justify-content:center}:where(.scalar-app) .justify-between{justify-content:space-between}:where(.scalar-app) .gap-0{gap:0px}:where(.scalar-app) .gap-0\.5{gap:2px}:where(.scalar-app) .gap-0\.75{gap:3px}:where(.scalar-app) .gap-1{gap:4px}:where(.scalar-app) .gap-1\.5{gap:6px}:where(.scalar-app) .gap-16{gap:64px}:where(.scalar-app) .gap-2{gap:8px}:where(.scalar-app) .gap-2\.5{gap:10px}:where(.scalar-app) .gap-3{gap:12px}:where(.scalar-app) .gap-4{gap:16px}:where(.scalar-app) .gap-x-4{-moz-column-gap:16px;column-gap:16px}:where(.scalar-app) .gap-y-8{row-gap:32px}:where(.scalar-app) .divide-y>:not([hidden])~:not([hidden]){--tw-divide-y-reverse: 0;border-top-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-y-reverse)));border-bottom-width:calc(var(--scalar-border-width) * var(--tw-divide-y-reverse))}:where(.scalar-app) .self-center{align-self:center}:where(.scalar-app) .overflow-auto{overflow:auto}:where(.scalar-app) .overflow-hidden{overflow:hidden}:where(.scalar-app) .overflow-visible{overflow:visible}:where(.scalar-app) .overflow-y-scroll{overflow-y:scroll}:where(.scalar-app) .overscroll-contain{overscroll-behavior:contain}:where(.scalar-app) .truncate{overflow:hidden;text-overflow:ellipsis;white-space:nowrap}:where(.scalar-app) .text-ellipsis{text-overflow:ellipsis}:where(.scalar-app) .whitespace-nowrap{white-space:nowrap}:where(.scalar-app) .text-wrap{text-wrap:wrap}:where(.scalar-app) .text-nowrap{text-wrap:nowrap}:where(.scalar-app) .rounded{border-radius:var(--scalar-radius)}:where(.scalar-app) .rounded-full{border-radius:9999px}:where(.scalar-app) .rounded-lg{border-radius:var(--scalar-radius-lg)}:where(.scalar-app) .rounded-md{border-radius:var(--scalar-radius)}:where(.scalar-app) .rounded-none{border-radius:0}:where(.scalar-app) .rounded-xl{border-radius:var(--scalar-radius-xl)}:where(.scalar-app) .rounded-t-none{border-top-left-radius:0;border-top-right-radius:0}:where(.scalar-app) .border{border-width:var(--scalar-border-width)}:where(.scalar-app) .border-0{border-width:0}:where(.scalar-app) .border-2{border-width:calc(var(--scalar-border-width) * 2)}:where(.scalar-app) .border-\[1px\]{border-width:1px}:where(.scalar-app) .border-b{border-bottom-width:var(--scalar-border-width)}:where(.scalar-app) .border-r{border-right-width:var(--scalar-border-width)}:where(.scalar-app) .border-t{border-top-width:var(--scalar-border-width)}:where(.scalar-app) .border-t-1\/2{border-top-width:calc(var(--scalar-border-width) / 2)}:where(.scalar-app) .border-solid{border-style:solid}:where(.scalar-app) .border-dashed{border-style:dashed}:where(.scalar-app) .border-none{border-style:none}:where(.scalar-app) .border-border{border-color:var(--scalar-border-color)}:where(.scalar-app) .border-c-1{border-color:var(--scalar-color-1)}:where(.scalar-app) .border-c-3{border-color:var(--scalar-color-3)}:where(.scalar-app) .border-red{border-color:var(--scalar-color-red)}:where(.scalar-app) .border-sidebar-border{border-color:var(--scalar-sidebar-border-color, var(--scalar-border-color))}:where(.scalar-app) .bg-b-1{background-color:var(--scalar-background-1)}:where(.scalar-app) .bg-b-2{background-color:var(--scalar-background-2)}:where(.scalar-app) .bg-b-3{background-color:var(--scalar-background-3)}:where(.scalar-app) .bg-b-btn{background-color:var(--scalar-button-1)}:where(.scalar-app) .bg-backdrop{background-color:#00000038}:where(.scalar-app) .bg-blue{background-color:var(--scalar-color-blue)}:where(.scalar-app) .bg-border{background-color:var(--scalar-border-color)}:where(.scalar-app) .bg-c-accent{background-color:var(--scalar-color-accent)}:where(.scalar-app) .bg-c-danger{background-color:var(--scalar-color-danger)}:where(.scalar-app) .bg-red{background-color:var(--scalar-color-red)}:where(.scalar-app) .bg-sidebar-b-1{background-color:var(--scalar-sidebar-background-1, var(--scalar-background-1))}:where(.scalar-app) .bg-transparent{background-color:transparent}:where(.scalar-app) .bg-white{--tw-bg-opacity: 1;background-color:rgb(255 255 255 / var(--tw-bg-opacity))}:where(.scalar-app) .p-0{padding:0}:where(.scalar-app) .p-0\.5{padding:2px}:where(.scalar-app) .p-0\.75{padding:3px}:where(.scalar-app) .p-1{padding:4px}:where(.scalar-app) .p-1\.5{padding:6px}:where(.scalar-app) .p-2{padding:8px}:where(.scalar-app) .p-3{padding:12px}:where(.scalar-app) .p-4{padding:16px}:where(.scalar-app) .p-6{padding:24px}:where(.scalar-app) .px-1{padding-left:4px;padding-right:4px}:where(.scalar-app) .px-1\.5{padding-left:6px;padding-right:6px}:where(.scalar-app) .px-2{padding-left:8px;padding-right:8px}:where(.scalar-app) .px-2\.5{padding-left:10px;padding-right:10px}:where(.scalar-app) .px-3{padding-left:12px;padding-right:12px}:where(.scalar-app) .px-4{padding-left:16px;padding-right:16px}:where(.scalar-app) .px-6{padding-left:24px;padding-right:24px}:where(.scalar-app) .px-9{padding-left:36px;padding-right:36px}:where(.scalar-app) .py-1{padding-top:4px;padding-bottom:4px}:where(.scalar-app) .py-1\.5{padding-top:6px;padding-bottom:6px}:where(.scalar-app) .py-2{padding-top:8px;padding-bottom:8px}:where(.scalar-app) .py-2\.5{padding-top:10px;padding-bottom:10px}:where(.scalar-app) .py-3{padding-top:12px;padding-bottom:12px}:where(.scalar-app) .py-4{padding-top:16px;padding-bottom:16px}:where(.scalar-app) .pb-0{padding-bottom:0}:where(.scalar-app) .pl-8{padding-left:32px}:where(.scalar-app) .pl-\[114px\]{padding-left:114px}:where(.scalar-app) .pl-\[24px\]{padding-left:24px}:where(.scalar-app) .pl-\[42px\]{padding-left:42px}:where(.scalar-app) .pl-\[60px\]{padding-left:60px}:where(.scalar-app) .pl-\[6px\]{padding-left:6px}:where(.scalar-app) .pl-\[78px\]{padding-left:78px}:where(.scalar-app) .pl-\[96px\]{padding-left:96px}:where(.scalar-app) .pr-3{padding-right:12px}:where(.scalar-app) .pt-3{padding-top:12px}:where(.scalar-app) .pt-6{padding-top:24px}:where(.scalar-app) .text-left{text-align:left}:where(.scalar-app) .font-code{font-family:var(--scalar-font-code)}:where(.scalar-app) .text-3xs{font-size:var(--scalar-font-size-7)}:where(.scalar-app) .text-lg{font-size:var(--scalar-font-size-2)}:where(.scalar-app) .text-sm{font-size:var(--scalar-font-size-4)}:where(.scalar-app) .text-xl{font-size:var(--scalar-font-size-1)}:where(.scalar-app) .text-xs{font-size:var(--scalar-font-size-5)}:where(.scalar-app) .text-xxs{font-size:var(--scalar-font-size-6)}:where(.scalar-app) .font-bold{font-weight:var(--scalar-bold)}:where(.scalar-app) .font-medium{font-weight:var(--scalar-semibold)}:where(.scalar-app) .font-normal{font-weight:var(--scalar-regular)}:where(.scalar-app) .leading-3{line-height:var(--scalar-line-height-3)}:where(.scalar-app) .leading-none{line-height:1}:where(.scalar-app) .leading-snug{line-height:1.375}:where(.scalar-app) .text-b-1{color:var(--scalar-background-1)}:where(.scalar-app) .text-blue{color:var(--scalar-color-blue)}:where(.scalar-app) .text-c-1{color:var(--scalar-color-1)}:where(.scalar-app) .text-c-2{color:var(--scalar-color-2)}:where(.scalar-app) .text-c-3{color:var(--scalar-color-3)}:where(.scalar-app) .text-c-accent{color:var(--scalar-color-accent)}:where(.scalar-app) .text-c-btn{color:var(--scalar-button-1-color)}:where(.scalar-app) .text-c-danger{color:var(--scalar-color-danger)}:where(.scalar-app) .text-c-ghost{color:var(--scalar-color-ghost)}:where(.scalar-app) .text-red{color:var(--scalar-color-red)}:where(.scalar-app) .text-sidebar-c-2{color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}:where(.scalar-app) .text-transparent{color:transparent}:where(.scalar-app) .text-white{--tw-text-opacity: 1;color:rgb(255 255 255 / var(--tw-text-opacity))}:where(.scalar-app) .no-underline{text-decoration-line:none}:where(.scalar-app) .opacity-0{opacity:0}:where(.scalar-app) .opacity-40{opacity:.4}:where(.scalar-app) .opacity-50{opacity:.5}:where(.scalar-app) .shadow{--tw-shadow: var(--scalar-shadow-1);--tw-shadow-colored: var(--scalar-shadow-1);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-border{--tw-shadow: inset 0 0 0 1px var(--scalar-border-color);--tw-shadow-colored: inset 0 0 0 1px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-lg{--tw-shadow: var(--scalar-shadow-2);--tw-shadow-colored: var(--scalar-shadow-2);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-none{--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-sm{--tw-shadow: rgba(0, 0, 0, .09) 0px 1px 4px;--tw-shadow-colored: 0px 1px 4px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-border{--tw-shadow-color: var(--scalar-border-color);--tw-shadow: var(--tw-shadow-colored)}:where(.scalar-app) .shadow-current{--tw-shadow-color: currentColor;--tw-shadow: var(--tw-shadow-colored)}:where(.scalar-app) .outline-none{outline:2px solid transparent;outline-offset:2px}:where(.scalar-app) .outline{outline-style:solid}:where(.scalar-app) .-outline-offset-1{outline-offset:-1px}:where(.scalar-app) .outline-offset-1{outline-offset:1px}:where(.scalar-app) .blur{--tw-blur: blur(8px);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .brightness-lifted{--tw-brightness: brightness(var(--scalar-lifted-brightness));filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .filter{filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .backdrop-blur{--tw-backdrop-blur: blur(8px);-webkit-backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia);backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia)}:where(.scalar-app) .transition{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,-webkit-backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter,-webkit-backdrop-filter;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .transition-colors{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .transition-opacity{transition-property:opacity;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .transition-transform{transition-property:transform;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .duration-100{transition-duration:.1s}:where(.scalar-app) .duration-300{transition-duration:.3s}:where(.scalar-app) .ease-in-out{transition-timing-function:cubic-bezier(.4,0,.2,1)}:where(.scalar-app) .bg-mix-transparent{background-color:color-mix(var(--tw-bg-mix-method, in srgb),transparent calc(var(--tw-bg-mix-amount, 0) * 1%),var(--tw-bg-base))}:where(.scalar-app) .bg-mix-amount-25{--tw-bg-mix-amount: 25}:where(.scalar-app) .bg-b-1{--tw-bg-base: var(--scalar-background-1)}:where(.scalar-app) .bg-b-2{--tw-bg-base: var(--scalar-background-2)}:where(.scalar-app) .bg-b-3{--tw-bg-base: var(--scalar-background-3)}:where(.scalar-app) .bg-b-btn{--tw-bg-base: var(--scalar-button-1)}:where(.scalar-app) .bg-backdrop{--tw-bg-base: rgba(0, 0, 0, .22)}:where(.scalar-app) .bg-blue{--tw-bg-base: var(--scalar-color-blue)}:where(.scalar-app) .bg-border{--tw-bg-base: var(--scalar-border-color)}:where(.scalar-app) .bg-c-accent{--tw-bg-base: var(--scalar-color-accent)}:where(.scalar-app) .bg-c-danger{--tw-bg-base: var(--scalar-color-danger)}:where(.scalar-app) .bg-red{--tw-bg-base: var(--scalar-color-red)}:where(.scalar-app) .bg-sidebar-b-1{--tw-bg-base: var(--scalar-sidebar-background-1, var(--scalar-background-1))}:where(.scalar-app) .bg-transparent{--tw-bg-base: transparent}:where(.scalar-app) .bg-white{--tw-bg-opacity: 1;--tw-bg-base: rgb(255 255 255 / var(--tw-bg-opacity))}:where(.scalar-app) .row,:where(.scalar-app) .scalar-row{display:flex;flex-direction:row}:where(.scalar-app) .col{display:flex;flex-direction:column}:where(.scalar-app) .centered-x{left:50%;--tw-translate-x: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .centered{left:50%;--tw-translate-x: -50%;top:50%;--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .\*\:flex>*{display:flex}:where(.scalar-app) .\*\:flex-1>*{flex:1 1 0%}:where(.scalar-app) .\*\:items-center>*{align-items:center}:where(.scalar-app) .\*\:gap-px>*{gap:1px}:where(.scalar-app) .hover\:bg-b-2:hover{background-color:var(--scalar-background-2)}:where(.scalar-app) .hover\:bg-b-3:hover{background-color:var(--scalar-background-3)}:where(.scalar-app) .hover\:bg-h-btn:hover{background-color:var(--scalar-button-1-hover)}:where(.scalar-app) .hover\:text-c-1:hover{color:var(--scalar-color-1)}:where(.scalar-app) .hover\:underline:hover{text-decoration-line:underline}:where(.scalar-app) .hover\:bg-b-2:hover{--tw-bg-base: var(--scalar-background-2)}:where(.scalar-app) .hover\:bg-b-3:hover{--tw-bg-base: var(--scalar-background-3)}:where(.scalar-app) .hover\:bg-h-btn:hover{--tw-bg-base: var(--scalar-button-1-hover)}:where(.scalar-app) .focus\:outline-none:focus{outline:2px solid transparent;outline-offset:2px}:where(.scalar-app) .focus-visible\:border-c-btn:focus-visible{border-color:var(--scalar-button-1-color)}:where(.scalar-app) .focus-visible\:ring-1:focus-visible{--tw-ring-offset-shadow: var(--tw-ring-inset) 0 0 0 var(--tw-ring-offset-width) var(--tw-ring-offset-color);--tw-ring-shadow: var(--tw-ring-inset) 0 0 0 calc(1px + var(--tw-ring-offset-width)) var(--tw-ring-color);box-shadow:var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow, 0 0 #0000)}:where(.scalar-app) .focus-visible\:ring-c-accent:focus-visible{--tw-ring-color: var(--scalar-color-accent)}:where(.scalar-app) .active\:bg-b-btn:active{background-color:var(--scalar-button-1)}:where(.scalar-app) .active\:text-c-1:active{color:var(--scalar-color-1)}:where(.scalar-app) .active\:shadow-none:active{--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .active\:brightness-90:active{--tw-brightness: brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .active\:bg-b-btn:active{--tw-bg-base: var(--scalar-button-1)}:where(.scalar-app) .group\/button:hover .group-hover\/button\:text-c-1,:where(.scalar-app) .group:hover .group-hover\:text-c-1{color:var(--scalar-color-1)}:where(.scalar-app) .group\/item:hover .group-hover\/item\:opacity-100{opacity:1}:where(.scalar-app) .group\/toggle:focus-visible .group-focus-visible\/toggle\:outline{outline-style:solid}:where(.scalar-app) .peer:checked~.peer-checked\:text-c-2{color:var(--scalar-color-2)}:where(.scalar-app) .peer:checked~.peer-checked\:opacity-100{opacity:1}:where(.scalar-app) .has-\[\:focus-visible\]\:bg-b-1:has(:focus-visible){background-color:var(--scalar-background-1)}:where(.scalar-app) .has-\[\:focus-visible\]\:outline:has(:focus-visible){outline-style:solid}:where(.scalar-app) .has-\[input\:focus-visible\]\:outline:has(input:focus-visible){outline-style:solid}:where(.scalar-app) .has-\[\:focus-visible\]\:bg-b-1:has(:focus-visible){--tw-bg-base: var(--scalar-background-1)}:where(.scalar-app) .ui-open\:rotate-180[data-headlessui-state~=open]{--tw-rotate: 180deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) :where([data-headlessui-state~=open]) .ui-open\:rotate-180{--tw-rotate: 180deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .has-actv-btn\:border:has(button:active){border-width:var(--scalar-border-width)}:where(.scalar-app) .has-actv-btn\:border-border:has(button:active){border-color:var(--scalar-border-color)}:where(.scalar-app) .\!empty\:flex:not(:empty){display:flex}:where(.scalar-app) .\!empty\:w-7:not(:empty){width:28px}:where(.scalar-app) .\!empty\:items-center:not(:empty){align-items:center}:where(.scalar-app) .\!empty\:pr-3:not(:empty){padding-right:12px}:where(.scalar-app) .hocus\:text-c-1:hover{color:var(--scalar-color-1)}:where(.scalar-app) .hocus\:brightness-90:hover{--tw-brightness: brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .hocus\:text-c-1:focus-visible{color:var(--scalar-color-1)}:where(.scalar-app) .hocus\:brightness-90:focus-visible{--tw-brightness: brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}@media (min-width: 800px){:where(.scalar-app) .md\:w-\[calc\(100vw-16px\)\]{width:calc(100vw - 16px)}:where(.scalar-app) .md\:row{display:flex;flex-direction:row}}@media (min-width: 1000px){:where(.scalar-app) .lg\:w-\[calc\(100vw-32px\)\]{width:calc(100vw - 32px)}:where(.scalar-app) .lg\:w-full{width:100%}}:where(.scalar-app) .dark\:bg-b-3:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-3)}:where(.scalar-app) .dark\:bg-backdropdark:where(.dark-mode,.dark-mode *){background-color:#00000073}:where(.scalar-app) .dark\:bg-b-3:where(.dark-mode,.dark-mode *){--tw-bg-base: var(--scalar-background-3)}:where(.scalar-app) .dark\:bg-backdropdark:where(.dark-mode,.dark-mode *){--tw-bg-base: rgba(0, 0, 0, .45)}:where(.scalar-app) .dark\:hover\:bg-b-3:hover:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-3);--tw-bg-base: var(--scalar-background-3)}.loader-wrapper[data-v-65265e87]{position:relative;display:flex;align-items:center;justify-content:center;--loader-size: 50%}.svg-loader[data-v-65265e87]{width:var(--loader-size);height:var(--loader-size);top:1rem;right:.9rem;overflow:visible;fill:none;background-color:transparent;stroke:currentColor}.svg-path[data-v-65265e87]{stroke-width:14;fill:none;transition:.3s}.svg-x-mark[data-v-65265e87]{stroke-dasharray:57;stroke-dashoffset:57;transition-delay:0s}.svg-check-mark[data-v-65265e87]{stroke-dasharray:149;stroke-dashoffset:149;transition-delay:0s}.icon-is-invalid .svg-x-mark[data-v-65265e87],.icon-is-valid .svg-check-mark[data-v-65265e87]{stroke-dashoffset:0;transition-delay:.3s}.circular-loader[data-v-65265e87]{animation:rotate-65265e87 .7s linear infinite,fade-in-65265e87 .4s;transform-origin:center center;transform:scale(5);background:transparent}.loader-path[data-v-65265e87]{stroke-dasharray:50,200;stroke-dashoffset:-100;stroke-linecap:round}.loader-path-off[data-v-65265e87]{stroke-dasharray:50,200;stroke-dashoffset:-100;transition:opacity .3s;opacity:0}@keyframes fade-in-65265e87{0%{opacity:0}70%{opacity:0}to{opacity:1}}@keyframes rotate-65265e87{0%{transform:scale(5) rotate(0)}to{transform:scale(5) rotate(360deg)}}.scalar-icon[data-v-f4f53d32],.scalar-icon[data-v-f4f53d32] *{stroke-width:var(--ce5c5c8a)}.scalar-code-block{background:inherit;position:relative;overflow:auto;padding:.75rem}.scalar-code-block:hover .copy-button,.copy-button:focus-visible{opacity:100}.scalar-codeblock-pre{all:unset;margin:0;background:transparent;text-wrap:nowrap;white-space-collapse:preserve;border-radius:0;width:-moz-fit-content;width:fit-content}.scalar-code-copy{display:flex;align-items:flex-start;justify-content:flex-end;position:sticky;top:0;right:0;bottom:0;left:0}.copy-button{position:relative;top:0;right:0;display:flex;align-items:center;background-color:var(--scalar-background-2);border:1px solid var(--scalar-border-color);border-radius:3px;color:var(--scalar-color-3);cursor:pointer;height:30px;margin-bottom:-30px;opacity:0;padding:6px;transition:opacity .15s ease-in-out,color .15s ease-in-out}.scalar-code-copy,.copy-button{background:inherit}.copy-button:hover{color:var(--scalar-color-1)}.copy-button svg{stroke-width:1.5}.toggle-icon-ellipse[data-v-8bd002b1]{width:7px;height:7px;border-radius:50%;box-shadow:inset 0 0 0 1px currentColor;overflow:hidden;position:relative;background:var(--scalar-background-1);display:inline-block;transition:width .3s ease-in-out,height .3s ease-in-out}.toggle-icon-moon-mask[data-v-8bd002b1]{width:100%;height:100%;border:1px solid currentColor;display:block;left:2.5px;bottom:2.5px;position:absolute;border-radius:50%;background:var(--scalar-background-1);transition:transform .3s ease-in-out;transform:translate3d(4px,-4px,0)}.toggle-icon-sun-ray[data-v-8bd002b1]{width:12px;height:1px;border-radius:8px;background:currentColor;position:absolute;transition:transform .3s ease-in-out}.toggle-icon-sun-ray[data-v-8bd002b1]:nth-of-type(2){transform:rotate(90deg)}.toggle-icon-sun-ray[data-v-8bd002b1]:nth-of-type(3){transform:rotate(45deg)}.toggle-icon-sun-ray[data-v-8bd002b1]:nth-of-type(4){transform:rotate(-45deg)}.toggle-icon-dark .toggle-icon-ellipse[data-v-8bd002b1]{width:10px;height:10px;-webkit-mask-image:radial-gradient(circle at bottom left,pink 10px,transparent 12px);mask-image:radial-gradient(circle at bottom left,pink 10px,transparent 12px)}.toggle-icon-dark .toggle-icon-sun-ray[data-v-8bd002b1]{transform:scale(0)}.toggle-icon-dark .toggle-icon-moon-mask[data-v-8bd002b1]{transform:translateZ(0)}.scalar-app code.hljs *{font-size:var(--scalar-small);font-family:var(--scalar-font-code);text-align:left;white-space:pre;word-spacing:normal;word-break:normal;word-wrap:normal;line-height:1.4;-moz-tab-size:4;-o-tab-size:4;tab-size:4}.scalar-app code.hljs{all:unset;font-size:var(--scalar-small);color:var(--scalar-color-2);font-family:var(--scalar-font-code);display:inline-block;counter-reset:linenumber}.scalar-app .hljs{background:transparent;color:var(--scalar-color-2)}.scalar-app .hljs .line:before{color:var(--scalar-color-3);display:inline-block;counter-increment:linenumber;content:counter(linenumber);margin-right:.7em;min-width:calc(var(--line-digits) * 1ch);text-align:right}.scalar-app .hljs-comment,.scalar-app .hljs-quote{color:var(--scalar-color-3);font-style:italic}.scalar-app .hljs-number{color:var(--scalar-color-orange)}.scalar-app .hljs-regexp,.scalar-app .hljs-string,.scalar-app .hljs-built_in{color:var(--scalar-color-blue)}.scalar-app .hljs-title.class_{color:var(--scalar-color-1)}.scalar-app .hljs-keyword{color:var(--scalar-color-purple)}.scalar-app .hljs-title.function_{color:var(--scalar-color-orange)}.scalar-app .hljs-subst,.scalar-app .hljs-name{color:var(--scalar-color-blue)}.scalar-app .hljs-attr,.scalar-app .hljs-attribute{color:var(--scalar-color-1)}.scalar-app .hljs-addition,.scalar-app .hljs-literal,.scalar-app .hljs-selector-tag,.scalar-app .hljs-type{color:var(--scalar-color-green)}.scalar-app .hljs-selector-attr,.scalar-app .hljs-selector-pseudo{color:var(--scalar-color-orange)}.scalar-app .hljs-doctag,.scalar-app .hljs-section,.scalar-app .hljs-title{color:var(--scalar-color-blue)}.scalar-app .hljs-selector-id,.scalar-app .hljs-template-variable,.scalar-app .hljs-variable{color:var(--scalar-color-1)}.scalar-app .hljs-name,.scalar-app .hljs-section,.scalar-app .hljs-strong{font-weight:var(--scalar-semibold)}.scalar-app .hljs-bullet,.scalar-app .hljs-link,.scalar-app .hljs-meta,.scalar-app .hljs-symbol{color:var(--scalar-color-blue)}.scalar-app .hljs-deletion{color:var(--scalar-color-red)}.scalar-app .hljs-formula{background:var(--scalar-color-1)}.scalar-app .hljs-emphasis{font-style:italic}.scalar-app .credentials{font-size:0!important;color:transparent}.scalar-app .credentials:after{content:"·····";font-size:var(--scalar-small);color:var(--scalar-color-3);-webkit-user-select:none;-moz-user-select:none;user-select:none}.hljs.language-html{color:var(--scalar-color-1)}.hljs.language-html .hljs-attr{color:var(--scalar-color-2)}.hljs.language-curl .hljs-keyword{color:var(--scalar-color-orange)}.hljs.language-curl .hljs-string{color:var(--scalar-color-1)}.hljs.language-curl .hljs-literal,.hljs.language-php .hljs-variable{color:var(--scalar-color-blue)}.hljs.language-objectivec .hljs-meta{color:var(--scalar-color-1)}.hljs.language-objectivec .hljs-built_in,.hljs-built_in{color:var(--scalar-color-orange)}.scalar-app .markdown{font-family:var(--scalar-font);color:var(--scalar-color-1);word-break:break-word}.scalar-app .markdown h1,.scalar-app .markdown h2,.scalar-app .markdown h3,.scalar-app .markdown h4,.scalar-app .markdown h5,.scalar-app .markdown h6,.scalar-app .markdown p,.scalar-app .markdown div,.scalar-app .markdown img,.scalar-app .markdown details,.scalar-app .markdown summary,.scalar-app .markdown ul,.scalar-app .markdown ol,.scalar-app .markdown table,.scalar-app .markdown blockquote,.scalar-app .markdown code{margin:12px 0}.scalar-app .markdown details{margin:12px 0;color:var(--scalar-color-1)}.scalar-app .markdown summary{display:block;margin:1em 0;padding-left:20px;position:relative;font-weight:var(--scalar-semibold);cursor:pointer;-webkit-user-select:none;-moz-user-select:none;user-select:none}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>");mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>")}.scalar-app .markdown summary:hover:after{background-color:var(--scalar-color-1)}.scalar-app .markdown details[open] summary:after{transform:rotate(90deg)}.scalar-app .markdown summary::-webkit-details-marker{display:none}.scalar-app .markdown img{overflow:hidden;border-radius:var(--scalar-radius);max-width:100%}.scalar-app .markdown>:first-child{margin-top:0;margin-bottom:0}.scalar-app .markdown h1{--font-size: 1.4em}.scalar-app .markdown h2{--font-size: 1.25em}.scalar-app .markdown h3{--font-size: 1.1em}.scalar-app .markdown h4,.scalar-app .markdown h6{--font-size: 1em}.scalar-app .markdown h1,.scalar-app .markdown h2,.scalar-app .markdown h3,.scalar-app .markdown h4,.scalar-app .markdown h5,.scalar-app .markdown h6{font-size:var(--font-size);margin:18px 0 6px;font-weight:var(--scalar-bold);display:block;line-height:1.45}.scalar-app .markdown b,.scalar-app .markdown strong{font-weight:var(--scalar-bold)}.scalar-app .markdown p{color:inherit;font-weight:var(--font-weight, var(--scalar-regular));line-height:1.5;margin-bottom:0;display:block}.scalar-app .markdown ul,.scalar-app .markdown ol{padding-left:24px;line-height:1.5;margin:12px 0;display:block}.scalar-app .markdown ul{list-style:disc}.scalar-app .markdown ol{list-style:decimal}.scalar-app .markdown ul.contains-task-list{list-style:none;padding-left:0}.scalar-app .markdown li{margin:6px 0;display:list-item}.scalar-app .markdown ul:first-of-type li:first-of-type{margin-top:0}.scalar-app .markdown a{color:var(--scalar-link-color, var(--scalar-color-accent));font-weight:var(--scalar-link-font-weight, inherit);-webkit-text-decoration:var(--scalar-text-decoration);text-decoration:var(--scalar-text-decoration);cursor:pointer}.scalar-app .markdown a:hover{color:var(--scalar-link-color-hover, var(--scalar-color-accent));-webkit-text-decoration:var(--scalar-text-decoration-hover);text-decoration:var(--scalar-text-decoration-hover)}.scalar-app .markdown em{font-style:italic}.scalar-app .markdown sup{font-size:var(--scalar-micro);vertical-align:super;font-weight:450}.scalar-app .markdown sub{font-size:var(--scalar-micro);vertical-align:sub;font-weight:450}.scalar-app .markdown del{text-decoration:line-through}.scalar-app .markdown code{font-family:var(--scalar-font-code);background-color:var(--scalar-background-2);box-shadow:0 0 0 var(--scalar-border-width) var(--scalar-border-color);font-size:var(--scalar-micro);border-radius:2px;padding:0 3px}.scalar-app .markdown .hljs{font-size:var(--scalar-small)}.scalar-app .markdown pre code{display:block;white-space:pre;padding:12px;line-height:1.5;margin:12px 0;-webkit-overflow-scrolling:touch;overflow-x:auto;max-width:100%;min-width:100px}.scalar-app .markdown hr{border:none;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.scalar-app .markdown blockquote{border-left:3px solid var(--scalar-border-color);padding-left:12px;margin:0;display:block}.scalar-app .markdown table{display:block;overflow-x:auto;position:relative;border-collapse:collapse;width:-moz-max-content;width:max-content;max-width:100%;margin:1em 0;box-shadow:0 0 0 var(--scalar-border-width) var(--scalar-border-color);border-radius:var(--scalar-radius-lg)}.scalar-app .markdown tbody{display:table-row-group;vertical-align:middle}.scalar-app .markdown thead{display:table-header-group;vertical-align:middle}.scalar-app .markdown tr{display:table-row;border-color:inherit;vertical-align:inherit}.scalar-app .markdown td,.scalar-app .markdown th{display:table-cell;vertical-align:inherit;min-width:1em;padding:6px 9px;vertical-align:top;line-height:1.5;position:relative;word-break:initial;font-size:var(--scalar-small);color:var(--scalar-color-1);font-weight:var(--font-weight, var(--scalar-regular));border-right:var(--scalar-border-width) solid var(--scalar-border-color);border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.scalar-app .markdown td>*,.scalar-app .markdown th>*{margin-bottom:0}.scalar-app .markdown th:empty{display:none}.scalar-app .markdown td:first-of-type,.scalar-app .markdown th:first-of-type{border-left:none}.scalar-app .markdown td:last-of-type,.scalar-app .markdown th:last-of-type{border-right:none}.scalar-app .markdown tr:last-of-type td{border-bottom:none}.scalar-app .markdown th{font-weight:var(--scalar-semibold)!important;text-align:left;border-left-color:transparent;background:var(--scalar-background-2)}.scalar-app .markdown tr>[align=left]{text-align:left}.scalar-app .markdown tr>[align=right]{text-align:right}.scalar-app .markdown tr>[align=center]{text-align:center}.dark-mode .scalar-dropdown-item[data-v-942bce7e]:hover,.dark-mode .scalar-dropdown-item[data-v-457c1cd3]:hover{filter:brightness(1.1)}.scalar-modal-layout[data-v-0f8e0211]{animation:fadein-layout-0f8e0211 ease-in-out .3s forwards}.scalar-modal[data-v-0f8e0211]{animation:fadein-modal-0f8e0211 ease-in-out .3s forwards;animation-delay:.1s;position:fixed;left:0;top:0;right:0;box-shadow:var(--scalar-shadow-2);transform:translate3d(0,10px,0)}.scalar-modal-layout-full[data-v-0f8e0211]{opacity:1!important;background:transparent!important}.dark-mode .scalar-modal[data-v-0f8e0211]{background-color:color-mix(in srgb,var(--scalar-background-1),black)}.scalar-modal.scalar-modal-search[data-v-0f8e0211]{max-width:540px;max-height:440px;background-color:transparent}.modal-content-search .modal-body[data-v-0f8e0211]{padding:0;overflow:hidden;display:flex;flex-direction:column;max-height:440px}@media (max-width: 1280px){.scalar-modal[data-v-0f8e0211]{max-height:calc(100% - 56px);top:28px}}@keyframes fadein-layout-0f8e0211{0%{opacity:0}to{opacity:1}}@keyframes fadein-modal-0f8e0211{0%{opacity:0;transform:translate3d(0,10px,0)}to{opacity:1;transform:translateZ(0)}}.full-size-styles[data-v-0f8e0211]{transform:translateZ(0);animation:fadein-layout-0f8e0211 ease-in-out .3s forwards;max-height:100%!important;top:0!important;left:0;position:absolute!important;margin:initial;border-radius:0!important;background-color:var(--scalar-background-1)!important;box-shadow:none!important;border-right:var(--scalar-border-width) solid var(--scalar-border-color)}@media (min-width: 800px){.full-size-styles[data-v-0f8e0211]{width:50dvw!important}}.full-size-styles[data-v-0f8e0211]:after{content:"";width:50dvw;height:100dvh;position:absolute;right:-50dvw;top:0}.search-background[data-v-2f2316c6]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2))}.search-background[data-v-2f2316c6]:focus-within{background:transparent}.scalar-input-wrapper-focus .scalar-input-label[data-v-4de7a24e]{color:var(--scalar-color-1)}.scalar-input-wrapper-error .scalar-input-label[data-v-4de7a24e]{color:var(--scalar-color-red)}.scalar-input[data-v-4de7a24e]::-moz-selection{color:var(--scalar-color-1);background:#ffa55859}.scalar-input[data-v-4de7a24e]::selection{color:var(--scalar-color-1);background:#ffa55859}.scalar-input[data-v-4de7a24e]:has(+.scalar-input-label){opacity:0}.scalar-input[data-v-4de7a24e]:not(:-moz-placeholder-shown){opacity:1;-moz-transition:opacity .2s ease-in-out .15s;transition:opacity .2s ease-in-out .15s}.scalar-input[data-v-4de7a24e]:not(:placeholder-shown),.scalar-input-wrapper-focus .scalar-input[data-v-4de7a24e]{opacity:1;transition:opacity .2s ease-in-out .15s}.scalar-input[data-v-4de7a24e]:-webkit-autofill,.scalar-input[data-v-4de7a24e]:-webkit-autofill:hover,.scalar-input[data-v-4de7a24e]:-webkit-autofill:focus,.scalar-input[data-v-4de7a24e]:-webkit-autofill:active,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill:hover,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill:focus,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill:active{-webkit-box-shadow:0 0 0px 1000px var(--scalar-background-1) inset!important;-webkit-text-fill-color:var(--scalar-color-1);color:var(--scalar-color-1);border-radius:var(--scalar-radius)}.scalar-input:not(:-moz-placeholder-shown)+.scalar-input-label[data-v-4de7a24e]{transform:translate3d(0,-20px,0) scale(.8);transform-origin:top left}.scalar-input-wrapper-focus .scalar-input-label[data-v-4de7a24e],.scalar-input:not(:placeholder-shown)+.scalar-input-label[data-v-4de7a24e]{transform:translate3d(0,-20px,0) scale(.8);transform-origin:top left}.scalar-input-wrapper-focus:has(button:active) .scalar-input-label[data-v-4de7a24e]{color:var(--scalar-color-3)!important}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}}:where(.scalar-app){font-family:var(--scalar-font);line-height:1.15;-webkit-text-size-adjust:100%;-moz-tab-size:4;-o-tab-size:4;tab-size:4}:where(.scalar-app) *,:where(.scalar-app) *:before,:where(.scalar-app) *:after{box-sizing:border-box;border-width:0;border-style:solid;border-color:var(--scalar-border-color);outline-width:1px;outline-style:none;outline-color:var(--scalar-color-accent);font-family:inherit;font-feature-settings:inherit;font-variation-settings:inherit;font-size:inherit;font-weight:inherit;line-height:inherit;color:inherit;margin:unset;padding:unset;text-rendering:optimizeLegibility;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale}:where(.scalar-app) :before,:where(.scalar-app) :after{--tw-content: ""}:where(.scalar-app) button,:where(.scalar-app) input,:where(.scalar-app) optgroup,:where(.scalar-app) select,:where(.scalar-app) textarea,:where(.scalar-app) ::file-selector-button{background:transparent}:where(.scalar-app) ol,:where(.scalar-app) ul,:where(.scalar-app) menu{list-style:none}:where(.scalar-app) input:where(:not([type=button],[type=reset],[type=submit])),:where(.scalar-app) select,:where(.scalar-app) textarea{border-radius:var(--scalar-radius);border-width:1px}:where(.scalar-app) input::-moz-placeholder{color:var(--scalar-color-3);font-family:var(--scalar-font)}:where(.scalar-app) input::placeholder{color:var(--scalar-color-3);font-family:var(--scalar-font)}:where(.scalar-app) input[type=search]::-webkit-search-cancel-button,:where(.scalar-app) input[type=search]::-webkit-search-decoration{-webkit-appearance:none;-moz-appearance:none;appearance:none}:where(.scalar-app) input:-webkit-autofill{-webkit-background-clip:text!important;background-clip:text!important}:where(.scalar-app) :focus-visible{outline-style:solid;border-radius:var(--scalar-radius)}:where(.scalar-app) button:focus-visible,:where(.scalar-app) [role=button]:focus-visible{outline-offset:-1px}:where(.scalar-app) button,:where(.scalar-app) [role=button]{cursor:pointer}:where(.scalar-app) :disabled{cursor:default}:where(.scalar-app) img,:where(.scalar-app) svg,:where(.scalar-app) video,:where(.scalar-app) canvas,:where(.scalar-app) audio,:where(.scalar-app) iframe,:where(.scalar-app) embed,:where(.scalar-app) object{display:block;vertical-align:middle}:where(.scalar-app) [hidden]{display:none}.scalar-app .cm-scroller,.scalar-app .custom-scroll{overflow-y:auto;scrollbar-color:transparent transparent;scrollbar-width:thin;-webkit-overflow-scrolling:touch}.scalar-app .custom-scroll-self-contain-overflow{overscroll-behavior:contain}.scalar-app .cm-scroller:hover,.scalar-app .custom-scroll:hover,.scalar-app.scalar-scrollbars-obtrusive .cm-scroller,.scalar-app.scalar-scrollbars-obtrusive .custom-scroll{scrollbar-color:var(--scalar-scrollbar-color, transparent) transparent}.scalar-app .cm-scroller:hover::-webkit-scrollbar-thumb,.scalar-app .custom-scroll:hover::-webkit-scrollbar-thumb{background:var(--scalar-scrollbar-color);background-clip:content-box;border:3px solid transparent}.scalar-app .cm-scroller::-webkit-scrollbar-thumb:active,.scalar-app .custom-scroll::-webkit-scrollbar-thumb:active{background:var(--scalar-scrollbar-color-active);background-clip:content-box;border:3px solid transparent}.scalar-app .cm-scroller::-webkit-scrollbar-corner,.scalar-app .custom-scroll::-webkit-scrollbar-corner{background:transparent}.scalar-app .cm-scroller::-webkit-scrollbar,.scalar-app .custom-scroll::-webkit-scrollbar{height:12px;width:12px}.scalar-app .cm-scroller::-webkit-scrollbar-track,.scalar-app .custom-scroll::-webkit-scrollbar-track{background:transparent}.scalar-app .cm-scroller::-webkit-scrollbar-thumb,.scalar-app .custom-scroll::-webkit-scrollbar-thumb{border-radius:20px;background:transparent;background-clip:content-box;border:3px solid transparent}@media (pointer: coarse){.scalar-app .cm-scroller,.scalar-app .custom-scroll{padding-right:12px}}@layer scalar-base{:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-client{line-height:1.5;color:var(--scalar-color-1)}.scalar-client .container{width:100%}@media (min-width: 400px){.scalar-client .container{max-width:400px}}@media (min-width: 600px){.scalar-client .container{max-width:600px}}@media (min-width: 800px){.scalar-client .container{max-width:800px}}@media (min-width: 1000px){.scalar-client .container{max-width:1000px}}@media (min-width: 1200px){.scalar-client .container{max-width:1200px}}.scalar-client .row{display:flex;flex-direction:row}.scalar-client .row-1{display:flex;flex-direction:row;flex:1 1 0%}.scalar-client .col{display:flex;flex-direction:column}.scalar-client .col-1{display:flex;flex-direction:column;flex:1 1 0%}.scalar-client .flex-center{display:flex;justify-content:center;align-items:center}.scalar-client .centered-x{left:50%;--tw-translate-x: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .centered-y{top:50%;--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .centered{left:50%;--tw-translate-x: -50%;top:50%;--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .floating-bg:before{background-color:var(--scalar-background-2);border-radius:var(--scalar-radius);content:"";height:calc(100% - 4px);left:-4px;opacity:0;position:absolute;top:2.5px;transition:opacity .2s ease-in-out;width:calc(100% + 8px);z-index:1}.scalar-client .floating-bg:hover:before{opacity:1}.scalar-client .sr-only{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);white-space:nowrap;border-width:0}.scalar-client .not-sr-only{position:static;width:auto;height:auto;padding:0;margin:0;overflow:visible;clip:auto;white-space:normal}.scalar-client .pointer-events-none{pointer-events:none}.scalar-client .pointer-events-auto{pointer-events:auto}.scalar-client .visible{visibility:visible}.scalar-client .static{position:static}.scalar-client .fixed{position:fixed}.scalar-client .absolute{position:absolute}.scalar-client .relative{position:relative}.scalar-client .sticky{position:sticky}.scalar-client .inset-0{top:0;right:0;bottom:0;left:0}.scalar-client .-inset-x-96{left:-384px;right:-384px}.scalar-client .inset-x-1{left:4px;right:4px}.scalar-client .-right-\[30px\]{right:-30px}.scalar-client .-top-\[--scalar-address-bar-height\]{top:calc(var(--scalar-address-bar-height) * -1)}.scalar-client .bottom-0{bottom:0}.scalar-client .bottom-1\/2{bottom:50%}.scalar-client .bottom-\[var\(--scalar-border-width\)\]{bottom:var(--scalar-border-width)}.scalar-client .left-0{left:0}.scalar-client .left-1\/2{left:50%}.scalar-client .right-0{right:0}.scalar-client .right-1{right:4px}.scalar-client .right-1\/2{right:50%}.scalar-client .right-14{right:56px}.scalar-client .right-16{right:64px}.scalar-client .right-2{right:8px}.scalar-client .right-4{right:16px}.scalar-client .right-7{right:28px}.scalar-client .top-0{top:0}.scalar-client .top-1\/2{top:50%}.scalar-client .top-12{top:48px}.scalar-client .top-2{top:8px}.scalar-client .-z-1{z-index:-1}.scalar-client .z-0{z-index:0}.scalar-client .z-1{z-index:1}.scalar-client .z-10{z-index:10}.scalar-client .z-20{z-index:20}.scalar-client .z-50{z-index:50}.scalar-client .z-\[10001\]{z-index:10001}.scalar-client .z-\[1\]{z-index:1}.scalar-client .z-context{z-index:1000}.scalar-client .z-context-plus{z-index:1001}.scalar-client .z-overlay{z-index:10000}.scalar-client .order-last{order:9999}.scalar-client .col-span-full{grid-column:1 / -1}.scalar-client .\!m-0{margin:0!important}.scalar-client .m-0{margin:0}.scalar-client .m-4{margin:16px}.scalar-client .m-auto{margin:auto}.scalar-client .\!mx-0{margin-left:0!important;margin-right:0!important}.scalar-client .-mx-1{margin-left:-4px;margin-right:-4px}.scalar-client .-my-1{margin-top:-4px;margin-bottom:-4px}.scalar-client .my-1{margin-top:4px;margin-bottom:4px}.scalar-client .my-1\.25{margin-top:5px;margin-bottom:5px}.scalar-client .my-12{margin-top:48px;margin-bottom:48px}.scalar-client .\!mb-0{margin-bottom:0!important}.scalar-client .\!mr-0{margin-right:0!important}.scalar-client .-mb-0{margin-bottom:-0px}.scalar-client .-mb-0\.25{margin-bottom:-1px}.scalar-client .-mb-\[var\(--scalar-border-width\)\]{margin-bottom:calc(var(--scalar-border-width) * -1)}.scalar-client .-ml-0{margin-left:-0px}.scalar-client .-ml-0\.5{margin-left:-2px}.scalar-client .-ml-12{margin-left:-48px}.scalar-client .-mr-1{margin-right:-4px}.scalar-client .-mr-1\.5{margin-right:-6px}.scalar-client .-mr-3{margin-right:-12px}.scalar-client .-mt-0{margin-top:-0px}.scalar-client .-mt-0\.25{margin-top:-1px}.scalar-client .-mt-\[\.5px\]{margin-top:-.5px}.scalar-client .mb-0{margin-bottom:0}.scalar-client .mb-1{margin-bottom:4px}.scalar-client .mb-1\.5{margin-bottom:6px}.scalar-client .mb-2{margin-bottom:8px}.scalar-client .mb-4{margin-bottom:16px}.scalar-client .mb-\[\.5px\]{margin-bottom:.5px}.scalar-client .ml-0{margin-left:0}.scalar-client .ml-0\.5{margin-left:2px}.scalar-client .ml-0\.75{margin-left:3px}.scalar-client .ml-1{margin-left:4px}.scalar-client .ml-2{margin-left:8px}.scalar-client .ml-auto{margin-left:auto}.scalar-client .mr-0{margin-right:0}.scalar-client .mr-0\.25{margin-right:1px}.scalar-client .mr-0\.75{margin-right:3px}.scalar-client .mr-1{margin-right:4px}.scalar-client .mr-1\.5{margin-right:6px}.scalar-client .mr-2{margin-right:8px}.scalar-client .mr-2\.5{margin-right:10px}.scalar-client .mr-3{margin-right:12px}.scalar-client .mr-auto{margin-right:auto}.scalar-client .mt-1{margin-top:4px}.scalar-client .mt-10{margin-top:40px}.scalar-client .mt-2{margin-top:8px}.scalar-client .mt-3{margin-top:12px}.scalar-client .mt-5{margin-top:20px}.scalar-client .mt-auto{margin-top:auto}.scalar-client .line-clamp-1{overflow:hidden;display:-webkit-box;-webkit-box-orient:vertical;-webkit-line-clamp:1}.scalar-client .block{display:block}.scalar-client .inline-block{display:inline-block}.scalar-client .\!flex{display:flex!important}.scalar-client .flex{display:flex}.scalar-client .inline-flex{display:inline-flex}.scalar-client .table{display:table}.scalar-client .grid{display:grid}.scalar-client .contents{display:contents}.scalar-client .\!hidden{display:none!important}.scalar-client .hidden{display:none}.scalar-client .aspect-\[4\/3\]{aspect-ratio:4/3}.scalar-client .aspect-square{aspect-ratio:1 / 1}.scalar-client .size-10{width:40px;height:40px}.scalar-client .size-2{width:8px;height:8px}.scalar-client .size-2\.5{width:10px;height:10px}.scalar-client .size-3{width:12px;height:12px}.scalar-client .size-3\.5{width:14px;height:14px}.scalar-client .size-3\/4{width:75%;height:75%}.scalar-client .size-4{width:16px;height:16px}.scalar-client .size-7{width:28px;height:28px}.scalar-client .h-1{height:4px}.scalar-client .h-1\.5{height:6px}.scalar-client .h-10{height:40px}.scalar-client .h-12{height:48px}.scalar-client .h-2{height:8px}.scalar-client .h-2\.5{height:10px}.scalar-client .h-24{height:96px}.scalar-client .h-3{height:12px}.scalar-client .h-3\.5{height:14px}.scalar-client .h-4{height:16px}.scalar-client .h-5{height:20px}.scalar-client .h-6{height:24px}.scalar-client .h-64{height:256px}.scalar-client .h-7{height:28px}.scalar-client .h-8{height:32px}.scalar-client .h-9{height:36px}.scalar-client .h-\[--scalar-address-bar-height\]{height:var(--scalar-address-bar-height)}.scalar-client .h-\[68px\]{height:68px}.scalar-client .h-\[calc\(100\%-273\.5px\)\]{height:calc(100% - 273.5px)}.scalar-client .h-\[calc\(100\%_-_50px\)\]{height:calc(100% - 50px)}.scalar-client .h-\[var\(--scalar-border-width\)\]{height:var(--scalar-border-width)}.scalar-client .h-auto{height:auto}.scalar-client .h-fit{height:-moz-fit-content;height:fit-content}.scalar-client .h-full{height:100%}.scalar-client .h-header{height:48px}.scalar-client .h-px{height:1px}.scalar-client .h-screen{height:100vh}.scalar-client .\!max-h-\[initial\]{max-height:initial!important}.scalar-client .max-h-40{max-height:160px}.scalar-client .max-h-8{max-height:32px}.scalar-client .max-h-\[40dvh\]{max-height:40dvh}.scalar-client .max-h-\[50dvh\]{max-height:50dvh}.scalar-client .max-h-\[60svh\]{max-height:60svh}.scalar-client .max-h-\[calc\(100\%-32px\)\]{max-height:calc(100% - 32px)}.scalar-client .max-h-\[inherit\]{max-height:inherit}.scalar-client .max-h-fit{max-height:-moz-fit-content;max-height:fit-content}.scalar-client .max-h-screen{max-height:100vh}.scalar-client .min-h-0{min-height:0px}.scalar-client .min-h-10{min-height:40px}.scalar-client .min-h-11{min-height:44px}.scalar-client .min-h-12{min-height:48px}.scalar-client .min-h-16{min-height:64px}.scalar-client .min-h-20{min-height:80px}.scalar-client .min-h-8{min-height:32px}.scalar-client .min-h-fit{min-height:-moz-fit-content;min-height:fit-content}.scalar-client .\!w-fit{width:-moz-fit-content!important;width:fit-content!important}.scalar-client .w-0{width:0px}.scalar-client .w-0\.5{width:2px}.scalar-client .w-1{width:4px}.scalar-client .w-1\.5{width:6px}.scalar-client .w-1\/2{width:50%}.scalar-client .w-10{width:40px}.scalar-client .w-2{width:8px}.scalar-client .w-2\.5{width:10px}.scalar-client .w-20{width:80px}.scalar-client .w-3{width:12px}.scalar-client .w-3\.5{width:14px}.scalar-client .w-4{width:16px}.scalar-client .w-5{width:20px}.scalar-client .w-56{width:224px}.scalar-client .w-6{width:24px}.scalar-client .w-64{width:256px}.scalar-client .w-72{width:288px}.scalar-client .w-8{width:32px}.scalar-client .w-\[60px\]{width:60px}.scalar-client .w-\[calc\(100\%-10px\)\]{width:calc(100% - 10px)}.scalar-client .w-\[calc\(100\%_-_8px\)\]{width:calc(100% - 8px)}.scalar-client .w-\[inherit\]{width:inherit}.scalar-client .w-auto{width:auto}.scalar-client .w-dvw{width:100dvw}.scalar-client .w-fit{width:-moz-fit-content;width:fit-content}.scalar-client .w-full{width:100%}.scalar-client .w-max{width:-moz-max-content;width:max-content}.scalar-client .min-w-0{min-width:0px}.scalar-client .min-w-2{min-width:8px}.scalar-client .min-w-2\.5{min-width:10px}.scalar-client .min-w-3{min-width:12px}.scalar-client .min-w-3\.5{min-width:14px}.scalar-client .min-w-48{min-width:192px}.scalar-client .min-w-8{min-width:32px}.scalar-client .min-w-\[296px\]{min-width:296px}.scalar-client .min-w-\[37px\]{min-width:37px}.scalar-client .min-w-fit{min-width:-moz-fit-content;min-width:fit-content}.scalar-client .min-w-full{min-width:100%}.scalar-client .max-w-\[100\%\]{max-width:100%}.scalar-client .max-w-\[14px\]{max-width:14px}.scalar-client .max-w-\[150px\]{max-width:150px}.scalar-client .max-w-\[320px\]{max-width:320px}.scalar-client .max-w-\[37px\]{max-width:37px}.scalar-client .max-w-\[380px\]{max-width:380px}.scalar-client .max-w-\[420px\]{max-width:420px}.scalar-client .max-w-\[720px\]{max-width:720px}.scalar-client .max-w-\[calc\(100dvw-24px\)\]{max-width:calc(100dvw - 24px)}.scalar-client .max-w-full{max-width:100%}.scalar-client .flex-1{flex:1 1 0%}.scalar-client .shrink-0{flex-shrink:0}.scalar-client .flex-grow{flex-grow:1}.scalar-client .-translate-x-1\/2{--tw-translate-x:-50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .-translate-y-1\/2{--tw-translate-y:-50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .translate-x-0{--tw-translate-x:0px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .translate-x-1\/2{--tw-translate-x:50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .translate-y-1\/2{--tw-translate-y:50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .rotate-90{--tw-rotate:90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .scale-75{--tw-scale-x:.75;--tw-scale-y:.75;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .transform{transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .cursor-auto{cursor:auto}.scalar-client .cursor-default{cursor:default}.scalar-client .cursor-help{cursor:help}.scalar-client .cursor-pointer{cursor:pointer}.scalar-client .cursor-text{cursor:text}.scalar-client .select-none{-webkit-user-select:none;-moz-user-select:none;user-select:none}.scalar-client .resize-none{resize:none}.scalar-client .resize{resize:both}.scalar-client .auto-rows-\[32px\]{grid-auto-rows:32px}.scalar-client .auto-rows-auto{grid-auto-rows:auto}.scalar-client .grid-cols-2{grid-template-columns:repeat(2,minmax(0,1fr))}.scalar-client .grid-cols-\[44px\,1fr\,repeat\(3\,auto\)\]{grid-template-columns:44px 1fr repeat(3,auto)}.scalar-client .grid-cols-\[auto\,1fr\]{grid-template-columns:auto 1fr}.scalar-client .grid-cols-\[repeat\(auto-fill\,minmax\(32px\,1fr\)\)\]{grid-template-columns:repeat(auto-fill,minmax(32px,1fr))}.scalar-client .flex-row{flex-direction:row}.scalar-client .flex-col{flex-direction:column}.scalar-client .flex-wrap{flex-wrap:wrap}.scalar-client .content-start{align-content:flex-start}.scalar-client .content-between{align-content:space-between}.scalar-client .items-start{align-items:flex-start}.scalar-client .items-end{align-items:flex-end}.scalar-client .items-center{align-items:center}.scalar-client .items-stretch{align-items:stretch}.scalar-client .justify-start{justify-content:flex-start}.scalar-client .justify-end{justify-content:flex-end}.scalar-client .justify-center{justify-content:center}.scalar-client .justify-between{justify-content:space-between}.scalar-client .justify-stretch{justify-content:stretch}.scalar-client .\!gap-2{gap:8px!important}.scalar-client .gap-0{gap:0px}.scalar-client .gap-0\.5{gap:2px}.scalar-client .gap-0\.75{gap:3px}.scalar-client .gap-1{gap:4px}.scalar-client .gap-1\.5{gap:6px}.scalar-client .gap-1\/2{gap:.5px}.scalar-client .gap-10{gap:40px}.scalar-client .gap-2{gap:8px}.scalar-client .gap-2\.5{gap:10px}.scalar-client .gap-3{gap:12px}.scalar-client .gap-4{gap:16px}.scalar-client .gap-6{gap:24px}.scalar-client .gap-8{gap:32px}.scalar-client .gap-\[1\.5px\]{gap:1.5px}.scalar-client .gap-px{gap:1px}.scalar-client .gap-x-2{-moz-column-gap:8px;column-gap:8px}.scalar-client .gap-x-2\.5{-moz-column-gap:10px;column-gap:10px}.scalar-client .space-x-1>:not([hidden])~:not([hidden]){--tw-space-x-reverse:0;margin-right:calc(4px * var(--tw-space-x-reverse));margin-left:calc(4px * calc(1 - var(--tw-space-x-reverse)))}.scalar-client .divide-x>:not([hidden])~:not([hidden]){--tw-divide-x-reverse:0;border-right-width:calc(var(--scalar-border-width) * var(--tw-divide-x-reverse));border-left-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-x-reverse)))}.scalar-client .divide-y>:not([hidden])~:not([hidden]){--tw-divide-y-reverse:0;border-top-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-y-reverse)));border-bottom-width:calc(var(--scalar-border-width) * var(--tw-divide-y-reverse))}.scalar-client .self-center{align-self:center}.scalar-client .justify-self-end{justify-self:end}.scalar-client .overflow-auto{overflow:auto}.scalar-client .overflow-hidden{overflow:hidden}.scalar-client .overflow-visible{overflow:visible}.scalar-client .overflow-x-auto{overflow-x:auto}.scalar-client .overflow-y-auto{overflow-y:auto}.scalar-client .truncate{overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.scalar-client .text-ellipsis{text-overflow:ellipsis}.scalar-client .whitespace-nowrap{white-space:nowrap}.scalar-client .whitespace-pre{white-space:pre}.scalar-client .text-balance{text-wrap:balance}.scalar-client .text-pretty{text-wrap:pretty}.scalar-client .break-words{overflow-wrap:break-word}.scalar-client .break-all{word-break:break-all}.scalar-client .\!rounded-none{border-radius:0!important}.scalar-client .rounded{border-radius:var(--scalar-radius)}.scalar-client .rounded-\[10px\]{border-radius:10px}.scalar-client .rounded-full{border-radius:9999px}.scalar-client .rounded-lg{border-radius:var(--scalar-radius-lg)}.scalar-client .rounded-md{border-radius:var(--scalar-radius)}.scalar-client .rounded-px{border-radius:1px}.scalar-client .rounded-xl{border-radius:var(--scalar-radius-xl)}.scalar-client .rounded-b{border-bottom-right-radius:var(--scalar-radius);border-bottom-left-radius:var(--scalar-radius)}.scalar-client .rounded-b-lg{border-bottom-right-radius:var(--scalar-radius-lg);border-bottom-left-radius:var(--scalar-radius-lg)}.scalar-client .rounded-t{border-top-left-radius:var(--scalar-radius);border-top-right-radius:var(--scalar-radius)}.scalar-client .rounded-t-lg{border-top-left-radius:var(--scalar-radius-lg);border-top-right-radius:var(--scalar-radius-lg)}.scalar-client .rounded-t-none{border-top-left-radius:0;border-top-right-radius:0}.scalar-client .\!border-0{border-width:0!important}.scalar-client .border{border-width:var(--scalar-border-width)}.scalar-client .border-0{border-width:0}.scalar-client .border-1\/2{border-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-\[1\.5px\]{border-width:1.5px}.scalar-client .border-\[1px\]{border-width:1px}.scalar-client .\!border-r-1\/2{border-right-width:calc(var(--scalar-border-width) / 2)!important}.scalar-client .border-b{border-bottom-width:var(--scalar-border-width)}.scalar-client .border-b-0{border-bottom-width:0}.scalar-client .border-b-1\/2{border-bottom-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-b-\[1px\]{border-bottom-width:1px}.scalar-client .border-l{border-left-width:var(--scalar-border-width)}.scalar-client .border-l-0{border-left-width:0}.scalar-client .border-r{border-right-width:var(--scalar-border-width)}.scalar-client .border-r-0{border-right-width:0}.scalar-client .border-r-1\/2{border-right-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-t{border-top-width:var(--scalar-border-width)}.scalar-client .border-t-0{border-top-width:0}.scalar-client .border-t-1\/2{border-top-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-solid{border-style:solid}.scalar-client .border-dashed{border-style:dashed}.scalar-client .border-none{border-style:none}.scalar-client .\!border-current{border-color:currentColor!important}.scalar-client .border-b-3{border-color:var(--scalar-background-3)}.scalar-client .border-c-3{border-color:var(--scalar-color-3)}.scalar-client .border-transparent{border-color:transparent}.scalar-client .border-r-transparent{border-right-color:transparent}.scalar-client .bg-b-1{background-color:var(--scalar-background-1)}.scalar-client .bg-b-2{background-color:var(--scalar-background-2)}.scalar-client .bg-b-3{background-color:var(--scalar-background-3)}.scalar-client .bg-blue{background-color:var(--scalar-color-blue)}.scalar-client .bg-border{background-color:var(--scalar-border-color)}.scalar-client .bg-c-2{background-color:var(--scalar-color-2)}.scalar-client .bg-c-3{background-color:var(--scalar-color-3)}.scalar-client .bg-c-accent{background-color:var(--scalar-color-accent)}.scalar-client .bg-current{background-color:currentColor}.scalar-client .bg-green{background-color:var(--scalar-color-green)}.scalar-client .bg-grey{background-color:var(--scalar-color-3)}.scalar-client .bg-orange{background-color:var(--scalar-color-orange)}.scalar-client .bg-purple{background-color:var(--scalar-color-purple)}.scalar-client .bg-red{background-color:var(--scalar-color-red)}.scalar-client .bg-sidebar-active-b{background-color:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .bg-yellow{background-color:var(--scalar-color-yellow)}.scalar-client .bg-none{background-image:none}.scalar-client .fill-current{fill:currentColor}.scalar-client .stroke-2{stroke-width:2}.scalar-client .stroke-\[1\.5\]{stroke-width:1.5}.scalar-client .stroke-\[1\.75\]{stroke-width:1.75}.scalar-client .stroke-\[2\.25\]{stroke-width:2.25}.scalar-client .object-contain{-o-object-fit:contain;object-fit:contain}.scalar-client .\!p-0{padding:0!important}.scalar-client .p-0{padding:0}.scalar-client .p-0\.5{padding:2px}.scalar-client .p-0\.75{padding:3px}.scalar-client .p-1{padding:4px}.scalar-client .p-1\.5{padding:6px}.scalar-client .p-1\.75{padding:7px}.scalar-client .p-2{padding:8px}.scalar-client .p-3{padding:12px}.scalar-client .p-4{padding:16px}.scalar-client .p-\[3px\]{padding:3px}.scalar-client .p-\[5px\]{padding:5px}.scalar-client .\!px-3{padding-left:12px!important;padding-right:12px!important}.scalar-client .\!py-1{padding-top:4px!important;padding-bottom:4px!important}.scalar-client .\!py-1\.5{padding-top:6px!important;padding-bottom:6px!important}.scalar-client .px-0{padding-left:0;padding-right:0}.scalar-client .px-0\.5{padding-left:2px;padding-right:2px}.scalar-client .px-1{padding-left:4px;padding-right:4px}.scalar-client .px-1\.5{padding-left:6px;padding-right:6px}.scalar-client .px-2{padding-left:8px;padding-right:8px}.scalar-client .px-2\.5{padding-left:10px;padding-right:10px}.scalar-client .px-3{padding-left:12px;padding-right:12px}.scalar-client .px-4{padding-left:16px;padding-right:16px}.scalar-client .px-5{padding-left:20px;padding-right:20px}.scalar-client .px-6{padding-left:24px;padding-right:24px}.scalar-client .px-8{padding-left:32px;padding-right:32px}.scalar-client .py-0{padding-top:0;padding-bottom:0}.scalar-client .py-0\.5{padding-top:2px;padding-bottom:2px}.scalar-client .py-0\.75{padding-top:3px;padding-bottom:3px}.scalar-client .py-1{padding-top:4px;padding-bottom:4px}.scalar-client .py-1\.25{padding-top:5px;padding-bottom:5px}.scalar-client .py-1\.5{padding-top:6px;padding-bottom:6px}.scalar-client .py-2{padding-top:8px;padding-bottom:8px}.scalar-client .py-2\.5{padding-top:10px;padding-bottom:10px}.scalar-client .py-3{padding-top:12px;padding-bottom:12px}.scalar-client .py-5{padding-top:20px;padding-bottom:20px}.scalar-client .py-8{padding-top:32px;padding-bottom:32px}.scalar-client .py-px{padding-top:1px;padding-bottom:1px}.scalar-client .\!pl-3{padding-left:12px!important}.scalar-client .\!pt-0{padding-top:0!important}.scalar-client .pb-0{padding-bottom:0}.scalar-client .pb-14{padding-bottom:56px}.scalar-client .pb-2{padding-bottom:8px}.scalar-client .pb-3{padding-bottom:12px}.scalar-client .pb-5{padding-bottom:20px}.scalar-client .pb-8{padding-bottom:32px}.scalar-client .pb-\[75px\]{padding-bottom:75px}.scalar-client .pl-1{padding-left:4px}.scalar-client .pl-1\.25{padding-left:5px}.scalar-client .pl-1\.5{padding-left:6px}.scalar-client .pl-2{padding-left:8px}.scalar-client .pl-3{padding-left:12px}.scalar-client .pl-6{padding-left:24px}.scalar-client .pl-8{padding-left:32px}.scalar-client .pl-9{padding-left:36px}.scalar-client .pl-px{padding-left:1px}.scalar-client .pr-0{padding-right:0}.scalar-client .pr-1{padding-right:4px}.scalar-client .pr-1\.5{padding-right:6px}.scalar-client .pr-2{padding-right:8px}.scalar-client .pr-2\.5{padding-right:10px}.scalar-client .pr-4{padding-right:16px}.scalar-client .pr-6{padding-right:24px}.scalar-client .pr-\[26px\]{padding-right:26px}.scalar-client .pr-px{padding-right:1px}.scalar-client .pt-0{padding-top:0}.scalar-client .pt-2{padding-top:8px}.scalar-client .pt-3{padding-top:12px}.scalar-client .pt-4{padding-top:16px}.scalar-client .pt-px{padding-top:1px}.scalar-client .text-left{text-align:left}.scalar-client .text-center{text-align:center}.scalar-client .text-right{text-align:right}.scalar-client .font-code{font-family:var(--scalar-font-code)}.scalar-client .text-3xs{font-size:var(--scalar-font-size-7)}.scalar-client .text-\[11px\]{font-size:11px}.scalar-client .text-\[21px\]{font-size:21px}.scalar-client .text-\[6px\]{font-size:6px}.scalar-client .text-base{font-size:var(--scalar-font-size-3)}.scalar-client .text-sm{font-size:var(--scalar-font-size-4)}.scalar-client .text-xl{font-size:var(--scalar-font-size-1)}.scalar-client .text-xs{font-size:var(--scalar-font-size-5)}.scalar-client .text-xxs{font-size:var(--scalar-font-size-6)}.scalar-client .\!font-normal{font-weight:var(--scalar-regular)!important}.scalar-client .font-bold{font-weight:var(--scalar-bold)}.scalar-client .font-medium{font-weight:var(--scalar-semibold)}.scalar-client .font-normal{font-weight:var(--scalar-regular)}.scalar-client .uppercase{text-transform:uppercase}.scalar-client .lowercase{text-transform:lowercase}.scalar-client .capitalize{text-transform:capitalize}.scalar-client .\!leading-\[6px\]{line-height:6px!important}.scalar-client .leading{line-height:1.5}.scalar-client .leading-3{line-height:var(--scalar-line-height-3)}.scalar-client .leading-5{line-height:var(--scalar-line-height-5)}.scalar-client .leading-\[1\.44\]{line-height:1.44}.scalar-client .leading-\[21px\]{line-height:21px}.scalar-client .leading-\[7px\]{line-height:7px}.scalar-client .leading-\[normal\]{line-height:normal}.scalar-client .leading-none{line-height:1}.scalar-client .leading-normal{line-height:1.5}.scalar-client .leading-snug{line-height:1.375}.scalar-client .\!text-c-1{color:var(--scalar-color-1)!important}.scalar-client .text-b-1{color:var(--scalar-background-1)}.scalar-client .text-blue{color:var(--scalar-color-blue)}.scalar-client .text-border{color:var(--scalar-border-color)}.scalar-client .text-c-1{color:var(--scalar-color-1)}.scalar-client .text-c-2{color:var(--scalar-color-2)}.scalar-client .text-c-3{color:var(--scalar-color-3)}.scalar-client .text-c-btn{color:var(--scalar-button-1-color)}.scalar-client .text-green{color:var(--scalar-color-green)}.scalar-client .text-grey{color:var(--scalar-color-3)}.scalar-client .text-orange{color:var(--scalar-color-orange)}.scalar-client .text-purple{color:var(--scalar-color-purple)}.scalar-client .text-red{color:var(--scalar-color-red)}.scalar-client .text-sidebar-active-c{color:var(--scalar-sidebar-color-active, currentColor)}.scalar-client .text-sidebar-c-2{color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}.scalar-client .text-transparent{color:transparent}.scalar-client .text-yellow{color:var(--scalar-color-yellow)}.scalar-client .underline{text-decoration-line:underline}.scalar-client .no-underline{text-decoration-line:none}.scalar-client .decoration-c-3{text-decoration-color:var(--scalar-color-3)}.scalar-client .underline-offset-2{text-underline-offset:2px}.scalar-client .opacity-0{opacity:0}.scalar-client .opacity-100{opacity:1}.scalar-client .opacity-25{opacity:.25}.scalar-client .shadow{--tw-shadow:var(--scalar-shadow-1);--tw-shadow-colored:var(--scalar-shadow-1);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-\[-8px_0_4px_var\(--scalar-background-1\)\]{--tw-shadow:-8px 0 4px var(--scalar-background-1);--tw-shadow-colored:-8px 0 4px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-\[0_-8px_0_8px_var\(--scalar-background-1\)\,0_0_8px_8px_var\(--scalar-background-1\)\]{--tw-shadow:0 -8px 0 8px var(--scalar-background-1),0 0 8px 8px var(--scalar-background-1);--tw-shadow-colored:0 -8px 0 8px var(--tw-shadow-color), 0 0 8px 8px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-border{--tw-shadow:inset 0 0 0 1px var(--scalar-border-color);--tw-shadow-colored:inset 0 0 0 1px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-lg{--tw-shadow:var(--scalar-shadow-2);--tw-shadow-colored:var(--scalar-shadow-2);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-none{--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-border{--tw-shadow-color:var(--scalar-border-color);--tw-shadow:var(--tw-shadow-colored)}.scalar-client .outline-none{outline:2px solid transparent;outline-offset:2px}.scalar-client .outline{outline-style:solid}.scalar-client .-outline-offset-1{outline-offset:-1px}.scalar-client .-outline-offset-2{outline-offset:-2px}.scalar-client .outline-offset-2{outline-offset:2px}.scalar-client .blur{--tw-blur:blur(8px);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .brightness-\[\.9\]{--tw-brightness:brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .brightness-lifted{--tw-brightness:brightness(var(--scalar-lifted-brightness));filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .filter{filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .backdrop-filter{-webkit-backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia);backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia)}.scalar-client .transition{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,-webkit-backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter,-webkit-backdrop-filter;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .transition-colors{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .transition-none{transition-property:none}.scalar-client .transition-opacity{transition-property:opacity;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .transition-transform{transition-property:transform;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .duration-150{transition-duration:.15s}.scalar-client .duration-200{transition-duration:.2s}.scalar-client .duration-300{transition-duration:.3s}.scalar-client .ease-in-out{transition-timing-function:cubic-bezier(.4,0,.2,1)}.scalar-client .bg-mix-transparent{background-color:color-mix(var(--tw-bg-mix-method, in srgb),transparent calc(var(--tw-bg-mix-amount, 0) * 1%),var(--tw-bg-base))}.scalar-client .bg-mix-amount-90{--tw-bg-mix-amount:90}.scalar-client .bg-mix-amount-95{--tw-bg-mix-amount:95}.scalar-client .bg-b-1{--tw-bg-base:var(--scalar-background-1)}.scalar-client .bg-b-2{--tw-bg-base:var(--scalar-background-2)}.scalar-client .bg-b-3{--tw-bg-base:var(--scalar-background-3)}.scalar-client .bg-blue{--tw-bg-base:var(--scalar-color-blue)}.scalar-client .bg-border{--tw-bg-base:var(--scalar-border-color)}.scalar-client .bg-c-2{--tw-bg-base:var(--scalar-color-2)}.scalar-client .bg-c-3{--tw-bg-base:var(--scalar-color-3)}.scalar-client .bg-c-accent{--tw-bg-base:var(--scalar-color-accent)}.scalar-client .bg-current{--tw-bg-base:currentColor}.scalar-client .bg-green{--tw-bg-base:var(--scalar-color-green)}.scalar-client .bg-grey{--tw-bg-base:var(--scalar-color-3)}.scalar-client .bg-orange{--tw-bg-base:var(--scalar-color-orange)}.scalar-client .bg-purple{--tw-bg-base:var(--scalar-color-purple)}.scalar-client .bg-red{--tw-bg-base:var(--scalar-color-red)}.scalar-client .bg-sidebar-active-b{--tw-bg-base:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .bg-yellow{--tw-bg-base:var(--scalar-color-yellow)}.scalar-client .app-drag-region{-webkit-app-region:drag}.scalar-client .app-no-drag-region{-webkit-app-region:no-drag}.scalar-client .\[--scalar-address-bar-height\:32px\]{--scalar-address-bar-height:32px}.scalar-client .\[--scalar-small\:--scalar-font-size-4\]{--scalar-small:var(--scalar-font-size-4)}.scalar-client .\*\:flex>*{display:flex}.scalar-client .\*\:h-8>*{height:32px}.scalar-client .\*\:cursor-pointer>*{cursor:pointer}.scalar-client .\*\:items-center>*{align-items:center}.scalar-client .\*\:rounded-none>*{border-radius:0}.scalar-client .\*\:border-t-1\/2>*{border-top-width:calc(var(--scalar-border-width) / 2)}.scalar-client .\*\:px-1>*{padding-left:4px;padding-right:4px}.scalar-client .\*\:px-1\.5>*{padding-left:6px;padding-right:6px}.scalar-client .\*\:pl-4>*{padding-left:16px}.scalar-client .before\:pointer-events-none:before{content:var(--tw-content);pointer-events:none}.scalar-client .before\:absolute:before{content:var(--tw-content);position:absolute}.scalar-client .before\:left-3:before{content:var(--tw-content);left:12px}.scalar-client .before\:left-\[calc\(\.75rem_\+_\.5px\)\]:before{content:var(--tw-content);left:calc(.75rem + .5px)}.scalar-client .before\:right-\[23px\]:before{content:var(--tw-content);right:23px}.scalar-client .before\:top-0:before{content:var(--tw-content);top:0}.scalar-client .before\:top-0\.5:before{content:var(--tw-content);top:2px}.scalar-client .before\:z-1:before{content:var(--tw-content);z-index:1}.scalar-client .before\:h-\[calc\(100\%_\+_\.5px\)\]:before{content:var(--tw-content);height:calc(100% + .5px)}.scalar-client .before\:min-h-\[calc\(100\%-4px\)\]:before{content:var(--tw-content);min-height:calc(100% - 4px)}.scalar-client .before\:w-3:before{content:var(--tw-content);width:12px}.scalar-client .before\:w-\[\.5px\]:before{content:var(--tw-content);width:.5px}.scalar-client .before\:bg-border:before{content:var(--tw-content);background-color:var(--scalar-border-color)}.scalar-client .before\:bg-gradient-to-r:before{content:var(--tw-content);background-image:linear-gradient(to right,var(--tw-gradient-stops))}.scalar-client .before\:from-transparent:before{content:var(--tw-content);--tw-gradient-from:transparent var(--tw-gradient-from-position);--tw-gradient-to:rgb(0 0 0 / 0) var(--tw-gradient-to-position);--tw-gradient-stops:var(--tw-gradient-from), var(--tw-gradient-to)}.scalar-client .before\:to-b-1:before{content:var(--tw-content);--tw-gradient-to:var(--scalar-background-1) var(--tw-gradient-to-position)}.scalar-client .before\:content-\[\'\'\]:before{--tw-content:"";content:var(--tw-content)}.scalar-client .before\:bg-border:before{content:var(--tw-content);--tw-bg-base:var(--scalar-border-color)}.scalar-client .after\:content-\[\'\:\'\]:after{--tw-content:":";content:var(--tw-content)}.scalar-client .first\:border-t:first-child{border-top-width:var(--scalar-border-width)}.scalar-client .first\:\*\:rounded-l>*:first-child{border-top-left-radius:var(--scalar-radius);border-bottom-left-radius:var(--scalar-radius)}.scalar-client .first\:\*\:border-t-0>*:first-child{border-top-width:0}.scalar-client .last\:mb-0:last-child{margin-bottom:0}.scalar-client .last\:border-r-0:last-child{border-right-width:0}.scalar-client .last\:\*\:rounded-r>*:last-child{border-top-right-radius:var(--scalar-radius);border-bottom-right-radius:var(--scalar-radius)}.scalar-client .last\:before\:h-full:last-child:before{content:var(--tw-content);height:100%}.scalar-client .focus-within\:z-20:focus-within{z-index:20}.scalar-client .focus-within\:border-b-3:focus-within{border-color:var(--scalar-background-3)}.scalar-client .focus-within\:bg-b-1:focus-within{background-color:var(--scalar-background-1)}.scalar-client .focus-within\:text-c-1:focus-within{color:var(--scalar-color-1)}.scalar-client .focus-within\:bg-b-1:focus-within{--tw-bg-base:var(--scalar-background-1)}.scalar-client .hover\:cursor-default:hover{cursor:default}.scalar-client .hover\:bg-b-2:hover{background-color:var(--scalar-background-2)}.scalar-client .hover\:bg-b-3:hover{background-color:var(--scalar-background-3)}.scalar-client .hover\:bg-inherit:hover{background-color:inherit}.scalar-client .hover\:bg-sidebar-active-b:hover{background-color:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .hover\:text-c-1:hover{color:var(--scalar-color-1)}.scalar-client .hover\:text-c-2:hover{color:var(--scalar-color-2)}.scalar-client .hover\:bg-b-2:hover{--tw-bg-base:var(--scalar-background-2)}.scalar-client .hover\:bg-b-3:hover{--tw-bg-base:var(--scalar-background-3)}.scalar-client .hover\:bg-inherit:hover{--tw-bg-base:inherit}.scalar-client .hover\:bg-sidebar-active-b:hover{--tw-bg-base:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .focus\:text-c-1:focus{color:var(--scalar-color-1)}.scalar-client .focus\:outline-none:focus{outline:2px solid transparent;outline-offset:2px}.scalar-client .focus-visible\:z-10:focus-visible{z-index:10}.scalar-client .active\:text-c-1:active{color:var(--scalar-color-1)}.scalar-client .disabled\:cursor-default:disabled{cursor:default}.scalar-client .disabled\:text-c-2:disabled{color:var(--scalar-color-2)}.scalar-client .group\/label:last-child .group-last\/label\:rounded-br-lg{border-bottom-right-radius:var(--scalar-radius-lg)}.scalar-client .group:last-child .group-last\:border-b-transparent{border-bottom-color:transparent}.scalar-client .group\/upload:hover .group-hover\/upload\:block,.scalar-client .group:hover .group-hover\:block{display:block}.scalar-client .group\/item:hover .group-hover\/item\:flex,.scalar-client .group:hover .group-hover\:flex{display:flex}.scalar-client .group:hover .group-hover\:hidden{display:none}.scalar-client .group:hover .group-hover\:pr-5{padding-right:20px}.scalar-client .group\/info:hover .group-hover\/info\:text-c-1{color:var(--scalar-color-1)}.scalar-client .group\/scopes-accordion:hover .group-hover\/scopes-accordion\:text-c-2{color:var(--scalar-color-2)}.scalar-client .group:hover .group-hover\:text-c-1{color:var(--scalar-color-1)}.scalar-client .group\/cell:hover .group-hover\/cell\:opacity-100,.scalar-client .group\/params:hover .group-hover\/params\:opacity-100,.scalar-client .group:hover .group-hover\:opacity-100{opacity:1}.scalar-client .group:hover .group-hover\:opacity-80{opacity:.8}.scalar-client .group:focus-visible .group-focus-visible\:opacity-100{opacity:1}.scalar-client .group:focus-visible .group-focus-visible\:outline{outline-style:solid}.scalar-client .group.alert .group-\[\.alert\]\:bg-b-alert{background-color:var(--scalar-background-alert)}.scalar-client .group.alert .group-\[\.alert\]\:bg-transparent{background-color:transparent}.scalar-client .group.alert .group-\[\.alert\]\:shadow-none{--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .group.alert .group-\[\.alert\]\:outline-orange{outline-color:var(--scalar-color-orange)}.scalar-client .group.alert .group-\[\.alert\]\:bg-b-alert{--tw-bg-base:var(--scalar-background-alert)}.scalar-client .group.alert .group-\[\.alert\]\:bg-transparent{--tw-bg-base:transparent}.scalar-client .group.alert .group-\[\.alert\]\:before\:to-b-alert:before{content:var(--tw-content);--tw-gradient-to:var(--scalar-background-alert) var(--tw-gradient-to-position)}.scalar-client .peer:checked~.peer-checked\:text-c-1{color:var(--scalar-color-1)}.scalar-client .has-\[\:focus-visible\]\:rounded-\[4px\]:has(:focus-visible){border-radius:4px}.scalar-client .has-\[\.empty-sidebar-item\]\:border-t-1\/2:has(.empty-sidebar-item){border-top-width:calc(var(--scalar-border-width) / 2)}.scalar-client .has-\[\:focus-visible\]\:opacity-100:has(:focus-visible){opacity:1}.scalar-client .has-\[\:focus-visible\]\:outline:has(:focus-visible){outline-style:solid}.scalar-client .group\/label:has(input) .group-has-\[input\]\/label\:mr-0{margin-right:0}.scalar-client .group\/cell:has(:focus-visible) .group-has-\[\:focus-visible\]\/cell\:border-c-accent{border-color:var(--scalar-color-accent)}.scalar-client .group\/cell:has(:focus-visible) .group-has-\[\:focus-visible\]\/cell\:opacity-100{opacity:1}.scalar-client .group:has(:focus-visible) .group-has-\[\:focus-visible\]\:opacity-100{opacity:1}.scalar-client .peer:has(.color-selector)~.peer-has-\[\.color-selector\]\:hidden{display:none}.scalar-client .peer:has(.cm-focused)~.peer-has-\[\.cm-focused\]\:opacity-0{opacity:0}.scalar-client .peer:has(:focus-visible)~.peer-has-\[\:focus-visible\]\:opacity-0{opacity:0}.scalar-client .ui-open\:rotate-90[data-headlessui-state~=open]{--tw-rotate:90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client :where([data-headlessui-state~=open]) .ui-open\:rotate-90{--tw-rotate:90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .ui-not-open\:invisible[data-headlessui-state]:not([data-headlessui-state~=open]){visibility:hidden}.scalar-client .ui-not-open\:rotate-0[data-headlessui-state]:not([data-headlessui-state~=open]){--tw-rotate:0deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client :where([data-headlessui-state]:not([data-headlessui-state~=open])) .ui-not-open\:invisible:not([data-headlessui-state]){visibility:hidden}.scalar-client :where([data-headlessui-state]:not([data-headlessui-state~=open])) .ui-not-open\:rotate-0:not([data-headlessui-state]){--tw-rotate:0deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .ui-checked\:bg-b-3[data-headlessui-state~=checked]{background-color:var(--scalar-background-3);--tw-bg-base:var(--scalar-background-3)}.scalar-client :where([data-headlessui-state~=checked]) .ui-checked\:bg-b-3{background-color:var(--scalar-background-3);--tw-bg-base:var(--scalar-background-3)}.scalar-client .ui-active\:bg-b-2[data-headlessui-state~=active]{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .\*\:ui-active\:bg-b-2[data-headlessui-state~=active]>*{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client :where([data-headlessui-state~=active]) .ui-active\:bg-b-2{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client :where([data-headlessui-state~=active]) .\*\:ui-active\:bg-b-2>*{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .app-platform-desktop .desktop\:rounded-lg{border-radius:var(--scalar-radius-lg)}.scalar-client .app-platform-mac .mac\:hidden{display:none}.scalar-client .app-platform-mac .mac\:pl-\[72px\]{padding-left:72px}@media (min-width: 600px){.scalar-client .sm\:not-sr-only{position:static;width:auto;height:auto;padding:0;margin:0;overflow:visible;clip:auto;white-space:normal}.scalar-client .sm\:order-none{order:0}.scalar-client .sm\:mb-1{margin-bottom:4px}.scalar-client .sm\:mb-1\.5{margin-bottom:6px}.scalar-client .sm\:ml-1{margin-left:4px}.scalar-client .sm\:ml-1\.5{margin-left:6px}.scalar-client .sm\:mr-1{margin-right:4px}.scalar-client .sm\:mr-1\.5{margin-right:6px}.scalar-client .sm\:flex{display:flex}.scalar-client .sm\:hidden{display:none}.scalar-client .sm\:min-w-max{min-width:-moz-max-content;min-width:max-content}.scalar-client .sm\:max-w-max{max-width:-moz-max-content;max-width:max-content}.scalar-client .sm\:flex-row{flex-direction:row}.scalar-client .sm\:flex-col{flex-direction:column}.scalar-client .sm\:justify-between{justify-content:space-between}.scalar-client .sm\:gap-px{gap:1px}.scalar-client .sm\:rounded{border-radius:var(--scalar-radius)}.scalar-client .sm\:rounded-lg{border-radius:var(--scalar-radius-lg)}.scalar-client .sm\:px-2{padding-left:8px;padding-right:8px}.scalar-client .sm\:px-3{padding-left:12px;padding-right:12px}.scalar-client .sm\:py-1{padding-top:4px;padding-bottom:4px}.scalar-client .sm\:py-1\.5{padding-top:6px;padding-bottom:6px}.scalar-client .sm\:\*\:rounded-lg>*{border-radius:var(--scalar-radius-lg)}}@media (min-width: 800px){.scalar-client .md\:bottom-10{bottom:40px}.scalar-client .md\:right-10{right:40px}.scalar-client .md\:-ml-1{margin-left:-4px}.scalar-client .md\:-ml-1\.25{margin-left:-5px}.scalar-client .md\:block{display:block}.scalar-client .md\:w-full{width:100%}.scalar-client .md\:min-w-fit{min-width:-moz-fit-content;min-width:fit-content}.scalar-client .md\:flex-none{flex:none}.scalar-client .md\:translate-x-0{--tw-translate-x:0px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .md\:translate-y-0{--tw-translate-y:0px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .md\:flex-row{flex-direction:row}.scalar-client .md\:border-b-0{border-bottom-width:0}.scalar-client .md\:border-r-1\/2{border-right-width:calc(var(--scalar-border-width) / 2)}.scalar-client .md\:p-1{padding:4px}.scalar-client .md\:p-1\.5{padding:6px}.scalar-client .md\:px-0{padding-left:0;padding-right:0}.scalar-client .md\:px-1{padding-left:4px;padding-right:4px}.scalar-client .md\:px-1\.5{padding-left:6px;padding-right:6px}.scalar-client .md\:px-2{padding-left:8px;padding-right:8px}.scalar-client .md\:px-2\.5{padding-left:10px;padding-right:10px}.scalar-client .md\:px-4{padding-left:16px;padding-right:16px}.scalar-client .md\:px-\[18px\]{padding-left:18px;padding-right:18px}.scalar-client .md\:py-2{padding-top:8px;padding-bottom:8px}.scalar-client .md\:py-2\.5{padding-top:10px;padding-bottom:10px}.scalar-client .md\:pb-2{padding-bottom:8px}.scalar-client .md\:pb-2\.5{padding-bottom:10px}.scalar-client .md\:pb-\[37px\]{padding-bottom:37px}.scalar-client .md\:\*\:border-t-0>*{border-top-width:0}}@media (min-width: 1000px){.scalar-client .lg\:order-none{order:0}.scalar-client .lg\:-mr-1{margin-right:-4px}.scalar-client .lg\:mb-0{margin-bottom:0}.scalar-client .lg\:flex{display:flex}.scalar-client .lg\:min-h-client-header{min-height:48px}.scalar-client .lg\:w-\[120px\]{width:120px}.scalar-client .lg\:w-auto{width:auto}.scalar-client .lg\:min-w-\[580px\]{min-width:580px}.scalar-client .lg\:max-w-\[580px\]{max-width:580px}.scalar-client .lg\:flex-1{flex:1 1 0%}.scalar-client .lg\:p-1{padding:4px}.scalar-client .lg\:px-1{padding-left:4px;padding-right:4px}.scalar-client .lg\:px-2{padding-left:8px;padding-right:8px}.scalar-client .lg\:px-2\.5{padding-left:10px;padding-right:10px}.scalar-client .lg\:pr-24{padding-right:96px}.scalar-client .lg\:pt-1{padding-top:4px}.scalar-client .lg\:text-sm{font-size:var(--scalar-font-size-4)}}@media (min-width: 1200px){.scalar-client .xl\:\!flex{display:flex!important}.scalar-client .xl\:flex{display:flex}.scalar-client .xl\:hidden{display:none}.scalar-client .xl\:h-full{height:100%}.scalar-client .xl\:min-h-client-header{min-height:48px}.scalar-client .xl\:min-w-0{min-width:0px}.scalar-client .xl\:min-w-\[720px\]{min-width:720px}.scalar-client .xl\:max-w-\[720px\]{max-width:720px}.scalar-client .xl\:flex-row{flex-direction:row}.scalar-client .xl\:overflow-hidden{overflow:hidden}.scalar-client .xl\:rounded-none{border-radius:0}.scalar-client .xl\:pl-2{padding-left:8px}.scalar-client .xl\:pr-0{padding-right:0}.scalar-client .xl\:pr-0\.5{padding-right:2px}.scalar-client .xl\:\*\:border-l-1\/2>*{border-left-width:calc(var(--scalar-border-width) / 2)}.scalar-client .xl\:\*\:border-t-0>*{border-top-width:0}.scalar-client .xl\:first\:\*\:border-l-0>*:first-child{border-left-width:0}}.scalar-client .dark\:bg-b-2:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .dark\:hover\:bg-b-2:hover:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .\[\&\>a\]\:pl-5>a{padding-left:20px}.scalar-client .\[\&\>a\]\:pl-\[1\.625rem\]>a{padding-left:1.625rem}html:has(.dark-mode){--scalar-background-1: #0f0f0f}html,body{background-color:var(--scalar-background-1);overscroll-behavior:none}#scalar-client{display:flex;flex-direction:column;height:100dvh;width:100dvw;position:relative;background-color:var(--scalar-background-2)}.dark-mode #scalar-client{background-color:color-mix(in srgb,var(--scalar-background-1) 65%,black)}.open-api-client-button[data-v-2e9f75f1]{cursor:pointer;width:100%;padding:9px 12px;height:31px;display:block;text-align:center;display:flex;align-items:center;justify-content:center;font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);line-height:1.385;text-decoration:none;border-radius:var(--scalar-radius);box-shadow:0 0 0 .5px var(--scalar-border-color);gap:6px;color:var(--scalar-sidebar-color-1)}.open-api-client-button[data-v-2e9f75f1]:hover{background:var( --scalar-sidebar-item-hover-background, var(--scalar-background-2) )}.addressbar-history-button[data-v-a323ca24]:hover{background:var(--scalar-background-3)}.addressbar-history-button[data-v-a323ca24]:focus-within{background:var(--scalar-background-2)}.addressbar-bg-states:has(.cm-focused) .codemirror-bg-switcher[data-v-a323ca24]{--scalar-background-1: var(--scalar-background-1)}.addressbar-bg-states[data-v-a323ca24]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2))}.addressbar-bg-states[data-v-a323ca24]:has(.cm-focused){background:var(--scalar-background-1);border-color:var(--scalar-border-color);outline:1px solid var(--scalar-color-accent)}.description[data-v-8eba7808] .markdown{font-weight:var(--scalar-semibold);color:var(--scalar-color--1);padding:0;display:block}.description[data-v-8eba7808] .markdown>*:first-child{margin-top:0}[data-v-b4d63e6c] .cm-editor{background-color:var(--scalar-background-1);height:100%;outline:none;width:100%}[data-v-b4d63e6c] .cm-content{padding:0;display:flex;align-items:center;font-size:var(--scalar-mini)}.scroll-timeline-x[data-v-b4d63e6c]{scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;-ms-overflow-style:none}.scroll-timeline-x-hidden[data-v-b4d63e6c]{overflow-x:auto}.scroll-timeline-x-hidden[data-v-b4d63e6c] .cm-scroller{scrollbar-width:none;-ms-overflow-style:none;padding-right:20px;overflow:auto}.scroll-timeline-x-hidden[data-v-b4d63e6c]::-webkit-scrollbar{width:0;height:0;display:none}.scroll-timeline-x-hidden[data-v-b4d63e6c] .cm-scroller::-webkit-scrollbar{width:0;height:0;display:none}.scroll-timeline-x-address[data-v-b4d63e6c]{line-height:27px;scrollbar-width:none}.scroll-timeline-x-address[data-v-b4d63e6c]:after{content:"";position:absolute;height:100%;width:24px;right:0;cursor:text}.scroll-timeline-x-address[data-v-b4d63e6c]:empty:before{content:"Enter URL or cURL request";color:var(--scalar-color-3);pointer-events:none}.fade-left[data-v-b4d63e6c],.fade-right[data-v-b4d63e6c]{content:"";position:sticky;height:100%;animation-name:fadein-b4d63e6c;animation-duration:1ms;animation-direction:reverse;animation-timeline:--scroll-timeline;pointer-events:none}.fade-left[data-v-b4d63e6c]{background:linear-gradient(-90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 30%,var(--scalar-background-1) 100%);left:0;min-width:3px;animation-direction:normal}.fade-right[data-v-b4d63e6c]{background:linear-gradient(90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 30%,var(--scalar-background-1) 100%);right:0;min-width:24px}@keyframes fadein-b4d63e6c{0%{opacity:0}1%{opacity:1}}.codemirror-bg-switcher[data-v-b4d63e6c]{--scalar-background-1: color-mix( in srgb, var(--scalar-background-1), var(--scalar-background-2) )}.addressbar-bg-states:has(.cm-focused) .codemirror-bg-switcher[data-v-b4d63e6c]{--scalar-background-1: var(--scalar-background-1)}.addressbar-bg-states[data-v-b4d63e6c]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2))}.addressbar-bg-states[data-v-b4d63e6c]:has(.cm-focused){background:var(--scalar-background-1);border-color:var(--scalar-border-color);outline:1px solid var(--scalar-color-accent)}.sidebar-height[data-v-5397f2a7]{min-height:100%}@media (min-width: 800px){.sidebar-mask[data-v-5397f2a7]{-webkit-mask-image:linear-gradient(0,transparent 0,transparent 0,var(--scalar-background-2) 30px);mask-image:linear-gradient(0,transparent 0,transparent 0,var(--scalar-background-2) 30px)}}.resizer[data-v-5397f2a7]{width:5px;cursor:col-resize;position:absolute;top:0;right:0;bottom:0;border-right:2px solid transparent;transition:border-right-color .3s}.resizer[data-v-5397f2a7]:hover,.dragging .resizer[data-v-5397f2a7]{border-right-color:var(--scalar-background-3)}.dragging[data-v-5397f2a7]{cursor:col-resize}.dragging[data-v-5397f2a7]:before{content:"";display:block;position:absolute;width:100%;height:100%}.scroll-timeline-x[data-v-b77d5bd3]{overflow:auto;scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;-ms-overflow-style:none;scrollbar-width:none}.gitbook-show[data-v-cac03082]{display:none}.app-exit-button[data-v-cac03082]{color:#fff;background:#0000001a}.app-exit-button[data-v-cac03082]:hover{background:#ffffff1a}.ellipsis-position[data-v-daf8ea6e]{transform:translate3d(calc(-100% - 4.5px),0,0)}.dragover-asChild[data-v-3449bf19],.dragover-above[data-v-3449bf19],.dragover-below[data-v-3449bf19]{position:relative}.dragover-above[data-v-3449bf19]:after,.dragover-below[data-v-3449bf19]:after{content:"";position:absolute;top:-1.5px;height:3px;width:100%;display:block;background:color-mix(in srgb,var(--scalar-color-blue),transparent 85%);pointer-events:none;border-radius:var(--scalar-radius)}.dragover-below[data-v-3449bf19]:after{top:initial;bottom:-1.5px}.dragover-asChild[data-v-3449bf19]:after{content:"";position:absolute;top:0;left:0;width:100%;height:100%;display:block;background:color-mix(in srgb,var(--scalar-color-blue),transparent 85%);pointer-events:none;border-radius:var(--scalar-radius)}.indent-border-line-offset[data-v-682d438e]:before{left:var(--0ff55690)}.indent-padding-left[data-v-682d438e]{padding-left:calc(var(--ecb155a4) + 6px)}.sidebar-folderitem[data-v-682d438e] .ellipsis-position{right:6px;transform:none}.search-button-fade[data-v-b7bf7c07]{background:linear-gradient(var(--scalar-background-1) 32px,color-mix(in srgb,var(--scalar-background-1),transparent) 38px,transparent)}.empty-sidebar-item-content[data-v-b7bf7c07]{display:none}.empty-sidebar-item .empty-sidebar-item-content[data-v-b7bf7c07]{display:block}.rabbitjump[data-v-b7bf7c07]{opacity:0}.empty-sidebar-item:hover .rabbitjump[data-v-b7bf7c07]{opacity:1;animation:rabbitAnimation-b7bf7c07 .5s steps(1) infinite}.empty-sidebar-item:hover .rabbitsit[data-v-b7bf7c07]{opacity:0;animation:rabbitAnimation2-b7bf7c07 .5s steps(1) infinite}.empty-sidebar-item:hover .rabbit-ascii[data-v-b7bf7c07]{animation:rabbitRun-b7bf7c07 8s infinite linear}@keyframes rabbitRun-b7bf7c07{0%{transform:translateZ(0)}25%{transform:translate3d(250px,0,0)}25.01%{transform:translate3d(-250px,0,0)}75%{transform:translate3d(250px,0,0)}75.01%{transform:translate3d(-250px,0,0)}to{transform:translateZ(0)}}@keyframes rabbitAnimation-b7bf7c07{0%,to{opacity:1}50%{opacity:0}}@keyframes rabbitAnimation2-b7bf7c07{0%,to{opacity:0}50%{opacity:1;transform:translate3d(0,-8px,0)}}.request-text-color-text[data-v-02ba9b2d]{color:var(--scalar-color-1);background:linear-gradient(var(--scalar-background-1),var(--scalar-background-3));box-shadow:0 0 0 1px var(--scalar-border-color)}@media screen and (max-width: 800px){.sidebar-active-hide-layout[data-v-02ba9b2d]{display:none}.sidebar-active-width[data-v-02ba9b2d]{width:100%}}.active-link[data-v-a6c1ff1c]{background-color:var(--scalar-background-2);color:var(--scalar-color-1);--tw-bg-base: var(--scalar-background-2)}.empty-variable-name[data-v-a6c1ff1c]:empty:before{content:"Untitled";color:var(--scalar-color-3)}.form-group[data-v-c25da223]{margin-bottom:1rem}.modal-actions[data-v-c25da223]{display:flex;justify-content:flex-end;gap:1rem}:root{--scalar-loaded-api-reference: true}@layer scalar-config{.scalar-api-reference[data-v-10065932]{--refs-sidebar-width: var(--scalar-sidebar-width, 0px);--refs-header-height: calc( var(--scalar-y-offset) + var(--scalar-header-height, 0px) );--refs-content-max-width: var(--scalar-content-max-width, 1540px)}.scalar-api-reference.references-classic[data-v-10065932]{--refs-content-max-width: var(--scalar-content-max-width, 1420px);min-height:100dvh;--refs-sidebar-width: 0}}.references-layout[data-v-10065932]{min-height:100dvh;min-width:100%;max-width:100%;flex:1;--full-height: 100dvh;display:grid;grid-template-rows:var(--scalar-header-height, 0px) repeat(2,auto);grid-template-columns:var(--refs-sidebar-width) 1fr;grid-template-areas:"header header" "navigation rendered" "footer footer";background:var(--scalar-background-1)}.references-header[data-v-10065932]{grid-area:header;position:sticky;top:var(--scalar-custom-header-height, 0px);z-index:10;height:var(--scalar-header-height, 0px)}.references-editor[data-v-10065932]{grid-area:editor;display:flex;min-width:0;background:var(--scalar-background-1);z-index:1}.references-navigation[data-v-10065932]{grid-area:navigation}.references-rendered[data-v-10065932]{position:relative;grid-area:rendered;min-width:0;background:var(--scalar-background-1)}.scalar-api-reference.references-classic[data-v-10065932],.references-classic .references-rendered[data-v-10065932]{--full-height: fit-content !important;height:initial!important;max-height:initial!important}.references-navigation-list[data-v-10065932]{position:sticky;top:var(--refs-header-height);height:calc(100dvh - var(--refs-header-height));background:var(--scalar-sidebar-background-1, var(--scalar-background-1));overflow-y:auto;display:flex;flex-direction:column}.references-editor-textarea[data-v-10065932]{position:sticky;top:var(--refs-header-height);height:calc(var(--full-height) - var(--refs-header-height));display:flex;min-width:0;flex:1}.references-editable[data-v-10065932]{grid-template-columns:var(--refs-sidebar-width) 1fr 1fr;grid-template-areas:"header header header" "navigation editor rendered" "footer footer footer"}@layer scalar-config{.references-sidebar[data-v-10065932]{--refs-sidebar-width: var(--scalar-sidebar-width, 280px)}}.references-footer[data-v-10065932]{grid-area:footer}@media (max-width: 1150px){.references-layout[data-v-10065932]{grid-template-columns:var(--refs-sidebar-width) 1fr 0px}}@media (max-width: 1000px){.references-layout[data-v-10065932]{grid-template-columns:auto;grid-template-rows:var(--scalar-header-height, 0px) 0px auto auto;grid-template-areas:"header" "navigation" "rendered" "footer"}.references-editable[data-v-10065932]{grid-template-areas:"header" "navigation" "editor"}.references-navigation[data-v-10065932],.references-rendered[data-v-10065932]{max-height:unset}.references-rendered[data-v-10065932]{position:static}.references-navigation[data-v-10065932]{display:none;z-index:10}.references-sidebar-mobile-open .references-navigation[data-v-10065932]{display:block;top:var(--refs-header-height);height:calc(100dvh - var(--refs-header-height));width:100%;position:sticky}.references-navigation-list[data-v-10065932]{position:absolute;top:-1px;height:calc(var(--full-height) - var(--refs-header-height) + 1px);width:100%;border-top:1px solid var(--scalar-border-color);display:flex;flex-direction:column}}.references-classic-header[data-v-a860f78f]{display:flex;align-items:center;justify-content:space-between;gap:12px;max-width:var(--refs-content-max-width);margin:auto;padding:12px 0}.references-classic-header-container[data-v-a860f78f]{padding:0 60px}@container narrow-references-container (max-width: 900px){.references-classic-header[data-v-a860f78f]{padding:12px 24px}.references-classic-header-container[data-v-a860f78f]{padding:0}}.references-classic-header-icon[data-v-a860f78f]{height:24px;color:var(--scalar-color-1)}a[data-v-829b4762]{text-decoration:none}.ref-search-container[data-v-829b4762]{display:flex;flex-direction:column;padding:12px 12px 0}.ref-search-results[data-v-829b4762]{padding:12px}.ref-search-meta[data-v-829b4762]{background:var(--scalar-background-3);padding:6px 12px;font-size:var(--scalar-font-size-4);color:var(--scalar-color-3);font-weight:var(--scalar-semibold);display:flex;gap:12px}.deprecated[data-v-829b4762]{text-decoration:line-through}.sidebar-search[data-v-9f0bd6b4]{display:flex;align-items:center;position:relative;padding:0 3px 0 9px;min-width:254px;max-width:100%;font-family:var(--scalar-font);background:var( --scalar-sidebar-search-background, var(--scalar-background-1) );color:var(--scalar-sidebar-color-2, var(--scalar-color-2));border-radius:var(--scalar-radius);box-shadow:0 0 0 .5px var(--scalar-sidebar-search-border-color, var(--scalar-border-color));cursor:pointer;-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none}.sidebar-search-input[data-v-9f0bd6b4]{font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);height:31px;-webkit-user-select:none;-moz-user-select:none;user-select:none;z-index:10;position:relative;display:flex;width:100%;justify-content:space-between;align-items:center}.sidebar-search-key[data-v-9f0bd6b4]{text-transform:uppercase;background-color:var(--scalar-background-2);padding:3px 5px;margin:2px;border-radius:var(--scalar-radius);color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}.scalar-search-icon[data-v-9f0bd6b4]{padding:0;margin-right:6px;width:12px}.references-mobile-header[data-v-b5fb4b8b]{display:none;align-items:center;height:100%;width:100%;padding:0 8px;background:var(--scalar-background-1);border-bottom:1px solid var(--scalar-border-color)}.references-mobile-breadcrumbs[data-v-b5fb4b8b]{flex:1;min-width:0;font-size:var(--scalar-small);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);overflow:hidden;white-space:nowrap;text-overflow:ellipsis}.references-mobile-header-actions[data-v-b5fb4b8b]{display:flex;flex-direction:row;gap:4px;height:24px;align-items:center;padding-left:4px}@media (max-width: 1000px){.references-mobile-header[data-v-b5fb4b8b]{display:flex}}@media (max-width: 1000px){.scalar-api-references-standalone-mobile{--scalar-header-height: 50px}}.scalar-api-references-standalone-search[data-v-3ca2ed96]{display:flex;flex-direction:column;padding:12px 12px 6px}.darklight-reference[data-v-3ca2ed96]{width:100%;margin-top:auto}@layer scalar-base{body{margin:0;background-color:var(--scalar-background-1)}}
/*! tailwindcss v4.0.7 | MIT License | https://tailwindcss.com */
@layer theme{:root,:host{--font-sans:"Lato",ui-sans-serif,system-ui,sans-serif,"Apple Color Emoji","Segoe UI Emoji","Segoe UI Symbol","Noto Color Emoji";--font-serif:"Lato",ui-serif,Georgia,Cambria,"Times New Roman",Times,serif;--font-mono:"Fira Mono",ui-monospace,SFMono-Regular,Menlo,Monaco,Consolas,"Liberation Mono","Courier New",monospace;--spacing:.25rem;--breakpoint-sm:40rem;--breakpoint-md:48rem;--breakpoint-lg:64rem;--breakpoint-xl:80rem;--breakpoint-2xl:96rem;--container-3xs:16rem;--container-2xs:18rem;--container-xs:20rem;--container-sm:24rem;--container-md:28rem;--container-lg:32rem;--container-xl:36rem;--container-2xl:42rem;--container-3xl:48rem;--container-4xl:56rem;--container-5xl:64rem;--container-6xl:72rem;--container-7xl:80rem;--text-xs:.75rem;--text-xs--line-height:calc(1/.75);--text-sm:.875rem;--text-sm--line-height:calc(1.25/.875);--text-base:1rem;--text-base--line-height:calc(1.5/1);--text-lg:1.125rem;--text-lg--line-height:calc(1.75/1.125);--text-xl:1.25rem;--text-xl--line-height:calc(1.75/1.25);--text-2xl:1.5rem;--text-2xl--line-height:calc(2/1.5);--text-3xl:1.875rem;--text-3xl--line-height:calc(2.25/1.875);--text-4xl:2.25rem;--text-4xl--line-height:calc(2.5/2.25);--text-5xl:3rem;--text-5xl--line-height:1;--text-6xl:3.75rem;--text-6xl--line-height:1;--text-7xl:4.5rem;--text-7xl--line-height:1;--text-8xl:6rem;--text-8xl--line-height:1;--text-9xl:8rem;--text-9xl--line-height:1;--font-weight-thin:100;--font-weight-extralight:200;--font-weight-light:300;--font-weight-normal:400;--font-weight-medium:500;--font-weight-semibold:600;--font-weight-bold:700;--font-weight-extrabold:800;--font-weight-black:900;--tracking-tighter:-.05em;--tracking-tight:-.025em;--tracking-normal:0em;--tracking-wide:.025em;--tracking-wider:.05em;--tracking-widest:.1em;--leading-tight:1.25;--leading-snug:1.375;--leading-normal:1.5;--leading-relaxed:1.625;--leading-loose:2;--radius-xs:.125rem;--radius-sm:.25rem;--radius-md:.375rem;--radius-lg:.5rem;--radius-xl:.75rem;--radius-2xl:1rem;--radius-3xl:1.5rem;--radius-4xl:2rem;--shadow-2xs:0 1px #0000000d;--shadow-xs:0 1px 2px 0 #0000000d;--shadow-sm:0 1px 3px 0 #0000001a,0 1px 2px -1px #0000001a;--shadow-md:0 4px 6px -1px #0000001a,0 2px 4px -2px #0000001a;--shadow-lg:0 10px 15px -3px #0000001a,0 4px 6px -4px #0000001a;--shadow-xl:0 20px 25px -5px #0000001a,0 8px 10px -6px #0000001a;--shadow-2xl:0 25px 50px -12px #00000040;--inset-shadow-2xs:inset 0 1px #0000000d;--inset-shadow-xs:inset 0 1px 1px #0000000d;--inset-shadow-sm:inset 0 2px 4px #0000000d;--drop-shadow-xs:0 1px 1px #0000000d;--drop-shadow-sm:0 1px 2px #00000026;--drop-shadow-md:0 3px 3px #0000001f;--drop-shadow-lg:0 4px 4px #00000026;--drop-shadow-xl:0 9px 7px #0000001a;--drop-shadow-2xl:0 25px 25px #00000026;--ease-in:cubic-bezier(.4,0,1,1);--ease-out:cubic-bezier(0,0,.2,1);--ease-in-out:cubic-bezier(.4,0,.2,1);--animate-spin:spin 1s linear infinite;--animate-ping:ping 1s cubic-bezier(0,0,.2,1)infinite;--animate-pulse:pulse 2s cubic-bezier(.4,0,.6,1)infinite;--animate-bounce:bounce 1s infinite;--blur-xs:4px;--blur-sm:8px;--blur-md:12px;--blur-lg:16px;--blur-xl:24px;--blur-2xl:40px;--blur-3xl:64px;--perspective-dramatic:100px;--perspective-near:300px;--perspective-normal:500px;--perspective-midrange:800px;--perspective-distant:1200px;--aspect-video:16/9;--default-transition-duration:.15s;--default-transition-timing-function:cubic-bezier(.4,0,.2,1);--default-font-family:var(--font-sans);--default-font-feature-settings:var(--font-sans--font-feature-settings);--default-font-variation-settings:var(--font-sans--font-variation-settings);--default-mono-font-family:var(--font-mono);--default-mono-font-feature-settings:var(--font-mono--font-feature-settings);--default-mono-font-variation-settings:var(--font-mono--font-variation-settings);--font-alternative:"Montserrat",ui-sans-serif,system-ui,sans-serif,"Apple Color Emoji","Segoe UI Emoji","Segoe UI Symbol","Noto Color Emoji";--color-vortico:teal;--color-vortico-50:#00dcdc;--color-vortico-100:#00d2d2;--color-vortico-200:#00bdbd;--color-vortico-300:#00a9a9;--color-vortico-400:#009494;--color-vortico-500:teal;--color-vortico-600:#007171;--color-vortico-700:#006161;--color-vortico-800:#005252;--color-vortico-900:#004343;--color-vortico-950:#003b3b;--color-bruma:#00bbd5;--color-bruma-50:#ebfeff;--color-bruma-100:#cefcff;--color-bruma-200:#a2f7ff;--color-bruma-300:#63effd;--color-bruma-400:#1cdcf4;--color-bruma-500:#00bbd5;--color-bruma-600:#0398b7;--color-bruma-700:#0a7994;--color-bruma-800:#126178;--color-bruma-900:#145165;--color-bruma-950:#063646;--color-bosque:#9e744f;--color-bosque-50:#f7f4ef;--color-bosque-100:#ebe3d6;--color-bosque-200:#d9c8af;--color-bosque-300:#c2a682;--color-bosque-400:#b0895f;--color-bosque-500:#9e744f;--color-bosque-600:#8a5f44;--color-bosque-700:#6f4939;--color-bosque-800:#5f3f34;--color-bosque-900:#523731;--color-bosque-950:#2f1d19;--color-ciclon:#00976e;--color-ciclon-50:#ebfef6;--color-ciclon-100:#d0fbe7;--color-ciclon-200:#a4f6d3;--color-ciclon-300:#6aebbd;--color-ciclon-400:#2fd8a1;--color-ciclon-500:#0abf8a;--color-ciclon-600:#00976e;--color-ciclon-700:#007c5e;--color-ciclon-800:#03624b;--color-ciclon-900:#04503f;--color-ciclon-950:#012d25;--color-flama:#e25822;--color-flama-50:#fdf5ef;--color-flama-100:#fbe8d9;--color-flama-200:#f6cdb2;--color-flama-300:#f0ac81;--color-flama-400:#e9804e;--color-flama-500:#e25822;--color-flama-600:#d54821;--color-flama-700:#b1351d;--color-flama-800:#8d2d1f;--color-flama-900:#72271c;--color-flama-950:#3d110d;--color-primary:#656775;--color-primary-50:#f5f5f6;--color-primary-100:#e5e6e8;--color-primary-200:#cdced4;--color-primary-300:#aaabb6;--color-primary-400:#808290;--color-primary-500:#656775;--color-primary-600:#575863;--color-primary-700:#4a4a54;--color-primary-800:#3f3f46;--color-primary-900:#3a3a3f;--color-primary-950:#242428;--breakpoint-3xl:120rem;--container-8xl:96rem;--container-9xl:120rem}}@layer base{*,:after,:before,::backdrop{box-sizing:border-box;border:0 solid;margin:0;padding:0}::file-selector-button{box-sizing:border-box;border:0 solid;margin:0;padding:0}html,:host{-webkit-text-size-adjust:100%;tab-size:4;line-height:1.5;font-family:var(--default-font-family,ui-sans-serif,system-ui,sans-serif,"Apple Color Emoji","Segoe UI Emoji","Segoe UI Symbol","Noto Color Emoji");font-feature-settings:var(--default-font-feature-settings,normal);font-variation-settings:var(--default-font-variation-settings,normal);-webkit-tap-highlight-color:transparent}body{line-height:inherit}hr{height:0;color:inherit;border-top-width:1px}abbr:where([title]){-webkit-text-decoration:underline dotted;text-decoration:underline dotted}h1,h2,h3,h4,h5,h6{font-size:inherit;font-weight:inherit}a{color:inherit;-webkit-text-decoration:inherit;-webkit-text-decoration:inherit;-webkit-text-decoration:inherit;text-decoration:inherit}b,strong{font-weight:bolder}code,kbd,samp,pre{font-family:var(--default-mono-font-family,ui-monospace,SFMono-Regular,Menlo,Monaco,Consolas,"Liberation Mono","Courier New",monospace);font-feature-settings:var(--default-mono-font-feature-settings,normal);font-variation-settings:var(--default-mono-font-variation-settings,normal);font-size:1em}small{font-size:80%}sub,sup{vertical-align:baseline;font-size:75%;line-height:0;position:relative}sub{bottom:-.25em}sup{top:-.5em}table{text-indent:0;border-color:inherit;border-collapse:collapse}:-moz-focusring{outline:auto}progress{vertical-align:baseline}summary{display:list-item}ol,ul,menu{list-style:none}img,svg,video,canvas,audio,iframe,embed,object{vertical-align:middle;display:block}img,video{max-width:100%;height:auto}button,input,select,optgroup,textarea{font:inherit;font-feature-settings:inherit;font-variation-settings:inherit;letter-spacing:inherit;color:inherit;opacity:1;background-color:#0000;border-radius:0}::file-selector-button{font:inherit;font-feature-settings:inherit;font-variation-settings:inherit;letter-spacing:inherit;color:inherit;opacity:1;background-color:#0000;border-radius:0}:where(select:is([multiple],[size])) optgroup{font-weight:bolder}:where(select:is([multiple],[size])) optgroup option{padding-inline-start:20px}::file-selector-button{margin-inline-end:4px}::placeholder{opacity:1;color:color-mix(in oklab,currentColor 50%,transparent)}textarea{resize:vertical}::-webkit-search-decoration{-webkit-appearance:none}::-webkit-date-and-time-value{min-height:1lh;text-align:inherit}::-webkit-datetime-edit{display:inline-flex}::-webkit-datetime-edit-fields-wrapper{padding:0}::-webkit-datetime-edit{padding-block:0}::-webkit-datetime-edit-year-field{padding-block:0}::-webkit-datetime-edit-month-field{padding-block:0}::-webkit-datetime-edit-day-field{padding-block:0}::-webkit-datetime-edit-hour-field{padding-block:0}::-webkit-datetime-edit-minute-field{padding-block:0}::-webkit-datetime-edit-second-field{padding-block:0}::-webkit-datetime-edit-millisecond-field{padding-block:0}::-webkit-datetime-edit-meridiem-field{padding-block:0}:-moz-ui-invalid{box-shadow:none}button,input:where([type=button],[type=reset],[type=submit]){appearance:button}::file-selector-button{appearance:button}::-webkit-inner-spin-button{height:auto}::-webkit-outer-spin-button{height:auto}[hidden]:where(:not([hidden=until-found])){display:none!important}}@layer components;@layer utilities{.absolute{position:absolute}.fixed{position:fixed}.relative{position:relative}.inset-0{inset:calc(var(--spacing)*0)}.inset-x-0{inset-inline:calc(var(--spacing)*0)}.inset-x-4{inset-inline:calc(var(--spacing)*4)}.inset-y-\[5vh\]{inset-block:5vh}.top-0{top:calc(var(--spacing)*0)}.top-4{top:calc(var(--spacing)*4)}.right-4{right:calc(var(--spacing)*4)}.-bottom-2{bottom:calc(var(--spacing)*-2)}.left-1\/2{left:50%}.z-10{z-index:10}.z-20{z-index:20}.z-50{z-index:50}.mx-auto{margin-inline:auto}.my-14{margin-block:calc(var(--spacing)*14)}.-mt-px{margin-top:-1px}.mt-\[calc\(3\.5rem\+1px\)\]{margin-top:calc(3.5rem + 1px)}.mt-px{margin-top:1px}.mb-px{margin-bottom:1px}.-ml-px{margin-left:-1px}.ml-1{margin-left:calc(var(--spacing)*1)}.block{display:block}.flex{display:flex}.hidden{display:none}.inline{display:inline}.table{display:table}.h-2{height:calc(var(--spacing)*2)}.h-4{height:calc(var(--spacing)*4)}.h-5{height:calc(var(--spacing)*5)}.h-6{height:calc(var(--spacing)*6)}.h-7{height:calc(var(--spacing)*7)}.h-8{height:calc(var(--spacing)*8)}.h-9{height:calc(var(--spacing)*9)}.h-12{height:calc(var(--spacing)*12)}.h-14{height:calc(var(--spacing)*14)}.h-\[672px\]{height:672px}.h-\[calc\(100vh-6\.5rem-2px\)\]{height:calc(100vh - 6.5rem - 2px)}.h-fit{height:fit-content}.h-full{height:100%}.h-screen{height:100vh}.max-h-0{max-height:calc(var(--spacing)*0)}.max-h-\[32px\]{max-height:32px}.max-h-\[calc\(85vh\)\]{max-height:85vh}.max-h-\[calc\(85vh\+2rem\+3px\)\]{max-height:calc(85vh + 2rem + 3px)}.max-h-\[calc\(100\%-2rem-1px\)\]{max-height:calc(100% - 2rem - 1px)}.max-h-full{max-height:100%}.min-h-full{min-height:100%}.w-2{width:calc(var(--spacing)*2)}.w-4{width:calc(var(--spacing)*4)}.w-5{width:calc(var(--spacing)*5)}.w-6{width:calc(var(--spacing)*6)}.w-7{width:calc(var(--spacing)*7)}.w-8{width:calc(var(--spacing)*8)}.w-9{width:calc(var(--spacing)*9)}.w-40{width:calc(var(--spacing)*40)}.w-fit{width:fit-content}.w-full{width:100%}.w-screen{width:100vw}.max-w-7xl{max-width:var(--container-7xl)}.max-w-8xl{max-width:var(--container-8xl)}.flex-auto{flex:auto}.flex-none{flex:none}.table-fixed{table-layout:fixed}.border-collapse{border-collapse:collapse}.-translate-x-1\/2{--tw-translate-x:calc(calc(1/2*100%)*-1);translate:var(--tw-translate-x)var(--tw-translate-y)}.translate-y-full{--tw-translate-y:100%;translate:var(--tw-translate-x)var(--tw-translate-y)}.rotate-90{rotate:90deg}.cursor-pointer{cursor:pointer}.flex-col{flex-direction:column}.items-center{align-items:center}.items-start{align-items:flex-start}.justify-between{justify-content:space-between}.justify-center{justify-content:center}.justify-end{justify-content:flex-end}.justify-start{justify-content:flex-start}.gap-1{gap:calc(var(--spacing)*1)}.gap-2{gap:calc(var(--spacing)*2)}.gap-8{gap:calc(var(--spacing)*8)}.gap-x-2{column-gap:calc(var(--spacing)*2)}:where(.divide-x>:not(:last-child)){--tw-divide-x-reverse:0;border-inline-style:var(--tw-border-style);border-inline-start-width:calc(1px*var(--tw-divide-x-reverse));border-inline-end-width:calc(1px*calc(1 - var(--tw-divide-x-reverse)))}:where(.divide-y>:not(:last-child)){--tw-divide-y-reverse:0;border-bottom-style:var(--tw-border-style);border-top-style:var(--tw-border-style);border-top-width:calc(1px*var(--tw-divide-y-reverse));border-bottom-width:calc(1px*calc(1 - var(--tw-divide-y-reverse)))}:where(.divide-flama-500\/50>:not(:last-child)){border-color:color-mix(in oklab,var(--color-flama-500)50%,transparent)}:where(.divide-vortico-500\/50>:not(:last-child)){border-color:color-mix(in oklab,var(--color-vortico-500)50%,transparent)}.truncate{text-overflow:ellipsis;white-space:nowrap;overflow:hidden}.overflow-auto{overflow:auto}.overflow-hidden{overflow:hidden}.overflow-visible{overflow:visible}.overflow-y-scroll{overflow-y:scroll}.rounded{border-radius:.25rem}.rounded-lg{border-radius:var(--radius-lg)}.border{border-style:var(--tw-border-style);border-width:1px}.border-t{border-top-style:var(--tw-border-style);border-top-width:1px}.border-r{border-right-style:var(--tw-border-style);border-right-width:1px}.border-b{border-bottom-style:var(--tw-border-style);border-bottom-width:1px}.border-l-4{border-left-style:var(--tw-border-style);border-left-width:4px}.border-current{border-color:currentColor}.border-flama-500{border-color:var(--color-flama-500)}.border-flama-500\/50{border-color:color-mix(in oklab,var(--color-flama-500)50%,transparent)}.border-primary-100{border-color:var(--color-primary-100)}.border-primary-300{border-color:var(--color-primary-300)}.border-primary-500{border-color:var(--color-primary-500)}.border-transparent{border-color:#0000}.border-vortico-500\/50{border-color:color-mix(in oklab,var(--color-vortico-500)50%,transparent)}.bg-flama-500\/25{background-color:color-mix(in oklab,var(--color-flama-500)25%,transparent)}.bg-flama-700{background-color:var(--color-flama-700)}.bg-primary-100{background-color:var(--color-primary-100)}.bg-primary-200{background-color:var(--color-primary-200)}.bg-primary-500\/30{background-color:color-mix(in oklab,var(--color-primary-500)30%,transparent)}.bg-primary-800{background-color:var(--color-primary-800)}.bg-primary-900{background-color:var(--color-primary-900)}.bg-primary-950\/30{background-color:color-mix(in oklab,var(--color-primary-950)30%,transparent)}.bg-linear-to-b{--tw-gradient-position:to bottom in oklab;background-image:linear-gradient(var(--tw-gradient-stops))}.from-flama-500\/10{--tw-gradient-from:color-mix(in oklab,var(--color-flama-500)10%,transparent);--tw-gradient-stops:var(--tw-gradient-via-stops,var(--tw-gradient-position),var(--tw-gradient-from)var(--tw-gradient-from-position),var(--tw-gradient-to)var(--tw-gradient-to-position))}.p-2{padding:calc(var(--spacing)*2)}.p-3{padding:calc(var(--spacing)*3)}.px-2{padding-inline:calc(var(--spacing)*2)}.px-4{padding-inline:calc(var(--spacing)*4)}.py-1{padding-block:calc(var(--spacing)*1)}.py-2{padding-block:calc(var(--spacing)*2)}.py-6{padding-block:calc(var(--spacing)*6)}.py-8{padding-block:calc(var(--spacing)*8)}.pr-3{padding-right:calc(var(--spacing)*3)}.pl-1{padding-left:calc(var(--spacing)*1)}.pl-2{padding-left:calc(var(--spacing)*2)}.pl-4{padding-left:calc(var(--spacing)*4)}.text-left{text-align:left}.text-right{text-align:right}.font-alternative{font-family:var(--font-alternative)}.font-mono{font-family:var(--font-mono)}.text-2xl{font-size:var(--text-2xl);line-height:var(--tw-leading,var(--text-2xl--line-height))}.text-base{font-size:var(--text-base);line-height:var(--tw-leading,var(--text-base--line-height))}.text-lg{font-size:var(--text-lg);line-height:var(--tw-leading,var(--text-lg--line-height))}.text-sm{font-size:var(--text-sm);line-height:var(--tw-leading,var(--text-sm--line-height))}.text-xl{font-size:var(--text-xl);line-height:var(--tw-leading,var(--text-xl--line-height))}.text-xs{font-size:var(--text-xs);line-height:var(--tw-leading,var(--text-xs--line-height))}.leading-4{--tw-leading:calc(var(--spacing)*4);line-height:calc(var(--spacing)*4)}.leading-5{--tw-leading:calc(var(--spacing)*5);line-height:calc(var(--spacing)*5)}.leading-6{--tw-leading:calc(var(--spacing)*6);line-height:calc(var(--spacing)*6)}.leading-7{--tw-leading:calc(var(--spacing)*7);line-height:calc(var(--spacing)*7)}.leading-8{--tw-leading:calc(var(--spacing)*8);line-height:calc(var(--spacing)*8)}.leading-9{--tw-leading:calc(var(--spacing)*9);line-height:calc(var(--spacing)*9)}.font-bold{--tw-font-weight:var(--font-weight-bold);font-weight:var(--font-weight-bold)}.font-medium{--tw-font-weight:var(--font-weight-medium);font-weight:var(--font-weight-medium)}.font-semibold{--tw-font-weight:var(--font-weight-semibold);font-weight:var(--font-weight-semibold)}.text-wrap{text-wrap:wrap}.whitespace-pre{white-space:pre}.text-bosque-200{color:var(--color-bosque-200)}.text-bosque-500{color:var(--color-bosque-500)}.text-bosque-800{color:var(--color-bosque-800)}.text-bruma-200{color:var(--color-bruma-200)}.text-bruma-500{color:var(--color-bruma-500)}.text-bruma-800{color:var(--color-bruma-800)}.text-ciclon-200{color:var(--color-ciclon-200)}.text-ciclon-500{color:var(--color-ciclon-500)}.text-ciclon-800{color:var(--color-ciclon-800)}.text-flama{color:var(--color-flama)}.text-flama-200{color:var(--color-flama-200)}.text-flama-400{color:var(--color-flama-400)}.text-flama-500{color:var(--color-flama-500)}.text-flama-800{color:var(--color-flama-800)}.text-primary-200{color:var(--color-primary-200)}.text-primary-300{color:var(--color-primary-300)}.text-primary-400{color:var(--color-primary-400)}.text-primary-500{color:var(--color-primary-500)}.text-primary-600{color:var(--color-primary-600)}.text-primary-700{color:var(--color-primary-700)}.text-primary-800{color:var(--color-primary-800)}.text-vortico-200{color:var(--color-vortico-200)}.text-vortico-500{color:var(--color-vortico-500)}.text-vortico-800{color:var(--color-vortico-800)}.capitalize{text-transform:capitalize}.italic{font-style:italic}.opacity-0{opacity:0}.opacity-100{opacity:1}.ring-1{--tw-ring-shadow:var(--tw-ring-inset,)0 0 0 calc(1px + var(--tw-ring-offset-width))var(--tw-ring-color,currentColor);box-shadow:var(--tw-inset-shadow),var(--tw-inset-ring-shadow),var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow)}.shadow-md{--tw-shadow:0 4px 6px -1px var(--tw-shadow-color,#0000001a),0 2px 4px -2px var(--tw-shadow-color,#0000001a);box-shadow:var(--tw-inset-shadow),var(--tw-inset-ring-shadow),var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow)}.shadow-primary-300{--tw-shadow-color:var(--color-primary-300)}.ring-primary-500\/50{--tw-ring-color:color-mix(in oklab,var(--color-primary-500)50%,transparent)}.filter{filter:var(--tw-blur,)var(--tw-brightness,)var(--tw-contrast,)var(--tw-grayscale,)var(--tw-hue-rotate,)var(--tw-invert,)var(--tw-saturate,)var(--tw-sepia,)var(--tw-drop-shadow,)}.backdrop-blur{--tw-backdrop-blur:blur(8px);-webkit-backdrop-filter:var(--tw-backdrop-blur,)var(--tw-backdrop-brightness,)var(--tw-backdrop-contrast,)var(--tw-backdrop-grayscale,)var(--tw-backdrop-hue-rotate,)var(--tw-backdrop-invert,)var(--tw-backdrop-opacity,)var(--tw-backdrop-saturate,)var(--tw-backdrop-sepia,);backdrop-filter:var(--tw-backdrop-blur,)var(--tw-backdrop-brightness,)var(--tw-backdrop-contrast,)var(--tw-backdrop-grayscale,)var(--tw-backdrop-hue-rotate,)var(--tw-backdrop-invert,)var(--tw-backdrop-opacity,)var(--tw-backdrop-saturate,)var(--tw-backdrop-sepia,)}.backdrop-blur-sm{--tw-backdrop-blur:blur(var(--blur-sm));-webkit-backdrop-filter:var(--tw-backdrop-blur,)var(--tw-backdrop-brightness,)var(--tw-backdrop-contrast,)var(--tw-backdrop-grayscale,)var(--tw-backdrop-hue-rotate,)var(--tw-backdrop-invert,)var(--tw-backdrop-opacity,)var(--tw-backdrop-saturate,)var(--tw-backdrop-sepia,);backdrop-filter:var(--tw-backdrop-blur,)var(--tw-backdrop-brightness,)var(--tw-backdrop-contrast,)var(--tw-backdrop-grayscale,)var(--tw-backdrop-hue-rotate,)var(--tw-backdrop-invert,)var(--tw-backdrop-opacity,)var(--tw-backdrop-saturate,)var(--tw-backdrop-sepia,)}.transition{transition-property:color,background-color,border-color,outline-color,text-decoration-color,fill,stroke,--tw-gradient-from,--tw-gradient-via,--tw-gradient-to,opacity,box-shadow,transform,translate,scale,rotate,filter,-webkit-backdrop-filter,backdrop-filter;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.transition-all{transition-property:all;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.transition-colors{transition-property:color,background-color,border-color,outline-color,text-decoration-color,fill,stroke,--tw-gradient-from,--tw-gradient-via,--tw-gradient-to;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.transition-opacity{transition-property:opacity;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.transition-transform{transition-property:transform,translate,scale,rotate;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.duration-200{--tw-duration:.2s;transition-duration:.2s}.duration-500{--tw-duration:.5s;transition-duration:.5s}.ring-inset{--tw-ring-inset:inset}.select-none{-webkit-user-select:none;user-select:none}@media (hover:hover){.group-hover\:block:is(:where(.group):hover *){display:block}.group-hover\:opacity-100:is(:where(.group):hover *){opacity:1}}.after\:absolute:after{content:var(--tw-content);position:absolute}.after\:bottom-\[100\%\]:after{content:var(--tw-content);bottom:100%}.after\:left-1\/2:after{content:var(--tw-content);left:50%}.after\:-translate-x-1\/2:after{content:var(--tw-content);--tw-translate-x:calc(calc(1/2*100%)*-1);translate:var(--tw-translate-x)var(--tw-translate-y)}.after\:border-8:after{content:var(--tw-content);border-style:var(--tw-border-style);border-width:8px}.after\:border-x-transparent:after{content:var(--tw-content);border-inline-color:#0000}.after\:border-t-transparent:after{content:var(--tw-content);border-top-color:#0000}.after\:border-b-primary-800:after{content:var(--tw-content);border-bottom-color:var(--color-primary-800)}.after\:content-\[\'\'\]:after{content:var(--tw-content);--tw-content:"";content:var(--tw-content)}.first\:pl-0:first-child{padding-left:calc(var(--spacing)*0)}.last\:pr-0:last-child{padding-right:calc(var(--spacing)*0)}@media (hover:hover){.hover\:border-l-4:hover{border-left-style:var(--tw-border-style);border-left-width:4px}.hover\:border-flama-500:hover{border-color:var(--color-flama-500)}.hover\:bg-flama-300:hover{background-color:var(--color-flama-300)}.hover\:pl-0:hover{padding-left:calc(var(--spacing)*0)}.hover\:text-bosque:hover{color:var(--color-bosque)}.hover\:text-ciclon:hover{color:var(--color-ciclon)}.hover\:text-flama:hover{color:var(--color-flama)}.hover\:text-flama-500:hover{color:var(--color-flama-500)}.hover\:text-primary-800:hover{color:var(--color-primary-800)}.hover\:text-vortico-500:hover{color:var(--color-vortico-500)}.hover\:shadow-lg:hover{--tw-shadow:0 10px 15px -3px var(--tw-shadow-color,#0000001a),0 4px 6px -4px var(--tw-shadow-color,#0000001a);box-shadow:var(--tw-inset-shadow),var(--tw-inset-ring-shadow),var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow)}.hover\:shadow-flama-300:hover{--tw-shadow-color:var(--color-flama-300)}}@media (width>=40rem){.sm\:inset-x-6{inset-inline:calc(var(--spacing)*6)}.sm\:px-6{padding-inline:calc(var(--spacing)*6)}.sm\:py-10{padding-block:calc(var(--spacing)*10)}}@media (width>=48rem){.md\:sticky{position:sticky}.md\:inset-x-8{inset-inline:calc(var(--spacing)*8)}.md\:left-0{left:calc(var(--spacing)*0)}.md\:block{display:block}.md\:flex{display:flex}.md\:hidden{display:none}.md\:h-5{height:calc(var(--spacing)*5)}.md\:h-6{height:calc(var(--spacing)*6)}.md\:h-7{height:calc(var(--spacing)*7)}.md\:h-8{height:calc(var(--spacing)*8)}.md\:h-9{height:calc(var(--spacing)*9)}.md\:h-10{height:calc(var(--spacing)*10)}.md\:w-5{width:calc(var(--spacing)*5)}.md\:w-6{width:calc(var(--spacing)*6)}.md\:w-7{width:calc(var(--spacing)*7)}.md\:w-8{width:calc(var(--spacing)*8)}.md\:w-9{width:calc(var(--spacing)*9)}.md\:w-10{width:calc(var(--spacing)*10)}.md\:px-8{padding-inline:calc(var(--spacing)*8)}.md\:py-12{padding-block:calc(var(--spacing)*12)}.md\:text-2xl{font-size:var(--text-2xl);line-height:var(--tw-leading,var(--text-2xl--line-height))}.md\:text-3xl{font-size:var(--text-3xl);line-height:var(--tw-leading,var(--text-3xl--line-height))}.md\:text-base{font-size:var(--text-base);line-height:var(--tw-leading,var(--text-base--line-height))}.md\:text-lg{font-size:var(--text-lg);line-height:var(--tw-leading,var(--text-lg--line-height))}.md\:text-sm{font-size:var(--text-sm);line-height:var(--tw-leading,var(--text-sm--line-height))}.md\:text-xl{font-size:var(--text-xl);line-height:var(--tw-leading,var(--text-xl--line-height))}.md\:leading-5{--tw-leading:calc(var(--spacing)*5);line-height:calc(var(--spacing)*5)}.md\:leading-6{--tw-leading:calc(var(--spacing)*6);line-height:calc(var(--spacing)*6)}.md\:leading-7{--tw-leading:calc(var(--spacing)*7);line-height:calc(var(--spacing)*7)}.md\:leading-8{--tw-leading:calc(var(--spacing)*8);line-height:calc(var(--spacing)*8)}.md\:leading-9{--tw-leading:calc(var(--spacing)*9);line-height:calc(var(--spacing)*9)}.md\:leading-10{--tw-leading:calc(var(--spacing)*10);line-height:calc(var(--spacing)*10)}}@media (width>=64rem){.lg\:w-xs{width:var(--container-xs)}.lg\:flex-row{flex-direction:row}}}:root{background-color:var(--color-primary-100);color:var(--color-primary-600);scrollbar-color:var(--color-primary-400)transparent;z-index:0}body{height:100%}@keyframes spin{to{transform:rotate(360deg)}}@keyframes ping{75%,to{opacity:0;transform:scale(2)}}@keyframes pulse{50%{opacity:.5}}@keyframes bounce{0%,to{animation-timing-function:cubic-bezier(.8,0,1,1);transform:translateY(-25%)}50%{animation-timing-function:cubic-bezier(0,0,.2,1);transform:none}}@property --tw-translate-x{syntax:"*";inherits:false;initial-value:0}@property --tw-translate-y{syntax:"*";inherits:false;initial-value:0}@property --tw-translate-z{syntax:"*";inherits:false;initial-value:0}@property --tw-divide-x-reverse{syntax:"*";inherits:false;initial-value:0}@property --tw-border-style{syntax:"*";inherits:false;initial-value:solid}@property --tw-divide-y-reverse{syntax:"*";inherits:false;initial-value:0}@property --tw-gradient-position{syntax:"*";inherits:false}@property --tw-gradient-from{syntax:"<color>";inherits:false;initial-value:#0000}@property --tw-gradient-via{syntax:"<color>";inherits:false;initial-value:#0000}@property --tw-gradient-to{syntax:"<color>";inherits:false;initial-value:#0000}@property --tw-gradient-stops{syntax:"*";inherits:false}@property --tw-gradient-via-stops{syntax:"*";inherits:false}@property --tw-gradient-from-position{syntax:"<length-percentage>";inherits:false;initial-value:0%}@property --tw-gradient-via-position{syntax:"<length-percentage>";inherits:false;initial-value:50%}@property --tw-gradient-to-position{syntax:"<length-percentage>";inherits:false;initial-value:100%}@property --tw-leading{syntax:"*";inherits:false}@property --tw-font-weight{syntax:"*";inherits:false}@property --tw-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-shadow-color{syntax:"*";inherits:false}@property --tw-inset-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-inset-shadow-color{syntax:"*";inherits:false}@property --tw-ring-color{syntax:"*";inherits:false}@property --tw-ring-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-inset-ring-color{syntax:"*";inherits:false}@property --tw-inset-ring-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-ring-inset{syntax:"*";inherits:false}@property --tw-ring-offset-width{syntax:"<length>";inherits:false;initial-value:0}@property --tw-ring-offset-color{syntax:"*";inherits:false;initial-value:#fff}@property --tw-ring-offset-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-blur{syntax:"*";inherits:false}@property --tw-brightness{syntax:"*";inherits:false}@property --tw-contrast{syntax:"*";inherits:false}@property --tw-grayscale{syntax:"*";inherits:false}@property --tw-hue-rotate{syntax:"*";inherits:false}@property --tw-invert{syntax:"*";inherits:false}@property --tw-opacity{syntax:"*";inherits:false}@property --tw-saturate{syntax:"*";inherits:false}@property --tw-sepia{syntax:"*";inherits:false}@property --tw-drop-shadow{syntax:"*";inherits:false}@property --tw-backdrop-blur{syntax:"*";inherits:false}@property --tw-backdrop-brightness{syntax:"*";inherits:false}@property --tw-backdrop-contrast{syntax:"*";inherits:false}@property --tw-backdrop-grayscale{syntax:"*";inherits:false}@property --tw-backdrop-hue-rotate{syntax:"*";inherits:false}@property --tw-backdrop-invert{syntax:"*";inherits:false}@property --tw-backdrop-opacity{syntax:"*";inherits:false}@property --tw-backdrop-saturate{syntax:"*";inherits:false}@property --tw-backdrop-sepia{syntax:"*";inherits:false}@property --tw-duration{syntax:"*";inherits:false}@property --tw-content{syntax:"*";inherits:false;initial-value:""}
</style></head><body><div id="app"/><script>(()=>{var e={25:(e,t,n)=>{"use strict";n.d(t,{A:()=>p});var r=n(641),a=n(953),o=n(33),i=n(6251),s=n(1266);const l={class:"bg-b-2 flex items-center"},c={class:"flex flex-1 items-center gap-1.5 text-c-1"},u={key:0,class:"bg-b-2 text-c-2 text-3xs inline-flex h-4 w-4 items-center justify-center rounded-full font-semibold border"},d={key:0,class:"ui-not-open:invisible flex items-center gap-2 pr-1.5"},p=(0,r.pM)({__name:"ViewLayoutCollapse",props:{defaultOpen:{type:Boolean,default:!0},itemCount:{default:0},layout:{default:"client"}},setup:e=>(e,t)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(i.EN),{as:"div",class:(0,o.C4)(["focus-within:text-c-1 text-c-2 request-item border-b",{"first:border-t":"client"===e.layout}]),defaultOpen:e.defaultOpen,static:"reference"===e.layout},{default:(0,r.k6)((({open:t})=>[(0,r.Lk)("div",l,[(0,r.bF)((0,a.R1)(i._q),{class:(0,o.C4)(["group hover:text-c-1 flex flex-1 items-center gap-2.5 overflow-hidden py-1.5 text-sm font-medium max-h-8 px-1 md:px-1.5 xl:pl-2 xl:pr-0.5 outline-none",{"!pl-3":"reference"===e.layout}]),disabled:"reference"===e.layout},{default:(0,r.k6)((()=>["reference"!==e.layout?((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.kv),{key:0,class:(0,o.C4)(["text-c-3 group-hover:text-c-1 group-focus-visible:outline ui-open:rotate-90 ui-not-open:rotate-0 rounded-px outline-offset-2"]),icon:"ChevronRight",size:"md"})):(0,r.Q3)("",!0),(0,r.Lk)("div",c,[(0,r.RG)(e.$slots,"title",{open:t}),!t&&e.itemCount?((0,r.uX)(),(0,r.CE)("span",u,(0,o.v_)(e.itemCount),1)):(0,r.Q3)("",!0)])])),_:2},1032,["class","disabled"]),e.$slots.actions?((0,r.uX)(),(0,r.CE)("div",d,[(0,r.RG)(e.$slots,"actions",{open:t})])):(0,r.Q3)("",!0)]),(0,r.bF)((0,a.R1)(i.kS),(0,r.v6)(e.$attrs,{class:"h-full max-h-fit rounded-b diclosure-panel"}),{default:(0,r.k6)((()=>[(0,r.RG)(e.$slots,"default",{open:t})])),_:2},1040)])),_:3},8,["class","defaultOpen","static"]))})},33:(e,t,n)=>{"use strict";function r(e){const t=Object.create(null);for(const n of e.split(","))t[n]=1;return e=>e in t}n.d(t,{$3:()=>h,$H:()=>L,BH:()=>W,BX:()=>ee,Bm:()=>O,C4:()=>G,CE:()=>m,CP:()=>c,DY:()=>B,Gv:()=>k,J$:()=>Y,Kg:()=>w,MZ:()=>a,Mp:()=>l,NO:()=>s,Oj:()=>o,PT:()=>D,Qd:()=>C,Ro:()=>U,SU:()=>A,TF:()=>d,Tg:()=>$,Tn:()=>y,Tr:()=>X,We:()=>j,X$:()=>u,Y2:()=>J,ZH:()=>N,Zf:()=>_,_B:()=>K,bB:()=>Q,cy:()=>f,gd:()=>b,pD:()=>r,rU:()=>I,tE:()=>i,u3:()=>te,vM:()=>g,v_:()=>re,yI:()=>T,yL:()=>x,yQ:()=>F});const a={},o=[],i=()=>{},s=()=>!1,l=e=>111===e.charCodeAt(0)&&110===e.charCodeAt(1)&&(e.charCodeAt(2)>122||e.charCodeAt(2)<97),c=e=>e.startsWith("onUpdate:"),u=Object.assign,d=(e,t)=>{const n=e.indexOf(t);n>-1&&e.splice(n,1)},p=Object.prototype.hasOwnProperty,h=(e,t)=>p.call(e,t),f=Array.isArray,m=e=>"[object Map]"===S(e),g=e=>"[object Set]"===S(e),v=e=>"[object Date]"===S(e),b=e=>"[object RegExp]"===S(e),y=e=>"function"==typeof e,w=e=>"string"==typeof e,O=e=>"symbol"==typeof e,k=e=>null!==e&&"object"==typeof e,x=e=>(k(e)||y(e))&&y(e.then)&&y(e.catch),E=Object.prototype.toString,S=e=>E.call(e),_=e=>S(e).slice(8,-1),C=e=>"[object Object]"===S(e),T=e=>w(e)&&"NaN"!==e&&"-"!==e[0]&&""+parseInt(e,10)===e,A=r(",key,ref,ref_for,ref_key,onVnodeBeforeMount,onVnodeMounted,onVnodeBeforeUpdate,onVnodeUpdated,onVnodeBeforeUnmount,onVnodeUnmounted"),R=e=>{const t=Object.create(null);return n=>t[n]||(t[n]=e(n))},P=/-(\w)/g,D=R((e=>e.replace(P,((e,t)=>t?t.toUpperCase():"")))),M=/\B([A-Z])/g,$=R((e=>e.replace(M,"-$1").toLowerCase())),N=R((e=>e.charAt(0).toUpperCase()+e.slice(1))),I=R((e=>e?`on${N(e)}`:"")),L=(e,t)=>!Object.is(e,t),B=(e,...t)=>{for(let n=0;n<e.length;n++)e[n](...t)},F=(e,t,n,r=!1)=>{Object.defineProperty(e,t,{configurable:!0,enumerable:!1,writable:r,value:n})},Q=e=>{const t=parseFloat(e);return isNaN(t)?e:t},U=e=>{const t=w(e)?Number(e):NaN;return isNaN(t)?e:t};let z;const j=()=>z||(z="undefined"!=typeof globalThis?globalThis:"undefined"!=typeof self?self:"undefined"!=typeof window?window:void 0!==n.g?n.g:{}),W=r("Infinity,undefined,NaN,isFinite,isNaN,parseFloat,parseInt,decodeURI,decodeURIComponent,encodeURI,encodeURIComponent,Math,Number,Date,Array,Object,Boolean,String,RegExp,Map,Set,JSON,Intl,BigInt,console,Error,Symbol");function X(e){if(f(e)){const t={};for(let n=0;n<e.length;n++){const r=e[n],a=w(r)?Z(r):X(r);if(a)for(const e in a)t[e]=a[e]}return t}if(w(e)||k(e))return e}const q=/;(?![^(]*\))/g,V=/:([^]+)/,H=/\/\*[^]*?\*\//g;function Z(e){const t={};return e.replace(H,"").split(q).forEach((e=>{if(e){const n=e.split(V);n.length>1&&(t[n[0].trim()]=n[1].trim())}})),t}function G(e){let t="";if(w(e))t=e;else if(f(e))for(let n=0;n<e.length;n++){const r=G(e[n]);r&&(t+=r+" ")}else if(k(e))for(const n in e)e[n]&&(t+=n+" ");return t.trim()}function K(e){if(!e)return null;let{class:t,style:n}=e;return t&&!w(t)&&(e.class=G(t)),n&&(e.style=X(n)),e}const Y=r("itemscope,allowfullscreen,formnovalidate,ismap,nomodule,novalidate,readonly");function J(e){return!!e||""===e}function ee(e,t){if(e===t)return!0;let n=v(e),r=v(t);if(n||r)return!(!n||!r)&&e.getTime()===t.getTime();if(n=O(e),r=O(t),n||r)return e===t;if(n=f(e),r=f(t),n||r)return!(!n||!r)&&function(e,t){if(e.length!==t.length)return!1;let n=!0;for(let r=0;n&&r<e.length;r++)n=ee(e[r],t[r]);return n}(e,t);if(n=k(e),r=k(t),n||r){if(!n||!r)return!1;if(Object.keys(e).length!==Object.keys(t).length)return!1;for(const n in e){const r=e.hasOwnProperty(n),a=t.hasOwnProperty(n);if(r&&!a||!r&&a||!ee(e[n],t[n]))return!1}}return String(e)===String(t)}function te(e,t){return e.findIndex((e=>ee(e,t)))}const ne=e=>!(!e||!0!==e.__v_isRef),re=e=>w(e)?e:null==e?"":f(e)||k(e)&&(e.toString===E||!y(e.toString))?ne(e)?re(e.value):JSON.stringify(e,ae,2):String(e),ae=(e,t)=>ne(t)?ae(e,t.value):m(t)?{[`Map(${t.size})`]:[...t.entries()].reduce(((e,[t,n],r)=>(e[oe(t,r)+" =>"]=n,e)),{})}:g(t)?{[`Set(${t.size})`]:[...t.values()].map((e=>oe(e)))}:O(t)?oe(t):!k(t)||f(t)||C(t)?t:String(t),oe=(e,t="")=>{var n;return O(e)?`Symbol(${null!=(n=e.description)?n:t})`:e}},68:(e,t,n)=>{"use strict";n.d(t,{A:()=>s});var r=n(641),a=n(953),o=n(6918),i=n(1266);const s=(0,r.pM)({__name:"EditSidebarListElement",props:{name:{}},emits:["close","edit"],setup(e,{emit:t}){const n=e,s=t,l=(0,a.KR)(n.name);return(e,t)=>((0,r.uX)(),(0,r.Wv)(o.A,{onCancel:t[1]||(t[1]=e=>s("close")),onSubmit:t[2]||(t[2]=e=>s("edit",l.value))},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(i.KB),{modelValue:l.value,"onUpdate:modelValue":t[0]||(t[0]=e=>l.value=e),autofocus:""},null,8,["modelValue"])])),_:1}))}})},165:(e,t,n)=>{"use strict";n.d(t,{GV:()=>a,o6:()=>o});var r=n(4048);function a(e,t){const n=e.match(r.QL.TEMPLATE_VARIABLE)??[];let a=e;return n.forEach((e=>{const n=function(e,t){const n=e.split(".").reduce(((e,t)=>null==e?void 0:e[t]),t);return"string"==typeof n?n:JSON.stringify(n)}(e.startsWith(":")?e.slice(1):e.replace(/[{}]/g,"").trim(),t);null!=n&&(a=a.replaceAll(e,n))})),a}function o(e){return function e(t,n){const r=Object.keys(t),a=[];return r.forEach((r=>{const o=n?`${n}.${r}`:r;"object"==typeof t[r]?a.push(...e(t[r],o)):a.push([o,String(t[r])])})),a}(e)}},180:(e,t,n)=>{"use strict";n.d(t,{A:()=>c});var r=n(641),a=n(953),o=n(33),i=n(1266),s=n(4048);const l=(0,r.pM)({__name:"HttpMethod",props:{isSquare:{type:Boolean,default:!1},method:{},isEditable:{type:Boolean,default:!1}},emits:["change"],setup(e,{emit:t}){const n=e,l=t,c=(0,r.EW)((()=>(0,s.gw)(n.method))),u=Object.entries(s.NT).map((([e])=>({id:e,label:e.toUpperCase(),color:(0,s.gw)(e).color}))),d=(0,r.EW)({get:()=>u.find((({id:e})=>e===n.method)),set:e=>(null==e?void 0:e.id)&&l("change",e.id)}),p=(0,i.Fj)({base:"text-center font-code text-3xs justify-center items-center flex",variants:{isSquare:{true:"px-2.5 whitespace-nowrap font-bold border-r h-fit m-auto",false:"rounded-full"},isEditable:{true:"http-bg-gradient rounded-md border-1/2 border-r-1/2",false:"cursor-auto"}}}),h=(0,r.EW)((()=>c.value.short));return(e,t)=>e.isEditable?((0,r.uX)(),(0,r.Wv)((0,a.R1)(i.cy),{key:0,modelValue:d.value,"onUpdate:modelValue":t[0]||(t[0]=e=>d.value=e),class:"mt-1 font-code text-sm scalar-client",options:(0,a.R1)(u)},{default:(0,r.k6)((()=>[(0,r.Lk)("div",{class:(0,o.C4)(["h-full",{"pointer-events-none":!e.isEditable}])},[(0,r.Lk)("button",{class:(0,o.C4)(["relative h-full",(0,a.R1)(i.cx)((0,a.R1)(p)({isSquare:e.isSquare,isEditable:e.isEditable}),c.value.color)]),type:"button"},[(0,r.Lk)("span",null,(0,o.v_)(h.value),1)],2)],2)])),_:1},8,["modelValue","options"])):((0,r.uX)(),(0,r.CE)("div",{key:1,class:(0,o.C4)(["relative gap-1 whitespace-nowrap",(0,a.R1)(i.cx)((0,a.R1)(p)({isSquare:e.isSquare,isEditable:e.isEditable}),c.value.color)]),type:"button"},(0,o.v_)(c.value.short),3))}}),c=(0,n(3526).A)(l,[["__scopeId","data-v-46723816"]])},264:(e,t,n)=>{"use strict";n.d(t,{x:()=>i});const r=e=>{try{return decodeURIComponent(e)}catch{return e}};var a=n(641),o=n(4851);function i(e){const t=(0,a.EW)((()=>{var t;const n=(null==(t=e.headers.find((e=>"content-type"===e.name.toLowerCase())))?void 0:t.value)??"";return new o(n)})),n=(0,a.EW)((()=>{var t;const n=(null==(t=e.headers.find((e=>"content-disposition"===e.name.toLowerCase())))?void 0:t.value)??"";return function(e){let t="";if(e){const n=e.match(/filename\s*=\s*"?([^";]+)"?/);"string"==typeof(null==n?void 0:n[1])&&(t=r(n[1].trim()))}return t}(n)})),i=(0,a.EW)((()=>e.data instanceof Blob?URL.createObjectURL(e.data):"string"==typeof e.data?URL.createObjectURL(new Blob([e.data],{type:t.value.toString()})):e.data instanceof Object&&Object.keys(e.data).length?URL.createObjectURL(new Blob([JSON.stringify(e.data)],{type:t.value.toString()})):""));return{mimeType:t,attachmentFilename:n,dataUrl:i}}},265:(e,t,n)=>{"use strict";n.d(t,{O:()=>r});const r=e=>null!=e},277:(e,t,n)=>{"use strict";n.d(t,{a:()=>o});var r=n(1259),a=n(5018);const o=(0,r.v)({properties:{ariaActiveDescendant:null,ariaAtomic:a.booleanish,ariaAutoComplete:null,ariaBusy:a.booleanish,ariaChecked:a.booleanish,ariaColCount:a.number,ariaColIndex:a.number,ariaColSpan:a.number,ariaControls:a.spaceSeparated,ariaCurrent:null,ariaDescribedBy:a.spaceSeparated,ariaDetails:null,ariaDisabled:a.booleanish,ariaDropEffect:a.spaceSeparated,ariaErrorMessage:null,ariaExpanded:a.booleanish,ariaFlowTo:a.spaceSeparated,ariaGrabbed:a.booleanish,ariaHasPopup:null,ariaHidden:a.booleanish,ariaInvalid:null,ariaKeyShortcuts:null,ariaLabel:null,ariaLabelledBy:a.spaceSeparated,ariaLevel:a.number,ariaLive:null,ariaModal:a.booleanish,ariaMultiLine:a.booleanish,ariaMultiSelectable:a.booleanish,ariaOrientation:null,ariaOwns:a.spaceSeparated,ariaPlaceholder:null,ariaPosInSet:a.number,ariaPressed:a.booleanish,ariaReadOnly:a.booleanish,ariaRelevant:null,ariaRequired:a.booleanish,ariaRoleDescription:a.spaceSeparated,ariaRowCount:a.number,ariaRowIndex:a.number,ariaRowSpan:a.number,ariaSelected:a.booleanish,ariaSetSize:a.number,ariaSort:null,ariaValueMax:a.number,ariaValueMin:a.number,ariaValueNow:a.number,ariaValueText:null,role:null},transform:(e,t)=>"role"===t?t:"aria-"+t.slice(4).toLowerCase()})},287:(e,t,n)=>{"use strict";function r(e={},t=null,n=[]){for(let[r,i]of Object.entries(e))o(n,a(t,r),i);return n}function a(e,t){return e?e+"["+t+"]":t}function o(e,t,n){if(Array.isArray(n))for(let[r,i]of n.entries())o(e,a(t,r.toString()),i);else n instanceof Date?e.push([t,n.toISOString()]):"boolean"==typeof n?e.push([t,n?"1":"0"]):"string"==typeof n?e.push([t,n]):"number"==typeof n?e.push([t,`${n}`]):null==n?e.push([t,""]):r(n,t,e)}function i(e){var t,n;let r=null!=(t=null==e?void 0:e.form)?t:e.closest("form");if(r){for(let t of r.elements)if(t!==e&&("INPUT"===t.tagName&&"submit"===t.type||"BUTTON"===t.tagName&&"submit"===t.type||"INPUT"===t.nodeName&&"image"===t.type))return void t.click();null==(n=r.requestSubmit)||n.call(r)}}n.d(t,{h:()=>r,q:()=>i})},487:(e,t,n)=>{"use strict";n.d(t,{E9:()=>T,WF:()=>C,go:()=>A,h7:()=>P,wh:()=>R});var r=n(641),a=n(953),o=n(6923),i=n(8869),s=n(3648),l=n(1972),c=n(6454),u=n(8428),d=n(805),p=n(7014),h=n(7838),f=n(1089),m=n(6413),g=n(1417),v=n(287),b=n(9388),y=n(3605);function w(e,t){return e===t}var O,k=((O=k||{})[O.Open=0]="Open",O[O.Closed=1]="Closed",O),x=(e=>(e[e.Single=0]="Single",e[e.Multi=1]="Multi",e))(x||{}),E=(e=>(e[e.Pointer=0]="Pointer",e[e.Other=1]="Other",e))(E||{});let S=Symbol("ListboxContext");function _(e){let t=(0,r.WQ)(S,null);if(null===t){let t=new Error(`<${e} /> is missing a parent <Listbox /> component.`);throw Error.captureStackTrace&&Error.captureStackTrace(t,_),t}return t}let C=(0,r.pM)({name:"Listbox",emits:{"update:modelValue":e=>!0},props:{as:{type:[Object,String],default:"template"},disabled:{type:[Boolean],default:!1},by:{type:[String,Function],default:()=>w},horizontal:{type:[Boolean],default:!1},modelValue:{type:[Object,String,Number,Boolean],default:void 0},defaultValue:{type:[Object,String,Number,Boolean],default:void 0},form:{type:String,optional:!0},name:{type:String,optional:!0},multiple:{type:[Boolean],default:!1}},inheritAttrs:!1,setup(e,{slots:t,attrs:n,emit:i}){let l=(0,a.KR)(1),c=(0,a.KR)(null),u=(0,a.KR)(null),h=(0,a.KR)(null),w=(0,a.KR)([]),O=(0,a.KR)(""),k=(0,a.KR)(null),x=(0,a.KR)(1);function E(e=e=>e){let t=null!==k.value?w.value[k.value]:null,n=(0,g.wl)(e(w.value.slice()),(e=>(0,m.t)(e.dataRef.domRef))),r=t?n.indexOf(t):null;return-1===r&&(r=null),{options:n,activeOptionIndex:r}}let _=(0,r.EW)((()=>e.multiple?1:0)),[C,T]=(0,o.P)((0,r.EW)((()=>e.modelValue)),(e=>i("update:modelValue",e)),(0,r.EW)((()=>e.defaultValue))),A=(0,r.EW)((()=>void 0===C.value?(0,b.Y)(_.value,{1:[],0:void 0}):C.value)),R={listboxState:l,value:A,mode:_,compare(t,n){if("string"==typeof e.by){let r=e.by;return(null==t?void 0:t[r])===(null==n?void 0:n[r])}return e.by(t,n)},orientation:(0,r.EW)((()=>e.horizontal?"horizontal":"vertical")),labelRef:c,buttonRef:u,optionsRef:h,disabled:(0,r.EW)((()=>e.disabled)),options:w,searchQuery:O,activeOptionIndex:k,activationTrigger:x,closeListbox(){e.disabled||1!==l.value&&(l.value=1,k.value=null)},openListbox(){e.disabled||0!==l.value&&(l.value=0)},goToOption(t,n,r){if(e.disabled||1===l.value)return;let a=E(),o=(0,f.X)(t===f.B.Specific?{focus:f.B.Specific,id:n}:{focus:t},{resolveItems:()=>a.options,resolveActiveIndex:()=>a.activeOptionIndex,resolveId:e=>e.id,resolveDisabled:e=>e.dataRef.disabled});O.value="",k.value=o,x.value=null!=r?r:1,w.value=a.options},search(t){if(e.disabled||1===l.value)return;let n=""!==O.value?0:1;O.value+=t.toLowerCase();let r=(null!==k.value?w.value.slice(k.value+n).concat(w.value.slice(0,k.value+n)):w.value).find((e=>e.dataRef.textValue.startsWith(O.value)&&!e.dataRef.disabled)),a=r?w.value.indexOf(r):-1;-1===a||a===k.value||(k.value=a,x.value=1)},clearSearch(){e.disabled||1!==l.value&&""!==O.value&&(O.value="")},registerOption(e,t){let n=E((n=>[...n,{id:e,dataRef:t}]));w.value=n.options,k.value=n.activeOptionIndex},unregisterOption(e){let t=E((t=>{let n=t.findIndex((t=>t.id===e));return-1!==n&&t.splice(n,1),t}));w.value=t.options,k.value=t.activeOptionIndex,x.value=1},theirOnChange(t){e.disabled||T(t)},select(t){e.disabled||T((0,b.Y)(_.value,{0:()=>t,1:()=>{let e=(0,a.ux)(R.value.value).slice(),n=(0,a.ux)(t),r=e.findIndex((e=>R.compare(n,(0,a.ux)(e))));return-1===r?e.push(n):e.splice(r,1),e}}))}};(0,s.j)([u,h],((e,t)=>{var n;R.closeListbox(),(0,g.Bm)(t,g.MZ.Loose)||(e.preventDefault(),null==(n=(0,m.t)(u))||n.focus())}),(0,r.EW)((()=>0===l.value))),(0,r.Gt)(S,R),(0,p.Fo)((0,r.EW)((()=>(0,b.Y)(l.value,{0:p.Uw.Open,1:p.Uw.Closed}))));let P=(0,r.EW)((()=>{var e;return null==(e=(0,m.t)(u))?void 0:e.closest("form")}));return(0,r.sV)((()=>{(0,r.wB)([P],(()=>{if(P.value&&void 0!==e.defaultValue)return P.value.addEventListener("reset",t),()=>{var e;null==(e=P.value)||e.removeEventListener("reset",t)};function t(){R.theirOnChange(e.defaultValue)}}),{immediate:!0})})),()=>{let{name:a,modelValue:o,disabled:i,form:s,...c}=e,u={open:0===l.value,disabled:i,value:A.value};return(0,r.h)(r.FK,[...null!=a&&null!=A.value?(0,v.h)({[a]:A.value}).map((([e,t])=>(0,r.h)(d.j,(0,y.oE)({features:d.O.Hidden,key:e,as:"input",type:"hidden",hidden:!0,readOnly:!0,form:s,disabled:i,name:e,value:t})))):[],(0,y.XX)({ourProps:{},theirProps:{...n,...(0,y.cJ)(c,["defaultValue","onUpdate:modelValue","horizontal","multiple","by"])},slot:u,slots:t,attrs:n,name:"Listbox"})])}}}),T=(0,r.pM)({name:"ListboxLabel",props:{as:{type:[Object,String],default:"label"},id:{type:String,default:null}},setup(e,{attrs:t,slots:n}){var r;let a=null!=(r=e.id)?r:`headlessui-listbox-label-${(0,i.B)()}`,o=_("ListboxLabel");function s(){var e;null==(e=(0,m.t)(o.buttonRef))||e.focus({preventScroll:!0})}return()=>{let r={open:0===o.listboxState.value,disabled:o.disabled.value},{...i}=e,l={id:a,ref:o.labelRef,onClick:s};return(0,y.XX)({ourProps:l,theirProps:i,slot:r,attrs:t,slots:n,name:"ListboxLabel"})}}}),A=(0,r.pM)({name:"ListboxButton",props:{as:{type:[Object,String],default:"button"},id:{type:String,default:null}},setup(e,{attrs:t,slots:n,expose:a}){var o;let s=null!=(o=e.id)?o:`headlessui-listbox-button-${(0,i.B)()}`,c=_("ListboxButton");function u(e){switch(e.key){case h.D.Space:case h.D.Enter:case h.D.ArrowDown:e.preventDefault(),c.openListbox(),(0,r.dY)((()=>{var e;null==(e=(0,m.t)(c.optionsRef))||e.focus({preventScroll:!0}),c.value.value||c.goToOption(f.B.First)}));break;case h.D.ArrowUp:e.preventDefault(),c.openListbox(),(0,r.dY)((()=>{var e;null==(e=(0,m.t)(c.optionsRef))||e.focus({preventScroll:!0}),c.value.value||c.goToOption(f.B.Last)}))}}function d(e){e.key===h.D.Space&&e.preventDefault()}function p(e){c.disabled.value||(0===c.listboxState.value?(c.closeListbox(),(0,r.dY)((()=>{var e;return null==(e=(0,m.t)(c.buttonRef))?void 0:e.focus({preventScroll:!0})}))):(e.preventDefault(),c.openListbox(),function(e){requestAnimationFrame((()=>requestAnimationFrame(e)))}((()=>{var e;return null==(e=(0,m.t)(c.optionsRef))?void 0:e.focus({preventScroll:!0})}))))}a({el:c.buttonRef,$el:c.buttonRef});let g=(0,l.c)((0,r.EW)((()=>({as:e.as,type:t.type}))),c.buttonRef);return()=>{var r,a;let o={open:0===c.listboxState.value,disabled:c.disabled.value,value:c.value.value},{...i}=e,l={ref:c.buttonRef,id:s,type:g.value,"aria-haspopup":"listbox","aria-controls":null==(r=(0,m.t)(c.optionsRef))?void 0:r.id,"aria-expanded":0===c.listboxState.value,"aria-labelledby":c.labelRef.value?[null==(a=(0,m.t)(c.labelRef))?void 0:a.id,s].join(" "):void 0,disabled:!0===c.disabled.value||void 0,onKeydown:u,onKeyup:d,onClick:p};return(0,y.XX)({ourProps:l,theirProps:i,slot:o,attrs:t,slots:n,name:"ListboxButton"})}}}),R=(0,r.pM)({name:"ListboxOptions",props:{as:{type:[Object,String],default:"ul"},static:{type:Boolean,default:!1},unmount:{type:Boolean,default:!0},id:{type:String,default:null}},setup(e,{attrs:t,slots:n,expose:o}){var s;let l=null!=(s=e.id)?s:`headlessui-listbox-options-${(0,i.B)()}`,c=_("ListboxOptions"),u=(0,a.KR)(null);function d(e){switch(u.value&&clearTimeout(u.value),e.key){case h.D.Space:if(""!==c.searchQuery.value)return e.preventDefault(),e.stopPropagation(),c.search(e.key);case h.D.Enter:if(e.preventDefault(),e.stopPropagation(),null!==c.activeOptionIndex.value){let e=c.options.value[c.activeOptionIndex.value];c.select(e.dataRef.value)}0===c.mode.value&&(c.closeListbox(),(0,r.dY)((()=>{var e;return null==(e=(0,m.t)(c.buttonRef))?void 0:e.focus({preventScroll:!0})})));break;case(0,b.Y)(c.orientation.value,{vertical:h.D.ArrowDown,horizontal:h.D.ArrowRight}):return e.preventDefault(),e.stopPropagation(),c.goToOption(f.B.Next);case(0,b.Y)(c.orientation.value,{vertical:h.D.ArrowUp,horizontal:h.D.ArrowLeft}):return e.preventDefault(),e.stopPropagation(),c.goToOption(f.B.Previous);case h.D.Home:case h.D.PageUp:return e.preventDefault(),e.stopPropagation(),c.goToOption(f.B.First);case h.D.End:case h.D.PageDown:return e.preventDefault(),e.stopPropagation(),c.goToOption(f.B.Last);case h.D.Escape:e.preventDefault(),e.stopPropagation(),c.closeListbox(),(0,r.dY)((()=>{var e;return null==(e=(0,m.t)(c.buttonRef))?void 0:e.focus({preventScroll:!0})}));break;case h.D.Tab:e.preventDefault(),e.stopPropagation();break;default:1===e.key.length&&(c.search(e.key),u.value=setTimeout((()=>c.clearSearch()),350))}}o({el:c.optionsRef,$el:c.optionsRef});let g=(0,p.O_)(),v=(0,r.EW)((()=>null!==g?(g.value&p.Uw.Open)===p.Uw.Open:0===c.listboxState.value));return()=>{var r,a;let o={open:0===c.listboxState.value},{...i}=e,s={"aria-activedescendant":null===c.activeOptionIndex.value||null==(r=c.options.value[c.activeOptionIndex.value])?void 0:r.id,"aria-multiselectable":1===c.mode.value||void 0,"aria-labelledby":null==(a=(0,m.t)(c.buttonRef))?void 0:a.id,"aria-orientation":c.orientation.value,id:l,onKeydown:d,role:"listbox",tabIndex:0,ref:c.optionsRef};return(0,y.XX)({ourProps:s,theirProps:i,slot:o,attrs:t,slots:n,features:y.O5.RenderStrategy|y.O5.Static,visible:v.value,name:"ListboxOptions"})}}}),P=(0,r.pM)({name:"ListboxOption",props:{as:{type:[Object,String],default:"li"},value:{type:[Object,String,Number,Boolean]},disabled:{type:Boolean,default:!1},id:{type:String,default:null}},setup(e,{slots:t,attrs:n,expose:o}){var s;let l=null!=(s=e.id)?s:`headlessui-listbox-option-${(0,i.B)()}`,d=_("ListboxOption"),p=(0,a.KR)(null);o({el:p,$el:p});let h=(0,r.EW)((()=>null!==d.activeOptionIndex.value&&d.options.value[d.activeOptionIndex.value].id===l)),g=(0,r.EW)((()=>(0,b.Y)(d.mode.value,{0:()=>d.compare((0,a.ux)(d.value.value),(0,a.ux)(e.value)),1:()=>(0,a.ux)(d.value.value).some((t=>d.compare((0,a.ux)(t),(0,a.ux)(e.value))))}))),v=(0,r.EW)((()=>(0,b.Y)(d.mode.value,{1:()=>{var e;let t=(0,a.ux)(d.value.value);return(null==(e=d.options.value.find((e=>t.some((t=>d.compare((0,a.ux)(t),(0,a.ux)(e.dataRef.value)))))))?void 0:e.id)===l},0:()=>g.value}))),w=(0,c.q)(p),O=(0,r.EW)((()=>({disabled:e.disabled,value:e.value,get textValue(){return w()},domRef:p})));function k(t){if(e.disabled)return t.preventDefault();d.select(e.value),0===d.mode.value&&(d.closeListbox(),(0,r.dY)((()=>{var e;return null==(e=(0,m.t)(d.buttonRef))?void 0:e.focus({preventScroll:!0})})))}function x(){if(e.disabled)return d.goToOption(f.B.Nothing);d.goToOption(f.B.Specific,l)}(0,r.sV)((()=>d.registerOption(l,O))),(0,r.hi)((()=>d.unregisterOption(l))),(0,r.sV)((()=>{(0,r.wB)([d.listboxState,g],(()=>{0===d.listboxState.value&&g.value&&(0,b.Y)(d.mode.value,{1:()=>{v.value&&d.goToOption(f.B.Specific,l)},0:()=>{d.goToOption(f.B.Specific,l)}})}),{immediate:!0})})),(0,r.nT)((()=>{0===d.listboxState.value&&h.value&&0!==d.activationTrigger.value&&(0,r.dY)((()=>{var e,t;return null==(t=null==(e=(0,m.t)(p))?void 0:e.scrollIntoView)?void 0:t.call(e,{block:"nearest"})}))}));let E=(0,u.J)();function S(e){E.update(e)}function C(t){E.wasMoved(t)&&(e.disabled||h.value||d.goToOption(f.B.Specific,l,0))}function T(t){E.wasMoved(t)&&(e.disabled||h.value&&d.goToOption(f.B.Nothing))}return()=>{let{disabled:r}=e,a={active:h.value,selected:g.value,disabled:r},{value:o,disabled:i,...s}=e,c={id:l,ref:p,role:"option",tabIndex:!0===r?void 0:-1,"aria-disabled":!0===r||void 0,"aria-selected":g.value,disabled:void 0,onClick:k,onFocus:x,onPointerenter:S,onMouseenter:S,onPointermove:C,onMousemove:C,onPointerleave:T,onMouseleave:T};return(0,y.XX)({ourProps:c,theirProps:s,slot:a,attrs:n,slots:t,name:"ListboxOption"})}}})},498:(e,t,n)=>{"use strict";function r(e,t){const n=(e,t,r)=>{let a="";if(e instanceof Array)for(let o=0,i=e.length;o<i;o++)a+=r+n(e[o],t,r+"\t")+"\n";else if("object"==typeof e){let o=!1;a+=r+"<"+t;for(const t in e)"@"==t.charAt(0)?a+=" "+t.substr(1)+'="'+e[t].toString()+'"':o=!0;if(a+=o?">":"/>",o){for(const t in e)"#text"==t?a+=e[t]:"#cdata"==t?a+="<![CDATA["+e[t]+"]]>":"@"!=t.charAt(0)&&(a+=n(e[t],t,r+"\t"));a+=("\n"==a.charAt(a.length-1)?r:"")+"</"+t+">"}}else a+=r+"<"+t+">"+e.toString()+"</"+t+">";return a};let r="";for(const t in e)r+=n(e[t],t,"");return t?r.replace(/\t/g,t):r.replace(/\t|\n/g,"")}n.d(t,{z:()=>u});var a=n(5868),o=n(6764),i=n(6941),s=n(3213);function l(e,t=""){return Object.entries(e).flatMap((([e,n])=>{const r=t?`${t}[${e}]`:e;return"object"==typeof n&&null!==n?l(n,r):[{name:r,value:n}]}))}const c=["application/json","application/octet-stream","application/x-www-form-urlencoded","application/xml","multipart/form-data","text/plain"];function u(e,t,n){const u=e.information?.requestBody?.content,d=(0,a.L)(u),p=c.find((e=>!!d?.[e]))??"application/json",h=d?.[p]?.examples??d?.["application/json"]?.examples,f=(h??{})?.[t??Object.keys(h??{})[0]];if(f)return{mimeType:p,text:(0,o.q)(f?.value)};const m=(0,s.I)(e,"body",!1);if(m.length>0)return{mimeType:"application/json",text:(0,o.q)(m[0].value)};const g=(0,s.I)(e,"formData",!1);if(g.length>0)return{mimeType:"application/x-www-form-urlencoded",params:g.map((e=>({name:e.name,value:"string"==typeof e.value?e.value:JSON.stringify(e.value)})))};if(!p)return null;const v=d?.[p],b=v?.example?v?.example:void 0;if("application/json"===p){const e=v?.schema?(0,i.M)(v?.schema,{mode:"write",omitEmptyAndOptionalProperties:n}):null,t=b??e;return{mimeType:p,text:"string"==typeof t?t:JSON.stringify(t,null,2)}}if("application/xml"===p){const e=v?.schema?(0,i.M)(v?.schema,{xml:!0,mode:"write"}):null;return{mimeType:p,text:b??r(e,"  ")}}if("application/octet-stream"===p)return{mimeType:p,text:"BINARY"};if("text/plain"===p){const e=v?.schema?(0,i.M)(v?.schema,{xml:!0,mode:"write"}):null;return{mimeType:p,text:b??e??""}}if("multipart/form-data"===p||"application/x-www-form-urlencoded"===p){const e=v?.schema?(0,i.M)(v?.schema,{xml:!0,mode:"write"}):null;return{mimeType:p,params:l(b??e??{})}}return null}},562:(e,t,n)=>{"use strict";n.d(t,{T:()=>o});var r=n(6413),a=n(9146);function o(e){if(a._.isServer)return null;if(e instanceof Node)return e.ownerDocument;if(null!=e&&e.hasOwnProperty("value")){let t=(0,r.t)(e);if(t)return t.ownerDocument}return document}},569:(e,t,n)=>{"use strict";n.d(t,{q:()=>i});var r=n(6530),a=n(1259),o=n(5018);const i=(0,a.v)({attributes:{acceptcharset:"accept-charset",classname:"class",htmlfor:"for",httpequiv:"http-equiv"},mustUseProperty:["checked","multiple","muted","selected"],properties:{abbr:null,accept:o.commaSeparated,acceptCharset:o.spaceSeparated,accessKey:o.spaceSeparated,action:null,allow:null,allowFullScreen:o.boolean,allowPaymentRequest:o.boolean,allowUserMedia:o.boolean,alt:null,as:null,async:o.boolean,autoCapitalize:null,autoComplete:o.spaceSeparated,autoFocus:o.boolean,autoPlay:o.boolean,blocking:o.spaceSeparated,capture:null,charSet:null,checked:o.boolean,cite:null,className:o.spaceSeparated,cols:o.number,colSpan:null,content:null,contentEditable:o.booleanish,controls:o.boolean,controlsList:o.spaceSeparated,coords:o.number|o.commaSeparated,crossOrigin:null,data:null,dateTime:null,decoding:null,default:o.boolean,defer:o.boolean,dir:null,dirName:null,disabled:o.boolean,download:o.overloadedBoolean,draggable:o.booleanish,encType:null,enterKeyHint:null,fetchPriority:null,form:null,formAction:null,formEncType:null,formMethod:null,formNoValidate:o.boolean,formTarget:null,headers:o.spaceSeparated,height:o.number,hidden:o.boolean,high:o.number,href:null,hrefLang:null,htmlFor:o.spaceSeparated,httpEquiv:o.spaceSeparated,id:null,imageSizes:null,imageSrcSet:null,inert:o.boolean,inputMode:null,integrity:null,is:null,isMap:o.boolean,itemId:null,itemProp:o.spaceSeparated,itemRef:o.spaceSeparated,itemScope:o.boolean,itemType:o.spaceSeparated,kind:null,label:null,lang:null,language:null,list:null,loading:null,loop:o.boolean,low:o.number,manifest:null,max:null,maxLength:o.number,media:null,method:null,min:null,minLength:o.number,multiple:o.boolean,muted:o.boolean,name:null,nonce:null,noModule:o.boolean,noValidate:o.boolean,onAbort:null,onAfterPrint:null,onAuxClick:null,onBeforeMatch:null,onBeforePrint:null,onBeforeToggle:null,onBeforeUnload:null,onBlur:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onContextLost:null,onContextMenu:null,onContextRestored:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnded:null,onError:null,onFocus:null,onFormData:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLanguageChange:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadEnd:null,onLoadStart:null,onMessage:null,onMessageError:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRejectionHandled:null,onReset:null,onResize:null,onScroll:null,onScrollEnd:null,onSecurityPolicyViolation:null,onSeeked:null,onSeeking:null,onSelect:null,onSlotChange:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnhandledRejection:null,onUnload:null,onVolumeChange:null,onWaiting:null,onWheel:null,open:o.boolean,optimum:o.number,pattern:null,ping:o.spaceSeparated,placeholder:null,playsInline:o.boolean,popover:null,popoverTarget:null,popoverTargetAction:null,poster:null,preload:null,readOnly:o.boolean,referrerPolicy:null,rel:o.spaceSeparated,required:o.boolean,reversed:o.boolean,rows:o.number,rowSpan:o.number,sandbox:o.spaceSeparated,scope:null,scoped:o.boolean,seamless:o.boolean,selected:o.boolean,shadowRootClonable:o.boolean,shadowRootDelegatesFocus:o.boolean,shadowRootMode:null,shape:null,size:o.number,sizes:null,slot:null,span:o.number,spellCheck:o.booleanish,src:null,srcDoc:null,srcLang:null,srcSet:null,start:o.number,step:null,style:null,tabIndex:o.number,target:null,title:null,translate:null,type:null,typeMustMatch:o.boolean,useMap:null,value:o.booleanish,width:o.number,wrap:null,writingSuggestions:null,align:null,aLink:null,archive:o.spaceSeparated,axis:null,background:null,bgColor:null,border:o.number,borderColor:null,bottomMargin:o.number,cellPadding:null,cellSpacing:null,char:null,charOff:null,classId:null,clear:null,code:null,codeBase:null,codeType:null,color:null,compact:o.boolean,declare:o.boolean,event:null,face:null,frame:null,frameBorder:null,hSpace:o.number,leftMargin:o.number,link:null,longDesc:null,lowSrc:null,marginHeight:o.number,marginWidth:o.number,noResize:o.boolean,noHref:o.boolean,noShade:o.boolean,noWrap:o.boolean,object:null,profile:null,prompt:null,rev:null,rightMargin:o.number,rules:null,scheme:null,scrolling:o.booleanish,standby:null,summary:null,text:null,topMargin:o.number,valueType:null,version:null,vAlign:null,vLink:null,vSpace:o.number,allowTransparency:null,autoCorrect:null,autoSave:null,disablePictureInPicture:o.boolean,disableRemotePlayback:o.boolean,prefix:null,property:null,results:o.number,security:null,unselectable:null},space:"html",transform:r._})},641:(e,t,n)=>{"use strict";n.d(t,{$u:()=>_e,$y:()=>Ie,Bi:()=>pe,CE:()=>gn,Df:()=>ue,Dl:()=>b,E3:()=>Sn,EW:()=>Kn,FK:()=>on,Fv:()=>Cn,Gt:()=>mt,Gy:()=>J,HF:()=>Ye,Ht:()=>He,Ic:()=>Se,Im:()=>H,K9:()=>$t,KC:()=>xe,LM:()=>Ut,Lk:()=>kn,MZ:()=>ce,Mw:()=>ln,Ng:()=>En,OA:()=>Ze,OW:()=>ie,Q3:()=>Tn,QP:()=>te,RG:()=>Ue,SS:()=>Ae,Tb:()=>je,U_:()=>Wt,WQ:()=>gt,Wv:()=>vn,Y4:()=>be,bF:()=>xn,bo:()=>L,dY:()=>m,eW:()=>_n,eX:()=>Qe,fn:()=>Zt,g2:()=>$e,h:()=>Yn,hi:()=>Te,jC:()=>cn,k6:()=>I,n:()=>ve,nI:()=>Ln,nT:()=>zt,p9:()=>jt,pI:()=>Fe,pM:()=>de,pR:()=>ae,qG:()=>De,qL:()=>i,rE:()=>Jn,sV:()=>Ee,uX:()=>pn,v6:()=>Dn,wB:()=>Xt,xo:()=>Ce,zz:()=>Je});var r=n(953),a=n(33);function o(e,t,n,r){try{return r?e(...r):e()}catch(e){s(e,t,n)}}function i(e,t,n,r){if((0,a.Tn)(e)){const i=o(e,t,n,r);return i&&(0,a.yL)(i)&&i.catch((e=>{s(e,t,n)})),i}if((0,a.cy)(e)){const a=[];for(let o=0;o<e.length;o++)a.push(i(e[o],t,n,r));return a}}function s(e,t,n,i=!0){t&&t.vnode;const{errorHandler:s,throwUnhandledErrorInProduction:l}=t&&t.appContext.config||a.MZ;if(t){let a=t.parent;const i=t.proxy,l=`https://vuejs.org/error-reference/#runtime-${n}`;for(;a;){const t=a.ec;if(t)for(let n=0;n<t.length;n++)if(!1===t[n](e,i,l))return;a=a.parent}if(s)return(0,r.C4)(),o(s,null,10,[e,i,l]),void(0,r.bl)()}!function(e,t,n,r=!0,a=!1){if(a)throw e;console.error(e)}(e,0,0,i,l)}const l=[];let c=-1;const u=[];let d=null,p=0;const h=Promise.resolve();let f=null;function m(e){const t=f||h;return e?t.then(this?e.bind(this):e):t}function g(e){if(!(1&e.flags)){const t=O(e),n=l[l.length-1];!n||!(2&e.flags)&&t>=O(n)?l.push(e):l.splice(function(e){let t=c+1,n=l.length;for(;t<n;){const r=t+n>>>1,a=l[r],o=O(a);o<e||o===e&&2&a.flags?t=r+1:n=r}return t}(t),0,e),e.flags|=1,v()}}function v(){f||(f=h.then(k))}function b(e){(0,a.cy)(e)?u.push(...e):d&&-1===e.id?d.splice(p+1,0,e):1&e.flags||(u.push(e),e.flags|=1),v()}function y(e,t,n=c+1){for(;n<l.length;n++){const t=l[n];if(t&&2&t.flags){if(e&&t.id!==e.uid)continue;l.splice(n,1),n--,4&t.flags&&(t.flags&=-2),t(),4&t.flags||(t.flags&=-2)}}}function w(e){if(u.length){const e=[...new Set(u)].sort(((e,t)=>O(e)-O(t)));if(u.length=0,d)return void d.push(...e);for(d=e,p=0;p<d.length;p++){const e=d[p];4&e.flags&&(e.flags&=-2),8&e.flags||e(),e.flags&=-2}d=null,p=0}}const O=e=>null==e.id?2&e.flags?-1:1/0:e.id;function k(e){a.tE;try{for(c=0;c<l.length;c++){const e=l[c];!e||8&e.flags||(4&e.flags&&(e.flags&=-2),o(e,e.i,e.i?15:14),4&e.flags||(e.flags&=-2))}}finally{for(;c<l.length;c++){const e=l[c];e&&(e.flags&=-2)}c=-1,l.length=0,w(),f=null,(l.length||u.length)&&k(e)}}let x,E=[],S=!1;function _(e,...t){x?x.emit(e,...t):S||E.push({event:e,args:t})}function C(e,t){var n,r;x=e,x?(x.enabled=!0,E.forEach((({event:e,args:t})=>x.emit(e,...t))),E=[]):"undefined"!=typeof window&&window.HTMLElement&&!(null==(r=null==(n=window.navigator)?void 0:n.userAgent)?void 0:r.includes("jsdom"))?((t.__VUE_DEVTOOLS_HOOK_REPLAY__=t.__VUE_DEVTOOLS_HOOK_REPLAY__||[]).push((e=>{C(e,t)})),setTimeout((()=>{x||(t.__VUE_DEVTOOLS_HOOK_REPLAY__=null,S=!0,E=[])}),3e3)):(S=!0,E=[])}const T=D("component:added"),A=D("component:updated"),R=D("component:removed"),P=e=>{x&&"function"==typeof x.cleanupBuffer&&!x.cleanupBuffer(e)&&R(e)};function D(e){return t=>{_(e,t.appContext.app,t.uid,t.parent?t.parent.uid:void 0,t)}}let M=null,$=null;function N(e){const t=M;return M=e,$=e&&e.type.__scopeId||null,t}function I(e,t=M,n){if(!t)return e;if(e._n)return e;const r=(...n)=>{r._d&&fn(-1);const a=N(t);let o;try{o=e(...n)}finally{N(a),r._d&&fn(1)}return __VUE_PROD_DEVTOOLS__&&A(t),o};return r._n=!0,r._c=!0,r._d=!0,r}function L(e,t){if(null===M)return e;const n=Gn(M),o=e.dirs||(e.dirs=[]);for(let e=0;e<t.length;e++){let[i,s,l,c=a.MZ]=t[e];i&&((0,a.Tn)(i)&&(i={mounted:i,updated:i}),i.deep&&(0,r.hV)(s),o.push({dir:i,instance:n,value:s,oldValue:void 0,arg:l,modifiers:c}))}return e}function B(e,t,n,a){const o=e.dirs,s=t&&t.dirs;for(let l=0;l<o.length;l++){const c=o[l];s&&(c.oldValue=s[l].value);let u=c.dir[a];u&&((0,r.C4)(),i(u,n,8,[e.el,c,e,t]),(0,r.bl)())}}const F=Symbol("_vte"),Q=e=>e.__isTeleport,U=e=>e&&(e.disabled||""===e.disabled),z=e=>e&&(e.defer||""===e.defer),j=e=>"undefined"!=typeof SVGElement&&e instanceof SVGElement,W=e=>"function"==typeof MathMLElement&&e instanceof MathMLElement,X=(e,t)=>{const n=e&&e.to;return(0,a.Kg)(n)?t?t(n):null:n},q={name:"Teleport",__isTeleport:!0,process(e,t,n,r,a,o,i,s,l,c){const{mc:u,pc:d,pbc:p,o:{insert:h,querySelector:f,createText:m,createComment:g}}=c,v=U(t.props);let{shapeFlag:b,children:y,dynamicChildren:w}=t;if(null==e){const e=t.el=m(""),c=t.anchor=m("");h(e,n,r),h(c,n,r);const d=(e,t)=>{16&b&&(a&&a.isCE&&(a.ce._teleportTarget=e),u(y,e,t,a,o,i,s,l))},p=()=>{const e=t.target=X(t.props,f),n=G(e,t,m,h);e&&("svg"!==i&&j(e)?i="svg":"mathml"!==i&&W(e)&&(i="mathml"),v||(d(e,n),Z(t,!1)))};v&&(d(n,c),Z(t,!0)),z(t.props)?Mt((()=>{p(),t.el.__isMounted=!0}),o):p()}else{if(z(t.props)&&!e.el.__isMounted)return void Mt((()=>{q.process(e,t,n,r,a,o,i,s,l,c),delete e.el.__isMounted}),o);t.el=e.el,t.targetStart=e.targetStart;const u=t.anchor=e.anchor,h=t.target=e.target,m=t.targetAnchor=e.targetAnchor,g=U(e.props),b=g?n:h,y=g?u:m;if("svg"===i||j(h)?i="svg":("mathml"===i||W(h))&&(i="mathml"),w?(p(e.dynamicChildren,w,b,a,o,i,s),Lt(e,t,!0)):l||d(e,t,b,y,a,o,i,s,!1),v)g?t.props&&e.props&&t.props.to!==e.props.to&&(t.props.to=e.props.to):V(t,n,u,c,1);else if((t.props&&t.props.to)!==(e.props&&e.props.to)){const e=t.target=X(t.props,f);e&&V(t,e,null,c,0)}else g&&V(t,h,m,c,1);Z(t,v)}},remove(e,t,n,{um:r,o:{remove:a}},o){const{shapeFlag:i,children:s,anchor:l,targetStart:c,targetAnchor:u,target:d,props:p}=e;if(d&&(a(c),a(u)),o&&a(l),16&i){const e=o||!U(p);for(let a=0;a<s.length;a++){const o=s[a];r(o,t,n,e,!!o.dynamicChildren)}}},move:V,hydrate:function(e,t,n,r,a,o,{o:{nextSibling:i,parentNode:s,querySelector:l,insert:c,createText:u}},d){const p=t.target=X(t.props,l);if(p){const l=U(t.props),h=p._lpa||p.firstChild;if(16&t.shapeFlag)if(l)t.anchor=d(i(e),t,s(e),n,r,a,o),t.targetStart=h,t.targetAnchor=h&&i(h);else{t.anchor=i(e);let s=h;for(;s;){if(s&&8===s.nodeType)if("teleport start anchor"===s.data)t.targetStart=s;else if("teleport anchor"===s.data){t.targetAnchor=s,p._lpa=t.targetAnchor&&i(t.targetAnchor);break}s=i(s)}t.targetAnchor||G(p,t,u,c),d(h&&i(h),t,p,n,r,a,o)}Z(t,l)}return t.anchor&&i(t.anchor)}};function V(e,t,n,{o:{insert:r},m:a},o=2){0===o&&r(e.targetAnchor,t,n);const{el:i,anchor:s,shapeFlag:l,children:c,props:u}=e,d=2===o;if(d&&r(i,t,n),(!d||U(u))&&16&l)for(let e=0;e<c.length;e++)a(c[e],t,n,2);d&&r(s,t,n)}const H=q;function Z(e,t){const n=e.ctx;if(n&&n.ut){let r,a;for(t?(r=e.el,a=e.anchor):(r=e.targetStart,a=e.targetAnchor);r&&r!==a;)1===r.nodeType&&r.setAttribute("data-v-owner",n.uid),r=r.nextSibling;n.ut()}}function G(e,t,n,r){const a=t.targetStart=n(""),o=t.targetAnchor=n("");return a[F]=o,e&&(r(a,e),r(o,e)),o}const K=Symbol("_leaveCb"),Y=Symbol("_enterCb");function J(){const e={isMounted:!1,isLeaving:!1,isUnmounting:!1,leavingVNodes:new Map};return Ee((()=>{e.isMounted=!0})),Ce((()=>{e.isUnmounting=!0})),e}const ee=[Function,Array],te={mode:String,appear:Boolean,persisted:Boolean,onBeforeEnter:ee,onEnter:ee,onAfterEnter:ee,onEnterCancelled:ee,onBeforeLeave:ee,onLeave:ee,onAfterLeave:ee,onLeaveCancelled:ee,onBeforeAppear:ee,onAppear:ee,onAfterAppear:ee,onAppearCancelled:ee},ne=e=>{const t=e.subTree;return t.component?ne(t.component):t};function re(e){let t=e[0];if(e.length>1){let n=!1;for(const r of e)if(r.type!==ln){t=r,n=!0;break}}return t}const ae={name:"BaseTransition",props:te,setup(e,{slots:t}){const n=Ln(),a=J();return()=>{const o=t.default&&ue(t.default(),!0);if(!o||!o.length)return;const i=re(o),s=(0,r.ux)(e),{mode:l}=s;if(a.isLeaving)return se(i);const c=le(i);if(!c)return se(i);let u=ie(c,s,a,n,(e=>u=e));c.type!==ln&&ce(c,u);let d=n.subTree&&le(n.subTree);if(d&&d.type!==ln&&!yn(c,d)&&ne(n).type!==ln){let e=ie(d,s,a,n);if(ce(d,e),"out-in"===l&&c.type!==ln)return a.isLeaving=!0,e.afterLeave=()=>{a.isLeaving=!1,8&n.job.flags||n.update(),delete e.afterLeave,d=void 0},se(i);"in-out"===l&&c.type!==ln?e.delayLeave=(e,t,n)=>{oe(a,d)[String(d.key)]=d,e[K]=()=>{t(),e[K]=void 0,delete u.delayedLeave,d=void 0},u.delayedLeave=()=>{n(),delete u.delayedLeave,d=void 0}}:d=void 0}else d&&(d=void 0);return i}}};function oe(e,t){const{leavingVNodes:n}=e;let r=n.get(t.type);return r||(r=Object.create(null),n.set(t.type,r)),r}function ie(e,t,n,r,o){const{appear:s,mode:l,persisted:c=!1,onBeforeEnter:u,onEnter:d,onAfterEnter:p,onEnterCancelled:h,onBeforeLeave:f,onLeave:m,onAfterLeave:g,onLeaveCancelled:v,onBeforeAppear:b,onAppear:y,onAfterAppear:w,onAppearCancelled:O}=t,k=String(e.key),x=oe(n,e),E=(e,t)=>{e&&i(e,r,9,t)},S=(e,t)=>{const n=t[1];E(e,t),(0,a.cy)(e)?e.every((e=>e.length<=1))&&n():e.length<=1&&n()},_={mode:l,persisted:c,beforeEnter(t){let r=u;if(!n.isMounted){if(!s)return;r=b||u}t[K]&&t[K](!0);const a=x[k];a&&yn(e,a)&&a.el[K]&&a.el[K](),E(r,[t])},enter(e){let t=d,r=p,a=h;if(!n.isMounted){if(!s)return;t=y||d,r=w||p,a=O||h}let o=!1;const i=e[Y]=t=>{o||(o=!0,E(t?a:r,[e]),_.delayedLeave&&_.delayedLeave(),e[Y]=void 0)};t?S(t,[e,i]):i()},leave(t,r){const a=String(e.key);if(t[Y]&&t[Y](!0),n.isUnmounting)return r();E(f,[t]);let o=!1;const i=t[K]=n=>{o||(o=!0,r(),E(n?v:g,[t]),t[K]=void 0,x[a]===e&&delete x[a])};x[a]=e,m?S(m,[t,i]):i()},clone(e){const a=ie(e,t,n,r,o);return o&&o(a),a}};return _}function se(e){if(ge(e))return(e=Sn(e)).children=null,e}function le(e){if(!ge(e))return Q(e.type)&&e.children?re(e.children):e;const{shapeFlag:t,children:n}=e;if(n){if(16&t)return n[0];if(32&t&&(0,a.Tn)(n.default))return n.default()}}function ce(e,t){6&e.shapeFlag&&e.component?(e.transition=t,ce(e.component.subTree,t)):128&e.shapeFlag?(e.ssContent.transition=t.clone(e.ssContent),e.ssFallback.transition=t.clone(e.ssFallback)):e.transition=t}function ue(e,t=!1,n){let r=[],a=0;for(let o=0;o<e.length;o++){let i=e[o];const s=null==n?i.key:String(n)+String(null!=i.key?i.key:o);i.type===on?(128&i.patchFlag&&a++,r=r.concat(ue(i.children,t,s))):(t||i.type!==ln)&&r.push(null!=s?Sn(i,{key:s}):i)}if(a>1)for(let e=0;e<r.length;e++)r[e].patchFlag=-2;return r}function de(e,t){return(0,a.Tn)(e)?(()=>(0,a.X$)({name:e.name},t,{setup:e}))():e}function pe(){const e=Ln();return e?(e.appContext.config.idPrefix||"v")+"-"+e.ids[0]+e.ids[1]++:""}function he(e){e.ids=[e.ids[0]+e.ids[2]+++"-",0,0]}function fe(e,t,n,i,s=!1){if((0,a.cy)(e))return void e.forEach(((e,r)=>fe(e,t&&((0,a.cy)(t)?t[r]:t),n,i,s)));if(me(i)&&!s)return void(512&i.shapeFlag&&i.type.__asyncResolved&&i.component.subTree.component&&fe(e,t,n,i.component.subTree));const l=4&i.shapeFlag?Gn(i.component):i.el,c=s?null:l,{i:u,r:d}=e,p=t&&t.r,h=u.refs===a.MZ?u.refs={}:u.refs,f=u.setupState,m=(0,r.ux)(f),g=f===a.MZ?()=>!1:e=>(0,a.$3)(m,e);if(null!=p&&p!==d&&((0,a.Kg)(p)?(h[p]=null,g(p)&&(f[p]=null)):(0,r.i9)(p)&&(p.value=null)),(0,a.Tn)(d))o(d,u,12,[c,h]);else{const t=(0,a.Kg)(d),o=(0,r.i9)(d);if(t||o){const r=()=>{if(e.f){const n=t?g(d)?f[d]:h[d]:d.value;s?(0,a.cy)(n)&&(0,a.TF)(n,l):(0,a.cy)(n)?n.includes(l)||n.push(l):t?(h[d]=[l],g(d)&&(f[d]=h[d])):(d.value=[l],e.k&&(h[e.k]=d.value))}else t?(h[d]=c,g(d)&&(f[d]=c)):o&&(d.value=c,e.k&&(h[e.k]=c))};c?(r.id=-1,Mt(r,n)):r()}}}(0,a.We)().requestIdleCallback,(0,a.We)().cancelIdleCallback;const me=e=>!!e.type.__asyncLoader,ge=e=>e.type.__isKeepAlive;function ve(e,t){ye(e,"a",t)}function be(e,t){ye(e,"da",t)}function ye(e,t,n=In){const r=e.__wdc||(e.__wdc=()=>{let t=n;for(;t;){if(t.isDeactivated)return;t=t.parent}return e()});if(Oe(t,r,n),n){let e=n.parent;for(;e&&e.parent;)ge(e.parent.vnode)&&we(r,t,n,e),e=e.parent}}function we(e,t,n,r){const o=Oe(t,e,r,!0);Te((()=>{(0,a.TF)(r[t],o)}),n)}function Oe(e,t,n=In,a=!1){if(n){const o=n[e]||(n[e]=[]),s=t.__weh||(t.__weh=(...a)=>{(0,r.C4)();const o=Qn(n),s=i(t,n,e,a);return o(),(0,r.bl)(),s});return a?o.unshift(s):o.push(s),s}}RegExp,RegExp;const ke=e=>(t,n=In)=>{Xn&&"sp"!==e||Oe(e,((...e)=>t(...e)),n)},xe=ke("bm"),Ee=ke("m"),Se=ke("bu"),_e=ke("u"),Ce=ke("bum"),Te=ke("um"),Ae=ke("sp"),Re=ke("rtg"),Pe=ke("rtc");function De(e,t=In){Oe("ec",e,t)}const Me="components";function $e(e,t){return Le(Me,e,!0,t)||e}const Ne=Symbol.for("v-ndc");function Ie(e){return(0,a.Kg)(e)?Le(Me,e,!1)||e:e||Ne}function Le(e,t,n=!0,r=!1){const o=M||In;if(o){const n=o.type;if(e===Me){const e=function(e,t=!0){return(0,a.Tn)(e)?e.displayName||e.name:e.name||t&&e.__name}(n,!1);if(e&&(e===t||e===(0,a.PT)(t)||e===(0,a.ZH)((0,a.PT)(t))))return n}const i=Be(o[e]||n[e],t)||Be(o.appContext[e],t);return!i&&r?n:i}}function Be(e,t){return e&&(e[t]||e[(0,a.PT)(t)]||e[(0,a.ZH)((0,a.PT)(t))])}function Fe(e,t,n,o){let i;const s=n&&n[o],l=(0,a.cy)(e);if(l||(0,a.Kg)(e)){let n=!1;l&&(0,r.g8)(e)&&(n=!(0,r.fE)(e),e=(0,r.qA)(e)),i=new Array(e.length);for(let a=0,o=e.length;a<o;a++)i[a]=t(n?(0,r.lJ)(e[a]):e[a],a,void 0,s&&s[a])}else if("number"==typeof e){i=new Array(e);for(let n=0;n<e;n++)i[n]=t(n+1,n,void 0,s&&s[n])}else if((0,a.Gv)(e))if(e[Symbol.iterator])i=Array.from(e,((e,n)=>t(e,n,void 0,s&&s[n])));else{const n=Object.keys(e);i=new Array(n.length);for(let r=0,a=n.length;r<a;r++){const a=n[r];i[r]=t(e[a],a,r,s&&s[r])}}else i=[];return n&&(n[o]=i),i}function Qe(e,t){for(let n=0;n<t.length;n++){const r=t[n];if((0,a.cy)(r))for(let t=0;t<r.length;t++)e[r[t].name]=r[t].fn;else r&&(e[r.name]=r.key?(...e)=>{const t=r.fn(...e);return t&&(t.key=r.key),t}:r.fn)}return e}function Ue(e,t,n={},r,o){if(M.ce||M.parent&&me(M.parent)&&M.parent.ce)return"default"!==t&&(n.name=t),pn(),vn(on,null,[xn("slot",n,r&&r())],64);let i=e[t];i&&i._c&&(i._d=!1),pn();const s=i&&ze(i(n)),l=n.key||s&&s.key,c=vn(on,{key:(l&&!(0,a.Bm)(l)?l:`_${t}`)+(!s&&r?"_fb":"")},s||(r?r():[]),s&&1===e._?64:-2);return!o&&c.scopeId&&(c.slotScopeIds=[c.scopeId+"-s"]),i&&i._c&&(i._d=!0),c}function ze(e){return e.some((e=>!bn(e)||e.type!==ln&&!(e.type===on&&!ze(e.children))))?e:null}function je(e,t){const n={};for(const r in e)n[t&&/[A-Z]/.test(r)?`on:${r}`:(0,a.rU)(r)]=e[r];return n}const We=e=>e?zn(e)?Gn(e):We(e.parent):null,Xe=(0,a.X$)(Object.create(null),{$:e=>e,$el:e=>e.vnode.el,$data:e=>e.data,$props:e=>e.props,$attrs:e=>e.attrs,$slots:e=>e.slots,$refs:e=>e.refs,$parent:e=>We(e.parent),$root:e=>We(e.root),$host:e=>e.ce,$emit:e=>e.emit,$options:e=>__VUE_OPTIONS_API__?rt(e):e.type,$forceUpdate:e=>e.f||(e.f=()=>{g(e.update)}),$nextTick:e=>e.n||(e.n=m.bind(e.proxy)),$watch:e=>__VUE_OPTIONS_API__?Vt.bind(e):a.tE}),qe=(e,t)=>e!==a.MZ&&!e.__isScriptSetup&&(0,a.$3)(e,t),Ve={get({_:e},t){if("__v_skip"===t)return!0;const{ctx:n,setupState:o,data:i,props:s,accessCache:l,type:c,appContext:u}=e;let d;if("$"!==t[0]){const r=l[t];if(void 0!==r)switch(r){case 1:return o[t];case 2:return i[t];case 4:return n[t];case 3:return s[t]}else{if(qe(o,t))return l[t]=1,o[t];if(i!==a.MZ&&(0,a.$3)(i,t))return l[t]=2,i[t];if((d=e.propsOptions[0])&&(0,a.$3)(d,t))return l[t]=3,s[t];if(n!==a.MZ&&(0,a.$3)(n,t))return l[t]=4,n[t];__VUE_OPTIONS_API__&&!et||(l[t]=0)}}const p=Xe[t];let h,f;return p?("$attrs"===t&&(0,r.u4)(e.attrs,"get",""),p(e)):(h=c.__cssModules)&&(h=h[t])?h:n!==a.MZ&&(0,a.$3)(n,t)?(l[t]=4,n[t]):(f=u.config.globalProperties,(0,a.$3)(f,t)?f[t]:void 0)},set({_:e},t,n){const{data:r,setupState:o,ctx:i}=e;return qe(o,t)?(o[t]=n,!0):r!==a.MZ&&(0,a.$3)(r,t)?(r[t]=n,!0):!((0,a.$3)(e.props,t)||"$"===t[0]&&t.slice(1)in e||(i[t]=n,0))},has({_:{data:e,setupState:t,accessCache:n,ctx:r,appContext:o,propsOptions:i}},s){let l;return!!n[s]||e!==a.MZ&&(0,a.$3)(e,s)||qe(t,s)||(l=i[0])&&(0,a.$3)(l,s)||(0,a.$3)(r,s)||(0,a.$3)(Xe,s)||(0,a.$3)(o.config.globalProperties,s)},defineProperty(e,t,n){return null!=n.get?e._.accessCache[t]=0:(0,a.$3)(n,"value")&&this.set(e,t,n.value,null),Reflect.defineProperty(e,t,n)}};function He(){return Ge().slots}function Ze(){return Ge().attrs}function Ge(){const e=Ln();return e.setupContext||(e.setupContext=Zn(e))}function Ke(e){return(0,a.cy)(e)?e.reduce(((e,t)=>(e[t]=null,e)),{}):e}function Ye(e,t){const n=Ke(e);for(const e in t){if(e.startsWith("__skip"))continue;let r=n[e];r?(0,a.cy)(r)||(0,a.Tn)(r)?r=n[e]={type:r,default:t[e]}:r.default=t[e]:null===r&&(r=n[e]={default:t[e]}),r&&t[`__skip_${e}`]&&(r.skipFactory=!0)}return n}function Je(e,t){return e&&t?(0,a.cy)(e)&&(0,a.cy)(t)?e.concat(t):(0,a.X$)({},Ke(e),Ke(t)):e||t}let et=!0;function tt(e,t,n){i((0,a.cy)(e)?e.map((e=>e.bind(t.proxy))):e.bind(t.proxy),t,n)}function nt(e,t,n,r){let o=r.includes(".")?Ht(n,r):()=>n[r];if((0,a.Kg)(e)){const n=t[e];(0,a.Tn)(n)&&Xt(o,n)}else if((0,a.Tn)(e))Xt(o,e.bind(n));else if((0,a.Gv)(e))if((0,a.cy)(e))e.forEach((e=>nt(e,t,n,r)));else{const r=(0,a.Tn)(e.handler)?e.handler.bind(n):t[e.handler];(0,a.Tn)(r)&&Xt(o,r,e)}}function rt(e){const t=e.type,{mixins:n,extends:r}=t,{mixins:o,optionsCache:i,config:{optionMergeStrategies:s}}=e.appContext,l=i.get(t);let c;return l?c=l:o.length||n||r?(c={},o.length&&o.forEach((e=>at(c,e,s,!0))),at(c,t,s)):c=t,(0,a.Gv)(t)&&i.set(t,c),c}function at(e,t,n,r=!1){const{mixins:a,extends:o}=t;o&&at(e,o,n,!0),a&&a.forEach((t=>at(e,t,n,!0)));for(const a in t)if(r&&"expose"===a);else{const r=ot[a]||n&&n[a];e[a]=r?r(e[a],t[a]):t[a]}return e}const ot={data:it,props:ut,emits:ut,methods:ct,computed:ct,beforeCreate:lt,created:lt,beforeMount:lt,mounted:lt,beforeUpdate:lt,updated:lt,beforeDestroy:lt,beforeUnmount:lt,destroyed:lt,unmounted:lt,activated:lt,deactivated:lt,errorCaptured:lt,serverPrefetch:lt,components:ct,directives:ct,watch:function(e,t){if(!e)return t;if(!t)return e;const n=(0,a.X$)(Object.create(null),e);for(const r in t)n[r]=lt(e[r],t[r]);return n},provide:it,inject:function(e,t){return ct(st(e),st(t))}};function it(e,t){return t?e?function(){return(0,a.X$)((0,a.Tn)(e)?e.call(this,this):e,(0,a.Tn)(t)?t.call(this,this):t)}:t:e}function st(e){if((0,a.cy)(e)){const t={};for(let n=0;n<e.length;n++)t[e[n]]=e[n];return t}return e}function lt(e,t){return e?[...new Set([].concat(e,t))]:t}function ct(e,t){return e?(0,a.X$)(Object.create(null),e,t):t}function ut(e,t){return e?(0,a.cy)(e)&&(0,a.cy)(t)?[...new Set([...e,...t])]:(0,a.X$)(Object.create(null),Ke(e),Ke(null!=t?t:{})):t}function dt(){return{app:null,config:{isNativeTag:a.NO,performance:!1,globalProperties:{},optionMergeStrategies:{},errorHandler:void 0,warnHandler:void 0,compilerOptions:{}},mixins:[],components:{},directives:{},provides:Object.create(null),optionsCache:new WeakMap,propsCache:new WeakMap,emitsCache:new WeakMap}}let pt=0;function ht(e,t){return function(n,r=null){(0,a.Tn)(n)||(n=(0,a.X$)({},n)),null==r||(0,a.Gv)(r)||(r=null);const o=dt(),s=new WeakSet,l=[];let c=!1;const u=o.app={_uid:pt++,_component:n,_props:r,_container:null,_context:o,_instance:null,version:Jn,get config(){return o.config},set config(e){},use:(e,...t)=>(s.has(e)||(e&&(0,a.Tn)(e.install)?(s.add(e),e.install(u,...t)):(0,a.Tn)(e)&&(s.add(e),e(u,...t))),u),mixin:e=>(__VUE_OPTIONS_API__&&(o.mixins.includes(e)||o.mixins.push(e)),u),component:(e,t)=>t?(o.components[e]=t,u):o.components[e],directive:(e,t)=>t?(o.directives[e]=t,u):o.directives[e],mount(a,i,s){if(!c){const l=u._ceVNode||xn(n,r);return l.appContext=o,!0===s?s="svg":!1===s&&(s=void 0),i&&t?t(l,a):e(l,a,s),c=!0,u._container=a,a.__vue_app__=u,__VUE_PROD_DEVTOOLS__&&(u._instance=l.component,function(e,t){_("app:init",e,t,{Fragment:on,Text:sn,Comment:ln,Static:cn})}(u,Jn)),Gn(l.component)}},onUnmount(e){l.push(e)},unmount(){c&&(i(l,u._instance,16),e(null,u._container),__VUE_PROD_DEVTOOLS__&&(u._instance=null,function(e){_("app:unmount",e)}(u)),delete u._container.__vue_app__)},provide:(e,t)=>(o.provides[e]=t,u),runWithContext(e){const t=ft;ft=u;try{return e()}finally{ft=t}}};return u}}let ft=null;function mt(e,t){if(In){let n=In.provides;const r=In.parent&&In.parent.provides;r===n&&(n=In.provides=Object.create(r)),n[e]=t}}function gt(e,t,n=!1){const r=In||M;if(r||ft){const o=ft?ft._context.provides:r?null==r.parent?r.vnode.appContext&&r.vnode.appContext.provides:r.parent.provides:void 0;if(o&&e in o)return o[e];if(arguments.length>1)return n&&(0,a.Tn)(t)?t.call(r&&r.proxy):t}}const vt={},bt=()=>Object.create(vt),yt=e=>Object.getPrototypeOf(e)===vt;function wt(e,t,n,o){const[i,s]=e.propsOptions;let l,c=!1;if(t)for(let r in t){if((0,a.SU)(r))continue;const u=t[r];let d;i&&(0,a.$3)(i,d=(0,a.PT)(r))?s&&s.includes(d)?(l||(l={}))[d]=u:n[d]=u:Jt(e.emitsOptions,r)||r in o&&u===o[r]||(o[r]=u,c=!0)}if(s){const t=(0,r.ux)(n),o=l||a.MZ;for(let r=0;r<s.length;r++){const l=s[r];n[l]=Ot(i,t,l,o[l],e,!(0,a.$3)(o,l))}}return c}function Ot(e,t,n,r,o,i){const s=e[n];if(null!=s){const e=(0,a.$3)(s,"default");if(e&&void 0===r){const e=s.default;if(s.type!==Function&&!s.skipFactory&&(0,a.Tn)(e)){const{propsDefaults:a}=o;if(n in a)r=a[n];else{const i=Qn(o);r=a[n]=e.call(null,t),i()}}else r=e;o.ce&&o.ce._setProp(n,r)}s[0]&&(i&&!e?r=!1:!s[1]||""!==r&&r!==(0,a.Tg)(n)||(r=!0))}return r}const kt=new WeakMap;function xt(e,t,n=!1){const r=__VUE_OPTIONS_API__&&n?kt:t.propsCache,o=r.get(e);if(o)return o;const i=e.props,s={},l=[];let c=!1;if(__VUE_OPTIONS_API__&&!(0,a.Tn)(e)){const r=e=>{c=!0;const[n,r]=xt(e,t,!0);(0,a.X$)(s,n),r&&l.push(...r)};!n&&t.mixins.length&&t.mixins.forEach(r),e.extends&&r(e.extends),e.mixins&&e.mixins.forEach(r)}if(!i&&!c)return(0,a.Gv)(e)&&r.set(e,a.Oj),a.Oj;if((0,a.cy)(i))for(let e=0;e<i.length;e++){const t=(0,a.PT)(i[e]);Et(t)&&(s[t]=a.MZ)}else if(i)for(const e in i){const t=(0,a.PT)(e);if(Et(t)){const n=i[e],r=s[t]=(0,a.cy)(n)||(0,a.Tn)(n)?{type:n}:(0,a.X$)({},n),o=r.type;let c=!1,u=!0;if((0,a.cy)(o))for(let e=0;e<o.length;++e){const t=o[e],n=(0,a.Tn)(t)&&t.name;if("Boolean"===n){c=!0;break}"String"===n&&(u=!1)}else c=(0,a.Tn)(o)&&"Boolean"===o.name;r[0]=c,r[1]=u,(c||(0,a.$3)(r,"default"))&&l.push(t)}}const u=[s,l];return(0,a.Gv)(e)&&r.set(e,u),u}function Et(e){return"$"!==e[0]&&!(0,a.SU)(e)}const St=e=>"_"===e[0]||"$stable"===e,_t=e=>(0,a.cy)(e)?e.map(An):[An(e)],Ct=(e,t,n)=>{if(t._n)return t;const r=I(((...e)=>_t(t(...e))),n);return r._c=!1,r},Tt=(e,t,n)=>{const r=e._ctx;for(const n in e){if(St(n))continue;const o=e[n];if((0,a.Tn)(o))t[n]=Ct(0,o,r);else if(null!=o){const e=_t(o);t[n]=()=>e}}},At=(e,t)=>{const n=_t(t);e.slots.default=()=>n},Rt=(e,t,n)=>{for(const r in t)(n||"_"!==r)&&(e[r]=t[r])},Pt=(e,t,n)=>{const r=e.slots=bt();if(32&e.vnode.shapeFlag){const e=t._;e?(Rt(r,t,n),n&&(0,a.yQ)(r,"_",e,!0)):Tt(t,r)}else t&&At(e,t)},Dt=(e,t,n)=>{const{vnode:r,slots:o}=e;let i=!0,s=a.MZ;if(32&r.shapeFlag){const e=t._;e?n&&1===e?i=!1:Rt(o,t,n):(i=!t.$stable,Tt(t,o)),s=t}else t&&(At(e,t),s={default:1});if(i)for(const e in o)St(e)||null!=s[e]||delete o[e]},Mt=function(e,t){t&&t.pendingBranch?(0,a.cy)(e)?t.effects.push(...e):t.effects.push(e):b(e)};function $t(e){return function(e,t){"boolean"!=typeof __VUE_OPTIONS_API__&&((0,a.We)().__VUE_OPTIONS_API__=!0),"boolean"!=typeof __VUE_PROD_DEVTOOLS__&&((0,a.We)().__VUE_PROD_DEVTOOLS__=!1),"boolean"!=typeof __VUE_PROD_HYDRATION_MISMATCH_DETAILS__&&((0,a.We)().__VUE_PROD_HYDRATION_MISMATCH_DETAILS__=!1);const n=(0,a.We)();n.__VUE__=!0,__VUE_PROD_DEVTOOLS__&&C(n.__VUE_DEVTOOLS_GLOBAL_HOOK__,n);const{insert:i,remove:l,patchProp:c,createElement:u,createText:d,createComment:p,setText:h,setElementText:f,parentNode:m,nextSibling:v,setScopeId:b=a.tE,insertStaticContent:O}=e,k=(e,t,n,r=null,a=null,o=null,i=void 0,s=null,l=!!t.dynamicChildren)=>{if(e===t)return;e&&!yn(e,t)&&(r=te(e),G(e,a,o,!0),e=null),-2===t.patchFlag&&(l=!1,t.dynamicChildren=null);const{type:c,ref:u,shapeFlag:d}=t;switch(c){case sn:x(e,t,n,r);break;case ln:E(e,t,n,r);break;case cn:null==e&&S(t,n,r,i);break;case on:Q(e,t,n,r,a,o,i,s,l);break;default:1&d?R(e,t,n,r,a,o,i,s,l):6&d?U(e,t,n,r,a,o,i,s,l):(64&d||128&d)&&c.process(e,t,n,r,a,o,i,s,l,ae)}null!=u&&a&&fe(u,e&&e.ref,o,t||e,!t)},x=(e,t,n,r)=>{if(null==e)i(t.el=d(t.children),n,r);else{const n=t.el=e.el;t.children!==e.children&&h(n,t.children)}},E=(e,t,n,r)=>{null==e?i(t.el=p(t.children||""),n,r):t.el=e.el},S=(e,t,n,r)=>{[e.el,e.anchor]=O(e.children,t,n,r,e.el,e.anchor)},_=({el:e,anchor:t})=>{let n;for(;e&&e!==t;)n=v(e),l(e),e=n;l(t)},R=(e,t,n,r,a,o,i,s,l)=>{"svg"===t.type?i="svg":"math"===t.type&&(i="mathml"),null==e?D(t,n,r,a,o,i,s,l):N(e,t,a,o,i,s,l)},D=(e,t,n,r,o,s,l,d)=>{let p,h;const{props:m,shapeFlag:g,transition:v,dirs:b}=e;if(p=e.el=u(e.type,s,m&&m.is,m),8&g?f(p,e.children):16&g&&$(e.children,p,null,r,o,Nt(e,s),l,d),b&&B(e,null,r,"created"),M(p,e,e.scopeId,l,r),m){for(const e in m)"value"===e||(0,a.SU)(e)||c(p,e,null,m[e],s,r);"value"in m&&c(p,"value",null,m.value,s),(h=m.onVnodeBeforeMount)&&Mn(h,r,e)}__VUE_PROD_DEVTOOLS__&&((0,a.yQ)(p,"__vnode",e,!0),(0,a.yQ)(p,"__vueParentComponent",r,!0)),b&&B(e,null,r,"beforeMount");const y=function(e,t){return(!e||e&&!e.pendingBranch)&&t&&!t.persisted}(o,v);y&&v.beforeEnter(p),i(p,t,n),((h=m&&m.onVnodeMounted)||y||b)&&Mt((()=>{h&&Mn(h,r,e),y&&v.enter(p),b&&B(e,null,r,"mounted")}),o)},M=(e,t,n,r,a)=>{if(n&&b(e,n),r)for(let t=0;t<r.length;t++)b(e,r[t]);if(a){let n=a.subTree;if(t===n||an(n.type)&&(n.ssContent===t||n.ssFallback===t)){const t=a.vnode;M(e,t,t.scopeId,t.slotScopeIds,a.parent)}}},$=(e,t,n,r,a,o,i,s,l=0)=>{for(let c=l;c<e.length;c++){const l=e[c]=s?Rn(e[c]):An(e[c]);k(null,l,t,n,r,a,o,i,s)}},N=(e,t,n,r,o,i,s)=>{const l=t.el=e.el;__VUE_PROD_DEVTOOLS__&&(l.__vnode=t);let{patchFlag:u,dynamicChildren:d,dirs:p}=t;u|=16&e.patchFlag;const h=e.props||a.MZ,m=t.props||a.MZ;let g;if(n&&It(n,!1),(g=m.onVnodeBeforeUpdate)&&Mn(g,n,t,e),p&&B(t,e,n,"beforeUpdate"),n&&It(n,!0),(h.innerHTML&&null==m.innerHTML||h.textContent&&null==m.textContent)&&f(l,""),d?I(e.dynamicChildren,d,l,n,r,Nt(t,o),i):s||q(e,t,l,null,n,r,Nt(t,o),i,!1),u>0){if(16&u)L(l,h,m,n,o);else if(2&u&&h.class!==m.class&&c(l,"class",null,m.class,o),4&u&&c(l,"style",h.style,m.style,o),8&u){const e=t.dynamicProps;for(let t=0;t<e.length;t++){const r=e[t],a=h[r],i=m[r];i===a&&"value"!==r||c(l,r,a,i,o,n)}}1&u&&e.children!==t.children&&f(l,t.children)}else s||null!=d||L(l,h,m,n,o);((g=m.onVnodeUpdated)||p)&&Mt((()=>{g&&Mn(g,n,t,e),p&&B(t,e,n,"updated")}),r)},I=(e,t,n,r,a,o,i)=>{for(let s=0;s<t.length;s++){const l=e[s],c=t[s],u=l.el&&(l.type===on||!yn(l,c)||70&l.shapeFlag)?m(l.el):n;k(l,c,u,null,r,a,o,i,!0)}},L=(e,t,n,r,o)=>{if(t!==n){if(t!==a.MZ)for(const i in t)(0,a.SU)(i)||i in n||c(e,i,t[i],null,o,r);for(const i in n){if((0,a.SU)(i))continue;const s=n[i],l=t[i];s!==l&&"value"!==i&&c(e,i,l,s,o,r)}"value"in n&&c(e,"value",t.value,n.value,o)}},Q=(e,t,n,r,a,o,s,l,c)=>{const u=t.el=e?e.el:d(""),p=t.anchor=e?e.anchor:d("");let{patchFlag:h,dynamicChildren:f,slotScopeIds:m}=t;m&&(l=l?l.concat(m):m),null==e?(i(u,n,r),i(p,n,r),$(t.children||[],n,p,a,o,s,l,c)):h>0&&64&h&&f&&e.dynamicChildren?(I(e.dynamicChildren,f,n,a,o,s,l),(null!=t.key||a&&t===a.subTree)&&Lt(e,t,!0)):q(e,t,n,p,a,o,s,l,c)},U=(e,t,n,r,a,o,i,s,l)=>{t.slotScopeIds=s,null==e?512&t.shapeFlag?a.ctx.activate(t,n,r,i,l):z(t,n,r,a,o,i,l):j(e,t,l)},z=(e,t,n,i,l,c,u)=>{const d=e.component=function(e,t,n){const o=e.type,i=(t?t.appContext:e.appContext)||$n,s={uid:Nn++,vnode:e,type:o,parent:t,appContext:i,root:null,next:null,subTree:null,effect:null,update:null,job:null,scope:new r.yC(!0),render:null,proxy:null,exposed:null,exposeProxy:null,withProxy:null,provides:t?t.provides:Object.create(i.provides),ids:t?t.ids:["",0,0],accessCache:null,renderCache:[],components:null,directives:null,propsOptions:xt(o,i),emitsOptions:Yt(o,i),emit:null,emitted:null,propsDefaults:a.MZ,inheritAttrs:o.inheritAttrs,ctx:a.MZ,data:a.MZ,props:a.MZ,attrs:a.MZ,slots:a.MZ,refs:a.MZ,setupState:a.MZ,setupContext:null,suspense:n,suspenseId:n?n.pendingId:0,asyncDep:null,asyncResolved:!1,isMounted:!1,isUnmounted:!1,isDeactivated:!1,bc:null,c:null,bm:null,m:null,bu:null,u:null,um:null,bum:null,da:null,a:null,rtg:null,rtc:null,ec:null,sp:null};return s.ctx={_:s},s.root=t?t.root:s,s.emit=Kt.bind(null,s),e.ce&&e.ce(s),s}(e,i,l);if(ge(e)&&(d.ctx.renderer=ae),function(e,t=!1,n=!1){t&&Fn(t);const{props:i,children:l}=e.vnode,c=zn(e);!function(e,t,n,a=!1){const o={},i=bt();e.propsDefaults=Object.create(null),wt(e,t,o,i);for(const t in e.propsOptions[0])t in o||(o[t]=void 0);n?e.props=a?o:(0,r.Gc)(o):e.type.props?e.props=o:e.props=i,e.attrs=i}(e,i,c,t),Pt(e,l,n);const u=c?function(e,t){const n=e.type;e.accessCache=Object.create(null),e.proxy=new Proxy(e.ctx,Ve);const{setup:i}=n;if(i){(0,r.C4)();const n=e.setupContext=i.length>1?Zn(e):null,l=Qn(e),c=o(i,e,0,[e.props,n]),u=(0,a.yL)(c);if((0,r.bl)(),l(),!u&&!e.sp||me(e)||he(e),u){if(c.then(Un,Un),t)return c.then((n=>{qn(e,n,t)})).catch((t=>{s(t,e,0)}));e.asyncDep=c}else qn(e,c,t)}else Vn(e,t)}(e,t):void 0;t&&Fn(!1)}(d,!1,u),d.asyncDep){if(l&&l.registerDep(d,W,u),!e.el){const e=d.subTree=xn(ln);E(null,e,t,n)}}else W(d,e,t,n,l,c,u)},j=(e,t,n)=>{const r=t.component=e.component;if(function(e,t,n){const{props:r,children:a,component:o}=e,{props:i,children:s,patchFlag:l}=t,c=o.emitsOptions;if(t.dirs||t.transition)return!0;if(!(n&&l>=0))return!(!a&&!s||s&&s.$stable)||r!==i&&(r?!i||rn(r,i,c):!!i);if(1024&l)return!0;if(16&l)return r?rn(r,i,c):!!i;if(8&l){const e=t.dynamicProps;for(let t=0;t<e.length;t++){const n=e[t];if(i[n]!==r[n]&&!Jt(c,n))return!0}}return!1}(e,t,n)){if(r.asyncDep&&!r.asyncResolved)return void X(r,t,n);r.next=t,r.update()}else t.el=e.el,r.vnode=t},W=(e,t,n,o,i,s,l)=>{const c=()=>{if(e.isMounted){let{next:t,bu:n,u:r,parent:o,vnode:u}=e;{const n=Bt(e);if(n)return t&&(t.el=u.el,X(e,t,l)),void n.asyncDep.then((()=>{e.isUnmounted||c()}))}let d,p=t;It(e,!1),t?(t.el=u.el,X(e,t,l)):t=u,n&&(0,a.DY)(n),(d=t.props&&t.props.onVnodeBeforeUpdate)&&Mn(d,o,t,u),It(e,!0);const h=en(e),f=e.subTree;e.subTree=h,k(f,h,m(f.el),te(f),e,i,s),t.el=h.el,null===p&&function({vnode:e,parent:t},n){for(;t;){const r=t.subTree;if(r.suspense&&r.suspense.activeBranch===e&&(r.el=e.el),r!==e)break;(e=t.vnode).el=n,t=t.parent}}(e,h.el),r&&Mt(r,i),(d=t.props&&t.props.onVnodeUpdated)&&Mt((()=>Mn(d,o,t,u)),i),__VUE_PROD_DEVTOOLS__&&A(e)}else{let r;const{el:l,props:c}=t,{bm:u,m:d,parent:p,root:h,type:f}=e,m=me(t);if(It(e,!1),u&&(0,a.DY)(u),!m&&(r=c&&c.onVnodeBeforeMount)&&Mn(r,p,t),It(e,!0),l&&ie){const t=()=>{e.subTree=en(e),ie(l,e.subTree,e,i,null)};m&&f.__asyncHydrate?f.__asyncHydrate(l,e,t):t()}else{h.ce&&h.ce._injectChildStyle(f);const r=e.subTree=en(e);k(null,r,n,o,e,i,s),t.el=r.el}if(d&&Mt(d,i),!m&&(r=c&&c.onVnodeMounted)){const e=t;Mt((()=>Mn(r,p,e)),i)}(256&t.shapeFlag||p&&me(p.vnode)&&256&p.vnode.shapeFlag)&&e.a&&Mt(e.a,i),e.isMounted=!0,__VUE_PROD_DEVTOOLS__&&T(e),t=n=o=null}};e.scope.on();const u=e.effect=new r.X2(c);e.scope.off();const d=e.update=u.run.bind(u),p=e.job=u.runIfDirty.bind(u);p.i=e,p.id=e.uid,u.scheduler=()=>g(p),It(e,!0),d()},X=(e,t,n)=>{t.component=e;const o=e.vnode.props;e.vnode=t,e.next=null,function(e,t,n,o){const{props:i,attrs:s,vnode:{patchFlag:l}}=e,c=(0,r.ux)(i),[u]=e.propsOptions;let d=!1;if(!(o||l>0)||16&l){let r;wt(e,t,i,s)&&(d=!0);for(const o in c)t&&((0,a.$3)(t,o)||(r=(0,a.Tg)(o))!==o&&(0,a.$3)(t,r))||(u?!n||void 0===n[o]&&void 0===n[r]||(i[o]=Ot(u,c,o,void 0,e,!0)):delete i[o]);if(s!==c)for(const e in s)t&&(0,a.$3)(t,e)||(delete s[e],d=!0)}else if(8&l){const n=e.vnode.dynamicProps;for(let r=0;r<n.length;r++){let o=n[r];if(Jt(e.emitsOptions,o))continue;const l=t[o];if(u)if((0,a.$3)(s,o))l!==s[o]&&(s[o]=l,d=!0);else{const t=(0,a.PT)(o);i[t]=Ot(u,c,t,l,e,!1)}else l!==s[o]&&(s[o]=l,d=!0)}}d&&(0,r.hZ)(e.attrs,"set","")}(e,t.props,o,n),Dt(e,t.children,n),(0,r.C4)(),y(e),(0,r.bl)()},q=(e,t,n,r,a,o,i,s,l=!1)=>{const c=e&&e.children,u=e?e.shapeFlag:0,d=t.children,{patchFlag:p,shapeFlag:h}=t;if(p>0){if(128&p)return void H(c,d,n,r,a,o,i,s,l);if(256&p)return void V(c,d,n,r,a,o,i,s,l)}8&h?(16&u&&ee(c,a,o),d!==c&&f(n,d)):16&u?16&h?H(c,d,n,r,a,o,i,s,l):ee(c,a,o,!0):(8&u&&f(n,""),16&h&&$(d,n,r,a,o,i,s,l))},V=(e,t,n,r,o,i,s,l,c)=>{e=e||a.Oj,t=t||a.Oj;const u=e.length,d=t.length,p=Math.min(u,d);let h;for(h=0;h<p;h++){const r=t[h]=c?Rn(t[h]):An(t[h]);k(e[h],r,n,null,o,i,s,l,c)}u>d?ee(e,o,i,!0,!1,p):$(t,n,r,o,i,s,l,c,p)},H=(e,t,n,r,o,i,s,l,c)=>{let u=0;const d=t.length;let p=e.length-1,h=d-1;for(;u<=p&&u<=h;){const r=e[u],a=t[u]=c?Rn(t[u]):An(t[u]);if(!yn(r,a))break;k(r,a,n,null,o,i,s,l,c),u++}for(;u<=p&&u<=h;){const r=e[p],a=t[h]=c?Rn(t[h]):An(t[h]);if(!yn(r,a))break;k(r,a,n,null,o,i,s,l,c),p--,h--}if(u>p){if(u<=h){const e=h+1,a=e<d?t[e].el:r;for(;u<=h;)k(null,t[u]=c?Rn(t[u]):An(t[u]),n,a,o,i,s,l,c),u++}}else if(u>h)for(;u<=p;)G(e[u],o,i,!0),u++;else{const f=u,m=u,g=new Map;for(u=m;u<=h;u++){const e=t[u]=c?Rn(t[u]):An(t[u]);null!=e.key&&g.set(e.key,u)}let v,b=0;const y=h-m+1;let w=!1,O=0;const x=new Array(y);for(u=0;u<y;u++)x[u]=0;for(u=f;u<=p;u++){const r=e[u];if(b>=y){G(r,o,i,!0);continue}let a;if(null!=r.key)a=g.get(r.key);else for(v=m;v<=h;v++)if(0===x[v-m]&&yn(r,t[v])){a=v;break}void 0===a?G(r,o,i,!0):(x[a-m]=u+1,a>=O?O=a:w=!0,k(r,t[a],n,null,o,i,s,l,c),b++)}const E=w?function(e){const t=e.slice(),n=[0];let r,a,o,i,s;const l=e.length;for(r=0;r<l;r++){const l=e[r];if(0!==l){if(a=n[n.length-1],e[a]<l){t[r]=a,n.push(r);continue}for(o=0,i=n.length-1;o<i;)s=o+i>>1,e[n[s]]<l?o=s+1:i=s;l<e[n[o]]&&(o>0&&(t[r]=n[o-1]),n[o]=r)}}for(o=n.length,i=n[o-1];o-- >0;)n[o]=i,i=t[i];return n}(x):a.Oj;for(v=E.length-1,u=y-1;u>=0;u--){const e=m+u,a=t[e],p=e+1<d?t[e+1].el:r;0===x[u]?k(null,a,n,p,o,i,s,l,c):w&&(v<0||u!==E[v]?Z(a,n,p,2):v--)}}},Z=(e,t,n,r,a=null)=>{const{el:o,type:s,transition:l,children:c,shapeFlag:u}=e;if(6&u)Z(e.component.subTree,t,n,r);else if(128&u)e.suspense.move(t,n,r);else if(64&u)s.move(e,t,n,ae);else if(s!==on)if(s!==cn)if(2!==r&&1&u&&l)if(0===r)l.beforeEnter(o),i(o,t,n),Mt((()=>l.enter(o)),a);else{const{leave:e,delayLeave:r,afterLeave:a}=l,s=()=>i(o,t,n),c=()=>{e(o,(()=>{s(),a&&a()}))};r?r(o,s,c):c()}else i(o,t,n);else(({el:e,anchor:t},n,r)=>{let a;for(;e&&e!==t;)a=v(e),i(e,n,r),e=a;i(t,n,r)})(e,t,n);else{i(o,t,n);for(let e=0;e<c.length;e++)Z(c[e],t,n,r);i(e.anchor,t,n)}},G=(e,t,n,r=!1,a=!1)=>{const{type:o,props:i,ref:s,children:l,dynamicChildren:c,shapeFlag:u,patchFlag:d,dirs:p,cacheIndex:h}=e;if(-2===d&&(a=!1),null!=s&&fe(s,null,n,e,!0),null!=h&&(t.renderCache[h]=void 0),256&u)return void t.ctx.deactivate(e);const f=1&u&&p,m=!me(e);let g;if(m&&(g=i&&i.onVnodeBeforeUnmount)&&Mn(g,t,e),6&u)J(e.component,n,r);else{if(128&u)return void e.suspense.unmount(n,r);f&&B(e,null,t,"beforeUnmount"),64&u?e.type.remove(e,t,n,ae,r):c&&!c.hasOnce&&(o!==on||d>0&&64&d)?ee(c,t,n,!1,!0):(o===on&&384&d||!a&&16&u)&&ee(l,t,n),r&&K(e)}(m&&(g=i&&i.onVnodeUnmounted)||f)&&Mt((()=>{g&&Mn(g,t,e),f&&B(e,null,t,"unmounted")}),n)},K=e=>{const{type:t,el:n,anchor:r,transition:a}=e;if(t===on)return void Y(n,r);if(t===cn)return void _(e);const o=()=>{l(n),a&&!a.persisted&&a.afterLeave&&a.afterLeave()};if(1&e.shapeFlag&&a&&!a.persisted){const{leave:t,delayLeave:r}=a,i=()=>t(n,o);r?r(e.el,o,i):i()}else o()},Y=(e,t)=>{let n;for(;e!==t;)n=v(e),l(e),e=n;l(t)},J=(e,t,n)=>{const{bum:r,scope:o,job:i,subTree:s,um:l,m:c,a:u}=e;Ft(c),Ft(u),r&&(0,a.DY)(r),o.stop(),i&&(i.flags|=8,G(s,e,t,n)),l&&Mt(l,t),Mt((()=>{e.isUnmounted=!0}),t),t&&t.pendingBranch&&!t.isUnmounted&&e.asyncDep&&!e.asyncResolved&&e.suspenseId===t.pendingId&&(t.deps--,0===t.deps&&t.resolve()),__VUE_PROD_DEVTOOLS__&&P(e)},ee=(e,t,n,r=!1,a=!1,o=0)=>{for(let i=o;i<e.length;i++)G(e[i],t,n,r,a)},te=e=>{if(6&e.shapeFlag)return te(e.component.subTree);if(128&e.shapeFlag)return e.suspense.next();const t=v(e.anchor||e.el),n=t&&t[F];return n?v(n):t};let ne=!1;const re=(e,t,n)=>{null==e?t._vnode&&G(t._vnode,null,null,!0):k(t._vnode||null,e,t,null,null,null,n),t._vnode=e,ne||(ne=!0,y(),w(),ne=!1)},ae={p:k,um:G,m:Z,r:K,mt:z,mc:$,pc:q,pbc:I,n:te,o:e};let oe,ie;return t&&([oe,ie]=t(ae)),{render:re,hydrate:oe,createApp:ht(re,oe)}}(e)}function Nt({type:e,props:t},n){return"svg"===n&&"foreignObject"===e||"mathml"===n&&"annotation-xml"===e&&t&&t.encoding&&t.encoding.includes("html")?void 0:n}function It({effect:e,job:t},n){n?(e.flags|=32,t.flags|=4):(e.flags&=-33,t.flags&=-5)}function Lt(e,t,n=!1){const r=e.children,o=t.children;if((0,a.cy)(r)&&(0,a.cy)(o))for(let e=0;e<r.length;e++){const t=r[e];let a=o[e];1&a.shapeFlag&&!a.dynamicChildren&&((a.patchFlag<=0||32===a.patchFlag)&&(a=o[e]=Rn(o[e]),a.el=t.el),n||-2===a.patchFlag||Lt(t,a)),a.type===sn&&(a.el=t.el)}}function Bt(e){const t=e.subTree.component;if(t)return t.asyncDep&&!t.asyncResolved?t:Bt(t)}function Ft(e){if(e)for(let t=0;t<e.length;t++)e[t].flags|=8}const Qt=Symbol.for("v-scx"),Ut=()=>gt(Qt);function zt(e,t){return qt(e,null,t)}function jt(e,t){return qt(e,null,{flush:"post"})}function Wt(e,t){return qt(e,null,{flush:"sync"})}function Xt(e,t,n){return qt(e,t,n)}function qt(e,t,n=a.MZ){const{immediate:o,deep:s,flush:l,once:c}=n,u=(0,a.X$)({},n),d=t&&o||!t&&"post"!==l;let p;if(Xn)if("sync"===l){const e=Ut();p=e.__watcherHandles||(e.__watcherHandles=[])}else if(!d){const e=()=>{};return e.stop=a.tE,e.resume=a.tE,e.pause=a.tE,e}const h=In;u.call=(e,t,n)=>i(e,h,t,n);let f=!1;"post"===l?u.scheduler=e=>{Mt(e,h&&h.suspense)}:"sync"!==l&&(f=!0,u.scheduler=(e,t)=>{t?e():g(e)}),u.augmentJob=e=>{t&&(e.flags|=4),f&&(e.flags|=2,h&&(e.id=h.uid,e.i=h))};const m=(0,r.wB)(e,t,u);return Xn&&(p?p.push(m):d&&m()),m}function Vt(e,t,n){const r=this.proxy,o=(0,a.Kg)(e)?e.includes(".")?Ht(r,e):()=>r[e]:e.bind(r,r);let i;(0,a.Tn)(t)?i=t:(i=t.handler,n=t);const s=Qn(this),l=qt(o,i.bind(r),n);return s(),l}function Ht(e,t){const n=t.split(".");return()=>{let t=e;for(let e=0;e<n.length&&t;e++)t=t[n[e]];return t}}function Zt(e,t,n=a.MZ){const o=Ln(),i=(0,a.PT)(t),s=(0,a.Tg)(t),l=Gt(e,i),c=(0,r.rY)(((r,l)=>{let c,u,d=a.MZ;return Wt((()=>{const t=e[i];(0,a.$H)(c,t)&&(c=t,l())})),{get:()=>(r(),n.get?n.get(c):c),set(e){const r=n.set?n.set(e):e;if(!((0,a.$H)(r,c)||d!==a.MZ&&(0,a.$H)(e,d)))return;const p=o.vnode.props;p&&(t in p||i in p||s in p)&&(`onUpdate:${t}`in p||`onUpdate:${i}`in p||`onUpdate:${s}`in p)||(c=e,l()),o.emit(`update:${t}`,r),(0,a.$H)(e,r)&&(0,a.$H)(e,d)&&!(0,a.$H)(r,u)&&l(),d=e,u=r}}}));return c[Symbol.iterator]=()=>{let e=0;return{next:()=>e<2?{value:e++?l||a.MZ:c,done:!1}:{done:!0}}},c}const Gt=(e,t)=>"modelValue"===t||"model-value"===t?e.modelModifiers:e[`${t}Modifiers`]||e[`${(0,a.PT)(t)}Modifiers`]||e[`${(0,a.Tg)(t)}Modifiers`];function Kt(e,t,...n){if(e.isUnmounted)return;const r=e.vnode.props||a.MZ;let o=n;const s=t.startsWith("update:"),l=s&&Gt(r,t.slice(7));let c;l&&(l.trim&&(o=n.map((e=>(0,a.Kg)(e)?e.trim():e))),l.number&&(o=n.map(a.bB))),__VUE_PROD_DEVTOOLS__&&function(e,t,n){_("component:emit",e.appContext.app,e,t,n)}(e,t,o);let u=r[c=(0,a.rU)(t)]||r[c=(0,a.rU)((0,a.PT)(t))];!u&&s&&(u=r[c=(0,a.rU)((0,a.Tg)(t))]),u&&i(u,e,6,o);const d=r[c+"Once"];if(d){if(e.emitted){if(e.emitted[c])return}else e.emitted={};e.emitted[c]=!0,i(d,e,6,o)}}function Yt(e,t,n=!1){const r=t.emitsCache,o=r.get(e);if(void 0!==o)return o;const i=e.emits;let s={},l=!1;if(__VUE_OPTIONS_API__&&!(0,a.Tn)(e)){const r=e=>{const n=Yt(e,t,!0);n&&(l=!0,(0,a.X$)(s,n))};!n&&t.mixins.length&&t.mixins.forEach(r),e.extends&&r(e.extends),e.mixins&&e.mixins.forEach(r)}return i||l?((0,a.cy)(i)?i.forEach((e=>s[e]=null)):(0,a.X$)(s,i),(0,a.Gv)(e)&&r.set(e,s),s):((0,a.Gv)(e)&&r.set(e,null),null)}function Jt(e,t){return!(!e||!(0,a.Mp)(t))&&(t=t.slice(2).replace(/Once$/,""),(0,a.$3)(e,t[0].toLowerCase()+t.slice(1))||(0,a.$3)(e,(0,a.Tg)(t))||(0,a.$3)(e,t))}function en(e){const{type:t,vnode:n,proxy:r,withProxy:o,propsOptions:[i],slots:l,attrs:c,emit:u,render:d,renderCache:p,props:h,data:f,setupState:m,ctx:g,inheritAttrs:v}=e,b=N(e);let y,w;try{if(4&n.shapeFlag){const e=o||r,t=e;y=An(d.call(t,e,p,h,m,f,g)),w=c}else{const e=t;y=An(e.length>1?e(h,{attrs:c,slots:l,emit:u}):e(h,null)),w=t.props?c:tn(c)}}catch(t){un.length=0,s(t,e,1),y=xn(ln)}let O=y;if(w&&!1!==v){const e=Object.keys(w),{shapeFlag:t}=O;e.length&&7&t&&(i&&e.some(a.CP)&&(w=nn(w,i)),O=Sn(O,w,!1,!0))}return n.dirs&&(O=Sn(O,null,!1,!0),O.dirs=O.dirs?O.dirs.concat(n.dirs):n.dirs),n.transition&&ce(O,n.transition),y=O,N(b),y}const tn=e=>{let t;for(const n in e)("class"===n||"style"===n||(0,a.Mp)(n))&&((t||(t={}))[n]=e[n]);return t},nn=(e,t)=>{const n={};for(const r in e)(0,a.CP)(r)&&r.slice(9)in t||(n[r]=e[r]);return n};function rn(e,t,n){const r=Object.keys(t);if(r.length!==Object.keys(e).length)return!0;for(let a=0;a<r.length;a++){const o=r[a];if(t[o]!==e[o]&&!Jt(n,o))return!0}return!1}const an=e=>e.__isSuspense;const on=Symbol.for("v-fgt"),sn=Symbol.for("v-txt"),ln=Symbol.for("v-cmt"),cn=Symbol.for("v-stc"),un=[];let dn=null;function pn(e=!1){un.push(dn=e?null:[])}let hn=1;function fn(e,t=!1){hn+=e,e<0&&dn&&t&&(dn.hasOnce=!0)}function mn(e){return e.dynamicChildren=hn>0?dn||a.Oj:null,un.pop(),dn=un[un.length-1]||null,hn>0&&dn&&dn.push(e),e}function gn(e,t,n,r,a,o){return mn(kn(e,t,n,r,a,o,!0))}function vn(e,t,n,r,a){return mn(xn(e,t,n,r,a,!0))}function bn(e){return!!e&&!0===e.__v_isVNode}function yn(e,t){return e.type===t.type&&e.key===t.key}const wn=({key:e})=>null!=e?e:null,On=({ref:e,ref_key:t,ref_for:n})=>("number"==typeof e&&(e=""+e),null!=e?(0,a.Kg)(e)||(0,r.i9)(e)||(0,a.Tn)(e)?{i:M,r:e,k:t,f:!!n}:e:null);function kn(e,t=null,n=null,r=0,o=null,i=(e===on?0:1),s=!1,l=!1){const c={__v_isVNode:!0,__v_skip:!0,type:e,props:t,key:t&&wn(t),ref:t&&On(t),scopeId:$,slotScopeIds:null,children:n,component:null,suspense:null,ssContent:null,ssFallback:null,dirs:null,transition:null,el:null,anchor:null,target:null,targetStart:null,targetAnchor:null,staticCount:0,shapeFlag:i,patchFlag:r,dynamicProps:o,dynamicChildren:null,appContext:null,ctx:M};return l?(Pn(c,n),128&i&&e.normalize(c)):n&&(c.shapeFlag|=(0,a.Kg)(n)?8:16),hn>0&&!s&&dn&&(c.patchFlag>0||6&i)&&32!==c.patchFlag&&dn.push(c),c}const xn=function(e,t=null,n=null,o=0,i=null,s=!1){if(e&&e!==Ne||(e=ln),bn(e)){const r=Sn(e,t,!0);return n&&Pn(r,n),hn>0&&!s&&dn&&(6&r.shapeFlag?dn[dn.indexOf(e)]=r:dn.push(r)),r.patchFlag=-2,r}if(l=e,(0,a.Tn)(l)&&"__vccOpts"in l&&(e=e.__vccOpts),t){t=En(t);let{class:e,style:n}=t;e&&!(0,a.Kg)(e)&&(t.class=(0,a.C4)(e)),(0,a.Gv)(n)&&((0,r.ju)(n)&&!(0,a.cy)(n)&&(n=(0,a.X$)({},n)),t.style=(0,a.Tr)(n))}var l;return kn(e,t,n,o,i,(0,a.Kg)(e)?1:an(e)?128:Q(e)?64:(0,a.Gv)(e)?4:(0,a.Tn)(e)?2:0,s,!0)};function En(e){return e?(0,r.ju)(e)||yt(e)?(0,a.X$)({},e):e:null}function Sn(e,t,n=!1,r=!1){const{props:o,ref:i,patchFlag:s,children:l,transition:c}=e,u=t?Dn(o||{},t):o,d={__v_isVNode:!0,__v_skip:!0,type:e.type,props:u,key:u&&wn(u),ref:t&&t.ref?n&&i?(0,a.cy)(i)?i.concat(On(t)):[i,On(t)]:On(t):i,scopeId:e.scopeId,slotScopeIds:e.slotScopeIds,children:l,target:e.target,targetStart:e.targetStart,targetAnchor:e.targetAnchor,staticCount:e.staticCount,shapeFlag:e.shapeFlag,patchFlag:t&&e.type!==on?-1===s?16:16|s:s,dynamicProps:e.dynamicProps,dynamicChildren:e.dynamicChildren,appContext:e.appContext,dirs:e.dirs,transition:c,component:e.component,suspense:e.suspense,ssContent:e.ssContent&&Sn(e.ssContent),ssFallback:e.ssFallback&&Sn(e.ssFallback),el:e.el,anchor:e.anchor,ctx:e.ctx,ce:e.ce};return c&&r&&ce(d,c.clone(d)),d}function _n(e=" ",t=0){return xn(sn,null,e,t)}function Cn(e,t){const n=xn(cn,null,e);return n.staticCount=t,n}function Tn(e="",t=!1){return t?(pn(),vn(ln,null,e)):xn(ln,null,e)}function An(e){return null==e||"boolean"==typeof e?xn(ln):(0,a.cy)(e)?xn(on,null,e.slice()):bn(e)?Rn(e):xn(sn,null,String(e))}function Rn(e){return null===e.el&&-1!==e.patchFlag||e.memo?e:Sn(e)}function Pn(e,t){let n=0;const{shapeFlag:r}=e;if(null==t)t=null;else if((0,a.cy)(t))n=16;else if("object"==typeof t){if(65&r){const n=t.default;return void(n&&(n._c&&(n._d=!1),Pn(e,n()),n._c&&(n._d=!0)))}{n=32;const r=t._;r||yt(t)?3===r&&M&&(1===M.slots._?t._=1:(t._=2,e.patchFlag|=1024)):t._ctx=M}}else(0,a.Tn)(t)?(t={default:t,_ctx:M},n=32):(t=String(t),64&r?(n=16,t=[_n(t)]):n=8);e.children=t,e.shapeFlag|=n}function Dn(...e){const t={};for(let n=0;n<e.length;n++){const r=e[n];for(const e in r)if("class"===e)t.class!==r.class&&(t.class=(0,a.C4)([t.class,r.class]));else if("style"===e)t.style=(0,a.Tr)([t.style,r.style]);else if((0,a.Mp)(e)){const n=t[e],o=r[e];!o||n===o||(0,a.cy)(n)&&n.includes(o)||(t[e]=n?[].concat(n,o):o)}else""!==e&&(t[e]=r[e])}return t}function Mn(e,t,n,r=null){i(e,t,7,[n,r])}const $n=dt();let Nn=0;let In=null;const Ln=()=>In||M;let Bn,Fn;{const e=(0,a.We)(),t=(t,n)=>{let r;return(r=e[t])||(r=e[t]=[]),r.push(n),e=>{r.length>1?r.forEach((t=>t(e))):r[0](e)}};Bn=t("__VUE_INSTANCE_SETTERS__",(e=>In=e)),Fn=t("__VUE_SSR_SETTERS__",(e=>Xn=e))}const Qn=e=>{const t=In;return Bn(e),e.scope.on(),()=>{e.scope.off(),Bn(t)}},Un=()=>{In&&In.scope.off(),Bn(null)};function zn(e){return 4&e.vnode.shapeFlag}let jn,Wn,Xn=!1;function qn(e,t,n){(0,a.Tn)(t)?e.type.__ssrInlineRender?e.ssrRender=t:e.render=t:(0,a.Gv)(t)&&(__VUE_PROD_DEVTOOLS__&&(e.devtoolsRawSetupState=t),e.setupState=(0,r.Pr)(t)),Vn(e,n)}function Vn(e,t,n){const o=e.type;if(!e.render){if(!t&&jn&&!o.render){const t=o.template||__VUE_OPTIONS_API__&&rt(e).template;if(t){const{isCustomElement:n,compilerOptions:r}=e.appContext.config,{delimiters:i,compilerOptions:s}=o,l=(0,a.X$)((0,a.X$)({isCustomElement:n,delimiters:i},r),s);o.render=jn(t,l)}}e.render=o.render||a.tE,Wn&&Wn(e)}if(__VUE_OPTIONS_API__){const t=Qn(e);(0,r.C4)();try{!function(e){const t=rt(e),n=e.proxy,o=e.ctx;et=!1,t.beforeCreate&&tt(t.beforeCreate,e,"bc");const{data:i,computed:s,methods:l,watch:c,provide:u,inject:d,created:p,beforeMount:h,mounted:f,beforeUpdate:m,updated:g,activated:v,deactivated:b,beforeDestroy:y,beforeUnmount:w,destroyed:O,unmounted:k,render:x,renderTracked:E,renderTriggered:S,errorCaptured:_,serverPrefetch:C,expose:T,inheritAttrs:A,components:R,directives:P,filters:D}=t;if(d&&function(e,t){(0,a.cy)(e)&&(e=st(e));for(const n in e){const o=e[n];let i;i=(0,a.Gv)(o)?"default"in o?gt(o.from||n,o.default,!0):gt(o.from||n):gt(o),(0,r.i9)(i)?Object.defineProperty(t,n,{enumerable:!0,configurable:!0,get:()=>i.value,set:e=>i.value=e}):t[n]=i}}(d,o),l)for(const e in l){const t=l[e];(0,a.Tn)(t)&&(o[e]=t.bind(n))}if(i){const t=i.call(n,n);(0,a.Gv)(t)&&(e.data=(0,r.Kh)(t))}if(et=!0,s)for(const e in s){const t=s[e],r=(0,a.Tn)(t)?t.bind(n,n):(0,a.Tn)(t.get)?t.get.bind(n,n):a.tE,i=!(0,a.Tn)(t)&&(0,a.Tn)(t.set)?t.set.bind(n):a.tE,l=Kn({get:r,set:i});Object.defineProperty(o,e,{enumerable:!0,configurable:!0,get:()=>l.value,set:e=>l.value=e})}if(c)for(const e in c)nt(c[e],o,n,e);if(u){const e=(0,a.Tn)(u)?u.call(n):u;Reflect.ownKeys(e).forEach((t=>{mt(t,e[t])}))}function M(e,t){(0,a.cy)(t)?t.forEach((t=>e(t.bind(n)))):t&&e(t.bind(n))}if(p&&tt(p,e,"c"),M(xe,h),M(Ee,f),M(Se,m),M(_e,g),M(ve,v),M(be,b),M(De,_),M(Pe,E),M(Re,S),M(Ce,w),M(Te,k),M(Ae,C),(0,a.cy)(T))if(T.length){const t=e.exposed||(e.exposed={});T.forEach((e=>{Object.defineProperty(t,e,{get:()=>n[e],set:t=>n[e]=t})}))}else e.exposed||(e.exposed={});x&&e.render===a.tE&&(e.render=x),null!=A&&(e.inheritAttrs=A),R&&(e.components=R),P&&(e.directives=P),C&&he(e)}(e)}finally{(0,r.bl)(),t()}}}const Hn={get:(e,t)=>((0,r.u4)(e,"get",""),e[t])};function Zn(e){return{attrs:new Proxy(e.attrs,Hn),slots:e.slots,emit:e.emit,expose:t=>{e.exposed=t||{}}}}function Gn(e){return e.exposed?e.exposeProxy||(e.exposeProxy=new Proxy((0,r.Pr)((0,r.IG)(e.exposed)),{get:(t,n)=>n in t?t[n]:n in Xe?Xe[n](e):void 0,has:(e,t)=>t in e||t in Xe})):e.proxy}const Kn=(e,t)=>(0,r.EW)(e,t,Xn);function Yn(e,t,n){const r=arguments.length;return 2===r?(0,a.Gv)(t)&&!(0,a.cy)(t)?bn(t)?xn(e,null,[t]):xn(e,t):xn(e,null,t):(r>3?n=Array.prototype.slice.call(arguments,2):3===r&&bn(n)&&(n=[n]),xn(e,t,n))}const Jn="3.5.13"},796:(e,t,n)=>{"use strict";function r(e,t,n){const r=t.split(".");return r.reduce(((e,t,a)=>(a===r.length-1&&(e[t]=n),e[t])),e),e}function a(e,t){return t.split(".").reduce(((e,t)=>e[t]),e)}n.d(t,{F:()=>r,L:()=>a})},805:(e,t,n)=>{"use strict";n.d(t,{O:()=>i,j:()=>s});var r,a=n(641),o=n(3605),i=((r=i||{})[r.None=1]="None",r[r.Focusable=2]="Focusable",r[r.Hidden=4]="Hidden",r);let s=(0,a.pM)({name:"Hidden",props:{as:{type:[Object,String],default:"div"},features:{type:Number,default:1}},setup:(e,{slots:t,attrs:n})=>()=>{var r;let{features:a,...i}=e,s={"aria-hidden":!(2&~a)||(null!=(r=i["aria-hidden"])?r:void 0),hidden:!(4&~a)||void 0,style:{position:"fixed",top:1,left:1,width:1,height:0,padding:0,margin:-1,overflow:"hidden",clip:"rect(0, 0, 0, 0)",whiteSpace:"nowrap",borderWidth:"0",...!(4&~a)&&!!(2&~a)&&{display:"none"}}};return(0,o.XX)({ourProps:s,theirProps:i,slot:{},attrs:n,slots:t,name:"Hidden"})}})},894:(e,t,n)=>{"use strict";n.r(t),n.d(t,{default:()=>L});var r=n(641),a=n(953),o=n(33),i=n(3751),s=n(1266),l=n(2667),c=n(8157),u=n(9351),d=n(3654),p=n(68),h=n(3858),f=n(4486),m=n(3966),g=n(1465),v=n(8082),b=n(9005),y=n(3395),w=n(2490),O=n(9135),k=n(6918);const x=["onClick"],E=["placeholder"],S=(0,r.pM)({__name:"EnvironmentColors",props:{activeColor:{},selector:{type:Boolean,default:!1}},emits:["select"],setup(e,{emit:t}){const n=e,l=t,c=(0,a.KR)(""),u=(0,a.KR)(null),d=(0,a.KR)(!1),p=(0,a.KR)(!1),h=[{color:"#8E8E8E"},{color:"#EF0006"},{color:"#EDBE20"},{color:"#069061"},{color:"#FB892C"},{color:"#0082D0"},{color:"#5203D1"},{color:"#FFC0CB"}],f=(0,r.EW)((()=>n.activeColor&&!h.some((e=>e.color===n.activeColor))||c.value?`background-color: ${n.activeColor||c.value};`:"background: linear-gradient(to right, rgb(235, 87, 87), rgb(242, 201, 76), rgb(76, 183, 130), rgb(78, 167, 252), rgb(250, 96, 122));")),m=()=>{d.value=!d.value,n.selector&&(p.value=!1),(0,r.dY)((()=>{u.value&&u.value.focus()}))};(0,r.wB)(c,(e=>{e&&!e.startsWith("#")&&(c.value=`#${e}`),d.value=!0}));const g=()=>{n.selector&&(p.value=!p.value)},v=e=>{l("select",e),n.selector&&(p.value=!1)};return(e,t)=>((0,r.uX)(),(0,r.CE)("div",null,[d.value?(0,r.Q3)("",!0):((0,r.uX)(),(0,r.CE)(r.FK,{key:0},[n.selector&&!p.value?((0,r.uX)(),(0,r.CE)("div",{key:0,class:(0,o.C4)(["flex cursor-pointer items-center justify-center rounded-full",n.selector?"h-4 w-4":"h-5 w-5"]),style:(0,o.Tr)({backgroundColor:e.activeColor}),onClick:g},[e.activeColor?((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.kv),{key:0,class:(0,o.C4)(["text-c-btn",n.selector&&"p-0.5"]),icon:"Checkmark",size:"xs"},null,8,["class"])):(0,r.Q3)("",!0)],6)):(0,r.Q3)("",!0),p.value||!n.selector?((0,r.uX)(),(0,r.CE)("div",{key:1,class:(0,o.C4)(["color-selector flex flex-row gap-1.5 items-center justify-between space-x-1",n.selector?"h-4":"min-h-10 min-w-[296px]"])},[((0,r.uX)(),(0,r.CE)(r.FK,null,(0,r.pI)(h,(t=>(0,r.Lk)("div",{key:t.color,class:(0,o.C4)(["flex cursor-pointer items-center justify-center rounded-full",n.selector?"h-4 w-4":"h-5 w-5"]),style:(0,o.Tr)({backgroundColor:t.color}),onClick:e=>v(t.color)},[e.activeColor!==t.color||c.value?(0,r.Q3)("",!0):((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.kv),{key:0,class:(0,o.C4)(["text-c-btn",n.selector&&"p-0.5"]),icon:"Checkmark",size:"xs"},null,8,["class"]))],14,x))),64)),t[2]||(t[2]=(0,r.Lk)("hr",{class:"w-0.5 h-5 border-ghost border-l"},null,-1)),(0,r.Lk)("label",{class:(0,o.C4)(["cursor-pointer flex flex-row justify-center gap-2 items-center rounded-full z-10",n.selector?"h-4 w-4":"h-5 w-5"]),style:(0,o.Tr)(f.value),onClick:m},[d.value||e.activeColor!==c.value&&(!e.activeColor||h.some((t=>t.color===e.activeColor)))?(0,r.Q3)("",!0):((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.kv),{key:0,class:"text-c-btn",icon:"Checkmark",size:"xs"}))],6)],2)):(0,r.Q3)("",!0)],64)),d.value?((0,r.uX)(),(0,r.CE)("div",{key:1,class:(0,o.C4)(["color-selector flex flex-1 gap-2 items-center rounded",n.selector?"h-4":"min-h-10"])},[(0,r.Lk)("span",{class:(0,o.C4)(["absolute border border-dashed rounded-full",n.selector?"h-4 w-4":"h-5 w-5"])},null,2),(0,r.Lk)("span",{class:(0,o.C4)(["rounded-full z-[1]",n.selector?"h-4 w-4":"h-5 w-5"]),style:(0,o.Tr)(f.value)},null,6),(0,r.bo)((0,r.Lk)("input",{ref_key:"customColorInputRef",ref:u,"onUpdate:modelValue":t[0]||(t[0]=e=>c.value=e),class:"border-transparent flex-1 outline-none w-full text-sm",placeholder:e.activeColor||"#000000",type:"text",onInput:t[1]||(t[1]=e=>v(c.value))},null,40,E),[[i.Jo,c.value]]),(0,r.Lk)("button",{class:"text-c-3 hover:bg-b-2 p-1.5 rounded-lg",type:"button",onClick:m},[(0,r.bF)((0,a.R1)(s.kv),{icon:"Checkmark",size:"xs"})])],2)):(0,r.Q3)("",!0)]))}}),_={class:"flex flex-col gap-4"},C=(0,r.pM)({__name:"EnvironmentColorModal",props:{state:{},selectedColor:{}},emits:["cancel","submit"],setup(e,{emit:t}){const n=e,o=t,i=(0,a.KR)(""),l=e=>{i.value=e},c=()=>{o("submit",i.value),i.value=""};return(e,t)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.uT),{size:"xxs",state:e.state,title:"Edit Environment Color"},{default:(0,r.k6)((()=>[(0,r.Lk)("div",_,[(0,r.bF)(S,{activeColor:i.value||n.selectedColor,class:"p-1 w-full",onSelect:l},null,8,["activeColor"]),(0,r.bF)(k.A,{onCancel:t[0]||(t[0]=e=>o("cancel")),onSubmit:c})])])),_:1},8,["state"]))}});var T=n(5411),A=n(4445),R=n(8309);const P={class:"flex items-start gap-2"},D=(0,r.pM)({__name:"EnvironmentModal",props:{state:{},activeWorkspaceCollections:{},collectionId:{}},emits:["cancel","submit"],setup(e,{emit:t}){const n=e,i=t,{events:l}=(0,R.hA)(),u=(0,a.KR)(""),d=(0,a.KR)("#8E8E8E"),p=(0,r.EW)((()=>[...n.activeWorkspaceCollections.filter((e=>{var t;return"Drafts"!==(null==(t=e.info)?void 0:t.title)})).map((e=>{var t;return{id:e.uid,label:(null==(t=e.info)?void 0:t.title)??"Untitled Collection"}}))])),h=(0,a.KR)(p.value.find((e=>e.id===n.collectionId))),{toast:f}=(0,c.Yf)(),m=e=>{d.value=e};(0,r.wB)((()=>n.state.open),(e=>{e&&(u.value="",d.value="#8E8E8E",n.collectionId?h.value=p.value.find((e=>e.id===n.collectionId)):h.value=void 0)}));const g=()=>{var e,t,n,r;null!=(e=h.value)&&e.id?i("submit",{name:u.value,color:d.value,type:"global"===(null==(t=h.value)?void 0:t.id)?"global":"collection",collectionId:"global"!==(null==(n=h.value)?void 0:n.id)?null==(r=h.value)?void 0:r.id:void 0}):f("Please select a collection before adding an environment.","error")},v=()=>{n.state.hide(),l.commandPalette.emit({commandName:"Create Collection"})};return(e,t)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.uT),{bodyClass:"border-t-0 rounded-t-lg",size:"xs",state:e.state},{default:(0,r.k6)((()=>[(0,r.bF)(T.A,{disabled:!h.value,onCancel:t[2]||(t[2]=e=>i("cancel")),onSubmit:g},{options:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(s.cy),{modelValue:h.value,"onUpdate:modelValue":t[1]||(t[1]=e=>h.value=e),options:p.value,placeholder:"Select Type"},{default:(0,r.k6)((()=>[p.value.length>0?((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.RO),{key:0,class:"hover:bg-b-2 max-h-8 w-fit justify-between gap-1 p-2 text-xs",variant:"outlined"},{default:(0,r.k6)((()=>[(0,r.Lk)("span",{class:(0,o.C4)(h.value?"text-c-1":"text-c-3")},(0,o.v_)(h.value?h.value.label:"Select Collection"),3),(0,r.bF)((0,a.R1)(s.kv),{class:"text-c-3",icon:"ChevronDown",size:"xs"})])),_:1})):((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.RO),{key:1,class:"hover:bg-b-2 max-h-8 justify-between gap-1 p-2 text-xs",variant:"outlined",onClick:v},{default:(0,r.k6)((()=>t[3]||(t[3]=[(0,r.Lk)("span",{class:"text-c-1"},"Create Collection",-1)]))),_:1}))])),_:1},8,["modelValue","options"])])),submit:(0,r.k6)((()=>t[4]||(t[4]=[(0,r.eW)(" Add Environment ")]))),default:(0,r.k6)((()=>[(0,r.Lk)("div",P,[(0,r.bF)(S,{activeColor:d.value,class:"peer",selector:"",onSelect:m},null,8,["activeColor"]),(0,r.bF)(A.A,{modelValue:u.value,"onUpdate:modelValue":t[0]||(t[0]=e=>u.value=e),class:"-mt-[.5px] !p-0 peer-has-[.color-selector]:hidden",placeholder:"Environment name"},null,8,["modelValue"])])])),_:1},8,["disabled"])])),_:1},8,["state"]))}});var M=n(6099);const $={class:"flex-1"},N=["onClick"],I={class:"flex h-5 max-w-[14px] items-center justify-center"},L=(0,r.pM)({__name:"Environment",setup(e){const t=(0,u.rd)(),n=(0,u.lq)(),{activeWorkspace:k,activeEnvironment:x,activeWorkspaceCollections:E,activeEnvVariables:S}=(0,O.JC)(),{events:_,workspaceMutators:T,collectionMutators:A}=(0,R.hA)(),{collapsedSidebarFolders:P,toggleSidebarFolder:L}=(0,M.c)(),B=(0,s.hS)(),F=(0,s.hS)(),Q=(0,s.hS)(),U=(0,a.KR)(null),z=(0,a.KR)("default"),j=(0,a.KR)(""),W=(0,a.KR)(void 0),X=(0,a.KR)(void 0),q=(0,a.KR)(void 0),{toast:V}=(0,c.Yf)();function H(e,t,n){var r;e&&(t.uid===n?V(`Environment name already used in ${null==(r=t.info)?void 0:r.title}`,"error"):V("Environment name already used in another collection","error"))}function Z(e){E.value.some((t=>{const n=Object.keys(t["x-scalar-environments"]||{}).includes(e.name);return H(n,t,e.collectionId),n}))||(e.collectionId&&(A.addEnvironment(e.name,{variables:{},color:e.color},e.collectionId),P[e.collectionId]||L(e.collectionId),t.push({name:"environment.collection",params:{[w.TU.Collection]:e.collectionId,[w.TU.Environment]:e.name}})),F.hide())}function G(e){var t,n;if(x){const r=(e=>JSON.parse(e))(e);if("default"===z.value)T.edit(null==(t=k.value)?void 0:t.uid,"environments",r);else{const e=E.value.find((e=>{var t;return null==(t=e["x-scalar-environments"])?void 0:t[z.value??""]}));if(null!=(n=null==e?void 0:e["x-scalar-environments"])&&n[z.value??""]){const t=e["x-scalar-environments"][z.value??""];t&&(t.variables=r,A.edit(e.uid,"x-scalar-environments",e["x-scalar-environments"]))}}}}const K=e=>{W.value=e,F.show()},Y=e=>{const t=U.value;"string"==typeof t&&(E.value.some((e=>{var n;return null==(n=e["x-scalar-environments"])?void 0:n[t]}))&&E.value.forEach((n=>{var r;null!=(r=n["x-scalar-environments"])&&r[t]&&(n["x-scalar-environments"][t].color=e,A.edit(n.uid,"x-scalar-environments",n["x-scalar-environments"]))})),B.hide())},J=()=>{var e,t,n,r;return"default"===z.value?JSON.stringify(null==(e=k.value)?void 0:e.environments,null,2):JSON.stringify(null==(r=null==(n=null==(t=E.value.find((e=>{var t;return null==(t=e["x-scalar-environments"])?void 0:t[z.value??""]})))?void 0:t["x-scalar-environments"])?void 0:n[z.value??""])?void 0:r.variables,null,2)},ee=e=>P[e],te=e=>{null!=e&&e.createNew&&"environment"===n.name&&K()};function ne(){X.value=void 0,W.value=void 0,q.value=void 0,Q.hide()}function re(e){E.value.some((t=>{const n=Object.keys(t["x-scalar-environments"]||{}).includes(e);return H(n,t,W.value),n}))||(e&&"default"!==X.value&&E.value.forEach((t=>{var n;if(null!=(n=t["x-scalar-environments"])&&n[X.value??""]){const n=t["x-scalar-environments"][X.value??""];n&&(delete t["x-scalar-environments"][X.value??""],t["x-scalar-environments"][e]=n,A.edit(t.uid,"x-scalar-environments",t["x-scalar-environments"]))}})),e&&z.value===X.value&&(z.value=e),X.value=void 0,W.value=void 0,q.value=void 0,Q.hide())}return(0,r.wB)((()=>[n.params[w.TU.Collection],n.params[w.TU.Environment]]),(([e,t])=>{z.value=e?t:"default"})),(0,r.sV)((()=>{z.value=n.params[w.TU.Environment]||"default",_.hotKeys.on(te);const e=n.params[w.TU.Collection];e&&!P[e]&&L(e)})),(0,r.xo)((()=>_.hotKeys.off(te))),(e,n)=>((0,r.uX)(),(0,r.Wv)(v.A,null,{default:(0,r.k6)((()=>[(0,r.bF)(h.A,{title:"Collections"},{content:(0,r.k6)((()=>[(0,r.Lk)("div",$,[(0,r.bF)(m.A,null,{default:(0,r.k6)((()=>[((0,r.uX)(),(0,r.Wv)(g.A,{key:"default",class:"text-xs",isCopyable:!1,type:"environment",variable:{name:"Global Environment",uid:"default",icon:"Globe",isDefault:!0}})),((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)((0,a.R1)(E),(e=>{var c;return(0,r.uX)(),(0,r.CE)("li",{key:e.uid,class:"gap-1/2 flex flex-col"},[(0,r.Lk)("button",{class:"hover:bg-b-2 group flex w-full items-center gap-1.5 break-words rounded p-1.5 text-left text-sm font-medium",type:"button",onClick:t=>(0,a.R1)(L)(e.uid)},[(0,r.Lk)("span",I,[(0,r.bF)((0,a.R1)(l.$n),{class:"text-sidebar-c-2 size-3.5 min-w-3.5 stroke-2 group-hover:hidden",src:e["x-scalar-icon"]||"interface-content-folder"},null,8,["src"]),(0,r.Lk)("div",{class:(0,o.C4)({"rotate-90":(0,a.R1)(P)[e.uid]})},[(0,r.bF)((0,a.R1)(s.kv),{class:"text-c-3 hover:text-c-1 hidden text-sm group-hover:block",icon:"ChevronRight",size:"md"})],2)]),(0,r.eW)(" "+(0,o.v_)((null==(c=e.info)?void 0:c.title)??""),1)],8,N),(0,r.bo)((0,r.Lk)("div",{class:(0,o.C4)({"before:bg-border before:z-1 relative mb-[.5px] before:pointer-events-none before:absolute before:left-3 before:top-0 before:h-[calc(100%_+_.5px)] before:w-[.5px] last:mb-0 last:before:h-full":Object.keys(e["x-scalar-environments"]||{}).length>0})},[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(e["x-scalar-environments"],((n,a)=>((0,r.uX)(),(0,r.Wv)(g.A,{key:a,class:"text-xs [&>a]:pl-5",collectionId:e.uid,isCopyable:!1,isDeletable:!0,isRenameable:!0,type:"environment",variable:{name:a,uid:a,color:n.color??"#8E8E8E",isDefault:!1},warningMessage:"Are you sure you want to delete this environment?",onClick:(0,i.D$)((n=>((e,n,r)=>{var a,o;const i=r?{name:"environment.collection",params:{[w.TU.Workspace]:null==(a=k.value)?void 0:a.uid,[w.TU.Collection]:r,[w.TU.Environment]:n}}:{name:"environment.default",params:{[w.TU.Workspace]:null==(o=k.value)?void 0:o.uid,[w.TU.Environment]:n}};e.metaKey?window.open(t.resolve(i).href,"_blank"):t.push(i)})(n,a,e.uid)),["prevent"]),onColorModal:e=>(e=>{var t,n,r;U.value=e,j.value=(null==(r=null==(n=null==(t=E.value.find((t=>{var n;return null==(n=t["x-scalar-environments"])?void 0:n[e]})))?void 0:t["x-scalar-environments"])?void 0:n[e])?void 0:r.color)??"",B.show()})(a),onDelete:e=>function(e){E.value.forEach((t=>{A.removeEnvironment(e,t.uid)}));const n=E.value.flatMap((e=>Object.keys(e["x-scalar-environments"]||{})));if(n.length>0){const e=n[n.length-1];if(!e)return;const r=E.value.find((t=>Object.keys(t["x-scalar-environments"]||{}).includes(e)));z.value=e,t.push({name:"environment.collection",params:{collectionId:null==r?void 0:r.uid,environmentId:e}}),r&&!P[r.uid]&&L(r.uid)}else z.value="default",t.push({name:"environment",params:{environment:"default"}})}(a),onRename:t=>((e,t)=>{X.value=e,W.value=t,q.value=e,Q.show()})(a,e.uid)},null,8,["collectionId","variable","onClick","onColorModal","onDelete","onRename"])))),128)),0===Object.keys(e["x-scalar-environments"]||{}).length?((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.RO),{key:0,class:"text-c-1 hover:bg-b-2 flex h-8 w-full justify-start gap-1.5 py-0 pl-6 text-xs",variant:"ghost",onClick:t=>K(e.uid)},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(s.kv),{icon:"Add",size:"sm"}),n[2]||(n[2]=(0,r.Lk)("span",null,"Add Environment",-1))])),_:2},1032,["onClick"])):(0,r.Q3)("",!0)],2),[[i.aG,ee(e.uid)]])])})),128))])),_:1})])])),button:(0,r.k6)((()=>[(0,r.bF)(f.A,{click:K,hotkey:"N"},{title:(0,r.k6)((()=>n[3]||(n[3]=[(0,r.eW)(" Add Environment ")]))),_:1})])),_:1}),(0,r.bF)(b.A,{class:"flex-1"},{default:(0,r.k6)((()=>[(0,r.bF)(y.A,null,(0,r.eX)({default:(0,r.k6)((()=>[z.value&&(0,a.R1)(k)?((0,r.uX)(),(0,r.Wv)(d.A,{key:0,class:"border-t py-2 pl-px pr-2 md:px-4",envVariables:(0,a.R1)(S),environment:(0,a.R1)(x),isCopyable:"",language:"json",lineNumbers:"",lint:"",modelValue:J(),workspace:(0,a.R1)(k),"onUpdate:modelValue":G},null,8,["envVariables","environment","modelValue","workspace"])):(0,r.Q3)("",!0)])),_:2},[z.value?{name:"title",fn:(0,r.k6)((()=>[(0,r.Lk)("span",null,(0,o.v_)("default"===z.value?"Global Environment":z.value),1)])),key:"0"}:void 0]),1024)])),_:1}),(0,r.bF)(C,{selectedColor:j.value,state:(0,a.R1)(B),onCancel:n[0]||(n[0]=e=>(0,a.R1)(B).hide()),onSubmit:Y},null,8,["selectedColor","state"]),(0,r.bF)(D,{activeWorkspaceCollections:(0,a.R1)(E),collectionId:W.value,state:(0,a.R1)(F),onCancel:n[1]||(n[1]=e=>(0,a.R1)(F).hide()),onSubmit:Z},null,8,["activeWorkspaceCollections","collectionId","state"]),(0,r.bF)((0,a.R1)(s.uT),{size:"xxs",state:(0,a.R1)(Q),title:`Edit ${X.value}`},{default:(0,r.k6)((()=>[(0,r.bF)(p.A,{name:q.value??"",onClose:ne,onEdit:re},null,8,["name"])])),_:1},8,["state","title"])])),_:1}))}})},953:(e,t,n)=>{"use strict";n.d(t,{BA:()=>Ie,C4:()=>_,EW:()=>Ve,Gc:()=>ve,IG:()=>_e,IJ:()=>Pe,KR:()=>Re,Kh:()=>ge,Pr:()=>Be,QW:()=>Ue,R1:()=>Ne,X2:()=>d,bl:()=>C,fE:()=>xe,g8:()=>Oe,hV:()=>Ye,hZ:()=>B,i9:()=>Ae,jr:()=>c,ju:()=>Ee,lJ:()=>Ce,lW:()=>We,mu:()=>$e,nD:()=>ye,o5:()=>l,qA:()=>Q,rY:()=>Qe,tB:()=>be,u4:()=>L,uY:()=>s,ux:()=>Se,wB:()=>Ke,yC:()=>i});var r=n(33);let a,o;class i{constructor(e=!1){this.detached=e,this._active=!0,this.effects=[],this.cleanups=[],this._isPaused=!1,this.parent=a,!e&&a&&(this.index=(a.scopes||(a.scopes=[])).push(this)-1)}get active(){return this._active}pause(){if(this._active){let e,t;if(this._isPaused=!0,this.scopes)for(e=0,t=this.scopes.length;e<t;e++)this.scopes[e].pause();for(e=0,t=this.effects.length;e<t;e++)this.effects[e].pause()}}resume(){if(this._active&&this._isPaused){let e,t;if(this._isPaused=!1,this.scopes)for(e=0,t=this.scopes.length;e<t;e++)this.scopes[e].resume();for(e=0,t=this.effects.length;e<t;e++)this.effects[e].resume()}}run(e){if(this._active){const t=a;try{return a=this,e()}finally{a=t}}}on(){a=this}off(){a=this.parent}stop(e){if(this._active){let t,n;for(this._active=!1,t=0,n=this.effects.length;t<n;t++)this.effects[t].stop();for(this.effects.length=0,t=0,n=this.cleanups.length;t<n;t++)this.cleanups[t]();if(this.cleanups.length=0,this.scopes){for(t=0,n=this.scopes.length;t<n;t++)this.scopes[t].stop(!0);this.scopes.length=0}if(!this.detached&&this.parent&&!e){const e=this.parent.scopes.pop();e&&e!==this&&(this.parent.scopes[this.index]=e,e.index=this.index)}this.parent=void 0}}}function s(e){return new i(e)}function l(){return a}function c(e,t=!1){a&&a.cleanups.push(e)}const u=new WeakSet;class d{constructor(e){this.fn=e,this.deps=void 0,this.depsTail=void 0,this.flags=5,this.next=void 0,this.cleanup=void 0,this.scheduler=void 0,a&&a.active&&a.effects.push(this)}pause(){this.flags|=64}resume(){64&this.flags&&(this.flags&=-65,u.has(this)&&(u.delete(this),this.trigger()))}notify(){2&this.flags&&!(32&this.flags)||8&this.flags||m(this)}run(){if(!(1&this.flags))return this.fn();this.flags|=2,T(this),b(this);const e=o,t=E;o=this,E=!0;try{return this.fn()}finally{y(this),o=e,E=t,this.flags&=-3}}stop(){if(1&this.flags){for(let e=this.deps;e;e=e.nextDep)k(e);this.deps=this.depsTail=void 0,T(this),this.onStop&&this.onStop(),this.flags&=-2}}trigger(){64&this.flags?u.add(this):this.scheduler?this.scheduler():this.runIfDirty()}runIfDirty(){w(this)&&this.run()}get dirty(){return w(this)}}let p,h,f=0;function m(e,t=!1){if(e.flags|=8,t)return e.next=h,void(h=e);e.next=p,p=e}function g(){f++}function v(){if(--f>0)return;if(h){let e=h;for(h=void 0;e;){const t=e.next;e.next=void 0,e.flags&=-9,e=t}}let e;for(;p;){let t=p;for(p=void 0;t;){const n=t.next;if(t.next=void 0,t.flags&=-9,1&t.flags)try{t.trigger()}catch(t){e||(e=t)}t=n}}if(e)throw e}function b(e){for(let t=e.deps;t;t=t.nextDep)t.version=-1,t.prevActiveLink=t.dep.activeLink,t.dep.activeLink=t}function y(e){let t,n=e.depsTail,r=n;for(;r;){const e=r.prevDep;-1===r.version?(r===n&&(n=e),k(r),x(r)):t=r,r.dep.activeLink=r.prevActiveLink,r.prevActiveLink=void 0,r=e}e.deps=t,e.depsTail=n}function w(e){for(let t=e.deps;t;t=t.nextDep)if(t.dep.version!==t.version||t.dep.computed&&(O(t.dep.computed)||t.dep.version!==t.version))return!0;return!!e._dirty}function O(e){if(4&e.flags&&!(16&e.flags))return;if(e.flags&=-17,e.globalVersion===A)return;e.globalVersion=A;const t=e.dep;if(e.flags|=2,t.version>0&&!e.isSSR&&e.deps&&!w(e))return void(e.flags&=-3);const n=o,a=E;o=e,E=!0;try{b(e);const n=e.fn(e._value);(0===t.version||(0,r.$H)(n,e._value))&&(e._value=n,t.version++)}catch(e){throw t.version++,e}finally{o=n,E=a,y(e),e.flags&=-3}}function k(e,t=!1){const{dep:n,prevSub:r,nextSub:a}=e;if(r&&(r.nextSub=a,e.prevSub=void 0),a&&(a.prevSub=r,e.nextSub=void 0),n.subs===e&&(n.subs=r,!r&&n.computed)){n.computed.flags&=-5;for(let e=n.computed.deps;e;e=e.nextDep)k(e,!0)}t||--n.sc||!n.map||n.map.delete(n.key)}function x(e){const{prevDep:t,nextDep:n}=e;t&&(t.nextDep=n,e.prevDep=void 0),n&&(n.prevDep=t,e.nextDep=void 0)}let E=!0;const S=[];function _(){S.push(E),E=!1}function C(){const e=S.pop();E=void 0===e||e}function T(e){const{cleanup:t}=e;if(e.cleanup=void 0,t){const e=o;o=void 0;try{t()}finally{o=e}}}let A=0;class R{constructor(e,t){this.sub=e,this.dep=t,this.version=t.version,this.nextDep=this.prevDep=this.nextSub=this.prevSub=this.prevActiveLink=void 0}}class P{constructor(e){this.computed=e,this.version=0,this.activeLink=void 0,this.subs=void 0,this.map=void 0,this.key=void 0,this.sc=0}track(e){if(!o||!E||o===this.computed)return;let t=this.activeLink;if(void 0===t||t.sub!==o)t=this.activeLink=new R(o,this),o.deps?(t.prevDep=o.depsTail,o.depsTail.nextDep=t,o.depsTail=t):o.deps=o.depsTail=t,D(t);else if(-1===t.version&&(t.version=this.version,t.nextDep)){const e=t.nextDep;e.prevDep=t.prevDep,t.prevDep&&(t.prevDep.nextDep=e),t.prevDep=o.depsTail,t.nextDep=void 0,o.depsTail.nextDep=t,o.depsTail=t,o.deps===t&&(o.deps=e)}return t}trigger(e){this.version++,A++,this.notify(e)}notify(e){g();try{for(let e=this.subs;e;e=e.prevSub)e.sub.notify()&&e.sub.dep.notify()}finally{v()}}}function D(e){if(e.dep.sc++,4&e.sub.flags){const t=e.dep.computed;if(t&&!e.dep.subs){t.flags|=20;for(let e=t.deps;e;e=e.nextDep)D(e)}const n=e.dep.subs;n!==e&&(e.prevSub=n,n&&(n.nextSub=e)),e.dep.subs=e}}const M=new WeakMap,$=Symbol(""),N=Symbol(""),I=Symbol("");function L(e,t,n){if(E&&o){let t=M.get(e);t||M.set(e,t=new Map);let r=t.get(n);r||(t.set(n,r=new P),r.map=t,r.key=n),r.track()}}function B(e,t,n,a,o,i){const s=M.get(e);if(!s)return void A++;const l=e=>{e&&e.trigger()};if(g(),"clear"===t)s.forEach(l);else{const o=(0,r.cy)(e),i=o&&(0,r.yI)(n);if(o&&"length"===n){const e=Number(a);s.forEach(((t,n)=>{("length"===n||n===I||!(0,r.Bm)(n)&&n>=e)&&l(t)}))}else switch((void 0!==n||s.has(void 0))&&l(s.get(n)),i&&l(s.get(I)),t){case"add":o?i&&l(s.get("length")):(l(s.get($)),(0,r.CE)(e)&&l(s.get(N)));break;case"delete":o||(l(s.get($)),(0,r.CE)(e)&&l(s.get(N)));break;case"set":(0,r.CE)(e)&&l(s.get($))}}v()}function F(e){const t=Se(e);return t===e?t:(L(t,0,I),xe(e)?t:t.map(Ce))}function Q(e){return L(e=Se(e),0,I),e}const U={__proto__:null,[Symbol.iterator](){return z(this,Symbol.iterator,Ce)},concat(...e){return F(this).concat(...e.map((e=>(0,r.cy)(e)?F(e):e)))},entries(){return z(this,"entries",(e=>(e[1]=Ce(e[1]),e)))},every(e,t){return W(this,"every",e,t,void 0,arguments)},filter(e,t){return W(this,"filter",e,t,(e=>e.map(Ce)),arguments)},find(e,t){return W(this,"find",e,t,Ce,arguments)},findIndex(e,t){return W(this,"findIndex",e,t,void 0,arguments)},findLast(e,t){return W(this,"findLast",e,t,Ce,arguments)},findLastIndex(e,t){return W(this,"findLastIndex",e,t,void 0,arguments)},forEach(e,t){return W(this,"forEach",e,t,void 0,arguments)},includes(...e){return q(this,"includes",e)},indexOf(...e){return q(this,"indexOf",e)},join(e){return F(this).join(e)},lastIndexOf(...e){return q(this,"lastIndexOf",e)},map(e,t){return W(this,"map",e,t,void 0,arguments)},pop(){return V(this,"pop")},push(...e){return V(this,"push",e)},reduce(e,...t){return X(this,"reduce",e,t)},reduceRight(e,...t){return X(this,"reduceRight",e,t)},shift(){return V(this,"shift")},some(e,t){return W(this,"some",e,t,void 0,arguments)},splice(...e){return V(this,"splice",e)},toReversed(){return F(this).toReversed()},toSorted(e){return F(this).toSorted(e)},toSpliced(...e){return F(this).toSpliced(...e)},unshift(...e){return V(this,"unshift",e)},values(){return z(this,"values",Ce)}};function z(e,t,n){const r=Q(e),a=r[t]();return r===e||xe(e)||(a._next=a.next,a.next=()=>{const e=a._next();return e.value&&(e.value=n(e.value)),e}),a}const j=Array.prototype;function W(e,t,n,r,a,o){const i=Q(e),s=i!==e&&!xe(e),l=i[t];if(l!==j[t]){const t=l.apply(e,o);return s?Ce(t):t}let c=n;i!==e&&(s?c=function(t,r){return n.call(this,Ce(t),r,e)}:n.length>2&&(c=function(t,r){return n.call(this,t,r,e)}));const u=l.call(i,c,r);return s&&a?a(u):u}function X(e,t,n,r){const a=Q(e);let o=n;return a!==e&&(xe(e)?n.length>3&&(o=function(t,r,a){return n.call(this,t,r,a,e)}):o=function(t,r,a){return n.call(this,t,Ce(r),a,e)}),a[t](o,...r)}function q(e,t,n){const r=Se(e);L(r,0,I);const a=r[t](...n);return-1!==a&&!1!==a||!Ee(n[0])?a:(n[0]=Se(n[0]),r[t](...n))}function V(e,t,n=[]){_(),g();const r=Se(e)[t].apply(e,n);return v(),C(),r}const H=(0,r.pD)("__proto__,__v_isRef,__isVue"),Z=new Set(Object.getOwnPropertyNames(Symbol).filter((e=>"arguments"!==e&&"caller"!==e)).map((e=>Symbol[e])).filter(r.Bm));function G(e){(0,r.Bm)(e)||(e=String(e));const t=Se(this);return L(t,0,e),t.hasOwnProperty(e)}class K{constructor(e=!1,t=!1){this._isReadonly=e,this._isShallow=t}get(e,t,n){if("__v_skip"===t)return e.__v_skip;const a=this._isReadonly,o=this._isShallow;if("__v_isReactive"===t)return!a;if("__v_isReadonly"===t)return a;if("__v_isShallow"===t)return o;if("__v_raw"===t)return n===(a?o?me:fe:o?he:pe).get(e)||Object.getPrototypeOf(e)===Object.getPrototypeOf(n)?e:void 0;const i=(0,r.cy)(e);if(!a){let e;if(i&&(e=U[t]))return e;if("hasOwnProperty"===t)return G}const s=Reflect.get(e,t,Ae(e)?e:n);return((0,r.Bm)(t)?Z.has(t):H(t))?s:(a||L(e,0,t),o?s:Ae(s)?i&&(0,r.yI)(t)?s:s.value:(0,r.Gv)(s)?a?be(s):ge(s):s)}}class Y extends K{constructor(e=!1){super(!1,e)}set(e,t,n,a){let o=e[t];if(!this._isShallow){const t=ke(o);if(xe(n)||ke(n)||(o=Se(o),n=Se(n)),!(0,r.cy)(e)&&Ae(o)&&!Ae(n))return!t&&(o.value=n,!0)}const i=(0,r.cy)(e)&&(0,r.yI)(t)?Number(t)<e.length:(0,r.$3)(e,t),s=Reflect.set(e,t,n,Ae(e)?e:a);return e===Se(a)&&(i?(0,r.$H)(n,o)&&B(e,"set",t,n):B(e,"add",t,n)),s}deleteProperty(e,t){const n=(0,r.$3)(e,t),a=(e[t],Reflect.deleteProperty(e,t));return a&&n&&B(e,"delete",t,void 0),a}has(e,t){const n=Reflect.has(e,t);return(0,r.Bm)(t)&&Z.has(t)||L(e,0,t),n}ownKeys(e){return L(e,0,(0,r.cy)(e)?"length":$),Reflect.ownKeys(e)}}class J extends K{constructor(e=!1){super(!0,e)}set(e,t){return!0}deleteProperty(e,t){return!0}}const ee=new Y,te=new J,ne=new Y(!0),re=new J(!0),ae=e=>e,oe=e=>Reflect.getPrototypeOf(e);function ie(e){return function(...t){return"delete"!==e&&("clear"===e?void 0:this)}}function se(e,t){const n=function(e,t){const n={get(n){const a=this.__v_raw,o=Se(a),i=Se(n);e||((0,r.$H)(n,i)&&L(o,0,n),L(o,0,i));const{has:s}=oe(o),l=t?ae:e?Te:Ce;return s.call(o,n)?l(a.get(n)):s.call(o,i)?l(a.get(i)):void(a!==o&&a.get(n))},get size(){const t=this.__v_raw;return!e&&L(Se(t),0,$),Reflect.get(t,"size",t)},has(t){const n=this.__v_raw,a=Se(n),o=Se(t);return e||((0,r.$H)(t,o)&&L(a,0,t),L(a,0,o)),t===o?n.has(t):n.has(t)||n.has(o)},forEach(n,r){const a=this,o=a.__v_raw,i=Se(o),s=t?ae:e?Te:Ce;return!e&&L(i,0,$),o.forEach(((e,t)=>n.call(r,s(e),s(t),a)))}};return(0,r.X$)(n,e?{add:ie("add"),set:ie("set"),delete:ie("delete"),clear:ie("clear")}:{add(e){t||xe(e)||ke(e)||(e=Se(e));const n=Se(this);return oe(n).has.call(n,e)||(n.add(e),B(n,"add",e,e)),this},set(e,n){t||xe(n)||ke(n)||(n=Se(n));const a=Se(this),{has:o,get:i}=oe(a);let s=o.call(a,e);s||(e=Se(e),s=o.call(a,e));const l=i.call(a,e);return a.set(e,n),s?(0,r.$H)(n,l)&&B(a,"set",e,n):B(a,"add",e,n),this},delete(e){const t=Se(this),{has:n,get:r}=oe(t);let a=n.call(t,e);a||(e=Se(e),a=n.call(t,e)),r&&r.call(t,e);const o=t.delete(e);return a&&B(t,"delete",e,void 0),o},clear(){const e=Se(this),t=0!==e.size,n=e.clear();return t&&B(e,"clear",void 0,void 0),n}}),["keys","values","entries",Symbol.iterator].forEach((a=>{n[a]=function(e,t,n){return function(...a){const o=this.__v_raw,i=Se(o),s=(0,r.CE)(i),l="entries"===e||e===Symbol.iterator&&s,c="keys"===e&&s,u=o[e](...a),d=n?ae:t?Te:Ce;return!t&&L(i,0,c?N:$),{next(){const{value:e,done:t}=u.next();return t?{value:e,done:t}:{value:l?[d(e[0]),d(e[1])]:d(e),done:t}},[Symbol.iterator](){return this}}}}(a,e,t)})),n}(e,t);return(t,a,o)=>"__v_isReactive"===a?!e:"__v_isReadonly"===a?e:"__v_raw"===a?t:Reflect.get((0,r.$3)(n,a)&&a in t?n:t,a,o)}const le={get:se(!1,!1)},ce={get:se(!1,!0)},ue={get:se(!0,!1)},de={get:se(!0,!0)},pe=new WeakMap,he=new WeakMap,fe=new WeakMap,me=new WeakMap;function ge(e){return ke(e)?e:we(e,!1,ee,le,pe)}function ve(e){return we(e,!1,ne,ce,he)}function be(e){return we(e,!0,te,ue,fe)}function ye(e){return we(e,!0,re,de,me)}function we(e,t,n,a,o){if(!(0,r.Gv)(e))return e;if(e.__v_raw&&(!t||!e.__v_isReactive))return e;const i=o.get(e);if(i)return i;const s=(l=e).__v_skip||!Object.isExtensible(l)?0:function(e){switch(e){case"Object":case"Array":return 1;case"Map":case"Set":case"WeakMap":case"WeakSet":return 2;default:return 0}}((0,r.Zf)(l));var l;if(0===s)return e;const c=new Proxy(e,2===s?a:n);return o.set(e,c),c}function Oe(e){return ke(e)?Oe(e.__v_raw):!(!e||!e.__v_isReactive)}function ke(e){return!(!e||!e.__v_isReadonly)}function xe(e){return!(!e||!e.__v_isShallow)}function Ee(e){return!!e&&!!e.__v_raw}function Se(e){const t=e&&e.__v_raw;return t?Se(t):e}function _e(e){return!(0,r.$3)(e,"__v_skip")&&Object.isExtensible(e)&&(0,r.yQ)(e,"__v_skip",!0),e}const Ce=e=>(0,r.Gv)(e)?ge(e):e,Te=e=>(0,r.Gv)(e)?be(e):e;function Ae(e){return!!e&&!0===e.__v_isRef}function Re(e){return De(e,!1)}function Pe(e){return De(e,!0)}function De(e,t){return Ae(e)?e:new Me(e,t)}class Me{constructor(e,t){this.dep=new P,this.__v_isRef=!0,this.__v_isShallow=!1,this._rawValue=t?e:Se(e),this._value=t?e:Ce(e),this.__v_isShallow=t}get value(){return this.dep.track(),this._value}set value(e){const t=this._rawValue,n=this.__v_isShallow||xe(e)||ke(e);e=n?e:Se(e),(0,r.$H)(e,t)&&(this._rawValue=e,this._value=n?e:Ce(e),this.dep.trigger())}}function $e(e){e.dep&&e.dep.trigger()}function Ne(e){return Ae(e)?e.value:e}function Ie(e){return(0,r.Tn)(e)?e():Ne(e)}const Le={get:(e,t,n)=>"__v_raw"===t?e:Ne(Reflect.get(e,t,n)),set:(e,t,n,r)=>{const a=e[t];return Ae(a)&&!Ae(n)?(a.value=n,!0):Reflect.set(e,t,n,r)}};function Be(e){return Oe(e)?e:new Proxy(e,Le)}class Fe{constructor(e){this.__v_isRef=!0,this._value=void 0;const t=this.dep=new P,{get:n,set:r}=e(t.track.bind(t),t.trigger.bind(t));this._get=n,this._set=r}get value(){return this._value=this._get()}set value(e){this._set(e)}}function Qe(e){return new Fe(e)}function Ue(e){const t=(0,r.cy)(e)?new Array(e.length):{};for(const n in e)t[n]=Xe(e,n);return t}class ze{constructor(e,t,n){this._object=e,this._key=t,this._defaultValue=n,this.__v_isRef=!0,this._value=void 0}get value(){const e=this._object[this._key];return this._value=void 0===e?this._defaultValue:e}set value(e){this._object[this._key]=e}get dep(){return function(e,t){const n=M.get(e);return n&&n.get(t)}(Se(this._object),this._key)}}class je{constructor(e){this._getter=e,this.__v_isRef=!0,this.__v_isReadonly=!0,this._value=void 0}get value(){return this._value=this._getter()}}function We(e,t,n){return Ae(e)?e:(0,r.Tn)(e)?new je(e):(0,r.Gv)(e)&&arguments.length>1?Xe(e,t,n):Re(e)}function Xe(e,t,n){const r=e[t];return Ae(r)?r:new ze(e,t,n)}class qe{constructor(e,t,n){this.fn=e,this.setter=t,this._value=void 0,this.dep=new P(this),this.__v_isRef=!0,this.deps=void 0,this.depsTail=void 0,this.flags=16,this.globalVersion=A-1,this.next=void 0,this.effect=this,this.__v_isReadonly=!t,this.isSSR=n}notify(){if(this.flags|=16,!(8&this.flags||o===this))return m(this,!0),!0}get value(){const e=this.dep.track();return O(this),e&&(e.version=this.dep.version),this._value}set value(e){this.setter&&this.setter(e)}}function Ve(e,t,n=!1){let a,o;return(0,r.Tn)(e)?a=e:(a=e.get,o=e.set),new qe(a,o,n)}const He={},Ze=new WeakMap;let Ge;function Ke(e,t,n=r.MZ){const{immediate:a,deep:o,once:i,scheduler:s,augmentJob:c,call:u}=n,p=e=>o?e:xe(e)||!1===o||0===o?Ye(e,1):Ye(e);let h,f,m,g,v=!1,b=!1;if(Ae(e)?(f=()=>e.value,v=xe(e)):Oe(e)?(f=()=>p(e),v=!0):(0,r.cy)(e)?(b=!0,v=e.some((e=>Oe(e)||xe(e))),f=()=>e.map((e=>Ae(e)?e.value:Oe(e)?p(e):(0,r.Tn)(e)?u?u(e,2):e():void 0))):f=(0,r.Tn)(e)?t?u?()=>u(e,2):e:()=>{if(m){_();try{m()}finally{C()}}const t=Ge;Ge=h;try{return u?u(e,3,[g]):e(g)}finally{Ge=t}}:r.tE,t&&o){const e=f,t=!0===o?1/0:o;f=()=>Ye(e(),t)}const y=l(),w=()=>{h.stop(),y&&y.active&&(0,r.TF)(y.effects,h)};if(i&&t){const e=t;t=(...t)=>{e(...t),w()}}let O=b?new Array(e.length).fill(He):He;const k=e=>{if(1&h.flags&&(h.dirty||e))if(t){const e=h.run();if(o||v||(b?e.some(((e,t)=>(0,r.$H)(e,O[t]))):(0,r.$H)(e,O))){m&&m();const n=Ge;Ge=h;try{const n=[e,O===He?void 0:b&&O[0]===He?[]:O,g];u?u(t,3,n):t(...n),O=e}finally{Ge=n}}}else h.run()};return c&&c(k),h=new d(f),h.scheduler=s?()=>s(k,!1):k,g=e=>function(e,t=!1,n=Ge){if(n){let t=Ze.get(n);t||Ze.set(n,t=[]),t.push(e)}}(e,!1,h),m=h.onStop=()=>{const e=Ze.get(h);if(e){if(u)u(e,4);else for(const t of e)t();Ze.delete(h)}},t?a?k(!0):O=h.run():s?s(k.bind(null,!0),!0):h.run(),w.pause=h.pause.bind(h),w.resume=h.resume.bind(h),w.stop=w,w}function Ye(e,t=1/0,n){if(t<=0||!(0,r.Gv)(e)||e.__v_skip)return e;if((n=n||new Set).has(e))return e;if(n.add(e),t--,Ae(e))Ye(e.value,t,n);else if((0,r.cy)(e))for(let r=0;r<e.length;r++)Ye(e[r],t,n);else if((0,r.vM)(e)||(0,r.CE)(e))e.forEach((e=>{Ye(e,t,n)}));else if((0,r.Qd)(e)){for(const r in e)Ye(e[r],t,n);for(const r of Object.getOwnPropertySymbols(e))Object.prototype.propertyIsEnumerable.call(e,r)&&Ye(e[r],t,n)}return e}},961:(e,t,n)=>{"use strict";!function e(){if("undefined"!=typeof __REACT_DEVTOOLS_GLOBAL_HOOK__&&"function"==typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE)try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(e){console.error(e)}}(),e.exports=n(6221)},1014:(e,t,n)=>{"use strict";n.d(t,{qL:()=>p,Ro:()=>s,kL:()=>i,MH:()=>l,ty:()=>c});const r="/* basic theme */\n.light-mode {\n  --scalar-background-1: #fff;\n  --scalar-background-2: #f6f6f6;\n  --scalar-background-3: #e7e7e7;\n  --scalar-background-accent: #8ab4f81f;\n\n  --scalar-color-1: #2a2f45;\n  --scalar-color-2: #757575;\n  --scalar-color-3: #8e8e8e;\n\n  --scalar-color-accent: #0099ff;\n  --scalar-border-color: #dfdfdf;\n}\n.dark-mode {\n  --scalar-background-1: #0f0f0f;\n  --scalar-background-2: #1a1a1a;\n  --scalar-background-3: #272727;\n\n  --scalar-color-1: #e7e7e7;\n  --scalar-color-2: #a4a4a4;\n  --scalar-color-3: #797979;\n\n  --scalar-color-accent: #3ea6ff;\n  --scalar-background-accent: #3ea6ff1f;\n\n  --scalar-border-color: #2d2d2d;\n}\n/* Document Sidebar */\n.light-mode .t-doc__sidebar,\n.dark-mode .t-doc__sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n\n  --scalar-sidebar-item-hover-background: var(--scalar-background-2);\n  --scalar-sidebar-item-hover-color: currentColor;\n\n  --scalar-sidebar-item-active-background: var(--scalar-background-2);\n  --scalar-sidebar-color-active: var(--scalar-color-1);\n\n  --scalar-sidebar-search-background: transparent;\n  --scalar-sidebar-search-color: var(--scalar-color-3);\n  --scalar-sidebar-search-border-color: var(--scalar-border-color);\n}\n/* advanced */\n.light-mode {\n  --scalar-color-green: #069061;\n  --scalar-color-red: #ef0006;\n  --scalar-color-yellow: #edbe20;\n  --scalar-color-blue: #0082d0;\n  --scalar-color-orange: #fb892c;\n  --scalar-color-purple: #5203d1;\n\n  --scalar-button-1: rgba(0, 0, 0, 1);\n  --scalar-button-1-hover: rgba(0, 0, 0, 0.8);\n  --scalar-button-1-color: rgba(255, 255, 255, 0.9);\n\n  --scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);\n\n  --scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);\n  --scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%);\n}\n.dark-mode {\n  --scalar-color-green: #00b648;\n  --scalar-color-red: #dc1b19;\n  --scalar-color-yellow: #ffc90d;\n  --scalar-color-blue: #4eb3ec;\n  --scalar-color-orange: #ff8d4d;\n  --scalar-color-purple: #b191f9;\n\n  --scalar-button-1: rgba(255, 255, 255, 1);\n  --scalar-button-1-hover: rgba(255, 255, 255, 0.9);\n  --scalar-button-1-color: black;\n\n  --scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);\n\n  --scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);\n  --scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%);\n}\n",a=[["--theme-","--scalar-"],["--sidebar-","--scalar-sidebar-"]],o=a.map((([e])=>e));function i(e){return o.some((t=>e.includes(t)))?(console.warn("DEPRECATION WARNING: It looks like you're using legacy CSS variables in your custom CSS string. Please migrate them to use the updated prefixes. See https://github.com/scalar/scalar#theme-prefix-changes"),a.reduce(((e,[t,n])=>e.replaceAll(t,n)),e)):e}function s(){if(typeof window>"u")return!1;const e=document.createElement("div");e.setAttribute("style","width:30px;height:30px;overflow-y:scroll;"),e.classList.add("scrollbar-test");const t=document.createElement("div");t.setAttribute("style","width:100%;height:40px"),e.appendChild(t),document.body.appendChild(e);const n=30-e.firstChild.clientWidth;return document.body.removeChild(e),!!n}const l=["alternate","default","moon","purple","solarized","bluePlanet","deepSpace","saturn","kepler","elysiajs","fastify","mars","none"],c={default:"Default",alternate:"Alternate",moon:"Moon",purple:"Purple",solarized:"Solarized",elysiajs:"Elysia.js",fastify:"Fastify",bluePlanet:"Blue Planet",saturn:"Saturn",kepler:"Kepler-11e",mars:"Mars",deepSpace:"Deep Space",none:""},u={alternate:"/* basic theme */\n:root {\n  --scalar-text-decoration: underline;\n  --scalar-text-decoration-hover: underline;\n}\n.light-mode,\n.light-mode .dark-mode {\n  --scalar-background-1: #f9f9f9;\n  --scalar-background-2: #f1f1f1;\n  --scalar-background-3: #e7e7e7;\n  --scalar-background-card: #fff;\n\n  --scalar-color-1: #2a2f45;\n  --scalar-color-2: #757575;\n  --scalar-color-3: #8e8e8e;\n\n  --scalar-color-accent: var(--scalar-color-1);\n  --scalar-background-accent: var(--scalar-background-3);\n\n  --scalar-border-color: rgba(0, 0, 0, 0.1);\n}\n.dark-mode {\n  --scalar-background-1: #131313;\n  --scalar-background-2: #1d1d1d;\n  --scalar-background-3: #272727;\n  --scalar-background-card: #1d1d1d;\n\n  --scalar-color-1: rgba(255, 255, 255, 0.9);\n  --scalar-color-2: rgba(255, 255, 255, 0.62);\n  --scalar-color-3: rgba(255, 255, 255, 0.44);\n\n  --scalar-color-accent: var(--scalar-color-1);\n  --scalar-background-accent: var(--scalar-background-3);\n\n  --scalar-border-color: #2a2b2a;\n}\n/* Document Sidebar */\n.light-mode .t-doc__sidebar,\n.dark-mode .t-doc__sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n\n  --scalar-sidebar-item-hover-background: var(--scalar-background-2);\n  --scalar-sidebar-item-hover-color: currentColor;\n\n  --scalar-sidebar-item-active-background: var(--scalar-background-accent);\n  --scalar-sidebar-color-active: var(--scalar-color-accent);\n\n  --scalar-sidebar-search-background: transparent;\n  --scalar-sidebar-search-color: var(--scalar-color-3);\n  --scalar-sidebar-search-border-color: var(--scalar-border-color);\n\n  --scalar-sidebar-indent-border: var(--scalar-sidebar-border-color);\n  --scalar-sidebar-indent-border-hover: var(--scalar-sidebar-border-color);\n  --scalar-sidebar-indent-border-active: var(--scalar-sidebar-border-color);\n}\n/* advanced */\n.light-mode .dark-mode,\n.light-mode {\n  --scalar-color-green: #069061;\n  --scalar-color-red: #ef0006;\n  --scalar-color-yellow: #edbe20;\n  --scalar-color-blue: #0082d0;\n  --scalar-color-orange: #fb892c;\n  --scalar-color-purple: #5203d1;\n\n  --scalar-button-1: rgba(0, 0, 0, 1);\n  --scalar-button-1-hover: rgba(0, 0, 0, 0.8);\n  --scalar-button-1-color: rgba(255, 255, 255, 0.9);\n}\n.dark-mode {\n  --scalar-color-green: #00b648;\n  --scalar-color-red: #dd2f2c;\n  --scalar-color-yellow: #ffc90d;\n  --scalar-color-blue: #4eb3ec;\n  --scalar-color-orange: #ff8d4d;\n  --scalar-color-purple: #b191f9;\n\n  --scalar-button-1: rgba(255, 255, 255, 1);\n  --scalar-button-1-hover: rgba(255, 255, 255, 0.9);\n  --scalar-button-1-color: black;\n}\n\n.scalar-api-client__item,\n.scalar-card,\n.dark-mode .dark-mode.scalar-card {\n  --scalar-background-1: var(--scalar-background-card);\n  --scalar-background-2: var(--scalar-background-1);\n  --scalar-background-3: var(--scalar-background-1);\n}\n.dark-mode .dark-mode.scalar-card {\n  --scalar-background-3: var(--scalar-background-3);\n}\n.t-doc__sidebar {\n  --scalar-color-green: var(--scalar-color-1);\n  --scalar-color-red: var(--scalar-color-1);\n  --scalar-color-yellow: var(--scalar-color-1);\n  --scalar-color-blue: var(--scalar-color-1);\n  --scalar-color-orange: var(--scalar-color-1);\n  --scalar-color-purple: var(--scalar-color-1);\n}\n.light-mode *::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-blue), transparent 70%);\n}\n.dark-mode *::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-blue), transparent 50%);\n}\n",default:r,moon:".light-mode {\n  color-scheme: light;\n  --scalar-color-1: #000000;\n  --scalar-color-2: #000000;\n  --scalar-color-3: #000000;\n  --scalar-color-accent: #645b0f;\n  --scalar-background-1: #ccc9b3;\n  --scalar-background-2: #c2bfaa;\n  --scalar-background-3: #b8b5a1;\n  --scalar-background-accent: #000000;\n\n  --scalar-border-color: rgba(0, 0, 0, 0.2);\n  --scalar-scrollbar-color: rgba(0, 0, 0, 0.18);\n  --scalar-scrollbar-color-active: rgba(0, 0, 0, 0.36);\n  --scalar-lifted-brightness: 1;\n  --scalar-backdrop-brightness: 1;\n\n  --scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, 0.11);\n  --scalar-shadow-2: rgba(0, 0, 0, 0.08) 0px 13px 20px 0px, rgba(0, 0, 0, 0.08) 0px 3px 8px 0px,\n    var(--scalar-border-color) 0px 0 0 1px;\n\n  --scalar-button-1: rgb(49 53 56);\n  --scalar-button-1-color: #fff;\n  --scalar-button-1-hover: rgb(28 31 33);\n\n  --scalar-color-red: #b91c1c;\n  --scalar-color-orange: #a16207;\n  --scalar-color-green: #047857;\n  --scalar-color-blue: #1d4ed8;\n  --scalar-color-orange: #c2410c;\n  --scalar-color-purple: #6d28d9;\n}\n\n.dark-mode {\n  color-scheme: dark;\n  --scalar-color-1: #fffef3;\n  --scalar-color-2: #fffef3;\n  --scalar-color-3: #fffef3;\n  --scalar-color-accent: #c3b531;\n  --scalar-background-1: #313332;\n  --scalar-background-2: #393b3a;\n  --scalar-background-3: #414342;\n  --scalar-background-accent: #fffef3;\n\n  --scalar-border-color: #505452;\n  --scalar-scrollbar-color: rgba(255, 255, 255, 0.24);\n  --scalar-scrollbar-color-active: rgba(255, 255, 255, 0.48);\n  --scalar-lifted-brightness: 1.45;\n  --scalar-backdrop-brightness: 0.5;\n\n  --scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, 0.11);\n  --scalar-shadow-2: rgba(15, 15, 15, 0.2) 0px 3px 6px, rgba(15, 15, 15, 0.4) 0px 9px 24px, 0 0 0 1px\n    rgba(255, 255, 255, 0.1);\n\n  --scalar-button-1: #f6f6f6;\n  --scalar-button-1-color: #000;\n  --scalar-button-1-hover: #e7e7e7;\n\n  --scalar-color-green: #00b648;\n  --scalar-color-red: #dc1b19;\n  --scalar-color-yellow: #ffc90d;\n  --scalar-color-blue: #4eb3ec;\n  --scalar-color-orange: #ff8d4d;\n  --scalar-color-purple: #b191f9;\n}\n\n/* Sidebar */\n.light-mode .t-doc__sidebar,\n.dark-mode .t-doc__sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n\n  --scalar-sidebar-item-hover-background: var(--scalar-background-2);\n  --scalar-sidebar-item-hover-color: currentColor;\n\n  --scalar-sidebar-item-active-background: var(--scalar-background-3);\n  --scalar-sidebar-color-active: var(--scalar-color-1);\n\n  --scalar-sidebar-search-background: transparent;\n  --scalar-sidebar-search-color: var(--scalar-color-3);\n  --scalar-sidebar-search-border-color: var(--scalar-border-color);\n}\n*::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-accent), transparent 80%);\n}\n",elysiajs:'.light-mode {\n  --scalar-color-1: #2a2f45;\n  --scalar-color-2: #757575;\n  --scalar-color-3: #8e8e8e;\n  --scalar-color-accent: #f06292;\n\n  --scalar-background-1: #fff;\n  --scalar-background-2: #f6f6f6;\n  --scalar-background-3: #e7e7e7;\n\n  --scalar-border-color: rgba(0, 0, 0, 0.1);\n}\n.dark-mode {\n  --scalar-color-1: rgba(255, 255, 255, 0.9);\n  --scalar-color-2: rgba(156, 163, 175, 1);\n  --scalar-color-3: rgba(255, 255, 255, 0.44);\n  --scalar-color-accent: #f06292;\n\n  --scalar-background-1: #111728;\n  --scalar-background-2: #1e293b;\n  --scalar-background-3: #334155;\n  --scalar-background-accent: #f062921f;\n\n  --scalar-border-color: rgba(255, 255, 255, 0.1);\n}\n\n/* Document Sidebar */\n.light-mode .t-doc__sidebar,\n.dark-mode .t-doc__sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n\n  --scalar-sidebar-item-hover-background: var(--scalar-background-2);\n  --scalar-sidebar-item-hover-color: currentColor;\n\n  --scalar-sidebar-item-active-background: #f062921f;\n  --scalar-sidebar-color-active: var(--scalar-color-accent);\n\n  --scalar-sidebar-search-background: transparent;\n  --scalar-sidebar-search-color: var(--scalar-color-3);\n  --scalar-sidebar-search-border-color: var(--scalar-border-color);\n}\n\n/* advanced */\n.light-mode {\n  --scalar-button-1: rgb(49 53 56);\n  --scalar-button-1-color: #fff;\n  --scalar-button-1-hover: rgb(28 31 33);\n\n  --scalar-color-green: #069061;\n  --scalar-color-red: #ef0006;\n  --scalar-color-yellow: #edbe20;\n  --scalar-color-blue: #0082d0;\n  --scalar-color-orange: #fb892c;\n  --scalar-color-purple: #5203d1;\n\n  --scalar-scrollbar-color: rgba(0, 0, 0, 0.18);\n  --scalar-scrollbar-color-active: rgba(0, 0, 0, 0.36);\n}\n.dark-mode {\n  --scalar-button-1: #f6f6f6;\n  --scalar-button-1-color: #000;\n  --scalar-button-1-hover: #e7e7e7;\n\n  --scalar-color-green: #a3ffa9;\n  --scalar-color-red: #ffa3a3;\n  --scalar-color-yellow: #fffca3;\n  --scalar-color-blue: #a5d6ff;\n  --scalar-color-orange: #e2ae83;\n  --scalar-color-purple: #d2a8ff;\n\n  --scalar-scrollbar-color: rgba(255, 255, 255, 0.24);\n  --scalar-scrollbar-color-active: rgba(255, 255, 255, 0.48);\n}\n.section-flare {\n  width: 100%;\n  height: 400px;\n  position: absolute;\n}\n.section-flare-item:first-of-type:before {\n  content: "";\n  position: absolute;\n  top: 0;\n  right: 0;\n  bottom: 0;\n  left: 0;\n  --stripes: repeating-linear-gradient(100deg, #fff 0%, #fff 0%, transparent 2%, transparent 12%, #fff 17%);\n  --stripesDark: repeating-linear-gradient(100deg, #000 0%, #000 0%, transparent 10%, transparent 12%, #000 17%);\n  --rainbow: repeating-linear-gradient(100deg, #60a5fa 10%, #e879f9 16%, #5eead4 22%, #60a5fa 30%);\n  contain: strict;\n  contain-intrinsic-size: 100vw 40vh;\n  background-image: var(--stripesDark), var(--rainbow);\n  background-size: 300%, 200%;\n  background-position: 50% 50%, 50% 50%;\n  filter: opacity(20%) saturate(200%);\n  -webkit-mask-image: radial-gradient(ellipse at 100% 0%, black 40%, transparent 70%);\n  mask-image: radial-gradient(ellipse at 100% 0%, black 40%, transparent 70%);\n  pointer-events: none;\n}\n.section-flare-item:first-of-type:after {\n  content: "";\n  position: absolute;\n  top: 0;\n  right: 0;\n  bottom: 0;\n  left: 0;\n  background-image: var(--stripes), var(--rainbow);\n  background-size: 200%, 100%;\n  background-attachment: fixed;\n  mix-blend-mode: difference;\n  background-image: var(--stripesDark), var(--rainbow);\n  pointer-events: none;\n}\n.light-mode .section-flare-item:first-of-type:after,\n.light-mode .section-flare-item:first-of-type:before {\n  background-image: var(--stripes), var(--rainbow);\n  filter: opacity(4%) saturate(200%);\n}\n',fastify:".light-mode {\n  color-scheme: light;\n  --scalar-color-1: #1c1e21;\n  --scalar-color-2: #757575;\n  --scalar-color-3: #8e8e8e;\n  --scalar-color-disabled: #b4b1b1;\n  --scalar-color-ghost: #a7a7a7;\n  --scalar-color-accent: #2f8555;\n  --scalar-background-1: #fff;\n  --scalar-background-2: #f5f5f5;\n  --scalar-background-3: #ededed;\n  --scalar-background-4: rgba(0, 0, 0, 0.06);\n  --scalar-background-accent: #2f85551f;\n\n  --scalar-border-color: rgba(0, 0, 0, 0.1);\n  --scalar-scrollbar-color: rgba(0, 0, 0, 0.18);\n  --scalar-scrollbar-color-active: rgba(0, 0, 0, 0.36);\n  --scalar-lifted-brightness: 1;\n  --scalar-backdrop-brightness: 1;\n\n  --scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, 0.11);\n  --scalar-shadow-2: rgba(0, 0, 0, 0.08) 0px 13px 20px 0px, rgba(0, 0, 0, 0.08) 0px 3px 8px 0px, #eeeeed 0px 0 0 1px;\n\n  --scalar-button-1: rgb(49 53 56);\n  --scalar-button-1-color: #fff;\n  --scalar-button-1-hover: rgb(28 31 33);\n\n  --scalar-color-green: #007300;\n  --scalar-color-red: #af272b;\n  --scalar-color-yellow: #b38200;\n  --scalar-color-blue: #3b8ba5;\n  --scalar-color-orange: #fb892c;\n  --scalar-color-purple: #5203d1;\n}\n\n.dark-mode {\n  color-scheme: dark;\n  --scalar-color-1: rgba(255, 255, 255, 0.9);\n  --scalar-color-2: rgba(255, 255, 255, 0.62);\n  --scalar-color-3: rgba(255, 255, 255, 0.44);\n  --scalar-color-disabled: rgba(255, 255, 255, 0.34);\n  --scalar-color-ghost: rgba(255, 255, 255, 0.26);\n  --scalar-color-accent: #27c2a0;\n  --scalar-background-1: #1b1b1d;\n  --scalar-background-2: #242526;\n  --scalar-background-3: #3b3b3b;\n  --scalar-background-4: rgba(255, 255, 255, 0.06);\n  --scalar-background-accent: #27c2a01f;\n\n  --scalar-border-color: rgba(255, 255, 255, 0.1);\n  --scalar-scrollbar-color: rgba(255, 255, 255, 0.24);\n  --scalar-scrollbar-color-active: rgba(255, 255, 255, 0.48);\n  --scalar-lifted-brightness: 1.45;\n  --scalar-backdrop-brightness: 0.5;\n\n  --scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, 0.1);\n  --scalar-shadow-2: rgba(15, 15, 15, 0.2) 0px 3px 6px, rgba(15, 15, 15, 0.4) 0px 9px 24px, 0 0 0 1px\n    rgba(255, 255, 255, 0.1);\n\n  --scalar-button-1: #f6f6f6;\n  --scalar-button-1-color: #000;\n  --scalar-button-1-hover: #e7e7e7;\n\n  --scalar-color-green: #26b226;\n  --scalar-color-red: #fb565b;\n  --scalar-color-yellow: #ffc426;\n  --scalar-color-blue: #6ecfef;\n  --scalar-color-orange: #ff8d4d;\n  --scalar-color-purple: #b191f9;\n}\n",purple:"/* basic theme */\n.light-mode {\n  --scalar-background-1: #fff;\n  --scalar-background-2: #f5f6f8;\n  --scalar-background-3: #eceef1;\n\n  --scalar-color-1: #2a2f45;\n  --scalar-color-2: #757575;\n  --scalar-color-3: #8e8e8e;\n\n  --scalar-color-accent: #5469d4;\n  --scalar-background-accent: #5469d41f;\n\n  --scalar-border-color: rgba(215, 215, 206, 0.68);\n}\n.dark-mode {\n  --scalar-background-1: #15171c;\n  --scalar-background-2: #1c1e24;\n  --scalar-background-3: #22252b;\n\n  --scalar-color-1: #fafafa;\n  --scalar-color-2: #c9ced8;\n  --scalar-color-3: #8c99ad;\n\n  --scalar-color-accent: #5469d4;\n  --scalar-background-accent: #5469d41f;\n\n  --scalar-border-color: #3f4145;\n}\n/* Document Sidebar */\n.light-mode .t-doc__sidebar,\n.dark-mode .t-doc__sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n\n  --scalar-sidebar-item-hover-color: currentColor;\n  --scalar-sidebar-item-hover-background: var(--scalar-background-3);\n\n  --scalar-sidebar-item-active-background: var(--scalar-background-accent);\n  --scalar-sidebar-color-active: var(--scalar-color-accent);\n\n  --scalar-sidebar-search-background: var(--scalar-background-1);\n  --scalar-sidebar-search-color: var(--scalar-color-3);\n  --scalar-sidebar-search-border-color: var(--scalar-border-color);\n}\n\n/* advanced */\n.light-mode {\n  --scalar-color-green: #17803d;\n  --scalar-color-red: #e10909;\n  --scalar-color-yellow: #edbe20;\n  --scalar-color-blue: #1763a6;\n  --scalar-color-orange: #e25b09;\n  --scalar-color-purple: #5c3993;\n\n  --scalar-button-1: rgba(0, 0, 0, 1);\n  --scalar-button-1-hover: rgba(0, 0, 0, 0.8);\n  --scalar-button-1-color: rgba(255, 255, 255, 0.9);\n}\n.dark-mode {\n  --scalar-color-green: #30a159;\n  --scalar-color-red: #dc1b19;\n  --scalar-color-yellow: #eec644;\n  --scalar-color-blue: #2b7abf;\n  --scalar-color-orange: #f07528;\n  --scalar-color-purple: #7a59b1;\n\n  --scalar-button-1: rgba(255, 255, 255, 1);\n  --scalar-button-1-hover: rgba(255, 255, 255, 0.9);\n  --scalar-button-1-color: black;\n}\n.light-mode *::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-accent), transparent 70%);\n}\n.dark-mode *::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-accent), transparent 50%);\n}\n",solarized:".light-mode {\n  color-scheme: light;\n  --scalar-color-1: #584c27;\n  --scalar-color-2: #616161;\n  --scalar-color-3: #a89f84;\n  --scalar-color-accent: #b58900;\n  --scalar-background-1: #fdf6e3;\n  --scalar-background-2: #eee8d5;\n  --scalar-background-3: #ddd6c1;\n  --scalar-background-accent: #b589001f;\n\n  --scalar-border-color: #ded8c8;\n  --scalar-scrollbar-color: rgba(0, 0, 0, 0.18);\n  --scalar-scrollbar-color-active: rgba(0, 0, 0, 0.36);\n  --scalar-lifted-brightness: 1;\n  --scalar-backdrop-brightness: 1;\n\n  --scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, 0.11);\n  --scalar-shadow-2: rgba(0, 0, 0, 0.08) 0px 13px 20px 0px, rgba(0, 0, 0, 0.08) 0px 3px 8px 0px, #eeeeed 0px 0 0 1px;\n\n  --scalar-button-1: rgb(49 53 56);\n  --scalar-button-1-color: #fff;\n  --scalar-button-1-hover: rgb(28 31 33);\n\n  --scalar-color-red: #b91c1c;\n  --scalar-color-orange: #a16207;\n  --scalar-color-green: #047857;\n  --scalar-color-blue: #1d4ed8;\n  --scalar-color-orange: #c2410c;\n  --scalar-color-purple: #6d28d9;\n}\n\n.dark-mode {\n  color-scheme: dark;\n  --scalar-color-1: #fff;\n  --scalar-color-2: #cccccc;\n  --scalar-color-3: #6d8890;\n  --scalar-color-accent: #007acc;\n  --scalar-background-1: #00212b;\n  --scalar-background-2: #012b36;\n  --scalar-background-3: #004052;\n  --scalar-background-accent: #015a6f;\n\n  --scalar-border-color: #2f4851;\n  --scalar-scrollbar-color: rgba(255, 255, 255, 0.24);\n  --scalar-scrollbar-color-active: rgba(255, 255, 255, 0.48);\n  --scalar-lifted-brightness: 1.45;\n  --scalar-backdrop-brightness: 0.5;\n\n  --scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, 0.1);\n  --scalar-shadow-2: rgba(15, 15, 15, 0.2) 0px 3px 6px, rgba(15, 15, 15, 0.4) 0px 9px 24px, 0 0 0 1px\n    rgba(255, 255, 255, 0.1);\n\n  --scalar-button-1: #f6f6f6;\n  --scalar-button-1-color: #000;\n  --scalar-button-1-hover: #e7e7e7;\n\n  --scalar-color-green: #00b648;\n  --scalar-color-red: #dc1b19;\n  --scalar-color-yellow: #ffc90d;\n  --scalar-color-blue: #4eb3ec;\n  --scalar-color-orange: #ff8d4d;\n  --scalar-color-purple: #b191f9;\n}\n\n/* Sidebar */\n.light-mode .t-doc__sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-item-hover-color: currentColor;\n  --scalar-sidebar-item-hover-background: var(--scalar-background-2);\n  --scalar-sidebar-item-active-background: var(--scalar-background-accent);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-color-active: var(--scalar-color-accent);\n  --scalar-sidebar-search-background: var(--scalar-background-2);\n  --scalar-sidebar-search-border-color: var(--scalar-sidebar-search-background);\n  --scalar-sidebar-search--color: var(--scalar-color-3);\n}\n\n.dark-mode .sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-item-hover-color: currentColor;\n  --scalar-sidebar-item-hover-background: var(--scalar-background-2);\n  --scalar-sidebar-item-active-background: var(--scalar-background-accent);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-color-active: var(--scalar-sidebar-color-1);\n  --scalar-sidebar-search-background: var(--scalar-background-2);\n  --scalar-sidebar-search-border-color: var(--scalar-sidebar-search-background);\n  --scalar-sidebar-search--color: var(--scalar-color-3);\n}\n*::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-accent), transparent 70%);\n}\n",bluePlanet:"/* basic theme */\n:root {\n  --scalar-text-decoration: underline;\n  --scalar-text-decoration-hover: underline;\n}\n.light-mode {\n  --scalar-background-1: #f0f2f5;\n  --scalar-background-2: #eaecf0;\n  --scalar-background-3: #e0e2e6;\n  --scalar-border-color: rgb(213 213 213);\n\n  --scalar-color-1: rgb(9, 9, 11);\n  --scalar-color-2: rgb(113, 113, 122);\n  --scalar-color-3: rgba(25, 25, 28, 0.5);\n\n  --scalar-color-accent: var(--scalar-color-1);\n  --scalar-background-accent: #8ab4f81f;\n}\n.light-mode .scalar-card.dark-mode,\n.dark-mode {\n  --scalar-background-1: #000e23;\n  --scalar-background-2: #01132e;\n  --scalar-background-3: #03193b;\n  --scalar-border-color: #2e394c;\n\n  --scalar-color-1: #fafafa;\n  --scalar-color-2: rgb(161, 161, 170);\n  --scalar-color-3: rgba(255, 255, 255, 0.533);\n\n  --scalar-color-accent: var(--scalar-color-1);\n  --scalar-background-accent: #8ab4f81f;\n\n  --scalar-code-language-color-supersede: var(--scalar-color-1);\n}\n/* Document Sidebar */\n.light-mode .t-doc__sidebar,\n.dark-mode .t-doc__sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n\n  --scalar-sidebar-item-hover-background: var(--scalar-background-2);\n  --scalar-sidebar-item-hover-color: currentColor;\n\n  --scalar-sidebar-item-active-background: var(--scalar-background-3);\n  --scalar-sidebar-color-active: var(--scalar-color-accent);\n\n  --scalar-sidebar-search-background: rgba(255, 255, 255, 0.1);\n  --scalar-sidebar-search-border-color: var(--scalar-border-color);\n  --scalar-sidebar-search-color: var(--scalar-color-3);\n  z-index: 1;\n}\n.light-mode .t-doc__sidebar {\n  --scalar-sidebar-search-background: white;\n}\n/* advanced */\n.light-mode {\n  --scalar-color-green: #069061;\n  --scalar-color-red: #ef0006;\n  --scalar-color-yellow: #edbe20;\n  --scalar-color-blue: #0082d0;\n  --scalar-color-orange: #fb892c;\n  --scalar-color-purple: #5203d1;\n\n  --scalar-button-1: rgba(0, 0, 0, 1);\n  --scalar-button-1-hover: rgba(0, 0, 0, 0.8);\n  --scalar-button-1-color: rgba(255, 255, 255, 0.9);\n}\n.dark-mode {\n  --scalar-color-green: rgba(69, 255, 165, 0.823);\n  --scalar-color-red: #ff8589;\n  --scalar-color-yellow: #ffcc4d;\n  --scalar-color-blue: #6bc1fe;\n  --scalar-color-orange: #f98943;\n  --scalar-color-purple: #b191f9;\n\n  --scalar-button-1: rgba(255, 255, 255, 1);\n  --scalar-button-1-hover: rgba(255, 255, 255, 0.9);\n  --scalar-button-1-color: black;\n}\n/* Custom theme */\n/* Document header */\n@keyframes headerbackground {\n  from {\n    background: transparent;\n    backdrop-filter: none;\n  }\n  to {\n    background: var(--header-background-1);\n    backdrop-filter: blur(12px);\n  }\n}\n.dark-mode h2.t-editor__heading,\n.dark-mode .t-editor__page-title h1,\n.dark-mode h1.section-header:not(::selection),\n.dark-mode .markdown h1,\n.dark-mode .markdown h2,\n.dark-mode .markdown h3,\n.dark-mode .markdown h4,\n.dark-mode .markdown h5,\n.dark-mode .markdown h6 {\n  -webkit-text-fill-color: transparent;\n  background-image: linear-gradient(to right bottom, rgb(255, 255, 255) 30%, rgba(255, 255, 255, 0.38));\n  -webkit-background-clip: text;\n  background-clip: text;\n}\n/* Hero Section Flare */\n.section-flare-item:nth-of-type(1) {\n  --c1: #ffffff;\n  --c2: #babfd8;\n  --c3: #2e8bb2;\n  --c4: #1a8593;\n  --c5: #0a143e;\n  --c6: #0a0f52;\n  --c7: #2341b8;\n\n  --solid: var(--c1), var(--c2), var(--c3), var(--c4), var(--c5), var(--c6), var(--c7);\n  --solid-wrap: var(--solid), var(--c1);\n  --trans:\n    var(--c1), transparent, var(--c2), transparent, var(--c3),\n    transparent, var(--c4), transparent, var(--c5), transparent, var(--c6),\n    transparent, var(--c7);\n  --trans-wrap: var(--trans), transparent, var(--c1);\n\n  background: radial-gradient(circle, var(--trans)), conic-gradient(from 180deg, var(--trans-wrap)),\n    radial-gradient(circle, var(--trans)), conic-gradient(var(--solid-wrap));\n  width: 70vw;\n  height: 700px;\n  border-radius: 50%;\n  filter: blur(100px);\n  z-index: 0;\n  right: 0;\n  position: absolute;\n  transform: rotate(-45deg);\n  top: -300px;\n  opacity: 0.3;\n}\n.section-flare-item:nth-of-type(3) {\n  --star-color: #6b9acc;\n  --star-color2: #446b8d;\n  --star-color3: #3e5879;\n  background-image: radial-gradient(2px 2px at 20px 30px, var(--star-color2), rgba(0, 0, 0, 0)),\n    radial-gradient(2px 2px at 40px 70px, var(--star-color), rgba(0, 0, 0, 0)),\n    radial-gradient(2px 2px at 50px 160px, var(--star-color3), rgba(0, 0, 0, 0)),\n    radial-gradient(2px 2px at 90px 40px, var(--star-color), rgba(0, 0, 0, 0)),\n    radial-gradient(2px 2px at 130px 80px, var(--star-color), rgba(0, 0, 0, 0)),\n    radial-gradient(2px 2px at 160px 120px, var(--star-color3), rgba(0, 0, 0, 0));\n  background-repeat: repeat;\n  background-size: 200px 200px;\n  width: 100%;\n  height: 100%;\n  mask-image: radial-gradient(ellipse at 100% 0%, black 40%, transparent 70%);\n}\n.section-flare {\n  top: -150px !important;\n  height: 100vh;\n  background: linear-gradient(#000, var(--scalar-background-1));\n  width: 100vw;\n  overflow-x: hidden;\n}\n.light-mode .section-flare {\n  display: none;\n}\n.light-mode .scalar-card {\n  --scalar-background-1: #fff;\n  --scalar-background-2: #fff;\n  --scalar-background-3: #fff;\n}\n\n*::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-blue), transparent 60%);\n}\n",deepSpace:'/* basic theme */\n:root {\n  --scalar-text-decoration: underline;\n  --scalar-text-decoration-hover: underline;\n}\n.light-mode {\n  --scalar-color-1: rgb(9, 9, 11);\n  --scalar-color-2: rgb(113, 113, 122);\n  --scalar-color-3: rgba(25, 25, 28, 0.5);\n  --scalar-color-accent: var(--scalar-color-1);\n\n  --scalar-background-1: #fff;\n  --scalar-background-2: #f4f4f5;\n  --scalar-background-3: #e3e3e6;\n  --scalar-background-accent: #8ab4f81f;\n\n  --scalar-border-color: rgb(228, 228, 231);\n  --scalar-code-language-color-supersede: var(--scalar-color-1);\n}\n.dark-mode {\n  --scalar-color-1: #fafafa;\n  --scalar-color-2: rgb(161, 161, 170);\n  --scalar-color-3: rgba(255, 255, 255, 0.533);\n  --scalar-color-accent: var(--scalar-color-1);\n\n  --scalar-background-1: #09090b;\n  --scalar-background-2: #18181b;\n  --scalar-background-3: #2c2c30;\n  --scalar-background-accent: #8ab4f81f;\n\n  --scalar-border-color: rgba(255, 255, 255, 0.16);\n  --scalar-code-language-color-supersede: var(--scalar-color-1);\n}\n\n/* Document Sidebar */\n.light-mode .t-doc__sidebar,\n.dark-mode .t-doc__sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n\n  --scalar-sidebar-item-hover-color: currentColor;\n  --scalar-sidebar-item-hover-background: var(--scalar-background-2);\n\n  --scalar-sidebar-item-active-background: var(--scalar-background-3);\n  --scalar-sidebar-color-active: var(--scalar-color-accent);\n\n  --scalar-sidebar-search-background: transparent;\n  --scalar-sidebar-search-border-color: var(--scalar-border-color);\n  --scalar-sidebar-search-color: var(--scalar-color-3);\n}\n.light-mode .t-doc__sidebar {\n  --scalar-sidebar-item-active-background: var(--scalar-background-2);\n}\n/* advanced */\n.light-mode {\n  --scalar-color-green: #069061;\n  --scalar-color-red: #ef0006;\n  --scalar-color-yellow: #edbe20;\n  --scalar-color-blue: #0082d0;\n  --scalar-color-orange: #fb892c;\n  --scalar-color-purple: #5203d1;\n\n  --scalar-button-1: rgba(0, 0, 0, 1);\n  --scalar-button-1-hover: rgba(0, 0, 0, 0.8);\n  --scalar-button-1-color: rgba(255, 255, 255, 0.9);\n}\n.dark-mode {\n  --scalar-color-green: rgba(69, 255, 165, 0.823);\n  --scalar-color-red: #ff8589;\n  --scalar-color-yellow: #ffcc4d;\n  --scalar-color-blue: #6bc1fe;\n  --scalar-color-orange: #f98943;\n  --scalar-color-purple: #b191f9;\n\n  --scalar-button-1: rgba(255, 255, 255, 1);\n  --scalar-button-1-hover: rgba(255, 255, 255, 0.9);\n  --scalar-button-1-color: black;\n}\n/* Custom theme */\n.dark-mode h2.t-editor__heading,\n.dark-mode .t-editor__page-title h1,\n.dark-mode h1.section-header:not(::selection),\n.dark-mode .markdown h1,\n.dark-mode .markdown h2,\n.dark-mode .markdown h3,\n.dark-mode .markdown h4,\n.dark-mode .markdown h5,\n.dark-mode .markdown h6 {\n  -webkit-text-fill-color: transparent;\n  background-image: linear-gradient(to right bottom, rgb(255, 255, 255) 30%, rgba(255, 255, 255, 0.38));\n  -webkit-background-clip: text;\n  background-clip: text;\n}\n.examples .scalar-card-footer {\n  --scalar-background-3: transparent;\n  padding-top: 0;\n}\n/* Hero section flare */\n.section-flare {\n  width: 100vw;\n  height: 550px;\n  position: absolute;\n}\n.section-flare-item:nth-of-type(1) {\n  position: absolute;\n  width: 100vw;\n  height: 550px;\n  --stripesDark: repeating-linear-gradient(100deg, #000 0%, #000 7%, transparent 10%, transparent 12%, #000 16%);\n  --rainbow: repeating-linear-gradient(100deg, #fff 10%, #fff 16%, #fff 22%, #fff 30%);\n  background-image: var(--stripesDark), var(--rainbow);\n  background-size: 300%, 200%;\n  background-position: 50% 50%, 50% 50%;\n  filter: invert(100%);\n  -webkit-mask-image: radial-gradient(ellipse at 100% 0%, black 40%, transparent 70%);\n  mask-image: radial-gradient(ellipse at 100% 0%, black 40%, transparent 70%);\n  pointer-events: none;\n  opacity: 0.07;\n}\n.dark-mode .section-flare-item:nth-of-type(1) {\n  background-image: var(--stripesDark), var(--rainbow);\n  filter: opacity(50%) saturate(200%);\n  opacity: 0.25;\n  height: 350px;\n}\n.section-flare-item:nth-of-type(1):after {\n  content: "";\n  position: absolute;\n  top: 0;\n  right: 0;\n  bottom: 0;\n  left: 0;\n  background-image: var(--stripesDark), var(--rainbow);\n  background-size: 200%, 100%;\n  background-attachment: fixed;\n  mix-blend-mode: difference;\n}\n.dark-mode .section-flare:after {\n  background-image: var(--stripesDark), var(--rainbow);\n}\n.section-flare-item:nth-of-type(2) {\n  --star-color: #fff;\n  --star-color2: #fff;\n  --star-color3: #fff;\n  width: 100%;\n  height: 100%;\n  position: absolute;\n  background-image: radial-gradient(2px 2px at 20px 30px, var(--star-color2), rgba(0, 0, 0, 0)),\n    radial-gradient(2px 2px at 40px 70px, var(--star-color), rgba(0, 0, 0, 0)),\n    radial-gradient(2px 2px at 50px 160px, var(--star-color3), rgba(0, 0, 0, 0)),\n    radial-gradient(2px 2px at 90px 40px, var(--star-color), rgba(0, 0, 0, 0)),\n    radial-gradient(2px 2px at 130px 80px, var(--star-color), rgba(0, 0, 0, 0)),\n    radial-gradient(2px 2px at 160px 120px, var(--star-color3), rgba(0, 0, 0, 0));\n  background-repeat: repeat;\n  background-size: 200px 200px;\n  mask-image: radial-gradient(ellipse at 100% 0%, black 40%, transparent 70%);\n  opacity: 0.2;\n}\n.light-mode *::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-blue), transparent 70%);\n}\n.dark-mode *::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-blue), transparent 50%);\n}\n',saturn:"/* basic theme */\n.light-mode {\n  --scalar-background-1: #f3f3ee;\n  --scalar-background-2: #e8e8e3;\n  --scalar-background-3: #e4e4df;\n  --scalar-border-color: rgba(215, 215, 206, 0.85);\n\n  --scalar-color-1: #2a2f45;\n  --scalar-color-2: #757575;\n  --scalar-color-3: #8e8e8e;\n\n  --scalar-color-accent: #1763a6;\n  --scalar-background-accent: #1f648e1f;\n}\n.dark-mode {\n  --scalar-background-1: #09090b;\n  --scalar-background-2: #18181b;\n  --scalar-background-3: #2c2c30;\n  --scalar-border-color: rgba(255, 255, 255, 0.17);\n\n  --scalar-color-1: #fafafa;\n  --scalar-color-2: rgb(161, 161, 170);\n  --scalar-color-3: rgba(255, 255, 255, 0.533);\n\n  --scalar-color-accent: #4eb3ec;\n  --scalar-background-accent: #8ab4f81f;\n}\n/* Document Sidebar */\n.light-mode .t-doc__sidebar,\n.dark-mode .t-doc__sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n\n  --scalar-sidebar-item-hover-background: var(--scalar-background-2);\n  --scalar-sidebar-item-hover-color: currentColor;\n\n  --scalar-sidebar-item-active-background: var(--scalar-background-3);\n  --scalar-sidebar-color-active: var(--scalar-color-1);\n\n  --scalar-sidebar-search-background: var(--scalar-background-1);\n  --scalar-sidebar-search-border-color: var(--scalar-border-color);\n  --scalar-sidebar-search-color: var(--scalar-color-3);\n}\n\n/* advanced */\n.light-mode {\n  --scalar-color-green: #17803d;\n  --scalar-color-red: #e10909;\n  --scalar-color-yellow: #edbe20;\n  --scalar-color-blue: #1763a6;\n  --scalar-color-orange: #e25b09;\n  --scalar-color-purple: #5c3993;\n\n  --scalar-button-1: rgba(0, 0, 0, 1);\n  --scalar-button-1-hover: rgba(0, 0, 0, 0.8);\n  --scalar-button-1-color: rgba(255, 255, 255, 0.9);\n}\n.dark-mode {\n  --scalar-color-green: #30a159;\n  --scalar-color-red: #dc1b19;\n  --scalar-color-yellow: #eec644;\n  --scalar-color-blue: #2b7abf;\n  --scalar-color-orange: #f07528;\n  --scalar-color-purple: #7a59b1;\n\n  --scalar-button-1: rgba(255, 255, 255, 1);\n  --scalar-button-1-hover: rgba(255, 255, 255, 0.9);\n  --scalar-button-1-color: black;\n}\n.dark-mode h2.t-editor__heading,\n.dark-mode .t-editor__page-title h1,\n.dark-mode h1.section-header:not(::selection),\n.dark-mode .markdown h1,\n.dark-mode .markdown h2,\n.dark-mode .markdown h3,\n.dark-mode .markdown h4,\n.dark-mode .markdown h5,\n.dark-mode .markdown h6 {\n  -webkit-text-fill-color: transparent;\n  background-image: linear-gradient(to right bottom, rgb(255, 255, 255) 30%, rgba(255, 255, 255, 0.38));\n  -webkit-background-clip: text;\n  background-clip: text;\n}\n.light-mode *::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-accent), transparent 70%);\n}\n.dark-mode *::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-accent), transparent 50%);\n}\n",kepler:"/* basic theme */\n.light-mode {\n  --scalar-color-1: #2a2f45;\n  --scalar-color-2: #757575;\n  --scalar-color-3: #8e8e8e;\n  --scalar-color-accent: #7070ff;\n\n  --scalar-background-1: #fff;\n  --scalar-background-2: #f6f6f6;\n  --scalar-background-3: #e7e7e7;\n  --scalar-background-accent: #7070ff1f;\n\n  --scalar-border-color: rgba(0, 0, 0, 0.1);\n\n  --scalar-code-language-color-supersede: var(--scalar-color-3);\n}\n.dark-mode {\n  --scalar-color-1: #f7f8f8;\n  --scalar-color-2: rgb(180, 188, 208);\n  --scalar-color-3: #b4bcd099;\n  --scalar-color-accent: #828fff;\n\n  --scalar-background-1: #000212;\n  --scalar-background-2: #0d0f1e;\n  --scalar-background-3: #232533;\n  --scalar-background-accent: #8ab4f81f;\n\n  --scalar-border-color: #313245;\n  --scalar-code-language-color-supersede: var(--scalar-color-3);\n}\n/* Document Sidebar */\n.light-mode .t-doc__sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-item-hover-color: currentColor;\n  --scalar-sidebar-item-hover-background: var(--scalar-background-2);\n  --scalar-sidebar-item-active-background: var(--scalar-background-accent);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-color-active: var(--scalar-color-accent);\n  --scalar-sidebar-search-background: rgba(0, 0, 0, 0.05);\n  --scalar-sidebar-search-border-color: 1px solid rgba(0, 0, 0, 0.05);\n  --scalar-sidebar-search-color: var(--scalar-color-3);\n  --scalar-background-2: rgba(0, 0, 0, 0.03);\n}\n.dark-mode .t-doc__sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-item-hover-color: currentColor;\n  --scalar-sidebar-item-hover-background: var(--scalar-background-2);\n  --scalar-sidebar-item-active-background: rgba(255, 255, 255, 0.1);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-color-active: var(--scalar-color-accent);\n  --scalar-sidebar-search-background: rgba(255, 255, 255, 0.1);\n  --scalar-sidebar-search-border-color: 1px solid rgba(255, 255, 255, 0.05);\n  --scalar-sidebar-search-color: var(--scalar-color-3);\n}\n/* advanced */\n.light-mode {\n  --scalar-color-green: #069061;\n  --scalar-color-red: #ef0006;\n  --scalar-color-yellow: #edbe20;\n  --scalar-color-blue: #0082d0;\n  --scalar-color-orange: #fb892c;\n  --scalar-color-purple: #5203d1;\n\n  --scalar-button-1: rgba(0, 0, 0, 1);\n  --scalar-button-1-hover: rgba(0, 0, 0, 0.8);\n  --scalar-button-1-color: rgba(255, 255, 255, 0.9);\n}\n.dark-mode {\n  --scalar-color-green: #00b648;\n  --scalar-color-red: #dc1b19;\n  --scalar-color-yellow: #ffc90d;\n  --scalar-color-blue: #4eb3ec;\n  --scalar-color-orange: #ff8d4d;\n  --scalar-color-purple: #b191f9;\n\n  --scalar-button-1: rgba(255, 255, 255, 1);\n  --scalar-button-1-hover: rgba(255, 255, 255, 0.9);\n  --scalar-button-1-color: black;\n}\n/* Custom Theme */\n.dark-mode h2.t-editor__heading,\n.dark-mode .t-editor__page-title h1,\n.dark-mode h1.section-header:not(::selection),\n.dark-mode .markdown h1,\n.dark-mode .markdown h2,\n.dark-mode .markdown h3,\n.dark-mode .markdown h4,\n.dark-mode .markdown h5,\n.dark-mode .markdown h6 {\n  -webkit-text-fill-color: transparent;\n  background-image: linear-gradient(to right bottom, rgb(255, 255, 255) 30%, rgba(255, 255, 255, 0.38));\n  -webkit-background-clip: text;\n  background-clip: text;\n}\n.sidebar-search {\n  backdrop-filter: blur(12px);\n}\n@keyframes headerbackground {\n  from {\n    background: transparent;\n    backdrop-filter: none;\n  }\n  to {\n    background: var(--header-background-1);\n    backdrop-filter: blur(12px);\n  }\n}\n.dark-mode .scalar-card {\n  background: rgba(255, 255, 255, 0.05) !important;\n}\n.dark-mode .scalar-card * {\n  --scalar-background-2: transparent !important;\n  --scalar-background-1: transparent !important;\n}\n.light-mode .dark-mode.scalar-card *,\n.light-mode .dark-mode.scalar-card {\n  --scalar-background-1: #0d0f1e !important;\n  --scalar-background-2: #0d0f1e !important;\n  --scalar-background-3: #191b29 !important;\n}\n.light-mode .dark-mode.scalar-card {\n  background: #191b29 !important;\n}\n.badge {\n  box-shadow: 0 0 0 1px var(--scalar-border-color);\n  margin-right: 6px;\n}\n\n.table-row.required-parameter .table-row-item:nth-of-type(2):after {\n  background: transparent;\n  box-shadow: none;\n}\n/* Hero Section Flare */\n.section-flare {\n  width: 100vw;\n  background: radial-gradient(ellipse 80% 50% at 50% -20%, rgba(120, 119, 198, 0.3), transparent);\n  height: 100vh;\n}\n.light-mode *::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-accent), transparent 70%);\n}\n.dark-mode *::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-accent), transparent 50%);\n}\n",mars:"/* basic theme */\n:root {\n  --scalar-text-decoration: underline;\n  --scalar-text-decoration-hover: underline;\n}\n.light-mode {\n  --scalar-background-1: #f9f6f0;\n  --scalar-background-2: #f2efe8;\n  --scalar-background-3: #e9e7e2;\n  --scalar-border-color: rgba(203, 165, 156, 0.6);\n\n  --scalar-color-1: #c75549;\n  --scalar-color-2: #c75549;\n  --scalar-color-3: #c75549;\n\n  --scalar-color-accent: #c75549;\n  --scalar-background-accent: #dcbfa81f;\n\n  --scalar-code-language-color-supersede: var(--scalar-color-1);\n}\n.dark-mode {\n  --scalar-background-1: #140507;\n  --scalar-background-2: #20090c;\n  --scalar-background-3: #321116;\n  --scalar-border-color: #3c3031;\n\n  --scalar-color-1: rgba(255, 255, 255, 0.9);\n  --scalar-color-2: rgba(255, 255, 255, 0.62);\n  --scalar-color-3: rgba(255, 255, 255, 0.44);\n\n  --scalar-color-accent: rgba(255, 255, 255, 0.9);\n  --scalar-background-accent: #441313;\n\n  --scalar-code-language-color-supersede: var(--scalar-color-1);\n}\n\n/* Document Sidebar */\n.light-mode .t-doc__sidebar,\n.dark-mode .t-doc__sidebar {\n  --scalar-sidebar-background-1: var(--scalar-background-1);\n  --scalar-sidebar-color-1: var(--scalar-color-1);\n  --scalar-sidebar-color-2: var(--scalar-color-2);\n  --scalar-sidebar-border-color: var(--scalar-border-color);\n\n  --scalar-sidebar-item-hover-color: currentColor;\n  --scalar-sidebar-item-hover-background: var(--scalar-background-2);\n\n  --scalar-sidebar-item-active-background: var(--scalar-background-3);\n  --scalar-sidebar-color-active: var(--scalar-color-accent);\n\n  --scalar-sidebar-search-background: rgba(255, 255, 255, 0.1);\n  --scalar-sidebar-search-color: var(--scalar-color-3);\n  --scalar-sidebar-search-border-color: var(--scalar-border-color);\n  z-index: 1;\n}\n/* advanced */\n.light-mode {\n  --scalar-color-green: #09533a;\n  --scalar-color-red: #aa181d;\n  --scalar-color-yellow: #ab8d2b;\n  --scalar-color-blue: #19689a;\n  --scalar-color-orange: #b26c34;\n  --scalar-color-purple: #4c2191;\n\n  --scalar-button-1: rgba(0, 0, 0, 1);\n  --scalar-button-1-hover: rgba(0, 0, 0, 0.8);\n  --scalar-button-1-color: rgba(255, 255, 255, 0.9);\n}\n.dark-mode {\n  --scalar-color-green: rgba(69, 255, 165, 0.823);\n  --scalar-color-red: #ff8589;\n  --scalar-color-yellow: #ffcc4d;\n  --scalar-color-blue: #6bc1fe;\n  --scalar-color-orange: #f98943;\n  --scalar-color-purple: #b191f9;\n\n  --scalar-button-1: rgba(255, 255, 255, 1);\n  --scalar-button-1-hover: rgba(255, 255, 255, 0.9);\n  --scalar-button-1-color: black;\n}\n/* Custom Theme */\n.dark-mode h2.t-editor__heading,\n.dark-mode .t-editor__page-title h1,\n.dark-mode h1.section-header:not(::selection),\n.dark-mode .markdown h1,\n.dark-mode .markdown h2,\n.dark-mode .markdown h3,\n.dark-mode .markdown h4,\n.dark-mode .markdown h5,\n.dark-mode .markdown h6 {\n  -webkit-text-fill-color: transparent;\n  background-image: linear-gradient(to right bottom, rgb(255, 255, 255) 30%, rgba(255, 255, 255, 0.38));\n  -webkit-background-clip: text;\n  background-clip: text;\n}\n.light-mode .t-doc__sidebar {\n  --scalar-sidebar-search-background: white;\n}\n.examples .scalar-card-footer {\n  --scalar-background-3: transparent;\n  padding-top: 0;\n}\n/* Hero section flare */\n.section-flare {\n  overflow-x: hidden;\n  height: 100vh;\n  left: initial;\n}\n.section-flare-item:nth-of-type(1) {\n  background: #d25019;\n  position: relative;\n  top: -150px;\n  right: -400px;\n  width: 80vw;\n  height: 500px;\n  margin-top: -150px;\n  border-radius: 50%;\n  filter: blur(100px);\n  z-index: 0;\n}\n.light-mode .section-flare {\n  display: none;\n}\n*::selection {\n  background-color: color-mix(in srgb, var(--scalar-color-red), transparent 75%);\n}\n"},d=(Object.keys(u),e=>"none"===e?"":u[e||"default"]??r),p=(e,t)=>{const{fonts:n=!0,layer:r="scalar-theme"}=t??{},a=[d(e),n?'/* Inter (--scalar-font) */\n/* cyrillic-ext */\n@font-face {\n  font-family: "Inter";\n  font-style: normal;\n  font-weight: 100 900;\n  font-display: swap;\n  src: url(https://fonts.scalar.com/inter-cyrillic-ext.woff2) format("woff2");\n  unicode-range: U+0460-052F, U+1C80-1C88, U+20B4, U+2DE0-2DFF, U+A640-A69F, U+FE2E-FE2F;\n}\n/* cyrillic */\n@font-face {\n  font-family: "Inter";\n  font-style: normal;\n  font-weight: 100 900;\n  font-display: swap;\n  src: url(https://fonts.scalar.com/inter-cyrillic.woff2) format("woff2");\n  unicode-range: U+0301, U+0400-045F, U+0490-0491, U+04B0-04B1, U+2116;\n}\n/* greek-ext */\n@font-face {\n  font-family: "Inter";\n  font-style: normal;\n  font-weight: 100 900;\n  font-display: swap;\n  src: url(https://fonts.scalar.com/inter-greek-ext.woff2) format("woff2");\n  unicode-range: U+1F00-1FFF;\n}\n/* greek */\n@font-face {\n  font-family: "Inter";\n  font-style: normal;\n  font-weight: 100 900;\n  font-display: swap;\n  src: url(https://fonts.scalar.com/inter-greek.woff2) format("woff2");\n  unicode-range: U+0370-0377, U+037A-037F, U+0384-038A, U+038C, U+038E-03A1, U+03A3-03FF;\n}\n/* vietnamese */\n@font-face {\n  font-family: "Inter";\n  font-style: normal;\n  font-weight: 100 900;\n  font-display: swap;\n  src: url(https://fonts.scalar.com/inter-vietnamese.woff2) format("woff2");\n  unicode-range:\n    U+0102-0103, U+0110-0111, U+0128-0129, U+0168-0169,\n    U+01A0-01A1, U+01AF-01B0, U+0300-0301, U+0303-0304, U+0308-0309, U+0323,\n    U+0329, U+1EA0-1EF9, U+20AB;\n}\n/* latin-ext */\n@font-face {\n  font-family: "Inter";\n  font-style: normal;\n  font-weight: 100 900;\n  font-display: swap;\n  src: url(https://fonts.scalar.com/inter-latin-ext.woff2) format("woff2");\n  unicode-range: U+0100-02AF, U+0304, U+0308, U+0329, U+1E00-1E9F, U+1EF2-1EFF, U+2020, U+20A0-20AB, U+20AD-20C0, U+2113,\n    U+2C60-2C7F, U+A720-A7FF;\n}\n/* latin */\n@font-face {\n  font-family: "Inter";\n  font-style: normal;\n  font-weight: 100 900;\n  font-display: swap;\n  src: url(https://fonts.scalar.com/inter-latin.woff2) format("woff2");\n  unicode-range:\n    U+0000-00FF, U+0131, U+0152-0153, U+02BB-02BC, U+02C6, U+02DA,\n    U+02DC, U+0304, U+0308, U+0329, U+2000-206F, U+2074, U+20AC, U+2122, U+2191,\n    U+2193, U+2212, U+2215, U+FEFF, U+FFFD;\n}\n/* JetBrains Mono (--scalar-font-code) */\n/* cyrillic-ext */\n@font-face {\n  font-family: "JetBrains Mono";\n  font-style: normal;\n  font-weight: 400;\n  src: url(https://fonts.scalar.com/mono-cyrillic-ext.woff2) format("woff2");\n  unicode-range: U+0460-052F, U+1C80-1C88, U+20B4, U+2DE0-2DFF, U+A640-A69F, U+FE2E-FE2F;\n}\n/* cyrillic */\n@font-face {\n  font-family: "JetBrains Mono";\n  font-style: normal;\n  font-weight: 400;\n  src: url(https://fonts.scalar.com/mono-cyrillic.woff2) format("woff2");\n  unicode-range: U+0301, U+0400-045F, U+0490-0491, U+04B0-04B1, U+2116;\n}\n/* greek */\n@font-face {\n  font-family: "JetBrains Mono";\n  font-style: normal;\n  font-weight: 400;\n  src: url(https://fonts.scalar.com/mono-greek.woff2) format("woff2");\n  unicode-range: U+0370-0377, U+037A-037F, U+0384-038A, U+038C, U+038E-03A1, U+03A3-03FF;\n}\n/* vietnamese */\n@font-face {\n  font-family: "JetBrains Mono";\n  font-style: normal;\n  font-weight: 400;\n  src: url(https://fonts.scalar.com/mono-vietnamese.woff2) format("woff2");\n  unicode-range:\n    U+0102-0103, U+0110-0111, U+0128-0129, U+0168-0169,\n    U+01A0-01A1, U+01AF-01B0, U+0300-0301, U+0303-0304, U+0308-0309, U+0323,\n    U+0329, U+1EA0-1EF9, U+20AB;\n}\n/* latin-ext */\n@font-face {\n  font-family: "JetBrains Mono";\n  font-style: normal;\n  font-weight: 400;\n  src: url(https://fonts.scalar.com/mono-latin-ext.woff2) format("woff2");\n  unicode-range: U+0100-02AF, U+0304, U+0308, U+0329, U+1E00-1E9F, U+1EF2-1EFF, U+2020, U+20A0-20AB, U+20AD-20C0, U+2113,\n    U+2C60-2C7F, U+A720-A7FF;\n}\n/* latin */\n@font-face {\n  font-family: "JetBrains Mono";\n  font-style: normal;\n  font-weight: 400;\n  src: url(https://fonts.scalar.com/mono-latin.woff2) format("woff2");\n  unicode-range:\n    U+0000-00FF, U+0131, U+0152-0153, U+02BB-02BC, U+02C6, U+02DA,\n    U+02DC, U+0304, U+0308, U+0329, U+2000-206F, U+2074, U+20AC, U+2122, U+2191,\n    U+2193, U+2212, U+2215, U+FEFF, U+FFFD;\n}\n':""].join("");return r?`@layer ${r} {\n${a}}`:a}},1089:(e,t,n)=>{"use strict";n.d(t,{B:()=>a,X:()=>o});var r,a=((r=a||{})[r.First=0]="First",r[r.Previous=1]="Previous",r[r.Next=2]="Next",r[r.Last=3]="Last",r[r.Specific=4]="Specific",r[r.Nothing=5]="Nothing",r);function o(e,t){let n=t.resolveItems();if(n.length<=0)return null;let r=t.resolveActiveIndex(),a=null!=r?r:-1;switch(e.focus){case 0:for(let e=0;e<n.length;++e)if(!t.resolveDisabled(n[e],e,n))return e;return r;case 1:-1===a&&(a=n.length);for(let e=a-1;e>=0;--e)if(!t.resolveDisabled(n[e],e,n))return e;return r;case 2:for(let e=a+1;e<n.length;++e)if(!t.resolveDisabled(n[e],e,n))return e;return r;case 3:for(let e=n.length-1;e>=0;--e)if(!t.resolveDisabled(n[e],e,n))return e;return r;case 4:for(let r=0;r<n.length;++r)if(t.resolveId(n[r],r,n)===e.id)return r;return r;case 5:return null;default:!function(e){throw new Error("Unexpected object: "+e)}(e)}}},1247:(e,t,n)=>{"use strict";var r=n(9982),a=n(6540),o=n(961);function i(e){var t="https://react.dev/errors/"+e;if(1<arguments.length){t+="?args[]="+encodeURIComponent(arguments[1]);for(var n=2;n<arguments.length;n++)t+="&args[]="+encodeURIComponent(arguments[n])}return"Minified React error #"+e+"; visit "+t+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function s(e){return!(!e||1!==e.nodeType&&9!==e.nodeType&&11!==e.nodeType)}var l=Symbol.for("react.element"),c=Symbol.for("react.transitional.element"),u=Symbol.for("react.portal"),d=Symbol.for("react.fragment"),p=Symbol.for("react.strict_mode"),h=Symbol.for("react.profiler"),f=Symbol.for("react.provider"),m=Symbol.for("react.consumer"),g=Symbol.for("react.context"),v=Symbol.for("react.forward_ref"),b=Symbol.for("react.suspense"),y=Symbol.for("react.suspense_list"),w=Symbol.for("react.memo"),O=Symbol.for("react.lazy");Symbol.for("react.scope"),Symbol.for("react.debug_trace_mode");var k=Symbol.for("react.offscreen");Symbol.for("react.legacy_hidden"),Symbol.for("react.tracing_marker");var x=Symbol.for("react.memo_cache_sentinel"),E=Symbol.iterator;function S(e){return null===e||"object"!=typeof e?null:"function"==typeof(e=E&&e[E]||e["@@iterator"])?e:null}var _=Symbol.for("react.client.reference");function C(e){if(null==e)return null;if("function"==typeof e)return e.$$typeof===_?null:e.displayName||e.name||null;if("string"==typeof e)return e;switch(e){case d:return"Fragment";case u:return"Portal";case h:return"Profiler";case p:return"StrictMode";case b:return"Suspense";case y:return"SuspenseList"}if("object"==typeof e)switch(e.$$typeof){case g:return(e.displayName||"Context")+".Provider";case m:return(e._context.displayName||"Context")+".Consumer";case v:var t=e.render;return(e=e.displayName)||(e=""!==(e=t.displayName||t.name||"")?"ForwardRef("+e+")":"ForwardRef"),e;case w:return null!==(t=e.displayName||null)?t:C(e.type)||"Memo";case O:t=e._payload,e=e._init;try{return C(e(t))}catch(e){}}return null}var T,A,R=a.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,P=Object.assign;function D(e){if(void 0===T)try{throw Error()}catch(e){var t=e.stack.trim().match(/\n( *(at )?)/);T=t&&t[1]||"",A=-1<e.stack.indexOf("\n    at")?" (<anonymous>)":-1<e.stack.indexOf("@")?"@unknown:0:0":""}return"\n"+T+e+A}var M=!1;function $(e,t){if(!e||M)return"";M=!0;var n=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{var r={DetermineComponentFrameRoot:function(){try{if(t){var n=function(){throw Error()};if(Object.defineProperty(n.prototype,"props",{set:function(){throw Error()}}),"object"==typeof Reflect&&Reflect.construct){try{Reflect.construct(n,[])}catch(e){var r=e}Reflect.construct(e,[],n)}else{try{n.call()}catch(e){r=e}e.call(n.prototype)}}else{try{throw Error()}catch(e){r=e}(n=e())&&"function"==typeof n.catch&&n.catch((function(){}))}}catch(e){if(e&&r&&"string"==typeof e.stack)return[e.stack,r.stack]}return[null,null]}};r.DetermineComponentFrameRoot.displayName="DetermineComponentFrameRoot";var a=Object.getOwnPropertyDescriptor(r.DetermineComponentFrameRoot,"name");a&&a.configurable&&Object.defineProperty(r.DetermineComponentFrameRoot,"name",{value:"DetermineComponentFrameRoot"});var o=r.DetermineComponentFrameRoot(),i=o[0],s=o[1];if(i&&s){var l=i.split("\n"),c=s.split("\n");for(a=r=0;r<l.length&&!l[r].includes("DetermineComponentFrameRoot");)r++;for(;a<c.length&&!c[a].includes("DetermineComponentFrameRoot");)a++;if(r===l.length||a===c.length)for(r=l.length-1,a=c.length-1;1<=r&&0<=a&&l[r]!==c[a];)a--;for(;1<=r&&0<=a;r--,a--)if(l[r]!==c[a]){if(1!==r||1!==a)do{if(r--,0>--a||l[r]!==c[a]){var u="\n"+l[r].replace(" at new "," at ");return e.displayName&&u.includes("<anonymous>")&&(u=u.replace("<anonymous>",e.displayName)),u}}while(1<=r&&0<=a);break}}}finally{M=!1,Error.prepareStackTrace=n}return(n=e?e.displayName||e.name:"")?D(n):""}function N(e){switch(e.tag){case 26:case 27:case 5:return D(e.type);case 16:return D("Lazy");case 13:return D("Suspense");case 19:return D("SuspenseList");case 0:case 15:return $(e.type,!1);case 11:return $(e.type.render,!1);case 1:return $(e.type,!0);default:return""}}function I(e){try{var t="";do{t+=N(e),e=e.return}while(e);return t}catch(e){return"\nError generating stack: "+e.message+"\n"+e.stack}}function L(e){var t=e,n=e;if(e.alternate)for(;t.return;)t=t.return;else{e=t;do{!!(4098&(t=e).flags)&&(n=t.return),e=t.return}while(e)}return 3===t.tag?n:null}function B(e){if(13===e.tag){var t=e.memoizedState;if(null===t&&null!==(e=e.alternate)&&(t=e.memoizedState),null!==t)return t.dehydrated}return null}function F(e){if(L(e)!==e)throw Error(i(188))}function Q(e){var t=e.tag;if(5===t||26===t||27===t||6===t)return e;for(e=e.child;null!==e;){if(null!==(t=Q(e)))return t;e=e.sibling}return null}var U=Array.isArray,z=o.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,j={pending:!1,data:null,method:null,action:null},W=[],X=-1;function q(e){return{current:e}}function V(e){0>X||(e.current=W[X],W[X]=null,X--)}function H(e,t){X++,W[X]=e.current,e.current=t}var Z=q(null),G=q(null),K=q(null),Y=q(null);function J(e,t){switch(H(K,t),H(G,e),H(Z,null),e=t.nodeType){case 9:case 11:t=(t=t.documentElement)&&(t=t.namespaceURI)?Gu(t):0;break;default:if(t=(e=8===e?t.parentNode:t).tagName,e=e.namespaceURI)t=Ku(e=Gu(e),t);else switch(t){case"svg":t=1;break;case"math":t=2;break;default:t=0}}V(Z),H(Z,t)}function ee(){V(Z),V(G),V(K)}function te(e){null!==e.memoizedState&&H(Y,e);var t=Z.current,n=Ku(t,e.type);t!==n&&(H(G,e),H(Z,n))}function ne(e){G.current===e&&(V(Z),V(G)),Y.current===e&&(V(Y),Id._currentValue=j)}var re=Object.prototype.hasOwnProperty,ae=r.unstable_scheduleCallback,oe=r.unstable_cancelCallback,ie=r.unstable_shouldYield,se=r.unstable_requestPaint,le=r.unstable_now,ce=r.unstable_getCurrentPriorityLevel,ue=r.unstable_ImmediatePriority,de=r.unstable_UserBlockingPriority,pe=r.unstable_NormalPriority,he=r.unstable_LowPriority,fe=r.unstable_IdlePriority,me=r.log,ge=r.unstable_setDisableYieldValue,ve=null,be=null;function ye(e){if("function"==typeof me&&ge(e),be&&"function"==typeof be.setStrictMode)try{be.setStrictMode(ve,e)}catch(e){}}var we=Math.clz32?Math.clz32:function(e){return 0==(e>>>=0)?32:31-(Oe(e)/ke|0)|0},Oe=Math.log,ke=Math.LN2,xe=128,Ee=4194304;function Se(e){var t=42&e;if(0!==t)return t;switch(e&-e){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:return 64;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return 4194176&e;case 4194304:case 8388608:case 16777216:case 33554432:return 62914560&e;case 67108864:return 67108864;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 0;default:return e}}function _e(e,t){var n=e.pendingLanes;if(0===n)return 0;var r=0,a=e.suspendedLanes,o=e.pingedLanes,i=e.warmLanes;e=0!==e.finishedLanes;var s=134217727&n;return 0!==s?0!=(n=s&~a)?r=Se(n):0!=(o&=s)?r=Se(o):e||0!=(i=s&~i)&&(r=Se(i)):0!=(s=n&~a)?r=Se(s):0!==o?r=Se(o):e||0!=(i=n&~i)&&(r=Se(i)),0===r?0:0===t||t===r||t&a||!((a=r&-r)>=(i=t&-t)||32===a&&4194176&i)?r:t}function Ce(e,t){return!(e.pendingLanes&~(e.suspendedLanes&~e.pingedLanes)&t)}function Te(e,t){switch(e){case 1:case 2:case 4:case 8:return t+250;case 16:case 32:case 64:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return t+5e3;default:return-1}}function Ae(){var e=xe;return!(4194176&(xe<<=1))&&(xe=128),e}function Re(){var e=Ee;return!(62914560&(Ee<<=1))&&(Ee=4194304),e}function Pe(e){for(var t=[],n=0;31>n;n++)t.push(e);return t}function De(e,t){e.pendingLanes|=t,268435456!==t&&(e.suspendedLanes=0,e.pingedLanes=0,e.warmLanes=0)}function Me(e,t,n){e.pendingLanes|=t,e.suspendedLanes&=~t;var r=31-we(t);e.entangledLanes|=t,e.entanglements[r]=1073741824|e.entanglements[r]|4194218&n}function $e(e,t){var n=e.entangledLanes|=t;for(e=e.entanglements;n;){var r=31-we(n),a=1<<r;a&t|e[r]&t&&(e[r]|=t),n&=~a}}function Ne(e){return 2<(e&=-e)?8<e?134217727&e?32:268435456:8:2}function Ie(){var e=z.p;return 0!==e?e:void 0===(e=window.event)?32:Kd(e.type)}var Le=Math.random().toString(36).slice(2),Be="__reactFiber$"+Le,Fe="__reactProps$"+Le,Qe="__reactContainer$"+Le,Ue="__reactEvents$"+Le,ze="__reactListeners$"+Le,je="__reactHandles$"+Le,We="__reactResources$"+Le,Xe="__reactMarker$"+Le;function qe(e){delete e[Be],delete e[Fe],delete e[Ue],delete e[ze],delete e[je]}function Ve(e){var t=e[Be];if(t)return t;for(var n=e.parentNode;n;){if(t=n[Qe]||n[Be]){if(n=t.alternate,null!==t.child||null!==n&&null!==n.child)for(e=ld(e);null!==e;){if(n=e[Be])return n;e=ld(e)}return t}n=(e=n).parentNode}return null}function He(e){if(e=e[Be]||e[Qe]){var t=e.tag;if(5===t||6===t||13===t||26===t||27===t||3===t)return e}return null}function Ze(e){var t=e.tag;if(5===t||26===t||27===t||6===t)return e.stateNode;throw Error(i(33))}function Ge(e){var t=e[We];return t||(t=e[We]={hoistableStyles:new Map,hoistableScripts:new Map}),t}function Ke(e){e[Xe]=!0}var Ye=new Set,Je={};function et(e,t){tt(e,t),tt(e+"Capture",t)}function tt(e,t){for(Je[e]=t,e=0;e<t.length;e++)Ye.add(t[e])}var nt=!("undefined"==typeof window||void 0===window.document||void 0===window.document.createElement),rt=RegExp("^[:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD][:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD\\-.0-9\\u00B7\\u0300-\\u036F\\u203F-\\u2040]*$"),at={},ot={};function it(e,t,n){if(a=t,re.call(ot,a)||!re.call(at,a)&&(rt.test(a)?ot[a]=!0:(at[a]=!0,0)))if(null===n)e.removeAttribute(t);else{switch(typeof n){case"undefined":case"function":case"symbol":return void e.removeAttribute(t);case"boolean":var r=t.toLowerCase().slice(0,5);if("data-"!==r&&"aria-"!==r)return void e.removeAttribute(t)}e.setAttribute(t,""+n)}var a}function st(e,t,n){if(null===n)e.removeAttribute(t);else{switch(typeof n){case"undefined":case"function":case"symbol":case"boolean":return void e.removeAttribute(t)}e.setAttribute(t,""+n)}}function lt(e,t,n,r){if(null===r)e.removeAttribute(n);else{switch(typeof r){case"undefined":case"function":case"symbol":case"boolean":return void e.removeAttribute(n)}e.setAttributeNS(t,n,""+r)}}function ct(e){switch(typeof e){case"bigint":case"boolean":case"number":case"string":case"undefined":case"object":return e;default:return""}}function ut(e){var t=e.type;return(e=e.nodeName)&&"input"===e.toLowerCase()&&("checkbox"===t||"radio"===t)}function dt(e){e._valueTracker||(e._valueTracker=function(e){var t=ut(e)?"checked":"value",n=Object.getOwnPropertyDescriptor(e.constructor.prototype,t),r=""+e[t];if(!e.hasOwnProperty(t)&&void 0!==n&&"function"==typeof n.get&&"function"==typeof n.set){var a=n.get,o=n.set;return Object.defineProperty(e,t,{configurable:!0,get:function(){return a.call(this)},set:function(e){r=""+e,o.call(this,e)}}),Object.defineProperty(e,t,{enumerable:n.enumerable}),{getValue:function(){return r},setValue:function(e){r=""+e},stopTracking:function(){e._valueTracker=null,delete e[t]}}}}(e))}function pt(e){if(!e)return!1;var t=e._valueTracker;if(!t)return!0;var n=t.getValue(),r="";return e&&(r=ut(e)?e.checked?"true":"false":e.value),(e=r)!==n&&(t.setValue(e),!0)}function ht(e){if(void 0===(e=e||("undefined"!=typeof document?document:void 0)))return null;try{return e.activeElement||e.body}catch(t){return e.body}}var ft=/[\n"\\]/g;function mt(e){return e.replace(ft,(function(e){return"\\"+e.charCodeAt(0).toString(16)+" "}))}function gt(e,t,n,r,a,o,i,s){e.name="",null!=i&&"function"!=typeof i&&"symbol"!=typeof i&&"boolean"!=typeof i?e.type=i:e.removeAttribute("type"),null!=t?"number"===i?(0===t&&""===e.value||e.value!=t)&&(e.value=""+ct(t)):e.value!==""+ct(t)&&(e.value=""+ct(t)):"submit"!==i&&"reset"!==i||e.removeAttribute("value"),null!=t?bt(e,i,ct(t)):null!=n?bt(e,i,ct(n)):null!=r&&e.removeAttribute("value"),null==a&&null!=o&&(e.defaultChecked=!!o),null!=a&&(e.checked=a&&"function"!=typeof a&&"symbol"!=typeof a),null!=s&&"function"!=typeof s&&"symbol"!=typeof s&&"boolean"!=typeof s?e.name=""+ct(s):e.removeAttribute("name")}function vt(e,t,n,r,a,o,i,s){if(null!=o&&"function"!=typeof o&&"symbol"!=typeof o&&"boolean"!=typeof o&&(e.type=o),null!=t||null!=n){if(("submit"===o||"reset"===o)&&null==t)return;n=null!=n?""+ct(n):"",t=null!=t?""+ct(t):n,s||t===e.value||(e.value=t),e.defaultValue=t}r="function"!=typeof(r=null!=r?r:a)&&"symbol"!=typeof r&&!!r,e.checked=s?e.checked:!!r,e.defaultChecked=!!r,null!=i&&"function"!=typeof i&&"symbol"!=typeof i&&"boolean"!=typeof i&&(e.name=i)}function bt(e,t,n){"number"===t&&ht(e.ownerDocument)===e||e.defaultValue===""+n||(e.defaultValue=""+n)}function yt(e,t,n,r){if(e=e.options,t){t={};for(var a=0;a<n.length;a++)t["$"+n[a]]=!0;for(n=0;n<e.length;n++)a=t.hasOwnProperty("$"+e[n].value),e[n].selected!==a&&(e[n].selected=a),a&&r&&(e[n].defaultSelected=!0)}else{for(n=""+ct(n),t=null,a=0;a<e.length;a++){if(e[a].value===n)return e[a].selected=!0,void(r&&(e[a].defaultSelected=!0));null!==t||e[a].disabled||(t=e[a])}null!==t&&(t.selected=!0)}}function wt(e,t,n){null==t||((t=""+ct(t))!==e.value&&(e.value=t),null!=n)?e.defaultValue=null!=n?""+ct(n):"":e.defaultValue!==t&&(e.defaultValue=t)}function Ot(e,t,n,r){if(null==t){if(null!=r){if(null!=n)throw Error(i(92));if(U(r)){if(1<r.length)throw Error(i(93));r=r[0]}n=r}null==n&&(n=""),t=n}n=ct(t),e.defaultValue=n,(r=e.textContent)===n&&""!==r&&null!==r&&(e.value=r)}function kt(e,t){if(t){var n=e.firstChild;if(n&&n===e.lastChild&&3===n.nodeType)return void(n.nodeValue=t)}e.textContent=t}var xt=new Set("animationIterationCount aspectRatio borderImageOutset borderImageSlice borderImageWidth boxFlex boxFlexGroup boxOrdinalGroup columnCount columns flex flexGrow flexPositive flexShrink flexNegative flexOrder gridArea gridRow gridRowEnd gridRowSpan gridRowStart gridColumn gridColumnEnd gridColumnSpan gridColumnStart fontWeight lineClamp lineHeight opacity order orphans scale tabSize widows zIndex zoom fillOpacity floodOpacity stopOpacity strokeDasharray strokeDashoffset strokeMiterlimit strokeOpacity strokeWidth MozAnimationIterationCount MozBoxFlex MozBoxFlexGroup MozLineClamp msAnimationIterationCount msFlex msZoom msFlexGrow msFlexNegative msFlexOrder msFlexPositive msFlexShrink msGridColumn msGridColumnSpan msGridRow msGridRowSpan WebkitAnimationIterationCount WebkitBoxFlex WebKitBoxFlexGroup WebkitBoxOrdinalGroup WebkitColumnCount WebkitColumns WebkitFlex WebkitFlexGrow WebkitFlexPositive WebkitFlexShrink WebkitLineClamp".split(" "));function Et(e,t,n){var r=0===t.indexOf("--");null==n||"boolean"==typeof n||""===n?r?e.setProperty(t,""):"float"===t?e.cssFloat="":e[t]="":r?e.setProperty(t,n):"number"!=typeof n||0===n||xt.has(t)?"float"===t?e.cssFloat=n:e[t]=(""+n).trim():e[t]=n+"px"}function St(e,t,n){if(null!=t&&"object"!=typeof t)throw Error(i(62));if(e=e.style,null!=n){for(var r in n)!n.hasOwnProperty(r)||null!=t&&t.hasOwnProperty(r)||(0===r.indexOf("--")?e.setProperty(r,""):"float"===r?e.cssFloat="":e[r]="");for(var a in t)r=t[a],t.hasOwnProperty(a)&&n[a]!==r&&Et(e,a,r)}else for(var o in t)t.hasOwnProperty(o)&&Et(e,o,t[o])}function _t(e){if(-1===e.indexOf("-"))return!1;switch(e){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}var Ct=new Map([["acceptCharset","accept-charset"],["htmlFor","for"],["httpEquiv","http-equiv"],["crossOrigin","crossorigin"],["accentHeight","accent-height"],["alignmentBaseline","alignment-baseline"],["arabicForm","arabic-form"],["baselineShift","baseline-shift"],["capHeight","cap-height"],["clipPath","clip-path"],["clipRule","clip-rule"],["colorInterpolation","color-interpolation"],["colorInterpolationFilters","color-interpolation-filters"],["colorProfile","color-profile"],["colorRendering","color-rendering"],["dominantBaseline","dominant-baseline"],["enableBackground","enable-background"],["fillOpacity","fill-opacity"],["fillRule","fill-rule"],["floodColor","flood-color"],["floodOpacity","flood-opacity"],["fontFamily","font-family"],["fontSize","font-size"],["fontSizeAdjust","font-size-adjust"],["fontStretch","font-stretch"],["fontStyle","font-style"],["fontVariant","font-variant"],["fontWeight","font-weight"],["glyphName","glyph-name"],["glyphOrientationHorizontal","glyph-orientation-horizontal"],["glyphOrientationVertical","glyph-orientation-vertical"],["horizAdvX","horiz-adv-x"],["horizOriginX","horiz-origin-x"],["imageRendering","image-rendering"],["letterSpacing","letter-spacing"],["lightingColor","lighting-color"],["markerEnd","marker-end"],["markerMid","marker-mid"],["markerStart","marker-start"],["overlinePosition","overline-position"],["overlineThickness","overline-thickness"],["paintOrder","paint-order"],["panose-1","panose-1"],["pointerEvents","pointer-events"],["renderingIntent","rendering-intent"],["shapeRendering","shape-rendering"],["stopColor","stop-color"],["stopOpacity","stop-opacity"],["strikethroughPosition","strikethrough-position"],["strikethroughThickness","strikethrough-thickness"],["strokeDasharray","stroke-dasharray"],["strokeDashoffset","stroke-dashoffset"],["strokeLinecap","stroke-linecap"],["strokeLinejoin","stroke-linejoin"],["strokeMiterlimit","stroke-miterlimit"],["strokeOpacity","stroke-opacity"],["strokeWidth","stroke-width"],["textAnchor","text-anchor"],["textDecoration","text-decoration"],["textRendering","text-rendering"],["transformOrigin","transform-origin"],["underlinePosition","underline-position"],["underlineThickness","underline-thickness"],["unicodeBidi","unicode-bidi"],["unicodeRange","unicode-range"],["unitsPerEm","units-per-em"],["vAlphabetic","v-alphabetic"],["vHanging","v-hanging"],["vIdeographic","v-ideographic"],["vMathematical","v-mathematical"],["vectorEffect","vector-effect"],["vertAdvY","vert-adv-y"],["vertOriginX","vert-origin-x"],["vertOriginY","vert-origin-y"],["wordSpacing","word-spacing"],["writingMode","writing-mode"],["xmlnsXlink","xmlns:xlink"],["xHeight","x-height"]]),Tt=/^[\u0000-\u001F ]*j[\r\n\t]*a[\r\n\t]*v[\r\n\t]*a[\r\n\t]*s[\r\n\t]*c[\r\n\t]*r[\r\n\t]*i[\r\n\t]*p[\r\n\t]*t[\r\n\t]*:/i;function At(e){return Tt.test(""+e)?"javascript:throw new Error('React has blocked a javascript: URL as a security precaution.')":e}var Rt=null;function Pt(e){return(e=e.target||e.srcElement||window).correspondingUseElement&&(e=e.correspondingUseElement),3===e.nodeType?e.parentNode:e}var Dt=null,Mt=null;function $t(e){var t=He(e);if(t&&(e=t.stateNode)){var n=e[Fe]||null;e:switch(e=t.stateNode,t.type){case"input":if(gt(e,n.value,n.defaultValue,n.defaultValue,n.checked,n.defaultChecked,n.type,n.name),t=n.name,"radio"===n.type&&null!=t){for(n=e;n.parentNode;)n=n.parentNode;for(n=n.querySelectorAll('input[name="'+mt(""+t)+'"][type="radio"]'),t=0;t<n.length;t++){var r=n[t];if(r!==e&&r.form===e.form){var a=r[Fe]||null;if(!a)throw Error(i(90));gt(r,a.value,a.defaultValue,a.defaultValue,a.checked,a.defaultChecked,a.type,a.name)}}for(t=0;t<n.length;t++)(r=n[t]).form===e.form&&pt(r)}break e;case"textarea":wt(e,n.value,n.defaultValue);break e;case"select":null!=(t=n.value)&&yt(e,!!n.multiple,t,!1)}}}var Nt=!1;function It(e,t,n){if(Nt)return e(t,n);Nt=!0;try{return e(t)}finally{if(Nt=!1,(null!==Dt||null!==Mt)&&(Bc(),Dt&&(t=Dt,e=Mt,Mt=Dt=null,$t(t),e)))for(t=0;t<e.length;t++)$t(e[t])}}function Lt(e,t){var n=e.stateNode;if(null===n)return null;var r=n[Fe]||null;if(null===r)return null;n=r[t];e:switch(t){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":case"onMouseEnter":(r=!r.disabled)||(r=!("button"===(e=e.type)||"input"===e||"select"===e||"textarea"===e)),e=!r;break e;default:e=!1}if(e)return null;if(n&&"function"!=typeof n)throw Error(i(231,t,typeof n));return n}var Bt=!1;if(nt)try{var Ft={};Object.defineProperty(Ft,"passive",{get:function(){Bt=!0}}),window.addEventListener("test",Ft,Ft),window.removeEventListener("test",Ft,Ft)}catch(e){Bt=!1}var Qt=null,Ut=null,zt=null;function jt(){if(zt)return zt;var e,t,n=Ut,r=n.length,a="value"in Qt?Qt.value:Qt.textContent,o=a.length;for(e=0;e<r&&n[e]===a[e];e++);var i=r-e;for(t=1;t<=i&&n[r-t]===a[o-t];t++);return zt=a.slice(e,1<t?1-t:void 0)}function Wt(e){var t=e.keyCode;return"charCode"in e?0===(e=e.charCode)&&13===t&&(e=13):e=t,10===e&&(e=13),32<=e||13===e?e:0}function Xt(){return!0}function qt(){return!1}function Vt(e){function t(t,n,r,a,o){for(var i in this._reactName=t,this._targetInst=r,this.type=n,this.nativeEvent=a,this.target=o,this.currentTarget=null,e)e.hasOwnProperty(i)&&(t=e[i],this[i]=t?t(a):a[i]);return this.isDefaultPrevented=(null!=a.defaultPrevented?a.defaultPrevented:!1===a.returnValue)?Xt:qt,this.isPropagationStopped=qt,this}return P(t.prototype,{preventDefault:function(){this.defaultPrevented=!0;var e=this.nativeEvent;e&&(e.preventDefault?e.preventDefault():"unknown"!=typeof e.returnValue&&(e.returnValue=!1),this.isDefaultPrevented=Xt)},stopPropagation:function(){var e=this.nativeEvent;e&&(e.stopPropagation?e.stopPropagation():"unknown"!=typeof e.cancelBubble&&(e.cancelBubble=!0),this.isPropagationStopped=Xt)},persist:function(){},isPersistent:Xt}),t}var Ht,Zt,Gt,Kt={eventPhase:0,bubbles:0,cancelable:0,timeStamp:function(e){return e.timeStamp||Date.now()},defaultPrevented:0,isTrusted:0},Yt=Vt(Kt),Jt=P({},Kt,{view:0,detail:0}),en=Vt(Jt),tn=P({},Jt,{screenX:0,screenY:0,clientX:0,clientY:0,pageX:0,pageY:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,getModifierState:hn,button:0,buttons:0,relatedTarget:function(e){return void 0===e.relatedTarget?e.fromElement===e.srcElement?e.toElement:e.fromElement:e.relatedTarget},movementX:function(e){return"movementX"in e?e.movementX:(e!==Gt&&(Gt&&"mousemove"===e.type?(Ht=e.screenX-Gt.screenX,Zt=e.screenY-Gt.screenY):Zt=Ht=0,Gt=e),Ht)},movementY:function(e){return"movementY"in e?e.movementY:Zt}}),nn=Vt(tn),rn=Vt(P({},tn,{dataTransfer:0})),an=Vt(P({},Jt,{relatedTarget:0})),on=Vt(P({},Kt,{animationName:0,elapsedTime:0,pseudoElement:0})),sn=Vt(P({},Kt,{clipboardData:function(e){return"clipboardData"in e?e.clipboardData:window.clipboardData}})),ln=Vt(P({},Kt,{data:0})),cn={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},un={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},dn={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function pn(e){var t=this.nativeEvent;return t.getModifierState?t.getModifierState(e):!!(e=dn[e])&&!!t[e]}function hn(){return pn}var fn=Vt(P({},Jt,{key:function(e){if(e.key){var t=cn[e.key]||e.key;if("Unidentified"!==t)return t}return"keypress"===e.type?13===(e=Wt(e))?"Enter":String.fromCharCode(e):"keydown"===e.type||"keyup"===e.type?un[e.keyCode]||"Unidentified":""},code:0,location:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,repeat:0,locale:0,getModifierState:hn,charCode:function(e){return"keypress"===e.type?Wt(e):0},keyCode:function(e){return"keydown"===e.type||"keyup"===e.type?e.keyCode:0},which:function(e){return"keypress"===e.type?Wt(e):"keydown"===e.type||"keyup"===e.type?e.keyCode:0}})),mn=Vt(P({},tn,{pointerId:0,width:0,height:0,pressure:0,tangentialPressure:0,tiltX:0,tiltY:0,twist:0,pointerType:0,isPrimary:0})),gn=Vt(P({},Jt,{touches:0,targetTouches:0,changedTouches:0,altKey:0,metaKey:0,ctrlKey:0,shiftKey:0,getModifierState:hn})),vn=Vt(P({},Kt,{propertyName:0,elapsedTime:0,pseudoElement:0})),bn=Vt(P({},tn,{deltaX:function(e){return"deltaX"in e?e.deltaX:"wheelDeltaX"in e?-e.wheelDeltaX:0},deltaY:function(e){return"deltaY"in e?e.deltaY:"wheelDeltaY"in e?-e.wheelDeltaY:"wheelDelta"in e?-e.wheelDelta:0},deltaZ:0,deltaMode:0})),yn=Vt(P({},Kt,{newState:0,oldState:0})),wn=[9,13,27,32],On=nt&&"CompositionEvent"in window,kn=null;nt&&"documentMode"in document&&(kn=document.documentMode);var xn=nt&&"TextEvent"in window&&!kn,En=nt&&(!On||kn&&8<kn&&11>=kn),Sn=String.fromCharCode(32),_n=!1;function Cn(e,t){switch(e){case"keyup":return-1!==wn.indexOf(t.keyCode);case"keydown":return 229!==t.keyCode;case"keypress":case"mousedown":case"focusout":return!0;default:return!1}}function Tn(e){return"object"==typeof(e=e.detail)&&"data"in e?e.data:null}var An=!1,Rn={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function Pn(e){var t=e&&e.nodeName&&e.nodeName.toLowerCase();return"input"===t?!!Rn[e.type]:"textarea"===t}function Dn(e,t,n,r){Dt?Mt?Mt.push(r):Mt=[r]:Dt=r,0<(t=Iu(t,"onChange")).length&&(n=new Yt("onChange","change",null,n,r),e.push({event:n,listeners:t}))}var Mn=null,$n=null;function Nn(e){Tu(e,0)}function In(e){if(pt(Ze(e)))return e}function Ln(e,t){if("change"===e)return t}var Bn=!1;if(nt){var Fn;if(nt){var Qn="oninput"in document;if(!Qn){var Un=document.createElement("div");Un.setAttribute("oninput","return;"),Qn="function"==typeof Un.oninput}Fn=Qn}else Fn=!1;Bn=Fn&&(!document.documentMode||9<document.documentMode)}function zn(){Mn&&(Mn.detachEvent("onpropertychange",jn),$n=Mn=null)}function jn(e){if("value"===e.propertyName&&In($n)){var t=[];Dn(t,$n,e,Pt(e)),It(Nn,t)}}function Wn(e,t,n){"focusin"===e?(zn(),$n=n,(Mn=t).attachEvent("onpropertychange",jn)):"focusout"===e&&zn()}function Xn(e){if("selectionchange"===e||"keyup"===e||"keydown"===e)return In($n)}function qn(e,t){if("click"===e)return In(t)}function Vn(e,t){if("input"===e||"change"===e)return In(t)}var Hn="function"==typeof Object.is?Object.is:function(e,t){return e===t&&(0!==e||1/e==1/t)||e!=e&&t!=t};function Zn(e,t){if(Hn(e,t))return!0;if("object"!=typeof e||null===e||"object"!=typeof t||null===t)return!1;var n=Object.keys(e),r=Object.keys(t);if(n.length!==r.length)return!1;for(r=0;r<n.length;r++){var a=n[r];if(!re.call(t,a)||!Hn(e[a],t[a]))return!1}return!0}function Gn(e){for(;e&&e.firstChild;)e=e.firstChild;return e}function Kn(e,t){var n,r=Gn(e);for(e=0;r;){if(3===r.nodeType){if(n=e+r.textContent.length,e<=t&&n>=t)return{node:r,offset:t-e};e=n}e:{for(;r;){if(r.nextSibling){r=r.nextSibling;break e}r=r.parentNode}r=void 0}r=Gn(r)}}function Yn(e,t){return!(!e||!t)&&(e===t||(!e||3!==e.nodeType)&&(t&&3===t.nodeType?Yn(e,t.parentNode):"contains"in e?e.contains(t):!!e.compareDocumentPosition&&!!(16&e.compareDocumentPosition(t))))}function Jn(e){for(var t=ht((e=null!=e&&null!=e.ownerDocument&&null!=e.ownerDocument.defaultView?e.ownerDocument.defaultView:window).document);t instanceof e.HTMLIFrameElement;){try{var n="string"==typeof t.contentWindow.location.href}catch(e){n=!1}if(!n)break;t=ht((e=t.contentWindow).document)}return t}function er(e){var t=e&&e.nodeName&&e.nodeName.toLowerCase();return t&&("input"===t&&("text"===e.type||"search"===e.type||"tel"===e.type||"url"===e.type||"password"===e.type)||"textarea"===t||"true"===e.contentEditable)}function tr(e,t){var n=Jn(t);t=e.focusedElem;var r=e.selectionRange;if(n!==t&&t&&t.ownerDocument&&Yn(t.ownerDocument.documentElement,t)){if(null!==r&&er(t))if(e=r.start,void 0===(n=r.end)&&(n=e),"selectionStart"in t)t.selectionStart=e,t.selectionEnd=Math.min(n,t.value.length);else if((n=(e=t.ownerDocument||document)&&e.defaultView||window).getSelection){n=n.getSelection();var a=t.textContent.length,o=Math.min(r.start,a);r=void 0===r.end?o:Math.min(r.end,a),!n.extend&&o>r&&(a=r,r=o,o=a),a=Kn(t,o);var i=Kn(t,r);a&&i&&(1!==n.rangeCount||n.anchorNode!==a.node||n.anchorOffset!==a.offset||n.focusNode!==i.node||n.focusOffset!==i.offset)&&((e=e.createRange()).setStart(a.node,a.offset),n.removeAllRanges(),o>r?(n.addRange(e),n.extend(i.node,i.offset)):(e.setEnd(i.node,i.offset),n.addRange(e)))}for(e=[],n=t;n=n.parentNode;)1===n.nodeType&&e.push({element:n,left:n.scrollLeft,top:n.scrollTop});for("function"==typeof t.focus&&t.focus(),t=0;t<e.length;t++)(n=e[t]).element.scrollLeft=n.left,n.element.scrollTop=n.top}}var nr=nt&&"documentMode"in document&&11>=document.documentMode,rr=null,ar=null,or=null,ir=!1;function sr(e,t,n){var r=n.window===n?n.document:9===n.nodeType?n:n.ownerDocument;ir||null==rr||rr!==ht(r)||(r="selectionStart"in(r=rr)&&er(r)?{start:r.selectionStart,end:r.selectionEnd}:{anchorNode:(r=(r.ownerDocument&&r.ownerDocument.defaultView||window).getSelection()).anchorNode,anchorOffset:r.anchorOffset,focusNode:r.focusNode,focusOffset:r.focusOffset},or&&Zn(or,r)||(or=r,0<(r=Iu(ar,"onSelect")).length&&(t=new Yt("onSelect","select",null,t,n),e.push({event:t,listeners:r}),t.target=rr)))}function lr(e,t){var n={};return n[e.toLowerCase()]=t.toLowerCase(),n["Webkit"+e]="webkit"+t,n["Moz"+e]="moz"+t,n}var cr={animationend:lr("Animation","AnimationEnd"),animationiteration:lr("Animation","AnimationIteration"),animationstart:lr("Animation","AnimationStart"),transitionrun:lr("Transition","TransitionRun"),transitionstart:lr("Transition","TransitionStart"),transitioncancel:lr("Transition","TransitionCancel"),transitionend:lr("Transition","TransitionEnd")},ur={},dr={};function pr(e){if(ur[e])return ur[e];if(!cr[e])return e;var t,n=cr[e];for(t in n)if(n.hasOwnProperty(t)&&t in dr)return ur[e]=n[t];return e}nt&&(dr=document.createElement("div").style,"AnimationEvent"in window||(delete cr.animationend.animation,delete cr.animationiteration.animation,delete cr.animationstart.animation),"TransitionEvent"in window||delete cr.transitionend.transition);var hr=pr("animationend"),fr=pr("animationiteration"),mr=pr("animationstart"),gr=pr("transitionrun"),vr=pr("transitionstart"),br=pr("transitioncancel"),yr=pr("transitionend"),wr=new Map,Or="abort auxClick beforeToggle cancel canPlay canPlayThrough click close contextMenu copy cut drag dragEnd dragEnter dragExit dragLeave dragOver dragStart drop durationChange emptied encrypted ended error gotPointerCapture input invalid keyDown keyPress keyUp load loadedData loadedMetadata loadStart lostPointerCapture mouseDown mouseMove mouseOut mouseOver mouseUp paste pause play playing pointerCancel pointerDown pointerMove pointerOut pointerOver pointerUp progress rateChange reset resize seeked seeking stalled submit suspend timeUpdate touchCancel touchEnd touchStart volumeChange scroll scrollEnd toggle touchMove waiting wheel".split(" ");function kr(e,t){wr.set(e,t),et(t,[e])}var xr=[],Er=0,Sr=0;function _r(){for(var e=Er,t=Sr=Er=0;t<e;){var n=xr[t];xr[t++]=null;var r=xr[t];xr[t++]=null;var a=xr[t];xr[t++]=null;var o=xr[t];if(xr[t++]=null,null!==r&&null!==a){var i=r.pending;null===i?a.next=a:(a.next=i.next,i.next=a),r.pending=a}0!==o&&Rr(n,a,o)}}function Cr(e,t,n,r){xr[Er++]=e,xr[Er++]=t,xr[Er++]=n,xr[Er++]=r,Sr|=r,e.lanes|=r,null!==(e=e.alternate)&&(e.lanes|=r)}function Tr(e,t,n,r){return Cr(e,t,n,r),Pr(e)}function Ar(e,t){return Cr(e,null,null,t),Pr(e)}function Rr(e,t,n){e.lanes|=n;var r=e.alternate;null!==r&&(r.lanes|=n);for(var a=!1,o=e.return;null!==o;)o.childLanes|=n,null!==(r=o.alternate)&&(r.childLanes|=n),22===o.tag&&(null===(e=o.stateNode)||1&e._visibility||(a=!0)),e=o,o=o.return;a&&null!==t&&3===e.tag&&(o=e.stateNode,a=31-we(n),null===(e=(o=o.hiddenUpdates)[a])?o[a]=[t]:e.push(t),t.lane=536870912|n)}function Pr(e){if(50<Tc)throw Tc=0,Ac=null,Error(i(185));for(var t=e.return;null!==t;)t=(e=t).return;return 3===e.tag?e.stateNode:null}var Dr={},Mr=new WeakMap;function $r(e,t){if("object"==typeof e&&null!==e){var n=Mr.get(e);return void 0!==n?n:(t={value:e,source:t,stack:I(t)},Mr.set(e,t),t)}return{value:e,source:t,stack:I(t)}}var Nr=[],Ir=0,Lr=null,Br=0,Fr=[],Qr=0,Ur=null,zr=1,jr="";function Wr(e,t){Nr[Ir++]=Br,Nr[Ir++]=Lr,Lr=e,Br=t}function Xr(e,t,n){Fr[Qr++]=zr,Fr[Qr++]=jr,Fr[Qr++]=Ur,Ur=e;var r=zr;e=jr;var a=32-we(r)-1;r&=~(1<<a),n+=1;var o=32-we(t)+a;if(30<o){var i=a-a%5;o=(r&(1<<i)-1).toString(32),r>>=i,a-=i,zr=1<<32-we(t)+a|n<<a|r,jr=o+e}else zr=1<<o|n<<a|r,jr=e}function qr(e){null!==e.return&&(Wr(e,1),Xr(e,1,0))}function Vr(e){for(;e===Lr;)Lr=Nr[--Ir],Nr[Ir]=null,Br=Nr[--Ir],Nr[Ir]=null;for(;e===Ur;)Ur=Fr[--Qr],Fr[Qr]=null,jr=Fr[--Qr],Fr[Qr]=null,zr=Fr[--Qr],Fr[Qr]=null}var Hr=null,Zr=null,Gr=!1,Kr=null,Yr=!1,Jr=Error(i(519));function ea(e){throw oa($r(Error(i(418,"")),e)),Jr}function ta(e){var t=e.stateNode,n=e.type,r=e.memoizedProps;switch(t[Be]=e,t[Fe]=r,n){case"dialog":Au("cancel",t),Au("close",t);break;case"iframe":case"object":case"embed":Au("load",t);break;case"video":case"audio":for(n=0;n<_u.length;n++)Au(_u[n],t);break;case"source":Au("error",t);break;case"img":case"image":case"link":Au("error",t),Au("load",t);break;case"details":Au("toggle",t);break;case"input":Au("invalid",t),vt(t,r.value,r.defaultValue,r.checked,r.defaultChecked,r.type,r.name,!0),dt(t);break;case"select":Au("invalid",t);break;case"textarea":Au("invalid",t),Ot(t,r.value,r.defaultValue,r.children),dt(t)}"string"!=typeof(n=r.children)&&"number"!=typeof n&&"bigint"!=typeof n||t.textContent===""+n||!0===r.suppressHydrationWarning||zu(t.textContent,n)?(null!=r.popover&&(Au("beforetoggle",t),Au("toggle",t)),null!=r.onScroll&&Au("scroll",t),null!=r.onScrollEnd&&Au("scrollend",t),null!=r.onClick&&(t.onclick=ju),t=!0):t=!1,t||ea(e)}function na(e){for(Hr=e.return;Hr;)switch(Hr.tag){case 3:case 27:return void(Yr=!0);case 5:case 13:return void(Yr=!1);default:Hr=Hr.return}}function ra(e){if(e!==Hr)return!1;if(!Gr)return na(e),Gr=!0,!1;var t,n=!1;if((t=3!==e.tag&&27!==e.tag)&&((t=5===e.tag)&&(t=!("form"!==(t=e.type)&&"button"!==t)||Yu(e.type,e.memoizedProps)),t=!t),t&&(n=!0),n&&Zr&&ea(e),na(e),13===e.tag){if(!(e=null!==(e=e.memoizedState)?e.dehydrated:null))throw Error(i(317));e:{for(e=e.nextSibling,n=0;e;){if(8===e.nodeType)if("/$"===(t=e.data)){if(0===n){Zr=sd(e.nextSibling);break e}n--}else"$"!==t&&"$!"!==t&&"$?"!==t||n++;e=e.nextSibling}Zr=null}}else Zr=Hr?sd(e.stateNode.nextSibling):null;return!0}function aa(){Zr=Hr=null,Gr=!1}function oa(e){null===Kr?Kr=[e]:Kr.push(e)}var ia=Error(i(460)),sa=Error(i(474)),la={then:function(){}};function ca(e){return"fulfilled"===(e=e.status)||"rejected"===e}function ua(){}function da(e,t,n){switch(void 0===(n=e[n])?e.push(t):n!==t&&(t.then(ua,ua),t=n),t.status){case"fulfilled":return t.value;case"rejected":if((e=t.reason)===ia)throw Error(i(483));throw e;default:if("string"==typeof t.status)t.then(ua,ua);else{if(null!==(e=tc)&&100<e.shellSuspendCounter)throw Error(i(482));(e=t).status="pending",e.then((function(e){if("pending"===t.status){var n=t;n.status="fulfilled",n.value=e}}),(function(e){if("pending"===t.status){var n=t;n.status="rejected",n.reason=e}}))}switch(t.status){case"fulfilled":return t.value;case"rejected":if((e=t.reason)===ia)throw Error(i(483));throw e}throw pa=t,ia}}var pa=null;function ha(){if(null===pa)throw Error(i(459));var e=pa;return pa=null,e}var fa=null,ma=0;function ga(e){var t=ma;return ma+=1,null===fa&&(fa=[]),da(fa,e,t)}function va(e,t){t=t.props.ref,e.ref=void 0!==t?t:null}function ba(e,t){if(t.$$typeof===l)throw Error(i(525));throw e=Object.prototype.toString.call(t),Error(i(31,"[object Object]"===e?"object with keys {"+Object.keys(t).join(", ")+"}":e))}function ya(e){return(0,e._init)(e._payload)}function wa(e){function t(t,n){if(e){var r=t.deletions;null===r?(t.deletions=[n],t.flags|=16):r.push(n)}}function n(n,r){if(!e)return null;for(;null!==r;)t(n,r),r=r.sibling;return null}function r(e){for(var t=new Map;null!==e;)null!==e.key?t.set(e.key,e):t.set(e.index,e),e=e.sibling;return t}function a(e,t){return(e=Ll(e,t)).index=0,e.sibling=null,e}function o(t,n,r){return t.index=r,e?null!==(r=t.alternate)?(r=r.index)<n?(t.flags|=33554434,n):r:(t.flags|=33554434,n):(t.flags|=1048576,n)}function s(t){return e&&null===t.alternate&&(t.flags|=33554434),t}function l(e,t,n,r){return null===t||6!==t.tag?((t=zl(n,e.mode,r)).return=e,t):((t=a(t,n)).return=e,t)}function p(e,t,n,r){var o=n.type;return o===d?f(e,t,n.props.children,r,n.key):null!==t&&(t.elementType===o||"object"==typeof o&&null!==o&&o.$$typeof===O&&ya(o)===t.type)?(va(t=a(t,n.props),n),t.return=e,t):(va(t=Fl(n.type,n.key,n.props,null,e.mode,r),n),t.return=e,t)}function h(e,t,n,r){return null===t||4!==t.tag||t.stateNode.containerInfo!==n.containerInfo||t.stateNode.implementation!==n.implementation?((t=jl(n,e.mode,r)).return=e,t):((t=a(t,n.children||[])).return=e,t)}function f(e,t,n,r,o){return null===t||7!==t.tag?((t=Ql(n,e.mode,r,o)).return=e,t):((t=a(t,n)).return=e,t)}function m(e,t,n){if("string"==typeof t&&""!==t||"number"==typeof t||"bigint"==typeof t)return(t=zl(""+t,e.mode,n)).return=e,t;if("object"==typeof t&&null!==t){switch(t.$$typeof){case c:return va(n=Fl(t.type,t.key,t.props,null,e.mode,n),t),n.return=e,n;case u:return(t=jl(t,e.mode,n)).return=e,t;case O:return m(e,t=(0,t._init)(t._payload),n)}if(U(t)||S(t))return(t=Ql(t,e.mode,n,null)).return=e,t;if("function"==typeof t.then)return m(e,ga(t),n);if(t.$$typeof===g)return m(e,_s(e,t),n);ba(e,t)}return null}function v(e,t,n,r){var a=null!==t?t.key:null;if("string"==typeof n&&""!==n||"number"==typeof n||"bigint"==typeof n)return null!==a?null:l(e,t,""+n,r);if("object"==typeof n&&null!==n){switch(n.$$typeof){case c:return n.key===a?p(e,t,n,r):null;case u:return n.key===a?h(e,t,n,r):null;case O:return v(e,t,n=(a=n._init)(n._payload),r)}if(U(n)||S(n))return null!==a?null:f(e,t,n,r,null);if("function"==typeof n.then)return v(e,t,ga(n),r);if(n.$$typeof===g)return v(e,t,_s(e,n),r);ba(e,n)}return null}function b(e,t,n,r,a){if("string"==typeof r&&""!==r||"number"==typeof r||"bigint"==typeof r)return l(t,e=e.get(n)||null,""+r,a);if("object"==typeof r&&null!==r){switch(r.$$typeof){case c:return p(t,e=e.get(null===r.key?n:r.key)||null,r,a);case u:return h(t,e=e.get(null===r.key?n:r.key)||null,r,a);case O:return b(e,t,n,r=(0,r._init)(r._payload),a)}if(U(r)||S(r))return f(t,e=e.get(n)||null,r,a,null);if("function"==typeof r.then)return b(e,t,n,ga(r),a);if(r.$$typeof===g)return b(e,t,n,_s(t,r),a);ba(t,r)}return null}function y(l,p,h,f){if("object"==typeof h&&null!==h&&h.type===d&&null===h.key&&(h=h.props.children),"object"==typeof h&&null!==h){switch(h.$$typeof){case c:e:{for(var w=h.key;null!==p;){if(p.key===w){if((w=h.type)===d){if(7===p.tag){n(l,p.sibling),(f=a(p,h.props.children)).return=l,l=f;break e}}else if(p.elementType===w||"object"==typeof w&&null!==w&&w.$$typeof===O&&ya(w)===p.type){n(l,p.sibling),va(f=a(p,h.props),h),f.return=l,l=f;break e}n(l,p);break}t(l,p),p=p.sibling}h.type===d?((f=Ql(h.props.children,l.mode,f,h.key)).return=l,l=f):(va(f=Fl(h.type,h.key,h.props,null,l.mode,f),h),f.return=l,l=f)}return s(l);case u:e:{for(w=h.key;null!==p;){if(p.key===w){if(4===p.tag&&p.stateNode.containerInfo===h.containerInfo&&p.stateNode.implementation===h.implementation){n(l,p.sibling),(f=a(p,h.children||[])).return=l,l=f;break e}n(l,p);break}t(l,p),p=p.sibling}(f=jl(h,l.mode,f)).return=l,l=f}return s(l);case O:return y(l,p,h=(w=h._init)(h._payload),f)}if(U(h))return function(a,i,s,l){for(var c=null,u=null,d=i,p=i=0,h=null;null!==d&&p<s.length;p++){d.index>p?(h=d,d=null):h=d.sibling;var f=v(a,d,s[p],l);if(null===f){null===d&&(d=h);break}e&&d&&null===f.alternate&&t(a,d),i=o(f,i,p),null===u?c=f:u.sibling=f,u=f,d=h}if(p===s.length)return n(a,d),Gr&&Wr(a,p),c;if(null===d){for(;p<s.length;p++)null!==(d=m(a,s[p],l))&&(i=o(d,i,p),null===u?c=d:u.sibling=d,u=d);return Gr&&Wr(a,p),c}for(d=r(d);p<s.length;p++)null!==(h=b(d,a,p,s[p],l))&&(e&&null!==h.alternate&&d.delete(null===h.key?p:h.key),i=o(h,i,p),null===u?c=h:u.sibling=h,u=h);return e&&d.forEach((function(e){return t(a,e)})),Gr&&Wr(a,p),c}(l,p,h,f);if(S(h)){if("function"!=typeof(w=S(h)))throw Error(i(150));return function(a,s,l,c){if(null==l)throw Error(i(151));for(var u=null,d=null,p=s,h=s=0,f=null,g=l.next();null!==p&&!g.done;h++,g=l.next()){p.index>h?(f=p,p=null):f=p.sibling;var y=v(a,p,g.value,c);if(null===y){null===p&&(p=f);break}e&&p&&null===y.alternate&&t(a,p),s=o(y,s,h),null===d?u=y:d.sibling=y,d=y,p=f}if(g.done)return n(a,p),Gr&&Wr(a,h),u;if(null===p){for(;!g.done;h++,g=l.next())null!==(g=m(a,g.value,c))&&(s=o(g,s,h),null===d?u=g:d.sibling=g,d=g);return Gr&&Wr(a,h),u}for(p=r(p);!g.done;h++,g=l.next())null!==(g=b(p,a,h,g.value,c))&&(e&&null!==g.alternate&&p.delete(null===g.key?h:g.key),s=o(g,s,h),null===d?u=g:d.sibling=g,d=g);return e&&p.forEach((function(e){return t(a,e)})),Gr&&Wr(a,h),u}(l,p,h=w.call(h),f)}if("function"==typeof h.then)return y(l,p,ga(h),f);if(h.$$typeof===g)return y(l,p,_s(l,h),f);ba(l,h)}return"string"==typeof h&&""!==h||"number"==typeof h||"bigint"==typeof h?(h=""+h,null!==p&&6===p.tag?(n(l,p.sibling),(f=a(p,h)).return=l,l=f):(n(l,p),(f=zl(h,l.mode,f)).return=l,l=f),s(l)):n(l,p)}return function(e,t,n,r){try{ma=0;var a=y(e,t,n,r);return fa=null,a}catch(t){if(t===ia)throw t;var o=Nl(29,t,null,e.mode);return o.lanes=r,o.return=e,o}}}var Oa=wa(!0),ka=wa(!1),xa=q(null),Ea=q(0);function Sa(e,t){H(Ea,e=cc),H(xa,t),cc=e|t.baseLanes}function _a(){H(Ea,cc),H(xa,xa.current)}function Ca(){cc=Ea.current,V(xa),V(Ea)}var Ta=q(null),Aa=null;function Ra(e){var t=e.alternate;H($a,1&$a.current),H(Ta,e),null===Aa&&(null===t||null!==xa.current||null!==t.memoizedState)&&(Aa=e)}function Pa(e){if(22===e.tag){if(H($a,$a.current),H(Ta,e),null===Aa){var t=e.alternate;null!==t&&null!==t.memoizedState&&(Aa=e)}}else Da()}function Da(){H($a,$a.current),H(Ta,Ta.current)}function Ma(e){V(Ta),Aa===e&&(Aa=null),V($a)}var $a=q(0);function Na(e){for(var t=e;null!==t;){if(13===t.tag){var n=t.memoizedState;if(null!==n&&(null===(n=n.dehydrated)||"$?"===n.data||"$!"===n.data))return t}else if(19===t.tag&&void 0!==t.memoizedProps.revealOrder){if(128&t.flags)return t}else if(null!==t.child){t.child.return=t,t=t.child;continue}if(t===e)break;for(;null===t.sibling;){if(null===t.return||t.return===e)return null;t=t.return}t.sibling.return=t.return,t=t.sibling}return null}var Ia="undefined"!=typeof AbortController?AbortController:function(){var e=[],t=this.signal={aborted:!1,addEventListener:function(t,n){e.push(n)}};this.abort=function(){t.aborted=!0,e.forEach((function(e){return e()}))}},La=r.unstable_scheduleCallback,Ba=r.unstable_NormalPriority,Fa={$$typeof:g,Consumer:null,Provider:null,_currentValue:null,_currentValue2:null,_threadCount:0};function Qa(){return{controller:new Ia,data:new Map,refCount:0}}function Ua(e){e.refCount--,0===e.refCount&&La(Ba,(function(){e.controller.abort()}))}var za=null,ja=0,Wa=0,Xa=null;function qa(){if(0==--ja&&null!==za){null!==Xa&&(Xa.status="fulfilled");var e=za;za=null,Wa=0,Xa=null;for(var t=0;t<e.length;t++)(0,e[t])()}}var Va=R.S;R.S=function(e,t){"object"==typeof t&&null!==t&&"function"==typeof t.then&&function(e,t){if(null===za){var n=za=[];ja=0,Wa=Ou(),Xa={status:"pending",value:void 0,then:function(e){n.push(e)}}}ja++,t.then(qa,qa)}(0,t),null!==Va&&Va(e,t)};var Ha=q(null);function Za(){var e=Ha.current;return null!==e?e:tc.pooledCache}function Ga(e,t){H(Ha,null===t?Ha.current:t.pool)}function Ka(){var e=Za();return null===e?null:{parent:Fa._currentValue,pool:e}}var Ya=0,Ja=null,eo=null,to=null,no=!1,ro=!1,ao=!1,oo=0,io=0,so=null,lo=0;function co(){throw Error(i(321))}function uo(e,t){if(null===t)return!1;for(var n=0;n<t.length&&n<e.length;n++)if(!Hn(e[n],t[n]))return!1;return!0}function po(e,t,n,r,a,o){return Ya=o,Ja=t,t.memoizedState=null,t.updateQueue=null,t.lanes=0,R.H=null===e||null===e.memoizedState?_i:Ci,ao=!1,o=n(r,a),ao=!1,ro&&(o=fo(t,n,r,a)),ho(e),o}function ho(e){R.H=Si;var t=null!==eo&&null!==eo.next;if(Ya=0,to=eo=Ja=null,no=!1,io=0,so=null,t)throw Error(i(300));null===e||Wi||null!==(e=e.dependencies)&&xs(e)&&(Wi=!0)}function fo(e,t,n,r){Ja=e;var a=0;do{if(ro&&(so=null),io=0,ro=!1,25<=a)throw Error(i(301));if(a+=1,to=eo=null,null!=e.updateQueue){var o=e.updateQueue;o.lastEffect=null,o.events=null,o.stores=null,null!=o.memoCache&&(o.memoCache.index=0)}R.H=Ti,o=t(n,r)}while(ro);return o}function mo(){var e=R.H,t=e.useState()[0];return t="function"==typeof t.then?Oo(t):t,e=e.useState()[0],(null!==eo?eo.memoizedState:null)!==e&&(Ja.flags|=1024),t}function go(){var e=0!==oo;return oo=0,e}function vo(e,t,n){t.updateQueue=e.updateQueue,t.flags&=-2053,e.lanes&=~n}function bo(e){if(no){for(e=e.memoizedState;null!==e;){var t=e.queue;null!==t&&(t.pending=null),e=e.next}no=!1}Ya=0,to=eo=Ja=null,ro=!1,io=oo=0,so=null}function yo(){var e={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return null===to?Ja.memoizedState=to=e:to=to.next=e,to}function wo(){if(null===eo){var e=Ja.alternate;e=null!==e?e.memoizedState:null}else e=eo.next;var t=null===to?Ja.memoizedState:to.next;if(null!==t)to=t,eo=e;else{if(null===e){if(null===Ja.alternate)throw Error(i(467));throw Error(i(310))}e={memoizedState:(eo=e).memoizedState,baseState:eo.baseState,baseQueue:eo.baseQueue,queue:eo.queue,next:null},null===to?Ja.memoizedState=to=e:to=to.next=e}return to}function Oo(e){var t=io;return io+=1,null===so&&(so=[]),e=da(so,e,t),t=Ja,null===(null===to?t.memoizedState:to.next)&&(t=t.alternate,R.H=null===t||null===t.memoizedState?_i:Ci),e}function ko(e){if(null!==e&&"object"==typeof e){if("function"==typeof e.then)return Oo(e);if(e.$$typeof===g)return Ss(e)}throw Error(i(438,String(e)))}function xo(e){var t=null,n=Ja.updateQueue;if(null!==n&&(t=n.memoCache),null==t){var r=Ja.alternate;null!==r&&null!==(r=r.updateQueue)&&null!=(r=r.memoCache)&&(t={data:r.data.map((function(e){return e.slice()})),index:0})}if(null==t&&(t={data:[],index:0}),null===n&&(n={lastEffect:null,events:null,stores:null,memoCache:null},Ja.updateQueue=n),n.memoCache=t,void 0===(n=t.data[t.index]))for(n=t.data[t.index]=Array(e),r=0;r<e;r++)n[r]=x;return t.index++,n}function Eo(e,t){return"function"==typeof t?t(e):t}function So(e){return _o(wo(),eo,e)}function _o(e,t,n){var r=e.queue;if(null===r)throw Error(i(311));r.lastRenderedReducer=n;var a=e.baseQueue,o=r.pending;if(null!==o){if(null!==a){var s=a.next;a.next=o.next,o.next=s}t.baseQueue=a=o,r.pending=null}if(o=e.baseState,null===a)e.memoizedState=o;else{var l=s=null,c=null,u=t=a.next,d=!1;do{var p=-536870913&u.lane;if(p!==u.lane?(rc&p)===p:(Ya&p)===p){var h=u.revertLane;if(0===h)null!==c&&(c=c.next={lane:0,revertLane:0,action:u.action,hasEagerState:u.hasEagerState,eagerState:u.eagerState,next:null}),p===Wa&&(d=!0);else{if((Ya&h)===h){u=u.next,h===Wa&&(d=!0);continue}p={lane:0,revertLane:u.revertLane,action:u.action,hasEagerState:u.hasEagerState,eagerState:u.eagerState,next:null},null===c?(l=c=p,s=o):c=c.next=p,Ja.lanes|=h,dc|=h}p=u.action,ao&&n(o,p),o=u.hasEagerState?u.eagerState:n(o,p)}else h={lane:p,revertLane:u.revertLane,action:u.action,hasEagerState:u.hasEagerState,eagerState:u.eagerState,next:null},null===c?(l=c=h,s=o):c=c.next=h,Ja.lanes|=p,dc|=p;u=u.next}while(null!==u&&u!==t);if(null===c?s=o:c.next=l,!Hn(o,e.memoizedState)&&(Wi=!0,d&&null!==(n=Xa)))throw n;e.memoizedState=o,e.baseState=s,e.baseQueue=c,r.lastRenderedState=o}return null===a&&(r.lanes=0),[e.memoizedState,r.dispatch]}function Co(e){var t=wo(),n=t.queue;if(null===n)throw Error(i(311));n.lastRenderedReducer=e;var r=n.dispatch,a=n.pending,o=t.memoizedState;if(null!==a){n.pending=null;var s=a=a.next;do{o=e(o,s.action),s=s.next}while(s!==a);Hn(o,t.memoizedState)||(Wi=!0),t.memoizedState=o,null===t.baseQueue&&(t.baseState=o),n.lastRenderedState=o}return[o,r]}function To(e,t,n){var r=Ja,a=wo(),o=Gr;if(o){if(void 0===n)throw Error(i(407));n=n()}else n=t();var s=!Hn((eo||a).memoizedState,n);if(s&&(a.memoizedState=n,Wi=!0),a=a.queue,Jo(Po.bind(null,r,a,e),[e]),a.getSnapshot!==t||s||null!==to&&1&to.memoizedState.tag){if(r.flags|=2048,Ho(9,Ro.bind(null,r,a,n,t),{destroy:void 0},null),null===tc)throw Error(i(349));o||60&Ya||Ao(r,t,n)}return n}function Ao(e,t,n){e.flags|=16384,e={getSnapshot:t,value:n},null===(t=Ja.updateQueue)?(t={lastEffect:null,events:null,stores:null,memoCache:null},Ja.updateQueue=t,t.stores=[e]):null===(n=t.stores)?t.stores=[e]:n.push(e)}function Ro(e,t,n,r){t.value=n,t.getSnapshot=r,Do(t)&&Mo(e)}function Po(e,t,n){return n((function(){Do(t)&&Mo(e)}))}function Do(e){var t=e.getSnapshot;e=e.value;try{var n=t();return!Hn(e,n)}catch(e){return!0}}function Mo(e){var t=Ar(e,2);null!==t&&Dc(t,0,2)}function $o(e){var t=yo();if("function"==typeof e){var n=e;if(e=n(),ao){ye(!0);try{n()}finally{ye(!1)}}}return t.memoizedState=t.baseState=e,t.queue={pending:null,lanes:0,dispatch:null,lastRenderedReducer:Eo,lastRenderedState:e},t}function No(e,t,n,r){return e.baseState=n,_o(e,eo,"function"==typeof r?r:Eo)}function Io(e,t,n,r,a){if(ki(e))throw Error(i(485));if(null!==(e=t.action)){var o={payload:a,action:e,next:null,isTransition:!0,status:"pending",value:null,reason:null,listeners:[],then:function(e){o.listeners.push(e)}};null!==R.T?n(!0):o.isTransition=!1,r(o),null===(n=t.pending)?(o.next=t.pending=o,Lo(t,o)):(o.next=n.next,t.pending=n.next=o)}}function Lo(e,t){var n=t.action,r=t.payload,a=e.state;if(t.isTransition){var o=R.T,i={};R.T=i;try{var s=n(a,r),l=R.S;null!==l&&l(i,s),Bo(e,t,s)}catch(n){Qo(e,t,n)}finally{R.T=o}}else try{Bo(e,t,o=n(a,r))}catch(n){Qo(e,t,n)}}function Bo(e,t,n){null!==n&&"object"==typeof n&&"function"==typeof n.then?n.then((function(n){Fo(e,t,n)}),(function(n){return Qo(e,t,n)})):Fo(e,t,n)}function Fo(e,t,n){t.status="fulfilled",t.value=n,Uo(t),e.state=n,null!==(t=e.pending)&&((n=t.next)===t?e.pending=null:(n=n.next,t.next=n,Lo(e,n)))}function Qo(e,t,n){var r=e.pending;if(e.pending=null,null!==r){r=r.next;do{t.status="rejected",t.reason=n,Uo(t),t=t.next}while(t!==r)}e.action=null}function Uo(e){e=e.listeners;for(var t=0;t<e.length;t++)(0,e[t])()}function zo(e,t){return t}function jo(e,t){if(Gr){var n=tc.formState;if(null!==n){e:{var r=Ja;if(Gr){if(Zr){t:{for(var a=Zr,o=Yr;8!==a.nodeType;){if(!o){a=null;break t}if(null===(a=sd(a.nextSibling))){a=null;break t}}a="F!"===(o=a.data)||"F"===o?a:null}if(a){Zr=sd(a.nextSibling),r="F!"===a.data;break e}}ea(r)}r=!1}r&&(t=n[0])}}return(n=yo()).memoizedState=n.baseState=t,r={pending:null,lanes:0,dispatch:null,lastRenderedReducer:zo,lastRenderedState:t},n.queue=r,n=yi.bind(null,Ja,r),r.dispatch=n,r=$o(!1),o=Oi.bind(null,Ja,!1,r.queue),a={state:t,dispatch:null,action:e,pending:null},(r=yo()).queue=a,n=Io.bind(null,Ja,a,o,n),a.dispatch=n,r.memoizedState=e,[t,n,!1]}function Wo(e){return Xo(wo(),eo,e)}function Xo(e,t,n){t=_o(e,t,zo)[0],e=So(Eo)[0],t="object"==typeof t&&null!==t&&"function"==typeof t.then?Oo(t):t;var r=wo(),a=r.queue,o=a.dispatch;return n!==r.memoizedState&&(Ja.flags|=2048,Ho(9,qo.bind(null,a,n),{destroy:void 0},null)),[t,o,e]}function qo(e,t){e.action=t}function Vo(e){var t=wo(),n=eo;if(null!==n)return Xo(t,n,e);wo(),t=t.memoizedState;var r=(n=wo()).queue.dispatch;return n.memoizedState=e,[t,r,!1]}function Ho(e,t,n,r){return e={tag:e,create:t,inst:n,deps:r,next:null},null===(t=Ja.updateQueue)&&(t={lastEffect:null,events:null,stores:null,memoCache:null},Ja.updateQueue=t),null===(n=t.lastEffect)?t.lastEffect=e.next=e:(r=n.next,n.next=e,e.next=r,t.lastEffect=e),e}function Zo(){return wo().memoizedState}function Go(e,t,n,r){var a=yo();Ja.flags|=e,a.memoizedState=Ho(1|t,n,{destroy:void 0},void 0===r?null:r)}function Ko(e,t,n,r){var a=wo();r=void 0===r?null:r;var o=a.memoizedState.inst;null!==eo&&null!==r&&uo(r,eo.memoizedState.deps)?a.memoizedState=Ho(t,n,o,r):(Ja.flags|=e,a.memoizedState=Ho(1|t,n,o,r))}function Yo(e,t){Go(8390656,8,e,t)}function Jo(e,t){Ko(2048,8,e,t)}function ei(e,t){return Ko(4,2,e,t)}function ti(e,t){return Ko(4,4,e,t)}function ni(e,t){if("function"==typeof t){e=e();var n=t(e);return function(){"function"==typeof n?n():t(null)}}if(null!=t)return e=e(),t.current=e,function(){t.current=null}}function ri(e,t,n){n=null!=n?n.concat([e]):null,Ko(4,4,ni.bind(null,t,e),n)}function ai(){}function oi(e,t){var n=wo();t=void 0===t?null:t;var r=n.memoizedState;return null!==t&&uo(t,r[1])?r[0]:(n.memoizedState=[e,t],e)}function ii(e,t){var n=wo();t=void 0===t?null:t;var r=n.memoizedState;if(null!==t&&uo(t,r[1]))return r[0];if(r=e(),ao){ye(!0);try{e()}finally{ye(!1)}}return n.memoizedState=[r,t],r}function si(e,t,n){return void 0===n||1073741824&Ya?e.memoizedState=t:(e.memoizedState=n,e=Pc(),Ja.lanes|=e,dc|=e,n)}function li(e,t,n,r){return Hn(n,t)?n:null!==xa.current?(e=si(e,n,r),Hn(e,t)||(Wi=!0),e):42&Ya?(e=Pc(),Ja.lanes|=e,dc|=e,t):(Wi=!0,e.memoizedState=n)}function ci(e,t,n,r,a){var o=z.p;z.p=0!==o&&8>o?o:8;var i,s,l,c=R.T,u={};R.T=u,Oi(e,!1,t,n);try{var d=a(),p=R.S;null!==p&&p(u,d),null!==d&&"object"==typeof d&&"function"==typeof d.then?wi(e,t,(i=r,s=[],l={status:"pending",value:null,reason:null,then:function(e){s.push(e)}},d.then((function(){l.status="fulfilled",l.value=i;for(var e=0;e<s.length;e++)(0,s[e])(i)}),(function(e){for(l.status="rejected",l.reason=e,e=0;e<s.length;e++)(0,s[e])(void 0)})),l),Rc()):wi(e,t,r,Rc())}catch(n){wi(e,t,{then:function(){},status:"rejected",reason:n},Rc())}finally{z.p=o,R.T=c}}function ui(){}function di(e,t,n,r){if(5!==e.tag)throw Error(i(476));var a=pi(e).queue;ci(e,a,t,j,null===n?ui:function(){return hi(e),n(r)})}function pi(e){var t=e.memoizedState;if(null!==t)return t;var n={};return(t={memoizedState:j,baseState:j,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:Eo,lastRenderedState:j},next:null}).next={memoizedState:n,baseState:n,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:Eo,lastRenderedState:n},next:null},e.memoizedState=t,null!==(e=e.alternate)&&(e.memoizedState=t),t}function hi(e){wi(e,pi(e).next.queue,{},Rc())}function fi(){return Ss(Id)}function mi(){return wo().memoizedState}function gi(){return wo().memoizedState}function vi(e){for(var t=e.return;null!==t;){switch(t.tag){case 24:case 3:var n=Rc(),r=Ds(t,e=Ps(n),n);return null!==r&&(Dc(r,0,n),Ms(r,t,n)),t={cache:Qa()},void(e.payload=t)}t=t.return}}function bi(e,t,n){var r=Rc();n={lane:r,revertLane:0,action:n,hasEagerState:!1,eagerState:null,next:null},ki(e)?xi(t,n):null!==(n=Tr(e,t,n,r))&&(Dc(n,0,r),Ei(n,t,r))}function yi(e,t,n){wi(e,t,n,Rc())}function wi(e,t,n,r){var a={lane:r,revertLane:0,action:n,hasEagerState:!1,eagerState:null,next:null};if(ki(e))xi(t,a);else{var o=e.alternate;if(0===e.lanes&&(null===o||0===o.lanes)&&null!==(o=t.lastRenderedReducer))try{var i=t.lastRenderedState,s=o(i,n);if(a.hasEagerState=!0,a.eagerState=s,Hn(s,i))return Cr(e,t,a,0),null===tc&&_r(),!1}catch(e){}if(null!==(n=Tr(e,t,a,r)))return Dc(n,0,r),Ei(n,t,r),!0}return!1}function Oi(e,t,n,r){if(r={lane:2,revertLane:Ou(),action:r,hasEagerState:!1,eagerState:null,next:null},ki(e)){if(t)throw Error(i(479))}else null!==(t=Tr(e,n,r,2))&&Dc(t,0,2)}function ki(e){var t=e.alternate;return e===Ja||null!==t&&t===Ja}function xi(e,t){ro=no=!0;var n=e.pending;null===n?t.next=t:(t.next=n.next,n.next=t),e.pending=t}function Ei(e,t,n){if(4194176&n){var r=t.lanes;n|=r&=e.pendingLanes,t.lanes=n,$e(e,n)}}var Si={readContext:Ss,use:ko,useCallback:co,useContext:co,useEffect:co,useImperativeHandle:co,useLayoutEffect:co,useInsertionEffect:co,useMemo:co,useReducer:co,useRef:co,useState:co,useDebugValue:co,useDeferredValue:co,useTransition:co,useSyncExternalStore:co,useId:co};Si.useCacheRefresh=co,Si.useMemoCache=co,Si.useHostTransitionStatus=co,Si.useFormState=co,Si.useActionState=co,Si.useOptimistic=co;var _i={readContext:Ss,use:ko,useCallback:function(e,t){return yo().memoizedState=[e,void 0===t?null:t],e},useContext:Ss,useEffect:Yo,useImperativeHandle:function(e,t,n){n=null!=n?n.concat([e]):null,Go(4194308,4,ni.bind(null,t,e),n)},useLayoutEffect:function(e,t){return Go(4194308,4,e,t)},useInsertionEffect:function(e,t){Go(4,2,e,t)},useMemo:function(e,t){var n=yo();t=void 0===t?null:t;var r=e();if(ao){ye(!0);try{e()}finally{ye(!1)}}return n.memoizedState=[r,t],r},useReducer:function(e,t,n){var r=yo();if(void 0!==n){var a=n(t);if(ao){ye(!0);try{n(t)}finally{ye(!1)}}}else a=t;return r.memoizedState=r.baseState=a,e={pending:null,lanes:0,dispatch:null,lastRenderedReducer:e,lastRenderedState:a},r.queue=e,e=e.dispatch=bi.bind(null,Ja,e),[r.memoizedState,e]},useRef:function(e){return e={current:e},yo().memoizedState=e},useState:function(e){var t=(e=$o(e)).queue,n=yi.bind(null,Ja,t);return t.dispatch=n,[e.memoizedState,n]},useDebugValue:ai,useDeferredValue:function(e,t){return si(yo(),e,t)},useTransition:function(){var e=$o(!1);return e=ci.bind(null,Ja,e.queue,!0,!1),yo().memoizedState=e,[!1,e]},useSyncExternalStore:function(e,t,n){var r=Ja,a=yo();if(Gr){if(void 0===n)throw Error(i(407));n=n()}else{if(n=t(),null===tc)throw Error(i(349));60&rc||Ao(r,t,n)}a.memoizedState=n;var o={value:n,getSnapshot:t};return a.queue=o,Yo(Po.bind(null,r,o,e),[e]),r.flags|=2048,Ho(9,Ro.bind(null,r,o,n,t),{destroy:void 0},null),n},useId:function(){var e=yo(),t=tc.identifierPrefix;if(Gr){var n=jr;t=":"+t+"R"+(n=(zr&~(1<<32-we(zr)-1)).toString(32)+n),0<(n=oo++)&&(t+="H"+n.toString(32)),t+=":"}else t=":"+t+"r"+(n=lo++).toString(32)+":";return e.memoizedState=t},useCacheRefresh:function(){return yo().memoizedState=vi.bind(null,Ja)}};_i.useMemoCache=xo,_i.useHostTransitionStatus=fi,_i.useFormState=jo,_i.useActionState=jo,_i.useOptimistic=function(e){var t=yo();t.memoizedState=t.baseState=e;var n={pending:null,lanes:0,dispatch:null,lastRenderedReducer:null,lastRenderedState:null};return t.queue=n,t=Oi.bind(null,Ja,!0,n),n.dispatch=t,[e,t]};var Ci={readContext:Ss,use:ko,useCallback:oi,useContext:Ss,useEffect:Jo,useImperativeHandle:ri,useInsertionEffect:ei,useLayoutEffect:ti,useMemo:ii,useReducer:So,useRef:Zo,useState:function(){return So(Eo)},useDebugValue:ai,useDeferredValue:function(e,t){return li(wo(),eo.memoizedState,e,t)},useTransition:function(){var e=So(Eo)[0],t=wo().memoizedState;return["boolean"==typeof e?e:Oo(e),t]},useSyncExternalStore:To,useId:mi};Ci.useCacheRefresh=gi,Ci.useMemoCache=xo,Ci.useHostTransitionStatus=fi,Ci.useFormState=Wo,Ci.useActionState=Wo,Ci.useOptimistic=function(e,t){return No(wo(),0,e,t)};var Ti={readContext:Ss,use:ko,useCallback:oi,useContext:Ss,useEffect:Jo,useImperativeHandle:ri,useInsertionEffect:ei,useLayoutEffect:ti,useMemo:ii,useReducer:Co,useRef:Zo,useState:function(){return Co(Eo)},useDebugValue:ai,useDeferredValue:function(e,t){var n=wo();return null===eo?si(n,e,t):li(n,eo.memoizedState,e,t)},useTransition:function(){var e=Co(Eo)[0],t=wo().memoizedState;return["boolean"==typeof e?e:Oo(e),t]},useSyncExternalStore:To,useId:mi};function Ai(e,t,n,r){n=null==(n=n(r,t=e.memoizedState))?t:P({},t,n),e.memoizedState=n,0===e.lanes&&(e.updateQueue.baseState=n)}Ti.useCacheRefresh=gi,Ti.useMemoCache=xo,Ti.useHostTransitionStatus=fi,Ti.useFormState=Vo,Ti.useActionState=Vo,Ti.useOptimistic=function(e,t){var n=wo();return null!==eo?No(n,0,e,t):(n.baseState=e,[e,n.queue.dispatch])};var Ri={isMounted:function(e){return!!(e=e._reactInternals)&&L(e)===e},enqueueSetState:function(e,t,n){e=e._reactInternals;var r=Rc(),a=Ps(r);a.payload=t,null!=n&&(a.callback=n),null!==(t=Ds(e,a,r))&&(Dc(t,0,r),Ms(t,e,r))},enqueueReplaceState:function(e,t,n){e=e._reactInternals;var r=Rc(),a=Ps(r);a.tag=1,a.payload=t,null!=n&&(a.callback=n),null!==(t=Ds(e,a,r))&&(Dc(t,0,r),Ms(t,e,r))},enqueueForceUpdate:function(e,t){e=e._reactInternals;var n=Rc(),r=Ps(n);r.tag=2,null!=t&&(r.callback=t),null!==(t=Ds(e,r,n))&&(Dc(t,0,n),Ms(t,e,n))}};function Pi(e,t,n,r,a,o,i){return"function"==typeof(e=e.stateNode).shouldComponentUpdate?e.shouldComponentUpdate(r,o,i):!(t.prototype&&t.prototype.isPureReactComponent&&Zn(n,r)&&Zn(a,o))}function Di(e,t,n,r){e=t.state,"function"==typeof t.componentWillReceiveProps&&t.componentWillReceiveProps(n,r),"function"==typeof t.UNSAFE_componentWillReceiveProps&&t.UNSAFE_componentWillReceiveProps(n,r),t.state!==e&&Ri.enqueueReplaceState(t,t.state,null)}function Mi(e,t){var n=t;if("ref"in t)for(var r in n={},t)"ref"!==r&&(n[r]=t[r]);if(e=e.defaultProps)for(var a in n===t&&(n=P({},n)),e)void 0===n[a]&&(n[a]=e[a]);return n}var $i="function"==typeof reportError?reportError:function(e){if("object"==typeof window&&"function"==typeof window.ErrorEvent){var t=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:"object"==typeof e&&null!==e&&"string"==typeof e.message?String(e.message):String(e),error:e});if(!window.dispatchEvent(t))return}else if("object"==typeof process&&"function"==typeof process.emit)return void process.emit("uncaughtException",e);console.error(e)};function Ni(e){$i(e)}function Ii(e){console.error(e)}function Li(e){$i(e)}function Bi(e,t){try{(0,e.onUncaughtError)(t.value,{componentStack:t.stack})}catch(e){setTimeout((function(){throw e}))}}function Fi(e,t,n){try{(0,e.onCaughtError)(n.value,{componentStack:n.stack,errorBoundary:1===t.tag?t.stateNode:null})}catch(e){setTimeout((function(){throw e}))}}function Qi(e,t,n){return(n=Ps(n)).tag=3,n.payload={element:null},n.callback=function(){Bi(e,t)},n}function Ui(e){return(e=Ps(e)).tag=3,e}function zi(e,t,n,r){var a=n.type.getDerivedStateFromError;if("function"==typeof a){var o=r.value;e.payload=function(){return a(o)},e.callback=function(){Fi(t,n,r)}}var i=n.stateNode;null!==i&&"function"==typeof i.componentDidCatch&&(e.callback=function(){Fi(t,n,r),"function"!=typeof a&&(null===kc?kc=new Set([this]):kc.add(this));var e=r.stack;this.componentDidCatch(r.value,{componentStack:null!==e?e:""})})}var ji=Error(i(461)),Wi=!1;function Xi(e,t,n,r){t.child=null===e?ka(t,null,n,r):Oa(t,e.child,n,r)}function qi(e,t,n,r,a){n=n.render;var o=t.ref;if("ref"in r){var i={};for(var s in r)"ref"!==s&&(i[s]=r[s])}else i=r;return Es(t),r=po(e,t,n,i,o,a),s=go(),null===e||Wi?(Gr&&s&&qr(t),t.flags|=1,Xi(e,t,r,a),t.child):(vo(e,t,a),ps(e,t,a))}function Vi(e,t,n,r,a){if(null===e){var o=n.type;return"function"!=typeof o||Il(o)||void 0!==o.defaultProps||null!==n.compare?((e=Fl(n.type,null,r,t,t.mode,a)).ref=t.ref,e.return=t,t.child=e):(t.tag=15,t.type=o,Hi(e,t,o,r,a))}if(o=e.child,!hs(e,a)){var i=o.memoizedProps;if((n=null!==(n=n.compare)?n:Zn)(i,r)&&e.ref===t.ref)return ps(e,t,a)}return t.flags|=1,(e=Ll(o,r)).ref=t.ref,e.return=t,t.child=e}function Hi(e,t,n,r,a){if(null!==e){var o=e.memoizedProps;if(Zn(o,r)&&e.ref===t.ref){if(Wi=!1,t.pendingProps=r=o,!hs(e,a))return t.lanes=e.lanes,ps(e,t,a);131072&e.flags&&(Wi=!0)}}return Yi(e,t,n,r,a)}function Zi(e,t,n){var r=t.pendingProps,a=r.children,o=!!(2&t.stateNode._pendingVisibility),i=null!==e?e.memoizedState:null;if(Ki(e,t),"hidden"===r.mode||o){if(128&t.flags){if(r=null!==i?i.baseLanes|n:n,null!==e){for(a=t.child=e.child,o=0;null!==a;)o=o|a.lanes|a.childLanes,a=a.sibling;t.childLanes=o&~r}else t.childLanes=0,t.child=null;return Gi(e,t,r,n)}if(!(536870912&n))return t.lanes=t.childLanes=536870912,Gi(e,t,null!==i?i.baseLanes|n:n,n);t.memoizedState={baseLanes:0,cachePool:null},null!==e&&Ga(0,null!==i?i.cachePool:null),null!==i?Sa(t,i):_a(),Pa(t)}else null!==i?(Ga(0,i.cachePool),Sa(t,i),Da(),t.memoizedState=null):(null!==e&&Ga(0,null),_a(),Da());return Xi(e,t,a,n),t.child}function Gi(e,t,n,r){var a=Za();return a=null===a?null:{parent:Fa._currentValue,pool:a},t.memoizedState={baseLanes:n,cachePool:a},null!==e&&Ga(0,null),_a(),Pa(t),null!==e&&ks(e,t,r,!0),null}function Ki(e,t){var n=t.ref;if(null===n)null!==e&&null!==e.ref&&(t.flags|=2097664);else{if("function"!=typeof n&&"object"!=typeof n)throw Error(i(284));null!==e&&e.ref===n||(t.flags|=2097664)}}function Yi(e,t,n,r,a){return Es(t),n=po(e,t,n,r,void 0,a),r=go(),null===e||Wi?(Gr&&r&&qr(t),t.flags|=1,Xi(e,t,n,a),t.child):(vo(e,t,a),ps(e,t,a))}function Ji(e,t,n,r,a,o){return Es(t),t.updateQueue=null,n=fo(t,r,n,a),ho(e),r=go(),null===e||Wi?(Gr&&r&&qr(t),t.flags|=1,Xi(e,t,n,o),t.child):(vo(e,t,o),ps(e,t,o))}function es(e,t,n,r,a){if(Es(t),null===t.stateNode){var o=Dr,i=n.contextType;"object"==typeof i&&null!==i&&(o=Ss(i)),o=new n(r,o),t.memoizedState=null!==o.state&&void 0!==o.state?o.state:null,o.updater=Ri,t.stateNode=o,o._reactInternals=t,(o=t.stateNode).props=r,o.state=t.memoizedState,o.refs={},As(t),i=n.contextType,o.context="object"==typeof i&&null!==i?Ss(i):Dr,o.state=t.memoizedState,"function"==typeof(i=n.getDerivedStateFromProps)&&(Ai(t,n,i,r),o.state=t.memoizedState),"function"==typeof n.getDerivedStateFromProps||"function"==typeof o.getSnapshotBeforeUpdate||"function"!=typeof o.UNSAFE_componentWillMount&&"function"!=typeof o.componentWillMount||(i=o.state,"function"==typeof o.componentWillMount&&o.componentWillMount(),"function"==typeof o.UNSAFE_componentWillMount&&o.UNSAFE_componentWillMount(),i!==o.state&&Ri.enqueueReplaceState(o,o.state,null),Ls(t,r,o,a),Is(),o.state=t.memoizedState),"function"==typeof o.componentDidMount&&(t.flags|=4194308),r=!0}else if(null===e){o=t.stateNode;var s=t.memoizedProps,l=Mi(n,s);o.props=l;var c=o.context,u=n.contextType;i=Dr,"object"==typeof u&&null!==u&&(i=Ss(u));var d=n.getDerivedStateFromProps;u="function"==typeof d||"function"==typeof o.getSnapshotBeforeUpdate,s=t.pendingProps!==s,u||"function"!=typeof o.UNSAFE_componentWillReceiveProps&&"function"!=typeof o.componentWillReceiveProps||(s||c!==i)&&Di(t,o,r,i),Ts=!1;var p=t.memoizedState;o.state=p,Ls(t,r,o,a),Is(),c=t.memoizedState,s||p!==c||Ts?("function"==typeof d&&(Ai(t,n,d,r),c=t.memoizedState),(l=Ts||Pi(t,n,l,r,p,c,i))?(u||"function"!=typeof o.UNSAFE_componentWillMount&&"function"!=typeof o.componentWillMount||("function"==typeof o.componentWillMount&&o.componentWillMount(),"function"==typeof o.UNSAFE_componentWillMount&&o.UNSAFE_componentWillMount()),"function"==typeof o.componentDidMount&&(t.flags|=4194308)):("function"==typeof o.componentDidMount&&(t.flags|=4194308),t.memoizedProps=r,t.memoizedState=c),o.props=r,o.state=c,o.context=i,r=l):("function"==typeof o.componentDidMount&&(t.flags|=4194308),r=!1)}else{o=t.stateNode,Rs(e,t),u=Mi(n,i=t.memoizedProps),o.props=u,d=t.pendingProps,p=o.context,c=n.contextType,l=Dr,"object"==typeof c&&null!==c&&(l=Ss(c)),(c="function"==typeof(s=n.getDerivedStateFromProps)||"function"==typeof o.getSnapshotBeforeUpdate)||"function"!=typeof o.UNSAFE_componentWillReceiveProps&&"function"!=typeof o.componentWillReceiveProps||(i!==d||p!==l)&&Di(t,o,r,l),Ts=!1,p=t.memoizedState,o.state=p,Ls(t,r,o,a),Is();var h=t.memoizedState;i!==d||p!==h||Ts||null!==e&&null!==e.dependencies&&xs(e.dependencies)?("function"==typeof s&&(Ai(t,n,s,r),h=t.memoizedState),(u=Ts||Pi(t,n,u,r,p,h,l)||null!==e&&null!==e.dependencies&&xs(e.dependencies))?(c||"function"!=typeof o.UNSAFE_componentWillUpdate&&"function"!=typeof o.componentWillUpdate||("function"==typeof o.componentWillUpdate&&o.componentWillUpdate(r,h,l),"function"==typeof o.UNSAFE_componentWillUpdate&&o.UNSAFE_componentWillUpdate(r,h,l)),"function"==typeof o.componentDidUpdate&&(t.flags|=4),"function"==typeof o.getSnapshotBeforeUpdate&&(t.flags|=1024)):("function"!=typeof o.componentDidUpdate||i===e.memoizedProps&&p===e.memoizedState||(t.flags|=4),"function"!=typeof o.getSnapshotBeforeUpdate||i===e.memoizedProps&&p===e.memoizedState||(t.flags|=1024),t.memoizedProps=r,t.memoizedState=h),o.props=r,o.state=h,o.context=l,r=u):("function"!=typeof o.componentDidUpdate||i===e.memoizedProps&&p===e.memoizedState||(t.flags|=4),"function"!=typeof o.getSnapshotBeforeUpdate||i===e.memoizedProps&&p===e.memoizedState||(t.flags|=1024),r=!1)}return o=r,Ki(e,t),r=!!(128&t.flags),o||r?(o=t.stateNode,n=r&&"function"!=typeof n.getDerivedStateFromError?null:o.render(),t.flags|=1,null!==e&&r?(t.child=Oa(t,e.child,null,a),t.child=Oa(t,null,n,a)):Xi(e,t,n,a),t.memoizedState=o.state,e=t.child):e=ps(e,t,a),e}function ts(e,t,n,r){return aa(),t.flags|=256,Xi(e,t,n,r),t.child}var ns={dehydrated:null,treeContext:null,retryLane:0};function rs(e){return{baseLanes:e,cachePool:Ka()}}function as(e,t,n){return e=null!==e?e.childLanes&~n:0,t&&(e|=fc),e}function os(e,t,n){var r,a=t.pendingProps,o=!1,s=!!(128&t.flags);if((r=s)||(r=(null===e||null!==e.memoizedState)&&!!(2&$a.current)),r&&(o=!0,t.flags&=-129),r=!!(32&t.flags),t.flags&=-33,null===e){if(Gr){if(o?Ra(t):Da(),Gr){var l,c=Zr;if(l=c){e:{for(l=c,c=Yr;8!==l.nodeType;){if(!c){c=null;break e}if(null===(l=sd(l.nextSibling))){c=null;break e}}c=l}null!==c?(t.memoizedState={dehydrated:c,treeContext:null!==Ur?{id:zr,overflow:jr}:null,retryLane:536870912},(l=Nl(18,null,null,0)).stateNode=c,l.return=t,t.child=l,Hr=t,Zr=null,l=!0):l=!1}l||ea(t)}if(null!==(c=t.memoizedState)&&null!==(c=c.dehydrated))return"$!"===c.data?t.lanes=16:t.lanes=536870912,null;Ma(t)}return c=a.children,a=a.fallback,o?(Da(),c=ss({mode:"hidden",children:c},o=t.mode),a=Ql(a,o,n,null),c.return=t,a.return=t,c.sibling=a,t.child=c,(o=t.child).memoizedState=rs(n),o.childLanes=as(e,r,n),t.memoizedState=ns,a):(Ra(t),is(t,c))}if(null!==(l=e.memoizedState)&&null!==(c=l.dehydrated)){if(s)256&t.flags?(Ra(t),t.flags&=-257,t=ls(e,t,n)):null!==t.memoizedState?(Da(),t.child=e.child,t.flags|=128,t=null):(Da(),o=a.fallback,c=t.mode,a=ss({mode:"visible",children:a.children},c),(o=Ql(o,c,n,null)).flags|=2,a.return=t,o.return=t,a.sibling=o,t.child=a,Oa(t,e.child,null,n),(a=t.child).memoizedState=rs(n),a.childLanes=as(e,r,n),t.memoizedState=ns,t=o);else if(Ra(t),"$!"===c.data){if(r=c.nextSibling&&c.nextSibling.dataset)var u=r.dgst;r=u,(a=Error(i(419))).stack="",a.digest=r,oa({value:a,source:null,stack:null}),t=ls(e,t,n)}else if(Wi||ks(e,t,n,!1),r=!!(n&e.childLanes),Wi||r){if(null!==(r=tc)){if(42&(a=n&-n))a=1;else switch(a){case 2:a=1;break;case 8:a=4;break;case 32:a=16;break;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:a=64;break;case 268435456:a=134217728;break;default:a=0}if(0!==(a=a&(r.suspendedLanes|n)?0:a)&&a!==l.retryLane)throw l.retryLane=a,Ar(e,a),Dc(r,0,a),ji}"$?"===c.data||Wc(),t=ls(e,t,n)}else"$?"===c.data?(t.flags|=128,t.child=e.child,t=su.bind(null,e),c._reactRetry=t,t=null):(e=l.treeContext,Zr=sd(c.nextSibling),Hr=t,Gr=!0,Kr=null,Yr=!1,null!==e&&(Fr[Qr++]=zr,Fr[Qr++]=jr,Fr[Qr++]=Ur,zr=e.id,jr=e.overflow,Ur=t),(t=is(t,a.children)).flags|=4096);return t}return o?(Da(),o=a.fallback,c=t.mode,u=(l=e.child).sibling,(a=Ll(l,{mode:"hidden",children:a.children})).subtreeFlags=31457280&l.subtreeFlags,null!==u?o=Ll(u,o):(o=Ql(o,c,n,null)).flags|=2,o.return=t,a.return=t,a.sibling=o,t.child=a,a=o,o=t.child,null===(c=e.child.memoizedState)?c=rs(n):(null!==(l=c.cachePool)?(u=Fa._currentValue,l=l.parent!==u?{parent:u,pool:u}:l):l=Ka(),c={baseLanes:c.baseLanes|n,cachePool:l}),o.memoizedState=c,o.childLanes=as(e,r,n),t.memoizedState=ns,a):(Ra(t),e=(n=e.child).sibling,(n=Ll(n,{mode:"visible",children:a.children})).return=t,n.sibling=null,null!==e&&(null===(r=t.deletions)?(t.deletions=[e],t.flags|=16):r.push(e)),t.child=n,t.memoizedState=null,n)}function is(e,t){return(t=ss({mode:"visible",children:t},e.mode)).return=e,e.child=t}function ss(e,t){return Ul(e,t,0,null)}function ls(e,t,n){return Oa(t,e.child,null,n),(e=is(t,t.pendingProps.children)).flags|=2,t.memoizedState=null,e}function cs(e,t,n){e.lanes|=t;var r=e.alternate;null!==r&&(r.lanes|=t),ws(e.return,t,n)}function us(e,t,n,r,a){var o=e.memoizedState;null===o?e.memoizedState={isBackwards:t,rendering:null,renderingStartTime:0,last:r,tail:n,tailMode:a}:(o.isBackwards=t,o.rendering=null,o.renderingStartTime=0,o.last=r,o.tail=n,o.tailMode=a)}function ds(e,t,n){var r=t.pendingProps,a=r.revealOrder,o=r.tail;if(Xi(e,t,r.children,n),2&(r=$a.current))r=1&r|2,t.flags|=128;else{if(null!==e&&128&e.flags)e:for(e=t.child;null!==e;){if(13===e.tag)null!==e.memoizedState&&cs(e,n,t);else if(19===e.tag)cs(e,n,t);else if(null!==e.child){e.child.return=e,e=e.child;continue}if(e===t)break e;for(;null===e.sibling;){if(null===e.return||e.return===t)break e;e=e.return}e.sibling.return=e.return,e=e.sibling}r&=1}switch(H($a,r),a){case"forwards":for(n=t.child,a=null;null!==n;)null!==(e=n.alternate)&&null===Na(e)&&(a=n),n=n.sibling;null===(n=a)?(a=t.child,t.child=null):(a=n.sibling,n.sibling=null),us(t,!1,a,n,o);break;case"backwards":for(n=null,a=t.child,t.child=null;null!==a;){if(null!==(e=a.alternate)&&null===Na(e)){t.child=a;break}e=a.sibling,a.sibling=n,n=a,a=e}us(t,!0,n,null,o);break;case"together":us(t,!1,null,null,void 0);break;default:t.memoizedState=null}return t.child}function ps(e,t,n){if(null!==e&&(t.dependencies=e.dependencies),dc|=t.lanes,!(n&t.childLanes)){if(null===e)return null;if(ks(e,t,n,!1),!(n&t.childLanes))return null}if(null!==e&&t.child!==e.child)throw Error(i(153));if(null!==t.child){for(n=Ll(e=t.child,e.pendingProps),t.child=n,n.return=t;null!==e.sibling;)e=e.sibling,(n=n.sibling=Ll(e,e.pendingProps)).return=t;n.sibling=null}return t.child}function hs(e,t){return!!(e.lanes&t)||!(null===(e=e.dependencies)||!xs(e))}function fs(e,t,n){if(null!==e)if(e.memoizedProps!==t.pendingProps)Wi=!0;else{if(!(hs(e,n)||128&t.flags))return Wi=!1,function(e,t,n){switch(t.tag){case 3:J(t,t.stateNode.containerInfo),bs(0,Fa,e.memoizedState.cache),aa();break;case 27:case 5:te(t);break;case 4:J(t,t.stateNode.containerInfo);break;case 10:bs(0,t.type,t.memoizedProps.value);break;case 13:var r=t.memoizedState;if(null!==r)return null!==r.dehydrated?(Ra(t),t.flags|=128,null):n&t.child.childLanes?os(e,t,n):(Ra(t),null!==(e=ps(e,t,n))?e.sibling:null);Ra(t);break;case 19:var a=!!(128&e.flags);if((r=!!(n&t.childLanes))||(ks(e,t,n,!1),r=!!(n&t.childLanes)),a){if(r)return ds(e,t,n);t.flags|=128}if(null!==(a=t.memoizedState)&&(a.rendering=null,a.tail=null,a.lastEffect=null),H($a,$a.current),r)break;return null;case 22:case 23:return t.lanes=0,Zi(e,t,n);case 24:bs(0,Fa,e.memoizedState.cache)}return ps(e,t,n)}(e,t,n);Wi=!!(131072&e.flags)}else Wi=!1,Gr&&1048576&t.flags&&Xr(t,Br,t.index);switch(t.lanes=0,t.tag){case 16:e:{e=t.pendingProps;var r=t.elementType,a=r._init;if(r=a(r._payload),t.type=r,"function"!=typeof r){if(null!=r){if((a=r.$$typeof)===v){t.tag=11,t=qi(null,t,r,e,n);break e}if(a===w){t.tag=14,t=Vi(null,t,r,e,n);break e}}throw t=C(r)||r,Error(i(306,t,""))}Il(r)?(e=Mi(r,e),t.tag=1,t=es(null,t,r,e,n)):(t.tag=0,t=Yi(null,t,r,e,n))}return t;case 0:return Yi(e,t,t.type,t.pendingProps,n);case 1:return es(e,t,r=t.type,a=Mi(r,t.pendingProps),n);case 3:e:{if(J(t,t.stateNode.containerInfo),null===e)throw Error(i(387));var o=t.pendingProps;r=(a=t.memoizedState).element,Rs(e,t),Ls(t,o,null,n);var s=t.memoizedState;if(o=s.cache,bs(0,Fa,o),o!==a.cache&&Os(t,[Fa],n,!0),Is(),o=s.element,a.isDehydrated){if(a={element:o,isDehydrated:!1,cache:s.cache},t.updateQueue.baseState=a,t.memoizedState=a,256&t.flags){t=ts(e,t,o,n);break e}if(o!==r){oa(r=$r(Error(i(424)),t)),t=ts(e,t,o,n);break e}for(Zr=sd(t.stateNode.containerInfo.firstChild),Hr=t,Gr=!0,Kr=null,Yr=!0,n=ka(t,null,o,n),t.child=n;n;)n.flags=-3&n.flags|4096,n=n.sibling}else{if(aa(),o===r){t=ps(e,t,n);break e}Xi(e,t,o,n)}t=t.child}return t;case 26:return Ki(e,t),null===e?(n=gd(t.type,null,t.pendingProps,null))?t.memoizedState=n:Gr||(n=t.type,e=t.pendingProps,(r=Zu(K.current).createElement(n))[Be]=t,r[Fe]=e,qu(r,n,e),Ke(r),t.stateNode=r):t.memoizedState=gd(t.type,e.memoizedProps,t.pendingProps,e.memoizedState),null;case 27:return te(t),null===e&&Gr&&(r=t.stateNode=cd(t.type,t.pendingProps,K.current),Hr=t,Yr=!0,Zr=sd(r.firstChild)),r=t.pendingProps.children,null!==e||Gr?Xi(e,t,r,n):t.child=Oa(t,null,r,n),Ki(e,t),t.child;case 5:return null===e&&Gr&&((a=r=Zr)&&(null!==(r=function(e,t,n,r){for(;1===e.nodeType;){var a=n;if(e.nodeName.toLowerCase()!==t.toLowerCase()){if(!r&&("INPUT"!==e.nodeName||"hidden"!==e.type))break}else if(r){if(!e[Xe])switch(t){case"meta":if(!e.hasAttribute("itemprop"))break;return e;case"link":if("stylesheet"===(o=e.getAttribute("rel"))&&e.hasAttribute("data-precedence"))break;if(o!==a.rel||e.getAttribute("href")!==(null==a.href?null:a.href)||e.getAttribute("crossorigin")!==(null==a.crossOrigin?null:a.crossOrigin)||e.getAttribute("title")!==(null==a.title?null:a.title))break;return e;case"style":if(e.hasAttribute("data-precedence"))break;return e;case"script":if(((o=e.getAttribute("src"))!==(null==a.src?null:a.src)||e.getAttribute("type")!==(null==a.type?null:a.type)||e.getAttribute("crossorigin")!==(null==a.crossOrigin?null:a.crossOrigin))&&o&&e.hasAttribute("async")&&!e.hasAttribute("itemprop"))break;return e;default:return e}}else{if("input"!==t||"hidden"!==e.type)return e;var o=null==a.name?null:""+a.name;if("hidden"===a.type&&e.getAttribute("name")===o)return e}if(null===(e=sd(e.nextSibling)))break}return null}(r,t.type,t.pendingProps,Yr))?(t.stateNode=r,Hr=t,Zr=sd(r.firstChild),Yr=!1,a=!0):a=!1),a||ea(t)),te(t),a=t.type,o=t.pendingProps,s=null!==e?e.memoizedProps:null,r=o.children,Yu(a,o)?r=null:null!==s&&Yu(a,s)&&(t.flags|=32),null!==t.memoizedState&&(a=po(e,t,mo,null,null,n),Id._currentValue=a),Ki(e,t),Xi(e,t,r,n),t.child;case 6:return null===e&&Gr&&((e=n=Zr)&&(null!==(n=function(e,t,n){if(""===t)return null;for(;3!==e.nodeType;){if((1!==e.nodeType||"INPUT"!==e.nodeName||"hidden"!==e.type)&&!n)return null;if(null===(e=sd(e.nextSibling)))return null}return e}(n,t.pendingProps,Yr))?(t.stateNode=n,Hr=t,Zr=null,e=!0):e=!1),e||ea(t)),null;case 13:return os(e,t,n);case 4:return J(t,t.stateNode.containerInfo),r=t.pendingProps,null===e?t.child=Oa(t,null,r,n):Xi(e,t,r,n),t.child;case 11:return qi(e,t,t.type,t.pendingProps,n);case 7:return Xi(e,t,t.pendingProps,n),t.child;case 8:case 12:return Xi(e,t,t.pendingProps.children,n),t.child;case 10:return r=t.pendingProps,bs(0,t.type,r.value),Xi(e,t,r.children,n),t.child;case 9:return a=t.type._context,r=t.pendingProps.children,Es(t),r=r(a=Ss(a)),t.flags|=1,Xi(e,t,r,n),t.child;case 14:return Vi(e,t,t.type,t.pendingProps,n);case 15:return Hi(e,t,t.type,t.pendingProps,n);case 19:return ds(e,t,n);case 22:return Zi(e,t,n);case 24:return Es(t),r=Ss(Fa),null===e?(null===(a=Za())&&(a=tc,o=Qa(),a.pooledCache=o,o.refCount++,null!==o&&(a.pooledCacheLanes|=n),a=o),t.memoizedState={parent:r,cache:a},As(t),bs(0,Fa,a)):(!!(e.lanes&n)&&(Rs(e,t),Ls(t,null,null,n),Is()),a=e.memoizedState,o=t.memoizedState,a.parent!==r?(a={parent:r,cache:r},t.memoizedState=a,0===t.lanes&&(t.memoizedState=t.updateQueue.baseState=a),bs(0,Fa,r)):(r=o.cache,bs(0,Fa,r),r!==a.cache&&Os(t,[Fa],n,!0))),Xi(e,t,t.pendingProps.children,n),t.child;case 29:throw t.pendingProps}throw Error(i(156,t.tag))}var ms=q(null),gs=null,vs=null;function bs(e,t,n){H(ms,t._currentValue),t._currentValue=n}function ys(e){e._currentValue=ms.current,V(ms)}function ws(e,t,n){for(;null!==e;){var r=e.alternate;if((e.childLanes&t)!==t?(e.childLanes|=t,null!==r&&(r.childLanes|=t)):null!==r&&(r.childLanes&t)!==t&&(r.childLanes|=t),e===n)break;e=e.return}}function Os(e,t,n,r){var a=e.child;for(null!==a&&(a.return=e);null!==a;){var o=a.dependencies;if(null!==o){var s=a.child;o=o.firstContext;e:for(;null!==o;){var l=o;o=a;for(var c=0;c<t.length;c++)if(l.context===t[c]){o.lanes|=n,null!==(l=o.alternate)&&(l.lanes|=n),ws(o.return,n,e),r||(s=null);break e}o=l.next}}else if(18===a.tag){if(null===(s=a.return))throw Error(i(341));s.lanes|=n,null!==(o=s.alternate)&&(o.lanes|=n),ws(s,n,e),s=null}else s=a.child;if(null!==s)s.return=a;else for(s=a;null!==s;){if(s===e){s=null;break}if(null!==(a=s.sibling)){a.return=s.return,s=a;break}s=s.return}a=s}}function ks(e,t,n,r){e=null;for(var a=t,o=!1;null!==a;){if(!o)if(524288&a.flags)o=!0;else if(262144&a.flags)break;if(10===a.tag){var s=a.alternate;if(null===s)throw Error(i(387));if(null!==(s=s.memoizedProps)){var l=a.type;Hn(a.pendingProps.value,s.value)||(null!==e?e.push(l):e=[l])}}else if(a===Y.current){if(null===(s=a.alternate))throw Error(i(387));s.memoizedState.memoizedState!==a.memoizedState.memoizedState&&(null!==e?e.push(Id):e=[Id])}a=a.return}null!==e&&Os(t,e,n,r),t.flags|=262144}function xs(e){for(e=e.firstContext;null!==e;){if(!Hn(e.context._currentValue,e.memoizedValue))return!0;e=e.next}return!1}function Es(e){gs=e,vs=null,null!==(e=e.dependencies)&&(e.firstContext=null)}function Ss(e){return Cs(gs,e)}function _s(e,t){return null===gs&&Es(e),Cs(e,t)}function Cs(e,t){var n=t._currentValue;if(t={context:t,memoizedValue:n,next:null},null===vs){if(null===e)throw Error(i(308));vs=t,e.dependencies={lanes:0,firstContext:t},e.flags|=524288}else vs=vs.next=t;return n}var Ts=!1;function As(e){e.updateQueue={baseState:e.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,lanes:0,hiddenCallbacks:null},callbacks:null}}function Rs(e,t){e=e.updateQueue,t.updateQueue===e&&(t.updateQueue={baseState:e.baseState,firstBaseUpdate:e.firstBaseUpdate,lastBaseUpdate:e.lastBaseUpdate,shared:e.shared,callbacks:null})}function Ps(e){return{lane:e,tag:0,payload:null,callback:null,next:null}}function Ds(e,t,n){var r=e.updateQueue;if(null===r)return null;if(r=r.shared,2&ec){var a=r.pending;return null===a?t.next=t:(t.next=a.next,a.next=t),r.pending=t,t=Pr(e),Rr(e,null,n),t}return Cr(e,r,t,n),Pr(e)}function Ms(e,t,n){if(null!==(t=t.updateQueue)&&(t=t.shared,4194176&n)){var r=t.lanes;n|=r&=e.pendingLanes,t.lanes=n,$e(e,n)}}function $s(e,t){var n=e.updateQueue,r=e.alternate;if(null!==r&&n===(r=r.updateQueue)){var a=null,o=null;if(null!==(n=n.firstBaseUpdate)){do{var i={lane:n.lane,tag:n.tag,payload:n.payload,callback:null,next:null};null===o?a=o=i:o=o.next=i,n=n.next}while(null!==n);null===o?a=o=t:o=o.next=t}else a=o=t;return n={baseState:r.baseState,firstBaseUpdate:a,lastBaseUpdate:o,shared:r.shared,callbacks:r.callbacks},void(e.updateQueue=n)}null===(e=n.lastBaseUpdate)?n.firstBaseUpdate=t:e.next=t,n.lastBaseUpdate=t}var Ns=!1;function Is(){if(Ns&&null!==Xa)throw Xa}function Ls(e,t,n,r){Ns=!1;var a=e.updateQueue;Ts=!1;var o=a.firstBaseUpdate,i=a.lastBaseUpdate,s=a.shared.pending;if(null!==s){a.shared.pending=null;var l=s,c=l.next;l.next=null,null===i?o=c:i.next=c,i=l;var u=e.alternate;null!==u&&(s=(u=u.updateQueue).lastBaseUpdate)!==i&&(null===s?u.firstBaseUpdate=c:s.next=c,u.lastBaseUpdate=l)}if(null!==o){var d=a.baseState;for(i=0,u=c=l=null,s=o;;){var p=-536870913&s.lane,h=p!==s.lane;if(h?(rc&p)===p:(r&p)===p){0!==p&&p===Wa&&(Ns=!0),null!==u&&(u=u.next={lane:0,tag:s.tag,payload:s.payload,callback:null,next:null});e:{var f=e,m=s;p=t;var g=n;switch(m.tag){case 1:if("function"==typeof(f=m.payload)){d=f.call(g,d,p);break e}d=f;break e;case 3:f.flags=-65537&f.flags|128;case 0:if(null==(p="function"==typeof(f=m.payload)?f.call(g,d,p):f))break e;d=P({},d,p);break e;case 2:Ts=!0}}null!==(p=s.callback)&&(e.flags|=64,h&&(e.flags|=8192),null===(h=a.callbacks)?a.callbacks=[p]:h.push(p))}else h={lane:p,tag:s.tag,payload:s.payload,callback:s.callback,next:null},null===u?(c=u=h,l=d):u=u.next=h,i|=p;if(null===(s=s.next)){if(null===(s=a.shared.pending))break;s=(h=s).next,h.next=null,a.lastBaseUpdate=h,a.shared.pending=null}}null===u&&(l=d),a.baseState=l,a.firstBaseUpdate=c,a.lastBaseUpdate=u,null===o&&(a.shared.lanes=0),dc|=i,e.lanes=i,e.memoizedState=d}}function Bs(e,t){if("function"!=typeof e)throw Error(i(191,e));e.call(t)}function Fs(e,t){var n=e.callbacks;if(null!==n)for(e.callbacks=null,e=0;e<n.length;e++)Bs(n[e],t)}function Qs(e,t){try{var n=t.updateQueue,r=null!==n?n.lastEffect:null;if(null!==r){var a=r.next;n=a;do{if((n.tag&e)===e){r=void 0;var o=n.create,i=n.inst;r=o(),i.destroy=r}n=n.next}while(n!==a)}}catch(e){ru(t,t.return,e)}}function Us(e,t,n){try{var r=t.updateQueue,a=null!==r?r.lastEffect:null;if(null!==a){var o=a.next;r=o;do{if((r.tag&e)===e){var i=r.inst,s=i.destroy;if(void 0!==s){i.destroy=void 0,a=t;var l=n;try{s()}catch(e){ru(a,l,e)}}}r=r.next}while(r!==o)}}catch(e){ru(t,t.return,e)}}function zs(e){var t=e.updateQueue;if(null!==t){var n=e.stateNode;try{Fs(t,n)}catch(t){ru(e,e.return,t)}}}function js(e,t,n){n.props=Mi(e.type,e.memoizedProps),n.state=e.memoizedState;try{n.componentWillUnmount()}catch(n){ru(e,t,n)}}function Ws(e,t){try{var n=e.ref;if(null!==n){var r=e.stateNode;switch(e.tag){case 26:case 27:case 5:var a=r;break;default:a=r}"function"==typeof n?e.refCleanup=n(a):n.current=a}}catch(n){ru(e,t,n)}}function Xs(e,t){var n=e.ref,r=e.refCleanup;if(null!==n)if("function"==typeof r)try{r()}catch(n){ru(e,t,n)}finally{e.refCleanup=null,null!=(e=e.alternate)&&(e.refCleanup=null)}else if("function"==typeof n)try{n(null)}catch(n){ru(e,t,n)}else n.current=null}function qs(e){var t=e.type,n=e.memoizedProps,r=e.stateNode;try{e:switch(t){case"button":case"input":case"select":case"textarea":n.autoFocus&&r.focus();break e;case"img":n.src?r.src=n.src:n.srcSet&&(r.srcset=n.srcSet)}}catch(t){ru(e,e.return,t)}}function Vs(e,t,n){try{var r=e.stateNode;!function(e,t,n,r){switch(t){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"input":var a=null,o=null,s=null,l=null,c=null,u=null,d=null;for(f in n){var p=n[f];if(n.hasOwnProperty(f)&&null!=p)switch(f){case"checked":case"value":break;case"defaultValue":c=p;default:r.hasOwnProperty(f)||Wu(e,t,f,null,r,p)}}for(var h in r){var f=r[h];if(p=n[h],r.hasOwnProperty(h)&&(null!=f||null!=p))switch(h){case"type":o=f;break;case"name":a=f;break;case"checked":u=f;break;case"defaultChecked":d=f;break;case"value":s=f;break;case"defaultValue":l=f;break;case"children":case"dangerouslySetInnerHTML":if(null!=f)throw Error(i(137,t));break;default:f!==p&&Wu(e,t,h,f,r,p)}}return void gt(e,s,l,c,u,d,o,a);case"select":for(o in f=s=l=h=null,n)if(c=n[o],n.hasOwnProperty(o)&&null!=c)switch(o){case"value":break;case"multiple":f=c;default:r.hasOwnProperty(o)||Wu(e,t,o,null,r,c)}for(a in r)if(o=r[a],c=n[a],r.hasOwnProperty(a)&&(null!=o||null!=c))switch(a){case"value":h=o;break;case"defaultValue":l=o;break;case"multiple":s=o;default:o!==c&&Wu(e,t,a,o,r,c)}return t=l,n=s,r=f,void(null!=h?yt(e,!!n,h,!1):!!r!=!!n&&(null!=t?yt(e,!!n,t,!0):yt(e,!!n,n?[]:"",!1)));case"textarea":for(l in f=h=null,n)if(a=n[l],n.hasOwnProperty(l)&&null!=a&&!r.hasOwnProperty(l))switch(l){case"value":case"children":break;default:Wu(e,t,l,null,r,a)}for(s in r)if(a=r[s],o=n[s],r.hasOwnProperty(s)&&(null!=a||null!=o))switch(s){case"value":h=a;break;case"defaultValue":f=a;break;case"children":break;case"dangerouslySetInnerHTML":if(null!=a)throw Error(i(91));break;default:a!==o&&Wu(e,t,s,a,r,o)}return void wt(e,h,f);case"option":for(var m in n)h=n[m],n.hasOwnProperty(m)&&null!=h&&!r.hasOwnProperty(m)&&("selected"===m?e.selected=!1:Wu(e,t,m,null,r,h));for(c in r)h=r[c],f=n[c],!r.hasOwnProperty(c)||h===f||null==h&&null==f||("selected"===c?e.selected=h&&"function"!=typeof h&&"symbol"!=typeof h:Wu(e,t,c,h,r,f));return;case"img":case"link":case"area":case"base":case"br":case"col":case"embed":case"hr":case"keygen":case"meta":case"param":case"source":case"track":case"wbr":case"menuitem":for(var g in n)h=n[g],n.hasOwnProperty(g)&&null!=h&&!r.hasOwnProperty(g)&&Wu(e,t,g,null,r,h);for(u in r)if(h=r[u],f=n[u],r.hasOwnProperty(u)&&h!==f&&(null!=h||null!=f))switch(u){case"children":case"dangerouslySetInnerHTML":if(null!=h)throw Error(i(137,t));break;default:Wu(e,t,u,h,r,f)}return;default:if(_t(t)){for(var v in n)h=n[v],n.hasOwnProperty(v)&&void 0!==h&&!r.hasOwnProperty(v)&&Xu(e,t,v,void 0,r,h);for(d in r)h=r[d],f=n[d],!r.hasOwnProperty(d)||h===f||void 0===h&&void 0===f||Xu(e,t,d,h,r,f);return}}for(var b in n)h=n[b],n.hasOwnProperty(b)&&null!=h&&!r.hasOwnProperty(b)&&Wu(e,t,b,null,r,h);for(p in r)h=r[p],f=n[p],!r.hasOwnProperty(p)||h===f||null==h&&null==f||Wu(e,t,p,h,r,f)}(r,e.type,n,t),r[Fe]=t}catch(t){ru(e,e.return,t)}}function Hs(e){return 5===e.tag||3===e.tag||26===e.tag||27===e.tag||4===e.tag}function Zs(e){e:for(;;){for(;null===e.sibling;){if(null===e.return||Hs(e.return))return null;e=e.return}for(e.sibling.return=e.return,e=e.sibling;5!==e.tag&&6!==e.tag&&27!==e.tag&&18!==e.tag;){if(2&e.flags)continue e;if(null===e.child||4===e.tag)continue e;e.child.return=e,e=e.child}if(!(2&e.flags))return e.stateNode}}function Gs(e,t,n){var r=e.tag;if(5===r||6===r)e=e.stateNode,t?8===n.nodeType?n.parentNode.insertBefore(e,t):n.insertBefore(e,t):(8===n.nodeType?(t=n.parentNode).insertBefore(e,n):(t=n).appendChild(e),null!=(n=n._reactRootContainer)||null!==t.onclick||(t.onclick=ju));else if(4!==r&&27!==r&&null!==(e=e.child))for(Gs(e,t,n),e=e.sibling;null!==e;)Gs(e,t,n),e=e.sibling}function Ks(e,t,n){var r=e.tag;if(5===r||6===r)e=e.stateNode,t?n.insertBefore(e,t):n.appendChild(e);else if(4!==r&&27!==r&&null!==(e=e.child))for(Ks(e,t,n),e=e.sibling;null!==e;)Ks(e,t,n),e=e.sibling}var Ys=!1,Js=!1,el=!1,tl="function"==typeof WeakSet?WeakSet:Set,nl=null,rl=!1;function al(e,t,n){var r=n.flags;switch(n.tag){case 0:case 11:case 15:vl(e,n),4&r&&Qs(5,n);break;case 1:if(vl(e,n),4&r)if(e=n.stateNode,null===t)try{e.componentDidMount()}catch(e){ru(n,n.return,e)}else{var a=Mi(n.type,t.memoizedProps);t=t.memoizedState;try{e.componentDidUpdate(a,t,e.__reactInternalSnapshotBeforeUpdate)}catch(e){ru(n,n.return,e)}}64&r&&zs(n),512&r&&Ws(n,n.return);break;case 3:if(vl(e,n),64&r&&null!==(r=n.updateQueue)){if(e=null,null!==n.child)switch(n.child.tag){case 27:case 5:case 1:e=n.child.stateNode}try{Fs(r,e)}catch(e){ru(n,n.return,e)}}break;case 26:vl(e,n),512&r&&Ws(n,n.return);break;case 27:case 5:vl(e,n),null===t&&4&r&&qs(n),512&r&&Ws(n,n.return);break;case 12:default:vl(e,n);break;case 13:vl(e,n),4&r&&ul(e,n);break;case 22:if(!(a=null!==n.memoizedState||Ys)){t=null!==t&&null!==t.memoizedState||Js;var o=Ys,i=Js;Ys=a,(Js=t)&&!i?yl(e,n,!!(8772&n.subtreeFlags)):vl(e,n),Ys=o,Js=i}512&r&&("manual"===n.memoizedProps.mode?Ws(n,n.return):Xs(n,n.return))}}function ol(e){var t=e.alternate;null!==t&&(e.alternate=null,ol(t)),e.child=null,e.deletions=null,e.sibling=null,5===e.tag&&null!==(t=e.stateNode)&&qe(t),e.stateNode=null,e.return=null,e.dependencies=null,e.memoizedProps=null,e.memoizedState=null,e.pendingProps=null,e.stateNode=null,e.updateQueue=null}var il=null,sl=!1;function ll(e,t,n){for(n=n.child;null!==n;)cl(e,t,n),n=n.sibling}function cl(e,t,n){if(be&&"function"==typeof be.onCommitFiberUnmount)try{be.onCommitFiberUnmount(ve,n)}catch(e){}switch(n.tag){case 26:Js||Xs(n,t),ll(e,t,n),n.memoizedState?n.memoizedState.count--:n.stateNode&&(n=n.stateNode).parentNode.removeChild(n);break;case 27:Js||Xs(n,t);var r=il,a=sl;for(il=n.stateNode,ll(e,t,n),t=(n=n.stateNode).attributes;t.length;)n.removeAttributeNode(t[0]);qe(n),il=r,sl=a;break;case 5:Js||Xs(n,t);case 6:a=il;var o=sl;if(il=null,ll(e,t,n),sl=o,null!==(il=a))if(sl)try{e=il,r=n.stateNode,8===e.nodeType?e.parentNode.removeChild(r):e.removeChild(r)}catch(e){ru(n,t,e)}else try{il.removeChild(n.stateNode)}catch(e){ru(n,t,e)}break;case 18:null!==il&&(sl?(t=il,n=n.stateNode,8===t.nodeType?od(t.parentNode,n):1===t.nodeType&&od(t,n),mp(t)):od(il,n.stateNode));break;case 4:r=il,a=sl,il=n.stateNode.containerInfo,sl=!0,ll(e,t,n),il=r,sl=a;break;case 0:case 11:case 14:case 15:Js||Us(2,n,t),Js||Us(4,n,t),ll(e,t,n);break;case 1:Js||(Xs(n,t),"function"==typeof(r=n.stateNode).componentWillUnmount&&js(n,t,r)),ll(e,t,n);break;case 21:ll(e,t,n);break;case 22:Js||Xs(n,t),Js=(r=Js)||null!==n.memoizedState,ll(e,t,n),Js=r;break;default:ll(e,t,n)}}function ul(e,t){if(null===t.memoizedState&&null!==(e=t.alternate)&&null!==(e=e.memoizedState)&&null!==(e=e.dehydrated))try{mp(e)}catch(e){ru(t,t.return,e)}}function dl(e,t){var n=function(e){switch(e.tag){case 13:case 19:var t=e.stateNode;return null===t&&(t=e.stateNode=new tl),t;case 22:return null===(t=(e=e.stateNode)._retryCache)&&(t=e._retryCache=new tl),t;default:throw Error(i(435,e.tag))}}(e);t.forEach((function(t){var r=lu.bind(null,e,t);n.has(t)||(n.add(t),t.then(r,r))}))}function pl(e,t){var n=t.deletions;if(null!==n)for(var r=0;r<n.length;r++){var a=n[r],o=e,s=t,l=s;e:for(;null!==l;){switch(l.tag){case 27:case 5:il=l.stateNode,sl=!1;break e;case 3:case 4:il=l.stateNode.containerInfo,sl=!0;break e}l=l.return}if(null===il)throw Error(i(160));cl(o,s,a),il=null,sl=!1,null!==(o=a.alternate)&&(o.return=null),a.return=null}if(13878&t.subtreeFlags)for(t=t.child;null!==t;)fl(t,e),t=t.sibling}var hl=null;function fl(e,t){var n=e.alternate,r=e.flags;switch(e.tag){case 0:case 11:case 14:case 15:pl(t,e),ml(e),4&r&&(Us(3,e,e.return),Qs(3,e),Us(5,e,e.return));break;case 1:pl(t,e),ml(e),512&r&&(Js||null===n||Xs(n,n.return)),64&r&&Ys&&null!==(e=e.updateQueue)&&null!==(r=e.callbacks)&&(n=e.shared.hiddenCallbacks,e.shared.hiddenCallbacks=null===n?r:n.concat(r));break;case 26:var a=hl;if(pl(t,e),ml(e),512&r&&(Js||null===n||Xs(n,n.return)),4&r){var o=null!==n?n.memoizedState:null;if(r=e.memoizedState,null===n)if(null===r)if(null===e.stateNode){e:{r=e.type,n=e.memoizedProps,a=a.ownerDocument||a;t:switch(r){case"title":(!(o=a.getElementsByTagName("title")[0])||o[Xe]||o[Be]||"http://www.w3.org/2000/svg"===o.namespaceURI||o.hasAttribute("itemprop"))&&(o=a.createElement(r),a.head.insertBefore(o,a.querySelector("head > title"))),qu(o,r,n),o[Be]=e,Ke(o),r=o;break e;case"link":var s=Cd("link","href",a).get(r+(n.href||""));if(s)for(var l=0;l<s.length;l++)if((o=s[l]).getAttribute("href")===(null==n.href?null:n.href)&&o.getAttribute("rel")===(null==n.rel?null:n.rel)&&o.getAttribute("title")===(null==n.title?null:n.title)&&o.getAttribute("crossorigin")===(null==n.crossOrigin?null:n.crossOrigin)){s.splice(l,1);break t}qu(o=a.createElement(r),r,n),a.head.appendChild(o);break;case"meta":if(s=Cd("meta","content",a).get(r+(n.content||"")))for(l=0;l<s.length;l++)if((o=s[l]).getAttribute("content")===(null==n.content?null:""+n.content)&&o.getAttribute("name")===(null==n.name?null:n.name)&&o.getAttribute("property")===(null==n.property?null:n.property)&&o.getAttribute("http-equiv")===(null==n.httpEquiv?null:n.httpEquiv)&&o.getAttribute("charset")===(null==n.charSet?null:n.charSet)){s.splice(l,1);break t}qu(o=a.createElement(r),r,n),a.head.appendChild(o);break;default:throw Error(i(468,r))}o[Be]=e,Ke(o),r=o}e.stateNode=r}else Td(a,e.type,e.stateNode);else e.stateNode=kd(a,r,e.memoizedProps);else o!==r?(null===o?null!==n.stateNode&&(n=n.stateNode).parentNode.removeChild(n):o.count--,null===r?Td(a,e.type,e.stateNode):kd(a,r,e.memoizedProps)):null===r&&null!==e.stateNode&&Vs(e,e.memoizedProps,n.memoizedProps)}break;case 27:if(4&r&&null===e.alternate){a=e.stateNode,o=e.memoizedProps;try{for(var c=a.firstChild;c;){var u=c.nextSibling,d=c.nodeName;c[Xe]||"HEAD"===d||"BODY"===d||"SCRIPT"===d||"STYLE"===d||"LINK"===d&&"stylesheet"===c.rel.toLowerCase()||a.removeChild(c),c=u}for(var p=e.type,h=a.attributes;h.length;)a.removeAttributeNode(h[0]);qu(a,p,o),a[Be]=e,a[Fe]=o}catch(t){ru(e,e.return,t)}}case 5:if(pl(t,e),ml(e),512&r&&(Js||null===n||Xs(n,n.return)),32&e.flags){a=e.stateNode;try{kt(a,"")}catch(t){ru(e,e.return,t)}}4&r&&null!=e.stateNode&&Vs(e,a=e.memoizedProps,null!==n?n.memoizedProps:a),1024&r&&(el=!0);break;case 6:if(pl(t,e),ml(e),4&r){if(null===e.stateNode)throw Error(i(162));r=e.memoizedProps,n=e.stateNode;try{n.nodeValue=r}catch(t){ru(e,e.return,t)}}break;case 3:if(_d=null,a=hl,hl=pd(t.containerInfo),pl(t,e),hl=a,ml(e),4&r&&null!==n&&n.memoizedState.isDehydrated)try{mp(t.containerInfo)}catch(t){ru(e,e.return,t)}el&&(el=!1,gl(e));break;case 4:r=hl,hl=pd(e.stateNode.containerInfo),pl(t,e),ml(e),hl=r;break;case 12:pl(t,e),ml(e);break;case 13:pl(t,e),ml(e),8192&e.child.flags&&null!==e.memoizedState!=(null!==n&&null!==n.memoizedState)&&(yc=le()),4&r&&null!==(r=e.updateQueue)&&(e.updateQueue=null,dl(e,r));break;case 22:if(512&r&&(Js||null===n||Xs(n,n.return)),c=null!==e.memoizedState,u=null!==n&&null!==n.memoizedState,Ys=(d=Ys)||c,Js=(p=Js)||u,pl(t,e),Js=p,Ys=d,ml(e),(t=e.stateNode)._current=e,t._visibility&=-3,t._visibility|=2&t._pendingVisibility,8192&r&&(t._visibility=c?-2&t._visibility:1|t._visibility,c&&(t=Ys||Js,null===n||u||t||bl(e)),null===e.memoizedProps||"manual"!==e.memoizedProps.mode))e:for(n=null,t=e;;){if(5===t.tag||26===t.tag||27===t.tag){if(null===n){u=n=t;try{if(a=u.stateNode,c)"function"==typeof(o=a.style).setProperty?o.setProperty("display","none","important"):o.display="none";else{s=u.stateNode;var f=null!=(l=u.memoizedProps.style)&&l.hasOwnProperty("display")?l.display:null;s.style.display=null==f||"boolean"==typeof f?"":(""+f).trim()}}catch(e){ru(u,u.return,e)}}}else if(6===t.tag){if(null===n){u=t;try{u.stateNode.nodeValue=c?"":u.memoizedProps}catch(e){ru(u,u.return,e)}}}else if((22!==t.tag&&23!==t.tag||null===t.memoizedState||t===e)&&null!==t.child){t.child.return=t,t=t.child;continue}if(t===e)break e;for(;null===t.sibling;){if(null===t.return||t.return===e)break e;n===t&&(n=null),t=t.return}n===t&&(n=null),t.sibling.return=t.return,t=t.sibling}4&r&&null!==(r=e.updateQueue)&&null!==(n=r.retryQueue)&&(r.retryQueue=null,dl(e,n));break;case 19:pl(t,e),ml(e),4&r&&null!==(r=e.updateQueue)&&(e.updateQueue=null,dl(e,r));break;case 21:break;default:pl(t,e),ml(e)}}function ml(e){var t=e.flags;if(2&t){try{if(27!==e.tag){e:{for(var n=e.return;null!==n;){if(Hs(n)){var r=n;break e}n=n.return}throw Error(i(160))}switch(r.tag){case 27:var a=r.stateNode;Ks(e,Zs(e),a);break;case 5:var o=r.stateNode;32&r.flags&&(kt(o,""),r.flags&=-33),Ks(e,Zs(e),o);break;case 3:case 4:var s=r.stateNode.containerInfo;Gs(e,Zs(e),s);break;default:throw Error(i(161))}}}catch(t){ru(e,e.return,t)}e.flags&=-3}4096&t&&(e.flags&=-4097)}function gl(e){if(1024&e.subtreeFlags)for(e=e.child;null!==e;){var t=e;gl(t),5===t.tag&&1024&t.flags&&t.stateNode.reset(),e=e.sibling}}function vl(e,t){if(8772&t.subtreeFlags)for(t=t.child;null!==t;)al(e,t.alternate,t),t=t.sibling}function bl(e){for(e=e.child;null!==e;){var t=e;switch(t.tag){case 0:case 11:case 14:case 15:Us(4,t,t.return),bl(t);break;case 1:Xs(t,t.return);var n=t.stateNode;"function"==typeof n.componentWillUnmount&&js(t,t.return,n),bl(t);break;case 26:case 27:case 5:Xs(t,t.return),bl(t);break;case 22:Xs(t,t.return),null===t.memoizedState&&bl(t);break;default:bl(t)}e=e.sibling}}function yl(e,t,n){for(n=n&&!!(8772&t.subtreeFlags),t=t.child;null!==t;){var r=t.alternate,a=e,o=t,i=o.flags;switch(o.tag){case 0:case 11:case 15:yl(a,o,n),Qs(4,o);break;case 1:if(yl(a,o,n),"function"==typeof(a=(r=o).stateNode).componentDidMount)try{a.componentDidMount()}catch(e){ru(r,r.return,e)}if(null!==(a=(r=o).updateQueue)){var s=r.stateNode;try{var l=a.shared.hiddenCallbacks;if(null!==l)for(a.shared.hiddenCallbacks=null,a=0;a<l.length;a++)Bs(l[a],s)}catch(e){ru(r,r.return,e)}}n&&64&i&&zs(o),Ws(o,o.return);break;case 26:case 27:case 5:yl(a,o,n),n&&null===r&&4&i&&qs(o),Ws(o,o.return);break;case 12:default:yl(a,o,n);break;case 13:yl(a,o,n),n&&4&i&&ul(a,o);break;case 22:null===o.memoizedState&&yl(a,o,n),Ws(o,o.return)}t=t.sibling}}function wl(e,t){var n=null;null!==e&&null!==e.memoizedState&&null!==e.memoizedState.cachePool&&(n=e.memoizedState.cachePool.pool),e=null,null!==t.memoizedState&&null!==t.memoizedState.cachePool&&(e=t.memoizedState.cachePool.pool),e!==n&&(null!=e&&e.refCount++,null!=n&&Ua(n))}function Ol(e,t){e=null,null!==t.alternate&&(e=t.alternate.memoizedState.cache),(t=t.memoizedState.cache)!==e&&(t.refCount++,null!=e&&Ua(e))}function kl(e,t,n,r){if(10256&t.subtreeFlags)for(t=t.child;null!==t;)xl(e,t,n,r),t=t.sibling}function xl(e,t,n,r){var a=t.flags;switch(t.tag){case 0:case 11:case 15:kl(e,t,n,r),2048&a&&Qs(9,t);break;case 3:kl(e,t,n,r),2048&a&&(e=null,null!==t.alternate&&(e=t.alternate.memoizedState.cache),(t=t.memoizedState.cache)!==e&&(t.refCount++,null!=e&&Ua(e)));break;case 12:if(2048&a){kl(e,t,n,r),e=t.stateNode;try{var o=t.memoizedProps,i=o.id,s=o.onPostCommit;"function"==typeof s&&s(i,null===t.alternate?"mount":"update",e.passiveEffectDuration,-0)}catch(e){ru(t,t.return,e)}}else kl(e,t,n,r);break;case 23:break;case 22:o=t.stateNode,null!==t.memoizedState?4&o._visibility?kl(e,t,n,r):Sl(e,t):4&o._visibility?kl(e,t,n,r):(o._visibility|=4,El(e,t,n,r,!!(10256&t.subtreeFlags))),2048&a&&wl(t.alternate,t);break;case 24:kl(e,t,n,r),2048&a&&Ol(t.alternate,t);break;default:kl(e,t,n,r)}}function El(e,t,n,r,a){for(a=a&&!!(10256&t.subtreeFlags),t=t.child;null!==t;){var o=e,i=t,s=n,l=r,c=i.flags;switch(i.tag){case 0:case 11:case 15:El(o,i,s,l,a),Qs(8,i);break;case 23:break;case 22:var u=i.stateNode;null!==i.memoizedState?4&u._visibility?El(o,i,s,l,a):Sl(o,i):(u._visibility|=4,El(o,i,s,l,a)),a&&2048&c&&wl(i.alternate,i);break;case 24:El(o,i,s,l,a),a&&2048&c&&Ol(i.alternate,i);break;default:El(o,i,s,l,a)}t=t.sibling}}function Sl(e,t){if(10256&t.subtreeFlags)for(t=t.child;null!==t;){var n=e,r=t,a=r.flags;switch(r.tag){case 22:Sl(n,r),2048&a&&wl(r.alternate,r);break;case 24:Sl(n,r),2048&a&&Ol(r.alternate,r);break;default:Sl(n,r)}t=t.sibling}}var _l=8192;function Cl(e){if(e.subtreeFlags&_l)for(e=e.child;null!==e;)Tl(e),e=e.sibling}function Tl(e){switch(e.tag){case 26:Cl(e),e.flags&_l&&null!==e.memoizedState&&function(e,t,n){if(null===Rd)throw Error(i(475));var r=Rd;if(!("stylesheet"!==t.type||"string"==typeof n.media&&!1===matchMedia(n.media).matches||4&t.state.loading)){if(null===t.instance){var a=vd(n.href),o=e.querySelector(bd(a));if(o)return null!==(e=o._p)&&"object"==typeof e&&"function"==typeof e.then&&(r.count++,r=Dd.bind(r),e.then(r,r)),t.state.loading|=4,t.instance=o,void Ke(o);o=e.ownerDocument||e,n=yd(n),(a=ud.get(a))&&Ed(n,a),Ke(o=o.createElement("link"));var s=o;s._p=new Promise((function(e,t){s.onload=e,s.onerror=t})),qu(o,"link",n),t.instance=o}null===r.stylesheets&&(r.stylesheets=new Map),r.stylesheets.set(t,e),(e=t.state.preload)&&!(3&t.state.loading)&&(r.count++,t=Dd.bind(r),e.addEventListener("load",t),e.addEventListener("error",t))}}(hl,e.memoizedState,e.memoizedProps);break;case 5:default:Cl(e);break;case 3:case 4:var t=hl;hl=pd(e.stateNode.containerInfo),Cl(e),hl=t;break;case 22:null===e.memoizedState&&(null!==(t=e.alternate)&&null!==t.memoizedState?(t=_l,_l=16777216,Cl(e),_l=t):Cl(e))}}function Al(e){var t=e.alternate;if(null!==t&&null!==(e=t.child)){t.child=null;do{t=e.sibling,e.sibling=null,e=t}while(null!==e)}}function Rl(e){var t=e.deletions;if(16&e.flags){if(null!==t)for(var n=0;n<t.length;n++){var r=t[n];nl=r,Ml(r,e)}Al(e)}if(10256&e.subtreeFlags)for(e=e.child;null!==e;)Pl(e),e=e.sibling}function Pl(e){switch(e.tag){case 0:case 11:case 15:Rl(e),2048&e.flags&&Us(9,e,e.return);break;case 3:case 12:default:Rl(e);break;case 22:var t=e.stateNode;null!==e.memoizedState&&4&t._visibility&&(null===e.return||13!==e.return.tag)?(t._visibility&=-5,Dl(e)):Rl(e)}}function Dl(e){var t=e.deletions;if(16&e.flags){if(null!==t)for(var n=0;n<t.length;n++){var r=t[n];nl=r,Ml(r,e)}Al(e)}for(e=e.child;null!==e;){switch((t=e).tag){case 0:case 11:case 15:Us(8,t,t.return),Dl(t);break;case 22:4&(n=t.stateNode)._visibility&&(n._visibility&=-5,Dl(t));break;default:Dl(t)}e=e.sibling}}function Ml(e,t){for(;null!==nl;){var n=nl;switch(n.tag){case 0:case 11:case 15:Us(8,n,t);break;case 23:case 22:if(null!==n.memoizedState&&null!==n.memoizedState.cachePool){var r=n.memoizedState.cachePool.pool;null!=r&&r.refCount++}break;case 24:Ua(n.memoizedState.cache)}if(null!==(r=n.child))r.return=n,nl=r;else e:for(n=e;null!==nl;){var a=(r=nl).sibling,o=r.return;if(ol(r),r===n){nl=null;break e}if(null!==a){a.return=o,nl=a;break e}nl=o}}}function $l(e,t,n,r){this.tag=e,this.key=n,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.refCleanup=this.ref=null,this.pendingProps=t,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=r,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function Nl(e,t,n,r){return new $l(e,t,n,r)}function Il(e){return!(!(e=e.prototype)||!e.isReactComponent)}function Ll(e,t){var n=e.alternate;return null===n?((n=Nl(e.tag,t,e.key,e.mode)).elementType=e.elementType,n.type=e.type,n.stateNode=e.stateNode,n.alternate=e,e.alternate=n):(n.pendingProps=t,n.type=e.type,n.flags=0,n.subtreeFlags=0,n.deletions=null),n.flags=31457280&e.flags,n.childLanes=e.childLanes,n.lanes=e.lanes,n.child=e.child,n.memoizedProps=e.memoizedProps,n.memoizedState=e.memoizedState,n.updateQueue=e.updateQueue,t=e.dependencies,n.dependencies=null===t?null:{lanes:t.lanes,firstContext:t.firstContext},n.sibling=e.sibling,n.index=e.index,n.ref=e.ref,n.refCleanup=e.refCleanup,n}function Bl(e,t){e.flags&=31457282;var n=e.alternate;return null===n?(e.childLanes=0,e.lanes=t,e.child=null,e.subtreeFlags=0,e.memoizedProps=null,e.memoizedState=null,e.updateQueue=null,e.dependencies=null,e.stateNode=null):(e.childLanes=n.childLanes,e.lanes=n.lanes,e.child=n.child,e.subtreeFlags=0,e.deletions=null,e.memoizedProps=n.memoizedProps,e.memoizedState=n.memoizedState,e.updateQueue=n.updateQueue,e.type=n.type,t=n.dependencies,e.dependencies=null===t?null:{lanes:t.lanes,firstContext:t.firstContext}),e}function Fl(e,t,n,r,a,o){var s=0;if(r=e,"function"==typeof e)Il(e)&&(s=1);else if("string"==typeof e)s=function(e,t,n){if(1===n||null!=t.itemProp)return!1;switch(e){case"meta":case"title":return!0;case"style":if("string"!=typeof t.precedence||"string"!=typeof t.href||""===t.href)break;return!0;case"link":if("string"!=typeof t.rel||"string"!=typeof t.href||""===t.href||t.onLoad||t.onError)break;return"stylesheet"!==t.rel||(e=t.disabled,"string"==typeof t.precedence&&null==e);case"script":if(t.async&&"function"!=typeof t.async&&"symbol"!=typeof t.async&&!t.onLoad&&!t.onError&&t.src&&"string"==typeof t.src)return!0}return!1}(e,n,Z.current)?26:"html"===e||"head"===e||"body"===e?27:5;else e:switch(e){case d:return Ql(n.children,a,o,t);case p:s=8,a|=24;break;case h:return(e=Nl(12,n,t,2|a)).elementType=h,e.lanes=o,e;case b:return(e=Nl(13,n,t,a)).elementType=b,e.lanes=o,e;case y:return(e=Nl(19,n,t,a)).elementType=y,e.lanes=o,e;case k:return Ul(n,a,o,t);default:if("object"==typeof e&&null!==e)switch(e.$$typeof){case f:case g:s=10;break e;case m:s=9;break e;case v:s=11;break e;case w:s=14;break e;case O:s=16,r=null;break e}s=29,n=Error(i(130,null===e?"null":typeof e,"")),r=null}return(t=Nl(s,n,t,a)).elementType=e,t.type=r,t.lanes=o,t}function Ql(e,t,n,r){return(e=Nl(7,e,r,t)).lanes=n,e}function Ul(e,t,n,r){(e=Nl(22,e,r,t)).elementType=k,e.lanes=n;var a={_visibility:1,_pendingVisibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null,_current:null,detach:function(){var e=a._current;if(null===e)throw Error(i(456));if(!(2&a._pendingVisibility)){var t=Ar(e,2);null!==t&&(a._pendingVisibility|=2,Dc(t,0,2))}},attach:function(){var e=a._current;if(null===e)throw Error(i(456));if(2&a._pendingVisibility){var t=Ar(e,2);null!==t&&(a._pendingVisibility&=-3,Dc(t,0,2))}}};return e.stateNode=a,e}function zl(e,t,n){return(e=Nl(6,e,null,t)).lanes=n,e}function jl(e,t,n){return(t=Nl(4,null!==e.children?e.children:[],e.key,t)).lanes=n,t.stateNode={containerInfo:e.containerInfo,pendingChildren:null,implementation:e.implementation},t}function Wl(e){e.flags|=4}function Xl(e,t){if("stylesheet"!==t.type||4&t.state.loading)e.flags&=-16777217;else if(e.flags|=16777216,!Ad(t)){if(null!==(t=Ta.current)&&((4194176&rc)===rc?null!==Aa:(62914560&rc)!==rc&&!(536870912&rc)||t!==Aa))throw pa=la,sa;e.flags|=8192}}function ql(e,t){null!==t&&(e.flags|=4),16384&e.flags&&(t=22!==e.tag?Re():536870912,e.lanes|=t,mc|=t)}function Vl(e,t){if(!Gr)switch(e.tailMode){case"hidden":t=e.tail;for(var n=null;null!==t;)null!==t.alternate&&(n=t),t=t.sibling;null===n?e.tail=null:n.sibling=null;break;case"collapsed":n=e.tail;for(var r=null;null!==n;)null!==n.alternate&&(r=n),n=n.sibling;null===r?t||null===e.tail?e.tail=null:e.tail.sibling=null:r.sibling=null}}function Hl(e){var t=null!==e.alternate&&e.alternate.child===e.child,n=0,r=0;if(t)for(var a=e.child;null!==a;)n|=a.lanes|a.childLanes,r|=31457280&a.subtreeFlags,r|=31457280&a.flags,a.return=e,a=a.sibling;else for(a=e.child;null!==a;)n|=a.lanes|a.childLanes,r|=a.subtreeFlags,r|=a.flags,a.return=e,a=a.sibling;return e.subtreeFlags|=r,e.childLanes=n,t}function Zl(e,t,n){var r=t.pendingProps;switch(Vr(t),t.tag){case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:case 1:return Hl(t),null;case 3:return n=t.stateNode,r=null,null!==e&&(r=e.memoizedState.cache),t.memoizedState.cache!==r&&(t.flags|=2048),ys(Fa),ee(),n.pendingContext&&(n.context=n.pendingContext,n.pendingContext=null),null!==e&&null!==e.child||(ra(t)?Wl(t):null===e||e.memoizedState.isDehydrated&&!(256&t.flags)||(t.flags|=1024,null!==Kr&&($c(Kr),Kr=null))),Hl(t),null;case 26:return n=t.memoizedState,null===e?(Wl(t),null!==n?(Hl(t),Xl(t,n)):(Hl(t),t.flags&=-16777217)):n?n!==e.memoizedState?(Wl(t),Hl(t),Xl(t,n)):(Hl(t),t.flags&=-16777217):(e.memoizedProps!==r&&Wl(t),Hl(t),t.flags&=-16777217),null;case 27:ne(t),n=K.current;var a=t.type;if(null!==e&&null!=t.stateNode)e.memoizedProps!==r&&Wl(t);else{if(!r){if(null===t.stateNode)throw Error(i(166));return Hl(t),null}e=Z.current,ra(t)?ta(t):(e=cd(a,r,n),t.stateNode=e,Wl(t))}return Hl(t),null;case 5:if(ne(t),n=t.type,null!==e&&null!=t.stateNode)e.memoizedProps!==r&&Wl(t);else{if(!r){if(null===t.stateNode)throw Error(i(166));return Hl(t),null}if(e=Z.current,ra(t))ta(t);else{switch(a=Zu(K.current),e){case 1:e=a.createElementNS("http://www.w3.org/2000/svg",n);break;case 2:e=a.createElementNS("http://www.w3.org/1998/Math/MathML",n);break;default:switch(n){case"svg":e=a.createElementNS("http://www.w3.org/2000/svg",n);break;case"math":e=a.createElementNS("http://www.w3.org/1998/Math/MathML",n);break;case"script":(e=a.createElement("div")).innerHTML="<script><\/script>",e=e.removeChild(e.firstChild);break;case"select":e="string"==typeof r.is?a.createElement("select",{is:r.is}):a.createElement("select"),r.multiple?e.multiple=!0:r.size&&(e.size=r.size);break;default:e="string"==typeof r.is?a.createElement(n,{is:r.is}):a.createElement(n)}}e[Be]=t,e[Fe]=r;e:for(a=t.child;null!==a;){if(5===a.tag||6===a.tag)e.appendChild(a.stateNode);else if(4!==a.tag&&27!==a.tag&&null!==a.child){a.child.return=a,a=a.child;continue}if(a===t)break e;for(;null===a.sibling;){if(null===a.return||a.return===t)break e;a=a.return}a.sibling.return=a.return,a=a.sibling}t.stateNode=e;e:switch(qu(e,n,r),n){case"button":case"input":case"select":case"textarea":e=!!r.autoFocus;break e;case"img":e=!0;break e;default:e=!1}e&&Wl(t)}}return Hl(t),t.flags&=-16777217,null;case 6:if(e&&null!=t.stateNode)e.memoizedProps!==r&&Wl(t);else{if("string"!=typeof r&&null===t.stateNode)throw Error(i(166));if(e=K.current,ra(t)){if(e=t.stateNode,n=t.memoizedProps,r=null,null!==(a=Hr))switch(a.tag){case 27:case 5:r=a.memoizedProps}e[Be]=t,(e=!!(e.nodeValue===n||null!==r&&!0===r.suppressHydrationWarning||zu(e.nodeValue,n)))||ea(t)}else(e=Zu(e).createTextNode(r))[Be]=t,t.stateNode=e}return Hl(t),null;case 13:if(r=t.memoizedState,null===e||null!==e.memoizedState&&null!==e.memoizedState.dehydrated){if(a=ra(t),null!==r&&null!==r.dehydrated){if(null===e){if(!a)throw Error(i(318));if(!(a=null!==(a=t.memoizedState)?a.dehydrated:null))throw Error(i(317));a[Be]=t}else aa(),!(128&t.flags)&&(t.memoizedState=null),t.flags|=4;Hl(t),a=!1}else null!==Kr&&($c(Kr),Kr=null),a=!0;if(!a)return 256&t.flags?(Ma(t),t):(Ma(t),null)}if(Ma(t),128&t.flags)return t.lanes=n,t;if(n=null!==r,e=null!==e&&null!==e.memoizedState,n){a=null,null!==(r=t.child).alternate&&null!==r.alternate.memoizedState&&null!==r.alternate.memoizedState.cachePool&&(a=r.alternate.memoizedState.cachePool.pool);var o=null;null!==r.memoizedState&&null!==r.memoizedState.cachePool&&(o=r.memoizedState.cachePool.pool),o!==a&&(r.flags|=2048)}return n!==e&&n&&(t.child.flags|=8192),ql(t,t.updateQueue),Hl(t),null;case 4:return ee(),null===e&&Du(t.stateNode.containerInfo),Hl(t),null;case 10:return ys(t.type),Hl(t),null;case 19:if(V($a),null===(a=t.memoizedState))return Hl(t),null;if(r=!!(128&t.flags),null===(o=a.rendering))if(r)Vl(a,!1);else{if(0!==uc||null!==e&&128&e.flags)for(e=t.child;null!==e;){if(null!==(o=Na(e))){for(t.flags|=128,Vl(a,!1),e=o.updateQueue,t.updateQueue=e,ql(t,e),t.subtreeFlags=0,e=n,n=t.child;null!==n;)Bl(n,e),n=n.sibling;return H($a,1&$a.current|2),t.child}e=e.sibling}null!==a.tail&&le()>wc&&(t.flags|=128,r=!0,Vl(a,!1),t.lanes=4194304)}else{if(!r)if(null!==(e=Na(o))){if(t.flags|=128,r=!0,e=e.updateQueue,t.updateQueue=e,ql(t,e),Vl(a,!0),null===a.tail&&"hidden"===a.tailMode&&!o.alternate&&!Gr)return Hl(t),null}else 2*le()-a.renderingStartTime>wc&&536870912!==n&&(t.flags|=128,r=!0,Vl(a,!1),t.lanes=4194304);a.isBackwards?(o.sibling=t.child,t.child=o):(null!==(e=a.last)?e.sibling=o:t.child=o,a.last=o)}return null!==a.tail?(t=a.tail,a.rendering=t,a.tail=t.sibling,a.renderingStartTime=le(),t.sibling=null,e=$a.current,H($a,r?1&e|2:1&e),t):(Hl(t),null);case 22:case 23:return Ma(t),Ca(),r=null!==t.memoizedState,null!==e?null!==e.memoizedState!==r&&(t.flags|=8192):r&&(t.flags|=8192),r?!!(536870912&n)&&!(128&t.flags)&&(Hl(t),6&t.subtreeFlags&&(t.flags|=8192)):Hl(t),null!==(n=t.updateQueue)&&ql(t,n.retryQueue),n=null,null!==e&&null!==e.memoizedState&&null!==e.memoizedState.cachePool&&(n=e.memoizedState.cachePool.pool),r=null,null!==t.memoizedState&&null!==t.memoizedState.cachePool&&(r=t.memoizedState.cachePool.pool),r!==n&&(t.flags|=2048),null!==e&&V(Ha),null;case 24:return n=null,null!==e&&(n=e.memoizedState.cache),t.memoizedState.cache!==n&&(t.flags|=2048),ys(Fa),Hl(t),null;case 25:return null}throw Error(i(156,t.tag))}function Gl(e,t){switch(Vr(t),t.tag){case 1:return 65536&(e=t.flags)?(t.flags=-65537&e|128,t):null;case 3:return ys(Fa),ee(),65536&(e=t.flags)&&!(128&e)?(t.flags=-65537&e|128,t):null;case 26:case 27:case 5:return ne(t),null;case 13:if(Ma(t),null!==(e=t.memoizedState)&&null!==e.dehydrated){if(null===t.alternate)throw Error(i(340));aa()}return 65536&(e=t.flags)?(t.flags=-65537&e|128,t):null;case 19:return V($a),null;case 4:return ee(),null;case 10:return ys(t.type),null;case 22:case 23:return Ma(t),Ca(),null!==e&&V(Ha),65536&(e=t.flags)?(t.flags=-65537&e|128,t):null;case 24:return ys(Fa),null;default:return null}}function Kl(e,t){switch(Vr(t),t.tag){case 3:ys(Fa),ee();break;case 26:case 27:case 5:ne(t);break;case 4:ee();break;case 13:Ma(t);break;case 19:V($a);break;case 10:ys(t.type);break;case 22:case 23:Ma(t),Ca(),null!==e&&V(Ha);break;case 24:ys(Fa)}}var Yl={getCacheForType:function(e){var t=Ss(Fa),n=t.data.get(e);return void 0===n&&(n=e(),t.data.set(e,n)),n}},Jl="function"==typeof WeakMap?WeakMap:Map,ec=0,tc=null,nc=null,rc=0,ac=0,oc=null,ic=!1,sc=!1,lc=!1,cc=0,uc=0,dc=0,pc=0,hc=0,fc=0,mc=0,gc=null,vc=null,bc=!1,yc=0,wc=1/0,Oc=null,kc=null,xc=!1,Ec=null,Sc=0,_c=0,Cc=null,Tc=0,Ac=null;function Rc(){return 2&ec&&0!==rc?rc&-rc:null!==R.T?0!==Wa?Wa:Ou():Ie()}function Pc(){0===fc&&(fc=536870912&rc&&!Gr?536870912:Ae());var e=Ta.current;return null!==e&&(e.flags|=32),fc}function Dc(e,t,n){(e===tc&&2===ac||null!==e.cancelPendingCommit)&&(Qc(e,0),Lc(e,rc,fc,!1)),De(e,n),2&ec&&e===tc||(e===tc&&(!(2&ec)&&(pc|=n),4===uc&&Lc(e,rc,fc,!1)),mu(e))}function Mc(e,t,n){if(6&ec)throw Error(i(327));for(var r=!n&&!(60&t)&&!(t&e.expiredLanes)||Ce(e,t),a=r?function(e,t){var n=ec;ec|=2;var r=zc(),a=jc();tc!==e||rc!==t?(Oc=null,wc=le()+500,Qc(e,t)):sc=Ce(e,t);e:for(;;)try{if(0!==ac&&null!==nc){t=nc;var o=oc;t:switch(ac){case 1:ac=0,oc=null,Gc(e,t,o,1);break;case 2:if(ca(o)){ac=0,oc=null,Zc(t);break}t=function(){2===ac&&tc===e&&(ac=7),mu(e)},o.then(t,t);break e;case 3:ac=7;break e;case 4:ac=5;break e;case 7:ca(o)?(ac=0,oc=null,Zc(t)):(ac=0,oc=null,Gc(e,t,o,7));break;case 5:var s=null;switch(nc.tag){case 26:s=nc.memoizedState;case 5:case 27:var l=nc;if(!s||Ad(s)){ac=0,oc=null;var c=l.sibling;if(null!==c)nc=c;else{var u=l.return;null!==u?(nc=u,Kc(u)):nc=null}break t}}ac=0,oc=null,Gc(e,t,o,5);break;case 6:ac=0,oc=null,Gc(e,t,o,6);break;case 8:Fc(),uc=6;break e;default:throw Error(i(462))}}Vc();break}catch(t){Uc(e,t)}return vs=gs=null,R.H=r,R.A=a,ec=n,null!==nc?0:(tc=null,rc=0,_r(),uc)}(e,t):Xc(e,t,!0),o=r;;){if(0===a){sc&&!r&&Lc(e,t,0,!1);break}if(6===a)Lc(e,t,0,!ic);else{if(n=e.current.alternate,o&&!Ic(n)){a=Xc(e,t,!1),o=!1;continue}if(2===a){if(o=t,e.errorRecoveryDisabledLanes&o)var s=0;else s=0!=(s=-536870913&e.pendingLanes)?s:536870912&s?536870912:0;if(0!==s){t=s;e:{var l=e;a=gc;var c=l.current.memoizedState.isDehydrated;if(c&&(Qc(l,s).flags|=256),2!==(s=Xc(l,s,!1))){if(lc&&!c){l.errorRecoveryDisabledLanes|=o,pc|=o,a=4;break e}o=vc,vc=a,null!==o&&$c(o)}a=s}if(o=!1,2!==a)continue}}if(1===a){Qc(e,0),Lc(e,t,0,!0);break}e:{switch(r=e,a){case 0:case 1:throw Error(i(345));case 4:if((4194176&t)===t){Lc(r,t,fc,!ic);break e}break;case 2:vc=null;break;case 3:case 5:break;default:throw Error(i(329))}if(r.finishedWork=n,r.finishedLanes=t,(62914560&t)===t&&10<(o=yc+300-le())){if(Lc(r,t,fc,!ic),0!==_e(r,0))break e;r.timeoutHandle=ed(Nc.bind(null,r,n,vc,Oc,bc,t,fc,pc,mc,ic,2,-0,0),o)}else Nc(r,n,vc,Oc,bc,t,fc,pc,mc,ic,0,-0,0)}}break}mu(e)}function $c(e){null===vc?vc=e:vc.push.apply(vc,e)}function Nc(e,t,n,r,a,o,s,l,c,u,d,p,h){var f=t.subtreeFlags;if((8192&f||!(16785408&~f))&&(Rd={stylesheets:null,count:0,unsuspend:Pd},Tl(t),null!==(t=function(){if(null===Rd)throw Error(i(475));var e=Rd;return e.stylesheets&&0===e.count&&$d(e,e.stylesheets),0<e.count?function(t){var n=setTimeout((function(){if(e.stylesheets&&$d(e,e.stylesheets),e.unsuspend){var t=e.unsuspend;e.unsuspend=null,t()}}),6e4);return e.unsuspend=t,function(){e.unsuspend=null,clearTimeout(n)}}:null}())))return e.cancelPendingCommit=t(Jc.bind(null,e,n,r,a,s,l,c,1,p,h)),void Lc(e,o,s,!u);Jc(e,n,r,a,s,l,c)}function Ic(e){for(var t=e;;){var n=t.tag;if((0===n||11===n||15===n)&&16384&t.flags&&null!==(n=t.updateQueue)&&null!==(n=n.stores))for(var r=0;r<n.length;r++){var a=n[r],o=a.getSnapshot;a=a.value;try{if(!Hn(o(),a))return!1}catch(e){return!1}}if(n=t.child,16384&t.subtreeFlags&&null!==n)n.return=t,t=n;else{if(t===e)break;for(;null===t.sibling;){if(null===t.return||t.return===e)return!0;t=t.return}t.sibling.return=t.return,t=t.sibling}}return!0}function Lc(e,t,n,r){t&=~hc,t&=~pc,e.suspendedLanes|=t,e.pingedLanes&=~t,r&&(e.warmLanes|=t),r=e.expirationTimes;for(var a=t;0<a;){var o=31-we(a),i=1<<o;r[o]=-1,a&=~i}0!==n&&Me(e,n,t)}function Bc(){return!!(6&ec)||(gu(0,!1),!1)}function Fc(){if(null!==nc){if(0===ac)var e=nc.return;else vs=gs=null,bo(e=nc),fa=null,ma=0,e=nc;for(;null!==e;)Kl(e.alternate,e),e=e.return;nc=null}}function Qc(e,t){e.finishedWork=null,e.finishedLanes=0;var n=e.timeoutHandle;-1!==n&&(e.timeoutHandle=-1,td(n)),null!==(n=e.cancelPendingCommit)&&(e.cancelPendingCommit=null,n()),Fc(),tc=e,nc=n=Ll(e.current,null),rc=t,ac=0,oc=null,ic=!1,sc=Ce(e,t),lc=!1,mc=fc=hc=pc=dc=uc=0,vc=gc=null,bc=!1,8&t&&(t|=32&t);var r=e.entangledLanes;if(0!==r)for(e=e.entanglements,r&=t;0<r;){var a=31-we(r),o=1<<a;t|=e[a],r&=~o}return cc=t,_r(),n}function Uc(e,t){Ja=null,R.H=Si,t===ia?(t=ha(),ac=3):t===sa?(t=ha(),ac=4):ac=t===ji?8:null!==t&&"object"==typeof t&&"function"==typeof t.then?6:1,oc=t,null===nc&&(uc=1,Bi(e,$r(t,e.current)))}function zc(){var e=R.H;return R.H=Si,null===e?Si:e}function jc(){var e=R.A;return R.A=Yl,e}function Wc(){uc=4,ic||(4194176&rc)!==rc&&null!==Ta.current||(sc=!0),!(134217727&dc)&&!(134217727&pc)||null===tc||Lc(tc,rc,fc,!1)}function Xc(e,t,n){var r=ec;ec|=2;var a=zc(),o=jc();tc===e&&rc===t||(Oc=null,Qc(e,t)),t=!1;var i=uc;e:for(;;)try{if(0!==ac&&null!==nc){var s=nc,l=oc;switch(ac){case 8:Fc(),i=6;break e;case 3:case 2:case 6:null===Ta.current&&(t=!0);var c=ac;if(ac=0,oc=null,Gc(e,s,l,c),n&&sc){i=0;break e}break;default:c=ac,ac=0,oc=null,Gc(e,s,l,c)}}qc(),i=uc;break}catch(t){Uc(e,t)}return t&&e.shellSuspendCounter++,vs=gs=null,ec=r,R.H=a,R.A=o,null===nc&&(tc=null,rc=0,_r()),i}function qc(){for(;null!==nc;)Hc(nc)}function Vc(){for(;null!==nc&&!ie();)Hc(nc)}function Hc(e){var t=fs(e.alternate,e,cc);e.memoizedProps=e.pendingProps,null===t?Kc(e):nc=t}function Zc(e){var t=e,n=t.alternate;switch(t.tag){case 15:case 0:t=Ji(n,t,t.pendingProps,t.type,void 0,rc);break;case 11:t=Ji(n,t,t.pendingProps,t.type.render,t.ref,rc);break;case 5:bo(t);default:Kl(n,t),t=fs(n,t=nc=Bl(t,cc),cc)}e.memoizedProps=e.pendingProps,null===t?Kc(e):nc=t}function Gc(e,t,n,r){vs=gs=null,bo(t),fa=null,ma=0;var a=t.return;try{if(function(e,t,n,r,a){if(n.flags|=32768,null!==r&&"object"==typeof r&&"function"==typeof r.then){if(null!==(t=n.alternate)&&ks(t,n,a,!0),null!==(n=Ta.current)){switch(n.tag){case 13:return null===Aa?Wc():null===n.alternate&&0===uc&&(uc=3),n.flags&=-257,n.flags|=65536,n.lanes=a,r===la?n.flags|=16384:(null===(t=n.updateQueue)?n.updateQueue=new Set([r]):t.add(r),au(e,r,a)),!1;case 22:return n.flags|=65536,r===la?n.flags|=16384:(null===(t=n.updateQueue)?(t={transitions:null,markerInstances:null,retryQueue:new Set([r])},n.updateQueue=t):null===(n=t.retryQueue)?t.retryQueue=new Set([r]):n.add(r),au(e,r,a)),!1}throw Error(i(435,n.tag))}return au(e,r,a),Wc(),!1}if(Gr)return null!==(t=Ta.current)?(!(65536&t.flags)&&(t.flags|=256),t.flags|=65536,t.lanes=a,r!==Jr&&oa($r(e=Error(i(422),{cause:r}),n))):(r!==Jr&&oa($r(t=Error(i(423),{cause:r}),n)),(e=e.current.alternate).flags|=65536,a&=-a,e.lanes|=a,r=$r(r,n),$s(e,a=Qi(e.stateNode,r,a)),4!==uc&&(uc=2)),!1;var o=Error(i(520),{cause:r});if(o=$r(o,n),null===gc?gc=[o]:gc.push(o),4!==uc&&(uc=2),null===t)return!0;r=$r(r,n),n=t;do{switch(n.tag){case 3:return n.flags|=65536,e=a&-a,n.lanes|=e,$s(n,e=Qi(n.stateNode,r,e)),!1;case 1:if(t=n.type,o=n.stateNode,!(128&n.flags||"function"!=typeof t.getDerivedStateFromError&&(null===o||"function"!=typeof o.componentDidCatch||null!==kc&&kc.has(o))))return n.flags|=65536,a&=-a,n.lanes|=a,zi(a=Ui(a),e,n,r),$s(n,a),!1}n=n.return}while(null!==n);return!1}(e,a,t,n,rc))return uc=1,Bi(e,$r(n,e.current)),void(nc=null)}catch(t){if(null!==a)throw nc=a,t;return uc=1,Bi(e,$r(n,e.current)),void(nc=null)}32768&t.flags?(Gr||1===r?e=!0:sc||536870912&rc?e=!1:(ic=e=!0,(2===r||3===r||6===r)&&null!==(r=Ta.current)&&13===r.tag&&(r.flags|=16384)),Yc(t,e)):Kc(t)}function Kc(e){var t=e;do{if(32768&t.flags)return void Yc(t,ic);e=t.return;var n=Zl(t.alternate,t,cc);if(null!==n)return void(nc=n);if(null!==(t=t.sibling))return void(nc=t);nc=t=e}while(null!==t);0===uc&&(uc=5)}function Yc(e,t){do{var n=Gl(e.alternate,e);if(null!==n)return n.flags&=32767,void(nc=n);if(null!==(n=e.return)&&(n.flags|=32768,n.subtreeFlags=0,n.deletions=null),!t&&null!==(e=e.sibling))return void(nc=e);nc=e=n}while(null!==e);uc=6,nc=null}function Jc(e,t,n,r,a,o,s,l,c,u){var d=R.T,p=z.p;try{z.p=2,R.T=null,function(e,t,n,r,a,o,s,l){do{tu()}while(null!==Ec);if(6&ec)throw Error(i(327));var c=e.finishedWork;if(r=e.finishedLanes,null===c)return null;if(e.finishedWork=null,e.finishedLanes=0,c===e.current)throw Error(i(177));e.callbackNode=null,e.callbackPriority=0,e.cancelPendingCommit=null;var u=c.lanes|c.childLanes;if(function(e,t,n,r,a,o){var i=e.pendingLanes;e.pendingLanes=n,e.suspendedLanes=0,e.pingedLanes=0,e.warmLanes=0,e.expiredLanes&=n,e.entangledLanes&=n,e.errorRecoveryDisabledLanes&=n,e.shellSuspendCounter=0;var s=e.entanglements,l=e.expirationTimes,c=e.hiddenUpdates;for(n=i&~n;0<n;){var u=31-we(n),d=1<<u;s[u]=0,l[u]=-1;var p=c[u];if(null!==p)for(c[u]=null,u=0;u<p.length;u++){var h=p[u];null!==h&&(h.lane&=-536870913)}n&=~d}0!==r&&Me(e,r,0),0!==o&&0===a&&0!==e.tag&&(e.suspendedLanes|=o&~(i&~t))}(e,r,u|=Sr,o,s,l),e===tc&&(nc=tc=null,rc=0),!(10256&c.subtreeFlags)&&!(10256&c.flags)||xc||(xc=!0,_c=u,Cc=n,ae(pe,(function(){return tu(),null}))),n=!!(15990&c.flags),15990&c.subtreeFlags||n?(n=R.T,R.T=null,o=z.p,z.p=2,s=ec,ec|=4,function(e,t){if(e=e.containerInfo,Vu=Wd,er(e=Jn(e))){if("selectionStart"in e)var n={start:e.selectionStart,end:e.selectionEnd};else e:{var r=(n=(n=e.ownerDocument)&&n.defaultView||window).getSelection&&n.getSelection();if(r&&0!==r.rangeCount){n=r.anchorNode;var a=r.anchorOffset,o=r.focusNode;r=r.focusOffset;try{n.nodeType,o.nodeType}catch(e){n=null;break e}var s=0,l=-1,c=-1,u=0,d=0,p=e,h=null;t:for(;;){for(var f;p!==n||0!==a&&3!==p.nodeType||(l=s+a),p!==o||0!==r&&3!==p.nodeType||(c=s+r),3===p.nodeType&&(s+=p.nodeValue.length),null!==(f=p.firstChild);)h=p,p=f;for(;;){if(p===e)break t;if(h===n&&++u===a&&(l=s),h===o&&++d===r&&(c=s),null!==(f=p.nextSibling))break;h=(p=h).parentNode}p=f}n=-1===l||-1===c?null:{start:l,end:c}}else n=null}n=n||{start:0,end:0}}else n=null;for(Hu={focusedElem:e,selectionRange:n},Wd=!1,nl=t;null!==nl;)if(e=(t=nl).child,1028&t.subtreeFlags&&null!==e)e.return=t,nl=e;else for(;null!==nl;){switch(o=(t=nl).alternate,e=t.flags,t.tag){case 0:case 11:case 15:case 5:case 26:case 27:case 6:case 4:case 17:break;case 1:if(1024&e&&null!==o){e=void 0,n=t,a=o.memoizedProps,o=o.memoizedState,r=n.stateNode;try{var m=Mi(n.type,a,(n.elementType,n.type));e=r.getSnapshotBeforeUpdate(m,o),r.__reactInternalSnapshotBeforeUpdate=e}catch(e){ru(n,n.return,e)}}break;case 3:if(1024&e)if(9===(n=(e=t.stateNode.containerInfo).nodeType))id(e);else if(1===n)switch(e.nodeName){case"HEAD":case"HTML":case"BODY":id(e);break;default:e.textContent=""}break;default:if(1024&e)throw Error(i(163))}if(null!==(e=t.sibling)){e.return=t.return,nl=e;break}nl=t.return}m=rl,rl=!1}(e,c),fl(c,e),tr(Hu,e.containerInfo),Wd=!!Vu,Hu=Vu=null,e.current=c,al(e,c.alternate,c),se(),ec=s,z.p=o,R.T=n):e.current=c,xc?(xc=!1,Ec=e,Sc=r):eu(e,u),0===(u=e.pendingLanes)&&(kc=null),function(e){if(be&&"function"==typeof be.onCommitFiberRoot)try{be.onCommitFiberRoot(ve,e,void 0,!(128&~e.current.flags))}catch(e){}}(c.stateNode),mu(e),null!==t)for(a=e.onRecoverableError,c=0;c<t.length;c++)a((u=t[c]).value,{componentStack:u.stack});!!(3&Sc)&&tu(),u=e.pendingLanes,4194218&r&&42&u?e===Ac?Tc++:(Tc=0,Ac=e):Tc=0,gu(0,!1)}(e,t,n,r,p,a,o,s)}finally{R.T=d,z.p=p}}function eu(e,t){0==(e.pooledCacheLanes&=t)&&null!=(t=e.pooledCache)&&(e.pooledCache=null,Ua(t))}function tu(){if(null!==Ec){var e=Ec,t=_c;_c=0;var n=Ne(Sc),r=R.T,a=z.p;try{if(z.p=32>n?32:n,R.T=null,null===Ec)var o=!1;else{n=Cc,Cc=null;var s=Ec,l=Sc;if(Ec=null,Sc=0,6&ec)throw Error(i(331));var c=ec;if(ec|=4,Pl(s.current),xl(s,s.current,l,n),ec=c,gu(0,!1),be&&"function"==typeof be.onPostCommitFiberRoot)try{be.onPostCommitFiberRoot(ve,s)}catch(e){}o=!0}return o}finally{z.p=a,R.T=r,eu(e,t)}}return!1}function nu(e,t,n){t=$r(n,t),null!==(e=Ds(e,t=Qi(e.stateNode,t,2),2))&&(De(e,2),mu(e))}function ru(e,t,n){if(3===e.tag)nu(e,e,n);else for(;null!==t;){if(3===t.tag){nu(t,e,n);break}if(1===t.tag){var r=t.stateNode;if("function"==typeof t.type.getDerivedStateFromError||"function"==typeof r.componentDidCatch&&(null===kc||!kc.has(r))){e=$r(n,e),null!==(r=Ds(t,n=Ui(2),2))&&(zi(n,r,t,e),De(r,2),mu(r));break}}t=t.return}}function au(e,t,n){var r=e.pingCache;if(null===r){r=e.pingCache=new Jl;var a=new Set;r.set(t,a)}else void 0===(a=r.get(t))&&(a=new Set,r.set(t,a));a.has(n)||(lc=!0,a.add(n),e=ou.bind(null,e,t,n),t.then(e,e))}function ou(e,t,n){var r=e.pingCache;null!==r&&r.delete(t),e.pingedLanes|=e.suspendedLanes&n,e.warmLanes&=~n,tc===e&&(rc&n)===n&&(4===uc||3===uc&&(62914560&rc)===rc&&300>le()-yc?!(2&ec)&&Qc(e,0):hc|=n,mc===rc&&(mc=0)),mu(e)}function iu(e,t){0===t&&(t=Re()),null!==(e=Ar(e,t))&&(De(e,t),mu(e))}function su(e){var t=e.memoizedState,n=0;null!==t&&(n=t.retryLane),iu(e,n)}function lu(e,t){var n=0;switch(e.tag){case 13:var r=e.stateNode,a=e.memoizedState;null!==a&&(n=a.retryLane);break;case 19:r=e.stateNode;break;case 22:r=e.stateNode._retryCache;break;default:throw Error(i(314))}null!==r&&r.delete(t),iu(e,n)}var cu=null,uu=null,du=!1,pu=!1,hu=!1,fu=0;function mu(e){var t;e!==uu&&null===e.next&&(null===uu?cu=uu=e:uu=uu.next=e),pu=!0,du||(du=!0,t=vu,rd((function(){6&ec?ae(ue,t):t()})))}function gu(e,t){if(!hu&&pu){hu=!0;do{for(var n=!1,r=cu;null!==r;){if(!t)if(0!==e){var a=r.pendingLanes;if(0===a)var o=0;else{var i=r.suspendedLanes,s=r.pingedLanes;o=(1<<31-we(42|e)+1)-1,o=201326677&(o&=a&~(i&~s))?201326677&o|1:o?2|o:0}0!==o&&(n=!0,wu(r,o))}else o=rc,!(3&(o=_e(r,r===tc?o:0)))||Ce(r,o)||(n=!0,wu(r,o));r=r.next}}while(n);hu=!1}}function vu(){pu=du=!1;var e,t=0;0!==fu&&(((e=window.event)&&"popstate"===e.type?e!==Ju&&(Ju=e,!0):(Ju=null,!1))&&(t=fu),fu=0);for(var n=le(),r=null,a=cu;null!==a;){var o=a.next,i=bu(a,n);0===i?(a.next=null,null===r?cu=o:r.next=o,null===o&&(uu=r)):(r=a,(0!==t||3&i)&&(pu=!0)),a=o}gu(t,!1)}function bu(e,t){for(var n=e.suspendedLanes,r=e.pingedLanes,a=e.expirationTimes,o=-62914561&e.pendingLanes;0<o;){var i=31-we(o),s=1<<i,l=a[i];-1===l?s&n&&!(s&r)||(a[i]=Te(s,t)):l<=t&&(e.expiredLanes|=s),o&=~s}if(n=rc,n=_e(e,e===(t=tc)?n:0),r=e.callbackNode,0===n||e===t&&2===ac||null!==e.cancelPendingCommit)return null!==r&&null!==r&&oe(r),e.callbackNode=null,e.callbackPriority=0;if(!(3&n)||Ce(e,n)){if((t=n&-n)===e.callbackPriority)return t;switch(null!==r&&oe(r),Ne(n)){case 2:case 8:n=de;break;case 32:default:n=pe;break;case 268435456:n=fe}return r=yu.bind(null,e),n=ae(n,r),e.callbackPriority=t,e.callbackNode=n,t}return null!==r&&null!==r&&oe(r),e.callbackPriority=2,e.callbackNode=null,2}function yu(e,t){var n=e.callbackNode;if(tu()&&e.callbackNode!==n)return null;var r=rc;return 0===(r=_e(e,e===tc?r:0))?null:(Mc(e,r,t),bu(e,le()),null!=e.callbackNode&&e.callbackNode===n?yu.bind(null,e):null)}function wu(e,t){if(tu())return null;Mc(e,t,!0)}function Ou(){return 0===fu&&(fu=Ae()),fu}function ku(e){return null==e||"symbol"==typeof e||"boolean"==typeof e?null:"function"==typeof e?e:At(""+e)}function xu(e,t){var n=t.ownerDocument.createElement("input");return n.name=t.name,n.value=t.value,e.id&&n.setAttribute("form",e.id),t.parentNode.insertBefore(n,t),e=new FormData(e),n.parentNode.removeChild(n),e}for(var Eu=0;Eu<Or.length;Eu++){var Su=Or[Eu];kr(Su.toLowerCase(),"on"+(Su[0].toUpperCase()+Su.slice(1)))}kr(hr,"onAnimationEnd"),kr(fr,"onAnimationIteration"),kr(mr,"onAnimationStart"),kr("dblclick","onDoubleClick"),kr("focusin","onFocus"),kr("focusout","onBlur"),kr(gr,"onTransitionRun"),kr(vr,"onTransitionStart"),kr(br,"onTransitionCancel"),kr(yr,"onTransitionEnd"),tt("onMouseEnter",["mouseout","mouseover"]),tt("onMouseLeave",["mouseout","mouseover"]),tt("onPointerEnter",["pointerout","pointerover"]),tt("onPointerLeave",["pointerout","pointerover"]),et("onChange","change click focusin focusout input keydown keyup selectionchange".split(" ")),et("onSelect","focusout contextmenu dragend focusin keydown keyup mousedown mouseup selectionchange".split(" ")),et("onBeforeInput",["compositionend","keypress","textInput","paste"]),et("onCompositionEnd","compositionend focusout keydown keypress keyup mousedown".split(" ")),et("onCompositionStart","compositionstart focusout keydown keypress keyup mousedown".split(" ")),et("onCompositionUpdate","compositionupdate focusout keydown keypress keyup mousedown".split(" "));var _u="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange resize seeked seeking stalled suspend timeupdate volumechange waiting".split(" "),Cu=new Set("beforetoggle cancel close invalid load scroll scrollend toggle".split(" ").concat(_u));function Tu(e,t){t=!!(4&t);for(var n=0;n<e.length;n++){var r=e[n],a=r.event;r=r.listeners;e:{var o=void 0;if(t)for(var i=r.length-1;0<=i;i--){var s=r[i],l=s.instance,c=s.currentTarget;if(s=s.listener,l!==o&&a.isPropagationStopped())break e;o=s,a.currentTarget=c;try{o(a)}catch(e){$i(e)}a.currentTarget=null,o=l}else for(i=0;i<r.length;i++){if(l=(s=r[i]).instance,c=s.currentTarget,s=s.listener,l!==o&&a.isPropagationStopped())break e;o=s,a.currentTarget=c;try{o(a)}catch(e){$i(e)}a.currentTarget=null,o=l}}}}function Au(e,t){var n=t[Ue];void 0===n&&(n=t[Ue]=new Set);var r=e+"__bubble";n.has(r)||(Mu(t,e,2,!1),n.add(r))}function Ru(e,t,n){var r=0;t&&(r|=4),Mu(n,e,r,t)}var Pu="_reactListening"+Math.random().toString(36).slice(2);function Du(e){if(!e[Pu]){e[Pu]=!0,Ye.forEach((function(t){"selectionchange"!==t&&(Cu.has(t)||Ru(t,!1,e),Ru(t,!0,e))}));var t=9===e.nodeType?e:e.ownerDocument;null===t||t[Pu]||(t[Pu]=!0,Ru("selectionchange",!1,t))}}function Mu(e,t,n,r){switch(Kd(t)){case 2:var a=Xd;break;case 8:a=qd;break;default:a=Vd}n=a.bind(null,t,n,e),a=void 0,!Bt||"touchstart"!==t&&"touchmove"!==t&&"wheel"!==t||(a=!0),r?void 0!==a?e.addEventListener(t,n,{capture:!0,passive:a}):e.addEventListener(t,n,!0):void 0!==a?e.addEventListener(t,n,{passive:a}):e.addEventListener(t,n,!1)}function $u(e,t,n,r,a){var o=r;if(!(1&t||2&t||null===r))e:for(;;){if(null===r)return;var i=r.tag;if(3===i||4===i){var s=r.stateNode.containerInfo;if(s===a||8===s.nodeType&&s.parentNode===a)break;if(4===i)for(i=r.return;null!==i;){var l=i.tag;if((3===l||4===l)&&((l=i.stateNode.containerInfo)===a||8===l.nodeType&&l.parentNode===a))return;i=i.return}for(;null!==s;){if(null===(i=Ve(s)))return;if(5===(l=i.tag)||6===l||26===l||27===l){r=o=i;continue e}s=s.parentNode}}r=r.return}It((function(){var r=o,a=Pt(n),i=[];e:{var s=wr.get(e);if(void 0!==s){var l=Yt,c=e;switch(e){case"keypress":if(0===Wt(n))break e;case"keydown":case"keyup":l=fn;break;case"focusin":c="focus",l=an;break;case"focusout":c="blur",l=an;break;case"beforeblur":case"afterblur":l=an;break;case"click":if(2===n.button)break e;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":l=nn;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":l=rn;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":l=gn;break;case hr:case fr:case mr:l=on;break;case yr:l=vn;break;case"scroll":case"scrollend":l=en;break;case"wheel":l=bn;break;case"copy":case"cut":case"paste":l=sn;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":l=mn;break;case"toggle":case"beforetoggle":l=yn}var u=!!(4&t),d=!u&&("scroll"===e||"scrollend"===e),p=u?null!==s?s+"Capture":null:s;u=[];for(var h,f=r;null!==f;){var m=f;if(h=m.stateNode,5!==(m=m.tag)&&26!==m&&27!==m||null===h||null===p||null!=(m=Lt(f,p))&&u.push(Nu(f,m,h)),d)break;f=f.return}0<u.length&&(s=new l(s,c,null,n,a),i.push({event:s,listeners:u}))}}if(!(7&t)){if(l="mouseout"===e||"pointerout"===e,(!(s="mouseover"===e||"pointerover"===e)||n===Rt||!(c=n.relatedTarget||n.fromElement)||!Ve(c)&&!c[Qe])&&(l||s)&&(s=a.window===a?a:(s=a.ownerDocument)?s.defaultView||s.parentWindow:window,l?(l=r,null!==(c=(c=n.relatedTarget||n.toElement)?Ve(c):null)&&(d=L(c),u=c.tag,c!==d||5!==u&&27!==u&&6!==u)&&(c=null)):(l=null,c=r),l!==c)){if(u=nn,m="onMouseLeave",p="onMouseEnter",f="mouse","pointerout"!==e&&"pointerover"!==e||(u=mn,m="onPointerLeave",p="onPointerEnter",f="pointer"),d=null==l?s:Ze(l),h=null==c?s:Ze(c),(s=new u(m,f+"leave",l,n,a)).target=d,s.relatedTarget=h,m=null,Ve(a)===r&&((u=new u(p,f+"enter",c,n,a)).target=h,u.relatedTarget=d,m=u),d=m,l&&c)e:{for(p=c,f=0,h=u=l;h;h=Lu(h))f++;for(h=0,m=p;m;m=Lu(m))h++;for(;0<f-h;)u=Lu(u),f--;for(;0<h-f;)p=Lu(p),h--;for(;f--;){if(u===p||null!==p&&u===p.alternate)break e;u=Lu(u),p=Lu(p)}u=null}else u=null;null!==l&&Bu(i,s,l,u,!1),null!==c&&null!==d&&Bu(i,d,c,u,!0)}if("select"===(l=(s=r?Ze(r):window).nodeName&&s.nodeName.toLowerCase())||"input"===l&&"file"===s.type)var g=Ln;else if(Pn(s))if(Bn)g=Vn;else{g=Xn;var v=Wn}else!(l=s.nodeName)||"input"!==l.toLowerCase()||"checkbox"!==s.type&&"radio"!==s.type?r&&_t(r.elementType)&&(g=Ln):g=qn;switch(g&&(g=g(e,r))?Dn(i,g,n,a):(v&&v(e,s,r),"focusout"===e&&r&&"number"===s.type&&null!=r.memoizedProps.value&&bt(s,"number",s.value)),v=r?Ze(r):window,e){case"focusin":(Pn(v)||"true"===v.contentEditable)&&(rr=v,ar=r,or=null);break;case"focusout":or=ar=rr=null;break;case"mousedown":ir=!0;break;case"contextmenu":case"mouseup":case"dragend":ir=!1,sr(i,n,a);break;case"selectionchange":if(nr)break;case"keydown":case"keyup":sr(i,n,a)}var b;if(On)e:{switch(e){case"compositionstart":var y="onCompositionStart";break e;case"compositionend":y="onCompositionEnd";break e;case"compositionupdate":y="onCompositionUpdate";break e}y=void 0}else An?Cn(e,n)&&(y="onCompositionEnd"):"keydown"===e&&229===n.keyCode&&(y="onCompositionStart");y&&(En&&"ko"!==n.locale&&(An||"onCompositionStart"!==y?"onCompositionEnd"===y&&An&&(b=jt()):(Ut="value"in(Qt=a)?Qt.value:Qt.textContent,An=!0)),0<(v=Iu(r,y)).length&&(y=new ln(y,e,null,n,a),i.push({event:y,listeners:v}),(b||null!==(b=Tn(n)))&&(y.data=b))),(b=xn?function(e,t){switch(e){case"compositionend":return Tn(t);case"keypress":return 32!==t.which?null:(_n=!0,Sn);case"textInput":return(e=t.data)===Sn&&_n?null:e;default:return null}}(e,n):function(e,t){if(An)return"compositionend"===e||!On&&Cn(e,t)?(e=jt(),zt=Ut=Qt=null,An=!1,e):null;switch(e){case"paste":default:return null;case"keypress":if(!(t.ctrlKey||t.altKey||t.metaKey)||t.ctrlKey&&t.altKey){if(t.char&&1<t.char.length)return t.char;if(t.which)return String.fromCharCode(t.which)}return null;case"compositionend":return En&&"ko"!==t.locale?null:t.data}}(e,n))&&0<(y=Iu(r,"onBeforeInput")).length&&(v=new ln("onBeforeInput","beforeinput",null,n,a),i.push({event:v,listeners:y}),v.data=b),function(e,t,n,r,a){if("submit"===t&&n&&n.stateNode===a){var o=ku((a[Fe]||null).action),i=r.submitter;i&&null!==(t=(t=i[Fe]||null)?ku(t.formAction):i.getAttribute("formAction"))&&(o=t,i=null);var s=new Yt("action","action",null,r,a);e.push({event:s,listeners:[{instance:null,listener:function(){if(r.defaultPrevented){if(0!==fu){var e=i?xu(a,i):new FormData(a);di(n,{pending:!0,data:e,method:a.method,action:o},null,e)}}else"function"==typeof o&&(s.preventDefault(),e=i?xu(a,i):new FormData(a),di(n,{pending:!0,data:e,method:a.method,action:o},o,e))},currentTarget:a}]})}}(i,e,r,n,a)}Tu(i,t)}))}function Nu(e,t,n){return{instance:e,listener:t,currentTarget:n}}function Iu(e,t){for(var n=t+"Capture",r=[];null!==e;){var a=e,o=a.stateNode;5!==(a=a.tag)&&26!==a&&27!==a||null===o||(null!=(a=Lt(e,n))&&r.unshift(Nu(e,a,o)),null!=(a=Lt(e,t))&&r.push(Nu(e,a,o))),e=e.return}return r}function Lu(e){if(null===e)return null;do{e=e.return}while(e&&5!==e.tag&&27!==e.tag);return e||null}function Bu(e,t,n,r,a){for(var o=t._reactName,i=[];null!==n&&n!==r;){var s=n,l=s.alternate,c=s.stateNode;if(s=s.tag,null!==l&&l===r)break;5!==s&&26!==s&&27!==s||null===c||(l=c,a?null!=(c=Lt(n,o))&&i.unshift(Nu(n,c,l)):a||null!=(c=Lt(n,o))&&i.push(Nu(n,c,l))),n=n.return}0!==i.length&&e.push({event:t,listeners:i})}var Fu=/\r\n?/g,Qu=/\u0000|\uFFFD/g;function Uu(e){return("string"==typeof e?e:""+e).replace(Fu,"\n").replace(Qu,"")}function zu(e,t){return t=Uu(t),Uu(e)===t}function ju(){}function Wu(e,t,n,r,a,o){switch(n){case"children":"string"==typeof r?"body"===t||"textarea"===t&&""===r||kt(e,r):("number"==typeof r||"bigint"==typeof r)&&"body"!==t&&kt(e,""+r);break;case"className":st(e,"class",r);break;case"tabIndex":st(e,"tabindex",r);break;case"dir":case"role":case"viewBox":case"width":case"height":st(e,n,r);break;case"style":St(e,r,o);break;case"data":if("object"!==t){st(e,"data",r);break}case"src":case"href":if(""===r&&("a"!==t||"href"!==n)){e.removeAttribute(n);break}if(null==r||"function"==typeof r||"symbol"==typeof r||"boolean"==typeof r){e.removeAttribute(n);break}r=At(""+r),e.setAttribute(n,r);break;case"action":case"formAction":if("function"==typeof r){e.setAttribute(n,"javascript:throw new Error('A React form was unexpectedly submitted. If you called form.submit() manually, consider using form.requestSubmit() instead. If you\\'re trying to use event.stopPropagation() in a submit event handler, consider also calling event.preventDefault().')");break}if("function"==typeof o&&("formAction"===n?("input"!==t&&Wu(e,t,"name",a.name,a,null),Wu(e,t,"formEncType",a.formEncType,a,null),Wu(e,t,"formMethod",a.formMethod,a,null),Wu(e,t,"formTarget",a.formTarget,a,null)):(Wu(e,t,"encType",a.encType,a,null),Wu(e,t,"method",a.method,a,null),Wu(e,t,"target",a.target,a,null))),null==r||"symbol"==typeof r||"boolean"==typeof r){e.removeAttribute(n);break}r=At(""+r),e.setAttribute(n,r);break;case"onClick":null!=r&&(e.onclick=ju);break;case"onScroll":null!=r&&Au("scroll",e);break;case"onScrollEnd":null!=r&&Au("scrollend",e);break;case"dangerouslySetInnerHTML":if(null!=r){if("object"!=typeof r||!("__html"in r))throw Error(i(61));if(null!=(n=r.__html)){if(null!=a.children)throw Error(i(60));e.innerHTML=n}}break;case"multiple":e.multiple=r&&"function"!=typeof r&&"symbol"!=typeof r;break;case"muted":e.muted=r&&"function"!=typeof r&&"symbol"!=typeof r;break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"defaultValue":case"defaultChecked":case"innerHTML":case"ref":case"autoFocus":break;case"xlinkHref":if(null==r||"function"==typeof r||"boolean"==typeof r||"symbol"==typeof r){e.removeAttribute("xlink:href");break}n=At(""+r),e.setAttributeNS("http://www.w3.org/1999/xlink","xlink:href",n);break;case"contentEditable":case"spellCheck":case"draggable":case"value":case"autoReverse":case"externalResourcesRequired":case"focusable":case"preserveAlpha":null!=r&&"function"!=typeof r&&"symbol"!=typeof r?e.setAttribute(n,""+r):e.removeAttribute(n);break;case"inert":case"allowFullScreen":case"async":case"autoPlay":case"controls":case"default":case"defer":case"disabled":case"disablePictureInPicture":case"disableRemotePlayback":case"formNoValidate":case"hidden":case"loop":case"noModule":case"noValidate":case"open":case"playsInline":case"readOnly":case"required":case"reversed":case"scoped":case"seamless":case"itemScope":r&&"function"!=typeof r&&"symbol"!=typeof r?e.setAttribute(n,""):e.removeAttribute(n);break;case"capture":case"download":!0===r?e.setAttribute(n,""):!1!==r&&null!=r&&"function"!=typeof r&&"symbol"!=typeof r?e.setAttribute(n,r):e.removeAttribute(n);break;case"cols":case"rows":case"size":case"span":null!=r&&"function"!=typeof r&&"symbol"!=typeof r&&!isNaN(r)&&1<=r?e.setAttribute(n,r):e.removeAttribute(n);break;case"rowSpan":case"start":null==r||"function"==typeof r||"symbol"==typeof r||isNaN(r)?e.removeAttribute(n):e.setAttribute(n,r);break;case"popover":Au("beforetoggle",e),Au("toggle",e),it(e,"popover",r);break;case"xlinkActuate":lt(e,"http://www.w3.org/1999/xlink","xlink:actuate",r);break;case"xlinkArcrole":lt(e,"http://www.w3.org/1999/xlink","xlink:arcrole",r);break;case"xlinkRole":lt(e,"http://www.w3.org/1999/xlink","xlink:role",r);break;case"xlinkShow":lt(e,"http://www.w3.org/1999/xlink","xlink:show",r);break;case"xlinkTitle":lt(e,"http://www.w3.org/1999/xlink","xlink:title",r);break;case"xlinkType":lt(e,"http://www.w3.org/1999/xlink","xlink:type",r);break;case"xmlBase":lt(e,"http://www.w3.org/XML/1998/namespace","xml:base",r);break;case"xmlLang":lt(e,"http://www.w3.org/XML/1998/namespace","xml:lang",r);break;case"xmlSpace":lt(e,"http://www.w3.org/XML/1998/namespace","xml:space",r);break;case"is":it(e,"is",r);break;case"innerText":case"textContent":break;default:(!(2<n.length)||"o"!==n[0]&&"O"!==n[0]||"n"!==n[1]&&"N"!==n[1])&&it(e,n=Ct.get(n)||n,r)}}function Xu(e,t,n,r,a,o){switch(n){case"style":St(e,r,o);break;case"dangerouslySetInnerHTML":if(null!=r){if("object"!=typeof r||!("__html"in r))throw Error(i(61));if(null!=(n=r.__html)){if(null!=a.children)throw Error(i(60));e.innerHTML=n}}break;case"children":"string"==typeof r?kt(e,r):("number"==typeof r||"bigint"==typeof r)&&kt(e,""+r);break;case"onScroll":null!=r&&Au("scroll",e);break;case"onScrollEnd":null!=r&&Au("scrollend",e);break;case"onClick":null!=r&&(e.onclick=ju);break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"innerHTML":case"ref":case"innerText":case"textContent":break;default:Je.hasOwnProperty(n)||("o"!==n[0]||"n"!==n[1]||(a=n.endsWith("Capture"),t=n.slice(2,a?n.length-7:void 0),"function"==typeof(o=null!=(o=e[Fe]||null)?o[n]:null)&&e.removeEventListener(t,o,a),"function"!=typeof r)?n in e?e[n]=r:!0===r?e.setAttribute(n,""):it(e,n,r):("function"!=typeof o&&null!==o&&(n in e?e[n]=null:e.hasAttribute(n)&&e.removeAttribute(n)),e.addEventListener(t,r,a)))}}function qu(e,t,n){switch(t){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"img":Au("error",e),Au("load",e);var r,a=!1,o=!1;for(r in n)if(n.hasOwnProperty(r)){var s=n[r];if(null!=s)switch(r){case"src":a=!0;break;case"srcSet":o=!0;break;case"children":case"dangerouslySetInnerHTML":throw Error(i(137,t));default:Wu(e,t,r,s,n,null)}}return o&&Wu(e,t,"srcSet",n.srcSet,n,null),void(a&&Wu(e,t,"src",n.src,n,null));case"input":Au("invalid",e);var l=r=s=o=null,c=null,u=null;for(a in n)if(n.hasOwnProperty(a)){var d=n[a];if(null!=d)switch(a){case"name":o=d;break;case"type":s=d;break;case"checked":c=d;break;case"defaultChecked":u=d;break;case"value":r=d;break;case"defaultValue":l=d;break;case"children":case"dangerouslySetInnerHTML":if(null!=d)throw Error(i(137,t));break;default:Wu(e,t,a,d,n,null)}}return vt(e,r,l,c,u,s,o,!1),void dt(e);case"select":for(o in Au("invalid",e),a=s=r=null,n)if(n.hasOwnProperty(o)&&null!=(l=n[o]))switch(o){case"value":r=l;break;case"defaultValue":s=l;break;case"multiple":a=l;default:Wu(e,t,o,l,n,null)}return t=r,n=s,e.multiple=!!a,void(null!=t?yt(e,!!a,t,!1):null!=n&&yt(e,!!a,n,!0));case"textarea":for(s in Au("invalid",e),r=o=a=null,n)if(n.hasOwnProperty(s)&&null!=(l=n[s]))switch(s){case"value":a=l;break;case"defaultValue":o=l;break;case"children":r=l;break;case"dangerouslySetInnerHTML":if(null!=l)throw Error(i(91));break;default:Wu(e,t,s,l,n,null)}return Ot(e,a,o,r),void dt(e);case"option":for(c in n)n.hasOwnProperty(c)&&null!=(a=n[c])&&("selected"===c?e.selected=a&&"function"!=typeof a&&"symbol"!=typeof a:Wu(e,t,c,a,n,null));return;case"dialog":Au("cancel",e),Au("close",e);break;case"iframe":case"object":Au("load",e);break;case"video":case"audio":for(a=0;a<_u.length;a++)Au(_u[a],e);break;case"image":Au("error",e),Au("load",e);break;case"details":Au("toggle",e);break;case"embed":case"source":case"link":Au("error",e),Au("load",e);case"area":case"base":case"br":case"col":case"hr":case"keygen":case"meta":case"param":case"track":case"wbr":case"menuitem":for(u in n)if(n.hasOwnProperty(u)&&null!=(a=n[u]))switch(u){case"children":case"dangerouslySetInnerHTML":throw Error(i(137,t));default:Wu(e,t,u,a,n,null)}return;default:if(_t(t)){for(d in n)n.hasOwnProperty(d)&&void 0!==(a=n[d])&&Xu(e,t,d,a,n,void 0);return}}for(l in n)n.hasOwnProperty(l)&&null!=(a=n[l])&&Wu(e,t,l,a,n,null)}var Vu=null,Hu=null;function Zu(e){return 9===e.nodeType?e:e.ownerDocument}function Gu(e){switch(e){case"http://www.w3.org/2000/svg":return 1;case"http://www.w3.org/1998/Math/MathML":return 2;default:return 0}}function Ku(e,t){if(0===e)switch(t){case"svg":return 1;case"math":return 2;default:return 0}return 1===e&&"foreignObject"===t?0:e}function Yu(e,t){return"textarea"===e||"noscript"===e||"string"==typeof t.children||"number"==typeof t.children||"bigint"==typeof t.children||"object"==typeof t.dangerouslySetInnerHTML&&null!==t.dangerouslySetInnerHTML&&null!=t.dangerouslySetInnerHTML.__html}var Ju=null,ed="function"==typeof setTimeout?setTimeout:void 0,td="function"==typeof clearTimeout?clearTimeout:void 0,nd="function"==typeof Promise?Promise:void 0,rd="function"==typeof queueMicrotask?queueMicrotask:void 0!==nd?function(e){return nd.resolve(null).then(e).catch(ad)}:ed;function ad(e){setTimeout((function(){throw e}))}function od(e,t){var n=t,r=0;do{var a=n.nextSibling;if(e.removeChild(n),a&&8===a.nodeType)if("/$"===(n=a.data)){if(0===r)return e.removeChild(a),void mp(t);r--}else"$"!==n&&"$?"!==n&&"$!"!==n||r++;n=a}while(n);mp(t)}function id(e){var t=e.firstChild;for(t&&10===t.nodeType&&(t=t.nextSibling);t;){var n=t;switch(t=t.nextSibling,n.nodeName){case"HTML":case"HEAD":case"BODY":id(n),qe(n);continue;case"SCRIPT":case"STYLE":continue;case"LINK":if("stylesheet"===n.rel.toLowerCase())continue}e.removeChild(n)}}function sd(e){for(;null!=e;e=e.nextSibling){var t=e.nodeType;if(1===t||3===t)break;if(8===t){if("$"===(t=e.data)||"$!"===t||"$?"===t||"F!"===t||"F"===t)break;if("/$"===t)return null}}return e}function ld(e){e=e.previousSibling;for(var t=0;e;){if(8===e.nodeType){var n=e.data;if("$"===n||"$!"===n||"$?"===n){if(0===t)return e;t--}else"/$"===n&&t++}e=e.previousSibling}return null}function cd(e,t,n){switch(t=Zu(n),e){case"html":if(!(e=t.documentElement))throw Error(i(452));return e;case"head":if(!(e=t.head))throw Error(i(453));return e;case"body":if(!(e=t.body))throw Error(i(454));return e;default:throw Error(i(451))}}var ud=new Map,dd=new Set;function pd(e){return"function"==typeof e.getRootNode?e.getRootNode():e.ownerDocument}var hd=z.d;z.d={f:function(){var e=hd.f(),t=Bc();return e||t},r:function(e){var t=He(e);null!==t&&5===t.tag&&"form"===t.type?hi(t):hd.r(e)},D:function(e){hd.D(e),md("dns-prefetch",e,null)},C:function(e,t){hd.C(e,t),md("preconnect",e,t)},L:function(e,t,n){hd.L(e,t,n);var r=fd;if(r&&e&&t){var a='link[rel="preload"][as="'+mt(t)+'"]';"image"===t&&n&&n.imageSrcSet?(a+='[imagesrcset="'+mt(n.imageSrcSet)+'"]',"string"==typeof n.imageSizes&&(a+='[imagesizes="'+mt(n.imageSizes)+'"]')):a+='[href="'+mt(e)+'"]';var o=a;switch(t){case"style":o=vd(e);break;case"script":o=wd(e)}ud.has(o)||(e=P({rel:"preload",href:"image"===t&&n&&n.imageSrcSet?void 0:e,as:t},n),ud.set(o,e),null!==r.querySelector(a)||"style"===t&&r.querySelector(bd(o))||"script"===t&&r.querySelector(Od(o))||(qu(t=r.createElement("link"),"link",e),Ke(t),r.head.appendChild(t)))}},m:function(e,t){hd.m(e,t);var n=fd;if(n&&e){var r=t&&"string"==typeof t.as?t.as:"script",a='link[rel="modulepreload"][as="'+mt(r)+'"][href="'+mt(e)+'"]',o=a;switch(r){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":o=wd(e)}if(!ud.has(o)&&(e=P({rel:"modulepreload",href:e},t),ud.set(o,e),null===n.querySelector(a))){switch(r){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":if(n.querySelector(Od(o)))return}qu(r=n.createElement("link"),"link",e),Ke(r),n.head.appendChild(r)}}},X:function(e,t){hd.X(e,t);var n=fd;if(n&&e){var r=Ge(n).hoistableScripts,a=wd(e),o=r.get(a);o||((o=n.querySelector(Od(a)))||(e=P({src:e,async:!0},t),(t=ud.get(a))&&Sd(e,t),Ke(o=n.createElement("script")),qu(o,"link",e),n.head.appendChild(o)),o={type:"script",instance:o,count:1,state:null},r.set(a,o))}},S:function(e,t,n){hd.S(e,t,n);var r=fd;if(r&&e){var a=Ge(r).hoistableStyles,o=vd(e);t=t||"default";var i=a.get(o);if(!i){var s={loading:0,preload:null};if(i=r.querySelector(bd(o)))s.loading=5;else{e=P({rel:"stylesheet",href:e,"data-precedence":t},n),(n=ud.get(o))&&Ed(e,n);var l=i=r.createElement("link");Ke(l),qu(l,"link",e),l._p=new Promise((function(e,t){l.onload=e,l.onerror=t})),l.addEventListener("load",(function(){s.loading|=1})),l.addEventListener("error",(function(){s.loading|=2})),s.loading|=4,xd(i,t,r)}i={type:"stylesheet",instance:i,count:1,state:s},a.set(o,i)}}},M:function(e,t){hd.M(e,t);var n=fd;if(n&&e){var r=Ge(n).hoistableScripts,a=wd(e),o=r.get(a);o||((o=n.querySelector(Od(a)))||(e=P({src:e,async:!0,type:"module"},t),(t=ud.get(a))&&Sd(e,t),Ke(o=n.createElement("script")),qu(o,"link",e),n.head.appendChild(o)),o={type:"script",instance:o,count:1,state:null},r.set(a,o))}}};var fd="undefined"==typeof document?null:document;function md(e,t,n){var r=fd;if(r&&"string"==typeof t&&t){var a=mt(t);a='link[rel="'+e+'"][href="'+a+'"]',"string"==typeof n&&(a+='[crossorigin="'+n+'"]'),dd.has(a)||(dd.add(a),e={rel:e,crossOrigin:n,href:t},null===r.querySelector(a)&&(qu(t=r.createElement("link"),"link",e),Ke(t),r.head.appendChild(t)))}}function gd(e,t,n,r){var a,o,s,l,c=(c=K.current)?pd(c):null;if(!c)throw Error(i(446));switch(e){case"meta":case"title":return null;case"style":return"string"==typeof n.precedence&&"string"==typeof n.href?(t=vd(n.href),(r=(n=Ge(c).hoistableStyles).get(t))||(r={type:"style",instance:null,count:0,state:null},n.set(t,r)),r):{type:"void",instance:null,count:0,state:null};case"link":if("stylesheet"===n.rel&&"string"==typeof n.href&&"string"==typeof n.precedence){e=vd(n.href);var u=Ge(c).hoistableStyles,d=u.get(e);if(d||(c=c.ownerDocument||c,d={type:"stylesheet",instance:null,count:0,state:{loading:0,preload:null}},u.set(e,d),(u=c.querySelector(bd(e)))&&!u._p&&(d.instance=u,d.state.loading=5),ud.has(e)||(n={rel:"preload",as:"style",href:n.href,crossOrigin:n.crossOrigin,integrity:n.integrity,media:n.media,hrefLang:n.hrefLang,referrerPolicy:n.referrerPolicy},ud.set(e,n),u||(a=c,o=e,s=n,l=d.state,a.querySelector('link[rel="preload"][as="style"]['+o+"]")?l.loading=1:(o=a.createElement("link"),l.preload=o,o.addEventListener("load",(function(){return l.loading|=1})),o.addEventListener("error",(function(){return l.loading|=2})),qu(o,"link",s),Ke(o),a.head.appendChild(o))))),t&&null===r)throw Error(i(528,""));return d}if(t&&null!==r)throw Error(i(529,""));return null;case"script":return t=n.async,"string"==typeof(n=n.src)&&t&&"function"!=typeof t&&"symbol"!=typeof t?(t=wd(n),(r=(n=Ge(c).hoistableScripts).get(t))||(r={type:"script",instance:null,count:0,state:null},n.set(t,r)),r):{type:"void",instance:null,count:0,state:null};default:throw Error(i(444,e))}}function vd(e){return'href="'+mt(e)+'"'}function bd(e){return'link[rel="stylesheet"]['+e+"]"}function yd(e){return P({},e,{"data-precedence":e.precedence,precedence:null})}function wd(e){return'[src="'+mt(e)+'"]'}function Od(e){return"script[async]"+e}function kd(e,t,n){if(t.count++,null===t.instance)switch(t.type){case"style":var r=e.querySelector('style[data-href~="'+mt(n.href)+'"]');if(r)return t.instance=r,Ke(r),r;var a=P({},n,{"data-href":n.href,"data-precedence":n.precedence,href:null,precedence:null});return Ke(r=(e.ownerDocument||e).createElement("style")),qu(r,"style",a),xd(r,n.precedence,e),t.instance=r;case"stylesheet":a=vd(n.href);var o=e.querySelector(bd(a));if(o)return t.state.loading|=4,t.instance=o,Ke(o),o;r=yd(n),(a=ud.get(a))&&Ed(r,a),Ke(o=(e.ownerDocument||e).createElement("link"));var s=o;return s._p=new Promise((function(e,t){s.onload=e,s.onerror=t})),qu(o,"link",r),t.state.loading|=4,xd(o,n.precedence,e),t.instance=o;case"script":return o=wd(n.src),(a=e.querySelector(Od(o)))?(t.instance=a,Ke(a),a):(r=n,(a=ud.get(o))&&Sd(r=P({},n),a),Ke(a=(e=e.ownerDocument||e).createElement("script")),qu(a,"link",r),e.head.appendChild(a),t.instance=a);case"void":return null;default:throw Error(i(443,t.type))}else"stylesheet"===t.type&&!(4&t.state.loading)&&(r=t.instance,t.state.loading|=4,xd(r,n.precedence,e));return t.instance}function xd(e,t,n){for(var r=n.querySelectorAll('link[rel="stylesheet"][data-precedence],style[data-precedence]'),a=r.length?r[r.length-1]:null,o=a,i=0;i<r.length;i++){var s=r[i];if(s.dataset.precedence===t)o=s;else if(o!==a)break}o?o.parentNode.insertBefore(e,o.nextSibling):(t=9===n.nodeType?n.head:n).insertBefore(e,t.firstChild)}function Ed(e,t){null==e.crossOrigin&&(e.crossOrigin=t.crossOrigin),null==e.referrerPolicy&&(e.referrerPolicy=t.referrerPolicy),null==e.title&&(e.title=t.title)}function Sd(e,t){null==e.crossOrigin&&(e.crossOrigin=t.crossOrigin),null==e.referrerPolicy&&(e.referrerPolicy=t.referrerPolicy),null==e.integrity&&(e.integrity=t.integrity)}var _d=null;function Cd(e,t,n){if(null===_d){var r=new Map,a=_d=new Map;a.set(n,r)}else(r=(a=_d).get(n))||(r=new Map,a.set(n,r));if(r.has(e))return r;for(r.set(e,null),n=n.getElementsByTagName(e),a=0;a<n.length;a++){var o=n[a];if(!(o[Xe]||o[Be]||"link"===e&&"stylesheet"===o.getAttribute("rel"))&&"http://www.w3.org/2000/svg"!==o.namespaceURI){var i=o.getAttribute(t)||"";i=e+i;var s=r.get(i);s?s.push(o):r.set(i,[o])}}return r}function Td(e,t,n){(e=e.ownerDocument||e).head.insertBefore(n,"title"===t?e.querySelector("head > title"):null)}function Ad(e){return!!("stylesheet"!==e.type||3&e.state.loading)}var Rd=null;function Pd(){}function Dd(){if(this.count--,0===this.count)if(this.stylesheets)$d(this,this.stylesheets);else if(this.unsuspend){var e=this.unsuspend;this.unsuspend=null,e()}}var Md=null;function $d(e,t){e.stylesheets=null,null!==e.unsuspend&&(e.count++,Md=new Map,t.forEach(Nd,e),Md=null,Dd.call(e))}function Nd(e,t){if(!(4&t.state.loading)){var n=Md.get(e);if(n)var r=n.get(null);else{n=new Map,Md.set(e,n);for(var a=e.querySelectorAll("link[data-precedence],style[data-precedence]"),o=0;o<a.length;o++){var i=a[o];"LINK"!==i.nodeName&&"not all"===i.getAttribute("media")||(n.set(i.dataset.precedence,i),r=i)}r&&n.set(null,r)}i=(a=t.instance).getAttribute("data-precedence"),(o=n.get(i)||r)===r&&n.set(null,a),n.set(i,a),this.count++,r=Dd.bind(this),a.addEventListener("load",r),a.addEventListener("error",r),o?o.parentNode.insertBefore(a,o.nextSibling):(e=9===e.nodeType?e.head:e).insertBefore(a,e.firstChild),t.state.loading|=4}}var Id={$$typeof:g,Provider:null,Consumer:null,_currentValue:j,_currentValue2:j,_threadCount:0};function Ld(e,t,n,r,a,o,i,s){this.tag=1,this.containerInfo=e,this.finishedWork=this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=-1,this.callbackNode=this.next=this.pendingContext=this.context=this.cancelPendingCommit=null,this.callbackPriority=0,this.expirationTimes=Pe(-1),this.entangledLanes=this.shellSuspendCounter=this.errorRecoveryDisabledLanes=this.finishedLanes=this.expiredLanes=this.warmLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=Pe(0),this.hiddenUpdates=Pe(null),this.identifierPrefix=r,this.onUncaughtError=a,this.onCaughtError=o,this.onRecoverableError=i,this.pooledCache=null,this.pooledCacheLanes=0,this.formState=s,this.incompleteTransitions=new Map}function Bd(e,t,n,r,a,o,i,s,l,c,u,d){return e=new Ld(e,t,n,i,s,l,c,d),t=1,!0===o&&(t|=24),o=Nl(3,null,null,t),e.current=o,o.stateNode=e,(t=Qa()).refCount++,e.pooledCache=t,t.refCount++,o.memoizedState={element:r,isDehydrated:n,cache:t},As(o),e}function Fd(e){return e?e=Dr:Dr}function Qd(e,t,n,r,a,o){a=Fd(a),null===r.context?r.context=a:r.pendingContext=a,(r=Ps(t)).payload={element:n},null!==(o=void 0===o?null:o)&&(r.callback=o),null!==(n=Ds(e,r,t))&&(Dc(n,0,t),Ms(n,e,t))}function Ud(e,t){if(null!==(e=e.memoizedState)&&null!==e.dehydrated){var n=e.retryLane;e.retryLane=0!==n&&n<t?n:t}}function zd(e,t){Ud(e,t),(e=e.alternate)&&Ud(e,t)}function jd(e){if(13===e.tag){var t=Ar(e,67108864);null!==t&&Dc(t,0,67108864),zd(e,67108864)}}var Wd=!0;function Xd(e,t,n,r){var a=R.T;R.T=null;var o=z.p;try{z.p=2,Vd(e,t,n,r)}finally{z.p=o,R.T=a}}function qd(e,t,n,r){var a=R.T;R.T=null;var o=z.p;try{z.p=8,Vd(e,t,n,r)}finally{z.p=o,R.T=a}}function Vd(e,t,n,r){if(Wd){var a=Hd(r);if(null===a)$u(e,t,r,Zd,n),ip(e,r);else if(function(e,t,n,r,a){switch(t){case"focusin":return Jd=sp(Jd,e,t,n,r,a),!0;case"dragenter":return ep=sp(ep,e,t,n,r,a),!0;case"mouseover":return tp=sp(tp,e,t,n,r,a),!0;case"pointerover":var o=a.pointerId;return np.set(o,sp(np.get(o)||null,e,t,n,r,a)),!0;case"gotpointercapture":return o=a.pointerId,rp.set(o,sp(rp.get(o)||null,e,t,n,r,a)),!0}return!1}(a,e,t,n,r))r.stopPropagation();else if(ip(e,r),4&t&&-1<op.indexOf(e)){for(;null!==a;){var o=He(a);if(null!==o)switch(o.tag){case 3:if((o=o.stateNode).current.memoizedState.isDehydrated){var i=Se(o.pendingLanes);if(0!==i){var s=o;for(s.pendingLanes|=2,s.entangledLanes|=2;i;){var l=1<<31-we(i);s.entanglements[1]|=l,i&=~l}mu(o),!(6&ec)&&(wc=le()+500,gu(0,!1))}}break;case 13:null!==(s=Ar(o,2))&&Dc(s,0,2),Bc(),zd(o,2)}if(null===(o=Hd(r))&&$u(e,t,r,Zd,n),o===a)break;a=o}null!==a&&r.stopPropagation()}else $u(e,t,r,null,n)}}function Hd(e){return Gd(e=Pt(e))}var Zd=null;function Gd(e){if(Zd=null,null!==(e=Ve(e))){var t=L(e);if(null===t)e=null;else{var n=t.tag;if(13===n){if(null!==(e=B(t)))return e;e=null}else if(3===n){if(t.stateNode.current.memoizedState.isDehydrated)return 3===t.tag?t.stateNode.containerInfo:null;e=null}else t!==e&&(e=null)}}return Zd=e,null}function Kd(e){switch(e){case"beforetoggle":case"cancel":case"click":case"close":case"contextmenu":case"copy":case"cut":case"auxclick":case"dblclick":case"dragend":case"dragstart":case"drop":case"focusin":case"focusout":case"input":case"invalid":case"keydown":case"keypress":case"keyup":case"mousedown":case"mouseup":case"paste":case"pause":case"play":case"pointercancel":case"pointerdown":case"pointerup":case"ratechange":case"reset":case"resize":case"seeked":case"submit":case"toggle":case"touchcancel":case"touchend":case"touchstart":case"volumechange":case"change":case"selectionchange":case"textInput":case"compositionstart":case"compositionend":case"compositionupdate":case"beforeblur":case"afterblur":case"beforeinput":case"blur":case"fullscreenchange":case"focus":case"hashchange":case"popstate":case"select":case"selectstart":return 2;case"drag":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"mousemove":case"mouseout":case"mouseover":case"pointermove":case"pointerout":case"pointerover":case"scroll":case"touchmove":case"wheel":case"mouseenter":case"mouseleave":case"pointerenter":case"pointerleave":return 8;case"message":switch(ce()){case ue:return 2;case de:return 8;case pe:case he:return 32;case fe:return 268435456;default:return 32}default:return 32}}var Yd=!1,Jd=null,ep=null,tp=null,np=new Map,rp=new Map,ap=[],op="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput copy cut paste click change contextmenu reset".split(" ");function ip(e,t){switch(e){case"focusin":case"focusout":Jd=null;break;case"dragenter":case"dragleave":ep=null;break;case"mouseover":case"mouseout":tp=null;break;case"pointerover":case"pointerout":np.delete(t.pointerId);break;case"gotpointercapture":case"lostpointercapture":rp.delete(t.pointerId)}}function sp(e,t,n,r,a,o){return null===e||e.nativeEvent!==o?(e={blockedOn:t,domEventName:n,eventSystemFlags:r,nativeEvent:o,targetContainers:[a]},null!==t&&null!==(t=He(t))&&jd(t),e):(e.eventSystemFlags|=r,t=e.targetContainers,null!==a&&-1===t.indexOf(a)&&t.push(a),e)}function lp(e){var t=Ve(e.target);if(null!==t){var n=L(t);if(null!==n)if(13===(t=n.tag)){if(null!==(t=B(n)))return e.blockedOn=t,void function(e){var t=z.p;try{return z.p=e,function(){if(13===n.tag){var e=Rc(),t=Ar(n,e);null!==t&&Dc(t,0,e),zd(n,e)}}()}finally{z.p=t}}(e.priority)}else if(3===t&&n.stateNode.current.memoizedState.isDehydrated)return void(e.blockedOn=3===n.tag?n.stateNode.containerInfo:null)}e.blockedOn=null}function cp(e){if(null!==e.blockedOn)return!1;for(var t=e.targetContainers;0<t.length;){var n=Hd(e.nativeEvent);if(null!==n)return null!==(t=He(n))&&jd(t),e.blockedOn=n,!1;var r=new(n=e.nativeEvent).constructor(n.type,n);Rt=r,n.target.dispatchEvent(r),Rt=null,t.shift()}return!0}function up(e,t,n){cp(e)&&n.delete(t)}function dp(){Yd=!1,null!==Jd&&cp(Jd)&&(Jd=null),null!==ep&&cp(ep)&&(ep=null),null!==tp&&cp(tp)&&(tp=null),np.forEach(up),rp.forEach(up)}function pp(e,t){e.blockedOn===t&&(e.blockedOn=null,Yd||(Yd=!0,r.unstable_scheduleCallback(r.unstable_NormalPriority,dp)))}var hp=null;function fp(e){hp!==e&&(hp=e,r.unstable_scheduleCallback(r.unstable_NormalPriority,(function(){hp===e&&(hp=null);for(var t=0;t<e.length;t+=3){var n=e[t],r=e[t+1],a=e[t+2];if("function"!=typeof r){if(null===Gd(r||n))continue;break}var o=He(n);null!==o&&(e.splice(t,3),t-=3,di(o,{pending:!0,data:a,method:n.method,action:r},r,a))}})))}function mp(e){function t(t){return pp(t,e)}null!==Jd&&pp(Jd,e),null!==ep&&pp(ep,e),null!==tp&&pp(tp,e),np.forEach(t),rp.forEach(t);for(var n=0;n<ap.length;n++){var r=ap[n];r.blockedOn===e&&(r.blockedOn=null)}for(;0<ap.length&&null===(n=ap[0]).blockedOn;)lp(n),null===n.blockedOn&&ap.shift();if(null!=(n=(e.ownerDocument||e).$$reactFormReplay))for(r=0;r<n.length;r+=3){var a=n[r],o=n[r+1],i=a[Fe]||null;if("function"==typeof o)i||fp(n);else if(i){var s=null;if(o&&o.hasAttribute("formAction")){if(a=o,i=o[Fe]||null)s=i.formAction;else if(null!==Gd(a))continue}else s=i.action;"function"==typeof s?n[r+1]=s:(n.splice(r,3),r-=3),fp(n)}}}function gp(e){this._internalRoot=e}function vp(e){this._internalRoot=e}vp.prototype.render=gp.prototype.render=function(e){var t=this._internalRoot;if(null===t)throw Error(i(409));Qd(t.current,Rc(),e,t,null,null)},vp.prototype.unmount=gp.prototype.unmount=function(){var e=this._internalRoot;if(null!==e){this._internalRoot=null;var t=e.containerInfo;0===e.tag&&tu(),Qd(e.current,2,null,e,null,null),Bc(),t[Qe]=null}},vp.prototype.unstable_scheduleHydration=function(e){if(e){var t=Ie();e={blockedOn:null,target:e,priority:t};for(var n=0;n<ap.length&&0!==t&&t<ap[n].priority;n++);ap.splice(n,0,e),0===n&&lp(e)}};var bp=a.version;if("19.0.0"!==bp)throw Error(i(527,bp,"19.0.0"));z.findDOMNode=function(e){var t=e._reactInternals;if(void 0===t){if("function"==typeof e.render)throw Error(i(188));throw e=Object.keys(e).join(","),Error(i(268,e))}return e=function(e){var t=e.alternate;if(!t){if(null===(t=L(e)))throw Error(i(188));return t!==e?null:e}for(var n=e,r=t;;){var a=n.return;if(null===a)break;var o=a.alternate;if(null===o){if(null!==(r=a.return)){n=r;continue}break}if(a.child===o.child){for(o=a.child;o;){if(o===n)return F(a),e;if(o===r)return F(a),t;o=o.sibling}throw Error(i(188))}if(n.return!==r.return)n=a,r=o;else{for(var s=!1,l=a.child;l;){if(l===n){s=!0,n=a,r=o;break}if(l===r){s=!0,r=a,n=o;break}l=l.sibling}if(!s){for(l=o.child;l;){if(l===n){s=!0,n=o,r=a;break}if(l===r){s=!0,r=o,n=a;break}l=l.sibling}if(!s)throw Error(i(189))}}if(n.alternate!==r)throw Error(i(190))}if(3!==n.tag)throw Error(i(188));return n.stateNode.current===n?e:t}(t),null===(e=null!==e?Q(e):null)?null:e.stateNode};var yp={bundleType:0,version:"19.0.0",rendererPackageName:"react-dom",currentDispatcherRef:R,findFiberByHostInstance:Ve,reconcilerVersion:"19.0.0"};if("undefined"!=typeof __REACT_DEVTOOLS_GLOBAL_HOOK__){var wp=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(!wp.isDisabled&&wp.supportsFiber)try{ve=wp.inject(yp),be=wp}catch(e){}}t.createRoot=function(e,t){if(!s(e))throw Error(i(299));var n=!1,r="",a=Ni,o=Ii,l=Li;return null!=t&&(!0===t.unstable_strictMode&&(n=!0),void 0!==t.identifierPrefix&&(r=t.identifierPrefix),void 0!==t.onUncaughtError&&(a=t.onUncaughtError),void 0!==t.onCaughtError&&(o=t.onCaughtError),void 0!==t.onRecoverableError&&(l=t.onRecoverableError),void 0!==t.unstable_transitionCallbacks&&t.unstable_transitionCallbacks),t=Bd(e,1,!1,null,0,n,r,a,o,l,0,null),e[Qe]=t.current,Du(8===e.nodeType?e.parentNode:e),new gp(t)},t.hydrateRoot=function(e,t,n){if(!s(e))throw Error(i(299));var r=!1,a="",o=Ni,l=Ii,c=Li,u=null;return null!=n&&(!0===n.unstable_strictMode&&(r=!0),void 0!==n.identifierPrefix&&(a=n.identifierPrefix),void 0!==n.onUncaughtError&&(o=n.onUncaughtError),void 0!==n.onCaughtError&&(l=n.onCaughtError),void 0!==n.onRecoverableError&&(c=n.onRecoverableError),void 0!==n.unstable_transitionCallbacks&&n.unstable_transitionCallbacks,void 0!==n.formState&&(u=n.formState)),(t=Bd(e,1,!0,t,0,r,a,o,l,c,0,u)).context=Fd(null),n=t.current,(a=Ps(r=Rc())).callback=null,Ds(n,a,r),t.current.lanes=r,De(t,r),mu(t),e[Qe]=t.current,Du(e),new vp(t)},t.version="19.0.0"},1259:(e,t,n)=>{"use strict";n.d(t,{v:()=>i});var r=n(7457),a=n(5147),o=n(7176);function i(e){const t={},n={};for(const[o,i]of Object.entries(e.properties)){const s=new a.E(o,e.transform(e.attributes||{},o),i,e.space);e.mustUseProperty&&e.mustUseProperty.includes(o)&&(s.mustUseProperty=!0),t[o]=s,n[(0,r.S)(o)]=o,n[(0,r.S)(s.attribute)]=o}return new o.S(t,n,e.space)}},1266:(e,t,n)=>{"use strict";n.d(t,{RO:()=>er,z5:()=>Ed,Jt:()=>Ad,en:()=>Cd,ev:()=>Wh,BR:()=>Xh,SM:()=>IO,kZ:()=>YO,t:()=>tk,rA:()=>Sk,dR:()=>NO,XO:()=>FO,Z5:()=>Uh,FI:()=>zh,kv:()=>wd,EX:()=>zO,cy:()=>qO,Pq:()=>Dd,XI:()=>Kn,mi:()=>ZO,uT:()=>ex,w:()=>nk,vW:()=>rx,rQ:()=>dk,oN:()=>rk,iQ:()=>fk,SL:()=>yk,N9:()=>Qh,ks:()=>Ok,KB:()=>cx,iy:()=>kk,$M:()=>xk,AG:()=>ux,Fj:()=>Fn,cx:()=>Qn,AX:()=>zn,R2:()=>Hn,hS:()=>Yk});var r={};n.r(r),n.d(r,{default:()=>Mr,render:()=>Dr});var a={};n.r(a),n.d(a,{default:()=>Ir,render:()=>Nr});var o={};n.r(o),n.d(o,{default:()=>Fr,render:()=>Br});var i={};n.r(i),n.d(i,{default:()=>zr,render:()=>Ur});var s={};n.r(s),n.d(s,{default:()=>Xr,render:()=>Wr});var l={};n.r(l),n.d(l,{default:()=>Hr,render:()=>Vr});var c={};n.r(c),n.d(c,{default:()=>Kr,render:()=>Gr});var u={};n.r(u),n.d(u,{default:()=>ea,render:()=>Jr});var d={};n.r(d),n.d(d,{default:()=>ra,render:()=>na});var p={};n.r(p),n.d(p,{default:()=>ia,render:()=>oa});var h={};n.r(h),n.d(h,{default:()=>ca,render:()=>la});var f={};n.r(f),n.d(f,{default:()=>pa,render:()=>da});var m={};n.r(m),n.d(m,{default:()=>ma,render:()=>fa});var g={};n.r(g),n.d(g,{default:()=>ba,render:()=>va});var v={};n.r(v),n.d(v,{default:()=>Oa,render:()=>wa});var b={};n.r(b),n.d(b,{default:()=>Ea,render:()=>xa});var y={};n.r(y),n.d(y,{default:()=>Ca,render:()=>_a});var w={};n.r(w),n.d(w,{default:()=>Ra,render:()=>Aa});var O={};n.r(O),n.d(O,{default:()=>Ma,render:()=>Da});var k={};n.r(k),n.d(k,{default:()=>Ia,render:()=>Na});var x={};n.r(x),n.d(x,{default:()=>Fa,render:()=>Ba});var E={};n.r(E),n.d(E,{default:()=>za,render:()=>Ua});var S={};n.r(S),n.d(S,{default:()=>Xa,render:()=>Wa});var _={};n.r(_),n.d(_,{default:()=>Ha,render:()=>Va});var C={};n.r(C),n.d(C,{default:()=>Ka,render:()=>Ga});var T={};n.r(T),n.d(T,{default:()=>eo,render:()=>Ja});var A={};n.r(A),n.d(A,{default:()=>ro,render:()=>no});var R={};n.r(R),n.d(R,{default:()=>io,render:()=>oo});var P={};n.r(P),n.d(P,{default:()=>co,render:()=>lo});var D={};n.r(D),n.d(D,{default:()=>ho,render:()=>po});var M={};n.r(M),n.d(M,{default:()=>go,render:()=>mo});var $={};n.r($),n.d($,{default:()=>yo,render:()=>bo});var N={};n.r(N),n.d(N,{default:()=>ko,render:()=>Oo});var I={};n.r(I),n.d(I,{default:()=>So,render:()=>Eo});var L={};n.r(L),n.d(L,{default:()=>To,render:()=>Co});var B={};n.r(B),n.d(B,{default:()=>Po,render:()=>Ro});var F={};n.r(F),n.d(F,{default:()=>$o,render:()=>Mo});var Q={};n.r(Q),n.d(Q,{default:()=>Lo,render:()=>Io});var U={};n.r(U),n.d(U,{default:()=>Qo,render:()=>Fo});var z={};n.r(z),n.d(z,{default:()=>jo,render:()=>zo});var j={};n.r(j),n.d(j,{default:()=>qo,render:()=>Xo});var W={};n.r(W),n.d(W,{default:()=>Zo,render:()=>Ho});var X={};n.r(X),n.d(X,{default:()=>Yo,render:()=>Ko});var q={};n.r(q),n.d(q,{default:()=>ti,render:()=>ei});var V={};n.r(V),n.d(V,{default:()=>ai,render:()=>ri});var H={};n.r(H),n.d(H,{default:()=>si,render:()=>ii});var Z={};n.r(Z),n.d(Z,{default:()=>ui,render:()=>ci});var G={};n.r(G),n.d(G,{default:()=>hi,render:()=>pi});var K={};n.r(K),n.d(K,{default:()=>gi,render:()=>mi});var Y={};n.r(Y),n.d(Y,{default:()=>yi,render:()=>bi});var J={};n.r(J),n.d(J,{default:()=>ki,render:()=>Oi});var ee={};n.r(ee),n.d(ee,{default:()=>Si,render:()=>Ei});var te={};n.r(te),n.d(te,{default:()=>Ti,render:()=>Ci});var ne={};n.r(ne),n.d(ne,{default:()=>Pi,render:()=>Ri});var re={};n.r(re),n.d(re,{default:()=>$i,render:()=>Mi});var ae={};n.r(ae),n.d(ae,{default:()=>Li,render:()=>Ii});var oe={};n.r(oe),n.d(oe,{default:()=>Qi,render:()=>Fi});var ie={};n.r(ie),n.d(ie,{default:()=>ji,render:()=>zi});var se={};n.r(se),n.d(se,{default:()=>qi,render:()=>Xi});var le={};n.r(le),n.d(le,{default:()=>Zi,render:()=>Hi});var ce={};n.r(ce),n.d(ce,{default:()=>Yi,render:()=>Ki});var ue={};n.r(ue),n.d(ue,{default:()=>ts,render:()=>es});var de={};n.r(de),n.d(de,{default:()=>as,render:()=>rs});var pe={};n.r(pe),n.d(pe,{default:()=>ss,render:()=>is});var he={};n.r(he),n.d(he,{default:()=>us,render:()=>cs});var fe={};n.r(fe),n.d(fe,{default:()=>hs,render:()=>ps});var me={};n.r(me),n.d(me,{default:()=>gs,render:()=>ms});var ge={};n.r(ge),n.d(ge,{default:()=>ys,render:()=>bs});var ve={};n.r(ve),n.d(ve,{default:()=>ks,render:()=>Os});var be={};n.r(be),n.d(be,{default:()=>Ss,render:()=>Es});var ye={};n.r(ye),n.d(ye,{default:()=>Ts,render:()=>Cs});var we={};n.r(we),n.d(we,{default:()=>Ps,render:()=>Rs});var Oe={};n.r(Oe),n.d(Oe,{default:()=>$s,render:()=>Ms});var ke={};n.r(ke),n.d(ke,{default:()=>Ls,render:()=>Is});var xe={};n.r(xe),n.d(xe,{default:()=>Qs,render:()=>Fs});var Ee={};n.r(Ee),n.d(Ee,{default:()=>js,render:()=>zs});var Se={};n.r(Se),n.d(Se,{default:()=>qs,render:()=>Xs});var _e={};n.r(_e),n.d(_e,{default:()=>Zs,render:()=>Hs});var Ce={};n.r(Ce),n.d(Ce,{default:()=>Ys,render:()=>Ks});var Te={};n.r(Te),n.d(Te,{default:()=>tl,render:()=>el});var Ae={};n.r(Ae),n.d(Ae,{default:()=>al,render:()=>rl});var Re={};n.r(Re),n.d(Re,{default:()=>sl,render:()=>il});var Pe={};n.r(Pe),n.d(Pe,{default:()=>ul,render:()=>cl});var De={};n.r(De),n.d(De,{default:()=>hl,render:()=>pl});var Me={};n.r(Me),n.d(Me,{default:()=>gl,render:()=>ml});var $e={};n.r($e),n.d($e,{default:()=>yl,render:()=>bl});var Ne={};n.r(Ne),n.d(Ne,{default:()=>kl,render:()=>Ol});var Ie={};n.r(Ie),n.d(Ie,{default:()=>Sl,render:()=>El});var Le={};n.r(Le),n.d(Le,{default:()=>Tl,render:()=>Cl});var Be={};n.r(Be),n.d(Be,{default:()=>Pl,render:()=>Rl});var Fe={};n.r(Fe),n.d(Fe,{default:()=>$l,render:()=>Ml});var Qe={};n.r(Qe),n.d(Qe,{default:()=>Ll,render:()=>Il});var Ue={};n.r(Ue),n.d(Ue,{default:()=>Ql,render:()=>Fl});var ze={};n.r(ze),n.d(ze,{default:()=>jl,render:()=>zl});var je={};n.r(je),n.d(je,{default:()=>ql,render:()=>Xl});var We={};n.r(We),n.d(We,{default:()=>Zl,render:()=>Hl});var Xe={};n.r(Xe),n.d(Xe,{default:()=>Yl,render:()=>Kl});var qe={};n.r(qe),n.d(qe,{default:()=>tc,render:()=>ec});var Ve={};n.r(Ve),n.d(Ve,{default:()=>ac,render:()=>rc});var He={};n.r(He),n.d(He,{default:()=>sc,render:()=>ic});var Ze={};n.r(Ze),n.d(Ze,{default:()=>uc,render:()=>cc});var Ge={};n.r(Ge),n.d(Ge,{default:()=>hc,render:()=>pc});var Ke={};n.r(Ke),n.d(Ke,{default:()=>gc,render:()=>mc});var Ye={};n.r(Ye),n.d(Ye,{default:()=>yc,render:()=>bc});var Je={};n.r(Je),n.d(Je,{default:()=>kc,render:()=>Oc});var et={};n.r(et),n.d(et,{default:()=>Sc,render:()=>Ec});var tt={};n.r(tt),n.d(tt,{default:()=>Tc,render:()=>Cc});var nt={};n.r(nt),n.d(nt,{default:()=>Pc,render:()=>Rc});var rt={};n.r(rt),n.d(rt,{default:()=>$c,render:()=>Mc});var at={};n.r(at),n.d(at,{default:()=>Lc,render:()=>Ic});var ot={};n.r(ot),n.d(ot,{default:()=>Qc,render:()=>Fc});var it={};n.r(it),n.d(it,{default:()=>jc,render:()=>zc});var st={};n.r(st),n.d(st,{default:()=>qc,render:()=>Xc});var lt={};n.r(lt),n.d(lt,{default:()=>Zc,render:()=>Hc});var ct={};n.r(ct),n.d(ct,{default:()=>Yc,render:()=>Kc});var ut={};n.r(ut),n.d(ut,{default:()=>tu,render:()=>eu});var dt={};n.r(dt),n.d(dt,{default:()=>au,render:()=>ru});var pt={};n.r(pt),n.d(pt,{default:()=>su,render:()=>iu});var ht={};n.r(ht),n.d(ht,{default:()=>uu,render:()=>cu});var ft={};n.r(ft),n.d(ft,{default:()=>hu,render:()=>pu});var mt={};n.r(mt),n.d(mt,{default:()=>gu,render:()=>mu});var gt={};n.r(gt),n.d(gt,{default:()=>yu,render:()=>bu});var vt={};n.r(vt),n.d(vt,{default:()=>ku,render:()=>Ou});var bt={};n.r(bt),n.d(bt,{default:()=>Su,render:()=>Eu});var yt={};n.r(yt),n.d(yt,{default:()=>Tu,render:()=>Cu});var wt={};n.r(wt),n.d(wt,{default:()=>Pu,render:()=>Ru});var Ot={};n.r(Ot),n.d(Ot,{default:()=>$u,render:()=>Mu});var kt={};n.r(kt),n.d(kt,{default:()=>Lu,render:()=>Iu});var xt={};n.r(xt),n.d(xt,{default:()=>Qu,render:()=>Fu});var Et={};n.r(Et),n.d(Et,{default:()=>ju,render:()=>zu});var St={};n.r(St),n.d(St,{default:()=>qu,render:()=>Xu});var _t={};n.r(_t),n.d(_t,{default:()=>Zu,render:()=>Hu});var Ct={};n.r(Ct),n.d(Ct,{default:()=>Yu,render:()=>Ku});var Tt={};n.r(Tt),n.d(Tt,{default:()=>td,render:()=>ed});var At={};n.r(At),n.d(At,{default:()=>ad,render:()=>rd});var Rt={};n.r(Rt),n.d(Rt,{default:()=>sd,render:()=>id});var Pt={};n.r(Pt),n.d(Pt,{default:()=>ud,render:()=>cd});var Dt={};n.r(Dt),n.d(Dt,{default:()=>hd,render:()=>pd});var Mt={};n.r(Mt),n.d(Mt,{default:()=>gd,render:()=>md});var $t=n(641);function Nt(e){var t,n,r="";if("string"==typeof e||"number"==typeof e)r+=e;else if("object"==typeof e)if(Array.isArray(e)){var a=e.length;for(t=0;t<a;t++)e[t]&&(n=Nt(e[t]))&&(r&&(r+=" "),r+=n)}else for(n in e)e[n]&&(r&&(r+=" "),r+=n);return r}function It(){for(var e,t,n=0,r="",a=arguments.length;n<a;n++)(e=arguments[n])&&(t=Nt(e))&&(r&&(r+=" "),r+=t);return r}const Lt=e=>"boolean"==typeof e?`${e}`:0===e?"0":e,Bt=e=>{const t=function(){for(var t=arguments.length,n=new Array(t),r=0;r<t;r++)n[r]=arguments[r];var a,o;return void 0!==(null==e||null===(a=e.hooks)||void 0===a?void 0:a["cx:done"])?null==e?void 0:e.hooks["cx:done"](It(n)):void 0!==(null==e||null===(o=e.hooks)||void 0===o?void 0:o.onComplete)?null==e?void 0:e.hooks.onComplete(It(n)):It(n)};return{compose:function(){for(var e=arguments.length,n=new Array(e),r=0;r<e;r++)n[r]=arguments[r];return e=>{const r=Object.fromEntries(Object.entries(e||{}).filter((e=>{let[t]=e;return!["class","className"].includes(t)})));return t(n.map((e=>e(r))),null==e?void 0:e.class,null==e?void 0:e.className)}},cva:e=>n=>{var r;if(null==(null==e?void 0:e.variants))return t(null==e?void 0:e.base,null==n?void 0:n.class,null==n?void 0:n.className);const{variants:a,defaultVariants:o}=e,i=Object.keys(a).map((e=>{const t=null==n?void 0:n[e],r=null==o?void 0:o[e],i=Lt(t)||Lt(r);return a[e][i]})),s={...o,...n&&Object.entries(n).reduce(((e,t)=>{let[n,r]=t;return void 0===r?e:{...e,[n]:r}}),{})},l=null==e||null===(r=e.compoundVariants)||void 0===r?void 0:r.reduce(((e,t)=>{let{class:n,className:r,...a}=t;return Object.entries(a).every((e=>{let[t,n]=e;const r=s[t];return Array.isArray(n)?n.includes(r):r===n}))?[...e,n,r]:e}),[]);return t(null==e?void 0:e.base,i,l,null==n?void 0:n.class,null==n?void 0:n.className)},cx:t}},{compose:Ft,cva:Qt,cx:Ut}=Bt(),zt=e=>{const t=qt(e),{conflictingClassGroups:n,conflictingClassGroupModifiers:r}=e;return{getClassGroupId:e=>{const n=e.split("-");return""===n[0]&&1!==n.length&&n.shift(),jt(n,t)||Xt(e)},getConflictingClassGroupIds:(e,t)=>{const a=n[e]||[];return t&&r[e]?[...a,...r[e]]:a}}},jt=(e,t)=>{if(0===e.length)return t.classGroupId;const n=e[0],r=t.nextPart.get(n),a=r?jt(e.slice(1),r):void 0;if(a)return a;if(0===t.validators.length)return;const o=e.join("-");return t.validators.find((({validator:e})=>e(o)))?.classGroupId},Wt=/^\[(.+)\]$/,Xt=e=>{if(Wt.test(e)){const t=Wt.exec(e)[1],n=t?.substring(0,t.indexOf(":"));if(n)return"arbitrary.."+n}},qt=e=>{const{theme:t,prefix:n}=e,r={nextPart:new Map,validators:[]};return Gt(Object.entries(e.classGroups),n).forEach((([e,n])=>{Vt(n,r,e,t)})),r},Vt=(e,t,n,r)=>{e.forEach((e=>{if("string"!=typeof e){if("function"==typeof e)return Zt(e)?void Vt(e(r),t,n,r):void t.validators.push({validator:e,classGroupId:n});Object.entries(e).forEach((([e,a])=>{Vt(a,Ht(t,e),n,r)}))}else(""===e?t:Ht(t,e)).classGroupId=n}))},Ht=(e,t)=>{let n=e;return t.split("-").forEach((e=>{n.nextPart.has(e)||n.nextPart.set(e,{nextPart:new Map,validators:[]}),n=n.nextPart.get(e)})),n},Zt=e=>e.isThemeGetter,Gt=(e,t)=>t?e.map((([e,n])=>[e,n.map((e=>"string"==typeof e?t+e:"object"==typeof e?Object.fromEntries(Object.entries(e).map((([e,n])=>[t+e,n]))):e))])):e,Kt=e=>{if(e<1)return{get:()=>{},set:()=>{}};let t=0,n=new Map,r=new Map;const a=(a,o)=>{n.set(a,o),t++,t>e&&(t=0,r=n,n=new Map)};return{get(e){let t=n.get(e);return void 0!==t?t:void 0!==(t=r.get(e))?(a(e,t),t):void 0},set(e,t){n.has(e)?n.set(e,t):a(e,t)}}},Yt=e=>{const{separator:t,experimentalParseClassName:n}=e,r=1===t.length,a=t[0],o=t.length,i=e=>{const n=[];let i,s=0,l=0;for(let c=0;c<e.length;c++){let u=e[c];if(0===s){if(u===a&&(r||e.slice(c,c+o)===t)){n.push(e.slice(l,c)),l=c+o;continue}if("/"===u){i=c;continue}}"["===u?s++:"]"===u&&s--}const c=0===n.length?e:e.substring(l),u=c.startsWith("!");return{modifiers:n,hasImportantModifier:u,baseClassName:u?c.substring(1):c,maybePostfixModifierPosition:i&&i>l?i-l:void 0}};return n?e=>n({className:e,parseClassName:i}):i},Jt=e=>{if(e.length<=1)return e;const t=[];let n=[];return e.forEach((e=>{"["===e[0]?(t.push(...n.sort(),e),n=[]):n.push(e)})),t.push(...n.sort()),t},en=/\s+/;function tn(){let e,t,n=0,r="";for(;n<arguments.length;)(e=arguments[n++])&&(t=nn(e))&&(r&&(r+=" "),r+=t);return r}const nn=e=>{if("string"==typeof e)return e;let t,n="";for(let r=0;r<e.length;r++)e[r]&&(t=nn(e[r]))&&(n&&(n+=" "),n+=t);return n};function rn(e,...t){let n,r,a,o=function(s){const l=t.reduce(((e,t)=>t(e)),e());return n=(e=>({cache:Kt(e.cacheSize),parseClassName:Yt(e),...zt(e)}))(l),r=n.cache.get,a=n.cache.set,o=i,i(s)};function i(e){const t=r(e);if(t)return t;const o=((e,t)=>{const{parseClassName:n,getClassGroupId:r,getConflictingClassGroupIds:a}=t,o=[],i=e.trim().split(en);let s="";for(let e=i.length-1;e>=0;e-=1){const t=i[e],{modifiers:l,hasImportantModifier:c,baseClassName:u,maybePostfixModifierPosition:d}=n(t);let p=Boolean(d),h=r(p?u.substring(0,d):u);if(!h){if(!p){s=t+(s.length>0?" "+s:s);continue}if(h=r(u),!h){s=t+(s.length>0?" "+s:s);continue}p=!1}const f=Jt(l).join(":"),m=c?f+"!":f,g=m+h;if(o.includes(g))continue;o.push(g);const v=a(h,p);for(let e=0;e<v.length;++e){const t=v[e];o.push(m+t)}s=t+(s.length>0?" "+s:s)}return s})(e,n);return a(e,o),o}return function(){return o(tn.apply(null,arguments))}}const an=e=>{const t=t=>t[e]||[];return t.isThemeGetter=!0,t},on=/^\[(?:([a-z-]+):)?(.+)\]$/i,sn=/^\d+\/\d+$/,ln=new Set(["px","full","screen"]),cn=/^(\d+(\.\d+)?)?(xs|sm|md|lg|xl)$/,un=/\d+(%|px|r?em|[sdl]?v([hwib]|min|max)|pt|pc|in|cm|mm|cap|ch|ex|r?lh|cq(w|h|i|b|min|max))|\b(calc|min|max|clamp)\(.+\)|^0$/,dn=/^(rgba?|hsla?|hwb|(ok)?(lab|lch))\(.+\)$/,pn=/^(inset_)?-?((\d+)?\.?(\d+)[a-z]+|0)_-?((\d+)?\.?(\d+)[a-z]+|0)/,hn=/^(url|image|image-set|cross-fade|element|(repeating-)?(linear|radial|conic)-gradient)\(.+\)$/,fn=e=>gn(e)||ln.has(e)||sn.test(e),mn=e=>An(e,"length",Rn),gn=e=>Boolean(e)&&!Number.isNaN(Number(e)),vn=e=>An(e,"number",gn),bn=e=>Boolean(e)&&Number.isInteger(Number(e)),yn=e=>e.endsWith("%")&&gn(e.slice(0,-1)),wn=e=>on.test(e),On=e=>cn.test(e),kn=new Set(["length","size","percentage"]),xn=e=>An(e,kn,Pn),En=e=>An(e,"position",Pn),Sn=new Set(["image","url"]),_n=e=>An(e,Sn,Mn),Cn=e=>An(e,"",Dn),Tn=()=>!0,An=(e,t,n)=>{const r=on.exec(e);return!!r&&(r[1]?"string"==typeof t?r[1]===t:t.has(r[1]):n(r[2]))},Rn=e=>un.test(e)&&!dn.test(e),Pn=()=>!1,Dn=e=>pn.test(e),Mn=e=>hn.test(e),$n=(Symbol.toStringTag,()=>{const e=an("colors"),t=an("spacing"),n=an("blur"),r=an("brightness"),a=an("borderColor"),o=an("borderRadius"),i=an("borderSpacing"),s=an("borderWidth"),l=an("contrast"),c=an("grayscale"),u=an("hueRotate"),d=an("invert"),p=an("gap"),h=an("gradientColorStops"),f=an("gradientColorStopPositions"),m=an("inset"),g=an("margin"),v=an("opacity"),b=an("padding"),y=an("saturate"),w=an("scale"),O=an("sepia"),k=an("skew"),x=an("space"),E=an("translate"),S=()=>["auto",wn,t],_=()=>[wn,t],C=()=>["",fn,mn],T=()=>["auto",gn,wn],A=()=>["","0",wn],R=()=>[gn,wn];return{cacheSize:500,separator:":",theme:{colors:[Tn],spacing:[fn,mn],blur:["none","",On,wn],brightness:R(),borderColor:[e],borderRadius:["none","","full",On,wn],borderSpacing:_(),borderWidth:C(),contrast:R(),grayscale:A(),hueRotate:R(),invert:A(),gap:_(),gradientColorStops:[e],gradientColorStopPositions:[yn,mn],inset:S(),margin:S(),opacity:R(),padding:_(),saturate:R(),scale:R(),sepia:A(),skew:R(),space:_(),translate:_()},classGroups:{aspect:[{aspect:["auto","square","video",wn]}],container:["container"],columns:[{columns:[On]}],"break-after":[{"break-after":["auto","avoid","all","avoid-page","page","left","right","column"]}],"break-before":[{"break-before":["auto","avoid","all","avoid-page","page","left","right","column"]}],"break-inside":[{"break-inside":["auto","avoid","avoid-page","avoid-column"]}],"box-decoration":[{"box-decoration":["slice","clone"]}],box:[{box:["border","content"]}],display:["block","inline-block","inline","flex","inline-flex","table","inline-table","table-caption","table-cell","table-column","table-column-group","table-footer-group","table-header-group","table-row-group","table-row","flow-root","grid","inline-grid","contents","list-item","hidden"],float:[{float:["right","left","none","start","end"]}],clear:[{clear:["left","right","both","none","start","end"]}],isolation:["isolate","isolation-auto"],"object-fit":[{object:["contain","cover","fill","none","scale-down"]}],"object-position":[{object:["bottom","center","left","left-bottom","left-top","right","right-bottom","right-top","top",wn]}],overflow:[{overflow:["auto","hidden","clip","visible","scroll"]}],"overflow-x":[{"overflow-x":["auto","hidden","clip","visible","scroll"]}],"overflow-y":[{"overflow-y":["auto","hidden","clip","visible","scroll"]}],overscroll:[{overscroll:["auto","contain","none"]}],"overscroll-x":[{"overscroll-x":["auto","contain","none"]}],"overscroll-y":[{"overscroll-y":["auto","contain","none"]}],position:["static","fixed","absolute","relative","sticky"],inset:[{inset:[m]}],"inset-x":[{"inset-x":[m]}],"inset-y":[{"inset-y":[m]}],start:[{start:[m]}],end:[{end:[m]}],top:[{top:[m]}],right:[{right:[m]}],bottom:[{bottom:[m]}],left:[{left:[m]}],visibility:["visible","invisible","collapse"],z:[{z:["auto",bn,wn]}],basis:[{basis:S()}],"flex-direction":[{flex:["row","row-reverse","col","col-reverse"]}],"flex-wrap":[{flex:["wrap","wrap-reverse","nowrap"]}],flex:[{flex:["1","auto","initial","none",wn]}],grow:[{grow:A()}],shrink:[{shrink:A()}],order:[{order:["first","last","none",bn,wn]}],"grid-cols":[{"grid-cols":[Tn]}],"col-start-end":[{col:["auto",{span:["full",bn,wn]},wn]}],"col-start":[{"col-start":T()}],"col-end":[{"col-end":T()}],"grid-rows":[{"grid-rows":[Tn]}],"row-start-end":[{row:["auto",{span:[bn,wn]},wn]}],"row-start":[{"row-start":T()}],"row-end":[{"row-end":T()}],"grid-flow":[{"grid-flow":["row","col","dense","row-dense","col-dense"]}],"auto-cols":[{"auto-cols":["auto","min","max","fr",wn]}],"auto-rows":[{"auto-rows":["auto","min","max","fr",wn]}],gap:[{gap:[p]}],"gap-x":[{"gap-x":[p]}],"gap-y":[{"gap-y":[p]}],"justify-content":[{justify:["normal","start","end","center","between","around","evenly","stretch"]}],"justify-items":[{"justify-items":["start","end","center","stretch"]}],"justify-self":[{"justify-self":["auto","start","end","center","stretch"]}],"align-content":[{content:["normal","start","end","center","between","around","evenly","stretch","baseline"]}],"align-items":[{items:["start","end","center","baseline","stretch"]}],"align-self":[{self:["auto","start","end","center","stretch","baseline"]}],"place-content":[{"place-content":["start","end","center","between","around","evenly","stretch","baseline"]}],"place-items":[{"place-items":["start","end","center","baseline","stretch"]}],"place-self":[{"place-self":["auto","start","end","center","stretch"]}],p:[{p:[b]}],px:[{px:[b]}],py:[{py:[b]}],ps:[{ps:[b]}],pe:[{pe:[b]}],pt:[{pt:[b]}],pr:[{pr:[b]}],pb:[{pb:[b]}],pl:[{pl:[b]}],m:[{m:[g]}],mx:[{mx:[g]}],my:[{my:[g]}],ms:[{ms:[g]}],me:[{me:[g]}],mt:[{mt:[g]}],mr:[{mr:[g]}],mb:[{mb:[g]}],ml:[{ml:[g]}],"space-x":[{"space-x":[x]}],"space-x-reverse":["space-x-reverse"],"space-y":[{"space-y":[x]}],"space-y-reverse":["space-y-reverse"],w:[{w:["auto","min","max","fit","svw","lvw","dvw",wn,t]}],"min-w":[{"min-w":[wn,t,"min","max","fit"]}],"max-w":[{"max-w":[wn,t,"none","full","min","max","fit","prose",{screen:[On]},On]}],h:[{h:[wn,t,"auto","min","max","fit","svh","lvh","dvh"]}],"min-h":[{"min-h":[wn,t,"min","max","fit","svh","lvh","dvh"]}],"max-h":[{"max-h":[wn,t,"min","max","fit","svh","lvh","dvh"]}],size:[{size:[wn,t,"auto","min","max","fit"]}],"font-size":[{text:["base",On,mn]}],"font-smoothing":["antialiased","subpixel-antialiased"],"font-style":["italic","not-italic"],"font-weight":[{font:["thin","extralight","light","normal","medium","semibold","bold","extrabold","black",vn]}],"font-family":[{font:[Tn]}],"fvn-normal":["normal-nums"],"fvn-ordinal":["ordinal"],"fvn-slashed-zero":["slashed-zero"],"fvn-figure":["lining-nums","oldstyle-nums"],"fvn-spacing":["proportional-nums","tabular-nums"],"fvn-fraction":["diagonal-fractions","stacked-fractions"],tracking:[{tracking:["tighter","tight","normal","wide","wider","widest",wn]}],"line-clamp":[{"line-clamp":["none",gn,vn]}],leading:[{leading:["none","tight","snug","normal","relaxed","loose",fn,wn]}],"list-image":[{"list-image":["none",wn]}],"list-style-type":[{list:["none","disc","decimal",wn]}],"list-style-position":[{list:["inside","outside"]}],"placeholder-color":[{placeholder:[e]}],"placeholder-opacity":[{"placeholder-opacity":[v]}],"text-alignment":[{text:["left","center","right","justify","start","end"]}],"text-color":[{text:[e]}],"text-opacity":[{"text-opacity":[v]}],"text-decoration":["underline","overline","line-through","no-underline"],"text-decoration-style":[{decoration:["solid","dashed","dotted","double","none","wavy"]}],"text-decoration-thickness":[{decoration:["auto","from-font",fn,mn]}],"underline-offset":[{"underline-offset":["auto",fn,wn]}],"text-decoration-color":[{decoration:[e]}],"text-transform":["uppercase","lowercase","capitalize","normal-case"],"text-overflow":["truncate","text-ellipsis","text-clip"],"text-wrap":[{text:["wrap","nowrap","balance","pretty"]}],indent:[{indent:_()}],"vertical-align":[{align:["baseline","top","middle","bottom","text-top","text-bottom","sub","super",wn]}],whitespace:[{whitespace:["normal","nowrap","pre","pre-line","pre-wrap","break-spaces"]}],break:[{break:["normal","words","all","keep"]}],hyphens:[{hyphens:["none","manual","auto"]}],content:[{content:["none",wn]}],"bg-attachment":[{bg:["fixed","local","scroll"]}],"bg-clip":[{"bg-clip":["border","padding","content","text"]}],"bg-opacity":[{"bg-opacity":[v]}],"bg-origin":[{"bg-origin":["border","padding","content"]}],"bg-position":[{bg:["bottom","center","left","left-bottom","left-top","right","right-bottom","right-top","top",En]}],"bg-repeat":[{bg:["no-repeat",{repeat:["","x","y","round","space"]}]}],"bg-size":[{bg:["auto","cover","contain",xn]}],"bg-image":[{bg:["none",{"gradient-to":["t","tr","r","br","b","bl","l","tl"]},_n]}],"bg-color":[{bg:[e]}],"gradient-from-pos":[{from:[f]}],"gradient-via-pos":[{via:[f]}],"gradient-to-pos":[{to:[f]}],"gradient-from":[{from:[h]}],"gradient-via":[{via:[h]}],"gradient-to":[{to:[h]}],rounded:[{rounded:[o]}],"rounded-s":[{"rounded-s":[o]}],"rounded-e":[{"rounded-e":[o]}],"rounded-t":[{"rounded-t":[o]}],"rounded-r":[{"rounded-r":[o]}],"rounded-b":[{"rounded-b":[o]}],"rounded-l":[{"rounded-l":[o]}],"rounded-ss":[{"rounded-ss":[o]}],"rounded-se":[{"rounded-se":[o]}],"rounded-ee":[{"rounded-ee":[o]}],"rounded-es":[{"rounded-es":[o]}],"rounded-tl":[{"rounded-tl":[o]}],"rounded-tr":[{"rounded-tr":[o]}],"rounded-br":[{"rounded-br":[o]}],"rounded-bl":[{"rounded-bl":[o]}],"border-w":[{border:[s]}],"border-w-x":[{"border-x":[s]}],"border-w-y":[{"border-y":[s]}],"border-w-s":[{"border-s":[s]}],"border-w-e":[{"border-e":[s]}],"border-w-t":[{"border-t":[s]}],"border-w-r":[{"border-r":[s]}],"border-w-b":[{"border-b":[s]}],"border-w-l":[{"border-l":[s]}],"border-opacity":[{"border-opacity":[v]}],"border-style":[{border:["solid","dashed","dotted","double","none","hidden"]}],"divide-x":[{"divide-x":[s]}],"divide-x-reverse":["divide-x-reverse"],"divide-y":[{"divide-y":[s]}],"divide-y-reverse":["divide-y-reverse"],"divide-opacity":[{"divide-opacity":[v]}],"divide-style":[{divide:["solid","dashed","dotted","double","none"]}],"border-color":[{border:[a]}],"border-color-x":[{"border-x":[a]}],"border-color-y":[{"border-y":[a]}],"border-color-s":[{"border-s":[a]}],"border-color-e":[{"border-e":[a]}],"border-color-t":[{"border-t":[a]}],"border-color-r":[{"border-r":[a]}],"border-color-b":[{"border-b":[a]}],"border-color-l":[{"border-l":[a]}],"divide-color":[{divide:[a]}],"outline-style":[{outline:["","solid","dashed","dotted","double","none"]}],"outline-offset":[{"outline-offset":[fn,wn]}],"outline-w":[{outline:[fn,mn]}],"outline-color":[{outline:[e]}],"ring-w":[{ring:C()}],"ring-w-inset":["ring-inset"],"ring-color":[{ring:[e]}],"ring-opacity":[{"ring-opacity":[v]}],"ring-offset-w":[{"ring-offset":[fn,mn]}],"ring-offset-color":[{"ring-offset":[e]}],shadow:[{shadow:["","inner","none",On,Cn]}],"shadow-color":[{shadow:[Tn]}],opacity:[{opacity:[v]}],"mix-blend":[{"mix-blend":["normal","multiply","screen","overlay","darken","lighten","color-dodge","color-burn","hard-light","soft-light","difference","exclusion","hue","saturation","color","luminosity","plus-lighter","plus-darker"]}],"bg-blend":[{"bg-blend":["normal","multiply","screen","overlay","darken","lighten","color-dodge","color-burn","hard-light","soft-light","difference","exclusion","hue","saturation","color","luminosity"]}],filter:[{filter:["","none"]}],blur:[{blur:[n]}],brightness:[{brightness:[r]}],contrast:[{contrast:[l]}],"drop-shadow":[{"drop-shadow":["","none",On,wn]}],grayscale:[{grayscale:[c]}],"hue-rotate":[{"hue-rotate":[u]}],invert:[{invert:[d]}],saturate:[{saturate:[y]}],sepia:[{sepia:[O]}],"backdrop-filter":[{"backdrop-filter":["","none"]}],"backdrop-blur":[{"backdrop-blur":[n]}],"backdrop-brightness":[{"backdrop-brightness":[r]}],"backdrop-contrast":[{"backdrop-contrast":[l]}],"backdrop-grayscale":[{"backdrop-grayscale":[c]}],"backdrop-hue-rotate":[{"backdrop-hue-rotate":[u]}],"backdrop-invert":[{"backdrop-invert":[d]}],"backdrop-opacity":[{"backdrop-opacity":[v]}],"backdrop-saturate":[{"backdrop-saturate":[y]}],"backdrop-sepia":[{"backdrop-sepia":[O]}],"border-collapse":[{border:["collapse","separate"]}],"border-spacing":[{"border-spacing":[i]}],"border-spacing-x":[{"border-spacing-x":[i]}],"border-spacing-y":[{"border-spacing-y":[i]}],"table-layout":[{table:["auto","fixed"]}],caption:[{caption:["top","bottom"]}],transition:[{transition:["none","all","","colors","opacity","shadow","transform",wn]}],duration:[{duration:R()}],ease:[{ease:["linear","in","out","in-out",wn]}],delay:[{delay:R()}],animate:[{animate:["none","spin","ping","pulse","bounce",wn]}],transform:[{transform:["","gpu","none"]}],scale:[{scale:[w]}],"scale-x":[{"scale-x":[w]}],"scale-y":[{"scale-y":[w]}],rotate:[{rotate:[bn,wn]}],"translate-x":[{"translate-x":[E]}],"translate-y":[{"translate-y":[E]}],"skew-x":[{"skew-x":[k]}],"skew-y":[{"skew-y":[k]}],"transform-origin":[{origin:["center","top","top-right","right","bottom-right","bottom","bottom-left","left","top-left",wn]}],accent:[{accent:["auto",e]}],appearance:[{appearance:["none","auto"]}],cursor:[{cursor:["auto","default","pointer","wait","text","move","help","not-allowed","none","context-menu","progress","cell","crosshair","vertical-text","alias","copy","no-drop","grab","grabbing","all-scroll","col-resize","row-resize","n-resize","e-resize","s-resize","w-resize","ne-resize","nw-resize","se-resize","sw-resize","ew-resize","ns-resize","nesw-resize","nwse-resize","zoom-in","zoom-out",wn]}],"caret-color":[{caret:[e]}],"pointer-events":[{"pointer-events":["none","auto"]}],resize:[{resize:["none","y","x",""]}],"scroll-behavior":[{scroll:["auto","smooth"]}],"scroll-m":[{"scroll-m":_()}],"scroll-mx":[{"scroll-mx":_()}],"scroll-my":[{"scroll-my":_()}],"scroll-ms":[{"scroll-ms":_()}],"scroll-me":[{"scroll-me":_()}],"scroll-mt":[{"scroll-mt":_()}],"scroll-mr":[{"scroll-mr":_()}],"scroll-mb":[{"scroll-mb":_()}],"scroll-ml":[{"scroll-ml":_()}],"scroll-p":[{"scroll-p":_()}],"scroll-px":[{"scroll-px":_()}],"scroll-py":[{"scroll-py":_()}],"scroll-ps":[{"scroll-ps":_()}],"scroll-pe":[{"scroll-pe":_()}],"scroll-pt":[{"scroll-pt":_()}],"scroll-pr":[{"scroll-pr":_()}],"scroll-pb":[{"scroll-pb":_()}],"scroll-pl":[{"scroll-pl":_()}],"snap-align":[{snap:["start","end","center","align-none"]}],"snap-stop":[{snap:["normal","always"]}],"snap-type":[{snap:["none","x","y","both"]}],"snap-strictness":[{snap:["mandatory","proximity"]}],touch:[{touch:["auto","none","manipulation"]}],"touch-x":[{"touch-pan":["x","left","right"]}],"touch-y":[{"touch-pan":["y","up","down"]}],"touch-pz":["touch-pinch-zoom"],select:[{select:["none","text","all","auto"]}],"will-change":[{"will-change":["auto","scroll","contents","transform",wn]}],fill:[{fill:[e,"none"]}],"stroke-w":[{stroke:[fn,mn,vn]}],stroke:[{stroke:[e,"none"]}],sr:["sr-only","not-sr-only"],"forced-color-adjust":[{"forced-color-adjust":["auto","none"]}]},conflictingClassGroups:{overflow:["overflow-x","overflow-y"],overscroll:["overscroll-x","overscroll-y"],inset:["inset-x","inset-y","start","end","top","right","bottom","left"],"inset-x":["right","left"],"inset-y":["top","bottom"],flex:["basis","grow","shrink"],gap:["gap-x","gap-y"],p:["px","py","ps","pe","pt","pr","pb","pl"],px:["pr","pl"],py:["pt","pb"],m:["mx","my","ms","me","mt","mr","mb","ml"],mx:["mr","ml"],my:["mt","mb"],size:["w","h"],"font-size":["leading"],"fvn-normal":["fvn-ordinal","fvn-slashed-zero","fvn-figure","fvn-spacing","fvn-fraction"],"fvn-ordinal":["fvn-normal"],"fvn-slashed-zero":["fvn-normal"],"fvn-figure":["fvn-normal"],"fvn-spacing":["fvn-normal"],"fvn-fraction":["fvn-normal"],"line-clamp":["display","overflow"],rounded:["rounded-s","rounded-e","rounded-t","rounded-r","rounded-b","rounded-l","rounded-ss","rounded-se","rounded-ee","rounded-es","rounded-tl","rounded-tr","rounded-br","rounded-bl"],"rounded-s":["rounded-ss","rounded-es"],"rounded-e":["rounded-se","rounded-ee"],"rounded-t":["rounded-tl","rounded-tr"],"rounded-r":["rounded-tr","rounded-br"],"rounded-b":["rounded-br","rounded-bl"],"rounded-l":["rounded-tl","rounded-bl"],"border-spacing":["border-spacing-x","border-spacing-y"],"border-w":["border-w-s","border-w-e","border-w-t","border-w-r","border-w-b","border-w-l"],"border-w-x":["border-w-r","border-w-l"],"border-w-y":["border-w-t","border-w-b"],"border-color":["border-color-s","border-color-e","border-color-t","border-color-r","border-color-b","border-color-l"],"border-color-x":["border-color-r","border-color-l"],"border-color-y":["border-color-t","border-color-b"],"scroll-m":["scroll-mx","scroll-my","scroll-ms","scroll-me","scroll-mt","scroll-mr","scroll-mb","scroll-ml"],"scroll-mx":["scroll-mr","scroll-ml"],"scroll-my":["scroll-mt","scroll-mb"],"scroll-p":["scroll-px","scroll-py","scroll-ps","scroll-pe","scroll-pt","scroll-pr","scroll-pb","scroll-pl"],"scroll-px":["scroll-pr","scroll-pl"],"scroll-py":["scroll-pt","scroll-pb"],touch:["touch-x","touch-y","touch-pz"],"touch-x":["touch"],"touch-y":["touch"],"touch-pz":["touch"]},conflictingClassGroupModifiers:{"font-size":["leading"]}}}),Nn=(e,t,n)=>{void 0!==n&&(e[t]=n)},In=(e,t)=>{if(t)for(const n in t)Nn(e,n,t[n])},Ln=(e,t)=>{if(t)for(const n in t){const r=t[n];void 0!==r&&(e[n]=(e[n]||[]).concat(r))}},Bn=((e,...t)=>"function"==typeof e?rn($n,e,...t):rn((()=>((e,{cacheSize:t,prefix:n,separator:r,experimentalParseClassName:a,extend:o={},override:i={}})=>{Nn(e,"cacheSize",t),Nn(e,"prefix",n),Nn(e,"separator",r),Nn(e,"experimentalParseClassName",a);for(const t in i)In(e[t],i[t]);for(const t in o)Ln(e[t],o[t]);return e})($n(),e)),...t))({extend:{classGroups:{"font-size":["text-3xs","text-xxs"]}}}),{cva:Fn,cx:Qn,compose:Un}=Bt({hooks:{onComplete:e=>Bn(e)}});function zn(){const e=(0,$t.OA)(),t=(0,$t.EW)((()=>{const{class:t,...n}=e;return{class:t||"",rest:n}}));return{cx:function(...e){return{class:Qn(...e,t.value.class),...t.value.rest}}}}var jn=n(953),Wn=n(33);const Xn={solid:["scalar-button-solid","bg-b-btn text-c-btn shadow-sm focus-visible:border-c-btn active:bg-b-btn active:shadow-none hover:bg-h-btn outline-offset-1"],outlined:["scalar-button-outlined","active:bg-btn-1 border border-solid border-border bg-b-1 text-c-1 hover:bg-b-2 shadow"],ghost:["scalar-button-ghost","bg-transparent text-c-3 active:text-c-1 hocus:text-c-1"],danger:["scalar-button-danger","bg-c-danger text-white active:brightness-90 hocus:brightness-90"]},qn=Fn({base:"scalar-button scalar-row cursor-pointer items-center justify-center rounded font-medium -outline-offset-1",variants:{disabled:{true:"bg-background-2 text-color-3 shadow-none"},fullWidth:{true:"w-full"},size:{sm:"px-2 py-1 text-xs",md:"h-10 px-6 text-sm"},variant:Xn},compoundVariants:[{disabled:!0,variant:["solid","outlined","ghost","danger"],class:"bg-b-2 text-c-3 shadow-none"}]}),Vn={key:0,class:"circular-loader"};function Hn(){return(0,jn.Kh)({isValid:!1,isInvalid:!1,isLoading:!1,startLoading(){this.isLoading=!0},stopLoading(){this.isLoading=!1},validate(e=800,t){this.isInvalid=!1,this.isValid=!0;const n=t?e-300:e;return new Promise((e=>setTimeout(t?()=>this.clear().then((()=>e(!0))):()=>e(!0),n)))},invalidate(e=1100,t){this.isValid=!1,this.isInvalid=!0;const n=t?e-300:e;return new Promise((e=>setTimeout(t?()=>this.clear().then((()=>e(!0))):()=>e(!0),n)))},clear(e=300){return this.isValid=!1,this.isInvalid=!1,this.isLoading=!1,new Promise((t=>{setTimeout((()=>{t(!0)}),e)}))}})}const Zn=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarLoading",props:{loadingState:{},size:{}},setup(e){const{cx:t}=zn(),n=Fn({variants:{size:{xs:"size-3",sm:"size-3.5",md:"size-4",lg:"size-5",xl:"size-6","2xl":"size-8","3xl":"size-10",full:"size-full"}},defaultVariants:{size:"full"}});return(e,r)=>e.loadingState?((0,$t.uX)(),(0,$t.CE)("div",(0,Wn._B)((0,$t.v6)({key:0},(0,jn.R1)(t)("loader-wrapper",(0,jn.R1)(n)({size:e.size})))),[((0,$t.uX)(),(0,$t.CE)("svg",{class:(0,Wn.C4)(["svg-loader",{"icon-is-valid":e.loadingState.isValid,"icon-is-invalid":e.loadingState.isInvalid}]),viewBox:"0 0 100 100",xmlns:"http://www.w3.org/2000/svg","xmlns:xlink":"http://www.w3.org/1999/xlink"},[r[0]||(r[0]=(0,$t.Fv)('<path class="svg-path svg-check-mark" d="m 0 60 l 30 30 l 70 -80" data-v-65265e87></path><path class="svg-path svg-x-mark" d="m 50 50 l 40 -40" data-v-65265e87></path><path class="svg-path svg-x-mark" d="m 50 50 l 40 40" data-v-65265e87></path><path class="svg-path svg-x-mark" d="m 50 50 l -40 -40" data-v-65265e87></path><path class="svg-path svg-x-mark" d="m 50 50 l -40 40" data-v-65265e87></path>',5)),e.loadingState.isLoading?((0,$t.uX)(),(0,$t.CE)("g",Vn,[(0,$t.Lk)("circle",{class:(0,Wn.C4)(["loader-path",{"loader-path-off":e.loadingState.isValid||e.loadingState.isInvalid}]),cx:"50",cy:"50",fill:"none",r:"20","stroke-width":"2"},null,2)])):(0,$t.Q3)("",!0)],2))],16)):(0,$t.Q3)("",!0)}}),Gn=(e,t)=>{const n=e.__vccOpts||e;for(const[e,r]of t)n[e]=r;return n},Kn=Gn(Zn,[["__scopeId","data-v-65265e87"]]),Yn=["ariaDisabled","type"],Jn={key:3,class:"centered-x absolute"},er=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarButton",props:{disabled:{type:Boolean},fullWidth:{type:Boolean,default:!1},loading:{},size:{default:"md"},variant:{default:"solid"},type:{default:"button"}},setup(e){const{cx:t}=zn();return(e,n)=>{var r,a,o,i;return(0,$t.uX)(),(0,$t.CE)("button",(0,$t.v6)({ariaDisabled:e.disabled||void 0,type:e.type},(0,jn.R1)(t)((0,jn.R1)(qn)({fullWidth:e.fullWidth,disabled:e.disabled,size:e.size,variant:e.variant}),{relative:null==(r=e.loading)?void 0:r.isLoading})),[e.$slots.icon?((0,$t.uX)(),(0,$t.CE)("div",{key:0,class:(0,Wn.C4)(["mr-2 h-4 w-4",{invisible:null==(a=e.loading)?void 0:a.isLoading}])},[(0,$t.RG)(e.$slots,"icon")],2)):(0,$t.Q3)("",!0),e.loading?((0,$t.uX)(),(0,$t.CE)("span",{key:1,class:(0,Wn.C4)({invisible:null==(o=e.loading)?void 0:o.isLoading})},[(0,$t.RG)(e.$slots,"default")],2)):(0,$t.RG)(e.$slots,"default",{key:2}),null!=(i=e.loading)&&i.isLoading?((0,$t.uX)(),(0,$t.CE)("div",Jn,[(0,$t.bF)((0,jn.R1)(Kn),{loadingState:e.loading,size:"xs"},null,8,["loadingState"])])):(0,$t.Q3)("",!0)],16,Yn)}}});var tr=n(7471),nr=n(9938),rr=n(4684),ar=n(8826),or=n(5946);const ir={abandonedHeadElementChild:{reason:"Unexpected metadata element after head",description:"Unexpected element after head. Expected the element before `</head>`",url:!1},abruptClosingOfEmptyComment:{reason:"Unexpected abruptly closed empty comment",description:"Unexpected `>` or `->`. Expected `--\x3e` to close comments"},abruptDoctypePublicIdentifier:{reason:"Unexpected abruptly closed public identifier",description:"Unexpected `>`. Expected a closing `\"` or `'` after the public identifier"},abruptDoctypeSystemIdentifier:{reason:"Unexpected abruptly closed system identifier",description:"Unexpected `>`. Expected a closing `\"` or `'` after the identifier identifier"},absenceOfDigitsInNumericCharacterReference:{reason:"Unexpected non-digit at start of numeric character reference",description:"Unexpected `%c`. Expected `[0-9]` for decimal references or `[0-9a-fA-F]` for hexadecimal references"},cdataInHtmlContent:{reason:"Unexpected CDATA section in HTML",description:"Unexpected `<![CDATA[` in HTML. Remove it, use a comment, or encode special characters instead"},characterReferenceOutsideUnicodeRange:{reason:"Unexpected too big numeric character reference",description:"Unexpectedly high character reference. Expected character references to be at most hexadecimal 10ffff (or decimal 1114111)"},closingOfElementWithOpenChildElements:{reason:"Unexpected closing tag with open child elements",description:"Unexpectedly closing tag. Expected other tags to be closed first",url:!1},controlCharacterInInputStream:{reason:"Unexpected control character",description:"Unexpected control character `%x`. Expected a non-control code point, 0x00, or ASCII whitespace"},controlCharacterReference:{reason:"Unexpected control character reference",description:"Unexpectedly control character in reference. Expected a non-control code point, 0x00, or ASCII whitespace"},disallowedContentInNoscriptInHead:{reason:"Disallowed content inside `<noscript>` in `<head>`",description:"Unexpected text character `%c`. Only use text in `<noscript>`s in `<body>`",url:!1},duplicateAttribute:{reason:"Unexpected duplicate attribute",description:"Unexpectedly double attribute. Expected attributes to occur only once"},endTagWithAttributes:{reason:"Unexpected attribute on closing tag",description:"Unexpected attribute. Expected `>` instead"},endTagWithTrailingSolidus:{reason:"Unexpected slash at end of closing tag",description:"Unexpected `%c-1`. Expected `>` instead"},endTagWithoutMatchingOpenElement:{reason:"Unexpected unopened end tag",description:"Unexpected end tag. Expected no end tag or another end tag",url:!1},eofBeforeTagName:{reason:"Unexpected end of file",description:"Unexpected end of file. Expected tag name instead"},eofInCdata:{reason:"Unexpected end of file in CDATA",description:"Unexpected end of file. Expected `]]>` to close the CDATA"},eofInComment:{reason:"Unexpected end of file in comment",description:"Unexpected end of file. Expected `--\x3e` to close the comment"},eofInDoctype:{reason:"Unexpected end of file in doctype",description:"Unexpected end of file. Expected a valid doctype (such as `<!doctype html>`)"},eofInElementThatCanContainOnlyText:{reason:"Unexpected end of file in element that can only contain text",description:"Unexpected end of file. Expected text or a closing tag",url:!1},eofInScriptHtmlCommentLikeText:{reason:"Unexpected end of file in comment inside script",description:"Unexpected end of file. Expected `--\x3e` to close the comment"},eofInTag:{reason:"Unexpected end of file in tag",description:"Unexpected end of file. Expected `>` to close the tag"},incorrectlyClosedComment:{reason:"Incorrectly closed comment",description:"Unexpected `%c-1`. Expected `--\x3e` to close the comment"},incorrectlyOpenedComment:{reason:"Incorrectly opened comment",description:"Unexpected `%c`. Expected `\x3c!--` to open the comment"},invalidCharacterSequenceAfterDoctypeName:{reason:"Invalid sequence after doctype name",description:"Unexpected sequence at `%c`. Expected `public` or `system`"},invalidFirstCharacterOfTagName:{reason:"Invalid first character in tag name",description:"Unexpected `%c`. Expected an ASCII letter instead"},misplacedDoctype:{reason:"Misplaced doctype",description:"Unexpected doctype. Expected doctype before head",url:!1},misplacedStartTagForHeadElement:{reason:"Misplaced `<head>` start tag",description:"Unexpected start tag `<head>`. Expected `<head>` directly after doctype",url:!1},missingAttributeValue:{reason:"Missing attribute value",description:"Unexpected `%c-1`. Expected an attribute value or no `%c-1` instead"},missingDoctype:{reason:"Missing doctype before other content",description:"Expected a `<!doctype html>` before anything else",url:!1},missingDoctypeName:{reason:"Missing doctype name",description:"Unexpected doctype end at `%c`. Expected `html` instead"},missingDoctypePublicIdentifier:{reason:"Missing public identifier in doctype",description:"Unexpected `%c`. Expected identifier for `public` instead"},missingDoctypeSystemIdentifier:{reason:"Missing system identifier in doctype",description:'Unexpected `%c`. Expected identifier for `system` instead (suggested: `"about:legacy-compat"`)'},missingEndTagName:{reason:"Missing name in end tag",description:"Unexpected `%c`. Expected an ASCII letter instead"},missingQuoteBeforeDoctypePublicIdentifier:{reason:"Missing quote before public identifier in doctype",description:"Unexpected `%c`. Expected `\"` or `'` instead"},missingQuoteBeforeDoctypeSystemIdentifier:{reason:"Missing quote before system identifier in doctype",description:"Unexpected `%c`. Expected `\"` or `'` instead"},missingSemicolonAfterCharacterReference:{reason:"Missing semicolon after character reference",description:"Unexpected `%c`. Expected `;` instead"},missingWhitespaceAfterDoctypePublicKeyword:{reason:"Missing whitespace after public identifier in doctype",description:"Unexpected `%c`. Expected ASCII whitespace instead"},missingWhitespaceAfterDoctypeSystemKeyword:{reason:"Missing whitespace after system identifier in doctype",description:"Unexpected `%c`. Expected ASCII whitespace instead"},missingWhitespaceBeforeDoctypeName:{reason:"Missing whitespace before doctype name",description:"Unexpected `%c`. Expected ASCII whitespace instead"},missingWhitespaceBetweenAttributes:{reason:"Missing whitespace between attributes",description:"Unexpected `%c`. Expected ASCII whitespace instead"},missingWhitespaceBetweenDoctypePublicAndSystemIdentifiers:{reason:"Missing whitespace between public and system identifiers in doctype",description:"Unexpected `%c`. Expected ASCII whitespace instead"},nestedComment:{reason:"Unexpected nested comment",description:"Unexpected `\x3c!--`. Expected `--\x3e`"},nestedNoscriptInHead:{reason:"Unexpected nested `<noscript>` in `<head>`",description:"Unexpected `<noscript>`. Expected a closing tag or a meta element",url:!1},nonConformingDoctype:{reason:"Unexpected non-conforming doctype declaration",description:'Expected `<!doctype html>` or `<!doctype html system "about:legacy-compat">`',url:!1},nonVoidHtmlElementStartTagWithTrailingSolidus:{reason:"Unexpected trailing slash on start tag of non-void element",description:"Unexpected `/`. Expected `>` instead"},noncharacterCharacterReference:{reason:"Unexpected noncharacter code point referenced by character reference",description:"Unexpected code point. Do not use noncharacters in HTML"},noncharacterInInputStream:{reason:"Unexpected noncharacter character",description:"Unexpected code point `%x`. Do not use noncharacters in HTML"},nullCharacterReference:{reason:"Unexpected NULL character referenced by character reference",description:"Unexpected code point. Do not use NULL characters in HTML"},openElementsLeftAfterEof:{reason:"Unexpected end of file",description:"Unexpected end of file. Expected closing tag instead",url:!1},surrogateCharacterReference:{reason:"Unexpected surrogate character referenced by character reference",description:"Unexpected code point. Do not use lone surrogate characters in HTML"},surrogateInInputStream:{reason:"Unexpected surrogate character",description:"Unexpected code point `%x`. Do not use lone surrogate characters in HTML"},unexpectedCharacterAfterDoctypeSystemIdentifier:{reason:"Invalid character after system identifier in doctype",description:"Unexpected character at `%c`. Expected `>`"},unexpectedCharacterInAttributeName:{reason:"Unexpected character in attribute name",description:"Unexpected `%c`. Expected whitespace, `/`, `>`, `=`, or probably an ASCII letter"},unexpectedCharacterInUnquotedAttributeValue:{reason:"Unexpected character in unquoted attribute value",description:"Unexpected `%c`. Quote the attribute value to include it"},unexpectedEqualsSignBeforeAttributeName:{reason:"Unexpected equals sign before attribute name",description:"Unexpected `%c`. Add an attribute name before it"},unexpectedNullCharacter:{reason:"Unexpected NULL character",description:"Unexpected code point `%x`. Do not use NULL characters in HTML"},unexpectedQuestionMarkInsteadOfTagName:{reason:"Unexpected question mark instead of tag name",description:"Unexpected `%c`. Expected an ASCII letter instead"},unexpectedSolidusInTag:{reason:"Unexpected slash in tag",description:"Unexpected `%c-1`. Expected it followed by `>` or in a quoted attribute value"},unknownNamedCharacterReference:{reason:"Unexpected unknown named character reference",description:"Unexpected character reference. Expected known named character references"}},sr=/-[a-z]/g,lr=/%c(?:([-+])(\d+))?/g,cr=/%x/g,ur={2:!0,1:!1,0:null},dr={};function pr(e){return e.charAt(1).toUpperCase()}function hr(e){const{emitParseErrors:t,...n}={...this.data("settings"),...e};this.parser=function(e,r){return function(e,t){const n=t||dr,r=n.onerror,a=e instanceof ar.T?e:new ar.T(e),o=n.fragment?rr.oH:rr.qg,i=String(a),s=o(i,{sourceCodeLocationInfo:!0,onParseError:n.onerror?function(e){const t=e.code,o=function(e){return e.replace(sr,pr)}(t),s=n[o],l=null==s||s,c="number"==typeof l?l:l?1:0;if(c){const n=ir[o];(0,tr.ok)(n,"expected known error from `parse5`");const i=new or.o(u(n.reason),{place:{start:{line:e.startLine,column:e.startCol,offset:e.startOffset},end:{line:e.endLine,column:e.endCol,offset:e.endOffset}},ruleId:t,source:"hast-util-from-html"});a.path&&(i.file=a.path,i.name=a.path+":"+i.name),i.fatal=ur[c],i.note=u(n.description),i.url=!1===n.url?void 0:"https://html.spec.whatwg.org/multipage/parsing.html#parse-error-"+t,(0,tr.ok)(r,"`internalOnerror` is not passed if `onerror` is not set"),r(i)}function u(t){return t.replace(lr,(function(t,n,r){const a=(r?Number.parseInt(r,10):0)*("-"===n?-1:1);return"`"===(o=i.charAt(e.startOffset+a))?"` ` `":o;var o})).replace(cr,(function(){return"0x"+i.charCodeAt(e.startOffset).toString(16).toUpperCase()}))}}:null,scriptingEnabled:!1});return(0,nr.D)(s,{file:a,space:n.space,verbose:n.verbose})}(e,{...n,onerror:t?function(e){r.path&&(e.name=r.path+":"+e.name,e.file=r.path),r.messages.push(e)}:void 0})}}var fr=n(8697),mr=n(9635),gr=n(7970),vr=n(9751);function br(e){return"text"===e?.type}function yr(e){return"element"===e?.type}function wr(e){return{type:"text",value:e}}function Or(){return e=>{(0,gr.YR)(e,"element",((e,t,n)=>{if("element"===n?.type&&"pre"===n.tagName&&"code"===e.tagName){let t=0;e.children=kr(e),e.children.forEach((e=>{if("element"===e.type&&"span"===e.tagName){const n=e.children[e.children.length-1];n&&(!br(n)||br(n)&&!Er(n))&&(e.children.push({type:"text",value:"\n"}),t++)}})),e.properties.style=[`--line-count: ${t};`,`--line-digits: ${t.toString().length};`]}}))}}function kr(e,t=[],n){const r=()=>t[t.length-1]??(t.push(xr())&&t[t.length-1]);return e.children.forEach((a=>{if(br(a)&&Er(a)){const o=a.value.split(/\n/);o.forEach(((a,i)=>{n?r().children.push({...e,children:[wr(a)]}):r().children.push(wr(a)),i!==o.length-1&&t.push(xr())}))}else yr(a)&&a.children.some(Er)?kr(a,t,!0):r().children.push(a)})),t}function xr(...e){return{type:"element",tagName:"span",properties:{class:["line"]},children:e}}function Er(e){return br(e)&&/\r?\n/.test(e.value)||yr(e)&&e.children.some(Er)}var Sr=n(9464);var _r=n(7526),Cr=n(3806);const Tr={parse:e=>{const t=JSON.parse(e);if("object"!=typeof t)throw Error("Invalid JSON object");return t},parseSafe(e,t){try{return Tr.parse(e)}catch(e){return"function"==typeof t?t(e):t}},stringify:e=>JSON.stringify(e)},Ar=e=>{if("string"==typeof e)return(e=>"string"==typeof e&&!!Tr.parseSafe(e,!1))(e)?JSON.stringify(JSON.parse(e),null,2):e;if("object"==typeof e)try{return JSON.stringify(e,null,2)}catch{return function(e){const t=new Set;return JSON.stringify(e,((e,n)=>{if("object"==typeof n&&null!==n){if(t.has(n))return"[Circular]";t.add(n)}return n}),2)}(e)}return e.toString()};var Rr=n(3751);const Pr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Dr(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Pr,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M5 11.988h14M12.006 5v14"},null,-1)]))}const Mr={render:Dr},$r={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Nr(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",$r,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M8.143 12h7.714M12 8.143v7.714M5.571 3H18.43A2.57 2.57 0 0 1 21 5.571V18.43A2.57 2.57 0 0 1 18.429 21H5.57A2.57 2.57 0 0 1 3 18.429V5.57A2.57 2.57 0 0 1 5.571 3"},null,-1)]))}const Ir={render:Nr},Lr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Br(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Lr,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 8v4m0 4h.01M22 12c0 5.523-4.477 10-10 10S2 17.523 2 12 6.477 2 12 2s10 4.477 10 10"},null,-1)]))}const Fr={render:Br},Qr={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24"};function Ur(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Qr,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M22 12H2m9.3-9.3L2 12l9.3 9.3"},null,-1)]))}const zr={render:Ur},jr={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24"};function Wr(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",jr,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M2 12h20m-9.3 9.3L22 12l-9.3-9.3"},null,-1)]))}const Xr={render:Wr},qr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Vr(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",qr,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M8 3H7a2 2 0 0 0-2 2v5a2 2 0 0 1-2 2 2 2 0 0 1 2 2v5c0 1.1.9 2 2 2h1m8 0h1a2 2 0 0 0 2-2v-5c0-1.1.9-2 2-2a2 2 0 0 1-2-2V5a2 2 0 0 0-2-2h-1"},null,-1)]))}const Hr={render:Vr},Zr={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function Gr(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Zr,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M16.4 6.8v5.6H7.6m0-5.6v10.4"},null,-1),(0,$t.Lk)("circle",{cx:"7.6",cy:"4.4",r:"2.4"},null,-1),(0,$t.Lk)("circle",{cx:"7.6",cy:"19.6",r:"2.4"},null,-1),(0,$t.Lk)("circle",{cx:"16.4",cy:"4.4",r:"2.4"},null,-1)]))}const Kr={render:Gr},Yr={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function Jr(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Yr,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M12 22.125V4.712S9.545 1.985 1.268 1.875a.5.5 0 0 0-.366.15.52.52 0 0 0-.152.369v16.375a.515.515 0 0 0 .5.519C9.543 19.4 12 22.125 12 22.125"},null,-1),(0,$t.Lk)("path",{d:"M9 11.18a21.3 21.3 0 0 0-5.26-1.055M9 15.375a21.3 21.3 0 0 0-5.26-1.055M15 11.18a21.3 21.3 0 0 1 5.26-1.055M15 15.375a21.3 21.3 0 0 1 5.26-1.055"},null,-1),(0,$t.Lk)("path",{d:"M12 22.125V4.712s2.455-2.727 10.732-2.837a.5.5 0 0 1 .366.15.52.52 0 0 1 .152.369v16.375a.515.515 0 0 1-.5.519C14.457 19.4 12 22.125 12 22.125"},null,-1)]))}const ea={render:Jr},ta={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24"};function na(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ta,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M21.4 4.6 10.6 19.4l-8-6.4"},null,-1)]))}const ra={render:na},aa={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function oa(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",aa,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m18 10-6 6-6-6"},null,-1)]))}const ia={render:oa},sa={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function la(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",sa,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m14 18-6-6 6-6"},null,-1)]))}const ca={render:la},ua={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function da(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ua,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m9 18 6-6-6-6"},null,-1)]))}const pa={render:da},ha={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function fa(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ha,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m18 15-6-6-6 6"},null,-1)]))}const ma={render:fa},ga={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function va(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ga,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M4 16c-1.1 0-2-.9-2-2V4c0-1.1.9-2 2-2h10c1.1 0 2 .9 2 2m-6 4h10a2 2 0 0 1 2 2v10a2 2 0 0 1-2 2H10a2 2 0 0 1-2-2V10a2 2 0 0 1 2-2"},null,-1)]))}const ba={render:va},ya={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function wa(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ya,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M18 6 6 18M6 6l12 12"},null,-1)]))}const Oa={render:wa},ka={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function xa(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ka,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m9.273 9.273 5.454 5.454m0-5.454-5.454 5.454M5.57 3h12.86A2.57 2.57 0 0 1 21 5.571V18.43A2.57 2.57 0 0 1 18.429 21H5.57A2.57 2.57 0 0 1 3 18.429V5.57A2.57 2.57 0 0 1 5.571 3Z"},null,-1)]))}const Ea={render:xa},Sa={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function _a(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Sa,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M4 16c-1.1 0-2-.9-2-2V4c0-1.1.9-2 2-2h10c1.1 0 2 .9 2 2m-3.121 8.879 4.242 4.242m0-4.242L12.88 17.12M10 8h10a2 2 0 0 1 2 2v10a2 2 0 0 1-2 2H10a2 2 0 0 1-2-2V10a2 2 0 0 1 2-2"},null,-1)]))}const Ca={render:_a},Ta={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function Aa(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ta,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M18.375 19.5a4.875 4.875 0 1 0-2.764-8.885A7.5 7.5 0 1 0 8.25 19.5z"},null,-1)]))}const Ra={render:Aa},Pa={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24"};function Da(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Pa,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M21.75 9V6a1.5 1.5 0 0 0-1.5-1.5h-12V3a1.5 1.5 0 0 0-1.5-1.5h-4.5A1.5 1.5 0 0 0 .75 3v17.8a1.7 1.7 0 0 0 3.336.438l2.352-11.154A1.5 1.5 0 0 1 7.879 9za1.5 1.5 0 0 1 1.45 1.886l-2.2 10.5a1.5 1.5 0 0 1-1.45 1.114H2.447"},null,-1),(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m12 13.5-2.25 2.25L12 18m4.5-4.5 2.25 2.25L16.5 18"},null,-1)]))}const Ma={render:Da},$a={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function Na(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",$a,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M10.546 2.438a1.957 1.957 0 0 0 2.908 0L14.4 1.4a1.959 1.959 0 0 1 3.41 1.413l-.071 1.4a1.96 1.96 0 0 0 2.051 2.054l1.4-.071a1.959 1.959 0 0 1 1.41 3.41l-1.042.94a1.96 1.96 0 0 0 0 2.909l1.042.94a1.959 1.959 0 0 1-1.413 3.41l-1.4-.071a1.96 1.96 0 0 0-2.056 2.056l.071 1.4A1.959 1.959 0 0 1 14.4 22.6l-.941-1.041a1.96 1.96 0 0 0-2.908 0L9.606 22.6A1.959 1.959 0 0 1 6.2 21.192l.072-1.4a1.96 1.96 0 0 0-2.056-2.056l-1.4.071A1.958 1.958 0 0 1 1.4 14.4l1.041-.94a1.96 1.96 0 0 0 0-2.909L1.4 9.606A1.958 1.958 0 0 1 2.809 6.2l1.4.071a1.96 1.96 0 0 0 2.058-2.06L6.2 2.81A1.959 1.959 0 0 1 9.606 1.4z"},null,-1),(0,$t.Lk)("path",{d:"M7.5 12.001a4.5 4.5 0 1 0 9 0 4.5 4.5 0 1 0-9 0"},null,-1)]))}const Ia={render:Na},La={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ba(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",La,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M20.86 16.6c0 .32-.16.64-.47.72l-6.3 2.83a.93.93 0 0 1-.8-.08c-.23-.16-.38-.4-.38-.63V4.56c0-.24.15-.55.39-.63.24-.16.47-.16.79-.08l6.3 2.83c.31.16.47.4.47.71v9.22zM10.7 19.52a.77.77 0 0 1-.7-.08.72.72 0 0 1-.32-.63V5.35c0-.24.15-.56.4-.63.23-.16.46-.16.78-.08M7.43 18.73a.77.77 0 0 1-.7-.08.72.72 0 0 1-.32-.63V6.05c0-.23.16-.55.4-.63.23-.15.47-.15.78-.07m-3.43 12.6a.77.77 0 0 1-.7-.08.72.72 0 0 1-.32-.63V6.84c0-.23.16-.55.4-.63.23-.16.47-.16.78-.08"},null,-1)]))}const Fa={render:Ba},Qa={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ua(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Qa,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M15.739 8.061c-2.02.404-4.041-1.01-4.445-3.03-.202-1.01 0-2.223.606-3.031C6.445 2 2 6.445 2 11.9s4.445 9.9 9.9 9.9 9.9-4.445 9.9-9.9v-.808c-1.616 1.212-4.04.808-5.253-.808-.606-.606-.808-1.415-.808-2.223"},null,-1),(0,$t.Lk)("path",{fill:"currentColor",d:"M8.667 15.133a1.212 1.212 0 1 0 0-2.425 1.212 1.212 0 0 0 0 2.425m4.243 1.212a1.212 1.212 0 1 0 0-2.424 1.212 1.212 0 0 0 0 2.424M9.88 9.88a1.212 1.212 0 1 0 0-2.425 1.212 1.212 0 0 0 0 2.425"},null,-1)]))}const za={render:Ua},ja={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function Wa(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ja,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M17.673 16.462C10.59 16.46 5.52 9.612 7.59 2.837.22 6.233-.71 16.333 5.917 21.016c5.864 4.144 14.064 1.34 16.163-5.526-1.38.64-2.884.972-4.407.972M11.996 3.484h4.866m-2.433-2.433v4.867m4.867 2.433h3.244m-1.622-1.622v3.244"},null,-1)]))}const Xa={render:Wa},qa={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Va(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",qa,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M3 6h18m-2 0v14c0 1-1 2-2 2H7c-1 0-2-1-2-2V6m3 0V4c0-1 1-2 2-2h4c1 0 2 1 2 2v2m-6 5v6m4-6v6"},null,-1)]))}const Ha={render:Va},Za={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Ga(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Za,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M20.317 4.37a19.8 19.8 0 0 0-4.885-1.515.074.074 0 0 0-.079.037c-.21.375-.444.864-.608 1.25a18.3 18.3 0 0 0-5.487 0 13 13 0 0 0-.617-1.25.08.08 0 0 0-.079-.037A19.7 19.7 0 0 0 3.677 4.37a.1.1 0 0 0-.032.027C.533 9.046-.32 13.58.099 18.057a.08.08 0 0 0 .031.057 19.9 19.9 0 0 0 5.993 3.03.08.08 0 0 0 .084-.028 14 14 0 0 0 1.226-1.994.076.076 0 0 0-.041-.106 13 13 0 0 1-1.872-.892.077.077 0 0 1-.008-.128 10 10 0 0 0 .372-.292.07.07 0 0 1 .077-.01c3.928 1.793 8.18 1.793 12.062 0a.07.07 0 0 1 .078.01q.181.149.373.292a.077.077 0 0 1-.006.127 12.3 12.3 0 0 1-1.873.892.077.077 0 0 0-.041.107c.36.698.772 1.362 1.225 1.993a.08.08 0 0 0 .084.028 19.8 19.8 0 0 0 6.002-3.03.08.08 0 0 0 .032-.054c.5-5.177-.838-9.674-3.549-13.66a.06.06 0 0 0-.031-.03M8.02 15.33c-1.182 0-2.157-1.085-2.157-2.419 0-1.333.956-2.419 2.157-2.419 1.21 0 2.176 1.096 2.157 2.42 0 1.333-.956 2.418-2.157 2.418m7.975 0c-1.183 0-2.157-1.085-2.157-2.419 0-1.333.955-2.419 2.157-2.419 1.21 0 2.176 1.096 2.157 2.42 0 1.333-.946 2.418-2.157 2.418"},null,-1)]))}const Ka={render:Ga},Ya={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function Ja(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ya,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M10.145 5.108a.584.584 0 0 1-.657-.37l-.24-.705a.51.51 0 0 0-.496-.35c-1.835.065-3.434.577-4.13.834a1.2 1.2 0 0 0-.6.474c-.75 1.142-3.085 5.242-3.02 11.944.003.4.169.78.485 1.025.716.556 2.259 1.576 4.702 2.314.358.108.75.004.962-.304.316-.462.597-1.04.796-1.494a.56.56 0 0 1 .669-.31c.87.248 1.993.439 3.39.439 1.393 0 2.512-.19 3.379-.437a.56.56 0 0 1 .669.31c.199.454.48 1.03.795 1.492.211.308.604.412.962.304 2.443-.738 3.986-1.758 4.703-2.314.315-.245.48-.626.485-1.025.063-6.652-2.236-10.74-3.003-11.917a1.25 1.25 0 0 0-.654-.5c-.75-.256-2.428-.763-4.094-.833a.505.505 0 0 0-.496.348l-.239.703a.586.586 0 0 1-.66.37 10.5 10.5 0 0 0-1.846-.15c-.827 0-1.461.08-1.862.152"},null,-1),(0,$t.Lk)("path",{d:"M10 12.553c0 1.38-1.008 2.5-2.25 2.5-1.243 0-2.25-1.12-2.25-2.5 0-1.381 1.007-2.5 2.25-2.5 1.242 0 2.25 1.119 2.25 2.5m4 0c0 1.38 1.007 2.5 2.25 2.5 1.242 0 2.25-1.12 2.25-2.5 0-1.381-1.008-2.5-2.25-2.5-1.243 0-2.25 1.119-2.25 2.5"},null,-1)]))}const eo={render:Ja},to={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24"};function no(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",to,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M5.4 22.2h13c1.1 0 2-.9 2-2V3.6c0-1.1-.9-2-2-2h-9l-6 6v12.6c0 1.1.9 2 2 2"},null,-1),(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M3.6 7.6h4c1.1 0 2-.9 2-2v-4zm9.4 0h4m-10 5h10M7 17.4h10"},null,-1)]))}const ro={render:no},ao={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function oo(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ao,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M12.001 3.75v12m-4.5-4.5 4.5 4.5 4.5-4.5m6.75 4.5v1.5a3 3 0 0 1-3 3h-16.5a3 3 0 0 1-3-3v-1.5"},null,-1)]))}const io={render:oo},so={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function lo(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",so,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M7.42 16.58H3.25a.83.83 0 0 1-.83-.83V3.25a.83.83 0 0 1 .83-.83h12.5a.83.83 0 0 1 .83.83v4.17"},null,-1),(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M8.25 7.42h12.5s.83 0 .83.83v12.5s0 .83-.83.83H8.25s-.83 0-.83-.83V8.25s0-.83.83-.83"},null,-1)]))}const co={render:lo},uo={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function po(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",uo,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M14 3H5.571A2.57 2.57 0 0 0 3 5.571V18.43A2.57 2.57 0 0 0 5.571 21H18.43A2.57 2.57 0 0 0 21 18.429V10m.96-5.35a1.845 1.845 0 0 0-2.61-2.61l-6.644 6.647a1.3 1.3 0 0 0-.328.543l-.864 2.849a.327.327 0 0 0 .408.407l2.849-.864c.205-.062.391-.174.543-.325l6.645-6.647Z"},null,-1)]))}const ho={render:po},fo={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function mo(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",fo,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M6 10c-1.1 0-2 .9-2 2s.9 2 2 2 2-.9 2-2-.9-2-2-2m12 0c-1.1 0-2 .9-2 2s.9 2 2 2 2-.9 2-2-.9-2-2-2m-6 0c-1.1 0-2 .9-2 2s.9 2 2 2 2-.9 2-2-.9-2-2-2"},null,-1)]))}const go={render:mo},vo={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function bo(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",vo,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M2.25 4.5h19.5s1.5 0 1.5 1.5v12s0 1.5-1.5 1.5H2.25s-1.5 0-1.5-1.5V6s0-1.5 1.5-1.5m13.437 5.475L19.5 13.5M8.313 9.975 4.5 13.5"},null,-1),(0,$t.Lk)("path",{d:"m22.88 5.014-9.513 6.56a2.41 2.41 0 0 1-2.734 0L1.12 5.014"},null,-1)]))}const yo={render:bo},wo={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Oo(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",wo,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M12 0a12.24 12.24 0 0 0-8.53 3.65A11.85 11.85 0 0 0 0 12.21 11.78 11.78 0 0 0 11.8 24h.2a12.11 12.11 0 0 0 12-12.21A11.77 11.77 0 0 0 12 0m-1.5 16.54A1.48 1.48 0 0 1 12 15a1.53 1.53 0 0 1 1.52 1.47A1.47 1.47 0 0 1 12.05 18a1.53 1.53 0 0 1-1.55-1.46m.5-4v-6a1 1 0 0 1 2 0v6a1 1 0 0 1-2 0"},null,-1)]))}const ko={render:Oo},xo={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Eo(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",xo,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m18.59 15.3 2.9-2.91a1.32 1.32 0 0 0 0-1.87l-1.85-1.85a3.3 3.3 0 1 1-4.31-4.31L13.48 2.5a1.32 1.32 0 0 0-1.87 0l-2.9 2.91a3.3 3.3 0 1 0-3.3 3.3l-2.9 2.9a1.32 1.32 0 0 0 0 1.87l1.85 1.85a3.3 3.3 0 1 1 4.31 4.31l1.85 1.85a1.32 1.32 0 0 0 1.87 0l2.9-2.9a3.3 3.3 0 1 0 3.3-3.3Z"},null,-1)]))}const So={render:Eo},_o={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Co(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",_o,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M6.4 4H20v13.59M4 20 20 4"},null,-1)]))}const To={render:Co},Ao={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ro(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ao,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M4 6h16M7.556 12h8.888m-6.222 6h3.556"},null,-1)]))}const Po={render:Ro},Do={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Mo(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Do,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M20 20a2 2 0 0 0 2-2V8a2 2 0 0 0-2-2h-7.9a2 2 0 0 1-1.69-.9L9.6 3.9A2 2 0 0 0 7.93 3H4a2 2 0 0 0-2 2v13a2 2 0 0 0 2 2z"},null,-1)]))}const $o={render:Mo},No={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 1024 1024"};function Io(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",No,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M512 0a512 512 0 0 0-162 998c26 4 35-11 35-25l-1-95c-128 23-161-32-172-60-6-15-31-61-52-73-18-10-44-33-1-33 40-1 69 37 78 52 46 78 120 56 149 43 5-33 18-56 33-69-114-13-234-56-234-253 0-56 20-101 53-137-5-13-23-65 5-136 0 0 43-13 141 53a487 487 0 0 1 256 0c98-66 141-53 141-53 28 71 10 123 5 136 33 36 53 81 53 137 0 197-120 240-234 253 18 16 35 47 35 95l-1 140c0 14 9 30 35 25A512 512 0 0 0 512 0"},null,-1)]))}const Lo={render:Io},Bo={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function Fo(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Bo,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M16.643 22.94v-3.135a3.68 3.68 0 0 0-1.026-2.852c3.43-.382 7.036-1.682 7.036-7.648a5.94 5.94 0 0 0-1.64-4.097 5.54 5.54 0 0 0-.097-4.12s-1.29-.382-4.272 1.618a14.6 14.6 0 0 0-7.649 0c-2.982-2-4.272-1.617-4.272-1.617a5.54 5.54 0 0 0-.098 4.119 5.94 5.94 0 0 0-1.639 4.13c0 5.922 3.606 7.222 7.036 7.648a3.68 3.68 0 0 0-1.027 2.819v3.136m0-2.186c-3.277 1.064-6.009 0-7.648-3.278"},null,-1)]))}const Qo={render:Fo},Uo={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function zo(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Uo,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M22 12c0 5.523-4.477 10-10 10m10-10c0-5.523-4.477-10-10-10m10 10H2m10 10C6.477 22 2 17.523 2 12m10 10a14.5 14.5 0 0 1 0-20m0 20a14.5 14.5 0 0 0 0-20M2 12C2 6.477 6.477 2 12 2"},null,-1)]))}const jo={render:zo},Wo={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Xo(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Wo,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M23.93 10.67a.76.76 0 0 0-.75-.67H12.75a.76.76 0 0 0-.75.75v3.5a.76.76 0 0 0 .75.75h5.58A7 7 0 1 1 12 5a7 7 0 0 1 4.23 1.42.74.74 0 0 0 1-.07l2.47-2.49a.75.75 0 0 0 .22-.57.77.77 0 0 0-.27-.54A12 12 0 1 0 24 12a12 12 0 0 0-.07-1.33"},null,-1)]))}const qo={render:Xo},Vo={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function Ho(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Vo,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M9.2 9.2c0-.5.1-1 .4-1.5s.6-.8 1.1-1q.75-.3 1.5-.3c.75 0 1 .2 1.4.5s.7.7 1 1.2c.2.5.3 1 .2 1.5s-.3 1-.6 1.4-.8.7-1.2.9c-.255.128-.5.3-.7.5s-.3.5-.3.8v.9m0 3.5c-.2 0-.4-.2-.4-.4s.2-.4.4-.4.4.2.4.4-.2.4-.4.4"},null,-1),(0,$t.Lk)("path",{d:"M12 22c5.524 0 10-4.476 10-10S17.524 2 12 2 2 6.476 2 12s4.476 10 10 10"},null,-1)]))}const Zo={render:Ho},Go={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ko(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Go,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M9.88 14.12a3 3 0 1 1 4.24-4.24m-3.39 9.04q.632.079 1.27.08c7 0 10-7 10-7a13.2 13.2 0 0 0-1.67-2.68M6.61 17.39A13.5 13.5 0 0 1 2 12s3-7 10-7a9.74 9.74 0 0 1 5.39 1.61M2 22 22 2"},null,-1)]))}const Yo={render:Ko},Jo={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function ei(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Jo,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M12.9 6.8v6.1h4.7"},null,-1),(0,$t.Lk)("path",{d:"M6.1 17.2c1.6 2.2 4.2 3.6 7.1 3.6 4.8 0 8.8-3.9 8.8-8.8s-3.9-8.8-8.8-8.8-8.8 4-8.8 8.8v1.7M2 11.3l2.4 2.4 2.4-2.4"},null,-1)]))}const ti={render:ei},ni={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function ri(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ni,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M22.1 9.2 12 1.5 1.8 9.2c-.2.2-.3.5-.4.7v11.7c0 .5.4.9.9.9h6.1v-4.9c0-1.9 1.6-3.5 3.5-3.5s3.5 1.6 3.5 3.5v4.9h6.1c.5 0 .9-.4.9-.9V9.9c0-.3-.1-.6-.4-.7z"},null,-1)]))}const ai={render:ri},oi={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function ii(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",oi,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M12 10.67v7.98M14.66 16 12 18.64 9.34 16"},null,-1),(0,$t.Lk)("path",{d:"M3.35 7.34h17.3s1.33 0 1.33 1.33v11.98s0 1.33-1.33 1.33H3.35s-1.33 0-1.33-1.33V8.67s0-1.33 1.33-1.33"},null,-1),(0,$t.Lk)("path",{d:"m2.2 8 2.13-5.19a1.33 1.33 0 0 1 1.22-.79h12.9a1.33 1.33 0 0 1 1.22.8l2.08 5.1M12 7.34V2.02"},null,-1)]))}const si={render:ii},li={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ci(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",li,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 22c5.523 0 10-4.477 10-10S17.523 2 12 2 2 6.477 2 12s4.477 10 10 10"},null,-1),(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 17v-6 0a.5.5 0 0 0-.5-.5l-.5.001h-1M12 17h-2m2 0h2"},null,-1),(0,$t.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M10.75 7.5a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const ui={render:ci},di={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function pi(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",di,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M12.82 14.51 15.36 12l1.76.47a1.56 1.56 0 0 0 1.51-.47 1.58 1.58 0 0 0 .37-1.49l-.47-1.75.82-.76 1.76.47A1.56 1.56 0 0 0 22.62 8 1.58 1.58 0 0 0 23 6.52l-.47-1.75a2.35 2.35 0 0 0-3.33-3.33l-9.71 9.74a6.29 6.29 0 1 0 3.33 3.33"},null,-1),(0,$t.Lk)("path",{d:"M4.26 17.84a1.9 1.9 0 1 0 3.8 0 1.9 1.9 0 1 0-3.8 0"},null,-1)]))}const hi={render:pi},fi={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function mi(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",fi,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M16.35 16.5v1c0 1.1-.9 2-2 2H2.75c-1.1 0-2-.9-2-2v-11c0-1.1.9-2 2-2h11.6c1.1 0 2 .9 2 2v1M5.25 12h18m0 0-3 3.1m3-3.1-3-2.9"},null,-1)]))}const gi={render:mi},vi={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function bi(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",vi,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M6.55 12a5.46 5.46 0 1 0 10.91.481A5.46 5.46 0 0 0 6.55 12m0 0H2m4.55 0H4.73m12.31 2.1a5.46 5.46 0 0 1-6.614-6.887M12 2v4.55m7.08-1.62-3.22 3.22M22 12h-4.54m1.62 7.08L16 16.01M12 22v-4m-7.07 1.08L8 16.01M4.93 4.93l3.22 3.22"},null,-1)]))}const yi={render:bi},wi={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24"};function Oi(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",wi,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M7.501 12a4.5 4.5 0 1 0 9 0 4.5 4.5 0 1 0-9 0m4.5-10.5v3m0 15v3m10.5-10.5h-3m-15 0h-3m17.924-7.424-2.121 2.121M6.697 17.304l-2.121 2.121m14.849 0-2.121-2.121M6.697 6.697 4.576 4.576"},null,-1)]))}const ki={render:Oi},xi={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function Ei(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",xi,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M13.5 18 12 19.5c-1 1-2.3 1.5-3.7 1.5s-2.7-.6-3.7-1.5c-1-1-1.5-2.3-1.5-3.7s.6-2.7 1.5-3.7l1.5-1.5M10.5 6 12 4.5c1-1 2.3-1.5 3.7-1.5s2.7.6 3.7 1.5c1 1 1.5 2.3 1.5 3.7s-.6 2.7-1.5 3.7l-1.5 1.5M9 15l6-6"},null,-1)]))}const Si={render:Ei},_i={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function Ci(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",_i,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M12 17a1 1 0 1 0 0-2 1 1 0 0 0 0 2"},null,-1),(0,$t.Lk)("path",{d:"M5.778 10h12.444c.982 0 1.778.895 1.778 2v8c0 1.105-.796 2-1.778 2H5.778C4.796 22 4 21.105 4 20v-8c0-1.105.796-2 1.778-2M7 10V7a5 5 0 1 1 10 0v3"},null,-1)]))}const Ti={render:Ci},Ai={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 593 593"};function Ri(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ai,t[0]||(t[0]=[(0,$t.Lk)("path",{"fill-rule":"evenodd",d:"M347 0c6 0 12 5 12 12v134l94-95c5-5 13-5 17 0l72 72c4 4 5 12 0 16v1l-95 94h134c7 0 12 5 12 12v101c0 7-5 12-12 12H447l95 94c4 5 5 13 0 17l-72 72c-4 4-12 5-16 0h-1l-94-95v134c0 7-5 12-12 12H246c-7 0-12-5-12-12v-70c0-22 9-43 24-59l130-130c14-14 14-37 0-51L259 142a84 84 0 0 1-25-59V12c0-7 5-12 12-12zM138 52h1l219 219c14 14 14 37 0 51L139 542c-4 5-12 5-17 0l-71-70c-4-5-5-12 0-17l95-96H12c-7 0-12-5-12-12V246c0-7 5-12 12-12h134l-95-94c-4-5-4-12 0-17l71-71c4-5 12-5 16 0"},null,-1)]))}const Pi={render:Ri},Di={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 662 662"};function Mi(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Di,t[0]||(t[0]=[(0,$t.Lk)("path",{"fill-rule":"evenodd",d:"M331 0a331 331 0 1 1 0 662 331 331 0 0 1 0-662m148.5 109 .6.9a187 187 0 1 1-297.6-.9 267 267 0 1 0 297 0"},null,-1)]))}const $i={render:Mi},Ni={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 620 620"};function Ii(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ni,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M-.653 309.29c-.078 82.272 31.848 159.735 89.739 218.178l.283.283c122.379 121.247 319.824 121.184 440.16-.283 119.967-121.1 119.409-316.933-.998-437.339l-.99-.99C468.681 30.845 390.716-1.017 307.893-.67c-82.83.354-160.443 32.994-218.814 91.79C31.209 149.555-.724 227.025-.653 309.29m469.271-160.789.778.778c87.9 87.9 88.325 231.005.707 319.33-77.753 78.46-199.786 87.412-288.959 26.862l198.513-198.513-.007 166.361 83.68-.07v-306.22H157.11l-.008 83.6h160.584L121.994 436.323c-59.347-87.689-50.536-208.59 26.51-286.343 87.836-88.686 231.436-89.31 320.106-1.486Z"},null,-1)]))}const Li={render:Ii},Bi={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 642 642"};function Fi(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Bi,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M321 0a321 321 0 1 1 0 642 321 321 0 0 1 0-642m51.3 63L321 189.5 269.7 63l1 136.5-95.8-97.2L228 228l-125.7-53.1 97.2 95.8-136.5-1L189.5 321 63 372.3l136.5-1-97.2 95.8 125.7-53-53.1 125.6 95.8-97.2-1 136.5L321 452.5 372.3 579l-1-136.5 95.8 97.2-53-125.7 125.6 53.1-97.2-95.8 136.5 1L452.5 321 579 269.7l-136.5 1 97.2-95.8L414 228l53.1-125.7-95.8 97.2z"},null,-1)]))}const Qi={render:Fi},Ui={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 620 620"};function zi(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ui,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M529.2 90.8A310 310 0 1 0 90.8 529.2 310 310 0 0 0 529.2 90.8M122.4 498.2c-54.9-55-15.4-183.5 88.2-287 103.6-103.7 232.1-143.2 287-88.3 54.9 54.8 15.4 183.4-88.3 287-103.5 103.6-232 143.1-286.9 88.3m261.8-262.7A104.7 104.7 0 1 1 236 383.7a104.7 104.7 0 0 1 148.2-148.2"},null,-1)]))}const ji={render:zi},Wi={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 14 14"};function Xi(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Wi,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M8.75 0a.75.75 0 0 1 .75.75v1.5a.75.75 0 0 1-1.5 0V.75A.75.75 0 0 1 8.75 0m4.03 2.28a.75.75 0 0 0-1.06-1.06l-1 1a.75.75 0 0 0 1.06 1.06zM9.707 5.707a1 1 0 0 0-1.414-1.414l-8 8a1 1 0 1 0 1.414 1.414zM11.75 4.5a.75.75 0 0 0 0 1.5h1.5a.75.75 0 0 0 0-1.5zm-1.53 2.22a.75.75 0 0 1 1.06 0l1 1a.75.75 0 0 1-1.06 1.06l-1-1a.75.75 0 0 1 0-1.06m-3.94-5a.75.75 0 0 0-1.06 1.06l1 1a.75.75 0 0 0 1.06-1.06z","clip-rule":"evenodd"},null,-1)]))}const qi={render:Xi},Vi={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24"};function Hi(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Vi,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M22 6.8H2m20 10.4H2"},null,-1)]))}const Zi={render:Hi},Gi={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function Ki(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Gi,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M17.333 5.34c1.39-.22 2.796-.334 4.205-.34.121 0 .238.043.329.122a.44.44 0 0 1 .133.314v13.73a.44.44 0 0 1-.133.306.46.46 0 0 1-.311.13C14.186 19.708 12 22 12 22m0 0V7.38S9.822 5.096 2.462 5a.5.5 0 0 0-.329.122.44.44 0 0 0-.133.314v13.73a.44.44 0 0 0 .133.306.46.46 0 0 0 .311.13C9.814 19.708 12 22 12 22"},null,-1),(0,$t.Lk)("path",{d:"M12 22a5.67 5.67 0 0 1 1.367-3.135c.967-1.141 2.237-1.923 3.633-2.236V2.443a.45.45 0 0 0-.051-.21.4.4 0 0 0-.14-.154.47.47 0 0 0-.409-.054 6.44 6.44 0 0 0-3.033 2.063A5.86 5.86 0 0 0 12 7.205"},null,-1)]))}const Yi={render:Ki},Ji={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 14 14"};function es(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ji,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M11.821.098a1.62 1.62 0 0 1 2.077 2.076l-3.574 10.712a1.62 1.62 0 0 1-1.168 1.069 1.6 1.6 0 0 1-1.52-.434l-1.918-1.909-2.014 1.042a.5.5 0 0 1-.73-.457l.083-3.184 7.045-5.117a.625.625 0 1 0-.735-1.012L2.203 8.088l-1.73-1.73a1.6 1.6 0 0 1-.437-1.447 1.62 1.62 0 0 1 1.069-1.238h.003z","clip-rule":"evenodd"},null,-1)]))}const ts={render:es},ns={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function rs(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ns,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M17 3a2.85 2.85 0 0 1 2.064-1.007 2.87 2.87 0 0 1 2.14.834 2.83 2.83 0 0 1 .825 2.131A2.8 2.8 0 0 1 21 7L7.5 20.5 2 22l1.5-5.5z"},null,-1)]))}const as={render:rs},os={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function is(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",os,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M6 6.663c0-1.582 1.75-2.538 3.082-1.682l8.301 5.337a2 2 0 0 1 0 3.364L9.082 19.02C7.75 19.875 6 18.919 6 17.337z"},null,-1)]))}const ss={render:is},ls={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 12 12"};function cs(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ls,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M11 4.3c-.5-.7-1.7-2.2-2.4-2.6C7.9 1.2 7 1 5.9 1s-2.6.5-3.5 1.5C1.5 3.4.9 4.7.9 6s.5 2.6 1.5 3.5c.9.9 2.2 1.5 3.6 1.5s2.6-.6 3.6-1.5c.5-.5.8-.9 1-1.6"},null,-1),(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M11 1v3.3H7.6"},null,-1)]))}const us={render:cs},ds={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function ps(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ds,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M16.22 2.663a1.47 1.47 0 0 0-1.09 0L8.87 5.167c-.35.14-.74.14-1.09 0L1.982 2.848a.735.735 0 0 0-1.008.686V17.62c0 .601.366 1.142.924 1.365l5.88 2.352c.35.14.742.14 1.092 0l6.258-2.504c.35-.14.741-.14 1.092 0l5.796 2.315a.735.735 0 0 0 1.008-.686V6.38a1.47 1.47 0 0 0-.924-1.365ZM8.325 5.272v16.17m7.35-18.884v16.17"},null,-1)]))}const hs={render:ps},fs={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ms(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",fs,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M3 15c1.895 3 3.79 4 6.632 4s6.631-3 6.631-7-2.842-7-5.684-7-4.737 1.5-4.737 4 1.895 5 5.684 5S19.492 11.547 21 9"},null,-1)]))}const gs={render:ms},vs={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function bs(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",vs,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M11 19a8 8 0 1 0 0-16 8 8 0 0 0 0 16m10 2-4.35-4.35"},null,-1)]))}const ys={render:bs},ws={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Os(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ws,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M20 12a2 2 0 0 0 2-2V6a2 2 0 0 0-2-2H4a2 2 0 0 0-2 2v4a2 2 0 0 0 2 2m16 0H4m16 0a2 2 0 0 1 2 2v4a2 2 0 0 1-2 2H4a2 2 0 0 1-2-2v-4a2 2 0 0 1 2-2"},null,-1),(0,$t.Lk)("path",{fill:"currentColor",d:"M18 9a1 1 0 1 0 0-2 1 1 0 0 0 0 2m0 8a1 1 0 1 0 0-2 1 1 0 0 0 0 2"},null,-1),(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M6 8h4m-4 8h4"},null,-1)]))}const ks={render:Os},xs={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Es(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",xs,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M20 7H9m0 0a2 2 0 1 1-4 0 2 2 0 0 1 4 0m7 10H5m11 0a2 2 0 1 0 4 0 2 2 0 0 0-4 0"},null,-1)]))}const Ss={render:Es},_s={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Cs(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",_s,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M2 12s3-7 10-7 10 7 10 7-3 7-10 7-10-7-10-7"},null,-1),(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 15a3 3 0 1 0 0-6 3 3 0 0 0 0 6"},null,-1)]))}const Ts={render:Cs},As={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24"};function Rs(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",As,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M2 5.4h20M2 12h20M2 18.6h13.3"},null,-1)]))}const Ps={render:Rs},Ds={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24"};function Ms(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ds,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M13 15h5M6 9l4 3-4 3"},null,-1),(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M21 2.8H3.1c-.7 0-1.4.6-1.4 1.4v15.6c0 .7.6 1.4 1.4 1.4H21c.7 0 1.4-.6 1.4-1.4V4.2c0-.8-.5-1.4-1.3-1.4z"},null,-1)]))}const $s={render:Ms},Ns={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Is(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ns,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M14.333 22.94c-.848 0-1.604-.5-1.96-1.251-1.783-3.772-3.568-6.116-6.2-7.75H5V2.69h1.779l6.5-1.5h7.82a1.375 1.375 0 0 1 0 2.75h-.5a.625.625 0 1 0 0 1.25h1.5a1.375 1.375 0 0 1 0 2.75h-.5a.625.625 0 1 0 0 1.25h1a1.375 1.375 0 0 1 0 2.75h-1a.625.625 0 0 0 0 1.25h1a1.375 1.375 0 0 1 0 2.75h-7.834c.618 1.618 1.2 3.36 1.497 4.743.276 1.28-.77 2.258-1.93 2.258ZM3.5 14.5a1 1 0 0 1-1 1H1a1 1 0 0 1-1-1V2a1 1 0 0 1 1-1h1.5a1 1 0 0 1 1 1z","clip-rule":"evenodd"},null,-1)]))}const Ls={render:Is},Bs={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Fs(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Bs,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M14.333.375c-.848 0-1.604.5-1.96 1.251-1.783 3.772-3.568 6.116-6.2 7.749H5v11.25h1.779l6.5 1.5h7.82a1.375 1.375 0 0 0 0-2.75h-.5a.625.625 0 1 1 0-1.25h1.5a1.375 1.375 0 0 0 0-2.75h-.5a.625.625 0 1 1 0-1.25h1a1.375 1.375 0 0 0 0-2.75h-1a.625.625 0 1 1 0-1.25h1a1.375 1.375 0 0 0 0-2.75h-7.834c.618-1.617 1.2-3.359 1.497-4.742.276-1.28-.77-2.258-1.93-2.258ZM3.5 8.815a1 1 0 0 0-1-1H1a1 1 0 0 0-1 1v12.5a1 1 0 0 0 1 1h1.5a1 1 0 0 0 1-1z","clip-rule":"evenodd"},null,-1)]))}const Qs={render:Fs},Us={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function zs(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Us,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m19.5 5.75-1.51 14.34a1.67 1.67 0 0 1-1.66 1.5H7.67A1.67 1.67 0 0 1 6 20.08L4.5 5.75m-2.08 0h19.16m-13.33 0v-2.5a.83.83 0 0 1 .83-.83h5.84a.83.83 0 0 1 .83.83v2.5M12 9.5v8.75m3.75-8.75-.42 8.75M8.25 9.5l.42 8.75"},null,-1)]))}const js={render:zs},Ws={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function Xs(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ws,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M12 17a1 1 0 1 0 0-2 1 1 0 0 0 0 2"},null,-1),(0,$t.Lk)("path",{d:"M5.778 10h12.444c.982 0 1.778.895 1.778 2v8c0 1.105-.796 2-1.778 2H5.778C4.796 22 4 21.105 4 20v-8c0-1.105.796-2 1.778-2M17 10V7a5 5 0 0 0-9.33-2.5"},null,-1)]))}const qs={render:Xs},Vs={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Hs(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Vs,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M4.9 19.1c-3.038-3.038-3.71-7.592-2.015-11.3M7.8 16.2c-1.397-1.397-1.945-3.346-1.646-5.2M16.2 7.8c2.3 2.3 2.3 6.1 0 8.5m2.9-11.4C23 8.8 23 15.1 19.1 19m-8.423-8.5a2 2 0 0 1 2.75 2.901M22 22 2 2"},null,-1)]))}const Zs={render:Hs},Gs={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Ks(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Gs,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M12 0a12 12 0 1 0 12 12A12 12 0 0 0 12 0m4.91 10.41A1 1 0 0 1 16 11h-2.25a.25.25 0 0 0-.25.25v7.25a1.5 1.5 0 0 1-3 0v-7.25a.25.25 0 0 0-.25-.25H8a1 1 0 0 1-.75-1.66l4-4.5a1 1 0 0 1 1.5 0l4 4.5a1 1 0 0 1 .16 1.07"},null,-1)]))}const Ys={render:Ks},Js={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function el(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Js,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 4.33v15.32m5.37-9.95L12 4.33 6.63 9.7"},null,-1)]))}const tl={render:el},nl={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function rl(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",nl,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M3.5 22.2c.3-4.5 4-8 8.5-8s8.2 3.5 8.5 8z"},null,-1),(0,$t.Lk)("circle",{cx:"12",cy:"6.2",r:"4.5"},null,-1)]))}const al={render:rl},ol={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function il(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ol,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M11.101 1.488H4.83a1 1 0 0 0-1 1v6.614m9.069 13.41h6.272a1 1 0 0 0 1-1v-6.614"},null,-1),(0,$t.Lk)("path",{d:"m1.23 6.5 2.6 2.6 2.601-2.6m16.339 11-2.6-2.6-2.601 2.6M5.505 18.546a2.551 2.551 0 1 0 0-5.103 2.551 2.551 0 0 0 0 5.103m4.277 3.965c-.559-1.83-2.29-3.202-4.276-3.202-1.988 0-3.718 1.372-4.277 3.202m17.264-15.92a2.551 2.551 0 1 0 0-5.103 2.551 2.551 0 0 0 0 5.103m4.277 3.965c-.559-1.83-2.289-3.202-4.276-3.202s-3.718 1.372-4.276 3.202"},null,-1)]))}const sl={render:il},ll={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function cl(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ll,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M3.375 7.875a4.125 4.125 0 1 0 8.25 0 4.125 4.125 0 1 0-8.25 0M.75 20.25a6.75 6.75 0 0 1 13.5 0m.001-15.834a4.125 4.125 0 1 1-.006 6.914M15 13.667a6.757 6.757 0 0 1 8.25 6.583"},null,-1)]))}const ul={render:cl},dl={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function pl(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",dl,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M2.97 2.771 13.908 5.88a3.77 3.77 0 0 1 2.592 3.5v10.51a2.052 2.052 0 0 1-2.6 2.167l-9.048-2.46a3.7 3.7 0 0 1-2.6-3.457V4.688a2.75 2.75 0 0 1 2.67-2.813H18.75a3 3 0 0 1 3 3v8.953a3.073 3.073 0 0 1-3.11 3.047H16.5"},null,-1),(0,$t.Lk)("path",{d:"M12.375 13.875a.375.375 0 0 1 0-.75m0 .75a.375.375 0 0 0 0-.75m9.375-3.75H16.5"},null,-1)]))}const hl={render:pl},fl={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ml(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",fl,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M4.9 19.1C1 15.2 1 8.8 4.9 4.9m2.9 11.3c-2.3-2.3-2.3-6.1 0-8.5m8.4.1c2.3 2.3 2.3 6.1 0 8.5m2.9-11.4C23 8.8 23 15.1 19.1 19M14 12a2 2 0 1 1-4 0 2 2 0 0 1 4 0"},null,-1)]))}const gl={render:ml},vl={xmlns:"http://www.w3.org/2000/svg",fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",viewBox:"0 0 24 24"};function bl(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",vl,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M20.25 3.75v19.5M3.75 3.75h16.5m-16.5 19.5V3.75m16.5 19.5H3.75m16.5-19.5H3.75l1.5-3h13.5zM.75 23.25h22.5m-9-2.25a2.25 2.25 0 0 0-4.5 0v2.25h4.5zM7.5 7.13a.37.37 0 1 1 0-.75m0 .75a.37.37 0 1 0 0-.75m4.5.75a.38.38 0 0 1 0-.75m0 .75a.38.38 0 0 0 0-.75m4.5.75a.38.38 0 0 1 0-.75m0 .75a.38.38 0 0 0 0-.75m-9 5.25a.38.38 0 0 1 0-.76m0 .76a.38.38 0 0 0 0-.76m4.5.76a.38.38 0 0 1 0-.76m0 .76a.38.38 0 0 0 0-.76m4.5.76a.38.38 0 0 1 0-.76m0 .76a.38.38 0 0 0 0-.76m-9 5.26a.38.38 0 0 1 0-.75m0 .75a.38.38 0 0 0 0-.75m4.5.75a.38.38 0 0 1 0-.75m0 .75a.38.38 0 0 0 0-.75m4.5.75a.38.38 0 0 1 0-.75m0 .75a.38.38 0 0 0 0-.75"},null,-1)]))}const yl={render:bl},wl={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Ol(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",wl,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M9.931 12.645h4.138l-2.07-4.908m0-7.737L.68 3.982l1.726 14.771L12 24l9.596-5.242L23.32 3.984zm7.064 18.31h-2.638l-1.422-3.503H8.996l-1.422 3.504h-2.64L12 2.65z"},null,-1)]))}const kl={render:Ol},xl={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function El(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",xl,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M16.074 16.86c-.72.616-2.157 1.035-3.812 1.035-2.032 0-3.735-.632-4.187-1.483-.161.488-.198 1.046-.198 1.402 0 0-.106 1.75 1.111 2.968 0-.632.513-1.145 1.145-1.145 1.083 0 1.082.945 1.081 1.712v.069c0 1.164.711 2.161 1.723 2.582a2.35 2.35 0 0 1-.236-1.029c0-1.11.652-1.523 1.41-2.003.602-.383 1.272-.807 1.733-1.66a3.1 3.1 0 0 0 .378-1.494 3.1 3.1 0 0 0-.148-.954M15.551.6c.196.244.296.572.496 1.229l4.368 14.347a18.2 18.2 0 0 0-5.222-1.768L12.35 4.8a.37.37 0 0 0-.71.002l-2.81 9.603a18.2 18.2 0 0 0-5.245 1.771L7.974 1.827c.2-.656.3-.984.497-1.227a1.6 1.6 0 0 1 .654-.484C9.415 0 9.757 0 10.443 0h3.135c.686 0 1.03 0 1.32.117A1.6 1.6 0 0 1 15.55.6z"},null,-1)]))}const Sl={render:El},_l={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Cl(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",_l,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M6.335.83a.75.75 0 0 0-.67 0L1.667 2.827a.75.75 0 0 0-.417.672V18c0 .271.146.52.382.654l7.989 4.493a.75.75 0 0 0 .747.007l8-4.5A.75.75 0 0 0 18.75 18v-4.537l3.585-1.792A.75.75 0 0 0 22.75 11V6a.75.75 0 0 0-.417-.672l-3.998-1.999a.75.75 0 0 0-.67 0l-3.997 1.999a.7.7 0 0 0-.246.194.75.75 0 0 0-.172.478v4.561l-2.5 1.407V3.5a.75.75 0 0 0-.43-.678L6.336.829Zm7.686 11.019 2.38 1.19-6.422 3.612-2.38-1.19zm3.229-.063-2.5-1.25V7.215l2.5 1.25zm1.5 0 2.5-1.25V7.215l-2.5 1.25zm-9.5 1.025V4.714l-2.5 1.25v8.254zm-1.005 4.65.743.372c.17.091.262.153.262.183v3.202l-6.5-3.657V4.714l2.5 1.25V15.5a.75.75 0 0 0 .415.67s1.122.591 2.108 1.064zm2.505.478 6.5-3.657v3.28l-6.5 3.656v-3.28ZM6 4.66 3.677 3.5 6 2.339 8.323 3.5zM20.323 6 18 4.839 15.677 6 18 7.161z","clip-rule":"evenodd"},null,-1)]))}const Tl={render:Cl},Al={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Rl(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Al,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M11.573 0a5 5 0 0 0-.359.007L10.85.04c-3.407.307-6.6 2.146-8.622 4.972a11.9 11.9 0 0 0-2.119 5.243c-.096.659-.108.854-.108 1.747s.012 1.089.108 1.748c.652 4.506 3.86 8.292 8.209 9.695.779.25 1.6.422 2.534.525.363.04 1.935.04 2.299 0 1.611-.178 2.977-.577 4.323-1.264.207-.106.247-.134.219-.158-.02-.013-.9-1.193-1.955-2.62l-1.919-2.592-2.404-3.558a339 339 0 0 0-2.422-3.556c-.009-.002-.018 1.579-.023 3.51-.007 3.38-.01 3.515-.052 3.595a.43.43 0 0 1-.206.214c-.075.037-.14.044-.495.044H7.81l-.108-.068a.44.44 0 0 1-.157-.171l-.05-.106.006-4.703.007-4.705.072-.092a.7.7 0 0 1 .174-.143c.096-.047.134-.051.54-.051.478 0 .558.018.682.154a467 467 0 0 1 2.895 4.361l4.735 7.17 1.9 2.879.096-.063a12.3 12.3 0 0 0 2.466-2.163 11.94 11.94 0 0 0 2.824-6.134c.096-.66.108-.854.108-1.748 0-.893-.012-1.088-.108-1.747-.652-4.506-3.859-8.292-8.208-9.695a12.6 12.6 0 0 0-2.499-.523A33 33 0 0 0 11.573 0m4.068 7.217c.347 0 .408.005.486.047a.47.47 0 0 1 .237.277c.018.06.023 1.365.018 4.304l-.006 4.218-.744-1.14-.746-1.14v-3.066c0-1.982.01-3.097.023-3.15a.48.48 0 0 1 .233-.296c.096-.05.13-.054.5-.054z"},null,-1)]))}const Pl={render:Rl},Dl={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Ml(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Dl,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M14.23 12.004a2.236 2.236 0 0 1-2.235 2.236 2.236 2.236 0 0 1-2.236-2.236 2.236 2.236 0 0 1 2.235-2.236 2.236 2.236 0 0 1 2.236 2.236m2.648-10.69c-1.346 0-3.107.96-4.888 2.622-1.78-1.653-3.542-2.602-4.887-2.602-.41 0-.783.093-1.106.278-1.375.793-1.683 3.264-.973 6.365C1.98 8.917 0 10.42 0 12.004c0 1.59 1.99 3.097 5.043 4.03-.704 3.113-.39 5.588.988 6.38.32.187.69.275 1.102.275 1.345 0 3.107-.96 4.888-2.624 1.78 1.654 3.542 2.603 4.887 2.603.41 0 .783-.09 1.106-.275 1.374-.792 1.683-3.263.973-6.365C22.02 15.096 24 13.59 24 12.004c0-1.59-1.99-3.097-5.043-4.032.704-3.11.39-5.587-.988-6.38a2.17 2.17 0 0 0-1.092-.278zm-.005 1.09v.006c.225 0 .406.044.558.127.666.382.955 1.835.73 3.704-.054.46-.142.945-.25 1.44a23.5 23.5 0 0 0-3.107-.534A24 24 0 0 0 12.769 4.7c1.592-1.48 3.087-2.292 4.105-2.295zm-9.77.02c1.012 0 2.514.808 4.11 2.28-.686.72-1.37 1.537-2.02 2.442a23 23 0 0 0-3.113.538 15 15 0 0 1-.254-1.42c-.23-1.868.054-3.32.714-3.707.19-.09.4-.127.563-.132zm4.882 3.05q.684.704 1.36 1.564c-.44-.02-.89-.034-1.345-.034q-.691-.001-1.36.034c.44-.572.895-1.096 1.345-1.565zM12 8.1c.74 0 1.477.034 2.202.093q.61.874 1.183 1.86.557.961 1.018 1.946c-.308.655-.646 1.31-1.013 1.95-.38.66-.773 1.288-1.18 1.87a25.6 25.6 0 0 1-4.412.005 27 27 0 0 1-1.183-1.86q-.557-.961-1.018-1.946a25 25 0 0 1 1.013-1.954c.38-.66.773-1.286 1.18-1.868A25 25 0 0 1 12 8.098zm-3.635.254c-.24.377-.48.763-.704 1.16q-.336.585-.635 1.174c-.265-.656-.49-1.31-.676-1.947.64-.15 1.315-.283 2.015-.386zm7.26 0q1.044.153 2.006.387c-.18.632-.405 1.282-.66 1.933a26 26 0 0 0-1.345-2.32zm3.063.675q.727.226 1.375.498c1.732.74 2.852 1.708 2.852 2.476-.005.768-1.125 1.74-2.857 2.475-.42.18-.88.342-1.355.493a24 24 0 0 0-1.1-2.98c.45-1.017.81-2.01 1.085-2.964zm-13.395.004c.278.96.645 1.957 1.1 2.98a23 23 0 0 0-1.086 2.964c-.484-.15-.944-.318-1.37-.5-1.732-.737-2.852-1.706-2.852-2.474s1.12-1.742 2.852-2.476c.42-.18.88-.342 1.356-.494m11.678 4.28c.265.657.49 1.312.676 1.948-.64.157-1.316.29-2.016.39a26 26 0 0 0 1.341-2.338zm-9.945.02c.2.392.41.783.64 1.175q.345.586.705 1.143a22 22 0 0 1-2.006-.386c.18-.63.406-1.282.66-1.933zM17.92 16.32c.112.493.2.968.254 1.423.23 1.868-.054 3.32-.714 3.708-.147.09-.338.128-.563.128-1.012 0-2.514-.807-4.11-2.28.686-.72 1.37-1.536 2.02-2.44 1.107-.118 2.154-.3 3.113-.54zm-11.83.01c.96.234 2.006.415 3.107.532.66.905 1.345 1.727 2.035 2.446-1.595 1.483-3.092 2.295-4.11 2.295a1.2 1.2 0 0 1-.553-.132c-.666-.38-.955-1.834-.73-3.703.054-.46.142-.944.25-1.438zm4.56.64q.661.032 1.345.034.691.001 1.36-.034c-.44.572-.895 1.095-1.345 1.565q-.684-.706-1.36-1.565"},null,-1)]))}const $l={render:Ml},Nl={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Il(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Nl,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M23.94 3a.5.5 0 0 0-.44-.25h-2.31l-8.55 13.91a.75.75 0 0 1-1.28 0L2.81 2.77H.5A.5.5 0 0 0 .06 3a.54.54 0 0 0 0 .51l11.5 19a.51.51 0 0 0 .86 0l11.5-19a.54.54 0 0 0 .02-.51"},null,-1),(0,$t.Lk)("path",{d:"m12 14.84 7.43-12.07H16a.47.47 0 0 0-.43.25L12 9.26 8.43 3A.47.47 0 0 0 8 2.77H4.57Z"},null,-1)]))}const Ll={render:Il},Bl={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 128 128"};function Fl(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Bl,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"m117.5 33.5.3-.2c-.6-1.1-1.5-2.1-2.4-2.6L67.1 2.9c-.8-.5-1.9-.7-3.1-.7s-2.3.3-3.1.7l-48 27.9c-1.7 1-2.9 3.5-2.9 5.4v55.7c0 1.1.2 2.3.9 3.4l-.2.1c.5.8 1.2 1.5 1.9 1.9l48.2 27.9c.8.5 1.9.7 3.1.7s2.3-.3 3.1-.7l48-27.9c1.7-1 2.9-3.5 2.9-5.4V36.1c.1-.8 0-1.7-.4-2.6M64 88.5c9.1 0 17.1-5 21.3-12.4l12.9 7.6c-6.8 11.8-19.6 19.8-34.2 19.8-21.8 0-39.5-17.7-39.5-39.5S42.2 24.5 64 24.5c14.7 0 27.5 8.1 34.3 20l-13 7.5C81.1 44.5 73.1 39.5 64 39.5c-13.5 0-24.5 11-24.5 24.5s11 24.5 24.5 24.5"},null,-1)]))}const Ql={render:Fl},Ul={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 128 128"};function zl(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ul,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M60.952.266C27.377.266.065 28.77.065 63.806s27.318 63.54 60.887 63.54c33.576 0 60.893-28.505 60.893-63.54S94.528.266 60.952.266m0 6.117c30.39 0 55.025 25.712 55.025 57.423 0 2.482-.167 4.924-.46 7.321-1.495 6.214-4.194 10.358-7.412 13.19-4.941 4.336-11.524 5.738-17.62 5.733a41 41 0 0 1-4.973-.289C97.506 77.4 99.336 57.866 89.864 43.315 80.39 28.763 62.315 23.338 46.882 30.416a24 24 0 0 0-1.68-1.114c-2.41-1.413-7.439-3.482-13.837-3.505-5.361-.022-10.615 1.568-15.133 4.58C26.554 15.317 43.21 6.386 60.952 6.394zM51.341 84.6c.894-4.07 3.343-10.431 5.548-15.655.623-1.481 1.23-2.872 1.777-4.104 3.451 12.822 5.635 20.449 9.546 25.65a18 18 0 0 0 1.913 2.114 28 28 0 0 1-9.173 1.56 28.2 28.2 0 0 1-10.023-1.86q-.098-1.227-.103-2.459a24.5 24.5 0 0 1 .515-5.24zm-7.444 3.77c-7.555-5.703-12.031-14.84-12.04-24.57.008-9.88 4.625-19.138 12.375-24.813 1.637.984 3.121 2.143 4.34 3.477 2.395 2.56 5.077 8.214 6.94 13.082.51 1.312.965 2.567 1.36 3.709C50.16 73.502 45.685 80.608 43.898 88.37zm21.802-18.226c-1.501-4.037-2.374-7.079-2.379-7.084l-.005-.012c-2.747-10.984-5.592-21.025-11.497-28.052a28 28 0 0 1 9.134-1.55c16.06.022 29.074 13.602 29.095 30.36-.007 9.673-4.432 18.764-11.914 24.474a15 15 0 0 1-2.184-.78c-1.137-.514-2.812-2.267-4.329-4.698-2.313-3.635-4.437-8.622-5.927-12.653zm-4.747 51.08c-30.39 0-55.025-25.707-55.025-57.424 0-2.086.114-4.144.32-6.173 4.508-16 15.338-21.92 25.595-22.02 2.08-.007 4.117.282 6.068.774-13.182 12.081-15.74 32.487-5.979 47.7 9.762 15.213 28.817 20.518 44.544 12.403 2.78 1.023 5.96 1.633 9.812 2.137 1.447.186 3.007.277 4.665.277a61.6 61.6 0 0 0 14.997-2.087c-10.29 15.308-27.088 24.417-45.007 24.407z"},null,-1)]))}const jl={render:zl},Wl={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 128 128"};function Xl(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Wl,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"m117.5 33.5.3-.2c-.6-1.1-1.5-2.1-2.4-2.6L67.1 2.9c-.8-.5-1.9-.7-3.1-.7s-2.3.3-3.1.7l-48 27.9c-1.7 1-2.9 3.5-2.9 5.4v55.7c0 1.1.2 2.3.9 3.4l-.2.1c.5.8 1.2 1.5 1.9 1.9l48.2 27.9c.8.5 1.9.7 3.1.7s2.3-.3 3.1-.7l48-27.9c1.7-1 2.9-3.5 2.9-5.4V36.1c.1-.8 0-1.7-.4-2.6m-53.5 70c-21.8 0-39.5-17.7-39.5-39.5S42.2 24.5 64 24.5c14.7 0 27.5 8.1 34.3 20l-13 7.5C81.1 44.5 73.1 39.5 64 39.5c-13.5 0-24.5 11-24.5 24.5s11 24.5 24.5 24.5c9.1 0 17.1-5 21.3-12.4l12.9 7.6c-6.8 11.8-19.6 19.8-34.2 19.8M115 62h-3.2l-.9 4h4.1v5h-5l-1.2 6h-4.9l1.2-6h-3.8l-1.2 6h-4.8l1.2-6H94v-5h3.5l.9-4H94v-5h5.3l1.2-6h4.9l-1.2 6h3.8l1.2-6h4.8l-1.2 6h2.2zm-12.7 4h3.8l.9-4h-3.8z"},null,-1)]))}const ql={render:Xl},Vl={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Hl(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Vl,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M1.5 0h21l-1.91 21.563L11.977 24l-8.565-2.438zm17.09 4.413L5.41 4.41l.213 2.622 10.125.002-.255 2.716h-6.64l.24 2.573h6.182l-.366 3.523-2.91.804-2.956-.81-.188-2.11h-2.61l.29 3.855L12 19.288l5.373-1.53z"},null,-1)]))}const Zl={render:Hl},Gl={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 50 50"};function Kl(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Gl,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"m7.91 9.32.09.09V35.3l-5.09-5.09a6.22 6.22 0 0 1-1.27-6.92zM35.3 8H9.41l-.09-.09 13.97-6.27a6.22 6.22 0 0 1 6.92 1.27zm2.29 31H11.71l-.54-.54A4 4 0 0 1 10 35.63V11.41zm-23.88 2H39v6c0 .55-.45 1-1 1H21.1c-.27 0-.52-.11-.71-.29l-4.7-4.7.02-.02zM48 21.1V38c0 .55-.45 1-1 1h-6.58L11.41 10h24.22c1.07 0 2.07.42 2.83 1.17l4.53 4.54.02-.02 4.7 4.7c.18.19.29.44.29.71"},null,-1)]))}const Yl={render:Kl},Jl={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 128 128"};function ec(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Jl,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M108.2 64.8c-.1-.1-.2-.2-.4-.2l-.1-.1c-.1-.1-.2-.1-.2-.2l-.1-.1c-.1 0-.2-.1-.2-.1l-.2-.1c-.1 0-.2-.1-.2-.1l-.2-.1c-.1 0-.2-.1-.2-.1-.1 0-.1 0-.2-.1l-.3-.1c-.1 0-.1 0-.2-.1l-.3-.1h-.1l-.4-.1h-.2c-.1 0-.2 0-.3-.1h-2.3c-.6-13.3.6-26.8-2.8-39.6 12.9-4.6 2.8-22.3-8.4-14.4-7.4-6.4-17.6-7.8-28.3-7.8-10.5.7-20.4 2.9-27.4 8.4-2.8-1.4-5.5-1.8-7.9-1.1v.1c-.1 0-.3.1-.4.2-.1 0-.3.1-.4.2h-.1c-.1 0-.2.1-.4.2h-.1l-.3.2h-.1l-.3.2h-.1l-.3.2s-.1 0-.1.1l-.3.2s-.1 0-.1.1l-.3.2s-.1 0-.1.1l-.3.2-.1.1c-.1.1-.2.1-.2.2l-.1.1-.2.2-.1.1c-.1.1-.1.2-.2.2l-.1.1c-.1.1-.1.2-.2.2l-.1.1c-.1.1-.1.2-.2.2l-.1.1c-.1.1-.1.2-.2.2l-.1.1c-.1.1-.1.2-.2.2l-.1.1-.1.3s0 .1-.1.1l-.1.3s0 .1-.1.1l-.1.3s0 .1-.1.1l-.1.3s0 .1-.1.1c.4.3.4.4.4.4v.1l-.1.3v.1c0 .1 0 .2-.1.3v3.1c0 .1 0 .2.1.3v.1l.1.3v.1l.1.3s0 .1.1.1l.1.3s0 .1.1.1l.1.3s0 .1.1.1l.2.3s0 .1.1.1l.2.3s0 .1.1.1l.2.3.1.1.3.3.3.3h.1c1 .9 2 1.6 4 2.2v-.2C23 37.3 26.5 50 26.7 63c-.6 0-.7.4-1.7.5h-.5c-.1 0-.3 0-.5.1-.1 0-.3 0-.4.1l-.4.1h-.1l-.4.1h-.1l-.3.1h-.1l-.3.1s-.1 0-.1.1l-.3.1-.2.1c-.1 0-.2.1-.2.1l-.2.1-.2.1c-.1 0-.2.1-.2.1l-.2.1-.4.3c-.1.1-.2.2-.3.2l-.4.4-.1.1c-.1.2-.3.4-.4.5l-.2.3-.3.6-.1.3v.3c0 .5.2.9.9 1.2.2 3.7 3.9 2 5.6.8l.1-.1c.2-.2.5-.3.6-.3h.1l.2-.1c.1 0 .1 0 .2-.1.2-.1.4-.1.5-.2.1 0 .1-.1.1-.2l.1-.1c.1-.2.2-.6.2-1.2l.1-1.3v1.8c-.5 13.1-4 30.7 3.3 42.5 1.3 2.1 2.9 3.9 4.7 5.4h-.5c-.2.2-.5.4-.8.6l-.9.6-.3.2-.6.4-.9.7-1.1 1c-.2.2-.3.4-.4.5l-.4.6-.2.3c-.1.2-.2.4-.2.6l-.1.3q-.3 1.2.6 2.7l.4.4h.2c.1 0 .2 0 .4.1.2.4 1.2 2.5 3.9.9 2.8-1.5 4.7-4.6 8.1-5.1l-.5-.6c5.9 2.8 12.8 4 19 4.2 8.7.3 18.6-.9 26.5-5.2 2.2.7 3.9 3.9 5.8 5.4l.1.1.1.1.1.1.1.1s.1 0 .1.1c0 0 .1 0 .1.1 0 0 .1 0 .1.1h2.1s.1 0 .1-.1h.1s.1 0 .1-.1h.1s.1 0 .1-.1c0 0 .1 0 .1-.1l.1-.1s.1 0 .1-.1l.1-.1h.1l.2-.2.2-.1h.1l.1-.1h.1l.1-.1.1-.1.1-.1.1-.1.1-.1.1-.1.1-.1v-.1s0-.1.1-.1v-.1s0-.1.1-.1v-.1s0-.1.1-.1v-1.4s-.3 0-.3-.1l-.3-.1v-.1l.3-.1s.2 0 .2-.1l.1-.1v-2.1s0-.1-.1-.1v-.1s0-.1-.1-.1v-.1s0-.1-.1-.1c0 0 0-.1-.1-.1 0 0 0-.1-.1-.1 0 0 0-.1-.1-.1 0 0 0-.1-.1-.1 0 0 0-.1-.1-.1 0 0 0-.1-.1-.1 0 0 0-.1-.1-.1 0 0 0-.1-.1-.1 0 0 0-.1-.1-.1 0 0 0-.1-.1-.1 0 0 0-.1-.1-.1 0 0 0-.1-.1-.1 0 0 0-.1-.1-.1l-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1v-.1l-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1-.1c2-1.9 3.8-4.2 5.1-6.9 5.9-11.8 4.9-26.2 4.1-39.2h.1c.1 0 .2.1.2.1h.3s.1 0 .1.1h.1s.1 0 .1.1l.2.1c1.7 1.2 5.4 2.9 5.6-.8 1.6.6-.3-1.8-1.3-2.5M36 23C32.8 7 58.4 4 59.3 19.6c.8 13-20 16.3-23.3 3.4m36.1 15c-1.3 1.4-2.7 1.2-4.1.7 0 1.9.4 3.9.1 5.9-.5.9-1.5 1-2.3 1.4-1.2-.2-2.1-.9-2.6-2l-.2-.1c-3.9 5.2-6.3-1.1-5.2-5-1.2.1-2.2-.2-3-1.5-1.4-2.6.7-5.8 3.4-6.3.7 3 8.7 2.6 10.1-.2 3.1 1.5 6.5 4.3 3.8 7.1m-7-17.5c-.9-13.8 20.3-17.5 23.4-4 3.5 15-20.8 18.9-23.4 4M41.7 17c-1.9 0-3.5 1.7-3.5 3.8s1.6 3.8 3.5 3.8 3.5-1.7 3.5-3.8-1.5-3.8-3.5-3.8m1.6 5.7c-.5 0-.8-.4-.8-1 0-.5.4-1 .8-1 .5 0 .8.4.8 1 0 .5-.3 1-.8 1m27.8-6.6c-1.9 0-3.4 1.7-3.4 3.8s1.5 3.8 3.4 3.8 3.4-1.7 3.4-3.8-1.5-3.8-3.4-3.8m1.6 5.6c-.4 0-.8-.4-.8-1 0-.5.4-1 .8-1s.8.4.8 1-.4 1-.8 1"},null,-1)]))}const tc={render:ec},nc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function rc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",nc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M1.5 0h21l-1.91 21.563L11.977 24l-8.564-2.438zm7.031 9.75-.232-2.718 10.059.003.23-2.622L5.412 4.41l.698 8.01h9.126l-.326 3.426-2.91.804-2.955-.81-.188-2.11H6.248l.33 4.171L12 19.351l5.379-1.443.744-8.157z"},null,-1)]))}const ac={render:rc},oc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 128 128"};function ic(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",oc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M111.7 29.7 128 64l-16.3 34.3H16.3L0 64l16.3-34.3zm-5.5 5.7H20L6.3 64l13.6 28.6h86.3L119.8 64zm-76.3 19v8h9.2v-8.1h4.8v20.4h-4.7v-8.4H30v8.4h-4.8V54.4zm34.3 0v3.8h-6.5v16.5H53V58.2h-6.5v-3.8zm18.3 0v3.8H76v16.5h-4.7V58.2h-6.5v-3.8zm16.1.9c1.2.5 2.3 1.4 3.1 2.6.7 1.2 1.1 2.5 1 3.9.1 1.4-.3 2.7-1 3.9-.7 1.1-1.8 2-3.1 2.6-1.5.6-3.1.9-4.7.9h-4.1v5.6H85V54.4h8.9c1.6 0 3.2.3 4.7.9m-5.1 2.9h-3.8v7.1h3.8q2.1 0 3.3-.9c.8-.6 1.2-1.6 1.1-2.6 0-1.1-.3-2.1-1.1-2.6-.8-.7-1.9-1-3.3-1"},null,-1)]))}const sc={render:ic},lc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function cc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",lc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M11.26 15.27c3.37-.15 7.15-1.09 7-2.21a26 26 0 0 1-4.76.59c-3.62.16-6.58-.24-6.61-.9 0-.2.24-.4.69-.58-2.39-.64-3.67.53-3.69 1.22-.1 1.45 3.22 2.06 7.37 1.88m5.91 1.64a23 23 0 0 1-4.07.5c-3.09.14-5.61-.21-5.64-.77 0-.17.21-.34.59-.49-2-.55-3.13.45-3.15 1-.05 1.24 2.78 1.77 6.33 1.61 2.87-.1 6.25-1 5.94-1.85M3.61 20.8a.93.93 0 0 1 .29-.65c-2.09.1-2.54 1.13-2.51 1.76.07 1.39 4.67 2.31 10 2.05 4.44-.22 9.16-1 9.16-2.34a26.3 26.3 0 0 1-7.2 1.23c-5.35.26-9.67-.66-9.74-2.05"},null,-1),(0,$t.Lk)("path",{d:"M22.4 12.65c-.55-1.48-2.67-2.34-4.74-1.57a2.47 2.47 0 0 0-1 .81c1.94-.56 3.83 0 4.34 1.41.45 1.21-.3 2.64-1.72 3.54a1.37 1.37 0 0 0 .92-.13c2.25-1.05 2.8-2.58 2.2-4.06m-9.88-1.38s-3.1-2.19-1.92-4.35S14.37 2.43 13.38 0c0 0 .12 1-3.12 3.91s-2.14 6.33 2.26 7.36"},null,-1),(0,$t.Lk)("path",{d:"M15.17 4.83s-4.24 1.51-2.83 3.53c1 1.43 2.13 2 1.42 4 0 0 2.83-1.51.71-3.53s.7-4 .7-4M8.69 19.26c-1.6-.43-2.46.35-2.47.81 0 1 2.18 1.39 5 1.27 2.26-.11 4.73-.95 4.67-1.49a16.6 16.6 0 0 1-3.2.4c-2.43.11-4.41-.16-4.43-.61-.04-.13.13-.26.43-.38"},null,-1)]))}const uc={render:cc},dc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function pc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",dc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M0 0h24v24H0zm22.034 18.276c-.175-1.095-.888-2.015-3.003-2.873-.736-.345-1.554-.585-1.797-1.14-.091-.33-.105-.51-.046-.705.15-.646.915-.84 1.515-.66.39.12.75.42.976.9 1.034-.676 1.034-.676 1.755-1.125-.27-.42-.404-.601-.586-.78-.63-.705-1.469-1.065-2.834-1.034l-.705.089c-.676.165-1.32.525-1.71 1.005-1.14 1.291-.811 3.541.569 4.471 1.365 1.02 3.361 1.244 3.616 2.205.24 1.17-.87 1.545-1.966 1.41-.811-.18-1.26-.586-1.755-1.336l-1.83 1.051c.21.48.45.689.81 1.109 1.74 1.756 6.09 1.666 6.871-1.004.029-.09.24-.705.074-1.65zm-8.983-7.245h-2.248c0 1.938-.009 3.864-.009 5.805 0 1.232.063 2.363-.138 2.711-.33.689-1.18.601-1.566.48-.396-.196-.597-.466-.83-.855-.063-.105-.11-.196-.127-.196l-1.825 1.125c.305.63.75 1.172 1.324 1.517.855.51 2.004.675 3.207.405.783-.226 1.458-.691 1.811-1.411.51-.93.402-2.07.397-3.346.012-2.054 0-4.109 0-6.179z"},null,-1)]))}const hc={render:pc},fc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function mc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",fc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M4.56 18.71c-1.41-4-.65-8.94 1.77-11.48A5.47 5.47 0 0 1 12 5.71h.15a.8.8 0 0 1 .22.12c4 1.65 4.28 8.57 1.28 11.41a4.15 4.15 0 0 0 3.08-1.35c2.12-2.23 2.76-6.62 1.49-10.22A7.75 7.75 0 0 0 12.81.54H12a11.5 11.5 0 0 0-5.42 21.6 10 10 0 0 1-2.02-3.43"},null,-1),(0,$t.Lk)("path",{d:"M23.5 12a11.49 11.49 0 0 0-6.09-10.14 9.9 9.9 0 0 1 2 3.43c1.41 4 .65 9-1.77 11.5a5.42 5.42 0 0 1-4 1.75A6.5 6.5 0 0 1 12 18.3h-.15a.8.8 0 0 1-.22-.12c-4-1.65-4.29-8.57-1.29-11.41a4.14 4.14 0 0 0-3.11 1.32c-2.12 2.22-2.76 6.61-1.49 10.2a7.77 7.77 0 0 0 5.45 5.17H12A11.5 11.5 0 0 0 23.5 12"},null,-1)]))}const gc={render:mc},vc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 128 128"};function bc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",vc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M0 0h61.4L0 60.4zm0 128L128 0H64.6L0 63.7zm128 0L64.6 66.6 3.3 128z"},null,-1)]))}const yc={render:bc},wc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Oc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",wc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M21.75 5.52 13 .28a2 2 0 0 0-2 0L2.25 5.52a2 2 0 0 0-1 1.69v9.58a2 2 0 0 0 1 1.69l2.62 1.57a3.12 3.12 0 0 0 4.73-2.68V8.12a1 1 0 0 0-2 0v9.25a1.13 1.13 0 0 1-.57 1 1.14 1.14 0 0 1-1.13 0l-2.61-1.58V7.24L12 2l8.73 5.21v9.55L12 22l-1.7-1a1 1 0 1 0-1 1.71l1.72 1a2 2 0 0 0 2 0l8.74-5.24a2 2 0 0 0 1-1.69V7.21a2 2 0 0 0-1.01-1.69"},null,-1),(0,$t.Lk)("path",{d:"M15.51 14.76a9.3 9.3 0 0 1-2.24 0 .85.85 0 0 1-.75-.85 1 1 0 0 0-2 0A2.87 2.87 0 0 0 13 16.75a12 12 0 0 0 1.36.08 12 12 0 0 0 1.37-.08 2.87 2.87 0 0 0 2.5-2.84A2.68 2.68 0 0 0 16 11.26l-2.91-.49a.69.69 0 0 1-.58-.68.85.85 0 0 1 .75-.85 9.3 9.3 0 0 1 2.24 0 .85.85 0 0 1 .75.85 1 1 0 0 0 2 0 2.87 2.87 0 0 0-2.5-2.84 12.5 12.5 0 0 0-2.73 0 2.87 2.87 0 0 0-2.51 2.84 2.68 2.68 0 0 0 2.25 2.65l2.91.49a.68.68 0 0 1 .58.68.85.85 0 0 1-.74.85"},null,-1)]))}const kc={render:Oc},xc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 128 128"};function Ec(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",xc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M63.877 125.392c-32.671 0-60.37-27.594-60.627-60.469a59.94 59.94 0 0 1 17.506-42.759 60.94 60.94 0 0 1 43.279-18.36 60.08 60.08 0 0 1 42.647 17.71 60.15 60.15 0 0 1 18.157 42.522c.151 33.604-26.864 61.021-60.469 61.363h-.493zm.19-118.406a57.77 57.77 0 0 0-41.01 17.427 56.78 56.78 0 0 0-16.63 40.484c.236 31.159 26.495 57.286 57.43 57.286h.414c31.863-.29 57.504-26.266 57.385-58.128a56.97 56.97 0 0 0-17.217-40.273A56.7 56.7 0 0 0 64.068 6.986z"},null,-1),(0,$t.Lk)("path",{d:"M16.89 82.383V46.865h8.64v3.183h-4.583v29.218h4.584v3.183zm29.323-18.111c0 6.478-3.933 10.167-9.26 10.167s-8.877-4.156-8.877-9.831c0-5.939 3.722-10.121 9.167-10.121s8.97 4.36 8.97 9.785m-14.415.29c0 3.932 1.973 7.05 5.36 7.05s5.333-3.183 5.333-7.195c0-3.643-1.796-7.083-5.334-7.083s-5.392 3.328-5.392 7.307zm17.407-9.404c1.69-.29 3.407-.434 5.123-.428a9.17 9.17 0 0 1 5.537 1.223 4.06 4.06 0 0 1 2.006 3.61 4.48 4.48 0 0 1-3.183 4.183c2.269.46 3.9 2.46 3.9 4.775a5.02 5.02 0 0 1-1.861 3.978c-1.368 1.21-3.643 1.796-7.162 1.796a34 34 0 0 1-4.327-.257zm3.499 7.622h1.795c2.433 0 3.801-1.145 3.801-2.782s-1.368-2.644-3.61-2.644a9.8 9.8 0 0 0-2.006.145zm0 8.878c.618.065 1.243.092 1.86.078 2.263 0 4.262-.861 4.262-3.182s-1.94-3.183-4.373-3.183h-1.75zM69.54 54.901h3.517v12.554c0 5.334-2.577 7.116-6.365 7.116a9.3 9.3 0 0 1-2.973-.507l.428-2.834c.703.224 1.44.335 2.183.349 2.006 0 3.183-.921 3.183-4.262l.026-12.416zm13.527 10.456v2.434h-7.32v-2.434zm17.091 8.273a12.3 12.3 0 0 1-4.978.862c-6.129 0-9.851-3.834-9.851-9.707-.283-5.353 3.827-9.923 9.18-10.206.375-.02.757-.02 1.131.006a11.1 11.1 0 0 1 4.775.862l-.783 2.801a9.5 9.5 0 0 0-3.788-.75c-3.932 0-6.76 2.467-6.76 7.116 0 4.235 2.499 6.971 6.734 6.971a9.8 9.8 0 0 0 3.834-.717zM111.2 46.766v35.61h-8.641v-3.182h4.583V49.949h-4.583v-3.183h8.64z"},null,-1)]))}const Sc={render:Ec},_c={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 128 128"};function Cc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",_c,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M65.004 115.355c-.461-.894-1.004-2.796-1.356-3.601-.378-.711-1.46-2.692-1.984-3.332-1.164-1.332-1.437-1.438-1.809-3.23-.628-3.067-2.148-8.462-4.042-12.227-1.004-2-2.626-3.606-4.067-5.07-1.246-1.247-4.121-3.31-4.668-3.227-4.766.894-6.226 5.586-8.457 9.27-1.27 2.062-2.516 3.769-3.52 5.937-.898 1.98-.812 4.23-2.331 5.938a15.4 15.4 0 0 0-3.333 5.855c-.195.453-.546 4.957-1.003 6.016l7.02-.438c6.585.461 4.687 2.961 14.858 2.438l16.098-.54a25 25 0 0 0-1.433-3.792zM111.793 8.254H16.207C7.312 8.23.086 15.457.086 24.352v35.105c2.352-.812 5.578-5.75 6.668-6.934 1.789-2.062 2.16-4.77 3.059-6.378 2.062-3.793 2.433-6.477 7.101-6.477 2.164 0 3.063.516 4.5 2.516.996 1.332 2.79 3.957 3.602 5.668 1.004 1.98 2.523 4.582 3.254 5.125.515.351.972.722 1.433.894.707.27 1.356-.27 1.902-.629.622-.539.895-1.52 1.52-2.953.895-2.086 1.813-4.418 2.332-5.312.914-1.461 1.273-3.254 2.25-4.067 1.461-1.246 3.441-1.355 3.957-1.437 2.98-.625 4.336 1.437 5.777 2.707.973.894 2.243 2.605 3.246 4.851.708 1.793 1.606 3.52 2.067 4.5.351.98 1.266 2.606 1.789 4.582.543 1.711 1.809 3.067 2.352 3.961 0 0 .812 2.164 5.476 4.145a35 35 0 0 0 4.336 1.52c2.066.734 4.047.644 6.563.374 1.789 0 2.793-2.625 3.601-4.683.438-1.254.98-4.774 1.25-5.758.27-.996-.437-1.707.192-2.625.722-.977 1.164-1.082 1.519-2.332.914-2.793 5.957-2.875 8.832-2.875 2.414 0 2.063 2.332 6.125 1.52 2.336-.434 4.586.273 7.023.995 2.063.543 4.043 1.168 5.204 2.524.73.898 2.629 5.312.73 5.476.164.188.36.645.625.817-.46 1.707-2.25.46-3.332.27-1.355-.27-2.332 0-3.684.624-2.335.996-5.668.918-7.726 2.625-1.715 1.438-1.715 4.582-2.543 6.371 0 0-2.254 5.696-6.996 9.192-1.278.914-3.715 3.058-8.918 3.871-2.356.355-4.586.355-7.024.27-1.164-.079-2.332-.079-3.52-.079-.706 0-3.062-.109-2.96.164l-.27.645c.024.29.063.602.164.895.102.515.102.976.192 1.437 0 .98-.086 2.063 0 3.066.082 2.063.894 3.957 1.004 6.102.078 2.355 1.246 4.875 2.414 6.77.46.707 1.086.789 1.355 1.71.352.98 0 2.141.188 3.227.625 4.227 1.875 8.73 3.773 12.61v.078c2.332-.352 4.77-1.247 7.836-1.684 5.664-.832 13.5-.461 18.54-.914 12.796-1.168 19.706 5.226 31.148 2.601V24.336c-.063-8.895-7.293-16.102-16.207-16.102zM64.086 83.855q0-.28 0 0m-34.457 14.75c.894-1.98 1.433-4.125 2.144-6.101.73-1.899 1.813-4.61 3.684-5.582-.246-.274-3.957-.375-4.934-.461-1.082-.086-2.171-.273-3.25-.438a135 135 0 0 1-6.125-1.265c-1.168-.274-5.21-1.715-6.02-2.067-2.085-.894-3.421-3.52-4.96-3.246-.977.188-1.98.54-2.605 1.54-.543.812-.731 2.242-1.083 3.226-.437 1.086-1.168 2.164-1.707 3.25-1.277 1.875-3.332 3.582-4.23 5.484-.191.457-.27.895-.457 1.356v21.683c1.082.188 2.16.371 3.328.73 8.996 2.438 11.164 2.606 19.98 1.63l.813-.11c.625-1.437 1.188-6.207 1.629-7.644.352-1.164.812-2.063.996-3.14.164-1.09 0-2.173-.102-3.15-.171-2.628 1.895-3.519 2.899-5.69zm0 0"},null,-1)]))}const Tc={render:Cc},Ac={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Rc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ac,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M7.01 10.207h-.944l-.515 2.648h.838q.834 0 1.242-.314.408-.315.55-1.049.137-.705-.124-.995-.262-.29-1.047-.29M12 5.688C5.373 5.688 0 8.514 0 12s5.373 6.313 12 6.313S24 15.486 24 12s-5.373-6.312-12-6.312m-3.26 7.451c-.261.25-.575.438-.917.551q-.505.163-1.285.164H5.357l-.327 1.681H3.652l1.23-6.326h2.65q1.195 0 1.744.628.549.627.33 1.752a2.8 2.8 0 0 1-.305.847q-.215.383-.561.703m4.024.715.543-2.799q.094-.478-.068-.651-.16-.174-.687-.174H11.46l-.704 3.625H9.388l1.23-6.327h1.367l-.327 1.682h1.218q1.15 0 1.586.401c.436.401.378.7.263 1.299l-.572 2.944zm7.597-2.265a2.8 2.8 0 0 1-.305.847q-.214.383-.561.703a2.44 2.44 0 0 1-.917.551q-.504.163-1.286.164h-1.18l-.327 1.682h-1.378l1.23-6.326h2.649q1.195 0 1.744.628.55.626.331 1.751m-2.595-1.382h-.943l-.516 2.648h.838q.835 0 1.242-.314.407-.315.551-1.049.137-.705-.125-.995c-.262-.29-.524-.29-1.047-.29"},null,-1)]))}const Pc={render:Rc},Dc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 121 121"};function Mc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Dc,t[0]||(t[0]=[(0,$t.Lk)("path",{"fill-rule":"evenodd",d:"M73.7 58.9c-1.5-1.8-3.2-3.9-5-5.9C58.5 41.7 48.4 30.3 38 19.2c-4.7-5.1-6.2-10-.6-15.1C43-.9 48.5-.4 53.7 5.3 68.1 21.2 82.4 37.2 97 52.9c5.5 5.9 4.2 9.9-1.8 14.2-23 16.7-46 33.5-68.9 50.4-5.1 3.8-10.1 4.8-14.2-.9-4.6-6.2-.3-9.9 4.4-13.3 17-12.4 34-24.8 50.9-37.4 2.5-1.7 5.9-2.8 6.3-7m7.3 59.8c-6.2 0-12.4.2-18.6-.1-4.9-.2-7.9-3-7.8-8.2.1-4.9 2.5-8.6 7.5-8.7 13.5-.4 27.1-.3 40.7 0 4.3.1 7.5 2.9 7.5 7.4 0 5.1-2.7 9.1-8.2 9.5-7.2.4-14.2.1-21.1.1","clip-rule":"evenodd"},null,-1)]))}const $c={render:Mc},Nc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Ic(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Nc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"m14.25.18.9.2.73.26.59.3.45.32.34.34.25.34.16.33.1.3.04.26.02.2-.01.13V8.5l-.05.63-.13.55-.21.46-.26.38-.3.31-.33.25-.35.19-.35.14-.33.1-.3.07-.26.04-.21.02H8.77l-.69.05-.59.14-.5.22-.41.27-.33.32-.27.35-.2.36-.15.37-.1.35-.07.32-.04.27-.02.21v3.06H3.17l-.21-.03-.28-.07-.32-.12-.35-.18-.36-.26-.36-.36-.35-.46-.32-.59-.28-.73-.21-.88-.14-1.05-.05-1.23.06-1.22.16-1.04.24-.87.32-.71.36-.57.4-.44.42-.33.42-.24.4-.16.36-.1.32-.05.24-.01h.16l.06.01h8.16v-.83H6.18l-.01-2.75-.02-.37.05-.34.11-.31.17-.28.25-.26.31-.23.38-.2.44-.18.51-.15.58-.12.64-.1.71-.06.77-.04.84-.02 1.27.05zm-6.3 1.98-.23.33-.08.41.08.41.23.34.33.22.41.09.41-.09.33-.22.23-.34.08-.41-.08-.41-.23-.33-.33-.22-.41-.09-.41.09zm13.09 3.95.28.06.32.12.35.18.36.27.36.35.35.47.32.59.28.73.21.88.14 1.04.05 1.23-.06 1.23-.16 1.04-.24.86-.32.71-.36.57-.4.45-.42.33-.42.24-.4.16-.36.09-.32.05-.24.02-.16-.01h-8.22v.82h5.84l.01 2.76.02.36-.05.34-.11.31-.17.29-.25.25-.31.24-.38.2-.44.17-.51.15-.58.13-.64.09-.71.07-.77.04-.84.01-1.27-.04-1.07-.14-.9-.2-.73-.25-.59-.3-.45-.33-.34-.34-.25-.34-.16-.33-.1-.3-.04-.25-.02-.2.01-.13v-5.34l.05-.64.13-.54.21-.46.26-.38.3-.32.33-.24.35-.2.35-.14.33-.1.3-.06.26-.04.21-.02.13-.01h5.84l.69-.05.59-.14.5-.21.41-.28.33-.32.27-.35.2-.36.15-.36.1-.35.07-.32.04-.28.02-.21V6.07h2.09l.14.01zm-6.47 14.25-.23.33-.08.41.08.41.23.33.33.23.41.08.41-.08.33-.23.23-.33.08-.41-.08-.41-.23-.33-.33-.23-.41-.08-.41.08z"},null,-1)]))}const Lc={render:Ic},Bc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 128 128"};function Fc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Bc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M64 14.648c-35.346 0-64 19.19-64 42.863C0 78.275 22.046 95.589 51.316 99.53V86.699c-15.55-4.89-26.166-14.693-26.166-25.991 0-16.183 21.779-29.303 48.646-29.303 26.866 0 46.693 8.975 46.693 29.303 0 10.486-5.273 17.95-14.066 22.72 1.204.908 2.22 2.072 2.904 3.419l.388.655C121.025 79.772 128 69.189 128 57.51c0-23.672-28.654-42.863-64-42.863zm20.1 74.88c-2.612.257-5.322.41-8.114.462l.002 9.63a88 88 0 0 0 12.474-2.492l-.501-.941c-.68-1.268-1.347-2.543-2.033-3.807a41 41 0 0 0-1.828-2.851z"},null,-1),(0,$t.Lk)("path",{"fill-rule":"evenodd",d:"M97.469 81.036s3.874 1.169 6.124 2.307c.78.396 2.132 1.184 3.106 2.22a8.4 8.4 0 0 1 1.42 2.04l15.266 25.74-24.674.01-11.537-21.666s-2.363-4.06-3.817-5.237c-1.213-.982-1.73-1.331-2.929-1.331h-5.862l.004 28.219-21.834.009V41.263h43.845s19.97.36 19.97 19.359S97.47 81.035 97.47 81.035zm-9.497-24.137-13.218-.009-.006 12.257 13.224-.004s6.124-.019 6.124-6.235c0-6.34-6.124-6.01-6.124-6.01z"},null,-1)]))}const Qc={render:Fc},Uc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function zc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Uc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M20.156.083c3.033.525 3.893 2.598 3.829 4.77L24 4.822 22.635 22.71 4.89 23.926h.016C3.433 23.864.15 23.729 0 19.139l1.645-3 2.819 6.586.503 1.172 2.805-9.144-.03.007.016-.03 9.255 2.956-1.396-5.431-.99-3.9 8.82-.569-.615-.51L16.5 2.114 20.159.073zM5.13 5.073c3.561-3.533 8.157-5.621 9.922-3.84 1.762 1.777-.105 6.105-3.673 9.636-3.563 3.532-8.103 5.734-9.864 3.957-1.766-1.777.045-6.217 3.612-9.75z"},null,-1)]))}const jc={render:zc},Wc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function Xc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Wc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M4.589 24c4.537 0 13.81-1.516 14.821-3v-5.729c-.957 1.408-10.284 2.912-14.821 2.912zm0-7.635c4.537 0 13.81-1.516 14.821-3V7.636c-.957 1.408-10.284 2.912-14.821 2.912zm0-7.636c4.537 0 13.81-1.516 14.821-3V0C18.453 1.408 9.126 2.912 4.589 2.912z"},null,-1)]))}const qc={render:Xc},Vc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 17 17"};function Hc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Vc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"m2.4 11.4 5.1-4.6-5.1-4.7L.7 3.7l3.4 3v.1l-3.4 3zm14.2 3.3v-2.2h-8v2.2z"},null,-1)]))}const Zc={render:Hc},Gc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 128 128"};function Kc(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Gc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M117.3 87.3c.2-.6.4-1.3.5-1.9 7-28-10.1-61.2-39.2-78.6C91.4 24 97 44.9 92 63.2c-.4 1.6-1 3.2-1.6 4.7-.6-.4-1.5-.9-2.5-1.5 0 0-28.9-17.9-60.2-49.4-.9-.8 16.7 25 36.6 46.1-9.4-5.3-35.5-24.3-52-39.4 2 3.4 4.4 6.7 7.1 9.8C33.1 51 51.2 72.6 72.7 89.2c-15.2 9.3-36.6 10-57.9 0C9.5 86.7 4.6 83.7 0 80.3c9 14.4 22.9 26.9 39.9 34.2 20.2 8.7 40.3 8.1 55.2.1.1 0 .1-.1.2-.1.7-.4 1.3-.7 2-1.1 7.2-3.7 21.3-7.4 28.9 7.4 1.7 3.6 5.7-15.6-8.9-33.5"},null,-1)]))}const Yc={render:Kc},Jc={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function eu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Jc,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M1.125 0C.502 0 0 .502 0 1.125v21.75C0 23.498.502 24 1.125 24h21.75c.623 0 1.125-.502 1.125-1.125V1.125C24 .502 23.498 0 22.875 0zm17.363 9.75q.918 0 1.627.111a6.4 6.4 0 0 1 1.306.34v2.458a4 4 0 0 0-.643-.361 5 5 0 0 0-.717-.26 5.5 5.5 0 0 0-1.426-.2q-.45 0-.819.086a2.1 2.1 0 0 0-.623.242q-.254.156-.393.374a.9.9 0 0 0-.14.49q0 .294.156.529.156.234.443.444c.287.21.423.276.696.41q.41.203.926.416.705.296 1.266.628.561.333.963.753.402.418.614.957.213.538.214 1.253 0 .986-.373 1.656a3 3 0 0 1-1.012 1.085 4.4 4.4 0 0 1-1.487.596q-.85.18-1.79.18a10 10 0 0 1-1.84-.164 5.5 5.5 0 0 1-1.512-.493v-2.63a5.03 5.03 0 0 0 3.237 1.2q.5 0 .872-.09.373-.09.623-.25.249-.162.373-.38a1.02 1.02 0 0 0-.074-1.089 2.1 2.1 0 0 0-.537-.5 5.6 5.6 0 0 0-.807-.444 28 28 0 0 0-1.007-.436q-1.377-.575-2.053-1.405t-.676-2.005q0-.92.369-1.582.368-.662 1.004-1.089a4.5 4.5 0 0 1 1.47-.629 7.5 7.5 0 0 1 1.77-.201m-15.113.188h9.563v2.166H9.506v9.646H6.789v-9.646H3.375z"},null,-1)]))}const tu={render:eu},nu={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24"};function ru(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",nu,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"none",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M7.8 4.6 1.4 11c-.6.6-.6 1.6 0 2l6.2 6.2m8.6-14.6 6.2 6.2c.6.6.6 1.6 0 2l-6.2 6.6"},null,-1)]))}const au={render:ru},ou={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function iu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ou,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M23.55 10.91 13.09.45a1.55 1.55 0 0 0-2.18 0L8.46 2.9 11 5.48a2.06 2.06 0 0 1 1-.23 2.25 2.25 0 0 1 2.25 2.25 2.2 2.2 0 0 1-.22 1l2.19 2.2a2 2 0 1 1-1.06 1.06L13 9.52a1 1 0 0 1-.21.09v5.54a2 2 0 1 1-1.5 0V9.61A2.24 2.24 0 0 1 9.75 7.5a2.2 2.2 0 0 1 .22-1L7.4 4 .45 11a1.55 1.55 0 0 0 0 2.18l10.46 10.37a1.55 1.55 0 0 0 2.18 0l10.46-10.46a1.55 1.55 0 0 0 0-2.18"},null,-1)]))}const su={render:iu},lu={xmlns:"http://www.w3.org/2000/svg",fill:"currentColor",viewBox:"0 0 24 24"};function cu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",lu,t[0]||(t[0]=[(0,$t.Lk)("path",{d:"M12.001 4.8q-4.8 0-6 4.8 1.8-2.4 4.2-1.8c.913.228 1.565.89 2.288 1.624C13.666 10.618 15.027 12 18.001 12q4.8 0 6-4.8-1.8 2.4-4.2 1.8c-.913-.228-1.565-.89-2.288-1.624C16.337 6.182 14.976 4.8 12.001 4.8m-6 7.2q-4.8 0-6 4.8 1.8-2.4 4.2-1.8c.913.228 1.565.89 2.288 1.624 1.177 1.194 2.538 2.576 5.512 2.576q4.8 0 6-4.8-1.8 2.4-4.2 1.8c-.913-.228-1.565-.89-2.288-1.624C10.337 13.382 8.976 12 6.001 12"},null,-1)]))}const uu={render:cu},du={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function pu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",du,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M2 11.997C2 20.066 3.934 22 11.997 22 20.066 22 22 20.066 22 11.997 22 3.934 20.06 2 11.997 2S2 3.934 2 11.997m4.034 2.081 3.129-7.132a2.96 2.96 0 0 1 2.828-1.848c1.285 0 2.302.649 2.834 1.848l3.135 7.132c.14.343.263.796.263 1.182 0 1.77-1.243 3.012-3.012 3.012-.606 0-1.084-.153-1.567-.306a5.1 5.1 0 0 0-1.653-.325c-.643 0-1.157.16-1.665.325-.49.153-.968.306-1.55.306A2.89 2.89 0 0 1 5.76 15.26c0-.392.122-.839.27-1.182zm5.963-5.889L8.905 15.2c.919-.429 1.972-.637 3.092-.637 1.084 0 2.18.208 3.06.637z","clip-rule":"evenodd"},null,-1)]))}const hu={render:pu},fu={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function mu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",fu,t[0]||(t[0]=[(0,$t.Fv)('<path fill="#fff" fill-rule="evenodd" d="M11.9 7.2h8.4v3.4h-8.4z" clip-rule="evenodd"></path><path fill="#3ECC5F" fill-rule="evenodd" d="M4.3 18.3c-.74 0-1.384-.403-1.73-1a2 2 0 0 0 1.73 3h2v-2z" clip-rule="evenodd"></path><path fill="#3ECC5F" fill-rule="evenodd" d="M13.298 7.738 20.3 7.3v-1a2 2 0 0 0-2-2h-9l-.25-.433a.29.29 0 0 0-.5 0L8.3 4.3l-.25-.433a.29.29 0 0 0-.5 0L7.3 4.3l-.25-.433a.29.29 0 0 0-.5 0L6.3 4.3h-.006l-.415-.414a.29.29 0 0 0-.482.13l-.137.51-.52-.14a.29.29 0 0 0-.353.354l.139.52-.51.136a.29.29 0 0 0-.13.483l.414.415V6.3l-.433.25a.29.29 0 0 0 0 .5l.433.25-.433.25a.29.29 0 0 0 0 .5l.433.25-.433.25a.29.29 0 0 0 0 .5l.433.25-.433.25a.29.29 0 0 0 0 .5l.433.25-.433.25a.29.29 0 0 0 0 .5l.433.25-.433.25a.29.29 0 0 0 0 .5l.433.25-.433.25a.29.29 0 0 0 0 .5l.433.25-.433.25a.29.29 0 0 0 0 .5l.433.25-.433.25a.29.29 0 0 0 0 .5l.433.25-.433.25a.29.29 0 0 0 0 .5l.433.25-.433.25a.29.29 0 0 0 0 .5l.433.25-.433.25a.29.29 0 0 0 0 .5l.433.25a2 2 0 0 0 2 2h12a2 2 0 0 0 2-2v-8l-7.002-.438a1.065 1.065 0 0 1 0-2.124" clip-rule="evenodd"></path><path fill="#3ECC5F" fill-rule="evenodd" d="M16.3 20.3h3v-4h-3z" clip-rule="evenodd"></path><path fill="#44D860" fill-rule="evenodd" d="M21.3 17.8q-.033.002-.064.006l-.012-.045a.5.5 0 1 0-.494-.857l-.033-.034a.496.496 0 0 0-.299-.788.5.5 0 0 0-.56.293l-.044-.011q.005-.031.006-.064a.5.5 0 0 0-1 0q.002.033.006.064l-.044.011a.5.5 0 1 0-.859.495 2 2 0 1 0 3.333 1.924q.031.005.064.006a.5.5 0 1 0 0-1" clip-rule="evenodd"></path><path fill="#3ECC5F" fill-rule="evenodd" d="M17.3 14.3h3v-2h-3z" clip-rule="evenodd"></path><path fill="#44D860" fill-rule="evenodd" d="M21.3 13.55a.25.25 0 1 0 0-.5l-.032.003-.006-.022a.25.25 0 0 0-.096-.481.25.25 0 0 0-.15.052l-.018-.017a.25.25 0 0 0-.005-.308.25.25 0 0 0-.424.06 1 1 0 1 0 0 1.925.25.25 0 0 0 .48-.098.25.25 0 0 0-.05-.149l.016-.017a.25.25 0 0 0 .31-.005.25.25 0 0 0-.063-.424l.006-.022z" clip-rule="evenodd"></path><path fill="#000" fill-rule="evenodd" d="M8.3 7.55a.25.25 0 0 1-.25-.25.75.75 0 0 0-1.5 0 .25.25 0 0 1-.5 0 1.25 1.25 0 0 1 2.5 0 .25.25 0 0 1-.25.25" clip-rule="evenodd"></path><path fill="#FFFF50" fill-rule="evenodd" d="M12.3 20.3h6a2 2 0 0 0 2-2v-7h-6a2 2 0 0 0-2 2z" clip-rule="evenodd"></path><path fill="#000" fill-rule="evenodd" d="M18.802 14.4h-5.004a.1.1 0 0 1 0-.2h5.004a.1.1 0 1 1 0 .2m0 2h-5.004a.1.1 0 0 1 0-.2h5.004a.1.1 0 1 1 0 .2m0 2h-5.004a.1.1 0 0 1 0-.2h5.004a.1.1 0 1 1 0 .2m0-4.981h-5.004a.1.1 0 0 1 0-.2h5.004a.1.1 0 1 1 0 .2m0 1.981h-5.004a.1.1 0 0 1 0-.2h5.004a.1.1 0 1 1 0 .2m0 2h-5.004a.1.1 0 0 1 0-.2h5.004a.1.1 0 1 1 0 .2M20.3 8.161h-.003c-.31.01-.456.32-.585.592-.134.285-.238.47-.409.465-.189-.007-.297-.22-.411-.446-.132-.26-.282-.554-.596-.543-.305.01-.452.28-.581.517-.138.253-.232.406-.412.4-.193-.008-.295-.181-.414-.382-.132-.224-.285-.475-.593-.467-.3.01-.447.239-.577.44-.137.213-.233.343-.416.336-.198-.007-.3-.151-.417-.318-.133-.187-.282-.4-.59-.39-.293.01-.44.198-.57.363-.123.156-.219.28-.422.272a.1.1 0 0 0-.008.2c.302.01.455-.18.588-.349.117-.15.219-.278.419-.285.192-.008.283.111.42.304.13.184.276.392.572.402.31.01.46-.221.592-.426.119-.184.22-.342.415-.349.18-.006.277.135.415.368.13.22.276.47.579.48.312.01.463-.262.594-.503.114-.21.222-.406.412-.413.18-.004.273.16.411.433.13.255.276.544.583.556h.017c.306 0 .452-.308.58-.58.114-.24.222-.468.407-.477z" clip-rule="evenodd"></path><path fill="#3ECC5F" fill-rule="evenodd" d="M10.3 20.3h4v-4h-4z" clip-rule="evenodd"></path><path fill="#D8D8D8" d="m20.276 9.922-7.778 2.084a.2.2 0 0 0-.141.245l.76 2.837a.2.2 0 0 0 .245.141l7.777-2.084a.2.2 0 0 0 .142-.245l-.76-2.836a.2.2 0 0 0-.245-.142"></path><path fill="#4A4A4A" d="m14.845 13.56-.278.074a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.123-.07m.559-.15-.278.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.123-.07m.559-.15-.278.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.123-.07m.559-.15-.278.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .122.071l.279-.075a.1.1 0 0 0 .07-.122l-.074-.278a.1.1 0 0 0-.123-.071m.559-.15-.278.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.123-.071m.559-.15-.278.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.123-.07Zm.56-.15-.28.075a.1.1 0 0 0-.07.123l.075.278a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.123-.07Zm.558-.15-.278.075a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .122.071l.279-.075a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.123-.07Zm.559-.15-.278.075a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .122.071l.279-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.123-.071Z"></path><path fill="#4A4A4A" fill-rule="evenodd" d="m13.082 14.032 1.187-.318a.1.1 0 0 1 .122.07l.075.279a.1.1 0 0 1-.07.122l-1.187.318a.1.1 0 0 1-.123-.07l-.075-.279a.1.1 0 0 1 .071-.122m6.533-1.75 1.187-.319a.1.1 0 0 1 .122.07l.075.28a.1.1 0 0 1-.07.122l-1.188.318a.1.1 0 0 1-.122-.07l-.075-.28a.1.1 0 0 1 .071-.122Zm-6.835.621.453-.121a.1.1 0 0 1 .122.07l.075.279a.1.1 0 0 1-.07.122l-.454.121a.1.1 0 0 1-.122-.07l-.075-.279a.1.1 0 0 1 .07-.122Z" clip-rule="evenodd"></path><path fill="#4A4A4A" d="m13.792 12.632-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.071l.278-.075a.1.1 0 0 0 .07-.122l-.074-.278a.1.1 0 0 0-.122-.071m.558-.15-.278.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07l.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.071Zm.56-.15-.279.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07l.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.07Zm.559-.15-.279.075a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.07l.278-.074a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.122-.07Zm.559-.149-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.071l.278-.075a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.122-.07m.559-.15-.279.075a.1.1 0 0 0-.07.122l.074.278a.1.1 0 0 0 .123.071l.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.071m.559-.15-.279.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07l.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.07Zm.559-.15-.279.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07l.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.07Zm.559-.149-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.122-.07m.559-.15-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.071l.278-.075a.1.1 0 0 0 .07-.122l-.074-.278a.1.1 0 0 0-.122-.071m.559-.15-.279.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07l.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.071m.558-.15-.278.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.07Zm.56-.15-.279.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.122-.07Z"></path><path fill="#4A4A4A" fill-rule="evenodd" d="m20.483 10.771-.454.122a.1.1 0 0 1-.122-.071l-.075-.278a.1.1 0 0 1 .071-.123l.453-.121a.1.1 0 0 1 .123.07l.074.279a.1.1 0 0 1-.07.122" clip-rule="evenodd"></path><path fill="#4A4A4A" d="m19.47 11.042.279-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07Zm-.558.15.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.071l-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .122.071Zm-.56.15.279-.075a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .122.071m-.558.15.278-.075a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.07Zm-.559.15.278-.075a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07Zm-.559.149.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.071l-.279.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07m-.559.15.278-.075a.1.1 0 0 0 .07-.122l-.074-.278a.1.1 0 0 0-.122-.071l-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.071m-.559.149.278-.074a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.07m-.559.15.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07m-.559.15.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07m-.56.15.28-.074a.1.1 0 0 0 .07-.123l-.075-.278a.1.1 0 0 0-.122-.071l-.279.075a.1.1 0 0 0-.07.122l.074.278a.1.1 0 0 0 .123.071Zm-.558.15.278-.075a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.071m-.559.15.278-.075a.1.1 0 0 0 .071-.122l-.075-.279a.1.1 0 0 0-.122-.07l-.278.074a.1.1 0 0 0-.071.123l.074.278a.1.1 0 0 0 .123.07Z"></path><path fill="#4A4A4A" d="m19.47 11.042.279-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07Zm-.558.15.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.071l-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .122.071Zm-.56.15.279-.075a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .122.071m-.558.15.278-.075a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.07Zm-.559.15.278-.075a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07Zm-.559.149.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.071l-.279.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07m-.559.15.278-.075a.1.1 0 0 0 .07-.122l-.074-.278a.1.1 0 0 0-.122-.071l-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.071m-.559.149.278-.074a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.07m-.559.15.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07m-.559.15.278-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07m-.56.15.28-.074a.1.1 0 0 0 .07-.123l-.075-.278a.1.1 0 0 0-.122-.071l-.279.075a.1.1 0 0 0-.07.122l.074.278a.1.1 0 0 0 .123.071Zm-.558.15.278-.075a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.122-.07l-.279.074a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.071m-.559.15.278-.075a.1.1 0 0 0 .071-.122l-.075-.279a.1.1 0 0 0-.122-.07l-.278.074a.1.1 0 0 0-.071.123l.074.278a.1.1 0 0 0 .123.07Z"></path><path fill="#4A4A4A" fill-rule="evenodd" d="m12.93 13.462.645-.173a.1.1 0 0 1 .122.07l.075.28a.1.1 0 0 1-.07.122l-.646.173a.1.1 0 0 1-.123-.071l-.074-.278a.1.1 0 0 1 .07-.123Z" clip-rule="evenodd"></path><path fill="#4A4A4A" d="m14.151 13.135-.278.074a.1.1 0 0 0-.071.123l.075.278a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.123-.07m.559-.15-.278.074a.1.1 0 0 0-.071.123l.075.278a.1.1 0 0 0 .122.071l.279-.075a.1.1 0 0 0 .07-.122l-.074-.278a.1.1 0 0 0-.123-.071m.56-.15-.28.075a.1.1 0 0 0-.07.122l.075.279a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.123-.071Zm.558-.15-.278.075a.1.1 0 0 0-.071.122l.075.279a.1.1 0 0 0 .122.07l.278-.074a.1.1 0 0 0 .071-.123l-.074-.278a.1.1 0 0 0-.123-.07Zm.559-.149-.278.074a.1.1 0 0 0-.071.123l.075.278a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.123-.07m.559-.15-.278.074a.1.1 0 0 0-.071.123l.075.278a.1.1 0 0 0 .122.071l.279-.075a.1.1 0 0 0 .07-.122l-.074-.279a.1.1 0 0 0-.123-.07m.559-.15-.278.074a.1.1 0 0 0-.071.123l.075.278a.1.1 0 0 0 .122.071l.278-.074a.1.1 0 0 0 .071-.123l-.074-.278a.1.1 0 0 0-.123-.071m.559-.15-.279.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07l.278-.074a.1.1 0 0 0 .071-.123l-.075-.278a.1.1 0 0 0-.122-.07Zm.559-.15-.279.075a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.07l.278-.074a.1.1 0 0 0 .071-.122l-.075-.279a.1.1 0 0 0-.122-.07Zm.559-.149-.278.074a.1.1 0 0 0-.071.123l.074.278a.1.1 0 0 0 .123.07l.278-.074a.1.1 0 0 0 .071-.122l-.075-.279a.1.1 0 0 0-.122-.07m.558-.15-.277.074a.1.1 0 0 0-.071.123l.074.278a.1.1 0 0 0 .123.071l.278-.075a.1.1 0 0 0 .071-.122l-.075-.279a.1.1 0 0 0-.122-.07Z"></path><path fill="#4A4A4A" fill-rule="evenodd" d="m20.039 11.557.61-.164a.1.1 0 0 1 .123.071l.074.279a.1.1 0 0 1-.07.122l-.61.164a.1.1 0 0 1-.123-.071l-.075-.278a.1.1 0 0 1 .07-.123z" clip-rule="evenodd"></path><path fill="#4A4A4A" d="m13.517 14.54-.279.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07l.278-.074a.1.1 0 0 0 .071-.123l-.075-.278a.1.1 0 0 0-.122-.07Zm.559-.15-.279.075a.1.1 0 0 0-.07.122l.074.279a.1.1 0 0 0 .123.07l.278-.074a.1.1 0 0 0 .071-.123l-.075-.278a.1.1 0 0 0-.122-.07Zm.559-.15-.279.075a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.07l.278-.074a.1.1 0 0 0 .071-.122l-.075-.279a.1.1 0 0 0-.122-.07Zm.559-.15-.279.075a.1.1 0 0 0-.07.123l.074.278a.1.1 0 0 0 .123.071l.278-.075a.1.1 0 0 0 .07-.122l-.074-.278a.1.1 0 0 0-.122-.071Z"></path><path fill="#4A4A4A" fill-rule="evenodd" d="m15.474 14.016 2.69-.72a.1.1 0 0 1 .122.07l.074.278a.1.1 0 0 1-.07.123l-2.69.72a.1.1 0 0 1-.122-.07l-.074-.279a.1.1 0 0 1 .07-.122m2.97-.796.348-.093a.1.1 0 0 1 .123.07l.074.279a.1.1 0 0 1-.07.122l-.349.094a.1.1 0 0 1-.122-.071l-.075-.279a.1.1 0 0 1 .07-.122Z" clip-rule="evenodd"></path><path fill="#4A4A4A" d="m19.351 12.977-.278.075a.1.1 0 0 0-.071.122l.075.279a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.123-.071m.559-.15-.278.075a.1.1 0 0 0-.071.122l.075.279a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.123-.07Zm1.048-.281-.278.075a.1.1 0 0 0-.071.122l.075.279a.1.1 0 0 0 .122.07l.279-.074a.1.1 0 0 0 .07-.123l-.074-.278a.1.1 0 0 0-.123-.07Zm-.845.442-.028-.104a.1.1 0 0 1 .07-.123l.28-.074a.1.1 0 0 1 .122.07l.027.105zm.49-.057.028.105a.1.1 0 0 1-.07.122l-.279.075a.1.1 0 0 1-.122-.07l-.028-.105zm-.344-2.893-7.652 2.05a.108.108 0 1 0 .057.21l7.65-2.05a.108.108 0 1 0-.055-.21"></path><path fill="#D8D8D8" d="m13.075 12-.3.08a.072.072 0 1 0 .037.14l.3-.08a.072.072 0 1 0-.037-.14m.603-.162-.3.081a.072.072 0 1 0 .037.14l.3-.08a.072.072 0 1 0-.037-.14Zm.464-.124-.3.08a.072.072 0 1 0 .037.14l.3-.08a.072.072 0 1 0-.037-.14m.464-.124-.3.08a.072.072 0 0 0 .038.14l.3-.08a.072.072 0 0 0-.038-.14m.464-.125-.3.081a.072.072 0 0 0 .037.14l.3-.08a.072.072 0 0 0-.037-.14Zm.58-.155-.3.08a.072.072 0 0 0 .037.14l.3-.08a.072.072 0 1 0-.037-.14m.464-.124-.3.08a.072.072 0 0 0 .037.14l.3-.08a.072.072 0 1 0-.038-.14Zm.464-.125-.3.081a.072.072 0 1 0 .037.14l.3-.08a.072.072 0 1 0-.037-.14Zm.464-.124-.3.08a.072.072 0 0 0 .038.14l.3-.08a.072.072 0 1 0-.037-.14Zm.58-.155-.3.08a.072.072 0 0 0 .037.14l.3-.08a.072.072 0 0 0-.038-.14Zm.464-.125-.3.081a.072.072 0 0 0 .037.14l.3-.08a.072.072 0 0 0-.037-.14Zm.464-.124-.3.08a.072.072 0 0 0 .038.14l.3-.08a.072.072 0 1 0-.037-.14Zm.58-.155-.3.08a.072.072 0 1 0 .037.14l.3-.08a.072.072 0 1 0-.038-.14Zm.464-.125-.301.08a.072.072 0 1 0 .037.14l.301-.08a.072.072 0 1 0-.037-.14m.464-.124-.301.08a.072.072 0 0 0 .037.14l.302-.08a.072.072 0 0 0-.038-.14" opacity=".136"></path><path fill="#44D860" fill-rule="evenodd" d="M16.3 17.8q-.033.002-.064.006l-.012-.045a.5.5 0 1 0-.494-.857l-.033-.034a.496.496 0 0 0-.299-.788.5.5 0 0 0-.56.293l-.044-.011q.005-.031.006-.064a.5.5 0 0 0-1 0q.002.033.006.064l-.044.011a.5.5 0 1 0-.859.495 2 2 0 1 0 3.333 1.924q.031.005.064.006a.5.5 0 1 0 0-1" clip-rule="evenodd"></path><path fill="#3ECC5F" fill-rule="evenodd" d="M10.3 14.3h4v-2h-4z" clip-rule="evenodd"></path><path fill="#44D860" fill-rule="evenodd" d="M15.3 13.55a.25.25 0 1 0 0-.5l-.032.003-.006-.022a.25.25 0 0 0-.096-.481.25.25 0 0 0-.15.052l-.018-.017a.25.25 0 0 0-.005-.308.25.25 0 0 0-.424.06 1 1 0 1 0 0 1.925.25.25 0 0 0 .48-.098.25.25 0 0 0-.05-.149l.016-.017a.25.25 0 0 0 .31-.005.25.25 0 0 0-.063-.424l.006-.022z" clip-rule="evenodd"></path><path fill="#000" fill-rule="evenodd" d="M16.3 6.175a.3.3 0 0 1-.096-.019.3.3 0 0 1-.081-.054.3.3 0 0 1-.054-.081.26.26 0 0 1 0-.192.26.26 0 0 1 .092-.112.3.3 0 0 1 .09-.037.26.26 0 0 1 .188.037l.038.031a.26.26 0 0 1 .073.177.25.25 0 0 1-.154.231.3.3 0 0 1-.096.02Zm2-.125a.26.26 0 0 1-.208-.11.26.26 0 0 1-.042-.139.26.26 0 0 1 .11-.209.3.3 0 0 1 .091-.037.26.26 0 0 1 .188.037l.038.031a.26.26 0 0 1 .054.273l-.023.043-.031.038a.26.26 0 0 1-.177.073" clip-rule="evenodd"></path>',29)]))}const gu={render:mu},vu={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function bu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",vu,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"#512BD4",d:"M22 2H2v20h20z"},null,-1),(0,$t.Lk)("path",{fill:"#fff",d:"M5.565 14.778a.5.5 0 0 1-.361-.142.47.47 0 0 1-.148-.344.47.47 0 0 1 .148-.348.5.5 0 0 1 .36-.145q.216 0 .365.145a.46.46 0 0 1 .152.348.46.46 0 0 1-.152.344.5.5 0 0 1-.364.142m5.653-.08h-.92l-2.421-3.82a1.7 1.7 0 0 1-.152-.301h-.021q.028.166.028.714v3.407h-.813V9.5h.98l2.34 3.73q.148.232.19.319h.015a5 5 0 0 1-.036-.7V9.5h.81zm3.96 0h-2.846V9.5h2.732v.732h-1.89v1.472h1.742v.728h-1.743v1.537h2.005zm4.044-4.466h-1.456v4.466h-.842v-4.466h-1.453V9.5h3.751z"},null,-1)]))}const yu={render:bu},wu={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ou(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",wu,t[0]||(t[0]=[(0,$t.Fv)('<path fill="#333" fill-rule="evenodd" d="M18.578 20.398A9.98 9.98 0 0 0 22 12.866c0-5.523-4.477-10-10-10s-10 4.477-10 10a9.97 9.97 0 0 0 3.24 7.369l.01-.064 13.123-4.297z" clip-rule="evenodd"></path><path fill="#CCC" d="m9.419 4.067-1.774-.264.265-.081q.242-.076.444-.228l.042-.032a1.22 1.22 0 0 0 .466-.778l.02-.124q.032-.202-.003-.403a.073.073 0 0 1 .1-.08l.072.03q.136.056.263.132l.302.182c.4.242.752.558 1.036.931l.2.264a1.33 1.33 0 0 0 .618.45l.14.05-.732.03a8 8 0 0 1-1.46-.08Z"></path><path fill="#CCC" d="m8.966 5.64.526-.962a.03.03 0 0 0 0-.03l-.179-.378a.68.68 0 0 0-.614-.388l-.05-.002a4.7 4.7 0 0 1-1.017-.148l-.065-.017a4 4 0 0 1-.79-.297l-.111-.055a3 3 0 0 1-.474-.3l-.147-.115a.1.1 0 0 0-.14.02 1 1 0 0 0-.185.47l-.027.235a2 2 0 0 0-.004.379l.065.734c.049.547.237 1.071.548 1.524l.066.097q.144.209.331.38a.1.1 0 0 0 .064.027A2.28 2.28 0 0 0 8.966 5.64"></path><path fill="#EDEDED" d="m6.976 6.832.625-.27a.45.45 0 1 0-.458-.769l-.148.117a1 1 0 0 0-.097.094l-.356.428a.15.15 0 0 0 .009.203l.11.126a.275.275 0 0 0 .315.071"></path><path fill="#B2B2B2" d="m7.78 3.758-.018.006.057.018c.247.08.504.125.763.138a.1.1 0 0 0 .038-.005l.118-.034.059-.02-.374-.08a.3.3 0 0 1-.134-.066.063.063 0 0 1 0-.096l.056-.046a.6.6 0 0 0 .157-.2c-.216.171-.46.301-.721.385Z"></path><path fill="#fff" d="M9.92 10.578 4.825 19.69a.078.078 0 0 1-.124.017l-.102-.103-.43-.488-.274-.354a8 8 0 0 1-.693-1.067l-.04-.072a8 8 0 0 1-.536-1.24l-.137-.412a5.55 5.55 0 0 1-.139-2.295l.005-.032a5.7 5.7 0 0 1 .24-.988l.226-.669.17-.356a5 5 0 0 1 1.067-1.473l.215-.204a3.8 3.8 0 0 1 1.576-.897 3 3 0 0 1-.422.084l-.281.033-.215.02-.235.04-.234.021-.156.017-.113.007c-.036.002-.053-.044-.024-.065l.157-.076a2.65 2.65 0 0 0 1.19-1.135l.174-.271a2.6 2.6 0 0 1 .808-.802l.067-.043q.182-.115.385-.187l.615-.217a.43.43 0 0 0 .253-.236l.009-.022a.34.34 0 0 0-.083-.38l-.108-.099-.018-.02-.696-.926-.555-.764a.97.97 0 0 1-.184-.517.04.04 0 0 1 .056-.038l.6.288a2.38 2.38 0 0 0 1.666.148q.195-.053.394-.075l.276-.028a5.7 5.7 0 0 1 2.064.17l.058.016.656.236a3 3 0 0 1 1.785 1.756l.138.536a2.17 2.17 0 0 0 .863 1.246l.35.244q.232.162.483.287l.775.388a.062.062 0 0 1 .022.093 2.1 2.1 0 0 1-.82.644l-.449.195a4 4 0 0 1-.55.193l-.33.088a5.5 5.5 0 0 1-1.417.188h-.79a1.3 1.3 0 0 1-.477-.091 1.3 1.3 0 0 0-1.615.576Z"></path><path fill="#666" d="m12.593 7.292.344.004c.047 0 .057-.065.013-.079a1.4 1.4 0 0 1-.212-.087l-.269-.139a2.8 2.8 0 0 1-.51-.338l-.053-.044q-.18-.15-.323-.334l-.06-.078a.034.034 0 0 0-.062.023c.04.598.533 1.065 1.132 1.072"></path><path fill="#CCC" d="m5.212 19.834 5.824-5.195a7.8 7.8 0 0 0 1.417-1.677l.035-.057a2.3 2.3 0 0 0 .237-1.239l-.052-.587a1.66 1.66 0 0 0-.424-.97.1.1 0 0 0-.036-.02l-.936-.192a4.2 4.2 0 0 0-2.092-.086l-.037.008-.544.127a6.74 6.74 0 0 0-3.94 2.63l-.023.034a6.4 6.4 0 0 0-.665 1.168l-.032.074a6.07 6.07 0 0 0-.441 3.31l.054.367a3.7 3.7 0 0 0 .932 1.96l.297.324a.304.304 0 0 0 .426.021"></path><path fill="#5E5E5E" d="m10.31 12.9-.018.01c-.492.281-.756.811-.73 1.377.007.16.168.308.323.264l.285-.081a2.6 2.6 0 0 0 1.001-.539l.829-.713.246-.182a6.6 6.6 0 0 1 4.979-1.209c.054.009.056.086.002.097l-.207.043-.095.033a1 1 0 0 0-.48.359.04.04 0 0 0 .044.062l.068-.021a3.2 3.2 0 0 1 1.184-.132l.157.012a2.86 2.86 0 0 1 2.159 1.137l.192.256c.57.758.812 1.714.67 2.652l-.074.49a4.7 4.7 0 0 1-.713 1.87l-.027.04-.41.547-.449.508-.411.395q-.566.542-1.227.962l-.14.088-.266.127c-.069.033-.126-.061-.065-.107l.37-.352.077-.08c.206-.216.355-.482.431-.77l.035-.134a.715.715 0 0 0-.549-.886.7.7 0 0 0-.583.146l-.367.299-.649.53a8 8 0 0 1-1.376.902l-.67.345-1.133.43-.238.064a8 8 0 0 1-1.263.235l-.316.032H9.891l-1.348-.146-.73-.158a4.55 4.55 0 0 1-1.91-.917l-.563-.458a3.02 3.02 0 0 1-1.177-2.168l-.034-.45.04-.631a4.6 4.6 0 0 1 1.956-3.483l.12-.084a5.2 5.2 0 0 1 1.646-.763l.396-.105a3.4 3.4 0 0 1 1.577-.042l.413.086c.104.022.126.16.034.212Z"></path>',9)]))}const ku={render:Ou},xu={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Eu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",xu,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"currentColor",d:"m16.378 12.034-4.148-5.53h.507c.369 0 .738.185.968.461l3.272 4.286 3.226-4.24c.322-.322.691-.507 1.06-.507h.553l-4.194 5.53L22 17.841h-.553c-.369 0-.737-.185-.968-.507l-3.456-4.516-3.456 4.516c-.23.322-.6.46-.968.46h-.553zm-13.456.461v.369c0 1.935 1.29 3.686 3.18 4.147a4.15 4.15 0 0 0 4.654-2.258c.138-.23.368-.415.645-.415h.46a5.07 5.07 0 0 1-4.792 3.687C4.212 18.025 2 15.583 2 12.541v-1.29c0-2.489 1.751-4.747 4.24-5.207a5.07 5.07 0 0 1 5.898 5.069v1.382zm0-.921h8.295v-.461a4.148 4.148 0 0 0-5.07-4.056c-1.935.461-3.225 2.212-3.225 4.378z"},null,-1)]))}const Su={render:Eu},_u={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Cu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",_u,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"currentColor",d:"M12 2a10 10 0 1 0 0 20 10 10 0 0 0 0-20m-.7 4.5h6.3L11.3 11h4.3l-9.2 6.4 2-4.4 1-2 2-4.5z"},null,-1)]))}const Tu={render:Cu},Au={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ru(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Au,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"currentColor",d:"M23.241 7.492 24 5.536l-.031-.125-7.479 1.971c.798-1.158.524-2.08.524-2.08s-2.386 1.525-4.193 1.486c-1.807-.04-2.386-.525-5.155.36-2.769.883-3.551 3.59-4.357 4.177-.798.579-3.309 2.51-3.309 2.51l2.269-.727s-.642.61-1.964 2.386v.008c.211.297 1.142 1.525 2.065 1.26.102-.032.22-.087.352-.15.415.236.97.462 1.573.525 0 0-.407-.47-.751-1.017l.281-.18-.047.031.869.313-.094-.813h.008l.844.313-.101-.744.32-.156.884-3.348 3.661-2.503-.289.735c-.743 1.823-2.136 2.253-2.136 2.253l-.578.227c-.438.508-.618.634-.767 2.354.352-.093.68-.11.986-.03 1.564.422 2.104 2.307 1.681 2.831-.101.125-.352.352-.672.61H7.76l-.008.516-.062.055h-.65l-.007.5-.173.126c-.602.016-1.376-.516-1.376-.516.008.485.406 1.228.406 1.228l.07-.04-.062.048s1.627 1.08 2.652.68c.908-.36 3.262-2.214 5.296-3.09l6.149-1.627.813-2.096-4.685 1.236V13.14l5.499-1.448.814-2.104-6.313 1.666V9.37zm-11.163 3.997 1.463-.383.016.07-.454 1.181-1.51.4zm.5 2.527-1.509.399.493-1.268 1.455-.383.023.07zm1.972-.423-1.51.4.493-1.268 1.455-.383.024.07z"},null,-1)]))}const Pu={render:Ru},Du={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Mu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Du,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"currentColor",d:"M1.81 10.714c-.046 0-.058-.023-.034-.058l.245-.316c.023-.035.082-.058.128-.058h4.17c.047 0 .059.035.035.07l-.198.304c-.024.035-.082.07-.117.07zM.048 11.789c-.047 0-.059-.024-.035-.059l.245-.315c.023-.035.082-.059.128-.059h5.327c.047 0 .07.036.058.07l-.093.28c-.012.048-.058.07-.105.07zm2.826 1.074c-.046 0-.058-.035-.035-.07l.164-.292c.023-.035.07-.07.117-.07h2.336c.047 0 .07.035.07.082l-.023.28c0 .047-.047.082-.082.082l-2.546-.012Zm12.125-2.36-1.962.515c-.175.046-.187.058-.339-.117a1.6 1.6 0 0 0-.549-.444 2.03 2.03 0 0 0-2.114.175 2.49 2.49 0 0 0-1.192 2.22c.012.934.655 1.705 1.577 1.833.795.106 1.46-.175 1.986-.77l.316-.433h-2.255c-.245 0-.304-.151-.222-.35.152-.362.432-.97.596-1.273a.32.32 0 0 1 .292-.187h4.252c-.024.315-.024.63-.07.946a5 5 0 0 1-.958 2.29 4.87 4.87 0 0 1-3.33 1.985 4.14 4.14 0 0 1-3.141-.77A3.67 3.67 0 0 1 6.4 13.528a4.58 4.58 0 0 1 .993-3.422 5.2 5.2 0 0 1 3.27-2.021c1.099-.199 2.15-.07 3.096.572.62.409 1.063.97 1.355 1.647.07.105.024.164-.117.199Z"},null,-1),(0,$t.Lk)("path",{fill:"currentColor",d:"M18.865 16.963a4.38 4.38 0 0 1-2.85-1.028 3.67 3.67 0 0 1-1.262-2.254 4.46 4.46 0 0 1 .946-3.528 4.9 4.9 0 0 1 3.27-1.95 4.32 4.32 0 0 1 3.33.595 3.63 3.63 0 0 1 1.647 2.605 4.51 4.51 0 0 1-1.343 3.96 5.35 5.35 0 0 1-2.804 1.495c-.315.058-.63.07-.934.105m2.78-4.719c-.012-.152-.012-.268-.035-.385a1.93 1.93 0 0 0-2.383-1.554c-1.087.245-1.787.935-2.044 2.033-.21.91.233 1.834 1.074 2.207.643.28 1.285.246 1.904-.07a2.56 2.56 0 0 0 1.484-2.23Z"},null,-1)]))}const $u={render:Mu},Nu={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Iu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Nu,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"#FF5B11",d:"M12.371 1.736q.08-.01.141.04a76 76 0 0 1 4.217 5.662 28 28 0 0 1 2.49 4.659q1.64 4.374-1.707 7.61-2.928 2.475-6.747 1.928-4.564-.93-6.205-5.281a7.4 7.4 0 0 1-.28-2.891 17.3 17.3 0 0 1 1.204-4.9 7.3 7.3 0 0 1 1.044-1.767q.525.625 1.004 1.285.222.232.462.442a38 38 0 0 1 4.377-6.787",opacity:".993"},null,-1),(0,$t.Lk)("path",{fill:"#FF9758",d:"M12.09 5.962a42.5 42.5 0 0 1 3.996 5.462q.56.959.924 2.008.758 2.982-1.627 4.92-2.306 1.622-5.06.963-2.97-.921-3.433-3.996a4.25 4.25 0 0 1 .2-1.887c.297-.754.66-1.48 1.085-2.168L9.38 9.495q1.367-1.758 2.71-3.534Z"},null,-1)]))}const Lu={render:Iu},Bu={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Fu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Bu,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"#FF2D20",d:"M21.217 6.8a.3.3 0 0 1 .012.076v4.078a.3.3 0 0 1-.15.26l-3.424 1.969v3.907a.3.3 0 0 1-.147.259l-7.15 4.111q-.023.014-.05.024l-.021.007a.3.3 0 0 1-.152 0l-.024-.01c-.017-.007-.033-.011-.048-.02l-7.142-4.112a.3.3 0 0 1-.15-.26V4.857q0-.043.012-.079c0-.01.007-.016.01-.023a.2.2 0 0 1 .019-.048l.021-.026.024-.036.03-.024.034-.023L6.492 2.54a.29.29 0 0 1 .298 0l3.573 2.053.033.024.029.024.026.036.022.023.019.048c0 .01.007.017.01.024a.4.4 0 0 1 .009.08v7.643l2.978-1.713v-3.91c0-.023.003-.052.01-.076l.01-.026.018-.047.022-.024.026-.036.029-.024.033-.023L17.21 4.56a.3.3 0 0 1 .298 0l3.57 2.058.034.024c.01.01.022.014.029.024q.017.015.026.035l.021.024c.01.017.015.03.02.048l.009.023zm-.58 3.98V7.392l-1.252.719-1.727.994v3.393l2.978-1.713zm-3.576 6.138v-3.395l-1.7.971-4.85 2.77v3.425l6.55-3.77ZM3.366 5.37v11.548l6.55 3.769V17.26l-3.421-1.934h-.003l-.03-.024c-.01-.01-.022-.014-.03-.024-.011-.01-.018-.024-.025-.03l-.024-.032-.014-.038q-.009-.015-.015-.033l-.007-.043-.002-.033V7.08l-1.728-.994-1.253-.716zm3.276-2.23L3.666 4.857 6.64 6.569l2.976-1.713-2.974-1.713zM8.19 13.833l1.727-.992V5.37l-1.247.718-1.727.995v7.47l1.249-.72zm9.169-8.667-2.974 1.713 2.974 1.713 2.974-1.713zm-.297 3.943-1.728-.995-1.249-.718v3.385l1.727.995 1.25.72v-3.39zm-6.845 7.639 4.365-2.491 2.182-1.244L13.79 11.3l-3.426 1.972-3.117 1.796z"},null,-1)]))}const Qu={render:Fu},Uu={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function zu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Uu,t[0]||(t[0]=[(0,$t.Fv)('<g fill="#EDB641" clip-path="url(#a)"><path d="m5.33 10.983.419 1.29H7.11l-1.1.796.42 1.29-1.101-.797-1.09.791.419-1.29-1.1-.796H4.92l.42-1.289-.011.005Zm1.954-4.245.362 1.111h1.169l-.944.687.362 1.111-.949-.681-.943.681.351-1.111-.943-.687h1.169l.361-1.11h.005ZM1.31 13.504l.21.65h.68l-.55.409.21.65-.55-.404-.556.404.21-.65-.551-.404h.681l.21-.65.005-.005Z"></path><path d="M9.197 15.684a19 19 0 0 0 2.505-1.378 22.6 22.6 0 0 0 3.092-2.6l-.613-.408-.534-.362 1.692-.063.545-.02.147-.525.461-1.635.587 1.588.189.514.545-.027 1.693-.062-1.332 1.048-.43.335.19.514.586 1.588-1.41-.944-.45-.298-.43.335-1.331 1.048.23-.812.231-.818.026-.105a21 21 0 0 1-5.24 2.815c-.315.104-.63.188-.944.267zm-3.186 1.242a16.1 16.1 0 0 0 8.081-2.494l-.535 1.923 3.145-2.479 3.328 2.222-1.384-3.752 3.144-2.474-4.003.158-1.384-3.758-1.09 3.852-3.999.157 2.317 1.546c-1.52 1.384-5.346 4.34-10.413 4.538A14 14 0 0 1 0 16.11s2.52.953 6.016.817h-.005Z"></path><path d="M16.408 4.412a7.588 7.588 0 0 0-7.242 9.862q.495-.25.943-.508a6.53 6.53 0 0 1 6.3-8.306 6.54 6.54 0 1 1-4.916 10.848c-.351.136-.703.262-1.059.367a7.588 7.588 0 1 0 5.974-12.263"></path></g><defs><clipPath id="a"><path fill="#fff" d="M0 0h24v24H0z"></path></clipPath></defs>',2)]))}const ju={render:zu},Wu={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Xu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Wu,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"#E0234E","fill-rule":"evenodd",d:"M13.776 2.04c-.144 0-.279.03-.403.071.264.176.409.409.48.672.006.036.016.062.021.098q.009.046.01.093c.021.455-.118.511-.216.78-.15.346-.109.718.072 1.017a.6.6 0 0 0 .062.114c-.196-1.307.894-1.503 1.095-1.91.016-.357-.279-.595-.511-.76a1.14 1.14 0 0 0-.61-.176Zm1.643.294c-.021.118-.005.087-.01.15-.006.04-.006.092-.01.134q-.016.063-.037.124a1 1 0 0 1-.042.124c-.02.041-.035.077-.056.118-.016.021-.026.042-.042.062l-.03.047a2 2 0 0 1-.078.103c-.03.031-.057.067-.093.093v.005c-.031.026-.062.057-.098.083-.108.082-.233.145-.346.222-.036.026-.073.046-.103.078a1 1 0 0 0-.099.082c-.036.031-.062.062-.093.098-.025.031-.056.067-.077.104q-.039.052-.072.108c-.021.041-.036.077-.057.119-.016.041-.031.077-.042.119a2 2 0 0 0-.036.129c-.005.02-.005.046-.01.067-.005.02-.005.041-.01.062 0 .041-.005.088-.005.129q-.001.046.005.093 0 .06.015.13.01.06.026.123l.041.124c.01.026.026.052.037.072l-1.188-.46a13 13 0 0 0-.6-.154l-.325-.078a10 10 0 0 0-.94-.14c-.01 0-.016-.004-.026-.004a9 9 0 0 0-1.601-.021c-.315.02-.63.062-.945.114q-.116.017-.233.04-.232.048-.465.104l-.232.062c-.077.031-.15.067-.222.098l-.17.078c-.01.005-.021.005-.026.01-.052.026-.098.047-.145.073a2 2 0 0 0-.196.098c-.036.015-.072.036-.104.051a3 3 0 0 0-.305.18c-.035.027-.071.048-.103.073-.005.005-.01.005-.015.01a1 1 0 0 0-.098.073l-.01.01-.078.062c-.01.005-.02.016-.03.02l-.078.068c-.006.01-.016.016-.021.02-.031.032-.062.058-.093.088-.005 0-.005.006-.01.011a2 2 0 0 0-.094.088c-.005.005-.005.01-.01.01a1 1 0 0 0-.077.083c-.01.01-.026.02-.036.03q-.04.047-.088.094c-.005.01-.016.015-.02.025-.042.042-.078.083-.12.124l-.015.016q-.123.134-.264.253-.14.125-.289.227a3.2 3.2 0 0 1-.635.351 4 4 0 0 1-.336.124c-.217.047-.439.135-.63.15-.041 0-.088.01-.129.016l-.129.03-.124.047a1 1 0 0 0-.124.057c-.036.026-.077.046-.114.072a1 1 0 0 0-.103.088 1 1 0 0 0-.103.093c-.031.036-.062.067-.088.103-.026.042-.057.078-.078.12a1 1 0 0 0-.072.118q-.032.067-.057.134l-.046.134a1 1 0 0 0-.026.124c0 .006-.005.011-.005.016-.01.047-.01.108-.016.14q-.009.05-.01.103a.5.5 0 0 0 .026.165q.014.046.036.093v.005a1 1 0 0 0 .057.093 1 1 0 0 0 .067.093q.042.044.088.083.045.045.098.082c.124.109.155.145.315.228q.038.02.083.041c.005 0 .01.005.015.005q-.001.013.005.026.01.062.026.124a1 1 0 0 0 .041.124c.016.03.026.062.042.093q.007.016.015.026l.062.113.078.109q.042.051.088.098.045.043.098.088s.005.005.01.005a1 1 0 0 0 .207.14 1 1 0 0 0 .118.056.6.6 0 0 0 .104.036c.005.005.01.005.02.01.021.006.047.011.068.016-.016.279-.021.543.02.635.047.104.274-.211.501-.573-.03.357-.051.775 0 .899.057.129.367-.274.636-.718 3.662-.847 7.003 1.684 7.354 5.258-.067-.558-.754-.868-1.069-.79-.155.382-.418.873-.841 1.177.037-.345.02-.693-.052-1.033a3.3 3.3 0 0 1-.64 1.302c-.491.036-.982-.202-1.24-.558-.02-.015-.026-.046-.042-.067a1 1 0 0 1-.04-.108.4.4 0 0 1-.032-.109q-.006-.052-.005-.114v-.077a1 1 0 0 1 .026-.109 1 1 0 0 1 .036-.108c.02-.036.036-.072.062-.108.088-.248.088-.45-.072-.569a.6.6 0 0 0-.099-.051c-.02-.006-.046-.016-.067-.021l-.041-.015a1 1 0 0 0-.109-.026.4.4 0 0 0-.108-.016 1 1 0 0 0-.114-.01c-.025 0-.051.005-.077.005a.4.4 0 0 0-.114.015q-.054.007-.108.021a1 1 0 0 0-.109.036l-.103.047c-.03.015-.062.036-.098.052-1.203.784-.486 2.623.336 3.155-.31.057-.625.124-.713.191l-.01.01q.334.202.697.347.415.134.837.248v.005a5 5 0 0 0 1.301.093c2.288-.16 4.163-1.9 4.504-4.194l.031.134c.015.093.036.191.046.29v.005q.014.07.021.134v.02q.009.072.01.135.01.085.01.17v.083c0 .026.006.057.006.083 0 .03-.005.062-.005.093v.072c0 .036-.005.067-.005.103q.001.029-.006.067c0 .036-.005.073-.005.114q-.006.022-.005.046l-.015.12q.001.022-.005.046c-.006.051-.016.098-.021.15v.01c-.01.046-.02.098-.031.144v.016l-.031.14c0 .005-.005.015-.005.02a1 1 0 0 1-.036.14v.015q-.025.076-.042.145-.006.009-.005.01l-.046.155c-.021.052-.037.098-.057.15s-.036.103-.057.15l-.062.15h-.005c-.021.046-.041.097-.067.144l-.016.036c-.005.005-.005.01-.01.015a4.9 4.9 0 0 1-1.451 1.736q-.063.04-.125.088c-.01.01-.026.015-.036.026q-.053.038-.113.077l.015.031h.005l.217-.03h.005q.202-.033.403-.073a1 1 0 0 0 .114-.026l.072-.015c.036-.006.072-.016.109-.021.03-.01.062-.016.093-.026a10 10 0 0 0 1.497-.496 8.3 8.3 0 0 1-3.207 2.619 8.5 8.5 0 0 0 1.761-.305 8.25 8.25 0 0 0 4.881-3.9 8.2 8.2 0 0 1-1.39 3.321c.501-.33.964-.714 1.38-1.146a8.2 8.2 0 0 0 2.174-4.39c.177.817.227 1.657.15 2.49 3.73-5.202.31-10.594-1.12-12.015-.006-.01-.01-.015-.01-.026-.006.006-.006.006-.006.011 0-.005 0-.005-.005-.01q-.002.092-.01.186a7 7 0 0 1-.052.345 8 8 0 0 1-.088.341 4 4 0 0 1-.124.33q-.072.162-.155.316a4 4 0 0 1-.398.573c-.077.093-.16.176-.242.259a3 3 0 0 1-.15.129l-.12.103c-.092.072-.185.14-.288.201a4 4 0 0 1-.305.176 9 9 0 0 1-.325.145 3.6 3.6 0 0 1-.682.186c-.12.02-.238.03-.351.04a4 4 0 0 1-.248.011c-.12 0-.238-.01-.351-.02a3 3 0 0 1-.352-.052 3 3 0 0 1-.346-.088h-.005c.114-.01.227-.02.341-.041a3.7 3.7 0 0 0 1.317-.501q.148-.099.29-.207.138-.108.263-.232c.087-.078.165-.165.242-.253q.117-.14.217-.28c.01-.015.02-.035.031-.05q.08-.125.15-.249a3.6 3.6 0 0 0 .367-.981c.02-.119.041-.233.052-.346.01-.119.02-.238.02-.351q-.001-.125-.01-.248-.017-.178-.041-.346a4 4 0 0 0-.078-.346c-.036-.109-.072-.222-.114-.331a3.5 3.5 0 0 0-.527-.91 11 11 0 0 0-.237-.262 3 3 0 0 0-.134-.13 9 9 0 0 0-.718-.506 1 1 0 0 0-.104-.051 2 2 0 0 0-.49-.217","clip-rule":"evenodd"},null,-1)]))}const qu={render:Xu},Vu={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Hu(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Vu,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"currentColor",d:"M11.345 2.005c-.043.004-.18.018-.303.028-2.84.256-5.501 1.788-7.187 4.144a9.9 9.9 0 0 0-1.765 4.369c-.08.549-.09.711-.09 1.456s.01.907.09 1.456c.543 3.755 3.216 6.91 6.84 8.08.65.208 1.334.351 2.112.437.303.033 1.613.033 1.916 0 1.343-.149 2.48-.48 3.603-1.053.172-.088.205-.112.182-.131a191 191 0 0 1-1.629-2.184l-1.599-2.16-2.004-2.965a284 284 0 0 0-2.017-2.964 161 161 0 0 0-.02 2.925c-.006 2.816-.008 2.93-.043 2.996a.36.36 0 0 1-.172.178c-.063.031-.117.037-.412.037h-.339l-.09-.056a.4.4 0 0 1-.13-.143l-.042-.088.004-3.92.006-3.92.06-.077a.5.5 0 0 1 .145-.12c.08-.038.112-.042.45-.042.399 0 .465.016.569.129.03.031 1.114 1.665 2.412 3.634l3.946 5.975 1.583 2.399.08-.053a10.3 10.3 0 0 0 2.055-1.802 9.95 9.95 0 0 0 2.354-5.112c.08-.549.09-.711.09-1.456s-.01-.907-.09-1.456c-.543-3.755-3.216-6.91-6.84-8.08a10.5 10.5 0 0 0-2.083-.435c-.187-.02-1.48-.041-1.642-.026m4.094 6.048c.094.047.17.137.197.231.016.05.02 1.138.016 3.587l-.006 3.515-.62-.95-.621-.95V10.93c0-1.652.008-2.58.02-2.625a.4.4 0 0 1 .193-.247c.08-.04.11-.045.416-.045.29 0 .34.004.405.04"},null,-1)]))}const Zu={render:Hu},Gu={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ku(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Gu,t[0]||(t[0]=[(0,$t.Fv)('<path fill="url(#a)" fill-rule="evenodd" d="M19.608 5.51C16.023 1.308 9.712.808 5.51 4.392 1.308 7.977.808 14.288 4.392 18.49c3.585 4.202 9.896 4.702 14.098 1.118 4.202-3.585 4.702-9.896 1.118-14.098m-4.982 5.249c.385 0 .631.417.438.75l-.056.096-3.774 6.21a.39.39 0 0 1-.33.185h-.36a.378.378 0 0 1-.366-.472l1.024-3.959a.5.5 0 0 0-.484-.625H9.145a.5.5 0 0 1-.484-.625l1.545-5.978a.46.46 0 0 1 .448-.34l.067.001h1.584a.5.5 0 0 1 .496.568l-.01.071-.604 3.533a.5.5 0 0 0 .493.585z" clip-rule="evenodd"></path><defs><radialGradient id="a" cx="0" cy="0" r="1" gradientTransform="matrix(19.5 0 0 198.851 4 12)" gradientUnits="userSpaceOnUse"><stop stop-color="#31B2F3"></stop><stop offset=".474" stop-color="#F27CEC"></stop><stop offset="1" stop-color="#FD6641"></stop></radialGradient></defs>',2)]))}const Yu={render:Ku},Ju={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ed(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",Ju,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"#00DC82",d:"M13.32 19.333h8.14c.26 0 .508-.09.733-.22.226-.129.457-.29.587-.513s.22-.476.22-.733c0-.258-.09-.51-.22-.734l-5.5-9.46a1.34 1.34 0 0 0-.513-.513 1.8 1.8 0 0 0-.807-.22c-.26 0-.508.09-.733.22-.216.12-.393.298-.514.513l-1.393 2.42L10.533 5.4c-.13-.223-.288-.458-.513-.587s-.474-.146-.733-.146c-.26 0-.509.018-.734.146-.24.146-.441.347-.586.587l-6.82 11.733c-.13.223-.147.476-.147.734 0 .257.017.51.147.733s.361.384.586.513c.225.13.474.22.734.22H7.6c2.034 0 3.515-.912 4.547-2.64l2.493-4.326 1.32-2.274 4.033 6.894H14.64zm-5.793-2.346H3.933l5.354-9.24L12 12.367l-1.797 3.132c-.687 1.122-1.468 1.488-2.676 1.488"},null,-1)]))}const td={render:ed},nd={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function rd(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",nd,t[0]||(t[0]=[(0,$t.Fv)('<path fill="#93D500" d="M7.285 13.5H2.007l.003.077.009.15.004.063.014.171.004.039a8.8 8.8 0 0 0 .387 1.871q0 .005.003.011l.058.173.016.044.08.216.048.124a5 5 0 0 0 .078.189l.052.117.031.07.066.14.022.046.079.16.012.024q.045.088.093.175l.004.008.02.036 4.508-2.715.017-.01a3.6 3.6 0 0 1-.33-1.179"></path><path fill="#4D5A31" d="m8.582 15.923-.013.013-3.717 3.717.057.053.105.094.055.048.127.107.034.029a8 8 0 0 0 .324.254l.006.004q.336.252.695.47l.025.016.137.08.084.048.08.045q.072.04.145.078a8 8 0 0 0 .431.21l.132-.32 1.87-4.54.007-.017a3.6 3.6 0 0 1-.584-.389"></path><path fill="#6BA43A" d="M8.195 15.553a3 3 0 0 1-.212-.262 4 4 0 0 1-.2-.296L3.27 17.713q.104.174.215.338l.007.012.003.005.007.01v.002l.019.026.001.002q.119.174.242.342l.006.007.005.007.11.142.015.019q.057.073.117.145l.028.034.155.181.094.105.053.058.024.026.07.075.028.03.125.127 3.723-3.723a3 3 0 0 1-.122-.13"></path><path fill="#4D5A31" d="m13.13 15.921-.124.096.009.016 2.714 4.506a8 8 0 0 0 1.134-.886l-3.718-3.717z"></path><path fill="#93D500" d="m15.335 20.576-.092-.153-2.534-4.206a4 4 0 0 1-.316.167 3.6 3.6 0 0 1-2.567.2c-.057-.017-.113-.04-.169-.06s-.113-.037-.168-.06L7.62 21.003l-.073.178-.063.152.014.006.016.006.004.002.133.05.167.065.085.032a9 9 0 0 0 1.88.442l.075.008.142.016.036.003.174.014.06.004.153.01.094.003.122.004.217.003a9 9 0 0 0 1.495-.126l.154-.028.088-.017.09-.02.15-.032.028-.007a9 9 0 0 0 2.124-.791l.13-.074.157-.088.107-.06.001-.001.002-.001h.001l.001-.001.002-.001.006-.004.034-.02zm-6.75-10.21.123-.097-.009-.015-2.714-4.506q-.189.123-.372.256a9 9 0 0 0-.762.629L8.57 10.35z"></path><path fill="#4D5A31" d="M4.594 6.88q-.064.065-.124.13l-.123.128a9 9 0 0 0-.753.933l-.032.046-.076.112-.104.16a8.83 8.83 0 0 0-1.375 4.397l-.005.178-.003.179h5.266q.002-.09.008-.179c.006-.089.004-.119.01-.178a3.57 3.57 0 0 1 .794-1.917q.059-.069.12-.133c.04-.044.077-.09.12-.133L4.594 6.881Zm10.864-1.306q-.01-.008-.021-.013l-.14-.083-.082-.047-.082-.046-.143-.077-.023-.012a9 9 0 0 0-.95-.426l-.024-.01a9 9 0 0 0-1.987-.498l-.074-.01-.075-.008-.143-.016-.04-.004-.17-.014-.066-.004-.146-.009-.078-.003v5.278c.27.028.54.086.8.174l3.884-3.884a9 9 0 0 0-.44-.288"></path><path fill="#6BA43A" d="M10.679 4.29q-.09 0-.179.003a9 9 0 0 0-1.113.115l-.024.004-.155.029-.087.016a6 6 0 0 0-.24.052l-.028.006a9 9 0 0 0-2.125.792l-.13.074-.156.087-.115.066h-.002l-.008.005-.03.017v.001l.093.153 2.625 4.359a4 4 0 0 1 .316-.167 3.6 3.6 0 0 1 1.536-.35V4.285q-.09.001-.178.004Z"></path><path fill="#4D5A31" d="M19.704 12.708a10 10 0 0 0-.05-.61l-.002-.014-.001-.007-.027-.201v-.006a9 9 0 0 0-.273-1.245l-.002-.008-.059-.19-.008-.025-.057-.17-.016-.046-.054-.147-.027-.07a5 5 0 0 0-.086-.216l-.04-.094-.052-.12-.03-.068-.066-.142-.021-.043-.08-.162-.012-.022-.093-.177-.003-.006a9 9 0 0 0-.504-.818l-3.886 3.886q.132.392.174.8h5.278q0-.04-.003-.08Z"></path><path fill="#6BA43A" d="M14.45 13.143q-.002.09-.01.179-.001.09-.01.178a3.57 3.57 0 0 1-.793 1.918c-.037.046-.08.088-.12.133-.04.044-.077.09-.12.132l3.723 3.723q.063-.064.124-.13l.123-.128q.415-.448.754-.935l.027-.039.082-.12.07-.108.034-.052a8.83 8.83 0 0 0 1.377-4.573l.003-.178z"></path><path fill="#424143" d="M21.279 2.722a2.467 2.467 0 0 0-3.953 2.84l-5.374 5.373a2.467 2.467 0 1 0 1.114 1.113l5.373-5.373a2.466 2.466 0 0 0 2.84-3.953"></path>',10)]))}const ad={render:rd},od={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function id(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",od,t[0]||(t[0]=[(0,$t.Lk)("path",{stroke:"#21FA90","stroke-linecap":"round","stroke-linejoin":"round",d:"M5.038 13.757v3.814m1.496-4.566 3.04 1.761 4.534-2.627V6.885L9.574 4.258 5.038 6.885v4.56m1.496 1.56v1.668m0-1.668 1.49-.91"},null,-1),(0,$t.Lk)("path",{stroke:"#00050B","stroke-linecap":"round","stroke-linejoin":"round","stroke-width":".885",d:"m21.719 10.719 1.839 1.09v1.09l-1.84 1.091M2.834 8.13.774 9.322v5.326l2.016 1.167m3.744 1.07v1.183l3.021 1.674 2.32-1.358-.02-2.471m2.282-1.387v1.3L16.31 17.1l3.175-1.9v-5.186L16.36 8.217"},null,-1)]))}const sd={render:id},ld={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function cd(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",ld,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"#61DAFB",d:"M12.5 14.014a1.783 1.783 0 1 0 0-3.565 1.783 1.783 0 0 0 0 3.565"},null,-1),(0,$t.Lk)("path",{stroke:"#61DAFB","stroke-width":".87",d:"M12.5 15.884c5.283 0 9.565-1.635 9.565-3.652S17.783 8.58 12.5 8.58s-9.565 1.635-9.565 3.652 4.282 3.652 9.565 3.652Z"},null,-1),(0,$t.Lk)("path",{stroke:"#61DAFB","stroke-width":".87",d:"M9.337 14.058c2.642 4.575 6.199 7.466 7.946 6.457 1.746-1.008 1.021-5.534-1.62-10.11C13.02 5.832 9.464 2.94 7.717 3.949s-1.021 5.535 1.62 10.11Z"},null,-1),(0,$t.Lk)("path",{stroke:"#61DAFB","stroke-width":".87",d:"M9.337 10.406c-2.641 4.575-3.366 9.1-1.62 10.11 1.747 1.008 5.304-1.883 7.946-6.458s3.366-9.101 1.62-10.11c-1.747-1.009-5.305 1.883-7.946 6.458Z"},null,-1)]))}const ud={render:cd},dd={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function pd(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",dd,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"currentColor",d:"m21.863 11.755-.839-.52-.024-.244.722-.673a.288.288 0 0 0-.096-.482l-.923-.345-.072-.239.575-.799a.288.288 0 0 0-.187-.455l-.973-.158-.117-.218.409-.897a.286.286 0 0 0-.273-.409l-.987.034-.157-.188.228-.962a.287.287 0 0 0-.348-.348l-.96.227-.19-.157.034-.986a.287.287 0 0 0-.409-.273l-.897.409-.218-.118-.158-.972a.288.288 0 0 0-.454-.189l-.8.575-.238-.071-.345-.923a.288.288 0 0 0-.482-.096L13.01 3l-.245-.023-.52-.84a.29.29 0 0 0-.49 0l-.52.84q-.123.01-.245.023l-.673-.722a.29.29 0 0 0-.482.096l-.345.923q-.12.034-.237.072l-.8-.576a.288.288 0 0 0-.455.188l-.158.973q-.11.057-.22.118l-.895-.41a.287.287 0 0 0-.409.273l.034.988-.19.156-.96-.227a.29.29 0 0 0-.348.348l.227.962-.157.188-.987-.034a.288.288 0 0 0-.273.409l.409.897a7 7 0 0 0-.117.22l-.972.156a.288.288 0 0 0-.189.455l.575.8-.071.237-.923.346a.29.29 0 0 0-.096.482L3 10.99l-.024.244-.84.52a.29.29 0 0 0 0 .492l.84.519q.01.123.025.245l-.723.673a.29.29 0 0 0 .096.482l.922.345q.035.12.073.237l-.576.8a.29.29 0 0 0 .188.454l.973.158q.056.11.117.22l-.41.896a.288.288 0 0 0 .274.409l.987-.035q.076.096.156.19l-.227.961a.287.287 0 0 0 .348.347l.961-.226.19.156-.035.987a.288.288 0 0 0 .41.273l.896-.41q.108.06.219.119l.158.971a.288.288 0 0 0 .455.19l.799-.577q.118.038.237.072l.345.923a.287.287 0 0 0 .482.096L10.99 21q.121.014.244.025l.52.839a.29.29 0 0 0 .492 0l.519-.84q.123-.01.245-.024l.674.722a.287.287 0 0 0 .48-.096l.346-.923q.12-.034.238-.072l.8.576a.29.29 0 0 0 .454-.189l.158-.972.219-.117.896.409a.288.288 0 0 0 .41-.273l-.035-.988a7 7 0 0 0 .19-.155l.96.226a.287.287 0 0 0 .348-.347l-.226-.961.155-.19.988.035a.29.29 0 0 0 .272-.41l-.409-.896q.06-.108.117-.219l.973-.158a.287.287 0 0 0 .188-.454l-.575-.8.072-.237.922-.345a.286.286 0 0 0 .096-.482L21 13.011l.025-.245.839-.52a.29.29 0 0 0 0-.49Zm-5.619 6.965a.595.595 0 0 1 .25-1.163.595.595 0 0 1-.25 1.163m-.286-1.93a.54.54 0 0 0-.642.417l-.299 1.391a7.3 7.3 0 0 1-3.017.65c-1.1 0-2.144-.243-3.08-.679l-.299-1.39a.54.54 0 0 0-.643-.417l-1.228.263a7 7 0 0 1-.635-.748h5.976c.068 0 .113-.012.113-.074V14.09c0-.062-.046-.074-.113-.074h-1.747v-1.34h1.89c.172 0 .922.049 1.162 1.007.075.295.24 1.254.352 1.561.113.344.57 1.032 1.059 1.032h2.977a1 1 0 0 0 .108-.01q-.31.421-.677.794l-1.256-.27Zm-8.265 1.9a.594.594 0 1 1-.014-1.17.594.594 0 0 1 .014 1.171ZM5.427 9.5a.594.594 0 1 1-1.086.481.594.594 0 0 1 1.086-.481m-.697 1.65 1.28-.569a.543.543 0 0 0 .274-.715l-.263-.597h1.036v4.672h-2.09a7.3 7.3 0 0 1-.237-2.792Zm5.614-.454V9.32h2.468c.128 0 .9.148.9.725 0 .48-.592.652-1.08.652zm8.969 1.24q0 .273-.02.542h-.75c-.076 0-.106.05-.106.123v.345c0 .81-.457.987-.858 1.032-.381.043-.805-.16-.857-.394-.225-1.265-.6-1.536-1.192-2.003.735-.467 1.5-1.155 1.5-2.077 0-.996-.683-1.623-1.148-1.93-.652-.43-1.375-.517-1.57-.517h-7.76a7.3 7.3 0 0 1 4.091-2.308l.915.959a.54.54 0 0 0 .765.018l1.024-.98a7.32 7.32 0 0 1 5.006 3.566l-.7 1.583a.543.543 0 0 0 .275.715l1.35.6q.034.357.035.726m-7.754-8.004a.595.595 0 1 1 .82.86.595.595 0 0 1-.82-.86m6.951 5.595a.594.594 0 1 1 1.086.481.594.594 0 0 1-1.086-.481"},null,-1)]))}const hd={render:pd},fd={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function md(e,t){return(0,$t.uX)(),(0,$t.CE)("svg",fd,t[0]||(t[0]=[(0,$t.Lk)("path",{fill:"#41B883",d:"m14.31 4.34-2.31 4-2.31-4H2l10 17.32L22 4.34z"},null,-1),(0,$t.Lk)("path",{fill:"#34495E",d:"m14.31 4.34-2.31 4-2.31-4H6l6 10.392L18 4.34z"},null,-1)]))}const gd={render:md},vd=Object.assign({"../icons/Add.svg":r,"../icons/AddTab.svg":a,"../icons/Alert.svg":o,"../icons/ArrowLeft.svg":i,"../icons/ArrowRight.svg":s,"../icons/Brackets.svg":l,"../icons/Branch.svg":c,"../icons/Changelog.svg":u,"../icons/Checkmark.svg":d,"../icons/ChevronDown.svg":p,"../icons/ChevronLeft.svg":h,"../icons/ChevronRight.svg":f,"../icons/ChevronUp.svg":m,"../icons/Clipboard.svg":g,"../icons/Close.svg":v,"../icons/CloseTab.svg":b,"../icons/CloseTabs.svg":y,"../icons/Cloud.svg":w,"../icons/CodeFolder.svg":O,"../icons/Cog.svg":k,"../icons/Collection.svg":x,"../icons/Cookie.svg":E,"../icons/DarkMode.svg":S,"../icons/Delete.svg":_,"../icons/Discord.svg":C,"../icons/DiscordLine.svg":T,"../icons/DocsPage.svg":A,"../icons/Download.svg":R,"../icons/Duplicate.svg":P,"../icons/Edit.svg":D,"../icons/Ellipses.svg":M,"../icons/Email.svg":$,"../icons/Error.svg":N,"../icons/Example.svg":I,"../icons/ExternalLink.svg":L,"../icons/FilterList.svg":B,"../icons/Folder.svg":F,"../icons/GitHub.svg":Q,"../icons/GitHubLine.svg":U,"../icons/Globe.svg":z,"../icons/Google.svg":j,"../icons/Help.svg":W,"../icons/Hide.svg":X,"../icons/History.svg":q,"../icons/House.svg":V,"../icons/Import.svg":H,"../icons/Info.svg":Z,"../icons/Key.svg":G,"../icons/Leave.svg":K,"../icons/LightDarkModeToggle.svg":Y,"../icons/LightMode.svg":J,"../icons/Link.svg":ee,"../icons/Lock.svg":te,"../icons/Logo.svg":ne,"../icons/LogoAPI.svg":re,"../icons/LogoClient.svg":ae,"../icons/LogoMarket.svg":oe,"../icons/LogoSwagger.svg":ie,"../icons/Magic.svg":se,"../icons/Menu.svg":le,"../icons/Page.svg":ce,"../icons/PaperAirplane.svg":ue,"../icons/Pencil.svg":de,"../icons/Play.svg":pe,"../icons/Refresh.svg":he,"../icons/Roadmap.svg":fe,"../icons/Scribble.svg":me,"../icons/Search.svg":ge,"../icons/Server.svg":ve,"../icons/Settings.svg":be,"../icons/Show.svg":ye,"../icons/Sort.svg":we,"../icons/Terminal.svg":Oe,"../icons/ThumbsDown.svg":ke,"../icons/ThumbsUp.svg":xe,"../icons/Trash.svg":Ee,"../icons/Unlock.svg":Se,"../icons/Unwatch.svg":_e,"../icons/Upload.svg":Ce,"../icons/UploadSimple.svg":Te,"../icons/User.svg":Ae,"../icons/UserSwitch.svg":Re,"../icons/Users.svg":Pe,"../icons/Wallet.svg":De,"../icons/Watch.svg":Me,"../icons/Workspace.svg":$e,"../icons/programming-framework-angular.svg":Ne,"../icons/programming-framework-astro.svg":Ie,"../icons/programming-framework-laravel.svg":Le,"../icons/programming-framework-nextdotjs.svg":Be,"../icons/programming-framework-react.svg":Fe,"../icons/programming-framework-vuedotjs.svg":Qe,"../icons/programming-language-c.svg":Ue,"../icons/programming-language-clojure.svg":ze,"../icons/programming-language-csharp.svg":je,"../icons/programming-language-css3.svg":We,"../icons/programming-language-dart.svg":Xe,"../icons/programming-language-go.svg":qe,"../icons/programming-language-html5.svg":Ve,"../icons/programming-language-http.svg":He,"../icons/programming-language-java.svg":Ze,"../icons/programming-language-javascript.svg":Ge,"../icons/programming-language-json.svg":Ke,"../icons/programming-language-kotlin.svg":Ye,"../icons/programming-language-node.svg":Je,"../icons/programming-language-objc.svg":et,"../icons/programming-language-ocaml.svg":tt,"../icons/programming-language-php.svg":nt,"../icons/programming-language-powershell.svg":rt,"../icons/programming-language-python.svg":at,"../icons/programming-language-r.svg":ot,"../icons/programming-language-ruby.svg":it,"../icons/programming-language-scala.svg":st,"../icons/programming-language-shell.svg":lt,"../icons/programming-language-swift.svg":ct,"../icons/programming-language-typescript.svg":ut,"../icons/programming-script-code.svg":dt,"../icons/programming-tool-git.svg":pt,"../icons/programming-tool-tailwindcss.svg":ht}),bd=Object.assign({"../logos/Adonisjs.svg":ft,"../logos/Docusaurus.svg":mt,"../logos/Dotnet.svg":gt,"../logos/Elysiajs.svg":vt,"../logos/Express.svg":bt,"../logos/Fastapi.svg":yt,"../logos/Fastify.svg":wt,"../logos/Go.svg":Ot,"../logos/Hono.svg":kt,"../logos/Laravel.svg":xt,"../logos/Litestar.svg":Et,"../logos/Nestjs.svg":St,"../logos/Nextjs.svg":_t,"../logos/Nitro.svg":Ct,"../logos/Nuxt.svg":Tt,"../logos/Openapi.svg":At,"../logos/Platformatic.svg":Rt,"../logos/React.svg":Pt,"../logos/Rust.svg":Dt,"../logos/Vue.svg":Mt}),yd=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarIcon",props:{icon:{},logo:{},size:{},thickness:{},label:{}},setup(e){(0,Rr.$9)((e=>({ce5c5c8a:a.value})));const t=e,{cx:n}=zn(),r=Fn({variants:{size:{xs:"size-3",sm:"size-3.5",md:"size-4",lg:"size-5",xl:"size-6","2xl":"size-8","3xl":"size-10",full:"size-full"}},defaultVariants:{size:"full"}}),a=(0,$t.EW)((()=>t.thickness??"2")),o=(0,$t.EW)((()=>t.label?{ariaLabel:t.label}:{ariaHidden:!0,role:"presentation"})),i=(0,$t.EW)((()=>t.icon?(e=>{const t=`../icons/${e}.svg`;return vd[t]?vd[t]:(console.warn(`Could not find icon: ${e}`),null)})(t.icon):t.logo?(e=>{const t=`../logos/${e}.svg`;return bd[t]?bd[t]:(console.warn(`Could not find icon: ${e}`),null)})(t.logo):void 0));return(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,$t.$y)(i.value),(0,Wn._B)((0,$t.Ng)({...(0,jn.R1)(n)("scalar-icon",(0,jn.R1)(r)({size:e.size})),...o.value})),null,16))}}),wd=Gn(yd,[["__scopeId","data-v-f4f53d32"]]),Od={class:"scalar-code-block custom-scroll min-h-12"},kd={key:0,class:"scalar-code-copy"},xd=["innerHTML"],Ed=(0,$t.pM)({__name:"ScalarCodeBlock",props:{content:{},lang:{default:"plaintext"},lineNumbers:{type:Boolean,default:!1},hideCredentials:{},copy:{type:Boolean,default:!0}},setup(e){const t=e,n=(0,$t.EW)((()=>function(e,t){const n=("string"==typeof t?.maskCredentials?[t.maskCredentials]:t?.maskCredentials??[]).filter((e=>!(e.length<3))),r=`language-${vr.I[t.lang]??t.lang}`;var a;const o=(0,mr.l)().use(hr,{fragment:!0}).use((a=e,()=>e=>{(0,gr.YR)(e,"element",(e=>{"code"===e.tagName&&e.children.push({type:"text",value:a})}))})).use(Sr.u,{languages:t.languages}).use(t?.lineNumbers?Or:()=>{}).use(fr.A).processSync(`<pre><code class="${r}"></code></pre>`).toString();return n.length?n.reduce(((e,t)=>e.split(t).join(`<span class="credentials">${t}</span>`)),o):o}(Ar(t.content),{lang:t.lang.trim(),languages:_r.n,lineNumbers:t.lineNumbers,maskCredentials:t.hideCredentials}).slice(5,-6))),{copyToClipboard:r}=(0,Cr.i)(),a=(0,$t.EW)((()=>null!==t.content&&"null"!==t.content&&"404 Not Found"!==t.content));return(e,o)=>((0,$t.uX)(),(0,$t.CE)("div",Od,[e.copy?((0,$t.uX)(),(0,$t.CE)("div",kd,[a.value?((0,$t.uX)(),(0,$t.CE)("button",{key:0,class:"copy-button",type:"button",onClick:o[0]||(o[0]=e=>(0,jn.R1)(r)((0,jn.R1)(Ar)(t.content)))},[o[1]||(o[1]=(0,$t.Lk)("span",{class:"sr-only"},"Copy content",-1)),(0,$t.bF)((0,jn.R1)(wd),{icon:"Clipboard",size:"md"})])):(0,$t.Q3)("",!0)])):(0,$t.Q3)("",!0),(0,$t.Lk)("pre",{class:"scalar-codeblock-pre",innerHTML:n.value},null,8,xd)]))}});var Sd=n(5841);const _d=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarColorModeToggleIcon",props:{is:{default:"button"},mode:{default:"light"}},setup(e){const{cx:t}=zn();return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,$t.$y)(e.is),(0,$t.v6)({class:`toggle-icon-${e.mode}`,type:"button"===e.is?"button":void 0},(0,jn.R1)(t)("size-3 flex items-center justify-center text-c-1")),{default:(0,$t.k6)((()=>n[0]||(n[0]=[(0,$t.Lk)("span",{class:"toggle-icon-sun-ray"},null,-1),(0,$t.Lk)("span",{class:"toggle-icon-sun-ray"},null,-1),(0,$t.Lk)("span",{class:"toggle-icon-sun-ray"},null,-1),(0,$t.Lk)("span",{class:"toggle-icon-sun-ray"},null,-1),(0,$t.Lk)("span",{class:"toggle-icon-ellipse"},[(0,$t.Lk)("span",{class:"toggle-icon-moon-mask"})],-1)]))),_:1},16,["class","type"]))}}),Cd=Gn(_d,[["__scopeId","data-v-8bd002b1"]]),Td=["aria-pressed"],Ad=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarColorModeToggleButton",props:{modelValue:{type:Boolean},modelModifiers:{}},emits:["update:modelValue"],setup(e){const{cx:t}=zn(),n=(0,$t.fn)(e,"modelValue");return(e,r)=>((0,$t.uX)(),(0,$t.CE)("button",(0,$t.v6)({"aria-pressed":n.value,type:"button"},(0,jn.R1)(t)("group/toggle flex h-6 w-[38px] brightness-lifted -mx-px items-center py-1.5 -my-1.5 relative outline-none"),{onClick:r[0]||(r[0]=e=>n.value=!n.value)}),[r[1]||(r[1]=(0,$t.Lk)("div",{class:"h-3 w-full bg-border mx-px rounded-xl group-focus-visible/toggle:outline -outline-offset-1"},null,-1)),(0,$t.Lk)("div",{class:(0,Wn.C4)(["size-[23px] left-border absolute border rounded-full flex items-center justify-center bg-b-1 group-focus-visible/toggle:outline -outline-offset-1 transition-transform duration-300 ease-in-out",{"translate-x-[14px]":n.value}])},[(0,$t.bF)(Cd,{is:"div",mode:n.value?"dark":"light"},null,8,["mode"])],2)],16,Td))}}),Rd=(0,$t.pM)({__name:"ScalarColorModeToggle",props:{variant:{default:"switch"}},setup(e){const{isDarkMode:t,toggleColorMode:n,darkLightMode:r}=(0,Sd.G)(),a=(0,$t.EW)((()=>t.value?"Set light mode":"Set dark mode"));return(e,o)=>"switch"===e.variant?((0,$t.uX)(),(0,$t.Wv)(Ad,{key:0,modelValue:(0,jn.R1)(t),"onUpdate:modelValue":o[0]||(o[0]=e=>(0,jn.i9)(t)?t.value=e:null),"aria-label":a.value},null,8,["modelValue","aria-label"])):((0,$t.uX)(),(0,$t.Wv)(Cd,{key:1,"aria-label":a.value,mode:(0,jn.R1)(r),onClick:(0,jn.R1)(n)},null,8,["aria-label","mode","onClick"]))}});let Pd=(e=21)=>{let t="",n=crypto.getRandomValues(new Uint8Array(e|=0));for(;e--;)t+="useandom-26T198340PX75pxJACKVERYMINDBUSHWOLF_GQZbfghjklqvwyzrict"[63&n[e]];return t};const Dd=(0,$t.pM)({__name:"ScalarListboxCheckbox",props:{selected:{type:Boolean},style:{}},setup:e=>(e,t)=>((0,$t.uX)(),(0,$t.CE)("div",{class:(0,Wn.C4)(["flex size-4 items-center justify-center p-0.75",[e.selected?"bg-c-accent text-b-1":"text-transparent shadow-border","checkbox"===e.style?"rounded":"rounded-full"]])},[e.selected?((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(wd),{key:0,class:"relative top-[0.5px] size-3",icon:"Checkmark",thickness:"3"})):(0,$t.Q3)("",!0)],2))}),Md={class:"inline-block min-w-0 flex-1 truncate text-c-1"},$d=(0,$t.pM)({__name:"ScalarComboboxOption",props:{active:{type:Boolean},selected:{type:Boolean},style:{},isDeletable:{type:Boolean}},emits:["delete"],setup(e){const t=Fn({base:["group/item","flex min-w-0 items-center gap-1.5 rounded px-2 py-1.5 text-left","truncate bg-transparent text-c-1","cursor-pointer hover:bg-b-2"],variants:{selected:{true:"text-c-1"},active:{true:"bg-b-2"}}});return(e,n)=>((0,$t.uX)(),(0,$t.CE)("li",{class:(0,Wn.C4)((0,jn.R1)(Qn)((0,jn.R1)(t)({active:e.active,selected:e.selected})))},[(0,$t.bF)((0,jn.R1)(Dd),{selected:e.selected,style:(0,Wn.Tr)(e.style)},null,8,["selected","style"]),(0,$t.Lk)("span",Md,[(0,$t.RG)(e.$slots,"default")]),e.isDeletable?((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(wd),{key:0,class:"text-c-2 opacity-0 group-hover/item:opacity-100",icon:"Delete",size:"md",thickness:"1.5",onClick:n[0]||(n[0]=(0,Rr.D$)((t=>e.$emit("delete")),["stop"]))})):(0,$t.Q3)("",!0)],2))}});function Nd(e){return function(e){return void 0!==e.options}(e[0])}const Id={class:"relative flex"},Ld=["placeholder"],Bd={key:0,class:"min-w-0 truncate px-2.5 py-1.5 text-left text-c-2"},Fd=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarComboboxOptions",props:{options:{},modelValue:{},placeholder:{},multiselect:{type:Boolean},isDeletable:{type:Boolean}},emits:["update:modelValue","delete"],setup(e,{emit:t}){var n;const r=e,a=t,o=`scalar-combobox-items-${Pd()}`;function i(e){return`${o}-${e.id}`}const s=(0,$t.EW)((()=>Nd(r.options)?r.options.flatMap((e=>e.options)):r.options)),l=(0,$t.EW)((()=>Nd(r.options)?r.options:[{label:"",options:r.options}])),c=(0,jn.KR)(""),u=(0,jn.KR)((null==(n=r.modelValue)?void 0:n[0])??s.value[0]);(0,$t.sV)((async()=>{var e;c.value="",u.value=(null==(e=r.modelValue)?void 0:e[0])??s.value[0],0!==p.value.length&&setTimeout((()=>{var e;null==(e=null==document?void 0:document.getElementById(i(p.value[0])))||e.scrollIntoView({block:"center"})}),10)})),(0,$t.wB)((()=>c.value),(()=>u.value=d.value[0]));const d=(0,$t.EW)((()=>""===c.value?s.value:s.value.filter((e=>e.label.toLowerCase().includes(c.value.toLowerCase()))))),p=(0,$t.EW)({get:()=>r.modelValue??[],set:e=>e&&a("update:modelValue",e)});function h(e){r.multiselect?p.value.some((t=>t.id===e.id))?p.value=p.value.filter((t=>t.id!==e.id)):p.value=[...p.value,e]:p.value=[e]}function f(e){var t;const n=d.value,r=n.findIndex((e=>e.id===u.value.id))+e;r<0||r>n.length-1||(u.value=n[r],null==(t=null==document?void 0:document.getElementById(i(u.value)))||t.scrollIntoView({behavior:"smooth",block:"nearest"}))}return(e,t)=>((0,$t.uX)(),(0,$t.CE)($t.FK,null,[(0,$t.Lk)("div",Id,[(0,$t.bF)((0,jn.R1)(wd),{class:"pointer-events-none absolute left-2.5 top-1/2 -translate-y-1/2 text-c-3",icon:"Search",size:"md"}),(0,$t.bo)((0,$t.Lk)("input",{"onUpdate:modelValue":t[0]||(t[0]=e=>c.value=e),"aria-autocomplete":"list","aria-controls":o,class:"min-w-0 flex-1 rounded-none border-0 py-2.5 pl-8 pr-3 leading-none text-c-1 outline-none","data-1p-ignore":"",placeholder:e.placeholder,role:"combobox",tabindex:"0",type:"text",onKeydown:[t[1]||(t[1]=(0,Rr.jR)((0,Rr.D$)((e=>f(1)),["prevent"]),["down"])),t[2]||(t[2]=(0,Rr.jR)((0,Rr.D$)((e=>h(u.value)),["prevent"]),["enter"])),t[3]||(t[3]=(0,Rr.jR)((0,Rr.D$)((e=>f(-1)),["prevent"]),["up"]))]},null,40,Ld),[[Rr.Jo,c.value]])]),(0,$t.bo)((0,$t.Lk)("ul",{id:o,class:"border-t p-0.75 custom-scroll overscroll-contain flex-1 min-h-0"},[(0,$t.RG)(e.$slots,"before"),((0,$t.uX)(!0),(0,$t.CE)($t.FK,null,(0,$t.pI)(l.value,((n,r)=>((0,$t.uX)(),(0,$t.CE)($t.FK,{key:r},[n.label&&n.options.some((e=>d.value.some((t=>t.id===e.id))))?((0,$t.uX)(),(0,$t.CE)("div",Bd,(0,Wn.v_)(n.label),1)):(0,$t.Q3)("",!0),((0,$t.uX)(!0),(0,$t.CE)($t.FK,null,(0,$t.pI)(d.value,(r=>{var a;return(0,$t.uX)(),(0,$t.CE)($t.FK,{key:r.id},[n.options.some((e=>e.id===r.id))?((0,$t.uX)(),(0,$t.Wv)($d,{key:0,id:i(r),active:(null==(a=u.value)?void 0:a.id)===r.id,isDeletable:r.isDeletable??e.isDeletable,selected:p.value.some((e=>e.id===r.id)),style:(0,Wn.Tr)(e.multiselect?"checkbox":"radio"),onClick:e=>h(r),onDelete:t=>e.$emit("delete",r),onMousedown:t[4]||(t[4]=(0,Rr.D$)((()=>{}),["prevent"])),onMouseenter:e=>u.value=r},{default:(0,$t.k6)((()=>[(0,$t.eW)((0,Wn.v_)(r.label),1)])),_:2},1032,["id","active","isDeletable","selected","style","onClick","onDelete","onMouseenter"])):(0,$t.Q3)("",!0)],64)})),128))],64)))),128)),(0,$t.RG)(e.$slots,"after")],512),[[Rr.aG,d.value.length||e.$slots.before||e.$slots.after]])],64))}});var Qd=n(3605);let Ud=Symbol("ForcePortalRootContext"),zd=(0,$t.pM)({name:"ForcePortalRoot",props:{as:{type:[Object,String],default:"template"},force:{type:Boolean,default:!1}},setup:(e,{slots:t,attrs:n})=>((0,$t.Gt)(Ud,e.force),()=>{let{force:r,...a}=e;return(0,Qd.XX)({theirProps:a,ourProps:{},slot:{},slots:t,attrs:n,name:"ForcePortalRoot"})})});var jd=n(6413),Wd=n(562);const Xd=new WeakMap;function qd(e,t){let n=t(function(e){var t;return null!=(t=Xd.get(e))?t:0}(e));return n<=0?Xd.delete(e):Xd.set(e,n),n}let Vd=(0,$t.pM)({name:"Portal",props:{as:{type:[Object,String],default:"div"}},setup(e,{slots:t,attrs:n}){let r=(0,jn.KR)(null),a=(0,$t.EW)((()=>(0,Wd.T)(r))),o=(0,$t.WQ)(Ud,!1),i=(0,$t.WQ)(Gd,null),s=(0,jn.KR)(!0===o||null==i?function(e){let t=(0,Wd.T)(e);if(!t){if(null===e)return null;throw new Error(`[Headless UI]: Cannot find ownerDocument for contextElement: ${e}`)}let n=t.getElementById("headlessui-portal-root");if(n)return n;let r=t.createElement("div");return r.setAttribute("id","headlessui-portal-root"),t.body.appendChild(r)}(r.value):i.resolveTarget());s.value&&qd(s.value,(e=>e+1));let l=(0,jn.KR)(!1);(0,$t.sV)((()=>{l.value=!0})),(0,$t.nT)((()=>{o||null!=i&&(s.value=i.resolveTarget())}));let c=(0,$t.WQ)(Hd,null),u=!1,d=(0,$t.nI)();return(0,$t.wB)(r,(()=>{if(u||!c)return;let e=(0,jd.t)(r);e&&((0,$t.hi)(c.register(e),d),u=!0)})),(0,$t.hi)((()=>{var e,t;let n=null==(e=a.value)?void 0:e.getElementById("headlessui-portal-root");!n||s.value!==n||qd(s.value,(e=>e-1))||s.value.children.length>0||null==(t=s.value.parentElement)||t.removeChild(s.value)})),()=>{if(!l.value||null===s.value)return null;let a={ref:r,"data-headlessui-portal":""};return(0,$t.h)($t.Im,{to:s.value},(0,Qd.XX)({ourProps:a,theirProps:e,slot:{},attrs:n,slots:t,name:"Portal"}))}}}),Hd=Symbol("PortalParentContext");function Zd(){let e=(0,$t.WQ)(Hd,null),t=(0,jn.KR)([]);function n(n){let r=t.value.indexOf(n);-1!==r&&t.value.splice(r,1),e&&e.unregister(n)}let r={register:function(r){return t.value.push(r),e&&e.register(r),()=>n(r)},unregister:n,portals:t};return[t,(0,$t.pM)({name:"PortalWrapper",setup:(e,{slots:t})=>((0,$t.Gt)(Hd,r),()=>{var e;return null==(e=t.default)?void 0:e.call(t)})})]}let Gd=Symbol("PortalGroupContext"),Kd=(0,$t.pM)({name:"PortalGroup",props:{as:{type:[Object,String],default:"template"},target:{type:Object,default:null}},setup(e,{attrs:t,slots:n}){let r=(0,jn.Kh)({resolveTarget:()=>e.target});return(0,$t.Gt)(Gd,r),()=>{let{target:r,...a}=e;return(0,Qd.XX)({theirProps:a,ourProps:{},slot:{},attrs:t,slots:n,name:"PortalGroup"})}}});var Yd=n(9146);function Jd(e,t,n,r){Yd._.isServer||(0,$t.nT)((a=>{(e=null!=e?e:window).addEventListener(t,n,r),a((()=>e.removeEventListener(t,n,r)))}))}var ep=n(8869),tp=n(3648),np=n(1972),rp=n(805);function ap({defaultContainers:e=[],portals:t,mainTreeNodeRef:n}={}){let r=(0,jn.KR)(null),a=(0,Wd.T)(r);function o(){var n,o,i;let s=[];for(let t of e)null!==t&&(t instanceof HTMLElement?s.push(t):"value"in t&&t.value instanceof HTMLElement&&s.push(t.value));if(null!=t&&t.value)for(let e of t.value)s.push(e);for(let e of null!=(n=null==a?void 0:a.querySelectorAll("html > *, body > *"))?n:[])e!==document.body&&e!==document.head&&e instanceof HTMLElement&&"headlessui-portal-root"!==e.id&&(e.contains((0,jd.t)(r))||e.contains(null==(i=null==(o=(0,jd.t)(r))?void 0:o.getRootNode())?void 0:i.host)||s.some((t=>e.contains(t)))||s.push(e));return s}return{resolveContainers:o,contains:e=>o().some((t=>t.contains(e))),mainTreeNodeRef:r,MainTreeNode:()=>null!=n?null:(0,$t.h)(rp.j,{features:rp.O.Hidden,ref:r})}}var op=n(2389),ip=(e=>(e[e.Forwards=0]="Forwards",e[e.Backwards=1]="Backwards",e))(ip||{});function sp(){let e=(0,jn.KR)(0);return(0,op.M)("keydown",(t=>{"Tab"===t.key&&(e.value=t.shiftKey?1:0)})),e}var lp=n(7014),cp=n(7838),up=n(1417),dp=n(9388),pp=(e=>(e[e.Open=0]="Open",e[e.Closed=1]="Closed",e))(pp||{});let hp=Symbol("PopoverContext");function fp(e){let t=(0,$t.WQ)(hp,null);if(null===t){let t=new Error(`<${e} /> is missing a parent <${bp.name} /> component.`);throw Error.captureStackTrace&&Error.captureStackTrace(t,fp),t}return t}let mp=Symbol("PopoverGroupContext");function gp(){return(0,$t.WQ)(mp,null)}let vp=Symbol("PopoverPanelContext"),bp=(0,$t.pM)({name:"Popover",inheritAttrs:!1,props:{as:{type:[Object,String],default:"div"}},setup(e,{slots:t,attrs:n,expose:r}){var a;let o=(0,jn.KR)(null);r({el:o,$el:o});let i=(0,jn.KR)(1),s=(0,jn.KR)(null),l=(0,jn.KR)(null),c=(0,jn.KR)(null),u=(0,jn.KR)(null),d=(0,$t.EW)((()=>(0,Wd.T)(o))),p=(0,$t.EW)((()=>{var e,t;if(!(0,jd.t)(s)||!(0,jd.t)(u))return!1;for(let e of document.querySelectorAll("body > *"))if(Number(null==e?void 0:e.contains((0,jd.t)(s)))^Number(null==e?void 0:e.contains((0,jd.t)(u))))return!0;let n=(0,up.iq)(),r=n.indexOf((0,jd.t)(s)),a=(r+n.length-1)%n.length,o=(r+1)%n.length,i=n[a],l=n[o];return!(null!=(e=(0,jd.t)(u))&&e.contains(i)||null!=(t=(0,jd.t)(u))&&t.contains(l))})),h={popoverState:i,buttonId:(0,jn.KR)(null),panelId:(0,jn.KR)(null),panel:u,button:s,isPortalled:p,beforePanelSentinel:l,afterPanelSentinel:c,togglePopover(){i.value=(0,dp.Y)(i.value,{0:1,1:0})},closePopover(){1!==i.value&&(i.value=1)},close(e){h.closePopover();let t=e?e instanceof HTMLElement?e:e.value instanceof HTMLElement?(0,jd.t)(e):(0,jd.t)(h.button):(0,jd.t)(h.button);null==t||t.focus()}};(0,$t.Gt)(hp,h),(0,lp.Fo)((0,$t.EW)((()=>(0,dp.Y)(i.value,{0:lp.Uw.Open,1:lp.Uw.Closed}))));let f={buttonId:h.buttonId,panelId:h.panelId,close(){h.closePopover()}},m=gp(),g=null==m?void 0:m.registerPopover,[v,b]=Zd(),y=ap({mainTreeNodeRef:null==m?void 0:m.mainTreeNodeRef,portals:v,defaultContainers:[s,u]});return(0,$t.nT)((()=>null==g?void 0:g(f))),Jd(null==(a=d.value)?void 0:a.defaultView,"focus",(e=>{var t,n;e.target!==window&&e.target instanceof HTMLElement&&0===i.value&&(function(){var e,t,n,r;return null!=(r=null==m?void 0:m.isFocusWithinPopoverGroup())?r:(null==(e=d.value)?void 0:e.activeElement)&&((null==(t=(0,jd.t)(s))?void 0:t.contains(d.value.activeElement))||(null==(n=(0,jd.t)(u))?void 0:n.contains(d.value.activeElement)))}()||s&&u&&(y.contains(e.target)||null!=(t=(0,jd.t)(h.beforePanelSentinel))&&t.contains(e.target)||null!=(n=(0,jd.t)(h.afterPanelSentinel))&&n.contains(e.target)||h.closePopover()))}),!0),(0,tp.j)(y.resolveContainers,((e,t)=>{var n;h.closePopover(),(0,up.Bm)(t,up.MZ.Loose)||(e.preventDefault(),null==(n=(0,jd.t)(s))||n.focus())}),(0,$t.EW)((()=>0===i.value))),()=>{let r={open:0===i.value,close:h.close};return(0,$t.h)($t.FK,[(0,$t.h)(b,{},(()=>(0,Qd.XX)({theirProps:{...e,...n},ourProps:{ref:o},slot:r,slots:t,attrs:n,name:"Popover"}))),(0,$t.h)(y.MainTreeNode)])}}}),yp=(0,$t.pM)({name:"PopoverButton",props:{as:{type:[Object,String],default:"button"},disabled:{type:[Boolean],default:!1},id:{type:String,default:null}},inheritAttrs:!1,setup(e,{attrs:t,slots:n,expose:r}){var a;let o=null!=(a=e.id)?a:`headlessui-popover-button-${(0,ep.B)()}`,i=fp("PopoverButton"),s=(0,$t.EW)((()=>(0,Wd.T)(i.button)));r({el:i.button,$el:i.button}),(0,$t.sV)((()=>{i.buttonId.value=o})),(0,$t.hi)((()=>{i.buttonId.value=null}));let l=gp(),c=null==l?void 0:l.closeOthers,u=(0,$t.WQ)(vp,null),d=(0,$t.EW)((()=>null!==u&&u.value===i.panelId.value)),p=(0,jn.KR)(null),h=`headlessui-focus-sentinel-${(0,ep.B)()}`;d.value||(0,$t.nT)((()=>{i.button.value=(0,jd.t)(p)}));let f=(0,np.c)((0,$t.EW)((()=>({as:e.as,type:t.type}))),p);function m(e){var t,n,r,a,o;if(d.value){if(1===i.popoverState.value)return;switch(e.key){case cp.D.Space:case cp.D.Enter:e.preventDefault(),null==(n=(t=e.target).click)||n.call(t),i.closePopover(),null==(r=(0,jd.t)(i.button))||r.focus()}}else switch(e.key){case cp.D.Space:case cp.D.Enter:e.preventDefault(),e.stopPropagation(),1===i.popoverState.value&&(null==c||c(i.buttonId.value)),i.togglePopover();break;case cp.D.Escape:if(0!==i.popoverState.value)return null==c?void 0:c(i.buttonId.value);if(!(0,jd.t)(i.button)||null!=(a=s.value)&&a.activeElement&&(null==(o=(0,jd.t)(i.button))||!o.contains(s.value.activeElement)))return;e.preventDefault(),e.stopPropagation(),i.closePopover()}}function g(e){d.value||e.key===cp.D.Space&&e.preventDefault()}function v(t){var n,r;e.disabled||(d.value?(i.closePopover(),null==(n=(0,jd.t)(i.button))||n.focus()):(t.preventDefault(),t.stopPropagation(),1===i.popoverState.value&&(null==c||c(i.buttonId.value)),i.togglePopover(),null==(r=(0,jd.t)(i.button))||r.focus()))}function b(e){e.preventDefault(),e.stopPropagation()}let y=sp();function w(){let e=(0,jd.t)(i.panel);e&&(0,dp.Y)(y.value,{[ip.Forwards]:()=>(0,up.CU)(e,up.BD.First),[ip.Backwards]:()=>(0,up.CU)(e,up.BD.Last)})===up.Me.Error&&(0,up.CU)((0,up.iq)().filter((e=>"true"!==e.dataset.headlessuiFocusGuard)),(0,dp.Y)(y.value,{[ip.Forwards]:up.BD.Next,[ip.Backwards]:up.BD.Previous}),{relativeTo:(0,jd.t)(i.button)})}return()=>{let r=0===i.popoverState.value,a={open:r},{...s}=e,l=d.value?{ref:p,type:f.value,onKeydown:m,onClick:v}:{ref:p,id:o,type:f.value,"aria-expanded":0===i.popoverState.value,"aria-controls":(0,jd.t)(i.panel)?i.panelId.value:void 0,disabled:!!e.disabled||void 0,onKeydown:m,onKeyup:g,onClick:v,onMousedown:b};return(0,$t.h)($t.FK,[(0,Qd.XX)({ourProps:l,theirProps:{...t,...s},slot:a,attrs:t,slots:n,name:"PopoverButton"}),r&&!d.value&&i.isPortalled.value&&(0,$t.h)(rp.j,{id:h,features:rp.O.Focusable,"data-headlessui-focus-guard":!0,as:"button",type:"button",onFocus:w})])}}}),wp=((0,$t.pM)({name:"PopoverOverlay",props:{as:{type:[Object,String],default:"div"},static:{type:Boolean,default:!1},unmount:{type:Boolean,default:!0}},setup(e,{attrs:t,slots:n}){let r=fp("PopoverOverlay"),a=`headlessui-popover-overlay-${(0,ep.B)()}`,o=(0,lp.O_)(),i=(0,$t.EW)((()=>null!==o?(o.value&lp.Uw.Open)===lp.Uw.Open:0===r.popoverState.value));function s(){r.closePopover()}return()=>{let o={open:0===r.popoverState.value};return(0,Qd.XX)({ourProps:{id:a,"aria-hidden":!0,onClick:s},theirProps:e,slot:o,attrs:t,slots:n,features:Qd.O5.RenderStrategy|Qd.O5.Static,visible:i.value,name:"PopoverOverlay"})}}}),(0,$t.pM)({name:"PopoverPanel",props:{as:{type:[Object,String],default:"div"},static:{type:Boolean,default:!1},unmount:{type:Boolean,default:!0},focus:{type:Boolean,default:!1},id:{type:String,default:null}},inheritAttrs:!1,setup(e,{attrs:t,slots:n,expose:r}){var a;let o=null!=(a=e.id)?a:`headlessui-popover-panel-${(0,ep.B)()}`,{focus:i}=e,s=fp("PopoverPanel"),l=(0,$t.EW)((()=>(0,Wd.T)(s.panel))),c=`headlessui-focus-sentinel-before-${(0,ep.B)()}`,u=`headlessui-focus-sentinel-after-${(0,ep.B)()}`;r({el:s.panel,$el:s.panel}),(0,$t.sV)((()=>{s.panelId.value=o})),(0,$t.hi)((()=>{s.panelId.value=null})),(0,$t.Gt)(vp,s.panelId),(0,$t.nT)((()=>{var e,t;if(!i||0!==s.popoverState.value||!s.panel)return;let n=null==(e=l.value)?void 0:e.activeElement;null!=(t=(0,jd.t)(s.panel))&&t.contains(n)||(0,up.CU)((0,jd.t)(s.panel),up.BD.First)}));let d=(0,lp.O_)(),p=(0,$t.EW)((()=>null!==d?(d.value&lp.Uw.Open)===lp.Uw.Open:0===s.popoverState.value));function h(e){var t,n;if(e.key===cp.D.Escape){if(0!==s.popoverState.value||!(0,jd.t)(s.panel)||l.value&&(null==(t=(0,jd.t)(s.panel))||!t.contains(l.value.activeElement)))return;e.preventDefault(),e.stopPropagation(),s.closePopover(),null==(n=(0,jd.t)(s.button))||n.focus()}}function f(e){var t,n,r,a,o;let i=e.relatedTarget;i&&(0,jd.t)(s.panel)&&(null!=(t=(0,jd.t)(s.panel))&&t.contains(i)||(s.closePopover(),(null!=(r=null==(n=(0,jd.t)(s.beforePanelSentinel))?void 0:n.contains)&&r.call(n,i)||null!=(o=null==(a=(0,jd.t)(s.afterPanelSentinel))?void 0:a.contains)&&o.call(a,i))&&i.focus({preventScroll:!0})))}let m=sp();function g(){let e=(0,jd.t)(s.panel);e&&(0,dp.Y)(m.value,{[ip.Forwards]:()=>{var t;(0,up.CU)(e,up.BD.First)===up.Me.Error&&(null==(t=(0,jd.t)(s.afterPanelSentinel))||t.focus())},[ip.Backwards]:()=>{var e;null==(e=(0,jd.t)(s.button))||e.focus({preventScroll:!0})}})}function v(){let e=(0,jd.t)(s.panel);e&&(0,dp.Y)(m.value,{[ip.Forwards]:()=>{let e=(0,jd.t)(s.button),t=(0,jd.t)(s.panel);if(!e)return;let n=(0,up.iq)(),r=n.indexOf(e),a=n.slice(0,r+1),o=[...n.slice(r+1),...a];for(let e of o.slice())if("true"===e.dataset.headlessuiFocusGuard||null!=t&&t.contains(e)){let t=o.indexOf(e);-1!==t&&o.splice(t,1)}(0,up.CU)(o,up.BD.First,{sorted:!1})},[ip.Backwards]:()=>{var t;(0,up.CU)(e,up.BD.Previous)===up.Me.Error&&(null==(t=(0,jd.t)(s.button))||t.focus())}})}return()=>{let r={open:0===s.popoverState.value,close:s.close},{focus:a,...l}=e,d={ref:s.panel,id:o,onKeydown:h,onFocusout:i&&0===s.popoverState.value?f:void 0,tabIndex:-1};return(0,Qd.XX)({ourProps:d,theirProps:{...t,...l},attrs:t,slot:r,slots:{...n,default:(...e)=>{var t;return[(0,$t.h)($t.FK,[p.value&&s.isPortalled.value&&(0,$t.h)(rp.j,{id:c,ref:s.beforePanelSentinel,features:rp.O.Focusable,"data-headlessui-focus-guard":!0,as:"button",type:"button",onFocus:g}),null==(t=n.default)?void 0:t.call(n,...e),p.value&&s.isPortalled.value&&(0,$t.h)(rp.j,{id:u,ref:s.afterPanelSentinel,features:rp.O.Focusable,"data-headlessui-focus-guard":!0,as:"button",type:"button",onFocus:v})])]}},features:Qd.O5.RenderStrategy|Qd.O5.Static,visible:p.value,name:"PopoverPanel"})}}}));(0,$t.pM)({name:"PopoverGroup",inheritAttrs:!1,props:{as:{type:[Object,String],default:"div"}},setup(e,{attrs:t,slots:n,expose:r}){let a=(0,jn.KR)(null),o=(0,jn.IJ)([]),i=(0,$t.EW)((()=>(0,Wd.T)(a))),s=function(){let e=(0,jn.KR)(null);return{mainTreeNodeRef:e,MainTreeNode:()=>(0,$t.h)(rp.j,{features:rp.O.Hidden,ref:e})}}();function l(e){let t=o.value.indexOf(e);-1!==t&&o.value.splice(t,1)}return r({el:a,$el:a}),(0,$t.Gt)(mp,{registerPopover:function(e){return o.value.push(e),()=>{l(e)}},unregisterPopover:l,isFocusWithinPopoverGroup:function(){var e;let t=i.value;if(!t)return!1;let n=t.activeElement;return!(null==(e=(0,jd.t)(a))||!e.contains(n))||o.value.some((e=>{var r,a;return(null==(r=t.getElementById(e.buttonId.value))?void 0:r.contains(n))||(null==(a=t.getElementById(e.panelId.value))?void 0:a.contains(n))}))},closeOthers:function(e){for(let t of o.value)t.buttonId.value!==e&&t.close()},mainTreeNodeRef:s.mainTreeNodeRef}),()=>(0,$t.h)($t.FK,[(0,Qd.XX)({ourProps:{ref:a},theirProps:{...e,...t},slot:{},attrs:t,slots:n,name:"PopoverGroup"}),(0,$t.h)(s.MainTreeNode)])}});const Op=["top","right","bottom","left"],kp=Math.min,xp=Math.max,Ep=Math.round,Sp=Math.floor,_p=e=>({x:e,y:e}),Cp={left:"right",right:"left",bottom:"top",top:"bottom"},Tp={start:"end",end:"start"};function Ap(e,t,n){return xp(e,kp(t,n))}function Rp(e,t){return"function"==typeof e?e(t):e}function Pp(e){return e.split("-")[0]}function Dp(e){return e.split("-")[1]}function Mp(e){return"x"===e?"y":"x"}function $p(e){return"y"===e?"height":"width"}function Np(e){return["top","bottom"].includes(Pp(e))?"y":"x"}function Ip(e){return Mp(Np(e))}function Lp(e){return e.replace(/start|end/g,(e=>Tp[e]))}function Bp(e){return e.replace(/left|right|bottom|top/g,(e=>Cp[e]))}function Fp(e){return"number"!=typeof e?function(e){return{top:0,right:0,bottom:0,left:0,...e}}(e):{top:e,right:e,bottom:e,left:e}}function Qp(e){const{x:t,y:n,width:r,height:a}=e;return{width:r,height:a,top:n,left:t,right:t+r,bottom:n+a,x:t,y:n}}function Up(e,t,n){let{reference:r,floating:a}=e;const o=Np(t),i=Ip(t),s=$p(i),l=Pp(t),c="y"===o,u=r.x+r.width/2-a.width/2,d=r.y+r.height/2-a.height/2,p=r[s]/2-a[s]/2;let h;switch(l){case"top":h={x:u,y:r.y-a.height};break;case"bottom":h={x:u,y:r.y+r.height};break;case"right":h={x:r.x+r.width,y:d};break;case"left":h={x:r.x-a.width,y:d};break;default:h={x:r.x,y:r.y}}switch(Dp(t)){case"start":h[i]-=p*(n&&c?-1:1);break;case"end":h[i]+=p*(n&&c?-1:1)}return h}async function zp(e,t){var n;void 0===t&&(t={});const{x:r,y:a,platform:o,rects:i,elements:s,strategy:l}=e,{boundary:c="clippingAncestors",rootBoundary:u="viewport",elementContext:d="floating",altBoundary:p=!1,padding:h=0}=Rp(t,e),f=Fp(h),m=s[p?"floating"===d?"reference":"floating":d],g=Qp(await o.getClippingRect({element:null==(n=await(null==o.isElement?void 0:o.isElement(m)))||n?m:m.contextElement||await(null==o.getDocumentElement?void 0:o.getDocumentElement(s.floating)),boundary:c,rootBoundary:u,strategy:l})),v="floating"===d?{x:r,y:a,width:i.floating.width,height:i.floating.height}:i.reference,b=await(null==o.getOffsetParent?void 0:o.getOffsetParent(s.floating)),y=await(null==o.isElement?void 0:o.isElement(b))&&await(null==o.getScale?void 0:o.getScale(b))||{x:1,y:1},w=Qp(o.convertOffsetParentRelativeRectToViewportRelativeRect?await o.convertOffsetParentRelativeRectToViewportRelativeRect({elements:s,rect:v,offsetParent:b,strategy:l}):v);return{top:(g.top-w.top+f.top)/y.y,bottom:(w.bottom-g.bottom+f.bottom)/y.y,left:(g.left-w.left+f.left)/y.x,right:(w.right-g.right+f.right)/y.x}}function jp(e,t){return{top:e.top-t.height,right:e.right-t.width,bottom:e.bottom-t.height,left:e.left-t.width}}function Wp(e){return Op.some((t=>e[t]>=0))}function Xp(){return"undefined"!=typeof window}function qp(e){return Zp(e)?(e.nodeName||"").toLowerCase():"#document"}function Vp(e){var t;return(null==e||null==(t=e.ownerDocument)?void 0:t.defaultView)||window}function Hp(e){var t;return null==(t=(Zp(e)?e.ownerDocument:e.document)||window.document)?void 0:t.documentElement}function Zp(e){return!!Xp()&&(e instanceof Node||e instanceof Vp(e).Node)}function Gp(e){return!!Xp()&&(e instanceof Element||e instanceof Vp(e).Element)}function Kp(e){return!!Xp()&&(e instanceof HTMLElement||e instanceof Vp(e).HTMLElement)}function Yp(e){return!(!Xp()||"undefined"==typeof ShadowRoot)&&(e instanceof ShadowRoot||e instanceof Vp(e).ShadowRoot)}function Jp(e){const{overflow:t,overflowX:n,overflowY:r,display:a}=oh(e);return/auto|scroll|overlay|hidden|clip/.test(t+r+n)&&!["inline","contents"].includes(a)}function eh(e){return["table","td","th"].includes(qp(e))}function th(e){return[":popover-open",":modal"].some((t=>{try{return e.matches(t)}catch(e){return!1}}))}function nh(e){const t=rh(),n=Gp(e)?oh(e):e;return["transform","translate","scale","rotate","perspective"].some((e=>!!n[e]&&"none"!==n[e]))||!!n.containerType&&"normal"!==n.containerType||!t&&!!n.backdropFilter&&"none"!==n.backdropFilter||!t&&!!n.filter&&"none"!==n.filter||["transform","translate","scale","rotate","perspective","filter"].some((e=>(n.willChange||"").includes(e)))||["paint","layout","strict","content"].some((e=>(n.contain||"").includes(e)))}function rh(){return!("undefined"==typeof CSS||!CSS.supports)&&CSS.supports("-webkit-backdrop-filter","none")}function ah(e){return["html","body","#document"].includes(qp(e))}function oh(e){return Vp(e).getComputedStyle(e)}function ih(e){return Gp(e)?{scrollLeft:e.scrollLeft,scrollTop:e.scrollTop}:{scrollLeft:e.scrollX,scrollTop:e.scrollY}}function sh(e){if("html"===qp(e))return e;const t=e.assignedSlot||e.parentNode||Yp(e)&&e.host||Hp(e);return Yp(t)?t.host:t}function lh(e){const t=sh(e);return ah(t)?e.ownerDocument?e.ownerDocument.body:e.body:Kp(t)&&Jp(t)?t:lh(t)}function ch(e,t,n){var r;void 0===t&&(t=[]),void 0===n&&(n=!0);const a=lh(e),o=a===(null==(r=e.ownerDocument)?void 0:r.body),i=Vp(a);if(o){const e=uh(i);return t.concat(i,i.visualViewport||[],Jp(a)?a:[],e&&n?ch(e):[])}return t.concat(a,ch(a,[],n))}function uh(e){return e.parent&&Object.getPrototypeOf(e.parent)?e.frameElement:null}function dh(e){const t=oh(e);let n=parseFloat(t.width)||0,r=parseFloat(t.height)||0;const a=Kp(e),o=a?e.offsetWidth:n,i=a?e.offsetHeight:r,s=Ep(n)!==o||Ep(r)!==i;return s&&(n=o,r=i),{width:n,height:r,$:s}}function ph(e){return Gp(e)?e:e.contextElement}function hh(e){const t=ph(e);if(!Kp(t))return _p(1);const n=t.getBoundingClientRect(),{width:r,height:a,$:o}=dh(t);let i=(o?Ep(n.width):n.width)/r,s=(o?Ep(n.height):n.height)/a;return i&&Number.isFinite(i)||(i=1),s&&Number.isFinite(s)||(s=1),{x:i,y:s}}const fh=_p(0);function mh(e){const t=Vp(e);return rh()&&t.visualViewport?{x:t.visualViewport.offsetLeft,y:t.visualViewport.offsetTop}:fh}function gh(e,t,n,r){void 0===t&&(t=!1),void 0===n&&(n=!1);const a=e.getBoundingClientRect(),o=ph(e);let i=_p(1);t&&(r?Gp(r)&&(i=hh(r)):i=hh(e));const s=function(e,t,n){return void 0===t&&(t=!1),!(!n||t&&n!==Vp(e))&&t}(o,n,r)?mh(o):_p(0);let l=(a.left+s.x)/i.x,c=(a.top+s.y)/i.y,u=a.width/i.x,d=a.height/i.y;if(o){const e=Vp(o),t=r&&Gp(r)?Vp(r):r;let n=e,a=uh(n);for(;a&&r&&t!==n;){const e=hh(a),t=a.getBoundingClientRect(),r=oh(a),o=t.left+(a.clientLeft+parseFloat(r.paddingLeft))*e.x,i=t.top+(a.clientTop+parseFloat(r.paddingTop))*e.y;l*=e.x,c*=e.y,u*=e.x,d*=e.y,l+=o,c+=i,n=Vp(a),a=uh(n)}}return Qp({width:u,height:d,x:l,y:c})}function vh(e,t){const n=ih(e).scrollLeft;return t?t.left+n:gh(Hp(e)).left+n}function bh(e,t,n){void 0===n&&(n=!1);const r=e.getBoundingClientRect();return{x:r.left+t.scrollLeft-(n?0:vh(e,r)),y:r.top+t.scrollTop}}function yh(e,t,n){let r;if("viewport"===t)r=function(e,t){const n=Vp(e),r=Hp(e),a=n.visualViewport;let o=r.clientWidth,i=r.clientHeight,s=0,l=0;if(a){o=a.width,i=a.height;const e=rh();(!e||e&&"fixed"===t)&&(s=a.offsetLeft,l=a.offsetTop)}return{width:o,height:i,x:s,y:l}}(e,n);else if("document"===t)r=function(e){const t=Hp(e),n=ih(e),r=e.ownerDocument.body,a=xp(t.scrollWidth,t.clientWidth,r.scrollWidth,r.clientWidth),o=xp(t.scrollHeight,t.clientHeight,r.scrollHeight,r.clientHeight);let i=-n.scrollLeft+vh(e);const s=-n.scrollTop;return"rtl"===oh(r).direction&&(i+=xp(t.clientWidth,r.clientWidth)-a),{width:a,height:o,x:i,y:s}}(Hp(e));else if(Gp(t))r=function(e,t){const n=gh(e,!0,"fixed"===t),r=n.top+e.clientTop,a=n.left+e.clientLeft,o=Kp(e)?hh(e):_p(1);return{width:e.clientWidth*o.x,height:e.clientHeight*o.y,x:a*o.x,y:r*o.y}}(t,n);else{const n=mh(e);r={x:t.x-n.x,y:t.y-n.y,width:t.width,height:t.height}}return Qp(r)}function wh(e,t){const n=sh(e);return!(n===t||!Gp(n)||ah(n))&&("fixed"===oh(n).position||wh(n,t))}function Oh(e,t,n){const r=Kp(t),a=Hp(t),o="fixed"===n,i=gh(e,!0,o,t);let s={scrollLeft:0,scrollTop:0};const l=_p(0);if(r||!r&&!o)if(("body"!==qp(t)||Jp(a))&&(s=ih(t)),r){const e=gh(t,!0,o,t);l.x=e.x+t.clientLeft,l.y=e.y+t.clientTop}else a&&(l.x=vh(a));const c=!a||r||o?_p(0):bh(a,s);return{x:i.left+s.scrollLeft-l.x-c.x,y:i.top+s.scrollTop-l.y-c.y,width:i.width,height:i.height}}function kh(e){return"static"===oh(e).position}function xh(e,t){if(!Kp(e)||"fixed"===oh(e).position)return null;if(t)return t(e);let n=e.offsetParent;return Hp(e)===n&&(n=n.ownerDocument.body),n}function Eh(e,t){const n=Vp(e);if(th(e))return n;if(!Kp(e)){let t=sh(e);for(;t&&!ah(t);){if(Gp(t)&&!kh(t))return t;t=sh(t)}return n}let r=xh(e,t);for(;r&&eh(r)&&kh(r);)r=xh(r,t);return r&&ah(r)&&kh(r)&&!nh(r)?n:r||function(e){let t=sh(e);for(;Kp(t)&&!ah(t);){if(nh(t))return t;if(th(t))return null;t=sh(t)}return null}(e)||n}const Sh={convertOffsetParentRelativeRectToViewportRelativeRect:function(e){let{elements:t,rect:n,offsetParent:r,strategy:a}=e;const o="fixed"===a,i=Hp(r),s=!!t&&th(t.floating);if(r===i||s&&o)return n;let l={scrollLeft:0,scrollTop:0},c=_p(1);const u=_p(0),d=Kp(r);if((d||!d&&!o)&&(("body"!==qp(r)||Jp(i))&&(l=ih(r)),Kp(r))){const e=gh(r);c=hh(r),u.x=e.x+r.clientLeft,u.y=e.y+r.clientTop}const p=!i||d||o?_p(0):bh(i,l,!0);return{width:n.width*c.x,height:n.height*c.y,x:n.x*c.x-l.scrollLeft*c.x+u.x+p.x,y:n.y*c.y-l.scrollTop*c.y+u.y+p.y}},getDocumentElement:Hp,getClippingRect:function(e){let{element:t,boundary:n,rootBoundary:r,strategy:a}=e;const o="clippingAncestors"===n?th(t)?[]:function(e,t){const n=t.get(e);if(n)return n;let r=ch(e,[],!1).filter((e=>Gp(e)&&"body"!==qp(e))),a=null;const o="fixed"===oh(e).position;let i=o?sh(e):e;for(;Gp(i)&&!ah(i);){const t=oh(i),n=nh(i);n||"fixed"!==t.position||(a=null),(o?!n&&!a:!n&&"static"===t.position&&a&&["absolute","fixed"].includes(a.position)||Jp(i)&&!n&&wh(e,i))?r=r.filter((e=>e!==i)):a=t,i=sh(i)}return t.set(e,r),r}(t,this._c):[].concat(n),i=[...o,r],s=i[0],l=i.reduce(((e,n)=>{const r=yh(t,n,a);return e.top=xp(r.top,e.top),e.right=kp(r.right,e.right),e.bottom=kp(r.bottom,e.bottom),e.left=xp(r.left,e.left),e}),yh(t,s,a));return{width:l.right-l.left,height:l.bottom-l.top,x:l.left,y:l.top}},getOffsetParent:Eh,getElementRects:async function(e){const t=this.getOffsetParent||Eh,n=this.getDimensions,r=await n(e.floating);return{reference:Oh(e.reference,await t(e.floating),e.strategy),floating:{x:0,y:0,width:r.width,height:r.height}}},getClientRects:function(e){return Array.from(e.getClientRects())},getDimensions:function(e){const{width:t,height:n}=dh(e);return{width:t,height:n}},getScale:hh,isElement:Gp,isRTL:function(e){return"rtl"===oh(e).direction}};function _h(e,t){return e.x===t.x&&e.y===t.y&&e.width===t.width&&e.height===t.height}function Ch(e,t,n,r){void 0===r&&(r={});const{ancestorScroll:a=!0,ancestorResize:o=!0,elementResize:i="function"==typeof ResizeObserver,layoutShift:s="function"==typeof IntersectionObserver,animationFrame:l=!1}=r,c=ph(e),u=a||o?[...c?ch(c):[],...ch(t)]:[];u.forEach((e=>{a&&e.addEventListener("scroll",n,{passive:!0}),o&&e.addEventListener("resize",n)}));const d=c&&s?function(e,t){let n,r=null;const a=Hp(e);function o(){var e;clearTimeout(n),null==(e=r)||e.disconnect(),r=null}return function i(s,l){void 0===s&&(s=!1),void 0===l&&(l=1),o();const c=e.getBoundingClientRect(),{left:u,top:d,width:p,height:h}=c;if(s||t(),!p||!h)return;const f={rootMargin:-Sp(d)+"px "+-Sp(a.clientWidth-(u+p))+"px "+-Sp(a.clientHeight-(d+h))+"px "+-Sp(u)+"px",threshold:xp(0,kp(1,l))||1};let m=!0;function g(t){const r=t[0].intersectionRatio;if(r!==l){if(!m)return i();r?i(!1,r):n=setTimeout((()=>{i(!1,1e-7)}),1e3)}1!==r||_h(c,e.getBoundingClientRect())||i(),m=!1}try{r=new IntersectionObserver(g,{...f,root:a.ownerDocument})}catch(e){r=new IntersectionObserver(g,f)}r.observe(e)}(!0),o}(c,n):null;let p,h=-1,f=null;i&&(f=new ResizeObserver((e=>{let[r]=e;r&&r.target===c&&f&&(f.unobserve(t),cancelAnimationFrame(h),h=requestAnimationFrame((()=>{var e;null==(e=f)||e.observe(t)}))),n()})),c&&!l&&f.observe(c),f.observe(t));let m=l?gh(e):null;return l&&function t(){const r=gh(e);m&&!_h(m,r)&&n(),m=r,p=requestAnimationFrame(t)}(),n(),()=>{var e;u.forEach((e=>{a&&e.removeEventListener("scroll",n),o&&e.removeEventListener("resize",n)})),null==d||d(),null==(e=f)||e.disconnect(),f=null,l&&cancelAnimationFrame(p)}}const Th=function(e){return void 0===e&&(e=0),{name:"offset",options:e,async fn(t){var n,r;const{x:a,y:o,placement:i,middlewareData:s}=t,l=await async function(e,t){const{placement:n,platform:r,elements:a}=e,o=await(null==r.isRTL?void 0:r.isRTL(a.floating)),i=Pp(n),s=Dp(n),l="y"===Np(n),c=["left","top"].includes(i)?-1:1,u=o&&l?-1:1,d=Rp(t,e);let{mainAxis:p,crossAxis:h,alignmentAxis:f}="number"==typeof d?{mainAxis:d,crossAxis:0,alignmentAxis:null}:{mainAxis:d.mainAxis||0,crossAxis:d.crossAxis||0,alignmentAxis:d.alignmentAxis};return s&&"number"==typeof f&&(h="end"===s?-1*f:f),l?{x:h*u,y:p*c}:{x:p*c,y:h*u}}(t,e);return i===(null==(n=s.offset)?void 0:n.placement)&&null!=(r=s.arrow)&&r.alignmentOffset?{}:{x:a+l.x,y:o+l.y,data:{...l,placement:i}}}}},Ah=function(e){return void 0===e&&(e={}),{name:"shift",options:e,async fn(t){const{x:n,y:r,placement:a}=t,{mainAxis:o=!0,crossAxis:i=!1,limiter:s={fn:e=>{let{x:t,y:n}=e;return{x:t,y:n}}},...l}=Rp(e,t),c={x:n,y:r},u=await zp(t,l),d=Np(Pp(a)),p=Mp(d);let h=c[p],f=c[d];if(o){const e="y"===p?"bottom":"right";h=Ap(h+u["y"===p?"top":"left"],h,h-u[e])}if(i){const e="y"===d?"bottom":"right";f=Ap(f+u["y"===d?"top":"left"],f,f-u[e])}const m=s.fn({...t,[p]:h,[d]:f});return{...m,data:{x:m.x-n,y:m.y-r,enabled:{[p]:o,[d]:i}}}}}},Rh=function(e){return void 0===e&&(e={}),{name:"flip",options:e,async fn(t){var n,r;const{placement:a,middlewareData:o,rects:i,initialPlacement:s,platform:l,elements:c}=t,{mainAxis:u=!0,crossAxis:d=!0,fallbackPlacements:p,fallbackStrategy:h="bestFit",fallbackAxisSideDirection:f="none",flipAlignment:m=!0,...g}=Rp(e,t);if(null!=(n=o.arrow)&&n.alignmentOffset)return{};const v=Pp(a),b=Np(s),y=Pp(s)===s,w=await(null==l.isRTL?void 0:l.isRTL(c.floating)),O=p||(y||!m?[Bp(s)]:function(e){const t=Bp(e);return[Lp(e),t,Lp(t)]}(s)),k="none"!==f;!p&&k&&O.push(...function(e,t,n,r){const a=Dp(e);let o=function(e,t,n){const r=["left","right"],a=["right","left"],o=["top","bottom"],i=["bottom","top"];switch(e){case"top":case"bottom":return n?t?a:r:t?r:a;case"left":case"right":return t?o:i;default:return[]}}(Pp(e),"start"===n,r);return a&&(o=o.map((e=>e+"-"+a)),t&&(o=o.concat(o.map(Lp)))),o}(s,m,f,w));const x=[s,...O],E=await zp(t,g),S=[];let _=(null==(r=o.flip)?void 0:r.overflows)||[];if(u&&S.push(E[v]),d){const e=function(e,t,n){void 0===n&&(n=!1);const r=Dp(e),a=Ip(e),o=$p(a);let i="x"===a?r===(n?"end":"start")?"right":"left":"start"===r?"bottom":"top";return t.reference[o]>t.floating[o]&&(i=Bp(i)),[i,Bp(i)]}(a,i,w);S.push(E[e[0]],E[e[1]])}if(_=[..._,{placement:a,overflows:S}],!S.every((e=>e<=0))){var C,T;const e=((null==(C=o.flip)?void 0:C.index)||0)+1,t=x[e];if(t)return{data:{index:e,overflows:_},reset:{placement:t}};let n=null==(T=_.filter((e=>e.overflows[0]<=0)).sort(((e,t)=>e.overflows[1]-t.overflows[1]))[0])?void 0:T.placement;if(!n)switch(h){case"bestFit":{var A;const e=null==(A=_.filter((e=>{if(k){const t=Np(e.placement);return t===b||"y"===t}return!0})).map((e=>[e.placement,e.overflows.filter((e=>e>0)).reduce(((e,t)=>e+t),0)])).sort(((e,t)=>e[1]-t[1]))[0])?void 0:A[0];e&&(n=e);break}case"initialPlacement":n=s}if(a!==n)return{reset:{placement:n}}}return{}}}},Ph=function(e){return void 0===e&&(e={}),{name:"size",options:e,async fn(t){var n,r;const{placement:a,rects:o,platform:i,elements:s}=t,{apply:l=()=>{},...c}=Rp(e,t),u=await zp(t,c),d=Pp(a),p=Dp(a),h="y"===Np(a),{width:f,height:m}=o.floating;let g,v;"top"===d||"bottom"===d?(g=d,v=p===(await(null==i.isRTL?void 0:i.isRTL(s.floating))?"start":"end")?"left":"right"):(v=d,g="end"===p?"top":"bottom");const b=m-u.top-u.bottom,y=f-u.left-u.right,w=kp(m-u[g],b),O=kp(f-u[v],y),k=!t.middlewareData.shift;let x=w,E=O;if(null!=(n=t.middlewareData.shift)&&n.enabled.x&&(E=y),null!=(r=t.middlewareData.shift)&&r.enabled.y&&(x=b),k&&!p){const e=xp(u.left,0),t=xp(u.right,0),n=xp(u.top,0),r=xp(u.bottom,0);h?E=f-2*(0!==e||0!==t?e+t:xp(u.left,u.right)):x=m-2*(0!==n||0!==r?n+r:xp(u.top,u.bottom))}await l({...t,availableWidth:E,availableHeight:x});const S=await i.getDimensions(s.floating);return f!==S.width||m!==S.height?{reset:{rects:!0}}:{}}}},Dh=function(e){return void 0===e&&(e={}),{name:"hide",options:e,async fn(t){const{rects:n}=t,{strategy:r="referenceHidden",...a}=Rp(e,t);switch(r){case"referenceHidden":{const e=jp(await zp(t,{...a,elementContext:"reference"}),n.reference);return{data:{referenceHiddenOffsets:e,referenceHidden:Wp(e)}}}case"escaped":{const e=jp(await zp(t,{...a,altBoundary:!0}),n.floating);return{data:{escapedOffsets:e,escaped:Wp(e)}}}default:return{}}}}},Mh=function(e){return void 0===e&&(e={}),{options:e,fn(t){const{x:n,y:r,placement:a,rects:o,middlewareData:i}=t,{offset:s=0,mainAxis:l=!0,crossAxis:c=!0}=Rp(e,t),u={x:n,y:r},d=Np(a),p=Mp(d);let h=u[p],f=u[d];const m=Rp(s,t),g="number"==typeof m?{mainAxis:m,crossAxis:0}:{mainAxis:0,crossAxis:0,...m};if(l){const e="y"===p?"height":"width",t=o.reference[p]-o.floating[e]+g.mainAxis,n=o.reference[p]+o.reference[e]-g.mainAxis;h<t?h=t:h>n&&(h=n)}if(c){var v,b;const e="y"===p?"width":"height",t=["top","left"].includes(Pp(a)),n=o.reference[d]-o.floating[e]+(t&&(null==(v=i.offset)?void 0:v[d])||0)+(t?0:g.crossAxis),r=o.reference[d]+o.reference[e]+(t?0:(null==(b=i.offset)?void 0:b[d])||0)-(t?g.crossAxis:0);f<n?f=n:f>r&&(f=r)}return{[p]:h,[d]:f}}}};function $h(e){if(function(e){return null!=e&&"object"==typeof e&&"$el"in e}(e)){const t=e.$el;return Zp(t)&&"#comment"===qp(t)?null:t}return e}function Nh(e){return"function"==typeof e?e():(0,jn.R1)(e)}function Ih(e){return"undefined"==typeof window?1:(e.ownerDocument.defaultView||window).devicePixelRatio||1}function Lh(e,t){const n=Ih(e);return Math.round(t*n)/n}function Bh(e,t,n){void 0===n&&(n={});const r=n.whileElementsMounted,a=(0,$t.EW)((()=>{var e;return null==(e=Nh(n.open))||e})),o=(0,$t.EW)((()=>Nh(n.middleware))),i=(0,$t.EW)((()=>{var e;return null!=(e=Nh(n.placement))?e:"bottom"})),s=(0,$t.EW)((()=>{var e;return null!=(e=Nh(n.strategy))?e:"absolute"})),l=(0,$t.EW)((()=>{var e;return null==(e=Nh(n.transform))||e})),c=(0,$t.EW)((()=>$h(e.value))),u=(0,$t.EW)((()=>$h(t.value))),d=(0,jn.KR)(0),p=(0,jn.KR)(0),h=(0,jn.KR)(s.value),f=(0,jn.KR)(i.value),m=(0,jn.IJ)({}),g=(0,jn.KR)(!1),v=(0,$t.EW)((()=>{const e={position:h.value,left:"0",top:"0"};if(!u.value)return e;const t=Lh(u.value,d.value),n=Lh(u.value,p.value);return l.value?{...e,transform:"translate("+t+"px, "+n+"px)",...Ih(u.value)>=1.5&&{willChange:"transform"}}:{position:h.value,left:t+"px",top:n+"px"}}));let b;function y(){if(null==c.value||null==u.value)return;const e=a.value;((e,t,n)=>{const r=new Map,a={platform:Sh,...n},o={...a.platform,_c:r};return(async(e,t,n)=>{const{placement:r="bottom",strategy:a="absolute",middleware:o=[],platform:i}=n,s=o.filter(Boolean),l=await(null==i.isRTL?void 0:i.isRTL(t));let c=await i.getElementRects({reference:e,floating:t,strategy:a}),{x:u,y:d}=Up(c,r,l),p=r,h={},f=0;for(let n=0;n<s.length;n++){const{name:o,fn:m}=s[n],{x:g,y:v,data:b,reset:y}=await m({x:u,y:d,initialPlacement:r,placement:p,strategy:a,middlewareData:h,rects:c,platform:i,elements:{reference:e,floating:t}});u=null!=g?g:u,d=null!=v?v:d,h={...h,[o]:{...h[o],...b}},y&&f<=50&&(f++,"object"==typeof y&&(y.placement&&(p=y.placement),y.rects&&(c=!0===y.rects?await i.getElementRects({reference:e,floating:t,strategy:a}):y.rects),({x:u,y:d}=Up(c,p,l))),n=-1)}return{x:u,y:d,placement:p,strategy:a,middlewareData:h}})(e,t,{...a,platform:o})})(c.value,u.value,{middleware:o.value,placement:i.value,strategy:s.value}).then((t=>{d.value=t.x,p.value=t.y,h.value=t.strategy,f.value=t.placement,m.value=t.middlewareData,g.value=!1!==e}))}function w(){"function"==typeof b&&(b(),b=void 0)}return(0,$t.wB)([o,i,s,a],y,{flush:"sync"}),(0,$t.wB)([c,u],(function(){w(),void 0!==r?null==c.value||null==u.value||(b=r(c.value,u.value,y)):y()}),{flush:"sync"}),(0,$t.wB)(a,(function(){a.value||(g.value=!1)}),{flush:"sync"}),(0,jn.o5)()&&(0,jn.jr)(w),{x:(0,jn.nD)(d),y:(0,jn.nD)(p),strategy:(0,jn.nD)(h),placement:(0,jn.nD)(f),middlewareData:(0,jn.nD)(m),isPositioned:(0,jn.nD)(g),floatingStyles:v,update:y}}const Fh=Symbol(),Qh=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarTeleport",props:{to:{},immediate:{type:Boolean},disabled:{type:Boolean}},setup(e){const t=(0,$t.WQ)(Fh,"body");return(e,n)=>((0,$t.uX)(),(0,$t.Wv)($t.Im,{defer:!e.immediate,disabled:e.disabled,to:e.to||(0,jn.R1)(t)},[(0,$t.Lk)("div",(0,$t.v6)({class:"scalar-app",style:{display:"contents"}},e.$attrs),[(0,$t.RG)(e.$slots,"default")],16)],8,["defer","disabled","to"]))}}),Uh=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarFloating",props:{placement:{default:"bottom"},offset:{default:5},resize:{type:Boolean,default:!1},target:{},middleware:{default:()=>[]},teleport:{type:[Boolean,String]}},setup(e){const t=(0,jn.KR)(null),n=(0,jn.KR)(null),r=(0,$t.EW)((()=>{var t;if(typeof window<"u"&&n.value){if("string"==typeof e.target){const t=document.getElementById(e.target);if(t)return t;console.warn(`ScalarFloating: Target with id="${e.target}" not found`)}else if(e.target instanceof HTMLElement)return e.target;return(null==(t=n.value.children)?void 0:t[0])||n.value}})),a=function(e,t={enabled:(0,jn.KR)(!0)}){const n=(0,jn.KR)(0),r=(0,jn.KR)(0),a=(0,jn.KR)();return typeof ResizeObserver<"u"&&(a.value=new ResizeObserver((([e])=>{var t,a;e&&(n.value=(null==(t=e.borderBoxSize[0])?void 0:t.inlineSize)??0,r.value=(null==(a=e.borderBoxSize[0])?void 0:a.blockSize)??0)}))),(0,$t.wB)([()=>(0,jn.BA)(t.enabled),()=>(0,jn.BA)(e)],(([e,t])=>{!t||!a.value||(e?a.value.observe(t):a.value.disconnect())}),{immediate:!0}),{width:(0,$t.EW)((()=>(0,jn.BA)(t.enabled)?`${n.value}px`:void 0)),height:(0,$t.EW)((()=>(0,jn.BA)(t.enabled)?`${r.value}px`:void 0))}}(r,{enabled:(0,$t.EW)((()=>e.resize))}),o=(0,$t.EW)((()=>"y"===Np(e.placement)?a.width.value:void 0)),i=(0,$t.EW)((()=>"x"===Np(e.placement)?a.height.value:void 0)),{floatingStyles:s,middlewareData:l}=Bh(r,t,{placement:(0,$t.EW)((()=>e.placement)),whileElementsMounted:Ch,middleware:(0,$t.EW)((()=>[Th(e.offset),Rh(),Ah(),Ph({apply({availableWidth:e,availableHeight:t,elements:n}){Object.assign(n.floating.style,{maxWidth:Math.max(0,e)-25+"px",maxHeight:Math.max(0,t)-25+"px"})}}),...e.middleware]))});return(e,r)=>((0,$t.uX)(),(0,$t.CE)($t.FK,null,[(0,$t.Lk)("div",{ref_key:"wrapperRef",ref:n,class:(0,Wn.C4)({contents:!!e.$slots.default})},[(0,$t.RG)(e.$slots,"default")],2),e.$slots.floating?((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Qh),{key:0,disabled:!e.teleport,to:"string"==typeof e.teleport?e.teleport:void 0},{default:(0,$t.k6)((()=>[(0,$t.Lk)("div",{ref_key:"floatingRef",ref:t,class:"relative z-context",style:(0,Wn.Tr)((0,jn.R1)(s))},[(0,$t.RG)(e.$slots,"floating",{data:(0,jn.R1)(l),height:i.value,width:o.value})],4)])),_:3},8,["disabled","to"])):(0,$t.Q3)("",!0)],64))}}),zh=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarFloatingBackdrop",setup(e){const{cx:t}=zn();return(e,n)=>((0,$t.uX)(),(0,$t.CE)("div",(0,Wn._B)((0,$t.Ng)((0,jn.R1)(t)("absolute inset-0 -z-1 rounded bg-b-1 shadow-lg brightness-lifted"))),[(0,$t.RG)(e.$slots,"default")],16))}}),jh=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarComboboxPopover",props:{placement:{},offset:{},resize:{type:Boolean},target:{},middleware:{},teleport:{type:[Boolean,String]}},setup(e,{expose:t}){const{cx:n}=zn(),r=(0,jn.KR)(null),a=e=>{var t;["ArrowUp","ArrowDown"].includes(e.key)&&(e.preventDefault(),null==(t=e.target)||t.dispatchEvent(new KeyboardEvent("keydown",{key:"Enter"})))};return t({popoverButtonRef:r}),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(bp),{as:"template"},{default:(0,$t.k6)((({open:t})=>[(0,$t.bF)((0,jn.R1)(Uh),(0,Wn._B)((0,$t.Ng)(e.$props)),(0,$t.eX)({default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(yp),{ref_key:"popoverButtonRef",ref:r,as:"template",onKeydown:a},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{open:t})])),_:2},1536)])),_:2},[t?{name:"floating",fn:(0,$t.k6)((({width:r})=>[(0,$t.bF)((0,jn.R1)(wp),(0,$t.v6)({focus:"",style:{width:r}},(0,jn.R1)(n)("relative flex flex-col max-h-[inherit] w-40 rounded text-sm")),{default:(0,$t.k6)((({close:n})=>[(0,$t.RG)(e.$slots,"popover",{close:n,open:t}),(0,$t.bF)((0,jn.R1)(zh))])),_:2},1040,["style"])])),key:"0"}:void 0]),1040)])),_:3}))}}),Wh=(0,$t.pM)({__name:"ScalarCombobox",props:{options:{},modelValue:{},placeholder:{},placement:{},offset:{},resize:{type:Boolean},target:{},middleware:{},teleport:{type:[Boolean,String]}},emits:["update:modelValue"],setup:e=>(e,t)=>((0,$t.uX)(),(0,$t.Wv)(jh,{middleware:e.middleware,offset:e.offset,placement:e.placement??"bottom-start",resize:e.resize,target:e.target,teleport:e.teleport},{default:(0,$t.k6)((({open:t})=>[(0,$t.RG)(e.$slots,"default",{open:t})])),popover:(0,$t.k6)((({open:t,close:n})=>[(0,$t.bF)(Fd,{modelValue:e.modelValue?[e.modelValue]:[],open:t,options:e.options,placeholder:e.placeholder,"onUpdate:modelValue":t=>(n(),e.$emit("update:modelValue",t[0]))},(0,$t.eX)({_:2},[e.$slots.before?{name:"before",fn:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"before",{open:t})])),key:"0"}:void 0,e.$slots.after?{name:"after",fn:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"after",{open:t})])),key:"1"}:void 0]),1032,["modelValue","open","options","placeholder","onUpdate:modelValue"])])),_:3},8,["middleware","offset","placement","resize","target","teleport"]))}),Xh=(0,$t.pM)({__name:"ScalarComboboxMultiselect",props:{options:{},modelValue:{},placeholder:{},isDeletable:{type:Boolean},placement:{},offset:{},resize:{type:Boolean},target:{},middleware:{},teleport:{type:[Boolean,String]}},emits:["update:modelValue","delete"],setup(e,{expose:t}){const n=(0,jn.KR)(null);return t({comboboxPopoverRef:n}),(e,t)=>((0,$t.uX)(),(0,$t.Wv)(jh,{ref_key:"comboboxPopoverRef",ref:n,middleware:e.middleware,offset:e.offset,placement:e.placement??"bottom-start",resize:e.resize,target:e.target,teleport:e.teleport},{popover:(0,$t.k6)((({open:n})=>{var r;return[null!=(r=e.options)&&r.length?((0,$t.uX)(),(0,$t.Wv)(Fd,{key:0,isDeletable:e.isDeletable,modelValue:e.modelValue,multiselect:"",open:n,options:e.options,placeholder:e.placeholder,onDelete:t[0]||(t[0]=t=>e.$emit("delete",t)),"onUpdate:modelValue":t[1]||(t[1]=t=>e.$emit("update:modelValue",t))},(0,$t.eX)({_:2},[e.$slots.before?{name:"before",fn:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"before",{open:n})])),key:"0"}:void 0,e.$slots.after?{name:"after",fn:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"after",{open:n})])),key:"1"}:void 0]),1032,["isDeletable","modelValue","open","options","placeholder"])):(0,$t.Q3)("",!0)]})),default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["middleware","offset","placement","resize","target","teleport"]))}});new Map;function qh(e,t){return e-t*Math.floor(e/t)}function Vh(e,t,n,r){let a=(t=Zh(e,t))-1,o=-2;return n<=2?o=0:Hh(t)&&(o=-1),1721425+365*a+Math.floor(a/4)-Math.floor(a/100)+Math.floor(a/400)+Math.floor((367*n-362)/12+o+r)}function Hh(e){return e%4==0&&(e%100!=0||e%400==0)}function Zh(e,t){return"BC"===e?1-t:t}function Gh(e){let t="AD";return e<=0&&(t="BC",e=1-e),[t,e]}const Kh={standard:[31,28,31,30,31,30,31,31,30,31,30,31],leapyear:[31,29,31,30,31,30,31,31,30,31,30,31]};class Yh{fromJulianDay(e){let t=e,n=t-1721426,r=Math.floor(n/146097),a=qh(n,146097),o=Math.floor(a/36524),i=qh(a,36524),s=Math.floor(i/1461),l=qh(i,1461),c=Math.floor(l/365),u=400*r+100*o+4*s+c+(4!==o&&4!==c?1:0),[d,p]=Gh(u),h=t-Vh(d,p,1,1),f=2;t<Vh(d,p,3,1)?f=0:Hh(p)&&(f=1);let m=Math.floor((12*(h+f)+373)/367),g=t-Vh(d,p,m,1)+1;return new Uf(d,p,m,g)}toJulianDay(e){return Vh(e.era,e.year,e.month,e.day)}getDaysInMonth(e){return Kh[Hh(e.year)?"leapyear":"standard"][e.month-1]}getMonthsInYear(e){return 12}getDaysInYear(e){return Hh(e.year)?366:365}getYearsInEra(e){return 9999}getEras(){return["BC","AD"]}isInverseEra(e){return"BC"===e.era}balanceDate(e){e.year<=0&&(e.era="BC"===e.era?"AD":"BC",e.year=1-e.year)}constructor(){this.identifier="gregory"}}function Jh(e,t){return t=yf(t,e.calendar),e.era===t.era&&e.year===t.year&&e.month===t.month&&e.day===t.day}function ef(e,t){return t=yf(t,e.calendar),e=lf(e),t=lf(t),e.era===t.era&&e.year===t.year&&e.month===t.month}function tf(e,t){return Jh(e,nf(t))}function nf(e){return t=function(e){return vf(Date.now(),e)}(e),new Uf(t.calendar,t.era,t.year,t.month,t.day);var t}function rf(e,t){return e.calendar.toJulianDay(e)-t.calendar.toJulianDay(t)}function af(e){return 36e5*e.hour+6e4*e.minute+1e3*e.second+e.millisecond}let of=null;function sf(){return null==of&&(of=(new Intl.DateTimeFormat).resolvedOptions().timeZone),of}function lf(e){return e.subtract({days:e.day-1})}new Map;function cf(e){return uf(Zh((e=yf(e,new Yh)).era,e.year),e.month,e.day,e.hour,e.minute,e.second,e.millisecond)}function uf(e,t,n,r,a,o,i){let s=new Date;return s.setUTCHours(r,a,o,i),s.setUTCFullYear(e,t-1,n),s.getTime()}function df(e,t){if("UTC"===t)return 0;if(e>0&&t===sf())return-6e4*new Date(e).getTimezoneOffset();let{year:n,month:r,day:a,hour:o,minute:i,second:s}=hf(e,t);return uf(n,r,a,o,i,s,0)-1e3*Math.floor(e/1e3)}const pf=new Map;function hf(e,t){let n=pf.get(t);n||(n=new Intl.DateTimeFormat("en-US",{timeZone:t,hour12:!1,era:"short",year:"numeric",month:"numeric",day:"numeric",hour:"numeric",minute:"numeric",second:"numeric"}),pf.set(t,n));let r=n.formatToParts(new Date(e)),a={};for(let e of r)"literal"!==e.type&&(a[e.type]=e.value);return{year:"BC"===a.era||"B"===a.era?1-a.year:+a.year,month:+a.month,day:+a.day,hour:"24"===a.hour?0:+a.hour,minute:+a.minute,second:+a.second}}const ff=864e5;function mf(e,t,n="compatible"){let r=bf(e);if("UTC"===t)return cf(r);if(t===sf()&&"compatible"===n){r=yf(r,new Yh);let e=new Date,t=Zh(r.era,r.year);return e.setFullYear(t,r.month-1,r.day),e.setHours(r.hour,r.minute,r.second,r.millisecond),e.getTime()}let a=cf(r),o=df(a-ff,t),i=df(a+ff,t),s=function(e,t,n,r){return(n===r?[n]:[n,r]).filter((n=>function(e,t,n){let r=hf(n,t);return e.year===r.year&&e.month===r.month&&e.day===r.day&&e.hour===r.hour&&e.minute===r.minute&&e.second===r.second}(e,t,n)))}(r,t,a-o,a-i);if(1===s.length)return s[0];if(s.length>1)switch(n){case"compatible":case"earlier":return s[0];case"later":return s[s.length-1];case"reject":throw new RangeError("Multiple possible absolute times found")}switch(n){case"earlier":return Math.min(a-o,a-i);case"compatible":case"later":return Math.max(a-o,a-i);case"reject":throw new RangeError("No such absolute time found")}}function gf(e,t,n="compatible"){return new Date(mf(e,t,n))}function vf(e,t){let n=df(e,t),r=new Date(e+n),a=r.getUTCFullYear(),o=r.getUTCMonth()+1,i=r.getUTCDate(),s=r.getUTCHours(),l=r.getUTCMinutes(),c=r.getUTCSeconds(),u=r.getUTCMilliseconds();return new Xf(a<1?"BC":"AD",a<1?1-a:a,o,i,t,n,s,l,c,u)}function bf(e,t){let n=0,r=0,a=0,o=0;if("timeZone"in e)({hour:n,minute:r,second:a,millisecond:o}=e);else if("hour"in e&&!t)return e;return t&&({hour:n,minute:r,second:a,millisecond:o}=t),new jf(e.calendar,e.era,e.year,e.month,e.day,n,r,a,o)}function yf(e,t){if(e.calendar.identifier===t.identifier)return e;let n=t.fromJulianDay(e.calendar.toJulianDay(e)),r=e.copy();return r.calendar=t,r.era=n.era,r.year=n.year,r.month=n.month,r.day=n.day,_f(r),r}function wf(e,t,n){return e instanceof Xf?e.timeZone===t?e:function(e,t){return yf(vf(cf(e)-e.offset,t),e.calendar)}(e,t):vf(mf(e,t,n),t)}const Of=36e5;function kf(e,t){let n=e.copy(),r="hour"in n?function(e,t){return e.hour+=t.hours||0,e.minute+=t.minutes||0,e.second+=t.seconds||0,e.millisecond+=t.milliseconds||0,function(e){e.second+=Math.floor(e.millisecond/1e3),e.millisecond=Pf(e.millisecond,1e3),e.minute+=Math.floor(e.second/60),e.second=Pf(e.second,60),e.hour+=Math.floor(e.minute/60),e.minute=Pf(e.minute,60);let t=Math.floor(e.hour/24);return e.hour=Pf(e.hour,24),t}(e)}(n,t):0;xf(n,t.years||0),n.calendar.balanceYearMonth&&n.calendar.balanceYearMonth(n,e),n.month+=t.months||0,Ef(n),Sf(n),n.day+=7*(t.weeks||0),n.day+=t.days||0,n.day+=r,function(e){for(;e.day<1;)e.month--,Ef(e),e.day+=e.calendar.getDaysInMonth(e);for(;e.day>e.calendar.getDaysInMonth(e);)e.day-=e.calendar.getDaysInMonth(e),e.month++,Ef(e)}(n),n.calendar.balanceDate&&n.calendar.balanceDate(n),n.year<1&&(n.year=1,n.month=1,n.day=1);let a=n.calendar.getYearsInEra(n);if(n.year>a){var o,i;let e=null===(o=(i=n.calendar).isInverseEra)||void 0===o?void 0:o.call(i,n);n.year=a,n.month=e?1:n.calendar.getMonthsInYear(n),n.day=e?1:n.calendar.getDaysInMonth(n)}n.month<1&&(n.month=1,n.day=1);let s=n.calendar.getMonthsInYear(n);return n.month>s&&(n.month=s,n.day=n.calendar.getDaysInMonth(n)),n.day=Math.max(1,Math.min(n.calendar.getDaysInMonth(n),n.day)),n}function xf(e,t){var n,r;(null===(n=(r=e.calendar).isInverseEra)||void 0===n?void 0:n.call(r,e))&&(t=-t),e.year+=t}function Ef(e){for(;e.month<1;)xf(e,-1),e.month+=e.calendar.getMonthsInYear(e);let t=0;for(;e.month>(t=e.calendar.getMonthsInYear(e));)e.month-=t,xf(e,1)}function Sf(e){e.month=Math.max(1,Math.min(e.calendar.getMonthsInYear(e),e.month)),e.day=Math.max(1,Math.min(e.calendar.getDaysInMonth(e),e.day))}function _f(e){e.calendar.constrainDate&&e.calendar.constrainDate(e),e.year=Math.max(1,Math.min(e.calendar.getYearsInEra(e),e.year)),Sf(e)}function Cf(e){let t={};for(let n in e)"number"==typeof e[n]&&(t[n]=-e[n]);return t}function Tf(e,t){return kf(e,Cf(t))}function Af(e,t){let n=e.copy();return null!=t.era&&(n.era=t.era),null!=t.year&&(n.year=t.year),null!=t.month&&(n.month=t.month),null!=t.day&&(n.day=t.day),_f(n),n}function Rf(e,t){let n=e.copy();var r;return null!=t.hour&&(n.hour=t.hour),null!=t.minute&&(n.minute=t.minute),null!=t.second&&(n.second=t.second),null!=t.millisecond&&(n.millisecond=t.millisecond),(r=n).millisecond=Math.max(0,Math.min(r.millisecond,1e3)),r.second=Math.max(0,Math.min(r.second,59)),r.minute=Math.max(0,Math.min(r.minute,59)),r.hour=Math.max(0,Math.min(r.hour,23)),n}function Pf(e,t){let n=e%t;return n<0&&(n+=t),n}function Df(e,t,n,r){let a=e.copy();switch(t){case"era":{let t=e.calendar.getEras(),o=t.indexOf(e.era);if(o<0)throw new Error("Invalid era: "+e.era);o=$f(o,n,0,t.length-1,null==r?void 0:r.round),a.era=t[o],_f(a);break}case"year":var o,i;(null===(o=(i=a.calendar).isInverseEra)||void 0===o?void 0:o.call(i,a))&&(n=-n),a.year=$f(e.year,n,-1/0,9999,null==r?void 0:r.round),a.year===-1/0&&(a.year=1),a.calendar.balanceYearMonth&&a.calendar.balanceYearMonth(a,e);break;case"month":a.month=$f(e.month,n,1,e.calendar.getMonthsInYear(e),null==r?void 0:r.round);break;case"day":a.day=$f(e.day,n,1,e.calendar.getDaysInMonth(e),null==r?void 0:r.round);break;default:throw new Error("Unsupported field "+t)}return e.calendar.balanceDate&&e.calendar.balanceDate(a),_f(a),a}function Mf(e,t,n,r){let a=e.copy();switch(t){case"hour":{let t=e.hour,o=0,i=23;if(12===(null==r?void 0:r.hourCycle)){let e=t>=12;o=e?12:0,i=e?23:11}a.hour=$f(t,n,o,i,null==r?void 0:r.round);break}case"minute":a.minute=$f(e.minute,n,0,59,null==r?void 0:r.round);break;case"second":a.second=$f(e.second,n,0,59,null==r?void 0:r.round);break;case"millisecond":a.millisecond=$f(e.millisecond,n,0,999,null==r?void 0:r.round);break;default:throw new Error("Unsupported field "+t)}return a}function $f(e,t,n,r,a=!1){if(a){(e+=Math.sign(t))<n&&(e=r);let a=Math.abs(t);(e=t>0?Math.ceil(e/a)*a:Math.floor(e/a)*a)>r&&(e=n)}else(e+=t)<n?e=r-(n-e-1):e>r&&(e=n+(e-r-1));return e}function Nf(e,t){let n;return n=null!=t.years&&0!==t.years||null!=t.months&&0!==t.months||null!=t.weeks&&0!==t.weeks||null!=t.days&&0!==t.days?mf(kf(bf(e),{years:t.years,months:t.months,weeks:t.weeks,days:t.days}),e.timeZone):cf(e)-e.offset,n+=t.milliseconds||0,n+=1e3*(t.seconds||0),n+=6e4*(t.minutes||0),n+=36e5*(t.hours||0),yf(vf(n,e.timeZone),e.calendar)}function If(e){let t,n=yf(e,new Yh);return t="BC"===n.era?1===n.year?"0000":"-"+String(Math.abs(1-n.year)).padStart(6,"00"):String(n.year).padStart(4,"0"),`${t}-${String(n.month).padStart(2,"0")}-${String(n.day).padStart(2,"0")}`}function Lf(e){return`${If(e)}T${t=e,`${String(t.hour).padStart(2,"0")}:${String(t.minute).padStart(2,"0")}:${String(t.second).padStart(2,"0")}${t.millisecond?String(t.millisecond/1e3).slice(1):""}`}`;var t}function Bf(e,t,n){!function(e,t){if(t.has(e))throw new TypeError("Cannot initialize the same private elements twice on an object")}(e,t),t.set(e,n)}function Ff(e){let t,n="object"==typeof e[0]?e.shift():new Yh;if("string"==typeof e[0])t=e.shift();else{let e=n.getEras();t=e[e.length-1]}return[n,t,e.shift(),e.shift(),e.shift()]}var Qf=new WeakMap;class Uf{copy(){return this.era?new Uf(this.calendar,this.era,this.year,this.month,this.day):new Uf(this.calendar,this.year,this.month,this.day)}add(e){return kf(this,e)}subtract(e){return Tf(this,e)}set(e){return Af(this,e)}cycle(e,t,n){return Df(this,e,t,n)}toDate(e){return gf(this,e)}toString(){return If(this)}compare(e){return rf(this,e)}constructor(...e){Bf(this,Qf,{writable:!0,value:void 0});let[t,n,r,a,o]=Ff(e);this.calendar=t,this.era=n,this.year=r,this.month=a,this.day=o,_f(this)}}var zf=new WeakMap;class jf{copy(){return this.era?new jf(this.calendar,this.era,this.year,this.month,this.day,this.hour,this.minute,this.second,this.millisecond):new jf(this.calendar,this.year,this.month,this.day,this.hour,this.minute,this.second,this.millisecond)}add(e){return kf(this,e)}subtract(e){return Tf(this,e)}set(e){return Af(Rf(this,e),e)}cycle(e,t,n){switch(e){case"era":case"year":case"month":case"day":return Df(this,e,t,n);default:return Mf(this,e,t,n)}}toDate(e,t){return gf(this,e,t)}toString(){return Lf(this)}compare(e){let t=rf(this,e);return 0===t?function(e,t){return af(e)-af(t)}(this,bf(e)):t}constructor(...e){Bf(this,zf,{writable:!0,value:void 0});let[t,n,r,a,o]=Ff(e);this.calendar=t,this.era=n,this.year=r,this.month=a,this.day=o,this.hour=e.shift()||0,this.minute=e.shift()||0,this.second=e.shift()||0,this.millisecond=e.shift()||0,_f(this)}}var Wf=new WeakMap;class Xf{copy(){return this.era?new Xf(this.calendar,this.era,this.year,this.month,this.day,this.timeZone,this.offset,this.hour,this.minute,this.second,this.millisecond):new Xf(this.calendar,this.year,this.month,this.day,this.timeZone,this.offset,this.hour,this.minute,this.second,this.millisecond)}add(e){return Nf(this,e)}subtract(e){return function(e,t){return Nf(e,Cf(t))}(this,e)}set(e,t){return function(e,t,n){let r=bf(e),a=Rf(Af(r,t),t);return 0===a.compare(r)?e:yf(vf(mf(a,e.timeZone,n),e.timeZone),e.calendar)}(this,e,t)}cycle(e,t,n){return function(e,t,n,r){switch(t){case"hour":{let t=0,a=23;if(12===(null==r?void 0:r.hourCycle)){let n=e.hour>=12;t=n?12:0,a=n?23:11}let o=bf(e),i=yf(Rf(o,{hour:t}),new Yh),s=[mf(i,e.timeZone,"earlier"),mf(i,e.timeZone,"later")].filter((t=>vf(t,e.timeZone).day===i.day))[0],l=yf(Rf(o,{hour:a}),new Yh),c=[mf(l,e.timeZone,"earlier"),mf(l,e.timeZone,"later")].filter((t=>vf(t,e.timeZone).day===l.day)).pop(),u=cf(e)-e.offset,d=Math.floor(u/Of),p=u%Of;return u=$f(d,n,Math.floor(s/Of),Math.floor(c/Of),null==r?void 0:r.round)*Of+p,yf(vf(u,e.timeZone),e.calendar)}case"minute":case"second":case"millisecond":return Mf(e,t,n,r);case"era":case"year":case"month":case"day":return yf(vf(mf(Df(bf(e),t,n,r),e.timeZone),e.timeZone),e.calendar);default:throw new Error("Unsupported field "+t)}}(this,e,t,n)}toDate(){return function(e){let t=cf(e)-e.offset;return new Date(t)}(this)}toString(){return`${Lf(this)}${function(e){let t=Math.sign(e)<0?"-":"+";e=Math.abs(e);let n=Math.floor(e/36e5),r=e%36e5/6e4;return`${t}${String(n).padStart(2,"0")}:${String(r).padStart(2,"0")}`}(this.offset)}[${this.timeZone}]`}toAbsoluteString(){return this.toDate().toISOString()}compare(e){return this.toDate().getTime()-wf(e,this.timeZone).toDate().getTime()}constructor(...e){Bf(this,Wf,{writable:!0,value:void 0});let[t,n,r,a,o]=Ff(e),i=e.shift(),s=e.shift();this.calendar=t,this.era=n,this.year=r,this.month=a,this.day=o,this.timeZone=i,this.offset=s,this.hour=e.shift()||0,this.minute=e.shift()||0,this.second=e.shift()||0,this.millisecond=e.shift()||0,_f(this)}}function qf(e,t=sf()){return Vf(e)?e.toDate():e.toDate(t)}function Vf(e){return e instanceof Xf}function Hf(e){if(e instanceof Date){const t=e.getFullYear(),n=e.getMonth()+1;return new Date(t,n,0).getDate()}return e.set({day:100}).day}new Map;let Zf=!1;try{Zf="exceptZero"===new Intl.NumberFormat("de-DE",{signDisplay:"exceptZero"}).resolvedOptions().signDisplay}catch{}let Gf=!1;try{Gf="unit"===new Intl.NumberFormat("de-DE",{style:"unit",unit:"degree"}).resolvedOptions().style}catch{}new RegExp("^.*\\(.*\\).*$");new Map;new Set(["decimal","fraction","integer","minusSign","plusSign","group"]);function Kf(e,t){const n=Symbol("string"!=typeof e||t?t:`${e}Context`);return[t=>{const r=(0,$t.WQ)(n,t);if(r||null===r)return r;throw new Error(`Injection \`${n.toString()}\` not found. Component must be used within ${Array.isArray(e)?`one of the following components: ${e.join(", ")}`:`\`${e}\``}`)},e=>((0,$t.Gt)(n,e),e)]}function Yf(e,t,n){const r=n.originalEvent.target,a=new CustomEvent(e,{bubbles:!1,cancelable:!0,detail:n});t&&r.addEventListener(e,t,{once:!0}),r.dispatchEvent(a)}function Jf(e,t=Number.NEGATIVE_INFINITY,n=Number.POSITIVE_INFINITY){return Math.min(n,Math.max(t,e))}function em(e){return null==e}const tm=["day","month","year"],nm=["hour","minute","second","dayPeriod"];function rm(e){return!!(0,jn.o5)()&&((0,jn.jr)(e),!0)}function am(){const e=new Set,t=t=>{e.delete(t)};return{on:n=>{e.add(n);const r=()=>t(n);return rm(r),{off:r}},off:t,trigger:(...t)=>Promise.all(Array.from(e).map((e=>e(...t))))}}function om(e){let t,n,r=0;const a=()=>{r-=1,n&&r<=0&&(n.stop(),t=void 0,n=void 0)};return(...o)=>(r+=1,t||(n=(0,jn.uY)(!0),t=n.run((()=>e(...o)))),rm(a),t)}function im(e){return"function"==typeof e?e():(0,jn.R1)(e)}const sm=typeof window<"u"&&typeof document<"u";typeof WorkerGlobalScope<"u"&&(globalThis,WorkerGlobalScope);const lm=Object.prototype.toString,cm=()=>{},um=dm();function dm(){var e,t;return sm&&(null==(e=null==window?void 0:window.navigator)?void 0:e.userAgent)&&(/iP(?:ad|hone|od)/.test(window.navigator.userAgent)||(null==(t=null==window?void 0:window.navigator)?void 0:t.maxTouchPoints)>2&&/iPad|Macintosh/.test(null==window?void 0:window.navigator.userAgent))}function pm(e,t){return function(...n){return new Promise(((r,a)=>{Promise.resolve(e((()=>t.apply(this,n)),{fn:t,thisArg:this,args:n})).then(r).catch(a)}))}}const hm=e=>e();function fm(e,t=1e4){return(0,jn.rY)(((n,r)=>{let a,o=im(e);return rm((()=>{clearTimeout(a)})),{get:()=>(n(),o),set(n){o=n,r(),clearTimeout(a),a=setTimeout((()=>{o=im(e),r()}),im(t))}}}))}function mm(e,t=200,n={}){return pm(function(e,t={}){let n,r,a=cm;const o=e=>{clearTimeout(e),a(),a=cm};return i=>{const s=im(e),l=im(t.maxWait);return n&&o(n),s<=0||void 0!==l&&l<=0?(r&&(o(r),r=null),Promise.resolve(i())):new Promise(((e,c)=>{a=t.rejectOnCancel?c:e,l&&!r&&(r=setTimeout((()=>{n&&o(n),r=null,e(i())}),l)),n=setTimeout((()=>{r&&o(r),r=null,e(i())}),s)}))}}(t,n),e)}function gm(e,t,n={}){const{eventFilter:r=hm,...a}=n;return(0,$t.wB)(e,pm(r,t),a)}function vm(e,t,n={}){const{eventFilter:r,...a}=n,{eventFilter:o,pause:i,resume:s,isActive:l}=function(e=hm){const t=(0,jn.KR)(!0);return{isActive:(0,jn.tB)(t),pause:function(){t.value=!1},resume:function(){t.value=!0},eventFilter:(...n)=>{t.value&&e(...n)}}}(r);return{stop:gm(e,t,{...a,eventFilter:o}),pause:i,resume:s,isActive:l}}function bm(e,t,n={}){const{immediate:r=!0}=n,a=(0,jn.KR)(!1);let o=null;function i(){o&&(clearTimeout(o),o=null)}function s(){a.value=!1,i()}function l(...n){i(),a.value=!0,o=setTimeout((()=>{a.value=!1,o=null,e(...n)}),im(t))}return r&&(a.value=!0,sm&&l()),rm(s),{isPending:(0,jn.tB)(a),start:l,stop:s}}function ym(e){var t;const n=im(e);return null!=(t=null==n?void 0:n.$el)?t:n}const wm=sm?window:void 0;function Om(...e){let t,n,r,a;if("string"==typeof e[0]||Array.isArray(e[0])?([n,r,a]=e,t=wm):[t,n,r,a]=e,!t)return cm;Array.isArray(n)||(n=[n]),Array.isArray(r)||(r=[r]);const o=[],i=()=>{o.forEach((e=>e())),o.length=0},s=(0,$t.wB)((()=>[ym(t),im(a)]),(([e,t])=>{if(i(),!e)return;const a=(e=>"[object Object]"===lm.call(e))(t)?{...t}:t;o.push(...n.flatMap((t=>r.map((n=>((e,t,n,r)=>(e.addEventListener(t,n,r),()=>e.removeEventListener(t,n,r)))(e,t,n,a))))))}),{immediate:!0,flush:"post"}),l=()=>{s(),i()};return rm(l),l}function km(...e){let t,n,r={};3===e.length?(t=e[0],n=e[1],r=e[2]):2===e.length?"object"==typeof e[1]?(t=!0,n=e[0],r=e[1]):(t=e[0],n=e[1]):(t=!0,n=e[0]);const{target:a=wm,eventName:o="keydown",passive:i=!1,dedupe:s=!1}=r,l=function(e){return"function"==typeof e?e:"string"==typeof e?t=>t.key===e:Array.isArray(e)?t=>e.includes(t.key):()=>!0}(t);return Om(a,o,(e=>{e.repeat&&im(s)||l(e)&&n(e)}),i)}function xm(){const e=(0,jn.KR)(!1),t=(0,$t.nI)();return t&&(0,$t.sV)((()=>{e.value=!0}),t),e}function Em(e){const t=xm();return(0,$t.EW)((()=>(t.value,!!e())))}function Sm(e,t={}){const{immediate:n=!0,fpsLimit:r,window:a=wm}=t,o=(0,jn.KR)(!1),i=r?1e3/r:null;let s=0,l=null;function c(t){if(!o.value||!a)return;s||(s=t);const n=t-s;i&&n<i||(s=t,e({delta:n,timestamp:t})),l=a.requestAnimationFrame(c)}function u(){!o.value&&a&&(o.value=!0,s=0,l=a.requestAnimationFrame(c))}function d(){o.value=!1,null!=l&&a&&(a.cancelAnimationFrame(l),l=null)}return n&&u(),rm(d),{isActive:(0,jn.tB)(o),pause:d,resume:u}}function _m(e,t,n={}){const{window:r=wm,...a}=n;let o;const i=Em((()=>r&&"ResizeObserver"in r)),s=()=>{o&&(o.disconnect(),o=void 0)},l=(0,$t.EW)((()=>Array.isArray(e)?e.map((e=>ym(e))):[ym(e)])),c=(0,$t.wB)(l,(e=>{if(s(),i.value&&r){o=new ResizeObserver(t);for(const t of e)t&&o.observe(t,a)}}),{immediate:!0,flush:"post"}),u=()=>{s(),c()};return rm(u),{isSupported:i,stop:u}}function Cm(e,t,n,r={}){var a,o,i;const{clone:s=!1,passive:l=!1,eventName:c,deep:u=!1,defaultValue:d,shouldEmit:p}=r,h=(0,$t.nI)(),f=n||(null==h?void 0:h.emit)||(null==(a=null==h?void 0:h.$emit)?void 0:a.bind(h))||(null==(i=null==(o=null==h?void 0:h.proxy)?void 0:o.$emit)?void 0:i.bind(null==h?void 0:h.proxy));let m=c;t||(t="modelValue"),m=m||`update:${t.toString()}`;const g=e=>s?"function"==typeof s?s(e):function(e){return JSON.parse(JSON.stringify(e))}(e):e,v=()=>(e=>typeof e<"u")(e[t])?g(e[t]):d,b=e=>{p?p(e)&&f(m,e):f(m,e)};if(l){const n=v(),r=(0,jn.KR)(n);let a=!1;return(0,$t.wB)((()=>e[t]),(e=>{a||(a=!0,r.value=g(e),(0,$t.dY)((()=>a=!1)))})),(0,$t.wB)(r,(n=>{!a&&(n!==e[t]||u)&&b(n)}),{deep:u}),r}return(0,$t.EW)({get:()=>v(),set(e){b(e)}})}function Tm(e){return e?e.flatMap((e=>e.type===$t.FK?Tm(e.children):[e])):[]}function Am(){let e=document.activeElement;if(null==e)return null;for(;null!=e&&null!=e.shadowRoot&&null!=e.shadowRoot.activeElement;)e=e.shadowRoot.activeElement;return e}const Rm=["INPUT","TEXTAREA"];function Pm(e,t,n,r={}){if(!t||r.enableIgnoredElement&&Rm.includes(t.nodeName))return null;const{arrowKeyOptions:a="both",attributeName:o="[data-radix-vue-collection-item]",itemsArray:i=[],loop:s=!0,dir:l="ltr",preventScroll:c=!0,focus:u=!1}=r,[d,p,h,f,m,g]=["ArrowRight"===e.key,"ArrowLeft"===e.key,"ArrowUp"===e.key,"ArrowDown"===e.key,"Home"===e.key,"End"===e.key],v=h||f,b=d||p;if(!m&&!g&&(!v&&!b||"vertical"===a&&b||"horizontal"===a&&v))return null;const y=n?Array.from(n.querySelectorAll(o)):i;if(!y.length)return null;c&&e.preventDefault();let w=null;return b||v?w=Dm(y,t,{goForward:v?f:"ltr"===l?d:p,loop:s}):m?w=y.at(0)||null:g&&(w=y.at(-1)||null),u&&(null==w||w.focus()),w}function Dm(e,t,n,r=e.length){if(0==--r)return null;const a=e.indexOf(t),o=n.goForward?a+1:a-1;if(!n.loop&&(o<0||o>=e.length))return null;const i=e[(o+e.length)%e.length];return i?i.hasAttribute("disabled")&&"false"!==i.getAttribute("disabled")?Dm(e,i,n,r):i:null}function Mm(e){if(null===e||"object"!=typeof e)return!1;const t=Object.getPrototypeOf(e);return!(null!==t&&t!==Object.prototype&&null!==Object.getPrototypeOf(t)||Symbol.iterator in e||Symbol.toStringTag in e&&"[object Module]"!==Object.prototype.toString.call(e))}function $m(e,t,n=".",r){if(!Mm(t))return $m(e,{},n,r);const a=Object.assign({},t);for(const t in e){if("__proto__"===t||"constructor"===t)continue;const o=e[t];null!=o&&(r&&r(a,t,o,n)||(Array.isArray(o)&&Array.isArray(a[t])?a[t]=[...o,...a[t]]:Mm(o)&&Mm(a[t])?a[t]=$m(o,a[t],(n?`${n}.`:"")+t.toString(),r):a[t]=o))}return a}const Nm=(...e)=>e.reduce(((e,t)=>$m(e,t,"",void 0)),{}),[Im,Lm]=Kf("ConfigProvider");Boolean;const Bm=om((()=>{const e=(0,jn.KR)(new Map),t=(0,jn.KR)(),n=(0,$t.EW)((()=>{for(const t of e.value.values())if(t)return!0;return!1})),r=Im({scrollBody:(0,jn.KR)(!0)});let a=null;return(0,$t.wB)(n,((e,n)=>{var o;if(!sm)return;if(!e)return void(n&&(document.body.style.paddingRight="",document.body.style.marginRight="",document.body.style.pointerEvents="",document.body.style.removeProperty("--scrollbar-width"),document.body.style.overflow=t.value??"",um&&(null==a||a()),t.value=void 0));void 0===t.value&&(t.value=document.body.style.overflow);const i=window.innerWidth-document.documentElement.clientWidth,s={padding:i,margin:0},l=null!=(o=r.scrollBody)&&o.value?"object"==typeof r.scrollBody.value?Nm({padding:!0===r.scrollBody.value.padding?i:r.scrollBody.value.padding,margin:!0===r.scrollBody.value.margin?i:r.scrollBody.value.margin},s):s:{padding:0,margin:0};i>0&&(document.body.style.paddingRight="number"==typeof l.padding?`${l.padding}px`:String(l.padding),document.body.style.marginRight="number"==typeof l.margin?`${l.margin}px`:String(l.margin),document.body.style.setProperty("--scrollbar-width",`${i}px`),document.body.style.overflow="hidden"),um&&(a=Om(document,"touchmove",(e=>function(e){const t=e||window.event,n=t.target;return!(n instanceof Element&&Qm(n))&&(t.touches.length>1||(t.preventDefault&&t.cancelable&&t.preventDefault(),!1))}(e)),{passive:!1})),(0,$t.dY)((()=>{document.body.style.pointerEvents="none",document.body.style.overflow="hidden"}))}),{immediate:!0,flush:"sync"}),e}));function Fm(e){const t=((e=21)=>{let t="",n=e;for(;n--;)t+="useandom-26T198340PX75pxJACKVERYMINDBUSHWOLF_GQZbfghjklqvwyzrict"[64*Math.random()|0];return t})(6),n=Bm();n.value.set(t,e??!1);const r=(0,$t.EW)({get:()=>n.value.get(t)??!1,set:e=>n.value.set(t,e)});return function(e){(0,$t.nI)()&&(0,$t.xo)(e,void 0)}((()=>{n.value.delete(t)})),r}function Qm(e){const t=window.getComputedStyle(e);if("scroll"===t.overflowX||"scroll"===t.overflowY||"auto"===t.overflowX&&e.clientWidth<e.scrollWidth||"auto"===t.overflowY&&e.clientHeight<e.scrollHeight)return!0;{const t=e.parentNode;return t instanceof Element&&"BODY"!==t.tagName&&Qm(t)}}function Um(e,t="data-radix-vue-collection-item"){const n=e??Symbol();return{createCollection:e=>{const r=(0,jn.KR)([]);function a(){const n=ym(e);return r.value=n?Array.from(n.querySelectorAll(`[${t}]:not([data-disabled])`)):[]}return(0,$t.Ic)((()=>{r.value=[]})),(0,$t.sV)(a),(0,$t.$u)(a),(0,$t.wB)((()=>null==e?void 0:e.value),a,{immediate:!0}),(0,$t.Gt)(n,r),r},injectCollection:()=>(0,$t.WQ)(n,(0,jn.KR)([]))}}function zm(e){const t=Im({dir:(0,jn.KR)("ltr")});return(0,$t.EW)((()=>{var n;return(null==e?void 0:e.value)||(null==(n=t.dir)?void 0:n.value)||"ltr"}))}function jm(e){const t=(0,$t.nI)(),n=null==t?void 0:t.type.emits,r={};return null!=n&&n.length||console.warn(`No emitted event found. Please check component: ${null==t?void 0:t.type.__name}`),null==n||n.forEach((t=>{r[(0,Wn.rU)((0,Wn.PT)(t))]=(...n)=>e(t,...n)})),r}let Wm=0;function Xm(){(0,$t.nT)((e=>{if(!sm)return;const t=document.querySelectorAll("[data-radix-focus-guard]");document.body.insertAdjacentElement("afterbegin",t[0]??qm()),document.body.insertAdjacentElement("beforeend",t[1]??qm()),Wm++,e((()=>{1===Wm&&document.querySelectorAll("[data-radix-focus-guard]").forEach((e=>e.remove())),Wm--}))}))}function qm(){const e=document.createElement("span");return e.setAttribute("data-radix-focus-guard",""),e.tabIndex=0,e.style.outline="none",e.style.opacity="0",e.style.position="fixed",e.style.pointerEvents="none",e}function Vm(e){const t=(0,$t.nI)(),n=Object.keys((null==t?void 0:t.type.props)??{}).reduce(((e,n)=>{const r=(null==t?void 0:t.type.props[n]).default;return void 0!==r&&(e[n]=r),e}),{}),r=(0,jn.lW)(e);return(0,$t.EW)((()=>{const e={},a=(null==t?void 0:t.vnode.props)??{};return Object.keys(a).forEach((t=>{e[(0,Wn.PT)(t)]=a[t]})),Object.keys({...n,...e}).reduce(((e,t)=>(void 0!==r.value[t]&&(e[t]=r.value[t]),e)),{})}))}function Hm(e,t){const n=Vm(e),r=t?jm(t):{};return(0,$t.EW)((()=>({...n.value,...r})))}function Zm(){const e=(0,$t.nI)(),t=(0,jn.KR)(),n=(0,$t.EW)((()=>{var e,n;return["#text","#comment"].includes(null==(e=t.value)?void 0:e.$el.nodeName)?null==(n=t.value)?void 0:n.$el.nextElementSibling:ym(t)})),r=Object.assign({},e.exposed),a={};for(const t in e.props)Object.defineProperty(a,t,{enumerable:!0,configurable:!0,get:()=>e.props[t]});if(Object.keys(r).length>0)for(const e in r)Object.defineProperty(a,e,{enumerable:!0,configurable:!0,get:()=>r[e]});return Object.defineProperty(a,"$el",{enumerable:!0,configurable:!0,get:()=>e.vnode.el}),e.exposed=a,{forwardRef:function(n){t.value=n,n&&(Object.defineProperty(a,"$el",{enumerable:!0,configurable:!0,get:()=>n instanceof Element?n:n.$el}),e.exposed=a)},currentRef:t,currentElement:n}}function Gm(e,t){const n=fm(!1,300),r=(0,jn.KR)(null),a=am();function o(){r.value=null,n.value=!1}function i(e,t){const a=e.currentTarget,o={x:e.clientX,y:e.clientY},i=function(e,t){const n=Math.abs(t.top-e.y),r=Math.abs(t.bottom-e.y),a=Math.abs(t.right-e.x),o=Math.abs(t.left-e.x);switch(Math.min(n,r,a,o)){case o:return"left";case a:return"right";case n:return"top";case r:return"bottom";default:throw new Error("unreachable")}}(o,a.getBoundingClientRect()),s=function(e,t,n=5){const r=[];switch(t){case"top":r.push({x:e.x-n,y:e.y+n},{x:e.x+n,y:e.y+n});break;case"bottom":r.push({x:e.x-n,y:e.y-n},{x:e.x+n,y:e.y-n});break;case"left":r.push({x:e.x+n,y:e.y-n},{x:e.x+n,y:e.y+n});break;case"right":r.push({x:e.x-n,y:e.y-n},{x:e.x-n,y:e.y+n})}return r}(o,i),l=function(e){const{top:t,right:n,bottom:r,left:a}=e;return[{x:a,y:t},{x:n,y:t},{x:n,y:r},{x:a,y:r}]}(t.getBoundingClientRect()),c=function(e){const t=e.slice();return t.sort(((e,t)=>e.x<t.x?-1:e.x>t.x?1:e.y<t.y?-1:e.y>t.y?1:0)),function(e){if(e.length<=1)return e.slice();const t=[];for(let n=0;n<e.length;n++){const r=e[n];for(;t.length>=2;){const e=t[t.length-1],n=t[t.length-2];if(!((e.x-n.x)*(r.y-n.y)>=(e.y-n.y)*(r.x-n.x)))break;t.pop()}t.push(r)}t.pop();const n=[];for(let t=e.length-1;t>=0;t--){const r=e[t];for(;n.length>=2;){const e=n[n.length-1],t=n[n.length-2];if(!((e.x-t.x)*(r.y-t.y)>=(e.y-t.y)*(r.x-t.x)))break;n.pop()}n.push(r)}return n.pop(),1===t.length&&1===n.length&&t[0].x===n[0].x&&t[0].y===n[0].y?t:t.concat(n)}(t)}([...s,...l]);r.value=c,n.value=!0}return(0,$t.nT)((n=>{if(e.value&&t.value){const r=e=>i(e,t.value),a=t=>i(t,e.value);e.value.addEventListener("pointerleave",r),t.value.addEventListener("pointerleave",a),n((()=>{var n,o;null==(n=e.value)||n.removeEventListener("pointerleave",r),null==(o=t.value)||o.removeEventListener("pointerleave",a)}))}})),(0,$t.nT)((n=>{var i;if(r.value){const s=n=>{var i,s;if(!r.value)return;const l=n.target,c={x:n.clientX,y:n.clientY},u=(null==(i=e.value)?void 0:i.contains(l))||(null==(s=t.value)?void 0:s.contains(l)),d=!function(e,t){const{x:n,y:r}=e;let a=!1;for(let e=0,o=t.length-1;e<t.length;o=e++){const i=t[e].x,s=t[e].y,l=t[o].x,c=t[o].y;s>r!=c>r&&n<(l-i)*(r-s)/(c-s)+i&&(a=!a)}return a}(c,r.value),p=!!l.closest("[data-grace-area-trigger]");u?o():(d||p)&&(o(),a.trigger())};null==(i=e.value)||i.ownerDocument.addEventListener("pointermove",s),n((()=>{var t;return null==(t=e.value)?void 0:t.ownerDocument.removeEventListener("pointermove",s)}))}})),{isPointerInTransit:n,onPointerExit:a.on}}var Km=new WeakMap,Ym=new WeakMap,Jm={},eg=0,tg=function(e){return e&&(e.host||tg(e.parentNode))},ng=function(e,t,n){void 0===n&&(n="data-aria-hidden");var r=Array.from(Array.isArray(e)?e:[e]),a=function(e){return typeof document>"u"?null:(Array.isArray(e)?e[0]:e).ownerDocument.body}(e);return a?(r.push.apply(r,Array.from(a.querySelectorAll("[aria-live]"))),function(e,t,n,r){var a=function(e,t){return t.map((function(t){if(e.contains(t))return t;var n=tg(t);return n&&e.contains(n)?n:(console.error("aria-hidden",t,"in not contained inside",e,". Doing nothing"),null)})).filter((function(e){return!!e}))}(t,Array.isArray(e)?e:[e]);Jm[n]||(Jm[n]=new WeakMap);var o=Jm[n],i=[],s=new Set,l=new Set(a),c=function(e){!e||s.has(e)||(s.add(e),c(e.parentNode))};a.forEach(c);var u=function(e){!e||l.has(e)||Array.prototype.forEach.call(e.children,(function(e){if(s.has(e))u(e);else try{var t=e.getAttribute(r),a=null!==t&&"false"!==t,l=(Km.get(e)||0)+1,c=(o.get(e)||0)+1;Km.set(e,l),o.set(e,c),i.push(e),1===l&&a&&Ym.set(e,!0),1===c&&e.setAttribute(n,"true"),a||e.setAttribute(r,"true")}catch(t){console.error("aria-hidden: cannot operate on ",e,t)}}))};return u(t),s.clear(),eg++,function(){i.forEach((function(e){var t=Km.get(e)-1,a=o.get(e)-1;Km.set(e,t),o.set(e,a),t||(Ym.has(e)||e.removeAttribute(r),Ym.delete(e)),a||e.removeAttribute(n)})),--eg||(Km=new WeakMap,Km=new WeakMap,Ym=new WeakMap,Jm={})}}(r,a,n,"aria-hidden")):function(){return null}};function rg(e){let t;(0,$t.wB)((()=>ym(e)),(e=>{e?t=ng(e):t&&t()})),(0,$t.hi)((()=>{t&&t()}))}let ag=0;function og(e,t="radix"){if(e)return e;const n=Im({useId:void 0});return $t.Bi?`${t}-${$t.Bi()}`:n.useId?`${t}-${n.useId()}`:`${t}-${++ag}`}function ig(e){const t=(0,jn.KR)(),n=(0,$t.EW)((()=>{var e;return(null==(e=t.value)?void 0:e.width)??0})),r=(0,$t.EW)((()=>{var e;return(null==(e=t.value)?void 0:e.height)??0}));return(0,$t.sV)((()=>{const n=ym(e);if(n){t.value={width:n.offsetWidth,height:n.offsetHeight};const e=new ResizeObserver((e=>{if(!Array.isArray(e)||!e.length)return;const r=e[0];let a,o;if("borderBoxSize"in r){const e=r.borderBoxSize,t=Array.isArray(e)?e[0]:e;a=t.inlineSize,o=t.blockSize}else a=n.offsetWidth,o=n.offsetHeight;t.value={width:a,height:o}}));return e.observe(n,{box:"border-box"}),()=>e.unobserve(n)}t.value=void 0})),{width:n,height:r}}function sg(e,t){const n=(0,jn.KR)(e);return{state:n,dispatch:e=>{n.value=function(e){return t[n.value][e]??n.value}(e)}}}function lg(e){const t=fm("",1e3);return{search:t,handleTypeaheadSearch:(n,r)=>{if(!(null!=e&&e.value||r))return;t.value=t.value+n;const a=(null==e?void 0:e.value)??r,o=Am(),i=a.map((e=>{var t;return{ref:e,textValue:(null==(t=(e.querySelector("[data-item-text]")??e).textContent)?void 0:t.trim())??""}})),s=i.find((e=>e.ref===o)),l=i.map((e=>e.textValue)),c=function(e,t,n){const r=t.length>1&&Array.from(t).every((e=>e===t[0]))?t[0]:t,a=n?e.indexOf(n):-1;let o=function(e,t){return e.map(((n,r)=>e[(t+r)%e.length]))}(e,Math.max(a,0));1===r.length&&(o=o.filter((e=>e!==n)));const i=o.find((e=>e.toLowerCase().startsWith(r.toLowerCase())));return i!==n?i:void 0}(l,t.value,null==s?void 0:s.textValue),u=i.find((e=>e.textValue===c));return u&&u.ref.focus(),null==u?void 0:u.ref},resetTypeahead:()=>{t.value=""}}}const cg=(0,$t.pM)({name:"PrimitiveSlot",inheritAttrs:!1,setup:(e,{attrs:t,slots:n})=>()=>{var e,r;if(!n.default)return null;const a=Tm(n.default()),o=a.findIndex((e=>e.type!==$t.Mw));if(-1===o)return a;const i=a[o];null==(e=i.props)||delete e.ref;const s=i.props?(0,$t.v6)(t,i.props):t;t.class&&null!=(r=i.props)&&r.class&&delete i.props.class;const l=(0,$t.E3)(i,s);for(const e in s)e.startsWith("on")&&(l.props||(l.props={}),l.props[e]=s[e]);return 1===a.length?l:(a[o]=l,a)}}),ug=(0,$t.pM)({name:"Primitive",inheritAttrs:!1,props:{asChild:{type:Boolean,default:!1},as:{type:[String,Object],default:"div"}},setup(e,{attrs:t,slots:n}){const r=e.asChild?"template":e.as;return"string"==typeof r&&["area","img","input"].includes(r)?()=>(0,$t.h)(r,t):"template"!==r?()=>(0,$t.h)(e.as,t,{default:n.default}):()=>(0,$t.h)(cg,t,{default:n.default})}});function dg(){const e=(0,jn.KR)(),t=(0,$t.EW)((()=>{var t,n;return["#text","#comment"].includes(null==(t=e.value)?void 0:t.$el.nodeName)?null==(n=e.value)?void 0:n.$el.nextElementSibling:ym(e)}));return{primitiveElement:e,currentElement:t}}const[pg,hg]=Kf("CollapsibleRoot");Boolean,Boolean,Boolean,Boolean,Boolean;function fg(e){return e&&getComputedStyle(e).animationName||"none"}const mg=(0,$t.pM)({name:"Presence",props:{present:{type:Boolean,required:!0},forceMount:{type:Boolean}},slots:{},setup(e,{slots:t,expose:n}){var r;const{present:a,forceMount:o}=(0,jn.QW)(e),i=(0,jn.KR)(),{isPresent:s}=function(e,t){var n;const r=(0,jn.KR)({}),a=(0,jn.KR)("none"),o=(0,jn.KR)(e),i=e.value?"mounted":"unmounted";let s;const l=(null==(n=t.value)?void 0:n.ownerDocument.defaultView)??wm,{state:c,dispatch:u}=sg(i,{mounted:{UNMOUNT:"unmounted",ANIMATION_OUT:"unmountSuspended"},unmountSuspended:{MOUNT:"mounted",ANIMATION_END:"unmounted"},unmounted:{MOUNT:"mounted"}}),d=e=>{var n;if(sm){const r=new CustomEvent(e,{bubbles:!1,cancelable:!1});null==(n=t.value)||n.dispatchEvent(r)}};(0,$t.wB)(e,(async(e,n)=>{var o;const i=n!==e;if(await(0,$t.dY)(),i){const i=a.value,s=fg(t.value);e?(u("MOUNT"),d("enter"),"none"===s&&d("after-enter")):"none"===s||"none"===(null==(o=r.value)?void 0:o.display)?(u("UNMOUNT"),d("leave"),d("after-leave")):n&&i!==s?(u("ANIMATION_OUT"),d("leave")):(u("UNMOUNT"),d("after-leave"))}}),{immediate:!0});const p=e=>{const n=fg(t.value),r=n.includes(e.animationName),a="mounted"===c.value?"enter":"leave";if(e.target===t.value&&r&&(d(`after-${a}`),u("ANIMATION_END"),!o.value)){const e=t.value.style.animationFillMode;t.value.style.animationFillMode="forwards",s=null==l?void 0:l.setTimeout((()=>{var n;"forwards"===(null==(n=t.value)?void 0:n.style.animationFillMode)&&(t.value.style.animationFillMode=e)}))}e.target===t.value&&"none"===n&&u("ANIMATION_END")},h=e=>{e.target===t.value&&(a.value=fg(t.value))},f=(0,$t.wB)(t,((e,t)=>{e?(r.value=getComputedStyle(e),e.addEventListener("animationstart",h),e.addEventListener("animationcancel",p),e.addEventListener("animationend",p)):(u("ANIMATION_END"),void 0!==s&&(null==l||l.clearTimeout(s)),null==t||t.removeEventListener("animationstart",h),null==t||t.removeEventListener("animationcancel",p),null==t||t.removeEventListener("animationend",p))}),{immediate:!0}),m=(0,$t.wB)(c,(()=>{const e=fg(t.value);a.value="mounted"===c.value?e:"none"}));return(0,$t.hi)((()=>{f(),m()})),{isPresent:(0,$t.EW)((()=>["mounted","unmountSuspended"].includes(c.value)))}}(a,i);n({present:s});let l=t.default({present:s});l=Tm(l||[]);const c=(0,$t.nI)();if(l&&(null==l?void 0:l.length)>1){const e=null!=(r=null==c?void 0:c.parent)&&r.type.name?`<${c.parent.type.name} />`:"component";throw new Error([`Detected an invalid children for \`${e}\` for  \`Presence\` component.`,"","Note: Presence works similarly to `v-if` directly, but it waits for animation/transition to finished before unmounting. So it expect only one direct child of valid VNode type.","You can apply a few solutions:",["Provide a single child element so that `presence` directive attach correctly.","Ensure the first child is an actual element instead of a raw text node or comment node."].map((e=>`  - ${e}`)).join("\n")].join("\n"))}return()=>o.value||a.value||s.value?(0,$t.h)(t.default({present:s})[0],{ref:e=>{const t=ym(e);return typeof(null==t?void 0:t.hasAttribute)>"u"||(null!=t&&t.hasAttribute("data-radix-popper-content-wrapper")?i.value=t.firstElementChild:i.value=t),t}}):null}});Boolean,Boolean;function gg({type:e,defaultValue:t,modelValue:n}){const r=n||t;if(em(e)&&em(n)&&em(t))throw new Error("Either the `type` or the `value` or `default-value` prop must be defined.");if(void 0!==n&&void 0!==t&&typeof n!=typeof t)throw new Error(`Invalid prop \`value\` of value \`${n}\` supplied, should be the same type as the \`defaultValue\` prop, which is \`${t}\`. The \`value\` prop must be:\n  ${"single"===e?"- a string":"multiple"===e?"- an array of strings":"- a string\n- an array of strings"}\n  - \`undefined\``);const a=void 0!==n||void 0!==t;if(e&&a){const r=Array.isArray(n)||Array.isArray(t),a=void 0!==n?"modelValue":"defaultValue",o="modelValue"===a?typeof n:typeof t;if("single"===e&&r)return console.error(`Invalid prop \`${a}\` of type ${o} supplied with type \`single\`. The \`modelValue\` prop must be a string or \`undefined\`.\n    You can remove the \`type\` prop to let the component infer the type from the ${a} prop.`),"multiple";if("multiple"===e&&!r)return console.error(`Invalid prop \`${a}\` of type ${o} supplied with type \`multiple\`. The \`modelValue\` prop must be an array of strings or \`undefined\`.\n    You can remove the \`type\` prop to let the component infer the type from the ${a} prop.`),"single"}return a?Array.isArray(r)?"multiple":"single":e}function vg({type:e,defaultValue:t}){return void 0!==t?t:"single"===e?void 0:[]}const[bg,yg]=Kf("AccordionRoot"),[wg,Og]=(Boolean,Boolean,Boolean,Kf("AccordionItem")),[kg,xg]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("DialogRoot")),Eg=(Boolean,Boolean,Boolean,Boolean,(0,$t.pM)({__name:"Teleport",props:{to:{default:"body"},disabled:{type:Boolean},forceMount:{type:Boolean}},setup(e){const t=xm();return(e,n)=>(0,jn.R1)(t)||e.forceMount?((0,$t.uX)(),(0,$t.Wv)($t.Im,{key:0,to:e.to,disabled:e.disabled},[(0,$t.RG)(e.$slots,"default")],8,["to","disabled"])):(0,$t.Q3)("",!0)}}));function Sg(e,t){const n=t.closest("[data-dismissable-layer]"),r=""===e.dataset.dismissableLayer?e:e.querySelector("[data-dismissable-layer]"),a=Array.from(e.ownerDocument.querySelectorAll("[data-dismissable-layer]"));return!!(n&&r===n||a.indexOf(r)<a.indexOf(n))}Boolean,Boolean;const _g=(0,jn.Kh)({layersRoot:new Set,layersWithOutsidePointerEventsDisabled:new Set,branches:new Set}),Cg=(0,$t.pM)({__name:"DismissableLayer",props:{disableOutsidePointerEvents:{type:Boolean,default:!1},asChild:{type:Boolean},as:{}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","dismiss"],setup(e,{emit:t}){const n=e,r=t,{forwardRef:a,currentElement:o}=Zm(),i=(0,$t.EW)((()=>{var e;return(null==(e=o.value)?void 0:e.ownerDocument)??globalThis.document})),s=(0,$t.EW)((()=>_g.layersRoot)),l=(0,$t.EW)((()=>o.value?Array.from(s.value).indexOf(o.value):-1)),c=(0,$t.EW)((()=>_g.layersWithOutsidePointerEventsDisabled.size>0)),u=(0,$t.EW)((()=>{const e=Array.from(s.value),[t]=[..._g.layersWithOutsidePointerEventsDisabled].slice(-1),n=e.indexOf(t);return l.value>=n})),d=function(e,t){var n;const r=(null==(n=null==t?void 0:t.value)?void 0:n.ownerDocument)??(null==globalThis?void 0:globalThis.document),a=(0,jn.KR)(!1),o=(0,jn.KR)((()=>{}));return(0,$t.nT)((n=>{if(!sm)return;const i=async n=>{const i=n.target;if(null!=t&&t.value){if(Sg(t.value,i))return void(a.value=!1);if(n.target&&!a.value){let t=function(){Yf("dismissableLayer.pointerDownOutside",e,a)};const a={originalEvent:n};"touch"===n.pointerType?(r.removeEventListener("click",o.value),o.value=t,r.addEventListener("click",o.value,{once:!0})):t()}else r.removeEventListener("click",o.value);a.value=!1}},s=window.setTimeout((()=>{r.addEventListener("pointerdown",i)}),0);n((()=>{window.clearTimeout(s),r.removeEventListener("pointerdown",i),r.removeEventListener("click",o.value)}))})),{onPointerDownCapture:()=>a.value=!0}}((async e=>{const t=[..._g.branches].some((t=>null==t?void 0:t.contains(e.target)));!u.value||t||(r("pointerDownOutside",e),r("interactOutside",e),await(0,$t.dY)(),e.defaultPrevented||r("dismiss"))}),o),p=function(e,t){var n;const r=(null==(n=null==t?void 0:t.value)?void 0:n.ownerDocument)??(null==globalThis?void 0:globalThis.document),a=(0,jn.KR)(!1);return(0,$t.nT)((n=>{if(!sm)return;const o=async n=>{null!=t&&t.value&&(await(0,$t.dY)(),t.value&&!Sg(t.value,n.target)&&n.target&&!a.value&&Yf("dismissableLayer.focusOutside",e,{originalEvent:n}))};r.addEventListener("focusin",o),n((()=>r.removeEventListener("focusin",o)))})),{onFocusCapture:()=>a.value=!0,onBlurCapture:()=>a.value=!1}}((e=>{[..._g.branches].some((t=>null==t?void 0:t.contains(e.target)))||(r("focusOutside",e),r("interactOutside",e),e.defaultPrevented||r("dismiss"))}),o);let h;return km("Escape",(e=>{l.value===s.value.size-1&&(r("escapeKeyDown",e),e.defaultPrevented||r("dismiss"))})),(0,$t.nT)((e=>{o.value&&(n.disableOutsidePointerEvents&&(0===_g.layersWithOutsidePointerEventsDisabled.size&&(h=i.value.body.style.pointerEvents,i.value.body.style.pointerEvents="none"),_g.layersWithOutsidePointerEventsDisabled.add(o.value)),s.value.add(o.value),e((()=>{n.disableOutsidePointerEvents&&1===_g.layersWithOutsidePointerEventsDisabled.size&&(i.value.body.style.pointerEvents=h)})))})),(0,$t.nT)((e=>{e((()=>{o.value&&(s.value.delete(o.value),_g.layersWithOutsidePointerEventsDisabled.delete(o.value))}))})),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),{ref:(0,jn.R1)(a),"as-child":e.asChild,as:e.as,"data-dismissable-layer":"",style:(0,Wn.Tr)({pointerEvents:c.value?u.value?"auto":"none":void 0}),onFocusCapture:(0,jn.R1)(p).onFocusCapture,onBlurCapture:(0,jn.R1)(p).onBlurCapture,onPointerdownCapture:(0,jn.R1)(d).onPointerDownCapture},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["as-child","as","style","onFocusCapture","onBlurCapture","onPointerdownCapture"]))}}),Tg=(Boolean,"focusScope.autoFocusOnMount"),Ag="focusScope.autoFocusOnUnmount",Rg={bubbles:!1,cancelable:!0};function Pg(e){const t=[],n=document.createTreeWalker(e,NodeFilter.SHOW_ELEMENT,{acceptNode:e=>{const t="INPUT"===e.tagName&&"hidden"===e.type;return e.disabled||e.hidden||t?NodeFilter.FILTER_SKIP:e.tabIndex>=0?NodeFilter.FILTER_ACCEPT:NodeFilter.FILTER_SKIP}});for(;n.nextNode();)t.push(n.currentNode);return t}function Dg(e,t){for(const n of e)if(!Mg(n,{upTo:t}))return n}function Mg(e,{upTo:t}){if("hidden"===getComputedStyle(e).visibility)return!0;for(;e;){if(void 0!==t&&e===t)return!1;if("none"===getComputedStyle(e).display)return!0;e=e.parentElement}return!1}function $g(e,{select:t=!1}={}){if(e&&e.focus){const n=Am();e.focus({preventScroll:!0}),e!==n&&function(e){return e instanceof HTMLInputElement&&"select"in e}(e)&&t&&e.select()}}const Ng=function(){let e,t=!1;const n=(0,jn.uY)(!0);return(...r)=>(t||(e=n.run((()=>(()=>(0,jn.KR)([]))(...r))),t=!0),e)}();function Ig(e,t){const n=[...e],r=n.indexOf(t);return-1!==r&&n.splice(r,1),n}const Lg=(0,$t.pM)({__name:"FocusScope",props:{loop:{type:Boolean,default:!1},trapped:{type:Boolean,default:!1},asChild:{type:Boolean},as:{}},emits:["mountAutoFocus","unmountAutoFocus"],setup(e,{emit:t}){const n=e,r=t,{currentRef:a,currentElement:o}=Zm(),i=(0,jn.KR)(null),s=function(){const e=Ng();return{add(t){const n=e.value[0];t!==n&&(null==n||n.pause()),e.value=Ig(e.value,t),e.value.unshift(t)},remove(t){var n;e.value=Ig(e.value,t),null==(n=e.value[0])||n.resume()}}}(),l=(0,jn.Kh)({paused:!1,pause(){this.paused=!0},resume(){this.paused=!1}});function c(e){if(!n.loop&&!n.trapped||l.paused)return;const t="Tab"===e.key&&!e.altKey&&!e.ctrlKey&&!e.metaKey,r=Am();if(t&&r){const t=e.currentTarget,[a,o]=function(e){const t=Pg(e);return[Dg(t,e),Dg(t.reverse(),e)]}(t);a&&o?e.shiftKey||r!==o?e.shiftKey&&r===a&&(e.preventDefault(),n.loop&&$g(o,{select:!0})):(e.preventDefault(),n.loop&&$g(a,{select:!0})):r===t&&e.preventDefault()}}return(0,$t.nT)((e=>{if(!sm)return;const t=o.value;if(!n.trapped)return;function r(e){if(l.paused||!t)return;const n=e.target;t.contains(n)?i.value=n:$g(i.value,{select:!0})}function a(e){if(l.paused||!t)return;const n=e.relatedTarget;null!==n&&(t.contains(n)||$g(i.value,{select:!0}))}document.addEventListener("focusin",r),document.addEventListener("focusout",a);const s=new MutationObserver((function(e){t.contains(i.value)||$g(t)}));t&&s.observe(t,{childList:!0,subtree:!0}),e((()=>{document.removeEventListener("focusin",r),document.removeEventListener("focusout",a),s.disconnect()}))})),(0,$t.nT)((async e=>{const t=o.value;if(await(0,$t.dY)(),!t)return;s.add(l);const n=Am();if(!t.contains(n)){const e=new CustomEvent(Tg,Rg);t.addEventListener(Tg,(e=>r("mountAutoFocus",e))),t.dispatchEvent(e),e.defaultPrevented||(function(e,{select:t=!1}={}){const n=Am();for(const r of e)if($g(r,{select:t}),Am()!==n)return!0}(function(e){return e.filter((e=>"A"!==e.tagName))}(Pg(t)),{select:!0}),Am()===n&&$g(t))}e((()=>{t.removeEventListener(Tg,(e=>r("mountAutoFocus",e)));const e=new CustomEvent(Ag,Rg),a=e=>{r("unmountAutoFocus",e)};t.addEventListener(Ag,a),t.dispatchEvent(e),setTimeout((()=>{e.defaultPrevented||$g(n??document.body,{select:!0}),t.removeEventListener(Ag,a),s.remove(l)}),0)}))})),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),{ref_key:"currentRef",ref:a,tabindex:"-1","as-child":e.asChild,as:e.as,onKeydown:c},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["as-child","as"]))}}),Bg=["Enter"," "],Fg=["ArrowUp","PageDown","End"],Qg=["ArrowDown","PageUp","Home",...Fg],Ug={ltr:[...Bg,"ArrowRight"],rtl:[...Bg,"ArrowLeft"]},zg={ltr:["ArrowLeft"],rtl:["ArrowRight"]};function jg(e){return e?"open":"closed"}function Wg(e){return"indeterminate"===e}function Xg(e){return Wg(e)?"indeterminate":e?"checked":"unchecked"}function qg(e){const t=Am();for(const n of e)if(n===t||(n.focus(),Am()!==t))return}function Vg(e){return"mouse"===e.pointerType}const Hg=(0,$t.pM)({__name:"DialogContentImpl",props:{forceMount:{type:Boolean},trapFocus:{type:Boolean},disableOutsidePointerEvents:{type:Boolean},asChild:{type:Boolean},as:{}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","openAutoFocus","closeAutoFocus"],setup(e,{emit:t}){const n=e,r=t,a=kg(),{forwardRef:o,currentElement:i}=Zm();return a.titleId||(a.titleId=og(void 0,"radix-vue-dialog-title")),a.descriptionId||(a.descriptionId=og(void 0,"radix-vue-dialog-description")),(0,$t.sV)((()=>{a.contentElement=i,Am()!==document.body&&(a.triggerElement.value=Am())})),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Lg),{"as-child":"",loop:"",trapped:n.trapFocus,onMountAutoFocus:t[5]||(t[5]=e=>r("openAutoFocus",e)),onUnmountAutoFocus:t[6]||(t[6]=e=>r("closeAutoFocus",e))},{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(Cg),(0,$t.v6)({id:(0,jn.R1)(a).contentId,ref:(0,jn.R1)(o),as:e.as,"as-child":e.asChild,"disable-outside-pointer-events":e.disableOutsidePointerEvents,role:"dialog","aria-describedby":(0,jn.R1)(a).descriptionId,"aria-labelledby":(0,jn.R1)(a).titleId,"data-state":(0,jn.R1)(jg)((0,jn.R1)(a).open.value)},e.$attrs,{onDismiss:t[0]||(t[0]=e=>(0,jn.R1)(a).onOpenChange(!1)),onEscapeKeyDown:t[1]||(t[1]=e=>r("escapeKeyDown",e)),onFocusOutside:t[2]||(t[2]=e=>r("focusOutside",e)),onInteractOutside:t[3]||(t[3]=e=>r("interactOutside",e)),onPointerDownOutside:t[4]||(t[4]=e=>r("pointerDownOutside",e))}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["id","as","as-child","disable-outside-pointer-events","aria-describedby","aria-labelledby","data-state"])])),_:3},8,["trapped"]))}}),Zg=(0,$t.pM)({__name:"DialogContentModal",props:{forceMount:{type:Boolean},trapFocus:{type:Boolean},disableOutsidePointerEvents:{type:Boolean},asChild:{type:Boolean},as:{}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","openAutoFocus","closeAutoFocus"],setup(e,{emit:t}){const n=e,r=t,a=kg(),o=jm(r),{forwardRef:i,currentElement:s}=Zm();return rg(s),(e,t)=>((0,$t.uX)(),(0,$t.Wv)(Hg,(0,$t.v6)({...n,...(0,jn.R1)(o)},{ref:(0,jn.R1)(i),"trap-focus":(0,jn.R1)(a).open.value,"disable-outside-pointer-events":!0,onCloseAutoFocus:t[0]||(t[0]=e=>{var t;e.defaultPrevented||(e.preventDefault(),null==(t=(0,jn.R1)(a).triggerElement.value)||t.focus())}),onPointerDownOutside:t[1]||(t[1]=e=>{const t=e.detail.originalEvent,n=0===t.button&&!0===t.ctrlKey;(2===t.button||n)&&e.preventDefault()}),onFocusOutside:t[2]||(t[2]=e=>{e.preventDefault()})}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["trap-focus"]))}}),Gg=(0,$t.pM)({__name:"DialogContentNonModal",props:{forceMount:{type:Boolean},trapFocus:{type:Boolean},disableOutsidePointerEvents:{type:Boolean},asChild:{type:Boolean},as:{}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","openAutoFocus","closeAutoFocus"],setup(e,{emit:t}){const n=e,r=jm(t);Zm();const a=kg(),o=(0,jn.KR)(!1),i=(0,jn.KR)(!1);return(e,t)=>((0,$t.uX)(),(0,$t.Wv)(Hg,(0,$t.v6)({...n,...(0,jn.R1)(r)},{"trap-focus":!1,"disable-outside-pointer-events":!1,onCloseAutoFocus:t[0]||(t[0]=e=>{var t;e.defaultPrevented||(o.value||null==(t=(0,jn.R1)(a).triggerElement.value)||t.focus(),e.preventDefault()),o.value=!1,i.value=!1}),onInteractOutside:t[1]||(t[1]=e=>{var t;e.defaultPrevented||(o.value=!0,"pointerdown"===e.detail.originalEvent.type&&(i.value=!0));const n=e.target;(null==(t=(0,jn.R1)(a).triggerElement.value)?void 0:t.contains(n))&&e.preventDefault(),"focusin"===e.detail.originalEvent.type&&i.value&&e.preventDefault()})}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),Kg=(Boolean,Boolean,Boolean,Boolean,(0,$t.pM)({__name:"DialogOverlayImpl",props:{asChild:{type:Boolean},as:{}},setup(e){const t=kg();return Fm(!0),Zm(),(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),{as:e.as,"as-child":e.asChild,"data-state":(0,jn.R1)(t).open.value?"open":"closed",style:{"pointer-events":"auto"}},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["as","as-child","data-state"]))}})),[Yg,Jg]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("AlertDialogContent")),[ev,tv]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("AvatarRoot"));Boolean,Boolean,Boolean;const[nv,rv]=Kf("CalendarRoot");Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean;const[av,ov]=Kf("CheckboxRoot"),[iv,sv]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("PopperRoot")),lv=(0,$t.pM)({inheritAttrs:!1,__name:"PopperRoot",setup(e){const t=(0,jn.KR)();return sv({anchor:t,onAnchorChange:e=>t.value=e}),(e,t)=>(0,$t.RG)(e.$slots,"default")}}),cv=(0,$t.pM)({__name:"PopperAnchor",props:{element:{},asChild:{type:Boolean},as:{}},setup(e){const t=e,{forwardRef:n,currentElement:r}=Zm(),a=iv();return(0,$t.nT)((()=>{a.onAnchorChange(t.element??r.value)})),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),{ref:(0,jn.R1)(n),as:e.as,"as-child":e.asChild},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["as","as-child"]))}});function uv(e){return null!==e}function dv(e){return{name:"transformOrigin",options:e,fn(t){var n,r,a;const{placement:o,rects:i,middlewareData:s}=t,l=0!==(null==(n=s.arrow)?void 0:n.centerOffset),c=l?0:e.arrowWidth,u=l?0:e.arrowHeight,[d,p]=pv(o),h={start:"0%",center:"50%",end:"100%"}[p],f=((null==(r=s.arrow)?void 0:r.x)??0)+c/2,m=((null==(a=s.arrow)?void 0:a.y)??0)+u/2;let g="",v="";return"bottom"===d?(g=l?h:`${f}px`,v=-u+"px"):"top"===d?(g=l?h:`${f}px`,v=`${i.floating.height+u}px`):"right"===d?(g=-u+"px",v=l?h:`${m}px`):"left"===d&&(g=`${i.floating.width+u}px`,v=l?h:`${m}px`),{data:{x:g,y:v}}}}}function pv(e){const[t,n="center"]=e.split("-");return[t,n]}const hv={side:"bottom",sideOffset:0,align:"center",alignOffset:0,arrowPadding:0,avoidCollisions:!0,collisionBoundary:()=>[],collisionPadding:0,sticky:"partial",hideWhenDetached:!1,updatePositionStrategy:"optimized",prioritizePosition:!1},[fv,mv]=Kf("PopperContent"),gv=(0,$t.pM)({inheritAttrs:!1,__name:"PopperContent",props:(0,$t.HF)({side:{},sideOffset:{},align:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean},updatePositionStrategy:{},prioritizePosition:{type:Boolean},asChild:{type:Boolean},as:{}},{...hv}),emits:["placed"],setup(e,{emit:t}){const n=e,r=t,a=iv(),{forwardRef:o,currentElement:i}=Zm(),s=(0,jn.KR)(),l=(0,jn.KR)(),{width:c,height:u}=ig(l),d=(0,$t.EW)((()=>n.side+("center"!==n.align?`-${n.align}`:""))),p=(0,$t.EW)((()=>"number"==typeof n.collisionPadding?n.collisionPadding:{top:0,right:0,bottom:0,left:0,...n.collisionPadding})),h=(0,$t.EW)((()=>Array.isArray(n.collisionBoundary)?n.collisionBoundary:[n.collisionBoundary])),f=(0,$t.EW)((()=>({padding:p.value,boundary:h.value.filter(uv),altBoundary:h.value.length>0}))),m=function(){const e=(0,jn.IJ)();return(0,$t.nT)((()=>{var t;e.value=[Th({mainAxis:n.sideOffset+u.value,alignmentAxis:n.alignOffset}),n.prioritizePosition&&n.avoidCollisions&&Rh({...f.value}),n.avoidCollisions&&Ah({mainAxis:!0,crossAxis:!!n.prioritizePosition,limiter:"partial"===n.sticky?Mh():void 0,...f.value}),!n.prioritizePosition&&n.avoidCollisions&&Rh({...f.value}),Ph({...f.value,apply:({elements:e,rects:t,availableWidth:n,availableHeight:r})=>{const{width:a,height:o}=t.reference,i=e.floating.style;i.setProperty("--radix-popper-available-width",`${n}px`),i.setProperty("--radix-popper-available-height",`${r}px`),i.setProperty("--radix-popper-anchor-width",`${a}px`),i.setProperty("--radix-popper-anchor-height",`${o}px`)}}),l.value&&(t={element:l.value,padding:n.arrowPadding},{name:"arrow",options:t,fn(e){const n=$h(Nh(t.element));return null==n?{}:(e=>({name:"arrow",options:e,async fn(t){const{x:n,y:r,placement:a,rects:o,platform:i,elements:s,middlewareData:l}=t,{element:c,padding:u=0}=Rp(e,t)||{};if(null==c)return{};const d=Fp(u),p={x:n,y:r},h=Ip(a),f=$p(h),m=await i.getDimensions(c),g="y"===h,v=g?"top":"left",b=g?"bottom":"right",y=g?"clientHeight":"clientWidth",w=o.reference[f]+o.reference[h]-p[h]-o.floating[f],O=p[h]-o.reference[h],k=await(null==i.getOffsetParent?void 0:i.getOffsetParent(c));let x=k?k[y]:0;x&&await(null==i.isElement?void 0:i.isElement(k))||(x=s.floating[y]||o.floating[f]);const E=w/2-O/2,S=x/2-m[f]/2-1,_=kp(d[v],S),C=kp(d[b],S),T=_,A=x-m[f]-C,R=x/2-m[f]/2+E,P=Ap(T,R,A),D=!l.arrow&&null!=Dp(a)&&R!==P&&o.reference[f]/2-(R<T?_:C)-m[f]/2<0,M=D?R<T?R-T:R-A:0;return{[h]:p[h]+M,data:{[h]:P,centerOffset:R-P-M,...D&&{alignmentOffset:M}},reset:D}}}))({element:n,padding:t.padding}).fn(e)}}),dv({arrowWidth:c.value,arrowHeight:u.value}),n.hideWhenDetached&&Dh({strategy:"referenceHidden",...f.value})]}),{flush:"sync"}),(0,jn.tB)(e)}(),{floatingStyles:g,placement:v,isPositioned:b,middlewareData:y}=Bh(a.anchor,s,{strategy:"fixed",placement:d,whileElementsMounted:(...e)=>Ch(...e,{animationFrame:"always"===n.updatePositionStrategy}),middleware:m}),w=(0,$t.EW)((()=>pv(v.value)[0])),O=(0,$t.EW)((()=>pv(v.value)[1]));(0,$t.p9)((()=>{b.value&&r("placed")}));const k=(0,$t.EW)((()=>{var e;return 0!==(null==(e=y.value.arrow)?void 0:e.centerOffset)})),x=(0,jn.KR)("");(0,$t.nT)((()=>{i.value&&(x.value=window.getComputedStyle(i.value).zIndex)}));const E=(0,$t.EW)((()=>{var e;return(null==(e=y.value.arrow)?void 0:e.x)??0})),S=(0,$t.EW)((()=>{var e;return(null==(e=y.value.arrow)?void 0:e.y)??0}));return mv({placedSide:w,onArrowChange:e=>l.value=e,arrowX:E,arrowY:S,shouldHideArrow:k}),(e,t)=>{var r,a,i;return(0,$t.uX)(),(0,$t.CE)("div",{ref_key:"floatingRef",ref:s,"data-radix-popper-content-wrapper":"",style:(0,Wn.Tr)({...(0,jn.R1)(g),transform:(0,jn.R1)(b)?(0,jn.R1)(g).transform:"translate(0, -200%)",minWidth:"max-content",zIndex:x.value,"--radix-popper-transform-origin":[null==(r=(0,jn.R1)(y).transformOrigin)?void 0:r.x,null==(a=(0,jn.R1)(y).transformOrigin)?void 0:a.y].join(" "),...(null==(i=(0,jn.R1)(y).hide)?void 0:i.referenceHidden)&&{visibility:"hidden",pointerEvents:"none"}})},[(0,$t.bF)((0,jn.R1)(ug),(0,$t.v6)({ref:(0,jn.R1)(o)},e.$attrs,{"as-child":n.asChild,as:e.as,"data-side":w.value,"data-align":O.value,style:{animation:(0,jn.R1)(b)?void 0:"none"}}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["as-child","as","data-side","data-align","style"])],4)}}}),vv=(0,$t.Lk)("polygon",{points:"0,0 30,0 15,10"},null,-1),bv=(0,$t.pM)({__name:"Arrow",props:{width:{default:10},height:{default:5},asChild:{type:Boolean},as:{default:"svg"}},setup(e){const t=e;return Zm(),(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),(0,$t.v6)(t,{width:e.width,height:e.height,viewBox:e.asChild?void 0:"0 0 30 10",preserveAspectRatio:e.asChild?void 0:"none"}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{},(()=>[vv]))])),_:3},16,["width","height","viewBox","preserveAspectRatio"]))}}),yv={top:"bottom",right:"left",bottom:"top",left:"right"},wv=(0,$t.pM)({inheritAttrs:!1,__name:"PopperArrow",props:{width:{},height:{},asChild:{type:Boolean},as:{default:"svg"}},setup(e){const{forwardRef:t}=Zm(),n=fv(),r=(0,$t.EW)((()=>yv[n.placedSide.value]));return(e,a)=>{var o,i,s,l;return(0,$t.uX)(),(0,$t.CE)("span",{ref:e=>{(0,jn.R1)(n).onArrowChange(e)},style:(0,Wn.Tr)({position:"absolute",left:null!=(o=(0,jn.R1)(n).arrowX)&&o.value?`${null==(i=(0,jn.R1)(n).arrowX)?void 0:i.value}px`:void 0,top:null!=(s=(0,jn.R1)(n).arrowY)&&s.value?`${null==(l=(0,jn.R1)(n).arrowY)?void 0:l.value}px`:void 0,[r.value]:0,transformOrigin:{top:"",right:"0 0",bottom:"center 0",left:"100% 0"}[(0,jn.R1)(n).placedSide.value],transform:{top:"translateY(100%)",right:"translateY(50%) rotate(90deg) translateX(-50%)",bottom:"rotate(180deg)",left:"translateY(50%) rotate(-90deg) translateX(50%)"}[(0,jn.R1)(n).placedSide.value],visibility:(0,jn.R1)(n).shouldHideArrow.value?"hidden":void 0})},[(0,$t.bF)(bv,(0,$t.v6)(e.$attrs,{ref:(0,jn.R1)(t),style:{display:"block"},as:e.as,"as-child":e.asChild,width:e.width,height:e.height}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["as","as-child","width","height"])],4)}}}),Ov=(0,$t.pM)({__name:"VisuallyHidden",props:{asChild:{type:Boolean},as:{default:"span"}},setup:e=>(Zm(),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),{as:e.as,"as-child":e.asChild,style:{position:"absolute",border:0,width:"1px",display:"inline-block",height:"1px",padding:0,margin:"-1px",overflow:"hidden",clip:"rect(0, 0, 0, 0)",whiteSpace:"nowrap",wordWrap:"normal"}},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["as","as-child"])))}),[kv,xv]=(Boolean,Boolean,Kf("CollectionProvider"));const Ev=(0,$t.pM)({name:"CollectionSlot",setup(e,{slots:t}){const n=kv(),{primitiveElement:r,currentElement:a}=dg();return(0,$t.wB)(a,(()=>{n.collectionRef.value=a.value})),()=>(0,$t.h)(cg,{ref:r},t)}}),Sv=(0,$t.pM)({name:"CollectionItem",inheritAttrs:!1,props:{value:{validator:()=>!0}},setup(e,{slots:t,attrs:n}){const r=kv(),{primitiveElement:a,currentElement:o}=dg();return(0,$t.nT)((t=>{if(o.value){const n=(0,jn.IG)(o.value);r.itemMap.value.set(n,{ref:o.value,value:e.value}),t((()=>r.itemMap.value.delete(n)))}})),()=>(0,$t.h)(cg,{...n,[r.attrName]:"",ref:a},t)}});function _v(e){const t=e??kv();return{getItems:()=>{const e=t.collectionRef.value;if(!e)return[];const n=Array.from(e.querySelectorAll(`[${t.attrName}]`));return Array.from(t.itemMap.value.values()).sort(((e,t)=>n.indexOf(e.ref)-n.indexOf(t.ref)))}}}const[Cv,Tv]=Kf("ComboboxRoot"),[Av,Rv]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("ComboboxGroup")),[Pv,Dv]=(Boolean,Boolean,Kf("ComboboxContent"));Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean;Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean;const[Mv,$v]=Kf("ComboboxItem"),Nv=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,(0,$t.pM)({__name:"MenuAnchor",props:{element:{},asChild:{type:Boolean},as:{}},setup(e){const t=e;return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(cv),(0,Wn._B)((0,$t.Ng)(t)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}})),Iv=(0,$t.pM)({__name:"MenuArrow",props:{width:{},height:{},asChild:{type:Boolean},as:{}},setup(e){const t=e;return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(wv),(0,Wn._B)((0,$t.Ng)(t)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),Lv=om((function(){const e=(0,jn.KR)(!1);return(0,$t.sV)((()=>{Om("keydown",(()=>{e.value=!0}),{capture:!0,passive:!0}),Om(["pointerdown","pointermove"],(()=>{e.value=!1}),{capture:!0,passive:!0})})),e})),[Bv,Fv]=Kf(["MenuRoot","MenuSub"],"MenuContext"),[Qv,Uv]=Kf("MenuRoot"),zv=(0,$t.pM)({__name:"MenuRoot",props:{open:{type:Boolean,default:!1},dir:{},modal:{type:Boolean,default:!0}},emits:["update:open"],setup(e,{emit:t}){const n=e,r=t,{modal:a,dir:o}=(0,jn.QW)(n),i=zm(o),s=Cm(n,"open",r),l=(0,jn.KR)(),c=Lv();return Fv({open:s,onOpenChange:e=>{s.value=e},content:l,onContentChange:e=>{l.value=e}}),Uv({onClose:()=>{s.value=!1},isUsingKeyboardRef:c,dir:i,modal:a}),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(lv),null,{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3}))}}),jv={bubbles:!1,cancelable:!0},Wv={ArrowLeft:"prev",ArrowUp:"prev",ArrowRight:"next",ArrowDown:"next",PageUp:"first",Home:"first",PageDown:"last",End:"last"};function Xv(e,t=!1){const n=Am();for(const r of e)if(r===n||(r.focus({preventScroll:t}),Am()!==n))return}const[qv,Vv]=Kf("RovingFocusGroup"),Hv=(0,$t.pM)({__name:"RovingFocusGroup",props:{orientation:{default:void 0},dir:{},loop:{type:Boolean,default:!1},currentTabStopId:{},defaultCurrentTabStopId:{},preventScrollOnEntryFocus:{type:Boolean,default:!1},asChild:{type:Boolean},as:{}},emits:["entryFocus","update:currentTabStopId"],setup(e,{expose:t,emit:n}){const r=e,a=n,{loop:o,orientation:i,dir:s}=(0,jn.QW)(r),l=zm(s),c=Cm(r,"currentTabStopId",a,{defaultValue:r.defaultCurrentTabStopId,passive:void 0===r.currentTabStopId}),u=(0,jn.KR)(!1),d=(0,jn.KR)(!1),p=(0,jn.KR)(0),{getItems:h}=function(e="data-radix-vue-collection-item"){const t=(0,jn.KR)(new Map),n=(0,jn.KR)(),r=xv({collectionRef:n,itemMap:t,attrName:e}),{getItems:a}=_v(r);return{getItems:a,reactiveItems:(0,$t.EW)((()=>Array.from(r.itemMap.value.values()))),itemMapSize:(0,$t.EW)((()=>r.itemMap.value.size))}}();function f(e){const t=!d.value;if(e.currentTarget&&e.target===e.currentTarget&&t&&!u.value){const t=new CustomEvent("rovingFocusGroup.onEntryFocus",jv);if(e.currentTarget.dispatchEvent(t),a("entryFocus",t),!t.defaultPrevented){const e=h().map((e=>e.ref)).filter((e=>""!==e.dataset.disabled)),t=e.find((e=>"true"===e.getAttribute("data-active"))),n=e.find((e=>e.id===c.value));Xv([t,n,...e].filter(Boolean),r.preventScrollOnEntryFocus)}}d.value=!1}function m(){setTimeout((()=>{d.value=!1}),1)}return t({getItems:h}),Vv({loop:o,dir:l,orientation:i,currentTabStopId:c,onItemFocus:e=>{c.value=e},onItemShiftTab:()=>{u.value=!0},onFocusableItemAdd:()=>{p.value++},onFocusableItemRemove:()=>{p.value--}}),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Ev),null,{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(ug),{tabindex:u.value||0===p.value?-1:0,"data-orientation":(0,jn.R1)(i),as:e.as,"as-child":e.asChild,dir:(0,jn.R1)(l),style:{outline:"none"},onMousedown:t[0]||(t[0]=e=>d.value=!0),onMouseup:m,onFocus:f,onBlur:t[1]||(t[1]=e=>u.value=!1)},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["tabindex","data-orientation","as","as-child","dir"])])),_:3}))}}),Zv=(0,$t.pM)({__name:"RovingFocusItem",props:{tabStopId:{},focusable:{type:Boolean,default:!0},active:{type:Boolean,default:!0},allowShiftKey:{type:Boolean},asChild:{type:Boolean},as:{default:"span"}},setup(e){const t=e,n=qv(),r=(0,$t.EW)((()=>t.tabStopId||og())),a=(0,$t.EW)((()=>n.currentTabStopId.value===r.value)),{getItems:o}=_v();function i(e){if("Tab"===e.key&&e.shiftKey)return void n.onItemShiftTab();if(e.target!==e.currentTarget)return;const r=function(e,t,n){const r=function(e,t){return"rtl"!==t?e:"ArrowLeft"===e?"ArrowRight":"ArrowRight"===e?"ArrowLeft":e}(e.key,n);if(!("vertical"===t&&["ArrowLeft","ArrowRight"].includes(r)||"horizontal"===t&&["ArrowUp","ArrowDown"].includes(r)))return Wv[r]}(e,n.orientation.value,n.dir.value);if(void 0!==r){if(e.metaKey||e.ctrlKey||e.altKey||!t.allowShiftKey&&e.shiftKey)return;e.preventDefault();let a=[...o().map((e=>e.ref)).filter((e=>""!==e.dataset.disabled))];if("last"===r)a.reverse();else if("prev"===r||"next"===r){"prev"===r&&a.reverse();const t=a.indexOf(e.currentTarget);a=n.loop.value?function(e,t){return e.map(((n,r)=>e[(t+r)%e.length]))}(a,t+1):a.slice(t+1)}(0,$t.dY)((()=>Xv(a)))}}return(0,$t.sV)((()=>{t.focusable&&n.onFocusableItemAdd()})),(0,$t.hi)((()=>{t.focusable&&n.onFocusableItemRemove()})),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Sv),null,{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(ug),{tabindex:a.value?0:-1,"data-orientation":(0,jn.R1)(n).orientation.value,"data-active":e.active,"data-disabled":e.focusable?void 0:"",as:e.as,"as-child":e.asChild,onMousedown:t[0]||(t[0]=t=>{e.focusable?(0,jn.R1)(n).onItemFocus(r.value):t.preventDefault()}),onFocus:t[1]||(t[1]=e=>(0,jn.R1)(n).onItemFocus(r.value)),onKeydown:i},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["tabindex","data-orientation","data-active","data-disabled","as","as-child"])])),_:3}))}}),[Gv,Kv]=Kf("MenuContent"),Yv=(0,$t.pM)({__name:"MenuContentImpl",props:(0,$t.HF)({loop:{type:Boolean},disableOutsidePointerEvents:{type:Boolean},disableOutsideScroll:{type:Boolean},trapFocus:{type:Boolean},side:{},sideOffset:{},align:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean},updatePositionStrategy:{},prioritizePosition:{type:Boolean},asChild:{type:Boolean},as:{}},{...hv}),emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","entryFocus","openAutoFocus","closeAutoFocus","dismiss"],setup(e,{emit:t}){const n=e,r=t,a=Bv(),o=Qv(),{trapFocus:i,disableOutsidePointerEvents:s,loop:l}=(0,jn.QW)(n);Xm(),Fm(s.value);const c=(0,jn.KR)(""),u=(0,jn.KR)(0),d=(0,jn.KR)(0),p=(0,jn.KR)(null),h=(0,jn.KR)("right"),f=(0,jn.KR)(0),m=(0,jn.KR)(null),{createCollection:g}=Um(),{forwardRef:v,currentElement:b}=Zm(),y=g(b);(0,$t.wB)(b,(e=>{a.onContentChange(e)}));const{handleTypeaheadSearch:w}=lg(y);function O(e){var t,n;return h.value===(null==(t=p.value)?void 0:t.side)&&function(e,t){return!!t&&function(e,t){const{x:n,y:r}=e;let a=!1;for(let e=0,o=t.length-1;e<t.length;o=e++){const i=t[e].x,s=t[e].y,l=t[o].x,c=t[o].y;s>r!=c>r&&n<(l-i)*(r-s)/(c-s)+i&&(a=!a)}return a}({x:e.clientX,y:e.clientY},t)}(e,null==(n=p.value)?void 0:n.area)}async function k(e){var t;r("openAutoFocus",e),!e.defaultPrevented&&(e.preventDefault(),null==(t=b.value)||t.focus({preventScroll:!0}))}function x(e){if(e.defaultPrevented)return;const t=e.target.closest("[data-radix-menu-content]")===e.currentTarget,n=e.ctrlKey||e.altKey||e.metaKey,r=1===e.key.length,a=Pm(e,Am(),b.value,{loop:l.value,arrowKeyOptions:"vertical",dir:null==o?void 0:o.dir.value,focus:!0,attributeName:"[data-radix-vue-collection-item]:not([data-disabled])"});if(a)return null==a?void 0:a.focus();if("Space"===e.code||(t&&("Tab"===e.key&&e.preventDefault(),!n&&r&&w(e.key)),e.target!==b.value)||!Qg.includes(e.key))return;e.preventDefault();const i=y.value;Fg.includes(e.key)&&i.reverse(),qg(i)}function E(e){var t,n;null!=(n=null==(t=null==e?void 0:e.currentTarget)?void 0:t.contains)&&n.call(t,e.target)||(window.clearTimeout(u.value),c.value="")}function S(e){var t;if(!Vg(e))return;const n=e.target,r=f.value!==e.clientX;if(null!=(t=null==e?void 0:e.currentTarget)&&t.contains(n)&&r){const t=e.clientX>f.value?"right":"left";h.value=t,f.value=e.clientX}}return(0,$t.hi)((()=>{window.clearTimeout(u.value)})),Kv({onItemEnter:e=>!!O(e),onItemLeave:e=>{var t;O(e)||(null==(t=b.value)||t.focus(),m.value=null)},onTriggerLeave:e=>!!O(e),searchRef:c,pointerGraceTimerRef:d,onPointerGraceIntentChange:e=>{p.value=e}}),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Lg),{"as-child":"",trapped:(0,jn.R1)(i),onMountAutoFocus:k,onUnmountAutoFocus:t[7]||(t[7]=e=>r("closeAutoFocus",e))},{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(Cg),{"as-child":"","disable-outside-pointer-events":(0,jn.R1)(s),onEscapeKeyDown:t[2]||(t[2]=e=>r("escapeKeyDown",e)),onPointerDownOutside:t[3]||(t[3]=e=>r("pointerDownOutside",e)),onFocusOutside:t[4]||(t[4]=e=>r("focusOutside",e)),onInteractOutside:t[5]||(t[5]=e=>r("interactOutside",e)),onDismiss:t[6]||(t[6]=e=>r("dismiss"))},{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(Hv),{"current-tab-stop-id":m.value,"onUpdate:currentTabStopId":t[0]||(t[0]=e=>m.value=e),"as-child":"",orientation:"vertical",dir:(0,jn.R1)(o).dir.value,loop:(0,jn.R1)(l),onEntryFocus:t[1]||(t[1]=e=>{r("entryFocus",e),(0,jn.R1)(o).isUsingKeyboardRef.value||e.preventDefault()})},{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(gv),{ref:(0,jn.R1)(v),role:"menu",as:e.as,"as-child":e.asChild,"aria-orientation":"vertical","data-radix-menu-content":"","data-state":(0,jn.R1)(jg)((0,jn.R1)(a).open.value),dir:(0,jn.R1)(o).dir.value,side:e.side,"side-offset":e.sideOffset,align:e.align,"align-offset":e.alignOffset,"avoid-collisions":e.avoidCollisions,"collision-boundary":e.collisionBoundary,"collision-padding":e.collisionPadding,"arrow-padding":e.arrowPadding,"prioritize-position":e.prioritizePosition,sticky:e.sticky,"hide-when-detached":e.hideWhenDetached,onKeydown:x,onBlur:E,onPointermove:S},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["as","as-child","data-state","dir","side","side-offset","align","align-offset","avoid-collisions","collision-boundary","collision-padding","arrow-padding","prioritize-position","sticky","hide-when-detached"])])),_:3},8,["current-tab-stop-id","dir","loop"])])),_:3},8,["disable-outside-pointer-events"])])),_:3},8,["trapped"]))}}),Jv=(0,$t.pM)({inheritAttrs:!1,__name:"MenuItemImpl",props:{disabled:{type:Boolean},textValue:{},asChild:{type:Boolean},as:{}},setup(e){const t=e,n=Gv(),{forwardRef:r}=Zm(),a=(0,jn.KR)(!1);async function o(e){if(!e.defaultPrevented&&Vg(e))if(t.disabled)n.onItemLeave(e);else if(!n.onItemEnter(e)){const t=e.currentTarget;null==t||t.focus({preventScroll:!0})}}async function i(e){await(0,$t.dY)(),!e.defaultPrevented&&Vg(e)&&n.onItemLeave(e)}return(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Sv),{value:{textValue:e.textValue}},{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(ug),(0,$t.v6)({ref:(0,jn.R1)(r),role:"menuitem",tabindex:"-1"},e.$attrs,{as:e.as,"as-child":e.asChild,"data-radix-vue-collection-item":"","aria-disabled":e.disabled||void 0,"data-disabled":e.disabled?"":void 0,"data-highlighted":a.value?"":void 0,onPointermove:o,onPointerleave:i,onFocus:t[0]||(t[0]=async t=>{await(0,$t.dY)(),!t.defaultPrevented&&!e.disabled&&(a.value=!0)}),onBlur:t[1]||(t[1]=async e=>{await(0,$t.dY)(),!e.defaultPrevented&&(a.value=!1)})}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["as","as-child","aria-disabled","data-disabled","data-highlighted"])])),_:3},8,["value"]))}}),eb=(0,$t.pM)({__name:"MenuItem",props:{disabled:{type:Boolean},textValue:{},asChild:{type:Boolean},as:{}},emits:["select"],setup(e,{emit:t}){const n=e,r=t,{forwardRef:a,currentElement:o}=Zm(),i=Qv(),s=Gv(),l=(0,jn.KR)(!1);async function c(){const e=o.value;if(!n.disabled&&e){const e=new CustomEvent("menu.itemSelect",{bubbles:!0,cancelable:!0});r("select",e),await(0,$t.dY)(),e.defaultPrevented?l.value=!1:i.onClose()}}return(e,t)=>((0,$t.uX)(),(0,$t.Wv)(Jv,(0,$t.v6)(n,{ref:(0,jn.R1)(a),onClick:c,onPointerdown:t[0]||(t[0]=()=>{l.value=!0}),onPointerup:t[1]||(t[1]=async e=>{var t;await(0,$t.dY)(),!e.defaultPrevented&&(l.value||null==(t=e.currentTarget)||t.click())}),onKeydown:t[2]||(t[2]=async t=>{const n=""!==(0,jn.R1)(s).searchRef.value;e.disabled||n&&" "===t.key||(0,jn.R1)(Bg).includes(t.key)&&(t.currentTarget.click(),t.preventDefault())})}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),[tb,nb]=Kf(["MenuCheckboxItem","MenuRadioItem"],"MenuItemIndicatorContext"),rb=(0,$t.pM)({__name:"MenuItemIndicator",props:{forceMount:{type:Boolean},asChild:{type:Boolean},as:{default:"span"}},setup(e){const t=tb({checked:(0,jn.KR)(!1)});return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(mg),{present:e.forceMount||(0,jn.R1)(Wg)((0,jn.R1)(t).checked.value)||!0===(0,jn.R1)(t).checked.value},{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(ug),{as:e.as,"as-child":e.asChild,"data-state":(0,jn.R1)(Xg)((0,jn.R1)(t).checked.value)},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["as","as-child","data-state"])])),_:3},8,["present"]))}}),ab=(0,$t.pM)({__name:"MenuCheckboxItem",props:{checked:{type:[Boolean,String],default:!1},disabled:{type:Boolean},textValue:{},asChild:{type:Boolean},as:{}},emits:["select","update:checked"],setup(e,{emit:t}){const n=e,r=t,a=Cm(n,"checked",r);return nb({checked:a}),(e,t)=>((0,$t.uX)(),(0,$t.Wv)(eb,(0,$t.v6)({role:"menuitemcheckbox"},n,{"aria-checked":(0,jn.R1)(Wg)((0,jn.R1)(a))?"mixed":(0,jn.R1)(a),"data-state":(0,jn.R1)(Xg)((0,jn.R1)(a)),onSelect:t[0]||(t[0]=async e=>{r("select",e),(0,jn.R1)(Wg)((0,jn.R1)(a))?a.value=!0:a.value=!(0,jn.R1)(a)})}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{checked:(0,jn.R1)(a)})])),_:3},16,["aria-checked","data-state"]))}}),ob=(0,$t.pM)({__name:"MenuRootContentModal",props:{loop:{type:Boolean},side:{},sideOffset:{},align:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean},updatePositionStrategy:{},prioritizePosition:{type:Boolean},asChild:{type:Boolean},as:{}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","entryFocus","openAutoFocus","closeAutoFocus"],setup(e,{emit:t}){const n=t,r=Hm(e,n),a=Bv(),{forwardRef:o,currentElement:i}=Zm();return rg(i),(e,t)=>((0,$t.uX)(),(0,$t.Wv)(Yv,(0,$t.v6)((0,jn.R1)(r),{ref:(0,jn.R1)(o),"trap-focus":(0,jn.R1)(a).open.value,"disable-outside-pointer-events":(0,jn.R1)(a).open.value,"disable-outside-scroll":!0,onDismiss:t[0]||(t[0]=e=>(0,jn.R1)(a).onOpenChange(!1)),onFocusOutside:t[1]||(t[1]=(0,Rr.D$)((e=>n("focusOutside",e)),["prevent"]))}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["trap-focus","disable-outside-pointer-events"]))}}),ib=(0,$t.pM)({__name:"MenuRootContentNonModal",props:{loop:{type:Boolean},side:{},sideOffset:{},align:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean},updatePositionStrategy:{},prioritizePosition:{type:Boolean},asChild:{type:Boolean},as:{}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","entryFocus","openAutoFocus","closeAutoFocus"],setup(e,{emit:t}){const n=Hm(e,t),r=Bv();return(e,t)=>((0,$t.uX)(),(0,$t.Wv)(Yv,(0,$t.v6)((0,jn.R1)(n),{"trap-focus":!1,"disable-outside-pointer-events":!1,"disable-outside-scroll":!1,onDismiss:t[0]||(t[0]=e=>(0,jn.R1)(r).onOpenChange(!1))}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),sb=(0,$t.pM)({__name:"MenuContent",props:{forceMount:{type:Boolean},loop:{type:Boolean},side:{},sideOffset:{},align:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean},updatePositionStrategy:{},prioritizePosition:{type:Boolean},asChild:{type:Boolean},as:{}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","entryFocus","openAutoFocus","closeAutoFocus"],setup(e,{emit:t}){const n=Hm(e,t),r=Bv(),a=Qv();return(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(mg),{present:e.forceMount||(0,jn.R1)(r).open.value},{default:(0,$t.k6)((()=>[(0,jn.R1)(a).modal.value?((0,$t.uX)(),(0,$t.Wv)(ob,(0,Wn._B)((0,$t.v6)({key:0},{...e.$attrs,...(0,jn.R1)(n)})),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16)):((0,$t.uX)(),(0,$t.Wv)(ib,(0,Wn._B)((0,$t.v6)({key:1},{...e.$attrs,...(0,jn.R1)(n)})),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))])),_:3},8,["present"]))}}),lb=(0,$t.pM)({__name:"MenuGroup",props:{asChild:{type:Boolean},as:{}},setup(e){const t=e;return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),(0,$t.v6)({role:"group"},t),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),cb=(0,$t.pM)({__name:"MenuLabel",props:{asChild:{type:Boolean},as:{default:"div"}},setup(e){const t=e;return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),(0,Wn._B)((0,$t.Ng)(t)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),ub=(0,$t.pM)({__name:"MenuPortal",props:{to:{},disabled:{type:Boolean},forceMount:{type:Boolean}},setup(e){const t=e;return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Eg),(0,Wn._B)((0,$t.Ng)(t)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),[db,pb]=Kf("MenuRadioGroup"),hb=(0,$t.pM)({__name:"MenuRadioGroup",props:{modelValue:{default:""},asChild:{type:Boolean},as:{}},emits:["update:modelValue"],setup(e,{emit:t}){const n=e,r=Cm(n,"modelValue",t);return pb({modelValue:r,onValueChange:e=>{r.value=e}}),(e,t)=>((0,$t.uX)(),(0,$t.Wv)(lb,(0,Wn._B)((0,$t.Ng)(n)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{modelValue:(0,jn.R1)(r)})])),_:3},16))}}),fb=(0,$t.pM)({__name:"MenuRadioItem",props:{value:{},disabled:{type:Boolean},textValue:{},asChild:{type:Boolean},as:{}},emits:["select"],setup(e,{emit:t}){const n=e,r=t,{value:a}=(0,jn.QW)(n),o=db(),i=(0,$t.EW)((()=>o.modelValue.value===(null==a?void 0:a.value)));return nb({checked:i}),(e,t)=>((0,$t.uX)(),(0,$t.Wv)(eb,(0,$t.v6)({role:"menuitemradio"},n,{"aria-checked":i.value,"data-state":(0,jn.R1)(Xg)(i.value),onSelect:t[0]||(t[0]=async e=>{r("select",e),(0,jn.R1)(o).onValueChange((0,jn.R1)(a))})}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["aria-checked","data-state"]))}}),mb=(0,$t.pM)({__name:"MenuSeparator",props:{asChild:{type:Boolean},as:{}},setup(e){const t=e;return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),(0,$t.v6)(t,{role:"separator","aria-orientation":"horizontal"}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),[gb,vb]=Kf("MenuSub"),bb=(0,$t.pM)({__name:"MenuSub",props:{open:{type:Boolean,default:void 0}},emits:["update:open"],setup(e,{emit:t}){const n=Cm(e,"open",t,{defaultValue:!1,passive:void 0===e.open}),r=Bv(),a=(0,jn.KR)(),o=(0,jn.KR)();return(0,$t.nT)((e=>{!1===(null==r?void 0:r.open.value)&&(n.value=!1),e((()=>n.value=!1))})),Fv({open:n,onOpenChange:e=>{n.value=e},content:o,onContentChange:e=>{o.value=e}}),vb({triggerId:"",contentId:"",trigger:a,onTriggerChange:e=>{a.value=e}}),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(lv),null,{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3}))}}),yb=(0,$t.pM)({__name:"MenuSubContent",props:{forceMount:{type:Boolean},loop:{type:Boolean},sideOffset:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean},updatePositionStrategy:{},prioritizePosition:{type:Boolean,default:!0},asChild:{type:Boolean},as:{}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","entryFocus","openAutoFocus","closeAutoFocus"],setup(e,{emit:t}){const n=Hm(e,t),r=Bv(),a=Qv(),o=gb(),{forwardRef:i,currentElement:s}=Zm();return o.contentId||(o.contentId=og(void 0,"radix-vue-menu-sub-content")),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(mg),{present:e.forceMount||(0,jn.R1)(r).open.value},{default:(0,$t.k6)((()=>[(0,$t.bF)(Yv,(0,$t.v6)((0,jn.R1)(n),{id:(0,jn.R1)(o).contentId,ref:(0,jn.R1)(i),"aria-labelledby":(0,jn.R1)(o).triggerId,align:"start",side:"rtl"===(0,jn.R1)(a).dir.value?"left":"right","disable-outside-pointer-events":!1,"disable-outside-scroll":!1,"trap-focus":!1,onOpenAutoFocus:t[0]||(t[0]=(0,Rr.D$)((e=>{var t;(0,jn.R1)(a).isUsingKeyboardRef.value&&(null==(t=(0,jn.R1)(s))||t.focus())}),["prevent"])),onCloseAutoFocus:t[1]||(t[1]=(0,Rr.D$)((()=>{}),["prevent"])),onFocusOutside:t[2]||(t[2]=e=>{e.defaultPrevented||e.target!==(0,jn.R1)(o).trigger.value&&(0,jn.R1)(r).onOpenChange(!1)}),onEscapeKeyDown:t[3]||(t[3]=e=>{(0,jn.R1)(a).onClose(),e.preventDefault()}),onKeydown:t[4]||(t[4]=e=>{var t,n;const i=null==(t=e.currentTarget)?void 0:t.contains(e.target),s=(0,jn.R1)(zg)[(0,jn.R1)(a).dir.value].includes(e.key);i&&s&&((0,jn.R1)(r).onOpenChange(!1),null==(n=(0,jn.R1)(o).trigger.value)||n.focus(),e.preventDefault())})}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["id","aria-labelledby","side"])])),_:3},8,["present"]))}}),wb=(0,$t.pM)({__name:"MenuSubTrigger",props:{disabled:{type:Boolean},textValue:{},asChild:{type:Boolean},as:{}},setup(e){const t=e,n=Bv(),r=Qv(),a=gb(),o=Gv(),i=(0,jn.KR)(null);function s(){i.value&&window.clearTimeout(i.value),i.value=null}function l(e){!Vg(e)||o.onItemEnter(e)||!t.disabled&&!n.open.value&&!i.value&&(o.onPointerGraceIntentChange(null),i.value=window.setTimeout((()=>{n.onOpenChange(!0),s()}),100))}async function c(e){var t,r;if(!Vg(e))return;s();const a=null==(t=n.content.value)?void 0:t.getBoundingClientRect();if(null!=a&&a.width){const t=null==(r=n.content.value)?void 0:r.dataset.side,i="right"===t,s=i?-5:5,l=a[i?"left":"right"],c=a[i?"right":"left"];o.onPointerGraceIntentChange({area:[{x:e.clientX+s,y:e.clientY},{x:l,y:a.top},{x:c,y:a.top},{x:c,y:a.bottom},{x:l,y:a.bottom}],side:t}),window.clearTimeout(o.pointerGraceTimerRef.value),o.pointerGraceTimerRef.value=window.setTimeout((()=>o.onPointerGraceIntentChange(null)),300)}else{if(o.onTriggerLeave(e))return;o.onPointerGraceIntentChange(null)}}async function u(e){var a;const i=""!==o.searchRef.value;t.disabled||i&&" "===e.key||Ug[r.dir.value].includes(e.key)&&(n.onOpenChange(!0),await(0,$t.dY)(),null==(a=n.content.value)||a.focus(),e.preventDefault())}return a.triggerId||(a.triggerId=og(void 0,"radix-vue-menu-sub-trigger")),(0,$t.hi)((()=>{s()})),(e,r)=>((0,$t.uX)(),(0,$t.Wv)(Nv,{"as-child":""},{default:(0,$t.k6)((()=>[(0,$t.bF)(Jv,(0,$t.v6)(t,{id:(0,jn.R1)(a).triggerId,ref:e=>{var t;null==(t=(0,jn.R1)(a))||t.onTriggerChange(null==e?void 0:e.$el)},"aria-haspopup":"menu","aria-expanded":(0,jn.R1)(n).open.value,"aria-controls":(0,jn.R1)(a).contentId,"data-state":(0,jn.R1)(jg)((0,jn.R1)(n).open.value),onClick:r[0]||(r[0]=async e=>{t.disabled||e.defaultPrevented||(e.currentTarget.focus(),(0,jn.R1)(n).open.value||(0,jn.R1)(n).onOpenChange(!0))}),onPointermove:l,onPointerleave:c,onKeydown:u}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["id","aria-expanded","aria-controls","data-state"])])),_:3}))}}),[Ob,kb]=Kf("ContextMenuRoot");Boolean;function xb(e){return"mouse"!==e.pointerType}Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean;function Eb(e){return e==="ArrowRight"||e==="ArrowLeft"}function Sb(e){return!Number.isNaN(Number.parseInt(e))}function _b(e){const t={ALT:"Alt",ARROW_DOWN:"ArrowDown",ARROW_LEFT:"ArrowLeft",ARROW_RIGHT:"ArrowRight",ARROW_UP:"ArrowUp",BACKSPACE:"Backspace",CAPS_LOCK:"CapsLock",CONTROL:"Control",DELETE:"Delete",END:"End",ENTER:"Enter",ESCAPE:"Escape",F1:"F1",F10:"F10",F11:"F11",F12:"F12",F2:"F2",F3:"F3",F4:"F4",F5:"F5",F6:"F6",F7:"F7",F8:"F8",F9:"F9",HOME:"Home",META:"Meta",PAGE_DOWN:"PageDown",PAGE_UP:"PageUp",SHIFT:"Shift",SPACE:" ",TAB:"Tab",CTRL:"Control",ASTERISK:"*",SPACE_CODE:"Space"};return!(![t.ENTER,t.ARROW_UP,t.ARROW_DOWN,t.ARROW_LEFT,t.ARROW_RIGHT,t.BACKSPACE,t.SPACE,"a","A","p","P"].includes(e)&&!Sb(e))}const[Cb,Tb]=Kf("DateFieldRoot");function Ab(e){return{role:"spinbutton",contenteditable:!0,tabindex:e.disabled?void 0:0,spellcheck:!1,inputmode:"numeric",autocorrect:"off",enterkeyhint:"next",style:"caret-color: transparent;"}}Boolean,Boolean,Boolean,Boolean,Boolean;const Rb={day:{attrs:function(e){const{segmentValues:t,placeholder:n}=e,r=null===t.day,a=t.day?n.set({day:t.day}):n,o=a.day,i=Hf(a),s=r?"Empty":`${o}`;return{...Ab(e),"aria-label":"day,","aria-valuemin":1,"aria-valuemax":i,"aria-valuenow":o,"aria-valuetext":s,"data-placeholder":r?"":void 0}}},month:{attrs:function(e){const{segmentValues:t,placeholder:n,formatter:r}=e,a=null===t.month,o=t.month?n.set({month:t.month}):n,i=o.month,s=a?"Empty":`${i} - ${r.fullMonth(qf(o))}`;return{...Ab(e),"aria-label":"month, ",contenteditable:!0,"aria-valuemin":1,"aria-valuemax":12,"aria-valuenow":i,"aria-valuetext":s,"data-placeholder":a?"":void 0}}},year:{attrs:function(e){const{segmentValues:t,placeholder:n}=e,r=null===t.year,a=(t.year?n.set({year:t.year}):n).year,o=r?"Empty":`${a}`;return{...Ab(e),"aria-label":"year, ","aria-valuemin":1,"aria-valuemax":9999,"aria-valuenow":a,"aria-valuetext":o,"data-placeholder":r?"":void 0}}},hour:{attrs:function(e){const{segmentValues:t,hourCycle:n,placeholder:r}=e;if(!("hour"in t)||!("hour"in r))return{};const a=null===t.hour,o=12===n?1:0,i=12===n?12:23,s=(t.hour?r.set({hour:t.hour}):r).hour,l=a?"Empty":`${s} ${t.dayPeriod??""}`;return{...Ab(e),"aria-label":"hour, ","aria-valuemin":o,"aria-valuemax":i,"aria-valuenow":s,"aria-valuetext":l,"data-placeholder":a?"":void 0}}},minute:{attrs:function(e){const{segmentValues:t,placeholder:n}=e;if(!("minute"in t)||!("minute"in n))return{};const r=null===t.minute,a=(t.minute?n.set({minute:t.minute}):n).minute,o=r?"Empty":`${a}`;return{...Ab(e),"aria-label":"minute, ","aria-valuemin":0,"aria-valuemax":59,"aria-valuenow":a,"aria-valuetext":o,"data-placeholder":r?"":void 0}}},second:{attrs:function(e){const{segmentValues:t,placeholder:n}=e;if(!("second"in t)||!("second"in n))return{};const r=null===t.second,a=(t.second?n.set({second:t.second}):n).second,o=r?"Empty":`${a}`;return{...Ab(e),"aria-label":"second, ","aria-valuemin":0,"aria-valuemax":59,"aria-valuenow":a,"aria-valuetext":o,"data-placeholder":r?"":void 0}}},dayPeriod:{attrs:function(e){const{segmentValues:t}=e;if(!("dayPeriod"in t))return{};const n=t.hour?t.hour>12?t.hour-12:t.hour:0,r=t.dayPeriod??"AM";return{...Ab(e),inputmode:"text","aria-label":"AM/PM","aria-valuemin":0,"aria-valuemax":12,"aria-valuenow":n,"aria-valuetext":r}}},literal:{attrs:function(e){return{"aria-hidden":!0,"data-segment":"literal"}}},timeZoneName:{attrs:function(e){return{role:"textbox","aria-label":"timezone, ","data-readonly":!0,"data-segment":"timeZoneName",tabindex:e.disabled?void 0:0,style:"caret-color: transparent;"}}},era:{attrs:function(e){const{segmentValues:t,placeholder:n}=e,r="era"in t?t.era:n.era;return{...Ab(e),"aria-label":"era","aria-valuemin":0,"aria-valuemax":0,"aria-valuenow":0,"aria-valuetext":r}}}};function Pb(e){const t={ALT:"Alt",ARROW_DOWN:"ArrowDown",ARROW_LEFT:"ArrowLeft",ARROW_RIGHT:"ArrowRight",ARROW_UP:"ArrowUp",BACKSPACE:"Backspace",CAPS_LOCK:"CapsLock",CONTROL:"Control",DELETE:"Delete",END:"End",ENTER:"Enter",ESCAPE:"Escape",F1:"F1",F10:"F10",F11:"F11",F12:"F12",F2:"F2",F3:"F3",F4:"F4",F5:"F5",F6:"F6",F7:"F7",F8:"F8",F9:"F9",HOME:"Home",META:"Meta",PAGE_DOWN:"PageDown",PAGE_UP:"PageUp",SHIFT:"Shift",SPACE:" ",TAB:"Tab",CTRL:"Control",ASTERISK:"*",SPACE_CODE:"Space"};function n({e,part:n,dateRef:r,prevValue:a}){const o=e.key===t.ARROW_UP?1:-1;if(null===a)return o>0?0:59;const i=[n,o];return r.set({[n]:a}).cycle(...i)[n]}function r(t){if(e.hasLeftFocus.value=!1,null===t)return t;const n=t.toString();return 1===n.length?(e.modelValue.value=void 0,null):Number.parseInt(n.slice(0,-1))}function a({e:n,part:r,dateRef:a,prevValue:o,hourCycle:i}){const s=n.key===t.ARROW_UP?1:-1;if(null===o)return a[r];if("hour"===r&&"hour"in a){const e=[r,s,{hourCycle:i}];return a.set({[r]:o}).cycle(...e)[r]}const l=[r,s];return"day"===r&&null!==e.segmentValues.value.month?a.set({[r]:o,month:e.segmentValues.value.month}).cycle(...l)[r]:a.set({[r]:o}).cycle(...l)[r]}function o(t,n,r){let a=!1;const o=Math.floor(t/10);if(e.hasLeftFocus.value&&(e.hasLeftFocus.value=!1,r=null),null===r)return 0===n?(e.lastKeyZero.value=!0,{value:null,moveToNext:a}):((e.lastKeyZero.value||n>o)&&(a=!0),e.lastKeyZero.value=!1,{value:n,moveToNext:a});const i=r.toString().length,s=Number.parseInt(r.toString()+n.toString());return 2===i||s>t?((n>o||s>t)&&(a=!0),{value:n,moveToNext:a}):(a=!0,{value:s,moveToNext:a})}function i(t,n){let r=!1;const a=Math.floor(5.9);if(e.hasLeftFocus.value&&(e.hasLeftFocus.value=!1,n=null),null===n)return 0===t?(e.lastKeyZero.value=!0,{value:0,moveToNext:r}):((e.lastKeyZero.value||t>a)&&(r=!0),e.lastKeyZero.value=!1,{value:t,moveToNext:r});const o=n.toString().length,i=Number.parseInt(n.toString()+t.toString());return 2===o||i>59?(t>a&&(r=!0),{value:t,moveToNext:r}):(r=!0,{value:i,moveToNext:r})}function s(n){if(!_b(n.key)||Eb(n.key))return;const i=e.segmentValues.value.day;if(n.key!==t.ARROW_DOWN&&n.key!==t.ARROW_UP){if(Sb(n.key)){const t=Number.parseInt(n.key),r=e.segmentValues.value.month,a=Hf(r?e.placeholder.value.set({month:r}):e.placeholder.value),{value:s,moveToNext:l}=o(a,t,i);e.segmentValues.value.day=s,l&&e.focusNext()}n.key===t.BACKSPACE&&(e.hasLeftFocus.value=!1,e.segmentValues.value.day=r(i))}else e.segmentValues.value.day=a({e:n,part:"day",dateRef:e.placeholder.value,prevValue:i})}function l(n){if(!_b(n.key)||Eb(n.key))return;const i=e.segmentValues.value.month;if(n.key!==t.ARROW_DOWN&&n.key!==t.ARROW_UP){if(Sb(n.key)){const t=Number.parseInt(n.key),{value:r,moveToNext:a}=o(12,t,i);e.segmentValues.value.month=r,a&&e.focusNext()}n.key===t.BACKSPACE&&(e.hasLeftFocus.value=!1,e.segmentValues.value.month=r(i))}else e.segmentValues.value.month=a({e:n,part:"month",dateRef:e.placeholder.value,prevValue:i})}function c(n){if(!_b(n.key)||Eb(n.key))return;const o=e.segmentValues.value.year;if(n.key!==t.ARROW_DOWN&&n.key!==t.ARROW_UP){if(Sb(n.key)){const t=Number.parseInt(n.key),{value:r,moveToNext:a}=function(t,n){let r=!1;if(e.hasLeftFocus.value&&(e.hasLeftFocus.value=!1,n=null),null===n)return{value:0===t?1:t,moveToNext:r};const a=n.toString()+t.toString();return a.length>4?{value:0===t?1:t,moveToNext:r}:(4===a.length&&(r=!0),{value:Number.parseInt(a),moveToNext:r})}(t,o);e.segmentValues.value.year=r,a&&e.focusNext()}n.key===t.BACKSPACE&&(e.hasLeftFocus.value=!1,e.segmentValues.value.year=r(o))}else e.segmentValues.value.year=a({e:n,part:"year",dateRef:e.placeholder.value,prevValue:o})}function u(n){const o=e.placeholder.value;if(!_b(n.key)||Eb(n.key)||!("hour"in o)||!("hour"in e.segmentValues.value))return;const i=e.segmentValues.value.hour,s=e.hourCycle;if(n.key===t.ARROW_UP||n.key===t.ARROW_DOWN)return e.segmentValues.value.hour=a({e:n,part:"hour",dateRef:e.placeholder.value,prevValue:i,hourCycle:s}),void("dayPeriod"in e.segmentValues.value&&(e.segmentValues.value.hour<12?e.segmentValues.value.dayPeriod="AM":e.segmentValues.value.hour&&(e.segmentValues.value.dayPeriod="PM")));if(Sb(n.key)){const t=Number.parseInt(n.key),{value:r,moveToNext:a}=function(t,n){let r=!1;const a=Math.floor(2.4);if(e.hasLeftFocus.value&&(e.hasLeftFocus.value=!1,n=null),null===n)return 0===t?(e.lastKeyZero.value=!0,{value:0,moveToNext:r}):((e.lastKeyZero.value||t>a)&&(r=!0),e.lastKeyZero.value=!1,{value:t,moveToNext:r});const o=n.toString().length,i=Number.parseInt(n.toString()+t.toString());return 2===o||i>24?(t>a&&(r=!0),{value:t,moveToNext:r}):(r=!0,{value:i,moveToNext:r})}(t,i);"dayPeriod"in e.segmentValues.value&&r&&r>12?e.segmentValues.value.dayPeriod="PM":"dayPeriod"in e.segmentValues.value&&r&&(e.segmentValues.value.dayPeriod="AM"),e.segmentValues.value.hour=r,a&&e.focusNext()}n.key===t.BACKSPACE&&(e.hasLeftFocus.value=!1,e.segmentValues.value.hour=r(i))}function d(a){const o=e.placeholder.value;if(!_b(a.key)||Eb(a.key)||!("minute"in o)||!("minute"in e.segmentValues.value))return;const s=e.segmentValues.value.minute;if(e.segmentValues.value.minute=n({e:a,part:"minute",dateRef:e.placeholder.value,prevValue:s}),Sb(a.key)){const t=Number.parseInt(a.key),{value:n,moveToNext:r}=i(t,s);e.segmentValues.value.minute=n,r&&e.focusNext()}a.key===t.BACKSPACE&&(e.hasLeftFocus.value=!1,e.segmentValues.value.minute=r(s))}function p(a){const o=e.placeholder.value;if(!_b(a.key)||Eb(a.key)||!("second"in o)||!("second"in e.segmentValues.value))return;const s=e.segmentValues.value.second;if(e.segmentValues.value.second=n({e:a,part:"second",dateRef:e.placeholder.value,prevValue:s}),Sb(a.key)){const t=Number.parseInt(a.key),{value:n,moveToNext:r}=i(t,s);e.segmentValues.value.second=n,r&&e.focusNext()}a.key===t.BACKSPACE&&(e.hasLeftFocus.value=!1,e.segmentValues.value.second=r(s))}function h(n){if((_b(n.key)&&!Eb(n.key)||"a"===n.key||"p"===n.key)&&"hour"in e.placeholder.value&&"dayPeriod"in e.segmentValues.value){if(n.key===t.ARROW_UP||n.key===t.ARROW_DOWN)return"AM"===e.segmentValues.value.dayPeriod?(e.segmentValues.value.dayPeriod="PM",void(e.segmentValues.value.hour=e.segmentValues.value.hour+12)):(e.segmentValues.value.dayPeriod="AM",void(e.segmentValues.value.hour=e.segmentValues.value.hour-12));if(["a","A"].includes(n.key)&&"AM"!==e.segmentValues.value.dayPeriod)return e.segmentValues.value.dayPeriod="AM",void(e.segmentValues.value.hour=e.segmentValues.value.hour-12);["p","P"].includes(n.key)&&"PM"!==e.segmentValues.value.dayPeriod&&(e.segmentValues.value.dayPeriod="PM",e.segmentValues.value.hour=e.segmentValues.value.hour+12)}}return{handleSegmentClick:function(t){e.disabled.value&&t.preventDefault()},handleSegmentKeydown:function(n){const r=e.disabled.value,a=e.readonly.value;if(n.key!==t.TAB&&n.preventDefault(),!r&&!a&&({day:s,month:l,year:c,hour:u,minute:d,second:p,dayPeriod:h,timeZoneName:()=>{}}[e.part](n),![t.ARROW_LEFT,t.ARROW_RIGHT].includes(n.key)&&n.key!==t.TAB&&n.key!==t.SHIFT&&_b(n.key)&&Object.values(e.segmentValues.value).every((e=>null!==e)))){const t={...e.segmentValues.value};let n=e.placeholder.value.copy();Object.keys(t).forEach((e=>{const r=t[e];n=n.set({[e]:r})})),e.modelValue.value=n.copy()}},attributes:(0,$t.EW)((()=>{var t;return(null==(t=Rb[e.part])?void 0:t.attrs({disabled:e.disabled.value,placeholder:e.placeholder.value,hourCycle:e.hourCycle,segmentValues:e.segmentValues.value,formatter:e.formatter}))??{}}))}}Boolean;const[Db,Mb]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("DatePickerRoot")),[$b,Nb]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("DateRangePickerRoot")),[Ib,Lb]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("DateRangeFieldRoot")),[Bb,Fb]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("DropdownMenuRoot")),Qb=(0,$t.pM)({__name:"DropdownMenuRoot",props:{defaultOpen:{type:Boolean},open:{type:Boolean,default:void 0},dir:{},modal:{type:Boolean,default:!0}},emits:["update:open"],setup(e,{emit:t}){const n=e,r=t;Zm();const a=Cm(n,"open",r,{defaultValue:n.defaultOpen,passive:void 0===n.open}),o=(0,jn.KR)(),{modal:i,dir:s}=(0,jn.QW)(n),l=zm(s);return Fb({open:a,onOpenChange:e=>{a.value=e},onOpenToggle:()=>{a.value=!a.value},triggerId:"",triggerElement:o,contentId:"",modal:i,dir:l}),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(zv),{open:(0,jn.R1)(a),"onUpdate:open":t[0]||(t[0]=e=>(0,jn.i9)(a)?a.value=e:null),dir:(0,jn.R1)(l),modal:(0,jn.R1)(i)},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{open:(0,jn.R1)(a)})])),_:3},8,["open","dir","modal"]))}}),Ub=(0,$t.pM)({__name:"DropdownMenuTrigger",props:{disabled:{type:Boolean},asChild:{type:Boolean},as:{default:"button"}},setup(e){const t=e,n=Bb(),{forwardRef:r,currentElement:a}=Zm();return(0,$t.sV)((()=>{n.triggerElement=a})),n.triggerId||(n.triggerId=og(void 0,"radix-vue-dropdown-menu-trigger")),(e,a)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Nv),{"as-child":""},{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(ug),{id:(0,jn.R1)(n).triggerId,ref:(0,jn.R1)(r),type:"button"===e.as?"button":void 0,"as-child":t.asChild,as:e.as,"aria-haspopup":"menu","aria-expanded":(0,jn.R1)(n).open.value,"aria-controls":(0,jn.R1)(n).open.value?(0,jn.R1)(n).contentId:void 0,"data-disabled":e.disabled?"":void 0,disabled:e.disabled,"data-state":(0,jn.R1)(n).open.value?"open":"closed",onClick:a[0]||(a[0]=async t=>{var r;!e.disabled&&0===t.button&&!1===t.ctrlKey&&(null==(r=(0,jn.R1)(n))||r.onOpenToggle(),await(0,$t.dY)(),(0,jn.R1)(n).open.value&&t.preventDefault())}),onKeydown:a[1]||(a[1]=(0,Rr.jR)((t=>{e.disabled||(["Enter"," "].includes(t.key)&&(0,jn.R1)(n).onOpenToggle(),"ArrowDown"===t.key&&(0,jn.R1)(n).onOpenChange(!0),["Enter"," ","ArrowDown"].includes(t.key)&&t.preventDefault())}),["enter","space","arrow-down"]))},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["id","type","as-child","as","aria-expanded","aria-controls","data-disabled","disabled","data-state"])])),_:3}))}}),zb=(0,$t.pM)({__name:"DropdownMenuPortal",props:{to:{},disabled:{type:Boolean},forceMount:{type:Boolean}},setup(e){const t=e;return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ub),(0,Wn._B)((0,$t.Ng)(t)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),jb=(0,$t.pM)({__name:"DropdownMenuContent",props:{forceMount:{type:Boolean},loop:{type:Boolean},side:{},sideOffset:{},align:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean},updatePositionStrategy:{},prioritizePosition:{type:Boolean},asChild:{type:Boolean},as:{}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","closeAutoFocus"],setup(e,{emit:t}){const n=Hm(e,t);Zm();const r=Bb(),a=(0,jn.KR)(!1);function o(e){e.defaultPrevented||(a.value||setTimeout((()=>{var e;null==(e=r.triggerElement.value)||e.focus()}),0),a.value=!1,e.preventDefault())}return r.contentId||(r.contentId=og(void 0,"radix-vue-dropdown-menu-content")),(e,t)=>{var i;return(0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(sb),(0,$t.v6)((0,jn.R1)(n),{id:(0,jn.R1)(r).contentId,"aria-labelledby":null==(i=(0,jn.R1)(r))?void 0:i.triggerId,style:{"--radix-dropdown-menu-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-dropdown-menu-content-available-width":"var(--radix-popper-available-width)","--radix-dropdown-menu-content-available-height":"var(--radix-popper-available-height)","--radix-dropdown-menu-trigger-width":"var(--radix-popper-anchor-width)","--radix-dropdown-menu-trigger-height":"var(--radix-popper-anchor-height)"},onCloseAutoFocus:o,onInteractOutside:t[0]||(t[0]=e=>{var t;if(e.defaultPrevented)return;const n=e.detail.originalEvent,o=0===n.button&&!0===n.ctrlKey,i=2===n.button||o;(!(0,jn.R1)(r).modal.value||i)&&(a.value=!0),null!=(t=(0,jn.R1)(r).triggerElement.value)&&t.contains(e.target)&&e.preventDefault()})}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["id","aria-labelledby"])}}}),Wb=(0,$t.pM)({__name:"DropdownMenuArrow",props:{width:{default:10},height:{default:5},asChild:{type:Boolean},as:{default:"svg"}},setup(e){const t=e;return Zm(),(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Iv),(0,Wn._B)((0,$t.Ng)(t)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),Xb=(0,$t.pM)({__name:"DropdownMenuItem",props:{disabled:{type:Boolean},textValue:{},asChild:{type:Boolean},as:{}},emits:["select"],setup(e,{emit:t}){const n=e,r=jm(t);return Zm(),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(eb),(0,Wn._B)((0,$t.Ng)({...n,...(0,jn.R1)(r)})),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),qb=(0,$t.pM)({__name:"DropdownMenuGroup",props:{asChild:{type:Boolean},as:{}},setup(e){const t=e;return Zm(),(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(lb),(0,Wn._B)((0,$t.Ng)(t)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),Vb=(0,$t.pM)({__name:"DropdownMenuSeparator",props:{asChild:{type:Boolean},as:{}},setup(e){const t=e;return Zm(),(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(mb),(0,Wn._B)((0,$t.Ng)(t)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),Hb=(0,$t.pM)({__name:"DropdownMenuCheckboxItem",props:{checked:{type:[Boolean,String]},disabled:{type:Boolean},textValue:{},asChild:{type:Boolean},as:{}},emits:["select","update:checked"],setup(e,{emit:t}){const n=e,r=jm(t);return Zm(),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ab),(0,Wn._B)((0,$t.Ng)({...n,...(0,jn.R1)(r)})),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),Zb=(0,$t.pM)({__name:"DropdownMenuItemIndicator",props:{forceMount:{type:Boolean},asChild:{type:Boolean},as:{}},setup(e){const t=e;return Zm(),(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(rb),(0,Wn._B)((0,$t.Ng)(t)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),Gb=(0,$t.pM)({__name:"DropdownMenuLabel",props:{asChild:{type:Boolean},as:{}},setup(e){const t=e;return Zm(),(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(cb),(0,Wn._B)((0,$t.Ng)(t)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),Kb=(0,$t.pM)({__name:"DropdownMenuRadioGroup",props:{modelValue:{},asChild:{type:Boolean},as:{}},emits:["update:modelValue"],setup(e,{emit:t}){const n=e,r=jm(t);return Zm(),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(hb),(0,Wn._B)((0,$t.Ng)({...n,...(0,jn.R1)(r)})),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),Yb=(0,$t.pM)({__name:"DropdownMenuRadioItem",props:{value:{},disabled:{type:Boolean},textValue:{},asChild:{type:Boolean},as:{}},emits:["select"],setup(e,{emit:t}){const n=Hm(e,t);return Zm(),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(fb),(0,Wn._B)((0,$t.Ng)((0,jn.R1)(n))),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),Jb=(0,$t.pM)({__name:"DropdownMenuSub",props:{defaultOpen:{type:Boolean},open:{type:Boolean,default:void 0}},emits:["update:open"],setup(e,{emit:t}){const n=e,r=Cm(n,"open",t,{passive:void 0===n.open,defaultValue:n.defaultOpen??!1});return Zm(),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(bb),{open:(0,jn.R1)(r),"onUpdate:open":t[0]||(t[0]=e=>(0,jn.i9)(r)?r.value=e:null)},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{open:(0,jn.R1)(r)})])),_:3},8,["open"]))}}),ey=(0,$t.pM)({__name:"DropdownMenuSubContent",props:{forceMount:{type:Boolean},loop:{type:Boolean},sideOffset:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean},updatePositionStrategy:{},prioritizePosition:{type:Boolean},asChild:{type:Boolean},as:{}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","entryFocus","openAutoFocus","closeAutoFocus"],setup(e,{emit:t}){const n=Hm(e,t);return Zm(),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(yb),(0,$t.v6)((0,jn.R1)(n),{style:{"--radix-dropdown-menu-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-dropdown-menu-content-available-width":"var(--radix-popper-available-width)","--radix-dropdown-menu-content-available-height":"var(--radix-popper-available-height)","--radix-dropdown-menu-trigger-width":"var(--radix-popper-anchor-width)","--radix-dropdown-menu-trigger-height":"var(--radix-popper-anchor-height)"}}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),ty=(0,$t.pM)({__name:"DropdownMenuSubTrigger",props:{disabled:{type:Boolean},textValue:{},asChild:{type:Boolean},as:{}},setup(e){const t=e;return Zm(),(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(wb),(0,Wn._B)((0,$t.Ng)(t)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),[ny,ry]=Kf("EditableRoot"),[ay,oy]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("HoverCardRoot"));Boolean,Boolean,Boolean,Boolean,Boolean;Boolean,Boolean,Boolean,Boolean;Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean;const[iy,sy]=Kf("ListboxRoot"),[ly,cy]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("ListboxItem"));Boolean,Boolean,Boolean,typeof window>"u"||window;const[uy,dy]=Kf("ListboxGroup"),[py,hy]=(Boolean,Boolean,Kf("MenubarRoot")),[fy,my]=(Boolean,Kf("MenubarMenu")),[gy,vy]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf(["NavigationMenuRoot","NavigationMenuSub"],"NavigationMenuContext"));function by(e){return e?"open":"closed"}Boolean,Boolean,Boolean;const yy="navigationMenu.rootContentDismiss";const[wy,Oy]=Kf("NavigationMenuItem");Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean;Boolean,Boolean,Boolean,Boolean;const[ky,xy]=Kf("NumberFieldRoot"),[Ey,Sy]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("PaginationRoot"));Boolean,Boolean,Boolean,Boolean,Boolean,Boolean;Boolean,Boolean,Boolean,Boolean;const[_y,Cy]=Kf("PinInputRoot"),[Ty,Ay]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("PopoverRoot")),Ry=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,(0,$t.pM)({__name:"PopoverContentImpl",props:{trapFocus:{type:Boolean},side:{},sideOffset:{},align:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean},updatePositionStrategy:{},prioritizePosition:{type:Boolean},asChild:{type:Boolean},as:{},disableOutsidePointerEvents:{type:Boolean}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","openAutoFocus","closeAutoFocus"],setup(e,{emit:t}){const n=t,r=Vm(e),{forwardRef:a}=Zm(),o=Ty();return Xm(),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Lg),{"as-child":"",loop:"",trapped:e.trapFocus,onMountAutoFocus:t[5]||(t[5]=e=>n("openAutoFocus",e)),onUnmountAutoFocus:t[6]||(t[6]=e=>n("closeAutoFocus",e))},{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(Cg),{"as-child":"","disable-outside-pointer-events":e.disableOutsidePointerEvents,onPointerDownOutside:t[0]||(t[0]=e=>n("pointerDownOutside",e)),onInteractOutside:t[1]||(t[1]=e=>n("interactOutside",e)),onEscapeKeyDown:t[2]||(t[2]=e=>n("escapeKeyDown",e)),onFocusOutside:t[3]||(t[3]=e=>n("focusOutside",e)),onDismiss:t[4]||(t[4]=e=>(0,jn.R1)(o).onOpenChange(!1))},{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(gv),(0,$t.v6)((0,jn.R1)(r),{id:(0,jn.R1)(o).contentId,ref:(0,jn.R1)(a),"data-state":(0,jn.R1)(o).open.value?"open":"closed",role:"dialog",style:{"--radix-popover-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-popover-content-available-width":"var(--radix-popper-available-width)","--radix-popover-content-available-height":"var(--radix-popper-available-height)","--radix-popover-trigger-width":"var(--radix-popper-anchor-width)","--radix-popover-trigger-height":"var(--radix-popper-anchor-height)"}}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["id","data-state"])])),_:3},8,["disable-outside-pointer-events"])])),_:3},8,["trapped"]))}})),Py=(0,$t.pM)({__name:"PopoverContentModal",props:{trapFocus:{type:Boolean},side:{},sideOffset:{},align:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean},updatePositionStrategy:{},prioritizePosition:{type:Boolean},asChild:{type:Boolean},as:{},disableOutsidePointerEvents:{type:Boolean}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","openAutoFocus","closeAutoFocus"],setup(e,{emit:t}){const n=e,r=t,a=Ty(),o=(0,jn.KR)(!1);Fm(!0);const i=Hm(n,r),{forwardRef:s,currentElement:l}=Zm();return rg(l),(e,t)=>((0,$t.uX)(),(0,$t.Wv)(Ry,(0,$t.v6)((0,jn.R1)(i),{ref:(0,jn.R1)(s),"trap-focus":(0,jn.R1)(a).open.value,"disable-outside-pointer-events":"",onCloseAutoFocus:t[0]||(t[0]=(0,Rr.D$)((e=>{var t;r("closeAutoFocus",e),o.value||null==(t=(0,jn.R1)(a).triggerElement.value)||t.focus()}),["prevent"])),onPointerDownOutside:t[1]||(t[1]=e=>{r("pointerDownOutside",e);const t=e.detail.originalEvent,n=0===t.button&&!0===t.ctrlKey,a=2===t.button||n;o.value=a}),onFocusOutside:t[2]||(t[2]=(0,Rr.D$)((()=>{}),["prevent"]))}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["trap-focus"]))}}),Dy=(0,$t.pM)({__name:"PopoverContentNonModal",props:{trapFocus:{type:Boolean},side:{},sideOffset:{},align:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean},updatePositionStrategy:{},prioritizePosition:{type:Boolean},asChild:{type:Boolean},as:{},disableOutsidePointerEvents:{type:Boolean}},emits:["escapeKeyDown","pointerDownOutside","focusOutside","interactOutside","openAutoFocus","closeAutoFocus"],setup(e,{emit:t}){const n=e,r=t,a=Ty(),o=(0,jn.KR)(!1),i=(0,jn.KR)(!1),s=Hm(n,r);return(e,t)=>((0,$t.uX)(),(0,$t.Wv)(Ry,(0,$t.v6)((0,jn.R1)(s),{"trap-focus":!1,"disable-outside-pointer-events":!1,onCloseAutoFocus:t[0]||(t[0]=e=>{var t;r("closeAutoFocus",e),e.defaultPrevented||(o.value||null==(t=(0,jn.R1)(a).triggerElement.value)||t.focus(),e.preventDefault()),o.value=!1,i.value=!1}),onInteractOutside:t[1]||(t[1]=async e=>{var t;r("interactOutside",e),e.defaultPrevented||(o.value=!0,"pointerdown"===e.detail.originalEvent.type&&(i.value=!0));const n=e.target;(null==(t=(0,jn.R1)(a).triggerElement.value)?void 0:t.contains(n))&&e.preventDefault(),"focusin"===e.detail.originalEvent.type&&i.value&&e.preventDefault()})}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),[My,$y]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("ProgressRoot"));Boolean,Boolean;const[Ny,Iy]=Kf("RadioGroupRoot"),Ly=(Boolean,Boolean,Boolean,Boolean,["value","checked","name","disabled","required"]),[By,Fy]=(Boolean,Boolean,Boolean,Boolean,Kf("RadioGroupItem"));Boolean,Boolean,Boolean,Boolean,Boolean;const[Qy,Uy]=Kf("RangeCalendarRoot"),[zy,jy]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("ScrollAreaRoot"));function Wy(e,t){return n=>{if(e[0]===e[1]||t[0]===t[1])return t[0];const r=(t[1]-t[0])/(e[1]-e[0]);return t[0]+r*(n-e[0])}}function Xy(e){const t=qy(e.viewport,e.content),n=e.scrollbar.paddingStart+e.scrollbar.paddingEnd,r=(e.scrollbar.size-n)*t;return Math.max(r,18)}function qy(e,t){const n=e/t;return Number.isNaN(n)?0:n}function Vy(e,t,n="ltr"){const r=Xy(t),a=t.scrollbar.paddingStart+t.scrollbar.paddingEnd,o=t.scrollbar.size-a,i=t.content-t.viewport,s=o-r,l="ltr"===n?[0,i]:[-1*i,0],c=Jf(e,l[0],l[1]);return Wy([0,i],[0,s])(c)}function Hy(e){return e?Number.parseInt(e,10):0}function Zy(e,t){return e>0&&e<t}Boolean,Boolean;const Gy=(0,$t.pM)({__name:"ScrollAreaScrollbarImpl",props:{isHorizontal:{type:Boolean}},emits:["onDragScroll","onWheelScroll","onThumbPointerDown"],setup(e,{emit:t}){const n=e,r=t,a=zy(),o=Jy(),i=rw(),{forwardRef:s,currentElement:l}=Zm(),c=(0,jn.KR)(""),u=(0,jn.KR)();function d(e){var t,n;if(u.value){const a=e.clientX-(null==(t=u.value)?void 0:t.left),o=e.clientY-(null==(n=u.value)?void 0:n.top);r("onDragScroll",{x:a,y:o})}}function p(e){0===e.button&&(e.target.setPointerCapture(e.pointerId),u.value=l.value.getBoundingClientRect(),c.value=document.body.style.webkitUserSelect,document.body.style.webkitUserSelect="none",a.viewport&&(a.viewport.value.style.scrollBehavior="auto"),d(e))}function h(e){d(e)}function f(e){const t=e.target;t.hasPointerCapture(e.pointerId)&&t.releasePointerCapture(e.pointerId),document.body.style.webkitUserSelect=c.value,a.viewport&&(a.viewport.value.style.scrollBehavior=""),u.value=void 0}function m(e){var t;const n=e.target,r=null==(t=l.value)?void 0:t.contains(n),a=o.sizes.value.content-o.sizes.value.viewport;r&&o.handleWheelScroll(e,a)}function g(){var e,t,r,i,s;l.value&&(n.isHorizontal?o.handleSizeChange({content:(null==(e=a.viewport.value)?void 0:e.scrollWidth)??0,viewport:(null==(t=a.viewport.value)?void 0:t.offsetWidth)??0,scrollbar:{size:l.value.clientWidth??0,paddingStart:Hy(getComputedStyle(l.value).paddingLeft),paddingEnd:Hy(getComputedStyle(l.value).paddingRight)}}):o.handleSizeChange({content:(null==(r=a.viewport.value)?void 0:r.scrollHeight)??0,viewport:(null==(i=a.viewport.value)?void 0:i.offsetHeight)??0,scrollbar:{size:(null==(s=l.value)?void 0:s.clientHeight)??0,paddingStart:Hy(getComputedStyle(l.value).paddingLeft),paddingEnd:Hy(getComputedStyle(l.value).paddingRight)}}))}return(0,$t.sV)((()=>{document.addEventListener("wheel",m,{passive:!1})})),(0,$t.hi)((()=>{document.removeEventListener("wheel",m)})),_m(l,g),_m(a.content,g),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),{ref:(0,jn.R1)(s),style:{position:"absolute"},"data-scrollbarimpl":"",as:(0,jn.R1)(i).as.value,"as-child":(0,jn.R1)(i).asChild.value,onPointerdown:p,onPointermove:h,onPointerup:f},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["as","as-child"]))}}),Ky=(0,$t.pM)({__name:"ScrollAreaScrollbarX",setup(e){const t=zy(),n=Jy(),{forwardRef:r,currentElement:a}=Zm();(0,$t.sV)((()=>{a.value&&t.onScrollbarXChange(a.value)}));const o=(0,$t.EW)((()=>n.sizes.value));return(e,a)=>((0,$t.uX)(),(0,$t.Wv)(Gy,{ref:(0,jn.R1)(r),"is-horizontal":!0,"data-orientation":"horizontal",style:(0,Wn.Tr)({bottom:0,left:"rtl"===(0,jn.R1)(t).dir.value?"var(--radix-scroll-area-corner-width)":0,right:"ltr"===(0,jn.R1)(t).dir.value?"var(--radix-scroll-area-corner-width)":0,"--radix-scroll-area-thumb-width":o.value?`${(0,jn.R1)(Xy)(o.value)}px`:void 0}),onOnDragScroll:a[0]||(a[0]=e=>(0,jn.R1)(n).onDragScroll(e.x))},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["style"]))}}),Yy=(0,$t.pM)({__name:"ScrollAreaScrollbarY",setup(e){const t=zy(),n=Jy(),{forwardRef:r,currentElement:a}=Zm();(0,$t.sV)((()=>{a.value&&t.onScrollbarYChange(a.value)}));const o=(0,$t.EW)((()=>n.sizes.value));return(e,a)=>((0,$t.uX)(),(0,$t.Wv)(Gy,{ref:(0,jn.R1)(r),"is-horizontal":!1,"data-orientation":"vertical",style:(0,Wn.Tr)({top:0,right:"ltr"===(0,jn.R1)(t).dir.value?0:void 0,left:"rtl"===(0,jn.R1)(t).dir.value?0:void 0,bottom:"var(--radix-scroll-area-corner-height)","--radix-scroll-area-thumb-height":o.value?`${(0,jn.R1)(Xy)(o.value)}px`:void 0}),onOnDragScroll:a[0]||(a[0]=e=>(0,jn.R1)(n).onDragScroll(e.y))},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["style"]))}}),[Jy,ew]=Kf("ScrollAreaScrollbarVisible"),tw=(0,$t.pM)({__name:"ScrollAreaScrollbarVisible",setup(e){const t=zy(),n=rw(),{forwardRef:r}=Zm(),a=(0,jn.KR)({content:0,viewport:0,scrollbar:{size:0,paddingStart:0,paddingEnd:0}}),o=(0,$t.EW)((()=>{const e=qy(a.value.viewport,a.value.content);return e>0&&e<1})),i=(0,jn.KR)(),s=(0,jn.KR)(0);function l(e,t){return function(e,t,n,r="ltr"){const a=Xy(n),o=t||a/2,i=a-o,s=n.scrollbar.paddingStart+o,l=n.scrollbar.size-n.scrollbar.paddingEnd-i,c=n.content-n.viewport;return Wy([s,l],"ltr"===r?[0,c]:[-1*c,0])(e)}(e,s.value,a.value,t)}const c=(0,$t.EW)((()=>n.isHorizontal.value));return ew({sizes:a,hasThumb:o,handleWheelScroll:function(e,n){if(c.value){const r=t.viewport.value.scrollLeft+e.deltaY;t.viewport.value.scrollLeft=r,Zy(r,n)&&e.preventDefault()}else{const r=t.viewport.value.scrollTop+e.deltaY;t.viewport.value.scrollTop=r,Zy(r,n)&&e.preventDefault()}},handleThumbDown:function(e,t){c.value?s.value=t.x:s.value=t.y},handleThumbUp:function(e){s.value=0},handleSizeChange:function(e){a.value=e},onThumbPositionChange:function(){if(c.value){if(t.viewport.value&&i.value){const e=Vy(t.viewport.value.scrollLeft,a.value,t.dir.value);i.value.style.transform=`translate3d(${e}px, 0, 0)`}}else if(t.viewport.value&&i.value){const e=Vy(t.viewport.value.scrollTop,a.value);i.value.style.transform=`translate3d(0, ${e}px, 0)`}},onThumbChange:function(e){i.value=e},onDragScroll:function(e){c.value?t.viewport.value.scrollLeft=l(e,t.dir.value):t.viewport.value.scrollTop=l(e)}}),(e,t)=>c.value?((0,$t.uX)(),(0,$t.Wv)(Ky,(0,$t.v6)({key:0},e.$attrs,{ref:(0,jn.R1)(r)}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16)):((0,$t.uX)(),(0,$t.Wv)(Yy,(0,$t.v6)({key:1},e.$attrs,{ref:(0,jn.R1)(r)}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),nw=(0,$t.pM)({__name:"ScrollAreaScrollbarAuto",props:{forceMount:{type:Boolean}},setup(e){const t=zy(),n=rw(),{forwardRef:r}=Zm(),a=(0,jn.KR)(!1),o=mm((()=>{if(t.viewport.value){const e=t.viewport.value.offsetWidth<t.viewport.value.scrollWidth,r=t.viewport.value.offsetHeight<t.viewport.value.scrollHeight;a.value=n.isHorizontal.value?e:r}}),10);return(0,$t.sV)((()=>o())),_m(t.viewport,o),_m(t.content,o),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(mg),{present:e.forceMount||a.value},{default:(0,$t.k6)((()=>[(0,$t.bF)(tw,(0,$t.v6)(e.$attrs,{ref:(0,jn.R1)(r),"data-state":a.value?"visible":"hidden"}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["data-state"])])),_:3},8,["present"]))}}),[rw,aw]=(Boolean,Boolean,Kf("ScrollAreaScrollbar")),ow=(Boolean,Boolean,Boolean,Boolean,["default-value"]),[iw,sw]=(Boolean,Boolean,Boolean,Boolean,Kf("SelectRoot")),[lw,cw]=Kf("SelectRoot"),uw=(Boolean,Boolean,Boolean,Boolean,10);Boolean,Boolean,Boolean,Boolean;const[dw,pw]=Kf("SelectItemAlignedPosition"),hw=(0,$t.pM)({inheritAttrs:!1,__name:"SelectItemAlignedPosition",props:{asChild:{type:Boolean},as:{}},emits:["placed"],setup(e,{emit:t}){const n=e,r=t,{injectCollection:a}=Um(),o=iw(),i=gw(),s=a(),l=(0,jn.KR)(!1),c=(0,jn.KR)(!0),u=(0,jn.KR)(),{forwardRef:d,currentElement:p}=Zm(),{viewport:h,selectedItem:f,selectedItemText:m,focusSelectedItem:g}=i;function v(){if(o.triggerElement.value&&o.valueElement.value&&u.value&&p.value&&null!=h&&h.value&&null!=f&&f.value&&null!=m&&m.value){const e=o.triggerElement.value.getBoundingClientRect(),t=p.value.getBoundingClientRect(),n=o.valueElement.value.getBoundingClientRect(),a=m.value.getBoundingClientRect();if("rtl"!==o.dir.value){const r=a.left-t.left,o=n.left-r,i=e.left-o,s=e.width+i,l=Math.max(s,t.width),c=window.innerWidth-uw,d=Jf(o,uw,Math.max(uw,c-l));u.value.style.minWidth=`${s}px`,u.value.style.left=`${d}px`}else{const r=t.right-a.right,o=window.innerWidth-n.right-r,i=window.innerWidth-e.right-o,s=e.width+i,l=Math.max(s,t.width),c=window.innerWidth-uw,d=Jf(o,uw,Math.max(uw,c-l));u.value.style.minWidth=`${s}px`,u.value.style.right=`${d}px`}const i=s.value,c=window.innerHeight-20,d=h.value.scrollHeight,g=window.getComputedStyle(p.value),v=Number.parseInt(g.borderTopWidth,10),b=Number.parseInt(g.paddingTop,10),y=Number.parseInt(g.borderBottomWidth,10),w=v+b+d+Number.parseInt(g.paddingBottom,10)+y,O=Math.min(5*f.value.offsetHeight,w),k=window.getComputedStyle(h.value),x=Number.parseInt(k.paddingTop,10),E=Number.parseInt(k.paddingBottom,10),S=e.top+e.height/2-uw,_=c-S,C=f.value.offsetHeight/2,T=v+b+(f.value.offsetTop+C),A=w-T;if(T<=S){const e=f.value===i[i.length-1];u.value.style.bottom="0px";const t=p.value.clientHeight-h.value.offsetTop-h.value.offsetHeight,n=T+Math.max(_,C+(e?E:0)+t+y);u.value.style.height=`${n}px`}else{const e=f.value===i[0];u.value.style.top="0px";const t=Math.max(S,v+h.value.offsetTop+(e?x:0)+C)+A;u.value.style.height=`${t}px`,h.value.scrollTop=T-S+h.value.offsetTop}u.value.style.margin="10px 0",u.value.style.minHeight=`${O}px`,u.value.style.maxHeight=`${c}px`,r("placed"),requestAnimationFrame((()=>l.value=!0))}}const b=(0,jn.KR)("");return(0,$t.sV)((async()=>{await(0,$t.dY)(),v(),p.value&&(b.value=window.getComputedStyle(p.value).zIndex)})),pw({contentWrapper:u,shouldExpandOnScrollRef:l,onScrollButtonChange:function(e){e&&!0===c.value&&(v(),null==g||g(),c.value=!1)}}),(e,t)=>((0,$t.uX)(),(0,$t.CE)("div",{ref_key:"contentWrapperElement",ref:u,style:(0,Wn.Tr)({display:"flex",flexDirection:"column",position:"fixed",zIndex:b.value})},[(0,$t.bF)((0,jn.R1)(ug),(0,$t.v6)({ref:(0,jn.R1)(d),style:{boxSizing:"border-box",maxHeight:"100%"}},{...e.$attrs,...n}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16)],4))}}),fw=(0,$t.pM)({__name:"SelectPopperPosition",props:{side:{},sideOffset:{},align:{default:"start"},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{default:uw},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean},updatePositionStrategy:{},prioritizePosition:{type:Boolean},asChild:{type:Boolean},as:{}},setup(e){const t=Vm(e);return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(gv),(0,$t.v6)((0,jn.R1)(t),{style:{boxSizing:"border-box","--radix-select-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-select-content-available-width":"var(--radix-popper-available-width)","--radix-select-content-available-height":"var(--radix-popper-available-height)","--radix-select-trigger-width":"var(--radix-popper-anchor-width)","--radix-select-trigger-height":"var(--radix-popper-anchor-height)"}}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),mw={onViewportChange:()=>{},itemTextRefCallback:()=>{},itemRefCallback:()=>{}},[gw,vw]=Kf("SelectContent"),[bw,yw]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("SelectItem")),[ww,Ow]=(Boolean,Boolean,Boolean,Kf("SelectGroup")),kw=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,(0,$t.pM)({__name:"BaseSeparator",props:{orientation:{default:"horizontal"},decorative:{type:Boolean},asChild:{type:Boolean},as:{}},setup(e){const t=e,n=["horizontal","vertical"],r=(0,$t.EW)((()=>function(e){return n.includes(e)}(t.orientation)?t.orientation:"horizontal")),a=(0,$t.EW)((()=>"vertical"===r.value?t.orientation:void 0)),o=(0,$t.EW)((()=>t.decorative?{role:"none"}:{"aria-orientation":a.value,role:"separator"}));return(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),(0,$t.v6)({as:e.as,"as-child":e.asChild,"data-orientation":r.value},o.value),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["as","as-child","data-orientation"]))}}));Boolean,Boolean;function xw(e,t){return n=>{if(e[0]===e[1]||t[0]===t[1])return t[0];const r=(t[1]-t[0])/(e[1]-e[0]);return t[0]+r*(n-e[0])}}const Ew=["PageUp","PageDown"],Sw=["ArrowUp","ArrowDown","ArrowLeft","ArrowRight"],_w={"from-left":["Home","PageDown","ArrowDown","ArrowLeft"],"from-right":["Home","PageDown","ArrowDown","ArrowRight"],"from-bottom":["Home","PageDown","ArrowDown","ArrowLeft"],"from-top":["Home","PageDown","ArrowUp","ArrowLeft"]},[Cw,Tw]=Kf(["SliderVertical","SliderHorizontal"]),Aw=(0,$t.pM)({__name:"SliderImpl",props:{asChild:{type:Boolean},as:{default:"span"}},emits:["slideStart","slideMove","slideEnd","homeKeyDown","endKeyDown","stepKeyDown"],setup(e,{emit:t}){const n=e,r=t,a=Rw();return(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),(0,$t.v6)({"data-slider-impl":""},n,{onKeydown:t[0]||(t[0]=e=>{"Home"===e.key?(r("homeKeyDown",e),e.preventDefault()):"End"===e.key?(r("endKeyDown",e),e.preventDefault()):(0,jn.R1)(Ew).concat((0,jn.R1)(Sw)).includes(e.key)&&(r("stepKeyDown",e),e.preventDefault())}),onPointerdown:t[1]||(t[1]=e=>{const t=e.target;t.setPointerCapture(e.pointerId),e.preventDefault(),(0,jn.R1)(a).thumbElements.value.includes(t)?t.focus():r("slideStart",e)}),onPointermove:t[2]||(t[2]=e=>{e.target.hasPointerCapture(e.pointerId)&&r("slideMove",e)}),onPointerup:t[3]||(t[3]=e=>{const t=e.target;t.hasPointerCapture(e.pointerId)&&(t.releasePointerCapture(e.pointerId),r("slideEnd",e))})}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),[Rw,Pw]=(Boolean,Boolean,Kf("SliderRoot"));Boolean,Boolean,Boolean,Boolean;Boolean,Boolean,Boolean;!function(){if("function"==typeof matchMedia)matchMedia("(pointer:coarse)").matches}();const[Dw,Mw]=Kf("PanelGroup");Boolean,Boolean,Boolean,Boolean,Boolean;const[$w,Nw]=Kf("StepperRoot"),[Iw,Lw]=(Boolean,Boolean,Kf("StepperItem")),[Bw,Fw]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("SwitchRoot")),[Qw,Uw]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("TabsRoot"));Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean;const[zw,jw]=Kf("TagsInputRoot"),[Ww,Xw]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("TagsInputItem")),[qw,Vw]=(Boolean,Boolean,Boolean,Boolean,Boolean,Kf("ToastProvider")),Hw="toast.viewportPause",Zw="toast.viewportResume";function Gw(e,t,n){const r=n.originalEvent.currentTarget,a=new CustomEvent(e,{bubbles:!1,cancelable:!0,detail:n});t&&r.addEventListener(e,t,{once:!0}),r.dispatchEvent(a)}function Kw(e,t,n=0){const r=Math.abs(e.x),a=Math.abs(e.y),o=r>a;return"left"===t||"right"===t?o&&r>n:!o&&a>n}function Yw(e){const t=[];return Array.from(e.childNodes).forEach((e=>{if(e.nodeType===e.TEXT_NODE&&e.textContent&&t.push(e.textContent),function(e){return e.nodeType===e.ELEMENT_NODE}(e)){const n=e.ariaHidden||e.hidden||"none"===e.style.display,r=""===e.dataset.radixToastAnnounceExclude;if(!n)if(r){const n=e.dataset.radixToastAnnounceAlt;n&&t.push(n)}else t.push(...Yw(e))}})),t}const Jw=(0,$t.pM)({__name:"ToastAnnounce",setup(e){const t=qw(),n=function(e=1e3,t={}){const{controls:n=!1,callback:r}=t,a=bm(r??cm,e,t),o=(0,$t.EW)((()=>!a.isPending.value));return n?{ready:o,...a}:o}(1e3),r=(0,jn.KR)(!1);return Sm((()=>{r.value=!0})),(e,a)=>(0,jn.R1)(n)||r.value?((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Ov),{key:0},{default:(0,$t.k6)((()=>[(0,$t.eW)((0,Wn.v_)((0,jn.R1)(t).label.value)+" ",1),(0,$t.RG)(e.$slots,"default")])),_:3})):(0,$t.Q3)("",!0)}}),[eO,tO]=Kf("ToastRoot"),nO=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,(0,$t.pM)({__name:"ToastAnnounceExclude",props:{altText:{},asChild:{type:Boolean},as:{}},setup:e=>(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),{as:e.as,"as-child":e.asChild,"data-radix-toast-announce-exclude":"","data-radix-toast-announce-alt":e.altText||void 0},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},8,["as","as-child","data-radix-toast-announce-alt"]))})),rO=(Boolean,Boolean,Boolean,Boolean,Boolean,(0,$t.pM)({__name:"Toggle",props:{defaultValue:{type:Boolean},pressed:{type:Boolean,default:void 0},disabled:{type:Boolean,default:!1},asChild:{type:Boolean},as:{default:"button"}},emits:["update:pressed"],setup(e,{emit:t}){const n=e,r=t;Zm();const a=Cm(n,"pressed",r,{defaultValue:n.defaultValue,passive:void 0===n.pressed});function o(){a.value=!a.value}const i=(0,$t.EW)((()=>a.value?"on":"off"));return(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(ug),{type:"button"===e.as?"button":void 0,"as-child":n.asChild,as:e.as,"aria-pressed":(0,jn.R1)(a),"data-state":i.value,"data-disabled":e.disabled?"":void 0,disabled:e.disabled,onClick:o},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{pressed:(0,jn.R1)(a)})])),_:3},8,["type","as-child","as","aria-pressed","data-state","data-disabled","disabled"]))}})),[aO,oO]=Kf("ToggleGroupRoot"),[iO,sO]=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Kf("ToolbarRoot")),lO=(Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,"tooltip.open"),[cO,uO]=Kf("TooltipProvider"),dO=(0,$t.pM)({inheritAttrs:!1,__name:"TooltipProvider",props:{delayDuration:{default:700},skipDelayDuration:{default:300},disableHoverableContent:{type:Boolean,default:!1},disableClosingTrigger:{type:Boolean},disabled:{type:Boolean},ignoreNonKeyboardFocus:{type:Boolean,default:!1}},setup(e){const t=e,{delayDuration:n,skipDelayDuration:r,disableHoverableContent:a,disableClosingTrigger:o,ignoreNonKeyboardFocus:i,disabled:s}=(0,jn.QW)(t);Zm();const l=(0,jn.KR)(!0),c=(0,jn.KR)(!1),{start:u,stop:d}=bm((()=>{l.value=!0}),r,{immediate:!1});return uO({isOpenDelayed:l,delayDuration:n,onOpen(){d(),l.value=!1},onClose(){u()},isPointerInTransitRef:c,disableHoverableContent:a,disableClosingTrigger:o,disabled:s,ignoreNonKeyboardFocus:i}),(e,t)=>(0,$t.RG)(e.$slots,"default")}}),[pO,hO]=Kf("TooltipRoot"),fO=(0,$t.pM)({__name:"TooltipRoot",props:{defaultOpen:{type:Boolean,default:!1},open:{type:Boolean,default:void 0},delayDuration:{default:void 0},disableHoverableContent:{type:Boolean,default:void 0},disableClosingTrigger:{type:Boolean,default:void 0},disabled:{type:Boolean,default:void 0},ignoreNonKeyboardFocus:{type:Boolean,default:void 0}},emits:["update:open"],setup(e,{emit:t}){const n=e,r=t;Zm();const a=cO(),o=(0,$t.EW)((()=>n.disableHoverableContent??a.disableHoverableContent.value)),i=(0,$t.EW)((()=>n.disableClosingTrigger??a.disableClosingTrigger.value)),s=(0,$t.EW)((()=>n.disabled??a.disabled.value)),l=(0,$t.EW)((()=>n.delayDuration??a.delayDuration.value)),c=(0,$t.EW)((()=>n.ignoreNonKeyboardFocus??a.ignoreNonKeyboardFocus.value)),u=Cm(n,"open",r,{defaultValue:n.defaultOpen,passive:void 0===n.open});(0,$t.wB)(u,(e=>{a.onClose&&(e?(a.onOpen(),document.dispatchEvent(new CustomEvent(lO))):a.onClose())}));const d=(0,jn.KR)(!1),p=(0,jn.KR)(),h=(0,$t.EW)((()=>u.value?d.value?"delayed-open":"instant-open":"closed")),{start:f,stop:m}=bm((()=>{d.value=!0,u.value=!0}),l,{immediate:!1});function g(){m(),d.value=!1,u.value=!0}function v(){m(),u.value=!1}return hO({contentId:"",open:u,stateAttribute:h,trigger:p,onTriggerChange(e){p.value=e},onTriggerEnter(){a.isOpenDelayed.value?f():g()},onTriggerLeave(){o.value?v():m()},onOpen:g,onClose:v,disableHoverableContent:o,disableClosingTrigger:i,disabled:s,ignoreNonKeyboardFocus:c}),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(lv),null,{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{open:(0,jn.R1)(u)})])),_:3}))}}),mO=(0,$t.pM)({__name:"TooltipTrigger",props:{asChild:{type:Boolean},as:{default:"button"}},setup(e){const t=e,n=pO(),r=cO();n.contentId||(n.contentId=og(void 0,"radix-vue-tooltip-content"));const{forwardRef:a,currentElement:o}=Zm(),i=(0,jn.KR)(!1),s=(0,jn.KR)(!1),l=(0,$t.EW)((()=>n.disabled.value?{}:{click:m,focus:h,pointermove:d,pointerleave:p,pointerdown:u,blur:f}));function c(){setTimeout((()=>{i.value=!1}),1)}function u(){i.value=!0,document.addEventListener("pointerup",c,{once:!0})}function d(e){"touch"!==e.pointerType&&!s.value&&!r.isPointerInTransitRef.value&&(n.onTriggerEnter(),s.value=!0)}function p(){n.onTriggerLeave(),s.value=!1}function h(e){var t,r;i.value||n.ignoreNonKeyboardFocus.value&&(null==(r=(t=e.target).matches)||!r.call(t,":focus-visible"))||n.onOpen()}function f(){n.onClose()}function m(){n.disableClosingTrigger.value||n.onClose()}return(0,$t.sV)((()=>{n.onTriggerChange(o.value)})),(e,r)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(cv),{"as-child":""},{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(ug),(0,$t.v6)({ref:(0,jn.R1)(a),"aria-describedby":(0,jn.R1)(n).open.value?(0,jn.R1)(n).contentId:void 0,"data-state":(0,jn.R1)(n).stateAttribute.value,as:e.as,"as-child":t.asChild,"data-grace-area-trigger":""},(0,$t.Tb)(l.value)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16,["aria-describedby","data-state","as","as-child"])])),_:3}))}}),gO=(0,$t.pM)({__name:"TooltipContentImpl",props:{ariaLabel:{},asChild:{type:Boolean},as:{},side:{default:"top"},sideOffset:{default:0},align:{default:"center"},alignOffset:{},avoidCollisions:{type:Boolean,default:!0},collisionBoundary:{default:()=>[]},collisionPadding:{default:0},arrowPadding:{default:0},sticky:{default:"partial"},hideWhenDetached:{type:Boolean,default:!1}},emits:["escapeKeyDown","pointerDownOutside"],setup(e,{emit:t}){const n=e,r=t,a=pO(),{forwardRef:o}=Zm(),i=(0,$t.Ht)(),s=(0,$t.EW)((()=>{var e;return null==(e=i.default)?void 0:e.call(i)})),l=(0,$t.EW)((()=>{var e;if(n.ariaLabel)return n.ariaLabel;let t="";function r(e){"string"==typeof e.children&&e.type!==$t.Mw?t+=e.children:Array.isArray(e.children)&&e.children.forEach((e=>r(e)))}return null==(e=s.value)||e.forEach((e=>r(e))),t})),c=(0,$t.EW)((()=>{const{ariaLabel:e,...t}=n;return t}));return(0,$t.sV)((()=>{Om(window,"scroll",(e=>{const t=e.target;null!=t&&t.contains(a.trigger.value)&&a.onClose()})),Om(window,lO,a.onClose)})),(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Cg),{"as-child":"","disable-outside-pointer-events":!1,onEscapeKeyDown:t[0]||(t[0]=e=>r("escapeKeyDown",e)),onPointerDownOutside:t[1]||(t[1]=e=>{var t;(0,jn.R1)(a).disableClosingTrigger.value&&null!=(t=(0,jn.R1)(a).trigger.value)&&t.contains(e.target)&&e.preventDefault(),r("pointerDownOutside",e)}),onFocusOutside:t[2]||(t[2]=(0,Rr.D$)((()=>{}),["prevent"])),onDismiss:t[3]||(t[3]=e=>(0,jn.R1)(a).onClose())},{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(gv),(0,$t.v6)({ref:(0,jn.R1)(o),"data-state":(0,jn.R1)(a).stateAttribute.value},{...e.$attrs,...c.value},{style:{"--radix-tooltip-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-tooltip-content-available-width":"var(--radix-popper-available-width)","--radix-tooltip-content-available-height":"var(--radix-popper-available-height)","--radix-tooltip-trigger-width":"var(--radix-popper-anchor-width)","--radix-tooltip-trigger-height":"var(--radix-popper-anchor-height)"}}),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default"),(0,$t.bF)((0,jn.R1)(Ov),{id:(0,jn.R1)(a).contentId,role:"tooltip"},{default:(0,$t.k6)((()=>[(0,$t.eW)((0,Wn.v_)(l.value),1)])),_:1},8,["id"])])),_:3},16,["data-state"])])),_:3}))}}),vO=(0,$t.pM)({__name:"TooltipContentHoverable",props:{ariaLabel:{},asChild:{type:Boolean},as:{},side:{},sideOffset:{},align:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean}},setup(e){const t=Vm(e),{forwardRef:n,currentElement:r}=Zm(),{trigger:a,onClose:o}=pO(),i=cO(),{isPointerInTransit:s,onPointerExit:l}=Gm(a,r);return i.isPointerInTransitRef=s,l((()=>{o()})),(e,r)=>((0,$t.uX)(),(0,$t.Wv)(gO,(0,$t.v6)({ref:(0,jn.R1)(n)},(0,jn.R1)(t)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))}}),bO=(0,$t.pM)({__name:"TooltipContent",props:{forceMount:{type:Boolean},ariaLabel:{},asChild:{type:Boolean},as:{},side:{default:"top"},sideOffset:{},align:{},alignOffset:{},avoidCollisions:{type:Boolean},collisionBoundary:{},collisionPadding:{},arrowPadding:{},sticky:{},hideWhenDetached:{type:Boolean}},emits:["escapeKeyDown","pointerDownOutside"],setup(e,{emit:t}){const n=e,r=t,a=pO(),o=Hm(n,r),{forwardRef:i}=Zm();return(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(mg),{present:e.forceMount||(0,jn.R1)(a).open.value},{default:(0,$t.k6)((()=>[((0,$t.uX)(),(0,$t.Wv)((0,$t.$y)((0,jn.R1)(a).disableHoverableContent.value?gO:vO),(0,$t.v6)({ref:(0,jn.R1)(i)},(0,jn.R1)(o)),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default")])),_:3},16))])),_:3},8,["present"]))}});Boolean,Boolean,Boolean;const[yO,wO]=Kf("TreeRoot");Boolean,Boolean,Boolean,Boolean,Boolean,Boolean,Boolean;var OO=n(6454),kO=n(8428),xO=n(8663),EO=n(1089),SO=(e=>(e[e.Open=0]="Open",e[e.Closed=1]="Closed",e))(SO||{}),_O=(e=>(e[e.Pointer=0]="Pointer",e[e.Other=1]="Other",e))(_O||{});let CO=Symbol("MenuContext");function TO(e){let t=(0,$t.WQ)(CO,null);if(null===t){let t=new Error(`<${e} /> is missing a parent <Menu /> component.`);throw Error.captureStackTrace&&Error.captureStackTrace(t,TO),t}return t}let AO=(0,$t.pM)({name:"Menu",props:{as:{type:[Object,String],default:"template"}},setup(e,{slots:t,attrs:n}){let r=(0,jn.KR)(1),a=(0,jn.KR)(null),o=(0,jn.KR)(null),i=(0,jn.KR)([]),s=(0,jn.KR)(""),l=(0,jn.KR)(null),c=(0,jn.KR)(1);function u(e=e=>e){let t=null!==l.value?i.value[l.value]:null,n=(0,up.wl)(e(i.value.slice()),(e=>(0,jd.t)(e.dataRef.domRef))),r=t?n.indexOf(t):null;return-1===r&&(r=null),{items:n,activeItemIndex:r}}let d={menuState:r,buttonRef:a,itemsRef:o,items:i,searchQuery:s,activeItemIndex:l,activationTrigger:c,closeMenu:()=>{r.value=1,l.value=null},openMenu:()=>r.value=0,goToItem(e,t,n){let r=u(),a=(0,EO.X)(e===EO.B.Specific?{focus:EO.B.Specific,id:t}:{focus:e},{resolveItems:()=>r.items,resolveActiveIndex:()=>r.activeItemIndex,resolveId:e=>e.id,resolveDisabled:e=>e.dataRef.disabled});s.value="",l.value=a,c.value=null!=n?n:1,i.value=r.items},search(e){let t=""!==s.value?0:1;s.value+=e.toLowerCase();let n=(null!==l.value?i.value.slice(l.value+t).concat(i.value.slice(0,l.value+t)):i.value).find((e=>e.dataRef.textValue.startsWith(s.value)&&!e.dataRef.disabled)),r=n?i.value.indexOf(n):-1;-1===r||r===l.value||(l.value=r,c.value=1)},clearSearch(){s.value=""},registerItem(e,t){let n=u((n=>[...n,{id:e,dataRef:t}]));i.value=n.items,l.value=n.activeItemIndex,c.value=1},unregisterItem(e){let t=u((t=>{let n=t.findIndex((t=>t.id===e));return-1!==n&&t.splice(n,1),t}));i.value=t.items,l.value=t.activeItemIndex,c.value=1}};return(0,tp.j)([a,o],((e,t)=>{var n;d.closeMenu(),(0,up.Bm)(t,up.MZ.Loose)||(e.preventDefault(),null==(n=(0,jd.t)(a))||n.focus())}),(0,$t.EW)((()=>0===r.value))),(0,$t.Gt)(CO,d),(0,lp.Fo)((0,$t.EW)((()=>(0,dp.Y)(r.value,{0:lp.Uw.Open,1:lp.Uw.Closed})))),()=>{let a={open:0===r.value,close:d.closeMenu};return(0,Qd.XX)({ourProps:{},theirProps:e,slot:a,slots:t,attrs:n,name:"Menu"})}}}),RO=(0,$t.pM)({name:"MenuButton",props:{disabled:{type:Boolean,default:!1},as:{type:[Object,String],default:"button"},id:{type:String,default:null}},setup(e,{attrs:t,slots:n,expose:r}){var a;let o=null!=(a=e.id)?a:`headlessui-menu-button-${(0,ep.B)()}`,i=TO("MenuButton");function s(e){switch(e.key){case cp.D.Space:case cp.D.Enter:case cp.D.ArrowDown:e.preventDefault(),e.stopPropagation(),i.openMenu(),(0,$t.dY)((()=>{var e;null==(e=(0,jd.t)(i.itemsRef))||e.focus({preventScroll:!0}),i.goToItem(EO.B.First)}));break;case cp.D.ArrowUp:e.preventDefault(),e.stopPropagation(),i.openMenu(),(0,$t.dY)((()=>{var e;null==(e=(0,jd.t)(i.itemsRef))||e.focus({preventScroll:!0}),i.goToItem(EO.B.Last)}))}}function l(e){e.key===cp.D.Space&&e.preventDefault()}function c(t){e.disabled||(0===i.menuState.value?(i.closeMenu(),(0,$t.dY)((()=>{var e;return null==(e=(0,jd.t)(i.buttonRef))?void 0:e.focus({preventScroll:!0})}))):(t.preventDefault(),i.openMenu(),function(e){requestAnimationFrame((()=>requestAnimationFrame(e)))}((()=>{var e;return null==(e=(0,jd.t)(i.itemsRef))?void 0:e.focus({preventScroll:!0})}))))}r({el:i.buttonRef,$el:i.buttonRef});let u=(0,np.c)((0,$t.EW)((()=>({as:e.as,type:t.type}))),i.buttonRef);return()=>{var r;let a={open:0===i.menuState.value},{...d}=e,p={ref:i.buttonRef,id:o,type:u.value,"aria-haspopup":"menu","aria-controls":null==(r=(0,jd.t)(i.itemsRef))?void 0:r.id,"aria-expanded":0===i.menuState.value,onKeydown:s,onKeyup:l,onClick:c};return(0,Qd.XX)({ourProps:p,theirProps:d,slot:a,attrs:t,slots:n,name:"MenuButton"})}}}),PO=(0,$t.pM)({name:"MenuItems",props:{as:{type:[Object,String],default:"div"},static:{type:Boolean,default:!1},unmount:{type:Boolean,default:!0},id:{type:String,default:null}},setup(e,{attrs:t,slots:n,expose:r}){var a;let o=null!=(a=e.id)?a:`headlessui-menu-items-${(0,ep.B)()}`,i=TO("MenuItems"),s=(0,jn.KR)(null);function l(e){var t;switch(s.value&&clearTimeout(s.value),e.key){case cp.D.Space:if(""!==i.searchQuery.value)return e.preventDefault(),e.stopPropagation(),i.search(e.key);case cp.D.Enter:if(e.preventDefault(),e.stopPropagation(),null!==i.activeItemIndex.value){let e=i.items.value[i.activeItemIndex.value];null==(t=(0,jd.t)(e.dataRef.domRef))||t.click()}i.closeMenu(),(0,up.Fh)((0,jd.t)(i.buttonRef));break;case cp.D.ArrowDown:return e.preventDefault(),e.stopPropagation(),i.goToItem(EO.B.Next);case cp.D.ArrowUp:return e.preventDefault(),e.stopPropagation(),i.goToItem(EO.B.Previous);case cp.D.Home:case cp.D.PageUp:return e.preventDefault(),e.stopPropagation(),i.goToItem(EO.B.First);case cp.D.End:case cp.D.PageDown:return e.preventDefault(),e.stopPropagation(),i.goToItem(EO.B.Last);case cp.D.Escape:e.preventDefault(),e.stopPropagation(),i.closeMenu(),(0,$t.dY)((()=>{var e;return null==(e=(0,jd.t)(i.buttonRef))?void 0:e.focus({preventScroll:!0})}));break;case cp.D.Tab:e.preventDefault(),e.stopPropagation(),i.closeMenu(),(0,$t.dY)((()=>(0,up.p9)((0,jd.t)(i.buttonRef),e.shiftKey?up.BD.Previous:up.BD.Next)));break;default:1===e.key.length&&(i.search(e.key),s.value=setTimeout((()=>i.clearSearch()),350))}}function c(e){e.key===cp.D.Space&&e.preventDefault()}r({el:i.itemsRef,$el:i.itemsRef}),(0,xO.i)({container:(0,$t.EW)((()=>(0,jd.t)(i.itemsRef))),enabled:(0,$t.EW)((()=>0===i.menuState.value)),accept:e=>"menuitem"===e.getAttribute("role")?NodeFilter.FILTER_REJECT:e.hasAttribute("role")?NodeFilter.FILTER_SKIP:NodeFilter.FILTER_ACCEPT,walk(e){e.setAttribute("role","none")}});let u=(0,lp.O_)(),d=(0,$t.EW)((()=>null!==u?(u.value&lp.Uw.Open)===lp.Uw.Open:0===i.menuState.value));return()=>{var r,a;let s={open:0===i.menuState.value},{...u}=e,p={"aria-activedescendant":null===i.activeItemIndex.value||null==(r=i.items.value[i.activeItemIndex.value])?void 0:r.id,"aria-labelledby":null==(a=(0,jd.t)(i.buttonRef))?void 0:a.id,id:o,onKeydown:l,onKeyup:c,role:"menu",tabIndex:0,ref:i.itemsRef};return(0,Qd.XX)({ourProps:p,theirProps:u,slot:s,attrs:t,slots:n,features:Qd.O5.RenderStrategy|Qd.O5.Static,visible:d.value,name:"MenuItems"})}}}),DO=(0,$t.pM)({name:"MenuItem",inheritAttrs:!1,props:{as:{type:[Object,String],default:"template"},disabled:{type:Boolean,default:!1},id:{type:String,default:null}},setup(e,{slots:t,attrs:n,expose:r}){var a;let o=null!=(a=e.id)?a:`headlessui-menu-item-${(0,ep.B)()}`,i=TO("MenuItem"),s=(0,jn.KR)(null);r({el:s,$el:s});let l=(0,$t.EW)((()=>null!==i.activeItemIndex.value&&i.items.value[i.activeItemIndex.value].id===o)),c=(0,OO.q)(s),u=(0,$t.EW)((()=>({disabled:e.disabled,get textValue(){return c()},domRef:s})));function d(t){if(e.disabled)return t.preventDefault();i.closeMenu(),(0,up.Fh)((0,jd.t)(i.buttonRef))}function p(){if(e.disabled)return i.goToItem(EO.B.Nothing);i.goToItem(EO.B.Specific,o)}(0,$t.sV)((()=>i.registerItem(o,u))),(0,$t.hi)((()=>i.unregisterItem(o))),(0,$t.nT)((()=>{0===i.menuState.value&&l.value&&0!==i.activationTrigger.value&&(0,$t.dY)((()=>{var e,t;return null==(t=null==(e=(0,jd.t)(s))?void 0:e.scrollIntoView)?void 0:t.call(e,{block:"nearest"})}))}));let h=(0,kO.J)();function f(e){h.update(e)}function m(t){h.wasMoved(t)&&(e.disabled||l.value||i.goToItem(EO.B.Specific,o,0))}function g(t){h.wasMoved(t)&&(e.disabled||l.value&&i.goToItem(EO.B.Nothing))}return()=>{let{disabled:r,...a}=e,c={active:l.value,disabled:r,close:i.closeMenu};return(0,Qd.XX)({ourProps:{id:o,ref:s,role:"menuitem",tabIndex:!0===r?void 0:-1,"aria-disabled":!0===r||void 0,onClick:d,onFocus:p,onPointerenter:f,onMouseenter:f,onPointermove:m,onMousemove:m,onPointerleave:g,onMouseleave:g},theirProps:{...n,...a},slot:c,attrs:n,slots:t,name:"MenuItem"})}}});const MO={class:"custom-scroll min-h-0 flex-1"},$O={class:"flex flex-col p-0.75"},NO=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarDropdownMenu",props:{is:{}},setup(e){const{cx:t}=zn();return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,$t.$y)(e.is??"div"),(0,$t.v6)({role:"menu",tabindex:"0"},(0,jn.R1)(t)("relative flex w-56")),{default:(0,$t.k6)((()=>[(0,$t.Lk)("div",MO,[(0,$t.Lk)("div",$O,[(0,$t.RG)(e.$slots,"default")]),(0,$t.RG)(e.$slots,"backdrop",{},(()=>[(0,$t.bF)((0,jn.R1)(zh))]))])])),_:3},16))}}),IO=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarDropdown",props:{placement:{},offset:{},resize:{type:Boolean},target:{},middleware:{},teleport:{type:[Boolean,String]}},setup(e){const{cx:t}=zn();return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(AO),null,{default:(0,$t.k6)((({open:n})=>[(0,$t.bF)((0,jn.R1)(Uh),(0,$t.v6)(e.$props,{placement:e.placement??"bottom-start"}),{floating:(0,$t.k6)((({width:r})=>[(0,$t.bF)(NO,(0,$t.v6)({is:(0,jn.R1)(PO),style:{width:r}},(0,jn.R1)(t)("max-h-[inherit]")),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"items",{open:n})])),_:2},1040,["is","style"])])),default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(RO),{as:"template"},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{open:n})])),_:2},1024)])),_:2},1040,["placement"])])),_:3}))}}),LO={key:1,class:"rounded border bg-b-2 p-3 text-sm"},BO={key:0,class:"mt-2 rounded border bg-b-1 p-2 font-code text-c-2"},FO=(0,$t.pM)({__name:"ScalarErrorBoundary",setup(e){const t=(0,jn.KR)(!1),n=(0,jn.KR)();return(0,$t.qG)(((e,r,a)=>(console.error("[ERROR]",e,a),t.value=!0,n.value=e,!1))),(e,r)=>{var a,o;return t.value?((0,$t.uX)(),(0,$t.CE)("div",LO,[r[0]||(r[0]=(0,$t.Lk)("div",{class:"p-2"},"Oops, something went wrong here.",-1)),n.value?((0,$t.uX)(),(0,$t.CE)("div",BO,(0,Wn.v_)(null==(a=n.value)?void 0:a.name)+": "+(0,Wn.v_)(null==(o=n.value)?void 0:o.message),1)):(0,$t.Q3)("",!0)])):(0,$t.RG)(e.$slots,"default",{key:0})}}}),QO=(Boolean,["ariaDisabled"]),UO={class:"sr-only"},zO=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarIconButton",props:{label:{},icon:{},disabled:{type:Boolean},variant:{default:"ghost"},size:{default:"md"},thickness:{}},setup(e){const{cx:t}=zn(),n=Fn({base:"scalar-icon-button grid aspect-square cursor-pointer rounded",variants:{size:{xxs:"size-3.5 p-0.5",xs:"size-5 p-1",sm:"size-6 p-1",md:"size-10 p-3",full:"h-full w-full"},disabled:{true:"cursor-not-allowed shadow-none"},variant:Xn}});return(e,r)=>((0,$t.uX)(),(0,$t.CE)("button",(0,$t.v6)({ariaDisabled:e.disabled||void 0,type:"button"},(0,jn.R1)(t)((0,jn.R1)(n)({size:e.size,variant:e.variant,disabled:e.disabled}))),[(0,$t.bF)((0,jn.R1)(wd),{icon:e.icon,thickness:e.thickness},null,8,["icon","thickness"]),(0,$t.Lk)("span",UO,(0,Wn.v_)(e.label),1)],16,QO))}});var jO=n(487);const WO=(0,$t.pM)({__name:"ScalarListboxItem",props:{option:{},style:{}},setup(e){const t=Fn({base:["group/item","flex min-w-0 items-center gap-1.5 rounded px-2 py-1.5 text-left","truncate bg-transparent text-c-1","cursor-pointer hover:bg-b-2"],variants:{selected:{true:"text-c-1"},active:{true:"bg-b-2"},disabled:{true:"pointer-events-none opacity-50"}}});return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(jO.h7),{as:"template",disabled:e.option.disabled,value:e.option},{default:(0,$t.k6)((({active:n,selected:r})=>[(0,$t.Lk)("li",{class:(0,Wn.C4)((0,jn.R1)(Qn)((0,jn.R1)(t)({active:n,selected:r,disabled:e.option.disabled})))},[(0,$t.bF)(Dd,{selected:r,style:(0,Wn.Tr)(e.style)},null,8,["selected","style"]),(0,$t.Lk)("span",{class:(0,Wn.C4)(["inline-block min-w-0 flex-1 truncate",e.option.color?e.option.color:"text-c-1"])},(0,Wn.v_)(e.option.label),3)],2)])),_:1},8,["disabled","value"]))}}),XO={class:"custom-scroll min-h-0 flex-1"},qO=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarListbox",props:{options:{},id:{},label:{},multiple:{type:Boolean},modelValue:{},placement:{},offset:{},resize:{type:Boolean},target:{},middleware:{},teleport:{type:[Boolean,String]}},emits:["update:modelValue"],setup(e){const{cx:t}=zn();return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(jO.WF),{modelValue:e.modelValue,multiple:e.multiple,"onUpdate:modelValue":n[0]||(n[0]=t=>e.$emit("update:modelValue",t))},{default:(0,$t.k6)((({open:n})=>[e.label?((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(jO.E9),{key:0,class:"sr-only"},{default:(0,$t.k6)((()=>[(0,$t.eW)((0,Wn.v_)(e.label),1)])),_:1})):(0,$t.Q3)("",!0),(0,$t.bF)((0,jn.R1)(Uh),(0,$t.v6)(e.$props,{placement:e.placement??"bottom-start"}),{floating:(0,$t.k6)((({width:r})=>[n?((0,$t.uX)(),(0,$t.CE)("div",(0,$t.v6)({key:0,style:{width:r}},(0,jn.R1)(t)("relative flex max-h-[inherit] w-40 rounded text-sm")),[(0,$t.Lk)("div",XO,[(0,$t.bF)((0,jn.R1)(jO.wh),{class:"flex flex-col gap-0.75 p-0.75"},{default:(0,$t.k6)((()=>[((0,$t.uX)(!0),(0,$t.CE)($t.FK,null,(0,$t.pI)(e.options,(t=>((0,$t.uX)(),(0,$t.Wv)(WO,{key:t.id,option:t,style:(0,Wn.Tr)(e.multiple?"checkbox":"radio")},null,8,["option","style"])))),128))])),_:1})]),(0,$t.bF)((0,jn.R1)(zh))],16)):(0,$t.Q3)("",!0)])),default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(jO.go),{id:e.id,as:"template",class:"justify-start focus:outline-none focus-visible:ring-1 focus-visible:ring-c-accent"},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{open:n})])),_:2},1032,["id"])])),_:2},1040,["placement"])])),_:3},8,["modelValue","multiple"]))}});var VO=n(9261);const HO=["innerHTML"],ZO=(0,$t.pM)({__name:"ScalarMarkdown",props:{value:{},withImages:{type:Boolean,default:!1},transform:{},transformType:{},clamp:{type:[String,Boolean]}},setup(e){const t=e,n=(0,$t.EW)((()=>(0,VO.Vg)(t.value??"",{removeTags:t.withImages?[]:["img","picture"],transform:t.transform,transformType:t.transformType})));return(0,$t.SS)((async()=>await new Promise((e=>setTimeout(e,1))))),(e,t)=>((0,$t.uX)(),(0,$t.CE)("div",{class:(0,Wn.C4)(["markdown text-ellipsis",{"line-clamp-4":e.clamp}]),style:(0,Wn.Tr)({"-webkit-line-clamp":"string"==typeof e.clamp?e.clamp:void 0}),innerHTML:n.value},null,14,HO))}}),GO={Root:Qb,Trigger:Ub,Portal:zb,Content:jb,Arrow:Wb,Item:Xb,Group:qb,Separator:Vb,CheckboxItem:Hb,ItemIndicator:Zb,Label:Gb,RadioGroup:Kb,RadioItem:Yb,Sub:Jb,SubContent:ey,SubTrigger:ty},KO=(Boolean,(0,$t.pM)({__name:"ScalarDropdownButton",props:{is:{default:"button"},active:{type:Boolean},disabled:{type:Boolean}},setup(e){const t=Fn({base:["flex h-8 min-w-0 items-center gap-1.5 rounded px-2.5 py-1.5 text-left","truncate  no-underline text-sm text-c-1","cursor-pointer hover:bg-b-2 hover:text-c-1"],variants:{disabled:{true:"pointer-events-none text-c-3"},active:{true:"bg-b-2 text-c-1"}}});return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,$t.$y)(e.is),{class:(0,Wn.C4)(["item",(0,jn.R1)(Qn)("scalar-dropdown-item",(0,jn.R1)(t)({active:e.active,disabled:e.disabled}))]),type:"button"===e.is?"button":void 0},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{},void 0,!0)])),_:3},8,["class","type"]))}})),YO=Gn(KO,[["__scopeId","data-v-942bce7e"]]),JO={key:1,class:"size-3"},ek=(Boolean,Boolean,{class:"-mx-0.75 my-0.75 h-border bg-border"}),tk=Gn({},[["render",function(e,t,n,r,a,o){return(0,$t.uX)(),(0,$t.CE)("div",ek)}]]),nk=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarPopover",props:{placement:{},offset:{},resize:{type:Boolean},target:{},middleware:{},teleport:{type:[Boolean,String]}},setup(e){const{cx:t}=zn();return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(bp),{as:"template"},{default:(0,$t.k6)((({open:n})=>[(0,$t.bF)((0,jn.R1)(Uh),(0,Wn._B)((0,$t.Ng)(e.$props)),{floating:(0,$t.k6)((({width:r,height:a})=>[(0,$t.bF)((0,jn.R1)(wp),(0,$t.v6)({style:{width:r,height:a}},(0,jn.R1)(t)("relative flex flex-col p-0.75")),{default:(0,$t.k6)((({close:t})=>[(0,$t.RG)(e.$slots,"popover",{close:()=>t(),open:n}),(0,$t.RG)(e.$slots,"backdrop",{open:n},(()=>[(0,$t.bF)((0,jn.R1)(zh))]))])),_:2},1040,["style"])])),default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(yp),{as:"template"},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{open:n})])),_:2},1024)])),_:2},1040)])),_:3}))}}),rk=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarSearchResultList",props:{noResults:{type:Boolean}},setup(e){const{cx:t}=zn();return(e,n)=>((0,$t.uX)(),(0,$t.CE)("ul",(0,$t.v6)({role:"listbox"},(0,jn.R1)(t)("flex flex-col")),[e.noResults?(0,$t.RG)(e.$slots,"noResults",{key:0},(()=>[n[0]||(n[0]=(0,$t.Lk)("div",{class:"flex flex-col items-center gap-2 px-3 py-4"},[(0,$t.Lk)("div",{class:"rotate-90 text-lg font-bold"},":("),(0,$t.Lk)("div",{class:"text-sm font-medium text-c-2"},"No results found")],-1))])):(0,$t.Q3)("",!0),(0,$t.RG)(e.$slots,"default")],16))}}),ak={class:"contents",role:"option"},ok={key:0,class:"flex h-fit items-center text-sm font-medium text-c-3 group-hover:text-c-1"},ik={class:"flex min-w-0 flex-1 flex-col gap-0.75"},sk={class:"flex items-center gap-1"},lk={class:"flex-1 truncate text-sm font-medium"},ck={key:0,class:"text-sm text-c-2"},uk={key:0,class:"truncate text-sm text-c-2"},dk=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarSearchResultItem",props:{icon:{},active:{type:Boolean}},setup(e){const{cx:t}=zn();return(e,n)=>((0,$t.uX)(),(0,$t.CE)("li",ak,[(0,$t.Lk)("a",(0,Wn._B)((0,$t.Ng)((0,jn.R1)(t)("group flex cursor-pointer gap-2.5 rounded px-3 py-1.5 no-underline hover:bg-b-2",{"bg-b-2":e.active}))),[e.icon?((0,$t.uX)(),(0,$t.CE)("div",ok,[(0,$t.RG)(e.$slots,"icon",{},(()=>[e.icon?((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(wd),{key:0,icon:e.icon,size:"sm"},null,8,["icon"])):(0,$t.Q3)("",!0)])),n[0]||(n[0]=(0,$t.Lk)("span",null," ",-1))])):(0,$t.Q3)("",!0),(0,$t.Lk)("div",ik,[(0,$t.Lk)("div",sk,[(0,$t.Lk)("div",lk,[(0,$t.RG)(e.$slots,"default")]),e.$slots.addon?((0,$t.uX)(),(0,$t.CE)("div",ck,[(0,$t.RG)(e.$slots,"addon")])):(0,$t.Q3)("",!0)]),e.$slots.description?((0,$t.uX)(),(0,$t.CE)("div",uk,[(0,$t.RG)(e.$slots,"description")])):(0,$t.Q3)("",!0)])],16)]))}}),pk={class:"flex items-center"},hk={class:"flex-1 min-w-0 flex items-center text-sm text-sidebar-c-2"},fk=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarSidebarFooter",setup(e){const{cx:t}=zn();return(e,n)=>((0,$t.uX)(),(0,$t.CE)("div",(0,Wn._B)((0,$t.Ng)((0,jn.R1)(t)("flex flex-col gap-3 p-3 border-t"))),[(0,$t.RG)(e.$slots,"default"),(0,$t.Lk)("div",pk,[(0,$t.Lk)("div",hk,[(0,$t.RG)(e.$slots,"description",{},(()=>[n[0]||(n[0]=(0,$t.Lk)("a",{class:"no-underline hover:underline",href:"https://www.scalar.com",target:"_blank"}," Powered by Scalar ",-1))]))]),(0,$t.RG)(e.$slots,"toggle",{},(()=>[(0,$t.bF)((0,jn.R1)(Rd))]))])],16))}}),mk={class:"flex items-center gap-1 flex-1"},gk={key:0,class:"size-3.5"},vk={key:0},bk=(Boolean,Boolean,{class:"sr-only"}),yk=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarSidebarGroupToggle",props:{is:{default:"div"},open:{type:Boolean,default:!1},icon:{default:"ChevronRight"}},setup(e){const t=Fn({base:"size-4 -m-px transition-transform duration-100",variants:{open:{true:"rotate-90"}},defaultVariants:{open:!1}}),{cx:n}=zn();return(e,r)=>((0,$t.uX)(),(0,$t.Wv)((0,$t.$y)(e.is),(0,$t.v6)({type:"button"===e.is?"button":void 0},(0,jn.R1)(n)((0,jn.R1)(t)({open:e.open}))),{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{open:e.open},(()=>[(0,$t.bF)((0,jn.R1)(wd),{icon:e.icon},null,8,["icon"])])),(0,$t.Lk)("span",bk,[(0,$t.RG)(e.$slots,"label",{open:e.open},(()=>[(0,$t.eW)((0,Wn.v_)(e.open?"Close":"Open")+" Group ",1)]))])])),_:3},16,["type"]))}}),wk=(Symbol(),Boolean,Boolean,Boolean,["id"]),Ok=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarTeleportRoot",props:{id:{}},setup(e){const t=(e=>{const t=e??`scalar-teleport-${Pd()}`;return(0,$t.Gt)(Fh,`#${t}`),t})(e.id);return(e,n)=>((0,$t.uX)(),(0,$t.CE)($t.FK,null,[(0,$t.RG)(e.$slots,"default"),(0,$t.Lk)("div",{id:(0,jn.R1)(t),class:"scalar-teleport-root contents"},null,8,wk)],64))}}),kk=(Boolean,Boolean,(0,$t.pM)({__name:"ScalarTooltip",props:{click:{},delay:{},skipDelay:{default:1e3},align:{default:"center"},side:{default:"top"},sideOffset:{},class:{},triggerClass:{},resize:{type:Boolean},as:{},disabled:{type:Boolean,default:!1}},emits:["click"],setup(e){const t=e;return(e,n)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(dO),{delayDuration:t.delay,skipDelayDuration:t.skipDelay},{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(fO),null,{default:(0,$t.k6)((()=>[(0,$t.bF)((0,jn.R1)(mO),{as:t.as||"button",class:(0,Wn.C4)(["flex items-center justify-center",[t.resize?"w-full":"",t.triggerClass]]),onClick:t.click},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"trigger")])),_:3},8,["as","class","onClick"]),(0,$t.bF)((0,jn.R1)(Qh),null,{default:(0,$t.k6)((()=>[t.disabled?(0,$t.Q3)("",!0):((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(bO),{key:0,align:t.align,class:(0,Wn.C4)(["scalar-app z-context",t.class]),side:t.side,sideOffset:t.sideOffset},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"content")])),_:3},8,["align","class","side","sideOffset"]))])),_:3})])),_:3})])),_:3},8,["delayDuration","skipDelayDuration"]))}})),xk=(0,$t.pM)({__name:"ScalarVirtualText",props:{text:{},lineHeight:{default:20},containerClass:{default:""},contentClass:{default:""},lineClass:{default:""}},setup(e){const t=e,n=(0,jn.KR)(null),r=(0,jn.KR)(null),a=(0,jn.KR)(0),o=(0,jn.KR)(0),i=(0,$t.EW)((()=>t.text.split("\n"))),s=(0,$t.EW)((()=>i.value.length*t.lineHeight)),l=(0,$t.EW)((()=>Math.floor(a.value/t.lineHeight))),c=(0,$t.EW)((()=>Math.min(Math.ceil((a.value+o.value)/t.lineHeight),i.value.length))),u=(0,$t.EW)((()=>{const e=Math.max(0,l.value-10),t=Math.min(i.value.length,c.value+10);return i.value.slice(e,t)})),d=(0,$t.EW)((()=>({height:`${s.value}px`,transform:`translateY(${Math.max(0,l.value-10)*t.lineHeight}px)`}))),p=()=>n.value&&(a.value=n.value.scrollTop),h=()=>n.value&&(o.value=n.value.clientHeight);return(0,$t.sV)((()=>{h(),window.addEventListener("resize",h)})),(0,$t.xo)((()=>{window.removeEventListener("resize",h)})),(0,$t.nT)((()=>{r.value&&(r.value.style.transform=`translateY(${Math.max(0,l.value-10)*t.lineHeight}px)`)})),(e,a)=>((0,$t.uX)(),(0,$t.CE)("div",{ref_key:"containerRef",ref:n,class:(0,Wn.C4)(["scalar-virtual-text overflow-auto",e.containerClass]),onScroll:p},[(0,$t.Lk)("code",{ref_key:"contentRef",ref:r,class:(0,Wn.C4)(["scalar-virtual-text-content",e.contentClass]),style:(0,Wn.Tr)(d.value)},[((0,$t.uX)(!0),(0,$t.CE)($t.FK,null,(0,$t.pI)(u.value,((n,r)=>((0,$t.uX)(),(0,$t.CE)("div",{key:l.value+r,class:(0,Wn.C4)(["scalar-virtual-text-line",e.lineClass]),style:(0,Wn.Tr)({height:`${t.lineHeight}px`,lineHeight:`${t.lineHeight}px`})},(0,Wn.v_)(n),7)))),128))],6)],34))}}),Ek=(0,$t.pM)({__name:"ScalarDropdownItem",props:{disabled:{type:Boolean}},emits:["click"],setup:e=>(e,t)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(DO),{disabled:e.disabled},{default:(0,$t.k6)((({active:n})=>[(0,$t.bF)(YO,{active:n,disabled:e.disabled,onClick:t[0]||(t[0]=t=>e.$emit("click",t))},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{},void 0,!0)])),_:2},1032,["active","disabled"])])),_:3},8,["disabled"]))}),Sk=Gn(Ek,[["__scopeId","data-v-457c1cd3"]]);let _k=[];!function(){function e(){"loading"!==document.readyState&&((()=>{function e(e){e.target instanceof HTMLElement&&e.target!==document.body&&_k[0]!==e.target&&(_k.unshift(e.target),_k=_k.filter((e=>null!=e&&e.isConnected)),_k.splice(10))}window.addEventListener("click",e,{capture:!0}),window.addEventListener("mousedown",e,{capture:!0}),window.addEventListener("focus",e,{capture:!0}),document.body.addEventListener("click",e,{capture:!0}),document.body.addEventListener("mousedown",e,{capture:!0}),document.body.addEventListener("focus",e,{capture:!0})})(),document.removeEventListener("DOMContentLoaded",e))}"undefined"!=typeof window&&"undefined"!=typeof document&&(document.addEventListener("DOMContentLoaded",e),e())}();var Ck=n(3827);function Tk(e){if(!e)return new Set;if("function"==typeof e)return new Set(e());let t=new Set;for(let n of e.value){let e=(0,jd.t)(n);e instanceof HTMLElement&&t.add(e)}return t}var Ak=(e=>(e[e.None=1]="None",e[e.InitialFocus=2]="InitialFocus",e[e.TabLock=4]="TabLock",e[e.FocusLock=8]="FocusLock",e[e.RestoreFocus=16]="RestoreFocus",e[e.All=30]="All",e))(Ak||{});let Rk=Object.assign((0,$t.pM)({name:"FocusTrap",props:{as:{type:[Object,String],default:"div"},initialFocus:{type:Object,default:null},features:{type:Number,default:30},containers:{type:[Object,Function],default:(0,jn.KR)(new Set)}},inheritAttrs:!1,setup(e,{attrs:t,slots:n,expose:r}){let a=(0,jn.KR)(null);r({el:a,$el:a});let o=(0,$t.EW)((()=>(0,Wd.T)(a))),i=(0,jn.KR)(!1);(0,$t.sV)((()=>i.value=!0)),(0,$t.hi)((()=>i.value=!1)),function({ownerDocument:e},t){let n=function(e){let t=(0,jn.KR)(_k.slice());return(0,$t.wB)([e],(([e],[n])=>{!0===n&&!1===e?(0,Ck._)((()=>{t.value.splice(0)})):!1===n&&!0===e&&(t.value=_k.slice())}),{flush:"post"}),()=>{var e;return null!=(e=t.value.find((e=>null!=e&&e.isConnected)))?e:null}}(t);(0,$t.sV)((()=>{(0,$t.nT)((()=>{var r,a;t.value||(null==(r=e.value)?void 0:r.activeElement)===(null==(a=e.value)?void 0:a.body)&&(0,up.pW)(n())}),{flush:"post"})})),(0,$t.hi)((()=>{t.value&&(0,up.pW)(n())}))}({ownerDocument:o},(0,$t.EW)((()=>i.value&&Boolean(16&e.features))));let s=function({ownerDocument:e,container:t,initialFocus:n},r){let a=(0,jn.KR)(null),o=(0,jn.KR)(!1);return(0,$t.sV)((()=>o.value=!0)),(0,$t.hi)((()=>o.value=!1)),(0,$t.sV)((()=>{(0,$t.wB)([t,n,r],((i,s)=>{if(i.every(((e,t)=>(null==s?void 0:s[t])===e))||!r.value)return;let l=(0,jd.t)(t);l&&(0,Ck._)((()=>{var t,r;if(!o.value)return;let i=(0,jd.t)(n),s=null==(t=e.value)?void 0:t.activeElement;if(i){if(i===s)return void(a.value=s)}else if(l.contains(s))return void(a.value=s);i?(0,up.pW)(i):(0,up.CU)(l,up.BD.First|up.BD.NoScroll)===up.Me.Error&&console.warn("There are no focusable elements inside the <FocusTrap />"),a.value=null==(r=e.value)?void 0:r.activeElement}))}),{immediate:!0,flush:"post"})})),a}({ownerDocument:o,container:a,initialFocus:(0,$t.EW)((()=>e.initialFocus))},(0,$t.EW)((()=>i.value&&Boolean(2&e.features))));!function({ownerDocument:e,container:t,containers:n,previousActiveElement:r},a){var o;Jd(null==(o=e.value)?void 0:o.defaultView,"focus",(e=>{if(!a.value)return;let o=Tk(n);(0,jd.t)(t)instanceof HTMLElement&&o.add((0,jd.t)(t));let i=r.value;if(!i)return;let s=e.target;s&&s instanceof HTMLElement?Pk(o,s)?(r.value=s,(0,up.pW)(s)):(e.preventDefault(),e.stopPropagation(),(0,up.pW)(i)):(0,up.pW)(r.value)}),!0)}({ownerDocument:o,container:a,containers:e.containers,previousActiveElement:s},(0,$t.EW)((()=>i.value&&Boolean(8&e.features))));let l=sp();function c(e){let t=(0,jd.t)(a);t&&(0,dp.Y)(l.value,{[ip.Forwards]:()=>{(0,up.CU)(t,up.BD.First,{skipElements:[e.relatedTarget]})},[ip.Backwards]:()=>{(0,up.CU)(t,up.BD.Last,{skipElements:[e.relatedTarget]})}})}let u=(0,jn.KR)(!1);function d(e){"Tab"===e.key&&(u.value=!0,requestAnimationFrame((()=>{u.value=!1})))}function p(t){if(!i.value)return;let n=Tk(e.containers);(0,jd.t)(a)instanceof HTMLElement&&n.add((0,jd.t)(a));let r=t.relatedTarget;r instanceof HTMLElement&&"true"!==r.dataset.headlessuiFocusGuard&&(Pk(n,r)||(u.value?(0,up.CU)((0,jd.t)(a),(0,dp.Y)(l.value,{[ip.Forwards]:()=>up.BD.Next,[ip.Backwards]:()=>up.BD.Previous})|up.BD.WrapAround,{relativeTo:t.target}):t.target instanceof HTMLElement&&(0,up.pW)(t.target)))}return()=>{let r={ref:a,onKeydown:d,onFocusout:p},{features:o,initialFocus:i,containers:s,...l}=e;return(0,$t.h)($t.FK,[Boolean(4&o)&&(0,$t.h)(rp.j,{as:"button",type:"button","data-headlessui-focus-guard":!0,onFocus:c,features:rp.O.Focusable}),(0,Qd.XX)({ourProps:r,theirProps:{...t,...l},slot:{},attrs:t,slots:n,name:"FocusTrap"}),Boolean(4&o)&&(0,$t.h)(rp.j,{as:"button",type:"button","data-headlessui-focus-guard":!0,onFocus:c,features:rp.O.Focusable})])}}}),{features:Ak});function Pk(e,t){for(let n of e)if(n.contains(t))return!0;return!1}function Dk(){let e=[],t={addEventListener:(e,n,r,a)=>(e.addEventListener(n,r,a),t.add((()=>e.removeEventListener(n,r,a)))),requestAnimationFrame(...e){let n=requestAnimationFrame(...e);t.add((()=>cancelAnimationFrame(n)))},nextFrame(...e){t.requestAnimationFrame((()=>{t.requestAnimationFrame(...e)}))},setTimeout(...e){let n=setTimeout(...e);t.add((()=>clearTimeout(n)))},microTask(...e){let n={current:!0};return(0,Ck._)((()=>{n.current&&e[0]()})),t.add((()=>{n.current=!1}))},style(e,t,n){let r=e.style.getPropertyValue(t);return Object.assign(e.style,{[t]:n}),this.add((()=>{Object.assign(e.style,{[t]:r})}))},group(e){let t=Dk();return e(t),this.add((()=>t.dispose()))},add:t=>(e.push(t),()=>{let n=e.indexOf(t);if(n>=0)for(let t of e.splice(n,1))t()}),dispose(){for(let t of e.splice(0))t()}};return t}function Mk(){let e;return{before({doc:t}){var n;let r=t.documentElement;e=(null!=(n=t.defaultView)?n:window).innerWidth-r.clientWidth},after({doc:t,d:n}){let r=t.documentElement,a=r.clientWidth-r.offsetWidth,o=e-a;n.style(r,"paddingRight",`${o}px`)}}}var $k=n(2966);function Nk(e){let t={};for(let n of e)Object.assign(t,n(t));return t}let Ik=function(e,t){let n=new Map,r=new Set;return{getSnapshot:()=>n,subscribe:e=>(r.add(e),()=>r.delete(e)),dispatch(e,...a){let o=t[e].call(n,...a);o&&(n=o,r.forEach((e=>e())))}}}(0,{PUSH(e,t){var n;let r=null!=(n=this.get(e))?n:{doc:e,count:0,d:Dk(),meta:new Set};return r.count++,r.meta.add(t),this.set(e,r),this},POP(e,t){let n=this.get(e);return n&&(n.count--,n.meta.delete(t)),this},SCROLL_PREVENT({doc:e,d:t,meta:n}){let r={doc:e,d:t,meta:Nk(n)},a=[(0,$k.un)()?{before({doc:e,d:t,meta:n}){function r(e){return n.containers.flatMap((e=>e())).some((t=>t.contains(e)))}t.microTask((()=>{var n;if("auto"!==window.getComputedStyle(e.documentElement).scrollBehavior){let n=Dk();n.style(e.documentElement,"scrollBehavior","auto"),t.add((()=>t.microTask((()=>n.dispose()))))}let a=null!=(n=window.scrollY)?n:window.pageYOffset,o=null;t.addEventListener(e,"click",(t=>{if(t.target instanceof HTMLElement)try{let n=t.target.closest("a");if(!n)return;let{hash:a}=new URL(n.href),i=e.querySelector(a);i&&!r(i)&&(o=i)}catch{}}),!0),t.addEventListener(e,"touchstart",(e=>{if(e.target instanceof HTMLElement)if(r(e.target)){let n=e.target;for(;n.parentElement&&r(n.parentElement);)n=n.parentElement;t.style(n,"overscrollBehavior","contain")}else t.style(e.target,"touchAction","none")})),t.addEventListener(e,"touchmove",(e=>{if(e.target instanceof HTMLElement){if("INPUT"===e.target.tagName)return;if(r(e.target)){let t=e.target;for(;t.parentElement&&""!==t.dataset.headlessuiPortal&&!(t.scrollHeight>t.clientHeight||t.scrollWidth>t.clientWidth);)t=t.parentElement;""===t.dataset.headlessuiPortal&&e.preventDefault()}else e.preventDefault()}}),{passive:!1}),t.add((()=>{var e;let t=null!=(e=window.scrollY)?e:window.pageYOffset;a!==t&&window.scrollTo(0,a),o&&o.isConnected&&(o.scrollIntoView({block:"nearest"}),o=null)}))}))}}:{},Mk(),{before({doc:e,d:t}){t.style(e.documentElement,"overflow","hidden")}}];a.forEach((({before:e})=>null==e?void 0:e(r))),a.forEach((({after:e})=>null==e?void 0:e(r)))},SCROLL_ALLOW({d:e}){e.dispose()},TEARDOWN({doc:e}){this.delete(e)}});Ik.subscribe((()=>{let e=Ik.getSnapshot(),t=new Map;for(let[n]of e)t.set(n,n.documentElement.style.overflow);for(let n of e.values()){let e="hidden"===t.get(n.doc),r=0!==n.count;(r&&!e||!r&&e)&&Ik.dispatch(n.count>0?"SCROLL_PREVENT":"SCROLL_ALLOW",n),0===n.count&&Ik.dispatch("TEARDOWN",n)}}));let Lk=new Map,Bk=new Map;function Fk(e,t=(0,jn.KR)(!0)){(0,$t.nT)((n=>{var r;if(!t.value)return;let a=(0,jd.t)(e);if(!a)return;n((function(){var e;if(!a)return;let t=null!=(e=Bk.get(a))?e:1;if(1===t?Bk.delete(a):Bk.set(a,t-1),1!==t)return;let n=Lk.get(a);n&&(null===n["aria-hidden"]?a.removeAttribute("aria-hidden"):a.setAttribute("aria-hidden",n["aria-hidden"]),a.inert=n.inert,Lk.delete(a))}));let o=null!=(r=Bk.get(a))?r:0;Bk.set(a,o+1),0===o&&(Lk.set(a,{"aria-hidden":a.getAttribute("aria-hidden"),inert:a.inert}),a.setAttribute("aria-hidden","true"),a.inert=!0)}))}let Qk=Symbol("StackContext");var Uk=(e=>(e[e.Add=0]="Add",e[e.Remove=1]="Remove",e))(Uk||{});var zk=n(1567),jk=(e=>(e[e.Open=0]="Open",e[e.Closed=1]="Closed",e))(jk||{});let Wk=Symbol("DialogContext");function Xk(e){let t=(0,$t.WQ)(Wk,null);if(null===t){let t=new Error(`<${e} /> is missing a parent <Dialog /> component.`);throw Error.captureStackTrace&&Error.captureStackTrace(t,Xk),t}return t}let qk="DC8F892D-2EBD-447C-A4C8-A03058436FF4",Vk=(0,$t.pM)({name:"Dialog",inheritAttrs:!1,props:{as:{type:[Object,String],default:"div"},static:{type:Boolean,default:!1},unmount:{type:Boolean,default:!0},open:{type:[Boolean,String],default:qk},initialFocus:{type:Object,default:null},id:{type:String,default:null},role:{type:String,default:"dialog"}},emits:{close:e=>!0},setup(e,{emit:t,attrs:n,slots:r,expose:a}){var o,i;let s=null!=(o=e.id)?o:`headlessui-dialog-${(0,ep.B)()}`,l=(0,jn.KR)(!1);(0,$t.sV)((()=>{l.value=!0}));let c=!1,u=(0,$t.EW)((()=>"dialog"===e.role||"alertdialog"===e.role?e.role:(c||(c=!0,console.warn(`Invalid role [${u}] passed to <Dialog />. Only \`dialog\` and and \`alertdialog\` are supported. Using \`dialog\` instead.`)),"dialog"))),d=(0,jn.KR)(0),p=(0,lp.O_)(),h=(0,$t.EW)((()=>e.open===qk&&null!==p?(p.value&lp.Uw.Open)===lp.Uw.Open:e.open)),f=(0,jn.KR)(null),m=(0,$t.EW)((()=>(0,Wd.T)(f)));if(a({el:f,$el:f}),e.open===qk&&null===p)throw new Error("You forgot to provide an `open` prop to the `Dialog`.");if("boolean"!=typeof h.value)throw new Error(`You provided an \`open\` prop to the \`Dialog\`, but the value is not a boolean. Received: ${h.value===qk?void 0:e.open}`);let g=(0,$t.EW)((()=>l.value&&h.value?0:1)),v=(0,$t.EW)((()=>0===g.value)),b=(0,$t.EW)((()=>d.value>1)),y=null!==(0,$t.WQ)(Wk,null),[w,O]=Zd(),{resolveContainers:k,mainTreeNodeRef:x,MainTreeNode:E}=ap({portals:w,defaultContainers:[(0,$t.EW)((()=>{var e;return null!=(e=M.panelRef.value)?e:f.value}))]}),S=(0,$t.EW)((()=>b.value?"parent":"leaf")),_=(0,$t.EW)((()=>null!==p&&(p.value&lp.Uw.Closing)===lp.Uw.Closing)),C=(0,$t.EW)((()=>!y&&!_.value&&v.value)),T=(0,$t.EW)((()=>{var e,t,n;return null!=(n=Array.from(null!=(t=null==(e=m.value)?void 0:e.querySelectorAll("body > *"))?t:[]).find((e=>"headlessui-portal-root"!==e.id&&e.contains((0,jd.t)(x))&&e instanceof HTMLElement)))?n:null}));Fk(T,C);let A=(0,$t.EW)((()=>!!b.value||v.value)),R=(0,$t.EW)((()=>{var e,t,n;return null!=(n=Array.from(null!=(t=null==(e=m.value)?void 0:e.querySelectorAll("[data-headlessui-portal]"))?t:[]).find((e=>e.contains((0,jd.t)(x))&&e instanceof HTMLElement)))?n:null}));Fk(R,A),function({type:e,enabled:t,element:n,onUpdate:r}){let a=(0,$t.WQ)(Qk,(()=>{}));function o(...e){null==r||r(...e),a(...e)}(0,$t.sV)((()=>{(0,$t.wB)(t,((t,r)=>{t?o(0,e,n):!0===r&&o(1,e,n)}),{immediate:!0,flush:"sync"})})),(0,$t.hi)((()=>{t.value&&o(1,e,n)})),(0,$t.Gt)(Qk,o)}({type:"Dialog",enabled:(0,$t.EW)((()=>0===g.value)),element:f,onUpdate:(e,t)=>{if("Dialog"===t)return(0,dp.Y)(e,{[Uk.Add]:()=>d.value+=1,[Uk.Remove]:()=>d.value-=1})}});let P=(0,zk.r)({name:"DialogDescription",slot:(0,$t.EW)((()=>({open:h.value})))}),D=(0,jn.KR)(null),M={titleId:D,panelRef:(0,jn.KR)(null),dialogState:g,setTitleId(e){D.value!==e&&(D.value=e)},close(){t("close",!1)}};(0,$t.Gt)(Wk,M);let $=(0,$t.EW)((()=>!(!v.value||b.value)));(0,tp.j)(k,((e,t)=>{e.preventDefault(),M.close(),(0,$t.dY)((()=>null==t?void 0:t.focus()))}),$);let N=(0,$t.EW)((()=>!(b.value||0!==g.value)));Jd(null==(i=m.value)?void 0:i.defaultView,"keydown",(e=>{N.value&&(e.defaultPrevented||e.key===cp.D.Escape&&(e.preventDefault(),e.stopPropagation(),M.close()))}));let I=(0,$t.EW)((()=>!(_.value||0!==g.value||y)));return function(e,t,n){let r=function(e){let t=(0,jn.IJ)(e.getSnapshot());return(0,$t.hi)(e.subscribe((()=>{t.value=e.getSnapshot()}))),t}(Ik),a=(0,$t.EW)((()=>{let t=e.value?r.value.get(e.value):void 0;return!!t&&t.count>0}));(0,$t.wB)([e,t],(([e,t],[r],a)=>{if(!e||!t)return;Ik.dispatch("PUSH",e,n);let o=!1;a((()=>{o||(Ik.dispatch("POP",null!=r?r:e,n),o=!0)}))}),{immediate:!0})}(m,I,(e=>{var t;return{containers:[...null!=(t=e.containers)?t:[],k]}})),(0,$t.nT)((e=>{if(0!==g.value)return;let t=(0,jd.t)(f);if(!t)return;let n=new ResizeObserver((e=>{for(let t of e){let e=t.target.getBoundingClientRect();0===e.x&&0===e.y&&0===e.width&&0===e.height&&M.close()}}));n.observe(t),e((()=>n.disconnect()))})),()=>{let{open:t,initialFocus:a,...o}=e,i={...n,ref:f,id:s,role:u.value,"aria-modal":0===g.value||void 0,"aria-labelledby":D.value,"aria-describedby":P.value},l={open:0===g.value};return(0,$t.h)(zd,{force:!0},(()=>[(0,$t.h)(Vd,(()=>(0,$t.h)(Kd,{target:f.value},(()=>(0,$t.h)(zd,{force:!1},(()=>(0,$t.h)(Rk,{initialFocus:a,containers:k,features:v.value?(0,dp.Y)(S.value,{parent:Rk.features.RestoreFocus,leaf:Rk.features.All&~Rk.features.FocusLock}):Rk.features.None},(()=>(0,$t.h)(O,{},(()=>(0,Qd.XX)({ourProps:i,theirProps:{...o,...n},slot:l,attrs:n,slots:r,visible:0===g.value,features:Qd.O5.RenderStrategy|Qd.O5.Static,name:"Dialog"}))))))))))),(0,$t.h)(E)]))}}}),Hk=((0,$t.pM)({name:"DialogOverlay",props:{as:{type:[Object,String],default:"div"},id:{type:String,default:null}},setup(e,{attrs:t,slots:n}){var r;let a=null!=(r=e.id)?r:`headlessui-dialog-overlay-${(0,ep.B)()}`,o=Xk("DialogOverlay");function i(e){e.target===e.currentTarget&&(e.preventDefault(),e.stopPropagation(),o.close())}return()=>{let{...r}=e;return(0,Qd.XX)({ourProps:{id:a,"aria-hidden":!0,onClick:i},theirProps:r,slot:{open:0===o.dialogState.value},attrs:t,slots:n,name:"DialogOverlay"})}}}),(0,$t.pM)({name:"DialogBackdrop",props:{as:{type:[Object,String],default:"div"},id:{type:String,default:null}},inheritAttrs:!1,setup(e,{attrs:t,slots:n,expose:r}){var a;let o=null!=(a=e.id)?a:`headlessui-dialog-backdrop-${(0,ep.B)()}`,i=Xk("DialogBackdrop"),s=(0,jn.KR)(null);return r({el:s,$el:s}),(0,$t.sV)((()=>{if(null===i.panelRef.value)throw new Error("A <DialogBackdrop /> component is being used, but a <DialogPanel /> component is missing.")})),()=>{let{...r}=e,a={id:o,ref:s,"aria-hidden":!0};return(0,$t.h)(zd,{force:!0},(()=>(0,$t.h)(Vd,(()=>(0,Qd.XX)({ourProps:a,theirProps:{...t,...r},slot:{open:0===i.dialogState.value},attrs:t,slots:n,name:"DialogBackdrop"})))))}}}),(0,$t.pM)({name:"DialogPanel",props:{as:{type:[Object,String],default:"div"},id:{type:String,default:null}},setup(e,{attrs:t,slots:n,expose:r}){var a;let o=null!=(a=e.id)?a:`headlessui-dialog-panel-${(0,ep.B)()}`,i=Xk("DialogPanel");function s(e){e.stopPropagation()}return r({el:i.panelRef,$el:i.panelRef}),()=>{let{...r}=e,a={id:o,ref:i.panelRef,onClick:s};return(0,Qd.XX)({ourProps:a,theirProps:r,slot:{open:0===i.dialogState.value},attrs:t,slots:n,name:"DialogPanel"})}}})),Zk=(0,$t.pM)({name:"DialogTitle",props:{as:{type:[Object,String],default:"h2"},id:{type:String,default:null}},setup(e,{attrs:t,slots:n}){var r;let a=null!=(r=e.id)?r:`headlessui-dialog-title-${(0,ep.B)()}`,o=Xk("DialogTitle");return(0,$t.sV)((()=>{o.setTitleId(a),(0,$t.hi)((()=>o.setTitleId(null)))})),()=>{let{...r}=e;return(0,Qd.XX)({ourProps:{id:a},theirProps:r,slot:{open:0===o.dialogState.value},attrs:t,slots:n,name:"DialogTitle"})}}}),Gk=zk.V;const Kk={key:0,class:"close-button z-10 fixed right-2 top-2"};function Yk(){return(0,jn.Kh)({open:!1,show(){this.open=!0},hide(){this.open=!1}})}const Jk=(0,$t.pM)({__name:"ScalarModal",props:{state:{},title:{},bodyClass:{},maxWidth:{},size:{default:"md"},variant:{}},setup(e){const t=Fn({base:["scalar-modal","col relative mx-auto mb-0 rounded-lg bg-b-1 p-0 text-left leading-snug text-c-1 opacity-0 md:w-[calc(100vw-16px)] lg:w-[calc(100vw-32px)]"].join(" "),variants:{size:{xxs:"mt-20 max-w-screen-xxs",xs:"mt-20 max-w-screen-xs",sm:"mt-20 max-w-screen-sm",md:"mt-20 max-w-screen-md",lg:"m-auto max-w-screen-lg",xl:"m-auto max-w-screen-xl",full:"full-size-styles mt-0 lg:w-full"},variant:{form:"scalar-modal-form",history:"scalar-modal-history bg-b-1",search:"scalar-modal-search",error:"scalar-modal-error"}}}),n=Fn({base:["scalar-modal-body","relative max-h-[calc(100dvh-240px)] rounded-lg rounded-t-none border-t-1/2 bg-b-1 p-3"].join(" "),variants:{variant:{form:"overflow-visible",history:"pt-3",search:"col !m-0 max-h-[440px] overflow-hidden p-0",error:"overflow-y-scroll"},size:{xxs:"max-h-[calc(100dvh-240px)]",xs:"max-h-[calc(100dvh-240px)]",sm:"max-h-[calc(100dvh-240px)]",md:"max-h-[calc(100dvh-240px)]",lg:"max-h-[calc(100dvh-180px)]",xl:"m-0 max-h-[calc(100dvh-120px)] p-0",full:"max-h-dvh rounded-none"}}});return(e,r)=>((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Vk),{open:e.state.open,onClose:r[1]||(r[1]=t=>e.state.hide())},{default:(0,$t.k6)((()=>[(0,$t.Lk)("div",{class:(0,Wn.C4)((0,jn.R1)(Qn)("full"===e.size?"scalar-modal-layout-full":"scalar-modal-layout","fixed left-0 top-0 flex items-start justify-center","z-[1001] h-[100dvh] w-[100dvw]","bg-backdrop opacity-0 dark:bg-backdropdark","full"===e.size&&"flex"))},[(0,$t.bF)((0,jn.R1)(Hk),{class:(0,Wn.C4)((0,jn.R1)(t)({size:e.size,variant:e.variant})),style:(0,Wn.Tr)({maxWidth:e.maxWidth})},{default:(0,$t.k6)((()=>[e.title?((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Zk),{key:0,class:(0,Wn.C4)(["scalar-modal-header m-0 -mb-1 rounded-lg p-3 text-left text-sm font-medium text-c-1",{"pb-0 pt-6":"history"===e.variant}])},{default:(0,$t.k6)((()=>[(0,$t.eW)((0,Wn.v_)(e.title),1)])),_:1},8,["class"])):(0,$t.Q3)("",!0),"full"===e.size?((0,$t.uX)(),(0,$t.CE)("div",{key:1,class:(0,Wn.C4)(e.bodyClass)},[(0,$t.RG)(e.$slots,"default",{},void 0,!0)],2)):((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Gk),{key:2,class:(0,Wn.C4)((0,jn.R1)(Qn)((0,jn.R1)(n)({size:e.size,variant:e.variant}),e.bodyClass))},{default:(0,$t.k6)((()=>[(0,$t.RG)(e.$slots,"default",{},void 0,!0)])),_:3},8,["class"]))])),_:3},8,["class","style"]),"full"===e.size?((0,$t.uX)(),(0,$t.CE)("div",Kk,[(0,$t.bF)((0,jn.R1)(zO),{class:"hover:bg-b-3 focus:outline-none",icon:"Close",label:"Clear Search",onClose:r[0]||(r[0]=t=>e.state.hide())})])):(0,$t.Q3)("",!0)],2)])),_:3},8,["open"]))}}),ex=Gn(Jk,[["__scopeId","data-v-0f8e0211"]]),tx=["aria-label","value"],nx=(0,$t.pM)({inheritAttrs:!1,__name:"ScalarSearchInput",props:{loading:{},modelValue:{},sidebar:{type:Boolean},label:{}},emits:["update:modelValue"],setup(e,{expose:t,emit:n}){const r=n,a=(0,jn.KR)(null);function o(e){r("update:modelValue",e.target.value)}const i=(0,$t.EW)((()=>{const{class:e,...t}=(0,$t.OA)();return{class:e||"",rest:t}})),s=Fn({base:"search-background flex items-center rounded border text-sm font-medium has-[:focus-visible]:bg-b-1 has-[:focus-visible]:outline",variants:{sidebar:{true:"h-8 gap-2 px-1.5",false:"h-10 p-3"}}});return t({focus:()=>{var e;null==(e=a.value)||e.focus()},blur:()=>{var e;null==(e=a.value)||e.blur()}}),(e,t)=>((0,$t.uX)(),(0,$t.CE)("label",{class:(0,Wn.C4)((0,jn.R1)(Qn)((0,jn.R1)(s)({sidebar:e.sidebar}),i.value.class))},[e.sidebar?((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(wd),{key:0,class:"text-c-2",icon:"Search",size:"xs",thickness:"2.5"})):(0,$t.Q3)("",!0),(0,$t.Lk)("input",(0,$t.v6)({ref_key:"inputRef",ref:a},i.value.rest,{"aria-label":e.label??"Enter search query",autocapitalize:"off",autocomplete:"off",autocorrect:"off",class:"flex-1 appearance-none rounded border-none bg-transparent outline-none",placeholder:"Search...",spellcheck:"false",type:"search",value:e.modelValue,onInput:o}),null,16,tx),e.loading&&e.loading.isLoading?((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(Kn),{key:1,class:"mr-3 self-center",loadingState:e.loading,size:"md"},null,8,["loadingState"])):e.modelValue?((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(zO),{key:2,class:(0,Wn.C4)((0,jn.R1)(Qn)("p-0",e.sidebar?"h-4 w-4":"h-5 w-5")),icon:"Close",label:"Clear Search",thickness:e.sidebar?"1.75":"1.5",onClick:t[0]||(t[0]=e=>r("update:modelValue",""))},null,8,["class","thickness"])):(0,$t.Q3)("",!0)],2))}}),rx=Gn(nx,[["__scopeId","data-v-2f2316c6"]]);var ax=n(3073);const ox={class:"scalar-input-container relative"},ix=["for"],sx={class:"icon-slot cursor-pointer text-c-ghost hover:text-c-1 !empty:flex !empty:w-7 !empty:items-center !empty:pr-3"},lx=(0,$t.pM)({__name:"ScalarTextField",props:{modelValue:{},placeholder:{},label:{},labelShadowColor:{default:"var(--scalar-background-1)"},error:{type:Boolean},isMultiline:{type:Boolean},helperText:{},emitOnBlur:{type:Boolean,default:!0},handleFieldSubmit:{},handleFieldChange:{},disableTrim:{type:Boolean,default:!1}},emits:["submit","update:modelValue"],setup(e,{emit:t}){const n=e,r=t,a=Fn({base:"scalar-input-wrapper relative flex items-center rounded border border-solid border-border",variants:{focus:{true:"scalar-input-wrapper-focus border-c-3 has-actv-btn:border has-actv-btn:border-border"},error:{true:"scalar-input-wrapper-error border-red"}}}),o=(0,$t.OA)(),i=o.id||`id-${Pd()}`,s=(0,jn.KR)(),l=(0,jn.KR)(!1);let c;if(n.isMultiline){const{triggerResize:e}=(0,ax.AHd)({element:s,input:n.modelValue});c=e}function u(e){const t=e.target.value;n.isMultiline&&c(),n.handleFieldChange?n.handleFieldChange(t):r("update:modelValue",t)}function d(e){const t=e.target,a=n.disableTrim||n.isMultiline?t.value:t.value.trim();n.handleFieldSubmit?n.handleFieldSubmit(a):r("submit",a)}function p(e){l.value=!1,n.emitOnBlur&&n.modelValue&&d(e)}return(0,$t.sV)((()=>{var e;Object.prototype.hasOwnProperty.call(o,"autofocus")&&(null==(e=s.value)||e.focus())})),(e,t)=>((0,$t.uX)(),(0,$t.CE)("div",ox,[(0,$t.Lk)("div",{class:(0,Wn.C4)((0,jn.R1)(a)({error:e.error,focus:l.value}))},[((0,$t.uX)(),(0,$t.Wv)((0,$t.$y)(e.isMultiline?"textarea":"input"),(0,$t.v6)({id:(0,jn.R1)(i)},e.$attrs,{ref_key:"input",ref:s,class:["scalar-input",(0,jn.R1)(Qn)({"min-h-[77px]":e.isMultiline},"z-10 w-full resize-none appearance-none border-0 bg-transparent p-3 text-sm text-c-1","outline-none transition-opacity")],placeholder:e.placeholder,value:e.modelValue,onBlur:p,onFocus:t[0]||(t[0]=e=>l.value=!0),onInput:u,onKeydown:(0,Rr.jR)(d,["enter"])}),null,16,["id","class","placeholder","value"])),e.label?((0,$t.uX)(),(0,$t.CE)("label",{key:0,class:(0,Wn.C4)((0,jn.R1)(Qn)("scalar-input-label pointer-events-none absolute left-0 top-0 mx-2 my-3 px-1 text-sm","z-10 origin-top-left rounded text-c-3 shadow-current transition-transform")),for:(0,jn.R1)(i),style:(0,Wn.Tr)({"box-shadow":`0 0 4px 4px ${e.labelShadowColor}`,"background-color":e.labelShadowColor})},(0,Wn.v_)(e.label),15,ix)):(0,$t.Q3)("",!0),(0,$t.Lk)("div",sx,[(0,$t.RG)(e.$slots,"default",{},void 0,!0)])],2),e.helperText?((0,$t.uX)(),(0,$t.CE)("span",{key:0,class:(0,Wn.C4)(["helper-text mt-1.5 flex items-center gap-1 text-sm",e.error?"font-medium text-red":"text-c-2"])},[e.error?((0,$t.uX)(),(0,$t.Wv)((0,jn.R1)(wd),{key:0,icon:"Error",size:"sm"})):(0,$t.Q3)("",!0),(0,$t.eW)(" "+(0,Wn.v_)(e.helperText),1)],2)):(0,$t.Q3)("",!0)]))}}),cx=Gn(lx,[["__scopeId","data-v-4de7a24e"]]),ux=()=>{const e=document.getElementById("headlessui-portal-root");if(e)e.classList.add("scalar-app"),e.classList.add("scalar-client");else{const e=new MutationObserver((t=>{const n=t.find((e=>Array.from(e.addedNodes).find((e=>"headlessui-portal-root"===e.id))));if(n){const t=n.addedNodes[0];t.classList.add("scalar-app"),t.classList.add("scalar-client"),e.disconnect()}}));e.observe(document.body,{childList:!0})}}},1284:(e,t,n)=>{"use strict";n.d(t,{As:()=>g,qg:()=>m});const{parse:r,stringify:a}=JSON,{keys:o}=Object,i=String,s="string",l={},c="object",u=(e,t)=>t,d=e=>e instanceof i?i(e):e,p=(e,t)=>typeof t===s?new i(t):t,h=(e,t,n,r)=>{const a=[];for(let s=o(n),{length:u}=s,d=0;d<u;d++){const o=s[d],u=n[o];if(u instanceof i){const i=e[u];typeof i!==c||t.has(i)?n[o]=r.call(n,o,i):(t.add(i),n[o]=l,a.push({k:o,a:[e,t,i,r]}))}else n[o]!==l&&(n[o]=r.call(n,o,u))}for(let{length:e}=a,t=0;t<e;t++){const{k:e,a:o}=a[t];n[e]=r.call(n,e,h.apply(null,o))}return n},f=(e,t,n)=>{const r=i(t.push(n)-1);return e.set(n,r),r},m=(e,t)=>{const n=r(e,p).map(d),a=n[0],o=t||u,i=typeof a===c&&a?h(n,new Set,a,o):a;return o.call({"":i},"",i)},g=(e,t,n)=>{const r=t&&typeof t===c?(e,n)=>""===e||-1<t.indexOf(e)?n:void 0:t||u,o=new Map,i=[],l=[];let d=+f(o,i,r.call({"":e},"",e)),p=!d;for(;d<i.length;)p=!0,l[d]=a(i[d++],h,n);return"["+l.join(",")+"]";function h(e,t){if(p)return p=!p,t;const n=r.call(this,e,t);switch(typeof n){case c:if(null===n)return n;case s:return o.get(n)||f(o,i,n)}return n}}},1347:(e,t,n)=>{"use strict";n.d(t,{CV:()=>r.CV,oE:()=>i.oE,I4:()=>l.I4,vA:()=>i.vA,$W:()=>i.$W,TF:()=>o.TF,NV:()=>l.NV,zT:()=>a.zT,c3:()=>s.c3});var r=n(7984),a=n(1608),o=n(7297),i=n(4703),s=n(8065),l=(n(4811),n(5785));n(2555)},1417:(e,t,n)=>{"use strict";n.d(t,{BD:()=>u,Bm:()=>m,CU:()=>k,Fh:()=>g,MZ:()=>f,Me:()=>d,iq:()=>h,p9:()=>O,pW:()=>b,wl:()=>w});var r=n(641),a=n(9388),o=n(562);let i=["[contentEditable=true]","[tabindex]","a[href]","area[href]","button:not([disabled])","iframe","input:not([disabled])","select:not([disabled])","textarea:not([disabled])"].map((e=>`${e}:not([tabindex='-1'])`)).join(",");var s,l,c,u=((c=u||{})[c.First=1]="First",c[c.Previous=2]="Previous",c[c.Next=4]="Next",c[c.Last=8]="Last",c[c.WrapAround=16]="WrapAround",c[c.NoScroll=32]="NoScroll",c),d=((l=d||{})[l.Error=0]="Error",l[l.Overflow=1]="Overflow",l[l.Success=2]="Success",l[l.Underflow=3]="Underflow",l),p=((s=p||{})[s.Previous=-1]="Previous",s[s.Next=1]="Next",s);function h(e=document.body){return null==e?[]:Array.from(e.querySelectorAll(i)).sort(((e,t)=>Math.sign((e.tabIndex||Number.MAX_SAFE_INTEGER)-(t.tabIndex||Number.MAX_SAFE_INTEGER))))}var f=(e=>(e[e.Strict=0]="Strict",e[e.Loose=1]="Loose",e))(f||{});function m(e,t=0){var n;return e!==(null==(n=(0,o.T)(e))?void 0:n.body)&&(0,a.Y)(t,{0:()=>e.matches(i),1(){let t=e;for(;null!==t;){if(t.matches(i))return!0;t=t.parentElement}return!1}})}function g(e){let t=(0,o.T)(e);(0,r.dY)((()=>{t&&!m(t.activeElement,0)&&b(e)}))}var v=(e=>(e[e.Keyboard=0]="Keyboard",e[e.Mouse=1]="Mouse",e))(v||{});function b(e){null==e||e.focus({preventScroll:!0})}"undefined"!=typeof window&&"undefined"!=typeof document&&(document.addEventListener("keydown",(e=>{e.metaKey||e.altKey||e.ctrlKey||(document.documentElement.dataset.headlessuiFocusVisible="")}),!0),document.addEventListener("click",(e=>{1===e.detail?delete document.documentElement.dataset.headlessuiFocusVisible:0===e.detail&&(document.documentElement.dataset.headlessuiFocusVisible="")}),!0));let y=["textarea","input"].join(",");function w(e,t=e=>e){return e.slice().sort(((e,n)=>{let r=t(e),a=t(n);if(null===r||null===a)return 0;let o=r.compareDocumentPosition(a);return o&Node.DOCUMENT_POSITION_FOLLOWING?-1:o&Node.DOCUMENT_POSITION_PRECEDING?1:0}))}function O(e,t){return k(h(),t,{relativeTo:e})}function k(e,t,{sorted:n=!0,relativeTo:r=null,skipElements:a=[]}={}){var o;let i=null!=(o=Array.isArray(e)?e.length>0?e[0].ownerDocument:document:null==e?void 0:e.ownerDocument)?o:document,s=Array.isArray(e)?n?w(e):e:h(e);a.length>0&&s.length>1&&(s=s.filter((e=>!a.includes(e)))),r=null!=r?r:i.activeElement;let l,c=(()=>{if(5&t)return 1;if(10&t)return-1;throw new Error("Missing Focus.First, Focus.Previous, Focus.Next or Focus.Last")})(),u=(()=>{if(1&t)return 0;if(2&t)return Math.max(0,s.indexOf(r))-1;if(4&t)return Math.max(0,s.indexOf(r))+1;if(8&t)return s.length-1;throw new Error("Missing Focus.First, Focus.Previous, Focus.Next or Focus.Last")})(),d=32&t?{preventScroll:!0}:{},p=0,f=s.length;do{if(p>=f||p+f<=0)return 0;let e=u+p;if(16&t)e=(e+f)%f;else{if(e<0)return 3;if(e>=f)return 1}l=s[e],null==l||l.focus(d),p+=c}while(l!==i.activeElement);return 6&t&&function(e){var t,n;return null!=(n=null==(t=null==e?void 0:e.matches)?void 0:t.call(e,y))&&n}(l)&&l.select(),2}},1465:(e,t,n)=>{"use strict";n.d(t,{A:()=>g});var r=n(641),a=n(33),o=n(953),i=n(3751),s=n(8547),l=n(1266),c=n(3806);const u={class:"absolute flex right-1 opacity-0 group-hover:opacity-100"},d=(0,r.pM)({__name:"SidebarListElementActions",props:{variable:{},warningMessage:{},isCopyable:{type:Boolean},isDeletable:{type:Boolean},isRenameable:{type:Boolean}},emits:["delete","rename"],setup(e,{emit:t}){const n=t,a=(0,o.KR)({action:"None",name:""}),d=(0,l.hS)(),{copyToClipboard:p}=(0,c.i)();function h(){d.hide(),a.value={action:"None",name:""}}return(t,c)=>((0,r.uX)(),(0,r.CE)(r.FK,null,[(0,r.Lk)("div",u,[t.isCopyable?((0,r.uX)(),(0,r.CE)("button",{key:0,class:"text-c-3 hover:bg-b-3 hover:text-c-1 rounded p-[5px]",type:"button",onClick:c[0]||(c[0]=e=>(0,o.R1)(p)(t.variable.name))},[(0,r.bF)((0,o.R1)(l.kv),{class:"h-3 w-3",icon:"Clipboard"})])):(0,r.Q3)("",!0),t.isRenameable?((0,r.uX)(),(0,r.CE)("button",{key:1,class:"text-c-3 hover:bg-b-3 hover:text-c-1 rounded p-[5px]",type:"button",onClick:c[1]||(c[1]=e=>n("rename",t.variable.uid))},[(0,r.bF)((0,o.R1)(l.kv),{class:"h-3 w-3",icon:"Edit"})])):(0,r.Q3)("",!0),!t.variable.isDefault&&t.isDeletable?((0,r.uX)(),(0,r.CE)("button",{key:2,class:"text-c-3 hover:bg-b-3 hover:text-c-1 rounded p-1",type:"button",onClick:c[2]||(c[2]=(0,i.D$)((t=>(a.value={action:"Delete",name:e.variable.name},void d.show())),["prevent"]))},[(0,r.bF)((0,o.R1)(l.kv),{class:"h-3.5 w-3.5",icon:"Close"})])):(0,r.Q3)("",!0)]),(0,r.bF)((0,o.R1)(l.uT),{size:"sm",state:(0,o.R1)(d),title:`${a.value.action} ${a.value.name}`},{default:(0,r.k6)((()=>["Delete"===a.value.action?((0,r.uX)(),(0,r.Wv)(s.A,{key:0,variableName:a.value.name,warningMessage:t.warningMessage,onClose:h,onDelete:c[3]||(c[3]=e=>function(e){n("delete",e),h()}(t.variable.uid))},null,8,["variableName","warningMessage"])):(0,r.Q3)("",!0)])),_:1},8,["state","title"])],64))}});var p=n(9135),h=n(9351);const f={class:"empty-variable-name text-sm line-clamp-1 break-all group-hover:pr-5"},m=(0,r.pM)({__name:"SidebarListElement",props:{variable:{},collectionId:{},warningMessage:{},isDeletable:{type:Boolean},isCopyable:{type:Boolean},isRenameable:{type:Boolean},type:{}},emits:["delete","colorModal","rename"],setup(e,{emit:t}){const n=e,s=t,c=(0,h.rd)(),{activeWorkspace:u}=(0,p.JC)(),m=e=>{s("delete",e)},g=e=>{s("rename",e)};return(e,t)=>{var p,h;const v=(0,r.g2)("router-link");return(0,r.uX)(),(0,r.CE)("li",null,[(0,r.bF)(v,{class:(0,a.C4)(["h-8 text-c-2 hover:bg-b-2 group relative block flex items-center gap-1.5 rounded py-1 pr-1.5 font-medium no-underline",[e.variable.color?"pl-1":"pl-1.5"]]),exactActiveClass:"active-link",role:"button",to:e.collectionId?`/workspace/${null==(p=(0,o.R1)(u))?void 0:p.uid}/${e.type}/${e.collectionId}/${e.variable.uid}`:`/workspace/${null==(h=(0,o.R1)(u))?void 0:h.uid}/${e.type}/${e.variable.uid}`,onClick:t[1]||(t[1]=(0,i.D$)((t=>((e,t,r)=>{var a;const o=null==(a=u.value)?void 0:a.uid,i=n.type,s=r?`/workspace/${o}/${i}/${r||void 0}/${t}`:`/workspace/${o}/${i}/${t}`;e.metaKey?window.open(s,"_blank"):c.push({path:s})})(t,e.variable.uid,e.collectionId)),["prevent"]))},{default:(0,r.k6)((()=>[e.variable.color?((0,r.uX)(),(0,r.CE)("button",{key:0,class:"hover:bg-b-3 rounded p-1.5",type:"button",onClick:t[0]||(t[0]=t=>(e=>{s("colorModal",e)})(e.variable.uid))},[(0,r.Lk)("div",{class:"h-2.5 w-2.5 rounded-xl",style:(0,a.Tr)({backgroundColor:e.variable.color})},null,4)])):(0,r.Q3)("",!0),e.variable.icon?((0,r.uX)(),(0,r.Wv)((0,o.R1)(l.kv),{key:1,class:"text-sidebar-c-2 size-3.5 stroke-[2.25]",icon:e.variable.icon},null,8,["icon"])):(0,r.Q3)("",!0),(0,r.Lk)("span",f,(0,a.v_)(e.variable.name),1),(0,r.bF)(d,{isCopyable:!!e.isCopyable,isDeletable:!!e.isDeletable,isRenameable:!!e.isRenameable,variable:{...e.variable,isDefault:e.variable.isDefault??!1},warningMessage:e.warningMessage,onDelete:m,onRename:g},null,8,["isCopyable","isDeletable","isRenameable","variable","warningMessage"])])),_:1},8,["class","to"])])}}}),g=(0,n(3526).A)(m,[["__scopeId","data-v-a6c1ff1c"]])},1504:(e,t,n)=>{"use strict";n.d(t,{A:()=>g});var r=n(641),a=n(953),o=n(33),i=n(3654),s=n(1266),l=n(4282),c=n(1760);const u={key:0,class:"text-c-1 flex items-center pl-3 pr-0"},d={class:"row-1 overflow-x-auto"},p=["readOnly","type","value"],h={key:2,class:"scalar-input-required absolute centered-y right-2 pt-px text-xxs text-c-3 bg-b-1 shadow-[-8px_0_4px_var(--scalar-background-1)] opacity-100 duration-150 transition-opacity peer-has-[:focus-visible]:opacity-0"},f={key:1,class:"absolute centered-y right-7 text-orange text-xs"},m=(0,r.pM)({inheritAttrs:!1,__name:"DataTableInput",props:{id:{},type:{},containerClass:{},required:{type:Boolean,default:!1},modelValue:{},canAddCustomEnumValue:{type:Boolean,default:!0},readOnly:{type:Boolean,default:!1},enum:{},min:{},max:{},environment:{},envVariables:{},workspace:{}},emits:["update:modelValue","inputFocus","inputBlur","selectVariable"],setup(e,{emit:t}){const n=e,m=t,g=(0,a.KR)(!0),v=(0,a.KR)(!1),b=()=>{v.value||m("inputBlur")},y=(0,r.EW)((()=>"password"===n.type?g.value?"password":"text":n.type??"text"));return(e,t)=>((0,r.uX)(),(0,r.Wv)(l.A,{class:(0,o.C4)(["relative row",e.containerClass])},{default:(0,r.k6)((()=>[e.$slots.default?((0,r.uX)(),(0,r.CE)("div",u,[(0,r.RG)(e.$slots,"default",{},void 0,!0),t[5]||(t[5]=(0,r.eW)(": "))])):(0,r.Q3)("",!0),(0,r.Lk)("div",d,[n.enum&&n.enum.length?((0,r.uX)(),(0,r.Wv)(c.A,{key:0,canAddCustomValue:n.canAddCustomEnumValue,modelValue:n.modelValue,value:n.enum,"onUpdate:modelValue":t[0]||(t[0]=e=>m("update:modelValue",e))},null,8,["canAddCustomValue","modelValue","value"])):((0,r.uX)(),(0,r.CE)(r.FK,{key:1},[g.value&&"password"===e.type?((0,r.uX)(),(0,r.CE)("input",(0,r.v6)({key:0},e.id?{...e.$attrs,id:e.id}:e.$attrs,{autocomplete:"off",class:"border-none text-c-1 disabled:text-c-2 min-w-0 w-full peer px-2 py-1.25 -outline-offset-2","data-1p-ignore":"",readOnly:e.readOnly,spellcheck:"false",type:y.value,value:e.modelValue,onInput:t[1]||(t[1]=e=>m("update:modelValue",e.target.value??""))}),null,16,p)):((0,r.uX)(),(0,r.Wv)(i.A,(0,r.v6)({key:1},e.$attrs,{id:e.id,class:"border-none text-c-1 disabled:text-c-2 min-w-0 w-full peer",disableCloseBrackets:"",disableTabIndent:"",envVariables:e.envVariables,environment:e.environment,max:e.max,min:e.min,modelValue:e.modelValue??"",readOnly:e.readOnly,required:!!e.required,spellcheck:"false",type:y.value,workspace:e.workspace,onBlur:b,onFocus:t[2]||(t[2]=e=>m("inputFocus")),"onUpdate:modelValue":t[3]||(t[3]=e=>m("update:modelValue",e))}),null,16,["id","envVariables","environment","max","min","modelValue","readOnly","required","type","workspace"])),e.required?((0,r.uX)(),(0,r.CE)("div",h," Required ")):(0,r.Q3)("",!0)],64))]),e.$slots.warning?((0,r.uX)(),(0,r.CE)("div",f,[(0,r.RG)(e.$slots,"warning",{},void 0,!0)])):(0,r.Q3)("",!0),(0,r.RG)(e.$slots,"icon",{},void 0,!0),"password"===e.type?((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.EX),{key:2,class:"-ml-.5 mr-0.75 h-6 w-6 self-center p-1.5",icon:g.value?"Show":"Hide",label:g.value?"Show Password":"Hide Password",onClick:t[4]||(t[4]=e=>g.value=!g.value)},null,8,["icon","label"])):(0,r.Q3)("",!0)])),_:3},8,["class"]))}}),g=(0,n(3526).A)(m,[["__scopeId","data-v-cb0ff5b1"]])},1507:(e,t,n)=>{"use strict";n.d(t,{Y:()=>i});var r=n(4279),a=n(1793),o=n(8007);function i(e,t={}){if((0,a.W)(e))return e;const n=(0,o.S)(e);return[{isEntrypoint:!0,specification:n,filename:null,dir:"./",references:(0,r.o)(n),...t}]}},1527:(e,t,n)=>{"use strict";n.d(t,{A:()=>w});var r=n(641),a=n(953),o=n(33),i=n(1266),s=n(4048),l=n(25),c=n(6584),u=n(8309),d=n(3833),p=n(9794);const h=Object.entries({apiKeyCookie:{label:"API Key in Cookies",payload:{type:"apiKey",in:"cookie",nameKey:"apiKeyCookie"}},apiKeyHeader:{label:"API Key in Headers",payload:{type:"apiKey",in:"header",nameKey:"apiKeyHeader"}},apiKeyQuery:{label:"API Key in Query Params",payload:{type:"apiKey",in:"query",nameKey:"apiKeyQuery"}},httpBasic:{label:"HTTP Basic",payload:{type:"http",scheme:"basic",nameKey:"httpBasic"}},httpBearer:{label:"HTTP Bearer",payload:{type:"http",scheme:"bearer",nameKey:"httpBearer"}},oauth2Implicit:{label:"Oauth2 Implicit Flow",payload:{type:"oauth2",nameKey:"oauth2Implicit",flows:{implicit:{type:"implicit"}}}},oauth2Password:{label:"Oauth2 Password Flow",payload:{type:"oauth2",nameKey:"oauth2Password",flows:{password:{type:"password"}}}},oauth2ClientCredentials:{label:"Oauth2 Client Credentials",payload:{type:"oauth2",nameKey:"oauth2ClientCredentials",flows:{clientCredentials:{type:"clientCredentials"}}}},oauth2AuthorizationFlow:{label:"Oauth2 Authorization Code",payload:{type:"oauth2",nameKey:"oauth2AuthorizationFlow",flows:{authorizationCode:{type:"authorizationCode"}}}}}).map((([e,t])=>({id:e,isDeletable:!1,...t}))),f=e=>({id:e.uid,label:"openIdConnect"===e.type?`${e.nameKey} (coming soon)`:e.nameKey}),m=(e,t)=>f(e.reduce(((e,n,r)=>{const a=t[n];return a&&(e.nameKey+=`${r>0?" & ":""}${a.nameKey}`,e.uid=`${e.uid}${r>0?",":""}${a.uid}`),e}),{type:"complex",nameKey:"",uid:""})),g={class:"inline-flex items-center gap-1"},v={class:"-mx-1 flex flex-1"},b={class:"text-c-1"},y=(0,r.pM)({__name:"RequestAuth",props:{collection:{},environment:{},envVariables:{},layout:{},operation:{},selectedSecuritySchemeUids:{},server:{},title:{},workspace:{}},setup(e){const{layout:t}=(0,c.g)(),{securitySchemes:n,securitySchemeMutators:y,requestMutators:w,collectionMutators:O}=(0,u.hA)(),k=(0,a.KR)(null),x=(0,i.hS)(),E=(0,a.KR)(null),S=(0,r.EW)((()=>{const t=((e,t)=>{var n;return"[{}]"===JSON.stringify(null==e?void 0:e.security)&&null!=(n=null==t?void 0:t.security)&&n.length?null!=t&&t.security.find((e=>"{}"===JSON.stringify(e)))?t.security:[...t.security,{}]:(null==e?void 0:e.security)??(null==t?void 0:t.security)??[]})(e.operation,e.collection);return{filteredRequirements:t.filter((e=>Object.keys(e).length)),requirements:t}})),_=(0,r.EW)((()=>{const{filteredRequirements:e,requirements:t}=S.value;if(!t.length)return null;const n=e.length<t.length,r=n?"Optional":"Required";return{icon:n?"Unlock":"Lock",text:`${1===e.length?Object.keys(e[0]||{})[0]:""} ${r}`}})),C=(0,r.EW)((()=>e.selectedSecuritySchemeUids.map((e=>{if(Array.isArray(e))return m(e,n);const t=n[e??""];return t?f(t):void 0})).filter(s.O9)));function T(t){var n;const r=t.find((e=>e.payload)),a=t.filter((e=>!e.payload)).map((({id:e})=>{const t=e.split(",");return t.length>1?t:e}));if(null!=r&&r.payload){const t=y.add(r.payload,null==(n=e.collection)?void 0:n.uid);t&&a.push(t.uid)}A(a)}const A=n=>{var r;"modal"===t||"reference"===e.layout?O.edit(e.collection.uid,"selectedSecuritySchemeUids",n):null!=(r=e.operation)&&r.uid&&w.edit(e.operation.uid,"selectedSecuritySchemeUids",n)};function R({id:e,label:t}){E.value={id:e,label:t},x.show()}const P=(0,r.EW)((()=>{var r;return((e,t,n,r=!1)=>{{const a=t.reduce(((e,t)=>{const r=n[t];return r&&(e[r.nameKey]=r),e}),{}),o=e.flatMap((e=>{const t=Object.keys(e);if(t.length>1){const e=t.map((e=>{var t;return null==(t=a[e])?void 0:t.uid})).filter(s.O9);return m(e,n)}if(t[0]){const e=a[t[0]];if(e)return f(e)}return[]})),i=t.filter((e=>!o.some((t=>t.id===e)))).map((e=>{const t=n[e];return t?f(t):null})).filter(s.O9),l=[{label:"Required authentication",options:o},{label:"Available authentication",options:i}];return r?o.length?l:i:(l.push({label:"Add new authentication",options:h}),l)}})(S.value.filteredRequirements,(null==(r=e.collection)?void 0:r.securitySchemes)??[],n,"modal"===t||"reference"===e.layout)}));return(n,s)=>((0,r.uX)(),(0,r.Wv)(l.A,{class:"group/params",itemCount:C.value.length,layout:n.layout},{title:(0,r.k6)((()=>[(0,r.Lk)("div",g,[(0,r.Lk)("span",null,(0,o.v_)(n.title),1),_.value?((0,r.uX)(),(0,r.CE)("span",{key:0,class:(0,o.C4)(["text-c-3 text-xs leading-[normal]",{"text-c-1":"Required"===_.value.text}])},(0,o.v_)(_.value.text),3)):(0,r.Q3)("",!0)])])),actions:(0,r.k6)((()=>[(0,r.Lk)("div",v,[(0,r.bF)((0,a.R1)(i.BR),{class:"w-72 text-xs",isDeletable:"modal"!==(0,a.R1)(t)&&"reference"!==n.layout,modelValue:C.value,multiple:"",options:P.value,onDelete:R,"onUpdate:modelValue":T},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(i.RO),{ref_key:"comboboxButtonRef",ref:k,class:"py-0.75 hover:bg-b-3 text-c-1 hover:text-c-1 h-auto px-1.5 font-normal",fullWidth:"",variant:"ghost"},{default:(0,r.k6)((()=>{var e;return[(0,r.Lk)("div",b,(0,o.v_)(0===C.value.length?"Auth Type":1===C.value.length?null==(e=C.value[0])?void 0:e.label:"Multiple"),1),(0,r.bF)((0,a.R1)(i.kv),{class:"ml-1 shrink-0",icon:"ChevronDown",size:"sm"})]})),_:1},512)])),_:1},8,["isDeletable","modelValue","options"])])])),default:(0,r.k6)((()=>[(0,r.bF)(p.A,{collection:n.collection,envVariables:n.envVariables,environment:n.environment,layout:n.layout,selectedSchemeOptions:C.value,server:n.server,workspace:n.workspace},null,8,["collection","envVariables","environment","layout","selectedSchemeOptions","server","workspace"]),(0,r.bF)(d.A,{scheme:E.value,state:(0,a.R1)(x),onClose:s[0]||(s[0]=e=>(0,a.R1)(x).hide()),onDelete:s[1]||(s[1]=t=>{var n;return(t=>{var n;if(!t)return;const r=e.selectedSecuritySchemeUids.filter((e=>{const n=t.split(",");return n.length>1&&Array.isArray(e)&&n.length===e.length?e.every((e=>!n.includes(e))):e!==t}));A(r),null==(n=k.value)||n.$el.focus(),x.hide()})(null==(n=E.value)?void 0:n.id)})},null,8,["scheme","state"])])),_:1},8,["itemCount","layout"]))}}),w=(0,n(3526).A)(y,[["__scopeId","data-v-3dc33a43"]])},1559:(e,t,n)=>{"use strict";n.d(t,{_:()=>r});const r=(0,n(1259).v)({properties:{xmlBase:null,xmlLang:null,xmlSpace:null},space:"xml",transform:(e,t)=>"xml:"+t.slice(3).toLowerCase()})},1564:(e,t,n)=>{"use strict";n.d(t,{J:()=>i});var r=n(6815),a=n(1259),o=n(5018);const i=(0,a.v)({attributes:{accentHeight:"accent-height",alignmentBaseline:"alignment-baseline",arabicForm:"arabic-form",baselineShift:"baseline-shift",capHeight:"cap-height",className:"class",clipPath:"clip-path",clipRule:"clip-rule",colorInterpolation:"color-interpolation",colorInterpolationFilters:"color-interpolation-filters",colorProfile:"color-profile",colorRendering:"color-rendering",crossOrigin:"crossorigin",dataType:"datatype",dominantBaseline:"dominant-baseline",enableBackground:"enable-background",fillOpacity:"fill-opacity",fillRule:"fill-rule",floodColor:"flood-color",floodOpacity:"flood-opacity",fontFamily:"font-family",fontSize:"font-size",fontSizeAdjust:"font-size-adjust",fontStretch:"font-stretch",fontStyle:"font-style",fontVariant:"font-variant",fontWeight:"font-weight",glyphName:"glyph-name",glyphOrientationHorizontal:"glyph-orientation-horizontal",glyphOrientationVertical:"glyph-orientation-vertical",hrefLang:"hreflang",horizAdvX:"horiz-adv-x",horizOriginX:"horiz-origin-x",horizOriginY:"horiz-origin-y",imageRendering:"image-rendering",letterSpacing:"letter-spacing",lightingColor:"lighting-color",markerEnd:"marker-end",markerMid:"marker-mid",markerStart:"marker-start",navDown:"nav-down",navDownLeft:"nav-down-left",navDownRight:"nav-down-right",navLeft:"nav-left",navNext:"nav-next",navPrev:"nav-prev",navRight:"nav-right",navUp:"nav-up",navUpLeft:"nav-up-left",navUpRight:"nav-up-right",onAbort:"onabort",onActivate:"onactivate",onAfterPrint:"onafterprint",onBeforePrint:"onbeforeprint",onBegin:"onbegin",onCancel:"oncancel",onCanPlay:"oncanplay",onCanPlayThrough:"oncanplaythrough",onChange:"onchange",onClick:"onclick",onClose:"onclose",onCopy:"oncopy",onCueChange:"oncuechange",onCut:"oncut",onDblClick:"ondblclick",onDrag:"ondrag",onDragEnd:"ondragend",onDragEnter:"ondragenter",onDragExit:"ondragexit",onDragLeave:"ondragleave",onDragOver:"ondragover",onDragStart:"ondragstart",onDrop:"ondrop",onDurationChange:"ondurationchange",onEmptied:"onemptied",onEnd:"onend",onEnded:"onended",onError:"onerror",onFocus:"onfocus",onFocusIn:"onfocusin",onFocusOut:"onfocusout",onHashChange:"onhashchange",onInput:"oninput",onInvalid:"oninvalid",onKeyDown:"onkeydown",onKeyPress:"onkeypress",onKeyUp:"onkeyup",onLoad:"onload",onLoadedData:"onloadeddata",onLoadedMetadata:"onloadedmetadata",onLoadStart:"onloadstart",onMessage:"onmessage",onMouseDown:"onmousedown",onMouseEnter:"onmouseenter",onMouseLeave:"onmouseleave",onMouseMove:"onmousemove",onMouseOut:"onmouseout",onMouseOver:"onmouseover",onMouseUp:"onmouseup",onMouseWheel:"onmousewheel",onOffline:"onoffline",onOnline:"ononline",onPageHide:"onpagehide",onPageShow:"onpageshow",onPaste:"onpaste",onPause:"onpause",onPlay:"onplay",onPlaying:"onplaying",onPopState:"onpopstate",onProgress:"onprogress",onRateChange:"onratechange",onRepeat:"onrepeat",onReset:"onreset",onResize:"onresize",onScroll:"onscroll",onSeeked:"onseeked",onSeeking:"onseeking",onSelect:"onselect",onShow:"onshow",onStalled:"onstalled",onStorage:"onstorage",onSubmit:"onsubmit",onSuspend:"onsuspend",onTimeUpdate:"ontimeupdate",onToggle:"ontoggle",onUnload:"onunload",onVolumeChange:"onvolumechange",onWaiting:"onwaiting",onZoom:"onzoom",overlinePosition:"overline-position",overlineThickness:"overline-thickness",paintOrder:"paint-order",panose1:"panose-1",pointerEvents:"pointer-events",referrerPolicy:"referrerpolicy",renderingIntent:"rendering-intent",shapeRendering:"shape-rendering",stopColor:"stop-color",stopOpacity:"stop-opacity",strikethroughPosition:"strikethrough-position",strikethroughThickness:"strikethrough-thickness",strokeDashArray:"stroke-dasharray",strokeDashOffset:"stroke-dashoffset",strokeLineCap:"stroke-linecap",strokeLineJoin:"stroke-linejoin",strokeMiterLimit:"stroke-miterlimit",strokeOpacity:"stroke-opacity",strokeWidth:"stroke-width",tabIndex:"tabindex",textAnchor:"text-anchor",textDecoration:"text-decoration",textRendering:"text-rendering",transformOrigin:"transform-origin",typeOf:"typeof",underlinePosition:"underline-position",underlineThickness:"underline-thickness",unicodeBidi:"unicode-bidi",unicodeRange:"unicode-range",unitsPerEm:"units-per-em",vAlphabetic:"v-alphabetic",vHanging:"v-hanging",vIdeographic:"v-ideographic",vMathematical:"v-mathematical",vectorEffect:"vector-effect",vertAdvY:"vert-adv-y",vertOriginX:"vert-origin-x",vertOriginY:"vert-origin-y",wordSpacing:"word-spacing",writingMode:"writing-mode",xHeight:"x-height",playbackOrder:"playbackorder",timelineBegin:"timelinebegin"},properties:{about:o.commaOrSpaceSeparated,accentHeight:o.number,accumulate:null,additive:null,alignmentBaseline:null,alphabetic:o.number,amplitude:o.number,arabicForm:null,ascent:o.number,attributeName:null,attributeType:null,azimuth:o.number,bandwidth:null,baselineShift:null,baseFrequency:null,baseProfile:null,bbox:null,begin:null,bias:o.number,by:null,calcMode:null,capHeight:o.number,className:o.spaceSeparated,clip:null,clipPath:null,clipPathUnits:null,clipRule:null,color:null,colorInterpolation:null,colorInterpolationFilters:null,colorProfile:null,colorRendering:null,content:null,contentScriptType:null,contentStyleType:null,crossOrigin:null,cursor:null,cx:null,cy:null,d:null,dataType:null,defaultAction:null,descent:o.number,diffuseConstant:o.number,direction:null,display:null,dur:null,divisor:o.number,dominantBaseline:null,download:o.boolean,dx:null,dy:null,edgeMode:null,editable:null,elevation:o.number,enableBackground:null,end:null,event:null,exponent:o.number,externalResourcesRequired:null,fill:null,fillOpacity:o.number,fillRule:null,filter:null,filterRes:null,filterUnits:null,floodColor:null,floodOpacity:null,focusable:null,focusHighlight:null,fontFamily:null,fontSize:null,fontSizeAdjust:null,fontStretch:null,fontStyle:null,fontVariant:null,fontWeight:null,format:null,fr:null,from:null,fx:null,fy:null,g1:o.commaSeparated,g2:o.commaSeparated,glyphName:o.commaSeparated,glyphOrientationHorizontal:null,glyphOrientationVertical:null,glyphRef:null,gradientTransform:null,gradientUnits:null,handler:null,hanging:o.number,hatchContentUnits:null,hatchUnits:null,height:null,href:null,hrefLang:null,horizAdvX:o.number,horizOriginX:o.number,horizOriginY:o.number,id:null,ideographic:o.number,imageRendering:null,initialVisibility:null,in:null,in2:null,intercept:o.number,k:o.number,k1:o.number,k2:o.number,k3:o.number,k4:o.number,kernelMatrix:o.commaOrSpaceSeparated,kernelUnitLength:null,keyPoints:null,keySplines:null,keyTimes:null,kerning:null,lang:null,lengthAdjust:null,letterSpacing:null,lightingColor:null,limitingConeAngle:o.number,local:null,markerEnd:null,markerMid:null,markerStart:null,markerHeight:null,markerUnits:null,markerWidth:null,mask:null,maskContentUnits:null,maskUnits:null,mathematical:null,max:null,media:null,mediaCharacterEncoding:null,mediaContentEncodings:null,mediaSize:o.number,mediaTime:null,method:null,min:null,mode:null,name:null,navDown:null,navDownLeft:null,navDownRight:null,navLeft:null,navNext:null,navPrev:null,navRight:null,navUp:null,navUpLeft:null,navUpRight:null,numOctaves:null,observer:null,offset:null,onAbort:null,onActivate:null,onAfterPrint:null,onBeforePrint:null,onBegin:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnd:null,onEnded:null,onError:null,onFocus:null,onFocusIn:null,onFocusOut:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadStart:null,onMessage:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onMouseWheel:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRepeat:null,onReset:null,onResize:null,onScroll:null,onSeeked:null,onSeeking:null,onSelect:null,onShow:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnload:null,onVolumeChange:null,onWaiting:null,onZoom:null,opacity:null,operator:null,order:null,orient:null,orientation:null,origin:null,overflow:null,overlay:null,overlinePosition:o.number,overlineThickness:o.number,paintOrder:null,panose1:null,path:null,pathLength:o.number,patternContentUnits:null,patternTransform:null,patternUnits:null,phase:null,ping:o.spaceSeparated,pitch:null,playbackOrder:null,pointerEvents:null,points:null,pointsAtX:o.number,pointsAtY:o.number,pointsAtZ:o.number,preserveAlpha:null,preserveAspectRatio:null,primitiveUnits:null,propagate:null,property:o.commaOrSpaceSeparated,r:null,radius:null,referrerPolicy:null,refX:null,refY:null,rel:o.commaOrSpaceSeparated,rev:o.commaOrSpaceSeparated,renderingIntent:null,repeatCount:null,repeatDur:null,requiredExtensions:o.commaOrSpaceSeparated,requiredFeatures:o.commaOrSpaceSeparated,requiredFonts:o.commaOrSpaceSeparated,requiredFormats:o.commaOrSpaceSeparated,resource:null,restart:null,result:null,rotate:null,rx:null,ry:null,scale:null,seed:null,shapeRendering:null,side:null,slope:null,snapshotTime:null,specularConstant:o.number,specularExponent:o.number,spreadMethod:null,spacing:null,startOffset:null,stdDeviation:null,stemh:null,stemv:null,stitchTiles:null,stopColor:null,stopOpacity:null,strikethroughPosition:o.number,strikethroughThickness:o.number,string:null,stroke:null,strokeDashArray:o.commaOrSpaceSeparated,strokeDashOffset:null,strokeLineCap:null,strokeLineJoin:null,strokeMiterLimit:o.number,strokeOpacity:o.number,strokeWidth:null,style:null,surfaceScale:o.number,syncBehavior:null,syncBehaviorDefault:null,syncMaster:null,syncTolerance:null,syncToleranceDefault:null,systemLanguage:o.commaOrSpaceSeparated,tabIndex:o.number,tableValues:null,target:null,targetX:o.number,targetY:o.number,textAnchor:null,textDecoration:null,textRendering:null,textLength:null,timelineBegin:null,title:null,transformBehavior:null,type:null,typeOf:o.commaOrSpaceSeparated,to:null,transform:null,transformOrigin:null,u1:null,u2:null,underlinePosition:o.number,underlineThickness:o.number,unicode:null,unicodeBidi:null,unicodeRange:null,unitsPerEm:o.number,values:null,vAlphabetic:o.number,vMathematical:o.number,vectorEffect:null,vHanging:o.number,vIdeographic:o.number,version:null,vertAdvY:o.number,vertOriginX:o.number,vertOriginY:o.number,viewBox:null,viewTarget:null,visibility:null,width:null,widths:null,wordSpacing:null,writingMode:null,x:null,x1:null,x2:null,xChannelSelector:null,xHeight:o.number,y:null,y1:null,y2:null,yChannelSelector:null,z:null,zoomAndPan:null},space:"svg",transform:r.B})},1567:(e,t,n)=>{"use strict";n.d(t,{V:()=>c,r:()=>l});var r=n(641),a=n(953),o=n(8869),i=n(3605);let s=Symbol("DescriptionContext");function l({slot:e=(0,a.KR)({}),name:t="Description",props:n={}}={}){let o=(0,a.KR)([]);return(0,r.Gt)(s,{register:function(e){return o.value.push(e),()=>{let t=o.value.indexOf(e);-1!==t&&o.value.splice(t,1)}},slot:e,name:t,props:n}),(0,r.EW)((()=>o.value.length>0?o.value.join(" "):void 0))}let c=(0,r.pM)({name:"Description",props:{as:{type:[Object,String],default:"p"},id:{type:String,default:null}},setup(e,{attrs:t,slots:n}){var l;let c=null!=(l=e.id)?l:`headlessui-description-${(0,o.B)()}`,u=function(){let e=(0,r.WQ)(s,null);if(null===e)throw new Error("Missing parent");return e}();return(0,r.sV)((()=>(0,r.hi)(u.register(c)))),()=>{let{name:r="Description",slot:o=(0,a.KR)({}),props:s={}}=u,{...l}=e,d={...Object.entries(s).reduce(((e,[t,n])=>Object.assign(e,{[t]:(0,a.R1)(n)})),{}),id:c};return(0,i.XX)({ourProps:d,theirProps:l,slot:o.value,attrs:t,slots:n,name:r})}}})},1608:(e,t,n)=>{"use strict";n.d(t,{zT:()=>i});var r=n(2071),a=n(4476);const o=a.z.object({enum:a.z.array(a.z.string()).optional(),default:a.z.string().optional(),description:a.z.string().optional()}).extend({value:a.z.string().optional()}).refine((e=>(Array.isArray(e.enum)&&!e.enum.includes(e.default??"")&&e.enum.length>0&&(e.default=e.enum[0]),Array.isArray(e.enum)&&0===e.enum.length&&delete e.enum,!0))),i=a.z.object({url:a.z.string(),description:a.z.string().optional(),variables:a.z.record(a.z.string(),o).optional()}).extend({uid:r.B.brand()})},1636:(e,t,n)=>{"use strict";n.d(t,{A:()=>p});var r=n(641),a=n(953);const o="          .:=+++++=================-\n         .--#*                       :.\n        .-:-**                        -.\n      .:-::-+*                        =:\n     .:-:::-=#                         ::\n    .-:::----**                        ..\n   .-:::::---=#                         ..\n  :-::::::----**                         ..\n.:-::::::::----*=                        ..\n.-::::::::------+-                        ..\n..::::::::-------=                         ..\n .:::::----------++                        ..\n   .:::----------+**+*++*+*++*+++*++++++++++:.\n   ------------+*+=-=======================.\n   .----------+*+=========================:.\n    :--------+*+=-========================.\n    .=------+*+=-========================:.\n     .=----+*+=-------=================+-.\n     .----+*+=----------================.\n     .:=-+*+=----------=-==============-\n      .-+*+=----------------===========.\n       .-+=------------------====-====:.";var i=n(5881),s=n(8839);const l={class:"scale-75 flex flex-center w-full"},c={class:"relative"},u={class:"relative -ml-12"},d=(0,r.pM)({__name:"EmptyState",setup:e=>(e,t)=>((0,r.uX)(),(0,r.CE)("div",l,[(0,r.Lk)("div",c,[(0,r.bF)(s.A,{class:"keycap-hotkey border-transparent right-14 text-xl",hotkey:""}),(0,r.bF)(i.A,{art:(0,a.R1)(o),class:"!leading-[6px] text-c-3"},null,8,["art"])]),(0,r.Lk)("div",u,[t[0]||(t[0]=(0,r.Lk)("div",{class:"keycap-hotkey right-16 text-xl"},"K",-1)),(0,r.bF)(i.A,{art:(0,a.R1)(o),class:"!leading-[6px] keycap-n"},null,8,["art"])])]))}),p=(0,n(3526).A)(d,[["__scopeId","data-v-6c848a44"]])},1660:(e,t,n)=>{"use strict";n.d(t,{T:()=>a});var r=n(2741);function a(e){return r.Q.PROTOCOL.test(e)?e:`http://${e.replace(/^\//,"")}`}},1760:(e,t,n)=>{"use strict";n.d(t,{A:()=>p});var r=n(641),a=n(953),o=n(33),i=n(3751),s=n(1266);const l={class:"group-[.alert]:outline-orange w-full pr-4 -outline-offset-1 has-[:focus-visible]:rounded-[4px] has-[:focus-visible]:outline"},c={class:"text-c-1 whitespace-nowrap"},u={class:"text-c-1"},d={class:"flex h-4 w-4 items-center justify-center"},p=(0,r.pM)({__name:"DataTableInputSelect",props:{modelValue:{},value:{},default:{},canAddCustomValue:{type:Boolean,default:!0},type:{}},emits:["update:modelValue"],setup(e,{emit:t}){const n=e,p=t,h=(0,r.EW)((()=>n.value??[])),f=(0,a.KR)(!1),m=(0,a.KR)(""),g=(0,a.KR)(null);(0,r.wB)(m,(e=>{p("update:modelValue",e)}));const v=e=>{p("update:modelValue",e),f.value=!1},b=()=>{m.value.trim()&&v(m.value)},y=()=>{m.value.trim()||p("update:modelValue",""),f.value=!1},w=e=>n.modelValue.toString()===e;(0,r.wB)(f,(e=>{e&&(0,r.dY)((()=>{var e;null==(e=g.value)||e.focus()}))}));const O=(0,r.EW)((()=>void 0!==n.modelValue?n.modelValue:n.default)),k=(0,r.EW)((()=>{const e=new Set(n.modelValue.toString().split(", "));return h.value.filter((t=>e.has(t))).map((e=>({id:e,label:e,value:e})))})),x=(0,r.EW)((()=>h.value.map((e=>({id:e,label:e,value:e}))))),E=e=>{const t=e.map((e=>e.value));p("update:modelValue",t.join(", "))};return(e,t)=>((0,r.uX)(),(0,r.CE)("div",l,["array"===e.type?((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.BR),{key:0,modelValue:k.value,options:x.value,"onUpdate:modelValue":E},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(s.RO),{class:"custom-scroll h-full justify-start gap-1.5 px-2 py-1.5 pr-6 font-normal outline-none",fullWidth:"",variant:"ghost"},{default:(0,r.k6)((()=>[(0,r.Lk)("span",c,(0,o.v_)(k.value.length>0?k.value.map((e=>e.label)).join(", "):"Select a value"),1),(0,r.bF)((0,a.R1)(s.kv),{icon:"ChevronDown",size:"md"})])),_:1})])),_:1},8,["modelValue","options"])):f.value?(0,r.bo)(((0,r.uX)(),(0,r.CE)("input",{key:1,ref_key:"inputRef",ref:g,"onUpdate:modelValue":t[0]||(t[0]=e=>m.value=e),class:"text-c-1 w-full min-w-0 border-none px-2 py-1.5 outline-none",placeholder:"Value",type:"text",onBlur:y,onKeyup:(0,i.jR)(b,["enter"])},null,544)),[[i.Jo,m.value]]):((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.SM),{key:2,resize:"",value:O.value},{items:(0,r.k6)((()=>[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(h.value,(e=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.rA),{key:e,class:"group/item flex items-center gap-1.5 overflow-hidden text-ellipsis whitespace-nowrap",value:e,onClick:t=>v(e)},{default:(0,r.k6)((()=>[(0,r.Lk)("div",{class:(0,o.C4)(["flex h-4 w-4 items-center justify-center rounded-full p-[3px]",w(e)?"bg-c-accent text-b-1":"shadow-border text-transparent"])},[(0,r.bF)((0,a.R1)(s.kv),{class:"size-2.5",icon:"Checkmark",thickness:"3"})],2),(0,r.eW)(" "+(0,o.v_)(e),1)])),_:2},1032,["value","onClick"])))),128)),e.canAddCustomValue?((0,r.uX)(),(0,r.CE)(r.FK,{key:0},[h.value.length?((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.t),{key:0})):(0,r.Q3)("",!0),(0,r.bF)((0,a.R1)(s.rA),{class:"flex items-center gap-1.5",onClick:t[1]||(t[1]=e=>f.value=!0)},{default:(0,r.k6)((()=>[(0,r.Lk)("div",d,[(0,r.bF)((0,a.R1)(s.kv),{icon:"Add",size:"sm"})]),t[2]||(t[2]=(0,r.Lk)("span",null,"Add value",-1))])),_:1})],64)):(0,r.Q3)("",!0)])),default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(s.RO),{class:"h-full justify-start gap-1.5 overflow-auto whitespace-nowrap px-2 py-1.5 font-normal outline-none",fullWidth:"",variant:"ghost"},{default:(0,r.k6)((()=>[(0,r.Lk)("span",u,(0,o.v_)(O.value||"Select a value"),1),(0,r.bF)((0,a.R1)(s.kv),{icon:"ChevronDown",size:"md"})])),_:1})])),_:1},8,["value"]))]))}})},1775:(e,t,n)=>{"use strict";n.d(t,{I:()=>c});var r=n(5147),a=n(8627),o=n(7457);const i=/[A-Z]/g,s=/-[a-z]/g,l=/^data[-\w.:]+$/i;function c(e,t){const n=(0,o.S)(t);let c=t,p=a.R;if(n in e.normal)return e.property[e.normal[n]];if(n.length>4&&"data"===n.slice(0,4)&&l.test(t)){if("-"===t.charAt(4)){const e=t.slice(5).replace(s,d);c="data"+e.charAt(0).toUpperCase()+e.slice(1)}else{const e=t.slice(4);if(!s.test(e)){let n=e.replace(i,u);"-"!==n.charAt(0)&&(n="-"+n),t="data"+n}}p=r.E}return new p(c,t)}function u(e){return"-"+e.toLowerCase()}function d(e){return e.charAt(1).toUpperCase()}},1793:(e,t,n)=>{"use strict";function r(e){return void 0!==e&&Array.isArray(e)&&e.length>0&&e.some((e=>!0===e.isEntrypoint))}n.d(t,{W:()=>r})},1845:(e,t,n)=>{"use strict";n.d(t,{RN:()=>s,rl:()=>l});var r=n(2402);const a=["Escape","ArrowDown","ArrowUp","Enter","F1","F2","F3","F4","F5","F6","F7","F8","F9","F10","F11","F12"],o={Escape:{event:"closeModal"},Enter:{event:"executeRequest",modifiers:["default"]},b:{event:"toggleSidebar",modifiers:["default"]},k:{event:"openCommandPalette",modifiers:["default"]}},i={Alt:"altKey",Control:"ctrlKey",Shift:"shiftKey",Meta:"metaKey"},s=e=>e.map((e=>"default"===e?(0,r.U)()?"metaKey":"ctrlKey":i[e])),l=(e,t,{hotKeys:n=o,modifiers:r=["default"]}={})=>{const i=" "===e.key?"Space":e.key,l=n[i];l&&("Escape"===i?t.emit({[l.event]:e}):s(l.modifiers||r).every((t=>!0===e[t]))?t.emit({[l.event]:e}):!(e=>{if(!(e.target instanceof HTMLElement))return!1;const t=e.target;return"INPUT"===t.tagName?!a.includes(e.key):!("TEXTAREA"!==t.tagName&&!t.getAttribute("contenteditable")&&"true"!==t.contentEditable)})(e)&&void 0===l.modifiers&&t.emit({[l.event]:e}))}},1902:(e,t,n)=>{"use strict";n.d(t,{L:()=>a});var r=n(4048);const a=({baseUrl:e="",method:t,body:n,path:a,cookies:o,headers:i,query:s})=>{var l,c,u;const d=(0,r.Hh)(e,a,void 0,!0),p={method:t.toUpperCase(),url:d.toString(),httpVersion:"HTTP/1.1",headers:[],queryString:[],cookies:[],headersSize:-1,bodySize:-1};if(o.length&&(p.cookies=o.filter((e=>e.enabled)).map((({key:e,value:t})=>({name:e,value:t})))),i.length&&(p.headers=i.filter((e=>e.enabled&&!("accept"===e.key.toLowerCase()&&"*/*"===e.value))).map((({key:e,value:t})=>({name:e.replace(/\b\w/g,(e=>e.toUpperCase())),value:t})))),s.length&&(p.queryString=s.filter((e=>e.enabled)).map((({key:e,value:t})=>({name:e,value:t})))),n)try{const e=(null==(l=i.find((e=>"content-type"===e.key.toLowerCase())))?void 0:l.value)||"application/json";if("formData"===n.activeBody&&n.formData){const t={};n.formData.value.forEach((({key:e,value:n,file:r,enabled:a})=>{a&&(r?t[e]={type:"file",text:"BINARY",name:e||"blob",size:r.size,fileName:r.name,mimeType:r.type||"application/octet-stream"}:t[e]?(Array.isArray(t[e])||(t[e]=[t[e]]),t[e].push(n)):t[e]=n)})),"urlencoded"===(null==(c=n.formData)?void 0:c.encoding)?p.postData={mimeType:e,text:new URLSearchParams(t).toString()}:p.postData={mimeType:e,text:JSON.stringify(t)}}else"raw"===n.activeBody&&n.raw&&(p.postData={mimeType:e,text:(null==(u=n.raw)?void 0:u.value)??""})}catch{}return p}},1972:(e,t,n)=>{"use strict";n.d(t,{c:()=>s});var r=n(953),a=n(641),o=n(6413);function i(e,t){if(e)return e;let n=null!=t?t:"button";return"string"==typeof n&&"button"===n.toLowerCase()?"button":void 0}function s(e,t){let n=(0,r.KR)(i(e.value.type,e.value.as));return(0,a.sV)((()=>{n.value=i(e.value.type,e.value.as)})),(0,a.nT)((()=>{var e;n.value||(0,o.t)(t)&&(0,o.t)(t)instanceof HTMLButtonElement&&(null==(e=(0,o.t)(t))||!e.hasAttribute("type"))&&(n.value="button")})),n}},1987:(e,t,n)=>{"use strict";n.d(t,{h:()=>a});var r=n(7176);function a(e,t){const n={},a={};for(const t of e)Object.assign(n,t.property),Object.assign(a,t.normal);return new r.S(n,a,t)}},2071:(e,t,n)=>{"use strict";n.d(t,{B:()=>a,I:()=>o});var r=n(4476);const a=r.z.string().min(7).optional().default((()=>((e=21)=>{let t="",n=crypto.getRandomValues(new Uint8Array(e|=0));for(;e--;)t+="useandom-26T198340PX75pxJACKVERYMINDBUSHWOLF_GQZbfghjklqvwyzrict"[63&n[e]];return t})())),o=r.z.union([r.z.string().brand(),r.z.string().brand().array()]).array().default([])},2282:(e,t,n)=>{"use strict";function r(e){const t=[],n=String(e||"");let r=n.indexOf(","),a=0,o=!1;for(;!o;){-1===r&&(r=n.length,o=!0);const e=n.slice(a,r).trim();!e&&o||t.push(e),a=r+1,r=n.indexOf(",",a)}return t}function a(e,t){const n=t||{};return(""===e[e.length-1]?[...e,""]:e).join((n.padRight?" ":"")+","+(!1===n.padLeft?"":" ")).trim()}n.d(t,{A:()=>a,q:()=>r})},2389:(e,t,n)=>{"use strict";n.d(t,{M:()=>o});var r=n(641),a=n(9146);function o(e,t,n){a._.isServer||(0,r.nT)((r=>{window.addEventListener(e,t,n),r((()=>window.removeEventListener(e,t,n)))}))}},2402:(e,t,n)=>{"use strict";function r(){return"undefined"!=typeof navigator&&/Mac/.test(navigator.platform)}n.d(t,{U:()=>r})},2453:(e,t,n)=>{"use strict";n.d(t,{J:()=>o});var r=n(6349),a=n(1902);const o=({operation:e,example:t,server:n,securitySchemes:o=[]})=>{const i=(0,r.A)(o,{},"YOUR_SECRET_TOKEN"),s=[...(null==t?void 0:t.parameters.headers)??[],...Object.entries(i.headers).map((([e,t])=>({key:e,value:t,enabled:!0})))],l=[...(null==t?void 0:t.parameters.cookies)??[],...i.cookies.map((e=>({key:e.name,value:e.value,enabled:!0})))],c=[...(null==t?void 0:t.parameters.query)??[],...Array.from(i.urlParams.entries()).map((([e,t])=>({key:e,value:t,enabled:!0})))];return(0,a.L)({baseUrl:null==n?void 0:n.url,method:(null==e?void 0:e.method)??"get",path:(null==e?void 0:e.path)??"/",body:null==t?void 0:t.body,cookies:l,headers:s,query:c})}},2485:(e,t,n)=>{"use strict";function r(e){const t=String(e||"").trim();return t?t.split(/[ \t\n\r\f]+/g):[]}function a(e){return e.join(" ").trim()}n.d(t,{A:()=>a,q:()=>r})},2490:(e,t,n)=>{"use strict";n.d(t,{Go:()=>a,TU:()=>r});var r=(e=>(e.Request="request",e.Examples="examples",e.Cookies="cookies",e.Collection="collection",e.Schema="schema",e.Environment="environment",e.Servers="servers",e.Workspace="workspace",e.Settings="settings",e))(r||{});const a=[{name:"root",path:"/",redirect:function(){const e=localStorage.getItem("activeWorkspace");return e?{name:"request.default",params:{workspace:e}}:{name:"workspace.default"}}},{name:"workspace.default",path:"/workspace",redirect:{name:"workspace",params:{workspace:"default"}}},{name:"workspace",path:"/workspace/:workspace",redirect:{name:"request.default"},children:[{name:"request.root",path:"",redirect:e=>({name:"request",params:{...e.params,request:"default"}})},{name:"request.default",path:"request",redirect:e=>({name:"request",params:{...e.params,request:"default"}})},{name:"request",path:"request/:request",component:()=>Promise.resolve().then(n.bind(n,2975))},{name:"request.examples",path:"request/:request/examples/:examples",component:()=>Promise.resolve().then(n.bind(n,2975))}]}]},2549:(e,t,n)=>{"use strict";n.d(t,{A:()=>w});var r=n(641),a=n(953),o=n(33),i=n(1266),s=n(6584),l=n(8309),c=n(3751),u=n(9347);const d=["aria-expanded"],p={class:"overflow-hidden text-ellipsis whitespace-nowrap"},h=["id"],f={key:0},m={class:"description text-c-3 px-3 py-1.5"},g=(0,r.pM)({__name:"ServerDropdownItem",props:{collection:{},operation:{},server:{},serverOption:{},type:{},layout:{}},emits:["update:variable"],setup(e,{emit:t}){const n=e,s=t,g=(0,r.Bi)(),{collectionMutators:v,requestMutators:b,servers:y}=(0,l.hA)(),w=(0,r.EW)((()=>{var e;return(null==(e=n.server)?void 0:e.uid)===n.serverOption.id})),O=e=>{if(!e)return!1;const t=y[e];return Object.keys((null==t?void 0:t.variables)??{}).length>0},k=(0,r.EW)((()=>{var e;return w.value&&O((null==(e=n.server)?void 0:e.uid)??"")})),x=(e,t)=>{s("update:variable",e,t)};return(e,t)=>{var s,l;return(0,r.uX)(),(0,r.CE)("div",{class:(0,o.C4)(["group/item flex min-h-fit flex-col rounded border",{"border-transparent":!w.value}])},[(0,r.Lk)("button",(0,r.v6)(k.value?{"aria-controls":(0,a.R1)(g)}:{},{"aria-expanded":k.value,class:["flex min-h-8 cursor-pointer items-center gap-1.5 rounded px-1.5",w.value?"text-c-1 bg-b-2":"hover:bg-b-2"],type:"button",onClick:t[0]||(t[0]=t=>((e,t)=>{var r,a;O(e)&&"reference"!==n.layout&&(null==t||t.stopPropagation()),"collection"===n.type&&n.collection?(null!=(a=null==(r=n.operation)?void 0:r.servers)&&a.length&&b.edit(n.operation.uid,"selectedServerUid",null),v.edit(n.collection.uid,"selectedServerUid",e)):"request"===n.type&&n.operation&&b.edit(n.operation.uid,"selectedServerUid",e)})(e.serverOption.id,t))}),[(0,r.bF)((0,a.R1)(i.Pq),{selected:w.value},null,8,["selected"]),(0,r.Lk)("span",p,(0,o.v_)(e.serverOption.label),1)],16,d),k.value&&"reference"!==n.layout?((0,r.uX)(),(0,r.CE)("div",{key:0,id:(0,a.R1)(g),class:"bg-b-2 divide divide-y rounded-b border-t *:pl-4",onClick:t[1]||(t[1]=(0,c.D$)((()=>{}),["stop"]))},[(0,r.bF)(u.A,{variables:null==(s=e.server)?void 0:s.variables,"onUpdate:variable":x},null,8,["variables"]),null!=(l=e.server)&&l.description?((0,r.uX)(),(0,r.CE)("div",f,[(0,r.Lk)("div",m,[(0,r.bF)((0,a.R1)(i.mi),{value:e.server.description},null,8,["value"])])])):(0,r.Q3)("",!0)],8,h)):(0,r.Q3)("",!0)],2)}}}),v=(0,n(3526).A)(g,[["__scopeId","data-v-8eba7808"]]),b=["onClick"],y={class:"flex h-4 w-4 items-center justify-center"},w=(0,r.pM)({__name:"ServerDropdown",props:{collection:{},operation:{},server:{},layout:{},target:{}},setup(e){const{layout:t}=(0,s.g)(),{servers:n,collectionMutators:c,events:u,serverMutators:d}=(0,l.hA)(),p=(0,r.EW)((()=>{var t,r;return null==(r=null==(t=e.operation)?void 0:t.servers)?void 0:r.map((e=>{var t;return{id:e,label:(null==(t=n[e])?void 0:t.url)??"Unknown server"}}))})),h=(0,r.EW)((()=>{var t,r;return null==(r=null==(t=e.collection)?void 0:t.servers)?void 0:r.filter((t=>{var n,r;return!(null!=(r=null==(n=e.operation)?void 0:n.servers)&&r.includes(t))})).map((e=>{var t;return{id:e,label:(null==(t=n[e])?void 0:t.url)??"Unknown server"}}))})),f=(0,r.EW)((()=>{var e,t;return(null==(e=p.value)?void 0:e.length)&&(null==(t=h.value)?void 0:t.length)}));(0,r.wB)([()=>e.collection,()=>e.operation],(([t,n])=>{var r;if(!t||t.selectedServerUid||null!=n&&n.selectedServerUid)return;const a=null==(r=e.collection.servers)?void 0:r[0];a&&c.edit(e.collection.uid,"selectedServerUid",a)}));const m=()=>u.commandPalette.emit({commandName:"Add Server"}),g=(0,r.EW)((()=>{var t,n,r;return null!=(n=null==(t=e.server)?void 0:t.url)&&n.endsWith("/")?e.server.url.slice(0,-1):(null==(r=e.server)?void 0:r.url)||""})),w=(t,n)=>{if(!e.server)return;const r=e.server.variables||{};r[t]={...r[t],default:n},d.edit(e.server.uid,"variables",r)},O=(0,i.Fj)({base:"gap-0.75 z-context-plus lg:text-sm text-xs whitespace-nowrap px-1.5 h-6.5",variants:{reference:{true:"!font-normal justify-start px-3 py-1.5 rounded-b-lg text-c-1 w-full",false:"border hover:bg-b-2 font-code ml-0.75 rounded text-c-2"}}});return(e,n)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(i.w),{class:"max-h-[inherit] p-0 text-sm",offset:"reference"===e.layout?6:0,placement:"bottom-start",resize:"",target:e.target,teleport:`#${e.target}`},{popover:(0,r.k6)((({close:s})=>[(0,r.Lk)("div",{class:(0,o.C4)(["custom-scroll flex max-h-[inherit] flex-col gap-1 p-1","reference"!==e.layout&&"border-t"]),onClick:s},[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(p.value,(t=>((0,r.uX)(),(0,r.Wv)(v,{key:t.id,collection:e.collection,layout:e.layout,operation:e.operation,server:e.server,serverOption:t,type:"request","onUpdate:variable":w},null,8,["collection","layout","operation","server","serverOption"])))),128)),f.value?((0,r.uX)(),(0,r.CE)(r.FK,{key:0},[(0,r.bF)((0,a.R1)(i.t)),n[1]||(n[1]=(0,r.Lk)("div",{class:"text-xxs text-c-2 px-2.5 py-1"},"Collection",-1))],64)):(0,r.Q3)("",!0),((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(h.value,(t=>((0,r.uX)(),(0,r.Wv)(v,{key:t.id,collection:e.collection,layout:e.layout,operation:e.operation,server:e.server,serverOption:t,type:"collection","onUpdate:variable":w},null,8,["collection","layout","operation","server","serverOption"])))),128)),"modal"!==(0,a.R1)(t)?((0,r.uX)(),(0,r.CE)("button",{key:1,class:"text-xxs p-1.75 hover:bg-b-2 flex cursor-pointer items-center gap-1.5 rounded",type:"button",onClick:m},[(0,r.Lk)("div",y,[(0,r.bF)((0,a.R1)(i.kv),{icon:"Add",size:"sm"})]),n[2]||(n[2]=(0,r.Lk)("span",null,"Add Server",-1))])):(0,r.Q3)("",!0)],10,b)])),backdrop:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(i.FI),{class:"-top-[--scalar-address-bar-height] rounded-lg"})])),default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(i.RO),{class:(0,o.C4)((0,a.R1)(i.cx)((0,a.R1)(O)({reference:"reference"===e.layout}))),variant:"ghost"},{default:(0,r.k6)((()=>[n[0]||(n[0]=(0,r.Lk)("span",{class:"sr-only"},"Server:",-1)),(0,r.eW)(" "+(0,o.v_)(g.value)+" ",1),"reference"===e.layout?((0,r.uX)(),(0,r.Wv)((0,a.R1)(i.kv),{key:0,class:"text-c-2",icon:"ChevronDown",size:"sm"})):(0,r.Q3)("",!0)])),_:1},8,["class"])])),_:1},8,["offset","target","teleport"]))}})},2555:(e,t,n)=>{"use strict";n.d(t,{O$:()=>i});var r=n(4476);const a=r.z.union([r.z.object({description:r.z.string().optional(),default:r.z.string().default("")}),r.z.string()]),o=r.z.object({description:r.z.string().optional(),color:r.z.string().optional(),variables:r.z.record(r.z.string(),a)}),i=r.z.record(r.z.string(),o)},2661:(e,t)=>{"use strict";t.removeLeadingAndTrailingHTTPWhitespace=e=>e.replace(/^[ \t\n\r]+/u,"").replace(/[ \t\n\r]+$/u,""),t.removeTrailingHTTPWhitespace=e=>e.replace(/[ \t\n\r]+$/u,""),t.isHTTPWhitespaceChar=e=>" "===e||"\t"===e||"\n"===e||"\r"===e,t.solelyContainsHTTPTokenCodePoints=e=>/^[-!#$%&'*+.^_`|~A-Za-z0-9]*$/u.test(e),t.soleyContainsHTTPQuotedStringTokenCodePoints=e=>/^[\t\u0020-\u007E\u0080-\u00FF]*$/u.test(e),t.asciiLowercase=e=>e.replace(/[A-Z]/gu,(e=>e.toLowerCase())),t.collectAnHTTPQuotedString=(e,t)=>{let n="";for(t++;;){for(;t<e.length&&'"'!==e[t]&&"\\"!==e[t];)n+=e[t],++t;if(t>=e.length)break;const r=e[t];if(++t,"\\"!==r)break;if(t>=e.length){n+="\\";break}n+=e[t],++t}return[n,t]}},2667:(e,t,n)=>{"use strict";n.d(t,{$n:()=>Oo,Sb:()=>yo});var r={};n.r(r),n.d(r,{default:()=>Ie,render:()=>Ne});var a={};n.r(a),n.d(a,{default:()=>Fe,render:()=>Be});var o={};n.r(o),n.d(o,{default:()=>ze,render:()=>Ue});var i={};n.r(i),n.d(i,{default:()=>Xe,render:()=>We});var s={};n.r(s),n.d(s,{default:()=>He,render:()=>Ve});var l={};n.r(l),n.d(l,{default:()=>Ke,render:()=>Ge});var c={};n.r(c),n.d(c,{default:()=>et,render:()=>Je});var u={};n.r(u),n.d(u,{default:()=>rt,render:()=>nt});var d={};n.r(d),n.d(d,{default:()=>it,render:()=>ot});var p={};n.r(p),n.d(p,{default:()=>ct,render:()=>lt});var h={};n.r(h),n.d(h,{default:()=>pt,render:()=>dt});var f={};n.r(f),n.d(f,{default:()=>mt,render:()=>ft});var m={};n.r(m),n.d(m,{default:()=>bt,render:()=>vt});var g={};n.r(g),n.d(g,{default:()=>Ot,render:()=>wt});var v={};n.r(v),n.d(v,{default:()=>Et,render:()=>xt});var b={};n.r(b),n.d(b,{default:()=>Ct,render:()=>_t});var y={};n.r(y),n.d(y,{default:()=>Rt,render:()=>At});var w={};n.r(w),n.d(w,{default:()=>Mt,render:()=>Dt});var O={};n.r(O),n.d(O,{default:()=>It,render:()=>Nt});var k={};n.r(k),n.d(k,{default:()=>Ft,render:()=>Bt});var x={};n.r(x),n.d(x,{default:()=>zt,render:()=>Ut});var E={};n.r(E),n.d(E,{default:()=>Xt,render:()=>Wt});var S={};n.r(S),n.d(S,{default:()=>Ht,render:()=>Vt});var _={};n.r(_),n.d(_,{default:()=>Kt,render:()=>Gt});var C={};n.r(C),n.d(C,{default:()=>en,render:()=>Jt});var T={};n.r(T),n.d(T,{default:()=>rn,render:()=>nn});var A={};n.r(A),n.d(A,{default:()=>sn,render:()=>on});var R={};n.r(R),n.d(R,{default:()=>un,render:()=>cn});var P={};n.r(P),n.d(P,{default:()=>hn,render:()=>pn});var D={};n.r(D),n.d(D,{default:()=>gn,render:()=>mn});var M={};n.r(M),n.d(M,{default:()=>yn,render:()=>bn});var $={};n.r($),n.d($,{default:()=>kn,render:()=>On});var N={};n.r(N),n.d(N,{default:()=>Sn,render:()=>En});var I={};n.r(I),n.d(I,{default:()=>Tn,render:()=>Cn});var L={};n.r(L),n.d(L,{default:()=>Pn,render:()=>Rn});var B={};n.r(B),n.d(B,{default:()=>$n,render:()=>Mn});var F={};n.r(F),n.d(F,{default:()=>Ln,render:()=>In});var Q={};n.r(Q),n.d(Q,{default:()=>Qn,render:()=>Fn});var U={};n.r(U),n.d(U,{default:()=>jn,render:()=>zn});var z={};n.r(z),n.d(z,{default:()=>qn,render:()=>Xn});var j={};n.r(j),n.d(j,{default:()=>Zn,render:()=>Hn});var W={};n.r(W),n.d(W,{default:()=>Yn,render:()=>Kn});var X={};n.r(X),n.d(X,{default:()=>tr,render:()=>er});var q={};n.r(q),n.d(q,{default:()=>ar,render:()=>rr});var V={};n.r(V),n.d(V,{default:()=>sr,render:()=>ir});var H={};n.r(H),n.d(H,{default:()=>ur,render:()=>cr});var Z={};n.r(Z),n.d(Z,{default:()=>hr,render:()=>pr});var G={};n.r(G),n.d(G,{default:()=>gr,render:()=>mr});var K={};n.r(K),n.d(K,{default:()=>yr,render:()=>br});var Y={};n.r(Y),n.d(Y,{default:()=>kr,render:()=>Or});var J={};n.r(J),n.d(J,{default:()=>Sr,render:()=>Er});var ee={};n.r(ee),n.d(ee,{default:()=>Tr,render:()=>Cr});var te={};n.r(te),n.d(te,{default:()=>Pr,render:()=>Rr});var ne={};n.r(ne),n.d(ne,{default:()=>$r,render:()=>Mr});var re={};n.r(re),n.d(re,{default:()=>Lr,render:()=>Ir});var ae={};n.r(ae),n.d(ae,{default:()=>Qr,render:()=>Fr});var oe={};n.r(oe),n.d(oe,{default:()=>jr,render:()=>zr});var ie={};n.r(ie),n.d(ie,{default:()=>qr,render:()=>Xr});var se={};n.r(se),n.d(se,{default:()=>Zr,render:()=>Hr});var le={};n.r(le),n.d(le,{default:()=>Yr,render:()=>Kr});var ce={};n.r(ce),n.d(ce,{default:()=>ta,render:()=>ea});var ue={};n.r(ue),n.d(ue,{default:()=>aa,render:()=>ra});var de={};n.r(de),n.d(de,{default:()=>sa,render:()=>ia});var pe={};n.r(pe),n.d(pe,{default:()=>ua,render:()=>ca});var he={};n.r(he),n.d(he,{default:()=>ha,render:()=>pa});var fe={};n.r(fe),n.d(fe,{default:()=>ga,render:()=>ma});var me={};n.r(me),n.d(me,{default:()=>ya,render:()=>ba});var ge={};n.r(ge),n.d(ge,{default:()=>ka,render:()=>Oa});var ve={};n.r(ve),n.d(ve,{default:()=>Sa,render:()=>Ea});var be={};n.r(be),n.d(be,{default:()=>Ta,render:()=>Ca});var ye={};n.r(ye),n.d(ye,{default:()=>Pa,render:()=>Ra});var we={};n.r(we),n.d(we,{default:()=>$a,render:()=>Ma});var Oe={};n.r(Oe),n.d(Oe,{default:()=>La,render:()=>Ia});var ke={};n.r(ke),n.d(ke,{default:()=>Qa,render:()=>Fa});var xe={};n.r(xe),n.d(xe,{default:()=>ja,render:()=>za});var Ee={};n.r(Ee),n.d(Ee,{default:()=>qa,render:()=>Xa});var Se={};n.r(Se),n.d(Se,{default:()=>Za,render:()=>Ha});var _e={};n.r(_e),n.d(_e,{default:()=>Ya,render:()=>Ka});var Ce={};n.r(Ce),n.d(Ce,{default:()=>to,render:()=>eo});var Te={};n.r(Te),n.d(Te,{default:()=>ao,render:()=>ro});var Ae={};n.r(Ae),n.d(Ae,{default:()=>so,render:()=>io});var Re={};n.r(Re),n.d(Re,{default:()=>uo,render:()=>co});var Pe={};n.r(Pe),n.d(Pe,{default:()=>fo,render:()=>ho});var De={};n.r(De),n.d(De,{default:()=>vo,render:()=>go});var Me=n(641);const $e={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ne(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",$e,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m10.812 2.492-8.32 8.32a1.68 1.68 0 0 0 0 2.377l8.32 8.319a1.68 1.68 0 0 0 2.377 0l8.319-8.32a1.68 1.68 0 0 0 0-2.377l-8.32-8.319a1.68 1.68 0 0 0-2.377 0Z"},null,-1)]))}const Ie={render:Ne},Le={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Be(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Le,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M20.711 16.963V7.044a1.37 1.37 0 0 0-.759-1.229l-7.338-3.67a1.37 1.37 0 0 0-1.227 0l-7.34 3.662a1.37 1.37 0 0 0-.758 1.23v9.919a1.37 1.37 0 0 0 .759 1.229l7.34 3.67a1.38 1.38 0 0 0 1.229 0l7.34-3.663a1.37 1.37 0 0 0 .754-1.23Z"},null,-1)]))}const Fe={render:Be},Qe={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ue(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Qe,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M2 12a10 10 0 1 0 20 0 10 10 0 0 0-20 0"},null,-1)]))}const ze={render:Ue},je={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function We(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",je,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M19 3H5a2 2 0 0 0-2 2v14a2 2 0 0 0 2 2h14a2 2 0 0 0 2-2V5a2 2 0 0 0-2-2"},null,-1)]))}const Xe={render:We},qe={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ve(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",qe,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M3.334 4.82v6.768a10.58 10.58 0 0 0 6.783 9.878l.926.356a2.67 2.67 0 0 0 1.914 0l.926-.356a10.58 10.58 0 0 0 6.783-9.878V4.821a1.32 1.32 0 0 0-.771-1.211A19.3 19.3 0 0 0 12 2a19.3 19.3 0 0 0-7.895 1.61 1.32 1.32 0 0 0-.771 1.21"},null,-1)]))}const He={render:Ve},Ze={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ge(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Ze,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M3 20h18M5 4h14a2 2 0 0 1 2 2v8a2 2 0 0 1-2 2H5a2 2 0 0 1-2-2V6a2 2 0 0 1 2-2"},null,-1)]))}const Ke={render:Ge},Ye={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Je(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Ye,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M18 8V6a2 2 0 0 0-2-2H4a2 2 0 0 0-2 2v7a2 2 0 0 0 2 2h8m-5 4h5m6-7h2a2 2 0 0 1 2 2v6a2 2 0 0 1-2 2h-2a2 2 0 0 1-2-2v-6a2 2 0 0 1 2-2"},null,-1)]))}const et={render:Je},tt={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function nt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",tt,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M20.522 14.167V6.64a2.13 2.13 0 0 0-2.13-2.13H5.608a2.13 2.13 0 0 0-2.13 2.13v7.526m17.044 0H3.478m17.044 0 1.364 4.778a1.066 1.066 0 0 1-.959 1.545H3.073a1.065 1.065 0 0 1-.959-1.545l1.364-4.778m6.855 3.333h3.334"},null,-1)]))}const rt={render:nt},at={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ot(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",at,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M8.333 2h7.334s2.666 0 2.666 2.667v14.666s0 2.667-2.666 2.667H8.333s-2.666 0-2.666-2.667V4.667S5.667 2 8.333 2"},null,-1),(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M9.186 4.563h1.246s.889 0 .889.889V9.88s0 .889-.889.889H9.186s-.89 0-.89-.89V5.453s0-.89.89-.89"},null,-1)]))}const it={render:ot},st={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function lt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",st,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M15.667 2H8.333a2.667 2.667 0 0 0-2.666 2.667v14.666A2.667 2.667 0 0 0 8.333 22h7.334a2.667 2.667 0 0 0 2.666-2.667V4.667A2.667 2.667 0 0 0 15.667 2"},null,-1),(0,Me.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M11 5a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const ct={render:lt},ut={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function dt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",ut,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M15.667 2H8.333a2.667 2.667 0 0 0-2.666 2.667v14.666A2.667 2.667 0 0 0 8.333 22h7.334a2.667 2.667 0 0 0 2.666-2.667V4.667A2.667 2.667 0 0 0 15.667 2"},null,-1),(0,Me.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M8 5a1 1 0 1 1 2 0 1 1 0 0 1-2 0m3 2a1 1 0 1 1 2 0 1 1 0 0 1-2 0M8 9a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const pt={render:dt},ht={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ft(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",ht,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M7.198 2h9.603s1.74 0 1.74 1.74v16.52s0 1.74-1.74 1.74H7.198s-1.74 0-1.74-1.74V3.74S5.459 2 7.199 2Z"},null,-1),(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M14.684 4.935H9.315a.94.94 0 0 1-.932-.8L8.067 2h7.865l-.315 2.133a.94.94 0 0 1-.933.802"},null,-1)]))}const mt={render:ft},gt={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function vt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",gt,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M21 11V7a2 2 0 0 0-2-2H5a2 2 0 0 0-2 2v10a2 2 0 0 0 2 2h6m1.034-7.319a.499.499 0 0 1 .647-.647l9 3.5a.5.5 0 0 1-.033.943l-3.444 1.068a1 1 0 0 0-.66.66l-1.067 3.443a.5.5 0 0 1-.943.033z"},null,-1)]))}const bt={render:vt},yt={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function wt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",yt,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M15.667 22H8.333a2.667 2.667 0 0 1-2.666-2.667V4.667A2.667 2.667 0 0 1 8.333 2h7.334a2.667 2.667 0 0 1 2.666 2.667v14.666A2.667 2.667 0 0 1 15.667 22"},null,-1),(0,Me.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M11 19a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const Ot={render:wt},kt={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function xt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",kt,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M10 7.5v2.813M14 7.5v2.813M14 7.5h2.5a1.125 1.125 0 0 1 1.125 1.125v5.062a1.125 1.125 0 0 1-1.125 1.125h-.543c-.373 0-.73.149-.994.412l-.864.864a1.4 1.4 0 0 1-.994.412h-2.21c-.373 0-.73-.148-.994-.412l-.864-.864a1.4 1.4 0 0 0-.994-.412H7.5a1.125 1.125 0 0 1-1.125-1.125V8.625A1.125 1.125 0 0 1 7.5 7.5zM4.2 3h15.6S21 3 21 4.2v15.6s0 1.2-1.2 1.2H4.2S3 21 3 19.8V4.2S3 3 4.2 3"},null,-1)]))}const Et={render:xt},St={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function _t(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",St,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M22 12c0 5.523-4.477 10-10 10m10-10c0-5.523-4.477-10-10-10m10 10H2m10 10C6.477 22 2 17.523 2 12m10 10a14.5 14.5 0 0 1 0-20m0 20a14.5 14.5 0 0 0 0-20M2 12C2 6.477 6.477 2 12 2"},null,-1)]))}const Ct={render:_t},Tt={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function At(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Tt,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M2 8.82a15 15 0 0 1 20 0M5 12.859a10 10 0 0 1 14 0m-10.5 3.57a5 5 0 0 1 7 0"},null,-1),(0,Me.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M11 20a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const Rt={render:At},Pt={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Dt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Pt,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M14.7 10.2v3.6m2.556-6.147a3.6 3.6 0 0 0-5.094 0m7.632-2.547a7.2 7.2 0 0 0-10.179 0M4.8 13.8h14.4a1.8 1.8 0 0 1 1.8 1.8v3.6a1.8 1.8 0 0 1-1.8 1.8H4.8A1.8 1.8 0 0 1 3 19.2v-3.6a1.8 1.8 0 0 1 1.8-1.8"},null,-1),(0,Me.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M6 17.5a1 1 0 1 1 2 0 1 1 0 0 1-2 0m4 0a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const Mt={render:Dt},$t={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Nt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",$t,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M8.5 2h7m-5.3 1v6.774a1.8 1.8 0 0 1-.19.807l-4.562 9.114A.9.9 0 0 0 6.258 21h11.484a.9.9 0 0 0 .81-1.305l-4.562-9.114a1.8 1.8 0 0 1-.19-.807V3M7.5 15.6h9"},null,-1)]))}const It={render:Nt},Lt={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Bt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Lt,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M4.802 13.8a.899.899 0 0 1-.702-1.467l8.908-9.178a.45.45 0 0 1 .774.414l-1.728 5.417A.9.9 0 0 0 12.9 10.2h6.298a.899.899 0 0 1 .702 1.467l-8.908 9.178a.45.45 0 0 1-.774-.414l1.728-5.417A.9.9 0 0 0 11.1 13.8z"},null,-1)]))}const Ft={render:Bt},Qt={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ut(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Qt,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 9.286a2.5 2.5 0 1 1 2.5 2.5m-2.5-2.5a2.5 2.5 0 1 0-2.5 2.5m2.5-2.5v.833m2.5 1.667a2.5 2.5 0 1 1-2.5 2.5m2.5-2.5h-.833m-4.167 0a2.5 2.5 0 1 0 2.5 2.5m-2.5-2.5h.833M12 10.119c-.92 0-1.667.746-1.667 1.667M12 10.119c.92 0 1.667.746 1.667 1.667M12 14.286v-.834m0 .834V20.5m-1.667-8.714c0 .92.746 1.666 1.667 1.666m1.667-1.666c0 .92-.746 1.666-1.667 1.666M4.2 3h15.6S21 3 21 4.2v15.6s0 1.2-1.2 1.2H4.2S3 21 3 19.8V4.2S3 3 4.2 3"},null,-1)]))}const zt={render:Ut},jt={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Wt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",jt,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m10.812 2.492-8.32 8.32a1.68 1.68 0 0 0 0 2.377l8.32 8.319a1.68 1.68 0 0 0 2.377 0l8.319-8.32a1.68 1.68 0 0 0 0-2.377l-8.32-8.319a1.68 1.68 0 0 0-2.377 0ZM12 8v4"},null,-1),(0,Me.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M11 16a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const Xt={render:Wt},qt={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Vt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",qt,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 9.57v4.296m7.926 7.516H4.074c-.798 0-1.318-.84-.96-1.554L11.04 3.976a1.074 1.074 0 0 1 1.92 0l7.927 15.852a1.074 1.074 0 0 1-.96 1.554Z"},null,-1),(0,Me.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M11 18a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const Ht={render:Vt},Zt={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Gt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Zt,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 22c5.523 0 10-4.477 10-10S17.523 2 12 2 2 6.477 2 12s4.477 10 10 10"},null,-1),(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round","stroke-miterlimit":"10",d:"M12 17v-6 0a.5.5 0 0 0-.5-.5l-.5.001h-1M12 17h-2m2 0h2"},null,-1),(0,Me.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M10.75 7.5a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const Kt={render:Gt},Yt={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Jt(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Yt,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M4.999 21.002H19M11.56 3.264a.5.5 0 0 1 .877 0l2.953 5.605a1 1 0 0 0 1.516.294L21.185 5.5a.5.5 0 0 1 .798.519l-2.835 10.248a1 1 0 0 1-.956.734H5.81a1 1 0 0 1-.957-.734L2.018 6.019a.5.5 0 0 1 .798-.52l4.277 3.665a1 1 0 0 0 1.516-.294l2.953-5.606Z"},null,-1)]))}const en={render:Jt},tn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function nn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",tn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M13.527 3.525a1.8 1.8 0 0 0-1.273-.527H5.798a1.8 1.8 0 0 0-1.8 1.8v6.456c0 .478.19.936.527 1.273l7.836 7.836a2.184 2.184 0 0 0 3.078 0l5.924-5.924a2.184 2.184 0 0 0 0-3.078z"},null,-1),(0,Me.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M8 8a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const rn={render:nn},an={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function on(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",an,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m19 21-7-4-7 4V5a2 2 0 0 1 2-2h10a2 2 0 0 1 2 2z"},null,-1)]))}const sn={render:on},ln={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function cn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",ln,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M11 14h1v4m4-16v4M3 10h18M8 2v4M5 4h14a2 2 0 0 1 2 2v14a2 2 0 0 1-2 2H5a2 2 0 0 1-2-2V6a2 2 0 0 1 2-2"},null,-1)]))}const un={render:cn},dn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function pn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",dn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M17.333 5.01c1.39-.224 2.796-.34 4.205-.346.121 0 .238.044.329.125.085.084.133.2.133.32v14a.453.453 0 0 1-.444.444C14.186 19.66 12 21.997 12 21.997m0 0V7.091m0 14.906s-2.187-2.337-9.556-2.444A.453.453 0 0 1 2 19.109v-14c0-.12.048-.236.133-.32a.5.5 0 0 1 .33-.125C9.822 4.762 12 7.091 12 7.091m0 0a5.5 5.5 0 0 1 1.458-3.05 6.9 6.9 0 0 1 3.235-2.017.53.53 0 0 1 .436.053.41.41 0 0 1 .204.356V16.31a7.27 7.27 0 0 0-3.875 2.186A5.33 5.33 0 0 0 12 21.562"},null,-1)]))}const hn={render:pn},fn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function mn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",fn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M11 3v8l3-3 3 3V3M5 3h14a2 2 0 0 1 2 2v14a2 2 0 0 1-2 2H5a2 2 0 0 1-2-2V5a2 2 0 0 1 2-2"},null,-1)]))}const gn={render:mn},vn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function bn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",vn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M15 3v4a2 2 0 0 0 2 2h4m-5-6H5a2 2 0 0 0-2 2v14a2 2 0 0 0 2 2h14a2 2 0 0 0 2-2V8z"},null,-1)]))}const yn={render:bn},wn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function On(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",wn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M13 6h6a2 2 0 0 1 2 2v11a2 2 0 0 1-2 2H5a2 2 0 0 1-2-2V5a2 2 0 0 1 2-2h3.93a2 2 0 0 1 1.67.9l.81 1.2a2 2 0 0 0 1.69.9"},null,-1)]))}const kn={render:On},xn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function En(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",xn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M16 4h2a2 2 0 0 1 2 2v14a2 2 0 0 1-2 2H6a2 2 0 0 1-2-2V6a2 2 0 0 1 2-2h2m1-2h6a1 1 0 0 1 1 1v2a1 1 0 0 1-1 1H9a1 1 0 0 1-1-1V3a1 1 0 0 1 1-1"},null,-1)]))}const Sn={render:En},_n={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Cn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",_n,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m20.72 11.12-8.51 8.51a5.559 5.559 0 1 1-7.86-7.861l7.935-7.936a3.71 3.71 0 0 1 5.25 5.241L9.581 17.01a1.853 1.853 0 1 1-2.62-2.62l7.861-7.853"},null,-1)]))}const Tn={render:Cn},An={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Rn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",An,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M10.2 16.8a3.6 3.6 0 1 1-7.2 0 3.6 3.6 0 0 1 7.2 0m0 0V5.7a2.1 2.1 0 0 0-4.125-.56L3.13 15.844m10.67.956a3.6 3.6 0 1 0 7.2 0 3.6 3.6 0 0 0-7.2 0m0 0V5.7a2.1 2.1 0 0 1 4.125-.56l2.945 10.7M10.2 9.6h3.6V12h-3.6z"},null,-1)]))}const Pn={render:Rn},Dn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Mn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Dn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M15 6V4m2 2.983 1.414-1.415M18 9h2m-3 2 1.2 1.2M4 20 15 9m-1.8-1.8L12 6"},null,-1)]))}const $n={render:Mn},Nn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function In(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Nn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M13.8 15.6v-1.8A1.8 1.8 0 0 0 12 12H4.8A1.8 1.8 0 0 1 3 10.2V7.5a1.8 1.8 0 0 1 1.8-1.8h1.8M19.2 3H8.4a1.8 1.8 0 0 0-1.8 1.8v1.8a1.8 1.8 0 0 0 1.8 1.8h10.8A1.8 1.8 0 0 0 21 6.6V4.8A1.8 1.8 0 0 0 19.2 3m-4.5 12.6h-1.8a.9.9 0 0 0-.9.9v3.6a.9.9 0 0 0 .9.9h1.8a.9.9 0 0 0 .9-.9v-3.6a.9.9 0 0 0-.9-.9"},null,-1)]))}const Ln={render:In},Bn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Fn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Bn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M16.494 3.913a2.56 2.56 0 0 1 1.855-.905 2.58 2.58 0 0 1 1.923.749 2.54 2.54 0 0 1 .741 1.915 2.53 2.53 0 0 1-.924 1.835l-12.13 12.13-4.943 1.348 1.348-4.942z"},null,-1)]))}const Qn={render:Fn},Un={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function zn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Un,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m15.477 12.89 1.515 8.526a.5.5 0 0 1-.81.47l-3.58-2.687a1 1 0 0 0-1.197 0l-3.586 2.686a.5.5 0 0 1-.81-.469l1.514-8.526M18 8A6 6 0 1 1 6 8a6 6 0 0 1 12 0"},null,-1)]))}const jn={render:zn},Wn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Xn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Wn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M4.8 14.7s.9-.9 3.6-.9 4.5 1.8 7.2 1.8 3.6-.9 3.6-.9V3.9s-.9.9-3.6.9S11.1 3 8.4 3s-3.6.9-3.6.9zm0 0V21"},null,-1)]))}const qn={render:Xn},Vn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Hn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Vn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M19 14c1.49-1.46 3-3.21 3-5.5A5.5 5.5 0 0 0 16.5 3c-1.76 0-3 .5-4.5 2-1.5-1.5-2.74-2-4.5-2A5.5 5.5 0 0 0 2 8.5c0 2.3 1.5 4.05 3 5.5l7 7z"},null,-1)]))}const Zn={render:Hn},Gn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Kn(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Gn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M11.525 2.295a.53.53 0 0 1 .95 0l2.31 4.679a2.12 2.12 0 0 0 1.595 1.16l5.166.756a.53.53 0 0 1 .294.904l-3.736 3.638a2.12 2.12 0 0 0-.611 1.878l.882 5.14a.53.53 0 0 1-.771.56l-4.618-2.428a2.12 2.12 0 0 0-1.973 0L6.396 21.01a.53.53 0 0 1-.77-.56l.881-5.139a2.12 2.12 0 0 0-.611-1.879L2.16 9.795a.53.53 0 0 1 .294-.906l5.165-.755a2.12 2.12 0 0 0 1.597-1.16z"},null,-1)]))}const Yn={render:Kn},Jn={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function er(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Jn,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M4 3v4M2 5h4m14 12v2m-1-1h2m-6.937-2.5a2 2 0 0 1 1.437-1.437l6.135-1.582a.5.5 0 0 0 0-.962L15.5 9.936A2 2 0 0 1 14.063 8.5l-1.582-6.135a.5.5 0 0 0-.963 0L9.937 8.5A2 2 0 0 1 8.5 9.937l-6.135 1.581a.5.5 0 0 0 0 .964L8.5 14.063A2 2 0 0 1 9.937 15.5l1.582 6.135a.5.5 0 0 0 .963 0z"},null,-1)]))}const tr={render:er},nr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function rr(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",nr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M7 11v4a2 2 0 0 0 2 2h4"},null,-1),(0,Me.Lk)("circle",{cx:"7",cy:"7",r:"3.25",stroke:"currentColor"},null,-1),(0,Me.Lk)("circle",{cx:"17",cy:"17",r:"3.25",stroke:"currentColor"},null,-1)]))}const ar={render:rr},or={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ir(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",or,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M15 21v-6a1 1 0 0 0-1-1h-4a1 1 0 0 0-1 1v6M3 10a2 2 0 0 1 .709-1.528l7-5.999a2 2 0 0 1 2.582 0l7 5.999A2 2 0 0 1 21 10v9a2 2 0 0 1-2 2H5a2 2 0 0 1-2-2z"},null,-1)]))}const sr={render:ir},lr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function cr(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",lr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M13.414 17.657 12 19.07A5 5 0 0 1 4.929 12l1.414-1.414m4.243-4.243L12 4.93A5 5 0 0 1 19.071 12l-1.414 1.414m-8.485 1.414 5.656-5.656"},null,-1)]))}const ur={render:cr},dr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function pr(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",dr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 2v2m0 16v2M4.93 4.93l1.41 1.41m11.32 11.32 1.41 1.41M2 12h2m16 0h2M6.34 17.66l-1.41 1.41M19.07 4.93l-1.41 1.41M16 12a4 4 0 1 1-8 0 4 4 0 0 1 8 0"},null,-1)]))}const hr={render:pr},fr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function mr(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",fr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M5.778 10h12.444c.982 0 1.778.895 1.778 2v8c0 1.105-.796 2-1.778 2H5.778C4.796 22 4 21.105 4 20v-8c0-1.105.796-2 1.778-2M7 10V7a5 5 0 1 1 10 0v3"},null,-1),(0,Me.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M11 16a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const gr={render:mr},vr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function br(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",vr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M5.778 10h12.444c.982 0 1.778.895 1.778 2v8c0 1.105-.796 2-1.778 2H5.778C4.796 22 4 21.105 4 20v-8c0-1.105.796-2 1.778-2M7 10V7a5 5 0 1 1 10 0v3"},null,-1),(0,Me.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M11 16a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const yr={render:br},wr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Or(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",wr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M7.5 21a5.5 5.5 0 1 0 0-11 5.5 5.5 0 0 0 0 11m3.9-9.4L19 4m-3.5 3.5 3 3M22 7l-3-3"},null,-1)]))}const kr={render:Or},xr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Er(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",xr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M11 19a8 8 0 1 0 0-16 8 8 0 0 0 0 16m10 2-4.35-4.35"},null,-1)]))}const Sr={render:Er},_r={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Cr(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",_r,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M2 11.78v.44a2 2 0 0 0 2 2h.18a2 2 0 0 1 1.73 1l.25.43a2 2 0 0 1 0 2l-.08.15a2 2 0 0 0 .73 2.73l.38.22a2 2 0 0 0 2.73-.73l.1-.15a2 2 0 0 1 1.72-1h.51a2 2 0 0 1 1.74 1l.09.15a2 2 0 0 0 2.73.73l.38-.22a2 2 0 0 0 .73-2.73l-.08-.15a2 2 0 0 1 0-2l.25-.43a2 2 0 0 1 1.73-1H20a2 2 0 0 0 2-2v-.44a2 2 0 0 0-2-2h-.18a2 2 0 0 1-1.73-1l-.25-.43a2 2 0 0 1 0-2l.08-.15a2 2 0 0 0-.73-2.73l-.39-.22a2 2 0 0 0-2.73.73l-.08.15a2 2 0 0 1-1.74 1h-.5a2 2 0 0 1-1.74-1l-.09-.15a2 2 0 0 0-2.73-.73l-.38.22a2 2 0 0 0-.73 2.73l.08.15a2 2 0 0 1 0 2l-.25.43a2 2 0 0 1-1.73 1H4a2 2 0 0 0-2 2"},null,-1),(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M15 12a3 3 0 1 0-6 0 3 3 0 0 0 6 0"},null,-1)]))}const Tr={render:Cr},Ar={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Rr(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Ar,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M16 9a5 5 0 0 1 0 6m3.364 3.364a9 9 0 0 0 0-12.728M11 4.702a.706.706 0 0 0-1.203-.498L6.413 7.587A1.4 1.4 0 0 1 5.416 8H3a1 1 0 0 0-1 1v6a1 1 0 0 0 1 1h2.416a1.4 1.4 0 0 1 .997.413l3.383 3.384A.705.705 0 0 0 11 19.298z"},null,-1)]))}const Pr={render:Rr},Dr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Mr(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Dr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M11.566 18.582v2.778s3.367-.611 4.445-2.222c.58-.87.599-2.196.465-3.334m-10.698-3.01H3s.611-3.367 2.222-4.445c.87-.58 2.196-.6 3.334-.465m-.82 8.883-1.571-1.571-.625-.625a20.2 20.2 0 0 1 1.837-4.667 10.76 10.76 0 0 1 5.199-4.825c2.26-1.003 4.914-1.314 7.712-.903.486 3.46.349 9.4-5.665 12.92a20.6 20.6 0 0 1-4.73 1.828l-.586-.585zm0 0L6.165 18.34"},null,-1)]))}const $r={render:Mr},Nr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ir(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Nr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m9 15 3-3m5 1a6 6 0 0 0-6-6m10 6A10 10 0 0 0 11 3m-7 7a7.31 7.31 0 0 0 10 10z"},null,-1)]))}const Lr={render:Ir},Br={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Fr(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Br,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M4 3v4M2 5h4m14 12v2m-1-1h2m-10.582-5.578L5.368 13.5a.763.763 0 0 1-.906-.574l-.439-1.757a.875.875 0 0 1 .565-1.034l11.036-3.629m-2.35 5.307 3.541-.755m-1.546 8.317L12.73 14.3M7.225 9.268l.91 3.642m.596 6.465 2.537-5.075m4.397-7.233a1.635 1.635 0 0 1 1.19-1.982l.89-.222a.817.817 0 0 1 .99.594l1.239 4.953a.82.82 0 0 1-.594.99l-.891.223a1.634 1.634 0 0 1-1.982-1.189zm-2.03 5.77a1.635 1.635 0 1 1-3.27 0 1.635 1.635 0 0 1 3.27 0"},null,-1)]))}const Qr={render:Fr},Ur={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function zr(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Ur,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m7.59 13.51 6.83 3.98m-.01-10.98-6.82 3.98M20 5a3 3 0 1 1-6 0 3 3 0 0 1 6 0M8 12a3 3 0 1 1-6 0 3 3 0 0 1 6 0m12 7a3 3 0 1 1-6 0 3 3 0 0 1 6 0"},null,-1)]))}const jr={render:zr},Wr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Xr(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Wr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M18 12h-1.488a1.2 1.2 0 0 0-1.158.876l-1.41 5.016a.15.15 0 0 1-.288 0L10.344 6.108a.15.15 0 0 0-.288 0l-1.41 5.016A1.2 1.2 0 0 1 7.494 12H6"},null,-1),(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M19 3H5a2 2 0 0 0-2 2v14a2 2 0 0 0 2 2h14a2 2 0 0 0 2-2V5a2 2 0 0 0-2-2"},null,-1)]))}const qr={render:Xr},Vr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Hr(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Vr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 6v6l4 2M2 12a10 10 0 1 0 20 0 10 10 0 0 0-20 0"},null,-1)]))}const Zr={render:Hr},Gr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Kr(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Gr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M17.263 5.71v.181c0 .558-.222 1.094-.617 1.489l-3.157 3.158a2.105 2.105 0 0 1-2.978 0L7.353 7.38a2.1 2.1 0 0 1-.616-1.489v-.18m10.526 0V3.605A2.105 2.105 0 0 0 15.158 1.5H8.842a2.105 2.105 0 0 0-2.105 2.105v2.106m10.526 0H6.737m0 15.789h10.526m-10.526 0 3.948-3.158a2.105 2.105 0 0 1 2.63 0l3.948 3.158m-10.526 0v-4.391c0-.559.222-1.094.616-1.489l3.158-3.158a2.105 2.105 0 0 1 2.978 0l3.157 3.158c.395.395.617.93.617 1.489V21.5"},null,-1)]))}const Yr={render:Kr},Jr={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ea(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Jr,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M16 21v-2a4 4 0 0 0-4-4H6a4 4 0 0 0-4 4v2m20 0v-2a4 4 0 0 0-3-3.87m-3-12a4 4 0 0 1 0 7.75M13 7a4 4 0 1 1-8 0 4 4 0 0 1 8 0"},null,-1)]))}const ta={render:ea},na={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ra(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",na,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 2a7.071 7.071 0 1 0 10 10M12 2a10 10 0 0 0-7.071 17.071M12 2a10 10 0 0 0-7.071 17.071M22 12a10 10 0 0 1-17.071 7.071M22 12a10 10 0 0 1-17.071 7.071"},null,-1)]))}const aa={render:ra},oa={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ia(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",oa,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M21 15a2 2 0 0 1-2 2H7l-4 4V5a2 2 0 0 1 2-2h14a2 2 0 0 1 2 2z"},null,-1)]))}const sa={render:ia},la={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ca(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",la,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M11.023 12.976a1.8 1.8 0 0 0-.603-.397L3.284 9.718a.45.45 0 0 1 .021-.843l17.098-5.85a.446.446 0 0 1 .572.572l-5.85 17.098a.45.45 0 0 1-.843.021l-2.861-7.138a1.8 1.8 0 0 0-.398-.601Zm0 0 9.845-9.843"},null,-1)]))}const ua={render:ca},da={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function pa(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",da,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m22 7-8.97 5.7a1.94 1.94 0 0 1-2.06 0L2 7m2-3h16a2 2 0 0 1 2 2v12a2 2 0 0 1-2 2H4a2 2 0 0 1-2-2V6a2 2 0 0 1 2-2"},null,-1)]))}const ha={render:pa},fa={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ma(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",fa,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M9 12h5M9 8h5m-9 9V5a2 2 0 0 1 2-2h13m0 0a2 2 0 0 0-2 2v14a2 2 0 0 1-2 2m4-18a2 2 0 0 1 2 2v2a1 1 0 0 1-1 1h-3m-2 13H4a2 2 0 0 1-2-2v-1a1 1 0 0 1 1-1h10a1 1 0 0 1 1 1v1a2 2 0 0 0 2 2"},null,-1)]))}const ga={render:ma},va={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ba(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",va,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 2v20m5-17H9.5a3.5 3.5 0 1 0 0 7h5a3.5 3.5 0 1 1 0 7H6"},null,-1)]))}const ya={render:ba},wa={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Oa(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",wa,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m22 7-8.5 8.5-5-5L2 17M22 7h-6m6 0v6"},null,-1)]))}const ka={render:Oa},xa={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ea(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",xa,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M8 17v-7m4 7v-4m4 4V8M5 21h14a2 2 0 0 0 2-2V5a2 2 0 0 0-2-2H5a2 2 0 0 0-2 2v14a2 2 0 0 0 2 2"},null,-1)]))}const Sa={render:Ea},_a={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ca(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",_a,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M3.011 20.99c1.124-.887 2.04-1.775 3.506-1.977 1.268 0 2.49.481 3.416 1.348a3.595 3.595 0 0 0 5.561-2.115 3.146 3.146 0 0 0 3.317-4.599 3.148 3.148 0 0 0 1.234-5.241 3.145 3.145 0 1 0-4.45-4.45 3.147 3.147 0 0 0-5.24 1.234 3.146 3.146 0 0 0-4.6 3.317 3.596 3.596 0 0 0-2.11 5.533c.714 1.035 1.532 2.197 1.344 3.444-.247 1.477-1.066 2.366-1.978 3.506m0 0L16.494 7.507"},null,-1)]))}const Ta={render:Ca},Aa={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ra(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Aa,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M21.038 16.44v2.71a1.81 1.81 0 0 1-1.97 1.807 17.9 17.9 0 0 1-7.794-2.773 17.6 17.6 0 0 1-5.42-5.42 17.9 17.9 0 0 1-2.773-7.83 1.806 1.806 0 0 1 1.798-1.97h2.71a1.806 1.806 0 0 1 1.806 1.554 11.6 11.6 0 0 0 .632 2.538 1.8 1.8 0 0 1-.406 1.906l-1.147 1.147a14.45 14.45 0 0 0 5.42 5.42l1.146-1.148a1.81 1.81 0 0 1 1.906-.406c.82.305 1.671.518 2.538.632a1.807 1.807 0 0 1 1.554 1.833"},null,-1)]))}const Pa={render:Ra},Da={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ma(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Da,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m8.21 2.5 1.782 1.781m4.016 0L15.79 2.5M9.158 7.36v-.947a2.845 2.845 0 1 1 5.684 0v.947M12 19.553c-3.126 0-5.684-2.558-5.684-5.685v-2.842a3.79 3.79 0 0 1 3.79-3.79h3.789a3.79 3.79 0 0 1 3.79 3.79v2.842c0 3.127-2.559 5.685-5.685 5.685m0 0v-8.527M6.818 9.132c-1.829-.19-3.344-1.8-3.344-3.79m2.842 7.58h-3.79m.948 7.578c0-1.99 1.61-3.695 3.6-3.79M20.498 5.342c0 1.99-1.516 3.6-3.316 3.79m4.292 3.79h-3.79m-.758 3.788c1.99.095 3.6 1.8 3.6 3.79"},null,-1)]))}const $a={render:Ma},Na={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ia(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Na,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M17.662 18.661a4.33 4.33 0 1 0-2.455-7.891 6.662 6.662 0 1 0-6.537 7.89h8.993Z"},null,-1)]))}const La={render:Ia},Ba={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Fa(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Ba,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M20 12a2 2 0 0 0 2-2V6a2 2 0 0 0-2-2H4a2 2 0 0 0-2 2v4a2 2 0 0 0 2 2m16 0H4m16 0a2 2 0 0 1 2 2v4a2 2 0 0 1-2 2H4a2 2 0 0 1-2-2v-4a2 2 0 0 1 2-2"},null,-1),(0,Me.Lk)("path",{fill:"currentColor",d:"M18 9a1 1 0 1 0 0-2 1 1 0 0 0 0 2m0 8a1 1 0 1 0 0-2 1 1 0 0 0 0 2"},null,-1),(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M6 8h4m-4 8h4"},null,-1)]))}const Qa={render:Fa},Ua={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function za(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Ua,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M3 6.601c0 .955.948 1.87 2.636 2.546S9.613 10.2 12 10.2s4.676-.38 6.364-1.054C20.052 8.47 21 7.556 21 6.6m-18 0c0-.955.948-1.87 2.636-2.546C7.324 3.38 9.613 3.001 12 3.001s4.676.38 6.364 1.054C20.052 4.731 21 5.646 21 6.601m-18 0v5.4m18-5.4v5.4M3 12c0 1.987 4.03 3.6 9 3.6s9-1.612 9-3.6M3 12v5.4c0 1.987 4.03 3.6 9 3.6s9-1.613 9-3.6V12"},null,-1)]))}const ja={render:za},Wa={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Xa(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Wa,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M9 3H4a1 1 0 0 0-1 1v5a1 1 0 0 0 1 1h5a1 1 0 0 0 1-1V4a1 1 0 0 0-1-1m11 0h-5a1 1 0 0 0-1 1v5a1 1 0 0 0 1 1h5a1 1 0 0 0 1-1V4a1 1 0 0 0-1-1m0 11h-5a1 1 0 0 0-1 1v5a1 1 0 0 0 1 1h5a1 1 0 0 0 1-1v-5a1 1 0 0 0-1-1M9 14H4a1 1 0 0 0-1 1v5a1 1 0 0 0 1 1h5a1 1 0 0 0 1-1v-5a1 1 0 0 0-1-1"},null,-1)]))}const qa={render:Xa},Va={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ha(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Va,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M20 14h-5a1 1 0 0 0-1 1v5a1 1 0 0 0 1 1h5a1 1 0 0 0 1-1v-5a1 1 0 0 0-1-1M9 14H4a1 1 0 0 0-1 1v5a1 1 0 0 0 1 1h5a1 1 0 0 0 1-1v-5a1 1 0 0 0-1-1m5-11H9a1 1 0 0 0-1 1v5a1 1 0 0 0 1 1h5a1 1 0 0 0 1-1V4a1 1 0 0 0-1-1"},null,-1)]))}const Za={render:Ha},Ga={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function Ka(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Ga,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m3.3 7 8.7 5m0 0 8.7-5M12 12v10m9-14a2 2 0 0 0-1-1.73l-7-4a2 2 0 0 0-2 0l-7 4A2 2 0 0 0 3 8v8a2 2 0 0 0 1 1.73l7 4a2 2 0 0 0 2 0l7-4A2 2 0 0 0 21 16z"},null,-1)]))}const Ya={render:Ka},Ja={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function eo(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",Ja,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"m7 14 2.5-2.5L7 9m5 7.5h5M5 4h14c1.105 0 2 .796 2 1.778v12.444c0 .982-.895 1.778-2 1.778H5c-1.105 0-2-.796-2-1.778V5.778C3 4.796 3.895 4 5 4"},null,-1)]))}const to={render:eo},no={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ro(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",no,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round","stroke-width":"1.347",d:"M2.02 3.015h1.796l2.39 11.156a1.8 1.8 0 0 0 1.796 1.42h8.784a1.796 1.796 0 0 0 1.752-1.411l1.482-6.674H4.777"},null,-1),(0,Me.Lk)("path",{fill:"currentColor","fill-rule":"evenodd",d:"M7 20a1 1 0 1 1 2 0 1 1 0 0 1-2 0m9 0a1 1 0 1 1 2 0 1 1 0 0 1-2 0","clip-rule":"evenodd"},null,-1)]))}const ao={render:ro},oo={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function io(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",oo,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M12 8v13m0-13c-.362-1.49-.985-2.765-1.787-3.657C9.41 3.451 8.465 2.983 7.5 3a2.5 2.5 0 1 0 0 5M12 8c.362-1.49.985-2.765 1.787-3.657.803-.892 1.748-1.36 2.713-1.343a2.5 2.5 0 0 1 0 5m2.5 4v7a2 2 0 0 1-2 2H7a2 2 0 0 1-2-2v-7M4 8h16a1 1 0 0 1 1 1v2a1 1 0 0 1-1 1H4a1 1 0 0 1-1-1V9a1 1 0 0 1 1-1"},null,-1)]))}const so={render:io},lo={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function co(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",lo,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M3 9h18M3 9v10a2 2 0 0 0 2 2h14a2 2 0 0 0 2-2V9M3 9l2.45-4.9A2 2 0 0 1 7.24 3h9.52a2 2 0 0 1 1.8 1.1L21 9m-9-6v6"},null,-1)]))}const uo={render:co},po={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function ho(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",po,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M4.929 19.071a10 10 0 0 0 14.142 0m0-14.142a10 10 0 0 0-14.142 0M4 15.5v-7l2.8 7v-7m5.6 7H11a1.4 1.4 0 0 1-1.4-1.4V9.9A1.4 1.4 0 0 1 11 8.5h1.4m-2.8 4.2h2.8m2.8-4.2v1.883a8.4 8.4 0 0 0 .601 3.119L16.6 15.5l1.4-4.9 1.4 4.9.8-1.998a8.4 8.4 0 0 0 .6-3.12V8.5"},null,-1)]))}const fo={render:ho},mo={xmlns:"http://www.w3.org/2000/svg",fill:"none",viewBox:"0 0 24 24"};function go(e,t){return(0,Me.uX)(),(0,Me.CE)("svg",mo,t[0]||(t[0]=[(0,Me.Lk)("path",{stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round",d:"M7 20.5 8 14h1.08a1.48 1.48 0 0 0 1.18-.58 1.44 1.44 0 0 0 .27-1.28l-.75-3A1.49 1.49 0 0 0 8.33 8H3m16-3h-3.753a1.49 1.49 0 0 0-1.45 1.14l-.75 3a1.44 1.44 0 0 0 .28 1.28 1.48 1.48 0 0 0 1.18.58h1.58l.78 4.75a1.5 1.5 0 0 0 1.46 1.25h2M2 12a10 10 0 1 0 20 0 10 10 0 0 0-20 0"},null,-1)]))}const vo={render:go},bo=function(e){const t=Object.entries(e).map((([e,t])=>{const n=e.replace("./icons/","").replace(".svg","");return{icon:{src:n,title:n.replaceAll("-"," "),tags:[]},rawData:t}}));return{iconDefinitionList:t.map((e=>e.icon)),iconDataMap:Object.fromEntries(t.map((e=>[e.icon.src,e.rawData])))}}(Object.assign({"./icons/basic-shape-diamond.svg":r,"./icons/basic-shape-hexagon.svg":a,"./icons/basic-shape-primary-circle-ellipse-round.svg":o,"./icons/basic-shape-primary-square-rectangle.svg":i,"./icons/basic-shape-shield.svg":s,"./icons/computer-device-desktop-monitor.svg":l,"./icons/computer-device-desktop.svg":c,"./icons/computer-device-laptop.svg":u,"./icons/computer-device-mobile-phone-android-samsung-back.svg":d,"./icons/computer-device-mobile-phone-android-samsung.svg":p,"./icons/computer-device-mobile-phone-iphone-x-back.svg":h,"./icons/computer-device-mobile-phone-iphone-x.svg":f,"./icons/computer-device-mobile-tablet-touch.svg":m,"./icons/computer-device-mobile-tablet.svg":g,"./icons/computer-device-network-ethernet-cat6.svg":v,"./icons/computer-device-network-lan-www.svg":b,"./icons/computer-device-network-wifi-connection.svg":y,"./icons/computer-device-network-wifi-router.svg":w,"./icons/ecology-science-erlenmeyer-flask.svg":O,"./icons/image-flash-lightning.svg":k,"./icons/image-picture-flower.svg":x,"./icons/interface-alert-exclamation-diamond.svg":E,"./icons/interface-alert-exclamation-triangle-warning.svg":S,"./icons/interface-alert-information-circle.svg":_,"./icons/interface-award-crown.svg":C,"./icons/interface-bookmark-tag.svg":T,"./icons/interface-bookmark.svg":A,"./icons/interface-calendar-date-one.svg":R,"./icons/interface-content-book-open-pages.svg":P,"./icons/interface-content-book-page.svg":D,"./icons/interface-content-file.svg":M,"./icons/interface-content-folder.svg":$,"./icons/interface-copy-clipboard.svg":N,"./icons/interface-edit-attachment.svg":I,"./icons/interface-edit-binocular.svg":L,"./icons/interface-edit-magic-wand.svg":B,"./icons/interface-edit-tool-paint-roller.svg":F,"./icons/interface-edit-tool-pencil.svg":Q,"./icons/interface-favorite-award.svg":U,"./icons/interface-favorite-flag.svg":z,"./icons/interface-favorite-heart.svg":j,"./icons/interface-favorite-star.svg":W,"./icons/interface-favorite-stars-sparkles.svg":X,"./icons/interface-hierarchy-flowchart.svg":q,"./icons/interface-home-house.svg":V,"./icons/interface-hyperlink.svg":H,"./icons/interface-lighting-brightness.svg":Z,"./icons/interface-lock-closed.svg":G,"./icons/interface-lock-open-unlock.svg":K,"./icons/interface-login-key.svg":Y,"./icons/interface-search.svg":J,"./icons/interface-setting-cog.svg":ee,"./icons/interface-share-megaphone-bullhorn.svg":te,"./icons/interface-share-rocket.svg":ne,"./icons/interface-share-satellite.svg":re,"./icons/interface-share-space-ship.svg":ae,"./icons/interface-share.svg":oe,"./icons/interface-signal-square.svg":ie,"./icons/interface-time-clock-circle.svg":se,"./icons/interface-time-hour-glass.svg":le,"./icons/interface-users-multiple.svg":ce,"./icons/interface-weather-moon.svg":ue,"./icons/mail-chat-bubble-square.svg":de,"./icons/mail-send-email-paper-airplane.svg":pe,"./icons/mail-send-envelope.svg":he,"./icons/money-cashier-receipt.svg":fe,"./icons/money-currency-dollar-pay.svg":me,"./icons/money-graph-arrow-increase.svg":ge,"./icons/money-graph-bar-chart-increase.svg":ve,"./icons/nature-ecology-leaf.svg":be,"./icons/phone-telephone.svg":ye,"./icons/programming-bug.svg":we,"./icons/programming-cloud.svg":Oe,"./icons/programming-computer-database-server.svg":ke,"./icons/programming-computer-database.svg":xe,"./icons/programming-module-four-layout.svg":Ee,"./icons/programming-module-three.svg":Se,"./icons/programming-module.svg":_e,"./icons/programming-script-code.svg":Ce,"./icons/shopping-cart.svg":Te,"./icons/shopping-gift-present.svg":Ae,"./icons/shopping-shipping-box-parcel-package.svg":Re,"./icons/tag-new-circle.svg":Pe,"./icons/travel-map-earth-globe.svg":De})),yo=bo.iconDefinitionList,wo=bo.iconDataMap,Oo=(0,Me.pM)({__name:"LibraryIcon",props:{src:{}},setup(e){const t=e,n=(0,Me.EW)((()=>(e=>wo[e])(t.src)));return(e,t)=>n.value?((0,Me.uX)(),(0,Me.Wv)((0,Me.$y)(n.value),{key:0})):(0,Me.Q3)("",!0)}})},2741:(e,t,n)=>{"use strict";n.d(t,{Q:()=>r});const r={PROTOCOL:/^(?:https?|ftp|file|mailto|tel|data|wss?)*:\/\//,MULTIPLE_SLASHES:/(?<!:)\/{2,}/g,VARIABLES:/{{((?:[^{}]|{[^{}]*})*)}}/g,PATH:/(?:{)([^{}]+)}(?!})/g,TEMPLATE_VARIABLE:/{{\s*([^}\s]+?)\s*}}|{\s*([^}\s]+?)\s*}|:\b[\w.]+\b/g}},2816:(e,t,n)=>{"use strict";n.d(t,{z:()=>a});var r=n(4048);const a=(e,t=[],n={})=>{const a=new Set(e.map((e=>Object.keys(e).sort().join(","))));return t.reduce(((e,t)=>{const o=Array.isArray(t)?t:[t],i=o.map((e=>{var t;return null==(t=n[e])?void 0:t.nameKey})).sort().join(",");return a.has(i)&&e.push(...o.map((e=>n[e])).filter(r.O9)),e}),[])}},2849:e=>{"use strict";var t=Object.prototype.hasOwnProperty,n=Object.prototype.toString,r=Object.defineProperty,a=Object.getOwnPropertyDescriptor,o=function(e){return"function"==typeof Array.isArray?Array.isArray(e):"[object Array]"===n.call(e)},i=function(e){if(!e||"[object Object]"!==n.call(e))return!1;var r,a=t.call(e,"constructor"),o=e.constructor&&e.constructor.prototype&&t.call(e.constructor.prototype,"isPrototypeOf");if(e.constructor&&!a&&!o)return!1;for(r in e);return void 0===r||t.call(e,r)},s=function(e,t){r&&"__proto__"===t.name?r(e,t.name,{enumerable:!0,configurable:!0,value:t.newValue,writable:!0}):e[t.name]=t.newValue},l=function(e,n){if("__proto__"===n){if(!t.call(e,n))return;if(a)return a(e,n).value}return e[n]};e.exports=function e(){var t,n,r,a,c,u,d=arguments[0],p=1,h=arguments.length,f=!1;for("boolean"==typeof d&&(f=d,d=arguments[1]||{},p=2),(null==d||"object"!=typeof d&&"function"!=typeof d)&&(d={});p<h;++p)if(null!=(t=arguments[p]))for(n in t)r=l(d,n),d!==(a=l(t,n))&&(f&&a&&(i(a)||(c=o(a)))?(c?(c=!1,u=r&&o(r)?r:[]):u=r&&i(r)?r:{},s(d,{name:n,newValue:e(f,u,a)})):void 0!==a&&s(d,{name:n,newValue:a}));return d}},2966:(e,t,n)=>{"use strict";function r(){return/iPhone/gi.test(window.navigator.platform)||/Mac/gi.test(window.navigator.platform)&&window.navigator.maxTouchPoints>0}function a(){return r()||/Android/gi.test(window.navigator.userAgent)}n.d(t,{Fr:()=>a,un:()=>r})},2975:(e,t,n)=>{"use strict";n.r(t),n.d(t,{default:()=>sr});var r=n(641),a=n(953),o=n(33),i=n(4048);const s={parse(e){try{return{error:!1,data:JSON.parse(e)}}catch(e){return{error:!0,message:e.message?String(e.message):"Unknown Error"}}}};var l=n(4325),c=n(8157),u=n(9351),d=n(1636),p=n(3751),h=n(180),f=n(9135),m=n(1266);const g={class:"border flex flex-row gap-2 h-9 items-center mb-4 mt-2 p-[3px] rounded text-sm"},v={class:"flex h-full"},b={class:"scroll-timeline-x whitespace-nowrap"},y={class:"flex gap-10 justify-between"},w=(0,r.pM)({__name:"ImportCurlModal",props:{parsedCurl:{},state:{},collectionUid:{}},emits:["importCurl","close"],setup(e,{emit:t}){const n=e,i=t,{toast:s}=(0,c.Yf)(),l=(0,a.KR)(""),{activeWorkspaceCollections:u,activeCollection:d}=(0,f.JC)(),w=(0,r.EW)((()=>!l.value.trim())),O=(0,r.EW)((()=>u.value.map((e=>{var t;return{id:e.uid,label:(null==(t=e.info)?void 0:t.title)??"Unititled Collection"}})))),k=(0,a.KR)(n.collectionUid?O.value.find((e=>e.id===n.collectionUid)):O.value.find((e=>{var t;return e.id===(null==(t=d.value)?void 0:t.uid)})));function x(){w.value?s("Please enter a name before importing your request.","error"):function(){var e;n.parsedCurl&&i("importCurl",{parsedCurl:n.parsedCurl,requestName:l.value,collectionUid:(null==(e=k.value)?void 0:e.id)??""})}()}return(0,r.wB)((()=>n.state.open),(e=>{e&&(l.value="")})),(e,t)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.uT),{size:"xs",state:e.state,title:"Import cURL request"},{default:(0,r.k6)((()=>{var e,i,s,c;return[(0,r.bF)((0,a.R1)(m.KB),{modelValue:l.value,"onUpdate:modelValue":t[0]||(t[0]=e=>l.value=e),autofocus:"",class:"h-9",placeholder:"Request Name",onKeyup:(0,p.jR)(x,["enter"])},null,8,["modelValue"]),(0,r.Lk)("div",g,[(0,r.Lk)("div",v,[(0,r.bF)(h.A,{isEditable:!1,isSquare:"",method:(null==(e=n.parsedCurl)?void 0:e.method)||"get"},null,8,["method"])]),(0,r.Lk)("span",b,(0,o.v_)((null==(s=null==(i=n.parsedCurl)?void 0:i.servers)?void 0:s[0])||"")+(0,o.v_)((null==(c=n.parsedCurl)?void 0:c.path)||""),1)]),(0,r.Lk)("div",y,[(0,r.bF)((0,a.R1)(m.cy),{modelValue:k.value,"onUpdate:modelValue":t[1]||(t[1]=e=>k.value=e),options:O.value},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.RO),{class:"justify-between p-2 max-h-8 gap-1 text-xs hover:bg-b-2",variant:"outlined"},{default:(0,r.k6)((()=>[(0,r.Lk)("span",{class:(0,o.C4)(["whitespace-nowrap",k.value?"text-c-1":"text-c-3"])},(0,o.v_)(k.value?k.value.label:"Select Collection"),3),(0,r.bF)((0,a.R1)(m.kv),{class:"text-c-3",icon:"ChevronDown",size:"md"})])),_:1})])),_:1},8,["modelValue","options"]),(0,r.bF)((0,a.R1)(m.RO),{class:"gap-1.5 font-medium px-2.5 h-8 shadow-none focus:outline-none whitespace-nowrap",disabled:w.value,type:"submit",onClick:x},{default:(0,r.k6)((()=>t[2]||(t[2]=[(0,r.eW)(" Import Request ")]))),_:1},8,["disabled"])])]})),_:1},8,["state"]))}});var O=n(3526);const k=(0,O.A)(w,[["__scopeId","data-v-b77d5bd3"]]);var x=n(8082),E=n(9005),S=n(5192);function _(e){const t={url:""},n=(0,S.parse)(e).map((e=>"object"==typeof e&&"op"in e&&"glob"===e.op?e.pattern.trim():"string"==typeof e?e.trim():e)).filter((e=>""!==e))[Symbol.iterator]();let r=n.next().value;for(;r;)"-X"===r||"--request"===r?C(n,t):"--url"===r?T(n,t):"-H"===r||"--header"===r?A(n,t):"--data"===r||"-d"===r||"--data-raw"===r||"--data-urlencode"===r||"--data-binary"===r||"--data-ascii"===r?N(n,t,e):"string"!=typeof r||t.url||!r.startsWith("http")&&!r.startsWith("www.")?"-P"===r?R(n,t):"string"==typeof r&&r.toLowerCase().includes("content-type")?D(r,t):"-u"===r||"--user"===r?M(n,t):("-b"===r||"--cookie"===r)&&$(n,t):T([r][Symbol.iterator](),t),r=n.next().value;return t}function C(e,t){t.method=e.next().value.toLowerCase()}function T(e,t){const n=new URL(e.next().value.replace(/['"]/g,""));t.servers=[n.origin],t.path="/"!==n.pathname?n.pathname:"",t.url=t.servers[0]+t.path;const r=P(n.search);t.queryParameters=t.queryParameters?[...t.queryParameters,...r]:r}function A(e,t){const n=e.next().value.split(/:(.*)/);t.headers=t.headers||{},void 0!==n[1]?t.headers[n[0].trim()]=n[1].trim():t.headers[n[0].trim()]=""}function R(e,t){const n=e.next().value.replace(/['"]/g,"").split("=");t.pathVariables=t.pathVariables||{},void 0!==n[1]?t.pathVariables[n[0].trim()]=n[1].trim():t.pathVariables[n[0].trim()]=""}function P(e){const t=[];return new URL(e,"http://example.com").searchParams.forEach(((e,n)=>{t.push({key:n,value:e})})),t}function D(e,t){const n=e.replace(/['"]/g,"").split(/:(.+)/);t.headers=t.headers||{},n[0]&&(void 0!==n[1]?t.headers[n[0].trim()]=n[1].trim():t.headers[n[0].trim()]="")}function M(e,t){const n=e.next().value;try{const e=btoa(n);t.headers=t.headers||{},t.headers.Authorization=`Basic ${e}`}catch(e){console.warn("Could not base64 encode these HTTP basic auth credentials:",n,e)}}function $(e,t){const n=e.next().value;t.headers=t.headers||{},t.headers.Cookie?t.headers.Cookie+=`; ${n}`:t.headers.Cookie=n.replace(/;$/,"")}function N(e,t,n){const r=e.next().value;if("string"==typeof r&&(r.startsWith("@")?t.body="":t.body=r,!t.url||n.includes("-G"))){const e=P(`?${t.body}`);t.queryParameters=t.queryParameters?[...t.queryParameters,...e]:e}}var I=n(2490),L=n(5191),B=n(1347),F=n(796);const Q={Date:!0,RegExp:!0,String:!0,Number:!0};function U(e,t,n={cyclesFix:!0},r=[]){let a=[];const o=Array.isArray(e);for(const i in e){const s=e[i],l=o?+i:i;if(!(i in t)){a.push({type:"REMOVE",path:[l],oldValue:e[i]});continue}const c=t[i],u="object"==typeof s&&"object"==typeof c&&Array.isArray(s)===Array.isArray(c);!(s&&c&&u)||Q[Object.getPrototypeOf(s)?.constructor?.name]||n.cyclesFix&&r.includes(s)?s===c||Number.isNaN(s)&&Number.isNaN(c)||u&&(isNaN(s)?s+""==c+"":+s==+c)||a.push({path:[l],type:"CHANGE",value:c,oldValue:s}):a.push.apply(a,U(s,c,n,n.cyclesFix?r.concat([s]):[]).map((e=>(e.path.unshift(l),e))))}const i=Array.isArray(t);for(const n in t)n in e||a.push({type:"CREATE",path:[i?+n:n],value:t[n]});return a}var z=n(4476);const j=(e,t=[])=>{const n=[];let r=!1;for(let a=0;a<e.length;a++){if(r){r=!1;continue}const o=e[a],i=e[a+1];if(o){if(t.length)o.path=[...t,...o.path],i&&(i.path=[...t,...i.path]);else if("paths"!==o.path[0]){n.push(o);continue}if("REMOVE"===o.type&&"CREATE"===(null==i?void 0:i.type)){const[,e,a]=o.path,[,s,l]=i.path,c=["paths",s].filter((e=>"string"==typeof e));if(e!==s&&n.push({type:"CHANGE",path:["paths","path"],oldValue:e,value:s}),a&&"string"==typeof l&&a!==l&&s&&(n.push({type:"CHANGE",path:["paths",s,"method"],oldValue:a,value:l}),c.push(l)),0===t.length){const e=U(o.oldValue,i.value);if(e.length){const t=j(e,c);n.push(...t)}}r=!0}else"CREATE"===o.type&&o.path.length>3&&"number"!=typeof o.path.at(-1)?n.push({...o,type:"CHANGE",oldValue:void 0}):"REMOVE"===o.type&&o.path.length>3&&"number"!=typeof o.path.at(-1)?n.push({...o,type:"CHANGE",value:void 0}):n.push(o)}}return n},W=(e,t,n)=>{for(const r of e){const e=t[r];if(e&&n(e))return e}return null},X=e=>e instanceof z.z.ZodOptional?X(e.unwrap()):e instanceof z.z.ZodDefault?X(e._def.innerType):e instanceof z.z.ZodEffects?X(e._def.schema):e,q=(e,t)=>{const n=((e,t)=>{let n=e;for(const e of t){if(n=X(n),n instanceof z.z.ZodAny)return n;if(n instanceof z.z.ZodObject&&"string"==typeof e&&e in n.shape)n=n.shape[e];else if(n instanceof z.z.ZodArray)if("number"==typeof e)n=n.element;else{if("string"!=typeof e)return null;if(n=n.element,!(n instanceof z.z.ZodObject&&e in n.shape))return null;n=n.shape[e]}else{if(!(n instanceof z.z.ZodRecord))return null;n=n.valueSchema}n=X(n)}return n})(e,t.path);if(!n)return null;const r=t.path.join("."),a=t.path.slice(0,-1).join(".");if("REMOVE"===t.type)return{path:r,pathMinusOne:a,value:void 0};const o=(0,i.TA)(t.value,n,!1);return o?{path:r,pathMinusOne:a,value:o}:null},V=(e,t)=>{const{requests:n,requestExamples:r,requestExampleMutators:a}=t,o=n[e];null==o||o.examples.forEach((e=>{var t;const n=(0,B.oE)(o,(null==(t=r[e])?void 0:t.name)??"Default");n&&a.set({...n,uid:e})}))};var H=n(4748),Z=n(3073),G=n(8309);const K=()=>{const{toast:e}=(0,c.Yf)(),t=(0,f.JC)(),n=(0,G.hA)(),{activeCollection:a,activeWorkspace:o}=t,{collectionMutators:s}=n,l=t=>e(`[useOpenApiWatcher] Changes to the ${t} were not applied`,"error"),u=e=>{"info"===e.path[0]||"security"===e.path[0]?((e,{activeCollection:t},{collectionMutators:n})=>{if(!t.value)return!1;if("number"!=typeof e.path[e.path.length-1]||"CREATE"!==e.type&&"REMOVE"!==e.type){const r=q(B.CV,e);if(!r)return!1;n.edit(t.value.uid,r.path,r.value)}else{const r=q(B.CV,{...e,path:e.path});if(!r)return!1;const a=[...(0,F.L)(t.value,r.pathMinusOne)];"CREATE"===e.type?a.push(r.value):"REMOVE"===e.type&&a.pop(),n.edit(t.value.uid,r.pathMinusOne,a)}return!0})(e,t,n)||l("collection"):"components"===e.path[0]&&"securitySchemes"===e.path[1]?((e,{activeCollection:t},{securitySchemes:n,securitySchemeMutators:r})=>{if(!t.value)return!1;const[,,a,...o]=e.path,i=n[a]??W(t.value.securitySchemes,n,(e=>e.nameKey===a));if(null!=o&&o.length){const t=((e,t,n)=>{const r=X(e);if(r instanceof z.z.ZodUnion||r instanceof z.z.ZodDiscriminatedUnion)for(const e of r.options)if(e instanceof z.z.ZodObject&&t in e.shape&&e.shape[t]instanceof z.z.ZodLiteral&&e.shape[t].value===n)return e;return null})(B.NV,"type",(null==i?void 0:i.type)??"");if(!t||!i)return!1;const n=q(t,{...e,path:o});if(!n)return!1;const a=n.path;r.edit(i.uid,a,n.value)}else if("REMOVE"===e.type){if(!i)return!1;r.delete(i.uid)}else"CREATE"===e.type&&r.add(B.NV.parse(e.value),t.value.uid);return!0})(e,t,n)||l("securitySchemes"):"servers"===e.path[0]?((e,{activeCollection:t},{servers:n,serverMutators:r})=>{if(!t.value)return!1;const[,a,...o]=e.path;if(null!=o&&o.length){const i=t.value.servers[a];if(!i)return!1;const s=n[i],l=q(B.zT,{...e,path:o});if(!s||!l)return!1;const c="REMOVE"===e.type&&"variables"===o[o.length-1]?{}:l.value;r.edit(i,l.path,c)}else if("REMOVE"===e.type){if(!t.value.servers[a])return!1;r.delete(t.value.servers[a],t.value.uid)}else if("CREATE"===e.type){const n=(0,i.TA)(e.value,B.zT,!1);if(!n)return!1;r.add(n,t.value.uid)}return!0})(e,t,n)||l("servers"):"tags"===e.path[0]?((e,{activeCollection:t},{tags:n,tagMutators:r})=>{if(!t.value)return!1;const[,a,...o]=e.path;if(null!=o&&o.length){const i=t.value.tags[a];if(!i)return!1;const s=n[i],l=q(B.c3,{...e,path:o});if(!s||!l)return!1;r.edit(i,l.path,l.value)}else if("REMOVE"===e.type){const e=t.value.tags[a];if(!e)return!1;const o=n[e];if(!o)return!1;r.delete(o,t.value.uid)}else if("CREATE"===e.type){const n=(0,i.TA)(e.value,B.c3,!1);if(!n)return!1;r.add(n,t.value.uid)}return!0})(e,t,n)||l("tags"):"paths"===e.path[0]&&(((e,{activeCollection:t},n)=>{if(!t.value)return!1;const{requests:r,requestMutators:a}=n,[,o,s,...l]=e.path;if("path"===o&&"CHANGE"===e.type)t.value.requests.forEach((t=>{var n;(null==(n=r[t])?void 0:n.path)===e.oldValue&&a.edit(t,"path",e.value)}));else if("method"===s&&"CHANGE"===e.type)t.value.requests.forEach((t=>{var n,i;(null==(n=r[t])?void 0:n.method)===e.oldValue&&(null==(i=r[t])?void 0:i.path)===o&&a.edit(t,"method",e.value)}));else if("CHANGE"!==e.type&&"number"==typeof l.at(-1)){const i=W(t.value.requests,r,(e=>e.path===o&&e.method===s)),l=q(B.TF,{...e,path:e.path.slice(3)});if(!i||!l)return!1;const c=[...(0,F.L)(i,l.pathMinusOne)];"CREATE"===e.type?c.push(l.value):"REMOVE"===e.type&&c.pop(),a.edit(i.uid,l.pathMinusOne,c),("parameters"===e.path[3]||"requestBody"===e.path[3])&&V(i.uid,n)}else if("CREATE"===e.type){const[n]=Object.entries(e.value??{}),[r,l]=n??[],c=s?e.value:l,u=s||r,d=B.zT.array().parse(c.servers??[]),{security:p,...h}=c,f={...h,method:(0,i.aO)(u)?u:"get",path:o,parameters:c.parameters??[],servers:d.map((e=>e.uid))};null!=p&&p.length&&(f.security=p.map((e=>{if(Object.keys(e).length){const[t]=Object.keys(e);return t?{[t]:e[t]}:e}return e})));const m=(0,i.TA)(f,B.TF,!1);if(!m)return!1;a.add(m,t.value.uid)}else if("REMOVE"===e.type){const e=W(t.value.requests,r,(e=>e.path===o&&e.method===s));if(!e)return!1;a.delete(e,t.value.uid)}else if("CHANGE"===e.type){const i=W(t.value.requests,r,(e=>e.path===o&&e.method===s)),c=q(B.TF,{...e,path:l});if(!i||!c)return!1;a.edit(i.uid,c.path,c.value),("parameters"===e.path[3]||"requestBody"===e.path[3])&&V(i.uid,n)}return!0})(e,t,n)||l("requests"))},{pause:d,resume:p}=(0,Z.Sj0)((async()=>{var t,n;const r=null==(t=a.value)?void 0:t.documentUrl;if(!r)return;const l=L.i[r];try{const e=await(0,i.tI)(r,null==(n=o.value)?void 0:n.proxyUrl,!1),t=(0,i.n1)(e);if(s.edit(a.value.uid,"watchModeStatus","WATCHING"),null!=l&&l.hash)if(l.hash&&l.hash!==t){const{schema:n}=await(0,H.D0)(e),a=U(l.schema,n),o=j(a);try{o.forEach(u),L.i[r]={hash:t,schema:n}}catch(e){console.error("[useOpenApiWatcher] Error:",e)}}else console.log("[useOpenApiWatcher] No changes detected yet…");else{const{schema:n}=await(0,H.D0)(e);n&&(L.i[r]={hash:t,schema:n})}}catch(t){console.error("[useOpenApiWatcher] Error:",t),console.info("[useOpenApiWatcher] Pausing watcher for 60 seconds"),d(),s.edit(a.value.uid,"watchModeStatus","ERROR"),e("[useOpenApiWatcher] Unable to fetch the spec file, paused the watcher for 60 seconds","error"),setTimeout((()=>{console.info("[useOpenApiWatcher] Resuming watcher"),p()}),6e4)}}),5e3);(0,r.wB)([()=>{var e;return null==(e=a.value)?void 0:e.documentUrl},()=>{var e;return null==(e=a.value)?void 0:e.watchMode}],(([e,t])=>{e&&t?(console.info(`[useOpenApiWatcher] Watching ${e} …`),p()):a.value&&(d(),s.edit(a.value.uid,"watchModeStatus","IDLE"))}),{immediate:!0})};var Y=n(7525),J=n(5336);const ee={class:"sr-only"},te={class:"size-4",fill:"none",viewBox:"0 0 24 24",xmlns:"http://www.w3.org/2000/svg"},ne={"clip-path":"url(#mask)"},re=(0,r.pM)({__name:"SidebarToggle",props:{modelValue:{type:Boolean}},emits:["update:modelValue"],setup:e=>(e,t)=>((0,r.uX)(),(0,r.CE)("button",{class:"scalar-sidebar-toggle text-c-3 hover:bg-b-2 active:text-c-1 p-2 rounded-lg",type:"button",onClick:t[0]||(t[0]=t=>e.$emit("update:modelValue",!e.modelValue))},[(0,r.Lk)("span",ee,(0,o.v_)(e.modelValue?"Hide":"Show")+" sidebar",1),((0,r.uX)(),(0,r.CE)("svg",te,[t[1]||(t[1]=(0,r.Lk)("defs",null,[(0,r.Lk)("clipPath",{id:"mask"},[(0,r.Lk)("path",{"clip-rule":"evenodd",d:"M9 3.2H4c-1.7 0-3 1.3-3 3v11.5c0 1.7 1.3 3 3 3h5V3.2z"})])],-1)),(0,r.Lk)("g",ne,[(0,r.Lk)("path",{class:(0,o.C4)(["transition-transform duration-300",e.modelValue?"translate-x-0":"-translate-x-1/2"]),d:"M1 3.2h8v17.5H1z",fill:"currentColor"},null,2)]),t[2]||(t[2]=(0,r.Lk)("path",{d:"M20 20.8H4c-1.7 0-3-1.3-3-3V6.2c0-1.7 1.3-3 3-3h16c1.7 0 3 1.3 3 3v11.5c0 1.7-1.3 3-3 3zM9 3.2v17.5",stroke:"currentColor","stroke-linecap":"round","stroke-linejoin":"round","stroke-width":"2"},null,-1))]))]))});var ae=n(8207),oe=n(6584);const ie={class:"lg:min-h-client-header flex items-center w-full justify-center p-2 pt-2 lg:pt-1 lg:p-1 flex-wrap t-app__top-container border-b-1/2"},se={class:"flex flex-row items-center gap-1 lg:px-1 lg:mb-0 mb-2 lg:flex-1 w-1/2"},le={class:"flex flex-row items-center gap-1 lg:px-2.5 lg:mb-0 mb-2 lg:flex-1 justify-end w-1/2"},ce=(0,r.pM)({__name:"RequestSubpageHeader",props:{collection:{},operation:{},server:{},environment:{},envVariables:{},workspace:{},modelValue:{type:Boolean}},emits:["update:modelValue","hideModal","importCurl"],setup(e){const{hideClientButton:t,showSidebar:n,integration:i}=(0,G.hA)(),{layout:s}=(0,oe.g)(),{currentRoute:l}=(0,u.rd)();return(e,c)=>((0,r.uX)(),(0,r.CE)("div",ie,[(0,r.Lk)("div",se,[(0,a.R1)(n)?((0,r.uX)(),(0,r.Wv)(re,{key:0,class:(0,o.C4)(["ml-1",[{hidden:e.modelValue},{"xl:!flex":!e.modelValue},{"!flex":"modal"===(0,a.R1)(s)},{"!hidden":"modal"===(0,a.R1)(s)&&e.modelValue}]]),modelValue:e.modelValue,"onUpdate:modelValue":c[0]||(c[0]=t=>e.$emit("update:modelValue",t))},null,8,["class","modelValue"])):(0,r.Q3)("",!0)]),(0,r.bF)(J.A,{collection:e.collection,envVariables:e.envVariables,environment:e.environment,operation:e.operation,server:e.server,workspace:e.workspace,onImportCurl:c[1]||(c[1]=t=>e.$emit("importCurl",t))},null,8,["collection","envVariables","environment","operation","server","workspace"]),(0,r.Lk)("div",le,["modal"===(0,a.R1)(s)&&e.collection.documentUrl&&!(0,a.R1)(t)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(ae.A),{key:0,buttonSource:"modal",class:"!w-fit lg:-mr-1",integration:(0,a.R1)(i)??e.collection.integration??null,source:"gitbook"===(0,a.R1)(l).query.source?"gitbook":"api-reference",url:e.collection.documentUrl},null,8,["integration","source","url"])):(0,r.Q3)("",!0),"modal"===(0,a.R1)(s)?((0,r.uX)(),(0,r.CE)("button",{key:1,class:"app-exit-button p-2 rounded-full fixed right-2 top-2 gitbook-hidden",type:"button",onClick:c[2]||(c[2]=t=>e.$emit("hideModal"))},[(0,r.bF)((0,a.R1)(m.kv),{icon:"Close",size:"lg",thickness:"2"}),c[4]||(c[4]=(0,r.Lk)("span",{class:"sr-only"},"Close Client",-1))])):(0,r.Q3)("",!0),"modal"===(0,a.R1)(s)?((0,r.uX)(),(0,r.CE)("button",{key:2,class:"text-c-1 hover:bg-b-2 active:text-c-1 p-2 rounded -mr-1.5 gitbook-show",type:"button",onClick:c[3]||(c[3]=t=>e.$emit("hideModal"))},[(0,r.bF)((0,a.R1)(m.kv),{icon:"Close",size:"md",thickness:"1.75"}),c[5]||(c[5]=(0,r.Lk)("span",{class:"sr-only"},"Close Client",-1))])):(0,r.Q3)("",!0)])]))}}),ue=(0,O.A)(ce,[["__scopeId","data-v-cac03082"]]);var de=n(7186),pe=n(3395),he=n(25),fe=n(264);const me={"application/epub+zip":{extension:".epub"},"application/gzip":{extension:".gz"},"application/java-archive":{extension:".jar"},"application/javascript":{extension:".js",raw:!0},"application/json":{extension:".json",raw:!0,language:"json"},"application/ld+json":{extension:".jsonld",raw:!0,language:"json"},"application/problem+json":{extension:".json",raw:!0,language:"json"},"application/msword":{extension:".doc"},"application/octet-stream":{extension:".bin"},"application/ogg":{extension:".ogx"},"application/pdf":{extension:".pdf",preview:"object"},"application/rtf":{extension:".rtf",raw:!0},"application/vnd.amazon.ebook":{extension:".azw"},"application/vnd.apple.installer+xml":{extension:".mpkg",raw:!0,language:"xml"},"application/vnd.mozilla.xul+xml":{extension:".xul",raw:!0,language:"xml"},"application/vnd.ms-excel":{extension:".xls"},"application/vnd.ms-fontobject":{extension:".eot"},"application/vnd.ms-powerpoint":{extension:".ppt"},"application/vnd.oasis.opendocument.presentation":{extension:".odp"},"application/vnd.oasis.opendocument.spreadsheet":{extension:".ods"},"application/vnd.oasis.opendocument.text":{extension:".odt"},"application/vnd.openxmlformats-officedocument.presentationml.presentation":{extension:".pptx"},"application/vnd.openxmlformats-officedocument.spreadsheetml.sheet":{extension:".xlsx"},"application/vnd.openxmlformats-officedocument.wordprocessingml.document":{extension:".docx"},"application/vnd.rar":{extension:".rar"},"application/vnd.visio":{extension:".vsd"},"application/x-7z-compressed":{extension:".7z"},"application/x-abiword":{extension:".abw"},"application/x-bzip":{extension:".bz"},"application/x-bzip2":{extension:".bz2"},"application/x-cdf":{extension:".cda"},"application/x-csh":{extension:".csh"},"application/x-freearc":{extension:".arc"},"application/x-httpd-php":{extension:".php",raw:!0},"application/x-sh":{extension:".sh",raw:!0},"application/x-tar":{extension:".tar"},"application/xhtml+xml":{extension:".xhtml",raw:!0,language:"html"},"application/xml":{extension:".xml",raw:!0,language:"xml"},"application/yaml":{extension:".yaml",raw:!0,language:"yaml"},"application/zip":{extension:".zip"},"audio/aac":{extension:".aac"},"audio/midi":{extension:".midi"},"audio/mpeg":{extension:".mp3",preview:"audio"},"audio/ogg":{extension:".oga"},"audio/wav":{extension:".wav"},"audio/webm":{extension:".weba"},"font/otf":{extension:".otf"},"font/ttf":{extension:".ttf"},"font/woff":{extension:".woff"},"font/woff2":{extension:".woff2"},"image/apng":{extension:".apng",preview:"image",alpha:!0},"image/avif":{extension:".avif",preview:"image"},"image/bmp":{extension:".bmp",preview:"image"},"image/gif":{extension:".gif",preview:"image",alpha:!0},"image/jpeg":{extension:".jpg",preview:"image"},"image/png":{extension:".png",preview:"image",alpha:!0},"image/svg+xml":{extension:".svg",raw:!0,language:"xml",preview:"image",alpha:!0},"image/tiff":{extension:".tiff"},"image/vnd.microsoft.icon":{extension:".ico",preview:"image"},"image/webp":{extension:".webp",preview:"image",alpha:!0},"text/calendar":{extension:".ics",raw:!0},"text/css":{extension:".css",raw:!0,language:"css"},"text/csv":{extension:".csv",raw:!0},"text/html":{extension:".html",raw:!0,language:"html",preview:"object"},"text/javascript":{extension:".js",raw:!0},"text/plain":{extension:".txt",raw:!0},"text/xml":{extension:".xml",raw:!0,language:"xml"},"text/yaml":{extension:".yaml",raw:!0,language:"yaml"},"video/3gpp":{extension:".3gp"},"audio/3gpp":{extension:".3gp"},"video/3gpp2":{extension:".3g2"},"audio/3gpp2":{extension:".3g2"},"video/mp2t":{extension:".ts"},"video/mp4":{extension:".mp4",preview:"video"},"video/mpeg":{extension:".mpeg"},"video/ogg":{extension:".ogv"},"video/webm":{extension:".webm",preview:"video"},"video/x-msvideo":{extension:".avi"}},ge=Object.entries(me).filter((([,e])=>null==e?void 0:e.raw)).map((([e])=>e)),ve=["download","href"],be=(0,r.pM)({__name:"ResponseBodyDownload",props:{href:{},type:{},filename:{}},setup(e){const t=e,n=(0,r.EW)((()=>{var e;const n=(null==(e=me[t.type??""])?void 0:e.extension)??".unknown";return t.filename?t.filename:`response${n}`}));return(e,t)=>((0,r.uX)(),(0,r.CE)("a",{class:"flex gap-1 text-c-3 text-xxs no-underline items-center hover:bg-b-3 rounded py-0.5 px-1.5",download:`${n.value}`,href:e.href,onClick:t[0]||(t[0]=(0,p.D$)((()=>{}),["stop"]))},[(0,r.bF)((0,a.R1)(m.kv),{icon:"Download",size:"xs"}),t[1]||(t[1]=(0,r.Lk)("span",null,[(0,r.Lk)("span",null,"Download"),(0,r.Lk)("span",{class:"sr-only"},"Response Body")],-1))],8,ve))}}),ye={},we={class:"flex justify-center px-2 py-3"},Oe={class:"text-c-3 p-2 text-sm"},ke=(0,O.A)(ye,[["render",function(e,t){return(0,r.uX)(),(0,r.CE)("div",we,[(0,r.Lk)("div",Oe,[(0,r.RG)(e.$slots,"default")])])}]]),xe=["src"],Ee=["src","type"],Se=["src","type"],_e=["data","type"],Ce=(0,r.pM)({__name:"ResponseBodyPreview",props:{src:{},type:{},mode:{},alpha:{type:Boolean,default:!1}},setup(e){const t=(0,a.KR)(!1);return(0,r.wB)((()=>e.src),(()=>t.value=!1)),(e,n)=>!t.value&&e.src?((0,r.uX)(),(0,r.CE)("div",{key:0,class:(0,o.C4)(["flex justify-center overflow-auto rounded-b",{"p-2 bg-preview":e.alpha}])},["image"===e.mode?((0,r.uX)(),(0,r.CE)("img",{key:0,class:(0,o.C4)(["h-full max-w-full",{rounded:e.alpha}]),src:e.src,onError:n[0]||(n[0]=e=>t.value=!0)},null,42,xe)):"video"===e.mode?((0,r.uX)(),(0,r.CE)("video",{key:1,autoplay:"",controls:"",width:"100%",onError:n[1]||(n[1]=e=>t.value=!0)},[(0,r.Lk)("source",{src:e.src,type:e.type},null,8,Ee)],32)):"audio"===e.mode?((0,r.uX)(),(0,r.CE)("audio",{key:2,class:"my-12",controls:"",onError:n[2]||(n[2]=e=>t.value=!0)},[(0,r.Lk)("source",{src:e.src,type:e.type},null,8,Se)],32)):((0,r.uX)(),(0,r.CE)("object",{key:3,class:"w-full aspect-[4/3]",data:e.src,type:e.type,onError:n[3]||(n[3]=e=>t.value=!0)},null,40,_e))],2)):((0,r.uX)(),(0,r.Wv)(ke,{key:1},{default:(0,r.k6)((()=>n[4]||(n[4]=[(0,r.eW)("Preview unavailable")]))),_:1}))}}),Te=(0,O.A)(Ce,[["__scopeId","data-v-5f138d26"]]);var Ae=n(9035),Re=n(3806);const Pe={class:"body-raw relative overflow-auto"},De={key:0,class:"scalar-code-copy"},Me=(0,r.pM)({__name:"ResponseBodyRaw",props:{content:{},language:{}},setup(e){const t=e,n=(0,a.KR)(null),{copyToClipboard:o}=(0,Re.i)(),{codeMirror:s}=(0,Ae.qF)({codeMirrorRef:n,readOnly:!0,lineNumbers:!0,content:(0,a.lW)((()=>(0,i.qX)(t.content))),language:(0,a.lW)((()=>t.language)),forceFoldGutter:!0}),l=()=>{var e;return(null==(e=s.value)?void 0:e.state.doc.toString())||""};return(e,t)=>((0,r.uX)(),(0,r.CE)("div",Pe,[l()?((0,r.uX)(),(0,r.CE)("div",De,[(0,r.Lk)("button",{class:"copy-button",type:"button",onClick:t[0]||(t[0]=e=>(0,a.R1)(o)(l()))},[t[1]||(t[1]=(0,r.Lk)("span",{class:"sr-only"},"Copy content",-1)),(0,r.bF)((0,a.R1)(m.kv),{icon:"Clipboard",size:"md"})])])):(0,r.Q3)("",!0),(0,r.Lk)("div",{ref_key:"codeMirrorRef",ref:n},null,512)]))}}),$e=(0,O.A)(Me,[["__scopeId","data-v-99361e9b"]]),Ne={class:"text-c-3 flex -my-1 justify-center rounded text-xxs p-0.5 gap-0.5"},Ie=(0,r.pM)({__name:"ResponseBodyToggle",props:{modelValue:{type:Boolean}},emits:["update:modelValue"],setup:e=>(e,t)=>((0,r.uX)(),(0,r.CE)("div",Ne,[(0,r.Lk)("button",{class:(0,o.C4)(["hover:bg-b-3 rounded px-1",{"bg-b-3 text-c-1 cursor-default":e.modelValue}]),type:"button",onClick:t[0]||(t[0]=(0,p.D$)((t=>e.$emit("update:modelValue",!0)),["stop"]))}," Preview ",2),(0,r.Lk)("button",{class:(0,o.C4)(["hover:bg-b-3 rounded px-1",{"bg-b-3 text-c-1 cursor-default":!e.modelValue}]),type:"button",onClick:t[1]||(t[1]=(0,p.D$)((t=>e.$emit("update:modelValue",!1)),["stop"]))}," Raw ",2)]))}),Le={key:0,class:"max-h-[calc(100%-32px)] border-t-1/2 flex flex-col bg-b-1 overflow-hidden"},Be={class:"flex justify-between items-center border-b-1/2 px-3 py-1.5"},Fe={class:"text-xxs leading-3 font-code"},Qe=(0,r.pM)({__name:"ResponseBody",props:{title:{},data:{},headers:{}},setup(e){const t=e,n=(0,a.KR)(!0),i=(0,r.EW)((()=>{var e;return!(null==(e=p.value)||!e.raw||!p.value.preview)})),s=(0,r.EW)((()=>n.value||!i.value)),l=(0,r.EW)((()=>!n.value||!i.value)),{mimeType:c,attachmentFilename:u,dataUrl:d}=(0,fe.x)({data:t.data,headers:t.headers}),p=(0,r.EW)((()=>me[c.value.essence]));return(e,t)=>((0,r.uX)(),(0,r.Wv)(he.A,{class:"max-h-content overflow-x-auto"},(0,r.eX)({title:(0,r.k6)((()=>[(0,r.eW)((0,o.v_)(e.title),1)])),default:(0,r.k6)((()=>{var u,h,f,m;return[e.data?((0,r.uX)(),(0,r.CE)("div",Le,[(0,r.Lk)("div",Be,[(0,r.Lk)("span",Fe,(0,o.v_)((0,a.R1)(c).essence),1),i.value?((0,r.uX)(),(0,r.Wv)(Ie,{key:0,modelValue:n.value,"onUpdate:modelValue":t[0]||(t[0]=e=>n.value=e)},null,8,["modelValue"])):(0,r.Q3)("",!0)]),null!=(u=p.value)&&u.raw&&l.value?((0,r.uX)(),(0,r.Wv)($e,{key:(0,a.R1)(d),content:e.data,language:p.value.language},null,8,["content","language"])):(0,r.Q3)("",!0),null!=(h=p.value)&&h.preview&&s.value?((0,r.uX)(),(0,r.Wv)(Te,{key:(0,a.R1)(d),alpha:p.value.alpha,mode:p.value.preview,src:(0,a.R1)(d),type:(0,a.R1)(c).essence},null,8,["alpha","mode","src","type"])):(0,r.Q3)("",!0),null!=(f=p.value)&&f.raw||null!=(m=p.value)&&m.preview?(0,r.Q3)("",!0):((0,r.uX)(),(0,r.Wv)(ke,{key:2},{default:(0,r.k6)((()=>t[1]||(t[1]=[(0,r.eW)(" Binary file ")]))),_:1}))])):(0,r.Q3)("",!0)]})),_:2},[e.data&&(0,a.R1)(d)?{name:"actions",fn:(0,r.k6)((()=>[(0,r.bF)(be,{filename:(0,a.R1)(u),href:(0,a.R1)(d),type:(0,a.R1)(c).essence},null,8,["filename","href","type"])])),key:"0"}:void 0]),1024))}}),Ue=(0,O.A)(Qe,[["__scopeId","data-v-cf80aa59"]]);var ze=n(5881),je=n(8839);const We={class:"border-t relative col-1 flex-center gap-6 p-2 capitalize"},Xe={key:0,class:"scalar-version-number"},qe={class:"text-c-3 right-4 mt-auto flex w-full flex-col items-end gap-2 text-sm"},Ve=(0,r.pM)({__name:"ResponseEmpty",props:{numWorkspaceRequests:{}},setup(e){const{events:t}=(0,G.hA)(),n=(0,u.lq)(),{layout:i}=(0,oe.g)(),s=()=>{t.commandPalette.emit({commandName:"Create Request"})},l=e=>{null!=e&&e.createNew&&"request"===n.name&&s()};return(0,r.sV)((()=>t.hotKeys.on(l))),(0,r.xo)((()=>t.hotKeys.off(l))),(e,n)=>((0,r.uX)(),(0,r.CE)("div",We,[(0,r.Lk)("div",{class:(0,o.C4)(["flex h-[calc(100%_-_50px)] flex-col items-center justify-center",{"hidden opacity-0":e.numWorkspaceRequests<=1&&"modal"!==(0,a.R1)(i)}])},["modal"!==(0,a.R1)(i)?((0,r.uX)(),(0,r.CE)("div",Xe,[(0,r.eW)(" Scalar App V"+(0,o.v_)((0,a.R1)("2.2.51"))+" Beta ",1),n[2]||(n[2]=(0,r.Lk)("div",{class:"mt-2"},[(0,r.Lk)("a",{href:"https://github.com/scalar/scalar/issues/2669",target:"_blank"}," Roadmap ")],-1))])):(0,r.Q3)("",!0),n[3]||(n[3]=(0,r.Lk)("a",{class:"gitbook-show scalar-version-number",href:"https://www.scalar.com",target:"_blank"}," Powered By Scalar.com ",-1)),(0,r.bF)(ze.A,{art:(0,a.R1)("                         .,,uod8B8bou,,.\n                ..,uod8BBBBBBBBBBBBBBBBRPFT?l!i:.\n           ||||||||||||||!?TFPRBBBBBBBBBBBBBBB8m=,\n           ||||   '\"\"^^!!||||||||||TFPRBBBVT!:...!\n           ||||            '\"\"^^!!|||||?!:.......!\n           ||||                     ||||.........!\n           ||||                     ||||.........!\n           ||||                     ||||.........!\n           ||||                     ||||.........!\n           ||||                     ||||.........!\n           ||||                     ||||.........!\n           ||||,                    ||||.........`\n           |||||!!-._               ||||.......;.\n           ':!|||||||||!!-._        ||||.....bBBBBWdou,.\n         bBBBBB86foi!|||||||!!-..:|||!..bBBBBBBBBBBBBBBY!\n         ::!?TFPRBBBBBB86foi!||||||||!!bBBBBBBBBBBBBBBY..!\n         :::::::::!?TFPRBBBBBB86ftiaabBBBBBBBBBBBBBBY....!\n         :::;`\"^!:;::::::!?TFPRBBBBBBBBBBBBBBBBBBBY......!\n         ;::::::...''^::::::::::!?TFPRBBBBBBBBBBY........!\n     .ob86foi;::::::::::::::::::::::::!?TFPRBY..........`\n    .b888888888886foi;:::::::::::::::::::::::..........`\n .b888888888888888888886foi;::::::::::::::::..........\n.b888888888888888888888888888886foi;:::::::::......`\n!Tf998888888888888888888888888888888886foi;:::....`\n  '\"^!|Tf9988888888888888888888888888888888!::..`\n       '\"^!|Tf998888888888888888888888889!! '`\n             '\"^!|Tf9988888888888888888!!`            iBBbo.\n                  '\"^!|Tf998888888889!`             WBBBBbo.\n                        '\"^!|Tf9989!`              YBBBP^'\n                              '\"^!`               `\n"),class:"text-c-3"},null,8,["art"])],2),"modal"!==(0,a.R1)(i)?((0,r.uX)(),(0,r.CE)("div",{key:0,class:(0,o.C4)(["h-[calc(100%_-_50px)] items-center justify-center hidden pb-5",{"!flex opacity-100":1==e.numWorkspaceRequests}])},[(0,r.bF)(d.A)],2)):(0,r.Q3)("",!0),(0,r.Lk)("div",qe,["modal"!==(0,a.R1)(i)?((0,r.uX)(),(0,r.CE)("button",{key:0,class:"flex items-center gap-1.5",type:"button",onClick:n[0]||(n[0]=e=>(0,a.R1)(t).commandPalette.emit())},[n[4]||(n[4]=(0,r.eW)(" Get Started ")),(0,r.bF)(je.A,{hotkey:"k"})])):(0,r.Q3)("",!0),"desktop"===(0,a.R1)(i)?((0,r.uX)(),(0,r.CE)("button",{key:1,class:"flex items-center gap-1.5",type:"button",onClick:s},[n[5]||(n[5]=(0,r.eW)(" New Request ")),(0,r.bF)(je.A,{hotkey:"N"})])):(0,r.Q3)("",!0),(0,r.Lk)("button",{class:"flex items-center gap-1.5",type:"button",onClick:n[1]||(n[1]=e=>(0,a.R1)(t).executeRequest.emit())},[n[6]||(n[6]=(0,r.eW)(" Send Request ")),(0,r.bF)(je.A,{hotkey:"↵"})])])]))}}),He=(0,O.A)(Ve,[["__scopeId","data-v-04a8f835"]]),Ze={key:0,class:"absolute inset-0 bg-b-1 flex flex-col gap-6 items-center justify-center"},Ge=(0,r.pM)({__name:"ResponseLoadingOverlay",setup(e){const{events:t}=(0,G.hA)(),n=(0,m.R2)(),o=(0,a.KR)();return t.requestStatus.on((e=>{"start"===e?o.value=setTimeout((()=>n.startLoading()),1e3):(clearTimeout(o.value),o.value=void 0,n.stopLoading())})),(e,o)=>((0,r.uX)(),(0,r.Wv)(p.eB,null,{default:(0,r.k6)((()=>[(0,a.R1)(n).isLoading?((0,r.uX)(),(0,r.CE)("div",Ze,[(0,r.bF)((0,a.R1)(m.XI),{class:"text-c-3",loadingState:(0,a.R1)(n),size:"3xl"},null,8,["loadingState"]),(0,r.bF)((0,a.R1)(m.RO),{variant:"ghost",onClick:o[0]||(o[0]=e=>(0,a.R1)(t).cancelRequest.emit())},{default:(0,r.k6)((()=>o[1]||(o[1]=[(0,r.eW)(" Cancel ")]))),_:1})])):(0,r.Q3)("",!0)])),_:1}))}}),Ke=(0,O.A)(Ge,[["__scopeId","data-v-ef5f83d8"]]),Ye=["href"],Je=(0,r.pM)({__name:"HelpfulLink",props:{href:{}},setup:e=>(e,t)=>((0,r.uX)(),(0,r.CE)("a",{class:"cursor-help decoration-c-3 underline underline-offset-2",href:e.href,rel:"noopener noreferrer",target:"_blank"},[(0,r.RG)(e.$slots,"default")],8,Ye))}),et=["B","kB","MB","GB","TB","PB","EB","ZB","YB"],tt=["B","KiB","MiB","GiB","TiB","PiB","EiB","ZiB","YiB"],nt=["b","kbit","Mbit","Gbit","Tbit","Pbit","Ebit","Zbit","Ybit"],rt=["b","kibit","Mibit","Gibit","Tibit","Pibit","Eibit","Zibit","Yibit"],at=(e,t,n)=>{let r=e;return"string"==typeof t||Array.isArray(t)?r=e.toLocaleString(t,n):!0!==t&&void 0===n||(r=e.toLocaleString(void 0,n)),r};function ot(e,t={}){if(!Number.isFinite(e))throw new TypeError("Expected a finite number");t.colonNotation&&(t.compact=!1,t.formatSubMilliseconds=!1,t.separateMilliseconds=!1,t.verbose=!1),t.compact&&(t.secondsDecimalDigits=0,t.millisecondsDecimalDigits=0);const n=[],r=(e,r,a,o)=>{if(!(0!==n.length&&t.colonNotation||0!==e||t.colonNotation&&"m"===a))return;let i,s;if(o=(o||e||"0").toString(),t.colonNotation){i=n.length>0?":":"",s="";const e=o.includes(".")?o.split(".")[0].length:o.length,t=n.length>0?2:1;o="0".repeat(Math.max(0,t-e))+o}else i="",s=t.verbose?" "+(l=r,1===e?l:`${l}s`):a;var l;n.push(i+o+s)},a=function(e){if("number"!=typeof e)throw new TypeError("Expected a number");const t=e>0?Math.floor:Math.ceil;return{days:t(e/864e5),hours:t(e/36e5)%24,minutes:t(e/6e4)%60,seconds:t(e/1e3)%60,milliseconds:t(e)%1e3,microseconds:t(1e3*e)%1e3,nanoseconds:t(1e6*e)%1e3}}(e);if(r(Math.trunc(a.days/365),"year","y"),r(a.days%365,"day","d"),r(a.hours,"hour","h"),r(a.minutes,"minute","m"),t.separateMilliseconds||t.formatSubMilliseconds||!t.colonNotation&&e<1e3)if(r(a.seconds,"second","s"),t.formatSubMilliseconds)r(a.milliseconds,"millisecond","ms"),r(a.microseconds,"microsecond","µs"),r(a.nanoseconds,"nanosecond","ns");else{const e=a.milliseconds+a.microseconds/1e3+a.nanoseconds/1e6,n="number"==typeof t.millisecondsDecimalDigits?t.millisecondsDecimalDigits:0,o=e>=1?Math.round(e):Math.ceil(e),i=n?e.toFixed(n):o;r(Number.parseFloat(i),"millisecond","ms",i)}else{const n=((e,t)=>{const n=Math.floor(e*10**t+1e-7);return(Math.round(n)/10**t).toFixed(t)})(e/1e3%60,"number"==typeof t.secondsDecimalDigits?t.secondsDecimalDigits:1),a=t.keepDecimalsOnWholeSeconds?n:n.replace(/\.0+$/,"");r(Number.parseFloat(a),"second","s",a)}if(0===n.length)return"0"+(t.verbose?" milliseconds":"ms");if(t.compact)return n[0];if("number"==typeof t.unitCount){const e=t.colonNotation?"":" ";return n.slice(0,Math.max(t.unitCount,1)).join(e)}return t.colonNotation?n.join(""):n.join(" ")}const it={class:"flex gap-1.5 text-c-1"},st={key:0},lt={key:0},ct={key:1},ut=(0,r.pM)({__name:"ResponseMetaInformation",props:{response:{}},setup(e){const t=e,{events:n}=(0,G.hA)(),s=(0,a.KR)(),l=(0,a.KR)(0);n.requestStatus.on((e=>{"start"===e?s.value=setInterval((()=>l.value+=1e3),1e3):(clearInterval(s.value),s.value=void 0,l.value=0)}));const c=e=>{var t,n;const r=parseInt((null==(t=e.headers)?void 0:t["Content-Length"])||(null==(n=e.headers)?void 0:n["content-length"])||"0",10);return r?function(e,t){if(!Number.isFinite(e))throw new TypeError(`Expected a finite number, got ${typeof e}: ${e}`);const n=(t={bits:!1,binary:!1,space:!0,...t}).bits?t.binary?rt:nt:t.binary?tt:et,r=t.space?" ":"";if(t.signed&&0===e)return` 0${r}${n[0]}`;const a=e<0,o=a?"-":t.signed?"+":"";let i;if(a&&(e=-e),void 0!==t.minimumFractionDigits&&(i={minimumFractionDigits:t.minimumFractionDigits}),void 0!==t.maximumFractionDigits&&(i={maximumFractionDigits:t.maximumFractionDigits,...i}),e<1)return o+at(e,t.locale,i)+r+n[0];const s=Math.min(Math.floor(t.binary?Math.log(e)/Math.log(1024):Math.log10(e)/3),n.length-1);return e/=(t.binary?1024:1e3)**s,i||(e=e.toPrecision(3)),o+at(Number(e),t.locale,i)+r+n[s]}(r):void 0},u=(0,r.EW)((()=>{const e=t.response.status;if(e)return i.Z7[e]??void 0}));return(e,t)=>((0,r.uX)(),(0,r.CE)("div",it,[s.value&&l.value?((0,r.uX)(),(0,r.CE)("span",st,(0,o.v_)((0,a.R1)(ot)(l.value)),1)):((0,r.uX)(),(0,r.CE)(r.FK,{key:1},[(0,r.Lk)("span",null,[t[0]||(t[0]=(0,r.Lk)("span",{class:"sr-only"},"Response Information, Duration:",-1)),(0,r.eW)(" "+(0,o.v_)((0,a.R1)(ot)(e.response.duration)),1)]),c(e.response)?((0,r.uX)(),(0,r.CE)("span",lt,[t[1]||(t[1]=(0,r.Lk)("span",{class:"sr-only"},", Size:",-1)),(0,r.eW)(" "+(0,o.v_)(c(e.response)),1)])):(0,r.Q3)("",!0),u.value?((0,r.uX)(),(0,r.CE)(r.FK,{key:1},[t[2]||(t[2]=(0,r.Lk)("span",{class:"sr-only"},", Status:",-1)),u.value.url?((0,r.uX)(),(0,r.Wv)(Je,{key:0,class:"flex items-center gap-1.5",href:u.value.url},{default:(0,r.k6)((()=>[(0,r.eW)((0,o.v_)(e.response.status)+" "+(0,o.v_)(u.value.name)+" ",1),(0,r.Lk)("span",{class:"block h-1.5 w-1.5 rounded-full",style:(0,o.Tr)({backgroundColor:u.value.color})},null,4)])),_:1},8,["href"])):((0,r.uX)(),(0,r.CE)("span",ct,[(0,r.eW)((0,o.v_)(e.response.status)+" "+(0,o.v_)(u.value.name)+" ",1),(0,r.Lk)("span",{class:"block h-1.5 w-1.5 rounded-full",style:(0,o.Tr)({backgroundColor:u.value.color})},null,4)]))],64)):(0,r.Q3)("",!0)],64))]))}}),dt=(0,r.pM)({__name:"ResponseBodyVirtual",props:{content:{},data:{},headers:{}},setup(e){const t=e,n=(0,r.EW)((()=>(0,i.nk)(t.content))),{mimeType:o,attachmentFilename:s,dataUrl:l}=(0,fe.x)({data:t.data,headers:t.headers});return(e,t)=>((0,r.uX)(),(0,r.Wv)(he.A,{class:"!max-h-100% overflow-x-auto response-body-virtual"},(0,r.eX)({title:(0,r.k6)((()=>[t[0]||(t[0]=(0,r.eW)("Body"))])),default:(0,r.k6)((()=>[t[1]||(t[1]=(0,r.Lk)("div",{class:"py-1.5 px-2.5 font-code text-xxs border-1/2 rounded-t border-b-0"}," This response body is massive! Syntax highlighting won’t work here. ",-1)),(0,r.bF)((0,a.R1)(m.$M),{containerClass:"custom-scroll scalar-code-block border-1/2 rounded-b flex flex-1 max-h-screen",contentClass:"language-plaintext whitespace-pre font-code text-base",lineHeight:20,text:n.value},null,8,["text"])])),_:2},[(0,a.R1)(l)?{name:"actions",fn:(0,r.k6)((()=>[(0,r.bF)(be,{filename:(0,a.R1)(s),href:(0,a.R1)(l),type:(0,a.R1)(o).essence},null,8,["filename","href","type"])])),key:"0"}:void 0]),1024))}});var pt=n(9742),ht=n(3016),ft=n(4282);const mt={class:"flex-1 px-2 py-1.5 whitespace-nowrap"},gt=(0,r.pM)({__name:"DataTableText",props:{text:{}},setup:e=>(e,t)=>((0,r.uX)(),(0,r.Wv)(ft.A,{class:"relative flex"},{default:(0,r.k6)((()=>[(0,r.Lk)("span",mt,[(0,r.RG)(e.$slots,"default",{},(()=>[(0,r.eW)((0,o.v_)(e.text),1)]))])])),_:3}))}),vt={key:1,class:"text-c-3 px-4 text-sm border-t-1/2 min-h-16 justify-center flex items-center bg-b-1"},bt=(0,r.pM)({__name:"ResponseCookies",props:{cookies:{}},setup:e=>(e,t)=>((0,r.uX)(),(0,r.Wv)(he.A,{defaultOpen:!1},{title:(0,r.k6)((()=>t[0]||(t[0]=[(0,r.eW)("Cookies")]))),default:(0,r.k6)((()=>[((0,r.uX)(),(0,r.CE)(r.FK,{key:0},[e.cookies.length?((0,r.uX)(),(0,r.Wv)(pt.A,{key:0,class:"flex-1",columns:["",""]},{default:(0,r.k6)((()=>[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(e.cookies,(e=>((0,r.uX)(),(0,r.Wv)(ht.A,{key:e.name},{default:(0,r.k6)((()=>[(0,r.bF)(gt,{text:e.name},null,8,["text"]),(0,r.bF)(gt,{text:e.value},null,8,["text"])])),_:2},1024)))),128))])),_:1})):((0,r.uX)(),(0,r.CE)("div",vt," No cookies "))],64))])),_:1}))}),yt=[{name:"Accept",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept"},{name:"Accept-CH",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept-CH"},{name:"Accept-CH-Lifetime",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept-CH-Lifetime"},{name:"Accept-Charset",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept-Charset"},{name:"Accept-Encoding",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept-Encoding"},{name:"Accept-Language",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept-Language"},{name:"Accept-Patch",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept-Patch"},{name:"Accept-Post",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept-Post"},{name:"Accept-Ranges",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept-Ranges"},{name:"Access-Control-Allow-Credentials",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Credentials"},{name:"Access-Control-Allow-Headers",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Headers"},{name:"Access-Control-Allow-Methods",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Methods"},{name:"Access-Control-Allow-Origin",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Origin"},{name:"Access-Control-Expose-Headers",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Expose-Headers"},{name:"Access-Control-Max-Age",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Max-Age"},{name:"Access-Control-Request-Headers",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Request-Headers"},{name:"Access-Control-Request-Method",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Request-Method"},{name:"Age",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Age"},{name:"Allow",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Allow"},{name:"Alt-Svc",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Alt-Svc"},{name:"Alt-Used",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Alt-Used"},{name:"Authorization",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Authorization"},{name:"Cache-Control",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cache-Control"},{name:"Clear-Site-Data",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Clear-Site-Data"},{name:"Connection",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Connection"},{name:"Content-Disposition",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Disposition"},{name:"Content-DPR",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-DPR"},{name:"Content-Encoding",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Encoding"},{name:"Content-Language",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Language"},{name:"Content-Length",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Length"},{name:"Content-Location",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Location"},{name:"Content-Range",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Range"},{name:"Content-Security-Policy",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Security-Policy"},{name:"Content-Security-Policy-Report-Only",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Security-Policy-Report-Only"},{name:"Content-Type",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Type"},{name:"Cookie",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cookie"},{name:"Critical-CH",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Critical-CH"},{name:"Cross-Origin-Embedder-Policy",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cross-Origin-Embedder-Policy"},{name:"Cross-Origin-Opener-Policy",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cross-Origin-Opener-Policy"},{name:"Cross-Origin-Resource-Policy",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cross-Origin-Resource-Policy"},{name:"Date",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Date"},{name:"Device-Memory",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Device-Memory"},{name:"Digest",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Digest"},{name:"DNT",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/DNT"},{name:"Downlink",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Downlink"},{name:"DPR",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/DPR"},{name:"Early-Data",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Early-Data"},{name:"ECT",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/ECT"},{name:"ETag",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/ETag"},{name:"Expect",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Expect"},{name:"Expect-CT",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Expect-CT"},{name:"Expires",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Expires"},{name:"Forwarded",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Forwarded"},{name:"From",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/From"},{name:"Host",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Host"},{name:"If-Match",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/If-Match"},{name:"If-Modified-Since",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/If-Modified-Since"},{name:"If-None-Match",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/If-None-Match"},{name:"If-Range",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/If-Range"},{name:"If-Unmodified-Since",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/If-Unmodified-Since"},{name:"Keep-Alive",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Keep-Alive"},{name:"Large-Allocation",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Large-Allocation"},{name:"Last-Modified",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Last-Modified"},{name:"Link",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Link"},{name:"Location",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Location"},{name:"Max-Forwards",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Max-Forwards"},{name:"NEL",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/NEL"},{name:"Origin",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Origin"},{name:"Permissions-Policy",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Permissions-Policy"},{name:"Pragma",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Pragma"},{name:"Proxy-Authenticate",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Proxy-Authenticate"},{name:"Proxy-Authorization",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Proxy-Authorization"},{name:"Range",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Range"},{name:"Referer",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Referer"},{name:"Referrer-Policy",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Referrer-Policy"},{name:"Retry-After",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Retry-After"},{name:"RTT",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/RTT"},{name:"Save-Data",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Save-Data"},{name:"Sec-CH-Prefers-Reduced-Motion",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-CH-Prefers-Reduced-Motion"},{name:"Sec-CH-Prefers-Reduced-Transparency",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-CH-Prefers-Reduced-Transparency"},{name:"Sec-CH-UA",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-CH-UA"},{name:"Sec-CH-UA-Arch",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-CH-UA-Arch"},{name:"Sec-CH-UA-Bitness",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-CH-UA-Bitness"},{name:"Sec-CH-UA-Full-Version",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-CH-UA-Full-Version"},{name:"Sec-CH-UA-Full-Version-List",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-CH-UA-Full-Version-List"},{name:"Sec-CH-UA-Mobile",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-CH-UA-Mobile"},{name:"Sec-CH-UA-Model",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-CH-UA-Model"},{name:"Sec-CH-UA-Platform",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-CH-UA-Platform"},{name:"Sec-CH-UA-Platform-Version",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-CH-UA-Platform-Version"},{name:"Sec-Fetch-Dest",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-Fetch-Dest"},{name:"Sec-Fetch-Mode",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-Fetch-Mode"},{name:"Sec-Fetch-Site",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-Fetch-Site"},{name:"Sec-Fetch-User",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-Fetch-User"},{name:"Sec-GPC",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-GPC"},{name:"Sec-Purpose",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-Purpose"},{name:"Sec-WebSocket-Accept",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Sec-WebSocket-Accept"},{name:"Server",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Server"},{name:"Server-Timing",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Server-Timing"},{name:"Service-Worker-Navigation-Preload",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Service-Worker-Navigation-Preload"},{name:"Set-Cookie",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie"},{name:"SourceMap",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/SourceMap"},{name:"Strict-Transport-Security",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Strict-Transport-Security"},{name:"TE",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/TE"},{name:"Timing-Allow-Origin",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Timing-Allow-Origin"},{name:"Tk",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Tk"},{name:"Trailer",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Trailer"},{name:"Transfer-Encoding",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Transfer-Encoding"},{name:"Upgrade",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Upgrade"},{name:"Upgrade-Insecure-Requests",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Upgrade-Insecure-Requests"},{name:"User-Agent",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/User-Agent"},{name:"Vary",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Vary"},{name:"Via",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Via"},{name:"Viewport-Width",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Viewport-Width"},{name:"Want-Digest",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Want-Digest"},{name:"Warning",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Warning"},{name:"Width",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Width"},{name:"WWW-Authenticate",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/WWW-Authenticate"},{name:"X-Content-Type-Options",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Content-Type-Options"},{name:"X-DNS-Prefetch-Control",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-DNS-Prefetch-Control"},{name:"Non-standard",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Non-standard"},{name:"X-Forwarded-For",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Forwarded-For"},{name:"Non-standard",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Non-standard"},{name:"X-Forwarded-Host",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Forwarded-Host"},{name:"Non-standard",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Non-standard"},{name:"X-Forwarded-Proto",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Forwarded-Proto"},{name:"Non-standard",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Non-standard"},{name:"X-Frame-Options",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Frame-Options"},{name:"X-XSS-Protection",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-XSS-Protection"},{name:"Cf-Cache-Status",url:"https://developers.cloudflare.com/cache/concepts/default-cache-behavior/#cloudflare-cache-responses"},{name:"Cf-Ray",url:"https://developers.cloudflare.com/fundamentals/get-started/reference/http-request-headers/#cf-ray"},{name:"Report-To",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Security-Policy/report-to"},{name:"X-Cloud-Trace-Context",url:"https://cloud.google.com/trace/docs/trace-context#legacy-http-header"},{name:"Speculation-Rules",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Speculation-Rules"}],wt={key:0,class:"border-t-1/2 border-b-1/2 max-h-[calc(100%-32px)] overflow-y-auto"},Ot={key:1,class:"text-c-3 px-4 text-sm border rounded min-h-12 justify-center flex items-center bg-b-1"},kt=(0,r.pM)({__name:"ResponseHeaders",props:{headers:{}},setup(e){const t=e=>yt.find((t=>t.name.toLowerCase()===e.toLowerCase()));return(e,n)=>((0,r.uX)(),(0,r.Wv)(he.A,{class:"overflow-auto",defaultOpen:!1,itemCount:e.headers.length},{title:(0,r.k6)((()=>n[0]||(n[0]=[(0,r.eW)("Headers")]))),default:(0,r.k6)((()=>[e.headers.length?((0,r.uX)(),(0,r.CE)("div",wt,[(0,r.bF)(pt.A,{class:"!border-0 !mx-0",columns:["minmax(auto, min-content)","minmax(50%, 1fr)"],scroll:""},{default:(0,r.k6)((()=>[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(e.headers,(e=>((0,r.uX)(),(0,r.Wv)(ht.A,{key:e.name,class:"text-c-1"},{default:(0,r.k6)((()=>[(0,r.bF)(gt,{class:"sticky left-0 z-1 bg-b-1 max-w-48"},{default:(0,r.k6)((()=>{var n;return["string"==typeof(null==(n=t(e.name))?void 0:n.url)?((0,r.uX)(),(0,r.Wv)(Je,{key:0,class:"decoration-c-3",href:t(e.name).url},{default:(0,r.k6)((()=>[(0,r.eW)((0,o.v_)(e.name),1)])),_:2},1032,["href"])):((0,r.uX)(),(0,r.CE)(r.FK,{key:1},[(0,r.eW)((0,o.v_)(e.name),1)],64))]})),_:2},1024),(0,r.bF)(gt,{class:"z-0",text:e.value},null,8,["text"])])),_:2},1024)))),128))])),_:1})])):((0,r.uX)(),(0,r.CE)("div",Ot," No Headers "))])),_:1},8,["itemCount"]))}}),xt={class:"flex items-center flex-1 h-8"},Et=2e5,St=(0,r.pM)({__name:"ResponseSection",props:{numWorkspaceRequests:{},response:{}},setup(e){const t=(0,r.EW)((()=>{var t;const n=null==(t=e.response)?void 0:t.headers;return n?Object.keys(n).map((e=>({name:e,value:n[e]??"",required:!1}))).filter((e=>!["rest-api-client-content-length","X-API-Client-Content-Length"].includes(e.name))):[]})),n=(0,r.EW)((()=>{var t;return(null==(t=e.response)?void 0:t.cookieHeaderKeys.flatMap((t=>{var n,r;const a=null==(r=null==(n=e.response)?void 0:n.headers)?void 0:r[t];return a?{name:t,value:a,required:!1}:[]})))??[]})),i=["All","Cookies","Headers","Body"],s=(0,a.KR)("All"),l=(0,r.EW)((()=>{var t,n;if(!e.response)return!1;const r=(null==(t=e.response.headers)?void 0:t["content-type"])||(null==(n=e.response.headers)?void 0:n["Content-Type"]);return!(!r||(e.response.size??0)<=Et||r.includes("text/html"))&&["text/","application/json","application/ld+json","application/problem+json","application/vnd.api+json","application/xml","application/atom+xml","application/rss+xml","application/problem+xml","application/javascript","application/ecmascript","application/x-yaml","application/yaml","application/x-httpd-php","application/x-sh","application/x-perl","application/x-python","application/x-ruby","application/x-java-source","application/x-www-form-urlencoded"].some((e=>r.includes(e)))&&(e.response.size??0)>Et}));return(e,a)=>((0,r.uX)(),(0,r.Wv)(pe.A,{"aria-label":"Response"},{title:(0,r.k6)((()=>[(0,r.Lk)("div",xt,[(0,r.Lk)("div",{"aria-live":"polite",class:(0,o.C4)(["flex items-center",{"animate-response-heading":e.response}])},[a[1]||(a[1]=(0,r.Lk)("span",{class:"response-heading absolute pointer-events-none"}," Response ",-1)),e.response?((0,r.uX)(),(0,r.Wv)(ut,{key:0,class:"animate-response-children",response:e.response},null,8,["response"])):(0,r.Q3)("",!0)],2),(0,r.bF)(de.A,{activeSection:s.value,sections:i,onSetActiveSection:a[0]||(a[0]=e=>s.value=e)},null,8,["activeSection"])])])),default:(0,r.k6)((()=>{var a,i,c;return[(0,r.Lk)("div",{class:(0,o.C4)(["custom-scroll h-full relative grid justify-stretch",{"content-start":e.response}])},[e.response?((0,r.uX)(),(0,r.CE)(r.FK,{key:1},["All"===s.value||"Cookies"===s.value?((0,r.uX)(),(0,r.Wv)(bt,{key:0,cookies:n.value},null,8,["cookies"])):(0,r.Q3)("",!0),"All"===s.value||"Headers"===s.value?((0,r.uX)(),(0,r.Wv)(kt,{key:1,headers:t.value},null,8,["headers"])):(0,r.Q3)("",!0),"All"===s.value||"Body"===s.value?((0,r.uX)(),(0,r.CE)(r.FK,{key:2},[l.value&&"string"==typeof(null==(a=e.response)?void 0:a.data)?((0,r.uX)(),(0,r.Wv)(dt,{key:0,content:e.response.data,data:null==(i=e.response)?void 0:i.data,headers:t.value},null,8,["content","data","headers"])):((0,r.uX)(),(0,r.Wv)(Ue,{key:1,active:!0,data:null==(c=e.response)?void 0:c.data,headers:t.value,title:"Body"},null,8,["data","headers"]))],64)):(0,r.Q3)("",!0)],64)):((0,r.uX)(),(0,r.Wv)(He,{key:0,numWorkspaceRequests:e.numWorkspaceRequests},null,8,["numWorkspaceRequests"])),(0,r.bF)(Ke)],2)]})),_:1}))}}),_t=(0,O.A)(St,[["__scopeId","data-v-ec732740"]]);var Ct=n(2667);const Tt={class:"font-medium m-0 flex gap-1.5 items-center whitespace-nowrap"},At={class:"flex items-center justify-center h-4 w-4"},Rt=(0,r.pM)({__name:"EnvironmentSelector",setup(e){const{activeCollection:t,activeWorkspace:n,activeEnvironment:i}=(0,f.JC)(),{collectionMutators:s}=(0,G.hA)(),{layout:l}=(0,oe.g)(),c=(0,u.rd)(),d=e=>{t.value&&n.value&&(s.edit(t.value.uid,"x-scalar-active-environment",e),n.value.activeEnvironmentId=e)},h=()=>{var e;return c.push({name:"environment",params:{environment:null==(e=n.value)?void 0:e.uid}})},g=(0,r.EW)((()=>{const{value:e}=i,{value:n}=t;return(null==e?void 0:e.uid)||(null==n?void 0:n["x-scalar-active-environment"])||"No Environment"})),v=(0,r.EW)((()=>{const{value:e}=t,n=null==e?void 0:e["x-scalar-environments"];return n?Object.entries(n).map((([e,t])=>({...t,uid:e,name:e}))):[]})),b=e=>{const r=e["x-scalar-active-environment"];r&&t.value&&n.value?(t.value["x-scalar-active-environment"]=r,n.value.activeEnvironmentId=r):n.value&&(n.value.activeEnvironmentId="")};return(0,r.wB)(t,(e=>e&&b(e))),(0,r.sV)((()=>{t.value&&b(t.value)})),(e,n)=>((0,r.uX)(),(0,r.CE)("div",null,[(0,r.bF)((0,a.R1)(m.SM),{placement:"bottom-end"},{items:(0,r.k6)((()=>[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(v.value,(e=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.rA),{key:e.uid,class:"flex gap-1.5 group/item items-center whitespace-nowrap text-ellipsis overflow-hidden",onClick:(0,p.D$)((t=>d(e.uid)),["stop"])},{default:(0,r.k6)((()=>{var n;return[(0,r.bF)((0,a.R1)(m.Pq),{selected:(null==(n=(0,a.R1)(t))?void 0:n["x-scalar-active-environment"])===e.uid},null,8,["selected"]),(0,r.eW)(" "+(0,o.v_)(e.name),1)]})),_:2},1032,["onClick"])))),128)),(0,r.bF)((0,a.R1)(m.rA),{class:"flex gap-1.5 group/item items-center whitespace-nowrap text-ellipsis overflow-hidden",onClick:n[0]||(n[0]=(0,p.D$)((e=>d("")),["stop"]))},{default:(0,r.k6)((()=>{var e,s;return[(0,r.Lk)("div",{class:(0,o.C4)(["flex items-center justify-center rounded-full p-[3px] w-4 h-4",""===(null==(e=(0,a.R1)(i))?void 0:e.uid)&&""===(null==(s=(0,a.R1)(t))?void 0:s["x-scalar-active-environment"])?"bg-c-accent text-b-1":"shadow-border text-transparent"])},[(0,r.bF)((0,a.R1)(m.kv),{class:"size-2.5",icon:"Checkmark",thickness:"3"})],2),n[1]||(n[1]=(0,r.eW)(" No Environment "))]})),_:1}),(0,r.bF)((0,a.R1)(m.t)),"modal"!==(0,a.R1)(l)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.rA),{key:0,class:"flex items-center gap-1.5",onClick:h},{default:(0,r.k6)((()=>[(0,r.Lk)("div",At,[(0,r.bF)((0,a.R1)(m.kv),{icon:"Brackets",size:"sm"})]),n[2]||(n[2]=(0,r.Lk)("span",{class:"leading-none"},"Manage Environments",-1))])),_:1})):(0,r.Q3)("",!0)])),default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.RO),{class:"font-normal h-auto justify-start py-1.5 px-1.5 pl-2 text-c-1 hover:bg-b-2 w-fit",fullWidth:"",variant:"ghost"},{default:(0,r.k6)((()=>[(0,r.Lk)("h2",Tt,(0,o.v_)(g.value),1)])),_:1})])),_:1})]))}});var Pt=n(5929);var Dt=n(4486),Mt=n(7356),$t=n(8547),Nt=n(6923),It=n(8869),Lt=n(8663),Bt=n(805),Ft=n(7838),Qt=n(6413),Ut=n(1417),zt=n(287),jt=n(562),Wt=n(3605),Xt=n(1567);let qt=Symbol("LabelContext");function Vt(){let e=(0,r.WQ)(qt,null);if(null===e){let e=new Error("You used a <Label /> component, but it is not inside a parent.");throw Error.captureStackTrace&&Error.captureStackTrace(e,Vt),e}return e}function Ht({slot:e={},name:t="Label",props:n={}}={}){let o=(0,a.KR)([]);return(0,r.Gt)(qt,{register:function(e){return o.value.push(e),()=>{let t=o.value.indexOf(e);-1!==t&&o.value.splice(t,1)}},slot:e,name:t,props:n}),(0,r.EW)((()=>o.value.length>0?o.value.join(" "):void 0))}let Zt=(0,r.pM)({name:"Label",props:{as:{type:[Object,String],default:"label"},passive:{type:[Boolean],default:!1},id:{type:String,default:null}},setup(e,{slots:t,attrs:n}){var o;let i=null!=(o=e.id)?o:`headlessui-label-${(0,It.B)()}`,s=Vt();return(0,r.sV)((()=>(0,r.hi)(s.register(i)))),()=>{let{name:r="Label",slot:o={},props:l={}}=s,{passive:c,...u}=e,d={...Object.entries(l).reduce(((e,[t,n])=>Object.assign(e,{[t]:(0,a.R1)(n)})),{}),id:i};return c&&(delete d.onClick,delete d.htmlFor,delete u.onClick),(0,Wt.XX)({ourProps:d,theirProps:u,slot:o,attrs:n,slots:t,name:r})}}});function Gt(e,t){return e===t}let Kt=Symbol("RadioGroupContext");function Yt(e){let t=(0,r.WQ)(Kt,null);if(null===t){let t=new Error(`<${e} /> is missing a parent <RadioGroup /> component.`);throw Error.captureStackTrace&&Error.captureStackTrace(t,Yt),t}return t}let Jt=(0,r.pM)({name:"RadioGroup",emits:{"update:modelValue":e=>!0},props:{as:{type:[Object,String],default:"div"},disabled:{type:[Boolean],default:!1},by:{type:[String,Function],default:()=>Gt},modelValue:{type:[Object,String,Number,Boolean],default:void 0},defaultValue:{type:[Object,String,Number,Boolean],default:void 0},form:{type:String,optional:!0},name:{type:String,optional:!0},id:{type:String,default:null}},inheritAttrs:!1,setup(e,{emit:t,attrs:n,slots:o,expose:i}){var s;let l=null!=(s=e.id)?s:`headlessui-radiogroup-${(0,It.B)()}`,c=(0,a.KR)(null),u=(0,a.KR)([]),d=Ht({name:"RadioGroupLabel"}),p=(0,Xt.r)({name:"RadioGroupDescription"});i({el:c,$el:c});let[h,f]=(0,Nt.P)((0,r.EW)((()=>e.modelValue)),(e=>t("update:modelValue",e)),(0,r.EW)((()=>e.defaultValue))),m={options:u,value:h,disabled:(0,r.EW)((()=>e.disabled)),firstOption:(0,r.EW)((()=>u.value.find((e=>!e.propsRef.disabled)))),containsCheckedOption:(0,r.EW)((()=>u.value.some((t=>m.compare((0,a.ux)(t.propsRef.value),(0,a.ux)(e.modelValue)))))),compare(t,n){if("string"==typeof e.by){let r=e.by;return(null==t?void 0:t[r])===(null==n?void 0:n[r])}return e.by(t,n)},change(t){var n;if(e.disabled||m.compare((0,a.ux)(h.value),(0,a.ux)(t)))return!1;let r=null==(n=u.value.find((e=>m.compare((0,a.ux)(e.propsRef.value),(0,a.ux)(t)))))?void 0:n.propsRef;return!(null!=r&&r.disabled||(f(t),0))},registerOption(e){u.value.push(e),u.value=(0,Ut.wl)(u.value,(e=>e.element))},unregisterOption(e){let t=u.value.findIndex((t=>t.id===e));-1!==t&&u.value.splice(t,1)}};function g(e){if(!c.value||!c.value.contains(e.target))return;let t=u.value.filter((e=>!1===e.propsRef.disabled)).map((e=>e.element));switch(e.key){case Ft.D.Enter:(0,zt.q)(e.currentTarget);break;case Ft.D.ArrowLeft:case Ft.D.ArrowUp:if(e.preventDefault(),e.stopPropagation(),(0,Ut.CU)(t,Ut.BD.Previous|Ut.BD.WrapAround)===Ut.Me.Success){let e=u.value.find((e=>{var t;return e.element===(null==(t=(0,jt.T)(c))?void 0:t.activeElement)}));e&&m.change(e.propsRef.value)}break;case Ft.D.ArrowRight:case Ft.D.ArrowDown:if(e.preventDefault(),e.stopPropagation(),(0,Ut.CU)(t,Ut.BD.Next|Ut.BD.WrapAround)===Ut.Me.Success){let e=u.value.find((e=>{var t;return e.element===(null==(t=(0,jt.T)(e.element))?void 0:t.activeElement)}));e&&m.change(e.propsRef.value)}break;case Ft.D.Space:{e.preventDefault(),e.stopPropagation();let t=u.value.find((e=>{var t;return e.element===(null==(t=(0,jt.T)(e.element))?void 0:t.activeElement)}));t&&m.change(t.propsRef.value)}}}(0,r.Gt)(Kt,m),(0,Lt.i)({container:(0,r.EW)((()=>(0,Qt.t)(c))),accept:e=>"radio"===e.getAttribute("role")?NodeFilter.FILTER_REJECT:e.hasAttribute("role")?NodeFilter.FILTER_SKIP:NodeFilter.FILTER_ACCEPT,walk(e){e.setAttribute("role","none")}});let v=(0,r.EW)((()=>{var e;return null==(e=(0,Qt.t)(c))?void 0:e.closest("form")}));return(0,r.sV)((()=>{(0,r.wB)([v],(()=>{if(v.value&&void 0!==e.defaultValue)return v.value.addEventListener("reset",t),()=>{var e;null==(e=v.value)||e.removeEventListener("reset",t)};function t(){m.change(e.defaultValue)}}),{immediate:!0})})),()=>{let{disabled:t,name:a,form:i,...s}=e,u={ref:c,id:l,role:"radiogroup","aria-labelledby":d.value,"aria-describedby":p.value,onKeydown:g};return(0,r.h)(r.FK,[...null!=a&&null!=h.value?(0,zt.h)({[a]:h.value}).map((([e,n])=>(0,r.h)(Bt.j,(0,Wt.oE)({features:Bt.O.Hidden,key:e,as:"input",type:"hidden",hidden:!0,readOnly:!0,form:i,disabled:t,name:e,value:n})))):[],(0,Wt.XX)({ourProps:u,theirProps:{...n,...(0,Wt.cJ)(s,["modelValue","defaultValue","by"])},slot:{},attrs:n,slots:o,name:"RadioGroup"})])}}});var en=(e=>(e[e.Empty=1]="Empty",e[e.Active=2]="Active",e))(en||{});let tn=(0,r.pM)({name:"RadioGroupOption",props:{as:{type:[Object,String],default:"div"},value:{type:[Object,String,Number,Boolean]},disabled:{type:Boolean,default:!1},id:{type:String,default:null}},setup(e,{attrs:t,slots:n,expose:o}){var i;let s=null!=(i=e.id)?i:`headlessui-radiogroup-option-${(0,It.B)()}`,l=Yt("RadioGroupOption"),c=Ht({name:"RadioGroupLabel"}),u=(0,Xt.r)({name:"RadioGroupDescription"}),d=(0,a.KR)(null),p=(0,r.EW)((()=>({value:e.value,disabled:e.disabled}))),h=(0,a.KR)(1);o({el:d,$el:d});let f=(0,r.EW)((()=>(0,Qt.t)(d)));(0,r.sV)((()=>l.registerOption({id:s,element:f,propsRef:p}))),(0,r.hi)((()=>l.unregisterOption(s)));let m=(0,r.EW)((()=>{var e;return(null==(e=l.firstOption.value)?void 0:e.id)===s})),g=(0,r.EW)((()=>l.disabled.value||e.disabled)),v=(0,r.EW)((()=>l.compare((0,a.ux)(l.value.value),(0,a.ux)(e.value)))),b=(0,r.EW)((()=>g.value?-1:v.value||!l.containsCheckedOption.value&&m.value?0:-1));function y(){var t;l.change(e.value)&&(h.value|=2,null==(t=(0,Qt.t)(d))||t.focus())}function w(){h.value|=2}function O(){h.value&=-3}return()=>{let{value:r,disabled:a,...o}=e,i={checked:v.value,disabled:g.value,active:Boolean(2&h.value)},l={id:s,ref:d,role:"radio","aria-checked":v.value?"true":"false","aria-labelledby":c.value,"aria-describedby":u.value,"aria-disabled":!!g.value||void 0,tabIndex:b.value,onClick:g.value?void 0:y,onFocus:g.value?void 0:w,onBlur:g.value?void 0:O};return(0,Wt.XX)({ourProps:l,theirProps:o,slot:i,attrs:t,slots:n,name:"RadioGroupOption"})}}}),nn=Zt;const rn={class:"flex text-sm"},an={class:"grid grid-cols-[repeat(auto-fill,minmax(32px,1fr))] auto-rows-[32px] justify-between content-between rounded bg-b-1 border p-1 max-w-[420px] w-dvw custom-scroll"},on=(0,r.pM)({__name:"IconSelector",props:{modelValue:{},placement:{}},emits:["update:modelValue"],setup(e,{emit:t}){const n=e,i=t,s=(0,r.EW)({get:()=>n.modelValue,set:e=>i("update:modelValue",e)});return(e,t)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.w),{class:"bg-b-2 rounded",placement:e.placement??"bottom"},{popover:(0,r.k6)((({close:n})=>[(0,r.bF)((0,a.R1)(Jt),{modelValue:s.value,"onUpdate:modelValue":t[0]||(t[0]=e=>s.value=e),class:"flex flex-col"},{default:(0,r.k6)((()=>[(0,r.Lk)("div",rn,[(0,r.bF)((0,a.R1)(nn),{class:"text-c-2 py-1 px-1"},{default:(0,r.k6)((()=>[(0,r.RG)(e.$slots,"title",{},(()=>[t[1]||(t[1]=(0,r.eW)("Select an icon"))]))])),_:3})]),(0,r.Lk)("ul",an,[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)((0,a.R1)(Ct.Sb),(e=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(tn),{key:e.src,as:"li",class:"p-2 rounded flex items-center justify-center text-c-3 cursor-pointer hover:text-c-2 hover:bg-b-2 ui-checked:bg-b-3 ui-active:bg-b-2",value:e.src,onClick:n},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(nn),{class:"sr-only"},{default:(0,r.k6)((()=>[(0,r.eW)((0,o.v_)(e.src.replaceAll("-"," "))+" Icon ",1)])),_:2},1024),(0,r.bF)((0,a.R1)(Ct.$n),{class:"stroke-[1.5]",src:e.src},null,8,["src"])])),_:2},1032,["value","onClick"])))),128))])])),_:2},1032,["modelValue"])])),default:(0,r.k6)((()=>[(0,r.RG)(e.$slots,"default")])),_:3},8,["placement"]))}});var sn=n(6918);const ln={class:"grid gap-2 grid-cols-[auto,1fr]"},cn={class:"flex aspect-square"},un=(0,r.pM)({__name:"EditSidebarListCollection",props:{name:{},icon:{}},emits:["close","edit"],setup(e,{emit:t}){const n=e,o=t,i=(0,a.KR)(n.name),s=(0,a.KR)(n.icon);return(e,t)=>((0,r.uX)(),(0,r.Wv)(sn.A,{onCancel:t[2]||(t[2]=e=>o("close")),onSubmit:t[3]||(t[3]=e=>o("edit",i.value,s.value))},{default:(0,r.k6)((()=>[(0,r.Lk)("div",ln,[(0,r.Lk)("div",cn,[(0,r.bF)(on,{modelValue:s.value,"onUpdate:modelValue":t[0]||(t[0]=e=>s.value=e),placement:"bottom-start"},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.RO),{class:"aspect-square px-0 h-auto",variant:"outlined"},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(Ct.$n),{class:"size-4 text-c-2",src:s.value},null,8,["src"])])),_:1})])),_:1},8,["modelValue"])]),(0,r.bF)((0,a.R1)(m.KB),{modelValue:i.value,"onUpdate:modelValue":t[1]||(t[1]=e=>i.value=e),autofocus:"",class:"flex-1"},null,8,["modelValue"])])])),_:1}))}});var dn=n(68);const pn=(0,r.pM)({__name:"RequestSidebarItemMenu",props:{menuItem:{}},emits:["closeMenu","toggleWatchMode","clearDrafts"],setup(e,{emit:t}){const n=e,i=t,{replace:s}=(0,u.rd)(),{activeRouterParams:l,activeWorkspaceCollections:c,activeWorkspaceRequests:d}=(0,f.JC)(),{events:h,requestMutators:g}=(0,G.hA)(),v=(0,m.hS)(),b=(0,m.hS)(),y=(0,m.hS)(),w=()=>{var e;return h.commandPalette.emit({commandName:"Add Example",metaData:{itemUid:null==(e=n.menuItem.item)?void 0:e.entity.uid}})},O=(e,t)=>{var r;null==(r=n.menuItem.item)||r.edit(e,t),v.hide()},k=()=>{var e,t,r;if(null==(e=n.menuItem.item)||e.delete(),!d.value.length){const{request:e}=(0,Mt.AJ)(),t=c.value.find((e=>{var t;return"Drafts"===(null==(t=e.info)?void 0:t.title)}));t&&(g.add(e,t.uid),s({name:"request",params:{[I.TU.Request]:e.uid}}))}if(l.value[I.TU.Request]===(null==(t=n.menuItem.item)?void 0:t.entity.uid)&&s({name:"request",params:{[I.TU.Request]:"default"}}),l.value[I.TU.Examples]===(null==(r=n.menuItem.item)?void 0:r.entity.uid)&&s({name:"request",params:{[I.TU.Request]:"default"}}),c.value[0]){const e=c.value[0].requests[0];s({name:"request",params:{[I.TU.Request]:e}})}b.hide()},x=(0,a.KR)(null);(0,r.wB)([()=>n.menuItem.open,x],(async([e])=>{var t,n;e&&null!=(n=null==(t=x.value)?void 0:t.$parent)&&n.$el&&x.value.$parent.$el.focus()}));const E=()=>n.menuItem.open&&i("closeMenu");(0,r.sV)((()=>window.addEventListener("click",E))),(0,r.xo)((()=>window.removeEventListener("click",E)));const S=()=>{i("toggleWatchMode",n.menuItem.item)},_=()=>{i("clearDrafts"),y.hide()},C=(0,r.EW)((()=>{var e;return"Drafts"===(null==(e=n.menuItem.item)?void 0:e.title)}));return(e,t)=>{var n,i;return(0,r.uX)(),(0,r.CE)(r.FK,null,[e.menuItem.targetRef&&e.menuItem.open?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.Z5),{key:0,placement:"right-start",target:e.menuItem.targetRef,teleport:""},{floating:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.dR),{onKeydown:t[3]||(t[3]=(0,p.jR)((t=>e.$emit("closeMenu")),["escape"]))},{default:(0,r.k6)((()=>{var n,i;return["request"===(null==(n=e.menuItem.item)?void 0:n.entity.type)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.kZ),{key:0,class:"flex gap-2",onClick:w},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.kv),{class:"inline-flex",icon:"Example",size:"md",thickness:"1.5"}),t[8]||(t[8]=(0,r.Lk)("span",null,"Add Example",-1))])),_:1})):(0,r.Q3)("",!0),C.value?(0,r.Q3)("",!0):((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.kZ),{key:1,ref_key:"menuRef",ref:x,class:"flex gap-2",onClick:t[0]||(t[0]=e=>(0,a.R1)(v).show())},{default:(0,r.k6)((()=>{var t;return[(0,r.bF)((0,a.R1)(m.kv),{class:"inline-flex",icon:"Edit",size:"md",thickness:"1.5"}),(0,r.Lk)("span",null,["collection"===(null==(t=e.menuItem.item)?void 0:t.entity.type)?((0,r.uX)(),(0,r.CE)(r.FK,{key:0},[(0,r.eW)(" Edit ")],64)):((0,r.uX)(),(0,r.CE)(r.FK,{key:1},[(0,r.eW)(" Rename ")],64))])]})),_:1},512)),null!=(i=e.menuItem.item)&&i.documentUrl?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.kZ),{key:2,ref_key:"menuRef",ref:x,class:"flex gap-2",onClick:S},{default:(0,r.k6)((()=>{var t,n;return[(0,r.bF)((0,a.R1)(m.kv),{class:"inline-flex",icon:null!=(t=e.menuItem.item)&&t.watchMode?"Unwatch":"Watch",size:"md",thickness:"1.5"},null,8,["icon"]),(0,r.Lk)("span",null,(0,o.v_)(null!=(n=e.menuItem.item)&&n.watchMode?"Disable Watch Mode":"Enable Watch Mode"),1)]})),_:1},512)):(0,r.Q3)("",!0),C.value?(0,r.Q3)("",!0):((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.kZ),{key:3,class:"flex gap-2",onClick:t[1]||(t[1]=e=>(0,a.R1)(b).show())},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.kv),{class:"inline-flex",icon:"Delete",size:"md",thickness:"1.5"}),t[9]||(t[9]=(0,r.Lk)("span",null,"Delete",-1))])),_:1})),C.value?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.kZ),{key:4,class:"flex gap-2",onClick:t[2]||(t[2]=e=>(0,a.R1)(y).show())},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.kv),{class:"inline-flex",icon:"Delete",size:"md",thickness:"1.5"}),t[10]||(t[10]=(0,r.Lk)("span",null,"Clear Drafts",-1))])),_:1})):(0,r.Q3)("",!0)]})),_:1})])),_:1},8,["target"])):(0,r.Q3)("",!0),(0,r.bF)((0,a.R1)(m.uT),{size:"xxs",state:(0,a.R1)(b),title:`Delete ${null==(n=e.menuItem.item)?void 0:n.resourceTitle}`},{default:(0,r.k6)((()=>{var n,o;return[(0,r.bF)($t.A,{variableName:(null==(n=e.menuItem.item)?void 0:n.title)??"",warningMessage:null==(o=e.menuItem.item)?void 0:o.warning,onClose:t[4]||(t[4]=e=>(0,a.R1)(b).hide()),onDelete:k},null,8,["variableName","warningMessage"])]})),_:1},8,["state","title"]),(0,r.bF)((0,a.R1)(m.uT),{size:"xxs",state:(0,a.R1)(v),title:`Edit ${null==(i=e.menuItem.item)?void 0:i.resourceTitle}`},{default:(0,r.k6)((()=>{var n,o,i,s;return["Collection"===(null==(n=e.menuItem.item)?void 0:n.resourceTitle)?((0,r.uX)(),(0,r.Wv)(un,{key:0,icon:(null==(o=e.menuItem.item)?void 0:o.icon)||"interface-content-folder",name:null==(i=e.menuItem.item)?void 0:i.title,onClose:t[5]||(t[5]=e=>(0,a.R1)(v).hide()),onEdit:O},null,8,["icon","name"])):((0,r.uX)(),(0,r.Wv)(dn.A,{key:1,name:(null==(s=e.menuItem.item)?void 0:s.title)??"",onClose:t[6]||(t[6]=e=>(0,a.R1)(v).hide()),onEdit:O},null,8,["name"]))]})),_:1},8,["state","title"]),(0,r.bF)((0,a.R1)(m.uT),{size:"xxs",state:(0,a.R1)(y),title:"Clear Drafts"},{default:(0,r.k6)((()=>[(0,r.bF)($t.A,{variableName:"All Drafts",warningMessage:"This action will clear all drafts. This cannot be undone.",onClose:t[7]||(t[7]=e=>(0,a.R1)(y).hide()),onDelete:_})])),_:1},8,["state"])],64)}}}),hn=(0,O.A)(pn,[["__scopeId","data-v-daf8ea6e"]]),fn=(0,a.KR)(null),mn=(0,a.KR)(null),gn=["draggable"],vn=(0,r.pM)({__name:"Draggable",props:{ceiling:{default:.8},floor:{default:.2},isDraggable:{type:Boolean,default:!0},isDroppable:{type:[Boolean,Function],default:!0},parentIds:{},id:{}},emits:["onDragEnd","onDragStart"],setup(e,{expose:t,emit:n}){const i=n,s=(0,r.EW)((()=>e.parentIds.at(-1)??null)),l=t=>{!t.dataTransfer||!(t.target instanceof HTMLElement)||!e.isDraggable||(t.target.classList.add("dragging"),t.dataTransfer.dropEffect="move",t.dataTransfer.effectAllowed="move",fn.value={id:e.id,parentId:s.value},i("onDragStart",{id:e.id,parentId:s.value}))},c=(e=>{let t=!1;return(...n)=>{t||(e(...n),t=!0,setTimeout((()=>t=!1),25))}})((t=>{var n,r;if(!fn.value||fn.value.id===e.id||e.parentIds.includes((null==(n=fn.value)?void 0:n.id)??""))return;const a=null==(r=mn.value)?void 0:r.offset,o=t.target.offsetHeight,i=e.floor*o,l=e.ceiling*o;let c=3;t.offsetY<=0&&a&&3!==a?c=a:t.offsetY<=i?c=0:t.offsetY>=l?c=1:t.offsetY>i&&t.offsetY<l&&(c=2),(t=>"function"==typeof e.isDroppable?e.isDroppable(fn.value,{id:e.id,parentId:s.value,offset:t}):e.isDroppable)(c)&&(mn.value={id:e.id,parentId:s.value,offset:c})})),u=["above","below","asChild"],d=(0,r.EW)((()=>{var t;let n="sidebar-indent-nested";return e.id===(null==(t=mn.value)?void 0:t.id)&&(n+=` dragover-${u[mn.value.offset]}`),n})),h=()=>{if(!mn.value||!fn.value)return;const e={...fn.value},t={...mn.value};fn.value=null,mn.value=null,document.querySelectorAll("div.dragging").forEach((e=>e.classList.remove("dragging"))),e.id!==t.id&&i("onDragEnd",e,t)};return t({draggingItem:fn,hoveredItem:mn}),(e,t)=>((0,r.uX)(),(0,r.CE)("div",{class:(0,o.C4)(d.value),draggable:e.isDraggable,onDragend:h,onDragover:t[0]||(t[0]=(0,p.D$)(((...e)=>(0,a.R1)(c)&&(0,a.R1)(c)(...e)),["prevent","stop"])),onDragstart:(0,p.D$)(l,["stop"])},[(0,r.RG)(e.$slots,"default",{},void 0,!0)],42,gn))}}),bn=(e,t)=>{const n=e.__vccOpts||e;for(const[e,r]of t)n[e]=r;return n},yn=bn(vn,[["__scopeId","data-v-3449bf19"]]);var wn=n(6099),On=n(1845);const kn={class:"line-clamp-1 w-full break-all pl-2 font-medium"},xn={class:"flex flex-row items-center gap-1"},En={class:"relative"},Sn={class:"flex items-start"},_n=["aria-expanded"],Cn={class:"flex h-5 max-w-[14px] items-center justify-center"},Tn={class:"flex flex-1 flex-row justify-between"},An={class:"line-clamp-1 w-full break-all text-left font-medium"},Rn={class:"relative flex h-fit justify-end"},Pn={class:"w-content bg-b-1 z-100 text-xxs text-c-1 pointer-events-none z-10 grid max-w-10 gap-1.5 rounded p-2 leading-5 shadow-lg"},Dn={class:"text-c-2 flex items-center"},Mn={class:"text-pretty break-all"},$n={key:2},Nn="hover:bg-sidebar-active-b indent-padding-left",In=(0,r.pM)({__name:"RequestSidebarItem",props:{isDraggable:{type:Boolean,default:!1},isDroppable:{type:[Boolean,Function],default:!1},parentUids:{},uid:{},menuItem:{}},emits:["onDragEnd","newTab","openMenu"],setup(e,{emit:t}){(0,p.$9)((e=>({"0ff55690":P.value,ecb155a4:D.value})));const n=t,{activeCollection:s,activeRequest:l,activeRouterParams:c,activeWorkspace:d}=(0,f.JC)(),{collections:g,tags:v,requests:b,requestExamples:y,collectionMutators:w,tagMutators:O,requestMutators:k,requestExampleMutators:x,events:E}=(0,G.hA)(),S=(0,u.rd)(),{collapsedSidebarFolders:_,toggleSidebarFolder:C}=(0,wn.c)(),{layout:T}=(0,oe.g)(),A=(0,r.EW)((()=>{var t,n,r,a;const o=g[e.uid],i=v[e.uid],s=b[e.uid],l=y[e.uid];return o?{title:(null==(t=o.info)?void 0:t.title)??"Unknown title",entity:o,resourceTitle:"Collection",children:o.children,icon:o["x-scalar-icon"],documentUrl:o.documentUrl,watchMode:o.watchMode,warning:"This cannot be undone. You’re about to delete the collection and all folders and requests inside it.",edit:(e,t)=>{w.edit(o.uid,"info.title",e),t&&w.edit(o.uid,"x-scalar-icon",t)},delete:()=>{d.value&&w.delete(o,d.value)}}:i?{title:i.name,entity:i,resourceTitle:"Tag",children:i.children,warning:"This cannot be undone. You’re about to delete the tag and all requests inside it",edit:e=>O.edit(i.uid,"name",e),delete:()=>e.parentUids[0]&&O.delete(i,e.parentUids[0])}:s?{title:s.summary??s.path,link:{name:"request",params:{workspace:null==(n=d.value)?void 0:n.uid,request:s.uid}},method:s.method,entity:s,resourceTitle:"Request",warning:"This cannot be undone. You’re about to delete the request.",children:s.examples.slice(1),edit:e=>k.edit(s.uid,"summary",e),delete:()=>e.parentUids[0]&&k.delete(s,e.parentUids[0])}:null!=l&&l.requestUid?{title:l.name,link:{name:"request.examples",params:{workspace:null==(r=d.value)?void 0:r.uid,request:l.requestUid,examples:l.uid}},method:null==(a=b[l.requestUid])?void 0:a.method,entity:l,resourceTitle:"Example",warning:"This cannot be undone. You’re about to delete the example from the request.",children:[],edit:e=>x.edit(l.uid,"name",e),delete:()=>x.delete(l)}:{title:"Unknown",entity:{uid:"",type:"unknown"},resourceTitle:"Unknown",children:[],edit:()=>null,delete:()=>null}})),R=(0,r.EW)((()=>"collection"===A.value.entity.type&&"Drafts"===A.value.title)),P=(0,r.EW)((()=>e.parentUids.length?"modal"===T?12*(e.parentUids.length-1)+"px":12*e.parentUids.length+"px":"12px")),D=(0,r.EW)((()=>e.parentUids.length?"modal"===T?12*(e.parentUids.length-1)+"px":12*e.parentUids.length+"px":"0px")),M=(0,r.EW)((()=>{var t;return _[e.uid]||(null==(t=l.value)?void 0:t.uid)===e.uid&&A.value.entity.examples.length>1})),$=(0,r.EW)((()=>{var t;return"default"===c.value[I.TU.Request]&&(null==(t=l.value)?void 0:t.uid)===e.uid})),N=(0,a.KR)(null),L=(0,r.EW)((()=>{let e=.5,t=.5;if(!N.value)return{ceiling:e,floor:t};const{draggingItem:n}=N.value;return g[null==n?void 0:n.id]||"collection"!==A.value.entity.type?"tag"===A.value.entity.type&&(e=.8,t=.2):(e=1,t=0),{ceiling:e,floor:t}})),B=(e,t)=>!("modal"===T||y[t.id]||g[e.id]);function F(){E.commandPalette.emit({commandName:"Create Request",metaData:{itemUid:e.uid,parentUid:e.parentUids[0]}})}const Q=(0,r.EW)((()=>{const{uid:e,watchModeStatus:t}=s.value||{};return e!==A.value.entity.uid?"text-c-3":"WATCHING"===t?"text-c-1":"ERROR"===t?"text-red":"text-c-3"})),U=(0,r.EW)((()=>"Drafts"==A.value.title&&"modal"!==T&&A.value.children.length>0)),z=(0,r.EW)((()=>{const t=b[e.uid];if(t)return!(0,i.pH)(t);const n=v[e.uid];return!n||!(0,i.pH)(n)}));return(e,t)=>{const i=(0,r.g2)("RequestSidebarItem",!0);return z.value?((0,r.uX)(),(0,r.CE)("li",{key:0,class:(0,o.C4)(["relative flex flex-row",["modal"===(0,a.R1)(T)&&e.parentUids.length>1||"modal"!==(0,a.R1)(T)&&e.parentUids.length?"before:bg-border before:z-1 indent-border-line-offset mb-[.5px] before:pointer-events-none before:absolute before:left-[calc(.75rem_+_.5px)] before:top-0 before:h-[calc(100%_+_.5px)] before:w-[.5px] last:mb-0 last:before:h-full":""]])},[(0,r.bF)((0,a.R1)(yn),{id:A.value.entity.uid,ref_key:"draggableRef",ref:N,ceiling:L.value.ceiling,class:"gap-1/2 flex flex-1 flex-col text-sm",floor:L.value.floor,isDraggable:e.isDraggable,isDroppable:e.isDroppable,parentIds:e.parentUids,onOnDragEnd:t[9]||(t[9]=(...t)=>e.$emit("onDragEnd",...t))},{default:(0,r.k6)((()=>{var s;return[A.value.link?((0,r.uX)(),(0,r.Wv)((0,a.R1)(u.Wk),{key:0,class:"group no-underline",to:A.value.link,onClick:t[1]||(t[1]=(0,p.D$)((e=>((e,t)=>{e&&((0,On.RN)(["default"]).some((t=>e[t]))?n("newTab",t.title||"",t.entity.uid):t.link&&S.push(t.link))})(e,A.value)),["prevent"]))},{default:(0,r.k6)((({isExactActive:n})=>{var i,s;return[(0,r.Lk)("div",{class:(0,o.C4)(["relative flex min-h-8 w-full cursor-pointer flex-row items-start justify-between gap-0.5 rounded py-1.5 pr-2",[Nn,n||$.value?"bg-sidebar-active-b text-sidebar-active-c transition-none":"text-sidebar-c-2"]])},[(0,r.Lk)("span",kn,(0,o.v_)(A.value.title),1),(0,r.Lk)("div",xn,[(0,r.Lk)("div",En,["modal"!==(0,a.R1)(T)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.RO),{key:0,class:(0,o.C4)(["hover:bg-b-3 hidden aspect-square h-fit px-0.5 py-0 opacity-0 group-hover:flex group-hover:opacity-100 group-focus-visible:opacity-100 group-has-[:focus-visible]:opacity-100",{flex:(null==(s=null==(i=e.menuItem)?void 0:i.item)?void 0:s.entity.uid)===A.value.entity.uid&&e.menuItem.open}]),size:"sm",type:"button",variant:"ghost",onClick:t[0]||(t[0]=(0,p.D$)((t=>e.$emit("openMenu",{item:A.value,parentUids:e.parentUids,targetRef:t.currentTarget,open:!e.menuItem.open})),["stop","prevent"]))},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.kv),{icon:"Ellipses",size:"md"})])),_:1},8,["class"])):(0,r.Q3)("",!0)]),(0,r.Lk)("span",Sn,[t[10]||(t[10]=(0,r.eW)("   ")),t[11]||(t[11]=(0,r.Lk)("span",{class:"sr-only"},"HTTP Method:",-1)),A.value.method?((0,r.uX)(),(0,r.Wv)((0,a.R1)(h.A),{key:0,class:"font-bold",method:A.value.method},null,8,["method"])):(0,r.Q3)("",!0)])])],2)]})),_:1},8,["to"])):"modal"!==(0,a.R1)(T)||e.parentUids.length?((0,r.uX)(),(0,r.CE)("button",{key:1,"aria-expanded":!!(0,a.R1)(_)[A.value.entity.uid],class:(0,o.C4)(["hover:bg-b-2 group relative flex w-full flex-row justify-start gap-1.5 rounded p-1.5 focus-visible:z-10",[Nn]]),type:"button",onClick:t[4]||(t[4]=e=>(0,a.R1)(C)(A.value.entity.uid))},[(0,r.Lk)("span",Cn,[(0,r.RG)(e.$slots,"leftIcon",{},(()=>[(0,r.bF)((0,a.R1)(m.SL),{class:"text-c-3 hover:text-c-1 shrink-0",open:!!(0,a.R1)(_)[A.value.entity.uid]},null,8,["open"])]),!0),t[12]||(t[12]=(0,r.eW)("   "))]),(0,r.Lk)("div",Tn,[(0,r.Lk)("span",An,(0,o.v_)(A.value.title),1),(0,r.Lk)("div",Rn,[(0,r.Lk)("div",{class:(0,o.C4)(["items-center gap-px opacity-0 group-hover:flex group-hover:opacity-100 group-focus-visible:opacity-100 group-has-[:focus-visible]:opacity-100",{flex:e.menuItem.open,hidden:!e.menuItem.open||(null==(s=e.menuItem.item)?void 0:s.entity.uid)!==A.value.entity.uid}])},["modal"!==(0,a.R1)(T)&&!R.value||R.value&&U.value?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.RO),{key:0,class:"hover:bg-b-3 hover:text-c-1 aspect-square h-fit px-0.5 py-0 group-focus-visible:opacity-100 group-has-[:focus-visible]:opacity-100",size:"sm",variant:"ghost",onClick:t[2]||(t[2]=(0,p.D$)((t=>e.$emit("openMenu",{item:A.value,parentUids:e.parentUids,targetRef:t.currentTarget.parentNode,open:!0})),["stop","prevent"]))},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.kv),{icon:"Ellipses",size:"md"})])),_:1})):(0,r.Q3)("",!0),"modal"!==(0,a.R1)(T)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.RO),{key:1,class:"hover:bg-b-3 hover:text-c-1 aspect-square h-fit px-0.5 py-0 group-focus-visible:opacity-100 group-has-[:focus-visible]:opacity-100",size:"sm",variant:"ghost",onClick:t[3]||(t[3]=(0,p.D$)((e=>F()),["stop","prevent"]))},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.kv),{icon:"Add",size:"md",thickness:"2"})])),_:1})):(0,r.Q3)("",!0)],2),A.value.watchMode?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.iy),{key:0,side:"right",sideOffset:12},{trigger:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.kv),{class:(0,o.C4)(["ml-0.5 text-sm",Q.value]),icon:"Watch",size:"md",thickness:"2"},null,8,["class"])])),content:(0,r.k6)((()=>[(0,r.Lk)("div",Pn,[(0,r.Lk)("div",Dn,[(0,r.Lk)("p",Mn," Watching: "+(0,o.v_)(A.value.documentUrl),1)])])])),_:1})):(0,r.Q3)("",!0),t[13]||(t[13]=(0,r.Lk)("span",null," ",-1))])])],10,_n)):(0,r.Q3)("",!0),M.value?((0,r.uX)(),(0,r.CE)("ul",$n,[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(A.value.children,(n=>((0,r.uX)(),(0,r.Wv)(i,{key:n,isDraggable:!(0,a.R1)(y)[n],isDroppable:B,menuItem:e.menuItem,parentUids:[...e.parentUids,e.uid],uid:n,onNewTab:t[5]||(t[5]=(t,n)=>e.$emit("newTab",t,n)),onOnDragEnd:t[6]||(t[6]=(...t)=>e.$emit("onDragEnd",...t)),onOpenMenu:t[7]||(t[7]=t=>e.$emit("openMenu",t))},null,8,["isDraggable","menuItem","parentUids","uid"])))),128)),0===A.value.children.length?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.RO),{key:0,class:(0,o.C4)(["text-c-1 hover:bg-b-2 flex h-8 w-full justify-start gap-1.5 py-0 text-xs",e.parentUids.length?"pl-9":""]),variant:"ghost",onClick:t[8]||(t[8]=e=>F())},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.kv),{icon:"Add",size:"sm"}),t[14]||(t[14]=(0,r.Lk)("span",null,"Add Request",-1))])),_:1},8,["class"])):(0,r.Q3)("",!0)])):(0,r.Q3)("",!0)]})),_:3},8,["id","ceiling","floor","isDraggable","isDroppable","parentIds"])],2)):(0,r.Q3)("",!0)}}}),Ln=(0,O.A)(In,[["__scopeId","data-v-682d438e"]]);var Bn=n(3858);const Fn={class:"flex w-[inherit] items-center text-sm"},Qn={class:"m-0 flex items-center gap-1.5 font-bold"},Un={class:"line-clamp-1 text-left"},zn={class:"overflow-hidden text-ellipsis"},jn={class:"flex h-4 w-4 items-center justify-center"},Wn=(0,r.pM)({__name:"WorkspaceDropdown",setup(e){const{activeWorkspace:t}=(0,f.JC)(),{workspaces:n,workspaceMutators:i,events:s}=(0,G.hA)(),{push:l}=(0,u.rd)(),c=(0,r.EW)((()=>1===Object.keys(n).length)),d=()=>s.commandPalette.emit({commandName:"Create Workspace"}),h=(0,a.KR)(""),g=(0,a.KR)(""),v=(0,m.hS)(),b=(0,m.hS)(),y=e=>{const t=n[e];t&&(h.value=t.name,g.value=e,v.show())},w=e=>{e.trim()&&(i.edit(g.value,"name",e.trim()),v.hide())},O=e=>{const t=n[e];t&&(h.value=t.name,g.value=e,b.show())},k=async()=>{var e;if(!c.value){const r=(null==(e=t.value)?void 0:e.uid)===g.value,a={...n};if(delete a[g.value],i.delete(g.value),r){const e=Object.keys(a)[0];await l({name:"workspace",params:{workspace:e}})}}b.hide()};return(e,i)=>((0,r.uX)(),(0,r.CE)("div",null,[(0,r.Lk)("div",Fn,[(0,r.bF)((0,a.R1)(m.SM),null,{items:(0,r.k6)((()=>[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)((0,a.R1)(n),((e,n)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.rA),{key:n,class:"group/item flex w-full items-center gap-1.5 overflow-hidden text-ellipsis whitespace-nowrap",onClick:(0,p.D$)((n=>(e=>{var n;e!==(null==(n=t.value)?void 0:n.uid)&&l({name:"workspace",params:{workspace:e}})})(e.uid)),["stop"])},{default:(0,r.k6)((()=>{var s;return[(0,r.bF)((0,a.R1)(m.Pq),{selected:(null==(s=(0,a.R1)(t))?void 0:s.uid)===n},null,8,["selected"]),(0,r.Lk)("span",zn,(0,o.v_)(e.name),1),(0,r.bF)((0,a.R1)(m.SM),{placement:"right-start",teleport:""},{items:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.rA),{class:"flex gap-2",onMousedown:t=>y(e.uid),onTouchend:(0,p.D$)((t=>y(e.uid)),["prevent"])},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.kv),{class:"inline-flex",icon:"Edit",size:"md",thickness:"1.5"}),i[4]||(i[4]=(0,r.Lk)("span",null,"Rename",-1))])),_:2},1032,["onMousedown","onTouchend"]),c.value?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.iy),{key:0,class:"z-overlay",side:"bottom"},{trigger:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.rA),{class:"flex w-full gap-2",disabled:"",onMousedown:i[0]||(i[0]=(0,p.D$)((()=>{}),["prevent"])),onTouchend:i[1]||(i[1]=(0,p.D$)((()=>{}),["prevent"]))},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.kv),{class:"inline-flex",icon:"Delete",size:"md",thickness:"1.5"}),i[5]||(i[5]=(0,r.Lk)("span",null,"Delete",-1))])),_:1})])),content:(0,r.k6)((()=>i[6]||(i[6]=[(0,r.Lk)("div",{class:"w-content bg-b-1 text-xxs text-c-1 pointer-events-none z-10 grid min-w-48 gap-1.5 rounded p-2 leading-5 shadow-lg"},[(0,r.Lk)("div",{class:"text-c-2 flex items-center"},[(0,r.Lk)("span",null,"Only workspace cannot be deleted.")])],-1)]))),_:1})):((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.rA),{key:1,class:"flex !gap-2",onMousedown:(0,p.D$)((t=>O(e.uid)),["prevent"]),onTouchend:(0,p.D$)((t=>O(e.uid)),["prevent"])},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.kv),{class:"inline-flex",icon:"Delete",size:"sm",thickness:"1.5"}),i[7]||(i[7]=(0,r.Lk)("span",null,"Delete",-1))])),_:2},1032,["onMousedown","onTouchend"]))])),default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.RO),{class:"hover:bg-b-3 -mr-1 ml-auto aspect-square h-fit px-0.5 py-0 group-hover/item:flex",size:"sm",type:"button",variant:"ghost"},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.kv),{icon:"Ellipses",size:"sm"})])),_:1})])),_:2},1024)]})),_:2},1032,["onClick"])))),128)),(0,r.bF)((0,a.R1)(m.t)),(0,r.bF)((0,a.R1)(m.rA),{class:"flex items-center gap-1.5",onClick:d},{default:(0,r.k6)((()=>[(0,r.Lk)("div",jn,[(0,r.bF)((0,a.R1)(m.kv),{icon:"Add",size:"sm"})]),i[8]||(i[8]=(0,r.Lk)("span",null,"Create Workspace",-1))])),_:1})])),default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(m.RO),{class:"text-c-1 hover:bg-b-2 line-clamp-1 h-full w-fit justify-start px-1.5 py-1.5 font-normal",fullWidth:"",variant:"ghost"},{default:(0,r.k6)((()=>{var e;return[(0,r.Lk)("div",Qn,[(0,r.Lk)("h2",Un,(0,o.v_)(null==(e=(0,a.R1)(t))?void 0:e.name),1)])]})),_:1})])),_:1})]),(0,r.bF)((0,a.R1)(m.uT),{size:"xxs",state:(0,a.R1)(b),title:"Delete workspace"},{default:(0,r.k6)((()=>[(0,r.bF)($t.A,{variableName:h.value,warningMessage:"This cannot be undone. You’re about to delete the workspace and everything inside it.",onClose:i[2]||(i[2]=e=>(0,a.R1)(b).hide()),onDelete:k},null,8,["variableName"])])),_:1},8,["state"]),(0,r.bF)((0,a.R1)(m.uT),{size:"xxs",state:(0,a.R1)(v),title:"Rename Workspace"},{default:(0,r.k6)((()=>[(0,r.bF)(dn.A,{name:h.value,onClose:i[3]||(i[3]=e=>(0,a.R1)(v).hide()),onEdit:w},null,8,["name"])])),_:1},8,["state"])]))}}),Xn={class:"bg-b-1 sticky top-0 z-20 flex h-12 items-center px-3"},qn={key:1,class:"text-c-3"},Vn={class:"search-button-fade sticky top-12 z-10 px-3 py-2.5 pt-0 focus-within:z-20",role:"search"},Hn={key:1,class:"contents"},Zn={class:"empty-sidebar-item-content px-2.5 py-2.5"},Gn={class:"rabbit-ascii relative m-auto mt-2 h-[68px] w-[60px]"},Kn=(0,r.pM)({__name:"RequestSidebar",props:{isSidebarOpen:{type:Boolean}},emits:["update:modelValue","update:isSidebarOpen","newTab","clearDrafts"],setup(e,{emit:t}){const n=e,s=t,{layout:l}=(0,oe.g)(),d=(0,G.hA)(),{activeWorkspaceCollections:g,activeRequest:v,activeWorkspaceRequests:b,activeWorkspace:y}=(0,f.JC)(),{findRequestParents:w,events:O,requestMutators:k,requests:x}=d,{handleDragEnd:E,isDroppable:S}=function(e,{collections:t,collectionMutators:n,tags:r,tagMutators:a,workspaceMutators:o}){const{layout:i}=(0,oe.g)(),s=(e,t)=>{"collection"===e.type?n.edit(e.uid,"children",t):"tag"===e.type&&a.edit(e.uid,"children",t)};return{handleDragEnd:(i,l)=>{var c,u,d,p;if(!i||!l)return;const{id:h,parentId:f}=i,{id:m,parentId:g,offset:v}=l;if(f?t[f]?n.edit(f,"children",t[f].children.filter((e=>e!==h))):r[f]&&a.edit(f,"children",r[f].children.filter((e=>e!==h))):o.edit(null==(c=e.value)?void 0:c.uid,"collections",(null==(u=e.value)?void 0:u.collections.filter((e=>e!==h)))??[]),2===v){const e=t[m]||r[m];e&&s(e,[...e.children??[],h])}else if(g){const e=t[g]||r[g];if(!e)return;const n=[...e.children??[]],a=n.findIndex((e=>m===e))??0;n.splice(a+v,0,h),s(e,n)}else{const t=[...(null==(d=e.value)?void 0:d.collections)??[]],n=t.findIndex((e=>m===e))??0;t.splice(n+v,0,h),o.edit(null==(p=e.value)?void 0:p.uid,"collections",t)}},isDroppable:(e,n)=>{var r,a;return!("modal"===i||!t[e.id]&&2!==n.offset||t[e.id]&&"Drafts"===(null==(a=null==(r=t[n.id])?void 0:r.info)?void 0:a.title))}}}(y,d),{collapsedSidebarFolders:_,setCollapsedSidebarFolder:C}=(0,wn.c)(),{replace:T}=(0,u.rd)(),A=()=>{O.commandPalette.emit({commandName:"Import from OpenAPI/Swagger/Postman"})},R=(0,r.Bi)(),{toast:P}=(0,c.Yf)(),D=(0,a.Kh)({open:!1}),M=(0,a.KR)(!1);(0,r.wB)(v,(e=>{e&&w(e).forEach((e=>C(e,!0)))}),{immediate:!0});const{searchText:$,searchResultsWithPlaceholderResults:N,selectedSearchResult:L,onSearchResultClick:B,fuseSearch:F,searchInputRef:Q,searchResultRefs:U,navigateSearchResults:z,selectSearchResult:j}=function(){const e=(0,u.rd)(),{activeWorkspace:t,activeWorkspaceRequests:n,activeWorkspaceCollections:o}=(0,f.JC)(),{requests:s,tags:l}=(0,G.hA)(),c=(0,a.KR)([]),d=(0,a.KR)([]),p=(0,a.KR)(0),h=(0,a.KR)(""),m=(0,a.KR)(null),g=(0,a.KR)([]),v=new Pt.A(c.value,{keys:["title","description","body"]}),b=n=>{c.value=n.filter((e=>!(0,i.pH)(e))).filter((e=>{var t;const n=null==(t=o.value)?void 0:t.find((t=>t.requests.includes(e.uid)));return!(null!=n&&n.tags.map((e=>l[e])).filter(i.O9).filter((t=>{var n;return null==(n=e.tags)?void 0:n.includes(t.name)})).filter((e=>(0,i.pH)(e))).length)})).map((n=>{var r,a;return{id:n.uid,title:n.summary??n.method,description:n.description??"",httpVerb:n.method,path:n.path,link:null==(a=null==e?void 0:e.resolve({name:"request",params:{[I.TU.Request]:n.uid,[I.TU.Workspace]:null==(r=t.value)?void 0:r.uid}}))?void 0:a.href}})),v.setCollection(c.value)},y=()=>{p.value=0,d.value=v.search(h.value)};(0,r.wB)(h,(e=>{e.length?y():d.value=[]}));const w=(0,r.EW)((()=>n.value.map((e=>s[e])).filter(i.O9)));(0,r.wB)(n,(()=>{b(w.value)}),{immediate:!0});const O=t=>{e.push(t.item.id),h.value="",p.value=0,d.value=[],m.value instanceof HTMLInputElement&&m.value.blur()},k=(0,r.EW)((()=>0===h.value.length?c.value.map((e=>({item:e}))):d.value));return{searchText:h,searchResultsWithPlaceholderResults:k,selectedSearchResult:p,onSearchResultClick:O,fuseSearch:y,searchInputRef:m,searchResultRefs:g,navigateSearchResults:e=>{const t="up"===e?-1:1,n=k.value.length;p.value=(p.value+t+n)%n,(0,r.dY)((()=>{const e=g.value[p.value];e instanceof HTMLElement&&e.scrollIntoView({behavior:"smooth",block:"center"})}))},selectSearchResult:()=>{if(p.value>=0){const e=k.value[p.value];e&&O(e)}},populateFuseDataArray:b}}(),W=e=>{var t;e&&(e.toggleSidebar&&s("update:isSidebarOpen",n.isSidebarOpen),e.focusRequestSearch&&(null==(t=Q.value)||t.focus()))};(0,r.sV)((()=>O.hotKeys.on(W))),(0,r.xo)((()=>{O.hotKeys.off(W)}));const X=e=>{if(null!=e&&e.documentUrl){e.watchMode=!e.watchMode;const t=g.value.find((t=>t.uid===e.entity.uid));t&&(t.watchMode=e.watchMode)}};(0,r.wB)((()=>g.value.map((e=>e.watchMode))),((e,t)=>{e.forEach(((e,n)=>{var r,a,o;if("modal"!==l&&e!==t[n]&&"Drafts"!==(null==(a=null==(r=g.value[n])?void 0:r.info)?void 0:a.title)&&g.value[n]){const t=g.value[n];if(!t)return;const r=`${null==(o=t.info)?void 0:o.title}: Watch Mode ${e?"enabled":"disabled"}`;P(r,"info")}}))}));const q=(0,r.EW)((()=>{var e;const t=N.value[L.value];return null!=(e=null==t?void 0:t.item)&&e.id?`#search-input-${t.item.id}`:void 0})),V=()=>{const e=g.value.find((e=>{var t;return"Drafts"===(null==(t=e.info)?void 0:t.title)}));if(e&&e.requests.forEach((t=>{x[t]&&k.delete(x[t],e.uid)})),b.value.length){const e=g.value[0],t=null==e?void 0:e.requests[0];t&&T({name:"request",params:{[I.TU.Request]:t}})}else{const{request:t}=(0,Mt.AJ)();e&&(k.add(t,e.uid),T({name:"request",params:{[I.TU.Request]:t.uid}}))}},H=()=>{M.value||($.value=""),M.value&&(0,r.dY)((()=>{var e;null==(e=Q.value)||e.focus()})),M.value=!M.value},Z=(0,r.EW)((()=>((e,t,n)=>{var r;const a=e.find((e=>{var t;return"Drafts"===(null==(t=e.info)?void 0:t.title)})),o=1===t.length;if(!t[0])return!1;const i=null==a?void 0:a.requests.includes(t[0]);if(!i)return!1;const s="My First Request"!==(null==(r=n[(null==a?void 0:a.requests[0])??""])?void 0:r.summary);return o&&i&&!s})(g.value,b.value,x)));return(e,t)=>((0,r.uX)(),(0,r.CE)(r.FK,null,[(0,r.bo)((0,r.bF)((0,a.R1)(Bn.A),{class:(0,o.C4)([e.isSidebarOpen?"sidebar-active-width":""]),isSidebarOpen:e.isSidebarOpen,"onUpdate:isSidebarOpen":t[9]||(t[9]=t=>e.$emit("update:isSidebarOpen",t))},(0,r.eX)({content:(0,r.k6)((()=>[(0,r.Lk)("div",Xn,[(0,r.bF)(re,{class:(0,o.C4)(["xl:hidden",[{"!flex":"modal"===(0,a.R1)(l)}]]),modelValue:e.isSidebarOpen,"onUpdate:modelValue":t[0]||(t[0]=t=>e.$emit("update:isSidebarOpen",t))},null,8,["class","modelValue"]),"modal"!==(0,a.R1)(l)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(Wn),{key:0})):(0,r.Q3)("",!0),"modal"!==(0,a.R1)(l)?((0,r.uX)(),(0,r.CE)("span",qn," / ")):(0,r.Q3)("",!0),"modal"!==(0,a.R1)(l)?((0,r.uX)(),(0,r.Wv)(Rt,{key:2})):(0,r.Q3)("",!0),(0,r.Lk)("button",{class:"ml-auto",type:"button",onClick:H},[(0,r.bF)((0,a.R1)(m.kv),{class:"text-c-3 hover:bg-b-2 p-1.75 max-h-8 max-w-8 rounded-lg text-sm",icon:"Search"})])]),(0,r.bo)((0,r.Lk)("div",Vn,[(0,r.bF)((0,a.R1)(m.vW),{ref_key:"searchInputRef",ref:Q,modelValue:(0,a.R1)($),"onUpdate:modelValue":t[1]||(t[1]=e=>(0,a.i9)($)?$.value=e:null),"aria-activedescendant":q.value,"aria-controls":(0,a.R1)(R),sidebar:"",onInput:(0,a.R1)(F),onKeydown:[t[2]||(t[2]=(0,p.jR)((0,p.D$)((e=>(0,a.R1)(z)("down")),["stop"]),["down"])),t[3]||(t[3]=(0,p.jR)((0,p.D$)((e=>(0,a.R1)(j)()),["stop"]),["enter"])),t[4]||(t[4]=(0,p.jR)((0,p.D$)((e=>(0,a.R1)(z)("up")),["stop"]),["up"]))]},null,8,["modelValue","aria-activedescendant","aria-controls","onInput"])],512),[[p.aG,M.value]]),(0,r.Lk)("div",{class:(0,o.C4)(["gap-1/2 flex flex-1 flex-col overflow-visible overflow-y-auto px-3 pb-3 pt-0",[{"pb-14":"modal"!==(0,a.R1)(l)},{"h-[calc(100%-273.5px)]":Z.value}]]),onDragenter:t[7]||(t[7]=(0,p.D$)((()=>{}),["prevent"])),onDragover:t[8]||(t[8]=(0,p.D$)((()=>{}),["prevent"]))},[(0,a.R1)($)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.oN),{key:0,id:(0,a.R1)(R),"aria-label":"Search Results",class:"gap-px",noResults:!(0,a.R1)(N).length},{default:(0,r.k6)((()=>[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)((0,a.R1)(N),((e,n)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.rQ),{id:`#search-input-${e.item.id}`,key:e.refIndex,ref_for:!0,ref:e=>(0,a.R1)(U)[n]=e,active:(0,a.R1)(L)===n,class:"px-2",href:e.item.link,onClick:(0,p.D$)((t=>(0,a.R1)(B)(e)),["prevent"]),onFocus:e=>L.value=n},{addon:(0,r.k6)((()=>[t[11]||(t[11]=(0,r.Lk)("span",{class:"sr-only"},"HTTP Method:",-1)),(0,r.bF)(h.A,{class:"font-bold",method:e.item.httpVerb??"get"},null,8,["method"])])),default:(0,r.k6)((()=>[(0,r.eW)((0,o.v_)(e.item.title)+" ",1)])),_:2},1032,["id","active","href","onClick","onFocus"])))),128))])),_:1},8,["id","noResults"])):((0,r.uX)(),(0,r.CE)("nav",Hn,[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)((0,a.R1)(g),(e=>{var n;return(0,r.uX)(),(0,r.Wv)(Ln,{key:e.uid,isDraggable:"modal"!==(0,a.R1)(l)&&"Drafts"!==(null==(n=e.info)?void 0:n.title),isDroppable:(0,a.R1)(S),menuItem:D,parentUids:[],uid:e.uid,onNewTab:t[5]||(t[5]=(e,t)=>s("newTab",{name:e,uid:t})),onOnDragEnd:(0,a.R1)(E),onOpenMenu:t[6]||(t[6]=e=>Object.assign(D,e))},{leftIcon:(0,r.k6)((()=>{var t;return["Drafts"===(null==(t=e.info)?void 0:t.title)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.kv),{key:0,class:"text-sidebar-c-2 group-hover:hidden",icon:"Scribble",thickness:"2.25"})):((0,r.uX)(),(0,r.Wv)((0,a.R1)(Ct.$n),{key:1,class:"text-sidebar-c-2 size-3.5 min-w-3.5 stroke-2 group-hover:hidden",src:e["x-scalar-icon"]||"interface-content-folder"},null,8,["src"])),(0,r.Lk)("div",{class:(0,o.C4)({"rotate-90":(0,a.R1)(_)[e.uid]})},[(0,r.bF)((0,a.R1)(m.kv),{class:"text-c-3 hover:text-c-1 hidden text-sm group-hover:block",icon:"ChevronRight",size:"md"})],2)]})),_:2},1032,["isDraggable","isDroppable","menuItem","uid","onOnDragEnd"])})),128))]))],34)])),button:(0,r.k6)((()=>[(0,r.Lk)("div",{class:(0,o.C4)({"empty-sidebar-item":Z.value})},[(0,r.Lk)("div",Zn,[(0,r.Lk)("div",Gn,[(0,r.bF)(ze.A,{art:(0,a.R1)('         ,\\\n         \\\\\\,_\n          \\` ,\\\n     __,.-" =__)\n   ."        )\n,_/   ,    \\/\\_\n\\_| )_-\\ \\_-`\n   `-----` `--`'),class:"rabbitsit font-bold"},null,8,["art"]),(0,r.bF)(ze.A,{art:(0,a.R1)('         __\n        // \\,_\n          \\` ,\\\n     __,.-" =__)\n   ."        )\n,_/   ,    \\/\\ \n\\_| // /  /  /\n   /  /    '),class:"rabbitjump absolute left-0 top-0 font-bold"},null,8,["art"])]),t[12]||(t[12]=(0,r.Lk)("div",{class:"mb-2 mt-2 text-balance text-center text-sm"},[(0,r.Lk)("b",{class:"font-medium"},"Let's Get Started"),(0,r.Lk)("p",{class:"mt-2"}," Create request, folder, collection or import from OpenAPI/Postman ")],-1))]),"modal"!==(0,a.R1)(l)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(m.RO),{key:0,class:(0,o.C4)(["mb-1.5 hidden h-fit w-full p-1.5 opacity-0",{"flex opacity-100":Z.value}]),onClick:A},{default:(0,r.k6)((()=>t[13]||(t[13]=[(0,r.eW)(" Import Collection ")]))),_:1},8,["class"])):(0,r.Q3)("",!0),"modal"!==(0,a.R1)(l)?((0,r.uX)(),(0,r.Wv)(Dt.A,{key:1,click:(0,a.R1)(O).commandPalette.emit,hotkey:"K"},{title:(0,r.k6)((()=>t[14]||(t[14]=[(0,r.eW)(" Add Item ")]))),_:1},8,["click"])):(0,r.Q3)("",!0)],2)])),_:2},["modal"!==(0,a.R1)(l)?{name:"header",fn:(0,r.k6)((()=>[])),key:"0"}:void 0]),1032,["class","isSidebarOpen"]),[[p.aG,e.isSidebarOpen]]),"modal"!==(0,a.R1)(l)&&D?((0,r.uX)(),(0,r.Wv)(hn,{key:0,menuItem:D,onClearDrafts:V,onCloseMenu:t[10]||(t[10]=e=>D.open=!1),onToggleWatchMode:X},null,8,["menuItem"])):(0,r.Q3)("",!0)],64))}}),Yn=(0,O.A)(Kn,[["__scopeId","data-v-b7bf7c07"]]);var Jn=n(7200);var er=n(165),tr=n(4851),nr=n(6736),rr=n(6349);const ar={class:"flex h-full"},or={key:1,class:"flex h-full flex-1 flex-col"},ir=(0,r.pM)({__name:"Request",emits:["newTab"],setup(e){const t=(0,G.hA)(),{toast:n}=(0,c.Yf)(),{layout:p}=(0,oe.g)(),{activeCollection:h,activeExample:m,activeEnvironment:g,activeRequest:v,activeWorkspace:b,activeServer:y,activeEnvVariables:w,activeWorkspaceCollections:O,activeWorkspaceRequests:S}=(0,f.JC)(),{cookies:C,modalState:T,requestHistory:A,showSidebar:R,securitySchemes:P,requestMutators:D,serverMutators:M,servers:$,events:N}=t,L=(0,a.KR)("modal"!==p),B=(0,a.KR)(),F=(0,a.KR)(),Q=(0,a.KR)(""),U=(0,u.rd)(),z=(0,r.EW)((()=>A.findLast((e=>{var t;return e.request.uid===(null==(t=m.value)?void 0:t.uid)})))),{mediaQueries:j}=(0,l.G)();(0,r.wB)(j.xl,(e=>L.value=e),{immediate:"modal"!==p});const W=(0,r.EW)((()=>{var e,t;return("modal"===p?null==(e=h.value)?void 0:e.selectedSecuritySchemeUids:null==(t=v.value)?void 0:t.selectedSecuritySchemeUids)??[]})),X=async()=>{var e,t,r,a;if(!v.value||!m.value||!h.value)return;const o="object"==typeof g.value?g.value.value:"{}",l=s.parse(o);l.error&&console.error("INVALID ENVIRONMENT!");const c=l.error||"object"!=typeof l.data?{}:l.data??{},u=(null==(e=b.value)?void 0:e.cookies.map((e=>C[e])).filter(i.O9))??[],d="Drafts"===(null==(r=null==(t=h.value)?void 0:t.info)?void 0:r.title)?void 0:y.value,[p,f]=(({environment:e,example:t,globalCookies:n,proxyUrl:r,request:a,securitySchemes:o,selectedSecuritySchemeUids:s=[],server:l,status:c})=>{try{const u=e??{},d=new AbortController,p=t.parameters.path.reduce(((e,t)=>(t.enabled&&(e[t.key]=(0,er.GV)(t.value,u)),e)),{}),h=(0,er.GV)((null==l?void 0:l.url)??"",u),f=(0,er.GV)(a.path,p);let m=h||f;if(!m)throw Jn.Sr.URL_EMPTY;Object.entries((null==l?void 0:l.variables)??{}).forEach((([e,t])=>{m=(0,er.GV)(m,{[e]:p[e]||t.default})}));const g=function(e,t){const n=new URLSearchParams;return e.parameters.query.forEach((e=>{e.enabled&&("array"===e.type?(0,er.GV)(e.value??"",t).split(","):[(0,er.GV)(e.value??"",t)]).forEach((t=>n.append(e.key,t.trim())))})),n}(t,u),v=function(e,t){const n={};return e.parameters.headers.forEach((e=>{const r=e.key.trim().toLowerCase();e.enabled&&("content-type"!==r||"multipart/form-data"!==e.value)&&(n[r]=(0,er.GV)(e.value,t))})),n}(t,u),{body:b}=function(e,t,n){var r,a,o;if(!(0,i.Ik)(e))return{body:void 0,contentType:void 0};if("formData"===t.body.activeBody&&t.body.formData){const e="form-data"===t.body.formData.encoding?"multipart/form-data":"application/x-www-form-urlencoded",r="form-data"===t.body.formData.encoding?new FormData:new URLSearchParams;return t.body.formData.value.forEach((e=>{!e.enabled||!e.key||(e.file&&r instanceof FormData?r.append(e.key,e.file,e.file.name):void 0!==e.value&&r.append(e.key,(0,er.GV)(e.value,n)))})),{body:r,contentType:e}}return"raw"===t.body.activeBody?{body:(0,er.GV)((null==(r=t.body.raw)?void 0:r.value)??"",n),contentType:null==(a=t.body.raw)?void 0:a.encoding}:"binary"===t.body.activeBody?{body:t.body.binary,contentType:null==(o=t.body.binary)?void 0:o.type}:{body:void 0,contentType:void 0}}(a.method,t,u),{cookieParams:y}=(0,nr.hF)({example:t,env:u,globalCookies:n,serverUrl:m,proxyUrl:r}),w=s.flat().map((e=>o[e])).filter(i.O9),O=(0,rr.A)(w,u),k={...Object.entries(O.headers).reduce(((e,[t,n])=>(e[t.toLowerCase()]=n,e)),{}),...v},x=[...y,...O.cookies],E=new URLSearchParams([...g,...O.urlParams]);m=(0,i.Hh)(m,f,E);const S=(0,er.GV)((0,nr.a2)(x,k.Cookie),u);S&&((0,i.p3)(r,m)?(console.warn("We’re using a `X-Scalar-Cookie` custom header to the request. The proxy will forward this as a `Cookie` header. We do this to avoid the browser omitting the `Cookie` header for cross-origin requests for security reasons."),k["X-Scalar-Cookie"]=S):(console.warn("We’re trying to add a Cookie header, but browsers often omit them for cross-origin requests for various security reasons. If it’s not working, that’s probably why. Here are the requirements for it to work:\n\n          - The browser URL must be on the same domain as the server URL.\n          - The connection must be made over HTTPS.\n          "),k.Cookie=S));const _=new URLSearchParams([["scalar_url",m.toString()]]),C=(0,i.p3)(r,m)?`${r}?${_.toString()}`:m,T=new Request(C,{method:a.method.toUpperCase(),body:b??null,headers:k});return[null,{request:T,sendRequest:async()=>{null==c||c.emit("start");const e=Date.now();try{const n=await fetch(T,{signal:d.signal});null==c||c.emit("stop");const o=((e,t=!1)=>{var n,r;const a=Object.fromEntries(e);t&&["Access-Control-Allow-Credentials","Access-Control-Allow-Headers","Access-Control-Allow-Methods","Access-Control-Allow-Origin","Access-Control-Expose-Headers"].map((e=>e.toLowerCase())).forEach((e=>delete a[e]));const o=Object.keys(a).find((e=>"x-scalar-modified-headers"===e.toLowerCase())),i=o?(null==(r=null==(n=a[o])?void 0:n.toString().split(", "))?void 0:r.map((e=>e.toLowerCase())))??[]:[];return Object.keys(a).forEach((e=>{i.includes(e.toLowerCase())&&delete a[e]})),o&&delete a[o],Object.keys(a).forEach((e=>{const t=/^x-scalar-original-/i;if(t.test(e)){const n=e.replace(t,"");a[e]&&(a[n]=a[e],delete a[e])}})),Object.keys(a).forEach((e=>{const t=(e=>e.split("-").map((e=>e.charAt(0).toUpperCase()+e.toLowerCase().slice(1))).join("-"))(e);e!==t&&a[e]&&(a[t]=a[e],delete a[e])})),Object.fromEntries(Object.entries(a).sort((([e],[t])=>e.localeCompare(t))))})(n.headers,(0,i.p3)(r,m)),s=n.headers.get("content-type")??"text/plain;charset=UTF-8",l=await n.arrayBuffer(),u=function(e,t){const n=new tr(t);return ge.includes(n.essence)?new TextDecoder(n.parameters.get("charset")).decode(e):new Blob([e],{type:n.essence})}(l,s),p="getSetCookie"in n.headers&&"function"==typeof n.headers.getSetCookie?n.headers.getSetCookie():[];return[null,{timestamp:Date.now(),request:t,response:{...n,headers:o,cookieHeaderKeys:p,data:u,size:l.byteLength,duration:Date.now()-e,method:a.method,status:n.status,path:f}}]}catch(e){return null==c||c.emit("abort"),[(0,Jn.cQ)(e,Jn.Sr.REQUEST_FAILED),null]}},controller:d}]}catch(e){return console.error(e),null==c||c.emit("abort"),[(0,Jn.cQ)(e),null]}})({request:v.value,example:m.value,selectedSecuritySchemeUids:W.value,proxyUrl:(null==(a=b.value)?void 0:a.proxyUrl)??"",environment:c,globalCookies:u,status:N.requestStatus,securitySchemes:P,server:d});if(p)return void n(p.message,"error");B.value=f.controller;const[w,O]=await f.sendRequest();w?n(w.message,"error"):A.push(O)},q=async()=>{var e;return null==(e=B.value)?void 0:e.abort(Jn.Sr.REQUEST_ABORTED)};function V({requestName:e,collectionUid:t}){var n,r,a,o,i,s,l;if(!F.value)return;const c=O.value.find((e=>e.uid===t));if(!c)return;const u="Drafts"===(null==(n=null==c?void 0:c.info)?void 0:n.title);if(!u&&F.value.servers){const e=Object.values($).find((e=>{var t,n;return e.url===(null==(n=null==(t=null==F?void 0:F.value)?void 0:t.servers)?void 0:n[0])}));Q.value=e?e.uid:M.add({url:F.value.servers[0]??"/"},c.uid).uid}const d=D.add({summary:e,path:u?null==(r=null==F?void 0:F.value)?void 0:r.url:null==(a=null==F?void 0:F.value)?void 0:a.path,method:null==(o=null==F?void 0:F.value)?void 0:o.method,parameters:null==(i=null==F?void 0:F.value)?void 0:i.parameters,selectedServerUid:u?void 0:Q.value,requestBody:null==(s=null==F?void 0:F.value)?void 0:s.requestBody},c.uid);d&&null!=(l=b.value)&&l.uid&&U.push({name:"request",params:{[I.TU.Workspace]:b.value.uid,[I.TU.Collection]:c.uid,[I.TU.Request]:d.uid}}),T.hide()}function H(e){F.value=function(e){const t=_(e),{method:n="get",url:r,body:a="",headers:o={},servers:i,queryParameters:s=[]}=t,l=new URL(r).pathname,c=null!=a&&a.includes("=")&&!a.startsWith("{")?"application/x-www-form-urlencoded":o["Content-Type"]||"",u=a?function(e){try{return JSON.parse(e)}catch{const t={};return e.split("&").forEach((e=>{const[n,r]=e.split("=");n&&r&&(t[decodeURIComponent(n)]=decodeURIComponent(r))})),t}}(a):{},d=[...Array.isArray(s)?s.map((({key:e,value:t})=>({name:e,in:"query",schema:{type:typeof t,examples:[t]}}))):[],...Object.entries(o||{}).map((([e,t])=>({name:e,in:"header",schema:{type:typeof t},example:t})))];return{method:n,url:r,path:l,headers:o,servers:i??[],...Object.keys(u).length>0&&{requestBody:{content:{[c]:{schema:{type:"object",properties:Object.fromEntries(Object.entries(u).map((([e,t])=>[e,{type:typeof t}])))},example:u}}}},parameters:d}}(e),T.show()}return(0,r.sV)((()=>{N.executeRequest.on(X),N.cancelRequest.on(q)})),K(),(0,r.xo)((()=>N.executeRequest.off(X))),(e,t)=>{var n;return(0,r.uX)(),(0,r.CE)(r.FK,null,[(0,a.R1)(h)&&(0,a.R1)(b)?((0,r.uX)(),(0,r.CE)("div",{key:0,class:(0,o.C4)(["bg-b-1 relative z-0 flex h-full flex-1 flex-col overflow-hidden pt-0",{"!mb-0 !mr-0 !border-0":"modal"===(0,a.R1)(p)}])},[(0,r.Lk)("div",ar,[(0,a.R1)(R)?((0,r.uX)(),(0,r.Wv)(Yn,{key:0,isSidebarOpen:L.value,onNewTab:t[0]||(t[0]=t=>e.$emit("newTab",t)),"onUpdate:isSidebarOpen":t[1]||(t[1]=e=>L.value=e)},null,8,["isSidebarOpen"])):(0,r.Q3)("",!0),(0,a.R1)(v)?((0,r.uX)(),(0,r.CE)("div",or,[(0,r.bF)(ue,{modelValue:L.value,"onUpdate:modelValue":t[2]||(t[2]=e=>L.value=e),collection:(0,a.R1)(h),envVariables:(0,a.R1)(w),environment:(0,a.R1)(g),operation:(0,a.R1)(v),server:(0,a.R1)(y),workspace:(0,a.R1)(b),onHideModal:t[3]||(t[3]=()=>(0,a.R1)(T).hide()),onImportCurl:H},null,8,["modelValue","collection","envVariables","environment","operation","server","workspace"]),(0,r.bF)(x.A,null,{default:(0,r.k6)((()=>[(0,a.R1)(m)?((0,r.uX)(),(0,r.Wv)(E.A,{key:0,class:(0,o.C4)(["flex-1",[L.value?"sidebar-active-hide-layout":""]])},{default:(0,r.k6)((()=>{var e;return[(0,r.bF)(Y.A,{collection:(0,a.R1)(h),envVariables:(0,a.R1)(w),environment:(0,a.R1)(g),example:(0,a.R1)(m),operation:(0,a.R1)(v),selectedSecuritySchemeUids:W.value,server:(0,a.R1)(y),workspace:(0,a.R1)(b)},null,8,["collection","envVariables","environment","example","operation","selectedSecuritySchemeUids","server","workspace"]),(0,r.bF)(_t,{numWorkspaceRequests:(0,a.R1)(S).length,response:null==(e=z.value)?void 0:e.response},null,8,["numWorkspaceRequests","response"])]})),_:1},8,["class"])):(0,r.Q3)("",!0)])),_:1})])):((0,r.uX)(),(0,r.Wv)(d.A,{key:2}))])],2)):((0,r.uX)(),(0,r.Wv)(d.A,{key:1})),F.value?((0,r.uX)(),(0,r.Wv)(k,{key:2,collectionUid:(null==(n=(0,a.R1)(h))?void 0:n.uid)??"",parsedCurl:F.value,state:(0,a.R1)(T),onClose:t[4]||(t[4]=e=>(0,a.R1)(T).hide()),onImportCurl:V},null,8,["collectionUid","parsedCurl","state"])):(0,r.Q3)("",!0)],64)}}}),sr=(0,O.A)(ir,[["__scopeId","data-v-02ba9b2d"]])},3016:(e,t,n)=>{"use strict";n.d(t,{A:()=>s});var r=n(641),a=n(3526);const o={},i={class:"group contents w-fit min-w-full"},s=(0,a.A)(o,[["render",function(e,t){return(0,r.uX)(),(0,r.CE)("tr",i,[(0,r.RG)(e.$slots,"default")])}]])},3050:(e,t,n)=>{"use strict";n.r(t),n.d(t,{default:()=>P});var r=n(641),a=n(953),o=n(3751),i=n(1266),s=n(6455),l=n(9351),c=n(1636),u=n(4486),d=n(3966),p=n(1465),h=n(8082),f=n(9005),m=n(2490),g=n(5936),v=n(9135),b=n(8309);const y=(0,r.pM)({__name:"CookieForm",setup(e){const{activeCookieId:t}=(0,v.JC)(),{cookies:n,cookieMutators:a}=(0,b.hA)(),o=[{label:"Name",key:"name",placeholder:"session_id"},{label:"Value",key:"value",placeholder:"my-cookie-session-id"},{label:"Domain",key:"domain",placeholder:"example.com"}],i=(0,r.EW)((()=>n[t.value]||s.c.parse({name:"",value:"",domain:"",path:""}))),l=(e,n)=>{a.edit(t.value,e,n)};return(e,t)=>((0,r.uX)(),(0,r.Wv)(g.A,{data:i.value,onUpdate:l,options:o},{title:(0,r.k6)((()=>t[0]||(t[0]=[(0,r.Lk)("span",{class:"text-c-2"},"Edit Cookie",-1)]))),_:1},8,["data"]))}});var w=n(5411),O=n(4445),k=n(8157);const x={class:"flex gap-2 h-8 items-start text-sm"},E={class:"flex gap-2 h-8 items-start text-sm"},S={class:"flex gap-2 h-8 items-start text-sm"},_=(0,r.pM)({__name:"CookieModal",props:{state:{}},emits:["cancel","submit"],setup(e,{emit:t}){const n=e,o=t,s=(0,a.KR)({name:"",value:"",domain:""}),{toast:l}=(0,k.Yf)(),c=()=>{s.value.name&&s.value.value?(o("submit",s.value),n.state.hide()):l("Please fill in all fields before adding a cookie.","error")};return(0,r.wB)((()=>n.state.open),(e=>{e&&(s.value={name:"",value:"",domain:""})})),(e,t)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(i.uT),{size:"xs",state:e.state,title:"Add Cookie"},{default:(0,r.k6)((()=>[(0,r.bF)(w.A,{disabled:!s.value.name||!s.value.value,onCancel:t[3]||(t[3]=e=>o("cancel")),onSubmit:c},{submit:(0,r.k6)((()=>t[7]||(t[7]=[(0,r.eW)("Add Cookie")]))),default:(0,r.k6)((()=>[(0,r.Lk)("div",x,[t[4]||(t[4]=(0,r.eW)(" Name: ")),(0,r.bF)(O.A,{modelValue:s.value.name,"onUpdate:modelValue":t[0]||(t[0]=e=>s.value.name=e),autofocus:"",class:"!p-0",placeholder:"session_id"},null,8,["modelValue"])]),(0,r.Lk)("div",E,[t[5]||(t[5]=(0,r.eW)(" Value: ")),(0,r.bF)(O.A,{modelValue:s.value.value,"onUpdate:modelValue":t[1]||(t[1]=e=>s.value.value=e),autofocus:"",class:"!p-0",placeholder:"my-cookie-session-id"},null,8,["modelValue"])]),(0,r.Lk)("div",S,[t[6]||(t[6]=(0,r.eW)(" Domain: ")),(0,r.bF)(O.A,{modelValue:s.value.domain,"onUpdate:modelValue":t[2]||(t[2]=e=>s.value.domain=e),autofocus:"",class:"!p-0",placeholder:"example.com"},null,8,["modelValue"])])])),_:1},8,["disabled"])])),_:1},8,["state"]))}}),C=(0,n(3526).A)(_,[["__scopeId","data-v-c25da223"]]);var T=n(3858);const A={class:"flex-1"},R={class:"relative mb-[.5px] last:mb-0"},P=(0,r.pM)({__name:"Cookies",setup(e){const{cookies:t,cookieMutators:n,events:g,workspaceMutators:w}=(0,b.hA)(),{activeWorkspace:O,activeCookieId:k}=(0,v.JC)(),x=(0,l.rd)(),E=(0,l.lq)(),S=(0,i.hS)(),_=e=>{var t,r;const a=s.c.parse({name:e.name,value:e.value,domain:e.domain,path:"/"});n.add(a),w.edit(null==(t=O.value)?void 0:t.uid,"cookies",[...(null==(r=O.value)?void 0:r.cookies)??[],a.uid]),x.push({name:"cookies",params:{cookies:a.uid}})},P=()=>{S.show()},D=e=>{null!=e&&e.createNew&&"cookies"===E.name&&P()};(0,r.sV)((()=>g.hotKeys.on(D))),(0,r.xo)((()=>g.hotKeys.off(D)));const M=(0,r.EW)((()=>t[k.value])),$=(0,r.EW)((()=>Object.keys(t).length>0&&M.value));return(e,i)=>((0,r.uX)(),(0,r.Wv)(h.A,null,{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(T.A),{title:"Cookies"},{content:(0,r.k6)((()=>[(0,r.Lk)("div",A,[(0,r.bF)(d.A,null,{default:(0,r.k6)((()=>[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(Object.values((0,a.R1)(t)),(e=>((0,r.uX)(),(0,r.CE)("li",{key:e.uid,class:"gap-1/2 flex flex-col"},[(0,r.Lk)("div",R,[((0,r.uX)(),(0,r.Wv)(p.A,{key:e.uid,class:"text-xs",isDeletable:"",type:"cookies",variable:{name:e.name,uid:e.uid},warningMessage:"Are you sure you want to delete this cookie?",onClick:(0,o.D$)((t=>((e,t)=>{var n;const r={name:"cookies",params:{workspace:(null==(n=O.value)?void 0:n.uid)??"default",cookies:t}};if(e.metaKey){const e=x.resolve(r).href;window.open(e,"_blank")}else x.push(r)})(t,e.uid)),["prevent"]),onDelete:r=>(e=>{var r,a;n.delete(e),w.edit(null==(r=O.value)?void 0:r.uid,"cookies",[...((null==(a=O.value)?void 0:a.cookies)??[]).filter((t=>t!==e))]);const o=Object.values(t).filter((t=>t.uid!==e));if(o.length>0){const e=o[o.length-1];e&&x.push(e.uid)}else x.push({name:"cookies",params:{[m.TU.Cookies]:"default"}})})(e.uid)},null,8,["variable","onClick","onDelete"]))])])))),128))])),_:1})])])),button:(0,r.k6)((()=>[(0,r.bF)(u.A,{click:P,hotkey:"N"},{title:(0,r.k6)((()=>i[1]||(i[1]=[(0,r.eW)(" Add Cookie ")]))),_:1})])),_:1}),(0,r.bF)(f.A,{class:"flex-1"},{default:(0,r.k6)((()=>[$.value?((0,r.uX)(),(0,r.Wv)(y,{key:0})):((0,r.uX)(),(0,r.Wv)(c.A,{key:1}))])),_:1}),(0,r.bF)(C,{state:(0,a.R1)(S),onCancel:i[0]||(i[0]=e=>(0,a.R1)(S).hide()),onSubmit:_},null,8,["state"])])),_:1}))}})},3073:(e,t,n)=>{"use strict";n.d(t,{X2F:()=>p,GF$:()=>g,iXd:()=>y,iwM:()=>w,BLq:()=>v,Ubj:()=>f,wYm:()=>m,AHd:()=>O,Sj0:()=>k});var r=n(9271),a=n(641),o=n(953),i=!1;function s(e){var t;const n=(0,r.BA)(e);return null!=(t=null==n?void 0:n.$el)?t:n}const l=r.oc?window:void 0,c=r.oc?window.document:void 0;function u(...e){let t,n,o,i;if("string"==typeof e[0]||Array.isArray(e[0])?([n,o,i]=e,t=l):[t,n,o,i]=e,!t)return r.lQ;Array.isArray(n)||(n=[n]),Array.isArray(o)||(o=[o]);const c=[],u=()=>{c.forEach((e=>e())),c.length=0},d=(0,a.wB)((()=>[s(t),(0,r.BA)(i)]),(([e,t])=>{if(u(),!e)return;const a=(0,r.Gv)(t)?{...t}:t;c.push(...n.flatMap((t=>o.map((n=>((e,t,n,r)=>(e.addEventListener(t,n,r),()=>e.removeEventListener(t,n,r)))(e,t,n,a))))))}),{immediate:!0,flush:"post"}),p=()=>{d(),u()};return(0,r.Uo)(p),p}r.oc&&window.navigator,r.oc&&window.location;let d=!1;function p(e,t,n={}){const{window:a=l,ignore:o=[],capture:i=!0,detectIframe:c=!1}=n;if(!a)return r.lQ;r.un&&!d&&(d=!0,Array.from(a.document.body.children).forEach((e=>e.addEventListener("click",r.lQ))),a.document.documentElement.addEventListener("click",r.lQ));let p=!0;const h=e=>o.some((t=>{if("string"==typeof t)return Array.from(a.document.querySelectorAll(t)).some((t=>t===e.target||e.composedPath().includes(t)));{const n=s(t);return n&&(e.target===n||e.composedPath().includes(n))}})),f=[u(a,"click",(n=>{const r=s(e);r&&r!==n.target&&!n.composedPath().includes(r)&&(0===n.detail&&(p=!h(n)),p?t(n):p=!0)}),{passive:!0,capture:i}),u(a,"pointerdown",(t=>{const n=s(e);p=!h(t)&&!(!n||t.composedPath().includes(n))}),{passive:!0}),c&&u(a,"blur",(n=>{setTimeout((()=>{var r;const o=s(e);"IFRAME"!==(null==(r=a.document.activeElement)?void 0:r.tagName)||(null==o?void 0:o.contains(a.document.activeElement))||t(n)}),0)}))].filter(Boolean);return()=>f.forEach((e=>e()))}function h(e){const t=function(){const e=(0,o.KR)(!1),t=(0,a.nI)();return t&&(0,a.sV)((()=>{e.value=!0}),i?void 0:t),e}();return(0,a.EW)((()=>(t.value,Boolean(e()))))}function f(e,t={}){const{window:n=l}=t,i=h((()=>n&&"matchMedia"in n&&"function"==typeof n.matchMedia));let s;const c=(0,o.KR)(!1),u=e=>{c.value=e.matches},d=()=>{s&&("removeEventListener"in s?s.removeEventListener("change",u):s.removeListener(u))},p=(0,a.nT)((()=>{i.value&&(d(),s=n.matchMedia((0,r.BA)(e)),"addEventListener"in s?s.addEventListener("change",u):s.addListener(u),c.value=s.matches)}));return(0,r.Uo)((()=>{p(),d(),s=void 0})),c}function m(e,t,n={}){const{window:o=l,...i}=n;let c;const u=h((()=>o&&"ResizeObserver"in o)),d=()=>{c&&(c.disconnect(),c=void 0)},p=(0,a.EW)((()=>Array.isArray(e)?e.map((e=>s(e))):[s(e)])),f=(0,a.wB)(p,(e=>{if(d(),u.value&&o){c=new ResizeObserver(t);for(const t of e)t&&c.observe(t,i)}}),{immediate:!0,flush:"post"}),m=()=>{d(),f()};return(0,r.Uo)(m),{isSupported:u,stop:m}}function g(e,t={}){const{delayEnter:n=0,delayLeave:r=0,window:a=l}=t,i=(0,o.KR)(!1);let s;const c=e=>{const t=e?n:r;s&&(clearTimeout(s),s=void 0),t?s=setTimeout((()=>i.value=e),t):i.value=e};return a?(u(e,"mouseenter",(()=>c(!0)),{passive:!0}),u(e,"mouseleave",(()=>c(!1)),{passive:!0}),i):i}function v(e,t,n={}){const{root:i,rootMargin:c="0px",threshold:u=.1,window:d=l,immediate:p=!0}=n,f=h((()=>d&&"IntersectionObserver"in d)),m=(0,a.EW)((()=>{const t=(0,r.BA)(e);return(Array.isArray(t)?t:[t]).map(s).filter(r.d6)}));let g=r.lQ;const v=(0,o.KR)(p),b=f.value?(0,a.wB)((()=>[m.value,s(i),v.value]),(([e,n])=>{if(g(),!v.value)return;if(!e.length)return;const a=new IntersectionObserver(t,{root:s(n),rootMargin:c,threshold:u});e.forEach((e=>e&&a.observe(e))),g=()=>{a.disconnect(),g=r.lQ}}),{immediate:p,flush:"post"}):r.lQ,y=()=>{g(),b(),v.value=!1};return(0,r.Uo)(y),{isSupported:f,isActive:v,pause(){g(),v.value=!1},resume(){v.value=!0},stop:y}}"undefined"!=typeof globalThis?globalThis:"undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self&&self;const b=new Map;function y(e){const t=(0,o.o5)();function n(n){var a;const o=b.get(e)||new Set;o.add(n),b.set(e,o);const i=()=>r(n);return null==(a=null==t?void 0:t.cleanups)||a.push(i),i}function r(t){const n=b.get(e);n&&(n.delete(t),n.size||a())}function a(){b.delete(e)}return{on:n,once:function(e){return n((function t(...n){r(t),e(...n)}))},off:r,emit:function(t,n){var r;null==(r=b.get(e))||r.forEach((e=>e(t,n)))},reset:a}}function w(e=null,t={}){const{baseUrl:n="",rel:o="icon",document:i=c}=t,s=(0,r.lW)(e);return(0,a.wB)(s,((e,t)=>{"string"==typeof e&&e!==t&&(e=>{const t=null==i?void 0:i.head.querySelectorAll(`link[rel*="${o}"]`);if(t&&0!==t.length)null==t||t.forEach((t=>t.href=`${n}${e}`));else{const t=null==i?void 0:i.createElement("link");t&&(t.rel=o,t.href=`${n}${e}`,t.type=`image/${e.split(".").pop()}`,null==i||i.head.append(t))}})(e)}),{immediate:!0}),s}function O(e){var t;const n=(0,o.KR)(null==e?void 0:e.element),i=(0,o.KR)(null==e?void 0:e.input),s=null!=(t=null==e?void 0:e.styleProp)?t:"height",l=(0,o.KR)(1);function c(){var t;if(!n.value)return;let a="";n.value.style[s]="1px",l.value=null==(t=n.value)?void 0:t.scrollHeight,(null==e?void 0:e.styleTarget)?(0,r.BA)(e.styleTarget).style[s]=`${l.value}px`:a=`${l.value}px`,n.value.style[s]=a}return(0,a.wB)([i,n],(()=>(0,a.dY)(c)),{immediate:!0}),(0,a.wB)(l,(()=>{var t;return null==(t=null==e?void 0:e.onResize)?void 0:t.call(e)})),m(n,(()=>c())),(null==e?void 0:e.watch)&&(0,a.wB)(e.watch,c,{immediate:!0,deep:!0}),{textarea:n,input:i,triggerResize:c}}function k(e,t,n){const{start:a}=(0,r.TO)(s,t,{immediate:!1}),i=(0,o.KR)(!1);async function s(){i.value&&(await e(),a())}function l(){i.value||(i.value=!0,s())}function c(){i.value=!1}return(null==n?void 0:n.immediate)&&l(),(0,r.Uo)(c),{isActive:i,pause:c,resume:l}}Number.POSITIVE_INFINITY;r.D_},3179:(e,t,n)=>{"use strict";n.d(t,{y:()=>i});var r=n(7786),a=n(2453);const o="ws://replace.me",i=(e,t,{operation:n,example:i,server:s,securitySchemes:l=[]})=>{var c;try{const u=(0,a.J)({operation:n,example:i,server:s,securitySchemes:l});if(!u.url)return[new Error("Please enter a URL to see a code snippet"),null];const d=u.url.startsWith("/")?"":"/";try{new URL(u.url)}catch{u.url=`${o}${d}${u.url}`}if("application/json"===(null==(c=u.postData)?void 0:c.mimeType))try{JSON.parse(u.postData.text||"{}")}catch{return[new Error("Invalid JSON body"),null]}const p=e.replace("javascript","js");if((0,r.E)().hasPlugin(p,t)){const e=(0,r.E)().print(p,t,u);return e?[null,e.replace(`${o}${d}`,"")]:[new Error("Error generating snippet"),null]}}catch(e){return console.error("[getSnippet]",e),[new Error("Error generating snippet"),null]}return[new Error("No snippet found"),null]}},3184:(e,t,n)=>{"use strict";n.d(t,{qg:()=>gn,As:()=>vn});const r=Symbol.for("yaml.alias"),a=Symbol.for("yaml.document"),o=Symbol.for("yaml.map"),i=Symbol.for("yaml.pair"),s=Symbol.for("yaml.scalar"),l=Symbol.for("yaml.seq"),c=Symbol.for("yaml.node.type"),u=e=>!!e&&"object"==typeof e&&e[c]===r,d=e=>!!e&&"object"==typeof e&&e[c]===a,p=e=>!!e&&"object"==typeof e&&e[c]===o,h=e=>!!e&&"object"==typeof e&&e[c]===i,f=e=>!!e&&"object"==typeof e&&e[c]===s,m=e=>!!e&&"object"==typeof e&&e[c]===l;function g(e){if(e&&"object"==typeof e)switch(e[c]){case o:case l:return!0}return!1}function v(e){if(e&&"object"==typeof e)switch(e[c]){case r:case o:case s:case l:return!0}return!1}const b=Symbol("break visit"),y=Symbol("skip children"),w=Symbol("remove node");function O(e,t){const n=S(t);d(e)?k(null,e.contents,n,Object.freeze([e]))===w&&(e.contents=null):k(null,e,n,Object.freeze([]))}function k(e,t,n,r){const a=_(e,t,n,r);if(v(a)||h(a))return C(e,r,a),k(e,a,n,r);if("symbol"!=typeof a)if(g(t)){r=Object.freeze(r.concat(t));for(let e=0;e<t.items.length;++e){const a=k(e,t.items[e],n,r);if("number"==typeof a)e=a-1;else{if(a===b)return b;a===w&&(t.items.splice(e,1),e-=1)}}}else if(h(t)){r=Object.freeze(r.concat(t));const e=k("key",t.key,n,r);if(e===b)return b;e===w&&(t.key=null);const a=k("value",t.value,n,r);if(a===b)return b;a===w&&(t.value=null)}return a}async function x(e,t){const n=S(t);d(e)?await E(null,e.contents,n,Object.freeze([e]))===w&&(e.contents=null):await E(null,e,n,Object.freeze([]))}async function E(e,t,n,r){const a=await _(e,t,n,r);if(v(a)||h(a))return C(e,r,a),E(e,a,n,r);if("symbol"!=typeof a)if(g(t)){r=Object.freeze(r.concat(t));for(let e=0;e<t.items.length;++e){const a=await E(e,t.items[e],n,r);if("number"==typeof a)e=a-1;else{if(a===b)return b;a===w&&(t.items.splice(e,1),e-=1)}}}else if(h(t)){r=Object.freeze(r.concat(t));const e=await E("key",t.key,n,r);if(e===b)return b;e===w&&(t.key=null);const a=await E("value",t.value,n,r);if(a===b)return b;a===w&&(t.value=null)}return a}function S(e){return"object"==typeof e&&(e.Collection||e.Node||e.Value)?Object.assign({Alias:e.Node,Map:e.Node,Scalar:e.Node,Seq:e.Node},e.Value&&{Map:e.Value,Scalar:e.Value,Seq:e.Value},e.Collection&&{Map:e.Collection,Seq:e.Collection},e):e}function _(e,t,n,r){return"function"==typeof n?n(e,t,r):p(t)?n.Map?.(e,t,r):m(t)?n.Seq?.(e,t,r):h(t)?n.Pair?.(e,t,r):f(t)?n.Scalar?.(e,t,r):u(t)?n.Alias?.(e,t,r):void 0}function C(e,t,n){const r=t[t.length-1];if(g(r))r.items[e]=n;else if(h(r))"key"===e?r.key=n:r.value=n;else{if(!d(r)){const e=u(r)?"alias":"scalar";throw new Error(`Cannot replace node with ${e} parent`)}r.contents=n}}O.BREAK=b,O.SKIP=y,O.REMOVE=w,x.BREAK=b,x.SKIP=y,x.REMOVE=w;const T={"!":"%21",",":"%2C","[":"%5B","]":"%5D","{":"%7B","}":"%7D"};class A{constructor(e,t){this.docStart=null,this.docEnd=!1,this.yaml=Object.assign({},A.defaultYaml,e),this.tags=Object.assign({},A.defaultTags,t)}clone(){const e=new A(this.yaml,this.tags);return e.docStart=this.docStart,e}atDocument(){const e=new A(this.yaml,this.tags);switch(this.yaml.version){case"1.1":this.atNextDocument=!0;break;case"1.2":this.atNextDocument=!1,this.yaml={explicit:A.defaultYaml.explicit,version:"1.2"},this.tags=Object.assign({},A.defaultTags)}return e}add(e,t){this.atNextDocument&&(this.yaml={explicit:A.defaultYaml.explicit,version:"1.1"},this.tags=Object.assign({},A.defaultTags),this.atNextDocument=!1);const n=e.trim().split(/[ \t]+/),r=n.shift();switch(r){case"%TAG":{if(2!==n.length&&(t(0,"%TAG directive should contain exactly two parts"),n.length<2))return!1;const[e,r]=n;return this.tags[e]=r,!0}case"%YAML":{if(this.yaml.explicit=!0,1!==n.length)return t(0,"%YAML directive should contain exactly one part"),!1;const[e]=n;return"1.1"===e||"1.2"===e?(this.yaml.version=e,!0):(t(6,`Unsupported YAML version ${e}`,/^\d+\.\d+$/.test(e)),!1)}default:return t(0,`Unknown directive ${r}`,!0),!1}}tagName(e,t){if("!"===e)return"!";if("!"!==e[0])return t(`Not a valid tag: ${e}`),null;if("<"===e[1]){const n=e.slice(2,-1);return"!"===n||"!!"===n?(t(`Verbatim tags aren't resolved, so ${e} is invalid.`),null):(">"!==e[e.length-1]&&t("Verbatim tags must end with a >"),n)}const[,n,r]=e.match(/^(.*!)([^!]*)$/s);r||t(`The ${e} tag has no suffix`);const a=this.tags[n];if(a)try{return a+decodeURIComponent(r)}catch(e){return t(String(e)),null}return"!"===n?e:(t(`Could not resolve tag: ${e}`),null)}tagString(e){for(const[t,n]of Object.entries(this.tags))if(e.startsWith(n))return t+e.substring(n.length).replace(/[!,[\]{}]/g,(e=>T[e]));return"!"===e[0]?e:`!<${e}>`}toString(e){const t=this.yaml.explicit?[`%YAML ${this.yaml.version||"1.2"}`]:[],n=Object.entries(this.tags);let r;if(e&&n.length>0&&v(e.contents)){const t={};O(e.contents,((e,n)=>{v(n)&&n.tag&&(t[n.tag]=!0)})),r=Object.keys(t)}else r=[];for(const[a,o]of n)"!!"===a&&"tag:yaml.org,2002:"===o||e&&!r.some((e=>e.startsWith(o)))||t.push(`%TAG ${a} ${o}`);return t.join("\n")}}function R(e){if(/[\x00-\x19\s,[\]{}]/.test(e)){const t=JSON.stringify(e);throw new Error(`Anchor must not contain whitespace or control characters: ${t}`)}return!0}function P(e){const t=new Set;return O(e,{Value(e,n){n.anchor&&t.add(n.anchor)}}),t}function D(e,t){for(let n=1;;++n){const r=`${e}${n}`;if(!t.has(r))return r}}function M(e,t,n,r){if(r&&"object"==typeof r)if(Array.isArray(r))for(let t=0,n=r.length;t<n;++t){const n=r[t],a=M(e,r,String(t),n);void 0===a?delete r[t]:a!==n&&(r[t]=a)}else if(r instanceof Map)for(const t of Array.from(r.keys())){const n=r.get(t),a=M(e,r,t,n);void 0===a?r.delete(t):a!==n&&r.set(t,a)}else if(r instanceof Set)for(const t of Array.from(r)){const n=M(e,r,t,t);void 0===n?r.delete(t):n!==t&&(r.delete(t),r.add(n))}else for(const[t,n]of Object.entries(r)){const a=M(e,r,t,n);void 0===a?delete r[t]:a!==n&&(r[t]=a)}return e.call(t,n,r)}function $(e,t,n){if(Array.isArray(e))return e.map(((e,t)=>$(e,String(t),n)));if(e&&"function"==typeof e.toJSON){if(!n||!f(r=e)&&!g(r)||!r.anchor)return e.toJSON(t,n);const a={aliasCount:0,count:1,res:void 0};n.anchors.set(e,a),n.onCreate=e=>{a.res=e,delete n.onCreate};const o=e.toJSON(t,n);return n.onCreate&&n.onCreate(o),o}var r;return"bigint"!=typeof e||n?.keep?e:Number(e)}A.defaultYaml={explicit:!1,version:"1.2"},A.defaultTags={"!!":"tag:yaml.org,2002:"};class N{constructor(e){Object.defineProperty(this,c,{value:e})}clone(){const e=Object.create(Object.getPrototypeOf(this),Object.getOwnPropertyDescriptors(this));return this.range&&(e.range=this.range.slice()),e}toJS(e,{mapAsMap:t,maxAliasCount:n,onAnchor:r,reviver:a}={}){if(!d(e))throw new TypeError("A document argument is required");const o={anchors:new Map,doc:e,keep:!0,mapAsMap:!0===t,mapKeyWarned:!1,maxAliasCount:"number"==typeof n?n:100},i=$(this,"",o);if("function"==typeof r)for(const{count:e,res:t}of o.anchors.values())r(t,e);return"function"==typeof a?M(a,{"":i},"",i):i}}class I extends N{constructor(e){super(r),this.source=e,Object.defineProperty(this,"tag",{set(){throw new Error("Alias nodes cannot have tags")}})}resolve(e){let t;return O(e,{Node:(e,n)=>{if(n===this)return O.BREAK;n.anchor===this.source&&(t=n)}}),t}toJSON(e,t){if(!t)return{source:this.source};const{anchors:n,doc:r,maxAliasCount:a}=t,o=this.resolve(r);if(!o){const e=`Unresolved alias (the anchor must be set before the alias): ${this.source}`;throw new ReferenceError(e)}let i=n.get(o);if(i||($(o,null,t),i=n.get(o)),!i||void 0===i.res)throw new ReferenceError("This should not happen: Alias anchor was not resolved?");if(a>=0&&(i.count+=1,0===i.aliasCount&&(i.aliasCount=L(r,o,n)),i.count*i.aliasCount>a))throw new ReferenceError("Excessive alias count indicates a resource exhaustion attack");return i.res}toString(e,t,n){const r=`*${this.source}`;if(e){if(R(this.source),e.options.verifyAliasOrder&&!e.anchors.has(this.source)){const e=`Unresolved alias (the anchor must be set before the alias): ${this.source}`;throw new Error(e)}if(e.implicitKey)return`${r} `}return r}}function L(e,t,n){if(u(t)){const r=t.resolve(e),a=n&&r&&n.get(r);return a?a.count*a.aliasCount:0}if(g(t)){let r=0;for(const a of t.items){const t=L(e,a,n);t>r&&(r=t)}return r}if(h(t)){const r=L(e,t.key,n),a=L(e,t.value,n);return Math.max(r,a)}return 1}const B=e=>!e||"function"!=typeof e&&"object"!=typeof e;class F extends N{constructor(e){super(s),this.value=e}toJSON(e,t){return t?.keep?this.value:$(this.value,e,t)}toString(){return String(this.value)}}function Q(e,t,n){if(d(e)&&(e=e.contents),v(e))return e;if(h(e)){const t=n.schema[o].createNode?.(n.schema,null,n);return t.items.push(e),t}(e instanceof String||e instanceof Number||e instanceof Boolean||"undefined"!=typeof BigInt&&e instanceof BigInt)&&(e=e.valueOf());const{aliasDuplicateObjects:r,onAnchor:a,onTagObj:i,schema:s,sourceObjects:c}=n;let u;if(r&&e&&"object"==typeof e){if(u=c.get(e),u)return u.anchor||(u.anchor=a(e)),new I(u.anchor);u={anchor:null,node:null},c.set(e,u)}t?.startsWith("!!")&&(t="tag:yaml.org,2002:"+t.slice(2));let p=function(e,t,n){if(t){const e=n.filter((e=>e.tag===t)),r=e.find((e=>!e.format))??e[0];if(!r)throw new Error(`Tag ${t} not found`);return r}return n.find((t=>t.identify?.(e)&&!t.format))}(e,t,s.tags);if(!p){if(e&&"function"==typeof e.toJSON&&(e=e.toJSON()),!e||"object"!=typeof e){const t=new F(e);return u&&(u.node=t),t}p=e instanceof Map?s[o]:Symbol.iterator in Object(e)?s[l]:s[o]}i&&(i(p),delete n.onTagObj);const f=p?.createNode?p.createNode(n.schema,e,n):"function"==typeof p?.nodeClass?.from?p.nodeClass.from(n.schema,e,n):new F(e);return t?f.tag=t:p.default||(f.tag=p.tag),u&&(u.node=f),f}function U(e,t,n){let r=n;for(let e=t.length-1;e>=0;--e){const n=t[e];if("number"==typeof n&&Number.isInteger(n)&&n>=0){const e=[];e[n]=r,r=e}else r=new Map([[n,r]])}return Q(r,void 0,{aliasDuplicateObjects:!1,keepUndefined:!1,onAnchor:()=>{throw new Error("This should not happen, please report a bug.")},schema:e,sourceObjects:new Map})}F.BLOCK_FOLDED="BLOCK_FOLDED",F.BLOCK_LITERAL="BLOCK_LITERAL",F.PLAIN="PLAIN",F.QUOTE_DOUBLE="QUOTE_DOUBLE",F.QUOTE_SINGLE="QUOTE_SINGLE";const z=e=>null==e||"object"==typeof e&&!!e[Symbol.iterator]().next().done;class j extends N{constructor(e,t){super(e),Object.defineProperty(this,"schema",{value:t,configurable:!0,enumerable:!1,writable:!0})}clone(e){const t=Object.create(Object.getPrototypeOf(this),Object.getOwnPropertyDescriptors(this));return e&&(t.schema=e),t.items=t.items.map((t=>v(t)||h(t)?t.clone(e):t)),this.range&&(t.range=this.range.slice()),t}addIn(e,t){if(z(e))this.add(t);else{const[n,...r]=e,a=this.get(n,!0);if(g(a))a.addIn(r,t);else{if(void 0!==a||!this.schema)throw new Error(`Expected YAML collection at ${n}. Remaining path: ${r}`);this.set(n,U(this.schema,r,t))}}}deleteIn(e){const[t,...n]=e;if(0===n.length)return this.delete(t);const r=this.get(t,!0);if(g(r))return r.deleteIn(n);throw new Error(`Expected YAML collection at ${t}. Remaining path: ${n}`)}getIn(e,t){const[n,...r]=e,a=this.get(n,!0);return 0===r.length?!t&&f(a)?a.value:a:g(a)?a.getIn(r,t):void 0}hasAllNullValues(e){return this.items.every((t=>{if(!h(t))return!1;const n=t.value;return null==n||e&&f(n)&&null==n.value&&!n.commentBefore&&!n.comment&&!n.tag}))}hasIn(e){const[t,...n]=e;if(0===n.length)return this.has(t);const r=this.get(t,!0);return!!g(r)&&r.hasIn(n)}setIn(e,t){const[n,...r]=e;if(0===r.length)this.set(n,t);else{const e=this.get(n,!0);if(g(e))e.setIn(r,t);else{if(void 0!==e||!this.schema)throw new Error(`Expected YAML collection at ${n}. Remaining path: ${r}`);this.set(n,U(this.schema,r,t))}}}}const W=e=>e.replace(/^(?!$)(?: $)?/gm,"#");function X(e,t){return/^\n+$/.test(e)?e.substring(1):t?e.replace(/^(?! *$)/gm,t):e}const q=(e,t,n)=>e.endsWith("\n")?X(n,t):n.includes("\n")?"\n"+X(n,t):(e.endsWith(" ")?"":" ")+n,V="flow",H="block",Z="quoted";function G(e,t,n="flow",{indentAtStart:r,lineWidth:a=80,minContentWidth:o=20,onFold:i,onOverflow:s}={}){if(!a||a<0)return e;a<o&&(o=0);const l=Math.max(1+o,1+a-t.length);if(e.length<=l)return e;const c=[],u={};let d,p,h=a-t.length;"number"==typeof r&&(r>a-Math.max(2,o)?c.push(0):h=a-r);let f=!1,m=-1,g=-1,v=-1;n===H&&(m=K(e,m,t.length),-1!==m&&(h=m+l));for(let r;r=e[m+=1];){if(n===Z&&"\\"===r){switch(g=m,e[m+1]){case"x":m+=3;break;case"u":m+=5;break;case"U":m+=9;break;default:m+=1}v=m}if("\n"===r)n===H&&(m=K(e,m,t.length)),h=m+t.length+l,d=void 0;else{if(" "===r&&p&&" "!==p&&"\n"!==p&&"\t"!==p){const t=e[m+1];t&&" "!==t&&"\n"!==t&&"\t"!==t&&(d=m)}if(m>=h)if(d)c.push(d),h=d+l,d=void 0;else if(n===Z){for(;" "===p||"\t"===p;)p=r,r=e[m+=1],f=!0;const t=m>v+1?m-2:g-1;if(u[t])return e;c.push(t),u[t]=!0,h=t+l,d=void 0}else f=!0}p=r}if(f&&s&&s(),0===c.length)return e;i&&i();let b=e.slice(0,c[0]);for(let r=0;r<c.length;++r){const a=c[r],o=c[r+1]||e.length;0===a?b=`\n${t}${e.slice(0,o)}`:(n===Z&&u[a]&&(b+=`${e[a]}\\`),b+=`\n${t}${e.slice(a+1,o)}`)}return b}function K(e,t,n){let r=t,a=t+1,o=e[a];for(;" "===o||"\t"===o;)if(t<a+n)o=e[++t];else{do{o=e[++t]}while(o&&"\n"!==o);r=t,a=t+1,o=e[a]}return r}const Y=(e,t)=>({indentAtStart:t?e.indent.length:e.indentAtStart,lineWidth:e.options.lineWidth,minContentWidth:e.options.minContentWidth}),J=e=>/^(%|---|\.\.\.)/m.test(e);function ee(e,t){const n=JSON.stringify(e);if(t.options.doubleQuotedAsJSON)return n;const{implicitKey:r}=t,a=t.options.doubleQuotedMinMultiLineLength,o=t.indent||(J(e)?"  ":"");let i="",s=0;for(let e=0,t=n[e];t;t=n[++e])if(" "===t&&"\\"===n[e+1]&&"n"===n[e+2]&&(i+=n.slice(s,e)+"\\ ",e+=1,s=e,t="\\"),"\\"===t)switch(n[e+1]){case"u":{i+=n.slice(s,e);const t=n.substr(e+2,4);switch(t){case"0000":i+="\\0";break;case"0007":i+="\\a";break;case"000b":i+="\\v";break;case"001b":i+="\\e";break;case"0085":i+="\\N";break;case"00a0":i+="\\_";break;case"2028":i+="\\L";break;case"2029":i+="\\P";break;default:"00"===t.substr(0,2)?i+="\\x"+t.substr(2):i+=n.substr(e,6)}e+=5,s=e+1}break;case"n":if(r||'"'===n[e+2]||n.length<a)e+=1;else{for(i+=n.slice(s,e)+"\n\n";"\\"===n[e+2]&&"n"===n[e+3]&&'"'!==n[e+4];)i+="\n",e+=2;i+=o," "===n[e+2]&&(i+="\\"),e+=1,s=e+1}break;default:e+=1}return i=s?i+n.slice(s):n,r?i:G(i,o,Z,Y(t,!1))}function te(e,t){if(!1===t.options.singleQuote||t.implicitKey&&e.includes("\n")||/[ \t]\n|\n[ \t]/.test(e))return ee(e,t);const n=t.indent||(J(e)?"  ":""),r="'"+e.replace(/'/g,"''").replace(/\n+/g,`$&\n${n}`)+"'";return t.implicitKey?r:G(r,n,V,Y(t,!1))}function ne(e,t){const{singleQuote:n}=t.options;let r;if(!1===n)r=ee;else{const t=e.includes('"'),a=e.includes("'");r=t&&!a?te:a&&!t?ee:n?te:ee}return r(e,t)}let re;try{re=new RegExp("(^|(?<!\n))\n+(?!\n|$)","g")}catch{re=/\n+(?!\n|$)/g}function ae({comment:e,type:t,value:n},r,a,o){const{blockQuote:i,commentString:s,lineWidth:l}=r.options;if(!i||/\n[\t ]+$/.test(n)||/^\s*$/.test(n))return ne(n,r);const c=r.indent||(r.forceBlockIndent||J(n)?"  ":""),u="literal"===i||"folded"!==i&&t!==F.BLOCK_FOLDED&&(t===F.BLOCK_LITERAL||!function(e,t,n){if(!t||t<0)return!1;const r=t-n,a=e.length;if(a<=r)return!1;for(let t=0,n=0;t<a;++t)if("\n"===e[t]){if(t-n>r)return!0;if(n=t+1,a-n<=r)return!1}return!0}(n,l,c.length));if(!n)return u?"|\n":">\n";let d,p;for(p=n.length;p>0;--p){const e=n[p-1];if("\n"!==e&&"\t"!==e&&" "!==e)break}let h=n.substring(p);const f=h.indexOf("\n");-1===f?d="-":n===h||f!==h.length-1?(d="+",o&&o()):d="",h&&(n=n.slice(0,-h.length),"\n"===h[h.length-1]&&(h=h.slice(0,-1)),h=h.replace(re,`$&${c}`));let m,g=!1,v=-1;for(m=0;m<n.length;++m){const e=n[m];if(" "===e)g=!0;else{if("\n"!==e)break;v=m}}let b=n.substring(0,v<m?v+1:m);b&&(n=n.substring(b.length),b=b.replace(/\n+/g,`$&${c}`));let y=(g?c?"2":"1":"")+d;if(e&&(y+=" "+s(e.replace(/ ?[\r\n]+/g," ")),a&&a()),!u){const e=n.replace(/\n+/g,"\n$&").replace(/(?:^|\n)([\t ].*)(?:([\n\t ]*)\n(?![\n\t ]))?/g,"$1$2").replace(/\n+/g,`$&${c}`);let a=!1;const o=Y(r,!0);"folded"!==i&&t!==F.BLOCK_FOLDED&&(o.onOverflow=()=>{a=!0});const s=G(`${b}${e}${h}`,c,H,o);if(!a)return`>${y}\n${c}${s}`}return`|${y}\n${c}${b}${n=n.replace(/\n+/g,`$&${c}`)}${h}`}function oe(e,t,n,r){const{implicitKey:a,inFlow:o}=t,i="string"==typeof e.value?e:Object.assign({},e,{value:String(e.value)});let{type:s}=e;s!==F.QUOTE_DOUBLE&&/[\x00-\x08\x0b-\x1f\x7f-\x9f\u{D800}-\u{DFFF}]/u.test(i.value)&&(s=F.QUOTE_DOUBLE);const l=e=>{switch(e){case F.BLOCK_FOLDED:case F.BLOCK_LITERAL:return a||o?ne(i.value,t):ae(i,t,n,r);case F.QUOTE_DOUBLE:return ee(i.value,t);case F.QUOTE_SINGLE:return te(i.value,t);case F.PLAIN:return function(e,t,n,r){const{type:a,value:o}=e,{actualString:i,implicitKey:s,indent:l,indentStep:c,inFlow:u}=t;if(s&&o.includes("\n")||u&&/[[\]{},]/.test(o))return ne(o,t);if(!o||/^[\n\t ,[\]{}#&*!|>'"%@`]|^[?-]$|^[?-][ \t]|[\n:][ \t]|[ \t]\n|[\n\t ]#|[\n\t :]$/.test(o))return s||u||!o.includes("\n")?ne(o,t):ae(e,t,n,r);if(!s&&!u&&a!==F.PLAIN&&o.includes("\n"))return ae(e,t,n,r);if(J(o)){if(""===l)return t.forceBlockIndent=!0,ae(e,t,n,r);if(s&&l===c)return ne(o,t)}const d=o.replace(/\n+/g,`$&\n${l}`);if(i){const e=e=>e.default&&"tag:yaml.org,2002:str"!==e.tag&&e.test?.test(d),{compat:n,tags:r}=t.doc.schema;if(r.some(e)||n?.some(e))return ne(o,t)}return s?d:G(d,l,V,Y(t,!1))}(i,t,n,r);default:return null}};let c=l(s);if(null===c){const{defaultKeyType:e,defaultStringType:n}=t.options,r=a&&e||n;if(c=l(r),null===c)throw new Error(`Unsupported default string type ${r}`)}return c}function ie(e,t){const n=Object.assign({blockQuote:!0,commentString:W,defaultKeyType:null,defaultStringType:"PLAIN",directives:null,doubleQuotedAsJSON:!1,doubleQuotedMinMultiLineLength:40,falseStr:"false",flowCollectionPadding:!0,indentSeq:!0,lineWidth:80,minContentWidth:20,nullStr:"null",simpleKeys:!1,singleQuote:null,trueStr:"true",verifyAliasOrder:!0},e.schema.toStringOptions,t);let r;switch(n.collectionStyle){case"block":r=!1;break;case"flow":r=!0;break;default:r=null}return{anchors:new Set,doc:e,flowCollectionPadding:n.flowCollectionPadding?" ":"",indent:"",indentStep:"number"==typeof n.indent?" ".repeat(n.indent):"  ",inFlow:r,options:n}}function se(e,t,n,r){if(h(e))return e.toString(t,n,r);if(u(e)){if(t.doc.directives)return e.toString(t);if(t.resolvedAliases?.has(e))throw new TypeError("Cannot stringify circular structure without alias nodes");t.resolvedAliases?t.resolvedAliases.add(e):t.resolvedAliases=new Set([e]),e=e.resolve(t.doc)}let a;const o=v(e)?e:t.doc.createNode(e,{onTagObj:e=>a=e});a||(a=function(e,t){if(t.tag){const n=e.filter((e=>e.tag===t.tag));if(n.length>0)return n.find((e=>e.format===t.format))??n[0]}let n,r;if(f(t)){r=t.value;let a=e.filter((e=>e.identify?.(r)));if(a.length>1){const e=a.filter((e=>e.test));e.length>0&&(a=e)}n=a.find((e=>e.format===t.format))??a.find((e=>!e.format))}else r=t,n=e.find((e=>e.nodeClass&&r instanceof e.nodeClass));if(!n)throw new Error(`Tag not resolved for ${r?.constructor?.name??typeof r} value`);return n}(t.doc.schema.tags,o));const i=function(e,t,{anchors:n,doc:r}){if(!r.directives)return"";const a=[],o=(f(e)||g(e))&&e.anchor;o&&R(o)&&(n.add(o),a.push(`&${o}`));const i=e.tag?e.tag:t.default?null:t.tag;return i&&a.push(r.directives.tagString(i)),a.join(" ")}(o,a,t);i.length>0&&(t.indentAtStart=(t.indentAtStart??0)+i.length+1);const s="function"==typeof a.stringify?a.stringify(o,t,n,r):f(o)?oe(o,t,n,r):o.toString(t,n,r);return i?f(o)||"{"===s[0]||"["===s[0]?`${i} ${s}`:`${i}\n${t.indent}${s}`:s}function le(e,t){"debug"!==e&&"warn"!==e||console.warn(t)}const ce="<<",ue={identify:e=>e===ce||"symbol"==typeof e&&e.description===ce,default:"key",tag:"tag:yaml.org,2002:merge",test:/^<<$/,resolve:()=>Object.assign(new F(Symbol(ce)),{addToJSMap:de}),stringify:()=>ce};function de(e,t,n){if(n=e&&u(n)?n.resolve(e.doc):n,m(n))for(const r of n.items)pe(e,t,r);else if(Array.isArray(n))for(const r of n)pe(e,t,r);else pe(e,t,n)}function pe(e,t,n){const r=e&&u(n)?n.resolve(e.doc):n;if(!p(r))throw new Error("Merge sources must be maps or map aliases");const a=r.toJSON(null,e,Map);for(const[e,n]of a)t instanceof Map?t.has(e)||t.set(e,n):t instanceof Set?t.add(e):Object.prototype.hasOwnProperty.call(t,e)||Object.defineProperty(t,e,{value:n,writable:!0,enumerable:!0,configurable:!0});return t}function he(e,t,{key:n,value:r}){if(v(n)&&n.addToJSMap)n.addToJSMap(e,t,r);else if(((e,t)=>(ue.identify(t)||f(t)&&(!t.type||t.type===F.PLAIN)&&ue.identify(t.value))&&e?.doc.schema.tags.some((e=>e.tag===ue.tag&&e.default)))(e,n))de(e,t,r);else{const a=$(n,"",e);if(t instanceof Map)t.set(a,$(r,a,e));else if(t instanceof Set)t.add(a);else{const o=function(e,t,n){if(null===t)return"";if("object"!=typeof t)return String(t);if(v(e)&&n?.doc){const t=ie(n.doc,{});t.anchors=new Set;for(const e of n.anchors.keys())t.anchors.add(e.anchor);t.inFlow=!0,t.inStringifyKey=!0;const r=e.toString(t);if(!n.mapKeyWarned){let e=JSON.stringify(r);e.length>40&&(e=e.substring(0,36)+'..."'),le(n.doc.options.logLevel,`Keys with collection values will be stringified due to JS Object restrictions: ${e}. Set mapAsMap: true to use object keys.`),n.mapKeyWarned=!0}return r}return JSON.stringify(t)}(n,a,e),i=$(r,o,e);o in t?Object.defineProperty(t,o,{value:i,writable:!0,enumerable:!0,configurable:!0}):t[o]=i}}return t}function fe(e,t,n){const r=Q(e,void 0,n),a=Q(t,void 0,n);return new me(r,a)}class me{constructor(e,t=null){Object.defineProperty(this,c,{value:i}),this.key=e,this.value=t}clone(e){let{key:t,value:n}=this;return v(t)&&(t=t.clone(e)),v(n)&&(n=n.clone(e)),new me(t,n)}toJSON(e,t){return he(t,t?.mapAsMap?new Map:{},this)}toString(e,t,n){return e?.doc?function({key:e,value:t},n,r,a){const{allNullValues:o,doc:i,indent:s,indentStep:l,options:{commentString:c,indentSeq:u,simpleKeys:d}}=n;let p=v(e)&&e.comment||null;if(d){if(p)throw new Error("With simple keys, key nodes cannot have comments");if(g(e)||!v(e)&&"object"==typeof e)throw new Error("With simple keys, collection cannot be used as a key value")}let h=!d&&(!e||p&&null==t&&!n.inFlow||g(e)||(f(e)?e.type===F.BLOCK_FOLDED||e.type===F.BLOCK_LITERAL:"object"==typeof e));n=Object.assign({},n,{allNullValues:!1,implicitKey:!h&&(d||!o),indent:s+l});let b,y,w,O=!1,k=!1,x=se(e,n,(()=>O=!0),(()=>k=!0));if(!h&&!n.inFlow&&x.length>1024){if(d)throw new Error("With simple keys, single line scalar must not span more than 1024 characters");h=!0}if(n.inFlow){if(o||null==t)return O&&r&&r(),""===x?"?":h?`? ${x}`:x}else if(o&&!d||null==t&&h)return x=`? ${x}`,p&&!O?x+=q(x,n.indent,c(p)):k&&a&&a(),x;O&&(p=null),h?(p&&(x+=q(x,n.indent,c(p))),x=`? ${x}\n${s}:`):(x=`${x}:`,p&&(x+=q(x,n.indent,c(p)))),v(t)?(b=!!t.spaceBefore,y=t.commentBefore,w=t.comment):(b=!1,y=null,w=null,t&&"object"==typeof t&&(t=i.createNode(t))),n.implicitKey=!1,h||p||!f(t)||(n.indentAtStart=x.length+1),k=!1,u||!(l.length>=2)||n.inFlow||h||!m(t)||t.flow||t.tag||t.anchor||(n.indent=n.indent.substring(2));let E=!1;const S=se(t,n,(()=>E=!0),(()=>k=!0));let _=" ";if(p||b||y)_=b?"\n":"",y&&(_+=`\n${X(c(y),n.indent)}`),""!==S||n.inFlow?_+=`\n${n.indent}`:"\n"===_&&(_="\n\n");else if(!h&&g(t)){const e=S[0],r=S.indexOf("\n"),a=-1!==r,o=n.inFlow??t.flow??0===t.items.length;if(a||!o){let t=!1;if(a&&("&"===e||"!"===e)){let n=S.indexOf(" ");"&"===e&&-1!==n&&n<r&&"!"===S[n+1]&&(n=S.indexOf(" ",n+1)),(-1===n||r<n)&&(t=!0)}t||(_=`\n${n.indent}`)}}else""!==S&&"\n"!==S[0]||(_="");return x+=_+S,n.inFlow?E&&r&&r():w&&!E?x+=q(x,n.indent,c(w)):k&&a&&a(),x}(this,e,t,n):JSON.stringify(this)}}function ge(e,t,n){return(t.inFlow??e.flow?be:ve)(e,t,n)}function ve({comment:e,items:t},n,{blockItemPrefix:r,flowChars:a,itemIndent:o,onChompKeep:i,onComment:s}){const{indent:l,options:{commentString:c}}=n,u=Object.assign({},n,{indent:o,type:null});let d=!1;const p=[];for(let e=0;e<t.length;++e){const a=t[e];let i=null;if(v(a))!d&&a.spaceBefore&&p.push(""),ye(n,p,a.commentBefore,d),a.comment&&(i=a.comment);else if(h(a)){const e=v(a.key)?a.key:null;e&&(!d&&e.spaceBefore&&p.push(""),ye(n,p,e.commentBefore,d))}d=!1;let s=se(a,u,(()=>i=null),(()=>d=!0));i&&(s+=q(s,o,c(i))),d&&i&&(d=!1),p.push(r+s)}let f;if(0===p.length)f=a.start+a.end;else{f=p[0];for(let e=1;e<p.length;++e){const t=p[e];f+=t?`\n${l}${t}`:"\n"}}return e?(f+="\n"+X(c(e),l),s&&s()):d&&i&&i(),f}function be({items:e},t,{flowChars:n,itemIndent:r}){const{indent:a,indentStep:o,flowCollectionPadding:i,options:{commentString:s}}=t;r+=o;const l=Object.assign({},t,{indent:r,inFlow:!0,type:null});let c=!1,u=0;const d=[];for(let n=0;n<e.length;++n){const a=e[n];let o=null;if(v(a))a.spaceBefore&&d.push(""),ye(t,d,a.commentBefore,!1),a.comment&&(o=a.comment);else if(h(a)){const e=v(a.key)?a.key:null;e&&(e.spaceBefore&&d.push(""),ye(t,d,e.commentBefore,!1),e.comment&&(c=!0));const n=v(a.value)?a.value:null;n?(n.comment&&(o=n.comment),n.commentBefore&&(c=!0)):null==a.value&&e?.comment&&(o=e.comment)}o&&(c=!0);let i=se(a,l,(()=>o=null));n<e.length-1&&(i+=","),o&&(i+=q(i,r,s(o))),!c&&(d.length>u||i.includes("\n"))&&(c=!0),d.push(i),u=d.length}const{start:p,end:f}=n;if(0===d.length)return p+f;if(!c){const e=d.reduce(((e,t)=>e+t.length+2),2);c=t.options.lineWidth>0&&e>t.options.lineWidth}if(c){let e=p;for(const t of d)e+=t?`\n${o}${a}${t}`:"\n";return`${e}\n${a}${f}`}return`${p}${i}${d.join(" ")}${i}${f}`}function ye({indent:e,options:{commentString:t}},n,r,a){if(r&&a&&(r=r.replace(/^\n+/,"")),r){const a=X(t(r),e);n.push(a.trimStart())}}function we(e,t){const n=f(t)?t.value:t;for(const r of e)if(h(r)){if(r.key===t||r.key===n)return r;if(f(r.key)&&r.key.value===n)return r}}class Oe extends j{static get tagName(){return"tag:yaml.org,2002:map"}constructor(e){super(o,e),this.items=[]}static from(e,t,n){const{keepUndefined:r,replacer:a}=n,o=new this(e),i=(e,i)=>{if("function"==typeof a)i=a.call(t,e,i);else if(Array.isArray(a)&&!a.includes(e))return;(void 0!==i||r)&&o.items.push(fe(e,i,n))};if(t instanceof Map)for(const[e,n]of t)i(e,n);else if(t&&"object"==typeof t)for(const e of Object.keys(t))i(e,t[e]);return"function"==typeof e.sortMapEntries&&o.items.sort(e.sortMapEntries),o}add(e,t){let n;n=h(e)?e:e&&"object"==typeof e&&"key"in e?new me(e.key,e.value):new me(e,e?.value);const r=we(this.items,n.key),a=this.schema?.sortMapEntries;if(r){if(!t)throw new Error(`Key ${n.key} already set`);f(r.value)&&B(n.value)?r.value.value=n.value:r.value=n.value}else if(a){const e=this.items.findIndex((e=>a(n,e)<0));-1===e?this.items.push(n):this.items.splice(e,0,n)}else this.items.push(n)}delete(e){const t=we(this.items,e);return!!t&&this.items.splice(this.items.indexOf(t),1).length>0}get(e,t){const n=we(this.items,e),r=n?.value;return(!t&&f(r)?r.value:r)??void 0}has(e){return!!we(this.items,e)}set(e,t){this.add(new me(e,t),!0)}toJSON(e,t,n){const r=n?new n:t?.mapAsMap?new Map:{};t?.onCreate&&t.onCreate(r);for(const e of this.items)he(t,r,e);return r}toString(e,t,n){if(!e)return JSON.stringify(this);for(const e of this.items)if(!h(e))throw new Error(`Map items must all be pairs; found ${JSON.stringify(e)} instead`);return!e.allNullValues&&this.hasAllNullValues(!1)&&(e=Object.assign({},e,{allNullValues:!0})),ge(this,e,{blockItemPrefix:"",flowChars:{start:"{",end:"}"},itemIndent:e.indent||"",onChompKeep:n,onComment:t})}}const ke={collection:"map",default:!0,nodeClass:Oe,tag:"tag:yaml.org,2002:map",resolve:(e,t)=>(p(e)||t("Expected a mapping for this tag"),e),createNode:(e,t,n)=>Oe.from(e,t,n)};class xe extends j{static get tagName(){return"tag:yaml.org,2002:seq"}constructor(e){super(l,e),this.items=[]}add(e){this.items.push(e)}delete(e){const t=Ee(e);return"number"==typeof t&&this.items.splice(t,1).length>0}get(e,t){const n=Ee(e);if("number"!=typeof n)return;const r=this.items[n];return!t&&f(r)?r.value:r}has(e){const t=Ee(e);return"number"==typeof t&&t<this.items.length}set(e,t){const n=Ee(e);if("number"!=typeof n)throw new Error(`Expected a valid index, not ${e}.`);const r=this.items[n];f(r)&&B(t)?r.value=t:this.items[n]=t}toJSON(e,t){const n=[];t?.onCreate&&t.onCreate(n);let r=0;for(const e of this.items)n.push($(e,String(r++),t));return n}toString(e,t,n){return e?ge(this,e,{blockItemPrefix:"- ",flowChars:{start:"[",end:"]"},itemIndent:(e.indent||"")+"  ",onChompKeep:n,onComment:t}):JSON.stringify(this)}static from(e,t,n){const{replacer:r}=n,a=new this(e);if(t&&Symbol.iterator in Object(t)){let e=0;for(let o of t){if("function"==typeof r){const n=t instanceof Set?o:String(e++);o=r.call(t,n,o)}a.items.push(Q(o,void 0,n))}}return a}}function Ee(e){let t=f(e)?e.value:e;return t&&"string"==typeof t&&(t=Number(t)),"number"==typeof t&&Number.isInteger(t)&&t>=0?t:null}const Se={collection:"seq",default:!0,nodeClass:xe,tag:"tag:yaml.org,2002:seq",resolve:(e,t)=>(m(e)||t("Expected a sequence for this tag"),e),createNode:(e,t,n)=>xe.from(e,t,n)},_e={identify:e=>"string"==typeof e,default:!0,tag:"tag:yaml.org,2002:str",resolve:e=>e,stringify:(e,t,n,r)=>oe(e,t=Object.assign({actualString:!0},t),n,r)},Ce={identify:e=>null==e,createNode:()=>new F(null),default:!0,tag:"tag:yaml.org,2002:null",test:/^(?:~|[Nn]ull|NULL)?$/,resolve:()=>new F(null),stringify:({source:e},t)=>"string"==typeof e&&Ce.test.test(e)?e:t.options.nullStr},Te={identify:e=>"boolean"==typeof e,default:!0,tag:"tag:yaml.org,2002:bool",test:/^(?:[Tt]rue|TRUE|[Ff]alse|FALSE)$/,resolve:e=>new F("t"===e[0]||"T"===e[0]),stringify:({source:e,value:t},n)=>e&&Te.test.test(e)&&t===("t"===e[0]||"T"===e[0])?e:t?n.options.trueStr:n.options.falseStr};function Ae({format:e,minFractionDigits:t,tag:n,value:r}){if("bigint"==typeof r)return String(r);const a="number"==typeof r?r:Number(r);if(!isFinite(a))return isNaN(a)?".nan":a<0?"-.inf":".inf";let o=JSON.stringify(r);if(!e&&t&&(!n||"tag:yaml.org,2002:float"===n)&&/^\d/.test(o)){let e=o.indexOf(".");e<0&&(e=o.length,o+=".");let n=t-(o.length-e-1);for(;n-- >0;)o+="0"}return o}const Re={identify:e=>"number"==typeof e,default:!0,tag:"tag:yaml.org,2002:float",test:/^(?:[-+]?\.(?:inf|Inf|INF)|\.nan|\.NaN|\.NAN)$/,resolve:e=>"nan"===e.slice(-3).toLowerCase()?NaN:"-"===e[0]?Number.NEGATIVE_INFINITY:Number.POSITIVE_INFINITY,stringify:Ae},Pe={identify:e=>"number"==typeof e,default:!0,tag:"tag:yaml.org,2002:float",format:"EXP",test:/^[-+]?(?:\.[0-9]+|[0-9]+(?:\.[0-9]*)?)[eE][-+]?[0-9]+$/,resolve:e=>parseFloat(e),stringify(e){const t=Number(e.value);return isFinite(t)?t.toExponential():Ae(e)}},De={identify:e=>"number"==typeof e,default:!0,tag:"tag:yaml.org,2002:float",test:/^[-+]?(?:\.[0-9]+|[0-9]+\.[0-9]*)$/,resolve(e){const t=new F(parseFloat(e)),n=e.indexOf(".");return-1!==n&&"0"===e[e.length-1]&&(t.minFractionDigits=e.length-n-1),t},stringify:Ae},Me=e=>"bigint"==typeof e||Number.isInteger(e),$e=(e,t,n,{intAsBigInt:r})=>r?BigInt(e):parseInt(e.substring(t),n);function Ne(e,t,n){const{value:r}=e;return Me(r)&&r>=0?n+r.toString(t):Ae(e)}const Ie={identify:e=>Me(e)&&e>=0,default:!0,tag:"tag:yaml.org,2002:int",format:"OCT",test:/^0o[0-7]+$/,resolve:(e,t,n)=>$e(e,2,8,n),stringify:e=>Ne(e,8,"0o")},Le={identify:Me,default:!0,tag:"tag:yaml.org,2002:int",test:/^[-+]?[0-9]+$/,resolve:(e,t,n)=>$e(e,0,10,n),stringify:Ae},Be={identify:e=>Me(e)&&e>=0,default:!0,tag:"tag:yaml.org,2002:int",format:"HEX",test:/^0x[0-9a-fA-F]+$/,resolve:(e,t,n)=>$e(e,2,16,n),stringify:e=>Ne(e,16,"0x")},Fe=[ke,Se,_e,Ce,Te,Ie,Le,Be,Re,Pe,De];function Qe(e){return"bigint"==typeof e||Number.isInteger(e)}const Ue=({value:e})=>JSON.stringify(e),ze=[ke,Se].concat([{identify:e=>"string"==typeof e,default:!0,tag:"tag:yaml.org,2002:str",resolve:e=>e,stringify:Ue},{identify:e=>null==e,createNode:()=>new F(null),default:!0,tag:"tag:yaml.org,2002:null",test:/^null$/,resolve:()=>null,stringify:Ue},{identify:e=>"boolean"==typeof e,default:!0,tag:"tag:yaml.org,2002:bool",test:/^true$|^false$/,resolve:e=>"true"===e,stringify:Ue},{identify:Qe,default:!0,tag:"tag:yaml.org,2002:int",test:/^-?(?:0|[1-9][0-9]*)$/,resolve:(e,t,{intAsBigInt:n})=>n?BigInt(e):parseInt(e,10),stringify:({value:e})=>Qe(e)?e.toString():JSON.stringify(e)},{identify:e=>"number"==typeof e,default:!0,tag:"tag:yaml.org,2002:float",test:/^-?(?:0|[1-9][0-9]*)(?:\.[0-9]*)?(?:[eE][-+]?[0-9]+)?$/,resolve:e=>parseFloat(e),stringify:Ue}],{default:!0,tag:"",test:/^/,resolve:(e,t)=>(t(`Unresolved plain scalar ${JSON.stringify(e)}`),e)}),je={identify:e=>e instanceof Uint8Array,default:!1,tag:"tag:yaml.org,2002:binary",resolve(e,t){if("function"==typeof atob){const t=atob(e.replace(/[\n\r]/g,"")),n=new Uint8Array(t.length);for(let e=0;e<t.length;++e)n[e]=t.charCodeAt(e);return n}return t("This environment does not support reading binary tags; either Buffer or atob is required"),e},stringify({comment:e,type:t,value:n},r,a,o){const i=n;let s;if("function"!=typeof btoa)throw new Error("This environment does not support writing binary tags; either Buffer or btoa is required");{let e="";for(let t=0;t<i.length;++t)e+=String.fromCharCode(i[t]);s=btoa(e)}if(t||(t=F.BLOCK_LITERAL),t!==F.QUOTE_DOUBLE){const e=Math.max(r.options.lineWidth-r.indent.length,r.options.minContentWidth),n=Math.ceil(s.length/e),a=new Array(n);for(let t=0,r=0;t<n;++t,r+=e)a[t]=s.substr(r,e);s=a.join(t===F.BLOCK_LITERAL?"\n":" ")}return oe({comment:e,type:t,value:s},r,a,o)}};function We(e,t){if(m(e))for(let n=0;n<e.items.length;++n){let r=e.items[n];if(!h(r)){if(p(r)){r.items.length>1&&t("Each pair must have its own sequence indicator");const e=r.items[0]||new me(new F(null));if(r.commentBefore&&(e.key.commentBefore=e.key.commentBefore?`${r.commentBefore}\n${e.key.commentBefore}`:r.commentBefore),r.comment){const t=e.value??e.key;t.comment=t.comment?`${r.comment}\n${t.comment}`:r.comment}r=e}e.items[n]=h(r)?r:new me(r)}}else t("Expected a sequence for this tag");return e}function Xe(e,t,n){const{replacer:r}=n,a=new xe(e);a.tag="tag:yaml.org,2002:pairs";let o=0;if(t&&Symbol.iterator in Object(t))for(let e of t){let i,s;if("function"==typeof r&&(e=r.call(t,String(o++),e)),Array.isArray(e)){if(2!==e.length)throw new TypeError(`Expected [key, value] tuple: ${e}`);i=e[0],s=e[1]}else if(e&&e instanceof Object){const t=Object.keys(e);if(1!==t.length)throw new TypeError(`Expected tuple with one key, not ${t.length} keys`);i=t[0],s=e[i]}else i=e;a.items.push(fe(i,s,n))}return a}const qe={collection:"seq",default:!1,tag:"tag:yaml.org,2002:pairs",resolve:We,createNode:Xe};class Ve extends xe{constructor(){super(),this.add=Oe.prototype.add.bind(this),this.delete=Oe.prototype.delete.bind(this),this.get=Oe.prototype.get.bind(this),this.has=Oe.prototype.has.bind(this),this.set=Oe.prototype.set.bind(this),this.tag=Ve.tag}toJSON(e,t){if(!t)return super.toJSON(e);const n=new Map;t?.onCreate&&t.onCreate(n);for(const e of this.items){let r,a;if(h(e)?(r=$(e.key,"",t),a=$(e.value,r,t)):r=$(e,"",t),n.has(r))throw new Error("Ordered maps must not include duplicate keys");n.set(r,a)}return n}static from(e,t,n){const r=Xe(e,t,n),a=new this;return a.items=r.items,a}}Ve.tag="tag:yaml.org,2002:omap";const He={collection:"seq",identify:e=>e instanceof Map,nodeClass:Ve,default:!1,tag:"tag:yaml.org,2002:omap",resolve(e,t){const n=We(e,t),r=[];for(const{key:e}of n.items)f(e)&&(r.includes(e.value)?t(`Ordered maps must not include duplicate keys: ${e.value}`):r.push(e.value));return Object.assign(new Ve,n)},createNode:(e,t,n)=>Ve.from(e,t,n)};function Ze({value:e,source:t},n){return t&&(e?Ge:Ke).test.test(t)?t:e?n.options.trueStr:n.options.falseStr}const Ge={identify:e=>!0===e,default:!0,tag:"tag:yaml.org,2002:bool",test:/^(?:Y|y|[Yy]es|YES|[Tt]rue|TRUE|[Oo]n|ON)$/,resolve:()=>new F(!0),stringify:Ze},Ke={identify:e=>!1===e,default:!0,tag:"tag:yaml.org,2002:bool",test:/^(?:N|n|[Nn]o|NO|[Ff]alse|FALSE|[Oo]ff|OFF)$/,resolve:()=>new F(!1),stringify:Ze},Ye={identify:e=>"number"==typeof e,default:!0,tag:"tag:yaml.org,2002:float",test:/^(?:[-+]?\.(?:inf|Inf|INF)|\.nan|\.NaN|\.NAN)$/,resolve:e=>"nan"===e.slice(-3).toLowerCase()?NaN:"-"===e[0]?Number.NEGATIVE_INFINITY:Number.POSITIVE_INFINITY,stringify:Ae},Je={identify:e=>"number"==typeof e,default:!0,tag:"tag:yaml.org,2002:float",format:"EXP",test:/^[-+]?(?:[0-9][0-9_]*)?(?:\.[0-9_]*)?[eE][-+]?[0-9]+$/,resolve:e=>parseFloat(e.replace(/_/g,"")),stringify(e){const t=Number(e.value);return isFinite(t)?t.toExponential():Ae(e)}},et={identify:e=>"number"==typeof e,default:!0,tag:"tag:yaml.org,2002:float",test:/^[-+]?(?:[0-9][0-9_]*)?\.[0-9_]*$/,resolve(e){const t=new F(parseFloat(e.replace(/_/g,""))),n=e.indexOf(".");if(-1!==n){const r=e.substring(n+1).replace(/_/g,"");"0"===r[r.length-1]&&(t.minFractionDigits=r.length)}return t},stringify:Ae},tt=e=>"bigint"==typeof e||Number.isInteger(e);function nt(e,t,n,{intAsBigInt:r}){const a=e[0];if("-"!==a&&"+"!==a||(t+=1),e=e.substring(t).replace(/_/g,""),r){switch(n){case 2:e=`0b${e}`;break;case 8:e=`0o${e}`;break;case 16:e=`0x${e}`}const t=BigInt(e);return"-"===a?BigInt(-1)*t:t}const o=parseInt(e,n);return"-"===a?-1*o:o}function rt(e,t,n){const{value:r}=e;if(tt(r)){const e=r.toString(t);return r<0?"-"+n+e.substr(1):n+e}return Ae(e)}const at={identify:tt,default:!0,tag:"tag:yaml.org,2002:int",format:"BIN",test:/^[-+]?0b[0-1_]+$/,resolve:(e,t,n)=>nt(e,2,2,n),stringify:e=>rt(e,2,"0b")},ot={identify:tt,default:!0,tag:"tag:yaml.org,2002:int",format:"OCT",test:/^[-+]?0[0-7_]+$/,resolve:(e,t,n)=>nt(e,1,8,n),stringify:e=>rt(e,8,"0")},it={identify:tt,default:!0,tag:"tag:yaml.org,2002:int",test:/^[-+]?[0-9][0-9_]*$/,resolve:(e,t,n)=>nt(e,0,10,n),stringify:Ae},st={identify:tt,default:!0,tag:"tag:yaml.org,2002:int",format:"HEX",test:/^[-+]?0x[0-9a-fA-F_]+$/,resolve:(e,t,n)=>nt(e,2,16,n),stringify:e=>rt(e,16,"0x")};class lt extends Oe{constructor(e){super(e),this.tag=lt.tag}add(e){let t;t=h(e)?e:e&&"object"==typeof e&&"key"in e&&"value"in e&&null===e.value?new me(e.key,null):new me(e,null),we(this.items,t.key)||this.items.push(t)}get(e,t){const n=we(this.items,e);return!t&&h(n)?f(n.key)?n.key.value:n.key:n}set(e,t){if("boolean"!=typeof t)throw new Error("Expected boolean value for set(key, value) in a YAML set, not "+typeof t);const n=we(this.items,e);n&&!t?this.items.splice(this.items.indexOf(n),1):!n&&t&&this.items.push(new me(e))}toJSON(e,t){return super.toJSON(e,t,Set)}toString(e,t,n){if(!e)return JSON.stringify(this);if(this.hasAllNullValues(!0))return super.toString(Object.assign({},e,{allNullValues:!0}),t,n);throw new Error("Set items must all have null values")}static from(e,t,n){const{replacer:r}=n,a=new this(e);if(t&&Symbol.iterator in Object(t))for(let e of t)"function"==typeof r&&(e=r.call(t,e,e)),a.items.push(fe(e,null,n));return a}}lt.tag="tag:yaml.org,2002:set";const ct={collection:"map",identify:e=>e instanceof Set,nodeClass:lt,default:!1,tag:"tag:yaml.org,2002:set",createNode:(e,t,n)=>lt.from(e,t,n),resolve(e,t){if(p(e)){if(e.hasAllNullValues(!0))return Object.assign(new lt,e);t("Set items must all have null values")}else t("Expected a mapping for this tag");return e}};function ut(e,t){const n=e[0],r="-"===n||"+"===n?e.substring(1):e,a=e=>t?BigInt(e):Number(e),o=r.replace(/_/g,"").split(":").reduce(((e,t)=>e*a(60)+a(t)),a(0));return"-"===n?a(-1)*o:o}function dt(e){let{value:t}=e,n=e=>e;if("bigint"==typeof t)n=e=>BigInt(e);else if(isNaN(t)||!isFinite(t))return Ae(e);let r="";t<0&&(r="-",t*=n(-1));const a=n(60),o=[t%a];return t<60?o.unshift(0):(t=(t-o[0])/a,o.unshift(t%a),t>=60&&(t=(t-o[0])/a,o.unshift(t))),r+o.map((e=>String(e).padStart(2,"0"))).join(":").replace(/000000\d*$/,"")}const pt={identify:e=>"bigint"==typeof e||Number.isInteger(e),default:!0,tag:"tag:yaml.org,2002:int",format:"TIME",test:/^[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+$/,resolve:(e,t,{intAsBigInt:n})=>ut(e,n),stringify:dt},ht={identify:e=>"number"==typeof e,default:!0,tag:"tag:yaml.org,2002:float",format:"TIME",test:/^[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\.[0-9_]*$/,resolve:e=>ut(e,!1),stringify:dt},ft={identify:e=>e instanceof Date,default:!0,tag:"tag:yaml.org,2002:timestamp",test:RegExp("^([0-9]{4})-([0-9]{1,2})-([0-9]{1,2})(?:(?:t|T|[ \\t]+)([0-9]{1,2}):([0-9]{1,2}):([0-9]{1,2}(\\.[0-9]+)?)(?:[ \\t]*(Z|[-+][012]?[0-9](?::[0-9]{2})?))?)?$"),resolve(e){const t=e.match(ft.test);if(!t)throw new Error("!!timestamp expects a date, starting with yyyy-mm-dd");const[,n,r,a,o,i,s]=t.map(Number),l=t[7]?Number((t[7]+"00").substr(1,3)):0;let c=Date.UTC(n,r-1,a,o||0,i||0,s||0,l);const u=t[8];if(u&&"Z"!==u){let e=ut(u,!1);Math.abs(e)<30&&(e*=60),c-=6e4*e}return new Date(c)},stringify:({value:e})=>e.toISOString().replace(/(T00:00:00)?\.000Z$/,"")},mt=[ke,Se,_e,Ce,Ge,Ke,at,ot,it,st,Ye,Je,et,je,ue,He,qe,ct,pt,ht,ft],gt=new Map([["core",Fe],["failsafe",[ke,Se,_e]],["json",ze],["yaml11",mt],["yaml-1.1",mt]]),vt={binary:je,bool:Te,float:De,floatExp:Pe,floatNaN:Re,floatTime:ht,int:Le,intHex:Be,intOct:Ie,intTime:pt,map:ke,merge:ue,null:Ce,omap:He,pairs:qe,seq:Se,set:ct,timestamp:ft},bt={"tag:yaml.org,2002:binary":je,"tag:yaml.org,2002:merge":ue,"tag:yaml.org,2002:omap":He,"tag:yaml.org,2002:pairs":qe,"tag:yaml.org,2002:set":ct,"tag:yaml.org,2002:timestamp":ft};function yt(e,t,n){const r=gt.get(t);if(r&&!e)return n&&!r.includes(ue)?r.concat(ue):r.slice();let a=r;if(!a){if(!Array.isArray(e)){const e=Array.from(gt.keys()).filter((e=>"yaml11"!==e)).map((e=>JSON.stringify(e))).join(", ");throw new Error(`Unknown schema "${t}"; use one of ${e} or define customTags array`)}a=[]}if(Array.isArray(e))for(const t of e)a=a.concat(t);else"function"==typeof e&&(a=e(a.slice()));return n&&(a=a.concat(ue)),a.reduce(((e,t)=>{const n="string"==typeof t?vt[t]:t;if(!n){const e=JSON.stringify(t),n=Object.keys(vt).map((e=>JSON.stringify(e))).join(", ");throw new Error(`Unknown custom tag ${e}; use one of ${n}`)}return e.includes(n)||e.push(n),e}),[])}const wt=(e,t)=>e.key<t.key?-1:e.key>t.key?1:0;class Ot{constructor({compat:e,customTags:t,merge:n,resolveKnownTags:r,schema:a,sortMapEntries:i,toStringDefaults:c}){this.compat=Array.isArray(e)?yt(e,"compat"):e?yt(null,e):null,this.name="string"==typeof a&&a||"core",this.knownTags=r?bt:{},this.tags=yt(t,this.name,n),this.toStringOptions=c??null,Object.defineProperty(this,o,{value:ke}),Object.defineProperty(this,s,{value:_e}),Object.defineProperty(this,l,{value:Se}),this.sortMapEntries="function"==typeof i?i:!0===i?wt:null}clone(){const e=Object.create(Ot.prototype,Object.getOwnPropertyDescriptors(this));return e.tags=this.tags.slice(),e}}class kt{constructor(e,t,n){this.commentBefore=null,this.comment=null,this.errors=[],this.warnings=[],Object.defineProperty(this,c,{value:a});let r=null;"function"==typeof t||Array.isArray(t)?r=t:void 0===n&&t&&(n=t,t=void 0);const o=Object.assign({intAsBigInt:!1,keepSourceTokens:!1,logLevel:"warn",prettyErrors:!0,strict:!0,stringKeys:!1,uniqueKeys:!0,version:"1.2"},n);this.options=o;let{version:i}=o;n?._directives?(this.directives=n._directives.atDocument(),this.directives.yaml.explicit&&(i=this.directives.yaml.version)):this.directives=new A({version:i}),this.setSchema(i,n),this.contents=void 0===e?null:this.createNode(e,r,n)}clone(){const e=Object.create(kt.prototype,{[c]:{value:a}});return e.commentBefore=this.commentBefore,e.comment=this.comment,e.errors=this.errors.slice(),e.warnings=this.warnings.slice(),e.options=Object.assign({},this.options),this.directives&&(e.directives=this.directives.clone()),e.schema=this.schema.clone(),e.contents=v(this.contents)?this.contents.clone(e.schema):this.contents,this.range&&(e.range=this.range.slice()),e}add(e){xt(this.contents)&&this.contents.add(e)}addIn(e,t){xt(this.contents)&&this.contents.addIn(e,t)}createAlias(e,t){if(!e.anchor){const n=P(this);e.anchor=!t||n.has(t)?D(t||"a",n):t}return new I(e.anchor)}createNode(e,t,n){let r;if("function"==typeof t)e=t.call({"":e},"",e),r=t;else if(Array.isArray(t)){const e=e=>"number"==typeof e||e instanceof String||e instanceof Number,n=t.filter(e).map(String);n.length>0&&(t=t.concat(n)),r=t}else void 0===n&&t&&(n=t,t=void 0);const{aliasDuplicateObjects:a,anchorPrefix:o,flow:i,keepUndefined:s,onTagObj:l,tag:c}=n??{},{onAnchor:u,setAnchors:d,sourceObjects:p}=function(e,t){const n=[],r=new Map;let a=null;return{onAnchor:r=>{n.push(r),a||(a=P(e));const o=D(t,a);return a.add(o),o},setAnchors:()=>{for(const e of n){const t=r.get(e);if("object"!=typeof t||!t.anchor||!f(t.node)&&!g(t.node)){const t=new Error("Failed to resolve repeated object (this should not happen)");throw t.source=e,t}t.node.anchor=t.anchor}},sourceObjects:r}}(this,o||"a"),h=Q(e,c,{aliasDuplicateObjects:a??!0,keepUndefined:s??!1,onAnchor:u,onTagObj:l,replacer:r,schema:this.schema,sourceObjects:p});return i&&g(h)&&(h.flow=!0),d(),h}createPair(e,t,n={}){const r=this.createNode(e,null,n),a=this.createNode(t,null,n);return new me(r,a)}delete(e){return!!xt(this.contents)&&this.contents.delete(e)}deleteIn(e){return z(e)?null!=this.contents&&(this.contents=null,!0):!!xt(this.contents)&&this.contents.deleteIn(e)}get(e,t){return g(this.contents)?this.contents.get(e,t):void 0}getIn(e,t){return z(e)?!t&&f(this.contents)?this.contents.value:this.contents:g(this.contents)?this.contents.getIn(e,t):void 0}has(e){return!!g(this.contents)&&this.contents.has(e)}hasIn(e){return z(e)?void 0!==this.contents:!!g(this.contents)&&this.contents.hasIn(e)}set(e,t){null==this.contents?this.contents=U(this.schema,[e],t):xt(this.contents)&&this.contents.set(e,t)}setIn(e,t){z(e)?this.contents=t:null==this.contents?this.contents=U(this.schema,Array.from(e),t):xt(this.contents)&&this.contents.setIn(e,t)}setSchema(e,t={}){let n;switch("number"==typeof e&&(e=String(e)),e){case"1.1":this.directives?this.directives.yaml.version="1.1":this.directives=new A({version:"1.1"}),n={resolveKnownTags:!1,schema:"yaml-1.1"};break;case"1.2":case"next":this.directives?this.directives.yaml.version=e:this.directives=new A({version:e}),n={resolveKnownTags:!0,schema:"core"};break;case null:this.directives&&delete this.directives,n=null;break;default:{const t=JSON.stringify(e);throw new Error(`Expected '1.1', '1.2' or null as first argument, but found: ${t}`)}}if(t.schema instanceof Object)this.schema=t.schema;else{if(!n)throw new Error("With a null YAML version, the { schema: Schema } option is required");this.schema=new Ot(Object.assign(n,t))}}toJS({json:e,jsonArg:t,mapAsMap:n,maxAliasCount:r,onAnchor:a,reviver:o}={}){const i={anchors:new Map,doc:this,keep:!e,mapAsMap:!0===n,mapKeyWarned:!1,maxAliasCount:"number"==typeof r?r:100},s=$(this.contents,t??"",i);if("function"==typeof a)for(const{count:e,res:t}of i.anchors.values())a(t,e);return"function"==typeof o?M(o,{"":s},"",s):s}toJSON(e,t){return this.toJS({json:!0,jsonArg:e,mapAsMap:!1,onAnchor:t})}toString(e={}){if(this.errors.length>0)throw new Error("Document with errors cannot be stringified");if("indent"in e&&(!Number.isInteger(e.indent)||Number(e.indent)<=0)){const t=JSON.stringify(e.indent);throw new Error(`"indent" option must be a positive integer, not ${t}`)}return function(e,t){const n=[];let r=!0===t.directives;if(!1!==t.directives&&e.directives){const t=e.directives.toString(e);t?(n.push(t),r=!0):e.directives.docStart&&(r=!0)}r&&n.push("---");const a=ie(e,t),{commentString:o}=a.options;if(e.commentBefore){1!==n.length&&n.unshift("");const t=o(e.commentBefore);n.unshift(X(t,""))}let i=!1,s=null;if(e.contents){if(v(e.contents)){if(e.contents.spaceBefore&&r&&n.push(""),e.contents.commentBefore){const t=o(e.contents.commentBefore);n.push(X(t,""))}a.forceBlockIndent=!!e.comment,s=e.contents.comment}const t=s?void 0:()=>i=!0;let l=se(e.contents,a,(()=>s=null),t);s&&(l+=q(l,"",o(s))),"|"!==l[0]&&">"!==l[0]||"---"!==n[n.length-1]?n.push(l):n[n.length-1]=`--- ${l}`}else n.push(se(e.contents,a));if(e.directives?.docEnd)if(e.comment){const t=o(e.comment);t.includes("\n")?(n.push("..."),n.push(X(t,""))):n.push(`... ${t}`)}else n.push("...");else{let t=e.comment;t&&i&&(t=t.replace(/^\n+/,"")),t&&(i&&!s||""===n[n.length-1]||n.push(""),n.push(X(o(t),"")))}return n.join("\n")+"\n"}(this,e)}}function xt(e){if(g(e))return!0;throw new Error("Expected a YAML collection as document contents")}class Et extends Error{constructor(e,t,n,r){super(),this.name=e,this.code=n,this.message=r,this.pos=t}}class St extends Et{constructor(e,t,n){super("YAMLParseError",e,t,n)}}class _t extends Et{constructor(e,t,n){super("YAMLWarning",e,t,n)}}const Ct=(e,t)=>n=>{if(-1===n.pos[0])return;n.linePos=n.pos.map((e=>t.linePos(e)));const{line:r,col:a}=n.linePos[0];n.message+=` at line ${r}, column ${a}`;let o=a-1,i=e.substring(t.lineStarts[r-1],t.lineStarts[r]).replace(/[\n\r]+$/,"");if(o>=60&&i.length>80){const e=Math.min(o-39,i.length-79);i="…"+i.substring(e),o-=e-1}if(i.length>80&&(i=i.substring(0,79)+"…"),r>1&&/^ *$/.test(i.substring(0,o))){let n=e.substring(t.lineStarts[r-2],t.lineStarts[r-1]);n.length>80&&(n=n.substring(0,79)+"…\n"),i=n+i}if(/[^ ]/.test(i)){let e=1;const t=n.linePos[1];t&&t.line===r&&t.col>a&&(e=Math.max(1,Math.min(t.col-a,80-o)));const s=" ".repeat(o)+"^".repeat(e);n.message+=`:\n\n${i}\n${s}\n`}};function Tt(e,{flow:t,indicator:n,next:r,offset:a,onError:o,parentIndent:i,startOnNewline:s}){let l=!1,c=s,u=s,d="",p="",h=!1,f=!1,m=null,g=null,v=null,b=null,y=null,w=null,O=null;for(const a of e)switch(f&&("space"!==a.type&&"newline"!==a.type&&"comma"!==a.type&&o(a.offset,"MISSING_CHAR","Tags and anchors must be separated from the next token by white space"),f=!1),m&&(c&&"comment"!==a.type&&"newline"!==a.type&&o(m,"TAB_AS_INDENT","Tabs are not allowed as indentation"),m=null),a.type){case"space":t||"doc-start"===n&&"flow-collection"===r?.type||!a.source.includes("\t")||(m=a),u=!0;break;case"comment":{u||o(a,"MISSING_CHAR","Comments must be separated from other tokens by white space characters");const e=a.source.substring(1)||" ";d?d+=p+e:d=e,p="",c=!1;break}case"newline":c?d?d+=a.source:w&&"seq-item-ind"===n||(l=!0):p+=a.source,c=!0,h=!0,(g||v)&&(b=a),u=!0;break;case"anchor":g&&o(a,"MULTIPLE_ANCHORS","A node can have at most one anchor"),a.source.endsWith(":")&&o(a.offset+a.source.length-1,"BAD_ALIAS","Anchor ending in : is ambiguous",!0),g=a,null===O&&(O=a.offset),c=!1,u=!1,f=!0;break;case"tag":v&&o(a,"MULTIPLE_TAGS","A node can have at most one tag"),v=a,null===O&&(O=a.offset),c=!1,u=!1,f=!0;break;case n:(g||v)&&o(a,"BAD_PROP_ORDER",`Anchors and tags must be after the ${a.source} indicator`),w&&o(a,"UNEXPECTED_TOKEN",`Unexpected ${a.source} in ${t??"collection"}`),w=a,c="seq-item-ind"===n||"explicit-key-ind"===n,u=!1;break;case"comma":if(t){y&&o(a,"UNEXPECTED_TOKEN",`Unexpected , in ${t}`),y=a,c=!1,u=!1;break}default:o(a,"UNEXPECTED_TOKEN",`Unexpected ${a.type} token`),c=!1,u=!1}const k=e[e.length-1],x=k?k.offset+k.source.length:a;return f&&r&&"space"!==r.type&&"newline"!==r.type&&"comma"!==r.type&&("scalar"!==r.type||""!==r.source)&&o(r.offset,"MISSING_CHAR","Tags and anchors must be separated from the next token by white space"),m&&(c&&m.indent<=i||"block-map"===r?.type||"block-seq"===r?.type)&&o(m,"TAB_AS_INDENT","Tabs are not allowed as indentation"),{comma:y,found:w,spaceBefore:l,comment:d,hasNewline:h,anchor:g,tag:v,newlineAfterProp:b,end:x,start:O??x}}function At(e){if(!e)return null;switch(e.type){case"alias":case"scalar":case"double-quoted-scalar":case"single-quoted-scalar":if(e.source.includes("\n"))return!0;if(e.end)for(const t of e.end)if("newline"===t.type)return!0;return!1;case"flow-collection":for(const t of e.items){for(const e of t.start)if("newline"===e.type)return!0;if(t.sep)for(const e of t.sep)if("newline"===e.type)return!0;if(At(t.key)||At(t.value))return!0}return!1;default:return!0}}function Rt(e,t,n){if("flow-collection"===t?.type){const r=t.end[0];r.indent!==e||"]"!==r.source&&"}"!==r.source||!At(t)||n(r,"BAD_INDENT","Flow end indicator should be more indented than parent",!0)}}function Pt(e,t,n){const{uniqueKeys:r}=e.options;if(!1===r)return!1;const a="function"==typeof r?r:(e,t)=>e===t||f(e)&&f(t)&&e.value===t.value;return t.some((e=>a(e.key,n)))}const Dt="All mapping items must start at the same column";function Mt(e,t,n,r){let a="";if(e){let o=!1,i="";for(const s of e){const{source:e,type:l}=s;switch(l){case"space":o=!0;break;case"comment":{n&&!o&&r(s,"MISSING_CHAR","Comments must be separated from other tokens by white space characters");const t=e.substring(1)||" ";a?a+=i+t:a=t,i="";break}case"newline":a&&(i+=e),o=!0;break;default:r(s,"UNEXPECTED_TOKEN",`Unexpected ${l} at node end`)}t+=e.length}}return{comment:a,offset:t}}const $t="Block collections are not allowed within flow collections",Nt=e=>e&&("block-map"===e.type||"block-seq"===e.type);function It(e,t,n,r,a,o){const i="block-map"===n.type?function({composeNode:e,composeEmptyNode:t},n,r,a,o){const i=new(o?.nodeClass??Oe)(n.schema);n.atRoot&&(n.atRoot=!1);let s=r.offset,l=null;for(const o of r.items){const{start:c,key:u,sep:d,value:p}=o,h=Tt(c,{indicator:"explicit-key-ind",next:u??d?.[0],offset:s,onError:a,parentIndent:r.indent,startOnNewline:!0}),f=!h.found;if(f){if(u&&("block-seq"===u.type?a(s,"BLOCK_AS_IMPLICIT_KEY","A block sequence may not be used as an implicit map key"):"indent"in u&&u.indent!==r.indent&&a(s,"BAD_INDENT",Dt)),!h.anchor&&!h.tag&&!d){l=h.end,h.comment&&(i.comment?i.comment+="\n"+h.comment:i.comment=h.comment);continue}(h.newlineAfterProp||At(u))&&a(u??c[c.length-1],"MULTILINE_IMPLICIT_KEY","Implicit keys need to be on a single line")}else h.found?.indent!==r.indent&&a(s,"BAD_INDENT",Dt);n.atKey=!0;const m=h.end,g=u?e(n,u,h,a):t(n,m,c,null,h,a);n.schema.compat&&Rt(r.indent,u,a),n.atKey=!1,Pt(n,i.items,g)&&a(m,"DUPLICATE_KEY","Map keys must be unique");const v=Tt(d??[],{indicator:"map-value-ind",next:p,offset:g.range[2],onError:a,parentIndent:r.indent,startOnNewline:!u||"block-scalar"===u.type});if(s=v.end,v.found){f&&("block-map"!==p?.type||v.hasNewline||a(s,"BLOCK_AS_IMPLICIT_KEY","Nested mappings are not allowed in compact mappings"),n.options.strict&&h.start<v.found.offset-1024&&a(g.range,"KEY_OVER_1024_CHARS","The : indicator must be at most 1024 chars after the start of an implicit block mapping key"));const l=p?e(n,p,v,a):t(n,s,d,null,v,a);n.schema.compat&&Rt(r.indent,p,a),s=l.range[2];const c=new me(g,l);n.options.keepSourceTokens&&(c.srcToken=o),i.items.push(c)}else{f&&a(g.range,"MISSING_CHAR","Implicit map keys need to be followed by map values"),v.comment&&(g.comment?g.comment+="\n"+v.comment:g.comment=v.comment);const e=new me(g);n.options.keepSourceTokens&&(e.srcToken=o),i.items.push(e)}}return l&&l<s&&a(l,"IMPOSSIBLE","Map comment with trailing content"),i.range=[r.offset,s,l??s],i}(e,t,n,r,o):"block-seq"===n.type?function({composeNode:e,composeEmptyNode:t},n,r,a,o){const i=new(o?.nodeClass??xe)(n.schema);n.atRoot&&(n.atRoot=!1),n.atKey&&(n.atKey=!1);let s=r.offset,l=null;for(const{start:o,value:c}of r.items){const u=Tt(o,{indicator:"seq-item-ind",next:c,offset:s,onError:a,parentIndent:r.indent,startOnNewline:!0});if(!u.found){if(!(u.anchor||u.tag||c)){l=u.end,u.comment&&(i.comment=u.comment);continue}c&&"block-seq"===c.type?a(u.end,"BAD_INDENT","All sequence items must start at the same column"):a(s,"MISSING_CHAR","Sequence item without - indicator")}const d=c?e(n,c,u,a):t(n,u.end,o,null,u,a);n.schema.compat&&Rt(r.indent,c,a),s=d.range[2],i.items.push(d)}return i.range=[r.offset,s,l??s],i}(e,t,n,r,o):function({composeNode:e,composeEmptyNode:t},n,r,a,o){const i="{"===r.start.source,s=i?"flow map":"flow sequence",l=new(o?.nodeClass??(i?Oe:xe))(n.schema);l.flow=!0;const c=n.atRoot;c&&(n.atRoot=!1),n.atKey&&(n.atKey=!1);let u=r.offset+r.start.source.length;for(let o=0;o<r.items.length;++o){const c=r.items[o],{start:d,key:p,sep:f,value:m}=c,g=Tt(d,{flow:s,indicator:"explicit-key-ind",next:p??f?.[0],offset:u,onError:a,parentIndent:r.indent,startOnNewline:!1});if(!g.found){if(!(g.anchor||g.tag||f||m)){0===o&&g.comma?a(g.comma,"UNEXPECTED_TOKEN",`Unexpected , in ${s}`):o<r.items.length-1&&a(g.start,"UNEXPECTED_TOKEN",`Unexpected empty item in ${s}`),g.comment&&(l.comment?l.comment+="\n"+g.comment:l.comment=g.comment),u=g.end;continue}!i&&n.options.strict&&At(p)&&a(p,"MULTILINE_IMPLICIT_KEY","Implicit keys of flow sequence pairs need to be on a single line")}if(0===o)g.comma&&a(g.comma,"UNEXPECTED_TOKEN",`Unexpected , in ${s}`);else if(g.comma||a(g.start,"MISSING_CHAR",`Missing , between ${s} items`),g.comment){let e="";e:for(const t of d)switch(t.type){case"comma":case"space":break;case"comment":e=t.source.substring(1);break e;default:break e}if(e){let t=l.items[l.items.length-1];h(t)&&(t=t.value??t.key),t.comment?t.comment+="\n"+e:t.comment=e,g.comment=g.comment.substring(e.length+1)}}if(i||f||g.found){n.atKey=!0;const o=g.end,h=p?e(n,p,g,a):t(n,o,d,null,g,a);Nt(p)&&a(h.range,"BLOCK_IN_FLOW",$t),n.atKey=!1;const v=Tt(f??[],{flow:s,indicator:"map-value-ind",next:m,offset:h.range[2],onError:a,parentIndent:r.indent,startOnNewline:!1});if(v.found){if(!i&&!g.found&&n.options.strict){if(f)for(const e of f){if(e===v.found)break;if("newline"===e.type){a(e,"MULTILINE_IMPLICIT_KEY","Implicit keys of flow sequence pairs need to be on a single line");break}}g.start<v.found.offset-1024&&a(v.found,"KEY_OVER_1024_CHARS","The : indicator must be at most 1024 chars after the start of an implicit flow sequence key")}}else m&&("source"in m&&m.source&&":"===m.source[0]?a(m,"MISSING_CHAR",`Missing space after : in ${s}`):a(v.start,"MISSING_CHAR",`Missing , or : between ${s} items`));const b=m?e(n,m,v,a):v.found?t(n,v.end,f,null,v,a):null;b?Nt(m)&&a(b.range,"BLOCK_IN_FLOW",$t):v.comment&&(h.comment?h.comment+="\n"+v.comment:h.comment=v.comment);const y=new me(h,b);if(n.options.keepSourceTokens&&(y.srcToken=c),i){const e=l;Pt(n,e.items,h)&&a(o,"DUPLICATE_KEY","Map keys must be unique"),e.items.push(y)}else{const e=new Oe(n.schema);e.flow=!0,e.items.push(y);const t=(b??h).range;e.range=[h.range[0],t[1],t[2]],l.items.push(e)}u=b?b.range[2]:v.end}else{const r=m?e(n,m,g,a):t(n,g.end,f,null,g,a);l.items.push(r),u=r.range[2],Nt(m)&&a(r.range,"BLOCK_IN_FLOW",$t)}}const d=i?"}":"]",[p,...f]=r.end;let m=u;if(p&&p.source===d)m=p.offset+p.source.length;else{const e=s[0].toUpperCase()+s.substring(1);a(u,c?"MISSING_CHAR":"BAD_INDENT",c?`${e} must end with a ${d}`:`${e} in block collection must be sufficiently indented and end with a ${d}`),p&&1!==p.source.length&&f.unshift(p)}if(f.length>0){const e=Mt(f,m,n.options.strict,a);e.comment&&(l.comment?l.comment+="\n"+e.comment:l.comment=e.comment),l.range=[r.offset,m,e.offset]}else l.range=[r.offset,m,m];return l}(e,t,n,r,o),s=i.constructor;return"!"===a||a===s.tagName?(i.tag=s.tagName,i):(a&&(i.tag=a),i)}function Lt(e){let t,n;try{t=new RegExp("(.*?)(?<![ \t])[ \t]*\r?\n","sy"),n=new RegExp("[ \t]*(.*?)(?:(?<![ \t])[ \t]*)?\r?\n","sy")}catch{t=/(.*?)[ \t]*\r?\n/sy,n=/[ \t]*(.*?)[ \t]*\r?\n/sy}let r=t.exec(e);if(!r)return e;let a=r[1],o=" ",i=t.lastIndex;for(n.lastIndex=i;r=n.exec(e);)""===r[1]?"\n"===o?a+=o:o="\n":(a+=o+r[1],o=" "),i=n.lastIndex;const s=/[ \t]*(.*)/sy;return s.lastIndex=i,r=s.exec(e),a+o+(r?.[1]??"")}function Bt(e,t){let n="",r=e[t+1];for(;!(" "!==r&&"\t"!==r&&"\n"!==r&&"\r"!==r||"\r"===r&&"\n"!==e[t+2]);)"\n"===r&&(n+="\n"),r=e[(t+=1)+1];return n||(n=" "),{fold:n,offset:t}}const Ft={0:"\0",a:"",b:"\b",e:"",f:"\f",n:"\n",r:"\r",t:"\t",v:"\v",N:"",_:" ",L:"\u2028",P:"\u2029"," ":" ",'"':'"',"/":"/","\\":"\\","\t":"\t"};function Qt(e,t,n,r){const a=e.substr(t,n),o=a.length===n&&/^[0-9a-fA-F]+$/.test(a)?parseInt(a,16):NaN;if(isNaN(o)){const a=e.substr(t-2,n+2);return r(t-2,"BAD_DQ_ESCAPE",`Invalid escape sequence ${a}`),a}return String.fromCodePoint(o)}function Ut(e,t,n,r){const{value:a,type:o,comment:i,range:l}="block-scalar"===t.type?function(e,t,n){const r=t.offset,a=function({offset:e,props:t},n,r){if("block-scalar-header"!==t[0].type)return r(t[0],"IMPOSSIBLE","Block scalar header not found"),null;const{source:a}=t[0],o=a[0];let i=0,s="",l=-1;for(let t=1;t<a.length;++t){const n=a[t];if(s||"-"!==n&&"+"!==n){const r=Number(n);!i&&r?i=r:-1===l&&(l=e+t)}else s=n}-1!==l&&r(l,"UNEXPECTED_TOKEN",`Block scalar header includes extra characters: ${a}`);let c=!1,u="",d=a.length;for(let e=1;e<t.length;++e){const a=t[e];switch(a.type){case"space":c=!0;case"newline":d+=a.source.length;break;case"comment":n&&!c&&r(a,"MISSING_CHAR","Comments must be separated from other tokens by white space characters"),d+=a.source.length,u=a.source.substring(1);break;case"error":r(a,"UNEXPECTED_TOKEN",a.message),d+=a.source.length;break;default:{r(a,"UNEXPECTED_TOKEN",`Unexpected token in block scalar header: ${a.type}`);const e=a.source;e&&"string"==typeof e&&(d+=e.length)}}}return{mode:o,indent:i,chomp:s,comment:u,length:d}}(t,e.options.strict,n);if(!a)return{value:"",type:null,comment:"",range:[r,r,r]};const o=">"===a.mode?F.BLOCK_FOLDED:F.BLOCK_LITERAL,i=t.source?function(e){const t=e.split(/\n( *)/),n=t[0],r=n.match(/^( *)/),a=[r?.[1]?[r[1],n.slice(r[1].length)]:["",n]];for(let e=1;e<t.length;e+=2)a.push([t[e],t[e+1]]);return a}(t.source):[];let s=i.length;for(let e=i.length-1;e>=0;--e){const t=i[e][1];if(""!==t&&"\r"!==t)break;s=e}if(0===s){const e="+"===a.chomp&&i.length>0?"\n".repeat(Math.max(1,i.length-1)):"";let n=r+a.length;return t.source&&(n+=t.source.length),{value:e,type:o,comment:a.comment,range:[r,n,n]}}let l=t.indent+a.indent,c=t.offset+a.length,u=0;for(let t=0;t<s;++t){const[r,o]=i[t];if(""!==o&&"\r"!==o){if(r.length<l){const e="Block scalars with more-indented leading empty lines must use an explicit indentation indicator";n(c+r.length,"MISSING_CHAR",e)}0===a.indent&&(l=r.length),u=t,0!==l||e.atRoot||n(c,"BAD_INDENT","Block scalar values in collections must be indented");break}0===a.indent&&r.length>l&&(l=r.length),c+=r.length+o.length+1}for(let e=i.length-1;e>=s;--e)i[e][0].length>l&&(s=e+1);let d="",p="",h=!1;for(let e=0;e<u;++e)d+=i[e][0].slice(l)+"\n";for(let e=u;e<s;++e){let[t,r]=i[e];c+=t.length+r.length+1;const s="\r"===r[r.length-1];if(s&&(r=r.slice(0,-1)),r&&t.length<l){const e="Block scalar lines must not be less indented than their "+(a.indent?"explicit indentation indicator":"first line");n(c-r.length-(s?2:1),"BAD_INDENT",e),t=""}o===F.BLOCK_LITERAL?(d+=p+t.slice(l)+r,p="\n"):t.length>l||"\t"===r[0]?(" "===p?p="\n":h||"\n"!==p||(p="\n\n"),d+=p+t.slice(l)+r,p="\n",h=!0):""===r?"\n"===p?d+="\n":p="\n":(d+=p+r,p=" ",h=!1)}switch(a.chomp){case"-":break;case"+":for(let e=s;e<i.length;++e)d+="\n"+i[e][0].slice(l);"\n"!==d[d.length-1]&&(d+="\n");break;default:d+="\n"}const f=r+a.length+t.source.length;return{value:d,type:o,comment:a.comment,range:[r,f,f]}}(e,t,r):function(e,t,n){const{offset:r,type:a,source:o,end:i}=e;let s,l;const c=(e,t,a)=>n(r+e,t,a);switch(a){case"scalar":s=F.PLAIN,l=function(e,t){let n="";switch(e[0]){case"\t":n="a tab character";break;case",":n="flow indicator character ,";break;case"%":n="directive indicator character %";break;case"|":case">":n=`block scalar indicator ${e[0]}`;break;case"@":case"`":n=`reserved character ${e[0]}`}return n&&t(0,"BAD_SCALAR_START",`Plain value cannot start with ${n}`),Lt(e)}(o,c);break;case"single-quoted-scalar":s=F.QUOTE_SINGLE,l=function(e,t){return"'"===e[e.length-1]&&1!==e.length||t(e.length,"MISSING_CHAR","Missing closing 'quote"),Lt(e.slice(1,-1)).replace(/''/g,"'")}(o,c);break;case"double-quoted-scalar":s=F.QUOTE_DOUBLE,l=function(e,t){let n="";for(let r=1;r<e.length-1;++r){const a=e[r];if("\r"!==a||"\n"!==e[r+1])if("\n"===a){const{fold:t,offset:a}=Bt(e,r);n+=t,r=a}else if("\\"===a){let a=e[++r];const o=Ft[a];if(o)n+=o;else if("\n"===a)for(a=e[r+1];" "===a||"\t"===a;)a=e[1+ ++r];else if("\r"===a&&"\n"===e[r+1])for(a=e[1+ ++r];" "===a||"\t"===a;)a=e[1+ ++r];else if("x"===a||"u"===a||"U"===a){const o={x:2,u:4,U:8}[a];n+=Qt(e,r+1,o,t),r+=o}else{const a=e.substr(r-1,2);t(r-1,"BAD_DQ_ESCAPE",`Invalid escape sequence ${a}`),n+=a}}else if(" "===a||"\t"===a){const t=r;let o=e[r+1];for(;" "===o||"\t"===o;)o=e[1+ ++r];"\n"===o||"\r"===o&&"\n"===e[r+2]||(n+=r>t?e.slice(t,r+1):a)}else n+=a}return'"'===e[e.length-1]&&1!==e.length||t(e.length,"MISSING_CHAR",'Missing closing "quote'),n}(o,c);break;default:return n(e,"UNEXPECTED_TOKEN",`Expected a flow scalar value, but found: ${a}`),{value:"",type:null,comment:"",range:[r,r+o.length,r+o.length]}}const u=r+o.length,d=Mt(i,u,t,n);return{value:l,type:s,comment:d.comment,range:[r,u,d.offset]}}(t,e.options.strict,r),c=n?e.directives.tagName(n.source,(e=>r(n,"TAG_RESOLVE_FAILED",e))):null;let u,d;u=e.options.stringKeys&&e.atKey?e.schema[s]:c?function(e,t,n,r,a){if("!"===n)return e[s];const o=[];for(const t of e.tags)if(!t.collection&&t.tag===n){if(!t.default||!t.test)return t;o.push(t)}for(const e of o)if(e.test?.test(t))return e;const i=e.knownTags[n];return i&&!i.collection?(e.tags.push(Object.assign({},i,{default:!1,test:void 0})),i):(a(r,"TAG_RESOLVE_FAILED",`Unresolved tag: ${n}`,"tag:yaml.org,2002:str"!==n),e[s])}(e.schema,a,c,n,r):"scalar"===t.type?function({atKey:e,directives:t,schema:n},r,a,o){const i=n.tags.find((t=>(!0===t.default||e&&"key"===t.default)&&t.test?.test(r)))||n[s];if(n.compat){const e=n.compat.find((e=>e.default&&e.test?.test(r)))??n[s];i.tag!==e.tag&&o(a,"TAG_RESOLVE_FAILED",`Value may be parsed as either ${t.tagString(i.tag)} or ${t.tagString(e.tag)}`,!0)}return i}(e,a,t,r):e.schema[s];try{const o=u.resolve(a,(e=>r(n??t,"TAG_RESOLVE_FAILED",e)),e.options);d=f(o)?o:new F(o)}catch(e){const o=e instanceof Error?e.message:String(e);r(n??t,"TAG_RESOLVE_FAILED",o),d=new F(a)}return d.range=l,d.source=a,o&&(d.type=o),c&&(d.tag=c),u.format&&(d.format=u.format),i&&(d.comment=i),d}function zt(e,t,n){if(t){null===n&&(n=t.length);for(let r=n-1;r>=0;--r){let n=t[r];switch(n.type){case"space":case"comment":case"newline":e-=n.source.length;continue}for(n=t[++r];"space"===n?.type;)e+=n.source.length,n=t[++r];break}}return e}const jt={composeNode:Wt,composeEmptyNode:Xt};function Wt(e,t,n,r){const a=e.atKey,{spaceBefore:o,comment:i,anchor:s,tag:l}=n;let c,u=!0;switch(t.type){case"alias":c=function({options:e},{offset:t,source:n,end:r},a){const o=new I(n.substring(1));""===o.source&&a(t,"BAD_ALIAS","Alias cannot be an empty string"),o.source.endsWith(":")&&a(t+n.length-1,"BAD_ALIAS","Alias ending in : is ambiguous",!0);const i=t+n.length,s=Mt(r,i,e.strict,a);return o.range=[t,i,s.offset],s.comment&&(o.comment=s.comment),o}(e,t,r),(s||l)&&r(t,"ALIAS_PROPS","An alias node must not specify any properties");break;case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":case"block-scalar":c=Ut(e,t,l,r),s&&(c.anchor=s.source.substring(1));break;case"block-map":case"block-seq":case"flow-collection":c=function(e,t,n,r,a){const o=r.tag,i=o?t.directives.tagName(o.source,(e=>a(o,"TAG_RESOLVE_FAILED",e))):null;if("block-seq"===n.type){const{anchor:e,newlineAfterProp:t}=r,n=e&&o?e.offset>o.offset?e:o:e??o;n&&(!t||t.offset<n.offset)&&a(n,"MISSING_CHAR","Missing newline after block sequence props")}const s="block-map"===n.type?"map":"block-seq"===n.type?"seq":"{"===n.start.source?"map":"seq";if(!o||!i||"!"===i||i===Oe.tagName&&"map"===s||i===xe.tagName&&"seq"===s)return It(e,t,n,a,i);let l=t.schema.tags.find((e=>e.tag===i&&e.collection===s));if(!l){const r=t.schema.knownTags[i];if(!r||r.collection!==s)return r?.collection?a(o,"BAD_COLLECTION_TYPE",`${r.tag} used for ${s} collection, but expects ${r.collection}`,!0):a(o,"TAG_RESOLVE_FAILED",`Unresolved tag: ${i}`,!0),It(e,t,n,a,i);t.schema.tags.push(Object.assign({},r,{default:!1})),l=r}const c=It(e,t,n,a,i,l),u=l.resolve?.(c,(e=>a(o,"TAG_RESOLVE_FAILED",e)),t.options)??c,d=v(u)?u:new F(u);return d.range=c.range,d.tag=i,l?.format&&(d.format=l.format),d}(jt,e,t,n,r),s&&(c.anchor=s.source.substring(1));break;default:r(t,"UNEXPECTED_TOKEN","error"===t.type?t.message:`Unsupported token (type: ${t.type})`),c=Xt(e,t.offset,void 0,null,n,r),u=!1}return s&&""===c.anchor&&r(s,"BAD_ALIAS","Anchor cannot be an empty string"),a&&e.options.stringKeys&&(!f(c)||"string"!=typeof c.value||c.tag&&"tag:yaml.org,2002:str"!==c.tag)&&r(l??t,"NON_STRING_KEY","With stringKeys, all keys must be strings"),o&&(c.spaceBefore=!0),i&&("scalar"===t.type&&""===t.source?c.comment=i:c.commentBefore=i),e.options.keepSourceTokens&&u&&(c.srcToken=t),c}function Xt(e,t,n,r,{spaceBefore:a,comment:o,anchor:i,tag:s,end:l},c){const u=Ut(e,{type:"scalar",offset:zt(t,n,r),indent:-1,source:""},s,c);return i&&(u.anchor=i.source.substring(1),""===u.anchor&&c(i,"BAD_ALIAS","Anchor cannot be an empty string")),a&&(u.spaceBefore=!0),o&&(u.comment=o,u.range[2]=l),u}function qt(e){if("number"==typeof e)return[e,e+1];if(Array.isArray(e))return 2===e.length?e:[e[0],e[1]];const{offset:t,source:n}=e;return[t,t+("string"==typeof n?n.length:1)]}function Vt(e){let t="",n=!1,r=!1;for(let a=0;a<e.length;++a){const o=e[a];switch(o[0]){case"#":t+=(""===t?"":r?"\n\n":"\n")+(o.substring(1)||" "),n=!0,r=!1;break;case"%":"#"!==e[a+1]?.[0]&&(a+=1),n=!1;break;default:n||(r=!0),n=!1}}return{comment:t,afterEmptyLine:r}}class Ht{constructor(e={}){this.doc=null,this.atDirectives=!1,this.prelude=[],this.errors=[],this.warnings=[],this.onError=(e,t,n,r)=>{const a=qt(e);r?this.warnings.push(new _t(a,t,n)):this.errors.push(new St(a,t,n))},this.directives=new A({version:e.version||"1.2"}),this.options=e}decorate(e,t){const{comment:n,afterEmptyLine:r}=Vt(this.prelude);if(n){const a=e.contents;if(t)e.comment=e.comment?`${e.comment}\n${n}`:n;else if(r||e.directives.docStart||!a)e.commentBefore=n;else if(g(a)&&!a.flow&&a.items.length>0){let e=a.items[0];h(e)&&(e=e.key);const t=e.commentBefore;e.commentBefore=t?`${n}\n${t}`:n}else{const e=a.commentBefore;a.commentBefore=e?`${n}\n${e}`:n}}t?(Array.prototype.push.apply(e.errors,this.errors),Array.prototype.push.apply(e.warnings,this.warnings)):(e.errors=this.errors,e.warnings=this.warnings),this.prelude=[],this.errors=[],this.warnings=[]}streamInfo(){return{comment:Vt(this.prelude).comment,directives:this.directives,errors:this.errors,warnings:this.warnings}}*compose(e,t=!1,n=-1){for(const t of e)yield*this.next(t);yield*this.end(t,n)}*next(e){switch(e.type){case"directive":this.directives.add(e.source,((t,n,r)=>{const a=qt(e);a[0]+=t,this.onError(a,"BAD_DIRECTIVE",n,r)})),this.prelude.push(e.source),this.atDirectives=!0;break;case"document":{const t=function(e,t,{offset:n,start:r,value:a,end:o},i){const s=Object.assign({_directives:t},e),l=new kt(void 0,s),c={atKey:!1,atRoot:!0,directives:l.directives,options:l.options,schema:l.schema},u=Tt(r,{indicator:"doc-start",next:a??o?.[0],offset:n,onError:i,parentIndent:0,startOnNewline:!0});u.found&&(l.directives.docStart=!0,!a||"block-map"!==a.type&&"block-seq"!==a.type||u.hasNewline||i(u.end,"MISSING_CHAR","Block collection cannot start on same line with directives-end marker")),l.contents=a?Wt(c,a,u,i):Xt(c,u.end,r,null,u,i);const d=l.contents.range[2],p=Mt(o,d,!1,i);return p.comment&&(l.comment=p.comment),l.range=[n,d,p.offset],l}(this.options,this.directives,e,this.onError);this.atDirectives&&!t.directives.docStart&&this.onError(e,"MISSING_CHAR","Missing directives-end/doc-start indicator line"),this.decorate(t,!1),this.doc&&(yield this.doc),this.doc=t,this.atDirectives=!1;break}case"byte-order-mark":case"space":break;case"comment":case"newline":this.prelude.push(e.source);break;case"error":{const t=e.source?`${e.message}: ${JSON.stringify(e.source)}`:e.message,n=new St(qt(e),"UNEXPECTED_TOKEN",t);this.atDirectives||!this.doc?this.errors.push(n):this.doc.errors.push(n);break}case"doc-end":{if(!this.doc){const t="Unexpected doc-end without preceding document";this.errors.push(new St(qt(e),"UNEXPECTED_TOKEN",t));break}this.doc.directives.docEnd=!0;const t=Mt(e.end,e.offset+e.source.length,this.doc.options.strict,this.onError);if(this.decorate(this.doc,!0),t.comment){const e=this.doc.comment;this.doc.comment=e?`${e}\n${t.comment}`:t.comment}this.doc.range[2]=t.offset;break}default:this.errors.push(new St(qt(e),"UNEXPECTED_TOKEN",`Unsupported token ${e.type}`))}}*end(e=!1,t=-1){if(this.doc)this.decorate(this.doc,!0),yield this.doc,this.doc=null;else if(e){const e=Object.assign({_directives:this.directives},this.options),n=new kt(void 0,e);this.atDirectives&&this.onError(t,"MISSING_CHAR","Missing directives-end indicator line"),n.range=[0,t,t],this.decorate(n,!1),yield n}}}const Zt=Symbol("break visit"),Gt=Symbol("skip children"),Kt=Symbol("remove item");function Yt(e,t){"type"in e&&"document"===e.type&&(e={start:e.start,value:e.value}),Jt(Object.freeze([]),e,t)}function Jt(e,t,n){let r=n(t,e);if("symbol"==typeof r)return r;for(const a of["key","value"]){const o=t[a];if(o&&"items"in o){for(let t=0;t<o.items.length;++t){const r=Jt(Object.freeze(e.concat([[a,t]])),o.items[t],n);if("number"==typeof r)t=r-1;else{if(r===Zt)return Zt;r===Kt&&(o.items.splice(t,1),t-=1)}}"function"==typeof r&&"key"===a&&(r=r(t,e))}}return"function"==typeof r?r(t,e):r}Yt.BREAK=Zt,Yt.SKIP=Gt,Yt.REMOVE=Kt,Yt.itemAtPath=(e,t)=>{let n=e;for(const[e,r]of t){const t=n?.[e];if(!t||!("items"in t))return;n=t.items[r]}return n},Yt.parentCollection=(e,t)=>{const n=Yt.itemAtPath(e,t.slice(0,-1)),r=t[t.length-1][0],a=n?.[r];if(a&&"items"in a)return a;throw new Error("Parent collection not found")};function en(e){switch(e){case void 0:case" ":case"\n":case"\r":case"\t":return!0;default:return!1}}const tn=new Set("0123456789ABCDEFabcdef"),nn=new Set("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-#;/?:@&=+$_.!~*'()"),rn=new Set(",[]{}"),an=new Set(" ,[]{}\n\r\t"),on=e=>!e||an.has(e);class sn{constructor(){this.atEnd=!1,this.blockScalarIndent=-1,this.blockScalarKeep=!1,this.buffer="",this.flowKey=!1,this.flowLevel=0,this.indentNext=0,this.indentValue=0,this.lineEndPos=null,this.next=null,this.pos=0}*lex(e,t=!1){if(e){if("string"!=typeof e)throw TypeError("source is not a string");this.buffer=this.buffer?this.buffer+e:e,this.lineEndPos=null}this.atEnd=!t;let n=this.next??"stream";for(;n&&(t||this.hasChars(1));)n=yield*this.parseNext(n)}atLineEnd(){let e=this.pos,t=this.buffer[e];for(;" "===t||"\t"===t;)t=this.buffer[++e];return!t||"#"===t||"\n"===t||"\r"===t&&"\n"===this.buffer[e+1]}charAt(e){return this.buffer[this.pos+e]}continueScalar(e){let t=this.buffer[e];if(this.indentNext>0){let n=0;for(;" "===t;)t=this.buffer[++n+e];if("\r"===t){const t=this.buffer[n+e+1];if("\n"===t||!t&&!this.atEnd)return e+n+1}return"\n"===t||n>=this.indentNext||!t&&!this.atEnd?e+n:-1}if("-"===t||"."===t){const t=this.buffer.substr(e,3);if(("---"===t||"..."===t)&&en(this.buffer[e+3]))return-1}return e}getLine(){let e=this.lineEndPos;return("number"!=typeof e||-1!==e&&e<this.pos)&&(e=this.buffer.indexOf("\n",this.pos),this.lineEndPos=e),-1===e?this.atEnd?this.buffer.substring(this.pos):null:("\r"===this.buffer[e-1]&&(e-=1),this.buffer.substring(this.pos,e))}hasChars(e){return this.pos+e<=this.buffer.length}setNext(e){return this.buffer=this.buffer.substring(this.pos),this.pos=0,this.lineEndPos=null,this.next=e,null}peek(e){return this.buffer.substr(this.pos,e)}*parseNext(e){switch(e){case"stream":return yield*this.parseStream();case"line-start":return yield*this.parseLineStart();case"block-start":return yield*this.parseBlockStart();case"doc":return yield*this.parseDocument();case"flow":return yield*this.parseFlowCollection();case"quoted-scalar":return yield*this.parseQuotedScalar();case"block-scalar":return yield*this.parseBlockScalar();case"plain-scalar":return yield*this.parsePlainScalar()}}*parseStream(){let e=this.getLine();if(null===e)return this.setNext("stream");if("\ufeff"===e[0]&&(yield*this.pushCount(1),e=e.substring(1)),"%"===e[0]){let t=e.length,n=e.indexOf("#");for(;-1!==n;){const r=e[n-1];if(" "===r||"\t"===r){t=n-1;break}n=e.indexOf("#",n+1)}for(;;){const n=e[t-1];if(" "!==n&&"\t"!==n)break;t-=1}const r=(yield*this.pushCount(t))+(yield*this.pushSpaces(!0));return yield*this.pushCount(e.length-r),this.pushNewline(),"stream"}if(this.atLineEnd()){const t=yield*this.pushSpaces(!0);return yield*this.pushCount(e.length-t),yield*this.pushNewline(),"stream"}return yield"",yield*this.parseLineStart()}*parseLineStart(){const e=this.charAt(0);if(!e&&!this.atEnd)return this.setNext("line-start");if("-"===e||"."===e){if(!this.atEnd&&!this.hasChars(4))return this.setNext("line-start");const e=this.peek(3);if(("---"===e||"..."===e)&&en(this.charAt(3)))return yield*this.pushCount(3),this.indentValue=0,this.indentNext=0,"---"===e?"doc":"stream"}return this.indentValue=yield*this.pushSpaces(!1),this.indentNext>this.indentValue&&!en(this.charAt(1))&&(this.indentNext=this.indentValue),yield*this.parseBlockStart()}*parseBlockStart(){const[e,t]=this.peek(2);if(!t&&!this.atEnd)return this.setNext("block-start");if(("-"===e||"?"===e||":"===e)&&en(t)){const e=(yield*this.pushCount(1))+(yield*this.pushSpaces(!0));return this.indentNext=this.indentValue+1,this.indentValue+=e,yield*this.parseBlockStart()}return"doc"}*parseDocument(){yield*this.pushSpaces(!0);const e=this.getLine();if(null===e)return this.setNext("doc");let t=yield*this.pushIndicators();switch(e[t]){case"#":yield*this.pushCount(e.length-t);case void 0:return yield*this.pushNewline(),yield*this.parseLineStart();case"{":case"[":return yield*this.pushCount(1),this.flowKey=!1,this.flowLevel=1,"flow";case"}":case"]":return yield*this.pushCount(1),"doc";case"*":return yield*this.pushUntil(on),"doc";case'"':case"'":return yield*this.parseQuotedScalar();case"|":case">":return t+=(yield*this.parseBlockScalarHeader()),t+=(yield*this.pushSpaces(!0)),yield*this.pushCount(e.length-t),yield*this.pushNewline(),yield*this.parseBlockScalar();default:return yield*this.parsePlainScalar()}}*parseFlowCollection(){let e,t,n=-1;do{e=yield*this.pushNewline(),e>0?(t=yield*this.pushSpaces(!1),this.indentValue=n=t):t=0,t+=(yield*this.pushSpaces(!0))}while(e+t>0);const r=this.getLine();if(null===r)return this.setNext("flow");if((-1!==n&&n<this.indentNext&&"#"!==r[0]||0===n&&(r.startsWith("---")||r.startsWith("..."))&&en(r[3]))&&(n!==this.indentNext-1||1!==this.flowLevel||"]"!==r[0]&&"}"!==r[0]))return this.flowLevel=0,yield"",yield*this.parseLineStart();let a=0;for(;","===r[a];)a+=(yield*this.pushCount(1)),a+=(yield*this.pushSpaces(!0)),this.flowKey=!1;switch(a+=(yield*this.pushIndicators()),r[a]){case void 0:return"flow";case"#":return yield*this.pushCount(r.length-a),"flow";case"{":case"[":return yield*this.pushCount(1),this.flowKey=!1,this.flowLevel+=1,"flow";case"}":case"]":return yield*this.pushCount(1),this.flowKey=!0,this.flowLevel-=1,this.flowLevel?"flow":"doc";case"*":return yield*this.pushUntil(on),"flow";case'"':case"'":return this.flowKey=!0,yield*this.parseQuotedScalar();case":":{const e=this.charAt(1);if(this.flowKey||en(e)||","===e)return this.flowKey=!1,yield*this.pushCount(1),yield*this.pushSpaces(!0),"flow"}default:return this.flowKey=!1,yield*this.parsePlainScalar()}}*parseQuotedScalar(){const e=this.charAt(0);let t=this.buffer.indexOf(e,this.pos+1);if("'"===e)for(;-1!==t&&"'"===this.buffer[t+1];)t=this.buffer.indexOf("'",t+2);else for(;-1!==t;){let e=0;for(;"\\"===this.buffer[t-1-e];)e+=1;if(e%2==0)break;t=this.buffer.indexOf('"',t+1)}const n=this.buffer.substring(0,t);let r=n.indexOf("\n",this.pos);if(-1!==r){for(;-1!==r;){const e=this.continueScalar(r+1);if(-1===e)break;r=n.indexOf("\n",e)}-1!==r&&(t=r-("\r"===n[r-1]?2:1))}if(-1===t){if(!this.atEnd)return this.setNext("quoted-scalar");t=this.buffer.length}return yield*this.pushToIndex(t+1,!1),this.flowLevel?"flow":"doc"}*parseBlockScalarHeader(){this.blockScalarIndent=-1,this.blockScalarKeep=!1;let e=this.pos;for(;;){const t=this.buffer[++e];if("+"===t)this.blockScalarKeep=!0;else if(t>"0"&&t<="9")this.blockScalarIndent=Number(t)-1;else if("-"!==t)break}return yield*this.pushUntil((e=>en(e)||"#"===e))}*parseBlockScalar(){let e,t=this.pos-1,n=0;e:for(let r=this.pos;e=this.buffer[r];++r)switch(e){case" ":n+=1;break;case"\n":t=r,n=0;break;case"\r":{const e=this.buffer[r+1];if(!e&&!this.atEnd)return this.setNext("block-scalar");if("\n"===e)break}default:break e}if(!e&&!this.atEnd)return this.setNext("block-scalar");if(n>=this.indentNext){-1===this.blockScalarIndent?this.indentNext=n:this.indentNext=this.blockScalarIndent+(0===this.indentNext?1:this.indentNext);do{const e=this.continueScalar(t+1);if(-1===e)break;t=this.buffer.indexOf("\n",e)}while(-1!==t);if(-1===t){if(!this.atEnd)return this.setNext("block-scalar");t=this.buffer.length}}let r=t+1;for(e=this.buffer[r];" "===e;)e=this.buffer[++r];if("\t"===e){for(;"\t"===e||" "===e||"\r"===e||"\n"===e;)e=this.buffer[++r];t=r-1}else if(!this.blockScalarKeep)for(;;){let e=t-1,r=this.buffer[e];"\r"===r&&(r=this.buffer[--e]);const a=e;for(;" "===r;)r=this.buffer[--e];if(!("\n"===r&&e>=this.pos&&e+1+n>a))break;t=e}return yield"",yield*this.pushToIndex(t+1,!0),yield*this.parseLineStart()}*parsePlainScalar(){const e=this.flowLevel>0;let t,n=this.pos-1,r=this.pos-1;for(;t=this.buffer[++r];)if(":"===t){const t=this.buffer[r+1];if(en(t)||e&&rn.has(t))break;n=r}else if(en(t)){let a=this.buffer[r+1];if("\r"===t&&("\n"===a?(r+=1,t="\n",a=this.buffer[r+1]):n=r),"#"===a||e&&rn.has(a))break;if("\n"===t){const e=this.continueScalar(r+1);if(-1===e)break;r=Math.max(r,e-2)}}else{if(e&&rn.has(t))break;n=r}return t||this.atEnd?(yield"",yield*this.pushToIndex(n+1,!0),e?"flow":"doc"):this.setNext("plain-scalar")}*pushCount(e){return e>0?(yield this.buffer.substr(this.pos,e),this.pos+=e,e):0}*pushToIndex(e,t){const n=this.buffer.slice(this.pos,e);return n?(yield n,this.pos+=n.length,n.length):(t&&(yield""),0)}*pushIndicators(){switch(this.charAt(0)){case"!":return(yield*this.pushTag())+(yield*this.pushSpaces(!0))+(yield*this.pushIndicators());case"&":return(yield*this.pushUntil(on))+(yield*this.pushSpaces(!0))+(yield*this.pushIndicators());case"-":case"?":case":":{const e=this.flowLevel>0,t=this.charAt(1);if(en(t)||e&&rn.has(t))return e?this.flowKey&&(this.flowKey=!1):this.indentNext=this.indentValue+1,(yield*this.pushCount(1))+(yield*this.pushSpaces(!0))+(yield*this.pushIndicators())}}return 0}*pushTag(){if("<"===this.charAt(1)){let e=this.pos+2,t=this.buffer[e];for(;!en(t)&&">"!==t;)t=this.buffer[++e];return yield*this.pushToIndex(">"===t?e+1:e,!1)}{let e=this.pos+1,t=this.buffer[e];for(;t;)if(nn.has(t))t=this.buffer[++e];else{if("%"!==t||!tn.has(this.buffer[e+1])||!tn.has(this.buffer[e+2]))break;t=this.buffer[e+=3]}return yield*this.pushToIndex(e,!1)}}*pushNewline(){const e=this.buffer[this.pos];return"\n"===e?yield*this.pushCount(1):"\r"===e&&"\n"===this.charAt(1)?yield*this.pushCount(2):0}*pushSpaces(e){let t,n=this.pos-1;do{t=this.buffer[++n]}while(" "===t||e&&"\t"===t);const r=n-this.pos;return r>0&&(yield this.buffer.substr(this.pos,r),this.pos=n),r}*pushUntil(e){let t=this.pos,n=this.buffer[t];for(;!e(n);)n=this.buffer[++t];return yield*this.pushToIndex(t,!1)}}function ln(e,t){for(let n=0;n<e.length;++n)if(e[n].type===t)return!0;return!1}function cn(e){for(let t=0;t<e.length;++t)switch(e[t].type){case"space":case"comment":case"newline":break;default:return t}return-1}function un(e){switch(e?.type){case"alias":case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":case"flow-collection":return!0;default:return!1}}function dn(e){switch(e.type){case"document":return e.start;case"block-map":{const t=e.items[e.items.length-1];return t.sep??t.start}case"block-seq":return e.items[e.items.length-1].start;default:return[]}}function pn(e){if(0===e.length)return[];let t=e.length;e:for(;--t>=0;)switch(e[t].type){case"doc-start":case"explicit-key-ind":case"map-value-ind":case"seq-item-ind":case"newline":break e}for(;"space"===e[++t]?.type;);return e.splice(t,e.length)}function hn(e){if("flow-seq-start"===e.start.type)for(const t of e.items)!t.sep||t.value||ln(t.start,"explicit-key-ind")||ln(t.sep,"map-value-ind")||(t.key&&(t.value=t.key),delete t.key,un(t.value)?t.value.end?Array.prototype.push.apply(t.value.end,t.sep):t.value.end=t.sep:Array.prototype.push.apply(t.start,t.sep),delete t.sep)}class fn{constructor(e){this.atNewLine=!0,this.atScalar=!1,this.indent=0,this.offset=0,this.onKeyLine=!1,this.stack=[],this.source="",this.type="",this.lexer=new sn,this.onNewLine=e}*parse(e,t=!1){this.onNewLine&&0===this.offset&&this.onNewLine(0);for(const n of this.lexer.lex(e,t))yield*this.next(n);t||(yield*this.end())}*next(e){if(this.source=e,this.atScalar)return this.atScalar=!1,yield*this.step(),void(this.offset+=e.length);const t=function(e){switch(e){case"\ufeff":return"byte-order-mark";case"":return"doc-mode";case"":return"flow-error-end";case"":return"scalar";case"---":return"doc-start";case"...":return"doc-end";case"":case"\n":case"\r\n":return"newline";case"-":return"seq-item-ind";case"?":return"explicit-key-ind";case":":return"map-value-ind";case"{":return"flow-map-start";case"}":return"flow-map-end";case"[":return"flow-seq-start";case"]":return"flow-seq-end";case",":return"comma"}switch(e[0]){case" ":case"\t":return"space";case"#":return"comment";case"%":return"directive-line";case"*":return"alias";case"&":return"anchor";case"!":return"tag";case"'":return"single-quoted-scalar";case'"':return"double-quoted-scalar";case"|":case">":return"block-scalar-header"}return null}(e);if(t)if("scalar"===t)this.atNewLine=!1,this.atScalar=!0,this.type="scalar";else{switch(this.type=t,yield*this.step(),t){case"newline":this.atNewLine=!0,this.indent=0,this.onNewLine&&this.onNewLine(this.offset+e.length);break;case"space":this.atNewLine&&" "===e[0]&&(this.indent+=e.length);break;case"explicit-key-ind":case"map-value-ind":case"seq-item-ind":this.atNewLine&&(this.indent+=e.length);break;case"doc-mode":case"flow-error-end":return;default:this.atNewLine=!1}this.offset+=e.length}else{const t=`Not a YAML token: ${e}`;yield*this.pop({type:"error",offset:this.offset,message:t,source:e}),this.offset+=e.length}}*end(){for(;this.stack.length>0;)yield*this.pop()}get sourceToken(){return{type:this.type,offset:this.offset,indent:this.indent,source:this.source}}*step(){const e=this.peek(1);if("doc-end"!==this.type||e&&"doc-end"===e.type){if(!e)return yield*this.stream();switch(e.type){case"document":return yield*this.document(e);case"alias":case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":return yield*this.scalar(e);case"block-scalar":return yield*this.blockScalar(e);case"block-map":return yield*this.blockMap(e);case"block-seq":return yield*this.blockSequence(e);case"flow-collection":return yield*this.flowCollection(e);case"doc-end":return yield*this.documentEnd(e)}yield*this.pop()}else{for(;this.stack.length>0;)yield*this.pop();this.stack.push({type:"doc-end",offset:this.offset,source:this.source})}}peek(e){return this.stack[this.stack.length-e]}*pop(e){const t=e??this.stack.pop();if(t)if(0===this.stack.length)yield t;else{const e=this.peek(1);switch("block-scalar"===t.type?t.indent="indent"in e?e.indent:0:"flow-collection"===t.type&&"document"===e.type&&(t.indent=0),"flow-collection"===t.type&&hn(t),e.type){case"document":e.value=t;break;case"block-scalar":e.props.push(t);break;case"block-map":{const n=e.items[e.items.length-1];if(n.value)return e.items.push({start:[],key:t,sep:[]}),void(this.onKeyLine=!0);if(!n.sep)return Object.assign(n,{key:t,sep:[]}),void(this.onKeyLine=!n.explicitKey);n.value=t;break}case"block-seq":{const n=e.items[e.items.length-1];n.value?e.items.push({start:[],value:t}):n.value=t;break}case"flow-collection":{const n=e.items[e.items.length-1];return void(!n||n.value?e.items.push({start:[],key:t,sep:[]}):n.sep?n.value=t:Object.assign(n,{key:t,sep:[]}))}default:yield*this.pop(),yield*this.pop(t)}if(!("document"!==e.type&&"block-map"!==e.type&&"block-seq"!==e.type||"block-map"!==t.type&&"block-seq"!==t.type)){const n=t.items[t.items.length-1];n&&!n.sep&&!n.value&&n.start.length>0&&-1===cn(n.start)&&(0===t.indent||n.start.every((e=>"comment"!==e.type||e.indent<t.indent)))&&("document"===e.type?e.end=n.start:e.items.push({start:n.start}),t.items.splice(-1,1))}}else{const e="Tried to pop an empty stack";yield{type:"error",offset:this.offset,source:"",message:e}}}*stream(){switch(this.type){case"directive-line":return void(yield{type:"directive",offset:this.offset,source:this.source});case"byte-order-mark":case"space":case"comment":case"newline":return void(yield this.sourceToken);case"doc-mode":case"doc-start":{const e={type:"document",offset:this.offset,start:[]};return"doc-start"===this.type&&e.start.push(this.sourceToken),void this.stack.push(e)}}yield{type:"error",offset:this.offset,message:`Unexpected ${this.type} token in YAML stream`,source:this.source}}*document(e){if(e.value)return yield*this.lineEnd(e);switch(this.type){case"doc-start":return void(-1!==cn(e.start)?(yield*this.pop(),yield*this.step()):e.start.push(this.sourceToken));case"anchor":case"tag":case"space":case"comment":case"newline":return void e.start.push(this.sourceToken)}const t=this.startBlockValue(e);t?this.stack.push(t):yield{type:"error",offset:this.offset,message:`Unexpected ${this.type} token in YAML document`,source:this.source}}*scalar(e){if("map-value-ind"===this.type){const t=pn(dn(this.peek(2)));let n;e.end?(n=e.end,n.push(this.sourceToken),delete e.end):n=[this.sourceToken];const r={type:"block-map",offset:e.offset,indent:e.indent,items:[{start:t,key:e,sep:n}]};this.onKeyLine=!0,this.stack[this.stack.length-1]=r}else yield*this.lineEnd(e)}*blockScalar(e){switch(this.type){case"space":case"comment":case"newline":return void e.props.push(this.sourceToken);case"scalar":if(e.source=this.source,this.atNewLine=!0,this.indent=0,this.onNewLine){let e=this.source.indexOf("\n")+1;for(;0!==e;)this.onNewLine(this.offset+e),e=this.source.indexOf("\n",e)+1}yield*this.pop();break;default:yield*this.pop(),yield*this.step()}}*blockMap(e){const t=e.items[e.items.length-1];switch(this.type){case"newline":if(this.onKeyLine=!1,t.value){const n="end"in t.value?t.value.end:void 0,r=Array.isArray(n)?n[n.length-1]:void 0;"comment"===r?.type?n?.push(this.sourceToken):e.items.push({start:[this.sourceToken]})}else t.sep?t.sep.push(this.sourceToken):t.start.push(this.sourceToken);return;case"space":case"comment":if(t.value)e.items.push({start:[this.sourceToken]});else if(t.sep)t.sep.push(this.sourceToken);else{if(this.atIndentedComment(t.start,e.indent)){const n=e.items[e.items.length-2],r=n?.value?.end;if(Array.isArray(r))return Array.prototype.push.apply(r,t.start),r.push(this.sourceToken),void e.items.pop()}t.start.push(this.sourceToken)}return}if(this.indent>=e.indent){const n=!this.onKeyLine&&this.indent===e.indent,r=n&&(t.sep||t.explicitKey)&&"seq-item-ind"!==this.type;let a=[];if(r&&t.sep&&!t.value){const n=[];for(let r=0;r<t.sep.length;++r){const a=t.sep[r];switch(a.type){case"newline":n.push(r);break;case"space":break;case"comment":a.indent>e.indent&&(n.length=0);break;default:n.length=0}}n.length>=2&&(a=t.sep.splice(n[1]))}switch(this.type){case"anchor":case"tag":return void(r||t.value?(a.push(this.sourceToken),e.items.push({start:a}),this.onKeyLine=!0):t.sep?t.sep.push(this.sourceToken):t.start.push(this.sourceToken));case"explicit-key-ind":return t.sep||t.explicitKey?r||t.value?(a.push(this.sourceToken),e.items.push({start:a,explicitKey:!0})):this.stack.push({type:"block-map",offset:this.offset,indent:this.indent,items:[{start:[this.sourceToken],explicitKey:!0}]}):(t.start.push(this.sourceToken),t.explicitKey=!0),void(this.onKeyLine=!0);case"map-value-ind":if(t.explicitKey)if(t.sep)if(t.value)e.items.push({start:[],key:null,sep:[this.sourceToken]});else if(ln(t.sep,"map-value-ind"))this.stack.push({type:"block-map",offset:this.offset,indent:this.indent,items:[{start:a,key:null,sep:[this.sourceToken]}]});else if(un(t.key)&&!ln(t.sep,"newline")){const e=pn(t.start),n=t.key,r=t.sep;r.push(this.sourceToken),delete t.key,delete t.sep,this.stack.push({type:"block-map",offset:this.offset,indent:this.indent,items:[{start:e,key:n,sep:r}]})}else a.length>0?t.sep=t.sep.concat(a,this.sourceToken):t.sep.push(this.sourceToken);else if(ln(t.start,"newline"))Object.assign(t,{key:null,sep:[this.sourceToken]});else{const e=pn(t.start);this.stack.push({type:"block-map",offset:this.offset,indent:this.indent,items:[{start:e,key:null,sep:[this.sourceToken]}]})}else t.sep?t.value||r?e.items.push({start:a,key:null,sep:[this.sourceToken]}):ln(t.sep,"map-value-ind")?this.stack.push({type:"block-map",offset:this.offset,indent:this.indent,items:[{start:[],key:null,sep:[this.sourceToken]}]}):t.sep.push(this.sourceToken):Object.assign(t,{key:null,sep:[this.sourceToken]});return void(this.onKeyLine=!0);case"alias":case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":{const n=this.flowScalar(this.type);return void(r||t.value?(e.items.push({start:a,key:n,sep:[]}),this.onKeyLine=!0):t.sep?this.stack.push(n):(Object.assign(t,{key:n,sep:[]}),this.onKeyLine=!0))}default:{const t=this.startBlockValue(e);if(t)return n&&"block-seq"!==t.type&&e.items.push({start:a}),void this.stack.push(t)}}}yield*this.pop(),yield*this.step()}*blockSequence(e){const t=e.items[e.items.length-1];switch(this.type){case"newline":if(t.value){const n="end"in t.value?t.value.end:void 0,r=Array.isArray(n)?n[n.length-1]:void 0;"comment"===r?.type?n?.push(this.sourceToken):e.items.push({start:[this.sourceToken]})}else t.start.push(this.sourceToken);return;case"space":case"comment":if(t.value)e.items.push({start:[this.sourceToken]});else{if(this.atIndentedComment(t.start,e.indent)){const n=e.items[e.items.length-2],r=n?.value?.end;if(Array.isArray(r))return Array.prototype.push.apply(r,t.start),r.push(this.sourceToken),void e.items.pop()}t.start.push(this.sourceToken)}return;case"anchor":case"tag":if(t.value||this.indent<=e.indent)break;return void t.start.push(this.sourceToken);case"seq-item-ind":if(this.indent!==e.indent)break;return void(t.value||ln(t.start,"seq-item-ind")?e.items.push({start:[this.sourceToken]}):t.start.push(this.sourceToken))}if(this.indent>e.indent){const t=this.startBlockValue(e);if(t)return void this.stack.push(t)}yield*this.pop(),yield*this.step()}*flowCollection(e){const t=e.items[e.items.length-1];if("flow-error-end"===this.type){let e;do{yield*this.pop(),e=this.peek(1)}while(e&&"flow-collection"===e.type)}else if(0===e.end.length){switch(this.type){case"comma":case"explicit-key-ind":return void(!t||t.sep?e.items.push({start:[this.sourceToken]}):t.start.push(this.sourceToken));case"map-value-ind":return void(!t||t.value?e.items.push({start:[],key:null,sep:[this.sourceToken]}):t.sep?t.sep.push(this.sourceToken):Object.assign(t,{key:null,sep:[this.sourceToken]}));case"space":case"comment":case"newline":case"anchor":case"tag":return void(!t||t.value?e.items.push({start:[this.sourceToken]}):t.sep?t.sep.push(this.sourceToken):t.start.push(this.sourceToken));case"alias":case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":{const n=this.flowScalar(this.type);return void(!t||t.value?e.items.push({start:[],key:n,sep:[]}):t.sep?this.stack.push(n):Object.assign(t,{key:n,sep:[]}))}case"flow-map-end":case"flow-seq-end":return void e.end.push(this.sourceToken)}const n=this.startBlockValue(e);n?this.stack.push(n):(yield*this.pop(),yield*this.step())}else{const t=this.peek(2);if("block-map"===t.type&&("map-value-ind"===this.type&&t.indent===e.indent||"newline"===this.type&&!t.items[t.items.length-1].sep))yield*this.pop(),yield*this.step();else if("map-value-ind"===this.type&&"flow-collection"!==t.type){const n=pn(dn(t));hn(e);const r=e.end.splice(1,e.end.length);r.push(this.sourceToken);const a={type:"block-map",offset:e.offset,indent:e.indent,items:[{start:n,key:e,sep:r}]};this.onKeyLine=!0,this.stack[this.stack.length-1]=a}else yield*this.lineEnd(e)}}flowScalar(e){if(this.onNewLine){let e=this.source.indexOf("\n")+1;for(;0!==e;)this.onNewLine(this.offset+e),e=this.source.indexOf("\n",e)+1}return{type:e,offset:this.offset,indent:this.indent,source:this.source}}startBlockValue(e){switch(this.type){case"alias":case"scalar":case"single-quoted-scalar":case"double-quoted-scalar":return this.flowScalar(this.type);case"block-scalar-header":return{type:"block-scalar",offset:this.offset,indent:this.indent,props:[this.sourceToken],source:""};case"flow-map-start":case"flow-seq-start":return{type:"flow-collection",offset:this.offset,indent:this.indent,start:this.sourceToken,items:[],end:[]};case"seq-item-ind":return{type:"block-seq",offset:this.offset,indent:this.indent,items:[{start:[this.sourceToken]}]};case"explicit-key-ind":{this.onKeyLine=!0;const t=pn(dn(e));return t.push(this.sourceToken),{type:"block-map",offset:this.offset,indent:this.indent,items:[{start:t,explicitKey:!0}]}}case"map-value-ind":{this.onKeyLine=!0;const t=pn(dn(e));return{type:"block-map",offset:this.offset,indent:this.indent,items:[{start:t,key:null,sep:[this.sourceToken]}]}}}return null}atIndentedComment(e,t){return"comment"===this.type&&!(this.indent<=t)&&e.every((e=>"newline"===e.type||"space"===e.type))}*documentEnd(e){"doc-mode"!==this.type&&(e.end?e.end.push(this.sourceToken):e.end=[this.sourceToken],"newline"===this.type&&(yield*this.pop()))}*lineEnd(e){switch(this.type){case"comma":case"doc-start":case"doc-end":case"flow-seq-end":case"flow-map-end":case"map-value-ind":yield*this.pop(),yield*this.step();break;case"newline":this.onKeyLine=!1;default:e.end?e.end.push(this.sourceToken):e.end=[this.sourceToken],"newline"===this.type&&(yield*this.pop())}}}class mn{constructor(){this.lineStarts=[],this.addNewLine=e=>this.lineStarts.push(e),this.linePos=e=>{let t=0,n=this.lineStarts.length;for(;t<n;){const r=t+n>>1;this.lineStarts[r]<e?t=r+1:n=r}return this.lineStarts[t]===e?{line:t+1,col:1}:0===t?{line:0,col:e}:{line:t,col:e-this.lineStarts[t-1]+1}}}}function gn(e,t,n){let r;"function"==typeof t?r=t:void 0===n&&t&&"object"==typeof t&&(n=t);const a=function(e,t={}){const{lineCounter:n,prettyErrors:r}=function(e){const t=!1!==e.prettyErrors;return{lineCounter:e.lineCounter||t&&new mn||null,prettyErrors:t}}(t),a=new fn(n?.addNewLine),o=new Ht(t);let i=null;for(const t of o.compose(a.parse(e),!0,e.length))if(i){if("silent"!==i.options.logLevel){i.errors.push(new St(t.range.slice(0,2),"MULTIPLE_DOCS","Source contains multiple documents; please use YAML.parseAllDocuments()"));break}}else i=t;return r&&n&&(i.errors.forEach(Ct(e,n)),i.warnings.forEach(Ct(e,n))),i}(e,n);if(!a)return null;if(a.warnings.forEach((e=>le(a.options.logLevel,e))),a.errors.length>0){if("silent"!==a.options.logLevel)throw a.errors[0];a.errors=[]}return a.toJS(Object.assign({reviver:r},n))}function vn(e,t,n){let r=null;if("function"==typeof t||Array.isArray(t)?r=t:void 0===n&&t&&(n=t),"string"==typeof n&&(n=n.length),"number"==typeof n){const e=Math.round(n);n=e<1?void 0:e>8?{indent:8}:{indent:e}}if(void 0===e){const{keepUndefined:e}=n??t??{};if(!e)return}return d(e)&&!r?e.toString(n):new kt(e,r,n).toString(n)}},3213:(e,t,n)=>{"use strict";n.d(t,{I:()=>a});var r=n(6941);function a(e,t,n=!0){return[...e.pathParameters||[],...e.information?.parameters||[]].filter((e=>e.in===t)).filter((e=>n&&e.required||!n)).map((e=>({name:e.name,description:e.description??null,value:e.example?e.example:e.schema?(0,r.M)(e.schema,{mode:"write"}):"",required:e.required??!1,enabled:e.required??!1}))).sort(((e,t)=>e.required&&!t.required?-1:!e.required&&t.required?1:0))}},3395:(e,t,n)=>{"use strict";n.d(t,{A:()=>l});var r=n(641),a=n(3526);const o={},i={class:"flex xl:h-full xl:min-w-0 flex-1 flex-col xl:custom-scroll bg-b-1"},s={class:"border-b min-h-11 -mb-0.25 flex items-center px-2.5 text-sm font-medium sticky top-0 bg-b-1 xl:rounded-none z-1"},l=(0,a.A)(o,[["render",function(e,t){return(0,r.uX)(),(0,r.CE)("section",i,[(0,r.Lk)("div",s,[(0,r.RG)(e.$slots,"title")]),(0,r.RG)(e.$slots,"default")])}]])},3412:(e,t,n)=>{"use strict";function r(e,t,n=[]){const a={};for(const[o,i]of Object.entries(e)){const e=[...n,o];Array.isArray(i)?a[o]=i.map(((n,a)=>"object"!=typeof n||Array.isArray(n)||null===n?n:r(n,t,[...e,a.toString()]))):"object"!=typeof i||Array.isArray(i)||null===i?a[o]=i:a[o]=r(i,t,e)}return t(a,n)}n.d(t,{h:()=>r})},3526:(e,t,n)=>{"use strict";n.d(t,{A:()=>r});const r=(e,t)=>{const n=e.__vccOpts||e;for(const[e,r]of t)n[e]=r;return n}},3605:(e,t,n)=>{"use strict";n.d(t,{O5:()=>s,XX:()=>c,cJ:()=>f,oE:()=>h});var r,a,o=n(641),i=n(9388),s=((a=s||{})[a.None=0]="None",a[a.RenderStrategy=1]="RenderStrategy",a[a.Static=2]="Static",a),l=((r=l||{})[r.Unmount=0]="Unmount",r[r.Hidden=1]="Hidden",r);function c({visible:e=!0,features:t=0,ourProps:n,theirProps:r,...a}){var o;let s=p(r,n),l=Object.assign(a,{props:s});if(e||2&t&&s.static)return u(l);if(1&t){let e=null==(o=s.unmount)||o?0:1;return(0,i.Y)(e,{0:()=>null,1:()=>u({...a,props:{...s,hidden:!0,style:{display:"none"}}})})}return u(l)}function u({props:e,attrs:t,slots:n,slot:r,name:a}){var i,s;let{as:l,...c}=f(e,["unmount","static"]),u=null==(i=n.default)?void 0:i.call(n,r),h={};if(r){let e=!1,t=[];for(let[n,a]of Object.entries(r))"boolean"==typeof a&&(e=!0),!0===a&&t.push(n);e&&(h["data-headlessui-state"]=t.join(" "))}if("template"===l){if(u=d(null!=u?u:[]),Object.keys(c).length>0||Object.keys(t).length>0){let[e,...n]=null!=u?u:[];if(!function(e){return null!=e&&("string"==typeof e.type||"object"==typeof e.type||"function"==typeof e.type)}(e)||n.length>0)throw new Error(['Passing props on "template"!',"",`The current component <${a} /> is rendering a "template".`,"However we need to passthrough the following props:",Object.keys(c).concat(Object.keys(t)).map((e=>e.trim())).filter(((e,t,n)=>n.indexOf(e)===t)).sort(((e,t)=>e.localeCompare(t))).map((e=>`  - ${e}`)).join("\n"),"","You can apply a few solutions:",['Add an `as="..."` prop, to ensure that we render an actual element instead of a "template".',"Render a single element as the child so that we can forward the props onto that element."].map((e=>`  - ${e}`)).join("\n")].join("\n"));let r=p(null!=(s=e.props)?s:{},c,h),i=(0,o.E3)(e,r,!0);for(let e in r)e.startsWith("on")&&(i.props||(i.props={}),i.props[e]=r[e]);return i}return Array.isArray(u)&&1===u.length?u[0]:u}return(0,o.h)(l,Object.assign({},c,h),{default:()=>u})}function d(e){return e.flatMap((e=>e.type===o.FK?d(e.children):[e]))}function p(...e){if(0===e.length)return{};if(1===e.length)return e[0];let t={},n={};for(let r of e)for(let e in r)e.startsWith("on")&&"function"==typeof r[e]?(null!=n[e]||(n[e]=[]),n[e].push(r[e])):t[e]=r[e];if(t.disabled||t["aria-disabled"])return Object.assign(t,Object.fromEntries(Object.keys(n).map((e=>[e,void 0]))));for(let e in n)Object.assign(t,{[e](t,...r){let a=n[e];for(let e of a){if(t instanceof Event&&t.defaultPrevented)return;e(t,...r)}}});return t}function h(e){let t=Object.assign({},e);for(let e in t)void 0===t[e]&&delete t[e];return t}function f(e,t=[]){let n=Object.assign({},e);for(let e of t)e in n&&delete n[e];return n}},3648:(e,t,n)=>{"use strict";n.d(t,{j:()=>d});var r=n(641),a=n(953),o=n(6413),i=n(1417),s=n(2966),l=n(9146);function c(e,t,n){l._.isServer||(0,r.nT)((r=>{document.addEventListener(e,t,n),r((()=>document.removeEventListener(e,t,n)))}))}var u=n(2389);function d(e,t,n=(0,r.EW)((()=>!0))){function l(r,a){if(!n.value||r.defaultPrevented)return;let s=a(r);if(null===s||!s.getRootNode().contains(s))return;let l=function e(t){return"function"==typeof t?e(t()):Array.isArray(t)||t instanceof Set?t:[t]}(e);for(let e of l){if(null===e)continue;let t=e instanceof HTMLElement?e:(0,o.t)(e);if(null!=t&&t.contains(s)||r.composed&&r.composedPath().includes(t))return}return!(0,i.Bm)(s,i.MZ.Loose)&&-1!==s.tabIndex&&r.preventDefault(),t(r,s)}let d=(0,a.KR)(null);c("pointerdown",(e=>{var t,r;n.value&&(d.value=(null==(r=null==(t=e.composedPath)?void 0:t.call(e))?void 0:r[0])||e.target)}),!0),c("mousedown",(e=>{var t,r;n.value&&(d.value=(null==(r=null==(t=e.composedPath)?void 0:t.call(e))?void 0:r[0])||e.target)}),!0),c("click",(e=>{(0,s.Fr)()||d.value&&(l(e,(()=>d.value)),d.value=null)}),!0),c("touchend",(e=>l(e,(()=>e.target instanceof HTMLElement?e.target:null))),!0),(0,u.M)("blur",(e=>l(e,(()=>window.document.activeElement instanceof HTMLIFrameElement?window.document.activeElement:null))),!0)}},3654:(e,t,n)=>{"use strict";n.d(t,{A:()=>M});var r=n(641),a=n(953),o=n(33),i=n(3751),s=n(1266),l=n(4048),c=n(9035),u=n(3806);const d=e=>e?e.color:"#8E8E8E";var p=n(3073),h=n(5929),f=n(9351);function m(e){return e.flatMap((e=>{if("value"===e.key)try{const t=JSON.parse(e.value);return Object.keys(t).map((n=>({key:n,value:t[n],source:e.source})))}catch{}return[e]}))}const g={key:0,class:"flex flex-col gap-1/2"},v=["onClick"],b={class:"flex items-center gap-1.5 whitespace-nowrap"},y={class:"w-20 overflow-hidden text-ellipsis text-right whitespace-nowrap"},w=(0,r.pM)({__name:"EnvironmentVariableDropdown",props:{query:{},environment:{},envVariables:{},dropdownPosition:{}},emits:["select"],setup(e,{expose:t,emit:n}){const i=e,l=n,c=(0,a.KR)(!0),u=(0,a.KR)(null),w=(0,a.KR)(0),O=(0,f.rd)(),k=()=>{if(!O)return;const{currentRoute:e,push:t}=O;t({name:"environment.default",params:{workspace:e.value.params.workspace}}),c.value=!1},x=new h.A(m(i.envVariables),{keys:["key","value"]}),E=(0,r.EW)((()=>{const e=i.query;if(!e)return m(i.envVariables).slice(-4).filter((({key:e,value:t})=>""!==e||""!==t));const t=x.search(e,{limit:10});return t.length>0?t.map((e=>e.item)).filter((({key:e,value:t})=>""!==e||""!==t)):[]})),S=e=>{l("select",e)};t({handleArrowKey:e=>{const t="up"===e?-1:1,n=E.value.length;0!==n&&(w.value=(w.value+t+n)%n)},handleSelect:()=>{if(w.value>=0){const e=E.value[w.value];e&&S(e.key)}}}),(0,r.sV)((()=>{w.value=0}));const _=(0,r.EW)((()=>{var e,t;return{left:((null==(e=i.dropdownPosition)?void 0:e.left)??0)+"px",top:((null==(t=i.dropdownPosition)?void 0:t.top)??0)+5+"px"}}));return(0,p.X2F)(u,(()=>{c.value=!1}),{ignore:[u]}),(e,t)=>c.value?((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.N9),{key:0,class:"scalar-client"},{default:(0,r.k6)((()=>[(0,r.Lk)("div",{ref_key:"dropdownRef",ref:u,class:"fixed left-0 top-0 flex flex-col p-0.75 max-h-[60svh] w-56 rounded border custom-scroll",style:(0,o.Tr)(_.value)},[E.value.length?((0,r.uX)(),(0,r.CE)("ul",g,[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(E.value,((t,n)=>((0,r.uX)(),(0,r.CE)("li",{key:t.key,class:(0,o.C4)(["h-8 font-code text-xxs hover:bg-b-2 flex cursor-pointer items-center justify-between gap-1.5 rounded p-1.5 transition-colors duration-150",{"bg-b-2":n===w.value}]),onClick:e=>S(t.key)},[(0,r.Lk)("div",b,["collection"===t.source?((0,r.uX)(),(0,r.CE)("span",{key:0,class:"h-2.5 w-2.5 min-w-2.5 rounded-full",style:(0,o.Tr)({backgroundColor:(0,a.R1)(d)(e.environment)})},null,4)):((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.kv),{key:1,class:"w-2.5",icon:"Globe"})),(0,r.eW)(" "+(0,o.v_)(t.key),1)]),(0,r.Lk)("span",y,(0,o.v_)(t.value),1)],10,v)))),128))])):(0,a.R1)(O)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.RO),{key:1,class:"font-code text-xxs hover:bg-b-2 flex h-8 w-full justify-start gap-2 px-1.5 transition-colors duration-150",variant:"secondary",onClick:k},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(s.kv),{icon:"Add",size:"sm"}),t[0]||(t[0]=(0,r.eW)(" Add Variable "))])),_:1})):(0,r.Q3)("",!0),t[1]||(t[1]=(0,r.Lk)("div",{class:"absolute inset-0 -z-1 rounded bg-b-1 shadow-lg brightness-lifted"},null,-1))],4)])),_:1})):(0,r.Q3)("",!0)}});var O=n(1760),k=Object.defineProperty,x=(e,t,n)=>((e,t,n)=>t in e?k(e,t,{enumerable:!0,configurable:!0,writable:!0,value:n}):e[t]=n)(e,"symbol"!=typeof t?t+"":t,n);class E extends c.xO{constructor(e,t,n,r,a){super(),x(this,"app"),x(this,"environment"),x(this,"envVariables"),x(this,"workspace"),x(this,"isReadOnly"),this.variableName=e,this.variableName=e,this.environment=t,this.envVariables=n,this.workspace=r,this.isReadOnly=a??!1}toDOM(){const e=document.createElement("span");e.className="cm-pill",e.textContent=`${this.variableName}`;const t=(0,r.pM)({props:{variableName:{type:String,default:null}},render:()=>{const t=this.envVariables?m(this.envVariables).find((e=>e.key===this.variableName)):void 0,n=t&&this.environment?d(this.environment):"#8E8E8E";e.style.setProperty("--tw-bg-base",n||"#8E8E8E"),e.style.opacity=null!=t&&t.value?"1":"0.5";const a=null!=t&&t.value?(0,r.h)("div",{class:"p-2"},t.value):(0,r.h)("div",{class:"divide-y divide-1/2 grid"},[(0,r.h)("span",{class:"p-2 opacity-25"},"No value"),!this.isReadOnly&&(0,r.h)("div",{class:"p-1"},[(0,r.h)(s.RO,{class:"gap-1.5 justify-start font-normal px-1 py-1.5 h-auto transition-colors rounded no-underline text-xxs w-full hover:bg-b-2",variant:"ghost",onClick:()=>{var e;window.location.href=`/workspace/${null==(e=this.workspace)?void 0:e.uid}/environment`}},{default:()=>[(0,r.h)(s.kv,{class:"w-2",icon:"Add",size:"xs"}),"Add variable"]})])]);return(0,r.h)(s.iy,{align:"center",class:"w-full",delay:0,side:"bottom",sideOffset:6},{trigger:()=>(0,r.h)("span",`${this.variableName}`),content:()=>(0,r.h)("div",{class:["border w-content rounded  bg-b-1 brightness-lifted text-xxs leading-5 text-c-1",null!=t&&t.value?"border-solid":"border-dashed"]},a)})}});return this.app=(0,i.Ef)(t,{variableName:this.variableName}),this.app.mount(e),e}destroy(){this.app&&this.app.unmount()}eq(e){return e instanceof E&&e.variableName===this.variableName}ignoreEvent(){return!1}}const S=e=>c.Z9.fromClass(class{constructor(e){x(this,"decorations"),this.decorations=this.buildDecorations(e)}update(e){(e.docChanged||e.viewportChanged)&&requestAnimationFrame((()=>{this.decorations=this.buildDecorations(e.view),e.view.update([])}))}buildDecorations(t){const n=new c.vB;for(const{from:r,to:a}of t.visibleRanges){const o=t.state.doc.sliceString(r,a);let i;for(;null!==(i=l.QL.VARIABLES.exec(o));){const t=r+i.index,a=t+i[0].length,o=i[1]??"";n.add(t,a,c.NZ.widget({widget:new E(o,e.environment,e.envVariables,e.workspace,e.isReadOnly),side:1}))}}return n.finish()}},{decorations:e=>e.decorations}),_=c.Lz.domEventHandlers({keydown(e,t){if("Backspace"===e.key){const{state:n}=t,{from:r,to:a}=n.selection.main;if(0===r&&a===n.doc.length)return t.dispatch({changes:{from:0,to:n.doc.length},selection:{anchor:0}}),e.preventDefault(),!0;if(r===a&&r>0&&"}}"===n.doc.sliceString(r-2,r))return t.dispatch({changes:{from:r-2,to:a},selection:{anchor:r-2}}),e.preventDefault(),!0}return!1}});var C=n(6584);const T=["id"],A={key:0,class:"scalar-code-copy z-context"},R={key:5,class:"centered-y text-orange absolute right-7 text-xs"},P={key:6,class:"required centered-y text-xxs text-c-3 bg-b-1 pointer-events-none absolute right-0 pr-2 pt-px opacity-100 shadow-[-8px_0_4px_var(--scalar-background-1)] transition-opacity duration-150 group-[.alert]:bg-transparent group-[.alert]:shadow-none peer-has-[.cm-focused]:opacity-0"},D=(0,r.pM)({inheritAttrs:!1,__name:"CodeInput",props:{colorPicker:{type:Boolean,default:!1},disabled:{type:Boolean,default:!1},modelValue:{},error:{type:Boolean},emitOnBlur:{type:Boolean,default:!0},lineNumbers:{type:Boolean},lint:{type:Boolean},disableTabIndent:{type:Boolean,default:!1},language:{},handleFieldSubmit:{},handleFieldChange:{},placeholder:{},required:{type:Boolean},disableEnter:{type:Boolean,default:!1},disableCloseBrackets:{type:Boolean,default:!1},enum:{},examples:{},type:{},nullable:{type:Boolean,default:!1},withVariables:{type:Boolean,default:!0},importCurl:{type:Boolean},isCopyable:{type:Boolean,default:!1},default:{},environment:{},envVariables:{},workspace:{}},emits:["submit","update:modelValue","curl"],setup(e,{expose:t,emit:n}){var d;const p=e,h=n,f=(0,r.OA)(),m=f.id||`id-${((e=21)=>{let t="",n=crypto.getRandomValues(new Uint8Array(e|=0));for(;e--;)t+="useandom-26T198340PX75pxJACKVERYMINDBUSHWOLF_GQZbfghjklqvwyzrict"[63&n[e]];return t})()}`,g=(0,a.KR)(!1),v=(0,a.KR)(!1),b=(0,a.KR)(""),y=(0,a.KR)({left:0,top:0}),k=(0,a.KR)(null),{layout:x}=(0,C.g)(),{copyToClipboard:E}=(0,u.i)();function D(e){return p.handleFieldSubmit?p.handleFieldSubmit(e):h("submit",e)}const M=[];p.colorPicker&&M.push(c.Ei),M.push(S({environment:p.environment,envVariables:p.envVariables,workspace:p.workspace,isReadOnly:"modal"===x}),_);const $=(0,a.KR)(null),{codeMirror:N}=(0,c.qF)({content:(0,a.lW)((()=>void 0!==p.modelValue?String(p.modelValue):"")),onChange:e=>{(function(e){var t;e===p.modelValue||(p.importCurl&&e.trim().toLowerCase().startsWith("curl")?(h("curl",e),null==(t=N.value)||t.dispatch({changes:{from:0,to:N.value.state.doc.length,insert:String(p.modelValue)}})):p.handleFieldChange?p.handleFieldChange(e):h("update:modelValue",e))})(e),L()},onFocus:()=>g.value=!0,onBlur:e=>function(e){g.value=!1,p.emitOnBlur&&p.modelValue&&D(e)}(e),codeMirrorRef:$,disableTabIndent:(0,a.lW)((()=>p.disableTabIndent)),disableEnter:(0,a.lW)((()=>p.disableEnter)),disableCloseBrackets:(0,a.lW)((()=>p.disableCloseBrackets)),lineNumbers:(0,a.lW)((()=>p.lineNumbers)),language:(0,a.lW)((()=>p.language)),lint:(0,a.lW)((()=>p.lint)),extensions:M,placeholder:(0,a.lW)((()=>p.placeholder))});null==(d=N.value)||d.focus(),(0,r.wB)(N,(()=>{N.value&&Object.prototype.hasOwnProperty.call(f,"autofocus")&&N.value.focus()}));const{handleDropdownSelect:I,updateDropdownVisibility:L}=(0,c.BK)({codeMirror:N,query:b,showDropdown:v,dropdownPosition:y}),B=(0,r.EW)((()=>{var e;return"boolean"===p.type||null!=(e=p.type)&&e.includes("boolean")||p.nullable?["true","false","null"]:["true","false"]}));t({focus:()=>{var e;null==(e=N.value)||e.focus()}});const F=(e,t)=>{var n,r,a;v.value?"down"===e?(t.preventDefault(),null==(n=k.value)||n.handleArrowKey("down")):"up"===e?(t.preventDefault(),null==(r=k.value)||r.handleArrowKey("up")):"enter"===e&&(t.preventDefault(),null==(a=k.value)||a.handleSelect()):"enter"===e&&t.target instanceof HTMLDivElement&&D(t.target.textContent??"")},Q=(0,r.EW)((()=>Array.isArray(p.type)?p.type.find((e=>"null"!==e))??"string":p.type));return(e,t)=>{var n;return(0,r.uX)(),(0,r.CE)(r.FK,null,[e.disabled?((0,r.uX)(),(0,r.CE)("div",{key:0,class:(0,o.C4)(["text-c-2 flex cursor-default items-center justify-center","modal"===(0,a.R1)(x)?"font-code pl-1 pr-2 text-sm":"px-2"])},[(0,r.Lk)("span",null,(0,o.v_)(e.modelValue),1)],2)):p.enum&&p.enum.length?((0,r.uX)(),(0,r.Wv)(O.A,{key:1,default:p.default,modelValue:e.modelValue,type:Q.value,value:p.enum,"onUpdate:modelValue":t[0]||(t[0]=e=>h("update:modelValue",e))},null,8,["default","modelValue","type","value"])):"boolean"===e.type||null!=(n=e.type)&&n.includes("boolean")?((0,r.uX)(),(0,r.Wv)(O.A,{key:2,default:p.default,modelValue:e.modelValue,value:B.value,"onUpdate:modelValue":t[1]||(t[1]=e=>h("update:modelValue",e))},null,8,["default","modelValue","value"])):p.examples&&p.examples.length?((0,r.uX)(),(0,r.Wv)(O.A,{key:3,default:p.default,modelValue:p.modelValue,value:p.examples,"onUpdate:modelValue":t[2]||(t[2]=e=>h("update:modelValue",e))},null,8,["default","modelValue","value"])):((0,r.uX)(),(0,r.CE)("div",(0,r.v6)({key:4,id:(0,a.R1)(m)},e.$attrs,{ref_key:"codeMirrorRef",ref:$,class:["group-[.alert]:outline-orange font-code peer relative w-full overflow-hidden whitespace-nowrap text-xs leading-[1.44] -outline-offset-1 has-[:focus-visible]:rounded-[4px] has-[:focus-visible]:outline",{"flow-code-input--error":e.error}],onKeydown:[t[4]||(t[4]=(0,i.jR)((0,i.D$)((e=>F("down",e)),["stop"]),["down"])),t[5]||(t[5]=(0,i.jR)((e=>F("enter",e)),["enter"])),t[6]||(t[6]=(0,i.jR)((0,i.D$)((e=>F("up",e)),["stop"]),["up"]))]}),[e.isCopyable?((0,r.uX)(),(0,r.CE)("div",A,[(0,r.Lk)("button",{class:"copy-button",type:"button",onClick:t[3]||(t[3]=t=>(0,a.R1)(E)((0,a.R1)(l.qX)(e.modelValue)))},[t[7]||(t[7]=(0,r.Lk)("span",{class:"sr-only"},"Copy content",-1)),(0,r.bF)((0,a.R1)(s.kv),{icon:"Clipboard",size:"md"})])])):(0,r.Q3)("",!0)],16,T)),e.$slots.warning?((0,r.uX)(),(0,r.CE)("div",R,[(0,r.RG)(e.$slots,"warning",{},void 0,!0)])):(0,r.Q3)("",!0),(0,r.RG)(e.$slots,"icon",{},void 0,!0),e.required?((0,r.uX)(),(0,r.CE)("div",P," Required ")):(0,r.Q3)("",!0),v.value&&e.withVariables&&"modal"!==(0,a.R1)(x)&&e.environment?((0,r.uX)(),(0,r.Wv)(w,{key:7,ref_key:"dropdownRef",ref:k,dropdownPosition:y.value,envVariables:e.envVariables,environment:e.environment,query:b.value,onSelect:(0,a.R1)(I)},null,8,["dropdownPosition","envVariables","environment","query","onSelect"])):(0,r.Q3)("",!0)],64)}}}),M=(0,n(3526).A)(D,[["__scopeId","data-v-4d46a9ea"]])},3687:(e,t,n)=>{"use strict";n.d(t,{Sr:()=>a,IY:()=>r});const r=Object.keys({"2.0":{title:"A JSON Schema for Swagger 2.0 API.",id:"http://swagger.io/v2/schema.json#",$schema:"http://json-schema.org/draft-04/schema#",type:"object",required:["swagger","info","paths"],additionalProperties:!1,patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}},properties:{swagger:{type:"string",enum:["2.0"],description:"The Swagger version of this document."},info:{$ref:"#/definitions/info"},host:{type:"string",pattern:"^[^{}/ :\\\\]+(?::\\d+)?$",description:"The host (name or ip) of the API. Example: 'swagger.io'"},basePath:{type:"string",pattern:"^/",description:"The base path to the API. Example: '/api'."},schemes:{$ref:"#/definitions/schemesList"},consumes:{description:"A list of MIME types accepted by the API.",allOf:[{$ref:"#/definitions/mediaTypeList"}]},produces:{description:"A list of MIME types the API can produce.",allOf:[{$ref:"#/definitions/mediaTypeList"}]},paths:{$ref:"#/definitions/paths"},definitions:{$ref:"#/definitions/definitions"},parameters:{$ref:"#/definitions/parameterDefinitions"},responses:{$ref:"#/definitions/responseDefinitions"},security:{$ref:"#/definitions/security"},securityDefinitions:{$ref:"#/definitions/securityDefinitions"},tags:{type:"array",items:{$ref:"#/definitions/tag"},uniqueItems:!0},externalDocs:{$ref:"#/definitions/externalDocs"}},definitions:{info:{type:"object",description:"General information about the API.",required:["version","title"],additionalProperties:!1,patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}},properties:{title:{type:"string",description:"A unique and precise title of the API."},version:{type:"string",description:"A semantic version number of the API."},description:{type:"string",description:"A longer description of the API. Should be different from the title.  GitHub Flavored Markdown is allowed."},termsOfService:{type:"string",description:"The terms of service for the API."},contact:{$ref:"#/definitions/contact"},license:{$ref:"#/definitions/license"}}},contact:{type:"object",description:"Contact information for the owners of the API.",additionalProperties:!1,properties:{name:{type:"string",description:"The identifying name of the contact person/organization."},url:{type:"string",description:"The URL pointing to the contact information.",format:"uri"},email:{type:"string",description:"The email address of the contact person/organization.",format:"email"}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},license:{type:"object",required:["name"],additionalProperties:!1,properties:{name:{type:"string",description:"The name of the license type. It's encouraged to use an OSI compatible license."},url:{type:"string",description:"The URL pointing to the license.",format:"uri"}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},paths:{type:"object",description:"Relative paths to the individual endpoints. They must be relative to the 'basePath'.",patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"},"^/":{$ref:"#/definitions/pathItem"}},additionalProperties:!1},definitions:{type:"object",additionalProperties:{$ref:"#/definitions/schema"},description:"One or more JSON objects describing the schemas being consumed and produced by the API."},parameterDefinitions:{type:"object",additionalProperties:{$ref:"#/definitions/parameter"},description:"One or more JSON representations for parameters"},responseDefinitions:{type:"object",additionalProperties:{$ref:"#/definitions/response"},description:"One or more JSON representations for responses"},externalDocs:{type:"object",additionalProperties:!1,description:"information about external documentation",required:["url"],properties:{description:{type:"string"},url:{type:"string",format:"uri"}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},examples:{type:"object",additionalProperties:!0},mimeType:{type:"string",description:"The MIME type of the HTTP message."},operation:{type:"object",required:["responses"],additionalProperties:!1,patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}},properties:{tags:{type:"array",items:{type:"string"},uniqueItems:!0},summary:{type:"string",description:"A brief summary of the operation."},description:{type:"string",description:"A longer description of the operation, GitHub Flavored Markdown is allowed."},externalDocs:{$ref:"#/definitions/externalDocs"},operationId:{type:"string",description:"A unique identifier of the operation."},produces:{description:"A list of MIME types the API can produce.",allOf:[{$ref:"#/definitions/mediaTypeList"}]},consumes:{description:"A list of MIME types the API can consume.",allOf:[{$ref:"#/definitions/mediaTypeList"}]},parameters:{$ref:"#/definitions/parametersList"},responses:{$ref:"#/definitions/responses"},schemes:{$ref:"#/definitions/schemesList"},deprecated:{type:"boolean",default:!1},security:{$ref:"#/definitions/security"}}},pathItem:{type:"object",additionalProperties:!1,patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}},properties:{$ref:{type:"string"},get:{$ref:"#/definitions/operation"},put:{$ref:"#/definitions/operation"},post:{$ref:"#/definitions/operation"},delete:{$ref:"#/definitions/operation"},options:{$ref:"#/definitions/operation"},head:{$ref:"#/definitions/operation"},patch:{$ref:"#/definitions/operation"},parameters:{$ref:"#/definitions/parametersList"}}},responses:{type:"object",description:"Response objects names can either be any valid HTTP status code or 'default'.",minProperties:1,additionalProperties:!1,patternProperties:{"^([0-9]{3})$|^(default)$":{$ref:"#/definitions/responseValue"},"^x-":{$ref:"#/definitions/vendorExtension"}},not:{type:"object",additionalProperties:!1,patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}}},responseValue:{oneOf:[{$ref:"#/definitions/response"},{$ref:"#/definitions/jsonReference"}]},response:{type:"object",required:["description"],properties:{description:{type:"string"},schema:{oneOf:[{$ref:"#/definitions/schema"},{$ref:"#/definitions/fileSchema"}]},headers:{$ref:"#/definitions/headers"},examples:{$ref:"#/definitions/examples"}},additionalProperties:!1,patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},headers:{type:"object",additionalProperties:{$ref:"#/definitions/header"}},header:{type:"object",additionalProperties:!1,required:["type"],properties:{type:{type:"string",enum:["string","number","integer","boolean","array"]},format:{type:"string"},items:{$ref:"#/definitions/primitivesItems"},collectionFormat:{$ref:"#/definitions/collectionFormat"},default:{$ref:"#/definitions/default"},maximum:{$ref:"#/definitions/maximum"},exclusiveMaximum:{$ref:"#/definitions/exclusiveMaximum"},minimum:{$ref:"#/definitions/minimum"},exclusiveMinimum:{$ref:"#/definitions/exclusiveMinimum"},maxLength:{$ref:"#/definitions/maxLength"},minLength:{$ref:"#/definitions/minLength"},pattern:{$ref:"#/definitions/pattern"},maxItems:{$ref:"#/definitions/maxItems"},minItems:{$ref:"#/definitions/minItems"},uniqueItems:{$ref:"#/definitions/uniqueItems"},enum:{$ref:"#/definitions/enum"},multipleOf:{$ref:"#/definitions/multipleOf"},description:{type:"string"}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},vendorExtension:{description:"Any property starting with x- is valid.",additionalProperties:!0,additionalItems:!0},bodyParameter:{type:"object",required:["name","in","schema"],patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}},properties:{description:{type:"string",description:"A brief description of the parameter. This could contain examples of use.  GitHub Flavored Markdown is allowed."},name:{type:"string",description:"The name of the parameter."},in:{type:"string",description:"Determines the location of the parameter.",enum:["body"]},required:{type:"boolean",description:"Determines whether or not this parameter is required or optional.",default:!1},schema:{$ref:"#/definitions/schema"}},additionalProperties:!1},headerParameterSubSchema:{additionalProperties:!1,patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}},properties:{required:{type:"boolean",description:"Determines whether or not this parameter is required or optional.",default:!1},in:{type:"string",description:"Determines the location of the parameter.",enum:["header"]},description:{type:"string",description:"A brief description of the parameter. This could contain examples of use.  GitHub Flavored Markdown is allowed."},name:{type:"string",description:"The name of the parameter."},type:{type:"string",enum:["string","number","boolean","integer","array"]},format:{type:"string"},items:{$ref:"#/definitions/primitivesItems"},collectionFormat:{$ref:"#/definitions/collectionFormat"},default:{$ref:"#/definitions/default"},maximum:{$ref:"#/definitions/maximum"},exclusiveMaximum:{$ref:"#/definitions/exclusiveMaximum"},minimum:{$ref:"#/definitions/minimum"},exclusiveMinimum:{$ref:"#/definitions/exclusiveMinimum"},maxLength:{$ref:"#/definitions/maxLength"},minLength:{$ref:"#/definitions/minLength"},pattern:{$ref:"#/definitions/pattern"},maxItems:{$ref:"#/definitions/maxItems"},minItems:{$ref:"#/definitions/minItems"},uniqueItems:{$ref:"#/definitions/uniqueItems"},enum:{$ref:"#/definitions/enum"},multipleOf:{$ref:"#/definitions/multipleOf"}}},queryParameterSubSchema:{additionalProperties:!1,patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}},properties:{required:{type:"boolean",description:"Determines whether or not this parameter is required or optional.",default:!1},in:{type:"string",description:"Determines the location of the parameter.",enum:["query"]},description:{type:"string",description:"A brief description of the parameter. This could contain examples of use.  GitHub Flavored Markdown is allowed."},name:{type:"string",description:"The name of the parameter."},allowEmptyValue:{type:"boolean",default:!1,description:"allows sending a parameter by name only or with an empty value."},type:{type:"string",enum:["string","number","boolean","integer","array"]},format:{type:"string"},items:{$ref:"#/definitions/primitivesItems"},collectionFormat:{$ref:"#/definitions/collectionFormatWithMulti"},default:{$ref:"#/definitions/default"},maximum:{$ref:"#/definitions/maximum"},exclusiveMaximum:{$ref:"#/definitions/exclusiveMaximum"},minimum:{$ref:"#/definitions/minimum"},exclusiveMinimum:{$ref:"#/definitions/exclusiveMinimum"},maxLength:{$ref:"#/definitions/maxLength"},minLength:{$ref:"#/definitions/minLength"},pattern:{$ref:"#/definitions/pattern"},maxItems:{$ref:"#/definitions/maxItems"},minItems:{$ref:"#/definitions/minItems"},uniqueItems:{$ref:"#/definitions/uniqueItems"},enum:{$ref:"#/definitions/enum"},multipleOf:{$ref:"#/definitions/multipleOf"}}},formDataParameterSubSchema:{additionalProperties:!1,patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}},properties:{required:{type:"boolean",description:"Determines whether or not this parameter is required or optional.",default:!1},in:{type:"string",description:"Determines the location of the parameter.",enum:["formData"]},description:{type:"string",description:"A brief description of the parameter. This could contain examples of use.  GitHub Flavored Markdown is allowed."},name:{type:"string",description:"The name of the parameter."},allowEmptyValue:{type:"boolean",default:!1,description:"allows sending a parameter by name only or with an empty value."},type:{type:"string",enum:["string","number","boolean","integer","array","file"]},format:{type:"string"},items:{$ref:"#/definitions/primitivesItems"},collectionFormat:{$ref:"#/definitions/collectionFormatWithMulti"},default:{$ref:"#/definitions/default"},maximum:{$ref:"#/definitions/maximum"},exclusiveMaximum:{$ref:"#/definitions/exclusiveMaximum"},minimum:{$ref:"#/definitions/minimum"},exclusiveMinimum:{$ref:"#/definitions/exclusiveMinimum"},maxLength:{$ref:"#/definitions/maxLength"},minLength:{$ref:"#/definitions/minLength"},pattern:{$ref:"#/definitions/pattern"},maxItems:{$ref:"#/definitions/maxItems"},minItems:{$ref:"#/definitions/minItems"},uniqueItems:{$ref:"#/definitions/uniqueItems"},enum:{$ref:"#/definitions/enum"},multipleOf:{$ref:"#/definitions/multipleOf"}}},pathParameterSubSchema:{additionalProperties:!1,patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}},required:["required"],properties:{required:{type:"boolean",enum:[!0],description:"Determines whether or not this parameter is required or optional."},in:{type:"string",description:"Determines the location of the parameter.",enum:["path"]},description:{type:"string",description:"A brief description of the parameter. This could contain examples of use.  GitHub Flavored Markdown is allowed."},name:{type:"string",description:"The name of the parameter."},type:{type:"string",enum:["string","number","boolean","integer","array"]},format:{type:"string"},items:{$ref:"#/definitions/primitivesItems"},collectionFormat:{$ref:"#/definitions/collectionFormat"},default:{$ref:"#/definitions/default"},maximum:{$ref:"#/definitions/maximum"},exclusiveMaximum:{$ref:"#/definitions/exclusiveMaximum"},minimum:{$ref:"#/definitions/minimum"},exclusiveMinimum:{$ref:"#/definitions/exclusiveMinimum"},maxLength:{$ref:"#/definitions/maxLength"},minLength:{$ref:"#/definitions/minLength"},pattern:{$ref:"#/definitions/pattern"},maxItems:{$ref:"#/definitions/maxItems"},minItems:{$ref:"#/definitions/minItems"},uniqueItems:{$ref:"#/definitions/uniqueItems"},enum:{$ref:"#/definitions/enum"},multipleOf:{$ref:"#/definitions/multipleOf"}}},nonBodyParameter:{type:"object",required:["name","in","type"],oneOf:[{$ref:"#/definitions/headerParameterSubSchema"},{$ref:"#/definitions/formDataParameterSubSchema"},{$ref:"#/definitions/queryParameterSubSchema"},{$ref:"#/definitions/pathParameterSubSchema"}]},parameter:{oneOf:[{$ref:"#/definitions/bodyParameter"},{$ref:"#/definitions/nonBodyParameter"}]},schema:{type:"object",description:"A deterministic version of a JSON Schema object.",patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}},properties:{$ref:{type:"string"},format:{type:"string"},title:{$ref:"http://json-schema.org/draft-04/schema#/properties/title"},description:{$ref:"http://json-schema.org/draft-04/schema#/properties/description"},default:{$ref:"http://json-schema.org/draft-04/schema#/properties/default"},multipleOf:{$ref:"http://json-schema.org/draft-04/schema#/properties/multipleOf"},maximum:{$ref:"http://json-schema.org/draft-04/schema#/properties/maximum"},exclusiveMaximum:{$ref:"http://json-schema.org/draft-04/schema#/properties/exclusiveMaximum"},minimum:{$ref:"http://json-schema.org/draft-04/schema#/properties/minimum"},exclusiveMinimum:{$ref:"http://json-schema.org/draft-04/schema#/properties/exclusiveMinimum"},maxLength:{$ref:"http://json-schema.org/draft-04/schema#/definitions/positiveInteger"},minLength:{$ref:"http://json-schema.org/draft-04/schema#/definitions/positiveIntegerDefault0"},pattern:{$ref:"http://json-schema.org/draft-04/schema#/properties/pattern"},maxItems:{$ref:"http://json-schema.org/draft-04/schema#/definitions/positiveInteger"},minItems:{$ref:"http://json-schema.org/draft-04/schema#/definitions/positiveIntegerDefault0"},uniqueItems:{$ref:"http://json-schema.org/draft-04/schema#/properties/uniqueItems"},maxProperties:{$ref:"http://json-schema.org/draft-04/schema#/definitions/positiveInteger"},minProperties:{$ref:"http://json-schema.org/draft-04/schema#/definitions/positiveIntegerDefault0"},required:{$ref:"http://json-schema.org/draft-04/schema#/definitions/stringArray"},enum:{$ref:"http://json-schema.org/draft-04/schema#/properties/enum"},additionalProperties:{anyOf:[{$ref:"#/definitions/schema"},{type:"boolean"}],default:{}},type:{$ref:"http://json-schema.org/draft-04/schema#/properties/type"},items:{anyOf:[{$ref:"#/definitions/schema"},{type:"array",minItems:1,items:{$ref:"#/definitions/schema"}}],default:{}},allOf:{type:"array",minItems:1,items:{$ref:"#/definitions/schema"}},properties:{type:"object",additionalProperties:{$ref:"#/definitions/schema"},default:{}},discriminator:{type:"string"},readOnly:{type:"boolean",default:!1},xml:{$ref:"#/definitions/xml"},externalDocs:{$ref:"#/definitions/externalDocs"},example:{}},additionalProperties:!1},fileSchema:{type:"object",description:"A deterministic version of a JSON Schema object.",patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}},required:["type"],properties:{format:{type:"string"},title:{$ref:"http://json-schema.org/draft-04/schema#/properties/title"},description:{$ref:"http://json-schema.org/draft-04/schema#/properties/description"},default:{$ref:"http://json-schema.org/draft-04/schema#/properties/default"},required:{$ref:"http://json-schema.org/draft-04/schema#/definitions/stringArray"},type:{type:"string",enum:["file"]},readOnly:{type:"boolean",default:!1},externalDocs:{$ref:"#/definitions/externalDocs"},example:{}},additionalProperties:!1},primitivesItems:{type:"object",additionalProperties:!1,properties:{type:{type:"string",enum:["string","number","integer","boolean","array"]},format:{type:"string"},items:{$ref:"#/definitions/primitivesItems"},collectionFormat:{$ref:"#/definitions/collectionFormat"},default:{$ref:"#/definitions/default"},maximum:{$ref:"#/definitions/maximum"},exclusiveMaximum:{$ref:"#/definitions/exclusiveMaximum"},minimum:{$ref:"#/definitions/minimum"},exclusiveMinimum:{$ref:"#/definitions/exclusiveMinimum"},maxLength:{$ref:"#/definitions/maxLength"},minLength:{$ref:"#/definitions/minLength"},pattern:{$ref:"#/definitions/pattern"},maxItems:{$ref:"#/definitions/maxItems"},minItems:{$ref:"#/definitions/minItems"},uniqueItems:{$ref:"#/definitions/uniqueItems"},enum:{$ref:"#/definitions/enum"},multipleOf:{$ref:"#/definitions/multipleOf"}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},security:{type:"array",items:{$ref:"#/definitions/securityRequirement"},uniqueItems:!0},securityRequirement:{type:"object",additionalProperties:{type:"array",items:{type:"string"},uniqueItems:!0}},xml:{type:"object",additionalProperties:!1,properties:{name:{type:"string"},namespace:{type:"string"},prefix:{type:"string"},attribute:{type:"boolean",default:!1},wrapped:{type:"boolean",default:!1}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},tag:{type:"object",additionalProperties:!1,required:["name"],properties:{name:{type:"string"},description:{type:"string"},externalDocs:{$ref:"#/definitions/externalDocs"}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},securityDefinitions:{type:"object",additionalProperties:{oneOf:[{$ref:"#/definitions/basicAuthenticationSecurity"},{$ref:"#/definitions/apiKeySecurity"},{$ref:"#/definitions/oauth2ImplicitSecurity"},{$ref:"#/definitions/oauth2PasswordSecurity"},{$ref:"#/definitions/oauth2ApplicationSecurity"},{$ref:"#/definitions/oauth2AccessCodeSecurity"}]}},basicAuthenticationSecurity:{type:"object",additionalProperties:!1,required:["type"],properties:{type:{type:"string",enum:["basic"]},description:{type:"string"}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},apiKeySecurity:{type:"object",additionalProperties:!1,required:["type","name","in"],properties:{type:{type:"string",enum:["apiKey"]},name:{type:"string"},in:{type:"string",enum:["header","query"]},description:{type:"string"}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},oauth2ImplicitSecurity:{type:"object",additionalProperties:!1,required:["type","flow","authorizationUrl"],properties:{type:{type:"string",enum:["oauth2"]},flow:{type:"string",enum:["implicit"]},scopes:{$ref:"#/definitions/oauth2Scopes"},authorizationUrl:{type:"string",format:"uri"},description:{type:"string"}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},oauth2PasswordSecurity:{type:"object",additionalProperties:!1,required:["type","flow","tokenUrl"],properties:{type:{type:"string",enum:["oauth2"]},flow:{type:"string",enum:["password"]},scopes:{$ref:"#/definitions/oauth2Scopes"},tokenUrl:{type:"string",format:"uri"},description:{type:"string"}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},oauth2ApplicationSecurity:{type:"object",additionalProperties:!1,required:["type","flow","tokenUrl"],properties:{type:{type:"string",enum:["oauth2"]},flow:{type:"string",enum:["application"]},scopes:{$ref:"#/definitions/oauth2Scopes"},tokenUrl:{type:"string",format:"uri"},description:{type:"string"}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},oauth2AccessCodeSecurity:{type:"object",additionalProperties:!1,required:["type","flow","authorizationUrl","tokenUrl"],properties:{type:{type:"string",enum:["oauth2"]},flow:{type:"string",enum:["accessCode"]},scopes:{$ref:"#/definitions/oauth2Scopes"},authorizationUrl:{type:"string",format:"uri"},tokenUrl:{type:"string",format:"uri"},description:{type:"string"}},patternProperties:{"^x-":{$ref:"#/definitions/vendorExtension"}}},oauth2Scopes:{type:"object",additionalProperties:{type:"string"}},mediaTypeList:{type:"array",items:{$ref:"#/definitions/mimeType"},uniqueItems:!0},parametersList:{type:"array",description:"The parameters needed to send a valid API call.",additionalItems:!1,items:{oneOf:[{$ref:"#/definitions/parameter"},{$ref:"#/definitions/jsonReference"}]},uniqueItems:!0},schemesList:{type:"array",description:"The transfer protocol of the API.",items:{type:"string",enum:["http","https","ws","wss"]},uniqueItems:!0},collectionFormat:{type:"string",enum:["csv","ssv","tsv","pipes"],default:"csv"},collectionFormatWithMulti:{type:"string",enum:["csv","ssv","tsv","pipes","multi"],default:"csv"},title:{$ref:"http://json-schema.org/draft-04/schema#/properties/title"},description:{$ref:"http://json-schema.org/draft-04/schema#/properties/description"},default:{$ref:"http://json-schema.org/draft-04/schema#/properties/default"},multipleOf:{$ref:"http://json-schema.org/draft-04/schema#/properties/multipleOf"},maximum:{$ref:"http://json-schema.org/draft-04/schema#/properties/maximum"},exclusiveMaximum:{$ref:"http://json-schema.org/draft-04/schema#/properties/exclusiveMaximum"},minimum:{$ref:"http://json-schema.org/draft-04/schema#/properties/minimum"},exclusiveMinimum:{$ref:"http://json-schema.org/draft-04/schema#/properties/exclusiveMinimum"},maxLength:{$ref:"http://json-schema.org/draft-04/schema#/definitions/positiveInteger"},minLength:{$ref:"http://json-schema.org/draft-04/schema#/definitions/positiveIntegerDefault0"},pattern:{$ref:"http://json-schema.org/draft-04/schema#/properties/pattern"},maxItems:{$ref:"http://json-schema.org/draft-04/schema#/definitions/positiveInteger"},minItems:{$ref:"http://json-schema.org/draft-04/schema#/definitions/positiveIntegerDefault0"},uniqueItems:{$ref:"http://json-schema.org/draft-04/schema#/properties/uniqueItems"},enum:{$ref:"http://json-schema.org/draft-04/schema#/properties/enum"},jsonReference:{type:"object",required:["$ref"],additionalProperties:!1,properties:{$ref:{type:"string"}}}}},"3.0":{id:"https://spec.openapis.org/oas/3.0/schema/2021-09-28",$schema:"http://json-schema.org/draft-04/schema#",description:"The description of OpenAPI v3.0.x documents, as defined by https://spec.openapis.org/oas/v3.0.3",type:"object",required:["openapi","info","paths"],properties:{openapi:{type:"string",pattern:"^3\\.0\\.\\d(-.+)?$"},info:{$ref:"#/definitions/Info"},externalDocs:{$ref:"#/definitions/ExternalDocumentation"},servers:{type:"array",items:{$ref:"#/definitions/Server"}},security:{type:"array",items:{$ref:"#/definitions/SecurityRequirement"}},tags:{type:"array",items:{$ref:"#/definitions/Tag"},uniqueItems:!0},paths:{$ref:"#/definitions/Paths"},components:{$ref:"#/definitions/Components"}},patternProperties:{"^x-":{}},additionalProperties:!1,definitions:{Reference:{type:"object",required:["$ref"],patternProperties:{"^\\$ref$":{type:"string",format:"uri-reference"}}},Info:{type:"object",required:["title","version"],properties:{title:{type:"string"},description:{type:"string"},termsOfService:{type:"string",format:"uri-reference"},contact:{$ref:"#/definitions/Contact"},license:{$ref:"#/definitions/License"},version:{type:"string"}},patternProperties:{"^x-":{}},additionalProperties:!1},Contact:{type:"object",properties:{name:{type:"string"},url:{type:"string",format:"uri-reference"},email:{type:"string",format:"email"}},patternProperties:{"^x-":{}},additionalProperties:!1},License:{type:"object",required:["name"],properties:{name:{type:"string"},url:{type:"string",format:"uri-reference"}},patternProperties:{"^x-":{}},additionalProperties:!1},Server:{type:"object",required:["url"],properties:{url:{type:"string"},description:{type:"string"},variables:{type:"object",additionalProperties:{$ref:"#/definitions/ServerVariable"}}},patternProperties:{"^x-":{}},additionalProperties:!1},ServerVariable:{type:"object",required:["default"],properties:{enum:{type:"array",items:{type:"string"}},default:{type:"string"},description:{type:"string"}},patternProperties:{"^x-":{}},additionalProperties:!1},Components:{type:"object",properties:{schemas:{type:"object",patternProperties:{"^[a-zA-Z0-9\\.\\-_]+$":{oneOf:[{$ref:"#/definitions/Schema"},{$ref:"#/definitions/Reference"}]}}},responses:{type:"object",patternProperties:{"^[a-zA-Z0-9\\.\\-_]+$":{oneOf:[{$ref:"#/definitions/Reference"},{$ref:"#/definitions/Response"}]}}},parameters:{type:"object",patternProperties:{"^[a-zA-Z0-9\\.\\-_]+$":{oneOf:[{$ref:"#/definitions/Reference"},{$ref:"#/definitions/Parameter"}]}}},examples:{type:"object",patternProperties:{"^[a-zA-Z0-9\\.\\-_]+$":{oneOf:[{$ref:"#/definitions/Reference"},{$ref:"#/definitions/Example"}]}}},requestBodies:{type:"object",patternProperties:{"^[a-zA-Z0-9\\.\\-_]+$":{oneOf:[{$ref:"#/definitions/Reference"},{$ref:"#/definitions/RequestBody"}]}}},headers:{type:"object",patternProperties:{"^[a-zA-Z0-9\\.\\-_]+$":{oneOf:[{$ref:"#/definitions/Reference"},{$ref:"#/definitions/Header"}]}}},securitySchemes:{type:"object",patternProperties:{"^[a-zA-Z0-9\\.\\-_]+$":{oneOf:[{$ref:"#/definitions/Reference"},{$ref:"#/definitions/SecurityScheme"}]}}},links:{type:"object",patternProperties:{"^[a-zA-Z0-9\\.\\-_]+$":{oneOf:[{$ref:"#/definitions/Reference"},{$ref:"#/definitions/Link"}]}}},callbacks:{type:"object",patternProperties:{"^[a-zA-Z0-9\\.\\-_]+$":{oneOf:[{$ref:"#/definitions/Reference"},{$ref:"#/definitions/Callback"}]}}}},patternProperties:{"^x-":{}},additionalProperties:!1},Schema:{type:"object",properties:{title:{type:"string"},multipleOf:{type:"number",minimum:0,exclusiveMinimum:!0},maximum:{type:"number"},exclusiveMaximum:{type:"boolean",default:!1},minimum:{type:"number"},exclusiveMinimum:{type:"boolean",default:!1},maxLength:{type:"integer",minimum:0},minLength:{type:"integer",minimum:0,default:0},pattern:{type:"string",format:"regex"},maxItems:{type:"integer",minimum:0},minItems:{type:"integer",minimum:0,default:0},uniqueItems:{type:"boolean",default:!1},maxProperties:{type:"integer",minimum:0},minProperties:{type:"integer",minimum:0,default:0},required:{type:"array",items:{type:"string"},minItems:1,uniqueItems:!0},enum:{type:"array",items:{},minItems:1,uniqueItems:!1},type:{type:"string",enum:["array","boolean","integer","number","object","string"]},not:{oneOf:[{$ref:"#/definitions/Schema"},{$ref:"#/definitions/Reference"}]},allOf:{type:"array",items:{oneOf:[{$ref:"#/definitions/Schema"},{$ref:"#/definitions/Reference"}]}},oneOf:{type:"array",items:{oneOf:[{$ref:"#/definitions/Schema"},{$ref:"#/definitions/Reference"}]}},anyOf:{type:"array",items:{oneOf:[{$ref:"#/definitions/Schema"},{$ref:"#/definitions/Reference"}]}},items:{oneOf:[{$ref:"#/definitions/Schema"},{$ref:"#/definitions/Reference"}]},properties:{type:"object",additionalProperties:{oneOf:[{$ref:"#/definitions/Schema"},{$ref:"#/definitions/Reference"}]}},additionalProperties:{oneOf:[{$ref:"#/definitions/Schema"},{$ref:"#/definitions/Reference"},{type:"boolean"}],default:!0},description:{type:"string"},format:{type:"string"},default:{},nullable:{type:"boolean",default:!1},discriminator:{$ref:"#/definitions/Discriminator"},readOnly:{type:"boolean",default:!1},writeOnly:{type:"boolean",default:!1},example:{},externalDocs:{$ref:"#/definitions/ExternalDocumentation"},deprecated:{type:"boolean",default:!1},xml:{$ref:"#/definitions/XML"}},patternProperties:{"^x-":{}},additionalProperties:!1},Discriminator:{type:"object",required:["propertyName"],properties:{propertyName:{type:"string"},mapping:{type:"object",additionalProperties:{type:"string"}}}},XML:{type:"object",properties:{name:{type:"string"},namespace:{type:"string",format:"uri"},prefix:{type:"string"},attribute:{type:"boolean",default:!1},wrapped:{type:"boolean",default:!1}},patternProperties:{"^x-":{}},additionalProperties:!1},Response:{type:"object",required:["description"],properties:{description:{type:"string"},headers:{type:"object",additionalProperties:{oneOf:[{$ref:"#/definitions/Header"},{$ref:"#/definitions/Reference"}]}},content:{type:"object",additionalProperties:{$ref:"#/definitions/MediaType"}},links:{type:"object",additionalProperties:{oneOf:[{$ref:"#/definitions/Link"},{$ref:"#/definitions/Reference"}]}}},patternProperties:{"^x-":{}},additionalProperties:!1},MediaType:{type:"object",properties:{schema:{oneOf:[{$ref:"#/definitions/Schema"},{$ref:"#/definitions/Reference"}]},example:{},examples:{type:"object",additionalProperties:{oneOf:[{$ref:"#/definitions/Example"},{$ref:"#/definitions/Reference"}]}},encoding:{type:"object",additionalProperties:{$ref:"#/definitions/Encoding"}}},patternProperties:{"^x-":{}},additionalProperties:!1,allOf:[{$ref:"#/definitions/ExampleXORExamples"}]},Example:{type:"object",properties:{summary:{type:"string"},description:{type:"string"},value:{},externalValue:{type:"string",format:"uri-reference"}},patternProperties:{"^x-":{}},additionalProperties:!1},Header:{type:"object",properties:{description:{type:"string"},required:{type:"boolean",default:!1},deprecated:{type:"boolean",default:!1},allowEmptyValue:{type:"boolean",default:!1},style:{type:"string",enum:["simple"],default:"simple"},explode:{type:"boolean"},allowReserved:{type:"boolean",default:!1},schema:{oneOf:[{$ref:"#/definitions/Schema"},{$ref:"#/definitions/Reference"}]},content:{type:"object",additionalProperties:{$ref:"#/definitions/MediaType"},minProperties:1,maxProperties:1},example:{},examples:{type:"object",additionalProperties:{oneOf:[{$ref:"#/definitions/Example"},{$ref:"#/definitions/Reference"}]}}},patternProperties:{"^x-":{}},additionalProperties:!1,allOf:[{$ref:"#/definitions/ExampleXORExamples"},{$ref:"#/definitions/SchemaXORContent"}]},Paths:{type:"object",patternProperties:{"^\\/":{$ref:"#/definitions/PathItem"},"^x-":{}},additionalProperties:!1},PathItem:{type:"object",properties:{$ref:{type:"string"},summary:{type:"string"},description:{type:"string"},servers:{type:"array",items:{$ref:"#/definitions/Server"}},parameters:{type:"array",items:{oneOf:[{$ref:"#/definitions/Parameter"},{$ref:"#/definitions/Reference"}]},uniqueItems:!0}},patternProperties:{"^(get|put|post|delete|options|head|patch|trace)$":{$ref:"#/definitions/Operation"},"^x-":{}},additionalProperties:!1},Operation:{type:"object",required:["responses"],properties:{tags:{type:"array",items:{type:"string"}},summary:{type:"string"},description:{type:"string"},externalDocs:{$ref:"#/definitions/ExternalDocumentation"},operationId:{type:"string"},parameters:{type:"array",items:{oneOf:[{$ref:"#/definitions/Parameter"},{$ref:"#/definitions/Reference"}]},uniqueItems:!0},requestBody:{oneOf:[{$ref:"#/definitions/RequestBody"},{$ref:"#/definitions/Reference"}]},responses:{$ref:"#/definitions/Responses"},callbacks:{type:"object",additionalProperties:{oneOf:[{$ref:"#/definitions/Callback"},{$ref:"#/definitions/Reference"}]}},deprecated:{type:"boolean",default:!1},security:{type:"array",items:{$ref:"#/definitions/SecurityRequirement"}},servers:{type:"array",items:{$ref:"#/definitions/Server"}}},patternProperties:{"^x-":{}},additionalProperties:!1},Responses:{type:"object",properties:{default:{oneOf:[{$ref:"#/definitions/Response"},{$ref:"#/definitions/Reference"}]}},patternProperties:{"^[1-5](?:\\d{2}|XX)$":{oneOf:[{$ref:"#/definitions/Response"},{$ref:"#/definitions/Reference"}]},"^x-":{}},minProperties:1,additionalProperties:!1},SecurityRequirement:{type:"object",additionalProperties:{type:"array",items:{type:"string"}}},Tag:{type:"object",required:["name"],properties:{name:{type:"string"},description:{type:"string"},externalDocs:{$ref:"#/definitions/ExternalDocumentation"}},patternProperties:{"^x-":{}},additionalProperties:!1},ExternalDocumentation:{type:"object",required:["url"],properties:{description:{type:"string"},url:{type:"string",format:"uri-reference"}},patternProperties:{"^x-":{}},additionalProperties:!1},ExampleXORExamples:{description:"Example and examples are mutually exclusive",not:{required:["example","examples"]}},SchemaXORContent:{description:"Schema and content are mutually exclusive, at least one is required",not:{required:["schema","content"]},oneOf:[{required:["schema"]},{required:["content"],description:"Some properties are not allowed if content is present",allOf:[{not:{required:["style"]}},{not:{required:["explode"]}},{not:{required:["allowReserved"]}},{not:{required:["example"]}},{not:{required:["examples"]}}]}]},Parameter:{type:"object",properties:{name:{type:"string"},in:{type:"string"},description:{type:"string"},required:{type:"boolean",default:!1},deprecated:{type:"boolean",default:!1},allowEmptyValue:{type:"boolean",default:!1},style:{type:"string"},explode:{type:"boolean"},allowReserved:{type:"boolean",default:!1},schema:{oneOf:[{$ref:"#/definitions/Schema"},{$ref:"#/definitions/Reference"}]},content:{type:"object",additionalProperties:{$ref:"#/definitions/MediaType"},minProperties:1,maxProperties:1},example:{},examples:{type:"object",additionalProperties:{oneOf:[{$ref:"#/definitions/Example"},{$ref:"#/definitions/Reference"}]}}},patternProperties:{"^x-":{}},additionalProperties:!1,required:["name","in"],allOf:[{$ref:"#/definitions/ExampleXORExamples"},{$ref:"#/definitions/SchemaXORContent"},{$ref:"#/definitions/ParameterLocation"}]},ParameterLocation:{description:"Parameter location",oneOf:[{description:"Parameter in path",required:["required"],properties:{in:{enum:["path"]},style:{enum:["matrix","label","simple"],default:"simple"},required:{enum:[!0]}}},{description:"Parameter in query",properties:{in:{enum:["query"]},style:{enum:["form","spaceDelimited","pipeDelimited","deepObject"],default:"form"}}},{description:"Parameter in header",properties:{in:{enum:["header"]},style:{enum:["simple"],default:"simple"}}},{description:"Parameter in cookie",properties:{in:{enum:["cookie"]},style:{enum:["form"],default:"form"}}}]},RequestBody:{type:"object",required:["content"],properties:{description:{type:"string"},content:{type:"object",additionalProperties:{$ref:"#/definitions/MediaType"}},required:{type:"boolean",default:!1}},patternProperties:{"^x-":{}},additionalProperties:!1},SecurityScheme:{oneOf:[{$ref:"#/definitions/APIKeySecurityScheme"},{$ref:"#/definitions/HTTPSecurityScheme"},{$ref:"#/definitions/OAuth2SecurityScheme"},{$ref:"#/definitions/OpenIdConnectSecurityScheme"}]},APIKeySecurityScheme:{type:"object",required:["type","name","in"],properties:{type:{type:"string",enum:["apiKey"]},name:{type:"string"},in:{type:"string",enum:["header","query","cookie"]},description:{type:"string"}},patternProperties:{"^x-":{}},additionalProperties:!1},HTTPSecurityScheme:{type:"object",required:["scheme","type"],properties:{scheme:{type:"string"},bearerFormat:{type:"string"},description:{type:"string"},type:{type:"string",enum:["http"]}},patternProperties:{"^x-":{}},additionalProperties:!1,oneOf:[{description:"Bearer",properties:{scheme:{type:"string",pattern:"^[Bb][Ee][Aa][Rr][Ee][Rr]$"}}},{description:"Non Bearer",not:{required:["bearerFormat"]},properties:{scheme:{not:{type:"string",pattern:"^[Bb][Ee][Aa][Rr][Ee][Rr]$"}}}}]},OAuth2SecurityScheme:{type:"object",required:["type","flows"],properties:{type:{type:"string",enum:["oauth2"]},flows:{$ref:"#/definitions/OAuthFlows"},description:{type:"string"}},patternProperties:{"^x-":{}},additionalProperties:!1},OpenIdConnectSecurityScheme:{type:"object",required:["type","openIdConnectUrl"],properties:{type:{type:"string",enum:["openIdConnect"]},openIdConnectUrl:{type:"string",format:"uri-reference"},description:{type:"string"}},patternProperties:{"^x-":{}},additionalProperties:!1},OAuthFlows:{type:"object",properties:{implicit:{$ref:"#/definitions/ImplicitOAuthFlow"},password:{$ref:"#/definitions/PasswordOAuthFlow"},clientCredentials:{$ref:"#/definitions/ClientCredentialsFlow"},authorizationCode:{$ref:"#/definitions/AuthorizationCodeOAuthFlow"}},patternProperties:{"^x-":{}},additionalProperties:!1},ImplicitOAuthFlow:{type:"object",required:["authorizationUrl","scopes"],properties:{authorizationUrl:{type:"string",format:"uri-reference"},refreshUrl:{type:"string",format:"uri-reference"},scopes:{type:"object",additionalProperties:{type:"string"}}},patternProperties:{"^x-":{}},additionalProperties:!1},PasswordOAuthFlow:{type:"object",required:["tokenUrl","scopes"],properties:{tokenUrl:{type:"string",format:"uri-reference"},refreshUrl:{type:"string",format:"uri-reference"},scopes:{type:"object",additionalProperties:{type:"string"}}},patternProperties:{"^x-":{}},additionalProperties:!1},ClientCredentialsFlow:{type:"object",required:["tokenUrl","scopes"],properties:{tokenUrl:{type:"string",format:"uri-reference"},refreshUrl:{type:"string",format:"uri-reference"},scopes:{type:"object",additionalProperties:{type:"string"}}},patternProperties:{"^x-":{}},additionalProperties:!1},AuthorizationCodeOAuthFlow:{type:"object",required:["authorizationUrl","tokenUrl","scopes"],properties:{authorizationUrl:{type:"string",format:"uri-reference"},tokenUrl:{type:"string",format:"uri-reference"},refreshUrl:{type:"string",format:"uri-reference"},scopes:{type:"object",additionalProperties:{type:"string"}}},patternProperties:{"^x-":{}},additionalProperties:!1},Link:{type:"object",properties:{operationId:{type:"string"},operationRef:{type:"string",format:"uri-reference"},parameters:{type:"object",additionalProperties:{}},requestBody:{},description:{type:"string"},server:{$ref:"#/definitions/Server"}},patternProperties:{"^x-":{}},additionalProperties:!1,not:{description:"Operation Id and Operation Ref are mutually exclusive",required:["operationId","operationRef"]}},Callback:{type:"object",additionalProperties:{$ref:"#/definitions/PathItem"},patternProperties:{"^x-":{}}},Encoding:{type:"object",properties:{contentType:{type:"string"},headers:{type:"object",additionalProperties:{oneOf:[{$ref:"#/definitions/Header"},{$ref:"#/definitions/Reference"}]}},style:{type:"string",enum:["form","spaceDelimited","pipeDelimited","deepObject"]},explode:{type:"boolean"},allowReserved:{type:"boolean",default:!1}},additionalProperties:!1}}},3.1:{$id:"https://spec.openapis.org/oas/3.1/schema/2022-10-07",$schema:"https://json-schema.org/draft/2020-12/schema",description:"The description of OpenAPI v3.1.x documents without schema validation, as defined by https://spec.openapis.org/oas/v3.1.0",type:"object",properties:{openapi:{type:"string",pattern:"^3\\.1\\.\\d+(-.+)?$"},info:{$ref:"#/$defs/info"},jsonSchemaDialect:{type:"string",format:"uri-reference",default:"https://spec.openapis.org/oas/3.1/dialect/base"},servers:{type:"array",items:{$ref:"#/$defs/server"},default:[{url:"/"}]},paths:{$ref:"#/$defs/paths"},webhooks:{type:"object",additionalProperties:{$ref:"#/$defs/path-item-or-reference"}},components:{$ref:"#/$defs/components"},security:{type:"array",items:{$ref:"#/$defs/security-requirement"}},tags:{type:"array",items:{$ref:"#/$defs/tag"}},externalDocs:{$ref:"#/$defs/external-documentation"}},required:["openapi","info"],anyOf:[{required:["paths"]},{required:["components"]},{required:["webhooks"]}],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1,$defs:{info:{$comment:"https://spec.openapis.org/oas/v3.1.0#info-object",type:"object",properties:{title:{type:"string"},summary:{type:"string"},description:{type:"string"},termsOfService:{type:"string",format:"uri-reference"},contact:{$ref:"#/$defs/contact"},license:{$ref:"#/$defs/license"},version:{type:"string"}},required:["title","version"],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},contact:{$comment:"https://spec.openapis.org/oas/v3.1.0#contact-object",type:"object",properties:{name:{type:"string"},url:{type:"string",format:"uri-reference"},email:{type:"string",format:"email"}},$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},license:{$comment:"https://spec.openapis.org/oas/v3.1.0#license-object",type:"object",properties:{name:{type:"string"},identifier:{type:"string"},url:{type:"string",format:"uri-reference"}},required:["name"],dependentSchemas:{identifier:{not:{required:["url"]}}},$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},server:{$comment:"https://spec.openapis.org/oas/v3.1.0#server-object",type:"object",properties:{url:{type:"string"},description:{type:"string"},variables:{type:"object",additionalProperties:{$ref:"#/$defs/server-variable"}}},required:["url"],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},"server-variable":{$comment:"https://spec.openapis.org/oas/v3.1.0#server-variable-object",type:"object",properties:{enum:{type:"array",items:{type:"string"},minItems:1},default:{type:"string"},description:{type:"string"}},required:["default"],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},components:{$comment:"https://spec.openapis.org/oas/v3.1.0#components-object",type:"object",properties:{schemas:{type:"object",additionalProperties:{$ref:"#/$defs/schema"}},responses:{type:"object",additionalProperties:{$ref:"#/$defs/response-or-reference"}},parameters:{type:"object",additionalProperties:{$ref:"#/$defs/parameter-or-reference"}},examples:{type:"object",additionalProperties:{$ref:"#/$defs/example-or-reference"}},requestBodies:{type:"object",additionalProperties:{$ref:"#/$defs/request-body-or-reference"}},headers:{type:"object",additionalProperties:{$ref:"#/$defs/header-or-reference"}},securitySchemes:{type:"object",additionalProperties:{$ref:"#/$defs/security-scheme-or-reference"}},links:{type:"object",additionalProperties:{$ref:"#/$defs/link-or-reference"}},callbacks:{type:"object",additionalProperties:{$ref:"#/$defs/callbacks-or-reference"}},pathItems:{type:"object",additionalProperties:{$ref:"#/$defs/path-item-or-reference"}}},patternProperties:{"^(schemas|responses|parameters|examples|requestBodies|headers|securitySchemes|links|callbacks|pathItems)$":{$comment:"Enumerating all of the property names in the regex above is necessary for unevaluatedProperties to work as expected",propertyNames:{pattern:"^[a-zA-Z0-9._-]+$"}}},$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},paths:{$comment:"https://spec.openapis.org/oas/v3.1.0#paths-object",type:"object",patternProperties:{"^/":{$ref:"#/$defs/path-item"}},$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},"path-item":{$comment:"https://spec.openapis.org/oas/v3.1.0#path-item-object",type:"object",properties:{summary:{type:"string"},description:{type:"string"},servers:{type:"array",items:{$ref:"#/$defs/server"}},parameters:{type:"array",items:{$ref:"#/$defs/parameter-or-reference"}},get:{$ref:"#/$defs/operation"},put:{$ref:"#/$defs/operation"},post:{$ref:"#/$defs/operation"},delete:{$ref:"#/$defs/operation"},options:{$ref:"#/$defs/operation"},head:{$ref:"#/$defs/operation"},patch:{$ref:"#/$defs/operation"},trace:{$ref:"#/$defs/operation"}},$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},"path-item-or-reference":{if:{type:"object",required:["$ref"]},then:{$ref:"#/$defs/reference"},else:{$ref:"#/$defs/path-item"}},operation:{$comment:"https://spec.openapis.org/oas/v3.1.0#operation-object",type:"object",properties:{tags:{type:"array",items:{type:"string"}},summary:{type:"string"},description:{type:"string"},externalDocs:{$ref:"#/$defs/external-documentation"},operationId:{type:"string"},parameters:{type:"array",items:{$ref:"#/$defs/parameter-or-reference"}},requestBody:{$ref:"#/$defs/request-body-or-reference"},responses:{$ref:"#/$defs/responses"},callbacks:{type:"object",additionalProperties:{$ref:"#/$defs/callbacks-or-reference"}},deprecated:{default:!1,type:"boolean"},security:{type:"array",items:{$ref:"#/$defs/security-requirement"}},servers:{type:"array",items:{$ref:"#/$defs/server"}}},$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},"external-documentation":{$comment:"https://spec.openapis.org/oas/v3.1.0#external-documentation-object",type:"object",properties:{description:{type:"string"},url:{type:"string",format:"uri-reference"}},required:["url"],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},parameter:{$comment:"https://spec.openapis.org/oas/v3.1.0#parameter-object",type:"object",properties:{name:{type:"string"},in:{enum:["query","header","path","cookie"]},description:{type:"string"},required:{default:!1,type:"boolean"},deprecated:{default:!1,type:"boolean"},schema:{$ref:"#/$defs/schema"},content:{$ref:"#/$defs/content",minProperties:1,maxProperties:1}},required:["name","in"],oneOf:[{required:["schema"]},{required:["content"]}],if:{properties:{in:{const:"query"}},required:["in"]},then:{properties:{allowEmptyValue:{default:!1,type:"boolean"}}},dependentSchemas:{schema:{properties:{style:{type:"string"},explode:{type:"boolean"}},allOf:[{$ref:"#/$defs/examples"},{$ref:"#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-path"},{$ref:"#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-header"},{$ref:"#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-query"},{$ref:"#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-cookie"},{$ref:"#/$defs/parameter/dependentSchemas/schema/$defs/styles-for-form"}],$defs:{"styles-for-path":{if:{properties:{in:{const:"path"}},required:["in"]},then:{properties:{name:{pattern:"[^/#?]+$"},style:{default:"simple",enum:["matrix","label","simple"]},required:{const:!0}},required:["required"]}},"styles-for-header":{if:{properties:{in:{const:"header"}},required:["in"]},then:{properties:{style:{default:"simple",const:"simple"}}}},"styles-for-query":{if:{properties:{in:{const:"query"}},required:["in"]},then:{properties:{style:{default:"form",enum:["form","spaceDelimited","pipeDelimited","deepObject"]},allowReserved:{default:!1,type:"boolean"}}}},"styles-for-cookie":{if:{properties:{in:{const:"cookie"}},required:["in"]},then:{properties:{style:{default:"form",const:"form"}}}},"styles-for-form":{if:{properties:{style:{const:"form"}},required:["style"]},then:{properties:{explode:{default:!0}}},else:{properties:{explode:{default:!1}}}}}}},$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},"parameter-or-reference":{if:{type:"object",required:["$ref"]},then:{$ref:"#/$defs/reference"},else:{$ref:"#/$defs/parameter"}},"request-body":{$comment:"https://spec.openapis.org/oas/v3.1.0#request-body-object",type:"object",properties:{description:{type:"string"},content:{$ref:"#/$defs/content"},required:{default:!1,type:"boolean"}},required:["content"],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},"request-body-or-reference":{if:{type:"object",required:["$ref"]},then:{$ref:"#/$defs/reference"},else:{$ref:"#/$defs/request-body"}},content:{$comment:"https://spec.openapis.org/oas/v3.1.0#fixed-fields-10",type:"object",additionalProperties:{$ref:"#/$defs/media-type"},propertyNames:{format:"media-range"}},"media-type":{$comment:"https://spec.openapis.org/oas/v3.1.0#media-type-object",type:"object",properties:{schema:{$ref:"#/$defs/schema"},encoding:{type:"object",additionalProperties:{$ref:"#/$defs/encoding"}}},allOf:[{$ref:"#/$defs/specification-extensions"},{$ref:"#/$defs/examples"}],unevaluatedProperties:!1},encoding:{$comment:"https://spec.openapis.org/oas/v3.1.0#encoding-object",type:"object",properties:{contentType:{type:"string",format:"media-range"},headers:{type:"object",additionalProperties:{$ref:"#/$defs/header-or-reference"}},style:{default:"form",enum:["form","spaceDelimited","pipeDelimited","deepObject"]},explode:{type:"boolean"},allowReserved:{default:!1,type:"boolean"}},allOf:[{$ref:"#/$defs/specification-extensions"},{$ref:"#/$defs/encoding/$defs/explode-default"}],unevaluatedProperties:!1,$defs:{"explode-default":{if:{properties:{style:{const:"form"}},required:["style"]},then:{properties:{explode:{default:!0}}},else:{properties:{explode:{default:!1}}}}}},responses:{$comment:"https://spec.openapis.org/oas/v3.1.0#responses-object",type:"object",properties:{default:{$ref:"#/$defs/response-or-reference"}},patternProperties:{"^[1-5](?:[0-9]{2}|XX)$":{$ref:"#/$defs/response-or-reference"}},minProperties:1,$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},response:{$comment:"https://spec.openapis.org/oas/v3.1.0#response-object",type:"object",properties:{description:{type:"string"},headers:{type:"object",additionalProperties:{$ref:"#/$defs/header-or-reference"}},content:{$ref:"#/$defs/content"},links:{type:"object",additionalProperties:{$ref:"#/$defs/link-or-reference"}}},required:["description"],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},"response-or-reference":{if:{type:"object",required:["$ref"]},then:{$ref:"#/$defs/reference"},else:{$ref:"#/$defs/response"}},callbacks:{$comment:"https://spec.openapis.org/oas/v3.1.0#callback-object",type:"object",$ref:"#/$defs/specification-extensions",additionalProperties:{$ref:"#/$defs/path-item-or-reference"}},"callbacks-or-reference":{if:{type:"object",required:["$ref"]},then:{$ref:"#/$defs/reference"},else:{$ref:"#/$defs/callbacks"}},example:{$comment:"https://spec.openapis.org/oas/v3.1.0#example-object",type:"object",properties:{summary:{type:"string"},description:{type:"string"},value:!0,externalValue:{type:"string",format:"uri-reference"}},not:{required:["value","externalValue"]},$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},"example-or-reference":{if:{type:"object",required:["$ref"]},then:{$ref:"#/$defs/reference"},else:{$ref:"#/$defs/example"}},link:{$comment:"https://spec.openapis.org/oas/v3.1.0#link-object",type:"object",properties:{operationRef:{type:"string",format:"uri-reference"},operationId:{type:"string"},parameters:{$ref:"#/$defs/map-of-strings"},requestBody:!0,description:{type:"string"},body:{$ref:"#/$defs/server"}},oneOf:[{required:["operationRef"]},{required:["operationId"]}],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},"link-or-reference":{if:{type:"object",required:["$ref"]},then:{$ref:"#/$defs/reference"},else:{$ref:"#/$defs/link"}},header:{$comment:"https://spec.openapis.org/oas/v3.1.0#header-object",type:"object",properties:{description:{type:"string"},required:{default:!1,type:"boolean"},deprecated:{default:!1,type:"boolean"},schema:{$ref:"#/$defs/schema"},content:{$ref:"#/$defs/content",minProperties:1,maxProperties:1}},oneOf:[{required:["schema"]},{required:["content"]}],dependentSchemas:{schema:{properties:{style:{default:"simple",const:"simple"},explode:{default:!1,type:"boolean"}},$ref:"#/$defs/examples"}},$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},"header-or-reference":{if:{type:"object",required:["$ref"]},then:{$ref:"#/$defs/reference"},else:{$ref:"#/$defs/header"}},tag:{$comment:"https://spec.openapis.org/oas/v3.1.0#tag-object",type:"object",properties:{name:{type:"string"},description:{type:"string"},externalDocs:{$ref:"#/$defs/external-documentation"}},required:["name"],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},reference:{$comment:"https://spec.openapis.org/oas/v3.1.0#reference-object",type:"object",properties:{$ref:{type:"string",format:"uri-reference"},summary:{type:"string"},description:{type:"string"}},unevaluatedProperties:!1},schema:{$comment:"https://spec.openapis.org/oas/v3.1.0#schema-object",$dynamicAnchor:"meta",type:["object","boolean"]},"security-scheme":{$comment:"https://spec.openapis.org/oas/v3.1.0#security-scheme-object",type:"object",properties:{type:{enum:["apiKey","http","mutualTLS","oauth2","openIdConnect"]},description:{type:"string"}},required:["type"],allOf:[{$ref:"#/$defs/specification-extensions"},{$ref:"#/$defs/security-scheme/$defs/type-apikey"},{$ref:"#/$defs/security-scheme/$defs/type-http"},{$ref:"#/$defs/security-scheme/$defs/type-http-bearer"},{$ref:"#/$defs/security-scheme/$defs/type-oauth2"},{$ref:"#/$defs/security-scheme/$defs/type-oidc"}],unevaluatedProperties:!1,$defs:{"type-apikey":{if:{properties:{type:{const:"apiKey"}},required:["type"]},then:{properties:{name:{type:"string"},in:{enum:["query","header","cookie"]}},required:["name","in"]}},"type-http":{if:{properties:{type:{const:"http"}},required:["type"]},then:{properties:{scheme:{type:"string"}},required:["scheme"]}},"type-http-bearer":{if:{properties:{type:{const:"http"},scheme:{type:"string",pattern:"^[Bb][Ee][Aa][Rr][Ee][Rr]$"}},required:["type","scheme"]},then:{properties:{bearerFormat:{type:"string"}}}},"type-oauth2":{if:{properties:{type:{const:"oauth2"}},required:["type"]},then:{properties:{flows:{$ref:"#/$defs/oauth-flows"}},required:["flows"]}},"type-oidc":{if:{properties:{type:{const:"openIdConnect"}},required:["type"]},then:{properties:{openIdConnectUrl:{type:"string",format:"uri-reference"}},required:["openIdConnectUrl"]}}}},"security-scheme-or-reference":{if:{type:"object",required:["$ref"]},then:{$ref:"#/$defs/reference"},else:{$ref:"#/$defs/security-scheme"}},"oauth-flows":{type:"object",properties:{implicit:{$ref:"#/$defs/oauth-flows/$defs/implicit"},password:{$ref:"#/$defs/oauth-flows/$defs/password"},clientCredentials:{$ref:"#/$defs/oauth-flows/$defs/client-credentials"},authorizationCode:{$ref:"#/$defs/oauth-flows/$defs/authorization-code"}},$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1,$defs:{implicit:{type:"object",properties:{authorizationUrl:{type:"string",format:"uri-reference"},refreshUrl:{type:"string",format:"uri-reference"},scopes:{$ref:"#/$defs/map-of-strings"}},required:["authorizationUrl","scopes"],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},password:{type:"object",properties:{tokenUrl:{type:"string",format:"uri-reference"},refreshUrl:{type:"string",format:"uri-reference"},scopes:{$ref:"#/$defs/map-of-strings"}},required:["tokenUrl","scopes"],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},"client-credentials":{type:"object",properties:{tokenUrl:{type:"string",format:"uri-reference"},refreshUrl:{type:"string",format:"uri-reference"},scopes:{$ref:"#/$defs/map-of-strings"}},required:["tokenUrl","scopes"],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1},"authorization-code":{type:"object",properties:{authorizationUrl:{type:"string",format:"uri-reference"},tokenUrl:{type:"string",format:"uri-reference"},refreshUrl:{type:"string",format:"uri-reference"},scopes:{$ref:"#/$defs/map-of-strings"}},required:["authorizationUrl","tokenUrl","scopes"],$ref:"#/$defs/specification-extensions",unevaluatedProperties:!1}}},"security-requirement":{$comment:"https://spec.openapis.org/oas/v3.1.0#security-requirement-object",type:"object",additionalProperties:{type:"array",items:{type:"string"}}},"specification-extensions":{$comment:"https://spec.openapis.org/oas/v3.1.0#specification-extensions",patternProperties:{"^x-":!0}},examples:{properties:{example:!0,examples:{type:"object",additionalProperties:{$ref:"#/$defs/example-or-reference"}}}},"map-of-strings":{type:"object",additionalProperties:{type:"string"}}}}}),a={EMPTY_OR_INVALID:"Can’t find JSON, YAML or filename in data",OPENAPI_VERSION_NOT_SUPPORTED:"Can’t find supported Swagger/OpenAPI version in specification, version must be a string.",INVALID_REFERENCE:"Can’t resolve reference: %s",EXTERNAL_REFERENCE_NOT_FOUND:"Can’t resolve external reference: %s",FILE_DOES_NOT_EXIST:"File does not exist: %s",NO_CONTENT:"No content found"}},3751:(e,t,n)=>{"use strict";n.d(t,{$9:()=>M,D$:()=>de,Ef:()=>be,Jo:()=>te,XL:()=>ae,XX:()=>ve,aG:()=>R,eB:()=>g,jR:()=>he,lH:()=>ne,u1:()=>oe});var r=n(641),a=n(33);let o;n(953);const i="undefined"!=typeof window&&window.trustedTypes;if(i)try{o=i.createPolicy("vue",{createHTML:e=>e})}catch(e){}const s=o?e=>o.createHTML(e):e=>e,l="undefined"!=typeof document?document:null,c=l&&l.createElement("template"),u={insert:(e,t,n)=>{t.insertBefore(e,n||null)},remove:e=>{const t=e.parentNode;t&&t.removeChild(e)},createElement:(e,t,n,r)=>{const a="svg"===t?l.createElementNS("http://www.w3.org/2000/svg",e):"mathml"===t?l.createElementNS("http://www.w3.org/1998/Math/MathML",e):n?l.createElement(e,{is:n}):l.createElement(e);return"select"===e&&r&&null!=r.multiple&&a.setAttribute("multiple",r.multiple),a},createText:e=>l.createTextNode(e),createComment:e=>l.createComment(e),setText:(e,t)=>{e.nodeValue=t},setElementText:(e,t)=>{e.textContent=t},parentNode:e=>e.parentNode,nextSibling:e=>e.nextSibling,querySelector:e=>l.querySelector(e),setScopeId(e,t){e.setAttribute(t,"")},insertStaticContent(e,t,n,r,a,o){const i=n?n.previousSibling:t.lastChild;if(a&&(a===o||a.nextSibling))for(;t.insertBefore(a.cloneNode(!0),n),a!==o&&(a=a.nextSibling););else{c.innerHTML=s("svg"===r?`<svg>${e}</svg>`:"mathml"===r?`<math>${e}</math>`:e);const a=c.content;if("svg"===r||"mathml"===r){const e=a.firstChild;for(;e.firstChild;)a.appendChild(e.firstChild);a.removeChild(e)}t.insertBefore(a,n)}return[i?i.nextSibling:t.firstChild,n?n.previousSibling:t.lastChild]}},d="transition",p="animation",h=Symbol("_vtc"),f={name:String,type:String,css:{type:Boolean,default:!0},duration:[String,Number,Object],enterFromClass:String,enterActiveClass:String,enterToClass:String,appearFromClass:String,appearActiveClass:String,appearToClass:String,leaveFromClass:String,leaveActiveClass:String,leaveToClass:String},m=(0,a.X$)({},r.QP,f),g=(e=>(e.displayName="Transition",e.props=m,e))(((e,{slots:t})=>(0,r.h)(r.pR,function(e){const t={};for(const n in e)n in f||(t[n]=e[n]);if(!1===e.css)return t;const{name:n="v",type:r,duration:o,enterFromClass:i=`${n}-enter-from`,enterActiveClass:s=`${n}-enter-active`,enterToClass:l=`${n}-enter-to`,appearFromClass:c=i,appearActiveClass:u=s,appearToClass:d=l,leaveFromClass:p=`${n}-leave-from`,leaveActiveClass:h=`${n}-leave-active`,leaveToClass:m=`${n}-leave-to`}=e,g=function(e){if(null==e)return null;if((0,a.Gv)(e))return[y(e.enter),y(e.leave)];{const t=y(e);return[t,t]}}(o),x=g&&g[0],S=g&&g[1],{onBeforeEnter:_,onEnter:T,onEnterCancelled:A,onLeave:R,onLeaveCancelled:P,onBeforeAppear:D=_,onAppear:M=T,onAppearCancelled:$=A}=t,N=(e,t,n,r)=>{e._enterCancelled=r,O(e,t?d:l),O(e,t?u:s),n&&n()},I=(e,t)=>{e._isLeaving=!1,O(e,p),O(e,m),O(e,h),t&&t()},L=e=>(t,n)=>{const a=e?M:T,o=()=>N(t,e,n);v(a,[t,o]),k((()=>{O(t,e?c:i),w(t,e?d:l),b(a)||E(t,r,x,o)}))};return(0,a.X$)(t,{onBeforeEnter(e){v(_,[e]),w(e,i),w(e,s)},onBeforeAppear(e){v(D,[e]),w(e,c),w(e,u)},onEnter:L(!1),onAppear:L(!0),onLeave(e,t){e._isLeaving=!0;const n=()=>I(e,t);w(e,p),e._enterCancelled?(w(e,h),C()):(C(),w(e,h)),k((()=>{e._isLeaving&&(O(e,p),w(e,m),b(R)||E(e,r,S,n))})),v(R,[e,n])},onEnterCancelled(e){N(e,!1,void 0,!0),v(A,[e])},onAppearCancelled(e){N(e,!0,void 0,!0),v($,[e])},onLeaveCancelled(e){I(e),v(P,[e])}})}(e),t))),v=(e,t=[])=>{(0,a.cy)(e)?e.forEach((e=>e(...t))):e&&e(...t)},b=e=>!!e&&((0,a.cy)(e)?e.some((e=>e.length>1)):e.length>1);function y(e){return(0,a.Ro)(e)}function w(e,t){t.split(/\s+/).forEach((t=>t&&e.classList.add(t))),(e[h]||(e[h]=new Set)).add(t)}function O(e,t){t.split(/\s+/).forEach((t=>t&&e.classList.remove(t)));const n=e[h];n&&(n.delete(t),n.size||(e[h]=void 0))}function k(e){requestAnimationFrame((()=>{requestAnimationFrame(e)}))}let x=0;function E(e,t,n,r){const a=e._endId=++x,o=()=>{a===e._endId&&r()};if(null!=n)return setTimeout(o,n);const{type:i,timeout:s,propCount:l}=function(e,t){const n=window.getComputedStyle(e),r=e=>(n[e]||"").split(", "),a=r(`${d}Delay`),o=r(`${d}Duration`),i=S(a,o),s=r(`${p}Delay`),l=r(`${p}Duration`),c=S(s,l);let u=null,h=0,f=0;return t===d?i>0&&(u=d,h=i,f=o.length):t===p?c>0&&(u=p,h=c,f=l.length):(h=Math.max(i,c),u=h>0?i>c?d:p:null,f=u?u===d?o.length:l.length:0),{type:u,timeout:h,propCount:f,hasTransform:u===d&&/\b(transform|all)(,|$)/.test(r(`${d}Property`).toString())}}(e,t);if(!i)return r();const c=i+"end";let u=0;const h=()=>{e.removeEventListener(c,f),o()},f=t=>{t.target===e&&++u>=l&&h()};setTimeout((()=>{u<l&&h()}),s+1),e.addEventListener(c,f)}function S(e,t){for(;e.length<t.length;)e=e.concat(e);return Math.max(...t.map(((t,n)=>_(t)+_(e[n]))))}function _(e){return"auto"===e?0:1e3*Number(e.slice(0,-1).replace(",","."))}function C(){return document.body.offsetHeight}const T=Symbol("_vod"),A=Symbol("_vsh"),R={beforeMount(e,{value:t},{transition:n}){e[T]="none"===e.style.display?"":e.style.display,n&&t?n.beforeEnter(e):P(e,t)},mounted(e,{value:t},{transition:n}){n&&t&&n.enter(e)},updated(e,{value:t,oldValue:n},{transition:r}){!t!=!n&&(r?t?(r.beforeEnter(e),P(e,!0),r.enter(e)):r.leave(e,(()=>{P(e,!1)})):P(e,t))},beforeUnmount(e,{value:t}){P(e,t)}};function P(e,t){e.style.display=t?e[T]:"none",e[A]=!t}const D=Symbol("");function M(e){const t=(0,r.nI)();if(!t)return;const n=t.ut=(n=e(t.proxy))=>{Array.from(document.querySelectorAll(`[data-v-owner="${t.uid}"]`)).forEach((e=>N(e,n)))},o=()=>{const r=e(t.proxy);t.ce?N(t.ce,r):$(t.subTree,r),n(r)};(0,r.Ic)((()=>{(0,r.Dl)(o)})),(0,r.sV)((()=>{(0,r.wB)(o,a.tE,{flush:"post"});const e=new MutationObserver(o);e.observe(t.subTree.el.parentNode,{childList:!0}),(0,r.hi)((()=>e.disconnect()))}))}function $(e,t){if(128&e.shapeFlag){const n=e.suspense;e=n.activeBranch,n.pendingBranch&&!n.isHydrating&&n.effects.push((()=>{$(n.activeBranch,t)}))}for(;e.component;)e=e.component.subTree;if(1&e.shapeFlag&&e.el)N(e.el,t);else if(e.type===r.FK)e.children.forEach((e=>$(e,t)));else if(e.type===r.jC){let{el:n,anchor:r}=e;for(;n&&(N(n,t),n!==r);)n=n.nextSibling}}function N(e,t){if(1===e.nodeType){const n=e.style;let r="";for(const e in t)n.setProperty(`--${e}`,t[e]),r+=`--${e}: ${t[e]};`;n[D]=r}}const I=/(^|;)\s*display\s*:/,L=/\s*!important$/;function B(e,t,n){if((0,a.cy)(n))n.forEach((n=>B(e,t,n)));else if(null==n&&(n=""),t.startsWith("--"))e.setProperty(t,n);else{const r=function(e,t){const n=Q[t];if(n)return n;let r=(0,a.PT)(t);if("filter"!==r&&r in e)return Q[t]=r;r=(0,a.ZH)(r);for(let n=0;n<F.length;n++){const a=F[n]+r;if(a in e)return Q[t]=a}return t}(e,t);L.test(n)?e.setProperty((0,a.Tg)(r),n.replace(L,""),"important"):e[r]=n}}const F=["Webkit","Moz","ms"],Q={},U="http://www.w3.org/1999/xlink";function z(e,t,n,r,o,i=(0,a.J$)(t)){r&&t.startsWith("xlink:")?null==n?e.removeAttributeNS(U,t.slice(6,t.length)):e.setAttributeNS(U,t,n):null==n||i&&!(0,a.Y2)(n)?e.removeAttribute(t):e.setAttribute(t,i?"":(0,a.Bm)(n)?String(n):n)}function j(e,t,n,r,o){if("innerHTML"===t||"textContent"===t)return void(null!=n&&(e[t]="innerHTML"===t?s(n):n));const i=e.tagName;if("value"===t&&"PROGRESS"!==i&&!i.includes("-")){const r="OPTION"===i?e.getAttribute("value")||"":e.value,a=null==n?"checkbox"===e.type?"on":"":String(n);return r===a&&"_value"in e||(e.value=a),null==n&&e.removeAttribute(t),void(e._value=n)}let l=!1;if(""===n||null==n){const r=typeof e[t];"boolean"===r?n=(0,a.Y2)(n):null==n&&"string"===r?(n="",l=!0):"number"===r&&(n=0,l=!0)}try{e[t]=n}catch(e){}l&&e.removeAttribute(o||t)}function W(e,t,n,r){e.addEventListener(t,n,r)}const X=Symbol("_vei");const q=/(?:Once|Passive|Capture)$/;let V=0;const H=Promise.resolve(),Z=()=>V||(H.then((()=>V=0)),V=Date.now()),G=e=>111===e.charCodeAt(0)&&110===e.charCodeAt(1)&&e.charCodeAt(2)>96&&e.charCodeAt(2)<123;"undefined"!=typeof HTMLElement&&HTMLElement,Symbol("_moveCb"),Symbol("_enterCb");const K=e=>{const t=e.props["onUpdate:modelValue"]||!1;return(0,a.cy)(t)?e=>(0,a.DY)(t,e):t};function Y(e){e.target.composing=!0}function J(e){const t=e.target;t.composing&&(t.composing=!1,t.dispatchEvent(new Event("input")))}const ee=Symbol("_assign"),te={created(e,{modifiers:{lazy:t,trim:n,number:r}},o){e[ee]=K(o);const i=r||o.props&&"number"===o.props.type;W(e,t?"change":"input",(t=>{if(t.target.composing)return;let r=e.value;n&&(r=r.trim()),i&&(r=(0,a.bB)(r)),e[ee](r)})),n&&W(e,"change",(()=>{e.value=e.value.trim()})),t||(W(e,"compositionstart",Y),W(e,"compositionend",J),W(e,"change",J))},mounted(e,{value:t}){e.value=null==t?"":t},beforeUpdate(e,{value:t,oldValue:n,modifiers:{lazy:r,trim:o,number:i}},s){if(e[ee]=K(s),e.composing)return;const l=null==t?"":t;if((!i&&"number"!==e.type||/^0\d/.test(e.value)?e.value:(0,a.bB)(e.value))!==l){if(document.activeElement===e&&"range"!==e.type){if(r&&t===n)return;if(o&&e.value.trim()===l)return}e.value=l}}},ne={deep:!0,created(e,t,n){e[ee]=K(n),W(e,"change",(()=>{const t=e._modelValue,n=se(e),r=e.checked,o=e[ee];if((0,a.cy)(t)){const e=(0,a.u3)(t,n),i=-1!==e;if(r&&!i)o(t.concat(n));else if(!r&&i){const n=[...t];n.splice(e,1),o(n)}}else if((0,a.vM)(t)){const e=new Set(t);r?e.add(n):e.delete(n),o(e)}else o(le(e,r))}))},mounted:re,beforeUpdate(e,t,n){e[ee]=K(n),re(e,t,n)}};function re(e,{value:t,oldValue:n},r){let o;if(e._modelValue=t,(0,a.cy)(t))o=(0,a.u3)(t,r.props.value)>-1;else if((0,a.vM)(t))o=t.has(r.props.value);else{if(t===n)return;o=(0,a.BX)(t,le(e,!0))}e.checked!==o&&(e.checked=o)}const ae={created(e,{value:t},n){e.checked=(0,a.BX)(t,n.props.value),e[ee]=K(n),W(e,"change",(()=>{e[ee](se(e))}))},beforeUpdate(e,{value:t,oldValue:n},r){e[ee]=K(r),t!==n&&(e.checked=(0,a.BX)(t,r.props.value))}},oe={deep:!0,created(e,{value:t,modifiers:{number:n}},o){const i=(0,a.vM)(t);W(e,"change",(()=>{const t=Array.prototype.filter.call(e.options,(e=>e.selected)).map((e=>n?(0,a.bB)(se(e)):se(e)));e[ee](e.multiple?i?new Set(t):t:t[0]),e._assigning=!0,(0,r.dY)((()=>{e._assigning=!1}))})),e[ee]=K(o)},mounted(e,{value:t}){ie(e,t)},beforeUpdate(e,t,n){e[ee]=K(n)},updated(e,{value:t}){e._assigning||ie(e,t)}};function ie(e,t){const n=e.multiple,r=(0,a.cy)(t);if(!n||r||(0,a.vM)(t)){for(let o=0,i=e.options.length;o<i;o++){const i=e.options[o],s=se(i);if(n)if(r){const e=typeof s;i.selected="string"===e||"number"===e?t.some((e=>String(e)===String(s))):(0,a.u3)(t,s)>-1}else i.selected=t.has(s);else if((0,a.BX)(se(i),t))return void(e.selectedIndex!==o&&(e.selectedIndex=o))}n||-1===e.selectedIndex||(e.selectedIndex=-1)}}function se(e){return"_value"in e?e._value:e.value}function le(e,t){const n=t?"_trueValue":"_falseValue";return n in e?e[n]:t}const ce=["ctrl","shift","alt","meta"],ue={stop:e=>e.stopPropagation(),prevent:e=>e.preventDefault(),self:e=>e.target!==e.currentTarget,ctrl:e=>!e.ctrlKey,shift:e=>!e.shiftKey,alt:e=>!e.altKey,meta:e=>!e.metaKey,left:e=>"button"in e&&0!==e.button,middle:e=>"button"in e&&1!==e.button,right:e=>"button"in e&&2!==e.button,exact:(e,t)=>ce.some((n=>e[`${n}Key`]&&!t.includes(n)))},de=(e,t)=>{const n=e._withMods||(e._withMods={}),r=t.join(".");return n[r]||(n[r]=(n,...r)=>{for(let e=0;e<t.length;e++){const r=ue[t[e]];if(r&&r(n,t))return}return e(n,...r)})},pe={esc:"escape",space:" ",up:"arrow-up",left:"arrow-left",right:"arrow-right",down:"arrow-down",delete:"backspace"},he=(e,t)=>{const n=e._withKeys||(e._withKeys={}),r=t.join(".");return n[r]||(n[r]=n=>{if(!("key"in n))return;const r=(0,a.Tg)(n.key);return t.some((e=>e===r||pe[e]===r))?e(n):void 0})},fe=(0,a.X$)({patchProp:(e,t,n,o,i,s)=>{const l="svg"===i;"class"===t?function(e,t,n){const r=e[h];r&&(t=(t?[t,...r]:[...r]).join(" ")),null==t?e.removeAttribute("class"):n?e.setAttribute("class",t):e.className=t}(e,o,l):"style"===t?function(e,t,n){const r=e.style,o=(0,a.Kg)(n);let i=!1;if(n&&!o){if(t)if((0,a.Kg)(t))for(const e of t.split(";")){const t=e.slice(0,e.indexOf(":")).trim();null==n[t]&&B(r,t,"")}else for(const e in t)null==n[e]&&B(r,e,"");for(const e in n)"display"===e&&(i=!0),B(r,e,n[e])}else if(o){if(t!==n){const e=r[D];e&&(n+=";"+e),r.cssText=n,i=I.test(n)}}else t&&e.removeAttribute("style");T in e&&(e[T]=i?r.display:"",e[A]&&(r.display="none"))}(e,n,o):(0,a.Mp)(t)?(0,a.CP)(t)||function(e,t,n,o,i=null){const s=e[X]||(e[X]={}),l=s[t];if(o&&l)l.value=o;else{const[n,c]=function(e){let t;if(q.test(e)){let n;for(t={};n=e.match(q);)e=e.slice(0,e.length-n[0].length),t[n[0].toLowerCase()]=!0}return[":"===e[2]?e.slice(3):(0,a.Tg)(e.slice(2)),t]}(t);if(o){const l=s[t]=function(e,t){const n=e=>{if(e._vts){if(e._vts<=n.attached)return}else e._vts=Date.now();(0,r.qL)(function(e,t){if((0,a.cy)(t)){const n=e.stopImmediatePropagation;return e.stopImmediatePropagation=()=>{n.call(e),e._stopped=!0},t.map((e=>t=>!t._stopped&&e&&e(t)))}return t}(e,n.value),t,5,[e])};return n.value=e,n.attached=Z(),n}(o,i);W(e,n,l,c)}else l&&(function(e,t,n,r){e.removeEventListener(t,n,r)}(e,n,l,c),s[t]=void 0)}}(e,t,0,o,s):("."===t[0]?(t=t.slice(1),1):"^"===t[0]?(t=t.slice(1),0):function(e,t,n,r){if(r)return"innerHTML"===t||"textContent"===t||!!(t in e&&G(t)&&(0,a.Tn)(n));if("spellcheck"===t||"draggable"===t||"translate"===t)return!1;if("form"===t)return!1;if("list"===t&&"INPUT"===e.tagName)return!1;if("type"===t&&"TEXTAREA"===e.tagName)return!1;if("width"===t||"height"===t){const t=e.tagName;if("IMG"===t||"VIDEO"===t||"CANVAS"===t||"SOURCE"===t)return!1}return(!G(t)||!(0,a.Kg)(n))&&t in e}(e,t,o,l))?(j(e,t,o),e.tagName.includes("-")||"value"!==t&&"checked"!==t&&"selected"!==t||z(e,t,o,l,0,"value"!==t)):!e._isVueCE||!/[A-Z]/.test(t)&&(0,a.Kg)(o)?("true-value"===t?e._trueValue=o:"false-value"===t&&(e._falseValue=o),z(e,t,o,l)):j(e,(0,a.PT)(t),o,0,t)}},u);let me;function ge(){return me||(me=(0,r.K9)(fe))}const ve=(...e)=>{ge().render(...e)},be=(...e)=>{const t=ge().createApp(...e),{mount:n}=t;return t.mount=e=>{const r=function(e){if((0,a.Kg)(e))return document.querySelector(e);return e}(e);if(!r)return;const o=t._component;(0,a.Tn)(o)||o.render||o.template||(o.template=r.innerHTML),1===r.nodeType&&(r.textContent="");const i=n(r,!1,function(e){return e instanceof SVGElement?"svg":"function"==typeof MathMLElement&&e instanceof MathMLElement?"mathml":void 0}(r));return r instanceof Element&&(r.removeAttribute("v-cloak"),r.setAttribute("data-v-app","")),i},t}},3762:(e,t,n)=>{"use strict";n.d(t,{Kb:()=>i,mg:()=>o,p3:()=>s});var r=n(2741);const a=["localhost","127.0.0.1","[::1]","0.0.0.0"];function o(e,t){if(!s(e,t))return t??"";const n=new URL(t);return n.href=e,n.searchParams.append("scalar_url",t),n.toString()}const i=e=>!r.Q.PROTOCOL.test(e)&&!/^[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+(\/|$)/.test(e);function s(e,t){return!(!e||!t||i(t)||function(e){try{const{hostname:t}=new URL(e);return a.includes(t)}catch{return!0}}(t))}},3806:(e,t,n)=>{"use strict";n.d(t,{i:()=>a});var r=n(8157);function a(e={}){const{notify:t=e=>n(e,"info")}=e,{toast:n}=(0,r.Yf)();return{copyToClipboard:async function(e){try{await navigator.clipboard.writeText(e),t("Copied to the clipboard")}catch(e){console.error(e.message),t("Failed to copy to clipboard")}}}}},3813:(e,t,n)=>{"use strict";n.d(t,{A:()=>u});var r=n(641),a=n(33),o=n(953),i=n(3179),s=n(1266),l=n(4048);const c={key:0,class:"text-c-3 px-4 text-sm min-h-16 justify-center flex items-center"},u=(0,r.pM)({__name:"CodeSnippet",props:{target:{},client:{},operation:{},server:{},example:{},securitySchemes:{default:()=>[]}},setup(e){const t=(0,r.EW)((()=>e.securitySchemes.flatMap((e=>"apiKey"===e.type?e.value:"http"===(null==e?void 0:e.type)?[e.token,e.password,btoa(`${e.username}:${e.password}`)]:"oauth2"===e.type?Object.values(e.flows).map((e=>null==e?void 0:e.token)).filter(l.O9):[])))),n=(0,r.EW)((()=>{const[t,n]=(0,i.y)(e.target,e.client,{operation:e.operation,example:e.example,server:e.server,securitySchemes:e.securitySchemes});return{error:t,payload:n}})),u=(0,r.EW)((()=>"shell"===e.target&&"curl"===e.client?"curl":e.target??"plaintext"));return(e,i)=>n.value.error?((0,r.uX)(),(0,r.CE)("div",c,(0,a.v_)(n.value.error.message),1)):n.value.payload?((0,r.uX)(),(0,r.Wv)((0,o.R1)(s.z5),{key:1,class:"w-full",content:n.value.payload,hideCredentials:t.value,lang:u.value,lineNumbers:""},null,8,["content","hideCredentials","lang"])):(0,r.Q3)("",!0)}})},3827:(e,t,n)=>{"use strict";function r(e){"function"==typeof queueMicrotask?queueMicrotask(e):Promise.resolve().then(e).catch((e=>setTimeout((()=>{throw e}))))}n.d(t,{_:()=>r})},3833:(e,t,n)=>{"use strict";n.d(t,{A:()=>u});var r=n(641),a=n(953),o=n(33),i=n(1266),s=n(8309);const l={class:"text-c-2 mb-4 text-sm leading-normal"},c={class:"flex justify-between gap-2"},u=(0,r.pM)({__name:"DeleteRequestAuthModal",props:{state:{},scheme:{}},emits:["close","delete"],setup(e,{emit:t}){const n=e,u=t,{securitySchemeMutators:d}=(0,s.hA)(),p=()=>{var e;null!=(e=n.scheme)&&e.id&&d.delete(n.scheme.id),u("delete")};return(e,t)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(i.uT),{size:"xxs",state:e.state,title:"Delete Security Scheme"},{default:(0,r.k6)((()=>{var n;return[(0,r.Lk)("p",l," This cannot be undone. You’re about to delete the "+(0,o.v_)(null==(n=e.scheme)?void 0:n.label)+" security scheme from the collection. ",1),(0,r.Lk)("div",c,[(0,r.bF)((0,a.R1)(i.RO),{class:"flex h-8 cursor-pointer items-center gap-1.5 px-3 shadow-none focus:outline-none",type:"button",variant:"outlined",onClick:t[0]||(t[0]=e=>u("close"))},{default:(0,r.k6)((()=>t[1]||(t[1]=[(0,r.eW)(" Cancel ")]))),_:1}),(0,r.bF)((0,a.R1)(i.RO),{class:"flex h-8 cursor-pointer items-center gap-1.5 px-3 shadow-none focus:outline-none",type:"submit",onClick:p},{default:(0,r.k6)((()=>{var t;return[(0,r.eW)(" Delete "+(0,o.v_)(null==(t=e.scheme)?void 0:t.label),1)]})),_:1})])]})),_:1},8,["state"]))}})},3858:(e,t,n)=>{"use strict";n.d(t,{A:()=>f});var r=n(641),a=n(953),o=n(33),i=n(3751),s=n(4325),l=n(8309),c=n(6584);const u={key:0,class:"min-h-12 xl:min-h-client-header flex items-center justify-between px-3 py-1.5 md:px-[18px] md:py-2.5 text-sm"},d={class:"font-medium m-0 text-sm whitespace-nowrap"},p={class:"bg-b-1 relative z-10 pt-0 md:px-2.5 md:pb-2.5 sticky bottom-0 w-[inherit] has-[.empty-sidebar-item]:border-t-1/2"},h=(0,r.pM)({__name:"Sidebar",props:{title:{},isSidebarOpen:{type:Boolean,default:!0}},setup(e){const t=e,{sidebarWidth:n,setSidebarWidth:h}=(0,l.hA)(),{layout:f}=(0,c.g)(),m=(0,a.KR)(!1),g=(0,a.KR)(null),{breakpoints:v}=(0,s.G)(),b=e=>{e.preventDefault();const t=e.clientX,r=parseInt(getComputedStyle(g.value).width||n.value,10),a=e=>{m.value=!0,document.body.classList.add("dragging");let n=r+e.clientX-t;n>420&&(n=420+.2*(n-420)),n<240&&(n=240),h(`${n}px`)},o=()=>{m.value=!1,document.body.classList.remove("dragging"),document.documentElement.removeEventListener("mousemove",a,!1),document.documentElement.removeEventListener("mouseup",o,!1),parseInt(n.value,10)>420?h("360px"):parseInt(n.value,10)<240&&h("240px")};document.documentElement.addEventListener("mousemove",a,!1),document.documentElement.addEventListener("mouseup",o,!1)};return(e,s)=>(0,r.bo)(((0,r.uX)(),(0,r.CE)("aside",{ref_key:"sidebarRef",ref:g,class:(0,o.C4)(["sidebar overflow-hidden relative flex flex-col flex-1 md:flex-none bg-b-1 md:border-b-0 md:border-r-1/2 min-w-full md:min-w-fit leading-3",{dragging:m.value}]),style:(0,o.Tr)({width:(0,a.R1)(v).lg?(0,a.R1)(n):"100%"})},[(0,r.RG)(e.$slots,"header",{},void 0,!0),"modal"!==(0,a.R1)(f)&&e.title?((0,r.uX)(),(0,r.CE)("div",u,[(0,r.Lk)("h2",d,(0,o.v_)(e.title),1),(0,a.R1)(v).lg?(0,r.Q3)("",!0):(0,r.RG)(e.$slots,"button",{key:0},void 0,!0)])):(0,r.Q3)("",!0),(0,r.Lk)("div",{class:(0,o.C4)(["custom-scroll sidebar-height pb-0 md:pb-[37px] w-[inherit]",{"sidebar-mask":"modal"!==(0,a.R1)(f)}])},[(0,r.RG)(e.$slots,"content",{},void 0,!0)],2),(0,a.R1)(v).lg?((0,r.uX)(),(0,r.CE)(r.FK,{key:1},[(0,r.Lk)("div",p,[(0,r.RG)(e.$slots,"button",{},void 0,!0)]),(0,r.Lk)("div",{class:"resizer",onMousedown:b},null,32)],64)):(0,r.Q3)("",!0)],6)),[[i.aG,t.isSidebarOpen]])}}),f=(0,n(3526).A)(h,[["__scopeId","data-v-5397f2a7"]])},3940:e=>{"use strict";e.exports=function(e){return e.map((function(e){return""===e?"''":e&&"object"==typeof e?e.op.replace(/(.)/g,"\\$1"):/["\s]/.test(e)&&!/'/.test(e)?"'"+e.replace(/(['\\])/g,"\\$1")+"'":/["'\s]/.test(e)?'"'+e.replace(/(["\\$`!])/g,"\\$1")+'"':String(e).replace(/([A-Za-z]:)?([#!"$&'()*,:;<=>?@[\\\]^`{|}])/g,"$1\\$2")})).join(" ")}},3966:(e,t,n)=>{"use strict";n.d(t,{A:()=>s});var r=n(641),a=n(3526);const o={},i={class:"gap-1/2 flex flex-col px-3 pb-[75px]"},s=(0,a.A)(o,[["render",function(e,t){return(0,r.uX)(),(0,r.CE)("ul",i,[(0,r.RG)(e.$slots,"default")])}]])},4048:(e,t,n)=>{"use strict";n.d(t,{nF:()=>v,QL:()=>p.Q,NT:()=>h.NT,Ik:()=>h.Ik,Dk:()=>o.Dk,n1:()=>i,nG:()=>x,tI:()=>d,nk:()=>l.nk,gw:()=>h.gw,Z7:()=>f,O9:()=>r.O,aO:()=>h.aO,Eo:()=>l.Eo,AY:()=>m,zR:()=>g,rP:()=>b,Hh:()=>o.Hh,LQ:()=>y.L,O1:()=>w,qX:()=>O.q,mg:()=>s.mg,TA:()=>k.T,pH:()=>a,p3:()=>s.p3,QJ:()=>E});var r=n(265);const a=e=>!0===e?.["x-internal"]||!0===e?.["x-scalar-ignore"];var o=n(5148);n(1660);const i=e=>{let t=0,n=0,r=0;if(!e?.length)return n;for(r=0;r<e.length;r++)t=e.charCodeAt(r),n=(n<<5)-n+t,n|=0;return n};var s=n(3762),l=n(8470);const c="https://api.scalar.com/request-proxy",u="https://proxy.scalar.com";async function d(e,t,n=!0){t===c&&(t=u);const r=await fetch(t?(0,s.mg)(t,e):e);if(200!==r.status)throw console.error(`[fetchSpecFromUrl] Failed to fetch the specification at ${e} (Status: ${r.status})`),t||console.warn(`[fetchSpecFromUrl] Tried to fetch the specification (url: ${e}) without a proxy. Are the CORS headers configured to allow cross-domain requests? https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS`),new Error(`Failed to fetch the specification (Status: ${r.status})`);return n?(0,l.nk)(await r.text()):await r.text()}var p=n(2741),h=n(4115);const f={100:{name:"Continue",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/100",color:"var(--scalar-color-blue)"},101:{name:"Switching Protocols",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/101",color:"var(--scalar-color-blue)"},102:{name:"Processing",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/102",color:"var(--scalar-color-blue)"},103:{name:"Early Hints",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/103",color:"var(--scalar-color-blue)"},200:{name:"OK",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/200",color:"var(--scalar-color-green)"},201:{name:"Created",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/201",color:"var(--scalar-color-green)"},202:{name:"Accepted",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/202",color:"var(--scalar-color-green)"},203:{name:"Non-Authoritative Information",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/203",color:"var(--scalar-color-green)"},204:{name:"No Content",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/204",color:"var(--scalar-color-green)"},205:{name:"Reset Content",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/205",color:"var(--scalar-color-green)"},206:{name:"Partial Content",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/206",color:"var(--scalar-color-green)"},207:{name:"Multi-Status",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/207",color:"var(--scalar-color-green)"},208:{name:"Already Reported",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/208",color:"var(--scalar-color-green)"},226:{name:"IM Used",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/226",color:"var(--scalar-color-green)"},300:{name:"Multiple Choices",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/300",color:"var(--scalar-color-yellow)"},301:{name:"Moved Permanently",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/301",color:"var(--scalar-color-yellow)"},302:{name:"Found",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/302",color:"var(--scalar-color-yellow)"},303:{name:"See Other",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/303",color:"var(--scalar-color-yellow)"},304:{name:"Not Modified",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/304",color:"var(--scalar-color-yellow)"},305:{name:"Use Proxy",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/305",color:"var(--scalar-color-yellow)"},306:{name:"(Unused)",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/306",color:"var(--scalar-color-yellow)"},307:{name:"Temporary Redirect",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/307",color:"var(--scalar-color-yellow)"},308:{name:"Permanent Redirect",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/308",color:"var(--scalar-color-yellow)"},400:{name:"Bad Request",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/400",color:"var(--scalar-color-red)"},401:{name:"Unauthorized",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/401",color:"var(--scalar-color-red)"},402:{name:"Payment Required",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/402",color:"var(--scalar-color-red)"},403:{name:"Forbidden",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/403",color:"var(--scalar-color-red)"},404:{name:"Not Found",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/404",color:"var(--scalar-color-red)"},405:{name:"Method Not Allowed",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/405",color:"var(--scalar-color-red)"},406:{name:"Not Acceptable",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/406",color:"var(--scalar-color-red)"},407:{name:"Proxy Authentication Required",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/407",color:"var(--scalar-color-red)"},408:{name:"Request Timeout",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/408",color:"var(--scalar-color-red)"},409:{name:"Conflict",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/409",color:"var(--scalar-color-red)"},410:{name:"Gone",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/410",color:"var(--scalar-color-red)"},411:{name:"Length Required",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/411",color:"var(--scalar-color-red)"},412:{name:"Precondition Failed",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/412",color:"var(--scalar-color-red)"},413:{name:"Content Too Large",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/413",color:"var(--scalar-color-red)"},414:{name:"URI Too Long",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/414",color:"var(--scalar-color-red)"},415:{name:"Unsupported Media Type",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/415",color:"var(--scalar-color-red)"},416:{name:"Range Not Satisfiable",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/416",color:"var(--scalar-color-red)"},417:{name:"Expectation Failed",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/417",color:"var(--scalar-color-red)"},418:{name:"I'm a teapot",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/418",color:"var(--scalar-color-red)"},421:{name:"Misdirected Request",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/421",color:"var(--scalar-color-red)"},422:{name:"Unprocessable Content",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/422",color:"var(--scalar-color-red)"},423:{name:"Locked",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/423",color:"var(--scalar-color-red)"},424:{name:"Failed Dependency",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/424",color:"var(--scalar-color-red)"},425:{name:"Too Early",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/425",color:"var(--scalar-color-red)"},426:{name:"Upgrade Required",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/426",color:"var(--scalar-color-red)"},428:{name:"Precondition Required",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/428",color:"var(--scalar-color-red)"},429:{name:"Too Many Requests",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/429",color:"var(--scalar-color-red)"},431:{name:"Request Header Fields Too Large",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/431",color:"var(--scalar-color-red)"},451:{name:"Unavailable For Legal Reasons",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/451",color:"var(--scalar-color-red)"},500:{name:"Internal Server Error",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/500",color:"var(--scalar-color-red)"},501:{name:"Not Implemented",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/501",color:"var(--scalar-color-red)"},502:{name:"Bad Gateway",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/502",color:"var(--scalar-color-red)"},503:{name:"Service Unavailable",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/503",color:"var(--scalar-color-red)"},504:{name:"Gateway Timeout",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/504",color:"var(--scalar-color-red)"},505:{name:"HTTP Version Not Supported",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/505",color:"var(--scalar-color-red)"},506:{name:"Variant Also Negotiates",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/506",color:"var(--scalar-color-red)"},507:{name:"Insufficient Storage",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/507",color:"var(--scalar-color-red)"},508:{name:"Loop Detected",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/508",color:"var(--scalar-color-red)"},510:{name:"Not Extended",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/510",color:"var(--scalar-color-red)"},511:{name:"Network Authentication Required",url:"https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/511",color:"var(--scalar-color-red)"}};function m(e){try{return Boolean(new URL(e))}catch{return!1}}const g=(e,t,n=" #")=>{if(!t(e))return e;const r=e.split(n),a=r.length>1?`${r.slice(0,-1).join()}${n}${Number(r.at(-1))+1}`:`${r.join()}${n}2`;return g(a,t,n)},v={COLLECTION:"collection",COOKIE:"cookie",ENVIRONMENT:"environment",REQUEST:"request",REQUEST_EXAMPLE:"requestExample",SECURITY_SCHEME:"securityScheme",SERVER:"server",TAG:"tag",WORKSPACE:"workspace"},b=(e,t)=>{if(!e||e.startsWith("http://")||e.startsWith("https://")||"undefined"==typeof window&&!t)return e;const n=t||window.location.href,r=n.split("?")[0]?.split("#")[0],a=r?.endsWith("/")?r:r?.substring(0,r?.lastIndexOf("/")+1);return new URL(e,a).toString()};var y=n(5868);function w(e,t){return Object.keys(e).forEach((n=>{Object.hasOwn(t,n)||delete e[n]})),Object.assign(e,t),e}var O=n(6764),k=n(5182);const x=()=>({}),E="undefined"!=typeof window?window.__SCALAR__??x():x()},4074:(e,t,n)=>{"use strict";n.d(t,{D:()=>d});var r=n(3687);function a(e){if(null===e)return{version:void 0,specificationType:void 0,specificationVersion:void 0};for(const t of new Set(r.IY)){const n="2.0"===t?"swagger":"openapi",r=e[n];if("string"==typeof r&&r.startsWith(t))return{version:t,specificationType:n,specificationVersion:r}}return{version:void 0,specificationType:void 0,specificationVersion:void 0}}var o=n(5362),i=n(1507),s=n(7427);function l(e,t,n,r=[]){const a=structuredClone(e),s=(0,i.Y)(a),l=(0,o.z)(s),u=n?.specification??l.specification;return c(u,s,n??l,new WeakSet,r,t),{valid:0===(r=r.filter(((e,t,n)=>t===n.findIndex((t=>t.message===e.message&&t.code===e.code))))).length,errors:r,schema:u}}function c(e,t,n,r,a,o){if(null!==e&&!r.has(e)){for(r.add(e);void 0!==e.$ref;){const r=u(e.$ref,o,n,t,i,a);if("object"!=typeof r||null===r)break;const s=e.$ref;delete e.$ref;for(const t of Object.keys(r))void 0===e[t]&&(e[t]=r[t]);s&&o?.onDereference?.({schema:e,ref:s})}for(const i of Object.values(e))"object"==typeof i&&null!==i&&c(i,t,n,r,a,o)}function i(e){return c(e.specification,t,e,r,a,o),e}}function u(e,t,n,a,o,i){if("string"!=typeof e){if(t?.throwOnError)throw new Error(r.Sr.INVALID_REFERENCE.replace("%s",e));return void i.push({code:"INVALID_REFERENCE",message:r.Sr.INVALID_REFERENCE.replace("%s",e)})}const[l,c]=e.split("#",2),d=l!==n.filename;if(l&&d){const e=a.find((e=>e.filename===l));if(!e){if(t?.throwOnError)throw new Error(r.Sr.EXTERNAL_REFERENCE_NOT_FOUND.replace("%s",l));return void i.push({code:"EXTERNAL_REFERENCE_NOT_FOUND",message:r.Sr.EXTERNAL_REFERENCE_NOT_FOUND.replace("%s",l)})}return void 0===c?e.specification:u(`#${c}`,t,o(e),a,o,i)}const p=function(e){return e.split("/").slice(1).map(s.K)}(c);try{return p.reduce(((e,t)=>e[t]),n.specification)}catch(n){if(t?.throwOnError)throw new Error(r.Sr.INVALID_REFERENCE.replace("%s",e));i.push({code:"INVALID_REFERENCE",message:r.Sr.INVALID_REFERENCE.replace("%s",e)})}}async function d(e,t){const n=(0,i.Y)(e),r=(0,o.z)(n),s=l(n,t);return{specification:r.specification,errors:s.errors,schema:s.schema,...a(r.specification)}}},4115:(e,t,n)=>{"use strict";n.d(t,{Ik:()=>i,NT:()=>a,aO:()=>l,gw:()=>s});var r=n(7297);const a={get:{short:"GET",color:"text-blue",backgroundColor:"bg-blue"},post:{short:"POST",color:"text-green",backgroundColor:"bg-green"},put:{short:"PUT",color:"text-orange",backgroundColor:"bg-orange"},patch:{short:"PATCH",color:"text-yellow",backgroundColor:"bg-yellow"},delete:{short:"DEL",color:"text-red",backgroundColor:"bg-red"},options:{short:"OPTS",color:"text-purple",backgroundColor:"bg-purple"},head:{short:"HEAD",color:"text-scalar-c-2",backgroundColor:"bg-c-2"},connect:{short:"CONN",color:"text-c-2",backgroundColor:"bg-c-2"},trace:{short:"TRACE",color:"text-c-2",backgroundColor:"bg-c-2"}},o=["post","put","patch","delete"],i=e=>o.includes(e),s=e=>{const t=e.trim().toLowerCase();return a[t]??{short:t,color:"text-c-2",backgroundColor:"bg-c-2"}},l=e=>!!e&&r.E1.includes(e)},4126:(e,t,n)=>{"use strict";const{solelyContainsHTTPTokenCodePoints:r}=n(2661);e.exports=e=>{let t=`${e.type}/${e.subtype}`;if(0===e.parameters.size)return t;for(let[n,a]of e.parameters)t+=";",t+=n,t+="=",r(a)&&0!==a.length||(a=a.replace(/(["\\])/gu,"\\$1"),a=`"${a}"`),t+=a;return t}},4199:(e,t,n)=>{"use strict";n.r(t),n.d(t,{default:()=>$});var r=n(641),a=n(953),o=n(8082),i=n(9005),s=n(33),l=n(1266),c=n(1014);const u=(0,r.pM)({__name:"IntegrationLogo",props:{integration:{}},setup(e){const t=e,n=["adonisjs","dotnet","elysiajs","express","fastapi","fastify","go","hono","laravel","litestar","nestjs","nextjs","nitro","nuxt","platformatic","react","rust"],o=(0,r.EW)((()=>{var e;const r="Openapi",a=null==(e=t.integration)?void 0:e.toLocaleLowerCase();if(!a)return r;const o=a.charAt(0).toUpperCase()+a.slice(1);return n.includes(a)?o:r}));return(e,t)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(l.kv),{class:"w-full h-full rounded-lg",logo:o.value},null,8,["logo"]))}});var d=n(9135),p=n(5841);const h={class:"flex flex-col gap-2"},f=(0,r.pM)({__name:"SettingsAppearance",setup(e){const{colorMode:t,setColorMode:n}=(0,p.G)(),o=(0,l.Fj)({base:"w-full shadow-none text-c-1 justify-start pl-2 gap-2 border-1/2",variants:{active:{true:"bg-primary text-c-1 hover:bg-inherit",false:"bg-b-1 hover:bg-b-2"}}});return(e,i)=>((0,r.uX)(),(0,r.CE)("div",h,[(0,r.bF)((0,a.R1)(l.RO),{class:(0,s.C4)((0,a.R1)(l.cx)((0,a.R1)(o)({active:"system"===(0,a.R1)(t)}))),onClick:i[0]||(i[0]=e=>(0,a.R1)(n)("system"))},{default:(0,r.k6)((()=>[(0,r.Lk)("div",{class:(0,s.C4)(["flex items-center justify-center w-5 h-5 rounded-full border-[1.5px] p-1",{"bg-c-accent border-transparent text-b-1":"system"===(0,a.R1)(t)}])},["system"===(0,a.R1)(t)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(l.kv),{key:0,icon:"Checkmark",size:"xs",thickness:"3.5"})):(0,r.Q3)("",!0)],2),i[3]||(i[3]=(0,r.eW)(" System Preference (default) "))])),_:1},8,["class"]),(0,r.bF)((0,a.R1)(l.RO),{class:(0,s.C4)((0,a.R1)(l.cx)((0,a.R1)(o)({active:"light"===(0,a.R1)(t)}))),onClick:i[1]||(i[1]=e=>(0,a.R1)(n)("light"))},{default:(0,r.k6)((()=>[(0,r.Lk)("div",{class:(0,s.C4)(["flex items-center justify-center w-5 h-5 rounded-full border-[1.5px] p-1",{"bg-c-accent border-transparent text-b-1":"light"===(0,a.R1)(t)}])},["light"===(0,a.R1)(t)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(l.kv),{key:0,icon:"Checkmark",size:"xs",thickness:"3.5"})):(0,r.Q3)("",!0)],2),i[4]||(i[4]=(0,r.eW)(" Light Mode Always "))])),_:1},8,["class"]),(0,r.bF)((0,a.R1)(l.RO),{class:(0,s.C4)((0,a.R1)(l.cx)((0,a.R1)(o)({active:"dark"===(0,a.R1)(t)}))),onClick:i[2]||(i[2]=e=>(0,a.R1)(n)("dark"))},{default:(0,r.k6)((()=>[(0,r.Lk)("div",{class:(0,s.C4)(["flex items-center justify-center w-5 h-5 rounded-full border-[1.5px] p-1",{"bg-c-accent border-transparent text-b-1":"dark"===(0,a.R1)(t)}])},["dark"===(0,a.R1)(t)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(l.kv),{key:0,icon:"Checkmark",size:"xs",thickness:"3.5"})):(0,r.Q3)("",!0)],2),i[5]||(i[5]=(0,r.eW)(" Dark Mode Always "))])),_:1},8,["class"])]))}}),m={class:"flex flex-col gap-2"},g={key:0,class:"font-bold"},v={key:1,class:"text-c-2 mb-4 leading-[21px]"},b=(0,r.pM)({__name:"SettingsSection",setup:e=>(e,t)=>((0,r.uX)(),(0,r.CE)("div",m,[e.$slots.title?((0,r.uX)(),(0,r.CE)("h3",g,[(0,r.RG)(e.$slots,"title")])):(0,r.Q3)("",!0),e.$slots.description?((0,r.uX)(),(0,r.CE)("p",v,[(0,r.RG)(e.$slots,"description")])):(0,r.Q3)("",!0),(0,r.RG)(e.$slots,"default")]))});var y=n(8309);const w={class:"bg-b-1 h-full w-full overflow-auto"},O={class:"ml-auto mr-auto w-full max-w-[720px] px-5 py-5"},k={class:"flex flex-col gap-8"},x={class:"flex flex-col gap-2"},E={class:"flex flex-col gap-2"},S={class:"grid grid-cols-2 gap-2"},_={class:"flex items-center gap-2"},C={class:"flex items-center gap-1"},T={class:"grid grid-cols-2 gap-2"},A={class:"flex items-center gap-2"},R={class:"flex items-center gap-1"},P={class:"size-7 rounded-xl"},D="https://proxy.scalar.com",M=(0,r.pM)({__name:"SettingsGeneral",setup(e){const{activeWorkspace:t}=(0,d.JC)(),{proxyUrl:n,workspaceMutators:o}=(0,y.hA)(),i=["default","alternate","purple","solarized","saturn","kepler"],p=["elysiajs","fastify"],h=e=>({default:{light:"#fff",dark:"#0f0f0f",accent:"#0099ff"},alternate:{light:"#f9f9f9",dark:"#131313",accent:"#e7e7e7"},moon:{light:"#ccc9b3",dark:"#313332",accent:"#645b0f"},purple:{light:"#f5f6f8",dark:"#22252b",accent:"#5469d4"},solarized:{light:"#fdf6e3",dark:"#00212b",accent:"#007acc"},bluePlanet:{light:"#f0f2f5",dark:"#000e23",accent:"#e0e2e6"},saturn:{light:"#e4e4df",dark:"#2c2c30",accent:"#1763a6"},kepler:{light:"#f6f6f6",dark:"#0d0f1e",accent:"#7070ff"},mars:{light:"#f2efe8",dark:"#321116",accent:"#c75549"},deepSpace:{light:"#f4f4f5",dark:"#09090b",accent:"#8ab4f8"},none:{light:"#ffffff",dark:"#000000",accent:"#3b82f6"}}[e]||{light:"#ffffff",dark:"#000000",accent:"#3b82f6"}),m=e=>{var n;return o.edit(null==(n=t.value)?void 0:n.uid,"themeId",e)},g=(0,l.Fj)({base:"w-full shadow-none text-c-1 justify-start pl-2 gap-2 border-1/2",variants:{active:{true:"bg-primary text-c-1 hover:bg-inherit",false:"bg-b-1 hover:bg-b-2"}}}),v=e=>{var n;return o.edit(null==(n=t.value)?void 0:n.uid,"proxyUrl",e)};return(e,o)=>((0,r.uX)(),(0,r.CE)("div",w,[(0,r.Lk)("div",O,[(0,r.Lk)("div",k,[o[13]||(o[13]=(0,r.Lk)("div",null,[(0,r.Lk)("h2",{class:"mt-10 text-xl font-bold"},"Settings")],-1)),(0,r.bF)(b,null,{title:(0,r.k6)((()=>o[3]||(o[3]=[(0,r.eW)(" CORS Proxy ")]))),description:(0,r.k6)((()=>o[4]||(o[4]=[(0,r.eW)(" Browsers block cross-origin requests for security. We provide a public proxy to "),(0,r.Lk)("a",{class:"hover:text-c-1 underline-offset-2",href:"https://en.wikipedia.org/wiki/Cross-origin_resource_sharing",target:"_blank"}," bypass CORS issues ",-1),(0,r.eW)(" . Check the "),(0,r.Lk)("a",{class:"hover:text-c-1 underline-offset-2",href:"https://github.com/scalar/scalar/tree/main/examples/proxy-server",target:"_blank"}," source code on GitHub ",-1),(0,r.eW)(" . ")]))),default:(0,r.k6)((()=>{var e,i,c;return[(0,r.Lk)("div",x,[(0,r.bF)((0,a.R1)(l.RO),{class:(0,s.C4)((0,a.R1)(l.cx)((0,a.R1)(g)({active:(null==(e=(0,a.R1)(t))?void 0:e.proxyUrl)===D}))),onClick:o[0]||(o[0]=e=>v(D))},{default:(0,r.k6)((()=>{var e,n;return[(0,r.Lk)("div",{class:(0,s.C4)(["flex h-5 w-5 items-center justify-center rounded-full border-[1.5px] p-1",{"bg-c-accent text-b-1 border-transparent":(null==(e=(0,a.R1)(t))?void 0:e.proxyUrl)===D}])},[(null==(n=(0,a.R1)(t))?void 0:n.proxyUrl)===D?((0,r.uX)(),(0,r.Wv)((0,a.R1)(l.kv),{key:0,icon:"Checkmark",size:"xs",thickness:"3.5"})):(0,r.Q3)("",!0)],2),o[5]||(o[5]=(0,r.eW)(" Use proxy.scalar.com (default) "))]})),_:1},8,["class"]),(0,a.R1)(n)&&(0,a.R1)(n)!==D?((0,r.uX)(),(0,r.Wv)((0,a.R1)(l.RO),{key:0,class:(0,s.C4)((0,a.R1)(l.cx)((0,a.R1)(g)({active:(null==(i=(0,a.R1)(t))?void 0:i.proxyUrl)===(0,a.R1)(n)}))),onClick:o[1]||(o[1]=e=>v((0,a.R1)(n)))},{default:(0,r.k6)((()=>{var e,o;return[(0,r.Lk)("div",{class:(0,s.C4)(["flex h-5 w-5 items-center justify-center rounded-full border-[1.5px] p-1",{"bg-c-accent text-b-1 border-transparent":(null==(e=(0,a.R1)(t))?void 0:e.proxyUrl)===(0,a.R1)(n)}])},[(null==(o=(0,a.R1)(t))?void 0:o.proxyUrl)===(0,a.R1)(n)?((0,r.uX)(),(0,r.Wv)((0,a.R1)(l.kv),{key:0,icon:"Checkmark",size:"xs",thickness:"3.5"})):(0,r.Q3)("",!0)],2),(0,r.eW)(" Use custom proxy ("+(0,s.v_)((0,a.R1)(n))+") ",1)]})),_:1},8,["class"])):(0,r.Q3)("",!0),(0,r.bF)((0,a.R1)(l.RO),{class:(0,s.C4)((0,a.R1)(l.cx)((0,a.R1)(g)({active:!(null!=(c=(0,a.R1)(t))&&c.proxyUrl)}))),onClick:o[2]||(o[2]=e=>v(void 0))},{default:(0,r.k6)((()=>{var e,n;return[(0,r.Lk)("div",{class:(0,s.C4)(["flex h-5 w-5 items-center justify-center rounded-full border-[1.5px] p-1",!(null!=(e=(0,a.R1)(t))&&e.proxyUrl)&&"bg-c-accent text-b-1 border-transparent"])},[null!=(n=(0,a.R1)(t))&&n.proxyUrl?(0,r.Q3)("",!0):((0,r.uX)(),(0,r.Wv)((0,a.R1)(l.kv),{key:0,icon:"Checkmark",size:"xs",thickness:"3.5"}))],2),o[6]||(o[6]=(0,r.eW)(" Skip the proxy "))]})),_:1},8,["class"])])]})),_:1}),(0,r.bF)(b,null,{title:(0,r.k6)((()=>o[7]||(o[7]=[(0,r.eW)(" Themes ")]))),description:(0,r.k6)((()=>o[8]||(o[8]=[(0,r.eW)(" We’ve got a whole rainbow of themes for you to play with: ")]))),default:(0,r.k6)((()=>[(0,r.Lk)("div",E,[(0,r.Lk)("div",S,[((0,r.uX)(),(0,r.CE)(r.FK,null,(0,r.pI)(i,(e=>{var n;return(0,r.bF)((0,a.R1)(l.RO),{key:e,class:(0,s.C4)((0,a.R1)(l.cx)((0,a.R1)(g)({active:(null==(n=(0,a.R1)(t))?void 0:n.themeId)===e}))),onClick:t=>m(e)},{default:(0,r.k6)((()=>{var n,o;return[(0,r.Lk)("div",_,[(0,r.Lk)("div",{class:(0,s.C4)(["flex h-5 w-5 items-center justify-center rounded-full border-[1.5px] p-1",{"bg-c-accent text-b-1 border-transparent":(null==(n=(0,a.R1)(t))?void 0:n.themeId)===e}])},[(null==(o=(0,a.R1)(t))?void 0:o.themeId)===e?((0,r.uX)(),(0,r.Wv)((0,a.R1)(l.kv),{key:0,icon:"Checkmark",size:"xs",thickness:"3.5"})):(0,r.Q3)("",!0)],2),(0,r.eW)(" "+(0,s.v_)((0,a.R1)(c.ty)[e]),1)]),(0,r.Lk)("div",C,[(0,r.Lk)("span",{class:"border-c-3 -mr-3 inline-block h-5 w-5 rounded-full",style:(0,s.Tr)({backgroundColor:h(e).light})},null,4),(0,r.Lk)("span",{class:"border-c-3 -mr-3 inline-block h-5 w-5 rounded-full",style:(0,s.Tr)({backgroundColor:h(e).dark})},null,4),(0,r.Lk)("span",{class:"border-c-3 inline-block h-5 w-5 rounded-full",style:(0,s.Tr)({backgroundColor:h(e).accent})},null,4)])]})),_:2},1032,["class","onClick"])})),64))])])])),_:1}),(0,r.bF)(b,null,{title:(0,r.k6)((()=>o[9]||(o[9]=[(0,r.eW)(" Framework Themes ")]))),description:(0,r.k6)((()=>o[10]||(o[10]=[(0,r.eW)(" Are you a real fan? Show your support by using your favorite framework’s theme! ")]))),default:(0,r.k6)((()=>[(0,r.Lk)("div",T,[((0,r.uX)(),(0,r.CE)(r.FK,null,(0,r.pI)(p,(e=>{var n;return(0,r.bF)((0,a.R1)(l.RO),{key:e,class:(0,s.C4)((0,a.R1)(l.cx)((0,a.R1)(g)({active:(null==(n=(0,a.R1)(t))?void 0:n.themeId)===e}))),onClick:t=>m(e)},{default:(0,r.k6)((()=>{var n,o;return[(0,r.Lk)("div",A,[(0,r.Lk)("div",{class:(0,s.C4)(["flex h-5 w-5 items-center justify-center rounded-full border-[1.5px] p-1",{"bg-c-accent text-b-1 border-transparent":(null==(n=(0,a.R1)(t))?void 0:n.themeId)===e}])},[(null==(o=(0,a.R1)(t))?void 0:o.themeId)===e?((0,r.uX)(),(0,r.Wv)((0,a.R1)(l.kv),{key:0,icon:"Checkmark",size:"xs",thickness:"3.5"})):(0,r.Q3)("",!0)],2),(0,r.eW)(" "+(0,s.v_)((0,a.R1)(c.ty)[e]),1)]),(0,r.Lk)("div",R,[(0,r.Lk)("div",P,[(0,r.bF)(u,{integration:e},null,8,["integration"])])])]})),_:2},1032,["class","onClick"])})),64))])])),_:1}),(0,r.bF)(b,null,{title:(0,r.k6)((()=>o[11]||(o[11]=[(0,r.eW)(" Appearance ")]))),description:(0,r.k6)((()=>o[12]||(o[12]=[(0,r.eW)(" Choose between light, dark, or system-based appearance for your workspace. ")]))),default:(0,r.k6)((()=>[(0,r.bF)(f)])),_:1})])])]))}}),$=(0,r.pM)({__name:"Settings",setup(e){const t={general:{component:M,title:"general"}},n=(0,a.KR)("general");return(e,a)=>((0,r.uX)(),(0,r.Wv)(o.A,null,{default:(0,r.k6)((()=>[(0,r.bF)(i.A,{class:"flex-1"},{default:(0,r.k6)((()=>[t[n.value]?((0,r.uX)(),(0,r.Wv)((0,r.$y)(t[n.value].component),{key:0})):(0,r.Q3)("",!0)])),_:1})])),_:1}))}})},4279:(e,t,n)=>{"use strict";n.d(t,{o:()=>a});var r=n(3412);function a(e){const t=[];return e&&"object"==typeof e?((0,r.h)(e,(e=>(e.$ref&&"string"==typeof e.$ref&&!e.$ref.startsWith("#")&&t.push(e.$ref.split("#")[0]),e))),[...new Set(t)]):t}},4282:(e,t,n)=>{"use strict";n.d(t,{A:()=>i});var r=n(641),a=n(953),o=n(1266);const i=(0,r.pM)({inheritAttrs:!1,__name:"DataTableCell",props:{is:{default:"td"}},setup(e){const{cx:t}=(0,o.AX)();return(e,n)=>((0,r.uX)(),(0,r.Wv)((0,r.$y)(e.is),(0,r.v6)((0,a.R1)(t)("max-h-8 min-h-8 min-w-8 border-l-0 border-t border-b-0 border-r flex text-sm last:border-r-0 group-last:border-b-transparent p-0 m-0 relative"),{class:"group-[.alert]:bg-b-alert",role:"cell"}),{default:(0,r.k6)((()=>[(0,r.RG)(e.$slots,"default")])),_:3},16))}})},4325:(e,t,n)=>{"use strict";n.d(t,{G:()=>R});const r="13px",a="14px",o="15px",i="16px",s="20px",l="24px",c="28px",u="32px",d="36px",p="40px",h="256px",f="288px",m="320px",g="384px",v="448px",b="512px",y="576px",w="672px",O="768px",k="896px",x="1024px",E="1152px",S="1280px",_={darkMode:["selector",".dark-mode"],presets:[{theme:{columns:{"3xs":h,"2xs":f,xs:m,sm:g,md:v,lg:b,xl:y,"2xl":w,"3xl":O,"4xl":k,"5xl":x,"6xl":E,"7xl":S},spacing:{0:"0px",.25:"1px",.5:"2px",.75:"3px",1:"4px",1.25:"5px",1.5:"6px",1.75:"7px",2:"8px",2.25:"9px",2.5:"10px",2.75:"11px",3:"12px",3.25:r,3.5:a,3.75:o,4:i,5:s,6:l,7:c,8:u,9:d,10:p,11:"44px",12:"48px",14:"56px",16:"64px",20:"80px",20.5:"82px",21:"84px",21.5:"86px",22:"88px",22.5:"90px",23:"92px",23.5:"94px",24:"96px",28:"112px",29:"116px",30:"120px",31:"124px",32:"128px",36:"144px",40:"160px",44:"176px",45:"180px",46:"184px",47:"188px",48:"192px",52:"208px",56:"224px",60:"240px",64:h,72:f,80:m,96:g},lineHeight:{3:"12px",3.25:r,3.5:a,3.75:o,4:i,4.25:"17px",4.5:"18px",4.75:"19px",5:s,5.25:"21px",5.5:"22px",5.75:"23px",6:l,7:c,8:u,9:d,10:p},maxWidth:()=>({0:"0px",xs:m,sm:g,md:v,lg:b,xl:y,"2xl":w,"3xl":O,"4xl":k,"5xl":x,"6xl":E,"7xl":S,full:"100%","min  ":"min-content",max:"max-content",fit:"fit-content"})}}],theme:{borderRadius:{DEFAULT:"var(--scalar-radius)",md:"var(--scalar-radius)",lg:"var(--scalar-radius-lg)",xl:"var(--scalar-radius-xl)",px:"1px",full:"9999px",none:"0px"},borderWidth:{DEFAULT:"var(--scalar-border-width)",0:"0","1/2":"calc(var(--scalar-border-width) / 2)",1:"var(--scalar-border-width)",2:"calc(var(--scalar-border-width) * 2)",4:"calc(var(--scalar-border-width) * 4)"},boxShadow:{DEFAULT:"var(--scalar-shadow-1)",lg:"var(--scalar-shadow-2)",md:"var(--scalar-shadow-1)",sm:"rgba(0, 0, 0, 0.09) 0px 1px 4px",none:"0 0 #0000",border:"inset 0 0 0 1px var(--scalar-border-color)","border-1/2":"inset 0 0 0 .5px var(--scalar-border-color)"},fontFamily:{DEFAULT:"var(--scalar-font)",code:"var(--scalar-font-code)"},fontSize:{"3xs":"var(--scalar-font-size-7)",xxs:"var(--scalar-font-size-6)",xs:"var(--scalar-font-size-5)",sm:"var(--scalar-font-size-4)",base:"var(--scalar-font-size-3)",lg:"var(--scalar-font-size-2)",xl:"var(--scalar-font-size-1)"},fontWeight:{DEFAULT:"var(--scalar-regular)",normal:"var(--scalar-regular)",medium:"var(--scalar-semibold)",bold:"var(--scalar-bold)"},colors:{current:"currentColor",inherit:"inherit",b:{1:"var(--scalar-background-1)",2:"var(--scalar-background-2)",3:"var(--scalar-background-3)",4:"var(--scalar-background-3)",accent:"var(--scalar-background-accent)",btn:"var(--scalar-button-1)",danger:"var(--scalar-background-danger)",alert:"var(--scalar-background-alert)"},c:{1:"var(--scalar-color-1)",2:"var(--scalar-color-2)",3:"var(--scalar-color-3)",accent:"var(--scalar-color-accent)",ghost:"var(--scalar-color-ghost)",disabled:"var(--scalar-color-disabled)",btn:"var(--scalar-button-1-color)",danger:"var(--scalar-color-danger)",alert:"var(--scalar-color-alert)"},h:{btn:"var(--scalar-button-1-hover)"},sidebar:{b:{1:"var(--scalar-sidebar-background-1, var(--scalar-background-1))"},c:{1:"var(--scalar-sidebar-color-1, var(--scalar-color-1))",2:"var(--scalar-sidebar-color-2, var(--scalar-color-2))"},border:"var(--scalar-sidebar-border-color, var(--scalar-border-color))"},backdrop:"rgba(0, 0, 0, 0.22)",backdropdark:"rgba(0, 0, 0, 0.45)",border:"var(--scalar-border-color)",brand:"var(--scalar-brand)",green:"var(--scalar-color-green)",red:"var(--scalar-color-red)",yellow:"var(--scalar-color-yellow)",blue:"var(--scalar-color-blue)",orange:"var(--scalar-color-orange)",purple:"var(--scalar-color-purple)",grey:"var(--scalar-color-3)",indigo:"var(--scalar-color-blue)",pink:"var(--scalar-color-pink)",white:"#FFF",transparent:"transparent"},lineHeight:{none:"1",tight:"1.25",snug:"1.375",normal:"1.5",relaxed:"1.625",loose:"2",DEFAULT:"1.5",1:"var(--scalar-line-height-1)",2:"var(--scalar-line-height-2)",3:"var(--scalar-line-height-3)",4:"var(--scalar-line-height-4)",5:"var(--scalar-line-height-5)"},screens:{xs:"400px",sm:"600px",md:"800px",lg:"1000px",xl:"1200px"},zIndex:{"-1":"-1",0:"0",1:"1"},extend:{borderColor:{DEFAULT:"var(--scalar-border-color)"},brightness:{lifted:"var(--scalar-lifted-brightness)",backdrop:"var(--scalar-backdrop-brightness)"},spacing:{px:"1px",header:"48px",border:"var(--scalar-border-width)"}}}};var C=n(3073),T=n(641),A=n(953);function R(){const e=_.theme.screens,t=Object.fromEntries(Object.entries(e).map((([e,t])=>[e,(0,C.Ubj)(`(min-width: ${t})`)]))),n=(0,T.EW)((()=>Object.fromEntries(Object.entries(t).map((([e,t])=>[e,(0,A.R1)(t)])))));return{screens:e,mediaQueries:t,breakpoints:n}}},4445:(e,t,n)=>{"use strict";n.d(t,{A:()=>s});var r=n(641),a=n(953),o=n(3751);const i=["placeholder"],s=(0,r.pM)({inheritAttrs:!1,__name:"CommandActionInput",props:{modelValue:{},placeholder:{},autofocus:{type:Boolean}},emits:["update:modelValue","onDelete"],setup(e,{emit:t}){const n=e,s=t,l=(0,a.KR)(null);(0,r.sV)((()=>(0,r.dY)((()=>{var e;n.autofocus||null==(e=l.value)||e.focus()}))));const c=(0,r.EW)({get:()=>n.modelValue??"",set:e=>s("update:modelValue",e)});return(e,t)=>(0,r.bo)(((0,r.uX)(),(0,r.CE)("textarea",(0,r.v6)({ref_key:"input",ref:l,"onUpdate:modelValue":t[0]||(t[0]=e=>c.value=e),class:"border-none outline-none flex-1 w-full pl-8 text-sm min-h-8 py-1.5 resize-none",placeholder:n.placeholder??"",wrap:"hard"},e.$attrs,{onKeydown:[t[1]||(t[1]=(0,o.jR)((e=>function(e){""===c.value&&(e.preventDefault(),e.stopPropagation(),s("onDelete",e))}(e)),["delete"])),t[2]||(t[2]=(0,o.jR)((e=>function(e){var t;if(e.shiftKey||!e.target)return;e.preventDefault();const n=e.target,r=new Event("submit",{cancelable:!0});null==(t=n.form)||t.dispatchEvent(r)}(e)),["enter"]))]}),null,16,i)),[[o.Jo,c.value]])}})},4476:(e,t,n)=>{"use strict";var r,a;n.d(t,{z:()=>_t}),function(e){e.assertEqual=e=>e,e.assertIs=function(e){},e.assertNever=function(e){throw new Error},e.arrayToEnum=e=>{const t={};for(const n of e)t[n]=n;return t},e.getValidEnumValues=t=>{const n=e.objectKeys(t).filter((e=>"number"!=typeof t[t[e]])),r={};for(const e of n)r[e]=t[e];return e.objectValues(r)},e.objectValues=t=>e.objectKeys(t).map((function(e){return t[e]})),e.objectKeys="function"==typeof Object.keys?e=>Object.keys(e):e=>{const t=[];for(const n in e)Object.prototype.hasOwnProperty.call(e,n)&&t.push(n);return t},e.find=(e,t)=>{for(const n of e)if(t(n))return n},e.isInteger="function"==typeof Number.isInteger?e=>Number.isInteger(e):e=>"number"==typeof e&&isFinite(e)&&Math.floor(e)===e,e.joinValues=function(e,t=" | "){return e.map((e=>"string"==typeof e?`'${e}'`:e)).join(t)},e.jsonStringifyReplacer=(e,t)=>"bigint"==typeof t?t.toString():t}(r||(r={})),function(e){e.mergeShapes=(e,t)=>({...e,...t})}(a||(a={}));const o=r.arrayToEnum(["string","nan","number","integer","float","boolean","date","bigint","symbol","function","undefined","null","array","object","unknown","promise","void","never","map","set"]),i=e=>{switch(typeof e){case"undefined":return o.undefined;case"string":return o.string;case"number":return isNaN(e)?o.nan:o.number;case"boolean":return o.boolean;case"function":return o.function;case"bigint":return o.bigint;case"symbol":return o.symbol;case"object":return Array.isArray(e)?o.array:null===e?o.null:e.then&&"function"==typeof e.then&&e.catch&&"function"==typeof e.catch?o.promise:"undefined"!=typeof Map&&e instanceof Map?o.map:"undefined"!=typeof Set&&e instanceof Set?o.set:"undefined"!=typeof Date&&e instanceof Date?o.date:o.object;default:return o.unknown}},s=r.arrayToEnum(["invalid_type","invalid_literal","custom","invalid_union","invalid_union_discriminator","invalid_enum_value","unrecognized_keys","invalid_arguments","invalid_return_type","invalid_date","invalid_string","too_small","too_big","invalid_intersection_types","not_multiple_of","not_finite"]);class l extends Error{get errors(){return this.issues}constructor(e){super(),this.issues=[],this.addIssue=e=>{this.issues=[...this.issues,e]},this.addIssues=(e=[])=>{this.issues=[...this.issues,...e]};const t=new.target.prototype;Object.setPrototypeOf?Object.setPrototypeOf(this,t):this.__proto__=t,this.name="ZodError",this.issues=e}format(e){const t=e||function(e){return e.message},n={_errors:[]},r=e=>{for(const a of e.issues)if("invalid_union"===a.code)a.unionErrors.map(r);else if("invalid_return_type"===a.code)r(a.returnTypeError);else if("invalid_arguments"===a.code)r(a.argumentsError);else if(0===a.path.length)n._errors.push(t(a));else{let e=n,r=0;for(;r<a.path.length;){const n=a.path[r];r===a.path.length-1?(e[n]=e[n]||{_errors:[]},e[n]._errors.push(t(a))):e[n]=e[n]||{_errors:[]},e=e[n],r++}}};return r(this),n}static assert(e){if(!(e instanceof l))throw new Error(`Not a ZodError: ${e}`)}toString(){return this.message}get message(){return JSON.stringify(this.issues,r.jsonStringifyReplacer,2)}get isEmpty(){return 0===this.issues.length}flatten(e=e=>e.message){const t={},n=[];for(const r of this.issues)r.path.length>0?(t[r.path[0]]=t[r.path[0]]||[],t[r.path[0]].push(e(r))):n.push(e(r));return{formErrors:n,fieldErrors:t}}get formErrors(){return this.flatten()}}l.create=e=>new l(e);const c=(e,t)=>{let n;switch(e.code){case s.invalid_type:n=e.received===o.undefined?"Required":`Expected ${e.expected}, received ${e.received}`;break;case s.invalid_literal:n=`Invalid literal value, expected ${JSON.stringify(e.expected,r.jsonStringifyReplacer)}`;break;case s.unrecognized_keys:n=`Unrecognized key(s) in object: ${r.joinValues(e.keys,", ")}`;break;case s.invalid_union:n="Invalid input";break;case s.invalid_union_discriminator:n=`Invalid discriminator value. Expected ${r.joinValues(e.options)}`;break;case s.invalid_enum_value:n=`Invalid enum value. Expected ${r.joinValues(e.options)}, received '${e.received}'`;break;case s.invalid_arguments:n="Invalid function arguments";break;case s.invalid_return_type:n="Invalid function return type";break;case s.invalid_date:n="Invalid date";break;case s.invalid_string:"object"==typeof e.validation?"includes"in e.validation?(n=`Invalid input: must include "${e.validation.includes}"`,"number"==typeof e.validation.position&&(n=`${n} at one or more positions greater than or equal to ${e.validation.position}`)):"startsWith"in e.validation?n=`Invalid input: must start with "${e.validation.startsWith}"`:"endsWith"in e.validation?n=`Invalid input: must end with "${e.validation.endsWith}"`:r.assertNever(e.validation):n="regex"!==e.validation?`Invalid ${e.validation}`:"Invalid";break;case s.too_small:n="array"===e.type?`Array must contain ${e.exact?"exactly":e.inclusive?"at least":"more than"} ${e.minimum} element(s)`:"string"===e.type?`String must contain ${e.exact?"exactly":e.inclusive?"at least":"over"} ${e.minimum} character(s)`:"number"===e.type?`Number must be ${e.exact?"exactly equal to ":e.inclusive?"greater than or equal to ":"greater than "}${e.minimum}`:"date"===e.type?`Date must be ${e.exact?"exactly equal to ":e.inclusive?"greater than or equal to ":"greater than "}${new Date(Number(e.minimum))}`:"Invalid input";break;case s.too_big:n="array"===e.type?`Array must contain ${e.exact?"exactly":e.inclusive?"at most":"less than"} ${e.maximum} element(s)`:"string"===e.type?`String must contain ${e.exact?"exactly":e.inclusive?"at most":"under"} ${e.maximum} character(s)`:"number"===e.type?`Number must be ${e.exact?"exactly":e.inclusive?"less than or equal to":"less than"} ${e.maximum}`:"bigint"===e.type?`BigInt must be ${e.exact?"exactly":e.inclusive?"less than or equal to":"less than"} ${e.maximum}`:"date"===e.type?`Date must be ${e.exact?"exactly":e.inclusive?"smaller than or equal to":"smaller than"} ${new Date(Number(e.maximum))}`:"Invalid input";break;case s.custom:n="Invalid input";break;case s.invalid_intersection_types:n="Intersection results could not be merged";break;case s.not_multiple_of:n=`Number must be a multiple of ${e.multipleOf}`;break;case s.not_finite:n="Number must be finite";break;default:n=t.defaultError,r.assertNever(e)}return{message:n}};let u=c;function d(){return u}const p=e=>{const{data:t,path:n,errorMaps:r,issueData:a}=e,o=[...n,...a.path||[]],i={...a,path:o};if(void 0!==a.message)return{...a,path:o,message:a.message};let s="";const l=r.filter((e=>!!e)).slice().reverse();for(const e of l)s=e(i,{data:t,defaultError:s}).message;return{...a,path:o,message:s}};function h(e,t){const n=d(),r=p({issueData:t,data:e.data,path:e.path,errorMaps:[e.common.contextualErrorMap,e.schemaErrorMap,n,n===c?void 0:c].filter((e=>!!e))});e.common.issues.push(r)}class f{constructor(){this.value="valid"}dirty(){"valid"===this.value&&(this.value="dirty")}abort(){"aborted"!==this.value&&(this.value="aborted")}static mergeArray(e,t){const n=[];for(const r of t){if("aborted"===r.status)return m;"dirty"===r.status&&e.dirty(),n.push(r.value)}return{status:e.value,value:n}}static async mergeObjectAsync(e,t){const n=[];for(const e of t){const t=await e.key,r=await e.value;n.push({key:t,value:r})}return f.mergeObjectSync(e,n)}static mergeObjectSync(e,t){const n={};for(const r of t){const{key:t,value:a}=r;if("aborted"===t.status)return m;if("aborted"===a.status)return m;"dirty"===t.status&&e.dirty(),"dirty"===a.status&&e.dirty(),"__proto__"===t.value||void 0===a.value&&!r.alwaysSet||(n[t.value]=a.value)}return{status:e.value,value:n}}}const m=Object.freeze({status:"aborted"}),g=e=>({status:"dirty",value:e}),v=e=>({status:"valid",value:e}),b=e=>"aborted"===e.status,y=e=>"dirty"===e.status,w=e=>"valid"===e.status,O=e=>"undefined"!=typeof Promise&&e instanceof Promise;function k(e,t,n,r){if("a"===n&&!r)throw new TypeError("Private accessor was defined without a getter");if("function"==typeof t?e!==t||!r:!t.has(e))throw new TypeError("Cannot read private member from an object whose class did not declare it");return"m"===n?r:"a"===n?r.call(e):r?r.value:t.get(e)}function x(e,t,n,r,a){if("m"===r)throw new TypeError("Private method is not writable");if("a"===r&&!a)throw new TypeError("Private accessor was defined without a setter");if("function"==typeof t?e!==t||!a:!t.has(e))throw new TypeError("Cannot write private member to an object whose class did not declare it");return"a"===r?a.call(e,n):a?a.value=n:t.set(e,n),n}var E,S,_;"function"==typeof SuppressedError&&SuppressedError,function(e){e.errToObj=e=>"string"==typeof e?{message:e}:e||{},e.toString=e=>"string"==typeof e?e:null==e?void 0:e.message}(E||(E={}));class C{constructor(e,t,n,r){this._cachedPath=[],this.parent=e,this.data=t,this._path=n,this._key=r}get path(){return this._cachedPath.length||(this._key instanceof Array?this._cachedPath.push(...this._path,...this._key):this._cachedPath.push(...this._path,this._key)),this._cachedPath}}const T=(e,t)=>{if(w(t))return{success:!0,data:t.value};if(!e.common.issues.length)throw new Error("Validation failed but no issues detected.");return{success:!1,get error(){if(this._error)return this._error;const t=new l(e.common.issues);return this._error=t,this._error}}};function A(e){if(!e)return{};const{errorMap:t,invalid_type_error:n,required_error:r,description:a}=e;if(t&&(n||r))throw new Error('Can\'t use "invalid_type_error" or "required_error" in conjunction with custom error map.');return t?{errorMap:t,description:a}:{errorMap:(t,a)=>{var o,i;const{message:s}=e;return"invalid_enum_value"===t.code?{message:null!=s?s:a.defaultError}:void 0===a.data?{message:null!==(o=null!=s?s:r)&&void 0!==o?o:a.defaultError}:"invalid_type"!==t.code?{message:a.defaultError}:{message:null!==(i=null!=s?s:n)&&void 0!==i?i:a.defaultError}},description:a}}class R{get description(){return this._def.description}_getType(e){return i(e.data)}_getOrReturnCtx(e,t){return t||{common:e.parent.common,data:e.data,parsedType:i(e.data),schemaErrorMap:this._def.errorMap,path:e.path,parent:e.parent}}_processInputParams(e){return{status:new f,ctx:{common:e.parent.common,data:e.data,parsedType:i(e.data),schemaErrorMap:this._def.errorMap,path:e.path,parent:e.parent}}}_parseSync(e){const t=this._parse(e);if(O(t))throw new Error("Synchronous parse encountered promise.");return t}_parseAsync(e){const t=this._parse(e);return Promise.resolve(t)}parse(e,t){const n=this.safeParse(e,t);if(n.success)return n.data;throw n.error}safeParse(e,t){var n;const r={common:{issues:[],async:null!==(n=null==t?void 0:t.async)&&void 0!==n&&n,contextualErrorMap:null==t?void 0:t.errorMap},path:(null==t?void 0:t.path)||[],schemaErrorMap:this._def.errorMap,parent:null,data:e,parsedType:i(e)},a=this._parseSync({data:e,path:r.path,parent:r});return T(r,a)}"~validate"(e){var t,n;const r={common:{issues:[],async:!!this["~standard"].async},path:[],schemaErrorMap:this._def.errorMap,parent:null,data:e,parsedType:i(e)};if(!this["~standard"].async)try{const t=this._parseSync({data:e,path:[],parent:r});return w(t)?{value:t.value}:{issues:r.common.issues}}catch(e){(null===(n=null===(t=null==e?void 0:e.message)||void 0===t?void 0:t.toLowerCase())||void 0===n?void 0:n.includes("encountered"))&&(this["~standard"].async=!0),r.common={issues:[],async:!0}}return this._parseAsync({data:e,path:[],parent:r}).then((e=>w(e)?{value:e.value}:{issues:r.common.issues}))}async parseAsync(e,t){const n=await this.safeParseAsync(e,t);if(n.success)return n.data;throw n.error}async safeParseAsync(e,t){const n={common:{issues:[],contextualErrorMap:null==t?void 0:t.errorMap,async:!0},path:(null==t?void 0:t.path)||[],schemaErrorMap:this._def.errorMap,parent:null,data:e,parsedType:i(e)},r=this._parse({data:e,path:n.path,parent:n}),a=await(O(r)?r:Promise.resolve(r));return T(n,a)}refine(e,t){const n=e=>"string"==typeof t||void 0===t?{message:t}:"function"==typeof t?t(e):t;return this._refinement(((t,r)=>{const a=e(t),o=()=>r.addIssue({code:s.custom,...n(t)});return"undefined"!=typeof Promise&&a instanceof Promise?a.then((e=>!!e||(o(),!1))):!!a||(o(),!1)}))}refinement(e,t){return this._refinement(((n,r)=>!!e(n)||(r.addIssue("function"==typeof t?t(n,r):t),!1)))}_refinement(e){return new Re({schema:this,typeName:je.ZodEffects,effect:{type:"refinement",refinement:e}})}superRefine(e){return this._refinement(e)}constructor(e){this.spa=this.safeParseAsync,this._def=e,this.parse=this.parse.bind(this),this.safeParse=this.safeParse.bind(this),this.parseAsync=this.parseAsync.bind(this),this.safeParseAsync=this.safeParseAsync.bind(this),this.spa=this.spa.bind(this),this.refine=this.refine.bind(this),this.refinement=this.refinement.bind(this),this.superRefine=this.superRefine.bind(this),this.optional=this.optional.bind(this),this.nullable=this.nullable.bind(this),this.nullish=this.nullish.bind(this),this.array=this.array.bind(this),this.promise=this.promise.bind(this),this.or=this.or.bind(this),this.and=this.and.bind(this),this.transform=this.transform.bind(this),this.brand=this.brand.bind(this),this.default=this.default.bind(this),this.catch=this.catch.bind(this),this.describe=this.describe.bind(this),this.pipe=this.pipe.bind(this),this.readonly=this.readonly.bind(this),this.isNullable=this.isNullable.bind(this),this.isOptional=this.isOptional.bind(this),this["~standard"]={version:1,vendor:"zod",validate:e=>this["~validate"](e)}}optional(){return Pe.create(this,this._def)}nullable(){return De.create(this,this._def)}nullish(){return this.nullable().optional()}array(){return de.create(this)}promise(){return Ae.create(this,this._def)}or(e){return fe.create([this,e],this._def)}and(e){return be.create(this,e,this._def)}transform(e){return new Re({...A(this._def),schema:this,typeName:je.ZodEffects,effect:{type:"transform",transform:e}})}default(e){const t="function"==typeof e?e:()=>e;return new Me({...A(this._def),innerType:this,defaultValue:t,typeName:je.ZodDefault})}brand(){return new Le({typeName:je.ZodBranded,type:this,...A(this._def)})}catch(e){const t="function"==typeof e?e:()=>e;return new $e({...A(this._def),innerType:this,catchValue:t,typeName:je.ZodCatch})}describe(e){return new(0,this.constructor)({...this._def,description:e})}pipe(e){return Be.create(this,e)}readonly(){return Fe.create(this)}isOptional(){return this.safeParse(void 0).success}isNullable(){return this.safeParse(null).success}}const P=/^c[^\s-]{8,}$/i,D=/^[0-9a-z]+$/,M=/^[0-9A-HJKMNP-TV-Z]{26}$/i,$=/^[0-9a-fA-F]{8}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{12}$/i,N=/^[a-z0-9_-]{21}$/i,I=/^[A-Za-z0-9-_]+\.[A-Za-z0-9-_]+\.[A-Za-z0-9-_]*$/,L=/^[-+]?P(?!$)(?:(?:[-+]?\d+Y)|(?:[-+]?\d+[.,]\d+Y$))?(?:(?:[-+]?\d+M)|(?:[-+]?\d+[.,]\d+M$))?(?:(?:[-+]?\d+W)|(?:[-+]?\d+[.,]\d+W$))?(?:(?:[-+]?\d+D)|(?:[-+]?\d+[.,]\d+D$))?(?:T(?=[\d+-])(?:(?:[-+]?\d+H)|(?:[-+]?\d+[.,]\d+H$))?(?:(?:[-+]?\d+M)|(?:[-+]?\d+[.,]\d+M$))?(?:[-+]?\d+(?:[.,]\d+)?S)?)??$/,B=/^(?!\.)(?!.*\.\.)([A-Z0-9_'+\-\.]*)[A-Z0-9_+-]@([A-Z0-9][A-Z0-9\-]*\.)+[A-Z]{2,}$/i;let F;const Q=/^(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])$/,U=/^(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\.){3}(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\/(3[0-2]|[12]?[0-9])$/,z=/^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))$/,j=/^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))\/(12[0-8]|1[01][0-9]|[1-9]?[0-9])$/,W=/^([0-9a-zA-Z+/]{4})*(([0-9a-zA-Z+/]{2}==)|([0-9a-zA-Z+/]{3}=))?$/,X=/^([0-9a-zA-Z-_]{4})*(([0-9a-zA-Z-_]{2}(==)?)|([0-9a-zA-Z-_]{3}(=)?))?$/,q="((\\d\\d[2468][048]|\\d\\d[13579][26]|\\d\\d0[48]|[02468][048]00|[13579][26]00)-02-29|\\d{4}-((0[13578]|1[02])-(0[1-9]|[12]\\d|3[01])|(0[469]|11)-(0[1-9]|[12]\\d|30)|(02)-(0[1-9]|1\\d|2[0-8])))",V=new RegExp(`^${q}$`);function H(e){let t="([01]\\d|2[0-3]):[0-5]\\d:[0-5]\\d";return e.precision?t=`${t}\\.\\d{${e.precision}}`:null==e.precision&&(t=`${t}(\\.\\d+)?`),t}function Z(e){let t=`${q}T${H(e)}`;const n=[];return n.push(e.local?"Z?":"Z"),e.offset&&n.push("([+-]\\d{2}:?\\d{2})"),t=`${t}(${n.join("|")})`,new RegExp(`^${t}$`)}function G(e,t){if(!I.test(e))return!1;try{const[n]=e.split("."),r=n.replace(/-/g,"+").replace(/_/g,"/").padEnd(n.length+(4-n.length%4)%4,"="),a=JSON.parse(atob(r));return!("object"!=typeof a||null===a||!a.typ||!a.alg||t&&a.alg!==t)}catch(e){return!1}}function K(e,t){return!("v4"!==t&&t||!U.test(e))||!("v6"!==t&&t||!j.test(e))}class Y extends R{_parse(e){if(this._def.coerce&&(e.data=String(e.data)),this._getType(e)!==o.string){const t=this._getOrReturnCtx(e);return h(t,{code:s.invalid_type,expected:o.string,received:t.parsedType}),m}const t=new f;let n;for(const o of this._def.checks)if("min"===o.kind)e.data.length<o.value&&(n=this._getOrReturnCtx(e,n),h(n,{code:s.too_small,minimum:o.value,type:"string",inclusive:!0,exact:!1,message:o.message}),t.dirty());else if("max"===o.kind)e.data.length>o.value&&(n=this._getOrReturnCtx(e,n),h(n,{code:s.too_big,maximum:o.value,type:"string",inclusive:!0,exact:!1,message:o.message}),t.dirty());else if("length"===o.kind){const r=e.data.length>o.value,a=e.data.length<o.value;(r||a)&&(n=this._getOrReturnCtx(e,n),r?h(n,{code:s.too_big,maximum:o.value,type:"string",inclusive:!0,exact:!0,message:o.message}):a&&h(n,{code:s.too_small,minimum:o.value,type:"string",inclusive:!0,exact:!0,message:o.message}),t.dirty())}else if("email"===o.kind)B.test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"email",code:s.invalid_string,message:o.message}),t.dirty());else if("emoji"===o.kind)F||(F=new RegExp("^(\\p{Extended_Pictographic}|\\p{Emoji_Component})+$","u")),F.test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"emoji",code:s.invalid_string,message:o.message}),t.dirty());else if("uuid"===o.kind)$.test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"uuid",code:s.invalid_string,message:o.message}),t.dirty());else if("nanoid"===o.kind)N.test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"nanoid",code:s.invalid_string,message:o.message}),t.dirty());else if("cuid"===o.kind)P.test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"cuid",code:s.invalid_string,message:o.message}),t.dirty());else if("cuid2"===o.kind)D.test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"cuid2",code:s.invalid_string,message:o.message}),t.dirty());else if("ulid"===o.kind)M.test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"ulid",code:s.invalid_string,message:o.message}),t.dirty());else if("url"===o.kind)try{new URL(e.data)}catch(r){n=this._getOrReturnCtx(e,n),h(n,{validation:"url",code:s.invalid_string,message:o.message}),t.dirty()}else"regex"===o.kind?(o.regex.lastIndex=0,o.regex.test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"regex",code:s.invalid_string,message:o.message}),t.dirty())):"trim"===o.kind?e.data=e.data.trim():"includes"===o.kind?e.data.includes(o.value,o.position)||(n=this._getOrReturnCtx(e,n),h(n,{code:s.invalid_string,validation:{includes:o.value,position:o.position},message:o.message}),t.dirty()):"toLowerCase"===o.kind?e.data=e.data.toLowerCase():"toUpperCase"===o.kind?e.data=e.data.toUpperCase():"startsWith"===o.kind?e.data.startsWith(o.value)||(n=this._getOrReturnCtx(e,n),h(n,{code:s.invalid_string,validation:{startsWith:o.value},message:o.message}),t.dirty()):"endsWith"===o.kind?e.data.endsWith(o.value)||(n=this._getOrReturnCtx(e,n),h(n,{code:s.invalid_string,validation:{endsWith:o.value},message:o.message}),t.dirty()):"datetime"===o.kind?Z(o).test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{code:s.invalid_string,validation:"datetime",message:o.message}),t.dirty()):"date"===o.kind?V.test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{code:s.invalid_string,validation:"date",message:o.message}),t.dirty()):"time"===o.kind?new RegExp(`^${H(o)}$`).test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{code:s.invalid_string,validation:"time",message:o.message}),t.dirty()):"duration"===o.kind?L.test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"duration",code:s.invalid_string,message:o.message}),t.dirty()):"ip"===o.kind?(a=e.data,("v4"===(i=o.version)||!i)&&Q.test(a)||("v6"===i||!i)&&z.test(a)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"ip",code:s.invalid_string,message:o.message}),t.dirty())):"jwt"===o.kind?G(e.data,o.alg)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"jwt",code:s.invalid_string,message:o.message}),t.dirty()):"cidr"===o.kind?K(e.data,o.version)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"cidr",code:s.invalid_string,message:o.message}),t.dirty()):"base64"===o.kind?W.test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"base64",code:s.invalid_string,message:o.message}),t.dirty()):"base64url"===o.kind?X.test(e.data)||(n=this._getOrReturnCtx(e,n),h(n,{validation:"base64url",code:s.invalid_string,message:o.message}),t.dirty()):r.assertNever(o);var a,i;return{status:t.value,value:e.data}}_regex(e,t,n){return this.refinement((t=>e.test(t)),{validation:t,code:s.invalid_string,...E.errToObj(n)})}_addCheck(e){return new Y({...this._def,checks:[...this._def.checks,e]})}email(e){return this._addCheck({kind:"email",...E.errToObj(e)})}url(e){return this._addCheck({kind:"url",...E.errToObj(e)})}emoji(e){return this._addCheck({kind:"emoji",...E.errToObj(e)})}uuid(e){return this._addCheck({kind:"uuid",...E.errToObj(e)})}nanoid(e){return this._addCheck({kind:"nanoid",...E.errToObj(e)})}cuid(e){return this._addCheck({kind:"cuid",...E.errToObj(e)})}cuid2(e){return this._addCheck({kind:"cuid2",...E.errToObj(e)})}ulid(e){return this._addCheck({kind:"ulid",...E.errToObj(e)})}base64(e){return this._addCheck({kind:"base64",...E.errToObj(e)})}base64url(e){return this._addCheck({kind:"base64url",...E.errToObj(e)})}jwt(e){return this._addCheck({kind:"jwt",...E.errToObj(e)})}ip(e){return this._addCheck({kind:"ip",...E.errToObj(e)})}cidr(e){return this._addCheck({kind:"cidr",...E.errToObj(e)})}datetime(e){var t,n;return"string"==typeof e?this._addCheck({kind:"datetime",precision:null,offset:!1,local:!1,message:e}):this._addCheck({kind:"datetime",precision:void 0===(null==e?void 0:e.precision)?null:null==e?void 0:e.precision,offset:null!==(t=null==e?void 0:e.offset)&&void 0!==t&&t,local:null!==(n=null==e?void 0:e.local)&&void 0!==n&&n,...E.errToObj(null==e?void 0:e.message)})}date(e){return this._addCheck({kind:"date",message:e})}time(e){return"string"==typeof e?this._addCheck({kind:"time",precision:null,message:e}):this._addCheck({kind:"time",precision:void 0===(null==e?void 0:e.precision)?null:null==e?void 0:e.precision,...E.errToObj(null==e?void 0:e.message)})}duration(e){return this._addCheck({kind:"duration",...E.errToObj(e)})}regex(e,t){return this._addCheck({kind:"regex",regex:e,...E.errToObj(t)})}includes(e,t){return this._addCheck({kind:"includes",value:e,position:null==t?void 0:t.position,...E.errToObj(null==t?void 0:t.message)})}startsWith(e,t){return this._addCheck({kind:"startsWith",value:e,...E.errToObj(t)})}endsWith(e,t){return this._addCheck({kind:"endsWith",value:e,...E.errToObj(t)})}min(e,t){return this._addCheck({kind:"min",value:e,...E.errToObj(t)})}max(e,t){return this._addCheck({kind:"max",value:e,...E.errToObj(t)})}length(e,t){return this._addCheck({kind:"length",value:e,...E.errToObj(t)})}nonempty(e){return this.min(1,E.errToObj(e))}trim(){return new Y({...this._def,checks:[...this._def.checks,{kind:"trim"}]})}toLowerCase(){return new Y({...this._def,checks:[...this._def.checks,{kind:"toLowerCase"}]})}toUpperCase(){return new Y({...this._def,checks:[...this._def.checks,{kind:"toUpperCase"}]})}get isDatetime(){return!!this._def.checks.find((e=>"datetime"===e.kind))}get isDate(){return!!this._def.checks.find((e=>"date"===e.kind))}get isTime(){return!!this._def.checks.find((e=>"time"===e.kind))}get isDuration(){return!!this._def.checks.find((e=>"duration"===e.kind))}get isEmail(){return!!this._def.checks.find((e=>"email"===e.kind))}get isURL(){return!!this._def.checks.find((e=>"url"===e.kind))}get isEmoji(){return!!this._def.checks.find((e=>"emoji"===e.kind))}get isUUID(){return!!this._def.checks.find((e=>"uuid"===e.kind))}get isNANOID(){return!!this._def.checks.find((e=>"nanoid"===e.kind))}get isCUID(){return!!this._def.checks.find((e=>"cuid"===e.kind))}get isCUID2(){return!!this._def.checks.find((e=>"cuid2"===e.kind))}get isULID(){return!!this._def.checks.find((e=>"ulid"===e.kind))}get isIP(){return!!this._def.checks.find((e=>"ip"===e.kind))}get isCIDR(){return!!this._def.checks.find((e=>"cidr"===e.kind))}get isBase64(){return!!this._def.checks.find((e=>"base64"===e.kind))}get isBase64url(){return!!this._def.checks.find((e=>"base64url"===e.kind))}get minLength(){let e=null;for(const t of this._def.checks)"min"===t.kind&&(null===e||t.value>e)&&(e=t.value);return e}get maxLength(){let e=null;for(const t of this._def.checks)"max"===t.kind&&(null===e||t.value<e)&&(e=t.value);return e}}function J(e,t){const n=(e.toString().split(".")[1]||"").length,r=(t.toString().split(".")[1]||"").length,a=n>r?n:r;return parseInt(e.toFixed(a).replace(".",""))%parseInt(t.toFixed(a).replace(".",""))/Math.pow(10,a)}Y.create=e=>{var t;return new Y({checks:[],typeName:je.ZodString,coerce:null!==(t=null==e?void 0:e.coerce)&&void 0!==t&&t,...A(e)})};class ee extends R{constructor(){super(...arguments),this.min=this.gte,this.max=this.lte,this.step=this.multipleOf}_parse(e){if(this._def.coerce&&(e.data=Number(e.data)),this._getType(e)!==o.number){const t=this._getOrReturnCtx(e);return h(t,{code:s.invalid_type,expected:o.number,received:t.parsedType}),m}let t;const n=new f;for(const a of this._def.checks)"int"===a.kind?r.isInteger(e.data)||(t=this._getOrReturnCtx(e,t),h(t,{code:s.invalid_type,expected:"integer",received:"float",message:a.message}),n.dirty()):"min"===a.kind?(a.inclusive?e.data<a.value:e.data<=a.value)&&(t=this._getOrReturnCtx(e,t),h(t,{code:s.too_small,minimum:a.value,type:"number",inclusive:a.inclusive,exact:!1,message:a.message}),n.dirty()):"max"===a.kind?(a.inclusive?e.data>a.value:e.data>=a.value)&&(t=this._getOrReturnCtx(e,t),h(t,{code:s.too_big,maximum:a.value,type:"number",inclusive:a.inclusive,exact:!1,message:a.message}),n.dirty()):"multipleOf"===a.kind?0!==J(e.data,a.value)&&(t=this._getOrReturnCtx(e,t),h(t,{code:s.not_multiple_of,multipleOf:a.value,message:a.message}),n.dirty()):"finite"===a.kind?Number.isFinite(e.data)||(t=this._getOrReturnCtx(e,t),h(t,{code:s.not_finite,message:a.message}),n.dirty()):r.assertNever(a);return{status:n.value,value:e.data}}gte(e,t){return this.setLimit("min",e,!0,E.toString(t))}gt(e,t){return this.setLimit("min",e,!1,E.toString(t))}lte(e,t){return this.setLimit("max",e,!0,E.toString(t))}lt(e,t){return this.setLimit("max",e,!1,E.toString(t))}setLimit(e,t,n,r){return new ee({...this._def,checks:[...this._def.checks,{kind:e,value:t,inclusive:n,message:E.toString(r)}]})}_addCheck(e){return new ee({...this._def,checks:[...this._def.checks,e]})}int(e){return this._addCheck({kind:"int",message:E.toString(e)})}positive(e){return this._addCheck({kind:"min",value:0,inclusive:!1,message:E.toString(e)})}negative(e){return this._addCheck({kind:"max",value:0,inclusive:!1,message:E.toString(e)})}nonpositive(e){return this._addCheck({kind:"max",value:0,inclusive:!0,message:E.toString(e)})}nonnegative(e){return this._addCheck({kind:"min",value:0,inclusive:!0,message:E.toString(e)})}multipleOf(e,t){return this._addCheck({kind:"multipleOf",value:e,message:E.toString(t)})}finite(e){return this._addCheck({kind:"finite",message:E.toString(e)})}safe(e){return this._addCheck({kind:"min",inclusive:!0,value:Number.MIN_SAFE_INTEGER,message:E.toString(e)})._addCheck({kind:"max",inclusive:!0,value:Number.MAX_SAFE_INTEGER,message:E.toString(e)})}get minValue(){let e=null;for(const t of this._def.checks)"min"===t.kind&&(null===e||t.value>e)&&(e=t.value);return e}get maxValue(){let e=null;for(const t of this._def.checks)"max"===t.kind&&(null===e||t.value<e)&&(e=t.value);return e}get isInt(){return!!this._def.checks.find((e=>"int"===e.kind||"multipleOf"===e.kind&&r.isInteger(e.value)))}get isFinite(){let e=null,t=null;for(const n of this._def.checks){if("finite"===n.kind||"int"===n.kind||"multipleOf"===n.kind)return!0;"min"===n.kind?(null===t||n.value>t)&&(t=n.value):"max"===n.kind&&(null===e||n.value<e)&&(e=n.value)}return Number.isFinite(t)&&Number.isFinite(e)}}ee.create=e=>new ee({checks:[],typeName:je.ZodNumber,coerce:(null==e?void 0:e.coerce)||!1,...A(e)});class te extends R{constructor(){super(...arguments),this.min=this.gte,this.max=this.lte}_parse(e){if(this._def.coerce)try{e.data=BigInt(e.data)}catch(t){return this._getInvalidInput(e)}if(this._getType(e)!==o.bigint)return this._getInvalidInput(e);let t;const n=new f;for(const a of this._def.checks)"min"===a.kind?(a.inclusive?e.data<a.value:e.data<=a.value)&&(t=this._getOrReturnCtx(e,t),h(t,{code:s.too_small,type:"bigint",minimum:a.value,inclusive:a.inclusive,message:a.message}),n.dirty()):"max"===a.kind?(a.inclusive?e.data>a.value:e.data>=a.value)&&(t=this._getOrReturnCtx(e,t),h(t,{code:s.too_big,type:"bigint",maximum:a.value,inclusive:a.inclusive,message:a.message}),n.dirty()):"multipleOf"===a.kind?e.data%a.value!==BigInt(0)&&(t=this._getOrReturnCtx(e,t),h(t,{code:s.not_multiple_of,multipleOf:a.value,message:a.message}),n.dirty()):r.assertNever(a);return{status:n.value,value:e.data}}_getInvalidInput(e){const t=this._getOrReturnCtx(e);return h(t,{code:s.invalid_type,expected:o.bigint,received:t.parsedType}),m}gte(e,t){return this.setLimit("min",e,!0,E.toString(t))}gt(e,t){return this.setLimit("min",e,!1,E.toString(t))}lte(e,t){return this.setLimit("max",e,!0,E.toString(t))}lt(e,t){return this.setLimit("max",e,!1,E.toString(t))}setLimit(e,t,n,r){return new te({...this._def,checks:[...this._def.checks,{kind:e,value:t,inclusive:n,message:E.toString(r)}]})}_addCheck(e){return new te({...this._def,checks:[...this._def.checks,e]})}positive(e){return this._addCheck({kind:"min",value:BigInt(0),inclusive:!1,message:E.toString(e)})}negative(e){return this._addCheck({kind:"max",value:BigInt(0),inclusive:!1,message:E.toString(e)})}nonpositive(e){return this._addCheck({kind:"max",value:BigInt(0),inclusive:!0,message:E.toString(e)})}nonnegative(e){return this._addCheck({kind:"min",value:BigInt(0),inclusive:!0,message:E.toString(e)})}multipleOf(e,t){return this._addCheck({kind:"multipleOf",value:e,message:E.toString(t)})}get minValue(){let e=null;for(const t of this._def.checks)"min"===t.kind&&(null===e||t.value>e)&&(e=t.value);return e}get maxValue(){let e=null;for(const t of this._def.checks)"max"===t.kind&&(null===e||t.value<e)&&(e=t.value);return e}}te.create=e=>{var t;return new te({checks:[],typeName:je.ZodBigInt,coerce:null!==(t=null==e?void 0:e.coerce)&&void 0!==t&&t,...A(e)})};class ne extends R{_parse(e){if(this._def.coerce&&(e.data=Boolean(e.data)),this._getType(e)!==o.boolean){const t=this._getOrReturnCtx(e);return h(t,{code:s.invalid_type,expected:o.boolean,received:t.parsedType}),m}return v(e.data)}}ne.create=e=>new ne({typeName:je.ZodBoolean,coerce:(null==e?void 0:e.coerce)||!1,...A(e)});class re extends R{_parse(e){if(this._def.coerce&&(e.data=new Date(e.data)),this._getType(e)!==o.date){const t=this._getOrReturnCtx(e);return h(t,{code:s.invalid_type,expected:o.date,received:t.parsedType}),m}if(isNaN(e.data.getTime()))return h(this._getOrReturnCtx(e),{code:s.invalid_date}),m;const t=new f;let n;for(const a of this._def.checks)"min"===a.kind?e.data.getTime()<a.value&&(n=this._getOrReturnCtx(e,n),h(n,{code:s.too_small,message:a.message,inclusive:!0,exact:!1,minimum:a.value,type:"date"}),t.dirty()):"max"===a.kind?e.data.getTime()>a.value&&(n=this._getOrReturnCtx(e,n),h(n,{code:s.too_big,message:a.message,inclusive:!0,exact:!1,maximum:a.value,type:"date"}),t.dirty()):r.assertNever(a);return{status:t.value,value:new Date(e.data.getTime())}}_addCheck(e){return new re({...this._def,checks:[...this._def.checks,e]})}min(e,t){return this._addCheck({kind:"min",value:e.getTime(),message:E.toString(t)})}max(e,t){return this._addCheck({kind:"max",value:e.getTime(),message:E.toString(t)})}get minDate(){let e=null;for(const t of this._def.checks)"min"===t.kind&&(null===e||t.value>e)&&(e=t.value);return null!=e?new Date(e):null}get maxDate(){let e=null;for(const t of this._def.checks)"max"===t.kind&&(null===e||t.value<e)&&(e=t.value);return null!=e?new Date(e):null}}re.create=e=>new re({checks:[],coerce:(null==e?void 0:e.coerce)||!1,typeName:je.ZodDate,...A(e)});class ae extends R{_parse(e){if(this._getType(e)!==o.symbol){const t=this._getOrReturnCtx(e);return h(t,{code:s.invalid_type,expected:o.symbol,received:t.parsedType}),m}return v(e.data)}}ae.create=e=>new ae({typeName:je.ZodSymbol,...A(e)});class oe extends R{_parse(e){if(this._getType(e)!==o.undefined){const t=this._getOrReturnCtx(e);return h(t,{code:s.invalid_type,expected:o.undefined,received:t.parsedType}),m}return v(e.data)}}oe.create=e=>new oe({typeName:je.ZodUndefined,...A(e)});class ie extends R{_parse(e){if(this._getType(e)!==o.null){const t=this._getOrReturnCtx(e);return h(t,{code:s.invalid_type,expected:o.null,received:t.parsedType}),m}return v(e.data)}}ie.create=e=>new ie({typeName:je.ZodNull,...A(e)});class se extends R{constructor(){super(...arguments),this._any=!0}_parse(e){return v(e.data)}}se.create=e=>new se({typeName:je.ZodAny,...A(e)});class le extends R{constructor(){super(...arguments),this._unknown=!0}_parse(e){return v(e.data)}}le.create=e=>new le({typeName:je.ZodUnknown,...A(e)});class ce extends R{_parse(e){const t=this._getOrReturnCtx(e);return h(t,{code:s.invalid_type,expected:o.never,received:t.parsedType}),m}}ce.create=e=>new ce({typeName:je.ZodNever,...A(e)});class ue extends R{_parse(e){if(this._getType(e)!==o.undefined){const t=this._getOrReturnCtx(e);return h(t,{code:s.invalid_type,expected:o.void,received:t.parsedType}),m}return v(e.data)}}ue.create=e=>new ue({typeName:je.ZodVoid,...A(e)});class de extends R{_parse(e){const{ctx:t,status:n}=this._processInputParams(e),r=this._def;if(t.parsedType!==o.array)return h(t,{code:s.invalid_type,expected:o.array,received:t.parsedType}),m;if(null!==r.exactLength){const e=t.data.length>r.exactLength.value,a=t.data.length<r.exactLength.value;(e||a)&&(h(t,{code:e?s.too_big:s.too_small,minimum:a?r.exactLength.value:void 0,maximum:e?r.exactLength.value:void 0,type:"array",inclusive:!0,exact:!0,message:r.exactLength.message}),n.dirty())}if(null!==r.minLength&&t.data.length<r.minLength.value&&(h(t,{code:s.too_small,minimum:r.minLength.value,type:"array",inclusive:!0,exact:!1,message:r.minLength.message}),n.dirty()),null!==r.maxLength&&t.data.length>r.maxLength.value&&(h(t,{code:s.too_big,maximum:r.maxLength.value,type:"array",inclusive:!0,exact:!1,message:r.maxLength.message}),n.dirty()),t.common.async)return Promise.all([...t.data].map(((e,n)=>r.type._parseAsync(new C(t,e,t.path,n))))).then((e=>f.mergeArray(n,e)));const a=[...t.data].map(((e,n)=>r.type._parseSync(new C(t,e,t.path,n))));return f.mergeArray(n,a)}get element(){return this._def.type}min(e,t){return new de({...this._def,minLength:{value:e,message:E.toString(t)}})}max(e,t){return new de({...this._def,maxLength:{value:e,message:E.toString(t)}})}length(e,t){return new de({...this._def,exactLength:{value:e,message:E.toString(t)}})}nonempty(e){return this.min(1,e)}}function pe(e){if(e instanceof he){const t={};for(const n in e.shape){const r=e.shape[n];t[n]=Pe.create(pe(r))}return new he({...e._def,shape:()=>t})}return e instanceof de?new de({...e._def,type:pe(e.element)}):e instanceof Pe?Pe.create(pe(e.unwrap())):e instanceof De?De.create(pe(e.unwrap())):e instanceof ye?ye.create(e.items.map((e=>pe(e)))):e}de.create=(e,t)=>new de({type:e,minLength:null,maxLength:null,exactLength:null,typeName:je.ZodArray,...A(t)});class he extends R{constructor(){super(...arguments),this._cached=null,this.nonstrict=this.passthrough,this.augment=this.extend}_getCached(){if(null!==this._cached)return this._cached;const e=this._def.shape(),t=r.objectKeys(e);return this._cached={shape:e,keys:t}}_parse(e){if(this._getType(e)!==o.object){const t=this._getOrReturnCtx(e);return h(t,{code:s.invalid_type,expected:o.object,received:t.parsedType}),m}const{status:t,ctx:n}=this._processInputParams(e),{shape:r,keys:a}=this._getCached(),i=[];if(!(this._def.catchall instanceof ce&&"strip"===this._def.unknownKeys))for(const e in n.data)a.includes(e)||i.push(e);const l=[];for(const e of a){const t=r[e],a=n.data[e];l.push({key:{status:"valid",value:e},value:t._parse(new C(n,a,n.path,e)),alwaysSet:e in n.data})}if(this._def.catchall instanceof ce){const e=this._def.unknownKeys;if("passthrough"===e)for(const e of i)l.push({key:{status:"valid",value:e},value:{status:"valid",value:n.data[e]}});else if("strict"===e)i.length>0&&(h(n,{code:s.unrecognized_keys,keys:i}),t.dirty());else if("strip"!==e)throw new Error("Internal ZodObject error: invalid unknownKeys value.")}else{const e=this._def.catchall;for(const t of i){const r=n.data[t];l.push({key:{status:"valid",value:t},value:e._parse(new C(n,r,n.path,t)),alwaysSet:t in n.data})}}return n.common.async?Promise.resolve().then((async()=>{const e=[];for(const t of l){const n=await t.key,r=await t.value;e.push({key:n,value:r,alwaysSet:t.alwaysSet})}return e})).then((e=>f.mergeObjectSync(t,e))):f.mergeObjectSync(t,l)}get shape(){return this._def.shape()}strict(e){return E.errToObj,new he({...this._def,unknownKeys:"strict",...void 0!==e?{errorMap:(t,n)=>{var r,a,o,i;const s=null!==(o=null===(a=(r=this._def).errorMap)||void 0===a?void 0:a.call(r,t,n).message)&&void 0!==o?o:n.defaultError;return"unrecognized_keys"===t.code?{message:null!==(i=E.errToObj(e).message)&&void 0!==i?i:s}:{message:s}}}:{}})}strip(){return new he({...this._def,unknownKeys:"strip"})}passthrough(){return new he({...this._def,unknownKeys:"passthrough"})}extend(e){return new he({...this._def,shape:()=>({...this._def.shape(),...e})})}merge(e){return new he({unknownKeys:e._def.unknownKeys,catchall:e._def.catchall,shape:()=>({...this._def.shape(),...e._def.shape()}),typeName:je.ZodObject})}setKey(e,t){return this.augment({[e]:t})}catchall(e){return new he({...this._def,catchall:e})}pick(e){const t={};return r.objectKeys(e).forEach((n=>{e[n]&&this.shape[n]&&(t[n]=this.shape[n])})),new he({...this._def,shape:()=>t})}omit(e){const t={};return r.objectKeys(this.shape).forEach((n=>{e[n]||(t[n]=this.shape[n])})),new he({...this._def,shape:()=>t})}deepPartial(){return pe(this)}partial(e){const t={};return r.objectKeys(this.shape).forEach((n=>{const r=this.shape[n];e&&!e[n]?t[n]=r:t[n]=r.optional()})),new he({...this._def,shape:()=>t})}required(e){const t={};return r.objectKeys(this.shape).forEach((n=>{if(e&&!e[n])t[n]=this.shape[n];else{let e=this.shape[n];for(;e instanceof Pe;)e=e._def.innerType;t[n]=e}})),new he({...this._def,shape:()=>t})}keyof(){return _e(r.objectKeys(this.shape))}}he.create=(e,t)=>new he({shape:()=>e,unknownKeys:"strip",catchall:ce.create(),typeName:je.ZodObject,...A(t)}),he.strictCreate=(e,t)=>new he({shape:()=>e,unknownKeys:"strict",catchall:ce.create(),typeName:je.ZodObject,...A(t)}),he.lazycreate=(e,t)=>new he({shape:e,unknownKeys:"strip",catchall:ce.create(),typeName:je.ZodObject,...A(t)});class fe extends R{_parse(e){const{ctx:t}=this._processInputParams(e),n=this._def.options;if(t.common.async)return Promise.all(n.map((async e=>{const n={...t,common:{...t.common,issues:[]},parent:null};return{result:await e._parseAsync({data:t.data,path:t.path,parent:n}),ctx:n}}))).then((function(e){for(const t of e)if("valid"===t.result.status)return t.result;for(const n of e)if("dirty"===n.result.status)return t.common.issues.push(...n.ctx.common.issues),n.result;const n=e.map((e=>new l(e.ctx.common.issues)));return h(t,{code:s.invalid_union,unionErrors:n}),m}));{let e;const r=[];for(const a of n){const n={...t,common:{...t.common,issues:[]},parent:null},o=a._parseSync({data:t.data,path:t.path,parent:n});if("valid"===o.status)return o;"dirty"!==o.status||e||(e={result:o,ctx:n}),n.common.issues.length&&r.push(n.common.issues)}if(e)return t.common.issues.push(...e.ctx.common.issues),e.result;const a=r.map((e=>new l(e)));return h(t,{code:s.invalid_union,unionErrors:a}),m}}get options(){return this._def.options}}fe.create=(e,t)=>new fe({options:e,typeName:je.ZodUnion,...A(t)});const me=e=>e instanceof Ee?me(e.schema):e instanceof Re?me(e.innerType()):e instanceof Se?[e.value]:e instanceof Ce?e.options:e instanceof Te?r.objectValues(e.enum):e instanceof Me?me(e._def.innerType):e instanceof oe?[void 0]:e instanceof ie?[null]:e instanceof Pe?[void 0,...me(e.unwrap())]:e instanceof De?[null,...me(e.unwrap())]:e instanceof Le||e instanceof Fe?me(e.unwrap()):e instanceof $e?me(e._def.innerType):[];class ge extends R{_parse(e){const{ctx:t}=this._processInputParams(e);if(t.parsedType!==o.object)return h(t,{code:s.invalid_type,expected:o.object,received:t.parsedType}),m;const n=this.discriminator,r=t.data[n],a=this.optionsMap.get(r);return a?t.common.async?a._parseAsync({data:t.data,path:t.path,parent:t}):a._parseSync({data:t.data,path:t.path,parent:t}):(h(t,{code:s.invalid_union_discriminator,options:Array.from(this.optionsMap.keys()),path:[n]}),m)}get discriminator(){return this._def.discriminator}get options(){return this._def.options}get optionsMap(){return this._def.optionsMap}static create(e,t,n){const r=new Map;for(const n of t){const t=me(n.shape[e]);if(!t.length)throw new Error(`A discriminator value for key \`${e}\` could not be extracted from all schema options`);for(const a of t){if(r.has(a))throw new Error(`Discriminator property ${String(e)} has duplicate value ${String(a)}`);r.set(a,n)}}return new ge({typeName:je.ZodDiscriminatedUnion,discriminator:e,options:t,optionsMap:r,...A(n)})}}function ve(e,t){const n=i(e),a=i(t);if(e===t)return{valid:!0,data:e};if(n===o.object&&a===o.object){const n=r.objectKeys(t),a=r.objectKeys(e).filter((e=>-1!==n.indexOf(e))),o={...e,...t};for(const n of a){const r=ve(e[n],t[n]);if(!r.valid)return{valid:!1};o[n]=r.data}return{valid:!0,data:o}}if(n===o.array&&a===o.array){if(e.length!==t.length)return{valid:!1};const n=[];for(let r=0;r<e.length;r++){const a=ve(e[r],t[r]);if(!a.valid)return{valid:!1};n.push(a.data)}return{valid:!0,data:n}}return n===o.date&&a===o.date&&+e==+t?{valid:!0,data:e}:{valid:!1}}class be extends R{_parse(e){const{status:t,ctx:n}=this._processInputParams(e),r=(e,r)=>{if(b(e)||b(r))return m;const a=ve(e.value,r.value);return a.valid?((y(e)||y(r))&&t.dirty(),{status:t.value,value:a.data}):(h(n,{code:s.invalid_intersection_types}),m)};return n.common.async?Promise.all([this._def.left._parseAsync({data:n.data,path:n.path,parent:n}),this._def.right._parseAsync({data:n.data,path:n.path,parent:n})]).then((([e,t])=>r(e,t))):r(this._def.left._parseSync({data:n.data,path:n.path,parent:n}),this._def.right._parseSync({data:n.data,path:n.path,parent:n}))}}be.create=(e,t,n)=>new be({left:e,right:t,typeName:je.ZodIntersection,...A(n)});class ye extends R{_parse(e){const{status:t,ctx:n}=this._processInputParams(e);if(n.parsedType!==o.array)return h(n,{code:s.invalid_type,expected:o.array,received:n.parsedType}),m;if(n.data.length<this._def.items.length)return h(n,{code:s.too_small,minimum:this._def.items.length,inclusive:!0,exact:!1,type:"array"}),m;!this._def.rest&&n.data.length>this._def.items.length&&(h(n,{code:s.too_big,maximum:this._def.items.length,inclusive:!0,exact:!1,type:"array"}),t.dirty());const r=[...n.data].map(((e,t)=>{const r=this._def.items[t]||this._def.rest;return r?r._parse(new C(n,e,n.path,t)):null})).filter((e=>!!e));return n.common.async?Promise.all(r).then((e=>f.mergeArray(t,e))):f.mergeArray(t,r)}get items(){return this._def.items}rest(e){return new ye({...this._def,rest:e})}}ye.create=(e,t)=>{if(!Array.isArray(e))throw new Error("You must pass an array of schemas to z.tuple([ ... ])");return new ye({items:e,typeName:je.ZodTuple,rest:null,...A(t)})};class we extends R{get keySchema(){return this._def.keyType}get valueSchema(){return this._def.valueType}_parse(e){const{status:t,ctx:n}=this._processInputParams(e);if(n.parsedType!==o.object)return h(n,{code:s.invalid_type,expected:o.object,received:n.parsedType}),m;const r=[],a=this._def.keyType,i=this._def.valueType;for(const e in n.data)r.push({key:a._parse(new C(n,e,n.path,e)),value:i._parse(new C(n,n.data[e],n.path,e)),alwaysSet:e in n.data});return n.common.async?f.mergeObjectAsync(t,r):f.mergeObjectSync(t,r)}get element(){return this._def.valueType}static create(e,t,n){return new we(t instanceof R?{keyType:e,valueType:t,typeName:je.ZodRecord,...A(n)}:{keyType:Y.create(),valueType:e,typeName:je.ZodRecord,...A(t)})}}class Oe extends R{get keySchema(){return this._def.keyType}get valueSchema(){return this._def.valueType}_parse(e){const{status:t,ctx:n}=this._processInputParams(e);if(n.parsedType!==o.map)return h(n,{code:s.invalid_type,expected:o.map,received:n.parsedType}),m;const r=this._def.keyType,a=this._def.valueType,i=[...n.data.entries()].map((([e,t],o)=>({key:r._parse(new C(n,e,n.path,[o,"key"])),value:a._parse(new C(n,t,n.path,[o,"value"]))})));if(n.common.async){const e=new Map;return Promise.resolve().then((async()=>{for(const n of i){const r=await n.key,a=await n.value;if("aborted"===r.status||"aborted"===a.status)return m;"dirty"!==r.status&&"dirty"!==a.status||t.dirty(),e.set(r.value,a.value)}return{status:t.value,value:e}}))}{const e=new Map;for(const n of i){const r=n.key,a=n.value;if("aborted"===r.status||"aborted"===a.status)return m;"dirty"!==r.status&&"dirty"!==a.status||t.dirty(),e.set(r.value,a.value)}return{status:t.value,value:e}}}}Oe.create=(e,t,n)=>new Oe({valueType:t,keyType:e,typeName:je.ZodMap,...A(n)});class ke extends R{_parse(e){const{status:t,ctx:n}=this._processInputParams(e);if(n.parsedType!==o.set)return h(n,{code:s.invalid_type,expected:o.set,received:n.parsedType}),m;const r=this._def;null!==r.minSize&&n.data.size<r.minSize.value&&(h(n,{code:s.too_small,minimum:r.minSize.value,type:"set",inclusive:!0,exact:!1,message:r.minSize.message}),t.dirty()),null!==r.maxSize&&n.data.size>r.maxSize.value&&(h(n,{code:s.too_big,maximum:r.maxSize.value,type:"set",inclusive:!0,exact:!1,message:r.maxSize.message}),t.dirty());const a=this._def.valueType;function i(e){const n=new Set;for(const r of e){if("aborted"===r.status)return m;"dirty"===r.status&&t.dirty(),n.add(r.value)}return{status:t.value,value:n}}const l=[...n.data.values()].map(((e,t)=>a._parse(new C(n,e,n.path,t))));return n.common.async?Promise.all(l).then((e=>i(e))):i(l)}min(e,t){return new ke({...this._def,minSize:{value:e,message:E.toString(t)}})}max(e,t){return new ke({...this._def,maxSize:{value:e,message:E.toString(t)}})}size(e,t){return this.min(e,t).max(e,t)}nonempty(e){return this.min(1,e)}}ke.create=(e,t)=>new ke({valueType:e,minSize:null,maxSize:null,typeName:je.ZodSet,...A(t)});class xe extends R{constructor(){super(...arguments),this.validate=this.implement}_parse(e){const{ctx:t}=this._processInputParams(e);if(t.parsedType!==o.function)return h(t,{code:s.invalid_type,expected:o.function,received:t.parsedType}),m;function n(e,n){return p({data:e,path:t.path,errorMaps:[t.common.contextualErrorMap,t.schemaErrorMap,d(),c].filter((e=>!!e)),issueData:{code:s.invalid_arguments,argumentsError:n}})}function r(e,n){return p({data:e,path:t.path,errorMaps:[t.common.contextualErrorMap,t.schemaErrorMap,d(),c].filter((e=>!!e)),issueData:{code:s.invalid_return_type,returnTypeError:n}})}const a={errorMap:t.common.contextualErrorMap},i=t.data;if(this._def.returns instanceof Ae){const e=this;return v((async function(...t){const o=new l([]),s=await e._def.args.parseAsync(t,a).catch((e=>{throw o.addIssue(n(t,e)),o})),c=await Reflect.apply(i,this,s);return await e._def.returns._def.type.parseAsync(c,a).catch((e=>{throw o.addIssue(r(c,e)),o}))}))}{const e=this;return v((function(...t){const o=e._def.args.safeParse(t,a);if(!o.success)throw new l([n(t,o.error)]);const s=Reflect.apply(i,this,o.data),c=e._def.returns.safeParse(s,a);if(!c.success)throw new l([r(s,c.error)]);return c.data}))}}parameters(){return this._def.args}returnType(){return this._def.returns}args(...e){return new xe({...this._def,args:ye.create(e).rest(le.create())})}returns(e){return new xe({...this._def,returns:e})}implement(e){return this.parse(e)}strictImplement(e){return this.parse(e)}static create(e,t,n){return new xe({args:e||ye.create([]).rest(le.create()),returns:t||le.create(),typeName:je.ZodFunction,...A(n)})}}class Ee extends R{get schema(){return this._def.getter()}_parse(e){const{ctx:t}=this._processInputParams(e);return this._def.getter()._parse({data:t.data,path:t.path,parent:t})}}Ee.create=(e,t)=>new Ee({getter:e,typeName:je.ZodLazy,...A(t)});class Se extends R{_parse(e){if(e.data!==this._def.value){const t=this._getOrReturnCtx(e);return h(t,{received:t.data,code:s.invalid_literal,expected:this._def.value}),m}return{status:"valid",value:e.data}}get value(){return this._def.value}}function _e(e,t){return new Ce({values:e,typeName:je.ZodEnum,...A(t)})}Se.create=(e,t)=>new Se({value:e,typeName:je.ZodLiteral,...A(t)});class Ce extends R{constructor(){super(...arguments),S.set(this,void 0)}_parse(e){if("string"!=typeof e.data){const t=this._getOrReturnCtx(e),n=this._def.values;return h(t,{expected:r.joinValues(n),received:t.parsedType,code:s.invalid_type}),m}if(k(this,S,"f")||x(this,S,new Set(this._def.values),"f"),!k(this,S,"f").has(e.data)){const t=this._getOrReturnCtx(e),n=this._def.values;return h(t,{received:t.data,code:s.invalid_enum_value,options:n}),m}return v(e.data)}get options(){return this._def.values}get enum(){const e={};for(const t of this._def.values)e[t]=t;return e}get Values(){const e={};for(const t of this._def.values)e[t]=t;return e}get Enum(){const e={};for(const t of this._def.values)e[t]=t;return e}extract(e,t=this._def){return Ce.create(e,{...this._def,...t})}exclude(e,t=this._def){return Ce.create(this.options.filter((t=>!e.includes(t))),{...this._def,...t})}}S=new WeakMap,Ce.create=_e;class Te extends R{constructor(){super(...arguments),_.set(this,void 0)}_parse(e){const t=r.getValidEnumValues(this._def.values),n=this._getOrReturnCtx(e);if(n.parsedType!==o.string&&n.parsedType!==o.number){const e=r.objectValues(t);return h(n,{expected:r.joinValues(e),received:n.parsedType,code:s.invalid_type}),m}if(k(this,_,"f")||x(this,_,new Set(r.getValidEnumValues(this._def.values)),"f"),!k(this,_,"f").has(e.data)){const e=r.objectValues(t);return h(n,{received:n.data,code:s.invalid_enum_value,options:e}),m}return v(e.data)}get enum(){return this._def.values}}_=new WeakMap,Te.create=(e,t)=>new Te({values:e,typeName:je.ZodNativeEnum,...A(t)});class Ae extends R{unwrap(){return this._def.type}_parse(e){const{ctx:t}=this._processInputParams(e);if(t.parsedType!==o.promise&&!1===t.common.async)return h(t,{code:s.invalid_type,expected:o.promise,received:t.parsedType}),m;const n=t.parsedType===o.promise?t.data:Promise.resolve(t.data);return v(n.then((e=>this._def.type.parseAsync(e,{path:t.path,errorMap:t.common.contextualErrorMap}))))}}Ae.create=(e,t)=>new Ae({type:e,typeName:je.ZodPromise,...A(t)});class Re extends R{innerType(){return this._def.schema}sourceType(){return this._def.schema._def.typeName===je.ZodEffects?this._def.schema.sourceType():this._def.schema}_parse(e){const{status:t,ctx:n}=this._processInputParams(e),a=this._def.effect||null,o={addIssue:e=>{h(n,e),e.fatal?t.abort():t.dirty()},get path(){return n.path}};if(o.addIssue=o.addIssue.bind(o),"preprocess"===a.type){const e=a.transform(n.data,o);if(n.common.async)return Promise.resolve(e).then((async e=>{if("aborted"===t.value)return m;const r=await this._def.schema._parseAsync({data:e,path:n.path,parent:n});return"aborted"===r.status?m:"dirty"===r.status||"dirty"===t.value?g(r.value):r}));{if("aborted"===t.value)return m;const r=this._def.schema._parseSync({data:e,path:n.path,parent:n});return"aborted"===r.status?m:"dirty"===r.status||"dirty"===t.value?g(r.value):r}}if("refinement"===a.type){const e=e=>{const t=a.refinement(e,o);if(n.common.async)return Promise.resolve(t);if(t instanceof Promise)throw new Error("Async refinement encountered during synchronous parse operation. Use .parseAsync instead.");return e};if(!1===n.common.async){const r=this._def.schema._parseSync({data:n.data,path:n.path,parent:n});return"aborted"===r.status?m:("dirty"===r.status&&t.dirty(),e(r.value),{status:t.value,value:r.value})}return this._def.schema._parseAsync({data:n.data,path:n.path,parent:n}).then((n=>"aborted"===n.status?m:("dirty"===n.status&&t.dirty(),e(n.value).then((()=>({status:t.value,value:n.value}))))))}if("transform"===a.type){if(!1===n.common.async){const e=this._def.schema._parseSync({data:n.data,path:n.path,parent:n});if(!w(e))return e;const r=a.transform(e.value,o);if(r instanceof Promise)throw new Error("Asynchronous transform encountered during synchronous parse operation. Use .parseAsync instead.");return{status:t.value,value:r}}return this._def.schema._parseAsync({data:n.data,path:n.path,parent:n}).then((e=>w(e)?Promise.resolve(a.transform(e.value,o)).then((e=>({status:t.value,value:e}))):e))}r.assertNever(a)}}Re.create=(e,t,n)=>new Re({schema:e,typeName:je.ZodEffects,effect:t,...A(n)}),Re.createWithPreprocess=(e,t,n)=>new Re({schema:t,effect:{type:"preprocess",transform:e},typeName:je.ZodEffects,...A(n)});class Pe extends R{_parse(e){return this._getType(e)===o.undefined?v(void 0):this._def.innerType._parse(e)}unwrap(){return this._def.innerType}}Pe.create=(e,t)=>new Pe({innerType:e,typeName:je.ZodOptional,...A(t)});class De extends R{_parse(e){return this._getType(e)===o.null?v(null):this._def.innerType._parse(e)}unwrap(){return this._def.innerType}}De.create=(e,t)=>new De({innerType:e,typeName:je.ZodNullable,...A(t)});class Me extends R{_parse(e){const{ctx:t}=this._processInputParams(e);let n=t.data;return t.parsedType===o.undefined&&(n=this._def.defaultValue()),this._def.innerType._parse({data:n,path:t.path,parent:t})}removeDefault(){return this._def.innerType}}Me.create=(e,t)=>new Me({innerType:e,typeName:je.ZodDefault,defaultValue:"function"==typeof t.default?t.default:()=>t.default,...A(t)});class $e extends R{_parse(e){const{ctx:t}=this._processInputParams(e),n={...t,common:{...t.common,issues:[]}},r=this._def.innerType._parse({data:n.data,path:n.path,parent:{...n}});return O(r)?r.then((e=>({status:"valid",value:"valid"===e.status?e.value:this._def.catchValue({get error(){return new l(n.common.issues)},input:n.data})}))):{status:"valid",value:"valid"===r.status?r.value:this._def.catchValue({get error(){return new l(n.common.issues)},input:n.data})}}removeCatch(){return this._def.innerType}}$e.create=(e,t)=>new $e({innerType:e,typeName:je.ZodCatch,catchValue:"function"==typeof t.catch?t.catch:()=>t.catch,...A(t)});class Ne extends R{_parse(e){if(this._getType(e)!==o.nan){const t=this._getOrReturnCtx(e);return h(t,{code:s.invalid_type,expected:o.nan,received:t.parsedType}),m}return{status:"valid",value:e.data}}}Ne.create=e=>new Ne({typeName:je.ZodNaN,...A(e)});const Ie=Symbol("zod_brand");class Le extends R{_parse(e){const{ctx:t}=this._processInputParams(e),n=t.data;return this._def.type._parse({data:n,path:t.path,parent:t})}unwrap(){return this._def.type}}class Be extends R{_parse(e){const{status:t,ctx:n}=this._processInputParams(e);if(n.common.async)return(async()=>{const e=await this._def.in._parseAsync({data:n.data,path:n.path,parent:n});return"aborted"===e.status?m:"dirty"===e.status?(t.dirty(),g(e.value)):this._def.out._parseAsync({data:e.value,path:n.path,parent:n})})();{const e=this._def.in._parseSync({data:n.data,path:n.path,parent:n});return"aborted"===e.status?m:"dirty"===e.status?(t.dirty(),{status:"dirty",value:e.value}):this._def.out._parseSync({data:e.value,path:n.path,parent:n})}}static create(e,t){return new Be({in:e,out:t,typeName:je.ZodPipeline})}}class Fe extends R{_parse(e){const t=this._def.innerType._parse(e),n=e=>(w(e)&&(e.value=Object.freeze(e.value)),e);return O(t)?t.then((e=>n(e))):n(t)}unwrap(){return this._def.innerType}}function Qe(e,t){const n="function"==typeof e?e(t):"string"==typeof e?{message:e}:e;return"string"==typeof n?{message:n}:n}function Ue(e,t={},n){return e?se.create().superRefine(((r,a)=>{var o,i;const s=e(r);if(s instanceof Promise)return s.then((e=>{var o,i;if(!e){const e=Qe(t,r),s=null===(i=null!==(o=e.fatal)&&void 0!==o?o:n)||void 0===i||i;a.addIssue({code:"custom",...e,fatal:s})}}));if(!s){const e=Qe(t,r),s=null===(i=null!==(o=e.fatal)&&void 0!==o?o:n)||void 0===i||i;a.addIssue({code:"custom",...e,fatal:s})}})):se.create()}Fe.create=(e,t)=>new Fe({innerType:e,typeName:je.ZodReadonly,...A(t)});const ze={object:he.lazycreate};var je;!function(e){e.ZodString="ZodString",e.ZodNumber="ZodNumber",e.ZodNaN="ZodNaN",e.ZodBigInt="ZodBigInt",e.ZodBoolean="ZodBoolean",e.ZodDate="ZodDate",e.ZodSymbol="ZodSymbol",e.ZodUndefined="ZodUndefined",e.ZodNull="ZodNull",e.ZodAny="ZodAny",e.ZodUnknown="ZodUnknown",e.ZodNever="ZodNever",e.ZodVoid="ZodVoid",e.ZodArray="ZodArray",e.ZodObject="ZodObject",e.ZodUnion="ZodUnion",e.ZodDiscriminatedUnion="ZodDiscriminatedUnion",e.ZodIntersection="ZodIntersection",e.ZodTuple="ZodTuple",e.ZodRecord="ZodRecord",e.ZodMap="ZodMap",e.ZodSet="ZodSet",e.ZodFunction="ZodFunction",e.ZodLazy="ZodLazy",e.ZodLiteral="ZodLiteral",e.ZodEnum="ZodEnum",e.ZodEffects="ZodEffects",e.ZodNativeEnum="ZodNativeEnum",e.ZodOptional="ZodOptional",e.ZodNullable="ZodNullable",e.ZodDefault="ZodDefault",e.ZodCatch="ZodCatch",e.ZodPromise="ZodPromise",e.ZodBranded="ZodBranded",e.ZodPipeline="ZodPipeline",e.ZodReadonly="ZodReadonly"}(je||(je={}));const We=Y.create,Xe=ee.create,qe=Ne.create,Ve=te.create,He=ne.create,Ze=re.create,Ge=ae.create,Ke=oe.create,Ye=ie.create,Je=se.create,et=le.create,tt=ce.create,nt=ue.create,rt=de.create,at=he.create,ot=he.strictCreate,it=fe.create,st=ge.create,lt=be.create,ct=ye.create,ut=we.create,dt=Oe.create,pt=ke.create,ht=xe.create,ft=Ee.create,mt=Se.create,gt=Ce.create,vt=Te.create,bt=Ae.create,yt=Re.create,wt=Pe.create,Ot=De.create,kt=Re.createWithPreprocess,xt=Be.create,Et={string:e=>Y.create({...e,coerce:!0}),number:e=>ee.create({...e,coerce:!0}),boolean:e=>ne.create({...e,coerce:!0}),bigint:e=>te.create({...e,coerce:!0}),date:e=>re.create({...e,coerce:!0})},St=m;var _t=Object.freeze({__proto__:null,defaultErrorMap:c,setErrorMap:function(e){u=e},getErrorMap:d,makeIssue:p,EMPTY_PATH:[],addIssueToContext:h,ParseStatus:f,INVALID:m,DIRTY:g,OK:v,isAborted:b,isDirty:y,isValid:w,isAsync:O,get util(){return r},get objectUtil(){return a},ZodParsedType:o,getParsedType:i,ZodType:R,datetimeRegex:Z,ZodString:Y,ZodNumber:ee,ZodBigInt:te,ZodBoolean:ne,ZodDate:re,ZodSymbol:ae,ZodUndefined:oe,ZodNull:ie,ZodAny:se,ZodUnknown:le,ZodNever:ce,ZodVoid:ue,ZodArray:de,ZodObject:he,ZodUnion:fe,ZodDiscriminatedUnion:ge,ZodIntersection:be,ZodTuple:ye,ZodRecord:we,ZodMap:Oe,ZodSet:ke,ZodFunction:xe,ZodLazy:Ee,ZodLiteral:Se,ZodEnum:Ce,ZodNativeEnum:Te,ZodPromise:Ae,ZodEffects:Re,ZodTransformer:Re,ZodOptional:Pe,ZodNullable:De,ZodDefault:Me,ZodCatch:$e,ZodNaN:Ne,BRAND:Ie,ZodBranded:Le,ZodPipeline:Be,ZodReadonly:Fe,custom:Ue,Schema:R,ZodSchema:R,late:ze,get ZodFirstPartyTypeKind(){return je},coerce:Et,any:Je,array:rt,bigint:Ve,boolean:He,date:Ze,discriminatedUnion:st,effect:yt,enum:gt,function:ht,instanceof:(e,t={message:`Input not instance of ${e.name}`})=>Ue((t=>t instanceof e),t),intersection:lt,lazy:ft,literal:mt,map:dt,nan:qe,nativeEnum:vt,never:tt,null:Ye,nullable:Ot,number:Xe,object:at,oboolean:()=>He().optional(),onumber:()=>Xe().optional(),optional:wt,ostring:()=>We().optional(),pipeline:xt,preprocess:kt,promise:bt,record:ut,set:pt,strictObject:ot,string:We,symbol:Ge,transformer:yt,tuple:ct,undefined:Ke,union:it,unknown:et,void:nt,NEVER:St,ZodIssueCode:s,quotelessJson:e=>JSON.stringify(e,null,2).replace(/"([^"]+)":/g,"$1:"),ZodError:l})},4477:(e,t)=>{"use strict";function n(e,t){var n=e.length;e.push(t);e:for(;0<n;){var r=n-1>>>1,a=e[r];if(!(0<o(a,t)))break e;e[r]=t,e[n]=a,n=r}}function r(e){return 0===e.length?null:e[0]}function a(e){if(0===e.length)return null;var t=e[0],n=e.pop();if(n!==t){e[0]=n;e:for(var r=0,a=e.length,i=a>>>1;r<i;){var s=2*(r+1)-1,l=e[s],c=s+1,u=e[c];if(0>o(l,n))c<a&&0>o(u,l)?(e[r]=u,e[c]=n,r=c):(e[r]=l,e[s]=n,r=s);else{if(!(c<a&&0>o(u,n)))break e;e[r]=u,e[c]=n,r=c}}}return t}function o(e,t){var n=e.sortIndex-t.sortIndex;return 0!==n?n:e.id-t.id}if(t.unstable_now=void 0,"object"==typeof performance&&"function"==typeof performance.now){var i=performance;t.unstable_now=function(){return i.now()}}else{var s=Date,l=s.now();t.unstable_now=function(){return s.now()-l}}var c=[],u=[],d=1,p=null,h=3,f=!1,m=!1,g=!1,v="function"==typeof setTimeout?setTimeout:null,b="function"==typeof clearTimeout?clearTimeout:null,y="undefined"!=typeof setImmediate?setImmediate:null;function w(e){for(var t=r(u);null!==t;){if(null===t.callback)a(u);else{if(!(t.startTime<=e))break;a(u),t.sortIndex=t.expirationTime,n(c,t)}t=r(u)}}function O(e){if(g=!1,w(e),!m)if(null!==r(c))m=!0,P();else{var t=r(u);null!==t&&D(O,t.startTime-e)}}var k,x=!1,E=-1,S=5,_=-1;function C(){return!(t.unstable_now()-_<S)}function T(){if(x){var e=t.unstable_now();_=e;var n=!0;try{e:{m=!1,g&&(g=!1,b(E),E=-1),f=!0;var o=h;try{t:{for(w(e),p=r(c);null!==p&&!(p.expirationTime>e&&C());){var i=p.callback;if("function"==typeof i){p.callback=null,h=p.priorityLevel;var s=i(p.expirationTime<=e);if(e=t.unstable_now(),"function"==typeof s){p.callback=s,w(e),n=!0;break t}p===r(c)&&a(c),w(e)}else a(c);p=r(c)}if(null!==p)n=!0;else{var l=r(u);null!==l&&D(O,l.startTime-e),n=!1}}break e}finally{p=null,h=o,f=!1}n=void 0}}finally{n?k():x=!1}}}if("function"==typeof y)k=function(){y(T)};else if("undefined"!=typeof MessageChannel){var A=new MessageChannel,R=A.port2;A.port1.onmessage=T,k=function(){R.postMessage(null)}}else k=function(){v(T,0)};function P(){x||(x=!0,k())}function D(e,n){E=v((function(){e(t.unstable_now())}),n)}t.unstable_IdlePriority=5,t.unstable_ImmediatePriority=1,t.unstable_LowPriority=4,t.unstable_NormalPriority=3,t.unstable_Profiling=null,t.unstable_UserBlockingPriority=2,t.unstable_cancelCallback=function(e){e.callback=null},t.unstable_continueExecution=function(){m||f||(m=!0,P())},t.unstable_forceFrameRate=function(e){0>e||125<e?console.error("forceFrameRate takes a positive int between 0 and 125, forcing frame rates higher than 125 fps is not supported"):S=0<e?Math.floor(1e3/e):5},t.unstable_getCurrentPriorityLevel=function(){return h},t.unstable_getFirstCallbackNode=function(){return r(c)},t.unstable_next=function(e){switch(h){case 1:case 2:case 3:var t=3;break;default:t=h}var n=h;h=t;try{return e()}finally{h=n}},t.unstable_pauseExecution=function(){},t.unstable_requestPaint=function(){},t.unstable_runWithPriority=function(e,t){switch(e){case 1:case 2:case 3:case 4:case 5:break;default:e=3}var n=h;h=e;try{return t()}finally{h=n}},t.unstable_scheduleCallback=function(e,a,o){var i=t.unstable_now();switch(o="object"==typeof o&&null!==o&&"number"==typeof(o=o.delay)&&0<o?i+o:i,e){case 1:var s=-1;break;case 2:s=250;break;case 5:s=1073741823;break;case 4:s=1e4;break;default:s=5e3}return e={id:d++,callback:a,priorityLevel:e,startTime:o,expirationTime:s=o+s,sortIndex:-1},o>i?(e.sortIndex=o,n(u,e),null===r(c)&&e===r(u)&&(g?(b(E),E=-1):g=!0,D(O,o-i))):(e.sortIndex=s,n(c,e),m||f||(m=!0,P())),e},t.unstable_shouldYield=C,t.unstable_wrapCallback=function(e){var t=h;return function(){var n=h;h=t;try{return e.apply(this,arguments)}finally{h=n}}}},4486:(e,t,n)=>{"use strict";n.d(t,{A:()=>l});var r=n(641),a=n(953),o=n(8839),i=n(1266),s=n(6584);const l=(0,r.pM)({__name:"SidebarButton",props:{click:{type:Function},hotkey:{}},setup(e){const t=e,{layout:n}=(0,s.g)(),l=()=>{t.click()};return(e,t)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(i.RO),{class:"bg-b-1 text-c-1 hover:bg-b-2 group relative w-auto md:w-full border-1/2 px-2 py-1 md:p-1.5 h-auto",icon:"Plus",variant:"outlined",onClick:l},{default:(0,r.k6)((()=>[(0,r.RG)(e.$slots,"title"),e.hotkey&&"desktop"===(0,a.R1)(n)?((0,r.uX)(),(0,r.Wv)(o.A,{key:0,class:"hidden md:block absolute right-2 group-hover:opacity-80 text-c-2 add-item-hotkey",hotkey:e.hotkey},null,8,["hotkey"])):(0,r.Q3)("",!0)])),_:3}))}})},4565:(e,t,n)=>{"use strict";n.d(t,{t:()=>r});const r={html:"http://www.w3.org/1999/xhtml",mathml:"http://www.w3.org/1998/Math/MathML",svg:"http://www.w3.org/2000/svg",xlink:"http://www.w3.org/1999/xlink",xml:"http://www.w3.org/XML/1998/namespace",xmlns:"http://www.w3.org/2000/xmlns/"}},4684:(e,t,n)=>{"use strict";n.d(t,{iX:()=>Me,ou:()=>r,V_:()=>X,qy:()=>a,qg:()=>Pt,oH:()=>Dt});var r={};n.r(r),n.d(r,{k:()=>m,r:()=>v});var a={};n.r(a),n.d(a,{ATTRS:()=>T,DOCUMENT_MODE:()=>A,NS:()=>C,NUMBERED_HEADERS:()=>U,SPECIAL_ELEMENTS:()=>Q,TAG_ID:()=>P,TAG_NAMES:()=>R,getTagID:()=>B,hasUnescapedText:()=>j});const o=new Set([65534,65535,131070,131071,196606,196607,262142,262143,327678,327679,393214,393215,458750,458751,524286,524287,589822,589823,655358,655359,720894,720895,786430,786431,851966,851967,917502,917503,983038,983039,1048574,1048575,1114110,1114111]),i="�";var s;!function(e){e[e.EOF=-1]="EOF",e[e.NULL=0]="NULL",e[e.TABULATION=9]="TABULATION",e[e.CARRIAGE_RETURN=13]="CARRIAGE_RETURN",e[e.LINE_FEED=10]="LINE_FEED",e[e.FORM_FEED=12]="FORM_FEED",e[e.SPACE=32]="SPACE",e[e.EXCLAMATION_MARK=33]="EXCLAMATION_MARK",e[e.QUOTATION_MARK=34]="QUOTATION_MARK",e[e.AMPERSAND=38]="AMPERSAND",e[e.APOSTROPHE=39]="APOSTROPHE",e[e.HYPHEN_MINUS=45]="HYPHEN_MINUS",e[e.SOLIDUS=47]="SOLIDUS",e[e.DIGIT_0=48]="DIGIT_0",e[e.DIGIT_9=57]="DIGIT_9",e[e.SEMICOLON=59]="SEMICOLON",e[e.LESS_THAN_SIGN=60]="LESS_THAN_SIGN",e[e.EQUALS_SIGN=61]="EQUALS_SIGN",e[e.GREATER_THAN_SIGN=62]="GREATER_THAN_SIGN",e[e.QUESTION_MARK=63]="QUESTION_MARK",e[e.LATIN_CAPITAL_A=65]="LATIN_CAPITAL_A",e[e.LATIN_CAPITAL_Z=90]="LATIN_CAPITAL_Z",e[e.RIGHT_SQUARE_BRACKET=93]="RIGHT_SQUARE_BRACKET",e[e.GRAVE_ACCENT=96]="GRAVE_ACCENT",e[e.LATIN_SMALL_A=97]="LATIN_SMALL_A",e[e.LATIN_SMALL_Z=122]="LATIN_SMALL_Z"}(s||(s={}));const l="[CDATA[",c="doctype",u="script";function d(e){return e>=55296&&e<=57343}function p(e){return 32!==e&&10!==e&&13!==e&&9!==e&&12!==e&&e>=1&&e<=31||e>=127&&e<=159}function h(e){return e>=64976&&e<=65007||o.has(e)}var f,m;!function(e){e.controlCharacterInInputStream="control-character-in-input-stream",e.noncharacterInInputStream="noncharacter-in-input-stream",e.surrogateInInputStream="surrogate-in-input-stream",e.nonVoidHtmlElementStartTagWithTrailingSolidus="non-void-html-element-start-tag-with-trailing-solidus",e.endTagWithAttributes="end-tag-with-attributes",e.endTagWithTrailingSolidus="end-tag-with-trailing-solidus",e.unexpectedSolidusInTag="unexpected-solidus-in-tag",e.unexpectedNullCharacter="unexpected-null-character",e.unexpectedQuestionMarkInsteadOfTagName="unexpected-question-mark-instead-of-tag-name",e.invalidFirstCharacterOfTagName="invalid-first-character-of-tag-name",e.unexpectedEqualsSignBeforeAttributeName="unexpected-equals-sign-before-attribute-name",e.missingEndTagName="missing-end-tag-name",e.unexpectedCharacterInAttributeName="unexpected-character-in-attribute-name",e.unknownNamedCharacterReference="unknown-named-character-reference",e.missingSemicolonAfterCharacterReference="missing-semicolon-after-character-reference",e.unexpectedCharacterAfterDoctypeSystemIdentifier="unexpected-character-after-doctype-system-identifier",e.unexpectedCharacterInUnquotedAttributeValue="unexpected-character-in-unquoted-attribute-value",e.eofBeforeTagName="eof-before-tag-name",e.eofInTag="eof-in-tag",e.missingAttributeValue="missing-attribute-value",e.missingWhitespaceBetweenAttributes="missing-whitespace-between-attributes",e.missingWhitespaceAfterDoctypePublicKeyword="missing-whitespace-after-doctype-public-keyword",e.missingWhitespaceBetweenDoctypePublicAndSystemIdentifiers="missing-whitespace-between-doctype-public-and-system-identifiers",e.missingWhitespaceAfterDoctypeSystemKeyword="missing-whitespace-after-doctype-system-keyword",e.missingQuoteBeforeDoctypePublicIdentifier="missing-quote-before-doctype-public-identifier",e.missingQuoteBeforeDoctypeSystemIdentifier="missing-quote-before-doctype-system-identifier",e.missingDoctypePublicIdentifier="missing-doctype-public-identifier",e.missingDoctypeSystemIdentifier="missing-doctype-system-identifier",e.abruptDoctypePublicIdentifier="abrupt-doctype-public-identifier",e.abruptDoctypeSystemIdentifier="abrupt-doctype-system-identifier",e.cdataInHtmlContent="cdata-in-html-content",e.incorrectlyOpenedComment="incorrectly-opened-comment",e.eofInScriptHtmlCommentLikeText="eof-in-script-html-comment-like-text",e.eofInDoctype="eof-in-doctype",e.nestedComment="nested-comment",e.abruptClosingOfEmptyComment="abrupt-closing-of-empty-comment",e.eofInComment="eof-in-comment",e.incorrectlyClosedComment="incorrectly-closed-comment",e.eofInCdata="eof-in-cdata",e.absenceOfDigitsInNumericCharacterReference="absence-of-digits-in-numeric-character-reference",e.nullCharacterReference="null-character-reference",e.surrogateCharacterReference="surrogate-character-reference",e.characterReferenceOutsideUnicodeRange="character-reference-outside-unicode-range",e.controlCharacterReference="control-character-reference",e.noncharacterCharacterReference="noncharacter-character-reference",e.missingWhitespaceBeforeDoctypeName="missing-whitespace-before-doctype-name",e.missingDoctypeName="missing-doctype-name",e.invalidCharacterSequenceAfterDoctypeName="invalid-character-sequence-after-doctype-name",e.duplicateAttribute="duplicate-attribute",e.nonConformingDoctype="non-conforming-doctype",e.missingDoctype="missing-doctype",e.misplacedDoctype="misplaced-doctype",e.endTagWithoutMatchingOpenElement="end-tag-without-matching-open-element",e.closingOfElementWithOpenChildElements="closing-of-element-with-open-child-elements",e.disallowedContentInNoscriptInHead="disallowed-content-in-noscript-in-head",e.openElementsLeftAfterEof="open-elements-left-after-eof",e.abandonedHeadElementChild="abandoned-head-element-child",e.misplacedStartTagForHeadElement="misplaced-start-tag-for-head-element",e.nestedNoscriptInHead="nested-noscript-in-head",e.eofInElementThatCanContainOnlyText="eof-in-element-that-can-contain-only-text"}(f||(f={}));class g{constructor(e){this.handler=e,this.html="",this.pos=-1,this.lastGapPos=-2,this.gapStack=[],this.skipNextNewLine=!1,this.lastChunkWritten=!1,this.endOfChunkHit=!1,this.bufferWaterline=65536,this.isEol=!1,this.lineStartPos=0,this.droppedBufferSize=0,this.line=1,this.lastErrOffset=-1}get col(){return this.pos-this.lineStartPos+Number(this.lastGapPos!==this.pos)}get offset(){return this.droppedBufferSize+this.pos}getError(e,t){const{line:n,col:r,offset:a}=this,o=r+t,i=a+t;return{code:e,startLine:n,endLine:n,startCol:o,endCol:o,startOffset:i,endOffset:i}}_err(e){this.handler.onParseError&&this.lastErrOffset!==this.offset&&(this.lastErrOffset=this.offset,this.handler.onParseError(this.getError(e,0)))}_addGap(){this.gapStack.push(this.lastGapPos),this.lastGapPos=this.pos}_processSurrogate(e){if(this.pos!==this.html.length-1){const t=this.html.charCodeAt(this.pos+1);if(function(e){return e>=56320&&e<=57343}(t))return this.pos++,this._addGap(),1024*(e-55296)+9216+t}else if(!this.lastChunkWritten)return this.endOfChunkHit=!0,s.EOF;return this._err(f.surrogateInInputStream),e}willDropParsedChunk(){return this.pos>this.bufferWaterline}dropParsedChunk(){this.willDropParsedChunk()&&(this.html=this.html.substring(this.pos),this.lineStartPos-=this.pos,this.droppedBufferSize+=this.pos,this.pos=0,this.lastGapPos=-2,this.gapStack.length=0)}write(e,t){this.html.length>0?this.html+=e:this.html=e,this.endOfChunkHit=!1,this.lastChunkWritten=t}insertHtmlAtCurrentPos(e){this.html=this.html.substring(0,this.pos+1)+e+this.html.substring(this.pos+1),this.endOfChunkHit=!1}startsWith(e,t){if(this.pos+e.length>this.html.length)return this.endOfChunkHit=!this.lastChunkWritten,!1;if(t)return this.html.startsWith(e,this.pos);for(let t=0;t<e.length;t++)if((32|this.html.charCodeAt(this.pos+t))!==e.charCodeAt(t))return!1;return!0}peek(e){const t=this.pos+e;if(t>=this.html.length)return this.endOfChunkHit=!this.lastChunkWritten,s.EOF;const n=this.html.charCodeAt(t);return n===s.CARRIAGE_RETURN?s.LINE_FEED:n}advance(){if(this.pos++,this.isEol&&(this.isEol=!1,this.line++,this.lineStartPos=this.pos),this.pos>=this.html.length)return this.endOfChunkHit=!this.lastChunkWritten,s.EOF;let e=this.html.charCodeAt(this.pos);return e===s.CARRIAGE_RETURN?(this.isEol=!0,this.skipNextNewLine=!0,s.LINE_FEED):e===s.LINE_FEED&&(this.isEol=!0,this.skipNextNewLine)?(this.line--,this.skipNextNewLine=!1,this._addGap(),this.advance()):(this.skipNextNewLine=!1,d(e)&&(e=this._processSurrogate(e)),null===this.handler.onParseError||e>31&&e<127||e===s.LINE_FEED||e===s.CARRIAGE_RETURN||e>159&&e<64976||this._checkForProblematicCharacters(e),e)}_checkForProblematicCharacters(e){p(e)?this._err(f.controlCharacterInInputStream):h(e)&&this._err(f.noncharacterInInputStream)}retreat(e){for(this.pos-=e;this.pos<this.lastGapPos;)this.lastGapPos=this.gapStack.pop(),this.pos--;this.isEol=!1}}function v(e,t){for(let n=e.attrs.length-1;n>=0;n--)if(e.attrs[n].name===t)return e.attrs[n].value;return null}!function(e){e[e.CHARACTER=0]="CHARACTER",e[e.NULL_CHARACTER=1]="NULL_CHARACTER",e[e.WHITESPACE_CHARACTER=2]="WHITESPACE_CHARACTER",e[e.START_TAG=3]="START_TAG",e[e.END_TAG=4]="END_TAG",e[e.COMMENT=5]="COMMENT",e[e.DOCTYPE=6]="DOCTYPE",e[e.EOF=7]="EOF",e[e.HIBERNATION=8]="HIBERNATION"}(m||(m={}));const b=new Uint16Array('ᵁ<Õıʊҝջאٵ۞ޢߖࠏ੊ઑඡ๭༉༦჊ረዡᐕᒝᓃᓟᔥ\0\0\0\0\0\0ᕫᛍᦍᰒᷝ὾⁠↰⊍⏀⏻⑂⠤⤒ⴈ⹈⿎〖㊺㘹㞬㣾㨨㩱㫠㬮ࠀEMabcfglmnoprstu\\bfms¦³¹ÈÏlig耻Æ䃆P耻&䀦cute耻Á䃁reve;䄂Āiyx}rc耻Â䃂;䐐r;쀀𝔄rave耻À䃀pha;䎑acr;䄀d;橓Āgp¡on;䄄f;쀀𝔸plyFunction;恡ing耻Å䃅Ācs¾Ãr;쀀𝒜ign;扔ilde耻Ã䃃ml耻Ä䃄ЀaceforsuåûþėĜĢħĪĀcrêòkslash;或Ŷöø;櫧ed;挆y;䐑ƀcrtąċĔause;戵noullis;愬a;䎒r;쀀𝔅pf;쀀𝔹eve;䋘còēmpeq;扎܀HOacdefhilorsuōőŖƀƞƢƵƷƺǜȕɳɸɾcy;䐧PY耻©䂩ƀcpyŝŢźute;䄆Ā;iŧŨ拒talDifferentialD;慅leys;愭ȀaeioƉƎƔƘron;䄌dil耻Ç䃇rc;䄈nint;戰ot;䄊ĀdnƧƭilla;䂸terDot;䂷òſi;䎧rcleȀDMPTǇǋǑǖot;抙inus;抖lus;投imes;抗oĀcsǢǸkwiseContourIntegral;戲eCurlyĀDQȃȏoubleQuote;思uote;怙ȀlnpuȞȨɇɕonĀ;eȥȦ户;橴ƀgitȯȶȺruent;扡nt;戯ourIntegral;戮ĀfrɌɎ;愂oduct;成nterClockwiseContourIntegral;戳oss;樯cr;쀀𝒞pĀ;Cʄʅ拓ap;才րDJSZacefiosʠʬʰʴʸˋ˗ˡ˦̳ҍĀ;oŹʥtrahd;椑cy;䐂cy;䐅cy;䐏ƀgrsʿ˄ˇger;怡r;憡hv;櫤Āayː˕ron;䄎;䐔lĀ;t˝˞戇a;䎔r;쀀𝔇Āaf˫̧Ācm˰̢riticalȀADGT̖̜̀̆cute;䂴oŴ̋̍;䋙bleAcute;䋝rave;䁠ilde;䋜ond;拄ferentialD;慆Ѱ̽\0\0\0͔͂\0Ѕf;쀀𝔻ƀ;DE͈͉͍䂨ot;惜qual;扐blèCDLRUVͣͲ΂ϏϢϸontourIntegraìȹoɴ͹\0\0ͻ»͉nArrow;懓Āeo·ΤftƀARTΐΖΡrrow;懐ightArrow;懔eåˊngĀLRΫτeftĀARγιrrow;柸ightArrow;柺ightArrow;柹ightĀATϘϞrrow;懒ee;抨pɁϩ\0\0ϯrrow;懑ownArrow;懕erticalBar;戥ǹABLRTaВЪаўѿͼrrowƀ;BUНОТ憓ar;椓pArrow;懵reve;䌑eft˒к\0ц\0ѐightVector;楐eeVector;楞ectorĀ;Bљњ憽ar;楖ightǔѧ\0ѱeeVector;楟ectorĀ;BѺѻ懁ar;楗eeĀ;A҆҇护rrow;憧ĀctҒҗr;쀀𝒟rok;䄐ࠀNTacdfglmopqstuxҽӀӄӋӞӢӧӮӵԡԯԶՒ՝ՠեG;䅊H耻Ð䃐cute耻É䃉ƀaiyӒӗӜron;䄚rc耻Ê䃊;䐭ot;䄖r;쀀𝔈rave耻È䃈ement;戈ĀapӺӾcr;䄒tyɓԆ\0\0ԒmallSquare;旻erySmallSquare;斫ĀgpԦԪon;䄘f;쀀𝔼silon;䎕uĀaiԼՉlĀ;TՂՃ橵ilde;扂librium;懌Āci՗՚r;愰m;橳a;䎗ml耻Ë䃋Āipժկsts;戃onentialE;慇ʀcfiosօֈ֍ֲ׌y;䐤r;쀀𝔉lledɓ֗\0\0֣mallSquare;旼erySmallSquare;斪Ͱֺ\0ֿ\0\0ׄf;쀀𝔽All;戀riertrf;愱cò׋؀JTabcdfgorstר׬ׯ׺؀ؒؖ؛؝أ٬ٲcy;䐃耻>䀾mmaĀ;d׷׸䎓;䏜reve;䄞ƀeiy؇،ؐdil;䄢rc;䄜;䐓ot;䄠r;쀀𝔊;拙pf;쀀𝔾eater̀EFGLSTصلَٖٛ٦qualĀ;Lؾؿ扥ess;招ullEqual;执reater;檢ess;扷lantEqual;橾ilde;扳cr;쀀𝒢;扫ЀAacfiosuڅڋږڛڞڪھۊRDcy;䐪Āctڐڔek;䋇;䁞irc;䄤r;愌lbertSpace;愋ǰگ\0ڲf;愍izontalLine;攀Āctۃۅòکrok;䄦mpńېۘownHumðįqual;扏܀EJOacdfgmnostuۺ۾܃܇܎ܚܞܡܨ݄ݸދޏޕcy;䐕lig;䄲cy;䐁cute耻Í䃍Āiyܓܘrc耻Î䃎;䐘ot;䄰r;愑rave耻Ì䃌ƀ;apܠܯܿĀcgܴܷr;䄪inaryI;慈lieóϝǴ݉\0ݢĀ;eݍݎ戬Āgrݓݘral;戫section;拂isibleĀCTݬݲomma;恣imes;恢ƀgptݿރވon;䄮f;쀀𝕀a;䎙cr;愐ilde;䄨ǫޚ\0ޞcy;䐆l耻Ï䃏ʀcfosuެ޷޼߂ߐĀiyޱ޵rc;䄴;䐙r;쀀𝔍pf;쀀𝕁ǣ߇\0ߌr;쀀𝒥rcy;䐈kcy;䐄΀HJacfosߤߨ߽߬߱ࠂࠈcy;䐥cy;䐌ppa;䎚Āey߶߻dil;䄶;䐚r;쀀𝔎pf;쀀𝕂cr;쀀𝒦րJTaceflmostࠥࠩࠬࡐࡣ঳সে্਷ੇcy;䐉耻<䀼ʀcmnpr࠷࠼ࡁࡄࡍute;䄹bda;䎛g;柪lacetrf;愒r;憞ƀaeyࡗ࡜ࡡron;䄽dil;䄻;䐛Āfsࡨ॰tԀACDFRTUVarࡾࢩࢱࣦ࣠ࣼयज़ΐ४Ānrࢃ࢏gleBracket;柨rowƀ;BR࢙࢚࢞憐ar;懤ightArrow;懆eiling;挈oǵࢷ\0ࣃbleBracket;柦nǔࣈ\0࣒eeVector;楡ectorĀ;Bࣛࣜ懃ar;楙loor;挊ightĀAV࣯ࣵrrow;憔ector;楎Āerँगeƀ;AVउऊऐ抣rrow;憤ector;楚iangleƀ;BEतथऩ抲ar;槏qual;抴pƀDTVषूौownVector;楑eeVector;楠ectorĀ;Bॖॗ憿ar;楘ectorĀ;B॥०憼ar;楒ightáΜs̀EFGLSTॾঋকঝঢভqualGreater;拚ullEqual;扦reater;扶ess;檡lantEqual;橽ilde;扲r;쀀𝔏Ā;eঽা拘ftarrow;懚idot;䄿ƀnpw৔ਖਛgȀLRlr৞৷ਂਐeftĀAR০৬rrow;柵ightArrow;柷ightArrow;柶eftĀarγਊightáοightáϊf;쀀𝕃erĀLRਢਬeftArrow;憙ightArrow;憘ƀchtਾੀੂòࡌ;憰rok;䅁;扪Ѐacefiosuਗ਼੝੠੷੼અઋ઎p;椅y;䐜Ādl੥੯iumSpace;恟lintrf;愳r;쀀𝔐nusPlus;戓pf;쀀𝕄cò੶;䎜ҀJacefostuણધભીଔଙඑ඗ඞcy;䐊cute;䅃ƀaey઴હાron;䅇dil;䅅;䐝ƀgswે૰଎ativeƀMTV૓૟૨ediumSpace;怋hiĀcn૦૘ë૙eryThiî૙tedĀGL૸ଆreaterGreateòٳessLesóੈLine;䀊r;쀀𝔑ȀBnptଢନଷ଺reak;恠BreakingSpace;䂠f;愕ڀ;CDEGHLNPRSTV୕ୖ୪୼஡௫ఄ౞಄ದ೘ൡඅ櫬Āou୛୤ngruent;扢pCap;扭oubleVerticalBar;戦ƀlqxஃஊ஛ement;戉ualĀ;Tஒஓ扠ilde;쀀≂̸ists;戄reater΀;EFGLSTஶஷ஽௉௓௘௥扯qual;扱ullEqual;쀀≧̸reater;쀀≫̸ess;批lantEqual;쀀⩾̸ilde;扵umpń௲௽ownHump;쀀≎̸qual;쀀≏̸eĀfsఊధtTriangleƀ;BEచఛడ拪ar;쀀⧏̸qual;括s̀;EGLSTవశ఼ౄోౘ扮qual;扰reater;扸ess;쀀≪̸lantEqual;쀀⩽̸ilde;扴estedĀGL౨౹reaterGreater;쀀⪢̸essLess;쀀⪡̸recedesƀ;ESಒಓಛ技qual;쀀⪯̸lantEqual;拠ĀeiಫಹverseElement;戌ghtTriangleƀ;BEೋೌ೒拫ar;쀀⧐̸qual;拭ĀquೝഌuareSuĀbp೨೹setĀ;E೰ೳ쀀⊏̸qual;拢ersetĀ;Eഃആ쀀⊐̸qual;拣ƀbcpഓതൎsetĀ;Eഛഞ쀀⊂⃒qual;抈ceedsȀ;ESTലള഻െ抁qual;쀀⪰̸lantEqual;拡ilde;쀀≿̸ersetĀ;E൘൛쀀⊃⃒qual;抉ildeȀ;EFT൮൯൵ൿ扁qual;扄ullEqual;扇ilde;扉erticalBar;戤cr;쀀𝒩ilde耻Ñ䃑;䎝܀Eacdfgmoprstuvලෂ෉෕ෛ෠෧෼ขภยา฿ไlig;䅒cute耻Ó䃓Āiy෎ීrc耻Ô䃔;䐞blac;䅐r;쀀𝔒rave耻Ò䃒ƀaei෮ෲ෶cr;䅌ga;䎩cron;䎟pf;쀀𝕆enCurlyĀDQฎบoubleQuote;怜uote;怘;橔Āclวฬr;쀀𝒪ash耻Ø䃘iŬื฼de耻Õ䃕es;樷ml耻Ö䃖erĀBP๋๠Āar๐๓r;怾acĀek๚๜;揞et;掴arenthesis;揜Ҁacfhilors๿ງຊຏຒດຝະ໼rtialD;戂y;䐟r;쀀𝔓i;䎦;䎠usMinus;䂱Āipຢອncareplanåڝf;愙Ȁ;eio຺ູ໠໤檻cedesȀ;EST່້໏໚扺qual;檯lantEqual;扼ilde;找me;怳Ādp໩໮uct;戏ortionĀ;aȥ໹l;戝Āci༁༆r;쀀𝒫;䎨ȀUfos༑༖༛༟OT耻"䀢r;쀀𝔔pf;愚cr;쀀𝒬؀BEacefhiorsu༾གྷཇའཱིྦྷྪྭ႖ႩႴႾarr;椐G耻®䂮ƀcnrཎནབute;䅔g;柫rĀ;tཛྷཝ憠l;椖ƀaeyཧཬཱron;䅘dil;䅖;䐠Ā;vླྀཹ愜erseĀEUྂྙĀlq྇ྎement;戋uilibrium;懋pEquilibrium;楯r»ཹo;䎡ghtЀACDFTUVa࿁࿫࿳ဢဨၛႇϘĀnr࿆࿒gleBracket;柩rowƀ;BL࿜࿝࿡憒ar;懥eftArrow;懄eiling;按oǵ࿹\0စbleBracket;柧nǔည\0နeeVector;楝ectorĀ;Bဝသ懂ar;楕loor;挋Āerိ၃eƀ;AVဵံြ抢rrow;憦ector;楛iangleƀ;BEၐၑၕ抳ar;槐qual;抵pƀDTVၣၮၸownVector;楏eeVector;楜ectorĀ;Bႂႃ憾ar;楔ectorĀ;B႑႒懀ar;楓Āpuႛ႞f;愝ndImplies;楰ightarrow;懛ĀchႹႼr;愛;憱leDelayed;槴ڀHOacfhimoqstuფჱჷჽᄙᄞᅑᅖᅡᅧᆵᆻᆿĀCcჩხHcy;䐩y;䐨FTcy;䐬cute;䅚ʀ;aeiyᄈᄉᄎᄓᄗ檼ron;䅠dil;䅞rc;䅜;䐡r;쀀𝔖ortȀDLRUᄪᄴᄾᅉownArrow»ОeftArrow»࢚ightArrow»࿝pArrow;憑gma;䎣allCircle;战pf;쀀𝕊ɲᅭ\0\0ᅰt;戚areȀ;ISUᅻᅼᆉᆯ斡ntersection;抓uĀbpᆏᆞsetĀ;Eᆗᆘ抏qual;抑ersetĀ;Eᆨᆩ抐qual;抒nion;抔cr;쀀𝒮ar;拆ȀbcmpᇈᇛሉላĀ;sᇍᇎ拐etĀ;Eᇍᇕqual;抆ĀchᇠህeedsȀ;ESTᇭᇮᇴᇿ扻qual;檰lantEqual;扽ilde;承Tháྌ;我ƀ;esሒሓሣ拑rsetĀ;Eሜም抃qual;抇et»ሓրHRSacfhiorsሾቄ቉ቕ቞ቱቶኟዂወዑORN耻Þ䃞ADE;愢ĀHc቎ቒcy;䐋y;䐦Ābuቚቜ;䀉;䎤ƀaeyብቪቯron;䅤dil;䅢;䐢r;쀀𝔗Āeiቻ኉ǲኀ\0ኇefore;戴a;䎘Ācn኎ኘkSpace;쀀  Space;怉ldeȀ;EFTካኬኲኼ戼qual;扃ullEqual;扅ilde;扈pf;쀀𝕋ipleDot;惛Āctዖዛr;쀀𝒯rok;䅦ૡዷጎጚጦ\0ጬጱ\0\0\0\0\0ጸጽ፷ᎅ\0᏿ᐄᐊᐐĀcrዻጁute耻Ú䃚rĀ;oጇገ憟cir;楉rǣጓ\0጖y;䐎ve;䅬Āiyጞጣrc耻Û䃛;䐣blac;䅰r;쀀𝔘rave耻Ù䃙acr;䅪Ādiፁ፩erĀBPፈ፝Āarፍፐr;䁟acĀekፗፙ;揟et;掵arenthesis;揝onĀ;P፰፱拃lus;抎Āgp፻፿on;䅲f;쀀𝕌ЀADETadps᎕ᎮᎸᏄϨᏒᏗᏳrrowƀ;BDᅐᎠᎤar;椒ownArrow;懅ownArrow;憕quilibrium;楮eeĀ;AᏋᏌ报rrow;憥ownáϳerĀLRᏞᏨeftArrow;憖ightArrow;憗iĀ;lᏹᏺ䏒on;䎥ing;䅮cr;쀀𝒰ilde;䅨ml耻Ü䃜ҀDbcdefosvᐧᐬᐰᐳᐾᒅᒊᒐᒖash;披ar;櫫y;䐒ashĀ;lᐻᐼ抩;櫦Āerᑃᑅ;拁ƀbtyᑌᑐᑺar;怖Ā;iᑏᑕcalȀBLSTᑡᑥᑪᑴar;戣ine;䁼eparator;杘ilde;所ThinSpace;怊r;쀀𝔙pf;쀀𝕍cr;쀀𝒱dash;抪ʀcefosᒧᒬᒱᒶᒼirc;䅴dge;拀r;쀀𝔚pf;쀀𝕎cr;쀀𝒲Ȁfiosᓋᓐᓒᓘr;쀀𝔛;䎞pf;쀀𝕏cr;쀀𝒳ҀAIUacfosuᓱᓵᓹᓽᔄᔏᔔᔚᔠcy;䐯cy;䐇cy;䐮cute耻Ý䃝Āiyᔉᔍrc;䅶;䐫r;쀀𝔜pf;쀀𝕐cr;쀀𝒴ml;䅸ЀHacdefosᔵᔹᔿᕋᕏᕝᕠᕤcy;䐖cute;䅹Āayᕄᕉron;䅽;䐗ot;䅻ǲᕔ\0ᕛoWidtè૙a;䎖r;愨pf;愤cr;쀀𝒵௡ᖃᖊᖐ\0ᖰᖶᖿ\0\0\0\0ᗆᗛᗫᙟ᙭\0ᚕ᚛ᚲᚹ\0ᚾcute耻á䃡reve;䄃̀;Ediuyᖜᖝᖡᖣᖨᖭ戾;쀀∾̳;房rc耻â䃢te肻´̆;䐰lig耻æ䃦Ā;r²ᖺ;쀀𝔞rave耻à䃠ĀepᗊᗖĀfpᗏᗔsym;愵èᗓha;䎱ĀapᗟcĀclᗤᗧr;䄁g;樿ɤᗰ\0\0ᘊʀ;adsvᗺᗻᗿᘁᘇ戧nd;橕;橜lope;橘;橚΀;elmrszᘘᘙᘛᘞᘿᙏᙙ戠;榤e»ᘙsdĀ;aᘥᘦ戡ѡᘰᘲᘴᘶᘸᘺᘼᘾ;榨;榩;榪;榫;榬;榭;榮;榯tĀ;vᙅᙆ戟bĀ;dᙌᙍ抾;榝Āptᙔᙗh;戢»¹arr;捼Āgpᙣᙧon;䄅f;쀀𝕒΀;Eaeiop዁ᙻᙽᚂᚄᚇᚊ;橰cir;橯;扊d;手s;䀧roxĀ;e዁ᚒñᚃing耻å䃥ƀctyᚡᚦᚨr;쀀𝒶;䀪mpĀ;e዁ᚯñʈilde耻ã䃣ml耻ä䃤Āciᛂᛈoninôɲnt;樑ࠀNabcdefiklnoprsu᛭ᛱᜰ᜼ᝃᝈ᝸᝽០៦ᠹᡐᜍ᤽᥈ᥰot;櫭Ācrᛶ᜞kȀcepsᜀᜅᜍᜓong;扌psilon;䏶rime;怵imĀ;e᜚᜛戽q;拍Ŷᜢᜦee;抽edĀ;gᜬᜭ挅e»ᜭrkĀ;t፜᜷brk;掶Āoyᜁᝁ;䐱quo;怞ʀcmprtᝓ᝛ᝡᝤᝨausĀ;eĊĉptyv;榰séᜌnoõēƀahwᝯ᝱ᝳ;䎲;愶een;扬r;쀀𝔟g΀costuvwឍឝឳេ៕៛៞ƀaiuបពរðݠrc;旯p»፱ƀdptឤឨឭot;樀lus;樁imes;樂ɱឹ\0\0ើcup;樆ar;昅riangleĀdu៍្own;施p;斳plus;樄eåᑄåᒭarow;植ƀako៭ᠦᠵĀcn៲ᠣkƀlst៺֫᠂ozenge;槫riangleȀ;dlr᠒᠓᠘᠝斴own;斾eft;旂ight;斸k;搣Ʊᠫ\0ᠳƲᠯ\0ᠱ;斒;斑4;斓ck;斈ĀeoᠾᡍĀ;qᡃᡆ쀀=⃥uiv;쀀≡⃥t;挐Ȁptwxᡙᡞᡧᡬf;쀀𝕓Ā;tᏋᡣom»Ꮜtie;拈؀DHUVbdhmptuvᢅᢖᢪᢻᣗᣛᣬ᣿ᤅᤊᤐᤡȀLRlrᢎᢐᢒᢔ;敗;敔;敖;敓ʀ;DUduᢡᢢᢤᢦᢨ敐;敦;敩;敤;敧ȀLRlrᢳᢵᢷᢹ;敝;敚;敜;教΀;HLRhlrᣊᣋᣍᣏᣑᣓᣕ救;敬;散;敠;敫;敢;敟ox;槉ȀLRlrᣤᣦᣨᣪ;敕;敒;攐;攌ʀ;DUduڽ᣷᣹᣻᣽;敥;敨;攬;攴inus;抟lus;択imes;抠ȀLRlrᤙᤛᤝ᤟;敛;敘;攘;攔΀;HLRhlrᤰᤱᤳᤵᤷ᤻᤹攂;敪;敡;敞;攼;攤;攜Āevģ᥂bar耻¦䂦Ȁceioᥑᥖᥚᥠr;쀀𝒷mi;恏mĀ;e᜚᜜lƀ;bhᥨᥩᥫ䁜;槅sub;柈Ŭᥴ᥾lĀ;e᥹᥺怢t»᥺pƀ;Eeįᦅᦇ;檮Ā;qۜۛೡᦧ\0᧨ᨑᨕᨲ\0ᨷᩐ\0\0᪴\0\0᫁\0\0ᬡᬮ᭍᭒\0᯽\0ᰌƀcpr᦭ᦲ᧝ute;䄇̀;abcdsᦿᧀᧄ᧊᧕᧙戩nd;橄rcup;橉Āau᧏᧒p;橋p;橇ot;橀;쀀∩︀Āeo᧢᧥t;恁îړȀaeiu᧰᧻ᨁᨅǰ᧵\0᧸s;橍on;䄍dil耻ç䃧rc;䄉psĀ;sᨌᨍ橌m;橐ot;䄋ƀdmnᨛᨠᨦil肻¸ƭptyv;榲t脀¢;eᨭᨮ䂢räƲr;쀀𝔠ƀceiᨽᩀᩍy;䑇ckĀ;mᩇᩈ朓ark»ᩈ;䏇r΀;Ecefms᩟᩠ᩢᩫ᪤᪪᪮旋;槃ƀ;elᩩᩪᩭ䋆q;扗eɡᩴ\0\0᪈rrowĀlr᩼᪁eft;憺ight;憻ʀRSacd᪒᪔᪖᪚᪟»ཇ;擈st;抛irc;抚ash;抝nint;樐id;櫯cir;槂ubsĀ;u᪻᪼晣it»᪼ˬ᫇᫔᫺\0ᬊonĀ;eᫍᫎ䀺Ā;qÇÆɭ᫙\0\0᫢aĀ;t᫞᫟䀬;䁀ƀ;fl᫨᫩᫫戁îᅠeĀmx᫱᫶ent»᫩eóɍǧ᫾\0ᬇĀ;dኻᬂot;橭nôɆƀfryᬐᬔᬗ;쀀𝕔oäɔ脀©;sŕᬝr;愗Āaoᬥᬩrr;憵ss;朗Ācuᬲᬷr;쀀𝒸Ābpᬼ᭄Ā;eᭁᭂ櫏;櫑Ā;eᭉᭊ櫐;櫒dot;拯΀delprvw᭠᭬᭷ᮂᮬᯔ᯹arrĀlr᭨᭪;椸;椵ɰ᭲\0\0᭵r;拞c;拟arrĀ;p᭿ᮀ憶;椽̀;bcdosᮏᮐᮖᮡᮥᮨ截rcap;橈Āauᮛᮞp;橆p;橊ot;抍r;橅;쀀∪︀Ȁalrv᮵ᮿᯞᯣrrĀ;mᮼᮽ憷;椼yƀevwᯇᯔᯘqɰᯎ\0\0ᯒreã᭳uã᭵ee;拎edge;拏en耻¤䂤earrowĀlrᯮ᯳eft»ᮀight»ᮽeäᯝĀciᰁᰇoninôǷnt;戱lcty;挭ঀAHabcdefhijlorstuwz᰸᰻᰿ᱝᱩᱵᲊᲞᲬᲷ᳻᳿ᴍᵻᶑᶫᶻ᷆᷍rò΁ar;楥Ȁglrs᱈ᱍ᱒᱔ger;怠eth;愸òᄳhĀ;vᱚᱛ怐»ऊūᱡᱧarow;椏aã̕Āayᱮᱳron;䄏;䐴ƀ;ao̲ᱼᲄĀgrʿᲁr;懊tseq;橷ƀglmᲑᲔᲘ耻°䂰ta;䎴ptyv;榱ĀirᲣᲨsht;楿;쀀𝔡arĀlrᲳᲵ»ࣜ»သʀaegsv᳂͸᳖᳜᳠mƀ;oș᳊᳔ndĀ;ș᳑uit;晦amma;䏝in;拲ƀ;io᳧᳨᳸䃷de脀÷;o᳧ᳰntimes;拇nø᳷cy;䑒cɯᴆ\0\0ᴊrn;挞op;挍ʀlptuwᴘᴝᴢᵉᵕlar;䀤f;쀀𝕕ʀ;emps̋ᴭᴷᴽᵂqĀ;d͒ᴳot;扑inus;戸lus;戔quare;抡blebarwedgåúnƀadhᄮᵝᵧownarrowóᲃarpoonĀlrᵲᵶefôᲴighôᲶŢᵿᶅkaro÷གɯᶊ\0\0ᶎrn;挟op;挌ƀcotᶘᶣᶦĀryᶝᶡ;쀀𝒹;䑕l;槶rok;䄑Ādrᶰᶴot;拱iĀ;fᶺ᠖斿Āah᷀᷃ròЩaòྦangle;榦Āci᷒ᷕy;䑟grarr;柿ऀDacdefglmnopqrstuxḁḉḙḸոḼṉṡṾấắẽỡἪἷὄ὎὚ĀDoḆᴴoôᲉĀcsḎḔute耻é䃩ter;橮ȀaioyḢḧḱḶron;䄛rĀ;cḭḮ扖耻ê䃪lon;払;䑍ot;䄗ĀDrṁṅot;扒;쀀𝔢ƀ;rsṐṑṗ檚ave耻è䃨Ā;dṜṝ檖ot;檘Ȁ;ilsṪṫṲṴ檙nters;揧;愓Ā;dṹṺ檕ot;檗ƀapsẅẉẗcr;䄓tyƀ;svẒẓẕ戅et»ẓpĀ1;ẝẤĳạả;怄;怅怃ĀgsẪẬ;䅋p;怂ĀgpẴẸon;䄙f;쀀𝕖ƀalsỄỎỒrĀ;sỊị拕l;槣us;橱iƀ;lvỚớở䎵on»ớ;䏵ȀcsuvỪỳἋἣĀioữḱrc»Ḯɩỹ\0\0ỻíՈantĀglἂἆtr»ṝess»Ṻƀaeiἒ἖Ἒls;䀽st;扟vĀ;DȵἠD;橸parsl;槥ĀDaἯἳot;打rr;楱ƀcdiἾὁỸr;愯oô͒ĀahὉὋ;䎷耻ð䃰Āmrὓὗl耻ë䃫o;悬ƀcipὡὤὧl;䀡sôծĀeoὬὴctatioîՙnentialåչৡᾒ\0ᾞ\0ᾡᾧ\0\0ῆῌ\0ΐ\0ῦῪ \0 ⁚llingdotseñṄy;䑄male;晀ƀilrᾭᾳ῁lig;耀ﬃɩᾹ\0\0᾽g;耀ﬀig;耀ﬄ;쀀𝔣lig;耀ﬁlig;쀀fjƀaltῙ῜ῡt;晭ig;耀ﬂns;斱of;䆒ǰ΅\0ῳf;쀀𝕗ĀakֿῷĀ;vῼ´拔;櫙artint;樍Āao‌⁕Ācs‑⁒α‚‰‸⁅⁈\0⁐β•‥‧‪‬\0‮耻½䂽;慓耻¼䂼;慕;慙;慛Ƴ‴\0‶;慔;慖ʴ‾⁁\0\0⁃耻¾䂾;慗;慜5;慘ƶ⁌\0⁎;慚;慝8;慞l;恄wn;挢cr;쀀𝒻ࢀEabcdefgijlnorstv₂₉₟₥₰₴⃰⃵⃺⃿℃ℒℸ̗ℾ⅒↞Ā;lٍ₇;檌ƀcmpₐₕ₝ute;䇵maĀ;dₜ᳚䎳;檆reve;䄟Āiy₪₮rc;䄝;䐳ot;䄡Ȁ;lqsؾق₽⃉ƀ;qsؾٌ⃄lanô٥Ȁ;cdl٥⃒⃥⃕c;檩otĀ;o⃜⃝檀Ā;l⃢⃣檂;檄Ā;e⃪⃭쀀⋛︀s;檔r;쀀𝔤Ā;gٳ؛mel;愷cy;䑓Ȁ;Eajٚℌℎℐ;檒;檥;檤ȀEaesℛℝ℩ℴ;扩pĀ;p℣ℤ檊rox»ℤĀ;q℮ℯ檈Ā;q℮ℛim;拧pf;쀀𝕘Āci⅃ⅆr;愊mƀ;el٫ⅎ⅐;檎;檐茀>;cdlqr׮ⅠⅪⅮⅳⅹĀciⅥⅧ;檧r;橺ot;拗Par;榕uest;橼ʀadelsↄⅪ←ٖ↛ǰ↉\0↎proø₞r;楸qĀlqؿ↖lesó₈ií٫Āen↣↭rtneqq;쀀≩︀Å↪ԀAabcefkosy⇄⇇⇱⇵⇺∘∝∯≨≽ròΠȀilmr⇐⇔⇗⇛rsðᒄf»․ilôکĀdr⇠⇤cy;䑊ƀ;cwࣴ⇫⇯ir;楈;憭ar;意irc;䄥ƀalr∁∎∓rtsĀ;u∉∊晥it»∊lip;怦con;抹r;쀀𝔥sĀew∣∩arow;椥arow;椦ʀamopr∺∾≃≞≣rr;懿tht;戻kĀlr≉≓eftarrow;憩ightarrow;憪f;쀀𝕙bar;怕ƀclt≯≴≸r;쀀𝒽asè⇴rok;䄧Ābp⊂⊇ull;恃hen»ᱛૡ⊣\0⊪\0⊸⋅⋎\0⋕⋳\0\0⋸⌢⍧⍢⍿\0⎆⎪⎴cute耻í䃭ƀ;iyݱ⊰⊵rc耻î䃮;䐸Ācx⊼⊿y;䐵cl耻¡䂡ĀfrΟ⋉;쀀𝔦rave耻ì䃬Ȁ;inoܾ⋝⋩⋮Āin⋢⋦nt;樌t;戭fin;槜ta;愩lig;䄳ƀaop⋾⌚⌝ƀcgt⌅⌈⌗r;䄫ƀelpܟ⌏⌓inåގarôܠh;䄱f;抷ed;䆵ʀ;cfotӴ⌬⌱⌽⍁are;愅inĀ;t⌸⌹戞ie;槝doô⌙ʀ;celpݗ⍌⍐⍛⍡al;抺Āgr⍕⍙eróᕣã⍍arhk;樗rod;樼Ȁcgpt⍯⍲⍶⍻y;䑑on;䄯f;쀀𝕚a;䎹uest耻¿䂿Āci⎊⎏r;쀀𝒾nʀ;EdsvӴ⎛⎝⎡ӳ;拹ot;拵Ā;v⎦⎧拴;拳Ā;iݷ⎮lde;䄩ǫ⎸\0⎼cy;䑖l耻ï䃯̀cfmosu⏌⏗⏜⏡⏧⏵Āiy⏑⏕rc;䄵;䐹r;쀀𝔧ath;䈷pf;쀀𝕛ǣ⏬\0⏱r;쀀𝒿rcy;䑘kcy;䑔Ѐacfghjos␋␖␢␧␭␱␵␻ppaĀ;v␓␔䎺;䏰Āey␛␠dil;䄷;䐺r;쀀𝔨reen;䄸cy;䑅cy;䑜pf;쀀𝕜cr;쀀𝓀஀ABEHabcdefghjlmnoprstuv⑰⒁⒆⒍⒑┎┽╚▀♎♞♥♹♽⚚⚲⛘❝❨➋⟀⠁⠒ƀart⑷⑺⑼rò৆òΕail;椛arr;椎Ā;gঔ⒋;檋ar;楢ॣ⒥\0⒪\0⒱\0\0\0\0\0⒵Ⓔ\0ⓆⓈⓍ\0⓹ute;䄺mptyv;榴raîࡌbda;䎻gƀ;dlࢎⓁⓃ;榑åࢎ;檅uo耻«䂫rЀ;bfhlpst࢙ⓞⓦⓩ⓫⓮⓱⓵Ā;f࢝ⓣs;椟s;椝ë≒p;憫l;椹im;楳l;憢ƀ;ae⓿─┄檫il;椙Ā;s┉┊檭;쀀⪭︀ƀabr┕┙┝rr;椌rk;杲Āak┢┬cĀek┨┪;䁻;䁛Āes┱┳;榋lĀdu┹┻;榏;榍Ȁaeuy╆╋╖╘ron;䄾Ādi═╔il;䄼ìࢰâ┩;䐻Ȁcqrs╣╦╭╽a;椶uoĀ;rนᝆĀdu╲╷har;楧shar;楋h;憲ʀ;fgqs▋▌উ◳◿扤tʀahlrt▘▤▷◂◨rrowĀ;t࢙□aé⓶arpoonĀdu▯▴own»њp»०eftarrows;懇ightƀahs◍◖◞rrowĀ;sࣴࢧarpoonó྘quigarro÷⇰hreetimes;拋ƀ;qs▋ও◺lanôবʀ;cdgsব☊☍☝☨c;檨otĀ;o☔☕橿Ā;r☚☛檁;檃Ā;e☢☥쀀⋚︀s;檓ʀadegs☳☹☽♉♋pproøⓆot;拖qĀgq♃♅ôউgtò⒌ôছiíলƀilr♕࣡♚sht;楼;쀀𝔩Ā;Eজ♣;檑š♩♶rĀdu▲♮Ā;l॥♳;楪lk;斄cy;䑙ʀ;achtੈ⚈⚋⚑⚖rò◁orneòᴈard;楫ri;旺Āio⚟⚤dot;䅀ustĀ;a⚬⚭掰che»⚭ȀEaes⚻⚽⛉⛔;扨pĀ;p⛃⛄檉rox»⛄Ā;q⛎⛏檇Ā;q⛎⚻im;拦Ѐabnoptwz⛩⛴⛷✚✯❁❇❐Ānr⛮⛱g;柬r;懽rëࣁgƀlmr⛿✍✔eftĀar০✇ightá৲apsto;柼ightá৽parrowĀlr✥✩efô⓭ight;憬ƀafl✶✹✽r;榅;쀀𝕝us;樭imes;樴š❋❏st;戗áፎƀ;ef❗❘᠀旊nge»❘arĀ;l❤❥䀨t;榓ʀachmt❳❶❼➅➇ròࢨorneòᶌarĀ;d྘➃;業;怎ri;抿̀achiqt➘➝ੀ➢➮➻quo;怹r;쀀𝓁mƀ;egল➪➬;檍;檏Ābu┪➳oĀ;rฟ➹;怚rok;䅂萀<;cdhilqrࠫ⟒☹⟜⟠⟥⟪⟰Āci⟗⟙;檦r;橹reå◲mes;拉arr;楶uest;橻ĀPi⟵⟹ar;榖ƀ;ef⠀भ᠛旃rĀdu⠇⠍shar;楊har;楦Āen⠗⠡rtneqq;쀀≨︀Å⠞܀Dacdefhilnopsu⡀⡅⢂⢎⢓⢠⢥⢨⣚⣢⣤ઃ⣳⤂Dot;戺Ȁclpr⡎⡒⡣⡽r耻¯䂯Āet⡗⡙;時Ā;e⡞⡟朠se»⡟Ā;sျ⡨toȀ;dluျ⡳⡷⡻owîҌefôएðᏑker;斮Āoy⢇⢌mma;権;䐼ash;怔asuredangle»ᘦr;쀀𝔪o;愧ƀcdn⢯⢴⣉ro耻µ䂵Ȁ;acdᑤ⢽⣀⣄sôᚧir;櫰ot肻·Ƶusƀ;bd⣒ᤃ⣓戒Ā;uᴼ⣘;横ţ⣞⣡p;櫛ò−ðઁĀdp⣩⣮els;抧f;쀀𝕞Āct⣸⣽r;쀀𝓂pos»ᖝƀ;lm⤉⤊⤍䎼timap;抸ఀGLRVabcdefghijlmoprstuvw⥂⥓⥾⦉⦘⧚⧩⨕⨚⩘⩝⪃⪕⪤⪨⬄⬇⭄⭿⮮ⰴⱧⱼ⳩Āgt⥇⥋;쀀⋙̸Ā;v⥐௏쀀≫⃒ƀelt⥚⥲⥶ftĀar⥡⥧rrow;懍ightarrow;懎;쀀⋘̸Ā;v⥻ే쀀≪⃒ightarrow;懏ĀDd⦎⦓ash;抯ash;抮ʀbcnpt⦣⦧⦬⦱⧌la»˞ute;䅄g;쀀∠⃒ʀ;Eiop඄⦼⧀⧅⧈;쀀⩰̸d;쀀≋̸s;䅉roø඄urĀ;a⧓⧔普lĀ;s⧓ସǳ⧟\0⧣p肻 ଷmpĀ;e௹ఀʀaeouy⧴⧾⨃⨐⨓ǰ⧹\0⧻;橃on;䅈dil;䅆ngĀ;dൾ⨊ot;쀀⩭̸p;橂;䐽ash;怓΀;Aadqsxஒ⨩⨭⨻⩁⩅⩐rr;懗rĀhr⨳⨶k;椤Ā;oᏲᏰot;쀀≐̸uiöୣĀei⩊⩎ar;椨í஘istĀ;s஠டr;쀀𝔫ȀEest௅⩦⩹⩼ƀ;qs஼⩭௡ƀ;qs஼௅⩴lanô௢ií௪Ā;rஶ⪁»ஷƀAap⪊⪍⪑rò⥱rr;憮ar;櫲ƀ;svྍ⪜ྌĀ;d⪡⪢拼;拺cy;䑚΀AEadest⪷⪺⪾⫂⫅⫶⫹rò⥦;쀀≦̸rr;憚r;急Ȁ;fqs఻⫎⫣⫯tĀar⫔⫙rro÷⫁ightarro÷⪐ƀ;qs఻⪺⫪lanôౕĀ;sౕ⫴»శiíౝĀ;rవ⫾iĀ;eచథiäඐĀpt⬌⬑f;쀀𝕟膀¬;in⬙⬚⬶䂬nȀ;Edvஉ⬤⬨⬮;쀀⋹̸ot;쀀⋵̸ǡஉ⬳⬵;拷;拶iĀ;vಸ⬼ǡಸ⭁⭃;拾;拽ƀaor⭋⭣⭩rȀ;ast୻⭕⭚⭟lleì୻l;쀀⫽⃥;쀀∂̸lint;樔ƀ;ceಒ⭰⭳uåಥĀ;cಘ⭸Ā;eಒ⭽ñಘȀAait⮈⮋⮝⮧rò⦈rrƀ;cw⮔⮕⮙憛;쀀⤳̸;쀀↝̸ghtarrow»⮕riĀ;eೋೖ΀chimpqu⮽⯍⯙⬄୸⯤⯯Ȁ;cerല⯆ഷ⯉uå൅;쀀𝓃ortɭ⬅\0\0⯖ará⭖mĀ;e൮⯟Ā;q൴൳suĀbp⯫⯭å೸åഋƀbcp⯶ⰑⰙȀ;Ees⯿ⰀഢⰄ抄;쀀⫅̸etĀ;eഛⰋqĀ;qണⰀcĀ;eലⰗñസȀ;EesⰢⰣൟⰧ抅;쀀⫆̸etĀ;e൘ⰮqĀ;qൠⰣȀgilrⰽⰿⱅⱇìௗlde耻ñ䃱çృiangleĀlrⱒⱜeftĀ;eచⱚñదightĀ;eೋⱥñ೗Ā;mⱬⱭ䎽ƀ;esⱴⱵⱹ䀣ro;愖p;怇ҀDHadgilrsⲏⲔⲙⲞⲣⲰⲶⳓⳣash;抭arr;椄p;쀀≍⃒ash;抬ĀetⲨⲬ;쀀≥⃒;쀀>⃒nfin;槞ƀAetⲽⳁⳅrr;椂;쀀≤⃒Ā;rⳊⳍ쀀<⃒ie;쀀⊴⃒ĀAtⳘⳜrr;椃rie;쀀⊵⃒im;쀀∼⃒ƀAan⳰⳴ⴂrr;懖rĀhr⳺⳽k;椣Ā;oᏧᏥear;椧ቓ᪕\0\0\0\0\0\0\0\0\0\0\0\0\0ⴭ\0ⴸⵈⵠⵥ⵲ⶄᬇ\0\0ⶍⶫ\0ⷈⷎ\0ⷜ⸙⸫⸾⹃Ācsⴱ᪗ute耻ó䃳ĀiyⴼⵅrĀ;c᪞ⵂ耻ô䃴;䐾ʀabios᪠ⵒⵗǈⵚlac;䅑v;樸old;榼lig;䅓Ācr⵩⵭ir;榿;쀀𝔬ͯ⵹\0\0⵼\0ⶂn;䋛ave耻ò䃲;槁Ābmⶈ෴ar;榵Ȁacitⶕ⶘ⶥⶨrò᪀Āir⶝ⶠr;榾oss;榻nå๒;槀ƀaeiⶱⶵⶹcr;䅍ga;䏉ƀcdnⷀⷅǍron;䎿;榶pf;쀀𝕠ƀaelⷔ⷗ǒr;榷rp;榹΀;adiosvⷪⷫⷮ⸈⸍⸐⸖戨rò᪆Ȁ;efmⷷⷸ⸂⸅橝rĀ;oⷾⷿ愴f»ⷿ耻ª䂪耻º䂺gof;抶r;橖lope;橗;橛ƀclo⸟⸡⸧ò⸁ash耻ø䃸l;折iŬⸯ⸴de耻õ䃵esĀ;aǛ⸺s;樶ml耻ö䃶bar;挽ૡ⹞\0⹽\0⺀⺝\0⺢⺹\0\0⻋ຜ\0⼓\0\0⼫⾼\0⿈rȀ;astЃ⹧⹲຅脀¶;l⹭⹮䂶leìЃɩ⹸\0\0⹻m;櫳;櫽y;䐿rʀcimpt⺋⺏⺓ᡥ⺗nt;䀥od;䀮il;怰enk;怱r;쀀𝔭ƀimo⺨⺰⺴Ā;v⺭⺮䏆;䏕maô੶ne;明ƀ;tv⺿⻀⻈䏀chfork»´;䏖Āau⻏⻟nĀck⻕⻝kĀ;h⇴⻛;愎ö⇴sҀ;abcdemst⻳⻴ᤈ⻹⻽⼄⼆⼊⼎䀫cir;樣ir;樢Āouᵀ⼂;樥;橲n肻±ຝim;樦wo;樧ƀipu⼙⼠⼥ntint;樕f;쀀𝕡nd耻£䂣Ԁ;Eaceinosu່⼿⽁⽄⽇⾁⾉⾒⽾⾶;檳p;檷uå໙Ā;c໎⽌̀;acens່⽙⽟⽦⽨⽾pproø⽃urlyeñ໙ñ໎ƀaes⽯⽶⽺pprox;檹qq;檵im;拨iíໟmeĀ;s⾈ຮ怲ƀEas⽸⾐⽺ð⽵ƀdfp໬⾙⾯ƀals⾠⾥⾪lar;挮ine;挒urf;挓Ā;t໻⾴ï໻rel;抰Āci⿀⿅r;쀀𝓅;䏈ncsp;怈̀fiopsu⿚⋢⿟⿥⿫⿱r;쀀𝔮pf;쀀𝕢rime;恗cr;쀀𝓆ƀaeo⿸〉〓tĀei⿾々rnionóڰnt;樖stĀ;e【】䀿ñἙô༔઀ABHabcdefhilmnoprstux぀けさすムㄎㄫㅇㅢㅲㆎ㈆㈕㈤㈩㉘㉮㉲㊐㊰㊷ƀartぇおがròႳòϝail;検aròᱥar;楤΀cdenqrtとふへみわゔヌĀeuねぱ;쀀∽̱te;䅕iãᅮmptyv;榳gȀ;del࿑らるろ;榒;榥å࿑uo耻»䂻rր;abcfhlpstw࿜ガクシスゼゾダッデナp;極Ā;f࿠ゴs;椠;椳s;椞ë≝ð✮l;楅im;楴l;憣;憝Āaiパフil;椚oĀ;nホボ戶aló༞ƀabrョリヮrò៥rk;杳ĀakンヽcĀekヹ・;䁽;䁝Āes㄂㄄;榌lĀduㄊㄌ;榎;榐Ȁaeuyㄗㄜㄧㄩron;䅙Ādiㄡㄥil;䅗ì࿲âヺ;䑀Ȁclqsㄴㄷㄽㅄa;椷dhar;楩uoĀ;rȎȍh;憳ƀacgㅎㅟངlȀ;ipsླྀㅘㅛႜnåႻarôྩt;断ƀilrㅩဣㅮsht;楽;쀀𝔯ĀaoㅷㆆrĀduㅽㅿ»ѻĀ;l႑ㆄ;楬Ā;vㆋㆌ䏁;䏱ƀgns㆕ㇹㇼht̀ahlrstㆤㆰ㇂㇘㇤㇮rrowĀ;t࿜ㆭaéトarpoonĀduㆻㆿowîㅾp»႒eftĀah㇊㇐rrowó࿪arpoonóՑightarrows;應quigarro÷ニhreetimes;拌g;䋚ingdotseñἲƀahm㈍㈐㈓rò࿪aòՑ;怏oustĀ;a㈞㈟掱che»㈟mid;櫮Ȁabpt㈲㈽㉀㉒Ānr㈷㈺g;柭r;懾rëဃƀafl㉇㉊㉎r;榆;쀀𝕣us;樮imes;樵Āap㉝㉧rĀ;g㉣㉤䀩t;榔olint;樒arò㇣Ȁachq㉻㊀Ⴜ㊅quo;怺r;쀀𝓇Ābu・㊊oĀ;rȔȓƀhir㊗㊛㊠reåㇸmes;拊iȀ;efl㊪ၙᠡ㊫方tri;槎luhar;楨;愞ൡ㋕㋛㋟㌬㌸㍱\0㍺㎤\0\0㏬㏰\0㐨㑈㑚㒭㒱㓊㓱\0㘖\0\0㘳cute;䅛quï➺Ԁ;Eaceinpsyᇭ㋳㋵㋿㌂㌋㌏㌟㌦㌩;檴ǰ㋺\0㋼;檸on;䅡uåᇾĀ;dᇳ㌇il;䅟rc;䅝ƀEas㌖㌘㌛;檶p;檺im;择olint;樓iíሄ;䑁otƀ;be㌴ᵇ㌵担;橦΀Aacmstx㍆㍊㍗㍛㍞㍣㍭rr;懘rĀhr㍐㍒ë∨Ā;oਸ਼਴t耻§䂧i;䀻war;椩mĀin㍩ðnuóñt;朶rĀ;o㍶⁕쀀𝔰Ȁacoy㎂㎆㎑㎠rp;景Āhy㎋㎏cy;䑉;䑈rtɭ㎙\0\0㎜iäᑤaraì⹯耻­䂭Āgm㎨㎴maƀ;fv㎱㎲㎲䏃;䏂Ѐ;deglnprካ㏅㏉㏎㏖㏞㏡㏦ot;橪Ā;q኱ኰĀ;E㏓㏔檞;檠Ā;E㏛㏜檝;檟e;扆lus;樤arr;楲aròᄽȀaeit㏸㐈㐏㐗Āls㏽㐄lsetmé㍪hp;樳parsl;槤Ādlᑣ㐔e;挣Ā;e㐜㐝檪Ā;s㐢㐣檬;쀀⪬︀ƀflp㐮㐳㑂tcy;䑌Ā;b㐸㐹䀯Ā;a㐾㐿槄r;挿f;쀀𝕤aĀdr㑍ЂesĀ;u㑔㑕晠it»㑕ƀcsu㑠㑹㒟Āau㑥㑯pĀ;sᆈ㑫;쀀⊓︀pĀ;sᆴ㑵;쀀⊔︀uĀbp㑿㒏ƀ;esᆗᆜ㒆etĀ;eᆗ㒍ñᆝƀ;esᆨᆭ㒖etĀ;eᆨ㒝ñᆮƀ;afᅻ㒦ְrť㒫ֱ»ᅼaròᅈȀcemt㒹㒾㓂㓅r;쀀𝓈tmîñiì㐕aræᆾĀar㓎㓕rĀ;f㓔ឿ昆Āan㓚㓭ightĀep㓣㓪psiloîỠhé⺯s»⡒ʀbcmnp㓻㕞ሉ㖋㖎Ҁ;Edemnprs㔎㔏㔑㔕㔞㔣㔬㔱㔶抂;櫅ot;檽Ā;dᇚ㔚ot;櫃ult;櫁ĀEe㔨㔪;櫋;把lus;檿arr;楹ƀeiu㔽㕒㕕tƀ;en㔎㕅㕋qĀ;qᇚ㔏eqĀ;q㔫㔨m;櫇Ābp㕚㕜;櫕;櫓c̀;acensᇭ㕬㕲㕹㕻㌦pproø㋺urlyeñᇾñᇳƀaes㖂㖈㌛pproø㌚qñ㌗g;晪ڀ123;Edehlmnps㖩㖬㖯ሜ㖲㖴㗀㗉㗕㗚㗟㗨㗭耻¹䂹耻²䂲耻³䂳;櫆Āos㖹㖼t;檾ub;櫘Ā;dሢ㗅ot;櫄sĀou㗏㗒l;柉b;櫗arr;楻ult;櫂ĀEe㗤㗦;櫌;抋lus;櫀ƀeiu㗴㘉㘌tƀ;enሜ㗼㘂qĀ;qሢ㖲eqĀ;q㗧㗤m;櫈Ābp㘑㘓;櫔;櫖ƀAan㘜㘠㘭rr;懙rĀhr㘦㘨ë∮Ā;oਫ਩war;椪lig耻ß䃟௡㙑㙝㙠ዎ㙳㙹\0㙾㛂\0\0\0\0\0㛛㜃\0㜉㝬\0\0\0㞇ɲ㙖\0\0㙛get;挖;䏄rë๟ƀaey㙦㙫㙰ron;䅥dil;䅣;䑂lrec;挕r;쀀𝔱Ȁeiko㚆㚝㚵㚼ǲ㚋\0㚑eĀ4fኄኁaƀ;sv㚘㚙㚛䎸ym;䏑Ācn㚢㚲kĀas㚨㚮pproø዁im»ኬsðኞĀas㚺㚮ð዁rn耻þ䃾Ǭ̟㛆⋧es膀×;bd㛏㛐㛘䃗Ā;aᤏ㛕r;樱;樰ƀeps㛡㛣㜀á⩍Ȁ;bcf҆㛬㛰㛴ot;挶ir;櫱Ā;o㛹㛼쀀𝕥rk;櫚á㍢rime;怴ƀaip㜏㜒㝤dåቈ΀adempst㜡㝍㝀㝑㝗㝜㝟ngleʀ;dlqr㜰㜱㜶㝀㝂斵own»ᶻeftĀ;e⠀㜾ñम;扜ightĀ;e㊪㝋ñၚot;旬inus;樺lus;樹b;槍ime;樻ezium;揢ƀcht㝲㝽㞁Āry㝷㝻;쀀𝓉;䑆cy;䑛rok;䅧Āio㞋㞎xô᝷headĀlr㞗㞠eftarro÷ࡏightarrow»ཝऀAHabcdfghlmoprstuw㟐㟓㟗㟤㟰㟼㠎㠜㠣㠴㡑㡝㡫㢩㣌㣒㣪㣶ròϭar;楣Ācr㟜㟢ute耻ú䃺òᅐrǣ㟪\0㟭y;䑞ve;䅭Āiy㟵㟺rc耻û䃻;䑃ƀabh㠃㠆㠋ròᎭlac;䅱aòᏃĀir㠓㠘sht;楾;쀀𝔲rave耻ù䃹š㠧㠱rĀlr㠬㠮»ॗ»ႃlk;斀Āct㠹㡍ɯ㠿\0\0㡊rnĀ;e㡅㡆挜r»㡆op;挏ri;旸Āal㡖㡚cr;䅫肻¨͉Āgp㡢㡦on;䅳f;쀀𝕦̀adhlsuᅋ㡸㡽፲㢑㢠ownáᎳarpoonĀlr㢈㢌efô㠭ighô㠯iƀ;hl㢙㢚㢜䏅»ᏺon»㢚parrows;懈ƀcit㢰㣄㣈ɯ㢶\0\0㣁rnĀ;e㢼㢽挝r»㢽op;挎ng;䅯ri;旹cr;쀀𝓊ƀdir㣙㣝㣢ot;拰lde;䅩iĀ;f㜰㣨»᠓Āam㣯㣲rò㢨l耻ü䃼angle;榧ހABDacdeflnoprsz㤜㤟㤩㤭㦵㦸㦽㧟㧤㧨㧳㧹㧽㨁㨠ròϷarĀ;v㤦㤧櫨;櫩asèϡĀnr㤲㤷grt;榜΀eknprst㓣㥆㥋㥒㥝㥤㦖appá␕othinçẖƀhir㓫⻈㥙opô⾵Ā;hᎷ㥢ïㆍĀiu㥩㥭gmá㎳Ābp㥲㦄setneqĀ;q㥽㦀쀀⊊︀;쀀⫋︀setneqĀ;q㦏㦒쀀⊋︀;쀀⫌︀Āhr㦛㦟etá㚜iangleĀlr㦪㦯eft»थight»ၑy;䐲ash»ံƀelr㧄㧒㧗ƀ;beⷪ㧋㧏ar;抻q;扚lip;拮Ābt㧜ᑨaòᑩr;쀀𝔳tré㦮suĀbp㧯㧱»ജ»൙pf;쀀𝕧roð໻tré㦴Ācu㨆㨋r;쀀𝓋Ābp㨐㨘nĀEe㦀㨖»㥾nĀEe㦒㨞»㦐igzag;榚΀cefoprs㨶㨻㩖㩛㩔㩡㩪irc;䅵Ādi㩀㩑Ābg㩅㩉ar;機eĀ;qᗺ㩏;扙erp;愘r;쀀𝔴pf;쀀𝕨Ā;eᑹ㩦atèᑹcr;쀀𝓌ૣណ㪇\0㪋\0㪐㪛\0\0㪝㪨㪫㪯\0\0㫃㫎\0㫘ៜ៟tré៑r;쀀𝔵ĀAa㪔㪗ròσrò৶;䎾ĀAa㪡㪤ròθrò৫að✓is;拻ƀdptឤ㪵㪾Āfl㪺ឩ;쀀𝕩imåឲĀAa㫇㫊ròώròਁĀcq㫒ីr;쀀𝓍Āpt៖㫜ré។Ѐacefiosu㫰㫽㬈㬌㬑㬕㬛㬡cĀuy㫶㫻te耻ý䃽;䑏Āiy㬂㬆rc;䅷;䑋n耻¥䂥r;쀀𝔶cy;䑗pf;쀀𝕪cr;쀀𝓎Ācm㬦㬩y;䑎l耻ÿ䃿Ԁacdefhiosw㭂㭈㭔㭘㭤㭩㭭㭴㭺㮀cute;䅺Āay㭍㭒ron;䅾;䐷ot;䅼Āet㭝㭡træᕟa;䎶r;쀀𝔷cy;䐶grarr;懝pf;쀀𝕫cr;쀀𝓏Ājn㮅㮇;怍j;怌'.split("").map((e=>e.charCodeAt(0)))),y=new Uint16Array("Ȁaglq\tɭ\0\0p;䀦os;䀧t;䀾t;䀼uot;䀢".split("").map((e=>e.charCodeAt(0))));var w;const O=new Map([[0,65533],[128,8364],[130,8218],[131,402],[132,8222],[133,8230],[134,8224],[135,8225],[136,710],[137,8240],[138,352],[139,8249],[140,338],[142,381],[145,8216],[146,8217],[147,8220],[148,8221],[149,8226],[150,8211],[151,8212],[152,732],[153,8482],[154,353],[155,8250],[156,339],[158,382],[159,376]]),k=null!==(w=String.fromCodePoint)&&void 0!==w?w:function(e){let t="";return e>65535&&(e-=65536,t+=String.fromCharCode(e>>>10&1023|55296),e=56320|1023&e),t+=String.fromCharCode(e),t};var x,E,S,_,C,T,A,R,P,D;function M(e){return e>=x.ZERO&&e<=x.NINE}!function(e){e[e.NUM=35]="NUM",e[e.SEMI=59]="SEMI",e[e.EQUALS=61]="EQUALS",e[e.ZERO=48]="ZERO",e[e.NINE=57]="NINE",e[e.LOWER_A=97]="LOWER_A",e[e.LOWER_F=102]="LOWER_F",e[e.LOWER_X=120]="LOWER_X",e[e.LOWER_Z=122]="LOWER_Z",e[e.UPPER_A=65]="UPPER_A",e[e.UPPER_F=70]="UPPER_F",e[e.UPPER_Z=90]="UPPER_Z"}(x||(x={})),function(e){e[e.VALUE_LENGTH=49152]="VALUE_LENGTH",e[e.BRANCH_LENGTH=16256]="BRANCH_LENGTH",e[e.JUMP_TABLE=127]="JUMP_TABLE"}(E||(E={})),function(e){e[e.EntityStart=0]="EntityStart",e[e.NumericStart=1]="NumericStart",e[e.NumericDecimal=2]="NumericDecimal",e[e.NumericHex=3]="NumericHex",e[e.NamedEntity=4]="NamedEntity"}(S||(S={})),function(e){e[e.Legacy=0]="Legacy",e[e.Strict=1]="Strict",e[e.Attribute=2]="Attribute"}(_||(_={}));class ${constructor(e,t,n){this.decodeTree=e,this.emitCodePoint=t,this.errors=n,this.state=S.EntityStart,this.consumed=1,this.result=0,this.treeIndex=0,this.excess=1,this.decodeMode=_.Strict}startEntity(e){this.decodeMode=e,this.state=S.EntityStart,this.result=0,this.treeIndex=0,this.excess=1,this.consumed=1}write(e,t){switch(this.state){case S.EntityStart:return e.charCodeAt(t)===x.NUM?(this.state=S.NumericStart,this.consumed+=1,this.stateNumericStart(e,t+1)):(this.state=S.NamedEntity,this.stateNamedEntity(e,t));case S.NumericStart:return this.stateNumericStart(e,t);case S.NumericDecimal:return this.stateNumericDecimal(e,t);case S.NumericHex:return this.stateNumericHex(e,t);case S.NamedEntity:return this.stateNamedEntity(e,t)}}stateNumericStart(e,t){return t>=e.length?-1:(32|e.charCodeAt(t))===x.LOWER_X?(this.state=S.NumericHex,this.consumed+=1,this.stateNumericHex(e,t+1)):(this.state=S.NumericDecimal,this.stateNumericDecimal(e,t))}addToNumericResult(e,t,n,r){if(t!==n){const a=n-t;this.result=this.result*Math.pow(r,a)+parseInt(e.substr(t,a),r),this.consumed+=a}}stateNumericHex(e,t){const n=t;for(;t<e.length;){const a=e.charCodeAt(t);if(!(M(a)||(r=a,r>=x.UPPER_A&&r<=x.UPPER_F||r>=x.LOWER_A&&r<=x.LOWER_F)))return this.addToNumericResult(e,n,t,16),this.emitNumericEntity(a,3);t+=1}var r;return this.addToNumericResult(e,n,t,16),-1}stateNumericDecimal(e,t){const n=t;for(;t<e.length;){const r=e.charCodeAt(t);if(!M(r))return this.addToNumericResult(e,n,t,10),this.emitNumericEntity(r,2);t+=1}return this.addToNumericResult(e,n,t,10),-1}emitNumericEntity(e,t){var n;if(this.consumed<=t)return null===(n=this.errors)||void 0===n||n.absenceOfDigitsInNumericCharacterReference(this.consumed),0;if(e===x.SEMI)this.consumed+=1;else if(this.decodeMode===_.Strict)return 0;return this.emitCodePoint(function(e){var t;return e>=55296&&e<=57343||e>1114111?65533:null!==(t=O.get(e))&&void 0!==t?t:e}(this.result),this.consumed),this.errors&&(e!==x.SEMI&&this.errors.missingSemicolonAfterCharacterReference(),this.errors.validateNumericCharacterReference(this.result)),this.consumed}stateNamedEntity(e,t){const{decodeTree:n}=this;let r=n[this.treeIndex],a=(r&E.VALUE_LENGTH)>>14;for(;t<e.length;t++,this.excess++){const i=e.charCodeAt(t);if(this.treeIndex=I(n,r,this.treeIndex+Math.max(1,a),i),this.treeIndex<0)return 0===this.result||this.decodeMode===_.Attribute&&(0===a||((o=i)===x.EQUALS||function(e){return e>=x.UPPER_A&&e<=x.UPPER_Z||e>=x.LOWER_A&&e<=x.LOWER_Z||M(e)}(o)))?0:this.emitNotTerminatedNamedEntity();if(r=n[this.treeIndex],a=(r&E.VALUE_LENGTH)>>14,0!==a){if(i===x.SEMI)return this.emitNamedEntityData(this.treeIndex,a,this.consumed+this.excess);this.decodeMode!==_.Strict&&(this.result=this.treeIndex,this.consumed+=this.excess,this.excess=0)}}var o;return-1}emitNotTerminatedNamedEntity(){var e;const{result:t,decodeTree:n}=this,r=(n[t]&E.VALUE_LENGTH)>>14;return this.emitNamedEntityData(t,r,this.consumed),null===(e=this.errors)||void 0===e||e.missingSemicolonAfterCharacterReference(),this.consumed}emitNamedEntityData(e,t,n){const{decodeTree:r}=this;return this.emitCodePoint(1===t?r[e]&~E.VALUE_LENGTH:r[e+1],n),3===t&&this.emitCodePoint(r[e+2],n),n}end(){var e;switch(this.state){case S.NamedEntity:return 0===this.result||this.decodeMode===_.Attribute&&this.result!==this.treeIndex?0:this.emitNotTerminatedNamedEntity();case S.NumericDecimal:return this.emitNumericEntity(0,2);case S.NumericHex:return this.emitNumericEntity(0,3);case S.NumericStart:return null===(e=this.errors)||void 0===e||e.absenceOfDigitsInNumericCharacterReference(this.consumed),0;case S.EntityStart:return 0}}}function N(e){let t="";const n=new $(e,(e=>t+=k(e)));return function(e,r){let a=0,o=0;for(;(o=e.indexOf("&",o))>=0;){t+=e.slice(a,o),n.startEntity(r);const i=n.write(e,o+1);if(i<0){a=o+n.end();break}a=o+i,o=0===i?a+1:a}const i=t+e.slice(a);return t="",i}}function I(e,t,n,r){const a=(t&E.BRANCH_LENGTH)>>7,o=t&E.JUMP_TABLE;if(0===a)return 0!==o&&r===o?n:-1;if(o){const t=r-o;return t<0||t>=a?-1:e[n+t]-1}let i=n,s=i+a-1;for(;i<=s;){const t=i+s>>>1,n=e[t];if(n<r)i=t+1;else{if(!(n>r))return e[t+a];s=t-1}}return-1}N(b),N(y),(D=C||(C={})).HTML="http://www.w3.org/1999/xhtml",D.MATHML="http://www.w3.org/1998/Math/MathML",D.SVG="http://www.w3.org/2000/svg",D.XLINK="http://www.w3.org/1999/xlink",D.XML="http://www.w3.org/XML/1998/namespace",D.XMLNS="http://www.w3.org/2000/xmlns/",function(e){e.TYPE="type",e.ACTION="action",e.ENCODING="encoding",e.PROMPT="prompt",e.NAME="name",e.COLOR="color",e.FACE="face",e.SIZE="size"}(T||(T={})),function(e){e.NO_QUIRKS="no-quirks",e.QUIRKS="quirks",e.LIMITED_QUIRKS="limited-quirks"}(A||(A={})),function(e){e.A="a",e.ADDRESS="address",e.ANNOTATION_XML="annotation-xml",e.APPLET="applet",e.AREA="area",e.ARTICLE="article",e.ASIDE="aside",e.B="b",e.BASE="base",e.BASEFONT="basefont",e.BGSOUND="bgsound",e.BIG="big",e.BLOCKQUOTE="blockquote",e.BODY="body",e.BR="br",e.BUTTON="button",e.CAPTION="caption",e.CENTER="center",e.CODE="code",e.COL="col",e.COLGROUP="colgroup",e.DD="dd",e.DESC="desc",e.DETAILS="details",e.DIALOG="dialog",e.DIR="dir",e.DIV="div",e.DL="dl",e.DT="dt",e.EM="em",e.EMBED="embed",e.FIELDSET="fieldset",e.FIGCAPTION="figcaption",e.FIGURE="figure",e.FONT="font",e.FOOTER="footer",e.FOREIGN_OBJECT="foreignObject",e.FORM="form",e.FRAME="frame",e.FRAMESET="frameset",e.H1="h1",e.H2="h2",e.H3="h3",e.H4="h4",e.H5="h5",e.H6="h6",e.HEAD="head",e.HEADER="header",e.HGROUP="hgroup",e.HR="hr",e.HTML="html",e.I="i",e.IMG="img",e.IMAGE="image",e.INPUT="input",e.IFRAME="iframe",e.KEYGEN="keygen",e.LABEL="label",e.LI="li",e.LINK="link",e.LISTING="listing",e.MAIN="main",e.MALIGNMARK="malignmark",e.MARQUEE="marquee",e.MATH="math",e.MENU="menu",e.META="meta",e.MGLYPH="mglyph",e.MI="mi",e.MO="mo",e.MN="mn",e.MS="ms",e.MTEXT="mtext",e.NAV="nav",e.NOBR="nobr",e.NOFRAMES="noframes",e.NOEMBED="noembed",e.NOSCRIPT="noscript",e.OBJECT="object",e.OL="ol",e.OPTGROUP="optgroup",e.OPTION="option",e.P="p",e.PARAM="param",e.PLAINTEXT="plaintext",e.PRE="pre",e.RB="rb",e.RP="rp",e.RT="rt",e.RTC="rtc",e.RUBY="ruby",e.S="s",e.SCRIPT="script",e.SEARCH="search",e.SECTION="section",e.SELECT="select",e.SOURCE="source",e.SMALL="small",e.SPAN="span",e.STRIKE="strike",e.STRONG="strong",e.STYLE="style",e.SUB="sub",e.SUMMARY="summary",e.SUP="sup",e.TABLE="table",e.TBODY="tbody",e.TEMPLATE="template",e.TEXTAREA="textarea",e.TFOOT="tfoot",e.TD="td",e.TH="th",e.THEAD="thead",e.TITLE="title",e.TR="tr",e.TRACK="track",e.TT="tt",e.U="u",e.UL="ul",e.SVG="svg",e.VAR="var",e.WBR="wbr",e.XMP="xmp"}(R||(R={})),function(e){e[e.UNKNOWN=0]="UNKNOWN",e[e.A=1]="A",e[e.ADDRESS=2]="ADDRESS",e[e.ANNOTATION_XML=3]="ANNOTATION_XML",e[e.APPLET=4]="APPLET",e[e.AREA=5]="AREA",e[e.ARTICLE=6]="ARTICLE",e[e.ASIDE=7]="ASIDE",e[e.B=8]="B",e[e.BASE=9]="BASE",e[e.BASEFONT=10]="BASEFONT",e[e.BGSOUND=11]="BGSOUND",e[e.BIG=12]="BIG",e[e.BLOCKQUOTE=13]="BLOCKQUOTE",e[e.BODY=14]="BODY",e[e.BR=15]="BR",e[e.BUTTON=16]="BUTTON",e[e.CAPTION=17]="CAPTION",e[e.CENTER=18]="CENTER",e[e.CODE=19]="CODE",e[e.COL=20]="COL",e[e.COLGROUP=21]="COLGROUP",e[e.DD=22]="DD",e[e.DESC=23]="DESC",e[e.DETAILS=24]="DETAILS",e[e.DIALOG=25]="DIALOG",e[e.DIR=26]="DIR",e[e.DIV=27]="DIV",e[e.DL=28]="DL",e[e.DT=29]="DT",e[e.EM=30]="EM",e[e.EMBED=31]="EMBED",e[e.FIELDSET=32]="FIELDSET",e[e.FIGCAPTION=33]="FIGCAPTION",e[e.FIGURE=34]="FIGURE",e[e.FONT=35]="FONT",e[e.FOOTER=36]="FOOTER",e[e.FOREIGN_OBJECT=37]="FOREIGN_OBJECT",e[e.FORM=38]="FORM",e[e.FRAME=39]="FRAME",e[e.FRAMESET=40]="FRAMESET",e[e.H1=41]="H1",e[e.H2=42]="H2",e[e.H3=43]="H3",e[e.H4=44]="H4",e[e.H5=45]="H5",e[e.H6=46]="H6",e[e.HEAD=47]="HEAD",e[e.HEADER=48]="HEADER",e[e.HGROUP=49]="HGROUP",e[e.HR=50]="HR",e[e.HTML=51]="HTML",e[e.I=52]="I",e[e.IMG=53]="IMG",e[e.IMAGE=54]="IMAGE",e[e.INPUT=55]="INPUT",e[e.IFRAME=56]="IFRAME",e[e.KEYGEN=57]="KEYGEN",e[e.LABEL=58]="LABEL",e[e.LI=59]="LI",e[e.LINK=60]="LINK",e[e.LISTING=61]="LISTING",e[e.MAIN=62]="MAIN",e[e.MALIGNMARK=63]="MALIGNMARK",e[e.MARQUEE=64]="MARQUEE",e[e.MATH=65]="MATH",e[e.MENU=66]="MENU",e[e.META=67]="META",e[e.MGLYPH=68]="MGLYPH",e[e.MI=69]="MI",e[e.MO=70]="MO",e[e.MN=71]="MN",e[e.MS=72]="MS",e[e.MTEXT=73]="MTEXT",e[e.NAV=74]="NAV",e[e.NOBR=75]="NOBR",e[e.NOFRAMES=76]="NOFRAMES",e[e.NOEMBED=77]="NOEMBED",e[e.NOSCRIPT=78]="NOSCRIPT",e[e.OBJECT=79]="OBJECT",e[e.OL=80]="OL",e[e.OPTGROUP=81]="OPTGROUP",e[e.OPTION=82]="OPTION",e[e.P=83]="P",e[e.PARAM=84]="PARAM",e[e.PLAINTEXT=85]="PLAINTEXT",e[e.PRE=86]="PRE",e[e.RB=87]="RB",e[e.RP=88]="RP",e[e.RT=89]="RT",e[e.RTC=90]="RTC",e[e.RUBY=91]="RUBY",e[e.S=92]="S",e[e.SCRIPT=93]="SCRIPT",e[e.SEARCH=94]="SEARCH",e[e.SECTION=95]="SECTION",e[e.SELECT=96]="SELECT",e[e.SOURCE=97]="SOURCE",e[e.SMALL=98]="SMALL",e[e.SPAN=99]="SPAN",e[e.STRIKE=100]="STRIKE",e[e.STRONG=101]="STRONG",e[e.STYLE=102]="STYLE",e[e.SUB=103]="SUB",e[e.SUMMARY=104]="SUMMARY",e[e.SUP=105]="SUP",e[e.TABLE=106]="TABLE",e[e.TBODY=107]="TBODY",e[e.TEMPLATE=108]="TEMPLATE",e[e.TEXTAREA=109]="TEXTAREA",e[e.TFOOT=110]="TFOOT",e[e.TD=111]="TD",e[e.TH=112]="TH",e[e.THEAD=113]="THEAD",e[e.TITLE=114]="TITLE",e[e.TR=115]="TR",e[e.TRACK=116]="TRACK",e[e.TT=117]="TT",e[e.U=118]="U",e[e.UL=119]="UL",e[e.SVG=120]="SVG",e[e.VAR=121]="VAR",e[e.WBR=122]="WBR",e[e.XMP=123]="XMP"}(P||(P={}));const L=new Map([[R.A,P.A],[R.ADDRESS,P.ADDRESS],[R.ANNOTATION_XML,P.ANNOTATION_XML],[R.APPLET,P.APPLET],[R.AREA,P.AREA],[R.ARTICLE,P.ARTICLE],[R.ASIDE,P.ASIDE],[R.B,P.B],[R.BASE,P.BASE],[R.BASEFONT,P.BASEFONT],[R.BGSOUND,P.BGSOUND],[R.BIG,P.BIG],[R.BLOCKQUOTE,P.BLOCKQUOTE],[R.BODY,P.BODY],[R.BR,P.BR],[R.BUTTON,P.BUTTON],[R.CAPTION,P.CAPTION],[R.CENTER,P.CENTER],[R.CODE,P.CODE],[R.COL,P.COL],[R.COLGROUP,P.COLGROUP],[R.DD,P.DD],[R.DESC,P.DESC],[R.DETAILS,P.DETAILS],[R.DIALOG,P.DIALOG],[R.DIR,P.DIR],[R.DIV,P.DIV],[R.DL,P.DL],[R.DT,P.DT],[R.EM,P.EM],[R.EMBED,P.EMBED],[R.FIELDSET,P.FIELDSET],[R.FIGCAPTION,P.FIGCAPTION],[R.FIGURE,P.FIGURE],[R.FONT,P.FONT],[R.FOOTER,P.FOOTER],[R.FOREIGN_OBJECT,P.FOREIGN_OBJECT],[R.FORM,P.FORM],[R.FRAME,P.FRAME],[R.FRAMESET,P.FRAMESET],[R.H1,P.H1],[R.H2,P.H2],[R.H3,P.H3],[R.H4,P.H4],[R.H5,P.H5],[R.H6,P.H6],[R.HEAD,P.HEAD],[R.HEADER,P.HEADER],[R.HGROUP,P.HGROUP],[R.HR,P.HR],[R.HTML,P.HTML],[R.I,P.I],[R.IMG,P.IMG],[R.IMAGE,P.IMAGE],[R.INPUT,P.INPUT],[R.IFRAME,P.IFRAME],[R.KEYGEN,P.KEYGEN],[R.LABEL,P.LABEL],[R.LI,P.LI],[R.LINK,P.LINK],[R.LISTING,P.LISTING],[R.MAIN,P.MAIN],[R.MALIGNMARK,P.MALIGNMARK],[R.MARQUEE,P.MARQUEE],[R.MATH,P.MATH],[R.MENU,P.MENU],[R.META,P.META],[R.MGLYPH,P.MGLYPH],[R.MI,P.MI],[R.MO,P.MO],[R.MN,P.MN],[R.MS,P.MS],[R.MTEXT,P.MTEXT],[R.NAV,P.NAV],[R.NOBR,P.NOBR],[R.NOFRAMES,P.NOFRAMES],[R.NOEMBED,P.NOEMBED],[R.NOSCRIPT,P.NOSCRIPT],[R.OBJECT,P.OBJECT],[R.OL,P.OL],[R.OPTGROUP,P.OPTGROUP],[R.OPTION,P.OPTION],[R.P,P.P],[R.PARAM,P.PARAM],[R.PLAINTEXT,P.PLAINTEXT],[R.PRE,P.PRE],[R.RB,P.RB],[R.RP,P.RP],[R.RT,P.RT],[R.RTC,P.RTC],[R.RUBY,P.RUBY],[R.S,P.S],[R.SCRIPT,P.SCRIPT],[R.SEARCH,P.SEARCH],[R.SECTION,P.SECTION],[R.SELECT,P.SELECT],[R.SOURCE,P.SOURCE],[R.SMALL,P.SMALL],[R.SPAN,P.SPAN],[R.STRIKE,P.STRIKE],[R.STRONG,P.STRONG],[R.STYLE,P.STYLE],[R.SUB,P.SUB],[R.SUMMARY,P.SUMMARY],[R.SUP,P.SUP],[R.TABLE,P.TABLE],[R.TBODY,P.TBODY],[R.TEMPLATE,P.TEMPLATE],[R.TEXTAREA,P.TEXTAREA],[R.TFOOT,P.TFOOT],[R.TD,P.TD],[R.TH,P.TH],[R.THEAD,P.THEAD],[R.TITLE,P.TITLE],[R.TR,P.TR],[R.TRACK,P.TRACK],[R.TT,P.TT],[R.U,P.U],[R.UL,P.UL],[R.SVG,P.SVG],[R.VAR,P.VAR],[R.WBR,P.WBR],[R.XMP,P.XMP]]);function B(e){var t;return null!==(t=L.get(e))&&void 0!==t?t:P.UNKNOWN}const F=P,Q={[C.HTML]:new Set([F.ADDRESS,F.APPLET,F.AREA,F.ARTICLE,F.ASIDE,F.BASE,F.BASEFONT,F.BGSOUND,F.BLOCKQUOTE,F.BODY,F.BR,F.BUTTON,F.CAPTION,F.CENTER,F.COL,F.COLGROUP,F.DD,F.DETAILS,F.DIR,F.DIV,F.DL,F.DT,F.EMBED,F.FIELDSET,F.FIGCAPTION,F.FIGURE,F.FOOTER,F.FORM,F.FRAME,F.FRAMESET,F.H1,F.H2,F.H3,F.H4,F.H5,F.H6,F.HEAD,F.HEADER,F.HGROUP,F.HR,F.HTML,F.IFRAME,F.IMG,F.INPUT,F.LI,F.LINK,F.LISTING,F.MAIN,F.MARQUEE,F.MENU,F.META,F.NAV,F.NOEMBED,F.NOFRAMES,F.NOSCRIPT,F.OBJECT,F.OL,F.P,F.PARAM,F.PLAINTEXT,F.PRE,F.SCRIPT,F.SECTION,F.SELECT,F.SOURCE,F.STYLE,F.SUMMARY,F.TABLE,F.TBODY,F.TD,F.TEMPLATE,F.TEXTAREA,F.TFOOT,F.TH,F.THEAD,F.TITLE,F.TR,F.TRACK,F.UL,F.WBR,F.XMP]),[C.MATHML]:new Set([F.MI,F.MO,F.MN,F.MS,F.MTEXT,F.ANNOTATION_XML]),[C.SVG]:new Set([F.TITLE,F.FOREIGN_OBJECT,F.DESC]),[C.XLINK]:new Set,[C.XML]:new Set,[C.XMLNS]:new Set},U=new Set([F.H1,F.H2,F.H3,F.H4,F.H5,F.H6]),z=new Set([R.STYLE,R.SCRIPT,R.XMP,R.IFRAME,R.NOEMBED,R.NOFRAMES,R.PLAINTEXT]);function j(e,t){return z.has(e)||t&&e===R.NOSCRIPT}var W;!function(e){e[e.DATA=0]="DATA",e[e.RCDATA=1]="RCDATA",e[e.RAWTEXT=2]="RAWTEXT",e[e.SCRIPT_DATA=3]="SCRIPT_DATA",e[e.PLAINTEXT=4]="PLAINTEXT",e[e.TAG_OPEN=5]="TAG_OPEN",e[e.END_TAG_OPEN=6]="END_TAG_OPEN",e[e.TAG_NAME=7]="TAG_NAME",e[e.RCDATA_LESS_THAN_SIGN=8]="RCDATA_LESS_THAN_SIGN",e[e.RCDATA_END_TAG_OPEN=9]="RCDATA_END_TAG_OPEN",e[e.RCDATA_END_TAG_NAME=10]="RCDATA_END_TAG_NAME",e[e.RAWTEXT_LESS_THAN_SIGN=11]="RAWTEXT_LESS_THAN_SIGN",e[e.RAWTEXT_END_TAG_OPEN=12]="RAWTEXT_END_TAG_OPEN",e[e.RAWTEXT_END_TAG_NAME=13]="RAWTEXT_END_TAG_NAME",e[e.SCRIPT_DATA_LESS_THAN_SIGN=14]="SCRIPT_DATA_LESS_THAN_SIGN",e[e.SCRIPT_DATA_END_TAG_OPEN=15]="SCRIPT_DATA_END_TAG_OPEN",e[e.SCRIPT_DATA_END_TAG_NAME=16]="SCRIPT_DATA_END_TAG_NAME",e[e.SCRIPT_DATA_ESCAPE_START=17]="SCRIPT_DATA_ESCAPE_START",e[e.SCRIPT_DATA_ESCAPE_START_DASH=18]="SCRIPT_DATA_ESCAPE_START_DASH",e[e.SCRIPT_DATA_ESCAPED=19]="SCRIPT_DATA_ESCAPED",e[e.SCRIPT_DATA_ESCAPED_DASH=20]="SCRIPT_DATA_ESCAPED_DASH",e[e.SCRIPT_DATA_ESCAPED_DASH_DASH=21]="SCRIPT_DATA_ESCAPED_DASH_DASH",e[e.SCRIPT_DATA_ESCAPED_LESS_THAN_SIGN=22]="SCRIPT_DATA_ESCAPED_LESS_THAN_SIGN",e[e.SCRIPT_DATA_ESCAPED_END_TAG_OPEN=23]="SCRIPT_DATA_ESCAPED_END_TAG_OPEN",e[e.SCRIPT_DATA_ESCAPED_END_TAG_NAME=24]="SCRIPT_DATA_ESCAPED_END_TAG_NAME",e[e.SCRIPT_DATA_DOUBLE_ESCAPE_START=25]="SCRIPT_DATA_DOUBLE_ESCAPE_START",e[e.SCRIPT_DATA_DOUBLE_ESCAPED=26]="SCRIPT_DATA_DOUBLE_ESCAPED",e[e.SCRIPT_DATA_DOUBLE_ESCAPED_DASH=27]="SCRIPT_DATA_DOUBLE_ESCAPED_DASH",e[e.SCRIPT_DATA_DOUBLE_ESCAPED_DASH_DASH=28]="SCRIPT_DATA_DOUBLE_ESCAPED_DASH_DASH",e[e.SCRIPT_DATA_DOUBLE_ESCAPED_LESS_THAN_SIGN=29]="SCRIPT_DATA_DOUBLE_ESCAPED_LESS_THAN_SIGN",e[e.SCRIPT_DATA_DOUBLE_ESCAPE_END=30]="SCRIPT_DATA_DOUBLE_ESCAPE_END",e[e.BEFORE_ATTRIBUTE_NAME=31]="BEFORE_ATTRIBUTE_NAME",e[e.ATTRIBUTE_NAME=32]="ATTRIBUTE_NAME",e[e.AFTER_ATTRIBUTE_NAME=33]="AFTER_ATTRIBUTE_NAME",e[e.BEFORE_ATTRIBUTE_VALUE=34]="BEFORE_ATTRIBUTE_VALUE",e[e.ATTRIBUTE_VALUE_DOUBLE_QUOTED=35]="ATTRIBUTE_VALUE_DOUBLE_QUOTED",e[e.ATTRIBUTE_VALUE_SINGLE_QUOTED=36]="ATTRIBUTE_VALUE_SINGLE_QUOTED",e[e.ATTRIBUTE_VALUE_UNQUOTED=37]="ATTRIBUTE_VALUE_UNQUOTED",e[e.AFTER_ATTRIBUTE_VALUE_QUOTED=38]="AFTER_ATTRIBUTE_VALUE_QUOTED",e[e.SELF_CLOSING_START_TAG=39]="SELF_CLOSING_START_TAG",e[e.BOGUS_COMMENT=40]="BOGUS_COMMENT",e[e.MARKUP_DECLARATION_OPEN=41]="MARKUP_DECLARATION_OPEN",e[e.COMMENT_START=42]="COMMENT_START",e[e.COMMENT_START_DASH=43]="COMMENT_START_DASH",e[e.COMMENT=44]="COMMENT",e[e.COMMENT_LESS_THAN_SIGN=45]="COMMENT_LESS_THAN_SIGN",e[e.COMMENT_LESS_THAN_SIGN_BANG=46]="COMMENT_LESS_THAN_SIGN_BANG",e[e.COMMENT_LESS_THAN_SIGN_BANG_DASH=47]="COMMENT_LESS_THAN_SIGN_BANG_DASH",e[e.COMMENT_LESS_THAN_SIGN_BANG_DASH_DASH=48]="COMMENT_LESS_THAN_SIGN_BANG_DASH_DASH",e[e.COMMENT_END_DASH=49]="COMMENT_END_DASH",e[e.COMMENT_END=50]="COMMENT_END",e[e.COMMENT_END_BANG=51]="COMMENT_END_BANG",e[e.DOCTYPE=52]="DOCTYPE",e[e.BEFORE_DOCTYPE_NAME=53]="BEFORE_DOCTYPE_NAME",e[e.DOCTYPE_NAME=54]="DOCTYPE_NAME",e[e.AFTER_DOCTYPE_NAME=55]="AFTER_DOCTYPE_NAME",e[e.AFTER_DOCTYPE_PUBLIC_KEYWORD=56]="AFTER_DOCTYPE_PUBLIC_KEYWORD",e[e.BEFORE_DOCTYPE_PUBLIC_IDENTIFIER=57]="BEFORE_DOCTYPE_PUBLIC_IDENTIFIER",e[e.DOCTYPE_PUBLIC_IDENTIFIER_DOUBLE_QUOTED=58]="DOCTYPE_PUBLIC_IDENTIFIER_DOUBLE_QUOTED",e[e.DOCTYPE_PUBLIC_IDENTIFIER_SINGLE_QUOTED=59]="DOCTYPE_PUBLIC_IDENTIFIER_SINGLE_QUOTED",e[e.AFTER_DOCTYPE_PUBLIC_IDENTIFIER=60]="AFTER_DOCTYPE_PUBLIC_IDENTIFIER",e[e.BETWEEN_DOCTYPE_PUBLIC_AND_SYSTEM_IDENTIFIERS=61]="BETWEEN_DOCTYPE_PUBLIC_AND_SYSTEM_IDENTIFIERS",e[e.AFTER_DOCTYPE_SYSTEM_KEYWORD=62]="AFTER_DOCTYPE_SYSTEM_KEYWORD",e[e.BEFORE_DOCTYPE_SYSTEM_IDENTIFIER=63]="BEFORE_DOCTYPE_SYSTEM_IDENTIFIER",e[e.DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED=64]="DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED",e[e.DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED=65]="DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED",e[e.AFTER_DOCTYPE_SYSTEM_IDENTIFIER=66]="AFTER_DOCTYPE_SYSTEM_IDENTIFIER",e[e.BOGUS_DOCTYPE=67]="BOGUS_DOCTYPE",e[e.CDATA_SECTION=68]="CDATA_SECTION",e[e.CDATA_SECTION_BRACKET=69]="CDATA_SECTION_BRACKET",e[e.CDATA_SECTION_END=70]="CDATA_SECTION_END",e[e.CHARACTER_REFERENCE=71]="CHARACTER_REFERENCE",e[e.AMBIGUOUS_AMPERSAND=72]="AMBIGUOUS_AMPERSAND"}(W||(W={}));const X={DATA:W.DATA,RCDATA:W.RCDATA,RAWTEXT:W.RAWTEXT,SCRIPT_DATA:W.SCRIPT_DATA,PLAINTEXT:W.PLAINTEXT,CDATA_SECTION:W.CDATA_SECTION};function q(e){return e>=s.LATIN_CAPITAL_A&&e<=s.LATIN_CAPITAL_Z}function V(e){return function(e){return e>=s.LATIN_SMALL_A&&e<=s.LATIN_SMALL_Z}(e)||q(e)}function H(e){return V(e)||function(e){return e>=s.DIGIT_0&&e<=s.DIGIT_9}(e)}function Z(e){return e+32}function G(e){return e===s.SPACE||e===s.LINE_FEED||e===s.TABULATION||e===s.FORM_FEED}function K(e){return G(e)||e===s.SOLIDUS||e===s.GREATER_THAN_SIGN}class Y{constructor(e,t){this.options=e,this.handler=t,this.paused=!1,this.inLoop=!1,this.inForeignNode=!1,this.lastStartTagName="",this.active=!1,this.state=W.DATA,this.returnState=W.DATA,this.entityStartPos=0,this.consumedAfterSnapshot=-1,this.currentCharacterToken=null,this.currentToken=null,this.currentAttr={name:"",value:""},this.preprocessor=new g(t),this.currentLocation=this.getCurrentLocation(-1),this.entityDecoder=new $(b,((e,t)=>{this.preprocessor.pos=this.entityStartPos+t-1,this._flushCodePointConsumedAsCharacterReference(e)}),t.onParseError?{missingSemicolonAfterCharacterReference:()=>{this._err(f.missingSemicolonAfterCharacterReference,1)},absenceOfDigitsInNumericCharacterReference:e=>{this._err(f.absenceOfDigitsInNumericCharacterReference,this.entityStartPos-this.preprocessor.pos+e)},validateNumericCharacterReference:e=>{const t=function(e){return e===s.NULL?f.nullCharacterReference:e>1114111?f.characterReferenceOutsideUnicodeRange:d(e)?f.surrogateCharacterReference:h(e)?f.noncharacterCharacterReference:p(e)||e===s.CARRIAGE_RETURN?f.controlCharacterReference:null}(e);t&&this._err(t,1)}}:void 0)}_err(e,t=0){var n,r;null===(r=(n=this.handler).onParseError)||void 0===r||r.call(n,this.preprocessor.getError(e,t))}getCurrentLocation(e){return this.options.sourceCodeLocationInfo?{startLine:this.preprocessor.line,startCol:this.preprocessor.col-e,startOffset:this.preprocessor.offset-e,endLine:-1,endCol:-1,endOffset:-1}:null}_runParsingLoop(){if(!this.inLoop){for(this.inLoop=!0;this.active&&!this.paused;){this.consumedAfterSnapshot=0;const e=this._consume();this._ensureHibernation()||this._callState(e)}this.inLoop=!1}}pause(){this.paused=!0}resume(e){if(!this.paused)throw new Error("Parser was already resumed");this.paused=!1,this.inLoop||(this._runParsingLoop(),this.paused||null==e||e())}write(e,t,n){this.active=!0,this.preprocessor.write(e,t),this._runParsingLoop(),this.paused||null==n||n()}insertHtmlAtCurrentPos(e){this.active=!0,this.preprocessor.insertHtmlAtCurrentPos(e),this._runParsingLoop()}_ensureHibernation(){return!!this.preprocessor.endOfChunkHit&&(this.preprocessor.retreat(this.consumedAfterSnapshot),this.consumedAfterSnapshot=0,this.active=!1,!0)}_consume(){return this.consumedAfterSnapshot++,this.preprocessor.advance()}_advanceBy(e){this.consumedAfterSnapshot+=e;for(let t=0;t<e;t++)this.preprocessor.advance()}_consumeSequenceIfMatch(e,t){return!!this.preprocessor.startsWith(e,t)&&(this._advanceBy(e.length-1),!0)}_createStartTagToken(){this.currentToken={type:m.START_TAG,tagName:"",tagID:P.UNKNOWN,selfClosing:!1,ackSelfClosing:!1,attrs:[],location:this.getCurrentLocation(1)}}_createEndTagToken(){this.currentToken={type:m.END_TAG,tagName:"",tagID:P.UNKNOWN,selfClosing:!1,ackSelfClosing:!1,attrs:[],location:this.getCurrentLocation(2)}}_createCommentToken(e){this.currentToken={type:m.COMMENT,data:"",location:this.getCurrentLocation(e)}}_createDoctypeToken(e){this.currentToken={type:m.DOCTYPE,name:e,forceQuirks:!1,publicId:null,systemId:null,location:this.currentLocation}}_createCharacterToken(e,t){this.currentCharacterToken={type:e,chars:t,location:this.currentLocation}}_createAttr(e){this.currentAttr={name:e,value:""},this.currentLocation=this.getCurrentLocation(0)}_leaveAttrName(){var e,t;const n=this.currentToken;null===v(n,this.currentAttr.name)?(n.attrs.push(this.currentAttr),n.location&&this.currentLocation&&((null!==(e=(t=n.location).attrs)&&void 0!==e?e:t.attrs=Object.create(null))[this.currentAttr.name]=this.currentLocation,this._leaveAttrValue())):this._err(f.duplicateAttribute)}_leaveAttrValue(){this.currentLocation&&(this.currentLocation.endLine=this.preprocessor.line,this.currentLocation.endCol=this.preprocessor.col,this.currentLocation.endOffset=this.preprocessor.offset)}prepareToken(e){this._emitCurrentCharacterToken(e.location),this.currentToken=null,e.location&&(e.location.endLine=this.preprocessor.line,e.location.endCol=this.preprocessor.col+1,e.location.endOffset=this.preprocessor.offset+1),this.currentLocation=this.getCurrentLocation(-1)}emitCurrentTagToken(){const e=this.currentToken;this.prepareToken(e),e.tagID=B(e.tagName),e.type===m.START_TAG?(this.lastStartTagName=e.tagName,this.handler.onStartTag(e)):(e.attrs.length>0&&this._err(f.endTagWithAttributes),e.selfClosing&&this._err(f.endTagWithTrailingSolidus),this.handler.onEndTag(e)),this.preprocessor.dropParsedChunk()}emitCurrentComment(e){this.prepareToken(e),this.handler.onComment(e),this.preprocessor.dropParsedChunk()}emitCurrentDoctype(e){this.prepareToken(e),this.handler.onDoctype(e),this.preprocessor.dropParsedChunk()}_emitCurrentCharacterToken(e){if(this.currentCharacterToken){switch(e&&this.currentCharacterToken.location&&(this.currentCharacterToken.location.endLine=e.startLine,this.currentCharacterToken.location.endCol=e.startCol,this.currentCharacterToken.location.endOffset=e.startOffset),this.currentCharacterToken.type){case m.CHARACTER:this.handler.onCharacter(this.currentCharacterToken);break;case m.NULL_CHARACTER:this.handler.onNullCharacter(this.currentCharacterToken);break;case m.WHITESPACE_CHARACTER:this.handler.onWhitespaceCharacter(this.currentCharacterToken)}this.currentCharacterToken=null}}_emitEOFToken(){const e=this.getCurrentLocation(0);e&&(e.endLine=e.startLine,e.endCol=e.startCol,e.endOffset=e.startOffset),this._emitCurrentCharacterToken(e),this.handler.onEof({type:m.EOF,location:e}),this.active=!1}_appendCharToCurrentCharacterToken(e,t){if(this.currentCharacterToken){if(this.currentCharacterToken.type===e)return void(this.currentCharacterToken.chars+=t);this.currentLocation=this.getCurrentLocation(0),this._emitCurrentCharacterToken(this.currentLocation),this.preprocessor.dropParsedChunk()}this._createCharacterToken(e,t)}_emitCodePoint(e){const t=G(e)?m.WHITESPACE_CHARACTER:e===s.NULL?m.NULL_CHARACTER:m.CHARACTER;this._appendCharToCurrentCharacterToken(t,String.fromCodePoint(e))}_emitChars(e){this._appendCharToCurrentCharacterToken(m.CHARACTER,e)}_startCharacterReference(){this.returnState=this.state,this.state=W.CHARACTER_REFERENCE,this.entityStartPos=this.preprocessor.pos,this.entityDecoder.startEntity(this._isCharacterReferenceInAttribute()?_.Attribute:_.Legacy)}_isCharacterReferenceInAttribute(){return this.returnState===W.ATTRIBUTE_VALUE_DOUBLE_QUOTED||this.returnState===W.ATTRIBUTE_VALUE_SINGLE_QUOTED||this.returnState===W.ATTRIBUTE_VALUE_UNQUOTED}_flushCodePointConsumedAsCharacterReference(e){this._isCharacterReferenceInAttribute()?this.currentAttr.value+=String.fromCodePoint(e):this._emitCodePoint(e)}_callState(e){switch(this.state){case W.DATA:this._stateData(e);break;case W.RCDATA:this._stateRcdata(e);break;case W.RAWTEXT:this._stateRawtext(e);break;case W.SCRIPT_DATA:this._stateScriptData(e);break;case W.PLAINTEXT:this._statePlaintext(e);break;case W.TAG_OPEN:this._stateTagOpen(e);break;case W.END_TAG_OPEN:this._stateEndTagOpen(e);break;case W.TAG_NAME:this._stateTagName(e);break;case W.RCDATA_LESS_THAN_SIGN:this._stateRcdataLessThanSign(e);break;case W.RCDATA_END_TAG_OPEN:this._stateRcdataEndTagOpen(e);break;case W.RCDATA_END_TAG_NAME:this._stateRcdataEndTagName(e);break;case W.RAWTEXT_LESS_THAN_SIGN:this._stateRawtextLessThanSign(e);break;case W.RAWTEXT_END_TAG_OPEN:this._stateRawtextEndTagOpen(e);break;case W.RAWTEXT_END_TAG_NAME:this._stateRawtextEndTagName(e);break;case W.SCRIPT_DATA_LESS_THAN_SIGN:this._stateScriptDataLessThanSign(e);break;case W.SCRIPT_DATA_END_TAG_OPEN:this._stateScriptDataEndTagOpen(e);break;case W.SCRIPT_DATA_END_TAG_NAME:this._stateScriptDataEndTagName(e);break;case W.SCRIPT_DATA_ESCAPE_START:this._stateScriptDataEscapeStart(e);break;case W.SCRIPT_DATA_ESCAPE_START_DASH:this._stateScriptDataEscapeStartDash(e);break;case W.SCRIPT_DATA_ESCAPED:this._stateScriptDataEscaped(e);break;case W.SCRIPT_DATA_ESCAPED_DASH:this._stateScriptDataEscapedDash(e);break;case W.SCRIPT_DATA_ESCAPED_DASH_DASH:this._stateScriptDataEscapedDashDash(e);break;case W.SCRIPT_DATA_ESCAPED_LESS_THAN_SIGN:this._stateScriptDataEscapedLessThanSign(e);break;case W.SCRIPT_DATA_ESCAPED_END_TAG_OPEN:this._stateScriptDataEscapedEndTagOpen(e);break;case W.SCRIPT_DATA_ESCAPED_END_TAG_NAME:this._stateScriptDataEscapedEndTagName(e);break;case W.SCRIPT_DATA_DOUBLE_ESCAPE_START:this._stateScriptDataDoubleEscapeStart(e);break;case W.SCRIPT_DATA_DOUBLE_ESCAPED:this._stateScriptDataDoubleEscaped(e);break;case W.SCRIPT_DATA_DOUBLE_ESCAPED_DASH:this._stateScriptDataDoubleEscapedDash(e);break;case W.SCRIPT_DATA_DOUBLE_ESCAPED_DASH_DASH:this._stateScriptDataDoubleEscapedDashDash(e);break;case W.SCRIPT_DATA_DOUBLE_ESCAPED_LESS_THAN_SIGN:this._stateScriptDataDoubleEscapedLessThanSign(e);break;case W.SCRIPT_DATA_DOUBLE_ESCAPE_END:this._stateScriptDataDoubleEscapeEnd(e);break;case W.BEFORE_ATTRIBUTE_NAME:this._stateBeforeAttributeName(e);break;case W.ATTRIBUTE_NAME:this._stateAttributeName(e);break;case W.AFTER_ATTRIBUTE_NAME:this._stateAfterAttributeName(e);break;case W.BEFORE_ATTRIBUTE_VALUE:this._stateBeforeAttributeValue(e);break;case W.ATTRIBUTE_VALUE_DOUBLE_QUOTED:this._stateAttributeValueDoubleQuoted(e);break;case W.ATTRIBUTE_VALUE_SINGLE_QUOTED:this._stateAttributeValueSingleQuoted(e);break;case W.ATTRIBUTE_VALUE_UNQUOTED:this._stateAttributeValueUnquoted(e);break;case W.AFTER_ATTRIBUTE_VALUE_QUOTED:this._stateAfterAttributeValueQuoted(e);break;case W.SELF_CLOSING_START_TAG:this._stateSelfClosingStartTag(e);break;case W.BOGUS_COMMENT:this._stateBogusComment(e);break;case W.MARKUP_DECLARATION_OPEN:this._stateMarkupDeclarationOpen(e);break;case W.COMMENT_START:this._stateCommentStart(e);break;case W.COMMENT_START_DASH:this._stateCommentStartDash(e);break;case W.COMMENT:this._stateComment(e);break;case W.COMMENT_LESS_THAN_SIGN:this._stateCommentLessThanSign(e);break;case W.COMMENT_LESS_THAN_SIGN_BANG:this._stateCommentLessThanSignBang(e);break;case W.COMMENT_LESS_THAN_SIGN_BANG_DASH:this._stateCommentLessThanSignBangDash(e);break;case W.COMMENT_LESS_THAN_SIGN_BANG_DASH_DASH:this._stateCommentLessThanSignBangDashDash(e);break;case W.COMMENT_END_DASH:this._stateCommentEndDash(e);break;case W.COMMENT_END:this._stateCommentEnd(e);break;case W.COMMENT_END_BANG:this._stateCommentEndBang(e);break;case W.DOCTYPE:this._stateDoctype(e);break;case W.BEFORE_DOCTYPE_NAME:this._stateBeforeDoctypeName(e);break;case W.DOCTYPE_NAME:this._stateDoctypeName(e);break;case W.AFTER_DOCTYPE_NAME:this._stateAfterDoctypeName(e);break;case W.AFTER_DOCTYPE_PUBLIC_KEYWORD:this._stateAfterDoctypePublicKeyword(e);break;case W.BEFORE_DOCTYPE_PUBLIC_IDENTIFIER:this._stateBeforeDoctypePublicIdentifier(e);break;case W.DOCTYPE_PUBLIC_IDENTIFIER_DOUBLE_QUOTED:this._stateDoctypePublicIdentifierDoubleQuoted(e);break;case W.DOCTYPE_PUBLIC_IDENTIFIER_SINGLE_QUOTED:this._stateDoctypePublicIdentifierSingleQuoted(e);break;case W.AFTER_DOCTYPE_PUBLIC_IDENTIFIER:this._stateAfterDoctypePublicIdentifier(e);break;case W.BETWEEN_DOCTYPE_PUBLIC_AND_SYSTEM_IDENTIFIERS:this._stateBetweenDoctypePublicAndSystemIdentifiers(e);break;case W.AFTER_DOCTYPE_SYSTEM_KEYWORD:this._stateAfterDoctypeSystemKeyword(e);break;case W.BEFORE_DOCTYPE_SYSTEM_IDENTIFIER:this._stateBeforeDoctypeSystemIdentifier(e);break;case W.DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED:this._stateDoctypeSystemIdentifierDoubleQuoted(e);break;case W.DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED:this._stateDoctypeSystemIdentifierSingleQuoted(e);break;case W.AFTER_DOCTYPE_SYSTEM_IDENTIFIER:this._stateAfterDoctypeSystemIdentifier(e);break;case W.BOGUS_DOCTYPE:this._stateBogusDoctype(e);break;case W.CDATA_SECTION:this._stateCdataSection(e);break;case W.CDATA_SECTION_BRACKET:this._stateCdataSectionBracket(e);break;case W.CDATA_SECTION_END:this._stateCdataSectionEnd(e);break;case W.CHARACTER_REFERENCE:this._stateCharacterReference();break;case W.AMBIGUOUS_AMPERSAND:this._stateAmbiguousAmpersand(e);break;default:throw new Error("Unknown state")}}_stateData(e){switch(e){case s.LESS_THAN_SIGN:this.state=W.TAG_OPEN;break;case s.AMPERSAND:this._startCharacterReference();break;case s.NULL:this._err(f.unexpectedNullCharacter),this._emitCodePoint(e);break;case s.EOF:this._emitEOFToken();break;default:this._emitCodePoint(e)}}_stateRcdata(e){switch(e){case s.AMPERSAND:this._startCharacterReference();break;case s.LESS_THAN_SIGN:this.state=W.RCDATA_LESS_THAN_SIGN;break;case s.NULL:this._err(f.unexpectedNullCharacter),this._emitChars(i);break;case s.EOF:this._emitEOFToken();break;default:this._emitCodePoint(e)}}_stateRawtext(e){switch(e){case s.LESS_THAN_SIGN:this.state=W.RAWTEXT_LESS_THAN_SIGN;break;case s.NULL:this._err(f.unexpectedNullCharacter),this._emitChars(i);break;case s.EOF:this._emitEOFToken();break;default:this._emitCodePoint(e)}}_stateScriptData(e){switch(e){case s.LESS_THAN_SIGN:this.state=W.SCRIPT_DATA_LESS_THAN_SIGN;break;case s.NULL:this._err(f.unexpectedNullCharacter),this._emitChars(i);break;case s.EOF:this._emitEOFToken();break;default:this._emitCodePoint(e)}}_statePlaintext(e){switch(e){case s.NULL:this._err(f.unexpectedNullCharacter),this._emitChars(i);break;case s.EOF:this._emitEOFToken();break;default:this._emitCodePoint(e)}}_stateTagOpen(e){if(V(e))this._createStartTagToken(),this.state=W.TAG_NAME,this._stateTagName(e);else switch(e){case s.EXCLAMATION_MARK:this.state=W.MARKUP_DECLARATION_OPEN;break;case s.SOLIDUS:this.state=W.END_TAG_OPEN;break;case s.QUESTION_MARK:this._err(f.unexpectedQuestionMarkInsteadOfTagName),this._createCommentToken(1),this.state=W.BOGUS_COMMENT,this._stateBogusComment(e);break;case s.EOF:this._err(f.eofBeforeTagName),this._emitChars("<"),this._emitEOFToken();break;default:this._err(f.invalidFirstCharacterOfTagName),this._emitChars("<"),this.state=W.DATA,this._stateData(e)}}_stateEndTagOpen(e){if(V(e))this._createEndTagToken(),this.state=W.TAG_NAME,this._stateTagName(e);else switch(e){case s.GREATER_THAN_SIGN:this._err(f.missingEndTagName),this.state=W.DATA;break;case s.EOF:this._err(f.eofBeforeTagName),this._emitChars("</"),this._emitEOFToken();break;default:this._err(f.invalidFirstCharacterOfTagName),this._createCommentToken(2),this.state=W.BOGUS_COMMENT,this._stateBogusComment(e)}}_stateTagName(e){const t=this.currentToken;switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:this.state=W.BEFORE_ATTRIBUTE_NAME;break;case s.SOLIDUS:this.state=W.SELF_CLOSING_START_TAG;break;case s.GREATER_THAN_SIGN:this.state=W.DATA,this.emitCurrentTagToken();break;case s.NULL:this._err(f.unexpectedNullCharacter),t.tagName+=i;break;case s.EOF:this._err(f.eofInTag),this._emitEOFToken();break;default:t.tagName+=String.fromCodePoint(q(e)?Z(e):e)}}_stateRcdataLessThanSign(e){e===s.SOLIDUS?this.state=W.RCDATA_END_TAG_OPEN:(this._emitChars("<"),this.state=W.RCDATA,this._stateRcdata(e))}_stateRcdataEndTagOpen(e){V(e)?(this.state=W.RCDATA_END_TAG_NAME,this._stateRcdataEndTagName(e)):(this._emitChars("</"),this.state=W.RCDATA,this._stateRcdata(e))}handleSpecialEndTag(e){if(!this.preprocessor.startsWith(this.lastStartTagName,!1))return!this._ensureHibernation();switch(this._createEndTagToken(),this.currentToken.tagName=this.lastStartTagName,this.preprocessor.peek(this.lastStartTagName.length)){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:return this._advanceBy(this.lastStartTagName.length),this.state=W.BEFORE_ATTRIBUTE_NAME,!1;case s.SOLIDUS:return this._advanceBy(this.lastStartTagName.length),this.state=W.SELF_CLOSING_START_TAG,!1;case s.GREATER_THAN_SIGN:return this._advanceBy(this.lastStartTagName.length),this.emitCurrentTagToken(),this.state=W.DATA,!1;default:return!this._ensureHibernation()}}_stateRcdataEndTagName(e){this.handleSpecialEndTag(e)&&(this._emitChars("</"),this.state=W.RCDATA,this._stateRcdata(e))}_stateRawtextLessThanSign(e){e===s.SOLIDUS?this.state=W.RAWTEXT_END_TAG_OPEN:(this._emitChars("<"),this.state=W.RAWTEXT,this._stateRawtext(e))}_stateRawtextEndTagOpen(e){V(e)?(this.state=W.RAWTEXT_END_TAG_NAME,this._stateRawtextEndTagName(e)):(this._emitChars("</"),this.state=W.RAWTEXT,this._stateRawtext(e))}_stateRawtextEndTagName(e){this.handleSpecialEndTag(e)&&(this._emitChars("</"),this.state=W.RAWTEXT,this._stateRawtext(e))}_stateScriptDataLessThanSign(e){switch(e){case s.SOLIDUS:this.state=W.SCRIPT_DATA_END_TAG_OPEN;break;case s.EXCLAMATION_MARK:this.state=W.SCRIPT_DATA_ESCAPE_START,this._emitChars("<!");break;default:this._emitChars("<"),this.state=W.SCRIPT_DATA,this._stateScriptData(e)}}_stateScriptDataEndTagOpen(e){V(e)?(this.state=W.SCRIPT_DATA_END_TAG_NAME,this._stateScriptDataEndTagName(e)):(this._emitChars("</"),this.state=W.SCRIPT_DATA,this._stateScriptData(e))}_stateScriptDataEndTagName(e){this.handleSpecialEndTag(e)&&(this._emitChars("</"),this.state=W.SCRIPT_DATA,this._stateScriptData(e))}_stateScriptDataEscapeStart(e){e===s.HYPHEN_MINUS?(this.state=W.SCRIPT_DATA_ESCAPE_START_DASH,this._emitChars("-")):(this.state=W.SCRIPT_DATA,this._stateScriptData(e))}_stateScriptDataEscapeStartDash(e){e===s.HYPHEN_MINUS?(this.state=W.SCRIPT_DATA_ESCAPED_DASH_DASH,this._emitChars("-")):(this.state=W.SCRIPT_DATA,this._stateScriptData(e))}_stateScriptDataEscaped(e){switch(e){case s.HYPHEN_MINUS:this.state=W.SCRIPT_DATA_ESCAPED_DASH,this._emitChars("-");break;case s.LESS_THAN_SIGN:this.state=W.SCRIPT_DATA_ESCAPED_LESS_THAN_SIGN;break;case s.NULL:this._err(f.unexpectedNullCharacter),this._emitChars(i);break;case s.EOF:this._err(f.eofInScriptHtmlCommentLikeText),this._emitEOFToken();break;default:this._emitCodePoint(e)}}_stateScriptDataEscapedDash(e){switch(e){case s.HYPHEN_MINUS:this.state=W.SCRIPT_DATA_ESCAPED_DASH_DASH,this._emitChars("-");break;case s.LESS_THAN_SIGN:this.state=W.SCRIPT_DATA_ESCAPED_LESS_THAN_SIGN;break;case s.NULL:this._err(f.unexpectedNullCharacter),this.state=W.SCRIPT_DATA_ESCAPED,this._emitChars(i);break;case s.EOF:this._err(f.eofInScriptHtmlCommentLikeText),this._emitEOFToken();break;default:this.state=W.SCRIPT_DATA_ESCAPED,this._emitCodePoint(e)}}_stateScriptDataEscapedDashDash(e){switch(e){case s.HYPHEN_MINUS:this._emitChars("-");break;case s.LESS_THAN_SIGN:this.state=W.SCRIPT_DATA_ESCAPED_LESS_THAN_SIGN;break;case s.GREATER_THAN_SIGN:this.state=W.SCRIPT_DATA,this._emitChars(">");break;case s.NULL:this._err(f.unexpectedNullCharacter),this.state=W.SCRIPT_DATA_ESCAPED,this._emitChars(i);break;case s.EOF:this._err(f.eofInScriptHtmlCommentLikeText),this._emitEOFToken();break;default:this.state=W.SCRIPT_DATA_ESCAPED,this._emitCodePoint(e)}}_stateScriptDataEscapedLessThanSign(e){e===s.SOLIDUS?this.state=W.SCRIPT_DATA_ESCAPED_END_TAG_OPEN:V(e)?(this._emitChars("<"),this.state=W.SCRIPT_DATA_DOUBLE_ESCAPE_START,this._stateScriptDataDoubleEscapeStart(e)):(this._emitChars("<"),this.state=W.SCRIPT_DATA_ESCAPED,this._stateScriptDataEscaped(e))}_stateScriptDataEscapedEndTagOpen(e){V(e)?(this.state=W.SCRIPT_DATA_ESCAPED_END_TAG_NAME,this._stateScriptDataEscapedEndTagName(e)):(this._emitChars("</"),this.state=W.SCRIPT_DATA_ESCAPED,this._stateScriptDataEscaped(e))}_stateScriptDataEscapedEndTagName(e){this.handleSpecialEndTag(e)&&(this._emitChars("</"),this.state=W.SCRIPT_DATA_ESCAPED,this._stateScriptDataEscaped(e))}_stateScriptDataDoubleEscapeStart(e){if(this.preprocessor.startsWith(u,!1)&&K(this.preprocessor.peek(6))){this._emitCodePoint(e);for(let e=0;e<6;e++)this._emitCodePoint(this._consume());this.state=W.SCRIPT_DATA_DOUBLE_ESCAPED}else this._ensureHibernation()||(this.state=W.SCRIPT_DATA_ESCAPED,this._stateScriptDataEscaped(e))}_stateScriptDataDoubleEscaped(e){switch(e){case s.HYPHEN_MINUS:this.state=W.SCRIPT_DATA_DOUBLE_ESCAPED_DASH,this._emitChars("-");break;case s.LESS_THAN_SIGN:this.state=W.SCRIPT_DATA_DOUBLE_ESCAPED_LESS_THAN_SIGN,this._emitChars("<");break;case s.NULL:this._err(f.unexpectedNullCharacter),this._emitChars(i);break;case s.EOF:this._err(f.eofInScriptHtmlCommentLikeText),this._emitEOFToken();break;default:this._emitCodePoint(e)}}_stateScriptDataDoubleEscapedDash(e){switch(e){case s.HYPHEN_MINUS:this.state=W.SCRIPT_DATA_DOUBLE_ESCAPED_DASH_DASH,this._emitChars("-");break;case s.LESS_THAN_SIGN:this.state=W.SCRIPT_DATA_DOUBLE_ESCAPED_LESS_THAN_SIGN,this._emitChars("<");break;case s.NULL:this._err(f.unexpectedNullCharacter),this.state=W.SCRIPT_DATA_DOUBLE_ESCAPED,this._emitChars(i);break;case s.EOF:this._err(f.eofInScriptHtmlCommentLikeText),this._emitEOFToken();break;default:this.state=W.SCRIPT_DATA_DOUBLE_ESCAPED,this._emitCodePoint(e)}}_stateScriptDataDoubleEscapedDashDash(e){switch(e){case s.HYPHEN_MINUS:this._emitChars("-");break;case s.LESS_THAN_SIGN:this.state=W.SCRIPT_DATA_DOUBLE_ESCAPED_LESS_THAN_SIGN,this._emitChars("<");break;case s.GREATER_THAN_SIGN:this.state=W.SCRIPT_DATA,this._emitChars(">");break;case s.NULL:this._err(f.unexpectedNullCharacter),this.state=W.SCRIPT_DATA_DOUBLE_ESCAPED,this._emitChars(i);break;case s.EOF:this._err(f.eofInScriptHtmlCommentLikeText),this._emitEOFToken();break;default:this.state=W.SCRIPT_DATA_DOUBLE_ESCAPED,this._emitCodePoint(e)}}_stateScriptDataDoubleEscapedLessThanSign(e){e===s.SOLIDUS?(this.state=W.SCRIPT_DATA_DOUBLE_ESCAPE_END,this._emitChars("/")):(this.state=W.SCRIPT_DATA_DOUBLE_ESCAPED,this._stateScriptDataDoubleEscaped(e))}_stateScriptDataDoubleEscapeEnd(e){if(this.preprocessor.startsWith(u,!1)&&K(this.preprocessor.peek(6))){this._emitCodePoint(e);for(let e=0;e<6;e++)this._emitCodePoint(this._consume());this.state=W.SCRIPT_DATA_ESCAPED}else this._ensureHibernation()||(this.state=W.SCRIPT_DATA_DOUBLE_ESCAPED,this._stateScriptDataDoubleEscaped(e))}_stateBeforeAttributeName(e){switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:break;case s.SOLIDUS:case s.GREATER_THAN_SIGN:case s.EOF:this.state=W.AFTER_ATTRIBUTE_NAME,this._stateAfterAttributeName(e);break;case s.EQUALS_SIGN:this._err(f.unexpectedEqualsSignBeforeAttributeName),this._createAttr("="),this.state=W.ATTRIBUTE_NAME;break;default:this._createAttr(""),this.state=W.ATTRIBUTE_NAME,this._stateAttributeName(e)}}_stateAttributeName(e){switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:case s.SOLIDUS:case s.GREATER_THAN_SIGN:case s.EOF:this._leaveAttrName(),this.state=W.AFTER_ATTRIBUTE_NAME,this._stateAfterAttributeName(e);break;case s.EQUALS_SIGN:this._leaveAttrName(),this.state=W.BEFORE_ATTRIBUTE_VALUE;break;case s.QUOTATION_MARK:case s.APOSTROPHE:case s.LESS_THAN_SIGN:this._err(f.unexpectedCharacterInAttributeName),this.currentAttr.name+=String.fromCodePoint(e);break;case s.NULL:this._err(f.unexpectedNullCharacter),this.currentAttr.name+=i;break;default:this.currentAttr.name+=String.fromCodePoint(q(e)?Z(e):e)}}_stateAfterAttributeName(e){switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:break;case s.SOLIDUS:this.state=W.SELF_CLOSING_START_TAG;break;case s.EQUALS_SIGN:this.state=W.BEFORE_ATTRIBUTE_VALUE;break;case s.GREATER_THAN_SIGN:this.state=W.DATA,this.emitCurrentTagToken();break;case s.EOF:this._err(f.eofInTag),this._emitEOFToken();break;default:this._createAttr(""),this.state=W.ATTRIBUTE_NAME,this._stateAttributeName(e)}}_stateBeforeAttributeValue(e){switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:break;case s.QUOTATION_MARK:this.state=W.ATTRIBUTE_VALUE_DOUBLE_QUOTED;break;case s.APOSTROPHE:this.state=W.ATTRIBUTE_VALUE_SINGLE_QUOTED;break;case s.GREATER_THAN_SIGN:this._err(f.missingAttributeValue),this.state=W.DATA,this.emitCurrentTagToken();break;default:this.state=W.ATTRIBUTE_VALUE_UNQUOTED,this._stateAttributeValueUnquoted(e)}}_stateAttributeValueDoubleQuoted(e){switch(e){case s.QUOTATION_MARK:this.state=W.AFTER_ATTRIBUTE_VALUE_QUOTED;break;case s.AMPERSAND:this._startCharacterReference();break;case s.NULL:this._err(f.unexpectedNullCharacter),this.currentAttr.value+=i;break;case s.EOF:this._err(f.eofInTag),this._emitEOFToken();break;default:this.currentAttr.value+=String.fromCodePoint(e)}}_stateAttributeValueSingleQuoted(e){switch(e){case s.APOSTROPHE:this.state=W.AFTER_ATTRIBUTE_VALUE_QUOTED;break;case s.AMPERSAND:this._startCharacterReference();break;case s.NULL:this._err(f.unexpectedNullCharacter),this.currentAttr.value+=i;break;case s.EOF:this._err(f.eofInTag),this._emitEOFToken();break;default:this.currentAttr.value+=String.fromCodePoint(e)}}_stateAttributeValueUnquoted(e){switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:this._leaveAttrValue(),this.state=W.BEFORE_ATTRIBUTE_NAME;break;case s.AMPERSAND:this._startCharacterReference();break;case s.GREATER_THAN_SIGN:this._leaveAttrValue(),this.state=W.DATA,this.emitCurrentTagToken();break;case s.NULL:this._err(f.unexpectedNullCharacter),this.currentAttr.value+=i;break;case s.QUOTATION_MARK:case s.APOSTROPHE:case s.LESS_THAN_SIGN:case s.EQUALS_SIGN:case s.GRAVE_ACCENT:this._err(f.unexpectedCharacterInUnquotedAttributeValue),this.currentAttr.value+=String.fromCodePoint(e);break;case s.EOF:this._err(f.eofInTag),this._emitEOFToken();break;default:this.currentAttr.value+=String.fromCodePoint(e)}}_stateAfterAttributeValueQuoted(e){switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:this._leaveAttrValue(),this.state=W.BEFORE_ATTRIBUTE_NAME;break;case s.SOLIDUS:this._leaveAttrValue(),this.state=W.SELF_CLOSING_START_TAG;break;case s.GREATER_THAN_SIGN:this._leaveAttrValue(),this.state=W.DATA,this.emitCurrentTagToken();break;case s.EOF:this._err(f.eofInTag),this._emitEOFToken();break;default:this._err(f.missingWhitespaceBetweenAttributes),this.state=W.BEFORE_ATTRIBUTE_NAME,this._stateBeforeAttributeName(e)}}_stateSelfClosingStartTag(e){switch(e){case s.GREATER_THAN_SIGN:this.currentToken.selfClosing=!0,this.state=W.DATA,this.emitCurrentTagToken();break;case s.EOF:this._err(f.eofInTag),this._emitEOFToken();break;default:this._err(f.unexpectedSolidusInTag),this.state=W.BEFORE_ATTRIBUTE_NAME,this._stateBeforeAttributeName(e)}}_stateBogusComment(e){const t=this.currentToken;switch(e){case s.GREATER_THAN_SIGN:this.state=W.DATA,this.emitCurrentComment(t);break;case s.EOF:this.emitCurrentComment(t),this._emitEOFToken();break;case s.NULL:this._err(f.unexpectedNullCharacter),t.data+=i;break;default:t.data+=String.fromCodePoint(e)}}_stateMarkupDeclarationOpen(e){this._consumeSequenceIfMatch("--",!0)?(this._createCommentToken(3),this.state=W.COMMENT_START):this._consumeSequenceIfMatch(c,!1)?(this.currentLocation=this.getCurrentLocation(8),this.state=W.DOCTYPE):this._consumeSequenceIfMatch(l,!0)?this.inForeignNode?this.state=W.CDATA_SECTION:(this._err(f.cdataInHtmlContent),this._createCommentToken(8),this.currentToken.data="[CDATA[",this.state=W.BOGUS_COMMENT):this._ensureHibernation()||(this._err(f.incorrectlyOpenedComment),this._createCommentToken(2),this.state=W.BOGUS_COMMENT,this._stateBogusComment(e))}_stateCommentStart(e){switch(e){case s.HYPHEN_MINUS:this.state=W.COMMENT_START_DASH;break;case s.GREATER_THAN_SIGN:{this._err(f.abruptClosingOfEmptyComment),this.state=W.DATA;const e=this.currentToken;this.emitCurrentComment(e);break}default:this.state=W.COMMENT,this._stateComment(e)}}_stateCommentStartDash(e){const t=this.currentToken;switch(e){case s.HYPHEN_MINUS:this.state=W.COMMENT_END;break;case s.GREATER_THAN_SIGN:this._err(f.abruptClosingOfEmptyComment),this.state=W.DATA,this.emitCurrentComment(t);break;case s.EOF:this._err(f.eofInComment),this.emitCurrentComment(t),this._emitEOFToken();break;default:t.data+="-",this.state=W.COMMENT,this._stateComment(e)}}_stateComment(e){const t=this.currentToken;switch(e){case s.HYPHEN_MINUS:this.state=W.COMMENT_END_DASH;break;case s.LESS_THAN_SIGN:t.data+="<",this.state=W.COMMENT_LESS_THAN_SIGN;break;case s.NULL:this._err(f.unexpectedNullCharacter),t.data+=i;break;case s.EOF:this._err(f.eofInComment),this.emitCurrentComment(t),this._emitEOFToken();break;default:t.data+=String.fromCodePoint(e)}}_stateCommentLessThanSign(e){const t=this.currentToken;switch(e){case s.EXCLAMATION_MARK:t.data+="!",this.state=W.COMMENT_LESS_THAN_SIGN_BANG;break;case s.LESS_THAN_SIGN:t.data+="<";break;default:this.state=W.COMMENT,this._stateComment(e)}}_stateCommentLessThanSignBang(e){e===s.HYPHEN_MINUS?this.state=W.COMMENT_LESS_THAN_SIGN_BANG_DASH:(this.state=W.COMMENT,this._stateComment(e))}_stateCommentLessThanSignBangDash(e){e===s.HYPHEN_MINUS?this.state=W.COMMENT_LESS_THAN_SIGN_BANG_DASH_DASH:(this.state=W.COMMENT_END_DASH,this._stateCommentEndDash(e))}_stateCommentLessThanSignBangDashDash(e){e!==s.GREATER_THAN_SIGN&&e!==s.EOF&&this._err(f.nestedComment),this.state=W.COMMENT_END,this._stateCommentEnd(e)}_stateCommentEndDash(e){const t=this.currentToken;switch(e){case s.HYPHEN_MINUS:this.state=W.COMMENT_END;break;case s.EOF:this._err(f.eofInComment),this.emitCurrentComment(t),this._emitEOFToken();break;default:t.data+="-",this.state=W.COMMENT,this._stateComment(e)}}_stateCommentEnd(e){const t=this.currentToken;switch(e){case s.GREATER_THAN_SIGN:this.state=W.DATA,this.emitCurrentComment(t);break;case s.EXCLAMATION_MARK:this.state=W.COMMENT_END_BANG;break;case s.HYPHEN_MINUS:t.data+="-";break;case s.EOF:this._err(f.eofInComment),this.emitCurrentComment(t),this._emitEOFToken();break;default:t.data+="--",this.state=W.COMMENT,this._stateComment(e)}}_stateCommentEndBang(e){const t=this.currentToken;switch(e){case s.HYPHEN_MINUS:t.data+="--!",this.state=W.COMMENT_END_DASH;break;case s.GREATER_THAN_SIGN:this._err(f.incorrectlyClosedComment),this.state=W.DATA,this.emitCurrentComment(t);break;case s.EOF:this._err(f.eofInComment),this.emitCurrentComment(t),this._emitEOFToken();break;default:t.data+="--!",this.state=W.COMMENT,this._stateComment(e)}}_stateDoctype(e){switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:this.state=W.BEFORE_DOCTYPE_NAME;break;case s.GREATER_THAN_SIGN:this.state=W.BEFORE_DOCTYPE_NAME,this._stateBeforeDoctypeName(e);break;case s.EOF:{this._err(f.eofInDoctype),this._createDoctypeToken(null);const e=this.currentToken;e.forceQuirks=!0,this.emitCurrentDoctype(e),this._emitEOFToken();break}default:this._err(f.missingWhitespaceBeforeDoctypeName),this.state=W.BEFORE_DOCTYPE_NAME,this._stateBeforeDoctypeName(e)}}_stateBeforeDoctypeName(e){if(q(e))this._createDoctypeToken(String.fromCharCode(Z(e))),this.state=W.DOCTYPE_NAME;else switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:break;case s.NULL:this._err(f.unexpectedNullCharacter),this._createDoctypeToken(i),this.state=W.DOCTYPE_NAME;break;case s.GREATER_THAN_SIGN:{this._err(f.missingDoctypeName),this._createDoctypeToken(null);const e=this.currentToken;e.forceQuirks=!0,this.emitCurrentDoctype(e),this.state=W.DATA;break}case s.EOF:{this._err(f.eofInDoctype),this._createDoctypeToken(null);const e=this.currentToken;e.forceQuirks=!0,this.emitCurrentDoctype(e),this._emitEOFToken();break}default:this._createDoctypeToken(String.fromCodePoint(e)),this.state=W.DOCTYPE_NAME}}_stateDoctypeName(e){const t=this.currentToken;switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:this.state=W.AFTER_DOCTYPE_NAME;break;case s.GREATER_THAN_SIGN:this.state=W.DATA,this.emitCurrentDoctype(t);break;case s.NULL:this._err(f.unexpectedNullCharacter),t.name+=i;break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:t.name+=String.fromCodePoint(q(e)?Z(e):e)}}_stateAfterDoctypeName(e){const t=this.currentToken;switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:break;case s.GREATER_THAN_SIGN:this.state=W.DATA,this.emitCurrentDoctype(t);break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:this._consumeSequenceIfMatch("public",!1)?this.state=W.AFTER_DOCTYPE_PUBLIC_KEYWORD:this._consumeSequenceIfMatch("system",!1)?this.state=W.AFTER_DOCTYPE_SYSTEM_KEYWORD:this._ensureHibernation()||(this._err(f.invalidCharacterSequenceAfterDoctypeName),t.forceQuirks=!0,this.state=W.BOGUS_DOCTYPE,this._stateBogusDoctype(e))}}_stateAfterDoctypePublicKeyword(e){const t=this.currentToken;switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:this.state=W.BEFORE_DOCTYPE_PUBLIC_IDENTIFIER;break;case s.QUOTATION_MARK:this._err(f.missingWhitespaceAfterDoctypePublicKeyword),t.publicId="",this.state=W.DOCTYPE_PUBLIC_IDENTIFIER_DOUBLE_QUOTED;break;case s.APOSTROPHE:this._err(f.missingWhitespaceAfterDoctypePublicKeyword),t.publicId="",this.state=W.DOCTYPE_PUBLIC_IDENTIFIER_SINGLE_QUOTED;break;case s.GREATER_THAN_SIGN:this._err(f.missingDoctypePublicIdentifier),t.forceQuirks=!0,this.state=W.DATA,this.emitCurrentDoctype(t);break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:this._err(f.missingQuoteBeforeDoctypePublicIdentifier),t.forceQuirks=!0,this.state=W.BOGUS_DOCTYPE,this._stateBogusDoctype(e)}}_stateBeforeDoctypePublicIdentifier(e){const t=this.currentToken;switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:break;case s.QUOTATION_MARK:t.publicId="",this.state=W.DOCTYPE_PUBLIC_IDENTIFIER_DOUBLE_QUOTED;break;case s.APOSTROPHE:t.publicId="",this.state=W.DOCTYPE_PUBLIC_IDENTIFIER_SINGLE_QUOTED;break;case s.GREATER_THAN_SIGN:this._err(f.missingDoctypePublicIdentifier),t.forceQuirks=!0,this.state=W.DATA,this.emitCurrentDoctype(t);break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:this._err(f.missingQuoteBeforeDoctypePublicIdentifier),t.forceQuirks=!0,this.state=W.BOGUS_DOCTYPE,this._stateBogusDoctype(e)}}_stateDoctypePublicIdentifierDoubleQuoted(e){const t=this.currentToken;switch(e){case s.QUOTATION_MARK:this.state=W.AFTER_DOCTYPE_PUBLIC_IDENTIFIER;break;case s.NULL:this._err(f.unexpectedNullCharacter),t.publicId+=i;break;case s.GREATER_THAN_SIGN:this._err(f.abruptDoctypePublicIdentifier),t.forceQuirks=!0,this.emitCurrentDoctype(t),this.state=W.DATA;break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:t.publicId+=String.fromCodePoint(e)}}_stateDoctypePublicIdentifierSingleQuoted(e){const t=this.currentToken;switch(e){case s.APOSTROPHE:this.state=W.AFTER_DOCTYPE_PUBLIC_IDENTIFIER;break;case s.NULL:this._err(f.unexpectedNullCharacter),t.publicId+=i;break;case s.GREATER_THAN_SIGN:this._err(f.abruptDoctypePublicIdentifier),t.forceQuirks=!0,this.emitCurrentDoctype(t),this.state=W.DATA;break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:t.publicId+=String.fromCodePoint(e)}}_stateAfterDoctypePublicIdentifier(e){const t=this.currentToken;switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:this.state=W.BETWEEN_DOCTYPE_PUBLIC_AND_SYSTEM_IDENTIFIERS;break;case s.GREATER_THAN_SIGN:this.state=W.DATA,this.emitCurrentDoctype(t);break;case s.QUOTATION_MARK:this._err(f.missingWhitespaceBetweenDoctypePublicAndSystemIdentifiers),t.systemId="",this.state=W.DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED;break;case s.APOSTROPHE:this._err(f.missingWhitespaceBetweenDoctypePublicAndSystemIdentifiers),t.systemId="",this.state=W.DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED;break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:this._err(f.missingQuoteBeforeDoctypeSystemIdentifier),t.forceQuirks=!0,this.state=W.BOGUS_DOCTYPE,this._stateBogusDoctype(e)}}_stateBetweenDoctypePublicAndSystemIdentifiers(e){const t=this.currentToken;switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:break;case s.GREATER_THAN_SIGN:this.emitCurrentDoctype(t),this.state=W.DATA;break;case s.QUOTATION_MARK:t.systemId="",this.state=W.DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED;break;case s.APOSTROPHE:t.systemId="",this.state=W.DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED;break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:this._err(f.missingQuoteBeforeDoctypeSystemIdentifier),t.forceQuirks=!0,this.state=W.BOGUS_DOCTYPE,this._stateBogusDoctype(e)}}_stateAfterDoctypeSystemKeyword(e){const t=this.currentToken;switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:this.state=W.BEFORE_DOCTYPE_SYSTEM_IDENTIFIER;break;case s.QUOTATION_MARK:this._err(f.missingWhitespaceAfterDoctypeSystemKeyword),t.systemId="",this.state=W.DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED;break;case s.APOSTROPHE:this._err(f.missingWhitespaceAfterDoctypeSystemKeyword),t.systemId="",this.state=W.DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED;break;case s.GREATER_THAN_SIGN:this._err(f.missingDoctypeSystemIdentifier),t.forceQuirks=!0,this.state=W.DATA,this.emitCurrentDoctype(t);break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:this._err(f.missingQuoteBeforeDoctypeSystemIdentifier),t.forceQuirks=!0,this.state=W.BOGUS_DOCTYPE,this._stateBogusDoctype(e)}}_stateBeforeDoctypeSystemIdentifier(e){const t=this.currentToken;switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:break;case s.QUOTATION_MARK:t.systemId="",this.state=W.DOCTYPE_SYSTEM_IDENTIFIER_DOUBLE_QUOTED;break;case s.APOSTROPHE:t.systemId="",this.state=W.DOCTYPE_SYSTEM_IDENTIFIER_SINGLE_QUOTED;break;case s.GREATER_THAN_SIGN:this._err(f.missingDoctypeSystemIdentifier),t.forceQuirks=!0,this.state=W.DATA,this.emitCurrentDoctype(t);break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:this._err(f.missingQuoteBeforeDoctypeSystemIdentifier),t.forceQuirks=!0,this.state=W.BOGUS_DOCTYPE,this._stateBogusDoctype(e)}}_stateDoctypeSystemIdentifierDoubleQuoted(e){const t=this.currentToken;switch(e){case s.QUOTATION_MARK:this.state=W.AFTER_DOCTYPE_SYSTEM_IDENTIFIER;break;case s.NULL:this._err(f.unexpectedNullCharacter),t.systemId+=i;break;case s.GREATER_THAN_SIGN:this._err(f.abruptDoctypeSystemIdentifier),t.forceQuirks=!0,this.emitCurrentDoctype(t),this.state=W.DATA;break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:t.systemId+=String.fromCodePoint(e)}}_stateDoctypeSystemIdentifierSingleQuoted(e){const t=this.currentToken;switch(e){case s.APOSTROPHE:this.state=W.AFTER_DOCTYPE_SYSTEM_IDENTIFIER;break;case s.NULL:this._err(f.unexpectedNullCharacter),t.systemId+=i;break;case s.GREATER_THAN_SIGN:this._err(f.abruptDoctypeSystemIdentifier),t.forceQuirks=!0,this.emitCurrentDoctype(t),this.state=W.DATA;break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:t.systemId+=String.fromCodePoint(e)}}_stateAfterDoctypeSystemIdentifier(e){const t=this.currentToken;switch(e){case s.SPACE:case s.LINE_FEED:case s.TABULATION:case s.FORM_FEED:break;case s.GREATER_THAN_SIGN:this.emitCurrentDoctype(t),this.state=W.DATA;break;case s.EOF:this._err(f.eofInDoctype),t.forceQuirks=!0,this.emitCurrentDoctype(t),this._emitEOFToken();break;default:this._err(f.unexpectedCharacterAfterDoctypeSystemIdentifier),this.state=W.BOGUS_DOCTYPE,this._stateBogusDoctype(e)}}_stateBogusDoctype(e){const t=this.currentToken;switch(e){case s.GREATER_THAN_SIGN:this.emitCurrentDoctype(t),this.state=W.DATA;break;case s.NULL:this._err(f.unexpectedNullCharacter);break;case s.EOF:this.emitCurrentDoctype(t),this._emitEOFToken()}}_stateCdataSection(e){switch(e){case s.RIGHT_SQUARE_BRACKET:this.state=W.CDATA_SECTION_BRACKET;break;case s.EOF:this._err(f.eofInCdata),this._emitEOFToken();break;default:this._emitCodePoint(e)}}_stateCdataSectionBracket(e){e===s.RIGHT_SQUARE_BRACKET?this.state=W.CDATA_SECTION_END:(this._emitChars("]"),this.state=W.CDATA_SECTION,this._stateCdataSection(e))}_stateCdataSectionEnd(e){switch(e){case s.GREATER_THAN_SIGN:this.state=W.DATA;break;case s.RIGHT_SQUARE_BRACKET:this._emitChars("]");break;default:this._emitChars("]]"),this.state=W.CDATA_SECTION,this._stateCdataSection(e)}}_stateCharacterReference(){let e=this.entityDecoder.write(this.preprocessor.html,this.preprocessor.pos);if(e<0){if(!this.preprocessor.lastChunkWritten)return this.active=!1,this.preprocessor.pos=this.preprocessor.html.length-1,this.consumedAfterSnapshot=0,void(this.preprocessor.endOfChunkHit=!0);e=this.entityDecoder.end()}0===e?(this.preprocessor.pos=this.entityStartPos,this._flushCodePointConsumedAsCharacterReference(s.AMPERSAND),this.state=!this._isCharacterReferenceInAttribute()&&H(this.preprocessor.peek(1))?W.AMBIGUOUS_AMPERSAND:this.returnState):this.state=this.returnState}_stateAmbiguousAmpersand(e){H(e)?this._flushCodePointConsumedAsCharacterReference(e):(e===s.SEMICOLON&&this._err(f.unknownNamedCharacterReference),this.state=this.returnState,this._callState(e))}}const J=new Set([P.DD,P.DT,P.LI,P.OPTGROUP,P.OPTION,P.P,P.RB,P.RP,P.RT,P.RTC]),ee=new Set([...J,P.CAPTION,P.COLGROUP,P.TBODY,P.TD,P.TFOOT,P.TH,P.THEAD,P.TR]),te=new Set([P.APPLET,P.CAPTION,P.HTML,P.MARQUEE,P.OBJECT,P.TABLE,P.TD,P.TEMPLATE,P.TH]),ne=new Set([...te,P.OL,P.UL]),re=new Set([...te,P.BUTTON]),ae=new Set([P.ANNOTATION_XML,P.MI,P.MN,P.MO,P.MS,P.MTEXT]),oe=new Set([P.DESC,P.FOREIGN_OBJECT,P.TITLE]),ie=new Set([P.TR,P.TEMPLATE,P.HTML]),se=new Set([P.TBODY,P.TFOOT,P.THEAD,P.TEMPLATE,P.HTML]),le=new Set([P.TABLE,P.TEMPLATE,P.HTML]),ce=new Set([P.TD,P.TH]);class ue{get currentTmplContentOrNode(){return this._isInTemplate()?this.treeAdapter.getTemplateContent(this.current):this.current}constructor(e,t,n){this.treeAdapter=t,this.handler=n,this.items=[],this.tagIDs=[],this.stackTop=-1,this.tmplCount=0,this.currentTagId=P.UNKNOWN,this.current=e}_indexOf(e){return this.items.lastIndexOf(e,this.stackTop)}_isInTemplate(){return this.currentTagId===P.TEMPLATE&&this.treeAdapter.getNamespaceURI(this.current)===C.HTML}_updateCurrentElement(){this.current=this.items[this.stackTop],this.currentTagId=this.tagIDs[this.stackTop]}push(e,t){this.stackTop++,this.items[this.stackTop]=e,this.current=e,this.tagIDs[this.stackTop]=t,this.currentTagId=t,this._isInTemplate()&&this.tmplCount++,this.handler.onItemPush(e,t,!0)}pop(){const e=this.current;this.tmplCount>0&&this._isInTemplate()&&this.tmplCount--,this.stackTop--,this._updateCurrentElement(),this.handler.onItemPop(e,!0)}replace(e,t){const n=this._indexOf(e);this.items[n]=t,n===this.stackTop&&(this.current=t)}insertAfter(e,t,n){const r=this._indexOf(e)+1;this.items.splice(r,0,t),this.tagIDs.splice(r,0,n),this.stackTop++,r===this.stackTop&&this._updateCurrentElement(),this.handler.onItemPush(this.current,this.currentTagId,r===this.stackTop)}popUntilTagNamePopped(e){let t=this.stackTop+1;do{t=this.tagIDs.lastIndexOf(e,t-1)}while(t>0&&this.treeAdapter.getNamespaceURI(this.items[t])!==C.HTML);this.shortenToLength(t<0?0:t)}shortenToLength(e){for(;this.stackTop>=e;){const t=this.current;this.tmplCount>0&&this._isInTemplate()&&(this.tmplCount-=1),this.stackTop--,this._updateCurrentElement(),this.handler.onItemPop(t,this.stackTop<e)}}popUntilElementPopped(e){const t=this._indexOf(e);this.shortenToLength(t<0?0:t)}popUntilPopped(e,t){const n=this._indexOfTagNames(e,t);this.shortenToLength(n<0?0:n)}popUntilNumberedHeaderPopped(){this.popUntilPopped(U,C.HTML)}popUntilTableCellPopped(){this.popUntilPopped(ce,C.HTML)}popAllUpToHtmlElement(){this.tmplCount=0,this.shortenToLength(1)}_indexOfTagNames(e,t){for(let n=this.stackTop;n>=0;n--)if(e.has(this.tagIDs[n])&&this.treeAdapter.getNamespaceURI(this.items[n])===t)return n;return-1}clearBackTo(e,t){const n=this._indexOfTagNames(e,t);this.shortenToLength(n+1)}clearBackToTableContext(){this.clearBackTo(le,C.HTML)}clearBackToTableBodyContext(){this.clearBackTo(se,C.HTML)}clearBackToTableRowContext(){this.clearBackTo(ie,C.HTML)}remove(e){const t=this._indexOf(e);t>=0&&(t===this.stackTop?this.pop():(this.items.splice(t,1),this.tagIDs.splice(t,1),this.stackTop--,this._updateCurrentElement(),this.handler.onItemPop(e,!1)))}tryPeekProperlyNestedBodyElement(){return this.stackTop>=1&&this.tagIDs[1]===P.BODY?this.items[1]:null}contains(e){return this._indexOf(e)>-1}getCommonAncestor(e){const t=this._indexOf(e)-1;return t>=0?this.items[t]:null}isRootHtmlElementCurrent(){return 0===this.stackTop&&this.tagIDs[0]===P.HTML}hasInDynamicScope(e,t){for(let n=this.stackTop;n>=0;n--){const r=this.tagIDs[n];switch(this.treeAdapter.getNamespaceURI(this.items[n])){case C.HTML:if(r===e)return!0;if(t.has(r))return!1;break;case C.SVG:if(oe.has(r))return!1;break;case C.MATHML:if(ae.has(r))return!1}}return!0}hasInScope(e){return this.hasInDynamicScope(e,te)}hasInListItemScope(e){return this.hasInDynamicScope(e,ne)}hasInButtonScope(e){return this.hasInDynamicScope(e,re)}hasNumberedHeaderInScope(){for(let e=this.stackTop;e>=0;e--){const t=this.tagIDs[e];switch(this.treeAdapter.getNamespaceURI(this.items[e])){case C.HTML:if(U.has(t))return!0;if(te.has(t))return!1;break;case C.SVG:if(oe.has(t))return!1;break;case C.MATHML:if(ae.has(t))return!1}}return!0}hasInTableScope(e){for(let t=this.stackTop;t>=0;t--)if(this.treeAdapter.getNamespaceURI(this.items[t])===C.HTML)switch(this.tagIDs[t]){case e:return!0;case P.TABLE:case P.HTML:return!1}return!0}hasTableBodyContextInTableScope(){for(let e=this.stackTop;e>=0;e--)if(this.treeAdapter.getNamespaceURI(this.items[e])===C.HTML)switch(this.tagIDs[e]){case P.TBODY:case P.THEAD:case P.TFOOT:return!0;case P.TABLE:case P.HTML:return!1}return!0}hasInSelectScope(e){for(let t=this.stackTop;t>=0;t--)if(this.treeAdapter.getNamespaceURI(this.items[t])===C.HTML)switch(this.tagIDs[t]){case e:return!0;case P.OPTION:case P.OPTGROUP:break;default:return!1}return!0}generateImpliedEndTags(){for(;J.has(this.currentTagId);)this.pop()}generateImpliedEndTagsThoroughly(){for(;ee.has(this.currentTagId);)this.pop()}generateImpliedEndTagsWithExclusion(e){for(;this.currentTagId!==e&&ee.has(this.currentTagId);)this.pop()}}var de;!function(e){e[e.Marker=0]="Marker",e[e.Element=1]="Element"}(de||(de={}));const pe={type:de.Marker};class he{constructor(e){this.treeAdapter=e,this.entries=[],this.bookmark=null}_getNoahArkConditionCandidates(e,t){const n=[],r=t.length,a=this.treeAdapter.getTagName(e),o=this.treeAdapter.getNamespaceURI(e);for(let e=0;e<this.entries.length;e++){const t=this.entries[e];if(t.type===de.Marker)break;const{element:i}=t;if(this.treeAdapter.getTagName(i)===a&&this.treeAdapter.getNamespaceURI(i)===o){const t=this.treeAdapter.getAttrList(i);t.length===r&&n.push({idx:e,attrs:t})}}return n}_ensureNoahArkCondition(e){if(this.entries.length<3)return;const t=this.treeAdapter.getAttrList(e),n=this._getNoahArkConditionCandidates(e,t);if(n.length<3)return;const r=new Map(t.map((e=>[e.name,e.value])));let a=0;for(let e=0;e<n.length;e++){const t=n[e];t.attrs.every((e=>r.get(e.name)===e.value))&&(a+=1,a>=3&&this.entries.splice(t.idx,1))}}insertMarker(){this.entries.unshift(pe)}pushElement(e,t){this._ensureNoahArkCondition(e),this.entries.unshift({type:de.Element,element:e,token:t})}insertElementAfterBookmark(e,t){const n=this.entries.indexOf(this.bookmark);this.entries.splice(n,0,{type:de.Element,element:e,token:t})}removeEntry(e){const t=this.entries.indexOf(e);t>=0&&this.entries.splice(t,1)}clearToLastMarker(){const e=this.entries.indexOf(pe);e>=0?this.entries.splice(0,e+1):this.entries.length=0}getElementEntryInScopeWithTagName(e){const t=this.entries.find((t=>t.type===de.Marker||this.treeAdapter.getTagName(t.element)===e));return t&&t.type===de.Element?t:null}getElementEntry(e){return this.entries.find((t=>t.type===de.Element&&t.element===e))}}const fe={createDocument:()=>({nodeName:"#document",mode:A.NO_QUIRKS,childNodes:[]}),createDocumentFragment:()=>({nodeName:"#document-fragment",childNodes:[]}),createElement:(e,t,n)=>({nodeName:e,tagName:e,attrs:n,namespaceURI:t,childNodes:[],parentNode:null}),createCommentNode:e=>({nodeName:"#comment",data:e,parentNode:null}),createTextNode:e=>({nodeName:"#text",value:e,parentNode:null}),appendChild(e,t){e.childNodes.push(t),t.parentNode=e},insertBefore(e,t,n){const r=e.childNodes.indexOf(n);e.childNodes.splice(r,0,t),t.parentNode=e},setTemplateContent(e,t){e.content=t},getTemplateContent:e=>e.content,setDocumentType(e,t,n,r){const a=e.childNodes.find((e=>"#documentType"===e.nodeName));if(a)a.name=t,a.publicId=n,a.systemId=r;else{const a={nodeName:"#documentType",name:t,publicId:n,systemId:r,parentNode:null};fe.appendChild(e,a)}},setDocumentMode(e,t){e.mode=t},getDocumentMode:e=>e.mode,detachNode(e){if(e.parentNode){const t=e.parentNode.childNodes.indexOf(e);e.parentNode.childNodes.splice(t,1),e.parentNode=null}},insertText(e,t){if(e.childNodes.length>0){const n=e.childNodes[e.childNodes.length-1];if(fe.isTextNode(n))return void(n.value+=t)}fe.appendChild(e,fe.createTextNode(t))},insertTextBefore(e,t,n){const r=e.childNodes[e.childNodes.indexOf(n)-1];r&&fe.isTextNode(r)?r.value+=t:fe.insertBefore(e,fe.createTextNode(t),n)},adoptAttributes(e,t){const n=new Set(e.attrs.map((e=>e.name)));for(let r=0;r<t.length;r++)n.has(t[r].name)||e.attrs.push(t[r])},getFirstChild:e=>e.childNodes[0],getChildNodes:e=>e.childNodes,getParentNode:e=>e.parentNode,getAttrList:e=>e.attrs,getTagName:e=>e.tagName,getNamespaceURI:e=>e.namespaceURI,getTextNodeContent:e=>e.value,getCommentNodeContent:e=>e.data,getDocumentTypeNodeName:e=>e.name,getDocumentTypeNodePublicId:e=>e.publicId,getDocumentTypeNodeSystemId:e=>e.systemId,isTextNode:e=>"#text"===e.nodeName,isCommentNode:e=>"#comment"===e.nodeName,isDocumentTypeNode:e=>"#documentType"===e.nodeName,isElementNode:e=>Object.prototype.hasOwnProperty.call(e,"tagName"),setNodeSourceCodeLocation(e,t){e.sourceCodeLocation=t},getNodeSourceCodeLocation:e=>e.sourceCodeLocation,updateNodeSourceCodeLocation(e,t){e.sourceCodeLocation={...e.sourceCodeLocation,...t}}},me="html",ge=["+//silmaril//dtd html pro v0r11 19970101//","-//as//dtd html 3.0 aswedit + extensions//","-//advasoft ltd//dtd html 3.0 aswedit + extensions//","-//ietf//dtd html 2.0 level 1//","-//ietf//dtd html 2.0 level 2//","-//ietf//dtd html 2.0 strict level 1//","-//ietf//dtd html 2.0 strict level 2//","-//ietf//dtd html 2.0 strict//","-//ietf//dtd html 2.0//","-//ietf//dtd html 2.1e//","-//ietf//dtd html 3.0//","-//ietf//dtd html 3.2 final//","-//ietf//dtd html 3.2//","-//ietf//dtd html 3//","-//ietf//dtd html level 0//","-//ietf//dtd html level 1//","-//ietf//dtd html level 2//","-//ietf//dtd html level 3//","-//ietf//dtd html strict level 0//","-//ietf//dtd html strict level 1//","-//ietf//dtd html strict level 2//","-//ietf//dtd html strict level 3//","-//ietf//dtd html strict//","-//ietf//dtd html//","-//metrius//dtd metrius presentational//","-//microsoft//dtd internet explorer 2.0 html strict//","-//microsoft//dtd internet explorer 2.0 html//","-//microsoft//dtd internet explorer 2.0 tables//","-//microsoft//dtd internet explorer 3.0 html strict//","-//microsoft//dtd internet explorer 3.0 html//","-//microsoft//dtd internet explorer 3.0 tables//","-//netscape comm. corp.//dtd html//","-//netscape comm. corp.//dtd strict html//","-//o'reilly and associates//dtd html 2.0//","-//o'reilly and associates//dtd html extended 1.0//","-//o'reilly and associates//dtd html extended relaxed 1.0//","-//sq//dtd html 2.0 hotmetal + extensions//","-//softquad software//dtd hotmetal pro 6.0::19990601::extensions to html 4.0//","-//softquad//dtd hotmetal pro 4.0::19971010::extensions to html 4.0//","-//spyglass//dtd html 2.0 extended//","-//sun microsystems corp.//dtd hotjava html//","-//sun microsystems corp.//dtd hotjava strict html//","-//w3c//dtd html 3 1995-03-24//","-//w3c//dtd html 3.2 draft//","-//w3c//dtd html 3.2 final//","-//w3c//dtd html 3.2//","-//w3c//dtd html 3.2s draft//","-//w3c//dtd html 4.0 frameset//","-//w3c//dtd html 4.0 transitional//","-//w3c//dtd html experimental 19960712//","-//w3c//dtd html experimental 970421//","-//w3c//dtd w3 html//","-//w3o//dtd w3 html 3.0//","-//webtechs//dtd mozilla html 2.0//","-//webtechs//dtd mozilla html//"],ve=[...ge,"-//w3c//dtd html 4.01 frameset//","-//w3c//dtd html 4.01 transitional//"],be=new Set(["-//w3o//dtd w3 html strict 3.0//en//","-/w3c/dtd html 4.0 transitional/en","html"]),ye=["-//w3c//dtd xhtml 1.0 frameset//","-//w3c//dtd xhtml 1.0 transitional//"],we=[...ye,"-//w3c//dtd html 4.01 frameset//","-//w3c//dtd html 4.01 transitional//"];function Oe(e,t){return t.some((t=>e.startsWith(t)))}const ke=new Map(["attributeName","attributeType","baseFrequency","baseProfile","calcMode","clipPathUnits","diffuseConstant","edgeMode","filterUnits","glyphRef","gradientTransform","gradientUnits","kernelMatrix","kernelUnitLength","keyPoints","keySplines","keyTimes","lengthAdjust","limitingConeAngle","markerHeight","markerUnits","markerWidth","maskContentUnits","maskUnits","numOctaves","pathLength","patternContentUnits","patternTransform","patternUnits","pointsAtX","pointsAtY","pointsAtZ","preserveAlpha","preserveAspectRatio","primitiveUnits","refX","refY","repeatCount","repeatDur","requiredExtensions","requiredFeatures","specularConstant","specularExponent","spreadMethod","startOffset","stdDeviation","stitchTiles","surfaceScale","systemLanguage","tableValues","targetX","targetY","textLength","viewBox","viewTarget","xChannelSelector","yChannelSelector","zoomAndPan"].map((e=>[e.toLowerCase(),e]))),xe=new Map([["xlink:actuate",{prefix:"xlink",name:"actuate",namespace:C.XLINK}],["xlink:arcrole",{prefix:"xlink",name:"arcrole",namespace:C.XLINK}],["xlink:href",{prefix:"xlink",name:"href",namespace:C.XLINK}],["xlink:role",{prefix:"xlink",name:"role",namespace:C.XLINK}],["xlink:show",{prefix:"xlink",name:"show",namespace:C.XLINK}],["xlink:title",{prefix:"xlink",name:"title",namespace:C.XLINK}],["xlink:type",{prefix:"xlink",name:"type",namespace:C.XLINK}],["xml:lang",{prefix:"xml",name:"lang",namespace:C.XML}],["xml:space",{prefix:"xml",name:"space",namespace:C.XML}],["xmlns",{prefix:"",name:"xmlns",namespace:C.XMLNS}],["xmlns:xlink",{prefix:"xmlns",name:"xlink",namespace:C.XMLNS}]]),Ee=new Map(["altGlyph","altGlyphDef","altGlyphItem","animateColor","animateMotion","animateTransform","clipPath","feBlend","feColorMatrix","feComponentTransfer","feComposite","feConvolveMatrix","feDiffuseLighting","feDisplacementMap","feDistantLight","feFlood","feFuncA","feFuncB","feFuncG","feFuncR","feGaussianBlur","feImage","feMerge","feMergeNode","feMorphology","feOffset","fePointLight","feSpecularLighting","feSpotLight","feTile","feTurbulence","foreignObject","glyphRef","linearGradient","radialGradient","textPath"].map((e=>[e.toLowerCase(),e]))),Se=new Set([P.B,P.BIG,P.BLOCKQUOTE,P.BODY,P.BR,P.CENTER,P.CODE,P.DD,P.DIV,P.DL,P.DT,P.EM,P.EMBED,P.H1,P.H2,P.H3,P.H4,P.H5,P.H6,P.HEAD,P.HR,P.I,P.IMG,P.LI,P.LISTING,P.MENU,P.META,P.NOBR,P.OL,P.P,P.PRE,P.RUBY,P.S,P.SMALL,P.SPAN,P.STRONG,P.STRIKE,P.SUB,P.SUP,P.TABLE,P.TT,P.U,P.UL,P.VAR]);function _e(e){for(let t=0;t<e.attrs.length;t++)if("definitionurl"===e.attrs[t].name){e.attrs[t].name="definitionURL";break}}function Ce(e){for(let t=0;t<e.attrs.length;t++){const n=ke.get(e.attrs[t].name);null!=n&&(e.attrs[t].name=n)}}function Te(e){for(let t=0;t<e.attrs.length;t++){const n=xe.get(e.attrs[t].name);n&&(e.attrs[t].prefix=n.prefix,e.attrs[t].name=n.name,e.attrs[t].namespace=n.namespace)}}var Ae;!function(e){e[e.INITIAL=0]="INITIAL",e[e.BEFORE_HTML=1]="BEFORE_HTML",e[e.BEFORE_HEAD=2]="BEFORE_HEAD",e[e.IN_HEAD=3]="IN_HEAD",e[e.IN_HEAD_NO_SCRIPT=4]="IN_HEAD_NO_SCRIPT",e[e.AFTER_HEAD=5]="AFTER_HEAD",e[e.IN_BODY=6]="IN_BODY",e[e.TEXT=7]="TEXT",e[e.IN_TABLE=8]="IN_TABLE",e[e.IN_TABLE_TEXT=9]="IN_TABLE_TEXT",e[e.IN_CAPTION=10]="IN_CAPTION",e[e.IN_COLUMN_GROUP=11]="IN_COLUMN_GROUP",e[e.IN_TABLE_BODY=12]="IN_TABLE_BODY",e[e.IN_ROW=13]="IN_ROW",e[e.IN_CELL=14]="IN_CELL",e[e.IN_SELECT=15]="IN_SELECT",e[e.IN_SELECT_IN_TABLE=16]="IN_SELECT_IN_TABLE",e[e.IN_TEMPLATE=17]="IN_TEMPLATE",e[e.AFTER_BODY=18]="AFTER_BODY",e[e.IN_FRAMESET=19]="IN_FRAMESET",e[e.AFTER_FRAMESET=20]="AFTER_FRAMESET",e[e.AFTER_AFTER_BODY=21]="AFTER_AFTER_BODY",e[e.AFTER_AFTER_FRAMESET=22]="AFTER_AFTER_FRAMESET"}(Ae||(Ae={}));const Re={startLine:-1,startCol:-1,startOffset:-1,endLine:-1,endCol:-1,endOffset:-1},Pe=new Set([P.TABLE,P.TBODY,P.TFOOT,P.THEAD,P.TR]),De={scriptingEnabled:!0,sourceCodeLocationInfo:!1,treeAdapter:fe,onParseError:null};class Me{constructor(e,t,n=null,r=null){this.fragmentContext=n,this.scriptHandler=r,this.currentToken=null,this.stopped=!1,this.insertionMode=Ae.INITIAL,this.originalInsertionMode=Ae.INITIAL,this.headElement=null,this.formElement=null,this.currentNotInHTML=!1,this.tmplInsertionModeStack=[],this.pendingCharacterTokens=[],this.hasNonWhitespacePendingCharacterToken=!1,this.framesetOk=!0,this.skipNextNewLine=!1,this.fosterParentingEnabled=!1,this.options={...De,...e},this.treeAdapter=this.options.treeAdapter,this.onParseError=this.options.onParseError,this.onParseError&&(this.options.sourceCodeLocationInfo=!0),this.document=null!=t?t:this.treeAdapter.createDocument(),this.tokenizer=new Y(this.options,this),this.activeFormattingElements=new he(this.treeAdapter),this.fragmentContextID=n?B(this.treeAdapter.getTagName(n)):P.UNKNOWN,this._setContextModes(null!=n?n:this.document,this.fragmentContextID),this.openElements=new ue(this.document,this.treeAdapter,this)}static parse(e,t){const n=new this(t);return n.tokenizer.write(e,!0),n.document}static getFragmentParser(e,t){const n={...De,...t};null!=e||(e=n.treeAdapter.createElement(R.TEMPLATE,C.HTML,[]));const r=n.treeAdapter.createElement("documentmock",C.HTML,[]),a=new this(n,r,e);return a.fragmentContextID===P.TEMPLATE&&a.tmplInsertionModeStack.unshift(Ae.IN_TEMPLATE),a._initTokenizerForFragmentParsing(),a._insertFakeRootElement(),a._resetInsertionMode(),a._findFormInFragmentContext(),a}getFragment(){const e=this.treeAdapter.getFirstChild(this.document),t=this.treeAdapter.createDocumentFragment();return this._adoptNodes(e,t),t}_err(e,t,n){var r;if(!this.onParseError)return;const a=null!==(r=e.location)&&void 0!==r?r:Re,o={code:t,startLine:a.startLine,startCol:a.startCol,startOffset:a.startOffset,endLine:n?a.startLine:a.endLine,endCol:n?a.startCol:a.endCol,endOffset:n?a.startOffset:a.endOffset};this.onParseError(o)}onItemPush(e,t,n){var r,a;null===(a=(r=this.treeAdapter).onItemPush)||void 0===a||a.call(r,e),n&&this.openElements.stackTop>0&&this._setContextModes(e,t)}onItemPop(e,t){var n,r;if(this.options.sourceCodeLocationInfo&&this._setEndLocation(e,this.currentToken),null===(r=(n=this.treeAdapter).onItemPop)||void 0===r||r.call(n,e,this.openElements.current),t){let e,t;0===this.openElements.stackTop&&this.fragmentContext?(e=this.fragmentContext,t=this.fragmentContextID):({current:e,currentTagId:t}=this.openElements),this._setContextModes(e,t)}}_setContextModes(e,t){const n=e===this.document||this.treeAdapter.getNamespaceURI(e)===C.HTML;this.currentNotInHTML=!n,this.tokenizer.inForeignNode=!n&&!this._isIntegrationPoint(t,e)}_switchToTextParsing(e,t){this._insertElement(e,C.HTML),this.tokenizer.state=t,this.originalInsertionMode=this.insertionMode,this.insertionMode=Ae.TEXT}switchToPlaintextParsing(){this.insertionMode=Ae.TEXT,this.originalInsertionMode=Ae.IN_BODY,this.tokenizer.state=X.PLAINTEXT}_getAdjustedCurrentElement(){return 0===this.openElements.stackTop&&this.fragmentContext?this.fragmentContext:this.openElements.current}_findFormInFragmentContext(){let e=this.fragmentContext;for(;e;){if(this.treeAdapter.getTagName(e)===R.FORM){this.formElement=e;break}e=this.treeAdapter.getParentNode(e)}}_initTokenizerForFragmentParsing(){if(this.fragmentContext&&this.treeAdapter.getNamespaceURI(this.fragmentContext)===C.HTML)switch(this.fragmentContextID){case P.TITLE:case P.TEXTAREA:this.tokenizer.state=X.RCDATA;break;case P.STYLE:case P.XMP:case P.IFRAME:case P.NOEMBED:case P.NOFRAMES:case P.NOSCRIPT:this.tokenizer.state=X.RAWTEXT;break;case P.SCRIPT:this.tokenizer.state=X.SCRIPT_DATA;break;case P.PLAINTEXT:this.tokenizer.state=X.PLAINTEXT}}_setDocumentType(e){const t=e.name||"",n=e.publicId||"",r=e.systemId||"";if(this.treeAdapter.setDocumentType(this.document,t,n,r),e.location){const t=this.treeAdapter.getChildNodes(this.document).find((e=>this.treeAdapter.isDocumentTypeNode(e)));t&&this.treeAdapter.setNodeSourceCodeLocation(t,e.location)}}_attachElementToTree(e,t){if(this.options.sourceCodeLocationInfo){const n=t&&{...t,startTag:t};this.treeAdapter.setNodeSourceCodeLocation(e,n)}if(this._shouldFosterParentOnInsertion())this._fosterParentElement(e);else{const t=this.openElements.currentTmplContentOrNode;this.treeAdapter.appendChild(t,e)}}_appendElement(e,t){const n=this.treeAdapter.createElement(e.tagName,t,e.attrs);this._attachElementToTree(n,e.location)}_insertElement(e,t){const n=this.treeAdapter.createElement(e.tagName,t,e.attrs);this._attachElementToTree(n,e.location),this.openElements.push(n,e.tagID)}_insertFakeElement(e,t){const n=this.treeAdapter.createElement(e,C.HTML,[]);this._attachElementToTree(n,null),this.openElements.push(n,t)}_insertTemplate(e){const t=this.treeAdapter.createElement(e.tagName,C.HTML,e.attrs),n=this.treeAdapter.createDocumentFragment();this.treeAdapter.setTemplateContent(t,n),this._attachElementToTree(t,e.location),this.openElements.push(t,e.tagID),this.options.sourceCodeLocationInfo&&this.treeAdapter.setNodeSourceCodeLocation(n,null)}_insertFakeRootElement(){const e=this.treeAdapter.createElement(R.HTML,C.HTML,[]);this.options.sourceCodeLocationInfo&&this.treeAdapter.setNodeSourceCodeLocation(e,null),this.treeAdapter.appendChild(this.openElements.current,e),this.openElements.push(e,P.HTML)}_appendCommentNode(e,t){const n=this.treeAdapter.createCommentNode(e.data);this.treeAdapter.appendChild(t,n),this.options.sourceCodeLocationInfo&&this.treeAdapter.setNodeSourceCodeLocation(n,e.location)}_insertCharacters(e){let t,n;if(this._shouldFosterParentOnInsertion()?(({parent:t,beforeElement:n}=this._findFosterParentingLocation()),n?this.treeAdapter.insertTextBefore(t,e.chars,n):this.treeAdapter.insertText(t,e.chars)):(t=this.openElements.currentTmplContentOrNode,this.treeAdapter.insertText(t,e.chars)),!e.location)return;const r=this.treeAdapter.getChildNodes(t),a=n?r.lastIndexOf(n):r.length,o=r[a-1];if(this.treeAdapter.getNodeSourceCodeLocation(o)){const{endLine:t,endCol:n,endOffset:r}=e.location;this.treeAdapter.updateNodeSourceCodeLocation(o,{endLine:t,endCol:n,endOffset:r})}else this.options.sourceCodeLocationInfo&&this.treeAdapter.setNodeSourceCodeLocation(o,e.location)}_adoptNodes(e,t){for(let n=this.treeAdapter.getFirstChild(e);n;n=this.treeAdapter.getFirstChild(e))this.treeAdapter.detachNode(n),this.treeAdapter.appendChild(t,n)}_setEndLocation(e,t){if(this.treeAdapter.getNodeSourceCodeLocation(e)&&t.location){const n=t.location,r=this.treeAdapter.getTagName(e),a=t.type===m.END_TAG&&r===t.tagName?{endTag:{...n},endLine:n.endLine,endCol:n.endCol,endOffset:n.endOffset}:{endLine:n.startLine,endCol:n.startCol,endOffset:n.startOffset};this.treeAdapter.updateNodeSourceCodeLocation(e,a)}}shouldProcessStartTagTokenInForeignContent(e){if(!this.currentNotInHTML)return!1;let t,n;return 0===this.openElements.stackTop&&this.fragmentContext?(t=this.fragmentContext,n=this.fragmentContextID):({current:t,currentTagId:n}=this.openElements),(e.tagID!==P.SVG||this.treeAdapter.getTagName(t)!==R.ANNOTATION_XML||this.treeAdapter.getNamespaceURI(t)!==C.MATHML)&&(this.tokenizer.inForeignNode||(e.tagID===P.MGLYPH||e.tagID===P.MALIGNMARK)&&!this._isIntegrationPoint(n,t,C.HTML))}_processToken(e){switch(e.type){case m.CHARACTER:this.onCharacter(e);break;case m.NULL_CHARACTER:this.onNullCharacter(e);break;case m.COMMENT:this.onComment(e);break;case m.DOCTYPE:this.onDoctype(e);break;case m.START_TAG:this._processStartTag(e);break;case m.END_TAG:this.onEndTag(e);break;case m.EOF:this.onEof(e);break;case m.WHITESPACE_CHARACTER:this.onWhitespaceCharacter(e)}}_isIntegrationPoint(e,t,n){return function(e,t,n,r){return(!r||r===C.HTML)&&function(e,t,n){if(t===C.MATHML&&e===P.ANNOTATION_XML)for(let e=0;e<n.length;e++)if(n[e].name===T.ENCODING){const t=n[e].value.toLowerCase();return"text/html"===t||"application/xhtml+xml"===t}return t===C.SVG&&(e===P.FOREIGN_OBJECT||e===P.DESC||e===P.TITLE)}(e,t,n)||(!r||r===C.MATHML)&&function(e,t){return t===C.MATHML&&(e===P.MI||e===P.MO||e===P.MN||e===P.MS||e===P.MTEXT)}(e,t)}(e,this.treeAdapter.getNamespaceURI(t),this.treeAdapter.getAttrList(t),n)}_reconstructActiveFormattingElements(){const e=this.activeFormattingElements.entries.length;if(e){const t=this.activeFormattingElements.entries.findIndex((e=>e.type===de.Marker||this.openElements.contains(e.element)));for(let n=t<0?e-1:t-1;n>=0;n--){const e=this.activeFormattingElements.entries[n];this._insertElement(e.token,this.treeAdapter.getNamespaceURI(e.element)),e.element=this.openElements.current}}}_closeTableCell(){this.openElements.generateImpliedEndTags(),this.openElements.popUntilTableCellPopped(),this.activeFormattingElements.clearToLastMarker(),this.insertionMode=Ae.IN_ROW}_closePElement(){this.openElements.generateImpliedEndTagsWithExclusion(P.P),this.openElements.popUntilTagNamePopped(P.P)}_resetInsertionMode(){for(let e=this.openElements.stackTop;e>=0;e--)switch(0===e&&this.fragmentContext?this.fragmentContextID:this.openElements.tagIDs[e]){case P.TR:return void(this.insertionMode=Ae.IN_ROW);case P.TBODY:case P.THEAD:case P.TFOOT:return void(this.insertionMode=Ae.IN_TABLE_BODY);case P.CAPTION:return void(this.insertionMode=Ae.IN_CAPTION);case P.COLGROUP:return void(this.insertionMode=Ae.IN_COLUMN_GROUP);case P.TABLE:return void(this.insertionMode=Ae.IN_TABLE);case P.BODY:return void(this.insertionMode=Ae.IN_BODY);case P.FRAMESET:return void(this.insertionMode=Ae.IN_FRAMESET);case P.SELECT:return void this._resetInsertionModeForSelect(e);case P.TEMPLATE:return void(this.insertionMode=this.tmplInsertionModeStack[0]);case P.HTML:return void(this.insertionMode=this.headElement?Ae.AFTER_HEAD:Ae.BEFORE_HEAD);case P.TD:case P.TH:if(e>0)return void(this.insertionMode=Ae.IN_CELL);break;case P.HEAD:if(e>0)return void(this.insertionMode=Ae.IN_HEAD)}this.insertionMode=Ae.IN_BODY}_resetInsertionModeForSelect(e){if(e>0)for(let t=e-1;t>0;t--){const e=this.openElements.tagIDs[t];if(e===P.TEMPLATE)break;if(e===P.TABLE)return void(this.insertionMode=Ae.IN_SELECT_IN_TABLE)}this.insertionMode=Ae.IN_SELECT}_isElementCausesFosterParenting(e){return Pe.has(e)}_shouldFosterParentOnInsertion(){return this.fosterParentingEnabled&&this._isElementCausesFosterParenting(this.openElements.currentTagId)}_findFosterParentingLocation(){for(let e=this.openElements.stackTop;e>=0;e--){const t=this.openElements.items[e];switch(this.openElements.tagIDs[e]){case P.TEMPLATE:if(this.treeAdapter.getNamespaceURI(t)===C.HTML)return{parent:this.treeAdapter.getTemplateContent(t),beforeElement:null};break;case P.TABLE:{const n=this.treeAdapter.getParentNode(t);return n?{parent:n,beforeElement:t}:{parent:this.openElements.items[e-1],beforeElement:null}}}}return{parent:this.openElements.items[0],beforeElement:null}}_fosterParentElement(e){const t=this._findFosterParentingLocation();t.beforeElement?this.treeAdapter.insertBefore(t.parent,e,t.beforeElement):this.treeAdapter.appendChild(t.parent,e)}_isSpecialElement(e,t){const n=this.treeAdapter.getNamespaceURI(e);return Q[n].has(t)}onCharacter(e){if(this.skipNextNewLine=!1,this.tokenizer.inForeignNode)!function(e,t){e._insertCharacters(t),e.framesetOk=!1}(this,e);else switch(this.insertionMode){case Ae.INITIAL:je(this,e);break;case Ae.BEFORE_HTML:We(this,e);break;case Ae.BEFORE_HEAD:Xe(this,e);break;case Ae.IN_HEAD:He(this,e);break;case Ae.IN_HEAD_NO_SCRIPT:Ze(this,e);break;case Ae.AFTER_HEAD:Ge(this,e);break;case Ae.IN_BODY:case Ae.IN_CAPTION:case Ae.IN_CELL:case Ae.IN_TEMPLATE:Je(this,e);break;case Ae.TEXT:case Ae.IN_SELECT:case Ae.IN_SELECT_IN_TABLE:this._insertCharacters(e);break;case Ae.IN_TABLE:case Ae.IN_TABLE_BODY:case Ae.IN_ROW:lt(this,e);break;case Ae.IN_TABLE_TEXT:ht(this,e);break;case Ae.IN_COLUMN_GROUP:vt(this,e);break;case Ae.AFTER_BODY:_t(this,e);break;case Ae.AFTER_AFTER_BODY:Ct(this,e)}}onNullCharacter(e){if(this.skipNextNewLine=!1,this.tokenizer.inForeignNode)!function(e,t){t.chars=i,e._insertCharacters(t)}(this,e);else switch(this.insertionMode){case Ae.INITIAL:je(this,e);break;case Ae.BEFORE_HTML:We(this,e);break;case Ae.BEFORE_HEAD:Xe(this,e);break;case Ae.IN_HEAD:He(this,e);break;case Ae.IN_HEAD_NO_SCRIPT:Ze(this,e);break;case Ae.AFTER_HEAD:Ge(this,e);break;case Ae.TEXT:this._insertCharacters(e);break;case Ae.IN_TABLE:case Ae.IN_TABLE_BODY:case Ae.IN_ROW:lt(this,e);break;case Ae.IN_COLUMN_GROUP:vt(this,e);break;case Ae.AFTER_BODY:_t(this,e);break;case Ae.AFTER_AFTER_BODY:Ct(this,e)}}onComment(e){if(this.skipNextNewLine=!1,this.currentNotInHTML)Ue(this,e);else switch(this.insertionMode){case Ae.INITIAL:case Ae.BEFORE_HTML:case Ae.BEFORE_HEAD:case Ae.IN_HEAD:case Ae.IN_HEAD_NO_SCRIPT:case Ae.AFTER_HEAD:case Ae.IN_BODY:case Ae.IN_TABLE:case Ae.IN_CAPTION:case Ae.IN_COLUMN_GROUP:case Ae.IN_TABLE_BODY:case Ae.IN_ROW:case Ae.IN_CELL:case Ae.IN_SELECT:case Ae.IN_SELECT_IN_TABLE:case Ae.IN_TEMPLATE:case Ae.IN_FRAMESET:case Ae.AFTER_FRAMESET:Ue(this,e);break;case Ae.IN_TABLE_TEXT:ft(this,e);break;case Ae.AFTER_BODY:!function(e,t){e._appendCommentNode(t,e.openElements.items[0])}(this,e);break;case Ae.AFTER_AFTER_BODY:case Ae.AFTER_AFTER_FRAMESET:!function(e,t){e._appendCommentNode(t,e.document)}(this,e)}}onDoctype(e){switch(this.skipNextNewLine=!1,this.insertionMode){case Ae.INITIAL:!function(e,t){e._setDocumentType(t);const n=t.forceQuirks?A.QUIRKS:function(e){if(e.name!==me)return A.QUIRKS;const{systemId:t}=e;if(t&&"http://www.ibm.com/data/dtd/v11/ibmxhtml1-transitional.dtd"===t.toLowerCase())return A.QUIRKS;let{publicId:n}=e;if(null!==n){if(n=n.toLowerCase(),be.has(n))return A.QUIRKS;let e=null===t?ve:ge;if(Oe(n,e))return A.QUIRKS;if(e=null===t?ye:we,Oe(n,e))return A.LIMITED_QUIRKS}return A.NO_QUIRKS}(t);(function(e){return e.name===me&&null===e.publicId&&(null===e.systemId||"about:legacy-compat"===e.systemId)})(t)||e._err(t,f.nonConformingDoctype),e.treeAdapter.setDocumentMode(e.document,n),e.insertionMode=Ae.BEFORE_HTML}(this,e);break;case Ae.BEFORE_HEAD:case Ae.IN_HEAD:case Ae.IN_HEAD_NO_SCRIPT:case Ae.AFTER_HEAD:this._err(e,f.misplacedDoctype);break;case Ae.IN_TABLE_TEXT:ft(this,e)}}onStartTag(e){this.skipNextNewLine=!1,this.currentToken=e,this._processStartTag(e),e.selfClosing&&!e.ackSelfClosing&&this._err(e,f.nonVoidHtmlElementStartTagWithTrailingSolidus)}_processStartTag(e){this.shouldProcessStartTagTokenInForeignContent(e)?function(e,t){if(function(e){const t=e.tagID;return t===P.FONT&&e.attrs.some((({name:e})=>e===T.COLOR||e===T.SIZE||e===T.FACE))||Se.has(t)}(t))Tt(e),e._startTagOutsideForeignContent(t);else{const n=e._getAdjustedCurrentElement(),r=e.treeAdapter.getNamespaceURI(n);r===C.MATHML?_e(t):r===C.SVG&&(function(e){const t=Ee.get(e.tagName);null!=t&&(e.tagName=t,e.tagID=B(e.tagName))}(t),Ce(t)),Te(t),t.selfClosing?e._appendElement(t,r):e._insertElement(t,r),t.ackSelfClosing=!0}}(this,e):this._startTagOutsideForeignContent(e)}_startTagOutsideForeignContent(e){switch(this.insertionMode){case Ae.INITIAL:je(this,e);break;case Ae.BEFORE_HTML:!function(e,t){t.tagID===P.HTML?(e._insertElement(t,C.HTML),e.insertionMode=Ae.BEFORE_HEAD):We(e,t)}(this,e);break;case Ae.BEFORE_HEAD:!function(e,t){switch(t.tagID){case P.HTML:at(e,t);break;case P.HEAD:e._insertElement(t,C.HTML),e.headElement=e.openElements.current,e.insertionMode=Ae.IN_HEAD;break;default:Xe(e,t)}}(this,e);break;case Ae.IN_HEAD:qe(this,e);break;case Ae.IN_HEAD_NO_SCRIPT:!function(e,t){switch(t.tagID){case P.HTML:at(e,t);break;case P.BASEFONT:case P.BGSOUND:case P.HEAD:case P.LINK:case P.META:case P.NOFRAMES:case P.STYLE:qe(e,t);break;case P.NOSCRIPT:e._err(t,f.nestedNoscriptInHead);break;default:Ze(e,t)}}(this,e);break;case Ae.AFTER_HEAD:!function(e,t){switch(t.tagID){case P.HTML:at(e,t);break;case P.BODY:e._insertElement(t,C.HTML),e.framesetOk=!1,e.insertionMode=Ae.IN_BODY;break;case P.FRAMESET:e._insertElement(t,C.HTML),e.insertionMode=Ae.IN_FRAMESET;break;case P.BASE:case P.BASEFONT:case P.BGSOUND:case P.LINK:case P.META:case P.NOFRAMES:case P.SCRIPT:case P.STYLE:case P.TEMPLATE:case P.TITLE:e._err(t,f.abandonedHeadElementChild),e.openElements.push(e.headElement,P.HEAD),qe(e,t),e.openElements.remove(e.headElement);break;case P.HEAD:e._err(t,f.misplacedStartTagForHeadElement);break;default:Ge(e,t)}}(this,e);break;case Ae.IN_BODY:at(this,e);break;case Ae.IN_TABLE:ct(this,e);break;case Ae.IN_TABLE_TEXT:ft(this,e);break;case Ae.IN_CAPTION:!function(e,t){const n=t.tagID;mt.has(n)?e.openElements.hasInTableScope(P.CAPTION)&&(e.openElements.generateImpliedEndTags(),e.openElements.popUntilTagNamePopped(P.CAPTION),e.activeFormattingElements.clearToLastMarker(),e.insertionMode=Ae.IN_TABLE,ct(e,t)):at(e,t)}(this,e);break;case Ae.IN_COLUMN_GROUP:gt(this,e);break;case Ae.IN_TABLE_BODY:bt(this,e);break;case Ae.IN_ROW:wt(this,e);break;case Ae.IN_CELL:!function(e,t){const n=t.tagID;mt.has(n)?(e.openElements.hasInTableScope(P.TD)||e.openElements.hasInTableScope(P.TH))&&(e._closeTableCell(),wt(e,t)):at(e,t)}(this,e);break;case Ae.IN_SELECT:kt(this,e);break;case Ae.IN_SELECT_IN_TABLE:!function(e,t){const n=t.tagID;n===P.CAPTION||n===P.TABLE||n===P.TBODY||n===P.TFOOT||n===P.THEAD||n===P.TR||n===P.TD||n===P.TH?(e.openElements.popUntilTagNamePopped(P.SELECT),e._resetInsertionMode(),e._processStartTag(t)):kt(e,t)}(this,e);break;case Ae.IN_TEMPLATE:!function(e,t){switch(t.tagID){case P.BASE:case P.BASEFONT:case P.BGSOUND:case P.LINK:case P.META:case P.NOFRAMES:case P.SCRIPT:case P.STYLE:case P.TEMPLATE:case P.TITLE:qe(e,t);break;case P.CAPTION:case P.COLGROUP:case P.TBODY:case P.TFOOT:case P.THEAD:e.tmplInsertionModeStack[0]=Ae.IN_TABLE,e.insertionMode=Ae.IN_TABLE,ct(e,t);break;case P.COL:e.tmplInsertionModeStack[0]=Ae.IN_COLUMN_GROUP,e.insertionMode=Ae.IN_COLUMN_GROUP,gt(e,t);break;case P.TR:e.tmplInsertionModeStack[0]=Ae.IN_TABLE_BODY,e.insertionMode=Ae.IN_TABLE_BODY,bt(e,t);break;case P.TD:case P.TH:e.tmplInsertionModeStack[0]=Ae.IN_ROW,e.insertionMode=Ae.IN_ROW,wt(e,t);break;default:e.tmplInsertionModeStack[0]=Ae.IN_BODY,e.insertionMode=Ae.IN_BODY,at(e,t)}}(this,e);break;case Ae.AFTER_BODY:!function(e,t){t.tagID===P.HTML?at(e,t):_t(e,t)}(this,e);break;case Ae.IN_FRAMESET:!function(e,t){switch(t.tagID){case P.HTML:at(e,t);break;case P.FRAMESET:e._insertElement(t,C.HTML);break;case P.FRAME:e._appendElement(t,C.HTML),t.ackSelfClosing=!0;break;case P.NOFRAMES:qe(e,t)}}(this,e);break;case Ae.AFTER_FRAMESET:!function(e,t){switch(t.tagID){case P.HTML:at(e,t);break;case P.NOFRAMES:qe(e,t)}}(this,e);break;case Ae.AFTER_AFTER_BODY:!function(e,t){t.tagID===P.HTML?at(e,t):Ct(e,t)}(this,e);break;case Ae.AFTER_AFTER_FRAMESET:!function(e,t){switch(t.tagID){case P.HTML:at(e,t);break;case P.NOFRAMES:qe(e,t)}}(this,e)}}onEndTag(e){this.skipNextNewLine=!1,this.currentToken=e,this.currentNotInHTML?function(e,t){if(t.tagID===P.P||t.tagID===P.BR)return Tt(e),void e._endTagOutsideForeignContent(t);for(let n=e.openElements.stackTop;n>0;n--){const r=e.openElements.items[n];if(e.treeAdapter.getNamespaceURI(r)===C.HTML){e._endTagOutsideForeignContent(t);break}const a=e.treeAdapter.getTagName(r);if(a.toLowerCase()===t.tagName){t.tagName=a,e.openElements.shortenToLength(n);break}}}(this,e):this._endTagOutsideForeignContent(e)}_endTagOutsideForeignContent(e){switch(this.insertionMode){case Ae.INITIAL:je(this,e);break;case Ae.BEFORE_HTML:!function(e,t){const n=t.tagID;n!==P.HTML&&n!==P.HEAD&&n!==P.BODY&&n!==P.BR||We(e,t)}(this,e);break;case Ae.BEFORE_HEAD:!function(e,t){const n=t.tagID;n===P.HEAD||n===P.BODY||n===P.HTML||n===P.BR?Xe(e,t):e._err(t,f.endTagWithoutMatchingOpenElement)}(this,e);break;case Ae.IN_HEAD:!function(e,t){switch(t.tagID){case P.HEAD:e.openElements.pop(),e.insertionMode=Ae.AFTER_HEAD;break;case P.BODY:case P.BR:case P.HTML:He(e,t);break;case P.TEMPLATE:Ve(e,t);break;default:e._err(t,f.endTagWithoutMatchingOpenElement)}}(this,e);break;case Ae.IN_HEAD_NO_SCRIPT:!function(e,t){switch(t.tagID){case P.NOSCRIPT:e.openElements.pop(),e.insertionMode=Ae.IN_HEAD;break;case P.BR:Ze(e,t);break;default:e._err(t,f.endTagWithoutMatchingOpenElement)}}(this,e);break;case Ae.AFTER_HEAD:!function(e,t){switch(t.tagID){case P.BODY:case P.HTML:case P.BR:Ge(e,t);break;case P.TEMPLATE:Ve(e,t);break;default:e._err(t,f.endTagWithoutMatchingOpenElement)}}(this,e);break;case Ae.IN_BODY:it(this,e);break;case Ae.TEXT:!function(e,t){var n;t.tagID===P.SCRIPT&&(null===(n=e.scriptHandler)||void 0===n||n.call(e,e.openElements.current)),e.openElements.pop(),e.insertionMode=e.originalInsertionMode}(this,e);break;case Ae.IN_TABLE:ut(this,e);break;case Ae.IN_TABLE_TEXT:ft(this,e);break;case Ae.IN_CAPTION:!function(e,t){const n=t.tagID;switch(n){case P.CAPTION:case P.TABLE:e.openElements.hasInTableScope(P.CAPTION)&&(e.openElements.generateImpliedEndTags(),e.openElements.popUntilTagNamePopped(P.CAPTION),e.activeFormattingElements.clearToLastMarker(),e.insertionMode=Ae.IN_TABLE,n===P.TABLE&&ut(e,t));break;case P.BODY:case P.COL:case P.COLGROUP:case P.HTML:case P.TBODY:case P.TD:case P.TFOOT:case P.TH:case P.THEAD:case P.TR:break;default:it(e,t)}}(this,e);break;case Ae.IN_COLUMN_GROUP:!function(e,t){switch(t.tagID){case P.COLGROUP:e.openElements.currentTagId===P.COLGROUP&&(e.openElements.pop(),e.insertionMode=Ae.IN_TABLE);break;case P.TEMPLATE:Ve(e,t);break;case P.COL:break;default:vt(e,t)}}(this,e);break;case Ae.IN_TABLE_BODY:yt(this,e);break;case Ae.IN_ROW:Ot(this,e);break;case Ae.IN_CELL:!function(e,t){const n=t.tagID;switch(n){case P.TD:case P.TH:e.openElements.hasInTableScope(n)&&(e.openElements.generateImpliedEndTags(),e.openElements.popUntilTagNamePopped(n),e.activeFormattingElements.clearToLastMarker(),e.insertionMode=Ae.IN_ROW);break;case P.TABLE:case P.TBODY:case P.TFOOT:case P.THEAD:case P.TR:e.openElements.hasInTableScope(n)&&(e._closeTableCell(),Ot(e,t));break;case P.BODY:case P.CAPTION:case P.COL:case P.COLGROUP:case P.HTML:break;default:it(e,t)}}(this,e);break;case Ae.IN_SELECT:xt(this,e);break;case Ae.IN_SELECT_IN_TABLE:!function(e,t){const n=t.tagID;n===P.CAPTION||n===P.TABLE||n===P.TBODY||n===P.TFOOT||n===P.THEAD||n===P.TR||n===P.TD||n===P.TH?e.openElements.hasInTableScope(n)&&(e.openElements.popUntilTagNamePopped(P.SELECT),e._resetInsertionMode(),e.onEndTag(t)):xt(e,t)}(this,e);break;case Ae.IN_TEMPLATE:!function(e,t){t.tagID===P.TEMPLATE&&Ve(e,t)}(this,e);break;case Ae.AFTER_BODY:St(this,e);break;case Ae.IN_FRAMESET:!function(e,t){t.tagID!==P.FRAMESET||e.openElements.isRootHtmlElementCurrent()||(e.openElements.pop(),e.fragmentContext||e.openElements.currentTagId===P.FRAMESET||(e.insertionMode=Ae.AFTER_FRAMESET))}(this,e);break;case Ae.AFTER_FRAMESET:!function(e,t){t.tagID===P.HTML&&(e.insertionMode=Ae.AFTER_AFTER_FRAMESET)}(this,e);break;case Ae.AFTER_AFTER_BODY:Ct(this,e)}}onEof(e){switch(this.insertionMode){case Ae.INITIAL:je(this,e);break;case Ae.BEFORE_HTML:We(this,e);break;case Ae.BEFORE_HEAD:Xe(this,e);break;case Ae.IN_HEAD:He(this,e);break;case Ae.IN_HEAD_NO_SCRIPT:Ze(this,e);break;case Ae.AFTER_HEAD:Ge(this,e);break;case Ae.IN_BODY:case Ae.IN_TABLE:case Ae.IN_CAPTION:case Ae.IN_COLUMN_GROUP:case Ae.IN_TABLE_BODY:case Ae.IN_ROW:case Ae.IN_CELL:case Ae.IN_SELECT:case Ae.IN_SELECT_IN_TABLE:st(this,e);break;case Ae.TEXT:!function(e,t){e._err(t,f.eofInElementThatCanContainOnlyText),e.openElements.pop(),e.insertionMode=e.originalInsertionMode,e.onEof(t)}(this,e);break;case Ae.IN_TABLE_TEXT:ft(this,e);break;case Ae.IN_TEMPLATE:Et(this,e);break;case Ae.AFTER_BODY:case Ae.IN_FRAMESET:case Ae.AFTER_FRAMESET:case Ae.AFTER_AFTER_BODY:case Ae.AFTER_AFTER_FRAMESET:ze(this,e)}}onWhitespaceCharacter(e){if(this.skipNextNewLine&&(this.skipNextNewLine=!1,e.chars.charCodeAt(0)===s.LINE_FEED)){if(1===e.chars.length)return;e.chars=e.chars.substr(1)}if(this.tokenizer.inForeignNode)this._insertCharacters(e);else switch(this.insertionMode){case Ae.IN_HEAD:case Ae.IN_HEAD_NO_SCRIPT:case Ae.AFTER_HEAD:case Ae.TEXT:case Ae.IN_COLUMN_GROUP:case Ae.IN_SELECT:case Ae.IN_SELECT_IN_TABLE:case Ae.IN_FRAMESET:case Ae.AFTER_FRAMESET:this._insertCharacters(e);break;case Ae.IN_BODY:case Ae.IN_CAPTION:case Ae.IN_CELL:case Ae.IN_TEMPLATE:case Ae.AFTER_BODY:case Ae.AFTER_AFTER_BODY:case Ae.AFTER_AFTER_FRAMESET:Ye(this,e);break;case Ae.IN_TABLE:case Ae.IN_TABLE_BODY:case Ae.IN_ROW:lt(this,e);break;case Ae.IN_TABLE_TEXT:pt(this,e)}}}function $e(e,t){let n=e.activeFormattingElements.getElementEntryInScopeWithTagName(t.tagName);return n?e.openElements.contains(n.element)?e.openElements.hasInScope(t.tagID)||(n=null):(e.activeFormattingElements.removeEntry(n),n=null):ot(e,t),n}function Ne(e,t){let n=null,r=e.openElements.stackTop;for(;r>=0;r--){const a=e.openElements.items[r];if(a===t.element)break;e._isSpecialElement(a,e.openElements.tagIDs[r])&&(n=a)}return n||(e.openElements.shortenToLength(r<0?0:r),e.activeFormattingElements.removeEntry(t)),n}function Ie(e,t,n){let r=t,a=e.openElements.getCommonAncestor(t);for(let o=0,i=a;i!==n;o++,i=a){a=e.openElements.getCommonAncestor(i);const n=e.activeFormattingElements.getElementEntry(i),s=n&&o>=3;!n||s?(s&&e.activeFormattingElements.removeEntry(n),e.openElements.remove(i)):(i=Le(e,n),r===t&&(e.activeFormattingElements.bookmark=n),e.treeAdapter.detachNode(r),e.treeAdapter.appendChild(i,r),r=i)}return r}function Le(e,t){const n=e.treeAdapter.getNamespaceURI(t.element),r=e.treeAdapter.createElement(t.token.tagName,n,t.token.attrs);return e.openElements.replace(t.element,r),t.element=r,r}function Be(e,t,n){const r=B(e.treeAdapter.getTagName(t));if(e._isElementCausesFosterParenting(r))e._fosterParentElement(n);else{const a=e.treeAdapter.getNamespaceURI(t);r===P.TEMPLATE&&a===C.HTML&&(t=e.treeAdapter.getTemplateContent(t)),e.treeAdapter.appendChild(t,n)}}function Fe(e,t,n){const r=e.treeAdapter.getNamespaceURI(n.element),{token:a}=n,o=e.treeAdapter.createElement(a.tagName,r,a.attrs);e._adoptNodes(t,o),e.treeAdapter.appendChild(t,o),e.activeFormattingElements.insertElementAfterBookmark(o,a),e.activeFormattingElements.removeEntry(n),e.openElements.remove(n.element),e.openElements.insertAfter(t,o,a.tagID)}function Qe(e,t){for(let n=0;n<8;n++){const n=$e(e,t);if(!n)break;const r=Ne(e,n);if(!r)break;e.activeFormattingElements.bookmark=n;const a=Ie(e,r,n.element),o=e.openElements.getCommonAncestor(n.element);e.treeAdapter.detachNode(a),o&&Be(e,o,a),Fe(e,r,n)}}function Ue(e,t){e._appendCommentNode(t,e.openElements.currentTmplContentOrNode)}function ze(e,t){if(e.stopped=!0,t.location){const n=e.fragmentContext?0:2;for(let r=e.openElements.stackTop;r>=n;r--)e._setEndLocation(e.openElements.items[r],t);if(!e.fragmentContext&&e.openElements.stackTop>=0){const n=e.openElements.items[0],r=e.treeAdapter.getNodeSourceCodeLocation(n);if(r&&!r.endTag&&(e._setEndLocation(n,t),e.openElements.stackTop>=1)){const n=e.openElements.items[1],r=e.treeAdapter.getNodeSourceCodeLocation(n);r&&!r.endTag&&e._setEndLocation(n,t)}}}}function je(e,t){e._err(t,f.missingDoctype,!0),e.treeAdapter.setDocumentMode(e.document,A.QUIRKS),e.insertionMode=Ae.BEFORE_HTML,e._processToken(t)}function We(e,t){e._insertFakeRootElement(),e.insertionMode=Ae.BEFORE_HEAD,e._processToken(t)}function Xe(e,t){e._insertFakeElement(R.HEAD,P.HEAD),e.headElement=e.openElements.current,e.insertionMode=Ae.IN_HEAD,e._processToken(t)}function qe(e,t){switch(t.tagID){case P.HTML:at(e,t);break;case P.BASE:case P.BASEFONT:case P.BGSOUND:case P.LINK:case P.META:e._appendElement(t,C.HTML),t.ackSelfClosing=!0;break;case P.TITLE:e._switchToTextParsing(t,X.RCDATA);break;case P.NOSCRIPT:e.options.scriptingEnabled?e._switchToTextParsing(t,X.RAWTEXT):(e._insertElement(t,C.HTML),e.insertionMode=Ae.IN_HEAD_NO_SCRIPT);break;case P.NOFRAMES:case P.STYLE:e._switchToTextParsing(t,X.RAWTEXT);break;case P.SCRIPT:e._switchToTextParsing(t,X.SCRIPT_DATA);break;case P.TEMPLATE:e._insertTemplate(t),e.activeFormattingElements.insertMarker(),e.framesetOk=!1,e.insertionMode=Ae.IN_TEMPLATE,e.tmplInsertionModeStack.unshift(Ae.IN_TEMPLATE);break;case P.HEAD:e._err(t,f.misplacedStartTagForHeadElement);break;default:He(e,t)}}function Ve(e,t){e.openElements.tmplCount>0?(e.openElements.generateImpliedEndTagsThoroughly(),e.openElements.currentTagId!==P.TEMPLATE&&e._err(t,f.closingOfElementWithOpenChildElements),e.openElements.popUntilTagNamePopped(P.TEMPLATE),e.activeFormattingElements.clearToLastMarker(),e.tmplInsertionModeStack.shift(),e._resetInsertionMode()):e._err(t,f.endTagWithoutMatchingOpenElement)}function He(e,t){e.openElements.pop(),e.insertionMode=Ae.AFTER_HEAD,e._processToken(t)}function Ze(e,t){const n=t.type===m.EOF?f.openElementsLeftAfterEof:f.disallowedContentInNoscriptInHead;e._err(t,n),e.openElements.pop(),e.insertionMode=Ae.IN_HEAD,e._processToken(t)}function Ge(e,t){e._insertFakeElement(R.BODY,P.BODY),e.insertionMode=Ae.IN_BODY,Ke(e,t)}function Ke(e,t){switch(t.type){case m.CHARACTER:Je(e,t);break;case m.WHITESPACE_CHARACTER:Ye(e,t);break;case m.COMMENT:Ue(e,t);break;case m.START_TAG:at(e,t);break;case m.END_TAG:it(e,t);break;case m.EOF:st(e,t)}}function Ye(e,t){e._reconstructActiveFormattingElements(),e._insertCharacters(t)}function Je(e,t){e._reconstructActiveFormattingElements(),e._insertCharacters(t),e.framesetOk=!1}function et(e,t){e._reconstructActiveFormattingElements(),e._appendElement(t,C.HTML),e.framesetOk=!1,t.ackSelfClosing=!0}function tt(e){const t=v(e,T.TYPE);return null!=t&&"hidden"===t.toLowerCase()}function nt(e,t){e._switchToTextParsing(t,X.RAWTEXT)}function rt(e,t){e._reconstructActiveFormattingElements(),e._insertElement(t,C.HTML)}function at(e,t){switch(t.tagID){case P.I:case P.S:case P.B:case P.U:case P.EM:case P.TT:case P.BIG:case P.CODE:case P.FONT:case P.SMALL:case P.STRIKE:case P.STRONG:!function(e,t){e._reconstructActiveFormattingElements(),e._insertElement(t,C.HTML),e.activeFormattingElements.pushElement(e.openElements.current,t)}(e,t);break;case P.A:!function(e,t){const n=e.activeFormattingElements.getElementEntryInScopeWithTagName(R.A);n&&(Qe(e,t),e.openElements.remove(n.element),e.activeFormattingElements.removeEntry(n)),e._reconstructActiveFormattingElements(),e._insertElement(t,C.HTML),e.activeFormattingElements.pushElement(e.openElements.current,t)}(e,t);break;case P.H1:case P.H2:case P.H3:case P.H4:case P.H5:case P.H6:!function(e,t){e.openElements.hasInButtonScope(P.P)&&e._closePElement(),U.has(e.openElements.currentTagId)&&e.openElements.pop(),e._insertElement(t,C.HTML)}(e,t);break;case P.P:case P.DL:case P.OL:case P.UL:case P.DIV:case P.DIR:case P.NAV:case P.MAIN:case P.MENU:case P.ASIDE:case P.CENTER:case P.FIGURE:case P.FOOTER:case P.HEADER:case P.HGROUP:case P.DIALOG:case P.DETAILS:case P.ADDRESS:case P.ARTICLE:case P.SEARCH:case P.SECTION:case P.SUMMARY:case P.FIELDSET:case P.BLOCKQUOTE:case P.FIGCAPTION:!function(e,t){e.openElements.hasInButtonScope(P.P)&&e._closePElement(),e._insertElement(t,C.HTML)}(e,t);break;case P.LI:case P.DD:case P.DT:!function(e,t){e.framesetOk=!1;const n=t.tagID;for(let t=e.openElements.stackTop;t>=0;t--){const r=e.openElements.tagIDs[t];if(n===P.LI&&r===P.LI||(n===P.DD||n===P.DT)&&(r===P.DD||r===P.DT)){e.openElements.generateImpliedEndTagsWithExclusion(r),e.openElements.popUntilTagNamePopped(r);break}if(r!==P.ADDRESS&&r!==P.DIV&&r!==P.P&&e._isSpecialElement(e.openElements.items[t],r))break}e.openElements.hasInButtonScope(P.P)&&e._closePElement(),e._insertElement(t,C.HTML)}(e,t);break;case P.BR:case P.IMG:case P.WBR:case P.AREA:case P.EMBED:case P.KEYGEN:et(e,t);break;case P.HR:!function(e,t){e.openElements.hasInButtonScope(P.P)&&e._closePElement(),e._appendElement(t,C.HTML),e.framesetOk=!1,t.ackSelfClosing=!0}(e,t);break;case P.RB:case P.RTC:!function(e,t){e.openElements.hasInScope(P.RUBY)&&e.openElements.generateImpliedEndTags(),e._insertElement(t,C.HTML)}(e,t);break;case P.RT:case P.RP:!function(e,t){e.openElements.hasInScope(P.RUBY)&&e.openElements.generateImpliedEndTagsWithExclusion(P.RTC),e._insertElement(t,C.HTML)}(e,t);break;case P.PRE:case P.LISTING:!function(e,t){e.openElements.hasInButtonScope(P.P)&&e._closePElement(),e._insertElement(t,C.HTML),e.skipNextNewLine=!0,e.framesetOk=!1}(e,t);break;case P.XMP:!function(e,t){e.openElements.hasInButtonScope(P.P)&&e._closePElement(),e._reconstructActiveFormattingElements(),e.framesetOk=!1,e._switchToTextParsing(t,X.RAWTEXT)}(e,t);break;case P.SVG:!function(e,t){e._reconstructActiveFormattingElements(),Ce(t),Te(t),t.selfClosing?e._appendElement(t,C.SVG):e._insertElement(t,C.SVG),t.ackSelfClosing=!0}(e,t);break;case P.HTML:!function(e,t){0===e.openElements.tmplCount&&e.treeAdapter.adoptAttributes(e.openElements.items[0],t.attrs)}(e,t);break;case P.BASE:case P.LINK:case P.META:case P.STYLE:case P.TITLE:case P.SCRIPT:case P.BGSOUND:case P.BASEFONT:case P.TEMPLATE:qe(e,t);break;case P.BODY:!function(e,t){const n=e.openElements.tryPeekProperlyNestedBodyElement();n&&0===e.openElements.tmplCount&&(e.framesetOk=!1,e.treeAdapter.adoptAttributes(n,t.attrs))}(e,t);break;case P.FORM:!function(e,t){const n=e.openElements.tmplCount>0;e.formElement&&!n||(e.openElements.hasInButtonScope(P.P)&&e._closePElement(),e._insertElement(t,C.HTML),n||(e.formElement=e.openElements.current))}(e,t);break;case P.NOBR:!function(e,t){e._reconstructActiveFormattingElements(),e.openElements.hasInScope(P.NOBR)&&(Qe(e,t),e._reconstructActiveFormattingElements()),e._insertElement(t,C.HTML),e.activeFormattingElements.pushElement(e.openElements.current,t)}(e,t);break;case P.MATH:!function(e,t){e._reconstructActiveFormattingElements(),_e(t),Te(t),t.selfClosing?e._appendElement(t,C.MATHML):e._insertElement(t,C.MATHML),t.ackSelfClosing=!0}(e,t);break;case P.TABLE:!function(e,t){e.treeAdapter.getDocumentMode(e.document)!==A.QUIRKS&&e.openElements.hasInButtonScope(P.P)&&e._closePElement(),e._insertElement(t,C.HTML),e.framesetOk=!1,e.insertionMode=Ae.IN_TABLE}(e,t);break;case P.INPUT:!function(e,t){e._reconstructActiveFormattingElements(),e._appendElement(t,C.HTML),tt(t)||(e.framesetOk=!1),t.ackSelfClosing=!0}(e,t);break;case P.PARAM:case P.TRACK:case P.SOURCE:!function(e,t){e._appendElement(t,C.HTML),t.ackSelfClosing=!0}(e,t);break;case P.IMAGE:!function(e,t){t.tagName=R.IMG,t.tagID=P.IMG,et(e,t)}(e,t);break;case P.BUTTON:!function(e,t){e.openElements.hasInScope(P.BUTTON)&&(e.openElements.generateImpliedEndTags(),e.openElements.popUntilTagNamePopped(P.BUTTON)),e._reconstructActiveFormattingElements(),e._insertElement(t,C.HTML),e.framesetOk=!1}(e,t);break;case P.APPLET:case P.OBJECT:case P.MARQUEE:!function(e,t){e._reconstructActiveFormattingElements(),e._insertElement(t,C.HTML),e.activeFormattingElements.insertMarker(),e.framesetOk=!1}(e,t);break;case P.IFRAME:!function(e,t){e.framesetOk=!1,e._switchToTextParsing(t,X.RAWTEXT)}(e,t);break;case P.SELECT:!function(e,t){e._reconstructActiveFormattingElements(),e._insertElement(t,C.HTML),e.framesetOk=!1,e.insertionMode=e.insertionMode===Ae.IN_TABLE||e.insertionMode===Ae.IN_CAPTION||e.insertionMode===Ae.IN_TABLE_BODY||e.insertionMode===Ae.IN_ROW||e.insertionMode===Ae.IN_CELL?Ae.IN_SELECT_IN_TABLE:Ae.IN_SELECT}(e,t);break;case P.OPTION:case P.OPTGROUP:!function(e,t){e.openElements.currentTagId===P.OPTION&&e.openElements.pop(),e._reconstructActiveFormattingElements(),e._insertElement(t,C.HTML)}(e,t);break;case P.NOEMBED:case P.NOFRAMES:nt(e,t);break;case P.FRAMESET:!function(e,t){const n=e.openElements.tryPeekProperlyNestedBodyElement();e.framesetOk&&n&&(e.treeAdapter.detachNode(n),e.openElements.popAllUpToHtmlElement(),e._insertElement(t,C.HTML),e.insertionMode=Ae.IN_FRAMESET)}(e,t);break;case P.TEXTAREA:!function(e,t){e._insertElement(t,C.HTML),e.skipNextNewLine=!0,e.tokenizer.state=X.RCDATA,e.originalInsertionMode=e.insertionMode,e.framesetOk=!1,e.insertionMode=Ae.TEXT}(e,t);break;case P.NOSCRIPT:e.options.scriptingEnabled?nt(e,t):rt(e,t);break;case P.PLAINTEXT:!function(e,t){e.openElements.hasInButtonScope(P.P)&&e._closePElement(),e._insertElement(t,C.HTML),e.tokenizer.state=X.PLAINTEXT}(e,t);break;case P.COL:case P.TH:case P.TD:case P.TR:case P.HEAD:case P.FRAME:case P.TBODY:case P.TFOOT:case P.THEAD:case P.CAPTION:case P.COLGROUP:break;default:rt(e,t)}}function ot(e,t){const n=t.tagName,r=t.tagID;for(let t=e.openElements.stackTop;t>0;t--){const a=e.openElements.items[t],o=e.openElements.tagIDs[t];if(r===o&&(r!==P.UNKNOWN||e.treeAdapter.getTagName(a)===n)){e.openElements.generateImpliedEndTagsWithExclusion(r),e.openElements.stackTop>=t&&e.openElements.shortenToLength(t);break}if(e._isSpecialElement(a,o))break}}function it(e,t){switch(t.tagID){case P.A:case P.B:case P.I:case P.S:case P.U:case P.EM:case P.TT:case P.BIG:case P.CODE:case P.FONT:case P.NOBR:case P.SMALL:case P.STRIKE:case P.STRONG:Qe(e,t);break;case P.P:!function(e){e.openElements.hasInButtonScope(P.P)||e._insertFakeElement(R.P,P.P),e._closePElement()}(e);break;case P.DL:case P.UL:case P.OL:case P.DIR:case P.DIV:case P.NAV:case P.PRE:case P.MAIN:case P.MENU:case P.ASIDE:case P.BUTTON:case P.CENTER:case P.FIGURE:case P.FOOTER:case P.HEADER:case P.HGROUP:case P.DIALOG:case P.ADDRESS:case P.ARTICLE:case P.DETAILS:case P.SEARCH:case P.SECTION:case P.SUMMARY:case P.LISTING:case P.FIELDSET:case P.BLOCKQUOTE:case P.FIGCAPTION:!function(e,t){const n=t.tagID;e.openElements.hasInScope(n)&&(e.openElements.generateImpliedEndTags(),e.openElements.popUntilTagNamePopped(n))}(e,t);break;case P.LI:!function(e){e.openElements.hasInListItemScope(P.LI)&&(e.openElements.generateImpliedEndTagsWithExclusion(P.LI),e.openElements.popUntilTagNamePopped(P.LI))}(e);break;case P.DD:case P.DT:!function(e,t){const n=t.tagID;e.openElements.hasInScope(n)&&(e.openElements.generateImpliedEndTagsWithExclusion(n),e.openElements.popUntilTagNamePopped(n))}(e,t);break;case P.H1:case P.H2:case P.H3:case P.H4:case P.H5:case P.H6:!function(e){e.openElements.hasNumberedHeaderInScope()&&(e.openElements.generateImpliedEndTags(),e.openElements.popUntilNumberedHeaderPopped())}(e);break;case P.BR:!function(e){e._reconstructActiveFormattingElements(),e._insertFakeElement(R.BR,P.BR),e.openElements.pop(),e.framesetOk=!1}(e);break;case P.BODY:!function(e,t){if(e.openElements.hasInScope(P.BODY)&&(e.insertionMode=Ae.AFTER_BODY,e.options.sourceCodeLocationInfo)){const n=e.openElements.tryPeekProperlyNestedBodyElement();n&&e._setEndLocation(n,t)}}(e,t);break;case P.HTML:!function(e,t){e.openElements.hasInScope(P.BODY)&&(e.insertionMode=Ae.AFTER_BODY,St(e,t))}(e,t);break;case P.FORM:!function(e){const t=e.openElements.tmplCount>0,{formElement:n}=e;t||(e.formElement=null),(n||t)&&e.openElements.hasInScope(P.FORM)&&(e.openElements.generateImpliedEndTags(),t?e.openElements.popUntilTagNamePopped(P.FORM):n&&e.openElements.remove(n))}(e);break;case P.APPLET:case P.OBJECT:case P.MARQUEE:!function(e,t){const n=t.tagID;e.openElements.hasInScope(n)&&(e.openElements.generateImpliedEndTags(),e.openElements.popUntilTagNamePopped(n),e.activeFormattingElements.clearToLastMarker())}(e,t);break;case P.TEMPLATE:Ve(e,t);break;default:ot(e,t)}}function st(e,t){e.tmplInsertionModeStack.length>0?Et(e,t):ze(e,t)}function lt(e,t){if(Pe.has(e.openElements.currentTagId))switch(e.pendingCharacterTokens.length=0,e.hasNonWhitespacePendingCharacterToken=!1,e.originalInsertionMode=e.insertionMode,e.insertionMode=Ae.IN_TABLE_TEXT,t.type){case m.CHARACTER:ht(e,t);break;case m.WHITESPACE_CHARACTER:pt(e,t)}else dt(e,t)}function ct(e,t){switch(t.tagID){case P.TD:case P.TH:case P.TR:!function(e,t){e.openElements.clearBackToTableContext(),e._insertFakeElement(R.TBODY,P.TBODY),e.insertionMode=Ae.IN_TABLE_BODY,bt(e,t)}(e,t);break;case P.STYLE:case P.SCRIPT:case P.TEMPLATE:qe(e,t);break;case P.COL:!function(e,t){e.openElements.clearBackToTableContext(),e._insertFakeElement(R.COLGROUP,P.COLGROUP),e.insertionMode=Ae.IN_COLUMN_GROUP,gt(e,t)}(e,t);break;case P.FORM:!function(e,t){e.formElement||0!==e.openElements.tmplCount||(e._insertElement(t,C.HTML),e.formElement=e.openElements.current,e.openElements.pop())}(e,t);break;case P.TABLE:!function(e,t){e.openElements.hasInTableScope(P.TABLE)&&(e.openElements.popUntilTagNamePopped(P.TABLE),e._resetInsertionMode(),e._processStartTag(t))}(e,t);break;case P.TBODY:case P.TFOOT:case P.THEAD:!function(e,t){e.openElements.clearBackToTableContext(),e._insertElement(t,C.HTML),e.insertionMode=Ae.IN_TABLE_BODY}(e,t);break;case P.INPUT:!function(e,t){tt(t)?e._appendElement(t,C.HTML):dt(e,t),t.ackSelfClosing=!0}(e,t);break;case P.CAPTION:!function(e,t){e.openElements.clearBackToTableContext(),e.activeFormattingElements.insertMarker(),e._insertElement(t,C.HTML),e.insertionMode=Ae.IN_CAPTION}(e,t);break;case P.COLGROUP:!function(e,t){e.openElements.clearBackToTableContext(),e._insertElement(t,C.HTML),e.insertionMode=Ae.IN_COLUMN_GROUP}(e,t);break;default:dt(e,t)}}function ut(e,t){switch(t.tagID){case P.TABLE:e.openElements.hasInTableScope(P.TABLE)&&(e.openElements.popUntilTagNamePopped(P.TABLE),e._resetInsertionMode());break;case P.TEMPLATE:Ve(e,t);break;case P.BODY:case P.CAPTION:case P.COL:case P.COLGROUP:case P.HTML:case P.TBODY:case P.TD:case P.TFOOT:case P.TH:case P.THEAD:case P.TR:break;default:dt(e,t)}}function dt(e,t){const n=e.fosterParentingEnabled;e.fosterParentingEnabled=!0,Ke(e,t),e.fosterParentingEnabled=n}function pt(e,t){e.pendingCharacterTokens.push(t)}function ht(e,t){e.pendingCharacterTokens.push(t),e.hasNonWhitespacePendingCharacterToken=!0}function ft(e,t){let n=0;if(e.hasNonWhitespacePendingCharacterToken)for(;n<e.pendingCharacterTokens.length;n++)dt(e,e.pendingCharacterTokens[n]);else for(;n<e.pendingCharacterTokens.length;n++)e._insertCharacters(e.pendingCharacterTokens[n]);e.insertionMode=e.originalInsertionMode,e._processToken(t)}const mt=new Set([P.CAPTION,P.COL,P.COLGROUP,P.TBODY,P.TD,P.TFOOT,P.TH,P.THEAD,P.TR]);function gt(e,t){switch(t.tagID){case P.HTML:at(e,t);break;case P.COL:e._appendElement(t,C.HTML),t.ackSelfClosing=!0;break;case P.TEMPLATE:qe(e,t);break;default:vt(e,t)}}function vt(e,t){e.openElements.currentTagId===P.COLGROUP&&(e.openElements.pop(),e.insertionMode=Ae.IN_TABLE,e._processToken(t))}function bt(e,t){switch(t.tagID){case P.TR:e.openElements.clearBackToTableBodyContext(),e._insertElement(t,C.HTML),e.insertionMode=Ae.IN_ROW;break;case P.TH:case P.TD:e.openElements.clearBackToTableBodyContext(),e._insertFakeElement(R.TR,P.TR),e.insertionMode=Ae.IN_ROW,wt(e,t);break;case P.CAPTION:case P.COL:case P.COLGROUP:case P.TBODY:case P.TFOOT:case P.THEAD:e.openElements.hasTableBodyContextInTableScope()&&(e.openElements.clearBackToTableBodyContext(),e.openElements.pop(),e.insertionMode=Ae.IN_TABLE,ct(e,t));break;default:ct(e,t)}}function yt(e,t){const n=t.tagID;switch(t.tagID){case P.TBODY:case P.TFOOT:case P.THEAD:e.openElements.hasInTableScope(n)&&(e.openElements.clearBackToTableBodyContext(),e.openElements.pop(),e.insertionMode=Ae.IN_TABLE);break;case P.TABLE:e.openElements.hasTableBodyContextInTableScope()&&(e.openElements.clearBackToTableBodyContext(),e.openElements.pop(),e.insertionMode=Ae.IN_TABLE,ut(e,t));break;case P.BODY:case P.CAPTION:case P.COL:case P.COLGROUP:case P.HTML:case P.TD:case P.TH:case P.TR:break;default:ut(e,t)}}function wt(e,t){switch(t.tagID){case P.TH:case P.TD:e.openElements.clearBackToTableRowContext(),e._insertElement(t,C.HTML),e.insertionMode=Ae.IN_CELL,e.activeFormattingElements.insertMarker();break;case P.CAPTION:case P.COL:case P.COLGROUP:case P.TBODY:case P.TFOOT:case P.THEAD:case P.TR:e.openElements.hasInTableScope(P.TR)&&(e.openElements.clearBackToTableRowContext(),e.openElements.pop(),e.insertionMode=Ae.IN_TABLE_BODY,bt(e,t));break;default:ct(e,t)}}function Ot(e,t){switch(t.tagID){case P.TR:e.openElements.hasInTableScope(P.TR)&&(e.openElements.clearBackToTableRowContext(),e.openElements.pop(),e.insertionMode=Ae.IN_TABLE_BODY);break;case P.TABLE:e.openElements.hasInTableScope(P.TR)&&(e.openElements.clearBackToTableRowContext(),e.openElements.pop(),e.insertionMode=Ae.IN_TABLE_BODY,yt(e,t));break;case P.TBODY:case P.TFOOT:case P.THEAD:(e.openElements.hasInTableScope(t.tagID)||e.openElements.hasInTableScope(P.TR))&&(e.openElements.clearBackToTableRowContext(),e.openElements.pop(),e.insertionMode=Ae.IN_TABLE_BODY,yt(e,t));break;case P.BODY:case P.CAPTION:case P.COL:case P.COLGROUP:case P.HTML:case P.TD:case P.TH:break;default:ut(e,t)}}function kt(e,t){switch(t.tagID){case P.HTML:at(e,t);break;case P.OPTION:e.openElements.currentTagId===P.OPTION&&e.openElements.pop(),e._insertElement(t,C.HTML);break;case P.OPTGROUP:e.openElements.currentTagId===P.OPTION&&e.openElements.pop(),e.openElements.currentTagId===P.OPTGROUP&&e.openElements.pop(),e._insertElement(t,C.HTML);break;case P.HR:e.openElements.currentTagId===P.OPTION&&e.openElements.pop(),e.openElements.currentTagId===P.OPTGROUP&&e.openElements.pop(),e._appendElement(t,C.HTML),t.ackSelfClosing=!0;break;case P.INPUT:case P.KEYGEN:case P.TEXTAREA:case P.SELECT:e.openElements.hasInSelectScope(P.SELECT)&&(e.openElements.popUntilTagNamePopped(P.SELECT),e._resetInsertionMode(),t.tagID!==P.SELECT&&e._processStartTag(t));break;case P.SCRIPT:case P.TEMPLATE:qe(e,t)}}function xt(e,t){switch(t.tagID){case P.OPTGROUP:e.openElements.stackTop>0&&e.openElements.currentTagId===P.OPTION&&e.openElements.tagIDs[e.openElements.stackTop-1]===P.OPTGROUP&&e.openElements.pop(),e.openElements.currentTagId===P.OPTGROUP&&e.openElements.pop();break;case P.OPTION:e.openElements.currentTagId===P.OPTION&&e.openElements.pop();break;case P.SELECT:e.openElements.hasInSelectScope(P.SELECT)&&(e.openElements.popUntilTagNamePopped(P.SELECT),e._resetInsertionMode());break;case P.TEMPLATE:Ve(e,t)}}function Et(e,t){e.openElements.tmplCount>0?(e.openElements.popUntilTagNamePopped(P.TEMPLATE),e.activeFormattingElements.clearToLastMarker(),e.tmplInsertionModeStack.shift(),e._resetInsertionMode(),e.onEof(t)):ze(e,t)}function St(e,t){var n;if(t.tagID===P.HTML){if(e.fragmentContext||(e.insertionMode=Ae.AFTER_AFTER_BODY),e.options.sourceCodeLocationInfo&&e.openElements.tagIDs[0]===P.HTML){e._setEndLocation(e.openElements.items[0],t);const r=e.openElements.items[1];r&&!(null===(n=e.treeAdapter.getNodeSourceCodeLocation(r))||void 0===n?void 0:n.endTag)&&e._setEndLocation(r,t)}}else _t(e,t)}function _t(e,t){e.insertionMode=Ae.IN_BODY,Ke(e,t)}function Ct(e,t){e.insertionMode=Ae.IN_BODY,Ke(e,t)}function Tt(e){for(;e.treeAdapter.getNamespaceURI(e.openElements.current)!==C.HTML&&!e._isIntegrationPoint(e.openElements.currentTagId,e.openElements.current);)e.openElements.pop()}const At=new Map([[34,"&quot;"],[38,"&amp;"],[39,"&apos;"],[60,"&lt;"],[62,"&gt;"]]);function Rt(e,t){return function(n){let r,a=0,o="";for(;r=e.exec(n);)a!==r.index&&(o+=n.substring(a,r.index)),o+=t.get(r[0].charCodeAt(0)),a=r.index+1;return o+n.substring(a)}}function Pt(e,t){return Me.parse(e,t)}function Dt(e,t,n){"string"==typeof e&&(n=t,t=e,e=null);const r=Me.getFragmentParser(e,n);return r.tokenizer.write(t,!0),r.getFragment()}String.prototype.codePointAt,Rt(/[&<>'"]/g,At),Rt(/["&\u00A0]/g,new Map([[34,"&quot;"],[38,"&amp;"],[160,"&nbsp;"]])),Rt(/[&<>\u00A0]/g,new Map([[38,"&amp;"],[60,"&lt;"],[62,"&gt;"],[160,"&nbsp;"]])),new Set([R.AREA,R.BASE,R.BASEFONT,R.BGSOUND,R.BR,R.COL,R.EMBED,R.FRAME,R.HR,R.IMG,R.INPUT,R.KEYGEN,R.LINK,R.META,R.PARAM,R.SOURCE,R.TRACK,R.WBR])},4703:(e,t,n)=>{"use strict";n.d(t,{oE:()=>m,vA:()=>l,$W:()=>p,kh:()=>f});var r=n(4476),a=n(2071),o=n(5182),i=n(498);const s=e=>{const t={};if(e.variables)for(const[n,r]of Object.entries(e.variables))t[n]=r.enum?.filter((e=>"string"==typeof e))??[r.default];return t},l=r.z.object({key:r.z.string().default(""),value:r.z.coerce.string().default(""),enabled:r.z.boolean().default(!0),file:r.z.any().optional(),description:r.z.string().optional(),required:r.z.boolean().optional(),enum:r.z.array(r.z.string()).optional(),examples:r.z.array(r.z.string()).optional(),type:r.z.union([r.z.string(),r.z.array(r.z.string())]).optional(),format:r.z.string().optional(),minimum:r.z.number().optional(),maximum:r.z.number().optional(),default:r.z.any().optional(),nullable:r.z.boolean().optional()}).transform((e=>{const t={...e};return Array.isArray(t.type)&&t.type.includes("null")&&(t.nullable=!0),Array.isArray(t.type)&&2===t.type.length&&t.type.includes("null")&&(t.type=t.type.find((e=>"null"!==e))),t})),c=r.z.object({url:r.z.string(),base64:r.z.string().optional()}).nullable(),u=(r.z.union([r.z.object({type:r.z.literal("string"),value:r.z.string()}),r.z.object({type:r.z.literal("file"),file:c})]),r.z.object({raw:r.z.object({encoding:r.z.enum(["json","text","html","javascript","xml","yaml","edn"]),value:r.z.string().default("")}).optional(),formData:r.z.object({encoding:r.z.union([r.z.literal("form-data"),r.z.literal("urlencoded")]).default("form-data"),value:l.array().default([])}).optional(),binary:r.z.instanceof(Blob).optional(),activeBody:r.z.union([r.z.literal("raw"),r.z.literal("formData"),r.z.literal("binary")]).default("raw")})),d=r.z.object({encoding:r.z.enum(["application/json","text/plain","text/html","application/javascript","application/xml","application/yaml","application/edn","application/octet-stream","application/x-www-form-urlencoded","multipart/form-data","binary"]),content:r.z.union([r.z.record(r.z.string(),r.z.any()),r.z.string()]),file:c.optional()}),p=r.z.object({uid:a.B.brand(),type:r.z.literal("requestExample").optional().default("requestExample"),requestUid:r.z.string().brand().optional(),name:r.z.string().optional().default("Name"),body:u.optional().default({}),parameters:r.z.object({path:l.array().default([]),query:l.array().default([]),headers:l.array().default([{key:"Accept",value:"*/*",enabled:!0}]),cookies:l.array().default([])}).optional().default({}),serverVariables:r.z.record(r.z.string(),r.z.array(r.z.string())).optional()}),h=r.z.record(r.z.string(),r.z.string()).optional(),f=r.z.object({name:r.z.string().optional(),body:d.optional(),parameters:r.z.object({path:h,query:h,headers:h,cookies:h})});function m(e,t,n){const r={path:[],query:[],cookie:[],header:[],headers:[{key:"Accept",value:"*/*",enabled:!0}]};e.parameters?.forEach((e=>r[e.in].push(function(e){const t=e.schema,n=Object.keys(e?.examples??{}),r=n.length?e.examples?.[n[0]]:null,a=String(t?.default??t?.examples?.[0]??t?.example??r?.value??e.example??""),i=t?.enum&&"string"!==t?.type?t.enum?.map(String):t?.items?.enum&&"array"===t?.type?t.items.enum.map(String):t?.enum,s=t?.examples&&"string"!==t?.type?t.examples?.map(String):t?.examples;return(0,o.T)({...t,key:e.name,value:a,description:e.description,required:e.required,enabled:!!e.required,enum:i,examples:s},l,!1)||(console.warn(`Example at ${e.name} is invalid.`),l.parse({}))}(e)))),r.header.length>0&&(r.headers=r.header,r.header=[]);const a={activeBody:"raw"};if(e.requestBody){const t=(0,i.z)({path:e.path,information:{requestBody:e.requestBody}});"application/json"===t?.mimeType&&(a.activeBody="raw",a.raw={encoding:"json",value:t.text??JSON.stringify({})}),"application/xml"===t?.mimeType&&(a.activeBody="raw",a.raw={encoding:"xml",value:t.text??""}),"application/octet-stream"===t?.mimeType&&(a.activeBody="binary",a.binary=void 0),"application/x-www-form-urlencoded"!==t?.mimeType&&"multipart/form-data"!==t?.mimeType||(a.activeBody="formData",a.formData={encoding:"application/x-www-form-urlencoded"===t.mimeType?"urlencoded":"form-data",value:(t.params||[]).map((e=>({key:e.name,value:e.value||"",enabled:!0})))}),t?.mimeType&&!r.headers.find((e=>"content-type"===e.key.toLowerCase()))&&r.headers.push({key:"Content-Type",value:t.mimeType,enabled:!0})}const c=n?s(n):{};return(0,o.T)({requestUid:e.uid,parameters:r,name:t,body:a,serverVariables:c},p,!1)||(console.warn(`Example at ${e.uid} is invalid.`),p.parse({}))}},4748:(e,t,n)=>{"use strict";n.d(t,{be:()=>k,D0:()=>w});var r=n(5785),a=n(4115),o=n(5182),i=n(5633),s=n(5362),l=n(1507),c=n(3412);function u(e){let t=e;return null!==t&&"string"==typeof t.openapi&&t.openapi.startsWith("3.0")?(t.openapi="3.1.1",t=(0,c.h)(t,(e=>("undefined"!==e.type&&!0===e.nullable&&(e.type=[e.type,"null"],delete e.nullable),e))),t=(0,c.h)(t,(e=>(!0===e.exclusiveMinimum?(e.exclusiveMinimum=e.minimum,delete e.minimum):!1===e.exclusiveMinimum&&delete e.exclusiveMinimum,!0===e.exclusiveMaximum?(e.exclusiveMaximum=e.maximum,delete e.maximum):!1===e.exclusiveMaximum&&delete e.exclusiveMaximum,e))),t=(0,c.h)(t,((e,t)=>(void 0!==e.example&&(function(e){return[["components","schemas"],"properties","items","allOf","anyOf","oneOf","not","additionalProperties"].some((t=>Array.isArray(t)?t.every(((t,n)=>e[n]===t)):e.includes(t)))||e.includes("schema")||e.some((e=>e.endsWith("Schema")))}(t)?e.examples=[e.example]:e.examples={default:{value:e.example}},delete e.example),e))),t=(0,c.h)(t,((e,t)=>{if("object"===e.type&&void 0!==e.properties&&t.slice(0,-1).some(((e,n)=>"content"===e&&"multipart/form-data"===t[n+1]))){const t=Object.entries(e.properties);for(const[e,n]of t)"object"==typeof n&&"string"===n.type&&"binary"===n.format&&(n.contentMediaType="application/octet-stream",delete n.format)}return e})),t=(0,c.h)(t,((e,t)=>t.includes("content")&&t.includes("application/octet-stream")?{}:"string"===e.type&&"binary"===e.format?{type:"string",contentMediaType:"application/octet-stream"}:e)),t=(0,c.h)(t,(e=>"string"===e.type&&"base64"===e.format?{type:"string",contentEncoding:"base64"}:e)),t):t}function d(e){let t=e;if(null===t||"string"!=typeof t.swagger||!t.swagger?.startsWith("2.0"))return t;if(t.openapi="3.0.4",delete t.swagger,console.warn("[upgradeFromTwoToThree] The upgrade from Swagger 2.0 to OpenAPI 3.0 documents is experimental and lacks features."),t.host){const e=Array.isArray(t.schemes)&&t.schemes?.length?t.schemes:["http"];t.servers=e.map((e=>({url:`${e}://${t.host}${t.basePath??""}`}))),delete t.basePath,delete t.schemes,delete t.host}if(t.definitions&&(t.components=Object.assign({},t.components,{schemas:t.definitions}),delete t.definitions,t=(0,c.h)(t,(e=>(e.$ref?.startsWith("#/definitions/")&&(e.$ref=e.$ref.replace(/^#\/definitions\//,"#/components/schemas/")),e)))),"object"==typeof t.paths)for(const e in t.paths)if(Object.hasOwn(t.paths,e)){const n=t.paths[e];for(const e in n)if(Object.hasOwn(n,e)){const r=n[e];if(r.parameters){const e=structuredClone(r.parameters.find((e=>"body"===e.in))??{});if(e&&Object.keys(e).length){delete e.name,delete e.in;const n=t.consumes??r.consumes??["application/json"];"object"!=typeof r.requestBody&&(r.requestBody={}),"object"!=typeof r.requestBody.content&&(r.requestBody.content={});const{schema:a,...o}=e;r.requestBody={...r.requestBody,...o};for(const e of n)r.requestBody.content[e]={schema:a}}r.parameters=r.parameters.filter((e=>"body"!==e.in)),delete r.consumes;const n=r.parameters.filter((e=>"formData"===e.in));if(n.length>0){"object"!=typeof r.requestBody&&(r.requestBody={}),"object"!=typeof r.requestBody.content&&(r.requestBody.content={}),r.requestBody.content["application/x-www-form-urlencoded"]={schema:{type:"object",properties:{},required:[]}};for(const e of n)r.requestBody.content["application/x-www-form-urlencoded"].schema.properties[e.name]={type:e.type,description:e.description},e.required&&r.requestBody.content["application/x-www-form-urlencoded"].schema.required.push(e.name);r.parameters=r.parameters.filter((e=>"formData"!==e.in))}}if(r.responses)for(const e in r.responses)if(Object.hasOwn(r.responses,e)){const n=r.responses[e];if(n.schema){const e=t.produces??r.produces??["application/json"];"object"!=typeof n.content&&(n.content={});for(const t of e)n.content[t]={schema:n.schema};delete n.schema}}delete r.produces,0===r.parameters?.length&&delete r.parameters}}if(t.securityDefinitions){"object"!=typeof t.components&&(t.components={}),t.components=t.components,Object.assign(t.components,{securitySchemes:{}});for(const[e,n]of Object.entries(t.securityDefinitions))if("object"==typeof n)if("type"in n&&"oauth2"===n.type){const{flow:r,authorizationUrl:a,tokenUrl:o,scopes:i}=n;Object.assign(t.components.securitySchemes,{[e]:{type:"oauth2",flows:{[r]:Object.assign({},a&&{authorizationUrl:a},o&&{tokenUrl:o},i&&{scopes:i})}}})}else Object.assign(t.components.securitySchemes,{[e]:n});delete t.securityDefinitions}return t}var p=n(4074),h=n(1608),f=n(265),m=n(7297),g=n(8065),v=n(4703),b=n(7984),y=n(5148);const w=async(e,{shouldLoad:t=!0}={})=>{if(null===e||"string"==typeof e&&""===e.trim())return console.warn("[@scalar/oas-utils] Empty OpenAPI document provided."),{schema:{},errors:[]};let n=e,r=[];if(t){const t=await(0,i.H)(e).catch((e=>({errors:[{code:e.code,message:e.message}],filesystem:[]})));n=t.filesystem,r=t.errors??[]}const{specification:a}=(o=n)?{specification:[d,u].reduce(((e,t)=>t(e)),(0,s.z)((0,l.Y)(o)).specification),version:"3.1"}:{specification:null,version:"3.1"};var o;const{schema:c,errors:h=[]}=await(0,p.D)(a);return c||console.warn("[@scalar/oas-utils] OpenAPI Parser Warning: Schema is undefined"),{schema:Array.isArray(c)?{}:c,errors:[...r,...h]}},O=(e,t=[],n)=>(e[0]&&!t.length?[e[0]]:t).map((e=>Array.isArray(e)?e.map((e=>n[e])):n[e]));async function k(e,{authentication:t,baseServerURL:n,documentUrl:i,servers:s,setCollectionSecurity:l=!1,shouldLoad:c,watchMode:u=!1}={}){const{schema:d,errors:p}=await w(e,{shouldLoad:c}),k=[...p.map((e=>e.message))];if(!d)return{importWarnings:k,error:!0};const E=performance.now(),S=[],_=function(e,{baseServerURL:t}={}){return e&&Array.isArray(e)?e.map((e=>{try{const n=h.zT.parse(e);if(n?.url?.startsWith("/")){if(t)return n.url=(0,y.Dk)(t,n.url),n;const e=x();if(e)return n.url=(0,y.Dk)(e,n.url.replace(/^\//,"")),n}return n}catch(t){return console.warn("Oops, that’s an invalid server configuration."),console.warn("Server:",e),void console.warn("Error:",t)}})).filter(f.O):[]}(s||d.servers,{baseServerURL:n});if(!_.length){const e=x();e&&_.push(h.zT.parse({url:e}))}const C=new Set,T=d.components?.securitySchemes??d?.securityDefinitions??{},A=Object.entries(T).map?.((([e,n])=>{const a={...n,nameKey:e};"oauth2"===a.type&&a.flows?Object.keys(a.flows).forEach((e=>{if(!a.flows?.[e])return;const n=a.flows[e];n.type=e,t?.oAuth2&&(t.oAuth2.accessToken&&(n.token=t.oAuth2.accessToken),"password"===n.type&&(n.username=t.oAuth2.username,n.password=t.oAuth2.password),t.oAuth2.scopes&&(n.selectedScopes=t.oAuth2.scopes,n["x-scalar-client-id"]=t.oAuth2.clientId)),Array.isArray(n.scopes)&&(n.scopes=n.scopes.reduce(((e,t)=>({...e,[t]:""})),{})),n["x-defaultClientId"]&&(n["x-scalar-client-id"]=n["x-defaultClientId"])})):t&&("apiKey"===a.type&&t.apiKey?.token?a.value=t.apiKey.token:"http"===a.type&&("basic"===a.scheme&&t.http?.basic?(a.username=t.http.basic.username??"",a.password=t.http.basic.password??""):"bearer"===a.scheme&&t.http?.bearer&&(a.token=t.http.bearer.token??"")));const i=(0,o.T)(a,r.NV,!1);return i||k.push(`Security scheme ${e} is invalid.`),i})).filter((e=>!!e)),R={};var P;A.forEach((e=>{R[e.nameKey]=e.uid})),(P=d.paths??{},P?Object.keys(P).map((e=>String(e))):[]).forEach((e=>{const n=d?.paths?.[e];if(!n)return;const r=h.zT.array().parse(n.servers??[]);_.push(...r),Object.keys(n).filter(a.aO).forEach((a=>{const i=n[a],s=h.zT.array().parse(i.servers??[]);_.push(...s),i.tags?.forEach((e=>C.add(e)));const{security:c,...u}=i,p=(c??d.security??[]).map((e=>{const t=Object.keys(e);return t.length>1?t:t[0]})).filter(f.O),g=[t?.preferredSecurityScheme??[]].flat().filter((e=>Array.isArray(e)?p.some((t=>Array.isArray(t)&&t.length===e.length&&t.every(((t,n)=>t===e[n])))):p.includes(e))),v=p.length&&!l?O(p,g,R):[],b={...u,method:a,path:e,security:c,selectedSecuritySchemeUids:v,parameters:[...n?.parameters??[],...i.parameters??[]],servers:[...r,...s].map((e=>e.uid))};b.examples&&(console.warn("[@scalar/api-client] operation.examples is not a valid openapi property"),delete b.examples),c?.length&&(b.security=c.map((e=>{if(Object.keys(e).length){const[t]=Object.keys(e);return{[t]:e[t]}}return e})));const y=(0,o.T)(b,m.TF,!1);y?S.push(y):k.push(`${a} Request at ${n} is invalid.`)}))}));const D=(0,o.T)(d?.tags??[],g.c3.array(),!1)??[];D.forEach((e=>C.delete(e.name))),C.forEach((e=>e&&D.push(g.c3.parse({name:e}))));const M={};D.forEach((e=>{M[e.name]=e}));const $=new Set(D.map((e=>e.uid)));D.forEach((e=>{e["x-scalar-children"]?.forEach((t=>{const n=M[t.tagName].uid;e.children.push(n),$.delete(n)}))})),S.forEach((e=>{e.tags?.length?e.tags.forEach((t=>{M[t].children.push(e.uid)})):$.add(e.uid)}));const N=[];S.forEach((e=>{const t=(0,v.oE)(e,"Default Example");N.push(t),e.examples.push(t.uid)}));const I=(d.security??[]).map((e=>{const t=Object.keys(e);return t.length>1?t:t[0]})).filter(f.O),L=[t?.preferredSecurityScheme??[]].flat(),B=(I.length||L?.length)&&l?O(I,L,R):[],F=b.CV.parse({...d,watchMode:u,documentUrl:i,requests:S.map((e=>e.uid)),servers:_.map((e=>e.uid)),tags:D.map((e=>e.uid)),children:[...$],security:d.security??[{}],selectedServerUid:_?.[0]?.uid,selectedSecuritySchemeUids:B,components:{...d.components},securitySchemes:A.map((e=>e.uid))}),Q=performance.now();return console.log(`workspace: ${Math.round(Q-E)} ms`),{error:!1,servers:_,schema:d,requests:S,examples:N,collection:F,tags:D,securitySchemes:A}}function x(){if("undefined"!=typeof window&&"string"==typeof window?.location?.origin)return window.location.origin}},4811:(e,t,n)=>{"use strict";n.d(t,{iZ:()=>i});var r=n(4476);const a=r.z.enum(["path","query","header","cookie"]),o=r.z.enum(["matrix","simple","form","label","spaceDelimited","pipeDelimited","deepObject"]),i=r.z.object({in:a,name:r.z.string(),description:r.z.string().optional(),required:r.z.boolean().optional().default(!1),deprecated:r.z.boolean().optional().default(!1),schema:r.z.unknown().optional(),content:r.z.unknown().optional(),style:o.optional(),example:r.z.unknown().optional(),examples:r.z.record(r.z.string(),r.z.object({value:r.z.unknown(),summary:r.z.string().optional()})).optional()})},4848:(e,t,n)=>{"use strict";e.exports=n(9698)},4851:(e,t,n)=>{"use strict";const r=n(8560),a=n(5817),o=n(4126),{asciiLowercase:i,solelyContainsHTTPTokenCodePoints:s}=n(2661);e.exports=class{constructor(e){e=String(e);const t=a(e);if(null===t)throw new Error(`Could not parse MIME type string "${e}"`);this._type=t.type,this._subtype=t.subtype,this._parameters=new r(t.parameters)}static parse(e){try{return new this(e)}catch(e){return null}}get essence(){return`${this.type}/${this.subtype}`}get type(){return this._type}set type(e){if(0===(e=i(String(e))).length)throw new Error("Invalid type: must be a non-empty string");if(!s(e))throw new Error(`Invalid type ${e}: must contain only HTTP token code points`);this._type=e}get subtype(){return this._subtype}set subtype(e){if(0===(e=i(String(e))).length)throw new Error("Invalid subtype: must be a non-empty string");if(!s(e))throw new Error(`Invalid subtype ${e}: must contain only HTTP token code points`);this._subtype=e}get parameters(){return this._parameters}toString(){return o(this)}isJavaScript({prohibitParameters:e=!1}={}){switch(this._type){case"text":switch(this._subtype){case"ecmascript":case"javascript":case"javascript1.0":case"javascript1.1":case"javascript1.2":case"javascript1.3":case"javascript1.4":case"javascript1.5":case"jscript":case"livescript":case"x-ecmascript":case"x-javascript":return!e||0===this._parameters.size;default:return!1}case"application":switch(this._subtype){case"ecmascript":case"javascript":case"x-ecmascript":case"x-javascript":return!e||0===this._parameters.size;default:return!1}default:return!1}}isXML(){return"xml"===this._subtype&&("text"===this._type||"application"===this._type)||this._subtype.endsWith("+xml")}isHTML(){return"html"===this._subtype&&"text"===this._type}}},5018:(e,t,n)=>{"use strict";n.r(t),n.d(t,{boolean:()=>a,booleanish:()=>o,commaOrSpaceSeparated:()=>u,commaSeparated:()=>c,number:()=>s,overloadedBoolean:()=>i,spaceSeparated:()=>l});let r=0;const a=d(),o=d(),i=d(),s=d(),l=d(),c=d(),u=d();function d(){return 2**++r}},5147:(e,t,n)=>{"use strict";n.d(t,{E:()=>i});var r=n(8627),a=n(5018);const o=Object.keys(a);class i extends r.R{constructor(e,t,n,r){let i=-1;if(super(e,t),s(this,"space",r),"number"==typeof n)for(;++i<o.length;){const e=o[i];s(this,o[i],(n&a[e])===a[e])}}}function s(e,t,n){n&&(e[t]=n)}i.prototype.defined=!0},5148:(e,t,n)=>{"use strict";n.d(t,{Dk:()=>i,Hh:()=>s});var r=n(2741),a=n(1660),o=n(3762);const i=(e,t)=>t&&e!==t?e?`${e.trim()}/${t.trim()}`.replace(r.Q.MULTIPLE_SLASHES,"/"):t.trim():e.trim(),s=(e,t,n=new URLSearchParams,r=!1)=>{if(e&&(!(0,o.Kb)(e)||"undefined"!=typeof window)){const s=r?e:(0,o.Kb)(e)?i(window.location.origin,e):(0,a.T)(e),[l="",c]=s.split("?"),u=new URLSearchParams(c||""),[d="",p]=t.split("?"),h=new URLSearchParams(p||""),f=e===t?l:i(l,d),m=((...e)=>{const t={};e.forEach((e=>{const n=Array.from(e.keys());new Set(n).forEach((n=>{const r=e.getAll(n),a=r.length>1?r:r[0]??"";t[n]=a}))}));const n=new URLSearchParams;return Object.entries(t).forEach((([e,t])=>{Array.isArray(t)?t.forEach((t=>n.append(e,t))):n.append(e,t)})),n})(u,h,n).toString();return m?`${f}?${m}`:f}return t?i(e,t):""}},5182:(e,t,n)=>{"use strict";function r(e,t,n=!0,r){const a=t.safeParse(e);if(a.success||(console.group("Schema Error"+(r?` - ${r}`:"")),console.warn(JSON.stringify(a.error.format(),null,2)),console.log("Received: ",e),console.groupEnd()),n&&!a.success)throw new Error("Zod validation failure");return a.data}n.d(t,{T:()=>r})},5191:(e,t,n)=>{"use strict";n.d(t,{i:()=>s,v:()=>l});var r=n(4048),a=n(4748),o=n(953),i=n(7200);const s={};function l({requestMutators:e,collectionMutators:t,serverMutators:n,tagMutators:l,securitySchemeMutators:c,requestExampleMutators:u,workspaceMutators:d,workspaces:p}){const h=async(i,h,f={})=>{var m;const g=(0,o.ux)(i),v=await(0,a.be)(g,f);return v.error?(console.group("IMPORT ERRORS"),v.importWarnings.forEach((e=>console.warn(e))),void console.groupEnd()):(f.documentUrl&&"string"==typeof g&&(s[f.documentUrl]={hash:(0,r.n1)(g),schema:v.schema}),v.examples.forEach((e=>u.add(e))),v.requests.forEach((t=>e.add(t))),v.tags.forEach((e=>l.add(e))),v.servers.forEach((e=>n.add(e))),v.securitySchemes.forEach((e=>c.add(e))),t.add(v.collection),d.edit(h,"collections",[...(null==(m=p[h])?void 0:m.collections)??[],v.collection.uid]),v)};return{importSpecFile:h,importSpecFromUrl:async function(e,t,{proxyUrl:n,...a}={}){try{const o=await(0,r.tI)(e,n);return[null,await h(o,t,{documentUrl:e,...a})]}catch(e){return console.error("Failed to fetch spec from URL:",e),[(0,i.cQ)(e),null]}}}}},5192:(e,t,n)=>{"use strict";n(3940),t.parse=n(9563)},5336:(e,t,n)=>{"use strict";n.d(t,{A:()=>D});var r=n(641),a=n(953),o=n(33),i=n(1266),s=n(4048),l=n(3654),c=n(180);const u=(e,t=2)=>e>1e3?(e/1e3).toFixed(t)+"s":e+"ms";var d=n(2490),p=n(487),h=n(9351);const f={100:{color:"text-yellow"},200:{color:"text-green"},202:{color:"text-green"},300:{color:"text-blue"},304:{color:"text-blue"},400:{color:"text-red"},401:{color:"text-orange"},422:{color:"text-yellow"},423:{color:"text-purple"},505:{color:"text-orange"}},m=e=>f[e]||{color:"text-grey"};var g=n(8309);const v={class:"min-w-0"},b={class:"min-w-0 truncate text-c-1"},y=(0,r.pM)({__name:"AddressBarHistory",props:{operation:{},target:{}},setup(e){const{requestHistory:t,requestExampleMutators:n}=(0,g.hA)(),l=(0,h.rd)(),f=(0,a.KR)(t[0]??null),y=(0,r.EW)((()=>t.filter((t=>t.request.requestUid===e.operation.uid)).slice().reverse()));return(t,h)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(p.WF),{modelValue:f.value,"onUpdate:modelValue":h[0]||(h[0]=e=>f.value=e)},{default:(0,r.k6)((({open:f})=>[(0,r.bF)((0,a.R1)(i.Z5),{offset:0,resize:"",target:t.target},(0,r.eX)({default:(0,r.k6)((()=>{var e;return[null!=(e=y.value)&&e.length?((0,r.uX)(),(0,r.Wv)((0,a.R1)(p.go),{key:0,class:"addressbar-history-button z-context-plus mr-1 rounded-lg p-1.5 text-c-3 focus:text-c-1"},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(i.kv),{icon:"History",size:"sm",thickness:"2.25"}),h[1]||(h[1]=(0,r.Lk)("span",{class:"sr-only"},"Request History",-1))])),_:1})):(0,r.Q3)("",!0)]})),_:2},[f?{name:"floating",fn:(0,r.k6)((({width:t})=>[(0,r.bF)((0,a.R1)(p.wh),{class:"address-bg-states border-t custom-scroll max-h-[inherit] p-0.75 grid grid-cols-[44px,1fr,repeat(3,auto)] items-center",style:(0,o.Tr)({width:t})},{default:(0,r.k6)((()=>[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(y.value,((t,i)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(p.h7),{key:t.timestamp,class:"contents font-code text-sm *:rounded-none first:*:rounded-l last:*:rounded-r *:h-8 *:ui-active:bg-b-2 *:flex *:items-center *:cursor-pointer *:px-1.5 text-c-2 font-medium",value:i,onClick:r=>function(t){const r=l.currentRoute.value.params.workspace;e.operation.uid!==t.request.requestUid&&l.push({name:"request",params:{[d.TU.Workspace]:r,[d.TU.Request]:t.request.requestUid}}),n.set({...t.request})}(t)},{default:(0,r.k6)((()=>{var e;return[t.response.method?((0,r.uX)(),(0,r.Wv)(c.A,{key:0,class:"text-[11px]",method:t.response.method},null,8,["method"])):(0,r.Q3)("",!0),(0,r.Lk)("div",v,[(0,r.Lk)("div",b,(0,o.v_)(t.response.path),1)]),(0,r.Lk)("div",null,(0,o.v_)((0,a.R1)(u)(t.response.duration)),1),(0,r.Lk)("div",{class:(0,o.C4)([(0,a.R1)(m)(t.response.status).color])},(0,o.v_)(t.response.status),3),(0,r.Lk)("div",null,(0,o.v_)(null==(e=(0,a.R1)(s.Z7)[t.response.status])?void 0:e.name),1)]})),_:2},1032,["value","onClick"])))),128))])),_:2},1032,["style"]),(0,r.bF)((0,a.R1)(i.FI),{class:"-top-[--scalar-address-bar-height] rounded-lg"})])),key:"0"}:void 0]),1032,["target"])])),_:1},8,["modelValue"]))}});var w=n(3526);const O=(0,w.A)(y,[["__scopeId","data-v-a323ca24"]]);var k=n(2549),x=n(6584);const E=["id"],S={class:"m-auto flex flex-row items-center"},_={class:"addressbar-bg-states text-xxs p-0.75 group relative order-last flex w-full max-w-[calc(100dvw-24px)] flex-1 flex-row items-stretch rounded-lg lg:order-none lg:min-w-[580px] lg:max-w-[580px] xl:min-w-[720px] xl:max-w-[720px]"},C={class:"pointer-events-none absolute left-0 top-0 block h-full w-full overflow-hidden rounded-lg border"},T={class:"z-context-plus flex gap-1"},A={class:"codemirror-bg-switcher scroll-timeline-x scroll-timeline-x-hidden z-context-plus relative flex w-full"},R={"aria-hidden":"true",class:"inline-flex items-center gap-1"},P=(0,r.pM)({__name:"AddressBar",props:{collection:{},operation:{},server:{},environment:{},envVariables:{},workspace:{}},emits:["importCurl"],setup(e){const t=(0,r.Bi)(),{requestMutators:n,events:u}=(0,g.hA)(),{layout:d}=(0,x.g)(),p=(0,a.KR)(null);(0,r.wB)((()=>e.operation.path),(t=>{t&&(t=>{e.operation.path!==t&&n.edit(e.operation.uid,"path",t)})(t)}));const h=(0,a.KR)(100),f=(0,a.KR)(0),m=(0,a.KR)(!1),v=(0,a.KR)();function b(){m.value?h.value-=(h.value-15)/60:h.value-=f.value/20,h.value<=0&&(clearInterval(v.value),v.value=void 0,h.value=100,m.value=!1)}function y(t){n.edit(e.operation.uid,"method",t)}function w(){const{method:t}=e.operation;return s.NT[t].backgroundColor}function P(){m.value||(m.value=!0,u.executeRequest.emit({requestUid:e.operation.uid}))}function D(t){n.edit(e.operation.uid,"path",t)}return u.requestStatus.on((e=>{"start"===e&&(v.value||(m.value=!0,v.value=setInterval(b,20))),"stop"===e&&(f.value=h.value,m.value=!1),"abort"===e&&(clearInterval(v.value),v.value=void 0,h.value=100,m.value=!1)})),u.hotKeys.on((e=>{var t;null!=e&&e.focusAddressBar&&(null==(t=p.value)||t.focus()),null!=e&&e.executeRequest&&P()})),(e,n)=>{var s;return(0,r.uX)(),(0,r.CE)("div",{id:(0,a.R1)(t),class:"scalar-address-bar order-last h-[--scalar-address-bar-height] w-full [--scalar-address-bar-height:32px] lg:order-none lg:w-auto"},[(0,r.Lk)("div",S,[(0,r.Lk)("div",_,[(0,r.Lk)("div",C,[(0,r.Lk)("div",{class:(0,o.C4)(["bg-mix-transparent bg-mix-amount-90 z-context absolute left-0 top-0 h-full w-full",w()]),style:(0,o.Tr)({transform:`translate3d(-${h.value}%,0,0)`})},null,6)]),(0,r.Lk)("div",T,[(0,r.bF)(c.A,{isEditable:"modal"!==(0,a.R1)(d),isSquare:"",method:e.operation.method,teleport:"",onChange:y},null,8,["isEditable","method"])]),(0,r.Lk)("div",A,[e.collection.servers.length?((0,r.uX)(),(0,r.Wv)((0,a.R1)(k.A),{key:0,collection:e.collection,layout:"client",operation:e.operation,server:e.server,target:(0,a.R1)(t)},null,8,["collection","operation","server","target"])):(0,r.Q3)("",!0),n[1]||(n[1]=(0,r.Lk)("div",{class:"fade-left"},null,-1)),(0,r.bF)(l.A,{ref_key:"addressBarRef",ref:p,"aria-label":"Path",class:"min-w-fit outline-none",disableCloseBrackets:"",disabled:"modal"===(0,a.R1)(d),disableEnter:"",disableTabIndent:"",emitOnBlur:!1,envVariables:e.envVariables,environment:e.environment,importCurl:"",modelValue:e.operation.path,placeholder:null!=(s=e.server)&&s.uid&&e.collection.servers.includes(e.server.uid)?"":"Enter a URL or cURL command",server:"",workspace:e.workspace,onCurl:n[0]||(n[0]=t=>e.$emit("importCurl",t)),onSubmit:P,"onUpdate:modelValue":D},null,8,["disabled","envVariables","environment","modelValue","placeholder","workspace"]),n[2]||(n[2]=(0,r.Lk)("div",{class:"fade-right"},null,-1))]),(0,r.bF)(O,{operation:e.operation,target:(0,a.R1)(t)},null,8,["operation","target"]),(0,r.bF)((0,a.R1)(i.RO),{class:"z-context-plus relative h-auto shrink-0 overflow-hidden py-1 pl-2 pr-2.5 font-bold",disabled:m.value,onClick:P},{default:(0,r.k6)((()=>[(0,r.Lk)("span",R,[(0,r.bF)((0,a.R1)(i.kv),{class:"relative shrink-0 fill-current",icon:"Play",size:"xs"}),n[3]||(n[3]=(0,r.Lk)("span",{class:"text-xxs hidden lg:flex"},"Send",-1))]),n[4]||(n[4]=(0,r.Lk)("span",{class:"sr-only"}," Send Request ",-1))])),_:1},8,["disabled"])])])],8,E)}}}),D=(0,w.A)(P,[["__scopeId","data-v-b4d63e6c"]])},5338:(e,t,n)=>{"use strict";!function e(){if("undefined"!=typeof __REACT_DEVTOOLS_GLOBAL_HOOK__&&"function"==typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE)try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(e){console.error(e)}}(),e.exports=n(1247)},5362:(e,t,n)=>{"use strict";function r(e){return e?.find((e=>e.isEntrypoint))}n.d(t,{z:()=>r})},5411:(e,t,n)=>{"use strict";n.d(t,{A:()=>u});var r=n(641),a=n(3751),o=n(953),i=n(1266);const s={class:"flex flex-col rounded min-h-20 relative"},l={class:"flex gap-2"},c={class:"flex flex-1 max-h-8"},u=(0,r.pM)({__name:"CommandActionForm",props:{loading:{},disabled:{type:Boolean,default:!1}},emits:["submit","cancel","back"],setup:e=>(e,t)=>((0,r.uX)(),(0,r.CE)("form",{class:"flex w-full flex-col gap-3",onKeydown:t[0]||(t[0]=(0,a.jR)((0,a.D$)((()=>{}),["stop"]),["enter"])),onSubmit:t[1]||(t[1]=(0,a.D$)((t=>e.$emit("submit")),["prevent","stop"]))},[(0,r.Lk)("div",s,[(0,r.RG)(e.$slots,"default")]),(0,r.Lk)("div",l,[(0,r.Lk)("div",c,[(0,r.RG)(e.$slots,"options")]),(0,r.bF)((0,o.R1)(i.RO),{class:"max-h-8 text-xs p-0 px-3",disabled:e.disabled,loading:e.loading,type:"submit"},{default:(0,r.k6)((()=>[(0,r.RG)(e.$slots,"submit",{},(()=>[t[2]||(t[2]=(0,r.eW)("Continue"))]))])),_:3},8,["disabled","loading"])])],32))})},5633:(e,t,n)=>{"use strict";n.d(t,{H:()=>l});var r=n(3687),a=n(5362),o=n(4279),i=n(1507),s=n(8007);async function l(e,t){const n=[];if(t?.filesystem?.find((t=>t.filename===e)))return{specification:(0,a.z)(t.filesystem)?.specification,filesystem:t.filesystem,errors:n};const c=t?.plugins?.find((t=>t.check(e)));let u;if(c)try{u=(0,s.S)(await c.get(e))}catch(a){if(t?.throwOnError)throw new Error(r.Sr.EXTERNAL_REFERENCE_NOT_FOUND.replace("%s",e));return n.push({code:"EXTERNAL_REFERENCE_NOT_FOUND",message:r.Sr.EXTERNAL_REFERENCE_NOT_FOUND.replace("%s",e)}),{specification:null,filesystem:[],errors:n}}else u=(0,s.S)(e);if(void 0===u){if(t?.throwOnError)throw new Error("No content to load");return n.push({code:"NO_CONTENT",message:r.Sr.NO_CONTENT}),{specification:null,filesystem:[],errors:n}}let d=(0,i.Y)(u,{filename:t?.filename??null});const p=(t?.filename?d.find((e=>e.filename===t?.filename)):(0,a.z)(d)).references??(0,o.o)(u);if(0===p.length)return{specification:(0,a.z)(d)?.specification,filesystem:d,errors:n};for(const r of p){const a=t?.plugins?.find((e=>e.check(r)));if(!a)continue;const o=a.check(r)&&a.resolvePath?a.resolvePath(e,r):r;if(d.find((e=>e.filename===r)))continue;const{filesystem:i,errors:s}=await l(o,{...t,filename:r});n.push(...s),d=[...d,...i.map((e=>({...e,isEntrypoint:!1})))]}return{specification:(0,a.z)(d)?.specification,filesystem:d,errors:n}}},5670:(e,t,n)=>{"use strict";n.d(t,{F:()=>r});const r=(0,n(1259).v)({properties:{xLinkActuate:null,xLinkArcRole:null,xLinkHref:null,xLinkRole:null,xLinkShow:null,xLinkTitle:null,xLinkType:null},space:"xlink",transform:(e,t)=>"xlink:"+t.slice(5).toLowerCase()})},5785:(e,t,n)=>{"use strict";n.d(t,{AH:()=>k,I4:()=>y,NV:()=>x});var r=n(4476),a=n(2071);const o=r.z.object({description:r.z.string().optional()}),i=r.z.object({uid:a.B.brand(),nameKey:r.z.string().optional().default("")}),s=o.extend({type:r.z.literal("apiKey"),name:r.z.string().optional().default(""),in:r.z.enum(["query","header","cookie"]).optional().default("header")}),l=r.z.object({value:r.z.string().default("")}),c=s.merge(i).merge(l),u=o.extend({type:r.z.literal("http"),scheme:r.z.string().toLowerCase().pipe(r.z.enum(["basic","bearer"])).optional().default("basic"),bearerFormat:r.z.union([r.z.literal("JWT"),r.z.string()]).optional().default("JWT")}),d=r.z.object({username:r.z.string().default(""),password:r.z.string().default(""),token:r.z.string().default("")}),p=u.merge(i).merge(d),h=o.extend({type:r.z.literal("openIdConnect"),openIdConnectUrl:r.z.string().optional().default("")}),f=h.merge(i),m=r.z.string().default(""),g=r.z.string().default(""),v=r.z.object({refreshUrl:r.z.string().optional().default(""),scopes:r.z.record(r.z.string(),r.z.string().optional().default("")).optional().default({}),selectedScopes:r.z.array(r.z.string()).optional().default([]),"x-scalar-client-id":r.z.string().optional().default(""),token:r.z.string().default("")}),b="undefined"!=typeof window?window.location.origin+window.location.pathname:"",y=["SHA-256","plain","no"],w=o.extend({type:r.z.literal("oauth2"),flows:r.z.object({implicit:v.extend({type:r.z.literal("implicit"),authorizationUrl:m,"x-scalar-redirect-uri":r.z.string().optional().default(b)}),password:v.extend({type:r.z.literal("password"),tokenUrl:g,clientSecret:r.z.string().default(""),username:r.z.string().default(""),password:r.z.string().default("")}),clientCredentials:v.extend({type:r.z.literal("clientCredentials"),tokenUrl:g,clientSecret:r.z.string().default("")}),authorizationCode:v.extend({type:r.z.literal("authorizationCode"),authorizationUrl:m,"x-usePkce":r.z.enum(y).optional().default("no"),"x-scalar-redirect-uri":r.z.string().optional().default(b),tokenUrl:g,clientSecret:r.z.string().default("")})}).partial().default({implicit:{type:"implicit",authorizationUrl:"http://localhost:8080"}})}),O=w.merge(i),k=r.z.record(r.z.string(),r.z.array(r.z.string()).optional().default([])),x=(r.z.union([s,u,w,h]),r.z.union([c,p,f,O]))},5817:(e,t,n)=>{"use strict";const{removeLeadingAndTrailingHTTPWhitespace:r,removeTrailingHTTPWhitespace:a,isHTTPWhitespaceChar:o,solelyContainsHTTPTokenCodePoints:i,soleyContainsHTTPQuotedStringTokenCodePoints:s,asciiLowercase:l,collectAnHTTPQuotedString:c}=n(2661);e.exports=e=>{e=r(e);let t=0,n="";for(;t<e.length&&"/"!==e[t];)n+=e[t],++t;if(0===n.length||!i(n))return null;if(t>=e.length)return null;++t;let u="";for(;t<e.length&&";"!==e[t];)u+=e[t],++t;if(u=a(u),0===u.length||!i(u))return null;const d={type:l(n),subtype:l(u),parameters:new Map};for(;t<e.length;){for(++t;o(e[t]);)++t;let n="";for(;t<e.length&&";"!==e[t]&&"="!==e[t];)n+=e[t],++t;if(n=l(n),t<e.length){if(";"===e[t])continue;++t}let r=null;if('"'===e[t])for([r,t]=c(e,t);t<e.length&&";"!==e[t];)++t;else{for(r="";t<e.length&&";"!==e[t];)r+=e[t],++t;if(r=a(r),""===r)continue}n.length>0&&i(n)&&s(r)&&!d.parameters.has(n)&&d.parameters.set(n,r)}return d}},5841:(e,t,n)=>{"use strict";n.d(t,{G:()=>l});var r=n(953),a=n(641),o=n(4476);const i=(0,r.KR)("dark"),s=o.z.enum(["dark","light","system"]).optional().catch(void 0);function l(e={}){var t;const{initialColorMode:n="system",overrideColorMode:r}=e;function o(e){var t;i.value=e,!(typeof window>"u")&&(null==(t=null==window?void 0:window.localStorage)||t.setItem("colorMode",i.value))}function l(){var e;return typeof window>"u"?"light":"function"!=typeof(null==window?void 0:window.matchMedia)||null!=(e=null==window?void 0:window.matchMedia("(prefers-color-scheme: dark)"))&&e.matches?"dark":"light"}const c=(0,a.EW)({get:()=>"system"===i.value?l():i.value,set:o}),u=(0,a.EW)({get:()=>"dark"===c.value,set:e=>o(e?"dark":"light")});function d(e){typeof document>"u"||typeof window>"u"||("dark"===(r??("system"===e?l():e))?(document.body.classList.add("dark-mode"),document.body.classList.remove("light-mode")):(document.body.classList.add("light-mode"),document.body.classList.remove("dark-mode")))}const p=s.parse(typeof window<"u"?null==(t=null==window?void 0:window.localStorage)?void 0:t.getItem("colorMode"):"system");return i.value=p??n,(0,a.wB)(i,d,{immediate:!0}),(0,a.sV)((()=>{if(typeof window<"u"&&"function"==typeof(null==window?void 0:window.matchMedia)){const e=window.matchMedia("(prefers-color-scheme: dark)"),t=()=>"system"===i.value&&d("system");e.addEventListener("change",t),(0,a.hi)((()=>{e.removeEventListener("change",t)}))}})),{colorMode:(0,a.EW)({get:()=>i.value,set:o}),darkLightMode:c,isDarkMode:u,toggleColorMode:function(){var e;i.value="dark"===c.value?"light":"dark",!(typeof window>"u")&&(null==(e=null==window?void 0:window.localStorage)||e.setItem("colorMode",i.value))},setColorMode:o,getSystemModePreference:l}}},5868:(e,t,n)=>{"use strict";function r(e){if(!e)return e;const t={...e};return Object.keys(t).forEach((e=>{const n=function(e){if("string"==typeof e)return e.replace(/;.*$/,"").replace(/\/.+\+/,"/").trim()}(e);void 0!==n&&(t[n]=t[e],e!==n&&delete t[e])})),t}n.d(t,{L:()=>r})},5881:(e,t,n)=>{"use strict";n.d(t,{A:()=>i});var r=n(641),a=n(33);const o=(0,r.pM)({__name:"ScalarAsciiArt",props:{art:{},animate:{type:Boolean}},setup(e){const t=e,n=(0,r.EW)((()=>t.art.split("\n"))),o=(e,t)=>{var r,a,o,i;return{animationDuration:100*e+"ms, 500ms",animationTimingFunction:`steps(${e}), step-end`,animationDelay:100*t+"ms, 0ms",animationIterationCount:"1, "+100*(((null==(r=n.value)?void 0:r.length)??0)+((null==(i=null==(o=n.value)?void 0:o[(null==(a=n.value)?void 0:a.length)-1])?void 0:i.length)??0)+5)/500}};return(e,t)=>((0,r.uX)(),(0,r.CE)("div",{class:(0,a.C4)(["ascii-art font-code flex flex-col items-start text-[6px] leading-[7px]",{"ascii-art-animate":e.animate}])},[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(n.value,((e,t)=>((0,r.uX)(),(0,r.CE)("span",{key:t,class:"inline-block",style:(0,a.Tr)({width:`calc(${e.length+1}ch)`})},[(0,r.Lk)("span",{class:"inline-block whitespace-pre overflow-hidden",style:(0,a.Tr)(o(e.length,t))},(0,a.v_)(e),5)],4)))),128))],2))}}),i=(0,n(3526).A)(o,[["__scopeId","data-v-0018c948"]])},5929:(e,t,n)=>{"use strict";function r(e){return Array.isArray?Array.isArray(e):"[object Array]"===c(e)}function a(e){return"string"==typeof e}function o(e){return"number"==typeof e}function i(e){return"object"==typeof e}function s(e){return null!=e}function l(e){return!e.trim().length}function c(e){return null==e?void 0===e?"[object Undefined]":"[object Null]":Object.prototype.toString.call(e)}n.d(t,{A:()=>U});const u=Object.prototype.hasOwnProperty;class d{constructor(e){this._keys=[],this._keyMap={};let t=0;e.forEach((e=>{let n=p(e);this._keys.push(n),this._keyMap[n.id]=n,t+=n.weight})),this._keys.forEach((e=>{e.weight/=t}))}get(e){return this._keyMap[e]}keys(){return this._keys}toJSON(){return JSON.stringify(this._keys)}}function p(e){let t=null,n=null,o=null,i=1,s=null;if(a(e)||r(e))o=e,t=h(e),n=f(e);else{if(!u.call(e,"name"))throw new Error("Missing name property in key");const r=e.name;if(o=r,u.call(e,"weight")&&(i=e.weight,i<=0))throw new Error((e=>`Property 'weight' in key '${e}' must be a positive integer`)(r));t=h(r),n=f(r),s=e.getFn}return{path:t,id:n,weight:i,src:o,getFn:s}}function h(e){return r(e)?e:e.split(".")}function f(e){return r(e)?e.join("."):e}var m={isCaseSensitive:!1,ignoreDiacritics:!1,includeScore:!1,keys:[],shouldSort:!0,sortFn:(e,t)=>e.score===t.score?e.idx<t.idx?-1:1:e.score<t.score?-1:1,includeMatches:!1,findAllMatches:!1,minMatchCharLength:1,location:0,threshold:.6,distance:100,useExtendedSearch:!1,getFn:function(e,t){let n=[],l=!1;const u=(e,t,d)=>{if(s(e))if(t[d]){const p=e[t[d]];if(!s(p))return;if(d===t.length-1&&(a(p)||o(p)||function(e){return!0===e||!1===e||function(e){return i(e)&&null!==e}(e)&&"[object Boolean]"==c(e)}(p)))n.push(function(e){return null==e?"":function(e){if("string"==typeof e)return e;let t=e+"";return"0"==t&&1/e==-1/0?"-0":t}(e)}(p));else if(r(p)){l=!0;for(let e=0,n=p.length;e<n;e+=1)u(p[e],t,d+1)}else t.length&&u(p,t,d+1)}else n.push(e)};return u(e,a(t)?t.split("."):t,0),l?n:n[0]},ignoreLocation:!1,ignoreFieldNorm:!1,fieldNormWeight:1};const g=/[^ ]+/g;class v{constructor({getFn:e=m.getFn,fieldNormWeight:t=m.fieldNormWeight}={}){this.norm=function(e=1,t=3){const n=new Map,r=Math.pow(10,t);return{get(t){const a=t.match(g).length;if(n.has(a))return n.get(a);const o=1/Math.pow(a,.5*e),i=parseFloat(Math.round(o*r)/r);return n.set(a,i),i},clear(){n.clear()}}}(t,3),this.getFn=e,this.isCreated=!1,this.setIndexRecords()}setSources(e=[]){this.docs=e}setIndexRecords(e=[]){this.records=e}setKeys(e=[]){this.keys=e,this._keysMap={},e.forEach(((e,t)=>{this._keysMap[e.id]=t}))}create(){!this.isCreated&&this.docs.length&&(this.isCreated=!0,a(this.docs[0])?this.docs.forEach(((e,t)=>{this._addString(e,t)})):this.docs.forEach(((e,t)=>{this._addObject(e,t)})),this.norm.clear())}add(e){const t=this.size();a(e)?this._addString(e,t):this._addObject(e,t)}removeAt(e){this.records.splice(e,1);for(let t=e,n=this.size();t<n;t+=1)this.records[t].i-=1}getValueForItemAtKeyId(e,t){return e[this._keysMap[t]]}size(){return this.records.length}_addString(e,t){if(!s(e)||l(e))return;let n={v:e,i:t,n:this.norm.get(e)};this.records.push(n)}_addObject(e,t){let n={i:t,$:{}};this.keys.forEach(((t,o)=>{let i=t.getFn?t.getFn(e):this.getFn(e,t.path);if(s(i))if(r(i)){let e=[];const t=[{nestedArrIndex:-1,value:i}];for(;t.length;){const{nestedArrIndex:n,value:o}=t.pop();if(s(o))if(a(o)&&!l(o)){let t={v:o,i:n,n:this.norm.get(o)};e.push(t)}else r(o)&&o.forEach(((e,n)=>{t.push({nestedArrIndex:n,value:e})}))}n.$[o]=e}else if(a(i)&&!l(i)){let e={v:i,n:this.norm.get(i)};n.$[o]=e}})),this.records.push(n)}toJSON(){return{keys:this.keys,records:this.records}}}function b(e,t,{getFn:n=m.getFn,fieldNormWeight:r=m.fieldNormWeight}={}){const a=new v({getFn:n,fieldNormWeight:r});return a.setKeys(e.map(p)),a.setSources(t),a.create(),a}function y(e,{errors:t=0,currentLocation:n=0,expectedLocation:r=0,distance:a=m.distance,ignoreLocation:o=m.ignoreLocation}={}){const i=t/e.length;if(o)return i;const s=Math.abs(r-n);return a?i+s/a:s?1:i}const w=32;function O(e){let t={};for(let n=0,r=e.length;n<r;n+=1){const a=e.charAt(n);t[a]=(t[a]||0)|1<<r-n-1}return t}const k=String.prototype.normalize?e=>e.normalize("NFD").replace(/[\u0300-\u036F\u0483-\u0489\u0591-\u05BD\u05BF\u05C1\u05C2\u05C4\u05C5\u05C7\u0610-\u061A\u064B-\u065F\u0670\u06D6-\u06DC\u06DF-\u06E4\u06E7\u06E8\u06EA-\u06ED\u0711\u0730-\u074A\u07A6-\u07B0\u07EB-\u07F3\u07FD\u0816-\u0819\u081B-\u0823\u0825-\u0827\u0829-\u082D\u0859-\u085B\u08D3-\u08E1\u08E3-\u0903\u093A-\u093C\u093E-\u094F\u0951-\u0957\u0962\u0963\u0981-\u0983\u09BC\u09BE-\u09C4\u09C7\u09C8\u09CB-\u09CD\u09D7\u09E2\u09E3\u09FE\u0A01-\u0A03\u0A3C\u0A3E-\u0A42\u0A47\u0A48\u0A4B-\u0A4D\u0A51\u0A70\u0A71\u0A75\u0A81-\u0A83\u0ABC\u0ABE-\u0AC5\u0AC7-\u0AC9\u0ACB-\u0ACD\u0AE2\u0AE3\u0AFA-\u0AFF\u0B01-\u0B03\u0B3C\u0B3E-\u0B44\u0B47\u0B48\u0B4B-\u0B4D\u0B56\u0B57\u0B62\u0B63\u0B82\u0BBE-\u0BC2\u0BC6-\u0BC8\u0BCA-\u0BCD\u0BD7\u0C00-\u0C04\u0C3E-\u0C44\u0C46-\u0C48\u0C4A-\u0C4D\u0C55\u0C56\u0C62\u0C63\u0C81-\u0C83\u0CBC\u0CBE-\u0CC4\u0CC6-\u0CC8\u0CCA-\u0CCD\u0CD5\u0CD6\u0CE2\u0CE3\u0D00-\u0D03\u0D3B\u0D3C\u0D3E-\u0D44\u0D46-\u0D48\u0D4A-\u0D4D\u0D57\u0D62\u0D63\u0D82\u0D83\u0DCA\u0DCF-\u0DD4\u0DD6\u0DD8-\u0DDF\u0DF2\u0DF3\u0E31\u0E34-\u0E3A\u0E47-\u0E4E\u0EB1\u0EB4-\u0EB9\u0EBB\u0EBC\u0EC8-\u0ECD\u0F18\u0F19\u0F35\u0F37\u0F39\u0F3E\u0F3F\u0F71-\u0F84\u0F86\u0F87\u0F8D-\u0F97\u0F99-\u0FBC\u0FC6\u102B-\u103E\u1056-\u1059\u105E-\u1060\u1062-\u1064\u1067-\u106D\u1071-\u1074\u1082-\u108D\u108F\u109A-\u109D\u135D-\u135F\u1712-\u1714\u1732-\u1734\u1752\u1753\u1772\u1773\u17B4-\u17D3\u17DD\u180B-\u180D\u1885\u1886\u18A9\u1920-\u192B\u1930-\u193B\u1A17-\u1A1B\u1A55-\u1A5E\u1A60-\u1A7C\u1A7F\u1AB0-\u1ABE\u1B00-\u1B04\u1B34-\u1B44\u1B6B-\u1B73\u1B80-\u1B82\u1BA1-\u1BAD\u1BE6-\u1BF3\u1C24-\u1C37\u1CD0-\u1CD2\u1CD4-\u1CE8\u1CED\u1CF2-\u1CF4\u1CF7-\u1CF9\u1DC0-\u1DF9\u1DFB-\u1DFF\u20D0-\u20F0\u2CEF-\u2CF1\u2D7F\u2DE0-\u2DFF\u302A-\u302F\u3099\u309A\uA66F-\uA672\uA674-\uA67D\uA69E\uA69F\uA6F0\uA6F1\uA802\uA806\uA80B\uA823-\uA827\uA880\uA881\uA8B4-\uA8C5\uA8E0-\uA8F1\uA8FF\uA926-\uA92D\uA947-\uA953\uA980-\uA983\uA9B3-\uA9C0\uA9E5\uAA29-\uAA36\uAA43\uAA4C\uAA4D\uAA7B-\uAA7D\uAAB0\uAAB2-\uAAB4\uAAB7\uAAB8\uAABE\uAABF\uAAC1\uAAEB-\uAAEF\uAAF5\uAAF6\uABE3-\uABEA\uABEC\uABED\uFB1E\uFE00-\uFE0F\uFE20-\uFE2F]/g,""):e=>e;class x{constructor(e,{location:t=m.location,threshold:n=m.threshold,distance:r=m.distance,includeMatches:a=m.includeMatches,findAllMatches:o=m.findAllMatches,minMatchCharLength:i=m.minMatchCharLength,isCaseSensitive:s=m.isCaseSensitive,ignoreDiacritics:l=m.ignoreDiacritics,ignoreLocation:c=m.ignoreLocation}={}){if(this.options={location:t,threshold:n,distance:r,includeMatches:a,findAllMatches:o,minMatchCharLength:i,isCaseSensitive:s,ignoreDiacritics:l,ignoreLocation:c},e=s?e:e.toLowerCase(),e=l?k(e):e,this.pattern=e,this.chunks=[],!this.pattern.length)return;const u=(e,t)=>{this.chunks.push({pattern:e,alphabet:O(e),startIndex:t})},d=this.pattern.length;if(d>w){let e=0;const t=d%w,n=d-t;for(;e<n;)u(this.pattern.substr(e,w),e),e+=w;if(t){const e=d-w;u(this.pattern.substr(e),e)}}else u(this.pattern,0)}searchIn(e){const{isCaseSensitive:t,ignoreDiacritics:n,includeMatches:r}=this.options;if(e=t?e:e.toLowerCase(),e=n?k(e):e,this.pattern===e){let t={isMatch:!0,score:0};return r&&(t.indices=[[0,e.length-1]]),t}const{location:a,distance:o,threshold:i,findAllMatches:s,minMatchCharLength:l,ignoreLocation:c}=this.options;let u=[],d=0,p=!1;this.chunks.forEach((({pattern:t,alphabet:n,startIndex:h})=>{const{isMatch:f,score:g,indices:v}=function(e,t,n,{location:r=m.location,distance:a=m.distance,threshold:o=m.threshold,findAllMatches:i=m.findAllMatches,minMatchCharLength:s=m.minMatchCharLength,includeMatches:l=m.includeMatches,ignoreLocation:c=m.ignoreLocation}={}){if(t.length>w)throw new Error("Pattern length exceeds max of 32.");const u=t.length,d=e.length,p=Math.max(0,Math.min(r,d));let h=o,f=p;const g=s>1||l,v=g?Array(d):[];let b;for(;(b=e.indexOf(t,f))>-1;){let e=y(t,{currentLocation:b,expectedLocation:p,distance:a,ignoreLocation:c});if(h=Math.min(e,h),f=b+u,g){let e=0;for(;e<u;)v[b+e]=1,e+=1}}f=-1;let O=[],k=1,x=u+d;const E=1<<u-1;for(let r=0;r<u;r+=1){let o=0,s=x;for(;o<s;)y(t,{errors:r,currentLocation:p+s,expectedLocation:p,distance:a,ignoreLocation:c})<=h?o=s:x=s,s=Math.floor((x-o)/2+o);x=s;let l=Math.max(1,p-s+1),m=i?d:Math.min(p+s,d)+u,b=Array(m+2);b[m+1]=(1<<r)-1;for(let o=m;o>=l;o-=1){let i=o-1,s=n[e.charAt(i)];if(g&&(v[i]=+!!s),b[o]=(b[o+1]<<1|1)&s,r&&(b[o]|=(O[o+1]|O[o])<<1|1|O[o+1]),b[o]&E&&(k=y(t,{errors:r,currentLocation:i,expectedLocation:p,distance:a,ignoreLocation:c}),k<=h)){if(h=k,f=i,f<=p)break;l=Math.max(1,2*p-f)}}if(y(t,{errors:r+1,currentLocation:p,expectedLocation:p,distance:a,ignoreLocation:c})>h)break;O=b}const S={isMatch:f>=0,score:Math.max(.001,k)};if(g){const e=function(e=[],t=m.minMatchCharLength){let n=[],r=-1,a=-1,o=0;for(let i=e.length;o<i;o+=1){let i=e[o];i&&-1===r?r=o:i||-1===r||(a=o-1,a-r+1>=t&&n.push([r,a]),r=-1)}return e[o-1]&&o-r>=t&&n.push([r,o-1]),n}(v,s);e.length?l&&(S.indices=e):S.isMatch=!1}return S}(e,t,n,{location:a+h,distance:o,threshold:i,findAllMatches:s,minMatchCharLength:l,includeMatches:r,ignoreLocation:c});f&&(p=!0),d+=g,f&&v&&(u=[...u,...v])}));let h={isMatch:p,score:p?d/this.chunks.length:1};return p&&r&&(h.indices=u),h}}class E{constructor(e){this.pattern=e}static isMultiMatch(e){return S(e,this.multiRegex)}static isSingleMatch(e){return S(e,this.singleRegex)}search(){}}function S(e,t){const n=e.match(t);return n?n[1]:null}class _ extends E{constructor(e,{location:t=m.location,threshold:n=m.threshold,distance:r=m.distance,includeMatches:a=m.includeMatches,findAllMatches:o=m.findAllMatches,minMatchCharLength:i=m.minMatchCharLength,isCaseSensitive:s=m.isCaseSensitive,ignoreDiacritics:l=m.ignoreDiacritics,ignoreLocation:c=m.ignoreLocation}={}){super(e),this._bitapSearch=new x(e,{location:t,threshold:n,distance:r,includeMatches:a,findAllMatches:o,minMatchCharLength:i,isCaseSensitive:s,ignoreDiacritics:l,ignoreLocation:c})}static get type(){return"fuzzy"}static get multiRegex(){return/^"(.*)"$/}static get singleRegex(){return/^(.*)$/}search(e){return this._bitapSearch.searchIn(e)}}class C extends E{constructor(e){super(e)}static get type(){return"include"}static get multiRegex(){return/^'"(.*)"$/}static get singleRegex(){return/^'(.*)$/}search(e){let t,n=0;const r=[],a=this.pattern.length;for(;(t=e.indexOf(this.pattern,n))>-1;)n=t+a,r.push([t,n-1]);const o=!!r.length;return{isMatch:o,score:o?0:1,indices:r}}}const T=[class extends E{constructor(e){super(e)}static get type(){return"exact"}static get multiRegex(){return/^="(.*)"$/}static get singleRegex(){return/^=(.*)$/}search(e){const t=e===this.pattern;return{isMatch:t,score:t?0:1,indices:[0,this.pattern.length-1]}}},C,class extends E{constructor(e){super(e)}static get type(){return"prefix-exact"}static get multiRegex(){return/^\^"(.*)"$/}static get singleRegex(){return/^\^(.*)$/}search(e){const t=e.startsWith(this.pattern);return{isMatch:t,score:t?0:1,indices:[0,this.pattern.length-1]}}},class extends E{constructor(e){super(e)}static get type(){return"inverse-prefix-exact"}static get multiRegex(){return/^!\^"(.*)"$/}static get singleRegex(){return/^!\^(.*)$/}search(e){const t=!e.startsWith(this.pattern);return{isMatch:t,score:t?0:1,indices:[0,e.length-1]}}},class extends E{constructor(e){super(e)}static get type(){return"inverse-suffix-exact"}static get multiRegex(){return/^!"(.*)"\$$/}static get singleRegex(){return/^!(.*)\$$/}search(e){const t=!e.endsWith(this.pattern);return{isMatch:t,score:t?0:1,indices:[0,e.length-1]}}},class extends E{constructor(e){super(e)}static get type(){return"suffix-exact"}static get multiRegex(){return/^"(.*)"\$$/}static get singleRegex(){return/^(.*)\$$/}search(e){const t=e.endsWith(this.pattern);return{isMatch:t,score:t?0:1,indices:[e.length-this.pattern.length,e.length-1]}}},class extends E{constructor(e){super(e)}static get type(){return"inverse-exact"}static get multiRegex(){return/^!"(.*)"$/}static get singleRegex(){return/^!(.*)$/}search(e){const t=-1===e.indexOf(this.pattern);return{isMatch:t,score:t?0:1,indices:[0,e.length-1]}}},_],A=T.length,R=/ +(?=(?:[^\"]*\"[^\"]*\")*[^\"]*$)/,P=new Set([_.type,C.type]);const D=[];function M(e,t){for(let n=0,r=D.length;n<r;n+=1){let r=D[n];if(r.condition(e,t))return new r(e,t)}return new x(e,t)}const $="$and",N="$path",I=e=>!(!e[$]&&!e.$or),L=e=>({[$]:Object.keys(e).map((t=>({[t]:e[t]})))});function B(e,t,{auto:n=!0}={}){const o=e=>{let s=Object.keys(e);const l=(e=>!!e[N])(e);if(!l&&s.length>1&&!I(e))return o(L(e));if((e=>!r(e)&&i(e)&&!I(e))(e)){const r=l?e[N]:s[0],o=l?e.$val:e[r];if(!a(o))throw new Error((e=>`Invalid value for key ${e}`)(r));const i={keyId:f(r),pattern:o};return n&&(i.searcher=M(o,t)),i}let c={children:[],operator:s[0]};return s.forEach((t=>{const n=e[t];r(n)&&n.forEach((e=>{c.children.push(o(e))}))})),c};return I(e)||(e=L(e)),o(e)}function F(e,t){const n=e.matches;t.matches=[],s(n)&&n.forEach((e=>{if(!s(e.indices)||!e.indices.length)return;const{indices:n,value:r}=e;let a={indices:n,value:r};e.key&&(a.key=e.key.src),e.idx>-1&&(a.refIndex=e.idx),t.matches.push(a)}))}function Q(e,t){t.score=e.score}class U{constructor(e,t={},n){this.options={...m,...t},this.options.useExtendedSearch,this._keyStore=new d(this.options.keys),this.setCollection(e,n)}setCollection(e,t){if(this._docs=e,t&&!(t instanceof v))throw new Error("Incorrect 'index' type");this._myIndex=t||b(this.options.keys,this._docs,{getFn:this.options.getFn,fieldNormWeight:this.options.fieldNormWeight})}add(e){s(e)&&(this._docs.push(e),this._myIndex.add(e))}remove(e=()=>!1){const t=[];for(let n=0,r=this._docs.length;n<r;n+=1){const a=this._docs[n];e(a,n)&&(this.removeAt(n),n-=1,r-=1,t.push(a))}return t}removeAt(e){this._docs.splice(e,1),this._myIndex.removeAt(e)}getIndex(){return this._myIndex}search(e,{limit:t=-1}={}){const{includeMatches:n,includeScore:r,shouldSort:i,sortFn:s,ignoreFieldNorm:l}=this.options;let c=a(e)?a(this._docs[0])?this._searchStringList(e):this._searchObjectList(e):this._searchLogical(e);return function(e,{ignoreFieldNorm:t=m.ignoreFieldNorm}){e.forEach((e=>{let n=1;e.matches.forEach((({key:e,norm:r,score:a})=>{const o=e?e.weight:null;n*=Math.pow(0===a&&o?Number.EPSILON:a,(o||1)*(t?1:r))})),e.score=n}))}(c,{ignoreFieldNorm:l}),i&&c.sort(s),o(t)&&t>-1&&(c=c.slice(0,t)),function(e,t,{includeMatches:n=m.includeMatches,includeScore:r=m.includeScore}={}){const a=[];return n&&a.push(F),r&&a.push(Q),e.map((e=>{const{idx:n}=e,r={item:t[n],refIndex:n};return a.length&&a.forEach((t=>{t(e,r)})),r}))}(c,this._docs,{includeMatches:n,includeScore:r})}_searchStringList(e){const t=M(e,this.options),{records:n}=this._myIndex,r=[];return n.forEach((({v:e,i:n,n:a})=>{if(!s(e))return;const{isMatch:o,score:i,indices:l}=t.searchIn(e);o&&r.push({item:e,idx:n,matches:[{score:i,value:e,norm:a,indices:l}]})})),r}_searchLogical(e){const t=B(e,this.options),n=(e,t,r)=>{if(!e.children){const{keyId:n,searcher:a}=e,o=this._findMatches({key:this._keyStore.get(n),value:this._myIndex.getValueForItemAtKeyId(t,n),searcher:a});return o&&o.length?[{idx:r,item:t,matches:o}]:[]}const a=[];for(let o=0,i=e.children.length;o<i;o+=1){const i=e.children[o],s=n(i,t,r);if(s.length)a.push(...s);else if(e.operator===$)return[]}return a},r=this._myIndex.records,a={},o=[];return r.forEach((({$:e,i:r})=>{if(s(e)){let i=n(t,e,r);i.length&&(a[r]||(a[r]={idx:r,item:e,matches:[]},o.push(a[r])),i.forEach((({matches:e})=>{a[r].matches.push(...e)})))}})),o}_searchObjectList(e){const t=M(e,this.options),{keys:n,records:r}=this._myIndex,a=[];return r.forEach((({$:e,i:r})=>{if(!s(e))return;let o=[];n.forEach(((n,r)=>{o.push(...this._findMatches({key:n,value:e[r],searcher:t}))})),o.length&&a.push({idx:r,item:e,matches:o})})),a}_findMatches({key:e,value:t,searcher:n}){if(!s(t))return[];let a=[];if(r(t))t.forEach((({v:t,i:r,n:o})=>{if(!s(t))return;const{isMatch:i,score:l,indices:c}=n.searchIn(t);i&&a.push({score:l,key:e,value:t,idx:r,norm:o,indices:c})}));else{const{v:r,n:o}=t,{isMatch:i,score:s,indices:l}=n.searchIn(r);i&&a.push({score:s,key:e,value:r,norm:o,indices:l})}return a}}U.version="7.1.0",U.createIndex=b,U.parseIndex=function(e,{getFn:t=m.getFn,fieldNormWeight:n=m.fieldNormWeight}={}){const{keys:r,records:a}=e,o=new v({getFn:t,fieldNormWeight:n});return o.setKeys(r),o.setIndexRecords(a),o},U.config=m,U.parseQuery=B,function(...e){D.push(...e)}(class{constructor(e,{isCaseSensitive:t=m.isCaseSensitive,ignoreDiacritics:n=m.ignoreDiacritics,includeMatches:r=m.includeMatches,minMatchCharLength:a=m.minMatchCharLength,ignoreLocation:o=m.ignoreLocation,findAllMatches:i=m.findAllMatches,location:s=m.location,threshold:l=m.threshold,distance:c=m.distance}={}){this.query=null,this.options={isCaseSensitive:t,ignoreDiacritics:n,includeMatches:r,minMatchCharLength:a,findAllMatches:i,ignoreLocation:o,location:s,threshold:l,distance:c},e=t?e:e.toLowerCase(),e=n?k(e):e,this.pattern=e,this.query=function(e,t={}){return e.split("|").map((e=>{let n=e.trim().split(R).filter((e=>e&&!!e.trim())),r=[];for(let e=0,a=n.length;e<a;e+=1){const a=n[e];let o=!1,i=-1;for(;!o&&++i<A;){const e=T[i];let n=e.isMultiMatch(a);n&&(r.push(new e(n,t)),o=!0)}if(!o)for(i=-1;++i<A;){const e=T[i];let n=e.isSingleMatch(a);if(n){r.push(new e(n,t));break}}}return r}))}(this.pattern,this.options)}static condition(e,t){return t.useExtendedSearch}searchIn(e){const t=this.query;if(!t)return{isMatch:!1,score:1};const{includeMatches:n,isCaseSensitive:r,ignoreDiacritics:a}=this.options;e=r?e:e.toLowerCase(),e=a?k(e):e;let o=0,i=[],s=0;for(let r=0,a=t.length;r<a;r+=1){const a=t[r];i.length=0,o=0;for(let t=0,r=a.length;t<r;t+=1){const r=a[t],{isMatch:l,indices:c,score:u}=r.search(e);if(!l){s=0,o=0,i.length=0;break}if(o+=1,s+=u,n){const e=r.constructor.type;P.has(e)?i=[...i,...c]:i.push(c)}}if(o){let e={isMatch:!0,score:s/o};return n&&(e.indices=i),e}}return{isMatch:!1,score:1}}})},5936:(e,t,n)=>{"use strict";n.d(t,{A:()=>h});var r=n(641),a=n(33),o=n(953),i=n(9742),s=n(1504),l=n(3016),c=n(3395),u=n(9135);const d={key:0},p={class:"custom-scroll flex flex-1 flex-col gap-1.5"},h=(0,r.pM)({__name:"Form",props:{title:{},options:{},data:{},onUpdate:{type:Function}},setup(e){const{activeEnvVariables:t,activeEnvironment:n,activeWorkspace:h}=(0,u.JC)();return(e,u)=>((0,r.uX)(),(0,r.Wv)(c.A,null,{title:(0,r.k6)((()=>[e.title?((0,r.uX)(),(0,r.CE)("span",d,(0,a.v_)(e.title),1)):(0,r.RG)(e.$slots,"title",{key:1})])),default:(0,r.k6)((()=>[(0,r.Lk)("div",p,[Object.keys(e.data).length>0&&(0,o.R1)(h)?((0,r.uX)(),(0,r.Wv)(i.A,{key:0,columns:[""]},{default:(0,r.k6)((()=>[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(e.options,((i,c)=>((0,r.uX)(),(0,r.Wv)(l.A,{key:c,class:(0,a.C4)({"border-t":0===c})},{default:(0,r.k6)((()=>[(0,r.bF)(s.A,{envVariables:(0,o.R1)(t),environment:(0,o.R1)(n),modelValue:e.data[i.key]??"",placeholder:i.placeholder,workspace:(0,o.R1)(h),"onUpdate:modelValue":t=>e.onUpdate(i.key,t)},{default:(0,r.k6)((()=>[(0,r.eW)((0,a.v_)(i.label),1)])),_:2},1032,["envVariables","environment","modelValue","placeholder","workspace","onUpdate:modelValue"])])),_:2},1032,["class"])))),128))])),_:1})):(0,r.Q3)("",!0)])])),_:3}))}})},5946:(e,t,n)=>{"use strict";n.d(t,{o:()=>a});var r=n(7188);class a extends Error{constructor(e,t,n){super(),"string"==typeof t&&(n=t,t=void 0);let a="",o={},i=!1;if(t&&(o="line"in t&&"column"in t||"start"in t&&"end"in t?{place:t}:"type"in t?{ancestors:[t],place:t.position}:{...t}),"string"==typeof e?a=e:!o.cause&&e&&(i=!0,a=e.message,o.cause=e),!o.ruleId&&!o.source&&"string"==typeof n){const e=n.indexOf(":");-1===e?o.ruleId=n:(o.source=n.slice(0,e),o.ruleId=n.slice(e+1))}if(!o.place&&o.ancestors&&o.ancestors){const e=o.ancestors[o.ancestors.length-1];e&&(o.place=e.position)}const s=o.place&&"start"in o.place?o.place.start:o.place;this.ancestors=o.ancestors||void 0,this.cause=o.cause||void 0,this.column=s?s.column:void 0,this.fatal=void 0,this.file,this.message=a,this.line=s?s.line:void 0,this.name=(0,r.L)(o.place)||"1:1",this.place=o.place||void 0,this.reason=this.message,this.ruleId=o.ruleId||void 0,this.source=o.source||void 0,this.stack=i&&o.cause&&"string"==typeof o.cause.stack?o.cause.stack:"",this.actual,this.expected,this.note,this.url}}a.prototype.file="",a.prototype.name="",a.prototype.reason="",a.prototype.message="",a.prototype.stack="",a.prototype.column=void 0,a.prototype.line=void 0,a.prototype.ancestors=void 0,a.prototype.cause=void 0,a.prototype.fatal=void 0,a.prototype.place=void 0,a.prototype.ruleId=void 0,a.prototype.source=void 0},6099:(e,t,n)=>{"use strict";n.d(t,{c:()=>i});const r=(0,n(953).Kh)({}),a=(e,t)=>r[e]=t,o=e=>{r[e]=!r[e]},i=()=>({collapsedSidebarFolders:r,setCollapsedSidebarFolder:a,toggleSidebarFolder:o})},6221:(e,t,n)=>{"use strict";var r=n(6540);function a(e){var t="https://react.dev/errors/"+e;if(1<arguments.length){t+="?args[]="+encodeURIComponent(arguments[1]);for(var n=2;n<arguments.length;n++)t+="&args[]="+encodeURIComponent(arguments[n])}return"Minified React error #"+e+"; visit "+t+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function o(){}var i={d:{f:o,r:function(){throw Error(a(522))},D:o,C:o,L:o,m:o,X:o,S:o,M:o},p:0,findDOMNode:null},s=Symbol.for("react.portal"),l=r.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE;function c(e,t){return"font"===e?"":"string"==typeof t?"use-credentials"===t?t:"":void 0}t.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=i,t.createPortal=function(e,t){var n=2<arguments.length&&void 0!==arguments[2]?arguments[2]:null;if(!t||1!==t.nodeType&&9!==t.nodeType&&11!==t.nodeType)throw Error(a(299));return function(e,t,n){var r=3<arguments.length&&void 0!==arguments[3]?arguments[3]:null;return{$$typeof:s,key:null==r?null:""+r,children:e,containerInfo:t,implementation:n}}(e,t,null,n)},t.flushSync=function(e){var t=l.T,n=i.p;try{if(l.T=null,i.p=2,e)return e()}finally{l.T=t,i.p=n,i.d.f()}},t.preconnect=function(e,t){"string"==typeof e&&(t=t?"string"==typeof(t=t.crossOrigin)?"use-credentials"===t?t:"":void 0:null,i.d.C(e,t))},t.prefetchDNS=function(e){"string"==typeof e&&i.d.D(e)},t.preinit=function(e,t){if("string"==typeof e&&t&&"string"==typeof t.as){var n=t.as,r=c(n,t.crossOrigin),a="string"==typeof t.integrity?t.integrity:void 0,o="string"==typeof t.fetchPriority?t.fetchPriority:void 0;"style"===n?i.d.S(e,"string"==typeof t.precedence?t.precedence:void 0,{crossOrigin:r,integrity:a,fetchPriority:o}):"script"===n&&i.d.X(e,{crossOrigin:r,integrity:a,fetchPriority:o,nonce:"string"==typeof t.nonce?t.nonce:void 0})}},t.preinitModule=function(e,t){if("string"==typeof e)if("object"==typeof t&&null!==t){if(null==t.as||"script"===t.as){var n=c(t.as,t.crossOrigin);i.d.M(e,{crossOrigin:n,integrity:"string"==typeof t.integrity?t.integrity:void 0,nonce:"string"==typeof t.nonce?t.nonce:void 0})}}else null==t&&i.d.M(e)},t.preload=function(e,t){if("string"==typeof e&&"object"==typeof t&&null!==t&&"string"==typeof t.as){var n=t.as,r=c(n,t.crossOrigin);i.d.L(e,n,{crossOrigin:r,integrity:"string"==typeof t.integrity?t.integrity:void 0,nonce:"string"==typeof t.nonce?t.nonce:void 0,type:"string"==typeof t.type?t.type:void 0,fetchPriority:"string"==typeof t.fetchPriority?t.fetchPriority:void 0,referrerPolicy:"string"==typeof t.referrerPolicy?t.referrerPolicy:void 0,imageSrcSet:"string"==typeof t.imageSrcSet?t.imageSrcSet:void 0,imageSizes:"string"==typeof t.imageSizes?t.imageSizes:void 0,media:"string"==typeof t.media?t.media:void 0})}},t.preloadModule=function(e,t){if("string"==typeof e)if(t){var n=c(t.as,t.crossOrigin);i.d.m(e,{as:"string"==typeof t.as&&"script"!==t.as?t.as:void 0,crossOrigin:n,integrity:"string"==typeof t.integrity?t.integrity:void 0})}else i.d.m(e)},t.requestFormReset=function(e){i.d.r(e)},t.unstable_batchedUpdates=function(e,t){return e(t)},t.useFormState=function(e,t,n){return l.H.useFormState(e,t,n)},t.useFormStatus=function(){return l.H.useHostTransitionStatus()},t.version="19.0.0"},6251:(e,t,n)=>{"use strict";n.d(t,{EN:()=>v,_q:()=>b,kS:()=>y});var r,a=n(641),o=n(953),i=n(8869),s=n(1972),l=n(7014),c=n(7838),u=n(6413),d=n(9388),p=n(3605),h=((r=h||{})[r.Open=0]="Open",r[r.Closed=1]="Closed",r);let f=Symbol("DisclosureContext");function m(e){let t=(0,a.WQ)(f,null);if(null===t){let t=new Error(`<${e} /> is missing a parent <Disclosure /> component.`);throw Error.captureStackTrace&&Error.captureStackTrace(t,m),t}return t}let g=Symbol("DisclosurePanelContext"),v=(0,a.pM)({name:"Disclosure",props:{as:{type:[Object,String],default:"template"},defaultOpen:{type:[Boolean],default:!1}},setup(e,{slots:t,attrs:n}){let r=(0,o.KR)(e.defaultOpen?0:1),s=(0,o.KR)(null),c=(0,o.KR)(null),h={buttonId:(0,o.KR)(`headlessui-disclosure-button-${(0,i.B)()}`),panelId:(0,o.KR)(`headlessui-disclosure-panel-${(0,i.B)()}`),disclosureState:r,panel:s,button:c,toggleDisclosure(){r.value=(0,d.Y)(r.value,{0:1,1:0})},closeDisclosure(){1!==r.value&&(r.value=1)},close(e){h.closeDisclosure();let t=e?e instanceof HTMLElement?e:e.value instanceof HTMLElement?(0,u.t)(e):(0,u.t)(h.button):(0,u.t)(h.button);null==t||t.focus()}};return(0,a.Gt)(f,h),(0,l.Fo)((0,a.EW)((()=>(0,d.Y)(r.value,{0:l.Uw.Open,1:l.Uw.Closed})))),()=>{let{defaultOpen:a,...o}=e,i={open:0===r.value,close:h.close};return(0,p.XX)({theirProps:o,ourProps:{},slot:i,slots:t,attrs:n,name:"Disclosure"})}}}),b=(0,a.pM)({name:"DisclosureButton",props:{as:{type:[Object,String],default:"button"},disabled:{type:[Boolean],default:!1},id:{type:String,default:null}},setup(e,{attrs:t,slots:n,expose:r}){let i=m("DisclosureButton"),l=(0,a.WQ)(g,null),d=(0,a.EW)((()=>null!==l&&l.value===i.panelId.value));(0,a.sV)((()=>{d.value||null!==e.id&&(i.buttonId.value=e.id)})),(0,a.hi)((()=>{d.value||(i.buttonId.value=null)}));let h=(0,o.KR)(null);r({el:h,$el:h}),d.value||(0,a.nT)((()=>{i.button.value=h.value}));let f=(0,s.c)((0,a.EW)((()=>({as:e.as,type:t.type}))),h);function v(){var t;e.disabled||(d.value?(i.toggleDisclosure(),null==(t=(0,u.t)(i.button))||t.focus()):i.toggleDisclosure())}function b(t){var n;if(!e.disabled)if(d.value)switch(t.key){case c.D.Space:case c.D.Enter:t.preventDefault(),t.stopPropagation(),i.toggleDisclosure(),null==(n=(0,u.t)(i.button))||n.focus()}else switch(t.key){case c.D.Space:case c.D.Enter:t.preventDefault(),t.stopPropagation(),i.toggleDisclosure()}}function y(e){e.key===c.D.Space&&e.preventDefault()}return()=>{var r;let a={open:0===i.disclosureState.value},{id:o,...s}=e,l=d.value?{ref:h,type:f.value,onClick:v,onKeydown:b}:{id:null!=(r=i.buttonId.value)?r:o,ref:h,type:f.value,"aria-expanded":0===i.disclosureState.value,"aria-controls":0===i.disclosureState.value||(0,u.t)(i.panel)?i.panelId.value:void 0,disabled:!!e.disabled||void 0,onClick:v,onKeydown:b,onKeyup:y};return(0,p.XX)({ourProps:l,theirProps:s,slot:a,attrs:t,slots:n,name:"DisclosureButton"})}}}),y=(0,a.pM)({name:"DisclosurePanel",props:{as:{type:[Object,String],default:"div"},static:{type:Boolean,default:!1},unmount:{type:Boolean,default:!0},id:{type:String,default:null}},setup(e,{attrs:t,slots:n,expose:r}){let o=m("DisclosurePanel");(0,a.sV)((()=>{null!==e.id&&(o.panelId.value=e.id)})),(0,a.hi)((()=>{o.panelId.value=null})),r({el:o.panel,$el:o.panel}),(0,a.Gt)(g,o.panelId);let i=(0,l.O_)(),s=(0,a.EW)((()=>null!==i?(i.value&l.Uw.Open)===l.Uw.Open:0===o.disclosureState.value));return()=>{var r;let a={open:0===o.disclosureState.value,close:o.close},{id:i,...l}=e,c={id:null!=(r=o.panelId.value)?r:i,ref:o.panel};return(0,p.XX)({ourProps:c,theirProps:l,slot:a,attrs:t,slots:n,features:p.O5.RenderStrategy|p.O5.Static,visible:s.value,name:"DisclosurePanel"})}}})},6349:(e,t,n)=>{"use strict";n.d(t,{A:()=>i});var r=n(165),a=n(6455),o=n(4048);const i=(e=[],t={},n="")=>{const i={},s=[],l=new URLSearchParams;return e.forEach((e=>{var c;if("apiKey"===e.type){const o=(0,r.GV)(e.value,t)||n;"header"===e.in&&(i[e.name]=o),"query"===e.in&&l.append(e.name,o),"cookie"===e.in&&s.push(a.c.parse({uid:e.uid,name:e.name,value:o,path:"/"}))}if("http"===e.type)if("basic"===e.scheme){const n=`${(0,r.GV)(e.username,t)}:${(0,r.GV)(e.password,t)}`;i.Authorization=`Basic ${":"===n?"username:password":btoa(n)}`}else{const a=(0,r.GV)(e.token,t);i.Authorization=`Bearer ${a||n}`}if("oauth2"===e.type){const t=null==(c=Object.values(e.flows).filter(o.O9).find((e=>e.token)))?void 0:c.token;i.Authorization=`Bearer ${t||n}`}})),{headers:i,cookies:s,urlParams:l}}},6413:(e,t,n)=>{"use strict";function r(e){var t;if(null==e||null==e.value)return null;let n=null!=(t=e.value.$el)?t:e.value;return n instanceof Node?n:null}n.d(t,{t:()=>r})},6454:(e,t,n)=>{"use strict";n.d(t,{q:()=>s});var r=n(953),a=n(6413);let o=/([\u2700-\u27BF]|[\uE000-\uF8FF]|\uD83C[\uDC00-\uDFFF]|\uD83D[\uDC00-\uDFFF]|[\u2011-\u26FF]|\uD83E[\uDD10-\uDDFF])/g;function i(e){var t,n;let r=null!=(t=e.innerText)?t:"",a=e.cloneNode(!0);if(!(a instanceof HTMLElement))return r;let i=!1;for(let e of a.querySelectorAll('[hidden],[aria-hidden],[role="img"]'))e.remove(),i=!0;let s=i?null!=(n=a.innerText)?n:"":r;return o.test(s)&&(s=s.replace(o,"")),s}function s(e){let t=(0,r.KR)(""),n=(0,r.KR)("");return()=>{let r=(0,a.t)(e);if(!r)return"";let o=r.innerText;if(t.value===o)return n.value;let s=function(e){let t=e.getAttribute("aria-label");if("string"==typeof t)return t.trim();let n=e.getAttribute("aria-labelledby");if(n){let e=n.split(" ").map((e=>{let t=document.getElementById(e);if(t){let e=t.getAttribute("aria-label");return"string"==typeof e?e.trim():i(t).trim()}return null})).filter(Boolean);if(e.length>0)return e.join(", ")}return i(e).trim()}(r).trim().toLowerCase();return t.value=o,n.value=s,s}}},6455:(e,t,n)=>{"use strict";n.d(t,{c:()=>o});var r=n(4476),a=n(2071);const o=r.z.object({uid:a.B.brand(),name:r.z.string().default(""),value:r.z.string().default(""),domain:r.z.string().optional(),path:r.z.string().optional()})},6530:(e,t,n)=>{"use strict";n.d(t,{_:()=>a});var r=n(6815);function a(e,t){return(0,r.B)(e,t.toLowerCase())}},6540:(e,t,n)=>{"use strict";e.exports=n(9869)},6584:(e,t,n)=>{"use strict";n.d(t,{K:()=>a,g:()=>o});var r=n(641);const a=Symbol(),o=()=>({layout:(0,r.WQ)(a,"desktop")})},6736:(e,t,n)=>{"use strict";n.d(t,{Kt:()=>c,a2:()=>u,hF:()=>s});var r=n(165),a=n(6455),o=n(4048);const i="/";function s({example:e,env:t,globalCookies:n,serverUrl:s,proxyUrl:u}){const d=[],p=(0,o.p3)(u,s),h=l(p?u:s??"http://localhost");return n.forEach((e=>{const{name:t,value:n,domain:r,...o}=e;!c(s,r)||!t||d.push(a.c.parse({name:t,value:n,domain:r,path:o.path}))})),e.parameters.cookies.forEach((e=>{!e.enabled||!e.key||d.push(a.c.parse({name:e.key,value:(0,r.GV)(e.value,t),domain:h,path:i}))})),{cookieParams:d}}const l=e=>{const t=new URL(e.startsWith("http")?e:`http://${e}`).hostname;return t.match(/^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/)||t.match(/^[a-fA-F0-9:]+$/)||t.startsWith(".")?t:`.${t}`},c=(e,t)=>{if(!e||!t)return!0;try{const n=e.startsWith("http")?e:`http://${e}`,r=new URL(n).hostname,a=!t,o=t===r,i=t.startsWith(".")&&t===`.${r}`,s=t.startsWith(".")&&(null==r?void 0:r.endsWith(t));return a||o||s||i}catch{return!1}},u=(e,t)=>{const n=e.map((e=>`${e.name}=${e.value}`)).join("; ");return t?`${t}; ${n}`.trim():n.trim()}},6764:(e,t,n)=>{"use strict";n.d(t,{q:()=>a});var r=n(8470);const a=e=>{if("string"==typeof e)return(0,r.Eo)(e)?JSON.stringify(JSON.parse(e),null,2):e;if("object"==typeof e)try{return JSON.stringify(e,null,2)}catch{return function(e){const t=new Set;return JSON.stringify(e,((e,n)=>{if("object"==typeof n&&null!==n){if(t.has(n))return"[Circular]";t.add(n)}return n}),2)}(e)}return e?.toString()??""}},6815:(e,t,n)=>{"use strict";function r(e,t){return t in e?e[t]:t}n.d(t,{B:()=>r})},6918:(e,t,n)=>{"use strict";n.d(t,{A:()=>u});var r=n(641),a=n(3751),o=n(953),i=n(33),s=n(1266);const l={class:"flex gap-10 justify-between"},c=(0,r.pM)({__name:"SidebarListElementForm",props:{danger:{type:Boolean},label:{}},emits:["cancel","submit"],setup(e,{emit:t}){const n=t;return(e,t)=>((0,r.uX)(),(0,r.CE)("form",{class:"flex flex-col gap-4 text-base",onSubmit:t[1]||(t[1]=(0,a.D$)((e=>n("submit")),["prevent"]))},[(0,r.RG)(e.$slots,"default",{},void 0,!0),(0,r.Lk)("div",l,[(0,r.bF)((0,o.R1)(s.RO),{class:"gap-1.5 px-2.5 h-8 shadow-none focus:outline-none flex items-center cursor-pointer",type:"button",variant:"outlined",onClick:t[0]||(t[0]=e=>n("cancel"))},{default:(0,r.k6)((()=>t[2]||(t[2]=[(0,r.eW)(" Cancel ")]))),_:1}),(0,r.bF)((0,o.R1)(s.RO),{class:"gap-1.5 font-medium px-2.5 h-8 shadow-none focus:outline-none custom-scroll whitespace-nowrap",type:"submit",variant:e.danger?"danger":"solid"},{default:(0,r.k6)((()=>[(0,r.eW)((0,i.v_)(e.label??"Save"),1)])),_:1},8,["variant"])])],32))}}),u=(0,n(3526).A)(c,[["__scopeId","data-v-b6a60dbc"]])},6923:(e,t,n)=>{"use strict";n.d(t,{P:()=>o});var r=n(953),a=n(641);function o(e,t,n){let o=(0,r.KR)(null==n?void 0:n.value),i=(0,a.EW)((()=>void 0!==e.value));return[(0,a.EW)((()=>i.value?e.value:o.value)),function(e){return i.value||(o.value=e),null==t?void 0:t(e)}]}},6932:(e,t,n)=>{"use strict";n.d(t,{b:()=>r});const r=["area","base","basefont","bgsound","br","col","command","embed","frame","hr","image","img","input","keygen","link","meta","param","source","track","wbr"]},6941:(e,t,n)=>{"use strict";n.d(t,{M:()=>s});const r={"date-time":(new Date).toISOString(),date:(new Date).toISOString().split("T")[0],email:"hello@example.com",hostname:"example.com","idn-email":"jane.doe@example.com","idn-hostname":"example.com",ipv4:"127.0.0.1",ipv6:"51d4:7fab:bfbf:b7d7:b2cb:d4b4:3dad:d998","iri-reference":"/entitiy/1",iri:"https://example.com/entity/123","json-pointer":"/nested/objects",password:"super-secret",regex:"/[a-z]/","relative-json-pointer":"1/nested/objects",time:(new Date).toISOString().split("T")[1].split(".")[0],"uri-reference":"../folder","uri-template":"https://example.com/{id}",uri:"https://example.com",uuid:"123e4567-e89b-12d3-a456-426614174000","object-id":"6592008029c8c3e4dc76256c"};function a(e,t=""){return r[e.format]??t}const o=new WeakMap;function i(e,t){return"object"!=typeof t||null===t||o.set(e,t),t}const s=(e,t,n=0,r,l)=>{if(o.has(e))return o.get(e);if(6===n)try{JSON.stringify(e)}catch{return"[Circular Reference]"}const c=!!t?.emptyString;if("write"===t?.mode&&e.readOnly||"read"===t?.mode&&e.writeOnly)return;if(e["x-variable"]){const n=t?.variables?.[e["x-variable"]];if(void 0!==n)return"number"===e.type||"integer"===e.type?parseInt(n,10):i(e,n)}if(Array.isArray(e.examples)&&e.examples.length>0)return i(e,e.examples[0]);if(void 0!==e.example)return i(e,e.example);if(void 0!==e.default)return i(e,e.default);if(Array.isArray(e.enum)&&e.enum.length>0)return i(e,e.enum[0]);if(!("object"===e.type||"array"===e.type||e.allOf?.at?.(0)||e.anyOf?.at?.(0)||e.oneOf?.at?.(0)||!0!==t?.omitEmptyAndOptionalProperties||!0===e.required||!0===r?.required||r?.required?.includes(l??e.name)))return;if("object"===e.type||void 0!==e.properties){const r={};if(void 0!==e.properties)for(const a in e.properties)if(Object.prototype.hasOwnProperty.call(e.properties,a)){const o=e.properties[a],i=t?.xml?o.xml?.name:void 0;r[i??a]=s(o,t,n+1,e,a),void 0===r[i??a]&&delete r[i??a]}if(void 0!==e.patternProperties)for(const a in e.patternProperties)if(Object.prototype.hasOwnProperty.call(e.patternProperties,a)){const o=e.patternProperties[a],i=a;r[i]=s(o,t,n+1,e,i)}return void 0!==e.additionalProperties&&(!0===e.additionalProperties||"object"==typeof e.additionalProperties&&!Object.keys(e.additionalProperties).length?r.ANY_ADDITIONAL_PROPERTY="anything":!1!==e.additionalProperties&&(r.ANY_ADDITIONAL_PROPERTY=s(e.additionalProperties,t,n+1))),void 0!==e.anyOf?Object.assign(r,s(e.anyOf[0],t,n+1)):void 0!==e.oneOf?Object.assign(r,s(e.oneOf[0],t,n+1)):void 0!==e.allOf&&Object.assign(r,...e.allOf.map((r=>s(r,t,n+1,e))).filter((e=>void 0!==e))),i(e,r)}if("array"===e.type||void 0!==e.items){const r=e?.items?.xml?.name,a=!!(t?.xml&&e.xml?.wrapped&&r);if(void 0!==e.example)return i(e,a?{[r]:e.example}:e.example);if(e.items){if(e.items.allOf){if("object"===e.items.allOf[0].type){const o=s({type:"object",allOf:e.items.allOf},t,n+1,e);return i(e,a?[{[r]:o}]:[o])}{const o=e.items.allOf.map((r=>s(r,t,n+1,e))).filter((e=>void 0!==e));return i(e,a?o.map((e=>({[r]:e}))):o)}}const o=["anyOf","oneOf"];for(const l of o){if(!e.items[l])continue;const o=e.items[l].slice(0,1).map((r=>s(r,t,n+1,e))).filter((e=>void 0!==e));return i(e,a?[{[r]:o}]:o)}}if(e.items?.type){const o=s(e.items,t,n+1);return a?[{[r]:o}]:[o]}return[]}const u={string:c?a(e,t?.emptyString):"",boolean:!0,integer:e.min??1,number:e.min??1,array:[]};if(void 0!==e.type&&void 0!==u[e.type])return i(e,u[e.type]);const d=e.oneOf||e.anyOf;if(Array.isArray(d)&&d.length>0){const e=d[0];return s(e,t,n+1)}if(Array.isArray(e.allOf)){let r=null;return e.allOf.forEach((e=>{const a=s(e,t,n+1);r="object"==typeof a&&"object"==typeof r?{...r??{},...a}:Array.isArray(a)&&Array.isArray(r)?[...r??{},...a]:a})),i(e,r)}if(Array.isArray(e.type)){if(e.type.includes("null"))return null;const t=u[e.type[0]];if(void 0!==t)return i(e,t)}return null}},7014:(e,t,n)=>{"use strict";n.d(t,{Fo:()=>l,O_:()=>s,Uw:()=>i});var r=n(641);let a=Symbol("Context");var o,i=((o=i||{})[o.Open=1]="Open",o[o.Closed=2]="Closed",o[o.Closing=4]="Closing",o[o.Opening=8]="Opening",o);function s(){return(0,r.WQ)(a,null)}function l(e){(0,r.Gt)(a,e)}},7088:e=>{e.exports=function(e){const t={className:"string",begin:/"/,end:/"/,contains:[e.BACKSLASH_ESCAPE,{className:"variable",begin:/\$\(/,end:/\)/,contains:[e.BACKSLASH_ESCAPE]}],relevance:0},n={className:"number",variants:[{begin:e.C_NUMBER_RE}],relevance:0};return{name:"curl",aliases:["curl"],keywords:"curl",case_insensitive:!0,contains:[{className:"literal",begin:/(--request|-X)\s/,contains:[{className:"symbol",begin:/(get|post|delete|options|head|put|patch|trace|connect)/,end:/\s/,returnEnd:!0}],returnEnd:!0,relevance:10},{className:"literal",begin:/--/,end:/[\s"]/,returnEnd:!0,relevance:0},{className:"literal",begin:/-\w/,end:/[\s"]/,returnEnd:!0,relevance:0},t,{className:"string",begin:/\\"/,relevance:0},{className:"string",begin:/'/,end:/'/,relevance:0},e.APOS_STRING_MODE,e.QUOTE_STRING_MODE,n,{match:/(\/[a-z._-]+)+/}]}}},7176:(e,t,n)=>{"use strict";n.d(t,{S:()=>r});class r{constructor(e,t,n){this.normal=t,this.property=e,n&&(this.space=n)}}r.prototype.normal={},r.prototype.property={},r.prototype.space=void 0},7186:(e,t,n)=>{"use strict";n.d(t,{A:()=>f});var r=n(641),a=n(33),o=n(953),i=n(3751),s=n(1266);const l={class:"hidden lg:flex ml-auto filter-hover context-bar-group lg:w-[120px]"},c={class:"request-section-content request-section-content-filter fade-request-section-content text-c-3 hidden xl:flex justify-end rounded text-xs gap-[1.5px] pointer-events-auto relative w-full py-2"},u=["name","value"],d={class:"flex items-center context-bar-group-hover:text-c-1 absolute -right-[30px] top-1/2 -translate-y-1/2"},p={class:"mr-1.5 context-bar-group-hover:hidden"},h=(0,r.pM)({__name:"ContextBar",props:{sections:{},activeSection:{}},emits:["setActiveSection"],setup(e,{emit:t}){const n=e,h=t,f=(0,r.Bi)(),m=(0,r.EW)({get:()=>n.activeSection,set:e=>h("setActiveSection",e)});return(e,t)=>((0,r.uX)(),(0,r.CE)("fieldset",l,[t[1]||(t[1]=(0,r.Lk)("legend",{class:"sr-only"},"Filter Sections",-1)),(0,r.Lk)("div",c,[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(e.sections,(e=>((0,r.uX)(),(0,r.CE)("label",{key:e,class:(0,a.C4)(["filter-hover-item flex items-center hover:bg-b-2 rounded p-1 px-2 cursor-pointer text-center font-medium w-fit whitespace-nowrap has-[:focus-visible]:outline",[m.value===e?"text-c-1  pointer-events-none":""]])},[(0,r.eW)((0,a.v_)(e)+" ",1),(0,r.bo)((0,r.Lk)("input",{"onUpdate:modelValue":t[0]||(t[0]=e=>m.value=e),class:"sr-only",name:(0,o.R1)(f),type:"radio",value:e},null,8,u),[[i.XL,m.value]])],2)))),128)),(0,r.Lk)("div",d,[(0,r.Lk)("span",p,(0,a.v_)(m.value),1),(0,r.bF)((0,o.R1)(s.kv),{icon:"FilterList",size:"md",thickness:"2"})])])]))}}),f=(0,n(3526).A)(h,[["__scopeId","data-v-7906217f"]])},7188:(e,t,n)=>{"use strict";function r(e){return e&&"object"==typeof e?"position"in e||"type"in e?o(e.position):"start"in e||"end"in e?o(e):"line"in e||"column"in e?a(e):"":""}function a(e){return i(e&&e.line)+":"+i(e&&e.column)}function o(e){return a(e&&e.start)+"-"+a(e&&e.end)}function i(e){return e&&"number"==typeof e?e:1}n.d(t,{L:()=>r})},7200:(e,t,n)=>{"use strict";n.d(t,{Sr:()=>r,cQ:()=>a});const r={DEFAULT:"An unknown error has occurred.",INVALID_URL:"The URL seems to be invalid. Try adding a valid URL.",INVALID_HEADER:"There is an invalid header present, please double check your params.",MISSING_FILE:"File uploads are not saved in history, you must re-upload the file.",REQUEST_ABORTED:"The request has been cancelled",REQUEST_FAILED:"An error occurred while making the request",URL_EMPTY:"The address bar input seems to be empty. Try adding a URL."},a=(e,t=r.DEFAULT)=>(console.error(e),e instanceof Error?(e.message=o(e.message),e):"string"==typeof e?new Error(o(e)):new Error(t)),o=e=>"Failed to execute 'append' on 'FormData': parameter 2 is not of type 'Blob'."===e?r.MISSING_FILE:"Failed to construct 'URL': Invalid URL"===e?r.INVALID_URL:"Failed to execute 'fetch' on 'Window': Invalid name"===e?r.INVALID_HEADER:e},7297:(e,t,n)=>{"use strict";n.d(t,{E1:()=>c,TF:()=>h});var r=n(2071),a=n(4476),o=n(4811),i=n(4703),s=n(5785),l=n(8065);const c=["connect","delete","get","head","options","patch","post","put","trace"],u=a.z.any(),d=a.z.object({tags:a.z.string().array().optional(),summary:a.z.string().optional(),description:a.z.string().optional(),operationId:a.z.string().optional(),security:a.z.array(s.AH).optional(),requestBody:u.optional(),parameters:o.iZ.array().optional(),externalDocs:l.MA.optional(),deprecated:a.z.boolean().optional(),responses:a.z.record(a.z.string(),a.z.any()).optional(),"x-scalar-examples":a.z.record(a.z.string(),i.kh).optional(),"x-internal":a.z.boolean().optional(),"x-scalar-ignore":a.z.boolean().optional()}),p=a.z.object({type:a.z.literal("request").optional().default("request"),uid:r.B.brand(),path:a.z.string().optional().default(""),method:a.z.enum(c).default("get"),servers:a.z.string().brand().array().default([]),selectedServerUid:a.z.string().brand().optional().nullable().default(null),examples:a.z.string().brand().array().default([]),selectedSecuritySchemeUids:r.I}),h=d.omit({"x-scalar-examples":!0}).merge(p)},7356:(e,t,n)=>{"use strict";n.d(t,{AJ:()=>u,D9:()=>l,I5:()=>s});var r=n(1347),a=n(4048),o=n(8186),i=n(953);function s(e){const t=(0,i.Kh)({});return{requests:t,requestMutators:(0,o.oT)(t,(0,i.Kh)({}),e&&a.nF.REQUEST)}}function l({requestExamples:e,requestExampleMutators:t,requestMutators:n,collectionMutators:o,collections:i,tags:s,tagMutators:l},u){return{addRequest:(c,d)=>{var p;const h=(0,a.TA)(c,r.TF,!1);if(!h)return console.error("INVALID REQUEST DATA",c);const f=i[d],m=(0,r.oE)(h,(0,a.zR)((h.summary??"Example")+" #1",(t=>h.examples.some((n=>{var r;return t===(null==(r=e[n])?void 0:r.name)})))));return h.examples.push(m.uid),n.add(h),t.add(m),f&&o.edit(d,"requests",[...f.requests,h.uid]),null!=(p=h.tags)&&p.length?h.tags.forEach((e=>{const t=null==f?void 0:f.tags.find((t=>{var n;return(null==(n=s[t])?void 0:n.name)===e}));t&&s[t]?l.edit(t,"children",[...s[t].children,h.uid]):u({name:e,children:[h.uid]},d)})):f&&o.edit(d,"children",[...f.children,h.uid]),h},deleteRequest:(e,r)=>{var a;const c=i[r];e.examples.forEach((e=>t.delete(e))),c&&(o.edit(r,"requests",c.requests.filter((t=>t!==e.uid))),o.edit(r,"children",c.children.filter((t=>t!==e.uid))),null==(a=e.tags)||a.forEach((t=>{var n;const r=c.tags.find((e=>{var n;return(null==(n=s[e])?void 0:n.name)===t}));r&&l.edit(r,"children",(null==(n=s[r])?void 0:n.children.filter((t=>t!==e.uid)))||[])}))),n.delete(e.uid)},findRequestParents:c({collections:i,tags:s})}}function c({collections:e,tags:t}){return function(n){const r=Object.values(e).find((e=>{var t;return null==(t=e.requests)?void 0:t.includes(n.uid)}));if(!r)return[];const a=Object.keys(t).reduce(((e,t)=>(e[t]=[],e)),{[null==r?void 0:r.uid]:[]});!function e(n,r){r.forEach((e=>{var t;return null==(t=a[e])?void 0:t.push(...n.children)})),n.children.forEach((n=>{t[n]&&e(t[n],[...r,n])}))}(r,[r.uid]);const o=new Set;return Object.entries(a).forEach((([e,t])=>{t.includes(n.uid)&&o.add(e)})),[...o]}}function u(){return{request:r.TF.parse({method:"get",parameters:[],path:"",summary:"My First Request",examples:[]})}}},7427:(e,t,n)=>{"use strict";function r(e){return decodeURI(e.replace(/~1/g,"/").replace(/~0/g,"~"))}n.d(t,{K:()=>r})},7457:(e,t,n)=>{"use strict";function r(e){return e.toLowerCase()}n.d(t,{S:()=>r})},7471:(e,t,n)=>{"use strict";function r(){}n.d(t,{ok:()=>r})},7525:(e,t,n)=>{"use strict";n.d(t,{A:()=>fe});var r=n(641),a=n(953),o=n(33),i=n(3751),s=n(7186),l=n(3395),c=n(6736),u=n(3654),d=n(9742),p=n(1266),h=n(4282);const f=(0,r.pM)({__name:"DataTableHeader",setup(e){const{cx:t}=(0,p.AX)();return(e,n)=>((0,r.uX)(),(0,r.Wv)(h.A,(0,r.v6)({is:"th"},(0,a.R1)(t)("truncate font-medium px-2 py-1.5")),{default:(0,r.k6)((()=>[(0,r.RG)(e.$slots,"default")])),_:3},16))}});var m=n(3016),g=n(25),v=n(1347),b=n(4048),y=n(9971);const w=e=>!!(e.description||e.type||e.default||e.format||e.minimum||e.maximum),O=e=>(0,r.EW)((()=>{if(void 0===e.value||""===e.value)return!1;if(e.type){if("integer"===e.type){const t=Number(e.value);if(isNaN(t)||!Number.isInteger(t))return"Value must be a whole number (e.g., 42)";if(void 0!==e.minimum&&t<e.minimum)return`Value must be ${e.minimum} or greater`;if(void 0!==e.maximum&&t>e.maximum)return`Value must be ${e.maximum} or less`}if("number"===e.type){const t=Number(e.value);if(isNaN(t))return"Value must be a number (e.g., 42.5)";if(void 0!==e.minimum&&t<e.minimum)return`Value must be ${e.minimum} or greater`;if(void 0!==e.maximum&&t>e.maximum)return`Value must be ${e.maximum} or less`}if("string"===e.type&&e.format){if("date"===e.format&&!/^\d{4}-\d{2}-\d{2}$/.test(e.value))return"Please enter a valid date in YYYY-MM-DD format (e.g., 2024-03-20)";if("date-time"===e.format&&!/^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?(Z|[+-]\d{2}:\d{2})$/.test(e.value))return"Please enter a valid date and time in RFC 3339 format (e.g., 2024-03-20T13:45:30Z)";if("email"===e.format&&!/^[^\s@]+@[^\s@]+\.[^\s@]+$/.test(e.value))return"Please enter a valid email address (e.g., user@example.com)";if("uri"===e.format&&!/^[a-zA-Z][a-zA-Z0-9+.-]*:.+$/.test(e.value))return"Please enter a valid URI (e.g., https://example.com)"}}return!1})),k={class:"bg-b-1 mr-0.25 pl-1 pr-1.5 group-[.alert]:bg-transparent"},x={class:"w-content bg-b-1 text-xxs text-c-1 pointer-events-none grid min-w-48 gap-1.5 rounded p-2 leading-5 shadow-lg"},E={key:0,class:"text-error-1"},S={key:1,class:"schema text-c-2 flex items-center"},_={key:0},C={key:1},T={key:2},A={key:3},R={key:4},P={key:2,class:"text-pretty text-sm leading-snug",style:{maxWidth:"16rem"}},D=(0,r.pM)({__name:"RequestTableTooltip",props:{item:{}},setup:e=>(e,t)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(p.iy),{align:"start",class:"w-full pr-px",delay:0,side:"left",triggerClass:"before:absolute before:content-[''] before:bg-gradient-to-r before:from-transparent before:to-b-1 group-[.alert]:before:to-b-alert before:min-h-[calc(100%-4px)] before:pointer-events-none before:right-[23px] before:top-0.5 before:w-3 absolute h-full right-0 -outline-offset-1"},{trigger:(0,r.k6)((()=>[(0,r.Lk)("div",k,[(0,r.bF)((0,a.R1)(p.kv),{class:(0,o.C4)((0,a.R1)(O)(e.item).value?"text-orange brightness-[.9]":"text-c-3 group-hover/info:text-c-1"),icon:(0,a.R1)(O)(e.item).value?"Alert":"Info",size:"sm",thickness:"1.5"},null,8,["class","icon"])])])),content:(0,r.k6)((()=>[(0,r.Lk)("div",x,[(0,a.R1)(O)(e.item).value?((0,r.uX)(),(0,r.CE)("div",E,(0,o.v_)((0,a.R1)(O)(e.item).value),1)):((0,r.uX)(),(0,r.CE)("div",S,[e.item.type?((0,r.uX)(),(0,r.CE)("span",_,(0,o.v_)(e.item.type),1)):(0,r.Q3)("",!0),e.item.format?((0,r.uX)(),(0,r.CE)("span",C,(0,o.v_)(e.item.format),1)):(0,r.Q3)("",!0),e.item.minimum?((0,r.uX)(),(0,r.CE)("span",T,"min: "+(0,o.v_)(e.item.minimum),1)):(0,r.Q3)("",!0),e.item.maximum?((0,r.uX)(),(0,r.CE)("span",A,"max: "+(0,o.v_)(e.item.maximum),1)):(0,r.Q3)("",!0),e.item.default?((0,r.uX)(),(0,r.CE)("span",R,"default: "+(0,o.v_)(e.item.default),1)):(0,r.Q3)("",!0)])),e.item.description&&!(0,a.R1)(O)(e.item).value?((0,r.uX)(),(0,r.CE)("span",P,(0,o.v_)(e.item.description),1)):(0,r.Q3)("",!0)])])),_:1}))});var M=n(3526);const $=(0,M.A)(D,[["__scopeId","data-v-30c955c6"]]),N={class:"contents"},I={class:"sr-only"},L={class:"text-c-2 filemask flex max-w-[100%] items-end justify-end overflow-hidden"},B=["onClick"],F=(0,r.pM)({__name:"RequestTable",props:{items:{},hasCheckboxDisabled:{type:Boolean,default:!1},showUploadButton:{type:Boolean,default:!1},isGlobal:{type:Boolean,default:!1},isReadOnly:{type:Boolean,default:!1},environment:{},envVariables:{},workspace:{}},emits:["updateRow","toggleRow","addRow","deleteRow","inputFocus","inputBlur","uploadFile","removeFile"],setup(e,{emit:t}){const n=e,i=t,s=["","","36px"],l=(e,t,n)=>{i("updateRow",e,t,n)},c=e=>Array.isArray(e.default)&&1===e.default.length?e.default[0]:e.default;return(e,t)=>{const f=(0,r.g2)("RouterLink");return(0,r.uX)(),(0,r.Wv)(d.A,{class:"group/table flex-1",columns:s},{default:(0,r.k6)((()=>[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(e.items,((s,d)=>((0,r.uX)(),(0,r.Wv)(m.A,{key:d,class:(0,o.C4)({alert:(0,a.R1)(O)(s).value})},{default:(0,r.k6)((()=>[(0,r.Lk)("label",N,[e.isGlobal?((0,r.uX)(),(0,r.Wv)(f,{key:0,class:"!border-r-1/2 border-t-1/2 text-c-2 flex items-center justify-center",to:s.route??{}},{default:(0,r.k6)((()=>[t[5]||(t[5]=(0,r.Lk)("span",{class:"sr-only"},"Global",-1)),(0,r.bF)((0,a.R1)(p.iy),{as:"div",side:"top"},{trigger:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(p.kv),{class:"text-c-1",icon:"Globe",size:"xs"})])),content:(0,r.k6)((()=>t[4]||(t[4]=[(0,r.Lk)("div",{class:"w-content bg-b-1 z-100 text-xxs text-c-1 pointer-events-none z-10 grid max-w-[320px] gap-1.5 rounded p-2 leading-5 shadow-lg"},[(0,r.Lk)("div",{class:"text-c-1 flex items-center"},[(0,r.Lk)("span",{class:"text-pretty"}," Global cookies are shared across the whole workspace. ")])],-1)]))),_:1})])),_:2},1032,["to"])):((0,r.uX)(),(0,r.CE)(r.FK,{key:1},[(0,r.Lk)("span",I," Row "+(0,o.v_)(s.enabled?"Enabled":"Disabled"),1),(0,r.bF)(y.A,{class:"!border-r-1/2",disabled:n.hasCheckboxDisabled,modelValue:s.enabled,"onUpdate:modelValue":e=>i("toggleRow",d,e)},null,8,["disabled","modelValue","onUpdate:modelValue"])],64))]),(0,r.bF)(h.A,null,{default:(0,r.k6)((()=>[(0,r.bF)(u.A,{disableCloseBrackets:"",disabled:n.isReadOnly,disableEnter:"",disableTabIndent:"",envVariables:e.envVariables,environment:e.environment,modelValue:s.key,placeholder:"Key",required:!!s.required,workspace:e.workspace,onBlur:t[0]||(t[0]=e=>i("inputBlur")),onFocus:t[1]||(t[1]=e=>i("inputFocus")),onSelectVariable:e=>l(d,"key",e),"onUpdate:modelValue":e=>i("updateRow",d,"key",e)},null,8,["disabled","envVariables","environment","modelValue","required","workspace","onSelectVariable","onUpdate:modelValue"])])),_:2},1024),(0,r.bF)(h.A,null,{default:(0,r.k6)((()=>[(0,r.bF)(u.A,{class:(0,o.C4)({"pr-6":(0,a.R1)(w)(s)}),default:s.default,disableCloseBrackets:"",disabled:n.isReadOnly,disableEnter:"",disableTabIndent:"",enum:s.enum??[],envVariables:e.envVariables,environment:e.environment,examples:s.examples??[],max:s.maximum,min:s.minimum,modelValue:s.value,nullable:!!s.nullable,placeholder:"Value",type:s.type,workspace:e.workspace,onBlur:t[2]||(t[2]=e=>i("inputBlur")),onFocus:t[3]||(t[3]=e=>i("inputFocus")),onSelectVariable:e=>l(d,"value",e),"onUpdate:modelValue":e=>i("updateRow",d,"value",e)},{icon:(0,r.k6)((()=>[(0,a.R1)(w)(s)?((0,r.uX)(),(0,r.Wv)($,{key:0,item:{...s,default:c(s)}},null,8,["item"])):(0,r.Q3)("",!0)])),_:2},1032,["class","default","disabled","enum","envVariables","environment","examples","max","min","modelValue","nullable","type","workspace","onSelectVariable","onUpdate:modelValue"])])),_:2},1024),e.showUploadButton?((0,r.uX)(),(0,r.Wv)(h.A,{key:0,class:"group/upload relative overflow-hidden text-ellipsis whitespace-nowrap p-1"},{default:(0,r.k6)((()=>{var e;return[s.file?((0,r.uX)(),(0,r.CE)(r.FK,{key:0},[(0,r.Lk)("div",L,[(0,r.Lk)("span",null,(0,o.v_)(null==(e=s.file)?void 0:e.name),1)]),(0,r.Lk)("button",{class:"bg-b-2 centered-x centered-y absolute hidden w-[calc(100%_-_8px)] rounded p-0.5 text-center text-xs font-medium group-hover/upload:block",type:"button",onClick:e=>i("removeFile",d)}," Delete ",8,B)],64)):((0,r.uX)(),(0,r.Wv)((0,a.R1)(p.RO),{key:1,class:"bg-b-2 hover:bg-b-3 text-c-2 border-0 py-px shadow-none",size:"sm",variant:"outlined",onClick:e=>(e=>{i("uploadFile",e)})(d)},{default:(0,r.k6)((()=>[t[6]||(t[6]=(0,r.Lk)("span",null,"File",-1)),(0,r.bF)((0,a.R1)(p.kv),{class:"ml-1",icon:"UploadSimple",size:"xs",thickness:"2.5"})])),_:2},1032,["onClick"]))]})),_:2},1024)):(0,r.Q3)("",!0)])),_:2},1032,["class"])))),128))])),_:1})}}}),Q=(0,M.A)(F,[["__scopeId","data-v-ccef6266"]]);var U=n(8309);function z({multiple:e,accept:t,onChange:n,onError:r}={}){const o=(0,a.KR)(null);let i;return typeof document<"u"&&(i=document.createElement("input"),i.type="file",i.onchange=e=>{const t=e.target;o.value=t.files,null==n||n(o.value)},i.onerror=()=>null==r?void 0:r(),i.multiple=e,i.accept=t),{files:(0,a.tB)(o),open:()=>{if(!i)return null==r?void 0:r();i.click()}}}const j={key:0,class:"border-t-1/2 text-c-3 flex min-h-10 w-full items-center justify-center p-2 text-sm"},W={key:1,class:"border-t flex items-center justify-center p-1.5 overflow-hidden"},X={class:"text-c-2 text-xs w-full border rounded py-1 px-1.5 max-w-full overflow-hidden whitespace-nowrap"},q=(0,r.pM)({__name:"RequestBody",props:{example:{},operation:{},environment:{},envVariables:{},workspace:{},title:{}},setup(e){const{requestExampleMutators:t}=(0,U.hA)(),n={json:"json",xml:"xml",yaml:"yaml",edn:"edn",other:"html"},i=Object.entries({multipartForm:"Multipart Form",formUrlEncoded:"Form URL Encoded",binaryFile:"Binary File",json:"JSON",xml:"XML",yaml:"YAML",edn:"EDN",other:"Other",none:"None"}).map((([e,t])=>({id:e,label:t}))),s=(0,r.EW)((()=>{var t;const{activeBody:n,formData:r,raw:a}=e.example.body;return"formData"===n?"urlencoded"===(null==r?void 0:r.encoding)?"formUrlEncoded":"multipartForm":"binary"===n?"binaryFile":"raw"===n&&null!=a&&a.encoding?"html"===a.encoding?"other":a.encoding:Object.keys((null==(t=e.operation.requestBody)?void 0:t.content)||{})[0]||"none"})),l=(0,r.EW)({get:()=>i.find((e=>e.id===s.value))??i[i.length-1]??i[0],set:e=>{null!=e&&e.id&&C(e.id)}}),c=(0,a.KR)(null),h=(0,r.EW)((()=>{var e;const t=null==(e=l.value)?void 0:e.id;return n[t]??"plaintext"}));function y(){console.log("deleteRow")}const w=(n,a,o)=>{var i,s,l,u,d,p,h;const f=O.value;if(f.length>n){const r=[...f];r[n]={...r[n],value:(null==(i=r[n])?void 0:i.value)||"",key:(null==(s=r[n])?void 0:s.key)||"",enabled:(null==(l=r[n])?void 0:l.enabled)??!1,[a]:o||""},(""!==(null==(u=r[n])?void 0:u.key)||""!==(null==(d=r[n])?void 0:d.value))&&(r[n].enabled=!0),""===(null==(p=r[n])?void 0:p.key)&&""===(null==(h=r[n])?void 0:h.value)&&r.splice(n,1),t.edit(e.example.uid,"body.formData.value",r)}else{const n=[v.vA.parse({[a]:o})];t.edit(e.example.uid,"body.formData.value",n),(0,r.dY)((()=>{var e;c.value&&(null==(e=c.value.querySelectorAll("input")["key"===a?0:1])||e.focus())}))}n===f.length-1&&x()},O=(0,r.EW)((()=>{var t;return(null==(t=e.example.body.formData)?void 0:t.value)??[]})),k=()=>{const e=O.value[O.value.length-1];(!e||""!==e.key||""!==e.value)&&x()},x=()=>{const n=v.vA.parse({enabled:!1}),r=[...O.value,n];e.example.body.formData?t.edit(e.example.uid,"body.formData.value",r):t.edit(e.example.uid,"body.formData",{value:r,encoding:"form-data"})},E=(n,r)=>{const a=O.value;if(a.length>n){const o=[...a];o[n]&&(o[n].enabled=r),t.edit(e.example.uid,"body.formData.value",o)}},S=n=>t.edit(e.example.uid,"body.raw.value",n),_=e=>"multipartForm"===e?{activeBody:"formData",encoding:"form-data",header:"multipart/form-data"}:"formUrlEncoded"===e?{activeBody:"formData",encoding:"urlencoded",header:"application/x-www-form-urlencoded"}:"binaryFile"===e?{activeBody:"binary",encoding:void 0,header:"application/octet-stream"}:"json"===e?{activeBody:"raw",encoding:"json",header:"application/json"}:"xml"===e?{activeBody:"raw",encoding:"xml",header:"application/xml"}:"yaml"===e?{activeBody:"raw",encoding:"yaml",header:"application/yaml"}:"edn"===e?{activeBody:"raw",encoding:"edn",header:"application/edn"}:"other"===e?{activeBody:"raw",encoding:"html",header:"application/html"}:{activeBody:"raw",encoding:void 0,header:void 0},C=n=>{var r,a;const{activeBody:o,encoding:i,header:s}=_(n);if(t.edit(e.example.uid,"body.activeBody",o),i&&"raw"===o)t.edit(e.example.uid,"body.raw",{encoding:i,value:(null==(r=e.example.body.raw)?void 0:r.value)??""});else if(i&&"formData"===o)t.edit(e.example.uid,"body.formData",{encoding:i,value:(null==(a=e.example.body.formData)?void 0:a.value)??[]});else if(!i&&"binary"!==o){const{raw:n,...r}=e.example.body;t.edit(e.example.uid,"body",r)}const l=[...e.example.parameters.headers],c=l.findIndex((e=>"content-type"===e.key.toLowerCase()));if(c>=0)s&&l[c]?l[c].value=s:l[c]&&l.splice(c,1);else if(s){const e=l[l.length-1];e&&""===e.key&&""===e.value?l.splice(l.length-1,0,{key:"Content-Type",value:s,enabled:!0}):l.push({key:"Content-Type",value:s,enabled:!0})}t.edit(e.example.uid,"parameters.headers",l)},T=async n=>{const{open:r}=z({onChange:async r=>{var a,o;const i=null==r?void 0:r[0];if(i){const r=[...O.value];r[n]={...r[n],file:i,value:(null==(a=r[n])?void 0:a.value)||i.name,key:(null==(o=r[n])?void 0:o.key)||i.name,enabled:!0},t.edit(e.example.uid,"body.formData.value",r)}},multiple:!1,accept:"*/*"});r()},A=()=>t.edit(e.example.uid,"body.binary",void 0);function R(n){const r=[...O.value];r[n]={...r[n],file:void 0,value:"",key:"",enabled:!1},t.edit(e.example.uid,"body.formData.value",r)}function P(){const{open:n}=z({onChange:async n=>{const r=null==n?void 0:n[0];r&&t.edit(e.example.uid,"body.binary",r)},multiple:!1,accept:"*/*"});n()}(0,r.wB)(l,(e=>{["multipartForm","formUrlEncoded"].includes((null==e?void 0:e.id)||"")&&k()}),{immediate:!0}),(0,r.wB)((()=>e.example.uid),(()=>{e.operation.method&&(0,b.Ik)(e.operation.method)&&C(s.value),["multipartForm","formUrlEncoded"].includes(s.value)&&k()}),{immediate:!0});const D=(0,r.EW)((()=>{var t,n,r;const a=null==(t=l.value)?void 0:t.id,{header:o}=_(a),i=(null==(n=e.operation.requestBody)?void 0:n.content)||{},s=o&&(null==(r=i[o])?void 0:r.examples)||{};return Object.entries(s).map((([e,t])=>({id:e,label:e,value:t})))})),M=(0,r.EW)({get:()=>{var t;const n=(null==(t=e.example.body.raw)?void 0:t.value)??"{}",r=JSON.parse(n);return D.value.find((e=>{const t=e.value;return JSON.stringify(t.value)===JSON.stringify(r)}))??D.value[0]},set:e=>{if(null!=e&&e.id){const t=D.value.find((t=>t.id===e.id));if(t){const e=t.value;S(JSON.stringify(e.value,null,2))}}}});return(e,t)=>((0,r.uX)(),(0,r.Wv)(g.A,null,{title:(0,r.k6)((()=>[(0,r.eW)((0,o.v_)(e.title),1)])),default:(0,r.k6)((()=>[(0,r.bF)(d.A,{columns:[""]},{default:(0,r.k6)((()=>[(0,r.bF)(m.A,null,{default:(0,r.k6)((()=>[(0,r.bF)(f,{class:"relative col-span-full flex h-8 cursor-pointer items-center justify-between !p-0"},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(p.cy),{modelValue:l.value,"onUpdate:modelValue":t[0]||(t[0]=e=>l.value=e),options:(0,a.R1)(i),teleport:""},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(p.RO),{class:"flex gap-1.5 h-full px-3 text-c-2 font-normal hover:text-c-1 w-fit",fullWidth:"",variant:"ghost"},{default:(0,r.k6)((()=>{var e;return[(0,r.Lk)("span",null,(0,o.v_)(null==(e=l.value)?void 0:e.label),1),(0,r.bF)((0,a.R1)(p.kv),{icon:"ChevronDown",size:"md"})]})),_:1})])),_:1},8,["modelValue","options"]),D.value.length>0?((0,r.uX)(),(0,r.Wv)((0,a.R1)(p.cy),{key:0,modelValue:M.value,"onUpdate:modelValue":t[1]||(t[1]=e=>M.value=e),options:D.value,side:"left",teleport:""},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(p.RO),{class:"flex gap-1.5 h-full px-2 text-c-2 font-normal hover:text-c-1 w-fit",fullWidth:"",variant:"ghost"},{default:(0,r.k6)((()=>{var e;return[(0,r.Lk)("span",null,(0,o.v_)(null==(e=M.value)?void 0:e.label),1),(0,r.bF)((0,a.R1)(p.kv),{icon:"ChevronDown",size:"md"})]})),_:1})])),_:1},8,["modelValue","options"])):(0,r.Q3)("",!0)])),_:1})])),_:1}),(0,r.bF)(m.A,null,{default:(0,r.k6)((()=>{var n,i,s,d,f,m;return["none"===(null==(n=l.value)?void 0:n.id)?((0,r.uX)(),(0,r.CE)("div",j,t[2]||(t[2]=[(0,r.Lk)("span",null,"No Body",-1)]))):"binaryFile"===(null==(i=l.value)?void 0:i.id)?((0,r.uX)(),(0,r.CE)("div",W,[e.example.body.binary?((0,r.uX)(),(0,r.CE)(r.FK,{key:0},[(0,r.Lk)("span",X,(0,o.v_)(e.example.body.binary.name),1),(0,r.bF)((0,a.R1)(p.RO),{class:"bg-b-2 hover:bg-b-3 border-0 text-c-2 ml-1 shadow-none",size:"sm",variant:"outlined",onClick:A},{default:(0,r.k6)((()=>t[3]||(t[3]=[(0,r.eW)(" Delete ")]))),_:1})],64)):((0,r.uX)(),(0,r.Wv)((0,a.R1)(p.RO),{key:1,class:"bg-b-2 hover:bg-b-3 border-0 text-c-2 shadow-none",size:"sm",variant:"outlined",onClick:P},{default:(0,r.k6)((()=>[t[4]||(t[4]=(0,r.Lk)("span",null,"Upload File",-1)),(0,r.bF)((0,a.R1)(p.kv),{class:"ml-1",icon:"UploadSimple",size:"xs",thickness:"2.5"})])),_:1}))])):"multipartForm"==(null==(s=l.value)?void 0:s.id)?((0,r.uX)(),(0,r.Wv)(Q,{key:2,ref_key:"tableWrapperRef",ref:c,class:"!m-0 rounded-t-none shadow-none border-l-0 border-r-0 border-t-0 border-b-0",columns:["32px","","","61px"],envVariables:e.envVariables,environment:e.environment,items:O.value,showUploadButton:"",workspace:e.workspace,onDeleteRow:y,onRemoveFile:R,onToggleRow:E,onUpdateRow:w,onUploadFile:T},null,8,["envVariables","environment","items","workspace"])):"formUrlEncoded"==(null==(d=l.value)?void 0:d.id)?((0,r.uX)(),(0,r.Wv)(Q,{key:3,ref_key:"tableWrapperRef",ref:c,class:"!m-0 rounded-t-none shadow-none border-l-0 border-r-0 border-t-0 border-b-0",columns:["32px","","","61px"],envVariables:e.envVariables,environment:e.environment,items:O.value,showUploadButton:"",workspace:e.workspace,onDeleteRow:y,onRemoveFile:R,onToggleRow:E,onUpdateRow:w,onUploadFile:T},null,8,["envVariables","environment","items","workspace"])):((0,r.uX)(),(0,r.Wv)(u.A,{key:4,class:"border-t-1/2 px-1",content:"",envVariables:e.envVariables,environment:e.environment,language:h.value,lineNumbers:"",lint:"",modelValue:(null==(m=null==(f=e.example.body)?void 0:f.raw)?void 0:m.value)??"",workspace:e.workspace,"onUpdate:modelValue":S},null,8,["envVariables","environment","language","modelValue","workspace"]))]})),_:1}),(0,r.bF)(m.A)])),_:1})])),_:1}))}}),V=(0,M.A)(q,[["__scopeId","data-v-3f1cd888"]]),H={class:"text-c-2 flex whitespace-nowrap opacity-0 group-hover/params:opacity-100 has-[:focus-visible]:opacity-100 request-meta-buttons"},Z={class:"sr-only"},G=(0,r.pM)({__name:"RequestParams",props:{example:{},environment:{},envVariables:{},workspace:{},title:{},paramKey:{},readOnlyEntries:{default:()=>[]}},setup(e){const{requestExampleMutators:t}=(0,U.hA)(),n=(0,r.EW)((()=>e.example.parameters[e.paramKey]??[]));(0,r.sV)((()=>h()));const s=()=>{const r=v.vA.parse({enabled:!1}),a=[...n.value,r];t.edit(e.example.uid,`parameters.${e.paramKey}`,a)},l=(0,a.KR)(null),c=(a,o,i)=>{const c=n.value;if(c.length>a){const n=[...c];if(!n[a])return;n[a]={...n[a],[o]:i},(""!==n[a].key||""!==n[a].value)&&(n[a].enabled=!0),""===n[a].key&&""===n[a].value&&n.splice(a,1),t.edit(e.example.uid,`parameters.${e.paramKey}`,n)}else{const n=[v.vA.parse({[o]:i})];t.edit(e.example.uid,`parameters.${e.paramKey}`,n),(0,r.dY)((()=>{var e;l.value&&(null==(e=l.value.querySelectorAll("input")["key"===o?0:1])||e.focus())}))}a===c.length-1&&s()},u=(n,r)=>t.edit(e.example.uid,`parameters.${e.paramKey}.${n}.enabled`,r),d=()=>{const a=n.value.filter((e=>e.required));t.edit(e.example.uid,`parameters.${e.paramKey}`,a),(0,r.dY)((()=>s()))};function h(){if(0===n.value.length)s();else if(n.value.length>=1){const e=n.value[n.value.length-1];e&&""!==e.key&&""!==e.value&&s()}}const f=(0,r.EW)((()=>n.value.filter((e=>e.key||e.value)).length)),m=(0,r.EW)((()=>n.value.length>1));(0,r.wB)((()=>e.example),((e,t)=>{e!==t&&h()}),{immediate:!0});const b=(0,r.EW)((()=>(e.readOnlyEntries??[]).length>0));return(e,t)=>((0,r.uX)(),(0,r.Wv)(g.A,{class:"group/params",itemCount:f.value},{title:(0,r.k6)((()=>[(0,r.eW)((0,o.v_)(e.title),1)])),actions:(0,r.k6)((()=>[(0,r.Lk)("div",H,[m.value?((0,r.uX)(),(0,r.Wv)((0,a.R1)(p.iy),{key:0,side:"right",sideOffset:12},{trigger:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(p.RO),{class:"px-1 transition-none",size:"sm",variant:"ghost",onClick:(0,i.D$)(d,["stop"])},{default:(0,r.k6)((()=>[t[0]||(t[0]=(0,r.eW)(" Clear ")),(0,r.Lk)("span",Z,"All "+(0,o.v_)(e.title),1)])),_:1})])),content:(0,r.k6)((()=>t[1]||(t[1]=[(0,r.Lk)("div",{class:"grid gap-1.5 pointer-events-none min-w-48 w-content shadow-lg rounded bg-b-1 p-2 text-xxs leading-5 z-10 text-c-1"},[(0,r.Lk)("div",{class:"flex items-center text-c-2"},[(0,r.Lk)("span",null,"Clear optional parameters")])],-1)]))),_:1})):(0,r.Q3)("",!0)])])),default:(0,r.k6)((()=>[(0,r.Lk)("div",{ref_key:"tableWrapperRef",ref:l},[b.value?((0,r.uX)(),(0,r.Wv)(Q,{key:0,class:(0,o.C4)(["flex-1",{"bg-mix-transparent bg-mix-amount-95 bg-c-3":b.value}]),columns:["32px","",""],envVariables:e.envVariables,environment:e.environment,isGlobal:"",isReadOnly:"",items:e.readOnlyEntries,workspace:e.workspace},null,8,["class","envVariables","environment","items","workspace"])):(0,r.Q3)("",!0),(0,r.bF)(Q,{class:"flex-1",columns:["32px","",""],envVariables:e.envVariables,environment:e.environment,items:n.value,workspace:e.workspace,onToggleRow:u,onUpdateRow:c},null,8,["envVariables","environment","items","workspace"])],512)])),_:1},8,["itemCount"]))}}),K=(0,r.pM)({__name:"RequestPathParams",props:{example:{},operation:{},paramKey:{},title:{},environment:{},envVariables:{},workspace:{}},setup(e){const{requestMutators:t,requestExampleMutators:n}=(0,U.hA)(),a=(0,r.EW)((()=>e.example.parameters[e.paramKey].map((e=>({...e,enum:e.enum}))))),i=(r,a,o)=>{var i,s;const l=e.example.parameters[e.paramKey],c=null==(i=l[r])?void 0:i.key;if(c){if("key"===a){if(null!=(s=l[r])&&s.required)return;if(o){const n=encodeURIComponent(c),r=encodeURIComponent(o),a=new RegExp(`(?<=/):${n}(?=[/?#]|$)`,"g"),i=e.operation.path.replace(a,`:${r}`);t.edit(e.operation.uid,"path",i)}else{l.splice(r,1);const n=new RegExp(`/:${encodeURIComponent(c)}(?=[/?#]|$)`,"g"),a=e.operation.path.replace(n,"");t.edit(e.operation.uid,"path",a)}}n.edit(e.example.uid,`parameters.${e.paramKey}.${r}.${a}`,o)}};return(0,r.wB)((()=>e.operation.path),(t=>{t&&(t=>{t&&(t=>{var r;const a=(null==(r=t.match(b.QL.PATH))?void 0:r.map((e=>e.slice(1,-1))))||[],o=e.example.parameters[e.paramKey],i=new Map(o.map((e=>[e.key,e]))),s=a.map((e=>i.get(e)||{key:e,value:"",enabled:!0}));o.forEach((e=>{!a.includes(e.key)&&(e.value||e.required)&&s.push(e)})),o.splice(0,o.length,...s),n.edit(e.example.uid,`parameters.${e.paramKey}`,o)})(t)})(t)})),(e,t)=>((0,r.uX)(),(0,r.Wv)(g.A,{itemCount:a.value.length},{title:(0,r.k6)((()=>[(0,r.eW)((0,o.v_)(e.title),1)])),default:(0,r.k6)((()=>[a.value.length?((0,r.uX)(),(0,r.Wv)(Q,{key:0,class:"flex-1",columns:["32px","",""],envVariables:e.envVariables,environment:e.environment,items:a.value,workspace:e.workspace,onUpdateRow:i},null,8,["envVariables","environment","items","workspace"])):(0,r.Q3)("",!0)])),_:1},8,["itemCount"]))}});var Y=n(1527),J=n(7786),ee=n(2816),te=n(3813);const ne={class:"w-full"},re={class:"flex flex-1 -mx-1"},ae={class:"bg-b-1 border-t flex items-center justify-center overflow-hidden"},oe=(0,r.pM)({__name:"RequestCodeExample",props:{collection:{},example:{},operation:{},server:{},workspace:{}},setup(e){const{securitySchemes:t,workspaceMutators:n}=(0,U.hA)(),i=(0,r.EW)((()=>(0,ee.z)(e.operation.security||e.collection.security||[],e.collection.selectedSecuritySchemeUids,t))),s=(0,r.EW)((()=>{const e={};return{options:(0,J.E)().clients().map((t=>({label:t.title,options:t.clients.map((n=>(e[`${t.key},${n.client}`]=n.title,{id:`${t.key},${n.client}`,label:n.title})))}))),dict:e}})),l=(0,r.EW)((()=>{var t;const n=e.workspace.selectedHttpClient;if(!n)return(null==(t=s.value.options[0])?void 0:t.options[0])??{id:"js,fetch",label:"Fetch"};const r=`${n.targetKey},${n.clientKey}`;return{id:r,label:s.value.dict[r]??"Unknown"}})),c=(0,r.EW)((()=>{var t;return(null==(t=e.workspace.selectedHttpClient)?void 0:t.targetKey)??"js"})),u=(0,r.EW)((()=>{var t;return(null==(t=e.workspace.selectedHttpClient)?void 0:t.clientKey)??"fetch"})),h=({id:t})=>{const[r,a]=t.split(",");!r||!a||n.edit(e.workspace.uid,"selectedHttpClient",{targetKey:r,clientKey:a})};return(e,t)=>((0,r.uX)(),(0,r.CE)("div",ne,[(0,r.bF)(g.A,{class:"group/preview -mt-0.25 w-full border-b-0",defaultOpen:!1,hasIcon:!1},{title:(0,r.k6)((()=>t[0]||(t[0]=[(0,r.eW)("Code Snippet")]))),actions:(0,r.k6)((()=>[(0,r.Lk)("div",re,[(0,r.bF)((0,a.R1)(p.ev),{modelValue:l.value,options:s.value.options,placement:"bottom-end","onUpdate:modelValue":h},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(p.RO),{class:"flex gap-1.5 h-full px-1.5 py-0.75 font-normal text-c-1 w-fit hover:bg-b-3",fullWidth:"",variant:"ghost"},{default:(0,r.k6)((()=>{var e;return[(0,r.Lk)("span",null,(0,o.v_)(null==(e=l.value)?void 0:e.label),1),(0,r.bF)((0,a.R1)(p.kv),{icon:"ChevronDown",size:"md"})]})),_:1})])),_:1},8,["modelValue","options"])])])),default:(0,r.k6)((()=>[(0,r.bF)(d.A,{columns:[""]},{default:(0,r.k6)((()=>[(0,r.bF)(m.A,null,{default:(0,r.k6)((()=>[(0,r.Lk)("div",ae,[(0,r.bF)((0,a.R1)(te.A),{class:"px-3 py-1.5",client:u.value,example:e.example,operation:e.operation,securitySchemes:i.value,server:e.server,target:c.value},null,8,["client","example","operation","securitySchemes","server","target"])])])),_:1})])),_:1})])),_:1})]))}}),ie=(0,M.A)(oe,[["__scopeId","data-v-c2f2c62b"]]);var se=n(6584);const le={class:"flex-1 flex gap-1 items-center lg:pr-24 pointer-events-none group"},ce={key:0,class:"absolute w-full h-full top-0 left-0 pointer-events-auto opacity-0 cursor-text",for:"requestname"},ue=["value"],de={key:2,class:"flex items-center text-c-1 h-8"},pe={class:"request-section-content custom-scroll flex flex-1 flex-col relative"},he=(0,r.pM)({__name:"RequestSection",props:{collection:{},environment:{},envVariables:{},example:{},operation:{},selectedSecuritySchemeUids:{},server:{},workspace:{}},setup(e){const{requestMutators:t,cookies:n,securitySchemes:u}=(0,U.hA)(),{layout:d}=(0,se.g)(),h=(0,r.EW)((()=>{const t=new Set(["All","Query","Auth","Variables","Cookies","Headers","Body"]);return e.example.parameters.path.length||t.delete("Variables"),(0,b.Ik)(e.operation.method??"get")||t.delete("Body"),f.value&&t.delete("Auth"),[...t]})),f=(0,r.EW)((()=>{var t;return"modal"===d&&0===(null==(t=e.operation.security)?void 0:t.length)})),m=(0,a.KR)("All");(0,r.wB)((()=>e.operation),(e=>{"Body"===m.value&&e&&!(0,b.Ik)(e.method)&&(m.value="All")}));const g=n=>{const r=n.target;t.edit(e.operation.uid,"summary",r.value)},v=(0,r.EW)((()=>(e.workspace.cookies??[]).map((e=>n[e])).filter(b.O9).filter((e=>e.name)).filter((t=>{var n;return(0,c.Kt)((null==(n=e.server)?void 0:n.url)||e.operation.path,t.domain)})).map((e=>({key:e.name,value:e.value,route:{name:"cookies",params:{cookies:e.uid}},enabled:!0})))));return(e,t)=>((0,r.uX)(),(0,r.Wv)(l.A,{"aria-label":`Request: ${e.operation.summary}`},{title:(0,r.k6)((()=>[(0,r.Lk)("div",le,["modal"!==(0,a.R1)(d)?((0,r.uX)(),(0,r.CE)("label",ce)):(0,r.Q3)("",!0),"modal"!==(0,a.R1)(d)?((0,r.uX)(),(0,r.CE)("input",{key:1,id:"requestname",class:"text-c-1 rounded pointer-events-auto relative w-full pl-1.25 -ml-0.5 md:-ml-1.25 h-8 group-hover-input has-[:focus-visible]:outline z-10",placeholder:"Request Name",value:e.operation.summary,onInput:g},null,40,ue)):((0,r.uX)(),(0,r.CE)("span",de,(0,o.v_)(e.operation.summary),1))]),(0,r.bF)(s.A,{activeSection:m.value,sections:h.value,onSetActiveSection:t[0]||(t[0]=e=>m.value=e)},null,8,["activeSection","sections"])])),default:(0,r.k6)((()=>[(0,r.Lk)("div",pe,[e.collection&&e.workspace&&("modal"!==(0,a.R1)(d)||Object.keys((0,a.R1)(u)??{}).length)?(0,r.bo)(((0,r.uX)(),(0,r.Wv)(Y.A,{key:0,collection:e.collection,envVariables:e.envVariables,environment:e.environment,layout:"client",operation:e.operation,selectedSecuritySchemeUids:e.selectedSecuritySchemeUids,server:e.server,title:"Authentication",workspace:e.workspace},null,8,["collection","envVariables","environment","operation","selectedSecuritySchemeUids","server","workspace"])),[[i.aG,!f.value&&("All"===m.value||"Auth"===m.value)]]):(0,r.Q3)("",!0),(0,r.bo)((0,r.bF)(K,{envVariables:e.envVariables,environment:e.environment,example:e.example,operation:e.operation,paramKey:"path",title:"Variables",workspace:e.workspace},null,8,["envVariables","environment","example","operation","workspace"]),[[i.aG,("All"===m.value||"Variables"===m.value)&&e.example.parameters.path.length]]),(0,r.bo)((0,r.bF)(G,{envVariables:e.envVariables,environment:e.environment,example:e.example,operation:e.operation,paramKey:"cookies",readOnlyEntries:v.value,title:"Cookies",workspace:e.workspace,workspaceParamKey:"cookies"},null,8,["envVariables","environment","example","operation","readOnlyEntries","workspace"]),[[i.aG,"All"===m.value||"Cookies"===m.value]]),(0,r.bo)((0,r.bF)(G,{envVariables:e.envVariables,environment:e.environment,example:e.example,operation:e.operation,paramKey:"headers",title:"Headers",workspace:e.workspace},null,8,["envVariables","environment","example","operation","workspace"]),[[i.aG,"All"===m.value||"Headers"===m.value]]),(0,r.bo)((0,r.bF)(G,{envVariables:e.envVariables,environment:e.environment,example:e.example,operation:e.operation,paramKey:"query",title:"Query Parameters",workspace:e.workspace},null,8,["envVariables","environment","example","operation","workspace"]),[[i.aG,"All"===m.value||"Query"===m.value]]),(0,r.bo)((0,r.bF)(V,{envVariables:e.envVariables,environment:e.environment,example:e.example,operation:e.operation,title:"Body",workspace:e.workspace},null,8,["envVariables","environment","example","operation","workspace"]),[[i.aG,e.operation.method&&("All"===m.value||"Body"===m.value)&&(0,a.R1)(b.Ik)(e.operation.method)]]),t[1]||(t[1]=(0,r.Lk)("div",{class:"flex flex-grow"},null,-1)),(0,r.bF)((0,a.R1)(p.XO),null,{default:(0,r.k6)((()=>[(0,r.bF)(ie,{collection:e.collection,example:e.example,operation:e.operation,server:e.server,workspace:e.workspace},null,8,["collection","example","operation","server","workspace"])])),_:1})])])),_:1},8,["aria-label"]))}}),fe=(0,M.A)(he,[["__scopeId","data-v-5ab558f8"]])},7526:(e,t,n)=>{"use strict";n.d(t,{n:()=>he});const r=["a","abbr","address","article","aside","audio","b","blockquote","body","button","canvas","caption","cite","code","dd","del","details","dfn","div","dl","dt","em","fieldset","figcaption","figure","footer","form","h1","h2","h3","h4","h5","h6","header","hgroup","html","i","iframe","img","input","ins","kbd","label","legend","li","main","mark","menu","nav","object","ol","optgroup","option","p","picture","q","quote","samp","section","select","source","span","strong","summary","sup","table","tbody","td","textarea","tfoot","th","thead","time","tr","ul","var","video","defs","g","marker","mask","pattern","svg","switch","symbol","feBlend","feColorMatrix","feComponentTransfer","feComposite","feConvolveMatrix","feDiffuseLighting","feDisplacementMap","feFlood","feGaussianBlur","feImage","feMerge","feMorphology","feOffset","feSpecularLighting","feTile","feTurbulence","linearGradient","radialGradient","stop","circle","ellipse","image","line","path","polygon","polyline","rect","text","use","textPath","tspan","foreignObject","clipPath"],a=["any-hover","any-pointer","aspect-ratio","color","color-gamut","color-index","device-aspect-ratio","device-height","device-width","display-mode","forced-colors","grid","height","hover","inverted-colors","monochrome","orientation","overflow-block","overflow-inline","pointer","prefers-color-scheme","prefers-contrast","prefers-reduced-motion","prefers-reduced-transparency","resolution","scan","scripting","update","width","min-width","max-width","min-height","max-height"].sort().reverse(),o=["active","any-link","blank","checked","current","default","defined","dir","disabled","drop","empty","enabled","first","first-child","first-of-type","fullscreen","future","focus","focus-visible","focus-within","has","host","host-context","hover","indeterminate","in-range","invalid","is","lang","last-child","last-of-type","left","link","local-link","not","nth-child","nth-col","nth-last-child","nth-last-col","nth-last-of-type","nth-of-type","only-child","only-of-type","optional","out-of-range","past","placeholder-shown","read-only","read-write","required","right","root","scope","target","target-within","user-invalid","valid","visited","where"].sort().reverse(),i=["after","backdrop","before","cue","cue-region","first-letter","first-line","grammar-error","marker","part","placeholder","selection","slotted","spelling-error"].sort().reverse(),s=["accent-color","align-content","align-items","align-self","alignment-baseline","all","anchor-name","animation","animation-composition","animation-delay","animation-direction","animation-duration","animation-fill-mode","animation-iteration-count","animation-name","animation-play-state","animation-range","animation-range-end","animation-range-start","animation-timeline","animation-timing-function","appearance","aspect-ratio","backdrop-filter","backface-visibility","background","background-attachment","background-blend-mode","background-clip","background-color","background-image","background-origin","background-position","background-position-x","background-position-y","background-repeat","background-size","baseline-shift","block-size","border","border-block","border-block-color","border-block-end","border-block-end-color","border-block-end-style","border-block-end-width","border-block-start","border-block-start-color","border-block-start-style","border-block-start-width","border-block-style","border-block-width","border-bottom","border-bottom-color","border-bottom-left-radius","border-bottom-right-radius","border-bottom-style","border-bottom-width","border-collapse","border-color","border-end-end-radius","border-end-start-radius","border-image","border-image-outset","border-image-repeat","border-image-slice","border-image-source","border-image-width","border-inline","border-inline-color","border-inline-end","border-inline-end-color","border-inline-end-style","border-inline-end-width","border-inline-start","border-inline-start-color","border-inline-start-style","border-inline-start-width","border-inline-style","border-inline-width","border-left","border-left-color","border-left-style","border-left-width","border-radius","border-right","border-right-color","border-right-style","border-right-width","border-spacing","border-start-end-radius","border-start-start-radius","border-style","border-top","border-top-color","border-top-left-radius","border-top-right-radius","border-top-style","border-top-width","border-width","bottom","box-align","box-decoration-break","box-direction","box-flex","box-flex-group","box-lines","box-ordinal-group","box-orient","box-pack","box-shadow","box-sizing","break-after","break-before","break-inside","caption-side","caret-color","clear","clip","clip-path","clip-rule","color","color-interpolation","color-interpolation-filters","color-profile","color-rendering","color-scheme","column-count","column-fill","column-gap","column-rule","column-rule-color","column-rule-style","column-rule-width","column-span","column-width","columns","contain","contain-intrinsic-block-size","contain-intrinsic-height","contain-intrinsic-inline-size","contain-intrinsic-size","contain-intrinsic-width","container","container-name","container-type","content","content-visibility","counter-increment","counter-reset","counter-set","cue","cue-after","cue-before","cursor","cx","cy","direction","display","dominant-baseline","empty-cells","enable-background","field-sizing","fill","fill-opacity","fill-rule","filter","flex","flex-basis","flex-direction","flex-flow","flex-grow","flex-shrink","flex-wrap","float","flood-color","flood-opacity","flow","font","font-display","font-family","font-feature-settings","font-kerning","font-language-override","font-optical-sizing","font-palette","font-size","font-size-adjust","font-smooth","font-smoothing","font-stretch","font-style","font-synthesis","font-synthesis-position","font-synthesis-small-caps","font-synthesis-style","font-synthesis-weight","font-variant","font-variant-alternates","font-variant-caps","font-variant-east-asian","font-variant-emoji","font-variant-ligatures","font-variant-numeric","font-variant-position","font-variation-settings","font-weight","forced-color-adjust","gap","glyph-orientation-horizontal","glyph-orientation-vertical","grid","grid-area","grid-auto-columns","grid-auto-flow","grid-auto-rows","grid-column","grid-column-end","grid-column-start","grid-gap","grid-row","grid-row-end","grid-row-start","grid-template","grid-template-areas","grid-template-columns","grid-template-rows","hanging-punctuation","height","hyphenate-character","hyphenate-limit-chars","hyphens","icon","image-orientation","image-rendering","image-resolution","ime-mode","initial-letter","initial-letter-align","inline-size","inset","inset-area","inset-block","inset-block-end","inset-block-start","inset-inline","inset-inline-end","inset-inline-start","isolation","justify-content","justify-items","justify-self","kerning","left","letter-spacing","lighting-color","line-break","line-height","line-height-step","list-style","list-style-image","list-style-position","list-style-type","margin","margin-block","margin-block-end","margin-block-start","margin-bottom","margin-inline","margin-inline-end","margin-inline-start","margin-left","margin-right","margin-top","margin-trim","marker","marker-end","marker-mid","marker-start","marks","mask","mask-border","mask-border-mode","mask-border-outset","mask-border-repeat","mask-border-slice","mask-border-source","mask-border-width","mask-clip","mask-composite","mask-image","mask-mode","mask-origin","mask-position","mask-repeat","mask-size","mask-type","masonry-auto-flow","math-depth","math-shift","math-style","max-block-size","max-height","max-inline-size","max-width","min-block-size","min-height","min-inline-size","min-width","mix-blend-mode","nav-down","nav-index","nav-left","nav-right","nav-up","none","normal","object-fit","object-position","offset","offset-anchor","offset-distance","offset-path","offset-position","offset-rotate","opacity","order","orphans","outline","outline-color","outline-offset","outline-style","outline-width","overflow","overflow-anchor","overflow-block","overflow-clip-margin","overflow-inline","overflow-wrap","overflow-x","overflow-y","overlay","overscroll-behavior","overscroll-behavior-block","overscroll-behavior-inline","overscroll-behavior-x","overscroll-behavior-y","padding","padding-block","padding-block-end","padding-block-start","padding-bottom","padding-inline","padding-inline-end","padding-inline-start","padding-left","padding-right","padding-top","page","page-break-after","page-break-before","page-break-inside","paint-order","pause","pause-after","pause-before","perspective","perspective-origin","place-content","place-items","place-self","pointer-events","position","position-anchor","position-visibility","print-color-adjust","quotes","r","resize","rest","rest-after","rest-before","right","rotate","row-gap","ruby-align","ruby-position","scale","scroll-behavior","scroll-margin","scroll-margin-block","scroll-margin-block-end","scroll-margin-block-start","scroll-margin-bottom","scroll-margin-inline","scroll-margin-inline-end","scroll-margin-inline-start","scroll-margin-left","scroll-margin-right","scroll-margin-top","scroll-padding","scroll-padding-block","scroll-padding-block-end","scroll-padding-block-start","scroll-padding-bottom","scroll-padding-inline","scroll-padding-inline-end","scroll-padding-inline-start","scroll-padding-left","scroll-padding-right","scroll-padding-top","scroll-snap-align","scroll-snap-stop","scroll-snap-type","scroll-timeline","scroll-timeline-axis","scroll-timeline-name","scrollbar-color","scrollbar-gutter","scrollbar-width","shape-image-threshold","shape-margin","shape-outside","shape-rendering","speak","speak-as","src","stop-color","stop-opacity","stroke","stroke-dasharray","stroke-dashoffset","stroke-linecap","stroke-linejoin","stroke-miterlimit","stroke-opacity","stroke-width","tab-size","table-layout","text-align","text-align-all","text-align-last","text-anchor","text-combine-upright","text-decoration","text-decoration-color","text-decoration-line","text-decoration-skip","text-decoration-skip-ink","text-decoration-style","text-decoration-thickness","text-emphasis","text-emphasis-color","text-emphasis-position","text-emphasis-style","text-indent","text-justify","text-orientation","text-overflow","text-rendering","text-shadow","text-size-adjust","text-transform","text-underline-offset","text-underline-position","text-wrap","text-wrap-mode","text-wrap-style","timeline-scope","top","touch-action","transform","transform-box","transform-origin","transform-style","transition","transition-behavior","transition-delay","transition-duration","transition-property","transition-timing-function","translate","unicode-bidi","user-modify","user-select","vector-effect","vertical-align","view-timeline","view-timeline-axis","view-timeline-inset","view-timeline-name","view-transition-name","visibility","voice-balance","voice-duration","voice-family","voice-pitch","voice-range","voice-rate","voice-stress","voice-volume","white-space","white-space-collapse","widows","width","will-change","word-break","word-spacing","word-wrap","writing-mode","x","y","z-index","zoom"].sort().reverse();var l="[0-9](_*[0-9])*",c=`\\.(${l})`,u="[0-9a-fA-F](_*[0-9a-fA-F])*",d={className:"number",variants:[{begin:`(\\b(${l})((${c})|\\.)?|(${c}))[eE][+-]?(${l})[fFdD]?\\b`},{begin:`\\b(${l})((${c})[fFdD]?\\b|\\.([fFdD]\\b)?)`},{begin:`(${c})[fFdD]?\\b`},{begin:`\\b(${l})[fFdD]\\b`},{begin:`\\b0[xX]((${u})\\.?|(${u})?\\.(${u}))[pP][+-]?(${l})[fFdD]?\\b`},{begin:"\\b(0|[1-9](_*[0-9])*)[lL]?\\b"},{begin:`\\b0[xX](${u})[lL]?\\b`},{begin:"\\b0(_*[0-7])*[lL]?\\b"},{begin:"\\b0[bB][01](_*[01])*[lL]?\\b"}],relevance:0};function p(e,t,n){return-1===n?"":e.replace(t,(r=>p(e,t,n-1)))}const h="[A-Za-z$_][0-9A-Za-z$_]*",f=["as","in","of","if","for","while","finally","var","new","function","do","return","void","else","break","catch","instanceof","with","throw","case","default","try","switch","continue","typeof","delete","let","yield","const","class","debugger","async","await","static","import","from","export","extends","using"],m=["true","false","null","undefined","NaN","Infinity"],g=["Object","Function","Boolean","Symbol","Math","Date","Number","BigInt","String","RegExp","Array","Float32Array","Float64Array","Int8Array","Uint8Array","Uint8ClampedArray","Int16Array","Int32Array","Uint16Array","Uint32Array","BigInt64Array","BigUint64Array","Set","Map","WeakSet","WeakMap","ArrayBuffer","SharedArrayBuffer","Atomics","DataView","JSON","Promise","Generator","GeneratorFunction","AsyncFunction","Reflect","Proxy","Intl","WebAssembly"],v=["Error","EvalError","InternalError","RangeError","ReferenceError","SyntaxError","TypeError","URIError"],b=["setInterval","setTimeout","clearInterval","clearTimeout","require","exports","eval","isFinite","isNaN","parseFloat","parseInt","decodeURI","decodeURIComponent","encodeURI","encodeURIComponent","escape","unescape"],y=["arguments","this","super","console","window","document","localStorage","sessionStorage","module","global"],w=[].concat(b,g,v);var O="[0-9](_*[0-9])*",k=`\\.(${O})`,x="[0-9a-fA-F](_*[0-9a-fA-F])*",E={className:"number",variants:[{begin:`(\\b(${O})((${k})|\\.)?|(${k}))[eE][+-]?(${O})[fFdD]?\\b`},{begin:`\\b(${O})((${k})[fFdD]?\\b|\\.([fFdD]\\b)?)`},{begin:`(${k})[fFdD]?\\b`},{begin:`\\b(${O})[fFdD]\\b`},{begin:`\\b0[xX]((${x})\\.?|(${x})?\\.(${x}))[pP][+-]?(${O})[fFdD]?\\b`},{begin:"\\b(0|[1-9](_*[0-9])*)[lL]?\\b"},{begin:`\\b0[xX](${x})[lL]?\\b`},{begin:"\\b0(_*[0-7])*[lL]?\\b"},{begin:"\\b0[bB][01](_*[01])*[lL]?\\b"}],relevance:0};const S=["a","abbr","address","article","aside","audio","b","blockquote","body","button","canvas","caption","cite","code","dd","del","details","dfn","div","dl","dt","em","fieldset","figcaption","figure","footer","form","h1","h2","h3","h4","h5","h6","header","hgroup","html","i","iframe","img","input","ins","kbd","label","legend","li","main","mark","menu","nav","object","ol","optgroup","option","p","picture","q","quote","samp","section","select","source","span","strong","summary","sup","table","tbody","td","textarea","tfoot","th","thead","time","tr","ul","var","video","defs","g","marker","mask","pattern","svg","switch","symbol","feBlend","feColorMatrix","feComponentTransfer","feComposite","feConvolveMatrix","feDiffuseLighting","feDisplacementMap","feFlood","feGaussianBlur","feImage","feMerge","feMorphology","feOffset","feSpecularLighting","feTile","feTurbulence","linearGradient","radialGradient","stop","circle","ellipse","image","line","path","polygon","polyline","rect","text","use","textPath","tspan","foreignObject","clipPath"],_=["any-hover","any-pointer","aspect-ratio","color","color-gamut","color-index","device-aspect-ratio","device-height","device-width","display-mode","forced-colors","grid","height","hover","inverted-colors","monochrome","orientation","overflow-block","overflow-inline","pointer","prefers-color-scheme","prefers-contrast","prefers-reduced-motion","prefers-reduced-transparency","resolution","scan","scripting","update","width","min-width","max-width","min-height","max-height"].sort().reverse(),C=["active","any-link","blank","checked","current","default","defined","dir","disabled","drop","empty","enabled","first","first-child","first-of-type","fullscreen","future","focus","focus-visible","focus-within","has","host","host-context","hover","indeterminate","in-range","invalid","is","lang","last-child","last-of-type","left","link","local-link","not","nth-child","nth-col","nth-last-child","nth-last-col","nth-last-of-type","nth-of-type","only-child","only-of-type","optional","out-of-range","past","placeholder-shown","read-only","read-write","required","right","root","scope","target","target-within","user-invalid","valid","visited","where"].sort().reverse(),T=["after","backdrop","before","cue","cue-region","first-letter","first-line","grammar-error","marker","part","placeholder","selection","slotted","spelling-error"].sort().reverse(),A=["accent-color","align-content","align-items","align-self","alignment-baseline","all","anchor-name","animation","animation-composition","animation-delay","animation-direction","animation-duration","animation-fill-mode","animation-iteration-count","animation-name","animation-play-state","animation-range","animation-range-end","animation-range-start","animation-timeline","animation-timing-function","appearance","aspect-ratio","backdrop-filter","backface-visibility","background","background-attachment","background-blend-mode","background-clip","background-color","background-image","background-origin","background-position","background-position-x","background-position-y","background-repeat","background-size","baseline-shift","block-size","border","border-block","border-block-color","border-block-end","border-block-end-color","border-block-end-style","border-block-end-width","border-block-start","border-block-start-color","border-block-start-style","border-block-start-width","border-block-style","border-block-width","border-bottom","border-bottom-color","border-bottom-left-radius","border-bottom-right-radius","border-bottom-style","border-bottom-width","border-collapse","border-color","border-end-end-radius","border-end-start-radius","border-image","border-image-outset","border-image-repeat","border-image-slice","border-image-source","border-image-width","border-inline","border-inline-color","border-inline-end","border-inline-end-color","border-inline-end-style","border-inline-end-width","border-inline-start","border-inline-start-color","border-inline-start-style","border-inline-start-width","border-inline-style","border-inline-width","border-left","border-left-color","border-left-style","border-left-width","border-radius","border-right","border-right-color","border-right-style","border-right-width","border-spacing","border-start-end-radius","border-start-start-radius","border-style","border-top","border-top-color","border-top-left-radius","border-top-right-radius","border-top-style","border-top-width","border-width","bottom","box-align","box-decoration-break","box-direction","box-flex","box-flex-group","box-lines","box-ordinal-group","box-orient","box-pack","box-shadow","box-sizing","break-after","break-before","break-inside","caption-side","caret-color","clear","clip","clip-path","clip-rule","color","color-interpolation","color-interpolation-filters","color-profile","color-rendering","color-scheme","column-count","column-fill","column-gap","column-rule","column-rule-color","column-rule-style","column-rule-width","column-span","column-width","columns","contain","contain-intrinsic-block-size","contain-intrinsic-height","contain-intrinsic-inline-size","contain-intrinsic-size","contain-intrinsic-width","container","container-name","container-type","content","content-visibility","counter-increment","counter-reset","counter-set","cue","cue-after","cue-before","cursor","cx","cy","direction","display","dominant-baseline","empty-cells","enable-background","field-sizing","fill","fill-opacity","fill-rule","filter","flex","flex-basis","flex-direction","flex-flow","flex-grow","flex-shrink","flex-wrap","float","flood-color","flood-opacity","flow","font","font-display","font-family","font-feature-settings","font-kerning","font-language-override","font-optical-sizing","font-palette","font-size","font-size-adjust","font-smooth","font-smoothing","font-stretch","font-style","font-synthesis","font-synthesis-position","font-synthesis-small-caps","font-synthesis-style","font-synthesis-weight","font-variant","font-variant-alternates","font-variant-caps","font-variant-east-asian","font-variant-emoji","font-variant-ligatures","font-variant-numeric","font-variant-position","font-variation-settings","font-weight","forced-color-adjust","gap","glyph-orientation-horizontal","glyph-orientation-vertical","grid","grid-area","grid-auto-columns","grid-auto-flow","grid-auto-rows","grid-column","grid-column-end","grid-column-start","grid-gap","grid-row","grid-row-end","grid-row-start","grid-template","grid-template-areas","grid-template-columns","grid-template-rows","hanging-punctuation","height","hyphenate-character","hyphenate-limit-chars","hyphens","icon","image-orientation","image-rendering","image-resolution","ime-mode","initial-letter","initial-letter-align","inline-size","inset","inset-area","inset-block","inset-block-end","inset-block-start","inset-inline","inset-inline-end","inset-inline-start","isolation","justify-content","justify-items","justify-self","kerning","left","letter-spacing","lighting-color","line-break","line-height","line-height-step","list-style","list-style-image","list-style-position","list-style-type","margin","margin-block","margin-block-end","margin-block-start","margin-bottom","margin-inline","margin-inline-end","margin-inline-start","margin-left","margin-right","margin-top","margin-trim","marker","marker-end","marker-mid","marker-start","marks","mask","mask-border","mask-border-mode","mask-border-outset","mask-border-repeat","mask-border-slice","mask-border-source","mask-border-width","mask-clip","mask-composite","mask-image","mask-mode","mask-origin","mask-position","mask-repeat","mask-size","mask-type","masonry-auto-flow","math-depth","math-shift","math-style","max-block-size","max-height","max-inline-size","max-width","min-block-size","min-height","min-inline-size","min-width","mix-blend-mode","nav-down","nav-index","nav-left","nav-right","nav-up","none","normal","object-fit","object-position","offset","offset-anchor","offset-distance","offset-path","offset-position","offset-rotate","opacity","order","orphans","outline","outline-color","outline-offset","outline-style","outline-width","overflow","overflow-anchor","overflow-block","overflow-clip-margin","overflow-inline","overflow-wrap","overflow-x","overflow-y","overlay","overscroll-behavior","overscroll-behavior-block","overscroll-behavior-inline","overscroll-behavior-x","overscroll-behavior-y","padding","padding-block","padding-block-end","padding-block-start","padding-bottom","padding-inline","padding-inline-end","padding-inline-start","padding-left","padding-right","padding-top","page","page-break-after","page-break-before","page-break-inside","paint-order","pause","pause-after","pause-before","perspective","perspective-origin","place-content","place-items","place-self","pointer-events","position","position-anchor","position-visibility","print-color-adjust","quotes","r","resize","rest","rest-after","rest-before","right","rotate","row-gap","ruby-align","ruby-position","scale","scroll-behavior","scroll-margin","scroll-margin-block","scroll-margin-block-end","scroll-margin-block-start","scroll-margin-bottom","scroll-margin-inline","scroll-margin-inline-end","scroll-margin-inline-start","scroll-margin-left","scroll-margin-right","scroll-margin-top","scroll-padding","scroll-padding-block","scroll-padding-block-end","scroll-padding-block-start","scroll-padding-bottom","scroll-padding-inline","scroll-padding-inline-end","scroll-padding-inline-start","scroll-padding-left","scroll-padding-right","scroll-padding-top","scroll-snap-align","scroll-snap-stop","scroll-snap-type","scroll-timeline","scroll-timeline-axis","scroll-timeline-name","scrollbar-color","scrollbar-gutter","scrollbar-width","shape-image-threshold","shape-margin","shape-outside","shape-rendering","speak","speak-as","src","stop-color","stop-opacity","stroke","stroke-dasharray","stroke-dashoffset","stroke-linecap","stroke-linejoin","stroke-miterlimit","stroke-opacity","stroke-width","tab-size","table-layout","text-align","text-align-all","text-align-last","text-anchor","text-combine-upright","text-decoration","text-decoration-color","text-decoration-line","text-decoration-skip","text-decoration-skip-ink","text-decoration-style","text-decoration-thickness","text-emphasis","text-emphasis-color","text-emphasis-position","text-emphasis-style","text-indent","text-justify","text-orientation","text-overflow","text-rendering","text-shadow","text-size-adjust","text-transform","text-underline-offset","text-underline-position","text-wrap","text-wrap-mode","text-wrap-style","timeline-scope","top","touch-action","transform","transform-box","transform-origin","transform-style","transition","transition-behavior","transition-delay","transition-duration","transition-property","transition-timing-function","translate","unicode-bidi","user-modify","user-select","vector-effect","vertical-align","view-timeline","view-timeline-axis","view-timeline-inset","view-timeline-name","view-transition-name","visibility","voice-balance","voice-duration","voice-family","voice-pitch","voice-range","voice-rate","voice-stress","voice-volume","white-space","white-space-collapse","widows","width","will-change","word-break","word-spacing","word-wrap","writing-mode","x","y","z-index","zoom"].sort().reverse(),R=C.concat(T).sort().reverse(),P=["a","abbr","address","article","aside","audio","b","blockquote","body","button","canvas","caption","cite","code","dd","del","details","dfn","div","dl","dt","em","fieldset","figcaption","figure","footer","form","h1","h2","h3","h4","h5","h6","header","hgroup","html","i","iframe","img","input","ins","kbd","label","legend","li","main","mark","menu","nav","object","ol","optgroup","option","p","picture","q","quote","samp","section","select","source","span","strong","summary","sup","table","tbody","td","textarea","tfoot","th","thead","time","tr","ul","var","video","defs","g","marker","mask","pattern","svg","switch","symbol","feBlend","feColorMatrix","feComponentTransfer","feComposite","feConvolveMatrix","feDiffuseLighting","feDisplacementMap","feFlood","feGaussianBlur","feImage","feMerge","feMorphology","feOffset","feSpecularLighting","feTile","feTurbulence","linearGradient","radialGradient","stop","circle","ellipse","image","line","path","polygon","polyline","rect","text","use","textPath","tspan","foreignObject","clipPath"],D=["any-hover","any-pointer","aspect-ratio","color","color-gamut","color-index","device-aspect-ratio","device-height","device-width","display-mode","forced-colors","grid","height","hover","inverted-colors","monochrome","orientation","overflow-block","overflow-inline","pointer","prefers-color-scheme","prefers-contrast","prefers-reduced-motion","prefers-reduced-transparency","resolution","scan","scripting","update","width","min-width","max-width","min-height","max-height"].sort().reverse(),M=["active","any-link","blank","checked","current","default","defined","dir","disabled","drop","empty","enabled","first","first-child","first-of-type","fullscreen","future","focus","focus-visible","focus-within","has","host","host-context","hover","indeterminate","in-range","invalid","is","lang","last-child","last-of-type","left","link","local-link","not","nth-child","nth-col","nth-last-child","nth-last-col","nth-last-of-type","nth-of-type","only-child","only-of-type","optional","out-of-range","past","placeholder-shown","read-only","read-write","required","right","root","scope","target","target-within","user-invalid","valid","visited","where"].sort().reverse(),$=["after","backdrop","before","cue","cue-region","first-letter","first-line","grammar-error","marker","part","placeholder","selection","slotted","spelling-error"].sort().reverse(),N=["accent-color","align-content","align-items","align-self","alignment-baseline","all","anchor-name","animation","animation-composition","animation-delay","animation-direction","animation-duration","animation-fill-mode","animation-iteration-count","animation-name","animation-play-state","animation-range","animation-range-end","animation-range-start","animation-timeline","animation-timing-function","appearance","aspect-ratio","backdrop-filter","backface-visibility","background","background-attachment","background-blend-mode","background-clip","background-color","background-image","background-origin","background-position","background-position-x","background-position-y","background-repeat","background-size","baseline-shift","block-size","border","border-block","border-block-color","border-block-end","border-block-end-color","border-block-end-style","border-block-end-width","border-block-start","border-block-start-color","border-block-start-style","border-block-start-width","border-block-style","border-block-width","border-bottom","border-bottom-color","border-bottom-left-radius","border-bottom-right-radius","border-bottom-style","border-bottom-width","border-collapse","border-color","border-end-end-radius","border-end-start-radius","border-image","border-image-outset","border-image-repeat","border-image-slice","border-image-source","border-image-width","border-inline","border-inline-color","border-inline-end","border-inline-end-color","border-inline-end-style","border-inline-end-width","border-inline-start","border-inline-start-color","border-inline-start-style","border-inline-start-width","border-inline-style","border-inline-width","border-left","border-left-color","border-left-style","border-left-width","border-radius","border-right","border-right-color","border-right-style","border-right-width","border-spacing","border-start-end-radius","border-start-start-radius","border-style","border-top","border-top-color","border-top-left-radius","border-top-right-radius","border-top-style","border-top-width","border-width","bottom","box-align","box-decoration-break","box-direction","box-flex","box-flex-group","box-lines","box-ordinal-group","box-orient","box-pack","box-shadow","box-sizing","break-after","break-before","break-inside","caption-side","caret-color","clear","clip","clip-path","clip-rule","color","color-interpolation","color-interpolation-filters","color-profile","color-rendering","color-scheme","column-count","column-fill","column-gap","column-rule","column-rule-color","column-rule-style","column-rule-width","column-span","column-width","columns","contain","contain-intrinsic-block-size","contain-intrinsic-height","contain-intrinsic-inline-size","contain-intrinsic-size","contain-intrinsic-width","container","container-name","container-type","content","content-visibility","counter-increment","counter-reset","counter-set","cue","cue-after","cue-before","cursor","cx","cy","direction","display","dominant-baseline","empty-cells","enable-background","field-sizing","fill","fill-opacity","fill-rule","filter","flex","flex-basis","flex-direction","flex-flow","flex-grow","flex-shrink","flex-wrap","float","flood-color","flood-opacity","flow","font","font-display","font-family","font-feature-settings","font-kerning","font-language-override","font-optical-sizing","font-palette","font-size","font-size-adjust","font-smooth","font-smoothing","font-stretch","font-style","font-synthesis","font-synthesis-position","font-synthesis-small-caps","font-synthesis-style","font-synthesis-weight","font-variant","font-variant-alternates","font-variant-caps","font-variant-east-asian","font-variant-emoji","font-variant-ligatures","font-variant-numeric","font-variant-position","font-variation-settings","font-weight","forced-color-adjust","gap","glyph-orientation-horizontal","glyph-orientation-vertical","grid","grid-area","grid-auto-columns","grid-auto-flow","grid-auto-rows","grid-column","grid-column-end","grid-column-start","grid-gap","grid-row","grid-row-end","grid-row-start","grid-template","grid-template-areas","grid-template-columns","grid-template-rows","hanging-punctuation","height","hyphenate-character","hyphenate-limit-chars","hyphens","icon","image-orientation","image-rendering","image-resolution","ime-mode","initial-letter","initial-letter-align","inline-size","inset","inset-area","inset-block","inset-block-end","inset-block-start","inset-inline","inset-inline-end","inset-inline-start","isolation","justify-content","justify-items","justify-self","kerning","left","letter-spacing","lighting-color","line-break","line-height","line-height-step","list-style","list-style-image","list-style-position","list-style-type","margin","margin-block","margin-block-end","margin-block-start","margin-bottom","margin-inline","margin-inline-end","margin-inline-start","margin-left","margin-right","margin-top","margin-trim","marker","marker-end","marker-mid","marker-start","marks","mask","mask-border","mask-border-mode","mask-border-outset","mask-border-repeat","mask-border-slice","mask-border-source","mask-border-width","mask-clip","mask-composite","mask-image","mask-mode","mask-origin","mask-position","mask-repeat","mask-size","mask-type","masonry-auto-flow","math-depth","math-shift","math-style","max-block-size","max-height","max-inline-size","max-width","min-block-size","min-height","min-inline-size","min-width","mix-blend-mode","nav-down","nav-index","nav-left","nav-right","nav-up","none","normal","object-fit","object-position","offset","offset-anchor","offset-distance","offset-path","offset-position","offset-rotate","opacity","order","orphans","outline","outline-color","outline-offset","outline-style","outline-width","overflow","overflow-anchor","overflow-block","overflow-clip-margin","overflow-inline","overflow-wrap","overflow-x","overflow-y","overlay","overscroll-behavior","overscroll-behavior-block","overscroll-behavior-inline","overscroll-behavior-x","overscroll-behavior-y","padding","padding-block","padding-block-end","padding-block-start","padding-bottom","padding-inline","padding-inline-end","padding-inline-start","padding-left","padding-right","padding-top","page","page-break-after","page-break-before","page-break-inside","paint-order","pause","pause-after","pause-before","perspective","perspective-origin","place-content","place-items","place-self","pointer-events","position","position-anchor","position-visibility","print-color-adjust","quotes","r","resize","rest","rest-after","rest-before","right","rotate","row-gap","ruby-align","ruby-position","scale","scroll-behavior","scroll-margin","scroll-margin-block","scroll-margin-block-end","scroll-margin-block-start","scroll-margin-bottom","scroll-margin-inline","scroll-margin-inline-end","scroll-margin-inline-start","scroll-margin-left","scroll-margin-right","scroll-margin-top","scroll-padding","scroll-padding-block","scroll-padding-block-end","scroll-padding-block-start","scroll-padding-bottom","scroll-padding-inline","scroll-padding-inline-end","scroll-padding-inline-start","scroll-padding-left","scroll-padding-right","scroll-padding-top","scroll-snap-align","scroll-snap-stop","scroll-snap-type","scroll-timeline","scroll-timeline-axis","scroll-timeline-name","scrollbar-color","scrollbar-gutter","scrollbar-width","shape-image-threshold","shape-margin","shape-outside","shape-rendering","speak","speak-as","src","stop-color","stop-opacity","stroke","stroke-dasharray","stroke-dashoffset","stroke-linecap","stroke-linejoin","stroke-miterlimit","stroke-opacity","stroke-width","tab-size","table-layout","text-align","text-align-all","text-align-last","text-anchor","text-combine-upright","text-decoration","text-decoration-color","text-decoration-line","text-decoration-skip","text-decoration-skip-ink","text-decoration-style","text-decoration-thickness","text-emphasis","text-emphasis-color","text-emphasis-position","text-emphasis-style","text-indent","text-justify","text-orientation","text-overflow","text-rendering","text-shadow","text-size-adjust","text-transform","text-underline-offset","text-underline-position","text-wrap","text-wrap-mode","text-wrap-style","timeline-scope","top","touch-action","transform","transform-box","transform-origin","transform-style","transition","transition-behavior","transition-delay","transition-duration","transition-property","transition-timing-function","translate","unicode-bidi","user-modify","user-select","vector-effect","vertical-align","view-timeline","view-timeline-axis","view-timeline-inset","view-timeline-name","view-transition-name","visibility","voice-balance","voice-duration","voice-family","voice-pitch","voice-range","voice-rate","voice-stress","voice-volume","white-space","white-space-collapse","widows","width","will-change","word-break","word-spacing","word-wrap","writing-mode","x","y","z-index","zoom"].sort().reverse();function I(e){return e?"string"==typeof e?e:e.source:null}function L(e){return B("(?=",e,")")}function B(...e){return e.map((e=>I(e))).join("")}function F(...e){const t=function(e){const t=e[e.length-1];return"object"==typeof t&&t.constructor===Object?(e.splice(e.length-1,1),t):{}}(e);return"("+(t.capture?"":"?:")+e.map((e=>I(e))).join("|")+")"}const Q=e=>B(/\b/,e,/\w$/.test(e)?/\b/:/\B/),U=["Protocol","Type"].map(Q),z=["init","self"].map(Q),j=["Any","Self"],W=["actor","any","associatedtype","async","await",/as\?/,/as!/,"as","borrowing","break","case","catch","class","consume","consuming","continue","convenience","copy","default","defer","deinit","didSet","distributed","do","dynamic","each","else","enum","extension","fallthrough",/fileprivate\(set\)/,"fileprivate","final","for","func","get","guard","if","import","indirect","infix",/init\?/,/init!/,"inout",/internal\(set\)/,"internal","in","is","isolated","nonisolated","lazy","let","macro","mutating","nonmutating",/open\(set\)/,"open","operator","optional","override","package","postfix","precedencegroup","prefix",/private\(set\)/,"private","protocol",/public\(set\)/,"public","repeat","required","rethrows","return","set","some","static","struct","subscript","super","switch","throws","throw",/try\?/,/try!/,"try","typealias",/unowned\(safe\)/,/unowned\(unsafe\)/,"unowned","var","weak","where","while","willSet"],X=["false","nil","true"],q=["assignment","associativity","higherThan","left","lowerThan","none","right"],V=["#colorLiteral","#column","#dsohandle","#else","#elseif","#endif","#error","#file","#fileID","#fileLiteral","#filePath","#function","#if","#imageLiteral","#keyPath","#line","#selector","#sourceLocation","#warning"],H=["abs","all","any","assert","assertionFailure","debugPrint","dump","fatalError","getVaList","isKnownUniquelyReferenced","max","min","numericCast","pointwiseMax","pointwiseMin","precondition","preconditionFailure","print","readLine","repeatElement","sequence","stride","swap","swift_unboxFromSwiftValueWithType","transcode","type","unsafeBitCast","unsafeDowncast","withExtendedLifetime","withUnsafeMutablePointer","withUnsafePointer","withVaList","withoutActuallyEscaping","zip"],Z=F(/[/=\-+!*%<>&|^~?]/,/[\u00A1-\u00A7]/,/[\u00A9\u00AB]/,/[\u00AC\u00AE]/,/[\u00B0\u00B1]/,/[\u00B6\u00BB\u00BF\u00D7\u00F7]/,/[\u2016-\u2017]/,/[\u2020-\u2027]/,/[\u2030-\u203E]/,/[\u2041-\u2053]/,/[\u2055-\u205E]/,/[\u2190-\u23FF]/,/[\u2500-\u2775]/,/[\u2794-\u2BFF]/,/[\u2E00-\u2E7F]/,/[\u3001-\u3003]/,/[\u3008-\u3020]/,/[\u3030]/),G=F(Z,/[\u0300-\u036F]/,/[\u1DC0-\u1DFF]/,/[\u20D0-\u20FF]/,/[\uFE00-\uFE0F]/,/[\uFE20-\uFE2F]/),K=B(Z,G,"*"),Y=F(/[a-zA-Z_]/,/[\u00A8\u00AA\u00AD\u00AF\u00B2-\u00B5\u00B7-\u00BA]/,/[\u00BC-\u00BE\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u00FF]/,/[\u0100-\u02FF\u0370-\u167F\u1681-\u180D\u180F-\u1DBF]/,/[\u1E00-\u1FFF]/,/[\u200B-\u200D\u202A-\u202E\u203F-\u2040\u2054\u2060-\u206F]/,/[\u2070-\u20CF\u2100-\u218F\u2460-\u24FF\u2776-\u2793]/,/[\u2C00-\u2DFF\u2E80-\u2FFF]/,/[\u3004-\u3007\u3021-\u302F\u3031-\u303F\u3040-\uD7FF]/,/[\uF900-\uFD3D\uFD40-\uFDCF\uFDF0-\uFE1F\uFE30-\uFE44]/,/[\uFE47-\uFEFE\uFF00-\uFFFD]/),J=F(Y,/\d/,/[\u0300-\u036F\u1DC0-\u1DFF\u20D0-\u20FF\uFE20-\uFE2F]/),ee=B(Y,J,"*"),te=B(/[A-Z]/,J,"*"),ne=["attached","autoclosure",B(/convention\(/,F("swift","block","c"),/\)/),"discardableResult","dynamicCallable","dynamicMemberLookup","escaping","freestanding","frozen","GKInspectable","IBAction","IBDesignable","IBInspectable","IBOutlet","IBSegueAction","inlinable","main","nonobjc","NSApplicationMain","NSCopying","NSManaged",B(/objc\(/,ee,/\)/),"objc","objcMembers","propertyWrapper","requires_stored_property_inits","resultBuilder","Sendable","testable","UIApplicationMain","unchecked","unknown","usableFromInline","warn_unqualified_access"],re=["iOS","iOSApplicationExtension","macOS","macOSApplicationExtension","macCatalyst","macCatalystApplicationExtension","watchOS","watchOSApplicationExtension","tvOS","tvOSApplicationExtension","swift"],ae="[A-Za-z$_][0-9A-Za-z$_]*",oe=["as","in","of","if","for","while","finally","var","new","function","do","return","void","else","break","catch","instanceof","with","throw","case","default","try","switch","continue","typeof","delete","let","yield","const","class","debugger","async","await","static","import","from","export","extends","using"],ie=["true","false","null","undefined","NaN","Infinity"],se=["Object","Function","Boolean","Symbol","Math","Date","Number","BigInt","String","RegExp","Array","Float32Array","Float64Array","Int8Array","Uint8Array","Uint8ClampedArray","Int16Array","Int32Array","Uint16Array","Uint32Array","BigInt64Array","BigUint64Array","Set","Map","WeakSet","WeakMap","ArrayBuffer","SharedArrayBuffer","Atomics","DataView","JSON","Promise","Generator","GeneratorFunction","AsyncFunction","Reflect","Proxy","Intl","WebAssembly"],le=["Error","EvalError","InternalError","RangeError","ReferenceError","SyntaxError","TypeError","URIError"],ce=["setInterval","setTimeout","clearInterval","clearTimeout","require","exports","eval","isFinite","isNaN","parseFloat","parseInt","decodeURI","decodeURIComponent","encodeURI","encodeURIComponent","escape","unescape"],ue=["arguments","this","super","console","window","document","localStorage","sessionStorage","module","global"],de=[].concat(ce,se,le);function pe(e){const t=e.regex,n=t.concat(/[\p{L}_]/u,t.optional(/[\p{L}0-9_.-]*:/u),/[\p{L}0-9_.-]*/u),r={className:"symbol",begin:/&[a-z]+;|&#[0-9]+;|&#x[a-f0-9]+;/},a={begin:/\s/,contains:[{className:"keyword",begin:/#?[a-z_][a-z1-9_-]+/,illegal:/\n/}]},o=e.inherit(a,{begin:/\(/,end:/\)/}),i=e.inherit(e.APOS_STRING_MODE,{className:"string"}),s=e.inherit(e.QUOTE_STRING_MODE,{className:"string"}),l={endsWithParent:!0,illegal:/</,relevance:0,contains:[{className:"attr",begin:/[\p{L}0-9._:-]+/u,relevance:0},{begin:/=\s*/,relevance:0,contains:[{className:"string",endsParent:!0,variants:[{begin:/"/,end:/"/,contains:[r]},{begin:/'/,end:/'/,contains:[r]},{begin:/[^\s"'=<>`]+/}]}]}]};return{name:"HTML, XML",aliases:["html","xhtml","rss","atom","xjb","xsd","xsl","plist","wsf","svg"],case_insensitive:!0,unicodeRegex:!0,contains:[{className:"meta",begin:/<![a-z]/,end:/>/,relevance:10,contains:[a,s,i,o,{begin:/\[/,end:/\]/,contains:[{className:"meta",begin:/<![a-z]/,end:/>/,contains:[a,o,s,i]}]}]},e.COMMENT(/<!--/,/-->/,{relevance:10}),{begin:/<!\[CDATA\[/,end:/\]\]>/,relevance:10},r,{className:"meta",end:/\?>/,variants:[{begin:/<\?xml/,relevance:10,contains:[s]},{begin:/<\?[a-z][a-z0-9]+/}]},{className:"tag",begin:/<style(?=\s|>)/,end:/>/,keywords:{name:"style"},contains:[l],starts:{end:/<\/style>/,returnEnd:!0,subLanguage:["css","xml"]}},{className:"tag",begin:/<script(?=\s|>)/,end:/>/,keywords:{name:"script"},contains:[l],starts:{end:/<\/script>/,returnEnd:!0,subLanguage:["javascript","handlebars","xml"]}},{className:"tag",begin:/<>|<\/>/},{className:"tag",begin:t.concat(/</,t.lookahead(t.concat(n,t.either(/\/>/,/>/,/\s/)))),end:/\/?>/,contains:[{className:"name",begin:n,relevance:0,starts:l}]},{className:"tag",begin:t.concat(/<\//,t.lookahead(t.concat(n,/>/))),contains:[{className:"name",begin:n,relevance:0},{begin:/>/,relevance:0,endsParent:!0}]}]}}const he={bash:function(e){const t=e.regex,n={},r={begin:/\$\{/,end:/\}/,contains:["self",{begin:/:-/,contains:[n]}]};Object.assign(n,{className:"variable",variants:[{begin:t.concat(/\$[\w\d#@][\w\d_]*/,"(?![\\w\\d])(?![$])")},r]});const a={className:"subst",begin:/\$\(/,end:/\)/,contains:[e.BACKSLASH_ESCAPE]},o=e.inherit(e.COMMENT(),{match:[/(^|\s)/,/#.*$/],scope:{2:"comment"}}),i={begin:/<<-?\s*(?=\w+)/,starts:{contains:[e.END_SAME_AS_BEGIN({begin:/(\w+)/,end:/(\w+)/,className:"string"})]}},s={className:"string",begin:/"/,end:/"/,contains:[e.BACKSLASH_ESCAPE,n,a]};a.contains.push(s);const l={begin:/\$?\(\(/,end:/\)\)/,contains:[{begin:/\d+#[0-9a-f]+/,className:"number"},e.NUMBER_MODE,n]},c=e.SHEBANG({binary:`(${["fish","bash","zsh","sh","csh","ksh","tcsh","dash","scsh"].join("|")})`,relevance:10}),u={className:"function",begin:/\w[\w\d_]*\s*\(\s*\)\s*\{/,returnBegin:!0,contains:[e.inherit(e.TITLE_MODE,{begin:/\w[\w\d_]*/})],relevance:0};return{name:"Bash",aliases:["sh","zsh"],keywords:{$pattern:/\b[a-z][a-z0-9._-]+\b/,keyword:["if","then","else","elif","fi","time","for","while","until","in","do","done","case","esac","coproc","function","select"],literal:["true","false"],built_in:["break","cd","continue","eval","exec","exit","export","getopts","hash","pwd","readonly","return","shift","test","times","trap","umask","unset","alias","bind","builtin","caller","command","declare","echo","enable","help","let","local","logout","mapfile","printf","read","readarray","source","sudo","type","typeset","ulimit","unalias","set","shopt","autoload","bg","bindkey","bye","cap","chdir","clone","comparguments","compcall","compctl","compdescribe","compfiles","compgroups","compquote","comptags","comptry","compvalues","dirs","disable","disown","echotc","echoti","emulate","fc","fg","float","functions","getcap","getln","history","integer","jobs","kill","limit","log","noglob","popd","print","pushd","pushln","rehash","sched","setcap","setopt","stat","suspend","ttyctl","unfunction","unhash","unlimit","unsetopt","vared","wait","whence","where","which","zcompile","zformat","zftp","zle","zmodload","zparseopts","zprof","zpty","zregexparse","zsocket","zstyle","ztcp","chcon","chgrp","chown","chmod","cp","dd","df","dir","dircolors","ln","ls","mkdir","mkfifo","mknod","mktemp","mv","realpath","rm","rmdir","shred","sync","touch","truncate","vdir","b2sum","base32","base64","cat","cksum","comm","csplit","cut","expand","fmt","fold","head","join","md5sum","nl","numfmt","od","paste","ptx","pr","sha1sum","sha224sum","sha256sum","sha384sum","sha512sum","shuf","sort","split","sum","tac","tail","tr","tsort","unexpand","uniq","wc","arch","basename","chroot","date","dirname","du","echo","env","expr","factor","groups","hostid","id","link","logname","nice","nohup","nproc","pathchk","pinky","printenv","printf","pwd","readlink","runcon","seq","sleep","stat","stdbuf","stty","tee","test","timeout","tty","uname","unlink","uptime","users","who","whoami","yes"]},contains:[c,e.SHEBANG(),u,l,o,i,{match:/(\/[a-z._-]+)+/},s,{match:/\\"/},{className:"string",begin:/'/,end:/'/},{match:/\\'/},n]}},c:function(e){const t=e.regex,n=e.COMMENT("//","$",{contains:[{begin:/\\\n/}]}),r="decltype\\(auto\\)",a="[a-zA-Z_]\\w*::",o="("+r+"|"+t.optional(a)+"[a-zA-Z_]\\w*"+t.optional("<[^<>]+>")+")",i={className:"type",variants:[{begin:"\\b[a-z\\d_]*_t\\b"},{match:/\batomic_[a-z]{3,6}\b/}]},s={className:"string",variants:[{begin:'(u8?|U|L)?"',end:'"',illegal:"\\n",contains:[e.BACKSLASH_ESCAPE]},{begin:"(u8?|U|L)?'(\\\\(x[0-9A-Fa-f]{2}|u[0-9A-Fa-f]{4,8}|[0-7]{3}|\\S)|.)",end:"'",illegal:"."},e.END_SAME_AS_BEGIN({begin:/(?:u8?|U|L)?R"([^()\\ ]{0,16})\(/,end:/\)([^()\\ ]{0,16})"/})]},l={className:"number",variants:[{match:/\b(0b[01']+)/},{match:/(-?)\b([\d']+(\.[\d']*)?|\.[\d']+)((ll|LL|l|L)(u|U)?|(u|U)(ll|LL|l|L)?|f|F|b|B)/},{match:/(-?)\b(0[xX][a-fA-F0-9]+(?:'[a-fA-F0-9]+)*(?:\.[a-fA-F0-9]*(?:'[a-fA-F0-9]*)*)?(?:[pP][-+]?[0-9]+)?(l|L)?(u|U)?)/},{match:/(-?)\b\d+(?:'\d+)*(?:\.\d*(?:'\d*)*)?(?:[eE][-+]?\d+)?/}],relevance:0},c={className:"meta",begin:/#\s*[a-z]+\b/,end:/$/,keywords:{keyword:"if else elif endif define undef warning error line pragma _Pragma ifdef ifndef elifdef elifndef include"},contains:[{begin:/\\\n/,relevance:0},e.inherit(s,{className:"string"}),{className:"string",begin:/<.*?>/},n,e.C_BLOCK_COMMENT_MODE]},u={className:"title",begin:t.optional(a)+e.IDENT_RE,relevance:0},d=t.optional(a)+e.IDENT_RE+"\\s*\\(",p={keyword:["asm","auto","break","case","continue","default","do","else","enum","extern","for","fortran","goto","if","inline","register","restrict","return","sizeof","typeof","typeof_unqual","struct","switch","typedef","union","volatile","while","_Alignas","_Alignof","_Atomic","_Generic","_Noreturn","_Static_assert","_Thread_local","alignas","alignof","noreturn","static_assert","thread_local","_Pragma"],type:["float","double","signed","unsigned","int","short","long","char","void","_Bool","_BitInt","_Complex","_Imaginary","_Decimal32","_Decimal64","_Decimal96","_Decimal128","_Decimal64x","_Decimal128x","_Float16","_Float32","_Float64","_Float128","_Float32x","_Float64x","_Float128x","const","static","constexpr","complex","bool","imaginary"],literal:"true false NULL",built_in:"std string wstring cin cout cerr clog stdin stdout stderr stringstream istringstream ostringstream auto_ptr deque list queue stack vector map set pair bitset multiset multimap unordered_set unordered_map unordered_multiset unordered_multimap priority_queue make_pair array shared_ptr abort terminate abs acos asin atan2 atan calloc ceil cosh cos exit exp fabs floor fmod fprintf fputs free frexp fscanf future isalnum isalpha iscntrl isdigit isgraph islower isprint ispunct isspace isupper isxdigit tolower toupper labs ldexp log10 log malloc realloc memchr memcmp memcpy memset modf pow printf putchar puts scanf sinh sin snprintf sprintf sqrt sscanf strcat strchr strcmp strcpy strcspn strlen strncat strncmp strncpy strpbrk strrchr strspn strstr tanh tan vfprintf vprintf vsprintf endl initializer_list unique_ptr"},h=[c,i,n,e.C_BLOCK_COMMENT_MODE,l,s],f={variants:[{begin:/=/,end:/;/},{begin:/\(/,end:/\)/},{beginKeywords:"new throw return else",end:/;/}],keywords:p,contains:h.concat([{begin:/\(/,end:/\)/,keywords:p,contains:h.concat(["self"]),relevance:0}]),relevance:0},m={begin:"("+o+"[\\*&\\s]+)+"+d,returnBegin:!0,end:/[{;=]/,excludeEnd:!0,keywords:p,illegal:/[^\w\s\*&:<>.]/,contains:[{begin:r,keywords:p,relevance:0},{begin:d,returnBegin:!0,contains:[e.inherit(u,{className:"title.function"})],relevance:0},{relevance:0,match:/,/},{className:"params",begin:/\(/,end:/\)/,keywords:p,relevance:0,contains:[n,e.C_BLOCK_COMMENT_MODE,s,l,i,{begin:/\(/,end:/\)/,keywords:p,relevance:0,contains:["self",n,e.C_BLOCK_COMMENT_MODE,s,l,i]}]},i,n,e.C_BLOCK_COMMENT_MODE,c]};return{name:"C",aliases:["h"],keywords:p,disableAutodetect:!0,illegal:"</",contains:[].concat(f,m,h,[c,{begin:e.IDENT_RE+"::",keywords:p},{className:"class",beginKeywords:"enum class struct union",end:/[{;:<>=]/,contains:[{beginKeywords:"final class struct"},e.TITLE_MODE]}]),exports:{preprocessor:c,strings:s,keywords:p}}},clojure:function(e){const t="a-zA-Z_\\-!.?+*=<>&'",n="[#]?["+t+"]["+t+"0-9/;:$#]*",r="def defonce defprotocol defstruct defmulti defmethod defn- defn defmacro deftype defrecord",a={$pattern:n,built_in:r+" cond apply if-not if-let if not not= =|0 <|0 >|0 <=|0 >=|0 ==|0 +|0 /|0 *|0 -|0 rem quot neg? pos? delay? symbol? keyword? true? false? integer? empty? coll? list? set? ifn? fn? associative? sequential? sorted? counted? reversible? number? decimal? class? distinct? isa? float? rational? reduced? ratio? odd? even? char? seq? vector? string? map? nil? contains? zero? instance? not-every? not-any? libspec? -> ->> .. . inc compare do dotimes mapcat take remove take-while drop letfn drop-last take-last drop-while while intern condp case reduced cycle split-at split-with repeat replicate iterate range merge zipmap declare line-seq sort comparator sort-by dorun doall nthnext nthrest partition eval doseq await await-for let agent atom send send-off release-pending-sends add-watch mapv filterv remove-watch agent-error restart-agent set-error-handler error-handler set-error-mode! error-mode shutdown-agents quote var fn loop recur throw try monitor-enter monitor-exit macroexpand macroexpand-1 for dosync and or when when-not when-let comp juxt partial sequence memoize constantly complement identity assert peek pop doto proxy first rest cons cast coll last butlast sigs reify second ffirst fnext nfirst nnext meta with-meta ns in-ns create-ns import refer keys select-keys vals key val rseq name namespace promise into transient persistent! conj! assoc! dissoc! pop! disj! use class type num float double short byte boolean bigint biginteger bigdec print-method print-dup throw-if printf format load compile get-in update-in pr pr-on newline flush read slurp read-line subvec with-open memfn time re-find re-groups rand-int rand mod locking assert-valid-fdecl alias resolve ref deref refset swap! reset! set-validator! compare-and-set! alter-meta! reset-meta! commute get-validator alter ref-set ref-history-count ref-min-history ref-max-history ensure sync io! new next conj set! to-array future future-call into-array aset gen-class reduce map filter find empty hash-map hash-set sorted-map sorted-map-by sorted-set sorted-set-by vec vector seq flatten reverse assoc dissoc list disj get union difference intersection extend extend-type extend-protocol int nth delay count concat chunk chunk-buffer chunk-append chunk-first chunk-rest max min dec unchecked-inc-int unchecked-inc unchecked-dec-inc unchecked-dec unchecked-negate unchecked-add-int unchecked-add unchecked-subtract-int unchecked-subtract chunk-next chunk-cons chunked-seq? prn vary-meta lazy-seq spread list* str find-keyword keyword symbol gensym force rationalize"},o={begin:n,relevance:0},i={scope:"number",relevance:0,variants:[{match:/[-+]?0[xX][0-9a-fA-F]+N?/},{match:/[-+]?0[0-7]+N?/},{match:/[-+]?[1-9][0-9]?[rR][0-9a-zA-Z]+N?/},{match:/[-+]?[0-9]+\/[0-9]+N?/},{match:/[-+]?[0-9]+((\.[0-9]*([eE][+-]?[0-9]+)?M?)|([eE][+-]?[0-9]+M?|M))/},{match:/[-+]?([1-9][0-9]*|0)N?/}]},s={scope:"character",variants:[{match:/\\o[0-3]?[0-7]{1,2}/},{match:/\\u[0-9a-fA-F]{4}/},{match:/\\(newline|space|tab|formfeed|backspace|return)/},{match:/\\\S/,relevance:0}]},l={scope:"regex",begin:/#"/,end:/"/,contains:[e.BACKSLASH_ESCAPE]},c=e.inherit(e.QUOTE_STRING_MODE,{illegal:null}),u={scope:"punctuation",match:/,/,relevance:0},d=e.COMMENT(";","$",{relevance:0}),p={className:"literal",begin:/\b(true|false|nil)\b/},h={begin:"\\[|(#::?"+n+")?\\{",end:"[\\]\\}]",relevance:0},f={className:"symbol",begin:"[:]{1,2}"+n},m={begin:"\\(",end:"\\)"},g={endsWithParent:!0,relevance:0},v={keywords:a,className:"name",begin:n,relevance:0,starts:g},b=[u,m,s,l,c,d,f,h,i,p,o],y={beginKeywords:r,keywords:{$pattern:n,keyword:r},end:'(\\[|#|\\d|"|:|\\{|\\)|\\(|$)',contains:[{className:"title",begin:n,relevance:0,excludeEnd:!0,endsParent:!0}].concat(b)};return m.contains=[y,v,g],g.contains=b,h.contains=b,{name:"Clojure",aliases:["clj","edn"],illegal:/\S/,contains:[u,m,s,l,c,d,f,h,i,p]}},cpp:function(e){const t=e.regex,n=e.COMMENT("//","$",{contains:[{begin:/\\\n/}]}),r="decltype\\(auto\\)",a="[a-zA-Z_]\\w*::",o="(?!struct)("+r+"|"+t.optional(a)+"[a-zA-Z_]\\w*"+t.optional("<[^<>]+>")+")",i={className:"type",begin:"\\b[a-z\\d_]*_t\\b"},s={className:"string",variants:[{begin:'(u8?|U|L)?"',end:'"',illegal:"\\n",contains:[e.BACKSLASH_ESCAPE]},{begin:"(u8?|U|L)?'(\\\\(x[0-9A-Fa-f]{2}|u[0-9A-Fa-f]{4,8}|[0-7]{3}|\\S)|.)",end:"'",illegal:"."},e.END_SAME_AS_BEGIN({begin:/(?:u8?|U|L)?R"([^()\\ ]{0,16})\(/,end:/\)([^()\\ ]{0,16})"/})]},l={className:"number",variants:[{begin:"[+-]?(?:(?:[0-9](?:'?[0-9])*\\.(?:[0-9](?:'?[0-9])*)?|\\.[0-9](?:'?[0-9])*)(?:[Ee][+-]?[0-9](?:'?[0-9])*)?|[0-9](?:'?[0-9])*[Ee][+-]?[0-9](?:'?[0-9])*|0[Xx](?:[0-9A-Fa-f](?:'?[0-9A-Fa-f])*(?:\\.(?:[0-9A-Fa-f](?:'?[0-9A-Fa-f])*)?)?|\\.[0-9A-Fa-f](?:'?[0-9A-Fa-f])*)[Pp][+-]?[0-9](?:'?[0-9])*)(?:[Ff](?:16|32|64|128)?|(BF|bf)16|[Ll]|)"},{begin:"[+-]?\\b(?:0[Bb][01](?:'?[01])*|0[Xx][0-9A-Fa-f](?:'?[0-9A-Fa-f])*|0(?:'?[0-7])*|[1-9](?:'?[0-9])*)(?:[Uu](?:LL?|ll?)|[Uu][Zz]?|(?:LL?|ll?)[Uu]?|[Zz][Uu]|)"}],relevance:0},c={className:"meta",begin:/#\s*[a-z]+\b/,end:/$/,keywords:{keyword:"if else elif endif define undef warning error line pragma _Pragma ifdef ifndef include"},contains:[{begin:/\\\n/,relevance:0},e.inherit(s,{className:"string"}),{className:"string",begin:/<.*?>/},n,e.C_BLOCK_COMMENT_MODE]},u={className:"title",begin:t.optional(a)+e.IDENT_RE,relevance:0},d=t.optional(a)+e.IDENT_RE+"\\s*\\(",p={type:["bool","char","char16_t","char32_t","char8_t","double","float","int","long","short","void","wchar_t","unsigned","signed","const","static"],keyword:["alignas","alignof","and","and_eq","asm","atomic_cancel","atomic_commit","atomic_noexcept","auto","bitand","bitor","break","case","catch","class","co_await","co_return","co_yield","compl","concept","const_cast|10","consteval","constexpr","constinit","continue","decltype","default","delete","do","dynamic_cast|10","else","enum","explicit","export","extern","false","final","for","friend","goto","if","import","inline","module","mutable","namespace","new","noexcept","not","not_eq","nullptr","operator","or","or_eq","override","private","protected","public","reflexpr","register","reinterpret_cast|10","requires","return","sizeof","static_assert","static_cast|10","struct","switch","synchronized","template","this","thread_local","throw","transaction_safe","transaction_safe_dynamic","true","try","typedef","typeid","typename","union","using","virtual","volatile","while","xor","xor_eq"],literal:["NULL","false","nullopt","nullptr","true"],built_in:["_Pragma"],_type_hints:["any","auto_ptr","barrier","binary_semaphore","bitset","complex","condition_variable","condition_variable_any","counting_semaphore","deque","false_type","flat_map","flat_set","future","imaginary","initializer_list","istringstream","jthread","latch","lock_guard","multimap","multiset","mutex","optional","ostringstream","packaged_task","pair","promise","priority_queue","queue","recursive_mutex","recursive_timed_mutex","scoped_lock","set","shared_future","shared_lock","shared_mutex","shared_timed_mutex","shared_ptr","stack","string_view","stringstream","timed_mutex","thread","true_type","tuple","unique_lock","unique_ptr","unordered_map","unordered_multimap","unordered_multiset","unordered_set","variant","vector","weak_ptr","wstring","wstring_view"]},h={className:"function.dispatch",relevance:0,keywords:{_hint:["abort","abs","acos","apply","as_const","asin","atan","atan2","calloc","ceil","cerr","cin","clog","cos","cosh","cout","declval","endl","exchange","exit","exp","fabs","floor","fmod","forward","fprintf","fputs","free","frexp","fscanf","future","invoke","isalnum","isalpha","iscntrl","isdigit","isgraph","islower","isprint","ispunct","isspace","isupper","isxdigit","labs","launder","ldexp","log","log10","make_pair","make_shared","make_shared_for_overwrite","make_tuple","make_unique","malloc","memchr","memcmp","memcpy","memset","modf","move","pow","printf","putchar","puts","realloc","scanf","sin","sinh","snprintf","sprintf","sqrt","sscanf","std","stderr","stdin","stdout","strcat","strchr","strcmp","strcpy","strcspn","strlen","strncat","strncmp","strncpy","strpbrk","strrchr","strspn","strstr","swap","tan","tanh","terminate","to_underlying","tolower","toupper","vfprintf","visit","vprintf","vsprintf"]},begin:t.concat(/\b/,/(?!decltype)/,/(?!if)/,/(?!for)/,/(?!switch)/,/(?!while)/,e.IDENT_RE,t.lookahead(/(<[^<>]+>|)\s*\(/))},f=[h,c,i,n,e.C_BLOCK_COMMENT_MODE,l,s],m={variants:[{begin:/=/,end:/;/},{begin:/\(/,end:/\)/},{beginKeywords:"new throw return else",end:/;/}],keywords:p,contains:f.concat([{begin:/\(/,end:/\)/,keywords:p,contains:f.concat(["self"]),relevance:0}]),relevance:0},g={className:"function",begin:"("+o+"[\\*&\\s]+)+"+d,returnBegin:!0,end:/[{;=]/,excludeEnd:!0,keywords:p,illegal:/[^\w\s\*&:<>.]/,contains:[{begin:r,keywords:p,relevance:0},{begin:d,returnBegin:!0,contains:[u],relevance:0},{begin:/::/,relevance:0},{begin:/:/,endsWithParent:!0,contains:[s,l]},{relevance:0,match:/,/},{className:"params",begin:/\(/,end:/\)/,keywords:p,relevance:0,contains:[n,e.C_BLOCK_COMMENT_MODE,s,l,i,{begin:/\(/,end:/\)/,keywords:p,relevance:0,contains:["self",n,e.C_BLOCK_COMMENT_MODE,s,l,i]}]},i,n,e.C_BLOCK_COMMENT_MODE,c]};return{name:"C++",aliases:["cc","c++","h++","hpp","hh","hxx","cxx"],keywords:p,illegal:"</",classNameAliases:{"function.dispatch":"built_in"},contains:[].concat(m,g,h,f,[c,{begin:"\\b(deque|list|queue|priority_queue|pair|stack|vector|map|set|bitset|multiset|multimap|unordered_map|unordered_set|unordered_multiset|unordered_multimap|array|tuple|optional|variant|function|flat_map|flat_set)\\s*<(?!<)",end:">",keywords:p,contains:["self",i]},{begin:e.IDENT_RE+"::",keywords:p},{match:[/\b(?:enum(?:\s+(?:class|struct))?|class|struct|union)/,/\s+/,/\w+/],className:{1:"keyword",3:"title.class"}}])}},css:function(e){const t=e.regex,n=(e=>({IMPORTANT:{scope:"meta",begin:"!important"},BLOCK_COMMENT:e.C_BLOCK_COMMENT_MODE,HEXCOLOR:{scope:"number",begin:/#(([0-9a-fA-F]{3,4})|(([0-9a-fA-F]{2}){3,4}))\b/},FUNCTION_DISPATCH:{className:"built_in",begin:/[\w-]+(?=\()/},ATTRIBUTE_SELECTOR_MODE:{scope:"selector-attr",begin:/\[/,end:/\]/,illegal:"$",contains:[e.APOS_STRING_MODE,e.QUOTE_STRING_MODE]},CSS_NUMBER_MODE:{scope:"number",begin:e.NUMBER_RE+"(%|em|ex|ch|rem|vw|vh|vmin|vmax|cm|mm|in|pt|pc|px|deg|grad|rad|turn|s|ms|Hz|kHz|dpi|dpcm|dppx)?",relevance:0},CSS_VARIABLE:{className:"attr",begin:/--[A-Za-z_][A-Za-z0-9_-]*/}}))(e),l=[e.APOS_STRING_MODE,e.QUOTE_STRING_MODE];return{name:"CSS",case_insensitive:!0,illegal:/[=|'\$]/,keywords:{keyframePosition:"from to"},classNameAliases:{keyframePosition:"selector-tag"},contains:[n.BLOCK_COMMENT,{begin:/-(webkit|moz|ms|o)-(?=[a-z])/},n.CSS_NUMBER_MODE,{className:"selector-id",begin:/#[A-Za-z0-9_-]+/,relevance:0},{className:"selector-class",begin:"\\.[a-zA-Z-][a-zA-Z0-9_-]*",relevance:0},n.ATTRIBUTE_SELECTOR_MODE,{className:"selector-pseudo",variants:[{begin:":("+o.join("|")+")"},{begin:":(:)?("+i.join("|")+")"}]},n.CSS_VARIABLE,{className:"attribute",begin:"\\b("+s.join("|")+")\\b"},{begin:/:/,end:/[;}{]/,contains:[n.BLOCK_COMMENT,n.HEXCOLOR,n.IMPORTANT,n.CSS_NUMBER_MODE,...l,{begin:/(url|data-uri)\(/,end:/\)/,relevance:0,keywords:{built_in:"url data-uri"},contains:[...l,{className:"string",begin:/[^)]/,endsWithParent:!0,excludeEnd:!0}]},n.FUNCTION_DISPATCH]},{begin:t.lookahead(/@/),end:"[{;]",relevance:0,illegal:/:/,contains:[{className:"keyword",begin:/@-?\w[\w]*(-\w+)*/},{begin:/\s/,endsWithParent:!0,excludeEnd:!0,relevance:0,keywords:{$pattern:/[a-z-]+/,keyword:"and or not only",attribute:a.join(" ")},contains:[{begin:/[a-z-]+(?=:)/,className:"attribute"},...l,n.CSS_NUMBER_MODE]}]},{className:"selector-tag",begin:"\\b("+r.join("|")+")\\b"}]}},curl:n(7088),csharp:function(e){const t={keyword:["abstract","as","base","break","case","catch","class","const","continue","do","else","event","explicit","extern","finally","fixed","for","foreach","goto","if","implicit","in","interface","internal","is","lock","namespace","new","operator","out","override","params","private","protected","public","readonly","record","ref","return","scoped","sealed","sizeof","stackalloc","static","struct","switch","this","throw","try","typeof","unchecked","unsafe","using","virtual","void","volatile","while"].concat(["add","alias","and","ascending","args","async","await","by","descending","dynamic","equals","file","from","get","global","group","init","into","join","let","nameof","not","notnull","on","or","orderby","partial","record","remove","required","scoped","select","set","unmanaged","value|0","var","when","where","with","yield"]),built_in:["bool","byte","char","decimal","delegate","double","dynamic","enum","float","int","long","nint","nuint","object","sbyte","short","string","ulong","uint","ushort"],literal:["default","false","null","true"]},n=e.inherit(e.TITLE_MODE,{begin:"[a-zA-Z](\\.?\\w)*"}),r={className:"number",variants:[{begin:"\\b(0b[01']+)"},{begin:"(-?)\\b([\\d']+(\\.[\\d']*)?|\\.[\\d']+)(u|U|l|L|ul|UL|f|F|b|B)"},{begin:"(-?)(\\b0[xX][a-fA-F0-9']+|(\\b[\\d']+(\\.[\\d']*)?|\\.[\\d']+)([eE][-+]?[\\d']+)?)"}],relevance:0},a={className:"string",begin:'@"',end:'"',contains:[{begin:'""'}]},o=e.inherit(a,{illegal:/\n/}),i={className:"subst",begin:/\{/,end:/\}/,keywords:t},s=e.inherit(i,{illegal:/\n/}),l={className:"string",begin:/\$"/,end:'"',illegal:/\n/,contains:[{begin:/\{\{/},{begin:/\}\}/},e.BACKSLASH_ESCAPE,s]},c={className:"string",begin:/\$@"/,end:'"',contains:[{begin:/\{\{/},{begin:/\}\}/},{begin:'""'},i]},u=e.inherit(c,{illegal:/\n/,contains:[{begin:/\{\{/},{begin:/\}\}/},{begin:'""'},s]});i.contains=[c,l,a,e.APOS_STRING_MODE,e.QUOTE_STRING_MODE,r,e.C_BLOCK_COMMENT_MODE],s.contains=[u,l,o,e.APOS_STRING_MODE,e.QUOTE_STRING_MODE,r,e.inherit(e.C_BLOCK_COMMENT_MODE,{illegal:/\n/})];const d={variants:[{className:"string",begin:/"""("*)(?!")(.|\n)*?"""\1/,relevance:1},c,l,a,e.APOS_STRING_MODE,e.QUOTE_STRING_MODE]},p={begin:"<",end:">",contains:[{beginKeywords:"in out"},n]},h=e.IDENT_RE+"(<"+e.IDENT_RE+"(\\s*,\\s*"+e.IDENT_RE+")*>)?(\\[\\])?",f={begin:"@"+e.IDENT_RE,relevance:0};return{name:"C#",aliases:["cs","c#"],keywords:t,illegal:/::/,contains:[e.COMMENT("///","$",{returnBegin:!0,contains:[{className:"doctag",variants:[{begin:"///",relevance:0},{begin:"\x3c!--|--\x3e"},{begin:"</?",end:">"}]}]}),e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE,{className:"meta",begin:"#",end:"$",keywords:{keyword:"if else elif endif define undef warning error line region endregion pragma checksum"}},d,r,{beginKeywords:"class interface",relevance:0,end:/[{;=]/,illegal:/[^\s:,]/,contains:[{beginKeywords:"where class"},n,p,e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE]},{beginKeywords:"namespace",relevance:0,end:/[{;=]/,illegal:/[^\s:]/,contains:[n,e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE]},{beginKeywords:"record",relevance:0,end:/[{;=]/,illegal:/[^\s:]/,contains:[n,p,e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE]},{className:"meta",begin:"^\\s*\\[(?=[\\w])",excludeBegin:!0,end:"\\]",excludeEnd:!0,contains:[{className:"string",begin:/"/,end:/"/}]},{beginKeywords:"new return throw await else",relevance:0},{className:"function",begin:"("+h+"\\s+)+"+e.IDENT_RE+"\\s*(<[^=]+>\\s*)?\\(",returnBegin:!0,end:/\s*[{;=]/,excludeEnd:!0,keywords:t,contains:[{beginKeywords:["public","private","protected","static","internal","protected","abstract","async","extern","override","unsafe","virtual","new","sealed","partial"].join(" "),relevance:0},{begin:e.IDENT_RE+"\\s*(<[^=]+>\\s*)?\\(",returnBegin:!0,contains:[e.TITLE_MODE,p],relevance:0},{match:/\(\)/},{className:"params",begin:/\(/,end:/\)/,excludeBegin:!0,excludeEnd:!0,keywords:t,relevance:0,contains:[d,r,e.C_BLOCK_COMMENT_MODE]},e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE]},f]}},elixir:function(e){const t=e.regex,n="[a-zA-Z_][a-zA-Z0-9_.]*(!|\\?)?",r={$pattern:n,keyword:["after","alias","and","case","catch","cond","defstruct","defguard","do","else","end","fn","for","if","import","in","not","or","quote","raise","receive","require","reraise","rescue","try","unless","unquote","unquote_splicing","use","when","with|0"],literal:["false","nil","true"]},a={className:"subst",begin:/#\{/,end:/\}/,keywords:r},o={match:/\\[\s\S]/,scope:"char.escape",relevance:0},i="[/|([{<\"']",s=[{begin:/"/,end:/"/},{begin:/'/,end:/'/},{begin:/\//,end:/\//},{begin:/\|/,end:/\|/},{begin:/\(/,end:/\)/},{begin:/\[/,end:/\]/},{begin:/\{/,end:/\}/},{begin:/</,end:/>/}],l=e=>({scope:"char.escape",begin:t.concat(/\\/,e),relevance:0}),c={className:"string",begin:"~[a-z](?="+i+")",contains:s.map((t=>e.inherit(t,{contains:[l(t.end),o,a]})))},u={className:"string",begin:"~[A-Z](?="+i+")",contains:s.map((t=>e.inherit(t,{contains:[l(t.end)]})))},d={className:"regex",variants:[{begin:"~r(?="+i+")",contains:s.map((n=>e.inherit(n,{end:t.concat(n.end,/[uismxfU]{0,7}/),contains:[l(n.end),o,a]})))},{begin:"~R(?="+i+")",contains:s.map((n=>e.inherit(n,{end:t.concat(n.end,/[uismxfU]{0,7}/),contains:[l(n.end)]})))}]},p={className:"string",contains:[e.BACKSLASH_ESCAPE,a],variants:[{begin:/"""/,end:/"""/},{begin:/'''/,end:/'''/},{begin:/~S"""/,end:/"""/,contains:[]},{begin:/~S"/,end:/"/,contains:[]},{begin:/~S'''/,end:/'''/,contains:[]},{begin:/~S'/,end:/'/,contains:[]},{begin:/'/,end:/'/},{begin:/"/,end:/"/}]},h={className:"function",beginKeywords:"def defp defmacro defmacrop",end:/\B\b/,contains:[e.inherit(e.TITLE_MODE,{begin:n,endsParent:!0})]},f=e.inherit(h,{className:"class",beginKeywords:"defimpl defmodule defprotocol defrecord",end:/\bdo\b|$|;/}),m=[p,d,u,c,e.HASH_COMMENT_MODE,f,h,{begin:"::"},{className:"symbol",begin:":(?![\\s:])",contains:[p,{begin:"[a-zA-Z_]\\w*[!?=]?|[-+~]@|<<|>>|=~|===?|<=>|[<>]=?|\\*\\*|[-/+%^&*~`|]|\\[\\]=?"}],relevance:0},{className:"symbol",begin:n+":(?!:)",relevance:0},{className:"title.class",begin:/(\b[A-Z][a-zA-Z0-9_]+)/,relevance:0},{className:"number",begin:"(\\b0o[0-7_]+)|(\\b0b[01_]+)|(\\b0x[0-9a-fA-F_]+)|(-?\\b[0-9][0-9_]*(\\.[0-9_]+([eE][-+]?[0-9]+)?)?)",relevance:0},{className:"variable",begin:"(\\$\\W)|((\\$|@@?)(\\w+))"}];return a.contains=m,{name:"Elixir",aliases:["ex","exs"],keywords:r,contains:m}},go:function(e){const t={keyword:["break","case","chan","const","continue","default","defer","else","fallthrough","for","func","go","goto","if","import","interface","map","package","range","return","select","struct","switch","type","var"],type:["bool","byte","complex64","complex128","error","float32","float64","int8","int16","int32","int64","string","uint8","uint16","uint32","uint64","int","uint","uintptr","rune"],literal:["true","false","iota","nil"],built_in:["append","cap","close","complex","copy","imag","len","make","new","panic","print","println","real","recover","delete"]};return{name:"Go",aliases:["golang"],keywords:t,illegal:"</",contains:[e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE,{className:"string",variants:[e.QUOTE_STRING_MODE,e.APOS_STRING_MODE,{begin:"`",end:"`"}]},{className:"number",variants:[{match:/-?\b0[xX]\.[a-fA-F0-9](_?[a-fA-F0-9])*[pP][+-]?\d(_?\d)*i?/,relevance:0},{match:/-?\b0[xX](_?[a-fA-F0-9])+((\.([a-fA-F0-9](_?[a-fA-F0-9])*)?)?[pP][+-]?\d(_?\d)*)?i?/,relevance:0},{match:/-?\b0[oO](_?[0-7])*i?/,relevance:0},{match:/-?\.\d(_?\d)*([eE][+-]?\d(_?\d)*)?i?/,relevance:0},{match:/-?\b\d(_?\d)*(\.(\d(_?\d)*)?)?([eE][+-]?\d(_?\d)*)?i?/,relevance:0}]},{begin:/:=/},{className:"function",beginKeywords:"func",end:"\\s*(\\{|$)",excludeEnd:!0,contains:[e.TITLE_MODE,{className:"params",begin:/\(/,end:/\)/,endsParent:!0,keywords:t,illegal:/["']/}]}]}},graphql:function(e){const t=e.regex;return{name:"GraphQL",aliases:["gql"],case_insensitive:!0,disableAutodetect:!1,keywords:{keyword:["query","mutation","subscription","type","input","schema","directive","interface","union","scalar","fragment","enum","on"],literal:["true","false","null"]},contains:[e.HASH_COMMENT_MODE,e.QUOTE_STRING_MODE,e.NUMBER_MODE,{scope:"punctuation",match:/[.]{3}/,relevance:0},{scope:"punctuation",begin:/[\!\(\)\:\=\[\]\{\|\}]{1}/,relevance:0},{scope:"variable",begin:/\$/,end:/\W/,excludeEnd:!0,relevance:0},{scope:"meta",match:/@\w+/,excludeEnd:!0},{scope:"symbol",begin:t.concat(/[_A-Za-z][_0-9A-Za-z]*/,t.lookahead(/\s*:/)),relevance:0}],illegal:[/[;<']/,/BEGIN/]}},html:pe,java:function(e){const t=e.regex,n="[À-ʸa-zA-Z_$][À-ʸa-zA-Z_$0-9]*",r=n+p("(?:<"+n+"~~~(?:\\s*,\\s*"+n+"~~~)*>)?",/~~~/g,2),a={keyword:["synchronized","abstract","private","var","static","if","const ","for","while","strictfp","finally","protected","import","native","final","void","enum","else","break","transient","catch","instanceof","volatile","case","assert","package","default","public","try","switch","continue","throws","protected","public","private","module","requires","exports","do","sealed","yield","permits","goto","when"],literal:["false","true","null"],type:["char","boolean","long","float","int","byte","short","double"],built_in:["super","this"]},o={className:"meta",begin:"@"+n,contains:[{begin:/\(/,end:/\)/,contains:["self"]}]},i={className:"params",begin:/\(/,end:/\)/,keywords:a,relevance:0,contains:[e.C_BLOCK_COMMENT_MODE],endsParent:!0};return{name:"Java",aliases:["jsp"],keywords:a,illegal:/<\/|#/,contains:[e.COMMENT("/\\*\\*","\\*/",{relevance:0,contains:[{begin:/\w+@/,relevance:0},{className:"doctag",begin:"@[A-Za-z]+"}]}),{begin:/import java\.[a-z]+\./,keywords:"import",relevance:2},e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE,{begin:/"""/,end:/"""/,className:"string",contains:[e.BACKSLASH_ESCAPE]},e.APOS_STRING_MODE,e.QUOTE_STRING_MODE,{match:[/\b(?:class|interface|enum|extends|implements|new)/,/\s+/,n],className:{1:"keyword",3:"title.class"}},{match:/non-sealed/,scope:"keyword"},{begin:[t.concat(/(?!else)/,n),/\s+/,n,/\s+/,/=(?!=)/],className:{1:"type",3:"variable",5:"operator"}},{begin:[/record/,/\s+/,n],className:{1:"keyword",3:"title.class"},contains:[i,e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE]},{beginKeywords:"new throw return else",relevance:0},{begin:["(?:"+r+"\\s+)",e.UNDERSCORE_IDENT_RE,/\s*(?=\()/],className:{2:"title.function"},keywords:a,contains:[{className:"params",begin:/\(/,end:/\)/,keywords:a,relevance:0,contains:[o,e.APOS_STRING_MODE,e.QUOTE_STRING_MODE,d,e.C_BLOCK_COMMENT_MODE]},e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE]},d,o]}},javascript:function(e){const t=e.regex,n=h,r={begin:/<[A-Za-z0-9\\._:-]+/,end:/\/[A-Za-z0-9\\._:-]+>|\/>/,isTrulyOpeningTag:(e,t)=>{const n=e[0].length+e.index,r=e.input[n];if("<"===r||","===r)return void t.ignoreMatch();let a;">"===r&&(((e,{after:t})=>{const n="</"+e[0].slice(1);return-1!==e.input.indexOf(n,t)})(e,{after:n})||t.ignoreMatch());const o=e.input.substring(n);((a=o.match(/^\s*=/))||(a=o.match(/^\s+extends\s+/))&&0===a.index)&&t.ignoreMatch()}},a={$pattern:h,keyword:f,literal:m,built_in:w,"variable.language":y},o="[0-9](_?[0-9])*",i=`\\.(${o})`,s="0|[1-9](_?[0-9])*|0[0-7]*[89][0-9]*",l={className:"number",variants:[{begin:`(\\b(${s})((${i})|\\.)?|(${i}))[eE][+-]?(${o})\\b`},{begin:`\\b(${s})\\b((${i})\\b|\\.)?|(${i})\\b`},{begin:"\\b(0|[1-9](_?[0-9])*)n\\b"},{begin:"\\b0[xX][0-9a-fA-F](_?[0-9a-fA-F])*n?\\b"},{begin:"\\b0[bB][0-1](_?[0-1])*n?\\b"},{begin:"\\b0[oO][0-7](_?[0-7])*n?\\b"},{begin:"\\b0[0-7]+n?\\b"}],relevance:0},c={className:"subst",begin:"\\$\\{",end:"\\}",keywords:a,contains:[]},u={begin:".?html`",end:"",starts:{end:"`",returnEnd:!1,contains:[e.BACKSLASH_ESCAPE,c],subLanguage:"xml"}},d={begin:".?css`",end:"",starts:{end:"`",returnEnd:!1,contains:[e.BACKSLASH_ESCAPE,c],subLanguage:"css"}},p={begin:".?gql`",end:"",starts:{end:"`",returnEnd:!1,contains:[e.BACKSLASH_ESCAPE,c],subLanguage:"graphql"}},O={className:"string",begin:"`",end:"`",contains:[e.BACKSLASH_ESCAPE,c]},k={className:"comment",variants:[e.COMMENT(/\/\*\*(?!\/)/,"\\*/",{relevance:0,contains:[{begin:"(?=@[A-Za-z]+)",relevance:0,contains:[{className:"doctag",begin:"@[A-Za-z]+"},{className:"type",begin:"\\{",end:"\\}",excludeEnd:!0,excludeBegin:!0,relevance:0},{className:"variable",begin:n+"(?=\\s*(-)|$)",endsParent:!0,relevance:0},{begin:/(?=[^\n])\s/,relevance:0}]}]}),e.C_BLOCK_COMMENT_MODE,e.C_LINE_COMMENT_MODE]},x=[e.APOS_STRING_MODE,e.QUOTE_STRING_MODE,u,d,p,O,{match:/\$\d+/},l];c.contains=x.concat({begin:/\{/,end:/\}/,keywords:a,contains:["self"].concat(x)});const E=[].concat(k,c.contains),S=E.concat([{begin:/(\s*)\(/,end:/\)/,keywords:a,contains:["self"].concat(E)}]),_={className:"params",begin:/(\s*)\(/,end:/\)/,excludeBegin:!0,excludeEnd:!0,keywords:a,contains:S},C={variants:[{match:[/class/,/\s+/,n,/\s+/,/extends/,/\s+/,t.concat(n,"(",t.concat(/\./,n),")*")],scope:{1:"keyword",3:"title.class",5:"keyword",7:"title.class.inherited"}},{match:[/class/,/\s+/,n],scope:{1:"keyword",3:"title.class"}}]},T={relevance:0,match:t.either(/\bJSON/,/\b[A-Z][a-z]+([A-Z][a-z]*|\d)*/,/\b[A-Z]{2,}([A-Z][a-z]+|\d)+([A-Z][a-z]*)*/,/\b[A-Z]{2,}[a-z]+([A-Z][a-z]+|\d)*([A-Z][a-z]*)*/),className:"title.class",keywords:{_:[...g,...v]}},A={variants:[{match:[/function/,/\s+/,n,/(?=\s*\()/]},{match:[/function/,/\s*(?=\()/]}],className:{1:"keyword",3:"title.function"},label:"func.def",contains:[_],illegal:/%/},R={match:t.concat(/\b/,(P=[...b,"super","import"].map((e=>`${e}\\s*\\(`)),t.concat("(?!",P.join("|"),")")),n,t.lookahead(/\s*\(/)),className:"title.function",relevance:0};var P;const D={begin:t.concat(/\./,t.lookahead(t.concat(n,/(?![0-9A-Za-z$_(])/))),end:n,excludeBegin:!0,keywords:"prototype",className:"property",relevance:0},M={match:[/get|set/,/\s+/,n,/(?=\()/],className:{1:"keyword",3:"title.function"},contains:[{begin:/\(\)/},_]},$="(\\([^()]*(\\([^()]*(\\([^()]*\\)[^()]*)*\\)[^()]*)*\\)|"+e.UNDERSCORE_IDENT_RE+")\\s*=>",N={match:[/const|var|let/,/\s+/,n,/\s*/,/=\s*/,/(async\s*)?/,t.lookahead($)],keywords:"async",className:{1:"keyword",3:"title.function"},contains:[_]};return{name:"JavaScript",aliases:["js","jsx","mjs","cjs"],keywords:a,exports:{PARAMS_CONTAINS:S,CLASS_REFERENCE:T},illegal:/#(?![$_A-z])/,contains:[e.SHEBANG({label:"shebang",binary:"node",relevance:5}),{label:"use_strict",className:"meta",relevance:10,begin:/^\s*['"]use (strict|asm)['"]/},e.APOS_STRING_MODE,e.QUOTE_STRING_MODE,u,d,p,O,k,{match:/\$\d+/},l,T,{scope:"attr",match:n+t.lookahead(":"),relevance:0},N,{begin:"("+e.RE_STARTERS_RE+"|\\b(case|return|throw)\\b)\\s*",keywords:"return throw case",relevance:0,contains:[k,e.REGEXP_MODE,{className:"function",begin:$,returnBegin:!0,end:"\\s*=>",contains:[{className:"params",variants:[{begin:e.UNDERSCORE_IDENT_RE,relevance:0},{className:null,begin:/\(\s*\)/,skip:!0},{begin:/(\s*)\(/,end:/\)/,excludeBegin:!0,excludeEnd:!0,keywords:a,contains:S}]}]},{begin:/,/,relevance:0},{match:/\s+/,relevance:0},{variants:[{begin:"<>",end:"</>"},{match:/<[A-Za-z0-9\\._:-]+\s*\/>/},{begin:r.begin,"on:begin":r.isTrulyOpeningTag,end:r.end}],subLanguage:"xml",contains:[{begin:r.begin,end:r.end,skip:!0,contains:["self"]}]}]},A,{beginKeywords:"while if switch catch for"},{begin:"\\b(?!function)"+e.UNDERSCORE_IDENT_RE+"\\([^()]*(\\([^()]*(\\([^()]*\\)[^()]*)*\\)[^()]*)*\\)\\s*\\{",returnBegin:!0,label:"func.def",contains:[_,e.inherit(e.TITLE_MODE,{begin:n,className:"title.function"})]},{match:/\.\.\./,relevance:0},D,{match:"\\$"+n,relevance:0},{match:[/\bconstructor(?=\s*\()/],className:{1:"title.function"},contains:[_]},R,{relevance:0,match:/\b[A-Z][A-Z_0-9]+\b/,className:"variable.constant"},C,M,{match:/\$[(.]/}]}},json:function(e){const t=["true","false","null"],n={scope:"literal",beginKeywords:t.join(" ")};return{name:"JSON",aliases:["jsonc"],keywords:{literal:t},contains:[{className:"attr",begin:/"(\\.|[^\\"\r\n])*"(?=\s*:)/,relevance:1.01},{match:/[{}[\],:]/,className:"punctuation",relevance:0},e.QUOTE_STRING_MODE,n,e.C_NUMBER_MODE,e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE],illegal:"\\S"}},kotlin:function(e){const t={keyword:"abstract as val var vararg get set class object open private protected public noinline crossinline dynamic final enum if else do while for when throw try catch finally import package is in fun override companion reified inline lateinit init interface annotation data sealed internal infix operator out by constructor super tailrec where const inner suspend typealias external expect actual",built_in:"Byte Short Char Int Long Boolean Float Double Void Unit Nothing",literal:"true false null"},n={className:"symbol",begin:e.UNDERSCORE_IDENT_RE+"@"},r={className:"subst",begin:/\$\{/,end:/\}/,contains:[e.C_NUMBER_MODE]},a={className:"variable",begin:"\\$"+e.UNDERSCORE_IDENT_RE},o={className:"string",variants:[{begin:'"""',end:'"""(?=[^"])',contains:[a,r]},{begin:"'",end:"'",illegal:/\n/,contains:[e.BACKSLASH_ESCAPE]},{begin:'"',end:'"',illegal:/\n/,contains:[e.BACKSLASH_ESCAPE,a,r]}]};r.contains.push(o);const i={className:"meta",begin:"@(?:file|property|field|get|set|receiver|param|setparam|delegate)\\s*:(?:\\s*"+e.UNDERSCORE_IDENT_RE+")?"},s={className:"meta",begin:"@"+e.UNDERSCORE_IDENT_RE,contains:[{begin:/\(/,end:/\)/,contains:[e.inherit(o,{className:"string"}),"self"]}]},l=E,c=e.COMMENT("/\\*","\\*/",{contains:[e.C_BLOCK_COMMENT_MODE]}),u={variants:[{className:"type",begin:e.UNDERSCORE_IDENT_RE},{begin:/\(/,end:/\)/,contains:[]}]},d=u;return d.variants[1].contains=[u],u.variants[1].contains=[d],{name:"Kotlin",aliases:["kt","kts"],keywords:t,contains:[e.COMMENT("/\\*\\*","\\*/",{relevance:0,contains:[{className:"doctag",begin:"@[A-Za-z]+"}]}),e.C_LINE_COMMENT_MODE,c,{className:"keyword",begin:/\b(break|continue|return|this)\b/,starts:{contains:[{className:"symbol",begin:/@\w+/}]}},n,i,s,{className:"function",beginKeywords:"fun",end:"[(]|$",returnBegin:!0,excludeEnd:!0,keywords:t,relevance:5,contains:[{begin:e.UNDERSCORE_IDENT_RE+"\\s*\\(",returnBegin:!0,relevance:0,contains:[e.UNDERSCORE_TITLE_MODE]},{className:"type",begin:/</,end:/>/,keywords:"reified",relevance:0},{className:"params",begin:/\(/,end:/\)/,endsParent:!0,keywords:t,relevance:0,contains:[{begin:/:/,end:/[=,\/]/,endsWithParent:!0,contains:[u,e.C_LINE_COMMENT_MODE,c],relevance:0},e.C_LINE_COMMENT_MODE,c,i,s,o,e.C_NUMBER_MODE]},c]},{begin:[/class|interface|trait/,/\s+/,e.UNDERSCORE_IDENT_RE],beginScope:{3:"title.class"},keywords:"class interface trait",end:/[:\{(]|$/,excludeEnd:!0,illegal:"extends implements",contains:[{beginKeywords:"public protected internal private constructor"},e.UNDERSCORE_TITLE_MODE,{className:"type",begin:/</,end:/>/,excludeBegin:!0,excludeEnd:!0,relevance:0},{className:"type",begin:/[,:]\s*/,end:/[<\(,){\s]|$/,excludeBegin:!0,returnEnd:!0},i,s]},o,{className:"meta",begin:"^#!/usr/bin/env",end:"$",illegal:"\n"},l]}},less:function(e){const t=(e=>({IMPORTANT:{scope:"meta",begin:"!important"},BLOCK_COMMENT:e.C_BLOCK_COMMENT_MODE,HEXCOLOR:{scope:"number",begin:/#(([0-9a-fA-F]{3,4})|(([0-9a-fA-F]{2}){3,4}))\b/},FUNCTION_DISPATCH:{className:"built_in",begin:/[\w-]+(?=\()/},ATTRIBUTE_SELECTOR_MODE:{scope:"selector-attr",begin:/\[/,end:/\]/,illegal:"$",contains:[e.APOS_STRING_MODE,e.QUOTE_STRING_MODE]},CSS_NUMBER_MODE:{scope:"number",begin:e.NUMBER_RE+"(%|em|ex|ch|rem|vw|vh|vmin|vmax|cm|mm|in|pt|pc|px|deg|grad|rad|turn|s|ms|Hz|kHz|dpi|dpcm|dppx)?",relevance:0},CSS_VARIABLE:{className:"attr",begin:/--[A-Za-z_][A-Za-z0-9_-]*/}}))(e),n=R,r="[\\w-]+",a="("+r+"|@\\{"+r+"\\})",o=[],i=[],s=function(e){return{className:"string",begin:"~?"+e+".*?"+e}},l=function(e,t,n){return{className:e,begin:t,relevance:n}},c={$pattern:/[a-z-]+/,keyword:"and or not only",attribute:_.join(" ")},u={begin:"\\(",end:"\\)",contains:i,keywords:c,relevance:0};i.push(e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE,s("'"),s('"'),t.CSS_NUMBER_MODE,{begin:"(url|data-uri)\\(",starts:{className:"string",end:"[\\)\\n]",excludeEnd:!0}},t.HEXCOLOR,u,l("variable","@@?"+r,10),l("variable","@\\{"+r+"\\}"),l("built_in","~?`[^`]*?`"),{className:"attribute",begin:r+"\\s*:",end:":",returnBegin:!0,excludeEnd:!0},t.IMPORTANT,{beginKeywords:"and not"},t.FUNCTION_DISPATCH);const d=i.concat({begin:/\{/,end:/\}/,contains:o}),p={beginKeywords:"when",endsWithParent:!0,contains:[{beginKeywords:"and not"}].concat(i)},h={begin:a+"\\s*:",returnBegin:!0,end:/[;}]/,relevance:0,contains:[{begin:/-(webkit|moz|ms|o)-/},t.CSS_VARIABLE,{className:"attribute",begin:"\\b("+A.join("|")+")\\b",end:/(?=:)/,starts:{endsWithParent:!0,illegal:"[<=$]",relevance:0,contains:i}}]},f={className:"keyword",begin:"@(import|media|charset|font-face|(-[a-z]+-)?keyframes|supports|document|namespace|page|viewport|host)\\b",starts:{end:"[;{}]",keywords:c,returnEnd:!0,contains:i,relevance:0}},m={className:"variable",variants:[{begin:"@"+r+"\\s*:",relevance:15},{begin:"@"+r}],starts:{end:"[;}]",returnEnd:!0,contains:d}},g={variants:[{begin:"[\\.#:&\\[>]",end:"[;{}]"},{begin:a,end:/\{/}],returnBegin:!0,returnEnd:!0,illegal:"[<='$\"]",relevance:0,contains:[e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE,p,l("keyword","all\\b"),l("variable","@\\{"+r+"\\}"),{begin:"\\b("+S.join("|")+")\\b",className:"selector-tag"},t.CSS_NUMBER_MODE,l("selector-tag",a,0),l("selector-id","#"+a),l("selector-class","\\."+a,0),l("selector-tag","&",0),t.ATTRIBUTE_SELECTOR_MODE,{className:"selector-pseudo",begin:":("+C.join("|")+")"},{className:"selector-pseudo",begin:":(:)?("+T.join("|")+")"},{begin:/\(/,end:/\)/,relevance:0,contains:d},{begin:"!important"},t.FUNCTION_DISPATCH]},v={begin:r+":(:)?"+`(${n.join("|")})`,returnBegin:!0,contains:[g]};return o.push(e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE,f,m,v,h,g,p,t.FUNCTION_DISPATCH),{name:"Less",case_insensitive:!0,illegal:"[=>'/<($\"]",contains:o}},makefile:function(e){const t={className:"variable",variants:[{begin:"\\$\\("+e.UNDERSCORE_IDENT_RE+"\\)",contains:[e.BACKSLASH_ESCAPE]},{begin:/\$[@%<?\^\+\*]/}]},n={className:"string",begin:/"/,end:/"/,contains:[e.BACKSLASH_ESCAPE,t]},r={className:"variable",begin:/\$\([\w-]+\s/,end:/\)/,keywords:{built_in:"subst patsubst strip findstring filter filter-out sort word wordlist firstword lastword dir notdir suffix basename addsuffix addprefix join wildcard realpath abspath error warning shell origin flavor foreach if or and call eval file value"},contains:[t,n]},a={begin:"^"+e.UNDERSCORE_IDENT_RE+"\\s*(?=[:+?]?=)"},o={className:"section",begin:/^[^\s]+:/,end:/$/,contains:[t]};return{name:"Makefile",aliases:["mk","mak","make"],keywords:{$pattern:/[\w-]+/,keyword:"define endef undefine ifdef ifndef ifeq ifneq else endif include -include sinclude override export unexport private vpath"},contains:[e.HASH_COMMENT_MODE,t,n,r,a,{className:"meta",begin:/^\.PHONY:/,end:/$/,keywords:{$pattern:/[\.\w]+/,keyword:".PHONY"}},o]}},markdown:function(e){const t={begin:/<\/?[A-Za-z_]/,end:">",subLanguage:"xml",relevance:0},n={variants:[{begin:/\[.+?\]\[.*?\]/,relevance:0},{begin:/\[.+?\]\(((data|javascript|mailto):|(?:http|ftp)s?:\/\/).*?\)/,relevance:2},{begin:e.regex.concat(/\[.+?\]\(/,/[A-Za-z][A-Za-z0-9+.-]*/,/:\/\/.*?\)/),relevance:2},{begin:/\[.+?\]\([./?&#].*?\)/,relevance:1},{begin:/\[.*?\]\(.*?\)/,relevance:0}],returnBegin:!0,contains:[{match:/\[(?=\])/},{className:"string",relevance:0,begin:"\\[",end:"\\]",excludeBegin:!0,returnEnd:!0},{className:"link",relevance:0,begin:"\\]\\(",end:"\\)",excludeBegin:!0,excludeEnd:!0},{className:"symbol",relevance:0,begin:"\\]\\[",end:"\\]",excludeBegin:!0,excludeEnd:!0}]},r={className:"strong",contains:[],variants:[{begin:/_{2}(?!\s)/,end:/_{2}/},{begin:/\*{2}(?!\s)/,end:/\*{2}/}]},a={className:"emphasis",contains:[],variants:[{begin:/\*(?![*\s])/,end:/\*/},{begin:/_(?![_\s])/,end:/_/,relevance:0}]},o=e.inherit(r,{contains:[]}),i=e.inherit(a,{contains:[]});r.contains.push(i),a.contains.push(o);let s=[t,n];return[r,a,o,i].forEach((e=>{e.contains=e.contains.concat(s)})),s=s.concat(r,a),{name:"Markdown",aliases:["md","mkdown","mkd"],contains:[{className:"section",variants:[{begin:"^#{1,6}",end:"$",contains:s},{begin:"(?=^.+?\\n[=-]{2,}$)",contains:[{begin:"^[=-]*$"},{begin:"^",end:"\\n",contains:s}]}]},t,{className:"bullet",begin:"^[ \t]*([*+-]|(\\d+\\.))(?=\\s+)",end:"\\s+",excludeEnd:!0},r,a,{className:"quote",begin:"^>\\s+",contains:s,end:"$"},{className:"code",variants:[{begin:"(`{3,})[^`](.|\\n)*?\\1`*[ ]*"},{begin:"(~{3,})[^~](.|\\n)*?\\1~*[ ]*"},{begin:"```",end:"```+[ ]*$"},{begin:"~~~",end:"~~~+[ ]*$"},{begin:"`.+?`"},{begin:"(?=^( {4}|\\t))",contains:[{begin:"^( {4}|\\t)",end:"(\\n)$"}],relevance:0}]},{begin:"^[-\\*]{3,}",end:"$"},n,{begin:/^\[[^\n]+\]:/,returnBegin:!0,contains:[{className:"symbol",begin:/\[/,end:/\]/,excludeBegin:!0,excludeEnd:!0},{className:"link",begin:/:\s*/,end:/$/,excludeBegin:!0}]},{scope:"literal",match:/&([a-zA-Z0-9]+|#[0-9]{1,7}|#[Xx][0-9a-fA-F]{1,6});/}]}},matlab:function(e){const t="('|\\.')+",n={relevance:0,contains:[{begin:t}]};return{name:"Matlab",keywords:{keyword:"arguments break case catch classdef continue else elseif end enumeration events for function global if methods otherwise parfor persistent properties return spmd switch try while",built_in:"sin sind sinh asin asind asinh cos cosd cosh acos acosd acosh tan tand tanh atan atand atan2 atanh sec secd sech asec asecd asech csc cscd csch acsc acscd acsch cot cotd coth acot acotd acoth hypot exp expm1 log log1p log10 log2 pow2 realpow reallog realsqrt sqrt nthroot nextpow2 abs angle complex conj imag real unwrap isreal cplxpair fix floor ceil round mod rem sign airy besselj bessely besselh besseli besselk beta betainc betaln ellipj ellipke erf erfc erfcx erfinv expint gamma gammainc gammaln psi legendre cross dot factor isprime primes gcd lcm rat rats perms nchoosek factorial cart2sph cart2pol pol2cart sph2cart hsv2rgb rgb2hsv zeros ones eye repmat rand randn linspace logspace freqspace meshgrid accumarray size length ndims numel disp isempty isequal isequalwithequalnans cat reshape diag blkdiag tril triu fliplr flipud flipdim rot90 find sub2ind ind2sub bsxfun ndgrid permute ipermute shiftdim circshift squeeze isscalar isvector ans eps realmax realmin pi i|0 inf nan isnan isinf isfinite j|0 why compan gallery hadamard hankel hilb invhilb magic pascal rosser toeplitz vander wilkinson max min nanmax nanmin mean nanmean type table readtable writetable sortrows sort figure plot plot3 scatter scatter3 cellfun legend intersect ismember procrustes hold num2cell "},illegal:'(//|"|#|/\\*|\\s+/\\w+)',contains:[{className:"function",beginKeywords:"function",end:"$",contains:[e.UNDERSCORE_TITLE_MODE,{className:"params",variants:[{begin:"\\(",end:"\\)"},{begin:"\\[",end:"\\]"}]}]},{className:"built_in",begin:/true|false/,relevance:0,starts:n},{begin:"[a-zA-Z][a-zA-Z_0-9]*"+t,relevance:0},{className:"number",begin:e.C_NUMBER_RE,relevance:0,starts:n},{className:"string",begin:"'",end:"'",contains:[{begin:"''"}]},{begin:/\]|\}|\)/,relevance:0,starts:n},{className:"string",begin:'"',end:'"',contains:[{begin:'""'}],starts:n},e.COMMENT("^\\s*%\\{\\s*$","^\\s*%\\}\\s*$"),e.COMMENT("%","$")]}},nginx:function(e){const t=e.regex,n={className:"variable",variants:[{begin:/\$\d+/},{begin:/\$\{\w+\}/},{begin:t.concat(/[$@]/,e.UNDERSCORE_IDENT_RE)}]},r={endsWithParent:!0,keywords:{$pattern:/[a-z_]{2,}|\/dev\/poll/,literal:["on","off","yes","no","true","false","none","blocked","debug","info","notice","warn","error","crit","select","break","last","permanent","redirect","kqueue","rtsig","epoll","poll","/dev/poll"]},relevance:0,illegal:"=>",contains:[e.HASH_COMMENT_MODE,{className:"string",contains:[e.BACKSLASH_ESCAPE,n],variants:[{begin:/"/,end:/"/},{begin:/'/,end:/'/}]},{begin:"([a-z]+):/",end:"\\s",endsWithParent:!0,excludeEnd:!0,contains:[n]},{className:"regexp",contains:[e.BACKSLASH_ESCAPE,n],variants:[{begin:"\\s\\^",end:"\\s|\\{|;",returnEnd:!0},{begin:"~\\*?\\s+",end:"\\s|\\{|;",returnEnd:!0},{begin:"\\*(\\.[a-z\\-]+)+"},{begin:"([a-z\\-]+\\.)+\\*"}]},{className:"number",begin:"\\b\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}(:\\d{1,5})?\\b"},{className:"number",begin:"\\b\\d+[kKmMgGdshdwy]?\\b",relevance:0},n]};return{name:"Nginx config",aliases:["nginxconf"],contains:[e.HASH_COMMENT_MODE,{beginKeywords:"upstream location",end:/;|\{/,contains:r.contains,keywords:{section:"upstream location"}},{className:"section",begin:t.concat(e.UNDERSCORE_IDENT_RE+t.lookahead(/\s+\{/)),relevance:0},{begin:t.lookahead(e.UNDERSCORE_IDENT_RE+"\\s"),end:";|\\{",contains:[{className:"attribute",begin:e.UNDERSCORE_IDENT_RE,starts:r}],relevance:0}],illegal:"[^\\s\\}\\{]"}},objectivec:function(e){const t=/[a-zA-Z@][a-zA-Z0-9_]*/,n={$pattern:t,keyword:["@interface","@class","@protocol","@implementation"]};return{name:"Objective-C",aliases:["mm","objc","obj-c","obj-c++","objective-c++"],keywords:{"variable.language":["this","super"],$pattern:t,keyword:["while","export","sizeof","typedef","const","struct","for","union","volatile","static","mutable","if","do","return","goto","enum","else","break","extern","asm","case","default","register","explicit","typename","switch","continue","inline","readonly","assign","readwrite","self","@synchronized","id","typeof","nonatomic","IBOutlet","IBAction","strong","weak","copy","in","out","inout","bycopy","byref","oneway","__strong","__weak","__block","__autoreleasing","@private","@protected","@public","@try","@property","@end","@throw","@catch","@finally","@autoreleasepool","@synthesize","@dynamic","@selector","@optional","@required","@encode","@package","@import","@defs","@compatibility_alias","__bridge","__bridge_transfer","__bridge_retained","__bridge_retain","__covariant","__contravariant","__kindof","_Nonnull","_Nullable","_Null_unspecified","__FUNCTION__","__PRETTY_FUNCTION__","__attribute__","getter","setter","retain","unsafe_unretained","nonnull","nullable","null_unspecified","null_resettable","class","instancetype","NS_DESIGNATED_INITIALIZER","NS_UNAVAILABLE","NS_REQUIRES_SUPER","NS_RETURNS_INNER_POINTER","NS_INLINE","NS_AVAILABLE","NS_DEPRECATED","NS_ENUM","NS_OPTIONS","NS_SWIFT_UNAVAILABLE","NS_ASSUME_NONNULL_BEGIN","NS_ASSUME_NONNULL_END","NS_REFINED_FOR_SWIFT","NS_SWIFT_NAME","NS_SWIFT_NOTHROW","NS_DURING","NS_HANDLER","NS_ENDHANDLER","NS_VALUERETURN","NS_VOIDRETURN"],literal:["false","true","FALSE","TRUE","nil","YES","NO","NULL"],built_in:["dispatch_once_t","dispatch_queue_t","dispatch_sync","dispatch_async","dispatch_once"],type:["int","float","char","unsigned","signed","short","long","double","wchar_t","unichar","void","bool","BOOL","id|0","_Bool"]},illegal:"</",contains:[{className:"built_in",begin:"\\b(AV|CA|CF|CG|CI|CL|CM|CN|CT|MK|MP|MTK|MTL|NS|SCN|SK|UI|WK|XC)\\w+"},e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE,e.C_NUMBER_MODE,e.QUOTE_STRING_MODE,e.APOS_STRING_MODE,{className:"string",variants:[{begin:'@"',end:'"',illegal:"\\n",contains:[e.BACKSLASH_ESCAPE]}]},{className:"meta",begin:/#\s*[a-z]+\b/,end:/$/,keywords:{keyword:"if else elif endif define undef warning error line pragma ifdef ifndef include"},contains:[{begin:/\\\n/,relevance:0},e.inherit(e.QUOTE_STRING_MODE,{className:"string"}),{className:"string",begin:/<.*?>/,end:/$/,illegal:"\\n"},e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE]},{className:"class",begin:"("+n.keyword.join("|")+")\\b",end:/(\{|$)/,excludeEnd:!0,keywords:n,contains:[e.UNDERSCORE_TITLE_MODE]},{begin:"\\."+e.UNDERSCORE_IDENT_RE,relevance:0}]}},ocaml:function(e){return{name:"OCaml",aliases:["ml"],keywords:{$pattern:"[a-z_]\\w*!?",keyword:"and as assert asr begin class constraint do done downto else end exception external for fun function functor if in include inherit! inherit initializer land lazy let lor lsl lsr lxor match method!|10 method mod module mutable new object of open! open or private rec sig struct then to try type val! val virtual when while with parser value",built_in:"array bool bytes char exn|5 float int int32 int64 list lazy_t|5 nativeint|5 string unit in_channel out_channel ref",literal:"true false"},illegal:/\/\/|>>/,contains:[{className:"literal",begin:"\\[(\\|\\|)?\\]|\\(\\)",relevance:0},e.COMMENT("\\(\\*","\\*\\)",{contains:["self"]}),{className:"symbol",begin:"'[A-Za-z_](?!')[\\w']*"},{className:"type",begin:"`[A-Z][\\w']*"},{className:"type",begin:"\\b[A-Z][\\w']*",relevance:0},{begin:"[a-z_]\\w*'[\\w']*",relevance:0},e.inherit(e.APOS_STRING_MODE,{className:"string",relevance:0}),e.inherit(e.QUOTE_STRING_MODE,{illegal:null}),{className:"number",begin:"\\b(0[xX][a-fA-F0-9_]+[Lln]?|0[oO][0-7_]+[Lln]?|0[bB][01_]+[Lln]?|[0-9][0-9_]*([Lln]|(\\.[0-9_]*)?([eE][-+]?[0-9_]+)?)?)",relevance:0},{begin:/->/}]}},php:function(e){const t=e.regex,n=/(?![A-Za-z0-9])(?![$])/,r=t.concat(/[a-zA-Z_\x7f-\xff][a-zA-Z0-9_\x7f-\xff]*/,n),a=t.concat(/(\\?[A-Z][a-z0-9_\x7f-\xff]+|\\?[A-Z]+(?=[A-Z][a-z0-9_\x7f-\xff])){1,}/,n),o=t.concat(/[A-Z]+/,n),i={scope:"variable",match:"\\$+"+r},s={scope:"subst",variants:[{begin:/\$\w+/},{begin:/\{\$/,end:/\}/}]},l=e.inherit(e.APOS_STRING_MODE,{illegal:null}),c="[ \t\n]",u={scope:"string",variants:[e.inherit(e.QUOTE_STRING_MODE,{illegal:null,contains:e.QUOTE_STRING_MODE.contains.concat(s)}),l,{begin:/<<<[ \t]*(?:(\w+)|"(\w+)")\n/,end:/[ \t]*(\w+)\b/,contains:e.QUOTE_STRING_MODE.contains.concat(s),"on:begin":(e,t)=>{t.data._beginMatch=e[1]||e[2]},"on:end":(e,t)=>{t.data._beginMatch!==e[1]&&t.ignoreMatch()}},e.END_SAME_AS_BEGIN({begin:/<<<[ \t]*'(\w+)'\n/,end:/[ \t]*(\w+)\b/})]},d={scope:"number",variants:[{begin:"\\b0[bB][01]+(?:_[01]+)*\\b"},{begin:"\\b0[oO][0-7]+(?:_[0-7]+)*\\b"},{begin:"\\b0[xX][\\da-fA-F]+(?:_[\\da-fA-F]+)*\\b"},{begin:"(?:\\b\\d+(?:_\\d+)*(\\.(?:\\d+(?:_\\d+)*))?|\\B\\.\\d+)(?:[eE][+-]?\\d+)?"}],relevance:0},p=["false","null","true"],h=["__CLASS__","__DIR__","__FILE__","__FUNCTION__","__COMPILER_HALT_OFFSET__","__LINE__","__METHOD__","__NAMESPACE__","__TRAIT__","die","echo","exit","include","include_once","print","require","require_once","array","abstract","and","as","binary","bool","boolean","break","callable","case","catch","class","clone","const","continue","declare","default","do","double","else","elseif","empty","enddeclare","endfor","endforeach","endif","endswitch","endwhile","enum","eval","extends","final","finally","float","for","foreach","from","global","goto","if","implements","instanceof","insteadof","int","integer","interface","isset","iterable","list","match|0","mixed","new","never","object","or","private","protected","public","readonly","real","return","string","switch","throw","trait","try","unset","use","var","void","while","xor","yield"],f=["Error|0","AppendIterator","ArgumentCountError","ArithmeticError","ArrayIterator","ArrayObject","AssertionError","BadFunctionCallException","BadMethodCallException","CachingIterator","CallbackFilterIterator","CompileError","Countable","DirectoryIterator","DivisionByZeroError","DomainException","EmptyIterator","ErrorException","Exception","FilesystemIterator","FilterIterator","GlobIterator","InfiniteIterator","InvalidArgumentException","IteratorIterator","LengthException","LimitIterator","LogicException","MultipleIterator","NoRewindIterator","OutOfBoundsException","OutOfRangeException","OuterIterator","OverflowException","ParentIterator","ParseError","RangeException","RecursiveArrayIterator","RecursiveCachingIterator","RecursiveCallbackFilterIterator","RecursiveDirectoryIterator","RecursiveFilterIterator","RecursiveIterator","RecursiveIteratorIterator","RecursiveRegexIterator","RecursiveTreeIterator","RegexIterator","RuntimeException","SeekableIterator","SplDoublyLinkedList","SplFileInfo","SplFileObject","SplFixedArray","SplHeap","SplMaxHeap","SplMinHeap","SplObjectStorage","SplObserver","SplPriorityQueue","SplQueue","SplStack","SplSubject","SplTempFileObject","TypeError","UnderflowException","UnexpectedValueException","UnhandledMatchError","ArrayAccess","BackedEnum","Closure","Fiber","Generator","Iterator","IteratorAggregate","Serializable","Stringable","Throwable","Traversable","UnitEnum","WeakReference","WeakMap","Directory","__PHP_Incomplete_Class","parent","php_user_filter","self","static","stdClass"],m={keyword:h,literal:(e=>{const t=[];return e.forEach((e=>{t.push(e),e.toLowerCase()===e?t.push(e.toUpperCase()):t.push(e.toLowerCase())})),t})(p),built_in:f},g=e=>e.map((e=>e.replace(/\|\d+$/,""))),v={variants:[{match:[/new/,t.concat(c,"+"),t.concat("(?!",g(f).join("\\b|"),"\\b)"),a],scope:{1:"keyword",4:"title.class"}}]},b=t.concat(r,"\\b(?!\\()"),y={variants:[{match:[t.concat(/::/,t.lookahead(/(?!class\b)/)),b],scope:{2:"variable.constant"}},{match:[/::/,/class/],scope:{2:"variable.language"}},{match:[a,t.concat(/::/,t.lookahead(/(?!class\b)/)),b],scope:{1:"title.class",3:"variable.constant"}},{match:[a,t.concat("::",t.lookahead(/(?!class\b)/))],scope:{1:"title.class"}},{match:[a,/::/,/class/],scope:{1:"title.class",3:"variable.language"}}]},w={scope:"attr",match:t.concat(r,t.lookahead(":"),t.lookahead(/(?!::)/))},O={relevance:0,begin:/\(/,end:/\)/,keywords:m,contains:[w,i,y,e.C_BLOCK_COMMENT_MODE,u,d,v]},k={relevance:0,match:[/\b/,t.concat("(?!fn\\b|function\\b|",g(h).join("\\b|"),"|",g(f).join("\\b|"),"\\b)"),r,t.concat(c,"*"),t.lookahead(/(?=\()/)],scope:{3:"title.function.invoke"},contains:[O]};O.contains.push(k);const x=[w,y,e.C_BLOCK_COMMENT_MODE,u,d,v],E={begin:t.concat(/#\[\s*\\?/,t.either(a,o)),beginScope:"meta",end:/]/,endScope:"meta",keywords:{literal:p,keyword:["new","array"]},contains:[{begin:/\[/,end:/]/,keywords:{literal:p,keyword:["new","array"]},contains:["self",...x]},...x,{scope:"meta",variants:[{match:a},{match:o}]}]};return{case_insensitive:!1,keywords:m,contains:[E,e.HASH_COMMENT_MODE,e.COMMENT("//","$"),e.COMMENT("/\\*","\\*/",{contains:[{scope:"doctag",match:"@[A-Za-z]+"}]}),{match:/__halt_compiler\(\);/,keywords:"__halt_compiler",starts:{scope:"comment",end:e.MATCH_NOTHING_RE,contains:[{match:/\?>/,scope:"meta",endsParent:!0}]}},{scope:"meta",variants:[{begin:/<\?php/,relevance:10},{begin:/<\?=/},{begin:/<\?/,relevance:.1},{begin:/\?>/}]},{scope:"variable.language",match:/\$this\b/},i,k,y,{match:[/const/,/\s/,r],scope:{1:"keyword",3:"variable.constant"}},v,{scope:"function",relevance:0,beginKeywords:"fn function",end:/[;{]/,excludeEnd:!0,illegal:"[$%\\[]",contains:[{beginKeywords:"use"},e.UNDERSCORE_TITLE_MODE,{begin:"=>",endsParent:!0},{scope:"params",begin:"\\(",end:"\\)",excludeBegin:!0,excludeEnd:!0,keywords:m,contains:["self",E,i,y,e.C_BLOCK_COMMENT_MODE,u,d]}]},{scope:"class",variants:[{beginKeywords:"enum",illegal:/[($"]/},{beginKeywords:"class interface trait",illegal:/[:($"]/}],relevance:0,end:/\{/,excludeEnd:!0,contains:[{beginKeywords:"extends implements"},e.UNDERSCORE_TITLE_MODE]},{beginKeywords:"namespace",relevance:0,end:";",illegal:/[.']/,contains:[e.inherit(e.UNDERSCORE_TITLE_MODE,{scope:"title.class"})]},{beginKeywords:"use",relevance:0,end:";",contains:[{match:/\b(as|const|function)\b/,scope:"keyword"},e.UNDERSCORE_TITLE_MODE]},u,d]}},plaintext:function(e){return{name:"Plain text",aliases:["text","txt"],disableAutodetect:!0}},powershell:function(e){const t={$pattern:/-?[A-z\.\-]+\b/,keyword:"if else foreach return do while until elseif begin for trap data dynamicparam end break throw param continue finally in switch exit filter try process catch hidden static parameter",built_in:"ac asnp cat cd CFS chdir clc clear clhy cli clp cls clv cnsn compare copy cp cpi cpp curl cvpa dbp del diff dir dnsn ebp echo|0 epal epcsv epsn erase etsn exsn fc fhx fl ft fw gal gbp gc gcb gci gcm gcs gdr gerr ghy gi gin gjb gl gm gmo gp gps gpv group gsn gsnp gsv gtz gu gv gwmi h history icm iex ihy ii ipal ipcsv ipmo ipsn irm ise iwmi iwr kill lp ls man md measure mi mount move mp mv nal ndr ni nmo npssc nsn nv ogv oh popd ps pushd pwd r rbp rcjb rcsn rd rdr ren ri rjb rm rmdir rmo rni rnp rp rsn rsnp rujb rv rvpa rwmi sajb sal saps sasv sbp sc scb select set shcm si sl sleep sls sort sp spjb spps spsv start stz sujb sv swmi tee trcm type wget where wjb write"},n={begin:"`[\\s\\S]",relevance:0},r={className:"variable",variants:[{begin:/\$\B/},{className:"keyword",begin:/\$this/},{begin:/\$[\w\d][\w\d_:]*/}]},a={className:"string",variants:[{begin:/"/,end:/"/},{begin:/@"/,end:/^"@/}],contains:[n,r,{className:"variable",begin:/\$[A-z]/,end:/[^A-z]/}]},o={className:"string",variants:[{begin:/'/,end:/'/},{begin:/@'/,end:/^'@/}]},i=e.inherit(e.COMMENT(null,null),{variants:[{begin:/#/,end:/$/},{begin:/<#/,end:/#>/}],contains:[{className:"doctag",variants:[{begin:/\.(synopsis|description|example|inputs|outputs|notes|link|component|role|functionality)/},{begin:/\.(parameter|forwardhelptargetname|forwardhelpcategory|remotehelprunspace|externalhelp)\s+\S+/}]}]}),s={className:"built_in",variants:[{begin:"(".concat("Add|Clear|Close|Copy|Enter|Exit|Find|Format|Get|Hide|Join|Lock|Move|New|Open|Optimize|Pop|Push|Redo|Remove|Rename|Reset|Resize|Search|Select|Set|Show|Skip|Split|Step|Switch|Undo|Unlock|Watch|Backup|Checkpoint|Compare|Compress|Convert|ConvertFrom|ConvertTo|Dismount|Edit|Expand|Export|Group|Import|Initialize|Limit|Merge|Mount|Out|Publish|Restore|Save|Sync|Unpublish|Update|Approve|Assert|Build|Complete|Confirm|Deny|Deploy|Disable|Enable|Install|Invoke|Register|Request|Restart|Resume|Start|Stop|Submit|Suspend|Uninstall|Unregister|Wait|Debug|Measure|Ping|Repair|Resolve|Test|Trace|Connect|Disconnect|Read|Receive|Send|Write|Block|Grant|Protect|Revoke|Unblock|Unprotect|Use|ForEach|Sort|Tee|Where",")+(-)[\\w\\d]+")}]},l={className:"class",beginKeywords:"class enum",end:/\s*[{]/,excludeEnd:!0,relevance:0,contains:[e.TITLE_MODE]},c={className:"function",begin:/function\s+/,end:/\s*\{|$/,excludeEnd:!0,returnBegin:!0,relevance:0,contains:[{begin:"function",relevance:0,className:"keyword"},{className:"title",begin:/\w[\w\d]*((-)[\w\d]+)*/,relevance:0},{begin:/\(/,end:/\)/,className:"params",relevance:0,contains:[r]}]},u={begin:/using\s/,end:/$/,returnBegin:!0,contains:[a,o,{className:"keyword",begin:/(using|assembly|command|module|namespace|type)/}]},d={variants:[{className:"operator",begin:"(".concat("-and|-as|-band|-bnot|-bor|-bxor|-casesensitive|-ccontains|-ceq|-cge|-cgt|-cle|-clike|-clt|-cmatch|-cne|-cnotcontains|-cnotlike|-cnotmatch|-contains|-creplace|-csplit|-eq|-exact|-f|-file|-ge|-gt|-icontains|-ieq|-ige|-igt|-ile|-ilike|-ilt|-imatch|-in|-ine|-inotcontains|-inotlike|-inotmatch|-ireplace|-is|-isnot|-isplit|-join|-le|-like|-lt|-match|-ne|-not|-notcontains|-notin|-notlike|-notmatch|-or|-regex|-replace|-shl|-shr|-split|-wildcard|-xor",")\\b")},{className:"literal",begin:/(-){1,2}[\w\d-]+/,relevance:0}]},p={className:"function",begin:/\[.*\]\s*[\w]+[ ]??\(/,end:/$/,returnBegin:!0,relevance:0,contains:[{className:"keyword",begin:"(".concat(t.keyword.toString().replace(/\s/g,"|"),")\\b"),endsParent:!0,relevance:0},e.inherit(e.TITLE_MODE,{endsParent:!0})]},h=[p,i,n,e.NUMBER_MODE,a,o,s,r,{className:"literal",begin:/\$(null|true|false)\b/},{className:"selector-tag",begin:/@\B/,relevance:0}],f={begin:/\[/,end:/\]/,excludeBegin:!0,excludeEnd:!0,relevance:0,contains:[].concat("self",h,{begin:"("+["string","char","byte","int","long","bool","decimal","single","double","DateTime","xml","array","hashtable","void"].join("|")+")",className:"built_in",relevance:0},{className:"type",begin:/[\.\w\d]+/,relevance:0})};return p.contains.unshift(f),{name:"PowerShell",aliases:["pwsh","ps","ps1"],case_insensitive:!0,keywords:t,contains:h.concat(l,c,u,d,f)}},python:function(e){const t=e.regex,n=/[\p{XID_Start}_]\p{XID_Continue}*/u,r=["and","as","assert","async","await","break","case","class","continue","def","del","elif","else","except","finally","for","from","global","if","import","in","is","lambda","match","nonlocal|10","not","or","pass","raise","return","try","while","with","yield"],a={$pattern:/[A-Za-z]\w+|__\w+__/,keyword:r,built_in:["__import__","abs","all","any","ascii","bin","bool","breakpoint","bytearray","bytes","callable","chr","classmethod","compile","complex","delattr","dict","dir","divmod","enumerate","eval","exec","filter","float","format","frozenset","getattr","globals","hasattr","hash","help","hex","id","input","int","isinstance","issubclass","iter","len","list","locals","map","max","memoryview","min","next","object","oct","open","ord","pow","print","property","range","repr","reversed","round","set","setattr","slice","sorted","staticmethod","str","sum","super","tuple","type","vars","zip"],literal:["__debug__","Ellipsis","False","None","NotImplemented","True"],type:["Any","Callable","Coroutine","Dict","List","Literal","Generic","Optional","Sequence","Set","Tuple","Type","Union"]},o={className:"meta",begin:/^(>>>|\.\.\.) /},i={className:"subst",begin:/\{/,end:/\}/,keywords:a,illegal:/#/},s={begin:/\{\{/,relevance:0},l={className:"string",contains:[e.BACKSLASH_ESCAPE],variants:[{begin:/([uU]|[bB]|[rR]|[bB][rR]|[rR][bB])?'''/,end:/'''/,contains:[e.BACKSLASH_ESCAPE,o],relevance:10},{begin:/([uU]|[bB]|[rR]|[bB][rR]|[rR][bB])?"""/,end:/"""/,contains:[e.BACKSLASH_ESCAPE,o],relevance:10},{begin:/([fF][rR]|[rR][fF]|[fF])'''/,end:/'''/,contains:[e.BACKSLASH_ESCAPE,o,s,i]},{begin:/([fF][rR]|[rR][fF]|[fF])"""/,end:/"""/,contains:[e.BACKSLASH_ESCAPE,o,s,i]},{begin:/([uU]|[rR])'/,end:/'/,relevance:10},{begin:/([uU]|[rR])"/,end:/"/,relevance:10},{begin:/([bB]|[bB][rR]|[rR][bB])'/,end:/'/},{begin:/([bB]|[bB][rR]|[rR][bB])"/,end:/"/},{begin:/([fF][rR]|[rR][fF]|[fF])'/,end:/'/,contains:[e.BACKSLASH_ESCAPE,s,i]},{begin:/([fF][rR]|[rR][fF]|[fF])"/,end:/"/,contains:[e.BACKSLASH_ESCAPE,s,i]},e.APOS_STRING_MODE,e.QUOTE_STRING_MODE]},c="[0-9](_?[0-9])*",u=`(\\b(${c}))?\\.(${c})|\\b(${c})\\.`,d=`\\b|${r.join("|")}`,p={className:"number",relevance:0,variants:[{begin:`(\\b(${c})|(${u}))[eE][+-]?(${c})[jJ]?(?=${d})`},{begin:`(${u})[jJ]?`},{begin:`\\b([1-9](_?[0-9])*|0+(_?0)*)[lLjJ]?(?=${d})`},{begin:`\\b0[bB](_?[01])+[lL]?(?=${d})`},{begin:`\\b0[oO](_?[0-7])+[lL]?(?=${d})`},{begin:`\\b0[xX](_?[0-9a-fA-F])+[lL]?(?=${d})`},{begin:`\\b(${c})[jJ](?=${d})`}]},h={className:"comment",begin:t.lookahead(/# type:/),end:/$/,keywords:a,contains:[{begin:/# type:/},{begin:/#/,end:/\b\B/,endsWithParent:!0}]},f={className:"params",variants:[{className:"",begin:/\(\s*\)/,skip:!0},{begin:/\(/,end:/\)/,excludeBegin:!0,excludeEnd:!0,keywords:a,contains:["self",o,p,l,e.HASH_COMMENT_MODE]}]};return i.contains=[l,p,o],{name:"Python",aliases:["py","gyp","ipython"],unicodeRegex:!0,keywords:a,illegal:/(<\/|\?)|=>/,contains:[o,p,{scope:"variable.language",match:/\bself\b/},{beginKeywords:"if",relevance:0},{match:/\bor\b/,scope:"keyword"},l,h,e.HASH_COMMENT_MODE,{match:[/\bdef/,/\s+/,n],scope:{1:"keyword",3:"title.function"},contains:[f]},{variants:[{match:[/\bclass/,/\s+/,n,/\s*/,/\(\s*/,n,/\s*\)/]},{match:[/\bclass/,/\s+/,n]}],scope:{1:"keyword",3:"title.class",6:"title.class.inherited"}},{className:"meta",begin:/^[\t ]*@/,end:/(?=#)|$/,contains:[p,f,l]}]}},r:function(e){const t=e.regex,n=/(?:(?:[a-zA-Z]|\.[._a-zA-Z])[._a-zA-Z0-9]*)|\.(?!\d)/,r=t.either(/0[xX][0-9a-fA-F]+\.[0-9a-fA-F]*[pP][+-]?\d+i?/,/0[xX][0-9a-fA-F]+(?:[pP][+-]?\d+)?[Li]?/,/(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?[Li]?/),a=/[=!<>:]=|\|\||&&|:::?|<-|<<-|->>|->|\|>|[-+*\/?!$&|:<=>@^~]|\*\*/,o=t.either(/[()]/,/[{}]/,/\[\[/,/[[\]]/,/\\/,/,/);return{name:"R",keywords:{$pattern:n,keyword:"function if in break next repeat else for while",literal:"NULL NA TRUE FALSE Inf NaN NA_integer_|10 NA_real_|10 NA_character_|10 NA_complex_|10",built_in:"LETTERS letters month.abb month.name pi T F abs acos acosh all any anyNA Arg as.call as.character as.complex as.double as.environment as.integer as.logical as.null.default as.numeric as.raw asin asinh atan atanh attr attributes baseenv browser c call ceiling class Conj cos cosh cospi cummax cummin cumprod cumsum digamma dim dimnames emptyenv exp expression floor forceAndCall gamma gc.time globalenv Im interactive invisible is.array is.atomic is.call is.character is.complex is.double is.environment is.expression is.finite is.function is.infinite is.integer is.language is.list is.logical is.matrix is.na is.name is.nan is.null is.numeric is.object is.pairlist is.raw is.recursive is.single is.symbol lazyLoadDBfetch length lgamma list log max min missing Mod names nargs nzchar oldClass on.exit pos.to.env proc.time prod quote range Re rep retracemem return round seq_along seq_len seq.int sign signif sin sinh sinpi sqrt standardGeneric substitute sum switch tan tanh tanpi tracemem trigamma trunc unclass untracemem UseMethod xtfrm"},contains:[e.COMMENT(/#'/,/$/,{contains:[{scope:"doctag",match:/@examples/,starts:{end:t.lookahead(t.either(/\n^#'\s*(?=@[a-zA-Z]+)/,/\n^(?!#')/)),endsParent:!0}},{scope:"doctag",begin:"@param",end:/$/,contains:[{scope:"variable",variants:[{match:n},{match:/`(?:\\.|[^`\\])+`/}],endsParent:!0}]},{scope:"doctag",match:/@[a-zA-Z]+/},{scope:"keyword",match:/\\[a-zA-Z]+/}]}),e.HASH_COMMENT_MODE,{scope:"string",contains:[e.BACKSLASH_ESCAPE],variants:[e.END_SAME_AS_BEGIN({begin:/[rR]"(-*)\(/,end:/\)(-*)"/}),e.END_SAME_AS_BEGIN({begin:/[rR]"(-*)\{/,end:/\}(-*)"/}),e.END_SAME_AS_BEGIN({begin:/[rR]"(-*)\[/,end:/\](-*)"/}),e.END_SAME_AS_BEGIN({begin:/[rR]'(-*)\(/,end:/\)(-*)'/}),e.END_SAME_AS_BEGIN({begin:/[rR]'(-*)\{/,end:/\}(-*)'/}),e.END_SAME_AS_BEGIN({begin:/[rR]'(-*)\[/,end:/\](-*)'/}),{begin:'"',end:'"',relevance:0},{begin:"'",end:"'",relevance:0}]},{relevance:0,variants:[{scope:{1:"operator",2:"number"},match:[a,r]},{scope:{1:"operator",2:"number"},match:[/%[^%]*%/,r]},{scope:{1:"punctuation",2:"number"},match:[o,r]},{scope:{2:"number"},match:[/[^a-zA-Z0-9._]|^/,r]}]},{scope:{3:"operator"},match:[n,/\s+/,/<-/,/\s+/]},{scope:"operator",relevance:0,variants:[{match:a},{match:/%[^%]*%/}]},{scope:"punctuation",relevance:0,match:o},{begin:"`",end:"`",contains:[{begin:/\\./}]}]}},ruby:function(e){const t=e.regex,n="([a-zA-Z_]\\w*[!?=]?|[-+~]@|<<|>>|=~|===?|<=>|[<>]=?|\\*\\*|[-/+%^&*~`|]|\\[\\]=?)",r=t.either(/\b([A-Z]+[a-z0-9]+)+/,/\b([A-Z]+[a-z0-9]+)+[A-Z]+/),a=t.concat(r,/(::\w+)*/),o={"variable.constant":["__FILE__","__LINE__","__ENCODING__"],"variable.language":["self","super"],keyword:["alias","and","begin","BEGIN","break","case","class","defined","do","else","elsif","end","END","ensure","for","if","in","module","next","not","or","redo","require","rescue","retry","return","then","undef","unless","until","when","while","yield","include","extend","prepend","public","private","protected","raise","throw"],built_in:["proc","lambda","attr_accessor","attr_reader","attr_writer","define_method","private_constant","module_function"],literal:["true","false","nil"]},i={className:"doctag",begin:"@[A-Za-z]+"},s={begin:"#<",end:">"},l=[e.COMMENT("#","$",{contains:[i]}),e.COMMENT("^=begin","^=end",{contains:[i],relevance:10}),e.COMMENT("^__END__",e.MATCH_NOTHING_RE)],c={className:"subst",begin:/#\{/,end:/\}/,keywords:o},u={className:"string",contains:[e.BACKSLASH_ESCAPE,c],variants:[{begin:/'/,end:/'/},{begin:/"/,end:/"/},{begin:/`/,end:/`/},{begin:/%[qQwWx]?\(/,end:/\)/},{begin:/%[qQwWx]?\[/,end:/\]/},{begin:/%[qQwWx]?\{/,end:/\}/},{begin:/%[qQwWx]?</,end:/>/},{begin:/%[qQwWx]?\//,end:/\//},{begin:/%[qQwWx]?%/,end:/%/},{begin:/%[qQwWx]?-/,end:/-/},{begin:/%[qQwWx]?\|/,end:/\|/},{begin:/\B\?(\\\d{1,3})/},{begin:/\B\?(\\x[A-Fa-f0-9]{1,2})/},{begin:/\B\?(\\u\{?[A-Fa-f0-9]{1,6}\}?)/},{begin:/\B\?(\\M-\\C-|\\M-\\c|\\c\\M-|\\M-|\\C-\\M-)[\x20-\x7e]/},{begin:/\B\?\\(c|C-)[\x20-\x7e]/},{begin:/\B\?\\?\S/},{begin:t.concat(/<<[-~]?'?/,t.lookahead(/(\w+)(?=\W)[^\n]*\n(?:[^\n]*\n)*?\s*\1\b/)),contains:[e.END_SAME_AS_BEGIN({begin:/(\w+)/,end:/(\w+)/,contains:[e.BACKSLASH_ESCAPE,c]})]}]},d="[0-9](_?[0-9])*",p={className:"number",relevance:0,variants:[{begin:`\\b([1-9](_?[0-9])*|0)(\\.(${d}))?([eE][+-]?(${d})|r)?i?\\b`},{begin:"\\b0[dD][0-9](_?[0-9])*r?i?\\b"},{begin:"\\b0[bB][0-1](_?[0-1])*r?i?\\b"},{begin:"\\b0[oO][0-7](_?[0-7])*r?i?\\b"},{begin:"\\b0[xX][0-9a-fA-F](_?[0-9a-fA-F])*r?i?\\b"},{begin:"\\b0(_?[0-7])+r?i?\\b"}]},h={variants:[{match:/\(\)/},{className:"params",begin:/\(/,end:/(?=\))/,excludeBegin:!0,endsParent:!0,keywords:o}]},f=[u,{variants:[{match:[/class\s+/,a,/\s+<\s+/,a]},{match:[/\b(class|module)\s+/,a]}],scope:{2:"title.class",4:"title.class.inherited"},keywords:o},{match:[/(include|extend)\s+/,a],scope:{2:"title.class"},keywords:o},{relevance:0,match:[a,/\.new[. (]/],scope:{1:"title.class"}},{relevance:0,match:/\b[A-Z][A-Z_0-9]+\b/,className:"variable.constant"},{relevance:0,match:r,scope:"title.class"},{match:[/def/,/\s+/,n],scope:{1:"keyword",3:"title.function"},contains:[h]},{begin:e.IDENT_RE+"::"},{className:"symbol",begin:e.UNDERSCORE_IDENT_RE+"(!|\\?)?:",relevance:0},{className:"symbol",begin:":(?!\\s)",contains:[u,{begin:n}],relevance:0},p,{className:"variable",begin:"(\\$\\W)|((\\$|@@?)(\\w+))(?=[^@$?])(?![A-Za-z])(?![@$?'])"},{className:"params",begin:/\|(?!=)/,end:/\|/,excludeBegin:!0,excludeEnd:!0,relevance:0,keywords:o},{begin:"("+e.RE_STARTERS_RE+"|unless)\\s*",keywords:"unless",contains:[{className:"regexp",contains:[e.BACKSLASH_ESCAPE,c],illegal:/\n/,variants:[{begin:"/",end:"/[a-z]*"},{begin:/%r\{/,end:/\}[a-z]*/},{begin:"%r\\(",end:"\\)[a-z]*"},{begin:"%r!",end:"![a-z]*"},{begin:"%r\\[",end:"\\][a-z]*"}]}].concat(s,l),relevance:0}].concat(s,l);c.contains=f,h.contains=f;const m=[{begin:/^\s*=>/,starts:{end:"$",contains:f}},{className:"meta.prompt",begin:"^([>?]>|[\\w#]+\\(\\w+\\):\\d+:\\d+[>*]|(\\w+-)?\\d+\\.\\d+\\.\\d+(p\\d+)?[^\\d][^>]+>)(?=[ ])",starts:{end:"$",keywords:o,contains:f}}];return l.unshift(s),{name:"Ruby",aliases:["rb","gemspec","podspec","thor","irb"],keywords:o,illegal:/\/\*/,contains:[e.SHEBANG({binary:"ruby"})].concat(m).concat(l).concat(f)}},rust:function(e){const t=e.regex,n=/(r#)?/,r=t.concat(n,e.UNDERSCORE_IDENT_RE),a=t.concat(n,e.IDENT_RE),o={className:"title.function.invoke",relevance:0,begin:t.concat(/\b/,/(?!let|for|while|if|else|match\b)/,a,t.lookahead(/\s*\(/))},i="([ui](8|16|32|64|128|size)|f(32|64))?",s=["drop ","Copy","Send","Sized","Sync","Drop","Fn","FnMut","FnOnce","ToOwned","Clone","Debug","PartialEq","PartialOrd","Eq","Ord","AsRef","AsMut","Into","From","Default","Iterator","Extend","IntoIterator","DoubleEndedIterator","ExactSizeIterator","SliceConcatExt","ToString","assert!","assert_eq!","bitflags!","bytes!","cfg!","col!","concat!","concat_idents!","debug_assert!","debug_assert_eq!","env!","eprintln!","panic!","file!","format!","format_args!","include_bytes!","include_str!","line!","local_data_key!","module_path!","option_env!","print!","println!","select!","stringify!","try!","unimplemented!","unreachable!","vec!","write!","writeln!","macro_rules!","assert_ne!","debug_assert_ne!"],l=["i8","i16","i32","i64","i128","isize","u8","u16","u32","u64","u128","usize","f32","f64","str","char","bool","Box","Option","Result","String","Vec"];return{name:"Rust",aliases:["rs"],keywords:{$pattern:e.IDENT_RE+"!?",type:l,keyword:["abstract","as","async","await","become","box","break","const","continue","crate","do","dyn","else","enum","extern","false","final","fn","for","if","impl","in","let","loop","macro","match","mod","move","mut","override","priv","pub","ref","return","self","Self","static","struct","super","trait","true","try","type","typeof","union","unsafe","unsized","use","virtual","where","while","yield"],literal:["true","false","Some","None","Ok","Err"],built_in:s},illegal:"</",contains:[e.C_LINE_COMMENT_MODE,e.COMMENT("/\\*","\\*/",{contains:["self"]}),e.inherit(e.QUOTE_STRING_MODE,{begin:/b?"/,illegal:null}),{className:"symbol",begin:/'[a-zA-Z_][a-zA-Z0-9_]*(?!')/},{scope:"string",variants:[{begin:/b?r(#*)"(.|\n)*?"\1(?!#)/},{begin:/b?'/,end:/'/,contains:[{scope:"char.escape",match:/\\('|\w|x\w{2}|u\w{4}|U\w{8})/}]}]},{className:"number",variants:[{begin:"\\b0b([01_]+)"+i},{begin:"\\b0o([0-7_]+)"+i},{begin:"\\b0x([A-Fa-f0-9_]+)"+i},{begin:"\\b(\\d[\\d_]*(\\.[0-9_]+)?([eE][+-]?[0-9_]+)?)"+i}],relevance:0},{begin:[/fn/,/\s+/,r],className:{1:"keyword",3:"title.function"}},{className:"meta",begin:"#!?\\[",end:"\\]",contains:[{className:"string",begin:/"/,end:/"/,contains:[e.BACKSLASH_ESCAPE]}]},{begin:[/let/,/\s+/,/(?:mut\s+)?/,r],className:{1:"keyword",3:"keyword",4:"variable"}},{begin:[/for/,/\s+/,r,/\s+/,/in/],className:{1:"keyword",3:"variable",5:"keyword"}},{begin:[/type/,/\s+/,r],className:{1:"keyword",3:"title.class"}},{begin:[/(?:trait|enum|struct|union|impl|for)/,/\s+/,r],className:{1:"keyword",3:"title.class"}},{begin:e.IDENT_RE+"::",keywords:{keyword:"Self",built_in:s,type:l}},{className:"punctuation",begin:"->"},o]}},scss:function(e){const t=(e=>({IMPORTANT:{scope:"meta",begin:"!important"},BLOCK_COMMENT:e.C_BLOCK_COMMENT_MODE,HEXCOLOR:{scope:"number",begin:/#(([0-9a-fA-F]{3,4})|(([0-9a-fA-F]{2}){3,4}))\b/},FUNCTION_DISPATCH:{className:"built_in",begin:/[\w-]+(?=\()/},ATTRIBUTE_SELECTOR_MODE:{scope:"selector-attr",begin:/\[/,end:/\]/,illegal:"$",contains:[e.APOS_STRING_MODE,e.QUOTE_STRING_MODE]},CSS_NUMBER_MODE:{scope:"number",begin:e.NUMBER_RE+"(%|em|ex|ch|rem|vw|vh|vmin|vmax|cm|mm|in|pt|pc|px|deg|grad|rad|turn|s|ms|Hz|kHz|dpi|dpcm|dppx)?",relevance:0},CSS_VARIABLE:{className:"attr",begin:/--[A-Za-z_][A-Za-z0-9_-]*/}}))(e),n=$,r=M,a="@[a-z-]+",o={className:"variable",begin:"(\\$[a-zA-Z-][a-zA-Z0-9_-]*)\\b",relevance:0};return{name:"SCSS",case_insensitive:!0,illegal:"[=/|']",contains:[e.C_LINE_COMMENT_MODE,e.C_BLOCK_COMMENT_MODE,t.CSS_NUMBER_MODE,{className:"selector-id",begin:"#[A-Za-z0-9_-]+",relevance:0},{className:"selector-class",begin:"\\.[A-Za-z0-9_-]+",relevance:0},t.ATTRIBUTE_SELECTOR_MODE,{className:"selector-tag",begin:"\\b("+P.join("|")+")\\b",relevance:0},{className:"selector-pseudo",begin:":("+r.join("|")+")"},{className:"selector-pseudo",begin:":(:)?("+n.join("|")+")"},o,{begin:/\(/,end:/\)/,contains:[t.CSS_NUMBER_MODE]},t.CSS_VARIABLE,{className:"attribute",begin:"\\b("+N.join("|")+")\\b"},{begin:"\\b(whitespace|wait|w-resize|visible|vertical-text|vertical-ideographic|uppercase|upper-roman|upper-alpha|underline|transparent|top|thin|thick|text|text-top|text-bottom|tb-rl|table-header-group|table-footer-group|sw-resize|super|strict|static|square|solid|small-caps|separate|se-resize|scroll|s-resize|rtl|row-resize|ridge|right|repeat|repeat-y|repeat-x|relative|progress|pointer|overline|outside|outset|oblique|nowrap|not-allowed|normal|none|nw-resize|no-repeat|no-drop|newspaper|ne-resize|n-resize|move|middle|medium|ltr|lr-tb|lowercase|lower-roman|lower-alpha|loose|list-item|line|line-through|line-edge|lighter|left|keep-all|justify|italic|inter-word|inter-ideograph|inside|inset|inline|inline-block|inherit|inactive|ideograph-space|ideograph-parenthesis|ideograph-numeric|ideograph-alpha|horizontal|hidden|help|hand|groove|fixed|ellipsis|e-resize|double|dotted|distribute|distribute-space|distribute-letter|distribute-all-lines|disc|disabled|default|decimal|dashed|crosshair|collapse|col-resize|circle|char|center|capitalize|break-word|break-all|bottom|both|bolder|bold|block|bidi-override|below|baseline|auto|always|all-scroll|absolute|table|table-cell)\\b"},{begin:/:/,end:/[;}{]/,relevance:0,contains:[t.BLOCK_COMMENT,o,t.HEXCOLOR,t.CSS_NUMBER_MODE,e.QUOTE_STRING_MODE,e.APOS_STRING_MODE,t.IMPORTANT,t.FUNCTION_DISPATCH]},{begin:"@(page|font-face)",keywords:{$pattern:a,keyword:"@page @font-face"}},{begin:"@",end:"[{;]",returnBegin:!0,keywords:{$pattern:/[a-z-]+/,keyword:"and or not only",attribute:D.join(" ")},contains:[{begin:a,className:"keyword"},{begin:/[a-z-]+(?=:)/,className:"attribute"},o,e.QUOTE_STRING_MODE,e.APOS_STRING_MODE,t.HEXCOLOR,t.CSS_NUMBER_MODE]},t.FUNCTION_DISPATCH]}},shell:function(e){return{name:"Shell Session",aliases:["console","shellsession"],contains:[{className:"meta.prompt",begin:/^\s{0,3}[/~\w\d[\]()@-]*[>%$#][ ]?/,starts:{end:/[^\\](?=\s*$)/,subLanguage:"bash"}}]}},sql:function(e){const t=e.regex,n=e.COMMENT("--","$"),r=["abs","acos","array_agg","asin","atan","avg","cast","ceil","ceiling","coalesce","corr","cos","cosh","count","covar_pop","covar_samp","cume_dist","dense_rank","deref","element","exp","extract","first_value","floor","json_array","json_arrayagg","json_exists","json_object","json_objectagg","json_query","json_table","json_table_primitive","json_value","lag","last_value","lead","listagg","ln","log","log10","lower","max","min","mod","nth_value","ntile","nullif","percent_rank","percentile_cont","percentile_disc","position","position_regex","power","rank","regr_avgx","regr_avgy","regr_count","regr_intercept","regr_r2","regr_slope","regr_sxx","regr_sxy","regr_syy","row_number","sin","sinh","sqrt","stddev_pop","stddev_samp","substring","substring_regex","sum","tan","tanh","translate","translate_regex","treat","trim","trim_array","unnest","upper","value_of","var_pop","var_samp","width_bucket"],a=r,o=["abs","acos","all","allocate","alter","and","any","are","array","array_agg","array_max_cardinality","as","asensitive","asin","asymmetric","at","atan","atomic","authorization","avg","begin","begin_frame","begin_partition","between","bigint","binary","blob","boolean","both","by","call","called","cardinality","cascaded","case","cast","ceil","ceiling","char","char_length","character","character_length","check","classifier","clob","close","coalesce","collate","collect","column","commit","condition","connect","constraint","contains","convert","copy","corr","corresponding","cos","cosh","count","covar_pop","covar_samp","create","cross","cube","cume_dist","current","current_catalog","current_date","current_default_transform_group","current_path","current_role","current_row","current_schema","current_time","current_timestamp","current_path","current_role","current_transform_group_for_type","current_user","cursor","cycle","date","day","deallocate","dec","decimal","decfloat","declare","default","define","delete","dense_rank","deref","describe","deterministic","disconnect","distinct","double","drop","dynamic","each","element","else","empty","end","end_frame","end_partition","end-exec","equals","escape","every","except","exec","execute","exists","exp","external","extract","false","fetch","filter","first_value","float","floor","for","foreign","frame_row","free","from","full","function","fusion","get","global","grant","group","grouping","groups","having","hold","hour","identity","in","indicator","initial","inner","inout","insensitive","insert","int","integer","intersect","intersection","interval","into","is","join","json_array","json_arrayagg","json_exists","json_object","json_objectagg","json_query","json_table","json_table_primitive","json_value","lag","language","large","last_value","lateral","lead","leading","left","like","like_regex","listagg","ln","local","localtime","localtimestamp","log","log10","lower","match","match_number","match_recognize","matches","max","member","merge","method","min","minute","mod","modifies","module","month","multiset","national","natural","nchar","nclob","new","no","none","normalize","not","nth_value","ntile","null","nullif","numeric","octet_length","occurrences_regex","of","offset","old","omit","on","one","only","open","or","order","out","outer","over","overlaps","overlay","parameter","partition","pattern","per","percent","percent_rank","percentile_cont","percentile_disc","period","portion","position","position_regex","power","precedes","precision","prepare","primary","procedure","ptf","range","rank","reads","real","recursive","ref","references","referencing","regr_avgx","regr_avgy","regr_count","regr_intercept","regr_r2","regr_slope","regr_sxx","regr_sxy","regr_syy","release","result","return","returns","revoke","right","rollback","rollup","row","row_number","rows","running","savepoint","scope","scroll","search","second","seek","select","sensitive","session_user","set","show","similar","sin","sinh","skip","smallint","some","specific","specifictype","sql","sqlexception","sqlstate","sqlwarning","sqrt","start","static","stddev_pop","stddev_samp","submultiset","subset","substring","substring_regex","succeeds","sum","symmetric","system","system_time","system_user","table","tablesample","tan","tanh","then","time","timestamp","timezone_hour","timezone_minute","to","trailing","translate","translate_regex","translation","treat","trigger","trim","trim_array","true","truncate","uescape","union","unique","unknown","unnest","update","upper","user","using","value","values","value_of","var_pop","var_samp","varbinary","varchar","varying","versioning","when","whenever","where","width_bucket","window","with","within","without","year","add","asc","collation","desc","final","first","last","view"].filter((e=>!r.includes(e))),i={match:t.concat(/\b/,t.either(...a),/\s*\(/),relevance:0,keywords:{built_in:a}};function s(e){return t.concat(/\b/,t.either(...e.map((e=>e.replace(/\s+/,"\\s+")))),/\b/)}const l={scope:"keyword",match:s(["create table","insert into","primary key","foreign key","not null","alter table","add constraint","grouping sets","on overflow","character set","respect nulls","ignore nulls","nulls first","nulls last","depth first","breadth first"]),relevance:0};return{name:"SQL",case_insensitive:!0,illegal:/[{}]|<\//,keywords:{$pattern:/\b[\w\.]+/,keyword:function(e,{exceptions:t,when:n}={}){const r=n;return t=t||[],e.map((e=>e.match(/\|\d+$/)||t.includes(e)?e:r(e)?`${e}|0`:e))}(o,{when:e=>e.length<3}),literal:["true","false","unknown"],type:["bigint","binary","blob","boolean","char","character","clob","date","dec","decfloat","decimal","float","int","integer","interval","nchar","nclob","national","numeric","real","row","smallint","time","timestamp","varchar","varying","varbinary"],built_in:["current_catalog","current_date","current_default_transform_group","current_path","current_role","current_schema","current_transform_group_for_type","current_user","session_user","system_time","system_user","current_time","localtime","current_timestamp","localtimestamp"]},contains:[{scope:"type",match:s(["double precision","large object","with timezone","without timezone"])},l,i,{scope:"variable",match:/@[a-z0-9][a-z0-9_]*/},{scope:"string",variants:[{begin:/'/,end:/'/,contains:[{match:/''/}]}]},{begin:/"/,end:/"/,contains:[{match:/""/}]},e.C_NUMBER_MODE,e.C_BLOCK_COMMENT_MODE,n,{scope:"operator",match:/[-+*/=%^~]|&&?|\|\|?|!=?|<(?:=>?|<|>)?|>[>=]?/,relevance:0}]}},swift:function(e){const t={match:/\s+/,relevance:0},n=e.COMMENT("/\\*","\\*/",{contains:["self"]}),r=[e.C_LINE_COMMENT_MODE,n],a={match:[/\./,F(...U,...z)],className:{2:"keyword"}},o={match:B(/\./,F(...W)),relevance:0},i=W.filter((e=>"string"==typeof e)).concat(["_|0"]),s={variants:[{className:"keyword",match:F(...W.filter((e=>"string"!=typeof e)).concat(j).map(Q),...z)}]},l={$pattern:F(/\b\w+/,/#\w+/),keyword:i.concat(V),literal:X},c=[a,o,s],u=[{match:B(/\./,F(...H)),relevance:0},{className:"built_in",match:B(/\b/,F(...H),/(?=\()/)}],d={match:/->/,relevance:0},p=[d,{className:"operator",relevance:0,variants:[{match:K},{match:`\\.(\\.|${G})+`}]}],h="([0-9]_*)+",f="([0-9a-fA-F]_*)+",m={className:"number",relevance:0,variants:[{match:`\\b(${h})(\\.(${h}))?([eE][+-]?(${h}))?\\b`},{match:`\\b0x(${f})(\\.(${f}))?([pP][+-]?(${h}))?\\b`},{match:/\b0o([0-7]_*)+\b/},{match:/\b0b([01]_*)+\b/}]},g=(e="")=>({className:"subst",variants:[{match:B(/\\/,e,/[0\\tnr"']/)},{match:B(/\\/,e,/u\{[0-9a-fA-F]{1,8}\}/)}]}),v=(e="")=>({className:"subst",match:B(/\\/,e,/[\t ]*(?:[\r\n]|\r\n)/)}),b=(e="")=>({className:"subst",label:"interpol",begin:B(/\\/,e,/\(/),end:/\)/}),y=(e="")=>({begin:B(e,/"""/),end:B(/"""/,e),contains:[g(e),v(e),b(e)]}),w=(e="")=>({begin:B(e,/"/),end:B(/"/,e),contains:[g(e),b(e)]}),O={className:"string",variants:[y(),y("#"),y("##"),y("###"),w(),w("#"),w("##"),w("###")]},k=[e.BACKSLASH_ESCAPE,{begin:/\[/,end:/\]/,relevance:0,contains:[e.BACKSLASH_ESCAPE]}],x={begin:/\/[^\s](?=[^/\n]*\/)/,end:/\//,contains:k},E=e=>{const t=B(e,/\//),n=B(/\//,e);return{begin:t,end:n,contains:[...k,{scope:"comment",begin:`#(?!.*${n})`,end:/$/}]}},S={scope:"regexp",variants:[E("###"),E("##"),E("#"),x]},_={match:B(/`/,ee,/`/)},C=[_,{className:"variable",match:/\$\d+/},{className:"variable",match:`\\$${J}+`}],T=[{match:/(@|#(un)?)available/,scope:"keyword",starts:{contains:[{begin:/\(/,end:/\)/,keywords:re,contains:[...p,m,O]}]}},{scope:"keyword",match:B(/@/,F(...ne),L(F(/\(/,/\s+/)))},{scope:"meta",match:B(/@/,ee)}],A={match:L(/\b[A-Z]/),relevance:0,contains:[{className:"type",match:B(/(AV|CA|CF|CG|CI|CL|CM|CN|CT|MK|MP|MTK|MTL|NS|SCN|SK|UI|WK|XC)/,J,"+")},{className:"type",match:te,relevance:0},{match:/[?!]+/,relevance:0},{match:/\.\.\./,relevance:0},{match:B(/\s+&\s+/,L(te)),relevance:0}]},R={begin:/</,end:/>/,keywords:l,contains:[...r,...c,...T,d,A]};A.contains.push(R);const P={begin:/\(/,end:/\)/,relevance:0,keywords:l,contains:["self",{match:B(ee,/\s*:/),keywords:"_|0",relevance:0},...r,S,...c,...u,...p,m,O,...C,...T,A]},D={begin:/</,end:/>/,keywords:"repeat each",contains:[...r,A]},M={begin:/\(/,end:/\)/,keywords:l,contains:[{begin:F(L(B(ee,/\s*:/)),L(B(ee,/\s+/,ee,/\s*:/))),end:/:/,relevance:0,contains:[{className:"keyword",match:/\b_\b/},{className:"params",match:ee}]},...r,...c,...p,m,O,...T,A,P],endsParent:!0,illegal:/["']/},$={match:[/(func|macro)/,/\s+/,F(_.match,ee,K)],className:{1:"keyword",3:"title.function"},contains:[D,M,t],illegal:[/\[/,/%/]},N={match:[/\b(?:subscript|init[?!]?)/,/\s*(?=[<(])/],className:{1:"keyword"},contains:[D,M,t],illegal:/\[|%/},I={match:[/operator/,/\s+/,K],className:{1:"keyword",3:"title"}},Z={begin:[/precedencegroup/,/\s+/,te],className:{1:"keyword",3:"title"},contains:[A],keywords:[...q,...X],end:/}/},Y={begin:[/(struct|protocol|class|extension|enum|actor)/,/\s+/,ee,/\s*/],beginScope:{1:"keyword",3:"title.class"},keywords:l,contains:[D,...c,{begin:/:/,end:/\{/,keywords:l,contains:[{scope:"title.class.inherited",match:te},...c],relevance:0}]};for(const e of O.variants){const t=e.contains.find((e=>"interpol"===e.label));t.keywords=l;const n=[...c,...u,...p,m,O,...C];t.contains=[...n,{begin:/\(/,end:/\)/,contains:["self",...n]}]}return{name:"Swift",keywords:l,contains:[...r,$,N,{match:[/class\b/,/\s+/,/func\b/,/\s+/,/\b[A-Za-z_][A-Za-z0-9_]*\b/],scope:{1:"keyword",3:"keyword",5:"title.function"}},{match:[/class\b/,/\s+/,/var\b/],scope:{1:"keyword",3:"keyword"}},Y,I,Z,{beginKeywords:"import",end:/$/,contains:[...r],relevance:0},S,...c,...u,...p,m,O,...C,...T,A,P]}},typescript:function(e){const t=e.regex,n=function(e){const t=e.regex,n=ae,r={begin:/<[A-Za-z0-9\\._:-]+/,end:/\/[A-Za-z0-9\\._:-]+>|\/>/,isTrulyOpeningTag:(e,t)=>{const n=e[0].length+e.index,r=e.input[n];if("<"===r||","===r)return void t.ignoreMatch();let a;">"===r&&(((e,{after:t})=>{const n="</"+e[0].slice(1);return-1!==e.input.indexOf(n,t)})(e,{after:n})||t.ignoreMatch());const o=e.input.substring(n);((a=o.match(/^\s*=/))||(a=o.match(/^\s+extends\s+/))&&0===a.index)&&t.ignoreMatch()}},a={$pattern:ae,keyword:oe,literal:ie,built_in:de,"variable.language":ue},o="[0-9](_?[0-9])*",i=`\\.(${o})`,s="0|[1-9](_?[0-9])*|0[0-7]*[89][0-9]*",l={className:"number",variants:[{begin:`(\\b(${s})((${i})|\\.)?|(${i}))[eE][+-]?(${o})\\b`},{begin:`\\b(${s})\\b((${i})\\b|\\.)?|(${i})\\b`},{begin:"\\b(0|[1-9](_?[0-9])*)n\\b"},{begin:"\\b0[xX][0-9a-fA-F](_?[0-9a-fA-F])*n?\\b"},{begin:"\\b0[bB][0-1](_?[0-1])*n?\\b"},{begin:"\\b0[oO][0-7](_?[0-7])*n?\\b"},{begin:"\\b0[0-7]+n?\\b"}],relevance:0},c={className:"subst",begin:"\\$\\{",end:"\\}",keywords:a,contains:[]},u={begin:".?html`",end:"",starts:{end:"`",returnEnd:!1,contains:[e.BACKSLASH_ESCAPE,c],subLanguage:"xml"}},d={begin:".?css`",end:"",starts:{end:"`",returnEnd:!1,contains:[e.BACKSLASH_ESCAPE,c],subLanguage:"css"}},p={begin:".?gql`",end:"",starts:{end:"`",returnEnd:!1,contains:[e.BACKSLASH_ESCAPE,c],subLanguage:"graphql"}},h={className:"string",begin:"`",end:"`",contains:[e.BACKSLASH_ESCAPE,c]},f={className:"comment",variants:[e.COMMENT(/\/\*\*(?!\/)/,"\\*/",{relevance:0,contains:[{begin:"(?=@[A-Za-z]+)",relevance:0,contains:[{className:"doctag",begin:"@[A-Za-z]+"},{className:"type",begin:"\\{",end:"\\}",excludeEnd:!0,excludeBegin:!0,relevance:0},{className:"variable",begin:n+"(?=\\s*(-)|$)",endsParent:!0,relevance:0},{begin:/(?=[^\n])\s/,relevance:0}]}]}),e.C_BLOCK_COMMENT_MODE,e.C_LINE_COMMENT_MODE]},m=[e.APOS_STRING_MODE,e.QUOTE_STRING_MODE,u,d,p,h,{match:/\$\d+/},l];c.contains=m.concat({begin:/\{/,end:/\}/,keywords:a,contains:["self"].concat(m)});const g=[].concat(f,c.contains),v=g.concat([{begin:/(\s*)\(/,end:/\)/,keywords:a,contains:["self"].concat(g)}]),b={className:"params",begin:/(\s*)\(/,end:/\)/,excludeBegin:!0,excludeEnd:!0,keywords:a,contains:v},y={variants:[{match:[/class/,/\s+/,n,/\s+/,/extends/,/\s+/,t.concat(n,"(",t.concat(/\./,n),")*")],scope:{1:"keyword",3:"title.class",5:"keyword",7:"title.class.inherited"}},{match:[/class/,/\s+/,n],scope:{1:"keyword",3:"title.class"}}]},w={relevance:0,match:t.either(/\bJSON/,/\b[A-Z][a-z]+([A-Z][a-z]*|\d)*/,/\b[A-Z]{2,}([A-Z][a-z]+|\d)+([A-Z][a-z]*)*/,/\b[A-Z]{2,}[a-z]+([A-Z][a-z]+|\d)*([A-Z][a-z]*)*/),className:"title.class",keywords:{_:[...se,...le]}},O={variants:[{match:[/function/,/\s+/,n,/(?=\s*\()/]},{match:[/function/,/\s*(?=\()/]}],className:{1:"keyword",3:"title.function"},label:"func.def",contains:[b],illegal:/%/},k={match:t.concat(/\b/,(x=[...ce,"super","import"].map((e=>`${e}\\s*\\(`)),t.concat("(?!",x.join("|"),")")),n,t.lookahead(/\s*\(/)),className:"title.function",relevance:0};var x;const E={begin:t.concat(/\./,t.lookahead(t.concat(n,/(?![0-9A-Za-z$_(])/))),end:n,excludeBegin:!0,keywords:"prototype",className:"property",relevance:0},S={match:[/get|set/,/\s+/,n,/(?=\()/],className:{1:"keyword",3:"title.function"},contains:[{begin:/\(\)/},b]},_="(\\([^()]*(\\([^()]*(\\([^()]*\\)[^()]*)*\\)[^()]*)*\\)|"+e.UNDERSCORE_IDENT_RE+")\\s*=>",C={match:[/const|var|let/,/\s+/,n,/\s*/,/=\s*/,/(async\s*)?/,t.lookahead(_)],keywords:"async",className:{1:"keyword",3:"title.function"},contains:[b]};return{name:"JavaScript",aliases:["js","jsx","mjs","cjs"],keywords:a,exports:{PARAMS_CONTAINS:v,CLASS_REFERENCE:w},illegal:/#(?![$_A-z])/,contains:[e.SHEBANG({label:"shebang",binary:"node",relevance:5}),{label:"use_strict",className:"meta",relevance:10,begin:/^\s*['"]use (strict|asm)['"]/},e.APOS_STRING_MODE,e.QUOTE_STRING_MODE,u,d,p,h,f,{match:/\$\d+/},l,w,{scope:"attr",match:n+t.lookahead(":"),relevance:0},C,{begin:"("+e.RE_STARTERS_RE+"|\\b(case|return|throw)\\b)\\s*",keywords:"return throw case",relevance:0,contains:[f,e.REGEXP_MODE,{className:"function",begin:_,returnBegin:!0,end:"\\s*=>",contains:[{className:"params",variants:[{begin:e.UNDERSCORE_IDENT_RE,relevance:0},{className:null,begin:/\(\s*\)/,skip:!0},{begin:/(\s*)\(/,end:/\)/,excludeBegin:!0,excludeEnd:!0,keywords:a,contains:v}]}]},{begin:/,/,relevance:0},{match:/\s+/,relevance:0},{variants:[{begin:"<>",end:"</>"},{match:/<[A-Za-z0-9\\._:-]+\s*\/>/},{begin:r.begin,"on:begin":r.isTrulyOpeningTag,end:r.end}],subLanguage:"xml",contains:[{begin:r.begin,end:r.end,skip:!0,contains:["self"]}]}]},O,{beginKeywords:"while if switch catch for"},{begin:"\\b(?!function)"+e.UNDERSCORE_IDENT_RE+"\\([^()]*(\\([^()]*(\\([^()]*\\)[^()]*)*\\)[^()]*)*\\)\\s*\\{",returnBegin:!0,label:"func.def",contains:[b,e.inherit(e.TITLE_MODE,{begin:n,className:"title.function"})]},{match:/\.\.\./,relevance:0},E,{match:"\\$"+n,relevance:0},{match:[/\bconstructor(?=\s*\()/],className:{1:"title.function"},contains:[b]},k,{relevance:0,match:/\b[A-Z][A-Z_0-9]+\b/,className:"variable.constant"},y,S,{match:/\$[(.]/}]}}(e),r=ae,a=["any","void","number","boolean","string","object","never","symbol","bigint","unknown"],o={begin:[/namespace/,/\s+/,e.IDENT_RE],beginScope:{1:"keyword",3:"title.class"}},i={beginKeywords:"interface",end:/\{/,excludeEnd:!0,keywords:{keyword:"interface extends",built_in:a},contains:[n.exports.CLASS_REFERENCE]},s={$pattern:ae,keyword:oe.concat(["type","interface","public","private","protected","implements","declare","abstract","readonly","enum","override","satisfies"]),literal:ie,built_in:de.concat(a),"variable.language":ue},l={className:"meta",begin:"@"+r},c=(e,t,n)=>{const r=e.contains.findIndex((e=>e.label===t));if(-1===r)throw new Error("can not find mode to replace");e.contains.splice(r,1,n)};Object.assign(n.keywords,s),n.exports.PARAMS_CONTAINS.push(l);const u=n.contains.find((e=>"attr"===e.scope)),d=Object.assign({},u,{match:t.concat(r,t.lookahead(/\s*\?:/))});return n.exports.PARAMS_CONTAINS.push([n.exports.CLASS_REFERENCE,u,d]),n.contains=n.contains.concat([l,o,i,d]),c(n,"shebang",e.SHEBANG()),c(n,"use_strict",{className:"meta",relevance:10,begin:/^\s*['"]use strict['"]/}),n.contains.find((e=>"func.def"===e.label)).relevance=0,Object.assign(n,{name:"TypeScript",aliases:["ts","tsx","mts","cts"]}),n},xml:pe,yaml:function(e){const t="true false yes no null",n="[\\w#;/?:@&=+$,.~*'()[\\]]+",r={className:"string",relevance:0,variants:[{begin:/"/,end:/"/},{begin:/\S+/}],contains:[e.BACKSLASH_ESCAPE,{className:"template-variable",variants:[{begin:/\{\{/,end:/\}\}/},{begin:/%\{/,end:/\}/}]}]},a=e.inherit(r,{variants:[{begin:/'/,end:/'/,contains:[{begin:/''/,relevance:0}]},{begin:/"/,end:/"/},{begin:/[^\s,{}[\]]+/}]}),o={end:",",endsWithParent:!0,excludeEnd:!0,keywords:t,relevance:0},i={begin:/\{/,end:/\}/,contains:[o],illegal:"\\n",relevance:0},s={begin:"\\[",end:"\\]",contains:[o],illegal:"\\n",relevance:0},l=[{className:"attr",variants:[{begin:/[\w*@][\w*@ :()\./-]*:(?=[ \t]|$)/},{begin:/"[\w*@][\w*@ :()\./-]*":(?=[ \t]|$)/},{begin:/'[\w*@][\w*@ :()\./-]*':(?=[ \t]|$)/}]},{className:"meta",begin:"^---\\s*$",relevance:10},{className:"string",begin:"[\\|>]([1-9]?[+-])?[ ]*\\n( +)[^ ][^\\n]*\\n(\\2[^\\n]+\\n?)*"},{begin:"<%[%=-]?",end:"[%-]?%>",subLanguage:"ruby",excludeBegin:!0,excludeEnd:!0,relevance:0},{className:"type",begin:"!\\w+!"+n},{className:"type",begin:"!<"+n+">"},{className:"type",begin:"!"+n},{className:"type",begin:"!!"+n},{className:"meta",begin:"&"+e.UNDERSCORE_IDENT_RE+"$"},{className:"meta",begin:"\\*"+e.UNDERSCORE_IDENT_RE+"$"},{className:"bullet",begin:"-(?=[ ]|$)",relevance:0},e.HASH_COMMENT_MODE,{beginKeywords:t,keywords:{literal:t}},{className:"number",begin:"\\b[0-9]{4}(-[0-9][0-9]){0,2}([Tt \\t][0-9][0-9]?(:[0-9][0-9]){2})?(\\.[0-9]*)?([ \\t])*(Z|[-+][0-9][0-9]?(:[0-9][0-9])?)?\\b"},{className:"number",begin:e.C_NUMBER_RE+"\\b",relevance:0},i,s,{className:"string",relevance:0,begin:/'/,end:/'/,contains:[{match:/''/,scope:"char.escape",relevance:0}]},r],c=[...l];return c.pop(),c.push(a),o.contains=c,{name:"YAML",case_insensitive:!0,aliases:["yml"],contains:l}}}},7786:(e,t,n)=>{"use strict";n.d(t,{E:()=>Ie});class r{constructor({indent:e,join:t}={}){this.postProcessors=[],this.code=[],this.indentationCharacter="",this.lineJoin="\n",this.indentLine=(e,t=0)=>`${this.indentationCharacter.repeat(t)}${e}`,this.unshift=(e,t)=>{const n=this.indentLine(e,t);this.code.unshift(n)},this.push=(e,t)=>{const n=this.indentLine(e,t);this.code.push(n)},this.blank=()=>{this.code.push("")},this.join=()=>{const e=this.code.join(this.lineJoin);return this.postProcessors.reduce(((e,t)=>t(e)),e)},this.addPostProcessor=e=>{this.postProcessors=[...this.postProcessors,e]},this.indentationCharacter=e||"",this.lineJoin=null!=t?t:"\n"}}function a(e,t={}){const{delimiter:n='"',escapeChar:r="\\",escapeNewlines:a=!0}=t;return[...e.toString()].map((e=>"\b"===e?`${r}b`:"\t"===e?`${r}t`:"\n"===e?a?`${r}n`:e:"\f"===e?`${r}f`:"\r"===e?a?`${r}r`:e:e===r?r+r:e===n?r+n:e<" "||e>"~"?JSON.stringify(e).slice(1,-1):e)).join("")}const o=e=>a(e,{delimiter:"'"}),i=e=>a(e,{delimiter:'"'}),s={info:{key:"libcurl",title:"Libcurl",link:"http://curl.haxx.se/libcurl",description:"Simple REST and HTTP API Client for C"},convert:({method:e,fullUrl:t,headersObj:n,allHeaders:a,postData:o})=>{const{push:s,blank:l,join:c}=new r;s("CURL *hnd = curl_easy_init();"),l(),s(`curl_easy_setopt(hnd, CURLOPT_CUSTOMREQUEST, "${e.toUpperCase()}");`),s(`curl_easy_setopt(hnd, CURLOPT_URL, "${t}");`);const u=Object.keys(n);return u.length&&(l(),s("struct curl_slist *headers = NULL;"),u.forEach((e=>{s(`headers = curl_slist_append(headers, "${e}: ${i(n[e])}");`)})),s("curl_easy_setopt(hnd, CURLOPT_HTTPHEADER, headers);")),a.cookie&&(l(),s(`curl_easy_setopt(hnd, CURLOPT_COOKIE, "${a.cookie}");`)),(null==o?void 0:o.text)&&(l(),s(`curl_easy_setopt(hnd, CURLOPT_POSTFIELDS, ${JSON.stringify(o.text)});`)),l(),s("CURLcode ret = curl_easy_perform(hnd);"),c()}};function l(e,t){const n=new URL(t?.url??""),r="/"===n.pathname?n.origin:n.toString(),a={method:t?.method??"GET",url:r,httpVersion:"HTTP/1.1",cookies:[],headers:t?.headers??[],headersSize:-1,bodySize:-1,queryString:Array.from(n.searchParams.entries()).map((([e,t])=>({name:e,value:t}))),postData:t?.postData},o=(a?.headers??[]).reduce(((e,t)=>({...e,[t.name]:t.value})),{}),i=(a.queryString??[]).reduce(((e,t)=>({...e,[t.name]:t.value})),{}),s=(a.cookies??[]).reduce(((e,t)=>({...e,[t.name]:t.value})),{}),l=new URL(a.url),c={protocol:l.protocol,hostname:l.hostname,host:l.hostname,port:l.port,pathname:l.pathname.split("/").map((e=>encodeURIComponent(decodeURIComponent(e)))).join("/")+l.search,path:l.pathname.split("/").map((e=>encodeURIComponent(decodeURIComponent(e)))).join("/")+l.search,search:l.search,hash:l.hash,href:l.href,origin:l.origin,password:l.password,searchParams:l.searchParams,username:l.username,toString:l.toString,toJSON:()=>l.toJSON()},u={url:a.url,uriObj:c,method:a.method?.toLocaleUpperCase()??"GET",httpVersion:a.httpVersion,cookies:a.cookies??[],headers:a.headers??[],headersSize:a.headersSize,headersObj:o??{},bodySize:a.bodySize,queryString:a.queryString??[],postData:a.postData?{mimeType:a.postData.mimeType??"application/json",text:a.postData.text??"",params:a.postData.params??[],paramsObj:a.postData.params?.reduce(((e,t)=>(t.name&&void 0!==t.value&&(e[t.name]=t.value),e)),{})??{}}:void 0,allHeaders:o??{},fullUrl:a.url,queryObj:i??{},cookiesObj:s??{}};return"function"==typeof e.convert?e.convert(u):""}const c={target:"c",client:"libcurl",title:"Libcurl",generate:e=>l(s,e)},u={info:{key:"nethttp",title:"java.net.http",link:"https://openjdk.java.net/groups/net/httpclient/intro.html",description:"Java Standardized HTTP Client API"},convert:({allHeaders:e,fullUrl:t,method:n,postData:a},o)=>{const s={indent:"  ",...o},{push:l,join:c}=new r({indent:s.indent});return l("HttpRequest request = HttpRequest.newBuilder()"),l(`.uri(URI.create("${t}"))`,2),Object.keys(e).forEach((t=>{l(`.header("${t}", "${i(e[t])}")`,2)})),(null==a?void 0:a.text)?l(`.method("${n.toUpperCase()}", HttpRequest.BodyPublishers.ofString(${JSON.stringify(a.text)}))`,2):l(`.method("${n.toUpperCase()}", HttpRequest.BodyPublishers.noBody())`,2),l(".build();",2),l("HttpResponse<String> response = HttpClient.newHttpClient().send(request, HttpResponse.BodyHandlers.ofString());"),l("System.out.println(response.body());"),c()}},d={target:"java",client:"nethttp",title:"java.net.http",generate:e=>l(u,e)},p={info:{key:"okhttp",title:"OkHttp",link:"http://square.github.io/okhttp/",description:"An HTTP Request Client Library"},convert:({postData:e,method:t,fullUrl:n,allHeaders:a},o)=>{const s={indent:"  ",...o},{push:l,blank:c,join:u}=new r({indent:s.indent});return l("OkHttpClient client = new OkHttpClient();"),c(),(null==e?void 0:e.text)&&(e.boundary?l(`MediaType mediaType = MediaType.parse("${e.mimeType}; boundary=${e.boundary}");`):l(`MediaType mediaType = MediaType.parse("${e.mimeType}");`),l(`RequestBody body = RequestBody.create(mediaType, ${JSON.stringify(e.text)});`)),l("Request request = new Request.Builder()"),l(`.url("${n}")`,1),["GET","POST","PUT","DELETE","PATCH","HEAD"].includes(t.toUpperCase())?["POST","PUT","DELETE","PATCH"].includes(t.toUpperCase())?(null==e?void 0:e.text)?l(`.${t.toLowerCase()}(body)`,1):l(`.${t.toLowerCase()}(null)`,1):l(`.${t.toLowerCase()}()`,1):(null==e?void 0:e.text)?l(`.method("${t.toUpperCase()}", body)`,1):l(`.method("${t.toUpperCase()}", null)`,1),Object.keys(a).forEach((e=>{l(`.addHeader("${e}", "${i(a[e])}")`,1)})),l(".build();",1),c(),l("Response response = client.newCall(request).execute();"),u()}},h={target:"java",client:"okhttp",title:"OkHttp",generate:e=>l(p,e)},f={info:{key:"unirest",title:"Unirest",link:"http://unirest.io/java.html",description:"Lightweight HTTP Request Client Library"},convert:({method:e,allHeaders:t,postData:n,fullUrl:a},o)=>{const s={indent:"  ",...o},{join:l,push:c}=new r({indent:s.indent});return["GET","POST","PUT","DELETE","PATCH","HEAD","OPTIONS"].includes(e.toUpperCase())?c(`HttpResponse<String> response = Unirest.${e.toLowerCase()}("${a}")`):c(`HttpResponse<String> response = Unirest.customMethod("${e.toUpperCase()}","${a}")`),Object.keys(t).forEach((e=>{c(`.header("${e}", "${i(t[e])}")`,1)})),(null==n?void 0:n.text)&&c(`.body(${JSON.stringify(n.text)})`,1),c(".asString();",1),l()}},m={target:"java",client:"unirest",title:"Unirest",generate:e=>l(f,e)},g=(e,t)=>Object.keys(e).find((e=>e.toLowerCase()===t.toLowerCase())),v=(e,t)=>{const n=g(e,t);if(n)return e[n]},b=(e,t)=>Boolean(g(e,t)),y={info:{key:"httpclient",title:"HttpClient",link:"https://docs.microsoft.com/en-us/dotnet/api/system.net.http.httpclient",description:".NET Standard HTTP Client"},convert:({allHeaders:e,postData:t,method:n,fullUrl:a},o)=>{let s,l;const c={indent:"    ",...o},{push:u,join:d}=new r({indent:c.indent});u("using System.Net.Http.Headers;");let p="";const h=Boolean(e.cookie),f=(e=>{let t=v(e,"accept-encoding");if(!t)return[];const n={gzip:"DecompressionMethods.GZip",deflate:"DecompressionMethods.Deflate"},r=[];return"string"==typeof t&&(t=[t]),t.forEach((e=>{e.split(",").forEach((e=>{const t=/\s*([^;\s]+)/.exec(e);if(t){const e=n[t[1]];e&&r.push(e)}}))})),r})(e);(h||f.length)&&(p="clientHandler",u("var clientHandler = new HttpClientHandler"),u("{"),h&&u("UseCookies = false,",1),f.length&&u(`AutomaticDecompression = ${f.join(" | ")},`,1),u("};")),u(`var client = new HttpClient(${p});`),u("var request = new HttpRequestMessage"),u("{"),n=(n=n.toUpperCase())&&["GET","POST","PUT","DELETE","PATCH","HEAD","OPTIONS","TRACE"].includes(n)?`HttpMethod.${n[0]}${n.substring(1).toLowerCase()}`:`new HttpMethod("${n}")`,u(`Method = ${n},`,1),u(`RequestUri = new Uri("${a}"),`,1);const m=Object.keys(e).filter((e=>{switch(e.toLowerCase()){case"content-type":case"content-length":case"accept-encoding":return!1;default:return!0}}));if(m.length&&(u("Headers =",1),u("{",1),m.forEach((t=>{u(`{ "${t}", "${i(e[t])}" },`,2)})),u("},",1)),null==t?void 0:t.text){const e=t.mimeType;switch(e){case"application/x-www-form-urlencoded":u("Content = new FormUrlEncodedContent(new Dictionary<string, string>",1),u("{",1),null===(s=t.params)||void 0===s||s.forEach((e=>{u(`{ "${e.name}", "${e.value}" },`,2)})),u("}),",1);break;case"multipart/form-data":u("Content = new MultipartFormDataContent",1),u("{",1),null===(l=t.params)||void 0===l||l.forEach((e=>{u(`new StringContent(${JSON.stringify(e.value||"")})`,2),u("{",2),u("Headers =",3),u("{",3),e.contentType&&u(`ContentType = new MediaTypeHeaderValue("${e.contentType}"),`,4),u('ContentDisposition = new ContentDispositionHeaderValue("form-data")',4),u("{",4),u(`Name = "${e.name}",`,5),e.fileName&&u(`FileName = "${e.fileName}",`,5),u("}",4),u("}",3),u("},",2)})),u("},",1);break;default:u(`Content = new StringContent(${JSON.stringify((null==t?void 0:t.text)||"")})`,1),u("{",1),u("Headers =",2),u("{",2),u(`ContentType = new MediaTypeHeaderValue("${e}")`,3),u("}",2),u("}",1)}}return u("};"),u("using (var response = await client.SendAsync(request))"),u("{"),u("response.EnsureSuccessStatusCode();",1),u("var body = await response.Content.ReadAsStringAsync();",1),u("Console.WriteLine(body);",1),u("}"),d()}},w={target:"csharp",client:"httpclient",title:"HttpClient",generate:e=>l(y,e)},O={info:{key:"restsharp",title:"RestSharp",link:"http://restsharp.org/",description:"Simple REST and HTTP API Client for .NET"},convert:({allHeaders:e,method:t,fullUrl:n,headersObj:a,cookies:o,postData:s})=>{const{push:l,join:c}=new r;if(!["GET","POST","PUT","DELETE","PATCH","HEAD","OPTIONS"].includes(t.toUpperCase()))return"Method not supported";if(l(`var client = new RestClient("${n}");`),l(`var request = new RestRequest(Method.${t.toUpperCase()});`),Object.keys(a).forEach((e=>{l(`request.AddHeader("${e}", "${i(a[e])}");`)})),null==o||o.forEach((({name:e,value:t})=>{l(`request.AddCookie("${e}", "${t}");`)})),null==s?void 0:s.text){const t=v(e,"content-type"),n=JSON.stringify(s.text);l(`request.AddParameter("${t}", ${n}, ParameterType.RequestBody);`)}return l("IRestResponse response = client.Execute(request);"),c()}},k={target:"csharp",client:"restsharp",title:"RestSharp",generate:e=>l(O,e)};class x{constructor(e){this.name="",this.toString=()=>`:${this.name}`,this.name=e}}class E{constructor(e){this.path="",this.toString=()=>`(clojure.java.io/file "${this.path}")`,this.path=e}}const S=e=>void 0===e?null:null===e?"null":e.constructor.name.toLowerCase(),_=e=>"object"===S(e)&&0===Object.keys(e).length,C=e=>(Object.keys(e).filter((t=>_(e[t]))).forEach((t=>{delete e[t]})),e),T=(e,t)=>{const n=" ".repeat(e);return t.replace(/\n/g,`\n${n}`)},A=e=>{switch(S(e)){case"string":return`"${e.replace(/"/g,'\\"')}"`;case"file":case"keyword":default:return e.toString();case"null":return"nil";case"regexp":return`#"${e.source}"`;case"object":{const t=Object.keys(e).reduce(((t,n)=>`${t}:${n} ${T(n.length+2,A(e[n]))}\n `),"").trim();return`{${T(1,t)}}`}case"array":{const t=e.reduce(((e,t)=>`${e} ${A(t)}`),"").trim();return`[${T(1,t)}]`}}},R={info:{key:"clj_http",title:"clj-http",link:"https://github.com/dakrone/clj-http",description:"An idiomatic clojure http client wrapping the apache client."},convert:({queryObj:e,method:t,postData:n,url:a,allHeaders:o},i)=>{const{push:s,join:l}=new r({indent:null==i?void 0:i.indent});if(t=t.toLowerCase(),!["get","post","put","delete","patch","head","options"].includes(t))return s("Method not supported"),l();const c={headers:o,"query-params":e};switch(e&&Object.keys(e).length>0&&(a=a.split("?")[0]),null==n?void 0:n.mimeType){case"application/json":{c["content-type"]=new x("json"),c["form-params"]=n.jsonObj;const e=g(c.headers,"content-type");e&&delete c.headers[e]}break;case"application/x-www-form-urlencoded":{c["form-params"]=n.paramsObj;const e=g(c.headers,"content-type");e&&delete c.headers[e]}break;case"text/plain":{c.body=n.text;const e=g(c.headers,"content-type");e&&delete c.headers[e]}break;case"multipart/form-data":if(n.params){c.multipart=n.params.map((e=>e.fileName&&!e.value?{name:e.name,content:new E(e.fileName)}:{name:e.name,content:e.value}));const e=g(c.headers,"content-type");e&&delete c.headers[e]}}if("application/json"===v(c.headers,"accept")){c.accept=new x("json");const e=g(c.headers,"accept");e&&delete c.headers[e]}if(s("(require '[clj-http.client :as client])\n"),_(C(c)))s(`(client/${t} "${a}")`);else{const e=11+t.length+a.length;s(`(client/${t} "${a}" ${T(e,A(C(c)))})`)}return l()}},P={target:"clojure",client:"clj_http",title:"clj-http",generate:e=>l(R,e)},D={target:"dart",client:"http",title:"Http",generate(e,t){const n={method:"GET",...e};n.method=n.method.toUpperCase();let r="import 'package:http/http.dart' as http;\n\nvoid main() async {\n",a="",o="";n.cookies&&n.cookies.length>0&&(o=n.cookies.map((e=>`${encodeURIComponent(e.name)}=${encodeURIComponent(e.value)}`)).join("; "),a=`  "Cookie": "${o}",\n`);const i=n.headers?.reduce(((e,t)=>(t.value&&!/[; ]/.test(t.name)&&(e[t.name]=t.value),e)),{})||{};if(t?.auth){const{username:e,password:n}=t.auth;if(e&&n){const t=`${e}:${n}`;i.Authorization=`'Basic ' + base64Encode(utf8.encode('${t}'))`}}if(a&&(i.Cookie=o),Object.keys(i).length>0){r+="  final headers = <String,String>{\n";for(const[e,t]of Object.entries(i))t.includes("utf8.encode")?r+=`    '${e}': ${t},\n`:r+=`    '${e}': '${t}',\n`;r+="  };\n\n"}const s=n.queryString?.length?"?"+n.queryString.map((e=>`${encodeURIComponent(e.name)}=${encodeURIComponent(e.value)}`)).join("&"):"",l=`${n.url}${s}`;let c="";if(n.postData)if("application/json"===n.postData.mimeType)c=`  final body = r'${n.postData.text}';\n\n`;else if("application/x-www-form-urlencoded"===n.postData.mimeType)c=`  final body = '${n.postData.params?.map((e=>`${encodeURIComponent(e.name)}=${encodeURIComponent(e.value??"")}`)).join("&")||""}';\n\n`;else if("multipart/form-data"===n.postData.mimeType){c="  final body = <String,String>{\n";for(const e of n.postData.params||[]){const t=e.value||"",n=e.fileName||"";c+=`    '${e.name}': '${n||t}',\n`}c+="  };\n\n"}else"application/octet-stream"===n.postData.mimeType&&(c=`  final body = '${n.postData.text}';\n\n`);return c&&(r+=c),r+=`  final response = await http.${n.method.toLowerCase()}(Uri.parse('${l}')${Object.keys(i).length>0?", headers: headers":""}${c?", body: body":""});\n`,r+="  print(response.body);\n",r+="}",r}},M={info:{key:"native",title:"NewRequest",link:"http://golang.org/pkg/net/http/#NewRequest",description:"Golang HTTP client request"},convert:({postData:e,method:t,allHeaders:n,fullUrl:a},o={})=>{const{blank:s,push:l,join:c}=new r({indent:"\t"}),{showBoilerplate:u=!0,checkErrors:d=!1,printBody:p=!0,timeout:h=-1,insecureSkipVerify:f=!1}=o,m=d?"err":"_",g=u?1:0,v=()=>{d&&(l("if err != nil {",g),l("panic(err)",g+1),l("}",g))};u&&(l("package main"),s(),l("import ("),l('"fmt"',g),h>0&&l('"time"',g),f&&l('"crypto/tls"',g),(null==e?void 0:e.text)&&l('"strings"',g),l('"net/http"',g),p&&l('"io"',g),l(")"),s(),l("func main() {"),s()),f&&(l("insecureTransport := http.DefaultTransport.(*http.Transport).Clone()",g),l("insecureTransport.TLSClientConfig = &tls.Config{InsecureSkipVerify: true}",g));const b=h>0,y=b||f,w=y?"client":"http.DefaultClient";return y&&(l("client := http.Client{",g),b&&l(`Timeout: time.Duration(${h} * time.Second),`,g+1),f&&l("Transport: insecureTransport,",g+1),l("}",g),s()),l(`url := "${a}"`,g),s(),(null==e?void 0:e.text)?(l(`payload := strings.NewReader(${JSON.stringify(e.text)})`,g),s(),l(`req, ${m} := http.NewRequest("${t}", url, payload)`,g),s()):(l(`req, ${m} := http.NewRequest("${t}", url, nil)`,g),s()),v(),Object.keys(n).length&&(Object.keys(n).forEach((e=>{l(`req.Header.Add("${e}", "${i(n[e])}")`,g)})),s()),l(`res, ${m} := ${w}.Do(req)`,g),v(),p&&(s(),l("defer res.Body.Close()",g),l(`body, ${m} := io.ReadAll(res.Body)`,g),v()),s(),l("fmt.Println(res)",g),p&&l("fmt.Println(string(body))",g),u&&(s(),l("}")),c()}},$={target:"go",client:"native",title:"NewRequest",generate:e=>l(M,e)},N={target:"http",client:"http1.1",title:"HTTP/1.1",generate(e){const t={method:"GET",headers:[],queryString:[],...e};let n,r;t.method=t.method.toUpperCase();try{n=new URL(t.url||"http://"),r=n.pathname+(n.search||"")}catch(e){r=t.url||"/"}const a=n?.hostname||"UNKNOWN_HOSTNAME";let o=`${t.method} ${r} HTTP/1.1\r\n`;if(t.queryString.length){const e=t.queryString.map((e=>`${encodeURIComponent(e.name)}=${encodeURIComponent(e.value)}`)).join("&");o=`${t.method} ${r}?${e} HTTP/1.1\r\n`}const i=new Map;if(i.set("Host",a),t.headers.forEach((e=>{i.has(e.name)?i.set(e.name,`${i.get(e.name)}, ${e.value}`):i.set(e.name,e.value)})),t.queryString.length){const e=t.queryString.map((e=>`${encodeURIComponent(e.name)}=${encodeURIComponent(e.value)}`)).join("&");o=`${t.method} ${r}?${e} HTTP/1.1\r\n`}let s="";if(t.postData)if("application/json"===t.postData.mimeType&&t.postData.text)i.set("Content-Type","application/json"),s=t.postData.text;else if("application/octet-stream"===t.postData.mimeType&&t.postData.text)i.set("Content-Type","application/octet-stream"),s=t.postData.text;else if("application/x-www-form-urlencoded"===t.postData.mimeType&&t.postData.params){const e=t.postData.params.map((e=>`${encodeURIComponent(e.name)}=${encodeURIComponent(e.value??"")}`)).join("&");i.set("Content-Type","application/x-www-form-urlencoded"),s=e}else if("multipart/form-data"===t.postData.mimeType&&t.postData.params){const e="----WebKitFormBoundary7MA4YWxkTrZu0gW";i.set("Content-Type",`multipart/form-data; boundary=${e}`),s=t.postData.params.map((t=>t.fileName?`--${e}\r\nContent-Disposition: form-data; name="${t.name}"; filename="${t.fileName}"\r\n\r\n`:`--${e}\r\nContent-Disposition: form-data; name="${t.name}"\r\n\r\n${t.value}\r\n`)).join("")+`--${e}--\r\n`}return i.forEach(((e,t)=>{o+=`${t}: ${e}\r\n`})),o+=`\r\n${s}`,o}},I={info:{key:"asynchttp",title:"AsyncHttp",link:"https://github.com/AsyncHttpClient/async-http-client",description:"Asynchronous Http and WebSocket Client library for Java"},convert:({method:e,allHeaders:t,postData:n,fullUrl:a},o)=>{const s={indent:"  ",...o},{blank:l,push:c,join:u}=new r({indent:s.indent});return c("AsyncHttpClient client = new DefaultAsyncHttpClient();"),c(`client.prepare("${e.toUpperCase()}", "${a}")`),Object.keys(t).forEach((e=>{c(`.setHeader("${e}", "${i(t[e])}")`,1)})),(null==n?void 0:n.text)&&c(`.setBody(${JSON.stringify(n.text)})`,1),c(".execute()",1),c(".toCompletableFuture()",1),c(".thenAccept(System.out::println)",1),c(".join();",1),l(),c("client.close();"),u()}},L={target:"java",client:"asynchttp",title:"AsyncHttp",generate:e=>l(I,e)};function B(e){return e.reduce(((e,t)=>(e[t.name]=t.value,e)),{})}function F(e){return/\s|-/.test(e)}function Q(e,t=0){const n=[],r=" ".repeat(t),a=" ".repeat(t+2);for(const[r,o]of Object.entries(e)){const e=F(r)?`'${r}'`:r;if(Array.isArray(o)){const r=o.map((e=>"string"==typeof e?`'${e}'`:e&&"object"==typeof e?Q(e,t+2):e)).join(`, ${a}`);n.push(`${a}${e}: [${r}]`)}else if(o&&"object"==typeof o)n.push(`${a}${e}: ${Q(o,t+2)}`);else if("string"==typeof o){let t=`${o}`;if(o.startsWith("JSON.stringify")){const e=o.split("\n");e.length>1&&(t=e.map(((e,t)=>0===t?e:`${a}${e}`)).join("\n"))}else t=`'${o}'`;n.push(`${a}${e}: ${t}`)}else n.push(`${a}${e}: ${o}`)}return`{\n${n.join(",\n")}\n${r}}`}const U={target:"js",client:"fetch",title:"Fetch",generate(e){const t={method:"GET",...e};t.method=t.method.toUpperCase();const n={method:"GET"===t.method?void 0:t.method},r=new URLSearchParams(t.queryString?B(t.queryString):void 0),a=r.size?`?${r.toString()}`:"";t.headers?.length&&(n.headers={},t.headers.forEach((e=>{n.headers[e.name]=e.value}))),t.cookies?.length&&(n.headers=n.headers||{},t.cookies.forEach((e=>{n.headers["Set-Cookie"]=n.headers["Set-Cookie"]?`${n.headers["Set-Cookie"]}; ${e.name}=${e.value}`:`${e.name}=${e.value}`}))),Object.keys(n).forEach((e=>{void 0===n[e]&&delete n[e]})),t.postData&&(n.body=t.postData.text,"application/json"===t.postData.mimeType&&(n.body=`JSON.stringify(${Q(JSON.parse(n.body))})`));const o=Object.keys(n).length?`, ${Q(n)}`:"";return`fetch('${t.url}${a}'${o})`}},{toString:z}=Object.prototype,{propertyIsEnumerable:j}=Object.prototype;function W(e,t,n){const r=[];return function e(t,n={},a=""){const o=n.indent||"\t";let i;i=void 0===n.inlineCharacterLimit?{newline:"\n",newlineOrSpace:"\n",pad:a,indent:a+o}:{newline:"@@__STRINGIFY_OBJECT_NEW_LINE__@@",newlineOrSpace:"@@__STRINGIFY_OBJECT_NEW_LINE_OR_SPACE__@@",pad:"@@__STRINGIFY_OBJECT_PAD__@@",indent:"@@__STRINGIFY_OBJECT_INDENT__@@"};const s=e=>{if(void 0===n.inlineCharacterLimit)return e;const t=e.replace(new RegExp(i.newline,"g"),"").replace(new RegExp(i.newlineOrSpace,"g")," ").replace(new RegExp(i.pad+"|"+i.indent,"g"),"");return t.length<=n.inlineCharacterLimit?t:e.replace(new RegExp(i.newline+"|"+i.newlineOrSpace,"g"),"\n").replace(new RegExp(i.pad,"g"),a).replace(new RegExp(i.indent,"g"),a+o)};if(r.includes(t))return'"[Circular]"';if(null==t||"number"==typeof t||"boolean"==typeof t||"function"==typeof t||"symbol"==typeof t||(l=t,"[object RegExp]"===z.call(l)))return String(t);var l,c;if(t instanceof Date)return`new Date('${t.toISOString()}')`;if(Array.isArray(t)){if(0===t.length)return"[]";r.push(t);const l="["+i.newline+t.map(((r,s)=>{const l=t.length-1===s?i.newline:","+i.newlineOrSpace;let c=e(r,n,a+o);return n.transform&&(c=n.transform(t,s,c)),i.indent+c+l})).join("")+i.pad+"]";return r.pop(),s(l)}if(function(e){const t=typeof e;return null!==e&&("object"===t||"function"===t)}(t)){let l=(c=t,[...Object.keys(c),...Object.getOwnPropertySymbols(c).filter((e=>j.call(c,e)))]);if(n.filter&&(l=l.filter((e=>n.filter(t,e)))),0===l.length)return"{}";r.push(t);const u="{"+i.newline+l.map(((r,s)=>{const c=l.length-1===s?i.newline:","+i.newlineOrSpace,u="symbol"==typeof r,d=!u&&/^[a-z$_][$\w]*$/i.test(r),p=u||d?r:e(r,n);let h=e(t[r],n,a+o);return n.transform&&(h=n.transform(t,r,h)),i.indent+String(p)+": "+h+c})).join("")+i.pad+"}";return r.pop(),s(u)}return t=t.replace(/\\/g,"\\\\"),t=String(t).replace(/[\r\n]/g,(e=>"\n"===e?"\\n":"\\r")),!1===n.singleQuotes?`"${t=t.replace(/"/g,'\\"')}"`:`'${t=t.replace(/'/g,"\\'")}'`}(e,t,n)}const X={info:{key:"axios",title:"Axios",link:"https://github.com/axios/axios",description:"Promise based HTTP client for the browser and node.js"},convert:({allHeaders:e,method:t,url:n,queryObj:a,postData:o},i)=>{const s={indent:"  ",...i},{blank:l,push:c,join:u,addPostProcessor:d}=new r({indent:s.indent});c("import axios from 'axios';"),l();const p={method:t,url:n};switch(Object.keys(a).length&&(p.params=a),Object.keys(e).length&&(p.headers=e),null==o?void 0:o.mimeType){case"application/x-www-form-urlencoded":o.params&&(c("const encodedParams = new URLSearchParams();"),o.params.forEach((e=>{c(`encodedParams.set('${e.name}', '${e.value}');`)})),l(),p.data="encodedParams,",d((e=>e.replace(/'encodedParams,'/,"encodedParams,"))));break;case"application/json":o.jsonObj&&(p.data=o.jsonObj);break;case"multipart/form-data":if(!o.params)break;c("const form = new FormData();"),o.params.forEach((e=>{c(`form.append('${e.name}', '${e.value||e.fileName||""}');`)})),l(),p.data="[form]";break;default:(null==o?void 0:o.text)&&(p.data=o.text)}const h=W(p,{indent:"  ",inlineCharacterLimit:80}).replace('"[form]"',"form");return c(`const options = ${h};`),l(),c("try {"),c("const { data } = await axios.request(options);",1),c("console.log(data);",1),c("} catch (error) {"),c("console.error(error);",1),c("}"),u()}},q={target:"js",client:"axios",title:"Axios",generate:e=>l(X,e)},V={target:"js",client:"ofetch",title:"ofetch",generate(e){const t={method:"GET",...e};t.method=t.method.toUpperCase();const n={method:"GET"===t.method?void 0:t.method},r=new URLSearchParams(t.queryString?B(t.queryString):void 0);r.size&&(n.query={},r.forEach(((e,t)=>{n.query[t]=e}))),t.headers?.length&&(n.headers={},t.headers.forEach((e=>{n.headers[e.name]=e.value}))),t.cookies?.length&&(n.headers=n.headers||{},t.cookies.forEach((e=>{n.headers["Set-Cookie"]=n.headers["Set-Cookie"]?`${n.headers["Set-Cookie"]}; ${e.name}=${e.value}`:`${e.name}=${e.value}`}))),Object.keys(n).forEach((e=>{void 0===n[e]&&delete n[e]})),t.postData&&(n.body=t.postData.text,"application/json"===t.postData.mimeType&&(n.body=JSON.parse(n.body)));const a=Object.keys(n).length?`, ${Q(n)}`:"";return`import { ofetch } from 'ofetch'\n\nofetch('${t.url}'${a})`}},H={info:{key:"jquery",title:"jQuery",link:"http://api.jquery.com/jquery.ajax/",description:"Perform an asynchronous HTTP (Ajax) requests with jQuery"},convert:({fullUrl:e,method:t,allHeaders:n,postData:a},o)=>{let i;const s={indent:"  ",...o},{blank:l,push:c,join:u}=new r({indent:s.indent}),d={async:!0,crossDomain:!0,url:e,method:t,headers:n};switch(null==a?void 0:a.mimeType){case"application/x-www-form-urlencoded":d.data=a.paramsObj?a.paramsObj:a.text;break;case"application/json":d.processData=!1,d.data=a.text;break;case"multipart/form-data":if(!a.params)break;if(c("const form = new FormData();"),a.params.forEach((e=>{c(`form.append('${e.name}', '${e.value||e.fileName||""}');`)})),d.processData=!1,d.contentType=!1,d.mimeType="multipart/form-data",d.data="[form]",b(n,"content-type")&&(null===(i=v(n,"content-type"))||void 0===i?void 0:i.includes("boundary"))){const e=g(n,"content-type");e&&delete d.headers[e]}l();break;default:(null==a?void 0:a.text)&&(d.data=a.text)}const p=W(d,{indent:s.indent}).replace("'[form]'","form");return c(`const settings = ${p};`),l(),c("$.ajax(settings).done(function (response) {"),c("console.log(response);",1),c("});"),u()}},Z={target:"js",client:"jquery",title:"jQuery",generate:e=>l(H,e)},G={info:{key:"xhr",title:"XMLHttpRequest",link:"https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest",description:"W3C Standard API that provides scripted client functionality"},convert:({postData:e,allHeaders:t,method:n,fullUrl:a},i)=>{let s;const l={indent:"  ",cors:!0,...i},{blank:c,push:u,join:d}=new r({indent:l.indent});switch(null==e?void 0:e.mimeType){case"application/json":u(`const data = JSON.stringify(${W(e.jsonObj,{indent:l.indent})});`),c();break;case"multipart/form-data":if(!e.params)break;if(u("const data = new FormData();"),e.params.forEach((e=>{u(`data.append('${e.name}', '${e.value||e.fileName||""}');`)})),b(t,"content-type")&&(null===(s=v(t,"content-type"))||void 0===s?void 0:s.includes("boundary"))){const e=g(t,"content-type");e&&delete t[e]}c();break;default:u(`const data = ${(null==e?void 0:e.text)?`'${e.text}'`:"null"};`),c()}return u("const xhr = new XMLHttpRequest();"),l.cors&&u("xhr.withCredentials = true;"),c(),u("xhr.addEventListener('readystatechange', function () {"),u("if (this.readyState === this.DONE) {",1),u("console.log(this.responseText);",2),u("}",1),u("});"),c(),u(`xhr.open('${n}', '${a}');`),Object.keys(t).forEach((e=>{u(`xhr.setRequestHeader('${e}', '${o(t[e])}');`)})),c(),u("xhr.send(data);"),d()}},K={target:"js",client:"xhr",title:"XHR",generate:e=>l(G,e)},Y={info:{key:"okhttp",title:"OkHttp",link:"http://square.github.io/okhttp/",description:"An HTTP Request Client Library"},convert:({postData:e,fullUrl:t,method:n,allHeaders:a},o)=>{const s={indent:"  ",...o},{blank:l,join:c,push:u}=new r({indent:s.indent});return u("val client = OkHttpClient()"),l(),(null==e?void 0:e.text)&&(e.boundary?u(`val mediaType = MediaType.parse("${e.mimeType}; boundary=${e.boundary}")`):u(`val mediaType = MediaType.parse("${e.mimeType}")`),u(`val body = RequestBody.create(mediaType, ${JSON.stringify(e.text)})`)),u("val request = Request.Builder()"),u(`.url("${t}")`,1),["GET","POST","PUT","DELETE","PATCH","HEAD"].includes(n.toUpperCase())?["POST","PUT","DELETE","PATCH"].includes(n.toUpperCase())?(null==e?void 0:e.text)?u(`.${n.toLowerCase()}(body)`,1):u(`.${n.toLowerCase()}(null)`,1):u(`.${n.toLowerCase()}()`,1):(null==e?void 0:e.text)?u(`.method("${n.toUpperCase()}", body)`,1):u(`.method("${n.toUpperCase()}", null)`,1),Object.keys(a).forEach((e=>{u(`.addHeader("${e}", "${i(a[e])}")`,1)})),u(".build()",1),l(),u("val response = client.newCall(request).execute()"),c()}},J={target:"kotlin",client:"okhttp",title:"OkHttp",generate:e=>l(Y,e)},ee={target:"node",client:"fetch",title:"Fetch",generate(e){const t={method:"GET",...e};t.method=t.method.toUpperCase();const n={method:"GET"===t.method?void 0:t.method},r=new URLSearchParams(t.queryString?B(t.queryString):void 0),a=r.size?`?${r.toString()}`:"";t.headers?.length&&(n.headers={},t.headers.forEach((e=>{n.headers[e.name]=e.value}))),t.cookies?.length&&(n.headers=n.headers||{},t.cookies.forEach((e=>{n.headers["Set-Cookie"]=n.headers["Set-Cookie"]?`${n.headers["Set-Cookie"]}; ${e.name}=${e.value}`:`${e.name}=${e.value}`}))),Object.keys(n).forEach((e=>{void 0===n[e]&&delete n[e]})),t.postData&&(n.body=t.postData.text,"application/json"===t.postData.mimeType&&(n.body=`JSON.stringify(${Q(JSON.parse(n.body))})`));const o=Object.keys(n).length?`, ${Q(n)}`:"";return`fetch('${t.url}${a}'${o})`}},te={info:{key:"axios",title:"Axios",link:"https://github.com/axios/axios",description:"Promise based HTTP client for the browser and node.js"},convert:({method:e,url:t,queryObj:n,allHeaders:a,postData:o},i)=>{const s={indent:"  ",...i},{blank:l,join:c,push:u,addPostProcessor:d}=new r({indent:s.indent});u("const axios = require('axios').default;");const p={method:e,url:t};switch(Object.keys(n).length&&(p.params=n),Object.keys(a).length&&(p.headers=a),null==o?void 0:o.mimeType){case"application/x-www-form-urlencoded":o.params&&(u("const { URLSearchParams } = require('url');"),l(),u("const encodedParams = new URLSearchParams();"),o.params.forEach((e=>{u(`encodedParams.set('${e.name}', '${e.value}');`)})),l(),p.data="encodedParams,",d((e=>e.replace(/'encodedParams,'/,"encodedParams,"))));break;case"application/json":l(),o.jsonObj&&(p.data=o.jsonObj);break;default:l(),(null==o?void 0:o.text)&&(p.data=o.text)}const h=W(p,{indent:"  ",inlineCharacterLimit:80});return u(`const options = ${h};`),l(),u("try {"),u("const { data } = await axios.request(options);",1),u("console.log(data);",1),u("} catch (error) {"),u("console.error(error);",1),u("}"),c()}},ne={target:"node",client:"axios",title:"Axios",generate:e=>l(te,e)},re={target:"node",client:"ofetch",title:"ofetch",generate(e){const t={method:"GET",...e};t.method=t.method.toUpperCase();const n={method:"GET"===t.method?void 0:t.method},r=new URLSearchParams(t.queryString?B(t.queryString):void 0);r.size&&(n.query={},r.forEach(((e,t)=>{n.query[t]=e}))),t.headers?.length&&(n.headers={},t.headers.forEach((e=>{n.headers[e.name]=e.value}))),t.cookies?.length&&(n.headers=n.headers||{},t.cookies.forEach((e=>{n.headers["Set-Cookie"]=n.headers["Set-Cookie"]?`${n.headers["Set-Cookie"]}; ${e.name}=${e.value}`:`${e.name}=${e.value}`}))),Object.keys(n).forEach((e=>{void 0===n[e]&&delete n[e]})),t.postData&&(n.body=t.postData.text,"application/json"===t.postData.mimeType&&(n.body=JSON.parse(n.body)));const a=Object.keys(n).length?`, ${Q(n)}`:"";return`import { ofetch } from 'ofetch'\n\nofetch('${t.url}'${a})`}},ae={target:"node",client:"undici",title:"undici",generate(e){const t={method:"GET",...e};t.method=t.method.toUpperCase();const n={method:"GET"===t.method?void 0:t.method},r=new URLSearchParams(t.queryString?B(t.queryString):void 0),a=r.size?`?${r.toString()}`:"";t.headers?.length&&(n.headers={},t.headers.forEach((e=>{n.headers[e.name]=e.value}))),t.cookies?.length&&(n.headers=n.headers||{},t.cookies.forEach((e=>{n.headers["Set-Cookie"]=n.headers["Set-Cookie"]?`${n.headers["Set-Cookie"]}; ${e.name}=${e.value}`:`${e.name}=${e.value}`}))),Object.keys(n).forEach((e=>{void 0===n[e]&&delete n[e]})),t.postData&&(n.body=t.postData.text,"application/json"===t.postData.mimeType&&(n.body=`JSON.stringify(${Q(JSON.parse(n.body))})`));const o=Object.keys(n).length?`, ${Q(n)}`:"";return`import { request } from 'undici'\n\nconst { statusCode, body } = await request('${t.url}${a}'${o})`}},oe=(e,t,n,r)=>{const a=`${e} *${t} = `;return`${a}${ie(n,r?a.length:void 0)};`},ie=(e,t)=>{const n=void 0===t?", ":`,\n   ${" ".repeat(t)}`;switch(Object.prototype.toString.call(e)){case"[object Number]":return`@${e}`;case"[object Array]":return`@[ ${e.map((e=>ie(e))).join(n)} ]`;case"[object Object]":{const t=[];for(const n in e)t.push(`@"${n}": ${ie(e[n])}`);return`@{ ${t.join(n)} }`}case"[object Boolean]":return e?"@YES":"@NO";default:return null==e?"":`@"${e.toString().replace(/"/g,'\\"')}"`}},se={info:{key:"nsurlsession",title:"NSURLSession",link:"https://developer.apple.com/library/mac/documentation/Foundation/Reference/NSURLSession_class/index.html",description:"Foundation's NSURLSession request"},convert:({allHeaders:e,postData:t,method:n,fullUrl:a},o)=>{let i;const s={indent:"    ",pretty:!0,timeout:10,...o},{push:l,join:c,blank:u}=new r({indent:s.indent}),d={hasHeaders:!1,hasBody:!1};if(l("#import <Foundation/Foundation.h>"),Object.keys(e).length&&(d.hasHeaders=!0,u(),l(oe("NSDictionary","headers",e,s.pretty))),t&&(t.text||t.jsonObj||t.params))switch(d.hasBody=!0,t.mimeType){case"application/x-www-form-urlencoded":if(null===(i=t.params)||void 0===i?void 0:i.length){u();const[e,...n]=t.params;l(`NSMutableData *postData = [[NSMutableData alloc] initWithData:[@"${e.name}=${e.value}" dataUsingEncoding:NSUTF8StringEncoding]];`),n.forEach((({name:e,value:t})=>{l(`[postData appendData:[@"&${e}=${t}" dataUsingEncoding:NSUTF8StringEncoding]];`)}))}else d.hasBody=!1;break;case"application/json":t.jsonObj&&(l(oe("NSDictionary","parameters",t.jsonObj,s.pretty)),u(),l("NSData *postData = [NSJSONSerialization dataWithJSONObject:parameters options:0 error:nil];"));break;case"multipart/form-data":l(oe("NSArray","parameters",t.params||[],s.pretty)),l(`NSString *boundary = @"${t.boundary}";`),u(),l("NSError *error;"),l("NSMutableString *body = [NSMutableString string];"),l("for (NSDictionary *param in parameters) {"),l('[body appendFormat:@"--%@\\r\\n", boundary];',1),l('if (param[@"fileName"]) {',1),l('[body appendFormat:@"Content-Disposition:form-data; name=\\"%@\\"; filename=\\"%@\\"\\r\\n", param[@"name"], param[@"fileName"]];',2),l('[body appendFormat:@"Content-Type: %@\\r\\n\\r\\n", param[@"contentType"]];',2),l('[body appendFormat:@"%@", [NSString stringWithContentsOfFile:param[@"fileName"] encoding:NSUTF8StringEncoding error:&error]];',2),l("if (error) {",2),l('NSLog(@"%@", error);',3),l("}",2),l("} else {",1),l('[body appendFormat:@"Content-Disposition:form-data; name=\\"%@\\"\\r\\n\\r\\n", param[@"name"]];',2),l('[body appendFormat:@"%@", param[@"value"]];',2),l("}",1),l("}"),l('[body appendFormat:@"\\r\\n--%@--\\r\\n", boundary];'),l("NSData *postData = [body dataUsingEncoding:NSUTF8StringEncoding];");break;default:u(),l(`NSData *postData = [[NSData alloc] initWithData:[@"${t.text}" dataUsingEncoding:NSUTF8StringEncoding]];`)}return u(),l(`NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:[NSURL URLWithString:@"${a}"]`),l("                                                       cachePolicy:NSURLRequestUseProtocolCachePolicy"),l(`                                                   timeoutInterval:${s.timeout.toFixed(1)}];`),l(`[request setHTTPMethod:@"${n}"];`),d.hasHeaders&&l("[request setAllHTTPHeaderFields:headers];"),d.hasBody&&l("[request setHTTPBody:postData];"),u(),l("NSURLSession *session = [NSURLSession sharedSession];"),l("NSURLSessionDataTask *dataTask = [session dataTaskWithRequest:request"),l("                                            completionHandler:^(NSData *data, NSURLResponse *response, NSError *error) {"),l("                                            if (error) {",1),l('                                            NSLog(@"%@", error);',2),l("                                            } else {",1),l("                                            NSHTTPURLResponse *httpResponse = (NSHTTPURLResponse *) response;",2),l('                                            NSLog(@"%@", httpResponse);',2),l("                                            }",1),l("                                            }];"),l("[dataTask resume];"),c()}},le={target:"objc",client:"nsurlsession",title:"NSURLSession",generate:e=>l(se,e)},ce={info:{key:"cohttp",title:"CoHTTP",link:"https://github.com/mirage/ocaml-cohttp",description:"Cohttp is a very lightweight HTTP server using Lwt or Async for OCaml"},convert:({fullUrl:e,allHeaders:t,postData:n,method:a},o)=>{const s={indent:"  ",...o},{push:l,blank:c,join:u}=new r({indent:s.indent});l("open Cohttp_lwt_unix"),l("open Cohttp"),l("open Lwt"),c(),l(`let uri = Uri.of_string "${e}" in`);const d=Object.keys(t);1===d.length?l(`let headers = Header.add (Header.init ()) "${d[0]}" "${i(t[d[0]])}" in`):d.length>1&&(l("let headers = Header.add_list (Header.init ()) ["),d.forEach((e=>{l(`("${e}", "${i(t[e])}");`,1)})),l("] in")),(null==n?void 0:n.text)&&l(`let body = Cohttp_lwt_body.of_string ${JSON.stringify(n.text)} in`),c();const p=d.length?"~headers ":"",h=(null==n?void 0:n.text)?"~body ":"",f=["get","post","head","delete","patch","put","options"].includes(a.toLowerCase())?`\`${a.toUpperCase()}`:`(Code.method_of_string "${a}")`;return l(`Client.call ${p}${h}${f} uri`),l(">>= fun (res, body_stream) ->"),l("(* Do stuff with the result *)",1),u()}},ue={target:"ocaml",client:"cohttp",title:"Cohttp",generate:e=>l(ce,e)},de=(e,t,n)=>{switch(n=n||"",t=t||"",Object.prototype.toString.call(e)){case"[object Null]":case"[object Undefined]":default:return"null";case"[object String]":return`'${a(e,{delimiter:"'",escapeNewlines:!1})}'`;case"[object Number]":return e.toString();case"[object Array]":{const r=e.map((e=>de(e,`${t}${t}`,t))).join(`,\n${t}`);return`[\n${t}${r}\n${n}]`}case"[object Object]":{const r=[];for(const n in e)Object.prototype.hasOwnProperty.call(e,n)&&r.push(`${de(n,t)} => ${de(e[n],`${t}${t}`,t)}`);return`[\n${t}${r.join(`,\n${t}`)}\n${n}]`}}},pe={info:{key:"curl",title:"cURL",link:"http://php.net/manual/en/book.curl.php",description:"PHP with ext-curl"},convert:({uriObj:e,postData:t,fullUrl:n,method:a,httpVersion:o,cookies:s,headersObj:l},c={})=>{const{closingTag:u=!1,indent:d="  ",maxRedirects:p=10,namedErrors:h=!1,noTags:f=!1,shortTags:m=!1,timeout:g=30}=c,{push:v,blank:b,join:y}=new r({indent:d});f||(v(m?"<?":"<?php"),b()),v("$curl = curl_init();"),b();const w=[{escape:!0,name:"CURLOPT_PORT",value:""===e.port?null:e.port},{escape:!0,name:"CURLOPT_URL",value:n},{escape:!1,name:"CURLOPT_RETURNTRANSFER",value:"true"},{escape:!0,name:"CURLOPT_ENCODING",value:""},{escape:!1,name:"CURLOPT_MAXREDIRS",value:p},{escape:!1,name:"CURLOPT_TIMEOUT",value:g},{escape:!1,name:"CURLOPT_HTTP_VERSION",value:"HTTP/1.0"===o?"CURL_HTTP_VERSION_1_0":"CURL_HTTP_VERSION_1_1"},{escape:!0,name:"CURLOPT_CUSTOMREQUEST",value:a},{escape:!(null==t?void 0:t.jsonObj),name:"CURLOPT_POSTFIELDS",value:t?t.jsonObj?`json_encode(${de(t.jsonObj,d.repeat(2),d)})`:t.text:void 0}];v("curl_setopt_array($curl, [");const O=new r({indent:d,join:`\n${d}`});w.forEach((({value:e,name:t,escape:n})=>{null!=e&&O.push(`${t} => ${n?JSON.stringify(e):e},`)}));const k=s.map((e=>`${encodeURIComponent(e.name)}=${encodeURIComponent(e.value)}`));k.length&&O.push(`CURLOPT_COOKIE => "${k.join("; ")}",`);const x=Object.keys(l).sort().map((e=>`"${e}: ${i(l[e])}"`));return x.length&&(O.push("CURLOPT_HTTPHEADER => ["),O.push(x.join(`,\n${d}${d}`),1),O.push("],")),v(O.join(),1),v("]);"),b(),v("$response = curl_exec($curl);"),v("$err = curl_error($curl);"),b(),v("curl_close($curl);"),b(),v("if ($err) {"),v(h?'echo array_flip(get_defined_constants(true)["curl"])[$err];':'echo "cURL Error #:" . $err;',1),v("} else {"),v("echo $response;",1),v("}"),!f&&u&&(b(),v("?>")),y()}},he={target:"php",client:"curl",title:"cURL",generate:e=>l(pe,e)},fe={info:{key:"guzzle",title:"Guzzle",link:"http://docs.guzzlephp.org/en/stable/",description:"PHP with Guzzle"},convert:({postData:e,fullUrl:t,method:n,cookies:a,headersObj:i},s)=>{let l;const c={closingTag:!1,indent:"  ",noTags:!1,shortTags:!1,...s},{push:u,blank:d,join:p}=new r({indent:c.indent}),{code:h,push:f,join:m}=new r({indent:c.indent});switch(c.noTags||(u(c.shortTags?"<?":"<?php"),d()),null==e?void 0:e.mimeType){case"application/x-www-form-urlencoded":f(`'form_params' => ${de(e.paramsObj,c.indent+c.indent,c.indent)},`,1);break;case"multipart/form-data":{const t=[];if(e.params&&e.params.forEach((e=>{if(e.fileName){const n={name:e.name,filename:e.fileName,contents:e.value};e.contentType&&(n.headers={"Content-Type":e.contentType}),t.push(n)}else e.value&&t.push({name:e.name,contents:e.value})})),t.length&&(f(`'multipart' => ${de(t,c.indent+c.indent,c.indent)}`,1),b(i,"content-type")&&(null===(l=v(i,"content-type"))||void 0===l?void 0:l.indexOf("boundary")))){const e=g(i,"content-type");e&&delete i[e]}break}default:(null==e?void 0:e.text)&&f(`'body' => ${de(e.text)},`,1)}const y=Object.keys(i).sort().map((e=>`${c.indent}${c.indent}'${e}' => '${o(i[e])}',`)),w=a.map((e=>`${encodeURIComponent(e.name)}=${encodeURIComponent(e.value)}`)).join("; ");return w.length&&y.push(`${c.indent}${c.indent}'cookie' => '${o(w)}',`),y.length&&(f("'headers' => [",1),f(y.join("\n")),f("],",1)),u("$client = new \\GuzzleHttp\\Client();"),d(),h.length?(u(`$response = $client->request('${n}', '${t}', [`),u(m()),u("]);")):u(`$response = $client->request('${n}', '${t}');`),d(),u("echo $response->getBody();"),!c.noTags&&c.closingTag&&(d(),u("?>")),p()}},me={target:"php",client:"guzzle",title:"Guzzle",generate:e=>l(fe,e)},ge=e=>({method:t,headersObj:n,cookies:o,uriObj:i,fullUrl:s,postData:l,allHeaders:c})=>{const{push:u,join:d}=new r;if(!["GET","POST","PUT","DELETE","PATCH","HEAD","OPTIONS"].includes(t.toUpperCase()))return"Method not supported";const p=[],h=Object.keys(n);return h.length&&(u("$headers=@{}"),h.forEach((e=>{"connection"!==e&&u(`$headers.Add("${e}", "${a(n[e],{escapeChar:"`"})}")`)})),p.push("-Headers $headers")),o.length&&(u("$session = New-Object Microsoft.PowerShell.Commands.WebRequestSession"),o.forEach((e=>{u("$cookie = New-Object System.Net.Cookie"),u(`$cookie.Name = '${e.name}'`),u(`$cookie.Value = '${e.value}'`),u(`$cookie.Domain = '${i.host}'`),u("$session.Cookies.Add($cookie)")})),p.push("-WebSession $session")),(null==l?void 0:l.text)&&(p.push(`-ContentType '${a(v(c,"content-type"),{delimiter:"'",escapeChar:"`"})}'`),p.push(`-Body '${l.text}'`)),u(`$response = ${e} -Uri '${s}' -Method ${t} ${p.join(" ")}`),d()},ve={info:{key:"webrequest",title:"Invoke-WebRequest",link:"https://docs.microsoft.com/en-us/powershell/module/Microsoft.PowerShell.Utility/Invoke-WebRequest",description:"Powershell Invoke-WebRequest client"},convert:ge("Invoke-WebRequest")},be={target:"powershell",client:"webrequest",title:"Invoke-WebRequest",generate:e=>l(ve,e)},ye={info:{key:"restmethod",title:"Invoke-RestMethod",link:"https://docs.microsoft.com/en-us/powershell/module/Microsoft.PowerShell.Utility/Invoke-RestMethod",description:"Powershell Invoke-RestMethod client"},convert:ge("Invoke-RestMethod")},we={info:{key:"python3",title:"http.client",link:"https://docs.python.org/3/library/http.client.html",description:"Python3 HTTP Client"},convert:({uriObj:{path:e,protocol:t,host:n},postData:a,allHeaders:o,method:s},l={})=>{const{insecureSkipVerify:c=!1}=l,{push:u,blank:d,join:p}=new r;u("import http.client"),c&&u("import ssl"),d(),"https:"===t?(u(`conn = http.client.HTTPSConnection("${n}"${c?", context = ssl._create_unverified_context()":""})`),d()):(u(`conn = http.client.HTTPConnection("${n}")`),d());const h=JSON.stringify(null==a?void 0:a.text);h&&(u(`payload = ${h}`),d());const f=o,m=Object.keys(f).length;if(1===m)for(const e in f)u(`headers = { '${e}': "${i(f[e])}" }`),d();else if(m>1){let e=1;u("headers = {");for(const t in f)e++!==m?u(`    '${t}': "${i(f[t])}",`):u(`    '${t}': "${i(f[t])}"`);u("}"),d()}return u(h&&m?`conn.request("${s}", "${e}", payload, headers)`:h&&!m?`conn.request("${s}", "${e}", payload)`:!h&&m?`conn.request("${s}", "${e}", headers=headers)`:`conn.request("${s}", "${e}")`),d(),u("res = conn.getresponse()"),u("data = res.read()"),d(),u('print(data.decode("utf-8"))'),p()}};function Oe(e,t,n,r,a){const o=r.repeat(a),i=r.repeat(a-1),s=n?`,\n${o}`:", ",l="object"===e?"{":"[",c="object"===e?"}":"]";return n?`${l}\n${o}${t.join(s)}\n${i}${c}`:"object"===e&&t.length>0?`${l} ${t.join(s)} ${c}`:`${l}${t.join(s)}${c}`}const ke=(e,t,n)=>{switch(n=void 0===n?1:n+1,Object.prototype.toString.call(e)){case"[object Number]":return e;case"[object Array]":{let r=!1;return Oe("array",e.map((e=>("[object Object]"===Object.prototype.toString.call(e)&&(r=Object.keys(e).length>1),ke(e,t,n)))),r,t.indent,n)}case"[object Object]":{const r=[];for(const a in e)r.push(`"${a}": ${ke(e[a],t,n)}`);return Oe("object",r,t.pretty&&r.length>1,t.indent,n)}case"[object Null]":return"None";case"[object Boolean]":return e?"True":"False";default:return null==e?"":`"${e.toString().replace(/"/g,'\\"')}"`}},xe=["HEAD","GET","POST","PUT","PATCH","DELETE","OPTIONS"],Ee={info:{key:"requests",title:"Requests",link:"http://docs.python-requests.org/en/latest/api/#requests.request",description:"Requests HTTP library"},convert:({queryObj:e,url:t,postData:n,allHeaders:a,method:o},s)=>{const l={indent:"    ",pretty:!0,...s},{push:c,blank:u,join:d}=new r({indent:l.indent});let p;c("import requests"),u(),c(`url = "${t}"`),u(),Object.keys(e).length&&(p=`querystring = ${JSON.stringify(e)}`,c(p),u());const h=a;let f={};const m={};let v=!1,b=!1,y=!1;switch(null==n?void 0:n.mimeType){case"application/json":n.jsonObj&&(c(`payload = ${ke(n.jsonObj,l)}`),y=!0,b=!0);break;case"multipart/form-data":if(!n.params)break;if(f={},n.params.forEach((e=>{e.fileName?(m[e.name]=`open('${e.fileName}', 'rb')`,v=!0):(f[e.name]=e.value,b=!0)})),v){c(`files = ${ke(m,l)}`),b&&c(`payload = ${ke(f,l)}`);const e=g(h,"content-type");e&&delete h[e]}else{const e=JSON.stringify(n.text);e&&(c(`payload = ${e}`),b=!0)}break;default:{if(!n)break;if("application/x-www-form-urlencoded"===n.mimeType&&n.paramsObj){c(`payload = ${ke(n.paramsObj,l)}`),b=!0;break}const e=JSON.stringify(n.text);e&&(c(`payload = ${e}`),b=!0)}}const w=Object.keys(h).length;if(0===w&&(b||v))u();else if(1===w)for(const e in h)c(`headers = {"${e}": "${i(h[e])}"}`),u();else if(w>1){let e=1;c("headers = {");for(const t in h)c(e!==w?`"${t}": "${i(h[t])}",`:`"${t}": "${i(h[t])}"`,1),e+=1;c("}"),u()}let O=xe.includes(o)?`response = requests.${o.toLowerCase()}(url`:`response = requests.request("${o}", url`;return b&&(O+=y?", json=payload":", data=payload"),v&&(O+=", files=files"),w>0&&(O+=", headers=headers"),p&&(O+=", params=querystring"),O+=")",c(O),u(),c("print(response.json())"),d()}},Se={info:{key:"httr",title:"httr",link:"https://cran.r-project.org/web/packages/httr/vignettes/quickstart.html",description:"httr: Tools for Working with URLs and HTTP"},convert:({url:e,queryObj:t,queryString:n,postData:a,allHeaders:s,method:l},c={})=>{let u,d;const{push:p,blank:h,join:f}=new r({indent:null!==(u=c.indent)&&void 0!==u?u:"  "});p("library(httr)"),h(),p(`url <- "${e}"`),h();const m=t;delete t.key;const g=Object.entries(m),b=g.length;if(1===b){const e=g[0];p(`queryString <- list(${e[0]} = "${e[1]}")`),h()}else b>1&&(p("queryString <- list("),g.forEach((([e,t],n)=>{p(`${e} = "${t}"${n!==b-1?",":""}`,1)})),p(")"),h());const y=JSON.stringify(null==a?void 0:a.text);if(y&&(p(`payload <- ${y}`),h()),a&&(a.text||a.jsonObj||a.params))switch(a.mimeType){case"application/x-www-form-urlencoded":p('encode <- "form"'),h();break;case"application/json":p('encode <- "json"'),h();break;case"multipart/form-data":p('encode <- "multipart"'),h();break;default:p('encode <- "raw"'),h()}const w=v(s,"cookie"),O=v(s,"accept"),k=w?`set_cookies(\`${String(w).replace(/;/g,'", `').replace(/` /g,"`").replace(/[=]/g,'` = "')}")`:void 0,x=O?`accept("${i(O)}")`:void 0,E=`content_type("${i(null!==(d=null==a?void 0:a.mimeType)&&void 0!==d?d:"application/octet-stream")}")`,S=Object.entries(s).filter((([e])=>!["cookie","accept","content-type"].includes(e.toLowerCase()))).map((([e,t])=>`'${e}' = '${o(t)}'`)).join(", "),_=S?`add_headers(${S})`:void 0;let C=`response <- VERB("${l}", url`;y&&(C+=", body = payload"),n.length&&(C+=", query = queryString");const T=[_,E,x,k].filter((e=>!!e)).join(", ");return T&&(C+=`, ${T}`),a&&(a.text||a.jsonObj||a.params)&&(C+=", encode = encode"),C+=")",p(C),h(),p('content(response, "text")'),f()}},_e={info:{key:"native",title:"net::http",link:"http://ruby-doc.org/stdlib-2.2.1/libdoc/net/http/rdoc/Net/HTTP.html",description:"Ruby HTTP client"},convert:({uriObj:e,method:t,fullUrl:n,postData:a,allHeaders:i},s={})=>{const{insecureSkipVerify:l=!1}=s,{push:c,blank:u,join:d}=new r;c("require 'uri'"),c("require 'net/http'"),u();const p=t.toUpperCase(),h=p.charAt(0)+p.substring(1).toLowerCase();["GET","POST","HEAD","DELETE","PATCH","PUT","OPTIONS","COPY","LOCK","UNLOCK","MOVE","TRACE"].includes(p)||(c(`class Net::HTTP::${h} < Net::HTTPRequest`),c(`  METHOD = '${p.toUpperCase()}'`),c(`  REQUEST_HAS_BODY = '${(null==a?void 0:a.text)?"true":"false"}'`),c("  RESPONSE_HAS_BODY = true"),c("end"),u()),c(`url = URI("${n}")`),u(),c("http = Net::HTTP.new(url.host, url.port)"),"https:"===e.protocol&&(c("http.use_ssl = true"),l&&c("http.verify_mode = OpenSSL::SSL::VERIFY_NONE")),u(),c(`request = Net::HTTP::${h}.new(url)`);const f=Object.keys(i);return f.length&&f.forEach((e=>{c(`request["${e}"] = '${o(i[e])}'`)})),(null==a?void 0:a.text)&&c(`request.body = ${JSON.stringify(a.text)}`),u(),c("response = http.request(request)"),c("puts response.read_body"),d()}},Ce=(e="")=>/^[a-z0-9-_/.@%^=:]+$/i.test(e)?e:`'${e.replace(/'/g,"'\\''")}'`,Te={info:{key:"wget",title:"Wget",link:"https://www.gnu.org/software/wget/",description:"a free software package for retrieving files using HTTP, HTTPS"},convert:({method:e,postData:t,allHeaders:n,fullUrl:a},o)=>{const i={indent:"  ",short:!1,verbose:!1,...o},{push:s,join:l}=new r({indent:i.indent,join:!1!==i.indent?` \\\n${i.indent}`:" "});var c;return i.verbose?s("wget "+(i.short?"-v":"--verbose")):s("wget "+(i.short?"-q":"--quiet")),s(`--method ${Ce(e)}`),Object.keys(n).forEach((e=>{const t=`${e}: ${n[e]}`;s(`--header ${Ce(t)}`)})),(null==t?void 0:t.text)&&s(`--body-data ${c=Ce(t.text),c.replace(/\r/g,"\\r").replace(/\n/g,"\\n")}`),s(i.short?"-O":"--output-document"),s(`- ${Ce(a)}`),l()}},Ae={info:{key:"httpie",title:"HTTPie",link:"http://httpie.org/",description:"a CLI, cURL-like tool for humans"},convert:({allHeaders:e,postData:t,queryObj:n,fullUrl:a,method:o,url:i},s)=>{const l={body:!1,cert:!1,headers:!1,indent:"  ",pretty:!1,print:!1,queryParams:!1,short:!1,style:!1,timeout:!1,verbose:!1,verify:!1,...s},{push:c,join:u,unshift:d}=new r({indent:l.indent,join:!1!==l.indent?` \\\n${l.indent}`:" "});let p=!1;const h=[];return l.headers&&h.push(l.short?"-h":"--headers"),l.body&&h.push(l.short?"-b":"--body"),l.verbose&&h.push(l.short?"-v":"--verbose"),l.print&&h.push(`${l.short?"-p":"--print"}=${l.print}`),l.verify&&h.push(`--verify=${l.verify}`),l.cert&&h.push(`--cert=${l.cert}`),l.pretty&&h.push(`--pretty=${l.pretty}`),l.style&&h.push(`--style=${l.style}`),l.timeout&&h.push(`--timeout=${l.timeout}`),l.queryParams&&Object.keys(n).forEach((e=>{const t=n[e];Array.isArray(t)?t.forEach((t=>{c(`${e}==${Ce(t)}`)})):c(`${e}==${Ce(t)}`)})),Object.keys(e).sort().forEach((t=>{c(`${t}:${Ce(e[t])}`)})),"application/x-www-form-urlencoded"===(null==t?void 0:t.mimeType)?t.params?.length&&(h.push(l.short?"-f":"--form"),t.params.forEach((e=>{c(`${e.name}=${Ce(e.value)}`)}))):p=!0,d(`http ${h.length?`${h.join(" ")} `:""}${o} ${i=Ce(l.queryParams?i:a)}`),p&&(null==t?void 0:t.text)&&d(`echo ${Ce(t.text)} | `),u()}},Re=(e,t)=>t.repeat(e),Pe=(e,t,n,r)=>{const a=Re(r,n),o=Re(r-1,n),i=t?`,\n${a}`:", ";return t?`[\n${a}${e.join(i)}\n${o}]`:`[${e.join(i)}]`},De=(e,t,n)=>`let ${e} = ${Me(t,n)}`,Me=(e,t,n)=>{switch(n=void 0===n?1:n+1,Object.prototype.toString.call(e)){case"[object Number]":return e;case"[object Array]":{let r=!1;const a=e.map((e=>("[object Object]"===Object.prototype.toString.call(e)&&(r=Object.keys(e).length>1),Me(e,t,n))));return Pe(a,r,t.indent,n)}case"[object Object]":{const r=[];for(const a in e)r.push(`"${a}": ${Me(e[a],t,n)}`);return Pe(r,t.pretty&&r.length>1,t.indent,n)}case"[object Boolean]":return e.toString();default:return null==e?"":`"${e.toString().replace(/"/g,'\\"')}"`}},$e={info:{key:"nsurlsession",title:"NSURLSession",link:"https://developer.apple.com/library/mac/documentation/Foundation/Reference/NSURLSession_class/index.html",description:"Foundation's NSURLSession request"},convert:({allHeaders:e,postData:t,fullUrl:n,method:a},o)=>{let i;const s={indent:"  ",pretty:!0,timeout:"10",...o},{push:l,blank:c,join:u}=new r({indent:s.indent}),d={hasHeaders:!1,hasBody:!1};if(l("import Foundation"),Object.keys(e).length&&(d.hasHeaders=!0,c(),l(De("headers",e,s))),t&&(t.text||t.jsonObj||t.params))switch(d.hasBody=!0,t.mimeType){case"application/x-www-form-urlencoded":if(c(),null===(i=t.params)||void 0===i?void 0:i.length){const[e,...n]=t.params;l(`let postData = NSMutableData(data: "${e.name}=${e.value}".data(using: String.Encoding.utf8)!)`),n.forEach((({name:e,value:t})=>{l(`postData.append("&${e}=${t}".data(using: String.Encoding.utf8)!)`)}))}else d.hasBody=!1;break;case"application/json":t.jsonObj&&(l(`${De("parameters",t.jsonObj,s)} as [String : Any]`),c(),l("let postData = JSONSerialization.data(withJSONObject: parameters, options: [])"));break;case"multipart/form-data":l(De("parameters",t.params,s)),c(),l(`let boundary = "${t.boundary}"`),c(),l('var body = ""'),l("var error: NSError? = nil"),l("for param in parameters {"),l('let paramName = param["name"]!',1),l('body += "--\\(boundary)\\r\\n"',1),l('body += "Content-Disposition:form-data; name=\\"\\(paramName)\\""',1),l('if let filename = param["fileName"] {',1),l('let contentType = param["content-type"]!',2),l("let fileContent = String(contentsOfFile: filename, encoding: String.Encoding.utf8)",2),l("if (error != nil) {",2),l("print(error as Any)",3),l("}",2),l('body += "; filename=\\"\\(filename)\\"\\r\\n"',2),l('body += "Content-Type: \\(contentType)\\r\\n\\r\\n"',2),l("body += fileContent",2),l('} else if let paramValue = param["value"] {',1),l('body += "\\r\\n\\r\\n\\(paramValue)"',2),l("}",1),l("}");break;default:c(),l(`let postData = NSData(data: "${t.text}".data(using: String.Encoding.utf8)!)`)}return c(),l(`let request = NSMutableURLRequest(url: NSURL(string: "${n}")! as URL,`),l("                                        cachePolicy: .useProtocolCachePolicy,"),l(`                                    timeoutInterval: ${Number.parseInt(s.timeout,10).toFixed(1)})`),l(`request.httpMethod = "${a}"`),d.hasHeaders&&l("request.allHTTPHeaderFields = headers"),d.hasBody&&l("request.httpBody = postData as Data"),c(),l("let session = URLSession.shared"),l("let dataTask = session.dataTask(with: request as URLRequest, completionHandler: { (data, response, error) -> Void in"),l("if (error != nil) {",1),l("print(error as Any)",2),l("} else {",1),l("let httpResponse = response as? HTTPURLResponse",2),l("print(httpResponse)",2),l("}",1),l("})"),c(),l("dataTask.resume()"),u()}},Ne=[{key:"c",title:"C",default:"libcurl",clients:[c]},{key:"csharp",title:"C#",default:"restsharp",clients:[w,k]},{key:"clojure",title:"Clojure",default:"clj_http",clients:[P]},{key:"dart",title:"Dart",default:"http",clients:[D]},{key:"go",title:"Go",default:"native",clients:[$]},{key:"http",title:"HTTP",default:"http1.1",clients:[N]},{key:"java",title:"Java",default:"unirest",clients:[L,d,h,m]},{key:"js",title:"JavaScript",default:"fetch",clients:[U,q,V,Z,K]},{key:"kotlin",title:"Kotlin",default:"okhttp",clients:[J]},{key:"node",title:"Node.js",default:"fetch",clients:[ee,ne,re,ae]},{key:"objc",title:"Objective-C",default:"nsurlsession",clients:[le]},{key:"ocaml",title:"OCaml",default:"cohttp",clients:[ue]},{key:"php",title:"PHP",default:"curl",clients:[he,me]},{key:"powershell",title:"Powershell",default:"webrequest",clients:[be,{target:"powershell",client:"restmethod",title:"Invoke-RestMethod",generate:e=>l(ye,e)}]},{key:"python",title:"Python",default:"python3",clients:[{target:"python",client:"python3",title:"http.client",generate:e=>l(we,e)},{target:"python",client:"requests",title:"Requests",generate:e=>l(Ee,e)}]},{key:"r",title:"R",default:"httr",clients:[{target:"r",client:"httr",title:"httr",generate:e=>l(Se,e)}]},{key:"ruby",title:"Ruby",default:"native",clients:[{target:"ruby",client:"native",title:"net::http",generate:e=>l(_e,e)}]},{key:"shell",title:"Shell",default:"curl",clients:[{target:"shell",client:"curl",title:"Curl",generate(e,t){const n={method:"GET",...e};n.method=n.method.toUpperCase();const r=["curl"],a=n.queryString?.length?"?"+n.queryString.map((e=>`${encodeURIComponent(e.name)}=${encodeURIComponent(e.value)}`)).join("&"):"",o=`${n.url}${a}`,i=/[\s<>[\]{}|\\^%]/.test(o),s=a||i?`'${o}'`:o;if(r[0]=`curl ${s}`,"GET"!==n.method&&r.push(`--request ${n.method}`),t?.auth?.username&&t?.auth?.password&&r.push(`--user '${t.auth.username}:${t.auth.password}'`),n.headers?.length){n.headers.forEach((e=>{r.push(`--header '${e.name}: ${e.value}'`)}));const e=n.headers.find((e=>"accept-encoding"===e.name.toLowerCase()));e&&/gzip|deflate/.test(e.value)&&r.push("--compressed")}if(n.cookies?.length){const e=n.cookies.map((e=>`${encodeURIComponent(e.name)}=${encodeURIComponent(e.value)}`)).join("; ");r.push(`--cookie '${e}'`)}if(n.postData)if("application/json"===n.postData.mimeType){if(n.postData.text){const e=JSON.parse(n.postData.text),t=JSON.stringify(e,null,2);r.push(`--data '${t}'`)}}else"application/octet-stream"===n.postData.mimeType?r.push(`--data-binary '${n.postData.text}'`):"application/x-www-form-urlencoded"===n.postData.mimeType&&n.postData.params?n.postData.params.forEach((e=>{r.push(`--data-urlencode '${encodeURIComponent(e.name)}=${e.value}'`)})):"multipart/form-data"===n.postData.mimeType&&n.postData.params?n.postData.params.forEach((e=>{void 0!==e.fileName?r.push(`--form '${e.name}=@${e.fileName}'`):r.push(`--form '${e.name}=${e.value}'`)})):r.push(`--data "${n.postData.text}"`);return r.join(" \\\n  ")}},{target:"shell",client:"wget",title:"Wget",generate:e=>l(Te,e)},{target:"shell",client:"httpie",title:"HTTPie",generate:e=>l(Ae,e)}]},{key:"swift",title:"Swift",default:"nsurlsession",clients:[{target:"swift",client:"nsurlsession",title:"NSURLSession",generate:e=>l($e,e)}]}];function Ie(){function e(e,t){return Ne.find((({key:t})=>t===e))?.clients.find((e=>e.client===t))}return{print:(t,n,r)=>e(t,n)?.generate(r),clients:()=>Ne,plugins:()=>Ne.flatMap((({key:e,clients:t})=>t.map((t=>({target:e,client:t.client}))))),findPlugin:e,hasPlugin:(t,n)=>Boolean(e(t,n))}}},7838:(e,t,n)=>{"use strict";n.d(t,{D:()=>a});var r,a=((r=a||{}).Space=" ",r.Enter="Enter",r.Escape="Escape",r.Backspace="Backspace",r.Delete="Delete",r.ArrowLeft="ArrowLeft",r.ArrowUp="ArrowUp",r.ArrowRight="ArrowRight",r.ArrowDown="ArrowDown",r.Home="Home",r.End="End",r.PageUp="PageUp",r.PageDown="PageDown",r.Tab="Tab",r)},7862:(e,t,n)=>{"use strict";n.d(t,{JW:()=>d,qy:()=>u});var r=n(1987),a=n(277),o=n(569),i=n(1564),s=n(5670),l=n(9072),c=n(1559);const u=(0,r.h)([a.a,o.q,s.F,l.z,c._],"html"),d=(0,r.h)([a.a,i.J,s.F,l.z,c._],"svg")},7970:(e,t,n)=>{"use strict";n.d(t,{YR:()=>a});var r=n(9656);function a(e,t,n,a){let o,i,s;"function"==typeof t&&"function"!=typeof n?(i=void 0,s=t,o=n):(i=t,s=n,o=a),(0,r.VG)(e,i,(function(e,t){const n=t[t.length-1],r=n?n.children.indexOf(e):void 0;return s(e,r,n)}),o)}},7984:(e,t,n)=>{"use strict";n.d(t,{CV:()=>p});var r=n(2071),a=n(2555),o=n(4476);const i=o.z.object({description:o.z.string().optional(),example:o.z.string().optional()}),s=o.z.record(o.z.string(),i);var l=n(5785),c=n(8065);const u=o.z.object({type:o.z.literal("collection").optional().default("collection"),openapi:o.z.union([o.z.string(),o.z.literal("3.0.0"),o.z.literal("3.1.0"),o.z.literal("4.0.0")]).optional().default("3.1.0"),jsonSchemaDialect:o.z.string().optional(),info:c.XX.optional(),security:o.z.array(l.AH).optional().default([]),externalDocs:c.MA.optional(),components:o.z.record(o.z.string(),o.z.unknown()).optional(),webhooks:o.z.record(o.z.string(),o.z.unknown()).optional(),"x-scalar-icon":o.z.string().optional().default("interface-content-folder"),"x-scalar-active-environment":o.z.string().optional(),"x-scalar-environments":a.O$.optional(),"x-scalar-secrets":s.optional()}),d=o.z.object({uid:r.B.brand(),securitySchemes:o.z.string().array().default([]),selectedSecuritySchemeUids:r.I,selectedServerUid:o.z.string().brand().optional(),servers:o.z.string().brand().array().default([]),requests:o.z.string().brand().array().default([]),tags:o.z.string().brand().array().default([]),children:o.z.union([o.z.string().brand(),o.z.string().brand()]).array().default([]),documentUrl:o.z.string().optional(),watchMode:o.z.boolean().optional().default(!1),integration:o.z.string().nullable().optional(),watchModeStatus:o.z.enum(["IDLE","WATCHING","ERROR"]).optional().default("IDLE")}),p=u.merge(d)},8007:(e,t,n)=>{"use strict";n.d(t,{S:()=>o});var r=n(3184),a=n(1793);function o(e){if(null===e)return{};if("string"==typeof e){if(""===e.trim())return{};try{return JSON.parse(e)}catch(t){return(0,r.qg)(e,{maxAliasCount:1e4})}}return(0,a.W)(e),e}},8065:(e,t,n)=>{"use strict";n.d(t,{MA:()=>l,XX:()=>s,c3:()=>u});var r=n(2071),a=n(4476);const o=a.z.object({name:a.z.string().optional().default("name"),identifier:a.z.string().optional(),url:a.z.string().optional()}),i=a.z.object({name:a.z.string().optional(),url:a.z.string().optional(),email:a.z.string().email().optional()}),s=a.z.object({title:a.z.string().optional().default("API"),summary:a.z.string().optional(),description:a.z.string().optional(),termsOfService:a.z.string().optional(),contact:i.optional(),license:o.optional(),version:a.z.string().optional().default("1.0")}),l=a.z.object({description:a.z.string().optional(),url:a.z.string().default("")}),c=a.z.object({tagName:a.z.string()}).array(),u=a.z.object({type:a.z.literal("tag").optional().default("tag"),name:a.z.string(),description:a.z.string().optional(),externalDocs:l.optional(),"x-scalar-children":c.default([]).optional(),"x-internal":a.z.boolean().optional(),"x-scalar-ignore":a.z.boolean().optional()}).extend({uid:r.B.brand(),children:a.z.union([a.z.string().brand(),a.z.string().brand()]).array().default([])})},8082:(e,t,n)=>{"use strict";n.d(t,{A:()=>s});var r=n(641),a=n(3526);const o={},i={class:"flex flex-col min-h-0 flex-1 *:border-t-1/2 first:*:border-t-0 md:*:border-t-0 xl:overflow-hidden md:flex-row leading-3"},s=(0,a.A)(o,[["render",function(e,t){return(0,r.uX)(),(0,r.CE)("div",i,[(0,r.RG)(e.$slots,"default")])}]])},8157:(e,t,n)=>{"use strict";n.d(t,{jx:()=>B,Yf:()=>L});var r=n(641),a=n(953),o=n(33),i=Object.defineProperty,s=(e,t,n)=>((e,t,n)=>t in e?i(e,t,{enumerable:!0,configurable:!0,writable:!0,value:n}):e[t]=n)(e,"symbol"!=typeof t?t+"":t,n);!function(e){if(typeof document>"u")return;let t=document.head||document.getElementsByTagName("head")[0],n=document.createElement("style");n.type="text/css",t.appendChild(n),n.styleSheet?n.styleSheet.cssText=e:n.appendChild(document.createTextNode(e))}(":where([data-sonner-toaster][dir=ltr]),:where(html[dir=ltr]){--toast-icon-margin-start:-3px;--toast-icon-margin-end:4px;--toast-svg-margin-start:-1px;--toast-svg-margin-end:0px;--toast-button-margin-start:auto;--toast-button-margin-end:0;--toast-close-button-start:0;--toast-close-button-end:unset;--toast-close-button-transform:translate(-35%, -35%)}:where([data-sonner-toaster][dir=rtl]),:where(html[dir=rtl]){--toast-icon-margin-start:4px;--toast-icon-margin-end:-3px;--toast-svg-margin-start:0px;--toast-svg-margin-end:-1px;--toast-button-margin-start:0;--toast-button-margin-end:auto;--toast-close-button-start:unset;--toast-close-button-end:0;--toast-close-button-transform:translate(35%, -35%)}:where([data-sonner-toaster]){position:fixed;width:var(--width);font-family:ui-sans-serif,system-ui,-apple-system,BlinkMacSystemFont,Segoe UI,Roboto,Helvetica Neue,Arial,Noto Sans,sans-serif,Apple Color Emoji,Segoe UI Emoji,Segoe UI Symbol,Noto Color Emoji;--gray1:hsl(0, 0%, 99%);--gray2:hsl(0, 0%, 97.3%);--gray3:hsl(0, 0%, 95.1%);--gray4:hsl(0, 0%, 93%);--gray5:hsl(0, 0%, 90.9%);--gray6:hsl(0, 0%, 88.7%);--gray7:hsl(0, 0%, 85.8%);--gray8:hsl(0, 0%, 78%);--gray9:hsl(0, 0%, 56.1%);--gray10:hsl(0, 0%, 52.3%);--gray11:hsl(0, 0%, 43.5%);--gray12:hsl(0, 0%, 9%);--border-radius:8px;box-sizing:border-box;padding:0;margin:0;list-style:none;outline:0;z-index:999999999;transition:transform .4s ease}:where([data-sonner-toaster][data-lifted=true]){transform:translateY(-10px)}@media (hover:none) and (pointer:coarse){:where([data-sonner-toaster][data-lifted=true]){transform:none}}:where([data-sonner-toaster][data-x-position=right]){right:max(var(--offset),env(safe-area-inset-right))}:where([data-sonner-toaster][data-x-position=left]){left:max(var(--offset),env(safe-area-inset-left))}:where([data-sonner-toaster][data-x-position=center]){left:50%;transform:translateX(-50%)}:where([data-sonner-toaster][data-y-position=top]){top:max(var(--offset),env(safe-area-inset-top))}:where([data-sonner-toaster][data-y-position=bottom]){bottom:max(var(--offset),env(safe-area-inset-bottom))}:where([data-sonner-toast]){--y:translateY(100%);--lift-amount:calc(var(--lift) * var(--gap));z-index:var(--z-index);position:absolute;opacity:0;transform:var(--y);filter:blur(0);touch-action:none;transition:transform .4s,opacity .4s,height .4s,box-shadow .2s;box-sizing:border-box;outline:0;overflow-wrap:anywhere}:where([data-sonner-toast][data-styled=true]){padding:16px;background:var(--normal-bg);border:1px solid var(--normal-border);color:var(--normal-text);border-radius:var(--border-radius);box-shadow:0 4px 12px rgba(0,0,0,.1);width:var(--width);font-size:13px;display:flex;align-items:center;gap:6px}:where([data-sonner-toast]:focus-visible){box-shadow:0 4px 12px rgba(0,0,0,.1),0 0 0 2px rgba(0,0,0,.2)}:where([data-sonner-toast][data-y-position=top]){top:0;--y:translateY(-100%);--lift:1;--lift-amount:calc(1 * var(--gap))}:where([data-sonner-toast][data-y-position=bottom]){bottom:0;--y:translateY(100%);--lift:-1;--lift-amount:calc(var(--lift) * var(--gap))}:where([data-sonner-toast]) :where([data-description]){font-weight:400;line-height:1.4;color:inherit}:where([data-sonner-toast]) :where([data-title]){font-weight:500;line-height:1.5;color:inherit}:where([data-sonner-toast]) :where([data-icon]){display:flex;height:16px;width:16px;position:relative;justify-content:flex-start;align-items:center;flex-shrink:0;margin-left:var(--toast-icon-margin-start);margin-right:var(--toast-icon-margin-end)}:where([data-sonner-toast][data-promise=true]) :where([data-icon])>svg{opacity:0;transform:scale(.8);transform-origin:center;animation:sonner-fade-in .3s ease forwards}:where([data-sonner-toast]) :where([data-icon])>*{flex-shrink:0}:where([data-sonner-toast]) :where([data-icon]) svg{margin-left:var(--toast-svg-margin-start);margin-right:var(--toast-svg-margin-end)}:where([data-sonner-toast]) :where([data-content]){display:flex;flex-direction:column;gap:2px}[data-sonner-toast][data-styled=true] [data-button]{border-radius:4px;padding-left:8px;padding-right:8px;height:24px;font-size:12px;color:var(--normal-bg);background:var(--normal-text);margin-left:var(--toast-button-margin-start);margin-right:var(--toast-button-margin-end);border:none;cursor:pointer;outline:0;display:flex;align-items:center;flex-shrink:0;transition:opacity .4s,box-shadow .2s}:where([data-sonner-toast]) :where([data-button]):focus-visible{box-shadow:0 0 0 2px rgba(0,0,0,.4)}:where([data-sonner-toast]) :where([data-button]):first-of-type{margin-left:var(--toast-button-margin-start);margin-right:var(--toast-button-margin-end)}:where([data-sonner-toast]) :where([data-cancel]){color:var(--normal-text);background:rgba(0,0,0,.08)}:where([data-sonner-toast][data-theme=dark]) :where([data-cancel]){background:rgba(255,255,255,.3)}[data-sonner-toast] [data-close-button]{position:absolute;left:var(--toast-close-button-start);right:var(--toast-close-button-end);top:0;height:20px;width:20px;display:flex;justify-content:center;align-items:center;padding:0;color:var(--gray12);border:1px solid var(--gray4);transform:var(--toast-close-button-transform);border-radius:50%;cursor:pointer;z-index:1;transition:opacity .1s,background .2s,border-color .2s}[data-sonner-toast] [data-close-button]{background:var(--gray1)}:where([data-sonner-toast]) :where([data-close-button]):focus-visible{box-shadow:0 4px 12px rgba(0,0,0,.1),0 0 0 2px rgba(0,0,0,.2)}:where([data-sonner-toast]) :where([data-disabled=true]){cursor:not-allowed}[data-sonner-toast]:hover [data-close-button]:hover{background:var(--gray2);border-color:var(--gray5)}:where([data-sonner-toast][data-swiping=true])::before{content:'';position:absolute;left:0;right:0;height:100%;z-index:-1}:where([data-sonner-toast][data-y-position=top][data-swiping=true])::before{bottom:50%;transform:scaleY(3) translateY(50%)}:where([data-sonner-toast][data-y-position=bottom][data-swiping=true])::before{top:50%;transform:scaleY(3) translateY(-50%)}:where([data-sonner-toast][data-swiping=false][data-removed=true])::before{content:'';position:absolute;inset:0;transform:scaleY(2)}:where([data-sonner-toast])::after{content:'';position:absolute;left:0;height:calc(var(--gap) + 1px);bottom:100%;width:100%}:where([data-sonner-toast][data-mounted=true]){--y:translateY(0);opacity:1}:where([data-sonner-toast][data-expanded=false][data-front=false]){--scale:var(--toasts-before) * 0.05 + 1;--y:translateY(calc(var(--lift-amount) * var(--toasts-before))) scale(calc(-1 * var(--scale)));height:var(--front-toast-height)}:where([data-sonner-toast])>*{transition:opacity .4s}:where([data-sonner-toast][data-expanded=false][data-front=false][data-styled=true])>*{opacity:0}:where([data-sonner-toast][data-visible=false]){opacity:0;pointer-events:none}:where([data-sonner-toast][data-mounted=true][data-expanded=true]){--y:translateY(calc(var(--lift) * var(--offset)));height:var(--initial-height)}:where([data-sonner-toast][data-removed=true][data-front=true][data-swipe-out=false]){--y:translateY(calc(var(--lift) * -100%));opacity:0}:where([data-sonner-toast][data-removed=true][data-front=false][data-swipe-out=false][data-expanded=true]){--y:translateY(calc(var(--lift) * var(--offset) + var(--lift) * -100%));opacity:0}:where([data-sonner-toast][data-removed=true][data-front=false][data-swipe-out=false][data-expanded=false]){--y:translateY(40%);opacity:0;transition:transform .5s,opacity .2s}:where([data-sonner-toast][data-removed=true][data-front=false])::before{height:calc(var(--initial-height) + 20%)}[data-sonner-toast][data-swiping=true]{transform:var(--y) translateY(var(--swipe-amount,0));transition:none}[data-sonner-toast][data-swiped=true]{user-select:none}[data-sonner-toast][data-swipe-out=true][data-y-position=bottom],[data-sonner-toast][data-swipe-out=true][data-y-position=top]{animation:swipe-out .2s ease-out forwards}@keyframes swipe-out{from{transform:translateY(calc(var(--lift) * var(--offset) + var(--swipe-amount)));opacity:1}to{transform:translateY(calc(var(--lift) * var(--offset) + var(--swipe-amount) + var(--lift) * -100%));opacity:0}}@media (max-width:600px){[data-sonner-toaster]{position:fixed;--mobile-offset:16px;right:var(--mobile-offset);left:var(--mobile-offset);width:100%}[data-sonner-toaster][dir=rtl]{left:calc(var(--mobile-offset) * -1)}[data-sonner-toaster] [data-sonner-toast]{left:0;right:0;width:calc(100% - var(--mobile-offset) * 2)}[data-sonner-toaster][data-x-position=left]{left:var(--mobile-offset)}[data-sonner-toaster][data-y-position=bottom]{bottom:20px}[data-sonner-toaster][data-y-position=top]{top:20px}[data-sonner-toaster][data-x-position=center]{left:var(--mobile-offset);right:var(--mobile-offset);transform:none}}[data-sonner-toaster][data-theme=light]{--normal-bg:#fff;--normal-border:var(--gray4);--normal-text:var(--gray12);--success-bg:hsl(143, 85%, 96%);--success-border:hsl(145, 92%, 91%);--success-text:hsl(140, 100%, 27%);--info-bg:hsl(208, 100%, 97%);--info-border:hsl(221, 91%, 91%);--info-text:hsl(210, 92%, 45%);--warning-bg:hsl(49, 100%, 97%);--warning-border:hsl(49, 91%, 91%);--warning-text:hsl(31, 92%, 45%);--error-bg:hsl(359, 100%, 97%);--error-border:hsl(359, 100%, 94%);--error-text:hsl(360, 100%, 45%)}[data-sonner-toaster][data-theme=light] [data-sonner-toast][data-invert=true]{--normal-bg:#000;--normal-border:hsl(0, 0%, 20%);--normal-text:var(--gray1)}[data-sonner-toaster][data-theme=dark] [data-sonner-toast][data-invert=true]{--normal-bg:#fff;--normal-border:var(--gray3);--normal-text:var(--gray12)}[data-sonner-toaster][data-theme=dark]{--normal-bg:#000;--normal-border:hsl(0, 0%, 20%);--normal-text:var(--gray1);--success-bg:hsl(150, 100%, 6%);--success-border:hsl(147, 100%, 12%);--success-text:hsl(150, 86%, 65%);--info-bg:hsl(215, 100%, 6%);--info-border:hsl(223, 100%, 12%);--info-text:hsl(216, 87%, 65%);--warning-bg:hsl(64, 100%, 6%);--warning-border:hsl(60, 100%, 12%);--warning-text:hsl(46, 87%, 65%);--error-bg:hsl(358, 76%, 10%);--error-border:hsl(357, 89%, 16%);--error-text:hsl(358, 100%, 81%)}[data-rich-colors=true][data-sonner-toast][data-type=success]{background:var(--success-bg);border-color:var(--success-border);color:var(--success-text)}[data-rich-colors=true][data-sonner-toast][data-type=success] [data-close-button]{background:var(--success-bg);border-color:var(--success-border);color:var(--success-text)}[data-rich-colors=true][data-sonner-toast][data-type=info]{background:var(--info-bg);border-color:var(--info-border);color:var(--info-text)}[data-rich-colors=true][data-sonner-toast][data-type=info] [data-close-button]{background:var(--info-bg);border-color:var(--info-border);color:var(--info-text)}[data-rich-colors=true][data-sonner-toast][data-type=warning]{background:var(--warning-bg);border-color:var(--warning-border);color:var(--warning-text)}[data-rich-colors=true][data-sonner-toast][data-type=warning] [data-close-button]{background:var(--warning-bg);border-color:var(--warning-border);color:var(--warning-text)}[data-rich-colors=true][data-sonner-toast][data-type=error]{background:var(--error-bg);border-color:var(--error-border);color:var(--error-text)}[data-rich-colors=true][data-sonner-toast][data-type=error] [data-close-button]{background:var(--error-bg);border-color:var(--error-border);color:var(--error-text)}.sonner-loading-wrapper{--size:16px;height:var(--size);width:var(--size);position:absolute;inset:0;z-index:10}.sonner-loading-wrapper[data-visible=false]{transform-origin:center;animation:sonner-fade-out .2s ease forwards}.sonner-spinner{position:relative;top:50%;left:50%;height:var(--size);width:var(--size)}.sonner-loading-bar{animation:sonner-spin 1.2s linear infinite;background:var(--gray11);border-radius:6px;height:8%;left:-10%;position:absolute;top:-3.9%;width:24%}.sonner-loading-bar:first-child{animation-delay:-1.2s;transform:rotate(.0001deg) translate(146%)}.sonner-loading-bar:nth-child(2){animation-delay:-1.1s;transform:rotate(30deg) translate(146%)}.sonner-loading-bar:nth-child(3){animation-delay:-1s;transform:rotate(60deg) translate(146%)}.sonner-loading-bar:nth-child(4){animation-delay:-.9s;transform:rotate(90deg) translate(146%)}.sonner-loading-bar:nth-child(5){animation-delay:-.8s;transform:rotate(120deg) translate(146%)}.sonner-loading-bar:nth-child(6){animation-delay:-.7s;transform:rotate(150deg) translate(146%)}.sonner-loading-bar:nth-child(7){animation-delay:-.6s;transform:rotate(180deg) translate(146%)}.sonner-loading-bar:nth-child(8){animation-delay:-.5s;transform:rotate(210deg) translate(146%)}.sonner-loading-bar:nth-child(9){animation-delay:-.4s;transform:rotate(240deg) translate(146%)}.sonner-loading-bar:nth-child(10){animation-delay:-.3s;transform:rotate(270deg) translate(146%)}.sonner-loading-bar:nth-child(11){animation-delay:-.2s;transform:rotate(300deg) translate(146%)}.sonner-loading-bar:nth-child(12){animation-delay:-.1s;transform:rotate(330deg) translate(146%)}@keyframes sonner-fade-in{0%{opacity:0;transform:scale(.8)}100%{opacity:1;transform:scale(1)}}@keyframes sonner-fade-out{0%{opacity:1;transform:scale(1)}100%{opacity:0;transform:scale(.8)}}@keyframes sonner-spin{0%{opacity:1}100%{opacity:.15}}@media (prefers-reduced-motion){.sonner-loading-bar,[data-sonner-toast],[data-sonner-toast]>*{transition:none!important;animation:none!important}}.sonner-loader{position:absolute;top:50%;left:50%;transform:translate(-50%,-50%);transform-origin:center;transition:opacity .2s,transform .2s}.sonner-loader[data-visible=false]{opacity:0;transform:scale(.8) translate(-50%,-50%)}");let l=0;const c=new class{constructor(){s(this,"subscribers"),s(this,"toasts"),s(this,"subscribe",(e=>(this.subscribers.push(e),()=>{const t=this.subscribers.indexOf(e);this.subscribers.splice(t,1)}))),s(this,"publish",(e=>{this.subscribers.forEach((t=>t(e)))})),s(this,"addToast",(e=>{this.publish(e),this.toasts=[...this.toasts,e]})),s(this,"create",(e=>{var t;const{message:n,...r}=e,a="number"==typeof e.id||e.id&&(null==(t=e.id)?void 0:t.length)>0?e.id:l++,o=this.toasts.find((e=>e.id===a)),i=void 0===e.dismissible||e.dismissible;return o?this.toasts=this.toasts.map((t=>t.id===a?(this.publish({...t,...e,id:a,title:n}),{...t,...e,id:a,dismissible:i,title:n}):t)):this.addToast({title:n,...r,dismissible:i,id:a}),a})),s(this,"dismiss",(e=>(e||this.toasts.forEach((e=>{this.subscribers.forEach((t=>t({id:e.id,dismiss:!0})))})),this.subscribers.forEach((t=>t({id:e,dismiss:!0}))),e))),s(this,"message",((e,t)=>this.create({...t,message:e,type:"default"}))),s(this,"error",((e,t)=>this.create({...t,type:"error",message:e}))),s(this,"success",((e,t)=>this.create({...t,type:"success",message:e}))),s(this,"info",((e,t)=>this.create({...t,type:"info",message:e}))),s(this,"warning",((e,t)=>this.create({...t,type:"warning",message:e}))),s(this,"loading",((e,t)=>this.create({...t,type:"loading",message:e}))),s(this,"promise",((e,t)=>{if(!t)return;let n;void 0!==t.loading&&(n=this.create({...t,promise:e,type:"loading",message:t.loading,description:"function"!=typeof t.description?t.description:void 0}));const r=e instanceof Promise?e:e();let a,o=void 0!==n;const i=r.then((async e=>{if(a=["resolve",e],u(e)&&!e.ok){o=!1;const r="function"==typeof t.error?await t.error(`HTTP error! status: ${e.status}`):t.error,a="function"==typeof t.description?await t.description(`HTTP error! status: ${e.status}`):t.description;this.create({id:n,type:"error",message:r,description:a})}else if(void 0!==t.success){o=!1;const r="function"==typeof t.success?await t.success(e):t.success,a="function"==typeof t.description?await t.description(e):t.description;this.create({id:n,type:"success",message:r,description:a})}})).catch((async e=>{if(a=["reject",e],void 0!==t.error){o=!1;const r="function"==typeof t.error?await t.error(e):t.error,a="function"==typeof t.description?await t.description(e):t.description;this.create({id:n,type:"error",message:r,description:a})}})).finally((()=>{var e;o&&(this.dismiss(n),n=void 0),null==(e=t.finally)||e.call(t)})),s=()=>new Promise(((e,t)=>i.then((()=>"reject"===a[0]?t(a[1]):e(a[1]))).catch(t)));return"string"!=typeof n&&"number"!=typeof n?{unwrap:s}:Object.assign(n,{unwrap:s})})),s(this,"custom",((e,t)=>{const n=(null==t?void 0:t.id)||l++;return this.publish({component:e,id:n,...t}),n})),this.subscribers=[],this.toasts=[]}},u=e=>e&&"object"==typeof e&&"ok"in e&&"boolean"==typeof e.ok&&"status"in e&&"number"==typeof e.status,d=Object.assign((function(e,t){const n=(null==t?void 0:t.id)||l++;return c.create({message:e,id:n,type:"default",...t}),n}),{success:c.success,info:c.info,warning:c.warning,error:c.error,custom:c.custom,message:c.message,promise:c.promise,dismiss:c.dismiss,loading:c.loading},{getHistory:()=>c.toasts});function p(e){return void 0!==e.label}const h=["aria-live","data-rich-colors","data-styled","data-mounted","data-promise","data-removed","data-visible","data-y-position","data-x-position","data-index","data-front","data-swiping","data-dismissible","data-type","data-invert","data-swipe-out","data-expanded"],f=["aria-label","data-disabled"],m=(0,r.pM)({__name:"Toast",props:{toast:{},toasts:{},index:{},expanded:{type:Boolean},invert:{type:Boolean},heights:{},gap:{},position:{},visibleToasts:{},expandByDefault:{type:Boolean},closeButton:{type:Boolean},interacting:{type:Boolean},style:{},cancelButtonStyle:{},actionButtonStyle:{},duration:{},class:{},unstyled:{type:Boolean},descriptionClass:{},loadingIcon:{},classes:{},icons:{},closeButtonAriaLabel:{},pauseWhenPageIsHidden:{type:Boolean},cn:{type:Function},defaultRichColors:{type:Boolean}},emits:["update:heights","removeToast"],setup(e,{emit:t}){const n=e,i=t,s=(0,a.KR)(!1),l=(0,a.KR)(!1),c=(0,a.KR)(!1),u=(0,a.KR)(!1),d=(0,a.KR)(!1),m=(0,a.KR)(0),g=(0,a.KR)(0),v=(0,a.KR)(n.toast.duration||n.duration||4e3),b=(0,a.KR)(null),y=(0,a.KR)(null),w=(0,r.EW)((()=>0===n.index)),O=(0,r.EW)((()=>n.index+1<=n.visibleToasts)),k=(0,r.EW)((()=>n.toast.type)),x=(0,r.EW)((()=>!1!==n.toast.dismissible)),E=(0,r.EW)((()=>n.toast.class||"")),S=(0,r.EW)((()=>n.descriptionClass||"")),_=n.toast.style||{},C=(0,r.EW)((()=>n.heights.findIndex((e=>e.toastId===n.toast.id))||0)),T=(0,r.EW)((()=>n.toast.closeButton??n.closeButton));(0,r.EW)((()=>n.toast.duration||n.duration||4e3));const A=(0,a.KR)(0),R=(0,a.KR)(0),P=(0,a.KR)(null),D=(0,r.EW)((()=>n.position.split("-"))),M=(0,r.EW)((()=>D.value[0])),$=(0,r.EW)((()=>D.value[1])),N=(0,r.EW)((()=>"string"!=typeof n.toast.title)),I=(0,r.EW)((()=>"string"!=typeof n.toast.description)),L=(0,r.EW)((()=>n.heights.reduce(((e,t,n)=>n>=C.value?e:e+t.height),0))),B=function(){const e=(0,a.KR)(!1);return(0,r.nT)((()=>{const t=()=>{e.value=document.hidden};return document.addEventListener("visibilitychange",t),()=>window.removeEventListener("visibilitychange",t)})),{isDocumentHidden:e}}(),F=(0,r.EW)((()=>n.toast.invert||n.invert)),Q=(0,r.EW)((()=>"loading"===k.value)),U=(0,r.EW)((()=>C.value*n.gap+L.value||0));function z(){l.value=!0,m.value=U.value;const e=n.heights.filter((e=>e.toastId!==n.toast.id));i("update:heights",e),setTimeout((()=>{i("removeToast",n.toast)}),200)}function j(){var e,t;if(Q.value||!x.value)return{};z(),null==(t=(e=n.toast).onDismiss)||t.call(e,n.toast)}function W(e){Q.value||!x.value||(b.value=new Date,m.value=U.value,e.target.setPointerCapture(e.pointerId),"BUTTON"!==e.target.tagName&&(c.value=!0,P.value={x:e.clientX,y:e.clientY}))}function X(){var e,t,r,a,o;if(u.value||!x)return;P.value=null;const i=Number((null==(e=y.value)?void 0:e.style.getPropertyValue("--swipe-amount").replace("px",""))||0),s=(new Date).getTime()-(null==(t=b.value)?void 0:t.getTime()),l=Math.abs(i)/s;if(Math.abs(i)>=20||l>.11)return m.value=U.value,null==(a=(r=n.toast).onDismiss)||a.call(r,n.toast),z(),u.value=!0,void(d.value=!1);null==(o=y.value)||o.style.setProperty("--swipe-amount","0px"),c.value=!1}function q(e){var t,n;if(!P.value||!x.value)return;const r=e.clientY-P.value.y,a=(null==(t=window.getSelection())?void 0:t.toString().length)>0,o="top"===M.value?Math.min(0,r):Math.max(0,r);Math.abs(o)>0&&(d.value=!0),!a&&(null==(n=y.value)||n.style.setProperty("--swipe-amount",`${o}px`))}return(0,r.sV)((()=>{if(!s.value)return;const e=y.value,t=null==e?void 0:e.style.height;e.style.height="auto";const r=e.getBoundingClientRect().height;let a;e.style.height=t,g.value=r,a=n.heights.find((e=>e.toastId===n.toast.id))?n.heights.map((e=>e.toastId===n.toast.id?{...e,height:r}:e)):[{toastId:n.toast.id,height:r,position:n.toast.position},...n.heights],i("update:heights",a)})),(0,r.nT)((e=>{if(n.toast.promise&&"loading"===k.value||n.toast.duration===1/0||"loading"===n.toast.type)return;let t;n.expanded||n.interacting||n.pauseWhenPageIsHidden&&B?(()=>{if(R.value<A.value){const e=(new Date).getTime()-A.value;v.value=v.value-e}R.value=(new Date).getTime()})():v.value!==1/0&&(A.value=(new Date).getTime(),t=setTimeout((()=>{var e,t;null==(t=(e=n.toast).onAutoClose)||t.call(e,n.toast),z()}),v.value)),e((()=>{clearTimeout(t)}))})),(0,r.wB)((()=>n.toast.delete),(()=>{n.toast.delete&&z()}),{deep:!0}),(0,r.sV)((()=>{if(s.value=!0,y.value){const e=y.value.getBoundingClientRect().height;g.value=e;const t=[{toastId:n.toast.id,height:e,position:n.toast.position},...n.heights];i("update:heights",t)}})),(0,r.xo)((()=>{if(y.value){const e=n.heights.filter((e=>e.toastId!==n.toast.id));i("update:heights",e)}})),(e,t)=>{var i,d,v,b,C,A,R,P,D,L,B,V,H,Z,G,K,Y,J,ee,te,ne,re,ae,oe,ie,se,le;return(0,r.uX)(),(0,r.CE)("li",{ref_key:"toastRef",ref:y,"aria-live":e.toast.important?"assertive":"polite","aria-atomic":"true",role:"status",tabindex:"0","data-sonner-toast":"true",class:(0,o.C4)(e.cn(n.class,E.value,null==(i=e.classes)?void 0:i.toast,null==(d=e.toast.classes)?void 0:d.toast,null==(v=e.classes)?void 0:v[k.value],null==(C=null==(b=e.toast)?void 0:b.classes)?void 0:C[k.value])),"data-rich-colors":e.toast.richColors??e.defaultRichColors,"data-styled":!(e.toast.component||null!=(A=e.toast)&&A.unstyled||e.unstyled),"data-mounted":s.value,"data-promise":!!e.toast.promise,"data-removed":l.value,"data-visible":O.value,"data-y-position":M.value,"data-x-position":$.value,"data-index":e.index,"data-front":w.value,"data-swiping":c.value,"data-dismissible":x.value,"data-type":k.value,"data-invert":F.value,"data-swipe-out":u.value,"data-expanded":!!(e.expanded||e.expandByDefault&&s.value),style:(0,o.Tr)({"--index":e.index,"--toasts-before":e.index,"--z-index":e.toasts.length-e.index,"--offset":`${l.value?m.value:U.value}px`,"--initial-height":e.expandByDefault?"auto":`${g.value}px`,...e.style,...(0,a.R1)(_)}),onPointerdown:W,onPointerup:X,onPointermove:q},[T.value&&!e.toast.component?((0,r.uX)(),(0,r.CE)("button",{key:0,"aria-label":e.closeButtonAriaLabel||"Close toast","data-disabled":Q.value,"data-close-button":"true",class:(0,o.C4)(e.cn(null==(R=e.classes)?void 0:R.closeButton,null==(D=null==(P=e.toast)?void 0:P.classes)?void 0:D.closeButton)),onClick:j},[null!=(L=e.icons)&&L.close?((0,r.uX)(),(0,r.Wv)((0,r.$y)(null==(B=e.icons)?void 0:B.close),{key:0})):(0,r.RG)(e.$slots,"close-icon",{key:1})],10,f)):(0,r.Q3)("",!0),e.toast.component?((0,r.uX)(),(0,r.Wv)((0,r.$y)(e.toast.component),(0,r.v6)({key:1},e.toast.componentProps,{onCloseToast:j}),null,16)):((0,r.uX)(),(0,r.CE)(r.FK,{key:2},["default"!==k.value||e.toast.icon||e.toast.promise?((0,r.uX)(),(0,r.CE)("div",{key:0,"data-icon":"",class:(0,o.C4)(e.cn(null==(V=e.classes)?void 0:V.icon,null==(Z=null==(H=e.toast)?void 0:H.classes)?void 0:Z.icon))},[!e.toast.promise&&"loading"!==k.value||e.toast.icon?(0,r.Q3)("",!0):(0,r.RG)(e.$slots,"loading-icon",{key:0}),e.toast.icon?((0,r.uX)(),(0,r.Wv)((0,r.$y)(e.toast.icon),{key:1})):((0,r.uX)(),(0,r.CE)(r.FK,{key:2},["success"===k.value?(0,r.RG)(e.$slots,"success-icon",{key:0}):"error"===k.value?(0,r.RG)(e.$slots,"error-icon",{key:1}):"warning"===k.value?(0,r.RG)(e.$slots,"warning-icon",{key:2}):"info"===k.value?(0,r.RG)(e.$slots,"info-icon",{key:3}):(0,r.Q3)("",!0)],64))],2)):(0,r.Q3)("",!0),(0,r.Lk)("div",{"data-content":"",class:(0,o.C4)(e.cn(null==(G=e.classes)?void 0:G.content,null==(Y=null==(K=e.toast)?void 0:K.classes)?void 0:Y.content))},[(0,r.Lk)("div",{"data-title":"",class:(0,o.C4)(e.cn(null==(J=e.classes)?void 0:J.title,null==(ee=e.toast.classes)?void 0:ee.title))},[N.value?((0,r.uX)(),(0,r.Wv)((0,r.$y)(e.toast.title),(0,o._B)((0,r.v6)({key:0},e.toast.componentProps)),null,16)):((0,r.uX)(),(0,r.CE)(r.FK,{key:1},[(0,r.eW)((0,o.v_)(e.toast.title),1)],64))],2),e.toast.description?((0,r.uX)(),(0,r.CE)("div",{key:0,"data-description":"",class:(0,o.C4)(e.cn(e.descriptionClass,S.value,null==(te=e.classes)?void 0:te.description,null==(ne=e.toast.classes)?void 0:ne.description))},[I.value?((0,r.uX)(),(0,r.Wv)((0,r.$y)(e.toast.description),(0,o._B)((0,r.v6)({key:0},e.toast.componentProps)),null,16)):((0,r.uX)(),(0,r.CE)(r.FK,{key:1},[(0,r.eW)((0,o.v_)(e.toast.description),1)],64))],2)):(0,r.Q3)("",!0)],2),e.toast.cancel?((0,r.uX)(),(0,r.CE)("button",{key:1,style:(0,o.Tr)(e.toast.cancelButtonStyle||e.cancelButtonStyle),class:(0,o.C4)(e.cn(null==(re=e.classes)?void 0:re.cancelButton,null==(ae=e.toast.classes)?void 0:ae.cancelButton)),"data-button":"","data-cancel":"",onClick:t[0]||(t[0]=t=>{var n,r;(0,a.R1)(p)(e.toast.cancel)&&x.value&&(null==(r=(n=e.toast.cancel).onClick)||r.call(n,t),z())})},(0,o.v_)((0,a.R1)(p)(e.toast.cancel)?null==(oe=e.toast.cancel)?void 0:oe.label:e.toast.cancel),7)):(0,r.Q3)("",!0),e.toast.action?((0,r.uX)(),(0,r.CE)("button",{key:2,style:(0,o.Tr)(e.toast.actionButtonStyle||e.actionButtonStyle),class:(0,o.C4)(e.cn(null==(ie=e.classes)?void 0:ie.actionButton,null==(se=e.toast.classes)?void 0:se.actionButton)),"data-button":"","data-action":"",onClick:t[1]||(t[1]=t=>{var n,r;(0,a.R1)(p)(e.toast.action)&&(t.defaultPrevented||(null==(r=(n=e.toast.action).onClick)||r.call(n,t),!t.defaultPrevented&&z()))})},(0,o.v_)((0,a.R1)(p)(e.toast.action)?null==(le=e.toast.action)?void 0:le.label:e.toast.action),7)):(0,r.Q3)("",!0)],64))],46,h)}}}),g=(e,t)=>{const n=e.__vccOpts||e;for(const[e,r]of t)n[e]=r;return n},v={xmlns:"http://www.w3.org/2000/svg",width:"12",height:"12",viewBox:"0 0 24 24",fill:"none",stroke:"currentColor","stoke-width":"1.5","stroke-linecap":"round","stroke-linejoin":"round"},b=g({},[["render",function(e,t){return(0,r.uX)(),(0,r.CE)("svg",v,t[0]||(t[0]=[(0,r.Lk)("line",{x1:"18",y1:"6",x2:"6",y2:"18"},null,-1),(0,r.Lk)("line",{x1:"6",y1:"6",x2:"18",y2:"18"},null,-1)]))}]]),y=["data-visible"],w={class:"sonner-spinner"},O=(0,r.pM)({__name:"Loader",props:{visible:{type:Boolean}},setup(e){const t=Array(12).fill(0);return(e,n)=>((0,r.uX)(),(0,r.CE)("div",{class:"sonner-loading-wrapper","data-visible":e.visible},[(0,r.Lk)("div",w,[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)((0,a.R1)(t),(e=>((0,r.uX)(),(0,r.CE)("div",{key:`spinner-bar-${e}`,class:"sonner-loading-bar"})))),128))])],8,y))}}),k={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 20 20",fill:"currentColor",height:"20",width:"20"},x=g({},[["render",function(e,t){return(0,r.uX)(),(0,r.CE)("svg",k,t[0]||(t[0]=[(0,r.Lk)("path",{"fill-rule":"evenodd",d:"M10 18a8 8 0 100-16 8 8 0 000 16zm3.857-9.809a.75.75 0 00-1.214-.882l-3.483 4.79-1.88-1.88a.75.75 0 10-1.06 1.061l2.5 2.5a.75.75 0 001.137-.089l4-5.5z","clip-rule":"evenodd"},null,-1)]))}]]),E={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 20 20",fill:"currentColor",height:"20",width:"20"},S=g({},[["render",function(e,t){return(0,r.uX)(),(0,r.CE)("svg",E,t[0]||(t[0]=[(0,r.Lk)("path",{"fill-rule":"evenodd",d:"M18 10a8 8 0 11-16 0 8 8 0 0116 0zm-7-4a1 1 0 11-2 0 1 1 0 012 0zM9 9a.75.75 0 000 1.5h.253a.25.25 0 01.244.304l-.459 2.066A1.75 1.75 0 0010.747 15H11a.75.75 0 000-1.5h-.253a.25.25 0 01-.244-.304l.459-2.066A1.75 1.75 0 009.253 9H9z","clip-rule":"evenodd"},null,-1)]))}]]),_={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 24 24",fill:"currentColor",height:"20",width:"20"},C=g({},[["render",function(e,t){return(0,r.uX)(),(0,r.CE)("svg",_,t[0]||(t[0]=[(0,r.Lk)("path",{"fill-rule":"evenodd",d:"M9.401 3.003c1.155-2 4.043-2 5.197 0l7.355 12.748c1.154 2-.29 4.5-2.599 4.5H4.645c-2.309 0-3.752-2.5-2.598-4.5L9.4 3.003zM12 8.25a.75.75 0 01.75.75v3.75a.75.75 0 01-1.5 0V9a.75.75 0 01.75-.75zm0 8.25a.75.75 0 100-1.5.75.75 0 000 1.5z","clip-rule":"evenodd"},null,-1)]))}]]),T={xmlns:"http://www.w3.org/2000/svg",viewBox:"0 0 20 20",fill:"currentColor",height:"20",width:"20"},A=g({},[["render",function(e,t){return(0,r.uX)(),(0,r.CE)("svg",T,t[0]||(t[0]=[(0,r.Lk)("path",{"fill-rule":"evenodd",d:"M18 10a8 8 0 11-16 0 8 8 0 0116 0zm-8-5a.75.75 0 01.75.75v4.5a.75.75 0 01-1.5 0v-4.5A.75.75 0 0110 5zm0 10a1 1 0 100-2 1 1 0 000 2z","clip-rule":"evenodd"},null,-1)]))}]]),R=["aria-label"],P=["dir","data-theme","data-rich-colors","data-y-position","data-x-position","data-lifted"],D="32px",M=typeof window<"u"&&typeof document<"u",$=(0,r.pM)({name:"Toaster",inheritAttrs:!1,__name:"Toaster",props:{invert:{type:Boolean,default:!1},theme:{default:"light"},position:{default:"bottom-right"},hotkey:{default:()=>["altKey","KeyT"]},richColors:{type:Boolean,default:!1},expand:{type:Boolean,default:!1},duration:{},gap:{default:14},visibleToasts:{default:3},closeButton:{type:Boolean,default:!1},toastOptions:{default:()=>({})},class:{default:""},style:{default:()=>({})},offset:{default:D},dir:{default:"auto"},icons:{},containerAriaLabel:{default:"Notifications"},pauseWhenPageIsHidden:{type:Boolean,default:!1},cn:{type:Function,default:function(...e){return e.filter(Boolean).join(" ")}}},setup(e){const t=e;function n(){if(typeof window>"u"||typeof document>"u")return"ltr";const e=document.documentElement.getAttribute("dir");return"auto"!==e&&e?e:window.getComputedStyle(document.documentElement).direction}const i=(0,r.OA)(),s=(0,a.KR)([]),l=(0,r.EW)((()=>(e,t)=>s.value.filter((n=>!n.position&&0===t||n.position===e)))),u=(0,r.EW)((()=>{const e=s.value.filter((e=>e.position)).map((e=>e.position));return e.length>0?Array.from(new Set([t.position].concat(e))):[t.position]})),d=(0,a.KR)([]),p=(0,a.KR)(!1),h=(0,a.KR)(!1),f=(0,a.KR)("system"!==t.theme?t.theme:typeof window<"u"&&window.matchMedia&&window.matchMedia("(prefers-color-scheme: dark)").matches?"dark":"light"),g=(0,a.KR)(null),v=(0,a.KR)(null),y=(0,a.KR)(!1),w=t.hotkey.join("+").replace(/Key/g,"").replace(/Digit/g,"");function k(e){var t;null!=(t=s.value.find((t=>t.id===e.id)))&&t.delete||c.dismiss(e.id),s.value=s.value.filter((({id:t})=>t!==e.id))}function E(e){var t,n;y.value&&(null==(n=null==(t=e.currentTarget)?void 0:t.contains)||!n.call(t,e.relatedTarget))&&(y.value=!1,v.value&&(v.value.focus({preventScroll:!0}),v.value=null))}function _(e){e.target instanceof HTMLElement&&"false"===e.target.dataset.dismissible||y.value||(y.value=!0,v.value=e.relatedTarget)}function T(e){e.target&&e.target instanceof HTMLElement&&"false"===e.target.dataset.dismissible||(h.value=!0)}return(0,r.nT)((e=>{e(c.subscribe((e=>{e.dismiss?s.value=s.value.map((t=>t.id===e.id?{...t,delete:!0}:t)):(0,r.dY)((()=>{const t=s.value.findIndex((t=>t.id===e.id));s.value=-1!==t?[...s.value.slice(0,t),{...s.value[t],...e},...s.value.slice(t+1)]:[e,...s.value]}))})))})),(0,r.wB)((()=>t.theme),(e=>{if("system"!==e)return void(f.value=e);if("system"===e&&(window.matchMedia&&window.matchMedia("(prefers-color-scheme: dark)").matches?f.value="dark":f.value="light"),typeof window>"u")return;const t=window.matchMedia("(prefers-color-scheme: dark)");try{t.addEventListener("change",(({matches:e})=>{f.value=e?"dark":"light"}))}catch{t.addListener((({matches:e})=>{try{f.value=e?"dark":"light"}catch(e){console.error(e)}}))}})),(0,r.nT)((()=>{g.value&&v.value&&(v.value.focus({preventScroll:!0}),v.value=null,y.value=!1)})),(0,r.nT)((()=>{s.value.length<=1&&(p.value=!1)})),(0,r.nT)((e=>{function n(e){const n=t.hotkey.every((t=>e[t]||e.code===t)),r=Array.isArray(g.value)?g.value[0]:g.value;n&&(p.value=!0,null==r||r.focus());const a=document.activeElement===g.value||(null==r?void 0:r.contains(document.activeElement));"Escape"===e.code&&a&&(p.value=!1)}M&&(document.addEventListener("keydown",n),e((()=>{document.removeEventListener("keydown",n)})))})),(e,c)=>((0,r.uX)(),(0,r.CE)("section",{"aria-label":`${e.containerAriaLabel} ${(0,a.R1)(w)}`,tabIndex:-1,"aria-live":"polite","aria-relevant":"additions text","aria-atomic":"false"},[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(u.value,((u,f)=>{var v;return(0,r.uX)(),(0,r.CE)("ol",(0,r.v6)({key:u,ref_for:!0,ref_key:"listRef",ref:g,"data-sonner-toaster":"",class:t.class,dir:"auto"===e.dir?n():e.dir,tabIndex:-1,"data-theme":e.theme,"data-rich-colors":e.richColors,"data-y-position":u.split("-")[0],"data-x-position":u.split("-")[1],"data-lifted":p.value&&s.value.length>1&&!e.expand,style:{"--front-toast-height":`${null==(v=d.value[0])?void 0:v.height}px`,"--offset":"number"==typeof e.offset?`${e.offset}px`:e.offset||D,"--width":"356px","--gap":`${e.gap}px`,...e.style,...(0,a.R1)(i).style}},e.$attrs,{onBlur:E,onFocus:_,onMouseenter:c[1]||(c[1]=()=>p.value=!0),onMousemove:c[2]||(c[2]=()=>p.value=!0),onMouseleave:c[3]||(c[3]=()=>{h.value||(p.value=!1)}),onPointerdown:T,onPointerup:c[4]||(c[4]=()=>h.value=!1)}),[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(l.value(u,f),((t,n)=>{var a,i,l,f,g,v,y,w,E;return(0,r.uX)(),(0,r.Wv)(m,{key:t.id,heights:d.value.filter((e=>e.position===t.position)),icons:e.icons,index:n,toast:t,defaultRichColors:e.richColors,duration:(null==(a=e.toastOptions)?void 0:a.duration)??e.duration,class:(0,o.C4)((null==(i=e.toastOptions)?void 0:i.class)??""),descriptionClass:null==(l=e.toastOptions)?void 0:l.descriptionClass,invert:e.invert,visibleToasts:e.visibleToasts,closeButton:(null==(f=e.toastOptions)?void 0:f.closeButton)??e.closeButton,interacting:h.value,position:u,style:(0,o.Tr)(null==(g=e.toastOptions)?void 0:g.style),unstyled:null==(v=e.toastOptions)?void 0:v.unstyled,classes:null==(y=e.toastOptions)?void 0:y.classes,cancelButtonStyle:null==(w=e.toastOptions)?void 0:w.cancelButtonStyle,actionButtonStyle:null==(E=e.toastOptions)?void 0:E.actionButtonStyle,toasts:s.value.filter((e=>e.position===t.position)),expandByDefault:e.expand,gap:e.gap,expanded:p.value,pauseWhenPageIsHidden:e.pauseWhenPageIsHidden,cn:e.cn,"onUpdate:heights":c[0]||(c[0]=e=>{d.value=e}),onRemoveToast:k},{"close-icon":(0,r.k6)((()=>[(0,r.RG)(e.$slots,"close-icon",{},(()=>[(0,r.bF)(b)]))])),"loading-icon":(0,r.k6)((()=>[(0,r.RG)(e.$slots,"loading-icon",{},(()=>[(0,r.bF)(O,{visible:"loading"===t.type},null,8,["visible"])]))])),"success-icon":(0,r.k6)((()=>[(0,r.RG)(e.$slots,"success-icon",{},(()=>[(0,r.bF)(x)]))])),"error-icon":(0,r.k6)((()=>[(0,r.RG)(e.$slots,"error-icon",{},(()=>[(0,r.bF)(A)]))])),"warning-icon":(0,r.k6)((()=>[(0,r.RG)(e.$slots,"warning-icon",{},(()=>[(0,r.bF)(C)]))])),"info-icon":(0,r.k6)((()=>[(0,r.RG)(e.$slots,"info-icon",{},(()=>[(0,r.bF)(S)]))])),_:2},1032,["heights","icons","index","toast","defaultRichColors","duration","class","descriptionClass","invert","visibleToasts","closeButton","interacting","position","style","unstyled","classes","cancelButtonStyle","actionButtonStyle","toasts","expandByDefault","gap","expanded","pauseWhenPageIsHidden","cn"])})),128))],16,P)})),128))],8,R))}}),N={toast:()=>null};function I(e){N.toast=e}function L(){return{initializeToasts:I,toast:(e,t="info",n={timeout:3e3})=>{N.toast(e,t,n)}}}const B=(0,r.pM)({__name:"ScalarToasts",setup(e){const t=(0,a.KR)(!1);(0,r.sV)((()=>t.value=!0));const n={success:d.success,error:d.error,warn:d.warning,info:d},{initializeToasts:o}=L();return o(((e,t="info",r={})=>{(n[t]||n.info)(e,{duration:r.timeout||3e3,description:r.description})})),(e,n)=>t.value?((0,r.uX)(),(0,r.Wv)((0,a.R1)($),{key:0,class:"scalar-toaster"})):(0,r.Q3)("",!0)}});!function(){try{if(typeof document<"u"){var e=document.createElement("style");e.appendChild(document.createTextNode(".scalar-toaster [data-sonner-toast][data-styled=true]{background:var(--scalar-background-1);color:var(--scalar-color-1);padding:18px;border:none;border-radius:var(--scalar-radius-lg);font-size:var(--scalar-font-size-3);font-weight:var(--scalar-font-medium);box-shadow:var(--scalar-shadow-2)}.scalar-toaster [data-sonner-toast] [data-icon]{align-self:flex-start;position:relative;top:2px}.scalar-toaster [data-sonner-toast][data-styled=true][data-expanded=true]{height:auto}.scalar-toaster [data-sonner-toast][data-type=error]{background:var(--scalar-background-1)}.scalar-toaster [data-sonner-toast][data-type=error] [data-icon]{color:color-mix(in srgb,var(--scalar-color-red) 75%,var(--scalar-color-1))}.scalar-toaster [data-sonner-toast][data-type=warning]{background:var(--scalar-background-1)}.scalar-toaster [data-sonner-toast][data-type=warning] [data-icon]{color:color-mix(in srgb,var(--scalar-color-orange) 90%,var(--scalar-color-1))}")),document.head.appendChild(e)}}catch(e){console.error("vite-plugin-css-injected-by-js",e)}}()},8186:(e,t,n)=>{"use strict";n.d(t,{oT:()=>l});const r={DEBOUNCE_MS:328,MAX_WAIT_MS:1e3};var a=n(796);class o{parentData;maxRecords;records=[];sideEffects=[];idx=0;debug;constructor(e,t=5e3,n=!1){this.maxRecords=t,this.parentData=e,this.debug=n}_unsavedMutate(e,t){(0,a.F)(this.parentData,e,t),this.runSideEffects(e)}addSideEffect(e,t,n,r=!0){this.sideEffects.push({triggers:e,effect:t,name:n}),r&&(t(this.parentData),this.debug&&console.info(`Running mutation side effect: ${n}`,"debug"))}runSideEffects(e){this.sideEffects.forEach((({effect:t,triggers:n,name:r})=>{(n.some((t=>e.includes(t)))||e.length<1)&&(t(this.parentData),this.debug&&console.info(`Running mutation side effect: ${r}`,"debug"))}))}mutate(e,t,n=null){this.idx<this.records.length-1&&this.records.splice(this.idx+1);const r=(0,a.L)(this.parentData,e);r!==t&&((0,a.F)(this.parentData,e,t),this.runSideEffects(e),this.records.push({prev:n??r,value:t,path:e}),this.idx=this.records.length-1,this.records.length>this.maxRecords&&this.records.shift(),this.debug&&console.info(`Set object '${this.idx}' '${e}' to ${t}`,"debug"))}undo(){if(this.idx<0||this.records.length<1)return!1;this.debug&&console.info("Undoing Mutation","debug");const e=this.records[this.idx];return this.idx-=1,e&&this._unsavedMutate(e.path,e.prev),!0}redo(){if(this.idx>this.records.length-2)return!1;this.debug&&console.info("Redoing Mutation","debug");const e=this.records[this.idx+1];return this.idx+=1,e&&this._unsavedMutate(e.path,e.value),!0}}var i=n(1284);const s=500;function l(e,t,n,a=s){function l(n){const r=t[n];return r||console.warn(`Missing ${e[n]?"mutator":"object"} for uid: ${n}`),r??null}const c=n?function(e,t,{maxWait:n}={}){let r=null,a=null;return function(...o){r&&clearTimeout(r),r=setTimeout((()=>{null!==a&&clearTimeout(a),a=null,e.apply(this,o)}),t),n&&!a&&(a=setTimeout((()=>{null!==r&&clearTimeout(r),a=null,e.apply(this,o)}),n))}}((()=>localStorage.setItem(n,(0,i.As)(e))),r.DEBOUNCE_MS,{maxWait:r.MAX_WAIT_MS}):()=>null;return{add:n=>{e[n.uid]=n,t[n.uid]=new o(n,a),c()},delete:n=>{delete e[n],delete t[n],c()},set:n=>{e[n.uid]=n,t[n.uid]=new o(n,a),c()},edit:(e,t,n)=>{const r=l(e);r?.mutate(t,n),c()},untrackedEdit:(e,t,n)=>{const r=l(e);r?._unsavedMutate(t,n),c()},undo:e=>{const t=l(e);t?.undo(),c()},redo:e=>{const t=l(e);t?.redo(),c()},reset:()=>{Object.keys(e).forEach((n=>{delete e[n],delete t[n]})),c()}}}},8207:(e,t,n)=>{"use strict";n.d(t,{A:()=>c});var r=n(641),a=n(953),o=n(1266),i=n(4048);const s=["href"],l=(0,r.pM)({__name:"OpenApiClientButton",props:{buttonSource:{},source:{default:"api-reference"},isDevelopment:{type:Boolean},integration:{},url:{}},setup(e){const t=(0,r.EW)((()=>{const t=e.url??(typeof window<"u"?window.location.href:void 0),n=(0,i.rP)(t);if(null==n||!n.length)return;const r=new URL(e.isDevelopment?"http://localhost:5065":"https://client.scalar.com");if(r.searchParams.set("url",n),null!==e.integration&&r.searchParams.set("integration",e.integration??"vue"),r.searchParams.set("utm_source","api-reference"),r.searchParams.set("utm_medium","button"),r.searchParams.set("utm_campaign",e.buttonSource),"gitbook"===e.source){r.searchParams.set("utm_source","gitbook");const e=document.querySelector("img.dark\\:block[alt='Logo']"),t=document.querySelector("img.dark\\:hidden[alt='Logo']");e&&e instanceof HTMLImageElement&&r.searchParams.set("dark_logo",encodeURIComponent(e.src)),t&&t instanceof HTMLImageElement&&r.searchParams.set("light_logo",encodeURIComponent(t.src))}return r.toString()}));return(e,n)=>t.value?((0,r.uX)(),(0,r.CE)("a",{key:0,class:"open-api-client-button",href:t.value,target:"_blank"},[(0,r.bF)((0,a.R1)(o.kv),{icon:"ExternalLink",size:"xs",thickness:"2.5"}),n[0]||(n[0]=(0,r.eW)(" Open API Client "))],8,s)):(0,r.Q3)("",!0)}}),c=(0,n(3526).A)(l,[["__scopeId","data-v-2e9f75f1"]])},8252:(e,t,n)=>{"use strict";n.r(t),n.d(t,{default:()=>T});var r=n(641),a=n(953),o=n(33),i=n(3751),s=n(1266),l=n(2667),c=n(9351),u=n(1636),d=n(4486),p=n(3966),h=n(1465),f=n(8082),m=n(9005),g=n(2490),v=n(9135),b=n(5936),y=n(8309);const w={class:"divide-0.5 flex w-full divide-x"},O=(0,r.pM)({__name:"ServerForm",setup(e){const{activeRouterParams:t,activeWorkspaceCollections:n}=(0,v.JC)(),{collections:a,servers:o,serverMutators:i}=(0,y.hA)(),s=[{label:"URL",key:"url",placeholder:"https://void.scalar.com/api"},{label:"Label",key:"description",placeholder:"Production"}],l=(0,r.EW)((()=>{const e=a[t.value.collection];if(!e)return;const n="string"==typeof t.value.servers&&"default"===t.value.servers?e.servers[0]:e.servers.find((e=>e===t.value.servers));return n?o[n]:void 0})),c=(e,t)=>{!n.value||!l.value||i.edit(l.value.uid,e,t)};return(e,t)=>((0,r.uX)(),(0,r.CE)("div",w,[l.value?((0,r.uX)(),(0,r.Wv)(b.A,{key:0,data:l.value,onUpdate:c,options:s,title:"Server"},null,8,["data"])):(0,r.Q3)("",!0)]))}});var k=n(3858),x=n(6099);const E={class:"flex-1"},S=["onClick"],_={class:"flex h-5 max-w-[14px] items-center justify-center pr-px"},C={class:"line-clamp-1 w-full break-all text-left font-medium"},T=(0,r.pM)({__name:"Servers",setup(e){const{activeWorkspaceCollections:t,activeWorkspace:n,activeRouterParams:b}=(0,v.JC)(),{servers:w,events:T,serverMutators:A}=(0,y.hA)(),{push:R,resolve:P}=(0,c.rd)(),{collapsedSidebarFolders:D,toggleSidebarFolder:M}=(0,x.c)(),$=e=>D[e];function N(e){T.commandPalette.emit({commandName:"Add Server",metaData:{parentUid:e}})}const I=(0,r.EW)((()=>t.value.filter((e=>{var t;return"Drafts"!==(null==(t=e.info)?void 0:t.title)})))),L=e=>A.delete(e,b.value.collection),B=(0,r.EW)((()=>Object.keys(w).length>0));return(e,t)=>((0,r.uX)(),(0,r.Wv)(f.A,null,{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(k.A),{title:"Servers"},{content:(0,r.k6)((()=>[(0,r.Lk)("div",E,[(0,r.bF)(p.A,null,{default:(0,r.k6)((()=>[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(I.value,(e=>{var c;return(0,r.uX)(),(0,r.CE)("li",{key:e.uid,class:"gap-1/2 flex flex-col"},[(0,r.Lk)("button",{class:"hover:bg-b-2 group flex w-full items-center gap-1.5 break-words rounded p-1.5 text-left text-sm font-medium",type:"button",onClick:t=>(0,a.R1)(M)(e.uid)},[(0,r.Lk)("span",_,[(0,r.bF)((0,a.R1)(l.$n),{class:"text-sidebar-c-2 size-3.5 min-w-3.5 stroke-2 group-hover:hidden",src:e["x-scalar-icon"]||"interface-content-folder"},null,8,["src"]),(0,r.Lk)("div",{class:(0,o.C4)({"rotate-90":(0,a.R1)(D)[e.uid]})},[(0,r.bF)((0,a.R1)(s.kv),{class:"text-c-3 hover:text-c-1 hidden text-sm group-hover:block",icon:"ChevronRight",size:"md"})],2)]),(0,r.Lk)("span",C,(0,o.v_)((null==(c=e.info)?void 0:c.title)??""),1),(0,r.bF)((0,a.R1)(s.RO),{class:"hover:bg-b-3 hover:text-c-1 hidden aspect-square h-fit px-0.5 py-0 group-hover:block group-focus-visible:opacity-100 group-has-[:focus-visible]:opacity-100",size:"sm",variant:"ghost",onClick:(0,i.D$)((t=>N(e.uid)),["stop","prevent"])},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(s.kv),{icon:"Add",size:"md",thickness:"2"})])),_:2},1032,["onClick"])],8,S),(0,r.bo)((0,r.Lk)("div",{class:(0,o.C4)({"before:bg-border before:z-1 gap-1/2 relative mb-[.5px] flex flex-col before:pointer-events-none before:absolute before:left-3 before:top-0 before:h-[calc(100%_+_.5px)] before:w-[.5px] last:mb-0 last:before:h-full":Object.keys(e.servers||{}).length>0})},[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(e.servers,(t=>{var o;return(0,r.uX)(),(0,r.Wv)(h.A,{key:t,class:"[&>a]:pl-[1.625rem]",collectionId:e.uid,isDeletable:!0,type:"servers",variable:{name:(null==(o=(0,a.R1)(w)[t])?void 0:o.url)??"",uid:t},onClick:r=>((e,t,r)=>{var a;const o={name:"servers",params:{[g.TU.Workspace]:null==(a=n.value)?void 0:a.uid,[g.TU.Collection]:r,[g.TU.Servers]:t}};e.metaKey?window.open(P(o).href,"_blank"):R(o)})(r,t,e.uid),onDelete:L},null,8,["collectionId","variable","onClick"])})),128)),0===Object.keys(e.servers||{}).length?((0,r.uX)(),(0,r.Wv)((0,a.R1)(s.RO),{key:0,class:"text-c-1 hover:bg-b-2 mb-[.5px] flex h-8 w-full justify-start gap-1.5 py-0 pl-6 text-xs",variant:"ghost",onClick:t=>N(e.uid)},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(s.kv),{icon:"Add",size:"sm"}),t[0]||(t[0]=(0,r.Lk)("span",null,"Add Server",-1))])),_:2},1032,["onClick"])):(0,r.Q3)("",!0)],2),[[i.aG,$(e.uid)]])])})),128))])),_:1})])])),button:(0,r.k6)((()=>[(0,r.bF)(d.A,{click:N},{title:(0,r.k6)((()=>t[1]||(t[1]=[(0,r.eW)(" Add Server ")]))),_:1})])),_:1}),(0,r.bF)(m.A,{class:"flex-1"},{default:(0,r.k6)((()=>[B.value?((0,r.uX)(),(0,r.Wv)(O,{key:0})):((0,r.uX)(),(0,r.Wv)(u.A,{key:1}))])),_:1})])),_:1}))}})},8256:(e,t,n)=>{"use strict";function r(e,t){const n=String(e);if("string"!=typeof t)throw new TypeError("Expected character");let r=0,a=n.indexOf(t);for(;-1!==a;)r++,a=n.indexOf(t,a+t.length);return r}n.d(t,{D:()=>r})},8309:(e,t,n)=>{"use strict";n.d(t,{pd:()=>v,Pg:()=>g,hA:()=>b});var r=n(1347),a=n(4048),o=n(8186),i=n(953);function s({requests:e,requestMutators:t,requestExamples:n,requestExampleMutators:a,workspaces:o,workspaceMutators:i,collections:s,collectionMutators:l,tagMutators:c,serverMutators:u}){return{addCollection:(e,t)=>{const n=r.CV.parse(e),a=o[t];return a&&i.edit(t,"collections",[...a.collections,n.uid]),l.add(n),n},deleteCollection:(r,o)=>{var d,p;if(o.uid){if("Drafts"===(null==(p=null==(d=s[r.uid])?void 0:d.info)?void 0:p.title))return void console.warn("The drafts collection cannot be deleted");if(1===Object.values(s).length)return void console.warn("You must have at least one collection");r.tags.forEach((e=>c.delete(e))),r.requests.forEach((r=>{const o=e[r];o&&(t.delete(r),o.examples.forEach((e=>n[e]&&a.delete(e))))})),r.servers.forEach((e=>{e&&u.delete(e)})),i.edit(o.uid,"collections",o.collections.filter((e=>e!==r.uid))),l.delete(r.uid)}},addCollectionEnvironment:(e,t,n)=>{const r=s[n];if(r){const a=r["x-scalar-environments"]||{};l.edit(n,"x-scalar-environments",{...a,[e]:t})}},removeCollectionEnvironment:(e,t)=>{const n=s[t];if(n){const r=n["x-scalar-environments"]||{};e in r&&(delete r[e],l.edit(t,"x-scalar-environments",r))}}}}var l=n(8803);function c(){const e=new Set;function t(t){return e.add(t),()=>n(t)}function n(t){e.delete(t)}return{on:t,once:function(e){return t((function t(...r){n(t),e(...r)}))},off:n,emit:function(t){null==e||e.forEach((e=>e(t)))},reset:function(){e.clear()},listeners:()=>Array.from(e)}}function u(){return{executeRequest:c(),cancelRequest:c(),requestStatus:c(),commandPalette:c(),hotKeys:c()}}var d=n(5191),p=n(7356),h=n(9582),f=n(1266),m=n(641);const g=({useLocalStorage:e=!0,showSidebar:t=!0,proxyUrl:n,themeId:c,hideClientButton:m=!1,integration:g})=>{const{collections:v,collectionMutators:b}=function(e){const t=(0,i.Kh)({});return{collections:t,collectionMutators:(0,o.oT)(t,(0,i.Kh)({}),e&&a.nF.COLLECTION)}}(e),{tags:y,tagMutators:w}=function(e){const t=(0,i.Kh)({});return{tags:t,tagMutators:(0,o.oT)(t,(0,i.Kh)({}),e&&a.nF.TAG)}}(e),{requests:O,requestMutators:k}=(0,p.I5)(e),{requestExamples:x,requestExampleMutators:E}=function(e){const t=(0,i.Kh)({});return{requestExamples:t,requestExampleMutators:(0,o.oT)(t,(0,i.Kh)({}),e&&a.nF.REQUEST_EXAMPLE)}}(e),{cookies:S,cookieMutators:_}=function(e){const t=(0,i.Kh)({});return{cookies:t,cookieMutators:(0,o.oT)(t,(0,i.Kh)({}),e&&a.nF.COOKIE)}}(e),{environments:C,environmentMutators:T}=function(e){const t=(0,i.Kh)({}),n=(0,o.oT)(t,(0,i.Kh)({}),e&&a.nF.ENVIRONMENT);return n.add(l.V.parse({uid:"default",name:"Default Environment",color:"#0082D0",value:JSON.stringify({exampleKey:"exampleValue"},null,2),isDefault:!0})),{environments:t,environmentMutators:n}}(e),{servers:A,serverMutators:R}=function(e){const t=(0,i.Kh)({});return{servers:t,serverMutators:(0,o.oT)(t,(0,i.Kh)({}),e&&a.nF.SERVER)}}(e),{securitySchemes:P,securitySchemeMutators:D}=function(e){const t=(0,i.Kh)({});return{securitySchemes:t,securitySchemeMutators:(0,o.oT)(t,(0,i.Kh)({}),e&&a.nF.SECURITY_SCHEME)}}(e),{workspaces:M,workspaceMutators:$}=function(e){const t=(0,i.Kh)({});return{workspaces:t,workspaceMutators:(0,o.oT)(t,(0,i.Kh)({}),e&&a.nF.WORKSPACE)}}(e),N={collections:v,collectionMutators:b,tags:y,tagMutators:w,requests:O,requestMutators:k,requestExamples:x,requestExampleMutators:E,cookies:S,cookieMutators:_,environments:C,environmentMutators:T,servers:A,serverMutators:R,securitySchemes:P,securitySchemeMutators:D,workspaces:M,workspaceMutators:$},{addTag:I,deleteTag:L}=function({collectionMutators:e,collections:t,requests:n,requestMutators:o,tagMutators:i}){return{addTag:(n,o)=>{const s=t[o],l=(0,a.TA)(n,r.c3,!1);return l&&s?(e.edit(o,"tags",[...s.tags,l.uid]),e.edit(o,"children",[...s.children,l.uid]),i.add(l),l):console.error("INVALID TAG DATA",n)},deleteTag:(r,a)=>{const s=t[a];s&&(e.edit(a,"tags",s.tags.filter((e=>e!==r.uid))),e.edit(a,"children",s.children.filter((e=>e!==r.uid))),r.children.forEach((t=>{var i;const l=n[t];if(!l)return;const c=null==(i=l.tags)?void 0:i.filter((e=>e!==r.name));o.edit(l.uid,"tags",c),(null==c||!c.length)&&!s.children.includes(t)&&e.edit(a,"children",[...s.children,t])})),i.delete(r.uid))}}}(N),{addRequest:B,deleteRequest:F,findRequestParents:Q}=(0,p.D9)(N,I),{deleteEnvironment:U}=function({environmentMutators:e}){return{deleteEnvironment:t=>{"default"!==t?e.delete(t):console.warn("Default environment cannot be deleted.")}}}(N),{addServer:z,deleteServer:j}=function({serverMutators:e,collections:t,collectionMutators:n,requests:a,requestMutators:o}){return{addServer:(i,s)=>{const l=r.zT.parse(i);return t[s]?n.edit(s,"servers",[...t[s].servers,l.uid]):a[s]&&o.edit(s,"servers",[...a[s].servers,l.uid]),e.add(l),l},deleteServer:(r,a)=>{t[a]&&(n.edit(a,"servers",t[a].servers.filter((e=>e!==r))),e.delete(r))}}}(N),{addCollection:W,deleteCollection:X}=s(N),{addRequestExample:q,deleteRequestExample:V}=function({requestExamples:e,requestExampleMutators:t,requestMutators:n,requests:o}){return{addRequestExample:(o,i)=>{const s=i??(0,a.zR)((o.summary??"Example")+" #1",(t=>o.examples.some((n=>{var r;return(null==(r=e[n])?void 0:r.name)===t})))),l=(0,r.oE)(o,s);return t.add(l),n.edit(o.uid,"examples",[...o.examples,l.uid]),l},deleteRequestExample:e=>{var r;e.requestUid&&(n.edit(e.requestUid,"examples",(null==(r=o[e.requestUid])?void 0:r.examples.filter((t=>t!==e.uid)))||[]),t.delete(e.uid))}}}(N),{addWorkspace:H,deleteWorkspace:Z}=function({workspaces:e,workspaceMutators:t,collectionMutators:n,requestMutators:a,requestExampleMutators:o}){return{addWorkspace:(e={})=>{const{request:i}=(0,p.AJ)(),s=r.$W.parse({name:"Example",requestUid:i.uid});i.examples.push(s.uid);const l=r.CV.parse({info:{title:"Drafts"},children:[i.uid],requests:[i.uid]}),c=h.N.parse({...e,collections:[l.uid]});return o.add(s),a.add(i),n.add(l),t.add(c),c},deleteWorkspace:n=>{Object.keys(e).length<=1?console.warn("The last workspace cannot be deleted."):t.delete(n)}}}(N),{addSecurityScheme:G,deleteSecurityScheme:K}=function({securitySchemeMutators:e,collectionMutators:t,collections:n,requests:a,requestMutators:o}){return{addSecurityScheme:(a,o)=>{const i=r.NV.parse(a);return e.add(i),o&&n[o]&&t.edit(o,"securitySchemes",[...n[o].securitySchemes,i.uid]),i},deleteSecurityScheme:r=>{Object.values(n).forEach((e=>{e.securitySchemes.includes(r)&&t.edit(e.uid,"securitySchemes",e.securitySchemes.filter((e=>e!==r)))})),Object.values(a).forEach((e=>{var t,n,i,s;null!=(t=e.security)&&t.some((e=>Object.keys(e).includes(r)))&&o.edit(e.uid,"security",null==(i=null==(n=a[e.uid])?void 0:n.security)?void 0:i.filter((e=>!Object.keys(e).includes(r)))),e.selectedSecuritySchemeUids.flat().includes(r)&&o.edit(e.uid,"selectedSecuritySchemeUids",null==(s=e.selectedSecuritySchemeUids)?void 0:s.filter((e=>Array.isArray(e)?!e.includes(r):e!==r)))})),e.delete(r)}}}(N),{addCollectionEnvironment:Y,removeCollectionEnvironment:J}=s(N),ee=(0,i.Kh)([]),{importSpecFile:te,importSpecFromUrl:ne}=(0,d.v)(N),re=(0,i.KR)((e?null==localStorage?void 0:localStorage.getItem("sidebarWidth"):void 0)||"280px"),ae=(0,f.hS)();Object.values(M).forEach((({uid:e})=>{n&&$.edit(e,"proxyUrl",n),c&&$.edit(e,"themeId",c)})),typeof window<"u"&&(window.dataDump=()=>({collections:(0,i.ux)(v),cookies:(0,i.ux)(S),environments:(0,i.ux)(C),requestExamples:(0,i.ux)(x),requests:(0,i.ux)(O),securitySchemes:(0,i.ux)(P),servers:(0,i.ux)(A),tags:(0,i.ux)(y),workspaces:(0,i.ux)(M)}));const oe=u();return{workspaces:M,collections:v,tags:y,cookies:S,environments:C,requestExamples:x,requests:O,servers:A,securitySchemes:P,modalState:ae,events:oe,sidebarWidth:re,setSidebarWidth:t=>{re.value=t,e&&(null==localStorage||localStorage.setItem("sidebarWidth",t))},proxyUrl:n,hideClientButton:m,showSidebar:t,integration:g,importSpecFile:te,importSpecFromUrl:ne,cookieMutators:_,collectionMutators:{...b,rawAdd:b.add,add:W,delete:X,addEnvironment:Y,removeEnvironment:J},environmentMutators:{...T,delete:U},requestMutators:{...k,rawAdd:k.add,add:B,delete:F},findRequestParents:Q,requestExampleMutators:{...E,rawAdd:E.add,add:q,delete:V},requestHistory:ee,securitySchemeMutators:{...D,rawAdd:D.add,add:G,delete:K},serverMutators:{...R,rawAdd:R.add,add:z,delete:j},tagMutators:{...w,rawAdd:w.add,add:I,delete:L},workspaceMutators:{...$,rawAdd:$.add,add:H,delete:Z},addCollectionEnvironment:Y,removeCollectionEnvironment:J}},v=Symbol(),b=()=>{const e=(0,m.WQ)(v);if(!e)throw new Error("Workspace store not provided");return e}},8416:e=>{function t(e){return e instanceof Map?e.clear=e.delete=e.set=function(){throw new Error("map is read-only")}:e instanceof Set&&(e.add=e.clear=e.delete=function(){throw new Error("set is read-only")}),Object.freeze(e),Object.getOwnPropertyNames(e).forEach((n=>{const r=e[n],a=typeof r;"object"!==a&&"function"!==a||Object.isFrozen(r)||t(r)})),e}class n{constructor(e){void 0===e.data&&(e.data={}),this.data=e.data,this.isMatchIgnored=!1}ignoreMatch(){this.isMatchIgnored=!0}}function r(e){return e.replace(/&/g,"&amp;").replace(/</g,"&lt;").replace(/>/g,"&gt;").replace(/"/g,"&quot;").replace(/'/g,"&#x27;")}function a(e,...t){const n=Object.create(null);for(const t in e)n[t]=e[t];return t.forEach((function(e){for(const t in e)n[t]=e[t]})),n}const o=e=>!!e.scope;class i{constructor(e,t){this.buffer="",this.classPrefix=t.classPrefix,e.walk(this)}addText(e){this.buffer+=r(e)}openNode(e){if(!o(e))return;const t=((e,{prefix:t})=>{if(e.startsWith("language:"))return e.replace("language:","language-");if(e.includes(".")){const n=e.split(".");return[`${t}${n.shift()}`,...n.map(((e,t)=>`${e}${"_".repeat(t+1)}`))].join(" ")}return`${t}${e}`})(e.scope,{prefix:this.classPrefix});this.span(t)}closeNode(e){o(e)&&(this.buffer+="</span>")}value(){return this.buffer}span(e){this.buffer+=`<span class="${e}">`}}const s=(e={})=>{const t={children:[]};return Object.assign(t,e),t};class l{constructor(){this.rootNode=s(),this.stack=[this.rootNode]}get top(){return this.stack[this.stack.length-1]}get root(){return this.rootNode}add(e){this.top.children.push(e)}openNode(e){const t=s({scope:e});this.add(t),this.stack.push(t)}closeNode(){if(this.stack.length>1)return this.stack.pop()}closeAllNodes(){for(;this.closeNode(););}toJSON(){return JSON.stringify(this.rootNode,null,4)}walk(e){return this.constructor._walk(e,this.rootNode)}static _walk(e,t){return"string"==typeof t?e.addText(t):t.children&&(e.openNode(t),t.children.forEach((t=>this._walk(e,t))),e.closeNode(t)),e}static _collapse(e){"string"!=typeof e&&e.children&&(e.children.every((e=>"string"==typeof e))?e.children=[e.children.join("")]:e.children.forEach((e=>{l._collapse(e)})))}}class c extends l{constructor(e){super(),this.options=e}addText(e){""!==e&&this.add(e)}startScope(e){this.openNode(e)}endScope(){this.closeNode()}__addSublanguage(e,t){const n=e.root;t&&(n.scope=`language:${t}`),this.add(n)}toHTML(){return new i(this,this.options).value()}finalize(){return this.closeAllNodes(),!0}}function u(e){return e?"string"==typeof e?e:e.source:null}function d(e){return f("(?=",e,")")}function p(e){return f("(?:",e,")*")}function h(e){return f("(?:",e,")?")}function f(...e){return e.map((e=>u(e))).join("")}function m(...e){const t=function(e){const t=e[e.length-1];return"object"==typeof t&&t.constructor===Object?(e.splice(e.length-1,1),t):{}}(e);return"("+(t.capture?"":"?:")+e.map((e=>u(e))).join("|")+")"}function g(e){return new RegExp(e.toString()+"|").exec("").length-1}const v=/\[(?:[^\\\]]|\\.)*\]|\(\??|\\([1-9][0-9]*)|\\./;function b(e,{joinWith:t}){let n=0;return e.map((e=>{n+=1;const t=n;let r=u(e),a="";for(;r.length>0;){const e=v.exec(r);if(!e){a+=r;break}a+=r.substring(0,e.index),r=r.substring(e.index+e[0].length),"\\"===e[0][0]&&e[1]?a+="\\"+String(Number(e[1])+t):(a+=e[0],"("===e[0]&&n++)}return a})).map((e=>`(${e})`)).join(t)}const y="[a-zA-Z]\\w*",w="[a-zA-Z_]\\w*",O="\\b\\d+(\\.\\d+)?",k="(-?)(\\b0[xX][a-fA-F0-9]+|(\\b\\d+(\\.\\d*)?|\\.\\d+)([eE][-+]?\\d+)?)",x="\\b(0b[01]+)",E={begin:"\\\\[\\s\\S]",relevance:0},S={scope:"string",begin:"'",end:"'",illegal:"\\n",contains:[E]},_={scope:"string",begin:'"',end:'"',illegal:"\\n",contains:[E]},C=function(e,t,n={}){const r=a({scope:"comment",begin:e,end:t,contains:[]},n);r.contains.push({scope:"doctag",begin:"[ ]*(?=(TODO|FIXME|NOTE|BUG|OPTIMIZE|HACK|XXX):)",end:/(TODO|FIXME|NOTE|BUG|OPTIMIZE|HACK|XXX):/,excludeBegin:!0,relevance:0});const o=m("I","a","is","so","us","to","at","if","in","it","on",/[A-Za-z]+['](d|ve|re|ll|t|s|n)/,/[A-Za-z]+[-][a-z]+/,/[A-Za-z][a-z]{2,}/);return r.contains.push({begin:f(/[ ]+/,"(",o,/[.]?[:]?([.][ ]|[ ])/,"){3}")}),r},T=C("//","$"),A=C("/\\*","\\*/"),R=C("#","$"),P={scope:"number",begin:O,relevance:0},D={scope:"number",begin:k,relevance:0},M={scope:"number",begin:x,relevance:0},$={scope:"regexp",begin:/\/(?=[^/\n]*\/)/,end:/\/[gimuy]*/,contains:[E,{begin:/\[/,end:/\]/,relevance:0,contains:[E]}]},N={scope:"title",begin:y,relevance:0},I={scope:"title",begin:w,relevance:0},L={begin:"\\.\\s*"+w,relevance:0};var B=Object.freeze({__proto__:null,APOS_STRING_MODE:S,BACKSLASH_ESCAPE:E,BINARY_NUMBER_MODE:M,BINARY_NUMBER_RE:x,COMMENT:C,C_BLOCK_COMMENT_MODE:A,C_LINE_COMMENT_MODE:T,C_NUMBER_MODE:D,C_NUMBER_RE:k,END_SAME_AS_BEGIN:function(e){return Object.assign(e,{"on:begin":(e,t)=>{t.data._beginMatch=e[1]},"on:end":(e,t)=>{t.data._beginMatch!==e[1]&&t.ignoreMatch()}})},HASH_COMMENT_MODE:R,IDENT_RE:y,MATCH_NOTHING_RE:/\b\B/,METHOD_GUARD:L,NUMBER_MODE:P,NUMBER_RE:O,PHRASAL_WORDS_MODE:{begin:/\b(a|an|the|are|I'm|isn't|don't|doesn't|won't|but|just|should|pretty|simply|enough|gonna|going|wtf|so|such|will|you|your|they|like|more)\b/},QUOTE_STRING_MODE:_,REGEXP_MODE:$,RE_STARTERS_RE:"!|!=|!==|%|%=|&|&&|&=|\\*|\\*=|\\+|\\+=|,|-|-=|/=|/|:|;|<<|<<=|<=|<|===|==|=|>>>=|>>=|>=|>>>|>>|>|\\?|\\[|\\{|\\(|\\^|\\^=|\\||\\|=|\\|\\||~",SHEBANG:(e={})=>{const t=/^#![ ]*\//;return e.binary&&(e.begin=f(t,/.*\b/,e.binary,/\b.*/)),a({scope:"meta",begin:t,end:/$/,relevance:0,"on:begin":(e,t)=>{0!==e.index&&t.ignoreMatch()}},e)},TITLE_MODE:N,UNDERSCORE_IDENT_RE:w,UNDERSCORE_TITLE_MODE:I});function F(e,t){"."===e.input[e.index-1]&&t.ignoreMatch()}function Q(e,t){void 0!==e.className&&(e.scope=e.className,delete e.className)}function U(e,t){t&&e.beginKeywords&&(e.begin="\\b("+e.beginKeywords.split(" ").join("|")+")(?!\\.)(?=\\b|\\s)",e.__beforeBegin=F,e.keywords=e.keywords||e.beginKeywords,delete e.beginKeywords,void 0===e.relevance&&(e.relevance=0))}function z(e,t){Array.isArray(e.illegal)&&(e.illegal=m(...e.illegal))}function j(e,t){if(e.match){if(e.begin||e.end)throw new Error("begin & end are not supported with match");e.begin=e.match,delete e.match}}function W(e,t){void 0===e.relevance&&(e.relevance=1)}const X=(e,t)=>{if(!e.beforeMatch)return;if(e.starts)throw new Error("beforeMatch cannot be used with starts");const n=Object.assign({},e);Object.keys(e).forEach((t=>{delete e[t]})),e.keywords=n.keywords,e.begin=f(n.beforeMatch,d(n.begin)),e.starts={relevance:0,contains:[Object.assign(n,{endsParent:!0})]},e.relevance=0,delete n.beforeMatch},q=["of","and","for","in","not","or","if","then","parent","list","value"];function V(e,t,n="keyword"){const r=Object.create(null);return"string"==typeof e?a(n,e.split(" ")):Array.isArray(e)?a(n,e):Object.keys(e).forEach((function(n){Object.assign(r,V(e[n],t,n))})),r;function a(e,n){t&&(n=n.map((e=>e.toLowerCase()))),n.forEach((function(t){const n=t.split("|");r[n[0]]=[e,H(n[0],n[1])]}))}}function H(e,t){return t?Number(t):function(e){return q.includes(e.toLowerCase())}(e)?0:1}const Z={},G=e=>{console.error(e)},K=(e,...t)=>{console.log(`WARN: ${e}`,...t)},Y=(e,t)=>{Z[`${e}/${t}`]||(console.log(`Deprecated as of ${e}. ${t}`),Z[`${e}/${t}`]=!0)},J=new Error;function ee(e,t,{key:n}){let r=0;const a=e[n],o={},i={};for(let e=1;e<=t.length;e++)i[e+r]=a[e],o[e+r]=!0,r+=g(t[e-1]);e[n]=i,e[n]._emit=o,e[n]._multi=!0}function te(e){!function(e){e.scope&&"object"==typeof e.scope&&null!==e.scope&&(e.beginScope=e.scope,delete e.scope)}(e),"string"==typeof e.beginScope&&(e.beginScope={_wrap:e.beginScope}),"string"==typeof e.endScope&&(e.endScope={_wrap:e.endScope}),function(e){if(Array.isArray(e.begin)){if(e.skip||e.excludeBegin||e.returnBegin)throw G("skip, excludeBegin, returnBegin not compatible with beginScope: {}"),J;if("object"!=typeof e.beginScope||null===e.beginScope)throw G("beginScope must be object"),J;ee(e,e.begin,{key:"beginScope"}),e.begin=b(e.begin,{joinWith:""})}}(e),function(e){if(Array.isArray(e.end)){if(e.skip||e.excludeEnd||e.returnEnd)throw G("skip, excludeEnd, returnEnd not compatible with endScope: {}"),J;if("object"!=typeof e.endScope||null===e.endScope)throw G("endScope must be object"),J;ee(e,e.end,{key:"endScope"}),e.end=b(e.end,{joinWith:""})}}(e)}function ne(e){function t(t,n){return new RegExp(u(t),"m"+(e.case_insensitive?"i":"")+(e.unicodeRegex?"u":"")+(n?"g":""))}class n{constructor(){this.matchIndexes={},this.regexes=[],this.matchAt=1,this.position=0}addRule(e,t){t.position=this.position++,this.matchIndexes[this.matchAt]=t,this.regexes.push([t,e]),this.matchAt+=g(e)+1}compile(){0===this.regexes.length&&(this.exec=()=>null);const e=this.regexes.map((e=>e[1]));this.matcherRe=t(b(e,{joinWith:"|"}),!0),this.lastIndex=0}exec(e){this.matcherRe.lastIndex=this.lastIndex;const t=this.matcherRe.exec(e);if(!t)return null;const n=t.findIndex(((e,t)=>t>0&&void 0!==e)),r=this.matchIndexes[n];return t.splice(0,n),Object.assign(t,r)}}class r{constructor(){this.rules=[],this.multiRegexes=[],this.count=0,this.lastIndex=0,this.regexIndex=0}getMatcher(e){if(this.multiRegexes[e])return this.multiRegexes[e];const t=new n;return this.rules.slice(e).forEach((([e,n])=>t.addRule(e,n))),t.compile(),this.multiRegexes[e]=t,t}resumingScanAtSamePosition(){return 0!==this.regexIndex}considerAll(){this.regexIndex=0}addRule(e,t){this.rules.push([e,t]),"begin"===t.type&&this.count++}exec(e){const t=this.getMatcher(this.regexIndex);t.lastIndex=this.lastIndex;let n=t.exec(e);if(this.resumingScanAtSamePosition())if(n&&n.index===this.lastIndex);else{const t=this.getMatcher(0);t.lastIndex=this.lastIndex+1,n=t.exec(e)}return n&&(this.regexIndex+=n.position+1,this.regexIndex===this.count&&this.considerAll()),n}}if(e.compilerExtensions||(e.compilerExtensions=[]),e.contains&&e.contains.includes("self"))throw new Error("ERR: contains `self` is not supported at the top-level of a language.  See documentation.");return e.classNameAliases=a(e.classNameAliases||{}),function n(o,i){const s=o;if(o.isCompiled)return s;[Q,j,te,X].forEach((e=>e(o,i))),e.compilerExtensions.forEach((e=>e(o,i))),o.__beforeBegin=null,[U,z,W].forEach((e=>e(o,i))),o.isCompiled=!0;let l=null;return"object"==typeof o.keywords&&o.keywords.$pattern&&(o.keywords=Object.assign({},o.keywords),l=o.keywords.$pattern,delete o.keywords.$pattern),l=l||/\w+/,o.keywords&&(o.keywords=V(o.keywords,e.case_insensitive)),s.keywordPatternRe=t(l,!0),i&&(o.begin||(o.begin=/\B|\b/),s.beginRe=t(s.begin),o.end||o.endsWithParent||(o.end=/\B|\b/),o.end&&(s.endRe=t(s.end)),s.terminatorEnd=u(s.end)||"",o.endsWithParent&&i.terminatorEnd&&(s.terminatorEnd+=(o.end?"|":"")+i.terminatorEnd)),o.illegal&&(s.illegalRe=t(o.illegal)),o.contains||(o.contains=[]),o.contains=[].concat(...o.contains.map((function(e){return function(e){return e.variants&&!e.cachedVariants&&(e.cachedVariants=e.variants.map((function(t){return a(e,{variants:null},t)}))),e.cachedVariants?e.cachedVariants:re(e)?a(e,{starts:e.starts?a(e.starts):null}):Object.isFrozen(e)?a(e):e}("self"===e?o:e)}))),o.contains.forEach((function(e){n(e,s)})),o.starts&&n(o.starts,i),s.matcher=function(e){const t=new r;return e.contains.forEach((e=>t.addRule(e.begin,{rule:e,type:"begin"}))),e.terminatorEnd&&t.addRule(e.terminatorEnd,{type:"end"}),e.illegal&&t.addRule(e.illegal,{type:"illegal"}),t}(s),s}(e)}function re(e){return!!e&&(e.endsWithParent||re(e.starts))}class ae extends Error{constructor(e,t){super(e),this.name="HTMLInjectionError",this.html=t}}const oe=r,ie=a,se=Symbol("nomatch"),le=function(e){const r=Object.create(null),a=Object.create(null),o=[];let i=!0;const s="Could not find the language '{}', did you forget to load/include a language module?",l={disableAutodetect:!0,name:"Plain text",contains:[]};let u={ignoreUnescapedHTML:!1,throwUnescapedHTML:!1,noHighlightRe:/^(no-?highlight)$/i,languageDetectRe:/\blang(?:uage)?-([\w-]+)\b/i,classPrefix:"hljs-",cssSelector:"pre code",languages:null,__emitter:c};function g(e){return u.noHighlightRe.test(e)}function v(e,t,n){let r="",a="";"object"==typeof t?(r=e,n=t.ignoreIllegals,a=t.language):(Y("10.7.0","highlight(lang, code, ...args) has been deprecated."),Y("10.7.0","Please use highlight(code, options) instead.\nhttps://github.com/highlightjs/highlight.js/issues/2277"),a=e,r=t),void 0===n&&(n=!0);const o={code:r,language:a};_("before:highlight",o);const i=o.result?o.result:b(o.language,o.code,n);return i.code=o.code,_("after:highlight",i),i}function b(e,t,a,o){const l=Object.create(null);function c(){if(!_.keywords)return void T.addText(A);let e=0;_.keywordPatternRe.lastIndex=0;let t=_.keywordPatternRe.exec(A),n="";for(;t;){n+=A.substring(e,t.index);const a=k.case_insensitive?t[0].toLowerCase():t[0],o=(r=a,_.keywords[r]);if(o){const[e,r]=o;if(T.addText(n),n="",l[a]=(l[a]||0)+1,l[a]<=7&&(R+=r),e.startsWith("_"))n+=t[0];else{const n=k.classNameAliases[e]||e;p(t[0],n)}}else n+=t[0];e=_.keywordPatternRe.lastIndex,t=_.keywordPatternRe.exec(A)}var r;n+=A.substring(e),T.addText(n)}function d(){null!=_.subLanguage?function(){if(""===A)return;let e=null;if("string"==typeof _.subLanguage){if(!r[_.subLanguage])return void T.addText(A);e=b(_.subLanguage,A,!0,C[_.subLanguage]),C[_.subLanguage]=e._top}else e=y(A,_.subLanguage.length?_.subLanguage:null);_.relevance>0&&(R+=e.relevance),T.__addSublanguage(e._emitter,e.language)}():c(),A=""}function p(e,t){""!==e&&(T.startScope(t),T.addText(e),T.endScope())}function h(e,t){let n=1;const r=t.length-1;for(;n<=r;){if(!e._emit[n]){n++;continue}const r=k.classNameAliases[e[n]]||e[n],a=t[n];r?p(a,r):(A=a,c(),A=""),n++}}function f(e,t){return e.scope&&"string"==typeof e.scope&&T.openNode(k.classNameAliases[e.scope]||e.scope),e.beginScope&&(e.beginScope._wrap?(p(A,k.classNameAliases[e.beginScope._wrap]||e.beginScope._wrap),A=""):e.beginScope._multi&&(h(e.beginScope,t),A="")),_=Object.create(e,{parent:{value:_}}),_}function m(e,t,r){let a=function(e,t){const n=e&&e.exec(t);return n&&0===n.index}(e.endRe,r);if(a){if(e["on:end"]){const r=new n(e);e["on:end"](t,r),r.isMatchIgnored&&(a=!1)}if(a){for(;e.endsParent&&e.parent;)e=e.parent;return e}}if(e.endsWithParent)return m(e.parent,t,r)}function g(e){return 0===_.matcher.regexIndex?(A+=e[0],1):(M=!0,0)}function v(e){const n=e[0],r=t.substring(e.index),a=m(_,e,r);if(!a)return se;const o=_;_.endScope&&_.endScope._wrap?(d(),p(n,_.endScope._wrap)):_.endScope&&_.endScope._multi?(d(),h(_.endScope,e)):o.skip?A+=n:(o.returnEnd||o.excludeEnd||(A+=n),d(),o.excludeEnd&&(A=n));do{_.scope&&T.closeNode(),_.skip||_.subLanguage||(R+=_.relevance),_=_.parent}while(_!==a.parent);return a.starts&&f(a.starts,e),o.returnEnd?0:n.length}let w={};function O(r,o){const s=o&&o[0];if(A+=r,null==s)return d(),0;if("begin"===w.type&&"end"===o.type&&w.index===o.index&&""===s){if(A+=t.slice(o.index,o.index+1),!i){const t=new Error(`0 width match regex (${e})`);throw t.languageName=e,t.badRule=w.rule,t}return 1}if(w=o,"begin"===o.type)return function(e){const t=e[0],r=e.rule,a=new n(r),o=[r.__beforeBegin,r["on:begin"]];for(const n of o)if(n&&(n(e,a),a.isMatchIgnored))return g(t);return r.skip?A+=t:(r.excludeBegin&&(A+=t),d(),r.returnBegin||r.excludeBegin||(A=t)),f(r,e),r.returnBegin?0:t.length}(o);if("illegal"===o.type&&!a){const e=new Error('Illegal lexeme "'+s+'" for mode "'+(_.scope||"<unnamed>")+'"');throw e.mode=_,e}if("end"===o.type){const e=v(o);if(e!==se)return e}if("illegal"===o.type&&""===s)return A+="\n",1;if(D>1e5&&D>3*o.index)throw new Error("potential infinite loop, way more iterations than matches");return A+=s,s.length}const k=x(e);if(!k)throw G(s.replace("{}",e)),new Error('Unknown language: "'+e+'"');const E=ne(k);let S="",_=o||E;const C={},T=new u.__emitter(u);!function(){const e=[];for(let t=_;t!==k;t=t.parent)t.scope&&e.unshift(t.scope);e.forEach((e=>T.openNode(e)))}();let A="",R=0,P=0,D=0,M=!1;try{if(k.__emitTokens)k.__emitTokens(t,T);else{for(_.matcher.considerAll();;){D++,M?M=!1:_.matcher.considerAll(),_.matcher.lastIndex=P;const e=_.matcher.exec(t);if(!e)break;const n=O(t.substring(P,e.index),e);P=e.index+n}O(t.substring(P))}return T.finalize(),S=T.toHTML(),{language:e,value:S,relevance:R,illegal:!1,_emitter:T,_top:_}}catch(n){if(n.message&&n.message.includes("Illegal"))return{language:e,value:oe(t),illegal:!0,relevance:0,_illegalBy:{message:n.message,index:P,context:t.slice(P-100,P+100),mode:n.mode,resultSoFar:S},_emitter:T};if(i)return{language:e,value:oe(t),illegal:!1,relevance:0,errorRaised:n,_emitter:T,_top:_};throw n}}function y(e,t){t=t||u.languages||Object.keys(r);const n=function(e){const t={value:oe(e),illegal:!1,relevance:0,_top:l,_emitter:new u.__emitter(u)};return t._emitter.addText(e),t}(e),a=t.filter(x).filter(S).map((t=>b(t,e,!1)));a.unshift(n);const o=a.sort(((e,t)=>{if(e.relevance!==t.relevance)return t.relevance-e.relevance;if(e.language&&t.language){if(x(e.language).supersetOf===t.language)return 1;if(x(t.language).supersetOf===e.language)return-1}return 0})),[i,s]=o,c=i;return c.secondBest=s,c}function w(e){let t=null;const n=function(e){let t=e.className+" ";t+=e.parentNode?e.parentNode.className:"";const n=u.languageDetectRe.exec(t);if(n){const t=x(n[1]);return t||(K(s.replace("{}",n[1])),K("Falling back to no-highlight mode for this block.",e)),t?n[1]:"no-highlight"}return t.split(/\s+/).find((e=>g(e)||x(e)))}(e);if(g(n))return;if(_("before:highlightElement",{el:e,language:n}),e.dataset.highlighted)return void console.log("Element previously highlighted. To highlight again, first unset `dataset.highlighted`.",e);if(e.children.length>0&&(u.ignoreUnescapedHTML||(console.warn("One of your code blocks includes unescaped HTML. This is a potentially serious security risk."),console.warn("https://github.com/highlightjs/highlight.js/wiki/security"),console.warn("The element with unescaped HTML:"),console.warn(e)),u.throwUnescapedHTML))throw new ae("One of your code blocks includes unescaped HTML.",e.innerHTML);t=e;const r=t.textContent,o=n?v(r,{language:n,ignoreIllegals:!0}):y(r);e.innerHTML=o.value,e.dataset.highlighted="yes",function(e,t,n){const r=t&&a[t]||n;e.classList.add("hljs"),e.classList.add(`language-${r}`)}(e,n,o.language),e.result={language:o.language,re:o.relevance,relevance:o.relevance},o.secondBest&&(e.secondBest={language:o.secondBest.language,relevance:o.secondBest.relevance}),_("after:highlightElement",{el:e,result:o,text:r})}let O=!1;function k(){if("loading"===document.readyState)return O||window.addEventListener("DOMContentLoaded",(function(){k()}),!1),void(O=!0);document.querySelectorAll(u.cssSelector).forEach(w)}function x(e){return e=(e||"").toLowerCase(),r[e]||r[a[e]]}function E(e,{languageName:t}){"string"==typeof e&&(e=[e]),e.forEach((e=>{a[e.toLowerCase()]=t}))}function S(e){const t=x(e);return t&&!t.disableAutodetect}function _(e,t){const n=e;o.forEach((function(e){e[n]&&e[n](t)}))}Object.assign(e,{highlight:v,highlightAuto:y,highlightAll:k,highlightElement:w,highlightBlock:function(e){return Y("10.7.0","highlightBlock will be removed entirely in v12.0"),Y("10.7.0","Please use highlightElement now."),w(e)},configure:function(e){u=ie(u,e)},initHighlighting:()=>{k(),Y("10.6.0","initHighlighting() deprecated.  Use highlightAll() now.")},initHighlightingOnLoad:function(){k(),Y("10.6.0","initHighlightingOnLoad() deprecated.  Use highlightAll() now.")},registerLanguage:function(t,n){let a=null;try{a=n(e)}catch(e){if(G("Language definition for '{}' could not be registered.".replace("{}",t)),!i)throw e;G(e),a=l}a.name||(a.name=t),r[t]=a,a.rawDefinition=n.bind(null,e),a.aliases&&E(a.aliases,{languageName:t})},unregisterLanguage:function(e){delete r[e];for(const t of Object.keys(a))a[t]===e&&delete a[t]},listLanguages:function(){return Object.keys(r)},getLanguage:x,registerAliases:E,autoDetection:S,inherit:ie,addPlugin:function(e){!function(e){e["before:highlightBlock"]&&!e["before:highlightElement"]&&(e["before:highlightElement"]=t=>{e["before:highlightBlock"](Object.assign({block:t.el},t))}),e["after:highlightBlock"]&&!e["after:highlightElement"]&&(e["after:highlightElement"]=t=>{e["after:highlightBlock"](Object.assign({block:t.el},t))})}(e),o.push(e)},removePlugin:function(e){const t=o.indexOf(e);-1!==t&&o.splice(t,1)}}),e.debugMode=function(){i=!1},e.safeMode=function(){i=!0},e.versionString="11.11.1",e.regex={concat:f,lookahead:d,either:m,optional:h,anyNumberOfTimes:p};for(const e in B)"object"==typeof B[e]&&t(B[e]);return Object.assign(e,B),e},ce=le({});ce.newInstance=()=>le({}),e.exports=ce,ce.HighlightJS=ce,ce.default=ce},8428:(e,t,n)=>{"use strict";n.d(t,{J:()=>o});var r=n(953);function a(e){return[e.screenX,e.screenY]}function o(){let e=(0,r.KR)([-1,-1]);return{wasMoved(t){let n=a(t);return(e.value[0]!==n[0]||e.value[1]!==n[1])&&(e.value=n,!0)},update(t){e.value=a(t)}}}},8470:(e,t,n)=>{"use strict";n.d(t,{Eo:()=>i,nk:()=>s});var r=n(3184);const a={parse:e=>{const t=(0,r.qg)(e);if("object"!=typeof t)throw Error("Invalid YAML object");return t},parseSafe(e,t){try{return a.parse(e)}catch(e){return"function"==typeof t?t(e):t}},stringify:r.As},o={parse:e=>{const t=JSON.parse(e);if("object"!=typeof t)throw Error("Invalid JSON object");return t},parseSafe(e,t){try{return o.parse(e)}catch(e){return"function"==typeof t?t(e):t}},stringify:e=>JSON.stringify(e)},i=e=>"string"==typeof e&&!!o.parseSafe(e,!1);function s(e){const t=e.trim();if("{"!==t[0]&&"["!==t[0])return e;try{return JSON.stringify(JSON.parse(e),null,2)}catch{return e}}},8547:(e,t,n)=>{"use strict";n.d(t,{A:()=>s});var r=n(641),a=n(33),o=n(6918);const i={key:0,class:"leading-normal text-c-2 text-sm text-pretty"},s=(0,r.pM)({__name:"DeleteSidebarListElement",props:{variableName:{},warningMessage:{}},emits:["close","delete"],setup(e,{emit:t}){const n=e,s=t,l=(0,r.EW)((()=>n.variableName.length>18?n.variableName.slice(0,18)+"…":n.variableName));return(e,t)=>((0,r.uX)(),(0,r.Wv)(o.A,{danger:"",label:`Delete ${l.value}`,onCancel:t[0]||(t[0]=e=>s("close")),onSubmit:t[1]||(t[1]=e=>s("delete"))},{default:(0,r.k6)((()=>[e.warningMessage?((0,r.uX)(),(0,r.CE)("p",i,(0,a.v_)(e.warningMessage),1)):(0,r.Q3)("",!0)])),_:1},8,["label"]))}})},8560:(e,t,n)=>{"use strict";const{asciiLowercase:r,solelyContainsHTTPTokenCodePoints:a,soleyContainsHTTPQuotedStringTokenCodePoints:o}=n(2661);e.exports=class{constructor(e){this._map=e}get size(){return this._map.size}get(e){return e=r(String(e)),this._map.get(e)}has(e){return e=r(String(e)),this._map.has(e)}set(e,t){if(e=r(String(e)),t=String(t),!a(e))throw new Error(`Invalid MIME type parameter name "${e}": only HTTP token code points are valid.`);if(!o(t))throw new Error(`Invalid MIME type parameter value "${t}": only HTTP quoted-string token code points are valid.`);return this._map.set(e,t)}clear(){this._map.clear()}delete(e){return e=r(String(e)),this._map.delete(e)}forEach(e,t){this._map.forEach(e,t)}keys(){return this._map.keys()}values(){return this._map.values()}entries(){return this._map.entries()}[Symbol.iterator](){return this._map[Symbol.iterator]()}}},8619:(e,t,n)=>{"use strict";n.d(t,{m:()=>a});const r=/[ \t\n\f\r]/g;function a(e){return"object"==typeof e?"text"===e.type&&o(e.value):o(e)}function o(e){return""===e.replace(r,"")}},8627:(e,t,n)=>{"use strict";n.d(t,{R:()=>r});class r{constructor(e,t){this.attribute=t,this.property=e}}r.prototype.attribute="",r.prototype.booleanish=!1,r.prototype.boolean=!1,r.prototype.commaOrSpaceSeparated=!1,r.prototype.commaSeparated=!1,r.prototype.defined=!1,r.prototype.mustUseProperty=!1,r.prototype.number=!1,r.prototype.overloadedBoolean=!1,r.prototype.property="",r.prototype.spaceSeparated=!1,r.prototype.space=void 0},8663:(e,t,n)=>{"use strict";n.d(t,{i:()=>o});var r=n(641),a=n(562);function o({container:e,accept:t,walk:n,enabled:o}){(0,r.nT)((()=>{let r=e.value;if(!r||void 0!==o&&!o.value)return;let i=(0,a.T)(e);if(!i)return;let s=Object.assign((e=>t(e)),{acceptNode:t}),l=i.createTreeWalker(r,NodeFilter.SHOW_ELEMENT,s,!1);for(;l.nextNode();)n(l.currentNode)}))}},8697:(e,t,n)=>{"use strict";n.d(t,{A:()=>K});var r=n(6932),a=n(7862),o=n(8845);const i=/["&'<>`]/g,s=/[\uD800-\uDBFF][\uDC00-\uDFFF]/g,l=/[\x01-\t\v\f\x0E-\x1F\x7F\x81\x8D\x8F\x90\x9D\xA0-\uFFFF]/g,c=/[|\\{}()[\]^$+*?.]/g,u=new WeakMap;const d=/[\dA-Fa-f]/,p=/\d/,h=["AElig","AMP","Aacute","Acirc","Agrave","Aring","Atilde","Auml","COPY","Ccedil","ETH","Eacute","Ecirc","Egrave","Euml","GT","Iacute","Icirc","Igrave","Iuml","LT","Ntilde","Oacute","Ocirc","Ograve","Oslash","Otilde","Ouml","QUOT","REG","THORN","Uacute","Ucirc","Ugrave","Uuml","Yacute","aacute","acirc","acute","aelig","agrave","amp","aring","atilde","auml","brvbar","ccedil","cedil","cent","copy","curren","deg","divide","eacute","ecirc","egrave","eth","euml","frac12","frac14","frac34","gt","iacute","icirc","iexcl","igrave","iquest","iuml","laquo","lt","macr","micro","middot","nbsp","not","ntilde","oacute","ocirc","ograve","ordf","ordm","oslash","otilde","ouml","para","plusmn","pound","quot","raquo","reg","sect","shy","sup1","sup2","sup3","szlig","thorn","times","uacute","ucirc","ugrave","uml","uuml","yacute","yen","yuml"],f={nbsp:" ",iexcl:"¡",cent:"¢",pound:"£",curren:"¤",yen:"¥",brvbar:"¦",sect:"§",uml:"¨",copy:"©",ordf:"ª",laquo:"«",not:"¬",shy:"­",reg:"®",macr:"¯",deg:"°",plusmn:"±",sup2:"²",sup3:"³",acute:"´",micro:"µ",para:"¶",middot:"·",cedil:"¸",sup1:"¹",ordm:"º",raquo:"»",frac14:"¼",frac12:"½",frac34:"¾",iquest:"¿",Agrave:"À",Aacute:"Á",Acirc:"Â",Atilde:"Ã",Auml:"Ä",Aring:"Å",AElig:"Æ",Ccedil:"Ç",Egrave:"È",Eacute:"É",Ecirc:"Ê",Euml:"Ë",Igrave:"Ì",Iacute:"Í",Icirc:"Î",Iuml:"Ï",ETH:"Ð",Ntilde:"Ñ",Ograve:"Ò",Oacute:"Ó",Ocirc:"Ô",Otilde:"Õ",Ouml:"Ö",times:"×",Oslash:"Ø",Ugrave:"Ù",Uacute:"Ú",Ucirc:"Û",Uuml:"Ü",Yacute:"Ý",THORN:"Þ",szlig:"ß",agrave:"à",aacute:"á",acirc:"â",atilde:"ã",auml:"ä",aring:"å",aelig:"æ",ccedil:"ç",egrave:"è",eacute:"é",ecirc:"ê",euml:"ë",igrave:"ì",iacute:"í",icirc:"î",iuml:"ï",eth:"ð",ntilde:"ñ",ograve:"ò",oacute:"ó",ocirc:"ô",otilde:"õ",ouml:"ö",divide:"÷",oslash:"ø",ugrave:"ù",uacute:"ú",ucirc:"û",uuml:"ü",yacute:"ý",thorn:"þ",yuml:"ÿ",fnof:"ƒ",Alpha:"Α",Beta:"Β",Gamma:"Γ",Delta:"Δ",Epsilon:"Ε",Zeta:"Ζ",Eta:"Η",Theta:"Θ",Iota:"Ι",Kappa:"Κ",Lambda:"Λ",Mu:"Μ",Nu:"Ν",Xi:"Ξ",Omicron:"Ο",Pi:"Π",Rho:"Ρ",Sigma:"Σ",Tau:"Τ",Upsilon:"Υ",Phi:"Φ",Chi:"Χ",Psi:"Ψ",Omega:"Ω",alpha:"α",beta:"β",gamma:"γ",delta:"δ",epsilon:"ε",zeta:"ζ",eta:"η",theta:"θ",iota:"ι",kappa:"κ",lambda:"λ",mu:"μ",nu:"ν",xi:"ξ",omicron:"ο",pi:"π",rho:"ρ",sigmaf:"ς",sigma:"σ",tau:"τ",upsilon:"υ",phi:"φ",chi:"χ",psi:"ψ",omega:"ω",thetasym:"ϑ",upsih:"ϒ",piv:"ϖ",bull:"•",hellip:"…",prime:"′",Prime:"″",oline:"‾",frasl:"⁄",weierp:"℘",image:"ℑ",real:"ℜ",trade:"™",alefsym:"ℵ",larr:"←",uarr:"↑",rarr:"→",darr:"↓",harr:"↔",crarr:"↵",lArr:"⇐",uArr:"⇑",rArr:"⇒",dArr:"⇓",hArr:"⇔",forall:"∀",part:"∂",exist:"∃",empty:"∅",nabla:"∇",isin:"∈",notin:"∉",ni:"∋",prod:"∏",sum:"∑",minus:"−",lowast:"∗",radic:"√",prop:"∝",infin:"∞",ang:"∠",and:"∧",or:"∨",cap:"∩",cup:"∪",int:"∫",there4:"∴",sim:"∼",cong:"≅",asymp:"≈",ne:"≠",equiv:"≡",le:"≤",ge:"≥",sub:"⊂",sup:"⊃",nsub:"⊄",sube:"⊆",supe:"⊇",oplus:"⊕",otimes:"⊗",perp:"⊥",sdot:"⋅",lceil:"⌈",rceil:"⌉",lfloor:"⌊",rfloor:"⌋",lang:"〈",rang:"〉",loz:"◊",spades:"♠",clubs:"♣",hearts:"♥",diams:"♦",quot:'"',amp:"&",lt:"<",gt:">",OElig:"Œ",oelig:"œ",Scaron:"Š",scaron:"š",Yuml:"Ÿ",circ:"ˆ",tilde:"˜",ensp:" ",emsp:" ",thinsp:" ",zwnj:"‌",zwj:"‍",lrm:"‎",rlm:"‏",ndash:"–",mdash:"—",lsquo:"‘",rsquo:"’",sbquo:"‚",ldquo:"“",rdquo:"”",bdquo:"„",dagger:"†",Dagger:"‡",permil:"‰",lsaquo:"‹",rsaquo:"›",euro:"€"},m=["cent","copy","divide","gt","lt","not","para","times"],g={}.hasOwnProperty,v={};let b;for(b in f)g.call(f,b)&&(v[f[b]]=b);const y=/[^\dA-Za-z]/;function w(e,t,n){let r,a=function(e,t,n){const r="&#x"+e.toString(16).toUpperCase();return n&&t&&!d.test(String.fromCharCode(t))?r:r+";"}(e,t,n.omitOptionalSemicolons);if((n.useNamedReferences||n.useShortestReferences)&&(r=function(e,t,n,r){const a=String.fromCharCode(e);if(g.call(v,a)){const e=v[a],o="&"+e;return n&&h.includes(e)&&!m.includes(e)&&(!r||t&&61!==t&&y.test(String.fromCharCode(t)))?o:o+";"}return""}(e,t,n.omitOptionalSemicolons,n.attribute)),(n.useShortestReferences||!r)&&n.useShortestReferences){const r=function(e,t,n){const r="&#"+String(e);return n&&t&&!p.test(String.fromCharCode(t))?r:r+";"}(e,t,n.omitOptionalSemicolons);r.length<a.length&&(a=r)}return r&&(!n.useShortestReferences||r.length<a.length)?r:a}function O(e,t){return function(e,t){return e=e.replace(t.subset?function(e){let t=u.get(e);return t||(t=function(e){const t=[];let n=-1;for(;++n<e.length;)t.push(e[n].replace(c,"\\$&"));return new RegExp("(?:"+t.join("|")+")","g")}(e),u.set(e,t)),t}(t.subset):i,n),t.subset||t.escapeOnly?e:e.replace(s,(function(e,n,r){return t.format(1024*(e.charCodeAt(0)-55296)+e.charCodeAt(1)-56320+65536,r.charCodeAt(n+2),t)})).replace(l,n);function n(e,n,r){return t.format(e.charCodeAt(0),r.charCodeAt(n+1),t)}}(e,Object.assign({format:w},t))}const k=/^>|^->|<!--|-->|--!>|<!-$/g,x=[">"],E=["<",">"];var S=n(8256),_=n(2282),C=n(1775),T=n(2485),A=n(8619);const R=M(1),P=M(-1),D=[];function M(e){return function(t,n,r){const a=t?t.children:D;let o=(n||0)+e,i=a[o];if(!r)for(;i&&(0,A.m)(i);)o+=e,i=a[o];return i}}const $={}.hasOwnProperty;function N(e){return function(t,n,r){return $.call(e,t.tagName)&&e[t.tagName](t,n,r)}}const I=N({body:function(e,t,n){const r=R(n,t);return!r||"comment"!==r.type},caption:L,colgroup:L,dd:function(e,t,n){const r=R(n,t);return!r||"element"===r.type&&("dt"===r.tagName||"dd"===r.tagName)},dt:function(e,t,n){const r=R(n,t);return Boolean(r&&"element"===r.type&&("dt"===r.tagName||"dd"===r.tagName))},head:L,html:function(e,t,n){const r=R(n,t);return!r||"comment"!==r.type},li:function(e,t,n){const r=R(n,t);return!r||"element"===r.type&&"li"===r.tagName},optgroup:function(e,t,n){const r=R(n,t);return!r||"element"===r.type&&"optgroup"===r.tagName},option:function(e,t,n){const r=R(n,t);return!r||"element"===r.type&&("option"===r.tagName||"optgroup"===r.tagName)},p:function(e,t,n){const r=R(n,t);return r?"element"===r.type&&("address"===r.tagName||"article"===r.tagName||"aside"===r.tagName||"blockquote"===r.tagName||"details"===r.tagName||"div"===r.tagName||"dl"===r.tagName||"fieldset"===r.tagName||"figcaption"===r.tagName||"figure"===r.tagName||"footer"===r.tagName||"form"===r.tagName||"h1"===r.tagName||"h2"===r.tagName||"h3"===r.tagName||"h4"===r.tagName||"h5"===r.tagName||"h6"===r.tagName||"header"===r.tagName||"hgroup"===r.tagName||"hr"===r.tagName||"main"===r.tagName||"menu"===r.tagName||"nav"===r.tagName||"ol"===r.tagName||"p"===r.tagName||"pre"===r.tagName||"section"===r.tagName||"table"===r.tagName||"ul"===r.tagName):!n||!("element"===n.type&&("a"===n.tagName||"audio"===n.tagName||"del"===n.tagName||"ins"===n.tagName||"map"===n.tagName||"noscript"===n.tagName||"video"===n.tagName))},rp:B,rt:B,tbody:function(e,t,n){const r=R(n,t);return!r||"element"===r.type&&("tbody"===r.tagName||"tfoot"===r.tagName)},td:F,tfoot:function(e,t,n){return!R(n,t)},th:F,thead:function(e,t,n){const r=R(n,t);return Boolean(r&&"element"===r.type&&("tbody"===r.tagName||"tfoot"===r.tagName))},tr:function(e,t,n){const r=R(n,t);return!r||"element"===r.type&&"tr"===r.tagName}});function L(e,t,n){const r=R(n,t,!0);return!r||"comment"!==r.type&&!("text"===r.type&&(0,A.m)(r.value.charAt(0)))}function B(e,t,n){const r=R(n,t);return!r||"element"===r.type&&("rp"===r.tagName||"rt"===r.tagName)}function F(e,t,n){const r=R(n,t);return!r||"element"===r.type&&("td"===r.tagName||"th"===r.tagName)}const Q=N({body:function(e){const t=R(e,-1,!0);return!(t&&("comment"===t.type||"text"===t.type&&(0,A.m)(t.value.charAt(0))||"element"===t.type&&("meta"===t.tagName||"link"===t.tagName||"script"===t.tagName||"style"===t.tagName||"template"===t.tagName)))},colgroup:function(e,t,n){const r=P(n,t),a=R(e,-1,!0);return!(n&&r&&"element"===r.type&&"colgroup"===r.tagName&&I(r,n.children.indexOf(r),n))&&Boolean(a&&"element"===a.type&&"col"===a.tagName)},head:function(e){const t=new Set;for(const n of e.children)if("element"===n.type&&("base"===n.tagName||"title"===n.tagName)){if(t.has(n.tagName))return!1;t.add(n.tagName)}const n=e.children[0];return!n||"element"===n.type},html:function(e){const t=R(e,-1);return!t||"comment"!==t.type},tbody:function(e,t,n){const r=P(n,t),a=R(e,-1);return(!n||!r||"element"!==r.type||"thead"!==r.tagName&&"tbody"!==r.tagName||!I(r,n.children.indexOf(r),n))&&Boolean(a&&"element"===a.type&&"tr"===a.tagName)}}),U={name:[["\t\n\f\r &/=>".split(""),"\t\n\f\r \"&'/=>`".split("")],["\0\t\n\f\r \"&'/<=>".split(""),"\0\t\n\f\r \"&'/<=>`".split("")]],unquoted:[["\t\n\f\r &>".split(""),"\0\t\n\f\r \"&'<=>`".split("")],["\0\t\n\f\r \"&'<=>`".split(""),"\0\t\n\f\r \"&'<=>`".split("")]],single:[["&'".split(""),"\"&'`".split("")],["\0&'".split(""),"\0\"&'`".split("")]],double:[['"&'.split(""),"\"&'`".split("")],['\0"&'.split(""),"\0\"&'`".split("")]]};function z(e,t,n){const r=(0,C.I)(e.schema,t),a=e.settings.allowParseErrors&&"html"===e.schema.space?0:1,o=e.settings.allowDangerousCharacters?0:1;let i,s=e.quote;if(!r.overloadedBoolean||n!==r.attribute&&""!==n?!r.boolean&&!r.overloadedBoolean||"string"==typeof n&&n!==r.attribute&&""!==n||(n=Boolean(n)):n=!0,null==n||!1===n||"number"==typeof n&&Number.isNaN(n))return"";const l=O(r.attribute,Object.assign({},e.settings.characterReferences,{subset:U.name[a][o]}));return!0===n?l:(n=Array.isArray(n)?(r.commaSeparated?_.A:T.A)(n,{padLeft:!e.settings.tightCommaSeparatedLists}):String(n),e.settings.collapseEmptyAttributes&&!n?l:(e.settings.preferUnquoted&&(i=O(n,Object.assign({},e.settings.characterReferences,{attribute:!0,subset:U.unquoted[a][o]}))),i!==n&&(e.settings.quoteSmart&&(0,S.D)(n,s)>(0,S.D)(n,e.alternative)&&(s=e.alternative),i=s+O(n,Object.assign({},e.settings.characterReferences,{subset:("'"===s?U.single:U.double)[a][o],attribute:!0}))+s),l+(i?"="+i:i)))}const j=["<","&"];function W(e,t,n,r){return!n||"element"!==n.type||"script"!==n.tagName&&"style"!==n.tagName?O(e.value,Object.assign({},r.settings.characterReferences,{subset:j})):e.value}const X=(0,o.A)("type",{invalid:function(e){throw new Error("Expected node, not `"+e+"`")},unknown:function(e){throw new Error("Cannot compile unknown node `"+e.type+"`")},handlers:{comment:function(e,t,n,r){return r.settings.bogusComments?"<?"+O(e.value,Object.assign({},r.settings.characterReferences,{subset:x}))+">":"\x3c!--"+e.value.replace(k,(function(e){return O(e,Object.assign({},r.settings.characterReferences,{subset:E}))}))+"--\x3e"},doctype:function(e,t,n,r){return"<!"+(r.settings.upperDoctype?"DOCTYPE":"doctype")+(r.settings.tightDoctype?"":" ")+"html>"},element:function(e,t,n,r){const o=r.schema,i="svg"!==o.space&&r.settings.omitOptionalTags;let s="svg"===o.space?r.settings.closeEmptyElements:r.settings.voids.includes(e.tagName.toLowerCase());const l=[];let c;"html"===o.space&&"svg"===e.tagName&&(r.schema=a.JW);const u=function(e,t){const n=[];let r,a=-1;if(t)for(r in t)if(null!==t[r]&&void 0!==t[r]){const a=z(e,r,t[r]);a&&n.push(a)}for(;++a<n.length;){const t=e.settings.tightAttributes?n[a].charAt(n[a].length-1):void 0;a!==n.length-1&&'"'!==t&&"'"!==t&&(n[a]+=" ")}return n.join("")}(r,e.properties),d=r.all("html"===o.space&&"template"===e.tagName?e.content:e);return r.schema=o,d&&(s=!1),!u&&i&&Q(e,t,n)||(l.push("<",e.tagName,u?" "+u:""),s&&("svg"===o.space||r.settings.closeSelfClosing)&&(c=u.charAt(u.length-1),(!r.settings.tightSelfClosing||"/"===c||c&&'"'!==c&&"'"!==c)&&l.push(" "),l.push("/")),l.push(">")),l.push(d),s||i&&I(e,t,n)||l.push("</"+e.tagName+">"),l.join("")},raw:function(e,t,n,r){return r.settings.allowDangerousHtml?e.value:W(e,0,n,r)},root:function(e,t,n,r){return r.all(e)},text:W}}),q={},V={},H=[];function Z(e,t,n){return X(e,t,n,this)}function G(e){const t=[],n=e&&e.children||H;let r=-1;for(;++r<n.length;)t[r]=this.one(n[r],r,e);return t.join("")}function K(e){const t={...this.data("settings"),...e};this.compiler=function(e){return function(e,t){const n=t||q,o=n.quote||'"',i='"'===o?"'":'"';if('"'!==o&&"'"!==o)throw new Error("Invalid quote `"+o+"`, expected `'` or `\"`");return{one:Z,all:G,settings:{omitOptionalTags:n.omitOptionalTags||!1,allowParseErrors:n.allowParseErrors||!1,allowDangerousCharacters:n.allowDangerousCharacters||!1,quoteSmart:n.quoteSmart||!1,preferUnquoted:n.preferUnquoted||!1,tightAttributes:n.tightAttributes||!1,upperDoctype:n.upperDoctype||!1,tightDoctype:n.tightDoctype||!1,bogusComments:n.bogusComments||!1,tightCommaSeparatedLists:n.tightCommaSeparatedLists||!1,tightSelfClosing:n.tightSelfClosing||!1,collapseEmptyAttributes:n.collapseEmptyAttributes||!1,allowDangerousHtml:n.allowDangerousHtml||!1,voids:n.voids||r.b,characterReferences:n.characterReferences||V,closeSelfClosing:n.closeSelfClosing||!1,closeEmptyElements:n.closeEmptyElements||!1},schema:"svg"===n.space?a.JW:a.qy,quote:o,alternative:i}.one(Array.isArray(e)?{type:"root",children:e}:e,void 0,void 0)}(e,t)}}},8803:(e,t,n)=>{"use strict";n.d(t,{V:()=>o});var r=n(4476),a=n(2071);const o=r.z.object({uid:a.B.brand(),name:r.z.string().optional().default("Default Environment"),color:r.z.string().optional().default("#0082D0"),value:r.z.string().default(""),isDefault:r.z.boolean().optional()})},8826:(e,t,n)=>{"use strict";n.d(t,{T:()=>h});var r=n(5946);const a=function(e,t){if(void 0!==t&&"string"!=typeof t)throw new TypeError('"ext" argument must be a string');c(e);let n,r=0,a=-1,o=e.length;if(void 0===t||0===t.length||t.length>e.length){for(;o--;)if(47===e.codePointAt(o)){if(n){r=o+1;break}}else a<0&&(n=!0,a=o+1);return a<0?"":e.slice(r,a)}if(t===e)return"";let i=-1,s=t.length-1;for(;o--;)if(47===e.codePointAt(o)){if(n){r=o+1;break}}else i<0&&(n=!0,i=o+1),s>-1&&(e.codePointAt(o)===t.codePointAt(s--)?s<0&&(a=o):(s=-1,a=i));return r===a?a=i:a<0&&(a=e.length),e.slice(r,a)},o=function(e){if(c(e),0===e.length)return".";let t,n=-1,r=e.length;for(;--r;)if(47===e.codePointAt(r)){if(t){n=r;break}}else t||(t=!0);return n<0?47===e.codePointAt(0)?"/":".":1===n&&47===e.codePointAt(0)?"//":e.slice(0,n)},i=function(e){c(e);let t,n=e.length,r=-1,a=0,o=-1,i=0;for(;n--;){const s=e.codePointAt(n);if(47!==s)r<0&&(t=!0,r=n+1),46===s?o<0?o=n:1!==i&&(i=1):o>-1&&(i=-1);else if(t){a=n+1;break}}return o<0||r<0||0===i||1===i&&o===r-1&&o===a+1?"":e.slice(o,r)},s=function(...e){let t,n=-1;for(;++n<e.length;)c(e[n]),e[n]&&(t=void 0===t?e[n]:t+"/"+e[n]);return void 0===t?".":function(e){c(e);const t=47===e.codePointAt(0);let n=function(e,t){let n,r,a="",o=0,i=-1,s=0,l=-1;for(;++l<=e.length;){if(l<e.length)n=e.codePointAt(l);else{if(47===n)break;n=47}if(47===n){if(i===l-1||1===s);else if(i!==l-1&&2===s){if(a.length<2||2!==o||46!==a.codePointAt(a.length-1)||46!==a.codePointAt(a.length-2))if(a.length>2){if(r=a.lastIndexOf("/"),r!==a.length-1){r<0?(a="",o=0):(a=a.slice(0,r),o=a.length-1-a.lastIndexOf("/")),i=l,s=0;continue}}else if(a.length>0){a="",o=0,i=l,s=0;continue}t&&(a=a.length>0?a+"/..":"..",o=2)}else a.length>0?a+="/"+e.slice(i+1,l):a=e.slice(i+1,l),o=l-i-1;i=l,s=0}else 46===n&&s>-1?s++:s=-1}return a}(e,!t);return 0!==n.length||t||(n="."),n.length>0&&47===e.codePointAt(e.length-1)&&(n+="/"),t?"/"+n:n}(t)},l="/";function c(e){if("string"!=typeof e)throw new TypeError("Path must be a string. Received "+JSON.stringify(e))}const u=function(){return"/"};function d(e){return Boolean(null!==e&&"object"==typeof e&&"href"in e&&e.href&&"protocol"in e&&e.protocol&&void 0===e.auth)}const p=["history","path","basename","stem","extname","dirname"];class h{constructor(e){let t;t=e?d(e)?{path:e}:"string"==typeof e||function(e){return Boolean(e&&"object"==typeof e&&"byteLength"in e&&"byteOffset"in e)}(e)?{value:e}:e:{},this.cwd="cwd"in t?"":u(),this.data={},this.history=[],this.messages=[],this.value,this.map,this.result,this.stored;let n,r=-1;for(;++r<p.length;){const e=p[r];e in t&&void 0!==t[e]&&null!==t[e]&&(this[e]="history"===e?[...t[e]]:t[e])}for(n in t)p.includes(n)||(this[n]=t[n])}get basename(){return"string"==typeof this.path?a(this.path):void 0}set basename(e){m(e,"basename"),f(e,"basename"),this.path=s(this.dirname||"",e)}get dirname(){return"string"==typeof this.path?o(this.path):void 0}set dirname(e){g(this.basename,"dirname"),this.path=s(e||"",this.basename)}get extname(){return"string"==typeof this.path?i(this.path):void 0}set extname(e){if(f(e,"extname"),g(this.dirname,"extname"),e){if(46!==e.codePointAt(0))throw new Error("`extname` must start with `.`");if(e.includes(".",1))throw new Error("`extname` cannot contain multiple dots")}this.path=s(this.dirname,this.stem+(e||""))}get path(){return this.history[this.history.length-1]}set path(e){d(e)&&(e=function(e){if("string"==typeof e)e=new URL(e);else if(!d(e)){const t=new TypeError('The "path" argument must be of type string or an instance of URL. Received `'+e+"`");throw t.code="ERR_INVALID_ARG_TYPE",t}if("file:"!==e.protocol){const e=new TypeError("The URL must be of scheme file");throw e.code="ERR_INVALID_URL_SCHEME",e}return function(e){if(""!==e.hostname){const e=new TypeError('File URL host must be "localhost" or empty on darwin');throw e.code="ERR_INVALID_FILE_URL_HOST",e}const t=e.pathname;let n=-1;for(;++n<t.length;)if(37===t.codePointAt(n)&&50===t.codePointAt(n+1)){const e=t.codePointAt(n+2);if(70===e||102===e){const e=new TypeError("File URL path must not include encoded / characters");throw e.code="ERR_INVALID_FILE_URL_PATH",e}}return decodeURIComponent(t)}(e)}(e)),m(e,"path"),this.path!==e&&this.history.push(e)}get stem(){return"string"==typeof this.path?a(this.path,this.extname):void 0}set stem(e){m(e,"stem"),f(e,"stem"),this.path=s(this.dirname||"",e+(this.extname||""))}fail(e,t,n){const r=this.message(e,t,n);throw r.fatal=!0,r}info(e,t,n){const r=this.message(e,t,n);return r.fatal=void 0,r}message(e,t,n){const a=new r.o(e,t,n);return this.path&&(a.name=this.path+":"+a.name,a.file=this.path),a.fatal=!1,this.messages.push(a),a}toString(e){return void 0===this.value?"":"string"==typeof this.value?this.value:new TextDecoder(e||void 0).decode(this.value)}}function f(e,t){if(e&&e.includes(l))throw new Error("`"+t+"` cannot be a path: did not expect `"+l+"`")}function m(e,t){if(!e)throw new Error("`"+t+"` cannot be empty")}function g(e,t){if(!e)throw new Error("Setting `"+t+"` requires `path` to be set too")}},8839:(e,t,n)=>{"use strict";n.d(t,{A:()=>l});var r=n(641),a=n(33),o=n(953),i=n(1266),s=n(2402);const l=(0,r.pM)({__name:"ScalarHotkey",props:{hotkey:{},modifier:{}},setup(e){const t=e,{cx:n}=(0,i.AX)(),l=(0,r.EW)((()=>t.modifier||"meta")),c=(0,r.EW)((()=>`${"meta"===l.value?(0,s.U)()?"⌘":"^":l.value} ${t.hotkey}`));return(e,t)=>((0,r.uX)(),(0,r.CE)("div",(0,a._B)((0,r.Ng)((0,o.R1)(n)("border-b-3 inline-block overflow-hidden rounded border-1/2 text-xxs rounded-b px-1 font-medium uppercase"))),(0,a.v_)(c.value),17))}})},8845:(e,t,n)=>{"use strict";n.d(t,{A:()=>a});const r={}.hasOwnProperty;function a(e,t){const n=t||{};function a(t,...n){let o=a.invalid;const i=a.handlers;if(t&&r.call(t,e)){const n=String(t[e]);o=r.call(i,n)?i[n]:a.unknown}if(o)return o.call(this,t,...n)}return a.handlers=n.handlers||{},a.invalid=n.invalid,a.unknown=n.unknown,a}},8869:(e,t,n)=>{"use strict";n.d(t,{B:()=>s,I:()=>l});var r,a=n(641);let o=Symbol("headlessui.useid"),i=0;const s=null!=(r=a.Bi)?r:function(){return a.WQ(o,(()=>""+ ++i))()};function l(e){a.Gt(o,e)}},9005:(e,t,n)=>{"use strict";n.d(t,{A:()=>s});var r=n(641),a=n(3526);const o={},i={class:"*:border-t-1/2 first:*:border-t-0 xl:*:border-t-0 xl:*:border-l-1/2 xl:first:*:border-l-0 flex xl:flex-row flex-col custom-scroll pr-0"},s=(0,a.A)(o,[["render",function(e,t){return(0,r.uX)(),(0,r.CE)("div",i,[(0,r.RG)(e.$slots,"default")])}]])},9035:(e,t,n)=>{"use strict";n.d(t,{NZ:()=>rn,Lz:()=>Ya,vB:()=>De,Z9:()=>er,xO:()=>tn,Ei:()=>xc,qF:()=>km,BK:()=>Sm});let r=[],a=[];function o(e){if(e<768)return!1;for(let t=0,n=r.length;;){let o=t+n>>1;if(e<r[o])n=o;else{if(!(e>=a[o]))return!0;t=o+1}if(t==n)return!1}}function i(e){return e>=127462&&e<=127487}function s(e,t,n=!0,r=!0){return(n?l:c)(e,t,r)}function l(e,t,n){if(t==e.length)return t;t&&d(e.charCodeAt(t))&&p(e.charCodeAt(t-1))&&t--;let r=u(e,t);for(t+=h(r);t<e.length;){let a=u(e,t);if(8205==r||8205==a||n&&o(a))t+=h(a),r=a;else{if(!i(a))break;{let n=0,r=t-2;for(;r>=0&&i(u(e,r));)n++,r-=2;if(n%2==0)break;t+=2}}}return t}function c(e,t,n){for(;t>0;){let r=l(e,t-2,n);if(r<t)return r;t--}return 0}function u(e,t){let n=e.charCodeAt(t);if(!p(n)||t+1==e.length)return n;let r=e.charCodeAt(t+1);return d(r)?r-56320+(n-55296<<10)+65536:n}function d(e){return e>=56320&&e<57344}function p(e){return e>=55296&&e<56320}function h(e){return e<65536?1:2}(()=>{let e="lc,34,7n,7,7b,19,,,,2,,2,,,20,b,1c,l,g,,2t,7,2,6,2,2,,4,z,,u,r,2j,b,1m,9,9,,o,4,,9,,3,,5,17,3,3b,f,,w,1j,,,,4,8,4,,3,7,a,2,t,,1m,,,,2,4,8,,9,,a,2,q,,2,2,1l,,4,2,4,2,2,3,3,,u,2,3,,b,2,1l,,4,5,,2,4,,k,2,m,6,,,1m,,,2,,4,8,,7,3,a,2,u,,1n,,,,c,,9,,14,,3,,1l,3,5,3,,4,7,2,b,2,t,,1m,,2,,2,,3,,5,2,7,2,b,2,s,2,1l,2,,,2,4,8,,9,,a,2,t,,20,,4,,2,3,,,8,,29,,2,7,c,8,2q,,2,9,b,6,22,2,r,,,,,,1j,e,,5,,2,5,b,,10,9,,2u,4,,6,,2,2,2,p,2,4,3,g,4,d,,2,2,6,,f,,jj,3,qa,3,t,3,t,2,u,2,1s,2,,7,8,,2,b,9,,19,3,3b,2,y,,3a,3,4,2,9,,6,3,63,2,2,,1m,,,7,,,,,2,8,6,a,2,,1c,h,1r,4,1c,7,,,5,,14,9,c,2,w,4,2,2,,3,1k,,,2,3,,,3,1m,8,2,2,48,3,,d,,7,4,,6,,3,2,5i,1m,,5,ek,,5f,x,2da,3,3x,,2o,w,fe,6,2x,2,n9w,4,,a,w,2,28,2,7k,,3,,4,,p,2,5,,47,2,q,i,d,,12,8,p,b,1a,3,1c,,2,4,2,2,13,,1v,6,2,2,2,2,c,,8,,1b,,1f,,,3,2,2,5,2,,,16,2,8,,6m,,2,,4,,fn4,,kh,g,g,g,a6,2,gt,,6a,,45,5,1ae,3,,2,5,4,14,3,4,,4l,2,fx,4,ar,2,49,b,4w,,1i,f,1k,3,1d,4,2,2,1x,3,10,5,,8,1q,,c,2,1g,9,a,4,2,,2n,3,2,,,2,6,,4g,,3,8,l,2,1l,2,,,,,m,,e,7,3,5,5f,8,2,3,,,n,,29,,2,6,,,2,,,2,,2,6j,,2,4,6,2,,2,r,2,2d,8,2,,,2,2y,,,,2,6,,,2t,3,2,4,,5,77,9,,2,6t,,a,2,,,4,,40,4,2,2,4,,w,a,14,6,2,4,8,,9,6,2,3,1a,d,,2,ba,7,,6,,,2a,m,2,7,,2,,2,3e,6,3,,,2,,7,,,20,2,3,,,,9n,2,f0b,5,1n,7,t4,,1r,4,29,,f5k,2,43q,,,3,4,5,8,8,2,7,u,4,44,3,1iz,1j,4,1e,8,,e,,m,5,,f,11s,7,,h,2,7,,2,,5,79,7,c5,4,15s,7,31,7,240,5,gx7k,2o,3k,6o".split(",").map((e=>e?parseInt(e,36):1));for(let t=0,n=0;t<e.length;t++)(t%2?a:r).push(n+=e[t])})();class f{lineAt(e){if(e<0||e>this.length)throw new RangeError(`Invalid position ${e} in document of length ${this.length}`);return this.lineInner(e,!1,1,0)}line(e){if(e<1||e>this.lines)throw new RangeError(`Invalid line number ${e} in ${this.lines}-line document`);return this.lineInner(e,!0,1,0)}replace(e,t,n){[e,t]=x(this,e,t);let r=[];return this.decompose(0,e,r,2),n.length&&n.decompose(0,n.length,r,3),this.decompose(t,this.length,r,1),g.from(r,this.length-(t-e)+n.length)}append(e){return this.replace(this.length,this.length,e)}slice(e,t=this.length){[e,t]=x(this,e,t);let n=[];return this.decompose(e,t,n,0),g.from(n,t-e)}eq(e){if(e==this)return!0;if(e.length!=this.length||e.lines!=this.lines)return!1;let t=this.scanIdentical(e,1),n=this.length-this.scanIdentical(e,-1),r=new y(this),a=new y(e);for(let e=t,o=t;;){if(r.next(e),a.next(e),e=0,r.lineBreak!=a.lineBreak||r.done!=a.done||r.value!=a.value)return!1;if(o+=r.value.length,r.done||o>=n)return!0}}iter(e=1){return new y(this,e)}iterRange(e,t=this.length){return new w(this,e,t)}iterLines(e,t){let n;if(null==e)n=this.iter();else{null==t&&(t=this.lines+1);let r=this.line(e).from;n=this.iterRange(r,Math.max(r,t==this.lines+1?this.length:t<=1?0:this.line(t-1).to))}return new O(n)}toString(){return this.sliceString(0)}toJSON(){let e=[];return this.flatten(e),e}constructor(){}static of(e){if(0==e.length)throw new RangeError("A document must have at least one line");return 1!=e.length||e[0]?e.length<=32?new m(e):g.from(m.split(e,[])):f.empty}}class m extends f{constructor(e,t=function(e){let t=-1;for(let n of e)t+=n.length+1;return t}(e)){super(),this.text=e,this.length=t}get lines(){return this.text.length}get children(){return null}lineInner(e,t,n,r){for(let a=0;;a++){let o=this.text[a],i=r+o.length;if((t?n:i)>=e)return new k(r,i,n,o);r=i+1,n++}}decompose(e,t,n,r){let a=e<=0&&t>=this.length?this:new m(b(this.text,e,t),Math.min(t,this.length)-Math.max(0,e));if(1&r){let e=n.pop(),t=v(a.text,e.text.slice(),0,a.length);if(t.length<=32)n.push(new m(t,e.length+a.length));else{let e=t.length>>1;n.push(new m(t.slice(0,e)),new m(t.slice(e)))}}else n.push(a)}replace(e,t,n){if(!(n instanceof m))return super.replace(e,t,n);[e,t]=x(this,e,t);let r=v(this.text,v(n.text,b(this.text,0,e)),t),a=this.length+n.length-(t-e);return r.length<=32?new m(r,a):g.from(m.split(r,[]),a)}sliceString(e,t=this.length,n="\n"){[e,t]=x(this,e,t);let r="";for(let a=0,o=0;a<=t&&o<this.text.length;o++){let i=this.text[o],s=a+i.length;a>e&&o&&(r+=n),e<s&&t>a&&(r+=i.slice(Math.max(0,e-a),t-a)),a=s+1}return r}flatten(e){for(let t of this.text)e.push(t)}scanIdentical(){return 0}static split(e,t){let n=[],r=-1;for(let a of e)n.push(a),r+=a.length+1,32==n.length&&(t.push(new m(n,r)),n=[],r=-1);return r>-1&&t.push(new m(n,r)),t}}class g extends f{constructor(e,t){super(),this.children=e,this.length=t,this.lines=0;for(let t of e)this.lines+=t.lines}lineInner(e,t,n,r){for(let a=0;;a++){let o=this.children[a],i=r+o.length,s=n+o.lines-1;if((t?s:i)>=e)return o.lineInner(e,t,n,r);r=i+1,n=s+1}}decompose(e,t,n,r){for(let a=0,o=0;o<=t&&a<this.children.length;a++){let i=this.children[a],s=o+i.length;if(e<=s&&t>=o){let a=r&((o<=e?1:0)|(s>=t?2:0));o>=e&&s<=t&&!a?n.push(i):i.decompose(e-o,t-o,n,a)}o=s+1}}replace(e,t,n){if([e,t]=x(this,e,t),n.lines<this.lines)for(let r=0,a=0;r<this.children.length;r++){let o=this.children[r],i=a+o.length;if(e>=a&&t<=i){let s=o.replace(e-a,t-a,n),l=this.lines-o.lines+s.lines;if(s.lines<l>>4&&s.lines>l>>6){let a=this.children.slice();return a[r]=s,new g(a,this.length-(t-e)+n.length)}return super.replace(a,i,s)}a=i+1}return super.replace(e,t,n)}sliceString(e,t=this.length,n="\n"){[e,t]=x(this,e,t);let r="";for(let a=0,o=0;a<this.children.length&&o<=t;a++){let i=this.children[a],s=o+i.length;o>e&&a&&(r+=n),e<s&&t>o&&(r+=i.sliceString(e-o,t-o,n)),o=s+1}return r}flatten(e){for(let t of this.children)t.flatten(e)}scanIdentical(e,t){if(!(e instanceof g))return 0;let n=0,[r,a,o,i]=t>0?[0,0,this.children.length,e.children.length]:[this.children.length-1,e.children.length-1,-1,-1];for(;;r+=t,a+=t){if(r==o||a==i)return n;let s=this.children[r],l=e.children[a];if(s!=l)return n+s.scanIdentical(l,t);n+=s.length+1}}static from(e,t=e.reduce(((e,t)=>e+t.length+1),-1)){let n=0;for(let t of e)n+=t.lines;if(n<32){let n=[];for(let t of e)t.flatten(n);return new m(n,t)}let r=Math.max(32,n>>5),a=r<<1,o=r>>1,i=[],s=0,l=-1,c=[];function u(e){let t;if(e.lines>a&&e instanceof g)for(let t of e.children)u(t);else e.lines>o&&(s>o||!s)?(d(),i.push(e)):e instanceof m&&s&&(t=c[c.length-1])instanceof m&&e.lines+t.lines<=32?(s+=e.lines,l+=e.length+1,c[c.length-1]=new m(t.text.concat(e.text),t.length+1+e.length)):(s+e.lines>r&&d(),s+=e.lines,l+=e.length+1,c.push(e))}function d(){0!=s&&(i.push(1==c.length?c[0]:g.from(c,l)),l=-1,s=c.length=0)}for(let t of e)u(t);return d(),1==i.length?i[0]:new g(i,t)}}function v(e,t,n=0,r=1e9){for(let a=0,o=0,i=!0;o<e.length&&a<=r;o++){let s=e[o],l=a+s.length;l>=n&&(l>r&&(s=s.slice(0,r-a)),a<n&&(s=s.slice(n-a)),i?(t[t.length-1]+=s,i=!1):t.push(s)),a=l+1}return t}function b(e,t,n){return v(e,[""],t,n)}f.empty=new m([""],0);class y{constructor(e,t=1){this.dir=t,this.done=!1,this.lineBreak=!1,this.value="",this.nodes=[e],this.offsets=[t>0?1:(e instanceof m?e.text.length:e.children.length)<<1]}nextInner(e,t){for(this.done=this.lineBreak=!1;;){let n=this.nodes.length-1,r=this.nodes[n],a=this.offsets[n],o=a>>1,i=r instanceof m?r.text.length:r.children.length;if(o==(t>0?i:0)){if(0==n)return this.done=!0,this.value="",this;t>0&&this.offsets[n-1]++,this.nodes.pop(),this.offsets.pop()}else if((1&a)==(t>0?0:1)){if(this.offsets[n]+=t,0==e)return this.lineBreak=!0,this.value="\n",this;e--}else if(r instanceof m){let a=r.text[o+(t<0?-1:0)];if(this.offsets[n]+=t,a.length>Math.max(0,e))return this.value=0==e?a:t>0?a.slice(e):a.slice(0,a.length-e),this;e-=a.length}else{let a=r.children[o+(t<0?-1:0)];e>a.length?(e-=a.length,this.offsets[n]+=t):(t<0&&this.offsets[n]--,this.nodes.push(a),this.offsets.push(t>0?1:(a instanceof m?a.text.length:a.children.length)<<1))}}}next(e=0){return e<0&&(this.nextInner(-e,-this.dir),e=this.value.length),this.nextInner(e,this.dir)}}class w{constructor(e,t,n){this.value="",this.done=!1,this.cursor=new y(e,t>n?-1:1),this.pos=t>n?e.length:0,this.from=Math.min(t,n),this.to=Math.max(t,n)}nextInner(e,t){if(t<0?this.pos<=this.from:this.pos>=this.to)return this.value="",this.done=!0,this;e+=Math.max(0,t<0?this.pos-this.to:this.from-this.pos);let n=t<0?this.pos-this.from:this.to-this.pos;e>n&&(e=n),n-=e;let{value:r}=this.cursor.next(e);return this.pos+=(r.length+e)*t,this.value=r.length<=n?r:t<0?r.slice(r.length-n):r.slice(0,n),this.done=!this.value,this}next(e=0){return e<0?e=Math.max(e,this.from-this.pos):e>0&&(e=Math.min(e,this.to-this.pos)),this.nextInner(e,this.cursor.dir)}get lineBreak(){return this.cursor.lineBreak&&""!=this.value}}class O{constructor(e){this.inner=e,this.afterBreak=!0,this.value="",this.done=!1}next(e=0){let{done:t,lineBreak:n,value:r}=this.inner.next(e);return t&&this.afterBreak?(this.value="",this.afterBreak=!1):t?(this.done=!0,this.value=""):n?this.afterBreak?this.value="":(this.afterBreak=!0,this.next()):(this.value=r,this.afterBreak=!1),this}get lineBreak(){return!1}}"undefined"!=typeof Symbol&&(f.prototype[Symbol.iterator]=function(){return this.iter()},y.prototype[Symbol.iterator]=w.prototype[Symbol.iterator]=O.prototype[Symbol.iterator]=function(){return this});class k{constructor(e,t,n,r){this.from=e,this.to=t,this.number=n,this.text=r}get length(){return this.to-this.from}}function x(e,t,n){return[t=Math.max(0,Math.min(e.length,t)),Math.max(t,Math.min(e.length,n))]}function E(e,t,n=!0,r=!0){return s(e,t,n,r)}function S(e,t){let n=e.charCodeAt(t);if(!((r=n)>=55296&&r<56320&&t+1!=e.length))return n;var r;let a=e.charCodeAt(t+1);return function(e){return e>=56320&&e<57344}(a)?a-56320+(n-55296<<10)+65536:n}function _(e){return e<=65535?String.fromCharCode(e):(e-=65536,String.fromCharCode(55296+(e>>10),56320+(1023&e)))}function C(e){return e<65536?1:2}const T=/\r\n?|\n/;var A=function(e){return e[e.Simple=0]="Simple",e[e.TrackDel=1]="TrackDel",e[e.TrackBefore=2]="TrackBefore",e[e.TrackAfter=3]="TrackAfter",e}(A||(A={}));class R{constructor(e){this.sections=e}get length(){let e=0;for(let t=0;t<this.sections.length;t+=2)e+=this.sections[t];return e}get newLength(){let e=0;for(let t=0;t<this.sections.length;t+=2){let n=this.sections[t+1];e+=n<0?this.sections[t]:n}return e}get empty(){return 0==this.sections.length||2==this.sections.length&&this.sections[1]<0}iterGaps(e){for(let t=0,n=0,r=0;t<this.sections.length;){let a=this.sections[t++],o=this.sections[t++];o<0?(e(n,r,a),r+=a):r+=o,n+=a}}iterChangedRanges(e,t=!1){$(this,e,t)}get invertedDesc(){let e=[];for(let t=0;t<this.sections.length;){let n=this.sections[t++],r=this.sections[t++];r<0?e.push(n,r):e.push(r,n)}return new R(e)}composeDesc(e){return this.empty?e:e.empty?this:I(this,e)}mapDesc(e,t=!1){return e.empty?this:N(this,e,t)}mapPos(e,t=-1,n=A.Simple){let r=0,a=0;for(let o=0;o<this.sections.length;){let i=this.sections[o++],s=this.sections[o++],l=r+i;if(s<0){if(l>e)return a+(e-r);a+=i}else{if(n!=A.Simple&&l>=e&&(n==A.TrackDel&&r<e&&l>e||n==A.TrackBefore&&r<e||n==A.TrackAfter&&l>e))return null;if(l>e||l==e&&t<0&&!i)return e==r||t<0?a:a+s;a+=s}r=l}if(e>r)throw new RangeError(`Position ${e} is out of range for changeset of length ${r}`);return a}touchesRange(e,t=e){for(let n=0,r=0;n<this.sections.length&&r<=t;){let a=r+this.sections[n++];if(this.sections[n++]>=0&&r<=t&&a>=e)return!(r<e&&a>t)||"cover";r=a}return!1}toString(){let e="";for(let t=0;t<this.sections.length;){let n=this.sections[t++],r=this.sections[t++];e+=(e?" ":"")+n+(r>=0?":"+r:"")}return e}toJSON(){return this.sections}static fromJSON(e){if(!Array.isArray(e)||e.length%2||e.some((e=>"number"!=typeof e)))throw new RangeError("Invalid JSON representation of ChangeDesc");return new R(e)}static create(e){return new R(e)}}class P extends R{constructor(e,t){super(e),this.inserted=t}apply(e){if(this.length!=e.length)throw new RangeError("Applying change set to a document with the wrong length");return $(this,((t,n,r,a,o)=>e=e.replace(r,r+(n-t),o)),!1),e}mapDesc(e,t=!1){return N(this,e,t,!0)}invert(e){let t=this.sections.slice(),n=[];for(let r=0,a=0;r<t.length;r+=2){let o=t[r],i=t[r+1];if(i>=0){t[r]=i,t[r+1]=o;let s=r>>1;for(;n.length<s;)n.push(f.empty);n.push(o?e.slice(a,a+o):f.empty)}a+=o}return new P(t,n)}compose(e){return this.empty?e:e.empty?this:I(this,e,!0)}map(e,t=!1){return e.empty?this:N(this,e,t,!0)}iterChanges(e,t=!1){$(this,e,t)}get desc(){return R.create(this.sections)}filter(e){let t=[],n=[],r=[],a=new L(this);e:for(let o=0,i=0;;){let s=o==e.length?1e9:e[o++];for(;i<s||i==s&&0==a.len;){if(a.done)break e;let e=Math.min(a.len,s-i);D(r,e,-1);let o=-1==a.ins?-1:0==a.off?a.ins:0;D(t,e,o),o>0&&M(n,t,a.text),a.forward(e),i+=e}let l=e[o++];for(;i<l;){if(a.done)break e;let e=Math.min(a.len,l-i);D(t,e,-1),D(r,e,-1==a.ins?-1:0==a.off?a.ins:0),a.forward(e),i+=e}}return{changes:new P(t,n),filtered:R.create(r)}}toJSON(){let e=[];for(let t=0;t<this.sections.length;t+=2){let n=this.sections[t],r=this.sections[t+1];r<0?e.push(n):0==r?e.push([n]):e.push([n].concat(this.inserted[t>>1].toJSON()))}return e}static of(e,t,n){let r=[],a=[],o=0,i=null;function s(e=!1){if(!e&&!r.length)return;o<t&&D(r,t-o,-1);let n=new P(r,a);i=i?i.compose(n.map(i)):n,r=[],a=[],o=0}return function e(l){if(Array.isArray(l))for(let t of l)e(t);else if(l instanceof P){if(l.length!=t)throw new RangeError(`Mismatched change set length (got ${l.length}, expected ${t})`);s(),i=i?i.compose(l.map(i)):l}else{let{from:e,to:i=e,insert:c}=l;if(e>i||e<0||i>t)throw new RangeError(`Invalid change range ${e} to ${i} (in doc of length ${t})`);let u=c?"string"==typeof c?f.of(c.split(n||T)):c:f.empty,d=u.length;if(e==i&&0==d)return;e<o&&s(),e>o&&D(r,e-o,-1),D(r,i-e,d),M(a,r,u),o=i}}(e),s(!i),i}static empty(e){return new P(e?[e,-1]:[],[])}static fromJSON(e){if(!Array.isArray(e))throw new RangeError("Invalid JSON representation of ChangeSet");let t=[],n=[];for(let r=0;r<e.length;r++){let a=e[r];if("number"==typeof a)t.push(a,-1);else{if(!Array.isArray(a)||"number"!=typeof a[0]||a.some(((e,t)=>t&&"string"!=typeof e)))throw new RangeError("Invalid JSON representation of ChangeSet");if(1==a.length)t.push(a[0],0);else{for(;n.length<r;)n.push(f.empty);n[r]=f.of(a.slice(1)),t.push(a[0],n[r].length)}}}return new P(t,n)}static createSet(e,t){return new P(e,t)}}function D(e,t,n,r=!1){if(0==t&&n<=0)return;let a=e.length-2;a>=0&&n<=0&&n==e[a+1]?e[a]+=t:a>=0&&0==t&&0==e[a]?e[a+1]+=n:r?(e[a]+=t,e[a+1]+=n):e.push(t,n)}function M(e,t,n){if(0==n.length)return;let r=t.length-2>>1;if(r<e.length)e[e.length-1]=e[e.length-1].append(n);else{for(;e.length<r;)e.push(f.empty);e.push(n)}}function $(e,t,n){let r=e.inserted;for(let a=0,o=0,i=0;i<e.sections.length;){let s=e.sections[i++],l=e.sections[i++];if(l<0)a+=s,o+=s;else{let c=a,u=o,d=f.empty;for(;c+=s,u+=l,l&&r&&(d=d.append(r[i-2>>1])),!(n||i==e.sections.length||e.sections[i+1]<0);)s=e.sections[i++],l=e.sections[i++];t(a,c,o,u,d),a=c,o=u}}}function N(e,t,n,r=!1){let a=[],o=r?[]:null,i=new L(e),s=new L(t);for(let e=-1;;){if(i.done&&s.len||s.done&&i.len)throw new Error("Mismatched change set lengths");if(-1==i.ins&&-1==s.ins){let e=Math.min(i.len,s.len);D(a,e,-1),i.forward(e),s.forward(e)}else if(s.ins>=0&&(i.ins<0||e==i.i||0==i.off&&(s.len<i.len||s.len==i.len&&!n))){let t=s.len;for(D(a,s.ins,-1);t;){let n=Math.min(i.len,t);i.ins>=0&&e<i.i&&i.len<=n&&(D(a,0,i.ins),o&&M(o,a,i.text),e=i.i),i.forward(n),t-=n}s.next()}else{if(!(i.ins>=0)){if(i.done&&s.done)return o?P.createSet(a,o):R.create(a);throw new Error("Mismatched change set lengths")}{let t=0,n=i.len;for(;n;)if(-1==s.ins){let e=Math.min(n,s.len);t+=e,n-=e,s.forward(e)}else{if(!(0==s.ins&&s.len<n))break;n-=s.len,s.next()}D(a,t,e<i.i?i.ins:0),o&&e<i.i&&M(o,a,i.text),e=i.i,i.forward(i.len-n)}}}}function I(e,t,n=!1){let r=[],a=n?[]:null,o=new L(e),i=new L(t);for(let e=!1;;){if(o.done&&i.done)return a?P.createSet(r,a):R.create(r);if(0==o.ins)D(r,o.len,0,e),o.next();else if(0!=i.len||i.done){if(o.done||i.done)throw new Error("Mismatched change set lengths");{let t=Math.min(o.len2,i.len),n=r.length;if(-1==o.ins){let n=-1==i.ins?-1:i.off?0:i.ins;D(r,t,n,e),a&&n&&M(a,r,i.text)}else-1==i.ins?(D(r,o.off?0:o.len,t,e),a&&M(a,r,o.textBit(t))):(D(r,o.off?0:o.len,i.off?0:i.ins,e),a&&!i.off&&M(a,r,i.text));e=(o.ins>t||i.ins>=0&&i.len>t)&&(e||r.length>n),o.forward2(t),i.forward(t)}}else D(r,0,i.ins,e),a&&M(a,r,i.text),i.next()}}class L{constructor(e){this.set=e,this.i=0,this.next()}next(){let{sections:e}=this.set;this.i<e.length?(this.len=e[this.i++],this.ins=e[this.i++]):(this.len=0,this.ins=-2),this.off=0}get done(){return-2==this.ins}get len2(){return this.ins<0?this.len:this.ins}get text(){let{inserted:e}=this.set,t=this.i-2>>1;return t>=e.length?f.empty:e[t]}textBit(e){let{inserted:t}=this.set,n=this.i-2>>1;return n>=t.length&&!e?f.empty:t[n].slice(this.off,null==e?void 0:this.off+e)}forward(e){e==this.len?this.next():(this.len-=e,this.off+=e)}forward2(e){-1==this.ins?this.forward(e):e==this.ins?this.next():(this.ins-=e,this.off+=e)}}class B{constructor(e,t,n){this.from=e,this.to=t,this.flags=n}get anchor(){return 32&this.flags?this.to:this.from}get head(){return 32&this.flags?this.from:this.to}get empty(){return this.from==this.to}get assoc(){return 8&this.flags?-1:16&this.flags?1:0}get bidiLevel(){let e=7&this.flags;return 7==e?null:e}get goalColumn(){let e=this.flags>>6;return 16777215==e?void 0:e}map(e,t=-1){let n,r;return this.empty?n=r=e.mapPos(this.from,t):(n=e.mapPos(this.from,1),r=e.mapPos(this.to,-1)),n==this.from&&r==this.to?this:new B(n,r,this.flags)}extend(e,t=e){if(e<=this.anchor&&t>=this.anchor)return F.range(e,t);let n=Math.abs(e-this.anchor)>Math.abs(t-this.anchor)?e:t;return F.range(this.anchor,n)}eq(e,t=!1){return!(this.anchor!=e.anchor||this.head!=e.head||t&&this.empty&&this.assoc!=e.assoc)}toJSON(){return{anchor:this.anchor,head:this.head}}static fromJSON(e){if(!e||"number"!=typeof e.anchor||"number"!=typeof e.head)throw new RangeError("Invalid JSON representation for SelectionRange");return F.range(e.anchor,e.head)}static create(e,t,n){return new B(e,t,n)}}class F{constructor(e,t){this.ranges=e,this.mainIndex=t}map(e,t=-1){return e.empty?this:F.create(this.ranges.map((n=>n.map(e,t))),this.mainIndex)}eq(e,t=!1){if(this.ranges.length!=e.ranges.length||this.mainIndex!=e.mainIndex)return!1;for(let n=0;n<this.ranges.length;n++)if(!this.ranges[n].eq(e.ranges[n],t))return!1;return!0}get main(){return this.ranges[this.mainIndex]}asSingle(){return 1==this.ranges.length?this:new F([this.main],0)}addRange(e,t=!0){return F.create([e].concat(this.ranges),t?0:this.mainIndex+1)}replaceRange(e,t=this.mainIndex){let n=this.ranges.slice();return n[t]=e,F.create(n,this.mainIndex)}toJSON(){return{ranges:this.ranges.map((e=>e.toJSON())),main:this.mainIndex}}static fromJSON(e){if(!e||!Array.isArray(e.ranges)||"number"!=typeof e.main||e.main>=e.ranges.length)throw new RangeError("Invalid JSON representation for EditorSelection");return new F(e.ranges.map((e=>B.fromJSON(e))),e.main)}static single(e,t=e){return new F([F.range(e,t)],0)}static create(e,t=0){if(0==e.length)throw new RangeError("A selection needs at least one range");for(let n=0,r=0;r<e.length;r++){let a=e[r];if(a.empty?a.from<=n:a.from<n)return F.normalized(e.slice(),t);n=a.to}return new F(e,t)}static cursor(e,t=0,n,r){return B.create(e,e,(0==t?0:t<0?8:16)|(null==n?7:Math.min(6,n))|(null!=r?r:16777215)<<6)}static range(e,t,n,r){let a=(null!=n?n:16777215)<<6|(null==r?7:Math.min(6,r));return t<e?B.create(t,e,48|a):B.create(e,t,(t>e?8:0)|a)}static normalized(e,t=0){let n=e[t];e.sort(((e,t)=>e.from-t.from)),t=e.indexOf(n);for(let n=1;n<e.length;n++){let r=e[n],a=e[n-1];if(r.empty?r.from<=a.to:r.from<a.to){let o=a.from,i=Math.max(r.to,a.to);n<=t&&t--,e.splice(--n,2,r.anchor>r.head?F.range(i,o):F.range(o,i))}}return new F(e,t)}}function Q(e,t){for(let n of e.ranges)if(n.to>t)throw new RangeError("Selection points outside of document")}let U=0;class z{constructor(e,t,n,r,a){this.combine=e,this.compareInput=t,this.compare=n,this.isStatic=r,this.id=U++,this.default=e([]),this.extensions="function"==typeof a?a(this):a}get reader(){return this}static define(e={}){return new z(e.combine||(e=>e),e.compareInput||((e,t)=>e===t),e.compare||(e.combine?(e,t)=>e===t:j),!!e.static,e.enables)}of(e){return new W([],this,0,e)}compute(e,t){if(this.isStatic)throw new Error("Can't compute a static facet");return new W(e,this,1,t)}computeN(e,t){if(this.isStatic)throw new Error("Can't compute a static facet");return new W(e,this,2,t)}from(e,t){return t||(t=e=>e),this.compute([e],(n=>t(n.field(e))))}}function j(e,t){return e==t||e.length==t.length&&e.every(((e,n)=>e===t[n]))}class W{constructor(e,t,n,r){this.dependencies=e,this.facet=t,this.type=n,this.value=r,this.id=U++}dynamicSlot(e){var t;let n=this.value,r=this.facet.compareInput,a=this.id,o=e[a]>>1,i=2==this.type,s=!1,l=!1,c=[];for(let n of this.dependencies)"doc"==n?s=!0:"selection"==n?l=!0:1&(null!==(t=e[n.id])&&void 0!==t?t:1)||c.push(e[n.id]);return{create:e=>(e.values[o]=n(e),1),update(e,t){if(s&&t.docChanged||l&&(t.docChanged||t.selection)||q(e,c)){let t=n(e);if(i?!X(t,e.values[o],r):!r(t,e.values[o]))return e.values[o]=t,1}return 0},reconfigure:(e,t)=>{let s,l=t.config.address[a];if(null!=l){let a=re(t,l);if(this.dependencies.every((n=>n instanceof z?t.facet(n)===e.facet(n):!(n instanceof Z)||t.field(n,!1)==e.field(n,!1)))||(i?X(s=n(e),a,r):r(s=n(e),a)))return e.values[o]=a,0}else s=n(e);return e.values[o]=s,1}}}}function X(e,t,n){if(e.length!=t.length)return!1;for(let r=0;r<e.length;r++)if(!n(e[r],t[r]))return!1;return!0}function q(e,t){let n=!1;for(let r of t)1&ne(e,r)&&(n=!0);return n}function V(e,t,n){let r=n.map((t=>e[t.id])),a=n.map((e=>e.type)),o=r.filter((e=>!(1&e))),i=e[t.id]>>1;function s(e){let n=[];for(let t=0;t<r.length;t++){let o=re(e,r[t]);if(2==a[t])for(let e of o)n.push(e);else n.push(o)}return t.combine(n)}return{create(e){for(let t of r)ne(e,t);return e.values[i]=s(e),1},update(e,n){if(!q(e,o))return 0;let r=s(e);return t.compare(r,e.values[i])?0:(e.values[i]=r,1)},reconfigure(e,a){let o=q(e,r),l=a.config.facets[t.id],c=a.facet(t);if(l&&!o&&j(n,l))return e.values[i]=c,0;let u=s(e);return t.compare(u,c)?(e.values[i]=c,0):(e.values[i]=u,1)}}}const H=z.define({static:!0});class Z{constructor(e,t,n,r,a){this.id=e,this.createF=t,this.updateF=n,this.compareF=r,this.spec=a,this.provides=void 0}static define(e){let t=new Z(U++,e.create,e.update,e.compare||((e,t)=>e===t),e);return e.provide&&(t.provides=e.provide(t)),t}create(e){let t=e.facet(H).find((e=>e.field==this));return((null==t?void 0:t.create)||this.createF)(e)}slot(e){let t=e[this.id]>>1;return{create:e=>(e.values[t]=this.create(e),1),update:(e,n)=>{let r=e.values[t],a=this.updateF(r,n);return this.compareF(r,a)?0:(e.values[t]=a,1)},reconfigure:(e,n)=>{let r,a=e.facet(H),o=n.facet(H);return(r=a.find((e=>e.field==this)))&&r!=o.find((e=>e.field==this))?(e.values[t]=r.create(e),1):null!=n.config.address[this.id]?(e.values[t]=n.field(this),0):(e.values[t]=this.create(e),1)}}}init(e){return[this,H.of({field:this,create:e})]}get extension(){return this}}function G(e){return t=>new Y(t,e)}const K={highest:G(0),high:G(1),default:G(2),low:G(3),lowest:G(4)};class Y{constructor(e,t){this.inner=e,this.prec=t}}class J{of(e){return new ee(this,e)}reconfigure(e){return J.reconfigure.of({compartment:this,extension:e})}get(e){return e.config.compartments.get(this)}}class ee{constructor(e,t){this.compartment=e,this.inner=t}}class te{constructor(e,t,n,r,a,o){for(this.base=e,this.compartments=t,this.dynamicSlots=n,this.address=r,this.staticValues=a,this.facets=o,this.statusTemplate=[];this.statusTemplate.length<n.length;)this.statusTemplate.push(0)}staticFacet(e){let t=this.address[e.id];return null==t?e.default:this.staticValues[t>>1]}static resolve(e,t,n){let r=[],a=Object.create(null),o=new Map;for(let n of function(e,t,n){let r=[[],[],[],[],[]],a=new Map;return function e(o,i){let s=a.get(o);if(null!=s){if(s<=i)return;let e=r[s].indexOf(o);e>-1&&r[s].splice(e,1),o instanceof ee&&n.delete(o.compartment)}if(a.set(o,i),Array.isArray(o))for(let t of o)e(t,i);else if(o instanceof ee){if(n.has(o.compartment))throw new RangeError("Duplicate use of compartment in extensions");let r=t.get(o.compartment)||o.inner;n.set(o.compartment,r),e(r,i)}else if(o instanceof Y)e(o.inner,o.prec);else if(o instanceof Z)r[i].push(o),o.provides&&e(o.provides,i);else if(o instanceof W)r[i].push(o),o.facet.extensions&&e(o.facet.extensions,2);else{let t=o.extension;if(!t)throw new Error(`Unrecognized extension value in extension set (${o}). This sometimes happens because multiple instances of @codemirror/state are loaded, breaking instanceof checks.`);e(t,i)}}(e,2),r.reduce(((e,t)=>e.concat(t)))}(e,t,o))n instanceof Z?r.push(n):(a[n.facet.id]||(a[n.facet.id]=[])).push(n);let i=Object.create(null),s=[],l=[];for(let e of r)i[e.id]=l.length<<1,l.push((t=>e.slot(t)));let c=null==n?void 0:n.config.facets;for(let e in a){let t=a[e],r=t[0].facet,o=c&&c[e]||[];if(t.every((e=>0==e.type)))if(i[r.id]=s.length<<1|1,j(o,t))s.push(n.facet(r));else{let e=r.combine(t.map((e=>e.value)));s.push(n&&r.compare(e,n.facet(r))?n.facet(r):e)}else{for(let e of t)0==e.type?(i[e.id]=s.length<<1|1,s.push(e.value)):(i[e.id]=l.length<<1,l.push((t=>e.dynamicSlot(t))));i[r.id]=l.length<<1,l.push((e=>V(e,r,t)))}}let u=l.map((e=>e(i)));return new te(e,o,u,i,s,a)}}function ne(e,t){if(1&t)return 2;let n=t>>1,r=e.status[n];if(4==r)throw new Error("Cyclic dependency between fields and/or facets");if(2&r)return r;e.status[n]=4;let a=e.computeSlot(e,e.config.dynamicSlots[n]);return e.status[n]=2|a}function re(e,t){return 1&t?e.config.staticValues[t>>1]:e.values[t>>1]}const ae=z.define(),oe=z.define({combine:e=>e.some((e=>e)),static:!0}),ie=z.define({combine:e=>e.length?e[0]:void 0,static:!0}),se=z.define(),le=z.define(),ce=z.define(),ue=z.define({combine:e=>!!e.length&&e[0]});class de{constructor(e,t){this.type=e,this.value=t}static define(){return new pe}}class pe{of(e){return new de(this,e)}}class he{constructor(e){this.map=e}of(e){return new fe(this,e)}}class fe{constructor(e,t){this.type=e,this.value=t}map(e){let t=this.type.map(this.value,e);return void 0===t?void 0:t==this.value?this:new fe(this.type,t)}is(e){return this.type==e}static define(e={}){return new he(e.map||(e=>e))}static mapEffects(e,t){if(!e.length)return e;let n=[];for(let r of e){let e=r.map(t);e&&n.push(e)}return n}}fe.reconfigure=fe.define(),fe.appendConfig=fe.define();class me{constructor(e,t,n,r,a,o){this.startState=e,this.changes=t,this.selection=n,this.effects=r,this.annotations=a,this.scrollIntoView=o,this._doc=null,this._state=null,n&&Q(n,t.newLength),a.some((e=>e.type==me.time))||(this.annotations=a.concat(me.time.of(Date.now())))}static create(e,t,n,r,a,o){return new me(e,t,n,r,a,o)}get newDoc(){return this._doc||(this._doc=this.changes.apply(this.startState.doc))}get newSelection(){return this.selection||this.startState.selection.map(this.changes)}get state(){return this._state||this.startState.applyTransaction(this),this._state}annotation(e){for(let t of this.annotations)if(t.type==e)return t.value}get docChanged(){return!this.changes.empty}get reconfigured(){return this.startState.config!=this.state.config}isUserEvent(e){let t=this.annotation(me.userEvent);return!(!t||!(t==e||t.length>e.length&&t.slice(0,e.length)==e&&"."==t[e.length]))}}function ge(e,t){let n=[];for(let r=0,a=0;;){let o,i;if(r<e.length&&(a==t.length||t[a]>=e[r]))o=e[r++],i=e[r++];else{if(!(a<t.length))return n;o=t[a++],i=t[a++]}!n.length||n[n.length-1]<o?n.push(o,i):n[n.length-1]<i&&(n[n.length-1]=i)}}function ve(e,t,n){var r;let a,o,i;return n?(a=t.changes,o=P.empty(t.changes.length),i=e.changes.compose(t.changes)):(a=t.changes.map(e.changes),o=e.changes.mapDesc(t.changes,!0),i=e.changes.compose(a)),{changes:i,selection:t.selection?t.selection.map(o):null===(r=e.selection)||void 0===r?void 0:r.map(a),effects:fe.mapEffects(e.effects,a).concat(fe.mapEffects(t.effects,o)),annotations:e.annotations.length?e.annotations.concat(t.annotations):t.annotations,scrollIntoView:e.scrollIntoView||t.scrollIntoView}}function be(e,t,n){let r=t.selection,a=Oe(t.annotations);return t.userEvent&&(a=a.concat(me.userEvent.of(t.userEvent))),{changes:t.changes instanceof P?t.changes:P.of(t.changes||[],n,e.facet(ie)),selection:r&&(r instanceof F?r:F.single(r.anchor,r.head)),effects:Oe(t.effects),annotations:a,scrollIntoView:!!t.scrollIntoView}}function ye(e,t,n){let r=be(e,t.length?t[0]:{},e.doc.length);t.length&&!1===t[0].filter&&(n=!1);for(let a=1;a<t.length;a++){!1===t[a].filter&&(n=!1);let o=!!t[a].sequential;r=ve(r,be(e,t[a],o?r.changes.newLength:e.doc.length),o)}let a=me.create(e,r.changes,r.selection,r.effects,r.annotations,r.scrollIntoView);return function(e){let t=e.startState,n=t.facet(ce),r=e;for(let a=n.length-1;a>=0;a--){let o=n[a](e);o&&Object.keys(o).length&&(r=ve(r,be(t,o,e.changes.newLength),!0))}return r==e?e:me.create(t,e.changes,e.selection,r.effects,r.annotations,r.scrollIntoView)}(n?function(e){let t=e.startState,n=!0;for(let r of t.facet(se)){let t=r(e);if(!1===t){n=!1;break}Array.isArray(t)&&(n=!0===n?t:ge(n,t))}if(!0!==n){let r,a;if(!1===n)a=e.changes.invertedDesc,r=P.empty(t.doc.length);else{let t=e.changes.filter(n);r=t.changes,a=t.filtered.mapDesc(t.changes).invertedDesc}e=me.create(t,r,e.selection&&e.selection.map(a),fe.mapEffects(e.effects,a),e.annotations,e.scrollIntoView)}let r=t.facet(le);for(let n=r.length-1;n>=0;n--){let a=r[n](e);e=a instanceof me?a:Array.isArray(a)&&1==a.length&&a[0]instanceof me?a[0]:ye(t,Oe(a),!1)}return e}(a):a)}me.time=de.define(),me.userEvent=de.define(),me.addToHistory=de.define(),me.remote=de.define();const we=[];function Oe(e){return null==e?we:Array.isArray(e)?e:[e]}var ke=function(e){return e[e.Word=0]="Word",e[e.Space=1]="Space",e[e.Other=2]="Other",e}(ke||(ke={}));const xe=/[\u00df\u0587\u0590-\u05f4\u0600-\u06ff\u3040-\u309f\u30a0-\u30ff\u3400-\u4db5\u4e00-\u9fcc\uac00-\ud7af]/;let Ee;try{Ee=new RegExp("[\\p{Alphabetic}\\p{Number}_]","u")}catch(e){}class Se{constructor(e,t,n,r,a,o){this.config=e,this.doc=t,this.selection=n,this.values=r,this.status=e.statusTemplate.slice(),this.computeSlot=a,o&&(o._state=this);for(let e=0;e<this.config.dynamicSlots.length;e++)ne(this,e<<1);this.computeSlot=null}field(e,t=!0){let n=this.config.address[e.id];if(null!=n)return ne(this,n),re(this,n);if(t)throw new RangeError("Field is not present in this state")}update(...e){return ye(this,e,!0)}applyTransaction(e){let t,n=this.config,{base:r,compartments:a}=n;for(let t of e.effects)t.is(J.reconfigure)?(n&&(a=new Map,n.compartments.forEach(((e,t)=>a.set(t,e))),n=null),a.set(t.value.compartment,t.value.extension)):t.is(fe.reconfigure)?(n=null,r=t.value):t.is(fe.appendConfig)&&(n=null,r=Oe(r).concat(t.value));n?t=e.startState.values.slice():(n=te.resolve(r,a,this),t=new Se(n,this.doc,this.selection,n.dynamicSlots.map((()=>null)),((e,t)=>t.reconfigure(e,this)),null).values);let o=e.startState.facet(oe)?e.newSelection:e.newSelection.asSingle();new Se(n,e.newDoc,o,t,((t,n)=>n.update(t,e)),e)}replaceSelection(e){return"string"==typeof e&&(e=this.toText(e)),this.changeByRange((t=>({changes:{from:t.from,to:t.to,insert:e},range:F.cursor(t.from+e.length)})))}changeByRange(e){let t=this.selection,n=e(t.ranges[0]),r=this.changes(n.changes),a=[n.range],o=Oe(n.effects);for(let n=1;n<t.ranges.length;n++){let i=e(t.ranges[n]),s=this.changes(i.changes),l=s.map(r);for(let e=0;e<n;e++)a[e]=a[e].map(l);let c=r.mapDesc(s,!0);a.push(i.range.map(c)),r=r.compose(l),o=fe.mapEffects(o,l).concat(fe.mapEffects(Oe(i.effects),c))}return{changes:r,selection:F.create(a,t.mainIndex),effects:o}}changes(e=[]){return e instanceof P?e:P.of(e,this.doc.length,this.facet(Se.lineSeparator))}toText(e){return f.of(e.split(this.facet(Se.lineSeparator)||T))}sliceDoc(e=0,t=this.doc.length){return this.doc.sliceString(e,t,this.lineBreak)}facet(e){let t=this.config.address[e.id];return null==t?e.default:(ne(this,t),re(this,t))}toJSON(e){let t={doc:this.sliceDoc(),selection:this.selection.toJSON()};if(e)for(let n in e){let r=e[n];r instanceof Z&&null!=this.config.address[r.id]&&(t[n]=r.spec.toJSON(this.field(e[n]),this))}return t}static fromJSON(e,t={},n){if(!e||"string"!=typeof e.doc)throw new RangeError("Invalid JSON representation for EditorState");let r=[];if(n)for(let t in n)if(Object.prototype.hasOwnProperty.call(e,t)){let a=n[t],o=e[t];r.push(a.init((e=>a.spec.fromJSON(o,e))))}return Se.create({doc:e.doc,selection:F.fromJSON(e.selection),extensions:t.extensions?r.concat([t.extensions]):r})}static create(e={}){let t=te.resolve(e.extensions||[],new Map),n=e.doc instanceof f?e.doc:f.of((e.doc||"").split(t.staticFacet(Se.lineSeparator)||T)),r=e.selection?e.selection instanceof F?e.selection:F.single(e.selection.anchor,e.selection.head):F.single(0);return Q(r,n.length),t.staticFacet(oe)||(r=r.asSingle()),new Se(t,n,r,t.dynamicSlots.map((()=>null)),((e,t)=>t.create(e)),null)}get tabSize(){return this.facet(Se.tabSize)}get lineBreak(){return this.facet(Se.lineSeparator)||"\n"}get readOnly(){return this.facet(ue)}phrase(e,...t){for(let t of this.facet(Se.phrases))if(Object.prototype.hasOwnProperty.call(t,e)){e=t[e];break}return t.length&&(e=e.replace(/\$(\$|\d*)/g,((e,n)=>{if("$"==n)return"$";let r=+(n||1);return!r||r>t.length?e:t[r-1]}))),e}languageDataAt(e,t,n=-1){let r=[];for(let a of this.facet(ae))for(let o of a(this,t,n))Object.prototype.hasOwnProperty.call(o,e)&&r.push(o[e]);return r}charCategorizer(e){return t=this.languageDataAt("wordChars",e).join(""),e=>{if(!/\S/.test(e))return ke.Space;if(function(e){if(Ee)return Ee.test(e);for(let t=0;t<e.length;t++){let n=e[t];if(/\w/.test(n)||n>""&&(n.toUpperCase()!=n.toLowerCase()||xe.test(n)))return!0}return!1}(e))return ke.Word;for(let n=0;n<t.length;n++)if(e.indexOf(t[n])>-1)return ke.Word;return ke.Other};var t}wordAt(e){let{text:t,from:n,length:r}=this.doc.lineAt(e),a=this.charCategorizer(e),o=e-n,i=e-n;for(;o>0;){let e=E(t,o,!1);if(a(t.slice(e,o))!=ke.Word)break;o=e}for(;i<r;){let e=E(t,i);if(a(t.slice(i,e))!=ke.Word)break;i=e}return o==i?null:F.range(o+n,i+n)}}function _e(e,t,n={}){let r={};for(let t of e)for(let e of Object.keys(t)){let a=t[e],o=r[e];if(void 0===o)r[e]=a;else if(o===a||void 0===a);else{if(!Object.hasOwnProperty.call(n,e))throw new Error("Config merge conflict for field "+e);r[e]=n[e](o,a)}}for(let e in t)void 0===r[e]&&(r[e]=t[e]);return r}Se.allowMultipleSelections=oe,Se.tabSize=z.define({combine:e=>e.length?e[0]:4}),Se.lineSeparator=ie,Se.readOnly=ue,Se.phrases=z.define({compare(e,t){let n=Object.keys(e),r=Object.keys(t);return n.length==r.length&&n.every((n=>e[n]==t[n]))}}),Se.languageData=ae,Se.changeFilter=se,Se.transactionFilter=le,Se.transactionExtender=ce,J.reconfigure=fe.define();class Ce{eq(e){return this==e}range(e,t=e){return Te.create(e,t,this)}}Ce.prototype.startSide=Ce.prototype.endSide=0,Ce.prototype.point=!1,Ce.prototype.mapMode=A.TrackDel;class Te{constructor(e,t,n){this.from=e,this.to=t,this.value=n}static create(e,t,n){return new Te(e,t,n)}}function Ae(e,t){return e.from-t.from||e.value.startSide-t.value.startSide}class Re{constructor(e,t,n,r){this.from=e,this.to=t,this.value=n,this.maxPoint=r}get length(){return this.to[this.to.length-1]}findIndex(e,t,n,r=0){let a=n?this.to:this.from;for(let o=r,i=a.length;;){if(o==i)return o;let r=o+i>>1,s=a[r]-e||(n?this.value[r].endSide:this.value[r].startSide)-t;if(r==o)return s>=0?o:i;s>=0?i=r:o=r+1}}between(e,t,n,r){for(let a=this.findIndex(t,-1e9,!0),o=this.findIndex(n,1e9,!1,a);a<o;a++)if(!1===r(this.from[a]+e,this.to[a]+e,this.value[a]))return!1}map(e,t){let n=[],r=[],a=[],o=-1,i=-1;for(let s=0;s<this.value.length;s++){let l,c,u=this.value[s],d=this.from[s]+e,p=this.to[s]+e;if(d==p){let e=t.mapPos(d,u.startSide,u.mapMode);if(null==e)continue;if(l=c=e,u.startSide!=u.endSide&&(c=t.mapPos(d,u.endSide),c<l))continue}else if(l=t.mapPos(d,u.startSide),c=t.mapPos(p,u.endSide),l>c||l==c&&u.startSide>0&&u.endSide<=0)continue;(c-l||u.endSide-u.startSide)<0||(o<0&&(o=l),u.point&&(i=Math.max(i,c-l)),n.push(u),r.push(l-o),a.push(c-o))}return{mapped:n.length?new Re(r,a,n,i):null,pos:o}}}class Pe{constructor(e,t,n,r){this.chunkPos=e,this.chunk=t,this.nextLayer=n,this.maxPoint=r}static create(e,t,n,r){return new Pe(e,t,n,r)}get length(){let e=this.chunk.length-1;return e<0?0:Math.max(this.chunkEnd(e),this.nextLayer.length)}get size(){if(this.isEmpty)return 0;let e=this.nextLayer.size;for(let t of this.chunk)e+=t.value.length;return e}chunkEnd(e){return this.chunkPos[e]+this.chunk[e].length}update(e){let{add:t=[],sort:n=!1,filterFrom:r=0,filterTo:a=this.length}=e,o=e.filter;if(0==t.length&&!o)return this;if(n&&(t=t.slice().sort(Ae)),this.isEmpty)return t.length?Pe.of(t):this;let i=new $e(this,null,-1).goto(0),s=0,l=[],c=new De;for(;i.value||s<t.length;)if(s<t.length&&(i.from-t[s].from||i.startSide-t[s].value.startSide)>=0){let e=t[s++];c.addInner(e.from,e.to,e.value)||l.push(e)}else 1==i.rangeIndex&&i.chunkIndex<this.chunk.length&&(s==t.length||this.chunkEnd(i.chunkIndex)<t[s].from)&&(!o||r>this.chunkEnd(i.chunkIndex)||a<this.chunkPos[i.chunkIndex])&&c.addChunk(this.chunkPos[i.chunkIndex],this.chunk[i.chunkIndex])?i.nextChunk():((!o||r>i.to||a<i.from||o(i.from,i.to,i.value))&&(c.addInner(i.from,i.to,i.value)||l.push(Te.create(i.from,i.to,i.value))),i.next());return c.finishInner(this.nextLayer.isEmpty&&!l.length?Pe.empty:this.nextLayer.update({add:l,filter:o,filterFrom:r,filterTo:a}))}map(e){if(e.empty||this.isEmpty)return this;let t=[],n=[],r=-1;for(let a=0;a<this.chunk.length;a++){let o=this.chunkPos[a],i=this.chunk[a],s=e.touchesRange(o,o+i.length);if(!1===s)r=Math.max(r,i.maxPoint),t.push(i),n.push(e.mapPos(o));else if(!0===s){let{mapped:a,pos:s}=i.map(o,e);a&&(r=Math.max(r,a.maxPoint),t.push(a),n.push(s))}}let a=this.nextLayer.map(e);return 0==t.length?a:new Pe(n,t,a||Pe.empty,r)}between(e,t,n){if(!this.isEmpty){for(let r=0;r<this.chunk.length;r++){let a=this.chunkPos[r],o=this.chunk[r];if(t>=a&&e<=a+o.length&&!1===o.between(a,e-a,t-a,n))return}this.nextLayer.between(e,t,n)}}iter(e=0){return Ne.from([this]).goto(e)}get isEmpty(){return this.nextLayer==this}static iter(e,t=0){return Ne.from(e).goto(t)}static compare(e,t,n,r,a=-1){let o=e.filter((e=>e.maxPoint>0||!e.isEmpty&&e.maxPoint>=a)),i=t.filter((e=>e.maxPoint>0||!e.isEmpty&&e.maxPoint>=a)),s=Me(o,i,n),l=new Le(o,s,a),c=new Le(i,s,a);n.iterGaps(((e,t,n)=>Be(l,e,c,t,n,r))),n.empty&&0==n.length&&Be(l,0,c,0,0,r)}static eq(e,t,n=0,r){null==r&&(r=999999999);let a=e.filter((e=>!e.isEmpty&&t.indexOf(e)<0)),o=t.filter((t=>!t.isEmpty&&e.indexOf(t)<0));if(a.length!=o.length)return!1;if(!a.length)return!0;let i=Me(a,o),s=new Le(a,i,0).goto(n),l=new Le(o,i,0).goto(n);for(;;){if(s.to!=l.to||!Fe(s.active,l.active)||s.point&&(!l.point||!s.point.eq(l.point)))return!1;if(s.to>r)return!0;s.next(),l.next()}}static spans(e,t,n,r,a=-1){let o=new Le(e,null,a).goto(t),i=t,s=o.openStart;for(;;){let e=Math.min(o.to,n);if(o.point){let n=o.activeForPoint(o.to),a=o.pointFrom<t?n.length+1:o.point.startSide<0?n.length:Math.min(n.length,s);r.point(i,e,o.point,n,a,o.pointRank),s=Math.min(o.openEnd(e),n.length)}else e>i&&(r.span(i,e,o.active,s),s=o.openEnd(e));if(o.to>n)return s+(o.point&&o.to>n?1:0);i=o.to,o.next()}}static of(e,t=!1){let n=new De;for(let r of e instanceof Te?[e]:t?function(e){if(e.length>1)for(let t=e[0],n=1;n<e.length;n++){let r=e[n];if(Ae(t,r)>0)return e.slice().sort(Ae);t=r}return e}(e):e)n.add(r.from,r.to,r.value);return n.finish()}static join(e){if(!e.length)return Pe.empty;let t=e[e.length-1];for(let n=e.length-2;n>=0;n--)for(let r=e[n];r!=Pe.empty;r=r.nextLayer)t=new Pe(r.chunkPos,r.chunk,t,Math.max(r.maxPoint,t.maxPoint));return t}}Pe.empty=new Pe([],[],null,-1),Pe.empty.nextLayer=Pe.empty;class De{finishChunk(e){this.chunks.push(new Re(this.from,this.to,this.value,this.maxPoint)),this.chunkPos.push(this.chunkStart),this.chunkStart=-1,this.setMaxPoint=Math.max(this.setMaxPoint,this.maxPoint),this.maxPoint=-1,e&&(this.from=[],this.to=[],this.value=[])}constructor(){this.chunks=[],this.chunkPos=[],this.chunkStart=-1,this.last=null,this.lastFrom=-1e9,this.lastTo=-1e9,this.from=[],this.to=[],this.value=[],this.maxPoint=-1,this.setMaxPoint=-1,this.nextLayer=null}add(e,t,n){this.addInner(e,t,n)||(this.nextLayer||(this.nextLayer=new De)).add(e,t,n)}addInner(e,t,n){let r=e-this.lastTo||n.startSide-this.last.endSide;if(r<=0&&(e-this.lastFrom||n.startSide-this.last.startSide)<0)throw new Error("Ranges must be added sorted by `from` position and `startSide`");return!(r<0||(250==this.from.length&&this.finishChunk(!0),this.chunkStart<0&&(this.chunkStart=e),this.from.push(e-this.chunkStart),this.to.push(t-this.chunkStart),this.last=n,this.lastFrom=e,this.lastTo=t,this.value.push(n),n.point&&(this.maxPoint=Math.max(this.maxPoint,t-e)),0))}addChunk(e,t){if((e-this.lastTo||t.value[0].startSide-this.last.endSide)<0)return!1;this.from.length&&this.finishChunk(!0),this.setMaxPoint=Math.max(this.setMaxPoint,t.maxPoint),this.chunks.push(t),this.chunkPos.push(e);let n=t.value.length-1;return this.last=t.value[n],this.lastFrom=t.from[n]+e,this.lastTo=t.to[n]+e,!0}finish(){return this.finishInner(Pe.empty)}finishInner(e){if(this.from.length&&this.finishChunk(!1),0==this.chunks.length)return e;let t=Pe.create(this.chunkPos,this.chunks,this.nextLayer?this.nextLayer.finishInner(e):e,this.setMaxPoint);return this.from=null,t}}function Me(e,t,n){let r=new Map;for(let t of e)for(let e=0;e<t.chunk.length;e++)t.chunk[e].maxPoint<=0&&r.set(t.chunk[e],t.chunkPos[e]);let a=new Set;for(let e of t)for(let t=0;t<e.chunk.length;t++){let o=r.get(e.chunk[t]);null==o||(n?n.mapPos(o):o)!=e.chunkPos[t]||(null==n?void 0:n.touchesRange(o,o+e.chunk[t].length))||a.add(e.chunk[t])}return a}class $e{constructor(e,t,n,r=0){this.layer=e,this.skip=t,this.minPoint=n,this.rank=r}get startSide(){return this.value?this.value.startSide:0}get endSide(){return this.value?this.value.endSide:0}goto(e,t=-1e9){return this.chunkIndex=this.rangeIndex=0,this.gotoInner(e,t,!1),this}gotoInner(e,t,n){for(;this.chunkIndex<this.layer.chunk.length;){let t=this.layer.chunk[this.chunkIndex];if(!(this.skip&&this.skip.has(t)||this.layer.chunkEnd(this.chunkIndex)<e||t.maxPoint<this.minPoint))break;this.chunkIndex++,n=!1}if(this.chunkIndex<this.layer.chunk.length){let r=this.layer.chunk[this.chunkIndex].findIndex(e-this.layer.chunkPos[this.chunkIndex],t,!0);(!n||this.rangeIndex<r)&&this.setRangeIndex(r)}this.next()}forward(e,t){(this.to-e||this.endSide-t)<0&&this.gotoInner(e,t,!0)}next(){for(;;){if(this.chunkIndex==this.layer.chunk.length){this.from=this.to=1e9,this.value=null;break}{let e=this.layer.chunkPos[this.chunkIndex],t=this.layer.chunk[this.chunkIndex],n=e+t.from[this.rangeIndex];if(this.from=n,this.to=e+t.to[this.rangeIndex],this.value=t.value[this.rangeIndex],this.setRangeIndex(this.rangeIndex+1),this.minPoint<0||this.value.point&&this.to-this.from>=this.minPoint)break}}}setRangeIndex(e){if(e==this.layer.chunk[this.chunkIndex].value.length){if(this.chunkIndex++,this.skip)for(;this.chunkIndex<this.layer.chunk.length&&this.skip.has(this.layer.chunk[this.chunkIndex]);)this.chunkIndex++;this.rangeIndex=0}else this.rangeIndex=e}nextChunk(){this.chunkIndex++,this.rangeIndex=0,this.next()}compare(e){return this.from-e.from||this.startSide-e.startSide||this.rank-e.rank||this.to-e.to||this.endSide-e.endSide}}class Ne{constructor(e){this.heap=e}static from(e,t=null,n=-1){let r=[];for(let a=0;a<e.length;a++)for(let o=e[a];!o.isEmpty;o=o.nextLayer)o.maxPoint>=n&&r.push(new $e(o,t,n,a));return 1==r.length?r[0]:new Ne(r)}get startSide(){return this.value?this.value.startSide:0}goto(e,t=-1e9){for(let n of this.heap)n.goto(e,t);for(let e=this.heap.length>>1;e>=0;e--)Ie(this.heap,e);return this.next(),this}forward(e,t){for(let n of this.heap)n.forward(e,t);for(let e=this.heap.length>>1;e>=0;e--)Ie(this.heap,e);(this.to-e||this.value.endSide-t)<0&&this.next()}next(){if(0==this.heap.length)this.from=this.to=1e9,this.value=null,this.rank=-1;else{let e=this.heap[0];this.from=e.from,this.to=e.to,this.value=e.value,this.rank=e.rank,e.value&&e.next(),Ie(this.heap,0)}}}function Ie(e,t){for(let n=e[t];;){let r=1+(t<<1);if(r>=e.length)break;let a=e[r];if(r+1<e.length&&a.compare(e[r+1])>=0&&(a=e[r+1],r++),n.compare(a)<0)break;e[r]=n,e[t]=a,t=r}}class Le{constructor(e,t,n){this.minPoint=n,this.active=[],this.activeTo=[],this.activeRank=[],this.minActive=-1,this.point=null,this.pointFrom=0,this.pointRank=0,this.to=-1e9,this.endSide=0,this.openStart=-1,this.cursor=Ne.from(e,t,n)}goto(e,t=-1e9){return this.cursor.goto(e,t),this.active.length=this.activeTo.length=this.activeRank.length=0,this.minActive=-1,this.to=e,this.endSide=t,this.openStart=-1,this.next(),this}forward(e,t){for(;this.minActive>-1&&(this.activeTo[this.minActive]-e||this.active[this.minActive].endSide-t)<0;)this.removeActive(this.minActive);this.cursor.forward(e,t)}removeActive(e){Qe(this.active,e),Qe(this.activeTo,e),Qe(this.activeRank,e),this.minActive=ze(this.active,this.activeTo)}addActive(e){let t=0,{value:n,to:r,rank:a}=this.cursor;for(;t<this.activeRank.length&&(a-this.activeRank[t]||r-this.activeTo[t])>0;)t++;Ue(this.active,t,n),Ue(this.activeTo,t,r),Ue(this.activeRank,t,a),e&&Ue(e,t,this.cursor.from),this.minActive=ze(this.active,this.activeTo)}next(){let e=this.to,t=this.point;this.point=null;let n=this.openStart<0?[]:null;for(;;){let r=this.minActive;if(r>-1&&(this.activeTo[r]-this.cursor.from||this.active[r].endSide-this.cursor.startSide)<0){if(this.activeTo[r]>e){this.to=this.activeTo[r],this.endSide=this.active[r].endSide;break}this.removeActive(r),n&&Qe(n,r)}else{if(!this.cursor.value){this.to=this.endSide=1e9;break}if(this.cursor.from>e){this.to=this.cursor.from,this.endSide=this.cursor.startSide;break}{let e=this.cursor.value;if(e.point){if(!(t&&this.cursor.to==this.to&&this.cursor.from<this.cursor.to)){this.point=e,this.pointFrom=this.cursor.from,this.pointRank=this.cursor.rank,this.to=this.cursor.to,this.endSide=e.endSide,this.cursor.next(),this.forward(this.to,this.endSide);break}this.cursor.next()}else this.addActive(n),this.cursor.next()}}}if(n){this.openStart=0;for(let t=n.length-1;t>=0&&n[t]<e;t--)this.openStart++}}activeForPoint(e){if(!this.active.length)return this.active;let t=[];for(let n=this.active.length-1;n>=0&&!(this.activeRank[n]<this.pointRank);n--)(this.activeTo[n]>e||this.activeTo[n]==e&&this.active[n].endSide>=this.point.endSide)&&t.push(this.active[n]);return t.reverse()}openEnd(e){let t=0;for(let n=this.activeTo.length-1;n>=0&&this.activeTo[n]>e;n--)t++;return t}}function Be(e,t,n,r,a,o){e.goto(t),n.goto(r);let i=r+a,s=r,l=r-t;for(;;){let t=e.to+l-n.to,r=t||e.endSide-n.endSide,a=r<0?e.to+l:n.to,c=Math.min(a,i);if(e.point||n.point?e.point&&n.point&&(e.point==n.point||e.point.eq(n.point))&&Fe(e.activeForPoint(e.to),n.activeForPoint(n.to))||o.comparePoint(s,c,e.point,n.point):c>s&&!Fe(e.active,n.active)&&o.compareRange(s,c,e.active,n.active),a>i)break;(t||e.openEnd!=n.openEnd)&&o.boundChange&&o.boundChange(a),s=a,r<=0&&e.next(),r>=0&&n.next()}}function Fe(e,t){if(e.length!=t.length)return!1;for(let n=0;n<e.length;n++)if(e[n]!=t[n]&&!e[n].eq(t[n]))return!1;return!0}function Qe(e,t){for(let n=t,r=e.length-1;n<r;n++)e[n]=e[n+1];e.pop()}function Ue(e,t,n){for(let n=e.length-1;n>=t;n--)e[n+1]=e[n];e[t]=n}function ze(e,t){let n=-1,r=1e9;for(let a=0;a<t.length;a++)(t[a]-r||e[a].endSide-e[n].endSide)<0&&(n=a,r=t[a]);return n}function je(e,t,n=e.length){let r=0;for(let a=0;a<n&&a<e.length;)9==e.charCodeAt(a)?(r+=t-r%t,a++):(r++,a=E(e,a));return r}const We="undefined"==typeof Symbol?"__ͼ":Symbol.for("ͼ"),Xe="undefined"==typeof Symbol?"__styleSet"+Math.floor(1e8*Math.random()):Symbol("styleSet"),qe="undefined"!=typeof globalThis?globalThis:"undefined"!=typeof window?window:{};class Ve{constructor(e,t){this.rules=[];let{finish:n}=t||{};function r(e){return/^@/.test(e)?[e]:e.split(/,\s*/)}function a(e,t,o,i){let s=[],l=/^@(\w+)\b/.exec(e[0]),c=l&&"keyframes"==l[1];if(l&&null==t)return o.push(e[0]+";");for(let n in t){let i=t[n];if(/&/.test(n))a(n.split(/,\s*/).map((t=>e.map((e=>t.replace(/&/,e))))).reduce(((e,t)=>e.concat(t))),i,o);else if(i&&"object"==typeof i){if(!l)throw new RangeError("The value of a property ("+n+") should be a primitive value.");a(r(n),i,s,c)}else null!=i&&s.push(n.replace(/_.*/,"").replace(/[A-Z]/g,(e=>"-"+e.toLowerCase()))+": "+i+";")}(s.length||c)&&o.push((!n||l||i?e:e.map(n)).join(", ")+" {"+s.join(" ")+"}")}for(let t in e)a(r(t),e[t],this.rules)}getRules(){return this.rules.join("\n")}static newName(){let e=qe[We]||1;return qe[We]=e+1,"ͼ"+e.toString(36)}static mount(e,t,n){let r=e[Xe],a=n&&n.nonce;r?a&&r.setNonce(a):r=new Ze(e,a),r.mount(Array.isArray(t)?t:[t],e)}}let He=new Map;class Ze{constructor(e,t){let n=e.ownerDocument||e,r=n.defaultView;if(!e.head&&e.adoptedStyleSheets&&r.CSSStyleSheet){let t=He.get(n);if(t)return e[Xe]=t;this.sheet=new r.CSSStyleSheet,He.set(n,this)}else this.styleTag=n.createElement("style"),t&&this.styleTag.setAttribute("nonce",t);this.modules=[],e[Xe]=this}mount(e,t){let n=this.sheet,r=0,a=0;for(let t=0;t<e.length;t++){let o=e[t],i=this.modules.indexOf(o);if(i<a&&i>-1&&(this.modules.splice(i,1),a--,i=-1),-1==i){if(this.modules.splice(a++,0,o),n)for(let e=0;e<o.rules.length;e++)n.insertRule(o.rules[e],r++)}else{for(;a<i;)r+=this.modules[a++].rules.length;r+=o.rules.length,a++}}if(n)t.adoptedStyleSheets.indexOf(this.sheet)<0&&(t.adoptedStyleSheets=[this.sheet,...t.adoptedStyleSheets]);else{let e="";for(let t=0;t<this.modules.length;t++)e+=this.modules[t].getRules()+"\n";this.styleTag.textContent=e;let n=t.head||t;this.styleTag.parentNode!=n&&n.insertBefore(this.styleTag,n.firstChild)}}setNonce(e){this.styleTag&&this.styleTag.getAttribute("nonce")!=e&&this.styleTag.setAttribute("nonce",e)}}for(var Ge={8:"Backspace",9:"Tab",10:"Enter",12:"NumLock",13:"Enter",16:"Shift",17:"Control",18:"Alt",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",44:"PrintScreen",45:"Insert",46:"Delete",59:";",61:"=",91:"Meta",92:"Meta",106:"*",107:"+",108:",",109:"-",110:".",111:"/",144:"NumLock",145:"ScrollLock",160:"Shift",161:"Shift",162:"Control",163:"Control",164:"Alt",165:"Alt",173:"-",186:";",187:"=",188:",",189:"-",190:".",191:"/",192:"`",219:"[",220:"\\",221:"]",222:"'"},Ke={48:")",49:"!",50:"@",51:"#",52:"$",53:"%",54:"^",55:"&",56:"*",57:"(",59:":",61:"+",173:"_",186:":",187:"+",188:"<",189:"_",190:">",191:"?",192:"~",219:"{",220:"|",221:"}",222:'"'},Ye="undefined"!=typeof navigator&&/Mac/.test(navigator.platform),Je="undefined"!=typeof navigator&&/MSIE \d|Trident\/(?:[7-9]|\d{2,})\..*rv:(\d+)/.exec(navigator.userAgent),et=0;et<10;et++)Ge[48+et]=Ge[96+et]=String(et);for(et=1;et<=24;et++)Ge[et+111]="F"+et;for(et=65;et<=90;et++)Ge[et]=String.fromCharCode(et+32),Ke[et]=String.fromCharCode(et);for(var tt in Ge)Ke.hasOwnProperty(tt)||(Ke[tt]=Ge[tt]);function nt(e){let t;return t=11==e.nodeType?e.getSelection?e:e.ownerDocument:e,t.getSelection()}function rt(e,t){return!!t&&(e==t||e.contains(1!=t.nodeType?t.parentNode:t))}function at(e,t){if(!t.anchorNode)return!1;try{return rt(e,t.anchorNode)}catch(e){return!1}}function ot(e){return 3==e.nodeType?bt(e,0,e.nodeValue.length).getClientRects():1==e.nodeType?e.getClientRects():[]}function it(e,t,n,r){return!!n&&(ct(e,t,n,r,-1)||ct(e,t,n,r,1))}function st(e){for(var t=0;;t++)if(!(e=e.previousSibling))return t}function lt(e){return 1==e.nodeType&&/^(DIV|P|LI|UL|OL|BLOCKQUOTE|DD|DT|H\d|SECTION|PRE)$/.test(e.nodeName)}function ct(e,t,n,r,a){for(;;){if(e==n&&t==r)return!0;if(t==(a<0?0:ut(e))){if("DIV"==e.nodeName)return!1;let n=e.parentNode;if(!n||1!=n.nodeType)return!1;t=st(e)+(a<0?0:1),e=n}else{if(1!=e.nodeType)return!1;if(1==(e=e.childNodes[t+(a<0?-1:0)]).nodeType&&"false"==e.contentEditable)return!1;t=a<0?ut(e):0}}}function ut(e){return 3==e.nodeType?e.nodeValue.length:e.childNodes.length}function dt(e,t){let n=t?e.left:e.right;return{left:n,right:n,top:e.top,bottom:e.bottom}}function pt(e){let t=e.visualViewport;return t?{left:0,right:t.width,top:0,bottom:t.height}:{left:0,right:e.innerWidth,top:0,bottom:e.innerHeight}}function ht(e,t){let n=t.width/e.offsetWidth,r=t.height/e.offsetHeight;return(n>.995&&n<1.005||!isFinite(n)||Math.abs(t.width-e.offsetWidth)<1)&&(n=1),(r>.995&&r<1.005||!isFinite(r)||Math.abs(t.height-e.offsetHeight)<1)&&(r=1),{scaleX:n,scaleY:r}}class ft{constructor(){this.anchorNode=null,this.anchorOffset=0,this.focusNode=null,this.focusOffset=0}eq(e){return this.anchorNode==e.anchorNode&&this.anchorOffset==e.anchorOffset&&this.focusNode==e.focusNode&&this.focusOffset==e.focusOffset}setRange(e){let{anchorNode:t,focusNode:n}=e;this.set(t,Math.min(e.anchorOffset,t?ut(t):0),n,Math.min(e.focusOffset,n?ut(n):0))}set(e,t,n,r){this.anchorNode=e,this.anchorOffset=t,this.focusNode=n,this.focusOffset=r}}let mt,gt=null;function vt(e){if(e.setActive)return e.setActive();if(gt)return e.focus(gt);let t=[];for(let n=e;n&&(t.push(n,n.scrollTop,n.scrollLeft),n!=n.ownerDocument);n=n.parentNode);if(e.focus(null==gt?{get preventScroll(){return gt={preventScroll:!0},!0}}:void 0),!gt){gt=!1;for(let e=0;e<t.length;){let n=t[e++],r=t[e++],a=t[e++];n.scrollTop!=r&&(n.scrollTop=r),n.scrollLeft!=a&&(n.scrollLeft=a)}}}function bt(e,t,n=t){let r=mt||(mt=document.createRange());return r.setEnd(e,n),r.setStart(e,t),r}function yt(e,t,n,r){let a={key:t,code:t,keyCode:n,which:n,cancelable:!0};r&&({altKey:a.altKey,ctrlKey:a.ctrlKey,shiftKey:a.shiftKey,metaKey:a.metaKey}=r);let o=new KeyboardEvent("keydown",a);o.synthetic=!0,e.dispatchEvent(o);let i=new KeyboardEvent("keyup",a);return i.synthetic=!0,e.dispatchEvent(i),o.defaultPrevented||i.defaultPrevented}function wt(e){for(;e.attributes.length;)e.removeAttributeNode(e.attributes[0])}function Ot(e){return e.scrollTop>Math.max(1,e.scrollHeight-e.clientHeight-4)}function kt(e,t){for(let n=e,r=t;;){if(3==n.nodeType&&r>0)return{node:n,offset:r};if(1==n.nodeType&&r>0){if("false"==n.contentEditable)return null;n=n.childNodes[r-1],r=ut(n)}else{if(!n.parentNode||lt(n))return null;r=st(n),n=n.parentNode}}}function xt(e,t){for(let n=e,r=t;;){if(3==n.nodeType&&r<n.nodeValue.length)return{node:n,offset:r};if(1==n.nodeType&&r<n.childNodes.length){if("false"==n.contentEditable)return null;n=n.childNodes[r],r=0}else{if(!n.parentNode||lt(n))return null;r=st(n)+1,n=n.parentNode}}}class Et{constructor(e,t,n=!0){this.node=e,this.offset=t,this.precise=n}static before(e,t){return new Et(e.parentNode,st(e),t)}static after(e,t){return new Et(e.parentNode,st(e)+1,t)}}const St=[];class _t{constructor(){this.parent=null,this.dom=null,this.flags=2}get overrideDOMText(){return null}get posAtStart(){return this.parent?this.parent.posBefore(this):0}get posAtEnd(){return this.posAtStart+this.length}posBefore(e){let t=this.posAtStart;for(let n of this.children){if(n==e)return t;t+=n.length+n.breakAfter}throw new RangeError("Invalid child in posBefore")}posAfter(e){return this.posBefore(e)+e.length}sync(e,t){if(2&this.flags){let n,r=this.dom,a=null;for(let o of this.children){if(7&o.flags){if(!o.dom&&(n=a?a.nextSibling:r.firstChild)){let e=_t.get(n);(!e||!e.parent&&e.canReuseDOM(o))&&o.reuseDOM(n)}o.sync(e,t),o.flags&=-8}if(n=a?a.nextSibling:r.firstChild,t&&!t.written&&t.node==r&&n!=o.dom&&(t.written=!0),o.dom.parentNode==r)for(;n&&n!=o.dom;)n=Ct(n);else r.insertBefore(o.dom,n);a=o.dom}for(n=a?a.nextSibling:r.firstChild,n&&t&&t.node==r&&(t.written=!0);n;)n=Ct(n)}else if(1&this.flags)for(let n of this.children)7&n.flags&&(n.sync(e,t),n.flags&=-8)}reuseDOM(e){}localPosFromDOM(e,t){let n;if(e==this.dom)n=this.dom.childNodes[t];else{let r=0==ut(e)?0:0==t?-1:1;for(;;){let t=e.parentNode;if(t==this.dom)break;0==r&&t.firstChild!=t.lastChild&&(r=e==t.firstChild?-1:1),e=t}n=r<0?e:e.nextSibling}if(n==this.dom.firstChild)return 0;for(;n&&!_t.get(n);)n=n.nextSibling;if(!n)return this.length;for(let e=0,t=0;;e++){let r=this.children[e];if(r.dom==n)return t;t+=r.length+r.breakAfter}}domBoundsAround(e,t,n=0){let r=-1,a=-1,o=-1,i=-1;for(let s=0,l=n,c=n;s<this.children.length;s++){let n=this.children[s],u=l+n.length;if(l<e&&u>t)return n.domBoundsAround(e,t,l);if(u>=e&&-1==r&&(r=s,a=l),l>t&&n.dom.parentNode==this.dom){o=s,i=c;break}c=u,l=u+n.breakAfter}return{from:a,to:i<0?n+this.length:i,startDOM:(r?this.children[r-1].dom.nextSibling:null)||this.dom.firstChild,endDOM:o<this.children.length&&o>=0?this.children[o].dom:null}}markDirty(e=!1){this.flags|=2,this.markParentsDirty(e)}markParentsDirty(e){for(let t=this.parent;t;t=t.parent){if(e&&(t.flags|=2),1&t.flags)return;t.flags|=1,e=!1}}setParent(e){this.parent!=e&&(this.parent=e,7&this.flags&&this.markParentsDirty(!0))}setDOM(e){this.dom!=e&&(this.dom&&(this.dom.cmView=null),this.dom=e,e.cmView=this)}get rootView(){for(let e=this;;){let t=e.parent;if(!t)return e;e=t}}replaceChildren(e,t,n=St){this.markDirty();for(let r=e;r<t;r++){let e=this.children[r];e.parent==this&&n.indexOf(e)<0&&e.destroy()}n.length<250?this.children.splice(e,t-e,...n):this.children=[].concat(this.children.slice(0,e),n,this.children.slice(t));for(let e=0;e<n.length;e++)n[e].setParent(this)}ignoreMutation(e){return!1}ignoreEvent(e){return!1}childCursor(e=this.length){return new Tt(this.children,e,this.children.length)}childPos(e,t=1){return this.childCursor().findPos(e,t)}toString(){let e=this.constructor.name.replace("View","");return e+(this.children.length?"("+this.children.join()+")":this.length?"["+("Text"==e?this.text:this.length)+"]":"")+(this.breakAfter?"#":"")}static get(e){return e.cmView}get isEditable(){return!0}get isWidget(){return!1}get isHidden(){return!1}merge(e,t,n,r,a,o){return!1}become(e){return!1}canReuseDOM(e){return e.constructor==this.constructor&&!(8&(this.flags|e.flags))}getSide(){return 0}destroy(){for(let e of this.children)e.parent==this&&e.destroy();this.parent=null}}function Ct(e){let t=e.nextSibling;return e.parentNode.removeChild(e),t}_t.prototype.breakAfter=0;class Tt{constructor(e,t,n){this.children=e,this.pos=t,this.i=n,this.off=0}findPos(e,t=1){for(;;){if(e>this.pos||e==this.pos&&(t>0||0==this.i||this.children[this.i-1].breakAfter))return this.off=e-this.pos,this;let n=this.children[--this.i];this.pos-=n.length+n.breakAfter}}}function At(e,t,n,r,a,o,i,s,l){let{children:c}=e,u=c.length?c[t]:null,d=o.length?o[o.length-1]:null,p=d?d.breakAfter:i;if(!(t==r&&u&&!i&&!p&&o.length<2&&u.merge(n,a,o.length?d:null,0==n,s,l))){if(r<c.length){let e=c[r];e&&(a<e.length||e.breakAfter&&(null==d?void 0:d.breakAfter))?(t==r&&(e=e.split(a),a=0),!p&&d&&e.merge(0,a,d,!0,0,l)?o[o.length-1]=e:((a||e.children.length&&!e.children[0].length)&&e.merge(0,a,null,!1,0,l),o.push(e))):(null==e?void 0:e.breakAfter)&&(d?d.breakAfter=1:i=1),r++}for(u&&(u.breakAfter=i,n>0&&(!i&&o.length&&u.merge(n,u.length,o[0],!1,s,0)?u.breakAfter=o.shift().breakAfter:(n<u.length||u.children.length&&0==u.children[u.children.length-1].length)&&u.merge(n,u.length,null,!1,s,0),t++));t<r&&o.length;)if(c[r-1].become(o[o.length-1]))r--,o.pop(),l=o.length?0:s;else{if(!c[t].become(o[0]))break;t++,o.shift(),s=o.length?0:l}!o.length&&t&&r<c.length&&!c[t-1].breakAfter&&c[r].merge(0,0,c[t-1],!1,s,l)&&t--,(t<r||o.length)&&e.replaceChildren(t,r,o)}}function Rt(e,t,n,r,a,o){let i=e.childCursor(),{i:s,off:l}=i.findPos(n,1),{i:c,off:u}=i.findPos(t,-1),d=t-n;for(let e of r)d+=e.length;e.length+=d,At(e,c,u,s,l,r,0,a,o)}let Pt="undefined"!=typeof navigator?navigator:{userAgent:"",vendor:"",platform:""},Dt="undefined"!=typeof document?document:{documentElement:{style:{}}};const Mt=/Edge\/(\d+)/.exec(Pt.userAgent),$t=/MSIE \d/.test(Pt.userAgent),Nt=/Trident\/(?:[7-9]|\d{2,})\..*rv:(\d+)/.exec(Pt.userAgent),It=!!($t||Nt||Mt),Lt=!It&&/gecko\/(\d+)/i.test(Pt.userAgent),Bt=!It&&/Chrome\/(\d+)/.exec(Pt.userAgent),Ft="webkitFontSmoothing"in Dt.documentElement.style,Qt=!It&&/Apple Computer/.test(Pt.vendor),Ut=Qt&&(/Mobile\/\w+/.test(Pt.userAgent)||Pt.maxTouchPoints>2);var zt={mac:Ut||/Mac/.test(Pt.platform),windows:/Win/.test(Pt.platform),linux:/Linux|X11/.test(Pt.platform),ie:It,ie_version:$t?Dt.documentMode||6:Nt?+Nt[1]:Mt?+Mt[1]:0,gecko:Lt,gecko_version:Lt?+(/Firefox\/(\d+)/.exec(Pt.userAgent)||[0,0])[1]:0,chrome:!!Bt,chrome_version:Bt?+Bt[1]:0,ios:Ut,android:/Android\b/.test(Pt.userAgent),webkit:Ft,safari:Qt,webkit_version:Ft?+(/\bAppleWebKit\/(\d+)/.exec(Pt.userAgent)||[0,0])[1]:0,tabSize:null!=Dt.documentElement.style.tabSize?"tab-size":"-moz-tab-size"};class jt extends _t{constructor(e){super(),this.text=e}get length(){return this.text.length}createDOM(e){this.setDOM(e||document.createTextNode(this.text))}sync(e,t){this.dom||this.createDOM(),this.dom.nodeValue!=this.text&&(t&&t.node==this.dom&&(t.written=!0),this.dom.nodeValue=this.text)}reuseDOM(e){3==e.nodeType&&this.createDOM(e)}merge(e,t,n){return!(8&this.flags||n&&(!(n instanceof jt)||this.length-(t-e)+n.length>256||8&n.flags)||(this.text=this.text.slice(0,e)+(n?n.text:"")+this.text.slice(t),this.markDirty(),0))}split(e){let t=new jt(this.text.slice(e));return this.text=this.text.slice(0,e),this.markDirty(),t.flags|=8&this.flags,t}localPosFromDOM(e,t){return e==this.dom?t:t?this.text.length:0}domAtPos(e){return new Et(this.dom,e)}domBoundsAround(e,t,n){return{from:n,to:n+this.length,startDOM:this.dom,endDOM:this.dom.nextSibling}}coordsAt(e,t){return function(e,t,n){let r=e.nodeValue.length;t>r&&(t=r);let a=t,o=t,i=0;0==t&&n<0||t==r&&n>=0?zt.chrome||zt.gecko||(t?(a--,i=1):o<r&&(o++,i=-1)):n<0?a--:o<r&&o++;let s=bt(e,a,o).getClientRects();if(!s.length)return null;let l=s[(i?i<0:n>=0)?0:s.length-1];return zt.safari&&!i&&0==l.width&&(l=Array.prototype.find.call(s,(e=>e.width))||l),i?dt(l,i<0):l||null}(this.dom,e,t)}}class Wt extends _t{constructor(e,t=[],n=0){super(),this.mark=e,this.children=t,this.length=n;for(let e of t)e.setParent(this)}setAttrs(e){if(wt(e),this.mark.class&&(e.className=this.mark.class),this.mark.attrs)for(let t in this.mark.attrs)e.setAttribute(t,this.mark.attrs[t]);return e}canReuseDOM(e){return super.canReuseDOM(e)&&!(8&(this.flags|e.flags))}reuseDOM(e){e.nodeName==this.mark.tagName.toUpperCase()&&(this.setDOM(e),this.flags|=6)}sync(e,t){this.dom?4&this.flags&&this.setAttrs(this.dom):this.setDOM(this.setAttrs(document.createElement(this.mark.tagName))),super.sync(e,t)}merge(e,t,n,r,a,o){return!(n&&(!(n instanceof Wt&&n.mark.eq(this.mark))||e&&a<=0||t<this.length&&o<=0)||(Rt(this,e,t,n?n.children.slice():[],a-1,o-1),this.markDirty(),0))}split(e){let t=[],n=0,r=-1,a=0;for(let o of this.children){let i=n+o.length;i>e&&t.push(n<e?o.split(e-n):o),r<0&&n>=e&&(r=a),n=i,a++}let o=this.length-e;return this.length=e,r>-1&&(this.children.length=r,this.markDirty()),new Wt(this.mark,t,o)}domAtPos(e){return Vt(this,e)}coordsAt(e,t){return Zt(this,e,t)}}class Xt extends _t{static create(e,t,n){return new Xt(e,t,n)}constructor(e,t,n){super(),this.widget=e,this.length=t,this.side=n,this.prevWidget=null}split(e){let t=Xt.create(this.widget,this.length-e,this.side);return this.length-=e,t}sync(e){this.dom&&this.widget.updateDOM(this.dom,e)||(this.dom&&this.prevWidget&&this.prevWidget.destroy(this.dom),this.prevWidget=null,this.setDOM(this.widget.toDOM(e)),this.widget.editable||(this.dom.contentEditable="false"))}getSide(){return this.side}merge(e,t,n,r,a,o){return!(n&&(!(n instanceof Xt&&this.widget.compare(n.widget))||e>0&&a<=0||t<this.length&&o<=0)||(this.length=e+(n?n.length:0)+(this.length-t),0))}become(e){return e instanceof Xt&&e.side==this.side&&this.widget.constructor==e.widget.constructor&&(this.widget.compare(e.widget)||this.markDirty(!0),this.dom&&!this.prevWidget&&(this.prevWidget=this.widget),this.widget=e.widget,this.length=e.length,!0)}ignoreMutation(){return!0}ignoreEvent(e){return this.widget.ignoreEvent(e)}get overrideDOMText(){if(0==this.length)return f.empty;let e=this;for(;e.parent;)e=e.parent;let{view:t}=e,n=t&&t.state.doc,r=this.posAtStart;return n?n.slice(r,r+this.length):f.empty}domAtPos(e){return(this.length?0==e:this.side>0)?Et.before(this.dom):Et.after(this.dom,e==this.length)}domBoundsAround(){return null}coordsAt(e,t){let n=this.widget.coordsAt(this.dom,e,t);if(n)return n;let r=this.dom.getClientRects(),a=null;if(!r.length)return null;let o=this.side?this.side<0:e>0;for(let t=o?r.length-1:0;a=r[t],!(e>0?0==t:t==r.length-1||a.top<a.bottom);t+=o?-1:1);return dt(a,!o)}get isEditable(){return!1}get isWidget(){return!0}get isHidden(){return this.widget.isHidden}destroy(){super.destroy(),this.dom&&this.widget.destroy(this.dom)}}class qt extends _t{constructor(e){super(),this.side=e}get length(){return 0}merge(){return!1}become(e){return e instanceof qt&&e.side==this.side}split(){return new qt(this.side)}sync(){if(!this.dom){let e=document.createElement("img");e.className="cm-widgetBuffer",e.setAttribute("aria-hidden","true"),this.setDOM(e)}}getSide(){return this.side}domAtPos(e){return this.side>0?Et.before(this.dom):Et.after(this.dom)}localPosFromDOM(){return 0}domBoundsAround(){return null}coordsAt(e){return this.dom.getBoundingClientRect()}get overrideDOMText(){return f.empty}get isHidden(){return!0}}function Vt(e,t){let n=e.dom,{children:r}=e,a=0;for(let e=0;a<r.length;a++){let o=r[a],i=e+o.length;if(!(i==e&&o.getSide()<=0)){if(t>e&&t<i&&o.dom.parentNode==n)return o.domAtPos(t-e);if(t<=e)break;e=i}}for(let e=a;e>0;e--){let t=r[e-1];if(t.dom.parentNode==n)return t.domAtPos(t.length)}for(let e=a;e<r.length;e++){let t=r[e];if(t.dom.parentNode==n)return t.domAtPos(0)}return new Et(n,0)}function Ht(e,t,n){let r,{children:a}=e;n>0&&t instanceof Wt&&a.length&&(r=a[a.length-1])instanceof Wt&&r.mark.eq(t.mark)?Ht(r,t.children[0],n-1):(a.push(t),t.setParent(e)),e.length+=t.length}function Zt(e,t,n){let r=null,a=-1,o=null,i=-1;!function e(t,s){for(let l=0,c=0;l<t.children.length&&c<=s;l++){let u=t.children[l],d=c+u.length;d>=s&&(u.children.length?e(u,s-c):(!o||o.isHidden&&n>0)&&(d>s||c==d&&u.getSide()>0)?(o=u,i=s-c):(c<s||c==d&&u.getSide()<0&&!u.isHidden)&&(r=u,a=s-c)),c=d}}(e,t);let s=(n<0?r:o)||r||o;return s?s.coordsAt(Math.max(0,s==r?a:i),n):function(e){let t=e.dom.lastChild;if(!t)return e.dom.getBoundingClientRect();let n=ot(t);return n[n.length-1]||null}(e)}function Gt(e,t){for(let n in e)"class"==n&&t.class?t.class+=" "+e.class:"style"==n&&t.style?t.style+=";"+e.style:t[n]=e[n];return t}jt.prototype.children=Xt.prototype.children=qt.prototype.children=St;const Kt=Object.create(null);function Yt(e,t,n){if(e==t)return!0;e||(e=Kt),t||(t=Kt);let r=Object.keys(e),a=Object.keys(t);if(r.length-(n&&r.indexOf(n)>-1?1:0)!=a.length-(n&&a.indexOf(n)>-1?1:0))return!1;for(let o of r)if(o!=n&&(-1==a.indexOf(o)||e[o]!==t[o]))return!1;return!0}function Jt(e,t,n){let r=!1;if(t)for(let a in t)n&&a in n||(r=!0,"style"==a?e.style.cssText="":e.removeAttribute(a));if(n)for(let a in n)t&&t[a]==n[a]||(r=!0,"style"==a?e.style.cssText=n[a]:e.setAttribute(a,n[a]));return r}function en(e){let t=Object.create(null);for(let n=0;n<e.attributes.length;n++){let r=e.attributes[n];t[r.name]=r.value}return t}class tn{eq(e){return!1}updateDOM(e,t){return!1}compare(e){return this==e||this.constructor==e.constructor&&this.eq(e)}get estimatedHeight(){return-1}get lineBreaks(){return 0}ignoreEvent(e){return!0}coordsAt(e,t,n){return null}get isHidden(){return!1}get editable(){return!1}destroy(e){}}var nn=function(e){return e[e.Text=0]="Text",e[e.WidgetBefore=1]="WidgetBefore",e[e.WidgetAfter=2]="WidgetAfter",e[e.WidgetRange=3]="WidgetRange",e}(nn||(nn={}));class rn extends Ce{constructor(e,t,n,r){super(),this.startSide=e,this.endSide=t,this.widget=n,this.spec=r}get heightRelevant(){return!1}static mark(e){return new an(e)}static widget(e){let t=Math.max(-1e4,Math.min(1e4,e.side||0)),n=!!e.block;return t+=n&&!e.inlineOrder?t>0?3e8:-4e8:t>0?1e8:-1e8,new sn(e,t,t,n,e.widget||null,!1)}static replace(e){let t,n,r=!!e.block;if(e.isBlockGap)t=-5e8,n=4e8;else{let{start:a,end:o}=ln(e,r);t=(a?r?-3e8:-1:5e8)-1,n=1+(o?r?2e8:1:-6e8)}return new sn(e,t,n,r,e.widget||null,!0)}static line(e){return new on(e)}static set(e,t=!1){return Pe.of(e,t)}hasHeight(){return!!this.widget&&this.widget.estimatedHeight>-1}}rn.none=Pe.empty;class an extends rn{constructor(e){let{start:t,end:n}=ln(e);super(t?-1:5e8,n?1:-6e8,null,e),this.tagName=e.tagName||"span",this.class=e.class||"",this.attrs=e.attributes||null}eq(e){var t,n;return this==e||e instanceof an&&this.tagName==e.tagName&&(this.class||(null===(t=this.attrs)||void 0===t?void 0:t.class))==(e.class||(null===(n=e.attrs)||void 0===n?void 0:n.class))&&Yt(this.attrs,e.attrs,"class")}range(e,t=e){if(e>=t)throw new RangeError("Mark decorations may not be empty");return super.range(e,t)}}an.prototype.point=!1;class on extends rn{constructor(e){super(-2e8,-2e8,null,e)}eq(e){return e instanceof on&&this.spec.class==e.spec.class&&Yt(this.spec.attributes,e.spec.attributes)}range(e,t=e){if(t!=e)throw new RangeError("Line decoration ranges must be zero-length");return super.range(e,t)}}on.prototype.mapMode=A.TrackBefore,on.prototype.point=!0;class sn extends rn{constructor(e,t,n,r,a,o){super(t,n,a,e),this.block=r,this.isReplace=o,this.mapMode=r?t<=0?A.TrackBefore:A.TrackAfter:A.TrackDel}get type(){return this.startSide!=this.endSide?nn.WidgetRange:this.startSide<=0?nn.WidgetBefore:nn.WidgetAfter}get heightRelevant(){return this.block||!!this.widget&&(this.widget.estimatedHeight>=5||this.widget.lineBreaks>0)}eq(e){return e instanceof sn&&(t=this.widget,n=e.widget,t==n||!!(t&&n&&t.compare(n)))&&this.block==e.block&&this.startSide==e.startSide&&this.endSide==e.endSide;var t,n}range(e,t=e){if(this.isReplace&&(e>t||e==t&&this.startSide>0&&this.endSide<=0))throw new RangeError("Invalid range for replacement decoration");if(!this.isReplace&&t!=e)throw new RangeError("Widget decorations can only have zero-length ranges");return super.range(e,t)}}function ln(e,t=!1){let{inclusiveStart:n,inclusiveEnd:r}=e;return null==n&&(n=e.inclusive),null==r&&(r=e.inclusive),{start:null!=n?n:t,end:null!=r?r:t}}function cn(e,t,n,r=0){let a=n.length-1;a>=0&&n[a]+r>=e?n[a]=Math.max(n[a],t):n.push(e,t)}sn.prototype.point=!0;class un extends _t{constructor(){super(...arguments),this.children=[],this.length=0,this.prevAttrs=void 0,this.attrs=null,this.breakAfter=0}merge(e,t,n,r,a,o){if(n){if(!(n instanceof un))return!1;this.dom||n.transferDOM(this)}return r&&this.setDeco(n?n.attrs:null),Rt(this,e,t,n?n.children.slice():[],a,o),!0}split(e){let t=new un;if(t.breakAfter=this.breakAfter,0==this.length)return t;let{i:n,off:r}=this.childPos(e);r&&(t.append(this.children[n].split(r),0),this.children[n].merge(r,this.children[n].length,null,!1,0,0),n++);for(let e=n;e<this.children.length;e++)t.append(this.children[e],0);for(;n>0&&0==this.children[n-1].length;)this.children[--n].destroy();return this.children.length=n,this.markDirty(),this.length=e,t}transferDOM(e){this.dom&&(this.markDirty(),e.setDOM(this.dom),e.prevAttrs=void 0===this.prevAttrs?this.attrs:this.prevAttrs,this.prevAttrs=void 0,this.dom=null)}setDeco(e){Yt(this.attrs,e)||(this.dom&&(this.prevAttrs=this.attrs,this.markDirty()),this.attrs=e)}append(e,t){Ht(this,e,t)}addLineDeco(e){let t=e.spec.attributes,n=e.spec.class;t&&(this.attrs=Gt(t,this.attrs||{})),n&&(this.attrs=Gt({class:n},this.attrs||{}))}domAtPos(e){return Vt(this,e)}reuseDOM(e){"DIV"==e.nodeName&&(this.setDOM(e),this.flags|=6)}sync(e,t){var n;this.dom?4&this.flags&&(wt(this.dom),this.dom.className="cm-line",this.prevAttrs=this.attrs?null:void 0):(this.setDOM(document.createElement("div")),this.dom.className="cm-line",this.prevAttrs=this.attrs?null:void 0),void 0!==this.prevAttrs&&(Jt(this.dom,this.prevAttrs,this.attrs),this.dom.classList.add("cm-line"),this.prevAttrs=void 0),super.sync(e,t);let r=this.dom.lastChild;for(;r&&_t.get(r)instanceof Wt;)r=r.lastChild;if(!(r&&this.length&&("BR"==r.nodeName||0!=(null===(n=_t.get(r))||void 0===n?void 0:n.isEditable)||zt.ios&&this.children.some((e=>e instanceof jt))))){let e=document.createElement("BR");e.cmIgnore=!0,this.dom.appendChild(e)}}measureTextSize(){if(0==this.children.length||this.length>20)return null;let e,t=0;for(let n of this.children){if(!(n instanceof jt)||/[^ -~]/.test(n.text))return null;let r=ot(n.dom);if(1!=r.length)return null;t+=r[0].width,e=r[0].height}return t?{lineHeight:this.dom.getBoundingClientRect().height,charWidth:t/this.length,textHeight:e}:null}coordsAt(e,t){let n=Zt(this,e,t);if(!this.children.length&&n&&this.parent){let{heightOracle:e}=this.parent.view.viewState,t=n.bottom-n.top;if(Math.abs(t-e.lineHeight)<2&&e.textHeight<t){let r=(t-e.textHeight)/2;return{top:n.top+r,bottom:n.bottom-r,left:n.left,right:n.left}}}return n}become(e){return e instanceof un&&0==this.children.length&&0==e.children.length&&Yt(this.attrs,e.attrs)&&this.breakAfter==e.breakAfter}covers(){return!0}static find(e,t){for(let n=0,r=0;n<e.children.length;n++){let a=e.children[n],o=r+a.length;if(o>=t){if(a instanceof un)return a;if(o>t)break}r=o+a.breakAfter}return null}}class dn extends _t{constructor(e,t,n){super(),this.widget=e,this.length=t,this.deco=n,this.breakAfter=0,this.prevWidget=null}merge(e,t,n,r,a,o){return!(n&&(!(n instanceof dn&&this.widget.compare(n.widget))||e>0&&a<=0||t<this.length&&o<=0)||(this.length=e+(n?n.length:0)+(this.length-t),0))}domAtPos(e){return 0==e?Et.before(this.dom):Et.after(this.dom,e==this.length)}split(e){let t=this.length-e;this.length=e;let n=new dn(this.widget,t,this.deco);return n.breakAfter=this.breakAfter,n}get children(){return St}sync(e){this.dom&&this.widget.updateDOM(this.dom,e)||(this.dom&&this.prevWidget&&this.prevWidget.destroy(this.dom),this.prevWidget=null,this.setDOM(this.widget.toDOM(e)),this.widget.editable||(this.dom.contentEditable="false"))}get overrideDOMText(){return this.parent?this.parent.view.state.doc.slice(this.posAtStart,this.posAtEnd):f.empty}domBoundsAround(){return null}become(e){return e instanceof dn&&e.widget.constructor==this.widget.constructor&&(e.widget.compare(this.widget)||this.markDirty(!0),this.dom&&!this.prevWidget&&(this.prevWidget=this.widget),this.widget=e.widget,this.length=e.length,this.deco=e.deco,this.breakAfter=e.breakAfter,!0)}ignoreMutation(){return!0}ignoreEvent(e){return this.widget.ignoreEvent(e)}get isEditable(){return!1}get isWidget(){return!0}coordsAt(e,t){return this.widget.coordsAt(this.dom,e,t)||(this.widget instanceof pn?null:dt(this.dom.getBoundingClientRect(),this.length?0==e:t<=0))}destroy(){super.destroy(),this.dom&&this.widget.destroy(this.dom)}covers(e){let{startSide:t,endSide:n}=this.deco;return t!=n&&(e<0?t<0:n>0)}}class pn extends tn{constructor(e){super(),this.height=e}toDOM(){let e=document.createElement("div");return e.className="cm-gap",this.updateDOM(e),e}eq(e){return e.height==this.height}updateDOM(e){return e.style.height=this.height+"px",!0}get editable(){return!0}get estimatedHeight(){return this.height}ignoreEvent(){return!1}}class hn{constructor(e,t,n,r){this.doc=e,this.pos=t,this.end=n,this.disallowBlockEffectsFor=r,this.content=[],this.curLine=null,this.breakAtStart=0,this.pendingBuffer=0,this.bufferMarks=[],this.atCursorPos=!0,this.openStart=-1,this.openEnd=-1,this.text="",this.textOff=0,this.cursor=e.iter(),this.skip=t}posCovered(){if(0==this.content.length)return!this.breakAtStart&&this.doc.lineAt(this.pos).from!=this.pos;let e=this.content[this.content.length-1];return!(e.breakAfter||e instanceof dn&&e.deco.endSide<0)}getLine(){return this.curLine||(this.content.push(this.curLine=new un),this.atCursorPos=!0),this.curLine}flushBuffer(e=this.bufferMarks){this.pendingBuffer&&(this.curLine.append(fn(new qt(-1),e),e.length),this.pendingBuffer=0)}addBlockWidget(e){this.flushBuffer(),this.curLine=null,this.content.push(e)}finish(e){this.pendingBuffer&&e<=this.bufferMarks.length?this.flushBuffer():this.pendingBuffer=0,this.posCovered()||e&&this.content.length&&this.content[this.content.length-1]instanceof dn||this.getLine()}buildText(e,t,n){for(;e>0;){if(this.textOff==this.text.length){let{value:t,lineBreak:n,done:r}=this.cursor.next(this.skip);if(this.skip=0,r)throw new Error("Ran out of text content when drawing inline views");if(n){this.posCovered()||this.getLine(),this.content.length?this.content[this.content.length-1].breakAfter=1:this.breakAtStart=1,this.flushBuffer(),this.curLine=null,this.atCursorPos=!0,e--;continue}this.text=t,this.textOff=0}let r=Math.min(this.text.length-this.textOff,e,512);this.flushBuffer(t.slice(t.length-n)),this.getLine().append(fn(new jt(this.text.slice(this.textOff,this.textOff+r)),t),n),this.atCursorPos=!0,this.textOff+=r,e-=r,n=0}}span(e,t,n,r){this.buildText(t-e,n,r),this.pos=t,this.openStart<0&&(this.openStart=r)}point(e,t,n,r,a,o){if(this.disallowBlockEffectsFor[o]&&n instanceof sn){if(n.block)throw new RangeError("Block decorations may not be specified via plugins");if(t>this.doc.lineAt(this.pos).to)throw new RangeError("Decorations that replace line breaks may not be specified via plugins")}let i=t-e;if(n instanceof sn)if(n.block)n.startSide>0&&!this.posCovered()&&this.getLine(),this.addBlockWidget(new dn(n.widget||mn.block,i,n));else{let o=Xt.create(n.widget||mn.inline,i,i?0:n.startSide),s=this.atCursorPos&&!o.isEditable&&a<=r.length&&(e<t||n.startSide>0),l=!o.isEditable&&(e<t||a>r.length||n.startSide<=0),c=this.getLine();2!=this.pendingBuffer||s||o.isEditable||(this.pendingBuffer=0),this.flushBuffer(r),s&&(c.append(fn(new qt(1),r),a),a=r.length+Math.max(0,a-r.length)),c.append(fn(o,r),a),this.atCursorPos=l,this.pendingBuffer=l?e<t||a>r.length?1:2:0,this.pendingBuffer&&(this.bufferMarks=r.slice())}else this.doc.lineAt(this.pos).from==this.pos&&this.getLine().addLineDeco(n);i&&(this.textOff+i<=this.text.length?this.textOff+=i:(this.skip+=i-(this.text.length-this.textOff),this.text="",this.textOff=0),this.pos=t),this.openStart<0&&(this.openStart=a)}static build(e,t,n,r,a){let o=new hn(e,t,n,a);return o.openEnd=Pe.spans(r,t,n,o),o.openStart<0&&(o.openStart=o.openEnd),o.finish(o.openEnd),o}}function fn(e,t){for(let n of t)e=new Wt(n,[e],e.length);return e}class mn extends tn{constructor(e){super(),this.tag=e}eq(e){return e.tag==this.tag}toDOM(){return document.createElement(this.tag)}updateDOM(e){return e.nodeName.toLowerCase()==this.tag}get isHidden(){return!0}}mn.inline=new mn("span"),mn.block=new mn("div");var gn=function(e){return e[e.LTR=0]="LTR",e[e.RTL=1]="RTL",e}(gn||(gn={}));const vn=gn.LTR,bn=gn.RTL;function yn(e){let t=[];for(let n=0;n<e.length;n++)t.push(1<<+e[n]);return t}const wn=yn("88888888888888888888888888888888888666888888787833333333337888888000000000000000000000000008888880000000000000000000000000088888888888888888888888888888888888887866668888088888663380888308888800000000000000000000000800000000000000000000000000000008"),On=yn("4444448826627288999999999992222222222222222222222222222222222222222222222229999999999999999999994444444444644222822222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222999999949999999229989999223333333333"),kn=Object.create(null),xn=[];for(let e of["()","[]","{}"]){let t=e.charCodeAt(0),n=e.charCodeAt(1);kn[t]=n,kn[n]=-t}function En(e){return e<=247?wn[e]:1424<=e&&e<=1524?2:1536<=e&&e<=1785?On[e-1536]:1774<=e&&e<=2220?4:8192<=e&&e<=8204?256:64336<=e&&e<=65023?4:1}const Sn=/[\u0590-\u05f4\u0600-\u06ff\u0700-\u08ac\ufb50-\ufdff]/;class _n{get dir(){return this.level%2?bn:vn}constructor(e,t,n){this.from=e,this.to=t,this.level=n}side(e,t){return this.dir==t==e?this.to:this.from}forward(e,t){return e==(this.dir==t)}static find(e,t,n,r){let a=-1;for(let o=0;o<e.length;o++){let i=e[o];if(i.from<=t&&i.to>=t){if(i.level==n)return o;(a<0||(0!=r?r<0?i.from<t:i.to>t:e[a].level>i.level))&&(a=o)}}if(a<0)throw new RangeError("Index out of range");return a}}function Cn(e,t){if(e.length!=t.length)return!1;for(let n=0;n<e.length;n++){let r=e[n],a=t[n];if(r.from!=a.from||r.to!=a.to||r.direction!=a.direction||!Cn(r.inner,a.inner))return!1}return!0}const Tn=[];function An(e,t,n,r,a,o,i){let s=r%2?2:1;if(r%2==a%2)for(let l=t,c=0;l<n;){let t=!0,u=!1;if(c==o.length||l<o[c].from){let e=Tn[l];e!=s&&(t=!1,u=16==e)}let d=t||1!=s?null:[],p=t?r:r+1,h=l;e:for(;;)if(c<o.length&&h==o[c].from){if(u)break e;let f=o[c];if(!t)for(let e=f.to,t=c+1;;){if(e==n)break e;if(!(t<o.length&&o[t].from==e)){if(Tn[e]==s)break e;break}e=o[t++].to}c++,d?d.push(f):(f.from>l&&i.push(new _n(l,f.from,p)),Rn(e,f.direction==vn!=!(p%2)?r+1:r,a,f.inner,f.from,f.to,i),l=f.to),h=f.to}else{if(h==n||(t?Tn[h]!=s:Tn[h]==s))break;h++}d?An(e,l,h,r+1,a,d,i):l<h&&i.push(new _n(l,h,p)),l=h}else for(let l=n,c=o.length;l>t;){let n=!0,u=!1;if(!c||l>o[c-1].to){let e=Tn[l-1];e!=s&&(n=!1,u=16==e)}let d=n||1!=s?null:[],p=n?r:r+1,h=l;e:for(;;)if(c&&h==o[c-1].to){if(u)break e;let f=o[--c];if(!n)for(let e=f.from,n=c;;){if(e==t)break e;if(!n||o[n-1].to!=e){if(Tn[e-1]==s)break e;break}e=o[--n].from}d?d.push(f):(f.to<l&&i.push(new _n(f.to,l,p)),Rn(e,f.direction==vn!=!(p%2)?r+1:r,a,f.inner,f.from,f.to,i),l=f.from),h=f.from}else{if(h==t||(n?Tn[h-1]!=s:Tn[h-1]==s))break;h--}d?An(e,h,l,r+1,a,d,i):h<l&&i.push(new _n(h,l,p)),l=h}}function Rn(e,t,n,r,a,o,i){let s=t%2?2:1;!function(e,t,n,r,a){for(let o=0;o<=r.length;o++){let i=o?r[o-1].to:t,s=o<r.length?r[o].from:n,l=o?256:a;for(let t=i,n=l,r=l;t<s;t++){let a=En(e.charCodeAt(t));512==a?a=n:8==a&&4==r&&(a=16),Tn[t]=4==a?2:a,7&a&&(r=a),n=a}for(let e=i,t=l,r=l;e<s;e++){let a=Tn[e];if(128==a)e<s-1&&t==Tn[e+1]&&24&t?a=Tn[e]=t:Tn[e]=256;else if(64==a){let a=e+1;for(;a<s&&64==Tn[a];)a++;let o=e&&8==t||a<n&&8==Tn[a]?1==r?1:8:256;for(let t=e;t<a;t++)Tn[t]=o;e=a-1}else 8==a&&1==r&&(Tn[e]=1);t=a,7&a&&(r=a)}}}(e,a,o,r,s),function(e,t,n,r,a){let o=1==a?2:1;for(let i=0,s=0,l=0;i<=r.length;i++){let c=i?r[i-1].to:t,u=i<r.length?r[i].from:n;for(let t,n,r,i=c;i<u;i++)if(n=kn[t=e.charCodeAt(i)])if(n<0){for(let e=s-3;e>=0;e-=3)if(xn[e+1]==-n){let t=xn[e+2],n=2&t?a:4&t?1&t?o:a:0;n&&(Tn[i]=Tn[xn[e]]=n),s=e;break}}else{if(189==xn.length)break;xn[s++]=i,xn[s++]=t,xn[s++]=l}else if(2==(r=Tn[i])||1==r){let e=r==a;l=e?0:1;for(let t=s-3;t>=0;t-=3){let n=xn[t+2];if(2&n)break;if(e)xn[t+2]|=2;else{if(4&n)break;xn[t+2]|=4}}}}}(e,a,o,r,s),function(e,t,n,r){for(let a=0,o=r;a<=n.length;a++){let i=a?n[a-1].to:e,s=a<n.length?n[a].from:t;for(let l=i;l<s;){let i=Tn[l];if(256==i){let i=l+1;for(;;)if(i==s){if(a==n.length)break;i=n[a++].to,s=a<n.length?n[a].from:t}else{if(256!=Tn[i])break;i++}let c=1==o,u=c==(1==(i<t?Tn[i]:r))?c?1:2:r;for(let t=i,r=a,o=r?n[r-1].to:e;t>l;)t==o&&(t=n[--r].from,o=r?n[r-1].to:e),Tn[--t]=u;l=i}else o=i,l++}}}(a,o,r,s),An(e,a,o,t,n,r,i)}function Pn(e){return[new _n(0,e,0)]}let Dn="";function Mn(e,t,n,r,a){var o;let i=r.head-e.from,s=_n.find(t,i,null!==(o=r.bidiLevel)&&void 0!==o?o:-1,r.assoc),l=t[s],c=l.side(a,n);if(i==c){let e=s+=a?1:-1;if(e<0||e>=t.length)return null;l=t[s=e],i=l.side(!a,n),c=l.side(a,n)}let u=E(e.text,i,l.forward(a,n));(u<l.from||u>l.to)&&(u=c),Dn=e.text.slice(Math.min(i,u),Math.max(i,u));let d=s==(a?t.length-1:0)?null:t[s+(a?1:-1)];return d&&u==c&&d.level+(a?0:1)<l.level?F.cursor(d.side(!a,n)+e.from,d.forward(a,n)?1:-1,d.level):F.cursor(u+e.from,l.forward(a,n)?-1:1,l.level)}function $n(e,t,n){for(let r=t;r<n;r++){let t=En(e.charCodeAt(r));if(1==t)return vn;if(2==t||4==t)return bn}return vn}const Nn=z.define(),In=z.define(),Ln=z.define(),Bn=z.define(),Fn=z.define(),Qn=z.define(),Un=z.define(),zn=z.define(),jn=z.define(),Wn=z.define({combine:e=>e.some((e=>e))}),Xn=z.define({combine:e=>e.some((e=>e))}),qn=z.define();class Vn{constructor(e,t="nearest",n="nearest",r=5,a=5,o=!1){this.range=e,this.y=t,this.x=n,this.yMargin=r,this.xMargin=a,this.isSnapshot=o}map(e){return e.empty?this:new Vn(this.range.map(e),this.y,this.x,this.yMargin,this.xMargin,this.isSnapshot)}clip(e){return this.range.to<=e.doc.length?this:new Vn(F.cursor(e.doc.length),this.y,this.x,this.yMargin,this.xMargin,this.isSnapshot)}}const Hn=fe.define({map:(e,t)=>e.map(t)}),Zn=fe.define();function Gn(e,t,n){let r=e.facet(Bn);r.length?r[0](t):window.onerror?window.onerror(String(t),n,void 0,void 0,t):n?console.error(n+":",t):console.error(t)}const Kn=z.define({combine:e=>!e.length||e[0]});let Yn=0;const Jn=z.define();class er{constructor(e,t,n,r,a){this.id=e,this.create=t,this.domEventHandlers=n,this.domEventObservers=r,this.extension=a(this)}static define(e,t){const{eventHandlers:n,eventObservers:r,provide:a,decorations:o}=t||{};return new er(Yn++,e,n,r,(e=>{let t=[Jn.of(e)];return o&&t.push(ar.of((t=>{let n=t.plugin(e);return n?o(n):rn.none}))),a&&t.push(a(e)),t}))}static fromClass(e,t){return er.define((t=>new e(t)),t)}}class tr{constructor(e){this.spec=e,this.mustUpdate=null,this.value=null}update(e){if(this.value){if(this.mustUpdate){let e=this.mustUpdate;if(this.mustUpdate=null,this.value.update)try{this.value.update(e)}catch(t){if(Gn(e.state,t,"CodeMirror plugin crashed"),this.value.destroy)try{this.value.destroy()}catch(e){}this.deactivate()}}}else if(this.spec)try{this.value=this.spec.create(e)}catch(t){Gn(e.state,t,"CodeMirror plugin crashed"),this.deactivate()}return this}destroy(e){var t;if(null===(t=this.value)||void 0===t?void 0:t.destroy)try{this.value.destroy()}catch(t){Gn(e.state,t,"CodeMirror plugin crashed")}}deactivate(){this.spec=this.value=null}}const nr=z.define(),rr=z.define(),ar=z.define(),or=z.define(),ir=z.define(),sr=z.define();function lr(e,t){let n=e.state.facet(sr);if(!n.length)return n;let r=n.map((t=>t instanceof Function?t(e):t)),a=[];return Pe.spans(r,t.from,t.to,{point(){},span(e,n,r,o){let i=e-t.from,s=n-t.from,l=a;for(let e=r.length-1;e>=0;e--,o--){let n,a=r[e].spec.bidiIsolate;if(null==a&&(a=$n(t.text,i,s)),o>0&&l.length&&(n=l[l.length-1]).to==i&&n.direction==a)n.to=s,l=n.inner;else{let e={from:i,to:s,direction:a,inner:[]};l.push(e),l=e.inner}}}}),a}const cr=z.define();function ur(e){let t=0,n=0,r=0,a=0;for(let o of e.state.facet(cr)){let i=o(e);i&&(null!=i.left&&(t=Math.max(t,i.left)),null!=i.right&&(n=Math.max(n,i.right)),null!=i.top&&(r=Math.max(r,i.top)),null!=i.bottom&&(a=Math.max(a,i.bottom)))}return{left:t,right:n,top:r,bottom:a}}const dr=z.define();class pr{constructor(e,t,n,r){this.fromA=e,this.toA=t,this.fromB=n,this.toB=r}join(e){return new pr(Math.min(this.fromA,e.fromA),Math.max(this.toA,e.toA),Math.min(this.fromB,e.fromB),Math.max(this.toB,e.toB))}addToSet(e){let t=e.length,n=this;for(;t>0;t--){let r=e[t-1];if(!(r.fromA>n.toA)){if(r.toA<n.fromA)break;n=n.join(r),e.splice(t-1,1)}}return e.splice(t,0,n),e}static extendWithRanges(e,t){if(0==t.length)return e;let n=[];for(let r=0,a=0,o=0,i=0;;r++){let s=r==e.length?null:e[r],l=o-i,c=s?s.fromB:1e9;for(;a<t.length&&t[a]<c;){let e=t[a],r=t[a+1],o=Math.max(i,e),s=Math.min(c,r);if(o<=s&&new pr(o+l,s+l,o,s).addToSet(n),r>c)break;a+=2}if(!s)return n;new pr(s.fromA,s.toA,s.fromB,s.toB).addToSet(n),o=s.toA,i=s.toB}}}class hr{constructor(e,t,n){this.view=e,this.state=t,this.transactions=n,this.flags=0,this.startState=e.state,this.changes=P.empty(this.startState.doc.length);for(let e of n)this.changes=this.changes.compose(e.changes);let r=[];this.changes.iterChangedRanges(((e,t,n,a)=>r.push(new pr(e,t,n,a)))),this.changedRanges=r}static create(e,t,n){return new hr(e,t,n)}get viewportChanged(){return(4&this.flags)>0}get viewportMoved(){return(8&this.flags)>0}get heightChanged(){return(2&this.flags)>0}get geometryChanged(){return this.docChanged||(18&this.flags)>0}get focusChanged(){return(1&this.flags)>0}get docChanged(){return!this.changes.empty}get selectionSet(){return this.transactions.some((e=>e.selection))}get empty(){return 0==this.flags&&0==this.transactions.length}}class fr extends _t{get length(){return this.view.state.doc.length}constructor(e){super(),this.view=e,this.decorations=[],this.dynamicDecorationMap=[!1],this.domChanged=null,this.hasComposition=null,this.markedForComposition=new Set,this.editContextFormatting=rn.none,this.lastCompositionAfterCursor=!1,this.minWidth=0,this.minWidthFrom=0,this.minWidthTo=0,this.impreciseAnchor=null,this.impreciseHead=null,this.forceSelection=!1,this.lastUpdate=Date.now(),this.setDOM(e.contentDOM),this.children=[new un],this.children[0].setParent(this),this.updateDeco(),this.updateInner([new pr(0,0,0,e.state.doc.length)],0,null)}update(e){var t;let n=e.changedRanges;this.minWidth>0&&n.length&&(n.every((({fromA:e,toA:t})=>t<this.minWidthFrom||e>this.minWidthTo))?(this.minWidthFrom=e.changes.mapPos(this.minWidthFrom,1),this.minWidthTo=e.changes.mapPos(this.minWidthTo,1)):this.minWidth=this.minWidthFrom=this.minWidthTo=0),this.updateEditContextFormatting(e);let r=-1;this.view.inputState.composing>=0&&!this.view.observer.editContext&&((null===(t=this.domChanged)||void 0===t?void 0:t.newSel)?r=this.domChanged.newSel.head:function(e,t){let n=!1;return t&&e.iterChangedRanges(((e,r)=>{e<t.to&&r>t.from&&(n=!0)})),n}(e.changes,this.hasComposition)||e.selectionSet||(r=e.state.selection.main.head));let a=r>-1?function(e,t,n){let r=mr(e,n);if(!r)return null;let{node:a,from:o,to:i}=r,s=a.nodeValue;if(/[\n\r]/.test(s))return null;if(e.state.doc.sliceString(r.from,r.to)!=s)return null;let l=t.invertedDesc,c=new pr(l.mapPos(o),l.mapPos(i),o,i),u=[];for(let t=a.parentNode;;t=t.parentNode){let n=_t.get(t);if(n instanceof Wt)u.push({node:t,deco:n.mark});else{if(n instanceof un||"DIV"==t.nodeName&&t.parentNode==e.contentDOM)return{range:c,text:a,marks:u,line:t};if(t==e.contentDOM)return null;u.push({node:t,deco:new an({inclusive:!0,attributes:en(t),tagName:t.tagName.toLowerCase()})})}}}(this.view,e.changes,r):null;if(this.domChanged=null,this.hasComposition){this.markedForComposition.clear();let{from:t,to:r}=this.hasComposition;n=new pr(t,r,e.changes.mapPos(t,-1),e.changes.mapPos(r,1)).addToSet(n.slice())}this.hasComposition=a?{from:a.range.fromB,to:a.range.toB}:null,(zt.ie||zt.chrome)&&!a&&e&&e.state.doc.lines!=e.startState.doc.lines&&(this.forceSelection=!0);let o=function(e,t,n){let r=new gr;return Pe.compare(e,t,n,r),r.changes}(this.decorations,this.updateDeco(),e.changes);return n=pr.extendWithRanges(n,o),!!(7&this.flags||0!=n.length)&&(this.updateInner(n,e.startState.doc.length,a),e.transactions.length&&(this.lastUpdate=Date.now()),!0)}updateInner(e,t,n){this.view.viewState.mustMeasureContent=!0,this.updateChildren(e,t,n);let{observer:r}=this.view;r.ignore((()=>{this.dom.style.height=this.view.viewState.contentHeight/this.view.scaleY+"px",this.dom.style.flexBasis=this.minWidth?this.minWidth+"px":"";let e=zt.chrome||zt.ios?{node:r.selectionRange.focusNode,written:!1}:void 0;this.sync(this.view,e),this.flags&=-8,e&&(e.written||r.selectionRange.focusNode!=e.node)&&(this.forceSelection=!0),this.dom.style.height=""})),this.markedForComposition.forEach((e=>e.flags&=-9));let a=[];if(this.view.viewport.from||this.view.viewport.to<this.view.state.doc.length)for(let e of this.children)e instanceof dn&&e.widget instanceof pn&&a.push(e.dom);r.updateGaps(a)}updateChildren(e,t,n){let r=n?n.range.addToSet(e.slice()):e,a=this.childCursor(t);for(let e=r.length-1;;e--){let t=e>=0?r[e]:null;if(!t)break;let o,i,s,l,{fromA:c,toA:u,fromB:d,toB:p}=t;if(n&&n.range.fromB<p&&n.range.toB>d){let e=hn.build(this.view.state.doc,d,n.range.fromB,this.decorations,this.dynamicDecorationMap),t=hn.build(this.view.state.doc,n.range.toB,p,this.decorations,this.dynamicDecorationMap);i=e.breakAtStart,s=e.openStart,l=t.openEnd;let r=this.compositionView(n);t.breakAtStart?r.breakAfter=1:t.content.length&&r.merge(r.length,r.length,t.content[0],!1,t.openStart,0)&&(r.breakAfter=t.content[0].breakAfter,t.content.shift()),e.content.length&&r.merge(0,0,e.content[e.content.length-1],!0,0,e.openEnd)&&e.content.pop(),o=e.content.concat(r).concat(t.content)}else({content:o,breakAtStart:i,openStart:s,openEnd:l}=hn.build(this.view.state.doc,d,p,this.decorations,this.dynamicDecorationMap));let{i:h,off:f}=a.findPos(u,1),{i:m,off:g}=a.findPos(c,-1);At(this,m,g,h,f,o,i,s,l)}n&&this.fixCompositionDOM(n)}updateEditContextFormatting(e){this.editContextFormatting=this.editContextFormatting.map(e.changes);for(let t of e.transactions)for(let e of t.effects)e.is(Zn)&&(this.editContextFormatting=e.value)}compositionView(e){let t=new jt(e.text.nodeValue);t.flags|=8;for(let{deco:n}of e.marks)t=new Wt(n,[t],t.length);let n=new un;return n.append(t,0),n}fixCompositionDOM(e){let t=(e,t)=>{t.flags|=8|(t.children.some((e=>7&e.flags))?1:0),this.markedForComposition.add(t);let n=_t.get(e);n&&n!=t&&(n.dom=null),t.setDOM(e)},n=this.childPos(e.range.fromB,1),r=this.children[n.i];t(e.line,r);for(let a=e.marks.length-1;a>=-1;a--)n=r.childPos(n.off,1),r=r.children[n.i],t(a>=0?e.marks[a].node:e.text,r)}updateSelection(e=!1,t=!1){!e&&this.view.observer.selectionRange.focusNode||this.view.observer.readSelectionRange();let n=this.view.root.activeElement,r=n==this.dom,a=!r&&!(this.view.state.facet(Kn)||this.dom.tabIndex>-1)&&at(this.dom,this.view.observer.selectionRange)&&!(n&&this.dom.contains(n));if(!(r||t||a))return;let o=this.forceSelection;this.forceSelection=!1;let i=this.view.state.selection.main,s=this.moveToLine(this.domAtPos(i.anchor)),l=i.empty?s:this.moveToLine(this.domAtPos(i.head));if(zt.gecko&&i.empty&&!this.hasComposition&&1==(c=s).node.nodeType&&c.node.firstChild&&(0==c.offset||"false"==c.node.childNodes[c.offset-1].contentEditable)&&(c.offset==c.node.childNodes.length||"false"==c.node.childNodes[c.offset].contentEditable)){let e=document.createTextNode("");this.view.observer.ignore((()=>s.node.insertBefore(e,s.node.childNodes[s.offset]||null))),s=l=new Et(e,0),o=!0}var c;let u=this.view.observer.selectionRange;!o&&u.focusNode&&(it(s.node,s.offset,u.anchorNode,u.anchorOffset)&&it(l.node,l.offset,u.focusNode,u.focusOffset)||this.suppressWidgetCursorChange(u,i))||(this.view.observer.ignore((()=>{zt.android&&zt.chrome&&this.dom.contains(u.focusNode)&&function(e,t){for(let n=e;n&&n!=t;n=n.assignedSlot||n.parentNode)if(1==n.nodeType&&"false"==n.contentEditable)return!0;return!1}(u.focusNode,this.dom)&&(this.dom.blur(),this.dom.focus({preventScroll:!0}));let e=nt(this.view.root);if(e)if(i.empty){if(zt.gecko){let e=(t=s.node,r=s.offset,1!=t.nodeType?0:(r&&"false"==t.childNodes[r-1].contentEditable?1:0)|(r<t.childNodes.length&&"false"==t.childNodes[r].contentEditable?2:0));if(e&&3!=e){let t=(1==e?kt:xt)(s.node,s.offset);t&&(s=new Et(t.node,t.offset))}}e.collapse(s.node,s.offset),null!=i.bidiLevel&&void 0!==e.caretBidiLevel&&(e.caretBidiLevel=i.bidiLevel)}else if(e.extend){e.collapse(s.node,s.offset);try{e.extend(l.node,l.offset)}catch(e){}}else{let t=document.createRange();i.anchor>i.head&&([s,l]=[l,s]),t.setEnd(l.node,l.offset),t.setStart(s.node,s.offset),e.removeAllRanges(),e.addRange(t)}var t,r;a&&this.view.root.activeElement==this.dom&&(this.dom.blur(),n&&n.focus())})),this.view.observer.setSelectionRange(s,l)),this.impreciseAnchor=s.precise?null:new Et(u.anchorNode,u.anchorOffset),this.impreciseHead=l.precise?null:new Et(u.focusNode,u.focusOffset)}suppressWidgetCursorChange(e,t){return this.hasComposition&&t.empty&&it(e.focusNode,e.focusOffset,e.anchorNode,e.anchorOffset)&&this.posFromDOM(e.focusNode,e.focusOffset)==t.head}enforceCursorAssoc(){if(this.hasComposition)return;let{view:e}=this,t=e.state.selection.main,n=nt(e.root),{anchorNode:r,anchorOffset:a}=e.observer.selectionRange;if(!(n&&t.empty&&t.assoc&&n.modify))return;let o=un.find(this,t.head);if(!o)return;let i=o.posAtStart;if(t.head==i||t.head==i+o.length)return;let s=this.coordsAt(t.head,-1),l=this.coordsAt(t.head,1);if(!s||!l||s.bottom>l.top)return;let c=this.domAtPos(t.head+t.assoc);n.collapse(c.node,c.offset),n.modify("move",t.assoc<0?"forward":"backward","lineboundary"),e.observer.readSelectionRange();let u=e.observer.selectionRange;e.docView.posFromDOM(u.anchorNode,u.anchorOffset)!=t.from&&n.collapse(r,a)}moveToLine(e){let t,n=this.dom;if(e.node!=n)return e;for(let r=e.offset;!t&&r<n.childNodes.length;r++){let e=_t.get(n.childNodes[r]);e instanceof un&&(t=e.domAtPos(0))}for(let r=e.offset-1;!t&&r>=0;r--){let e=_t.get(n.childNodes[r]);e instanceof un&&(t=e.domAtPos(e.length))}return t?new Et(t.node,t.offset,!0):e}nearest(e){for(let t=e;t;){let e=_t.get(t);if(e&&e.rootView==this)return e;t=t.parentNode}return null}posFromDOM(e,t){let n=this.nearest(e);if(!n)throw new RangeError("Trying to find position for a DOM position outside of the document");return n.localPosFromDOM(e,t)+n.posAtStart}domAtPos(e){let{i:t,off:n}=this.childCursor().findPos(e,-1);for(;t<this.children.length-1;){let e=this.children[t];if(n<e.length||e instanceof un)break;t++,n=0}return this.children[t].domAtPos(n)}coordsAt(e,t){let n=null,r=0;for(let a=this.length,o=this.children.length-1;o>=0;o--){let i=this.children[o],s=a-i.breakAfter,l=s-i.length;if(s<e)break;if(l<=e&&(l<e||i.covers(-1))&&(s>e||i.covers(1))&&(!n||i instanceof un&&!(n instanceof un&&t>=0)))n=i,r=l;else if(n&&l==e&&s==e&&i instanceof dn&&Math.abs(t)<2){if(i.deco.startSide<0)break;o&&(n=null)}a=l}return n?n.coordsAt(e-r,t):null}coordsForChar(e){let{i:t,off:n}=this.childPos(e,1),r=this.children[t];if(!(r instanceof un))return null;for(;r.children.length;){let{i:e,off:t}=r.childPos(n,1);for(;;e++){if(e==r.children.length)return null;if((r=r.children[e]).length)break}n=t}if(!(r instanceof jt))return null;let a=E(r.text,n);if(a==n)return null;let o=bt(r.dom,n,a).getClientRects();for(let e=0;e<o.length;e++){let t=o[e];if(e==o.length-1||t.top<t.bottom&&t.left<t.right)return t}return null}measureVisibleLineHeights(e){let t=[],{from:n,to:r}=e,a=this.view.contentDOM.clientWidth,o=a>Math.max(this.view.scrollDOM.clientWidth,this.minWidth)+1,i=-1,s=this.view.textDirection==gn.LTR;for(let e=0,l=0;l<this.children.length;l++){let c=this.children[l],u=e+c.length;if(u>r)break;if(e>=n){let n=c.dom.getBoundingClientRect();if(t.push(n.height),o){let t=c.dom.lastChild,r=t?ot(t):[];if(r.length){let t=r[r.length-1],o=s?t.right-n.left:n.right-t.left;o>i&&(i=o,this.minWidth=a,this.minWidthFrom=e,this.minWidthTo=u)}}}e=u+c.breakAfter}return t}textDirectionAt(e){let{i:t}=this.childPos(e,1);return"rtl"==getComputedStyle(this.children[t].dom).direction?gn.RTL:gn.LTR}measureTextSize(){for(let e of this.children)if(e instanceof un){let t=e.measureTextSize();if(t)return t}let e,t,n,r=document.createElement("div");return r.className="cm-line",r.style.width="99999px",r.style.position="absolute",r.textContent="abc def ghi jkl mno pqr stu",this.view.observer.ignore((()=>{this.dom.appendChild(r);let a=ot(r.firstChild)[0];e=r.getBoundingClientRect().height,t=a?a.width/27:7,n=a?a.height:e,r.remove()})),{lineHeight:e,charWidth:t,textHeight:n}}childCursor(e=this.length){let t=this.children.length;return t&&(e-=this.children[--t].length),new Tt(this.children,e,t)}computeBlockGapDeco(){let e=[],t=this.view.viewState;for(let n=0,r=0;;r++){let a=r==t.viewports.length?null:t.viewports[r],o=a?a.from-1:this.length;if(o>n){let r=(t.lineBlockAt(o).bottom-t.lineBlockAt(n).top)/this.view.scaleY;e.push(rn.replace({widget:new pn(r),block:!0,inclusive:!0,isBlockGap:!0}).range(n,o))}if(!a)break;n=a.to+1}return rn.set(e)}updateDeco(){let e=1,t=this.view.state.facet(ar).map((t=>(this.dynamicDecorationMap[e++]="function"==typeof t)?t(this.view):t)),n=!1,r=this.view.state.facet(or).map(((e,t)=>{let r="function"==typeof e;return r&&(n=!0),r?e(this.view):e}));for(r.length&&(this.dynamicDecorationMap[e++]=n,t.push(Pe.join(r))),this.decorations=[this.editContextFormatting,...t,this.computeBlockGapDeco(),this.view.viewState.lineGapDeco];e<this.decorations.length;)this.dynamicDecorationMap[e++]=!1;return this.decorations}scrollIntoView(e){if(e.isSnapshot){let t=this.view.viewState.lineBlockAt(e.range.head);return this.view.scrollDOM.scrollTop=t.top-e.yMargin,void(this.view.scrollDOM.scrollLeft=e.xMargin)}for(let t of this.view.state.facet(qn))try{if(t(this.view,e.range,e))return!0}catch(e){Gn(this.view.state,e,"scroll handler")}let t,{range:n}=e,r=this.coordsAt(n.head,n.empty?n.assoc:n.head>n.anchor?-1:1);if(!r)return;!n.empty&&(t=this.coordsAt(n.anchor,n.anchor>n.head?-1:1))&&(r={left:Math.min(r.left,t.left),top:Math.min(r.top,t.top),right:Math.max(r.right,t.right),bottom:Math.max(r.bottom,t.bottom)});let a=ur(this.view),o={left:r.left-a.left,top:r.top-a.top,right:r.right+a.right,bottom:r.bottom+a.bottom},{offsetWidth:i,offsetHeight:s}=this.view.scrollDOM;!function(e,t,n,r,a,o,i,s){let l=e.ownerDocument,c=l.defaultView||window;for(let u=e,d=!1;u&&!d;)if(1==u.nodeType){let e,p=u==l.body,h=1,f=1;if(p)e=pt(c);else{if(/^(fixed|sticky)$/.test(getComputedStyle(u).position)&&(d=!0),u.scrollHeight<=u.clientHeight&&u.scrollWidth<=u.clientWidth){u=u.assignedSlot||u.parentNode;continue}let t=u.getBoundingClientRect();({scaleX:h,scaleY:f}=ht(u,t)),e={left:t.left,right:t.left+u.clientWidth*h,top:t.top,bottom:t.top+u.clientHeight*f}}let m=0,g=0;if("nearest"==a)t.top<e.top?(g=-(e.top-t.top+i),n>0&&t.bottom>e.bottom+g&&(g=t.bottom-e.bottom+g+i)):t.bottom>e.bottom&&(g=t.bottom-e.bottom+i,n<0&&t.top-g<e.top&&(g=-(e.top+g-t.top+i)));else{let r=t.bottom-t.top,o=e.bottom-e.top;g=("center"==a&&r<=o?t.top+r/2-o/2:"start"==a||"center"==a&&n<0?t.top-i:t.bottom-o+i)-e.top}if("nearest"==r?t.left<e.left?(m=-(e.left-t.left+o),n>0&&t.right>e.right+m&&(m=t.right-e.right+m+o)):t.right>e.right&&(m=t.right-e.right+o,n<0&&t.left<e.left+m&&(m=-(e.left+m-t.left+o))):m=("center"==r?t.left+(t.right-t.left)/2-(e.right-e.left)/2:"start"==r==s?t.left-o:t.right-(e.right-e.left)+o)-e.left,m||g)if(p)c.scrollBy(m,g);else{let e=0,n=0;if(g){let e=u.scrollTop;u.scrollTop+=g/f,n=(u.scrollTop-e)*f}if(m){let t=u.scrollLeft;u.scrollLeft+=m/h,e=(u.scrollLeft-t)*h}t={left:t.left-e,top:t.top-n,right:t.right-e,bottom:t.bottom-n},e&&Math.abs(e-m)<1&&(r="nearest"),n&&Math.abs(n-g)<1&&(a="nearest")}if(p)break;u=u.assignedSlot||u.parentNode}else{if(11!=u.nodeType)break;u=u.host}}(this.view.scrollDOM,o,n.head<n.anchor?-1:1,e.x,e.y,Math.max(Math.min(e.xMargin,i),-i),Math.max(Math.min(e.yMargin,s),-s),this.view.textDirection==gn.LTR)}}function mr(e,t){let n=e.observer.selectionRange;if(!n.focusNode)return null;let r=kt(n.focusNode,n.focusOffset),a=xt(n.focusNode,n.focusOffset),o=r||a;if(a&&r&&a.node!=r.node){let t=_t.get(a.node);if(!t||t instanceof jt&&t.text!=a.node.nodeValue)o=a;else if(e.docView.lastCompositionAfterCursor){let e=_t.get(r.node);!e||e instanceof jt&&e.text!=r.node.nodeValue||(o=a)}}if(e.docView.lastCompositionAfterCursor=o!=r,!o)return null;let i=t-o.offset;return{from:i,to:i+o.node.nodeValue.length,node:o.node}}let gr=class{constructor(){this.changes=[]}compareRange(e,t){cn(e,t,this.changes)}comparePoint(e,t){cn(e,t,this.changes)}boundChange(e){cn(e,e,this.changes)}};function vr(e,t){return t.left>e?t.left-e:Math.max(0,e-t.right)}function br(e,t){return t.top>e?t.top-e:Math.max(0,e-t.bottom)}function yr(e,t){return e.top<t.bottom-1&&e.bottom>t.top+1}function wr(e,t){return t<e.top?{top:t,left:e.left,right:e.right,bottom:e.bottom}:e}function Or(e,t){return t>e.bottom?{top:e.top,left:e.left,right:e.right,bottom:t}:e}function kr(e,t,n){let r,a,o,i,s,l,c,u,d=!1;for(let p=e.firstChild;p;p=p.nextSibling){let e=ot(p);for(let h=0;h<e.length;h++){let f=e[h];a&&yr(a,f)&&(f=wr(Or(f,a.bottom),a.top));let m=vr(t,f),g=br(n,f);if(0==m&&0==g)return 3==p.nodeType?xr(p,t,n):kr(p,t,n);if(!r||i>g||i==g&&o>m){r=p,a=f,o=m,i=g;let s=g?n<f.top?-1:1:m?t<f.left?-1:1:0;d=!s||(s>0?h<e.length-1:h>0)}0==m?n>f.bottom&&(!c||c.bottom<f.bottom)?(s=p,c=f):n<f.top&&(!u||u.top>f.top)&&(l=p,u=f):c&&yr(c,f)?c=Or(c,f.bottom):u&&yr(u,f)&&(u=wr(u,f.top))}}if(c&&c.bottom>=n?(r=s,a=c):u&&u.top<=n&&(r=l,a=u),!r)return{node:e,offset:0};let p=Math.max(a.left,Math.min(a.right,t));return 3==r.nodeType?xr(r,p,n):d&&"false"!=r.contentEditable?kr(r,p,n):{node:e,offset:Array.prototype.indexOf.call(e.childNodes,r)+(t>=(a.left+a.right)/2?1:0)}}function xr(e,t,n){let r=e.nodeValue.length,a=-1,o=1e9,i=0;for(let s=0;s<r;s++){let r=bt(e,s,s+1).getClientRects();for(let l=0;l<r.length;l++){let c=r[l];if(c.top==c.bottom)continue;i||(i=t-c.left);let u=(c.top>n?c.top-n:n-c.bottom)-1;if(c.left-1<=t&&c.right+1>=t&&u<o){let n=t>=(c.left+c.right)/2,r=n;if((zt.chrome||zt.gecko)&&bt(e,s).getBoundingClientRect().left==c.right&&(r=!n),u<=0)return{node:e,offset:s+(r?1:0)};a=s+(r?1:0),o=u}}}return{node:e,offset:a>-1?a:i>0?e.nodeValue.length:0}}function Er(e,t,n,r=-1){var a,o;let i,s=e.contentDOM.getBoundingClientRect(),l=s.top+e.viewState.paddingTop,{docHeight:c}=e.viewState,{x:u,y:d}=t,p=d-l;if(p<0)return 0;if(p>c)return e.state.doc.length;for(let t=e.viewState.heightOracle.textHeight/2,a=!1;i=e.elementAtHeight(p),i.type!=nn.Text;)for(;p=r>0?i.bottom+t:i.top-t,!(p>=0&&p<=c);){if(a)return n?null:0;a=!0,r=-r}d=l+p;let h=i.from;if(h<e.viewport.from)return 0==e.viewport.from?0:n?null:Sr(e,s,i,u,d);if(h>e.viewport.to)return e.viewport.to==e.state.doc.length?e.state.doc.length:n?null:Sr(e,s,i,u,d);let f=e.dom.ownerDocument,m=e.root.elementFromPoint?e.root:f,g=m.elementFromPoint(u,d);g&&!e.contentDOM.contains(g)&&(g=null),g||(u=Math.max(s.left+1,Math.min(s.right-1,u)),g=m.elementFromPoint(u,d),g&&!e.contentDOM.contains(g)&&(g=null));let v,b=-1;if(g&&0!=(null===(a=e.docView.nearest(g))||void 0===a?void 0:a.isEditable)){if(f.caretPositionFromPoint){let e=f.caretPositionFromPoint(u,d);e&&({offsetNode:v,offset:b}=e)}else if(f.caretRangeFromPoint){let t=f.caretRangeFromPoint(u,d);t&&(({startContainer:v,startOffset:b}=t),(!e.contentDOM.contains(v)||zt.safari&&function(e,t,n){let r;if(3!=e.nodeType||t!=(r=e.nodeValue.length))return!1;for(let t=e.nextSibling;t;t=t.nextSibling)if(1!=t.nodeType||"BR"!=t.nodeName)return!1;return bt(e,r-1,r).getBoundingClientRect().left>n}(v,b,u)||zt.chrome&&function(e,t,n){if(0!=t)return!1;for(let t=e;;){let e=t.parentNode;if(!e||1!=e.nodeType||e.firstChild!=t)return!1;if(e.classList.contains("cm-line"))break;t=e}return n-(1==e.nodeType?e.getBoundingClientRect():bt(e,0,Math.max(e.nodeValue.length,1)).getBoundingClientRect()).left>5}(v,b,u))&&(v=void 0))}v&&(b=Math.min(ut(v),b))}if(!v||!e.docView.dom.contains(v)){let t=un.find(e.docView,h);if(!t)return p>i.top+i.height/2?i.to:i.from;({node:v,offset:b}=kr(t.dom,u,d))}let y=e.docView.nearest(v);if(!y)return null;if(y.isWidget&&1==(null===(o=y.dom)||void 0===o?void 0:o.nodeType)){let e=y.dom.getBoundingClientRect();return t.y<e.top||t.y<=e.bottom&&t.x<=(e.left+e.right)/2?y.posAtStart:y.posAtEnd}return y.localPosFromDOM(v,b)+y.posAtStart}function Sr(e,t,n,r,a){let o=Math.round((r-t.left)*e.defaultCharacterWidth);if(e.lineWrapping&&n.height>1.5*e.defaultLineHeight){let t=e.viewState.heightOracle.textHeight;o+=Math.floor((a-n.top-.5*(e.defaultLineHeight-t))/t)*e.viewState.heightOracle.lineLength}let i=e.state.sliceDoc(n.from,n.to);return n.from+function(e,t,n){for(let r=0,a=0;;){if(a>=t)return r;if(r==e.length)break;a+=9==e.charCodeAt(r)?n-a%n:1,r=E(e,r)}return e.length}(i,o,e.state.tabSize)}function _r(e,t,n,r){let a=e.state.doc.lineAt(t.head),o=e.bidiSpans(a),i=e.textDirectionAt(a.from);for(let s=t,l=null;;){let t=Mn(a,o,i,s,n),c=Dn;if(!t){if(a.number==(n?e.state.doc.lines:1))return s;c="\n",a=e.state.doc.line(a.number+(n?1:-1)),o=e.bidiSpans(a),t=e.visualLineSide(a,!n)}if(l){if(!l(c))return s}else{if(!r)return t;l=r(c)}s=t}}function Cr(e,t,n){for(;;){let r=0;for(let a of e)a.between(t-1,t+1,((e,a,o)=>{if(t>e&&t<a){let o=r||n||(t-e<a-t?-1:1);t=o<0?e:a,r=o}}));if(!r)return t}}function Tr(e,t,n){let r=Cr(e.state.facet(ir).map((t=>t(e))),n.from,t.head>n.from?-1:1);return r==n.from?n:F.cursor(r,r<n.from?1:-1)}const Ar="";class Rr{constructor(e,t){this.points=e,this.text="",this.lineSeparator=t.facet(Se.lineSeparator)}append(e){this.text+=e}lineBreak(){this.text+=Ar}readRange(e,t){if(!e)return this;let n=e.parentNode;for(let r=e;;){this.findPointBefore(n,r);let e=this.text.length;this.readNode(r);let a=r.nextSibling;if(a==t)break;let o=_t.get(r),i=_t.get(a);(o&&i?o.breakAfter:(o?o.breakAfter:lt(r))||lt(a)&&("BR"!=r.nodeName||r.cmIgnore)&&this.text.length>e)&&this.lineBreak(),r=a}return this.findPointBefore(n,t),this}readTextNode(e){let t=e.nodeValue;for(let n of this.points)n.node==e&&(n.pos=this.text.length+Math.min(n.offset,t.length));for(let n=0,r=this.lineSeparator?null:/\r\n?|\n/g;;){let a,o=-1,i=1;if(this.lineSeparator?(o=t.indexOf(this.lineSeparator,n),i=this.lineSeparator.length):(a=r.exec(t))&&(o=a.index,i=a[0].length),this.append(t.slice(n,o<0?t.length:o)),o<0)break;if(this.lineBreak(),i>1)for(let t of this.points)t.node==e&&t.pos>this.text.length&&(t.pos-=i-1);n=o+i}}readNode(e){if(e.cmIgnore)return;let t=_t.get(e),n=t&&t.overrideDOMText;if(null!=n){this.findPointInside(e,n.length);for(let e=n.iter();!e.next().done;)e.lineBreak?this.lineBreak():this.append(e.value)}else 3==e.nodeType?this.readTextNode(e):"BR"==e.nodeName?e.nextSibling&&this.lineBreak():1==e.nodeType&&this.readRange(e.firstChild,null)}findPointBefore(e,t){for(let n of this.points)n.node==e&&e.childNodes[n.offset]==t&&(n.pos=this.text.length)}findPointInside(e,t){for(let n of this.points)(3==e.nodeType?n.node==e:e.contains(n.node))&&(n.pos=this.text.length+(Pr(e,n.node,n.offset)?t:0))}}function Pr(e,t,n){for(;;){if(!t||n<ut(t))return!1;if(t==e)return!0;n=st(t)+1,t=t.parentNode}}class Dr{constructor(e,t){this.node=e,this.offset=t,this.pos=-1}}class Mr{constructor(e,t,n,r){this.typeOver=r,this.bounds=null,this.text="",this.domChanged=t>-1;let{impreciseHead:a,impreciseAnchor:o}=e.docView;if(e.state.readOnly&&t>-1)this.newSel=null;else if(t>-1&&(this.bounds=e.docView.domBoundsAround(t,n,0))){let t=a||o?[]:function(e){let t=[];if(e.root.activeElement!=e.contentDOM)return t;let{anchorNode:n,anchorOffset:r,focusNode:a,focusOffset:o}=e.observer.selectionRange;return n&&(t.push(new Dr(n,r)),a==n&&o==r||t.push(new Dr(a,o))),t}(e),n=new Rr(t,e.state);n.readRange(this.bounds.startDOM,this.bounds.endDOM),this.text=n.text,this.newSel=function(e,t){if(0==e.length)return null;let n=e[0].pos,r=2==e.length?e[1].pos:n;return n>-1&&r>-1?F.single(n+t,r+t):null}(t,this.bounds.from)}else{let t=e.observer.selectionRange,n=a&&a.node==t.focusNode&&a.offset==t.focusOffset||!rt(e.contentDOM,t.focusNode)?e.state.selection.main.head:e.docView.posFromDOM(t.focusNode,t.focusOffset),r=o&&o.node==t.anchorNode&&o.offset==t.anchorOffset||!rt(e.contentDOM,t.anchorNode)?e.state.selection.main.anchor:e.docView.posFromDOM(t.anchorNode,t.anchorOffset),i=e.viewport;if((zt.ios||zt.chrome)&&e.state.selection.main.empty&&n!=r&&(i.from>0||i.to<e.state.doc.length)){let t=Math.min(n,r),a=Math.max(n,r),o=i.from-t,s=i.to-a;0!=o&&1!=o&&0!=t||0!=s&&-1!=s&&a!=e.state.doc.length||(n=0,r=e.state.doc.length)}this.newSel=F.single(r,n)}}}function $r(e,t){let n,{newSel:r}=t,a=e.state.selection.main,o=e.inputState.lastKeyTime>Date.now()-100?e.inputState.lastKeyCode:-1;if(t.bounds){let{from:r,to:i}=t.bounds,s=a.from,l=null;(8===o||zt.android&&t.text.length<i-r)&&(s=a.to,l="end");let c=function(e,t,n,r){let a=Math.min(e.length,t.length),o=0;for(;o<a&&e.charCodeAt(o)==t.charCodeAt(o);)o++;if(o==a&&e.length==t.length)return null;let i=e.length,s=t.length;for(;i>0&&s>0&&e.charCodeAt(i-1)==t.charCodeAt(s-1);)i--,s--;return"end"==r&&(n-=i+Math.max(0,o-Math.min(i,s))-o),i<o&&e.length<t.length?(o-=n<=o&&n>=i?o-n:0,s=o+(s-i),i=o):s<o&&(o-=n<=o&&n>=s?o-n:0,i=o+(i-s),s=o),{from:o,toA:i,toB:s}}(e.state.doc.sliceString(r,i,Ar),t.text,s-r,l);c&&(zt.chrome&&13==o&&c.toB==c.from+2&&t.text.slice(c.from,c.toB)==Ar+Ar&&c.toB--,n={from:r+c.from,to:r+c.toA,insert:f.of(t.text.slice(c.from,c.toB).split(Ar))})}else r&&(!e.hasFocus&&e.state.facet(Kn)||r.main.eq(a))&&(r=null);if(!n&&!r)return!1;if(!n&&t.typeOver&&!a.empty&&r&&r.main.empty?n={from:a.from,to:a.to,insert:e.state.doc.slice(a.from,a.to)}:(zt.mac||zt.android)&&n&&n.from==n.to&&n.from==a.head-1&&/^\. ?$/.test(n.insert.toString())&&"off"==e.contentDOM.getAttribute("autocorrect")?(r&&2==n.insert.length&&(r=F.single(r.main.anchor-1,r.main.head-1)),n={from:n.from,to:n.to,insert:f.of([n.insert.toString().replace("."," ")])}):n&&n.from>=a.from&&n.to<=a.to&&(n.from!=a.from||n.to!=a.to)&&a.to-a.from-(n.to-n.from)<=4?n={from:a.from,to:a.to,insert:e.state.doc.slice(a.from,n.from).append(n.insert).append(e.state.doc.slice(n.to,a.to))}:zt.chrome&&n&&n.from==n.to&&n.from==a.head&&"\n "==n.insert.toString()&&e.lineWrapping&&(r&&(r=F.single(r.main.anchor-1,r.main.head-1)),n={from:a.from,to:a.to,insert:f.of([" "])}),n)return Nr(e,n,r,o);if(r&&!r.main.eq(a)){let t=!1,n="select";return e.inputState.lastSelectionTime>Date.now()-50&&("select"==e.inputState.lastSelectionOrigin&&(t=!0),n=e.inputState.lastSelectionOrigin),e.dispatch({selection:r,scrollIntoView:t,userEvent:n}),!0}return!1}function Nr(e,t,n,r=-1){if(zt.ios&&e.inputState.flushIOSKey(t))return!0;let a=e.state.selection.main;if(zt.android&&(t.to==a.to&&(t.from==a.from||t.from==a.from-1&&" "==e.state.sliceDoc(t.from,a.from))&&1==t.insert.length&&2==t.insert.lines&&yt(e.contentDOM,"Enter",13)||(t.from==a.from-1&&t.to==a.to&&0==t.insert.length||8==r&&t.insert.length<t.to-t.from&&t.to>a.head)&&yt(e.contentDOM,"Backspace",8)||t.from==a.from&&t.to==a.to+1&&0==t.insert.length&&yt(e.contentDOM,"Delete",46)))return!0;let o,i=t.insert.toString();e.inputState.composing>=0&&e.inputState.composing++;let s=()=>o||(o=function(e,t,n){let r,a=e.state,o=a.selection.main;if(t.from>=o.from&&t.to<=o.to&&t.to-t.from>=(o.to-o.from)/3&&(!n||n.main.empty&&n.main.from==t.from+t.insert.length)&&e.inputState.composing<0){let n=o.from<t.from?a.sliceDoc(o.from,t.from):"",i=o.to>t.to?a.sliceDoc(t.to,o.to):"";r=a.replaceSelection(e.state.toText(n+t.insert.sliceString(0,void 0,e.state.lineBreak)+i))}else{let i=a.changes(t),s=n&&n.main.to<=i.newLength?n.main:void 0;if(a.selection.ranges.length>1&&e.inputState.composing>=0&&t.to<=o.to&&t.to>=o.to-10){let l,c=e.state.sliceDoc(t.from,t.to),u=n&&mr(e,n.main.head);if(u){let e=t.insert.length-(t.to-t.from);l={from:u.from,to:u.to-e}}else l=e.state.doc.lineAt(o.head);let d=o.to-t.to,p=o.to-o.from;r=a.changeByRange((n=>{if(n.from==o.from&&n.to==o.to)return{changes:i,range:s||n.map(i)};let r=n.to-d,u=r-c.length;if(n.to-n.from!=p||e.state.sliceDoc(u,r)!=c||n.to>=l.from&&n.from<=l.to)return{range:n};let h=a.changes({from:u,to:r,insert:t.insert}),f=n.to-o.to;return{changes:h,range:s?F.range(Math.max(0,s.anchor+f),Math.max(0,s.head+f)):n.map(h)}}))}else r={changes:i,selection:s&&a.selection.replaceRange(s)}}let i="input.type";return(e.composing||e.inputState.compositionPendingChange&&e.inputState.compositionEndedAt>Date.now()-50)&&(e.inputState.compositionPendingChange=!1,i+=".compose",e.inputState.compositionFirstChange&&(i+=".start",e.inputState.compositionFirstChange=!1)),a.update(r,{userEvent:i,scrollIntoView:!0})}(e,t,n));return e.state.facet(Qn).some((n=>n(e,t.from,t.to,i,s)))||e.dispatch(s()),!0}class Ir{setSelectionOrigin(e){this.lastSelectionOrigin=e,this.lastSelectionTime=Date.now()}constructor(e){this.view=e,this.lastKeyCode=0,this.lastKeyTime=0,this.lastTouchTime=0,this.lastFocusTime=0,this.lastScrollTop=0,this.lastScrollLeft=0,this.pendingIOSKey=void 0,this.tabFocusMode=-1,this.lastSelectionOrigin=null,this.lastSelectionTime=0,this.lastContextMenu=0,this.scrollHandlers=[],this.handlers=Object.create(null),this.composing=-1,this.compositionFirstChange=null,this.compositionEndedAt=0,this.compositionPendingKey=!1,this.compositionPendingChange=!1,this.mouseSelection=null,this.draggedContent=null,this.handleEvent=this.handleEvent.bind(this),this.notifiedFocused=e.hasFocus,zt.safari&&e.contentDOM.addEventListener("input",(()=>null)),zt.gecko&&function(e){ca.has(e)||(ca.add(e),e.addEventListener("copy",(()=>{})),e.addEventListener("cut",(()=>{})))}(e.contentDOM.ownerDocument)}handleEvent(e){(function(e,t){if(!t.bubbles)return!0;if(t.defaultPrevented)return!1;for(let n,r=t.target;r!=e.contentDOM;r=r.parentNode)if(!r||11==r.nodeType||(n=_t.get(r))&&n.ignoreEvent(t))return!1;return!0})(this.view,e)&&!this.ignoreDuringComposition(e)&&("keydown"==e.type&&this.keydown(e)||(0!=this.view.updateState?Promise.resolve().then((()=>this.runHandlers(e.type,e))):this.runHandlers(e.type,e)))}runHandlers(e,t){let n=this.handlers[e];if(n){for(let e of n.observers)e(this.view,t);for(let e of n.handlers){if(t.defaultPrevented)break;if(e(this.view,t)){t.preventDefault();break}}}}ensureHandlers(e){let t=Br(e),n=this.handlers,r=this.view.contentDOM;for(let e in t)if("scroll"!=e){let a=!t[e].handlers.length,o=n[e];o&&a!=!o.handlers.length&&(r.removeEventListener(e,this.handleEvent),o=null),o||r.addEventListener(e,this.handleEvent,{passive:a})}for(let e in n)"scroll"==e||t[e]||r.removeEventListener(e,this.handleEvent);this.handlers=t}keydown(e){if(this.lastKeyCode=e.keyCode,this.lastKeyTime=Date.now(),9==e.keyCode&&this.tabFocusMode>-1&&(!this.tabFocusMode||Date.now()<=this.tabFocusMode))return!0;if(this.tabFocusMode>0&&27!=e.keyCode&&Ur.indexOf(e.keyCode)<0&&(this.tabFocusMode=-1),zt.android&&zt.chrome&&!e.synthetic&&(13==e.keyCode||8==e.keyCode))return this.view.observer.delayAndroidKey(e.key,e.keyCode),!0;let t;return!zt.ios||e.synthetic||e.altKey||e.metaKey||!((t=Fr.find((t=>t.keyCode==e.keyCode)))&&!e.ctrlKey||Qr.indexOf(e.key)>-1&&e.ctrlKey&&!e.shiftKey)?(229!=e.keyCode&&this.view.observer.forceFlush(),!1):(this.pendingIOSKey=t||e,setTimeout((()=>this.flushIOSKey()),250),!0)}flushIOSKey(e){let t=this.pendingIOSKey;return!!t&&!("Enter"==t.key&&e&&e.from<e.to&&/^\S+$/.test(e.insert.toString()))&&(this.pendingIOSKey=void 0,yt(this.view.contentDOM,t.key,t.keyCode,t instanceof KeyboardEvent?t:void 0))}ignoreDuringComposition(e){return!!/^key/.test(e.type)&&(this.composing>0||!!(zt.safari&&!zt.ios&&this.compositionPendingKey&&Date.now()-this.compositionEndedAt<100)&&(this.compositionPendingKey=!1,!0))}startMouseSelection(e){this.mouseSelection&&this.mouseSelection.destroy(),this.mouseSelection=e}update(e){this.view.observer.update(e),this.mouseSelection&&this.mouseSelection.update(e),this.draggedContent&&e.docChanged&&(this.draggedContent=this.draggedContent.map(e.changes)),e.transactions.length&&(this.lastKeyCode=this.lastSelectionTime=0)}destroy(){this.mouseSelection&&this.mouseSelection.destroy()}}function Lr(e,t){return(n,r)=>{try{return t.call(e,r,n)}catch(e){Gn(n.state,e)}}}function Br(e){let t=Object.create(null);function n(e){return t[e]||(t[e]={observers:[],handlers:[]})}for(let t of e){let e=t.spec;if(e&&e.domEventHandlers)for(let r in e.domEventHandlers){let a=e.domEventHandlers[r];a&&n(r).handlers.push(Lr(t.value,a))}if(e&&e.domEventObservers)for(let r in e.domEventObservers){let a=e.domEventObservers[r];a&&n(r).observers.push(Lr(t.value,a))}}for(let e in Wr)n(e).handlers.push(Wr[e]);for(let e in Xr)n(e).observers.push(Xr[e]);return t}const Fr=[{key:"Backspace",keyCode:8,inputType:"deleteContentBackward"},{key:"Enter",keyCode:13,inputType:"insertParagraph"},{key:"Enter",keyCode:13,inputType:"insertLineBreak"},{key:"Delete",keyCode:46,inputType:"deleteContentForward"}],Qr="dthko",Ur=[16,17,18,20,91,92,224,225];function zr(e){return.7*Math.max(0,e)+8}class jr{constructor(e,t,n,r){this.view=e,this.startEvent=t,this.style=n,this.mustSelect=r,this.scrollSpeed={x:0,y:0},this.scrolling=-1,this.lastEvent=t,this.scrollParents=function(e){let t,n,r=e.ownerDocument;for(let a=e.parentNode;a&&!(a==r.body||t&&n);)if(1==a.nodeType)!n&&a.scrollHeight>a.clientHeight&&(n=a),!t&&a.scrollWidth>a.clientWidth&&(t=a),a=a.assignedSlot||a.parentNode;else{if(11!=a.nodeType)break;a=a.host}return{x:t,y:n}}(e.contentDOM),this.atoms=e.state.facet(ir).map((t=>t(e)));let a=e.contentDOM.ownerDocument;a.addEventListener("mousemove",this.move=this.move.bind(this)),a.addEventListener("mouseup",this.up=this.up.bind(this)),this.extend=t.shiftKey,this.multiple=e.state.facet(Se.allowMultipleSelections)&&function(e,t){let n=e.state.facet(Nn);return n.length?n[0](t):zt.mac?t.metaKey:t.ctrlKey}(e,t),this.dragging=!(!function(e,t){let{main:n}=e.state.selection;if(n.empty)return!1;let r=nt(e.root);if(!r||0==r.rangeCount)return!0;let a=r.getRangeAt(0).getClientRects();for(let e=0;e<a.length;e++){let n=a[e];if(n.left<=t.clientX&&n.right>=t.clientX&&n.top<=t.clientY&&n.bottom>=t.clientY)return!0}return!1}(e,t)||1!=ra(t))&&null}start(e){!1===this.dragging&&this.select(e)}move(e){if(0==e.buttons)return this.destroy();if(this.dragging||null==this.dragging&&(t=this.startEvent,n=e,Math.max(Math.abs(t.clientX-n.clientX),Math.abs(t.clientY-n.clientY))<10))return;var t,n;this.select(this.lastEvent=e);let r=0,a=0,o=0,i=0,s=this.view.win.innerWidth,l=this.view.win.innerHeight;this.scrollParents.x&&({left:o,right:s}=this.scrollParents.x.getBoundingClientRect()),this.scrollParents.y&&({top:i,bottom:l}=this.scrollParents.y.getBoundingClientRect());let c=ur(this.view);e.clientX-c.left<=o+6?r=-zr(o-e.clientX):e.clientX+c.right>=s-6&&(r=zr(e.clientX-s)),e.clientY-c.top<=i+6?a=-zr(i-e.clientY):e.clientY+c.bottom>=l-6&&(a=zr(e.clientY-l)),this.setScrollSpeed(r,a)}up(e){null==this.dragging&&this.select(this.lastEvent),this.dragging||e.preventDefault(),this.destroy()}destroy(){this.setScrollSpeed(0,0);let e=this.view.contentDOM.ownerDocument;e.removeEventListener("mousemove",this.move),e.removeEventListener("mouseup",this.up),this.view.inputState.mouseSelection=this.view.inputState.draggedContent=null}setScrollSpeed(e,t){this.scrollSpeed={x:e,y:t},e||t?this.scrolling<0&&(this.scrolling=setInterval((()=>this.scroll()),50)):this.scrolling>-1&&(clearInterval(this.scrolling),this.scrolling=-1)}scroll(){let{x:e,y:t}=this.scrollSpeed;e&&this.scrollParents.x&&(this.scrollParents.x.scrollLeft+=e,e=0),t&&this.scrollParents.y&&(this.scrollParents.y.scrollTop+=t,t=0),(e||t)&&this.view.win.scrollBy(e,t),!1===this.dragging&&this.select(this.lastEvent)}skipAtoms(e){let t=null;for(let n=0;n<e.ranges.length;n++){let r=e.ranges[n],a=null;if(r.empty){let e=Cr(this.atoms,r.from,0);e!=r.from&&(a=F.cursor(e,-1))}else{let e=Cr(this.atoms,r.from,-1),t=Cr(this.atoms,r.to,1);e==r.from&&t==r.to||(a=F.range(r.from==r.anchor?e:t,r.from==r.head?e:t))}a&&(t||(t=e.ranges.slice()),t[n]=a)}return t?F.create(t,e.mainIndex):e}select(e){let{view:t}=this,n=this.skipAtoms(this.style.get(e,this.extend,this.multiple));!this.mustSelect&&n.eq(t.state.selection,!1===this.dragging)||this.view.dispatch({selection:n,userEvent:"select.pointer"}),this.mustSelect=!1}update(e){e.transactions.some((e=>e.isUserEvent("input.type")))?this.destroy():this.style.update(e)&&setTimeout((()=>this.select(this.lastEvent)),20)}}const Wr=Object.create(null),Xr=Object.create(null),qr=zt.ie&&zt.ie_version<15||zt.ios&&zt.webkit_version<604;function Vr(e,t,n){for(let r of e.facet(t))n=r(n,e);return n}function Hr(e,t){t=Vr(e.state,zn,t);let n,{state:r}=e,a=1,o=r.toText(t),i=o.lines==r.selection.ranges.length;if(null!=oa&&r.selection.ranges.every((e=>e.empty))&&oa==o.toString()){let e=-1;n=r.changeByRange((n=>{let s=r.doc.lineAt(n.from);if(s.from==e)return{range:n};e=s.from;let l=r.toText((i?o.line(a++).text:t)+r.lineBreak);return{changes:{from:s.from,insert:l},range:F.cursor(n.from+l.length)}}))}else n=i?r.changeByRange((e=>{let t=o.line(a++);return{changes:{from:e.from,to:e.to,insert:t.text},range:F.cursor(e.from+t.length)}})):r.replaceSelection(o);e.dispatch(n,{userEvent:"input.paste",scrollIntoView:!0})}function Zr(e,t,n,r){if(1==r)return F.cursor(t,n);if(2==r)return function(e,t,n=1){let r=e.charCategorizer(t),a=e.doc.lineAt(t),o=t-a.from;if(0==a.length)return F.cursor(t);0==o?n=1:o==a.length&&(n=-1);let i=o,s=o;n<0?i=E(a.text,o,!1):s=E(a.text,o);let l=r(a.text.slice(i,s));for(;i>0;){let e=E(a.text,i,!1);if(r(a.text.slice(e,i))!=l)break;i=e}for(;s<a.length;){let e=E(a.text,s);if(r(a.text.slice(s,e))!=l)break;s=e}return F.range(i+a.from,s+a.from)}(e.state,t,n);{let n=un.find(e.docView,t),r=e.state.doc.lineAt(n?n.posAtEnd:t),a=n?n.posAtStart:r.from,o=n?n.posAtEnd:r.to;return o<e.state.doc.length&&o==r.to&&o++,F.range(a,o)}}Xr.scroll=e=>{e.inputState.lastScrollTop=e.scrollDOM.scrollTop,e.inputState.lastScrollLeft=e.scrollDOM.scrollLeft},Wr.keydown=(e,t)=>(e.inputState.setSelectionOrigin("select"),27==t.keyCode&&0!=e.inputState.tabFocusMode&&(e.inputState.tabFocusMode=Date.now()+2e3),!1),Xr.touchstart=(e,t)=>{e.inputState.lastTouchTime=Date.now(),e.inputState.setSelectionOrigin("select.pointer")},Xr.touchmove=e=>{e.inputState.setSelectionOrigin("select.pointer")},Wr.mousedown=(e,t)=>{if(e.observer.flush(),e.inputState.lastTouchTime>Date.now()-2e3)return!1;let n=null;for(let r of e.state.facet(Ln))if(n=r(e,t),n)break;if(n||0!=t.button||(n=function(e,t){let n=Yr(e,t),r=ra(t),a=e.state.selection;return{update(e){e.docChanged&&(n.pos=e.changes.mapPos(n.pos),a=a.map(e.changes))},get(t,o,i){let s,l=Yr(e,t),c=Zr(e,l.pos,l.bias,r);if(n.pos!=l.pos&&!o){let t=Zr(e,n.pos,n.bias,r),a=Math.min(t.from,c.from),o=Math.max(t.to,c.to);c=a<c.from?F.range(a,o):F.range(o,a)}return o?a.replaceRange(a.main.extend(c.from,c.to)):i&&1==r&&a.ranges.length>1&&(s=function(e,t){for(let n=0;n<e.ranges.length;n++){let{from:r,to:a}=e.ranges[n];if(r<=t&&a>=t)return F.create(e.ranges.slice(0,n).concat(e.ranges.slice(n+1)),e.mainIndex==n?0:e.mainIndex-(e.mainIndex>n?1:0))}return null}(a,l.pos))?s:i?a.addRange(c):F.create([c])}}}(e,t)),n){let r=!e.hasFocus;e.inputState.startMouseSelection(new jr(e,t,n,r)),r&&e.observer.ignore((()=>{vt(e.contentDOM);let t=e.root.activeElement;t&&!t.contains(e.contentDOM)&&t.blur()}));let a=e.inputState.mouseSelection;if(a)return a.start(t),!1===a.dragging}return!1};let Gr=(e,t,n)=>t>=n.top&&t<=n.bottom&&e>=n.left&&e<=n.right;function Kr(e,t,n,r){let a=un.find(e.docView,t);if(!a)return 1;let o=t-a.posAtStart;if(0==o)return 1;if(o==a.length)return-1;let i=a.coordsAt(o,-1);if(i&&Gr(n,r,i))return-1;let s=a.coordsAt(o,1);return s&&Gr(n,r,s)?1:i&&i.bottom>=r?-1:1}function Yr(e,t){let n=e.posAtCoords({x:t.clientX,y:t.clientY},!1);return{pos:n,bias:Kr(e,n,t.clientX,t.clientY)}}const Jr=zt.ie&&zt.ie_version<=11;let ea=null,ta=0,na=0;function ra(e){if(!Jr)return e.detail;let t=ea,n=na;return ea=e,na=Date.now(),ta=!t||n>Date.now()-400&&Math.abs(t.clientX-e.clientX)<2&&Math.abs(t.clientY-e.clientY)<2?(ta+1)%3:1}function aa(e,t,n,r){if(!(n=Vr(e.state,zn,n)))return;let a=e.posAtCoords({x:t.clientX,y:t.clientY},!1),{draggedContent:o}=e.inputState,i=r&&o&&function(e,t){let n=e.state.facet(In);return n.length?n[0](t):zt.mac?!t.altKey:!t.ctrlKey}(e,t)?{from:o.from,to:o.to}:null,s={from:a,insert:n},l=e.state.changes(i?[i,s]:s);e.focus(),e.dispatch({changes:l,selection:{anchor:l.mapPos(a,-1),head:l.mapPos(a,1)},userEvent:i?"move.drop":"input.drop"}),e.inputState.draggedContent=null}Wr.dragstart=(e,t)=>{let{selection:{main:n}}=e.state;if(t.target.draggable){let r=e.docView.nearest(t.target);if(r&&r.isWidget){let e=r.posAtStart,t=e+r.length;(e>=n.to||t<=n.from)&&(n=F.range(e,t))}}let{inputState:r}=e;return r.mouseSelection&&(r.mouseSelection.dragging=!0),r.draggedContent=n,t.dataTransfer&&(t.dataTransfer.setData("Text",Vr(e.state,jn,e.state.sliceDoc(n.from,n.to))),t.dataTransfer.effectAllowed="copyMove"),!1},Wr.dragend=e=>(e.inputState.draggedContent=null,!1),Wr.drop=(e,t)=>{if(!t.dataTransfer)return!1;if(e.state.readOnly)return!0;let n=t.dataTransfer.files;if(n&&n.length){let r=Array(n.length),a=0,o=()=>{++a==n.length&&aa(e,t,r.filter((e=>null!=e)).join(e.state.lineBreak),!1)};for(let e=0;e<n.length;e++){let t=new FileReader;t.onerror=o,t.onload=()=>{/[\x00-\x08\x0e-\x1f]{2}/.test(t.result)||(r[e]=t.result),o()},t.readAsText(n[e])}return!0}{let n=t.dataTransfer.getData("Text");if(n)return aa(e,t,n,!0),!0}return!1},Wr.paste=(e,t)=>{if(e.state.readOnly)return!0;e.observer.flush();let n=qr?null:t.clipboardData;return n?(Hr(e,n.getData("text/plain")||n.getData("text/uri-list")),!0):(function(e){let t=e.dom.parentNode;if(!t)return;let n=t.appendChild(document.createElement("textarea"));n.style.cssText="position: fixed; left: -10000px; top: 10px",n.focus(),setTimeout((()=>{e.focus(),n.remove(),Hr(e,n.value)}),50)}(e),!1)};let oa=null;Wr.copy=Wr.cut=(e,t)=>{let{text:n,ranges:r,linewise:a}=function(e){let t=[],n=[],r=!1;for(let r of e.selection.ranges)r.empty||(t.push(e.sliceDoc(r.from,r.to)),n.push(r));if(!t.length){let a=-1;for(let{from:r}of e.selection.ranges){let o=e.doc.lineAt(r);o.number>a&&(t.push(o.text),n.push({from:o.from,to:Math.min(e.doc.length,o.to+1)})),a=o.number}r=!0}return{text:Vr(e,jn,t.join(e.lineBreak)),ranges:n,linewise:r}}(e.state);if(!n&&!a)return!1;oa=a?n:null,"cut"!=t.type||e.state.readOnly||e.dispatch({changes:r,scrollIntoView:!0,userEvent:"delete.cut"});let o=qr?null:t.clipboardData;return o?(o.clearData(),o.setData("text/plain",n),!0):(function(e,t){let n=e.dom.parentNode;if(!n)return;let r=n.appendChild(document.createElement("textarea"));r.style.cssText="position: fixed; left: -10000px; top: 10px",r.value=t,r.focus(),r.selectionEnd=t.length,r.selectionStart=0,setTimeout((()=>{r.remove(),e.focus()}),50)}(e,n),!1)};const ia=de.define();function sa(e,t){let n=[];for(let r of e.facet(Un)){let a=r(e,t);a&&n.push(a)}return n?e.update({effects:n,annotations:ia.of(!0)}):null}function la(e){setTimeout((()=>{let t=e.hasFocus;if(t!=e.inputState.notifiedFocused){let n=sa(e.state,t);n?e.dispatch(n):e.update([])}}),10)}Xr.focus=e=>{e.inputState.lastFocusTime=Date.now(),e.scrollDOM.scrollTop||!e.inputState.lastScrollTop&&!e.inputState.lastScrollLeft||(e.scrollDOM.scrollTop=e.inputState.lastScrollTop,e.scrollDOM.scrollLeft=e.inputState.lastScrollLeft),la(e)},Xr.blur=e=>{e.observer.clearSelectionRange(),la(e)},Xr.compositionstart=Xr.compositionupdate=e=>{e.observer.editContext||(null==e.inputState.compositionFirstChange&&(e.inputState.compositionFirstChange=!0),e.inputState.composing<0&&(e.inputState.composing=0))},Xr.compositionend=e=>{e.observer.editContext||(e.inputState.composing=-1,e.inputState.compositionEndedAt=Date.now(),e.inputState.compositionPendingKey=!0,e.inputState.compositionPendingChange=e.observer.pendingRecords().length>0,e.inputState.compositionFirstChange=null,zt.chrome&&zt.android?e.observer.flushSoon():e.inputState.compositionPendingChange?Promise.resolve().then((()=>e.observer.flush())):setTimeout((()=>{e.inputState.composing<0&&e.docView.hasComposition&&e.update([])}),50))},Xr.contextmenu=e=>{e.inputState.lastContextMenu=Date.now()},Wr.beforeinput=(e,t)=>{var n,r;if("insertReplacementText"==t.inputType&&e.observer.editContext){let r=null===(n=t.dataTransfer)||void 0===n?void 0:n.getData("text/plain"),a=t.getTargetRanges();if(r&&a.length){let t=a[0],n=e.posAtDOM(t.startContainer,t.startOffset),o=e.posAtDOM(t.endContainer,t.endOffset);return Nr(e,{from:n,to:o,insert:e.state.toText(r)},null),!0}}let a;if(zt.chrome&&zt.android&&(a=Fr.find((e=>e.inputType==t.inputType)))&&(e.observer.delayAndroidKey(a.key,a.keyCode),"Backspace"==a.key||"Delete"==a.key)){let t=(null===(r=window.visualViewport)||void 0===r?void 0:r.height)||0;setTimeout((()=>{var n;((null===(n=window.visualViewport)||void 0===n?void 0:n.height)||0)>t+10&&e.hasFocus&&(e.contentDOM.blur(),e.focus())}),100)}return zt.ios&&"deleteContentForward"==t.inputType&&e.observer.flushSoon(),zt.safari&&"insertText"==t.inputType&&e.inputState.composing>=0&&setTimeout((()=>Xr.compositionend(e,t)),20),!1};const ca=new Set,ua=["pre-wrap","normal","pre-line","break-spaces"];let da=!1;function pa(){da=!1}class ha{constructor(e){this.lineWrapping=e,this.doc=f.empty,this.heightSamples={},this.lineHeight=14,this.charWidth=7,this.textHeight=14,this.lineLength=30}heightForGap(e,t){let n=this.doc.lineAt(t).number-this.doc.lineAt(e).number+1;return this.lineWrapping&&(n+=Math.max(0,Math.ceil((t-e-n*this.lineLength*.5)/this.lineLength))),this.lineHeight*n}heightForLine(e){return this.lineWrapping?(1+Math.max(0,Math.ceil((e-this.lineLength)/(this.lineLength-5))))*this.lineHeight:this.lineHeight}setDoc(e){return this.doc=e,this}mustRefreshForWrapping(e){return ua.indexOf(e)>-1!=this.lineWrapping}mustRefreshForHeights(e){let t=!1;for(let n=0;n<e.length;n++){let r=e[n];r<0?n++:this.heightSamples[Math.floor(10*r)]||(t=!0,this.heightSamples[Math.floor(10*r)]=!0)}return t}refresh(e,t,n,r,a,o){let i=ua.indexOf(e)>-1,s=Math.round(t)!=Math.round(this.lineHeight)||this.lineWrapping!=i;if(this.lineWrapping=i,this.lineHeight=t,this.charWidth=n,this.textHeight=r,this.lineLength=a,s){this.heightSamples={};for(let e=0;e<o.length;e++){let t=o[e];t<0?e++:this.heightSamples[Math.floor(10*t)]=!0}}return s}}class fa{constructor(e,t){this.from=e,this.heights=t,this.index=0}get more(){return this.index<this.heights.length}}class ma{constructor(e,t,n,r,a){this.from=e,this.length=t,this.top=n,this.height=r,this._content=a}get type(){return"number"==typeof this._content?nn.Text:Array.isArray(this._content)?this._content:this._content.type}get to(){return this.from+this.length}get bottom(){return this.top+this.height}get widget(){return this._content instanceof sn?this._content.widget:null}get widgetLineBreaks(){return"number"==typeof this._content?this._content:0}join(e){let t=(Array.isArray(this._content)?this._content:[this]).concat(Array.isArray(e._content)?e._content:[e]);return new ma(this.from,this.length+e.length,this.top,this.height+e.height,t)}}var ga=function(e){return e[e.ByPos=0]="ByPos",e[e.ByHeight=1]="ByHeight",e[e.ByPosNoHeight=2]="ByPosNoHeight",e}(ga||(ga={}));const va=.001;class ba{constructor(e,t,n=2){this.length=e,this.height=t,this.flags=n}get outdated(){return(2&this.flags)>0}set outdated(e){this.flags=(e?2:0)|-3&this.flags}setHeight(e){this.height!=e&&(Math.abs(this.height-e)>va&&(da=!0),this.height=e)}replace(e,t,n){return ba.of(n)}decomposeLeft(e,t){t.push(this)}decomposeRight(e,t){t.push(this)}applyChanges(e,t,n,r){let a=this,o=n.doc;for(let i=r.length-1;i>=0;i--){let{fromA:s,toA:l,fromB:c,toB:u}=r[i],d=a.lineAt(s,ga.ByPosNoHeight,n.setDoc(t),0,0),p=d.to>=l?d:a.lineAt(l,ga.ByPosNoHeight,n,0,0);for(u+=p.to-l,l=p.to;i>0&&d.from<=r[i-1].toA;)s=r[i-1].fromA,c=r[i-1].fromB,i--,s<d.from&&(d=a.lineAt(s,ga.ByPosNoHeight,n,0,0));c+=d.from-s,s=d.from;let h=Sa.build(n.setDoc(o),e,c,u);a=ya(a,a.replace(s,l,h))}return a.updateHeight(n,0)}static empty(){return new Oa(0,0)}static of(e){if(1==e.length)return e[0];let t=0,n=e.length,r=0,a=0;for(;;)if(t==n)if(r>2*a){let a=e[t-1];a.break?e.splice(--t,1,a.left,null,a.right):e.splice(--t,1,a.left,a.right),n+=1+a.break,r-=a.size}else{if(!(a>2*r))break;{let t=e[n];t.break?e.splice(n,1,t.left,null,t.right):e.splice(n,1,t.left,t.right),n+=2+t.break,a-=t.size}}else if(r<a){let n=e[t++];n&&(r+=n.size)}else{let t=e[--n];t&&(a+=t.size)}let o=0;return null==e[t-1]?(o=1,t--):null==e[t]&&(o=1,n++),new xa(ba.of(e.slice(0,t)),o,ba.of(e.slice(n)))}}function ya(e,t){return e==t?e:(e.constructor!=t.constructor&&(da=!0),t)}ba.prototype.size=1;class wa extends ba{constructor(e,t,n){super(e,t),this.deco=n}blockAt(e,t,n,r){return new ma(r,this.length,n,this.height,this.deco||0)}lineAt(e,t,n,r,a){return this.blockAt(0,n,r,a)}forEachLine(e,t,n,r,a,o){e<=a+this.length&&t>=a&&o(this.blockAt(0,n,r,a))}updateHeight(e,t=0,n=!1,r){return r&&r.from<=t&&r.more&&this.setHeight(r.heights[r.index++]),this.outdated=!1,this}toString(){return`block(${this.length})`}}class Oa extends wa{constructor(e,t){super(e,t,null),this.collapsed=0,this.widgetHeight=0,this.breaks=0}blockAt(e,t,n,r){return new ma(r,this.length,n,this.height,this.breaks)}replace(e,t,n){let r=n[0];return 1==n.length&&(r instanceof Oa||r instanceof ka&&4&r.flags)&&Math.abs(this.length-r.length)<10?(r instanceof ka?r=new Oa(r.length,this.height):r.height=this.height,this.outdated||(r.outdated=!1),r):ba.of(n)}updateHeight(e,t=0,n=!1,r){return r&&r.from<=t&&r.more?this.setHeight(r.heights[r.index++]):(n||this.outdated)&&this.setHeight(Math.max(this.widgetHeight,e.heightForLine(this.length-this.collapsed))+this.breaks*e.lineHeight),this.outdated=!1,this}toString(){return`line(${this.length}${this.collapsed?-this.collapsed:""}${this.widgetHeight?":"+this.widgetHeight:""})`}}class ka extends ba{constructor(e){super(e,0)}heightMetrics(e,t){let n,r=e.doc.lineAt(t).number,a=e.doc.lineAt(t+this.length).number,o=a-r+1,i=0;if(e.lineWrapping){let t=Math.min(this.height,e.lineHeight*o);n=t/o,this.length>o+1&&(i=(this.height-t)/(this.length-o-1))}else n=this.height/o;return{firstLine:r,lastLine:a,perLine:n,perChar:i}}blockAt(e,t,n,r){let{firstLine:a,lastLine:o,perLine:i,perChar:s}=this.heightMetrics(t,r);if(t.lineWrapping){let a=r+(e<t.lineHeight?0:Math.round(Math.max(0,Math.min(1,(e-n)/this.height))*this.length)),o=t.doc.lineAt(a),l=i+o.length*s,c=Math.max(n,e-l/2);return new ma(o.from,o.length,c,l,0)}{let r=Math.max(0,Math.min(o-a,Math.floor((e-n)/i))),{from:s,length:l}=t.doc.line(a+r);return new ma(s,l,n+i*r,i,0)}}lineAt(e,t,n,r,a){if(t==ga.ByHeight)return this.blockAt(e,n,r,a);if(t==ga.ByPosNoHeight){let{from:t,to:r}=n.doc.lineAt(e);return new ma(t,r-t,0,0,0)}let{firstLine:o,perLine:i,perChar:s}=this.heightMetrics(n,a),l=n.doc.lineAt(e),c=i+l.length*s,u=l.number-o,d=r+i*u+s*(l.from-a-u);return new ma(l.from,l.length,Math.max(r,Math.min(d,r+this.height-c)),c,0)}forEachLine(e,t,n,r,a,o){e=Math.max(e,a),t=Math.min(t,a+this.length);let{firstLine:i,perLine:s,perChar:l}=this.heightMetrics(n,a);for(let c=e,u=r;c<=t;){let t=n.doc.lineAt(c);if(c==e){let n=t.number-i;u+=s*n+l*(e-a-n)}let r=s+l*t.length;o(new ma(t.from,t.length,u,r,0)),u+=r,c=t.to+1}}replace(e,t,n){let r=this.length-t;if(r>0){let e=n[n.length-1];e instanceof ka?n[n.length-1]=new ka(e.length+r):n.push(null,new ka(r-1))}if(e>0){let t=n[0];t instanceof ka?n[0]=new ka(e+t.length):n.unshift(new ka(e-1),null)}return ba.of(n)}decomposeLeft(e,t){t.push(new ka(e-1),null)}decomposeRight(e,t){t.push(null,new ka(this.length-e-1))}updateHeight(e,t=0,n=!1,r){let a=t+this.length;if(r&&r.from<=t+this.length&&r.more){let n=[],o=Math.max(t,r.from),i=-1;for(r.from>t&&n.push(new ka(r.from-t-1).updateHeight(e,t));o<=a&&r.more;){let t=e.doc.lineAt(o).length;n.length&&n.push(null);let a=r.heights[r.index++];-1==i?i=a:Math.abs(a-i)>=va&&(i=-2);let s=new Oa(t,a);s.outdated=!1,n.push(s),o+=t+1}o<=a&&n.push(null,new ka(a-o).updateHeight(e,o));let s=ba.of(n);return(i<0||Math.abs(s.height-this.height)>=va||Math.abs(i-this.heightMetrics(e,t).perLine)>=va)&&(da=!0),ya(this,s)}return(n||this.outdated)&&(this.setHeight(e.heightForGap(t,t+this.length)),this.outdated=!1),this}toString(){return`gap(${this.length})`}}class xa extends ba{constructor(e,t,n){super(e.length+t+n.length,e.height+n.height,t|(e.outdated||n.outdated?2:0)),this.left=e,this.right=n,this.size=e.size+n.size}get break(){return 1&this.flags}blockAt(e,t,n,r){let a=n+this.left.height;return e<a?this.left.blockAt(e,t,n,r):this.right.blockAt(e,t,a,r+this.left.length+this.break)}lineAt(e,t,n,r,a){let o=r+this.left.height,i=a+this.left.length+this.break,s=t==ga.ByHeight?e<o:e<i,l=s?this.left.lineAt(e,t,n,r,a):this.right.lineAt(e,t,n,o,i);if(this.break||(s?l.to<i:l.from>i))return l;let c=t==ga.ByPosNoHeight?ga.ByPosNoHeight:ga.ByPos;return s?l.join(this.right.lineAt(i,c,n,o,i)):this.left.lineAt(i,c,n,r,a).join(l)}forEachLine(e,t,n,r,a,o){let i=r+this.left.height,s=a+this.left.length+this.break;if(this.break)e<s&&this.left.forEachLine(e,t,n,r,a,o),t>=s&&this.right.forEachLine(e,t,n,i,s,o);else{let l=this.lineAt(s,ga.ByPos,n,r,a);e<l.from&&this.left.forEachLine(e,l.from-1,n,r,a,o),l.to>=e&&l.from<=t&&o(l),t>l.to&&this.right.forEachLine(l.to+1,t,n,i,s,o)}}replace(e,t,n){let r=this.left.length+this.break;if(t<r)return this.balanced(this.left.replace(e,t,n),this.right);if(e>this.left.length)return this.balanced(this.left,this.right.replace(e-r,t-r,n));let a=[];e>0&&this.decomposeLeft(e,a);let o=a.length;for(let e of n)a.push(e);if(e>0&&Ea(a,o-1),t<this.length){let e=a.length;this.decomposeRight(t,a),Ea(a,e)}return ba.of(a)}decomposeLeft(e,t){let n=this.left.length;if(e<=n)return this.left.decomposeLeft(e,t);t.push(this.left),this.break&&(n++,e>=n&&t.push(null)),e>n&&this.right.decomposeLeft(e-n,t)}decomposeRight(e,t){let n=this.left.length,r=n+this.break;if(e>=r)return this.right.decomposeRight(e-r,t);e<n&&this.left.decomposeRight(e,t),this.break&&e<r&&t.push(null),t.push(this.right)}balanced(e,t){return e.size>2*t.size||t.size>2*e.size?ba.of(this.break?[e,null,t]:[e,t]):(this.left=ya(this.left,e),this.right=ya(this.right,t),this.setHeight(e.height+t.height),this.outdated=e.outdated||t.outdated,this.size=e.size+t.size,this.length=e.length+this.break+t.length,this)}updateHeight(e,t=0,n=!1,r){let{left:a,right:o}=this,i=t+a.length+this.break,s=null;return r&&r.from<=t+a.length&&r.more?s=a=a.updateHeight(e,t,n,r):a.updateHeight(e,t,n),r&&r.from<=i+o.length&&r.more?s=o=o.updateHeight(e,i,n,r):o.updateHeight(e,i,n),s?this.balanced(a,o):(this.height=this.left.height+this.right.height,this.outdated=!1,this)}toString(){return this.left+(this.break?" ":"-")+this.right}}function Ea(e,t){let n,r;null==e[t]&&(n=e[t-1])instanceof ka&&(r=e[t+1])instanceof ka&&e.splice(t-1,3,new ka(n.length+1+r.length))}class Sa{constructor(e,t){this.pos=e,this.oracle=t,this.nodes=[],this.lineStart=-1,this.lineEnd=-1,this.covering=null,this.writtenTo=e}get isCovered(){return this.covering&&this.nodes[this.nodes.length-1]==this.covering}span(e,t){if(this.lineStart>-1){let e=Math.min(t,this.lineEnd),n=this.nodes[this.nodes.length-1];n instanceof Oa?n.length+=e-this.pos:(e>this.pos||!this.isCovered)&&this.nodes.push(new Oa(e-this.pos,-1)),this.writtenTo=e,t>e&&(this.nodes.push(null),this.writtenTo++,this.lineStart=-1)}this.pos=t}point(e,t,n){if(e<t||n.heightRelevant){let r=n.widget?n.widget.estimatedHeight:0,a=n.widget?n.widget.lineBreaks:0;r<0&&(r=this.oracle.lineHeight);let o=t-e;n.block?this.addBlock(new wa(o,r,n)):(o||a||r>=5)&&this.addLineDeco(r,a,o)}else t>e&&this.span(e,t);this.lineEnd>-1&&this.lineEnd<this.pos&&(this.lineEnd=this.oracle.doc.lineAt(this.pos).to)}enterLine(){if(this.lineStart>-1)return;let{from:e,to:t}=this.oracle.doc.lineAt(this.pos);this.lineStart=e,this.lineEnd=t,this.writtenTo<e&&((this.writtenTo<e-1||null==this.nodes[this.nodes.length-1])&&this.nodes.push(this.blankContent(this.writtenTo,e-1)),this.nodes.push(null)),this.pos>e&&this.nodes.push(new Oa(this.pos-e,-1)),this.writtenTo=this.pos}blankContent(e,t){let n=new ka(t-e);return this.oracle.doc.lineAt(e).to==t&&(n.flags|=4),n}ensureLine(){this.enterLine();let e=this.nodes.length?this.nodes[this.nodes.length-1]:null;if(e instanceof Oa)return e;let t=new Oa(0,-1);return this.nodes.push(t),t}addBlock(e){this.enterLine();let t=e.deco;t&&t.startSide>0&&!this.isCovered&&this.ensureLine(),this.nodes.push(e),this.writtenTo=this.pos=this.pos+e.length,t&&t.endSide>0&&(this.covering=e)}addLineDeco(e,t,n){let r=this.ensureLine();r.length+=n,r.collapsed+=n,r.widgetHeight=Math.max(r.widgetHeight,e),r.breaks+=t,this.writtenTo=this.pos=this.pos+n}finish(e){let t=0==this.nodes.length?null:this.nodes[this.nodes.length-1];!(this.lineStart>-1)||t instanceof Oa||this.isCovered?(this.writtenTo<this.pos||null==t)&&this.nodes.push(this.blankContent(this.writtenTo,this.pos)):this.nodes.push(new Oa(0,-1));let n=e;for(let e of this.nodes)e instanceof Oa&&e.updateHeight(this.oracle,n),n+=e?e.length:1;return this.nodes}static build(e,t,n,r){let a=new Sa(n,e);return Pe.spans(t,n,r,a,0),a.finish(n)}}class _a{constructor(){this.changes=[]}compareRange(){}comparePoint(e,t,n,r){(e<t||n&&n.heightRelevant||r&&r.heightRelevant)&&cn(e,t,this.changes,5)}}function Ca(e,t){let n=e.getBoundingClientRect(),r=e.ownerDocument,a=r.defaultView||window,o=Math.max(0,n.left),i=Math.min(a.innerWidth,n.right),s=Math.max(0,n.top),l=Math.min(a.innerHeight,n.bottom);for(let t=e.parentNode;t&&t!=r.body;)if(1==t.nodeType){let n=t,r=window.getComputedStyle(n);if((n.scrollHeight>n.clientHeight||n.scrollWidth>n.clientWidth)&&"visible"!=r.overflow){let r=n.getBoundingClientRect();o=Math.max(o,r.left),i=Math.min(i,r.right),s=Math.max(s,r.top),l=Math.min(t==e.parentNode?a.innerHeight:l,r.bottom)}t="absolute"==r.position||"fixed"==r.position?n.offsetParent:n.parentNode}else{if(11!=t.nodeType)break;t=t.host}return{left:o-n.left,right:Math.max(o,i)-n.left,top:s-(n.top+t),bottom:Math.max(s,l)-(n.top+t)}}function Ta(e,t){let n=e.getBoundingClientRect();return{left:0,right:n.right-n.left,top:t,bottom:n.bottom-(n.top+t)}}class Aa{constructor(e,t,n,r){this.from=e,this.to=t,this.size=n,this.displaySize=r}static same(e,t){if(e.length!=t.length)return!1;for(let n=0;n<e.length;n++){let r=e[n],a=t[n];if(r.from!=a.from||r.to!=a.to||r.size!=a.size)return!1}return!0}draw(e,t){return rn.replace({widget:new Ra(this.displaySize*(t?e.scaleY:e.scaleX),t)}).range(this.from,this.to)}}class Ra extends tn{constructor(e,t){super(),this.size=e,this.vertical=t}eq(e){return e.size==this.size&&e.vertical==this.vertical}toDOM(){let e=document.createElement("div");return this.vertical?e.style.height=this.size+"px":(e.style.width=this.size+"px",e.style.height="2px",e.style.display="inline-block"),e}get estimatedHeight(){return this.vertical?this.size:-1}}class Pa{constructor(e){this.state=e,this.pixelViewport={left:0,right:window.innerWidth,top:0,bottom:0},this.inView=!0,this.paddingTop=0,this.paddingBottom=0,this.contentDOMWidth=0,this.contentDOMHeight=0,this.editorHeight=0,this.editorWidth=0,this.scrollTop=0,this.scrolledToBottom=!1,this.scaleX=1,this.scaleY=1,this.scrollAnchorPos=0,this.scrollAnchorHeight=-1,this.scaler=Na,this.scrollTarget=null,this.printing=!1,this.mustMeasureContent=!0,this.defaultTextDirection=gn.LTR,this.visibleRanges=[],this.mustEnforceCursorAssoc=!1;let t=e.facet(rr).some((e=>"function"!=typeof e&&"cm-lineWrapping"==e.class));this.heightOracle=new ha(t),this.stateDeco=e.facet(ar).filter((e=>"function"!=typeof e)),this.heightMap=ba.empty().applyChanges(this.stateDeco,f.empty,this.heightOracle.setDoc(e.doc),[new pr(0,0,0,e.doc.length)]);for(let e=0;e<2&&(this.viewport=this.getViewport(0,null),this.updateForViewport());e++);this.updateViewportLines(),this.lineGaps=this.ensureLineGaps([]),this.lineGapDeco=rn.set(this.lineGaps.map((e=>e.draw(this,!1)))),this.computeVisibleRanges()}updateForViewport(){let e=[this.viewport],{main:t}=this.state.selection;for(let n=0;n<=1;n++){let r=n?t.head:t.anchor;if(!e.some((({from:e,to:t})=>r>=e&&r<=t))){let{from:t,to:n}=this.lineBlockAt(r);e.push(new Da(t,n))}}return this.viewports=e.sort(((e,t)=>e.from-t.from)),this.updateScaler()}updateScaler(){let e=this.scaler;return this.scaler=this.heightMap.height<=7e6?Na:new Ia(this.heightOracle,this.heightMap,this.viewports),e.eq(this.scaler)?0:2}updateViewportLines(){this.viewportLines=[],this.heightMap.forEachLine(this.viewport.from,this.viewport.to,this.heightOracle.setDoc(this.state.doc),0,0,(e=>{this.viewportLines.push(La(e,this.scaler))}))}update(e,t=null){this.state=e.state;let n=this.stateDeco;this.stateDeco=this.state.facet(ar).filter((e=>"function"!=typeof e));let r=e.changedRanges,a=pr.extendWithRanges(r,function(e,t,n){let r=new _a;return Pe.compare(e,t,n,r,0),r.changes}(n,this.stateDeco,e?e.changes:P.empty(this.state.doc.length))),o=this.heightMap.height,i=this.scrolledToBottom?null:this.scrollAnchorAt(this.scrollTop);pa(),this.heightMap=this.heightMap.applyChanges(this.stateDeco,e.startState.doc,this.heightOracle.setDoc(this.state.doc),a),(this.heightMap.height!=o||da)&&(e.flags|=2),i?(this.scrollAnchorPos=e.changes.mapPos(i.from,-1),this.scrollAnchorHeight=i.top):(this.scrollAnchorPos=-1,this.scrollAnchorHeight=this.heightMap.height);let s=a.length?this.mapViewport(this.viewport,e.changes):this.viewport;(t&&(t.range.head<s.from||t.range.head>s.to)||!this.viewportIsAppropriate(s))&&(s=this.getViewport(0,t));let l=s.from!=this.viewport.from||s.to!=this.viewport.to;this.viewport=s,e.flags|=this.updateForViewport(),(l||!e.changes.empty||2&e.flags)&&this.updateViewportLines(),(this.lineGaps.length||this.viewport.to-this.viewport.from>4e3)&&this.updateLineGaps(this.ensureLineGaps(this.mapLineGaps(this.lineGaps,e.changes))),e.flags|=this.computeVisibleRanges(e.changes),t&&(this.scrollTarget=t),!this.mustEnforceCursorAssoc&&e.selectionSet&&e.view.lineWrapping&&e.state.selection.main.empty&&e.state.selection.main.assoc&&!e.state.facet(Xn)&&(this.mustEnforceCursorAssoc=!0)}measure(e){let t=e.contentDOM,n=window.getComputedStyle(t),r=this.heightOracle,a=n.whiteSpace;this.defaultTextDirection="rtl"==n.direction?gn.RTL:gn.LTR;let o=this.heightOracle.mustRefreshForWrapping(a),i=t.getBoundingClientRect(),s=o||this.mustMeasureContent||this.contentDOMHeight!=i.height;this.contentDOMHeight=i.height,this.mustMeasureContent=!1;let l=0,c=0;if(i.width&&i.height){let{scaleX:e,scaleY:n}=ht(t,i);(e>.005&&Math.abs(this.scaleX-e)>.005||n>.005&&Math.abs(this.scaleY-n)>.005)&&(this.scaleX=e,this.scaleY=n,l|=16,o=s=!0)}let u=(parseInt(n.paddingTop)||0)*this.scaleY,d=(parseInt(n.paddingBottom)||0)*this.scaleY;this.paddingTop==u&&this.paddingBottom==d||(this.paddingTop=u,this.paddingBottom=d,l|=18),this.editorWidth!=e.scrollDOM.clientWidth&&(r.lineWrapping&&(s=!0),this.editorWidth=e.scrollDOM.clientWidth,l|=16);let p=e.scrollDOM.scrollTop*this.scaleY;this.scrollTop!=p&&(this.scrollAnchorHeight=-1,this.scrollTop=p),this.scrolledToBottom=Ot(e.scrollDOM);let h=(this.printing?Ta:Ca)(t,this.paddingTop),m=h.top-this.pixelViewport.top,g=h.bottom-this.pixelViewport.bottom;this.pixelViewport=h;let v=this.pixelViewport.bottom>this.pixelViewport.top&&this.pixelViewport.right>this.pixelViewport.left;if(v!=this.inView&&(this.inView=v,v&&(s=!0)),!this.inView&&!this.scrollTarget&&!function(e){let t=e.getBoundingClientRect(),n=e.ownerDocument.defaultView||window;return t.left<n.innerWidth&&t.right>0&&t.top<n.innerHeight&&t.bottom>0}(e.dom))return 0;let b=i.width;if(this.contentDOMWidth==b&&this.editorHeight==e.scrollDOM.clientHeight||(this.contentDOMWidth=i.width,this.editorHeight=e.scrollDOM.clientHeight,l|=16),s){let t=e.docView.measureVisibleLineHeights(this.viewport);if(r.mustRefreshForHeights(t)&&(o=!0),o||r.lineWrapping&&Math.abs(b-this.contentDOMWidth)>r.charWidth){let{lineHeight:n,charWidth:i,textHeight:s}=e.docView.measureTextSize();o=n>0&&r.refresh(a,n,i,s,b/i,t),o&&(e.docView.minWidth=0,l|=16)}m>0&&g>0?c=Math.max(m,g):m<0&&g<0&&(c=Math.min(m,g)),pa();for(let n of this.viewports){let a=n.from==this.viewport.from?t:e.docView.measureVisibleLineHeights(n);this.heightMap=(o?ba.empty().applyChanges(this.stateDeco,f.empty,this.heightOracle,[new pr(0,0,0,e.state.doc.length)]):this.heightMap).updateHeight(r,0,o,new fa(n.from,a))}da&&(l|=2)}let y=!this.viewportIsAppropriate(this.viewport,c)||this.scrollTarget&&(this.scrollTarget.range.head<this.viewport.from||this.scrollTarget.range.head>this.viewport.to);return y&&(2&l&&(l|=this.updateScaler()),this.viewport=this.getViewport(c,this.scrollTarget),l|=this.updateForViewport()),(2&l||y)&&this.updateViewportLines(),(this.lineGaps.length||this.viewport.to-this.viewport.from>4e3)&&this.updateLineGaps(this.ensureLineGaps(o?[]:this.lineGaps,e)),l|=this.computeVisibleRanges(),this.mustEnforceCursorAssoc&&(this.mustEnforceCursorAssoc=!1,e.docView.enforceCursorAssoc()),l}get visibleTop(){return this.scaler.fromDOM(this.pixelViewport.top)}get visibleBottom(){return this.scaler.fromDOM(this.pixelViewport.bottom)}getViewport(e,t){let n=.5-Math.max(-.5,Math.min(.5,e/1e3/2)),r=this.heightMap,a=this.heightOracle,{visibleTop:o,visibleBottom:i}=this,s=new Da(r.lineAt(o-1e3*n,ga.ByHeight,a,0,0).from,r.lineAt(i+1e3*(1-n),ga.ByHeight,a,0,0).to);if(t){let{head:e}=t.range;if(e<s.from||e>s.to){let n,o=Math.min(this.editorHeight,this.pixelViewport.bottom-this.pixelViewport.top),i=r.lineAt(e,ga.ByPos,a,0,0);n="center"==t.y?(i.top+i.bottom)/2-o/2:"start"==t.y||"nearest"==t.y&&e<s.from?i.top:i.bottom-o,s=new Da(r.lineAt(n-500,ga.ByHeight,a,0,0).from,r.lineAt(n+o+500,ga.ByHeight,a,0,0).to)}}return s}mapViewport(e,t){let n=t.mapPos(e.from,-1),r=t.mapPos(e.to,1);return new Da(this.heightMap.lineAt(n,ga.ByPos,this.heightOracle,0,0).from,this.heightMap.lineAt(r,ga.ByPos,this.heightOracle,0,0).to)}viewportIsAppropriate({from:e,to:t},n=0){if(!this.inView)return!0;let{top:r}=this.heightMap.lineAt(e,ga.ByPos,this.heightOracle,0,0),{bottom:a}=this.heightMap.lineAt(t,ga.ByPos,this.heightOracle,0,0),{visibleTop:o,visibleBottom:i}=this;return(0==e||r<=o-Math.max(10,Math.min(-n,250)))&&(t==this.state.doc.length||a>=i+Math.max(10,Math.min(n,250)))&&r>o-2e3&&a<i+2e3}mapLineGaps(e,t){if(!e.length||t.empty)return e;let n=[];for(let r of e)t.touchesRange(r.from,r.to)||n.push(new Aa(t.mapPos(r.from),t.mapPos(r.to),r.size,r.displaySize));return n}ensureLineGaps(e,t){let n=this.heightOracle.lineWrapping,r=n?1e4:2e3,a=r>>1,o=r<<1;if(this.defaultTextDirection!=gn.LTR&&!n)return[];let i=[],s=(r,o,l,c)=>{if(o-r<a)return;let u=this.state.selection.main,d=[u.from];u.empty||d.push(u.to);for(let e of d)if(e>r&&e<o)return s(r,e-10,l,c),void s(e+10,o,l,c);let p=function(e,t){for(let n of e)if(t(n))return n}(e,(e=>e.from>=l.from&&e.to<=l.to&&Math.abs(e.from-r)<a&&Math.abs(e.to-o)<a&&!d.some((t=>e.from<t&&e.to>t))));if(!p){if(o<l.to&&t&&n&&t.visibleRanges.some((e=>e.from<=o&&e.to>=o))){let e=t.moveToLineBoundary(F.cursor(o),!1,!0).head;e>r&&(o=e)}let e=this.gapSize(l,r,o,c);p=new Aa(r,o,e,n||e<2e6?e:2e6)}i.push(p)},l=t=>{if(t.length<o||t.type!=nn.Text)return;let a=function(e,t,n){let r=[],a=e,o=0;return Pe.spans(n,e,t,{span(){},point(e,t){e>a&&(r.push({from:a,to:e}),o+=e-a),a=t}},20),a<t&&(r.push({from:a,to:t}),o+=t-a),{total:o,ranges:r}}(t.from,t.to,this.stateDeco);if(a.total<o)return;let i,l,c=this.scrollTarget?this.scrollTarget.range.head:null;if(n){let e,n,o=r/this.heightOracle.lineLength*this.heightOracle.lineHeight;if(null!=c){let r=$a(a,c),i=((this.visibleBottom-this.visibleTop)/2+o)/t.height;e=r-i,n=r+i}else e=(this.visibleTop-t.top-o)/t.height,n=(this.visibleBottom-t.top+o)/t.height;i=Ma(a,e),l=Ma(a,n)}else{let n=a.total*this.heightOracle.charWidth,o=r*this.heightOracle.charWidth,s=0;if(n>2e6)for(let n of e)n.from>=t.from&&n.from<t.to&&n.size!=n.displaySize&&n.from*this.heightOracle.charWidth+s<this.pixelViewport.left&&(s=n.size-n.displaySize);let u,d,p=this.pixelViewport.left+s,h=this.pixelViewport.right+s;if(null!=c){let e=$a(a,c),t=((h-p)/2+o)/n;u=e-t,d=e+t}else u=(p-o)/n,d=(h+o)/n;i=Ma(a,u),l=Ma(a,d)}i>t.from&&s(t.from,i,t,a),l<t.to&&s(l,t.to,t,a)};for(let e of this.viewportLines)Array.isArray(e.type)?e.type.forEach(l):l(e);return i}gapSize(e,t,n,r){let a=$a(r,n)-$a(r,t);return this.heightOracle.lineWrapping?e.height*a:r.total*this.heightOracle.charWidth*a}updateLineGaps(e){Aa.same(e,this.lineGaps)||(this.lineGaps=e,this.lineGapDeco=rn.set(e.map((e=>e.draw(this,this.heightOracle.lineWrapping)))))}computeVisibleRanges(e){let t=this.stateDeco;this.lineGaps.length&&(t=t.concat(this.lineGapDeco));let n=[];Pe.spans(t,this.viewport.from,this.viewport.to,{span(e,t){n.push({from:e,to:t})},point(){}},20);let r=0;if(n.length!=this.visibleRanges.length)r=12;else for(let t=0;t<n.length&&!(8&r);t++){let a=this.visibleRanges[t],o=n[t];a.from==o.from&&a.to==o.to||(r|=4,e&&e.mapPos(a.from,-1)==o.from&&e.mapPos(a.to,1)==o.to||(r|=8))}return this.visibleRanges=n,r}lineBlockAt(e){return e>=this.viewport.from&&e<=this.viewport.to&&this.viewportLines.find((t=>t.from<=e&&t.to>=e))||La(this.heightMap.lineAt(e,ga.ByPos,this.heightOracle,0,0),this.scaler)}lineBlockAtHeight(e){return e>=this.viewportLines[0].top&&e<=this.viewportLines[this.viewportLines.length-1].bottom&&this.viewportLines.find((t=>t.top<=e&&t.bottom>=e))||La(this.heightMap.lineAt(this.scaler.fromDOM(e),ga.ByHeight,this.heightOracle,0,0),this.scaler)}scrollAnchorAt(e){let t=this.lineBlockAtHeight(e+8);return t.from>=this.viewport.from||this.viewportLines[0].top-e>200?t:this.viewportLines[0]}elementAtHeight(e){return La(this.heightMap.blockAt(this.scaler.fromDOM(e),this.heightOracle,0,0),this.scaler)}get docHeight(){return this.scaler.toDOM(this.heightMap.height)}get contentHeight(){return this.docHeight+this.paddingTop+this.paddingBottom}}class Da{constructor(e,t){this.from=e,this.to=t}}function Ma({total:e,ranges:t},n){if(n<=0)return t[0].from;if(n>=1)return t[t.length-1].to;let r=Math.floor(e*n);for(let e=0;;e++){let{from:n,to:a}=t[e],o=a-n;if(r<=o)return n+r;r-=o}}function $a(e,t){let n=0;for(let{from:r,to:a}of e.ranges){if(t<=a){n+=t-r;break}n+=a-r}return n/e.total}const Na={toDOM:e=>e,fromDOM:e=>e,scale:1,eq(e){return e==this}};class Ia{constructor(e,t,n){let r=0,a=0,o=0;this.viewports=n.map((({from:n,to:a})=>{let o=t.lineAt(n,ga.ByPos,e,0,0).top,i=t.lineAt(a,ga.ByPos,e,0,0).bottom;return r+=i-o,{from:n,to:a,top:o,bottom:i,domTop:0,domBottom:0}})),this.scale=(7e6-r)/(t.height-r);for(let e of this.viewports)e.domTop=o+(e.top-a)*this.scale,o=e.domBottom=e.domTop+(e.bottom-e.top),a=e.bottom}toDOM(e){for(let t=0,n=0,r=0;;t++){let a=t<this.viewports.length?this.viewports[t]:null;if(!a||e<a.top)return r+(e-n)*this.scale;if(e<=a.bottom)return a.domTop+(e-a.top);n=a.bottom,r=a.domBottom}}fromDOM(e){for(let t=0,n=0,r=0;;t++){let a=t<this.viewports.length?this.viewports[t]:null;if(!a||e<a.domTop)return n+(e-r)/this.scale;if(e<=a.domBottom)return a.top+(e-a.domTop);n=a.bottom,r=a.domBottom}}eq(e){return e instanceof Ia&&this.scale==e.scale&&this.viewports.length==e.viewports.length&&this.viewports.every(((t,n)=>t.from==e.viewports[n].from&&t.to==e.viewports[n].to))}}function La(e,t){if(1==t.scale)return e;let n=t.toDOM(e.top),r=t.toDOM(e.bottom);return new ma(e.from,e.length,n,r-n,Array.isArray(e._content)?e._content.map((e=>La(e,t))):e._content)}const Ba=z.define({combine:e=>e.join(" ")}),Fa=z.define({combine:e=>e.indexOf(!0)>-1}),Qa=Ve.newName(),Ua=Ve.newName(),za=Ve.newName(),ja={"&light":"."+Ua,"&dark":"."+za};function Wa(e,t,n){return new Ve(t,{finish:t=>/&/.test(t)?t.replace(/&\w*/,(t=>{if("&"==t)return e;if(!n||!n[t])throw new RangeError(`Unsupported selector: ${t}`);return n[t]})):e+" "+t})}const Xa=Wa("."+Qa,{"&":{position:"relative !important",boxSizing:"border-box","&.cm-focused":{outline:"1px dotted #212121"},display:"flex !important",flexDirection:"column"},".cm-scroller":{display:"flex !important",alignItems:"flex-start !important",fontFamily:"monospace",lineHeight:1.4,height:"100%",overflowX:"auto",position:"relative",zIndex:0,overflowAnchor:"none"},".cm-content":{margin:0,flexGrow:2,flexShrink:0,display:"block",whiteSpace:"pre",wordWrap:"normal",boxSizing:"border-box",minHeight:"100%",padding:"4px 0",outline:"none","&[contenteditable=true]":{WebkitUserModify:"read-write-plaintext-only"}},".cm-lineWrapping":{whiteSpace_fallback:"pre-wrap",whiteSpace:"break-spaces",wordBreak:"break-word",overflowWrap:"anywhere",flexShrink:1},"&light .cm-content":{caretColor:"black"},"&dark .cm-content":{caretColor:"white"},".cm-line":{display:"block",padding:"0 2px 0 6px"},".cm-layer":{position:"absolute",left:0,top:0,contain:"size style","& > *":{position:"absolute"}},"&light .cm-selectionBackground":{background:"#d9d9d9"},"&dark .cm-selectionBackground":{background:"#222"},"&light.cm-focused > .cm-scroller > .cm-selectionLayer .cm-selectionBackground":{background:"#d7d4f0"},"&dark.cm-focused > .cm-scroller > .cm-selectionLayer .cm-selectionBackground":{background:"#233"},".cm-cursorLayer":{pointerEvents:"none"},"&.cm-focused > .cm-scroller > .cm-cursorLayer":{animation:"steps(1) cm-blink 1.2s infinite"},"@keyframes cm-blink":{"0%":{},"50%":{opacity:0},"100%":{}},"@keyframes cm-blink2":{"0%":{},"50%":{opacity:0},"100%":{}},".cm-cursor, .cm-dropCursor":{borderLeft:"1.2px solid black",marginLeft:"-0.6px",pointerEvents:"none"},".cm-cursor":{display:"none"},"&dark .cm-cursor":{borderLeftColor:"#ddd"},".cm-dropCursor":{position:"absolute"},"&.cm-focused > .cm-scroller > .cm-cursorLayer .cm-cursor":{display:"block"},".cm-iso":{unicodeBidi:"isolate"},".cm-announced":{position:"fixed",top:"-10000px"},"@media print":{".cm-announced":{display:"none"}},"&light .cm-activeLine":{backgroundColor:"#cceeff44"},"&dark .cm-activeLine":{backgroundColor:"#99eeff33"},"&light .cm-specialChar":{color:"red"},"&dark .cm-specialChar":{color:"#f78"},".cm-gutters":{flexShrink:0,display:"flex",height:"100%",boxSizing:"border-box",insetInlineStart:0,zIndex:200},"&light .cm-gutters":{backgroundColor:"#f5f5f5",color:"#6c6c6c",borderRight:"1px solid #ddd"},"&dark .cm-gutters":{backgroundColor:"#333338",color:"#ccc"},".cm-gutter":{display:"flex !important",flexDirection:"column",flexShrink:0,boxSizing:"border-box",minHeight:"100%",overflow:"hidden"},".cm-gutterElement":{boxSizing:"border-box"},".cm-lineNumbers .cm-gutterElement":{padding:"0 3px 0 5px",minWidth:"20px",textAlign:"right",whiteSpace:"nowrap"},"&light .cm-activeLineGutter":{backgroundColor:"#e2f2ff"},"&dark .cm-activeLineGutter":{backgroundColor:"#222227"},".cm-panels":{boxSizing:"border-box",position:"sticky",left:0,right:0,zIndex:300},"&light .cm-panels":{backgroundColor:"#f5f5f5",color:"black"},"&light .cm-panels-top":{borderBottom:"1px solid #ddd"},"&light .cm-panels-bottom":{borderTop:"1px solid #ddd"},"&dark .cm-panels":{backgroundColor:"#333338",color:"white"},".cm-tab":{display:"inline-block",overflow:"hidden",verticalAlign:"bottom"},".cm-widgetBuffer":{verticalAlign:"text-top",height:"1em",width:0,display:"inline"},".cm-placeholder":{color:"#888",display:"inline-block",verticalAlign:"top"},".cm-highlightSpace":{backgroundImage:"radial-gradient(circle at 50% 55%, #aaa 20%, transparent 5%)",backgroundPosition:"center"},".cm-highlightTab":{backgroundImage:'url(\'data:image/svg+xml,<svg xmlns="http://www.w3.org/2000/svg" width="200" height="20"><path stroke="%23888" stroke-width="1" fill="none" d="M1 10H196L190 5M190 15L196 10M197 4L197 16"/></svg>\')',backgroundSize:"auto 100%",backgroundPosition:"right 90%",backgroundRepeat:"no-repeat"},".cm-trailingSpace":{backgroundColor:"#ff332255"},".cm-button":{verticalAlign:"middle",color:"inherit",fontSize:"70%",padding:".2em 1em",borderRadius:"1px"},"&light .cm-button":{backgroundImage:"linear-gradient(#eff1f5, #d9d9df)",border:"1px solid #888","&:active":{backgroundImage:"linear-gradient(#b4b4b4, #d0d3d6)"}},"&dark .cm-button":{backgroundImage:"linear-gradient(#393939, #111)",border:"1px solid #888","&:active":{backgroundImage:"linear-gradient(#111, #333)"}},".cm-textfield":{verticalAlign:"middle",color:"inherit",fontSize:"70%",border:"1px solid silver",padding:".2em .5em"},"&light .cm-textfield":{backgroundColor:"white"},"&dark .cm-textfield":{border:"1px solid #555",backgroundColor:"inherit"}},ja),qa={childList:!0,characterData:!0,subtree:!0,attributes:!0,characterDataOldValue:!0},Va=zt.ie&&zt.ie_version<=11;class Ha{constructor(e){this.view=e,this.active=!1,this.editContext=null,this.selectionRange=new ft,this.selectionChanged=!1,this.delayedFlush=-1,this.resizeTimeout=-1,this.queue=[],this.delayedAndroidKey=null,this.flushingAndroidKey=-1,this.lastChange=0,this.scrollTargets=[],this.intersection=null,this.resizeScroll=null,this.intersecting=!1,this.gapIntersection=null,this.gaps=[],this.printQuery=null,this.parentCheck=-1,this.dom=e.contentDOM,this.observer=new MutationObserver((t=>{for(let e of t)this.queue.push(e);(zt.ie&&zt.ie_version<=11||zt.ios&&e.composing)&&t.some((e=>"childList"==e.type&&e.removedNodes.length||"characterData"==e.type&&e.oldValue.length>e.target.nodeValue.length))?this.flushSoon():this.flush()})),!window.EditContext||!1===e.constructor.EDIT_CONTEXT||zt.chrome&&zt.chrome_version<126||(this.editContext=new Ka(e),e.state.facet(Kn)&&(e.contentDOM.editContext=this.editContext.editContext)),Va&&(this.onCharData=e=>{this.queue.push({target:e.target,type:"characterData",oldValue:e.prevValue}),this.flushSoon()}),this.onSelectionChange=this.onSelectionChange.bind(this),this.onResize=this.onResize.bind(this),this.onPrint=this.onPrint.bind(this),this.onScroll=this.onScroll.bind(this),window.matchMedia&&(this.printQuery=window.matchMedia("print")),"function"==typeof ResizeObserver&&(this.resizeScroll=new ResizeObserver((()=>{var e;(null===(e=this.view.docView)||void 0===e?void 0:e.lastUpdate)<Date.now()-75&&this.onResize()})),this.resizeScroll.observe(e.scrollDOM)),this.addWindowListeners(this.win=e.win),this.start(),"function"==typeof IntersectionObserver&&(this.intersection=new IntersectionObserver((e=>{this.parentCheck<0&&(this.parentCheck=setTimeout(this.listenForScroll.bind(this),1e3)),e.length>0&&e[e.length-1].intersectionRatio>0!=this.intersecting&&(this.intersecting=!this.intersecting,this.intersecting!=this.view.inView&&this.onScrollChanged(document.createEvent("Event")))}),{threshold:[0,.001]}),this.intersection.observe(this.dom),this.gapIntersection=new IntersectionObserver((e=>{e.length>0&&e[e.length-1].intersectionRatio>0&&this.onScrollChanged(document.createEvent("Event"))}),{})),this.listenForScroll(),this.readSelectionRange()}onScrollChanged(e){this.view.inputState.runHandlers("scroll",e),this.intersecting&&this.view.measure()}onScroll(e){this.intersecting&&this.flush(!1),this.editContext&&this.view.requestMeasure(this.editContext.measureReq),this.onScrollChanged(e)}onResize(){this.resizeTimeout<0&&(this.resizeTimeout=setTimeout((()=>{this.resizeTimeout=-1,this.view.requestMeasure()}),50))}onPrint(e){("change"!=e.type&&e.type||e.matches)&&(this.view.viewState.printing=!0,this.view.measure(),setTimeout((()=>{this.view.viewState.printing=!1,this.view.requestMeasure()}),500))}updateGaps(e){if(this.gapIntersection&&(e.length!=this.gaps.length||this.gaps.some(((t,n)=>t!=e[n])))){this.gapIntersection.disconnect();for(let t of e)this.gapIntersection.observe(t);this.gaps=e}}onSelectionChange(e){let t=this.selectionChanged;if(!this.readSelectionRange()||this.delayedAndroidKey)return;let{view:n}=this,r=this.selectionRange;if(n.state.facet(Kn)?n.root.activeElement!=this.dom:!at(this.dom,r))return;let a=r.anchorNode&&n.docView.nearest(r.anchorNode);a&&a.ignoreEvent(e)?t||(this.selectionChanged=!1):(zt.ie&&zt.ie_version<=11||zt.android&&zt.chrome)&&!n.state.selection.main.empty&&r.focusNode&&it(r.focusNode,r.focusOffset,r.anchorNode,r.anchorOffset)?this.flushSoon():this.flush(!1)}readSelectionRange(){let{view:e}=this,t=nt(e.root);if(!t)return!1;let n=zt.safari&&11==e.root.nodeType&&e.root.activeElement==this.dom&&function(e,t){if(t.getComposedRanges){let n=t.getComposedRanges(e.root)[0];if(n)return Ga(e,n)}let n=null;function r(e){e.preventDefault(),e.stopImmediatePropagation(),n=e.getTargetRanges()[0]}return e.contentDOM.addEventListener("beforeinput",r,!0),e.dom.ownerDocument.execCommand("indent"),e.contentDOM.removeEventListener("beforeinput",r,!0),n?Ga(e,n):null}(this.view,t)||t;if(!n||this.selectionRange.eq(n))return!1;let r=at(this.dom,n);return r&&!this.selectionChanged&&e.inputState.lastFocusTime>Date.now()-200&&e.inputState.lastTouchTime<Date.now()-300&&function(e,t){let n=t.focusNode,r=t.focusOffset;if(!n||t.anchorNode!=n||t.anchorOffset!=r)return!1;for(r=Math.min(r,ut(n));;)if(r){if(1!=n.nodeType)return!1;let e=n.childNodes[r-1];"false"==e.contentEditable?r--:(n=e,r=ut(n))}else{if(n==e)return!0;r=st(n),n=n.parentNode}}(this.dom,n)?(this.view.inputState.lastFocusTime=0,e.docView.updateSelection(),!1):(this.selectionRange.setRange(n),r&&(this.selectionChanged=!0),!0)}setSelectionRange(e,t){this.selectionRange.set(e.node,e.offset,t.node,t.offset),this.selectionChanged=!1}clearSelectionRange(){this.selectionRange.set(null,0,null,0)}listenForScroll(){this.parentCheck=-1;let e=0,t=null;for(let n=this.dom;n;)if(1==n.nodeType)!t&&e<this.scrollTargets.length&&this.scrollTargets[e]==n?e++:t||(t=this.scrollTargets.slice(0,e)),t&&t.push(n),n=n.assignedSlot||n.parentNode;else{if(11!=n.nodeType)break;n=n.host}if(e<this.scrollTargets.length&&!t&&(t=this.scrollTargets.slice(0,e)),t){for(let e of this.scrollTargets)e.removeEventListener("scroll",this.onScroll);for(let e of this.scrollTargets=t)e.addEventListener("scroll",this.onScroll)}}ignore(e){if(!this.active)return e();try{return this.stop(),e()}finally{this.start(),this.clear()}}start(){this.active||(this.observer.observe(this.dom,qa),Va&&this.dom.addEventListener("DOMCharacterDataModified",this.onCharData),this.active=!0)}stop(){this.active&&(this.active=!1,this.observer.disconnect(),Va&&this.dom.removeEventListener("DOMCharacterDataModified",this.onCharData))}clear(){this.processRecords(),this.queue.length=0,this.selectionChanged=!1}delayAndroidKey(e,t){var n;if(!this.delayedAndroidKey){let e=()=>{let e=this.delayedAndroidKey;e&&(this.clearDelayedAndroidKey(),this.view.inputState.lastKeyCode=e.keyCode,this.view.inputState.lastKeyTime=Date.now(),!this.flush()&&e.force&&yt(this.dom,e.key,e.keyCode))};this.flushingAndroidKey=this.view.win.requestAnimationFrame(e)}this.delayedAndroidKey&&"Enter"!=e||(this.delayedAndroidKey={key:e,keyCode:t,force:this.lastChange<Date.now()-50||!!(null===(n=this.delayedAndroidKey)||void 0===n?void 0:n.force)})}clearDelayedAndroidKey(){this.win.cancelAnimationFrame(this.flushingAndroidKey),this.delayedAndroidKey=null,this.flushingAndroidKey=-1}flushSoon(){this.delayedFlush<0&&(this.delayedFlush=this.view.win.requestAnimationFrame((()=>{this.delayedFlush=-1,this.flush()})))}forceFlush(){this.delayedFlush>=0&&(this.view.win.cancelAnimationFrame(this.delayedFlush),this.delayedFlush=-1),this.flush()}pendingRecords(){for(let e of this.observer.takeRecords())this.queue.push(e);return this.queue}processRecords(){let e=this.pendingRecords();e.length&&(this.queue=[]);let t=-1,n=-1,r=!1;for(let a of e){let e=this.readMutation(a);e&&(e.typeOver&&(r=!0),-1==t?({from:t,to:n}=e):(t=Math.min(e.from,t),n=Math.max(e.to,n)))}return{from:t,to:n,typeOver:r}}readChange(){let{from:e,to:t,typeOver:n}=this.processRecords(),r=this.selectionChanged&&at(this.dom,this.selectionRange);if(e<0&&!r)return null;e>-1&&(this.lastChange=Date.now()),this.view.inputState.lastFocusTime=0,this.selectionChanged=!1;let a=new Mr(this.view,e,t,n);return this.view.docView.domChanged={newSel:a.newSel?a.newSel.main:null},a}flush(e=!0){if(this.delayedFlush>=0||this.delayedAndroidKey)return!1;e&&this.readSelectionRange();let t=this.readChange();if(!t)return this.view.requestMeasure(),!1;let n=this.view.state,r=$r(this.view,t);return this.view.state==n&&(t.domChanged||t.newSel&&!t.newSel.main.eq(this.view.state.selection.main))&&this.view.update([]),r}readMutation(e){let t=this.view.docView.nearest(e.target);if(!t||t.ignoreMutation(e))return null;if(t.markDirty("attributes"==e.type),"attributes"==e.type&&(t.flags|=4),"childList"==e.type){let n=Za(t,e.previousSibling||e.target.previousSibling,-1),r=Za(t,e.nextSibling||e.target.nextSibling,1);return{from:n?t.posAfter(n):t.posAtStart,to:r?t.posBefore(r):t.posAtEnd,typeOver:!1}}return"characterData"==e.type?{from:t.posAtStart,to:t.posAtEnd,typeOver:e.target.nodeValue==e.oldValue}:null}setWindow(e){e!=this.win&&(this.removeWindowListeners(this.win),this.win=e,this.addWindowListeners(this.win))}addWindowListeners(e){e.addEventListener("resize",this.onResize),this.printQuery?this.printQuery.addEventListener?this.printQuery.addEventListener("change",this.onPrint):this.printQuery.addListener(this.onPrint):e.addEventListener("beforeprint",this.onPrint),e.addEventListener("scroll",this.onScroll),e.document.addEventListener("selectionchange",this.onSelectionChange)}removeWindowListeners(e){e.removeEventListener("scroll",this.onScroll),e.removeEventListener("resize",this.onResize),this.printQuery?this.printQuery.removeEventListener?this.printQuery.removeEventListener("change",this.onPrint):this.printQuery.removeListener(this.onPrint):e.removeEventListener("beforeprint",this.onPrint),e.document.removeEventListener("selectionchange",this.onSelectionChange)}update(e){this.editContext&&(this.editContext.update(e),e.startState.facet(Kn)!=e.state.facet(Kn)&&(e.view.contentDOM.editContext=e.state.facet(Kn)?this.editContext.editContext:null))}destroy(){var e,t,n;this.stop(),null===(e=this.intersection)||void 0===e||e.disconnect(),null===(t=this.gapIntersection)||void 0===t||t.disconnect(),null===(n=this.resizeScroll)||void 0===n||n.disconnect();for(let e of this.scrollTargets)e.removeEventListener("scroll",this.onScroll);this.removeWindowListeners(this.win),clearTimeout(this.parentCheck),clearTimeout(this.resizeTimeout),this.win.cancelAnimationFrame(this.delayedFlush),this.win.cancelAnimationFrame(this.flushingAndroidKey),this.editContext&&(this.view.contentDOM.editContext=null,this.editContext.destroy())}}function Za(e,t,n){for(;t;){let r=_t.get(t);if(r&&r.parent==e)return r;let a=t.parentNode;t=a!=e.dom?a:n>0?t.nextSibling:t.previousSibling}return null}function Ga(e,t){let n=t.startContainer,r=t.startOffset,a=t.endContainer,o=t.endOffset,i=e.docView.domAtPos(e.state.selection.main.anchor);return it(i.node,i.offset,a,o)&&([n,r,a,o]=[a,o,n,r]),{anchorNode:n,anchorOffset:r,focusNode:a,focusOffset:o}}class Ka{constructor(e){this.from=0,this.to=0,this.pendingContextChange=null,this.handlers=Object.create(null),this.composing=null,this.resetRange(e.state);let t=this.editContext=new window.EditContext({text:e.state.doc.sliceString(this.from,this.to),selectionStart:this.toContextPos(Math.max(this.from,Math.min(this.to,e.state.selection.main.anchor))),selectionEnd:this.toContextPos(e.state.selection.main.head)});this.handlers.textupdate=t=>{let n=e.state.selection.main,{anchor:r,head:a}=n,o=this.toEditorPos(t.updateRangeStart),i=this.toEditorPos(t.updateRangeEnd);e.inputState.composing>=0&&!this.composing&&(this.composing={contextBase:t.updateRangeStart,editorBase:o,drifted:!1});let s={from:o,to:i,insert:f.of(t.text.split("\n"))};if(s.from==this.from&&r<this.from?s.from=r:s.to==this.to&&r>this.to&&(s.to=r),s.from!=s.to||s.insert.length){if((zt.mac||zt.android)&&s.from==a-1&&/^\. ?$/.test(t.text)&&"off"==e.contentDOM.getAttribute("autocorrect")&&(s={from:o,to:i,insert:f.of([t.text.replace("."," ")])}),this.pendingContextChange=s,!e.state.readOnly){let n=this.to-this.from+(s.to-s.from+s.insert.length);Nr(e,s,F.single(this.toEditorPos(t.selectionStart,n),this.toEditorPos(t.selectionEnd,n)))}this.pendingContextChange&&(this.revertPending(e.state),this.setSelection(e.state))}else{let r=F.single(this.toEditorPos(t.selectionStart),this.toEditorPos(t.selectionEnd));r.main.eq(n)||e.dispatch({selection:r,userEvent:"select"})}},this.handlers.characterboundsupdate=n=>{let r=[],a=null;for(let t=this.toEditorPos(n.rangeStart),o=this.toEditorPos(n.rangeEnd);t<o;t++){let n=e.coordsForChar(t);a=n&&new DOMRect(n.left,n.top,n.right-n.left,n.bottom-n.top)||a||new DOMRect,r.push(a)}t.updateCharacterBounds(n.rangeStart,r)},this.handlers.textformatupdate=t=>{let n=[];for(let e of t.getTextFormats()){let t=e.underlineStyle,r=e.underlineThickness;if("None"!=t&&"None"!=r){let a=this.toEditorPos(e.rangeStart),o=this.toEditorPos(e.rangeEnd);if(a<o){let e=`text-decoration: underline ${"Dashed"==t?"dashed ":"Squiggle"==t?"wavy ":""}${"Thin"==r?1:2}px`;n.push(rn.mark({attributes:{style:e}}).range(a,o))}}}e.dispatch({effects:Zn.of(rn.set(n))})},this.handlers.compositionstart=()=>{e.inputState.composing<0&&(e.inputState.composing=0,e.inputState.compositionFirstChange=!0)},this.handlers.compositionend=()=>{if(e.inputState.composing=-1,e.inputState.compositionFirstChange=null,this.composing){let{drifted:t}=this.composing;this.composing=null,t&&this.reset(e.state)}};for(let e in this.handlers)t.addEventListener(e,this.handlers[e]);this.measureReq={read:e=>{this.editContext.updateControlBounds(e.contentDOM.getBoundingClientRect());let t=nt(e.root);t&&t.rangeCount&&this.editContext.updateSelectionBounds(t.getRangeAt(0).getBoundingClientRect())}}}applyEdits(e){let t=0,n=!1,r=this.pendingContextChange;return e.changes.iterChanges(((a,o,i,s,l)=>{if(n)return;let c=l.length-(o-a);if(r&&o>=r.to){if(r.from==a&&r.to==o&&r.insert.eq(l))return r=this.pendingContextChange=null,t+=c,void(this.to+=c);r=null,this.revertPending(e.state)}if(a+=t,(o+=t)<=this.from)this.from+=c,this.to+=c;else if(a<this.to){if(a<this.from||o>this.to||this.to-this.from+l.length>3e4)return void(n=!0);this.editContext.updateText(this.toContextPos(a),this.toContextPos(o),l.toString()),this.to+=c}t+=c})),r&&!n&&this.revertPending(e.state),!n}update(e){let t=this.pendingContextChange,n=e.startState.selection.main;this.composing&&(this.composing.drifted||!e.changes.touchesRange(n.from,n.to)&&e.transactions.some((e=>!e.isUserEvent("input.type")&&e.changes.touchesRange(this.from,this.to))))?(this.composing.drifted=!0,this.composing.editorBase=e.changes.mapPos(this.composing.editorBase)):this.applyEdits(e)&&this.rangeIsValid(e.state)?(e.docChanged||e.selectionSet||t)&&this.setSelection(e.state):(this.pendingContextChange=null,this.reset(e.state)),(e.geometryChanged||e.docChanged||e.selectionSet)&&e.view.requestMeasure(this.measureReq)}resetRange(e){let{head:t}=e.selection.main;this.from=Math.max(0,t-1e4),this.to=Math.min(e.doc.length,t+1e4)}reset(e){this.resetRange(e),this.editContext.updateText(0,this.editContext.text.length,e.doc.sliceString(this.from,this.to)),this.setSelection(e)}revertPending(e){let t=this.pendingContextChange;this.pendingContextChange=null,this.editContext.updateText(this.toContextPos(t.from),this.toContextPos(t.from+t.insert.length),e.doc.sliceString(t.from,t.to))}setSelection(e){let{main:t}=e.selection,n=this.toContextPos(Math.max(this.from,Math.min(this.to,t.anchor))),r=this.toContextPos(t.head);this.editContext.selectionStart==n&&this.editContext.selectionEnd==r||this.editContext.updateSelection(n,r)}rangeIsValid(e){let{head:t}=e.selection.main;return!(this.from>0&&t-this.from<500||this.to<e.doc.length&&this.to-t<500||this.to-this.from>3e4)}toEditorPos(e,t=this.to-this.from){e=Math.min(e,t);let n=this.composing;return n&&n.drifted?n.editorBase+(e-n.contextBase):e+this.from}toContextPos(e){let t=this.composing;return t&&t.drifted?t.contextBase+(e-t.editorBase):e-this.from}destroy(){for(let e in this.handlers)this.editContext.removeEventListener(e,this.handlers[e])}}class Ya{get state(){return this.viewState.state}get viewport(){return this.viewState.viewport}get visibleRanges(){return this.viewState.visibleRanges}get inView(){return this.viewState.inView}get composing(){return this.inputState.composing>0}get compositionStarted(){return this.inputState.composing>=0}get root(){return this._root}get win(){return this.dom.ownerDocument.defaultView||window}constructor(e={}){var t;this.plugins=[],this.pluginMap=new Map,this.editorAttrs={},this.contentAttrs={},this.bidiCache=[],this.destroyed=!1,this.updateState=2,this.measureScheduled=-1,this.measureRequests=[],this.contentDOM=document.createElement("div"),this.scrollDOM=document.createElement("div"),this.scrollDOM.tabIndex=-1,this.scrollDOM.className="cm-scroller",this.scrollDOM.appendChild(this.contentDOM),this.announceDOM=document.createElement("div"),this.announceDOM.className="cm-announced",this.announceDOM.setAttribute("aria-live","polite"),this.dom=document.createElement("div"),this.dom.appendChild(this.announceDOM),this.dom.appendChild(this.scrollDOM),e.parent&&e.parent.appendChild(this.dom);let{dispatch:n}=e;this.dispatchTransactions=e.dispatchTransactions||n&&(e=>e.forEach((e=>n(e,this))))||(e=>this.update(e)),this.dispatch=this.dispatch.bind(this),this._root=e.root||function(e){for(;e;){if(e&&(9==e.nodeType||11==e.nodeType&&e.host))return e;e=e.assignedSlot||e.parentNode}return null}(e.parent)||document,this.viewState=new Pa(e.state||Se.create(e)),e.scrollTo&&e.scrollTo.is(Hn)&&(this.viewState.scrollTarget=e.scrollTo.value.clip(this.viewState.state)),this.plugins=this.state.facet(Jn).map((e=>new tr(e)));for(let e of this.plugins)e.update(this);this.observer=new Ha(this),this.inputState=new Ir(this),this.inputState.ensureHandlers(this.plugins),this.docView=new fr(this),this.mountStyles(),this.updateAttrs(),this.updateState=0,this.requestMeasure(),(null===(t=document.fonts)||void 0===t?void 0:t.ready)&&document.fonts.ready.then((()=>this.requestMeasure()))}dispatch(...e){let t=1==e.length&&e[0]instanceof me?e:1==e.length&&Array.isArray(e[0])?e[0]:[this.state.update(...e)];this.dispatchTransactions(t,this)}update(e){if(0!=this.updateState)throw new Error("Calls to EditorView.update are not allowed while an update is in progress");let t,n=!1,r=!1,a=this.state;for(let t of e){if(t.startState!=a)throw new RangeError("Trying to update state with a transaction that doesn't start from the previous state.");a=t.state}if(this.destroyed)return void(this.viewState.state=a);let o=this.hasFocus,i=0,s=null;e.some((e=>e.annotation(ia)))?(this.inputState.notifiedFocused=o,i=1):o!=this.inputState.notifiedFocused&&(this.inputState.notifiedFocused=o,s=sa(a,o),s||(i=1));let l=this.observer.delayedAndroidKey,c=null;if(l?(this.observer.clearDelayedAndroidKey(),c=this.observer.readChange(),(c&&!this.state.doc.eq(a.doc)||!this.state.selection.eq(a.selection))&&(c=null)):this.observer.clear(),a.facet(Se.phrases)!=this.state.facet(Se.phrases))return this.setState(a);t=hr.create(this,a,e),t.flags|=i;let u=this.viewState.scrollTarget;try{this.updateState=2;for(let t of e){if(u&&(u=u.map(t.changes)),t.scrollIntoView){let{main:e}=t.state.selection;u=new Vn(e.empty?e:F.cursor(e.head,e.head>e.anchor?-1:1))}for(let e of t.effects)e.is(Hn)&&(u=e.value.clip(this.state))}this.viewState.update(t,u),this.bidiCache=to.update(this.bidiCache,t.changes),t.empty||(this.updatePlugins(t),this.inputState.update(t)),n=this.docView.update(t),this.state.facet(dr)!=this.styleModules&&this.mountStyles(),r=this.updateAttrs(),this.showAnnouncements(e),this.docView.updateSelection(n,e.some((e=>e.isUserEvent("select.pointer"))))}finally{this.updateState=0}if(t.startState.facet(Ba)!=t.state.facet(Ba)&&(this.viewState.mustMeasureContent=!0),(n||r||u||this.viewState.mustEnforceCursorAssoc||this.viewState.mustMeasureContent)&&this.requestMeasure(),n&&this.docViewUpdate(),!t.empty)for(let e of this.state.facet(Fn))try{e(t)}catch(e){Gn(this.state,e,"update listener")}(s||c)&&Promise.resolve().then((()=>{s&&this.state==s.startState&&this.dispatch(s),c&&!$r(this,c)&&l.force&&yt(this.contentDOM,l.key,l.keyCode)}))}setState(e){if(0!=this.updateState)throw new Error("Calls to EditorView.setState are not allowed while an update is in progress");if(this.destroyed)return void(this.viewState.state=e);this.updateState=2;let t=this.hasFocus;try{for(let e of this.plugins)e.destroy(this);this.viewState=new Pa(e),this.plugins=e.facet(Jn).map((e=>new tr(e))),this.pluginMap.clear();for(let e of this.plugins)e.update(this);this.docView.destroy(),this.docView=new fr(this),this.inputState.ensureHandlers(this.plugins),this.mountStyles(),this.updateAttrs(),this.bidiCache=[]}finally{this.updateState=0}t&&this.focus(),this.requestMeasure()}updatePlugins(e){let t=e.startState.facet(Jn),n=e.state.facet(Jn);if(t!=n){let r=[];for(let a of n){let n=t.indexOf(a);if(n<0)r.push(new tr(a));else{let t=this.plugins[n];t.mustUpdate=e,r.push(t)}}for(let t of this.plugins)t.mustUpdate!=e&&t.destroy(this);this.plugins=r,this.pluginMap.clear()}else for(let t of this.plugins)t.mustUpdate=e;for(let e=0;e<this.plugins.length;e++)this.plugins[e].update(this);t!=n&&this.inputState.ensureHandlers(this.plugins)}docViewUpdate(){for(let e of this.plugins){let t=e.value;if(t&&t.docViewUpdate)try{t.docViewUpdate(this)}catch(e){Gn(this.state,e,"doc view update listener")}}}measure(e=!0){if(this.destroyed)return;if(this.measureScheduled>-1&&this.win.cancelAnimationFrame(this.measureScheduled),this.observer.delayedAndroidKey)return this.measureScheduled=-1,void this.requestMeasure();this.measureScheduled=0,e&&this.observer.forceFlush();let t=null,n=this.scrollDOM,r=n.scrollTop*this.scaleY,{scrollAnchorPos:a,scrollAnchorHeight:o}=this.viewState;Math.abs(r-this.viewState.scrollTop)>1&&(o=-1),this.viewState.scrollAnchorHeight=-1;try{for(let e=0;;e++){if(o<0)if(Ot(n))a=-1,o=this.viewState.heightMap.height;else{let e=this.viewState.scrollAnchorAt(r);a=e.from,o=e.top}this.updateState=1;let i=this.viewState.measure(this);if(!i&&!this.measureRequests.length&&null==this.viewState.scrollTarget)break;if(e>5){console.warn(this.measureRequests.length?"Measure loop restarted more than 5 times":"Viewport failed to stabilize");break}let s=[];4&i||([this.measureRequests,s]=[s,this.measureRequests]);let l=s.map((e=>{try{return e.read(this)}catch(e){return Gn(this.state,e),eo}})),c=hr.create(this,this.state,[]),u=!1;c.flags|=i,t?t.flags|=i:t=c,this.updateState=2,c.empty||(this.updatePlugins(c),this.inputState.update(c),this.updateAttrs(),u=this.docView.update(c),u&&this.docViewUpdate());for(let e=0;e<s.length;e++)if(l[e]!=eo)try{let t=s[e];t.write&&t.write(l[e],this)}catch(e){Gn(this.state,e)}if(u&&this.docView.updateSelection(!0),!c.viewportChanged&&0==this.measureRequests.length){if(this.viewState.editorHeight){if(this.viewState.scrollTarget){this.docView.scrollIntoView(this.viewState.scrollTarget),this.viewState.scrollTarget=null,o=-1;continue}{let e=(a<0?this.viewState.heightMap.height:this.viewState.lineBlockAt(a).top)-o;if(e>1||e<-1){r+=e,n.scrollTop=r/this.scaleY,o=-1;continue}}}break}}}finally{this.updateState=0,this.measureScheduled=-1}if(t&&!t.empty)for(let e of this.state.facet(Fn))e(t)}get themeClasses(){return Qa+" "+(this.state.facet(Fa)?za:Ua)+" "+this.state.facet(Ba)}updateAttrs(){let e=no(this,nr,{class:"cm-editor"+(this.hasFocus?" cm-focused ":" ")+this.themeClasses}),t={spellcheck:"false",autocorrect:"off",autocapitalize:"off",writingsuggestions:"false",translate:"no",contenteditable:this.state.facet(Kn)?"true":"false",class:"cm-content",style:`${zt.tabSize}: ${this.state.tabSize}`,role:"textbox","aria-multiline":"true"};this.state.readOnly&&(t["aria-readonly"]="true"),no(this,rr,t);let n=this.observer.ignore((()=>{let n=Jt(this.contentDOM,this.contentAttrs,t),r=Jt(this.dom,this.editorAttrs,e);return n||r}));return this.editorAttrs=e,this.contentAttrs=t,n}showAnnouncements(e){let t=!0;for(let n of e)for(let e of n.effects)e.is(Ya.announce)&&(t&&(this.announceDOM.textContent=""),t=!1,this.announceDOM.appendChild(document.createElement("div")).textContent=e.value)}mountStyles(){this.styleModules=this.state.facet(dr);let e=this.state.facet(Ya.cspNonce);Ve.mount(this.root,this.styleModules.concat(Xa).reverse(),e?{nonce:e}:void 0)}readMeasured(){if(2==this.updateState)throw new Error("Reading the editor layout isn't allowed during an update");0==this.updateState&&this.measureScheduled>-1&&this.measure(!1)}requestMeasure(e){if(this.measureScheduled<0&&(this.measureScheduled=this.win.requestAnimationFrame((()=>this.measure()))),e){if(this.measureRequests.indexOf(e)>-1)return;if(null!=e.key)for(let t=0;t<this.measureRequests.length;t++)if(this.measureRequests[t].key===e.key)return void(this.measureRequests[t]=e);this.measureRequests.push(e)}}plugin(e){let t=this.pluginMap.get(e);return(void 0===t||t&&t.spec!=e)&&this.pluginMap.set(e,t=this.plugins.find((t=>t.spec==e))||null),t&&t.update(this).value}get documentTop(){return this.contentDOM.getBoundingClientRect().top+this.viewState.paddingTop}get documentPadding(){return{top:this.viewState.paddingTop,bottom:this.viewState.paddingBottom}}get scaleX(){return this.viewState.scaleX}get scaleY(){return this.viewState.scaleY}elementAtHeight(e){return this.readMeasured(),this.viewState.elementAtHeight(e)}lineBlockAtHeight(e){return this.readMeasured(),this.viewState.lineBlockAtHeight(e)}get viewportLineBlocks(){return this.viewState.viewportLines}lineBlockAt(e){return this.viewState.lineBlockAt(e)}get contentHeight(){return this.viewState.contentHeight}moveByChar(e,t,n){return Tr(this,e,_r(this,e,t,n))}moveByGroup(e,t){return Tr(this,e,_r(this,e,t,(t=>function(e,t,n){let r=e.state.charCategorizer(t),a=r(n);return e=>{let t=r(e);return a==ke.Space&&(a=t),a==t}}(this,e.head,t))))}visualLineSide(e,t){let n=this.bidiSpans(e),r=this.textDirectionAt(e.from),a=n[t?n.length-1:0];return F.cursor(a.side(t,r)+e.from,a.forward(!t,r)?1:-1)}moveToLineBoundary(e,t,n=!0){return function(e,t,n,r){let a=function(e,t){let n=e.lineBlockAt(t);if(Array.isArray(n.type))for(let e of n.type)if(e.to>t||e.to==t&&(e.to==n.to||e.type==nn.Text))return e;return n}(e,t.head),o=r&&a.type==nn.Text&&(e.lineWrapping||a.widgetLineBreaks)?e.coordsAtPos(t.assoc<0&&t.head>a.from?t.head-1:t.head):null;if(o){let t=e.dom.getBoundingClientRect(),r=e.textDirectionAt(a.from),i=e.posAtCoords({x:n==(r==gn.LTR)?t.right-1:t.left+1,y:(o.top+o.bottom)/2});if(null!=i)return F.cursor(i,n?-1:1)}return F.cursor(n?a.to:a.from,n?-1:1)}(this,e,t,n)}moveVertically(e,t,n){return Tr(this,e,function(e,t,n,r){let a=t.head,o=n?1:-1;if(a==(n?e.state.doc.length:0))return F.cursor(a,t.assoc);let i,s=t.goalColumn,l=e.contentDOM.getBoundingClientRect(),c=e.coordsAtPos(a,t.assoc||-1),u=e.documentTop;if(c)null==s&&(s=c.left-l.left),i=o<0?c.top:c.bottom;else{let t=e.viewState.lineBlockAt(a);null==s&&(s=Math.min(l.right-l.left,e.defaultCharacterWidth*(a-t.from))),i=(o<0?t.top:t.bottom)+u}let d=l.left+s,p=null!=r?r:e.viewState.heightOracle.textHeight>>1;for(let t=0;;t+=10){let n=i+(p+t)*o,r=Er(e,{x:d,y:n},!1,o);if(n<l.top||n>l.bottom||(o<0?r<a:r>a)){let t=e.docView.coordsForChar(r),a=!t||n<t.top?-1:1;return F.cursor(r,a,void 0,s)}}}(this,e,t,n))}domAtPos(e){return this.docView.domAtPos(e)}posAtDOM(e,t=0){return this.docView.posFromDOM(e,t)}posAtCoords(e,t=!0){return this.readMeasured(),Er(this,e,t)}coordsAtPos(e,t=1){this.readMeasured();let n=this.docView.coordsAt(e,t);if(!n||n.left==n.right)return n;let r=this.state.doc.lineAt(e),a=this.bidiSpans(r);return dt(n,a[_n.find(a,e-r.from,-1,t)].dir==gn.LTR==t>0)}coordsForChar(e){return this.readMeasured(),this.docView.coordsForChar(e)}get defaultCharacterWidth(){return this.viewState.heightOracle.charWidth}get defaultLineHeight(){return this.viewState.heightOracle.lineHeight}get textDirection(){return this.viewState.defaultTextDirection}textDirectionAt(e){return!this.state.facet(Wn)||e<this.viewport.from||e>this.viewport.to?this.textDirection:(this.readMeasured(),this.docView.textDirectionAt(e))}get lineWrapping(){return this.viewState.heightOracle.lineWrapping}bidiSpans(e){if(e.length>Ja)return Pn(e.length);let t,n=this.textDirectionAt(e.from);for(let r of this.bidiCache)if(r.from==e.from&&r.dir==n&&(r.fresh||Cn(r.isolates,t=lr(this,e))))return r.order;t||(t=lr(this,e));let r=function(e,t,n){if(!e)return[new _n(0,0,t==bn?1:0)];if(t==vn&&!n.length&&!Sn.test(e))return Pn(e.length);if(n.length)for(;e.length>Tn.length;)Tn[Tn.length]=256;let r=[],a=t==vn?0:1;return Rn(e,a,a,n,0,e.length,r),r}(e.text,n,t);return this.bidiCache.push(new to(e.from,e.to,n,t,!0,r)),r}get hasFocus(){var e;return(this.dom.ownerDocument.hasFocus()||zt.safari&&(null===(e=this.inputState)||void 0===e?void 0:e.lastContextMenu)>Date.now()-3e4)&&this.root.activeElement==this.contentDOM}focus(){this.observer.ignore((()=>{vt(this.contentDOM),this.docView.updateSelection()}))}setRoot(e){this._root!=e&&(this._root=e,this.observer.setWindow((9==e.nodeType?e:e.ownerDocument).defaultView||window),this.mountStyles())}destroy(){this.root.activeElement==this.contentDOM&&this.contentDOM.blur();for(let e of this.plugins)e.destroy(this);this.plugins=[],this.inputState.destroy(),this.docView.destroy(),this.dom.remove(),this.observer.destroy(),this.measureScheduled>-1&&this.win.cancelAnimationFrame(this.measureScheduled),this.destroyed=!0}static scrollIntoView(e,t={}){return Hn.of(new Vn("number"==typeof e?F.cursor(e):e,t.y,t.x,t.yMargin,t.xMargin))}scrollSnapshot(){let{scrollTop:e,scrollLeft:t}=this.scrollDOM,n=this.viewState.scrollAnchorAt(e);return Hn.of(new Vn(F.cursor(n.from),"start","start",n.top-e,t,!0))}setTabFocusMode(e){null==e?this.inputState.tabFocusMode=this.inputState.tabFocusMode<0?0:-1:"boolean"==typeof e?this.inputState.tabFocusMode=e?0:-1:0!=this.inputState.tabFocusMode&&(this.inputState.tabFocusMode=Date.now()+e)}static domEventHandlers(e){return er.define((()=>({})),{eventHandlers:e})}static domEventObservers(e){return er.define((()=>({})),{eventObservers:e})}static theme(e,t){let n=Ve.newName(),r=[Ba.of(n),dr.of(Wa(`.${n}`,e))];return t&&t.dark&&r.push(Fa.of(!0)),r}static baseTheme(e){return K.lowest(dr.of(Wa("."+Qa,e,ja)))}static findFromDOM(e){var t;let n=e.querySelector(".cm-content"),r=n&&_t.get(n)||_t.get(e);return(null===(t=null==r?void 0:r.rootView)||void 0===t?void 0:t.view)||null}}Ya.styleModule=dr,Ya.inputHandler=Qn,Ya.clipboardInputFilter=zn,Ya.clipboardOutputFilter=jn,Ya.scrollHandler=qn,Ya.focusChangeEffect=Un,Ya.perLineTextDirection=Wn,Ya.exceptionSink=Bn,Ya.updateListener=Fn,Ya.editable=Kn,Ya.mouseSelectionStyle=Ln,Ya.dragMovesSelection=In,Ya.clickAddsSelectionRange=Nn,Ya.decorations=ar,Ya.outerDecorations=or,Ya.atomicRanges=ir,Ya.bidiIsolatedRanges=sr,Ya.scrollMargins=cr,Ya.darkTheme=Fa,Ya.cspNonce=z.define({combine:e=>e.length?e[0]:""}),Ya.contentAttributes=rr,Ya.editorAttributes=nr,Ya.lineWrapping=Ya.contentAttributes.of({class:"cm-lineWrapping"}),Ya.announce=fe.define();const Ja=4096,eo={};class to{constructor(e,t,n,r,a,o){this.from=e,this.to=t,this.dir=n,this.isolates=r,this.fresh=a,this.order=o}static update(e,t){if(t.empty&&!e.some((e=>e.fresh)))return e;let n=[],r=e.length?e[e.length-1].dir:gn.LTR;for(let a=Math.max(0,e.length-10);a<e.length;a++){let o=e[a];o.dir!=r||t.touchesRange(o.from,o.to)||n.push(new to(t.mapPos(o.from,1),t.mapPos(o.to,-1),o.dir,o.isolates,!1,o.order))}return n}}function no(e,t,n){for(let r=e.state.facet(t),a=r.length-1;a>=0;a--){let t=r[a],o="function"==typeof t?t(e):t;o&&Gt(o,n)}return n}const ro=zt.mac?"mac":zt.windows?"win":zt.linux?"linux":"key";function ao(e,t,n){return t.altKey&&(e="Alt-"+e),t.ctrlKey&&(e="Ctrl-"+e),t.metaKey&&(e="Meta-"+e),!1!==n&&t.shiftKey&&(e="Shift-"+e),e}const oo=K.default(Ya.domEventHandlers({keydown:(e,t)=>function(e,t,n,r){uo=t;let a=function(e){var t=!(Ye&&e.metaKey&&e.shiftKey&&!e.ctrlKey&&!e.altKey||Je&&e.shiftKey&&e.key&&1==e.key.length||"Unidentified"==e.key)&&e.key||(e.shiftKey?Ke:Ge)[e.keyCode]||e.key||"Unidentified";return"Esc"==t&&(t="Escape"),"Del"==t&&(t="Delete"),"Left"==t&&(t="ArrowLeft"),"Up"==t&&(t="ArrowUp"),"Right"==t&&(t="ArrowRight"),"Down"==t&&(t="ArrowDown"),t}(t),o=C(S(a,0))==a.length&&" "!=a,i="",s=!1,l=!1,c=!1;lo&&lo.view==n&&lo.scope==r&&(i=lo.prefix+" ",Ur.indexOf(t.keyCode)<0&&(l=!0,lo=null));let u,d,p=new Set,h=e=>{if(e){for(let t of e.run)if(!p.has(t)&&(p.add(t),t(n)))return e.stopPropagation&&(c=!0),!0;e.preventDefault&&(e.stopPropagation&&(c=!0),l=!0)}return!1},f=e[r];return f&&(h(f[i+ao(a,t,!o)])?s=!0:o&&(t.altKey||t.metaKey||t.ctrlKey)&&!(zt.windows&&t.ctrlKey&&t.altKey)&&(u=Ge[t.keyCode])&&u!=a?(h(f[i+ao(u,t,!0)])||t.shiftKey&&(d=Ke[t.keyCode])!=a&&d!=u&&h(f[i+ao(d,t,!1)]))&&(s=!0):o&&t.shiftKey&&h(f[i+ao(a,t,!0)])&&(s=!0),!s&&h(f._any)&&(s=!0)),l&&(s=!0),s&&c&&t.stopPropagation(),uo=null,s}(function(e){let t=e.facet(io),n=so.get(t);return n||so.set(t,n=function(e,t=ro){let n=Object.create(null),r=Object.create(null),a=(e,t)=>{let n=r[e];if(null==n)r[e]=t;else if(n!=t)throw new Error("Key binding "+e+" is used both as a regular binding and as a multi-stroke prefix")},o=(e,r,o,i,s)=>{var l,c;let u=n[e]||(n[e]=Object.create(null)),d=r.split(/ (?!$)/).map((e=>function(e,t){const n=e.split(/-(?!$)/);let r,a,o,i,s=n[n.length-1];"Space"==s&&(s=" ");for(let e=0;e<n.length-1;++e){const s=n[e];if(/^(cmd|meta|m)$/i.test(s))i=!0;else if(/^a(lt)?$/i.test(s))r=!0;else if(/^(c|ctrl|control)$/i.test(s))a=!0;else if(/^s(hift)?$/i.test(s))o=!0;else{if(!/^mod$/i.test(s))throw new Error("Unrecognized modifier name: "+s);"mac"==t?i=!0:a=!0}}return r&&(s="Alt-"+s),a&&(s="Ctrl-"+s),i&&(s="Meta-"+s),o&&(s="Shift-"+s),s}(e,t)));for(let t=1;t<d.length;t++){let n=d.slice(0,t).join(" ");a(n,!0),u[n]||(u[n]={preventDefault:!0,stopPropagation:!1,run:[t=>{let r=lo={view:t,prefix:n,scope:e};return setTimeout((()=>{lo==r&&(lo=null)}),co),!0}]})}let p=d.join(" ");a(p,!1);let h=u[p]||(u[p]={preventDefault:!1,stopPropagation:!1,run:(null===(c=null===(l=u._any)||void 0===l?void 0:l.run)||void 0===c?void 0:c.slice())||[]});o&&h.run.push(o),i&&(h.preventDefault=!0),s&&(h.stopPropagation=!0)};for(let r of e){let e=r.scope?r.scope.split(" "):["editor"];if(r.any)for(let t of e){let e=n[t]||(n[t]=Object.create(null));e._any||(e._any={preventDefault:!1,stopPropagation:!1,run:[]});let{any:a}=r;for(let t in e)e[t].run.push((e=>a(e,uo)))}let a=r[t]||r.key;if(a)for(let t of e)o(t,a,r.run,r.preventDefault,r.stopPropagation),r.shift&&o(t,"Shift-"+a,r.shift,r.preventDefault,r.stopPropagation)}return n}(t.reduce(((e,t)=>e.concat(t)),[]))),n}(t.state),e,t,"editor")})),io=z.define({enables:oo}),so=new WeakMap;let lo=null;const co=4e3;let uo=null;const po={".cm-line":{"& ::selection, &::selection":{backgroundColor:"transparent !important"}},".cm-content":{"& :focus":{caretColor:"initial !important","&::selection, & ::selection":{backgroundColor:"Highlight !important"}}}};function ho(e,t,n,r,a){t.lastIndex=0;for(let o,i=e.iterRange(n,r),s=n;!i.next().done;s+=i.value.length)if(!i.lineBreak)for(;o=t.exec(i.value);)a(s+o.index,o)}!(zt.ios&&zt.webkit&&zt.webkit_version<534)&&(po[".cm-line"].caretColor=po[".cm-content"].caretColor="transparent !important");class fo{constructor(e){const{regexp:t,decoration:n,decorate:r,boundary:a,maxLength:o=1e3}=e;if(!t.global)throw new RangeError("The regular expression given to MatchDecorator should have its 'g' flag set");if(this.regexp=t,r)this.addMatch=(e,t,n,a)=>r(a,n,n+e[0].length,e,t);else if("function"==typeof n)this.addMatch=(e,t,r,a)=>{let o=n(e,t,r);o&&a(r,r+e[0].length,o)};else{if(!n)throw new RangeError("Either 'decorate' or 'decoration' should be provided to MatchDecorator");this.addMatch=(e,t,r,a)=>a(r,r+e[0].length,n)}this.boundary=a,this.maxLength=o}createDeco(e){let t=new De,n=t.add.bind(t);for(let{from:t,to:r}of function(e,t){let n=e.visibleRanges;if(1==n.length&&n[0].from==e.viewport.from&&n[0].to==e.viewport.to)return n;let r=[];for(let{from:a,to:o}of n)a=Math.max(e.state.doc.lineAt(a).from,a-t),o=Math.min(e.state.doc.lineAt(o).to,o+t),r.length&&r[r.length-1].to>=a?r[r.length-1].to=o:r.push({from:a,to:o});return r}(e,this.maxLength))ho(e.state.doc,this.regexp,t,r,((t,r)=>this.addMatch(r,e,t,n)));return t.finish()}updateDeco(e,t){let n=1e9,r=-1;return e.docChanged&&e.changes.iterChanges(((t,a,o,i)=>{i>=e.view.viewport.from&&o<=e.view.viewport.to&&(n=Math.min(o,n),r=Math.max(i,r))})),e.viewportMoved||r-n>1e3?this.createDeco(e.view):r>-1?this.updateRange(e.view,t.map(e.changes),n,r):t}updateRange(e,t,n,r){for(let a of e.visibleRanges){let o=Math.max(a.from,n),i=Math.min(a.to,r);if(i>o){let n=e.state.doc.lineAt(o),r=n.to<i?e.state.doc.lineAt(i):n,s=Math.max(a.from,n.from),l=Math.min(a.to,r.to);if(this.boundary){for(;o>n.from;o--)if(this.boundary.test(n.text[o-1-n.from])){s=o;break}for(;i<r.to;i++)if(this.boundary.test(r.text[i-r.from])){l=i;break}}let c,u=[],d=(e,t,n)=>u.push(n.range(e,t));if(n==r)for(this.regexp.lastIndex=s-n.from;(c=this.regexp.exec(n.text))&&c.index<l-n.from;)this.addMatch(c,e,c.index+n.from,d);else ho(e.state.doc,this.regexp,s,l,((t,n)=>this.addMatch(n,e,t,d)));t=t.update({filterFrom:s,filterTo:l,filter:(e,t)=>e<s||t>l,add:u})}}return t}}const mo=null!=/x/.unicode?"gu":"g",go=new RegExp("[\0-\b\n--­؜​‎‏\u2028\u2029‭‮⁦⁧⁩\ufeff￹-￼]",mo),vo={0:"null",7:"bell",8:"backspace",10:"newline",11:"vertical tab",13:"carriage return",27:"escape",8203:"zero width space",8204:"zero width non-joiner",8205:"zero width joiner",8206:"left-to-right mark",8207:"right-to-left mark",8232:"line separator",8237:"left-to-right override",8238:"right-to-left override",8294:"left-to-right isolate",8295:"right-to-left isolate",8297:"pop directional isolate",8233:"paragraph separator",65279:"zero width no-break space",65532:"object replacement"};let bo=null;const yo=z.define({combine(e){let t=_e(e,{render:null,specialChars:go,addSpecialChars:null});return(t.replaceTabs=!function(){var e;if(null==bo&&"undefined"!=typeof document&&document.body){let t=document.body.style;bo=null!=(null!==(e=t.tabSize)&&void 0!==e?e:t.MozTabSize)}return bo||!1}())&&(t.specialChars=new RegExp("\t|"+t.specialChars.source,mo)),t.addSpecialChars&&(t.specialChars=new RegExp(t.specialChars.source+"|"+t.addSpecialChars.source,mo)),t}});function wo(e={}){return[yo.of(e),Oo||(Oo=er.fromClass(class{constructor(e){this.view=e,this.decorations=rn.none,this.decorationCache=Object.create(null),this.decorator=this.makeDecorator(e.state.facet(yo)),this.decorations=this.decorator.createDeco(e)}makeDecorator(e){return new fo({regexp:e.specialChars,decoration:(t,n,r)=>{let{doc:a}=n.state,o=S(t[0],0);if(9==o){let e=a.lineAt(r),t=n.state.tabSize,o=je(e.text,t,r-e.from);return rn.replace({widget:new xo((t-o%t)*this.view.defaultCharacterWidth/this.view.scaleX)})}return this.decorationCache[o]||(this.decorationCache[o]=rn.replace({widget:new ko(e,o)}))},boundary:e.replaceTabs?void 0:/[^]/})}update(e){let t=e.state.facet(yo);e.startState.facet(yo)!=t?(this.decorator=this.makeDecorator(t),this.decorations=this.decorator.createDeco(e.view)):this.decorations=this.decorator.updateDeco(e,this.decorations)}},{decorations:e=>e.decorations}))]}let Oo=null;class ko extends tn{constructor(e,t){super(),this.options=e,this.code=t}eq(e){return e.code==this.code}toDOM(e){let t=function(e){return e>=32?"•":10==e?"␤":String.fromCharCode(9216+e)}(this.code),n=e.state.phrase("Control character")+" "+(vo[this.code]||"0x"+this.code.toString(16)),r=this.options.render&&this.options.render(this.code,n,t);if(r)return r;let a=document.createElement("span");return a.textContent=t,a.title=n,a.setAttribute("aria-label",n),a.className="cm-specialChar",a}ignoreEvent(){return!1}}class xo extends tn{constructor(e){super(),this.width=e}eq(e){return e.width==this.width}toDOM(){let e=document.createElement("span");return e.textContent="\t",e.className="cm-tab",e.style.width=this.width+"px",e}ignoreEvent(){return!1}}class Eo extends tn{constructor(e){super(),this.content=e}toDOM(e){let t=document.createElement("span");return t.className="cm-placeholder",t.style.pointerEvents="none",t.appendChild("string"==typeof this.content?document.createTextNode(this.content):"function"==typeof this.content?this.content(e):this.content.cloneNode(!0)),"string"==typeof this.content?t.setAttribute("aria-label","placeholder "+this.content):t.setAttribute("aria-hidden","true"),t}coordsAt(e){let t=e.firstChild?ot(e.firstChild):[];if(!t.length)return null;let n=window.getComputedStyle(e.parentNode),r=dt(t[0],"rtl"!=n.direction),a=parseInt(n.lineHeight);return r.bottom-r.top>1.5*a?{left:r.left,right:r.right,top:r.top,bottom:r.top+a}:r}ignoreEvent(){return!1}}const So="-10000px";class _o{constructor(e,t,n,r){this.facet=t,this.createTooltipView=n,this.removeTooltipView=r,this.input=e.state.facet(t),this.tooltips=this.input.filter((e=>e));let a=null;this.tooltipViews=this.tooltips.map((e=>a=n(e,a)))}update(e,t){var n;let r=e.state.facet(this.facet),a=r.filter((e=>e));if(r===this.input){for(let t of this.tooltipViews)t.update&&t.update(e);return!1}let o=[],i=t?[]:null;for(let n=0;n<a.length;n++){let r=a[n],s=-1;if(r){for(let e=0;e<this.tooltips.length;e++){let t=this.tooltips[e];t&&t.create==r.create&&(s=e)}if(s<0)o[n]=this.createTooltipView(r,n?o[n-1]:null),i&&(i[n]=!!r.above);else{let r=o[n]=this.tooltipViews[s];i&&(i[n]=t[s]),r.update&&r.update(e)}}}for(let e of this.tooltipViews)o.indexOf(e)<0&&(this.removeTooltipView(e),null===(n=e.destroy)||void 0===n||n.call(e));return t&&(i.forEach(((e,n)=>t[n]=e)),t.length=i.length),this.input=r,this.tooltips=a,this.tooltipViews=o,!0}}function Co(e){let t=e.dom.ownerDocument.documentElement;return{top:0,left:0,bottom:t.clientHeight,right:t.clientWidth}}const To=z.define({combine:e=>{var t,n,r;return{position:zt.ios?"absolute":(null===(t=e.find((e=>e.position)))||void 0===t?void 0:t.position)||"fixed",parent:(null===(n=e.find((e=>e.parent)))||void 0===n?void 0:n.parent)||null,tooltipSpace:(null===(r=e.find((e=>e.tooltipSpace)))||void 0===r?void 0:r.tooltipSpace)||Co}}}),Ao=new WeakMap,Ro=er.fromClass(class{constructor(e){this.view=e,this.above=[],this.inView=!0,this.madeAbsolute=!1,this.lastTransaction=0,this.measureTimeout=-1;let t=e.state.facet(To);this.position=t.position,this.parent=t.parent,this.classes=e.themeClasses,this.createContainer(),this.measureReq={read:this.readMeasure.bind(this),write:this.writeMeasure.bind(this),key:this},this.resizeObserver="function"==typeof ResizeObserver?new ResizeObserver((()=>this.measureSoon())):null,this.manager=new _o(e,$o,((e,t)=>this.createTooltip(e,t)),(e=>{this.resizeObserver&&this.resizeObserver.unobserve(e.dom),e.dom.remove()})),this.above=this.manager.tooltips.map((e=>!!e.above)),this.intersectionObserver="function"==typeof IntersectionObserver?new IntersectionObserver((e=>{Date.now()>this.lastTransaction-50&&e.length>0&&e[e.length-1].intersectionRatio<1&&this.measureSoon()}),{threshold:[1]}):null,this.observeIntersection(),e.win.addEventListener("resize",this.measureSoon=this.measureSoon.bind(this)),this.maybeMeasure()}createContainer(){this.parent?(this.container=document.createElement("div"),this.container.style.position="relative",this.container.className=this.view.themeClasses,this.parent.appendChild(this.container)):this.container=this.view.dom}observeIntersection(){if(this.intersectionObserver){this.intersectionObserver.disconnect();for(let e of this.manager.tooltipViews)this.intersectionObserver.observe(e.dom)}}measureSoon(){this.measureTimeout<0&&(this.measureTimeout=setTimeout((()=>{this.measureTimeout=-1,this.maybeMeasure()}),50))}update(e){e.transactions.length&&(this.lastTransaction=Date.now());let t=this.manager.update(e,this.above);t&&this.observeIntersection();let n=t||e.geometryChanged,r=e.state.facet(To);if(r.position!=this.position&&!this.madeAbsolute){this.position=r.position;for(let e of this.manager.tooltipViews)e.dom.style.position=this.position;n=!0}if(r.parent!=this.parent){this.parent&&this.container.remove(),this.parent=r.parent,this.createContainer();for(let e of this.manager.tooltipViews)this.container.appendChild(e.dom);n=!0}else this.parent&&this.view.themeClasses!=this.classes&&(this.classes=this.container.className=this.view.themeClasses);n&&this.maybeMeasure()}createTooltip(e,t){let n=e.create(this.view),r=t?t.dom:null;if(n.dom.classList.add("cm-tooltip"),e.arrow&&!n.dom.querySelector(".cm-tooltip > .cm-tooltip-arrow")){let e=document.createElement("div");e.className="cm-tooltip-arrow",n.dom.appendChild(e)}return n.dom.style.position=this.position,n.dom.style.top=So,n.dom.style.left="0px",this.container.insertBefore(n.dom,r),n.mount&&n.mount(this.view),this.resizeObserver&&this.resizeObserver.observe(n.dom),n}destroy(){var e,t,n;this.view.win.removeEventListener("resize",this.measureSoon);for(let t of this.manager.tooltipViews)t.dom.remove(),null===(e=t.destroy)||void 0===e||e.call(t);this.parent&&this.container.remove(),null===(t=this.resizeObserver)||void 0===t||t.disconnect(),null===(n=this.intersectionObserver)||void 0===n||n.disconnect(),clearTimeout(this.measureTimeout)}readMeasure(){let e=1,t=1,n=!1;if("fixed"==this.position&&this.manager.tooltipViews.length){let{dom:e}=this.manager.tooltipViews[0];if(zt.gecko)n=e.offsetParent!=this.container.ownerDocument.body;else if(e.style.top==So&&"0px"==e.style.left){let t=e.getBoundingClientRect();n=Math.abs(t.top+1e4)>1||Math.abs(t.left)>1}}if(n||"absolute"==this.position)if(this.parent){let n=this.parent.getBoundingClientRect();n.width&&n.height&&(e=n.width/this.parent.offsetWidth,t=n.height/this.parent.offsetHeight)}else({scaleX:e,scaleY:t}=this.view.viewState);let r=this.view.scrollDOM.getBoundingClientRect(),a=ur(this.view);return{visible:{left:r.left+a.left,top:r.top+a.top,right:r.right-a.right,bottom:r.bottom-a.bottom},parent:this.parent?this.container.getBoundingClientRect():this.view.dom.getBoundingClientRect(),pos:this.manager.tooltips.map(((e,t)=>{let n=this.manager.tooltipViews[t];return n.getCoords?n.getCoords(e.pos):this.view.coordsAtPos(e.pos)})),size:this.manager.tooltipViews.map((({dom:e})=>e.getBoundingClientRect())),space:this.view.state.facet(To).tooltipSpace(this.view),scaleX:e,scaleY:t,makeAbsolute:n}}writeMeasure(e){var t;if(e.makeAbsolute){this.madeAbsolute=!0,this.position="absolute";for(let e of this.manager.tooltipViews)e.dom.style.position="absolute"}let{visible:n,space:r,scaleX:a,scaleY:o}=e,i=[];for(let s=0;s<this.manager.tooltips.length;s++){let l=this.manager.tooltips[s],c=this.manager.tooltipViews[s],{dom:u}=c,d=e.pos[s],p=e.size[s];if(!d||!1!==l.clip&&(d.bottom<=Math.max(n.top,r.top)||d.top>=Math.min(n.bottom,r.bottom)||d.right<Math.max(n.left,r.left)-.1||d.left>Math.min(n.right,r.right)+.1)){u.style.top=So;continue}let h=l.arrow?c.dom.querySelector(".cm-tooltip-arrow"):null,f=h?7:0,m=p.right-p.left,g=null!==(t=Ao.get(c))&&void 0!==t?t:p.bottom-p.top,v=c.offset||Mo,b=this.view.textDirection==gn.LTR,y=p.width>r.right-r.left?b?r.left:r.right-p.width:b?Math.max(r.left,Math.min(d.left-(h?14:0)+v.x,r.right-m)):Math.min(Math.max(r.left,d.left-m+(h?14:0)-v.x),r.right-m),w=this.above[s];!l.strictSide&&(w?d.top-g-f-v.y<r.top:d.bottom+g+f+v.y>r.bottom)&&w==r.bottom-d.bottom>d.top-r.top&&(w=this.above[s]=!w);let O=(w?d.top-r.top:r.bottom-d.bottom)-f;if(O<g&&!1!==c.resize){if(O<this.view.defaultLineHeight){u.style.top=So;continue}Ao.set(c,g),u.style.height=(g=O)/o+"px"}else u.style.height&&(u.style.height="");let k=w?d.top-g-f-v.y:d.bottom+f+v.y,x=y+m;if(!0!==c.overlap)for(let e of i)e.left<x&&e.right>y&&e.top<k+g&&e.bottom>k&&(k=w?e.top-g-2-f:e.bottom+f+2);if("absolute"==this.position?(u.style.top=(k-e.parent.top)/o+"px",Po(u,(y-e.parent.left)/a)):(u.style.top=k/o+"px",Po(u,y/a)),h){let e=d.left+(b?v.x:-v.x)-(y+14-7);h.style.left=e/a+"px"}!0!==c.overlap&&i.push({left:y,top:k,right:x,bottom:k+g}),u.classList.toggle("cm-tooltip-above",w),u.classList.toggle("cm-tooltip-below",!w),c.positioned&&c.positioned(e.space)}}maybeMeasure(){if(this.manager.tooltips.length&&(this.view.inView&&this.view.requestMeasure(this.measureReq),this.inView!=this.view.inView&&(this.inView=this.view.inView,!this.inView)))for(let e of this.manager.tooltipViews)e.dom.style.top=So}},{eventObservers:{scroll(){this.maybeMeasure()}}});function Po(e,t){let n=parseInt(e.style.left,10);(isNaN(n)||Math.abs(t-n)>1)&&(e.style.left=t+"px")}const Do=Ya.baseTheme({".cm-tooltip":{zIndex:500,boxSizing:"border-box"},"&light .cm-tooltip":{border:"1px solid #bbb",backgroundColor:"#f5f5f5"},"&light .cm-tooltip-section:not(:first-child)":{borderTop:"1px solid #bbb"},"&dark .cm-tooltip":{backgroundColor:"#333338",color:"white"},".cm-tooltip-arrow":{height:"7px",width:"14px",position:"absolute",zIndex:-1,overflow:"hidden","&:before, &:after":{content:"''",position:"absolute",width:0,height:0,borderLeft:"7px solid transparent",borderRight:"7px solid transparent"},".cm-tooltip-above &":{bottom:"-7px","&:before":{borderTop:"7px solid #bbb"},"&:after":{borderTop:"7px solid #f5f5f5",bottom:"1px"}},".cm-tooltip-below &":{top:"-7px","&:before":{borderBottom:"7px solid #bbb"},"&:after":{borderBottom:"7px solid #f5f5f5",top:"1px"}}},"&dark .cm-tooltip .cm-tooltip-arrow":{"&:before":{borderTopColor:"#333338",borderBottomColor:"#333338"},"&:after":{borderTopColor:"transparent",borderBottomColor:"transparent"}}}),Mo={x:0,y:0},$o=z.define({enables:[Ro,Do]}),No=z.define({combine:e=>e.reduce(((e,t)=>e.concat(t)),[])});class Io{static create(e){return new Io(e)}constructor(e){this.view=e,this.mounted=!1,this.dom=document.createElement("div"),this.dom.classList.add("cm-tooltip-hover"),this.manager=new _o(e,No,((e,t)=>this.createHostedView(e,t)),(e=>e.dom.remove()))}createHostedView(e,t){let n=e.create(this.view);return n.dom.classList.add("cm-tooltip-section"),this.dom.insertBefore(n.dom,t?t.dom.nextSibling:this.dom.firstChild),this.mounted&&n.mount&&n.mount(this.view),n}mount(e){for(let t of this.manager.tooltipViews)t.mount&&t.mount(e);this.mounted=!0}positioned(e){for(let t of this.manager.tooltipViews)t.positioned&&t.positioned(e)}update(e){this.manager.update(e)}destroy(){var e;for(let t of this.manager.tooltipViews)null===(e=t.destroy)||void 0===e||e.call(t)}passProp(e){let t;for(let n of this.manager.tooltipViews){let r=n[e];if(void 0!==r)if(void 0===t)t=r;else if(t!==r)return}return t}get offset(){return this.passProp("offset")}get getCoords(){return this.passProp("getCoords")}get overlap(){return this.passProp("overlap")}get resize(){return this.passProp("resize")}}const Lo=$o.compute([No],(e=>{let t=e.facet(No);return 0===t.length?null:{pos:Math.min(...t.map((e=>e.pos))),end:Math.max(...t.map((e=>{var t;return null!==(t=e.end)&&void 0!==t?t:e.pos}))),create:Io.create,above:t[0].above,arrow:t.some((e=>e.arrow))}}));class Bo{constructor(e,t,n,r,a){this.view=e,this.source=t,this.field=n,this.setHover=r,this.hoverTime=a,this.hoverTimeout=-1,this.restartTimeout=-1,this.pending=null,this.lastMove={x:0,y:0,target:e.dom,time:0},this.checkHover=this.checkHover.bind(this),e.dom.addEventListener("mouseleave",this.mouseleave=this.mouseleave.bind(this)),e.dom.addEventListener("mousemove",this.mousemove=this.mousemove.bind(this))}update(){this.pending&&(this.pending=null,clearTimeout(this.restartTimeout),this.restartTimeout=setTimeout((()=>this.startHover()),20))}get active(){return this.view.state.field(this.field)}checkHover(){if(this.hoverTimeout=-1,this.active.length)return;let e=Date.now()-this.lastMove.time;e<this.hoverTime?this.hoverTimeout=setTimeout(this.checkHover,this.hoverTime-e):this.startHover()}startHover(){clearTimeout(this.restartTimeout);let{view:e,lastMove:t}=this,n=e.docView.nearest(t.target);if(!n)return;let r,a=1;if(n instanceof Xt)r=n.posAtStart;else{if(r=e.posAtCoords(t),null==r)return;let n=e.coordsAtPos(r);if(!n||t.y<n.top||t.y>n.bottom||t.x<n.left-e.defaultCharacterWidth||t.x>n.right+e.defaultCharacterWidth)return;let o=e.bidiSpans(e.state.doc.lineAt(r)).find((e=>e.from<=r&&e.to>=r)),i=o&&o.dir==gn.RTL?-1:1;a=t.x<n.left?-i:i}let o=this.source(e,r,a);if(null==o?void 0:o.then){let t=this.pending={pos:r};o.then((n=>{this.pending==t&&(this.pending=null,!n||Array.isArray(n)&&!n.length||e.dispatch({effects:this.setHover.of(Array.isArray(n)?n:[n])}))}),(t=>Gn(e.state,t,"hover tooltip")))}else!o||Array.isArray(o)&&!o.length||e.dispatch({effects:this.setHover.of(Array.isArray(o)?o:[o])})}get tooltip(){let e=this.view.plugin(Ro),t=e?e.manager.tooltips.findIndex((e=>e.create==Io.create)):-1;return t>-1?e.manager.tooltipViews[t]:null}mousemove(e){var t,n;this.lastMove={x:e.clientX,y:e.clientY,target:e.target,time:Date.now()},this.hoverTimeout<0&&(this.hoverTimeout=setTimeout(this.checkHover,this.hoverTime));let{active:r,tooltip:a}=this;if(r.length&&a&&!function(e,t){let n,{left:r,right:a,top:o,bottom:i}=e.getBoundingClientRect();if(n=e.querySelector(".cm-tooltip-arrow")){let e=n.getBoundingClientRect();o=Math.min(e.top,o),i=Math.max(e.bottom,i)}return t.clientX>=r-Fo&&t.clientX<=a+Fo&&t.clientY>=o-Fo&&t.clientY<=i+Fo}(a.dom,e)||this.pending){let{pos:a}=r[0]||this.pending,o=null!==(n=null===(t=r[0])||void 0===t?void 0:t.end)&&void 0!==n?n:a;(a==o?this.view.posAtCoords(this.lastMove)==a:function(e,t,n,r,a){let o=e.scrollDOM.getBoundingClientRect(),i=e.documentTop+e.documentPadding.top+e.contentHeight;if(o.left>r||o.right<r||o.top>a||Math.min(o.bottom,i)<a)return!1;let s=e.posAtCoords({x:r,y:a},!1);return s>=t&&s<=n}(this.view,a,o,e.clientX,e.clientY))||(this.view.dispatch({effects:this.setHover.of([])}),this.pending=null)}}mouseleave(e){clearTimeout(this.hoverTimeout),this.hoverTimeout=-1;let{active:t}=this;if(t.length){let{tooltip:t}=this;t&&t.dom.contains(e.relatedTarget)?this.watchTooltipLeave(t.dom):this.view.dispatch({effects:this.setHover.of([])})}}watchTooltipLeave(e){let t=n=>{e.removeEventListener("mouseleave",t),this.active.length&&!this.view.dom.contains(n.relatedTarget)&&this.view.dispatch({effects:this.setHover.of([])})};e.addEventListener("mouseleave",t)}destroy(){clearTimeout(this.hoverTimeout),this.view.dom.removeEventListener("mouseleave",this.mouseleave),this.view.dom.removeEventListener("mousemove",this.mousemove)}}const Fo=4;function Qo(e,t={}){let n=fe.define(),r=Z.define({create:()=>[],update(e,r){if(e.length&&(t.hideOnChange&&(r.docChanged||r.selection)?e=[]:t.hideOn&&(e=e.filter((e=>!t.hideOn(r,e)))),r.docChanged)){let t=[];for(let n of e){let e=r.changes.mapPos(n.pos,-1,A.TrackDel);if(null!=e){let a=Object.assign(Object.create(null),n);a.pos=e,null!=a.end&&(a.end=r.changes.mapPos(a.end)),t.push(a)}}e=t}for(let t of r.effects)t.is(n)&&(e=t.value),t.is(zo)&&(e=[]);return e},provide:e=>No.from(e)});return{active:r,extension:[r,er.define((a=>new Bo(a,e,r,n,t.hoverTime||300))),Lo]}}function Uo(e,t){let n=e.plugin(Ro);if(!n)return null;let r=n.manager.tooltips.indexOf(t);return r<0?null:n.manager.tooltipViews[r]}const zo=fe.define(),jo=z.define({combine(e){let t,n;for(let r of e)t=t||r.topContainer,n=n||r.bottomContainer;return{topContainer:t,bottomContainer:n}}}),Wo=er.fromClass(class{constructor(e){this.input=e.state.facet(Vo),this.specs=this.input.filter((e=>e)),this.panels=this.specs.map((t=>t(e)));let t=e.state.facet(jo);this.top=new Xo(e,!0,t.topContainer),this.bottom=new Xo(e,!1,t.bottomContainer),this.top.sync(this.panels.filter((e=>e.top))),this.bottom.sync(this.panels.filter((e=>!e.top)));for(let e of this.panels)e.dom.classList.add("cm-panel"),e.mount&&e.mount()}update(e){let t=e.state.facet(jo);this.top.container!=t.topContainer&&(this.top.sync([]),this.top=new Xo(e.view,!0,t.topContainer)),this.bottom.container!=t.bottomContainer&&(this.bottom.sync([]),this.bottom=new Xo(e.view,!1,t.bottomContainer)),this.top.syncClasses(),this.bottom.syncClasses();let n=e.state.facet(Vo);if(n!=this.input){let t=n.filter((e=>e)),r=[],a=[],o=[],i=[];for(let n of t){let t,s=this.specs.indexOf(n);s<0?(t=n(e.view),i.push(t)):(t=this.panels[s],t.update&&t.update(e)),r.push(t),(t.top?a:o).push(t)}this.specs=t,this.panels=r,this.top.sync(a),this.bottom.sync(o);for(let e of i)e.dom.classList.add("cm-panel"),e.mount&&e.mount()}else for(let t of this.panels)t.update&&t.update(e)}destroy(){this.top.sync([]),this.bottom.sync([])}},{provide:e=>Ya.scrollMargins.of((t=>{let n=t.plugin(e);return n&&{top:n.top.scrollMargin(),bottom:n.bottom.scrollMargin()}}))});class Xo{constructor(e,t,n){this.view=e,this.top=t,this.container=n,this.dom=void 0,this.classes="",this.panels=[],this.syncClasses()}sync(e){for(let t of this.panels)t.destroy&&e.indexOf(t)<0&&t.destroy();this.panels=e,this.syncDOM()}syncDOM(){if(0==this.panels.length)return void(this.dom&&(this.dom.remove(),this.dom=void 0));if(!this.dom){this.dom=document.createElement("div"),this.dom.className=this.top?"cm-panels cm-panels-top":"cm-panels cm-panels-bottom",this.dom.style[this.top?"top":"bottom"]="0";let e=this.container||this.view.dom;e.insertBefore(this.dom,this.top?e.firstChild:null)}let e=this.dom.firstChild;for(let t of this.panels)if(t.dom.parentNode==this.dom){for(;e!=t.dom;)e=qo(e);e=e.nextSibling}else this.dom.insertBefore(t.dom,e);for(;e;)e=qo(e)}scrollMargin(){return!this.dom||this.container?0:Math.max(0,this.top?this.dom.getBoundingClientRect().bottom-Math.max(0,this.view.scrollDOM.getBoundingClientRect().top):Math.min(innerHeight,this.view.scrollDOM.getBoundingClientRect().bottom)-this.dom.getBoundingClientRect().top)}syncClasses(){if(this.container&&this.classes!=this.view.themeClasses){for(let e of this.classes.split(" "))e&&this.container.classList.remove(e);for(let e of(this.classes=this.view.themeClasses).split(" "))e&&this.container.classList.add(e)}}}function qo(e){let t=e.nextSibling;return e.remove(),t}const Vo=z.define({enables:Wo});class Ho extends Ce{compare(e){return this==e||this.constructor==e.constructor&&this.eq(e)}eq(e){return!1}destroy(e){}}Ho.prototype.elementClass="",Ho.prototype.toDOM=void 0,Ho.prototype.mapMode=A.TrackBefore,Ho.prototype.startSide=Ho.prototype.endSide=-1,Ho.prototype.point=!0;const Zo=z.define(),Go=z.define(),Ko={class:"",renderEmptyElements:!1,elementStyle:"",markers:()=>Pe.empty,lineMarker:()=>null,widgetMarker:()=>null,lineMarkerChange:null,initialSpacer:null,updateSpacer:null,domEventHandlers:{}},Yo=z.define();function Jo(e){return[ti(),Yo.of(Object.assign(Object.assign({},Ko),e))]}const ei=z.define({combine:e=>e.some((e=>e))});function ti(e){let t=[ni];return e&&!1===e.fixed&&t.push(ei.of(!0)),t}const ni=er.fromClass(class{constructor(e){this.view=e,this.prevViewport=e.viewport,this.dom=document.createElement("div"),this.dom.className="cm-gutters",this.dom.setAttribute("aria-hidden","true"),this.dom.style.minHeight=this.view.contentHeight/this.view.scaleY+"px",this.gutters=e.state.facet(Yo).map((t=>new ii(e,t)));for(let e of this.gutters)this.dom.appendChild(e.dom);this.fixed=!e.state.facet(ei),this.fixed&&(this.dom.style.position="sticky"),this.syncGutters(!1),e.scrollDOM.insertBefore(this.dom,e.contentDOM)}update(e){if(this.updateGutters(e)){let t=this.prevViewport,n=e.view.viewport,r=Math.min(t.to,n.to)-Math.max(t.from,n.from);this.syncGutters(r<.8*(n.to-n.from))}e.geometryChanged&&(this.dom.style.minHeight=this.view.contentHeight/this.view.scaleY+"px"),this.view.state.facet(ei)!=!this.fixed&&(this.fixed=!this.fixed,this.dom.style.position=this.fixed?"sticky":""),this.prevViewport=e.view.viewport}syncGutters(e){let t=this.dom.nextSibling;e&&this.dom.remove();let n=Pe.iter(this.view.state.facet(Zo),this.view.viewport.from),r=[],a=this.gutters.map((e=>new oi(e,this.view.viewport,-this.view.documentPadding.top)));for(let e of this.view.viewportLineBlocks)if(r.length&&(r=[]),Array.isArray(e.type)){let t=!0;for(let o of e.type)if(o.type==nn.Text&&t){ai(n,r,o.from);for(let e of a)e.line(this.view,o,r);t=!1}else if(o.widget)for(let e of a)e.widget(this.view,o)}else if(e.type==nn.Text){ai(n,r,e.from);for(let t of a)t.line(this.view,e,r)}else if(e.widget)for(let t of a)t.widget(this.view,e);for(let e of a)e.finish();e&&this.view.scrollDOM.insertBefore(this.dom,t)}updateGutters(e){let t=e.startState.facet(Yo),n=e.state.facet(Yo),r=e.docChanged||e.heightChanged||e.viewportChanged||!Pe.eq(e.startState.facet(Zo),e.state.facet(Zo),e.view.viewport.from,e.view.viewport.to);if(t==n)for(let t of this.gutters)t.update(e)&&(r=!0);else{r=!0;let a=[];for(let r of n){let n=t.indexOf(r);n<0?a.push(new ii(this.view,r)):(this.gutters[n].update(e),a.push(this.gutters[n]))}for(let e of this.gutters)e.dom.remove(),a.indexOf(e)<0&&e.destroy();for(let e of a)this.dom.appendChild(e.dom);this.gutters=a}return r}destroy(){for(let e of this.gutters)e.destroy();this.dom.remove()}},{provide:e=>Ya.scrollMargins.of((t=>{let n=t.plugin(e);return n&&0!=n.gutters.length&&n.fixed?t.textDirection==gn.LTR?{left:n.dom.offsetWidth*t.scaleX}:{right:n.dom.offsetWidth*t.scaleX}:null}))});function ri(e){return Array.isArray(e)?e:[e]}function ai(e,t,n){for(;e.value&&e.from<=n;)e.from==n&&t.push(e.value),e.next()}class oi{constructor(e,t,n){this.gutter=e,this.height=n,this.i=0,this.cursor=Pe.iter(e.markers,t.from)}addElement(e,t,n){let{gutter:r}=this,a=(t.top-this.height)/e.scaleY,o=t.height/e.scaleY;if(this.i==r.elements.length){let t=new si(e,o,a,n);r.elements.push(t),r.dom.appendChild(t.dom)}else r.elements[this.i].update(e,o,a,n);this.height=t.bottom,this.i++}line(e,t,n){let r=[];ai(this.cursor,r,t.from),n.length&&(r=r.concat(n));let a=this.gutter.config.lineMarker(e,t,r);a&&r.unshift(a);let o=this.gutter;(0!=r.length||o.config.renderEmptyElements)&&this.addElement(e,t,r)}widget(e,t){let n=this.gutter.config.widgetMarker(e,t.widget,t),r=n?[n]:null;for(let n of e.state.facet(Go)){let a=n(e,t.widget,t);a&&(r||(r=[])).push(a)}r&&this.addElement(e,t,r)}finish(){let e=this.gutter;for(;e.elements.length>this.i;){let t=e.elements.pop();e.dom.removeChild(t.dom),t.destroy()}}}class ii{constructor(e,t){this.view=e,this.config=t,this.elements=[],this.spacer=null,this.dom=document.createElement("div"),this.dom.className="cm-gutter"+(this.config.class?" "+this.config.class:"");for(let n in t.domEventHandlers)this.dom.addEventListener(n,(r=>{let a,o=r.target;if(o!=this.dom&&this.dom.contains(o)){for(;o.parentNode!=this.dom;)o=o.parentNode;let e=o.getBoundingClientRect();a=(e.top+e.bottom)/2}else a=r.clientY;let i=e.lineBlockAtHeight(a-e.documentTop);t.domEventHandlers[n](e,i,r)&&r.preventDefault()}));this.markers=ri(t.markers(e)),t.initialSpacer&&(this.spacer=new si(e,0,0,[t.initialSpacer(e)]),this.dom.appendChild(this.spacer.dom),this.spacer.dom.style.cssText+="visibility: hidden; pointer-events: none")}update(e){let t=this.markers;if(this.markers=ri(this.config.markers(e.view)),this.spacer&&this.config.updateSpacer){let t=this.config.updateSpacer(this.spacer.markers[0],e);t!=this.spacer.markers[0]&&this.spacer.update(e.view,0,0,[t])}let n=e.view.viewport;return!Pe.eq(this.markers,t,n.from,n.to)||!!this.config.lineMarkerChange&&this.config.lineMarkerChange(e)}destroy(){for(let e of this.elements)e.destroy()}}class si{constructor(e,t,n,r){this.height=-1,this.above=0,this.markers=[],this.dom=document.createElement("div"),this.dom.className="cm-gutterElement",this.update(e,t,n,r)}update(e,t,n,r){this.height!=t&&(this.height=t,this.dom.style.height=t+"px"),this.above!=n&&(this.dom.style.marginTop=(this.above=n)?n+"px":""),function(e,t){if(e.length!=t.length)return!1;for(let n=0;n<e.length;n++)if(!e[n].compare(t[n]))return!1;return!0}(this.markers,r)||this.setMarkers(e,r)}setMarkers(e,t){let n="cm-gutterElement",r=this.dom.firstChild;for(let a=0,o=0;;){let i=o,s=a<t.length?t[a++]:null,l=!1;if(s){let e=s.elementClass;e&&(n+=" "+e);for(let e=o;e<this.markers.length;e++)if(this.markers[e].compare(s)){i=e,l=!0;break}}else i=this.markers.length;for(;o<i;){let e=this.markers[o++];if(e.toDOM){e.destroy(r);let t=r.nextSibling;r.remove(),r=t}}if(!s)break;s.toDOM&&(l?r=r.nextSibling:this.dom.insertBefore(s.toDOM(e),r)),l&&o++}this.dom.className=n,this.markers=t}destroy(){this.setMarkers(null,[])}}const li=z.define(),ci=z.define(),ui=z.define({combine:e=>_e(e,{formatNumber:String,domEventHandlers:{}},{domEventHandlers(e,t){let n=Object.assign({},e);for(let e in t){let r=n[e],a=t[e];n[e]=r?(e,t,n)=>r(e,t,n)||a(e,t,n):a}return n}})});class di extends Ho{constructor(e){super(),this.number=e}eq(e){return this.number==e.number}toDOM(){return document.createTextNode(this.number)}}function pi(e,t){return e.state.facet(ui).formatNumber(t,e.state)}const hi=Yo.compute([ui],(e=>({class:"cm-lineNumbers",renderEmptyElements:!1,markers:e=>e.state.facet(li),lineMarker:(e,t,n)=>n.some((e=>e.toDOM))?null:new di(pi(e,e.state.doc.lineAt(t.from).number)),widgetMarker:(e,t,n)=>{for(let r of e.state.facet(ci)){let a=r(e,t,n);if(a)return a}return null},lineMarkerChange:e=>e.startState.facet(ui)!=e.state.facet(ui),initialSpacer:e=>new di(pi(e,fi(e.state.doc.lines))),updateSpacer(e,t){let n=pi(t.view,fi(t.view.state.doc.lines));return n==e.number?e:new di(n)},domEventHandlers:e.facet(ui).domEventHandlers})));function fi(e){let t=9;for(;t<e;)t=10*t+9;return t}const mi=1024;let gi=0;class vi{constructor(e,t){this.from=e,this.to=t}}class bi{constructor(e={}){this.id=gi++,this.perNode=!!e.perNode,this.deserialize=e.deserialize||(()=>{throw new Error("This node type doesn't define a deserialize function")})}add(e){if(this.perNode)throw new RangeError("Can't add per-node props to node types");return"function"!=typeof e&&(e=Oi.match(e)),t=>{let n=e(t);return void 0===n?null:[this,n]}}}bi.closedBy=new bi({deserialize:e=>e.split(" ")}),bi.openedBy=new bi({deserialize:e=>e.split(" ")}),bi.group=new bi({deserialize:e=>e.split(" ")}),bi.isolate=new bi({deserialize:e=>{if(e&&"rtl"!=e&&"ltr"!=e&&"auto"!=e)throw new RangeError("Invalid value for isolate: "+e);return e||"auto"}}),bi.contextHash=new bi({perNode:!0}),bi.lookAhead=new bi({perNode:!0}),bi.mounted=new bi({perNode:!0});class yi{constructor(e,t,n){this.tree=e,this.overlay=t,this.parser=n}static get(e){return e&&e.props&&e.props[bi.mounted.id]}}const wi=Object.create(null);class Oi{constructor(e,t,n,r=0){this.name=e,this.props=t,this.id=n,this.flags=r}static define(e){let t=e.props&&e.props.length?Object.create(null):wi,n=(e.top?1:0)|(e.skipped?2:0)|(e.error?4:0)|(null==e.name?8:0),r=new Oi(e.name||"",t,e.id,n);if(e.props)for(let n of e.props)if(Array.isArray(n)||(n=n(r)),n){if(n[0].perNode)throw new RangeError("Can't store a per-node prop on a node type");t[n[0].id]=n[1]}return r}prop(e){return this.props[e.id]}get isTop(){return(1&this.flags)>0}get isSkipped(){return(2&this.flags)>0}get isError(){return(4&this.flags)>0}get isAnonymous(){return(8&this.flags)>0}is(e){if("string"==typeof e){if(this.name==e)return!0;let t=this.prop(bi.group);return!!t&&t.indexOf(e)>-1}return this.id==e}static match(e){let t=Object.create(null);for(let n in e)for(let r of n.split(" "))t[r]=e[n];return e=>{for(let n=e.prop(bi.group),r=-1;r<(n?n.length:0);r++){let a=t[r<0?e.name:n[r]];if(a)return a}}}}Oi.none=new Oi("",Object.create(null),0,8);class ki{constructor(e){this.types=e;for(let t=0;t<e.length;t++)if(e[t].id!=t)throw new RangeError("Node type ids should correspond to array positions when creating a node set")}extend(...e){let t=[];for(let n of this.types){let r=null;for(let t of e){let e=t(n);e&&(r||(r=Object.assign({},n.props)),r[e[0].id]=e[1])}t.push(r?new Oi(n.name,r,n.id,n.flags):n)}return new ki(t)}}const xi=new WeakMap,Ei=new WeakMap;var Si;!function(e){e[e.ExcludeBuffers=1]="ExcludeBuffers",e[e.IncludeAnonymous=2]="IncludeAnonymous",e[e.IgnoreMounts=4]="IgnoreMounts",e[e.IgnoreOverlays=8]="IgnoreOverlays"}(Si||(Si={}));class _i{constructor(e,t,n,r,a){if(this.type=e,this.children=t,this.positions=n,this.length=r,this.props=null,a&&a.length){this.props=Object.create(null);for(let[e,t]of a)this.props["number"==typeof e?e:e.id]=t}}toString(){let e=yi.get(this);if(e&&!e.overlay)return e.tree.toString();let t="";for(let e of this.children){let n=e.toString();n&&(t&&(t+=","),t+=n)}return this.type.name?(/\W/.test(this.type.name)&&!this.type.isError?JSON.stringify(this.type.name):this.type.name)+(t.length?"("+t+")":""):t}cursor(e=0){return new Fi(this.topNode,e)}cursorAt(e,t=0,n=0){let r=xi.get(this)||this.topNode,a=new Fi(r);return a.moveTo(e,t),xi.set(this,a._tree),a}get topNode(){return new Di(this,0,0,null)}resolve(e,t=0){let n=Ri(xi.get(this)||this.topNode,e,t,!1);return xi.set(this,n),n}resolveInner(e,t=0){let n=Ri(Ei.get(this)||this.topNode,e,t,!0);return Ei.set(this,n),n}resolveStack(e,t=0){return function(e,t,n){let r=e.resolveInner(t,n),a=null;for(let e=r instanceof Di?r:r.context.parent;e;e=e.parent)if(e.index<0){let o=e.parent;(a||(a=[r])).push(o.resolve(t,n)),e=o}else{let o=yi.get(e.tree);if(o&&o.overlay&&o.overlay[0].from<=t&&o.overlay[o.overlay.length-1].to>=t){let i=new Di(o.tree,o.overlay[0].from+e.from,-1,e);(a||(a=[r])).push(Ri(i,t,n,!1))}}return a?Li(a):r}(this,e,t)}iterate(e){let{enter:t,leave:n,from:r=0,to:a=this.length}=e,o=e.mode||0,i=(o&Si.IncludeAnonymous)>0;for(let e=this.cursor(o|Si.IncludeAnonymous);;){let o=!1;if(e.from<=a&&e.to>=r&&(!i&&e.type.isAnonymous||!1!==t(e))){if(e.firstChild())continue;o=!0}for(;o&&n&&(i||!e.type.isAnonymous)&&n(e),!e.nextSibling();){if(!e.parent())return;o=!0}}}prop(e){return e.perNode?this.props?this.props[e.id]:void 0:this.type.prop(e)}get propValues(){let e=[];if(this.props)for(let t in this.props)e.push([+t,this.props[t]]);return e}balance(e={}){return this.children.length<=8?this:ji(Oi.none,this.children,this.positions,0,this.children.length,0,this.length,((e,t,n)=>new _i(this.type,e,t,n,this.propValues)),e.makeTree||((e,t,n)=>new _i(Oi.none,e,t,n)))}static build(e){return function(e){var t;let{buffer:n,nodeSet:r,maxBufferLength:a=mi,reused:o=[],minRepeatType:i=r.types.length}=e,s=Array.isArray(n)?new Ci(n,n.length):n,l=r.types,c=0,u=0;function d(e,t,n,g,v,b){let{id:y,start:w,end:O,size:k}=s,x=u,E=c;for(;k<0;){if(s.next(),-1==k){let t=o[y];return n.push(t),void g.push(w-e)}if(-3==k)return void(c=y);if(-4==k)return void(u=y);throw new RangeError(`Unrecognized record size: ${k}`)}let S,_,C=l[y],T=w-e;if(O-w<=a&&(_=function(e,t){let n=s.fork(),r=0,o=0,l=0,c=n.end-a,u={size:0,start:0,skip:0};e:for(let a=n.pos-e;n.pos>a;){let e=n.size;if(n.id==t&&e>=0){u.size=r,u.start=o,u.skip=l,l+=4,r+=4,n.next();continue}let s=n.pos-e;if(e<0||s<a||n.start<c)break;let d=n.id>=i?4:0,p=n.start;for(n.next();n.pos>s;){if(n.size<0){if(-3!=n.size)break e;d+=4}else n.id>=i&&(d+=4);n.next()}o=p,r+=e,l+=d}return(t<0||r==e)&&(u.size=r,u.start=o,u.skip=l),u.size>4?u:void 0}(s.pos-t,v))){let t=new Uint16Array(_.size-_.skip),n=s.pos-_.size,a=t.length;for(;s.pos>n;)a=m(_.start,t,a);S=new Ti(t,O-_.start,r),T=_.start-e}else{let e=s.pos-k;s.next();let t=[],n=[],r=y>=i?y:-1,o=0,l=O;for(;s.pos>e;)r>=0&&s.id==r&&s.size>=0?(s.end<=l-a&&(h(t,n,w,o,s.end,l,r,x,E),o=t.length,l=s.end),s.next()):b>2500?p(w,e,t,n):d(w,e,t,n,r,b+1);if(r>=0&&o>0&&o<t.length&&h(t,n,w,o,w,l,r,x,E),t.reverse(),n.reverse(),r>-1&&o>0){let e=function(e,t){return(n,r,a)=>{let o,i,s=0,l=n.length-1;if(l>=0&&(o=n[l])instanceof _i){if(!l&&o.type==e&&o.length==a)return o;(i=o.prop(bi.lookAhead))&&(s=r[l]+o.length+i)}return f(e,n,r,a,s,t)}}(C,E);S=ji(C,t,n,0,t.length,0,O-w,e,e)}else S=f(C,t,n,O-w,x-O,E)}n.push(S),g.push(T)}function p(e,t,n,o){let i=[],l=0,c=-1;for(;s.pos>t;){let{id:e,start:t,end:n,size:r}=s;if(r>4)s.next();else{if(c>-1&&t<c)break;c<0&&(c=n-a),i.push(e,t,n),l++,s.next()}}if(l){let t=new Uint16Array(4*l),a=i[i.length-2];for(let e=i.length-3,n=0;e>=0;e-=3)t[n++]=i[e],t[n++]=i[e+1]-a,t[n++]=i[e+2]-a,t[n++]=n;n.push(new Ti(t,i[2]-a,r)),o.push(a-e)}}function h(e,t,n,a,o,i,s,l,c){let u=[],d=[];for(;e.length>a;)u.push(e.pop()),d.push(t.pop()+n-o);e.push(f(r.types[s],u,d,i-o,l-i,c)),t.push(o-n)}function f(e,t,n,r,a,o,i){if(o){let e=[bi.contextHash,o];i=i?[e].concat(i):[e]}if(a>25){let e=[bi.lookAhead,a];i=i?[e].concat(i):[e]}return new _i(e,t,n,r,i)}function m(e,t,n){let{id:r,start:a,end:o,size:l}=s;if(s.next(),l>=0&&r<i){let i=n;if(l>4){let r=s.pos-(l-4);for(;s.pos>r;)n=m(e,t,n)}t[--n]=i,t[--n]=o-e,t[--n]=a-e,t[--n]=r}else-3==l?c=r:-4==l&&(u=r);return n}let g=[],v=[];for(;s.pos>0;)d(e.start||0,e.bufferStart||0,g,v,-1,0);let b=null!==(t=e.length)&&void 0!==t?t:g.length?v[0]+g[0].length:0;return new _i(l[e.topID],g.reverse(),v.reverse(),b)}(e)}}_i.empty=new _i(Oi.none,[],[],0);class Ci{constructor(e,t){this.buffer=e,this.index=t}get id(){return this.buffer[this.index-4]}get start(){return this.buffer[this.index-3]}get end(){return this.buffer[this.index-2]}get size(){return this.buffer[this.index-1]}get pos(){return this.index}next(){this.index-=4}fork(){return new Ci(this.buffer,this.index)}}class Ti{constructor(e,t,n){this.buffer=e,this.length=t,this.set=n}get type(){return Oi.none}toString(){let e=[];for(let t=0;t<this.buffer.length;)e.push(this.childString(t)),t=this.buffer[t+3];return e.join(",")}childString(e){let t=this.buffer[e],n=this.buffer[e+3],r=this.set.types[t],a=r.name;if(/\W/.test(a)&&!r.isError&&(a=JSON.stringify(a)),n==(e+=4))return a;let o=[];for(;e<n;)o.push(this.childString(e)),e=this.buffer[e+3];return a+"("+o.join(",")+")"}findChild(e,t,n,r,a){let{buffer:o}=this,i=-1;for(let s=e;s!=t&&!(Ai(a,r,o[s+1],o[s+2])&&(i=s,n>0));s=o[s+3]);return i}slice(e,t,n){let r=this.buffer,a=new Uint16Array(t-e),o=0;for(let i=e,s=0;i<t;){a[s++]=r[i++],a[s++]=r[i++]-n;let t=a[s++]=r[i++]-n;a[s++]=r[i++]-e,o=Math.max(o,t)}return new Ti(a,o,this.set)}}function Ai(e,t,n,r){switch(e){case-2:return n<t;case-1:return r>=t&&n<t;case 0:return n<t&&r>t;case 1:return n<=t&&r>t;case 2:return r>t;case 4:return!0}}function Ri(e,t,n,r){for(var a;e.from==e.to||(n<1?e.from>=t:e.from>t)||(n>-1?e.to<=t:e.to<t);){let t=!r&&e instanceof Di&&e.index<0?null:e.parent;if(!t)return e;e=t}let o=r?0:Si.IgnoreOverlays;if(r)for(let r=e,i=r.parent;i;r=i,i=r.parent)r instanceof Di&&r.index<0&&(null===(a=i.enter(t,n,o))||void 0===a?void 0:a.from)!=r.from&&(e=i);for(;;){let r=e.enter(t,n,o);if(!r)return e;e=r}}class Pi{cursor(e=0){return new Fi(this,e)}getChild(e,t=null,n=null){let r=Mi(this,e,t,n);return r.length?r[0]:null}getChildren(e,t=null,n=null){return Mi(this,e,t,n)}resolve(e,t=0){return Ri(this,e,t,!1)}resolveInner(e,t=0){return Ri(this,e,t,!0)}matchContext(e){return $i(this.parent,e)}enterUnfinishedNodesBefore(e){let t=this.childBefore(e),n=this;for(;t;){let e=t.lastChild;if(!e||e.to!=t.to)break;e.type.isError&&e.from==e.to?(n=t,t=e.prevSibling):t=e}return n}get node(){return this}get next(){return this.parent}}class Di extends Pi{constructor(e,t,n,r){super(),this._tree=e,this.from=t,this.index=n,this._parent=r}get type(){return this._tree.type}get name(){return this._tree.type.name}get to(){return this.from+this._tree.length}nextChild(e,t,n,r,a=0){for(let o=this;;){for(let{children:i,positions:s}=o._tree,l=t>0?i.length:-1;e!=l;e+=t){let l=i[e],c=s[e]+o.from;if(Ai(r,n,c,c+l.length))if(l instanceof Ti){if(a&Si.ExcludeBuffers)continue;let i=l.findChild(0,l.buffer.length,t,n-c,r);if(i>-1)return new Ii(new Ni(o,l,e,c),null,i)}else if(a&Si.IncludeAnonymous||!l.type.isAnonymous||Qi(l)){let i;if(!(a&Si.IgnoreMounts)&&(i=yi.get(l))&&!i.overlay)return new Di(i.tree,c,e,o);let s=new Di(l,c,e,o);return a&Si.IncludeAnonymous||!s.type.isAnonymous?s:s.nextChild(t<0?l.children.length-1:0,t,n,r)}}if(a&Si.IncludeAnonymous||!o.type.isAnonymous)return null;if(e=o.index>=0?o.index+t:t<0?-1:o._parent._tree.children.length,o=o._parent,!o)return null}}get firstChild(){return this.nextChild(0,1,0,4)}get lastChild(){return this.nextChild(this._tree.children.length-1,-1,0,4)}childAfter(e){return this.nextChild(0,1,e,2)}childBefore(e){return this.nextChild(this._tree.children.length-1,-1,e,-2)}enter(e,t,n=0){let r;if(!(n&Si.IgnoreOverlays)&&(r=yi.get(this._tree))&&r.overlay){let n=e-this.from;for(let{from:e,to:a}of r.overlay)if((t>0?e<=n:e<n)&&(t<0?a>=n:a>n))return new Di(r.tree,r.overlay[0].from+this.from,-1,this)}return this.nextChild(0,1,e,t,n)}nextSignificantParent(){let e=this;for(;e.type.isAnonymous&&e._parent;)e=e._parent;return e}get parent(){return this._parent?this._parent.nextSignificantParent():null}get nextSibling(){return this._parent&&this.index>=0?this._parent.nextChild(this.index+1,1,0,4):null}get prevSibling(){return this._parent&&this.index>=0?this._parent.nextChild(this.index-1,-1,0,4):null}get tree(){return this._tree}toTree(){return this._tree}toString(){return this._tree.toString()}}function Mi(e,t,n,r){let a=e.cursor(),o=[];if(!a.firstChild())return o;if(null!=n)for(let e=!1;!e;)if(e=a.type.is(n),!a.nextSibling())return o;for(;;){if(null!=r&&a.type.is(r))return o;if(a.type.is(t)&&o.push(a.node),!a.nextSibling())return null==r?o:[]}}function $i(e,t,n=t.length-1){for(let r=e;n>=0;r=r.parent){if(!r)return!1;if(!r.type.isAnonymous){if(t[n]&&t[n]!=r.name)return!1;n--}}return!0}class Ni{constructor(e,t,n,r){this.parent=e,this.buffer=t,this.index=n,this.start=r}}class Ii extends Pi{get name(){return this.type.name}get from(){return this.context.start+this.context.buffer.buffer[this.index+1]}get to(){return this.context.start+this.context.buffer.buffer[this.index+2]}constructor(e,t,n){super(),this.context=e,this._parent=t,this.index=n,this.type=e.buffer.set.types[e.buffer.buffer[n]]}child(e,t,n){let{buffer:r}=this.context,a=r.findChild(this.index+4,r.buffer[this.index+3],e,t-this.context.start,n);return a<0?null:new Ii(this.context,this,a)}get firstChild(){return this.child(1,0,4)}get lastChild(){return this.child(-1,0,4)}childAfter(e){return this.child(1,e,2)}childBefore(e){return this.child(-1,e,-2)}enter(e,t,n=0){if(n&Si.ExcludeBuffers)return null;let{buffer:r}=this.context,a=r.findChild(this.index+4,r.buffer[this.index+3],t>0?1:-1,e-this.context.start,t);return a<0?null:new Ii(this.context,this,a)}get parent(){return this._parent||this.context.parent.nextSignificantParent()}externalSibling(e){return this._parent?null:this.context.parent.nextChild(this.context.index+e,e,0,4)}get nextSibling(){let{buffer:e}=this.context,t=e.buffer[this.index+3];return t<(this._parent?e.buffer[this._parent.index+3]:e.buffer.length)?new Ii(this.context,this._parent,t):this.externalSibling(1)}get prevSibling(){let{buffer:e}=this.context,t=this._parent?this._parent.index+4:0;return this.index==t?this.externalSibling(-1):new Ii(this.context,this._parent,e.findChild(t,this.index,-1,0,4))}get tree(){return null}toTree(){let e=[],t=[],{buffer:n}=this.context,r=this.index+4,a=n.buffer[this.index+3];if(a>r){let o=n.buffer[this.index+1];e.push(n.slice(r,a,o)),t.push(0)}return new _i(this.type,e,t,this.to-this.from)}toString(){return this.context.buffer.childString(this.index)}}function Li(e){if(!e.length)return null;let t=0,n=e[0];for(let r=1;r<e.length;r++){let a=e[r];(a.from>n.from||a.to<n.to)&&(n=a,t=r)}let r=n instanceof Di&&n.index<0?null:n.parent,a=e.slice();return r?a[t]=r:a.splice(t,1),new Bi(a,n)}class Bi{constructor(e,t){this.heads=e,this.node=t}get next(){return Li(this.heads)}}class Fi{get name(){return this.type.name}constructor(e,t=0){if(this.mode=t,this.buffer=null,this.stack=[],this.index=0,this.bufferNode=null,e instanceof Di)this.yieldNode(e);else{this._tree=e.context.parent,this.buffer=e.context;for(let t=e._parent;t;t=t._parent)this.stack.unshift(t.index);this.bufferNode=e,this.yieldBuf(e.index)}}yieldNode(e){return!!e&&(this._tree=e,this.type=e.type,this.from=e.from,this.to=e.to,!0)}yieldBuf(e,t){this.index=e;let{start:n,buffer:r}=this.buffer;return this.type=t||r.set.types[r.buffer[e]],this.from=n+r.buffer[e+1],this.to=n+r.buffer[e+2],!0}yield(e){return!!e&&(e instanceof Di?(this.buffer=null,this.yieldNode(e)):(this.buffer=e.context,this.yieldBuf(e.index,e.type)))}toString(){return this.buffer?this.buffer.buffer.childString(this.index):this._tree.toString()}enterChild(e,t,n){if(!this.buffer)return this.yield(this._tree.nextChild(e<0?this._tree._tree.children.length-1:0,e,t,n,this.mode));let{buffer:r}=this.buffer,a=r.findChild(this.index+4,r.buffer[this.index+3],e,t-this.buffer.start,n);return!(a<0)&&(this.stack.push(this.index),this.yieldBuf(a))}firstChild(){return this.enterChild(1,0,4)}lastChild(){return this.enterChild(-1,0,4)}childAfter(e){return this.enterChild(1,e,2)}childBefore(e){return this.enterChild(-1,e,-2)}enter(e,t,n=this.mode){return this.buffer?!(n&Si.ExcludeBuffers)&&this.enterChild(1,e,t):this.yield(this._tree.enter(e,t,n))}parent(){if(!this.buffer)return this.yieldNode(this.mode&Si.IncludeAnonymous?this._tree._parent:this._tree.parent);if(this.stack.length)return this.yieldBuf(this.stack.pop());let e=this.mode&Si.IncludeAnonymous?this.buffer.parent:this.buffer.parent.nextSignificantParent();return this.buffer=null,this.yieldNode(e)}sibling(e){if(!this.buffer)return!!this._tree._parent&&this.yield(this._tree.index<0?null:this._tree._parent.nextChild(this._tree.index+e,e,0,4,this.mode));let{buffer:t}=this.buffer,n=this.stack.length-1;if(e<0){let e=n<0?0:this.stack[n]+4;if(this.index!=e)return this.yieldBuf(t.findChild(e,this.index,-1,0,4))}else{let e=t.buffer[this.index+3];if(e<(n<0?t.buffer.length:t.buffer[this.stack[n]+3]))return this.yieldBuf(e)}return n<0&&this.yield(this.buffer.parent.nextChild(this.buffer.index+e,e,0,4,this.mode))}nextSibling(){return this.sibling(1)}prevSibling(){return this.sibling(-1)}atLastNode(e){let t,n,{buffer:r}=this;if(r){if(e>0){if(this.index<r.buffer.buffer.length)return!1}else for(let e=0;e<this.index;e++)if(r.buffer.buffer[e+3]<this.index)return!1;({index:t,parent:n}=r)}else({index:t,_parent:n}=this._tree);for(;n;({index:t,_parent:n}=n))if(t>-1)for(let r=t+e,a=e<0?-1:n._tree.children.length;r!=a;r+=e){let e=n._tree.children[r];if(this.mode&Si.IncludeAnonymous||e instanceof Ti||!e.type.isAnonymous||Qi(e))return!1}return!0}move(e,t){if(t&&this.enterChild(e,0,4))return!0;for(;;){if(this.sibling(e))return!0;if(this.atLastNode(e)||!this.parent())return!1}}next(e=!0){return this.move(1,e)}prev(e=!0){return this.move(-1,e)}moveTo(e,t=0){for(;(this.from==this.to||(t<1?this.from>=e:this.from>e)||(t>-1?this.to<=e:this.to<e))&&this.parent(););for(;this.enterChild(1,e,t););return this}get node(){if(!this.buffer)return this._tree;let e=this.bufferNode,t=null,n=0;if(e&&e.context==this.buffer)e:for(let r=this.index,a=this.stack.length;a>=0;){for(let o=e;o;o=o._parent)if(o.index==r){if(r==this.index)return o;t=o,n=a+1;break e}r=this.stack[--a]}for(let e=n;e<this.stack.length;e++)t=new Ii(this.buffer,t,this.stack[e]);return this.bufferNode=new Ii(this.buffer,t,this.index)}get tree(){return this.buffer?null:this._tree._tree}iterate(e,t){for(let n=0;;){let r=!1;if(this.type.isAnonymous||!1!==e(this)){if(this.firstChild()){n++;continue}this.type.isAnonymous||(r=!0)}for(;;){if(r&&t&&t(this),r=this.type.isAnonymous,!n)return;if(this.nextSibling())break;this.parent(),n--,r=!0}}}matchContext(e){if(!this.buffer)return $i(this.node.parent,e);let{buffer:t}=this.buffer,{types:n}=t.set;for(let r=e.length-1,a=this.stack.length-1;r>=0;a--){if(a<0)return $i(this._tree,e,r);let o=n[t.buffer[this.stack[a]]];if(!o.isAnonymous){if(e[r]&&e[r]!=o.name)return!1;r--}}return!0}}function Qi(e){return e.children.some((e=>e instanceof Ti||!e.type.isAnonymous||Qi(e)))}const Ui=new WeakMap;function zi(e,t){if(!e.isAnonymous||t instanceof Ti||t.type!=e)return 1;let n=Ui.get(t);if(null==n){n=1;for(let r of t.children){if(r.type!=e||!(r instanceof _i)){n=1;break}n+=zi(e,r)}Ui.set(t,n)}return n}function ji(e,t,n,r,a,o,i,s,l){let c=0;for(let n=r;n<a;n++)c+=zi(e,t[n]);let u=Math.ceil(1.5*c/8),d=[],p=[];return function t(n,r,a,i,s){for(let c=a;c<i;){let a=c,h=r[c],f=zi(e,n[c]);for(c++;c<i;c++){let t=zi(e,n[c]);if(f+t>=u)break;f+=t}if(c==a+1){if(f>u){let e=n[a];t(e.children,e.positions,0,e.children.length,r[a]+s);continue}d.push(n[a])}else{let t=r[c-1]+n[c-1].length-h;d.push(ji(e,n,r,a,c,h,t,null,l))}p.push(h+s-o)}}(t,n,r,a,0),(s||l)(d,p,i)}class Wi{constructor(){this.map=new WeakMap}setBuffer(e,t,n){let r=this.map.get(e);r||this.map.set(e,r=new Map),r.set(t,n)}getBuffer(e,t){let n=this.map.get(e);return n&&n.get(t)}set(e,t){e instanceof Ii?this.setBuffer(e.context.buffer,e.index,t):e instanceof Di&&this.map.set(e.tree,t)}get(e){return e instanceof Ii?this.getBuffer(e.context.buffer,e.index):e instanceof Di?this.map.get(e.tree):void 0}cursorSet(e,t){e.buffer?this.setBuffer(e.buffer.buffer,e.index,t):this.map.set(e.tree,t)}cursorGet(e){return e.buffer?this.getBuffer(e.buffer.buffer,e.index):this.map.get(e.tree)}}class Xi{constructor(e,t,n,r,a=!1,o=!1){this.from=e,this.to=t,this.tree=n,this.offset=r,this.open=(a?1:0)|(o?2:0)}get openStart(){return(1&this.open)>0}get openEnd(){return(2&this.open)>0}static addTree(e,t=[],n=!1){let r=[new Xi(0,e.length,e,0,!1,n)];for(let n of t)n.to>e.length&&r.push(n);return r}static applyChanges(e,t,n=128){if(!t.length)return e;let r=[],a=1,o=e.length?e[0]:null;for(let i=0,s=0,l=0;;i++){let c=i<t.length?t[i]:null,u=c?c.fromA:1e9;if(u-s>=n)for(;o&&o.from<u;){let t=o;if(s>=t.from||u<=t.to||l){let e=Math.max(t.from,s)-l,n=Math.min(t.to,u)-l;t=e>=n?null:new Xi(e,n,t.tree,t.offset+l,i>0,!!c)}if(t&&r.push(t),o.to>u)break;o=a<e.length?e[a++]:null}if(!c)break;s=c.toA,l=c.toA-c.toB}return r}}class qi{startParse(e,t,n){return"string"==typeof e&&(e=new Vi(e)),n=n?n.length?n.map((e=>new vi(e.from,e.to))):[new vi(0,0)]:[new vi(0,e.length)],this.createParse(e,t||[],n)}parse(e,t,n){let r=this.startParse(e,t,n);for(;;){let e=r.advance();if(e)return e}}}class Vi{constructor(e){this.string=e}get length(){return this.string.length}chunk(e){return this.string.slice(e)}get lineChunks(){return!1}read(e,t){return this.string.slice(e,t)}}class Hi{constructor(e,t,n,r,a){this.parser=e,this.parse=t,this.overlay=n,this.target=r,this.from=a}}function Zi(e){if(!e.length||e.some((e=>e.from>=e.to)))throw new RangeError("Invalid inner parse ranges given: "+JSON.stringify(e))}class Gi{constructor(e,t,n,r,a,o,i){this.parser=e,this.predicate=t,this.mounts=n,this.index=r,this.start=a,this.target=o,this.prev=i,this.depth=0,this.ranges=[]}}const Ki=new bi({perNode:!0});class Yi{constructor(e,t,n,r,a){this.nest=t,this.input=n,this.fragments=r,this.ranges=a,this.inner=[],this.innerDone=0,this.baseTree=null,this.stoppedAt=null,this.baseParse=e}advance(){if(this.baseParse){let e=this.baseParse.advance();if(!e)return null;if(this.baseParse=null,this.baseTree=e,this.startInner(),null!=this.stoppedAt)for(let e of this.inner)e.parse.stopAt(this.stoppedAt)}if(this.innerDone==this.inner.length){let e=this.baseTree;return null!=this.stoppedAt&&(e=new _i(e.type,e.children,e.positions,e.length,e.propValues.concat([[Ki,this.stoppedAt]]))),e}let e=this.inner[this.innerDone],t=e.parse.advance();if(t){this.innerDone++;let n=Object.assign(Object.create(null),e.target.props);n[bi.mounted.id]=new yi(t,e.overlay,e.parser),e.target.props=n}return null}get parsedPos(){if(this.baseParse)return 0;let e=this.input.length;for(let t=this.innerDone;t<this.inner.length;t++)this.inner[t].from<e&&(e=Math.min(e,this.inner[t].parse.parsedPos));return e}stopAt(e){if(this.stoppedAt=e,this.baseParse)this.baseParse.stopAt(e);else for(let t=this.innerDone;t<this.inner.length;t++)this.inner[t].parse.stopAt(e)}startInner(){let e=new rs(this.fragments),t=null,n=null,r=new Fi(new Di(this.baseTree,this.ranges[0].from,0,null),Si.IncludeAnonymous|Si.IgnoreMounts);e:for(let a,o;;){let i,s=!0;if(null!=this.stoppedAt&&r.from>=this.stoppedAt)s=!1;else if(e.hasNode(r)){if(t){let e=t.mounts.find((e=>e.frag.from<=r.from&&e.frag.to>=r.to&&e.mount.overlay));if(e)for(let n of e.mount.overlay){let a=n.from+e.pos,o=n.to+e.pos;a>=r.from&&o<=r.to&&!t.ranges.some((e=>e.from<o&&e.to>a))&&t.ranges.push({from:a,to:o})}}s=!1}else if(n&&(o=Ji(n.ranges,r.from,r.to)))s=2!=o;else if(!r.type.isAnonymous&&(a=this.nest(r,this.input))&&(r.from<r.to||!a.overlay)){r.tree||ts(r);let o=e.findMounts(r.from,a.parser);if("function"==typeof a.overlay)t=new Gi(a.parser,a.overlay,o,this.inner.length,r.from,r.tree,t);else{let e=as(this.ranges,a.overlay||(r.from<r.to?[new vi(r.from,r.to)]:[]));e.length&&Zi(e),!e.length&&a.overlay||this.inner.push(new Hi(a.parser,e.length?a.parser.startParse(this.input,is(o,e),e):a.parser.startParse(""),a.overlay?a.overlay.map((e=>new vi(e.from-r.from,e.to-r.from))):null,r.tree,e.length?e[0].from:r.from)),a.overlay?e.length&&(n={ranges:e,depth:0,prev:n}):s=!1}}else if(t&&(i=t.predicate(r))&&(!0===i&&(i=new vi(r.from,r.to)),i.from<i.to)){let e=t.ranges.length-1;e>=0&&t.ranges[e].to==i.from?t.ranges[e]={from:t.ranges[e].from,to:i.to}:t.ranges.push(i)}if(s&&r.firstChild())t&&t.depth++,n&&n.depth++;else for(;!r.nextSibling();){if(!r.parent())break e;if(t&&! --t.depth){let e=as(this.ranges,t.ranges);e.length&&(Zi(e),this.inner.splice(t.index,0,new Hi(t.parser,t.parser.startParse(this.input,is(t.mounts,e),e),t.ranges.map((e=>new vi(e.from-t.start,e.to-t.start))),t.target,e[0].from))),t=t.prev}n&&! --n.depth&&(n=n.prev)}}}}function Ji(e,t,n){for(let r of e){if(r.from>=n)break;if(r.to>t)return r.from<=t&&r.to>=n?2:1}return 0}function es(e,t,n,r,a,o){if(t<n){let i=e.buffer[t+1];r.push(e.slice(t,n,i)),a.push(i-o)}}function ts(e){let{node:t}=e,n=[],r=t.context.buffer;do{n.push(e.index),e.parent()}while(!e.tree);let a=e.tree,o=a.children.indexOf(r),i=a.children[o],s=i.buffer,l=[o];a.children[o]=function e(r,a,o,c,u,d){let p=n[d],h=[],f=[];es(i,r,p,h,f,c);let m=s[p+1],g=s[p+2];l.push(h.length);let v=d?e(p+4,s[p+3],i.set.types[s[p]],m,g-m,d-1):t.toTree();return h.push(v),f.push(m-c),es(i,s[p+3],a,h,f,c),new _i(o,h,f,u)}(0,s.length,Oi.none,0,i.length,n.length-1);for(let t of l){let n=e.tree.children[t],r=e.tree.positions[t];e.yield(new Di(n,r+e.from,t,e._tree))}}class ns{constructor(e,t){this.offset=t,this.done=!1,this.cursor=e.cursor(Si.IncludeAnonymous|Si.IgnoreMounts)}moveTo(e){let{cursor:t}=this,n=e-this.offset;for(;!this.done&&t.from<n;)t.to>=e&&t.enter(n,1,Si.IgnoreOverlays|Si.ExcludeBuffers)||t.next(!1)||(this.done=!0)}hasNode(e){if(this.moveTo(e.from),!this.done&&this.cursor.from+this.offset==e.from&&this.cursor.tree)for(let t=this.cursor.tree;;){if(t==e.tree)return!0;if(!(t.children.length&&0==t.positions[0]&&t.children[0]instanceof _i))break;t=t.children[0]}return!1}}class rs{constructor(e){var t;if(this.fragments=e,this.curTo=0,this.fragI=0,e.length){let n=this.curFrag=e[0];this.curTo=null!==(t=n.tree.prop(Ki))&&void 0!==t?t:n.to,this.inner=new ns(n.tree,-n.offset)}else this.curFrag=this.inner=null}hasNode(e){for(;this.curFrag&&e.from>=this.curTo;)this.nextFrag();return this.curFrag&&this.curFrag.from<=e.from&&this.curTo>=e.to&&this.inner.hasNode(e)}nextFrag(){var e;if(this.fragI++,this.fragI==this.fragments.length)this.curFrag=this.inner=null;else{let t=this.curFrag=this.fragments[this.fragI];this.curTo=null!==(e=t.tree.prop(Ki))&&void 0!==e?e:t.to,this.inner=new ns(t.tree,-t.offset)}}findMounts(e,t){var n;let r=[];if(this.inner){this.inner.cursor.moveTo(e,1);for(let e=this.inner.cursor.node;e;e=e.parent){let a=null===(n=e.tree)||void 0===n?void 0:n.prop(bi.mounted);if(a&&a.parser==t)for(let t=this.fragI;t<this.fragments.length;t++){let n=this.fragments[t];if(n.from>=e.to)break;n.tree==this.curFrag.tree&&r.push({frag:n,pos:e.from-n.offset,mount:a})}}}return r}}function as(e,t){let n=null,r=t;for(let a=1,o=0;a<e.length;a++){let i=e[a-1].to,s=e[a].from;for(;o<r.length;o++){let e=r[o];if(e.from>=s)break;e.to<=i||(n||(r=n=t.slice()),e.from<i?(n[o]=new vi(e.from,i),e.to>s&&n.splice(o+1,0,new vi(s,e.to))):e.to>s?n[o--]=new vi(s,e.to):n.splice(o--,1))}}return r}function os(e,t,n,r){let a=0,o=0,i=!1,s=!1,l=-1e9,c=[];for(;;){let u=a==e.length?1e9:i?e[a].to:e[a].from,d=o==t.length?1e9:s?t[o].to:t[o].from;if(i!=s){let e=Math.max(l,n),t=Math.min(u,d,r);e<t&&c.push(new vi(e,t))}if(l=Math.min(u,d),1e9==l)break;u==l&&(i?(i=!1,a++):i=!0),d==l&&(s?(s=!1,o++):s=!0)}return c}function is(e,t){let n=[];for(let{pos:r,mount:a,frag:o}of e){let e=r+(a.overlay?a.overlay[0].from:0),i=e+a.tree.length,s=Math.max(o.from,e),l=Math.min(o.to,i);if(a.overlay){let i=os(t,a.overlay.map((e=>new vi(e.from+r,e.to+r))),s,l);for(let t=0,r=s;;t++){let s=t==i.length,c=s?l:i[t].from;if(c>r&&n.push(new Xi(r,c,a.tree,-e,o.from>=r||o.openStart,o.to<=c||o.openEnd)),s)break;r=i[t].to}}else n.push(new Xi(s,l,a.tree,-e,o.from>=e||o.openStart,o.to<=i||o.openEnd))}return n}let ss=0;class ls{constructor(e,t,n,r){this.name=e,this.set=t,this.base=n,this.modified=r,this.id=ss++}toString(){let{name:e}=this;for(let t of this.modified)t.name&&(e=`${t.name}(${e})`);return e}static define(e,t){let n="string"==typeof e?e:"?";if(e instanceof ls&&(t=e),null==t?void 0:t.base)throw new Error("Can not derive from a modified tag");let r=new ls(n,[],null,[]);if(r.set.push(r),t)for(let e of t.set)r.set.push(e);return r}static defineModifier(e){let t=new us(e);return e=>e.modified.indexOf(t)>-1?e:us.get(e.base||e,e.modified.concat(t).sort(((e,t)=>e.id-t.id)))}}let cs=0;class us{constructor(e){this.name=e,this.instances=[],this.id=cs++}static get(e,t){if(!t.length)return e;let n=t[0].instances.find((n=>{return n.base==e&&(r=t,a=n.modified,r.length==a.length&&r.every(((e,t)=>e==a[t])));var r,a}));if(n)return n;let r=[],a=new ls(e.name,r,e,t);for(let e of t)e.instances.push(a);let o=function(e){let t=[[]];for(let n=0;n<e.length;n++)for(let r=0,a=t.length;r<a;r++)t.push(t[r].concat(e[n]));return t.sort(((e,t)=>t.length-e.length))}(t);for(let t of e.set)if(!t.modified.length)for(let e of o)r.push(us.get(t,e));return a}}function ds(e){let t=Object.create(null);for(let n in e){let r=e[n];Array.isArray(r)||(r=[r]);for(let e of n.split(" "))if(e){let n=[],a=2,o=e;for(let t=0;;){if("..."==o&&t>0&&t+3==e.length){a=1;break}let r=/^"(?:[^"\\]|\\.)*?"|[^\/!]+/.exec(o);if(!r)throw new RangeError("Invalid path: "+e);if(n.push("*"==r[0]?"":'"'==r[0][0]?JSON.parse(r[0]):r[0]),t+=r[0].length,t==e.length)break;let i=e[t++];if(t==e.length&&"!"==i){a=0;break}if("/"!=i)throw new RangeError("Invalid path: "+e);o=e.slice(t)}let i=n.length-1,s=n[i];if(!s)throw new RangeError("Invalid path: "+e);let l=new hs(r,a,i>0?n.slice(0,i):null);t[s]=l.sort(t[s])}}return ps.add(t)}const ps=new bi;class hs{constructor(e,t,n,r){this.tags=e,this.mode=t,this.context=n,this.next=r}get opaque(){return 0==this.mode}get inherit(){return 1==this.mode}sort(e){return!e||e.depth<this.depth?(this.next=e,this):(e.next=this.sort(e.next),e)}get depth(){return this.context?this.context.length:0}}function fs(e,t){let n=Object.create(null);for(let t of e)if(Array.isArray(t.tag))for(let e of t.tag)n[e.id]=t.class;else n[t.tag.id]=t.class;let{scope:r,all:a=null}=t||{};return{style:e=>{let t=a;for(let r of e)for(let e of r.set){let r=n[e.id];if(r){t=t?t+" "+r:r;break}}return t},scope:r}}function ms(e,t,n,r=0,a=e.length){let o=new gs(r,Array.isArray(t)?t:[t],n);o.highlightRange(e.cursor(),r,a,"",o.highlighters),o.flush(a)}hs.empty=new hs([],2,null);class gs{constructor(e,t,n){this.at=e,this.highlighters=t,this.span=n,this.class=""}startSpan(e,t){t!=this.class&&(this.flush(e),e>this.at&&(this.at=e),this.class=t)}flush(e){e>this.at&&this.class&&this.span(this.at,e,this.class)}highlightRange(e,t,n,r,a){let{type:o,from:i,to:s}=e;if(i>=n||s<=t)return;o.isTop&&(a=this.highlighters.filter((e=>!e.scope||e.scope(o))));let l=r,c=function(e){let t=e.type.prop(ps);for(;t&&t.context&&!e.matchContext(t.context);)t=t.next;return t||null}(e)||hs.empty,u=function(e,t){let n=null;for(let r of e){let e=r.style(t);e&&(n=n?n+" "+e:e)}return n}(a,c.tags);if(u&&(l&&(l+=" "),l+=u,1==c.mode&&(r+=(r?" ":"")+u)),this.startSpan(Math.max(t,i),l),c.opaque)return;let d=e.tree&&e.tree.prop(bi.mounted);if(d&&d.overlay){let o=e.node.enter(d.overlay[0].from+i,1),c=this.highlighters.filter((e=>!e.scope||e.scope(d.tree.type))),u=e.firstChild();for(let p=0,h=i;;p++){let f=p<d.overlay.length?d.overlay[p]:null,m=f?f.from+i:s,g=Math.max(t,h),v=Math.min(n,m);if(g<v&&u)for(;e.from<v&&(this.highlightRange(e,g,v,r,a),this.startSpan(Math.min(v,e.to),l),!(e.to>=m)&&e.nextSibling()););if(!f||m>n)break;h=f.to+i,h>t&&(this.highlightRange(o.cursor(),Math.max(t,f.from+i),Math.min(n,h),"",c),this.startSpan(Math.min(n,h),l))}u&&e.parent()}else if(e.firstChild()){d&&(r="");do{if(!(e.to<=t)){if(e.from>=n)break;this.highlightRange(e,t,n,r,a),this.startSpan(Math.min(n,e.to),l)}}while(e.nextSibling());e.parent()}}}const vs=ls.define,bs=vs(),ys=vs(),ws=vs(ys),Os=vs(ys),ks=vs(),xs=vs(ks),Es=vs(ks),Ss=vs(),_s=vs(Ss),Cs=vs(),Ts=vs(),As=vs(),Rs=vs(As),Ps=vs(),Ds={comment:bs,lineComment:vs(bs),blockComment:vs(bs),docComment:vs(bs),name:ys,variableName:vs(ys),typeName:ws,tagName:vs(ws),propertyName:Os,attributeName:vs(Os),className:vs(ys),labelName:vs(ys),namespace:vs(ys),macroName:vs(ys),literal:ks,string:xs,docString:vs(xs),character:vs(xs),attributeValue:vs(xs),number:Es,integer:vs(Es),float:vs(Es),bool:vs(ks),regexp:vs(ks),escape:vs(ks),color:vs(ks),url:vs(ks),keyword:Cs,self:vs(Cs),null:vs(Cs),atom:vs(Cs),unit:vs(Cs),modifier:vs(Cs),operatorKeyword:vs(Cs),controlKeyword:vs(Cs),definitionKeyword:vs(Cs),moduleKeyword:vs(Cs),operator:Ts,derefOperator:vs(Ts),arithmeticOperator:vs(Ts),logicOperator:vs(Ts),bitwiseOperator:vs(Ts),compareOperator:vs(Ts),updateOperator:vs(Ts),definitionOperator:vs(Ts),typeOperator:vs(Ts),controlOperator:vs(Ts),punctuation:As,separator:vs(As),bracket:Rs,angleBracket:vs(Rs),squareBracket:vs(Rs),paren:vs(Rs),brace:vs(Rs),content:Ss,heading:_s,heading1:vs(_s),heading2:vs(_s),heading3:vs(_s),heading4:vs(_s),heading5:vs(_s),heading6:vs(_s),contentSeparator:vs(Ss),list:vs(Ss),quote:vs(Ss),emphasis:vs(Ss),strong:vs(Ss),link:vs(Ss),monospace:vs(Ss),strikethrough:vs(Ss),inserted:vs(),deleted:vs(),changed:vs(),invalid:vs(),meta:Ps,documentMeta:vs(Ps),annotation:vs(Ps),processingInstruction:vs(Ps),definition:ls.defineModifier("definition"),constant:ls.defineModifier("constant"),function:ls.defineModifier("function"),standard:ls.defineModifier("standard"),local:ls.defineModifier("local"),special:ls.defineModifier("special")};for(let e in Ds){let t=Ds[e];t instanceof ls&&(t.name=e)}var Ms;fs([{tag:Ds.link,class:"tok-link"},{tag:Ds.heading,class:"tok-heading"},{tag:Ds.emphasis,class:"tok-emphasis"},{tag:Ds.strong,class:"tok-strong"},{tag:Ds.keyword,class:"tok-keyword"},{tag:Ds.atom,class:"tok-atom"},{tag:Ds.bool,class:"tok-bool"},{tag:Ds.url,class:"tok-url"},{tag:Ds.labelName,class:"tok-labelName"},{tag:Ds.inserted,class:"tok-inserted"},{tag:Ds.deleted,class:"tok-deleted"},{tag:Ds.literal,class:"tok-literal"},{tag:Ds.string,class:"tok-string"},{tag:Ds.number,class:"tok-number"},{tag:[Ds.regexp,Ds.escape,Ds.special(Ds.string)],class:"tok-string2"},{tag:Ds.variableName,class:"tok-variableName"},{tag:Ds.local(Ds.variableName),class:"tok-variableName tok-local"},{tag:Ds.definition(Ds.variableName),class:"tok-variableName tok-definition"},{tag:Ds.special(Ds.variableName),class:"tok-variableName2"},{tag:Ds.definition(Ds.propertyName),class:"tok-propertyName tok-definition"},{tag:Ds.typeName,class:"tok-typeName"},{tag:Ds.namespace,class:"tok-namespace"},{tag:Ds.className,class:"tok-className"},{tag:Ds.macroName,class:"tok-macroName"},{tag:Ds.propertyName,class:"tok-propertyName"},{tag:Ds.operator,class:"tok-operator"},{tag:Ds.comment,class:"tok-comment"},{tag:Ds.meta,class:"tok-meta"},{tag:Ds.invalid,class:"tok-invalid"},{tag:Ds.punctuation,class:"tok-punctuation"}]);const $s=new bi;function Ns(e){return z.define({combine:e?t=>t.concat(e):void 0})}const Is=new bi;class Ls{constructor(e,t,n=[],r=""){this.data=e,this.name=r,Se.prototype.hasOwnProperty("tree")||Object.defineProperty(Se.prototype,"tree",{get(){return Qs(this)}}),this.parser=t,this.extension=[Zs.of(this),Se.languageData.of(((e,t,n)=>{let r=Bs(e,t,n),a=r.type.prop($s);if(!a)return[];let o=e.facet(a),i=r.type.prop(Is);if(i){let a=r.resolve(t-r.from,n);for(let t of i)if(t.test(a,e)){let n=e.facet(t.facet);return"replace"==t.type?n:n.concat(o)}}return o}))].concat(n)}isActiveAt(e,t,n=-1){return Bs(e,t,n).type.prop($s)==this.data}findRegions(e){let t=e.facet(Zs);if((null==t?void 0:t.data)==this.data)return[{from:0,to:e.doc.length}];if(!t||!t.allowsNesting)return[];let n=[],r=(e,t)=>{if(e.prop($s)==this.data)return void n.push({from:t,to:t+e.length});let a=e.prop(bi.mounted);if(a){if(a.tree.prop($s)==this.data){if(a.overlay)for(let e of a.overlay)n.push({from:e.from+t,to:e.to+t});else n.push({from:t,to:t+e.length});return}if(a.overlay){let e=n.length;if(r(a.tree,a.overlay[0].from+t),n.length>e)return}}for(let n=0;n<e.children.length;n++){let a=e.children[n];a instanceof _i&&r(a,e.positions[n]+t)}};return r(Qs(e),0),n}get allowsNesting(){return!0}}function Bs(e,t,n){let r=e.facet(Zs),a=Qs(e).topNode;if(!r||r.allowsNesting)for(let e=a;e;e=e.enter(t,n,Si.ExcludeBuffers))e.type.isTop&&(a=e);return a}Ls.setState=fe.define();class Fs extends Ls{constructor(e,t,n){super(e,t,[],n),this.parser=t}static define(e){let t=Ns(e.languageData);return new Fs(t,e.parser.configure({props:[$s.add((e=>e.isTop?t:void 0))]}),e.name)}configure(e,t){return new Fs(this.data,this.parser.configure(e),t||this.name)}get allowsNesting(){return this.parser.hasWrappers()}}function Qs(e){let t=e.field(Ls.state,!1);return t?t.tree:_i.empty}class Us{constructor(e){this.doc=e,this.cursorPos=0,this.string="",this.cursor=e.iter()}get length(){return this.doc.length}syncTo(e){return this.string=this.cursor.next(e-this.cursorPos).value,this.cursorPos=e+this.string.length,this.cursorPos-this.string.length}chunk(e){return this.syncTo(e),this.string}get lineChunks(){return!0}read(e,t){let n=this.cursorPos-this.string.length;return e<n||t>=this.cursorPos?this.doc.sliceString(e,t):this.string.slice(e-n,t-n)}}let zs=null;class js{constructor(e,t,n=[],r,a,o,i,s){this.parser=e,this.state=t,this.fragments=n,this.tree=r,this.treeLen=a,this.viewport=o,this.skipped=i,this.scheduleOn=s,this.parse=null,this.tempSkipped=[]}static create(e,t,n){return new js(e,t,[],_i.empty,0,n,[],null)}startParse(){return this.parser.startParse(new Us(this.state.doc),this.fragments)}work(e,t){return null!=t&&t>=this.state.doc.length&&(t=void 0),this.tree!=_i.empty&&this.isDone(null!=t?t:this.state.doc.length)?(this.takeTree(),!0):this.withContext((()=>{var n;if("number"==typeof e){let t=Date.now()+e;e=()=>Date.now()>t}for(this.parse||(this.parse=this.startParse()),null!=t&&(null==this.parse.stoppedAt||this.parse.stoppedAt>t)&&t<this.state.doc.length&&this.parse.stopAt(t);;){let r=this.parse.advance();if(r){if(this.fragments=this.withoutTempSkipped(Xi.addTree(r,this.fragments,null!=this.parse.stoppedAt)),this.treeLen=null!==(n=this.parse.stoppedAt)&&void 0!==n?n:this.state.doc.length,this.tree=r,this.parse=null,!(this.treeLen<(null!=t?t:this.state.doc.length)))return!0;this.parse=this.startParse()}if(e())return!1}}))}takeTree(){let e,t;this.parse&&(e=this.parse.parsedPos)>=this.treeLen&&((null==this.parse.stoppedAt||this.parse.stoppedAt>e)&&this.parse.stopAt(e),this.withContext((()=>{for(;!(t=this.parse.advance()););})),this.treeLen=e,this.tree=t,this.fragments=this.withoutTempSkipped(Xi.addTree(this.tree,this.fragments,!0)),this.parse=null)}withContext(e){let t=zs;zs=this;try{return e()}finally{zs=t}}withoutTempSkipped(e){for(let t;t=this.tempSkipped.pop();)e=Ws(e,t.from,t.to);return e}changes(e,t){let{fragments:n,tree:r,treeLen:a,viewport:o,skipped:i}=this;if(this.takeTree(),!e.empty){let t=[];if(e.iterChangedRanges(((e,n,r,a)=>t.push({fromA:e,toA:n,fromB:r,toB:a}))),n=Xi.applyChanges(n,t),r=_i.empty,a=0,o={from:e.mapPos(o.from,-1),to:e.mapPos(o.to,1)},this.skipped.length){i=[];for(let t of this.skipped){let n=e.mapPos(t.from,1),r=e.mapPos(t.to,-1);n<r&&i.push({from:n,to:r})}}}return new js(this.parser,t,n,r,a,o,i,this.scheduleOn)}updateViewport(e){if(this.viewport.from==e.from&&this.viewport.to==e.to)return!1;this.viewport=e;let t=this.skipped.length;for(let t=0;t<this.skipped.length;t++){let{from:n,to:r}=this.skipped[t];n<e.to&&r>e.from&&(this.fragments=Ws(this.fragments,n,r),this.skipped.splice(t--,1))}return!(this.skipped.length>=t||(this.reset(),0))}reset(){this.parse&&(this.takeTree(),this.parse=null)}skipUntilInView(e,t){this.skipped.push({from:e,to:t})}static getSkippingParser(e){return new class extends qi{createParse(t,n,r){let a=r[0].from,o=r[r.length-1].to;return{parsedPos:a,advance(){let t=zs;if(t){for(let e of r)t.tempSkipped.push(e);e&&(t.scheduleOn=t.scheduleOn?Promise.all([t.scheduleOn,e]):e)}return this.parsedPos=o,new _i(Oi.none,[],[],o-a)},stoppedAt:null,stopAt(){}}}}}isDone(e){e=Math.min(e,this.state.doc.length);let t=this.fragments;return this.treeLen>=e&&t.length&&0==t[0].from&&t[0].to>=e}static get(){return zs}}function Ws(e,t,n){return Xi.applyChanges(e,[{fromA:t,toA:n,fromB:t,toB:n}])}class Xs{constructor(e){this.context=e,this.tree=e.tree}apply(e){if(!e.docChanged&&this.tree==this.context.tree)return this;let t=this.context.changes(e.changes,e.state),n=this.context.treeLen==e.startState.doc.length?void 0:Math.max(e.changes.mapPos(this.context.treeLen),t.viewport.to);return t.work(20,n)||t.takeTree(),new Xs(t)}static init(e){let t=Math.min(3e3,e.doc.length),n=js.create(e.facet(Zs).parser,e,{from:0,to:t});return n.work(20,t)||n.takeTree(),new Xs(n)}}Ls.state=Z.define({create:Xs.init,update(e,t){for(let e of t.effects)if(e.is(Ls.setState))return e.value;return t.startState.facet(Zs)!=t.state.facet(Zs)?Xs.init(t.state):e.apply(t)}});let qs=e=>{let t=setTimeout((()=>e()),500);return()=>clearTimeout(t)};"undefined"!=typeof requestIdleCallback&&(qs=e=>{let t=-1,n=setTimeout((()=>{t=requestIdleCallback(e,{timeout:400})}),100);return()=>t<0?clearTimeout(n):cancelIdleCallback(t)});const Vs="undefined"!=typeof navigator&&(null===(Ms=navigator.scheduling)||void 0===Ms?void 0:Ms.isInputPending)?()=>navigator.scheduling.isInputPending():null,Hs=er.fromClass(class{constructor(e){this.view=e,this.working=null,this.workScheduled=0,this.chunkEnd=-1,this.chunkBudget=-1,this.work=this.work.bind(this),this.scheduleWork()}update(e){let t=this.view.state.field(Ls.state).context;(t.updateViewport(e.view.viewport)||this.view.viewport.to>t.treeLen)&&this.scheduleWork(),(e.docChanged||e.selectionSet)&&(this.view.hasFocus&&(this.chunkBudget+=50),this.scheduleWork()),this.checkAsyncSchedule(t)}scheduleWork(){if(this.working)return;let{state:e}=this.view,t=e.field(Ls.state);t.tree==t.context.tree&&t.context.isDone(e.doc.length)||(this.working=qs(this.work))}work(e){this.working=null;let t=Date.now();if(this.chunkEnd<t&&(this.chunkEnd<0||this.view.hasFocus)&&(this.chunkEnd=t+3e4,this.chunkBudget=3e3),this.chunkBudget<=0)return;let{state:n,viewport:{to:r}}=this.view,a=n.field(Ls.state);if(a.tree==a.context.tree&&a.context.isDone(r+1e5))return;let o=Date.now()+Math.min(this.chunkBudget,100,e&&!Vs?Math.max(25,e.timeRemaining()-5):1e9),i=a.context.treeLen<r&&n.doc.length>r+1e3,s=a.context.work((()=>Vs&&Vs()||Date.now()>o),r+(i?0:1e5));this.chunkBudget-=Date.now()-t,(s||this.chunkBudget<=0)&&(a.context.takeTree(),this.view.dispatch({effects:Ls.setState.of(new Xs(a.context))})),this.chunkBudget>0&&(!s||i)&&this.scheduleWork(),this.checkAsyncSchedule(a.context)}checkAsyncSchedule(e){e.scheduleOn&&(this.workScheduled++,e.scheduleOn.then((()=>this.scheduleWork())).catch((e=>Gn(this.view.state,e))).then((()=>this.workScheduled--)),e.scheduleOn=null)}destroy(){this.working&&this.working()}isWorking(){return!!(this.working||this.workScheduled>0)}},{eventHandlers:{focus(){this.scheduleWork()}}}),Zs=z.define({combine:e=>e.length?e[0]:null,enables:e=>[Ls.state,Hs,Ya.contentAttributes.compute([e],(t=>{let n=t.facet(e);return n&&n.name?{"data-language":n.name}:{}}))]});class Gs{constructor(e,t=[]){this.language=e,this.support=t,this.extension=[e,t]}}const Ks=z.define(),Ys=z.define({combine:e=>{if(!e.length)return"  ";let t=e[0];if(!t||/\S/.test(t)||Array.from(t).some((e=>e!=t[0])))throw new Error("Invalid indent unit: "+JSON.stringify(e[0]));return t}});function Js(e){let t=e.facet(Ys);return 9==t.charCodeAt(0)?e.tabSize*t.length:t.length}function el(e,t){let n="",r=e.tabSize,a=e.facet(Ys)[0];if("\t"==a){for(;t>=r;)n+="\t",t-=r;a=" "}for(let e=0;e<t;e++)n+=a;return n}function tl(e,t){e instanceof Se&&(e=new nl(e));for(let n of e.state.facet(Ks)){let r=n(e,t);if(void 0!==r)return r}let n=Qs(e.state);return n.length>=t?function(e,t,n){let r=t.resolveStack(n),a=t.resolveInner(n,-1).resolve(n,0).enterUnfinishedNodesBefore(n);if(a!=r.node){let e=[];for(let t=a;t&&(t.from!=r.node.from||t.type!=r.node.type);t=t.parent)e.push(t);for(let t=e.length-1;t>=0;t--)r={node:e[t],next:r}}return al(r,e,n)}(e,n,t):null}class nl{constructor(e,t={}){this.state=e,this.options=t,this.unit=Js(e)}lineAt(e,t=1){let n=this.state.doc.lineAt(e),{simulateBreak:r,simulateDoubleBreak:a}=this.options;return null!=r&&r>=n.from&&r<=n.to?a&&r==e?{text:"",from:e}:(t<0?r<e:r<=e)?{text:n.text.slice(r-n.from),from:r}:{text:n.text.slice(0,r-n.from),from:n.from}:n}textAfterPos(e,t=1){if(this.options.simulateDoubleBreak&&e==this.options.simulateBreak)return"";let{text:n,from:r}=this.lineAt(e,t);return n.slice(e-r,Math.min(n.length,e+100-r))}column(e,t=1){let{text:n,from:r}=this.lineAt(e,t),a=this.countColumn(n,e-r),o=this.options.overrideIndentation?this.options.overrideIndentation(r):-1;return o>-1&&(a+=o-this.countColumn(n,n.search(/\S|$/))),a}countColumn(e,t=e.length){return je(e,this.state.tabSize,t)}lineIndent(e,t=1){let{text:n,from:r}=this.lineAt(e,t),a=this.options.overrideIndentation;if(a){let e=a(r);if(e>-1)return e}return this.countColumn(n,n.search(/\S|$/))}get simulatedBreak(){return this.options.simulateBreak||null}}const rl=new bi;function al(e,t,n){for(let r=e;r;r=r.next){let e=ol(r.node);if(e)return e(sl.create(t,n,r))}return 0}function ol(e){let t=e.type.prop(rl);if(t)return t;let n,r=e.firstChild;if(r&&(n=r.type.prop(bi.closedBy))){let t=e.lastChild,r=t&&n.indexOf(t.name)>-1;return e=>ul(e,!0,1,void 0,r&&!function(e){return e.pos==e.options.simulateBreak&&e.options.simulateDoubleBreak}(e)?t.from:void 0)}return null==e.parent?il:null}function il(){return 0}class sl extends nl{constructor(e,t,n){super(e.state,e.options),this.base=e,this.pos=t,this.context=n}get node(){return this.context.node}static create(e,t,n){return new sl(e,t,n)}get textAfter(){return this.textAfterPos(this.pos)}get baseIndent(){return this.baseIndentFor(this.node)}baseIndentFor(e){let t=this.state.doc.lineAt(e.from);for(;;){let n=e.resolve(t.from);for(;n.parent&&n.parent.from==n.from;)n=n.parent;if(ll(n,e))break;t=this.state.doc.lineAt(n.from)}return this.lineIndent(t.from)}continue(){return al(this.context.next,this.base,this.pos)}}function ll(e,t){for(let n=t;n;n=n.parent)if(e==n)return!0;return!1}function cl({closing:e,align:t=!0,units:n=1}){return r=>ul(r,t,n,e)}function ul(e,t,n,r,a){let o=e.textAfter,i=o.match(/^\s*/)[0].length,s=r&&o.slice(i,i+r.length)==r||a==e.pos+i,l=t?function(e){let t=e.node,n=t.childAfter(t.from),r=t.lastChild;if(!n)return null;let a=e.options.simulateBreak,o=e.state.doc.lineAt(n.from),i=null==a||a<=o.from?o.to:Math.min(o.to,a);for(let e=n.to;;){let a=t.childAfter(e);if(!a||a==r)return null;if(!a.type.isSkipped){if(a.from>=i)return null;let e=/^ */.exec(o.text.slice(n.to-o.from))[0].length;return{from:n.from,to:n.to+e}}e=a.to}}(e):null;return l?s?e.column(l.from):e.column(l.to):e.baseIndent+(s?0:e.unit*n)}function dl({except:e,units:t=1}={}){return n=>{let r=e&&e.test(n.textAfter);return n.baseIndent+(r?0:t*n.unit)}}const pl=z.define(),hl=new bi;function fl(e){let t=e.firstChild,n=e.lastChild;return t&&t.to<n.from?{from:t.to,to:n.type.isError?e.to:n.from}:null}function ml(e){let t=e.lastChild;return t&&t.to==e.to&&t.type.isError}function gl(e,t,n){for(let r of e.facet(pl)){let a=r(e,t,n);if(a)return a}return function(e,t,n){let r=Qs(e);if(r.length<n)return null;let a=null;for(let o=r.resolveStack(n,1);o;o=o.next){let i=o.node;if(i.to<=n||i.from>n)continue;if(a&&i.from<t)break;let s=i.type.prop(hl);if(s&&(i.to<r.length-50||r.length==e.doc.length||!ml(i))){let r=s(i,e);r&&r.from<=n&&r.from>=t&&r.to>n&&(a=r)}}return a}(e,t,n)}function vl(e,t){let n=t.mapPos(e.from,1),r=t.mapPos(e.to,-1);return n>=r?void 0:{from:n,to:r}}const bl=fe.define({map:vl}),yl=fe.define({map:vl}),wl=Z.define({create:()=>rn.none,update(e,t){e=e.map(t.changes);for(let n of t.effects)if(n.is(bl)&&!kl(e,n.value.from,n.value.to)){let{preparePlaceholder:r}=t.state.facet(El),a=r?rn.replace({widget:new Tl(r(t.state,n.value))}):Cl;e=e.update({add:[a.range(n.value.from,n.value.to)]})}else n.is(yl)&&(e=e.update({filter:(e,t)=>n.value.from!=e||n.value.to!=t,filterFrom:n.value.from,filterTo:n.value.to}));if(t.selection){let n=!1,{head:r}=t.selection.main;e.between(r,r,((e,t)=>{e<r&&t>r&&(n=!0)})),n&&(e=e.update({filterFrom:r,filterTo:r,filter:(e,t)=>t<=r||e>=r}))}return e},provide:e=>Ya.decorations.from(e),toJSON(e,t){let n=[];return e.between(0,t.doc.length,((e,t)=>{n.push(e,t)})),n},fromJSON(e){if(!Array.isArray(e)||e.length%2)throw new RangeError("Invalid JSON for fold state");let t=[];for(let n=0;n<e.length;){let r=e[n++],a=e[n++];if("number"!=typeof r||"number"!=typeof a)throw new RangeError("Invalid JSON for fold state");t.push(Cl.range(r,a))}return rn.set(t,!0)}});function Ol(e,t,n){var r;let a=null;return null===(r=e.field(wl,!1))||void 0===r||r.between(t,n,((e,t)=>{(!a||a.from>e)&&(a={from:e,to:t})})),a}function kl(e,t,n){let r=!1;return e.between(t,t,((e,a)=>{e==t&&a==n&&(r=!0)})),r}const xl={placeholderDOM:null,preparePlaceholder:null,placeholderText:"…"},El=z.define({combine:e=>_e(e,xl)});function Sl(e){let t=[wl,Dl];return e&&t.push(El.of(e)),t}function _l(e,t){let{state:n}=e,r=n.facet(El),a=t=>{let n=e.lineBlockAt(e.posAtDOM(t.target)),r=Ol(e.state,n.from,n.to);r&&e.dispatch({effects:yl.of(r)}),t.preventDefault()};if(r.placeholderDOM)return r.placeholderDOM(e,a,t);let o=document.createElement("span");return o.textContent=r.placeholderText,o.setAttribute("aria-label",n.phrase("folded code")),o.title=n.phrase("unfold"),o.className="cm-foldPlaceholder",o.onclick=a,o}const Cl=rn.replace({widget:new class extends tn{toDOM(e){return _l(e,null)}}});class Tl extends tn{constructor(e){super(),this.value=e}eq(e){return this.value==e.value}toDOM(e){return _l(e,this.value)}}const Al={openText:"⌄",closedText:"›",markerDOM:null,domEventHandlers:{},foldingChanged:()=>!1};class Rl extends Ho{constructor(e,t){super(),this.config=e,this.open=t}eq(e){return this.config==e.config&&this.open==e.open}toDOM(e){if(this.config.markerDOM)return this.config.markerDOM(this.open);let t=document.createElement("span");return t.textContent=this.open?this.config.openText:this.config.closedText,t.title=e.state.phrase(this.open?"Fold line":"Unfold line"),t}}function Pl(e={}){let t=Object.assign(Object.assign({},Al),e),n=new Rl(t,!0),r=new Rl(t,!1),a=er.fromClass(class{constructor(e){this.from=e.viewport.from,this.markers=this.buildMarkers(e)}update(e){(e.docChanged||e.viewportChanged||e.startState.facet(Zs)!=e.state.facet(Zs)||e.startState.field(wl,!1)!=e.state.field(wl,!1)||Qs(e.startState)!=Qs(e.state)||t.foldingChanged(e))&&(this.markers=this.buildMarkers(e.view))}buildMarkers(e){let t=new De;for(let a of e.viewportLineBlocks){let o=Ol(e.state,a.from,a.to)?r:gl(e.state,a.from,a.to)?n:null;o&&t.add(a.from,a.from,o)}return t.finish()}}),{domEventHandlers:o}=t;return[a,Jo({class:"cm-foldGutter",markers(e){var t;return(null===(t=e.plugin(a))||void 0===t?void 0:t.markers)||Pe.empty},initialSpacer:()=>new Rl(t,!1),domEventHandlers:Object.assign(Object.assign({},o),{click:(e,t,n)=>{if(o.click&&o.click(e,t,n))return!0;let r=Ol(e.state,t.from,t.to);if(r)return e.dispatch({effects:yl.of(r)}),!0;let a=gl(e.state,t.from,t.to);return!!a&&(e.dispatch({effects:bl.of(a)}),!0)}})}),Sl()]}const Dl=Ya.baseTheme({".cm-foldPlaceholder":{backgroundColor:"#eee",border:"1px solid #ddd",color:"#888",borderRadius:".2em",margin:"0 1px",padding:"0 1px",cursor:"pointer"},".cm-foldGutter span":{padding:"0 1px",cursor:"pointer"}});class Ml{constructor(e,t){let n;function r(e){let t=Ve.newName();return(n||(n=Object.create(null)))["."+t]=e,t}this.specs=e;const a="string"==typeof t.all?t.all:t.all?r(t.all):void 0,o=t.scope;this.scope=o instanceof Ls?e=>e.prop($s)==o.data:o?e=>e==o:void 0,this.style=fs(e.map((e=>({tag:e.tag,class:e.class||r(Object.assign({},e,{tag:null}))}))),{all:a}).style,this.module=n?new Ve(n):null,this.themeType=t.themeType}static define(e,t){return new Ml(e,t||{})}}const $l=z.define(),Nl=z.define({combine:e=>e.length?[e[0]]:null});function Il(e){let t=e.facet($l);return t.length?t:e.facet(Nl)}function Ll(e,t){let n,r=[Fl];return e instanceof Ml&&(e.module&&r.push(Ya.styleModule.of(e.module)),n=e.themeType),(null==t?void 0:t.fallback)?r.push(Nl.of(e)):n?r.push($l.computeN([Ya.darkTheme],(t=>t.facet(Ya.darkTheme)==("dark"==n)?[e]:[]))):r.push($l.of(e)),r}class Bl{constructor(e){this.markCache=Object.create(null),this.tree=Qs(e.state),this.decorations=this.buildDeco(e,Il(e.state)),this.decoratedTo=e.viewport.to}update(e){let t=Qs(e.state),n=Il(e.state),r=n!=Il(e.startState),{viewport:a}=e.view,o=e.changes.mapPos(this.decoratedTo,1);t.length<a.to&&!r&&t.type==this.tree.type&&o>=a.to?(this.decorations=this.decorations.map(e.changes),this.decoratedTo=o):(t!=this.tree||e.viewportChanged||r)&&(this.tree=t,this.decorations=this.buildDeco(e.view,n),this.decoratedTo=a.to)}buildDeco(e,t){if(!t||!this.tree.length)return rn.none;let n=new De;for(let{from:r,to:a}of e.visibleRanges)ms(this.tree,t,((e,t,r)=>{n.add(e,t,this.markCache[r]||(this.markCache[r]=rn.mark({class:r})))}),r,a);return n.finish()}}const Fl=K.high(er.fromClass(Bl,{decorations:e=>e.decorations})),Ql=Ml.define([{tag:Ds.meta,color:"#404740"},{tag:Ds.link,textDecoration:"underline"},{tag:Ds.heading,textDecoration:"underline",fontWeight:"bold"},{tag:Ds.emphasis,fontStyle:"italic"},{tag:Ds.strong,fontWeight:"bold"},{tag:Ds.strikethrough,textDecoration:"line-through"},{tag:Ds.keyword,color:"#708"},{tag:[Ds.atom,Ds.bool,Ds.url,Ds.contentSeparator,Ds.labelName],color:"#219"},{tag:[Ds.literal,Ds.inserted],color:"#164"},{tag:[Ds.string,Ds.deleted],color:"#a11"},{tag:[Ds.regexp,Ds.escape,Ds.special(Ds.string)],color:"#e40"},{tag:Ds.definition(Ds.variableName),color:"#00f"},{tag:Ds.local(Ds.variableName),color:"#30a"},{tag:[Ds.typeName,Ds.namespace],color:"#085"},{tag:Ds.className,color:"#167"},{tag:[Ds.special(Ds.variableName),Ds.macroName],color:"#256"},{tag:Ds.definition(Ds.propertyName),color:"#00c"},{tag:Ds.comment,color:"#940"},{tag:Ds.invalid,color:"#f00"}]),Ul=Ya.baseTheme({"&.cm-focused .cm-matchingBracket":{backgroundColor:"#328c8252"},"&.cm-focused .cm-nonmatchingBracket":{backgroundColor:"#bb555544"}}),zl="()[]{}",jl=z.define({combine:e=>_e(e,{afterCursor:!0,brackets:zl,maxScanDistance:1e4,renderMatch:ql})}),Wl=rn.mark({class:"cm-matchingBracket"}),Xl=rn.mark({class:"cm-nonmatchingBracket"});function ql(e){let t=[],n=e.matched?Wl:Xl;return t.push(n.range(e.start.from,e.start.to)),e.end&&t.push(n.range(e.end.from,e.end.to)),t}const Vl=Z.define({create:()=>rn.none,update(e,t){if(!t.docChanged&&!t.selection)return e;let n=[],r=t.state.facet(jl);for(let e of t.state.selection.ranges){if(!e.empty)continue;let a=Jl(t.state,e.head,-1,r)||e.head>0&&Jl(t.state,e.head-1,1,r)||r.afterCursor&&(Jl(t.state,e.head,1,r)||e.head<t.state.doc.length&&Jl(t.state,e.head+1,-1,r));a&&(n=n.concat(r.renderMatch(a,t.state)))}return rn.set(n,!0)},provide:e=>Ya.decorations.from(e)}),Hl=[Vl,Ul];function Zl(e={}){return[jl.of(e),Hl]}const Gl=new bi;function Kl(e,t,n){let r=e.prop(t<0?bi.openedBy:bi.closedBy);if(r)return r;if(1==e.name.length){let r=n.indexOf(e.name);if(r>-1&&r%2==(t<0?1:0))return[n[r+t]]}return null}function Yl(e){let t=e.type.prop(Gl);return t?t(e.node):e}function Jl(e,t,n,r={}){let a=r.maxScanDistance||1e4,o=r.brackets||zl,i=Qs(e),s=i.resolveInner(t,n);for(let e=s;e;e=e.parent){let r=Kl(e.type,n,o);if(r&&e.from<e.to){let a=Yl(e);if(a&&(n>0?t>=a.from&&t<a.to:t>a.from&&t<=a.to))return ec(0,0,n,e,a,r,o)}}return function(e,t,n,r,a,o,i){let s=n<0?e.sliceDoc(t-1,t):e.sliceDoc(t,t+1),l=i.indexOf(s);if(l<0||l%2==0!=n>0)return null;let c={from:n<0?t-1:t,to:n>0?t+1:t},u=e.doc.iterRange(t,n>0?e.doc.length:0),d=0;for(let e=0;!u.next().done&&e<=o;){let o=u.value;n<0&&(e+=o.length);let s=t+e*n;for(let e=n>0?0:o.length-1,t=n>0?o.length:-1;e!=t;e+=n){let t=i.indexOf(o[e]);if(!(t<0||r.resolveInner(s+e,1).type!=a))if(t%2==0==n>0)d++;else{if(1==d)return{start:c,end:{from:s+e,to:s+e+1},matched:t>>1==l>>1};d--}}n>0&&(e+=o.length)}return u.done?{start:c,matched:!1}:null}(e,t,n,i,s.type,a,o)}function ec(e,t,n,r,a,o,i){let s=r.parent,l={from:a.from,to:a.to},c=0,u=null==s?void 0:s.cursor();if(u&&(n<0?u.childBefore(r.from):u.childAfter(r.to)))do{if(n<0?u.to<=r.from:u.from>=r.to){if(0==c&&o.indexOf(u.type.name)>-1&&u.from<u.to){let e=Yl(u);return{start:l,end:e?{from:e.from,to:e.to}:void 0,matched:!0}}if(Kl(u.type,n,i))c++;else if(Kl(u.type,-n,i)){if(0==c){let e=Yl(u);return{start:l,end:e&&e.from<e.to?{from:e.from,to:e.to}:void 0,matched:!1}}c--}}}while(n<0?u.prevSibling():u.nextSibling());return{start:l,matched:!1}}const tc=Object.create(null),nc=[Oi.none],rc=[],ac=Object.create(null),oc=Object.create(null);for(let[e,t]of[["variable","variableName"],["variable-2","variableName.special"],["string-2","string.special"],["def","variableName.definition"],["tag","tagName"],["attribute","attributeName"],["type","typeName"],["builtin","variableName.standard"],["qualifier","modifier"],["error","invalid"],["header","heading"],["property","propertyName"]])oc[e]=sc(tc,t);function ic(e,t){rc.indexOf(e)>-1||(rc.push(e),console.warn(t))}function sc(e,t){let n=[];for(let r of t.split(" ")){let t=[];for(let n of r.split(".")){let r=e[n]||Ds[n];r?"function"==typeof r?t.length?t=t.map(r):ic(n,`Modifier ${n} used at start of tag`):t.length?ic(n,`Tag ${n} used as modifier`):t=Array.isArray(r)?r:[r]:ic(n,`Unknown highlighting tag ${n}`)}for(let e of t)n.push(e)}if(!n.length)return 0;let r=t.replace(/ /g,"_"),a=r+" "+n.map((e=>e.id)),o=ac[a];if(o)return o.id;let i=ac[a]=Oi.define({id:nc.length,name:r,props:[ds({[r]:n})]});return nc.push(i),i.id}gn.RTL,gn.LTR;const lc=new Map([["aliceblue","#f0f8ff"],["antiquewhite","#faebd7"],["aqua","#00ffff"],["aquamarine","#7fffd4"],["azure","#f0ffff"],["beige","#f5f5dc"],["bisque","#ffe4c4"],["black","#000000"],["blanchedalmond","#ffebcd"],["blue","#0000ff"],["blueviolet","#8a2be2"],["brown","#a52a2a"],["burlywood","#deb887"],["cadetblue","#5f9ea0"],["chartreuse","#7fff00"],["chocolate","#d2691e"],["coral","#ff7f50"],["cornflowerblue","#6495ed"],["cornsilk","#fff8dc"],["crimson","#dc143c"],["cyan","#00ffff"],["darkblue","#00008b"],["darkcyan","#008b8b"],["darkgoldenrod","#b8860b"],["darkgray","#a9a9a9"],["darkgreen","#006400"],["darkgrey","#a9a9a9"],["darkkhaki","#bdb76b"],["darkmagenta","#8b008b"],["darkolivegreen","#556b2f"],["darkorange","#ff8c00"],["darkorchid","#9932cc"],["darkred","#8b0000"],["darksalmon","#e9967a"],["darkseagreen","#8fbc8f"],["darkslateblue","#483d8b"],["darkslategray","#2f4f4f"],["darkslategrey","#2f4f4f"],["darkturquoise","#00ced1"],["darkviolet","#9400d3"],["deeppink","#ff1493"],["deepskyblue","#00bfff"],["dimgray","#696969"],["dimgrey","#696969"],["dodgerblue","#1e90ff"],["firebrick","#b22222"],["floralwhite","#fffaf0"],["forestgreen","#228b22"],["fuchsia","#ff00ff"],["gainsboro","#dcdcdc"],["ghostwhite","#f8f8ff"],["goldenrod","#daa520"],["gold","#ffd700"],["gray","#808080"],["green","#008000"],["greenyellow","#adff2f"],["grey","#808080"],["honeydew","#f0fff0"],["hotpink","#ff69b4"],["indianred","#cd5c5c"],["indigo","#4b0082"],["ivory","#fffff0"],["khaki","#f0e68c"],["lavenderblush","#fff0f5"],["lavender","#e6e6fa"],["lawngreen","#7cfc00"],["lemonchiffon","#fffacd"],["lightblue","#add8e6"],["lightcoral","#f08080"],["lightcyan","#e0ffff"],["lightgoldenrodyellow","#fafad2"],["lightgray","#d3d3d3"],["lightgreen","#90ee90"],["lightgrey","#d3d3d3"],["lightpink","#ffb6c1"],["lightsalmon","#ffa07a"],["lightseagreen","#20b2aa"],["lightskyblue","#87cefa"],["lightslategray","#778899"],["lightslategrey","#778899"],["lightsteelblue","#b0c4de"],["lightyellow","#ffffe0"],["lime","#00ff00"],["limegreen","#32cd32"],["linen","#faf0e6"],["magenta","#ff00ff"],["maroon","#800000"],["mediumaquamarine","#66cdaa"],["mediumblue","#0000cd"],["mediumorchid","#ba55d3"],["mediumpurple","#9370db"],["mediumseagreen","#3cb371"],["mediumslateblue","#7b68ee"],["mediumspringgreen","#00fa9a"],["mediumturquoise","#48d1cc"],["mediumvioletred","#c71585"],["midnightblue","#191970"],["mintcream","#f5fffa"],["mistyrose","#ffe4e1"],["moccasin","#ffe4b5"],["navajowhite","#ffdead"],["navy","#000080"],["oldlace","#fdf5e6"],["olive","#808000"],["olivedrab","#6b8e23"],["orange","#ffa500"],["orangered","#ff4500"],["orchid","#da70d6"],["palegoldenrod","#eee8aa"],["palegreen","#98fb98"],["paleturquoise","#afeeee"],["palevioletred","#db7093"],["papayawhip","#ffefd5"],["peachpuff","#ffdab9"],["peru","#cd853f"],["pink","#ffc0cb"],["plum","#dda0dd"],["powderblue","#b0e0e6"],["purple","#800080"],["rebeccapurple","#663399"],["red","#ff0000"],["rosybrown","#bc8f8f"],["royalblue","#4169e1"],["saddlebrown","#8b4513"],["salmon","#fa8072"],["sandybrown","#f4a460"],["seagreen","#2e8b57"],["seashell","#fff5ee"],["sienna","#a0522d"],["silver","#c0c0c0"],["skyblue","#87ceeb"],["slateblue","#6a5acd"],["slategray","#708090"],["slategrey","#708090"],["snow","#fffafa"],["springgreen","#00ff7f"],["steelblue","#4682b4"],["tan","#d2b48c"],["teal","#008080"],["thistle","#d8bfd8"],["tomato","#ff6347"],["turquoise","#40e0d0"],["violet","#ee82ee"],["wheat","#f5deb3"],["white","#ffffff"],["whitesmoke","#f5f5f5"],["yellow","#ffff00"],["yellowgreen","#9acd32"]]);const cc=new WeakMap;var uc=function(e){return e.rgb="RGB",e.hex="HEX",e.named="NAMED",e.hsl="HSL",e}(uc||(uc={}));const dc=/rgb(?:a)?\(\s*(\d{1,3}%?)\s*,?\s*(\d{1,3}%?)\s*,?\s*(\d{1,3}%?)\s*([,/]\s*0?\.?\d+%?)?\)/,pc=/hsl\(\s*(\d{1,3})\s*,\s*(\d{1,3})%\s*,\s*(\d{1,3})%\s*(,\s*0?\.\d+)?\)/,hc=/(^|\b)(#[0-9a-f]{3,9})(\b|$)/i;function fc(e,t){const n=[],r=Qs(e.state);for(const a of e.visibleRanges)r.iterate({from:a.from,to:a.to,enter:({type:a,from:o,to:i})=>{var s;const l=t(r,o,i,a.name,e.state.doc,null===(s=e.state.facet(Zs))||void 0===s?void 0:s.name);if(l)if(Array.isArray(l))for(const e of l)n.push(rn.widget({widget:new Oc(e),side:1}).range(e.from));else n.push(rn.widget({widget:new Oc(l),side:1}).range(l.from))}});return rn.set(n)}function mc(e){let t;if(e.endsWith("%")){const n=Number(e.slice(0,-1));t=Math.round(n/100*255)}else t=Number(e);return gc(t)}function gc(e){const t=e.toString(16);return 1===t.length?"0"+t:t}function vc(e){const t=e.slice(1,3),n=e.slice(3,5),r=e.slice(5,7);return[parseInt(t,16),parseInt(n,16),parseInt(r,16)]}function bc(e){return e<0?e+1:e>1?e-1:e}function yc(e,t,n){return 6*n<1?t+6*(e-t)*n:2*n<1?e:3*n<2?t+(e-t)*(.666-n)*6:t}const wc="cm-css-color-picker-wrapper";class Oc extends tn{constructor(e){var{color:t}=e,n=function(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var a=0;for(r=Object.getOwnPropertySymbols(e);a<r.length;a++)t.indexOf(r[a])<0&&Object.prototype.propertyIsEnumerable.call(e,r[a])&&(n[r[a]]=e[r[a]])}return n}(e,["color"]);super(),this.state=n,this.color=t}eq(e){return e.state.colorType===this.state.colorType&&e.color===this.color&&e.state.from===this.state.from&&e.state.to===this.state.to&&e.state.alpha===this.state.alpha}toDOM(){const e=document.createElement("input");cc.set(e,this.state),e.type="color",e.value=this.color;const t=document.createElement("span");return t.appendChild(e),t.className=wc,t}ignoreEvent(){return!1}}const kc=Ya.baseTheme({[`.${wc}`]:{display:"inline-block",outline:"1px solid #eee",marginRight:"0.6ch",height:"1em",width:"1em",transform:"translateY(1px)"},[`.${wc} input[type="color"]`]:{cursor:"pointer",height:"100%",width:"100%",padding:0,border:"none","&::-webkit-color-swatch-wrapper":{padding:0},"&::-webkit-color-swatch":{border:"none"},"&::-moz-color-swatch":{border:"none"}}}),xc=[(e=>er.fromClass(class{constructor(t){this.decorations=fc(t,e.discoverColors)}update(t){(t.docChanged||t.viewportChanged)&&(this.decorations=fc(t.view,e.discoverColors))}},{decorations:e=>e.decorations,eventHandlers:{change:(e,t)=>{const n=e.target;if("INPUT"!==n.nodeName||!n.parentElement||!n.parentElement.classList.contains(wc))return!1;const r=cc.get(n);let a=n.value+r.alpha;if(r.colorType===uc.rgb)a=`rgb(${vc(n.value).join(", ")}${r.alpha})`;else if(r.colorType===uc.named)for(const[e,t]of lc.entries())t===n.value&&(a=e);else if(r.colorType===uc.hsl){const[e,t,o]=vc(n.value),[i,s,l]=function(e,t,n){const r=e/255,a=t/255,o=n/255,i=Math.min(r,a,o),s=Math.max(r,a,o),l=(s+i)/2;if(s===i)return[0,0,l];let c,u;for(c=l<=.5?(s-i)/(s+i):(s-i)/(2-s-i),u=s===r?(a-o)/(s-i):a===s?2+(o-r)/(s-i):4+(r-a)/(s-i),u=Math.round(60*u);u<0;)u+=360;return[u,c,l]}(e,t,o);a=`hsl(${i}, ${Math.round(100*s)}%, ${Math.round(100*l)}%${r.alpha})`}return t.dispatch({changes:{from:r.from,to:r.to,insert:a}}),!0}}}))({discoverColors:function e(t,n,r,a,o,i){var s;switch(a){case"AttributeValue":{const r=t.resolveInner(n,0).tree;if(!r)return null;const a=null===(s=r.prop(bi.mounted))||void 0===s?void 0:s.tree;if("Styles"!==(null==a?void 0:a.type.name))return null;const i=[];return a.iterate({from:0,to:a.length,enter:({type:r,from:a,to:s})=>{const l=e(t,n+1+a,n+1+s,r.name,o);if(l){if(Array.isArray(l))throw new Error("Unexpected nested overlays");i.push(l)}}}),i}case"CallExpression":{const e=function(e){switch(e.slice(0,3)){case"rgb":{const t=dc.exec(e);if(!t)return null;const[n,r,a,o,i]=t,s=function(e,t,n){return`#${mc(e)}${mc(t)}${mc(n)}`}(r,a,o);return{colorType:uc.rgb,color:s,alpha:i||""}}case"hsl":{const t=pc.exec(e);if(!t)return null;const[n,r,a,o,i]=t,s=function(e,t,n){const r=Number(t)/100,a=Number(n)/100,[o,i,s]=function(e,t,n){if(0===t){const e=Math.round(255*n);return[e,e,e]}let r;r=n<.5?n*(1+t):n+t-n*t;const a=2*n-r,o=bc(.333+(e/=360)),i=e,s=bc(e-.333),l=yc(r,a,o),c=yc(r,a,i),u=yc(r,a,s);return[Math.round(255*l),Math.round(255*c),Math.round(255*u)]}(Number(e),r,a);return`#${gc(o)}${gc(i)}${gc(s)}`}(r,a,o);return{colorType:uc.hsl,color:s,alpha:i||""}}default:return null}}(o.sliceString(n,r));return e?Object.assign(Object.assign({},e),{from:n,to:r}):null}case"ColorLiteral":{const e=function(e){if(!hc.exec(e))return null;const[t,n]=function(e){return 4===e.length?[`#${e[1].repeat(2)}${e[2].repeat(2)}${e[3].repeat(2)}`,""]:5===e.length?[`#${e[1].repeat(2)}${e[2].repeat(2)}${e[3].repeat(2)}`,e[4].repeat(2)]:9===e.length?[`#${e.slice(1,-2)}`,e.slice(-2)]:[e,""]}(e);return{colorType:uc.hex,color:t,alpha:n}}(o.sliceString(n,r));return e?Object.assign(Object.assign({},e),{from:n,to:r}):null}case"ValueName":{const e=function(e){const t=lc.get(e);return t?{colorType:uc.named,color:t,alpha:""}:null}(o.sliceString(n,r));return e?Object.assign(Object.assign({},e),{from:n,to:r}):null}default:return null}}}),kc];class Ec{constructor(e,t,n,r){this.state=e,this.pos=t,this.explicit=n,this.view=r,this.abortListeners=[],this.abortOnDocChange=!1}tokenBefore(e){let t=Qs(this.state).resolveInner(this.pos,-1);for(;t&&e.indexOf(t.name)<0;)t=t.parent;return t?{from:t.from,to:this.pos,text:this.state.sliceDoc(t.from,this.pos),type:t.type}:null}matchBefore(e){let t=this.state.doc.lineAt(this.pos),n=Math.max(t.from,this.pos-250),r=t.text.slice(n-t.from,this.pos-t.from),a=r.search(Ac(e,!1));return a<0?null:{from:n+a,to:this.pos,text:r.slice(a)}}get aborted(){return null==this.abortListeners}addEventListener(e,t,n){"abort"==e&&this.abortListeners&&(this.abortListeners.push(t),n&&n.onDocChange&&(this.abortOnDocChange=!0))}}function Sc(e){let t=Object.keys(e).join(""),n=/\w/.test(t);return n&&(t=t.replace(/\w/g,"")),`[${n?"\\w":""}${t.replace(/[^\w\s]/g,"\\$&")}]`}function _c(e){let t=e.map((e=>"string"==typeof e?{label:e}:e)),[n,r]=t.every((e=>/^\w+$/.test(e.label)))?[/\w*$/,/\w+$/]:function(e){let t=Object.create(null),n=Object.create(null);for(let{label:r}of e){t[r[0]]=!0;for(let e=1;e<r.length;e++)n[r[e]]=!0}let r=Sc(t)+Sc(n)+"*$";return[new RegExp("^"+r),new RegExp(r)]}(t);return e=>{let a=e.matchBefore(r);return a||e.explicit?{from:a?a.from:e.pos,options:t,validFor:n}:null}}class Cc{constructor(e,t,n,r){this.completion=e,this.source=t,this.match=n,this.score=r}}function Tc(e){return e.selection.main.from}function Ac(e,t){var n;let{source:r}=e,a=t&&"^"!=r[0],o="$"!=r[r.length-1];return a||o?new RegExp(`${a?"^":""}(?:${r})${o?"$":""}`,null!==(n=e.flags)&&void 0!==n?n:e.ignoreCase?"i":""):e}const Rc=de.define(),Pc=new WeakMap;function Dc(e){if(!Array.isArray(e))return e;let t=Pc.get(e);return t||Pc.set(e,t=_c(e)),t}const Mc=fe.define(),$c=fe.define();class Nc{constructor(e){this.pattern=e,this.chars=[],this.folded=[],this.any=[],this.precise=[],this.byWord=[],this.score=0,this.matched=[];for(let t=0;t<e.length;){let n=S(e,t),r=C(n);this.chars.push(n);let a=e.slice(t,t+r),o=a.toUpperCase();this.folded.push(S(o==a?a.toLowerCase():o,0)),t+=r}this.astral=e.length!=this.chars.length}ret(e,t){return this.score=e,this.matched=t,this}match(e){if(0==this.pattern.length)return this.ret(-100,[]);if(e.length<this.pattern.length)return null;let{chars:t,folded:n,any:r,precise:a,byWord:o}=this;if(1==t.length){let r=S(e,0),a=C(r),o=a==e.length?0:-100;if(r==t[0]);else{if(r!=n[0])return null;o+=-200}return this.ret(o,[0,a])}let i=e.indexOf(this.pattern);if(0==i)return this.ret(e.length==this.pattern.length?0:-100,[0,this.pattern.length]);let s=t.length,l=0;if(i<0){for(let a=0,o=Math.min(e.length,200);a<o&&l<s;){let o=S(e,a);o!=t[l]&&o!=n[l]||(r[l++]=a),a+=C(o)}if(l<s)return null}let c=0,u=0,d=!1,p=0,h=-1,f=-1,m=/[a-z]/.test(e),g=!0;for(let r=0,l=Math.min(e.length,200),v=0;r<l&&u<s;){let l=S(e,r);i<0&&(c<s&&l==t[c]&&(a[c++]=r),p<s&&(l==t[p]||l==n[p]?(0==p&&(h=r),f=r+1,p++):p=0));let b,y=l<255?l>=48&&l<=57||l>=97&&l<=122?2:l>=65&&l<=90?1:0:(b=_(l))!=b.toLowerCase()?1:b!=b.toUpperCase()?2:0;(!r||1==y&&m||0==v&&0!=y)&&(t[u]==l||n[u]==l&&(d=!0)?o[u++]=r:o.length&&(g=!1)),v=y,r+=C(l)}return u==s&&0==o[0]&&g?this.result((d?-200:0)-100,o,e):p==s&&0==h?this.ret(-200-e.length+(f==e.length?0:-100),[0,f]):i>-1?this.ret(-700-e.length,[i,i+this.pattern.length]):p==s?this.ret(-900-e.length,[h,f]):u==s?this.result((d?-200:0)-100-700+(g?0:-1100),o,e):2==t.length?null:this.result((r[0]?-700:0)-200-1100,r,e)}result(e,t,n){let r=[],a=0;for(let e of t){let t=e+(this.astral?C(S(n,e)):1);a&&r[a-1]==e?r[a-1]=t:(r[a++]=e,r[a++]=t)}return this.ret(e-n.length,r)}}class Ic{constructor(e){this.pattern=e,this.matched=[],this.score=0,this.folded=e.toLowerCase()}match(e){if(e.length<this.pattern.length)return null;let t=e.slice(0,this.pattern.length),n=t==this.pattern?0:t.toLowerCase()==this.folded?-200:null;return null==n?null:(this.matched=[0,t.length],this.score=n+(e.length==this.pattern.length?0:-100),this)}}const Lc=z.define({combine:e=>_e(e,{activateOnTyping:!0,activateOnCompletion:()=>!1,activateOnTypingDelay:100,selectOnOpen:!0,override:null,closeOnBlur:!0,maxRenderedOptions:100,defaultKeymap:!0,tooltipClass:()=>"",optionClass:()=>"",aboveCursor:!1,icons:!0,addToOptions:[],positionInfo:Fc,filterStrict:!1,compareCompletions:(e,t)=>e.label.localeCompare(t.label),interactionDelay:75,updateSyncTime:100},{defaultKeymap:(e,t)=>e&&t,closeOnBlur:(e,t)=>e&&t,icons:(e,t)=>e&&t,tooltipClass:(e,t)=>n=>Bc(e(n),t(n)),optionClass:(e,t)=>n=>Bc(e(n),t(n)),addToOptions:(e,t)=>e.concat(t),filterStrict:(e,t)=>e||t})});function Bc(e,t){return e?t?e+" "+t:e:t}function Fc(e,t,n,r,a,o){let i,s,l=e.textDirection==gn.RTL,c=l,u=!1,d="top",p=t.left-a.left,h=a.right-t.right,f=r.right-r.left,m=r.bottom-r.top;if(c&&p<Math.min(f,h)?c=!1:!c&&h<Math.min(f,p)&&(c=!0),f<=(c?p:h))i=Math.max(a.top,Math.min(n.top,a.bottom-m))-t.top,s=Math.min(400,c?p:h);else{u=!0,s=Math.min(400,(l?t.right:a.right-t.left)-30);let e=a.bottom-t.bottom;e>=m||e>t.top?i=n.bottom-t.top:(d="bottom",i=t.bottom-n.top)}return{style:`${d}: ${i/((t.bottom-t.top)/o.offsetHeight)}px; max-width: ${s/((t.right-t.left)/o.offsetWidth)}px`,class:"cm-completionInfo-"+(u?l?"left-narrow":"right-narrow":c?"left":"right")}}function Qc(e,t,n){if(e<=n)return{from:0,to:e};if(t<0&&(t=0),t<=e>>1){let e=Math.floor(t/n);return{from:e*n,to:(e+1)*n}}let r=Math.floor((e-t)/n);return{from:e-(r+1)*n,to:e-r*n}}class Uc{constructor(e,t,n){this.view=e,this.stateField=t,this.applyCompletion=n,this.info=null,this.infoDestroy=null,this.placeInfoReq={read:()=>this.measureInfo(),write:e=>this.placeInfo(e),key:this},this.space=null,this.currentClass="";let r=e.state.field(t),{options:a,selected:o}=r.open,i=e.state.facet(Lc);this.optionContent=function(e){let t=e.addToOptions.slice();return e.icons&&t.push({render(e){let t=document.createElement("div");return t.classList.add("cm-completionIcon"),e.type&&t.classList.add(...e.type.split(/\s+/g).map((e=>"cm-completionIcon-"+e))),t.setAttribute("aria-hidden","true"),t},position:20}),t.push({render(e,t,n,r){let a=document.createElement("span");a.className="cm-completionLabel";let o=e.displayLabel||e.label,i=0;for(let e=0;e<r.length;){let t=r[e++],n=r[e++];t>i&&a.appendChild(document.createTextNode(o.slice(i,t)));let s=a.appendChild(document.createElement("span"));s.appendChild(document.createTextNode(o.slice(t,n))),s.className="cm-completionMatchedText",i=n}return i<o.length&&a.appendChild(document.createTextNode(o.slice(i))),a},position:50},{render(e){if(!e.detail)return null;let t=document.createElement("span");return t.className="cm-completionDetail",t.textContent=e.detail,t},position:80}),t.sort(((e,t)=>e.position-t.position)).map((e=>e.render))}(i),this.optionClass=i.optionClass,this.tooltipClass=i.tooltipClass,this.range=Qc(a.length,o,i.maxRenderedOptions),this.dom=document.createElement("div"),this.dom.className="cm-tooltip-autocomplete",this.updateTooltipClass(e.state),this.dom.addEventListener("mousedown",(n=>{let{options:r}=e.state.field(t).open;for(let t,a=n.target;a&&a!=this.dom;a=a.parentNode)if("LI"==a.nodeName&&(t=/-(\d+)$/.exec(a.id))&&+t[1]<r.length)return this.applyCompletion(e,r[+t[1]]),void n.preventDefault()})),this.dom.addEventListener("focusout",(t=>{let n=e.state.field(this.stateField,!1);n&&n.tooltip&&e.state.facet(Lc).closeOnBlur&&t.relatedTarget!=e.contentDOM&&e.dispatch({effects:$c.of(null)})})),this.showOptions(a,r.id)}mount(){this.updateSel()}showOptions(e,t){this.list&&this.list.remove(),this.list=this.dom.appendChild(this.createListBox(e,t,this.range)),this.list.addEventListener("scroll",(()=>{this.info&&this.view.requestMeasure(this.placeInfoReq)}))}update(e){var t;let n=e.state.field(this.stateField),r=e.startState.field(this.stateField);if(this.updateTooltipClass(e.state),n!=r){let{options:a,selected:o,disabled:i}=n.open;r.open&&r.open.options==a||(this.range=Qc(a.length,o,e.state.facet(Lc).maxRenderedOptions),this.showOptions(a,n.id)),this.updateSel(),i!=(null===(t=r.open)||void 0===t?void 0:t.disabled)&&this.dom.classList.toggle("cm-tooltip-autocomplete-disabled",!!i)}}updateTooltipClass(e){let t=this.tooltipClass(e);if(t!=this.currentClass){for(let e of this.currentClass.split(" "))e&&this.dom.classList.remove(e);for(let e of t.split(" "))e&&this.dom.classList.add(e);this.currentClass=t}}positioned(e){this.space=e,this.info&&this.view.requestMeasure(this.placeInfoReq)}updateSel(){let e=this.view.state.field(this.stateField),t=e.open;if((t.selected>-1&&t.selected<this.range.from||t.selected>=this.range.to)&&(this.range=Qc(t.options.length,t.selected,this.view.state.facet(Lc).maxRenderedOptions),this.showOptions(t.options,e.id)),this.updateSelectedOption(t.selected)){this.destroyInfo();let{completion:n}=t.options[t.selected],{info:r}=n;if(!r)return;let a="string"==typeof r?document.createTextNode(r):r(n);if(!a)return;"then"in a?a.then((t=>{t&&this.view.state.field(this.stateField,!1)==e&&this.addInfoPane(t,n)})).catch((e=>Gn(this.view.state,e,"completion info"))):this.addInfoPane(a,n)}}addInfoPane(e,t){this.destroyInfo();let n=this.info=document.createElement("div");if(n.className="cm-tooltip cm-completionInfo",null!=e.nodeType)n.appendChild(e),this.infoDestroy=null;else{let{dom:t,destroy:r}=e;n.appendChild(t),this.infoDestroy=r||null}this.dom.appendChild(n),this.view.requestMeasure(this.placeInfoReq)}updateSelectedOption(e){let t=null;for(let n=this.list.firstChild,r=this.range.from;n;n=n.nextSibling,r++)"LI"==n.nodeName&&n.id?r==e?n.hasAttribute("aria-selected")||(n.setAttribute("aria-selected","true"),t=n):n.hasAttribute("aria-selected")&&n.removeAttribute("aria-selected"):r--;return t&&function(e,t){let n=e.getBoundingClientRect(),r=t.getBoundingClientRect(),a=n.height/e.offsetHeight;r.top<n.top?e.scrollTop-=(n.top-r.top)/a:r.bottom>n.bottom&&(e.scrollTop+=(r.bottom-n.bottom)/a)}(this.list,t),t}measureInfo(){let e=this.dom.querySelector("[aria-selected]");if(!e||!this.info)return null;let t=this.dom.getBoundingClientRect(),n=this.info.getBoundingClientRect(),r=e.getBoundingClientRect(),a=this.space;if(!a){let e=this.dom.ownerDocument.documentElement;a={left:0,top:0,right:e.clientWidth,bottom:e.clientHeight}}return r.top>Math.min(a.bottom,t.bottom)-10||r.bottom<Math.max(a.top,t.top)+10?null:this.view.state.facet(Lc).positionInfo(this.view,t,r,n,a,this.dom)}placeInfo(e){this.info&&(e?(e.style&&(this.info.style.cssText=e.style),this.info.className="cm-tooltip cm-completionInfo "+(e.class||"")):this.info.style.cssText="top: -1e6px")}createListBox(e,t,n){const r=document.createElement("ul");r.id=t,r.setAttribute("role","listbox"),r.setAttribute("aria-expanded","true"),r.setAttribute("aria-label",this.view.state.phrase("Completions")),r.addEventListener("mousedown",(e=>{e.target==r&&e.preventDefault()}));let a=null;for(let o=n.from;o<n.to;o++){let{completion:i,match:s}=e[o],{section:l}=i;if(l){let e="string"==typeof l?l:l.name;e!=a&&(o>n.from||0==n.from)&&(a=e,"string"!=typeof l&&l.header?r.appendChild(l.header(l)):r.appendChild(document.createElement("completion-section")).textContent=e)}const c=r.appendChild(document.createElement("li"));c.id=t+"-"+o,c.setAttribute("role","option");let u=this.optionClass(i);u&&(c.className=u);for(let e of this.optionContent){let t=e(i,this.view.state,this.view,s);t&&c.appendChild(t)}}return n.from&&r.classList.add("cm-completionListIncompleteTop"),n.to<e.length&&r.classList.add("cm-completionListIncompleteBottom"),r}destroyInfo(){this.info&&(this.infoDestroy&&this.infoDestroy(),this.info.remove(),this.info=null)}destroy(){this.destroyInfo()}}function zc(e,t){return n=>new Uc(n,e,t)}function jc(e){return 100*(e.boost||0)+(e.apply?10:0)+(e.info?5:0)+(e.type?1:0)}class Wc{constructor(e,t,n,r,a,o){this.options=e,this.attrs=t,this.tooltip=n,this.timestamp=r,this.selected=a,this.disabled=o}setSelected(e,t){return e==this.selected||e>=this.options.length?this:new Wc(this.options,Hc(t,e),this.tooltip,this.timestamp,e,this.disabled)}static build(e,t,n,r,a,o){if(r&&!o&&e.some((e=>e.isPending)))return r.setDisabled();let i=function(e,t){let n=[],r=null,a=e=>{n.push(e);let{section:t}=e.completion;if(t){r||(r=[]);let e="string"==typeof t?t:t.name;r.some((t=>t.name==e))||r.push("string"==typeof t?{name:e}:t)}},o=t.facet(Lc);for(let r of e)if(r.hasResult()){let e=r.result.getMatch;if(!1===r.result.filter)for(let t of r.result.options)a(new Cc(t,r.source,e?e(t):[],1e9-n.length));else{let n,i=t.sliceDoc(r.from,r.to),s=o.filterStrict?new Ic(i):new Nc(i);for(let t of r.result.options)if(n=s.match(t.label)){let o=t.displayLabel?e?e(t,n.matched):[]:n.matched;a(new Cc(t,r.source,o,n.score+(t.boost||0)))}}}if(r){let e=Object.create(null),t=0,a=(e,t)=>{var n,r;return(null!==(n=e.rank)&&void 0!==n?n:1e9)-(null!==(r=t.rank)&&void 0!==r?r:1e9)||(e.name<t.name?-1:1)};for(let n of r.sort(a))t-=1e5,e[n.name]=t;for(let t of n){let{section:n}=t.completion;n&&(t.score+=e["string"==typeof n?n:n.name])}}let i=[],s=null,l=o.compareCompletions;for(let e of n.sort(((e,t)=>t.score-e.score||l(e.completion,t.completion)))){let t=e.completion;!s||s.label!=t.label||s.detail!=t.detail||null!=s.type&&null!=t.type&&s.type!=t.type||s.apply!=t.apply||s.boost!=t.boost?i.push(e):jc(e.completion)>jc(s)&&(i[i.length-1]=e),s=e.completion}return i}(e,t);if(!i.length)return r&&e.some((e=>e.isPending))?r.setDisabled():null;let s=t.facet(Lc).selectOnOpen?0:-1;if(r&&r.selected!=s&&-1!=r.selected){let e=r.options[r.selected].completion;for(let t=0;t<i.length;t++)if(i[t].completion==e){s=t;break}}return new Wc(i,Hc(n,s),{pos:e.reduce(((e,t)=>t.hasResult()?Math.min(e,t.from):e),1e8),create:ru,above:a.aboveCursor},r?r.timestamp:Date.now(),s,!1)}map(e){return new Wc(this.options,this.attrs,Object.assign(Object.assign({},this.tooltip),{pos:e.mapPos(this.tooltip.pos)}),this.timestamp,this.selected,this.disabled)}setDisabled(){return new Wc(this.options,this.attrs,this.tooltip,this.timestamp,this.selected,!0)}}class Xc{constructor(e,t,n){this.active=e,this.id=t,this.open=n}static start(){return new Xc(Zc,"cm-ac-"+Math.floor(2e6*Math.random()).toString(36),null)}update(e){let{state:t}=e,n=t.facet(Lc),r=(n.override||t.languageDataAt("autocomplete",Tc(t)).map(Dc)).map((t=>(this.active.find((e=>e.source==t))||new Kc(t,this.active.some((e=>0!=e.state))?1:0)).update(e,n)));r.length==this.active.length&&r.every(((e,t)=>e==this.active[t]))&&(r=this.active);let a=this.open,o=e.effects.some((e=>e.is(Jc)));a&&e.docChanged&&(a=a.map(e.changes)),e.selection||r.some((t=>t.hasResult()&&e.changes.touchesRange(t.from,t.to)))||!function(e,t){if(e==t)return!0;for(let n=0,r=0;;){for(;n<e.length&&!e[n].hasResult();)n++;for(;r<t.length&&!t[r].hasResult();)r++;let a=n==e.length,o=r==t.length;if(a||o)return a==o;if(e[n++].result!=t[r++].result)return!1}}(r,this.active)||o?a=Wc.build(r,t,this.id,a,n,o):a&&a.disabled&&!r.some((e=>e.isPending))&&(a=null),!a&&r.every((e=>!e.isPending))&&r.some((e=>e.hasResult()))&&(r=r.map((e=>e.hasResult()?new Kc(e.source,0):e)));for(let t of e.effects)t.is(eu)&&(a=a&&a.setSelected(t.value,this.id));return r==this.active&&a==this.open?this:new Xc(r,this.id,a)}get tooltip(){return this.open?this.open.tooltip:null}get attrs(){return this.open?this.open.attrs:this.active.length?qc:Vc}}const qc={"aria-autocomplete":"list"},Vc={};function Hc(e,t){let n={"aria-autocomplete":"list","aria-haspopup":"listbox","aria-controls":e};return t>-1&&(n["aria-activedescendant"]=e+"-"+t),n}const Zc=[];function Gc(e,t){if(e.isUserEvent("input.complete")){let n=e.annotation(Rc);if(n&&t.activateOnCompletion(n))return 12}let n=e.isUserEvent("input.type");return n&&t.activateOnTyping?5:n?1:e.isUserEvent("delete.backward")?2:e.selection?8:e.docChanged?16:0}class Kc{constructor(e,t,n=!1){this.source=e,this.state=t,this.explicit=n}hasResult(){return!1}get isPending(){return 1==this.state}update(e,t){let n=Gc(e,t),r=this;(8&n||16&n&&this.touches(e))&&(r=new Kc(r.source,0)),4&n&&0==r.state&&(r=new Kc(this.source,1)),r=r.updateFor(e,n);for(let t of e.effects)if(t.is(Mc))r=new Kc(r.source,1,t.value);else if(t.is($c))r=new Kc(r.source,0);else if(t.is(Jc))for(let e of t.value)e.source==r.source&&(r=e);return r}updateFor(e,t){return this.map(e.changes)}map(e){return this}touches(e){return e.changes.touchesRange(Tc(e.state))}}class Yc extends Kc{constructor(e,t,n,r,a,o){super(e,3,t),this.limit=n,this.result=r,this.from=a,this.to=o}hasResult(){return!0}updateFor(e,t){var n;if(!(3&t))return this.map(e.changes);let r=this.result;r.map&&!e.changes.empty&&(r=r.map(r,e.changes));let a=e.changes.mapPos(this.from),o=e.changes.mapPos(this.to,1),i=Tc(e.state);if(i>o||!r||2&t&&(Tc(e.startState)==this.from||i<this.limit))return new Kc(this.source,4&t?1:0);let s=e.changes.mapPos(this.limit);return function(e,t,n,r){if(!e)return!1;let a=t.sliceDoc(n,r);return"function"==typeof e?e(a,n,r,t):Ac(e,!0).test(a)}(r.validFor,e.state,a,o)?new Yc(this.source,this.explicit,s,r,a,o):r.update&&(r=r.update(r,a,o,new Ec(e.state,i,!1)))?new Yc(this.source,this.explicit,s,r,r.from,null!==(n=r.to)&&void 0!==n?n:Tc(e.state)):new Kc(this.source,1,this.explicit)}map(e){return e.empty?this:(this.result.map?this.result.map(this.result,e):this.result)?new Yc(this.source,this.explicit,e.mapPos(this.limit),this.result,e.mapPos(this.from),e.mapPos(this.to,1)):new Kc(this.source,0)}touches(e){return e.changes.touchesRange(this.from,this.to)}}const Jc=fe.define({map:(e,t)=>e.map((e=>e.map(t)))}),eu=fe.define(),tu=Z.define({create:()=>Xc.start(),update:(e,t)=>e.update(t),provide:e=>[$o.from(e,(e=>e.tooltip)),Ya.contentAttributes.from(e,(e=>e.attrs))]});function nu(e,t){const n=t.completion.apply||t.completion.label;let r=e.state.field(tu).active.find((e=>e.source==t.source));return r instanceof Yc&&("string"==typeof n?e.dispatch(Object.assign(Object.assign({},function(e,t,n,r){let{main:a}=e.selection,o=n-a.from,i=r-a.from;return Object.assign(Object.assign({},e.changeByRange((s=>{if(s!=a&&n!=r&&e.sliceDoc(s.from+o,s.from+i)!=e.sliceDoc(n,r))return{range:s};let l=e.toText(t);return{changes:{from:s.from+o,to:r==a.from?s.to:s.from+i,insert:l},range:F.cursor(s.from+o+l.length)}}))),{scrollIntoView:!0,userEvent:"input.complete"})}(e.state,n,r.from,r.to)),{annotations:Rc.of(t.completion)})):n(e,t.completion,r.from,r.to),!0)}const ru=zc(tu,nu);function au(e,t="option"){return n=>{let r=n.state.field(tu,!1);if(!r||!r.open||r.open.disabled||Date.now()-r.open.timestamp<n.state.facet(Lc).interactionDelay)return!1;let a,o=1;"page"==t&&(a=Uo(n,r.open.tooltip))&&(o=Math.max(2,Math.floor(a.dom.offsetHeight/a.dom.querySelector("li").offsetHeight)-1));let{length:i}=r.open.options,s=r.open.selected>-1?r.open.selected+o*(e?1:-1):e?0:i-1;return s<0?s="page"==t?0:i-1:s>=i&&(s="page"==t?i-1:0),n.dispatch({effects:eu.of(s)}),!0}}const ou=e=>!!e.state.field(tu,!1)&&(e.dispatch({effects:Mc.of(!0)}),!0);class iu{constructor(e,t){this.active=e,this.context=t,this.time=Date.now(),this.updates=[],this.done=void 0}}const su=er.fromClass(class{constructor(e){this.view=e,this.debounceUpdate=-1,this.running=[],this.debounceAccept=-1,this.pendingStart=!1,this.composing=0;for(let t of e.state.field(tu).active)t.isPending&&this.startQuery(t)}update(e){let t=e.state.field(tu),n=e.state.facet(Lc);if(!e.selectionSet&&!e.docChanged&&e.startState.field(tu)==t)return;let r=e.transactions.some((e=>{let t=Gc(e,n);return 8&t||(e.selection||e.docChanged)&&!(3&t)}));for(let t=0;t<this.running.length;t++){let n=this.running[t];if(r||n.context.abortOnDocChange&&e.docChanged||n.updates.length+e.transactions.length>50&&Date.now()-n.time>1e3){for(let e of n.context.abortListeners)try{e()}catch(e){Gn(this.view.state,e)}n.context.abortListeners=null,this.running.splice(t--,1)}else n.updates.push(...e.transactions)}this.debounceUpdate>-1&&clearTimeout(this.debounceUpdate),e.transactions.some((e=>e.effects.some((e=>e.is(Mc)))))&&(this.pendingStart=!0);let a=this.pendingStart?50:n.activateOnTypingDelay;if(this.debounceUpdate=t.active.some((e=>e.isPending&&!this.running.some((t=>t.active.source==e.source))))?setTimeout((()=>this.startUpdate()),a):-1,0!=this.composing)for(let t of e.transactions)t.isUserEvent("input.type")?this.composing=2:2==this.composing&&t.selection&&(this.composing=3)}startUpdate(){this.debounceUpdate=-1,this.pendingStart=!1;let{state:e}=this.view,t=e.field(tu);for(let e of t.active)e.isPending&&!this.running.some((t=>t.active.source==e.source))&&this.startQuery(e);this.running.length&&t.open&&t.open.disabled&&(this.debounceAccept=setTimeout((()=>this.accept()),this.view.state.facet(Lc).updateSyncTime))}startQuery(e){let{state:t}=this.view,n=Tc(t),r=new Ec(t,n,e.explicit,this.view),a=new iu(e,r);this.running.push(a),Promise.resolve(e.source(r)).then((e=>{a.context.aborted||(a.done=e||null,this.scheduleAccept())}),(e=>{this.view.dispatch({effects:$c.of(null)}),Gn(this.view.state,e)}))}scheduleAccept(){this.running.every((e=>void 0!==e.done))?this.accept():this.debounceAccept<0&&(this.debounceAccept=setTimeout((()=>this.accept()),this.view.state.facet(Lc).updateSyncTime))}accept(){var e;this.debounceAccept>-1&&clearTimeout(this.debounceAccept),this.debounceAccept=-1;let t=[],n=this.view.state.facet(Lc),r=this.view.state.field(tu);for(let a=0;a<this.running.length;a++){let o=this.running[a];if(void 0===o.done)continue;if(this.running.splice(a--,1),o.done){let r=Tc(o.updates.length?o.updates[0].startState:this.view.state),a=Math.min(r,o.done.from+(o.active.explicit?0:1)),i=new Yc(o.active.source,o.active.explicit,a,o.done,o.done.from,null!==(e=o.done.to)&&void 0!==e?e:r);for(let e of o.updates)i=i.update(e,n);if(i.hasResult()){t.push(i);continue}}let i=r.active.find((e=>e.source==o.active.source));if(i&&i.isPending)if(null==o.done){let e=new Kc(o.active.source,0);for(let t of o.updates)e=e.update(t,n);e.isPending||t.push(e)}else this.startQuery(i)}(t.length||r.open&&r.open.disabled)&&this.view.dispatch({effects:Jc.of(t)})}},{eventHandlers:{blur(e){let t=this.view.state.field(tu,!1);if(t&&t.tooltip&&this.view.state.facet(Lc).closeOnBlur){let n=t.open&&Uo(this.view,t.open.tooltip);n&&n.dom.contains(e.relatedTarget)||setTimeout((()=>this.view.dispatch({effects:$c.of(null)})),10)}},compositionstart(){this.composing=1},compositionend(){3==this.composing&&setTimeout((()=>this.view.dispatch({effects:Mc.of(!1)})),20),this.composing=0}}}),lu="object"==typeof navigator&&/Win/.test(navigator.platform),cu=K.highest(Ya.domEventHandlers({keydown(e,t){let n=t.state.field(tu,!1);if(!n||!n.open||n.open.disabled||n.open.selected<0||e.key.length>1||e.ctrlKey&&(!lu||!e.altKey)||e.metaKey)return!1;let r=n.open.options[n.open.selected],a=n.active.find((e=>e.source==r.source)),o=r.completion.commitCharacters||a.result.commitCharacters;return o&&o.indexOf(e.key)>-1&&nu(t,r),!1}})),uu=Ya.baseTheme({".cm-tooltip.cm-tooltip-autocomplete":{"& > ul":{fontFamily:"monospace",whiteSpace:"nowrap",overflow:"hidden auto",maxWidth_fallback:"700px",maxWidth:"min(700px, 95vw)",minWidth:"250px",maxHeight:"10em",height:"100%",listStyle:"none",margin:0,padding:0,"& > li, & > completion-section":{padding:"1px 3px",lineHeight:1.2},"& > li":{overflowX:"hidden",textOverflow:"ellipsis",cursor:"pointer"},"& > completion-section":{display:"list-item",borderBottom:"1px solid silver",paddingLeft:"0.5em",opacity:.7}}},"&light .cm-tooltip-autocomplete ul li[aria-selected]":{background:"#17c",color:"white"},"&light .cm-tooltip-autocomplete-disabled ul li[aria-selected]":{background:"#777"},"&dark .cm-tooltip-autocomplete ul li[aria-selected]":{background:"#347",color:"white"},"&dark .cm-tooltip-autocomplete-disabled ul li[aria-selected]":{background:"#444"},".cm-completionListIncompleteTop:before, .cm-completionListIncompleteBottom:after":{content:'"···"',opacity:.5,display:"block",textAlign:"center"},".cm-tooltip.cm-completionInfo":{position:"absolute",padding:"3px 9px",width:"max-content",maxWidth:"400px",boxSizing:"border-box",whiteSpace:"pre-line"},".cm-completionInfo.cm-completionInfo-left":{right:"100%"},".cm-completionInfo.cm-completionInfo-right":{left:"100%"},".cm-completionInfo.cm-completionInfo-left-narrow":{right:"30px"},".cm-completionInfo.cm-completionInfo-right-narrow":{left:"30px"},"&light .cm-snippetField":{backgroundColor:"#00000022"},"&dark .cm-snippetField":{backgroundColor:"#ffffff22"},".cm-snippetFieldPosition":{verticalAlign:"text-top",width:0,height:"1.15em",display:"inline-block",margin:"0 -0.7px -.7em",borderLeft:"1.4px dotted #888"},".cm-completionMatchedText":{textDecoration:"underline"},".cm-completionDetail":{marginLeft:"0.5em",fontStyle:"italic"},".cm-completionIcon":{fontSize:"90%",width:".8em",display:"inline-block",textAlign:"center",paddingRight:".6em",opacity:"0.6",boxSizing:"content-box"},".cm-completionIcon-function, .cm-completionIcon-method":{"&:after":{content:"'ƒ'"}},".cm-completionIcon-class":{"&:after":{content:"'○'"}},".cm-completionIcon-interface":{"&:after":{content:"'◌'"}},".cm-completionIcon-variable":{"&:after":{content:"'𝑥'"}},".cm-completionIcon-constant":{"&:after":{content:"'𝐶'"}},".cm-completionIcon-type":{"&:after":{content:"'𝑡'"}},".cm-completionIcon-enum":{"&:after":{content:"'∪'"}},".cm-completionIcon-property":{"&:after":{content:"'□'"}},".cm-completionIcon-keyword":{"&:after":{content:"'🔑︎'"}},".cm-completionIcon-namespace":{"&:after":{content:"'▢'"}},".cm-completionIcon-text":{"&:after":{content:"'abc'",fontSize:"50%",verticalAlign:"middle"}}});class du{constructor(e,t,n,r){this.field=e,this.line=t,this.from=n,this.to=r}}class pu{constructor(e,t,n){this.field=e,this.from=t,this.to=n}map(e){let t=e.mapPos(this.from,-1,A.TrackDel),n=e.mapPos(this.to,1,A.TrackDel);return null==t||null==n?null:new pu(this.field,t,n)}}class hu{constructor(e,t){this.lines=e,this.fieldPositions=t}instantiate(e,t){let n=[],r=[t],a=e.doc.lineAt(t),o=/^\s*/.exec(a.text)[0];for(let a of this.lines){if(n.length){let n=o,i=/^\t*/.exec(a)[0].length;for(let t=0;t<i;t++)n+=e.facet(Ys);r.push(t+n.length-i),a=n+a.slice(i)}n.push(a),t+=a.length+1}let i=this.fieldPositions.map((e=>new pu(e.field,r[e.line]+e.from,r[e.line]+e.to)));return{text:n,ranges:i}}static parse(e){let t,n=[],r=[],a=[];for(let o of e.split(/\r\n?|\n/)){for(;t=/[#$]\{(?:(\d+)(?::([^}]*))?|((?:\\[{}]|[^}])*))\}/.exec(o);){let e=t[1]?+t[1]:null,i=t[2]||t[3]||"",s=-1,l=i.replace(/\\[{}]/g,(e=>e[1]));for(let t=0;t<n.length;t++)(null!=e?n[t].seq==e:l&&n[t].name==l)&&(s=t);if(s<0){let t=0;for(;t<n.length&&(null==e||null!=n[t].seq&&n[t].seq<e);)t++;n.splice(t,0,{seq:e,name:l}),s=t;for(let e of a)e.field>=s&&e.field++}a.push(new du(s,r.length,t.index,t.index+l.length)),o=o.slice(0,t.index)+i+o.slice(t.index+t[0].length)}o=o.replace(/\\([{}])/g,((e,t,n)=>{for(let e of a)e.line==r.length&&e.from>n&&(e.from--,e.to--);return t})),r.push(o)}return new hu(r,a)}}let fu=rn.widget({widget:new class extends tn{toDOM(){let e=document.createElement("span");return e.className="cm-snippetFieldPosition",e}ignoreEvent(){return!1}}}),mu=rn.mark({class:"cm-snippetField"});class gu{constructor(e,t){this.ranges=e,this.active=t,this.deco=rn.set(e.map((e=>(e.from==e.to?fu:mu).range(e.from,e.to))))}map(e){let t=[];for(let n of this.ranges){let r=n.map(e);if(!r)return null;t.push(r)}return new gu(t,this.active)}selectionInsideField(e){return e.ranges.every((e=>this.ranges.some((t=>t.field==this.active&&t.from<=e.from&&t.to>=e.to))))}}const vu=fe.define({map:(e,t)=>e&&e.map(t)}),bu=fe.define(),yu=Z.define({create:()=>null,update(e,t){for(let n of t.effects){if(n.is(vu))return n.value;if(n.is(bu)&&e)return new gu(e.ranges,n.value)}return e&&t.docChanged&&(e=e.map(t.changes)),e&&t.selection&&!e.selectionInsideField(t.selection)&&(e=null),e},provide:e=>Ya.decorations.from(e,(e=>e?e.deco:rn.none))});function wu(e,t){return F.create(e.filter((e=>e.field==t)).map((e=>F.range(e.from,e.to))))}function Ou(e){let t=hu.parse(e);return(e,n,r,a)=>{let{text:o,ranges:i}=t.instantiate(e.state,r),{main:s}=e.state.selection,l={changes:{from:r,to:a==s.from?s.to:a,insert:f.of(o)},scrollIntoView:!0,annotations:n?[Rc.of(n),me.userEvent.of("input.complete")]:void 0};if(i.length&&(l.selection=wu(i,0)),i.some((e=>e.field>0))){let t=new gu(i,0),n=l.effects=[vu.of(t)];void 0===e.state.field(yu,!1)&&n.push(fe.appendConfig.of([yu,Su,Cu,uu]))}e.dispatch(e.state.update(l))}}function ku(e){return({state:t,dispatch:n})=>{let r=t.field(yu,!1);if(!r||e<0&&0==r.active)return!1;let a=r.active+e,o=e>0&&!r.ranges.some((t=>t.field==a+e));return n(t.update({selection:wu(r.ranges,a),effects:vu.of(o?null:new gu(r.ranges,a)),scrollIntoView:!0})),!0}}const xu=[{key:"Tab",run:ku(1),shift:ku(-1)},{key:"Escape",run:({state:e,dispatch:t})=>!!e.field(yu,!1)&&(t(e.update({effects:vu.of(null)})),!0)}],Eu=z.define({combine:e=>e.length?e[0]:xu}),Su=K.highest(io.compute([Eu],(e=>e.facet(Eu))));function _u(e,t){return Object.assign(Object.assign({},t),{apply:Ou(e)})}const Cu=Ya.domEventHandlers({mousedown(e,t){let n,r=t.state.field(yu,!1);if(!r||null==(n=t.posAtCoords({x:e.clientX,y:e.clientY})))return!1;let a=r.ranges.find((e=>e.from<=n&&e.to>=n));return!(!a||a.field==r.active||(t.dispatch({selection:wu(r.ranges,a.field),effects:vu.of(r.ranges.some((e=>e.field>a.field))?new gu(r.ranges,a.field):null),scrollIntoView:!0}),0))}}),Tu={brackets:["(","[","{","'",'"'],before:")]}:;>",stringPrefixes:[]},Au=fe.define({map(e,t){let n=t.mapPos(e,-1,A.TrackAfter);return null==n?void 0:n}}),Ru=new class extends Ce{};Ru.startSide=1,Ru.endSide=-1;const Pu=Z.define({create:()=>Pe.empty,update(e,t){if(e=e.map(t.changes),t.selection){let n=t.state.doc.lineAt(t.selection.main.head);e=e.update({filter:e=>e>=n.from&&e<=n.to})}for(let n of t.effects)n.is(Au)&&(e=e.update({add:[Ru.range(n.value,n.value+1)]}));return e}}),Du="()[]{}<>«»»«［］｛｝";function Mu(e){for(let t=0;t<16;t+=2)if(Du.charCodeAt(t)==e)return Du.charAt(t+1);return _(e<128?e:e+1)}function $u(e,t){return e.languageDataAt("closeBrackets",t)[0]||Tu}const Nu="object"==typeof navigator&&/Android\b/.test(navigator.userAgent),Iu=Ya.inputHandler.of(((e,t,n,r)=>{if((Nu?e.composing:e.compositionStarted)||e.state.readOnly)return!1;let a=e.state.selection.main;if(r.length>2||2==r.length&&1==C(S(r,0))||t!=a.from||n!=a.to)return!1;let o=function(e,t){let n=$u(e,e.selection.main.head),r=n.brackets||Tu.brackets;for(let a of r){let o=Mu(S(a,0));if(t==a)return o==a?zu(e,a,r.indexOf(a+a+a)>-1,n):Qu(e,a,o,n.before||Tu.before);if(t==o&&Bu(e,e.selection.main.from))return Uu(e,0,o)}return null}(e.state,r);return!!o&&(e.dispatch(o),!0)})),Lu=[{key:"Backspace",run:({state:e,dispatch:t})=>{if(e.readOnly)return!1;let n=$u(e,e.selection.main.head).brackets||Tu.brackets,r=null,a=e.changeByRange((t=>{if(t.empty){let r=function(e,t){let n=e.sliceString(t-2,t);return C(S(n,0))==n.length?n:n.slice(1)}(e.doc,t.head);for(let a of n)if(a==r&&Fu(e.doc,t.head)==Mu(S(a,0)))return{changes:{from:t.head-a.length,to:t.head+a.length},range:F.cursor(t.head-a.length)}}return{range:r=t}}));return r||t(e.update(a,{scrollIntoView:!0,userEvent:"delete.backward"})),!r}}];function Bu(e,t){let n=!1;return e.field(Pu).between(0,e.doc.length,(e=>{e==t&&(n=!0)})),n}function Fu(e,t){let n=e.sliceString(t,t+2);return n.slice(0,C(S(n,0)))}function Qu(e,t,n,r){let a=null,o=e.changeByRange((o=>{if(!o.empty)return{changes:[{insert:t,from:o.from},{insert:n,from:o.to}],effects:Au.of(o.to+t.length),range:F.range(o.anchor+t.length,o.head+t.length)};let i=Fu(e.doc,o.head);return!i||/\s/.test(i)||r.indexOf(i)>-1?{changes:{insert:t+n,from:o.head},effects:Au.of(o.head+t.length),range:F.cursor(o.head+t.length)}:{range:a=o}}));return a?null:e.update(o,{scrollIntoView:!0,userEvent:"input.type"})}function Uu(e,t,n){let r=null,a=e.changeByRange((t=>t.empty&&Fu(e.doc,t.head)==n?{changes:{from:t.head,to:t.head+n.length,insert:n},range:F.cursor(t.head+n.length)}:r={range:t}));return r?null:e.update(a,{scrollIntoView:!0,userEvent:"input.type"})}function zu(e,t,n,r){let a=r.stringPrefixes||Tu.stringPrefixes,o=null,i=e.changeByRange((r=>{if(!r.empty)return{changes:[{insert:t,from:r.from},{insert:t,from:r.to}],effects:Au.of(r.to+t.length),range:F.range(r.anchor+t.length,r.head+t.length)};let i,s=r.head,l=Fu(e.doc,s);if(l==t){if(ju(e,s))return{changes:{insert:t+t,from:s},effects:Au.of(s+t.length),range:F.cursor(s+t.length)};if(Bu(e,s)){let r=n&&e.sliceDoc(s,s+3*t.length)==t+t+t?t+t+t:t;return{changes:{from:s,to:s+r.length,insert:r},range:F.cursor(s+r.length)}}}else{if(n&&e.sliceDoc(s-2*t.length,s)==t+t&&(i=Wu(e,s-2*t.length,a))>-1&&ju(e,i))return{changes:{insert:t+t+t+t,from:s},effects:Au.of(s+t.length),range:F.cursor(s+t.length)};if(e.charCategorizer(s)(l)!=ke.Word&&Wu(e,s,a)>-1&&!function(e,t,n,r){let a=Qs(e).resolveInner(t,-1),o=r.reduce(((e,t)=>Math.max(e,t.length)),0);for(let i=0;i<5;i++){let i=e.sliceDoc(a.from,Math.min(a.to,a.from+n.length+o)),s=i.indexOf(n);if(!s||s>-1&&r.indexOf(i.slice(0,s))>-1){let t=a.firstChild;for(;t&&t.from==a.from&&t.to-t.from>n.length+s;){if(e.sliceDoc(t.to-n.length,t.to)==n)return!1;t=t.firstChild}return!0}let l=a.to==t&&a.parent;if(!l)break;a=l}return!1}(e,s,t,a))return{changes:{insert:t+t,from:s},effects:Au.of(s+t.length),range:F.cursor(s+t.length)}}return{range:o=r}}));return o?null:e.update(i,{scrollIntoView:!0,userEvent:"input.type"})}function ju(e,t){let n=Qs(e).resolveInner(t+1);return n.parent&&n.from==t}function Wu(e,t,n){let r=e.charCategorizer(t);if(r(e.sliceDoc(t-1,t))!=ke.Word)return t;for(let a of n){let n=t-a.length;if(e.sliceDoc(n,t)==a&&r(e.sliceDoc(n-1,n))!=ke.Word)return n}return-1}const Xu=[{key:"Ctrl-Space",run:ou},{mac:"Alt-`",run:ou},{key:"Escape",run:e=>{let t=e.state.field(tu,!1);return!(!t||!t.active.some((e=>0!=e.state))||(e.dispatch({effects:$c.of(null)}),0))}},{key:"ArrowDown",run:au(!0)},{key:"ArrowUp",run:au(!1)},{key:"PageDown",run:au(!0,"page")},{key:"PageUp",run:au(!1,"page")},{key:"Enter",run:e=>{let t=e.state.field(tu,!1);return!(e.state.readOnly||!t||!t.open||t.open.selected<0||t.open.disabled||Date.now()-t.open.timestamp<e.state.facet(Lc).interactionDelay)&&nu(e,t.open.options[t.open.selected])}}],qu=K.highest(io.computeN([Lc],(e=>e.facet(Lc).defaultKeymap?[Xu]:[]))),Vu=de.define(),Hu=z.define();class Zu{constructor(e,t,n,r,a){this.changes=e,this.effects=t,this.mapped=n,this.startSelection=r,this.selectionsAfter=a}setSelAfter(e){return new Zu(this.changes,this.effects,this.mapped,this.startSelection,e)}toJSON(){var e,t,n;return{changes:null===(e=this.changes)||void 0===e?void 0:e.toJSON(),mapped:null===(t=this.mapped)||void 0===t?void 0:t.toJSON(),startSelection:null===(n=this.startSelection)||void 0===n?void 0:n.toJSON(),selectionsAfter:this.selectionsAfter.map((e=>e.toJSON()))}}static fromJSON(e){return new Zu(e.changes&&P.fromJSON(e.changes),[],e.mapped&&R.fromJSON(e.mapped),e.startSelection&&F.fromJSON(e.startSelection),e.selectionsAfter.map(F.fromJSON))}static fromTransaction(e,t){let n=Yu;for(let t of e.startState.facet(Hu)){let r=t(e);r.length&&(n=n.concat(r))}return!n.length&&e.changes.empty?null:new Zu(e.changes.invert(e.startState.doc),n,void 0,t||e.startState.selection,Yu)}static selection(e){return new Zu(void 0,Yu,void 0,void 0,e)}}function Gu(e,t,n,r){let a=t+1>n+20?t-n-1:0,o=e.slice(a,t);return o.push(r),o}function Ku(e,t){return e.length?t.length?e.concat(t):e:t}const Yu=[];function Ju(e){let t=e[e.length-1],n=e.slice();return n[e.length-1]=t.setSelAfter(t.selectionsAfter.slice(0,t.selectionsAfter.length-1)),n}function ed(e,t){if(!e.length)return e;let n=e.length,r=Yu;for(;n;){let a=td(e[n-1],t,r);if(a.changes&&!a.changes.empty||a.effects.length){let t=e.slice(0,n);return t[n-1]=a,t}t=a.mapped,n--,r=a.selectionsAfter}return r.length?[Zu.selection(r)]:Yu}function td(e,t,n){let r=Ku(e.selectionsAfter.length?e.selectionsAfter.map((e=>e.map(t))):Yu,n);if(!e.changes)return Zu.selection(r);let a=e.changes.map(t),o=t.mapDesc(e.changes,!0),i=e.mapped?e.mapped.composeDesc(o):o;return new Zu(a,fe.mapEffects(e.effects,t),i,e.startSelection.map(o),r)}const nd=/^(input\.type|delete)($|\.)/;class rd{constructor(e,t,n=0,r=void 0){this.done=e,this.undone=t,this.prevTime=n,this.prevUserEvent=r}isolate(){return this.prevTime?new rd(this.done,this.undone):this}addChanges(e,t,n,r,a){let o=this.done,i=o[o.length-1];return o=i&&i.changes&&!i.changes.empty&&e.changes&&(!n||nd.test(n))&&(!i.selectionsAfter.length&&t-this.prevTime<r.newGroupDelay&&r.joinToEvent(a,function(e,t){let n=[],r=!1;return e.iterChangedRanges(((e,t)=>n.push(e,t))),t.iterChangedRanges(((e,t,a,o)=>{for(let e=0;e<n.length;){let t=n[e++],i=n[e++];o>=t&&a<=i&&(r=!0)}})),r}(i.changes,e.changes))||"input.type.compose"==n)?Gu(o,o.length-1,r.minDepth,new Zu(e.changes.compose(i.changes),Ku(fe.mapEffects(e.effects,i.changes),i.effects),i.mapped,i.startSelection,Yu)):Gu(o,o.length,r.minDepth,e),new rd(o,Yu,t,n)}addSelection(e,t,n,r){let a=this.done.length?this.done[this.done.length-1].selectionsAfter:Yu;return a.length>0&&t-this.prevTime<r&&n==this.prevUserEvent&&n&&/^select($|\.)/.test(n)&&(o=a[a.length-1],i=e,o.ranges.length==i.ranges.length&&0===o.ranges.filter(((e,t)=>e.empty!=i.ranges[t].empty)).length)?this:new rd(function(e,t){if(e.length){let n=e[e.length-1],r=n.selectionsAfter.slice(Math.max(0,n.selectionsAfter.length-200));return r.length&&r[r.length-1].eq(t)?e:(r.push(t),Gu(e,e.length-1,1e9,n.setSelAfter(r)))}return[Zu.selection([t])]}(this.done,e),this.undone,t,n);var o,i}addMapping(e){return new rd(ed(this.done,e),ed(this.undone,e),this.prevTime,this.prevUserEvent)}pop(e,t,n){let r=0==e?this.done:this.undone;if(0==r.length)return null;let a=r[r.length-1],o=a.selectionsAfter[0]||t.selection;if(n&&a.selectionsAfter.length)return t.update({selection:a.selectionsAfter[a.selectionsAfter.length-1],annotations:Vu.of({side:e,rest:Ju(r),selection:o}),userEvent:0==e?"select.undo":"select.redo",scrollIntoView:!0});if(a.changes){let n=1==r.length?Yu:r.slice(0,r.length-1);return a.mapped&&(n=ed(n,a.mapped)),t.update({changes:a.changes,selection:a.startSelection,effects:a.effects,annotations:Vu.of({side:e,rest:n,selection:o}),filter:!1,userEvent:0==e?"undo":"redo",scrollIntoView:!0})}return null}}rd.empty=new rd(Yu,Yu),"undefined"!=typeof Intl&&Intl.Segmenter;const ad=({state:e,dispatch:t})=>(t(e.update(e.replaceSelection(e.lineBreak),{scrollIntoView:!0,userEvent:"input"})),!0);function od(e,t){let n=-1;return e.changeByRange((r=>{let a=[];for(let o=r.from;o<=r.to;){let i=e.doc.lineAt(o);i.number>n&&(r.empty||r.to>i.from)&&(t(i,a,r),n=i.number),o=i.to+1}let o=e.changes(a);return{changes:a,range:F.range(o.mapPos(r.anchor,1),o.mapPos(r.head,1))}}))}const id={key:"Tab",run:({state:e,dispatch:t})=>!e.readOnly&&(t(e.update(od(e,((t,n)=>{n.push({from:t.from,insert:e.facet(Ys)})})),{userEvent:"input.indent"})),!0),shift:({state:e,dispatch:t})=>!e.readOnly&&(t(e.update(od(e,((t,n)=>{let r=/^\s*/.exec(t.text)[0];if(!r)return;let a=je(r,e.tabSize),o=0,i=el(e,Math.max(0,a-Js(e)));for(;o<r.length&&o<i.length&&r.charCodeAt(o)==i.charCodeAt(o);)o++;n.push({from:t.from+o,to:t.from+r.length,insert:i.slice(o)})})),{userEvent:"delete.dedent"})),!0)};class sd{constructor(e,t,n,r,a,o,i,s,l,c=0,u){this.p=e,this.stack=t,this.state=n,this.reducePos=r,this.pos=a,this.score=o,this.buffer=i,this.bufferBase=s,this.curContext=l,this.lookAhead=c,this.parent=u}toString(){return`[${this.stack.filter(((e,t)=>t%3==0)).concat(this.state)}]@${this.pos}${this.score?"!"+this.score:""}`}static start(e,t,n=0){let r=e.parser.context;return new sd(e,[],t,n,n,0,[],0,r?new ld(r,r.start):null,0,null)}get context(){return this.curContext?this.curContext.context:null}pushState(e,t){this.stack.push(this.state,t,this.bufferBase+this.buffer.length),this.state=e}reduce(e){var t;let n=e>>19,r=65535&e,{parser:a}=this.p,o=this.reducePos<this.pos-25;o&&this.setLookAhead(this.pos);let i=a.dynamicPrecedence(r);if(i&&(this.score+=i),0==n)return this.pushState(a.getGoto(this.state,r,!0),this.reducePos),r<a.minRepeatTerm&&this.storeNode(r,this.reducePos,this.reducePos,o?8:4,!0),void this.reduceContext(r,this.reducePos);let s=this.stack.length-3*(n-1)-(262144&e?6:0),l=s?this.stack[s-2]:this.p.ranges[0].from,c=this.reducePos-l;c>=2e3&&!(null===(t=this.p.parser.nodeSet.types[r])||void 0===t?void 0:t.isAnonymous)&&(l==this.p.lastBigReductionStart?(this.p.bigReductionCount++,this.p.lastBigReductionSize=c):this.p.lastBigReductionSize<c&&(this.p.bigReductionCount=1,this.p.lastBigReductionStart=l,this.p.lastBigReductionSize=c));let u=s?this.stack[s-1]:0,d=this.bufferBase+this.buffer.length-u;if(r<a.minRepeatTerm||131072&e){let e=a.stateFlag(this.state,1)?this.pos:this.reducePos;this.storeNode(r,l,e,d+4,!0)}if(262144&e)this.state=this.stack[s];else{let e=this.stack[s-3];this.state=a.getGoto(e,r,!0)}for(;this.stack.length>s;)this.stack.pop();this.reduceContext(r,l)}storeNode(e,t,n,r=4,a=!1){if(0==e&&(!this.stack.length||this.stack[this.stack.length-1]<this.buffer.length+this.bufferBase)){let e=this,r=this.buffer.length;if(0==r&&e.parent&&(r=e.bufferBase-e.parent.bufferBase,e=e.parent),r>0&&0==e.buffer[r-4]&&e.buffer[r-1]>-1){if(t==n)return;if(e.buffer[r-2]>=t)return void(e.buffer[r-2]=n)}}if(a&&this.pos!=n){let a=this.buffer.length;if(a>0&&0!=this.buffer[a-4]){let e=!1;for(let t=a;t>0&&this.buffer[t-2]>n;t-=4)if(this.buffer[t-1]>=0){e=!0;break}if(e)for(;a>0&&this.buffer[a-2]>n;)this.buffer[a]=this.buffer[a-4],this.buffer[a+1]=this.buffer[a-3],this.buffer[a+2]=this.buffer[a-2],this.buffer[a+3]=this.buffer[a-1],a-=4,r>4&&(r-=4)}this.buffer[a]=e,this.buffer[a+1]=t,this.buffer[a+2]=n,this.buffer[a+3]=r}else this.buffer.push(e,t,n,r)}shift(e,t,n,r){if(131072&e)this.pushState(65535&e,this.pos);else if(262144&e)this.pos=r,this.shiftContext(t,n),t<=this.p.parser.maxNode&&this.buffer.push(t,n,r,4);else{let a=e,{parser:o}=this.p;(r>this.pos||t<=o.maxNode)&&(this.pos=r,o.stateFlag(a,1)||(this.reducePos=r)),this.pushState(a,n),this.shiftContext(t,n),t<=o.maxNode&&this.buffer.push(t,n,r,4)}}apply(e,t,n,r){65536&e?this.reduce(e):this.shift(e,t,n,r)}useNode(e,t){let n=this.p.reused.length-1;(n<0||this.p.reused[n]!=e)&&(this.p.reused.push(e),n++);let r=this.pos;this.reducePos=this.pos=r+e.length,this.pushState(t,r),this.buffer.push(n,r,this.reducePos,-1),this.curContext&&this.updateContext(this.curContext.tracker.reuse(this.curContext.context,e,this,this.p.stream.reset(this.pos-e.length)))}split(){let e=this,t=e.buffer.length;for(;t>0&&e.buffer[t-2]>e.reducePos;)t-=4;let n=e.buffer.slice(t),r=e.bufferBase+t;for(;e&&r==e.bufferBase;)e=e.parent;return new sd(this.p,this.stack.slice(),this.state,this.reducePos,this.pos,this.score,n,r,this.curContext,this.lookAhead,e)}recoverByDelete(e,t){let n=e<=this.p.parser.maxNode;n&&this.storeNode(e,this.pos,t,4),this.storeNode(0,this.pos,t,n?8:4),this.pos=this.reducePos=t,this.score-=190}canShift(e){for(let t=new cd(this);;){let n=this.p.parser.stateSlot(t.state,4)||this.p.parser.hasAction(t.state,e);if(0==n)return!1;if(!(65536&n))return!0;t.reduce(n)}}recoverByInsert(e){if(this.stack.length>=300)return[];let t=this.p.parser.nextStates(this.state);if(t.length>8||this.stack.length>=120){let n=[];for(let r,a=0;a<t.length;a+=2)(r=t[a+1])!=this.state&&this.p.parser.hasAction(r,e)&&n.push(t[a],r);if(this.stack.length<120)for(let e=0;n.length<8&&e<t.length;e+=2){let r=t[e+1];n.some(((e,t)=>1&t&&e==r))||n.push(t[e],r)}t=n}let n=[];for(let e=0;e<t.length&&n.length<4;e+=2){let r=t[e+1];if(r==this.state)continue;let a=this.split();a.pushState(r,this.pos),a.storeNode(0,a.pos,a.pos,4,!0),a.shiftContext(t[e],this.pos),a.reducePos=this.pos,a.score-=200,n.push(a)}return n}forceReduce(){let{parser:e}=this.p,t=e.stateSlot(this.state,5);if(!(65536&t))return!1;if(!e.validAction(this.state,t)){let n=t>>19,r=65535&t,a=this.stack.length-3*n;if(a<0||e.getGoto(this.stack[a],r,!1)<0){let e=this.findForcedReduction();if(null==e)return!1;t=e}this.storeNode(0,this.pos,this.pos,4,!0),this.score-=100}return this.reducePos=this.pos,this.reduce(t),!0}findForcedReduction(){let{parser:e}=this.p,t=[],n=(r,a)=>{if(!t.includes(r))return t.push(r),e.allActions(r,(t=>{if(393216&t);else if(65536&t){let n=(t>>19)-a;if(n>1){let r=65535&t,a=this.stack.length-3*n;if(a>=0&&e.getGoto(this.stack[a],r,!1)>=0)return n<<19|65536|r}}else{let e=n(t,a+1);if(null!=e)return e}}))};return n(this.state,0)}forceAll(){for(;!this.p.parser.stateFlag(this.state,2);)if(!this.forceReduce()){this.storeNode(0,this.pos,this.pos,4,!0);break}return this}get deadEnd(){if(3!=this.stack.length)return!1;let{parser:e}=this.p;return 65535==e.data[e.stateSlot(this.state,1)]&&!e.stateSlot(this.state,4)}restart(){this.storeNode(0,this.pos,this.pos,4,!0),this.state=this.stack[0],this.stack.length=0}sameState(e){if(this.state!=e.state||this.stack.length!=e.stack.length)return!1;for(let t=0;t<this.stack.length;t+=3)if(this.stack[t]!=e.stack[t])return!1;return!0}get parser(){return this.p.parser}dialectEnabled(e){return this.p.parser.dialect.flags[e]}shiftContext(e,t){this.curContext&&this.updateContext(this.curContext.tracker.shift(this.curContext.context,e,this,this.p.stream.reset(t)))}reduceContext(e,t){this.curContext&&this.updateContext(this.curContext.tracker.reduce(this.curContext.context,e,this,this.p.stream.reset(t)))}emitContext(){let e=this.buffer.length-1;(e<0||-3!=this.buffer[e])&&this.buffer.push(this.curContext.hash,this.pos,this.pos,-3)}emitLookAhead(){let e=this.buffer.length-1;(e<0||-4!=this.buffer[e])&&this.buffer.push(this.lookAhead,this.pos,this.pos,-4)}updateContext(e){if(e!=this.curContext.context){let t=new ld(this.curContext.tracker,e);t.hash!=this.curContext.hash&&this.emitContext(),this.curContext=t}}setLookAhead(e){e>this.lookAhead&&(this.emitLookAhead(),this.lookAhead=e)}close(){this.curContext&&this.curContext.tracker.strict&&this.emitContext(),this.lookAhead>0&&this.emitLookAhead()}}class ld{constructor(e,t){this.tracker=e,this.context=t,this.hash=e.strict?e.hash(t):0}}class cd{constructor(e){this.start=e,this.state=e.state,this.stack=e.stack,this.base=this.stack.length}reduce(e){let t=65535&e,n=e>>19;0==n?(this.stack==this.start.stack&&(this.stack=this.stack.slice()),this.stack.push(this.state,0,0),this.base+=3):this.base-=3*(n-1);let r=this.start.p.parser.getGoto(this.stack[this.base-3],t,!0);this.state=r}}class ud{constructor(e,t,n){this.stack=e,this.pos=t,this.index=n,this.buffer=e.buffer,0==this.index&&this.maybeNext()}static create(e,t=e.bufferBase+e.buffer.length){return new ud(e,t,t-e.bufferBase)}maybeNext(){let e=this.stack.parent;null!=e&&(this.index=this.stack.bufferBase-e.bufferBase,this.stack=e,this.buffer=e.buffer)}get id(){return this.buffer[this.index-4]}get start(){return this.buffer[this.index-3]}get end(){return this.buffer[this.index-2]}get size(){return this.buffer[this.index-1]}next(){this.index-=4,this.pos-=4,0==this.index&&this.maybeNext()}fork(){return new ud(this.stack,this.pos,this.index)}}function dd(e,t=Uint16Array){if("string"!=typeof e)return e;let n=null;for(let r=0,a=0;r<e.length;){let o=0;for(;;){let t=e.charCodeAt(r++),n=!1;if(126==t){o=65535;break}t>=92&&t--,t>=34&&t--;let a=t-32;if(a>=46&&(a-=46,n=!0),o+=a,n)break;o*=46}n?n[a++]=o:n=new t(o)}return n}class pd{constructor(){this.start=-1,this.value=-1,this.end=-1,this.extended=-1,this.lookAhead=0,this.mask=0,this.context=0}}const hd=new pd;class fd{constructor(e,t){this.input=e,this.ranges=t,this.chunk="",this.chunkOff=0,this.chunk2="",this.chunk2Pos=0,this.next=-1,this.token=hd,this.rangeIndex=0,this.pos=this.chunkPos=t[0].from,this.range=t[0],this.end=t[t.length-1].to,this.readNext()}resolveOffset(e,t){let n=this.range,r=this.rangeIndex,a=this.pos+e;for(;a<n.from;){if(!r)return null;let e=this.ranges[--r];a-=n.from-e.to,n=e}for(;t<0?a>n.to:a>=n.to;){if(r==this.ranges.length-1)return null;let e=this.ranges[++r];a+=e.from-n.to,n=e}return a}clipPos(e){if(e>=this.range.from&&e<this.range.to)return e;for(let t of this.ranges)if(t.to>e)return Math.max(e,t.from);return this.end}peek(e){let t,n,r=this.chunkOff+e;if(r>=0&&r<this.chunk.length)t=this.pos+e,n=this.chunk.charCodeAt(r);else{let r=this.resolveOffset(e,1);if(null==r)return-1;if(t=r,t>=this.chunk2Pos&&t<this.chunk2Pos+this.chunk2.length)n=this.chunk2.charCodeAt(t-this.chunk2Pos);else{let e=this.rangeIndex,r=this.range;for(;r.to<=t;)r=this.ranges[++e];this.chunk2=this.input.chunk(this.chunk2Pos=t),t+this.chunk2.length>r.to&&(this.chunk2=this.chunk2.slice(0,r.to-t)),n=this.chunk2.charCodeAt(0)}}return t>=this.token.lookAhead&&(this.token.lookAhead=t+1),n}acceptToken(e,t=0){let n=t?this.resolveOffset(t,-1):this.pos;if(null==n||n<this.token.start)throw new RangeError("Token end out of bounds");this.token.value=e,this.token.end=n}acceptTokenTo(e,t){this.token.value=e,this.token.end=t}getChunk(){if(this.pos>=this.chunk2Pos&&this.pos<this.chunk2Pos+this.chunk2.length){let{chunk:e,chunkPos:t}=this;this.chunk=this.chunk2,this.chunkPos=this.chunk2Pos,this.chunk2=e,this.chunk2Pos=t,this.chunkOff=this.pos-this.chunkPos}else{this.chunk2=this.chunk,this.chunk2Pos=this.chunkPos;let e=this.input.chunk(this.pos),t=this.pos+e.length;this.chunk=t>this.range.to?e.slice(0,this.range.to-this.pos):e,this.chunkPos=this.pos,this.chunkOff=0}}readNext(){return this.chunkOff>=this.chunk.length&&(this.getChunk(),this.chunkOff==this.chunk.length)?this.next=-1:this.next=this.chunk.charCodeAt(this.chunkOff)}advance(e=1){for(this.chunkOff+=e;this.pos+e>=this.range.to;){if(this.rangeIndex==this.ranges.length-1)return this.setDone();e-=this.range.to-this.pos,this.range=this.ranges[++this.rangeIndex],this.pos=this.range.from}return this.pos+=e,this.pos>=this.token.lookAhead&&(this.token.lookAhead=this.pos+1),this.readNext()}setDone(){return this.pos=this.chunkPos=this.end,this.range=this.ranges[this.rangeIndex=this.ranges.length-1],this.chunk="",this.next=-1}reset(e,t){if(t?(this.token=t,t.start=e,t.lookAhead=e+1,t.value=t.extended=-1):this.token=hd,this.pos!=e){if(this.pos=e,e==this.end)return this.setDone(),this;for(;e<this.range.from;)this.range=this.ranges[--this.rangeIndex];for(;e>=this.range.to;)this.range=this.ranges[++this.rangeIndex];e>=this.chunkPos&&e<this.chunkPos+this.chunk.length?this.chunkOff=e-this.chunkPos:(this.chunk="",this.chunkOff=0),this.readNext()}return this}read(e,t){if(e>=this.chunkPos&&t<=this.chunkPos+this.chunk.length)return this.chunk.slice(e-this.chunkPos,t-this.chunkPos);if(e>=this.chunk2Pos&&t<=this.chunk2Pos+this.chunk2.length)return this.chunk2.slice(e-this.chunk2Pos,t-this.chunk2Pos);if(e>=this.range.from&&t<=this.range.to)return this.input.read(e,t);let n="";for(let r of this.ranges){if(r.from>=t)break;r.to>e&&(n+=this.input.read(Math.max(r.from,e),Math.min(r.to,t)))}return n}}class md{constructor(e,t){this.data=e,this.id=t}token(e,t){let{parser:n}=t.p;bd(this.data,e,t,this.id,n.data,n.tokenPrecTable)}}md.prototype.contextual=md.prototype.fallback=md.prototype.extend=!1;class gd{constructor(e,t,n){this.precTable=t,this.elseToken=n,this.data="string"==typeof e?dd(e):e}token(e,t){let n=e.pos,r=0;for(;;){let n=e.next<0,a=e.resolveOffset(1,1);if(bd(this.data,e,t,0,this.data,this.precTable),e.token.value>-1)break;if(null==this.elseToken)return;if(n||r++,null==a)break;e.reset(a,e.token)}r&&(e.reset(n,e.token),e.acceptToken(this.elseToken,r))}}gd.prototype.contextual=md.prototype.fallback=md.prototype.extend=!1;class vd{constructor(e,t={}){this.token=e,this.contextual=!!t.contextual,this.fallback=!!t.fallback,this.extend=!!t.extend}}function bd(e,t,n,r,a,o){let i=0,s=1<<r,{dialect:l}=n.p.parser;e:for(;s&e[i];){let n=e[i+1];for(let r=i+3;r<n;r+=2)if((e[r+1]&s)>0){let n=e[r];if(l.allows(n)&&(-1==t.token.value||t.token.value==n||wd(n,t.token.value,a,o))){t.acceptToken(n);break}}let r=t.next,c=0,u=e[i+2];if(!(t.next<0&&u>c&&65535==e[n+3*u-3])){for(;c<u;){let a=c+u>>1,o=n+a+(a<<1),s=e[o],l=e[o+1]||65536;if(r<s)u=a;else{if(!(r>=l)){i=e[o+2],t.advance();continue e}c=a+1}}break}i=e[n+3*u-1]}}function yd(e,t,n){for(let r,a=t;65535!=(r=e[a]);a++)if(r==n)return a-t;return-1}function wd(e,t,n,r){let a=yd(n,r,t);return a<0||yd(n,r,e)<a}const Od="undefined"!=typeof process&&process.env&&/\bparse\b/.test(process.env.LOG);let kd=null;function xd(e,t,n){let r=e.cursor(Si.IncludeAnonymous);for(r.moveTo(t);;)if(!(n<0?r.childBefore(t):r.childAfter(t)))for(;;){if((n<0?r.to<t:r.from>t)&&!r.type.isError)return n<0?Math.max(0,Math.min(r.to-1,t-25)):Math.min(e.length,Math.max(r.from+1,t+25));if(n<0?r.prevSibling():r.nextSibling())break;if(!r.parent())return n<0?0:e.length}}class Ed{constructor(e,t){this.fragments=e,this.nodeSet=t,this.i=0,this.fragment=null,this.safeFrom=-1,this.safeTo=-1,this.trees=[],this.start=[],this.index=[],this.nextFragment()}nextFragment(){let e=this.fragment=this.i==this.fragments.length?null:this.fragments[this.i++];if(e){for(this.safeFrom=e.openStart?xd(e.tree,e.from+e.offset,1)-e.offset:e.from,this.safeTo=e.openEnd?xd(e.tree,e.to+e.offset,-1)-e.offset:e.to;this.trees.length;)this.trees.pop(),this.start.pop(),this.index.pop();this.trees.push(e.tree),this.start.push(-e.offset),this.index.push(0),this.nextStart=this.safeFrom}else this.nextStart=1e9}nodeAt(e){if(e<this.nextStart)return null;for(;this.fragment&&this.safeTo<=e;)this.nextFragment();if(!this.fragment)return null;for(;;){let t=this.trees.length-1;if(t<0)return this.nextFragment(),null;let n=this.trees[t],r=this.index[t];if(r==n.children.length){this.trees.pop(),this.start.pop(),this.index.pop();continue}let a=n.children[r],o=this.start[t]+n.positions[r];if(o>e)return this.nextStart=o,null;if(a instanceof _i){if(o==e){if(o<this.safeFrom)return null;let e=o+a.length;if(e<=this.safeTo){let t=a.prop(bi.lookAhead);if(!t||e+t<this.fragment.to)return a}}this.index[t]++,o+a.length>=Math.max(this.safeFrom,e)&&(this.trees.push(a),this.start.push(o),this.index.push(0))}else this.index[t]++,this.nextStart=o+a.length}}}class Sd{constructor(e,t){this.stream=t,this.tokens=[],this.mainToken=null,this.actions=[],this.tokens=e.tokenizers.map((e=>new pd))}getActions(e){let t=0,n=null,{parser:r}=e.p,{tokenizers:a}=r,o=r.stateSlot(e.state,3),i=e.curContext?e.curContext.hash:0,s=0;for(let r=0;r<a.length;r++){if(!(1<<r&o))continue;let l=a[r],c=this.tokens[r];if((!n||l.fallback)&&((l.contextual||c.start!=e.pos||c.mask!=o||c.context!=i)&&(this.updateCachedToken(c,l,e),c.mask=o,c.context=i),c.lookAhead>c.end+25&&(s=Math.max(c.lookAhead,s)),0!=c.value)){let r=t;if(c.extended>-1&&(t=this.addActions(e,c.extended,c.end,t)),t=this.addActions(e,c.value,c.end,t),!l.extend&&(n=c,t>r))break}}for(;this.actions.length>t;)this.actions.pop();return s&&e.setLookAhead(s),n||e.pos!=this.stream.end||(n=new pd,n.value=e.p.parser.eofTerm,n.start=n.end=e.pos,t=this.addActions(e,n.value,n.end,t)),this.mainToken=n,this.actions}getMainToken(e){if(this.mainToken)return this.mainToken;let t=new pd,{pos:n,p:r}=e;return t.start=n,t.end=Math.min(n+1,r.stream.end),t.value=n==r.stream.end?r.parser.eofTerm:0,t}updateCachedToken(e,t,n){let r=this.stream.clipPos(n.pos);if(t.token(this.stream.reset(r,e),n),e.value>-1){let{parser:t}=n.p;for(let r=0;r<t.specialized.length;r++)if(t.specialized[r]==e.value){let a=t.specializers[r](this.stream.read(e.start,e.end),n);if(a>=0&&n.p.parser.dialect.allows(a>>1)){1&a?e.extended=a>>1:e.value=a>>1;break}}}else e.value=0,e.end=this.stream.clipPos(r+1)}putAction(e,t,n,r){for(let t=0;t<r;t+=3)if(this.actions[t]==e)return r;return this.actions[r++]=e,this.actions[r++]=t,this.actions[r++]=n,r}addActions(e,t,n,r){let{state:a}=e,{parser:o}=e.p,{data:i}=o;for(let e=0;e<2;e++)for(let s=o.stateSlot(a,e?2:1);;s+=3){if(65535==i[s]){if(1!=i[s+1]){0==r&&2==i[s+1]&&(r=this.putAction(Dd(i,s+2),t,n,r));break}s=Dd(i,s+2)}i[s]==t&&(r=this.putAction(Dd(i,s+1),t,n,r))}return r}}class _d{constructor(e,t,n,r){this.parser=e,this.input=t,this.ranges=r,this.recovering=0,this.nextStackID=9812,this.minStackPos=0,this.reused=[],this.stoppedAt=null,this.lastBigReductionStart=-1,this.lastBigReductionSize=0,this.bigReductionCount=0,this.stream=new fd(t,r),this.tokens=new Sd(e,this.stream),this.topTerm=e.top[1];let{from:a}=r[0];this.stacks=[sd.start(this,e.top[0],a)],this.fragments=n.length&&this.stream.end-a>4*e.bufferLength?new Ed(n,e.nodeSet):null}get parsedPos(){return this.minStackPos}advance(){let e,t,n=this.stacks,r=this.minStackPos,a=this.stacks=[];if(this.bigReductionCount>300&&1==n.length){let[e]=n;for(;e.forceReduce()&&e.stack.length&&e.stack[e.stack.length-2]>=this.lastBigReductionStart;);this.bigReductionCount=this.lastBigReductionSize=0}for(let o=0;o<n.length;o++){let i=n[o];for(;;){if(this.tokens.mainToken=null,i.pos>r)a.push(i);else{if(this.advanceStack(i,a,n))continue;{e||(e=[],t=[]),e.push(i);let n=this.tokens.getMainToken(i);t.push(n.value,n.end)}}break}}if(!a.length){let t=e&&function(e){let t=null;for(let n of e){let e=n.p.stoppedAt;(n.pos==n.p.stream.end||null!=e&&n.pos>e)&&n.p.parser.stateFlag(n.state,2)&&(!t||t.score<n.score)&&(t=n)}return t}(e);if(t)return Od&&console.log("Finish with "+this.stackID(t)),this.stackToTree(t);if(this.parser.strict)throw Od&&e&&console.log("Stuck with token "+(this.tokens.mainToken?this.parser.getName(this.tokens.mainToken.value):"none")),new SyntaxError("No parse at "+r);this.recovering||(this.recovering=5)}if(this.recovering&&e){let n=null!=this.stoppedAt&&e[0].pos>this.stoppedAt?e[0]:this.runRecovery(e,t,a);if(n)return Od&&console.log("Force-finish "+this.stackID(n)),this.stackToTree(n.forceAll())}if(this.recovering){let e=1==this.recovering?1:3*this.recovering;if(a.length>e)for(a.sort(((e,t)=>t.score-e.score));a.length>e;)a.pop();a.some((e=>e.reducePos>r))&&this.recovering--}else if(a.length>1){e:for(let e=0;e<a.length-1;e++){let t=a[e];for(let n=e+1;n<a.length;n++){let r=a[n];if(t.sameState(r)||t.buffer.length>500&&r.buffer.length>500){if(!((t.score-r.score||t.buffer.length-r.buffer.length)>0)){a.splice(e--,1);continue e}a.splice(n--,1)}}}a.length>12&&a.splice(12,a.length-12)}this.minStackPos=a[0].pos;for(let e=1;e<a.length;e++)a[e].pos<this.minStackPos&&(this.minStackPos=a[e].pos);return null}stopAt(e){if(null!=this.stoppedAt&&this.stoppedAt<e)throw new RangeError("Can't move stoppedAt forward");this.stoppedAt=e}advanceStack(e,t,n){let r=e.pos,{parser:a}=this,o=Od?this.stackID(e)+" -> ":"";if(null!=this.stoppedAt&&r>this.stoppedAt)return e.forceReduce()?e:null;if(this.fragments){let t=e.curContext&&e.curContext.tracker.strict,n=t?e.curContext.hash:0;for(let i=this.fragments.nodeAt(r);i;){let r=this.parser.nodeSet.types[i.type.id]==i.type?a.getGoto(e.state,i.type.id):-1;if(r>-1&&i.length&&(!t||(i.prop(bi.contextHash)||0)==n))return e.useNode(i,r),Od&&console.log(o+this.stackID(e)+` (via reuse of ${a.getName(i.type.id)})`),!0;if(!(i instanceof _i)||0==i.children.length||i.positions[0]>0)break;let s=i.children[0];if(!(s instanceof _i&&0==i.positions[0]))break;i=s}}let i=a.stateSlot(e.state,4);if(i>0)return e.reduce(i),Od&&console.log(o+this.stackID(e)+` (via always-reduce ${a.getName(65535&i)})`),!0;if(e.stack.length>=8400)for(;e.stack.length>6e3&&e.forceReduce(););let s=this.tokens.getActions(e);for(let i=0;i<s.length;){let l=s[i++],c=s[i++],u=s[i++],d=i==s.length||!n,p=d?e:e.split(),h=this.tokens.mainToken;if(p.apply(l,c,h?h.start:p.pos,u),Od&&console.log(o+this.stackID(p)+` (via ${65536&l?`reduce of ${a.getName(65535&l)}`:"shift"} for ${a.getName(c)} @ ${r}${p==e?"":", split"})`),d)return!0;p.pos>r?t.push(p):n.push(p)}return!1}advanceFully(e,t){let n=e.pos;for(;;){if(!this.advanceStack(e,null,null))return!1;if(e.pos>n)return Cd(e,t),!0}}runRecovery(e,t,n){let r=null,a=!1;for(let o=0;o<e.length;o++){let i=e[o],s=t[o<<1],l=t[1+(o<<1)],c=Od?this.stackID(i)+" -> ":"";if(i.deadEnd){if(a)continue;if(a=!0,i.restart(),Od&&console.log(c+this.stackID(i)+" (restarted)"),this.advanceFully(i,n))continue}let u=i.split(),d=c;for(let e=0;u.forceReduce()&&e<10&&(Od&&console.log(d+this.stackID(u)+" (via force-reduce)"),!this.advanceFully(u,n));e++)Od&&(d=this.stackID(u)+" -> ");for(let e of i.recoverByInsert(s))Od&&console.log(c+this.stackID(e)+" (via recover-insert)"),this.advanceFully(e,n);this.stream.end>i.pos?(l==i.pos&&(l++,s=0),i.recoverByDelete(s,l),Od&&console.log(c+this.stackID(i)+` (via recover-delete ${this.parser.getName(s)})`),Cd(i,n)):(!r||r.score<i.score)&&(r=i)}return r}stackToTree(e){return e.close(),_i.build({buffer:ud.create(e),nodeSet:this.parser.nodeSet,topID:this.topTerm,maxBufferLength:this.parser.bufferLength,reused:this.reused,start:this.ranges[0].from,length:e.pos-this.ranges[0].from,minRepeatType:this.parser.minRepeatTerm})}stackID(e){let t=(kd||(kd=new WeakMap)).get(e);return t||kd.set(e,t=String.fromCodePoint(this.nextStackID++)),t+e}}function Cd(e,t){for(let n=0;n<t.length;n++){let r=t[n];if(r.pos==e.pos&&r.sameState(e))return void(t[n].score<e.score&&(t[n]=e))}t.push(e)}class Td{constructor(e,t,n){this.source=e,this.flags=t,this.disabled=n}allows(e){return!this.disabled||0==this.disabled[e]}}const Ad=e=>e;class Rd{constructor(e){this.start=e.start,this.shift=e.shift||Ad,this.reduce=e.reduce||Ad,this.reuse=e.reuse||Ad,this.hash=e.hash||(()=>0),this.strict=!1!==e.strict}}class Pd extends qi{constructor(e){if(super(),this.wrappers=[],14!=e.version)throw new RangeError(`Parser version (${e.version}) doesn't match runtime version (14)`);let t=e.nodeNames.split(" ");this.minRepeatTerm=t.length;for(let n=0;n<e.repeatNodeCount;n++)t.push("");let n=Object.keys(e.topRules).map((t=>e.topRules[t][1])),r=[];for(let e=0;e<t.length;e++)r.push([]);function a(e,t,n){r[e].push([t,t.deserialize(String(n))])}if(e.nodeProps)for(let t of e.nodeProps){let e=t[0];"string"==typeof e&&(e=bi[e]);for(let n=1;n<t.length;){let r=t[n++];if(r>=0)a(r,e,t[n++]);else{let o=t[n+-r];for(let i=-r;i>0;i--)a(t[n++],e,o);n++}}}this.nodeSet=new ki(t.map(((t,a)=>Oi.define({name:a>=this.minRepeatTerm?void 0:t,id:a,props:r[a],top:n.indexOf(a)>-1,error:0==a,skipped:e.skippedNodes&&e.skippedNodes.indexOf(a)>-1})))),e.propSources&&(this.nodeSet=this.nodeSet.extend(...e.propSources)),this.strict=!1,this.bufferLength=mi;let o=dd(e.tokenData);this.context=e.context,this.specializerSpecs=e.specialized||[],this.specialized=new Uint16Array(this.specializerSpecs.length);for(let e=0;e<this.specializerSpecs.length;e++)this.specialized[e]=this.specializerSpecs[e].term;this.specializers=this.specializerSpecs.map(Md),this.states=dd(e.states,Uint32Array),this.data=dd(e.stateData),this.goto=dd(e.goto),this.maxTerm=e.maxTerm,this.tokenizers=e.tokenizers.map((e=>"number"==typeof e?new md(o,e):e)),this.topRules=e.topRules,this.dialects=e.dialects||{},this.dynamicPrecedences=e.dynamicPrecedences||null,this.tokenPrecTable=e.tokenPrec,this.termNames=e.termNames||null,this.maxNode=this.nodeSet.types.length-1,this.dialect=this.parseDialect(),this.top=this.topRules[Object.keys(this.topRules)[0]]}createParse(e,t,n){let r=new _d(this,e,t,n);for(let a of this.wrappers)r=a(r,e,t,n);return r}getGoto(e,t,n=!1){let r=this.goto;if(t>=r[0])return-1;for(let a=r[t+1];;){let t=r[a++],o=1&t,i=r[a++];if(o&&n)return i;for(let n=a+(t>>1);a<n;a++)if(r[a]==e)return i;if(o)return-1}}hasAction(e,t){let n=this.data;for(let r=0;r<2;r++)for(let a,o=this.stateSlot(e,r?2:1);;o+=3){if(65535==(a=n[o])){if(1!=n[o+1]){if(2==n[o+1])return Dd(n,o+2);break}a=n[o=Dd(n,o+2)]}if(a==t||0==a)return Dd(n,o+1)}return 0}stateSlot(e,t){return this.states[6*e+t]}stateFlag(e,t){return(this.stateSlot(e,0)&t)>0}validAction(e,t){return!!this.allActions(e,(e=>e==t||null))}allActions(e,t){let n=this.stateSlot(e,4),r=n?t(n):void 0;for(let n=this.stateSlot(e,1);null==r;n+=3){if(65535==this.data[n]){if(1!=this.data[n+1])break;n=Dd(this.data,n+2)}r=t(Dd(this.data,n+1))}return r}nextStates(e){let t=[];for(let n=this.stateSlot(e,1);;n+=3){if(65535==this.data[n]){if(1!=this.data[n+1])break;n=Dd(this.data,n+2)}if(!(1&this.data[n+2])){let e=this.data[n+1];t.some(((t,n)=>1&n&&t==e))||t.push(this.data[n],e)}}return t}configure(e){let t=Object.assign(Object.create(Pd.prototype),this);if(e.props&&(t.nodeSet=this.nodeSet.extend(...e.props)),e.top){let n=this.topRules[e.top];if(!n)throw new RangeError(`Invalid top rule name ${e.top}`);t.top=n}return e.tokenizers&&(t.tokenizers=this.tokenizers.map((t=>{let n=e.tokenizers.find((e=>e.from==t));return n?n.to:t}))),e.specializers&&(t.specializers=this.specializers.slice(),t.specializerSpecs=this.specializerSpecs.map(((n,r)=>{let a=e.specializers.find((e=>e.from==n.external));if(!a)return n;let o=Object.assign(Object.assign({},n),{external:a.to});return t.specializers[r]=Md(o),o}))),e.contextTracker&&(t.context=e.contextTracker),e.dialect&&(t.dialect=this.parseDialect(e.dialect)),null!=e.strict&&(t.strict=e.strict),e.wrap&&(t.wrappers=t.wrappers.concat(e.wrap)),null!=e.bufferLength&&(t.bufferLength=e.bufferLength),t}hasWrappers(){return this.wrappers.length>0}getName(e){return this.termNames?this.termNames[e]:String(e<=this.maxNode&&this.nodeSet.types[e].name||e)}get eofTerm(){return this.maxNode+1}get topNode(){return this.nodeSet.types[this.top[1]]}dynamicPrecedence(e){let t=this.dynamicPrecedences;return null==t?0:t[e]||0}parseDialect(e){let t=Object.keys(this.dialects),n=t.map((()=>!1));if(e)for(let r of e.split(" ")){let e=t.indexOf(r);e>=0&&(n[e]=!0)}let r=null;for(let e=0;e<t.length;e++)if(!n[e])for(let n,a=this.dialects[t[e]];65535!=(n=this.data[a++]);)(r||(r=new Uint8Array(this.maxTerm+1)))[n]=1;return new Td(e,n,r)}static deserialize(e){return new Pd(e)}}function Dd(e,t){return e[t]|e[t+1]<<16}function Md(e){if(e.external){let t=e.extend?1:0;return(n,r)=>e.external(n,r)<<1|t}return e.get}const $d=[9,10,11,12,13,32,133,160,5760,8192,8193,8194,8195,8196,8197,8198,8199,8200,8201,8202,8232,8233,8239,8287,12288];function Nd(e){return e>=65&&e<=90||e>=97&&e<=122||e>=161}function Id(e){return e>=48&&e<=57}const Ld=new vd(((e,t)=>{for(let n=!1,r=0,a=0;;a++){let{next:o}=e;if(Nd(o)||45==o||95==o||n&&Id(o))!n&&(45!=o||a>0)&&(n=!0),r===a&&45==o&&r++,e.advance();else{if(92!=o||10==e.peek(1)){n&&e.acceptToken(40==o?101:2==r&&t.canShift(2)?2:102);break}e.advance(),e.next>-1&&e.advance(),n=!0}}})),Bd=new vd((e=>{if($d.includes(e.peek(-1))){let{next:t}=e;(Nd(t)||95==t||35==t||46==t||91==t||58==t&&Nd(e.peek(1))||45==t||38==t)&&e.acceptToken(100)}})),Fd=new vd((e=>{if(!$d.includes(e.peek(-1))){let{next:t}=e;if(37==t&&(e.advance(),e.acceptToken(1)),Nd(t)){do{e.advance()}while(Nd(e.next)||Id(e.next));e.acceptToken(1)}}})),Qd=ds({"AtKeyword import charset namespace keyframes media supports":Ds.definitionKeyword,"from to selector":Ds.keyword,NamespaceName:Ds.namespace,KeyframeName:Ds.labelName,KeyframeRangeName:Ds.operatorKeyword,TagName:Ds.tagName,ClassName:Ds.className,PseudoClassName:Ds.constant(Ds.className),IdName:Ds.labelName,"FeatureName PropertyName":Ds.propertyName,AttributeName:Ds.attributeName,NumberLiteral:Ds.number,KeywordQuery:Ds.keyword,UnaryQueryOp:Ds.operatorKeyword,"CallTag ValueName":Ds.atom,VariableName:Ds.variableName,Callee:Ds.operatorKeyword,Unit:Ds.unit,"UniversalSelector NestingSelector":Ds.definitionOperator,MatchOp:Ds.compareOperator,"ChildOp SiblingOp, LogicOp":Ds.logicOperator,BinOp:Ds.arithmeticOperator,Important:Ds.modifier,Comment:Ds.blockComment,ColorLiteral:Ds.color,"ParenthesizedContent StringLiteral":Ds.string,":":Ds.punctuation,"PseudoOp #":Ds.derefOperator,"; ,":Ds.separator,"( )":Ds.paren,"[ ]":Ds.squareBracket,"{ }":Ds.brace}),Ud={__proto__:null,lang:34,"nth-child":34,"nth-last-child":34,"nth-of-type":34,"nth-last-of-type":34,dir:34,"host-context":34,url:62,"url-prefix":62,domain:62,regexp:62,selector:140},zd={__proto__:null,"@import":120,"@media":144,"@charset":148,"@namespace":152,"@keyframes":158,"@supports":170},jd={__proto__:null,not:134,only:134},Wd=Pd.deserialize({version:14,states:":jQYQ[OOO#_Q[OOP#fOWOOOOQP'#Cd'#CdOOQP'#Cc'#CcO#kQ[O'#CfO$_QXO'#CaO$fQ[O'#CiO$qQ[O'#DUO$vQ[O'#DXOOQP'#En'#EnO${QdO'#DhO%jQ[O'#DuO${QdO'#DwO%{Q[O'#DyO&WQ[O'#D|O&`Q[O'#ESO&nQ[O'#EUOOQS'#Em'#EmOOQS'#EX'#EXQYQ[OOO&uQXO'#CdO'jQWO'#DdO'oQWO'#EsO'zQ[O'#EsQOQWOOP(UO#tO'#C_POOO)C@])C@]OOQP'#Ch'#ChOOQP,59Q,59QO#kQ[O,59QO(aQ[O'#E]O({QWO,58{O)TQ[O,59TO$qQ[O,59pO$vQ[O,59sO(aQ[O,59vO(aQ[O,59xO(aQ[O,59yO)`Q[O'#DcOOQS,58{,58{OOQP'#Cl'#ClOOQO'#DS'#DSOOQP,59T,59TO)gQWO,59TO)lQWO,59TOOQP'#DW'#DWOOQP,59p,59pOOQO'#DY'#DYO)qQ`O,59sOOQS'#Cq'#CqO${QdO'#CrO)yQvO'#CtO+ZQtO,5:SOOQO'#Cy'#CyO)lQWO'#CxO+oQWO'#CzO+tQ[O'#DPOOQS'#Ep'#EpOOQO'#Dk'#DkO+|Q[O'#DrO,[QWO'#EtO&`Q[O'#DpO,jQWO'#DsOOQO'#Eu'#EuO)OQWO,5:aO,oQpO,5:cOOQS'#D{'#D{O,wQWO,5:eO,|Q[O,5:eOOQO'#EO'#EOO-UQWO,5:hO-ZQWO,5:nO-cQWO,5:pOOQS-E8V-E8VO-kQdO,5:OO-{Q[O'#E_O.YQWO,5;_O.YQWO,5;_POOO'#EW'#EWP.eO#tO,58yPOOO,58y,58yOOQP1G.l1G.lO/[QXO,5:wOOQO-E8Z-E8ZOOQS1G.g1G.gOOQP1G.o1G.oO)gQWO1G.oO)lQWO1G.oOOQP1G/[1G/[O/iQ`O1G/_O0SQXO1G/bO0jQXO1G/dO1QQXO1G/eO1hQWO,59}O1mQ[O'#DTO1tQdO'#CpOOQP1G/_1G/_O${QdO1G/_O1{QpO,59^OOQS,59`,59`O${QdO,59bO2TQWO1G/nOOQS,59d,59dO2YQ!bO,59fOOQS'#DQ'#DQOOQS'#EZ'#EZO2eQ[O,59kOOQS,59k,59kO2mQWO'#DkO2xQWO,5:WO2}QWO,5:^O&`Q[O,5:YO&`Q[O'#E`O3VQWO,5;`O3bQWO,5:[O(aQ[O,5:_OOQS1G/{1G/{OOQS1G/}1G/}OOQS1G0P1G0PO3sQWO1G0PO3xQdO'#EPOOQS1G0S1G0SOOQS1G0Y1G0YOOQS1G0[1G0[O4TQtO1G/jOOQO1G/j1G/jOOQO,5:y,5:yO4kQ[O,5:yOOQO-E8]-E8]O4xQWO1G0yPOOO-E8U-E8UPOOO1G.e1G.eOOQP7+$Z7+$ZOOQP7+$y7+$yO${QdO7+$yOOQS1G/i1G/iO5TQXO'#ErO5[QWO,59oO5aQtO'#EYO6XQdO'#EoO6cQWO,59[O6hQpO7+$yOOQS1G.x1G.xOOQS1G.|1G.|OOQS7+%Y7+%YOOQS1G/Q1G/QO6pQWO1G/QOOQS-E8X-E8XOOQS1G/V1G/VO${QdO1G/rOOQO1G/x1G/xOOQO1G/t1G/tO6uQWO,5:zOOQO-E8^-E8^O7TQXO1G/yOOQS7+%k7+%kO7[QYO'#CtOOQO'#ER'#ERO7gQ`O'#EQOOQO'#EQ'#EQO7rQWO'#EaO7zQdO,5:kOOQS,5:k,5:kO8VQtO'#E^O${QdO'#E^O9WQdO7+%UOOQO7+%U7+%UOOQO1G0e1G0eO9kQpO<<HeO9sQWO,5;^OOQP1G/Z1G/ZOOQS-E8W-E8WO${QdO'#E[O9{QWO,5;ZOOQT1G.v1G.vOOQP<<He<<HeOOQS7+$l7+$lO:TQdO7+%^OOQO7+%e7+%eOOQO,5:l,5:lO3{QdO'#EbO7rQWO,5:{OOQS,5:{,5:{OOQS-E8_-E8_OOQS1G0V1G0VO:[QtO,5:xOOQS-E8[-E8[OOQO<<Hp<<HpOOQPAN>PAN>PO;]QdO,5:vOOQO-E8Y-E8YOOQO<<Hx<<HxOOQO,5:|,5:|OOQO-E8`-E8`OOQS1G0g1G0g",stateData:";o~O#[OS#]QQ~OUYOXYOZTO^VO_VOrXOyWO!]aO!^ZO!j[O!l]O!n^O!q_O!w`O#YRO~OQfOUYOXYOZTO^VO_VOrXOyWO!]aO!^ZO!j[O!l]O!n^O!q_O!w`O#YeO~O#V#gP~P!ZO#]jO~O#YlO~OZnO^qO_qOrsOuoOyrO!PtO!SvO#WuO~O!UwO~P#pOa}O#XzO#YyO~O#Y!OO~O#Y!QO~OQ![Oc!TOg![Oi![Oo!YOr!ZO#X!WO#Y!SO#e!UO~Oc!^O!e!`O!h!aO#Y!]O!U#hP~Oi!fOo!YO#Y!eO~Oi!hO#Y!hO~Oc!^O!e!`O!h!aO#Y!]O~O!Z#hP~P%jOZWX^WX^!XX_WXrWXuWXyWX!PWX!SWX!UWX#WWX~O^!mO~O!Z!nO#V#gX!T#gX~O#V#gX!T#gX~P!ZO#^!qO#_!qO#`!sO~OUYOXYOZTO^VO_VOrXOyWO#YRO~OuoO!UwO~Oa!zO#XzO#YyO~O!T#gP~P!ZOc#RO~Oc#SO~Oq#TO}#UO~OP#WOchXkhX!ZhX!ehX!hhX#YhXbhXQhXghXihXohXrhXuhX!YhX#VhX#XhX#ehXqhX!ThX~Oc!^Ok#XO!e!`O!h!aO#Y!]O!Z#hP~Oc#[O~Oq#`O#Y#]O~Oc!^O!e!`O!h!aO#Y#aO~Ou#eO!c#dO!U#hX!Z#hX~Oc#hO~Ok#XO!Z#jO~O!Z#kO~Oi#lOo!YO~O!U#mO~O!UwO!c#dO~O!UwO!Z#pO~O!Y#rO!Z!Wa#V!Wa!T!Wa~P${O!Z#RX#V#RX!T#RX~P!ZO!Z!nO#V#ga!T#ga~O#^!qO#_!qO#`#xO~OZnO^qO_qOrsOyrO!PtO!SvO#WuO~Ou#Pa!U#Pab#Pa~P.pOq#zO}#{O~OZnO^qO_qOrsOyrO~Ou!Oi!P!Oi!S!Oi!U!Oi#W!Oib!Oi~P/qOu!Qi!P!Qi!S!Qi!U!Qi#W!Qib!Qi~P/qOu!Ri!P!Ri!S!Ri!U!Ri#W!Rib!Ri~P/qO!T#|O~Ob#fP~P(aOb#cP~P${Ob$TOk#XO~O!Z$VO~Ob$WOi$XOp$XO~Oq$ZO#Y#]O~O^!aXb!_X!c!_X~O^$[O~Ob$]O!c#dO~Ou#eO!U#ha!Z#ha~O!c#dOu!da!U!da!Z!dab!da~O!Z$bO~O!T$iO#Y$dO#e$cO~Ok#XOu$kO!Y$mO!Z!Wi#V!Wi!T!Wi~P${O!Z#Ra#V#Ra!T#Ra~P!ZO!Z!nO#V#gi!T#gi~Ob#fX~P#pOb$qO~Ok#XOQ!|Xb!|Xc!|Xg!|Xi!|Xo!|Xr!|Xu!|X#X!|X#Y!|X#e!|X~Ou$sOb#cX~P${Ob$uO~Ok#XOq$vO~Ob$wO~O!c#dOu#Sa!U#Sa!Z#Sa~Ob$yO~P.pOP#WOuhX!UhX~O#e$cOu!tX!U!tX~Ou${O!UwO~O!T%PO#Y$dO#e$cO~Ok#XOQ#QXc#QXg#QXi#QXo#QXr#QXu#QX!Y#QX!Z#QX#V#QX#X#QX#Y#QX#e#QX!T#QX~Ou$kO!Y%SO!Z!Wq#V!Wq!T!Wq~P${Ok#XOq%TO~OuoOb#fa~Ou$sOb#ca~Ob%WO~P${Ok#XOQ#Qac#Qag#Qai#Qao#Qar#Qau#Qa!Y#Qa!Z#Qa#V#Qa#X#Qa#Y#Qa#e#Qa!T#Qa~Ob#Oau#Oa~P${O#[p#]#ek!S#e~",goto:"-g#jPPP#kP#nP#w$WP#wP$g#wPP$mPPP$s$|$|P%`P$|P$|%z&^PPPP$|&vP&z'Q#wP'W#w'^P#wP#w#wPPP'd'y(WPP#nPP(_(_(i(_P(_P(_(_P#nP#nP#nP(l#nP(o(r(u(|#nP#nP)R)X)h)v)|*S*^*d*n*t*zPPPPPPPPPP+Q+Z+v+yP,o,r,x-RRkQ_bOPdhw!n#tkYOPdhotuvw!n#R#h#tkSOPdhotuvw!n#R#h#tQmTR!tnQ{VR!xqQ!x}Q#Z!XR#y!zq![Z]!T!m#S#U#X#q#{$Q$[$k$l$s$x%Up![Z]!T!m#S#U#X#q#{$Q$[$k$l$s$x%UU$f#m$h${R$z$eq!XZ]!T!m#S#U#X#q#{$Q$[$k$l$s$x%Up![Z]!T!m#S#U#X#q#{$Q$[$k$l$s$x%UQ!f^R#l!gT#^!Z#_Q|VR!yqQ!x|R#y!yQ!PWR!{rQ!RXR!|sQxUQ!wpQ#i!cQ#o!jQ#p!kQ$}$gR%Z$|SgPwQ!phQ#s!nR$n#tZfPhw!n#ta!b[`a!V!^!`#d#eR#b!^R!g^R!i_R#n!iS$g#m$hR%X${V$e#m$h${Q!rjR#w!rQdOShPwU!ldh#tR#t!nQ$Q#SU$r$Q$x%UQ$x$[R%U$sQ#_!ZR$Y#_Q$t$QR%V$tQpUS!vp$pR$p#}Q$l#qR%R$lQ!ogS#u!o#vR#v!pQ#f!_R$`#fQ$h#mR%O$hQ$|$gR%Y$|_cOPdhw!n#t^UOPdhw!n#tQ!uoQ!}tQ#OuQ#PvQ#}#RR$a#hR$R#SQ!VZQ!d]Q#V!TQ#q!m[$P#S$Q$[$s$x%UQ$S#UQ$U#XS$j#q$lQ$o#{R%Q$kR$O#RQiPR#QwQ!c[Q!kaR#Y!VU!_[a!VQ!j`Q#c!^Q#g!`Q$^#dR$_#e",nodeNames:"⚠ Unit VariableName Comment StyleSheet RuleSet UniversalSelector TagSelector TagName NestingSelector ClassSelector . ClassName PseudoClassSelector : :: PseudoClassName PseudoClassName ) ( ArgList ValueName ParenthesizedValue ColorLiteral NumberLiteral StringLiteral BinaryExpression BinOp CallExpression Callee CallLiteral CallTag ParenthesizedContent ] [ LineNames LineName , PseudoClassName ArgList IdSelector # IdName AttributeSelector AttributeName MatchOp ChildSelector ChildOp DescendantSelector SiblingSelector SiblingOp } { Block Declaration PropertyName Important ; ImportStatement AtKeyword import KeywordQuery FeatureQuery FeatureName BinaryQuery LogicOp UnaryQuery UnaryQueryOp ParenthesizedQuery SelectorQuery selector MediaStatement media CharsetStatement charset NamespaceStatement namespace NamespaceName KeyframesStatement keyframes KeyframeName KeyframeList KeyframeSelector KeyframeRangeName SupportsStatement supports AtRule Styles",maxTerm:117,nodeProps:[["isolate",-2,3,25,""],["openedBy",18,"(",33,"[",51,"{"],["closedBy",19,")",34,"]",52,"}"]],propSources:[Qd],skippedNodes:[0,3,88],repeatNodeCount:11,tokenData:"J^~R!^OX$}X^%u^p$}pq%uqr)Xrs.Rst/utu6duv$}vw7^wx7oxy9^yz9oz{9t{|:_|}?Q}!O?c!O!P@Q!P!Q@i!Q![Ab![!]B]!]!^CX!^!_$}!_!`Cj!`!aC{!a!b$}!b!cDw!c!}$}!}#OFa#O#P$}#P#QFr#Q#R6d#R#T$}#T#UGT#U#c$}#c#dHf#d#o$}#o#pH{#p#q6d#q#rI^#r#sIo#s#y$}#y#z%u#z$f$}$f$g%u$g#BY$}#BY#BZ%u#BZ$IS$}$IS$I_%u$I_$I|$}$I|$JO%u$JO$JT$}$JT$JU%u$JU$KV$}$KV$KW%u$KW&FU$}&FU&FV%u&FV;'S$};'S;=`JW<%lO$}`%QSOy%^z;'S%^;'S;=`%o<%lO%^`%cSp`Oy%^z;'S%^;'S;=`%o<%lO%^`%rP;=`<%l%^~%zh#[~OX%^X^'f^p%^pq'fqy%^z#y%^#y#z'f#z$f%^$f$g'f$g#BY%^#BY#BZ'f#BZ$IS%^$IS$I_'f$I_$I|%^$I|$JO'f$JO$JT%^$JT$JU'f$JU$KV%^$KV$KW'f$KW&FU%^&FU&FV'f&FV;'S%^;'S;=`%o<%lO%^~'mh#[~p`OX%^X^'f^p%^pq'fqy%^z#y%^#y#z'f#z$f%^$f$g'f$g#BY%^#BY#BZ'f#BZ$IS%^$IS$I_'f$I_$I|%^$I|$JO'f$JO$JT%^$JT$JU'f$JU$KV%^$KV$KW'f$KW&FU%^&FU&FV'f&FV;'S%^;'S;=`%o<%lO%^l)[UOy%^z#]%^#]#^)n#^;'S%^;'S;=`%o<%lO%^l)sUp`Oy%^z#a%^#a#b*V#b;'S%^;'S;=`%o<%lO%^l*[Up`Oy%^z#d%^#d#e*n#e;'S%^;'S;=`%o<%lO%^l*sUp`Oy%^z#c%^#c#d+V#d;'S%^;'S;=`%o<%lO%^l+[Up`Oy%^z#f%^#f#g+n#g;'S%^;'S;=`%o<%lO%^l+sUp`Oy%^z#h%^#h#i,V#i;'S%^;'S;=`%o<%lO%^l,[Up`Oy%^z#T%^#T#U,n#U;'S%^;'S;=`%o<%lO%^l,sUp`Oy%^z#b%^#b#c-V#c;'S%^;'S;=`%o<%lO%^l-[Up`Oy%^z#h%^#h#i-n#i;'S%^;'S;=`%o<%lO%^l-uS!Y[p`Oy%^z;'S%^;'S;=`%o<%lO%^~.UWOY.RZr.Rrs.ns#O.R#O#P.s#P;'S.R;'S;=`/o<%lO.R~.sOi~~.vRO;'S.R;'S;=`/P;=`O.R~/SXOY.RZr.Rrs.ns#O.R#O#P.s#P;'S.R;'S;=`/o;=`<%l.R<%lO.R~/rP;=`<%l.Rn/zYyQOy%^z!Q%^!Q![0j![!c%^!c!i0j!i#T%^#T#Z0j#Z;'S%^;'S;=`%o<%lO%^l0oYp`Oy%^z!Q%^!Q![1_![!c%^!c!i1_!i#T%^#T#Z1_#Z;'S%^;'S;=`%o<%lO%^l1dYp`Oy%^z!Q%^!Q![2S![!c%^!c!i2S!i#T%^#T#Z2S#Z;'S%^;'S;=`%o<%lO%^l2ZYg[p`Oy%^z!Q%^!Q![2y![!c%^!c!i2y!i#T%^#T#Z2y#Z;'S%^;'S;=`%o<%lO%^l3QYg[p`Oy%^z!Q%^!Q![3p![!c%^!c!i3p!i#T%^#T#Z3p#Z;'S%^;'S;=`%o<%lO%^l3uYp`Oy%^z!Q%^!Q![4e![!c%^!c!i4e!i#T%^#T#Z4e#Z;'S%^;'S;=`%o<%lO%^l4lYg[p`Oy%^z!Q%^!Q![5[![!c%^!c!i5[!i#T%^#T#Z5[#Z;'S%^;'S;=`%o<%lO%^l5aYp`Oy%^z!Q%^!Q![6P![!c%^!c!i6P!i#T%^#T#Z6P#Z;'S%^;'S;=`%o<%lO%^l6WSg[p`Oy%^z;'S%^;'S;=`%o<%lO%^d6gUOy%^z!_%^!_!`6y!`;'S%^;'S;=`%o<%lO%^d7QS}Sp`Oy%^z;'S%^;'S;=`%o<%lO%^b7cSXQOy%^z;'S%^;'S;=`%o<%lO%^~7rWOY7oZw7owx.nx#O7o#O#P8[#P;'S7o;'S;=`9W<%lO7o~8_RO;'S7o;'S;=`8h;=`O7o~8kXOY7oZw7owx.nx#O7o#O#P8[#P;'S7o;'S;=`9W;=`<%l7o<%lO7o~9ZP;=`<%l7on9cSc^Oy%^z;'S%^;'S;=`%o<%lO%^~9tOb~n9{UUQkWOy%^z!_%^!_!`6y!`;'S%^;'S;=`%o<%lO%^n:fWkW!SQOy%^z!O%^!O!P;O!P!Q%^!Q![>T![;'S%^;'S;=`%o<%lO%^l;TUp`Oy%^z!Q%^!Q![;g![;'S%^;'S;=`%o<%lO%^l;nYp`#e[Oy%^z!Q%^!Q![;g![!g%^!g!h<^!h#X%^#X#Y<^#Y;'S%^;'S;=`%o<%lO%^l<cYp`Oy%^z{%^{|=R|}%^}!O=R!O!Q%^!Q![=j![;'S%^;'S;=`%o<%lO%^l=WUp`Oy%^z!Q%^!Q![=j![;'S%^;'S;=`%o<%lO%^l=qUp`#e[Oy%^z!Q%^!Q![=j![;'S%^;'S;=`%o<%lO%^l>[[p`#e[Oy%^z!O%^!O!P;g!P!Q%^!Q![>T![!g%^!g!h<^!h#X%^#X#Y<^#Y;'S%^;'S;=`%o<%lO%^n?VSu^Oy%^z;'S%^;'S;=`%o<%lO%^l?hWkWOy%^z!O%^!O!P;O!P!Q%^!Q![>T![;'S%^;'S;=`%o<%lO%^n@VUZQOy%^z!Q%^!Q![;g![;'S%^;'S;=`%o<%lO%^~@nTkWOy%^z{@}{;'S%^;'S;=`%o<%lO%^~AUSp`#]~Oy%^z;'S%^;'S;=`%o<%lO%^lAg[#e[Oy%^z!O%^!O!P;g!P!Q%^!Q![>T![!g%^!g!h<^!h#X%^#X#Y<^#Y;'S%^;'S;=`%o<%lO%^bBbU^QOy%^z![%^![!]Bt!];'S%^;'S;=`%o<%lO%^bB{S_Qp`Oy%^z;'S%^;'S;=`%o<%lO%^nC^S!Z^Oy%^z;'S%^;'S;=`%o<%lO%^dCoS}SOy%^z;'S%^;'S;=`%o<%lO%^bDQU!PQOy%^z!`%^!`!aDd!a;'S%^;'S;=`%o<%lO%^bDkS!PQp`Oy%^z;'S%^;'S;=`%o<%lO%^bDzWOy%^z!c%^!c!}Ed!}#T%^#T#oEd#o;'S%^;'S;=`%o<%lO%^bEk[!]Qp`Oy%^z}%^}!OEd!O!Q%^!Q![Ed![!c%^!c!}Ed!}#T%^#T#oEd#o;'S%^;'S;=`%o<%lO%^nFfSr^Oy%^z;'S%^;'S;=`%o<%lO%^nFwSq^Oy%^z;'S%^;'S;=`%o<%lO%^bGWUOy%^z#b%^#b#cGj#c;'S%^;'S;=`%o<%lO%^bGoUp`Oy%^z#W%^#W#XHR#X;'S%^;'S;=`%o<%lO%^bHYS!cQp`Oy%^z;'S%^;'S;=`%o<%lO%^bHiUOy%^z#f%^#f#gHR#g;'S%^;'S;=`%o<%lO%^fIQS!UUOy%^z;'S%^;'S;=`%o<%lO%^nIcS!T^Oy%^z;'S%^;'S;=`%o<%lO%^fItU!SQOy%^z!_%^!_!`6y!`;'S%^;'S;=`%o<%lO%^`JZP;=`<%l$}",tokenizers:[Bd,Fd,Ld,1,2,3,4,new gd("m~RRYZ[z{a~~g~aO#_~~dP!P!Qg~lO#`~~",28,106)],topRules:{StyleSheet:[0,4],Styles:[1,87]},specialized:[{term:101,get:e=>Ud[e]||-1},{term:59,get:e=>zd[e]||-1},{term:102,get:e=>jd[e]||-1}],tokenPrec:1219});let Xd=null;function qd(){if(!Xd&&"object"==typeof document&&document.body){let{style:e}=document.body,t=[],n=new Set;for(let r in e)"cssText"!=r&&"cssFloat"!=r&&"string"==typeof e[r]&&(/[A-Z]/.test(r)&&(r=r.replace(/[A-Z]/g,(e=>"-"+e.toLowerCase()))),n.has(r)||(t.push(r),n.add(r)));Xd=t.sort().map((e=>({type:"property",label:e,apply:e+": "})))}return Xd||[]}const Vd=["active","after","any-link","autofill","backdrop","before","checked","cue","default","defined","disabled","empty","enabled","file-selector-button","first","first-child","first-letter","first-line","first-of-type","focus","focus-visible","focus-within","fullscreen","has","host","host-context","hover","in-range","indeterminate","invalid","is","lang","last-child","last-of-type","left","link","marker","modal","not","nth-child","nth-last-child","nth-last-of-type","nth-of-type","only-child","only-of-type","optional","out-of-range","part","placeholder","placeholder-shown","read-only","read-write","required","right","root","scope","selection","slotted","target","target-text","valid","visited","where"].map((e=>({type:"class",label:e}))),Hd=["above","absolute","activeborder","additive","activecaption","after-white-space","ahead","alias","all","all-scroll","alphabetic","alternate","always","antialiased","appworkspace","asterisks","attr","auto","auto-flow","avoid","avoid-column","avoid-page","avoid-region","axis-pan","background","backwards","baseline","below","bidi-override","blink","block","block-axis","bold","bolder","border","border-box","both","bottom","break","break-all","break-word","bullets","button","button-bevel","buttonface","buttonhighlight","buttonshadow","buttontext","calc","capitalize","caps-lock-indicator","caption","captiontext","caret","cell","center","checkbox","circle","cjk-decimal","clear","clip","close-quote","col-resize","collapse","color","color-burn","color-dodge","column","column-reverse","compact","condensed","contain","content","contents","content-box","context-menu","continuous","copy","counter","counters","cover","crop","cross","crosshair","currentcolor","cursive","cyclic","darken","dashed","decimal","decimal-leading-zero","default","default-button","dense","destination-atop","destination-in","destination-out","destination-over","difference","disc","discard","disclosure-closed","disclosure-open","document","dot-dash","dot-dot-dash","dotted","double","down","e-resize","ease","ease-in","ease-in-out","ease-out","element","ellipse","ellipsis","embed","end","ethiopic-abegede-gez","ethiopic-halehame-aa-er","ethiopic-halehame-gez","ew-resize","exclusion","expanded","extends","extra-condensed","extra-expanded","fantasy","fast","fill","fill-box","fixed","flat","flex","flex-end","flex-start","footnotes","forwards","from","geometricPrecision","graytext","grid","groove","hand","hard-light","help","hidden","hide","higher","highlight","highlighttext","horizontal","hsl","hsla","hue","icon","ignore","inactiveborder","inactivecaption","inactivecaptiontext","infinite","infobackground","infotext","inherit","initial","inline","inline-axis","inline-block","inline-flex","inline-grid","inline-table","inset","inside","intrinsic","invert","italic","justify","keep-all","landscape","large","larger","left","level","lighter","lighten","line-through","linear","linear-gradient","lines","list-item","listbox","listitem","local","logical","loud","lower","lower-hexadecimal","lower-latin","lower-norwegian","lowercase","ltr","luminosity","manipulation","match","matrix","matrix3d","medium","menu","menutext","message-box","middle","min-intrinsic","mix","monospace","move","multiple","multiple_mask_images","multiply","n-resize","narrower","ne-resize","nesw-resize","no-close-quote","no-drop","no-open-quote","no-repeat","none","normal","not-allowed","nowrap","ns-resize","numbers","numeric","nw-resize","nwse-resize","oblique","opacity","open-quote","optimizeLegibility","optimizeSpeed","outset","outside","outside-shape","overlay","overline","padding","padding-box","painted","page","paused","perspective","pinch-zoom","plus-darker","plus-lighter","pointer","polygon","portrait","pre","pre-line","pre-wrap","preserve-3d","progress","push-button","radial-gradient","radio","read-only","read-write","read-write-plaintext-only","rectangle","region","relative","repeat","repeating-linear-gradient","repeating-radial-gradient","repeat-x","repeat-y","reset","reverse","rgb","rgba","ridge","right","rotate","rotate3d","rotateX","rotateY","rotateZ","round","row","row-resize","row-reverse","rtl","run-in","running","s-resize","sans-serif","saturation","scale","scale3d","scaleX","scaleY","scaleZ","screen","scroll","scrollbar","scroll-position","se-resize","self-start","self-end","semi-condensed","semi-expanded","separate","serif","show","single","skew","skewX","skewY","skip-white-space","slide","slider-horizontal","slider-vertical","sliderthumb-horizontal","sliderthumb-vertical","slow","small","small-caps","small-caption","smaller","soft-light","solid","source-atop","source-in","source-out","source-over","space","space-around","space-between","space-evenly","spell-out","square","start","static","status-bar","stretch","stroke","stroke-box","sub","subpixel-antialiased","svg_masks","super","sw-resize","symbolic","symbols","system-ui","table","table-caption","table-cell","table-column","table-column-group","table-footer-group","table-header-group","table-row","table-row-group","text","text-bottom","text-top","textarea","textfield","thick","thin","threeddarkshadow","threedface","threedhighlight","threedlightshadow","threedshadow","to","top","transform","translate","translate3d","translateX","translateY","translateZ","transparent","ultra-condensed","ultra-expanded","underline","unidirectional-pan","unset","up","upper-latin","uppercase","url","var","vertical","vertical-text","view-box","visible","visibleFill","visiblePainted","visibleStroke","visual","w-resize","wait","wave","wider","window","windowframe","windowtext","words","wrap","wrap-reverse","x-large","x-small","xor","xx-large","xx-small"].map((e=>({type:"keyword",label:e}))).concat(["aliceblue","antiquewhite","aqua","aquamarine","azure","beige","bisque","black","blanchedalmond","blue","blueviolet","brown","burlywood","cadetblue","chartreuse","chocolate","coral","cornflowerblue","cornsilk","crimson","cyan","darkblue","darkcyan","darkgoldenrod","darkgray","darkgreen","darkkhaki","darkmagenta","darkolivegreen","darkorange","darkorchid","darkred","darksalmon","darkseagreen","darkslateblue","darkslategray","darkturquoise","darkviolet","deeppink","deepskyblue","dimgray","dodgerblue","firebrick","floralwhite","forestgreen","fuchsia","gainsboro","ghostwhite","gold","goldenrod","gray","grey","green","greenyellow","honeydew","hotpink","indianred","indigo","ivory","khaki","lavender","lavenderblush","lawngreen","lemonchiffon","lightblue","lightcoral","lightcyan","lightgoldenrodyellow","lightgray","lightgreen","lightpink","lightsalmon","lightseagreen","lightskyblue","lightslategray","lightsteelblue","lightyellow","lime","limegreen","linen","magenta","maroon","mediumaquamarine","mediumblue","mediumorchid","mediumpurple","mediumseagreen","mediumslateblue","mediumspringgreen","mediumturquoise","mediumvioletred","midnightblue","mintcream","mistyrose","moccasin","navajowhite","navy","oldlace","olive","olivedrab","orange","orangered","orchid","palegoldenrod","palegreen","paleturquoise","palevioletred","papayawhip","peachpuff","peru","pink","plum","powderblue","purple","rebeccapurple","red","rosybrown","royalblue","saddlebrown","salmon","sandybrown","seagreen","seashell","sienna","silver","skyblue","slateblue","slategray","snow","springgreen","steelblue","tan","teal","thistle","tomato","turquoise","violet","wheat","white","whitesmoke","yellow","yellowgreen"].map((e=>({type:"constant",label:e})))),Zd=["a","abbr","address","article","aside","b","bdi","bdo","blockquote","body","br","button","canvas","caption","cite","code","col","colgroup","dd","del","details","dfn","dialog","div","dl","dt","em","figcaption","figure","footer","form","header","hgroup","h1","h2","h3","h4","h5","h6","hr","html","i","iframe","img","input","ins","kbd","label","legend","li","main","meter","nav","ol","output","p","pre","ruby","section","select","small","source","span","strong","sub","summary","sup","table","tbody","td","template","textarea","tfoot","th","thead","tr","u","ul"].map((e=>({type:"type",label:e}))),Gd=["@charset","@color-profile","@container","@counter-style","@font-face","@font-feature-values","@font-palette-values","@import","@keyframes","@layer","@media","@namespace","@page","@position-try","@property","@scope","@starting-style","@supports","@view-transition"].map((e=>({type:"keyword",label:e}))),Kd=/^(\w[\w-]*|-\w[\w-]*|)$/,Yd=/^-(-[\w-]*)?$/,Jd=new Wi,ep=["Declaration"];function tp(e){for(let t=e;;){if(t.type.isTop)return t;if(!(t=t.parent))return e}}function np(e,t,n){if(t.to-t.from>4096){let r=Jd.get(t);if(r)return r;let a=[],o=new Set,i=t.cursor(Si.IncludeAnonymous);if(i.firstChild())do{for(let t of np(e,i.node,n))o.has(t.label)||(o.add(t.label),a.push(t))}while(i.nextSibling());return Jd.set(t,a),a}{let r=[],a=new Set;return t.cursor().iterate((t=>{var o;if(n(t)&&t.matchContext(ep)&&":"==(null===(o=t.node.nextSibling)||void 0===o?void 0:o.name)){let n=e.sliceString(t.from,t.to);a.has(n)||(a.add(n),r.push({label:n,type:"variable"}))}})),r}}const rp=e=>t=>{let{state:n,pos:r}=t,a=Qs(n).resolveInner(r,-1),o=a.type.isError&&a.from==a.to-1&&"-"==n.doc.sliceString(a.from,a.to);if("PropertyName"==a.name||(o||"TagName"==a.name)&&/^(Block|Styles)$/.test(a.resolve(a.to).name))return{from:a.from,options:qd(),validFor:Kd};if("ValueName"==a.name)return{from:a.from,options:Hd,validFor:Kd};if("PseudoClassName"==a.name)return{from:a.from,options:Vd,validFor:Kd};if(e(a)||(t.explicit||o)&&function(e,t){var n;if(("("==e.name||e.type.isError)&&(e=e.parent||e),"ArgList"!=e.name)return!1;let r=null===(n=e.parent)||void 0===n?void 0:n.firstChild;return"Callee"==(null==r?void 0:r.name)&&"var"==t.sliceString(r.from,r.to)}(a,n.doc))return{from:e(a)||o?a.from:r,options:np(n.doc,tp(a),e),validFor:Yd};if("TagName"==a.name){for(let{parent:e}=a;e;e=e.parent)if("Block"==e.name)return{from:a.from,options:qd(),validFor:Kd};return{from:a.from,options:Zd,validFor:Kd}}if("AtKeyword"==a.name)return{from:a.from,options:Gd,validFor:Kd};if(!t.explicit)return null;let i=a.resolve(r),s=i.childBefore(r);return s&&":"==s.name&&"PseudoClassSelector"==i.name?{from:r,options:Vd,validFor:Kd}:s&&":"==s.name&&"Declaration"==i.name||"ArgList"==i.name?{from:r,options:Hd,validFor:Kd}:"Block"==i.name||"Styles"==i.name?{from:r,options:qd(),validFor:Kd}:null},ap=rp((e=>"VariableName"==e.name)),op=Fs.define({name:"css",parser:Wd.configure({props:[rl.add({Declaration:dl()}),hl.add({"Block KeyframeList":fl})]}),languageData:{commentTokens:{block:{open:"/*",close:"*/"}},indentOnInput:/^\s*\}$/,wordChars:"-"}});function ip(){return new Gs(op,op.data.of({autocomplete:ap}))}const sp={area:!0,base:!0,br:!0,col:!0,command:!0,embed:!0,frame:!0,hr:!0,img:!0,input:!0,keygen:!0,link:!0,meta:!0,param:!0,source:!0,track:!0,wbr:!0,menuitem:!0},lp={dd:!0,li:!0,optgroup:!0,option:!0,p:!0,rp:!0,rt:!0,tbody:!0,td:!0,tfoot:!0,th:!0,tr:!0},cp={dd:{dd:!0,dt:!0},dt:{dd:!0,dt:!0},li:{li:!0},option:{option:!0,optgroup:!0},optgroup:{optgroup:!0},p:{address:!0,article:!0,aside:!0,blockquote:!0,dir:!0,div:!0,dl:!0,fieldset:!0,footer:!0,form:!0,h1:!0,h2:!0,h3:!0,h4:!0,h5:!0,h6:!0,header:!0,hgroup:!0,hr:!0,menu:!0,nav:!0,ol:!0,p:!0,pre:!0,section:!0,table:!0,ul:!0},rp:{rp:!0,rt:!0},rt:{rp:!0,rt:!0},tbody:{tbody:!0,tfoot:!0},td:{td:!0,th:!0},tfoot:{tbody:!0},th:{td:!0,th:!0},thead:{tbody:!0,tfoot:!0},tr:{tr:!0}};function up(e){return 9==e||10==e||13==e||32==e}let dp=null,pp=null,hp=0;function fp(e,t){let n=e.pos+t;if(hp==n&&pp==e)return dp;let r=e.peek(t);for(;up(r);)r=e.peek(++t);let a="";for(;45==(o=r)||46==o||58==o||o>=65&&o<=90||95==o||o>=97&&o<=122||o>=161;)a+=String.fromCharCode(r),r=e.peek(++t);var o;return pp=e,hp=n,dp=a?a.toLowerCase():r==mp||r==gp?void 0:null}const mp=63,gp=33;function vp(e,t){this.name=e,this.parent=t}const bp=[6,10,7,8,9],yp=new Rd({start:null,shift:(e,t,n,r)=>bp.indexOf(t)>-1?new vp(fp(r,1)||"",e):e,reduce:(e,t)=>20==t&&e?e.parent:e,reuse(e,t,n,r){let a=t.type.id;return 6==a||36==a?new vp(fp(r,1)||"",e):e},strict:!1}),wp=new vd(((e,t)=>{if(60!=e.next)return void(e.next<0&&t.context&&e.acceptToken(57));e.advance();let n=47==e.next;n&&e.advance();let r=fp(e,0);if(void 0===r)return;if(!r)return e.acceptToken(n?14:6);let a=t.context?t.context.name:null;if(n){if(r==a)return e.acceptToken(11);if(a&&lp[a])return e.acceptToken(57,-2);if(t.dialectEnabled(0))return e.acceptToken(12);for(let e=t.context;e;e=e.parent)if(e.name==r)return;e.acceptToken(13)}else{if("script"==r)return e.acceptToken(7);if("style"==r)return e.acceptToken(8);if("textarea"==r)return e.acceptToken(9);if(sp.hasOwnProperty(r))return e.acceptToken(10);a&&cp[a]&&cp[a][r]?e.acceptToken(57,-1):e.acceptToken(6)}}),{contextual:!0}),Op=new vd((e=>{for(let t=0,n=0;;n++){if(e.next<0){n&&e.acceptToken(58);break}if(45==e.next)t++;else{if(62==e.next&&t>=2){n>=3&&e.acceptToken(58,-2);break}t=0}e.advance()}})),kp=new vd(((e,t)=>{if(47==e.next&&62==e.peek(1)){let n=t.dialectEnabled(1)||function(e){for(;e;e=e.parent)if("svg"==e.name||"math"==e.name)return!0;return!1}(t.context);e.acceptToken(n?5:4,2)}else 62==e.next&&e.acceptToken(4,1)}));function xp(e,t,n){let r=2+e.length;return new vd((a=>{for(let o=0,i=0,s=0;;s++){if(a.next<0){s&&a.acceptToken(t);break}if(0==o&&60==a.next||1==o&&47==a.next||o>=2&&o<r&&a.next==e.charCodeAt(o-2))o++,i++;else if(2!=o&&o!=r||!up(a.next)){if(o==r&&62==a.next){s>i?a.acceptToken(t,-i):a.acceptToken(n,-(i-2));break}if((10==a.next||13==a.next)&&s){a.acceptToken(t,1);break}o=i=0}else i++;a.advance()}}))}const Ep=xp("script",54,1),Sp=xp("style",55,2),_p=xp("textarea",56,3),Cp=ds({"Text RawText":Ds.content,"StartTag StartCloseTag SelfClosingEndTag EndTag":Ds.angleBracket,TagName:Ds.tagName,"MismatchedCloseTag/TagName":[Ds.tagName,Ds.invalid],AttributeName:Ds.attributeName,"AttributeValue UnquotedAttributeValue":Ds.attributeValue,Is:Ds.definitionOperator,"EntityReference CharacterReference":Ds.character,Comment:Ds.blockComment,ProcessingInst:Ds.processingInstruction,DoctypeDecl:Ds.documentMeta}),Tp=Pd.deserialize({version:14,states:",xOVO!rOOO!WQ#tO'#CqO!]Q#tO'#CzO!bQ#tO'#C}O!gQ#tO'#DQO!lQ#tO'#DSO!qOaO'#CpO!|ObO'#CpO#XOdO'#CpO$eO!rO'#CpOOO`'#Cp'#CpO$lO$fO'#DTO$tQ#tO'#DVO$yQ#tO'#DWOOO`'#Dk'#DkOOO`'#DY'#DYQVO!rOOO%OQ&rO,59]O%ZQ&rO,59fO%fQ&rO,59iO%qQ&rO,59lO%|Q&rO,59nOOOa'#D^'#D^O&XOaO'#CxO&dOaO,59[OOOb'#D_'#D_O&lObO'#C{O&wObO,59[OOOd'#D`'#D`O'POdO'#DOO'[OdO,59[OOO`'#Da'#DaO'dO!rO,59[O'kQ#tO'#DROOO`,59[,59[OOOp'#Db'#DbO'pO$fO,59oOOO`,59o,59oO'xQ#|O,59qO'}Q#|O,59rOOO`-E7W-E7WO(SQ&rO'#CsOOQW'#DZ'#DZO(bQ&rO1G.wOOOa1G.w1G.wOOO`1G/Y1G/YO(mQ&rO1G/QOOOb1G/Q1G/QO(xQ&rO1G/TOOOd1G/T1G/TO)TQ&rO1G/WOOO`1G/W1G/WO)`Q&rO1G/YOOOa-E7[-E7[O)kQ#tO'#CyOOO`1G.v1G.vOOOb-E7]-E7]O)pQ#tO'#C|OOOd-E7^-E7^O)uQ#tO'#DPOOO`-E7_-E7_O)zQ#|O,59mOOOp-E7`-E7`OOO`1G/Z1G/ZOOO`1G/]1G/]OOO`1G/^1G/^O*PQ,UO,59_OOQW-E7X-E7XOOOa7+$c7+$cOOO`7+$t7+$tOOOb7+$l7+$lOOOd7+$o7+$oOOO`7+$r7+$rO*[Q#|O,59eO*aQ#|O,59hO*fQ#|O,59kOOO`1G/X1G/XO*kO7[O'#CvO*|OMhO'#CvOOQW1G.y1G.yOOO`1G/P1G/POOO`1G/S1G/SOOO`1G/V1G/VOOOO'#D['#D[O+_O7[O,59bOOQW,59b,59bOOOO'#D]'#D]O+pOMhO,59bOOOO-E7Y-E7YOOQW1G.|1G.|OOOO-E7Z-E7Z",stateData:",]~O!^OS~OUSOVPOWQOXROYTO[]O][O^^O`^Oa^Ob^Oc^Ox^O{_O!dZO~OfaO~OfbO~OfcO~OfdO~OfeO~O!WfOPlP!ZlP~O!XiOQoP!ZoP~O!YlORrP!ZrP~OUSOVPOWQOXROYTOZqO[]O][O^^O`^Oa^Ob^Oc^Ox^O!dZO~O!ZrO~P#dO![sO!euO~OfvO~OfwO~OS|OT}OhyO~OS!POT}OhyO~OS!ROT}OhyO~OS!TOT}OhyO~OS}OT}OhyO~O!WfOPlX!ZlX~OP!WO!Z!XO~O!XiOQoX!ZoX~OQ!ZO!Z!XO~O!YlORrX!ZrX~OR!]O!Z!XO~O!Z!XO~P#dOf!_O~O![sO!e!aO~OS!bO~OS!cO~Oi!dOSgXTgXhgX~OS!fOT!gOhyO~OS!hOT!gOhyO~OS!iOT!gOhyO~OS!jOT!gOhyO~OS!gOT!gOhyO~Of!kO~Of!lO~Of!mO~OS!nO~Ok!qO!`!oO!b!pO~OS!rO~OS!sO~OS!tO~Oa!uOb!uOc!uO!`!wO!a!uO~Oa!xOb!xOc!xO!b!wO!c!xO~Oa!uOb!uOc!uO!`!{O!a!uO~Oa!xOb!xOc!xO!b!{O!c!xO~OT~bac!dx{!d~",goto:"%p!`PPPPPPPPPPPPPPPPPPPP!a!gP!mPP!yP!|#P#S#Y#]#`#f#i#l#r#x!aP!a!aP$O$U$l$r$x%O%U%[%bPPPPPPPP%hX^OX`pXUOX`pezabcde{!O!Q!S!UR!q!dRhUR!XhXVOX`pRkVR!XkXWOX`pRnWR!XnXXOX`pQrXR!XpXYOX`pQ`ORx`Q{aQ!ObQ!QcQ!SdQ!UeZ!e{!O!Q!S!UQ!v!oR!z!vQ!y!pR!|!yQgUR!VgQjVR!YjQmWR![mQpXR!^pQtZR!`tS_O`ToXp",nodeNames:"⚠ StartCloseTag StartCloseTag StartCloseTag EndTag SelfClosingEndTag StartTag StartTag StartTag StartTag StartTag StartCloseTag StartCloseTag StartCloseTag IncompleteCloseTag Document Text EntityReference CharacterReference InvalidEntity Element OpenTag TagName Attribute AttributeName Is AttributeValue UnquotedAttributeValue ScriptText CloseTag OpenTag StyleText CloseTag OpenTag TextareaText CloseTag OpenTag CloseTag SelfClosingTag Comment ProcessingInst MismatchedCloseTag CloseTag DoctypeDecl",maxTerm:67,context:yp,nodeProps:[["closedBy",-10,1,2,3,7,8,9,10,11,12,13,"EndTag",6,"EndTag SelfClosingEndTag",-4,21,30,33,36,"CloseTag"],["openedBy",4,"StartTag StartCloseTag",5,"StartTag",-4,29,32,35,37,"OpenTag"],["group",-9,14,17,18,19,20,39,40,41,42,"Entity",16,"Entity TextContent",-3,28,31,34,"TextContent Entity"],["isolate",-11,21,29,30,32,33,35,36,37,38,41,42,"ltr",-3,26,27,39,""]],propSources:[Cp],skippedNodes:[0],repeatNodeCount:9,tokenData:"!<p!aR!YOX$qXY,QYZ,QZ[$q[]&X]^,Q^p$qpq,Qqr-_rs3_sv-_vw3}wxHYx}-_}!OH{!O!P-_!P!Q$q!Q![-_![!]Mz!]!^-_!^!_!$S!_!`!;x!`!a&X!a!c-_!c!}Mz!}#R-_#R#SMz#S#T1k#T#oMz#o#s-_#s$f$q$f%W-_%W%oMz%o%p-_%p&aMz&a&b-_&b1pMz1p4U-_4U4dMz4d4e-_4e$ISMz$IS$I`-_$I`$IbMz$Ib$Kh-_$Kh%#tMz%#t&/x-_&/x&EtMz&Et&FV-_&FV;'SMz;'S;:j!#|;:j;=`3X<%l?&r-_?&r?AhMz?Ah?BY$q?BY?MnMz?MnO$q!Z$|c`PkW!a`!cpOX$qXZ&XZ[$q[^&X^p$qpq&Xqr$qrs&}sv$qvw+Pwx(tx!^$q!^!_*V!_!a&X!a#S$q#S#T&X#T;'S$q;'S;=`+z<%lO$q!R&bX`P!a`!cpOr&Xrs&}sv&Xwx(tx!^&X!^!_*V!_;'S&X;'S;=`*y<%lO&Xq'UV`P!cpOv&}wx'kx!^&}!^!_(V!_;'S&};'S;=`(n<%lO&}P'pT`POv'kw!^'k!_;'S'k;'S;=`(P<%lO'kP(SP;=`<%l'kp([S!cpOv(Vx;'S(V;'S;=`(h<%lO(Vp(kP;=`<%l(Vq(qP;=`<%l&}a({W`P!a`Or(trs'ksv(tw!^(t!^!_)e!_;'S(t;'S;=`*P<%lO(t`)jT!a`Or)esv)ew;'S)e;'S;=`)y<%lO)e`)|P;=`<%l)ea*SP;=`<%l(t!Q*^V!a`!cpOr*Vrs(Vsv*Vwx)ex;'S*V;'S;=`*s<%lO*V!Q*vP;=`<%l*V!R*|P;=`<%l&XW+UYkWOX+PZ[+P^p+Pqr+Psw+Px!^+P!a#S+P#T;'S+P;'S;=`+t<%lO+PW+wP;=`<%l+P!Z+}P;=`<%l$q!a,]``P!a`!cp!^^OX&XXY,QYZ,QZ]&X]^,Q^p&Xpq,Qqr&Xrs&}sv&Xwx(tx!^&X!^!_*V!_;'S&X;'S;=`*y<%lO&X!_-ljhS`PkW!a`!cpOX$qXZ&XZ[$q[^&X^p$qpq&Xqr-_rs&}sv-_vw/^wx(tx!P-_!P!Q$q!Q!^-_!^!_*V!_!a&X!a#S-_#S#T1k#T#s-_#s$f$q$f;'S-_;'S;=`3X<%l?Ah-_?Ah?BY$q?BY?Mn-_?MnO$q[/ebhSkWOX+PZ[+P^p+Pqr/^sw/^x!P/^!P!Q+P!Q!^/^!a#S/^#S#T0m#T#s/^#s$f+P$f;'S/^;'S;=`1e<%l?Ah/^?Ah?BY+P?BY?Mn/^?MnO+PS0rXhSqr0msw0mx!P0m!Q!^0m!a#s0m$f;'S0m;'S;=`1_<%l?Ah0m?BY?Mn0mS1bP;=`<%l0m[1hP;=`<%l/^!V1vchS`P!a`!cpOq&Xqr1krs&}sv1kvw0mwx(tx!P1k!P!Q&X!Q!^1k!^!_*V!_!a&X!a#s1k#s$f&X$f;'S1k;'S;=`3R<%l?Ah1k?Ah?BY&X?BY?Mn1k?MnO&X!V3UP;=`<%l1k!_3[P;=`<%l-_!Z3hV!`h`P!cpOv&}wx'kx!^&}!^!_(V!_;'S&};'S;=`(n<%lO&}!_4WihSkWc!ROX5uXZ7SZ[5u[^7S^p5uqr8trs7Sst>]tw8twx7Sx!P8t!P!Q5u!Q!]8t!]!^/^!^!a7S!a#S8t#S#T;{#T#s8t#s$f5u$f;'S8t;'S;=`>V<%l?Ah8t?Ah?BY5u?BY?Mn8t?MnO5u!Z5zbkWOX5uXZ7SZ[5u[^7S^p5uqr5urs7Sst+Ptw5uwx7Sx!]5u!]!^7w!^!a7S!a#S5u#S#T7S#T;'S5u;'S;=`8n<%lO5u!R7VVOp7Sqs7St!]7S!]!^7l!^;'S7S;'S;=`7q<%lO7S!R7qOa!R!R7tP;=`<%l7S!Z8OYkWa!ROX+PZ[+P^p+Pqr+Psw+Px!^+P!a#S+P#T;'S+P;'S;=`+t<%lO+P!Z8qP;=`<%l5u!_8{ihSkWOX5uXZ7SZ[5u[^7S^p5uqr8trs7Sst/^tw8twx7Sx!P8t!P!Q5u!Q!]8t!]!^:j!^!a7S!a#S8t#S#T;{#T#s8t#s$f5u$f;'S8t;'S;=`>V<%l?Ah8t?Ah?BY5u?BY?Mn8t?MnO5u!_:sbhSkWa!ROX+PZ[+P^p+Pqr/^sw/^x!P/^!P!Q+P!Q!^/^!a#S/^#S#T0m#T#s/^#s$f+P$f;'S/^;'S;=`1e<%l?Ah/^?Ah?BY+P?BY?Mn/^?MnO+P!V<QchSOp7Sqr;{rs7Sst0mtw;{wx7Sx!P;{!P!Q7S!Q!];{!]!^=]!^!a7S!a#s;{#s$f7S$f;'S;{;'S;=`>P<%l?Ah;{?Ah?BY7S?BY?Mn;{?MnO7S!V=dXhSa!Rqr0msw0mx!P0m!Q!^0m!a#s0m$f;'S0m;'S;=`1_<%l?Ah0m?BY?Mn0m!V>SP;=`<%l;{!_>YP;=`<%l8t!_>dhhSkWOX@OXZAYZ[@O[^AY^p@OqrBwrsAYswBwwxAYx!PBw!P!Q@O!Q!]Bw!]!^/^!^!aAY!a#SBw#S#TE{#T#sBw#s$f@O$f;'SBw;'S;=`HS<%l?AhBw?Ah?BY@O?BY?MnBw?MnO@O!Z@TakWOX@OXZAYZ[@O[^AY^p@Oqr@OrsAYsw@OwxAYx!]@O!]!^Az!^!aAY!a#S@O#S#TAY#T;'S@O;'S;=`Bq<%lO@O!RA]UOpAYq!]AY!]!^Ao!^;'SAY;'S;=`At<%lOAY!RAtOb!R!RAwP;=`<%lAY!ZBRYkWb!ROX+PZ[+P^p+Pqr+Psw+Px!^+P!a#S+P#T;'S+P;'S;=`+t<%lO+P!ZBtP;=`<%l@O!_COhhSkWOX@OXZAYZ[@O[^AY^p@OqrBwrsAYswBwwxAYx!PBw!P!Q@O!Q!]Bw!]!^Dj!^!aAY!a#SBw#S#TE{#T#sBw#s$f@O$f;'SBw;'S;=`HS<%l?AhBw?Ah?BY@O?BY?MnBw?MnO@O!_DsbhSkWb!ROX+PZ[+P^p+Pqr/^sw/^x!P/^!P!Q+P!Q!^/^!a#S/^#S#T0m#T#s/^#s$f+P$f;'S/^;'S;=`1e<%l?Ah/^?Ah?BY+P?BY?Mn/^?MnO+P!VFQbhSOpAYqrE{rsAYswE{wxAYx!PE{!P!QAY!Q!]E{!]!^GY!^!aAY!a#sE{#s$fAY$f;'SE{;'S;=`G|<%l?AhE{?Ah?BYAY?BY?MnE{?MnOAY!VGaXhSb!Rqr0msw0mx!P0m!Q!^0m!a#s0m$f;'S0m;'S;=`1_<%l?Ah0m?BY?Mn0m!VHPP;=`<%lE{!_HVP;=`<%lBw!ZHcW!bx`P!a`Or(trs'ksv(tw!^(t!^!_)e!_;'S(t;'S;=`*P<%lO(t!aIYlhS`PkW!a`!cpOX$qXZ&XZ[$q[^&X^p$qpq&Xqr-_rs&}sv-_vw/^wx(tx}-_}!OKQ!O!P-_!P!Q$q!Q!^-_!^!_*V!_!a&X!a#S-_#S#T1k#T#s-_#s$f$q$f;'S-_;'S;=`3X<%l?Ah-_?Ah?BY$q?BY?Mn-_?MnO$q!aK_khS`PkW!a`!cpOX$qXZ&XZ[$q[^&X^p$qpq&Xqr-_rs&}sv-_vw/^wx(tx!P-_!P!Q$q!Q!^-_!^!_*V!_!`&X!`!aMS!a#S-_#S#T1k#T#s-_#s$f$q$f;'S-_;'S;=`3X<%l?Ah-_?Ah?BY$q?BY?Mn-_?MnO$q!TM_X`P!a`!cp!eQOr&Xrs&}sv&Xwx(tx!^&X!^!_*V!_;'S&X;'S;=`*y<%lO&X!aNZ!ZhSfQ`PkW!a`!cpOX$qXZ&XZ[$q[^&X^p$qpq&Xqr-_rs&}sv-_vw/^wx(tx}-_}!OMz!O!PMz!P!Q$q!Q![Mz![!]Mz!]!^-_!^!_*V!_!a&X!a!c-_!c!}Mz!}#R-_#R#SMz#S#T1k#T#oMz#o#s-_#s$f$q$f$}-_$}%OMz%O%W-_%W%oMz%o%p-_%p&aMz&a&b-_&b1pMz1p4UMz4U4dMz4d4e-_4e$ISMz$IS$I`-_$I`$IbMz$Ib$Je-_$Je$JgMz$Jg$Kh-_$Kh%#tMz%#t&/x-_&/x&EtMz&Et&FV-_&FV;'SMz;'S;:j!#|;:j;=`3X<%l?&r-_?&r?AhMz?Ah?BY$q?BY?MnMz?MnO$q!a!$PP;=`<%lMz!R!$ZY!a`!cpOq*Vqr!$yrs(Vsv*Vwx)ex!a*V!a!b!4t!b;'S*V;'S;=`*s<%lO*V!R!%Q]!a`!cpOr*Vrs(Vsv*Vwx)ex}*V}!O!%y!O!f*V!f!g!']!g#W*V#W#X!0`#X;'S*V;'S;=`*s<%lO*V!R!&QX!a`!cpOr*Vrs(Vsv*Vwx)ex}*V}!O!&m!O;'S*V;'S;=`*s<%lO*V!R!&vV!a`!cp!dPOr*Vrs(Vsv*Vwx)ex;'S*V;'S;=`*s<%lO*V!R!'dX!a`!cpOr*Vrs(Vsv*Vwx)ex!q*V!q!r!(P!r;'S*V;'S;=`*s<%lO*V!R!(WX!a`!cpOr*Vrs(Vsv*Vwx)ex!e*V!e!f!(s!f;'S*V;'S;=`*s<%lO*V!R!(zX!a`!cpOr*Vrs(Vsv*Vwx)ex!v*V!v!w!)g!w;'S*V;'S;=`*s<%lO*V!R!)nX!a`!cpOr*Vrs(Vsv*Vwx)ex!{*V!{!|!*Z!|;'S*V;'S;=`*s<%lO*V!R!*bX!a`!cpOr*Vrs(Vsv*Vwx)ex!r*V!r!s!*}!s;'S*V;'S;=`*s<%lO*V!R!+UX!a`!cpOr*Vrs(Vsv*Vwx)ex!g*V!g!h!+q!h;'S*V;'S;=`*s<%lO*V!R!+xY!a`!cpOr!+qrs!,hsv!+qvw!-Swx!.[x!`!+q!`!a!/j!a;'S!+q;'S;=`!0Y<%lO!+qq!,mV!cpOv!,hvx!-Sx!`!,h!`!a!-q!a;'S!,h;'S;=`!.U<%lO!,hP!-VTO!`!-S!`!a!-f!a;'S!-S;'S;=`!-k<%lO!-SP!-kO{PP!-nP;=`<%l!-Sq!-xS!cp{POv(Vx;'S(V;'S;=`(h<%lO(Vq!.XP;=`<%l!,ha!.aX!a`Or!.[rs!-Ssv!.[vw!-Sw!`!.[!`!a!.|!a;'S!.[;'S;=`!/d<%lO!.[a!/TT!a`{POr)esv)ew;'S)e;'S;=`)y<%lO)ea!/gP;=`<%l!.[!R!/sV!a`!cp{POr*Vrs(Vsv*Vwx)ex;'S*V;'S;=`*s<%lO*V!R!0]P;=`<%l!+q!R!0gX!a`!cpOr*Vrs(Vsv*Vwx)ex#c*V#c#d!1S#d;'S*V;'S;=`*s<%lO*V!R!1ZX!a`!cpOr*Vrs(Vsv*Vwx)ex#V*V#V#W!1v#W;'S*V;'S;=`*s<%lO*V!R!1}X!a`!cpOr*Vrs(Vsv*Vwx)ex#h*V#h#i!2j#i;'S*V;'S;=`*s<%lO*V!R!2qX!a`!cpOr*Vrs(Vsv*Vwx)ex#m*V#m#n!3^#n;'S*V;'S;=`*s<%lO*V!R!3eX!a`!cpOr*Vrs(Vsv*Vwx)ex#d*V#d#e!4Q#e;'S*V;'S;=`*s<%lO*V!R!4XX!a`!cpOr*Vrs(Vsv*Vwx)ex#X*V#X#Y!+q#Y;'S*V;'S;=`*s<%lO*V!R!4{Y!a`!cpOr!4trs!5ksv!4tvw!6Vwx!8]x!a!4t!a!b!:]!b;'S!4t;'S;=`!;r<%lO!4tq!5pV!cpOv!5kvx!6Vx!a!5k!a!b!7W!b;'S!5k;'S;=`!8V<%lO!5kP!6YTO!a!6V!a!b!6i!b;'S!6V;'S;=`!7Q<%lO!6VP!6lTO!`!6V!`!a!6{!a;'S!6V;'S;=`!7Q<%lO!6VP!7QOxPP!7TP;=`<%l!6Vq!7]V!cpOv!5kvx!6Vx!`!5k!`!a!7r!a;'S!5k;'S;=`!8V<%lO!5kq!7yS!cpxPOv(Vx;'S(V;'S;=`(h<%lO(Vq!8YP;=`<%l!5ka!8bX!a`Or!8]rs!6Vsv!8]vw!6Vw!a!8]!a!b!8}!b;'S!8];'S;=`!:V<%lO!8]a!9SX!a`Or!8]rs!6Vsv!8]vw!6Vw!`!8]!`!a!9o!a;'S!8];'S;=`!:V<%lO!8]a!9vT!a`xPOr)esv)ew;'S)e;'S;=`)y<%lO)ea!:YP;=`<%l!8]!R!:dY!a`!cpOr!4trs!5ksv!4tvw!6Vwx!8]x!`!4t!`!a!;S!a;'S!4t;'S;=`!;r<%lO!4t!R!;]V!a`!cpxPOr*Vrs(Vsv*Vwx)ex;'S*V;'S;=`*s<%lO*V!R!;uP;=`<%l!4t!V!<TXiS`P!a`!cpOr&Xrs&}sv&Xwx(tx!^&X!^!_*V!_;'S&X;'S;=`*y<%lO&X",tokenizers:[Ep,Sp,_p,kp,wp,Op,0,1,2,3,4,5],topRules:{Document:[0,15]},dialects:{noMatch:0,selfClosing:509},tokenPrec:511});function Ap(e,t){let n=Object.create(null);for(let r of e.getChildren(23)){let e=r.getChild(24),a=r.getChild(26)||r.getChild(27);e&&(n[t.read(e.from,e.to)]=a?26==a.type.id?t.read(a.from+1,a.to-1):t.read(a.from,a.to):"")}return n}function Rp(e,t){let n=e.getChild(22);return n?t.read(n.from,n.to):" "}function Pp(e,t,n){let r;for(let a of n)if(!a.attrs||a.attrs(r||(r=Ap(e.node.parent.firstChild,t))))return{parser:a.parser};return null}function Dp(e=[],t=[]){let n=[],r=[],a=[],o=[];for(let t of e)("script"==t.tag?n:"style"==t.tag?r:"textarea"==t.tag?a:o).push(t);let i=t.length?Object.create(null):null;for(let e of t)(i[e.name]||(i[e.name]=[])).push(e);return s=(e,t)=>{let s=e.type.id;if(28==s)return Pp(e,t,n);if(31==s)return Pp(e,t,r);if(34==s)return Pp(e,t,a);if(20==s&&o.length){let n,r=e.node,a=r.firstChild,i=a&&Rp(a,t);if(i)for(let e of o)if(e.tag==i&&(!e.attrs||e.attrs(n||(n=Ap(a,t))))){let t=r.lastChild,n=37==t.type.id?t.from:r.to;if(n>a.to)return{parser:e.parser,overlay:[{from:a.to,to:n}]}}}if(i&&23==s){let n,r=e.node;if(n=r.firstChild){let e=i[t.read(n.from,n.to)];if(e)for(let n of e){if(n.tagName&&n.tagName!=Rp(r.parent,t))continue;let e=r.lastChild;if(26==e.type.id){let t=e.from+1,r=e.lastChild,a=e.to-(r&&r.isError?0:1);if(a>t)return{parser:n.parser,overlay:[{from:t,to:a}]}}else if(27==e.type.id)return{parser:n.parser,overlay:[{from:e.from,to:e.to}]}}}}return null},(e,t,n,r)=>new Yi(e,s,t,n,r);var s}const Mp=[9,10,11,12,13,32,133,160,5760,8192,8193,8194,8195,8196,8197,8198,8199,8200,8201,8202,8232,8233,8239,8287,12288],$p=new Rd({start:!1,shift:(e,t)=>5==t||6==t||318==t?e:319==t,strict:!1}),Np=new vd(((e,t)=>{let{next:n}=e;(125==n||-1==n||t.context)&&e.acceptToken(316)}),{contextual:!0,fallback:!0}),Ip=new vd(((e,t)=>{let n,{next:r}=e;Mp.indexOf(r)>-1||(47!=r||47!=(n=e.peek(1))&&42!=n)&&(125==r||59==r||-1==r||t.context||e.acceptToken(314))}),{contextual:!0}),Lp=new vd(((e,t)=>{91!=e.next||t.context||e.acceptToken(315)}),{contextual:!0}),Bp=new vd(((e,t)=>{let{next:n}=e;if(43==n||45==n){if(e.advance(),n==e.next){e.advance();let n=!t.context&&t.canShift(1);e.acceptToken(n?1:2)}}else 63==n&&46==e.peek(1)&&(e.advance(),e.advance(),(e.next<48||e.next>57)&&e.acceptToken(3))}),{contextual:!0});function Fp(e,t){return e>=65&&e<=90||e>=97&&e<=122||95==e||e>=192||!t&&e>=48&&e<=57}const Qp=new vd(((e,t)=>{if(60!=e.next||!t.dialectEnabled(0))return;if(e.advance(),47==e.next)return;let n=0;for(;Mp.indexOf(e.next)>-1;)e.advance(),n++;if(Fp(e.next,!0)){for(e.advance(),n++;Fp(e.next,!1);)e.advance(),n++;for(;Mp.indexOf(e.next)>-1;)e.advance(),n++;if(44==e.next)return;for(let t=0;;t++){if(7==t){if(!Fp(e.next,!0))return;break}if(e.next!="extends".charCodeAt(t))break;e.advance(),n++}}e.acceptToken(4,-n)})),Up=ds({"get set async static":Ds.modifier,"for while do if else switch try catch finally return throw break continue default case":Ds.controlKeyword,"in of await yield void typeof delete instanceof":Ds.operatorKeyword,"let var const using function class extends":Ds.definitionKeyword,"import export from":Ds.moduleKeyword,"with debugger as new":Ds.keyword,TemplateString:Ds.special(Ds.string),super:Ds.atom,BooleanLiteral:Ds.bool,this:Ds.self,null:Ds.null,Star:Ds.modifier,VariableName:Ds.variableName,"CallExpression/VariableName TaggedTemplateExpression/VariableName":Ds.function(Ds.variableName),VariableDefinition:Ds.definition(Ds.variableName),Label:Ds.labelName,PropertyName:Ds.propertyName,PrivatePropertyName:Ds.special(Ds.propertyName),"CallExpression/MemberExpression/PropertyName":Ds.function(Ds.propertyName),"FunctionDeclaration/VariableDefinition":Ds.function(Ds.definition(Ds.variableName)),"ClassDeclaration/VariableDefinition":Ds.definition(Ds.className),"NewExpression/VariableName":Ds.className,PropertyDefinition:Ds.definition(Ds.propertyName),PrivatePropertyDefinition:Ds.definition(Ds.special(Ds.propertyName)),UpdateOp:Ds.updateOperator,"LineComment Hashbang":Ds.lineComment,BlockComment:Ds.blockComment,Number:Ds.number,String:Ds.string,Escape:Ds.escape,ArithOp:Ds.arithmeticOperator,LogicOp:Ds.logicOperator,BitOp:Ds.bitwiseOperator,CompareOp:Ds.compareOperator,RegExp:Ds.regexp,Equals:Ds.definitionOperator,Arrow:Ds.function(Ds.punctuation),": Spread":Ds.punctuation,"( )":Ds.paren,"[ ]":Ds.squareBracket,"{ }":Ds.brace,"InterpolationStart InterpolationEnd":Ds.special(Ds.brace),".":Ds.derefOperator,", ;":Ds.separator,"@":Ds.meta,TypeName:Ds.typeName,TypeDefinition:Ds.definition(Ds.typeName),"type enum interface implements namespace module declare":Ds.definitionKeyword,"abstract global Privacy readonly override":Ds.modifier,"is keyof unique infer asserts":Ds.operatorKeyword,JSXAttributeValue:Ds.attributeValue,JSXText:Ds.content,"JSXStartTag JSXStartCloseTag JSXSelfCloseEndTag JSXEndTag":Ds.angleBracket,"JSXIdentifier JSXNameSpacedName":Ds.tagName,"JSXAttribute/JSXIdentifier JSXAttribute/JSXNameSpacedName":Ds.attributeName,"JSXBuiltin/JSXIdentifier":Ds.standard(Ds.tagName)}),zp={__proto__:null,export:20,as:25,from:33,default:36,async:41,function:42,const:52,extends:56,this:60,true:68,false:68,null:80,void:84,typeof:88,super:104,new:138,delete:150,yield:159,await:163,class:168,public:231,private:231,protected:231,readonly:233,instanceof:252,satisfies:255,in:256,import:290,keyof:347,unique:351,infer:357,asserts:393,is:395,abstract:415,implements:417,type:419,let:422,var:424,using:427,interface:433,enum:437,namespace:443,module:445,declare:449,global:453,for:472,of:481,while:484,with:488,do:492,if:496,else:498,switch:502,case:508,try:514,catch:518,finally:522,return:526,throw:530,break:534,continue:538,debugger:542},jp={__proto__:null,async:125,get:127,set:129,declare:191,public:193,private:193,protected:193,static:195,abstract:197,override:199,readonly:205,accessor:207,new:399},Wp={__proto__:null,"<":189},Xp=Pd.deserialize({version:14,states:"$EOQ%TQlOOO%[QlOOO'_QpOOP(lO`OOO*zQ!0MxO'#CiO+RO#tO'#CjO+aO&jO'#CjO+oO#@ItO'#D_O.QQlO'#DeO.bQlO'#DpO%[QlO'#DxO0fQlO'#EQOOQ!0Lf'#EY'#EYO1PQ`O'#EVOOQO'#En'#EnOOQO'#Ij'#IjO1XQ`O'#GrO1dQ`O'#EmO1iQ`O'#EmO3hQ!0MxO'#JpO6[Q!0MxO'#JqO6uQ`O'#F[O6zQ,UO'#FsOOQ!0Lf'#Fe'#FeO7VO7dO'#FeO7eQMhO'#F{O9UQ`O'#FzOOQ!0Lf'#Jq'#JqOOQ!0Lb'#Jp'#JpO9ZQ`O'#GvOOQ['#K]'#K]O9fQ`O'#IWO9kQ!0LrO'#IXOOQ['#J^'#J^OOQ['#I]'#I]Q`QlOOQ`QlOOO9sQ!L^O'#DtO9zQlO'#D|O:RQlO'#EOO9aQ`O'#GrO:YQMhO'#CoO:hQ`O'#ElO:sQ`O'#EwO:xQMhO'#FdO;gQ`O'#GrOOQO'#K^'#K^O;lQ`O'#K^O;zQ`O'#GzO;zQ`O'#G{O;zQ`O'#G}O9aQ`O'#HQO<qQ`O'#HTO>YQ`O'#CeO>jQ`O'#HaO>rQ`O'#HgO>rQ`O'#HiO`QlO'#HkO>rQ`O'#HmO>rQ`O'#HpO>wQ`O'#HvO>|Q!0LsO'#H|O%[QlO'#IOO?XQ!0LsO'#IQO?dQ!0LsO'#ISO9kQ!0LrO'#IUO?oQ!0MxO'#CiO@qQpO'#DjQOQ`OOO%[QlO'#EOOAXQ`O'#ERO:YQMhO'#ElOAdQ`O'#ElOAoQ!bO'#FdOOQ['#Cg'#CgOOQ!0Lb'#Do'#DoOOQ!0Lb'#Jt'#JtO%[QlO'#JtOOQO'#Jw'#JwOOQO'#If'#IfOBoQpO'#EeOOQ!0Lb'#Ed'#EdOOQ!0Lb'#J{'#J{OCkQ!0MSO'#EeOCuQpO'#EUOOQO'#Jv'#JvODZQpO'#JwOEhQpO'#EUOCuQpO'#EePEuO&2DjO'#CbPOOO)CD{)CD{OOOO'#I^'#I^OFQO#tO,59UOOQ!0Lh,59U,59UOOOO'#I_'#I_OF`O&jO,59UOFnQ!L^O'#DaOOOO'#Ia'#IaOFuO#@ItO,59yOOQ!0Lf,59y,59yOGTQlO'#IbOGhQ`O'#JrOIgQ!fO'#JrO+}QlO'#JrOInQ`O,5:POJUQ`O'#EnOJcQ`O'#KROJnQ`O'#KQOJnQ`O'#KQOJvQ`O,5;[OJ{Q`O'#KPOOQ!0Ln,5:[,5:[OKSQlO,5:[OMQQ!0MxO,5:dOMqQ`O,5:lON[Q!0LrO'#KOONcQ`O'#J}O9ZQ`O'#J}ONwQ`O'#J}O! PQ`O,5;ZO! UQ`O'#J}O!#ZQ!fO'#JqOOQ!0Lh'#Ci'#CiO%[QlO'#EQO!#yQ!fO,5:qOOQS'#Jx'#JxOOQO-E<h-E<hO9aQ`O,5=^O!$aQ`O,5=^O!$fQlO,5;XO!&iQMhO'#EiO!(SQ`O,5;XO!(XQlO'#DwO!(cQpO,5;bO!(kQpO,5;bO%[QlO,5;bOOQ['#FS'#FSOOQ['#FU'#FUO%[QlO,5;cO%[QlO,5;cO%[QlO,5;cO%[QlO,5;cO%[QlO,5;cO%[QlO,5;cO%[QlO,5;cO%[QlO,5;cO%[QlO,5;cO%[QlO,5;cOOQ['#FY'#FYO!(yQlO,5;sOOQ!0Lf,5;x,5;xOOQ!0Lf,5;y,5;yOOQ!0Lf,5;{,5;{O%[QlO'#InO!*|Q!0LrO,5<hO%[QlO,5;cO!&iQMhO,5;cO!+kQMhO,5;cO!-]QMhO'#E[O%[QlO,5;vOOQ!0Lf,5;z,5;zO!-dQ,UO'#FiO!.aQ,UO'#KVO!-{Q,UO'#KVO!.hQ,UO'#KVOOQO'#KV'#KVO!.|Q,UO,5<ROOOW,5<_,5<_O!/_QlO'#FuOOOW'#Im'#ImO7VO7dO,5<PO!/fQ,UO'#FwOOQ!0Lf,5<P,5<PO!0VQ$IUO'#CwOOQ!0Lh'#C{'#C{O!0jO#@ItO'#DPO!1WQMjO,5<dO!1_Q`O,5<gO!2zQ(CWO'#GWO!3XQ`O'#GXO!3^Q`O'#GXO!4|Q(CWO'#G]O!6RQpO'#GaOOQO'#Gm'#GmO!+rQMhO'#GlOOQO'#Go'#GoO!+rQMhO'#GnO!6tQ$IUO'#JjOOQ!0Lh'#Jj'#JjO!7OQ`O'#JiO!7^Q`O'#JhO!7fQ`O'#CuOOQ!0Lh'#Cy'#CyO!7qQ`O'#C{OOQ!0Lh'#DT'#DTOOQ!0Lh'#DV'#DVO1SQ`O'#DXO!+rQMhO'#GOO!+rQMhO'#GQO!7vQ`O'#GSO!7{Q`O'#GTO!3^Q`O'#GZO!+rQMhO'#G`O;zQ`O'#JiO!8QQ`O'#EoO!8oQ`O,5<fOOQ!0Lb'#Cr'#CrO!8wQ`O'#EpO!9qQpO'#EqOOQ!0Lb'#KP'#KPO!9xQ!0LrO'#K_O9kQ!0LrO,5=bO`QlO,5>rOOQ['#Jf'#JfOOQ[,5>s,5>sOOQ[-E<Z-E<ZO!;wQ!0MxO,5:`O!9lQpO,5:^O!>bQ!0MxO,5:hO%[QlO,5:hO!@xQ!0MxO,5:jOOQO,5@x,5@xO!AiQMhO,5=^O!AwQ!0LrO'#JgO9UQ`O'#JgO!BYQ!0LrO,59ZO!BeQpO,59ZO!BmQMhO,59ZO:YQMhO,59ZO!BxQ`O,5;XO!CQQ`O'#H`O!CfQ`O'#KbO%[QlO,5;|O!9lQpO,5<OO!CnQ`O,5=yO!CsQ`O,5=yO!CxQ`O,5=yO9kQ!0LrO,5=yO;zQ`O,5=iOOQO'#Cw'#CwO!DWQpO,5=fO!D`QMhO,5=gO!DkQ`O,5=iO!DpQ!bO,5=lO!DxQ`O'#K^O>wQ`O'#HVO9aQ`O'#HXO!D}Q`O'#HXO:YQMhO'#HZO!ESQ`O'#HZOOQ[,5=o,5=oO!EXQ`O'#H[O!EjQ`O'#CoO!EoQ`O,59PO!EyQ`O,59PO!HOQlO,59POOQ[,59P,59PO!H`Q!0LrO,59PO%[QlO,59PO!JkQlO'#HcOOQ['#Hd'#HdOOQ['#He'#HeO`QlO,5={O!KRQ`O,5={O`QlO,5>RO`QlO,5>TO!KWQ`O,5>VO`QlO,5>XO!K]Q`O,5>[O!KbQlO,5>bOOQ[,5>h,5>hO%[QlO,5>hO9kQ!0LrO,5>jOOQ[,5>l,5>lO# lQ`O,5>lOOQ[,5>n,5>nO# lQ`O,5>nOOQ[,5>p,5>pO#!YQpO'#D]O%[QlO'#JtO#!{QpO'#JtO##VQpO'#DkO##hQpO'#DkO#%yQlO'#DkO#&QQ`O'#JsO#&YQ`O,5:UO#&_Q`O'#ErO#&mQ`O'#KSO#&uQ`O,5;]O#&zQpO'#DkO#'XQpO'#ETOOQ!0Lf,5:m,5:mO%[QlO,5:mO#'`Q`O,5:mO>wQ`O,5;WO!BeQpO,5;WO!BmQMhO,5;WO:YQMhO,5;WO#'hQ`O,5@`O#'mQ07dO,5:qOOQO-E<d-E<dO#(sQ!0MSO,5;POCuQpO,5:pO#(}QpO,5:pOCuQpO,5;PO!BYQ!0LrO,5:pOOQ!0Lb'#Eh'#EhOOQO,5;P,5;PO%[QlO,5;PO#)[Q!0LrO,5;PO#)gQ!0LrO,5;PO!BeQpO,5:pOOQO,5;V,5;VO#)uQ!0LrO,5;PPOOO'#I['#I[P#*ZO&2DjO,58|POOO,58|,58|OOOO-E<[-E<[OOQ!0Lh1G.p1G.pOOOO-E<]-E<]OOOO,59{,59{O#*fQ!bO,59{OOOO-E<_-E<_OOQ!0Lf1G/e1G/eO#*kQ!fO,5>|O+}QlO,5>|OOQO,5?S,5?SO#*uQlO'#IbOOQO-E<`-E<`O#+SQ`O,5@^O#+[Q!fO,5@^O#+cQ`O,5@lOOQ!0Lf1G/k1G/kO%[QlO,5@mO#+kQ`O'#IhOOQO-E<f-E<fO#+cQ`O,5@lOOQ!0Lb1G0v1G0vOOQ!0Ln1G/v1G/vOOQ!0Ln1G0W1G0WO%[QlO,5@jO#,PQ!0LrO,5@jO#,bQ!0LrO,5@jO#,iQ`O,5@iO9ZQ`O,5@iO#,qQ`O,5@iO#-PQ`O'#IkO#,iQ`O,5@iOOQ!0Lb1G0u1G0uO!(cQpO,5:sO!(nQpO,5:sOOQS,5:u,5:uO#-qQdO,5:uO#-yQMhO1G2xO9aQ`O1G2xOOQ!0Lf1G0s1G0sO#.XQ!0MxO1G0sO#/^Q!0MvO,5;TOOQ!0Lh'#GV'#GVO#/zQ!0MzO'#JjO!$fQlO1G0sO#2VQ!fO'#JuO%[QlO'#JuO#2aQ`O,5:cOOQ!0Lh'#D]'#D]OOQ!0Lf1G0|1G0|O%[QlO1G0|OOQ!0Lf1G1e1G1eO#2fQ`O1G0|O#4zQ!0MxO1G0}O#5RQ!0MxO1G0}O#7iQ!0MxO1G0}O#7pQ!0MxO1G0}O#:WQ!0MxO1G0}O#<nQ!0MxO1G0}O#<uQ!0MxO1G0}O#<|Q!0MxO1G0}O#?dQ!0MxO1G0}O#?kQ!0MxO1G0}O#AxQ?MtO'#CiO#CsQ?MtO1G1_O#CzQ?MtO'#JqO#D_Q!0MxO,5?YOOQ!0Lb-E<l-E<lO#FlQ!0MxO1G0}O#GiQ!0MzO1G0}OOQ!0Lf1G0}1G0}O#HlQMjO'#JzO#HvQ`O,5:vO#H{Q!0MxO1G1bO#IoQ,UO,5<VO#IwQ,UO,5<WO#JPQ,UO'#FnO#JhQ`O'#FmOOQO'#KW'#KWOOQO'#Il'#IlO#JmQ,UO1G1mOOQ!0Lf1G1m1G1mOOOW1G1x1G1xO#KOQ?MtO'#JpO#KYQ`O,5<aO!(yQlO,5<aOOOW-E<k-E<kOOQ!0Lf1G1k1G1kO#K_QpO'#KVOOQ!0Lf,5<c,5<cO#KgQpO,5<cO#KlQMhO'#DROOOO'#I`'#I`O#KsO#@ItO,59kOOQ!0Lh,59k,59kO%[QlO1G2OO!7{Q`O'#IpO#LOQ`O,5<yOOQ!0Lh,5<v,5<vO!+rQMhO'#IsO#LlQMjO,5=WO!+rQMhO'#IuO#M_QMjO,5=YO!&iQMhO,5=[OOQO1G2R1G2RO#MiQ!dO'#CrO#M|Q(CWO'#EpO$ RQpO'#GaO$ iQ!dO,5<rO$ pQ`O'#KYO9ZQ`O'#KYO$!OQ`O,5<tO!+rQMhO,5<sO$!TQ`O'#GYO$!fQ`O,5<sO$!kQ!dO'#GVO$!xQ!dO'#KZO$#SQ`O'#KZO!&iQMhO'#KZO$#XQ`O,5<wO$#^QlO'#JtO$#hQpO'#GbO##hQpO'#GbO$#yQ`O'#GfO!3^Q`O'#GjO$$OQ!0LrO'#IrO$$ZQpO,5<{OOQ!0Lp,5<{,5<{O$$bQpO'#GbO$$oQpO'#GcO$%QQpO'#GcO$%VQMjO,5=WO$%gQMjO,5=YOOQ!0Lh,5=],5=]O!+rQMhO,5@TO!+rQMhO,5@TO$%wQ`O'#IwO$&VQ`O,5@SO$&_Q`O,59aOOQ!0Lh,59g,59gO$'UQ$IYO,59sOOQ!0Lh'#Jn'#JnO$'wQMjO,5<jO$(jQMjO,5<lO@iQ`O,5<nOOQ!0Lh,5<o,5<oO$(tQ`O,5<uO$(yQMjO,5<zO$)ZQ`O,5@TO$)iQ`O'#J}O!$fQlO1G2QO$)nQ`O1G2QO9ZQ`O'#KQO9ZQ`O'#ErO%[QlO'#ErO9ZQ`O'#IyO$)sQ!0LrO,5@yOOQ[1G2|1G2|OOQ[1G4^1G4^OOQ!0Lf1G/z1G/zOOQ!0Lf1G/x1G/xO$+uQ!0MxO1G0SOOQ[1G2x1G2xO!&iQMhO1G2xO%[QlO1G2xO#-|Q`O1G2xO$-yQMhO'#EiOOQ!0Lb,5@R,5@RO$.WQ!0LrO,5@ROOQ[1G.u1G.uO!BYQ!0LrO1G.uO!BeQpO1G.uO!BmQMhO1G.uO$.iQ`O1G0sO$.nQ`O'#CiO$.yQ`O'#KcO$/RQ`O,5=zO$/WQ`O'#KcO$/]Q`O'#KcO$/kQ`O'#JPO$/yQ`O,5@|O$0RQ!fO1G1hOOQ!0Lf1G1j1G1jO9aQ`O1G3eO@iQ`O1G3eO$0YQ`O1G3eO$0_Q`O1G3eOOQ[1G3e1G3eO!DkQ`O1G3TO!&iQMhO1G3QO$0dQ`O1G3QOOQ[1G3R1G3RO!&iQMhO1G3RO$0iQ`O1G3RO$0qQpO'#HPOOQ[1G3T1G3TO!5|QpO'#I{O!DpQ!bO1G3WOOQ[1G3W1G3WOOQ[,5=q,5=qO$0yQMhO,5=sO9aQ`O,5=sO$#yQ`O,5=uO9UQ`O,5=uO!BeQpO,5=uO!BmQMhO,5=uO:YQMhO,5=uO$1XQ`O'#KaO$1dQ`O,5=vOOQ[1G.k1G.kO$1iQ!0LrO1G.kO@iQ`O1G.kO$1tQ`O1G.kO9kQ!0LrO1G.kO$3|Q!fO,5AOO$4ZQ`O,5AOO9ZQ`O,5AOO$4fQlO,5=}O$4mQ`O,5=}OOQ[1G3g1G3gO`QlO1G3gOOQ[1G3m1G3mOOQ[1G3o1G3oO>rQ`O1G3qO$4rQlO1G3sO$8vQlO'#HrOOQ[1G3v1G3vO$9TQ`O'#HxO>wQ`O'#HzOOQ[1G3|1G3|O$9]QlO1G3|O9kQ!0LrO1G4SOOQ[1G4U1G4UOOQ!0Lb'#G^'#G^O9kQ!0LrO1G4WO9kQ!0LrO1G4YO$=dQ`O,5@`O!(yQlO,5;^O9ZQ`O,5;^O>wQ`O,5:VO!(yQlO,5:VO!BeQpO,5:VO$=iQ?MtO,5:VOOQO,5;^,5;^O$=sQpO'#IcO$>ZQ`O,5@_OOQ!0Lf1G/p1G/pO$>cQpO'#IiO$>mQ`O,5@nOOQ!0Lb1G0w1G0wO##hQpO,5:VOOQO'#Ie'#IeO$>uQpO,5:oOOQ!0Ln,5:o,5:oO#'cQ`O1G0XOOQ!0Lf1G0X1G0XO%[QlO1G0XOOQ!0Lf1G0r1G0rO>wQ`O1G0rO!BeQpO1G0rO!BmQMhO1G0rOOQ!0Lb1G5z1G5zO!BYQ!0LrO1G0[OOQO1G0k1G0kO%[QlO1G0kO$>|Q!0LrO1G0kO$?XQ!0LrO1G0kO!BeQpO1G0[OCuQpO1G0[O$?gQ!0LrO1G0kOOQO1G0[1G0[O$?{Q!0MxO1G0kPOOO-E<Y-E<YPOOO1G.h1G.hOOOO1G/g1G/gO$@VQ!bO,5<hO$@_Q!fO1G4hOOQO1G4n1G4nO%[QlO,5>|O$@iQ`O1G5xO$@qQ`O1G6WO$@yQ!fO1G6XO9ZQ`O,5?SO$ATQ!0MxO1G6UO%[QlO1G6UO$AeQ!0LrO1G6UO$AvQ`O1G6TO$AvQ`O1G6TO9ZQ`O1G6TO$BOQ`O,5?VO9ZQ`O,5?VOOQO,5?V,5?VO$BdQ`O,5?VO$)iQ`O,5?VOOQO-E<i-E<iOOQS1G0_1G0_OOQS1G0a1G0aO#-tQ`O1G0aOOQ[7+(d7+(dO!&iQMhO7+(dO%[QlO7+(dO$BrQ`O7+(dO$B}QMhO7+(dO$C]Q!0MzO,5=WO$EhQ!0MzO,5=YO$GsQ!0MzO,5=WO$JUQ!0MzO,5=YO$LgQ!0MzO,59sO$NlQ!0MzO,5<jO%!wQ!0MzO,5<lO%%SQ!0MzO,5<zOOQ!0Lf7+&_7+&_O%'eQ!0MxO7+&_O%(XQlO'#IdO%(fQ`O,5@aO%(nQ!fO,5@aOOQ!0Lf1G/}1G/}O%(xQ`O7+&hOOQ!0Lf7+&h7+&hO%(}Q?MtO,5:dO%[QlO7+&yO%)XQ?MtO,5:`O%)fQ?MtO,5:hO%)pQ?MtO,5:jO%)zQMhO'#IgO%*UQ`O,5@fOOQ!0Lh1G0b1G0bOOQO1G1q1G1qOOQO1G1r1G1rO%*^Q!jO,5<YO!(yQlO,5<XOOQO-E<j-E<jOOQ!0Lf7+'X7+'XOOOW7+'d7+'dOOOW1G1{1G1{O%*iQ`O1G1{OOQ!0Lf1G1}1G1}OOOO,59m,59mO%*nQ!dO,59mOOOO-E<^-E<^OOQ!0Lh1G/V1G/VO%*uQ!0MxO7+'jOOQ!0Lh,5?[,5?[O%+iQMhO1G2eP%+pQ`O'#IpPOQ!0Lh-E<n-E<nO%,^QMjO,5?_OOQ!0Lh-E<q-E<qO%-PQMjO,5?aOOQ!0Lh-E<s-E<sO%-ZQ!dO1G2vO%-bQ!dO'#CrO%-xQMhO'#KQO$#^QlO'#JtOOQ!0Lh1G2^1G2^O%.PQ`O'#IoO%.eQ`O,5@tO%.eQ`O,5@tO%.mQ`O,5@tO%.xQ`O,5@tOOQO1G2`1G2`O%/WQMjO1G2_O!+rQMhO1G2_O%/hQ(CWO'#IqO%/uQ`O,5@uO!&iQMhO,5@uO%/}Q!dO,5@uOOQ!0Lh1G2c1G2cO%2_Q!fO'#CiO%2iQ`O,5=OOOQ!0Lb,5<|,5<|O%2qQpO,5<|OOQ!0Lb,5<},5<}OCfQ`O,5<|O%2|QpO,5<|OOQ!0Lb,5=Q,5=QO$)iQ`O,5=UOOQO,5?^,5?^OOQO-E<p-E<pOOQ!0Lp1G2g1G2gO##hQpO,5<|O$#^QlO,5=OO%3[Q`O,5<}O%3gQpO,5<}O!+rQMhO'#IsO%4aQMjO1G2rO!+rQMhO'#IuO%5SQMjO1G2tO%5^QMjO1G5oO%5hQMjO1G5oOOQO,5?c,5?cOOQO-E<u-E<uOOQO1G.{1G.{O!9lQpO,59uO%[QlO,59uOOQ!0Lh,5<i,5<iO%5uQ`O1G2YO!+rQMhO1G2aO!+rQMhO1G5oO!+rQMhO1G5oO%5zQ!0MxO7+'lOOQ!0Lf7+'l7+'lO!$fQlO7+'lO%6nQ`O,5;^OOQ!0Lb,5?e,5?eOOQ!0Lb-E<w-E<wO%6sQ!dO'#K[O#'cQ`O7+(dO4UQ!fO7+(dO$BuQ`O7+(dO%6}Q!0MvO'#CiO%7nQ!0LrO,5=RO%8PQ!0MvO,5=RO%8dQ`O,5=ROOQ!0Lb1G5m1G5mOOQ[7+$a7+$aO!BYQ!0LrO7+$aO!BeQpO7+$aO!$fQlO7+&_O%8lQ`O'#JOO%9TQ`O,5@}OOQO1G3f1G3fO9aQ`O,5@}O%9TQ`O,5@}O%9]Q`O,5@}OOQO,5?k,5?kOOQO-E<}-E<}OOQ!0Lf7+'S7+'SO%9bQ`O7+)PO9kQ!0LrO7+)PO9aQ`O7+)PO@iQ`O7+)POOQ[7+(o7+(oO%9gQ!0MvO7+(lO!&iQMhO7+(lO!DfQ`O7+(mOOQ[7+(m7+(mO!&iQMhO7+(mO%9qQ`O'#K`O%9|Q`O,5=kOOQO,5?g,5?gOOQO-E<y-E<yOOQ[7+(r7+(rO%;`QpO'#HYOOQ[1G3_1G3_O!&iQMhO1G3_O%[QlO1G3_O%;gQ`O1G3_O%;rQMhO1G3_O9kQ!0LrO1G3aO$#yQ`O1G3aO9UQ`O1G3aO!BeQpO1G3aO!BmQMhO1G3aO%<QQ`O'#I}O%<fQ`O,5@{O%<nQpO,5@{OOQ!0Lb1G3b1G3bOOQ[7+$V7+$VO@iQ`O7+$VO9kQ!0LrO7+$VO%<yQ`O7+$VO%[QlO1G6jO%[QlO1G6kO%=OQ!0LrO1G6jO%=YQlO1G3iO%=aQ`O1G3iO%=fQlO1G3iOOQ[7+)R7+)RO9kQ!0LrO7+)]O`QlO7+)_OOQ['#Kf'#KfOOQ['#JQ'#JQO%=mQlO,5>^OOQ[,5>^,5>^O%[QlO'#HsO%=zQ`O'#HuOOQ[,5>d,5>dO9ZQ`O,5>dOOQ[,5>f,5>fOOQ[7+)h7+)hOOQ[7+)n7+)nOOQ[7+)r7+)rOOQ[7+)t7+)tO%>PQpO1G5zO%>kQ?MtO1G0xO%>uQ`O1G0xOOQO1G/q1G/qO%?QQ?MtO1G/qO>wQ`O1G/qO!(yQlO'#DkOOQO,5>},5>}OOQO-E<a-E<aOOQO,5?T,5?TOOQO-E<g-E<gO!BeQpO1G/qOOQO-E<c-E<cOOQ!0Ln1G0Z1G0ZOOQ!0Lf7+%s7+%sO#'cQ`O7+%sOOQ!0Lf7+&^7+&^O>wQ`O7+&^O!BeQpO7+&^OOQO7+%v7+%vO$?{Q!0MxO7+&VOOQO7+&V7+&VO%[QlO7+&VO%?[Q!0LrO7+&VO!BYQ!0LrO7+%vO!BeQpO7+%vO%?gQ!0LrO7+&VO%?uQ!0MxO7++pO%[QlO7++pO%@VQ`O7++oO%@VQ`O7++oOOQO1G4q1G4qO9ZQ`O1G4qO%@_Q`O1G4qOOQS7+%{7+%{O#'cQ`O<<LOO4UQ!fO<<LOO%@mQ`O<<LOOOQ[<<LO<<LOO!&iQMhO<<LOO%[QlO<<LOO%@uQ`O<<LOO%AQQ!0MzO,5?_O%C]Q!0MzO,5?aO%EhQ!0MzO1G2_O%GyQ!0MzO1G2rO%JUQ!0MzO1G2tO%LaQ!fO,5?OO%[QlO,5?OOOQO-E<b-E<bO%LkQ`O1G5{OOQ!0Lf<<JS<<JSO%LsQ?MtO1G0sO%NzQ?MtO1G0}O& RQ?MtO1G0}O&#SQ?MtO1G0}O&#ZQ?MtO1G0}O&%[Q?MtO1G0}O&']Q?MtO1G0}O&'dQ?MtO1G0}O&'kQ?MtO1G0}O&)lQ?MtO1G0}O&)sQ?MtO1G0}O&)zQ!0MxO<<JeO&+rQ?MtO1G0}O&,oQ?MvO1G0}O&-rQ?MvO'#JjO&/xQ?MtO1G1bO&0VQ?MtO1G0SO&0aQMjO,5?ROOQO-E<e-E<eO!(yQlO'#FpOOQO'#KX'#KXOOQO1G1t1G1tO&0kQ`O1G1sO&0pQ?MtO,5?YOOOW7+'g7+'gOOOO1G/X1G/XO&0zQ!dO1G4vOOQ!0Lh7+(P7+(PP!&iQMhO,5?[O!+rQMhO7+(bO&1RQ`O,5?ZO9ZQ`O,5?ZOOQO-E<m-E<mO&1aQ`O1G6`O&1aQ`O1G6`O&1iQ`O1G6`O&1tQMjO7+'yO&2UQ!dO,5?]O&2`Q`O,5?]O!&iQMhO,5?]OOQO-E<o-E<oO&2eQ!dO1G6aO&2oQ`O1G6aO&2wQ`O1G2jO!&iQMhO1G2jOOQ!0Lb1G2h1G2hOOQ!0Lb1G2i1G2iO%2qQpO1G2hO!BeQpO1G2hOCfQ`O1G2hOOQ!0Lb1G2p1G2pO&2|QpO1G2hO&3[Q`O1G2jO$)iQ`O1G2iOCfQ`O1G2iO$#^QlO1G2jO&3dQ`O1G2iO&4WQMjO,5?_OOQ!0Lh-E<r-E<rO&4yQMjO,5?aOOQ!0Lh-E<t-E<tO!+rQMhO7++ZOOQ!0Lh1G/a1G/aO&5TQ`O1G/aOOQ!0Lh7+'t7+'tO&5YQMjO7+'{O&5jQMjO7++ZO&5tQMjO7++ZO&6RQ!0MxO<<KWOOQ!0Lf<<KW<<KWO&6uQ`O1G0xO!&iQMhO'#IxO&6zQ`O,5@vO&8|Q!fO<<LOO!&iQMhO1G2mO&9TQ!0LrO1G2mOOQ[<<G{<<G{O!BYQ!0LrO<<G{O&9fQ!0MxO<<IyOOQ!0Lf<<Iy<<IyOOQO,5?j,5?jO&:YQ`O,5?jO&:_Q`O,5?jOOQO-E<|-E<|O&:mQ`O1G6iO&:mQ`O1G6iO9aQ`O1G6iO@iQ`O<<LkOOQ[<<Lk<<LkO&:uQ`O<<LkO9kQ!0LrO<<LkOOQ[<<LW<<LWO%9gQ!0MvO<<LWOOQ[<<LX<<LXO!DfQ`O<<LXO&:zQpO'#IzO&;VQ`O,5@zO!(yQlO,5@zOOQ[1G3V1G3VOOQO'#I|'#I|O9kQ!0LrO'#I|O&;_QpO,5=tOOQ[,5=t,5=tO&;fQpO'#EeO&;mQpO'#GdO&;rQ`O7+(yO&;wQ`O7+(yOOQ[7+(y7+(yO!&iQMhO7+(yO%[QlO7+(yO&<PQ`O7+(yOOQ[7+({7+({O9kQ!0LrO7+({O$#yQ`O7+({O9UQ`O7+({O!BeQpO7+({O&<[Q`O,5?iOOQO-E<{-E<{OOQO'#H]'#H]O&<gQ`O1G6gO9kQ!0LrO<<GqOOQ[<<Gq<<GqO@iQ`O<<GqO&<oQ`O7+,UO&<tQ`O7+,VO%[QlO7+,UO%[QlO7+,VOOQ[7+)T7+)TO&<yQ`O7+)TO&=OQlO7+)TO&=VQ`O7+)TOOQ[<<Lw<<LwOOQ[<<Ly<<LyOOQ[-E=O-E=OOOQ[1G3x1G3xO&=[Q`O,5>_OOQ[,5>a,5>aO&=aQ`O1G4OO9ZQ`O7+&dO!(yQlO7+&dOOQO7+%]7+%]O&=fQ?MtO1G6XO>wQ`O7+%]OOQ!0Lf<<I_<<I_OOQ!0Lf<<Ix<<IxO>wQ`O<<IxOOQO<<Iq<<IqO$?{Q!0MxO<<IqO%[QlO<<IqOOQO<<Ib<<IbO!BYQ!0LrO<<IbO&=pQ!0LrO<<IqO&={Q!0MxO<= [O&>]Q`O<= ZOOQO7+*]7+*]O9ZQ`O7+*]OOQ[ANAjANAjO&>eQ!fOANAjO!&iQMhOANAjO#'cQ`OANAjO4UQ!fOANAjO&>lQ`OANAjO%[QlOANAjO&>tQ!0MzO7+'yO&AVQ!0MzO,5?_O&CbQ!0MzO,5?aO&EmQ!0MzO7+'{O&HOQ!fO1G4jO&HYQ?MtO7+&_O&J^Q?MvO,5=WO&LeQ?MvO,5=YO&LuQ?MvO,5=WO&MVQ?MvO,5=YO&MgQ?MvO,59sO' mQ?MvO,5<jO'#pQ?MvO,5<lO'&UQ?MvO,5<zO''zQ?MtO7+'jO'(XQ?MtO7+'lO'(fQ`O,5<[OOQO7+'_7+'_OOQ!0Lh7+*b7+*bO'(kQMjO<<K|OOQO1G4u1G4uO'(rQ`O1G4uO'(}Q`O1G4uO')]Q`O7++zO')]Q`O7++zO!&iQMhO1G4wO')eQ!dO1G4wO')oQ`O7++{O')wQ`O7+(UO'*SQ!dO7+(UOOQ!0Lb7+(S7+(SOOQ!0Lb7+(T7+(TO!BeQpO7+(SOCfQ`O7+(SO'*^Q`O7+(UO!&iQMhO7+(UO$)iQ`O7+(TO'*cQ`O7+(UOCfQ`O7+(TO'*kQMjO<<NuOOQ!0Lh7+${7+${O!+rQMhO<<NuO'*uQ!dO,5?dOOQO-E<v-E<vO'+PQ!0MvO7+(XO!&iQMhO7+(XOOQ[AN=gAN=gO9aQ`O1G5UOOQO1G5U1G5UO'+aQ`O1G5UO'+fQ`O7+,TO'+fQ`O7+,TO9kQ!0LrOANBVO@iQ`OANBVOOQ[ANBVANBVOOQ[ANArANArOOQ[ANAsANAsO'+nQ`O,5?fOOQO-E<x-E<xO'+yQ?MtO1G6fOOQO,5?h,5?hOOQO-E<z-E<zOOQ[1G3`1G3`O',TQ`O,5=OOOQ[<<Le<<LeO!&iQMhO<<LeO&;rQ`O<<LeO',YQ`O<<LeO%[QlO<<LeOOQ[<<Lg<<LgO9kQ!0LrO<<LgO$#yQ`O<<LgO9UQ`O<<LgO',bQpO1G5TO',mQ`O7+,ROOQ[AN=]AN=]O9kQ!0LrOAN=]OOQ[<= p<= pOOQ[<= q<= qO',uQ`O<= pO',zQ`O<= qOOQ[<<Lo<<LoO'-PQ`O<<LoO'-UQlO<<LoOOQ[1G3y1G3yO>wQ`O7+)jO'-]Q`O<<JOO'-hQ?MtO<<JOOOQO<<Hw<<HwOOQ!0LfAN?dAN?dOOQOAN?]AN?]O$?{Q!0MxOAN?]OOQOAN>|AN>|O%[QlOAN?]OOQO<<Mw<<MwOOQ[G27UG27UO!&iQMhOG27UO#'cQ`OG27UO'-rQ!fOG27UO4UQ!fOG27UO'-yQ`OG27UO'.RQ?MtO<<JeO'.`Q?MvO1G2_O'0UQ?MvO,5?_O'2XQ?MvO,5?aO'4[Q?MvO1G2rO'6_Q?MvO1G2tO'8bQ?MtO<<KWO'8oQ?MtO<<IyOOQO1G1v1G1vO!+rQMhOANAhOOQO7+*a7+*aO'8|Q`O7+*aO'9XQ`O<= fO'9aQ!dO7+*cOOQ!0Lb<<Kp<<KpO$)iQ`O<<KpOCfQ`O<<KpO'9kQ`O<<KpO!&iQMhO<<KpOOQ!0Lb<<Kn<<KnO!BeQpO<<KnO'9vQ!dO<<KpOOQ!0Lb<<Ko<<KoO':QQ`O<<KpO!&iQMhO<<KpO$)iQ`O<<KoO':VQMjOANDaO':aQ!0MvO<<KsOOQO7+*p7+*pO9aQ`O7+*pO':qQ`O<= oOOQ[G27qG27qO9kQ!0LrOG27qO!(yQlO1G5QO':yQ`O7+,QO';RQ`O1G2jO&;rQ`OANBPOOQ[ANBPANBPO!&iQMhOANBPO';WQ`OANBPOOQ[ANBRANBRO9kQ!0LrOANBRO$#yQ`OANBROOQO'#H^'#H^OOQO7+*o7+*oOOQ[G22wG22wOOQ[ANE[ANE[OOQ[ANE]ANE]OOQ[ANBZANBZO';`Q`OANBZOOQ[<<MU<<MUO!(yQlOAN?jOOQOG24wG24wO$?{Q!0MxOG24wO#'cQ`OLD,pOOQ[LD,pLD,pO!&iQMhOLD,pO';eQ!fOLD,pO';lQ?MvO7+'yO'=bQ?MvO,5?_O'?eQ?MvO,5?aO'AhQ?MvO7+'{O'C^QMjOG27SOOQO<<M{<<M{OOQ!0LbANA[ANA[O$)iQ`OANA[OCfQ`OANA[O'CnQ!dOANA[OOQ!0LbANAYANAYO'CuQ`OANA[O!&iQMhOANA[O'DQQ!dOANA[OOQ!0LbANAZANAZOOQO<<N[<<N[OOQ[LD-]LD-]O'D[Q?MtO7+*lOOQO'#Ge'#GeOOQ[G27kG27kO&;rQ`OG27kO!&iQMhOG27kOOQ[G27mG27mO9kQ!0LrOG27mOOQ[G27uG27uO'DfQ?MtOG25UOOQOLD*cLD*cOOQ[!$(![!$(![O#'cQ`O!$(![O!&iQMhO!$(![O'DpQ!0MzOG27SOOQ!0LbG26vG26vO$)iQ`OG26vO'GRQ`OG26vOCfQ`OG26vO'G^Q!dOG26vO!&iQMhOG26vOOQ[LD-VLD-VO&;rQ`OLD-VOOQ[LD-XLD-XOOQ[!)9Ev!)9EvO#'cQ`O!)9EvOOQ!0LbLD,bLD,bO$)iQ`OLD,bOCfQ`OLD,bO'GeQ`OLD,bO'GpQ!dOLD,bOOQ[!$(!q!$(!qOOQ[!.K;b!.K;bO'GwQ?MvOG27SOOQ!0Lb!$( |!$( |O$)iQ`O!$( |OCfQ`O!$( |O'ImQ`O!$( |OOQ!0Lb!)9Eh!)9EhO$)iQ`O!)9EhOCfQ`O!)9EhOOQ!0Lb!.K;S!.K;SO$)iQ`O!.K;SOOQ!0Lb!4/0n!4/0nO!(yQlO'#DxO1PQ`O'#EVO'IxQ!fO'#JpO'JPQ!L^O'#DtO'JWQlO'#D|O'J_Q!fO'#CiO'LuQ!fO'#CiO!(yQlO'#EOO'MVQlO,5;XO!(yQlO,5;cO!(yQlO,5;cO!(yQlO,5;cO!(yQlO,5;cO!(yQlO,5;cO!(yQlO,5;cO!(yQlO,5;cO!(yQlO,5;cO!(yQlO,5;cO!(yQlO,5;cO!(yQlO'#InO( YQ`O,5<hO!(yQlO,5;cO( bQMhO,5;cO(!{QMhO,5;cO!(yQlO,5;vO!&iQMhO'#GlO( bQMhO'#GlO!&iQMhO'#GnO( bQMhO'#GnO1SQ`O'#DXO1SQ`O'#DXO!&iQMhO'#GOO( bQMhO'#GOO!&iQMhO'#GQO( bQMhO'#GQO!&iQMhO'#G`O( bQMhO'#G`O!(yQlO,5:hO(#SQpO'#D]O(#^QpO'#JtO!(yQlO,5@mO'MVQlO1G0sO(#hQ?MtO'#CiO!(yQlO1G2OO!&iQMhO'#IsO( bQMhO'#IsO!&iQMhO'#IuO( bQMhO'#IuO(#rQ!dO'#CrO!&iQMhO,5<sO( bQMhO,5<sO'MVQlO1G2QO!(yQlO7+&yO!&iQMhO1G2_O( bQMhO1G2_O!&iQMhO'#IsO( bQMhO'#IsO!&iQMhO'#IuO( bQMhO'#IuO!&iQMhO1G2aO( bQMhO1G2aO'MVQlO7+'lO'MVQlO7+&_O!&iQMhOANAhO( bQMhOANAhO($VQ`O'#EmO($[Q`O'#EmO($dQ`O'#F[O($iQ`O'#EwO($nQ`O'#KRO($yQ`O'#KPO(%UQ`O,5;XO(%ZQMjO,5<dO(%bQ`O'#GXO(%gQ`O'#GXO(%lQ`O,5<fO(%tQ`O,5;XO(%|Q?MtO1G1_O(&TQ`O,5<sO(&YQ`O,5<sO(&_Q`O,5<uO(&dQ`O,5<uO(&iQ`O1G2QO(&nQ`O1G0sO(&sQMjO<<K|O(&zQMjO<<K|O7eQMhO'#F{O9UQ`O'#FzOAdQ`O'#ElO!(yQlO,5;sO!3^Q`O'#GXO!3^Q`O'#GXO!3^Q`O'#GZO!3^Q`O'#GZO!+rQMhO7+(bO!+rQMhO7+(bO%-ZQ!dO1G2vO%-ZQ!dO1G2vO!&iQMhO,5=[O!&iQMhO,5=[",stateData:"((P~O'zOS'{OSTOS'|RQ~OPYOQYOSfOY!VOaqOdzOeyOj!POnkOpYOqkOrkOxkOzYO|YO!QWO!UkO!VkO!]XO!guO!jZO!mYO!nYO!oYO!qvO!swO!vxO!z]O$V|O$miO%g}O%i!QO%k!OO%l!OO%m!OO%p!RO%r!SO%u!TO%v!TO%x!UO&U!WO&[!XO&^!YO&`!ZO&b![O&e!]O&k!^O&q!_O&s!`O&u!aO&w!bO&y!cO(RSO(TTO(WUO(_VO(m[O~OWtO~P`OPYOQYOSfOd!jOe!iOnkOpYOqkOrkOxkOzYO|YO!QWO!UkO!VkO!]!eO!guO!jZO!mYO!nYO!oYO!qvO!s!gO!v!hO$V!kO$miO(R!dO(TTO(WUO(_VO(m[O~Oa!wOq!nO!Q!oO!`!yO!a!vO!b!vO!z;wO#R!pO#S!pO#T!xO#U!pO#V!pO#Y!zO#Z!zO(S!lO(TTO(WUO(c!mO(m!sO~O'|!{O~OP]XR]X[]Xa]Xp]X!O]X!Q]X!Z]X!j]X!n]X#P]X#Q]X#^]X#ifX#l]X#m]X#n]X#o]X#p]X#q]X#r]X#s]X#t]X#u]X#w]X#y]X#z]X$P]X'x]X(_]X(p]X(w]X(x]X~O!e%QX~P(qO_!}O(T#PO(U!}O(V#PO~O_#QO(V#PO(W#PO(X#QO~Ov#SO!S#TO(`#TO(a#VO~OPYOQYOSfOd!jOe!iOnkOpYOqkOrkOxkOzYO|YO!QWO!UkO!VkO!]!eO!guO!jZO!mYO!nYO!oYO!qvO!s!gO!v!hO$V!kO$miO(R;{O(TTO(WUO(_VO(m[O~O!Y#ZO!Z#WO!W(fP!W(tP~P+}O![#cO~P`OPYOQYOSfOd!jOe!iOpYOqkOrkOxkOzYO|YO!QWO!UkO!VkO!]!eO!guO!jZO!mYO!nYO!oYO!qvO!s!gO!v!hO$V!kO$miO(TTO(WUO(_VO(m[O~On#mO!Y#iO!z]O#g#lO#h#iO(R;|O!i(qP~P.iO!j#oO(R#nO~O!v#sO!z]O%g#tO~O#i#uO~O!e#vO#i#uO~OP$[OR#zO[$cOp$aO!O#yO!Q#{O!Z$_O!j#xO!n$[O#P$RO#l$OO#m$PO#n$PO#o$PO#p$QO#q$RO#r$RO#s$bO#t$RO#u$SO#w$UO#y$WO#z$XO(_VO(p$YO(w#|O(x#}O~Oa(dX'x(dX'u(dX!i(dX!W(dX!](dX%h(dX!e(dX~P1qO#Q$dO#^$eO$P$eOP(eXR(eX[(eXp(eX!O(eX!Q(eX!Z(eX!j(eX!n(eX#P(eX#l(eX#m(eX#n(eX#o(eX#p(eX#q(eX#r(eX#s(eX#t(eX#u(eX#w(eX#y(eX#z(eX(_(eX(p(eX(w(eX(x(eX!](eX%h(eX~Oa(eX'x(eX'u(eX!W(eX!i(eXt(eX!e(eX~P4UO#^$eO~O$[$hO$^$gO$e$mO~OSfO!]$nO$h$oO$j$qO~Oh%VOj%cOn%WOp%XOq$tOr$tOx%YOz%ZO|%[O!Q${O!]$|O!g%aO!j$xO#h%bO$V%_O$s%]O$u%^O$x%`O(R$sO(TTO(WUO(_$uO(w$}O(x%POg([P~O!j%dO~O!Q%gO!]%hO(R%fO~O!e%lO~Oa%mO'x%mO~O!O%qO~P%[O(S!lO~P%[O%m%uO~P%[Oh%VO!j%dO(R%fO(S!lO~Oe%|O!j%dO(R%fO~O#t$RO~O!O&RO!]&OO!j&QO%i&UO(R%fO(S!lO(TTO(WUO`)UP~O!v#sO~O%r&WO!Q)QX!])QX(R)QX~O(R&XO~Oj!PO!s&^O%i!QO%k!OO%l!OO%m!OO%p!RO%r!SO%u!TO%v!TO~Od&cOe&bO!v&`O%g&aO%z&_O~P<POd&fOeyOj!PO!]&eO!s&^O!vxO!z]O%g}O%k!OO%l!OO%m!OO%p!RO%r!SO%u!TO%v!TO%x!UO~Ob&iO#^&lO%i&gO(S!lO~P=UO!j&mO!s&qO~O!j#oO~O!]XO~Oa%mO'v&yO'x%mO~Oa%mO'v&|O'x%mO~Oa%mO'v'OO'x%mO~O'u]X!W]Xt]X!i]X&Y]X!]]X%h]X!e]X~P(qO!`']O!a'UO!b'UO(S!lO(TTO(WUO~Oq'SO!Q'RO!Y'VO(c'QO![(gP![(vP~P@]Ol'`O!]'^O(R%fO~Oe'eO!j%dO(R%fO~O!O&RO!j&QO~Oq!nO!Q!oO!z;wO#R!pO#S!pO#U!pO#V!pO(S!lO(TTO(WUO(c!mO(m!sO~O!`'kO!a'jO!b'jO#T!pO#Y'lO#Z'lO~PAwOa%mOh%VO!e#vO!j%dO'x%mO(p'nO~O!n'rO#^'pO~PCVOq!nO!Q!oO(TTO(WUO(c!mO(m!sO~O!]XOq(kX!Q(kX!`(kX!a(kX!b(kX!z(kX#R(kX#S(kX#T(kX#U(kX#V(kX#Y(kX#Z(kX(S(kX(T(kX(W(kX(c(kX(m(kX~O!a'jO!b'jO(S!lO~PCuO'}'vO(O'vO(P'xO~O_!}O(T'zO(U!}O(V'zO~O_#QO(V'zO(W'zO(X#QO~Ot'|O~P%[Ov#SO!S#TO(`#TO(a(PO~O!Y(RO!W'UX!W'[X!Z'UX!Z'[X~P+}O!Z(TO!W(fX~OP$[OR#zO[$cOp$aO!O#yO!Q#{O!Z(TO!j#xO!n$[O#P$RO#l$OO#m$PO#n$PO#o$PO#p$QO#q$RO#r$RO#s$bO#t$RO#u$SO#w$UO#y$WO#z$XO(_VO(p$YO(w#|O(x#}O~O!W(fX~PGpO!W(YO~O!W(sX!Z(sX!e(sX!i(sX(p(sX~O#^(sX#i#bX![(sX~PIsO#^(ZO!W(uX!Z(uX~O!Z([O!W(tX~O!W(_O~O#^$eO~PIsO![(`O~P`OR#zO!O#yO!Q#{O!j#xO(_VOP!la[!lap!la!Z!la!n!la#P!la#l!la#m!la#n!la#o!la#p!la#q!la#r!la#s!la#t!la#u!la#w!la#y!la#z!la(p!la(w!la(x!la~Oa!la'x!la'u!la!W!la!i!lat!la!]!la%h!la!e!la~PKZO!i(aO~O!e#vO#^(bO(p'nO!Z(rXa(rX'x(rX~O!i(rX~PMvO!Q%gO!]%hO!z]O#g(gO#h(fO(R%fO~O!Z(hO!i(qX~O!i(jO~O!Q%gO!]%hO#h(fO(R%fO~OP(eXR(eX[(eXp(eX!O(eX!Q(eX!Z(eX!j(eX!n(eX#P(eX#l(eX#m(eX#n(eX#o(eX#p(eX#q(eX#r(eX#s(eX#t(eX#u(eX#w(eX#y(eX#z(eX(_(eX(p(eX(w(eX(x(eX~O!e#vO!i(eX~P! dOR(lO!O(kO!j#xO#Q$dO!z!ya!Q!ya~O!v!ya%g!ya!]!ya#g!ya#h!ya(R!ya~P!#eO!v(pO~OPYOQYOSfOd!jOe!iOnkOpYOqkOrkOxkOzYO|YO!QWO!UkO!VkO!]XO!guO!jZO!mYO!nYO!oYO!qvO!s!gO!v!hO$V!kO$miO(R!dO(TTO(WUO(_VO(m[O~Oh%VOn%WOp%XOq$tOr$tOx%YOz%ZO|<eO!Q${O!]$|O!g=vO!j$xO#h<kO$V%_O$s<gO$u<iO$x%`O(R(tO(TTO(WUO(_$uO(w$}O(x%PO~O#i(vO~O!Y(xO!i(iP~P%[O(c(zO(m[O~O!Q(|O!j#xO(c(zO(m[O~OP;vOQ;vOSfOd=rOe!iOnkOp;vOqkOrkOxkOz;vO|;vO!QWO!UkO!VkO!]!eO!g;yO!jZO!m;vO!n;vO!o;vO!q;zO!s;}O!v!hO$V!kO$m=pO(R)ZO(TTO(WUO(_VO(m[O~O!Z$_Oa$pa'x$pa'u$pa!i$pa!W$pa!]$pa%h$pa!e$pa~Oj)bO~P!&iOh%VOn%WOp%XOq$tOr$tOx%YOz%ZO|%[O!Q${O!]$|O!g%aO!j$xO#h%bO$V%_O$s%]O$u%^O$x%`O(R(tO(TTO(WUO(_$uO(w$}O(x%PO~Og(nP~P!+rO!O)gO!e)fO!]$]X$Y$]X$[$]X$^$]X$e$]X~O!e)fO!](yX$Y(yX$[(yX$^(yX$e(yX~O!O)gO~P!-{O!O)gO!](yX$Y(yX$[(yX$^(yX$e(yX~O!])iO$Y)mO$[)hO$^)hO$e)nO~O!Y)qO~P!(yO$[$hO$^$gO$e)uO~Ol$yX!O$yX#Q$yX'w$yX(w$yX(x$yX~OgkXg$yXlkX!ZkX#^kX~P!/qOv)wO(`)xO(a)zO~Ol*TO!O)|O'w)}O(w$}O(x%PO~Og){O~P!0uOg*UO~Oh%VOn%WOp%XOq$tOr$tOx%YOz%ZO|<eO!Q*WO!]*XO!g=vO!j$xO#h<kO$V%_O$s<gO$u<iO$x%`O(TTO(WUO(_$uO(w$}O(x%PO~O!Y*[O(R*VO!i(|P~P!1dO#i*^O~O!j*_O~Oh%VOn%WOp%XOq$tOr$tOx%YOz%ZO|<eO!Q${O!]$|O!g=vO!j$xO#h<kO$V%_O$s<gO$u<iO$x%`O(R*aO(TTO(WUO(_$uO(w$}O(x%PO~O!Y*dO!W(}P~P!3cOp*pOq!nO!Q*fO!`*nO!a*hO!b*hO!j*_O#Y*oO%_*jO(S!lO(TTO(WUO(c!mO~O![*mO~P!5WO#Q$dOl(^X!O(^X'w(^X(w(^X(x(^X!Z(^X#^(^X~Og(^X#}(^X~P!6YOl*uO#^*tOg(]X!Z(]X~O!Z*vOg([X~Oj%cO(R&XOg([P~Oq*yO~O!j+OO~O(R(tO~On+TO!Q%gO!Y#iO!]%hO!z]O#g#lO#h#iO(R%fO!i(qP~O!e#vO#i+UO~O!Q%gO!Y+WO!Z([O!]%hO(R%fO!W(tP~Oq'YO!Q+YO!Y+XO(TTO(WUO(c(zO~O![(vP~P!9]O!Z+ZOa)RX'x)RX~OP$[OR#zO[$cOp$aO!O#yO!Q#{O!j#xO!n$[O#P$RO#l$OO#m$PO#n$PO#o$PO#p$QO#q$RO#r$RO#s$bO#t$RO#u$SO#w$UO#y$WO#z$XO(_VO(p$YO(w#|O(x#}O~Oa!ha!Z!ha'x!ha'u!ha!W!ha!i!hat!ha!]!ha%h!ha!e!ha~P!:TOR#zO!O#yO!Q#{O!j#xO(_VOP!pa[!pap!pa!Z!pa!n!pa#P!pa#l!pa#m!pa#n!pa#o!pa#p!pa#q!pa#r!pa#s!pa#t!pa#u!pa#w!pa#y!pa#z!pa(p!pa(w!pa(x!pa~Oa!pa'x!pa'u!pa!W!pa!i!pat!pa!]!pa%h!pa!e!pa~P!<kOR#zO!O#yO!Q#{O!j#xO(_VOP!ra[!rap!ra!Z!ra!n!ra#P!ra#l!ra#m!ra#n!ra#o!ra#p!ra#q!ra#r!ra#s!ra#t!ra#u!ra#w!ra#y!ra#z!ra(p!ra(w!ra(x!ra~Oa!ra'x!ra'u!ra!W!ra!i!rat!ra!]!ra%h!ra!e!ra~P!?ROh%VOl+dO!]'^O%h+cO~O!e+fOa(ZX!](ZX'x(ZX!Z(ZX~Oa%mO!]XO'x%mO~Oh%VO!j%dO~Oh%VO!j%dO(R%fO~O!e#vO#i(vO~Ob+qO%i+rO(R+nO(TTO(WUO![)VP~O!Z+sO`)UX~O[+wO~O`+xO~O!]&OO(R%fO(S!lO`)UP~Oh%VO#^+}O~Oh%VOl,QO!]$|O~O!],SO~O!O,UO!]XO~O%m%uO~O!v,ZO~Oe,`O~Ob,aO(R#nO(TTO(WUO![)TP~Oe%|O~O%i!QO(R&XO~P=UO[,fO`,eO~OPYOQYOSfOdzOeyOnkOpYOqkOrkOxkOzYO|YO!QWO!UkO!VkO!guO!jZO!mYO!nYO!oYO!qvO!vxO!z]O$miO%g}O(TTO(WUO(_VO(m[O~O!]!eO!s!gO$V!kO(R!dO~P!FRO`,eOa%mO'x%mO~OPYOQYOSfOd!jOe!iOnkOpYOqkOrkOxkOzYO|YO!QWO!UkO!VkO!]!eO!guO!jZO!mYO!nYO!oYO!qvO!v!hO$V!kO$miO(R!dO(TTO(WUO(_VO(m[O~Oa,kOj!OO!swO%k!OO%l!OO%m!OO~P!HkO!j&mO~O&[,qO~O!],sO~O&m,uO&o,vOP&jaQ&jaS&jaY&jaa&jad&jae&jaj&jan&jap&jaq&jar&jax&jaz&ja|&ja!Q&ja!U&ja!V&ja!]&ja!g&ja!j&ja!m&ja!n&ja!o&ja!q&ja!s&ja!v&ja!z&ja$V&ja$m&ja%g&ja%i&ja%k&ja%l&ja%m&ja%p&ja%r&ja%u&ja%v&ja%x&ja&U&ja&[&ja&^&ja&`&ja&b&ja&e&ja&k&ja&q&ja&s&ja&u&ja&w&ja&y&ja'u&ja(R&ja(T&ja(W&ja(_&ja(m&ja![&ja&c&jab&ja&h&ja~O(R,{O~Oh!cX!Z!PX![!PX!e!PX!e!cX!j!cX#^!PX~O!Z!cX![!cX~P# qO!e-QO#^-POh(hX!Z#fX![#fX!e(hX!j(hX~O!Z(hX![(hX~P#!dOh%VO!e-SO!j%dO!Z!_X![!_X~Oq!nO!Q!oO(TTO(WUO(c!mO~OP;vOQ;vOSfOd=rOe!iOnkOp;vOqkOrkOxkOz;vO|;vO!QWO!UkO!VkO!]!eO!g;yO!jZO!m;vO!n;vO!o;vO!q;zO!s;}O!v!hO$V!kO$m=pO(TTO(WUO(_VO(m[O~O(R<rO~P##yO!Z-WO![(gX~O![-YO~O!e-QO#^-PO!Z#fX![#fX~O!Z-ZO![(vX~O![-]O~O!a-^O!b-^O(S!lO~P##hO![-aO~P'_Ol-dO!]'^O~O!W-iO~Oq!ya!`!ya!a!ya!b!ya#R!ya#S!ya#T!ya#U!ya#V!ya#Y!ya#Z!ya(S!ya(T!ya(W!ya(c!ya(m!ya~P!#eO!n-nO#^-lO~PCVO!a-pO!b-pO(S!lO~PCuOa%mO#^-lO'x%mO~Oa%mO!e#vO#^-lO'x%mO~Oa%mO!e#vO!n-nO#^-lO'x%mO(p'nO~O'}'vO(O'vO(P-uO~Ot-vO~O!W'Ua!Z'Ua~P!:TO!Y-zO!W'UX!Z'UX~P%[O!Z(TO!W(fa~O!W(fa~PGpO!Z([O!W(ta~O!Q%gO!Y.OO!]%hO(R%fO!W'[X!Z'[X~O#^.QO!Z(ra!i(raa(ra'x(ra~O!e#vO~P#,PO!Z(hO!i(qa~O!Q%gO!]%hO#h.UO(R%fO~On.ZO!Q%gO!Y.WO!]%hO!z]O#g.YO#h.WO(R%fO!Z'_X!i'_X~OR._O!j#xO~Oh%VOl.bO!]'^O%h.aO~Oa#ai!Z#ai'x#ai'u#ai!W#ai!i#ait#ai!]#ai%h#ai!e#ai~P!:TOl=|O!O)|O'w)}O(w$}O(x%PO~O#i#]aa#]a#^#]a'x#]a!Z#]a!i#]a!]#]a!W#]a~P#.{O#i(^XP(^XR(^X[(^Xa(^Xp(^X!Q(^X!j(^X!n(^X#P(^X#l(^X#m(^X#n(^X#o(^X#p(^X#q(^X#r(^X#s(^X#t(^X#u(^X#w(^X#y(^X#z(^X'x(^X(_(^X(p(^X!i(^X!W(^X'u(^Xt(^X!](^X%h(^X!e(^X~P!6YO!Z.oO!i(iX~P!:TO!i.rO~O!W.tO~OP$[OR#zO!O#yO!Q#{O!j#xO!n$[O(_VO[#kia#kip#ki!Z#ki#P#ki#m#ki#n#ki#o#ki#p#ki#q#ki#r#ki#s#ki#t#ki#u#ki#w#ki#y#ki#z#ki'x#ki(p#ki(w#ki(x#ki'u#ki!W#ki!i#kit#ki!]#ki%h#ki!e#ki~O#l#ki~P#2kO#l$OO~P#2kOP$[OR#zOp$aO!O#yO!Q#{O!j#xO!n$[O#l$OO#m$PO#n$PO#o$PO(_VO[#kia#ki!Z#ki#P#ki#q#ki#r#ki#s#ki#t#ki#u#ki#w#ki#y#ki#z#ki'x#ki(p#ki(w#ki(x#ki'u#ki!W#ki!i#kit#ki!]#ki%h#ki!e#ki~O#p#ki~P#5YO#p$QO~P#5YOP$[OR#zO[$cOp$aO!O#yO!Q#{O!j#xO!n$[O#P$RO#l$OO#m$PO#n$PO#o$PO#p$QO#q$RO#r$RO#s$bO#t$RO(_VOa#ki!Z#ki#w#ki#y#ki#z#ki'x#ki(p#ki(w#ki(x#ki'u#ki!W#ki!i#kit#ki!]#ki%h#ki!e#ki~O#u#ki~P#7wOP$[OR#zO[$cOp$aO!O#yO!Q#{O!j#xO!n$[O#P$RO#l$OO#m$PO#n$PO#o$PO#p$QO#q$RO#r$RO#s$bO#t$RO#u$SO(_VO(x#}Oa#ki!Z#ki#y#ki#z#ki'x#ki(p#ki(w#ki'u#ki!W#ki!i#kit#ki!]#ki%h#ki!e#ki~O#w$UO~P#:_O#w#ki~P#:_O#u$SO~P#7wOP$[OR#zO[$cOp$aO!O#yO!Q#{O!j#xO!n$[O#P$RO#l$OO#m$PO#n$PO#o$PO#p$QO#q$RO#r$RO#s$bO#t$RO#u$SO#w$UO(_VO(w#|O(x#}Oa#ki!Z#ki#z#ki'x#ki(p#ki'u#ki!W#ki!i#kit#ki!]#ki%h#ki!e#ki~O#y#ki~P#=TO#y$WO~P#=TOP]XR]X[]Xp]X!O]X!Q]X!j]X!n]X#P]X#Q]X#^]X#ifX#l]X#m]X#n]X#o]X#p]X#q]X#r]X#s]X#t]X#u]X#w]X#y]X#z]X$P]X(_]X(p]X(w]X(x]X!Z]X![]X~O#}]X~P#?rOP$[OR#zO[<_Op<]O!O#yO!Q#{O!j#xO!n$[O#P<SO#l<PO#m<QO#n<QO#o<QO#p<RO#q<SO#r<SO#s<^O#t<SO#u<TO#w<VO#y<XO#z<YO(_VO(p$YO(w#|O(x#}O~O#}.vO~P#BPO#Q$dO#^<`O$P<`O#}(eX![(eX~P! dOa'ba!Z'ba'x'ba'u'ba!i'ba!W'bat'ba!]'ba%h'ba!e'ba~P!:TO[#kia#kip#ki!Z#ki#P#ki#p#ki#q#ki#r#ki#s#ki#t#ki#u#ki#w#ki#y#ki#z#ki'x#ki(p#ki'u#ki!W#ki!i#kit#ki!]#ki%h#ki!e#ki~OP$[OR#zO!O#yO!Q#{O!j#xO!n$[O#l$OO#m$PO#n$PO#o$PO(_VO(w#ki(x#ki~P#EROl=|O!O)|O'w)}O(w$}O(x%POP#kiR#ki!Q#ki!j#ki!n#ki#l#ki#m#ki#n#ki#o#ki(_#ki~P#ERO!Z.zOg(nX~P!0uOg.|O~Oa$Oi!Z$Oi'x$Oi'u$Oi!W$Oi!i$Oit$Oi!]$Oi%h$Oi!e$Oi~P!:TO$[.}O$^.}O~O$[/OO$^/OO~O!e)fO#^/PO!]$bX$Y$bX$[$bX$^$bX$e$bX~O!Y/QO~O!])iO$Y/SO$[)hO$^)hO$e/TO~O!Z<ZO![(dX~P#BPO![/UO~O!e)fO$e(yX~O$e/WO~Ot/XO~P!&iOv)wO(`)xO(a/[O~O!Q/_O~O(w$}Ol%`a!O%`a'w%`a(x%`a!Z%`a#^%`a~Og%`a#}%`a~P#LTO(x%POl%ba!O%ba'w%ba(w%ba!Z%ba#^%ba~Og%ba#}%ba~P#LvO!ZfX!efX!ifX!i$yX(pfX~P!/qO!Y/hO!Z([O(R/gO!W(tP!W(}P~P!1dOp*pO!`*nO!a*hO!b*hO!j*_O#Y*oO%_*jO(S!lO(TTO(WUO~Oq<oO!Q/iO!Y+XO![*mO(c<nO![(vP~P#NaO!i/jO~P#.{O!Z/kO!e#vO(p'nO!i(|X~O!i/pO~O!Q%gO!Y*[O!]%hO(R%fO!i(|P~O#i/rO~O!W$yX!Z$yX!e%QX~P!/qO!Z/sO!W(}X~P#.{O!e/uO~O!W/wO~OnkO(R/xO~P.iOh%VOp/}O!e#vO!j%dO(p'nO~O!e+fO~Oa%mO!Z0RO'x%mO~O![0TO~P!5WO!a0UO!b0UO(S!lO~P##hOq!nO!Q0VO(TTO(WUO(c!mO~O#Y0XO~Og%`a!Z%`a#^%`a#}%`a~P!0uOg%ba!Z%ba#^%ba#}%ba~P!0uOj%cO(R&XOg'kX!Z'kX~O!Z*vOg([a~Og0bO~OR0cO!O0cO!Q0dO#Q$dOl{a'w{a(w{a(x{a!Z{a#^{a~Og{a#}{a~P$&dO!O)|O'w)}Ol$ra(w$ra(x$ra!Z$ra#^$ra~Og$ra#}$ra~P$'`O!O)|O'w)}Ol$ta(w$ta(x$ta!Z$ta#^$ta~Og$ta#}$ta~P$(RO#i0gO~Og%Sa!Z%Sa#^%Sa#}%Sa~P!0uOl0iO#^0hOg(]a!Z(]a~O!e#vO~O#i0lO~O!Z+ZOa)Ra'x)Ra~OR#zO!O#yO!Q#{O!j#xO(_VOP!pi[!pip!pi!Z!pi!n!pi#P!pi#l!pi#m!pi#n!pi#o!pi#p!pi#q!pi#r!pi#s!pi#t!pi#u!pi#w!pi#y!pi#z!pi(p!pi(w!pi(x!pi~Oa!pi'x!pi'u!pi!W!pi!i!pit!pi!]!pi%h!pi!e!pi~P$*OOh%VOp%XOq$tOr$tOx%YOz%ZO|<eO!Q${O!]$|O!g=vO!j$xO#h<kO$V%_O$s<gO$u<iO$x%`O(TTO(WUO(_$uO(w$}O(x%PO~On0vO%[0wO(R0tO~P$,fO!e+fOa(Za!](Za'x(Za!Z(Za~O#i0|O~O[]X!ZfX![fX~O!Z0}O![)VX~O![1PO~O[1QO~Ob1SO(R+nO(TTO(WUO~O!]&OO(R%fO`'sX!Z'sX~O!Z+sO`)Ua~O!i1VO~P!:TO[1YO~O`1ZO~O#^1^O~Ol1aO!]$|O~O(c(zO![)SP~Oh%VOl1jO!]1gO%h1iO~O[1tO!Z1rO![)TX~O![1uO~O`1wOa%mO'x%mO~O(R#nO(TTO(WUO~O#Q$dO#^$eO$P$eOP(eXR(eX[(eXp(eX!O(eX!Q(eX!Z(eX!j(eX!n(eX#P(eX#l(eX#m(eX#n(eX#o(eX#p(eX#q(eX#r(eX#s(eX#u(eX#w(eX#y(eX#z(eX(_(eX(p(eX(w(eX(x(eX~O#t1zO&Y1{Oa(eX~P$2PO#^$eO#t1zO&Y1{O~Oa1}O~P%[Oa2PO~O&c2SOP&aiQ&aiS&aiY&aia&aid&aie&aij&ain&aip&aiq&air&aix&aiz&ai|&ai!Q&ai!U&ai!V&ai!]&ai!g&ai!j&ai!m&ai!n&ai!o&ai!q&ai!s&ai!v&ai!z&ai$V&ai$m&ai%g&ai%i&ai%k&ai%l&ai%m&ai%p&ai%r&ai%u&ai%v&ai%x&ai&U&ai&[&ai&^&ai&`&ai&b&ai&e&ai&k&ai&q&ai&s&ai&u&ai&w&ai&y&ai'u&ai(R&ai(T&ai(W&ai(_&ai(m&ai![&aib&ai&h&ai~Ob2YO![2WO&h2XO~P`O!]XO!j2[O~O&o,vOP&jiQ&jiS&jiY&jia&jid&jie&jij&jin&jip&jiq&jir&jix&jiz&ji|&ji!Q&ji!U&ji!V&ji!]&ji!g&ji!j&ji!m&ji!n&ji!o&ji!q&ji!s&ji!v&ji!z&ji$V&ji$m&ji%g&ji%i&ji%k&ji%l&ji%m&ji%p&ji%r&ji%u&ji%v&ji%x&ji&U&ji&[&ji&^&ji&`&ji&b&ji&e&ji&k&ji&q&ji&s&ji&u&ji&w&ji&y&ji'u&ji(R&ji(T&ji(W&ji(_&ji(m&ji![&ji&c&jib&ji&h&ji~O!W2bO~O!Z!_a![!_a~P#BPOq!nO!Q!oO!Y2hO(c!mO!Z'VX!['VX~P@]O!Z-WO![(ga~O!Z']X![']X~P!9]O!Z-ZO![(va~O![2oO~P'_Oa%mO#^2xO'x%mO~Oa%mO!e#vO#^2xO'x%mO~Oa%mO!e#vO!n2|O#^2xO'x%mO(p'nO~Oa%mO'x%mO~P!:TO!Z$_Ot$pa~O!W'Ui!Z'Ui~P!:TO!Z(TO!W(fi~O!Z([O!W(ti~O!W(ui!Z(ui~P!:TO!Z(ri!i(ria(ri'x(ri~P!:TO#^3OO!Z(ri!i(ria(ri'x(ri~O!Z(hO!i(qi~O!Q%gO!]%hO!z]O#g3TO#h3SO(R%fO~O!Q%gO!]%hO#h3SO(R%fO~Ol3[O!]'^O%h3ZO~Oh%VOl3[O!]'^O%h3ZO~O#i%`aP%`aR%`a[%`aa%`ap%`a!Q%`a!j%`a!n%`a#P%`a#l%`a#m%`a#n%`a#o%`a#p%`a#q%`a#r%`a#s%`a#t%`a#u%`a#w%`a#y%`a#z%`a'x%`a(_%`a(p%`a!i%`a!W%`a'u%`at%`a!]%`a%h%`a!e%`a~P#LTO#i%baP%baR%ba[%baa%bap%ba!Q%ba!j%ba!n%ba#P%ba#l%ba#m%ba#n%ba#o%ba#p%ba#q%ba#r%ba#s%ba#t%ba#u%ba#w%ba#y%ba#z%ba'x%ba(_%ba(p%ba!i%ba!W%ba'u%bat%ba!]%ba%h%ba!e%ba~P#LvO#i%`aP%`aR%`a[%`aa%`ap%`a!Q%`a!Z%`a!j%`a!n%`a#P%`a#l%`a#m%`a#n%`a#o%`a#p%`a#q%`a#r%`a#s%`a#t%`a#u%`a#w%`a#y%`a#z%`a'x%`a(_%`a(p%`a!i%`a!W%`a'u%`a#^%`at%`a!]%`a%h%`a!e%`a~P#.{O#i%baP%baR%ba[%baa%bap%ba!Q%ba!Z%ba!j%ba!n%ba#P%ba#l%ba#m%ba#n%ba#o%ba#p%ba#q%ba#r%ba#s%ba#t%ba#u%ba#w%ba#y%ba#z%ba'x%ba(_%ba(p%ba!i%ba!W%ba'u%ba#^%bat%ba!]%ba%h%ba!e%ba~P#.{O#i{aP{a[{aa{ap{a!j{a!n{a#P{a#l{a#m{a#n{a#o{a#p{a#q{a#r{a#s{a#t{a#u{a#w{a#y{a#z{a'x{a(_{a(p{a!i{a!W{a'u{at{a!]{a%h{a!e{a~P$&dO#i$raP$raR$ra[$raa$rap$ra!Q$ra!j$ra!n$ra#P$ra#l$ra#m$ra#n$ra#o$ra#p$ra#q$ra#r$ra#s$ra#t$ra#u$ra#w$ra#y$ra#z$ra'x$ra(_$ra(p$ra!i$ra!W$ra'u$rat$ra!]$ra%h$ra!e$ra~P$'`O#i$taP$taR$ta[$taa$tap$ta!Q$ta!j$ta!n$ta#P$ta#l$ta#m$ta#n$ta#o$ta#p$ta#q$ta#r$ta#s$ta#t$ta#u$ta#w$ta#y$ta#z$ta'x$ta(_$ta(p$ta!i$ta!W$ta'u$tat$ta!]$ta%h$ta!e$ta~P$(RO#i%SaP%SaR%Sa[%Saa%Sap%Sa!Q%Sa!Z%Sa!j%Sa!n%Sa#P%Sa#l%Sa#m%Sa#n%Sa#o%Sa#p%Sa#q%Sa#r%Sa#s%Sa#t%Sa#u%Sa#w%Sa#y%Sa#z%Sa'x%Sa(_%Sa(p%Sa!i%Sa!W%Sa'u%Sa#^%Sat%Sa!]%Sa%h%Sa!e%Sa~P#.{Oa#aq!Z#aq'x#aq'u#aq!W#aq!i#aqt#aq!]#aq%h#aq!e#aq~P!:TO!Y3dO!Z'WX!i'WX~P%[O!Z.oO!i(ia~O!Z.oO!i(ia~P!:TO!W3gO~O#}!la![!la~PKZO#}!ha!Z!ha![!ha~P#BPO#}!pa![!pa~P!<kO#}!ra![!ra~P!?ROg'ZX!Z'ZX~P!+rO!Z.zOg(na~OSfO!]3{O$c3|O~O![4QO~Ot4RO~P#.{Oa$lq!Z$lq'x$lq'u$lq!W$lq!i$lqt$lq!]$lq%h$lq!e$lq~P!:TO!W4TO~P!&iO!Q4UO~O!O)|O'w)}O(x%POl'ga(w'ga!Z'ga#^'ga~Og'ga#}'ga~P%+uO!O)|O'w)}Ol'ia(w'ia(x'ia!Z'ia#^'ia~Og'ia#}'ia~P%,hO(p$YO~P#.{O!WfX!W$yX!ZfX!Z$yX!e%QX#^fX~P!/qO(R<xO~P!1dO!Q%gO!Y4XO!]%hO(R%fO!Z'cX!i'cX~O!Z/kO!i(|a~O!Z/kO!e#vO!i(|a~O!Z/kO!e#vO(p'nO!i(|a~Og${i!Z${i#^${i#}${i~P!0uO!Y4aO!W'eX!Z'eX~P!3cO!Z/sO!W(}a~O!Z/sO!W(}a~P#.{OP]XR]X[]Xp]X!O]X!Q]X!W]X!Z]X!j]X!n]X#P]X#Q]X#^]X#ifX#l]X#m]X#n]X#o]X#p]X#q]X#r]X#s]X#t]X#u]X#w]X#y]X#z]X$P]X(_]X(p]X(w]X(x]X~O!e%XX#t%XX~P%0XO!e#vO#t4fO~Oh%VO!e#vO!j%dO~Oh%VOp4kO!j%dO(p'nO~Op4pO!e#vO(p'nO~Oq!nO!Q4qO(TTO(WUO(c!mO~O(w$}Ol%`i!O%`i'w%`i(x%`i!Z%`i#^%`i~Og%`i#}%`i~P%3xO(x%POl%bi!O%bi'w%bi(w%bi!Z%bi#^%bi~Og%bi#}%bi~P%4kOg(]i!Z(]i~P!0uO#^4wOg(]i!Z(]i~P!0uO!i4zO~Oa$nq!Z$nq'x$nq'u$nq!W$nq!i$nqt$nq!]$nq%h$nq!e$nq~P!:TO!W5QO~O!Z5RO!])OX~P#.{Oa]Xa$yX!]]X!]$yX%]]X'x]X'x$yX!Z]X!Z$yX~P!/qO%]5UOa%Za!]%Za'x%Za!Z%Za~OlmX!OmX'wmX(wmX(xmX~P%7nOn5VO(R#nO~Ob5]O%i5^O(R+nO(TTO(WUO!Z'rX!['rX~O!Z0}O![)Va~O[5bO~O`5cO~Oa%mO'x%mO~P#.{O!Z5kO#^5mO![)SX~O![5nO~Op5tOq!nO!Q*fO!`!yO!a!vO!b!vO!z;wO#R!pO#S!pO#T!pO#U!pO#V!pO#Y5sO#Z!zO(S!lO(TTO(WUO(c!mO(m!sO~O![5rO~P%:ROl5yO!]1gO%h5xO~Oh%VOl5yO!]1gO%h5xO~Ob6QO(R#nO(TTO(WUO!Z'qX!['qX~O!Z1rO![)Ta~O(TTO(WUO(c6SO~O`6WO~O#t6ZO&Y6[O~PMvO!i6]O~P%[Oa6_O~Oa6_O~P%[Ob2YO![6dO&h2XO~P`O!e6fO~O!e6hOh(hi!Z(hi![(hi!e(hi!j(hip(hi(p(hi~O!Z#fi![#fi~P#BPO#^6iO!Z#fi![#fi~O!Z!_i![!_i~P#BPOa%mO#^6rO'x%mO~Oa%mO!e#vO#^6rO'x%mO~O!Z(rq!i(rqa(rq'x(rq~P!:TO!Z(hO!i(qq~O!Q%gO!]%hO#h6yO(R%fO~O!]'^O%h6|O~Ol7QO!]'^O%h6|O~O#i'gaP'gaR'ga['gaa'gap'ga!Q'ga!j'ga!n'ga#P'ga#l'ga#m'ga#n'ga#o'ga#p'ga#q'ga#r'ga#s'ga#t'ga#u'ga#w'ga#y'ga#z'ga'x'ga(_'ga(p'ga!i'ga!W'ga'u'gat'ga!]'ga%h'ga!e'ga~P%+uO#i'iaP'iaR'ia['iaa'iap'ia!Q'ia!j'ia!n'ia#P'ia#l'ia#m'ia#n'ia#o'ia#p'ia#q'ia#r'ia#s'ia#t'ia#u'ia#w'ia#y'ia#z'ia'x'ia(_'ia(p'ia!i'ia!W'ia'u'iat'ia!]'ia%h'ia!e'ia~P%,hO#i${iP${iR${i[${ia${ip${i!Q${i!Z${i!j${i!n${i#P${i#l${i#m${i#n${i#o${i#p${i#q${i#r${i#s${i#t${i#u${i#w${i#y${i#z${i'x${i(_${i(p${i!i${i!W${i'u${i#^${it${i!]${i%h${i!e${i~P#.{O#i%`iP%`iR%`i[%`ia%`ip%`i!Q%`i!j%`i!n%`i#P%`i#l%`i#m%`i#n%`i#o%`i#p%`i#q%`i#r%`i#s%`i#t%`i#u%`i#w%`i#y%`i#z%`i'x%`i(_%`i(p%`i!i%`i!W%`i'u%`it%`i!]%`i%h%`i!e%`i~P%3xO#i%biP%biR%bi[%bia%bip%bi!Q%bi!j%bi!n%bi#P%bi#l%bi#m%bi#n%bi#o%bi#p%bi#q%bi#r%bi#s%bi#t%bi#u%bi#w%bi#y%bi#z%bi'x%bi(_%bi(p%bi!i%bi!W%bi'u%bit%bi!]%bi%h%bi!e%bi~P%4kO!Z'Wa!i'Wa~P!:TO!Z.oO!i(ii~O#}#ai!Z#ai![#ai~P#BPOP$[OR#zO!O#yO!Q#{O!j#xO!n$[O(_VO[#kip#ki#P#ki#m#ki#n#ki#o#ki#p#ki#q#ki#r#ki#s#ki#t#ki#u#ki#w#ki#y#ki#z#ki#}#ki(p#ki(w#ki(x#ki!Z#ki![#ki~O#l#ki~P%MQO#l<PO~P%MQOP$[OR#zOp<]O!O#yO!Q#{O!j#xO!n$[O#l<PO#m<QO#n<QO#o<QO(_VO[#ki#P#ki#q#ki#r#ki#s#ki#t#ki#u#ki#w#ki#y#ki#z#ki#}#ki(p#ki(w#ki(x#ki!Z#ki![#ki~O#p#ki~P& YO#p<RO~P& YOP$[OR#zO[<_Op<]O!O#yO!Q#{O!j#xO!n$[O#P<SO#l<PO#m<QO#n<QO#o<QO#p<RO#q<SO#r<SO#s<^O#t<SO(_VO#w#ki#y#ki#z#ki#}#ki(p#ki(w#ki(x#ki!Z#ki![#ki~O#u#ki~P&#bOP$[OR#zO[<_Op<]O!O#yO!Q#{O!j#xO!n$[O#P<SO#l<PO#m<QO#n<QO#o<QO#p<RO#q<SO#r<SO#s<^O#t<SO#u<TO(_VO(x#}O#y#ki#z#ki#}#ki(p#ki(w#ki!Z#ki![#ki~O#w<VO~P&%cO#w#ki~P&%cO#u<TO~P&#bOP$[OR#zO[<_Op<]O!O#yO!Q#{O!j#xO!n$[O#P<SO#l<PO#m<QO#n<QO#o<QO#p<RO#q<SO#r<SO#s<^O#t<SO#u<TO#w<VO(_VO(w#|O(x#}O#z#ki#}#ki(p#ki!Z#ki![#ki~O#y#ki~P&'rO#y<XO~P&'rOa#{y!Z#{y'x#{y'u#{y!W#{y!i#{yt#{y!]#{y%h#{y!e#{y~P!:TO[#kip#ki#P#ki#p#ki#q#ki#r#ki#s#ki#t#ki#u#ki#w#ki#y#ki#z#ki#}#ki(p#ki!Z#ki![#ki~OP$[OR#zO!O#yO!Q#{O!j#xO!n$[O#l<PO#m<QO#n<QO#o<QO(_VO(w#ki(x#ki~P&*nOl=}O!O)|O'w)}O(w$}O(x%POP#kiR#ki!Q#ki!j#ki!n#ki#l#ki#m#ki#n#ki#o#ki(_#ki~P&*nO#Q$dOP(^XR(^X[(^Xl(^Xp(^X!O(^X!Q(^X!j(^X!n(^X#P(^X#l(^X#m(^X#n(^X#o(^X#p(^X#q(^X#r(^X#s(^X#t(^X#u(^X#w(^X#y(^X#z(^X#}(^X'w(^X(_(^X(p(^X(w(^X(x(^X!Z(^X![(^X~O#}$Oi!Z$Oi![$Oi~P#BPO#}!pi![!pi~P$*OOg'Za!Z'Za~P!0uO![7dO~O!Z'ba!['ba~P#BPO!W7eO~P#.{O!e#vO(p'nO!Z'ca!i'ca~O!Z/kO!i(|i~O!Z/kO!e#vO!i(|i~Og${q!Z${q#^${q#}${q~P!0uO!W'ea!Z'ea~P#.{O!e7lO~O!Z/sO!W(}i~P#.{O!Z/sO!W(}i~O!W7oO~Oh%VOp7tO!j%dO(p'nO~O!e#vO#t7vO~Op7yO!e#vO(p'nO~O!O)|O'w)}O(x%POl'ha(w'ha!Z'ha#^'ha~Og'ha#}'ha~P&3oO!O)|O'w)}Ol'ja(w'ja(x'ja!Z'ja#^'ja~Og'ja#}'ja~P&4bO!W7{O~Og$}q!Z$}q#^$}q#}$}q~P!0uOg(]q!Z(]q~P!0uO#^7|Og(]q!Z(]q~P!0uOa$ny!Z$ny'x$ny'u$ny!W$ny!i$nyt$ny!]$ny%h$ny!e$ny~P!:TO!e6hO~O!Z5RO!])Oa~O!]'^OP$SaR$Sa[$Sap$Sa!O$Sa!Q$Sa!Z$Sa!j$Sa!n$Sa#P$Sa#l$Sa#m$Sa#n$Sa#o$Sa#p$Sa#q$Sa#r$Sa#s$Sa#t$Sa#u$Sa#w$Sa#y$Sa#z$Sa(_$Sa(p$Sa(w$Sa(x$Sa~O%h6|O~P&7SO%]8QOa%Zi!]%Zi'x%Zi!Z%Zi~Oa#ay!Z#ay'x#ay'u#ay!W#ay!i#ayt#ay!]#ay%h#ay!e#ay~P!:TO[8SO~Ob8UO(R+nO(TTO(WUO~O!Z0}O![)Vi~O`8YO~O(c(zO!Z'nX!['nX~O!Z5kO![)Sa~O![8cO~P%:RO(m!sO~P$$oO#Y8dO~O!]1gO~O!]1gO%h8fO~Ol8iO!]1gO%h8fO~O[8nO!Z'qa!['qa~O!Z1rO![)Ti~O!i8rO~O!i8sO~O!i8vO~O!i8vO~P%[Oa8xO~O!e8yO~O!i8zO~O!Z(ui![(ui~P#BPOa%mO#^9SO'x%mO~O!Z(ry!i(rya(ry'x(ry~P!:TO!Z(hO!i(qy~O%h9VO~P&7SO!]'^O%h9VO~O#i${qP${qR${q[${qa${qp${q!Q${q!Z${q!j${q!n${q#P${q#l${q#m${q#n${q#o${q#p${q#q${q#r${q#s${q#t${q#u${q#w${q#y${q#z${q'x${q(_${q(p${q!i${q!W${q'u${q#^${qt${q!]${q%h${q!e${q~P#.{O#i'haP'haR'ha['haa'hap'ha!Q'ha!j'ha!n'ha#P'ha#l'ha#m'ha#n'ha#o'ha#p'ha#q'ha#r'ha#s'ha#t'ha#u'ha#w'ha#y'ha#z'ha'x'ha(_'ha(p'ha!i'ha!W'ha'u'hat'ha!]'ha%h'ha!e'ha~P&3oO#i'jaP'jaR'ja['jaa'jap'ja!Q'ja!j'ja!n'ja#P'ja#l'ja#m'ja#n'ja#o'ja#p'ja#q'ja#r'ja#s'ja#t'ja#u'ja#w'ja#y'ja#z'ja'x'ja(_'ja(p'ja!i'ja!W'ja'u'jat'ja!]'ja%h'ja!e'ja~P&4bO#i$}qP$}qR$}q[$}qa$}qp$}q!Q$}q!Z$}q!j$}q!n$}q#P$}q#l$}q#m$}q#n$}q#o$}q#p$}q#q$}q#r$}q#s$}q#t$}q#u$}q#w$}q#y$}q#z$}q'x$}q(_$}q(p$}q!i$}q!W$}q'u$}q#^$}qt$}q!]$}q%h$}q!e$}q~P#.{O!Z'Wi!i'Wi~P!:TO#}#aq!Z#aq![#aq~P#BPO(w$}OP%`aR%`a[%`ap%`a!Q%`a!j%`a!n%`a#P%`a#l%`a#m%`a#n%`a#o%`a#p%`a#q%`a#r%`a#s%`a#t%`a#u%`a#w%`a#y%`a#z%`a#}%`a(_%`a(p%`a!Z%`a![%`a~Ol%`a!O%`a'w%`a(x%`a~P&HgO(x%POP%baR%ba[%bap%ba!Q%ba!j%ba!n%ba#P%ba#l%ba#m%ba#n%ba#o%ba#p%ba#q%ba#r%ba#s%ba#t%ba#u%ba#w%ba#y%ba#z%ba#}%ba(_%ba(p%ba!Z%ba![%ba~Ol%ba!O%ba'w%ba(w%ba~P&JnOl=}O!O)|O'w)}O(x%PO~P&HgOl=}O!O)|O'w)}O(w$}O~P&JnOR0cO!O0cO!Q0dO#Q$dOP{a[{al{ap{a!j{a!n{a#P{a#l{a#m{a#n{a#o{a#p{a#q{a#r{a#s{a#t{a#u{a#w{a#y{a#z{a#}{a'w{a(_{a(p{a(w{a(x{a!Z{a![{a~O!O)|O'w)}OP$raR$ra[$ral$rap$ra!Q$ra!j$ra!n$ra#P$ra#l$ra#m$ra#n$ra#o$ra#p$ra#q$ra#r$ra#s$ra#t$ra#u$ra#w$ra#y$ra#z$ra#}$ra(_$ra(p$ra(w$ra(x$ra!Z$ra![$ra~O!O)|O'w)}OP$taR$ta[$tal$tap$ta!Q$ta!j$ta!n$ta#P$ta#l$ta#m$ta#n$ta#o$ta#p$ta#q$ta#r$ta#s$ta#t$ta#u$ta#w$ta#y$ta#z$ta#}$ta(_$ta(p$ta(w$ta(x$ta!Z$ta![$ta~Ol=}O!O)|O'w)}O(w$}O(x%PO~OP%SaR%Sa[%Sap%Sa!Q%Sa!j%Sa!n%Sa#P%Sa#l%Sa#m%Sa#n%Sa#o%Sa#p%Sa#q%Sa#r%Sa#s%Sa#t%Sa#u%Sa#w%Sa#y%Sa#z%Sa#}%Sa(_%Sa(p%Sa!Z%Sa![%Sa~P'%sO#}$lq!Z$lq![$lq~P#BPO#}$nq!Z$nq![$nq~P#BPO![9dO~O#}9eO~P!0uO!e#vO!Z'ci!i'ci~O!e#vO(p'nO!Z'ci!i'ci~O!Z/kO!i(|q~O!W'ei!Z'ei~P#.{O!Z/sO!W(}q~Op9lO!e#vO(p'nO~O[9nO!W9mO~P#.{O!W9mO~O!e#vO#t9tO~Og(]y!Z(]y~P!0uO!Z'la!]'la~P#.{Oa%Zq!]%Zq'x%Zq!Z%Zq~P#.{O[9yO~O!Z0}O![)Vq~O#^9}O!Z'na!['na~O!Z5kO![)Si~P#BPO!Q:PO~O!]1gO%h:SO~O(TTO(WUO(c:XO~O!Z1rO![)Tq~O!i:[O~O!i:]O~O!i:^O~O!i:^O~P%[O#^:aO!Z#fy![#fy~O!Z#fy![#fy~P#BPO%h:fO~P&7SO!]'^O%h:fO~O#}#{y!Z#{y![#{y~P#BPOP${iR${i[${ip${i!Q${i!j${i!n${i#P${i#l${i#m${i#n${i#o${i#p${i#q${i#r${i#s${i#t${i#u${i#w${i#y${i#z${i#}${i(_${i(p${i!Z${i![${i~P'%sO!O)|O'w)}O(x%POP'gaR'ga['gal'gap'ga!Q'ga!j'ga!n'ga#P'ga#l'ga#m'ga#n'ga#o'ga#p'ga#q'ga#r'ga#s'ga#t'ga#u'ga#w'ga#y'ga#z'ga#}'ga(_'ga(p'ga(w'ga!Z'ga!['ga~O!O)|O'w)}OP'iaR'ia['ial'iap'ia!Q'ia!j'ia!n'ia#P'ia#l'ia#m'ia#n'ia#o'ia#p'ia#q'ia#r'ia#s'ia#t'ia#u'ia#w'ia#y'ia#z'ia#}'ia(_'ia(p'ia(w'ia(x'ia!Z'ia!['ia~O(w$}OP%`iR%`i[%`il%`ip%`i!O%`i!Q%`i!j%`i!n%`i#P%`i#l%`i#m%`i#n%`i#o%`i#p%`i#q%`i#r%`i#s%`i#t%`i#u%`i#w%`i#y%`i#z%`i#}%`i'w%`i(_%`i(p%`i(x%`i!Z%`i![%`i~O(x%POP%biR%bi[%bil%bip%bi!O%bi!Q%bi!j%bi!n%bi#P%bi#l%bi#m%bi#n%bi#o%bi#p%bi#q%bi#r%bi#s%bi#t%bi#u%bi#w%bi#y%bi#z%bi#}%bi'w%bi(_%bi(p%bi(w%bi!Z%bi![%bi~O#}$ny!Z$ny![$ny~P#BPO#}#ay!Z#ay![#ay~P#BPO!e#vO!Z'cq!i'cq~O!Z/kO!i(|y~O!W'eq!Z'eq~P#.{Op:pO!e#vO(p'nO~O[:tO!W:sO~P#.{O!W:sO~Og(]!R!Z(]!R~P!0uOa%Zy!]%Zy'x%Zy!Z%Zy~P#.{O!Z0}O![)Vy~O!Z5kO![)Sq~O(R:zO~O!]1gO%h:}O~O!i;QO~O%h;VO~P&7SOP${qR${q[${qp${q!Q${q!j${q!n${q#P${q#l${q#m${q#n${q#o${q#p${q#q${q#r${q#s${q#t${q#u${q#w${q#y${q#z${q#}${q(_${q(p${q!Z${q![${q~P'%sO!O)|O'w)}O(x%POP'haR'ha['hal'hap'ha!Q'ha!j'ha!n'ha#P'ha#l'ha#m'ha#n'ha#o'ha#p'ha#q'ha#r'ha#s'ha#t'ha#u'ha#w'ha#y'ha#z'ha#}'ha(_'ha(p'ha(w'ha!Z'ha!['ha~O!O)|O'w)}OP'jaR'ja['jal'jap'ja!Q'ja!j'ja!n'ja#P'ja#l'ja#m'ja#n'ja#o'ja#p'ja#q'ja#r'ja#s'ja#t'ja#u'ja#w'ja#y'ja#z'ja#}'ja(_'ja(p'ja(w'ja(x'ja!Z'ja!['ja~OP$}qR$}q[$}qp$}q!Q$}q!j$}q!n$}q#P$}q#l$}q#m$}q#n$}q#o$}q#p$}q#q$}q#r$}q#s$}q#t$}q#u$}q#w$}q#y$}q#z$}q#}$}q(_$}q(p$}q!Z$}q![$}q~P'%sOg%d!Z!Z%d!Z#^%d!Z#}%d!Z~P!0uO!W;ZO~P#.{Op;[O!e#vO(p'nO~O[;^O!W;ZO~P#.{O!Z'nq!['nq~P#BPO!Z#f!Z![#f!Z~P#BPO#i%d!ZP%d!ZR%d!Z[%d!Za%d!Zp%d!Z!Q%d!Z!Z%d!Z!j%d!Z!n%d!Z#P%d!Z#l%d!Z#m%d!Z#n%d!Z#o%d!Z#p%d!Z#q%d!Z#r%d!Z#s%d!Z#t%d!Z#u%d!Z#w%d!Z#y%d!Z#z%d!Z'x%d!Z(_%d!Z(p%d!Z!i%d!Z!W%d!Z'u%d!Z#^%d!Zt%d!Z!]%d!Z%h%d!Z!e%d!Z~P#.{Op;fO!e#vO(p'nO~O!W;gO~P#.{Op;nO!e#vO(p'nO~O!W;oO~P#.{OP%d!ZR%d!Z[%d!Zp%d!Z!Q%d!Z!j%d!Z!n%d!Z#P%d!Z#l%d!Z#m%d!Z#n%d!Z#o%d!Z#p%d!Z#q%d!Z#r%d!Z#s%d!Z#t%d!Z#u%d!Z#w%d!Z#y%d!Z#z%d!Z#}%d!Z(_%d!Z(p%d!Z!Z%d!Z![%d!Z~P'%sOp;rO!e#vO(p'nO~Ot(dX~P1qO!O%qO~P!(yO(S!lO~P!(yO!WfX!ZfX#^fX~P%0XOP]XR]X[]Xp]X!O]X!Q]X!Z]X!ZfX!j]X!n]X#P]X#Q]X#^]X#^fX#ifX#l]X#m]X#n]X#o]X#p]X#q]X#r]X#s]X#t]X#u]X#w]X#y]X#z]X$P]X(_]X(p]X(w]X(x]X~O!efX!i]X!ifX(pfX~P'JlOP;vOQ;vOSfOd=rOe!iOnkOp;vOqkOrkOxkOz;vO|;vO!QWO!UkO!VkO!]XO!g;yO!jZO!m;vO!n;vO!o;vO!q;zO!s;}O!v!hO$V!kO$m=pO(R)ZO(TTO(WUO(_VO(m[O~O!Z<ZO![$pa~Oh%VOn%WOp%XOq$tOr$tOx%YOz%ZO|<fO!Q${O!]$|O!g=wO!j$xO#h<lO$V%_O$s<hO$u<jO$x%`O(R(tO(TTO(WUO(_$uO(w$}O(x%PO~Oj)bO~P( bOp!cX(p!cX~P# qOp(hX(p(hX~P#!dO![]X![fX~P'JlO!WfX!W$yX!ZfX!Z$yX#^fX~P!/qO#i<OO~O!e#vO#i<OO~O#^<`O~O#t<SO~O#^<pO!Z(uX![(uX~O#^<`O!Z(sX![(sX~O#i<qO~Og<sO~P!0uO#i<yO~O#i<zO~O!e#vO#i<{O~O!e#vO#i<qO~O#}<|O~P#BPO#i<}O~O#i=OO~O#i=TO~O#i=UO~O#i=VO~O#i=WO~O#}=XO~P!0uO#}=YO~P!0uO#Q#R#S#U#V#Y#g#h#s$m$s$u$x%[%]%g%h%i%p%r%u%v%x%z~'|T#m!V'z(S#nq#l#op!O'{$['{(R$^(c~",goto:"$8f)ZPPPPPP)[PP)_P)pP+Q/VPPPP6aPP6wPP<oP@cP@yP@yPPP@yPCRP@yP@yP@yPCVPC[PCyPHsPPPHwPPPPHwKzPPPLQLrPHwPHwPP! QHwPPPHwPHwP!#XHwP!&o!'t!'}P!(q!(u!(q!,SPPPPPPP!,s!'tPP!-T!.uP!2RHwHw!2W!5d!:Q!:Q!>PPPP!>XHwPPPPPPPPPP!AhP!BuPPHw!DWPHwPHwHwHwHwHwPHw!EjP!HtP!KzP!LO!LY!L^!L^P!HqP!Lb!LbP# hP# lHwPHw# r#$wCV@yP@yP@y@yP#&U@y@y#(h@y#+`@y#-l@y@y#.[#0p#0p#0u#1O#0p#1ZPP#0pP@y#1s@y#5r@y@y6aPPP#9wPPP#:b#:bP#:bP#:x#:bPP#;OP#:uP#:u#;c#:u#;}#<T#<W)_#<Z)_P#<b#<b#<bP)_P)_P)_P)_PP)_P#<h#<kP#<k)_P#<oP#<rP)_P)_P)_P)_P)_P)_)_PP#<x#=O#=Z#=a#=g#=m#=s#>R#>X#>c#>i#>s#>y#?Z#?a#@R#@e#@k#@q#AP#Af#CZ#Ci#Cp#E[#Ej#G[#Gj#Gp#Gv#G|#HW#H^#Hd#Hn#IQ#IWPPPPPPPPPPP#I^PPPPPPP#JR#MY#Nr#Ny$ RPPP$&mP$&v$)o$0Y$0]$0`$1_$1b$1i$1qP$1w$1zP$2h$2l$3d$4r$4w$5_PP$5d$5j$5n$5q$5u$5y$6u$7^$7u$7y$7|$8P$8V$8Y$8^$8bR!|RoqOXst!Z#d%l&p&r&s&u,n,s2S2VY!vQ'^-`1g5qQ%svQ%{yQ&S|Q&h!VS'U!e-WQ'd!iS'j!r!yU*h$|*X*lQ+l%|Q+y&UQ,_&bQ-^']Q-h'eQ-p'kQ0U*nQ1q,`R<m;z%SdOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&p&r&s&u&y'R'`'p(R(T(Z(b(v(x(|){*f+U+Y,k,n,s-d-l-z.Q.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3d4q5y6Z6[6_6r8i8x9SS#q];w!r)]$Z$n'V)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sU*{%[<e<fQ+q&OQ,a&eQ,h&mQ0r+dQ0u+fQ1S+rQ1y,fQ3W.bQ5V0wQ5]0}Q6Q1rQ7O3[Q8U5^R9Y7Q'QkOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=s!S!nQ!r!v!y!z$|'U']'^'j'k'l*h*l*n*o-W-^-`-p0U0X1g5q5s%[$ti#v$b$c$d$x${%O%Q%]%^%b)w*P*R*T*W*^*d*t*u+c+f+},Q.a.z/_/h/r/s/u0Y0[0g0h0i1^1a1i3Z4U4V4a4f4w5R5U5x6|7l7v7|8Q8f9V9e9n9t:S:f:t:};V;^<^<_<a<b<c<d<g<h<i<j<k<l<t<u<v<w<y<z<}=O=P=Q=R=S=T=U=X=Y=p=x=y=|=}Q&V|Q'S!eS'Y%h-ZQ+q&OQ,a&eQ0f+OQ1S+rQ1X+xQ1x,eQ1y,fQ5]0}Q5f1ZQ6Q1rQ6T1tQ6U1wQ8U5^Q8X5cQ8q6WQ9|8YQ:Y8nR<o*XrnOXst!V!Z#d%l&g&p&r&s&u,n,s2S2VR,c&i&z^OPXYstuvwz!Z!`!g!j!o#S#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n%l%s&Q&i&l&m&p&r&s&u&y'R'`'p(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=r=s[#]WZ#W#Z'V(R!b%im#h#i#l$x%d%g([(f(g(h*W*[*_+W+X+Z,j-Q.O.U.V.W.Y/h/k2[3S3T4X6h6yQ%vxQ%zyS&P|&UQ&]!TQ'a!hQ'c!iQ(o#sS+k%{%|Q+o&OQ,Y&`Q,^&bS-g'd'eQ.d(pQ0{+lQ1R+rQ1T+sQ1W+wQ1l,ZS1p,_,`Q2t-hQ5[0}Q5`1QQ5e1YQ6P1qQ8T5^Q8W5bQ9x8SR:w9y!U$zi$d%O%Q%]%^%b*P*R*^*t*u.z/r0Y0[0g0h0i4V4w7|9e=p=x=y!^%xy!i!u%z%{%|'T'c'd'e'i's*g+k+l-T-g-h-o/{0O0{2m2t2{4i4j4m7s9pQ+e%vQ,O&YQ,R&ZQ,]&bQ.c(oQ1k,YU1o,^,_,`Q3].dQ5z1lS6O1p1qQ8m6P#f=t#v$b$c$x${)w*T*W*d+c+f+},Q.a/_/h/s/u1^1a1i3Z4U4a4f5R5U5x6|7l7v8Q8f9V9n9t:S:f:t:};V;^<a<c<g<i<k<t<v<y<}=P=R=T=X=|=}o=u<^<_<b<d<h<j<l<u<w<z=O=Q=S=U=YW%Ti%V*v=pS&Y!Q&gQ&Z!RQ&[!SQ+S%cR+|&W%]%Si#v$b$c$d$x${%O%Q%]%^%b)w*P*R*T*W*^*d*t*u+c+f+},Q.a.z/_/h/r/s/u0Y0[0g0h0i1^1a1i3Z4U4V4a4f4w5R5U5x6|7l7v7|8Q8f9V9e9n9t:S:f:t:};V;^<^<_<a<b<c<d<g<h<i<j<k<l<t<u<v<w<y<z<}=O=P=Q=R=S=T=U=X=Y=p=x=y=|=}T)x$u)yV*{%[<e<fW'Y!e%h*X-ZS({#y#zQ+`%qQ+v&RS.](k(lQ1b,SQ4x0cR8^5k'QkOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=s$i$^c#Y#e%p%r%t(Q(W(r(w)P)Q)R)S)T)U)V)W)X)Y)[)^)`)e)o+a+u-U-s-x-}.P.n.q.u.w.x.y/]0j2c2f2v2}3c3h3i3j3k3l3m3n3o3p3q3r3s3t3w3x4P5O5Y6k6q6v7V7W7a7b8`8|9Q9[9b9c:c:y;R;x=gT#TV#U'RkOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sQ'W!eR2i-W!W!nQ!e!r!v!y!z$|'U']'^'j'k'l*X*h*l*n*o-W-^-`-p0U0X1g5q5sR1d,UnqOXst!Z#d%l&p&r&s&u,n,s2S2VQ&w!^Q't!xS(q#u<OQ+i%yQ,W&]Q,X&_Q-e'bQ-r'mS.m(v<qS0k+U<{Q0y+jQ1f,VQ2Z,uQ2],vQ2e-RQ2r-fQ2u-jS5P0l=VQ5W0zS5Z0|=WQ6j2gQ6n2sQ6s2zQ8R5XQ8}6lQ9O6oQ9R6tR:`8z$d$]c#Y#e%r%t(Q(W(r(w)P)Q)R)S)T)U)V)W)X)Y)[)^)`)e)o+a+u-U-s-x-}.P.n.q.u.x.y/]0j2c2f2v2}3c3h3i3j3k3l3m3n3o3p3q3r3s3t3w3x4P5O5Y6k6q6v7V7W7a7b8`8|9Q9[9b9c:c:y;R;x=gS(m#p'gQ(}#zS+_%p.wS.^(l(nR3U._'QkOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sS#q];wQ&r!XQ&s!YQ&u![Q&v!]R2R,qQ'_!hQ+b%vQ-c'aS.`(o+eQ2p-bW3Y.c.d0q0sQ6m2qW6z3V3X3]5TU9U6{6}7PU:e9W9X9ZS;T:d:gQ;b;UR;j;cU!wQ'^-`T5o1g5q!Q_OXZ`st!V!Z#d#h%d%l&g&i&p&r&s&u(h,n,s.V2S2V]!pQ!r'^-`1g5qT#q];w%^{OPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&m&p&r&s&u&y'R'`'p(R(T(Z(b(v(x(|){*f+U+Y+d,k,n,s-d-l-z.Q.b.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3[3d4q5y6Z6[6_6r7Q8i8x9SS({#y#zS.](k(l!s=^$Z$n'V)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sU$fd)],hS(n#p'gU*s%R(u3vU0e*z.i7]Q5T0rQ6{3WQ9X7OR:g9Ym!tQ!r!v!y!z'^'j'k'l-`-p1g5q5sQ'r!uS(d#g1|S-n'i'uQ/n*ZQ/{*gQ2|-qQ4]/oQ4i/}Q4j0OQ4o0WQ7h4WS7s4k4mS7w4p4rQ9g7iQ9k7oQ9p7tQ9u7yS:o9l9mS;Y:p:sS;e;Z;[S;m;f;gS;q;n;oR;t;rQ#wbQ'q!uS(c#g1|S(e#m+TQ+V%eQ+g%wQ+m%}U-m'i'r'uQ.R(dQ/m*ZQ/|*gQ0P*iQ0x+hQ1m,[S2y-n-qQ3R.ZS4[/n/oQ4e/yS4h/{0WQ4l0QQ5|1nQ6u2|Q7g4WQ7k4]U7r4i4o4rQ7u4nQ8k5}S9f7h7iQ9j7oQ9r7wQ9s7xQ:V8lQ:m9gS:n9k9mQ:v9uQ;P:WS;X:o:sS;d;Y;ZS;l;e;gS;p;m;oQ;s;qQ;u;tQ=a=[Q=l=eR=m=fV!wQ'^-`%^aOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&m&p&r&s&u&y'R'`'p(R(T(Z(b(v(x(|){*f+U+Y+d,k,n,s-d-l-z.Q.b.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3[3d4q5y6Z6[6_6r7Q8i8x9SS#wz!j!r=Z$Z$n'V)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sR=a=r%^bOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&m&p&r&s&u&y'R'`'p(R(T(Z(b(v(x(|){*f+U+Y+d,k,n,s-d-l-z.Q.b.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3[3d4q5y6Z6[6_6r7Q8i8x9SQ%ej!^%wy!i!u%z%{%|'T'c'd'e'i's*g+k+l-T-g-h-o/{0O0{2m2t2{4i4j4m7s9pS%}z!jQ+h%xQ,[&bW1n,],^,_,`U5}1o1p1qS8l6O6PQ:W8m!r=[$Z$n'V)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sQ=e=qR=f=r%QeOPXYstuvw!Z!`!g!o#S#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&p&r&s&u&y'R'`'p(T(Z(b(v(x(|){*f+U+Y+d,k,n,s-d-l-z.Q.b.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3[3d4q5y6Z6[6_6r7Q8i8x9SY#bWZ#W#Z(R!b%im#h#i#l$x%d%g([(f(g(h*W*[*_+W+X+Z,j-Q.O.U.V.W.Y/h/k2[3S3T4X6h6yQ,i&m!p=]$Z$n)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sR=`'VU'Z!e%h*XR2k-Z%SdOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&p&r&s&u&y'R'`'p(R(T(Z(b(v(x(|){*f+U+Y,k,n,s-d-l-z.Q.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3d4q5y6Z6[6_6r8i8x9S!r)]$Z$n'V)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sQ,h&mQ0r+dQ3W.bQ7O3[R9Y7Q!b$Tc#Y%p(Q(W(r(w)X)Y)^)e+u-s-x-}.P.n.q/]0j2v2}3c3s5O5Y6q6v7V9Q:c;x!P<U)[)o-U.w2c2f3h3q3r3w4P6k7W7a7b8`8|9[9b9c:y;R=g!f$Vc#Y%p(Q(W(r(w)U)V)X)Y)^)e+u-s-x-}.P.n.q/]0j2v2}3c3s5O5Y6q6v7V9Q:c;x!T<W)[)o-U.w2c2f3h3n3o3q3r3w4P6k7W7a7b8`8|9[9b9c:y;R=g!^$Zc#Y%p(Q(W(r(w)^)e+u-s-x-}.P.n.q/]0j2v2}3c3s5O5Y6q6v7V9Q:c;xQ4V/fz=s)[)o-U.w2c2f3h3w4P6k7W7a7b8`8|9[9b9c:y;R=gQ=x=zR=y={'QkOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sS$oh$pR3|/P'XgOPWXYZhstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n$p%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/P/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sT$kf$qQ$ifS)h$l)lR)t$qT$jf$qT)j$l)l'XhOPWXYZhstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$Z$_$a$e$n$p%l%s&Q&i&l&m&p&r&s&u&y'R'V'`'p(R(T(Z(b(v(x(|)q){*f+U+Y+d,k,n,s-P-S-d-l-z.Q.b.o.v/P/Q/i0V0d0l0|1j1z1{1}2P2S2V2X2h2x3O3[3d3{4q5m5y6Z6[6_6i6r7Q8i8x9S9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=sT$oh$pQ$rhR)s$p%^jOPWXYZstuvw!Z!`!g!o#S#W#Z#d#o#u#x#{$O$P$Q$R$S$T$U$V$W$X$_$a$e%l%s&Q&i&l&m&p&r&s&u&y'R'`'p(R(T(Z(b(v(x(|){*f+U+Y+d,k,n,s-d-l-z.Q.b.o.v/i0V0d0l0|1j1z1{1}2P2S2V2X2x3O3[3d4q5y6Z6[6_6r7Q8i8x9S!s=q$Z$n'V)q-P-S/Q2h3{5m6i9}:a;v;y;z;}<O<P<Q<R<S<T<U<V<W<X<Y<Z<]<`<m<p<q<s<{<|=V=W=s#glOPXZst!Z!`!o#S#d#o#{$n%l&i&l&m&p&r&s&u&y'R'`(|)q*f+Y+d,k,n,s-d.b/Q/i0V0d1j1z1{1}2P2S2V2X3[3{4q5y6Z6[6_7Q8i8x!U%Ri$d%O%Q%]%^%b*P*R*^*t*u.z/r0Y0[0g0h0i4V4w7|9e=p=x=y#f(u#v$b$c$x${)w*T*W*d+c+f+},Q.a/_/h/s/u1^1a1i3Z4U4a4f5R5U5x6|7l7v8Q8f9V9n9t:S:f:t:};V;^<a<c<g<i<k<t<v<y<}=P=R=T=X=|=}Q+P%`Q/^)|o3v<^<_<b<d<h<j<l<u<w<z=O=Q=S=U=Y!U$yi$d%O%Q%]%^%b*P*R*^*t*u.z/r0Y0[0g0h0i4V4w7|9e=p=x=yQ*`$zU*i$|*X*lQ+Q%aQ0Q*j#f=c#v$b$c$x${)w*T*W*d+c+f+},Q.a/_/h/s/u1^1a1i3Z4U4a4f5R5U5x6|7l7v8Q8f9V9n9t:S:f:t:};V;^<a<c<g<i<k<t<v<y<}=P=R=T=X=|=}n=d<^<_<b<d<h<j<l<u<w<z=O=Q=S=U=YQ=h=tQ=i=uQ=j=vR=k=w!U%Ri$d%O%Q%]%^%b*P*R*^*t*u.z/r0Y0[0g0h0i4V4w7|9e=p=x=y#f(u#v$b$c$x${)w*T*W*d+c+f+},Q.a/_/h/s/u1^1a1i3Z4U4a4f5R5U5x6|7l7v8Q8f9V9n9t:S:f:t:};V;^<a<c<g<i<k<t<v<y<}=P=R=T=X=|=}o3v<^<_<b<d<h<j<l<u<w<z=O=Q=S=U=YnoOXst!Z#d%l&p&r&s&u,n,s2S2VS*c${*WQ,|&|Q,}'OR4`/s%[%Si#v$b$c$d$x${%O%Q%]%^%b)w*P*R*T*W*^*d*t*u+c+f+},Q.a.z/_/h/r/s/u0Y0[0g0h0i1^1a1i3Z4U4V4a4f4w5R5U5x6|7l7v7|8Q8f9V9e9n9t:S:f:t:};V;^<^<_<a<b<c<d<g<h<i<j<k<l<t<u<v<w<y<z<}=O=P=Q=R=S=T=U=X=Y=p=x=y=|=}Q,P&ZQ1`,RQ5i1_R8]5jV*k$|*X*lU*k$|*X*lT5p1g5qS/y*f/iQ4n0VT7x4q:PQ+g%wQ0P*iQ0x+hQ1m,[Q5|1nQ8k5}Q:V8lR;P:W!U%Oi$d%O%Q%]%^%b*P*R*^*t*u.z/r0Y0[0g0h0i4V4w7|9e=p=x=yx*P$v)c*Q*r+R/q0^0_3y4^4{4|4}7f7z9v:l=b=n=oS0Y*q0Z#f<a#v$b$c$x${)w*T*W*d+c+f+},Q.a/_/h/s/u1^1a1i3Z4U4a4f5R5U5x6|7l7v8Q8f9V9n9t:S:f:t:};V;^<a<c<g<i<k<t<v<y<}=P=R=T=X=|=}n<b<^<_<b<d<h<j<l<u<w<z=O=Q=S=U=Y!d<t(s)a*Y*b.e.h.l/Y/f/v0p1]3`4S4_4c5h7R7U7m7p7}8P9i9q9w:q:u;W;];h=z={`<u3u7X7[7`9]:h:k;kS=P.g3aT=Q7Z9`!U%Qi$d%O%Q%]%^%b*P*R*^*t*u.z/r0Y0[0g0h0i4V4w7|9e=p=x=y|*R$v)c*S*q+R/b/q0^0_3y4^4s4{4|4}7f7z9v:l=b=n=oS0[*r0]#f<c#v$b$c$x${)w*T*W*d+c+f+},Q.a/_/h/s/u1^1a1i3Z4U4a4f5R5U5x6|7l7v8Q8f9V9n9t:S:f:t:};V;^<a<c<g<i<k<t<v<y<}=P=R=T=X=|=}n<d<^<_<b<d<h<j<l<u<w<z=O=Q=S=U=Y!h<v(s)a*Y*b.f.g.l/Y/f/v0p1]3^3`4S4_4c5h7R7S7U7m7p7}8P9i9q9w:q:u;W;];h=z={d<w3u7Y7Z7`9]9^:h:i:k;kS=R.h3bT=S7[9arnOXst!V!Z#d%l&g&p&r&s&u,n,s2S2VQ&d!UR,k&mrnOXst!V!Z#d%l&g&p&r&s&u,n,s2S2VR&d!UQ,T&[R1[+|snOXst!V!Z#d%l&g&p&r&s&u,n,s2S2VQ1h,YS5w1k1lU8e5u5v5zS:R8g8hS:{:Q:TQ;_:|R;i;`Q&k!VR,d&gR6T1tR:Y8nS&P|&UR1T+sQ&p!WR,n&qR,t&vT2T,s2VR,x&wQ,w&wR2^,xQ'w!{R-t'wSsOtQ#dXT%os#dQ#OTR'y#OQ#RUR'{#RQ)y$uR/Z)yQ#UVR(O#UQ#XWU(U#X(V-{Q(V#YR-{(WQ-X'WR2j-XQ.p(wS3e.p3fR3f.qQ-`'^R2n-`Y!rQ'^-`1g5qR'h!rQ.{)cR3z.{U#_W%g*WU(]#_(^-|Q(^#`R-|(XQ-['ZR2l-[t`OXst!V!Z#d%l&g&i&p&r&s&u,n,s2S2VS#hZ%dU#r`#h.VR.V(hQ(i#jQ.S(eW.[(i.S3P6wQ3P.TR6w3QQ)l$lR/R)lQ$phR)r$pQ$`cU)_$`-w<[Q-w;xR<[)oQ/l*ZW4Y/l4Z7j9hU4Z/m/n/oS7j4[4]R9h7k$e*O$v(s)a)c*Y*b*q*r*|*}+R.g.h.j.k.l/Y/b/d/f/q/v0^0_0p1]3^3_3`3u3y4S4^4_4c4s4u4{4|4}5h7R7S7T7U7Z7[7^7_7`7f7m7p7z7}8P9]9^9_9i9q9v9w:h:i:j:k:l:q:u;W;];h;k=b=n=o=z={Q/t*bU4b/t4d7nQ4d/vR7n4cS*l$|*XR0S*lx*Q$v)c*q*r+R/q0^0_3y4^4{4|4}7f7z9v:l=b=n=o!d.e(s)a*Y*b.g.h.l/Y/f/v0p1]3`4S4_4c5h7R7U7m7p7}8P9i9q9w:q:u;W;];h=z={U/c*Q.e7Xa7X3u7Z7[7`9]:h:k;kQ0Z*qQ3a.gU4t0Z3a9`R9`7Z|*S$v)c*q*r+R/b/q0^0_3y4^4s4{4|4}7f7z9v:l=b=n=o!h.f(s)a*Y*b.g.h.l/Y/f/v0p1]3^3`4S4_4c5h7R7S7U7m7p7}8P9i9q9w:q:u;W;];h=z={U/e*S.f7Ye7Y3u7Z7[7`9]9^:h:i:k;kQ0]*rQ3b.hU4v0]3b9aR9a7[Q*w%UR0a*wQ5S0pR8O5SQ+[%jR0o+[Q5l1bS8_5l:OR:O8`Q,V&]R1e,VQ5q1gR8b5qQ1s,aS6R1s8oR8o6TQ1O+oW5_1O5a8V9zQ5a1RQ8V5`R9z8WQ+t&PR1U+tQ2V,sR6c2VYrOXst#dQ&t!ZQ+^%lQ,m&pQ,o&rQ,p&sQ,r&uQ2Q,nS2T,s2VR6b2SQ%npQ&x!_Q&{!aQ&}!bQ'P!cQ'o!uQ+]%kQ+i%yQ+{&VQ,c&kQ,z&zW-k'i'q'r'uQ-r'mQ0R*kQ0y+jS1v,d,gQ2_,yQ2`,|Q2a,}Q2u-jW2w-m-n-q-sQ5W0zQ5d1XQ5g1]Q5{1mQ6V1xQ6a2RU6p2v2y2|Q6s2zQ8R5XQ8Z5fQ8[5hQ8a5pQ8j5|Q8p6US9P6q6uQ9R6tQ9{8XQ:U8kQ:Z8qQ:b9QQ:x9|Q;O:VQ;S:cR;a;PQ%yyQ'b!iQ'm!uU+j%z%{%|Q-R'TU-f'c'd'eS-j'i'sQ/z*gS0z+k+lQ2g-TS2s-g-hQ2z-oS4g/{0OQ5X0{Q6l2mQ6o2tQ6t2{U7q4i4j4mQ9o7sR:r9pS$wi=pR*x%VU%Ui%V=pR0`*vQ$viS(s#v+fS)a$b$cQ)c$dQ*Y$xS*b${*WQ*q%OQ*r%QQ*|%]Q*}%^Q+R%bQ.g<aQ.h<cQ.j<gQ.k<iQ.l<kQ/Y)wQ/b*PQ/d*RQ/f*TQ/q*^S/v*d/hQ0^*tQ0_*ul0p+c,Q.a1a1i3Z5x6|8f9V:S:f:};VQ1]+}Q3^<tQ3_<vQ3`<yS3u<^<_Q3y.zS4S/_4UQ4^/rQ4_/sQ4c/uQ4s0YQ4u0[Q4{0gQ4|0hQ4}0iQ5h1^Q7R<}Q7S=PQ7T=RQ7U=TQ7Z<bQ7[<dQ7^<hQ7_<jQ7`<lQ7f4VQ7m4aQ7p4fQ7z4wQ7}5RQ8P5UQ9]<zQ9^<uQ9_<wQ9i7lQ9q7vQ9v7|Q9w8QQ:h=OQ:i=QQ:j=SQ:k=UQ:l9eQ:q9nQ:u9tQ;W=XQ;]:tQ;h;^Q;k=YQ=b=pQ=n=xQ=o=yQ=z=|R={=}Q*z%[Q.i<eR7]<fnpOXst!Z#d%l&p&r&s&u,n,s2S2VQ!fPS#fZ#oQ&z!`W'f!o*f0V4qQ'}#SQ)O#{Q)p$nS,g&i&lQ,l&mQ,y&yS-O'R/iQ-b'`Q.s(|Q/V)qQ0m+YQ0s+dQ2O,kQ2q-dQ3X.bQ4O/QQ4y0dQ5v1jQ6X1zQ6Y1{Q6^1}Q6`2PQ6e2XQ7P3[Q7c3{Q8h5yQ8t6ZQ8u6[Q8w6_Q9Z7QQ:T8iR:_8x#[cOPXZst!Z!`!o#d#o#{%l&i&l&m&p&r&s&u&y'R'`(|*f+Y+d,k,n,s-d.b/i0V0d1j1z1{1}2P2S2V2X3[4q5y6Z6[6_7Q8i8xQ#YWQ#eYQ%puQ%rvS%tw!gS(Q#W(TQ(W#ZQ(r#uQ(w#xQ)P$OQ)Q$PQ)R$QQ)S$RQ)T$SQ)U$TQ)V$UQ)W$VQ)X$WQ)Y$XQ)[$ZQ)^$_Q)`$aQ)e$eW)o$n)q/Q3{Q+a%sQ+u&QS-U'V2hQ-s'pS-x(R-zQ-}(ZQ.P(bQ.n(vQ.q(xQ.u;vQ.w;yQ.x;zQ.y;}Q/]){Q0j+UQ2c-PQ2f-SQ2v-lQ2}.QQ3c.oQ3h<OQ3i<PQ3j<QQ3k<RQ3l<SQ3m<TQ3n<UQ3o<VQ3p<WQ3q<XQ3r<YQ3s.vQ3t<]Q3w<`Q3x<mQ4P<ZQ5O0lQ5Y0|Q6k<pQ6q2xQ6v3OQ7V3dQ7W<qQ7a<sQ7b<{Q8`5mQ8|6iQ9Q6rQ9[<|Q9b=VQ9c=WQ:c9SQ:y9}Q;R:aQ;x#SR=g=sR#[WR'X!el!tQ!r!v!y!z'^'j'k'l-`-p1g5q5sS'T!e-WU*g$|*X*lS-T'U']S0O*h*nQ0W*oQ2m-^Q4m0UR4r0XR(y#xQ!fQT-_'^-`]!qQ!r'^-`1g5qQ#p]R'g;wR)d$dY!uQ'^-`1g5qQ'i!rS's!v!yS'u!z5sS-o'j'kQ-q'lR2{-pT#kZ%dS#jZ%dS%jm,jU(e#h#i#lS.T(f(gQ.X(hQ0n+ZQ3Q.UU3R.V.W.YS6x3S3TR9T6yd#^W#W#Z%g(R([*W+W.O/hr#gZm#h#i#l%d(f(g(h+Z.U.V.W.Y3S3T6yS*Z$x*_Q/o*[Q1|,jQ2d-QQ4W/kQ6g2[Q7i4XQ8{6hT=_'V+XV#aW%g*WU#`W%g*WS(S#W([U(X#Z+W/hS-V'V+XT-y(R.OV'[!e%h*XQ$lfR)v$qT)k$l)lR3}/PT*]$x*_T*e${*WQ0q+cQ1_,QQ3V.aQ5j1aQ5u1iQ6}3ZQ8g5xQ9W6|Q:Q8fQ:d9VQ:|:SQ;U:fQ;`:}R;c;VnqOXst!Z#d%l&p&r&s&u,n,s2S2VQ&j!VR,c&gtmOXst!U!V!Z#d%l&g&p&r&s&u,n,s2S2VR,j&mT%km,jR1c,SR,b&eQ&T|R+z&UR+p&OT&n!W&qT&o!W&qT2U,s2V",nodeNames:"⚠ ArithOp ArithOp ?. JSXStartTag LineComment BlockComment Script Hashbang ExportDeclaration export Star as VariableName String Escape from ; default FunctionDeclaration async function VariableDefinition > < TypeParamList const TypeDefinition extends ThisType this LiteralType ArithOp Number BooleanLiteral TemplateType InterpolationEnd Interpolation InterpolationStart NullType null VoidType void TypeofType typeof MemberExpression . PropertyName [ TemplateString Escape Interpolation super RegExp ] ArrayExpression Spread , } { ObjectExpression Property async get set PropertyDefinition Block : NewTarget new NewExpression ) ( ArgList UnaryExpression delete LogicOp BitOp YieldExpression yield AwaitExpression await ParenthesizedExpression ClassExpression class ClassBody MethodDeclaration Decorator @ MemberExpression PrivatePropertyName CallExpression TypeArgList CompareOp < declare Privacy static abstract override PrivatePropertyDefinition PropertyDeclaration readonly accessor Optional TypeAnnotation Equals StaticBlock FunctionExpression ArrowFunction ParamList ParamList ArrayPattern ObjectPattern PatternProperty Privacy readonly Arrow MemberExpression BinaryExpression ArithOp ArithOp ArithOp ArithOp BitOp CompareOp instanceof satisfies in CompareOp BitOp BitOp BitOp LogicOp LogicOp ConditionalExpression LogicOp LogicOp AssignmentExpression UpdateOp PostfixExpression CallExpression InstantiationExpression TaggedTemplateExpression DynamicImport import ImportMeta JSXElement JSXSelfCloseEndTag JSXSelfClosingTag JSXIdentifier JSXBuiltin JSXIdentifier JSXNamespacedName JSXMemberExpression JSXSpreadAttribute JSXAttribute JSXAttributeValue JSXEscape JSXEndTag JSXOpenTag JSXFragmentTag JSXText JSXEscape JSXStartCloseTag JSXCloseTag PrefixCast < ArrowFunction TypeParamList SequenceExpression InstantiationExpression KeyofType keyof UniqueType unique ImportType InferredType infer TypeName ParenthesizedType FunctionSignature ParamList NewSignature IndexedType TupleType Label ArrayType ReadonlyType ObjectType MethodType PropertyType IndexSignature PropertyDefinition CallSignature TypePredicate asserts is NewSignature new UnionType LogicOp IntersectionType LogicOp ConditionalType ParameterizedType ClassDeclaration abstract implements type VariableDeclaration let var using TypeAliasDeclaration InterfaceDeclaration interface EnumDeclaration enum EnumBody NamespaceDeclaration namespace module AmbientDeclaration declare GlobalDeclaration global ClassDeclaration ClassBody AmbientFunctionDeclaration ExportGroup VariableName VariableName ImportDeclaration ImportGroup ForStatement for ForSpec ForInSpec ForOfSpec of WhileStatement while WithStatement with DoStatement do IfStatement if else SwitchStatement switch SwitchBody CaseLabel case DefaultLabel TryStatement try CatchClause catch FinallyClause finally ReturnStatement return ThrowStatement throw BreakStatement break ContinueStatement continue DebuggerStatement debugger LabeledStatement ExpressionStatement SingleExpression SingleClassItem",maxTerm:378,context:$p,nodeProps:[["isolate",-8,5,6,14,35,37,49,51,53,""],["group",-26,9,17,19,66,206,210,214,215,217,220,223,233,235,241,243,245,247,250,256,262,264,266,268,270,272,273,"Statement",-34,13,14,30,33,34,40,49,52,53,55,60,68,70,74,78,80,82,83,108,109,118,119,135,138,140,141,142,143,144,146,147,166,168,170,"Expression",-23,29,31,35,39,41,43,172,174,176,177,179,180,181,183,184,185,187,188,189,200,202,204,205,"Type",-3,86,101,107,"ClassItem"],["openedBy",23,"<",36,"InterpolationStart",54,"[",58,"{",71,"(",159,"JSXStartCloseTag"],["closedBy",-2,24,167,">",38,"InterpolationEnd",48,"]",59,"}",72,")",164,"JSXEndTag"]],propSources:[Up],skippedNodes:[0,5,6,276],repeatNodeCount:37,tokenData:"$Fq07[R!bOX%ZXY+gYZ-yZ[+g[]%Z]^.c^p%Zpq+gqr/mrs3cst:_tuEruvJSvwLkwx! Yxy!'iyz!(sz{!)}{|!,q|}!.O}!O!,q!O!P!/Y!P!Q!9j!Q!R#:O!R![#<_![!]#I_!]!^#Jk!^!_#Ku!_!`$![!`!a$$v!a!b$*T!b!c$,r!c!}Er!}#O$-|#O#P$/W#P#Q$4o#Q#R$5y#R#SEr#S#T$7W#T#o$8b#o#p$<r#p#q$=h#q#r$>x#r#s$@U#s$f%Z$f$g+g$g#BYEr#BY#BZ$A`#BZ$ISEr$IS$I_$A`$I_$I|Er$I|$I}$Dk$I}$JO$Dk$JO$JTEr$JT$JU$A`$JU$KVEr$KV$KW$A`$KW&FUEr&FU&FV$A`&FV;'SEr;'S;=`I|<%l?HTEr?HT?HU$A`?HUOEr(n%d_$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z&j&hT$h&jO!^&c!_#o&c#p;'S&c;'S;=`&w<%lO&c&j&zP;=`<%l&c'|'U]$h&j(X!bOY&}YZ&cZw&}wx&cx!^&}!^!_'}!_#O&}#O#P&c#P#o&}#o#p'}#p;'S&};'S;=`(l<%lO&}!b(SU(X!bOY'}Zw'}x#O'}#P;'S'};'S;=`(f<%lO'}!b(iP;=`<%l'}'|(oP;=`<%l&}'[(y]$h&j(UpOY(rYZ&cZr(rrs&cs!^(r!^!_)r!_#O(r#O#P&c#P#o(r#o#p)r#p;'S(r;'S;=`*a<%lO(rp)wU(UpOY)rZr)rs#O)r#P;'S)r;'S;=`*Z<%lO)rp*^P;=`<%l)r'[*dP;=`<%l(r#S*nX(Up(X!bOY*gZr*grs'}sw*gwx)rx#O*g#P;'S*g;'S;=`+Z<%lO*g#S+^P;=`<%l*g(n+dP;=`<%l%Z07[+rq$h&j(Up(X!b'z0/lOX%ZXY+gYZ&cZ[+g[p%Zpq+gqr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p$f%Z$f$g+g$g#BY%Z#BY#BZ+g#BZ$IS%Z$IS$I_+g$I_$JT%Z$JT$JU+g$JU$KV%Z$KV$KW+g$KW&FU%Z&FU&FV+g&FV;'S%Z;'S;=`+a<%l?HT%Z?HT?HU+g?HUO%Z07[.ST(V#S$h&j'{0/lO!^&c!_#o&c#p;'S&c;'S;=`&w<%lO&c07[.n_$h&j(Up(X!b'{0/lOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z)3p/x`$h&j!n),Q(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`0z!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW1V`#u(Ch$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`2X!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW2d_#u(Ch$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'At3l_(T':f$h&j(X!bOY4kYZ5qZr4krs7nsw4kwx5qx!^4k!^!_8p!_#O4k#O#P5q#P#o4k#o#p8p#p;'S4k;'S;=`:X<%lO4k(^4r_$h&j(X!bOY4kYZ5qZr4krs7nsw4kwx5qx!^4k!^!_8p!_#O4k#O#P5q#P#o4k#o#p8p#p;'S4k;'S;=`:X<%lO4k&z5vX$h&jOr5qrs6cs!^5q!^!_6y!_#o5q#o#p6y#p;'S5q;'S;=`7h<%lO5q&z6jT$c`$h&jO!^&c!_#o&c#p;'S&c;'S;=`&w<%lO&c`6|TOr6yrs7]s;'S6y;'S;=`7b<%lO6y`7bO$c``7eP;=`<%l6y&z7kP;=`<%l5q(^7w]$c`$h&j(X!bOY&}YZ&cZw&}wx&cx!^&}!^!_'}!_#O&}#O#P&c#P#o&}#o#p'}#p;'S&};'S;=`(l<%lO&}!r8uZ(X!bOY8pYZ6yZr8prs9hsw8pwx6yx#O8p#O#P6y#P;'S8p;'S;=`:R<%lO8p!r9oU$c`(X!bOY'}Zw'}x#O'}#P;'S'};'S;=`(f<%lO'}!r:UP;=`<%l8p(^:[P;=`<%l4k%9[:hh$h&j(Up(X!bOY%ZYZ&cZq%Zqr<Srs&}st%ZtuCruw%Zwx(rx!^%Z!^!_*g!_!c%Z!c!}Cr!}#O%Z#O#P&c#P#R%Z#R#SCr#S#T%Z#T#oCr#o#p*g#p$g%Z$g;'SCr;'S;=`El<%lOCr(r<__WS$h&j(Up(X!bOY<SYZ&cZr<Srs=^sw<Swx@nx!^<S!^!_Bm!_#O<S#O#P>`#P#o<S#o#pBm#p;'S<S;'S;=`Cl<%lO<S(Q=g]WS$h&j(X!bOY=^YZ&cZw=^wx>`x!^=^!^!_?q!_#O=^#O#P>`#P#o=^#o#p?q#p;'S=^;'S;=`@h<%lO=^&n>gXWS$h&jOY>`YZ&cZ!^>`!^!_?S!_#o>`#o#p?S#p;'S>`;'S;=`?k<%lO>`S?XSWSOY?SZ;'S?S;'S;=`?e<%lO?SS?hP;=`<%l?S&n?nP;=`<%l>`!f?xWWS(X!bOY?qZw?qwx?Sx#O?q#O#P?S#P;'S?q;'S;=`@b<%lO?q!f@eP;=`<%l?q(Q@kP;=`<%l=^'`@w]WS$h&j(UpOY@nYZ&cZr@nrs>`s!^@n!^!_Ap!_#O@n#O#P>`#P#o@n#o#pAp#p;'S@n;'S;=`Bg<%lO@ntAwWWS(UpOYApZrAprs?Ss#OAp#O#P?S#P;'SAp;'S;=`Ba<%lOAptBdP;=`<%lAp'`BjP;=`<%l@n#WBvYWS(Up(X!bOYBmZrBmrs?qswBmwxApx#OBm#O#P?S#P;'SBm;'S;=`Cf<%lOBm#WCiP;=`<%lBm(rCoP;=`<%l<S%9[C}i$h&j(m%1l(Up(X!bOY%ZYZ&cZr%Zrs&}st%ZtuCruw%Zwx(rx!Q%Z!Q![Cr![!^%Z!^!_*g!_!c%Z!c!}Cr!}#O%Z#O#P&c#P#R%Z#R#SCr#S#T%Z#T#oCr#o#p*g#p$g%Z$g;'SCr;'S;=`El<%lOCr%9[EoP;=`<%lCr07[FRk$h&j(Up(X!b$[#t(R,2j(c$I[OY%ZYZ&cZr%Zrs&}st%ZtuEruw%Zwx(rx}%Z}!OGv!O!Q%Z!Q![Er![!^%Z!^!_*g!_!c%Z!c!}Er!}#O%Z#O#P&c#P#R%Z#R#SEr#S#T%Z#T#oEr#o#p*g#p$g%Z$g;'SEr;'S;=`I|<%lOEr+dHRk$h&j(Up(X!b$[#tOY%ZYZ&cZr%Zrs&}st%ZtuGvuw%Zwx(rx}%Z}!OGv!O!Q%Z!Q![Gv![!^%Z!^!_*g!_!c%Z!c!}Gv!}#O%Z#O#P&c#P#R%Z#R#SGv#S#T%Z#T#oGv#o#p*g#p$g%Z$g;'SGv;'S;=`Iv<%lOGv+dIyP;=`<%lGv07[JPP;=`<%lEr(KWJ_`$h&j(Up(X!b#n(ChOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KWKl_$h&j$P(Ch(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z,#xLva(x+JY$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sv%ZvwM{wx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KWNW`$h&j#y(Ch(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'At! c_(W';W$h&j(UpOY!!bYZ!#hZr!!brs!#hsw!!bwx!$xx!^!!b!^!_!%z!_#O!!b#O#P!#h#P#o!!b#o#p!%z#p;'S!!b;'S;=`!'c<%lO!!b'l!!i_$h&j(UpOY!!bYZ!#hZr!!brs!#hsw!!bwx!$xx!^!!b!^!_!%z!_#O!!b#O#P!#h#P#o!!b#o#p!%z#p;'S!!b;'S;=`!'c<%lO!!b&z!#mX$h&jOw!#hwx6cx!^!#h!^!_!$Y!_#o!#h#o#p!$Y#p;'S!#h;'S;=`!$r<%lO!#h`!$]TOw!$Ywx7]x;'S!$Y;'S;=`!$l<%lO!$Y`!$oP;=`<%l!$Y&z!$uP;=`<%l!#h'l!%R]$c`$h&j(UpOY(rYZ&cZr(rrs&cs!^(r!^!_)r!_#O(r#O#P&c#P#o(r#o#p)r#p;'S(r;'S;=`*a<%lO(r!Q!&PZ(UpOY!%zYZ!$YZr!%zrs!$Ysw!%zwx!&rx#O!%z#O#P!$Y#P;'S!%z;'S;=`!']<%lO!%z!Q!&yU$c`(UpOY)rZr)rs#O)r#P;'S)r;'S;=`*Z<%lO)r!Q!'`P;=`<%l!%z'l!'fP;=`<%l!!b/5|!'t_!j/.^$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z#&U!)O_!i!Lf$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z-!n!*[b$h&j(Up(X!b(S%&f#o(ChOY%ZYZ&cZr%Zrs&}sw%Zwx(rxz%Zz{!+d{!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW!+o`$h&j(Up(X!b#l(ChOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z+;x!,|`$h&j(Up(X!bp+4YOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z,$U!.Z_!Z+Jf$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z07[!/ec$h&j(Up(X!b!O.2^OY%ZYZ&cZr%Zrs&}sw%Zwx(rx!O%Z!O!P!0p!P!Q%Z!Q![!3Y![!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z#%|!0ya$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!O%Z!O!P!2O!P!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z#%|!2Z_!Y!L^$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad!3eg$h&j(Up(X!bq'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q![!3Y![!^%Z!^!_*g!_!g%Z!g!h!4|!h#O%Z#O#P&c#P#R%Z#R#S!3Y#S#X%Z#X#Y!4|#Y#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad!5Vg$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx{%Z{|!6n|}%Z}!O!6n!O!Q%Z!Q![!8S![!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S!8S#S#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad!6wc$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q![!8S![!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S!8S#S#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad!8_c$h&j(Up(X!bq'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q![!8S![!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S!8S#S#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z07[!9uf$h&j(Up(X!b#m(ChOY!;ZYZ&cZr!;Zrs!<nsw!;Zwx!Lcxz!;Zz{#-}{!P!;Z!P!Q#/d!Q!^!;Z!^!_#(i!_!`#7S!`!a#8i!a!}!;Z!}#O#,f#O#P!Dy#P#o!;Z#o#p#(i#p;'S!;Z;'S;=`#-w<%lO!;Z?O!;fb$h&j(Up(X!b!V7`OY!;ZYZ&cZr!;Zrs!<nsw!;Zwx!Lcx!P!;Z!P!Q#&`!Q!^!;Z!^!_#(i!_!}!;Z!}#O#,f#O#P!Dy#P#o!;Z#o#p#(i#p;'S!;Z;'S;=`#-w<%lO!;Z>^!<w`$h&j(X!b!V7`OY!<nYZ&cZw!<nwx!=yx!P!<n!P!Q!Eq!Q!^!<n!^!_!Gr!_!}!<n!}#O!KS#O#P!Dy#P#o!<n#o#p!Gr#p;'S!<n;'S;=`!L]<%lO!<n<z!>Q^$h&j!V7`OY!=yYZ&cZ!P!=y!P!Q!>|!Q!^!=y!^!_!@c!_!}!=y!}#O!CW#O#P!Dy#P#o!=y#o#p!@c#p;'S!=y;'S;=`!Ek<%lO!=y<z!?Td$h&j!V7`O!^&c!_#W&c#W#X!>|#X#Z&c#Z#[!>|#[#]&c#]#^!>|#^#a&c#a#b!>|#b#g&c#g#h!>|#h#i&c#i#j!>|#j#k!>|#k#m&c#m#n!>|#n#o&c#p;'S&c;'S;=`&w<%lO&c7`!@hX!V7`OY!@cZ!P!@c!P!Q!AT!Q!}!@c!}#O!Ar#O#P!Bq#P;'S!@c;'S;=`!CQ<%lO!@c7`!AYW!V7`#W#X!AT#Z#[!AT#]#^!AT#a#b!AT#g#h!AT#i#j!AT#j#k!AT#m#n!AT7`!AuVOY!ArZ#O!Ar#O#P!B[#P#Q!@c#Q;'S!Ar;'S;=`!Bk<%lO!Ar7`!B_SOY!ArZ;'S!Ar;'S;=`!Bk<%lO!Ar7`!BnP;=`<%l!Ar7`!BtSOY!@cZ;'S!@c;'S;=`!CQ<%lO!@c7`!CTP;=`<%l!@c<z!C][$h&jOY!CWYZ&cZ!^!CW!^!_!Ar!_#O!CW#O#P!DR#P#Q!=y#Q#o!CW#o#p!Ar#p;'S!CW;'S;=`!Ds<%lO!CW<z!DWX$h&jOY!CWYZ&cZ!^!CW!^!_!Ar!_#o!CW#o#p!Ar#p;'S!CW;'S;=`!Ds<%lO!CW<z!DvP;=`<%l!CW<z!EOX$h&jOY!=yYZ&cZ!^!=y!^!_!@c!_#o!=y#o#p!@c#p;'S!=y;'S;=`!Ek<%lO!=y<z!EnP;=`<%l!=y>^!Ezl$h&j(X!b!V7`OY&}YZ&cZw&}wx&cx!^&}!^!_'}!_#O&}#O#P&c#P#W&}#W#X!Eq#X#Z&}#Z#[!Eq#[#]&}#]#^!Eq#^#a&}#a#b!Eq#b#g&}#g#h!Eq#h#i&}#i#j!Eq#j#k!Eq#k#m&}#m#n!Eq#n#o&}#o#p'}#p;'S&};'S;=`(l<%lO&}8r!GyZ(X!b!V7`OY!GrZw!Grwx!@cx!P!Gr!P!Q!Hl!Q!}!Gr!}#O!JU#O#P!Bq#P;'S!Gr;'S;=`!J|<%lO!Gr8r!Hse(X!b!V7`OY'}Zw'}x#O'}#P#W'}#W#X!Hl#X#Z'}#Z#[!Hl#[#]'}#]#^!Hl#^#a'}#a#b!Hl#b#g'}#g#h!Hl#h#i'}#i#j!Hl#j#k!Hl#k#m'}#m#n!Hl#n;'S'};'S;=`(f<%lO'}8r!JZX(X!bOY!JUZw!JUwx!Arx#O!JU#O#P!B[#P#Q!Gr#Q;'S!JU;'S;=`!Jv<%lO!JU8r!JyP;=`<%l!JU8r!KPP;=`<%l!Gr>^!KZ^$h&j(X!bOY!KSYZ&cZw!KSwx!CWx!^!KS!^!_!JU!_#O!KS#O#P!DR#P#Q!<n#Q#o!KS#o#p!JU#p;'S!KS;'S;=`!LV<%lO!KS>^!LYP;=`<%l!KS>^!L`P;=`<%l!<n=l!Ll`$h&j(Up!V7`OY!LcYZ&cZr!Lcrs!=ys!P!Lc!P!Q!Mn!Q!^!Lc!^!_# o!_!}!Lc!}#O#%P#O#P!Dy#P#o!Lc#o#p# o#p;'S!Lc;'S;=`#&Y<%lO!Lc=l!Mwl$h&j(Up!V7`OY(rYZ&cZr(rrs&cs!^(r!^!_)r!_#O(r#O#P&c#P#W(r#W#X!Mn#X#Z(r#Z#[!Mn#[#](r#]#^!Mn#^#a(r#a#b!Mn#b#g(r#g#h!Mn#h#i(r#i#j!Mn#j#k!Mn#k#m(r#m#n!Mn#n#o(r#o#p)r#p;'S(r;'S;=`*a<%lO(r8Q# vZ(Up!V7`OY# oZr# ors!@cs!P# o!P!Q#!i!Q!}# o!}#O#$R#O#P!Bq#P;'S# o;'S;=`#$y<%lO# o8Q#!pe(Up!V7`OY)rZr)rs#O)r#P#W)r#W#X#!i#X#Z)r#Z#[#!i#[#])r#]#^#!i#^#a)r#a#b#!i#b#g)r#g#h#!i#h#i)r#i#j#!i#j#k#!i#k#m)r#m#n#!i#n;'S)r;'S;=`*Z<%lO)r8Q#$WX(UpOY#$RZr#$Rrs!Ars#O#$R#O#P!B[#P#Q# o#Q;'S#$R;'S;=`#$s<%lO#$R8Q#$vP;=`<%l#$R8Q#$|P;=`<%l# o=l#%W^$h&j(UpOY#%PYZ&cZr#%Prs!CWs!^#%P!^!_#$R!_#O#%P#O#P!DR#P#Q!Lc#Q#o#%P#o#p#$R#p;'S#%P;'S;=`#&S<%lO#%P=l#&VP;=`<%l#%P=l#&]P;=`<%l!Lc?O#&kn$h&j(Up(X!b!V7`OY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#W%Z#W#X#&`#X#Z%Z#Z#[#&`#[#]%Z#]#^#&`#^#a%Z#a#b#&`#b#g%Z#g#h#&`#h#i%Z#i#j#&`#j#k#&`#k#m%Z#m#n#&`#n#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z9d#(r](Up(X!b!V7`OY#(iZr#(irs!Grsw#(iwx# ox!P#(i!P!Q#)k!Q!}#(i!}#O#+`#O#P!Bq#P;'S#(i;'S;=`#,`<%lO#(i9d#)th(Up(X!b!V7`OY*gZr*grs'}sw*gwx)rx#O*g#P#W*g#W#X#)k#X#Z*g#Z#[#)k#[#]*g#]#^#)k#^#a*g#a#b#)k#b#g*g#g#h#)k#h#i*g#i#j#)k#j#k#)k#k#m*g#m#n#)k#n;'S*g;'S;=`+Z<%lO*g9d#+gZ(Up(X!bOY#+`Zr#+`rs!JUsw#+`wx#$Rx#O#+`#O#P!B[#P#Q#(i#Q;'S#+`;'S;=`#,Y<%lO#+`9d#,]P;=`<%l#+`9d#,cP;=`<%l#(i?O#,o`$h&j(Up(X!bOY#,fYZ&cZr#,frs!KSsw#,fwx#%Px!^#,f!^!_#+`!_#O#,f#O#P!DR#P#Q!;Z#Q#o#,f#o#p#+`#p;'S#,f;'S;=`#-q<%lO#,f?O#-tP;=`<%l#,f?O#-zP;=`<%l!;Z07[#.[b$h&j(Up(X!b'|0/l!V7`OY!;ZYZ&cZr!;Zrs!<nsw!;Zwx!Lcx!P!;Z!P!Q#&`!Q!^!;Z!^!_#(i!_!}!;Z!}#O#,f#O#P!Dy#P#o!;Z#o#p#(i#p;'S!;Z;'S;=`#-w<%lO!;Z07[#/o_$h&j(Up(X!bT0/lOY#/dYZ&cZr#/drs#0nsw#/dwx#4Ox!^#/d!^!_#5}!_#O#/d#O#P#1p#P#o#/d#o#p#5}#p;'S#/d;'S;=`#6|<%lO#/d06j#0w]$h&j(X!bT0/lOY#0nYZ&cZw#0nwx#1px!^#0n!^!_#3R!_#O#0n#O#P#1p#P#o#0n#o#p#3R#p;'S#0n;'S;=`#3x<%lO#0n05W#1wX$h&jT0/lOY#1pYZ&cZ!^#1p!^!_#2d!_#o#1p#o#p#2d#p;'S#1p;'S;=`#2{<%lO#1p0/l#2iST0/lOY#2dZ;'S#2d;'S;=`#2u<%lO#2d0/l#2xP;=`<%l#2d05W#3OP;=`<%l#1p01O#3YW(X!bT0/lOY#3RZw#3Rwx#2dx#O#3R#O#P#2d#P;'S#3R;'S;=`#3r<%lO#3R01O#3uP;=`<%l#3R06j#3{P;=`<%l#0n05x#4X]$h&j(UpT0/lOY#4OYZ&cZr#4Ors#1ps!^#4O!^!_#5Q!_#O#4O#O#P#1p#P#o#4O#o#p#5Q#p;'S#4O;'S;=`#5w<%lO#4O00^#5XW(UpT0/lOY#5QZr#5Qrs#2ds#O#5Q#O#P#2d#P;'S#5Q;'S;=`#5q<%lO#5Q00^#5tP;=`<%l#5Q05x#5zP;=`<%l#4O01p#6WY(Up(X!bT0/lOY#5}Zr#5}rs#3Rsw#5}wx#5Qx#O#5}#O#P#2d#P;'S#5};'S;=`#6v<%lO#5}01p#6yP;=`<%l#5}07[#7PP;=`<%l#/d)3h#7ab$h&j$P(Ch(Up(X!b!V7`OY!;ZYZ&cZr!;Zrs!<nsw!;Zwx!Lcx!P!;Z!P!Q#&`!Q!^!;Z!^!_#(i!_!}!;Z!}#O#,f#O#P!Dy#P#o!;Z#o#p#(i#p;'S!;Z;'S;=`#-w<%lO!;ZAt#8vb$Y#t$h&j(Up(X!b!V7`OY!;ZYZ&cZr!;Zrs!<nsw!;Zwx!Lcx!P!;Z!P!Q#&`!Q!^!;Z!^!_#(i!_!}!;Z!}#O#,f#O#P!Dy#P#o!;Z#o#p#(i#p;'S!;Z;'S;=`#-w<%lO!;Z'Ad#:Zp$h&j(Up(X!bq'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!O%Z!O!P!3Y!P!Q%Z!Q![#<_![!^%Z!^!_*g!_!g%Z!g!h!4|!h#O%Z#O#P&c#P#R%Z#R#S#<_#S#U%Z#U#V#?i#V#X%Z#X#Y!4|#Y#b%Z#b#c#>_#c#d#Bq#d#l%Z#l#m#Es#m#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#<jk$h&j(Up(X!bq'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!O%Z!O!P!3Y!P!Q%Z!Q![#<_![!^%Z!^!_*g!_!g%Z!g!h!4|!h#O%Z#O#P&c#P#R%Z#R#S#<_#S#X%Z#X#Y!4|#Y#b%Z#b#c#>_#c#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#>j_$h&j(Up(X!bq'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#?rd$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q!R#AQ!R!S#AQ!S!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S#AQ#S#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#A]f$h&j(Up(X!bq'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q!R#AQ!R!S#AQ!S!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S#AQ#S#b%Z#b#c#>_#c#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#Bzc$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q!Y#DV!Y!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S#DV#S#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#Dbe$h&j(Up(X!bq'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q!Y#DV!Y!^%Z!^!_*g!_#O%Z#O#P&c#P#R%Z#R#S#DV#S#b%Z#b#c#>_#c#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#E|g$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q![#Ge![!^%Z!^!_*g!_!c%Z!c!i#Ge!i#O%Z#O#P&c#P#R%Z#R#S#Ge#S#T%Z#T#Z#Ge#Z#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z'Ad#Gpi$h&j(Up(X!bq'9tOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!Q%Z!Q![#Ge![!^%Z!^!_*g!_!c%Z!c!i#Ge!i#O%Z#O#P&c#P#R%Z#R#S#Ge#S#T%Z#T#Z#Ge#Z#b%Z#b#c#>_#c#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z*)x#Il_!e$b$h&j#})Lv(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z)[#Jv_al$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z04f#LS^h#)`#P-<U(Up(X!b$m7`OY*gZr*grs'}sw*gwx)rx!P*g!P!Q#MO!Q!^*g!^!_#Mt!_!`$ f!`#O*g#P;'S*g;'S;=`+Z<%lO*g(n#MXX$j&j(Up(X!bOY*gZr*grs'}sw*gwx)rx#O*g#P;'S*g;'S;=`+Z<%lO*g(El#M}Z#p(Ch(Up(X!bOY*gZr*grs'}sw*gwx)rx!_*g!_!`#Np!`#O*g#P;'S*g;'S;=`+Z<%lO*g(El#NyX$P(Ch(Up(X!bOY*gZr*grs'}sw*gwx)rx#O*g#P;'S*g;'S;=`+Z<%lO*g(El$ oX#q(Ch(Up(X!bOY*gZr*grs'}sw*gwx)rx#O*g#P;'S*g;'S;=`+Z<%lO*g*)x$!ga#^*!Y$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`0z!`!a$#l!a#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(K[$#w_#i(Cl$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z*)x$%Vag!*r#q(Ch$e#|$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`$&[!`!a$'f!a#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW$&g_#q(Ch$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW$'qa#p(Ch$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`!a$(v!a#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW$)R`#p(Ch$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(Kd$*`a(p(Ct$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!a%Z!a!b$+e!b#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW$+p`$h&j#z(Ch(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z%#`$,}_!z$Ip$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z04f$.X_!Q0,v$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(n$/]Z$h&jO!^$0O!^!_$0f!_#i$0O#i#j$0k#j#l$0O#l#m$2^#m#o$0O#o#p$0f#p;'S$0O;'S;=`$4i<%lO$0O(n$0VT_#S$h&jO!^&c!_#o&c#p;'S&c;'S;=`&w<%lO&c#S$0kO_#S(n$0p[$h&jO!Q&c!Q![$1f![!^&c!_!c&c!c!i$1f!i#T&c#T#Z$1f#Z#o&c#o#p$3|#p;'S&c;'S;=`&w<%lO&c(n$1kZ$h&jO!Q&c!Q![$2^![!^&c!_!c&c!c!i$2^!i#T&c#T#Z$2^#Z#o&c#p;'S&c;'S;=`&w<%lO&c(n$2cZ$h&jO!Q&c!Q![$3U![!^&c!_!c&c!c!i$3U!i#T&c#T#Z$3U#Z#o&c#p;'S&c;'S;=`&w<%lO&c(n$3ZZ$h&jO!Q&c!Q![$0O![!^&c!_!c&c!c!i$0O!i#T&c#T#Z$0O#Z#o&c#p;'S&c;'S;=`&w<%lO&c#S$4PR!Q![$4Y!c!i$4Y#T#Z$4Y#S$4]S!Q![$4Y!c!i$4Y#T#Z$4Y#q#r$0f(n$4lP;=`<%l$0O#1[$4z_!W#)l$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z(KW$6U`#w(Ch$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z+;p$7c_$h&j(Up(X!b(_+4QOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z07[$8qk$h&j(Up(X!b(R,2j$^#t(c$I[OY%ZYZ&cZr%Zrs&}st%Ztu$8buw%Zwx(rx}%Z}!O$:f!O!Q%Z!Q![$8b![!^%Z!^!_*g!_!c%Z!c!}$8b!}#O%Z#O#P&c#P#R%Z#R#S$8b#S#T%Z#T#o$8b#o#p*g#p$g%Z$g;'S$8b;'S;=`$<l<%lO$8b+d$:qk$h&j(Up(X!b$^#tOY%ZYZ&cZr%Zrs&}st%Ztu$:fuw%Zwx(rx}%Z}!O$:f!O!Q%Z!Q![$:f![!^%Z!^!_*g!_!c%Z!c!}$:f!}#O%Z#O#P&c#P#R%Z#R#S$:f#S#T%Z#T#o$:f#o#p*g#p$g%Z$g;'S$:f;'S;=`$<f<%lO$:f+d$<iP;=`<%l$:f07[$<oP;=`<%l$8b#Jf$<{X!]#Hb(Up(X!bOY*gZr*grs'}sw*gwx)rx#O*g#P;'S*g;'S;=`+Z<%lO*g,#x$=sa(w+JY$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_!`Ka!`#O%Z#O#P&c#P#o%Z#o#p*g#p#q$+e#q;'S%Z;'S;=`+a<%lO%Z)>v$?V_![(CdtBr$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z?O$@a_!o7`$h&j(Up(X!bOY%ZYZ&cZr%Zrs&}sw%Zwx(rx!^%Z!^!_*g!_#O%Z#O#P&c#P#o%Z#o#p*g#p;'S%Z;'S;=`+a<%lO%Z07[$Aq|$h&j(Up(X!b'z0/l$[#t(R,2j(c$I[OX%ZXY+gYZ&cZ[+g[p%Zpq+gqr%Zrs&}st%ZtuEruw%Zwx(rx}%Z}!OGv!O!Q%Z!Q![Er![!^%Z!^!_*g!_!c%Z!c!}Er!}#O%Z#O#P&c#P#R%Z#R#SEr#S#T%Z#T#oEr#o#p*g#p$f%Z$f$g+g$g#BYEr#BY#BZ$A`#BZ$ISEr$IS$I_$A`$I_$JTEr$JT$JU$A`$JU$KVEr$KV$KW$A`$KW&FUEr&FU&FV$A`&FV;'SEr;'S;=`I|<%l?HTEr?HT?HU$A`?HUOEr07[$D|k$h&j(Up(X!b'{0/l$[#t(R,2j(c$I[OY%ZYZ&cZr%Zrs&}st%ZtuEruw%Zwx(rx}%Z}!OGv!O!Q%Z!Q![Er![!^%Z!^!_*g!_!c%Z!c!}Er!}#O%Z#O#P&c#P#R%Z#R#SEr#S#T%Z#T#oEr#o#p*g#p$g%Z$g;'SEr;'S;=`I|<%lOEr",tokenizers:[Ip,Lp,Bp,Qp,2,3,4,5,6,7,8,9,10,11,12,13,14,Np,new gd("$S~RRtu[#O#Pg#S#T#|~_P#o#pb~gOv~~jVO#i!P#i#j!U#j#l!P#l#m!q#m;'S!P;'S;=`#v<%lO!P~!UO!S~~!XS!Q![!e!c!i!e#T#Z!e#o#p#Z~!hR!Q![!q!c!i!q#T#Z!q~!tR!Q![!}!c!i!}#T#Z!}~#QR!Q![!P!c!i!P#T#Z!P~#^R!Q![#g!c!i#g#T#Z#g~#jS!Q![#g!c!i#g#T#Z#g#q#r!P~#yP;=`<%l!P~$RO(a~~",141,338),new gd("j~RQYZXz{^~^O(O~~aP!P!Qd~iO(P~~",25,321)],topRules:{Script:[0,7],SingleExpression:[1,274],SingleClassItem:[2,275]},dialects:{jsx:0,ts:15091},dynamicPrecedences:{78:1,80:1,92:1,168:1,198:1},specialized:[{term:325,get:e=>zp[e]||-1},{term:341,get:e=>jp[e]||-1},{term:93,get:e=>Wp[e]||-1}],tokenPrec:15116}),qp=[_u("function ${name}(${params}) {\n\t${}\n}",{label:"function",detail:"definition",type:"keyword"}),_u("for (let ${index} = 0; ${index} < ${bound}; ${index}++) {\n\t${}\n}",{label:"for",detail:"loop",type:"keyword"}),_u("for (let ${name} of ${collection}) {\n\t${}\n}",{label:"for",detail:"of loop",type:"keyword"}),_u("do {\n\t${}\n} while (${})",{label:"do",detail:"loop",type:"keyword"}),_u("while (${}) {\n\t${}\n}",{label:"while",detail:"loop",type:"keyword"}),_u("try {\n\t${}\n} catch (${error}) {\n\t${}\n}",{label:"try",detail:"/ catch block",type:"keyword"}),_u("if (${}) {\n\t${}\n}",{label:"if",detail:"block",type:"keyword"}),_u("if (${}) {\n\t${}\n} else {\n\t${}\n}",{label:"if",detail:"/ else block",type:"keyword"}),_u("class ${name} {\n\tconstructor(${params}) {\n\t\t${}\n\t}\n}",{label:"class",detail:"definition",type:"keyword"}),_u('import {${names}} from "${module}"\n${}',{label:"import",detail:"named",type:"keyword"}),_u('import ${name} from "${module}"\n${}',{label:"import",detail:"default",type:"keyword"})],Vp=qp.concat([_u("interface ${name} {\n\t${}\n}",{label:"interface",detail:"definition",type:"keyword"}),_u("type ${name} = ${type}",{label:"type",detail:"definition",type:"keyword"}),_u("enum ${name} {\n\t${}\n}",{label:"enum",detail:"definition",type:"keyword"})]),Hp=new Wi,Zp=new Set(["Script","Block","FunctionExpression","FunctionDeclaration","ArrowFunction","MethodDeclaration","ForStatement"]);function Gp(e){return(t,n)=>{let r=t.node.getChild("VariableDefinition");return r&&n(r,e),!0}}const Kp=["FunctionDeclaration"],Yp={FunctionDeclaration:Gp("function"),ClassDeclaration:Gp("class"),ClassExpression:()=>!0,EnumDeclaration:Gp("constant"),TypeAliasDeclaration:Gp("type"),NamespaceDeclaration:Gp("namespace"),VariableDefinition(e,t){e.matchContext(Kp)||t(e,"variable")},TypeDefinition(e,t){t(e,"type")},__proto__:null};function Jp(e,t){let n=Hp.get(t);if(n)return n;let r=[],a=!0;function o(t,n){let a=e.sliceString(t.from,t.to);r.push({label:a,type:n})}return t.cursor(Si.IncludeAnonymous).iterate((t=>{if(a)a=!1;else if(t.name){let e=Yp[t.name];if(e&&e(t,o)||Zp.has(t.name))return!1}else if(t.to-t.from>8192){for(let n of Jp(e,t.node))r.push(n);return!1}})),Hp.set(t,r),r}const eh=/^[\w$\xa1-\uffff][\w$\d\xa1-\uffff]*$/,th=["TemplateString","String","RegExp","LineComment","BlockComment","VariableDefinition","TypeDefinition","Label","PropertyDefinition","PropertyName","PrivatePropertyDefinition","PrivatePropertyName","JSXText","JSXAttributeValue","JSXOpenTag","JSXCloseTag","JSXSelfClosingTag",".","?."];function nh(e){let t=Qs(e.state).resolveInner(e.pos,-1);if(th.indexOf(t.name)>-1)return null;let n="VariableName"==t.name||t.to-t.from<20&&eh.test(e.state.sliceDoc(t.from,t.to));if(!n&&!e.explicit)return null;let r=[];for(let n=t;n;n=n.parent)Zp.has(n.name)&&(r=r.concat(Jp(e.state.doc,n)));return{options:r,from:n?t.from:e.pos,validFor:eh}}const rh=Fs.define({name:"javascript",parser:Xp.configure({props:[rl.add({IfStatement:dl({except:/^\s*({|else\b)/}),TryStatement:dl({except:/^\s*({|catch\b|finally\b)/}),LabeledStatement:e=>e.baseIndent,SwitchBody:e=>{let t=e.textAfter,n=/^\s*\}/.test(t),r=/^\s*(case|default)\b/.test(t);return e.baseIndent+(n?0:r?1:2)*e.unit},Block:cl({closing:"}"}),ArrowFunction:e=>e.baseIndent+e.unit,"TemplateString BlockComment":()=>null,"Statement Property":dl({except:/^{/}),JSXElement(e){let t=/^\s*<\//.test(e.textAfter);return e.lineIndent(e.node.from)+(t?0:e.unit)},JSXEscape(e){let t=/\s*\}/.test(e.textAfter);return e.lineIndent(e.node.from)+(t?0:e.unit)},"JSXOpenTag JSXSelfClosingTag":e=>e.column(e.node.from)+e.unit}),hl.add({"Block ClassBody SwitchBody EnumBody ObjectExpression ArrayExpression ObjectType":fl,BlockComment:e=>({from:e.from+2,to:e.to-2})})]}),languageData:{closeBrackets:{brackets:["(","[","{","'",'"',"`"]},commentTokens:{line:"//",block:{open:"/*",close:"*/"}},indentOnInput:/^\s*(?:case |default:|\{|\}|<\/)$/,wordChars:"$"}}),ah={test:e=>/^JSX/.test(e.name),facet:Ns({commentTokens:{block:{open:"{/*",close:"*/}"}}})},oh=rh.configure({dialect:"ts"},"typescript"),ih=rh.configure({dialect:"jsx",props:[Is.add((e=>e.isTop?[ah]:void 0))]}),sh=rh.configure({dialect:"jsx ts",props:[Is.add((e=>e.isTop?[ah]:void 0))]},"typescript");let lh=e=>({label:e,type:"keyword"});const ch="break case const continue default delete export extends false finally in instanceof let new return static super switch this throw true typeof var yield".split(" ").map(lh),uh=ch.concat(["declare","implements","private","protected","public"].map(lh));function dh(e={}){let t=e.jsx?e.typescript?sh:ih:e.typescript?oh:rh,n=e.typescript?Vp.concat(uh):qp.concat(ch);return new Gs(t,[rh.data.of({autocomplete:(r=th,a=_c(n),e=>{for(let t=Qs(e.state).resolveInner(e.pos,-1);t;t=t.parent){if(r.indexOf(t.name)>-1)return null;if(t.type.isTop)break}return a(e)})}),rh.data.of({autocomplete:nh}),e.jsx?fh:[]]);var r,a}function ph(e,t,n=e.length){for(let r=null==t?void 0:t.firstChild;r;r=r.nextSibling)if("JSXIdentifier"==r.name||"JSXBuiltin"==r.name||"JSXNamespacedName"==r.name||"JSXMemberExpression"==r.name)return e.sliceString(r.from,Math.min(r.to,n));return""}const hh="object"==typeof navigator&&/Android\b/.test(navigator.userAgent),fh=Ya.inputHandler.of(((e,t,n,r,a)=>{if((hh?e.composing:e.compositionStarted)||e.state.readOnly||t!=n||">"!=r&&"/"!=r||!rh.isActiveAt(e.state,t,-1))return!1;let o=a(),{state:i}=o,s=i.changeByRange((e=>{var t;let n,{head:a}=e,o=Qs(i).resolveInner(a-1,-1);if("JSXStartTag"==o.name&&(o=o.parent),i.doc.sliceString(a-1,a)!=r||"JSXAttributeValue"==o.name&&o.to>a);else{if(">"==r&&"JSXFragmentTag"==o.name)return{range:e,changes:{from:a,insert:"</>"}};if("/"==r&&"JSXStartCloseTag"==o.name){let e=o.parent,r=e.parent;if(r&&e.from==a-2&&((n=ph(i.doc,r.firstChild,a))||"JSXFragmentTag"==(null===(t=r.firstChild)||void 0===t?void 0:t.name))){let e=`${n}>`;return{range:F.cursor(a+e.length,-1),changes:{from:a,insert:e}}}}else if(">"==r){let t=function(e){for(;;){if("JSXOpenTag"==e.name||"JSXSelfClosingTag"==e.name||"JSXFragmentTag"==e.name)return e;if("JSXEscape"==e.name||!e.parent)return null;e=e.parent}}(o);if(t&&"JSXOpenTag"==t.name&&!/^\/?>|^<\//.test(i.doc.sliceString(a,a+2))&&(n=ph(i.doc,t,a)))return{range:e,changes:{from:a,insert:`</${n}>`}}}}return{range:e}}));return!s.changes.empty&&(e.dispatch([o,i.update(s,{userEvent:"input.complete",scrollIntoView:!0})]),!0)})),mh=["_blank","_self","_top","_parent"],gh=["ascii","utf-8","utf-16","latin1","latin1"],vh=["get","post","put","delete"],bh=["application/x-www-form-urlencoded","multipart/form-data","text/plain"],yh=["true","false"],wh={},Oh={a:{attrs:{href:null,ping:null,type:null,media:null,target:mh,hreflang:null}},abbr:wh,address:wh,area:{attrs:{alt:null,coords:null,href:null,target:null,ping:null,media:null,hreflang:null,type:null,shape:["default","rect","circle","poly"]}},article:wh,aside:wh,audio:{attrs:{src:null,mediagroup:null,crossorigin:["anonymous","use-credentials"],preload:["none","metadata","auto"],autoplay:["autoplay"],loop:["loop"],controls:["controls"]}},b:wh,base:{attrs:{href:null,target:mh}},bdi:wh,bdo:wh,blockquote:{attrs:{cite:null}},body:wh,br:wh,button:{attrs:{form:null,formaction:null,name:null,value:null,autofocus:["autofocus"],disabled:["autofocus"],formenctype:bh,formmethod:vh,formnovalidate:["novalidate"],formtarget:mh,type:["submit","reset","button"]}},canvas:{attrs:{width:null,height:null}},caption:wh,center:wh,cite:wh,code:wh,col:{attrs:{span:null}},colgroup:{attrs:{span:null}},command:{attrs:{type:["command","checkbox","radio"],label:null,icon:null,radiogroup:null,command:null,title:null,disabled:["disabled"],checked:["checked"]}},data:{attrs:{value:null}},datagrid:{attrs:{disabled:["disabled"],multiple:["multiple"]}},datalist:{attrs:{data:null}},dd:wh,del:{attrs:{cite:null,datetime:null}},details:{attrs:{open:["open"]}},dfn:wh,div:wh,dl:wh,dt:wh,em:wh,embed:{attrs:{src:null,type:null,width:null,height:null}},eventsource:{attrs:{src:null}},fieldset:{attrs:{disabled:["disabled"],form:null,name:null}},figcaption:wh,figure:wh,footer:wh,form:{attrs:{action:null,name:null,"accept-charset":gh,autocomplete:["on","off"],enctype:bh,method:vh,novalidate:["novalidate"],target:mh}},h1:wh,h2:wh,h3:wh,h4:wh,h5:wh,h6:wh,head:{children:["title","base","link","style","meta","script","noscript","command"]},header:wh,hgroup:wh,hr:wh,html:{attrs:{manifest:null}},i:wh,iframe:{attrs:{src:null,srcdoc:null,name:null,width:null,height:null,sandbox:["allow-top-navigation","allow-same-origin","allow-forms","allow-scripts"],seamless:["seamless"]}},img:{attrs:{alt:null,src:null,ismap:null,usemap:null,width:null,height:null,crossorigin:["anonymous","use-credentials"]}},input:{attrs:{alt:null,dirname:null,form:null,formaction:null,height:null,list:null,max:null,maxlength:null,min:null,name:null,pattern:null,placeholder:null,size:null,src:null,step:null,value:null,width:null,accept:["audio/*","video/*","image/*"],autocomplete:["on","off"],autofocus:["autofocus"],checked:["checked"],disabled:["disabled"],formenctype:bh,formmethod:vh,formnovalidate:["novalidate"],formtarget:mh,multiple:["multiple"],readonly:["readonly"],required:["required"],type:["hidden","text","search","tel","url","email","password","datetime","date","month","week","time","datetime-local","number","range","color","checkbox","radio","file","submit","image","reset","button"]}},ins:{attrs:{cite:null,datetime:null}},kbd:wh,keygen:{attrs:{challenge:null,form:null,name:null,autofocus:["autofocus"],disabled:["disabled"],keytype:["RSA"]}},label:{attrs:{for:null,form:null}},legend:wh,li:{attrs:{value:null}},link:{attrs:{href:null,type:null,hreflang:null,media:null,sizes:["all","16x16","16x16 32x32","16x16 32x32 64x64"]}},map:{attrs:{name:null}},mark:wh,menu:{attrs:{label:null,type:["list","context","toolbar"]}},meta:{attrs:{content:null,charset:gh,name:["viewport","application-name","author","description","generator","keywords"],"http-equiv":["content-language","content-type","default-style","refresh"]}},meter:{attrs:{value:null,min:null,low:null,high:null,max:null,optimum:null}},nav:wh,noscript:wh,object:{attrs:{data:null,type:null,name:null,usemap:null,form:null,width:null,height:null,typemustmatch:["typemustmatch"]}},ol:{attrs:{reversed:["reversed"],start:null,type:["1","a","A","i","I"]},children:["li","script","template","ul","ol"]},optgroup:{attrs:{disabled:["disabled"],label:null}},option:{attrs:{disabled:["disabled"],label:null,selected:["selected"],value:null}},output:{attrs:{for:null,form:null,name:null}},p:wh,param:{attrs:{name:null,value:null}},pre:wh,progress:{attrs:{value:null,max:null}},q:{attrs:{cite:null}},rp:wh,rt:wh,ruby:wh,samp:wh,script:{attrs:{type:["text/javascript"],src:null,async:["async"],defer:["defer"],charset:gh}},section:wh,select:{attrs:{form:null,name:null,size:null,autofocus:["autofocus"],disabled:["disabled"],multiple:["multiple"]}},slot:{attrs:{name:null}},small:wh,source:{attrs:{src:null,type:null,media:null}},span:wh,strong:wh,style:{attrs:{type:["text/css"],media:null,scoped:null}},sub:wh,summary:wh,sup:wh,table:wh,tbody:wh,td:{attrs:{colspan:null,rowspan:null,headers:null}},template:wh,textarea:{attrs:{dirname:null,form:null,maxlength:null,name:null,placeholder:null,rows:null,cols:null,autofocus:["autofocus"],disabled:["disabled"],readonly:["readonly"],required:["required"],wrap:["soft","hard"]}},tfoot:wh,th:{attrs:{colspan:null,rowspan:null,headers:null,scope:["row","col","rowgroup","colgroup"]}},thead:wh,time:{attrs:{datetime:null}},title:wh,tr:wh,track:{attrs:{src:null,label:null,default:null,kind:["subtitles","captions","descriptions","chapters","metadata"],srclang:null}},ul:{children:["li","script","template","ul","ol"]},var:wh,video:{attrs:{src:null,poster:null,width:null,height:null,crossorigin:["anonymous","use-credentials"],preload:["auto","metadata","none"],autoplay:["autoplay"],mediagroup:["movie"],muted:["muted"],controls:["controls"]}},wbr:wh},kh={accesskey:null,class:null,contenteditable:yh,contextmenu:null,dir:["ltr","rtl","auto"],draggable:["true","false","auto"],dropzone:["copy","move","link","string:","file:"],hidden:["hidden"],id:null,inert:["inert"],itemid:null,itemprop:null,itemref:null,itemscope:["itemscope"],itemtype:null,lang:["ar","bn","de","en-GB","en-US","es","fr","hi","id","ja","pa","pt","ru","tr","zh"],spellcheck:yh,autocorrect:yh,autocapitalize:yh,style:null,tabindex:null,title:null,translate:["yes","no"],rel:["stylesheet","alternate","author","bookmark","help","license","next","nofollow","noreferrer","prefetch","prev","search","tag"],role:"alert application article banner button cell checkbox complementary contentinfo dialog document feed figure form grid gridcell heading img list listbox listitem main navigation region row rowgroup search switch tab table tabpanel textbox timer".split(" "),"aria-activedescendant":null,"aria-atomic":yh,"aria-autocomplete":["inline","list","both","none"],"aria-busy":yh,"aria-checked":["true","false","mixed","undefined"],"aria-controls":null,"aria-describedby":null,"aria-disabled":yh,"aria-dropeffect":null,"aria-expanded":["true","false","undefined"],"aria-flowto":null,"aria-grabbed":["true","false","undefined"],"aria-haspopup":yh,"aria-hidden":yh,"aria-invalid":["true","false","grammar","spelling"],"aria-label":null,"aria-labelledby":null,"aria-level":null,"aria-live":["off","polite","assertive"],"aria-multiline":yh,"aria-multiselectable":yh,"aria-owns":null,"aria-posinset":null,"aria-pressed":["true","false","mixed","undefined"],"aria-readonly":yh,"aria-relevant":null,"aria-required":yh,"aria-selected":["true","false","undefined"],"aria-setsize":null,"aria-sort":["ascending","descending","none","other"],"aria-valuemax":null,"aria-valuemin":null,"aria-valuenow":null,"aria-valuetext":null},xh="beforeunload copy cut dragstart dragover dragleave dragenter dragend drag paste focus blur change click load mousedown mouseenter mouseleave mouseup keydown keyup resize scroll unload".split(" ").map((e=>"on"+e));for(let e of xh)kh[e]=null;class Eh{constructor(e,t){this.tags=Object.assign(Object.assign({},Oh),e),this.globalAttrs=Object.assign(Object.assign({},kh),t),this.allTags=Object.keys(this.tags),this.globalAttrNames=Object.keys(this.globalAttrs)}}function Sh(e,t,n=e.length){if(!t)return"";let r=t.firstChild,a=r&&r.getChild("TagName");return a?e.sliceString(a.from,Math.min(a.to,n)):""}function _h(e,t=!1){for(;e;e=e.parent)if("Element"==e.name){if(!t)return e;t=!1}return null}function Ch(e,t,n){let r=n.tags[Sh(e,_h(t))];return(null==r?void 0:r.children)||n.allTags}function Th(e,t){let n=[];for(let r=_h(t);r&&!r.type.isTop;r=_h(r.parent)){let a=Sh(e,r);if(a&&"CloseTag"==r.lastChild.name)break;a&&n.indexOf(a)<0&&("EndTag"==t.name||t.from>=r.firstChild.to)&&n.push(a)}return n}Eh.default=new Eh;const Ah=/^[:\-\.\w\u00b7-\uffff]*$/;function Rh(e,t,n,r,a){let o=/\s*>/.test(e.sliceDoc(a,a+5))?"":">",i=_h(n,!0);return{from:r,to:a,options:Ch(e.doc,i,t).map((e=>({label:e,type:"type"}))).concat(Th(e.doc,n).map(((e,t)=>({label:"/"+e,apply:"/"+e+o,type:"type",boost:99-t})))),validFor:/^\/?[:\-\.\w\u00b7-\uffff]*$/}}function Ph(e,t,n,r){let a=/\s*>/.test(e.sliceDoc(r,r+5))?"":">";return{from:n,to:r,options:Th(e.doc,t).map(((e,t)=>({label:e,apply:e+a,type:"type",boost:99-t}))),validFor:Ah}}function Dh(e){let{extraTags:t,extraGlobalAttributes:n}=e,r=n||t?new Eh(t,n):Eh.default;return e=>function(e,t){let{state:n,pos:r}=t,a=Qs(n).resolveInner(r,-1),o=a.resolve(r);for(let e,t=r;o==a&&(e=a.childBefore(t));){let n=e.lastChild;if(!n||!n.type.isError||n.from<n.to)break;o=a=e,t=n.from}return"TagName"==a.name?a.parent&&/CloseTag$/.test(a.parent.name)?Ph(n,a,a.from,r):Rh(n,e,a,a.from,r):"StartTag"==a.name?Rh(n,e,a,r,r):"StartCloseTag"==a.name||"IncompleteCloseTag"==a.name?Ph(n,a,r,r):"OpenTag"==a.name||"SelfClosingTag"==a.name||"AttributeName"==a.name?function(e,t,n,r,a){let o=_h(n),i=o?t.tags[Sh(e.doc,o)]:null,s=i&&i.attrs?Object.keys(i.attrs):[];return{from:r,to:a,options:(i&&!1===i.globalAttrs?s:s.length?s.concat(t.globalAttrNames):t.globalAttrNames).map((e=>({label:e,type:"property"}))),validFor:Ah}}(n,e,a,"AttributeName"==a.name?a.from:r,r):"Is"==a.name||"AttributeValue"==a.name||"UnquotedAttributeValue"==a.name?function(e,t,n,r,a){var o;let i,s=null===(o=n.parent)||void 0===o?void 0:o.getChild("AttributeName"),l=[];if(s){let o=e.sliceDoc(s.from,s.to),c=t.globalAttrs[o];if(!c){let r=_h(n),a=r?t.tags[Sh(e.doc,r)]:null;c=(null==a?void 0:a.attrs)&&a.attrs[o]}if(c){let t=e.sliceDoc(r,a).toLowerCase(),n='"',o='"';/^['"]/.test(t)?(i='"'==t[0]?/^[^"]*$/:/^[^']*$/,n="",o=e.sliceDoc(a,a+1)==t[0]?"":t[0],t=t.slice(1),r++):i=/^[^\s<>='"]*$/;for(let e of c)l.push({label:e,apply:n+e+o,type:"constant"})}}return{from:r,to:a,options:l,validFor:i}}(n,e,a,"Is"==a.name?r:a.from,r):!t.explicit||"Element"!=o.name&&"Text"!=o.name&&"Document"!=o.name?null:function(e,t,n,r){let a=[],o=0;for(let r of Ch(e.doc,n,t))a.push({label:"<"+r,type:"type"});for(let t of Th(e.doc,n))a.push({label:"</"+t+">",type:"type",boost:99-o++});return{from:r,to:r,options:a,validFor:/^<\/?[:\-\.\w\u00b7-\uffff]*$/}}(n,e,a,r)}(r,e)}const Mh=rh.parser.configure({top:"SingleExpression"}),$h=[{tag:"script",attrs:e=>"text/typescript"==e.type||"ts"==e.lang,parser:oh.parser},{tag:"script",attrs:e=>"text/babel"==e.type||"text/jsx"==e.type,parser:ih.parser},{tag:"script",attrs:e=>"text/typescript-jsx"==e.type,parser:sh.parser},{tag:"script",attrs:e=>/^(importmap|speculationrules|application\/(.+\+)?json)$/i.test(e.type),parser:Mh},{tag:"script",attrs:e=>!e.type||/^(?:text|application)\/(?:x-)?(?:java|ecma)script$|^module$|^$/i.test(e.type),parser:rh.parser},{tag:"style",attrs:e=>(!e.lang||"css"==e.lang)&&(!e.type||/^(text\/)?(x-)?(stylesheet|css)$/i.test(e.type)),parser:op.parser}],Nh=[{name:"style",parser:op.parser.configure({top:"Styles"})}].concat(xh.map((e=>({name:e,parser:rh.parser})))),Ih=Fs.define({name:"html",parser:Tp.configure({props:[rl.add({Element(e){let t=/^(\s*)(<\/)?/.exec(e.textAfter);return e.node.to<=e.pos+t[0].length?e.continue():e.lineIndent(e.node.from)+(t[2]?0:e.unit)},"OpenTag CloseTag SelfClosingTag":e=>e.column(e.node.from)+e.unit,Document(e){if(e.pos+/\s*/.exec(e.textAfter)[0].length<e.node.to)return e.continue();let t,n=null;for(let t=e.node;;){let e=t.lastChild;if(!e||"Element"!=e.name||e.to!=t.to)break;n=t=e}return n&&(!(t=n.lastChild)||"CloseTag"!=t.name&&"SelfClosingTag"!=t.name)?e.lineIndent(n.from)+e.unit:null}}),hl.add({Element(e){let t=e.firstChild,n=e.lastChild;return t&&"OpenTag"==t.name?{from:t.to,to:"CloseTag"==n.name?n.from:e.to}:null}}),Gl.add({"OpenTag CloseTag":e=>e.getChild("TagName")})]}),languageData:{commentTokens:{block:{open:"\x3c!--",close:"--\x3e"}},indentOnInput:/^\s*<\/\w+\W$/,wordChars:"-._"}}),Lh=Ih.configure({wrap:Dp($h,Nh)}),Bh=new Set("area base br col command embed frame hr img input keygen link meta param source track wbr menuitem".split(" ")),Fh=Ya.inputHandler.of(((e,t,n,r,a)=>{if(e.composing||e.state.readOnly||t!=n||">"!=r&&"/"!=r||!Lh.isActiveAt(e.state,t,-1))return!1;let o=a(),{state:i}=o,s=i.changeByRange((e=>{var t,n,a;let o,s=i.doc.sliceString(e.from-1,e.to)==r,{head:l}=e,c=Qs(i).resolveInner(l,-1);if(s&&">"==r&&"EndTag"==c.name){let r=c.parent;if("CloseTag"!=(null===(n=null===(t=r.parent)||void 0===t?void 0:t.lastChild)||void 0===n?void 0:n.name)&&(o=Sh(i.doc,r.parent,l))&&!Bh.has(o))return{range:e,changes:{from:l,to:l+(">"===i.doc.sliceString(l,l+1)?1:0),insert:`</${o}>`}}}else if(s&&"/"==r&&"IncompleteCloseTag"==c.name){let e=c.parent;if(c.from==l-2&&"CloseTag"!=(null===(a=e.lastChild)||void 0===a?void 0:a.name)&&(o=Sh(i.doc,e,l))&&!Bh.has(o)){let e=l+(">"===i.doc.sliceString(l,l+1)?1:0),t=`${o}>`;return{range:F.cursor(l+t.length,-1),changes:{from:l,to:e,insert:t}}}}return{range:e}}));return!s.changes.empty&&(e.dispatch([o,i.update(s,{userEvent:"input.complete",scrollIntoView:!0})]),!0)})),Qh=ds({String:Ds.string,Number:Ds.number,"True False":Ds.bool,PropertyName:Ds.propertyName,Null:Ds.null,", :":Ds.separator,"[ ]":Ds.squareBracket,"{ }":Ds.brace}),Uh=Pd.deserialize({version:14,states:"$bOVQPOOOOQO'#Cb'#CbOnQPO'#CeOvQPO'#ClOOQO'#Cr'#CrQOQPOOOOQO'#Cg'#CgO}QPO'#CfO!SQPO'#CtOOQO,59P,59PO![QPO,59PO!aQPO'#CuOOQO,59W,59WO!iQPO,59WOVQPO,59QOqQPO'#CmO!nQPO,59`OOQO1G.k1G.kOVQPO'#CnO!vQPO,59aOOQO1G.r1G.rOOQO1G.l1G.lOOQO,59X,59XOOQO-E6k-E6kOOQO,59Y,59YOOQO-E6l-E6l",stateData:"#O~OeOS~OQSORSOSSOTSOWQO_ROgPO~OVXOgUO~O^[O~PVO[^O~O]_OVhX~OVaO~O]bO^iX~O^dO~O]_OVha~O]bO^ia~O",goto:"!kjPPPPPPkPPkqwPPPPk{!RPPP!XP!e!hXSOR^bQWQRf_TVQ_Q`WRg`QcZRicQTOQZRQe^RhbRYQR]R",nodeNames:"⚠ JsonText True False Null Number String } { Object Property PropertyName : , ] [ Array",maxTerm:25,nodeProps:[["isolate",-2,6,11,""],["openedBy",7,"{",14,"["],["closedBy",8,"}",15,"]"]],propSources:[Qh],skippedNodes:[0],repeatNodeCount:2,tokenData:"(|~RaXY!WYZ!W]^!Wpq!Wrs!]|}$u}!O$z!Q!R%T!R![&c![!]&t!}#O&y#P#Q'O#Y#Z'T#b#c'r#h#i(Z#o#p(r#q#r(w~!]Oe~~!`Wpq!]qr!]rs!xs#O!]#O#P!}#P;'S!];'S;=`$o<%lO!]~!}Og~~#QXrs!]!P!Q!]#O#P!]#U#V!]#Y#Z!]#b#c!]#f#g!]#h#i!]#i#j#m~#pR!Q![#y!c!i#y#T#Z#y~#|R!Q![$V!c!i$V#T#Z$V~$YR!Q![$c!c!i$c#T#Z$c~$fR!Q![!]!c!i!]#T#Z!]~$rP;=`<%l!]~$zO]~~$}Q!Q!R%T!R![&c~%YRT~!O!P%c!g!h%w#X#Y%w~%fP!Q![%i~%nRT~!Q![%i!g!h%w#X#Y%w~%zR{|&T}!O&T!Q![&Z~&WP!Q![&Z~&`PT~!Q![&Z~&hST~!O!P%c!Q![&c!g!h%w#X#Y%w~&yO[~~'OO_~~'TO^~~'WP#T#U'Z~'^P#`#a'a~'dP#g#h'g~'jP#X#Y'm~'rOR~~'uP#i#j'x~'{P#`#a(O~(RP#`#a(U~(ZOS~~(^P#f#g(a~(dP#i#j(g~(jP#X#Y(m~(rOQ~~(wOW~~(|OV~",tokenizers:[0],topRules:{JsonText:[0,1]},tokenPrec:0}),zh=Fs.define({name:"json",parser:Uh.configure({props:[rl.add({Object:dl({except:/^\s*\}/}),Array:dl({except:/^\s*\]/})}),hl.add({"Object Array":fl})]}),languageData:{closeBrackets:{brackets:["[","{",'"']},indentOnInput:/^\s*[\}\]]$/}});function jh(e){return 45==e||46==e||58==e||e>=65&&e<=90||95==e||e>=97&&e<=122||e>=161}let Wh=null,Xh=null,qh=0;function Vh(e,t){let n=e.pos+t;if(Xh==e&&qh==n)return Wh;for(;9==(r=e.peek(t))||10==r||13==r||32==r;)t++;var r;let a="";for(;;){let n=e.peek(t);if(!jh(n))break;a+=String.fromCharCode(n),t++}return Xh=e,qh=n,Wh=a||null}function Hh(e,t){this.name=e,this.parent=t}const Zh=new Rd({start:null,shift:(e,t,n,r)=>1==t?new Hh(Vh(r,1)||"",e):e,reduce:(e,t)=>11==t&&e?e.parent:e,reuse(e,t,n,r){let a=t.type.id;return 1==a||13==a?new Hh(Vh(r,1)||"",e):e},strict:!1}),Gh=new vd(((e,t)=>{if(60==e.next)if(e.advance(),47==e.next){e.advance();let n=Vh(e,0);if(!n)return e.acceptToken(5);if(t.context&&n==t.context.name)return e.acceptToken(2);for(let r=t.context;r;r=r.parent)if(r.name==n)return e.acceptToken(3,-2);e.acceptToken(4)}else if(33!=e.next&&63!=e.next)return e.acceptToken(1)}),{contextual:!0});function Kh(e,t){return new vd((n=>{let r=0,a=t.charCodeAt(0);e:for(;!(n.next<0);n.advance(),r++)if(n.next==a){for(let e=1;e<t.length;e++)if(n.peek(e)!=t.charCodeAt(e))continue e;break}r&&n.acceptToken(e)}))}const Yh=Kh(36,"--\x3e"),Jh=Kh(37,"?>"),ef=Kh(38,"]]>"),tf=ds({Text:Ds.content,"StartTag StartCloseTag EndTag SelfCloseEndTag":Ds.angleBracket,TagName:Ds.tagName,"MismatchedCloseTag/TagName":[Ds.tagName,Ds.invalid],AttributeName:Ds.attributeName,AttributeValue:Ds.attributeValue,Is:Ds.definitionOperator,"EntityReference CharacterReference":Ds.character,Comment:Ds.blockComment,ProcessingInst:Ds.processingInstruction,DoctypeDecl:Ds.documentMeta,Cdata:Ds.special(Ds.string)}),nf=Pd.deserialize({version:14,states:",lOQOaOOOrOxO'#CfOzOpO'#CiO!tOaO'#CgOOOP'#Cg'#CgO!{OrO'#CrO#TOtO'#CsO#]OpO'#CtOOOP'#DT'#DTOOOP'#Cv'#CvQQOaOOOOOW'#Cw'#CwO#eOxO,59QOOOP,59Q,59QOOOO'#Cx'#CxO#mOpO,59TO#uO!bO,59TOOOP'#C|'#C|O$TOaO,59RO$[OpO'#CoOOOP,59R,59ROOOQ'#C}'#C}O$dOrO,59^OOOP,59^,59^OOOS'#DO'#DOO$lOtO,59_OOOP,59_,59_O$tOpO,59`O$|OpO,59`OOOP-E6t-E6tOOOW-E6u-E6uOOOP1G.l1G.lOOOO-E6v-E6vO%UO!bO1G.oO%UO!bO1G.oO%dOpO'#CkO%lO!bO'#CyO%zO!bO1G.oOOOP1G.o1G.oOOOP1G.w1G.wOOOP-E6z-E6zOOOP1G.m1G.mO&VOpO,59ZO&_OpO,59ZOOOQ-E6{-E6{OOOP1G.x1G.xOOOS-E6|-E6|OOOP1G.y1G.yO&gOpO1G.zO&gOpO1G.zOOOP1G.z1G.zO&oO!bO7+$ZO&}O!bO7+$ZOOOP7+$Z7+$ZOOOP7+$c7+$cO'YOpO,59VO'bOpO,59VO'mO!bO,59eOOOO-E6w-E6wO'{OpO1G.uO'{OpO1G.uOOOP1G.u1G.uO(TOpO7+$fOOOP7+$f7+$fO(]O!bO<<GuOOOP<<Gu<<GuOOOP<<G}<<G}O'bOpO1G.qO'bOpO1G.qO(hO#tO'#CnO(vO&jO'#CnOOOO1G.q1G.qO)UOpO7+$aOOOP7+$a7+$aOOOP<<HQ<<HQOOOPAN=aAN=aOOOPAN=iAN=iO'bOpO7+$]OOOO7+$]7+$]OOOO'#Cz'#CzO)^O#tO,59YOOOO,59Y,59YOOOO'#C{'#C{O)lO&jO,59YOOOP<<G{<<G{OOOO<<Gw<<GwOOOO-E6x-E6xOOOO1G.t1G.tOOOO-E6y-E6y",stateData:")z~OPQOSVOTWOVWOWWOXWOiXOyPO!QTO!SUO~OvZOx]O~O^`Oz^O~OPQOQcOSVOTWOVWOWWOXWOyPO!QTO!SUO~ORdO~P!SOteO!PgO~OuhO!RjO~O^lOz^O~OvZOxoO~O^qOz^O~O[vO`sOdwOz^O~ORyO~P!SO^{Oz^O~OteO!P}O~OuhO!R!PO~O^!QOz^O~O[!SOz^O~O[!VO`sOd!WOz^O~Oa!YOz^O~Oz^O[mX`mXdmX~O[!VO`sOd!WO~O^!]Oz^O~O[!_Oz^O~O[!aOz^O~O[!cO`sOd!dOz^O~O[!cO`sOd!dO~Oa!eOz^O~Oz^O{!gO}!hO~Oz^O[ma`madma~O[!kOz^O~O[!lOz^O~O[!mO`sOd!nO~OW!qOX!qO{!sO|!qO~OW!tOX!tO}!sO!O!tO~O[!vOz^O~OW!qOX!qO{!yO|!qO~OW!tOX!tO}!yO!O!tO~O",goto:"%cxPPPPPPPPPPyyP!PP!VPP!`!jP!pyyyP!v!|#S$[$k$q$w$}%TPPPP%ZXWORYbXRORYb_t`qru!T!U!bQ!i!YS!p!e!fR!w!oQdRRybXSORYbQYORmYQ[PRn[Q_QQkVjp_krz!R!T!X!Z!^!`!f!j!oQr`QzcQ!RlQ!TqQ!XsQ!ZtQ!^{Q!`!QQ!f!YQ!j!]R!o!eQu`S!UqrU![u!U!bR!b!TQ!r!gR!x!rQ!u!hR!z!uQbRRxbQfTR|fQiUR!OiSXOYTaRb",nodeNames:"⚠ StartTag StartCloseTag MissingCloseTag StartCloseTag StartCloseTag Document Text EntityReference CharacterReference Cdata Element EndTag OpenTag TagName Attribute AttributeName Is AttributeValue CloseTag SelfCloseEndTag SelfClosingTag Comment ProcessingInst MismatchedCloseTag DoctypeDecl",maxTerm:50,context:Zh,nodeProps:[["closedBy",1,"SelfCloseEndTag EndTag",13,"CloseTag MissingCloseTag"],["openedBy",12,"StartTag StartCloseTag",19,"OpenTag",20,"StartTag"],["isolate",-6,13,18,19,21,22,24,""]],propSources:[tf],skippedNodes:[0],repeatNodeCount:9,tokenData:"!)v~R!YOX$qXY)iYZ)iZ]$q]^)i^p$qpq)iqr$qrs*vsv$qvw+fwx/ix}$q}!O0[!O!P$q!P!Q2z!Q![$q![!]4n!]!^$q!^!_8U!_!`!#t!`!a!$l!a!b!%d!b!c$q!c!}4n!}#P$q#P#Q!'W#Q#R$q#R#S4n#S#T$q#T#o4n#o%W$q%W%o4n%o%p$q%p&a4n&a&b$q&b1p4n1p4U$q4U4d4n4d4e$q4e$IS4n$IS$I`$q$I`$Ib4n$Ib$Kh$q$Kh%#t4n%#t&/x$q&/x&Et4n&Et&FV$q&FV;'S4n;'S;:j8O;:j;=`)c<%l?&r$q?&r?Ah4n?Ah?BY$q?BY?Mn4n?MnO$qi$zXVP|W!O`Or$qrs%gsv$qwx'^x!^$q!^!_(o!_;'S$q;'S;=`)c<%lO$qa%nVVP!O`Ov%gwx&Tx!^%g!^!_&o!_;'S%g;'S;=`'W<%lO%gP&YTVPOv&Tw!^&T!_;'S&T;'S;=`&i<%lO&TP&lP;=`<%l&T`&tS!O`Ov&ox;'S&o;'S;=`'Q<%lO&o`'TP;=`<%l&oa'ZP;=`<%l%gX'eWVP|WOr'^rs&Tsv'^w!^'^!^!_'}!_;'S'^;'S;=`(i<%lO'^W(ST|WOr'}sv'}w;'S'};'S;=`(c<%lO'}W(fP;=`<%l'}X(lP;=`<%l'^h(vV|W!O`Or(ors&osv(owx'}x;'S(o;'S;=`)]<%lO(oh)`P;=`<%l(oi)fP;=`<%l$qo)t`VP|W!O`zUOX$qXY)iYZ)iZ]$q]^)i^p$qpq)iqr$qrs%gsv$qwx'^x!^$q!^!_(o!_;'S$q;'S;=`)c<%lO$qk+PV{YVP!O`Ov%gwx&Tx!^%g!^!_&o!_;'S%g;'S;=`'W<%lO%g~+iast,n![!]-r!c!}-r#R#S-r#T#o-r%W%o-r%p&a-r&b1p-r4U4d-r4e$IS-r$I`$Ib-r$Kh%#t-r&/x&Et-r&FV;'S-r;'S;:j/c?&r?Ah-r?BY?Mn-r~,qQ!Q![,w#l#m-V~,zQ!Q![,w!]!^-Q~-VOX~~-YR!Q![-c!c!i-c#T#Z-c~-fS!Q![-c!]!^-Q!c!i-c#T#Z-c~-ug}!O-r!O!P-r!Q![-r![!]-r!]!^/^!c!}-r#R#S-r#T#o-r$}%O-r%W%o-r%p&a-r&b1p-r1p4U-r4U4d-r4e$IS-r$I`$Ib-r$Je$Jg-r$Kh%#t-r&/x&Et-r&FV;'S-r;'S;:j/c?&r?Ah-r?BY?Mn-r~/cOW~~/fP;=`<%l-rk/rW}bVP|WOr'^rs&Tsv'^w!^'^!^!_'}!_;'S'^;'S;=`(i<%lO'^k0eZVP|W!O`Or$qrs%gsv$qwx'^x}$q}!O1W!O!^$q!^!_(o!_;'S$q;'S;=`)c<%lO$qk1aZVP|W!O`Or$qrs%gsv$qwx'^x!^$q!^!_(o!_!`$q!`!a2S!a;'S$q;'S;=`)c<%lO$qk2_X!PQVP|W!O`Or$qrs%gsv$qwx'^x!^$q!^!_(o!_;'S$q;'S;=`)c<%lO$qm3TZVP|W!O`Or$qrs%gsv$qwx'^x!^$q!^!_(o!_!`$q!`!a3v!a;'S$q;'S;=`)c<%lO$qm4RXdSVP|W!O`Or$qrs%gsv$qwx'^x!^$q!^!_(o!_;'S$q;'S;=`)c<%lO$qo4{!P`S^QVP|W!O`Or$qrs%gsv$qwx'^x}$q}!O4n!O!P4n!P!Q$q!Q![4n![!]4n!]!^$q!^!_(o!_!c$q!c!}4n!}#R$q#R#S4n#S#T$q#T#o4n#o$}$q$}%O4n%O%W$q%W%o4n%o%p$q%p&a4n&a&b$q&b1p4n1p4U4n4U4d4n4d4e$q4e$IS4n$IS$I`$q$I`$Ib4n$Ib$Je$q$Je$Jg4n$Jg$Kh$q$Kh%#t4n%#t&/x$q&/x&Et4n&Et&FV$q&FV;'S4n;'S;:j8O;:j;=`)c<%l?&r$q?&r?Ah4n?Ah?BY$q?BY?Mn4n?MnO$qo8RP;=`<%l4ni8]Y|W!O`Oq(oqr8{rs&osv(owx'}x!a(o!a!b!#U!b;'S(o;'S;=`)]<%lO(oi9S_|W!O`Or(ors&osv(owx'}x}(o}!O:R!O!f(o!f!g;e!g!}(o!}#ODh#O#W(o#W#XLp#X;'S(o;'S;=`)]<%lO(oi:YX|W!O`Or(ors&osv(owx'}x}(o}!O:u!O;'S(o;'S;=`)]<%lO(oi;OV!QP|W!O`Or(ors&osv(owx'}x;'S(o;'S;=`)]<%lO(oi;lX|W!O`Or(ors&osv(owx'}x!q(o!q!r<X!r;'S(o;'S;=`)]<%lO(oi<`X|W!O`Or(ors&osv(owx'}x!e(o!e!f<{!f;'S(o;'S;=`)]<%lO(oi=SX|W!O`Or(ors&osv(owx'}x!v(o!v!w=o!w;'S(o;'S;=`)]<%lO(oi=vX|W!O`Or(ors&osv(owx'}x!{(o!{!|>c!|;'S(o;'S;=`)]<%lO(oi>jX|W!O`Or(ors&osv(owx'}x!r(o!r!s?V!s;'S(o;'S;=`)]<%lO(oi?^X|W!O`Or(ors&osv(owx'}x!g(o!g!h?y!h;'S(o;'S;=`)]<%lO(oi@QY|W!O`Or?yrs@psv?yvwA[wxBdx!`?y!`!aCr!a;'S?y;'S;=`Db<%lO?ya@uV!O`Ov@pvxA[x!`@p!`!aAy!a;'S@p;'S;=`B^<%lO@pPA_TO!`A[!`!aAn!a;'SA[;'S;=`As<%lOA[PAsOiPPAvP;=`<%lA[aBQSiP!O`Ov&ox;'S&o;'S;=`'Q<%lO&oaBaP;=`<%l@pXBiX|WOrBdrsA[svBdvwA[w!`Bd!`!aCU!a;'SBd;'S;=`Cl<%lOBdXC]TiP|WOr'}sv'}w;'S'};'S;=`(c<%lO'}XCoP;=`<%lBdiC{ViP|W!O`Or(ors&osv(owx'}x;'S(o;'S;=`)]<%lO(oiDeP;=`<%l?yiDoZ|W!O`Or(ors&osv(owx'}x!e(o!e!fEb!f#V(o#V#WIr#W;'S(o;'S;=`)]<%lO(oiEiX|W!O`Or(ors&osv(owx'}x!f(o!f!gFU!g;'S(o;'S;=`)]<%lO(oiF]X|W!O`Or(ors&osv(owx'}x!c(o!c!dFx!d;'S(o;'S;=`)]<%lO(oiGPX|W!O`Or(ors&osv(owx'}x!v(o!v!wGl!w;'S(o;'S;=`)]<%lO(oiGsX|W!O`Or(ors&osv(owx'}x!c(o!c!dH`!d;'S(o;'S;=`)]<%lO(oiHgX|W!O`Or(ors&osv(owx'}x!}(o!}#OIS#O;'S(o;'S;=`)]<%lO(oiI]V|W!O`yPOr(ors&osv(owx'}x;'S(o;'S;=`)]<%lO(oiIyX|W!O`Or(ors&osv(owx'}x#W(o#W#XJf#X;'S(o;'S;=`)]<%lO(oiJmX|W!O`Or(ors&osv(owx'}x#T(o#T#UKY#U;'S(o;'S;=`)]<%lO(oiKaX|W!O`Or(ors&osv(owx'}x#h(o#h#iK|#i;'S(o;'S;=`)]<%lO(oiLTX|W!O`Or(ors&osv(owx'}x#T(o#T#UH`#U;'S(o;'S;=`)]<%lO(oiLwX|W!O`Or(ors&osv(owx'}x#c(o#c#dMd#d;'S(o;'S;=`)]<%lO(oiMkX|W!O`Or(ors&osv(owx'}x#V(o#V#WNW#W;'S(o;'S;=`)]<%lO(oiN_X|W!O`Or(ors&osv(owx'}x#h(o#h#iNz#i;'S(o;'S;=`)]<%lO(oi! RX|W!O`Or(ors&osv(owx'}x#m(o#m#n! n#n;'S(o;'S;=`)]<%lO(oi! uX|W!O`Or(ors&osv(owx'}x#d(o#d#e!!b#e;'S(o;'S;=`)]<%lO(oi!!iX|W!O`Or(ors&osv(owx'}x#X(o#X#Y?y#Y;'S(o;'S;=`)]<%lO(oi!#_V!SP|W!O`Or(ors&osv(owx'}x;'S(o;'S;=`)]<%lO(ok!$PXaQVP|W!O`Or$qrs%gsv$qwx'^x!^$q!^!_(o!_;'S$q;'S;=`)c<%lO$qo!$wX[UVP|W!O`Or$qrs%gsv$qwx'^x!^$q!^!_(o!_;'S$q;'S;=`)c<%lO$qk!%mZVP|W!O`Or$qrs%gsv$qwx'^x!^$q!^!_(o!_!`$q!`!a!&`!a;'S$q;'S;=`)c<%lO$qk!&kX!RQVP|W!O`Or$qrs%gsv$qwx'^x!^$q!^!_(o!_;'S$q;'S;=`)c<%lO$qk!'aZVP|W!O`Or$qrs%gsv$qwx'^x!^$q!^!_(o!_#P$q#P#Q!(S#Q;'S$q;'S;=`)c<%lO$qk!(]ZVP|W!O`Or$qrs%gsv$qwx'^x!^$q!^!_(o!_!`$q!`!a!)O!a;'S$q;'S;=`)c<%lO$qk!)ZXxQVP|W!O`Or$qrs%gsv$qwx'^x!^$q!^!_(o!_;'S$q;'S;=`)c<%lO$q",tokenizers:[Gh,Yh,Jh,ef,0,1,2,3,4],topRules:{Document:[0,6]},tokenPrec:0});function rf(e,t){let n=t&&t.getChild("TagName");return n?e.sliceString(n.from,n.to):""}function af(e,t){let n=t&&t.firstChild;return n&&"OpenTag"==n.name?rf(e,n):""}function of(e){for(let t=e&&e.parent;t;t=t.parent)if("Element"==t.name)return t;return null}class sf{constructor(e,t,n){this.attrs=t,this.attrValues=n,this.children=[],this.name=e.name,this.completion=Object.assign(Object.assign({type:"type"},e.completion||{}),{label:this.name}),this.openCompletion=Object.assign(Object.assign({},this.completion),{label:"<"+this.name}),this.closeCompletion=Object.assign(Object.assign({},this.completion),{label:"</"+this.name+">",boost:2}),this.closeNameCompletion=Object.assign(Object.assign({},this.completion),{label:this.name+">"}),this.text=e.textContent?e.textContent.map((e=>({label:e,type:"text"}))):[]}}const lf=/^[:\-\.\w\u00b7-\uffff]*$/;function cf(e){return Object.assign(Object.assign({type:"property"},e.completion||{}),{label:e.name})}function uf(e){return"string"==typeof e?{label:`"${e}"`,type:"constant"}:/^"/.test(e.label)?e:Object.assign(Object.assign({},e),{label:`"${e.label}"`})}function df(e,t){let n=[],r=[],a=Object.create(null);for(let e of t){let t=cf(e);n.push(t),e.global&&r.push(t),e.values&&(a[e.name]=e.values.map(uf))}let o=[],i=[],s=Object.create(null);for(let t of e){let e=r,l=a;t.attributes&&(e=e.concat(t.attributes.map((e=>"string"==typeof e?n.find((t=>t.label==e))||{label:e,type:"property"}:(e.values&&(l==a&&(l=Object.create(l)),l[e.name]=e.values.map(uf)),cf(e))))));let c=new sf(t,e,l);s[c.name]=c,o.push(c),t.top&&i.push(c)}i.length||(i=o);for(let t=0;t<o.length;t++){let n=e[t],r=o[t];if(n.children)for(let e of n.children)s[e]&&r.children.push(s[e]);else r.children=o}return e=>{var t;let{doc:n}=e.state,l=function(e,t){var n;let r=Qs(e).resolveInner(t,-1),a=null;for(let e=r;!a&&e.parent;e=e.parent)"OpenTag"!=e.name&&"CloseTag"!=e.name&&"SelfClosingTag"!=e.name&&"MismatchedCloseTag"!=e.name||(a=e);if(a&&(a.to>t||a.lastChild.type.isError)){let e=a.parent;if("TagName"==r.name)return"CloseTag"==a.name||"MismatchedCloseTag"==a.name?{type:"closeTag",from:r.from,context:e}:{type:"openTag",from:r.from,context:of(e)};if("AttributeName"==r.name)return{type:"attrName",from:r.from,context:a};if("AttributeValue"==r.name)return{type:"attrValue",from:r.from,context:a};let n=r==a||"Attribute"==r.name?r.childBefore(t):r;return"StartTag"==(null==n?void 0:n.name)?{type:"openTag",from:t,context:of(e)}:"StartCloseTag"==(null==n?void 0:n.name)&&n.to<=t?{type:"closeTag",from:t,context:e}:"Is"==(null==n?void 0:n.name)?{type:"attrValue",from:t,context:a}:n?{type:"attrName",from:t,context:a}:null}if("StartCloseTag"==r.name)return{type:"closeTag",from:t,context:r.parent};for(;r.parent&&r.to==t&&!(null===(n=r.lastChild)||void 0===n?void 0:n.type.isError);)r=r.parent;return"Element"==r.name||"Text"==r.name||"Document"==r.name?{type:"tag",from:t,context:"Element"==r.name?r:of(r)}:null}(e.state,e.pos);if(!l||"tag"==l.type&&!e.explicit)return null;let{type:c,from:u,context:d}=l;if("openTag"==c){let e=i,t=af(n,d);if(t){let n=s[t];e=(null==n?void 0:n.children)||o}return{from:u,options:e.map((e=>e.completion)),validFor:lf}}if("closeTag"==c){let r=af(n,d);return r?{from:u,to:e.pos+(">"==n.sliceString(e.pos,e.pos+1)?1:0),options:[(null===(t=s[r])||void 0===t?void 0:t.closeNameCompletion)||{label:r+">",type:"type"}],validFor:lf}:null}if("attrName"==c){let e=s[rf(n,d)];return{from:u,options:(null==e?void 0:e.attrs)||r,validFor:lf}}if("attrValue"==c){let t=function(e,t,n){let r=t&&t.getChildren("Attribute").find((e=>e.from<=n&&e.to>=n)),a=r&&r.getChild("AttributeName");return a?e.sliceString(a.from,a.to):""}(n,d,u);if(!t)return null;let r=s[rf(n,d)],o=((null==r?void 0:r.attrValues)||a)[t];return o&&o.length?{from:u,to:e.pos+('"'==n.sliceString(e.pos,e.pos+1)?1:0),options:o,validFor:/^"[^"]*"?$/}:null}if("tag"==c){let t=af(n,d),r=s[t],a=[],l=d&&d.lastChild;!t||l&&"CloseTag"==l.name&&rf(n,l)==t||a.push(r?r.closeCompletion:{label:"</"+t+">",type:"type",boost:2});let c=a.concat(((null==r?void 0:r.children)||(d?o:i)).map((e=>e.openCompletion)));if(d&&(null==r?void 0:r.text.length)){let t=d.firstChild;t.to>e.pos-20&&!/\S/.test(e.state.sliceDoc(t.to,e.pos))&&(c=c.concat(r.text))}return{from:u,options:c,validFor:/^<\/?[:\-\.\w\u00b7-\uffff]*$/}}return null}}const pf=Fs.define({name:"xml",parser:nf.configure({props:[rl.add({Element(e){let t=/^\s*<\//.test(e.textAfter);return e.lineIndent(e.node.from)+(t?0:e.unit)},"OpenTag CloseTag SelfClosingTag":e=>e.column(e.node.from)+e.unit}),hl.add({Element(e){let t=e.firstChild,n=e.lastChild;return t&&"OpenTag"==t.name?{from:t.to,to:"CloseTag"==n.name?n.from:e.to}:null}}),Gl.add({"OpenTag CloseTag":e=>e.getChild("TagName")})]}),languageData:{commentTokens:{block:{open:"\x3c!--",close:"--\x3e"}},indentOnInput:/^\s*<\/$/}});function hf(e,t,n=e.length){if(!t)return"";let r=t.firstChild,a=r&&r.getChild("TagName");return a?e.sliceString(a.from,Math.min(a.to,n)):""}const ff=Ya.inputHandler.of(((e,t,n,r,a)=>{if(e.composing||e.state.readOnly||t!=n||">"!=r&&"/"!=r||!pf.isActiveAt(e.state,t,-1))return!1;let o=a(),{state:i}=o,s=i.changeByRange((e=>{var t,n,a;let o,{head:s}=e,l=i.doc.sliceString(s-1,s)==r,c=Qs(i).resolveInner(s,-1);if(l&&">"==r&&"EndTag"==c.name){let r=c.parent;if("CloseTag"!=(null===(n=null===(t=r.parent)||void 0===t?void 0:t.lastChild)||void 0===n?void 0:n.name)&&(o=hf(i.doc,r.parent,s)))return{range:e,changes:{from:s,to:s+(">"===i.doc.sliceString(s,s+1)?1:0),insert:`</${o}>`}}}else if(l&&"/"==r&&"StartCloseTag"==c.name){let e=c.parent;if(c.from==s-2&&"CloseTag"!=(null===(a=e.lastChild)||void 0===a?void 0:a.name)&&(o=hf(i.doc,e,s))){let e=s+(">"===i.doc.sliceString(s,s+1)?1:0),t=`${o}>`;return{range:F.cursor(s+t.length,-1),changes:{from:s,to:e,insert:t}}}}return{range:e}}));return!s.changes.empty&&(e.dispatch([o,i.update(s,{userEvent:"input.complete",scrollIntoView:!0})]),!0)})),mf=63;class gf{constructor(e,t,n){this.parent=e,this.depth=t,this.type=n,this.hash=(e?e.hash+e.hash<<8:0)+t+(t<<4)+n}}function vf(e,t){for(let n=0,r=t-e.pos-1;;r--,n++){let t=e.peek(r);if(yf(t)||-1==t)return n}}function bf(e){return 32==e||9==e}function yf(e){return 10==e||13==e}function wf(e){return bf(e)||yf(e)}function Of(e){return e<0||wf(e)}gf.top=new gf(null,-1,0);const kf=new Rd({start:gf.top,reduce:(e,t)=>3!=e.type||20!=t&&34!=t?e:e.parent,shift(e,t,n,r){if(3==t)return new gf(e,vf(r,r.pos),1);if(65==t||5==t)return new gf(e,vf(r,r.pos),2);if(t==mf)return e.parent;if(19==t||33==t)return new gf(e,0,3);if(13==t&&4==e.type)return e.parent;if(47==t){let t=/[1-9]/.exec(r.read(r.pos,n.pos));if(t)return new gf(e,e.depth+ +t[0],4)}return e},hash:e=>e.hash});function xf(e,t,n=0){return e.peek(n)==t&&e.peek(n+1)==t&&e.peek(n+2)==t&&Of(e.peek(n+3))}const Ef=new vd(((e,t)=>{if(-1==e.next&&t.canShift(64))return e.acceptToken(64);let n=e.peek(-1);if((yf(n)||n<0)&&3!=t.context.type){if(xf(e,45)){if(!t.canShift(mf))return e.acceptToken(1,3);e.acceptToken(mf)}if(xf(e,46)){if(!t.canShift(mf))return e.acceptToken(2,3);e.acceptToken(mf)}let n=0;for(;32==e.next;)n++,e.advance();!(n<t.context.depth)&&(n!=t.context.depth||1!=t.context.type||45==e.next&&Of(e.peek(1)))||-1==e.next||yf(e.next)||35==e.next||e.acceptToken(mf,-n)}}),{contextual:!0}),Sf=new vd(((e,t)=>{if(3!=t.context.type)if(45==e.next)e.advance(),Of(e.next)&&e.acceptToken(1==t.context.type&&t.context.depth==vf(e,e.pos-1)?4:3);else if(63==e.next)e.advance(),Of(e.next)&&e.acceptToken(2==t.context.type&&t.context.depth==vf(e,e.pos-1)?6:5);else{let n=e.pos;for(;;)if(bf(e.next)){if(e.pos==n)return;e.advance()}else if(33==e.next)Tf(e);else{if(38!=e.next){if(42==e.next){Af(e);break}if(39==e.next||34==e.next){if(Rf(e,!0))break;return}if(91==e.next||123==e.next){if(!Pf(e))return;break}Nf(e,!0,!1,0);break}Af(e)}for(;bf(e.next);)e.advance();if(58==e.next){if(e.pos==n&&t.canShift(29))return;Of(e.peek(1))&&e.acceptTokenTo(2==t.context.type&&t.context.depth==vf(e,n)?66:65,n)}}else 63==e.next&&(e.advance(),Of(e.next)&&e.acceptToken(7))}),{contextual:!0});function _f(e){return e>=48&&e<=57||e>=97&&e<=102||e>=65&&e<=70}function Cf(e,t){return 37==e.next?(e.advance(),_f(e.next)&&e.advance(),_f(e.next)&&e.advance(),!0):!!((n=e.next)>32&&n<127&&34!=n&&37!=n&&44!=n&&60!=n&&62!=n&&92!=n&&94!=n&&96!=n&&123!=n&&124!=n&&125!=n||t&&44==e.next)&&(e.advance(),!0);var n}function Tf(e){if(e.advance(),60==e.next){for(e.advance();;)if(!Cf(e,!0)){62==e.next&&e.advance();break}}else for(;Cf(e,!1););}function Af(e){for(e.advance();!Of(e.next)&&"f"!=Mf(e.tag);)e.advance()}function Rf(e,t){let n=e.next,r=!1,a=e.pos;for(e.advance();;){let o=e.next;if(o<0)break;if(e.advance(),o==n){if(39!=o)break;if(39!=e.next)break;e.advance()}else if(92==o&&34==n)e.next>=0&&e.advance();else if(yf(o)){if(t)return!1;r=!0}else if(t&&e.pos>=a+1024)return!1}return!r}function Pf(e){for(let t=[],n=e.pos+1024;;)if(91==e.next||123==e.next)t.push(e.next),e.advance();else if(39==e.next||34==e.next){if(!Rf(e,!0))return!1}else if(93==e.next||125==e.next){if(t[t.length-1]!=e.next-2)return!1;if(t.pop(),e.advance(),!t.length)return!0}else{if(e.next<0||e.pos>n||yf(e.next))return!1;e.advance()}}const Df="iiisiiissisfissssssssssssisssiiissssssssssssssssssssssssssfsfssissssssssssssssssssssssssssfif";function Mf(e){return e<33?"u":e>125?"s":Df[e-33]}function $f(e,t){let n=Mf(e);return"u"!=n&&!(t&&"f"==n)}function Nf(e,t,n,r){if("s"!=Mf(e.next)&&(63!=e.next&&58!=e.next&&45!=e.next||!$f(e.peek(1),n)))return!1;e.advance();let a=e.pos;for(;;){let o=e.next,i=0,s=r+1;for(;wf(o);){if(yf(o)){if(t)return!1;s=0}else s++;o=e.peek(++i)}if(!(o>=0&&(58==o?$f(e.peek(i+1),n):35==o?32!=e.peek(i-1):$f(o,n)))||!n&&s<=r||0==s&&!n&&(xf(e,45,i)||xf(e,46,i)))break;if(t&&"f"==Mf(o))return!1;for(let t=i;t>=0;t--)e.advance();if(t&&e.pos>a+1024)return!1}return!0}const If=new vd(((e,t)=>{if(33==e.next)Tf(e),e.acceptToken(12);else if(38==e.next||42==e.next){let t=38==e.next?10:11;Af(e),e.acceptToken(t)}else 39==e.next||34==e.next?(Rf(e,!1),e.acceptToken(9)):Nf(e,!1,3==t.context.type,t.context.depth)&&e.acceptToken(8)})),Lf=new vd(((e,t)=>{let n=4==t.context.type?t.context.depth:-1,r=e.pos;e:for(;;){let a=0,o=e.next;for(;32==o;)o=e.peek(++a);if(!a&&(xf(e,45,a)||xf(e,46,a)))break;if(!yf(o)&&(n<0&&(n=Math.max(t.context.depth+1,a)),a<n))break;for(;;){if(e.next<0)break e;let t=yf(e.next);if(e.advance(),t)continue e;r=e.pos}}e.acceptTokenTo(13,r)})),Bf=ds({DirectiveName:Ds.keyword,DirectiveContent:Ds.attributeValue,"DirectiveEnd DocEnd":Ds.meta,QuotedLiteral:Ds.string,BlockLiteralHeader:Ds.special(Ds.string),BlockLiteralContent:Ds.content,Literal:Ds.content,"Key/Literal Key/QuotedLiteral":Ds.definition(Ds.propertyName),"Anchor Alias":Ds.labelName,Tag:Ds.typeName,Comment:Ds.lineComment,": , -":Ds.separator,"?":Ds.punctuation,"[ ]":Ds.squareBracket,"{ }":Ds.brace}),Ff=Pd.deserialize({version:14,states:"5lQ!ZQgOOO#PQfO'#CpO#uQfO'#DOOOQR'#Dv'#DvO$qQgO'#DRO%gQdO'#DUO%nQgO'#DUO&ROaO'#D[OOQR'#Du'#DuO&{QgO'#D^O'rQgO'#D`OOQR'#Dt'#DtO(iOqO'#DbOOQP'#Dj'#DjO(zQaO'#CmO)YQgO'#CmOOQP'#Cm'#CmQ)jQaOOQ)uQgOOQ]QgOOO*PQdO'#CrO*nQdO'#CtOOQO'#Dw'#DwO+]Q`O'#CxO+hQdO'#CwO+rQ`O'#CwOOQO'#Cv'#CvO+wQdO'#CvOOQO'#Cq'#CqO,UQ`O,59[O,^QfO,59[OOQR,59[,59[OOQO'#Cx'#CxO,eQ`O'#DPO,pQdO'#DPOOQO'#Dx'#DxO,zQdO'#DxO-XQ`O,59jO-aQfO,59jOOQR,59j,59jOOQR'#DS'#DSO-hQcO,59mO-sQgO'#DVO.TQ`O'#DVO.YQcO,59pOOQR'#DX'#DXO#|QfO'#DWO.hQcO'#DWOOQR,59v,59vO.yOWO,59vO/OOaO,59vO/WOaO,59vO/cQgO'#D_OOQR,59x,59xO0VQgO'#DaOOQR,59z,59zOOQP,59|,59|O0yOaO,59|O1ROaO,59|O1aOqO,59|OOQP-E7h-E7hO1oQgO,59XOOQP,59X,59XO2PQaO'#DeO2_QgO'#DeO2oQgO'#DkOOQP'#Dk'#DkQ)jQaOOO3PQdO'#CsOOQO,59^,59^O3kQdO'#CuOOQO,59`,59`OOQO,59c,59cO4VQdO,59cO4aQdO'#CzO4kQ`O'#CzOOQO,59b,59bOOQU,5:Q,5:QOOQR1G.v1G.vO4pQ`O1G.vOOQU-E7d-E7dO4xQdO,59kOOQO,59k,59kO5SQdO'#DQO5^Q`O'#DQOOQO,5:d,5:dOOQU,5:R,5:ROOQR1G/U1G/UO5cQ`O1G/UOOQU-E7e-E7eO5kQgO'#DhO5xQcO1G/XOOQR1G/X1G/XOOQR,59q,59qO6TQgO,59qO6eQdO'#DiO6lQgO'#DiO7PQcO1G/[OOQR1G/[1G/[OOQR,59r,59rO#|QfO,59rOOQR1G/b1G/bO7_OWO1G/bO7dOaO1G/bOOQR,59y,59yOOQR,59{,59{OOQP1G/h1G/hO7lOaO1G/hO7tOaO1G/hO8POaO1G/hOOQP1G.s1G.sO8_QgO,5:POOQP,5:P,5:POOQP,5:V,5:VOOQP-E7i-E7iOOQO,59_,59_OOQO,59a,59aOOQO1G.}1G.}OOQO,59f,59fO8oQdO,59fOOQR7+$b7+$bP,XQ`O'#DfOOQO1G/V1G/VOOQO,59l,59lO8yQdO,59lOOQR7+$p7+$pP9TQ`O'#DgOOQR'#DT'#DTOOQR,5:S,5:SOOQR-E7f-E7fOOQR7+$s7+$sOOQR1G/]1G/]O9YQgO'#DYO9jQ`O'#DYOOQR,5:T,5:TO#|QfO'#DZO9oQcO'#DZOOQR-E7g-E7gOOQR7+$v7+$vOOQR1G/^1G/^OOQR7+$|7+$|O:QOWO7+$|OOQP7+%S7+%SO:VOaO7+%SO:_OaO7+%SOOQP1G/k1G/kOOQO1G/Q1G/QOOQO1G/W1G/WOOQR,59t,59tO:jQgO,59tOOQR,59u,59uO#|QfO,59uOOQR<<Hh<<HhOOQP<<Hn<<HnO:zOaO<<HnOOQR1G/`1G/`OOQR1G/a1G/aOOQPAN>YAN>Y",stateData:";S~O!fOS!gOS^OS~OP_OQbORSOTUOWROXROYYOZZO[XOcPOqQO!PVO!V[O!cTO~O`cO~P]OVkOWROXROYeOZfO[dOcPOmhOqQO~OboO~P!bOVtOWROXROYeOZfO[dOcPOmrOqQO~OpwO~P#WORSOTUOWROXROYYOZZO[XOcPOqQO!PVO!cTO~OSvP!avP!bvP~P#|OWROXROYeOZfO[dOcPOqQO~OmzO~P%OOm!OOUzP!azP!bzP!dzP~P#|O^!SO!b!QO!f!TO!g!RO~ORSOTUOWROXROcPOqQO!PVO!cTO~OY!UOP!QXQ!QX!V!QX!`!QXS!QX!a!QX!b!QXU!QXm!QX!d!QX~P&aO[!WOP!SXQ!SX!V!SX!`!SXS!SX!a!SX!b!SXU!SXm!SX!d!SX~P&aO^!ZO!W![O!b!YO!f!]O!g!YO~OP!_O!V[OQaX!`aX~OPaXQaX!VaX!`aX~P#|OP!bOQ!cO!V[O~OP_O!V[O~P#|OWROXROY!fOcPOqQObfXmfXofXpfX~OWROXRO[!hOcPOqQObhXmhXohXphX~ObeXmlXoeX~ObkXokX~P%OOm!kO~Om!lObnPonP~P%OOb!pOo!oO~Ob!pO~P!bOm!sOosXpsX~OosXpsX~P%OOm!uOotPptP~P%OOo!xOp!yO~Op!yO~P#WOS!|O!a#OO!b#OO~OUyX!ayX!byX!dyX~P#|Om#QO~OU#SO!a#UO!b#UO!d#RO~Om#WOUzX!azX!bzX!dzX~O]#XO~O!b#XO!g#YO~O^#ZO!b#XO!g#YO~OP!RXQ!RX!V!RX!`!RXS!RX!a!RX!b!RXU!RXm!RX!d!RX~P&aOP!TXQ!TX!V!TX!`!TXS!TX!a!TX!b!TXU!TXm!TX!d!TX~P&aO!b#^O!g#^O~O^#_O!b#^O!f#`O!g#^O~O^#_O!W#aO!b#^O!g#^O~OPaaQaa!Vaa!`aa~P#|OP#cO!V[OQ!XX!`!XX~OP!XXQ!XX!V!XX!`!XX~P#|OP_O!V[OQ!_X!`!_X~P#|OWROXROcPOqQObgXmgXogXpgX~OWROXROcPOqQObiXmiXoiXpiX~Obkaoka~P%OObnXonX~P%OOm#kO~Ob#lOo!oO~Oosapsa~P%OOotXptX~P%OOm#pO~Oo!xOp#qO~OSwP!awP!bwP~P#|OS!|O!a#vO!b#vO~OUya!aya!bya!dya~P#|Om#xO~P%OOm#{OU}P!a}P!b}P!d}P~P#|OU#SO!a$OO!b$OO!d#RO~O]$QO~O!b$QO!g$RO~O!b$SO!g$SO~O^$TO!b$SO!g$SO~O^$TO!b$SO!f$UO!g$SO~OP!XaQ!Xa!V!Xa!`!Xa~P#|Obnaona~P%OOotapta~P%OOo!xO~OU|X!a|X!b|X!d|X~P#|Om$ZO~Om$]OU}X!a}X!b}X!d}X~O]$^O~O!b$_O!g$_O~O^$`O!b$_O!g$_O~OU|a!a|a!b|a!d|a~P#|O!b$cO!g$cO~O",goto:",]!mPPPPPPPPPPPPPPPPP!nPP!v#v#|$`#|$c$f$j$nP%VPPP!v%Y%^%a%{&O%a&R&U&X&_&b%aP&e&{&e'O'RPP']'a'g'm's'y(XPPPPPPPP(_)e*X+c,VUaObcR#e!c!{ROPQSTUXY_bcdehknrtvz!O!U!W!_!b!c!f!h!k!l!s!u!|#Q#R#S#W#c#k#p#x#{$Z$]QmPR!qnqfPQThknrtv!k!l!s!u#R#k#pR!gdR!ieTlPnTjPnSiPnSqQvQ{TQ!mkQ!trQ!vtR#y#RR!nkTsQvR!wt!RWOSUXY_bcz!O!U!W!_!b!c!|#Q#S#W#c#x#{$Z$]RySR#t!|R|TR|UQ!PUR#|#SR#z#RR#z#SyZOSU_bcz!O!_!b!c!|#Q#S#W#c#x#{$Z$]R!VXR!XYa]O^abc!a!c!eT!da!eQnPR!rnQvQR!{vQ!}yR#u!}Q#T|R#}#TW^Obc!cS!^^!aT!aa!eQ!eaR#f!eW`Obc!cQxSS}U#SQ!`_Q#PzQ#V!OQ#b!_Q#d!bQ#s!|Q#w#QQ$P#WQ$V#cQ$Y#xQ$[#{Q$a$ZR$b$]xZOSU_bcz!O!_!b!c!|#Q#S#W#c#x#{$Z$]Q!VXQ!XYQ#[!UR#]!W!QWOSUXY_bcz!O!U!W!_!b!c!|#Q#S#W#c#x#{$Z$]pfPQThknrtv!k!l!s!u#R#k#pQ!gdQ!ieQ#g!fR#h!hSgPn^pQTkrtv#RQ!jhQ#i!kQ#j!lQ#n!sQ#o!uQ$W#kR$X#pQuQR!zv",nodeNames:"⚠ DirectiveEnd DocEnd - - ? ? ? Literal QuotedLiteral Anchor Alias Tag BlockLiteralContent Comment Stream BOM Document ] [ FlowSequence Item Tagged Anchored Anchored Tagged FlowMapping Pair Key : Pair , } { FlowMapping Pair Pair BlockSequence Item Item BlockMapping Pair Pair Key Pair Pair BlockLiteral BlockLiteralHeader Tagged Anchored Anchored Tagged Directive DirectiveName DirectiveContent Document",maxTerm:74,context:kf,nodeProps:[["isolate",-3,8,9,14,""],["openedBy",18,"[",32,"{"],["closedBy",19,"]",33,"}"]],propSources:[Bf],skippedNodes:[0],repeatNodeCount:6,tokenData:"-Y~RnOX#PXY$QYZ$]Z]#P]^$]^p#Ppq$Qqs#Pst$btu#Puv$yv|#P|}&e}![#P![!]'O!]!`#P!`!a'i!a!}#P!}#O*g#O#P#P#P#Q+Q#Q#o#P#o#p+k#p#q'i#q#r,U#r;'S#P;'S;=`#z<%l?HT#P?HT?HU,o?HUO#PQ#UU!WQOY#PZp#Ppq#hq;'S#P;'S;=`#z<%lO#PQ#kTOY#PZs#Pt;'S#P;'S;=`#z<%lO#PQ#}P;=`<%l#P~$VQ!f~XY$Qpq$Q~$bO!g~~$gS^~OY$bZ;'S$b;'S;=`$s<%lO$b~$vP;=`<%l$bR%OX!WQOX%kXY#PZ]%k]^#P^p%kpq#hq;'S%k;'S;=`&_<%lO%kR%rX!WQ!VPOX%kXY#PZ]%k]^#P^p%kpq#hq;'S%k;'S;=`&_<%lO%kR&bP;=`<%l%kR&lUoP!WQOY#PZp#Ppq#hq;'S#P;'S;=`#z<%lO#PR'VUmP!WQOY#PZp#Ppq#hq;'S#P;'S;=`#z<%lO#PR'p[!PP!WQOY#PZp#Ppq#hq{#P{|(f|}#P}!O(f!O!R#P!R![)p![;'S#P;'S;=`#z<%lO#PR(mW!PP!WQOY#PZp#Ppq#hq!R#P!R![)V![;'S#P;'S;=`#z<%lO#PR)^U!PP!WQOY#PZp#Ppq#hq;'S#P;'S;=`#z<%lO#PR)wY!PP!WQOY#PZp#Ppq#hq{#P{|)V|}#P}!O)V!O;'S#P;'S;=`#z<%lO#PR*nUcP!WQOY#PZp#Ppq#hq;'S#P;'S;=`#z<%lO#PR+XUbP!WQOY#PZp#Ppq#hq;'S#P;'S;=`#z<%lO#PR+rUqP!WQOY#PZp#Ppq#hq;'S#P;'S;=`#z<%lO#PR,]UpP!WQOY#PZp#Ppq#hq;'S#P;'S;=`#z<%lO#PR,vU`P!WQOY#PZp#Ppq#hq;'S#P;'S;=`#z<%lO#P",tokenizers:[Ef,Sf,If,Lf,0,1],topRules:{Stream:[0,15]},tokenPrec:0}),Qf=Fs.define({name:"yaml",parser:Ff.configure({props:[rl.add({Stream:e=>{for(let t=e.node.resolve(e.pos,-1);t&&t.to>=e.pos;t=t.parent){if("BlockLiteralContent"==t.name&&t.from<t.to)return e.baseIndentFor(t);if("BlockLiteral"==t.name)return e.baseIndentFor(t)+e.unit;if("BlockSequence"==t.name||"BlockMapping"==t.name)return e.column(t.from,1);if("QuotedLiteral"==t.name)return null;if("Literal"==t.name){let n=e.column(t.from,1);if(n==e.lineIndent(t.from,1))return n;if(t.to>e.pos)return null}}return null},FlowMapping:cl({closing:"}"}),FlowSequence:cl({closing:"]"})}),hl.add({"FlowMapping FlowSequence":fl,"Item Pair BlockLiteral":(e,t)=>({from:t.doc.lineAt(e.from).to,to:e.to})})]}),languageData:{commentTokens:{line:"#"},indentOnInput:/^\s*[\]\}]$/}});function Uf(){var e=arguments[0];"string"==typeof e&&(e=document.createElement(e));var t=1,n=arguments[1];if(n&&"object"==typeof n&&null==n.nodeType&&!Array.isArray(n)){for(var r in n)if(Object.prototype.hasOwnProperty.call(n,r)){var a=n[r];"string"==typeof a?e.setAttribute(r,a):null!=a&&(e[r]=a)}t++}for(;t<arguments.length;t++)zf(e,arguments[t]);return e}function zf(e,t){if("string"==typeof t)e.appendChild(document.createTextNode(t));else if(null==t);else if(null!=t.nodeType)e.appendChild(t);else{if(!Array.isArray(t))throw new RangeError("Unsupported child node: "+t);for(var n=0;n<t.length;n++)zf(e,t[n])}}Ds.meta;class jf{constructor(e,t,n){this.from=e,this.to=t,this.diagnostic=n}}class Wf{constructor(e,t,n){this.diagnostics=e,this.panel=t,this.selected=n}static init(e,t,n){let r=e,a=n.facet(nm).markerFilter;a&&(r=a(r,n));let o=e.slice().sort(((e,t)=>e.from-t.from||e.to-t.to)),i=new De,s=[],l=0;for(let e=0;;){let t,r,a=e==o.length?null:o[e];if(!a&&!s.length)break;for(s.length?(t=l,r=s.reduce(((e,t)=>Math.min(e,t.to)),a&&a.from>t?a.from:1e8)):(t=a.from,r=a.to,s.push(a),e++);e<o.length;){let n=o[e];if(n.from!=t||!(n.to>n.from||n.to==t)){r=Math.min(n.from,r);break}s.push(n),e++,r=Math.min(n.to,r)}let c=dm(s);if(s.some((e=>e.from==e.to||e.from==e.to-1&&n.doc.lineAt(e.from).to==e.from)))i.add(t,t,rn.widget({widget:new om(c),diagnostics:s.slice()}));else{let e=s.reduce(((e,t)=>t.markClass?e+" "+t.markClass:e),"");i.add(t,r,rn.mark({class:"cm-lintRange cm-lintRange-"+c+e,diagnostics:s.slice(),inclusiveEnd:s.some((e=>e.to>r))}))}l=r;for(let e=0;e<s.length;e++)s[e].to<=l&&s.splice(e--,1)}let c=i.finish();return new Wf(c,t,Xf(c))}}function Xf(e,t=null,n=0){let r=null;return e.between(n,1e9,((e,n,{spec:a})=>{if(!(t&&a.diagnostics.indexOf(t)<0))if(r){if(a.diagnostics.indexOf(r.diagnostic)<0)return!1;r=new jf(r.from,n,r.diagnostic)}else r=new jf(e,n,t||a.diagnostics[0])})),r}function qf(e,t){return e.field(Gf,!1)?t:t.concat(fe.appendConfig.of(pm))}const Vf=fe.define(),Hf=fe.define(),Zf=fe.define(),Gf=Z.define({create:()=>new Wf(rn.none,null,null),update(e,t){if(t.docChanged&&e.diagnostics.size){let n=e.diagnostics.map(t.changes),r=null,a=e.panel;if(e.selected){let a=t.changes.mapPos(e.selected.from,1);r=Xf(n,e.selected.diagnostic,a)||Xf(n,null,a)}!n.size&&a&&t.state.facet(nm).autoPanel&&(a=null),e=new Wf(n,a,r)}for(let n of t.effects)if(n.is(Vf)){let r=t.state.facet(nm).autoPanel?n.value.length?sm.open:null:e.panel;e=Wf.init(n.value,r,t.state)}else n.is(Hf)?e=new Wf(e.diagnostics,n.value?sm.open:null,e.selected):n.is(Zf)&&(e=new Wf(e.diagnostics,e.panel,n.value));return e},provide:e=>[Vo.from(e,(e=>e.panel)),Ya.decorations.from(e,(e=>e.diagnostics))]}),Kf=rn.mark({class:"cm-lintRange cm-lintRange-active"});function Yf(e,t,n){let r,{diagnostics:a}=e.state.field(Gf),o=-1,i=-1;a.between(t-(n<0?1:0),t+(n>0?1:0),((e,a,{spec:s})=>{if(t>=e&&t<=a&&(e==a||(t>e||n>0)&&(t<a||n<0)))return r=s.diagnostics,o=e,i=a,!1}));let s=e.state.facet(nm).tooltipFilter;return r&&s&&(r=s(r,e.state)),r?{pos:o,end:i,above:e.state.doc.lineAt(o).to<i,create:()=>({dom:Jf(e,r)})}:null}function Jf(e,t){return Uf("ul",{class:"cm-tooltip-lint"},t.map((t=>am(e,t,!1))))}const em=e=>{let t=e.state.field(Gf,!1);return!(!t||!t.panel||(e.dispatch({effects:Hf.of(!1)}),0))},tm=er.fromClass(class{constructor(e){this.view=e,this.timeout=-1,this.set=!0;let{delay:t}=e.state.facet(nm);this.lintTime=Date.now()+t,this.run=this.run.bind(this),this.timeout=setTimeout(this.run,t)}run(){clearTimeout(this.timeout);let e=Date.now();if(e<this.lintTime-10)this.timeout=setTimeout(this.run,this.lintTime-e);else{this.set=!1;let{state:e}=this.view,{sources:t}=e.facet(nm);t.length&&function(e,t,n){let r=[],a=-1;for(let o of e)o.then((n=>{r.push(n),clearTimeout(a),r.length==e.length?t(r):a=setTimeout((()=>t(r)),200)}),n)}(t.map((e=>Promise.resolve(e(this.view)))),(t=>{this.view.state.doc==e.doc&&this.view.dispatch(function(e,t){return{effects:qf(e,[Vf.of(t)])}}(this.view.state,t.reduce(((e,t)=>e.concat(t)))))}),(e=>{Gn(this.view.state,e)}))}}update(e){let t=e.state.facet(nm);(e.docChanged||t!=e.startState.facet(nm)||t.needsRefresh&&t.needsRefresh(e))&&(this.lintTime=Date.now()+t.delay,this.set||(this.set=!0,this.timeout=setTimeout(this.run,t.delay)))}force(){this.set&&(this.lintTime=Date.now(),this.run())}destroy(){clearTimeout(this.timeout)}}),nm=z.define({combine:e=>Object.assign({sources:e.map((e=>e.source)).filter((e=>null!=e))},_e(e.map((e=>e.config)),{delay:750,markerFilter:null,tooltipFilter:null,needsRefresh:null,hideOn:()=>null},{needsRefresh:(e,t)=>e?t?n=>e(n)||t(n):e:t}))});function rm(e){let t=[];if(e)e:for(let{name:n}of e){for(let e=0;e<n.length;e++){let r=n[e];if(/[a-zA-Z]/.test(r)&&!t.some((e=>e.toLowerCase()==r.toLowerCase()))){t.push(r);continue e}}t.push("")}return t}function am(e,t,n){var r;let a=n?rm(t.actions):[];return Uf("li",{class:"cm-diagnostic cm-diagnostic-"+t.severity},Uf("span",{class:"cm-diagnosticText"},t.renderMessage?t.renderMessage(e):t.message),null===(r=t.actions)||void 0===r?void 0:r.map(((n,r)=>{let o=!1,i=r=>{if(r.preventDefault(),o)return;o=!0;let a=Xf(e.state.field(Gf).diagnostics,t);a&&n.apply(e,a.from,a.to)},{name:s}=n,l=a[r]?s.indexOf(a[r]):-1,c=l<0?s:[s.slice(0,l),Uf("u",s.slice(l,l+1)),s.slice(l+1)];return Uf("button",{type:"button",class:"cm-diagnosticAction",onclick:i,onmousedown:i,"aria-label":` Action: ${s}${l<0?"":` (access key "${a[r]})"`}.`},c)})),t.source&&Uf("div",{class:"cm-diagnosticSource"},t.source))}class om extends tn{constructor(e){super(),this.sev=e}eq(e){return e.sev==this.sev}toDOM(){return Uf("span",{class:"cm-lintPoint cm-lintPoint-"+this.sev})}}class im{constructor(e,t){this.diagnostic=t,this.id="item_"+Math.floor(4294967295*Math.random()).toString(16),this.dom=am(e,t,!0),this.dom.id=this.id,this.dom.setAttribute("role","option")}}class sm{constructor(e){this.view=e,this.items=[],this.list=Uf("ul",{tabIndex:0,role:"listbox","aria-label":this.view.state.phrase("Diagnostics"),onkeydown:t=>{if(27==t.keyCode)em(this.view),this.view.focus();else if(38==t.keyCode||33==t.keyCode)this.moveSelection((this.selectedIndex-1+this.items.length)%this.items.length);else if(40==t.keyCode||34==t.keyCode)this.moveSelection((this.selectedIndex+1)%this.items.length);else if(36==t.keyCode)this.moveSelection(0);else if(35==t.keyCode)this.moveSelection(this.items.length-1);else if(13==t.keyCode)this.view.focus();else{if(!(t.keyCode>=65&&t.keyCode<=90&&this.selectedIndex>=0))return;{let{diagnostic:n}=this.items[this.selectedIndex],r=rm(n.actions);for(let a=0;a<r.length;a++)if(r[a].toUpperCase().charCodeAt(0)==t.keyCode){let t=Xf(this.view.state.field(Gf).diagnostics,n);t&&n.actions[a].apply(e,t.from,t.to)}}}t.preventDefault()},onclick:e=>{for(let t=0;t<this.items.length;t++)this.items[t].dom.contains(e.target)&&this.moveSelection(t)}}),this.dom=Uf("div",{class:"cm-panel-lint"},this.list,Uf("button",{type:"button",name:"close","aria-label":this.view.state.phrase("close"),onclick:()=>em(this.view)},"×")),this.update()}get selectedIndex(){let e=this.view.state.field(Gf).selected;if(!e)return-1;for(let t=0;t<this.items.length;t++)if(this.items[t].diagnostic==e.diagnostic)return t;return-1}update(){let{diagnostics:e,selected:t}=this.view.state.field(Gf),n=0,r=!1,a=null,o=new Set;for(e.between(0,this.view.state.doc.length,((e,i,{spec:s})=>{for(let e of s.diagnostics){if(o.has(e))continue;o.add(e);let i,s=-1;for(let t=n;t<this.items.length;t++)if(this.items[t].diagnostic==e){s=t;break}s<0?(i=new im(this.view,e),this.items.splice(n,0,i),r=!0):(i=this.items[s],s>n&&(this.items.splice(n,s-n),r=!0)),t&&i.diagnostic==t.diagnostic?i.dom.hasAttribute("aria-selected")||(i.dom.setAttribute("aria-selected","true"),a=i):i.dom.hasAttribute("aria-selected")&&i.dom.removeAttribute("aria-selected"),n++}}));n<this.items.length&&!(1==this.items.length&&this.items[0].diagnostic.from<0);)r=!0,this.items.pop();0==this.items.length&&(this.items.push(new im(this.view,{from:-1,to:-1,severity:"info",message:this.view.state.phrase("No diagnostics")})),r=!0),a?(this.list.setAttribute("aria-activedescendant",a.id),this.view.requestMeasure({key:this,read:()=>({sel:a.dom.getBoundingClientRect(),panel:this.list.getBoundingClientRect()}),write:({sel:e,panel:t})=>{let n=t.height/this.list.offsetHeight;e.top<t.top?this.list.scrollTop-=(t.top-e.top)/n:e.bottom>t.bottom&&(this.list.scrollTop+=(e.bottom-t.bottom)/n)}})):this.selectedIndex<0&&this.list.removeAttribute("aria-activedescendant"),r&&this.sync()}sync(){let e=this.list.firstChild;function t(){let t=e;e=t.nextSibling,t.remove()}for(let n of this.items)if(n.dom.parentNode==this.list){for(;e!=n.dom;)t();e=n.dom.nextSibling}else this.list.insertBefore(n.dom,e);for(;e;)t()}moveSelection(e){if(this.selectedIndex<0)return;let t=Xf(this.view.state.field(Gf).diagnostics,this.items[e].diagnostic);t&&this.view.dispatch({selection:{anchor:t.from,head:t.to},scrollIntoView:!0,effects:Zf.of(t)})}static open(e){return new sm(e)}}function lm(e){return function(e,t='viewBox="0 0 40 40"'){return`url('data:image/svg+xml,<svg xmlns="http://www.w3.org/2000/svg" ${t}>${encodeURIComponent(e)}</svg>')`}(`<path d="m0 2.5 l2 -1.5 l1 0 l2 1.5 l1 0" stroke="${e}" fill="none" stroke-width=".7"/>`,'width="6" height="3"')}const cm=Ya.baseTheme({".cm-diagnostic":{padding:"3px 6px 3px 8px",marginLeft:"-1px",display:"block",whiteSpace:"pre-wrap"},".cm-diagnostic-error":{borderLeft:"5px solid #d11"},".cm-diagnostic-warning":{borderLeft:"5px solid orange"},".cm-diagnostic-info":{borderLeft:"5px solid #999"},".cm-diagnostic-hint":{borderLeft:"5px solid #66d"},".cm-diagnosticAction":{font:"inherit",border:"none",padding:"2px 4px",backgroundColor:"#444",color:"white",borderRadius:"3px",marginLeft:"8px",cursor:"pointer"},".cm-diagnosticSource":{fontSize:"70%",opacity:.7},".cm-lintRange":{backgroundPosition:"left bottom",backgroundRepeat:"repeat-x",paddingBottom:"0.7px"},".cm-lintRange-error":{backgroundImage:lm("#d11")},".cm-lintRange-warning":{backgroundImage:lm("orange")},".cm-lintRange-info":{backgroundImage:lm("#999")},".cm-lintRange-hint":{backgroundImage:lm("#66d")},".cm-lintRange-active":{backgroundColor:"#ffdd9980"},".cm-tooltip-lint":{padding:0,margin:0},".cm-lintPoint":{position:"relative","&:after":{content:'""',position:"absolute",bottom:0,left:"-2px",borderLeft:"3px solid transparent",borderRight:"3px solid transparent",borderBottom:"4px solid #d11"}},".cm-lintPoint-warning":{"&:after":{borderBottomColor:"orange"}},".cm-lintPoint-info":{"&:after":{borderBottomColor:"#999"}},".cm-lintPoint-hint":{"&:after":{borderBottomColor:"#66d"}},".cm-panel.cm-panel-lint":{position:"relative","& ul":{maxHeight:"100px",overflowY:"auto","& [aria-selected]":{backgroundColor:"#ddd","& u":{textDecoration:"underline"}},"&:focus [aria-selected]":{background_fallback:"#bdf",backgroundColor:"Highlight",color_fallback:"white",color:"HighlightText"},"& u":{textDecoration:"none"},padding:0,margin:0},"& [name=close]":{position:"absolute",top:"0",right:"2px",background:"inherit",border:"none",font:"inherit",padding:0,margin:0}}});function um(e){return"error"==e?4:"warning"==e?3:"info"==e?2:1}function dm(e){let t="hint",n=1;for(let r of e){let e=um(r.severity);e>n&&(n=e,t=r.severity)}return t}const pm=[Gf,Ya.decorations.compute([Gf],(e=>{let{selected:t,panel:n}=e.field(Gf);return t&&n&&t.from!=t.to?rn.set([Kf.range(t.from,t.to)]):rn.none})),Qo(Yf,{hideOn:function(e,t){let n=t.pos,r=t.end||n,a=e.state.facet(nm).hideOn(e,n,r);if(null!=a)return a;let o=e.startState.doc.lineAt(t.pos);return!(!e.effects.some((e=>e.is(Vf)))&&!e.changes.touchesRange(o.from,Math.max(o.to,r)))}}),cm];var hm=n(1266),fm=n(953),mm=n(641),gm=n(3751);const vm=(({theme:e,settings:t={},styles:n=[]})=>{const r={".cm-gutters":{}},a={};t.background&&(a.backgroundColor=t.background),t.backgroundImage&&(a.backgroundImage=t.backgroundImage),t.foreground&&(a.color=t.foreground),t.fontSize&&(a.fontSize=t.fontSize),(t.background||t.foreground)&&(r["&"]=a),t.fontFamily&&(r["&.cm-editor .cm-scroller"]={fontFamily:t.fontFamily}),t.gutterBackground&&(r[".cm-gutters"].backgroundColor=t.gutterBackground),t.gutterForeground&&(r[".cm-gutters"].color=t.gutterForeground),t.gutterBorder&&(r[".cm-gutters"].borderRightColor=t.gutterBorder),t.caret&&(r[".cm-content"]={caretColor:t.caret},r[".cm-cursor, .cm-dropCursor"]={borderLeftColor:t.caret});const o={};return t.gutterActiveForeground&&(o.color=t.gutterActiveForeground),t.lineHighlight&&(r[".cm-activeLine"]={backgroundColor:t.lineHighlight},o.backgroundColor=t.lineHighlight),r[".cm-activeLineGutter"]=o,t.selection&&(r["&.cm-focused .cm-selectionBackground, & .cm-line::selection, & .cm-selectionLayer .cm-selectionBackground, .cm-content ::selection"]={background:t.selection+" !important"}),t.selectionMatch&&(r["& .cm-selectionMatch"]={backgroundColor:t.selectionMatch}),[Ya.theme(r,{dark:"dark"===e}),Ll(Ml.define(n))]})({theme:"light",settings:{background:"var(--scalar-background-2)",foreground:"var(--scalar-color-1)",caret:"var(--scalar-color-1)",selectionMatch:"#e3dcce",gutterBackground:"var(--scalar-background-2)",gutterForeground:"var(--scalar-color-3)",gutterBorder:"transparent",lineHighlight:"var(--scalar-background-3)",fontFamily:"var(--scalar-font-code)"},styles:[{tag:[Ds.standard(Ds.tagName),Ds.tagName],color:"var(--scalar-color-purple)"},{tag:[Ds.comment],color:"var(--scalar-color-3)"},{tag:[Ds.className],color:"var(--scalar-color-orange)"},{tag:[Ds.variableName,Ds.propertyName,Ds.attributeName],color:"var(--scalar-color-1)"},{tag:[Ds.operator],color:"var(--scalar-color-2)"},{tag:[Ds.keyword,Ds.typeName,Ds.typeOperator],color:"var(--scalar-color-green)"},{tag:[Ds.string],color:"var(--scalar-color-blue)"},{tag:[Ds.bracket,Ds.regexp,Ds.meta],color:"var(--scalar-color-3)"},{tag:[Ds.number],color:"var(--scalar-color-orange)"},{tag:[Ds.name,Ds.quote],color:"var(--scalar-color-3)"},{tag:[Ds.heading],color:"var(--scalar-color-3)",fontWeight:"bold"},{tag:[Ds.emphasis],color:"var(--scalar-color-3)",fontStyle:"italic"},{tag:[Ds.deleted],color:"var(--scalar-color-3)",backgroundColor:"transparent"},{tag:[Ds.atom,Ds.bool,Ds.special(Ds.variableName)],color:"var(--scalar-color-3)"},{tag:[Ds.url,Ds.escape,Ds.regexp,Ds.link],color:"var(--scalar-color-1)"},{tag:Ds.link,textDecoration:"underline"},{tag:Ds.strikethrough,textDecoration:"line-through"},{tag:Ds.invalid,color:"var(--scalar-color-3)"}]}),bm=new fo({regexp:/(\{[^}]+\})/g,decoration:()=>rn.mark({attributes:{class:"api-client-url-variable"}})}),ym=()=>er.fromClass(class{variables;constructor(e){this.variables=bm.createDeco(e)}update(e){this.variables=bm.updateDeco(e,this.variables)}},{decorations:e=>e.variables,provide:e=>Ya.atomicRanges.of((t=>t.plugin(e)?.variables||rn.none))}),wm=e=>"provider"in e&&!!(0,fm.BA)(e.provider),Om={key:"Mod-a",run:e=>(e.dispatch({selection:{anchor:0,head:e.state.doc.length},scrollIntoView:!1}),!0)},km=e=>{const t=(0,fm.KR)(null),n=(e="")=>{t.value&&t.value.state.doc.toString()!==e&&t.value.dispatch({changes:{from:0,to:t.value.state.doc.length,insert:e},selection:{anchor:Math.min(t.value.state.selection.main.anchor,e.length)}})},r=(0,mm.EW)((()=>({onChange:e.onChange,onBlur:e.onBlur,onFocus:e.onFocus,disableTabIndent:(0,fm.BA)(e.disableTabIndent),language:(0,fm.BA)(e.language),classes:(0,fm.BA)(e.classes),readOnly:(0,fm.BA)(e.readOnly),lineNumbers:(0,fm.BA)(e.lineNumbers),withVariables:(0,fm.BA)(e.withVariables),forceFoldGutter:(0,fm.BA)(e.forceFoldGutter),disableEnter:(0,fm.BA)(e.disableEnter),disableCloseBrackets:(0,fm.BA)(e.disableCloseBrackets),withoutTheme:(0,fm.BA)(e.withoutTheme),lint:(0,fm.BA)(e.lint),additionalExtensions:(0,fm.BA)(e.extensions),placeholder:(0,fm.BA)(e.placeholder)})));function a(){if(e.codeMirrorRef.value){const a=wm(e)?(0,fm.BA)(e.provider):null,o=Em({...r.value,provider:a});t.value=new Ya({parent:e.codeMirrorRef.value,extensions:o}),wm(e)||n((0,fm.BA)(e.content))}}return(0,mm.wB)(e.codeMirrorRef,(()=>{t.value?.destroy(),a()}),{immediate:!0}),(0,mm.xo)((()=>t.value?.destroy())),(0,mm.wB)((()=>wm(e)?(0,fm.BA)(e.provider):null),(()=>{wm(e)&&(t.value?.destroy(),a())})),(0,mm.wB)(r,(()=>{if(!t.value)return;const n=wm(e)?(0,fm.BA)(e.provider):null,a=Em({...r.value,provider:n});requestAnimationFrame((()=>{t.value?.dispatch({effects:fe.reconfigure.of(a)})}))}),{immediate:!0}),(0,mm.wB)((()=>(0,fm.BA)(e.content)),(()=>{wm(e)||n((0,fm.BA)(e.content))}),{immediate:!0}),{setCodeMirrorContent:n,codeMirror:t}},xm={html:function(e={}){let t,n="";!1===e.matchClosingTags&&(n="noMatch"),!0===e.selfClosingTags&&(n=(n?n+" ":"")+"selfClosing"),(e.nestedLanguages&&e.nestedLanguages.length||e.nestedAttributes&&e.nestedAttributes.length)&&(t=Dp((e.nestedLanguages||[]).concat($h),(e.nestedAttributes||[]).concat(Nh)));let r=t?Ih.configure({wrap:t,dialect:n}):n?Lh.configure({dialect:n}):Lh;return new Gs(r,[Lh.data.of({autocomplete:Dh(e)}),!1!==e.autoCloseTags?Fh:[],dh().support,ip().support])},json:function(){return new Gs(zh)},yaml:function(){return new Gs(Qf)},css:ip,xml:function(e={}){let t=[pf.data.of({autocomplete:df(e.elements||[],e.attributes||[])})];return!1!==e.autoCloseTags&&t.push(ff),new Gs(pf,t)}};function Em({onChange:e,onBlur:t,onFocus:n,provider:r,language:a,classes:o=[],readOnly:i=!1,lineNumbers:s=!1,withVariables:l=!1,forceFoldGutter:c=!1,disableEnter:u=!1,disableCloseBrackets:d=!1,disableTabIndent:p=!1,withoutTheme:h=!1,lint:f=!1,additionalExtensions:m=[],placeholder:g}){const v=[wo(),Ll(Ql,{fallback:!0}),Ya.theme({".cm-line":{lineHeight:"20px",padding:"0 2px 0 4px"},".cm-gutterElement":{lineHeight:"20px"},".cm-tooltip":{border:"1px solid #f5c6cb",fontSize:"12px"},".cm-tooltip-lint":{backgroundColor:"#ffffff"},".cm-diagnostic-error":{borderLeft:"0",color:"#dc1b19"},".cm-foldPlaceholder":{background:"var(--scalar-background-1)",border:"none",fontFamily:"var(--scalar-font)"}}),Ya.updateListener.of((t=>{t.docChanged&&e?.(t.state.doc.toString())})),Ya.domEventHandlers({blur:(e,n)=>{t?.(n.state.doc.toString())},focus:(e,t)=>{n?.(t.state.doc.toString())}}),Ya.editorAttributes.of({class:o.join(" ")}),...m];if(r&&v.push(r),h||v.push(vm),i?v.push(Ya.editable.of(!1)):(v.push(Se.transactionFilter.of((e=>{if(!e.docChanged||!e.isUserEvent("input.type")&&!e.isUserEvent("input.complete"))return e;let t=e.startState.languageDataAt("indentOnInput",e.startState.selection.main.head);if(!t.length)return e;let n=e.newDoc,{head:r}=e.newSelection.main,a=n.lineAt(r);if(r>a.from+200)return e;let o=n.sliceString(a.from,r);if(!t.some((e=>e.test(o))))return e;let{state:i}=e,s=-1,l=[];for(let{head:e}of i.selection.ranges){let t=i.doc.lineAt(e);if(t.from==s)continue;s=t.from;let n=tl(i,t.from);if(null==n)continue;let r=/^\s*/.exec(t.text)[0],a=el(i,n);r!=a&&l.push({from:t.from,to:t.from+r.length,insert:a})}return l.length?[e,{changes:l,sequential:!0}]:e})),Zl(),function(e={}){return[cu,tu,Lc.of(e),su,qu,uu]}(),io.of([...Xu,Om]),Zl()),d||v.push([Iu,Pu],io.of([...Lu])),p?v.push(io.of([{key:"Tab",run:()=>!1,shift:()=>!1}])):v.push(io.of([id]))),g&&v.push(function(e){return er.fromClass(class{constructor(t){this.view=t,this.placeholder=e?rn.set([rn.widget({widget:new Eo(e),side:1}).range(0)]):rn.none}get decorations(){return this.view.state.doc.length?rn.none:this.placeholder}},{decorations:e=>e.decorations})}(g)),s&&v.push(function(e={}){return[ui.of(e),ti(),hi]}()),c&&v.push(Pl({markerDOM:e=>{const t=document.createElement("div");t.classList.add("cm-foldMarker");const n=(0,mm.h)(hm.kv,{icon:e?"ChevronDown":"ChevronRight",size:"md"});return(0,gm.XX)(n,t),t}})),a&&xm[a]&&(v.push(xm[a]()),c||v.push(Pl({markerDOM:e=>{const t=document.createElement("div");t.classList.add("cm-foldMarker");const n=(0,mm.h)(hm.kv,{icon:e?"ChevronDown":"ChevronRight",size:"md"});return(0,gm.XX)(n,t),t}}))),f&&"json"===a){const e=function(e,t={}){return[nm.of({source:e,config:t}),tm,pm]}((e=>{const t=[],n=e.state.doc.toString();if(n.trim())try{JSON.parse(n)}catch(n){n instanceof Error&&t.push({from:0,to:e.state.doc.length,severity:"error",message:n.message})}return t}));v.push(e)}return l&&v.push(ym()),u?v.push(io.of([{key:"Enter",run:()=>!0},{key:"Ctrl-Enter",mac:"Cmd-Enter",run:()=>!0},{key:"Shift-Enter",run:()=>!0}])):v.push(io.of([{key:"Enter",run:ad}])),v}function Sm(e){const{codeMirror:t,query:n,showDropdown:r,dropdownPosition:a}=e;function o(){return t.value?.state.selection.main.head||0}function i(e){return t.value?.coordsAtPos(e)}function s(){const e=o();requestAnimationFrame((()=>{const t=i(e-n.value.length-2);t&&(a.value={left:t.left,top:Math.max(t.bottom)})}))}function l(){const e=o(),l=t.value?.state.doc.sliceString(0,e)||"",c=l.lastIndexOf("{{");c>l.lastIndexOf("}}")?(n.value=l.slice(c+2),r.value=!0,a.value=i(e)||{left:0,top:0},""===n.value&&s()):r.value=!1}return(0,mm.wB)(n,s),(0,mm.wB)((()=>t.value?.state.doc.toString()),l),{handleDropdownSelect:function(e){const a=`{{${e}}}`,i=o(),s=Math.max(0,i-n.value.length-2),l=i;t.value?.dispatch({changes:{from:s,to:l,insert:a}}),r.value=!1},updateDropdownVisibility:l}}},9072:(e,t,n)=>{"use strict";n.d(t,{z:()=>o});var r=n(1259),a=n(6530);const o=(0,r.v)({attributes:{xmlnsxlink:"xmlns:xlink"},properties:{xmlnsXLink:null,xmlns:null},space:"xmlns",transform:a._})},9135:(e,t,n)=>{"use strict";n.d(t,{UF:()=>c,T7:()=>l,JC:()=>u});var r=n(165),a=n(8803),o=n(4048),i=n(641),s=n(2490);const l=({collections:e,requestExamples:t,requests:n,router:l,servers:c,workspaces:u})=>{const d=(0,i.EW)((x=l,()=>{const e={[s.TU.Collection]:"default",[s.TU.Environment]:"default",[s.TU.Request]:"default",[s.TU.Examples]:"default",[s.TU.Schema]:"default",[s.TU.Cookies]:"default",[s.TU.Servers]:"default",[s.TU.Workspace]:"default",[s.TU.Settings]:"default"},t=null==x?void 0:x.currentRoute.value;return t&&Object.keys(e).forEach((n=>{t.params[n]&&(e[n]=t.params[n])})),e})),p=(0,i.EW)((()=>u[d.value[s.TU.Workspace]]??Object.values(u)[0])),h=(0,i.EW)((()=>{var t;return(null==(t=p.value)?void 0:t.collections.map((t=>e[t])).filter(o.O9).sort(((e,t)=>{var n,r;return"Drafts"===(null==(n=e.info)?void 0:n.title)?1:"Drafts"===(null==(r=t.info)?void 0:r.title)?-1:0})))??[]})),f=(0,i.EW)((()=>{var e;return null==(e=h.value)?void 0:e.flatMap((e=>e.servers.map((e=>c[e]))))})),m=(0,i.EW)((()=>{var e;return(null==(e=h.value)?void 0:e.flatMap((e=>e.requests)))??[]})),g=(0,i.EW)((()=>{var e,t,n,r,o,i,s,l,c;if(null==(e=p.value)||!e.activeEnvironmentId)return a.V.parse({uid:"default",color:"#0082D0",name:"No Environment",value:JSON.stringify(null==(t=p.value)?void 0:t.environments,null,2)});const u=h.value.find((e=>{var t,n;return null==(n=e["x-scalar-environments"])?void 0:n[(null==(t=p.value)?void 0:t.activeEnvironmentId)??""]}));return u&&null!=(n=p.value)&&n.activeEnvironmentId?a.V.parse({uid:p.value.activeEnvironmentId,name:p.value.activeEnvironmentId,value:JSON.stringify(null==(i=null==(o=u["x-scalar-environments"])?void 0:o[null==(r=p.value)?void 0:r.activeEnvironmentId])?void 0:i.variables,null,2),color:(null==(c=null==(l=u["x-scalar-environments"])?void 0:l[null==(s=p.value)?void 0:s.activeEnvironmentId])?void 0:c.color)||"#0082D0",isDefault:!1}):a.V.parse({uid:"default",color:"#0082D0",name:"No Environment",value:JSON.stringify(p.value.environments,null,2)})})),v=(0,i.EW)((()=>{var t;const r=d.value[s.TU.Request],a=e[d.value.collection]||e[(null==(t=p.value)?void 0:t.collections[0])??""];return n[r]||n[(null==a?void 0:a.requests[0])??""]||Object.values(n)[0]})),b=(0,i.EW)((()=>{var e;const n="default"===d.value[s.TU.Examples]?(null==(e=v.value)?void 0:e.examples[0])||"":d.value[s.TU.Examples];return t[n]})),y=(0,i.EW)((()=>{var t,n,r;const a=null==(t=v.value)?void 0:t.uid;if(a)return Object.values(e).find((e=>{var t;return null==(t=e.requests)?void 0:t.includes(a)}));const o=(null==(n=p.value)?void 0:n.collections[0])??(null==(r=e[0])?void 0:r.uid)??"";return e[o]})),w=(0,i.EW)((()=>{var e,t,n;return c[((null==(e=v.value)?void 0:e.selectedServerUid)||(null==(t=y.value)?void 0:t.selectedServerUid)||(null==(n=y.value)?void 0:n.servers[0]))??""]})),O=(0,i.EW)((()=>{var e;return"default"===d.value[s.TU.Cookies]?(null==(e=p.value)?void 0:e.cookies[0])??"default":d.value[s.TU.Cookies]})),k=(0,i.EW)((()=>{var e;const t=(null==(e=p.value)?void 0:e.environments)??{},n=g.value.uid?JSON.parse(g.value.value):{},a=(0,r.o6)(t).map((([e,t])=>({key:e,value:t,source:"global"}))),o=(0,r.o6)(n).map((([e,t])=>({key:e,value:t,source:"collection"}))),i=new Map;return o.forEach((e=>{i.set(e.key,e)})),a.forEach((e=>{i.has(e.key)||i.set(e.key,e)})),Array.from(i.values())}));var x;return{activeCollection:y,activeCookieId:O,activeExample:b,activeRequest:v,activeRouterParams:d,activeEnvironment:g,activeServer:w,activeWorkspace:p,activeWorkspaceCollections:h,activeWorkspaceServers:f,activeEnvVariables:k,activeWorkspaceRequests:m}},c=Symbol(),u=()=>{const e=(0,i.WQ)(c);if(!e)throw new Error("Active entities store not provided");return e}},9143:(e,t,n)=>{"use strict";n.d(t,{C:()=>r});const r=function(e){if(null==e)return o;if("function"==typeof e)return a(e);if("object"==typeof e)return Array.isArray(e)?function(e){const t=[];let n=-1;for(;++n<e.length;)t[n]=r(e[n]);return a((function(...e){let n=-1;for(;++n<t.length;)if(t[n].apply(this,e))return!0;return!1}))}(e):function(e){const t=e;return a((function(n){const r=n;let a;for(a in e)if(r[a]!==t[a])return!1;return!0}))}(e);if("string"==typeof e)return t=e,a((function(e){return e&&e.type===t}));var t;throw new Error("Expected function, string, or object as test")};function a(e){return function(t,n,r){return Boolean(function(e){return null!==e&&"object"==typeof e&&"type"in e}(t)&&e.call(this,t,"number"==typeof n?n:void 0,r||void 0))}}function o(){return!0}},9146:(e,t,n)=>{"use strict";n.d(t,{_:()=>o});var r=Object.defineProperty,a=(e,t,n)=>(((e,t,n)=>{t in e?r(e,t,{enumerable:!0,configurable:!0,writable:!0,value:n}):e[t]=n})(e,"symbol"!=typeof t?t+"":t,n),n);let o=new class{constructor(){a(this,"current",this.detect()),a(this,"currentId",0)}set(e){this.current!==e&&(this.currentId=0,this.current=e)}reset(){this.set(this.detect())}nextId(){return++this.currentId}get isServer(){return"server"===this.current}get isClient(){return"client"===this.current}detect(){return"undefined"==typeof window||"undefined"==typeof document?"server":"client"}}},9261:(e,t,n)=>{"use strict";n.d(t,{cf:()=>ii,Vg:()=>ai,sV:()=>li});var r={};n.r(r),n.d(r,{boolean:()=>se,booleanish:()=>le,commaOrSpaceSeparated:()=>he,commaSeparated:()=>pe,number:()=>ue,overloadedBoolean:()=>ce,spaceSeparated:()=>de});var a={};n.r(a),n.d(a,{attentionMarkers:()=>Ka,contentInitial:()=>Xa,disable:()=>Ya,document:()=>Wa,flow:()=>Va,flowInitial:()=>qa,insideSpan:()=>Ga,string:()=>Ha,text:()=>Za});const o="object"==typeof self?self:globalThis,i=e=>((e,t)=>{const n=(t,n)=>(e.set(n,t),t),r=a=>{if(e.has(a))return e.get(a);const[i,s]=t[a];switch(i){case 0:case-1:return n(s,a);case 1:{const e=n([],a);for(const t of s)e.push(r(t));return e}case 2:{const e=n({},a);for(const[t,n]of s)e[r(t)]=r(n);return e}case 3:return n(new Date(s),a);case 4:{const{source:e,flags:t}=s;return n(new RegExp(e,t),a)}case 5:{const e=n(new Map,a);for(const[t,n]of s)e.set(r(t),r(n));return e}case 6:{const e=n(new Set,a);for(const t of s)e.add(r(t));return e}case 7:{const{name:e,message:t}=s;return n(new o[e](t),a)}case 8:return n(BigInt(s),a);case"BigInt":return n(Object(BigInt(s)),a);case"ArrayBuffer":return n(new Uint8Array(s).buffer,s);case"DataView":{const{buffer:e}=new Uint8Array(s);return n(new DataView(e),s)}}return n(new o[i](s),a)};return r})(new Map,e)(0),s="",{toString:l}={},{keys:c}=Object,u=e=>{const t=typeof e;if("object"!==t||!e)return[0,t];const n=l.call(e).slice(8,-1);switch(n){case"Array":return[1,s];case"Object":return[2,s];case"Date":return[3,s];case"RegExp":return[4,s];case"Map":return[5,s];case"Set":return[6,s];case"DataView":return[1,n]}return n.includes("Array")?[1,n]:n.includes("Error")?[7,n]:[2,n]},d=([e,t])=>0===e&&("function"===t||"symbol"===t),p=(e,{json:t,lossy:n}={})=>{const r=[];return((e,t,n,r)=>{const a=(e,t)=>{const a=r.push(e)-1;return n.set(t,a),a},o=r=>{if(n.has(r))return n.get(r);let[i,s]=u(r);switch(i){case 0:{let t=r;switch(s){case"bigint":i=8,t=r.toString();break;case"function":case"symbol":if(e)throw new TypeError("unable to serialize "+s);t=null;break;case"undefined":return a([-1],r)}return a([i,t],r)}case 1:{if(s){let e=r;return"DataView"===s?e=new Uint8Array(r.buffer):"ArrayBuffer"===s&&(e=new Uint8Array(r)),a([s,[...e]],r)}const e=[],t=a([i,e],r);for(const t of r)e.push(o(t));return t}case 2:{if(s)switch(s){case"BigInt":return a([s,r.toString()],r);case"Boolean":case"Number":case"String":return a([s,r.valueOf()],r)}if(t&&"toJSON"in r)return o(r.toJSON());const n=[],l=a([i,n],r);for(const t of c(r))!e&&d(u(r[t]))||n.push([o(t),o(r[t])]);return l}case 3:return a([i,r.toISOString()],r);case 4:{const{source:e,flags:t}=r;return a([i,{source:e,flags:t}],r)}case 5:{const t=[],n=a([i,t],r);for(const[n,a]of r)(e||!d(u(n))&&!d(u(a)))&&t.push([o(n),o(a)]);return n}case 6:{const t=[],n=a([i,t],r);for(const n of r)!e&&d(u(n))||t.push(o(n));return n}}const{message:l}=r;return a([i,{name:s,message:l}],r)};return o})(!(t||n),!!t,new Map,r)(e),r},h="function"==typeof structuredClone?(e,t)=>t&&("json"in t||"lossy"in t)?i(p(e,t)):structuredClone(e):(e,t)=>i(p(e,t));var f=n(9747);const m=/^[a-zA-Z][a-zA-Z\d+\-.]*?:/,g=/^[a-zA-Z]:\\/;var v=n(2485),b=n(7970);const y=["http","https"],w=["nofollow"],O={};function k(e){const t=e||O,n=t.protocols||y,r=(0,f.a)(t.test);return function(e){(0,b.YR)(e,"element",(function(e,a,o){if("a"===e.tagName&&"string"==typeof e.properties.href&&r(e,a,o)){const r=e.properties.href;if(function(e){if("string"!=typeof e)throw new TypeError(`Expected a \`string\`, got \`${typeof e}\``);return!g.test(e)&&m.test(e)}(r)?n.includes(r.slice(0,r.indexOf(":"))):r.startsWith("//")){const n=x(t.content,e),r=n&&!Array.isArray(n)?[n]:n,a=x(t.rel,e)||w,o="string"==typeof a?(0,v.q)(a):a,i=x(t.target,e),s=x(t.properties,e);if(s&&Object.assign(e.properties,h(s)),o.length>0&&(e.properties.rel=[...o]),i&&(e.properties.target=i),r){const n=x(t.contentProperties,e)||{};e.children.push({type:"element",tagName:"span",properties:h(n),children:h(r)})}}}}))}}function x(e,t){return"function"==typeof e?e(t):e}const E=(0,f.a)((function(e){return"audio"===e.tagName||"canvas"===e.tagName||"embed"===e.tagName||"iframe"===e.tagName||"img"===e.tagName||"math"===e.tagName||"object"===e.tagName||"picture"===e.tagName||"svg"===e.tagName||"video"===e.tagName}));var S=n(8619),_=n(9143);const C=["address","article","aside","blockquote","body","br","caption","center","col","colgroup","dd","dialog","dir","div","dl","dt","figcaption","figure","footer","form","h1","h2","h3","h4","h5","h6","head","header","hgroup","hr","html","legend","li","li","listing","main","menu","nav","ol","optgroup","option","p","plaintext","pre","section","summary","table","tbody","td","td","tfoot","th","th","thead","tr","ul","wbr","xmp"],T=["button","input","select","textarea"],A=["area","base","basefont","dialog","datalist","head","link","meta","noembed","noframes","param","rp","script","source","style","template","track","title"],R={},P=(0,_.C)(["comment","doctype"]);function D(e,t){if("children"in e){const n={...t};return("root"===e.type||I(e))&&(n.before=!0,n.after=!0),n.whitespace=function(e,t){if("tagName"in e&&e.properties)switch(e.tagName){case"listing":case"plaintext":case"script":case"style":case"xmp":return"pre";case"nobr":return"nowrap";case"pre":return e.properties.wrap?"pre-wrap":"pre";case"td":case"th":return e.properties.noWrap?"nowrap":t.whitespace;case"textarea":return"pre-wrap"}return t.whitespace}(e,t),function(e,t){let n=t.before;const r=t.after,a=e.children;let o=a.length,i=-1;for(;++i<o;){const e=D(a[i],{...t,after:M(a,i,r),before:n});e.remove?(a.splice(i,1),i--,o--):e.ignore||(n=e.stripAtStart),N(a[i])&&(n=!1)}return{ignore:!1,stripAtStart:Boolean(n||r),remove:!1}}(e,n)}if("text"===e.type){if("normal"===t.whitespace)return function(e,t){const n=t.collapse(e.value),r={ignore:!1,stripAtStart:!1,remove:!1};let a=0,o=n.length;return t.before&&B(n.charAt(0))&&a++,a!==o&&B(n.charAt(o-1))&&(t.after?o--:r.stripAtStart=!0),a===o?r.remove=!0:e.value=n.slice(a,o),r}(e,t);"nowrap"===t.whitespace&&(e.value=t.collapse(e.value))}return{ignore:P(e),stripAtStart:!1,remove:!1}}function M(e,t,n){for(;++t<e.length;){const n=e[t];let r=$(n);if(void 0===r&&"children"in n&&!L(n)&&(r=M(n.children,-1)),"boolean"==typeof r)return r}return n}function $(e){if("element"===e.type){if(N(e))return!1;if(I(e))return!0}else if("text"===e.type){if(!(0,S.m)(e))return!1}else if(!P(e))return!1}function N(e){return E(e)||(0,f.v)(e,T)}function I(e){return(0,f.v)(e,C)}function L(e){return Boolean("element"===e.type&&e.properties.hidden)||P(e)||(0,f.v)(e,A)}function B(e){return" "===e||"\n"===e}function F(e){const t=/\r?\n|\r/.exec(e);return t?t[0]:" "}function Q(){return" "}function U(e){return function(t){return String(t).replace(/[\t\n\v\f\r ]+/g,e)}}const z={}.hasOwnProperty,j=new Set(["pingback","prefetch","stylesheet"]),W=(0,f.a)(["a","abbr","area","b","bdi","bdo","br","button","cite","code","data","datalist","del","dfn","em","i","input","ins","kbd","keygen","label","map","mark","meter","noscript","output","progress","q","ruby","s","samp","script","select","small","span","strong","sub","sup","template","textarea","time","u","var","wbr"]),X=(0,f.a)("meta");const q=["pre","script","style","textarea"];var V=n(9656);const H={};function Z(e,t){const n=t||H,r={blanks:n.blanks||[],head:!1,indentInitial:!1!==n.indentInitial,indent:"number"==typeof n.indent?" ".repeat(n.indent):"string"==typeof n.indent?n.indent:"  "};function a(e,t,n){const a=e[e.length-1],o=a&&(0,S.m)(a)?e[e.length-2]:a,i=(G(r,o)&&G(r,n)?"\n\n":"\n")+r.indent.repeat(Math.max(t,0));a&&"text"===a.type?a.value=(0,S.m)(a)?i:a.value+i:e.push({type:"text",value:i})}!function(e){D(e,{collapse:U(({newlines:!0}||R).newlines?F:Q),whitespace:"normal"})}(e),(0,V.VG)(e,(function(e,t){if(!("children"in e))return;if("element"===e.type&&"head"===e.tagName&&(r.head=!0),r.head&&"element"===e.type&&"body"===e.tagName&&(r.head=!1),"element"===e.type&&q.includes(e.tagName))return V._Z;if(0===e.children.length||!K(r,e))return;let n=t.length;r.indentInitial||n--;let o=!1;for(const t of e.children)"comment"!==t.type&&"text"!==t.type||(t.value.includes("\n")&&(o=!0),t.value=t.value.replace(/ *\n/g,"$&"+r.indent.repeat(n)));const i=[];let s;for(const t of e.children)(K(r,t)||o&&!s)&&(a(i,n,t),o=!0),s=t,i.push(t);s&&(o||K(r,s))&&((0,S.m)(s)&&(i.pop(),s=i[i.length-1]),a(i,n-1)),e.children=i}))}function G(e,t){return Boolean(t&&"element"===t.type&&e.blanks.length>0&&e.blanks.includes(t.tagName))}function K(e,t){return"root"===t.type||"element"===t.type&&(e.head||"script"===t.tagName||E(t)||(n=t,!Boolean("text"===n.type||W(n)||E(n)||function(e){if("element"!==e.type||"link"!==e.tagName)return!1;if(e.properties.itemProp)return!0;const t=e.properties.rel;let n=-1;if(!Array.isArray(t)||0===t.length)return!1;for(;++n<t.length;)if(!j.has(String(t[n])))return!1;return!0}(n)||X(n)&&function(e,t){const n="element"===e.type&&z.call(e.properties,t)&&e.properties[t];return null!=n&&!1!==n}(n,"itemProp"))));var n}function Y(e){return function(t){Z(t,e)}}var J=n(9938),ee=n(2282),te=n(7471);class ne{constructor(e,t,n){this.property=e,this.normal=t,n&&(this.space=n)}}function re(e,t){const n={},r={};let a=-1;for(;++a<e.length;)Object.assign(n,e[a].property),Object.assign(r,e[a].normal);return new ne(n,r,t)}function ae(e){return e.toLowerCase()}ne.prototype.property={},ne.prototype.normal={},ne.prototype.space=null;class oe{constructor(e,t){this.property=e,this.attribute=t}}oe.prototype.space=null,oe.prototype.boolean=!1,oe.prototype.booleanish=!1,oe.prototype.overloadedBoolean=!1,oe.prototype.number=!1,oe.prototype.commaSeparated=!1,oe.prototype.spaceSeparated=!1,oe.prototype.commaOrSpaceSeparated=!1,oe.prototype.mustUseProperty=!1,oe.prototype.defined=!1;let ie=0;const se=fe(),le=fe(),ce=fe(),ue=fe(),de=fe(),pe=fe(),he=fe();function fe(){return 2**++ie}const me=Object.keys(r);class ge extends oe{constructor(e,t,n,a){let o=-1;if(super(e,t),ve(this,"space",a),"number"==typeof n)for(;++o<me.length;){const e=me[o];ve(this,me[o],(n&r[e])===r[e])}}}function ve(e,t,n){n&&(e[t]=n)}ge.prototype.defined=!0;const be={}.hasOwnProperty;function ye(e){const t={},n={};let r;for(r in e.properties)if(be.call(e.properties,r)){const a=e.properties[r],o=new ge(r,e.transform(e.attributes||{},r),a,e.space);e.mustUseProperty&&e.mustUseProperty.includes(r)&&(o.mustUseProperty=!0),t[r]=o,n[ae(r)]=r,n[ae(o.attribute)]=r}return new ne(t,n,e.space)}const we=ye({space:"xlink",transform:(e,t)=>"xlink:"+t.slice(5).toLowerCase(),properties:{xLinkActuate:null,xLinkArcRole:null,xLinkHref:null,xLinkRole:null,xLinkShow:null,xLinkTitle:null,xLinkType:null}}),Oe=ye({space:"xml",transform:(e,t)=>"xml:"+t.slice(3).toLowerCase(),properties:{xmlLang:null,xmlBase:null,xmlSpace:null}});function ke(e,t){return t in e?e[t]:t}function xe(e,t){return ke(e,t.toLowerCase())}const Ee=ye({space:"xmlns",attributes:{xmlnsxlink:"xmlns:xlink"},transform:xe,properties:{xmlns:null,xmlnsXLink:null}}),Se=ye({transform:(e,t)=>"role"===t?t:"aria-"+t.slice(4).toLowerCase(),properties:{ariaActiveDescendant:null,ariaAtomic:le,ariaAutoComplete:null,ariaBusy:le,ariaChecked:le,ariaColCount:ue,ariaColIndex:ue,ariaColSpan:ue,ariaControls:de,ariaCurrent:null,ariaDescribedBy:de,ariaDetails:null,ariaDisabled:le,ariaDropEffect:de,ariaErrorMessage:null,ariaExpanded:le,ariaFlowTo:de,ariaGrabbed:le,ariaHasPopup:null,ariaHidden:le,ariaInvalid:null,ariaKeyShortcuts:null,ariaLabel:null,ariaLabelledBy:de,ariaLevel:ue,ariaLive:null,ariaModal:le,ariaMultiLine:le,ariaMultiSelectable:le,ariaOrientation:null,ariaOwns:de,ariaPlaceholder:null,ariaPosInSet:ue,ariaPressed:le,ariaReadOnly:le,ariaRelevant:null,ariaRequired:le,ariaRoleDescription:de,ariaRowCount:ue,ariaRowIndex:ue,ariaRowSpan:ue,ariaSelected:le,ariaSetSize:ue,ariaSort:null,ariaValueMax:ue,ariaValueMin:ue,ariaValueNow:ue,ariaValueText:null,role:null}}),_e=ye({space:"html",attributes:{acceptcharset:"accept-charset",classname:"class",htmlfor:"for",httpequiv:"http-equiv"},transform:xe,mustUseProperty:["checked","multiple","muted","selected"],properties:{abbr:null,accept:pe,acceptCharset:de,accessKey:de,action:null,allow:null,allowFullScreen:se,allowPaymentRequest:se,allowUserMedia:se,alt:null,as:null,async:se,autoCapitalize:null,autoComplete:de,autoFocus:se,autoPlay:se,blocking:de,capture:null,charSet:null,checked:se,cite:null,className:de,cols:ue,colSpan:null,content:null,contentEditable:le,controls:se,controlsList:de,coords:ue|pe,crossOrigin:null,data:null,dateTime:null,decoding:null,default:se,defer:se,dir:null,dirName:null,disabled:se,download:ce,draggable:le,encType:null,enterKeyHint:null,fetchPriority:null,form:null,formAction:null,formEncType:null,formMethod:null,formNoValidate:se,formTarget:null,headers:de,height:ue,hidden:se,high:ue,href:null,hrefLang:null,htmlFor:de,httpEquiv:de,id:null,imageSizes:null,imageSrcSet:null,inert:se,inputMode:null,integrity:null,is:null,isMap:se,itemId:null,itemProp:de,itemRef:de,itemScope:se,itemType:de,kind:null,label:null,lang:null,language:null,list:null,loading:null,loop:se,low:ue,manifest:null,max:null,maxLength:ue,media:null,method:null,min:null,minLength:ue,multiple:se,muted:se,name:null,nonce:null,noModule:se,noValidate:se,onAbort:null,onAfterPrint:null,onAuxClick:null,onBeforeMatch:null,onBeforePrint:null,onBeforeToggle:null,onBeforeUnload:null,onBlur:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onContextLost:null,onContextMenu:null,onContextRestored:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnded:null,onError:null,onFocus:null,onFormData:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLanguageChange:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadEnd:null,onLoadStart:null,onMessage:null,onMessageError:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRejectionHandled:null,onReset:null,onResize:null,onScroll:null,onScrollEnd:null,onSecurityPolicyViolation:null,onSeeked:null,onSeeking:null,onSelect:null,onSlotChange:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnhandledRejection:null,onUnload:null,onVolumeChange:null,onWaiting:null,onWheel:null,open:se,optimum:ue,pattern:null,ping:de,placeholder:null,playsInline:se,popover:null,popoverTarget:null,popoverTargetAction:null,poster:null,preload:null,readOnly:se,referrerPolicy:null,rel:de,required:se,reversed:se,rows:ue,rowSpan:ue,sandbox:de,scope:null,scoped:se,seamless:se,selected:se,shadowRootClonable:se,shadowRootDelegatesFocus:se,shadowRootMode:null,shape:null,size:ue,sizes:null,slot:null,span:ue,spellCheck:le,src:null,srcDoc:null,srcLang:null,srcSet:null,start:ue,step:null,style:null,tabIndex:ue,target:null,title:null,translate:null,type:null,typeMustMatch:se,useMap:null,value:le,width:ue,wrap:null,writingSuggestions:null,align:null,aLink:null,archive:de,axis:null,background:null,bgColor:null,border:ue,borderColor:null,bottomMargin:ue,cellPadding:null,cellSpacing:null,char:null,charOff:null,classId:null,clear:null,code:null,codeBase:null,codeType:null,color:null,compact:se,declare:se,event:null,face:null,frame:null,frameBorder:null,hSpace:ue,leftMargin:ue,link:null,longDesc:null,lowSrc:null,marginHeight:ue,marginWidth:ue,noResize:se,noHref:se,noShade:se,noWrap:se,object:null,profile:null,prompt:null,rev:null,rightMargin:ue,rules:null,scheme:null,scrolling:le,standby:null,summary:null,text:null,topMargin:ue,valueType:null,version:null,vAlign:null,vLink:null,vSpace:ue,allowTransparency:null,autoCorrect:null,autoSave:null,disablePictureInPicture:se,disableRemotePlayback:se,prefix:null,property:null,results:ue,security:null,unselectable:null}}),Ce=ye({space:"svg",attributes:{accentHeight:"accent-height",alignmentBaseline:"alignment-baseline",arabicForm:"arabic-form",baselineShift:"baseline-shift",capHeight:"cap-height",className:"class",clipPath:"clip-path",clipRule:"clip-rule",colorInterpolation:"color-interpolation",colorInterpolationFilters:"color-interpolation-filters",colorProfile:"color-profile",colorRendering:"color-rendering",crossOrigin:"crossorigin",dataType:"datatype",dominantBaseline:"dominant-baseline",enableBackground:"enable-background",fillOpacity:"fill-opacity",fillRule:"fill-rule",floodColor:"flood-color",floodOpacity:"flood-opacity",fontFamily:"font-family",fontSize:"font-size",fontSizeAdjust:"font-size-adjust",fontStretch:"font-stretch",fontStyle:"font-style",fontVariant:"font-variant",fontWeight:"font-weight",glyphName:"glyph-name",glyphOrientationHorizontal:"glyph-orientation-horizontal",glyphOrientationVertical:"glyph-orientation-vertical",hrefLang:"hreflang",horizAdvX:"horiz-adv-x",horizOriginX:"horiz-origin-x",horizOriginY:"horiz-origin-y",imageRendering:"image-rendering",letterSpacing:"letter-spacing",lightingColor:"lighting-color",markerEnd:"marker-end",markerMid:"marker-mid",markerStart:"marker-start",navDown:"nav-down",navDownLeft:"nav-down-left",navDownRight:"nav-down-right",navLeft:"nav-left",navNext:"nav-next",navPrev:"nav-prev",navRight:"nav-right",navUp:"nav-up",navUpLeft:"nav-up-left",navUpRight:"nav-up-right",onAbort:"onabort",onActivate:"onactivate",onAfterPrint:"onafterprint",onBeforePrint:"onbeforeprint",onBegin:"onbegin",onCancel:"oncancel",onCanPlay:"oncanplay",onCanPlayThrough:"oncanplaythrough",onChange:"onchange",onClick:"onclick",onClose:"onclose",onCopy:"oncopy",onCueChange:"oncuechange",onCut:"oncut",onDblClick:"ondblclick",onDrag:"ondrag",onDragEnd:"ondragend",onDragEnter:"ondragenter",onDragExit:"ondragexit",onDragLeave:"ondragleave",onDragOver:"ondragover",onDragStart:"ondragstart",onDrop:"ondrop",onDurationChange:"ondurationchange",onEmptied:"onemptied",onEnd:"onend",onEnded:"onended",onError:"onerror",onFocus:"onfocus",onFocusIn:"onfocusin",onFocusOut:"onfocusout",onHashChange:"onhashchange",onInput:"oninput",onInvalid:"oninvalid",onKeyDown:"onkeydown",onKeyPress:"onkeypress",onKeyUp:"onkeyup",onLoad:"onload",onLoadedData:"onloadeddata",onLoadedMetadata:"onloadedmetadata",onLoadStart:"onloadstart",onMessage:"onmessage",onMouseDown:"onmousedown",onMouseEnter:"onmouseenter",onMouseLeave:"onmouseleave",onMouseMove:"onmousemove",onMouseOut:"onmouseout",onMouseOver:"onmouseover",onMouseUp:"onmouseup",onMouseWheel:"onmousewheel",onOffline:"onoffline",onOnline:"ononline",onPageHide:"onpagehide",onPageShow:"onpageshow",onPaste:"onpaste",onPause:"onpause",onPlay:"onplay",onPlaying:"onplaying",onPopState:"onpopstate",onProgress:"onprogress",onRateChange:"onratechange",onRepeat:"onrepeat",onReset:"onreset",onResize:"onresize",onScroll:"onscroll",onSeeked:"onseeked",onSeeking:"onseeking",onSelect:"onselect",onShow:"onshow",onStalled:"onstalled",onStorage:"onstorage",onSubmit:"onsubmit",onSuspend:"onsuspend",onTimeUpdate:"ontimeupdate",onToggle:"ontoggle",onUnload:"onunload",onVolumeChange:"onvolumechange",onWaiting:"onwaiting",onZoom:"onzoom",overlinePosition:"overline-position",overlineThickness:"overline-thickness",paintOrder:"paint-order",panose1:"panose-1",pointerEvents:"pointer-events",referrerPolicy:"referrerpolicy",renderingIntent:"rendering-intent",shapeRendering:"shape-rendering",stopColor:"stop-color",stopOpacity:"stop-opacity",strikethroughPosition:"strikethrough-position",strikethroughThickness:"strikethrough-thickness",strokeDashArray:"stroke-dasharray",strokeDashOffset:"stroke-dashoffset",strokeLineCap:"stroke-linecap",strokeLineJoin:"stroke-linejoin",strokeMiterLimit:"stroke-miterlimit",strokeOpacity:"stroke-opacity",strokeWidth:"stroke-width",tabIndex:"tabindex",textAnchor:"text-anchor",textDecoration:"text-decoration",textRendering:"text-rendering",transformOrigin:"transform-origin",typeOf:"typeof",underlinePosition:"underline-position",underlineThickness:"underline-thickness",unicodeBidi:"unicode-bidi",unicodeRange:"unicode-range",unitsPerEm:"units-per-em",vAlphabetic:"v-alphabetic",vHanging:"v-hanging",vIdeographic:"v-ideographic",vMathematical:"v-mathematical",vectorEffect:"vector-effect",vertAdvY:"vert-adv-y",vertOriginX:"vert-origin-x",vertOriginY:"vert-origin-y",wordSpacing:"word-spacing",writingMode:"writing-mode",xHeight:"x-height",playbackOrder:"playbackorder",timelineBegin:"timelinebegin"},transform:ke,properties:{about:he,accentHeight:ue,accumulate:null,additive:null,alignmentBaseline:null,alphabetic:ue,amplitude:ue,arabicForm:null,ascent:ue,attributeName:null,attributeType:null,azimuth:ue,bandwidth:null,baselineShift:null,baseFrequency:null,baseProfile:null,bbox:null,begin:null,bias:ue,by:null,calcMode:null,capHeight:ue,className:de,clip:null,clipPath:null,clipPathUnits:null,clipRule:null,color:null,colorInterpolation:null,colorInterpolationFilters:null,colorProfile:null,colorRendering:null,content:null,contentScriptType:null,contentStyleType:null,crossOrigin:null,cursor:null,cx:null,cy:null,d:null,dataType:null,defaultAction:null,descent:ue,diffuseConstant:ue,direction:null,display:null,dur:null,divisor:ue,dominantBaseline:null,download:se,dx:null,dy:null,edgeMode:null,editable:null,elevation:ue,enableBackground:null,end:null,event:null,exponent:ue,externalResourcesRequired:null,fill:null,fillOpacity:ue,fillRule:null,filter:null,filterRes:null,filterUnits:null,floodColor:null,floodOpacity:null,focusable:null,focusHighlight:null,fontFamily:null,fontSize:null,fontSizeAdjust:null,fontStretch:null,fontStyle:null,fontVariant:null,fontWeight:null,format:null,fr:null,from:null,fx:null,fy:null,g1:pe,g2:pe,glyphName:pe,glyphOrientationHorizontal:null,glyphOrientationVertical:null,glyphRef:null,gradientTransform:null,gradientUnits:null,handler:null,hanging:ue,hatchContentUnits:null,hatchUnits:null,height:null,href:null,hrefLang:null,horizAdvX:ue,horizOriginX:ue,horizOriginY:ue,id:null,ideographic:ue,imageRendering:null,initialVisibility:null,in:null,in2:null,intercept:ue,k:ue,k1:ue,k2:ue,k3:ue,k4:ue,kernelMatrix:he,kernelUnitLength:null,keyPoints:null,keySplines:null,keyTimes:null,kerning:null,lang:null,lengthAdjust:null,letterSpacing:null,lightingColor:null,limitingConeAngle:ue,local:null,markerEnd:null,markerMid:null,markerStart:null,markerHeight:null,markerUnits:null,markerWidth:null,mask:null,maskContentUnits:null,maskUnits:null,mathematical:null,max:null,media:null,mediaCharacterEncoding:null,mediaContentEncodings:null,mediaSize:ue,mediaTime:null,method:null,min:null,mode:null,name:null,navDown:null,navDownLeft:null,navDownRight:null,navLeft:null,navNext:null,navPrev:null,navRight:null,navUp:null,navUpLeft:null,navUpRight:null,numOctaves:null,observer:null,offset:null,onAbort:null,onActivate:null,onAfterPrint:null,onBeforePrint:null,onBegin:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnd:null,onEnded:null,onError:null,onFocus:null,onFocusIn:null,onFocusOut:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadStart:null,onMessage:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onMouseWheel:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRepeat:null,onReset:null,onResize:null,onScroll:null,onSeeked:null,onSeeking:null,onSelect:null,onShow:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnload:null,onVolumeChange:null,onWaiting:null,onZoom:null,opacity:null,operator:null,order:null,orient:null,orientation:null,origin:null,overflow:null,overlay:null,overlinePosition:ue,overlineThickness:ue,paintOrder:null,panose1:null,path:null,pathLength:ue,patternContentUnits:null,patternTransform:null,patternUnits:null,phase:null,ping:de,pitch:null,playbackOrder:null,pointerEvents:null,points:null,pointsAtX:ue,pointsAtY:ue,pointsAtZ:ue,preserveAlpha:null,preserveAspectRatio:null,primitiveUnits:null,propagate:null,property:he,r:null,radius:null,referrerPolicy:null,refX:null,refY:null,rel:he,rev:he,renderingIntent:null,repeatCount:null,repeatDur:null,requiredExtensions:he,requiredFeatures:he,requiredFonts:he,requiredFormats:he,resource:null,restart:null,result:null,rotate:null,rx:null,ry:null,scale:null,seed:null,shapeRendering:null,side:null,slope:null,snapshotTime:null,specularConstant:ue,specularExponent:ue,spreadMethod:null,spacing:null,startOffset:null,stdDeviation:null,stemh:null,stemv:null,stitchTiles:null,stopColor:null,stopOpacity:null,strikethroughPosition:ue,strikethroughThickness:ue,string:null,stroke:null,strokeDashArray:he,strokeDashOffset:null,strokeLineCap:null,strokeLineJoin:null,strokeMiterLimit:ue,strokeOpacity:ue,strokeWidth:null,style:null,surfaceScale:ue,syncBehavior:null,syncBehaviorDefault:null,syncMaster:null,syncTolerance:null,syncToleranceDefault:null,systemLanguage:he,tabIndex:ue,tableValues:null,target:null,targetX:ue,targetY:ue,textAnchor:null,textDecoration:null,textRendering:null,textLength:null,timelineBegin:null,title:null,transformBehavior:null,type:null,typeOf:he,to:null,transform:null,transformOrigin:null,u1:null,u2:null,underlinePosition:ue,underlineThickness:ue,unicode:null,unicodeBidi:null,unicodeRange:null,unitsPerEm:ue,values:null,vAlphabetic:ue,vMathematical:ue,vectorEffect:null,vHanging:ue,vIdeographic:ue,version:null,vertAdvY:ue,vertOriginX:ue,vertOriginY:ue,viewBox:null,viewTarget:null,visibility:null,width:null,widths:null,wordSpacing:null,writingMode:null,x:null,x1:null,x2:null,xChannelSelector:null,xHeight:ue,y:null,y1:null,y2:null,yChannelSelector:null,z:null,zoomAndPan:null}}),Te=re([Oe,we,Ee,Se,_e],"html"),Ae=re([Oe,we,Ee,Se,Ce],"svg"),Re=/^data[-\w.:]+$/i,Pe=/-[a-z]/g,De=/[A-Z]/g;function Me(e){return"-"+e.toLowerCase()}function $e(e){return e.charAt(1).toUpperCase()}var Ne=n(4565),Ie=n(8845);const Le={},Be={}.hasOwnProperty,Fe=(0,Ie.A)("type",{handlers:{root:function(e,t){const n={nodeName:"#document",mode:(e.data||{}).quirksMode?"quirks":"no-quirks",childNodes:[]};return n.childNodes=Ue(e.children,n,t),ze(e,n),n},element:function(e,t){const n=t;let r=n;"element"===e.type&&"svg"===e.tagName.toLowerCase()&&"html"===n.space&&(r=Ae);const a=[];let o;if(e.properties)for(o in e.properties)if("children"!==o&&Be.call(e.properties,o)){const t=Qe(r,o,e.properties[o]);t&&a.push(t)}const i=r.space;(0,te.ok)(i);const s={nodeName:e.tagName,tagName:e.tagName,attrs:a,namespaceURI:Ne.t[i],childNodes:[],parentNode:null};return s.childNodes=Ue(e.children,s,r),ze(e,s),"template"===e.tagName&&e.content&&(s.content=function(e,t){const n={nodeName:"#document-fragment",childNodes:[]};return n.childNodes=Ue(e.children,n,t),ze(e,n),n}(e.content,r)),s},text:function(e){const t={nodeName:"#text",value:e.value,parentNode:null};return ze(e,t),t},comment:function(e){const t={nodeName:"#comment",data:e.value,parentNode:null};return ze(e,t),t},doctype:function(e){const t={nodeName:"#documentType",name:"html",publicId:"",systemId:"",parentNode:null};return ze(e,t),t}}});function Qe(e,t,n){const r=function(e,t){const n=ae(t);let r=t,a=oe;if(n in e.normal)return e.property[e.normal[n]];if(n.length>4&&"data"===n.slice(0,4)&&Re.test(t)){if("-"===t.charAt(4)){const e=t.slice(5).replace(Pe,$e);r="data"+e.charAt(0).toUpperCase()+e.slice(1)}else{const e=t.slice(4);if(!Pe.test(e)){let n=e.replace(De,Me);"-"!==n.charAt(0)&&(n="-"+n),t="data"+n}}a=ge}return new a(r,t)}(e,t);if(!1===n||null==n||"number"==typeof n&&Number.isNaN(n)||!n&&r.boolean)return;Array.isArray(n)&&(n=r.commaSeparated?(0,ee.A)(n):(0,v.A)(n));const a={name:r.attribute,value:!0===n?"":String(n)};if(r.space&&"html"!==r.space&&"svg"!==r.space){const e=a.name.indexOf(":");e<0?a.prefix="":(a.name=a.name.slice(e+1),a.prefix=r.attribute.slice(0,e)),a.namespace=Ne.t[r.space]}return a}function Ue(e,t,n){let r=-1;const a=[];if(e)for(;++r<e.length;){const o=Fe(e[r],n);o.parentNode=t,a.push(o)}return a}function ze(e,t){const n=e.position;n&&n.start&&n.end&&((0,te.ok)("number"==typeof n.start.offset),(0,te.ok)("number"==typeof n.end.offset),t.sourceCodeLocation={startLine:n.start.line,startCol:n.start.column,startOffset:n.start.offset,endLine:n.end.line,endCol:n.end.column,endOffset:n.end.offset})}var je=n(6932),We=n(4684);const Xe=Ve("end"),qe=Ve("start");function Ve(e){return function(t){const n=t&&t.position&&t.position[e]||{};if("number"==typeof n.line&&n.line>0&&"number"==typeof n.column&&n.column>0)return{line:n.line,column:n.column,offset:"number"==typeof n.offset&&n.offset>-1?n.offset:void 0}}}function He(e){const t=qe(e),n=Xe(e);if(t&&n)return{start:t,end:n}}const Ze=/<(\/?)(iframe|noembed|noframes|plaintext|script|style|textarea|title|xmp)(?=[\t\n\f\r />])/gi,Ge=new Set(["mdxFlowExpression","mdxJsxFlowElement","mdxJsxTextElement","mdxTextExpression","mdxjsEsm"]),Ke={sourceCodeLocationInfo:!0,scriptingEnabled:!1};function Ye(e,t){const n=function(e){const t="root"===e.type?e.children[0]:e;return Boolean(t&&("doctype"===t.type||"element"===t.type&&"html"===t.tagName.toLowerCase()))}(e),r=(0,Ie.A)("type",{handlers:{root:et,element:tt,text:nt,comment:at,doctype:rt,raw:ot},unknown:it}),a={parser:n?new We.iX(Ke):We.iX.getFragmentParser(void 0,Ke),handle(e){r(e,a)},stitches:!1,options:t||{}};r(e,a),st(a,qe());const o=n?a.parser.document:a.parser.getFragment(),i=(0,J.D)(o,{file:a.options.file});return a.stitches&&(0,b.YR)(i,"comment",(function(e,t,n){const r=e;if(r.value.stitch&&n&&void 0!==t)return n.children[t]=r.value.stitch,t})),"root"===i.type&&1===i.children.length&&i.children[0].type===e.type?i.children[0]:i}function Je(e,t){let n=-1;if(e)for(;++n<e.length;)t.handle(e[n])}function et(e,t){Je(e.children,t)}function tt(e,t){!function(e,t){const n=e.tagName.toLowerCase();if(t.parser.tokenizer.state===We.V_.PLAINTEXT)return;st(t,qe(e));const r=t.parser.openElements.current;let a="namespaceURI"in r?r.namespaceURI:Ne.t.html;a===Ne.t.html&&"svg"===n&&(a=Ne.t.svg);const o=function(e,t){const n=(t||Le).space;return Fe(e,"svg"===n?Ae:Te)}({...e,children:[]},{space:a===Ne.t.svg?"svg":"html"}),i={type:We.ou.k.START_TAG,tagName:n,tagID:We.qy.getTagID(n),selfClosing:!1,ackSelfClosing:!1,attrs:"attrs"in o?o.attrs:[],location:ct(e)};t.parser.currentToken=i,t.parser._processToken(t.parser.currentToken),t.parser.tokenizer.lastStartTagName=n}(e,t),Je(e.children,t),function(e,t){const n=e.tagName.toLowerCase();if(!t.parser.tokenizer.inForeignNode&&je.b.includes(n))return;if(t.parser.tokenizer.state===We.V_.PLAINTEXT)return;st(t,Xe(e));const r={type:We.ou.k.END_TAG,tagName:n,tagID:We.qy.getTagID(n),selfClosing:!1,ackSelfClosing:!1,attrs:[],location:ct(e)};t.parser.currentToken=r,t.parser._processToken(t.parser.currentToken),n!==t.parser.tokenizer.lastStartTagName||t.parser.tokenizer.state!==We.V_.RCDATA&&t.parser.tokenizer.state!==We.V_.RAWTEXT&&t.parser.tokenizer.state!==We.V_.SCRIPT_DATA||(t.parser.tokenizer.state=We.V_.DATA)}(e,t)}function nt(e,t){t.parser.tokenizer.state>4&&(t.parser.tokenizer.state=0);const n={type:We.ou.k.CHARACTER,chars:e.value,location:ct(e)};st(t,qe(e)),t.parser.currentToken=n,t.parser._processToken(t.parser.currentToken)}function rt(e,t){const n={type:We.ou.k.DOCTYPE,name:"html",forceQuirks:!1,publicId:"",systemId:"",location:ct(e)};st(t,qe(e)),t.parser.currentToken=n,t.parser._processToken(t.parser.currentToken)}function at(e,t){const n=e.value,r={type:We.ou.k.COMMENT,data:n,location:ct(e)};st(t,qe(e)),t.parser.currentToken=r,t.parser._processToken(t.parser.currentToken)}function ot(e,t){if(t.parser.tokenizer.preprocessor.html="",t.parser.tokenizer.preprocessor.pos=-1,t.parser.tokenizer.preprocessor.lastGapPos=-2,t.parser.tokenizer.preprocessor.gapStack=[],t.parser.tokenizer.preprocessor.skipNextNewLine=!1,t.parser.tokenizer.preprocessor.lastChunkWritten=!1,t.parser.tokenizer.preprocessor.endOfChunkHit=!1,t.parser.tokenizer.preprocessor.isEol=!1,lt(t,qe(e)),t.parser.tokenizer.write(t.options.tagfilter?e.value.replace(Ze,"&lt;$1$2"):e.value,!1),t.parser.tokenizer._runParsingLoop(),72===t.parser.tokenizer.state||78===t.parser.tokenizer.state){t.parser.tokenizer.preprocessor.lastChunkWritten=!0;const e=t.parser.tokenizer._consume();t.parser.tokenizer._callState(e)}}function it(e,t){const n=e;if(!t.options.passThrough||!t.options.passThrough.includes(n.type)){let e="";throw Ge.has(n.type)&&(e=". It looks like you are using MDX nodes with `hast-util-raw` (or `rehype-raw`). If you use this because you are using remark or rehype plugins that inject `'html'` nodes, then please raise an issue with that plugin, as its a bad and slow idea. If you use this because you are using markdown syntax, then you have to configure this utility (or plugin) to pass through these nodes (see `passThrough` in docs), but you can also migrate to use the MDX syntax"),new Error("Cannot compile `"+n.type+"` node"+e)}!function(e,t){t.stitches=!0;const n=function(e){return h("children"in e?{...e,children:[]}:e)}(e);if("children"in e&&"children"in n){const r=Ye({type:"root",children:e.children},t.options);n.children=r.children}at({type:"comment",value:{stitch:n}},t)}(n,t)}function st(e,t){lt(e,t);const n=e.parser.tokenizer.currentCharacterToken;n&&n.location&&(n.location.endLine=e.parser.tokenizer.preprocessor.line,n.location.endCol=e.parser.tokenizer.preprocessor.col+1,n.location.endOffset=e.parser.tokenizer.preprocessor.offset+1,e.parser.currentToken=n,e.parser._processToken(e.parser.currentToken)),e.parser.tokenizer.paused=!1,e.parser.tokenizer.inLoop=!1,e.parser.tokenizer.active=!1,e.parser.tokenizer.returnState=We.V_.DATA,e.parser.tokenizer.charRefCode=-1,e.parser.tokenizer.consumedAfterSnapshot=-1,e.parser.tokenizer.currentLocation=null,e.parser.tokenizer.currentCharacterToken=null,e.parser.tokenizer.currentToken=null,e.parser.tokenizer.currentAttr={name:"",value:""}}function lt(e,t){if(t&&void 0!==t.offset){const n={startLine:t.line,startCol:t.column,startOffset:t.offset,endLine:-1,endCol:-1,endOffset:-1};e.parser.tokenizer.preprocessor.lineStartPos=1-t.column,e.parser.tokenizer.preprocessor.droppedBufferSize=t.offset,e.parser.tokenizer.preprocessor.line=t.line,e.parser.tokenizer.currentLocation=n}}function ct(e){const t=qe(e)||{line:void 0,column:void 0,offset:void 0},n=Xe(e)||{line:void 0,column:void 0,offset:void 0};return{startLine:t.line,startCol:t.column,startOffset:t.offset,endLine:n.line,endCol:n.column,endOffset:n.offset}}function ut(e){return function(t,n){return Ye(t,{...e,file:n})}}const dt=["ariaDescribedBy","ariaLabel","ariaLabelledBy"],pt={ancestors:{tbody:["table"],td:["table"],th:["table"],thead:["table"],tfoot:["table"],tr:["table"]},attributes:{a:[...dt,"dataFootnoteBackref","dataFootnoteRef",["className","data-footnote-backref"],"href"],blockquote:["cite"],code:[["className",/^language-./]],del:["cite"],div:["itemScope","itemType"],dl:[...dt],h2:[["className","sr-only"]],img:[...dt,"longDesc","src"],input:[["disabled",!0],["type","checkbox"]],ins:["cite"],li:[["className","task-list-item"]],ol:[...dt,["className","contains-task-list"]],q:["cite"],section:["dataFootnotes",["className","footnotes"]],source:["srcSet"],summary:[...dt],table:[...dt],ul:[...dt,["className","contains-task-list"]],"*":["abbr","accept","acceptCharset","accessKey","action","align","alt","axis","border","cellPadding","cellSpacing","char","charOff","charSet","checked","clear","colSpan","color","cols","compact","coords","dateTime","dir","encType","frame","hSpace","headers","height","hrefLang","htmlFor","id","isMap","itemProp","label","lang","maxLength","media","method","multiple","name","noHref","noShade","noWrap","open","prompt","readOnly","rev","rowSpan","rows","rules","scope","selected","shape","size","span","start","summary","tabIndex","title","useMap","vAlign","value","width"]},clobber:["ariaDescribedBy","ariaLabelledBy","id","name"],clobberPrefix:"user-content-",protocols:{cite:["http","https"],href:["http","https","irc","ircs","mailto","xmpp"],longDesc:["http","https"],src:["http","https"]},required:{input:{disabled:!0,type:"checkbox"}},strip:["script"],tagNames:["a","b","blockquote","br","code","dd","del","details","div","dl","dt","em","h1","h2","h3","h4","h5","h6","hr","i","img","input","ins","kbd","li","ol","p","picture","pre","q","rp","rt","ruby","s","samp","section","source","span","strike","strong","sub","summary","sup","table","tbody","td","tfoot","th","thead","tr","tt","ul","var"]},ht={}.hasOwnProperty;function ft(e,t){if(t&&"object"==typeof t){const n=t;switch("string"==typeof n.type?n.type:""){case"comment":return function(e,t){if(e.schema.allowComments){const e="string"==typeof t.value?t.value:"",n=e.indexOf("--\x3e"),r={type:"comment",value:n<0?e:e.slice(0,n)};return bt(r,t),r}}(e,n);case"doctype":return function(e,t){if(e.schema.allowDoctypes){const e={type:"doctype"};return bt(e,t),e}}(e,n);case"element":return function(e,t){const n="string"==typeof t.tagName?t.tagName:"";e.stack.push(n);const r=mt(e,t.children),a=function(e,t){const n=e.stack[e.stack.length-1],r=e.schema.attributes,a=e.schema.required,o=r&&ht.call(r,n)?r[n]:void 0,i=r&&ht.call(r,"*")?r["*"]:void 0,s=t&&"object"==typeof t?t:{},l={};let c;for(c in s)if(ht.call(s,c)){const t=s[c];let n=gt(e,yt(o,c),c,t);null==n&&(n=gt(e,yt(i,c),c,t)),null!=n&&(l[c]=n)}if(a&&ht.call(a,n)){const e=a[n];for(c in e)ht.call(e,c)&&!ht.call(l,c)&&(l[c]=e[c])}return l}(e,t.properties);e.stack.pop();let o=!1;if(n&&"*"!==n&&(!e.schema.tagNames||e.schema.tagNames.includes(n))&&(o=!0,e.schema.ancestors&&ht.call(e.schema.ancestors,n))){const t=e.schema.ancestors[n];let r=-1;for(o=!1;++r<t.length;)e.stack.includes(t[r])&&(o=!0)}if(!o)return e.schema.strip&&!e.schema.strip.includes(n)?r:void 0;const i={type:"element",tagName:n,properties:a,children:r};return bt(i,t),i}(e,n);case"root":return function(e,t){const n={type:"root",children:mt(e,t.children)};return bt(n,t),n}(e,n);case"text":return function(e,t){const n={type:"text",value:"string"==typeof t.value?t.value:""};return bt(n,t),n}(0,n)}}}function mt(e,t){const n=[];if(Array.isArray(t)){const r=t;let a=-1;for(;++a<r.length;){const t=ft(e,r[a]);t&&(Array.isArray(t)?n.push(...t):n.push(t))}}return n}function gt(e,t,n,r){return t?Array.isArray(r)?function(e,t,n,r){let a=-1;const o=[];for(;++a<r.length;){const i=vt(e,t,n,r[a]);"number"!=typeof i&&"string"!=typeof i||o.push(i)}return o}(e,t,n,r):vt(e,t,n,r):void 0}function vt(e,t,n,r){if(("boolean"==typeof r||"number"==typeof r||"string"==typeof r)&&function(e,t,n){const r=e.schema.protocols&&ht.call(e.schema.protocols,t)?e.schema.protocols[t]:void 0;if(!r||0===r.length)return!0;const a=String(n),o=a.indexOf(":"),i=a.indexOf("?"),s=a.indexOf("#"),l=a.indexOf("/");if(o<0||l>-1&&o>l||i>-1&&o>i||s>-1&&o>s)return!0;let c=-1;for(;++c<r.length;){const e=r[c];if(o===e.length&&a.slice(0,e.length)===e)return!0}return!1}(e,n,r)){if("object"==typeof t&&t.length>1){let e=!1,n=0;for(;++n<t.length;){const a=t[n];if(a&&"object"==typeof a&&"flags"in a){if(a.test(String(r))){e=!0;break}}else if(a===r){e=!0;break}}if(!e)return}return e.schema.clobber&&e.schema.clobberPrefix&&e.schema.clobber.includes(n)?e.schema.clobberPrefix+r:r}}function bt(e,t){const n=He(t);t.data&&(e.data=h(t.data)),n&&(e.position=n)}function yt(e,t){let n,r=-1;if(e)for(;++r<e.length;){const a=e[r],o="string"==typeof a?a:a[0];if(o===t)return a;"data*"===o&&(n=a)}if(t.length>4&&"data"===t.slice(0,4).toLowerCase())return n}function wt(e){return function(t){const n=function(e,t){let n={type:"root",children:[]};const r=ft({schema:t?{...pt,...t}:pt,stack:[]},e);return r&&(Array.isArray(r)?1===r.length?n=r[0]:n.children=r:n=r),n}(t,e);return n}}var Ot=n(8697),kt=n(8256);const xt=Nt(/[A-Za-z]/),Et=Nt(/[\dA-Za-z]/),St=Nt(/[#-'*+\--9=?A-Z^-~]/);function _t(e){return null!==e&&(e<32||127===e)}const Ct=Nt(/\d/),Tt=Nt(/[\dA-Fa-f]/),At=Nt(/[!-/:-@[-`{-~]/);function Rt(e){return null!==e&&e<-2}function Pt(e){return null!==e&&(e<0||32===e)}function Dt(e){return-2===e||-1===e||32===e}const Mt=Nt(/\p{P}|\p{S}/u),$t=Nt(/\s/);function Nt(e){return function(t){return null!==t&&t>-1&&e.test(String.fromCharCode(t))}}function It(e){return"function"==typeof e?e:function(){return e}}const Lt="phrasing",Bt=["autolink","link","image","label"];function Ft(e){this.enter({type:"link",title:null,url:"",children:[]},e)}function Qt(e){this.config.enter.autolinkProtocol.call(this,e)}function Ut(e){this.config.exit.autolinkProtocol.call(this,e)}function zt(e){this.config.exit.data.call(this,e);const t=this.stack[this.stack.length-1];(0,te.ok)("link"===t.type),t.url="http://"+this.sliceSerialize(e)}function jt(e){this.config.exit.autolinkEmail.call(this,e)}function Wt(e){this.exit(e)}function Xt(e){!function(e,t,n){const r=n||{},a=(0,_.C)(r.ignore||[]),o=function(e){const t=[];if(!Array.isArray(e))throw new TypeError("Expected find and replace tuple or list of tuples");const n=!e[0]||Array.isArray(e[0])?e:[e];let r=-1;for(;++r<n.length;){const e=n[r];t.push([(a=e[0],"string"==typeof a?new RegExp(function(e){if("string"!=typeof e)throw new TypeError("Expected a string");return e.replace(/[|\\{}()[\]^$+*?.]/g,"\\$&").replace(/-/g,"\\x2d")}(a),"g"):a),It(e[1])])}var a;return t}(t);let i=-1;for(;++i<o.length;)(0,V.VG)(e,"text",s);function s(e,t){let n,r=-1;for(;++r<t.length;){const e=t[r],o=n?n.children:void 0;if(a(e,o?o.indexOf(e):void 0,n))return;n=e}if(n)return function(e,t){const n=t[t.length-1],r=o[i][0],a=o[i][1];let s=0;const l=n.children.indexOf(e);let c=!1,u=[];r.lastIndex=0;let d=r.exec(e.value);for(;d;){const n=d.index,o={index:d.index,input:d.input,stack:[...t,e]};let i=a(...d,o);if("string"==typeof i&&(i=i.length>0?{type:"text",value:i}:void 0),!1===i?r.lastIndex=n+1:(s!==n&&u.push({type:"text",value:e.value.slice(s,n)}),Array.isArray(i)?u.push(...i):i&&u.push(i),s=n+d[0].length,c=!0),!r.global)break;d=r.exec(e.value)}return c?(s<e.value.length&&u.push({type:"text",value:e.value.slice(s)}),n.children.splice(l,1,...u)):u=[e],l+u.length}(e,t)}}(e,[[/(https?:\/\/|www(?=\.))([-.\w]+)([^ \t\r\n]*)/gi,qt],[/(?<=^|\s|\p{P}|\p{S})([-.\w+]+)@([-\w]+(?:\.[-\w]+)+)/gu,Vt]],{ignore:["link","linkReference"]})}function qt(e,t,n,r,a){let o="";if(!Ht(a))return!1;if(/^w/i.test(t)&&(n=t+n,t="",o="http://"),!function(e){const t=e.split(".");return!(t.length<2||t[t.length-1]&&(/_/.test(t[t.length-1])||!/[a-zA-Z\d]/.test(t[t.length-1]))||t[t.length-2]&&(/_/.test(t[t.length-2])||!/[a-zA-Z\d]/.test(t[t.length-2])))}(n))return!1;const i=function(e){const t=/[!"&'),.:;<>?\]}]+$/.exec(e);if(!t)return[e,void 0];e=e.slice(0,t.index);let n=t[0],r=n.indexOf(")");const a=(0,kt.D)(e,"(");let o=(0,kt.D)(e,")");for(;-1!==r&&a>o;)e+=n.slice(0,r+1),n=n.slice(r+1),r=n.indexOf(")"),o++;return[e,n]}(n+r);if(!i[0])return!1;const s={type:"link",title:null,url:o+t+i[0],children:[{type:"text",value:t+i[0]}]};return i[1]?[s,{type:"text",value:i[1]}]:s}function Vt(e,t,n,r){return!(!Ht(r,!0)||/[-\d_]$/.test(n))&&{type:"link",title:null,url:"mailto:"+t+"@"+n,children:[{type:"text",value:t+"@"+n}]}}function Ht(e,t){const n=e.input.charCodeAt(e.index-1);return(0===e.index||$t(n)||Mt(n))&&(!t||47!==n)}function Zt(e){return e.replace(/[\t\n\r ]+/g," ").replace(/^ | $/g,"").toLowerCase().toUpperCase()}function Gt(){this.buffer()}function Kt(e){this.enter({type:"footnoteReference",identifier:"",label:""},e)}function Yt(){this.buffer()}function Jt(e){this.enter({type:"footnoteDefinition",identifier:"",label:"",children:[]},e)}function en(e){const t=this.resume(),n=this.stack[this.stack.length-1];(0,te.ok)("footnoteReference"===n.type),n.identifier=Zt(this.sliceSerialize(e)).toLowerCase(),n.label=t}function tn(e){this.exit(e)}function nn(e){const t=this.resume(),n=this.stack[this.stack.length-1];(0,te.ok)("footnoteDefinition"===n.type),n.identifier=Zt(this.sliceSerialize(e)).toLowerCase(),n.label=t}function rn(e){this.exit(e)}function an(e,t,n,r){const a=n.createTracker(r);let o=a.move("[^");const i=n.enter("footnoteReference"),s=n.enter("reference");return o+=a.move(n.safe(n.associationId(e),{after:"]",before:o})),s(),i(),o+=a.move("]"),o}function on(e){let t=!1;return e&&e.firstLineBlank&&(t=!0),{handlers:{footnoteDefinition:function(e,n,r,a){const o=r.createTracker(a);let i=o.move("[^");const s=r.enter("footnoteDefinition"),l=r.enter("label");return i+=o.move(r.safe(r.associationId(e),{before:i,after:"]"})),l(),i+=o.move("]:"),e.children&&e.children.length>0&&(o.shift(4),i+=o.move((t?"\n":" ")+r.indentLines(r.containerFlow(e,o.current()),t?ln:sn))),s(),i},footnoteReference:an},unsafe:[{character:"[",inConstruct:["label","phrasing","reference"]}]}}function sn(e,t,n){return 0===t?e:ln(e,0,n)}function ln(e,t,n){return(n?"":"    ")+e}an.peek=function(){return"["};const cn=["autolink","destinationLiteral","destinationRaw","reference","titleQuote","titleApostrophe"];function un(e){this.enter({type:"delete",children:[]},e)}function dn(e){this.exit(e)}function pn(e,t,n,r){const a=n.createTracker(r),o=n.enter("strikethrough");let i=a.move("~~");return i+=n.containerPhrasing(e,{...a.current(),before:i,after:"~"}),i+=a.move("~~"),o(),i}function hn(e){return e.length}function fn(e){const t="string"==typeof e?e.codePointAt(0):0;return 67===t||99===t?99:76===t||108===t?108:82===t||114===t?114:0}function mn(e,t,n){return">"+(n?"":" ")+e}function gn(e,t){return vn(e,t.inConstruct,!0)&&!vn(e,t.notInConstruct,!1)}function vn(e,t,n){if("string"==typeof t&&(t=[t]),!t||0===t.length)return n;let r=-1;for(;++r<t.length;)if(e.includes(t[r]))return!0;return!1}function bn(e,t,n,r){let a=-1;for(;++a<n.unsafe.length;)if("\n"===n.unsafe[a].character&&gn(n.stack,n.unsafe[a]))return/[ \t]/.test(r.before)?"":" ";return"\\\n"}function yn(e,t){return Boolean(!1===t.options.fences&&e.value&&!e.lang&&/[^ \r\n]/.test(e.value)&&!/^[\t ]*(?:[\r\n]|$)|(?:^|[\r\n])[\t ]*$/.test(e.value))}function wn(e,t,n){return(n?"":"    ")+e}function On(e){const t=e.options.quote||'"';if('"'!==t&&"'"!==t)throw new Error("Cannot serialize title with `"+t+"` for `options.quote`, expected `\"`, or `'`");return t}function kn(e){return"&#x"+e.toString(16).toUpperCase()+";"}function xn(e){return null===e||Pt(e)||$t(e)?1:Mt(e)?2:void 0}function En(e,t,n){const r=xn(e),a=xn(t);return void 0===r?void 0===a?"_"===n?{inside:!0,outside:!0}:{inside:!1,outside:!1}:1===a?{inside:!0,outside:!0}:{inside:!1,outside:!0}:1===r?void 0===a?{inside:!1,outside:!1}:1===a?{inside:!0,outside:!0}:{inside:!1,outside:!1}:void 0===a?{inside:!1,outside:!1}:1===a?{inside:!0,outside:!1}:{inside:!1,outside:!1}}function Sn(e,t,n,r){const a=function(e){const t=e.options.emphasis||"*";if("*"!==t&&"_"!==t)throw new Error("Cannot serialize emphasis with `"+t+"` for `options.emphasis`, expected `*`, or `_`");return t}(n),o=n.enter("emphasis"),i=n.createTracker(r),s=i.move(a);let l=i.move(n.containerPhrasing(e,{after:a,before:s,...i.current()}));const c=l.charCodeAt(0),u=En(r.before.charCodeAt(r.before.length-1),c,a);u.inside&&(l=kn(c)+l.slice(1));const d=l.charCodeAt(l.length-1),p=En(r.after.charCodeAt(0),d,a);p.inside&&(l=l.slice(0,-1)+kn(d));const h=i.move(a);return o(),n.attentionEncodeSurroundingInfo={after:p.outside,before:u.outside},s+l+h}pn.peek=function(){return"~"},Sn.peek=function(e,t,n){return n.options.emphasis||"*"};const _n={};function Cn(e,t){const n=t||_n;return Tn(e,"boolean"!=typeof n.includeImageAlt||n.includeImageAlt,"boolean"!=typeof n.includeHtml||n.includeHtml)}function Tn(e,t,n){if(function(e){return Boolean(e&&"object"==typeof e)}(e)){if("value"in e)return"html"!==e.type||n?e.value:"";if(t&&"alt"in e&&e.alt)return e.alt;if("children"in e)return An(e.children,t,n)}return Array.isArray(e)?An(e,t,n):""}function An(e,t,n){const r=[];let a=-1;for(;++a<e.length;)r[a]=Tn(e[a],t,n);return r.join("")}function Rn(e,t){let n=!1;return(0,b.YR)(e,(function(e){if("value"in e&&/\r?\n|\r/.test(e.value)||"break"===e.type)return n=!0,V.dc})),Boolean((!e.depth||e.depth<3)&&Cn(e)&&(t.options.setext||n))}function Pn(e){return e.value||""}function Dn(e,t,n,r){const a=On(n),o='"'===a?"Quote":"Apostrophe",i=n.enter("image");let s=n.enter("label");const l=n.createTracker(r);let c=l.move("![");return c+=l.move(n.safe(e.alt,{before:c,after:"]",...l.current()})),c+=l.move("]("),s(),!e.url&&e.title||/[\0- \u007F]/.test(e.url)?(s=n.enter("destinationLiteral"),c+=l.move("<"),c+=l.move(n.safe(e.url,{before:c,after:">",...l.current()})),c+=l.move(">")):(s=n.enter("destinationRaw"),c+=l.move(n.safe(e.url,{before:c,after:e.title?" ":")",...l.current()}))),s(),e.title&&(s=n.enter(`title${o}`),c+=l.move(" "+a),c+=l.move(n.safe(e.title,{before:c,after:a,...l.current()})),c+=l.move(a),s()),c+=l.move(")"),i(),c}function Mn(e,t,n,r){const a=e.referenceType,o=n.enter("imageReference");let i=n.enter("label");const s=n.createTracker(r);let l=s.move("![");const c=n.safe(e.alt,{before:l,after:"]",...s.current()});l+=s.move(c+"]["),i();const u=n.stack;n.stack=[],i=n.enter("reference");const d=n.safe(n.associationId(e),{before:l,after:"]",...s.current()});return i(),n.stack=u,o(),"full"!==a&&c&&c===d?"shortcut"===a?l=l.slice(0,-1):l+=s.move("]"):l+=s.move(d+"]"),l}function $n(e,t,n){let r=e.value||"",a="`",o=-1;for(;new RegExp("(^|[^`])"+a+"([^`]|$)").test(r);)a+="`";for(/[^ \r\n]/.test(r)&&(/^[ \r\n]/.test(r)&&/[ \r\n]$/.test(r)||/^`|`$/.test(r))&&(r=" "+r+" ");++o<n.unsafe.length;){const e=n.unsafe[o],t=n.compilePattern(e);let a;if(e.atBreak)for(;a=t.exec(r);){let e=a.index;10===r.charCodeAt(e)&&13===r.charCodeAt(e-1)&&e--,r=r.slice(0,e)+" "+r.slice(a.index+1)}}return a+r+a}function Nn(e,t){const n=Cn(e);return Boolean(!t.options.resourceLink&&e.url&&!e.title&&e.children&&1===e.children.length&&"text"===e.children[0].type&&(n===e.url||"mailto:"+n===e.url)&&/^[a-z][a-z+.-]+:/i.test(e.url)&&!/[\0- <>\u007F]/.test(e.url))}function In(e,t,n,r){const a=On(n),o='"'===a?"Quote":"Apostrophe",i=n.createTracker(r);let s,l;if(Nn(e,n)){const t=n.stack;n.stack=[],s=n.enter("autolink");let r=i.move("<");return r+=i.move(n.containerPhrasing(e,{before:r,after:">",...i.current()})),r+=i.move(">"),s(),n.stack=t,r}s=n.enter("link"),l=n.enter("label");let c=i.move("[");return c+=i.move(n.containerPhrasing(e,{before:c,after:"](",...i.current()})),c+=i.move("]("),l(),!e.url&&e.title||/[\0- \u007F]/.test(e.url)?(l=n.enter("destinationLiteral"),c+=i.move("<"),c+=i.move(n.safe(e.url,{before:c,after:">",...i.current()})),c+=i.move(">")):(l=n.enter("destinationRaw"),c+=i.move(n.safe(e.url,{before:c,after:e.title?" ":")",...i.current()}))),l(),e.title&&(l=n.enter(`title${o}`),c+=i.move(" "+a),c+=i.move(n.safe(e.title,{before:c,after:a,...i.current()})),c+=i.move(a),l()),c+=i.move(")"),s(),c}function Ln(e,t,n,r){const a=e.referenceType,o=n.enter("linkReference");let i=n.enter("label");const s=n.createTracker(r);let l=s.move("[");const c=n.containerPhrasing(e,{before:l,after:"]",...s.current()});l+=s.move(c+"]["),i();const u=n.stack;n.stack=[],i=n.enter("reference");const d=n.safe(n.associationId(e),{before:l,after:"]",...s.current()});return i(),n.stack=u,o(),"full"!==a&&c&&c===d?"shortcut"===a?l=l.slice(0,-1):l+=s.move("]"):l+=s.move(d+"]"),l}function Bn(e){const t=e.options.bullet||"*";if("*"!==t&&"+"!==t&&"-"!==t)throw new Error("Cannot serialize items with `"+t+"` for `options.bullet`, expected `*`, `+`, or `-`");return t}function Fn(e){const t=e.options.rule||"*";if("*"!==t&&"-"!==t&&"_"!==t)throw new Error("Cannot serialize rules with `"+t+"` for `options.rule`, expected `*`, `-`, or `_`");return t}Pn.peek=function(){return"<"},Dn.peek=function(){return"!"},Mn.peek=function(){return"!"},$n.peek=function(){return"`"},In.peek=function(e,t,n){return Nn(e,n)?"<":"["},Ln.peek=function(){return"["};const Qn=(0,_.C)(["break","delete","emphasis","footnote","footnoteReference","image","imageReference","inlineCode","inlineMath","link","linkReference","mdxJsxTextElement","mdxTextExpression","strong","text","textDirective"]);function Un(e,t,n,r){const a=function(e){const t=e.options.strong||"*";if("*"!==t&&"_"!==t)throw new Error("Cannot serialize strong with `"+t+"` for `options.strong`, expected `*`, or `_`");return t}(n),o=n.enter("strong"),i=n.createTracker(r),s=i.move(a+a);let l=i.move(n.containerPhrasing(e,{after:a,before:s,...i.current()}));const c=l.charCodeAt(0),u=En(r.before.charCodeAt(r.before.length-1),c,a);u.inside&&(l=kn(c)+l.slice(1));const d=l.charCodeAt(l.length-1),p=En(r.after.charCodeAt(0),d,a);p.inside&&(l=l.slice(0,-1)+kn(d));const h=i.move(a+a);return o(),n.attentionEncodeSurroundingInfo={after:p.outside,before:u.outside},s+l+h}Un.peek=function(e,t,n){return n.options.strong||"*"};const zn={blockquote:function(e,t,n,r){const a=n.enter("blockquote"),o=n.createTracker(r);o.move("> "),o.shift(2);const i=n.indentLines(n.containerFlow(e,o.current()),mn);return a(),i},break:bn,code:function(e,t,n,r){const a=function(e){const t=e.options.fence||"`";if("`"!==t&&"~"!==t)throw new Error("Cannot serialize code with `"+t+"` for `options.fence`, expected `` ` `` or `~`");return t}(n),o=e.value||"",i="`"===a?"GraveAccent":"Tilde";if(yn(e,n)){const e=n.enter("codeIndented"),t=n.indentLines(o,wn);return e(),t}const s=n.createTracker(r),l=a.repeat(Math.max(function(e,t){const n=String(e);let r=n.indexOf(t),a=r,o=0,i=0;if("string"!=typeof t)throw new TypeError("Expected substring");for(;-1!==r;)r===a?++o>i&&(i=o):o=1,a=r+t.length,r=n.indexOf(t,a);return i}(o,a)+1,3)),c=n.enter("codeFenced");let u=s.move(l);if(e.lang){const t=n.enter(`codeFencedLang${i}`);u+=s.move(n.safe(e.lang,{before:u,after:" ",encode:["`"],...s.current()})),t()}if(e.lang&&e.meta){const t=n.enter(`codeFencedMeta${i}`);u+=s.move(" "),u+=s.move(n.safe(e.meta,{before:u,after:"\n",encode:["`"],...s.current()})),t()}return u+=s.move("\n"),o&&(u+=s.move(o+"\n")),u+=s.move(l),c(),u},definition:function(e,t,n,r){const a=On(n),o='"'===a?"Quote":"Apostrophe",i=n.enter("definition");let s=n.enter("label");const l=n.createTracker(r);let c=l.move("[");return c+=l.move(n.safe(n.associationId(e),{before:c,after:"]",...l.current()})),c+=l.move("]: "),s(),!e.url||/[\0- \u007F]/.test(e.url)?(s=n.enter("destinationLiteral"),c+=l.move("<"),c+=l.move(n.safe(e.url,{before:c,after:">",...l.current()})),c+=l.move(">")):(s=n.enter("destinationRaw"),c+=l.move(n.safe(e.url,{before:c,after:e.title?" ":"\n",...l.current()}))),s(),e.title&&(s=n.enter(`title${o}`),c+=l.move(" "+a),c+=l.move(n.safe(e.title,{before:c,after:a,...l.current()})),c+=l.move(a),s()),i(),c},emphasis:Sn,hardBreak:bn,heading:function(e,t,n,r){const a=Math.max(Math.min(6,e.depth||1),1),o=n.createTracker(r);if(Rn(e,n)){const t=n.enter("headingSetext"),r=n.enter("phrasing"),i=n.containerPhrasing(e,{...o.current(),before:"\n",after:"\n"});return r(),t(),i+"\n"+(1===a?"=":"-").repeat(i.length-(Math.max(i.lastIndexOf("\r"),i.lastIndexOf("\n"))+1))}const i="#".repeat(a),s=n.enter("headingAtx"),l=n.enter("phrasing");o.move(i+" ");let c=n.containerPhrasing(e,{before:"# ",after:"\n",...o.current()});return/^[\t ]/.test(c)&&(c=kn(c.charCodeAt(0))+c.slice(1)),c=c?i+" "+c:i,n.options.closeAtx&&(c+=" "+i),l(),s(),c},html:Pn,image:Dn,imageReference:Mn,inlineCode:$n,link:In,linkReference:Ln,list:function(e,t,n,r){const a=n.enter("list"),o=n.bulletCurrent;let i=e.ordered?function(e){const t=e.options.bulletOrdered||".";if("."!==t&&")"!==t)throw new Error("Cannot serialize items with `"+t+"` for `options.bulletOrdered`, expected `.` or `)`");return t}(n):Bn(n);const s=e.ordered?"."===i?")":".":function(e){const t=Bn(e),n=e.options.bulletOther;if(!n)return"*"===t?"-":"*";if("*"!==n&&"+"!==n&&"-"!==n)throw new Error("Cannot serialize items with `"+n+"` for `options.bulletOther`, expected `*`, `+`, or `-`");if(n===t)throw new Error("Expected `bullet` (`"+t+"`) and `bulletOther` (`"+n+"`) to be different");return n}(n);let l=!(!t||!n.bulletLastUsed)&&i===n.bulletLastUsed;if(!e.ordered){const t=e.children?e.children[0]:void 0;if("*"!==i&&"-"!==i||!t||t.children&&t.children[0]||"list"!==n.stack[n.stack.length-1]||"listItem"!==n.stack[n.stack.length-2]||"list"!==n.stack[n.stack.length-3]||"listItem"!==n.stack[n.stack.length-4]||0!==n.indexStack[n.indexStack.length-1]||0!==n.indexStack[n.indexStack.length-2]||0!==n.indexStack[n.indexStack.length-3]||(l=!0),Fn(n)===i&&t){let t=-1;for(;++t<e.children.length;){const n=e.children[t];if(n&&"listItem"===n.type&&n.children&&n.children[0]&&"thematicBreak"===n.children[0].type){l=!0;break}}}}l&&(i=s),n.bulletCurrent=i;const c=n.containerFlow(e,r);return n.bulletLastUsed=i,n.bulletCurrent=o,a(),c},listItem:function(e,t,n,r){const a=function(e){const t=e.options.listItemIndent||"one";if("tab"!==t&&"one"!==t&&"mixed"!==t)throw new Error("Cannot serialize items with `"+t+"` for `options.listItemIndent`, expected `tab`, `one`, or `mixed`");return t}(n);let o=n.bulletCurrent||Bn(n);t&&"list"===t.type&&t.ordered&&(o=("number"==typeof t.start&&t.start>-1?t.start:1)+(!1===n.options.incrementListMarker?0:t.children.indexOf(e))+o);let i=o.length+1;("tab"===a||"mixed"===a&&(t&&"list"===t.type&&t.spread||e.spread))&&(i=4*Math.ceil(i/4));const s=n.createTracker(r);s.move(o+" ".repeat(i-o.length)),s.shift(i);const l=n.enter("listItem"),c=n.indentLines(n.containerFlow(e,s.current()),(function(e,t,n){return t?(n?"":" ".repeat(i))+e:(n?o:o+" ".repeat(i-o.length))+e}));return l(),c},paragraph:function(e,t,n,r){const a=n.enter("paragraph"),o=n.enter("phrasing"),i=n.containerPhrasing(e,r);return o(),a(),i},root:function(e,t,n,r){return(e.children.some((function(e){return Qn(e)}))?n.containerPhrasing:n.containerFlow).call(n,e,r)},strong:Un,text:function(e,t,n,r){return n.safe(e.value,r)},thematicBreak:function(e,t,n){const r=(Fn(n)+(n.options.ruleSpaces?" ":"")).repeat(function(e){const t=e.options.ruleRepetition||3;if(t<3)throw new Error("Cannot serialize rules with repetition `"+t+"` for `options.ruleRepetition`, expected `3` or more");return t}(n));return n.options.ruleSpaces?r.slice(0,-1):r}};function jn(e){const t=e._align;(0,te.ok)(t,"expected `_align` on table"),this.enter({type:"table",align:t.map((function(e){return"none"===e?null:e})),children:[]},e),this.data.inTable=!0}function Wn(e){this.exit(e),this.data.inTable=void 0}function Xn(e){this.enter({type:"tableRow",children:[]},e)}function qn(e){this.exit(e)}function Vn(e){this.enter({type:"tableCell",children:[]},e)}function Hn(e){let t=this.resume();this.data.inTable&&(t=t.replace(/\\([\\|])/g,Zn));const n=this.stack[this.stack.length-1];(0,te.ok)("inlineCode"===n.type),n.value=t,this.exit(e)}function Zn(e,t){return"|"===t?t:e}function Gn(e){const t=e||{},n=t.tableCellPadding,r=t.tablePipeAlign,a=t.stringLength,o=n?" ":"|";return{unsafe:[{character:"\r",inConstruct:"tableCell"},{character:"\n",inConstruct:"tableCell"},{atBreak:!0,character:"|",after:"[\t :-]"},{character:"|",inConstruct:"tableCell"},{atBreak:!0,character:":",after:"-"},{atBreak:!0,character:"-",after:"[:|-]"}],handlers:{inlineCode:function(e,t,n){let r=zn.inlineCode(e,t,n);return n.stack.includes("tableCell")&&(r=r.replace(/\|/g,"\\$&")),r},table:function(e,t,n,r){return s(function(e,t,n){const r=e.children;let a=-1;const o=[],i=t.enter("table");for(;++a<r.length;)o[a]=l(r[a],t,n);return i(),o}(e,n,r),e.align)},tableCell:i,tableRow:function(e,t,n,r){const a=s([l(e,n,r)]);return a.slice(0,a.indexOf("\n"))}}};function i(e,t,n,r){const a=n.enter("tableCell"),i=n.enter("phrasing"),s=n.containerPhrasing(e,{...r,before:o,after:o});return i(),a(),s}function s(e,t){return function(e,t){const n=t||{},r=(n.align||[]).concat(),a=n.stringLength||hn,o=[],i=[],s=[],l=[];let c=0,u=-1;for(;++u<e.length;){const t=[],r=[];let o=-1;for(e[u].length>c&&(c=e[u].length);++o<e[u].length;){const i=null==(d=e[u][o])?"":String(d);if(!1!==n.alignDelimiters){const e=a(i);r[o]=e,(void 0===l[o]||e>l[o])&&(l[o]=e)}t.push(i)}i[u]=t,s[u]=r}var d;let p=-1;if("object"==typeof r&&"length"in r)for(;++p<c;)o[p]=fn(r[p]);else{const e=fn(r);for(;++p<c;)o[p]=e}p=-1;const h=[],f=[];for(;++p<c;){const e=o[p];let t="",r="";99===e?(t=":",r=":"):108===e?t=":":114===e&&(r=":");let a=!1===n.alignDelimiters?1:Math.max(1,l[p]-t.length-r.length);const i=t+"-".repeat(a)+r;!1!==n.alignDelimiters&&(a=t.length+a+r.length,a>l[p]&&(l[p]=a),f[p]=a),h[p]=i}i.splice(1,0,h),s.splice(1,0,f),u=-1;const m=[];for(;++u<i.length;){const e=i[u],t=s[u];p=-1;const r=[];for(;++p<c;){const a=e[p]||"";let i="",s="";if(!1!==n.alignDelimiters){const e=l[p]-(t[p]||0),n=o[p];114===n?i=" ".repeat(e):99===n?e%2?(i=" ".repeat(e/2+.5),s=" ".repeat(e/2-.5)):(i=" ".repeat(e/2),s=i):s=" ".repeat(e)}!1===n.delimiterStart||p||r.push("|"),!1===n.padding||!1===n.alignDelimiters&&""===a||!1===n.delimiterStart&&!p||r.push(" "),!1!==n.alignDelimiters&&r.push(i),r.push(a),!1!==n.alignDelimiters&&r.push(s),!1!==n.padding&&r.push(" "),!1===n.delimiterEnd&&p===c-1||r.push("|")}m.push(!1===n.delimiterEnd?r.join("").replace(/ +$/,""):r.join(""))}return m.join("\n")}(e,{align:t,alignDelimiters:r,padding:n,stringLength:a})}function l(e,t,n){const r=e.children;let a=-1;const o=[],s=t.enter("tableRow");for(;++a<r.length;)o[a]=i(r[a],0,t,n);return s(),o}}function Kn(e){const t=this.stack[this.stack.length-2];(0,te.ok)("listItem"===t.type),t.checked="taskListCheckValueChecked"===e.type}function Yn(e){const t=this.stack[this.stack.length-2];if(t&&"listItem"===t.type&&"boolean"==typeof t.checked){const e=this.stack[this.stack.length-1];(0,te.ok)("paragraph"===e.type);const n=e.children[0];if(n&&"text"===n.type){const r=t.children;let a,o=-1;for(;++o<r.length;){const e=r[o];if("paragraph"===e.type){a=e;break}}a===e&&(n.value=n.value.slice(1),0===n.value.length?e.children.shift():e.position&&n.position&&"number"==typeof n.position.start.offset&&(n.position.start.column++,n.position.start.offset++,e.position.start=Object.assign({},n.position.start)))}}this.exit(e)}function Jn(e,t,n,r){const a=e.children[0],o="boolean"==typeof e.checked&&a&&"paragraph"===a.type,i="["+(e.checked?"x":" ")+"] ",s=n.createTracker(r);o&&s.move(i);let l=zn.listItem(e,t,n,{...r,...s.current()});return o&&(l=l.replace(/^(?:[*+-]|\d+\.)([\r\n]| {1,3})/,(function(e){return e+i}))),l}function er(e,t,n,r){const a=e.length;let o,i=0;if(t=t<0?-t>a?0:a+t:t>a?a:t,n=n>0?n:0,r.length<1e4)o=Array.from(r),o.unshift(t,n),e.splice(...o);else for(n&&e.splice(t,n);i<r.length;)o=r.slice(i,i+1e4),o.unshift(t,0),e.splice(...o),i+=1e4,t+=1e4}function tr(e,t){return e.length>0?(er(e,e.length,0,t),e):t}const nr={}.hasOwnProperty;function rr(e){const t={};let n=-1;for(;++n<e.length;)ar(t,e[n]);return t}function ar(e,t){let n;for(n in t){const r=(nr.call(e,n)?e[n]:void 0)||(e[n]={}),a=t[n];let o;if(a)for(o in a){nr.call(r,o)||(r[o]=[]);const e=a[o];or(r[o],Array.isArray(e)?e:e?[e]:[])}}}function or(e,t){let n=-1;const r=[];for(;++n<t.length;)("after"===t[n].add?e:r).push(t[n]);er(e,0,0,r)}const ir={tokenize:function(e,t,n){let r=0;return function t(o){return(87===o||119===o)&&r<3?(r++,e.consume(o),t):46===o&&3===r?(e.consume(o),a):n(o)};function a(e){return null===e?n(e):t(e)}},partial:!0},sr={tokenize:function(e,t,n){let r,a,o;return i;function i(t){return 46===t||95===t?e.check(cr,l,s)(t):null===t||Pt(t)||$t(t)||45!==t&&Mt(t)?l(t):(o=!0,e.consume(t),i)}function s(t){return 95===t?r=!0:(a=r,r=void 0),e.consume(t),i}function l(e){return a||r||!o?n(e):t(e)}},partial:!0},lr={tokenize:function(e,t){let n=0,r=0;return a;function a(i){return 40===i?(n++,e.consume(i),a):41===i&&r<n?o(i):33===i||34===i||38===i||39===i||41===i||42===i||44===i||46===i||58===i||59===i||60===i||63===i||93===i||95===i||126===i?e.check(cr,t,o)(i):null===i||Pt(i)||$t(i)?t(i):(e.consume(i),a)}function o(t){return 41===t&&r++,e.consume(t),a}},partial:!0},cr={tokenize:function(e,t,n){return r;function r(i){return 33===i||34===i||39===i||41===i||42===i||44===i||46===i||58===i||59===i||63===i||95===i||126===i?(e.consume(i),r):38===i?(e.consume(i),o):93===i?(e.consume(i),a):60===i||null===i||Pt(i)||$t(i)?t(i):n(i)}function a(e){return null===e||40===e||91===e||Pt(e)||$t(e)?t(e):r(e)}function o(e){return xt(e)?i(e):n(e)}function i(t){return 59===t?(e.consume(t),r):xt(t)?(e.consume(t),i):n(t)}},partial:!0},ur={tokenize:function(e,t,n){return function(t){return e.consume(t),r};function r(e){return Et(e)?n(e):t(e)}},partial:!0},dr={name:"wwwAutolink",tokenize:function(e,t,n){const r=this;return function(t){return 87!==t&&119!==t||!gr.call(r,r.previous)||wr(r.events)?n(t):(e.enter("literalAutolink"),e.enter("literalAutolinkWww"),e.check(ir,e.attempt(sr,e.attempt(lr,a),n),n)(t))};function a(n){return e.exit("literalAutolinkWww"),e.exit("literalAutolink"),t(n)}},previous:gr},pr={name:"protocolAutolink",tokenize:function(e,t,n){const r=this;let a="",o=!1;return function(t){return 72!==t&&104!==t||!vr.call(r,r.previous)||wr(r.events)?n(t):(e.enter("literalAutolink"),e.enter("literalAutolinkHttp"),a+=String.fromCodePoint(t),e.consume(t),i)};function i(t){if(xt(t)&&a.length<5)return a+=String.fromCodePoint(t),e.consume(t),i;if(58===t){const n=a.toLowerCase();if("http"===n||"https"===n)return e.consume(t),s}return n(t)}function s(t){return 47===t?(e.consume(t),o?l:(o=!0,s)):n(t)}function l(t){return null===t||_t(t)||Pt(t)||$t(t)||Mt(t)?n(t):e.attempt(sr,e.attempt(lr,c),n)(t)}function c(n){return e.exit("literalAutolinkHttp"),e.exit("literalAutolink"),t(n)}},previous:vr},hr={name:"emailAutolink",tokenize:function(e,t,n){const r=this;let a,o;return function(t){return yr(t)&&br.call(r,r.previous)&&!wr(r.events)?(e.enter("literalAutolink"),e.enter("literalAutolinkEmail"),i(t)):n(t)};function i(t){return yr(t)?(e.consume(t),i):64===t?(e.consume(t),s):n(t)}function s(t){return 46===t?e.check(ur,c,l)(t):45===t||95===t||Et(t)?(o=!0,e.consume(t),s):c(t)}function l(t){return e.consume(t),a=!0,s}function c(i){return o&&a&&xt(r.previous)?(e.exit("literalAutolinkEmail"),e.exit("literalAutolink"),t(i)):n(i)}},previous:br},fr={};let mr=48;for(;mr<123;)fr[mr]=hr,mr++,58===mr?mr=65:91===mr&&(mr=97);function gr(e){return null===e||40===e||42===e||95===e||91===e||93===e||126===e||Pt(e)}function vr(e){return!xt(e)}function br(e){return!(47===e||yr(e))}function yr(e){return 43===e||45===e||46===e||95===e||Et(e)}function wr(e){let t=e.length,n=!1;for(;t--;){const r=e[t][1];if(("labelLink"===r.type||"labelImage"===r.type)&&!r._balanced){n=!0;break}if(r._gfmAutolinkLiteralWalkedInto){n=!1;break}}return e.length>0&&!n&&(e[e.length-1][1]._gfmAutolinkLiteralWalkedInto=!0),n}function Or(e,t,n,r){const a=r?r-1:Number.POSITIVE_INFINITY;let o=0;return function(r){return Dt(r)?(e.enter(n),i(r)):t(r)};function i(r){return Dt(r)&&o++<a?(e.consume(r),i):(e.exit(n),t(r))}}fr[43]=hr,fr[45]=hr,fr[46]=hr,fr[95]=hr,fr[72]=[hr,pr],fr[104]=[hr,pr],fr[87]=[hr,dr],fr[119]=[hr,dr];const kr={partial:!0,tokenize:function(e,t,n){return function(t){return Dt(t)?Or(e,r,"linePrefix")(t):r(t)};function r(e){return null===e||Rt(e)?t(e):n(e)}}},xr={tokenize:function(e,t,n){const r=this;return Or(e,(function(e){const a=r.events[r.events.length-1];return a&&"gfmFootnoteDefinitionIndent"===a[1].type&&4===a[2].sliceSerialize(a[1],!0).length?t(e):n(e)}),"gfmFootnoteDefinitionIndent",5)},partial:!0};function Er(e,t,n){const r=this;let a=r.events.length;const o=r.parser.gfmFootnotes||(r.parser.gfmFootnotes=[]);let i;for(;a--;){const e=r.events[a][1];if("labelImage"===e.type){i=e;break}if("gfmFootnoteCall"===e.type||"labelLink"===e.type||"label"===e.type||"image"===e.type||"link"===e.type)break}return function(a){if(!i||!i._balanced)return n(a);const s=Zt(r.sliceSerialize({start:i.end,end:r.now()}));return 94===s.codePointAt(0)&&o.includes(s.slice(1))?(e.enter("gfmFootnoteCallLabelMarker"),e.consume(a),e.exit("gfmFootnoteCallLabelMarker"),t(a)):n(a)}}function Sr(e,t){let n,r=e.length;for(;r--;)if("labelImage"===e[r][1].type&&"enter"===e[r][0]){n=e[r][1];break}e[r+1][1].type="data",e[r+3][1].type="gfmFootnoteCallLabelMarker";const a={type:"gfmFootnoteCall",start:Object.assign({},e[r+3][1].start),end:Object.assign({},e[e.length-1][1].end)},o={type:"gfmFootnoteCallMarker",start:Object.assign({},e[r+3][1].end),end:Object.assign({},e[r+3][1].end)};o.end.column++,o.end.offset++,o.end._bufferIndex++;const i={type:"gfmFootnoteCallString",start:Object.assign({},o.end),end:Object.assign({},e[e.length-1][1].start)},s={type:"chunkString",contentType:"string",start:Object.assign({},i.start),end:Object.assign({},i.end)},l=[e[r+1],e[r+2],["enter",a,t],e[r+3],e[r+4],["enter",o,t],["exit",o,t],["enter",i,t],["enter",s,t],["exit",s,t],["exit",i,t],e[e.length-2],e[e.length-1],["exit",a,t]];return e.splice(r,e.length-r+1,...l),e}function _r(e,t,n){const r=this,a=r.parser.gfmFootnotes||(r.parser.gfmFootnotes=[]);let o,i=0;return function(t){return e.enter("gfmFootnoteCall"),e.enter("gfmFootnoteCallLabelMarker"),e.consume(t),e.exit("gfmFootnoteCallLabelMarker"),s};function s(t){return 94!==t?n(t):(e.enter("gfmFootnoteCallMarker"),e.consume(t),e.exit("gfmFootnoteCallMarker"),e.enter("gfmFootnoteCallString"),e.enter("chunkString").contentType="string",l)}function l(s){if(i>999||93===s&&!o||null===s||91===s||Pt(s))return n(s);if(93===s){e.exit("chunkString");const o=e.exit("gfmFootnoteCallString");return a.includes(Zt(r.sliceSerialize(o)))?(e.enter("gfmFootnoteCallLabelMarker"),e.consume(s),e.exit("gfmFootnoteCallLabelMarker"),e.exit("gfmFootnoteCall"),t):n(s)}return Pt(s)||(o=!0),i++,e.consume(s),92===s?c:l}function c(t){return 91===t||92===t||93===t?(e.consume(t),i++,l):l(t)}}function Cr(e,t,n){const r=this,a=r.parser.gfmFootnotes||(r.parser.gfmFootnotes=[]);let o,i,s=0;return function(t){return e.enter("gfmFootnoteDefinition")._container=!0,e.enter("gfmFootnoteDefinitionLabel"),e.enter("gfmFootnoteDefinitionLabelMarker"),e.consume(t),e.exit("gfmFootnoteDefinitionLabelMarker"),l};function l(t){return 94===t?(e.enter("gfmFootnoteDefinitionMarker"),e.consume(t),e.exit("gfmFootnoteDefinitionMarker"),e.enter("gfmFootnoteDefinitionLabelString"),e.enter("chunkString").contentType="string",c):n(t)}function c(t){if(s>999||93===t&&!i||null===t||91===t||Pt(t))return n(t);if(93===t){e.exit("chunkString");const n=e.exit("gfmFootnoteDefinitionLabelString");return o=Zt(r.sliceSerialize(n)),e.enter("gfmFootnoteDefinitionLabelMarker"),e.consume(t),e.exit("gfmFootnoteDefinitionLabelMarker"),e.exit("gfmFootnoteDefinitionLabel"),d}return Pt(t)||(i=!0),s++,e.consume(t),92===t?u:c}function u(t){return 91===t||92===t||93===t?(e.consume(t),s++,c):c(t)}function d(t){return 58===t?(e.enter("definitionMarker"),e.consume(t),e.exit("definitionMarker"),a.includes(o)||a.push(o),Or(e,p,"gfmFootnoteDefinitionWhitespace")):n(t)}function p(e){return t(e)}}function Tr(e,t,n){return e.check(kr,t,e.attempt(xr,t,n))}function Ar(e){e.exit("gfmFootnoteDefinition")}function Rr(e,t,n){const r=[];let a=-1;for(;++a<e.length;){const o=e[a].resolveAll;o&&!r.includes(o)&&(t=o(t,n),r.push(o))}return t}function Pr(e){let t=(e||{}).singleTilde;const n={name:"strikethrough",tokenize:function(e,n,r){const a=this.previous,o=this.events;let i=0;return function(t){return 126===a&&"characterEscape"!==o[o.length-1][1].type?r(t):(e.enter("strikethroughSequenceTemporary"),s(t))};function s(o){const l=xn(a);if(126===o)return i>1?r(o):(e.consume(o),i++,s);if(i<2&&!t)return r(o);const c=e.exit("strikethroughSequenceTemporary"),u=xn(o);return c._open=!u||2===u&&Boolean(l),c._close=!l||2===l&&Boolean(u),n(o)}},resolveAll:function(e,t){let n=-1;for(;++n<e.length;)if("enter"===e[n][0]&&"strikethroughSequenceTemporary"===e[n][1].type&&e[n][1]._close){let r=n;for(;r--;)if("exit"===e[r][0]&&"strikethroughSequenceTemporary"===e[r][1].type&&e[r][1]._open&&e[n][1].end.offset-e[n][1].start.offset==e[r][1].end.offset-e[r][1].start.offset){e[n][1].type="strikethroughSequence",e[r][1].type="strikethroughSequence";const a={type:"strikethrough",start:Object.assign({},e[r][1].start),end:Object.assign({},e[n][1].end)},o={type:"strikethroughText",start:Object.assign({},e[r][1].end),end:Object.assign({},e[n][1].start)},i=[["enter",a,t],["enter",e[r][1],t],["exit",e[r][1],t],["enter",o,t]],s=t.parser.constructs.insideSpan.null;s&&er(i,i.length,0,Rr(s,e.slice(r+1,n),t)),er(i,i.length,0,[["exit",o,t],["enter",e[n][1],t],["exit",e[n][1],t],["exit",a,t]]),er(e,r-1,n-r+3,i),n=r+i.length-2;break}}for(n=-1;++n<e.length;)"strikethroughSequenceTemporary"===e[n][1].type&&(e[n][1].type="data");return e}};return null==t&&(t=!0),{text:{126:n},insideSpan:{null:[n]},attentionMarkers:{null:[126]}}}class Dr{constructor(){this.map=[]}add(e,t,n){!function(e,t,n,r){let a=0;if(0!==n||0!==r.length){for(;a<e.map.length;){if(e.map[a][0]===t)return e.map[a][1]+=n,void e.map[a][2].push(...r);a+=1}e.map.push([t,n,r])}}(this,e,t,n)}consume(e){if(this.map.sort((function(e,t){return e[0]-t[0]})),0===this.map.length)return;let t=this.map.length;const n=[];for(;t>0;)t-=1,n.push(e.slice(this.map[t][0]+this.map[t][1]),this.map[t][2]),e.length=this.map[t][0];n.push(e.slice()),e.length=0;let r=n.pop();for(;r;){for(const t of r)e.push(t);r=n.pop()}this.map.length=0}}function Mr(e,t){let n=!1;const r=[];for(;t<e.length;){const a=e[t];if(n){if("enter"===a[0])"tableContent"===a[1].type&&r.push("tableDelimiterMarker"===e[t+1][1].type?"left":"none");else if("tableContent"===a[1].type){if("tableDelimiterMarker"===e[t-1][1].type){const e=r.length-1;r[e]="left"===r[e]?"center":"right"}}else if("tableDelimiterRow"===a[1].type)break}else"enter"===a[0]&&"tableDelimiterRow"===a[1].type&&(n=!0);t+=1}return r}function $r(e,t,n){const r=this;let a,o=0,i=0;return function(e){let t=r.events.length-1;for(;t>-1;){const e=r.events[t][1].type;if("lineEnding"!==e&&"linePrefix"!==e)break;t--}const a=t>-1?r.events[t][1].type:null,o="tableHead"===a||"tableRow"===a?w:s;return o===w&&r.parser.lazy[r.now().line]?n(e):o(e)};function s(t){return e.enter("tableHead"),e.enter("tableRow"),function(e){return 124===e||(a=!0,i+=1),l(e)}(t)}function l(t){return null===t?n(t):Rt(t)?i>1?(i=0,r.interrupt=!0,e.exit("tableRow"),e.enter("lineEnding"),e.consume(t),e.exit("lineEnding"),d):n(t):Dt(t)?Or(e,l,"whitespace")(t):(i+=1,a&&(a=!1,o+=1),124===t?(e.enter("tableCellDivider"),e.consume(t),e.exit("tableCellDivider"),a=!0,l):(e.enter("data"),c(t)))}function c(t){return null===t||124===t||Pt(t)?(e.exit("data"),l(t)):(e.consume(t),92===t?u:c)}function u(t){return 92===t||124===t?(e.consume(t),c):c(t)}function d(t){return r.interrupt=!1,r.parser.lazy[r.now().line]?n(t):(e.enter("tableDelimiterRow"),a=!1,Dt(t)?Or(e,p,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(t):p(t))}function p(t){return 45===t||58===t?f(t):124===t?(a=!0,e.enter("tableCellDivider"),e.consume(t),e.exit("tableCellDivider"),h):y(t)}function h(t){return Dt(t)?Or(e,f,"whitespace")(t):f(t)}function f(t){return 58===t?(i+=1,a=!0,e.enter("tableDelimiterMarker"),e.consume(t),e.exit("tableDelimiterMarker"),m):45===t?(i+=1,m(t)):null===t||Rt(t)?b(t):y(t)}function m(t){return 45===t?(e.enter("tableDelimiterFiller"),g(t)):y(t)}function g(t){return 45===t?(e.consume(t),g):58===t?(a=!0,e.exit("tableDelimiterFiller"),e.enter("tableDelimiterMarker"),e.consume(t),e.exit("tableDelimiterMarker"),v):(e.exit("tableDelimiterFiller"),v(t))}function v(t){return Dt(t)?Or(e,b,"whitespace")(t):b(t)}function b(n){return 124===n?p(n):(null===n||Rt(n))&&a&&o===i?(e.exit("tableDelimiterRow"),e.exit("tableHead"),t(n)):y(n)}function y(e){return n(e)}function w(t){return e.enter("tableRow"),O(t)}function O(n){return 124===n?(e.enter("tableCellDivider"),e.consume(n),e.exit("tableCellDivider"),O):null===n||Rt(n)?(e.exit("tableRow"),t(n)):Dt(n)?Or(e,O,"whitespace")(n):(e.enter("data"),k(n))}function k(t){return null===t||124===t||Pt(t)?(e.exit("data"),O(t)):(e.consume(t),92===t?x:k)}function x(t){return 92===t||124===t?(e.consume(t),k):k(t)}}function Nr(e,t){let n,r,a,o=-1,i=!0,s=0,l=[0,0,0,0],c=[0,0,0,0],u=!1,d=0;const p=new Dr;for(;++o<e.length;){const h=e[o],f=h[1];"enter"===h[0]?"tableHead"===f.type?(u=!1,0!==d&&(Lr(p,t,d,n,r),r=void 0,d=0),n={type:"table",start:Object.assign({},f.start),end:Object.assign({},f.end)},p.add(o,0,[["enter",n,t]])):"tableRow"===f.type||"tableDelimiterRow"===f.type?(i=!0,a=void 0,l=[0,0,0,0],c=[0,o+1,0,0],u&&(u=!1,r={type:"tableBody",start:Object.assign({},f.start),end:Object.assign({},f.end)},p.add(o,0,[["enter",r,t]])),s="tableDelimiterRow"===f.type?2:r?3:1):!s||"data"!==f.type&&"tableDelimiterMarker"!==f.type&&"tableDelimiterFiller"!==f.type?"tableCellDivider"===f.type&&(i?i=!1:(0!==l[1]&&(c[0]=c[1],a=Ir(p,t,l,s,void 0,a)),l=c,c=[l[1],o,0,0])):(i=!1,0===c[2]&&(0!==l[1]&&(c[0]=c[1],a=Ir(p,t,l,s,void 0,a),l=[0,0,0,0]),c[2]=o)):"tableHead"===f.type?(u=!0,d=o):"tableRow"===f.type||"tableDelimiterRow"===f.type?(d=o,0!==l[1]?(c[0]=c[1],a=Ir(p,t,l,s,o,a)):0!==c[1]&&(a=Ir(p,t,c,s,o,a)),s=0):!s||"data"!==f.type&&"tableDelimiterMarker"!==f.type&&"tableDelimiterFiller"!==f.type||(c[3]=o)}for(0!==d&&Lr(p,t,d,n,r),p.consume(t.events),o=-1;++o<t.events.length;){const e=t.events[o];"enter"===e[0]&&"table"===e[1].type&&(e[1]._align=Mr(t.events,o))}return e}function Ir(e,t,n,r,a,o){const i=1===r?"tableHeader":2===r?"tableDelimiter":"tableData";0!==n[0]&&(o.end=Object.assign({},Br(t.events,n[0])),e.add(n[0],0,[["exit",o,t]]));const s=Br(t.events,n[1]);if(o={type:i,start:Object.assign({},s),end:Object.assign({},s)},e.add(n[1],0,[["enter",o,t]]),0!==n[2]){const a=Br(t.events,n[2]),o=Br(t.events,n[3]),i={type:"tableContent",start:Object.assign({},a),end:Object.assign({},o)};if(e.add(n[2],0,[["enter",i,t]]),2!==r){const r=t.events[n[2]],a=t.events[n[3]];if(r[1].end=Object.assign({},a[1].end),r[1].type="chunkText",r[1].contentType="text",n[3]>n[2]+1){const t=n[2]+1,r=n[3]-n[2]-1;e.add(t,r,[])}}e.add(n[3]+1,0,[["exit",i,t]])}return void 0!==a&&(o.end=Object.assign({},Br(t.events,a)),e.add(a,0,[["exit",o,t]]),o=void 0),o}function Lr(e,t,n,r,a){const o=[],i=Br(t.events,n);a&&(a.end=Object.assign({},i),o.push(["exit",a,t])),r.end=Object.assign({},i),o.push(["exit",r,t]),e.add(n+1,0,o)}function Br(e,t){const n=e[t],r="enter"===n[0]?"start":"end";return n[1][r]}const Fr={name:"tasklistCheck",tokenize:function(e,t,n){const r=this;return function(t){return null===r.previous&&r._gfmTasklistFirstContentOfListItem?(e.enter("taskListCheck"),e.enter("taskListCheckMarker"),e.consume(t),e.exit("taskListCheckMarker"),a):n(t)};function a(t){return Pt(t)?(e.enter("taskListCheckValueUnchecked"),e.consume(t),e.exit("taskListCheckValueUnchecked"),o):88===t||120===t?(e.enter("taskListCheckValueChecked"),e.consume(t),e.exit("taskListCheckValueChecked"),o):n(t)}function o(t){return 93===t?(e.enter("taskListCheckMarker"),e.consume(t),e.exit("taskListCheckMarker"),e.exit("taskListCheck"),i):n(t)}function i(r){return Rt(r)?t(r):Dt(r)?e.check({tokenize:Qr},t,n)(r):n(r)}}};function Qr(e,t,n){return Or(e,(function(e){return null===e?n(e):t(e)}),"whitespace")}const Ur={};function zr(e){const t=e||Ur,n=this.data(),r=n.micromarkExtensions||(n.micromarkExtensions=[]),a=n.fromMarkdownExtensions||(n.fromMarkdownExtensions=[]),o=n.toMarkdownExtensions||(n.toMarkdownExtensions=[]);r.push(function(e){return rr([{text:fr},{document:{91:{name:"gfmFootnoteDefinition",tokenize:Cr,continuation:{tokenize:Tr},exit:Ar}},text:{91:{name:"gfmFootnoteCall",tokenize:_r},93:{name:"gfmPotentialFootnoteCall",add:"after",tokenize:Er,resolveTo:Sr}}},Pr(e),{flow:{null:{name:"table",tokenize:$r,resolveAll:Nr}}},{text:{91:Fr}}])}(t)),a.push([{transforms:[Xt],enter:{literalAutolink:Ft,literalAutolinkEmail:Qt,literalAutolinkHttp:Qt,literalAutolinkWww:Qt},exit:{literalAutolink:Wt,literalAutolinkEmail:jt,literalAutolinkHttp:Ut,literalAutolinkWww:zt}},{enter:{gfmFootnoteCallString:Gt,gfmFootnoteCall:Kt,gfmFootnoteDefinitionLabelString:Yt,gfmFootnoteDefinition:Jt},exit:{gfmFootnoteCallString:en,gfmFootnoteCall:tn,gfmFootnoteDefinitionLabelString:nn,gfmFootnoteDefinition:rn}},{canContainEols:["delete"],enter:{strikethrough:un},exit:{strikethrough:dn}},{enter:{table:jn,tableData:Vn,tableHeader:Vn,tableRow:Xn},exit:{codeText:Hn,table:Wn,tableData:qn,tableHeader:qn,tableRow:qn}},{exit:{taskListCheckValueChecked:Kn,taskListCheckValueUnchecked:Kn,paragraph:Yn}}]),o.push(function(e){return{extensions:[{unsafe:[{character:"@",before:"[+\\-.\\w]",after:"[\\-.\\w]",inConstruct:Lt,notInConstruct:Bt},{character:".",before:"[Ww]",after:"[\\-.\\w]",inConstruct:Lt,notInConstruct:Bt},{character:":",before:"[ps]",after:"\\/",inConstruct:Lt,notInConstruct:Bt}]},on(e),{unsafe:[{character:"~",inConstruct:"phrasing",notInConstruct:cn}],handlers:{delete:pn}},Gn(e),{unsafe:[{atBreak:!0,character:"-",after:"[:|-]"}],handlers:{listItem:Jn}}]}}(t))}class jr{constructor(e){this.left=e?[...e]:[],this.right=[]}get(e){if(e<0||e>=this.left.length+this.right.length)throw new RangeError("Cannot access index `"+e+"` in a splice buffer of size `"+(this.left.length+this.right.length)+"`");return e<this.left.length?this.left[e]:this.right[this.right.length-e+this.left.length-1]}get length(){return this.left.length+this.right.length}shift(){return this.setCursor(0),this.right.pop()}slice(e,t){const n=null==t?Number.POSITIVE_INFINITY:t;return n<this.left.length?this.left.slice(e,n):e>this.left.length?this.right.slice(this.right.length-n+this.left.length,this.right.length-e+this.left.length).reverse():this.left.slice(e).concat(this.right.slice(this.right.length-n+this.left.length).reverse())}splice(e,t,n){const r=t||0;this.setCursor(Math.trunc(e));const a=this.right.splice(this.right.length-r,Number.POSITIVE_INFINITY);return n&&Wr(this.left,n),a.reverse()}pop(){return this.setCursor(Number.POSITIVE_INFINITY),this.left.pop()}push(e){this.setCursor(Number.POSITIVE_INFINITY),this.left.push(e)}pushMany(e){this.setCursor(Number.POSITIVE_INFINITY),Wr(this.left,e)}unshift(e){this.setCursor(0),this.right.push(e)}unshiftMany(e){this.setCursor(0),Wr(this.right,e.reverse())}setCursor(e){if(!(e===this.left.length||e>this.left.length&&0===this.right.length||e<0&&0===this.left.length))if(e<this.left.length){const t=this.left.splice(e,Number.POSITIVE_INFINITY);Wr(this.right,t.reverse())}else{const t=this.right.splice(this.left.length+this.right.length-e,Number.POSITIVE_INFINITY);Wr(this.left,t.reverse())}}}function Wr(e,t){let n=0;if(t.length<1e4)e.push(...t);else for(;n<t.length;)e.push(...t.slice(n,n+1e4)),n+=1e4}function Xr(e){const t={};let n,r,a,o,i,s,l,c=-1;const u=new jr(e);for(;++c<u.length;){for(;c in t;)c=t[c];if(n=u.get(c),c&&"chunkFlow"===n[1].type&&"listItemPrefix"===u.get(c-1)[1].type&&(s=n[1]._tokenizer.events,a=0,a<s.length&&"lineEndingBlank"===s[a][1].type&&(a+=2),a<s.length&&"content"===s[a][1].type))for(;++a<s.length&&"content"!==s[a][1].type;)"chunkText"===s[a][1].type&&(s[a][1]._isInFirstContentOfListItem=!0,a++);if("enter"===n[0])n[1].contentType&&(Object.assign(t,qr(u,c)),c=t[c],l=!0);else if(n[1]._container){for(a=c,r=void 0;a--;)if(o=u.get(a),"lineEnding"===o[1].type||"lineEndingBlank"===o[1].type)"enter"===o[0]&&(r&&(u.get(r)[1].type="lineEndingBlank"),o[1].type="lineEnding",r=a);else if("linePrefix"!==o[1].type)break;r&&(n[1].end={...u.get(r)[1].start},i=u.slice(r,c),i.unshift(n),u.splice(r,c-r+1,i))}}return er(e,0,Number.POSITIVE_INFINITY,u.slice(0)),!l}function qr(e,t){const n=e.get(t)[1],r=e.get(t)[2];let a=t-1;const o=[],i=n._tokenizer||r.parser[n.contentType](n.start),s=i.events,l=[],c={};let u,d,p=-1,h=n,f=0,m=0;const g=[m];for(;h;){for(;e.get(++a)[1]!==h;);o.push(a),h._tokenizer||(u=r.sliceStream(h),h.next||u.push(null),d&&i.defineSkip(h.start),h._isInFirstContentOfListItem&&(i._gfmTasklistFirstContentOfListItem=!0),i.write(u),h._isInFirstContentOfListItem&&(i._gfmTasklistFirstContentOfListItem=void 0)),d=h,h=h.next}for(h=n;++p<s.length;)"exit"===s[p][0]&&"enter"===s[p-1][0]&&s[p][1].type===s[p-1][1].type&&s[p][1].start.line!==s[p][1].end.line&&(m=p+1,g.push(m),h._tokenizer=void 0,h.previous=void 0,h=h.next);for(i.events=[],h?(h._tokenizer=void 0,h.previous=void 0):g.pop(),p=g.length;p--;){const t=s.slice(g[p],g[p+1]),n=o.pop();l.push([n,n+t.length-1]),e.splice(n,2,t)}for(l.reverse(),p=-1;++p<l.length;)c[f+l[p][0]]=f+l[p][1],f+=l[p][1]-l[p][0]-1;return c}const Vr={tokenize:function(e){const t=e.attempt(this.parser.constructs.contentInitial,(function(n){if(null!==n)return e.enter("lineEnding"),e.consume(n),e.exit("lineEnding"),Or(e,t,"linePrefix");e.consume(n)}),(function(t){return e.enter("paragraph"),r(t)}));let n;return t;function r(t){const r=e.enter("chunkText",{contentType:"text",previous:n});return n&&(n.next=r),n=r,a(t)}function a(t){return null===t?(e.exit("chunkText"),e.exit("paragraph"),void e.consume(t)):Rt(t)?(e.consume(t),e.exit("chunkText"),r):(e.consume(t),a)}}},Hr={tokenize:function(e){const t=this,n=[];let r,a,o,i=0;return s;function s(r){if(i<n.length){const a=n[i];return t.containerState=a[1],e.attempt(a[0].continuation,l,c)(r)}return c(r)}function l(e){if(i++,t.containerState._closeFlow){t.containerState._closeFlow=void 0,r&&b();const n=t.events.length;let a,o=n;for(;o--;)if("exit"===t.events[o][0]&&"chunkFlow"===t.events[o][1].type){a=t.events[o][1].end;break}v(i);let s=n;for(;s<t.events.length;)t.events[s][1].end={...a},s++;return er(t.events,o+1,0,t.events.slice(n)),t.events.length=s,c(e)}return s(e)}function c(a){if(i===n.length){if(!r)return p(a);if(r.currentConstruct&&r.currentConstruct.concrete)return f(a);t.interrupt=Boolean(r.currentConstruct&&!r._gfmTableDynamicInterruptHack)}return t.containerState={},e.check(Zr,u,d)(a)}function u(e){return r&&b(),v(i),p(e)}function d(e){return t.parser.lazy[t.now().line]=i!==n.length,o=t.now().offset,f(e)}function p(n){return t.containerState={},e.attempt(Zr,h,f)(n)}function h(e){return i++,n.push([t.currentConstruct,t.containerState]),p(e)}function f(n){return null===n?(r&&b(),v(0),void e.consume(n)):(r=r||t.parser.flow(t.now()),e.enter("chunkFlow",{_tokenizer:r,contentType:"flow",previous:a}),m(n))}function m(n){return null===n?(g(e.exit("chunkFlow"),!0),v(0),void e.consume(n)):Rt(n)?(e.consume(n),g(e.exit("chunkFlow")),i=0,t.interrupt=void 0,s):(e.consume(n),m)}function g(e,n){const s=t.sliceStream(e);if(n&&s.push(null),e.previous=a,a&&(a.next=e),a=e,r.defineSkip(e.start),r.write(s),t.parser.lazy[e.start.line]){let e=r.events.length;for(;e--;)if(r.events[e][1].start.offset<o&&(!r.events[e][1].end||r.events[e][1].end.offset>o))return;const n=t.events.length;let a,s,l=n;for(;l--;)if("exit"===t.events[l][0]&&"chunkFlow"===t.events[l][1].type){if(a){s=t.events[l][1].end;break}a=!0}for(v(i),e=n;e<t.events.length;)t.events[e][1].end={...s},e++;er(t.events,l+1,0,t.events.slice(n)),t.events.length=e}}function v(r){let a=n.length;for(;a-- >r;){const r=n[a];t.containerState=r[1],r[0].exit.call(t,e)}n.length=r}function b(){r.write([null]),a=void 0,r=void 0,t.containerState._closeFlow=void 0}}},Zr={tokenize:function(e,t,n){return Or(e,e.attempt(this.parser.constructs.document,t,n),"linePrefix",this.parser.constructs.disable.null.includes("codeIndented")?void 0:4)}},Gr={resolve:function(e){return Xr(e),e},tokenize:function(e,t){let n;return function(t){return e.enter("content"),n=e.enter("chunkContent",{contentType:"content"}),r(t)};function r(t){return null===t?a(t):Rt(t)?e.check(Kr,o,a)(t):(e.consume(t),r)}function a(n){return e.exit("chunkContent"),e.exit("content"),t(n)}function o(t){return e.consume(t),e.exit("chunkContent"),n.next=e.enter("chunkContent",{contentType:"content",previous:n}),n=n.next,r}}},Kr={partial:!0,tokenize:function(e,t,n){const r=this;return function(t){return e.exit("chunkContent"),e.enter("lineEnding"),e.consume(t),e.exit("lineEnding"),Or(e,a,"linePrefix")};function a(a){if(null===a||Rt(a))return n(a);const o=r.events[r.events.length-1];return!r.parser.constructs.disable.null.includes("codeIndented")&&o&&"linePrefix"===o[1].type&&o[2].sliceSerialize(o[1],!0).length>=4?t(a):e.interrupt(r.parser.constructs.flow,n,t)(a)}}},Yr={tokenize:function(e){const t=this,n=e.attempt(kr,(function(r){if(null!==r)return e.enter("lineEndingBlank"),e.consume(r),e.exit("lineEndingBlank"),t.currentConstruct=void 0,n;e.consume(r)}),e.attempt(this.parser.constructs.flowInitial,r,Or(e,e.attempt(this.parser.constructs.flow,r,e.attempt(Gr,r)),"linePrefix")));return n;function r(r){if(null!==r)return e.enter("lineEnding"),e.consume(r),e.exit("lineEnding"),t.currentConstruct=void 0,n;e.consume(r)}}},Jr={resolveAll:ra()},ea=na("string"),ta=na("text");function na(e){return{resolveAll:ra("text"===e?aa:void 0),tokenize:function(t){const n=this,r=this.parser.constructs[e],a=t.attempt(r,o,i);return o;function o(e){return l(e)?a(e):i(e)}function i(e){if(null!==e)return t.enter("data"),t.consume(e),s;t.consume(e)}function s(e){return l(e)?(t.exit("data"),a(e)):(t.consume(e),s)}function l(e){if(null===e)return!0;const t=r[e];let a=-1;if(t)for(;++a<t.length;){const e=t[a];if(!e.previous||e.previous.call(n,n.previous))return!0}return!1}}}}function ra(e){return function(t,n){let r,a=-1;for(;++a<=t.length;)void 0===r?t[a]&&"data"===t[a][1].type&&(r=a,a++):t[a]&&"data"===t[a][1].type||(a!==r+2&&(t[r][1].end=t[a-1][1].end,t.splice(r+2,a-r-2),a=r+2),r=void 0);return e?e(t,n):t}}function aa(e,t){let n=0;for(;++n<=e.length;)if((n===e.length||"lineEnding"===e[n][1].type)&&"data"===e[n-1][1].type){const r=e[n-1][1],a=t.sliceStream(r);let o,i=a.length,s=-1,l=0;for(;i--;){const e=a[i];if("string"==typeof e){for(s=e.length;32===e.charCodeAt(s-1);)l++,s--;if(s)break;s=-1}else if(-2===e)o=!0,l++;else if(-1!==e){i++;break}}if(l){const a={type:n===e.length||o||l<2?"lineSuffix":"hardBreakTrailing",start:{_bufferIndex:i?s:r.start._bufferIndex+s,_index:r.start._index+i,line:r.end.line,column:r.end.column-l,offset:r.end.offset-l},end:{...r.end}};r.end={...a.start},r.start.offset===r.end.offset?Object.assign(r,a):(e.splice(n,0,["enter",a,t],["exit",a,t]),n+=2)}n++}return e}const oa={name:"thematicBreak",tokenize:function(e,t,n){let r,a=0;return function(t){return e.enter("thematicBreak"),function(e){return r=e,o(e)}(t)};function o(o){return o===r?(e.enter("thematicBreakSequence"),i(o)):a>=3&&(null===o||Rt(o))?(e.exit("thematicBreak"),t(o)):n(o)}function i(t){return t===r?(e.consume(t),a++,i):(e.exit("thematicBreakSequence"),Dt(t)?Or(e,o,"whitespace")(t):o(t))}}},ia={continuation:{tokenize:function(e,t,n){const r=this;return r.containerState._closeFlow=void 0,e.check(kr,(function(n){return r.containerState.furtherBlankLines=r.containerState.furtherBlankLines||r.containerState.initialBlankLine,Or(e,t,"listItemIndent",r.containerState.size+1)(n)}),(function(n){return r.containerState.furtherBlankLines||!Dt(n)?(r.containerState.furtherBlankLines=void 0,r.containerState.initialBlankLine=void 0,a(n)):(r.containerState.furtherBlankLines=void 0,r.containerState.initialBlankLine=void 0,e.attempt(la,t,a)(n))}));function a(a){return r.containerState._closeFlow=!0,r.interrupt=void 0,Or(e,e.attempt(ia,t,n),"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(a)}}},exit:function(e){e.exit(this.containerState.type)},name:"list",tokenize:function(e,t,n){const r=this,a=r.events[r.events.length-1];let o=a&&"linePrefix"===a[1].type?a[2].sliceSerialize(a[1],!0).length:0,i=0;return function(t){const a=r.containerState.type||(42===t||43===t||45===t?"listUnordered":"listOrdered");if("listUnordered"===a?!r.containerState.marker||t===r.containerState.marker:Ct(t)){if(r.containerState.type||(r.containerState.type=a,e.enter(a,{_container:!0})),"listUnordered"===a)return e.enter("listItemPrefix"),42===t||45===t?e.check(oa,n,l)(t):l(t);if(!r.interrupt||49===t)return e.enter("listItemPrefix"),e.enter("listItemValue"),s(t)}return n(t)};function s(t){return Ct(t)&&++i<10?(e.consume(t),s):(!r.interrupt||i<2)&&(r.containerState.marker?t===r.containerState.marker:41===t||46===t)?(e.exit("listItemValue"),l(t)):n(t)}function l(t){return e.enter("listItemMarker"),e.consume(t),e.exit("listItemMarker"),r.containerState.marker=r.containerState.marker||t,e.check(kr,r.interrupt?n:c,e.attempt(sa,d,u))}function c(e){return r.containerState.initialBlankLine=!0,o++,d(e)}function u(t){return Dt(t)?(e.enter("listItemPrefixWhitespace"),e.consume(t),e.exit("listItemPrefixWhitespace"),d):n(t)}function d(n){return r.containerState.size=o+r.sliceSerialize(e.exit("listItemPrefix"),!0).length,t(n)}}},sa={partial:!0,tokenize:function(e,t,n){const r=this;return Or(e,(function(e){const a=r.events[r.events.length-1];return!Dt(e)&&a&&"listItemPrefixWhitespace"===a[1].type?t(e):n(e)}),"listItemPrefixWhitespace",r.parser.constructs.disable.null.includes("codeIndented")?void 0:5)}},la={partial:!0,tokenize:function(e,t,n){const r=this;return Or(e,(function(e){const a=r.events[r.events.length-1];return a&&"listItemIndent"===a[1].type&&a[2].sliceSerialize(a[1],!0).length===r.containerState.size?t(e):n(e)}),"listItemIndent",r.containerState.size+1)}},ca={continuation:{tokenize:function(e,t,n){const r=this;return function(t){return Dt(t)?Or(e,a,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(t):a(t)};function a(r){return e.attempt(ca,t,n)(r)}}},exit:function(e){e.exit("blockQuote")},name:"blockQuote",tokenize:function(e,t,n){const r=this;return function(t){if(62===t){const n=r.containerState;return n.open||(e.enter("blockQuote",{_container:!0}),n.open=!0),e.enter("blockQuotePrefix"),e.enter("blockQuoteMarker"),e.consume(t),e.exit("blockQuoteMarker"),a}return n(t)};function a(n){return Dt(n)?(e.enter("blockQuotePrefixWhitespace"),e.consume(n),e.exit("blockQuotePrefixWhitespace"),e.exit("blockQuotePrefix"),t):(e.exit("blockQuotePrefix"),t(n))}}};function ua(e,t,n,r,a,o,i,s,l){const c=l||Number.POSITIVE_INFINITY;let u=0;return function(t){return 60===t?(e.enter(r),e.enter(a),e.enter(o),e.consume(t),e.exit(o),d):null===t||32===t||41===t||_t(t)?n(t):(e.enter(r),e.enter(i),e.enter(s),e.enter("chunkString",{contentType:"string"}),f(t))};function d(n){return 62===n?(e.enter(o),e.consume(n),e.exit(o),e.exit(a),e.exit(r),t):(e.enter(s),e.enter("chunkString",{contentType:"string"}),p(n))}function p(t){return 62===t?(e.exit("chunkString"),e.exit(s),d(t)):null===t||60===t||Rt(t)?n(t):(e.consume(t),92===t?h:p)}function h(t){return 60===t||62===t||92===t?(e.consume(t),p):p(t)}function f(a){return u||null!==a&&41!==a&&!Pt(a)?u<c&&40===a?(e.consume(a),u++,f):41===a?(e.consume(a),u--,f):null===a||32===a||40===a||_t(a)?n(a):(e.consume(a),92===a?m:f):(e.exit("chunkString"),e.exit(s),e.exit(i),e.exit(r),t(a))}function m(t){return 40===t||41===t||92===t?(e.consume(t),f):f(t)}}function da(e,t,n,r,a,o){const i=this;let s,l=0;return function(t){return e.enter(r),e.enter(a),e.consume(t),e.exit(a),e.enter(o),c};function c(d){return l>999||null===d||91===d||93===d&&!s||94===d&&!l&&"_hiddenFootnoteSupport"in i.parser.constructs?n(d):93===d?(e.exit(o),e.enter(a),e.consume(d),e.exit(a),e.exit(r),t):Rt(d)?(e.enter("lineEnding"),e.consume(d),e.exit("lineEnding"),c):(e.enter("chunkString",{contentType:"string"}),u(d))}function u(t){return null===t||91===t||93===t||Rt(t)||l++>999?(e.exit("chunkString"),c(t)):(e.consume(t),s||(s=!Dt(t)),92===t?d:u)}function d(t){return 91===t||92===t||93===t?(e.consume(t),l++,u):u(t)}}function pa(e,t,n,r,a,o){let i;return function(t){return 34===t||39===t||40===t?(e.enter(r),e.enter(a),e.consume(t),e.exit(a),i=40===t?41:t,s):n(t)};function s(n){return n===i?(e.enter(a),e.consume(n),e.exit(a),e.exit(r),t):(e.enter(o),l(n))}function l(t){return t===i?(e.exit(o),s(i)):null===t?n(t):Rt(t)?(e.enter("lineEnding"),e.consume(t),e.exit("lineEnding"),Or(e,l,"linePrefix")):(e.enter("chunkString",{contentType:"string"}),c(t))}function c(t){return t===i||null===t||Rt(t)?(e.exit("chunkString"),l(t)):(e.consume(t),92===t?u:c)}function u(t){return t===i||92===t?(e.consume(t),c):c(t)}}function ha(e,t){let n;return function r(a){return Rt(a)?(e.enter("lineEnding"),e.consume(a),e.exit("lineEnding"),n=!0,r):Dt(a)?Or(e,r,n?"linePrefix":"lineSuffix")(a):t(a)}}const fa={name:"definition",tokenize:function(e,t,n){const r=this;let a;return function(t){return e.enter("definition"),function(t){return da.call(r,e,o,n,"definitionLabel","definitionLabelMarker","definitionLabelString")(t)}(t)};function o(t){return a=Zt(r.sliceSerialize(r.events[r.events.length-1][1]).slice(1,-1)),58===t?(e.enter("definitionMarker"),e.consume(t),e.exit("definitionMarker"),i):n(t)}function i(t){return Pt(t)?ha(e,s)(t):s(t)}function s(t){return ua(e,l,n,"definitionDestination","definitionDestinationLiteral","definitionDestinationLiteralMarker","definitionDestinationRaw","definitionDestinationString")(t)}function l(t){return e.attempt(ma,c,c)(t)}function c(t){return Dt(t)?Or(e,u,"whitespace")(t):u(t)}function u(o){return null===o||Rt(o)?(e.exit("definition"),r.parser.defined.push(a),t(o)):n(o)}}},ma={partial:!0,tokenize:function(e,t,n){return function(t){return Pt(t)?ha(e,r)(t):n(t)};function r(t){return pa(e,a,n,"definitionTitle","definitionTitleMarker","definitionTitleString")(t)}function a(t){return Dt(t)?Or(e,o,"whitespace")(t):o(t)}function o(e){return null===e||Rt(e)?t(e):n(e)}}},ga={name:"codeIndented",tokenize:function(e,t,n){const r=this;return function(t){return e.enter("codeIndented"),Or(e,a,"linePrefix",5)(t)};function a(e){const t=r.events[r.events.length-1];return t&&"linePrefix"===t[1].type&&t[2].sliceSerialize(t[1],!0).length>=4?o(e):n(e)}function o(t){return null===t?s(t):Rt(t)?e.attempt(va,o,s)(t):(e.enter("codeFlowValue"),i(t))}function i(t){return null===t||Rt(t)?(e.exit("codeFlowValue"),o(t)):(e.consume(t),i)}function s(n){return e.exit("codeIndented"),t(n)}}},va={partial:!0,tokenize:function(e,t,n){const r=this;return a;function a(t){return r.parser.lazy[r.now().line]?n(t):Rt(t)?(e.enter("lineEnding"),e.consume(t),e.exit("lineEnding"),a):Or(e,o,"linePrefix",5)(t)}function o(e){const o=r.events[r.events.length-1];return o&&"linePrefix"===o[1].type&&o[2].sliceSerialize(o[1],!0).length>=4?t(e):Rt(e)?a(e):n(e)}}},ba={name:"headingAtx",resolve:function(e,t){let n,r,a=e.length-2,o=3;return"whitespace"===e[o][1].type&&(o+=2),a-2>o&&"whitespace"===e[a][1].type&&(a-=2),"atxHeadingSequence"===e[a][1].type&&(o===a-1||a-4>o&&"whitespace"===e[a-2][1].type)&&(a-=o+1===a?2:4),a>o&&(n={type:"atxHeadingText",start:e[o][1].start,end:e[a][1].end},r={type:"chunkText",start:e[o][1].start,end:e[a][1].end,contentType:"text"},er(e,o,a-o+1,[["enter",n,t],["enter",r,t],["exit",r,t],["exit",n,t]])),e},tokenize:function(e,t,n){let r=0;return function(t){return e.enter("atxHeading"),function(t){return e.enter("atxHeadingSequence"),a(t)}(t)};function a(t){return 35===t&&r++<6?(e.consume(t),a):null===t||Pt(t)?(e.exit("atxHeadingSequence"),o(t)):n(t)}function o(n){return 35===n?(e.enter("atxHeadingSequence"),i(n)):null===n||Rt(n)?(e.exit("atxHeading"),t(n)):Dt(n)?Or(e,o,"whitespace")(n):(e.enter("atxHeadingText"),s(n))}function i(t){return 35===t?(e.consume(t),i):(e.exit("atxHeadingSequence"),o(t))}function s(t){return null===t||35===t||Pt(t)?(e.exit("atxHeadingText"),o(t)):(e.consume(t),s)}}},ya={name:"setextUnderline",resolveTo:function(e,t){let n,r,a,o=e.length;for(;o--;)if("enter"===e[o][0]){if("content"===e[o][1].type){n=o;break}"paragraph"===e[o][1].type&&(r=o)}else"content"===e[o][1].type&&e.splice(o,1),a||"definition"!==e[o][1].type||(a=o);const i={type:"setextHeading",start:{...e[r][1].start},end:{...e[e.length-1][1].end}};return e[r][1].type="setextHeadingText",a?(e.splice(r,0,["enter",i,t]),e.splice(a+1,0,["exit",e[n][1],t]),e[n][1].end={...e[a][1].end}):e[n][1]=i,e.push(["exit",i,t]),e},tokenize:function(e,t,n){const r=this;let a;return function(t){let i,s=r.events.length;for(;s--;)if("lineEnding"!==r.events[s][1].type&&"linePrefix"!==r.events[s][1].type&&"content"!==r.events[s][1].type){i="paragraph"===r.events[s][1].type;break}return r.parser.lazy[r.now().line]||!r.interrupt&&!i?n(t):(e.enter("setextHeadingLine"),a=t,function(t){return e.enter("setextHeadingLineSequence"),o(t)}(t))};function o(t){return t===a?(e.consume(t),o):(e.exit("setextHeadingLineSequence"),Dt(t)?Or(e,i,"lineSuffix")(t):i(t))}function i(r){return null===r||Rt(r)?(e.exit("setextHeadingLine"),t(r)):n(r)}}},wa=["address","article","aside","base","basefont","blockquote","body","caption","center","col","colgroup","dd","details","dialog","dir","div","dl","dt","fieldset","figcaption","figure","footer","form","frame","frameset","h1","h2","h3","h4","h5","h6","head","header","hr","html","iframe","legend","li","link","main","menu","menuitem","nav","noframes","ol","optgroup","option","p","param","search","section","summary","table","tbody","td","tfoot","th","thead","title","tr","track","ul"],Oa=["pre","script","style","textarea"],ka={concrete:!0,name:"htmlFlow",resolveTo:function(e){let t=e.length;for(;t--&&("enter"!==e[t][0]||"htmlFlow"!==e[t][1].type););return t>1&&"linePrefix"===e[t-2][1].type&&(e[t][1].start=e[t-2][1].start,e[t+1][1].start=e[t-2][1].start,e.splice(t-2,2)),e},tokenize:function(e,t,n){const r=this;let a,o,i,s,l;return function(t){return function(t){return e.enter("htmlFlow"),e.enter("htmlFlowData"),e.consume(t),c}(t)};function c(s){return 33===s?(e.consume(s),u):47===s?(e.consume(s),o=!0,h):63===s?(e.consume(s),a=3,r.interrupt?t:$):xt(s)?(e.consume(s),i=String.fromCharCode(s),f):n(s)}function u(o){return 45===o?(e.consume(o),a=2,d):91===o?(e.consume(o),a=5,s=0,p):xt(o)?(e.consume(o),a=4,r.interrupt?t:$):n(o)}function d(a){return 45===a?(e.consume(a),r.interrupt?t:$):n(a)}function p(a){return a==="CDATA[".charCodeAt(s++)?(e.consume(a),6===s?r.interrupt?t:_:p):n(a)}function h(t){return xt(t)?(e.consume(t),i=String.fromCharCode(t),f):n(t)}function f(s){if(null===s||47===s||62===s||Pt(s)){const l=47===s,c=i.toLowerCase();return l||o||!Oa.includes(c)?wa.includes(i.toLowerCase())?(a=6,l?(e.consume(s),m):r.interrupt?t(s):_(s)):(a=7,r.interrupt&&!r.parser.lazy[r.now().line]?n(s):o?g(s):v(s)):(a=1,r.interrupt?t(s):_(s))}return 45===s||Et(s)?(e.consume(s),i+=String.fromCharCode(s),f):n(s)}function m(a){return 62===a?(e.consume(a),r.interrupt?t:_):n(a)}function g(t){return Dt(t)?(e.consume(t),g):E(t)}function v(t){return 47===t?(e.consume(t),E):58===t||95===t||xt(t)?(e.consume(t),b):Dt(t)?(e.consume(t),v):E(t)}function b(t){return 45===t||46===t||58===t||95===t||Et(t)?(e.consume(t),b):y(t)}function y(t){return 61===t?(e.consume(t),w):Dt(t)?(e.consume(t),y):v(t)}function w(t){return null===t||60===t||61===t||62===t||96===t?n(t):34===t||39===t?(e.consume(t),l=t,O):Dt(t)?(e.consume(t),w):k(t)}function O(t){return t===l?(e.consume(t),l=null,x):null===t||Rt(t)?n(t):(e.consume(t),O)}function k(t){return null===t||34===t||39===t||47===t||60===t||61===t||62===t||96===t||Pt(t)?y(t):(e.consume(t),k)}function x(e){return 47===e||62===e||Dt(e)?v(e):n(e)}function E(t){return 62===t?(e.consume(t),S):n(t)}function S(t){return null===t||Rt(t)?_(t):Dt(t)?(e.consume(t),S):n(t)}function _(t){return 45===t&&2===a?(e.consume(t),R):60===t&&1===a?(e.consume(t),P):62===t&&4===a?(e.consume(t),N):63===t&&3===a?(e.consume(t),$):93===t&&5===a?(e.consume(t),M):!Rt(t)||6!==a&&7!==a?null===t||Rt(t)?(e.exit("htmlFlowData"),C(t)):(e.consume(t),_):(e.exit("htmlFlowData"),e.check(xa,I,C)(t))}function C(t){return e.check(Ea,T,I)(t)}function T(t){return e.enter("lineEnding"),e.consume(t),e.exit("lineEnding"),A}function A(t){return null===t||Rt(t)?C(t):(e.enter("htmlFlowData"),_(t))}function R(t){return 45===t?(e.consume(t),$):_(t)}function P(t){return 47===t?(e.consume(t),i="",D):_(t)}function D(t){if(62===t){const n=i.toLowerCase();return Oa.includes(n)?(e.consume(t),N):_(t)}return xt(t)&&i.length<8?(e.consume(t),i+=String.fromCharCode(t),D):_(t)}function M(t){return 93===t?(e.consume(t),$):_(t)}function $(t){return 62===t?(e.consume(t),N):45===t&&2===a?(e.consume(t),$):_(t)}function N(t){return null===t||Rt(t)?(e.exit("htmlFlowData"),I(t)):(e.consume(t),N)}function I(n){return e.exit("htmlFlow"),t(n)}}},xa={partial:!0,tokenize:function(e,t,n){return function(r){return e.enter("lineEnding"),e.consume(r),e.exit("lineEnding"),e.attempt(kr,t,n)}}},Ea={partial:!0,tokenize:function(e,t,n){const r=this;return function(t){return Rt(t)?(e.enter("lineEnding"),e.consume(t),e.exit("lineEnding"),a):n(t)};function a(e){return r.parser.lazy[r.now().line]?n(e):t(e)}}},Sa={partial:!0,tokenize:function(e,t,n){const r=this;return function(t){return null===t?n(t):(e.enter("lineEnding"),e.consume(t),e.exit("lineEnding"),a)};function a(e){return r.parser.lazy[r.now().line]?n(e):t(e)}}},_a={concrete:!0,name:"codeFenced",tokenize:function(e,t,n){const r=this,a={partial:!0,tokenize:function(e,t,n){let a=0;return function(t){return e.enter("lineEnding"),e.consume(t),e.exit("lineEnding"),i};function i(t){return e.enter("codeFencedFence"),Dt(t)?Or(e,l,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(t):l(t)}function l(t){return t===o?(e.enter("codeFencedFenceSequence"),c(t)):n(t)}function c(t){return t===o?(a++,e.consume(t),c):a>=s?(e.exit("codeFencedFenceSequence"),Dt(t)?Or(e,u,"whitespace")(t):u(t)):n(t)}function u(r){return null===r||Rt(r)?(e.exit("codeFencedFence"),t(r)):n(r)}}};let o,i=0,s=0;return function(t){return function(t){const n=r.events[r.events.length-1];return i=n&&"linePrefix"===n[1].type?n[2].sliceSerialize(n[1],!0).length:0,o=t,e.enter("codeFenced"),e.enter("codeFencedFence"),e.enter("codeFencedFenceSequence"),l(t)}(t)};function l(t){return t===o?(s++,e.consume(t),l):s<3?n(t):(e.exit("codeFencedFenceSequence"),Dt(t)?Or(e,c,"whitespace")(t):c(t))}function c(n){return null===n||Rt(n)?(e.exit("codeFencedFence"),r.interrupt?t(n):e.check(Sa,h,b)(n)):(e.enter("codeFencedFenceInfo"),e.enter("chunkString",{contentType:"string"}),u(n))}function u(t){return null===t||Rt(t)?(e.exit("chunkString"),e.exit("codeFencedFenceInfo"),c(t)):Dt(t)?(e.exit("chunkString"),e.exit("codeFencedFenceInfo"),Or(e,d,"whitespace")(t)):96===t&&t===o?n(t):(e.consume(t),u)}function d(t){return null===t||Rt(t)?c(t):(e.enter("codeFencedFenceMeta"),e.enter("chunkString",{contentType:"string"}),p(t))}function p(t){return null===t||Rt(t)?(e.exit("chunkString"),e.exit("codeFencedFenceMeta"),c(t)):96===t&&t===o?n(t):(e.consume(t),p)}function h(t){return e.attempt(a,b,f)(t)}function f(t){return e.enter("lineEnding"),e.consume(t),e.exit("lineEnding"),m}function m(t){return i>0&&Dt(t)?Or(e,g,"linePrefix",i+1)(t):g(t)}function g(t){return null===t||Rt(t)?e.check(Sa,h,b)(t):(e.enter("codeFlowValue"),v(t))}function v(t){return null===t||Rt(t)?(e.exit("codeFlowValue"),g(t)):(e.consume(t),v)}function b(n){return e.exit("codeFenced"),t(n)}}},Ca=document.createElement("i");function Ta(e){const t="&"+e+";";Ca.innerHTML=t;const n=Ca.textContent;return(59!==n.charCodeAt(n.length-1)||"semi"===e)&&n!==t&&n}const Aa={name:"characterReference",tokenize:function(e,t,n){const r=this;let a,o,i=0;return function(t){return e.enter("characterReference"),e.enter("characterReferenceMarker"),e.consume(t),e.exit("characterReferenceMarker"),s};function s(t){return 35===t?(e.enter("characterReferenceMarkerNumeric"),e.consume(t),e.exit("characterReferenceMarkerNumeric"),l):(e.enter("characterReferenceValue"),a=31,o=Et,c(t))}function l(t){return 88===t||120===t?(e.enter("characterReferenceMarkerHexadecimal"),e.consume(t),e.exit("characterReferenceMarkerHexadecimal"),e.enter("characterReferenceValue"),a=6,o=Tt,c):(e.enter("characterReferenceValue"),a=7,o=Ct,c(t))}function c(s){if(59===s&&i){const a=e.exit("characterReferenceValue");return o!==Et||Ta(r.sliceSerialize(a))?(e.enter("characterReferenceMarker"),e.consume(s),e.exit("characterReferenceMarker"),e.exit("characterReference"),t):n(s)}return o(s)&&i++<a?(e.consume(s),c):n(s)}}},Ra={name:"characterEscape",tokenize:function(e,t,n){return function(t){return e.enter("characterEscape"),e.enter("escapeMarker"),e.consume(t),e.exit("escapeMarker"),r};function r(r){return At(r)?(e.enter("characterEscapeValue"),e.consume(r),e.exit("characterEscapeValue"),e.exit("characterEscape"),t):n(r)}}},Pa={name:"lineEnding",tokenize:function(e,t){return function(n){return e.enter("lineEnding"),e.consume(n),e.exit("lineEnding"),Or(e,t,"linePrefix")}}},Da={name:"labelEnd",resolveAll:function(e){let t=-1;const n=[];for(;++t<e.length;){const r=e[t][1];if(n.push(e[t]),"labelImage"===r.type||"labelLink"===r.type||"labelEnd"===r.type){const e="labelImage"===r.type?4:2;r.type="data",t+=e}}return e.length!==n.length&&er(e,0,e.length,n),e},resolveTo:function(e,t){let n,r,a,o,i=e.length,s=0;for(;i--;)if(n=e[i][1],r){if("link"===n.type||"labelLink"===n.type&&n._inactive)break;"enter"===e[i][0]&&"labelLink"===n.type&&(n._inactive=!0)}else if(a){if("enter"===e[i][0]&&("labelImage"===n.type||"labelLink"===n.type)&&!n._balanced&&(r=i,"labelLink"!==n.type)){s=2;break}}else"labelEnd"===n.type&&(a=i);const l={type:"labelLink"===e[r][1].type?"link":"image",start:{...e[r][1].start},end:{...e[e.length-1][1].end}},c={type:"label",start:{...e[r][1].start},end:{...e[a][1].end}},u={type:"labelText",start:{...e[r+s+2][1].end},end:{...e[a-2][1].start}};return o=[["enter",l,t],["enter",c,t]],o=tr(o,e.slice(r+1,r+s+3)),o=tr(o,[["enter",u,t]]),o=tr(o,Rr(t.parser.constructs.insideSpan.null,e.slice(r+s+4,a-3),t)),o=tr(o,[["exit",u,t],e[a-2],e[a-1],["exit",c,t]]),o=tr(o,e.slice(a+1)),o=tr(o,[["exit",l,t]]),er(e,r,e.length,o),e},tokenize:function(e,t,n){const r=this;let a,o,i=r.events.length;for(;i--;)if(("labelImage"===r.events[i][1].type||"labelLink"===r.events[i][1].type)&&!r.events[i][1]._balanced){a=r.events[i][1];break}return function(t){return a?a._inactive?u(t):(o=r.parser.defined.includes(Zt(r.sliceSerialize({start:a.end,end:r.now()}))),e.enter("labelEnd"),e.enter("labelMarker"),e.consume(t),e.exit("labelMarker"),e.exit("labelEnd"),s):n(t)};function s(t){return 40===t?e.attempt(Ma,c,o?c:u)(t):91===t?e.attempt($a,c,o?l:u)(t):o?c(t):u(t)}function l(t){return e.attempt(Na,c,u)(t)}function c(e){return t(e)}function u(e){return a._balanced=!0,n(e)}}},Ma={tokenize:function(e,t,n){return function(t){return e.enter("resource"),e.enter("resourceMarker"),e.consume(t),e.exit("resourceMarker"),r};function r(t){return Pt(t)?ha(e,a)(t):a(t)}function a(t){return 41===t?c(t):ua(e,o,i,"resourceDestination","resourceDestinationLiteral","resourceDestinationLiteralMarker","resourceDestinationRaw","resourceDestinationString",32)(t)}function o(t){return Pt(t)?ha(e,s)(t):c(t)}function i(e){return n(e)}function s(t){return 34===t||39===t||40===t?pa(e,l,n,"resourceTitle","resourceTitleMarker","resourceTitleString")(t):c(t)}function l(t){return Pt(t)?ha(e,c)(t):c(t)}function c(r){return 41===r?(e.enter("resourceMarker"),e.consume(r),e.exit("resourceMarker"),e.exit("resource"),t):n(r)}}},$a={tokenize:function(e,t,n){const r=this;return function(t){return da.call(r,e,a,o,"reference","referenceMarker","referenceString")(t)};function a(e){return r.parser.defined.includes(Zt(r.sliceSerialize(r.events[r.events.length-1][1]).slice(1,-1)))?t(e):n(e)}function o(e){return n(e)}}},Na={tokenize:function(e,t,n){return function(t){return e.enter("reference"),e.enter("referenceMarker"),e.consume(t),e.exit("referenceMarker"),r};function r(r){return 93===r?(e.enter("referenceMarker"),e.consume(r),e.exit("referenceMarker"),e.exit("reference"),t):n(r)}}},Ia={name:"labelStartImage",resolveAll:Da.resolveAll,tokenize:function(e,t,n){const r=this;return function(t){return e.enter("labelImage"),e.enter("labelImageMarker"),e.consume(t),e.exit("labelImageMarker"),a};function a(t){return 91===t?(e.enter("labelMarker"),e.consume(t),e.exit("labelMarker"),e.exit("labelImage"),o):n(t)}function o(e){return 94===e&&"_hiddenFootnoteSupport"in r.parser.constructs?n(e):t(e)}}},La={name:"attention",resolveAll:function(e,t){let n,r,a,o,i,s,l,c,u=-1;for(;++u<e.length;)if("enter"===e[u][0]&&"attentionSequence"===e[u][1].type&&e[u][1]._close)for(n=u;n--;)if("exit"===e[n][0]&&"attentionSequence"===e[n][1].type&&e[n][1]._open&&t.sliceSerialize(e[n][1]).charCodeAt(0)===t.sliceSerialize(e[u][1]).charCodeAt(0)){if((e[n][1]._close||e[u][1]._open)&&(e[u][1].end.offset-e[u][1].start.offset)%3&&!((e[n][1].end.offset-e[n][1].start.offset+e[u][1].end.offset-e[u][1].start.offset)%3))continue;s=e[n][1].end.offset-e[n][1].start.offset>1&&e[u][1].end.offset-e[u][1].start.offset>1?2:1;const d={...e[n][1].end},p={...e[u][1].start};Ba(d,-s),Ba(p,s),o={type:s>1?"strongSequence":"emphasisSequence",start:d,end:{...e[n][1].end}},i={type:s>1?"strongSequence":"emphasisSequence",start:{...e[u][1].start},end:p},a={type:s>1?"strongText":"emphasisText",start:{...e[n][1].end},end:{...e[u][1].start}},r={type:s>1?"strong":"emphasis",start:{...o.start},end:{...i.end}},e[n][1].end={...o.start},e[u][1].start={...i.end},l=[],e[n][1].end.offset-e[n][1].start.offset&&(l=tr(l,[["enter",e[n][1],t],["exit",e[n][1],t]])),l=tr(l,[["enter",r,t],["enter",o,t],["exit",o,t],["enter",a,t]]),l=tr(l,Rr(t.parser.constructs.insideSpan.null,e.slice(n+1,u),t)),l=tr(l,[["exit",a,t],["enter",i,t],["exit",i,t],["exit",r,t]]),e[u][1].end.offset-e[u][1].start.offset?(c=2,l=tr(l,[["enter",e[u][1],t],["exit",e[u][1],t]])):c=0,er(e,n-1,u-n+3,l),u=n+l.length-c-2;break}for(u=-1;++u<e.length;)"attentionSequence"===e[u][1].type&&(e[u][1].type="data");return e},tokenize:function(e,t){const n=this.parser.constructs.attentionMarkers.null,r=this.previous,a=xn(r);let o;return function(t){return o=t,e.enter("attentionSequence"),i(t)};function i(s){if(s===o)return e.consume(s),i;const l=e.exit("attentionSequence"),c=xn(s),u=!c||2===c&&a||n.includes(s),d=!a||2===a&&c||n.includes(r);return l._open=Boolean(42===o?u:u&&(a||!d)),l._close=Boolean(42===o?d:d&&(c||!u)),t(s)}}};function Ba(e,t){e.column+=t,e.offset+=t,e._bufferIndex+=t}const Fa={name:"autolink",tokenize:function(e,t,n){let r=0;return function(t){return e.enter("autolink"),e.enter("autolinkMarker"),e.consume(t),e.exit("autolinkMarker"),e.enter("autolinkProtocol"),a};function a(t){return xt(t)?(e.consume(t),o):64===t?n(t):l(t)}function o(e){return 43===e||45===e||46===e||Et(e)?(r=1,i(e)):l(e)}function i(t){return 58===t?(e.consume(t),r=0,s):(43===t||45===t||46===t||Et(t))&&r++<32?(e.consume(t),i):(r=0,l(t))}function s(r){return 62===r?(e.exit("autolinkProtocol"),e.enter("autolinkMarker"),e.consume(r),e.exit("autolinkMarker"),e.exit("autolink"),t):null===r||32===r||60===r||_t(r)?n(r):(e.consume(r),s)}function l(t){return 64===t?(e.consume(t),c):St(t)?(e.consume(t),l):n(t)}function c(e){return Et(e)?u(e):n(e)}function u(n){return 46===n?(e.consume(n),r=0,c):62===n?(e.exit("autolinkProtocol").type="autolinkEmail",e.enter("autolinkMarker"),e.consume(n),e.exit("autolinkMarker"),e.exit("autolink"),t):d(n)}function d(t){if((45===t||Et(t))&&r++<63){const n=45===t?d:u;return e.consume(t),n}return n(t)}}},Qa={name:"htmlText",tokenize:function(e,t,n){const r=this;let a,o,i;return function(t){return e.enter("htmlText"),e.enter("htmlTextData"),e.consume(t),s};function s(t){return 33===t?(e.consume(t),l):47===t?(e.consume(t),w):63===t?(e.consume(t),b):xt(t)?(e.consume(t),x):n(t)}function l(t){return 45===t?(e.consume(t),c):91===t?(e.consume(t),o=0,h):xt(t)?(e.consume(t),v):n(t)}function c(t){return 45===t?(e.consume(t),p):n(t)}function u(t){return null===t?n(t):45===t?(e.consume(t),d):Rt(t)?(i=u,D(t)):(e.consume(t),u)}function d(t){return 45===t?(e.consume(t),p):u(t)}function p(e){return 62===e?P(e):45===e?d(e):u(e)}function h(t){return t==="CDATA[".charCodeAt(o++)?(e.consume(t),6===o?f:h):n(t)}function f(t){return null===t?n(t):93===t?(e.consume(t),m):Rt(t)?(i=f,D(t)):(e.consume(t),f)}function m(t){return 93===t?(e.consume(t),g):f(t)}function g(t){return 62===t?P(t):93===t?(e.consume(t),g):f(t)}function v(t){return null===t||62===t?P(t):Rt(t)?(i=v,D(t)):(e.consume(t),v)}function b(t){return null===t?n(t):63===t?(e.consume(t),y):Rt(t)?(i=b,D(t)):(e.consume(t),b)}function y(e){return 62===e?P(e):b(e)}function w(t){return xt(t)?(e.consume(t),O):n(t)}function O(t){return 45===t||Et(t)?(e.consume(t),O):k(t)}function k(t){return Rt(t)?(i=k,D(t)):Dt(t)?(e.consume(t),k):P(t)}function x(t){return 45===t||Et(t)?(e.consume(t),x):47===t||62===t||Pt(t)?E(t):n(t)}function E(t){return 47===t?(e.consume(t),P):58===t||95===t||xt(t)?(e.consume(t),S):Rt(t)?(i=E,D(t)):Dt(t)?(e.consume(t),E):P(t)}function S(t){return 45===t||46===t||58===t||95===t||Et(t)?(e.consume(t),S):_(t)}function _(t){return 61===t?(e.consume(t),C):Rt(t)?(i=_,D(t)):Dt(t)?(e.consume(t),_):E(t)}function C(t){return null===t||60===t||61===t||62===t||96===t?n(t):34===t||39===t?(e.consume(t),a=t,T):Rt(t)?(i=C,D(t)):Dt(t)?(e.consume(t),C):(e.consume(t),A)}function T(t){return t===a?(e.consume(t),a=void 0,R):null===t?n(t):Rt(t)?(i=T,D(t)):(e.consume(t),T)}function A(t){return null===t||34===t||39===t||60===t||61===t||96===t?n(t):47===t||62===t||Pt(t)?E(t):(e.consume(t),A)}function R(e){return 47===e||62===e||Pt(e)?E(e):n(e)}function P(r){return 62===r?(e.consume(r),e.exit("htmlTextData"),e.exit("htmlText"),t):n(r)}function D(t){return e.exit("htmlTextData"),e.enter("lineEnding"),e.consume(t),e.exit("lineEnding"),M}function M(t){return Dt(t)?Or(e,$,"linePrefix",r.parser.constructs.disable.null.includes("codeIndented")?void 0:4)(t):$(t)}function $(t){return e.enter("htmlTextData"),i(t)}}},Ua={name:"labelStartLink",resolveAll:Da.resolveAll,tokenize:function(e,t,n){const r=this;return function(t){return e.enter("labelLink"),e.enter("labelMarker"),e.consume(t),e.exit("labelMarker"),e.exit("labelLink"),a};function a(e){return 94===e&&"_hiddenFootnoteSupport"in r.parser.constructs?n(e):t(e)}}},za={name:"hardBreakEscape",tokenize:function(e,t,n){return function(t){return e.enter("hardBreakEscape"),e.consume(t),r};function r(r){return Rt(r)?(e.exit("hardBreakEscape"),t(r)):n(r)}}},ja={name:"codeText",previous:function(e){return 96!==e||"characterEscape"===this.events[this.events.length-1][1].type},resolve:function(e){let t,n,r=e.length-4,a=3;if(!("lineEnding"!==e[a][1].type&&"space"!==e[a][1].type||"lineEnding"!==e[r][1].type&&"space"!==e[r][1].type))for(t=a;++t<r;)if("codeTextData"===e[t][1].type){e[a][1].type="codeTextPadding",e[r][1].type="codeTextPadding",a+=2,r-=2;break}for(t=a-1,r++;++t<=r;)void 0===n?t!==r&&"lineEnding"!==e[t][1].type&&(n=t):t!==r&&"lineEnding"!==e[t][1].type||(e[n][1].type="codeTextData",t!==n+2&&(e[n][1].end=e[t-1][1].end,e.splice(n+2,t-n-2),r-=t-n-2,t=n+2),n=void 0);return e},tokenize:function(e,t,n){let r,a,o=0;return function(t){return e.enter("codeText"),e.enter("codeTextSequence"),i(t)};function i(t){return 96===t?(e.consume(t),o++,i):(e.exit("codeTextSequence"),s(t))}function s(t){return null===t?n(t):32===t?(e.enter("space"),e.consume(t),e.exit("space"),s):96===t?(a=e.enter("codeTextSequence"),r=0,c(t)):Rt(t)?(e.enter("lineEnding"),e.consume(t),e.exit("lineEnding"),s):(e.enter("codeTextData"),l(t))}function l(t){return null===t||32===t||96===t||Rt(t)?(e.exit("codeTextData"),s(t)):(e.consume(t),l)}function c(n){return 96===n?(e.consume(n),r++,c):r===o?(e.exit("codeTextSequence"),e.exit("codeText"),t(n)):(a.type="codeTextData",l(n))}}},Wa={42:ia,43:ia,45:ia,48:ia,49:ia,50:ia,51:ia,52:ia,53:ia,54:ia,55:ia,56:ia,57:ia,62:ca},Xa={91:fa},qa={[-2]:ga,[-1]:ga,32:ga},Va={35:ba,42:oa,45:[ya,oa],60:ka,61:ya,95:oa,96:_a,126:_a},Ha={38:Aa,92:Ra},Za={[-5]:Pa,[-4]:Pa,[-3]:Pa,33:Ia,38:Aa,42:La,60:[Fa,Qa],91:Ua,92:[za,Ra],93:Da,95:La,96:ja},Ga={null:[La,Jr]},Ka={null:[42,95]},Ya={null:[]};function Ja(e,t,n){let r={_bufferIndex:-1,_index:0,line:n&&n.line||1,column:n&&n.column||1,offset:n&&n.offset||0};const a={},o=[];let i=[],s=[],l=!0;const c={attempt:v((function(e,t){b(e,t.from)})),check:v(g),consume:function(e){Rt(e)?(r.line++,r.column=1,r.offset+=-3===e?2:1,y()):-1!==e&&(r.column++,r.offset++),r._bufferIndex<0?r._index++:(r._bufferIndex++,r._bufferIndex===i[r._index].length&&(r._bufferIndex=-1,r._index++)),u.previous=e,l=!0},enter:function(e,t){const n=t||{};return n.type=e,n.start=f(),u.events.push(["enter",n,u]),s.push(n),n},exit:function(e){const t=s.pop();return t.end=f(),u.events.push(["exit",t,u]),t},interrupt:v(g,{interrupt:!0})},u={code:null,containerState:{},defineSkip:function(e){a[e.line]=e.column,y()},events:[],now:f,parser:e,previous:null,sliceSerialize:function(e,t){return function(e,t){let n=-1;const r=[];let a;for(;++n<e.length;){const o=e[n];let i;if("string"==typeof o)i=o;else switch(o){case-5:i="\r";break;case-4:i="\n";break;case-3:i="\r\n";break;case-2:i=t?" ":"\t";break;case-1:if(!t&&a)continue;i=" ";break;default:i=String.fromCharCode(o)}a=-2===o,r.push(i)}return r.join("")}(h(e),t)},sliceStream:h,write:function(e){return i=tr(i,e),function(){let e;for(;r._index<i.length;){const t=i[r._index];if("string"==typeof t)for(e=r._index,r._bufferIndex<0&&(r._bufferIndex=0);r._index===e&&r._bufferIndex<t.length;)m(t.charCodeAt(r._bufferIndex));else m(t)}}(),null!==i[i.length-1]?[]:(b(t,0),u.events=Rr(o,u.events,u),u.events)}};let d,p=t.tokenize.call(u,c);return t.resolveAll&&o.push(t),u;function h(e){return function(e,t){const n=t.start._index,r=t.start._bufferIndex,a=t.end._index,o=t.end._bufferIndex;let i;if(n===a)i=[e[n].slice(r,o)];else{if(i=e.slice(n,a),r>-1){const e=i[0];"string"==typeof e?i[0]=e.slice(r):i.shift()}o>0&&i.push(e[a].slice(0,o))}return i}(i,e)}function f(){const{_bufferIndex:e,_index:t,line:n,column:a,offset:o}=r;return{_bufferIndex:e,_index:t,line:n,column:a,offset:o}}function m(e){l=void 0,d=e,p=p(e)}function g(e,t){t.restore()}function v(e,t){return function(n,a,o){let i,d,p,h;return Array.isArray(n)?m(n):"tokenize"in n?m([n]):function(e){return function(t){const n=null!==t&&e[t],r=null!==t&&e.null;return m([...Array.isArray(n)?n:n?[n]:[],...Array.isArray(r)?r:r?[r]:[]])(t)}}(n);function m(e){return i=e,d=0,0===e.length?o:g(e[d])}function g(e){return function(n){return h=function(){const e=f(),t=u.previous,n=u.currentConstruct,a=u.events.length,o=Array.from(s);return{from:a,restore:function(){r=e,u.previous=t,u.currentConstruct=n,u.events.length=a,s=o,y()}}}(),p=e,e.partial||(u.currentConstruct=e),e.name&&u.parser.constructs.disable.null.includes(e.name)?b():e.tokenize.call(t?Object.assign(Object.create(u),t):u,c,v,b)(n)}}function v(t){return l=!0,e(p,h),a}function b(e){return l=!0,h.restore(),++d<i.length?g(i[d]):o}}}function b(e,t){e.resolveAll&&!o.includes(e)&&o.push(e),e.resolve&&er(u.events,t,u.events.length-t,e.resolve(u.events.slice(t),u)),e.resolveTo&&(u.events=e.resolveTo(u.events,u))}function y(){r.line in a&&r.column<2&&(r.column=a[r.line],r.offset+=a[r.line]-1)}}const eo=/[\0\t\n\r]/g;function to(e,t){const n=Number.parseInt(e,t);return n<9||11===n||n>13&&n<32||n>126&&n<160||n>55295&&n<57344||n>64975&&n<65008||!(65535&~n)||65534==(65535&n)||n>1114111?"�":String.fromCodePoint(n)}const no=/\\([!-/:-@[-`{-~])|&(#(?:\d{1,7}|x[\da-f]{1,6})|[\da-z]{1,31});/gi;function ro(e){return e.replace(no,ao)}function ao(e,t,n){if(t)return t;if(35===n.charCodeAt(0)){const e=n.charCodeAt(1),t=120===e||88===e;return to(n.slice(t?2:1),t?16:10)}return Ta(n)||e}var oo=n(7188);const io={}.hasOwnProperty;function so(e,t,n){return"string"!=typeof t&&(n=t,t=void 0),function(e){const t={transforms:[],canContainEols:["emphasis","fragment","heading","paragraph","strong"],enter:{autolink:a(b),autolinkProtocol:u,autolinkEmail:u,atxHeading:a(m),blockQuote:a((function(){return{type:"blockquote",children:[]}})),characterEscape:u,characterReference:u,codeFenced:a(f),codeFencedFenceInfo:o,codeFencedFenceMeta:o,codeIndented:a(f,o),codeText:a((function(){return{type:"inlineCode",value:""}}),o),codeTextData:u,data:u,codeFlowValue:u,definition:a((function(){return{type:"definition",identifier:"",label:null,title:null,url:""}})),definitionDestinationString:o,definitionLabelString:o,definitionTitleString:o,emphasis:a((function(){return{type:"emphasis",children:[]}})),hardBreakEscape:a(g),hardBreakTrailing:a(g),htmlFlow:a(v,o),htmlFlowData:u,htmlText:a(v,o),htmlTextData:u,image:a((function(){return{type:"image",title:null,url:"",alt:null}})),label:o,link:a(b),listItem:a((function(e){return{type:"listItem",spread:e._spread,checked:null,children:[]}})),listItemValue:function(e){this.data.expectingFirstListItemValue&&(this.stack[this.stack.length-2].start=Number.parseInt(this.sliceSerialize(e),10),this.data.expectingFirstListItemValue=void 0)},listOrdered:a(y,(function(){this.data.expectingFirstListItemValue=!0})),listUnordered:a(y),paragraph:a((function(){return{type:"paragraph",children:[]}})),reference:function(){this.data.referenceType="collapsed"},referenceString:o,resourceDestinationString:o,resourceTitleString:o,setextHeading:a(m),strong:a((function(){return{type:"strong",children:[]}})),thematicBreak:a((function(){return{type:"thematicBreak"}}))},exit:{atxHeading:s(),atxHeadingSequence:function(e){const t=this.stack[this.stack.length-1];if(!t.depth){const n=this.sliceSerialize(e).length;t.depth=n}},autolink:s(),autolinkEmail:function(e){d.call(this,e),this.stack[this.stack.length-1].url="mailto:"+this.sliceSerialize(e)},autolinkProtocol:function(e){d.call(this,e),this.stack[this.stack.length-1].url=this.sliceSerialize(e)},blockQuote:s(),characterEscapeValue:d,characterReferenceMarkerHexadecimal:h,characterReferenceMarkerNumeric:h,characterReferenceValue:function(e){const t=this.sliceSerialize(e),n=this.data.characterReferenceType;let r;n?(r=to(t,"characterReferenceMarkerNumeric"===n?10:16),this.data.characterReferenceType=void 0):r=Ta(t);this.stack[this.stack.length-1].value+=r},characterReference:function(e){this.stack.pop().position.end=lo(e.end)},codeFenced:s((function(){const e=this.resume();this.stack[this.stack.length-1].value=e.replace(/^(\r?\n|\r)|(\r?\n|\r)$/g,""),this.data.flowCodeInside=void 0})),codeFencedFence:function(){this.data.flowCodeInside||(this.buffer(),this.data.flowCodeInside=!0)},codeFencedFenceInfo:function(){const e=this.resume();this.stack[this.stack.length-1].lang=e},codeFencedFenceMeta:function(){const e=this.resume();this.stack[this.stack.length-1].meta=e},codeFlowValue:d,codeIndented:s((function(){const e=this.resume();this.stack[this.stack.length-1].value=e.replace(/(\r?\n|\r)$/g,"")})),codeText:s((function(){const e=this.resume();this.stack[this.stack.length-1].value=e})),codeTextData:d,data:d,definition:s(),definitionDestinationString:function(){const e=this.resume();this.stack[this.stack.length-1].url=e},definitionLabelString:function(e){const t=this.resume(),n=this.stack[this.stack.length-1];n.label=t,n.identifier=Zt(this.sliceSerialize(e)).toLowerCase()},definitionTitleString:function(){const e=this.resume();this.stack[this.stack.length-1].title=e},emphasis:s(),hardBreakEscape:s(p),hardBreakTrailing:s(p),htmlFlow:s((function(){const e=this.resume();this.stack[this.stack.length-1].value=e})),htmlFlowData:d,htmlText:s((function(){const e=this.resume();this.stack[this.stack.length-1].value=e})),htmlTextData:d,image:s((function(){const e=this.stack[this.stack.length-1];if(this.data.inReference){const t=this.data.referenceType||"shortcut";e.type+="Reference",e.referenceType=t,delete e.url,delete e.title}else delete e.identifier,delete e.label;this.data.referenceType=void 0})),label:function(){const e=this.stack[this.stack.length-1],t=this.resume(),n=this.stack[this.stack.length-1];if(this.data.inReference=!0,"link"===n.type){const t=e.children;n.children=t}else n.alt=t},labelText:function(e){const t=this.sliceSerialize(e),n=this.stack[this.stack.length-2];n.label=ro(t),n.identifier=Zt(t).toLowerCase()},lineEnding:function(e){const n=this.stack[this.stack.length-1];if(this.data.atHardBreak)return n.children[n.children.length-1].position.end=lo(e.end),void(this.data.atHardBreak=void 0);!this.data.setextHeadingSlurpLineEnding&&t.canContainEols.includes(n.type)&&(u.call(this,e),d.call(this,e))},link:s((function(){const e=this.stack[this.stack.length-1];if(this.data.inReference){const t=this.data.referenceType||"shortcut";e.type+="Reference",e.referenceType=t,delete e.url,delete e.title}else delete e.identifier,delete e.label;this.data.referenceType=void 0})),listItem:s(),listOrdered:s(),listUnordered:s(),paragraph:s(),referenceString:function(e){const t=this.resume(),n=this.stack[this.stack.length-1];n.label=t,n.identifier=Zt(this.sliceSerialize(e)).toLowerCase(),this.data.referenceType="full"},resourceDestinationString:function(){const e=this.resume();this.stack[this.stack.length-1].url=e},resourceTitleString:function(){const e=this.resume();this.stack[this.stack.length-1].title=e},resource:function(){this.data.inReference=void 0},setextHeading:s((function(){this.data.setextHeadingSlurpLineEnding=void 0})),setextHeadingLineSequence:function(e){this.stack[this.stack.length-1].depth=61===this.sliceSerialize(e).codePointAt(0)?1:2},setextHeadingText:function(){this.data.setextHeadingSlurpLineEnding=!0},strong:s(),thematicBreak:s()}};co(t,(e||{}).mdastExtensions||[]);const n={};return function(e){let a={type:"root",children:[]};const s={stack:[a],tokenStack:[],config:t,enter:i,exit:l,buffer:o,resume:c,data:n},u=[];let d=-1;for(;++d<e.length;)"listOrdered"!==e[d][1].type&&"listUnordered"!==e[d][1].type||("enter"===e[d][0]?u.push(d):d=r(e,u.pop(),d));for(d=-1;++d<e.length;){const n=t[e[d][0]];io.call(n,e[d][1].type)&&n[e[d][1].type].call(Object.assign({sliceSerialize:e[d][2].sliceSerialize},s),e[d][1])}if(s.tokenStack.length>0){const e=s.tokenStack[s.tokenStack.length-1];(e[1]||po).call(s,void 0,e[0])}for(a.position={start:lo(e.length>0?e[0][1].start:{line:1,column:1,offset:0}),end:lo(e.length>0?e[e.length-2][1].end:{line:1,column:1,offset:0})},d=-1;++d<t.transforms.length;)a=t.transforms[d](a)||a;return a};function r(e,t,n){let r,a,o,i,s=t-1,l=-1,c=!1;for(;++s<=n;){const t=e[s];switch(t[1].type){case"listUnordered":case"listOrdered":case"blockQuote":"enter"===t[0]?l++:l--,i=void 0;break;case"lineEndingBlank":"enter"===t[0]&&(!r||i||l||o||(o=s),i=void 0);break;case"linePrefix":case"listItemValue":case"listItemMarker":case"listItemPrefix":case"listItemPrefixWhitespace":break;default:i=void 0}if(!l&&"enter"===t[0]&&"listItemPrefix"===t[1].type||-1===l&&"exit"===t[0]&&("listUnordered"===t[1].type||"listOrdered"===t[1].type)){if(r){let i=s;for(a=void 0;i--;){const t=e[i];if("lineEnding"===t[1].type||"lineEndingBlank"===t[1].type){if("exit"===t[0])continue;a&&(e[a][1].type="lineEndingBlank",c=!0),t[1].type="lineEnding",a=i}else if("linePrefix"!==t[1].type&&"blockQuotePrefix"!==t[1].type&&"blockQuotePrefixWhitespace"!==t[1].type&&"blockQuoteMarker"!==t[1].type&&"listItemIndent"!==t[1].type)break}o&&(!a||o<a)&&(r._spread=!0),r.end=Object.assign({},a?e[a][1].start:t[1].end),e.splice(a||s,0,["exit",r,t[2]]),s++,n++}if("listItemPrefix"===t[1].type){const a={type:"listItem",_spread:!1,start:Object.assign({},t[1].start),end:void 0};r=a,e.splice(s,0,["enter",a,t[2]]),s++,n++,o=void 0,i=!0}}}return e[t][1]._spread=c,n}function a(e,t){return function(n){i.call(this,e(n),n),t&&t.call(this,n)}}function o(){this.stack.push({type:"fragment",children:[]})}function i(e,t,n){this.stack[this.stack.length-1].children.push(e),this.stack.push(e),this.tokenStack.push([t,n||void 0]),e.position={start:lo(t.start),end:void 0}}function s(e){return function(t){e&&e.call(this,t),l.call(this,t)}}function l(e,t){const n=this.stack.pop(),r=this.tokenStack.pop();if(!r)throw new Error("Cannot close `"+e.type+"` ("+(0,oo.L)({start:e.start,end:e.end})+"): it’s not open");r[0].type!==e.type&&(t?t.call(this,e,r[0]):(r[1]||po).call(this,e,r[0])),n.position.end=lo(e.end)}function c(){return Cn(this.stack.pop())}function u(e){const t=this.stack[this.stack.length-1].children;let n=t[t.length-1];n&&"text"===n.type||(n={type:"text",value:""},n.position={start:lo(e.start),end:void 0},t.push(n)),this.stack.push(n)}function d(e){const t=this.stack.pop();t.value+=this.sliceSerialize(e),t.position.end=lo(e.end)}function p(){this.data.atHardBreak=!0}function h(e){this.data.characterReferenceType=e.type}function f(){return{type:"code",lang:null,meta:null,value:""}}function m(){return{type:"heading",depth:0,children:[]}}function g(){return{type:"break"}}function v(){return{type:"html",value:""}}function b(){return{type:"link",title:null,url:"",children:[]}}function y(e){return{type:"list",ordered:"listOrdered"===e.type,start:null,spread:e._spread,children:[]}}}(n)(function(e){for(;!Xr(e););return e}(function(e){const t={constructs:rr([a,...(e||{}).extensions||[]]),content:n(Vr),defined:[],document:n(Hr),flow:n(Yr),lazy:{},string:n(ea),text:n(ta)};return t;function n(e){return function(n){return Ja(t,e,n)}}}(n).document().write(function(){let e,t=1,n="",r=!0;return function(a,o,i){const s=[];let l,c,u,d,p;for(a=n+("string"==typeof a?a.toString():new TextDecoder(o||void 0).decode(a)),u=0,n="",r&&(65279===a.charCodeAt(0)&&u++,r=void 0);u<a.length;){if(eo.lastIndex=u,l=eo.exec(a),d=l&&void 0!==l.index?l.index:a.length,p=a.charCodeAt(d),!l){n=a.slice(u);break}if(10===p&&u===d&&e)s.push(-3),e=void 0;else switch(e&&(s.push(-5),e=void 0),u<d&&(s.push(a.slice(u,d)),t+=d-u),p){case 0:s.push(65533),t++;break;case 9:for(c=4*Math.ceil(t/4),s.push(-2);t++<c;)s.push(-1);break;case 10:s.push(-4),t=1;break;default:e=!0,t=1}u=d+1}return i&&(e&&s.push(-5),n&&s.push(n),s.push(null)),s}}()(e,t,!0))))}function lo(e){return{line:e.line,column:e.column,offset:e.offset}}function co(e,t){let n=-1;for(;++n<t.length;){const r=t[n];Array.isArray(r)?co(e,r):uo(e,r)}}function uo(e,t){let n;for(n in t)if(io.call(t,n))switch(n){case"canContainEols":{const r=t[n];r&&e[n].push(...r);break}case"transforms":{const r=t[n];r&&e[n].push(...r);break}case"enter":case"exit":{const r=t[n];r&&Object.assign(e[n],r);break}}}function po(e,t){throw e?new Error("Cannot close `"+e.type+"` ("+(0,oo.L)({start:e.start,end:e.end})+"): a different token (`"+t.type+"`, "+(0,oo.L)({start:t.start,end:t.end})+") is open"):new Error("Cannot close document, a token (`"+t.type+"`, "+(0,oo.L)({start:t.start,end:t.end})+") is still open")}function ho(e){const t=this;t.parser=function(n){return so(n,{...t.data("settings"),...e,extensions:t.data("micromarkExtensions")||[],mdastExtensions:t.data("fromMarkdownExtensions")||[]})}}function fo(e){const t=[];let n=-1,r=0,a=0;for(;++n<e.length;){const o=e.charCodeAt(n);let i="";if(37===o&&Et(e.charCodeAt(n+1))&&Et(e.charCodeAt(n+2)))a=2;else if(o<128)/[!#$&-;=?-Z_a-z~]/.test(String.fromCharCode(o))||(i=String.fromCharCode(o));else if(o>55295&&o<57344){const t=e.charCodeAt(n+1);o<56320&&t>56319&&t<57344?(i=String.fromCharCode(o,t),a=1):i="�"}else i=String.fromCharCode(o);i&&(t.push(e.slice(r,n),encodeURIComponent(i)),r=n+a+1,i=""),a&&(n+=a,a=0)}return t.join("")+e.slice(r)}function mo(e,t){const n=[{type:"text",value:"↩"}];return t>1&&n.push({type:"element",tagName:"sup",properties:{},children:[{type:"text",value:String(t)}]}),n}function go(e,t){return"Back to reference "+(e+1)+(t>1?"-"+t:"")}function vo(e,t){const n=t.referenceType;let r="]";if("collapsed"===n?r+="[]":"full"===n&&(r+="["+(t.label||t.identifier)+"]"),"imageReference"===t.type)return[{type:"text",value:"!["+t.alt+r}];const a=e.all(t),o=a[0];o&&"text"===o.type?o.value="["+o.value:a.unshift({type:"text",value:"["});const i=a[a.length-1];return i&&"text"===i.type?i.value+=r:a.push({type:"text",value:r}),a}function bo(e){const t=e.spread;return null==t?e.children.length>1:t}function yo(e){const t=String(e),n=/\r?\n|\r/g;let r=n.exec(t),a=0;const o=[];for(;r;)o.push(wo(t.slice(a,r.index),a>0,!0),r[0]),a=r.index+r[0].length,r=n.exec(t);return o.push(wo(t.slice(a),a>0,!1)),o.join("")}function wo(e,t,n){let r=0,a=e.length;if(t){let t=e.codePointAt(r);for(;9===t||32===t;)r++,t=e.codePointAt(r)}if(n){let t=e.codePointAt(a-1);for(;9===t||32===t;)a--,t=e.codePointAt(a-1)}return a>r?e.slice(r,a):""}const Oo={blockquote:function(e,t){const n={type:"element",tagName:"blockquote",properties:{},children:e.wrap(e.all(t),!0)};return e.patch(t,n),e.applyData(t,n)},break:function(e,t){const n={type:"element",tagName:"br",properties:{},children:[]};return e.patch(t,n),[e.applyData(t,n),{type:"text",value:"\n"}]},code:function(e,t){const n=t.value?t.value+"\n":"",r={};t.lang&&(r.className=["language-"+t.lang]);let a={type:"element",tagName:"code",properties:r,children:[{type:"text",value:n}]};return t.meta&&(a.data={meta:t.meta}),e.patch(t,a),a=e.applyData(t,a),a={type:"element",tagName:"pre",properties:{},children:[a]},e.patch(t,a),a},delete:function(e,t){const n={type:"element",tagName:"del",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)},emphasis:function(e,t){const n={type:"element",tagName:"em",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)},footnoteReference:function(e,t){const n="string"==typeof e.options.clobberPrefix?e.options.clobberPrefix:"user-content-",r=String(t.identifier).toUpperCase(),a=fo(r.toLowerCase()),o=e.footnoteOrder.indexOf(r);let i,s=e.footnoteCounts.get(r);void 0===s?(s=0,e.footnoteOrder.push(r),i=e.footnoteOrder.length):i=o+1,s+=1,e.footnoteCounts.set(r,s);const l={type:"element",tagName:"a",properties:{href:"#"+n+"fn-"+a,id:n+"fnref-"+a+(s>1?"-"+s:""),dataFootnoteRef:!0,ariaDescribedBy:["footnote-label"]},children:[{type:"text",value:String(i)}]};e.patch(t,l);const c={type:"element",tagName:"sup",properties:{},children:[l]};return e.patch(t,c),e.applyData(t,c)},heading:function(e,t){const n={type:"element",tagName:"h"+t.depth,properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)},html:function(e,t){if(e.options.allowDangerousHtml){const n={type:"raw",value:t.value};return e.patch(t,n),e.applyData(t,n)}},imageReference:function(e,t){const n=String(t.identifier).toUpperCase(),r=e.definitionById.get(n);if(!r)return vo(e,t);const a={src:fo(r.url||""),alt:t.alt};null!==r.title&&void 0!==r.title&&(a.title=r.title);const o={type:"element",tagName:"img",properties:a,children:[]};return e.patch(t,o),e.applyData(t,o)},image:function(e,t){const n={src:fo(t.url)};null!==t.alt&&void 0!==t.alt&&(n.alt=t.alt),null!==t.title&&void 0!==t.title&&(n.title=t.title);const r={type:"element",tagName:"img",properties:n,children:[]};return e.patch(t,r),e.applyData(t,r)},inlineCode:function(e,t){const n={type:"text",value:t.value.replace(/\r?\n|\r/g," ")};e.patch(t,n);const r={type:"element",tagName:"code",properties:{},children:[n]};return e.patch(t,r),e.applyData(t,r)},linkReference:function(e,t){const n=String(t.identifier).toUpperCase(),r=e.definitionById.get(n);if(!r)return vo(e,t);const a={href:fo(r.url||"")};null!==r.title&&void 0!==r.title&&(a.title=r.title);const o={type:"element",tagName:"a",properties:a,children:e.all(t)};return e.patch(t,o),e.applyData(t,o)},link:function(e,t){const n={href:fo(t.url)};null!==t.title&&void 0!==t.title&&(n.title=t.title);const r={type:"element",tagName:"a",properties:n,children:e.all(t)};return e.patch(t,r),e.applyData(t,r)},listItem:function(e,t,n){const r=e.all(t),a=n?function(e){let t=!1;if("list"===e.type){t=e.spread||!1;const n=e.children;let r=-1;for(;!t&&++r<n.length;)t=bo(n[r])}return t}(n):bo(t),o={},i=[];if("boolean"==typeof t.checked){const e=r[0];let n;e&&"element"===e.type&&"p"===e.tagName?n=e:(n={type:"element",tagName:"p",properties:{},children:[]},r.unshift(n)),n.children.length>0&&n.children.unshift({type:"text",value:" "}),n.children.unshift({type:"element",tagName:"input",properties:{type:"checkbox",checked:t.checked,disabled:!0},children:[]}),o.className=["task-list-item"]}let s=-1;for(;++s<r.length;){const e=r[s];(a||0!==s||"element"!==e.type||"p"!==e.tagName)&&i.push({type:"text",value:"\n"}),"element"!==e.type||"p"!==e.tagName||a?i.push(e):i.push(...e.children)}const l=r[r.length-1];l&&(a||"element"!==l.type||"p"!==l.tagName)&&i.push({type:"text",value:"\n"});const c={type:"element",tagName:"li",properties:o,children:i};return e.patch(t,c),e.applyData(t,c)},list:function(e,t){const n={},r=e.all(t);let a=-1;for("number"==typeof t.start&&1!==t.start&&(n.start=t.start);++a<r.length;){const e=r[a];if("element"===e.type&&"li"===e.tagName&&e.properties&&Array.isArray(e.properties.className)&&e.properties.className.includes("task-list-item")){n.className=["contains-task-list"];break}}const o={type:"element",tagName:t.ordered?"ol":"ul",properties:n,children:e.wrap(r,!0)};return e.patch(t,o),e.applyData(t,o)},paragraph:function(e,t){const n={type:"element",tagName:"p",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)},root:function(e,t){const n={type:"root",children:e.wrap(e.all(t))};return e.patch(t,n),e.applyData(t,n)},strong:function(e,t){const n={type:"element",tagName:"strong",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)},table:function(e,t){const n=e.all(t),r=n.shift(),a=[];if(r){const n={type:"element",tagName:"thead",properties:{},children:e.wrap([r],!0)};e.patch(t.children[0],n),a.push(n)}if(n.length>0){const r={type:"element",tagName:"tbody",properties:{},children:e.wrap(n,!0)},o=qe(t.children[1]),i=Xe(t.children[t.children.length-1]);o&&i&&(r.position={start:o,end:i}),a.push(r)}const o={type:"element",tagName:"table",properties:{},children:e.wrap(a,!0)};return e.patch(t,o),e.applyData(t,o)},tableCell:function(e,t){const n={type:"element",tagName:"td",properties:{},children:e.all(t)};return e.patch(t,n),e.applyData(t,n)},tableRow:function(e,t,n){const r=n?n.children:void 0,a=0===(r?r.indexOf(t):1)?"th":"td",o=n&&"table"===n.type?n.align:void 0,i=o?o.length:t.children.length;let s=-1;const l=[];for(;++s<i;){const n=t.children[s],r={},i=o?o[s]:void 0;i&&(r.align=i);let c={type:"element",tagName:a,properties:r,children:[]};n&&(c.children=e.all(n),e.patch(n,c),c=e.applyData(n,c)),l.push(c)}const c={type:"element",tagName:"tr",properties:{},children:e.wrap(l,!0)};return e.patch(t,c),e.applyData(t,c)},text:function(e,t){const n={type:"text",value:yo(String(t.value))};return e.patch(t,n),e.applyData(t,n)},thematicBreak:function(e,t){const n={type:"element",tagName:"hr",properties:{},children:[]};return e.patch(t,n),e.applyData(t,n)},toml:ko,yaml:ko,definition:ko,footnoteDefinition:ko};function ko(){}const xo={}.hasOwnProperty,Eo={};function So(e,t){e.position&&(t.position=He(e))}function _o(e,t){let n=t;if(e&&e.data){const t=e.data.hName,r=e.data.hChildren,a=e.data.hProperties;"string"==typeof t&&("element"===n.type?n.tagName=t:n={type:"element",tagName:t,properties:{},children:"children"in n?n.children:[n]}),"element"===n.type&&a&&Object.assign(n.properties,h(a)),"children"in n&&n.children&&null!=r&&(n.children=r)}return n}function Co(e,t){const n=t.data||{},r=!("value"in t)||xo.call(n,"hProperties")||xo.call(n,"hChildren")?{type:"element",tagName:"div",properties:{},children:e.all(t)}:{type:"text",value:t.value};return e.patch(t,r),e.applyData(t,r)}function To(e,t){const n=[];let r=-1;for(t&&n.push({type:"text",value:"\n"});++r<e.length;)r&&n.push({type:"text",value:"\n"}),n.push(e[r]);return t&&e.length>0&&n.push({type:"text",value:"\n"}),n}function Ao(e){let t=0,n=e.charCodeAt(t);for(;9===n||32===n;)t++,n=e.charCodeAt(t);return e.slice(t)}function Ro(e,t){const n=function(e,t){const n=t||Eo,r=new Map,a=new Map,o=new Map,i={...Oo,...n.handlers},s={all:function(e){const t=[];if("children"in e){const n=e.children;let r=-1;for(;++r<n.length;){const a=s.one(n[r],e);if(a){if(r&&"break"===n[r-1].type&&(Array.isArray(a)||"text"!==a.type||(a.value=Ao(a.value)),!Array.isArray(a)&&"element"===a.type)){const e=a.children[0];e&&"text"===e.type&&(e.value=Ao(e.value))}Array.isArray(a)?t.push(...a):t.push(a)}}}return t},applyData:_o,definitionById:r,footnoteById:a,footnoteCounts:o,footnoteOrder:[],handlers:i,one:function(e,t){const n=e.type,r=s.handlers[n];if(xo.call(s.handlers,n)&&r)return r(s,e,t);if(s.options.passThrough&&s.options.passThrough.includes(n)){if("children"in e){const{children:t,...n}=e,r=h(n);return r.children=s.all(e),r}return h(e)}return(s.options.unknownHandler||Co)(s,e,t)},options:n,patch:So,wrap:To};return(0,b.YR)(e,(function(e){if("definition"===e.type||"footnoteDefinition"===e.type){const t="definition"===e.type?r:a,n=String(e.identifier).toUpperCase();t.has(n)||t.set(n,e)}})),s}(e,t),r=n.one(e,void 0),a=function(e){const t="string"==typeof e.options.clobberPrefix?e.options.clobberPrefix:"user-content-",n=e.options.footnoteBackContent||mo,r=e.options.footnoteBackLabel||go,a=e.options.footnoteLabel||"Footnotes",o=e.options.footnoteLabelTagName||"h2",i=e.options.footnoteLabelProperties||{className:["sr-only"]},s=[];let l=-1;for(;++l<e.footnoteOrder.length;){const a=e.footnoteById.get(e.footnoteOrder[l]);if(!a)continue;const o=e.all(a),i=String(a.identifier).toUpperCase(),c=fo(i.toLowerCase());let u=0;const d=[],p=e.footnoteCounts.get(i);for(;void 0!==p&&++u<=p;){d.length>0&&d.push({type:"text",value:" "});let e="string"==typeof n?n:n(l,u);"string"==typeof e&&(e={type:"text",value:e}),d.push({type:"element",tagName:"a",properties:{href:"#"+t+"fnref-"+c+(u>1?"-"+u:""),dataFootnoteBackref:"",ariaLabel:"string"==typeof r?r:r(l,u),className:["data-footnote-backref"]},children:Array.isArray(e)?e:[e]})}const h=o[o.length-1];if(h&&"element"===h.type&&"p"===h.tagName){const e=h.children[h.children.length-1];e&&"text"===e.type?e.value+=" ":h.children.push({type:"text",value:" "}),h.children.push(...d)}else o.push(...d);const f={type:"element",tagName:"li",properties:{id:t+"fn-"+c},children:e.wrap(o,!0)};e.patch(a,f),s.push(f)}if(0!==s.length)return{type:"element",tagName:"section",properties:{dataFootnotes:!0,className:["footnotes"]},children:[{type:"element",tagName:o,properties:{...h(i),id:"footnote-label"},children:[{type:"text",value:a}]},{type:"text",value:"\n"},{type:"element",tagName:"ol",properties:{},children:e.wrap(s,!0)},{type:"text",value:"\n"}]}}(n),o=Array.isArray(r)?{type:"root",children:r}:r||{type:"root",children:[]};return a&&((0,te.ok)("children"in o),o.children.push({type:"text",value:"\n"},a)),o}function Po(e,t){return e&&"run"in e?async function(n,r){const a=Ro(n,{file:r,...t});await e.run(a,r)}:function(n,r){return Ro(n,{file:r,...e||t})}}const Do={}.hasOwnProperty;function Mo(e,t){let n,r=-1;if(t.extensions)for(;++r<t.extensions.length;)Mo(e,t.extensions[r]);for(n in t)if(Do.call(t,n))switch(n){case"extensions":break;case"unsafe":case"join":a=e[n],(o=t[n])&&a.push(...o);break;case"handlers":$o(e[n],t[n]);break;default:e.options[n]=t[n]}var a,o;return e}function $o(e,t){t&&Object.assign(e,t)}const No=[function(e,t,n,r){if("code"===t.type&&yn(t,r)&&("list"===e.type||e.type===t.type&&yn(e,r)))return!1;if("spread"in n&&"boolean"==typeof n.spread){if("paragraph"===e.type&&(e.type===t.type||"definition"===t.type||"heading"===t.type&&Rn(t,r)))return;return n.spread?1:0}}],Io=["autolink","destinationLiteral","destinationRaw","reference","titleQuote","titleApostrophe"],Lo=[{character:"\t",after:"[\\r\\n]",inConstruct:"phrasing"},{character:"\t",before:"[\\r\\n]",inConstruct:"phrasing"},{character:"\t",inConstruct:["codeFencedLangGraveAccent","codeFencedLangTilde"]},{character:"\r",inConstruct:["codeFencedLangGraveAccent","codeFencedLangTilde","codeFencedMetaGraveAccent","codeFencedMetaTilde","destinationLiteral","headingAtx"]},{character:"\n",inConstruct:["codeFencedLangGraveAccent","codeFencedLangTilde","codeFencedMetaGraveAccent","codeFencedMetaTilde","destinationLiteral","headingAtx"]},{character:" ",after:"[\\r\\n]",inConstruct:"phrasing"},{character:" ",before:"[\\r\\n]",inConstruct:"phrasing"},{character:" ",inConstruct:["codeFencedLangGraveAccent","codeFencedLangTilde"]},{character:"!",after:"\\[",inConstruct:"phrasing",notInConstruct:Io},{character:'"',inConstruct:"titleQuote"},{atBreak:!0,character:"#"},{character:"#",inConstruct:"headingAtx",after:"(?:[\r\n]|$)"},{character:"&",after:"[#A-Za-z]",inConstruct:"phrasing"},{character:"'",inConstruct:"titleApostrophe"},{character:"(",inConstruct:"destinationRaw"},{before:"\\]",character:"(",inConstruct:"phrasing",notInConstruct:Io},{atBreak:!0,before:"\\d+",character:")"},{character:")",inConstruct:"destinationRaw"},{atBreak:!0,character:"*",after:"(?:[ \t\r\n*])"},{character:"*",inConstruct:"phrasing",notInConstruct:Io},{atBreak:!0,character:"+",after:"(?:[ \t\r\n])"},{atBreak:!0,character:"-",after:"(?:[ \t\r\n-])"},{atBreak:!0,before:"\\d+",character:".",after:"(?:[ \t\r\n]|$)"},{atBreak:!0,character:"<",after:"[!/?A-Za-z]"},{character:"<",after:"[!/?A-Za-z]",inConstruct:"phrasing",notInConstruct:Io},{character:"<",inConstruct:"destinationLiteral"},{atBreak:!0,character:"="},{atBreak:!0,character:">"},{character:">",inConstruct:"destinationLiteral"},{atBreak:!0,character:"["},{character:"[",inConstruct:"phrasing",notInConstruct:Io},{character:"[",inConstruct:["label","reference"]},{character:"\\",after:"[\\r\\n]",inConstruct:"phrasing"},{character:"]",inConstruct:["label","reference"]},{atBreak:!0,character:"_"},{character:"_",inConstruct:"phrasing",notInConstruct:Io},{atBreak:!0,character:"`"},{character:"`",inConstruct:["codeFencedLangGraveAccent","codeFencedMetaGraveAccent"]},{character:"`",inConstruct:"phrasing",notInConstruct:Io},{atBreak:!0,character:"~"}];function Bo(e){return e.label||!e.identifier?e.label||"":ro(e.identifier)}function Fo(e){if(!e._compiled){const t=(e.atBreak?"[\\r\\n][\\t ]*":"")+(e.before?"(?:"+e.before+")":"");e._compiled=new RegExp((t?"("+t+")":"")+(/[|\\{}()[\]^$+*?.-]/.test(e.character)?"\\":"")+e.character+(e.after?"(?:"+e.after+")":""),"g")}return e._compiled}function Qo(e,t,n,r){let a=r.join.length;for(;a--;){const o=r.join[a](e,t,n,r);if(!0===o||1===o)break;if("number"==typeof o)return"\n".repeat(1+o);if(!1===o)return"\n\n\x3c!----\x3e\n\n"}return"\n\n"}const Uo=/\r?\n|\r/g;function zo(e,t){const n=[];let r,a=0,o=0;for(;r=Uo.exec(e);)i(e.slice(a,r.index)),n.push(r[0]),a=r.index+r[0].length,o++;return i(e.slice(a)),n.join("");function i(e){n.push(t(e,o,!e))}}function jo(e,t){return e-t}function Wo(e,t){const n=/\\(?=[!-/:-@[-`{-~])/g,r=[],a=[],o=e+t;let i,s=-1,l=0;for(;i=n.exec(o);)r.push(i.index);for(;++s<r.length;)l!==r[s]&&a.push(e.slice(l,r[s])),a.push("\\"),l=r[s];return a.push(e.slice(l)),a.join("")}function Xo(e){const t=e||{},n=t.now||{};let r=t.lineShift||0,a=n.line||1,o=n.column||1;return{move:function(e){const t=e||"",n=t.split(/\r?\n|\r/g),i=n[n.length-1];return a+=n.length-1,o=1===n.length?o+i.length:1+i.length+r,t},current:function(){return{now:{line:a,column:o},lineShift:r}},shift:function(e){r+=e}}}function qo(e,t){const n=t||{},r={associationId:Bo,containerPhrasing:Go,containerFlow:Ko,createTracker:Xo,compilePattern:Fo,enter:function(e){return r.stack.push(e),function(){r.stack.pop()}},handlers:{...zn},handle:void 0,indentLines:zo,indexStack:[],join:[...No],options:{},safe:Yo,stack:[],unsafe:[...Lo]};Mo(r,n),r.options.tightDefinitions&&r.join.push(Zo),r.handle=(0,Ie.A)("type",{invalid:Vo,unknown:Ho,handlers:r.handlers});let a=r.handle(e,void 0,r,{before:"\n",after:"\n",now:{line:1,column:1},lineShift:0});return a&&10!==a.charCodeAt(a.length-1)&&13!==a.charCodeAt(a.length-1)&&(a+="\n"),a}function Vo(e){throw new Error("Cannot handle value `"+e+"`, expected node")}function Ho(e){throw new Error("Cannot handle unknown node `"+e.type+"`")}function Zo(e,t){if("definition"===e.type&&e.type===t.type)return 0}function Go(e,t){return function(e,t,n){const r=t.indexStack,a=e.children||[],o=[];let i,s=-1,l=n.before;r.push(-1);let c=t.createTracker(n);for(;++s<a.length;){const u=a[s];let d;if(r[r.length-1]=s,s+1<a.length){let n=t.handle.handlers[a[s+1].type];n&&n.peek&&(n=n.peek),d=n?n(a[s+1],e,t,{before:"",after:"",...c.current()}).charAt(0):""}else d=n.after;o.length>0&&("\r"===l||"\n"===l)&&"html"===u.type&&(o[o.length-1]=o[o.length-1].replace(/(\r?\n|\r)$/," "),l=" ",c=t.createTracker(n),c.move(o.join("")));let p=t.handle(u,e,t,{...c.current(),after:d,before:l});i&&i===p.slice(0,1)&&(p=kn(i.charCodeAt(0))+p.slice(1));const h=t.attentionEncodeSurroundingInfo;t.attentionEncodeSurroundingInfo=void 0,i=void 0,h&&(o.length>0&&h.before&&l===o[o.length-1].slice(-1)&&(o[o.length-1]=o[o.length-1].slice(0,-1)+kn(l.charCodeAt(0))),h.after&&(i=d)),c.move(p),o.push(p),l=p.slice(-1)}return r.pop(),o.join("")}(e,this,t)}function Ko(e,t){return function(e,t,n){const r=t.indexStack,a=e.children||[],o=t.createTracker(n),i=[];let s=-1;for(r.push(-1);++s<a.length;){const n=a[s];r[r.length-1]=s,i.push(o.move(t.handle(n,e,t,{before:"\n",after:"\n",...o.current()}))),"list"!==n.type&&(t.bulletLastUsed=void 0),s<a.length-1&&i.push(o.move(Qo(n,a[s+1],e,t)))}return r.pop(),i.join("")}(e,this,t)}function Yo(e,t){return function(e,t,n){const r=(n.before||"")+(t||"")+(n.after||""),a=[],o=[],i={};let s=-1;for(;++s<e.unsafe.length;){const t=e.unsafe[s];if(!gn(e.stack,t))continue;const n=e.compilePattern(t);let o;for(;o=n.exec(r);){const e="before"in t||Boolean(t.atBreak),n="after"in t,r=o.index+(e?o[1].length:0);a.includes(r)?(i[r].before&&!e&&(i[r].before=!1),i[r].after&&!n&&(i[r].after=!1)):(a.push(r),i[r]={before:e,after:n})}}a.sort(jo);let l=n.before?n.before.length:0;const c=r.length-(n.after?n.after.length:0);for(s=-1;++s<a.length;){const e=a[s];e<l||e>=c||e+1<c&&a[s+1]===e+1&&i[e].after&&!i[e+1].before&&!i[e+1].after||a[s-1]===e-1&&i[e].before&&!i[e-1].before&&!i[e-1].after||(l!==e&&o.push(Wo(r.slice(l,e),"\\")),l=e,!/[!-/:-@[-`{-~]/.test(r.charAt(e))||n.encode&&n.encode.includes(r.charAt(e))?(o.push(kn(r.charCodeAt(e))),l++):o.push("\\"))}return o.push(Wo(r.slice(l,c),n.after)),o.join("")}(this,e,t)}function Jo(e){const t=this;t.compiler=function(n){return qo(n,{...t.data("settings"),...e,extensions:t.data("toMarkdownExtensions")||[]})}}var ei=n(9635),ti=n(9464),ni=n(7526);const ri=(e,...t)=>t=>{e?.transform&&e?.type&&(0,b.YR)(t,e?.type,(t=>(e?.transform&&e?.transform(t),V._Z)))};function ai(e,t){const n=t?.removeTags??[],r=[...pt.tagNames??[],...t?.allowTags??[]].filter((e=>!n.includes(e)));return(0,ei.l)().use(ho).use(zr).use(ri,{transform:t?.transform,type:t?.transformType}).use(Po,{allowDangerousHtml:!0}).use(ut).use(wt,{...pt,clobberPrefix:"",tagNames:r,attributes:{...pt.attributes,abbr:["title"]}}).use(ti.u,{languages:ni.n,detect:!0}).use(k,{target:"_blank"}).use(Y).use(Ot.A).processSync(e).toString()}function oi(e){return(0,ei.l)().use(ho).use(zr).parse(e)}function ii(e,t=1){const n=oi(e),r=[];return(0,b.YR)(n,"heading",(e=>{const n=si(e);n&&r.push({depth:e.depth??t,value:n.value})})),r}function si(e){if("text"===e.type)return e;if("children"in e&&e.children)for(const t of e.children){const e=si(t);if(e)return e}return null}function li(e){const t=oi(e),n=[];let r=[];return t.children?.forEach((e=>{"heading"===e.type?(r.length&&n.push(r),n.push([e]),r=[]):r.push(e)})),r.length&&n.push(r),n.map((e=>function(e){return(0,ei.l)().use(Jo).use(zr).stringify({type:"root",children:e}).trim()}(e)))}},9271:(e,t,n)=>{"use strict";n.d(t,{D_:()=>w,oc:()=>s,un:()=>p,Gv:()=>u,lQ:()=>d,d6:()=>l,lW:()=>O,BA:()=>i,Uo:()=>o,Q0:()=>k,TO:()=>x,Th:()=>E});var r=n(953),a=n(641);function o(e){return!!(0,r.o5)()&&((0,r.jr)(e),!0)}function i(e){return"function"==typeof e?e():(0,r.R1)(e)}const s="undefined"!=typeof window&&"undefined"!=typeof document,l=("undefined"!=typeof WorkerGlobalScope&&(globalThis,WorkerGlobalScope),e=>null!=e),c=Object.prototype.toString,u=e=>"[object Object]"===c.call(e),d=()=>{},p=h();function h(){var e,t;return s&&(null==(e=null==window?void 0:window.navigator)?void 0:e.userAgent)&&(/iP(?:ad|hone|od)/.test(window.navigator.userAgent)||(null==(t=null==window?void 0:window.navigator)?void 0:t.maxTouchPoints)>2&&/iPad|Macintosh/.test(null==window?void 0:window.navigator.userAgent))}function f(e,t){return function(...n){return new Promise(((r,a)=>{Promise.resolve(e((()=>t.apply(this,n)),{fn:t,thisArg:this,args:n})).then(r).catch(a)}))}}const m=e=>e();function g(e,t={}){let n,r,a=d;const o=e=>{clearTimeout(e),a(),a=d};return s=>{const l=i(e),c=i(t.maxWait);return n&&o(n),l<=0||void 0!==c&&c<=0?(r&&(o(r),r=null),Promise.resolve(s())):new Promise(((e,i)=>{a=t.rejectOnCancel?i:e,c&&!r&&(r=setTimeout((()=>{n&&o(n),r=null,e(s())}),c)),n=setTimeout((()=>{r&&o(r),r=null,e(s())}),l)}))}}function v(e){const t=Object.create(null);return n=>t[n]||(t[n]=e(n))}const b=/\B([A-Z])/g,y=(v((e=>e.replace(b,"-$1").toLowerCase())),/-(\w)/g);function w(e){return e}function O(...e){if(1!==e.length)return(0,r.lW)(...e);const t=e[0];return"function"==typeof t?(0,r.tB)((0,r.rY)((()=>({get:t,set:d})))):(0,r.KR)(t)}function k(e,t=200,n={}){return f(g(t,n),e)}function x(e,t,n={}){const{immediate:a=!0}=n,l=(0,r.KR)(!1);let c=null;function u(){c&&(clearTimeout(c),c=null)}function d(){l.value=!1,u()}function p(...n){u(),l.value=!0,c=setTimeout((()=>{l.value=!1,c=null,e(...n)}),i(t))}return a&&(l.value=!0,s&&p()),o(d),{isPending:(0,r.tB)(l),start:p,stop:d}}function E(e,t,n={}){const{debounce:r=0,maxWait:o,...i}=n;return function(e,t,n={}){const{eventFilter:r=m,...o}=n;return(0,a.wB)(e,f(r,t),o)}(e,t,{...i,eventFilter:g(r,{maxWait:o})})}v((e=>e.replace(y,((e,t)=>t?t.toUpperCase():""))))},9347:(e,t,n)=>{"use strict";n.d(t,{A:()=>p});var r=n(641),a=n(33),o=n(953),i=n(1266);const s={key:0,class:"sr-only"},l=(0,r.pM)({__name:"ServerVariablesSelect",props:{enum:{},value:{},controls:{}},emits:["change"],setup(e,{emit:t}){const n=e,l=t,c=(0,r.EW)((()=>n.enum.map((e=>({id:e,label:e}))))),u=(0,r.EW)({get:()=>c.value.find((e=>e.id===n.value)),set:e=>l("change",(null==e?void 0:e.id)??"")});return(e,t)=>((0,r.uX)(),(0,r.Wv)((0,o.R1)(i.cy),{modelValue:u.value,"onUpdate:modelValue":t[0]||(t[0]=e=>u.value=e),options:c.value},{default:(0,r.k6)((()=>[(0,r.bF)((0,o.R1)(i.RO),{"aria-controls":e.controls,class:"h-8 p-0 py-1.5 w-full font-normal",variant:"ghost"},{default:(0,r.k6)((()=>[(0,r.Lk)("span",{class:(0,a.C4)({"text-c-1":e.value})},[e.value?((0,r.uX)(),(0,r.CE)("span",s," Selected: ")):(0,r.Q3)("",!0),(0,r.eW)(" "+(0,a.v_)(e.value||"Select value"),1)],2),(0,r.bF)((0,o.R1)(i.kv),{class:"ml-1",icon:"ChevronDown",size:"sm"})])),_:1},8,["aria-controls"])])),_:1},8,["modelValue","options"]))}});var c=n(3751);const u=(0,r.pM)({__name:"ServerVariablesTextbox",props:{value:{},controls:{}},emits:["change"],setup(e,{emit:t}){const n=e,a=t,o=(0,r.EW)({get:()=>n.value,set:e=>a("change",e)});return(e,t)=>(0,r.bo)(((0,r.uX)(),(0,r.CE)("input",(0,r.v6)({"onUpdate:modelValue":t[0]||(t[0]=e=>o.value=e)},e.controls?{...e.$attrs,"aria-controls":e.controls}:{},{autocomplete:"off",class:"px-1.5 text-c-1 border-transparent -outline-offset-1 w-full group-last/label:rounded-br-lg",placeholder:"value",spellcheck:"false",type:"text"}),null,16)),[[c.Jo,o.value]])}}),d={class:"flex items-center pl-3 py-1.5 after:content-[':'] mr-1.5 group-has-[input]/label:mr-0"},p=(0,r.pM)({__name:"ServerVariablesForm",props:{variables:{},values:{},controls:{}},emits:["update:variable"],setup(e,{emit:t}){const n=e,o=t;function i(e,t){o("update:variable",e,t)}const s=e=>{var t,r,a;return((null==(t=n.values)?void 0:t[e])??(null==(a=null==(r=n.variables)?void 0:r[e])?void 0:a.default)??"").toString()};return(e,t)=>e.variables&&Object.keys(e.variables??{}).length?((0,r.uX)(!0),(0,r.CE)(r.FK,{key:0},(0,r.pI)(Object.keys(e.variables),(t=>{var n,o,c,p,h;return(0,r.uX)(),(0,r.CE)("label",{key:t,class:"group/label flex w-full"},[(0,r.Lk)("span",d,(0,a.v_)(t),1),null!=(c=null==(o=null==(n=e.variables)?void 0:n[t])?void 0:o.enum)&&c.length?((0,r.uX)(),(0,r.Wv)(l,{key:0,controls:e.controls,enum:(null==(h=null==(p=e.variables[t])?void 0:p.enum)?void 0:h.map((e=>`${e}`)))??[],label:t,value:s(t),onChange:e=>i(t,e)},null,8,["controls","enum","label","value","onChange"])):((0,r.uX)(),(0,r.Wv)(u,{key:1,controls:e.controls,value:s(t),onChange:e=>i(t,e)},null,8,["controls","value","onChange"]))])})),128)):(0,r.Q3)("",!0)}})},9351:(e,t,n)=>{"use strict";n.d(t,{Wk:()=>De,Tp:()=>Ie,sC:()=>K,aE:()=>nt,lq:()=>at,rd:()=>rt});var r=n(641),a=n(953);function o(){return"undefined"!=typeof navigator&&"undefined"!=typeof window?window:"undefined"!=typeof globalThis?globalThis:{}}const i="function"==typeof Proxy,s="devtools-plugin:setup";let l,c;class u{constructor(e,t){this.target=null,this.targetQueue=[],this.onQueue=[],this.plugin=e,this.hook=t;const n={};if(e.settings)for(const t in e.settings){const r=e.settings[t];n[t]=r.defaultValue}const r=`__vue-devtools-plugin-settings__${e.id}`;let a=Object.assign({},n);try{const e=localStorage.getItem(r),t=JSON.parse(e);Object.assign(a,t)}catch(e){}this.fallbacks={getSettings:()=>a,setSettings(e){try{localStorage.setItem(r,JSON.stringify(e))}catch(e){}a=e},now:()=>{return void 0!==l||("undefined"!=typeof window&&window.performance?(l=!0,c=window.performance):"undefined"!=typeof globalThis&&(null===(e=globalThis.perf_hooks)||void 0===e?void 0:e.performance)?(l=!0,c=globalThis.perf_hooks.performance):l=!1),l?c.now():Date.now();var e}},t&&t.on("plugin:settings:set",((e,t)=>{e===this.plugin.id&&this.fallbacks.setSettings(t)})),this.proxiedOn=new Proxy({},{get:(e,t)=>this.target?this.target.on[t]:(...e)=>{this.onQueue.push({method:t,args:e})}}),this.proxiedTarget=new Proxy({},{get:(e,t)=>this.target?this.target[t]:"on"===t?this.proxiedOn:Object.keys(this.fallbacks).includes(t)?(...e)=>(this.targetQueue.push({method:t,args:e,resolve:()=>{}}),this.fallbacks[t](...e)):(...e)=>new Promise((n=>{this.targetQueue.push({method:t,args:e,resolve:n})}))})}async setRealTarget(e){this.target=e;for(const e of this.onQueue)this.target.on[e.method](...e.args);for(const e of this.targetQueue)e.resolve(await this.target[e.method](...e.args))}}const d="undefined"!=typeof document;function p(e){return"object"==typeof e||"displayName"in e||"props"in e||"__vccOpts"in e}const h=Object.assign;function f(e,t){const n={};for(const r in t){const a=t[r];n[r]=g(a)?a.map(e):e(a)}return n}const m=()=>{},g=Array.isArray,v=/#/g,b=/&/g,y=/\//g,w=/=/g,O=/\?/g,k=/\+/g,x=/%5B/g,E=/%5D/g,S=/%5E/g,_=/%60/g,C=/%7B/g,T=/%7C/g,A=/%7D/g,R=/%20/g;function P(e){return encodeURI(""+e).replace(T,"|").replace(x,"[").replace(E,"]")}function D(e){return P(e).replace(k,"%2B").replace(R,"+").replace(v,"%23").replace(b,"%26").replace(_,"`").replace(C,"{").replace(A,"}").replace(S,"^")}function M(e){return null==e?"":function(e){return P(e).replace(v,"%23").replace(O,"%3F")}(e).replace(y,"%2F")}function $(e){try{return decodeURIComponent(""+e)}catch(e){}return""+e}const N=/\/$/,I=e=>e.replace(N,"");function L(e,t,n="/"){let r,a={},o="",i="";const s=t.indexOf("#");let l=t.indexOf("?");return s<l&&s>=0&&(l=-1),l>-1&&(r=t.slice(0,l),o=t.slice(l+1,s>-1?s:t.length),a=e(o)),s>-1&&(r=r||t.slice(0,s),i=t.slice(s,t.length)),r=function(e,t){if(e.startsWith("/"))return e;if(!e)return t;const n=t.split("/"),r=e.split("/"),a=r[r.length-1];".."!==a&&"."!==a||r.push("");let o,i,s=n.length-1;for(o=0;o<r.length;o++)if(i=r[o],"."!==i){if(".."!==i)break;s>1&&s--}return n.slice(0,s).join("/")+"/"+r.slice(o).join("/")}(null!=r?r:t,n),{fullPath:r+(o&&"?")+o+i,path:r,query:a,hash:$(i)}}function B(e,t){return(e.aliasOf||e)===(t.aliasOf||t)}function F(e,t){if(Object.keys(e).length!==Object.keys(t).length)return!1;for(const n in e)if(!Q(e[n],t[n]))return!1;return!0}function Q(e,t){return g(e)?U(e,t):g(t)?U(t,e):e===t}function U(e,t){return g(t)?e.length===t.length&&e.every(((e,n)=>e===t[n])):1===e.length&&e[0]===t}const z={path:"/",name:void 0,params:{},query:{},hash:"",fullPath:"/",matched:[],meta:{},redirectedFrom:void 0};var j,W;!function(e){e.pop="pop",e.push="push"}(j||(j={})),function(e){e.back="back",e.forward="forward",e.unknown=""}(W||(W={}));const X="";const q=/^[^#]+#/;function V(e,t){return e.replace(q,"#")+t}const H=()=>({left:window.scrollX,top:window.scrollY});function Z(e,t){return(history.state?history.state.position-t:-1)+e}const G=new Map;function K(e=""){let t=[],n=[X],r=0;function a(e){r++,r!==n.length&&n.splice(r),n.push(e)}e=function(e){if(!e)if(d){const t=document.querySelector("base");e=(e=t&&t.getAttribute("href")||"/").replace(/^\w+:\/\/[^\/]+/,"")}else e="/";return"/"!==e[0]&&"#"!==e[0]&&(e="/"+e),I(e)}(e);const o={location:X,state:{},base:e,createHref:V.bind(null,e),replace(e){n.splice(r--,1),a(e)},push(e,t){a(e)},listen:e=>(t.push(e),()=>{const n=t.indexOf(e);n>-1&&t.splice(n,1)}),destroy(){t=[],n=[X],r=0},go(e,a=!0){const o=this.location,i=e<0?W.back:W.forward;r=Math.max(0,Math.min(r+e,n.length-1)),a&&function(e,n,{direction:r,delta:a}){const o={direction:r,delta:a,type:j.pop};for(const r of t)r(e,n,o)}(this.location,o,{direction:i,delta:e})}};return Object.defineProperty(o,"location",{enumerable:!0,get:()=>n[r]}),o}function Y(e){return"string"==typeof e||e&&"object"==typeof e}function J(e){return"string"==typeof e||"symbol"==typeof e}const ee=Symbol("");var te;function ne(e,t){return h(new Error,{type:e,[ee]:!0},t)}function re(e,t){return e instanceof Error&&ee in e&&(null==t||!!(e.type&t))}!function(e){e[e.aborted=4]="aborted",e[e.cancelled=8]="cancelled",e[e.duplicated=16]="duplicated"}(te||(te={}));const ae="[^/]+?",oe={sensitive:!1,strict:!1,start:!0,end:!0},ie=/[.+*?^${}()[\]/\\]/g;function se(e,t){let n=0;for(;n<e.length&&n<t.length;){const r=t[n]-e[n];if(r)return r;n++}return e.length<t.length?1===e.length&&80===e[0]?-1:1:e.length>t.length?1===t.length&&80===t[0]?1:-1:0}function le(e,t){let n=0;const r=e.score,a=t.score;for(;n<r.length&&n<a.length;){const e=se(r[n],a[n]);if(e)return e;n++}if(1===Math.abs(a.length-r.length)){if(ce(r))return 1;if(ce(a))return-1}return a.length-r.length}function ce(e){const t=e[e.length-1];return e.length>0&&t[t.length-1]<0}const ue={type:0,value:""},de=/[a-zA-Z0-9_]/;function pe(e,t,n){const r=function(e,t){const n=h({},oe,t),r=[];let a=n.start?"^":"";const o=[];for(const t of e){const e=t.length?[]:[90];n.strict&&!t.length&&(a+="/");for(let r=0;r<t.length;r++){const i=t[r];let s=40+(n.sensitive?.25:0);if(0===i.type)r||(a+="/"),a+=i.value.replace(ie,"\\$&"),s+=40;else if(1===i.type){const{value:e,repeatable:n,optional:l,regexp:c}=i;o.push({name:e,repeatable:n,optional:l});const u=c||ae;if(u!==ae){s+=10;try{new RegExp(`(${u})`)}catch(t){throw new Error(`Invalid custom RegExp for param "${e}" (${u}): `+t.message)}}let d=n?`((?:${u})(?:/(?:${u}))*)`:`(${u})`;r||(d=l&&t.length<2?`(?:/${d})`:"/"+d),l&&(d+="?"),a+=d,s+=20,l&&(s+=-8),n&&(s+=-20),".*"===u&&(s+=-50)}e.push(s)}r.push(e)}if(n.strict&&n.end){const e=r.length-1;r[e][r[e].length-1]+=.7000000000000001}n.strict||(a+="/?"),n.end?a+="$":n.strict&&!a.endsWith("/")&&(a+="(?:/|$)");const i=new RegExp(a,n.sensitive?"":"i");return{re:i,score:r,keys:o,parse:function(e){const t=e.match(i),n={};if(!t)return null;for(let e=1;e<t.length;e++){const r=t[e]||"",a=o[e-1];n[a.name]=r&&a.repeatable?r.split("/"):r}return n},stringify:function(t){let n="",r=!1;for(const a of e){r&&n.endsWith("/")||(n+="/"),r=!1;for(const e of a)if(0===e.type)n+=e.value;else if(1===e.type){const{value:o,repeatable:i,optional:s}=e,l=o in t?t[o]:"";if(g(l)&&!i)throw new Error(`Provided param "${o}" is an array but it is not repeatable (* or + modifiers)`);const c=g(l)?l.join("/"):l;if(!c){if(!s)throw new Error(`Missing required param "${o}"`);a.length<2&&(n.endsWith("/")?n=n.slice(0,-1):r=!0)}n+=c}}return n||"/"}}}(function(e){if(!e)return[[]];if("/"===e)return[[ue]];if(!e.startsWith("/"))throw new Error(`Invalid path "${e}"`);function t(e){throw new Error(`ERR (${n})/"${c}": ${e}`)}let n=0,r=n;const a=[];let o;function i(){o&&a.push(o),o=[]}let s,l=0,c="",u="";function d(){c&&(0===n?o.push({type:0,value:c}):1===n||2===n||3===n?(o.length>1&&("*"===s||"+"===s)&&t(`A repeatable param (${c}) must be alone in its segment. eg: '/:ids+.`),o.push({type:1,value:c,regexp:u,repeatable:"*"===s||"+"===s,optional:"*"===s||"?"===s})):t("Invalid state to consume buffer"),c="")}function p(){c+=s}for(;l<e.length;)if(s=e[l++],"\\"!==s||2===n)switch(n){case 0:"/"===s?(c&&d(),i()):":"===s?(d(),n=1):p();break;case 4:p(),n=r;break;case 1:"("===s?n=2:de.test(s)?p():(d(),n=0,"*"!==s&&"?"!==s&&"+"!==s&&l--);break;case 2:")"===s?"\\"==u[u.length-1]?u=u.slice(0,-1)+s:n=3:u+=s;break;case 3:d(),n=0,"*"!==s&&"?"!==s&&"+"!==s&&l--,u="";break;default:t("Unknown state")}else r=n,n=4;return 2===n&&t(`Unfinished custom RegExp for param "${c}"`),d(),i(),a}(e.path),n),a=h(r,{record:e,parent:t,children:[],alias:[]});return t&&!a.record.aliasOf==!t.record.aliasOf&&t.children.push(a),a}function he(e,t){const n={};for(const r of t)r in e&&(n[r]=e[r]);return n}function fe(e){const t={path:e.path,redirect:e.redirect,name:e.name,meta:e.meta||{},aliasOf:e.aliasOf,beforeEnter:e.beforeEnter,props:me(e),children:e.children||[],instances:{},leaveGuards:new Set,updateGuards:new Set,enterCallbacks:{},components:"components"in e?e.components||null:e.component&&{default:e.component}};return Object.defineProperty(t,"mods",{value:{}}),t}function me(e){const t={},n=e.props||!1;if("component"in e)t.default=n;else for(const r in e.components)t[r]="object"==typeof n?n[r]:n;return t}function ge(e){for(;e;){if(e.record.aliasOf)return!0;e=e.parent}return!1}function ve(e){return e.reduce(((e,t)=>h(e,t.meta)),{})}function be(e,t){const n={};for(const r in e)n[r]=r in t?t[r]:e[r];return n}function ye({record:e}){return!!(e.name||e.components&&Object.keys(e.components).length||e.redirect)}function we(e){const t={};if(""===e||"?"===e)return t;const n=("?"===e[0]?e.slice(1):e).split("&");for(let e=0;e<n.length;++e){const r=n[e].replace(k," "),a=r.indexOf("="),o=$(a<0?r:r.slice(0,a)),i=a<0?null:$(r.slice(a+1));if(o in t){let e=t[o];g(e)||(e=t[o]=[e]),e.push(i)}else t[o]=i}return t}function Oe(e){let t="";for(let n in e){const r=e[n];(n=D(n).replace(w,"%3D"),null!=r)?(g(r)?r.map((e=>e&&D(e))):[r&&D(r)]).forEach((e=>{void 0!==e&&(t+=(t.length?"&":"")+n,null!=e&&(t+="="+e))})):void 0!==r&&(t+=(t.length?"&":"")+n)}return t}function ke(e){const t={};for(const n in e){const r=e[n];void 0!==r&&(t[n]=g(r)?r.map((e=>null==e?null:""+e)):null==r?r:""+r)}return t}const xe=Symbol(""),Ee=Symbol(""),Se=Symbol(""),_e=Symbol(""),Ce=Symbol("");function Te(){let e=[];return{add:function(t){return e.push(t),()=>{const n=e.indexOf(t);n>-1&&e.splice(n,1)}},list:()=>e.slice(),reset:function(){e=[]}}}function Ae(e,t,n,r,a,o=e=>e()){const i=r&&(r.enterCallbacks[a]=r.enterCallbacks[a]||[]);return()=>new Promise(((s,l)=>{const c=e=>{!1===e?l(ne(4,{from:n,to:t})):e instanceof Error?l(e):Y(e)?l(ne(2,{from:t,to:e})):(i&&r.enterCallbacks[a]===i&&"function"==typeof e&&i.push(e),s())},u=o((()=>e.call(r&&r.instances[a],t,n,c)));let d=Promise.resolve(u);e.length<3&&(d=d.then(c)),d.catch((e=>l(e)))}))}function Re(e,t,n,r,a=e=>e()){const o=[];for(const i of e)for(const e in i.components){let s=i.components[e];if("beforeRouteEnter"===t||i.instances[e])if(p(s)){const l=(s.__vccOpts||s)[t];l&&o.push(Ae(l,n,r,i,e,a))}else{let l=s();o.push((()=>l.then((o=>{if(!o)throw new Error(`Couldn't resolve component "${e}" at "${i.path}"`);const s=(l=o).__esModule||"Module"===l[Symbol.toStringTag]||l.default&&p(l.default)?o.default:o;var l;i.mods[e]=o,i.components[e]=s;const c=(s.__vccOpts||s)[t];return c&&Ae(c,n,r,i,e,a)()}))))}}return o}function Pe(e){const t=(0,r.WQ)(Se),n=(0,r.WQ)(_e),o=(0,r.EW)((()=>{const n=(0,a.R1)(e.to);return t.resolve(n)})),i=(0,r.EW)((()=>{const{matched:e}=o.value,{length:t}=e,r=e[t-1],a=n.matched;if(!r||!a.length)return-1;const i=a.findIndex(B.bind(null,r));if(i>-1)return i;const s=Me(e[t-2]);return t>1&&Me(r)===s&&a[a.length-1].path!==s?a.findIndex(B.bind(null,e[t-2])):i})),s=(0,r.EW)((()=>i.value>-1&&function(e,t){for(const n in t){const r=t[n],a=e[n];if("string"==typeof r){if(r!==a)return!1}else if(!g(a)||a.length!==r.length||r.some(((e,t)=>e!==a[t])))return!1}return!0}(n.params,o.value.params))),l=(0,r.EW)((()=>i.value>-1&&i.value===n.matched.length-1&&F(n.params,o.value.params)));if(__VUE_PROD_DEVTOOLS__&&d){const t=(0,r.nI)();if(t){const n={route:o.value,isActive:s.value,isExactActive:l.value,error:null};t.__vrl_devtools=t.__vrl_devtools||[],t.__vrl_devtools.push(n),(0,r.nT)((()=>{n.route=o.value,n.isActive=s.value,n.isExactActive=l.value,n.error=Y((0,a.R1)(e.to))?null:'Invalid "to" value'}),{flush:"post"})}}return{route:o,href:(0,r.EW)((()=>o.value.href)),isActive:s,isExactActive:l,navigate:function(n={}){if(function(e){if(!(e.metaKey||e.altKey||e.ctrlKey||e.shiftKey||e.defaultPrevented||void 0!==e.button&&0!==e.button)){if(e.currentTarget&&e.currentTarget.getAttribute){const t=e.currentTarget.getAttribute("target");if(/\b_blank\b/i.test(t))return}return e.preventDefault&&e.preventDefault(),!0}}(n)){const n=t[(0,a.R1)(e.replace)?"replace":"push"]((0,a.R1)(e.to)).catch(m);return e.viewTransition&&"undefined"!=typeof document&&"startViewTransition"in document&&document.startViewTransition((()=>n)),n}return Promise.resolve()}}}const De=(0,r.pM)({name:"RouterLink",compatConfig:{MODE:3},props:{to:{type:[String,Object],required:!0},replace:Boolean,activeClass:String,exactActiveClass:String,custom:Boolean,ariaCurrentValue:{type:String,default:"page"}},useLink:Pe,setup(e,{slots:t}){const n=(0,a.Kh)(Pe(e)),{options:o}=(0,r.WQ)(Se),i=(0,r.EW)((()=>({[$e(e.activeClass,o.linkActiveClass,"router-link-active")]:n.isActive,[$e(e.exactActiveClass,o.linkExactActiveClass,"router-link-exact-active")]:n.isExactActive})));return()=>{const a=t.default&&(1===(o=t.default(n)).length?o[0]:o);var o;return e.custom?a:(0,r.h)("a",{"aria-current":n.isExactActive?e.ariaCurrentValue:null,href:n.href,onClick:n.navigate,class:i.value},a)}}});function Me(e){return e?e.aliasOf?e.aliasOf.path:e.path:""}const $e=(e,t,n)=>null!=e?e:null!=t?t:n;function Ne(e,t){if(!e)return null;const n=e(t);return 1===n.length?n[0]:n}const Ie=(0,r.pM)({name:"RouterView",inheritAttrs:!1,props:{name:{type:String,default:"default"},route:Object},compatConfig:{MODE:3},setup(e,{attrs:t,slots:n}){const o=(0,r.WQ)(Ce),i=(0,r.EW)((()=>e.route||o.value)),s=(0,r.WQ)(Ee,0),l=(0,r.EW)((()=>{let e=(0,a.R1)(s);const{matched:t}=i.value;let n;for(;(n=t[e])&&!n.components;)e++;return e})),c=(0,r.EW)((()=>i.value.matched[l.value]));(0,r.Gt)(Ee,(0,r.EW)((()=>l.value+1))),(0,r.Gt)(xe,c),(0,r.Gt)(Ce,i);const u=(0,a.KR)();return(0,r.wB)((()=>[u.value,c.value,e.name]),(([e,t,n],[r,a,o])=>{t&&(t.instances[n]=e,a&&a!==t&&e&&e===r&&(t.leaveGuards.size||(t.leaveGuards=a.leaveGuards),t.updateGuards.size||(t.updateGuards=a.updateGuards))),!e||!t||a&&B(t,a)&&r||(t.enterCallbacks[n]||[]).forEach((t=>t(e)))}),{flush:"post"}),()=>{const a=i.value,o=e.name,s=c.value,p=s&&s.components[o];if(!p)return Ne(n.default,{Component:p,route:a});const f=s.props[o],m=f?!0===f?a.params:"function"==typeof f?f(a):f:null,v=(0,r.h)(p,h({},m,t,{onVnodeUnmounted:e=>{e.component.isUnmounted&&(s.instances[o]=null)},ref:u}));if(__VUE_PROD_DEVTOOLS__&&d&&v.ref){const e={depth:l.value,name:s.name,path:s.path,meta:s.meta};(g(v.ref)?v.ref.map((e=>e.i)):[v.ref.i]).forEach((t=>{t.__vrv_devtools=e}))}return Ne(n.default,{Component:v,route:a})||v}}});function Le(e,t){const n=h({},e,{matched:e.matched.map((e=>function(e,t){const n={};for(const r in e)t.includes(r)||(n[r]=e[r]);return n}(e,["instances","children","aliasOf"])))});return{_custom:{type:null,readOnly:!0,display:e.fullPath,tooltip:t,value:n}}}function Be(e){return{_custom:{display:e}}}let Fe=0;function Qe(e,t,n){if(t.__hasDevtools)return;t.__hasDevtools=!0;const a=Fe++;!function(e,t){const n=e,r=o(),a=o().__VUE_DEVTOOLS_GLOBAL_HOOK__,l=i&&n.enableEarlyProxy;if(!a||!r.__VUE_DEVTOOLS_PLUGIN_API_AVAILABLE__&&l){const e=l?new u(n,a):null;(r.__VUE_DEVTOOLS_PLUGINS__=r.__VUE_DEVTOOLS_PLUGINS__||[]).push({pluginDescriptor:n,setupFn:t,proxy:e}),e&&t(e.proxiedTarget)}else a.emit(s,e,t)}({id:"org.vuejs.router"+(a?"."+a:""),label:"Vue Router",packageName:"vue-router",homepage:"https://router.vuejs.org",logo:"https://router.vuejs.org/logo.png",componentStateTypes:["Routing"],app:e},(o=>{"function"!=typeof o.now&&console.warn("[Vue Router]: You seem to be using an outdated version of Vue Devtools. Are you still using the Beta release instead of the stable one? You can find the links at https://devtools.vuejs.org/guide/installation.html."),o.on.inspectComponent(((e,n)=>{e.instanceData&&e.instanceData.state.push({type:"Routing",key:"$route",editable:!1,value:Le(t.currentRoute.value,"Current Route")})})),o.on.visitComponentTree((({treeNode:e,componentInstance:t})=>{if(t.__vrv_devtools){const n=t.__vrv_devtools;e.tags.push({label:(n.name?`${n.name.toString()}: `:"")+n.path,textColor:0,tooltip:"This component is rendered by &lt;router-view&gt;",backgroundColor:ze})}g(t.__vrl_devtools)&&(t.__devtoolsApi=o,t.__vrl_devtools.forEach((t=>{let n=t.route.path,r=qe,a="",o=0;t.error?(n=t.error,r=He,o=Ze):t.isExactActive?(r=We,a="This is exactly active"):t.isActive&&(r=je,a="This link is active"),e.tags.push({label:n,textColor:o,tooltip:a,backgroundColor:r})})))})),(0,r.wB)(t.currentRoute,(()=>{c(),o.notifyComponentUpdate(),o.sendInspectorTree(l),o.sendInspectorState(l)}));const i="router:navigations:"+a;o.addTimelineLayer({id:i,label:`Router${a?" "+a:""} Navigations`,color:4237508}),t.onError(((e,t)=>{o.addTimelineEvent({layerId:i,event:{title:"Error during Navigation",subtitle:t.fullPath,logType:"error",time:o.now(),data:{error:e},groupId:t.meta.__navigationId}})}));let s=0;t.beforeEach(((e,t)=>{const n={guard:Be("beforeEach"),from:Le(t,"Current Location during this navigation"),to:Le(e,"Target location")};Object.defineProperty(e.meta,"__navigationId",{value:s++}),o.addTimelineEvent({layerId:i,event:{time:o.now(),title:"Start of navigation",subtitle:e.fullPath,data:n,groupId:e.meta.__navigationId}})})),t.afterEach(((e,t,n)=>{const r={guard:Be("afterEach")};n?(r.failure={_custom:{type:Error,readOnly:!0,display:n?n.message:"",tooltip:"Navigation Failure",value:n}},r.status=Be("❌")):r.status=Be("✅"),r.from=Le(t,"Current Location during this navigation"),r.to=Le(e,"Target location"),o.addTimelineEvent({layerId:i,event:{title:"End of navigation",subtitle:e.fullPath,time:o.now(),data:r,logType:n?"warning":"default",groupId:e.meta.__navigationId}})}));const l="router-inspector:"+a;function c(){if(!u)return;const e=u;let r=n.getRoutes().filter((e=>!e.parent||!e.parent.record.components));r.forEach(et),e.filter&&(r=r.filter((t=>tt(t,e.filter.toLowerCase())))),r.forEach((e=>Je(e,t.currentRoute.value))),e.rootNodes=r.map(Ge)}let u;o.addInspector({id:l,label:"Routes"+(a?" "+a:""),icon:"book",treeFilterPlaceholder:"Search routes"}),o.on.getInspectorTree((t=>{u=t,t.app===e&&t.inspectorId===l&&c()})),o.on.getInspectorState((t=>{if(t.app===e&&t.inspectorId===l){const e=n.getRoutes().find((e=>e.record.__vd_id===t.nodeId));e&&(t.state={options:Ue(e)})}})),o.sendInspectorTree(l),o.sendInspectorState(l)}))}function Ue(e){const{record:t}=e,n=[{editable:!1,key:"path",value:t.path}];return null!=t.name&&n.push({editable:!1,key:"name",value:t.name}),n.push({editable:!1,key:"regexp",value:e.re}),e.keys.length&&n.push({editable:!1,key:"keys",value:{_custom:{type:null,readOnly:!0,display:e.keys.map((e=>`${e.name}${function(e){return e.optional?e.repeatable?"*":"?":e.repeatable?"+":""}(e)}`)).join(" "),tooltip:"Param keys",value:e.keys}}}),null!=t.redirect&&n.push({editable:!1,key:"redirect",value:t.redirect}),e.alias.length&&n.push({editable:!1,key:"aliases",value:e.alias.map((e=>e.record.path))}),Object.keys(e.record.meta).length&&n.push({editable:!1,key:"meta",value:e.record.meta}),n.push({key:"score",editable:!1,value:{_custom:{type:null,readOnly:!0,display:e.score.map((e=>e.join(", "))).join(" | "),tooltip:"Score used to sort routes",value:e.score}}}),n}const ze=15485081,je=2450411,We=8702998,Xe=2282478,qe=16486972,Ve=6710886,He=16704226,Ze=12131356;function Ge(e){const t=[],{record:n}=e;null!=n.name&&t.push({label:String(n.name),textColor:0,backgroundColor:Xe}),n.aliasOf&&t.push({label:"alias",textColor:0,backgroundColor:qe}),e.__vd_match&&t.push({label:"matches",textColor:0,backgroundColor:ze}),e.__vd_exactActive&&t.push({label:"exact",textColor:0,backgroundColor:We}),e.__vd_active&&t.push({label:"active",textColor:0,backgroundColor:je}),n.redirect&&t.push({label:"string"==typeof n.redirect?`redirect: ${n.redirect}`:"redirects",textColor:16777215,backgroundColor:Ve});let r=n.__vd_id;return null==r&&(r=String(Ke++),n.__vd_id=r),{id:r,label:n.path,tags:t,children:e.children.map(Ge)}}let Ke=0;const Ye=/^\/(.*)\/([a-z]*)$/;function Je(e,t){const n=t.matched.length&&B(t.matched[t.matched.length-1],e.record);e.__vd_exactActive=e.__vd_active=n,n||(e.__vd_active=t.matched.some((t=>B(t,e.record)))),e.children.forEach((e=>Je(e,t)))}function et(e){e.__vd_match=!1,e.children.forEach(et)}function tt(e,t){const n=String(e.re).match(Ye);if(e.__vd_match=!1,!n||n.length<3)return!1;if(new RegExp(n[1].replace(/\$$/,""),n[2]).test(t))return e.children.forEach((e=>tt(e,t))),("/"!==e.record.path||"/"===t)&&(e.__vd_match=e.re.test(t),!0);const r=e.record.path.toLowerCase(),a=$(r);return!(t.startsWith("/")||!a.includes(t)&&!r.includes(t))||!(!a.startsWith(t)&&!r.startsWith(t))||!(!e.record.name||!String(e.record.name).includes(t))||e.children.some((e=>tt(e,t)))}function nt(e){const t=function(e,t){const n=[],r=new Map;function a(e,n,r){const s=!r,l=fe(e);l.aliasOf=r&&r.record;const c=be(t,e),u=[l];if("alias"in e){const t="string"==typeof e.alias?[e.alias]:e.alias;for(const e of t)u.push(fe(h({},l,{components:r?r.record.components:l.components,path:e,aliasOf:r?r.record:l})))}let d,p;for(const t of u){const{path:u}=t;if(n&&"/"!==u[0]){const e=n.record.path,r="/"===e[e.length-1]?"":"/";t.path=n.record.path+(u&&r+u)}if(d=pe(t,n,c),r?r.alias.push(d):(p=p||d,p!==d&&p.alias.push(d),s&&e.name&&!ge(d)&&o(e.name)),ye(d)&&i(d),l.children){const e=l.children;for(let t=0;t<e.length;t++)a(e[t],d,r&&r.children[t])}r=r||d}return p?()=>{o(p)}:m}function o(e){if(J(e)){const t=r.get(e);t&&(r.delete(e),n.splice(n.indexOf(t),1),t.children.forEach(o),t.alias.forEach(o))}else{const t=n.indexOf(e);t>-1&&(n.splice(t,1),e.record.name&&r.delete(e.record.name),e.children.forEach(o),e.alias.forEach(o))}}function i(e){const t=function(e,t){let n=0,r=t.length;for(;n!==r;){const a=n+r>>1;le(e,t[a])<0?r=a:n=a+1}const a=function(e){let t=e;for(;t=t.parent;)if(ye(t)&&0===le(e,t))return t}(e);return a&&(r=t.lastIndexOf(a,r-1)),r}(e,n);n.splice(t,0,e),e.record.name&&!ge(e)&&r.set(e.record.name,e)}return t=be({strict:!1,end:!0,sensitive:!1},t),e.forEach((e=>a(e))),{addRoute:a,resolve:function(e,t){let a,o,i,s={};if("name"in e&&e.name){if(a=r.get(e.name),!a)throw ne(1,{location:e});i=a.record.name,s=h(he(t.params,a.keys.filter((e=>!e.optional)).concat(a.parent?a.parent.keys.filter((e=>e.optional)):[]).map((e=>e.name))),e.params&&he(e.params,a.keys.map((e=>e.name)))),o=a.stringify(s)}else if(null!=e.path)o=e.path,a=n.find((e=>e.re.test(o))),a&&(s=a.parse(o),i=a.record.name);else{if(a=t.name?r.get(t.name):n.find((e=>e.re.test(t.path))),!a)throw ne(1,{location:e,currentLocation:t});i=a.record.name,s=h({},t.params,e.params),o=a.stringify(s)}const l=[];let c=a;for(;c;)l.unshift(c.record),c=c.parent;return{name:i,path:o,params:s,matched:l,meta:ve(l)}},removeRoute:o,clearRoutes:function(){n.length=0,r.clear()},getRoutes:function(){return n},getRecordMatcher:function(e){return r.get(e)}}}(e.routes,e),n=e.parseQuery||we,o=e.stringifyQuery||Oe,i=e.history,s=Te(),l=Te(),c=Te(),u=(0,a.IJ)(z);let p=z;d&&e.scrollBehavior&&"scrollRestoration"in history&&(history.scrollRestoration="manual");const v=f.bind(null,(e=>""+e)),b=f.bind(null,M),y=f.bind(null,$);function w(e,r){if(r=h({},r||u.value),"string"==typeof e){const a=L(n,e,r.path),o=t.resolve({path:a.path},r),s=i.createHref(a.fullPath);return h(a,o,{params:y(o.params),hash:$(a.hash),redirectedFrom:void 0,href:s})}let a;if(null!=e.path)a=h({},e,{path:L(n,e.path,r.path).path});else{const t=h({},e.params);for(const e in t)null==t[e]&&delete t[e];a=h({},e,{params:b(t)}),r.params=b(r.params)}const s=t.resolve(a,r),l=e.hash||"";s.params=v(y(s.params));const c=function(e,t){const n=t.query?e(t.query):"";return t.path+(n&&"?")+n+(t.hash||"")}(o,h({},e,{hash:(d=l,P(d).replace(C,"{").replace(A,"}").replace(S,"^")),path:s.path}));var d;const p=i.createHref(c);return h({fullPath:c,hash:l,query:o===Oe?ke(e.query):e.query||{}},s,{redirectedFrom:void 0,href:p})}function O(e){return"string"==typeof e?L(n,e,u.value.path):h({},e)}function k(e,t){if(p!==e)return ne(8,{from:t,to:e})}function x(e){return _(e)}function E(e){const t=e.matched[e.matched.length-1];if(t&&t.redirect){const{redirect:n}=t;let r="function"==typeof n?n(e):n;return"string"==typeof r&&(r=r.includes("?")||r.includes("#")?r=O(r):{path:r},r.params={}),h({query:e.query,hash:e.hash,params:null!=r.path?{}:e.params},r)}}function _(e,t){const n=p=w(e),r=u.value,a=e.state,i=e.force,s=!0===e.replace,l=E(n);if(l)return _(h(O(l),{state:"object"==typeof l?h({},a,l.state):a,force:i,replace:s}),t||n);const c=n;let d;return c.redirectedFrom=t,!i&&function(e,t,n){const r=t.matched.length-1,a=n.matched.length-1;return r>-1&&r===a&&B(t.matched[r],n.matched[a])&&F(t.params,n.params)&&e(t.query)===e(n.query)&&t.hash===n.hash}(o,r,n)&&(d=ne(16,{to:c,from:r}),K(r,r,!0,!1)),(d?Promise.resolve(d):D(c,r)).catch((e=>re(e)?re(e,2)?e:V(e):q(e,c,r))).then((e=>{if(e){if(re(e,2))return _(h({replace:s},O(e.to),{state:"object"==typeof e.to?h({},a,e.to.state):a,force:i}),t||c)}else e=I(c,r,!0,s,a);return N(c,r,e),e}))}function T(e,t){const n=k(e,t);return n?Promise.reject(n):Promise.resolve()}function R(e){const t=te.values().next().value;return t&&"function"==typeof t.runWithContext?t.runWithContext(e):e()}function D(e,t){let n;const[r,a,o]=function(e,t){const n=[],r=[],a=[],o=Math.max(t.matched.length,e.matched.length);for(let i=0;i<o;i++){const o=t.matched[i];o&&(e.matched.find((e=>B(e,o)))?r.push(o):n.push(o));const s=e.matched[i];s&&(t.matched.find((e=>B(e,s)))||a.push(s))}return[n,r,a]}(e,t);n=Re(r.reverse(),"beforeRouteLeave",e,t);for(const a of r)a.leaveGuards.forEach((r=>{n.push(Ae(r,e,t))}));const i=T.bind(null,e,t);return n.push(i),oe(n).then((()=>{n=[];for(const r of s.list())n.push(Ae(r,e,t));return n.push(i),oe(n)})).then((()=>{n=Re(a,"beforeRouteUpdate",e,t);for(const r of a)r.updateGuards.forEach((r=>{n.push(Ae(r,e,t))}));return n.push(i),oe(n)})).then((()=>{n=[];for(const r of o)if(r.beforeEnter)if(g(r.beforeEnter))for(const a of r.beforeEnter)n.push(Ae(a,e,t));else n.push(Ae(r.beforeEnter,e,t));return n.push(i),oe(n)})).then((()=>(e.matched.forEach((e=>e.enterCallbacks={})),n=Re(o,"beforeRouteEnter",e,t,R),n.push(i),oe(n)))).then((()=>{n=[];for(const r of l.list())n.push(Ae(r,e,t));return n.push(i),oe(n)})).catch((e=>re(e,8)?e:Promise.reject(e)))}function N(e,t,n){c.list().forEach((r=>R((()=>r(e,t,n)))))}function I(e,t,n,r,a){const o=k(e,t);if(o)return o;const s=t===z,l=d?history.state:{};n&&(r||s?i.replace(e.fullPath,h({scroll:s&&l&&l.scroll},a)):i.push(e.fullPath,a)),u.value=e,K(e,t,n,s),V()}let Q;let U,W=Te(),X=Te();function q(e,t,n){V(e);const r=X.list();return r.length?r.forEach((r=>r(e,t,n))):console.error(e),Promise.reject(e)}function V(e){return U||(U=!e,Q||(Q=i.listen(((e,t,n)=>{if(!ae.listening)return;const r=w(e),a=E(r);if(a)return void _(h(a,{replace:!0,force:!0}),r).catch(m);p=r;const o=u.value;var s,l;d&&(s=Z(o.fullPath,n.delta),l=H(),G.set(s,l)),D(r,o).catch((e=>re(e,12)?e:re(e,2)?(_(h(O(e.to),{force:!0}),r).then((e=>{re(e,20)&&!n.delta&&n.type===j.pop&&i.go(-1,!1)})).catch(m),Promise.reject()):(n.delta&&i.go(-n.delta,!1),q(e,r,o)))).then((e=>{(e=e||I(r,o,!1))&&(n.delta&&!re(e,8)?i.go(-n.delta,!1):n.type===j.pop&&re(e,20)&&i.go(-1,!1)),N(r,o,e)})).catch(m)}))),W.list().forEach((([t,n])=>e?n(e):t())),W.reset()),e}function K(t,n,a,o){const{scrollBehavior:i}=e;if(!d||!i)return Promise.resolve();const s=!a&&function(e){const t=G.get(e);return G.delete(e),t}(Z(t.fullPath,0))||(o||!a)&&history.state&&history.state.scroll||null;return(0,r.dY)().then((()=>i(t,n,s))).then((e=>e&&function(e){let t;if("el"in e){const n=e.el,r="string"==typeof n&&n.startsWith("#"),a="string"==typeof n?r?document.getElementById(n.slice(1)):document.querySelector(n):n;if(!a)return;t=function(e,t){const n=document.documentElement.getBoundingClientRect(),r=e.getBoundingClientRect();return{behavior:t.behavior,left:r.left-n.left-(t.left||0),top:r.top-n.top-(t.top||0)}}(a,e)}else t=e;"scrollBehavior"in document.documentElement.style?window.scrollTo(t):window.scrollTo(null!=t.left?t.left:window.scrollX,null!=t.top?t.top:window.scrollY)}(e))).catch((e=>q(e,t,n)))}const Y=e=>i.go(e);let ee;const te=new Set,ae={currentRoute:u,listening:!0,addRoute:function(e,n){let r,a;return J(e)?(r=t.getRecordMatcher(e),a=n):a=e,t.addRoute(a,r)},removeRoute:function(e){const n=t.getRecordMatcher(e);n&&t.removeRoute(n)},clearRoutes:t.clearRoutes,hasRoute:function(e){return!!t.getRecordMatcher(e)},getRoutes:function(){return t.getRoutes().map((e=>e.record))},resolve:w,options:e,push:x,replace:function(e){return x(h(O(e),{replace:!0}))},go:Y,back:()=>Y(-1),forward:()=>Y(1),beforeEach:s.add,beforeResolve:l.add,afterEach:c.add,onError:X.add,isReady:function(){return U&&u.value!==z?Promise.resolve():new Promise(((e,t)=>{W.add([e,t])}))},install(e){const n=this;e.component("RouterLink",De),e.component("RouterView",Ie),e.config.globalProperties.$router=n,Object.defineProperty(e.config.globalProperties,"$route",{enumerable:!0,get:()=>(0,a.R1)(u)}),d&&!ee&&u.value===z&&(ee=!0,x(i.location).catch((e=>{})));const r={};for(const e in z)Object.defineProperty(r,e,{get:()=>u.value[e],enumerable:!0});e.provide(Se,n),e.provide(_e,(0,a.Gc)(r)),e.provide(Ce,u);const o=e.unmount;te.add(e),e.unmount=function(){te.delete(e),te.size<1&&(p=z,Q&&Q(),Q=null,u.value=z,ee=!1,U=!1),o()},__VUE_PROD_DEVTOOLS__&&d&&Qe(e,n,t)}};function oe(e){return e.reduce(((e,t)=>e.then((()=>R(t)))),Promise.resolve())}return ae}function rt(){return(0,r.WQ)(Se)}function at(e){return(0,r.WQ)(_e)}},9388:(e,t,n)=>{"use strict";function r(e,t,...n){if(e in t){let r=t[e];return"function"==typeof r?r(...n):r}let a=new Error(`Tried to handle "${e}" but there is no handler defined. Only defined handlers are: ${Object.keys(t).map((e=>`"${e}"`)).join(", ")}.`);throw Error.captureStackTrace&&Error.captureStackTrace(a,r),a}n.d(t,{Y:()=>r})},9464:(e,t,n)=>{"use strict";n.d(t,{u:()=>C});var r=n(9143);const a=function(e,t,n){const a=(0,r.C)(n);if(!e||!e.type||!e.children)throw new Error("Expected parent node");if("number"==typeof t){if(t<0||t===Number.POSITIVE_INFINITY)throw new Error("Expected positive finite number as index")}else if((t=e.children.indexOf(t))<0)throw new Error("Expected child node or index");for(;++t<e.children.length;)if(a(e.children[t],t,e))return e.children[t]};var o=n(9747);const i=/\n/g,s=/[\t ]+/g,l=(0,o.a)("br"),c=(0,o.a)((function(e){return"td"===e.tagName||"th"===e.tagName})),u=(0,o.a)("p"),d=(0,o.a)("tr"),p=(0,o.a)(["datalist","head","noembed","noframes","noscript","rp","script","style","template","title",function(e){return Boolean((e.properties||{}).hidden)},function(e){return"dialog"===e.tagName&&!(e.properties||{}).open}]),h=(0,o.a)(["address","article","aside","blockquote","body","caption","center","dd","dialog","dir","dl","dt","div","figure","figcaption","footer","form,","h1","h2","h3","h4","h5","h6","header","hgroup","hr","html","legend","li","listing","main","menu","nav","ol","p","plaintext","pre","section","ul","xmp"]);function f(e,t){const n=t||{},r="children"in e?e.children:[],a=h(e),o=b(e,{whitespace:n.whitespace||"normal",breakBefore:!1,breakAfter:!1}),i=[];"text"!==e.type&&"comment"!==e.type||i.push(...g(e,{whitespace:o,breakBefore:!0,breakAfter:!0}));let s=-1;for(;++s<r.length;)i.push(...m(r[s],e,{whitespace:o,breakBefore:s?void 0:a,breakAfter:s<r.length-1?l(r[s+1]):a}));const c=[];let u;for(s=-1;++s<i.length;){const e=i[s];"number"==typeof e?void 0!==u&&e>u&&(u=e):e&&(void 0!==u&&u>-1&&c.push("\n".repeat(u)||" "),u=-1,c.push(e))}return c.join("")}function m(e,t,n){return"element"===e.type?function(e,t,n){const r=b(e,n),o=e.children||[];let i,s,f=-1,g=[];if(p(e))return g;for(l(e)||d(e)&&a(t,e,d)?s="\n":u(e)?(i=2,s=2):h(e)&&(i=1,s=1);++f<o.length;)g=g.concat(m(o[f],e,{whitespace:r,breakBefore:f?void 0:i,breakAfter:f<o.length-1?l(o[f+1]):s}));return c(e)&&a(t,e,c)&&g.push("\t"),i&&g.unshift(i),s&&g.push(s),g}(e,t,n):"text"===e.type?"normal"===n.whitespace?g(e,n):function(e){return[String(e.value)]}(e):[]}function g(e,t){const n=String(e.value),r=[],a=[];let o=0;for(;o<=n.length;){i.lastIndex=o;const e=i.exec(n),a=e&&"index"in e?e.index:n.length;r.push(v(n.slice(o,a).replace(/[\u061C\u200E\u200F\u202A-\u202E\u2066-\u2069]/g,""),0!==o||t.breakBefore,a!==n.length||t.breakAfter)),o=a+1}let s,l=-1;for(;++l<r.length;)8203===r[l].charCodeAt(r[l].length-1)||l<r.length-1&&8203===r[l+1].charCodeAt(0)?(a.push(r[l]),s=void 0):r[l]?("number"==typeof s&&a.push(s),a.push(r[l]),s=0):0!==l&&l!==r.length-1||a.push(0);return a}function v(e,t,n){const r=[];let a,o=0;for(;o<e.length;){s.lastIndex=o;const n=s.exec(e);a=n?n.index:e.length,o||a||!n||t||r.push(""),o!==a&&r.push(e.slice(o,a)),o=n?a+n[0].length:a}return o===a||n||r.push(""),r.join(" ")}function b(e,t){if("element"===e.type){const n=e.properties||{};switch(e.tagName){case"listing":case"plaintext":case"xmp":return"pre";case"nobr":return"nowrap";case"pre":return n.wrap?"pre-wrap":"pre";case"td":case"th":return n.noWrap?"nowrap":t.whitespace;case"textarea":return"pre-wrap"}}return t.whitespace}var y=n(7471);const w=n(8416),O={},k="hljs-";class x{constructor(e){this.options=e,this.root={type:"root",children:[],data:{language:void 0,relevance:0}},this.stack=[this.root]}addText(e){if(""===e)return;const t=this.stack[this.stack.length-1],n=t.children[t.children.length-1];n&&"text"===n.type?n.value+=e:t.children.push({type:"text",value:e})}startScope(e){this.openNode(String(e))}endScope(){this.closeNode()}__addSublanguage(e,t){const n=this.stack[this.stack.length-1],r=e.root.children;t?n.children.push({type:"element",tagName:"span",properties:{className:[t]},children:r}):n.children.push(...r)}openNode(e){const t=this,n={type:"element",tagName:"span",properties:{className:e.split(".").map((function(e,n){return n?e+"_".repeat(n):t.options.classPrefix+e}))},children:[]};this.stack[this.stack.length-1].children.push(n),this.stack.push(n)}closeNode(){this.stack.pop()}finalize(){}toHTML(){return""}}var E=n(7970),S=n(9751);const _={};function C(e){const t=e||_,n=t.aliases,r=e?.detect??!1,a=t.languages,o=t.plainText,i=t.prefix,s=t.subset;let l="hljs";const c=e?.lowlight??function(e){const t=w.newInstance();return e&&a(e),{highlight:n,highlightAuto:function(e,a){(0,y.ok)("string"==typeof e,"expected `string` as `value`");const o=(a||O).subset||r();let i,s=-1,l=0;for(;++s<o.length;){const r=o[s];if(!t.getLanguage(r))continue;const c=n(r,e,a);c.data&&void 0!==c.data.relevance&&c.data.relevance>l&&(l=c.data.relevance,i=c)}return i||{type:"root",children:[],data:{language:void 0,relevance:l}}},listLanguages:r,register:a,registerAlias:function(e,n){if("string"==typeof e)(0,y.ok)(void 0!==n),t.registerAliases("string"==typeof n?n:[...n],{languageName:e});else{let n;for(n in e)if(Object.hasOwn(e,n)){const r=e[n];t.registerAliases("string"==typeof r?r:[...r],{languageName:n})}}},registered:function(e){return Boolean(t.getLanguage(e))}};function n(e,n,r){(0,y.ok)("string"==typeof e,"expected `string` as `name`"),(0,y.ok)("string"==typeof n,"expected `string` as `value`");const a=r||O,o="string"==typeof a.prefix?a.prefix:k;if(!t.getLanguage(e))throw new Error("Unknown language: `"+e+"` is not registered");t.configure({__emitter:x,classPrefix:o});const i=t.highlight(n,{ignoreIllegals:!0,language:e});if(i.errorRaised)throw new Error("Could not highlight with `Highlight.js`",{cause:i.errorRaised});const s=i._emitter.root,l=s.data;return l.language=i.language,l.relevance=i.relevance,s}function r(){return t.listLanguages()}function a(e,n){if("string"==typeof e)(0,y.ok)(void 0!==n,"expected `grammar`"),t.registerLanguage(e,n);else{let n;for(n in e)Object.hasOwn(e,n)&&t.registerLanguage(n,e[n])}}}(a);if(n&&c.registerAlias(n),i){const e=i.indexOf("-");l=e>-1?i.slice(0,e):i}return(e,t)=>{(0,E.YR)(e,"element",((e,n,a)=>{if("code"!==e.tagName||!a||"element"!==a.type||"pre"!==a.tagName)return;const u=function(e){const t=e.properties.className;if(!Array.isArray(t))return"";const n=t.reduce(((e,t)=>{if(e)return e;const n=String(t);return"no-highlight"===n||"nohighlight"===n?"no-highlight":"lang-"===n.slice(0,5)?n.slice(5):"language-"===n.slice(0,9)?n.slice(9):e}),"");return S.I[n||""]||n}(e);if("no-highlight"===u||!u&&!r||u&&o?.includes(u))return;let d;Array.isArray(e.properties.className)||(e.properties.className=[]),e.properties.className.includes(l)||e.properties.className.unshift(l);try{d=u?c.highlight(u,f(a),{prefix:i}):c.highlightAuto(f(a),{prefix:i,subset:s})}catch(n){const r=n;if(u&&/Unknown language/.test(r.message))return void t.message("Cannot highlight as `"+u+"`, it’s not registered",{ancestors:[a,e],cause:r,place:e.position,ruleId:"missing-language",source:"rehype-highlight"});throw r}!u&&d.data?.language&&e.properties.className.push("language-"+d.data.language),d.children.length>0&&(e.children=d.children)}))}}},9563:e=>{"use strict";for(var t="(?:"+["\\|\\|","\\&\\&",";;","\\|\\&","\\<\\(","\\<\\<\\<",">>",">\\&","<\\&","[&;()|<>]"].join("|")+")",n=new RegExp("^"+t+"$"),r="|&;()<> \\t",a=/^#$/,o="",i=0;i<4;i++)o+=(4294967296*Math.random()).toString(16);var s=new RegExp("^"+o);e.exports=function(e,i,l){var c=function(e,i,s){s||(s={});var l=s.escape||"\\",c=new RegExp(["("+t+")","((\\"+l+"['\""+r+"]|[^\\s'\""+r+"])+|\"((\\\\\"|[^\"])*?)\"|'((\\\\'|[^'])*?)')+"].join("|"),"g"),u=function(e,t){for(var n,r=t.lastIndex,a=[];n=t.exec(e);)a.push(n),t.lastIndex===n.index&&(t.lastIndex+=1);return t.lastIndex=r,a}(e,c);if(0===u.length)return[];i||(i={});var d=!1;return u.map((function(t){var r=t[0];if(r&&!d){if(n.test(r))return{op:r};var s,c=!1,u=!1,p="",h=!1;for(s=0;s<r.length;s++){var f=r.charAt(s);if(h=h||!c&&("*"===f||"?"===f),u)p+=f,u=!1;else if(c)f===c?c=!1:"'"==c?p+=f:f===l?(s+=1,p+='"'===(f=r.charAt(s))||f===l||"$"===f?f:l+f):p+="$"===f?g():f;else if('"'===f||"'"===f)c=f;else{if(n.test(f))return{op:r};if(a.test(f)){d=!0;var m={comment:e.slice(t.index+s+1)};return p.length?[p,m]:[m]}f===l?u=!0:p+="$"===f?g():f}}return h?{op:"glob",pattern:p}:p}function g(){var e,t;s+=1;var n=r.charAt(s);if("{"===n){if(s+=1,"}"===r.charAt(s))throw new Error("Bad substitution: "+r.slice(s-2,s+1));if((e=r.indexOf("}",s))<0)throw new Error("Bad substitution: "+r.slice(s));t=r.slice(s,e),s=e}else if(/[*@#?$!_-]/.test(n))t=n,s+=1;else{var a=r.slice(s);(e=a.match(/[^\w\d_]/))?(t=a.slice(0,e.index),s+=e.index-1):(t=a,s=r.length)}return function(e,t,n){var r="function"==typeof e?e(n):e[n];return void 0===r&&""!=n?r="":void 0===r&&(r="$"),"object"==typeof r?""+o+JSON.stringify(r)+o:""+r}(i,0,t)}})).reduce((function(e,t){return void 0===t?e:e.concat(t)}),[])}(e,i,l);return"function"!=typeof i?c:c.reduce((function(e,t){if("object"==typeof t)return e.concat(t);var n=t.split(RegExp("("+o+".*?"+o+")","g"));return 1===n.length?e.concat(n[0]):e.concat(n.filter(Boolean).map((function(e){return s.test(e)?JSON.parse(e.split(o)[1]):e})))}),[])}},9582:(e,t,n)=>{"use strict";n.d(t,{N:()=>u});var r=n(1014),a=n(4476),o=n(2071);const i=a.z.enum(["Meta","Control","Shift","Alt","default"]).optional().default("default"),s=a.z.array(i).optional().default(["default"]),l=a.z.record(a.z.enum(["Space","Backspace","Tab","Enter","Escape","ArrowDown","ArrowLeft","ArrowRight","ArrowUp","End","Home","PageDown","PageUp","Delete","0","1","2","3","4","5","6","7","8","9","a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z","0","1","2","3","4","5","6","7","8","9","*","+","-",".","/","F1","F2","F3","F4","F5","F6","F7","F8","F9","F10","F11","F12",";","=",",","-",".","/","`","[","\\","]",""]),a.z.object({modifiers:s.optional(),event:a.z.enum(["addTopNav","closeModal","closeTopNav","createNew","executeRequest","focusAddressBar","focusRequestSearch","jumpToLastTab","jumpToTab","navigateSearchResultsDown","navigateSearchResultsUp","navigateTopNavLeft","navigateTopNavRight","openCommandPalette","selectSearchResult","toggleSidebar"])})),c=a.z.object({modifiers:s,hotKeys:l.optional()}).optional(),u=a.z.object({uid:o.B.brand(),name:a.z.string().default("Default Workspace"),description:a.z.string().default("Basic Scalar Workspace"),collections:a.z.array(a.z.string().brand()).default([]),environments:a.z.record(a.z.string()).default({}),hotKeyConfig:c,activeEnvironmentId:a.z.string().optional().default("default"),cookies:a.z.array(a.z.string().brand()).default([]),proxyUrl:a.z.string().optional(),themeId:a.z.enum(r.MH).optional().default("default"),selectedHttpClient:a.z.object({targetKey:a.z.string(),clientKey:a.z.string()}).optional().default({targetKey:"shell",clientKey:"curl"})})},9635:(e,t,n)=>{"use strict";function r(e){if(e)throw e}n.d(t,{l:()=>d});var a=n(2849),o=n(7471);function i(e){if("object"!=typeof e||null===e)return!1;const t=Object.getPrototypeOf(e);return!(null!==t&&t!==Object.prototype&&null!==Object.getPrototypeOf(t)||Symbol.toStringTag in e||Symbol.iterator in e)}var s=n(8826);const l=function(e){const t=this.constructor.prototype,n=t[e],r=function(){return n.apply(r,arguments)};return Object.setPrototypeOf(r,t),r},c={}.hasOwnProperty;class u extends l{constructor(){super("copy"),this.Compiler=void 0,this.Parser=void 0,this.attachers=[],this.compiler=void 0,this.freezeIndex=-1,this.frozen=void 0,this.namespace={},this.parser=void 0,this.transformers=function(){const e=[],t={run:function(...t){let n=-1;const r=t.pop();if("function"!=typeof r)throw new TypeError("Expected function as last argument, not "+r);!function a(o,...i){const s=e[++n];let l=-1;if(o)r(o);else{for(;++l<t.length;)null!==i[l]&&void 0!==i[l]||(i[l]=t[l]);t=i,s?function(e,t){let n;return function(...t){const o=e.length>t.length;let i;o&&t.push(r);try{i=e.apply(this,t)}catch(e){if(o&&n)throw e;return r(e)}o||(i&&i.then&&"function"==typeof i.then?i.then(a,r):i instanceof Error?r(i):a(i))};function r(e,...r){n||(n=!0,t(e,...r))}function a(e){r(null,e)}}(s,a)(...i):r(null,...i)}}(null,...t)},use:function(n){if("function"!=typeof n)throw new TypeError("Expected `middelware` to be a function, not "+n);return e.push(n),t}};return t}()}copy(){const e=new u;let t=-1;for(;++t<this.attachers.length;){const n=this.attachers[t];e.use(...n)}return e.data(a(!0,{},this.namespace)),e}data(e,t){return"string"==typeof e?2===arguments.length?(f("data",this.frozen),this.namespace[e]=t,this):c.call(this.namespace,e)&&this.namespace[e]||void 0:e?(f("data",this.frozen),this.namespace=e,this):this.namespace}freeze(){if(this.frozen)return this;const e=this;for(;++this.freezeIndex<this.attachers.length;){const[t,...n]=this.attachers[this.freezeIndex];if(!1===n[0])continue;!0===n[0]&&(n[0]=void 0);const r=t.call(e,...n);"function"==typeof r&&this.transformers.use(r)}return this.frozen=!0,this.freezeIndex=Number.POSITIVE_INFINITY,this}parse(e){this.freeze();const t=v(e),n=this.parser||this.Parser;return p("parse",n),n(String(t),t)}process(e,t){const n=this;return this.freeze(),p("process",this.parser||this.Parser),h("process",this.compiler||this.Compiler),t?r(void 0,t):new Promise(r);function r(r,a){const i=v(e),s=n.parse(i);function l(e,n){e||!n?a(e):r?r(n):((0,o.ok)(t,"`done` is defined if `resolve` is not"),t(void 0,n))}n.run(s,i,(function(e,t,r){if(e||!t||!r)return l(e);const a=t,o=n.stringify(a,r);var i;"string"==typeof(i=o)||function(e){return Boolean(e&&"object"==typeof e&&"byteLength"in e&&"byteOffset"in e)}(i)?r.value=o:r.result=o,l(e,r)}))}}processSync(e){let t,n=!1;return this.freeze(),p("processSync",this.parser||this.Parser),h("processSync",this.compiler||this.Compiler),this.process(e,(function(e,a){n=!0,r(e),t=a})),g("processSync","process",n),(0,o.ok)(t,"we either bailed on an error or have a tree"),t}run(e,t,n){m(e),this.freeze();const r=this.transformers;return n||"function"!=typeof t||(n=t,t=void 0),n?a(void 0,n):new Promise(a);function a(a,i){(0,o.ok)("function"!=typeof t,"`file` can’t be a `done` anymore, we checked");const s=v(t);r.run(e,s,(function(t,r,s){const l=r||e;t?i(t):a?a(l):((0,o.ok)(n,"`done` is defined if `resolve` is not"),n(void 0,l,s))}))}}runSync(e,t){let n,a=!1;return this.run(e,t,(function(e,t){r(e),n=t,a=!0})),g("runSync","run",a),(0,o.ok)(n,"we either bailed on an error or have a tree"),n}stringify(e,t){this.freeze();const n=v(t),r=this.compiler||this.Compiler;return h("stringify",r),m(e),r(e,n)}use(e,...t){const n=this.attachers,r=this.namespace;if(f("use",this.frozen),null==e);else if("function"==typeof e)c(e,t);else{if("object"!=typeof e)throw new TypeError("Expected usable value, not `"+e+"`");Array.isArray(e)?l(e):s(e)}return this;function o(e){if("function"==typeof e)c(e,[]);else{if("object"!=typeof e)throw new TypeError("Expected usable value, not `"+e+"`");if(Array.isArray(e)){const[t,...n]=e;c(t,n)}else s(e)}}function s(e){if(!("plugins"in e)&&!("settings"in e))throw new Error("Expected usable value but received an empty preset, which is probably a mistake: presets typically come with `plugins` and sometimes with `settings`, but this has neither");l(e.plugins),e.settings&&(r.settings=a(!0,r.settings,e.settings))}function l(e){let t=-1;if(null==e);else{if(!Array.isArray(e))throw new TypeError("Expected a list of plugins, not `"+e+"`");for(;++t<e.length;)o(e[t])}}function c(e,t){let r=-1,o=-1;for(;++r<n.length;)if(n[r][0]===e){o=r;break}if(-1===o)n.push([e,...t]);else if(t.length>0){let[r,...s]=t;const l=n[o][1];i(l)&&i(r)&&(r=a(!0,l,r)),n[o]=[e,r,...s]}}}}const d=(new u).freeze();function p(e,t){if("function"!=typeof t)throw new TypeError("Cannot `"+e+"` without `parser`")}function h(e,t){if("function"!=typeof t)throw new TypeError("Cannot `"+e+"` without `compiler`")}function f(e,t){if(t)throw new Error("Cannot call `"+e+"` on a frozen processor.\nCreate a new processor first, by calling it: use `processor()` instead of `processor`.")}function m(e){if(!i(e)||"string"!=typeof e.type)throw new TypeError("Expected node, got `"+e+"`")}function g(e,t,n){if(!n)throw new Error("`"+e+"` finished async. Use `"+t+"` instead")}function v(e){return function(e){return Boolean(e&&"object"==typeof e&&"message"in e&&"messages"in e)}(e)?e:new s.T(e)}},9656:(e,t,n)=>{"use strict";n.d(t,{dc:()=>i,_Z:()=>s,VG:()=>l});var r=n(9143);const a=[],o=!0,i=!1,s="skip";function l(e,t,n,l){let c;"function"==typeof t&&"function"!=typeof n?(l=n,n=t):c=t;const u=(0,r.C)(c),d=l?-1:1;!function e(r,c,p){const h=r&&"object"==typeof r?r:{};if("string"==typeof h.type){const e="string"==typeof h.tagName?h.tagName:"string"==typeof h.name?h.name:void 0;Object.defineProperty(f,"name",{value:"node ("+r.type+(e?"<"+e+">":"")+")"})}return f;function f(){let h,f,m,g=a;if((!t||u(r,c,p[p.length-1]||void 0))&&(g=function(e){return Array.isArray(e)?e:"number"==typeof e?[o,e]:null==e?a:[e]}(n(r,p)),g[0]===i))return g;if("children"in r&&r.children){const t=r;if(t.children&&g[0]!==s)for(f=(l?t.children.length:-1)+d,m=p.concat(t);f>-1&&f<t.children.length;){const n=t.children[f];if(h=e(n,f,m)(),h[0]===i)return h;f="number"==typeof h[1]?h[1]:f+d}}return g}}(e,void 0,[])()}},9698:(e,t)=>{"use strict";var n=Symbol.for("react.transitional.element"),r=Symbol.for("react.fragment");function a(e,t,r){var a=null;if(void 0!==r&&(a=""+r),void 0!==t.key&&(a=""+t.key),"key"in t)for(var o in r={},t)"key"!==o&&(r[o]=t[o]);else r=t;return t=r.ref,{$$typeof:n,type:e,key:a,ref:void 0!==t?t:null,props:r}}t.Fragment=r,t.jsx=a,t.jsxs=a},9742:(e,t,n)=>{"use strict";n.d(t,{A:()=>l});var r=n(641),a=n(33),o=n(953),i=n(1266);const s={key:0,class:"sr-only"},l=(0,r.pM)({__name:"DataTable",props:{columns:{},scroll:{type:Boolean}},setup(e){const{cx:t}=(0,i.AX)();return(e,n)=>((0,r.uX)(),(0,r.CE)("div",(0,a._B)((0,r.Ng)((0,o.R1)(t)(e.scroll?"overflow-x-auto custom-scroll":"overflow-visible","scalar-data-table bg-b-1"))),[(0,r.Lk)("table",{class:(0,a.C4)(["grid auto-rows-auto min-h-8 mb-0",{"min-w-full w-max":e.scroll}]),style:(0,a.Tr)({gridTemplateColumns:e.columns.map((e=>e||"1fr")).join(" ")})},[e.$slots.caption?((0,r.uX)(),(0,r.CE)("caption",s,[(0,r.RG)(e.$slots,"caption")])):(0,r.Q3)("",!0),(0,r.RG)(e.$slots,"default")],6)],16))}})},9747:(e,t,n)=>{"use strict";n.d(t,{a:()=>a,v:()=>r});const r=function(e,t,n,r,o){const i=a(t);if(null!=n&&("number"!=typeof n||n<0||n===Number.POSITIVE_INFINITY))throw new Error("Expected positive finite `index`");if(!(null==r||r.type&&r.children))throw new Error("Expected valid `parent`");if(null==n!=(null==r))throw new Error("Expected both `index` and `parent`");return!!s(e)&&i.call(o,e,n,r)},a=function(e){if(null==e)return i;if("string"==typeof e)return t=e,o((function(e){return e.tagName===t}));var t;if("object"==typeof e)return function(e){const t=[];let n=-1;for(;++n<e.length;)t[n]=a(e[n]);return o((function(...e){let n=-1;for(;++n<t.length;)if(t[n].apply(this,e))return!0;return!1}))}(e);if("function"==typeof e)return o(e);throw new Error("Expected function, string, or array as `test`")};function o(e){return function(t,n,r){return Boolean(s(t)&&e.call(this,t,"number"==typeof n?n:void 0,r||void 0))}}function i(e){return Boolean(e&&"object"==typeof e&&"type"in e&&"element"===e.type&&"tagName"in e&&"string"==typeof e.tagName)}function s(e){return null!==e&&"object"==typeof e&&"type"in e&&"tagName"in e}},9751:(e,t,n)=>{"use strict";n.d(t,{I:()=>r});const r={ts:"typescript",js:"javascript",py:"python",py3:"python","c#":"csharp","c++":"cpp",node:"javascript"}},9794:(e,t,n)=>{"use strict";n.d(t,{A:()=>X});var r=n(641),a=n(953),o=n(3751),i=n(33),s=n(1266),l=n(3833),c=n(8309),u=n(1347),d=n(8157),p=n(6251),h=n(4282),f=n(3016),m=n(9971);const g={class:"flex h-fit w-full"},v={class:"flex-1"},b={class:"grid auto-rows-auto",style:{gridTemplateColumns:"1fr auto"}},y={key:0},w={class:"font-code text-xs"},O=(0,r.pM)({__name:"OAuthScopesInput",props:{flow:{},updateScheme:{type:Function}},setup(e){const t=(0,r.EW)((()=>{var t;return Object.entries((null==(t=e.flow)?void 0:t.scopes)??{}).map((([e,t])=>({id:e,label:e,description:t})))})),n=(0,r.EW)((()=>{var t;return(null==(t=e.flow)?void 0:t.selectedScopes)||[]}));return(o,l)=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(h.A),{class:"items-center min-h-8 h-auto !max-h-[initial]"},{default:(0,r.k6)((()=>[(0,r.Lk)("div",g,[l[0]||(l[0]=(0,r.Lk)("div",{class:"text-c-1 items-center h-full"},null,-1)),(0,r.bF)((0,a.R1)(p.EN),{as:"div",class:"flex flex-col w-full bl"},{default:(0,r.k6)((()=>{var l,c;return[(0,r.bF)((0,a.R1)(p._q),{class:(0,i.C4)(["group/scopes-accordion flex items-center text-left min-h-8 gap-1.5 h-auto pl-3 pr-2 hover:text-c-1 cursor-pointer",((null==(c=null==(l=o.flow)?void 0:l.selectedScopes)?void 0:c.length)||0)>0?"text-c-1":"text-c-3"])},{default:(0,r.k6)((({open:e})=>{var t,n,l;return[(0,r.Lk)("div",v," Scopes Selected "+(0,i.v_)((null==(n=null==(t=o.flow)?void 0:t.selectedScopes)?void 0:n.length)||0)+" / "+(0,i.v_)(Object.keys((null==(l=o.flow)?void 0:l.scopes)??{}).length||0),1),(0,r.bF)((0,a.R1)(s.kv),{class:"text-c-3 group-hover/scopes-accordion:text-c-2",icon:e?"ChevronDown":"ChevronRight",size:"sm"},null,8,["icon"])]})),_:1},8,["class"]),(0,r.bF)((0,a.R1)(p.kS),{as:"template"},{default:(0,r.k6)((()=>[(0,r.Lk)("table",b,[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(t.value,(({id:t,label:o,description:s})=>((0,r.uX)(),(0,r.Wv)((0,a.R1)(f.A),{key:t,class:"text-c-2",onClick:r=>function(t,r){r?e.updateScheme(`flows.${e.flow.type}.selectedScopes`,[...n.value,t]):e.updateScheme(`flows.${e.flow.type}.selectedScopes`,n.value.filter((e=>e!==t)))}(t,!n.value.includes(t))},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(h.A),{class:"w-full px-3 py-1.5 hover:text-c-1 cursor-pointer !max-h-[initial]"},{default:(0,r.k6)((()=>[(0,r.Lk)("span",null,[s?((0,r.uX)(),(0,r.CE)("span",y,[(0,r.Lk)("span",w,(0,i.v_)(o),1),(0,r.eW)(" – "+(0,i.v_)(s),1)])):(0,r.Q3)("",!0)])])),_:2},1024),(0,r.bF)((0,a.R1)(m.A),{modelValue:n.value.includes(t),"onUpdate:modelValue":()=>{}},null,8,["modelValue"])])),_:2},1032,["onClick"])))),128))])])),_:1})]})),_:1})])])),_:1}))}});var k=n(1504);const x=["for"],E=(0,r.pM)({__name:"RequestAuthDataTableInput",props:{type:{},containerClass:{},required:{type:Boolean,default:!1},modelValue:{},readOnly:{type:Boolean,default:!1},environment:{},envVariables:{},workspace:{}},emits:["update:modelValue","inputFocus","inputBlur","selectVariable"],setup(e,{emit:t}){const n=e,o=t,i=(0,r.Bi)();return(e,t)=>((0,r.uX)(),(0,r.Wv)(k.A,(0,r.v6)({id:(0,a.R1)(i),canAddCustomEnumValue:!n.readOnly,containerClass:n.containerClass},e.$attrs,{envVariables:n.envVariables,environment:n.environment,modelValue:n.modelValue,readOnly:n.readOnly,required:n.required,type:n.type,workspace:n.workspace,onInputBlur:t[0]||(t[0]=e=>o("inputBlur")),onInputFocus:t[1]||(t[1]=e=>o("inputFocus")),onSelectVariable:t[2]||(t[2]=e=>o("selectVariable",e)),"onUpdate:modelValue":t[3]||(t[3]=e=>o("update:modelValue",e))}),{default:(0,r.k6)((()=>[(0,r.Lk)("label",{for:(0,a.R1)(i)},[(0,r.RG)(e.$slots,"default")],8,x)])),icon:(0,r.k6)((()=>[(0,r.RG)(e.$slots,"icon")])),_:3},16,["id","canAddCustomEnumValue","containerClass","envVariables","environment","modelValue","readOnly","required","type","workspace"]))}});var S=n(4048);const _=async(e,t,{code:n,pkce:r,proxyUrl:a}={})=>{if(!e)return[new Error("OAuth2 flow was not defined"),null];const o=new URLSearchParams;o.set("client_id",e["x-scalar-client-id"]),t&&("clientCredentials"===e.type||"password"===e.type)&&o.set("scope",t),e.clientSecret&&o.set("client_secret",e.clientSecret),"x-scalar-redirect-uri"in e&&e["x-scalar-redirect-uri"]&&o.set("redirect_uri",e["x-scalar-redirect-uri"]),n?(o.set("code",n),o.set("grant_type","authorization_code"),r&&o.set("code_verifier",r.codeVerifier)):"password"===e.type?(o.set("grant_type","password"),o.set("username",e.username),o.set("password",e.password)):o.set("grant_type","client_credentials");try{const t={"Content-Type":"application/x-www-form-urlencoded"};t.Authorization=`Basic ${btoa(`${e["x-scalar-client-id"]}:${e.clientSecret}`)}`;const n=(0,S.p3)(a,e.tokenUrl)?`${a}?${new URLSearchParams([["scalar_url",e.tokenUrl]]).toString()}`:e.tokenUrl,r=await fetch(n,{method:"POST",headers:t,body:o}),{access_token:i}=await r.json();return[null,i]}catch{return[new Error("Failed to get an access token. Please check your credentials."),null]}},C={class:"h-8 flex items-center justify-self-end"},T={class:"h-8 flex items-center justify-end border-t-1/2 w-full"},A=(0,r.pM)({__name:"OAuth2",props:{collection:{},environment:{},envVariables:{},flow:{},scheme:{},server:{},workspace:{}},setup(e){const t=(0,s.R2)(),{toast:n}=(0,d.Yf)(),{securitySchemeMutators:o}=(0,c.hA)(),i=(t,n)=>o.edit(e.scheme.uid,t,n),l=async()=>{var r,a;if(t.isLoading||null==(r=e.collection)||!r.uid)return;if(t.startLoading(),!e.server)return void n("No server selected","error");const[o,s]=await(async(e,t,n)=>{try{if(!e)return[new Error("Flow not found"),null];const r=e.selectedScopes.join(" ");if("clientCredentials"===e.type||"password"===e.type)return _(e,r,{proxyUrl:n});const a=(Math.random()+1).toString(36).substring(2,10),o=new URL(e.authorizationUrl);let i=null;if("implicit"===e.type)o.searchParams.set("response_type","token");else if("authorizationCode"===e.type&&(o.searchParams.set("response_type","code"),"no"!==e["x-usePkce"])){const t=(()=>{const e=new Uint8Array(32);return crypto.getRandomValues(e),btoa(String.fromCharCode(...e)).replace(/\+/g,"-").replace(/\//g,"_").replace(/=/g,"")})(),n=await(async(e,t)=>{if("plain"===t)return e;const n=(new TextEncoder).encode(e),r=await crypto.subtle.digest("SHA-256",n);return btoa(String.fromCharCode(...new Uint8Array(r))).replace(/\+/g,"-").replace(/\//g,"_").replace(/=+$/,"")})(t,e["x-usePkce"]);i={codeVerifier:t,codeChallenge:n,codeChallengeMethod:"SHA-256"===e["x-usePkce"]?"S256":"plain"},o.searchParams.set("code_challenge",n),o.searchParams.set("code_challenge_method",i.codeChallengeMethod)}if(e["x-scalar-redirect-uri"].startsWith("/")){const n=t.url||window.location.origin+window.location.pathname,r=new URL(e["x-scalar-redirect-uri"],n).toString();o.searchParams.set("redirect_uri",r)}else o.searchParams.set("redirect_uri",e["x-scalar-redirect-uri"]);o.searchParams.set("client_id",e["x-scalar-client-id"]),o.searchParams.set("state",a),r&&o.searchParams.set("scope",r);const s=window.open(o,"openAuth2Window","left=100,top=100,width=800,height=600");return s?new Promise((t=>{const o=setInterval((()=>{var l;let c=null,u=null;try{const e=new URL(s.location.href).searchParams;c=e.get("access_token"),u=e.get("code");const t=new URLSearchParams(s.location.href.split("#")[1]);c||(c=t.get("access_token")),u||(u=t.get("code"))}catch{}if(s.closed||c||u)if(clearInterval(o),s.close(),c){const e=null==(l=s.location.href.match(/state=([^&]*)/))?void 0:l[1];t(e===a?[null,c]:[new Error("State mismatch"),null])}else u?new URL(s.location.href).searchParams.get("state")===a?_(e,r,{code:u,pkce:i,proxyUrl:n}).then(t):t([new Error("State mismatch"),null]):(clearInterval(o),t([new Error("Window was closed without granting authorization"),null]))}),200)})):[new Error("Failed to open auth window"),null]}catch{return[new Error("Failed to authorize oauth2 flow"),null]}})(e.flow,e.server,null==(a=e.workspace)?void 0:a.proxyUrl).finally((()=>t.stopLoading()));s?i(`flows.${e.flow.type}.token`,s):(console.error(o),n((null==o?void 0:o.message)??"Failed to authorize","error"))},p={environment:e.environment,envVariables:e.envVariables,workspace:e.workspace};return(e,n)=>((0,r.uX)(),(0,r.CE)(r.FK,null,[e.flow.token?((0,r.uX)(),(0,r.CE)(r.FK,{key:0},[(0,r.bF)((0,a.R1)(f.A),null,{default:(0,r.k6)((()=>[(0,r.bF)(E,(0,r.v6)(p,{class:"border-r-transparent",modelValue:e.flow.token,placeholder:"QUxMIFlPVVIgQkFTRSBBUkUgQkVMT05HIFRPIFVT",type:"password","onUpdate:modelValue":n[0]||(n[0]=t=>i(`flows.${e.flow.type}.token`,t))}),{default:(0,r.k6)((()=>n[10]||(n[10]=[(0,r.eW)(" Access Token ")]))),_:1},16,["modelValue"])])),_:1}),(0,r.bF)((0,a.R1)(f.A),{class:"min-w-full"},{default:(0,r.k6)((()=>[(0,r.Lk)("div",C,[(0,r.bF)((0,a.R1)(s.RO),{class:"p-0 py-0.5 px-2 mr-1",loading:(0,a.R1)(t),size:"sm",variant:"outlined",onClick:n[1]||(n[1]=t=>i(`flows.${e.flow.type}.token`,""))},{default:(0,r.k6)((()=>n[11]||(n[11]=[(0,r.eW)(" Clear ")]))),_:1},8,["loading"])])])),_:1})],64)):((0,r.uX)(),(0,r.CE)(r.FK,{key:1},[(0,r.bF)((0,a.R1)(f.A),null,{default:(0,r.k6)((()=>["authorizationUrl"in e.flow?((0,r.uX)(),(0,r.Wv)(E,(0,r.v6)({key:0},p,{modelValue:e.flow.authorizationUrl,placeholder:"https://galaxy.scalar.com/authorize","onUpdate:modelValue":n[2]||(n[2]=t=>i(`flows.${e.flow.type}.authorizationUrl`,t))}),{default:(0,r.k6)((()=>n[12]||(n[12]=[(0,r.eW)(" Auth URL ")]))),_:1},16,["modelValue"])):(0,r.Q3)("",!0),"tokenUrl"in e.flow?((0,r.uX)(),(0,r.Wv)(E,(0,r.v6)({key:1},p,{modelValue:e.flow.tokenUrl,placeholder:"https://galaxy.scalar.com/token","onUpdate:modelValue":n[3]||(n[3]=t=>i(`flows.${e.flow.type}.tokenUrl`,t))}),{default:(0,r.k6)((()=>n[13]||(n[13]=[(0,r.eW)(" Token URL ")]))),_:1},16,["modelValue"])):(0,r.Q3)("",!0)])),_:1}),"x-scalar-redirect-uri"in e.flow?((0,r.uX)(),(0,r.Wv)((0,a.R1)(f.A),{key:0},{default:(0,r.k6)((()=>[(0,r.bF)(E,(0,r.v6)(p,{modelValue:e.flow["x-scalar-redirect-uri"],placeholder:"https://galaxy.scalar.com/callback","onUpdate:modelValue":n[4]||(n[4]=t=>i(`flows.${e.flow.type}.x-scalar-redirect-uri`,t))}),{default:(0,r.k6)((()=>n[14]||(n[14]=[(0,r.eW)(" Redirect URL ")]))),_:1},16,["modelValue"])])),_:1})):(0,r.Q3)("",!0),"password"===e.flow.type?((0,r.uX)(),(0,r.CE)(r.FK,{key:1},[(0,r.bF)((0,a.R1)(f.A),null,{default:(0,r.k6)((()=>[(0,r.bF)(E,(0,r.v6)(p,{class:"text-c-2",modelValue:e.flow.username,placeholder:"janedoe","onUpdate:modelValue":n[5]||(n[5]=t=>i(`flows.${e.flow.type}.username`,t))}),{default:(0,r.k6)((()=>n[15]||(n[15]=[(0,r.eW)(" Username ")]))),_:1},16,["modelValue"])])),_:1}),(0,r.bF)((0,a.R1)(f.A),null,{default:(0,r.k6)((()=>[(0,r.bF)(E,(0,r.v6)(p,{modelValue:e.flow.password,placeholder:"********",type:"password","onUpdate:modelValue":n[6]||(n[6]=t=>i(`flows.${e.flow.type}.password`,t))}),{default:(0,r.k6)((()=>n[16]||(n[16]=[(0,r.eW)(" Password ")]))),_:1},16,["modelValue"])])),_:1})],64)):(0,r.Q3)("",!0),(0,r.bF)((0,a.R1)(f.A),null,{default:(0,r.k6)((()=>[(0,r.bF)(E,(0,r.v6)(p,{modelValue:e.flow["x-scalar-client-id"],placeholder:"12345","onUpdate:modelValue":n[7]||(n[7]=t=>i(`flows.${e.flow.type}.x-scalar-client-id`,t))}),{default:(0,r.k6)((()=>n[17]||(n[17]=[(0,r.eW)(" Client ID ")]))),_:1},16,["modelValue"])])),_:1}),"clientSecret"in e.flow?((0,r.uX)(),(0,r.Wv)((0,a.R1)(f.A),{key:2},{default:(0,r.k6)((()=>[(0,r.bF)(E,(0,r.v6)(p,{modelValue:e.flow.clientSecret,placeholder:"XYZ123",type:"password","onUpdate:modelValue":n[8]||(n[8]=t=>i(`flows.${e.flow.type}.clientSecret`,t))}),{default:(0,r.k6)((()=>n[18]||(n[18]=[(0,r.eW)(" Client Secret ")]))),_:1},16,["modelValue"])])),_:1})):(0,r.Q3)("",!0),"x-usePkce"in e.flow?((0,r.uX)(),(0,r.Wv)((0,a.R1)(f.A),{key:3},{default:(0,r.k6)((()=>[(0,r.bF)(E,(0,r.v6)(p,{enum:(0,a.R1)(u.I4),modelValue:e.flow["x-usePkce"],readOnly:"","onUpdate:modelValue":n[9]||(n[9]=t=>i(`flows.${e.flow.type}.x-usePkce`,t))}),{default:(0,r.k6)((()=>n[19]||(n[19]=[(0,r.eW)(" Use PKCE ")]))),_:1},16,["enum","modelValue"])])),_:1})):(0,r.Q3)("",!0),Object.keys(e.flow.scopes??{}).length?((0,r.uX)(),(0,r.Wv)((0,a.R1)(f.A),{key:4},{default:(0,r.k6)((()=>[(0,r.bF)(O,{flow:e.flow,updateScheme:i},null,8,["flow"])])),_:1})):(0,r.Q3)("",!0)],64)),e.flow.token?(0,r.Q3)("",!0):((0,r.uX)(),(0,r.Wv)((0,a.R1)(f.A),{key:2,class:"min-w-full"},{default:(0,r.k6)((()=>[(0,r.Lk)("div",T,[(0,r.bF)((0,a.R1)(s.RO),{class:"p-0 py-0.5 px-2 mr-1",loading:(0,a.R1)(t),size:"sm",variant:"outlined",onClick:l},{default:(0,r.k6)((()=>n[20]||(n[20]=[(0,r.eW)(" Authorize ")]))),_:1},8,["loading"])])])),_:1}))],64))}}),R={key:0,class:"flex min-h-8 border-t text-sm"},P={class:"flex h-8 max-w-full gap-2.5 overflow-x-auto px-3"},D=["onClick"],M={class:"relative z-10"},$={key:4,class:"text-c-3 bg-b-1 flex min-h-16 items-center justify-center border-t px-4 text-sm"},N=(0,r.pM)({__name:"RequestAuthTab",props:{collection:{},environment:{},envVariables:{},layout:{},securitySchemeUids:{},server:{},workspace:{}},setup(e){const{securitySchemes:t,securitySchemeMutators:n}=(0,c.hA)(),o=(0,r.EW)((()=>e.securitySchemeUids.map((e=>({scheme:t[e]}))))),s=(0,a.KR)(""),l=e=>{if("apiKey"===e.type)return`${(0,i.ZH)(e.nameKey)}: ${e.in}`;if("oauth2"===e.type){const t=Object.values(e.flows??{})[0];return`${(0,i.ZH)(e.nameKey)}: ${s.value?s.value:(null==t?void 0:t.type)??""}`}return"http"===e.type?`${(0,i.ZH)(e.nameKey)}: ${e.scheme}`:`${(0,i.ZH)(e.nameKey)}: ${e.type}`},u=(e,t,r)=>n.edit(e,t,r),d={environment:e.environment,envVariables:e.envVariables,workspace:e.workspace};return(e,t)=>((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(o.value,(({scheme:n},c)=>((0,r.uX)(),(0,r.CE)(r.FK,{key:null==n?void 0:n.uid},[o.value.length>1?((0,r.uX)(),(0,r.Wv)((0,a.R1)(f.A),{key:0,class:(0,i.C4)({"request-example-references-header":"reference"===e.layout})},{default:(0,r.k6)((()=>[(0,r.bF)((0,a.R1)(h.A),{class:(0,i.C4)(["text-c-3 flex items-center pl-2 font-medium","reference"===e.layout&&"border-t "+(0!==c?"mt-2":"")])},{default:(0,r.k6)((()=>[(0,r.eW)((0,i.v_)(l(n)),1)])),_:2},1032,["class"])])),_:2},1032,["class"])):(0,r.Q3)("",!0),"http"===(null==n?void 0:n.type)?((0,r.uX)(),(0,r.CE)(r.FK,{key:1},["bearer"===n.scheme?((0,r.uX)(),(0,r.Wv)((0,a.R1)(f.A),{key:0},{default:(0,r.k6)((()=>[(0,r.bF)(E,(0,r.v6)({ref_for:!0},d,{containerClass:"reference"===e.layout&&"border-t",modelValue:n.token,placeholder:"Token",type:"password","onUpdate:modelValue":e=>u(n.uid,"token",e)}),{default:(0,r.k6)((()=>t[0]||(t[0]=[(0,r.eW)(" Bearer Token ")]))),_:2},1040,["containerClass","modelValue","onUpdate:modelValue"])])),_:2},1024)):"basic"===(null==n?void 0:n.scheme)?((0,r.uX)(),(0,r.CE)(r.FK,{key:1},[(0,r.bF)((0,a.R1)(f.A),null,{default:(0,r.k6)((()=>[(0,r.bF)(E,(0,r.v6)({ref_for:!0},d,{class:"text-c-2",containerClass:"reference"===e.layout&&"auth-blend-required border-t",modelValue:n.username,placeholder:"janedoe",required:"","onUpdate:modelValue":e=>u(n.uid,"username",e)}),{default:(0,r.k6)((()=>t[1]||(t[1]=[(0,r.eW)(" Username ")]))),_:2},1040,["containerClass","modelValue","onUpdate:modelValue"])])),_:2},1024),(0,r.bF)((0,a.R1)(f.A),null,{default:(0,r.k6)((()=>[(0,r.bF)(E,(0,r.v6)({ref_for:!0},d,{modelValue:n.password,placeholder:"********",type:"password","onUpdate:modelValue":e=>u(n.uid,"password",e)}),{default:(0,r.k6)((()=>t[2]||(t[2]=[(0,r.eW)(" Password ")]))),_:2},1040,["modelValue","onUpdate:modelValue"])])),_:2},1024)],64)):(0,r.Q3)("",!0)],64)):"apiKey"===(null==n?void 0:n.type)?((0,r.uX)(),(0,r.CE)(r.FK,{key:2},[(0,r.bF)((0,a.R1)(f.A),null,{default:(0,r.k6)((()=>[(0,r.bF)(E,(0,r.v6)({ref_for:!0},d,{containerClass:"reference"===e.layout&&"border-t",modelValue:n.name,placeholder:"api-key","onUpdate:modelValue":e=>u(n.uid,"name",e)}),{default:(0,r.k6)((()=>t[3]||(t[3]=[(0,r.eW)(" Name ")]))),_:2},1040,["containerClass","modelValue","onUpdate:modelValue"])])),_:2},1024),(0,r.bF)((0,a.R1)(f.A),null,{default:(0,r.k6)((()=>[(0,r.bF)(E,(0,r.v6)({ref_for:!0},d,{modelValue:n.value,placeholder:"QUxMIFlPVVIgQkFTRSBBUkUgQkVMT05HIFRPIFVT","onUpdate:modelValue":e=>u(n.uid,"value",e)}),{default:(0,r.k6)((()=>t[4]||(t[4]=[(0,r.eW)(" Value ")]))),_:2},1040,["modelValue","onUpdate:modelValue"])])),_:2},1024)],64)):"oauth2"===(null==n?void 0:n.type)?((0,r.uX)(),(0,r.CE)(r.FK,{key:3},[(0,r.bF)((0,a.R1)(f.A),null,{default:(0,r.k6)((()=>[Object.keys(n.flows).length>1?((0,r.uX)(),(0,r.CE)("div",R,[(0,r.Lk)("div",P,[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(null==n?void 0:n.flows,((t,n,a)=>((0,r.uX)(),(0,r.CE)("button",{key:n,class:(0,i.C4)(["floating-bg text-c-3 relative cursor-pointer border-b-[1px] border-transparent py-1 text-sm font-medium",{"!text-c-1 !rounded-none border-b-[1px] !border-current":"reference"!==e.layout&&(s.value===n||0===a&&!s.value),"!text-c-1 !rounded-none border-b-[1px] !border-current opacity-100":"reference"===e.layout&&(s.value===n||0===a&&!s.value)}]),type:"button",onClick:e=>s.value=n},[(0,r.Lk)("span",M,(0,i.v_)(n),1)],10,D)))),128))])])):(0,r.Q3)("",!0)])),_:2},1024),((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(null==n?void 0:n.flows,((t,a,o)=>((0,r.uX)(),(0,r.CE)(r.FK,{key:a},[s.value===a||0===o&&!s.value?((0,r.uX)(),(0,r.Wv)(A,(0,r.v6)({key:0,ref_for:!0},d,{collection:e.collection,flow:t,scheme:n,server:e.server,workspace:e.workspace}),null,16,["collection","flow","scheme","server","workspace"])):(0,r.Q3)("",!0)],64)))),128))],64)):"openIdConnect"===(null==n?void 0:n.type)?((0,r.uX)(),(0,r.CE)("div",$," Coming soon ")):(0,r.Q3)("",!0)],64)))),128))}});var I=n(3526);const L=(0,I.A)(N,[["__scopeId","data-v-2b678dcb"]]);var B=n(9742);const F={key:0,class:"flex flex-wrap gap-x-2.5 overflow-hidden border-t px-3"},Q=["onClick"],U={class:"relative z-10 whitespace-nowrap font-medium"},z={key:0,class:"z-1 absolute inset-x-1 bottom-[var(--scalar-border-width)] left-1/2 h-px w-full -translate-x-1/2 bg-current"},j={key:2,class:"text-c-3 bg-b-1 flex min-h-16 items-center justify-center border-t px-4 text-sm"},W=(0,r.pM)({__name:"RequestAuthDataTable",props:{collection:{},environment:{},envVariables:{},layout:{default:"client"},selectedSchemeOptions:{default:()=>[]},server:{},workspace:{}},setup(e){const t=(0,s.hS)(),n=(0,a.KR)(null),c=(0,a.KR)(0),u=(0,r.EW)((()=>{const t=e.selectedSchemeOptions[c.value];if(!t)return[];const n=null==t?void 0:t.id.split(",");return n.length>1?n:[t.id]}));return(0,r.wB)((()=>e.selectedSchemeOptions),(e=>{e[c.value]||(c.value=Math.max(0,c.value-1))})),(e,s)=>((0,r.uX)(),(0,r.CE)("form",{onSubmit:s[1]||(s[1]=(0,o.D$)((()=>{}),["prevent"]))},[e.selectedSchemeOptions.length>1?((0,r.uX)(),(0,r.CE)("div",F,[((0,r.uX)(!0),(0,r.CE)(r.FK,null,(0,r.pI)(e.selectedSchemeOptions,((e,t)=>((0,r.uX)(),(0,r.CE)("div",{key:e.id,class:(0,i.C4)(["z-1 relative -mb-[var(--scalar-border-width)] flex h-8 cursor-pointer",[c.value===t?"text-c-1":"text-c-3"]])},[(0,r.Lk)("button",{class:"floating-bg relative cursor-pointer border-b-[1px] border-transparent py-1 text-sm font-medium",type:"button",onClick:e=>c.value=t},[(0,r.Lk)("span",U,(0,i.v_)(e.label),1)],8,Q),s[2]||(s[2]=(0,r.Lk)("div",{class:"bg-border absolute -inset-x-96 bottom-0 z-0 h-[var(--scalar-border-width)]"},null,-1)),c.value===t?((0,r.uX)(),(0,r.CE)("div",z)):(0,r.Q3)("",!0)],2)))),128))])):(0,r.Q3)("",!0),u.value.length?((0,r.uX)(),(0,r.Wv)((0,a.R1)(B.A),{key:1,class:(0,i.C4)(["flex-1","reference"===e.layout&&"border-0"]),columns:[""]},{default:(0,r.k6)((()=>[(0,r.bF)(L,{collection:e.collection,envVariables:e.envVariables,environment:e.environment,layout:e.layout,securitySchemeUids:u.value,server:e.server,workspace:e.workspace},null,8,["collection","envVariables","environment","layout","securitySchemeUids","server","workspace"])])),_:1},8,["class"])):((0,r.uX)(),(0,r.CE)("div",j," No authentication selected ")),(0,r.bF)(l.A,{scheme:n.value,state:(0,a.R1)(t),onClose:s[0]||(s[0]=e=>(0,a.R1)(t).hide())},null,8,["scheme","state"])],32))}}),X=(0,I.A)(W,[["__scopeId","data-v-3569c0b2"]])},9869:(e,t)=>{"use strict";var n=Symbol.for("react.transitional.element"),r=Symbol.for("react.portal"),a=Symbol.for("react.fragment"),o=Symbol.for("react.strict_mode"),i=Symbol.for("react.profiler"),s=Symbol.for("react.consumer"),l=Symbol.for("react.context"),c=Symbol.for("react.forward_ref"),u=Symbol.for("react.suspense"),d=Symbol.for("react.memo"),p=Symbol.for("react.lazy"),h=Symbol.iterator,f={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},m=Object.assign,g={};function v(e,t,n){this.props=e,this.context=t,this.refs=g,this.updater=n||f}function b(){}function y(e,t,n){this.props=e,this.context=t,this.refs=g,this.updater=n||f}v.prototype.isReactComponent={},v.prototype.setState=function(e,t){if("object"!=typeof e&&"function"!=typeof e&&null!=e)throw Error("takes an object of state variables to update or a function which returns an object of state variables.");this.updater.enqueueSetState(this,e,t,"setState")},v.prototype.forceUpdate=function(e){this.updater.enqueueForceUpdate(this,e,"forceUpdate")},b.prototype=v.prototype;var w=y.prototype=new b;w.constructor=y,m(w,v.prototype),w.isPureReactComponent=!0;var O=Array.isArray,k={H:null,A:null,T:null,S:null},x=Object.prototype.hasOwnProperty;function E(e,t,r,a,o,i){return r=i.ref,{$$typeof:n,type:e,key:t,ref:void 0!==r?r:null,props:i}}function S(e){return"object"==typeof e&&null!==e&&e.$$typeof===n}var _=/\/+/g;function C(e,t){return"object"==typeof e&&null!==e&&null!=e.key?(n=""+e.key,r={"=":"=0",":":"=2"},"$"+n.replace(/[=:]/g,(function(e){return r[e]}))):t.toString(36);var n,r}function T(){}function A(e,t,a,o,i){var s=typeof e;"undefined"!==s&&"boolean"!==s||(e=null);var l,c,u=!1;if(null===e)u=!0;else switch(s){case"bigint":case"string":case"number":u=!0;break;case"object":switch(e.$$typeof){case n:case r:u=!0;break;case p:return A((u=e._init)(e._payload),t,a,o,i)}}if(u)return i=i(e),u=""===o?"."+C(e,0):o,O(i)?(a="",null!=u&&(a=u.replace(_,"$&/")+"/"),A(i,t,a,"",(function(e){return e}))):null!=i&&(S(i)&&(l=i,c=a+(null==i.key||e&&e.key===i.key?"":(""+i.key).replace(_,"$&/")+"/")+u,i=E(l.type,c,void 0,0,0,l.props)),t.push(i)),1;u=0;var d,f=""===o?".":o+":";if(O(e))for(var m=0;m<e.length;m++)u+=A(o=e[m],t,a,s=f+C(o,m),i);else if("function"==typeof(m=null===(d=e)||"object"!=typeof d?null:"function"==typeof(d=h&&d[h]||d["@@iterator"])?d:null))for(e=m.call(e),m=0;!(o=e.next()).done;)u+=A(o=o.value,t,a,s=f+C(o,m++),i);else if("object"===s){if("function"==typeof e.then)return A(function(e){switch(e.status){case"fulfilled":return e.value;case"rejected":throw e.reason;default:switch("string"==typeof e.status?e.then(T,T):(e.status="pending",e.then((function(t){"pending"===e.status&&(e.status="fulfilled",e.value=t)}),(function(t){"pending"===e.status&&(e.status="rejected",e.reason=t)}))),e.status){case"fulfilled":return e.value;case"rejected":throw e.reason}}throw e}(e),t,a,o,i);throw t=String(e),Error("Objects are not valid as a React child (found: "+("[object Object]"===t?"object with keys {"+Object.keys(e).join(", ")+"}":t)+"). If you meant to render a collection of children, use an array instead.")}return u}function R(e,t,n){if(null==e)return e;var r=[],a=0;return A(e,r,"","",(function(e){return t.call(n,e,a++)})),r}function P(e){if(-1===e._status){var t=e._result;(t=t()).then((function(t){0!==e._status&&-1!==e._status||(e._status=1,e._result=t)}),(function(t){0!==e._status&&-1!==e._status||(e._status=2,e._result=t)})),-1===e._status&&(e._status=0,e._result=t)}if(1===e._status)return e._result.default;throw e._result}var D="function"==typeof reportError?reportError:function(e){if("object"==typeof window&&"function"==typeof window.ErrorEvent){var t=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:"object"==typeof e&&null!==e&&"string"==typeof e.message?String(e.message):String(e),error:e});if(!window.dispatchEvent(t))return}else if("object"==typeof process&&"function"==typeof process.emit)return void process.emit("uncaughtException",e);console.error(e)};function M(){}t.Children={map:R,forEach:function(e,t,n){R(e,(function(){t.apply(this,arguments)}),n)},count:function(e){var t=0;return R(e,(function(){t++})),t},toArray:function(e){return R(e,(function(e){return e}))||[]},only:function(e){if(!S(e))throw Error("React.Children.only expected to receive a single React element child.");return e}},t.Component=v,t.Fragment=a,t.Profiler=i,t.PureComponent=y,t.StrictMode=o,t.Suspense=u,t.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=k,t.act=function(){throw Error("act(...) is not supported in production builds of React.")},t.cache=function(e){return function(){return e.apply(null,arguments)}},t.cloneElement=function(e,t,n){if(null==e)throw Error("The argument must be a React element, but you passed "+e+".");var r=m({},e.props),a=e.key;if(null!=t)for(o in t.ref,void 0!==t.key&&(a=""+t.key),t)!x.call(t,o)||"key"===o||"__self"===o||"__source"===o||"ref"===o&&void 0===t.ref||(r[o]=t[o]);var o=arguments.length-2;if(1===o)r.children=n;else if(1<o){for(var i=Array(o),s=0;s<o;s++)i[s]=arguments[s+2];r.children=i}return E(e.type,a,void 0,0,0,r)},t.createContext=function(e){return(e={$$typeof:l,_currentValue:e,_currentValue2:e,_threadCount:0,Provider:null,Consumer:null}).Provider=e,e.Consumer={$$typeof:s,_context:e},e},t.createElement=function(e,t,n){var r,a={},o=null;if(null!=t)for(r in void 0!==t.key&&(o=""+t.key),t)x.call(t,r)&&"key"!==r&&"__self"!==r&&"__source"!==r&&(a[r]=t[r]);var i=arguments.length-2;if(1===i)a.children=n;else if(1<i){for(var s=Array(i),l=0;l<i;l++)s[l]=arguments[l+2];a.children=s}if(e&&e.defaultProps)for(r in i=e.defaultProps)void 0===a[r]&&(a[r]=i[r]);return E(e,o,void 0,0,0,a)},t.createRef=function(){return{current:null}},t.forwardRef=function(e){return{$$typeof:c,render:e}},t.isValidElement=S,t.lazy=function(e){return{$$typeof:p,_payload:{_status:-1,_result:e},_init:P}},t.memo=function(e,t){return{$$typeof:d,type:e,compare:void 0===t?null:t}},t.startTransition=function(e){var t=k.T,n={};k.T=n;try{var r=e(),a=k.S;null!==a&&a(n,r),"object"==typeof r&&null!==r&&"function"==typeof r.then&&r.then(M,D)}catch(e){D(e)}finally{k.T=t}},t.unstable_useCacheRefresh=function(){return k.H.useCacheRefresh()},t.use=function(e){return k.H.use(e)},t.useActionState=function(e,t,n){return k.H.useActionState(e,t,n)},t.useCallback=function(e,t){return k.H.useCallback(e,t)},t.useContext=function(e){return k.H.useContext(e)},t.useDebugValue=function(){},t.useDeferredValue=function(e,t){return k.H.useDeferredValue(e,t)},t.useEffect=function(e,t){return k.H.useEffect(e,t)},t.useId=function(){return k.H.useId()},t.useImperativeHandle=function(e,t,n){return k.H.useImperativeHandle(e,t,n)},t.useInsertionEffect=function(e,t){return k.H.useInsertionEffect(e,t)},t.useLayoutEffect=function(e,t){return k.H.useLayoutEffect(e,t)},t.useMemo=function(e,t){return k.H.useMemo(e,t)},t.useOptimistic=function(e,t){return k.H.useOptimistic(e,t)},t.useReducer=function(e,t,n){return k.H.useReducer(e,t,n)},t.useRef=function(e){return k.H.useRef(e)},t.useState=function(e){return k.H.useState(e)},t.useSyncExternalStore=function(e,t,n){return k.H.useSyncExternalStore(e,t,n)},t.useTransition=function(){return k.H.useTransition()},t.version="19.0.0"},9938:(e,t,n)=>{"use strict";n.d(t,{D:()=>w});var r=n(7471),a=n(7862),o=n(2282);const i=/[#.]/g;var s=n(1775),l=n(7457),c=n(2485);function u(e,t,n){const r=n?function(e){const t=new Map;for(const n of e)t.set(n.toLowerCase(),n);return t}(n):void 0;return function(n,a,...o){let s;if(null==n){s={type:"root",children:[]};const e=a;o.unshift(e)}else{s=function(e,t){const n=e||"",r={};let a,o,s=0;for(;s<n.length;){i.lastIndex=s;const e=i.exec(n),t=n.slice(s,e?e.index:n.length);t&&(a?"#"===a?r.id=t:Array.isArray(r.className)?r.className.push(t):r.className=[t]:o=t,s+=t.length),e&&(a=e[0],s++)}return{type:"element",tagName:o||t||"div",properties:r,children:[]}}(n,t);const l=s.tagName.toLowerCase(),c=r?r.get(l):void 0;if(s.tagName=c||l,function(e){if(null===e||"object"!=typeof e||Array.isArray(e))return!0;if("string"!=typeof e.type)return!1;const t=e,n=Object.keys(e);for(const e of n){const n=t[e];if(n&&"object"==typeof n){if(!Array.isArray(n))return!0;const e=n;for(const t of e)if("number"!=typeof t&&"string"!=typeof t)return!0}}return!(!("children"in e)||!Array.isArray(e.children))}(a))o.unshift(a);else for(const[t,n]of Object.entries(a))d(e,s.properties,t,n)}for(const e of o)p(s.children,e);return"element"===s.type&&"template"===s.tagName&&(s.content={type:"root",children:s.children},s.children=[]),s}}function d(e,t,n,r){const a=(0,s.I)(e,n);let i;if(null!=r){if("number"==typeof r){if(Number.isNaN(r))return;i=r}else i="boolean"==typeof r?r:"string"==typeof r?a.spaceSeparated?(0,c.q)(r):a.commaSeparated?(0,o.q)(r):a.commaOrSpaceSeparated?(0,c.q)((0,o.q)(r).join(" ")):h(a,a.property,r):Array.isArray(r)?[...r]:"style"===a.property?function(e){const t=[];for(const[n,r]of Object.entries(e))t.push([n,r].join(": "));return t.join("; ")}(r):String(r);if(Array.isArray(i)){const e=[];for(const t of i)e.push(h(a,a.property,t));i=e}"className"===a.property&&Array.isArray(t.className)&&(i=t.className.concat(i)),t[a.property]=i}}function p(e,t){if(null==t);else if("number"==typeof t||"string"==typeof t)e.push({type:"text",value:String(t)});else if(Array.isArray(t))for(const n of t)p(e,n);else{if("object"!=typeof t||!("type"in t))throw new Error("Expected node, nodes, or string, got `"+t+"`");"root"===t.type?p(e,t.children):e.push(t)}}function h(e,t,n){if("string"==typeof n){if(e.number&&n&&!Number.isNaN(Number(n)))return Number(n);if((e.boolean||e.overloadedBoolean)&&(""===n||(0,l.S)(n)===(0,l.S)(t)))return!0}return n}const f=u(a.qy,"div"),m=u(a.JW,"g",["altGlyph","altGlyphDef","altGlyphItem","animateColor","animateMotion","animateTransform","clipPath","feBlend","feColorMatrix","feComponentTransfer","feComposite","feConvolveMatrix","feDiffuseLighting","feDisplacementMap","feDistantLight","feDropShadow","feFlood","feFuncA","feFuncB","feFuncG","feFuncR","feGaussianBlur","feImage","feMerge","feMergeNode","feMorphology","feOffset","fePointLight","feSpecularLighting","feSpotLight","feTile","feTurbulence","foreignObject","glyphRef","linearGradient","radialGradient","solidColor","textArea","textPath"]);function g(e,t){const n=e.indexOf("\r",t),r=e.indexOf("\n",t);return-1===r?n:-1===n||n+1===r?r:n<r?n:r}var v=n(4565);const b={}.hasOwnProperty,y=Object.prototype;function w(e,t){const n=t||{};return O({file:n.file||void 0,location:!1,schema:"svg"===n.space?a.JW:a.qy,verbose:n.verbose||!1},e)}function O(e,t){let n;switch(t.nodeName){case"#comment":{const r=t;return n={type:"comment",value:r.data},x(e,r,n),n}case"#document":case"#document-fragment":{const a=t,o="mode"in a&&("quirks"===a.mode||"limited-quirks"===a.mode);if(n={type:"root",children:k(e,t.childNodes),data:{quirksMode:o}},e.file&&e.location){const t=String(e.file),a=function(e){const t=String(e),n=[];return{toOffset:function(e){if(e&&"number"==typeof e.line&&"number"==typeof e.column&&!Number.isNaN(e.line)&&!Number.isNaN(e.column)){for(;n.length<e.line;){const e=n[n.length-1],r=g(t,e),a=-1===r?t.length+1:r+1;if(e===a)break;n.push(a)}const r=(e.line>1?n[e.line-2]:0)+e.column-1;if(r<n[e.line-1])return r}},toPoint:function(e){if("number"==typeof e&&e>-1&&e<=t.length){let r=0;for(;;){let a=n[r];if(void 0===a){const e=g(t,n[r-1]);a=-1===e?t.length+1:e+1,n[r]=a}if(a>e)return{line:r+1,column:e-(r>0?n[r-1]:0)+1,offset:e};r++}}}}}(t),o=a.toPoint(0),i=a.toPoint(t.length);(0,r.ok)(o,"expected `start`"),(0,r.ok)(i,"expected `end`"),n.position={start:o,end:i}}return n}case"#documentType":return n={type:"doctype"},x(e,t,n),n;case"#text":{const r=t;return n={type:"text",value:r.value},x(e,r,n),n}default:return n=function(e,t){const n=e.schema;e.schema=t.namespaceURI===v.t.svg?a.JW:a.qy;let r=-1;const o={};for(;++r<t.attrs.length;){const e=t.attrs[r],n=(e.prefix?e.prefix+":":"")+e.name;b.call(y,n)||(o[n]=e.value)}const i=("svg"===e.schema.space?m:f)(t.tagName,o,k(e,t.childNodes));if(x(e,t,i),"template"===i.tagName){const n=t,r=n.sourceCodeLocation,a=r&&r.startTag&&S(r.startTag),o=r&&r.endTag&&S(r.endTag),s=O(e,n.content);a&&o&&e.file&&(s.position={start:a.end,end:o.start}),i.content=s}return e.schema=n,i}(e,t),n}}function k(e,t){let n=-1;const r=[];for(;++n<t.length;){const a=O(e,t[n]);r.push(a)}return r}function x(e,t,n){if("sourceCodeLocation"in t&&t.sourceCodeLocation&&e.file){const r=E(e,n,t.sourceCodeLocation);r&&(e.location=!0,n.position=r)}}function E(e,t,n){const a=S(n);if("element"===t.type){const o=t.children[t.children.length-1];if(a&&!n.endTag&&o&&o.position&&o.position.end&&(a.end=Object.assign({},o.position.end)),e.verbose){const a={};let o;if(n.attrs)for(o in n.attrs)b.call(n.attrs,o)&&(a[(0,s.I)(e.schema,o).property]=S(n.attrs[o]));(0,r.ok)(n.startTag,"a start tag should exist");const i=S(n.startTag),l=n.endTag?S(n.endTag):void 0,c={opening:i};l&&(c.closing=l),c.properties=a,t.data={position:c}}}return a}function S(e){const t=_({line:e.startLine,column:e.startCol,offset:e.startOffset}),n=_({line:e.endLine,column:e.endCol,offset:e.endOffset});return t||n?{start:t,end:n}:void 0}function _(e){return e.line&&e.column?e:void 0}},9971:(e,t,n)=>{"use strict";n.d(t,{A:()=>c});var r=n(641),a=n(33),o=n(953),i=n(1266),s=n(4282);const l=["checked","disabled"],c=(0,r.pM)({__name:"DataTableCheckbox",props:{modelValue:{type:Boolean},disabled:{type:Boolean},align:{default:"center"}},emits:["update:modelValue"],setup(e){const t=(0,i.Fj)({base:"w-8 h-8 flex items-center justify-center text-border peer-checked:text-c-1 pointer-events-none absolute",variants:{align:{left:"left-0",center:"centered"}}});return(e,n)=>((0,r.uX)(),(0,r.Wv)(s.A,{class:"group/cell relative flex min-w-8"},{default:(0,r.k6)((()=>[(0,r.Lk)("input",{checked:e.modelValue,class:"peer absolute inset-0 opacity-0 disabled:cursor-default cursor-pointer",disabled:!!e.disabled,type:"checkbox",onChange:n[0]||(n[0]=t=>e.$emit("update:modelValue",t.target.checked))},null,40,l),(0,r.Lk)("div",{class:(0,a.C4)((0,o.R1)(t)({align:e.align}))},[(0,r.Lk)("div",{class:(0,a.C4)(["absolute opacity-0 border-[1px] rounded size-3/4 m-auto",!e.disabled&&"group-hover/cell:opacity-100 group-has-[:focus-visible]/cell:opacity-100 group-has-[:focus-visible]/cell:border-c-accent"])},null,2),(0,r.bF)((0,o.R1)(i.kv),{icon:"Checkmark",size:"xs",thickness:"2.5"})],2)])),_:1}))}})},9982:(e,t,n)=>{"use strict";e.exports=n(4477)}},t={};function n(r){var a=t[r];if(void 0!==a)return a.exports;var o=t[r]={exports:{}};return e[r](o,o.exports,n),o.exports}n.d=(e,t)=>{for(var r in t)n.o(t,r)&&!n.o(e,r)&&Object.defineProperty(e,r,{enumerable:!0,get:t[r]})},n.g=function(){if("object"==typeof globalThis)return globalThis;try{return this||new Function("return this")()}catch(e){if("object"==typeof window)return window}}(),n.o=(e,t)=>Object.prototype.hasOwnProperty.call(e,t),n.r=e=>{"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(e,"__esModule",{value:!0})},(()=>{"use strict";var e=n(5338),t=n(6540),r={outline:{xmlns:"http://www.w3.org/2000/svg",width:24,height:24,viewBox:"0 0 24 24",fill:"none",stroke:"currentColor",strokeWidth:2,strokeLinecap:"round",strokeLinejoin:"round"},filled:{xmlns:"http://www.w3.org/2000/svg",width:24,height:24,viewBox:"0 0 24 24",fill:"currentColor",stroke:"none"}};const a=(e,n,a,o)=>{const i=(0,t.forwardRef)((({color:a="currentColor",size:i=24,stroke:s=2,title:l,className:c,children:u,...d},p)=>(0,t.createElement)("svg",{ref:p,...r[e],width:i,height:i,className:["tabler-icon",`tabler-icon-${n}`,c].join(" "),..."filled"===e?{fill:a}:{strokeWidth:s,stroke:a},...d},[l&&(0,t.createElement)("title",{key:"svg-title"},l),...o.map((([e,n])=>(0,t.createElement)(e,n))),...Array.isArray(u)?u:[u]])));return i.displayName=`${a}`,i};var o=a("outline","book","IconBook",[["path",{d:"M3 19a9 9 0 0 1 9 0a9 9 0 0 1 9 0",key:"svg-0"}],["path",{d:"M3 6a9 9 0 0 1 9 0a9 9 0 0 1 9 0",key:"svg-1"}],["path",{d:"M3 6l0 13",key:"svg-2"}],["path",{d:"M12 6l0 13",key:"svg-3"}],["path",{d:"M21 6l0 13",key:"svg-4"}]]),i=a("outline","script","IconScript",[["path",{d:"M17 20h-11a3 3 0 0 1 0 -6h11a3 3 0 0 0 0 6h1a3 3 0 0 0 3 -3v-11a2 2 0 0 0 -2 -2h-10a2 2 0 0 0 -2 2v8",key:"svg-0"}]]),s=a("outline","brand-github","IconBrandGithub",[["path",{d:"M9 19c-4.3 1.4 -4.3 -2.5 -6 -3m12 5v-3.5c0 -1 .1 -1.4 -.5 -2c2.8 -.3 5.5 -1.4 5.5 -6a4.6 4.6 0 0 0 -1.3 -3.2a4.2 4.2 0 0 0 -.1 -3.2s-1.1 -.3 -3.5 1.3a12.3 12.3 0 0 0 -6.2 0c-2.4 -1.6 -3.5 -1.3 -3.5 -1.3a4.2 4.2 0 0 0 -.1 3.2a4.6 4.6 0 0 0 -1.3 3.2c0 4.6 2.7 5.7 5.5 6c-.6 .6 -.6 1.2 -.5 2v3.5",key:"svg-0"}]]),l=a("outline","x","IconX",[["path",{d:"M18 6l-12 12",key:"svg-0"}],["path",{d:"M6 6l12 12",key:"svg-1"}]]),c=a("outline","menu-2","IconMenu2",[["path",{d:"M4 6l16 0",key:"svg-0"}],["path",{d:"M4 12l16 0",key:"svg-1"}],["path",{d:"M4 18l16 0",key:"svg-2"}]]),u=n(961),d=n(4848);function p(e){let{...t}=e;return(0,d.jsx)("svg",{viewBox:"0 0 1024 1024",fill:"currentColor",xmlns:"http://www.w3.org/2000/svg",...t,children:(0,d.jsx)("path",{d:"M734.61,441.99c-7.05-15.75-9.96-32.93-8.26-50.83,1.72-18.08,13.06-35.59,26.3-52.84,.66-1.73,8.25-11.78-7.07-5.84-3.79,.74-16.01,5.48-19.74,6.61-32.26,9.73-43.32,29.76-52.49,66.59-3.39,13.6-3.19,28.47-4.11,42.8-1.09,17.02-1.47,34.1-2.49,51.13-.4,6.65-.96,13.51-2.72,19.8-9.42,33.62-43.08,55.79-72.92,48.45-24.22-5.95-39.46-26.85-40.25-55.39-.36-13.17,1.9-25.66,8.56-36.6,4.8-7.89,9.78-15.64,14.95-23.2,14.1-20.64,26-42.69,33.76-67.61,4.88-15.67,5.79-31.99,6.1-48.32,.25-12.84,.78-25.97-1.13-38.5-5.01-32.85-16.97-62.77-33.87-89.52-19.85-31.43-45.02-56.86-72.02-79.83-40.98-32.52-5.03,50.15-4.46,81.96,.35,19.63,.31,39.17-3.85,58.36-6.6,30.4-20.91,55.67-39.07,78.36-20.7,25.85-44.14,48.14-67.7,70.28-27.96,26.28-56.18,52.2-81.3,82.27-27.42,32.82-48.89,69.9-61.12,113.69-5.24,18.77-10.29,37.67-11.06,57.49-.77,19.99-.76,40.03-.77,60.04,0,7.1,.32,14.35,1.55,21.29,10.52,59.38,34.5,109.72,75.2,148.62,29.12,27.83,52.9,41.17,85.72,53.58-10.19-10.79-16.94-20.99-26.23-35.86-35.68-57.1-48.33-121.18-40.92-190.73,3.28-30.8,11.37-60.07,22.68-88.23,14.75-36.72,19.81-50.3,45.56-81.01,21.79-31.69,23.03-10.04,22.49-6.18-3.86,27.43-6.75,54.98-4.67,82.8,2.13,28.45,12.22,52.83,29.27,72.9,21.31,25.08,47.84,30.81,76.54,23.2,19.23-5.1,38.46-8.62,58.18-7.81,8.4,.35,16.84,.07,24.9,4.1,27.05,13.52,45.64,36.41,55.25,68.43,8.76,29.2,7.75,59.16,1.55,88.8-4.91,23.47-19.9,51.68-36.69,69.87,9.99-7.51,18.59-17.66,29.05-26,11.97-8.06,46.48-42.15,72.77-72.74,117.87-120.71,14.8-391.43,4.53-414.37Z",fill:"#e05922"})})}function h(e){let{...t}=e;return(0,d.jsxs)("svg",{viewBox:"0 0 1024 1024",xmlns:"http://www.w3.org/2000/svg",...t,children:[(0,d.jsx)("path",{d:"M341.49,413.75c-4.13,3.29-7.85,7.08-11.69,10.63-3.67,3.77-7.31,7.52-10.71,11.48-13.54,15.88-25.22,33.01-34.21,51.46-18.33,36.64-27.95,76.95-29.36,116.91-.64,19.99,.85,39.91,4.29,59.3,3.76,19.34,9.4,38.15,17.34,55.83,7.83,17.71,18.02,34.17,30.16,49.5,12.02,15.37,25.79,29.57,40.97,42.3,15.05,12.88,31.44,24.38,48.76,34.48,17.22,10.29,35.46,19.01,54.23,26.65,18.76,7.69,38.19,13.94,57.97,19.23,19.85,5.05,40.06,9.06,60.53,12.06,20.48,3,41.23,4.61,62.11,5.66,20.89,.87,41.91,.7,63.02-.11,21.08-1.09,42.29-2.61,63.49-5.17,21.26-2.26,125.44-27.98,33.77,3.27-10.2,3.64-20.42,7.38-30.83,10.52-20.71,6.81-41.99,11.73-63.46,16.19-21.55,3.99-43.37,7.12-65.41,8.72-5.5,.51-11.03,.67-16.56,.93-5.53,.27-11.07,.48-16.62,.43-5.55,0-11.1,.12-16.66,.01l-16.68-.66c-44.48-2.39-89.13-10.67-131.81-26.02-21.35-7.61-42.13-17.1-62.13-28.2-19.9-11.28-38.93-24.35-56.6-39.25-17.54-15.05-33.78-31.87-47.88-50.57-13.9-18.74-26.19-39.3-34.7-61.32-8.61-21.93-14.47-44.79-17.21-67.89-2.73-23.1-2.51-46.35,.21-69.14,5.74-45.52,21.81-89.49,47.76-126.49,19.27-27.67,45.15-50.46,73.59-67.06,15.58-9.09,12.19-.88,4.33,6.3Z",fill:"#e05922"}),(0,d.jsx)("path",{d:"M459.32,516.03c4.44-3.84,10.57,1.79,7.52,6.81-4.99,8.18-7.54,17.19-6.95,25.72,.51,8.41,3.67,15.98,8.7,22.02,4.93,5.69,11.76,10.38,19.26,13.32,15.06,6.17,31.86,5.47,44.44-.91,12.66-6.28,23.08-18.15,28.39-31.95,5.51-13.78,5.92-28.56,1.92-41.61-1.95-6.6-5.4-12.76-9.53-18.72l-1.51-2.25-1.65-1.92-1.6-1.95-1.88-1.88c-1.19-1.28-2.47-2.53-3.91-3.69-1.37-1.2-2.7-2.43-4.26-3.48-11.78-9.12-26.26-15.18-40.74-17.29-7.29-.85-14.59-1.1-21.9-.32-1.83,.23-3.64,.65-5.47,.93l-2.74,.44-2.71,.76-5.19,1.38c-.17,.05-.34,.1-.51,.16l-5.12,1.95c-14.45,5.37-27.44,13.84-37.15,25.22-10.01,11.1-17.29,24.81-21.02,39.89-3.75,15.11-3.76,31.41-.7,47.65,1.48,8.14,4.12,16.18,7.32,24.12,2.98,7.83,6.67,14.57,11.45,21.26,18.93,26.56,53.25,45,90.6,51.63,18.72,3.52,38.35,4.07,57.9,1.5,9.88-1.38,19.8-3.54,29.53-6.75,5.52-1.82,9.38,5.57,4.65,8.96-8.94,6.4-18.81,11.76-29.12,16-20.39,8.5-42.63,12.95-65.24,13.53-22.6,.48-45.76-2.69-68.01-10.85-11.16-3.97-22-9.31-32.32-15.87-10.29-6.58-20.16-14.38-28.69-23.73-8.53-9.18-16.21-19.97-21.97-30.92-5.74-10.74-10.56-22.15-14.02-34.12-7.19-23.87-8.8-49.99-4.32-75.53,4.59-25.55,16.38-50.09,33.24-70.47,17-20.42,39.62-35.63,63.87-44.89l9.02-3.2c.14-.05,.28-.09,.42-.13l9.38-2.44,4.81-1.18,4.93-.81c3.3-.5,6.57-1.09,9.89-1.45,13.33-1.25,26.85-.88,40.07,1.34,26.3,5.06,51,16.33,70.94,33.99,2.56,2.1,4.94,4.49,7.31,6.88,2.41,2.31,4.69,4.83,6.88,7.48l3.18,3.77c.08,.09,.15,.19,.22,.29l2.99,4.1,2.92,4.11c.08,.11,.15,.22,.22,.34l2.5,4.17c6.99,11.47,12.28,24.45,14.99,38.08,2.75,13.59,3.16,27.69,.94,41.21-2.22,13.53-6.8,26.43-13.32,38.13-6.47,11.72-15.09,22.19-25.47,30.81-10.17,8.62-22.47,15.42-35.6,19.14-13.12,3.76-26.95,4.55-39.89,2.27-12.95-2.27-25.1-7.13-35.66-14.31-10.47-7.25-19.72-16.64-25.64-28.65-5.59-11.7-7.51-25.35-4.7-37.54,1.25-6.13,3.74-11.79,6.69-16.94,3.23-5.02,6.85-9.61,11.16-13.36,.09-.07,.17-.15,.26-.22Z",fill:"#018080"}),(0,d.jsx)("path",{d:"M98.42,646.77l.07-16.47,.52-16.51c1.03-22.04,3.8-44.08,8.11-65.99,8.71-43.73,24.57-87.01,49.25-125.97,24.47-38.99,58.12-73.12,97.84-98.05,19.84-12.47,41.14-22.6,63.18-30.52,22.13-7.67,44.99-13.18,68.18-16.14,46.35-5.78,94-2.55,139.16,12.63,11.27,3.81,22.39,8.57,33.11,14.01,10.64,5.57,20.7,11.96,30.36,18.8,9.61,6.92,18.64,14.52,27.17,22.58,8.41,8.19,16.39,16.76,23.76,25.83,29.1,36.36,48.92,80.72,52.58,126.94,1.85,23.01-.5,46.21-7.07,67.84-6.51,21.64-22.14,38.76-19.52,28.34,3.07-10.23,4.93-20.71,6.18-31.1,1.03-10.45,1.32-20.85,.77-31.14-.86-10.29-1.86-20.48-4.35-30.41-9.05-39.83-30.33-75.81-58.1-104.69-27.76-29.04-62.14-50.98-99.26-61.61-37.74-10.7-78.7-12.91-118.28-7.67-39.65,5.17-78.28,17.85-112.81,37.81-34.75,19.7-64.98,47.12-89.77,79.46-24.82,32.41-43.74,69.97-58.02,109.77-7.22,19.9-13.15,40.44-18.35,61.32l-3.68,15.75-3.28,15.89-3.14,15.99c-2.34,13.37-4.27-5.72-4.62-16.7Z",fill:"#1ab7d1"}),(0,d.jsx)("path",{d:"M922.46,330.55c.86,10.35,1.65,20.72,2.14,31.12,1.23,20.81,1.19,41.7,.44,62.66-1.75,41.88-7.37,84.19-19.75,125.55-12.21,41.26-31.57,81.7-59.3,116.49-13.72,17.46-29.51,33.28-46.44,47.55-17.03,14.18-35.39,26.53-54.36,37.43-19.06,10.8-38.88,19.76-58.94,27.65-10.03,4-20.23,7.31-30.39,10.74l-15.32,4.71c-5.03,1.57-10.71,3.08-16.05,4.29-21.66,5.07-43.48,8.01-65.46,8.76-21.96,.74-44.01-.54-65.72-4.41-43.33-7.44-85.59-25.6-118.6-54.31-16.49-14.26-30.84-30.66-42.15-48.7-2.71-4.6-5.55-9.06-8.08-13.7-2.4-4.72-4.98-9.3-7.18-14.06-2.05-4.85-4.2-9.62-6.12-14.46-1.81-4.9,.67-6.16,3.5-1.87,2.98,4.2,6.16,8.26,9.2,12.35,3.1,4.05,6.42,7.92,9.59,11.86,3.28,3.86,6.83,7.48,10.19,11.2,13.95,14.48,29.35,27.1,45.95,37.33,33.09,20.71,70.7,31.76,108.92,34.74,38.17,3.02,77.34-1.07,113.89-11.4,4.64-1.31,8.86-2.66,13.62-4.25l14.06-4.78c9.24-3.45,18.56-6.75,27.54-10.66,18.15-7.41,35.87-15.62,52.57-25.27,33.62-18.89,63.84-42.56,88.67-71.32,12.47-14.33,23.53-29.95,33.31-46.54,9.68-16.64,18.18-34.2,25.49-52.44,14.8-36.45,24.81-75.52,32.69-115.35,7.99-39.88,21.11-101.25,22.1-90.9Z",fill:"#9d744f"}),(0,d.jsx)("path",{d:"M733.6,626.97c-2.93,5.29-10.97,2.03-9.45-3.82,14.58-55.99,21.84-112.28,17.4-167.2-4.75-66.45-28.39-129.68-68.89-179.51-40.28-50.26-98.17-84.83-162.02-104.09-31.99-9.66-65.52-16.11-99.82-19.24-8.56-1.08-17.19-1.22-25.85-1.97-8.67-.88-17.35-.84-26.1-1.18-8.74-.25-17.55-.77-26.32-.41-8.81-.04-12.77-4.21-5.81-6.22,4.67-1.35,9.39-2.55,14.14-3.62,14.27-3.21,28.63-6.77,43.25-8.46,8.91-1.08,17.84-2.5,26.84-3.13,35.99-3.21,72.78-1.35,109.11,5.38,36.3,6.8,72.1,19.32,105.08,37.84,4.24,2.1,8.22,4.67,12.17,7.29l11.92,7.76c7.78,5.31,15.07,11.35,22.61,17.04,.11,.08,.22,.17,.32,.26,7.01,6.34,14.07,12.66,20.79,19.32l9.73,10.5c3.25,3.5,6.54,6.96,9.37,10.81,5.76,7.61,11.94,14.92,16.9,23.09l7.81,11.98c2.49,4.06,4.67,8.3,6.94,12.49,2.19,4.23,4.6,8.35,6.58,12.66l5.73,13.05c4.09,8.59,6.86,17.68,9.95,26.62,11.69,36,16.61,73.75,15.42,110.81-1.86,61.23-19.21,120.32-47.78,171.95Z",fill:"#00966f"})]})}const f={xs:"h-4 w-4 md:h-5 md:w-5",sm:"h-5 w-5 md:h-6 md:w-6",md:"h-6 w-6 md:h-7 md:w-7",lg:"h-7 w-7 md:h-8 md:w-8",xl:"h-8 w-8 md:h-9 md:w-9","2xl":"h-9 w-9 md:h-10 md:w-10"},m={xs:"text-xs leading-4 md:text-sm md:leading-5",sm:"text-sm leading-5 md:text-base md:leading-6",md:"text-base leading-6 md:text-lg md:leading-7",lg:"text-lg leading-7 md:text-xl md:leading-8",xl:"text-xl leading-8 md:text-2xl md:leading-9","2xl":"text-2xl leading-9 md:text-3xl md:leading-10"},g={vortico:"text-vortico-500",bosque:"text-bosque-500",bruma:"text-bruma-500",ciclon:"text-ciclon-500",flama:"text-flama-500",primary:"text-primary-500","vortico-light":"text-vortico-200","bosque-light":"text-bosque-200","bruma-light":"text-bruma-200","ciclon-light":"text-ciclon-200","flama-light":"text-flama-200","primary-light":"text-primary-200","vortico-dark":"text-vortico-800","bosque-dark":"text-bosque-800","bruma-dark":"text-bruma-800","ciclon-dark":"text-ciclon-800","flama-dark":"text-flama-800","primary-dark":"text-primary-800"},v={flama:{text:"Flama",icon:(0,d.jsx)(p,{})},vortico:{text:"Vortico",icon:(0,d.jsx)(h,{})}};function b(e){let{logo:t,color:n,size:r="md"}=e;const{text:a,icon:o}=v[t],i=n?g[n]:"";return(0,d.jsxs)("div",{className:"flex items-center justify-center gap-1","aria-label":`${a} logo`,children:[(0,d.jsx)("div",{className:`${f[r]} ${i}`,children:o}),(0,d.jsx)("span",{className:`text-brand ${m[r]} ${i}`,children:a})]})}function y(e){let{onClick:t,...n}=e;const r=[{href:"https://flama.dev/docs/",title:"Docs",icon:(0,d.jsx)(o,{}),label:"Flama docs"},{href:"https://flama.dev/blog/",title:"Blog",icon:(0,d.jsx)(i,{}),label:"Flama blog"}];return(0,d.jsx)("nav",{children:(0,d.jsx)("ul",{...n,children:r.map(((e,n)=>{let{href:r,title:a,icon:o,label:i}=e;return(0,d.jsx)("li",{onClick:t,children:(0,d.jsx)("a",{href:r,"aria-label":i,children:(0,d.jsxs)("div",{className:"flex h-12 cursor-pointer items-center justify-start gap-2",children:[(0,d.jsx)("div",{className:"text-primary-400 hover:text-flama-500 h-6 w-6 transition-colors duration-200 md:hidden",children:o}),(0,d.jsx)("div",{className:"text-primary-400 hover:text-flama-500 text-md font-medium transition-colors duration-200",children:a})]})})},n)}))})})}function w(e){let{...t}=e;const n=[{href:"https://github.com/vortico/flama",title:"GitHub",icon:(0,d.jsx)(s,{}),label:"Flama on GitHub"}];return(0,d.jsx)("div",{children:(0,d.jsx)("ul",{...t,children:n.map(((e,t)=>{let{href:n,title:r,icon:a,label:o}=e;return(0,d.jsx)("li",{children:(0,d.jsx)("a",{href:n,"aria-label":o,children:(0,d.jsxs)("div",{className:"flex h-12 cursor-pointer items-center justify-start gap-2",children:[(0,d.jsx)("div",{className:"text-primary-400 hover:text-flama-500 h-6 w-6 transition-colors duration-200",children:a}),(0,d.jsx)("div",{className:"text-primary-400 hover:text-flama-500 text-md font-medium transition-colors duration-200 md:hidden",children:r})]})})},t)}))})})}function O(e){let{onClose:t}=e;return(0,d.jsxs)("div",{className:"bg-primary-100 fixed inset-x-0 top-0 z-50 h-screen w-screen overflow-auto","aria-modal":"true",role:"dialog",children:[(0,d.jsxs)("div",{className:"border-flama-500 flex h-14 w-full items-center justify-between border-b px-4 sm:px-6 md:px-8",children:[(0,d.jsx)("a",{href:"https://flama.dev",className:"block cursor-pointer","aria-label":"Flama website",children:(0,d.jsx)(b,{logo:"flama",color:"flama",size:"lg"})}),(0,d.jsx)("button",{className:"h-6 w-6",onClick:t,"aria-label":"Close menu",children:(0,d.jsx)(l,{className:"text-primary-400 hover:text-flama-500 h-full w-full transition-colors duration-200"})})]}),(0,d.jsxs)("div",{className:"w-full px-4 sm:px-6 md:px-8",children:[(0,d.jsx)(y,{className:"flex flex-col items-start justify-start",onClick:t}),(0,d.jsx)(w,{className:"flex flex-col items-start justify-start gap-8"})]})]})}function k(e){let{onOpen:t}=e;return(0,d.jsx)(d.Fragment,{children:(0,d.jsxs)("div",{className:"flex h-14 w-full items-center justify-between",children:[(0,d.jsx)("a",{href:"https://flama.dev",className:"block cursor-pointer","aria-label":"Flama website",children:(0,d.jsx)(b,{logo:"flama",color:"flama",size:"lg"})}),(0,d.jsxs)("div",{className:"hidden items-center justify-between gap-8 md:flex",children:[(0,d.jsx)(y,{className:"flex items-center justify-start gap-8"}),(0,d.jsx)(w,{className:"flex items-center justify-start gap-8"})]}),(0,d.jsx)("button",{className:"block h-6 w-6 cursor-pointer md:hidden",onClick:t,"aria-label":"Open menu",children:(0,d.jsx)(c,{className:"h-full w-full"})})]})})}function x(){const[e,n]=(0,t.useState)(!1),r=(0,t.useCallback)((()=>n(!0)),[n]),a=(0,t.useCallback)((()=>n(!1)),[n]);return(0,d.jsxs)("div",{className:"border-flama-500 bg-primary-100 fixed inset-x-0 top-0 z-10 mx-auto border-b",children:[(0,d.jsx)("div",{className:"divide-flama-500/50 mx-auto w-full divide-y px-4 sm:px-6 md:px-8",children:(0,d.jsx)(k,{onOpen:r})}),e&&(0,u.createPortal)((0,d.jsx)(O,{onClose:a}),document.body)]})}var E=n(641),S=n(953),_=n(33),C=n(1014),T=n(5841);function A(e,t){const n=[],r=t.resolveKeyData||(e=>e.key),a=t.resolveValueData||(e=>e.value);for(const[o,i]of Object.entries(e))n.push(...(Array.isArray(i)?i:[i]).map((e=>{const n={key:o,value:e},i=a(n);return"object"==typeof i?A(i,t):Array.isArray(i)?i:{["function"==typeof t.key?t.key(n):t.key]:r(n),["function"==typeof t.value?t.value(n):t.value]:i}})).flat());return n}function R(e,t){return Object.entries(e).map((([e,n])=>{if("object"==typeof n&&(n=R(n,t)),t.resolve){const r=t.resolve({key:e,value:n});if(void 0!==r)return r}return"number"==typeof n&&(n=n.toString()),"string"==typeof n&&t.wrapValue&&(n=n.replace(new RegExp(t.wrapValue,"g"),`\\${t.wrapValue}`),n=`${t.wrapValue}${n}${t.wrapValue}`),`${e}${t.keyValueSeparator||""}${n}`})).join(t.entrySeparator||"")}const P=new Set(["title","titleTemplate","script","style","noscript"]),D=new Set(["base","meta","link","style","script","noscript"]),M=new Set(["title","titleTemplate","templateParams","base","htmlAttrs","bodyAttrs","meta","link","style","script","noscript"]),$=new Set(["base","title","titleTemplate","bodyAttrs","htmlAttrs","templateParams"]),N=new Set(["tagPosition","tagPriority","tagDuplicateStrategy","children","innerHTML","textContent","processTemplateParams"]),I="undefined"!=typeof window;function L(e){let t=9;for(let n=0;n<e.length;)t=Math.imul(t^e.charCodeAt(n++),9**9);return(65536+(t^t>>>9)).toString(16).substring(1,8).toLowerCase()}function B(e){if(e._h)return e._h;if(e._d)return L(e._d);let t=`${e.tag}:${e.textContent||e.innerHTML||""}:`;for(const n in e.props)t+=`${n}:${String(e.props[n])},`;return L(t)}const F=e=>({keyValue:e,metaKey:"property"}),Q=e=>({keyValue:e}),U={appleItunesApp:{unpack:{entrySeparator:", ",resolve:({key:e,value:t})=>`${W(e)}=${t}`}},articleExpirationTime:F("article:expiration_time"),articleModifiedTime:F("article:modified_time"),articlePublishedTime:F("article:published_time"),bookReleaseDate:F("book:release_date"),charset:{metaKey:"charset"},contentSecurityPolicy:{unpack:{entrySeparator:"; ",resolve:({key:e,value:t})=>`${W(e)} ${t}`},metaKey:"http-equiv"},contentType:{metaKey:"http-equiv"},defaultStyle:{metaKey:"http-equiv"},fbAppId:F("fb:app_id"),msapplicationConfig:Q("msapplication-Config"),msapplicationTileColor:Q("msapplication-TileColor"),msapplicationTileImage:Q("msapplication-TileImage"),ogAudioSecureUrl:F("og:audio:secure_url"),ogAudioUrl:F("og:audio"),ogImageSecureUrl:F("og:image:secure_url"),ogImageUrl:F("og:image"),ogSiteName:F("og:site_name"),ogVideoSecureUrl:F("og:video:secure_url"),ogVideoUrl:F("og:video"),profileFirstName:F("profile:first_name"),profileLastName:F("profile:last_name"),profileUsername:F("profile:username"),refresh:{metaKey:"http-equiv",unpack:{entrySeparator:";",resolve({key:e,value:t}){if("seconds"===e)return`${t}`}}},robots:{unpack:{entrySeparator:", ",resolve:({key:e,value:t})=>"boolean"==typeof t?`${W(e)}`:`${W(e)}:${t}`}},xUaCompatible:{metaKey:"http-equiv"}},z=new Set(["og","book","article","profile"]);function j(e){const t=W(e),n=t.indexOf(":");return z.has(t.substring(0,n))?"property":U[e]?.metaKey||"name"}function W(e){const t=e.replace(/([A-Z])/g,"-$1").toLowerCase(),n=t.indexOf("-"),r=t.substring(0,n);return"twitter"===r||z.has(r)?e.replace(/([A-Z])/g,":$1").toLowerCase():t}function X(e){if(Array.isArray(e))return e.map((e=>X(e)));if("object"!=typeof e||Array.isArray(e))return e;const t={};for(const n in e)Object.prototype.hasOwnProperty.call(e,n)&&(t[W(n)]=X(e[n]));return t}function q(e,t){const n=U[t];return"refresh"===t?`${e.seconds};url=${e.url}`:R(X(e),{keyValueSeparator:"=",entrySeparator:", ",resolve:({value:e,key:t})=>null===e?"":"boolean"==typeof e?`${t}`:void 0,...n?.unpack})}const V=new Set(["og:image","og:video","og:audio","twitter:image"]);function H(e){const t={};for(const n in e){if(!Object.prototype.hasOwnProperty.call(e,n))continue;const r=e[n];"false"!==String(r)&&n&&(t[n]=r)}return t}function Z(e,t){const n=H(t),r=W(e),a=j(r);if(V.has(r)){const t={};for(const r in n)Object.prototype.hasOwnProperty.call(n,r)&&(t[`${e}${"url"===r?"":`${r[0].toUpperCase()}${r.slice(1)}`}`]=n[r]);return G(t).sort(((e,t)=>(e[a]?.length||0)-(t[a]?.length||0)))}return[{[a]:r,...n}]}function G(e){const t=[],n={};for(const r in e){if(!Object.prototype.hasOwnProperty.call(e,r))continue;const a=e[r];if(Array.isArray(a))for(const e of a)t.push(..."string"==typeof e?G({[r]:e}):Z(r,e));else if("object"==typeof a&&a){if(V.has(W(r))){t.push(...Z(r,a));continue}n[r]=H(a)}else n[r]=a}const r=A(n,{key:({key:e})=>j(e),value:({key:e})=>"charset"===e?"charset":"content",resolveKeyData:({key:e})=>function(e){return U[e]?.keyValue||W(e)}(e),resolveValueData:({value:e,key:t})=>null===e?"_null":"object"==typeof e?q(e,t):"number"==typeof e?e.toString():e});return[...t,...r].map((e=>("_null"===e.content&&(e.content=null),e)))}function K(e,t,n,r){const a=r||ee("object"!=typeof t||"function"==typeof t||t instanceof Promise?{["script"===e||"noscript"===e||"style"===e?"innerHTML":"textContent"]:t}:{...t},"templateParams"===e||"titleTemplate"===e);if(a instanceof Promise)return a.then((r=>K(e,t,n,r)));const o={tag:e,props:a};for(const e of N){const t=void 0!==o.props[e]?o.props[e]:n[e];void 0!==t&&(("innerHTML"!==e&&"textContent"!==e&&"children"!==e||P.has(o.tag))&&(o["children"===e?"innerHTML":e]=t),delete o.props[e])}return o.props.body&&(o.tagPosition="bodyClose",delete o.props.body),"script"===o.tag&&"object"==typeof o.innerHTML&&(o.innerHTML=JSON.stringify(o.innerHTML),o.props.type=o.props.type||"application/json"),Array.isArray(o.props.content)?o.props.content.map((e=>({...o,props:{...o.props,content:e}}))):o}function Y(e,t){const n="class"===e?" ":";";return t&&"object"==typeof t&&!Array.isArray(t)&&(t=Object.entries(t).filter((([,e])=>e)).map((([t,n])=>"style"===e?`${t}:${n}`:t))),String(Array.isArray(t)?t.join(n):t)?.split(n).filter((e=>Boolean(e.trim()))).join(n)}function J(e,t,n,r){for(let a=r;a<n.length;a+=1){const r=n[a];if("class"!==r&&"style"!==r){if(e[r]instanceof Promise)return e[r].then((o=>(e[r]=o,J(e,t,n,a))));if(!t&&!N.has(r)){const t=String(e[r]),n=r.startsWith("data-");"true"===t||""===t?e[r]=!n||"true":e[r]||(n&&"false"===t?e[r]="false":delete e[r])}}else e[r]=Y(r,e[r])}}function ee(e,t=!1){const n=J(e,t,Object.keys(e),0);return n instanceof Promise?n.then((()=>e)):e}function te(e,t,n){for(let r=n;r<t.length;r+=1){const n=t[r];if(n instanceof Promise)return n.then((n=>(t[r]=n,te(e,t,r))));Array.isArray(n)?e.push(...n):e.push(n)}}function ne(e){const t=[],n=e.resolvedInput;for(const r in n){if(!Object.prototype.hasOwnProperty.call(n,r))continue;const a=n[r];if(void 0!==a&&M.has(r))if(Array.isArray(a))for(const n of a)t.push(K(r,n,e));else t.push(K(r,a,e))}if(0===t.length)return[];const r=[];return o=()=>r.map(((t,n)=>(t._e=e._i,e.mode&&(t._m=e.mode),t._p=(e._i<<10)+n,t))),(a=te(r,t,0))instanceof Promise?a.then(o):o();var a,o}const re=new Set(["onload","onerror","onabort","onprogress","onloadstart"]),ae={base:-10,title:10},oe={critical:-80,high:-10,low:20};function ie(e){const t=e.tagPriority;if("number"==typeof t)return t;let n=100;return"meta"===e.tag?"content-security-policy"===e.props["http-equiv"]?n=-30:e.props.charset?n=-20:"viewport"===e.props.name&&(n=-15):"link"===e.tag&&"preconnect"===e.props.rel?n=20:e.tag in ae&&(n=ae[e.tag]),t&&t in oe?n+oe[t]:n}const se=[{prefix:"before:",offset:-1},{prefix:"after:",offset:1}],le=["name","property","http-equiv"];function ce(e){const{props:t,tag:n}=e;if($.has(n))return n;if("link"===n&&"canonical"===t.rel)return"canonical";if(t.charset)return"charset";if(t.id)return`${n}:id:${t.id}`;for(const e of le)if(void 0!==t[e])return`${n}:${e}:${t[e]}`;return!1}const ue="%separator",de=new RegExp(`${ue}(?:\\s*${ue})*`,"g");function pe(e,t,n,r=!1){if("string"!=typeof e||!e.includes("%"))return e;let a=e;try{a=decodeURI(e)}catch{}const o=a.match(/%\w+(?:\.\w+)?/g);if(!o)return e;const i=e.includes(ue);return e=e.replace(/%\w+(?:\.\w+)?/g,(e=>{if(e===ue||!o.includes(e))return e;const n=function(e,t,n=!1){let r;if("s"===t||"pageTitle"===t)r=e.pageTitle;else if(t.includes(".")){const n=t.indexOf(".");r=e[t.substring(0,n)]?.[t.substring(n+1)]}else r=e[t];if(void 0!==r)return n?(r||"").replace(/"/g,'\\"'):r||""}(t,e.slice(1),r);return void 0!==n?n:e})).trim(),i&&(e.endsWith(ue)&&(e=e.slice(0,-10)),e.startsWith(ue)&&(e=e.slice(10)),e=e.replace(de,n).trim()),e}function he(e,t){return null==e?t||null:"function"==typeof e?e(t):e}function fe(e,t={},n){for(const r in e){const a=e[r],o=n?`${n}:${r}`:r;"object"==typeof a&&null!==a?fe(a,t,o):"function"==typeof a&&(t[o]=a)}return t}const me={run:e=>e()},ge=void 0!==console.createTask?console.createTask:()=>me;function ve(e,t){const n=t.shift(),r=ge(n);return e.reduce(((e,n)=>e.then((()=>r.run((()=>n(...t)))))),Promise.resolve())}function be(e,t){const n=t.shift(),r=ge(n);return Promise.all(e.map((e=>r.run((()=>e(...t))))))}function ye(e,t){for(const n of[...e])n(t)}class we{constructor(){this._hooks={},this._before=void 0,this._after=void 0,this._deprecatedMessages=void 0,this._deprecatedHooks={},this.hook=this.hook.bind(this),this.callHook=this.callHook.bind(this),this.callHookWith=this.callHookWith.bind(this)}hook(e,t,n={}){if(!e||"function"!=typeof t)return()=>{};const r=e;let a;for(;this._deprecatedHooks[e];)a=this._deprecatedHooks[e],e=a.to;if(a&&!n.allowDeprecated){let e=a.message;e||(e=`${r} hook has been deprecated`+(a.to?`, please use ${a.to}`:"")),this._deprecatedMessages||(this._deprecatedMessages=new Set),this._deprecatedMessages.has(e)||(console.warn(e),this._deprecatedMessages.add(e))}if(!t.name)try{Object.defineProperty(t,"name",{get:()=>"_"+e.replace(/\W+/g,"_")+"_hook_cb",configurable:!0})}catch{}return this._hooks[e]=this._hooks[e]||[],this._hooks[e].push(t),()=>{t&&(this.removeHook(e,t),t=void 0)}}hookOnce(e,t){let n,r=(...e)=>("function"==typeof n&&n(),n=void 0,r=void 0,t(...e));return n=this.hook(e,r),n}removeHook(e,t){if(this._hooks[e]){const n=this._hooks[e].indexOf(t);-1!==n&&this._hooks[e].splice(n,1),0===this._hooks[e].length&&delete this._hooks[e]}}deprecateHook(e,t){this._deprecatedHooks[e]="string"==typeof t?{to:t}:t;const n=this._hooks[e]||[];delete this._hooks[e];for(const t of n)this.hook(e,t)}deprecateHooks(e){Object.assign(this._deprecatedHooks,e);for(const t in e)this.deprecateHook(t,e[t])}addHooks(e){const t=fe(e),n=Object.keys(t).map((e=>this.hook(e,t[e])));return()=>{for(const e of n.splice(0,n.length))e()}}removeHooks(e){const t=fe(e);for(const e in t)this.removeHook(e,t[e])}removeAllHooks(){for(const e in this._hooks)delete this._hooks[e]}callHook(e,...t){return t.unshift(e),this.callHookWith(ve,e,...t)}callHookParallel(e,...t){return t.unshift(e),this.callHookWith(be,e,...t)}callHookWith(e,t,...n){const r=this._before||this._after?{name:t,args:n,context:{}}:void 0;this._before&&ye(this._before,r);const a=e(t in this._hooks?[...this._hooks[t]]:[],n);return a instanceof Promise?a.finally((()=>{this._after&&r&&ye(this._after,r)})):(this._after&&r&&ye(this._after,r),a)}beforeEach(e){return this._before=this._before||[],this._before.push(e),()=>{if(void 0!==this._before){const t=this._before.indexOf(e);-1!==t&&this._before.splice(t,1)}}}afterEach(e){return this._after=this._after||[],this._after.push(e),()=>{if(void 0!==this._after){const t=this._after.indexOf(e);-1!==t&&this._after.splice(t,1)}}}}const Oe=new Set(["templateParams","htmlAttrs","bodyAttrs"]),ke={hooks:{"tag:normalise":({tag:e})=>{e.props.hid&&(e.key=e.props.hid,delete e.props.hid),e.props.vmid&&(e.key=e.props.vmid,delete e.props.vmid),e.props.key&&(e.key=e.props.key,delete e.props.key);const t=ce(e);!t||t.startsWith("meta:og:")||t.startsWith("meta:twitter:")||delete e.key;const n=t||!!e.key&&`${e.tag}:${e.key}`;n&&(e._d=n)},"tags:resolve":e=>{const t=Object.create(null);for(const n of e.tags){const e=(n.key?`${n.tag}:${n.key}`:n._d)||B(n),r=t[e];if(r){let a=n?.tagDuplicateStrategy;if(!a&&Oe.has(n.tag)&&(a="merge"),"merge"===a){const a=r.props;a.style&&n.props.style&&(";"!==a.style[a.style.length-1]&&(a.style+=";"),n.props.style=`${a.style} ${n.props.style}`),a.class&&n.props.class?n.props.class=`${a.class} ${n.props.class}`:a.class&&(n.props.class=a.class),t[e].props={...a,...n.props};continue}if(n._e===r._e){r._duped=r._duped||[],n._d=`${r._d}:${r._duped.length+1}`,r._duped.push(n);continue}if(ie(n)>ie(r))continue}n.innerHTML||n.textContent||0!==Object.keys(n.props).length||!D.has(n.tag)?t[e]=n:delete t[e]}const n=[];for(const e in t){const r=t[e],a=r._duped;n.push(r),a&&(delete r._duped,n.push(...a))}e.tags=n,e.tags=e.tags.filter((e=>!("meta"===e.tag&&(e.props.name||e.props.property)&&!e.props.content)))}}},xe=new Set(["script","link","bodyAttrs"]),Ee=e=>({hooks:{"tags:resolve":t=>{for(const n of t.tags){if(!xe.has(n.tag))continue;const t=n.props;for(const r in t){if("o"!==r[0]||"n"!==r[1])continue;if(!Object.prototype.hasOwnProperty.call(t,r))continue;const a=t[r];"function"==typeof a&&(e.ssr&&re.has(r)?t[r]=`this.dataset.${r}fired = true`:delete t[r],n._eventHandlers=n._eventHandlers||{},n._eventHandlers[r]=a)}e.ssr&&n._eventHandlers&&(n.props.src||n.props.href)&&(n.key=n.key||L(n.props.src||n.props.href))}},"dom:renderTag":({$el:e,tag:t})=>{const n=e?.dataset;if(n)for(const r in n){if(!r.endsWith("fired"))continue;const n=r.slice(0,-5);re.has(n)&&t._eventHandlers?.[n]?.call(e,new Event(n.substring(2)))}}}}),Se=new Set(["link","style","script","noscript"]),_e={hooks:{"tag:normalise":({tag:e})=>{e.key&&Se.has(e.tag)&&(e.props["data-hid"]=e._h=L(e.key))}}},Ce={mode:"server",hooks:{"tags:beforeResolve":e=>{const t={};let n=!1;for(const r of e.tags)"server"!==r._m||"titleTemplate"!==r.tag&&"templateParams"!==r.tag&&"title"!==r.tag||(t[r.tag]="title"===r.tag||"titleTemplate"===r.tag?r.textContent:r.props,n=!0);n&&e.tags.push({tag:"script",innerHTML:JSON.stringify(t),props:{id:"unhead:payload",type:"application/json"}})}}},Te={hooks:{"tags:resolve":e=>{for(const t of e.tags)if("string"==typeof t.tagPriority)for(const{prefix:n,offset:r}of se){if(!t.tagPriority.startsWith(n))continue;const a=t.tagPriority.substring(n.length),o=e.tags.find((e=>e._d===a))?._p;if(void 0!==o){t._p=o+r;break}}e.tags.sort(((e,t)=>{const n=ie(e),r=ie(t);return n<r?-1:n>r?1:e._p-t._p}))}}},Ae={meta:"content",link:"href",htmlAttrs:"lang"},Re=["innerHTML","textContent"],Pe=e=>({hooks:{"tags:resolve":t=>{const{tags:n}=t;let r;for(let e=0;e<n.length;e+=1)"templateParams"===n[e].tag&&(r=t.tags.splice(e,1)[0].props,e-=1);const a=r||{},o=a.separator||"|";delete a.separator,a.pageTitle=pe(a.pageTitle||n.find((e=>"title"===e.tag))?.textContent||"",a,o);for(const e of n){if(!1===e.processTemplateParams)continue;const t=Ae[e.tag];if(t&&"string"==typeof e.props[t])e.props[t]=pe(e.props[t],a,o);else if(e.processTemplateParams||"titleTemplate"===e.tag||"title"===e.tag)for(const t of Re)"string"==typeof e[t]&&(e[t]=pe(e[t],a,o,"script"===e.tag&&e.props.type.endsWith("json")))}e._templateParams=a,e._separator=o},"tags:afterResolve":({tags:t})=>{let n;for(let e=0;e<t.length;e+=1){const r=t[e];"title"===r.tag&&!1!==r.processTemplateParams&&(n=r)}n?.textContent&&(n.textContent=pe(n.textContent,e._templateParams,e._separator))}}}),De={hooks:{"tags:resolve":e=>{const{tags:t}=e;let n,r;for(let e=0;e<t.length;e+=1){const a=t[e];"title"===a.tag?n=a:"titleTemplate"===a.tag&&(r=a)}if(r&&n){const t=he(r.textContent,n.textContent);null!==t?n.textContent=t||n.textContent:e.tags.splice(e.tags.indexOf(n),1)}else if(r){const e=he(r.textContent);null!==e&&(r.textContent=e,r.tag="title",r=void 0)}r&&e.tags.splice(e.tags.indexOf(r),1)}}},Me={hooks:{"tags:afterResolve":e=>{for(const t of e.tags)"string"==typeof t.innerHTML&&(!t.innerHTML||"application/ld+json"!==t.props.type&&"application/json"!==t.props.type?t.innerHTML=t.innerHTML.replace(new RegExp(`</${t.tag}`,"g"),`<\\/${t.tag}`):t.innerHTML=t.innerHTML.replace(/</g,"\\u003C"))}}};let $e;function Ne(e={}){const t=function(e={}){const t=new we;t.addHooks(e.hooks||{}),e.document=e.document||(I?document:void 0);const n=!e.document,r=()=>{s.dirty=!0,t.callHook("entries:updated",s)};let a=0,o=[];const i=[],s={plugins:i,dirty:!1,resolvedOptions:e,hooks:t,headEntries:()=>o,use(e){const r="function"==typeof e?e(s):e;r.key&&i.some((e=>e.key===r.key))||(i.push(r),Ie(r.mode,n)&&t.addHooks(r.hooks||{}))},push(e,t){delete t?.head;const i={_i:a++,input:e,...t};return Ie(i.mode,n)&&(o.push(i),r()),{dispose(){o=o.filter((e=>e._i!==i._i)),r()},patch(e){for(const t of o)t._i===i._i&&(t.input=i.input=e);r()}}},async resolveTags(){const e={tags:[],entries:[...o]};await t.callHook("entries:resolve",e);for(const n of e.entries){const r=n.resolvedInput||n.input;if(n.resolvedInput=await(n.transform?n.transform(r):r),n.resolvedInput)for(const r of await ne(n)){const a={tag:r,entry:n,resolvedOptions:s.resolvedOptions};await t.callHook("tag:normalise",a),e.tags.push(a.tag)}}return await t.callHook("tags:beforeResolve",e),await t.callHook("tags:resolve",e),await t.callHook("tags:afterResolve",e),e.tags},ssr:n};return[ke,Ce,Ee,_e,Te,Pe,De,Me,...e?.plugins||[]].forEach((e=>s.use(e))),s.hooks.callHook("init",s),s}(e);return t.use((e=>{const t=e.resolvedOptions.document?.head.querySelector('script[id="unhead:payload"]')?.innerHTML||!1;return t&&e.push(JSON.parse(t)),{mode:"client",hooks:{"entries:updated":e=>{!function(e,t={}){const n=t.delayFn||(e=>setTimeout(e,10));e._domDebouncedUpdatePromise=e._domDebouncedUpdatePromise||new Promise((r=>n((()=>async function(e,t={}){const n=t.document||e.resolvedOptions.document;if(!n||!e.dirty)return;const r={shouldRender:!0,tags:[]};return await e.hooks.callHook("dom:beforeRender",r),r.shouldRender?(e._domUpdatePromise||(e._domUpdatePromise=new Promise((async t=>{const r=(await e.resolveTags()).map((e=>({tag:e,id:D.has(e.tag)?B(e):e.tag,shouldRender:!0})));let a=e._dom;if(!a){a={elMap:{htmlAttrs:n.documentElement,bodyAttrs:n.body}};const e=new Set;for(const t of["body","head"]){const r=n[t]?.children;for(const t of r){const n=t.tagName.toLowerCase();if(!D.has(n))continue;const r={tag:n,props:await ee(t.getAttributeNames().reduce(((e,n)=>({...e,[n]:t.getAttribute(n)})),{})),innerHTML:t.innerHTML},o=ce(r);let i=o,s=1;for(;i&&e.has(i);)i=`${o}:${s++}`;i&&(r._d=i,e.add(i)),a.elMap[t.getAttribute("data-hid")||B(r)]=t}}}function o(e,t,n){const r=`${e}:${t}`;a.sideEffects[r]=n,delete a.pendingSideEffects[r]}function i({id:e,$el:t,tag:r}){const i=r.tag.endsWith("Attrs");if(a.elMap[e]=t,i||(r.textContent&&r.textContent!==t.textContent&&(t.textContent=r.textContent),r.innerHTML&&r.innerHTML!==t.innerHTML&&(t.innerHTML=r.innerHTML),o(e,"el",(()=>{a.elMap[e]?.remove(),delete a.elMap[e]}))),r._eventHandlers)for(const e in r._eventHandlers)Object.prototype.hasOwnProperty.call(r._eventHandlers,e)&&""!==t.getAttribute(`data-${e}`)&&(("bodyAttrs"===r.tag?n.defaultView:t).addEventListener(e.substring(2),r._eventHandlers[e].bind(t)),t.setAttribute(`data-${e}`,""));for(const n in r.props){if(!Object.prototype.hasOwnProperty.call(r.props,n))continue;const a=r.props[n],s=`attr:${n}`;if("class"===n){if(!a)continue;for(const n of a.split(" "))i&&o(e,`${s}:${n}`,(()=>t.classList.remove(n))),!t.classList.contains(n)&&t.classList.add(n)}else if("style"===n){if(!a)continue;for(const n of a.split(";")){const r=n.indexOf(":"),a=n.substring(0,r).trim(),i=n.substring(r+1).trim();o(e,`${s}:${a}`,(()=>{t.style.removeProperty(a)})),t.style.setProperty(a,i)}}else t.getAttribute(n)!==a&&t.setAttribute(n,!0===a?"":String(a)),i&&o(e,s,(()=>t.removeAttribute(n)))}}a.pendingSideEffects={...a.sideEffects},a.sideEffects={};const s=[],l={bodyClose:void 0,bodyOpen:void 0,head:void 0};for(const e of r){const{tag:t,shouldRender:r,id:o}=e;r&&("title"!==t.tag?(e.$el=e.$el||a.elMap[o],e.$el?i(e):D.has(t.tag)&&s.push(e)):n.title=t.textContent)}for(const e of s){const t=e.tag.tagPosition||"head";e.$el=n.createElement(e.tag.tag),i(e),l[t]=l[t]||n.createDocumentFragment(),l[t].appendChild(e.$el)}for(const t of r)await e.hooks.callHook("dom:renderTag",t,n,o);l.head&&n.head.appendChild(l.head),l.bodyOpen&&n.body.insertBefore(l.bodyOpen,n.body.firstChild),l.bodyClose&&n.body.appendChild(l.bodyClose);for(const e in a.pendingSideEffects)a.pendingSideEffects[e]();e._dom=a,await e.hooks.callHook("dom:rendered",{renders:r}),t()})).finally((()=>{e._domUpdatePromise=void 0,e.dirty=!1}))),e._domUpdatePromise):void 0}(e,t).then((()=>{delete e._domDebouncedUpdatePromise,r()}))))))}(e,void 0)}}}})),$e=t}function Ie(e,t){return!e||"server"===e&&t||"client"===e&&!t}Symbol("ScriptProxyTarget");const Le="3"===E.rE[0];function Be(e){if(e instanceof Promise||e instanceof Date||e instanceof RegExp)return e;const t="function"==typeof(n=e)?n():(0,S.R1)(n);var n;if(!e||!t)return t;if(Array.isArray(t))return t.map((e=>Be(e)));if("object"==typeof t){const e={};for(const n in t)Object.prototype.hasOwnProperty.call(t,n)&&("titleTemplate"===n||"o"===n[0]&&"n"===n[1]?e[n]=(0,S.R1)(t[n]):e[n]=Be(t[n]));return e}return t}const Fe={hooks:{"entries:resolve":e=>{for(const t of e.entries)t.resolvedInput=Be(t.input)}}},Qe="usehead";const Ue="undefined"!=typeof globalThis?globalThis:"undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self?self:{},ze="__unhead_injection_handler__";function je(e,t={}){const n=t.head||function(){if(ze in Ue)return Ue[ze]();return(0,E.WQ)(Qe)||$e}();if(n)return n.ssr?n.push(e,t):function(e,t,n={}){const r=(0,S.KR)(!1),a=(0,S.KR)({});(0,E.nT)((()=>{a.value=r.value?{}:Be(t)}));const o=e.push(a.value,n);return(0,E.wB)(a,(e=>{o.patch(e)})),(0,E.nI)()&&((0,E.xo)((()=>{o.dispose()})),(0,E.Y4)((()=>{r.value=!0})),(0,E.n)((()=>{r.value=!1}))),o}(n,e,t)}var We=n(3073),Xe=n(4048),qe=n(5633),Ve=n(4074);const He={limit:20},Ze=e=>{let t=0;const n={...He,...e};return{check:e=>"string"==typeof e&&!(!e.startsWith("http://")&&!e.startsWith("https://")),async get(e){if(!1!==n?.limit&&t>=n?.limit)console.warn(`[fetchUrls] Maximum number of requests reeached (${n?.limit}), skipping request`);else try{t++;const r=await(n?.fetch?n.fetch(e):fetch(e));return await r.text()}catch(t){return void console.error("[fetchUrls]",t.message,`(${e})`)}}}};function Ge(e,t){for(const[n,r]of Object.entries(e))null!==r&&"object"==typeof r?(t[n]??(t[n]=new r.__proto__.constructor),Ge(r,t[n])):typeof r<"u"&&(t[n]=r);return t}function Ke(e){return Ge(e??{},{info:{title:"",description:"",termsOfService:"",version:"",license:{name:"",url:""},contact:{email:""}},externalDocs:{description:"",url:""},servers:[],tags:[]})}const Ye=["GET","POST","PUT","HEAD","DELETE","PATCH","OPTIONS","CONNECT","TRACE"],Je="get",et=e=>{if("string"!=typeof e)return console.warn(`Request method is not a string. Using ${Je} as the default.`),Je;const t=e.trim().toUpperCase();return n=t,Ye.includes(n)?t:(console.warn(`${e} is not a valid request method. Using ${Je} as the default.`),Je);var n},tt=e=>{var t,n;let r={};r=e&&"object"==typeof e?structuredClone(e):Ke(),r.tags||(r.tags=[]),r.paths||(r.paths={});const a={};return Object.keys(r.webhooks??{}).forEach((e=>{var t;Object.keys((null==(t=r.webhooks)?void 0:t[e])??{}).forEach((t=>{var n,o,i,s;const l=null==(n=r.webhooks)?void 0:n[e][t];if(l&&!(0,Xe.pH)(l)){if(Array.isArray(l.tags)){const e=null==(o=l.tags)?void 0:o.map((e=>{var t;return null==(t=r.tags)?void 0:t.find((t=>t.name===e))}));if(l.tags=null==e?void 0:e.filter((e=>!(0,Xe.pH)(e))),null!=e&&e.some((e=>(0,Xe.pH)(e))))return}void 0===a[e]&&(a[e]={}),a[e][t]={httpVerb:et(t),path:e,operationId:(null==l?void 0:l.operationId)||e,name:(null==l?void 0:l.summary)||e||"",description:(null==l?void 0:l.description)||"",pathParameters:null==(s=null==(i=r.paths)?void 0:i[e])?void 0:s.parameters,information:{...l}}}}))})),Object.keys((null==(t=r.components)?void 0:t.schemas)??{}).forEach((e=>{var t,n,a,o;(0,Xe.pH)(null==(n=null==(t=r.components)?void 0:t.schemas)?void 0:n[e])&&(null==(o=null==(a=r.components)?void 0:a.schemas)||delete o[e])})),Object.keys(r.paths).forEach((e=>{Object.keys(r.paths[e]).filter((e=>Ye.includes(e.toUpperCase()))).forEach((t=>{var n,a,o,i,s,l;const c=r.paths[e][t];if(void 0===c||(0,Xe.pH)(c))return;const u={httpVerb:et(t),path:e,operationId:c.operationId||e,name:c.summary||e||"",description:c.description||"",information:{...c},pathParameters:null==(a=null==(n=r.paths)?void 0:n[e])?void 0:a.parameters};if(c.tags&&0!==c.tags.length)c.tags.forEach((e=>{var t,n,a;const o=null==(t=r.tags)?void 0:t.findIndex((t=>t.name===e));-1===o&&(null==(n=r.tags)||n.push({name:e,description:""}));const i=-1!==o?o:r.tags.length-1;typeof(null==(a=r.tags[i])?void 0:a.operations)>"u"&&(r.tags[i].operations=[]),r.tags[i].operations.push(u)}));else{null!=(o=r.tags)&&o.find((e=>"default"===e.name))||null==(i=r.tags)||i.push({name:"default",description:"",operations:[]});const e=null==(s=r.tags)?void 0:s.findIndex((e=>"default"===e.name));e>=0&&(null==(l=r.tags[e])||l.operations.push(u))}}))})),r.tags=null==(n=r.tags)?void 0:n.filter((e=>!(0,Xe.pH)(e))),{...r,webhooks:a}};function nt({specConfig:e,proxyUrl:t}){const n=(0,S.KR)(""),r=(0,S.Kh)(Ke()),a=(0,S.KR)(null);return(0,E.wB)((()=>(0,S.BA)(e)),(async e=>{var r;if(e){const a=null==(r=await(async({url:e,content:t},n)=>{if(e){const t=performance.now();try{const r=(0,Xe.AY)(e)?await(0,Xe.tI)(e,n):await(0,Xe.tI)(e),a=performance.now();return console.log(`fetch: ${Math.round(a-t)} ms (${e})`),console.log("size:",Math.round(r.length/1024),"kB"),r}catch(e){console.error("Failed to fetch spec from URL:",e)}}const r="function"==typeof t?t():t;return"string"==typeof r?r:"object"==typeof r?(0,Xe.qX)(r):void 0})(e,(0,S.BA)(t)))?void 0:r.trim();"string"==typeof a&&(n.value=a)}}),{immediate:!0,deep:!0}),(0,E.wB)(n,(()=>{var e;(e=n.value)?((e,{proxyUrl:t}={})=>new Promise((async(n,r)=>{var a;try{if(!e)return n(tt(Ke()));const o=performance.now(),{filesystem:i}=await(0,qe.H)(e,{plugins:[Ze({fetch:e=>fetch(t?(0,Xe.mg)(t,e):e)})]}),{schema:s,errors:l}=await(0,Ve.D)(i),c=performance.now();return console.log(`dereference: ${Math.round(c-o)} ms`),null!=l&&l.length&&console.warn("Please open an issue on https://github.com/scalar/scalar\n","Scalar OpenAPI Parser Warning:\n",l),void 0===s?(r((null==(a=null==l?void 0:l[0])?void 0:a.message)??"Failed to parse the OpenAPI file."),n(tt(Ke()))):n(tt(s))}catch(e){console.error("[@scalar/api-reference]","Failed to parse the OpenAPI document. It might be invalid?"),console.error(e),r(e)}return n(tt(Ke()))})))(e,{proxyUrl:t?(0,S.BA)(t):void 0}).then((e=>{a.value=null,Object.assign(r,{servers:[],...e})})).catch((e=>{a.value=e.toString()})):Object.assign(r,Ke())})),{rawSpec:n,parsedSpec:r,specErrors:a}}var rt=n(3751);const at=Symbol(),ot=()=>(0,E.WQ)(at,{});var it=n(7786);const st={targetKey:"shell",clientKey:"curl"},{clients:lt}=(0,it.E)();function ct(e){var t;return(null==(t=ht.value.find((t=>t.key===e.targetKey)))?void 0:t.title)??e.targetKey}function ut(e){var t,n;return(null==(n=null==(t=ht.value.find((t=>t.key===e.targetKey)))?void 0:t.clients.find((t=>t.client===e.clientKey)))?void 0:n.title)??e.clientKey}const dt=(0,E.EW)((()=>ct(wt))),pt=(0,E.EW)((()=>ut(wt))),ht=(0,E.EW)((()=>{return e=lt(),!0===(t=ft).value?[]:e.flatMap((e=>{var n;return"object"!=typeof t.value?[]:Array.isArray(t.value)?(e.clients=e.clients.filter((e=>!t.value.includes(e.client))),e.clients.length?[e]:[]):!0===t.value[e.key]?[]:(Array.isArray(t.value[e.key])&&(e.clients=e.clients.filter((n=>!t.value[e.key].includes(n.client)))),null!=(n=null==e?void 0:e.clients)&&n.length?[e]:[])}));var e,t})),ft=(0,S.KR)({}),mt=(0,S.KR)();function gt(e){void 0!==e&&(mt.value=e,Ot(vt()))}const vt=()=>{var e,t,n,r;return bt(mt.value)?mt.value:bt(st)?st:{targetKey:null==(e=ht.value[0])?void 0:e.key,clientKey:null==(r=null==(n=null==(t=ht.value[0])?void 0:t.clients)?void 0:n[0])?void 0:r.client}};function bt(e){return void 0!==e&&!!ht.value.find((t=>t.key===e.targetKey&&t.clients.find((t=>t.client===e.clientKey))))}function yt(){(0,Xe.O1)(wt,vt())}const wt=(0,S.Kh)(vt()),Ot=e=>{Object.assign(wt,{...wt,...e})},kt=()=>({httpClient:(0,S.tB)(wt),resetState:yt,setHttpClient:Ot,setDefaultHttpClient:gt,excludedClients:(0,S.tB)(ft),setExcludedClients:e=>{ft.value=e,(0,Xe.O1)(wt,vt())},availableTargets:ht,getClientTitle:ut,getTargetTitle:ct,httpTargetTitle:dt,httpClientTitle:pt});var xt=n(8869),Et=(n(6099),n(6584)),St=(n(264),n(8309)),_t=n(9135),Ct=n(1266),Tt=n(8157),At=n(9271);const Rt={class:"start custom-scroll"},Pt={class:"start-copy"},Dt={class:"start-cta"},Mt={class:"start-row"},$t={class:"start-section start-section-colors"},Nt=["onClick"],It=(e,t)=>{const n=e.__vccOpts||e;for(const[e,r]of t)n[e]=r;return n},Lt=It((0,E.pM)({__name:"GettingStarted",props:{theme:{}},emits:["changeTheme","loadSwaggerFile","linkSwaggerFile","updateContent"],setup(e,{emit:t}){const n=t,r=["default","alternate","moon","purple","solarized","bluePlanet","saturn","kepler","mars","deepSpace"];async function a(){const e=await fetch("https://cdn.jsdelivr.net/npm/@scalar/galaxy/dist/latest.yaml");n("updateContent",await e.text())}return(e,t)=>((0,E.uX)(),(0,E.CE)("div",Rt,[(0,E.Lk)("div",Pt,[t[3]||(t[3]=(0,E.Fv)('<div class="start-logo" data-v-391b4e37><svg height="36" viewBox="0 0 36 36" width="36" xmlns="http://www.w3.org/2000/svg" data-v-391b4e37><path d="M18 0a18 18 0 1 1 0 36 18 18 0 0 1 0-36Zm11.2 6.5c-3.3-3.3-11.1-1-17.4 5.3-6.2 6.3-8.6 14-5.3 17.4 3.3 3.3 11 .9 17.3-5.4 6.3-6.2 8.7-14 5.4-17.3ZM17.6 12a6.3 6.3 0 1 1 0 12.7 6.3 6.3 0 0 1 0-12.7Z" fill="currentColor" fill-rule="evenodd" data-v-391b4e37></path></svg></div><h1 class="start-h1" data-v-391b4e37>Swagger Editor</h1><p class="start-p" data-v-391b4e37> Welcome to the Scalar API References + Swagger Editor, a Free &amp; Open-Source tool that takes your Swagger/OAS file and generates Beautiful API references. </p>',3)),(0,E.Lk)("div",Dt,[(0,E.bF)((0,S.R1)(Ct.RO),{fullWidth:"",onClick:a},{default:(0,E.k6)((()=>t[1]||(t[1]=[(0,E.eW)(" Show Example ")]))),_:1}),(0,E.bF)((0,S.R1)(Ct.RO),{fullWidth:"",variant:"outlined",onClick:t[0]||(t[0]=t=>e.$emit("loadSwaggerFile"))},{default:(0,E.k6)((()=>t[2]||(t[2]=[(0,E.eW)(" Upload File ")]))),_:1})])]),(0,E.Lk)("div",Mt,[t[5]||(t[5]=(0,E.Fv)('<div class="start-section start-section-integrations" data-v-391b4e37><div class="start-h2" data-v-391b4e37>INTEGRATIONS</div><a class="start-item" href="https://github.com/scalar/scalar/tree/main/packages/fastify-api-reference#readme" target="_blank" data-v-391b4e37><svg fill="currentColor" height="16" viewBox="0 0 19 16" width="19" xmlns="http://www.w3.org/2000/svg" data-v-391b4e37><path d="m18.2 3.1.8-2V.9l-4.7 1.3C15.2 1 15 0 15 0s-2.5 1.6-4.3 1.5c-2 0-3.6.8-4 1-1.8 1.2-2.5 3.3-3.2 3.8L0 8.9 2.3 8l-2 2.5c.2.3 1.2 1.6 2.1 1.3l.4-.1 1.6.5-.7-1 .2-.2.9.3-.1-.8.9.3-.1-.8.3-.1 1-3.5 3.7-2.6-.3.7A4 4 0 0 1 8 7l-.6.2c-.5.5-.7.7-.8 2.5a2 2 0 0 1 1 0c1.6.4 2.2 2.3 1.7 2.9l-.7.6H8v.6h-.7v.5l-.2.2c-.7 0-1.4-.6-1.4-.6 0 .5.4 1.3.4 1.3s1.7 1.1 2.7.7c1-.4.7-2.3 2.8-3.2l3.3-.9.8-2.2-1.7.5v-2l2.5-.6.9-2.2-3.4.9v-2l4.2-1.1Z" fill="currentColor" fill-rule="nonzero" data-v-391b4e37></path></svg><span data-v-391b4e37>Fastify</span></a><a class="start-item" href="https://github.com/scalar/scalar/tree/main#from-a-cdn" target="_blank" data-v-391b4e37><svg fill="currentColor" viewBox="0 0 24 24" xmlns="http://www.w3.org/2000/svg" data-v-391b4e37><g data-v-391b4e37><path d="M22.5 1.5h-21A1.5 1.5 0 0 0 0 3v3a1.5 1.5 0 0 0 1.5 1.5h21A1.5 1.5 0 0 0 24 6V3a1.5 1.5 0 0 0-1.5-1.5Zm-19.25 3A1.25 1.25 0 1 1 4.5 5.75 1.25 1.25 0 0 1 3.25 4.5ZM8.5 5.75A1.25 1.25 0 1 1 9.75 4.5 1.25 1.25 0 0 1 8.5 5.75Z" fill="currentColor" data-v-391b4e37></path><path d="M22.5 9h-21A1.5 1.5 0 0 0 0 10.5v3A1.5 1.5 0 0 0 1.5 15h21a1.5 1.5 0 0 0 1.5-1.5v-3A1.5 1.5 0 0 0 22.5 9ZM3.25 12a1.25 1.25 0 1 1 1.25 1.25A1.25 1.25 0 0 1 3.25 12Zm5.25 1.25A1.25 1.25 0 1 1 9.75 12a1.25 1.25 0 0 1-1.25 1.25Z" fill="currentColor" data-v-391b4e37></path><path d="M22.5 16.5h-21A1.5 1.5 0 0 0 0 18v3a1.5 1.5 0 0 0 1.5 1.5h21A1.5 1.5 0 0 0 24 21v-3a1.5 1.5 0 0 0-1.5-1.5Zm-19.25 3a1.25 1.25 0 1 1 1.25 1.25 1.25 1.25 0 0 1-1.25-1.25Zm5.25 1.25a1.25 1.25 0 1 1 1.25-1.25 1.25 1.25 0 0 1-1.25 1.25Z" fill="currentColor" data-v-391b4e37></path></g></svg><span data-v-391b4e37>CDN</span></a><a class="start-item" href="https://github.com/scalar/scalar/tree/main#with-vuejs" target="_blank" data-v-391b4e37><svg height="170" viewBox="0 0 196.3 170" width="196.3" xmlns="http://www.w3.org/2000/svg" data-v-391b4e37><g fill="currentColor" fill-rule="nonzero" data-v-391b4e37><polygon points="39.23 0 0 0 2.9450761 5.1010782 98.16 170.02 196.32 0 157.06 0 98.16 102.01 42.175701 5.0991171" data-v-391b4e37></polygon><polygon points="75.5 2.009956e-14 0 2.009956e-14 2.94 5.1 78.44871 5.1 98.16 39.26 117.87937 5.1 193.38 5.1 196.325 0 120.82 7.8065636e-15 114.97322 2.009956e-14 98.16 29.037153 81.35 2.009956e-14" data-v-391b4e37></polygon></g></svg><span data-v-391b4e37>Vue</span></a><a class="start-item" href="https://github.com/scalar/scalar/tree/main#with-react" target="_blank" data-v-391b4e37><svg height="23.3" viewBox="0 0 22 23.3" width="22" xmlns="http://www.w3.org/2000/svg" data-v-391b4e37><g fill="none" fill-rule="evenodd" data-v-391b4e37><circle cx="11" cy="11.6" fill="currentColor" fill-rule="nonzero" r="2" data-v-391b4e37></circle><g stroke="currentColor" data-v-391b4e37><ellipse cx="11" cy="11.6" rx="11" ry="4.2" data-v-391b4e37></ellipse><ellipse cx="11" cy="11.6" rx="11" ry="4.2" transform="rotate(60 11 11.6)" data-v-391b4e37></ellipse><ellipse cx="11" cy="11.6" rx="11" ry="4.2" transform="rotate(120 11 11.6)" data-v-391b4e37></ellipse></g></g></svg><span data-v-391b4e37>React</span></a></div>',1)),(0,E.Lk)("div",$t,[t[4]||(t[4]=(0,E.Lk)("p",{class:"start-h2"},"THEMING",-1)),((0,E.uX)(),(0,E.CE)(E.FK,null,(0,E.pI)(r,(t=>(0,E.Lk)("div",{key:t,class:(0,_.C4)(["start-item",{"start-item-active":t===e.theme}]),onClick:n=>e.$emit("changeTheme",{id:t,label:(0,S.R1)(C.ty)[t]})},(0,_.v_)((0,S.R1)(C.ty)[t]),11,Nt))),64))])]),t[6]||(t[6]=(0,E.Fv)('<p class="start-h1" data-v-391b4e37>Features</p><ul class="start-ul" data-v-391b4e37><li data-v-391b4e37><p class="start-h3" data-v-391b4e37>Customize</p> Bring your typography &amp; color palettes, or use our themes! </li><li data-v-391b4e37><p class="start-h3" data-v-391b4e37>Testing</p> A deeply integrated Rest API Client (Also Free &amp; Open-Source) </li><li data-v-391b4e37><p class="start-h3" data-v-391b4e37>Search</p> Fully integrated Search (Using fuse.js) </li><li data-v-391b4e37><p class="start-h3" data-v-391b4e37>Hosting</p> Free subdomain hosting on https://apidocumentation.com </li><li data-v-391b4e37><p class="start-h3" data-v-391b4e37>OpenAPI &amp; Swagger</p> Support for OpenAPI 3.1, OpenAPI 3.0, and Swagger 2.0 </li><li data-v-391b4e37><p class="start-h3" data-v-391b4e37>Code Samples</p> Code samples to show off your API in most popular languages </li></ul>',2))]))}}),[["__scopeId","data-v-391b4e37"]]),Bt=Symbol("downloadSpec"),Ft=(0,We.iXd)(Bt),Qt="https://api.scalar.com/request-proxy",Ut="https://proxy.scalar.com",zt=["GET","POST","PUT","HEAD","DELETE","PATCH","OPTIONS","CONNECT","TRACE"],jt={POST:"var(--scalar-color-green)",DELETE:"var(--scalar-color-red)",PATCH:"var(--scalar-color-yellow)",GET:"var(--scalar-color-blue)",PUT:"var(--scalar-color-orange)",OPTIONS:"var(--scalar-color-purple)",HEAD:"var(--scalar-color-2)",CONNECT:"var(--scalar-color-2)",TRACE:"var(--scalar-color-2)"},Wt={POST:"POST",DELETE:"DEL",PATCH:"PATCH",GET:"GET",PUT:"PUT",OPTIONS:"OPTS",HEAD:"HEAD",CONNECT:"CONN",TRACE:"TRACE"};function Xt(e){return zt.includes(e)}const qt=(0,E.pM)({__name:"HttpMethod",props:{as:{},property:{},short:{type:Boolean},method:{}},setup(e){const t=e,n=(0,E.EW)((()=>t.method.trim().toUpperCase())),r=(0,E.EW)((()=>Xt(n.value)?Wt[n.value]:n.value.slice(0,4))),a=(0,E.EW)((()=>Xt(n.value)?jt[n.value]:"var(--scalar-color-ghost)"));return(e,t)=>((0,E.uX)(),(0,E.Wv)((0,E.$y)(e.as??"span"),{style:(0,_.Tr)({[e.property||"color"]:a.value})},{default:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"default",(0,_._B)((0,E.Ng)({normalized:n.value,abbreviated:r.value,color:a.value})),(()=>[(0,E.eW)((0,_.v_)(e.short?r.value:n.value),1)]))])),_:3},8,["style"]))}}),Vt=It((0,E.pM)({__name:"SidebarHttpBadge",props:{active:{type:Boolean},method:{}},setup:e=>(e,t)=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(qt),{class:(0,_.C4)(["sidebar-heading-type",{"sidebar-heading-type-active":e.active}]),method:e.method,property:"--method-color",short:""},null,8,["class","method"]))}),[["__scopeId","data-v-f51e2805"]]),Ht=e=>new Promise((t=>setTimeout(t,e))),Zt=/[\0-\x1F!-,\.\/:-@\[-\^`\{-\xA9\xAB-\xB4\xB6-\xB9\xBB-\xBF\xD7\xF7\u02C2-\u02C5\u02D2-\u02DF\u02E5-\u02EB\u02ED\u02EF-\u02FF\u0375\u0378\u0379\u037E\u0380-\u0385\u0387\u038B\u038D\u03A2\u03F6\u0482\u0530\u0557\u0558\u055A-\u055F\u0589-\u0590\u05BE\u05C0\u05C3\u05C6\u05C8-\u05CF\u05EB-\u05EE\u05F3-\u060F\u061B-\u061F\u066A-\u066D\u06D4\u06DD\u06DE\u06E9\u06FD\u06FE\u0700-\u070F\u074B\u074C\u07B2-\u07BF\u07F6-\u07F9\u07FB\u07FC\u07FE\u07FF\u082E-\u083F\u085C-\u085F\u086B-\u089F\u08B5\u08C8-\u08D2\u08E2\u0964\u0965\u0970\u0984\u098D\u098E\u0991\u0992\u09A9\u09B1\u09B3-\u09B5\u09BA\u09BB\u09C5\u09C6\u09C9\u09CA\u09CF-\u09D6\u09D8-\u09DB\u09DE\u09E4\u09E5\u09F2-\u09FB\u09FD\u09FF\u0A00\u0A04\u0A0B-\u0A0E\u0A11\u0A12\u0A29\u0A31\u0A34\u0A37\u0A3A\u0A3B\u0A3D\u0A43-\u0A46\u0A49\u0A4A\u0A4E-\u0A50\u0A52-\u0A58\u0A5D\u0A5F-\u0A65\u0A76-\u0A80\u0A84\u0A8E\u0A92\u0AA9\u0AB1\u0AB4\u0ABA\u0ABB\u0AC6\u0ACA\u0ACE\u0ACF\u0AD1-\u0ADF\u0AE4\u0AE5\u0AF0-\u0AF8\u0B00\u0B04\u0B0D\u0B0E\u0B11\u0B12\u0B29\u0B31\u0B34\u0B3A\u0B3B\u0B45\u0B46\u0B49\u0B4A\u0B4E-\u0B54\u0B58-\u0B5B\u0B5E\u0B64\u0B65\u0B70\u0B72-\u0B81\u0B84\u0B8B-\u0B8D\u0B91\u0B96-\u0B98\u0B9B\u0B9D\u0BA0-\u0BA2\u0BA5-\u0BA7\u0BAB-\u0BAD\u0BBA-\u0BBD\u0BC3-\u0BC5\u0BC9\u0BCE\u0BCF\u0BD1-\u0BD6\u0BD8-\u0BE5\u0BF0-\u0BFF\u0C0D\u0C11\u0C29\u0C3A-\u0C3C\u0C45\u0C49\u0C4E-\u0C54\u0C57\u0C5B-\u0C5F\u0C64\u0C65\u0C70-\u0C7F\u0C84\u0C8D\u0C91\u0CA9\u0CB4\u0CBA\u0CBB\u0CC5\u0CC9\u0CCE-\u0CD4\u0CD7-\u0CDD\u0CDF\u0CE4\u0CE5\u0CF0\u0CF3-\u0CFF\u0D0D\u0D11\u0D45\u0D49\u0D4F-\u0D53\u0D58-\u0D5E\u0D64\u0D65\u0D70-\u0D79\u0D80\u0D84\u0D97-\u0D99\u0DB2\u0DBC\u0DBE\u0DBF\u0DC7-\u0DC9\u0DCB-\u0DCE\u0DD5\u0DD7\u0DE0-\u0DE5\u0DF0\u0DF1\u0DF4-\u0E00\u0E3B-\u0E3F\u0E4F\u0E5A-\u0E80\u0E83\u0E85\u0E8B\u0EA4\u0EA6\u0EBE\u0EBF\u0EC5\u0EC7\u0ECE\u0ECF\u0EDA\u0EDB\u0EE0-\u0EFF\u0F01-\u0F17\u0F1A-\u0F1F\u0F2A-\u0F34\u0F36\u0F38\u0F3A-\u0F3D\u0F48\u0F6D-\u0F70\u0F85\u0F98\u0FBD-\u0FC5\u0FC7-\u0FFF\u104A-\u104F\u109E\u109F\u10C6\u10C8-\u10CC\u10CE\u10CF\u10FB\u1249\u124E\u124F\u1257\u1259\u125E\u125F\u1289\u128E\u128F\u12B1\u12B6\u12B7\u12BF\u12C1\u12C6\u12C7\u12D7\u1311\u1316\u1317\u135B\u135C\u1360-\u137F\u1390-\u139F\u13F6\u13F7\u13FE-\u1400\u166D\u166E\u1680\u169B-\u169F\u16EB-\u16ED\u16F9-\u16FF\u170D\u1715-\u171F\u1735-\u173F\u1754-\u175F\u176D\u1771\u1774-\u177F\u17D4-\u17D6\u17D8-\u17DB\u17DE\u17DF\u17EA-\u180A\u180E\u180F\u181A-\u181F\u1879-\u187F\u18AB-\u18AF\u18F6-\u18FF\u191F\u192C-\u192F\u193C-\u1945\u196E\u196F\u1975-\u197F\u19AC-\u19AF\u19CA-\u19CF\u19DA-\u19FF\u1A1C-\u1A1F\u1A5F\u1A7D\u1A7E\u1A8A-\u1A8F\u1A9A-\u1AA6\u1AA8-\u1AAF\u1AC1-\u1AFF\u1B4C-\u1B4F\u1B5A-\u1B6A\u1B74-\u1B7F\u1BF4-\u1BFF\u1C38-\u1C3F\u1C4A-\u1C4C\u1C7E\u1C7F\u1C89-\u1C8F\u1CBB\u1CBC\u1CC0-\u1CCF\u1CD3\u1CFB-\u1CFF\u1DFA\u1F16\u1F17\u1F1E\u1F1F\u1F46\u1F47\u1F4E\u1F4F\u1F58\u1F5A\u1F5C\u1F5E\u1F7E\u1F7F\u1FB5\u1FBD\u1FBF-\u1FC1\u1FC5\u1FCD-\u1FCF\u1FD4\u1FD5\u1FDC-\u1FDF\u1FED-\u1FF1\u1FF5\u1FFD-\u203E\u2041-\u2053\u2055-\u2070\u2072-\u207E\u2080-\u208F\u209D-\u20CF\u20F1-\u2101\u2103-\u2106\u2108\u2109\u2114\u2116-\u2118\u211E-\u2123\u2125\u2127\u2129\u212E\u213A\u213B\u2140-\u2144\u214A-\u214D\u214F-\u215F\u2189-\u24B5\u24EA-\u2BFF\u2C2F\u2C5F\u2CE5-\u2CEA\u2CF4-\u2CFF\u2D26\u2D28-\u2D2C\u2D2E\u2D2F\u2D68-\u2D6E\u2D70-\u2D7E\u2D97-\u2D9F\u2DA7\u2DAF\u2DB7\u2DBF\u2DC7\u2DCF\u2DD7\u2DDF\u2E00-\u2E2E\u2E30-\u3004\u3008-\u3020\u3030\u3036\u3037\u303D-\u3040\u3097\u3098\u309B\u309C\u30A0\u30FB\u3100-\u3104\u3130\u318F-\u319F\u31C0-\u31EF\u3200-\u33FF\u4DC0-\u4DFF\u9FFD-\u9FFF\uA48D-\uA4CF\uA4FE\uA4FF\uA60D-\uA60F\uA62C-\uA63F\uA673\uA67E\uA6F2-\uA716\uA720\uA721\uA789\uA78A\uA7C0\uA7C1\uA7CB-\uA7F4\uA828-\uA82B\uA82D-\uA83F\uA874-\uA87F\uA8C6-\uA8CF\uA8DA-\uA8DF\uA8F8-\uA8FA\uA8FC\uA92E\uA92F\uA954-\uA95F\uA97D-\uA97F\uA9C1-\uA9CE\uA9DA-\uA9DF\uA9FF\uAA37-\uAA3F\uAA4E\uAA4F\uAA5A-\uAA5F\uAA77-\uAA79\uAAC3-\uAADA\uAADE\uAADF\uAAF0\uAAF1\uAAF7-\uAB00\uAB07\uAB08\uAB0F\uAB10\uAB17-\uAB1F\uAB27\uAB2F\uAB5B\uAB6A-\uAB6F\uABEB\uABEE\uABEF\uABFA-\uABFF\uD7A4-\uD7AF\uD7C7-\uD7CA\uD7FC-\uD7FF\uE000-\uF8FF\uFA6E\uFA6F\uFADA-\uFAFF\uFB07-\uFB12\uFB18-\uFB1C\uFB29\uFB37\uFB3D\uFB3F\uFB42\uFB45\uFBB2-\uFBD2\uFD3E-\uFD4F\uFD90\uFD91\uFDC8-\uFDEF\uFDFC-\uFDFF\uFE10-\uFE1F\uFE30-\uFE32\uFE35-\uFE4C\uFE50-\uFE6F\uFE75\uFEFD-\uFF0F\uFF1A-\uFF20\uFF3B-\uFF3E\uFF40\uFF5B-\uFF65\uFFBF-\uFFC1\uFFC8\uFFC9\uFFD0\uFFD1\uFFD8\uFFD9\uFFDD-\uFFFF]|\uD800[\uDC0C\uDC27\uDC3B\uDC3E\uDC4E\uDC4F\uDC5E-\uDC7F\uDCFB-\uDD3F\uDD75-\uDDFC\uDDFE-\uDE7F\uDE9D-\uDE9F\uDED1-\uDEDF\uDEE1-\uDEFF\uDF20-\uDF2C\uDF4B-\uDF4F\uDF7B-\uDF7F\uDF9E\uDF9F\uDFC4-\uDFC7\uDFD0\uDFD6-\uDFFF]|\uD801[\uDC9E\uDC9F\uDCAA-\uDCAF\uDCD4-\uDCD7\uDCFC-\uDCFF\uDD28-\uDD2F\uDD64-\uDDFF\uDF37-\uDF3F\uDF56-\uDF5F\uDF68-\uDFFF]|\uD802[\uDC06\uDC07\uDC09\uDC36\uDC39-\uDC3B\uDC3D\uDC3E\uDC56-\uDC5F\uDC77-\uDC7F\uDC9F-\uDCDF\uDCF3\uDCF6-\uDCFF\uDD16-\uDD1F\uDD3A-\uDD7F\uDDB8-\uDDBD\uDDC0-\uDDFF\uDE04\uDE07-\uDE0B\uDE14\uDE18\uDE36\uDE37\uDE3B-\uDE3E\uDE40-\uDE5F\uDE7D-\uDE7F\uDE9D-\uDEBF\uDEC8\uDEE7-\uDEFF\uDF36-\uDF3F\uDF56-\uDF5F\uDF73-\uDF7F\uDF92-\uDFFF]|\uD803[\uDC49-\uDC7F\uDCB3-\uDCBF\uDCF3-\uDCFF\uDD28-\uDD2F\uDD3A-\uDE7F\uDEAA\uDEAD-\uDEAF\uDEB2-\uDEFF\uDF1D-\uDF26\uDF28-\uDF2F\uDF51-\uDFAF\uDFC5-\uDFDF\uDFF7-\uDFFF]|\uD804[\uDC47-\uDC65\uDC70-\uDC7E\uDCBB-\uDCCF\uDCE9-\uDCEF\uDCFA-\uDCFF\uDD35\uDD40-\uDD43\uDD48-\uDD4F\uDD74\uDD75\uDD77-\uDD7F\uDDC5-\uDDC8\uDDCD\uDDDB\uDDDD-\uDDFF\uDE12\uDE38-\uDE3D\uDE3F-\uDE7F\uDE87\uDE89\uDE8E\uDE9E\uDEA9-\uDEAF\uDEEB-\uDEEF\uDEFA-\uDEFF\uDF04\uDF0D\uDF0E\uDF11\uDF12\uDF29\uDF31\uDF34\uDF3A\uDF45\uDF46\uDF49\uDF4A\uDF4E\uDF4F\uDF51-\uDF56\uDF58-\uDF5C\uDF64\uDF65\uDF6D-\uDF6F\uDF75-\uDFFF]|\uD805[\uDC4B-\uDC4F\uDC5A-\uDC5D\uDC62-\uDC7F\uDCC6\uDCC8-\uDCCF\uDCDA-\uDD7F\uDDB6\uDDB7\uDDC1-\uDDD7\uDDDE-\uDDFF\uDE41-\uDE43\uDE45-\uDE4F\uDE5A-\uDE7F\uDEB9-\uDEBF\uDECA-\uDEFF\uDF1B\uDF1C\uDF2C-\uDF2F\uDF3A-\uDFFF]|\uD806[\uDC3B-\uDC9F\uDCEA-\uDCFE\uDD07\uDD08\uDD0A\uDD0B\uDD14\uDD17\uDD36\uDD39\uDD3A\uDD44-\uDD4F\uDD5A-\uDD9F\uDDA8\uDDA9\uDDD8\uDDD9\uDDE2\uDDE5-\uDDFF\uDE3F-\uDE46\uDE48-\uDE4F\uDE9A-\uDE9C\uDE9E-\uDEBF\uDEF9-\uDFFF]|\uD807[\uDC09\uDC37\uDC41-\uDC4F\uDC5A-\uDC71\uDC90\uDC91\uDCA8\uDCB7-\uDCFF\uDD07\uDD0A\uDD37-\uDD39\uDD3B\uDD3E\uDD48-\uDD4F\uDD5A-\uDD5F\uDD66\uDD69\uDD8F\uDD92\uDD99-\uDD9F\uDDAA-\uDEDF\uDEF7-\uDFAF\uDFB1-\uDFFF]|\uD808[\uDF9A-\uDFFF]|\uD809[\uDC6F-\uDC7F\uDD44-\uDFFF]|[\uD80A\uD80B\uD80E-\uD810\uD812-\uD819\uD824-\uD82B\uD82D\uD82E\uD830-\uD833\uD837\uD839\uD83D\uD83F\uD87B-\uD87D\uD87F\uD885-\uDB3F\uDB41-\uDBFF][\uDC00-\uDFFF]|\uD80D[\uDC2F-\uDFFF]|\uD811[\uDE47-\uDFFF]|\uD81A[\uDE39-\uDE3F\uDE5F\uDE6A-\uDECF\uDEEE\uDEEF\uDEF5-\uDEFF\uDF37-\uDF3F\uDF44-\uDF4F\uDF5A-\uDF62\uDF78-\uDF7C\uDF90-\uDFFF]|\uD81B[\uDC00-\uDE3F\uDE80-\uDEFF\uDF4B-\uDF4E\uDF88-\uDF8E\uDFA0-\uDFDF\uDFE2\uDFE5-\uDFEF\uDFF2-\uDFFF]|\uD821[\uDFF8-\uDFFF]|\uD823[\uDCD6-\uDCFF\uDD09-\uDFFF]|\uD82C[\uDD1F-\uDD4F\uDD53-\uDD63\uDD68-\uDD6F\uDEFC-\uDFFF]|\uD82F[\uDC6B-\uDC6F\uDC7D-\uDC7F\uDC89-\uDC8F\uDC9A-\uDC9C\uDC9F-\uDFFF]|\uD834[\uDC00-\uDD64\uDD6A-\uDD6C\uDD73-\uDD7A\uDD83\uDD84\uDD8C-\uDDA9\uDDAE-\uDE41\uDE45-\uDFFF]|\uD835[\uDC55\uDC9D\uDCA0\uDCA1\uDCA3\uDCA4\uDCA7\uDCA8\uDCAD\uDCBA\uDCBC\uDCC4\uDD06\uDD0B\uDD0C\uDD15\uDD1D\uDD3A\uDD3F\uDD45\uDD47-\uDD49\uDD51\uDEA6\uDEA7\uDEC1\uDEDB\uDEFB\uDF15\uDF35\uDF4F\uDF6F\uDF89\uDFA9\uDFC3\uDFCC\uDFCD]|\uD836[\uDC00-\uDDFF\uDE37-\uDE3A\uDE6D-\uDE74\uDE76-\uDE83\uDE85-\uDE9A\uDEA0\uDEB0-\uDFFF]|\uD838[\uDC07\uDC19\uDC1A\uDC22\uDC25\uDC2B-\uDCFF\uDD2D-\uDD2F\uDD3E\uDD3F\uDD4A-\uDD4D\uDD4F-\uDEBF\uDEFA-\uDFFF]|\uD83A[\uDCC5-\uDCCF\uDCD7-\uDCFF\uDD4C-\uDD4F\uDD5A-\uDFFF]|\uD83B[\uDC00-\uDDFF\uDE04\uDE20\uDE23\uDE25\uDE26\uDE28\uDE33\uDE38\uDE3A\uDE3C-\uDE41\uDE43-\uDE46\uDE48\uDE4A\uDE4C\uDE50\uDE53\uDE55\uDE56\uDE58\uDE5A\uDE5C\uDE5E\uDE60\uDE63\uDE65\uDE66\uDE6B\uDE73\uDE78\uDE7D\uDE7F\uDE8A\uDE9C-\uDEA0\uDEA4\uDEAA\uDEBC-\uDFFF]|\uD83C[\uDC00-\uDD2F\uDD4A-\uDD4F\uDD6A-\uDD6F\uDD8A-\uDFFF]|\uD83E[\uDC00-\uDFEF\uDFFA-\uDFFF]|\uD869[\uDEDE-\uDEFF]|\uD86D[\uDF35-\uDF3F]|\uD86E[\uDC1E\uDC1F]|\uD873[\uDEA2-\uDEAF]|\uD87A[\uDFE1-\uDFFF]|\uD87E[\uDE1E-\uDFFF]|\uD884[\uDF4B-\uDFFF]|\uDB40[\uDC00-\uDCFF\uDDF0-\uDFFF]/g,Gt=Object.hasOwnProperty;class Kt{constructor(){this.occurrences,this.reset()}slug(e,t){const n=this;let r=Yt(e,!0===t);const a=r;for(;Gt.call(n.occurrences,r);)n.occurrences[a]++,r=a+"-"+n.occurrences[a];return n.occurrences[r]=0,r}reset(){this.occurrences=Object.create(null)}}function Yt(e,t){return"string"!=typeof e?"":(t||(e=e.toLowerCase()),e.replace(Zt,"").replace(/ /g,"-"))}const Jt=(0,S.KR)(""),en=(0,S.KR)(Xe.QJ.hash??""),tn=(0,S.KR)(),nn=(0,S.KR)(!1),rn=e=>{var t;if(!tn.value)return"";const n=new RegExp("^"+(null==(t=tn.value)?void 0:t.basePath)+"/?");return decodeURIComponent(e.replace(n,""))},an=(e=en.value)=>{var t;const n=null==(t=e.match(/(tag\/[^/]+)/))?void 0:t[0],r=e.startsWith("model")?"models":"",a=e.startsWith("webhook")?"webhooks":"";return n||r||a},on=()=>{en.value=tn.value?rn(window.location.pathname):decodeURIComponent(window.location.hash.replace(/^#/,"")).slice(Jt.value.length)},sn=(e,t=window.location.href)=>{const n=new URL(t);tn.value?n.pathname=(0,Xe.Dk)(tn.value.basePath,e):n.hash=Jt.value+e,en.value=e,window.history.replaceState({},"",n)},ln=(e,t=window.location.href,n=window.location.search)=>{const r=new URL(t);return r.hash=Jt.value+e,r.search=n,r.toString()},cn=(e=en.value)=>`${Jt.value}${e}`,un=()=>decodeURIComponent(window.location.hash.replace(/^#/,"").slice(Jt.value.length)),dn=()=>{const e=ot(),t=t=>"function"==typeof(null==e?void 0:e.generateTagSlug)?`tag/${e.generateTagSlug(t)}`:`tag/${Yt(t.name)}`;return{hash:en,setHashPrefix:e=>{Jt.value=e},getFullHash:cn,getHashedUrl:ln,replaceUrlState:sn,getReferenceHash:un,getWebhookId:t=>null!=t&&t.name?"function"==typeof(null==e?void 0:e.generateWebhookSlug)?`webhook/${e.generateWebhookSlug(t)}`:`webhook/${t.method}/${Yt(t.name)}`:"webhooks",getModelId:t=>null!=t&&t.name?"function"==typeof(null==e?void 0:e.generateModelSlug)?`model/${e.generateModelSlug(t)}`:`model/${Yt(t.name)}`:"models",getHeadingId:t=>"function"==typeof(null==e?void 0:e.generateHeadingSlug)?`${e.generateHeadingSlug(t)}`:t.slug?`description/${t.slug}`:"",getOperationId:(n,r)=>{var a;return"function"==typeof(null==e?void 0:e.generateOperationSlug)?`${t(r)}/${e.generateOperationSlug({path:n.path,operationId:n.operationId,method:n.httpVerb,summary:null==(a=n.information)?void 0:a.summary})}`:`${t(r)}/${n.httpVerb}${n.path}`},getPathRoutingId:rn,getSectionId:an,getTagId:t,isIntersectionEnabled:nn,pathRouting:tn,updateHash:on}},pn=async e=>{var t;null==(t=document.getElementById(e))||t.scrollIntoView()},hn=["id"],fn={key:0,class:"sidebar-heading-chevron"},mn=["aria-expanded"],gn=["href","tabindex"],vn={class:"sidebar-heading-link-title"},bn={key:1,class:"sidebar-heading-link-method"},yn={key:1,class:"action-menu"},wn=It((0,E.pM)({__name:"SidebarElement",props:{id:{},item:{},isActive:{type:Boolean},hasChildren:{type:Boolean},open:{type:Boolean}},emits:["toggleOpen"],setup(e,{emit:t}){const n=e,r=t,{getFullHash:a,isIntersectionEnabled:o,pathRouting:i,replaceUrlState:s}=dn(),l=async()=>{var e,t;n.hasChildren&&r("toggleOpen"),null==(t=null==(e=n.item)?void 0:e.select)||t.call(e),n.open&&(o.value=!1,await Ht(100),o.value=!0)},c=()=>{if(i.value)return(0,Xe.Dk)(i.value.basePath,n.item.id);if(typeof window<"u"){const e=new URL(window.location.href);return e.hash=a(n.item.id),`${e.pathname}${e.search}${e.hash}`}return`#${a(n.item.id)}`},u=async e=>{var t,a;i.value&&(e.preventDefault(),n.hasChildren&&r("toggleOpen"),null==(a=null==(t=n.item)?void 0:t.select)||a.call(t),r("toggleOpen"),o.value=!1,s(n.item.id),pn(n.item.id),await Ht(100),o.value=!0)};return(e,t)=>{var n,r;return(0,E.uX)(),(0,E.CE)("li",{id:e.id,class:"sidebar-group-item"},[(0,E.Lk)("div",{class:(0,_.C4)(["sidebar-heading",{"sidebar-group-item__folder":e.hasChildren,active_page:e.isActive,deprecated:e.item.deprecated??!1}]),onClick:l},[e.hasChildren?((0,E.uX)(),(0,E.CE)("p",fn,[(0,E.Lk)("button",{"aria-expanded":e.open,class:"toggle-nested-icon",type:"button",onClick:(0,rt.D$)(l,["stop"])},[(0,E.bF)((0,S.R1)(Ct.SL),{open:e.open},{label:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(e.open?"Collapse":"Expand")+" "+(0,_.v_)(e.item.title),1)])),_:1},8,["open"])],8,mn),t[0]||(t[0]=(0,E.eW)("   "))])):(0,E.Q3)("",!0),(0,E.Lk)("a",{class:"sidebar-heading-link",href:c(),tabindex:e.hasChildren?-1:0,onClick:u},[null!=(r=null==(n=e.item)?void 0:n.icon)&&r.src?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.kv),{key:0,class:"sidebar-icon",icon:e.item.icon.src},null,8,["icon"])):(0,E.Q3)("",!0),(0,E.Lk)("p",vn,(0,_.v_)(e.item.title),1),e.item.httpVerb&&!e.hasChildren?((0,E.uX)(),(0,E.CE)("p",bn,[t[1]||(t[1]=(0,E.eW)("   ")),t[2]||(t[2]=(0,E.Lk)("span",{class:"sr-only"},"HTTP Method: ",-1)),(0,E.bF)(Vt,{active:e.isActive,method:e.item.httpVerb},null,8,["active","method"])])):(0,E.Q3)("",!0)],8,gn)],2),e.open?(0,E.RG)(e.$slots,"default",{key:0},void 0,!0):(0,E.Q3)("",!0),e.$slots["action-menu"]?((0,E.uX)(),(0,E.CE)("div",yn,[(0,E.RG)(e.$slots,"action-menu",{},void 0,!0)])):(0,E.Q3)("",!0)],8,hn)}}}),[["__scopeId","data-v-18e6dc3d"]]),On=It((0,E.pM)({__name:"SidebarGroup",props:{level:{}},setup:e=>(e,t)=>((0,E.uX)(),(0,E.CE)("ul",{class:"sidebar-group sidebar-indent-nested",style:(0,_.Tr)({"--scalar-sidebar-level":e.level})},[(0,E.RG)(e.$slots,"default",{},void 0,!0)],4))}),[["__scopeId","data-v-39c84840"]]);var kn,xn;function En(e){var t,n,r;return void 0!==(null==(t=e.information)?void 0:t.deprecated)?null==(n=e.information)?void 0:n.deprecated:!(null==(r=e.information)||!r["x-scalar-stability"]||e.information["x-scalar-stability"]!==kn.Deprecated)}function Sn(e){var t,n;return null!=(t=e.information)&&t.deprecated?kn.Deprecated:null==(n=e.information)?void 0:n["x-scalar-stability"]}function _n(e){const t=Sn(e);return t===kn.Deprecated?"text-red":t===kn.Experimental?"text-orange":t===kn.Stable?"text-green":""}(xn=kn||(kn={})).Deprecated="deprecated",xn.Experimental="experimental",xn.Stable="stable";const Cn=Symbol(),Tn=(0,We.iXd)(Cn),An=e=>{const t=Math.min(...e.map((e=>e.depth)));return t>=1&&t<=6?t:1};function Rn(e){var t,n;if(!e)return{};const r=Object.keys((null==(t=null==e?void 0:e.components)?void 0:t.schemas)??{}).length?null==(n=null==e?void 0:e.components)?void 0:n.schemas:Object.keys((null==e?void 0:e.definitions)??{}).length?null==e?void 0:e.definitions:{};return Object.keys(r??{}).forEach((e=>{var t;!0===(null==(t=r[e])?void 0:t["x-internal"])&&delete r[e]})),r}const Pn=e=>!!e&&!!Object.keys(Rn(e)??{}).length;var Dn=n(9261);function Mn(e){const t=new Kt;return((e,t)=>e.map((e=>({...e,slug:t.slug(e.value)}))))((0,Dn.cf)(e),t)}const $n=(0,S.KR)(null),Nn=(0,S.KR)(void 0),In=(0,S.Kh)({});function Ln(e){var t,n,r;let a;return"alpha"===In.tagsSorter?e.tags=null==(t=e.tags)?void 0:t.sort(((e,t)=>e.name.localeCompare(t.name))):"function"==typeof In.tagsSorter&&(e.tags=null==(n=e.tags)?void 0:n.sort(In.tagsSorter)),"alpha"===In.operationsSorter?a=(e,t)=>{const n=e.name??e.path,r=t.name??t.path;return n.localeCompare(r)}:"method"===In.operationsSorter?a=(e,t)=>e.httpVerb.localeCompare(t.httpVerb):"function"==typeof In.operationsSorter&&(a=In.operationsSorter),a&&(null==(r=e.tags)||r.forEach((e=>{var t;e.operations=null==(t=e.operations)?void 0:t.sort(a)}))),Nn.value=e}const Bn=(0,S.KR)(!1),Fn=(0,S.KR)(!1),Qn=(0,S.Kh)(Xe.QJ["useSidebarContent-collapsedSidebarItems"]??{});function Un(e){Qn[e]=!Qn[e]}function zn(e,t){Qn[e]=t}const jn=(0,S.KR)([]),Wn=(0,E.EW)((()=>{var e,t,n,r,a,o,i,s,l,c;if(!$n.value)return{entries:[],titles:{}};const{getHeadingId:u,getModelId:d,getOperationId:p,getTagId:h,getWebhookId:f}=$n.value,m={},g=[];let v=null;jn.value.forEach((e=>{var t;e.depth===An(jn.value)?(v={id:u(e),title:e.value,show:!0,children:[]},g.push(v)):v&&(null==(t=v.children)||t.push({id:u(e),title:e.value,show:!0}))}));const b=null==(t=null==(e=Nn.value)?void 0:e.tags)?void 0:t[0],y=b&&(1!==(null==(w=null==(n=Nn.value)?void 0:n.tags)?void 0:w.length)||"default"!==w[0].name||""!==w[0].description)?null==(a=null==(r=Nn.value)?void 0:r.tags)?void 0:a.filter((e=>{var t;return(null==(t=e.operations)?void 0:t.length)>0})).map((e=>{var t;return{id:h(e),title:e.name,displayTitle:e["x-displayName"]??e.name,show:!0,children:null==(t=e.operations)?void 0:t.map((t=>{const n=p(t,e),r=t.name??t.path;return m[n]=r,{id:n,title:r,httpVerb:t.httpVerb,deprecated:En(t),show:!0,select:()=>{}}}))}})):null==(o=null==b?void 0:b.operations)?void 0:o.map((e=>{const t=p(e,b),n=e.name??e.path;return m[t]=n,{id:t,title:n,httpVerb:e.httpVerb,deprecated:En(e),show:!0,select:()=>{}}}));var w;let O=Pn(Nn.value)&&!Bn.value?[{id:d(),title:"Models",show:!0,children:Object.keys(Rn(Nn.value)??{}).map((e=>{var t;const n=d({name:e});return m[n]=e,{id:n,title:(null==(t=Rn(Nn.value))?void 0:t[e]).title??e,show:!0}}))}]:[],k=(e=>!!e&&!!Object.keys((null==e?void 0:e.webhooks)??{}).length)(Nn.value)?[{id:f(),title:"Webhooks",show:!0,children:Object.keys((null==(i=Nn.value)?void 0:i.webhooks)??{}).map((e=>{var t,n;const r=f({name:e});return m[r]=e,Object.keys((null==(n=null==(t=Nn.value)?void 0:t.webhooks)?void 0:n[e])??{}).map((t=>{var n,r,a;return{id:f({name:e,method:t}),title:null==(a=null==(r=null==(n=Nn.value)?void 0:n.webhooks)?void 0:r[e][t])?void 0:a.name,httpVerb:t,show:!0}}))})).flat()}]:[];const x=null!=(s=Nn.value)&&s["x-tagGroups"]?null==(c=null==(l=Nn.value)?void 0:l["x-tagGroups"])?void 0:c.map((e=>{var t;const n=[];return null==(t=e.tags)||t.map((e=>{if("models"===e&&O.length>0)n.push(O[0]),O=[];else if("webhooks"===e&&k.length>0)n.push(k[0]),k=[];else{const t=null==y?void 0:y.find((t=>t.title===e));t&&n.push(t)}})),{id:e.name,title:e.name,children:n,show:!0,isGroup:!0}})):void 0,E=[...g,...x??y??[],...k,...O];return Fn.value&&E.forEach((e=>{zn(e.id,!0),e.show=!0})),{entries:E,titles:m}})),Xn=(0,S.KR)(!1),qn=(0,E.EW)((()=>{var e,t,n;return(null==(n=null==(e=Wn.value)?void 0:e.titles)?void 0:n[(null==(t=$n.value)?void 0:t.hash)??""])??""})),Vn=e=>{var t;const n=null==(t=$n.value)?void 0:t.getSectionId(e);if(n&&n!==e)if(Qn[n])pn(e);else{const t=Tn.on((n=>{n.id===e&&(pn(e),t())}));zn(n,!0)}};function Hn(e){Object.assign(In,e);const t=dn();$n.value=t;const{hash:n,getSectionId:r,getTagId:a}=t;return null!=e&&e.parsedSpec&&(Ln(e.parsedSpec),(0,E.wB)((()=>{var e,t;return null==(t=null==(e=Nn.value)?void 0:e.tags)?void 0:t.length}),(()=>{var e,t;if(n.value){const e=r(n.value);e&&zn(e,!0)}else{const n=null==(t=null==(e=Nn.value)?void 0:e.tags)?void 0:t[0];n&&zn(a(n),!0)}})),(0,E.wB)((()=>{var e,t;return null==(t=null==(e=Nn.value)?void 0:e.info)?void 0:t.description}),(()=>{var e,t;const n=null==(t=null==(e=Nn.value)?void 0:e.info)?void 0:t.description;return jn.value=n?function(e){const t=Mn(e),n=An(t);return t.filter((e=>e.depth===n||e.depth===n+1))}(n):[]}),{immediate:!0})),{breadcrumb:qn,items:Wn,isSidebarOpen:Xn,collapsedSidebarItems:Qn,toggleCollapsedSidebarItem:Un,setCollapsedSidebarItem:zn,hideModels:Bn,setParsedSpec:Ln,defaultOpenAllTags:Fn,scrollToOperation:Vn}}const Zn={class:"sidebar"},Gn=["aria-label"],Kn={class:"sidebar-group-title"},Yn=It((0,E.pM)({__name:"Sidebar",props:{parsedSpec:{},tagsSorter:{type:[String,Function]},operationsSorter:{type:[String,Function]}},setup(e){const t=e,{hash:n,isIntersectionEnabled:r}=dn(),{items:a,toggleCollapsedSidebarItem:o,collapsedSidebarItems:i}=Hn({parsedSpec:t.parsedSpec,tagsSorter:t.tagsSorter,operationsSorter:t.operationsSorter}),s=(0,S.KR)(null),l=(0,S.KR)(!0);(0,E.wB)(n,(e=>{!r.value||l.value||typeof window>"u"||c(e)}));const c=e=>{var t,n,r,a;const o=document.getElementById(`sidebar-${e}`);if(!o||!s.value)return;let i=-160;"heading"===o.getAttribute("data-sidebar-type")?i+=o.offsetTop+(null==(t=o.getElementsByClassName("sidebar-heading"))?void 0:t[0]).offsetHeight:i+=o.offsetTop+((null==(n=o.parentElement)?void 0:n.offsetTop)??0)+((null==(a=null==(r=o.parentElement)?void 0:r.parentElement)?void 0:a.offsetTop)??0),s.value.scrollTo({top:i,behavior:"smooth"})};return(0,E.sV)((()=>{setTimeout((()=>c(n.value)),500),l.value=!1})),(e,t)=>{var r;return(0,E.uX)(),(0,E.CE)("div",Zn,[(0,E.RG)(e.$slots,"sidebar-start",{},void 0,!0),(0,E.Lk)("nav",{ref_key:"scrollerEl",ref:s,"aria-label":`Contents of ${null==(r=e.parsedSpec.info)?void 0:r.title}`,class:"sidebar-pages custom-scroll custom-scroll-self-contain-overflow"},[(0,E.bF)(On,{level:0},{default:(0,E.k6)((()=>[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)((0,S.R1)(a).entries,(e=>((0,E.uX)(),(0,E.CE)(E.FK,{key:e.id},[e.isGroup?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[(0,E.Lk)("li",Kn,(0,_.v_)(e.displayTitle??e.title),1),((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(e.children,(t=>((0,E.uX)(),(0,E.Wv)(wn,{key:t.id,id:`sidebar-${t.id}`,"data-sidebar-type":"heading",hasChildren:t.children&&t.children.length>0,isActive:(0,S.R1)(n)===t.id,item:{id:t.id,title:t.displayTitle??t.title,select:t.select,httpVerb:t.httpVerb,deprecated:t.deprecated??!1},open:(0,S.R1)(i)[t.id]??!1,onToggleOpen:async()=>{l.value=!0,(0,S.R1)(o)(t.id),await(0,S.R1)(Ht)(100),l.value=!1}},{default:(0,E.k6)((()=>{var r;return[t.children&&(null==(r=t.children)?void 0:r.length)>0?((0,E.uX)(),(0,E.Wv)(On,{key:0,level:1},{default:(0,E.k6)((()=>[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(t.children,(t=>((0,E.uX)(),(0,E.CE)(E.FK,{key:t.id},[e.show?((0,E.uX)(),(0,E.Wv)(wn,{key:0,id:`sidebar-${t.id}`,isActive:(0,S.R1)(n)===t.id,item:{id:t.id,title:t.displayTitle??t.title,select:t.select,httpVerb:t.httpVerb,deprecated:t.deprecated??!1}},null,8,["id","isActive","item"])):(0,E.Q3)("",!0)],64)))),128))])),_:2},1024)):(0,E.Q3)("",!0)]})),_:2},1032,["id","hasChildren","isActive","item","open","onToggleOpen"])))),128))],64)):((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[e.show?((0,E.uX)(),(0,E.Wv)(wn,{key:0,id:`sidebar-${e.id}`,"data-sidebar-type":"heading",hasChildren:e.children&&e.children.length>0,isActive:(0,S.R1)(n)===e.id,item:{id:e.id,title:e.displayTitle??e.title,select:e.select,httpVerb:e.httpVerb,deprecated:e.deprecated??!1},open:(0,S.R1)(i)[e.id]??!1,onToggleOpen:async()=>{l.value=!0,(0,S.R1)(o)(e.id),await(0,S.R1)(Ht)(100),l.value=!1}},{default:(0,E.k6)((()=>{var t;return[e.children&&(null==(t=e.children)?void 0:t.length)>0?((0,E.uX)(),(0,E.Wv)(On,{key:0,level:1},{default:(0,E.k6)((()=>[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(e.children,(t=>((0,E.uX)(),(0,E.CE)(E.FK,{key:t.id},[e.show?((0,E.uX)(),(0,E.Wv)(wn,{key:0,id:`sidebar-${t.id}`,isActive:(0,S.R1)(n)===t.id,item:{id:t.id,title:t.displayTitle??t.title,select:t.select,httpVerb:t.httpVerb,deprecated:t.deprecated??!1}},null,8,["id","isActive","item"])):(0,E.Q3)("",!0)],64)))),128))])),_:2},1024)):(0,E.Q3)("",!0)]})),_:2},1032,["id","hasChildren","isActive","item","open","onToggleOpen"])):(0,E.Q3)("",!0)],64))],64)))),128))])),_:1})],8,Gn),(0,E.RG)(e.$slots,"sidebar-end",{},void 0,!0)])}}}),[["__scopeId","data-v-b863366c"]]);n(9794);var Jn=n(1527);const er=It((0,E.pM)({__name:"ShowMoreButton",props:{id:{}},setup(e){const{setCollapsedSidebarItem:t}=Hn();return(e,n)=>((0,E.uX)(),(0,E.CE)("button",{class:"show-more",type:"button",onClick:n[0]||(n[0]=n=>(0,S.R1)(t)(e.id,!0))},[n[1]||(n[1]=(0,E.eW)(" Show More ")),(0,E.bF)((0,S.R1)(Ct.kv),{class:"show-more-icon",icon:"ChevronDown"})]))}}),[["__scopeId","data-v-252cd03d"]]),tr={key:0,class:"screenreader-only"},nr=It((0,E.pM)({__name:"ScreenReader",props:{if:{type:Boolean,default:!0}},setup:e=>(e,t)=>e.$props.if?((0,E.uX)(),(0,E.CE)("span",tr,[(0,E.RG)(e.$slots,"default",{},void 0,!0)])):(0,E.RG)(e.$slots,"default",{key:1},void 0,!0)}),[["__scopeId","data-v-df2e1026"]]),rr=(0,E.pM)({__name:"IntersectionObserver",props:{id:{},is:{}},emits:["intersecting"],setup(e,{emit:t}){const n=e,r=t,a=(0,S.KR)(),o=e=>{const t=e.offsetHeight;return`${t/2}px 0px ${t/2}px 0px`},i=e=>e.offsetHeight<window.innerHeight?.8:.5;return(0,E.sV)((()=>{if(a.value){const e={rootMargin:o(a.value),threshold:i(a.value)};(0,We.BLq)(a,(([{isIntersecting:e}])=>{e&&n.id&&r("intersecting")}),e)}})),(e,t)=>((0,E.uX)(),(0,E.Wv)((0,E.$y)(e.is??"div"),{id:e.id,ref_key:"intersectionObserverRef",ref:a},{default:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"default")])),_:3},8,["id"]))}}),ar=It((0,E.pM)({__name:"Section",props:{id:{},label:{}},setup(e){const t=e,{getSectionId:n,isIntersectionEnabled:r,replaceUrlState:a}=dn(),{setCollapsedSidebarItem:o}=Hn();function i(){var e,i;!t.label||!r.value||(a(t.id??""),(null!=(e=t.id)&&e.startsWith("model")||null!=(i=t.id)&&i.startsWith("webhook"))&&o(n(t.id),!0))}return(e,t)=>((0,E.uX)(),(0,E.Wv)(rr,{is:"section",id:e.id,class:"section",onIntersecting:i},{default:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"default",{},void 0,!0)])),_:3},8,["id"]))}}),[["__scopeId","data-v-64ccd510"]]),or=It((0,E.pM)({__name:"LoadingSkeleton",props:{lines:{default:1}},setup:e=>(e,t)=>((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)([...Array(e.lines).keys()],(t=>((0,E.uX)(),(0,E.CE)("div",{key:t,class:(0,_.C4)(["loading",{"single-line":1===e.lines}])},null,2)))),128))}),[["__scopeId","data-v-c90b2c46"]]),ir={class:"section-header-wrapper"},sr=It((0,E.pM)({__name:"SectionHeader",props:{loading:{type:Boolean,default:!1},tight:{type:Boolean,default:!1},level:{default:1}},setup:e=>(e,t)=>((0,E.uX)(),(0,E.CE)("div",ir,[e.loading?((0,E.uX)(),(0,E.Wv)(or,{key:0})):((0,E.uX)(),(0,E.Wv)((0,E.$y)(`h${e.level}`),{key:1,class:(0,_.C4)(["section-header",{tight:e.tight}])},{default:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"default",{},void 0,!0)])),_:3},8,["class"]))]))}),[["__scopeId","data-v-b6aef53d"]]);var lr=n(3806);const cr={class:"label"},ur={class:"anchor"},dr=It((0,E.pM)({__name:"Anchor",props:{id:{}},setup(e){const{copyToClipboard:t}=(0,lr.i)(),{getHashedUrl:n}=dn();return(e,r)=>((0,E.uX)(),(0,E.CE)("span",cr,[(0,E.RG)(e.$slots,"default",{},void 0,!0),(0,E.Lk)("span",ur,[r[4]||(r[4]=(0,E.Lk)("span",null,"​",-1)),(0,E.Lk)("button",{class:"anchor-copy",type:"button",onClick:r[0]||(r[0]=(0,rt.D$)((r=>(0,S.R1)(t)((e=>n(e))(e.id))),["stop"]))},[r[3]||(r[3]=(0,E.Lk)("span",{"aria-hidden":"true"},"#",-1)),(0,E.bF)(nr,null,{default:(0,E.k6)((()=>[r[1]||(r[1]=(0,E.eW)('Copy link to "')),(0,E.RG)(e.$slots,"default",{},void 0,!0),r[2]||(r[2]=(0,E.eW)('"'))])),_:3})])])]))}}),[["__scopeId","data-v-3cb2dd80"]]),pr={class:"section-content"},hr=It((0,E.pM)({__name:"SectionContent",props:{loading:{type:Boolean,default:!1}},setup:e=>(e,t)=>((0,E.uX)(),(0,E.CE)("div",pr,[e.loading?((0,E.uX)(),(0,E.Wv)(or,{key:1,lines:8})):(0,E.RG)(e.$slots,"default",{key:0},void 0,!0)]))}),[["__scopeId","data-v-9735459e"]]),fr={class:"section-columns"},mr=It({},[["render",function(e,t){return(0,E.uX)(),(0,E.CE)("div",fr,[(0,E.RG)(e.$slots,"default",{},void 0,!0)])}],["__scopeId","data-v-8b9602bf"]]),gr={class:"section-column"},vr=It({},[["render",function(e,t){return(0,E.uX)(),(0,E.CE)("div",gr,[(0,E.RG)(e.$slots,"default",{},void 0,!0)])}],["__scopeId","data-v-699c28e3"]]),br={class:"scalar-card"},yr=It({},[["render",function(e,t){return(0,E.uX)(),(0,E.CE)("div",br,[(0,E.RG)(e.$slots,"default",{},void 0,!0)])}],["__scopeId","data-v-38801681"]]),wr=It((0,E.pM)({__name:"CardContent",props:{muted:{type:Boolean},contrast:{type:Boolean},frameless:{type:Boolean},transparent:{type:Boolean},borderless:{type:Boolean}},setup:e=>(e,t)=>((0,E.uX)(),(0,E.CE)("div",{class:(0,_.C4)(["scalar-card-content",{"scalar-card--muted":e.muted,"scalar-card--contrast":e.contrast,"scalar-card--frameless":e.frameless,"scalar-card--transparent":e.transparent,"scalar-card--borderless":e.borderless}])},[(0,E.RG)(e.$slots,"default",{},void 0,!0)],2))}),[["__scopeId","data-v-dd83e46e"]]),Or={class:"scalar-card-header-slots"},kr={class:"scalar-card-header-slot scalar-card-header-title"},xr={class:"scalar-card-header-slot scalar-card-header-actions"},Er=It((0,E.pM)({__name:"CardHeader",props:{muted:{type:Boolean},contrast:{type:Boolean},frameless:{type:Boolean},transparent:{type:Boolean},borderless:{type:Boolean}},setup(e){const t=e;return(e,n)=>((0,E.uX)(),(0,E.Wv)(wr,(0,E.v6)(t,{class:"scalar-card-header"}),{default:(0,E.k6)((()=>[(0,E.Lk)("div",Or,[(0,E.Lk)("div",kr,[(0,E.RG)(e.$slots,"default",{},void 0,!0)]),(0,E.Lk)("div",xr,[(0,E.RG)(e.$slots,"actions",{},void 0,!0)])])])),_:3},16))}}),[["__scopeId","data-v-838dc259"]]),Sr=["aria-label"],_r=["href","onClick"],Cr=It((0,E.pM)({__name:"Endpoints",props:{id:{},tag:{},isCollapsed:{type:Boolean}},setup(e){const t=e,{getOperationId:n,getTagId:r}=dn(),{scrollToOperation:a}=Hn(),o=(0,E.EW)((()=>t.tag["x-displayName"]??t.tag.name));return(e,i)=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(ar),{id:e.id,label:e.tag.name.toUpperCase()},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(sr),{level:2},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(dr),{id:(0,S.R1)(r)(e.tag)},{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(o.value),1)])),_:1},8,["id"])])),_:1}),(0,E.bF)((0,S.R1)(hr),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(mr),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(vr),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Ct.mi),{clamp:!!e.isCollapsed&&"7",value:e.tag.description,withImages:""},null,8,["clamp","value"])])),_:1}),(0,E.bF)((0,S.R1)(vr),null,{default:(0,E.k6)((()=>{var r;return[(null==(r=e.tag.operations)?void 0:r.length)>0?((0,E.uX)(),(0,E.Wv)((0,S.R1)(yr),{key:0,class:"scalar-card-sticky"},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Er),{muted:""},{default:(0,E.k6)((()=>[(0,E.bF)(nr,null,{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(o.value),1)])),_:1}),i[0]||(i[0]=(0,E.eW)(" Endpoints "))])),_:1}),(0,E.bF)((0,S.R1)(wr),{class:"custom-scroll",muted:""},{default:(0,E.k6)((()=>[(0,E.Lk)("ul",{"aria-label":`${o.value} endpoints`,class:"endpoints"},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(e.tag.operations,(r=>((0,E.uX)(),(0,E.CE)("li",{key:(0,S.R1)(n)(r,e.tag),class:"contents"},[(0,E.Lk)("a",{class:"endpoint",href:`#${(0,S.R1)(n)(r,e.tag)}`,onClick:(0,rt.D$)((e=>(async e=>{const r=n(e,t.tag);a(r)})(r)),["prevent"])},[(0,E.bF)((0,S.R1)(qt),{method:r.httpVerb},null,8,["method"]),(0,E.Lk)("span",{class:(0,_.C4)(["endpoint-path",{deprecated:(0,S.R1)(En)(r)}])},(0,_.v_)(r.path),3)],8,_r)])))),128))],8,Sr)])),_:1})])),_:1})):(0,E.Q3)("",!0)]})),_:1})])),_:1})])),_:1})])),_:1},8,["id","label"]))}}),[["__scopeId","data-v-cfef28a1"]]),Tr={class:"section-container"},Ar=It({},[["render",function(e,t){return(0,E.uX)(),(0,E.CE)("div",Tr,[(0,E.RG)(e.$slots,"default",{},void 0,!0)])}],["__scopeId","data-v-1913755a"]]),Rr=It((0,E.pM)({__name:"Tag",props:{id:{},tag:{},spec:{}},setup(e){const t=e,n=(0,S.KR)(),r=(0,S.KR)(),{collapsedSidebarItems:a}=Hn(),{getTagId:o}=dn(),i=(0,E.EW)((()=>{var e,n;return(null==(e=t.spec.tags)?void 0:e.length)&&(null==(n=t.spec.tags)?void 0:n.length)>1})),s=(0,E.EW)((()=>{var e,n;return i.value||"default"!==(null==(e=t.tag)?void 0:e.name)||""!==(null==(n=t.tag)?void 0:n.description)}));async function l(){var e,t;await(0,E.dY)(),null==(t=null==(e=r.value)?void 0:e.querySelector("button"))||t.focus()}return(e,t)=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ar),{ref_key:"sectionContainerRef",ref:n,class:"tag-section-container"},{default:(0,E.k6)((()=>[s.value?((0,E.uX)(),(0,E.Wv)(Cr,{key:0,id:e.id,isCollapsed:!(0,S.R1)(a)[(0,S.R1)(o)(e.tag)],tag:e.tag},null,8,["id","isCollapsed","tag"])):(0,E.Q3)("",!0),!(0,S.R1)(a)[(0,S.R1)(o)(e.tag)]&&i.value?((0,E.uX)(),(0,E.Wv)(er,{key:1,id:e.id??"","aria-label":`Show all ${e.tag["x-displayName"]??e.tag.name} endpoints`,onClick:l},null,8,["id","aria-label"])):((0,E.uX)(),(0,E.CE)("div",{key:2,ref_key:"contentsRef",ref:r,class:"contents"},[(0,E.RG)(e.$slots,"default",{},void 0,!0)],512))])),_:3},512))}}),[["__scopeId","data-v-d47a803a"]]);function Pr(e){e.unshift("#");const t=e.map((e=>e.trim().replace(/~/g,"~0").replace(/\//g,"~1"))).filter(Boolean).join("/");if("#"===t)throw new Error("Path cannot be empty. Please provide a path to the operation you want to display.");return t}var Dr=n(7427);const Mr={key:0},$r=It((0,E.pM)({__name:"OperationPath",props:{path:{},deprecated:{type:Boolean}},setup(e){const t=e,n=e=>e.startsWith("{")&&e.endsWith("}"),r=(0,E.EW)((()=>t.path.split(/({[^}]+})/)));return(e,t)=>((0,E.uX)(),(0,E.CE)("span",{class:(0,_.C4)(["operation-path",{deprecated:e.deprecated}])},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(r.value,((e,t)=>((0,E.uX)(),(0,E.CE)(E.FK,{key:t},[n(e)?((0,E.uX)(),(0,E.CE)("em",Mr,(0,_.v_)(e),1)):((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[(0,E.eW)((0,_.v_)(e),1)],64))],64)))),128))],2))}}),[["__scopeId","data-v-ec6c8861"]]);var Nr=n(6251);const Ir=["value"],Lr=["value"],Br=It((0,E.pM)({__name:"ContentTypeSelect",props:{requestBody:{},defaultValue:{}},emits:["selectContentType"],setup(e,{emit:t}){const n=e,r=t,a=(0,E.EW)((()=>{var e;return null!=(e=n.requestBody)&&e.content?Object.keys(n.requestBody.content):[]})),o=(0,S.KR)(n.defaultValue||a.value[0]);return(e,t)=>{var i;return(0,E.uX)(),(0,E.CE)("div",{class:(0,_.C4)(["content-type-select",{"content-type-no-select":a.value.length<=1}])},[(0,E.Lk)("span",null,(0,_.v_)(o.value),1),null!=n&&n.requestBody&&a.value.length>1?((0,E.uX)(),(0,E.CE)("select",{key:0,value:o.value,onChange:t[0]||(t[0]=e=>(e=>{const t=e.target.value;o.value=t,r("selectContentType",{contentType:t})})(e))},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(null==(i=n.requestBody)?void 0:i.content,((e,t)=>((0,E.uX)(),(0,E.CE)("option",{key:t,value:t},(0,_.v_)(t),9,Lr)))),128))],40,Ir)):(0,E.Q3)("",!0)],2)}}}),[["__scopeId","data-v-4b712394"]]);function Fr(e){return Array.isArray(e)?`[${e.map((e=>"string"==typeof e?`"${e.toString().trim()}"`:"object"==typeof e?JSON.stringify(e):void 0===e?"undefined":null===e?"null":e)).join(", ")}]`:"object"==typeof e?JSON.stringify(e):null===e?"null":void 0===e?"undefined":"string"==typeof e?e.trim():e.toString().trim()}const Qr=["oneOf","anyOf","allOf","not"],Ur={key:0,class:"schema-type"},zr=["title"],jr=It((0,E.pM)({__name:"SchemaHeading",props:{value:{type:[Object,Boolean]},name:{}},setup:e=>(e,t)=>{var n,r,a,o,i,s;return"object"==typeof e.value?((0,E.uX)(),(0,E.CE)("span",Ur,[(0,E.Lk)("span",{class:"schema-type-icon",title:"string"==typeof e.value.type?e.value.type:Array.isArray(e.value.type)?e.value.type.join(" | "):"unkown type"},["object"===e.value.type?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[(0,E.eW)(" {} ")],64)):(0,E.Q3)("",!0),"array"===e.value.type?((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[(0,E.eW)(" [] ")],64)):(0,E.Q3)("",!0),e.value.enum?((0,E.uX)(),(0,E.CE)(E.FK,{key:2},[(0,E.eW)(" enum ")],64)):(0,E.Q3)("",!0)],8,zr),null!=(r=null==(n=e.value)?void 0:n.xml)&&r.name&&"##default"!==(null==(o=null==(a=e.value)?void 0:a.xml)?void 0:o.name)?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[(0,E.eW)(" <"+(0,_.v_)(null==(s=null==(i=e.value)?void 0:i.xml)?void 0:s.name)+" /> ",1)],64)):e.name?((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[(0,E.eW)((0,_.v_)(e.name),1)],64)):((0,E.uX)(),(0,E.CE)(E.FK,{key:2},[(0,E.eW)((0,_.v_)(e.value.type),1)],64))])):(0,E.Q3)("",!0)}}),[["__scopeId","data-v-2a52eb0b"]]),Wr={key:0,class:"schema-card-description"},Xr=It((0,E.pM)({__name:"Schema",props:{value:{type:[Object,Boolean]},level:{default:0},name:{},compact:{type:Boolean},noncollapsible:{type:Boolean},hideHeading:{type:Boolean}},setup(e){const t=e,n=(0,E.EW)((()=>!(t.noncollapsible||0===t.level))),r=e=>t.noncollapsible&&e.stopPropagation();return(e,t)=>"object"==typeof e.value&&Object.keys(e.value).length?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Nr.EN),{key:0,defaultOpen:e.noncollapsible},{default:(0,E.k6)((({open:t})=>{var a;return[(0,E.Lk)("div",{class:(0,_.C4)(["schema-card",[`schema-card--level-${e.level}`,{"schema-card--compact":e.compact,"schema-card--open":t}]])},[null!=(a=e.value)&&a.description&&"string"==typeof e.value.description?((0,E.uX)(),(0,E.CE)("div",Wr,[(0,E.bF)((0,S.R1)(Ct.mi),{value:e.value.description},null,8,["value"])])):(0,E.Q3)("",!0),(0,E.Lk)("div",{class:(0,_.C4)(["schema-properties",{"schema-properties-open":t}])},[(0,E.bo)((0,E.bF)((0,S.R1)(Nr._q),{as:e.noncollapsible?"div":"button",class:(0,_.C4)(["schema-card-title",{"schema-card-title--compact":e.compact}]),style:(0,_.Tr)({top:`calc(var(--refs-header-height) +  calc(var(--schema-title-height) * ${e.level}))`}),onClickCapture:r},{default:(0,E.k6)((()=>{var r,a,o;return[e.compact?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[n.value?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.kv),{key:0,class:(0,_.C4)(["schema-card-title-icon",{"schema-card-title-icon--open":t}]),icon:"Add",size:"sm"},null,8,["class"])):(0,E.Q3)("",!0),t?((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[(0,E.eW)(" Hide "+(0,_.v_)((null==(r=e.value)?void 0:r.title)??"Child Attributes"),1)],64)):((0,E.uX)(),(0,E.CE)(E.FK,{key:2},[(0,E.eW)(" Show "+(0,_.v_)((null==(a=e.value)?void 0:a.title)??"Child Attributes"),1)],64))],64)):((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[n.value?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.kv),{key:0,class:(0,_.C4)(["schema-card-title-icon",{"schema-card-title-icon--open":t}]),icon:"Add",size:"sm"},null,8,["class"])):(0,E.Q3)("",!0),(0,E.bF)(jr,{name:(null==(o=e.value)?void 0:o.title)??e.name,value:e.value},null,8,["name","value"])],64))]})),_:2},1032,["as","class","style"]),[[rt.aG,!(e.hideHeading||e.noncollapsible&&e.compact)]]),(0,E.bF)((0,S.R1)(Nr.kS),{static:e.noncollapsible},{default:(0,E.k6)((()=>{var t,n;return[e.value.properties||e.value.additionalProperties||e.value.patternProperties?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[e.value.properties?((0,E.uX)(!0),(0,E.CE)(E.FK,{key:0},(0,E.pI)(Object.keys(null==(t=e.value)?void 0:t.properties),(t=>{var n,r,a,o;return(0,E.uX)(),(0,E.Wv)(Ea,{key:t,compact:e.compact,level:e.level,name:t,required:(null==(n=e.value.required)?void 0:n.includes(t))||!0===(null==(a=null==(r=e.value.properties)?void 0:r[t])?void 0:a.required),value:null==(o=e.value.properties)?void 0:o[t]},null,8,["compact","level","name","required","value"])})),128)):(0,E.Q3)("",!0),e.value.patternProperties?((0,E.uX)(!0),(0,E.CE)(E.FK,{key:1},(0,E.pI)(Object.keys(null==(n=e.value)?void 0:n.patternProperties),(t=>{var n;return(0,E.uX)(),(0,E.Wv)(Ea,{key:t,compact:e.compact,level:e.level,name:t,pattern:"",value:null==(n=e.value.patternProperties)?void 0:n[t]},null,8,["compact","level","name","value"])})),128)):(0,E.Q3)("",!0),e.value.additionalProperties?((0,E.uX)(),(0,E.CE)(E.FK,{key:2},[!0!==e.value.additionalProperties&&0!==Object.keys(e.value.additionalProperties).length&&e.value.additionalProperties.type?((0,E.uX)(),(0,E.Wv)(Ea,{key:1,additional:"",compact:e.compact,level:e.level,noncollapsible:"",value:e.value.additionalProperties},null,8,["compact","level","value"])):((0,E.uX)(),(0,E.Wv)(Ea,{key:0,additional:"",compact:e.compact,level:e.level,noncollapsible:"",value:{type:"anything",..."object"==typeof e.value.additionalProperties?e.value.additionalProperties:{}}},null,8,["compact","level","value"]))],64)):(0,E.Q3)("",!0)],64)):((0,E.uX)(),(0,E.Wv)(Ea,{key:1,compact:e.compact,level:e.level,name:e.value.name,value:e.value},null,8,["compact","level","name","value"]))]})),_:1},8,["static"])],2)],2)]})),_:1},8,["defaultOpen"])):(0,E.Q3)("",!0)}}),[["__scopeId","data-v-f48d0197"]]),qr={key:0,class:"property-detail-prefix"},Vr={key:1,class:"property-detail-value"},Hr={key:2,class:"property-detail-value"},Zr=It((0,E.pM)({__name:"SchemaPropertyDetail",props:{truncate:{type:Boolean},code:{type:Boolean}},setup:e=>(e,t)=>((0,E.uX)(),(0,E.CE)("span",{class:(0,_.C4)(["property-detail",{"property-detail-truncate":e.truncate}])},[e.$slots.prefix?((0,E.uX)(),(0,E.CE)("div",qr,[(0,E.RG)(e.$slots,"prefix",{},void 0,!0),t[0]||(t[0]=(0,E.eW)("  "))])):(0,E.Q3)("",!0),e.code?((0,E.uX)(),(0,E.CE)("code",Vr,[(0,E.RG)(e.$slots,"default",{},void 0,!0)])):((0,E.uX)(),(0,E.CE)("span",Hr,[(0,E.RG)(e.$slots,"default",{},void 0,!0)]))],2))}),[["__scopeId","data-v-3ae98a1b"]]),Gr={class:"badge"},Kr=It({},[["render",function(e,t){return(0,E.uX)(),(0,E.CE)("div",Gr,[(0,E.RG)(e.$slots,"default",{},void 0,!0)])}],["__scopeId","data-v-190574ac"]]),Yr={class:"property-heading"},Jr={key:0,class:"property-name"},ea={key:1,class:"property-additional"},ta={key:2,class:"property-pattern"},na={key:3,class:"property-deprecated"},ra={key:4,class:"property-const"},aa={key:7,class:"property-write-only"},oa={key:8,class:"property-read-only"},ia={key:9,class:"property-required"},sa=It((0,E.pM)({__name:"SchemaPropertyHeading",props:{value:{},enum:{type:Boolean},required:{type:Boolean,default:!1},additional:{type:Boolean},pattern:{type:Boolean}},setup(e){const t=["oneOf","anyOf","allOf","not"],n=e=>Array.isArray(null==e?void 0:e.default)&&1===e.default.length?e.default[0]:null==e?void 0:e.default;return(e,r)=>{var a,o,i,s,l,c,u,d;return(0,E.uX)(),(0,E.CE)("div",Yr,[e.$slots.name?((0,E.uX)(),(0,E.CE)("div",Jr,[e.pattern?((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[r[0]||(r[0]=(0,E.eW)("/")),(0,E.RG)(e.$slots,"name",{},void 0,!0),r[1]||(r[1]=(0,E.eW)("/"))],64)):(0,E.RG)(e.$slots,"name",{key:0},void 0,!0)])):(0,E.Q3)("",!0),e.additional?((0,E.uX)(),(0,E.CE)("div",ea,[null!=(a=e.value)&&a["x-additionalPropertiesName"]?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[(0,E.eW)((0,_.v_)(e.value["x-additionalPropertiesName"]),1)],64)):((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[(0,E.eW)("additional properties")],64))])):(0,E.Q3)("",!0),e.pattern?((0,E.uX)(),(0,E.CE)("div",ta,[(0,E.bF)((0,S.R1)(Kr),null,{default:(0,E.k6)((()=>r[2]||(r[2]=[(0,E.eW)("pattern")]))),_:1})])):(0,E.Q3)("",!0),null!=(o=e.value)&&o.deprecated?((0,E.uX)(),(0,E.CE)("div",na,[(0,E.bF)((0,S.R1)(Kr),null,{default:(0,E.k6)((()=>r[3]||(r[3]=[(0,E.eW)("deprecated")]))),_:1})])):(0,E.Q3)("",!0),null!=(i=e.value)&&i.const||null!=(s=e.value)&&s.enum&&1===e.value.enum.length?((0,E.uX)(),(0,E.CE)("div",ra,[(0,E.bF)(Zr,{truncate:""},{prefix:(0,E.k6)((()=>r[4]||(r[4]=[(0,E.eW)("const:")]))),default:(0,E.k6)((()=>[(0,E.eW)(" "+(0,_.v_)(e.value.const??e.value.enum[0]),1)])),_:1})])):null!=(l=e.value)&&l.type?((0,E.uX)(),(0,E.CE)(E.FK,{key:5},[(0,E.bF)(Zr,null,{default:(0,E.k6)((()=>{var t,n,r;return[null!=(n=null==(t=e.value)?void 0:t.items)&&n.type?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[(0,E.eW)((0,_.v_)(e.value.type)+" "+(0,_.v_)(e.value.items.type)+"[] ",1)],64)):((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[(0,E.eW)((0,_.v_)(Array.isArray(e.value.type)?e.value.type.join(" | "):e.value.type)+" "+(0,_.v_)(null!=(r=e.value)&&r.nullable?" | nullable":""),1)],64)),e.value.minItems||e.value.maxItems?((0,E.uX)(),(0,E.CE)(E.FK,{key:2},[(0,E.eW)((0,_.v_)(e.value.minItems)+"…"+(0,_.v_)(e.value.maxItems),1)],64)):(0,E.Q3)("",!0)]})),_:1}),e.value.minLength?((0,E.uX)(),(0,E.Wv)(Zr,{key:0},{prefix:(0,E.k6)((()=>r[5]||(r[5]=[(0,E.eW)("min:")]))),default:(0,E.k6)((()=>[(0,E.eW)(" "+(0,_.v_)(e.value.minLength),1)])),_:1})):(0,E.Q3)("",!0),e.value.maxLength?((0,E.uX)(),(0,E.Wv)(Zr,{key:1},{prefix:(0,E.k6)((()=>r[6]||(r[6]=[(0,E.eW)("max:")]))),default:(0,E.k6)((()=>[(0,E.eW)(" "+(0,_.v_)(e.value.maxLength),1)])),_:1})):(0,E.Q3)("",!0),e.value.uniqueItems?((0,E.uX)(),(0,E.Wv)(Zr,{key:2},{default:(0,E.k6)((()=>r[7]||(r[7]=[(0,E.eW)(" unique! ")]))),_:1})):(0,E.Q3)("",!0),e.value.format?((0,E.uX)(),(0,E.Wv)(Zr,{key:3},{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(e.value.format),1)])),_:1})):(0,E.Q3)("",!0),void 0!==e.value.minimum&&e.value.exclusiveMinimum?((0,E.uX)(),(0,E.Wv)(Zr,{key:4},{prefix:(0,E.k6)((()=>r[8]||(r[8]=[(0,E.eW)("greater than:")]))),default:(0,E.k6)((()=>[(0,E.eW)(" "+(0,_.v_)(e.value.minimum),1)])),_:1})):(0,E.Q3)("",!0),void 0===e.value.minimum||e.value.exclusiveMinimum?(0,E.Q3)("",!0):((0,E.uX)(),(0,E.Wv)(Zr,{key:5},{prefix:(0,E.k6)((()=>r[9]||(r[9]=[(0,E.eW)("min:")]))),default:(0,E.k6)((()=>[(0,E.eW)(" "+(0,_.v_)(e.value.minimum),1)])),_:1})),void 0!==e.value.maximum&&e.value.exclusiveMaximum?((0,E.uX)(),(0,E.Wv)(Zr,{key:6},{prefix:(0,E.k6)((()=>r[10]||(r[10]=[(0,E.eW)("less than:")]))),default:(0,E.k6)((()=>[(0,E.eW)(" "+(0,_.v_)(e.value.maximum),1)])),_:1})):(0,E.Q3)("",!0),void 0===e.value.maximum||e.value.exclusiveMaximum?(0,E.Q3)("",!0):((0,E.uX)(),(0,E.Wv)(Zr,{key:7},{prefix:(0,E.k6)((()=>r[11]||(r[11]=[(0,E.eW)("max:")]))),default:(0,E.k6)((()=>[(0,E.eW)(" "+(0,_.v_)(e.value.maximum),1)])),_:1})),e.value.pattern?((0,E.uX)(),(0,E.Wv)(Zr,{key:8,code:"",truncate:""},{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(e.value.pattern),1)])),_:1})):(0,E.Q3)("",!0),e.$props.enum?((0,E.uX)(),(0,E.Wv)(Zr,{key:9},{default:(0,E.k6)((()=>r[12]||(r[12]=[(0,E.eW)("enum")]))),_:1})):(0,E.Q3)("",!0),void 0!==e.value.default?((0,E.uX)(),(0,E.Wv)(Zr,{key:10,truncate:""},{prefix:(0,E.k6)((()=>r[13]||(r[13]=[(0,E.eW)("default:")]))),default:(0,E.k6)((()=>[(0,E.eW)(" "+(0,_.v_)(n(e.value)),1)])),_:1})):(0,E.Q3)("",!0)],64)):((0,E.uX)(),(0,E.CE)(E.FK,{key:6},[!0===(null==(c=e.value)?void 0:c.nullable)?((0,E.uX)(),(0,E.Wv)(Zr,{key:0},{default:(0,E.k6)((()=>r[14]||(r[14]=[(0,E.eW)(" nullable ")]))),_:1})):(0,E.Q3)("",!0)],64)),null!=(u=e.value)&&u.writeOnly?((0,E.uX)(),(0,E.CE)("div",aa," write-only ")):null!=(d=e.value)&&d.readOnly?((0,E.uX)(),(0,E.CE)("div",oa," read-only ")):(0,E.Q3)("",!0),((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(t.filter((t=>{var n,r,a;return(null==(n=e.value)?void 0:n[t])||(null==(a=null==(r=e.value)?void 0:r.items)?void 0:a[t])})),(t=>{var n,r;return(0,E.uX)(),(0,E.CE)(E.FK,{key:t},[(null==(r=null==(n=e.value)?void 0:n[t])?void 0:r.length)>1?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Kr),{key:0},{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(t),1)])),_:2},1024)):(0,E.Q3)("",!0)],64)})),128)),e.required?((0,E.uX)(),(0,E.CE)("div",ia," required ")):(0,E.Q3)("",!0)])}}}),[["__scopeId","data-v-dd95f4d7"]]),la={key:0,class:"property-description"},ca={key:1,class:"property-description"},ua={key:2,class:"property-example custom-scroll"},da={class:"property-example-value"},pa={key:3,class:"property-example custom-scroll"},ha={class:"property-example-label"},fa={key:4,class:"property-enum"},ma={key:0,class:"property-list"},ga={class:"property-heading"},va={class:"property-name"},ba={class:"property-description"},ya={key:1,class:"property-enum-values"},wa={key:5,class:"children"},Oa={key:0,class:"children"},ka={key:0,class:"property-rule"},xa={key:1,class:"property-rule"},Ea=It((0,E.pM)({__name:"SchemaProperty",props:{value:{},level:{default:0},name:{},required:{type:Boolean,default:!1},compact:{type:Boolean,default:!1},description:{},additional:{type:Boolean},pattern:{type:Boolean},withExamples:{type:Boolean,default:!0}},setup(e){const t=e,n={integer:{_default:"Integer numbers.",int32:"Signed 32-bit integers (commonly used integer type).",int64:"Signed 64-bit integers (long type)."},string:{date:"full-date notation as defined by RFC 3339, section 5.6, for example, 2017-07-21","date-time":"the date-time notation as defined by RFC 3339, section 5.6, for example, 2017-07-21T17:32:28Z",password:"a hint to UIs to mask the input",byte:"base64-encoded characters, for example, U3dhZ2dlciByb2Nrcw==",binary:"binary data, used to describe files"}},r=(e,t)=>null!=t&&t.properties||null!=t&&t.additionalProperties||null!=t&&t.patternProperties?null:e||(null==t?void 0:t.description)||null,a=e=>e&&n[e.type]?n[e.type][e.format||"_default"]:null,o=e=>{var t;return(null==e?void 0:e.enum)||(null==(t=null==e?void 0:e.items)?void 0:t.enum)||[]},i=(0,E.EW)((()=>o(u.value).length>9)),s=(0,E.EW)((()=>i.value?5:9)),l=(0,E.EW)((()=>o(u.value).slice(0,s.value))),c=(0,E.EW)((()=>o(u.value).slice(s.value))),u=(0,E.EW)((()=>function(e){var t;if(!e||"object"!=typeof e)return e;let n={...e};const r=Qr.find((e=>null==n?void 0:n[e]));if(!r||"not"===r)return n;const a=null==n?void 0:n[r];if(!Array.isArray(a))return n;a.some((e=>"null"===e.type))&&(n.nullable=!0);const o=a.filter((e=>"null"!==e.type));return 1===o.length&&null!=n&&n[r]?(n={...n,...o[0]},null==n||delete n[r],n):(Array.isArray(null==n?void 0:n[r])&&(null==(t=null==n?void 0:n[r])?void 0:t.length)>1&&(n[r]=o),n)}(t.value))),d=Qr.find((e=>!(!u.value||"object"!=typeof u.value)&&(e in u.value||u.value.items&&"object"==typeof u.value.items&&e in u.value.items)));return(e,t)=>{var n,s,p,h,f,m,g,v,b,y,w,O,k,x,C,T;return(0,E.uX)(),(0,E.CE)("div",{class:(0,_.C4)(["property",[`property--level-${e.level}`,{"property--compact":e.compact,"property--deprecated":null==(n=u.value)?void 0:n.deprecated}]])},[(0,E.bF)(sa,{additional:e.additional,enum:o(u.value).length>0,pattern:e.pattern,required:e.required,value:u.value},(0,E.eX)({_:2},[e.name?{name:"name",fn:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(e.name),1)])),key:"0"}:void 0,null!=(s=u.value)&&s.example?{name:"example",fn:(0,E.k6)((()=>[(0,E.eW)(" Example: "+(0,_.v_)(u.value.example),1)])),key:"1"}:void 0]),1032,["additional","enum","pattern","required","value"]),r(e.description,u.value)?((0,E.uX)(),(0,E.CE)("div",la,[(0,E.bF)((0,S.R1)(Ct.mi),{value:r(e.description,u.value)},null,8,["value"])])):a(u.value)?((0,E.uX)(),(0,E.CE)("div",ca,[(0,E.bF)((0,S.R1)(Ct.mi),{value:a(u.value)||""},null,8,["value"])])):(0,E.Q3)("",!0),e.withExamples&&(null!=(p=u.value)&&p.example||null!=(f=null==(h=u.value)?void 0:h.items)&&f.example)?((0,E.uX)(),(0,E.CE)("div",ua,[t[0]||(t[0]=(0,E.Lk)("span",{class:"property-example-label"},"Example",-1)),(0,E.Lk)("code",da,(0,_.v_)((0,S.R1)(Fr)((null==(m=u.value)?void 0:m.example)||(0,S.R1)(d)&&(null==(g=u.value)?void 0:g.items)&&"object"==typeof u.value.items&&u.value.items[(0,S.R1)(d)])),1)])):(0,E.Q3)("",!0),null!=(v=u.value)&&v.examples&&"object"==typeof u.value.examples&&Object.keys(u.value.examples).length>0?((0,E.uX)(),(0,E.CE)("div",pa,[(0,E.Lk)("span",ha,(0,_.v_)(1===Object.keys(u.value.examples).length?"Example":"Examples"),1),((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(u.value.examples,((e,t)=>((0,E.uX)(),(0,E.CE)("code",{key:t,class:"property-example-value"},(0,_.v_)(e),1)))),128))])):(0,E.Q3)("",!0),(null==(b=o(u.value))?void 0:b.length)>0?((0,E.uX)(),(0,E.CE)("div",fa,[Array.isArray(null==(y=u.value)?void 0:y["x-enumDescriptions"])?((0,E.uX)(),(0,E.CE)("div",ma,[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(o(u.value),(e=>((0,E.uX)(),(0,E.CE)("div",{key:e,class:"property"},[(0,E.Lk)("div",ga,[(0,E.Lk)("div",va,(0,_.v_)(e),1)]),(0,E.Lk)("div",ba,[(0,E.bF)((0,S.R1)(Ct.mi),{value:u.value["x-enumDescriptions"][e]},null,8,["value"])])])))),128))])):((0,E.uX)(),(0,E.CE)("ul",ya,[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(l.value,(e=>((0,E.uX)(),(0,E.CE)("li",{key:e,class:"property-enum-value"},(0,_.v_)(e),1)))),128)),i.value?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Nr.EN),{key:0},{default:(0,E.k6)((({open:e})=>[(0,E.bF)((0,S.R1)(Nr.kS),null,{default:(0,E.k6)((()=>[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(c.value,(e=>((0,E.uX)(),(0,E.CE)("li",{key:e,class:"property-enum-value"},(0,_.v_)(e),1)))),128))])),_:1}),(0,E.bF)((0,S.R1)(Nr._q),{class:"enum-toggle-button"},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Ct.kv),{class:(0,_.C4)(["enum-toggle-button-icon",{"enum-toggle-button-icon--open":e}]),icon:"Add",size:"sm"},null,8,["class"]),(0,E.eW)(" "+(0,_.v_)(e?"Hide values":"Show all values"),1)])),_:2},1024)])),_:1})):(0,E.Q3)("",!0)]))])):(0,E.Q3)("",!0),"object"===(null==(w=u.value)?void 0:w.type)&&(null!=(O=u.value)&&O.properties||null!=(k=u.value)&&k.additionalProperties)?((0,E.uX)(),(0,E.CE)("div",wa,[(0,E.bF)(Xr,{compact:e.compact,level:e.level+1,value:u.value},null,8,["compact","level","value"])])):(0,E.Q3)("",!0),null!=(x=u.value)&&x.items&&"object"==typeof u.value.items&&"type"in u.value.items&&"string"==typeof u.value.items.type?((0,E.uX)(),(0,E.CE)(E.FK,{key:6},[["object"].includes(null==(T=null==(C=u.value)?void 0:C.items)?void 0:T.type)?((0,E.uX)(),(0,E.CE)("div",Oa,[(0,E.bF)(Xr,{compact:e.compact,level:e.level+1,value:u.value.items},null,8,["compact","level","value"])])):(0,E.Q3)("",!0)],64)):(0,E.Q3)("",!0),((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)((0,S.R1)(Qr),(t=>{var n,r;return(0,E.uX)(),(0,E.CE)(E.FK,{key:t},[null!=(n=u.value)&&n[t]?((0,E.uX)(),(0,E.CE)("div",ka,[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(u.value[t],(n=>{var r,a;return(0,E.uX)(),(0,E.Wv)(Xr,{key:n.id,compact:e.compact,level:e.level+1,noncollapsible:Array.isArray(null==(r=u.value)?void 0:r[t])&&1===(null==(a=u.value)?void 0:a[t].length),value:n},null,8,["compact","level","noncollapsible","value"])})),128))])):(0,E.Q3)("",!0),null!=(r=u.value)&&r.items&&"string"==typeof t&&"object"==typeof u.value.items&&t in u.value.items&&Array.isArray(u.value.items[t])&&e.level<3?((0,E.uX)(),(0,E.CE)("div",xa,[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(u.value.items[t],(t=>((0,E.uX)(),(0,E.Wv)(Xr,{key:t.id,compact:e.compact,level:e.level+1,value:t},null,8,["compact","level","value"])))),128))])):(0,E.Q3)("",!0)],64)})),128))],2)}}}),[["__scopeId","data-v-f5c211dc"]]),Sa=It((0,E.pM)({__name:"ParameterHeaders",props:{headers:{}},setup(e){function t(e){return void 0!==e.schema}return(e,n)=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Nr.EN),null,{default:(0,E.k6)((({open:n})=>[(0,E.Lk)("div",{class:(0,_.C4)(["headers-card headers-card--compact",[{"headers-card--open":n}]])},[(0,E.Lk)("div",{class:(0,_.C4)(["headers-properties",{"headers-properties-open":n}])},[(0,E.bF)((0,S.R1)(Nr._q),{class:"headers-card-title headers-card-title--compact",style:{top:"calc(var(--refs-header-height)))"}},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Ct.kv),{class:(0,_.C4)(["headers-card-title-icon",{"headers-card-title-icon--open":n}]),icon:"Add",size:"sm"},null,8,["class"]),n?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[(0,E.eW)(" Hide Headers ")],64)):((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[(0,E.eW)(" Show Headers ")],64))])),_:2},1024),(0,E.bF)((0,S.R1)(Nr.kS),null,{default:(0,E.k6)((()=>[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(e.headers,((e,n)=>((0,E.uX)(),(0,E.Wv)(Ea,{key:n,description:e.description,name:`${n}`,value:t(e)?e.schema:void 0},null,8,["description","name","value"])))),128))])),_:1})],2)],2)])),_:1}))}}),[["__scopeId","data-v-2f99eae6"]]),_a={key:0,class:"parameter-item"},Ca={class:"parameter-item-name"},Ta={class:"parameter-item-type"},Aa={class:"absolute right-0"},Ra=It((0,E.pM)({__name:"ParameterListItem",props:{parameter:{},showChildren:{type:Boolean,default:!1},collapsableItems:{type:Boolean,default:!1},withExamples:{type:Boolean,default:!0}},setup(e){const t=e,n=(0,E.EW)((()=>t.parameter.content?Object.keys(t.parameter.content):[])),r=(0,S.KR)(n.value[0]);t.parameter.content&&"application/json"in t.parameter.content&&(r.value="application/json");const a=(0,E.EW)((()=>!(!t.collapsableItems||!t.parameter.content))),o=(0,E.EW)((()=>!0!==t.parameter.readOnly));return(e,n)=>o.value?((0,E.uX)(),(0,E.CE)("li",_a,[(0,E.bF)((0,S.R1)(Nr.EN),null,{default:(0,E.k6)((({open:o})=>[a.value?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Nr._q),{key:0,class:(0,_.C4)(["parameter-item-trigger flex",{"parameter-item-trigger-open":o}])},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Ct.kv),{class:"parameter-item-icon",icon:o?"ChevronDown":"ChevronRight",thickness:"1.5"},null,8,["icon"]),(0,E.bF)(nr,null,{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(o?"Collapse":"Expand"),1)])),_:2},1024),(0,E.Lk)("span",Ca,(0,_.v_)(e.parameter.name),1),(0,E.Lk)("span",Ta,[e.parameter.description?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.mi),{key:0,class:"markdown",value:e.parameter.description},null,8,["value"])):(0,E.Q3)("",!0)]),(0,E.Lk)("div",Aa,[a.value&&t.parameter.content?((0,E.uX)(),(0,E.Wv)(Br,{key:0,class:"parameter-item-content-type",defaultValue:r.value,requestBody:t.parameter,onSelectContentType:n[0]||(n[0]=({contentType:e})=>r.value=e)},null,8,["defaultValue","requestBody"])):(0,E.Q3)("",!0)])])),_:2},1032,["class"])):(0,E.Q3)("",!0),(0,E.bF)((0,S.R1)(Nr.kS),{class:"parameter-item-container parameter-item-container-markdown",static:!a.value},{default:(0,E.k6)((()=>{var t,n;return[e.parameter.headers?((0,E.uX)(),(0,E.Wv)(Sa,{key:0,headers:e.parameter.headers},null,8,["headers"])):(0,E.Q3)("",!0),(0,E.bF)((0,S.R1)(Ea),{compact:"",description:a.value?"":e.parameter.description,name:a.value?"":e.parameter.name,noncollapsible:e.showChildren,required:e.parameter.required,value:e.parameter.content?null==(n=null==(t=e.parameter.content)?void 0:t[r.value])?void 0:n.schema:e.parameter.schema,withExamples:e.withExamples},null,8,["description","name","noncollapsible","required","value","withExamples"])]})),_:1},8,["static"])])),_:1})])):(0,E.Q3)("",!0)}}),[["__scopeId","data-v-e1efed5d"]]),Pa={key:0,class:"parameter-list"},Da={class:"parameter-list-title"},Ma={class:"parameter-list-items"},$a=It((0,E.pM)({__name:"ParameterList",props:{parameters:{},showChildren:{type:Boolean,default:!1},collapsableItems:{type:Boolean,default:!1},withExamples:{type:Boolean,default:!0}},setup:e=>(e,t)=>{var n;return null!=(n=e.parameters)&&n.length?((0,E.uX)(),(0,E.CE)("div",Pa,[(0,E.Lk)("div",Da,[(0,E.RG)(e.$slots,"title",{},void 0,!0)]),(0,E.Lk)("ul",Ma,[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(e.parameters,(t=>((0,E.uX)(),(0,E.Wv)(Ra,{key:t.name,collapsableItems:e.collapsableItems,parameter:t,showChildren:e.showChildren,withExamples:e.withExamples},null,8,["collapsableItems","parameter","showChildren","withExamples"])))),128))])])):(0,E.Q3)("",!0)}}),[["__scopeId","data-v-b05cf5ff"]]),Na={key:0},Ia={class:"request-body-title"},La=["value"],Ba={key:0,class:"request-body-description"},Fa={key:0,class:"request-body-schema"},Qa=It((0,E.pM)({__name:"RequestBody",props:{requestBody:{}},setup(e){var t;const n=(0,E.EW)((()=>{var t;return Object.keys((null==(t=e.requestBody)?void 0:t.content)??{})})),r=(0,S.KR)("application/json");return null!=(t=e.requestBody)&&t.content&&n.value.length>0&&(r.value=n.value[0]),(e,t)=>{var a,o,i,s;return e.requestBody?((0,E.uX)(),(0,E.CE)("div",Na,[(0,E.Lk)("div",Ia,[(0,E.RG)(e.$slots,"title",{},void 0,!0),(0,E.Lk)("div",{class:(0,_.C4)(["request-body-title-select",{"request-body-title-no-select":n.value.length<=1}])},[(0,E.Lk)("span",null,(0,_.v_)(r.value),1),e.requestBody&&n.value.length>1?(0,E.bo)(((0,E.uX)(),(0,E.CE)("select",{key:0,"onUpdate:modelValue":t[0]||(t[0]=e=>r.value=e)},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(null==(a=e.requestBody)?void 0:a.content,((e,t)=>((0,E.uX)(),(0,E.CE)("option",{key:t,value:t},(0,_.v_)(t),9,La)))),128))],512)),[[rt.u1,r.value]]):(0,E.Q3)("",!0)],2),e.requestBody.description?((0,E.uX)(),(0,E.CE)("div",Ba,[(0,E.bF)((0,S.R1)(Ct.mi),{value:e.requestBody.description},null,8,["value"])])):(0,E.Q3)("",!0)]),null!=(o=e.requestBody.content)&&o[r.value]?((0,E.uX)(),(0,E.CE)("div",Fa,[(0,E.bF)((0,S.R1)(Xr),{compact:"",noncollapsible:"",value:null==(s=null==(i=e.requestBody.content)?void 0:i[r.value])?void 0:s.schema},null,8,["value"])])):(0,E.Q3)("",!0)])):(0,E.Q3)("",!0)}}}),[["__scopeId","data-v-dc8f9923"]]),Ua=(0,E.pM)({__name:"OperationParameters",props:{operation:{}},setup(e){const t=e,n=e=>{var n,r;return(null==(r=null==(n=t.operation)?void 0:n.parameters)?void 0:r.filter((t=>t.in===e)))??[]};return(e,t)=>{var r;return(0,E.uX)(),(0,E.CE)(E.FK,null,[(0,E.bF)($a,{parameters:n("path")},{title:(0,E.k6)((()=>t[0]||(t[0]=[(0,E.eW)("Path Parameters")]))),_:1},8,["parameters"]),(0,E.bF)($a,{parameters:n("query")},{title:(0,E.k6)((()=>t[1]||(t[1]=[(0,E.eW)("Query Parameters")]))),_:1},8,["parameters"]),(0,E.bF)($a,{parameters:n("header")},{title:(0,E.k6)((()=>t[2]||(t[2]=[(0,E.eW)("Headers")]))),_:1},8,["parameters"]),(0,E.bF)($a,{parameters:n("cookie")},{title:(0,E.k6)((()=>t[3]||(t[3]=[(0,E.eW)("Cookies")]))),_:1},8,["parameters"]),null!=(r=e.operation)&&r.requestBody?((0,E.uX)(),(0,E.Wv)(Qa,{key:0,requestBody:e.operation.requestBody},{title:(0,E.k6)((()=>t[4]||(t[4]=[(0,E.eW)("Body")]))),_:1},8,["requestBody"])):(0,E.Q3)("",!0)],64)}}}),za=(0,E.pM)({__name:"OperationResponses",props:{operation:{},collapsableItems:{type:Boolean,default:!0}},setup(e){const t=e,{responses:n}=(r=t.operation,{responses:(0,E.EW)((()=>{if(!r.information)return[];const{responses:e}=r.information,t=[];return e&&Object.entries(e).forEach((([e,n])=>{t.push({name:e,description:n.description,content:n.content,headers:n.headers})})),t}))});var r;return(e,t)=>((0,E.uX)(),(0,E.Wv)($a,{collapsableItems:e.collapsableItems,parameters:(0,S.R1)(n),withExamples:!1},{title:(0,E.k6)((()=>t[0]||(t[0]=[(0,E.eW)("Responses")]))),_:1},8,["collapsableItems","parameters"]))}}),ja={class:"section-accordion-button-content"},Wa={key:0,class:"section-accordion-button-actions"},Xa={key:0,class:"section-accordion-description"},qa={class:"section-accordion-content-card"},Va=It((0,E.pM)({__name:"SectionAccordion",props:{id:{},transparent:{type:Boolean}},setup(e){const t=(0,S.KR)(),n=(0,We.GF$)(t);return(e,r)=>((0,E.uX)(),(0,E.Wv)(rr,{id:e.id,class:"section-wrapper"},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Nr.EN),{as:"section",class:(0,_.C4)(["section-accordion",{"section-accordion-transparent":e.transparent}])},{default:(0,E.k6)((({open:r})=>[(0,E.bF)((0,S.R1)(Nr._q),{ref_key:"button",ref:t,class:"section-accordion-button"},{default:(0,E.k6)((()=>[(0,E.Lk)("div",ja,[(0,E.RG)(e.$slots,"title",{},void 0,!0)]),e.$slots.actions?((0,E.uX)(),(0,E.CE)("div",Wa,[(0,E.RG)(e.$slots,"actions",{active:(0,S.R1)(n)||r},void 0,!0)])):(0,E.Q3)("",!0),(0,E.bF)((0,S.R1)(Ct.kv),{class:"section-accordion-chevron",icon:r?"ChevronDown":"ChevronRight"},null,8,["icon"])])),_:2},1536),(0,E.bF)((0,S.R1)(Nr.kS),{class:"section-accordion-content"},{default:(0,E.k6)((()=>[e.$slots.description?((0,E.uX)(),(0,E.CE)("div",Xa,[(0,E.RG)(e.$slots,"description",{},void 0,!0)])):(0,E.Q3)("",!0),(0,E.Lk)("div",qa,[(0,E.RG)(e.$slots,"default",{},void 0,!0)])])),_:3})])),_:3},8,["class"])])),_:3},8,["id"]))}}),[["__scopeId","data-v-07cabcb0"]]);function Ha(e){return"function"==typeof e?e():(0,S.R1)(e)}const Za="undefined"!=typeof window&&"undefined"!=typeof document,Ga=("undefined"!=typeof WorkerGlobalScope&&(globalThis,WorkerGlobalScope),e=>null!=e);function Ka(e){const t=Object.create(null);return n=>t[n]||(t[n]=e(n))}Object.prototype.toString;const Ya=/\B([A-Z])/g,Ja=(Ka((e=>e.replace(Ya,"-$1").toLowerCase())),/-(\w)/g);Ka((e=>e.replace(Ja,((e,t)=>t?t.toUpperCase():"")))),Za&&window,Za&&window.document,Za&&window.navigator,Za&&window.location,"undefined"!=typeof globalThis?globalThis:"undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self&&self,Number.POSITIVE_INFINITY;var eo=["input:not([inert])","select:not([inert])","textarea:not([inert])","a[href]:not([inert])","button:not([inert])","[tabindex]:not(slot):not([inert])","audio[controls]:not([inert])","video[controls]:not([inert])",'[contenteditable]:not([contenteditable="false"]):not([inert])',"details>summary:first-of-type:not([inert])","details:not([inert])"],to=eo.join(","),no="undefined"==typeof Element,ro=no?function(){}:Element.prototype.matches||Element.prototype.msMatchesSelector||Element.prototype.webkitMatchesSelector,ao=!no&&Element.prototype.getRootNode?function(e){var t;return null==e||null===(t=e.getRootNode)||void 0===t?void 0:t.call(e)}:function(e){return null==e?void 0:e.ownerDocument},oo=function e(t,n){var r;void 0===n&&(n=!0);var a=null==t||null===(r=t.getAttribute)||void 0===r?void 0:r.call(t,"inert");return""===a||"true"===a||n&&t&&e(t.parentNode)},io=function(e,t,n){if(oo(e))return[];var r=Array.prototype.slice.apply(e.querySelectorAll(to));return t&&ro.call(e,to)&&r.unshift(e),r.filter(n)},so=function e(t,n,r){for(var a=[],o=Array.from(t);o.length;){var i=o.shift();if(!oo(i,!1))if("SLOT"===i.tagName){var s=i.assignedElements(),l=e(s.length?s:i.children,!0,r);r.flatten?a.push.apply(a,l):a.push({scopeParent:i,candidates:l})}else{ro.call(i,to)&&r.filter(i)&&(n||!t.includes(i))&&a.push(i);var c=i.shadowRoot||"function"==typeof r.getShadowRoot&&r.getShadowRoot(i),u=!oo(c,!1)&&(!r.shadowRootFilter||r.shadowRootFilter(i));if(c&&u){var d=e(!0===c?i.children:c.children,!0,r);r.flatten?a.push.apply(a,d):a.push({scopeParent:i,candidates:d})}else o.unshift.apply(o,i.children)}}return a},lo=function(e){return!isNaN(parseInt(e.getAttribute("tabindex"),10))},co=function(e){if(!e)throw new Error("No node provided");return e.tabIndex<0&&(/^(AUDIO|VIDEO|DETAILS)$/.test(e.tagName)||function(e){var t,n=null==e||null===(t=e.getAttribute)||void 0===t?void 0:t.call(e,"contenteditable");return""===n||"true"===n}(e))&&!lo(e)?0:e.tabIndex},uo=function(e,t){return e.tabIndex===t.tabIndex?e.documentOrder-t.documentOrder:e.tabIndex-t.tabIndex},po=function(e){return"INPUT"===e.tagName},ho=function(e){var t=e.getBoundingClientRect(),n=t.width,r=t.height;return 0===n&&0===r},fo=function(e,t){return!(t.disabled||oo(t)||function(e){return po(e)&&"hidden"===e.type}(t)||function(e,t){var n=t.displayCheck,r=t.getShadowRoot;if("hidden"===getComputedStyle(e).visibility)return!0;var a=ro.call(e,"details>summary:first-of-type")?e.parentElement:e;if(ro.call(a,"details:not([open]) *"))return!0;if(n&&"full"!==n&&"legacy-full"!==n){if("non-zero-area"===n)return ho(e)}else{if("function"==typeof r){for(var o=e;e;){var i=e.parentElement,s=ao(e);if(i&&!i.shadowRoot&&!0===r(i))return ho(e);e=e.assignedSlot?e.assignedSlot:i||s===e.ownerDocument?i:s.host}e=o}if(function(e){var t,n,r,a,o=e&&ao(e),i=null===(t=o)||void 0===t?void 0:t.host,s=!1;if(o&&o!==e)for(s=!!(null!==(n=i)&&void 0!==n&&null!==(r=n.ownerDocument)&&void 0!==r&&r.contains(i)||null!=e&&null!==(a=e.ownerDocument)&&void 0!==a&&a.contains(e));!s&&i;){var l,c,u;s=!(null===(c=i=null===(l=o=ao(i))||void 0===l?void 0:l.host)||void 0===c||null===(u=c.ownerDocument)||void 0===u||!u.contains(i))}return s}(e))return!e.getClientRects().length;if("legacy-full"!==n)return!0}return!1}(t,e)||function(e){return"DETAILS"===e.tagName&&Array.prototype.slice.apply(e.children).some((function(e){return"SUMMARY"===e.tagName}))}(t)||function(e){if(/^(INPUT|BUTTON|SELECT|TEXTAREA)$/.test(e.tagName))for(var t=e.parentElement;t;){if("FIELDSET"===t.tagName&&t.disabled){for(var n=0;n<t.children.length;n++){var r=t.children.item(n);if("LEGEND"===r.tagName)return!!ro.call(t,"fieldset[disabled] *")||!r.contains(e)}return!0}t=t.parentElement}return!1}(t))},mo=function(e,t){return!(function(e){return function(e){return po(e)&&"radio"===e.type}(e)&&!function(e){if(!e.name)return!0;var t,n=e.form||ao(e),r=function(e){return n.querySelectorAll('input[type="radio"][name="'+e+'"]')};if("undefined"!=typeof window&&void 0!==window.CSS&&"function"==typeof window.CSS.escape)t=r(window.CSS.escape(e.name));else try{t=r(e.name)}catch(e){return console.error("Looks like you have a radio button with a name attribute containing invalid CSS selector characters and need the CSS.escape polyfill: %s",e.message),!1}var a=function(e,t){for(var n=0;n<e.length;n++)if(e[n].checked&&e[n].form===t)return e[n]}(t,e.form);return!a||a===e}(e)}(t)||co(t)<0||!fo(e,t))},go=function(e){var t=parseInt(e.getAttribute("tabindex"),10);return!!(isNaN(t)||t>=0)},vo=function e(t){var n=[],r=[];return t.forEach((function(t,a){var o=!!t.scopeParent,i=o?t.scopeParent:t,s=function(e,t){var n=co(e);return n<0&&t&&!lo(e)?0:n}(i,o),l=o?e(t.candidates):i;0===s?o?n.push.apply(n,l):n.push(i):r.push({documentOrder:a,tabIndex:s,item:t,isScope:o,content:l})})),r.sort(uo).reduce((function(e,t){return t.isScope?e.push.apply(e,t.content):e.push(t.content),e}),[]).concat(n)},bo=function(e,t){if(t=t||{},!e)throw new Error("No node provided");return!1!==ro.call(e,to)&&mo(t,e)},yo=eo.concat("iframe").join(","),wo=function(e,t){if(t=t||{},!e)throw new Error("No node provided");return!1!==ro.call(e,yo)&&fo(t,e)};function Oo(e,t){(null==t||t>e.length)&&(t=e.length);for(var n=0,r=Array(t);n<t;n++)r[n]=e[n];return r}function ko(e,t,n){return(t=function(e){var t=function(e){if("object"!=typeof e||!e)return e;var t=e[Symbol.toPrimitive];if(void 0!==t){var n=t.call(e,"string");if("object"!=typeof n)return n;throw new TypeError("@@toPrimitive must return a primitive value.")}return String(e)}(e);return"symbol"==typeof t?t:t+""}(t))in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function xo(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter((function(t){return Object.getOwnPropertyDescriptor(e,t).enumerable}))),n.push.apply(n,r)}return n}function Eo(e){for(var t=1;t<arguments.length;t++){var n=null!=arguments[t]?arguments[t]:{};t%2?xo(Object(n),!0).forEach((function(t){ko(e,t,n[t])})):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):xo(Object(n)).forEach((function(t){Object.defineProperty(e,t,Object.getOwnPropertyDescriptor(n,t))}))}return e}var So=function(e){return"Tab"===(null==e?void 0:e.key)||9===(null==e?void 0:e.keyCode)},_o=function(e){return So(e)&&!e.shiftKey},Co=function(e){return So(e)&&e.shiftKey},To=function(e){return setTimeout(e,0)},Ao=function(e){for(var t=arguments.length,n=new Array(t>1?t-1:0),r=1;r<t;r++)n[r-1]=arguments[r];return"function"==typeof e?e.apply(void 0,n):e},Ro=function(e){return e.target.shadowRoot&&"function"==typeof e.composedPath?e.composedPath()[0]:e.target},Po=[],Do=function(e,t){var n,r=(null==t?void 0:t.document)||document,a=(null==t?void 0:t.trapStack)||Po,o=Eo({returnFocusOnDeactivate:!0,escapeDeactivates:!0,delayInitialFocus:!0,isKeyForward:_o,isKeyBackward:Co},t),i={containers:[],containerGroups:[],tabbableGroups:[],nodeFocusedBeforeActivation:null,mostRecentlyFocusedNode:null,active:!1,paused:!1,manuallyPaused:!1,delayInitialFocusTimer:void 0,recentNavEvent:void 0},s=function(e,t,n){return e&&void 0!==e[t]?e[t]:o[n||t]},l=function(e,t){var n="function"==typeof(null==t?void 0:t.composedPath)?t.composedPath():void 0;return i.containerGroups.findIndex((function(t){var r=t.container,a=t.tabbableNodes;return r.contains(e)||(null==n?void 0:n.includes(r))||a.find((function(t){return t===e}))}))},c=function(e){var t,n=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{},a=n.hasFallback,i=void 0!==a&&a,s=n.params,l=void 0===s?[]:s,c=o[e];if("function"==typeof c&&(c=c.apply(void 0,function(e){if(Array.isArray(e))return Oo(e)}(t=l)||function(e){if("undefined"!=typeof Symbol&&null!=e[Symbol.iterator]||null!=e["@@iterator"])return Array.from(e)}(t)||function(e,t){if(e){if("string"==typeof e)return Oo(e,t);var n={}.toString.call(e).slice(8,-1);return"Object"===n&&e.constructor&&(n=e.constructor.name),"Map"===n||"Set"===n?Array.from(e):"Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)?Oo(e,t):void 0}}(t)||function(){throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}())),!0===c&&(c=void 0),!c){if(void 0===c||!1===c)return c;throw new Error("`".concat(e,"` was specified but was not a node, or did not return a node"))}var u=c;if("string"==typeof c){try{u=r.querySelector(c)}catch(t){throw new Error("`".concat(e,'` appears to be an invalid selector; error="').concat(t.message,'"'))}if(!u&&!i)throw new Error("`".concat(e,"` as selector refers to no known node"))}return u},u=function(){var e=c("initialFocus",{hasFallback:!0});if(!1===e)return!1;if(void 0===e||e&&!wo(e,o.tabbableOptions))if(l(r.activeElement)>=0)e=r.activeElement;else{var t=i.tabbableGroups[0];e=t&&t.firstTabbableNode||c("fallbackFocus")}else null===e&&(e=c("fallbackFocus"));if(!e)throw new Error("Your focus-trap needs to have at least one focusable element");return e},d=function(){if(i.containerGroups=i.containers.map((function(e){var t=function(e,t){var n;return n=(t=t||{}).getShadowRoot?so([e],t.includeContainer,{filter:mo.bind(null,t),flatten:!1,getShadowRoot:t.getShadowRoot,shadowRootFilter:go}):io(e,t.includeContainer,mo.bind(null,t)),vo(n)}(e,o.tabbableOptions),n=function(e,t){return(t=t||{}).getShadowRoot?so([e],t.includeContainer,{filter:fo.bind(null,t),flatten:!0,getShadowRoot:t.getShadowRoot}):io(e,t.includeContainer,fo.bind(null,t))}(e,o.tabbableOptions),r=t.length>0?t[0]:void 0,a=t.length>0?t[t.length-1]:void 0,i=n.find((function(e){return bo(e)})),s=n.slice().reverse().find((function(e){return bo(e)})),l=!!t.find((function(e){return co(e)>0}));return{container:e,tabbableNodes:t,focusableNodes:n,posTabIndexesFound:l,firstTabbableNode:r,lastTabbableNode:a,firstDomTabbableNode:i,lastDomTabbableNode:s,nextTabbableNode:function(e){var r=!(arguments.length>1&&void 0!==arguments[1])||arguments[1],a=t.indexOf(e);return a<0?r?n.slice(n.indexOf(e)+1).find((function(e){return bo(e)})):n.slice(0,n.indexOf(e)).reverse().find((function(e){return bo(e)})):t[a+(r?1:-1)]}}})),i.tabbableGroups=i.containerGroups.filter((function(e){return e.tabbableNodes.length>0})),i.tabbableGroups.length<=0&&!c("fallbackFocus"))throw new Error("Your focus-trap must have at least one container with at least one tabbable node in it at all times");if(i.containerGroups.find((function(e){return e.posTabIndexesFound}))&&i.containerGroups.length>1)throw new Error("At least one node with a positive tabindex was found in one of your focus-trap's multiple containers. Positive tabindexes are only supported in single-container focus-traps.")},p=function(e){var t=e.activeElement;if(t)return t.shadowRoot&&null!==t.shadowRoot.activeElement?p(t.shadowRoot):t},h=function(e){!1!==e&&e!==p(document)&&(e&&e.focus?(e.focus({preventScroll:!!o.preventScroll}),i.mostRecentlyFocusedNode=e,function(e){return e.tagName&&"input"===e.tagName.toLowerCase()&&"function"==typeof e.select}(e)&&e.select()):h(u()))},f=function(e){var t=c("setReturnFocus",{params:[e]});return t||!1!==t&&e},m=function(e){var t=e.target,n=e.event,r=e.isBackward,a=void 0!==r&&r;t=t||Ro(n),d();var s=null;if(i.tabbableGroups.length>0){var u=l(t,n),p=u>=0?i.containerGroups[u]:void 0;if(u<0)s=a?i.tabbableGroups[i.tabbableGroups.length-1].lastTabbableNode:i.tabbableGroups[0].firstTabbableNode;else if(a){var h=i.tabbableGroups.findIndex((function(e){var n=e.firstTabbableNode;return t===n}));if(h<0&&(p.container===t||wo(t,o.tabbableOptions)&&!bo(t,o.tabbableOptions)&&!p.nextTabbableNode(t,!1))&&(h=u),h>=0){var f=0===h?i.tabbableGroups.length-1:h-1,m=i.tabbableGroups[f];s=co(t)>=0?m.lastTabbableNode:m.lastDomTabbableNode}else So(n)||(s=p.nextTabbableNode(t,!1))}else{var g=i.tabbableGroups.findIndex((function(e){var n=e.lastTabbableNode;return t===n}));if(g<0&&(p.container===t||wo(t,o.tabbableOptions)&&!bo(t,o.tabbableOptions)&&!p.nextTabbableNode(t))&&(g=u),g>=0){var v=g===i.tabbableGroups.length-1?0:g+1,b=i.tabbableGroups[v];s=co(t)>=0?b.firstTabbableNode:b.firstDomTabbableNode}else So(n)||(s=p.nextTabbableNode(t))}}else s=c("fallbackFocus");return s},g=function(e){var t=Ro(e);l(t,e)>=0||(Ao(o.clickOutsideDeactivates,e)?n.deactivate({returnFocus:o.returnFocusOnDeactivate}):Ao(o.allowOutsideClick,e)||e.preventDefault())},v=function(e){var t=Ro(e),n=l(t,e)>=0;if(n||t instanceof Document)n&&(i.mostRecentlyFocusedNode=t);else{var r;e.stopImmediatePropagation();var a=!0;if(i.mostRecentlyFocusedNode)if(co(i.mostRecentlyFocusedNode)>0){var s=l(i.mostRecentlyFocusedNode),c=i.containerGroups[s].tabbableNodes;if(c.length>0){var d=c.findIndex((function(e){return e===i.mostRecentlyFocusedNode}));d>=0&&(o.isKeyForward(i.recentNavEvent)?d+1<c.length&&(r=c[d+1],a=!1):d-1>=0&&(r=c[d-1],a=!1))}}else i.containerGroups.some((function(e){return e.tabbableNodes.some((function(e){return co(e)>0}))}))||(a=!1);else a=!1;a&&(r=m({target:i.mostRecentlyFocusedNode,isBackward:o.isKeyBackward(i.recentNavEvent)})),h(r||i.mostRecentlyFocusedNode||u())}i.recentNavEvent=void 0},b=function(e){(o.isKeyForward(e)||o.isKeyBackward(e))&&function(e){var t=arguments.length>1&&void 0!==arguments[1]&&arguments[1];i.recentNavEvent=e;var n=m({event:e,isBackward:t});n&&(So(e)&&e.preventDefault(),h(n))}(e,o.isKeyBackward(e))},y=function(e){var t;("Escape"===(null==(t=e)?void 0:t.key)||"Esc"===(null==t?void 0:t.key)||27===(null==t?void 0:t.keyCode))&&!1!==Ao(o.escapeDeactivates,e)&&(e.preventDefault(),n.deactivate())},w=function(e){var t=Ro(e);l(t,e)>=0||Ao(o.clickOutsideDeactivates,e)||Ao(o.allowOutsideClick,e)||(e.preventDefault(),e.stopImmediatePropagation())},O=function(){if(i.active)return function(e,t){if(e.length>0){var n=e[e.length-1];n!==t&&n._setPausedState(!0)}var r=e.indexOf(t);-1===r||e.splice(r,1),e.push(t)}(a,n),i.delayInitialFocusTimer=o.delayInitialFocus?To((function(){h(u())})):h(u()),r.addEventListener("focusin",v,!0),r.addEventListener("mousedown",g,{capture:!0,passive:!1}),r.addEventListener("touchstart",g,{capture:!0,passive:!1}),r.addEventListener("click",w,{capture:!0,passive:!1}),r.addEventListener("keydown",b,{capture:!0,passive:!1}),r.addEventListener("keydown",y),n},k=function(){if(i.active)return r.removeEventListener("focusin",v,!0),r.removeEventListener("mousedown",g,!0),r.removeEventListener("touchstart",g,!0),r.removeEventListener("click",w,!0),r.removeEventListener("keydown",b,!0),r.removeEventListener("keydown",y),n},x="undefined"!=typeof window&&"MutationObserver"in window?new MutationObserver((function(e){e.some((function(e){return Array.from(e.removedNodes).some((function(e){return e===i.mostRecentlyFocusedNode}))}))&&h(u())})):void 0,E=function(){x&&(x.disconnect(),i.active&&!i.paused&&i.containers.map((function(e){x.observe(e,{subtree:!0,childList:!0})})))};return n={get active(){return i.active},get paused(){return i.paused},activate:function(e){if(i.active)return this;var t=s(e,"onActivate"),n=s(e,"onPostActivate"),a=s(e,"checkCanFocusTrap");a||d(),i.active=!0,i.paused=!1,i.nodeFocusedBeforeActivation=r.activeElement,null==t||t();var o=function(){a&&d(),O(),E(),null==n||n()};return a?(a(i.containers.concat()).then(o,o),this):(o(),this)},deactivate:function(e){if(!i.active)return this;var t=Eo({onDeactivate:o.onDeactivate,onPostDeactivate:o.onPostDeactivate,checkCanReturnFocus:o.checkCanReturnFocus},e);clearTimeout(i.delayInitialFocusTimer),i.delayInitialFocusTimer=void 0,k(),i.active=!1,i.paused=!1,E(),function(e,t){var n=e.indexOf(t);-1!==n&&e.splice(n,1),e.length>0&&!e[e.length-1]._isManuallyPaused()&&e[e.length-1]._setPausedState(!1)}(a,n);var r=s(t,"onDeactivate"),l=s(t,"onPostDeactivate"),c=s(t,"checkCanReturnFocus"),u=s(t,"returnFocus","returnFocusOnDeactivate");null==r||r();var d=function(){To((function(){u&&h(f(i.nodeFocusedBeforeActivation)),null==l||l()}))};return u&&c?(c(f(i.nodeFocusedBeforeActivation)).then(d,d),this):(d(),this)},pause:function(e){return i.active?(i.manuallyPaused=!0,this._setPausedState(!0,e)):this},unpause:function(e){return i.active?(i.manuallyPaused=!1,a[a.length-1]!==this?this:this._setPausedState(!1,e)):this},updateContainerElements:function(e){var t=[].concat(e).filter(Boolean);return i.containers=t.map((function(e){return"string"==typeof e?r.querySelector(e):e})),i.active&&d(),E(),this}},Object.defineProperties(n,{_isManuallyPaused:{value:function(){return i.manuallyPaused}},_setPausedState:{value:function(e,t){if(i.paused===e)return this;if(i.paused=e,e){var n=s(t,"onPause"),r=s(t,"onPostPause");null==n||n(),k(),E(),null==r||r()}else{var a=s(t,"onUnpause"),o=s(t,"onPostUnpause");null==a||a(),d(),O(),E(),null==o||o()}return this}}}),n.updateContainerElements(e),n},Mo=n(9351),$o=n(1845);const No={class:"scalar scalar-app"},Io={class:"scalar-container"},Lo=["id"],Bo=(0,E.pM)({__name:"ApiClientModal",setup(e){const{activeWorkspace:t}=(0,_t.JC)(),{modalState:n,events:r}=(0,St.hA)(),a=(0,S.KR)(null),o=(0,E.Bi)(),{activate:i,deactivate:s}=function(e,t={}){let n;const{immediate:r,...a}=t,o=(0,S.KR)(!1),i=(0,S.KR)(!1),s=e=>n&&n.activate(e),l=e=>n&&n.deactivate(e),c=(0,E.EW)((()=>{const t=Ha(e);return(Array.isArray(t)?t:[t]).map((e=>{const t=Ha(e);return"string"==typeof t?t:function(e){var t;const n=Ha(e);return null!=(t=null==n?void 0:n.$el)?t:n}(t)})).filter(Ga)}));return(0,E.wB)(c,(e=>{e.length&&(n=Do(e,{...a,onActivate(){o.value=!0,t.onActivate&&t.onActivate()},onDeactivate(){o.value=!1,t.onDeactivate&&t.onDeactivate()}}),r&&s())}),{flush:"post"}),u=()=>l(),!!(0,S.o5)()&&(0,S.jr)(u),{hasFocus:o,isPaused:i,activate:s,deactivate:l,pause:()=>{n&&(n.pause(),i.value=!0)},unpause:()=>{n&&(n.unpause(),i.value=!1)}};var u}(a,{allowOutsideClick:!0,initialFocus:()=>a.value,fallbackFocus:`#${o}`}),l=e=>{var n;return(0,$o.rl)(e,r.hotKeys,null==(n=t.value)?void 0:n.hotKeyConfig)};(0,E.wB)((()=>n.open),(e=>{e?(window.addEventListener("keydown",l),document.documentElement.style.overflow="hidden",i()):(window.removeEventListener("keydown",l),document.documentElement.style.removeProperty("overflow"),s())})),(0,E.KC)((()=>(0,Ct.AG)()));const c=e=>(null==e?void 0:e.closeModal)&&n.open&&n.hide();return(0,E.sV)((()=>r.hotKeys.on(c))),(0,E.xo)((()=>{document.documentElement.style.removeProperty("overflow"),r.hotKeys.off(c)})),(e,t)=>(0,E.bo)(((0,E.uX)(),(0,E.CE)("div",No,[(0,E.Lk)("div",Io,[(0,E.Lk)("div",{id:(0,S.R1)(o),ref_key:"client",ref:a,"aria-label":"API Client","aria-modal":"true",class:"scalar-app-layout scalar-client",role:"dialog",tabindex:"0"},[(0,E.bF)((0,S.R1)(Ct.ks),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Mo.Tp),{key:"$route.fullPath"})])),_:1})],8,Lo),(0,E.Lk)("div",{class:"scalar-app-exit -z-1",onClick:t[0]||(t[0]=e=>(0,S.R1)(n).hide())})])],512)),[[rt.aG,(0,S.R1)(n).open]])}}),Fo=(0,n(3526).A)(Bo,[["__scopeId","data-v-cc5e3e81"]]);var Qo=n(2490),Uo=n(6455),zo=n(8803),jo=n(1347),Wo=n(9582);const Xo="2.5.0",qo="scalar_api_client_data_version";var Vo=n(1284);const Ho=e=>{const t=localStorage.getItem(e)||"[{}]";return"["===t[0]?(0,Vo.qg)(t):JSON.parse(t)},Zo=(e,t)=>{const n=e=>{const[t,n=0,r=0]=e.split(".").map((e=>parseInt(e,10)));return{major:t,minor:n,patch:r}},{major:r,minor:a,patch:o}=n(e),{major:i,minor:s,patch:l}=n(t);return r<i||!(r>i)&&(a<s||!(a>s)&&o<l)},Go=e=>e[0].toUpperCase()+e.slice(1),Ko=(e,t,n)=>e.forEach((e=>{const r=(0,Xe.TA)(e,t,!1);r&&n(r)})),Yo=({el:e,appComponent:t,configuration:n={},isReadOnly:r=!1,store:a,persistData:o=!0,mountOnInitialize:i=!0,layout:s="desktop",router:l})=>{const c=a||(0,St.Pg)({proxyUrl:n.proxyUrl,themeId:n.themeId,showSidebar:n.showSidebar??!0,hideClientButton:n.hideClientButton??!1,integration:n._integration,useLocalStorage:o}),u=(0,_t.T7)({...c,router:l});if(localStorage.getItem(Xe.nF.WORKSPACE)&&!r){const e={};let t=0,n=0,r="";for(r in localStorage)Object.prototype.hasOwnProperty.call(localStorage,r)&&(n=2*(localStorage[r].length+r.length),t+=n,e[r]=(n/1024).toFixed(2)+" KB");e.Total=(t/1024).toFixed(2)+" KB",console.table(e),(e=>{const{collectionMutators:t,cookieMutators:n,environmentMutators:r,tagMutators:a,requestExampleMutators:o,requestMutators:i,serverMutators:s,securitySchemeMutators:l,workspaceMutators:c}=e;try{const{collections:e,cookies:u,environments:d,requestExamples:p,requests:h,servers:f,securitySchemes:m,tags:g,workspaces:v}=(()=>{const e=(()=>{const e=localStorage.getItem("collection"),t=localStorage.getItem(qo);if(t)return t;if(!e?.length||"{"===e?.[0])return"0.0.0";try{const[t]=Object.values((0,Vo.qg)(e)??{});return"collection"===t?.type?"2.1.0":"0.0.0"}catch(e){return console.error(e),"0.0.0"}})();console.info("Data version: "+e);let t={collections:Ho("collection"),cookies:Ho("cookie"),environments:Ho("environment"),requestExamples:Ho("requestExample"),requests:Ho("request"),securitySchemes:Ho("securityScheme"),servers:Ho("server"),tags:Ho("tag"),workspaces:Ho("workspace")};return Zo(e,"2.1.0")&&(t=(e=>{console.info("Performing data migration v-0.0.0 to v-2.1.0");const t={...e,folders:Ho("folder")},n=e=>e.reduce(((e,r)=>{const a=t.requests[r];if(a)e.requestUids.add(r),a.securitySchemeUids?.forEach((t=>e.authUids.add(t)));else if(t.folders[r]){const{requestUids:a,tagUids:o,authUids:i}=n(t.folders[r].childUids??[]);e.tagUids.add(r),a.forEach((t=>e.requestUids.add(t))),o.forEach((t=>e.tagUids.add(t))),i.forEach((t=>e.authUids.add(t)))}return e}),{requestUids:new Set,tagUids:new Set,authUids:new Set}),r={},a=Object.values(t.collections??{}).reduce(((e,a)=>{const{requestUids:o,tagUids:i,authUids:s}=n(a.childUids??[]),l=[...new Set([...s,...Object.values(a.securitySchemeDict??{})])];o.forEach((e=>r[e]=l));const c=l.reduce(((e,n)=>{const r=t.securitySchemes[n];return r?.uid&&e&&(e[n]=(e=>"apiKey"===e.type?{type:"apiKey",name:e.name,value:e.value??""}:"http"===e.type?{type:"http",username:e.value??"",password:e.secondValue??"",token:e.value??""}:"oauth2"===e.type&&"implicit"===e.flow.type?{type:"oauth-implicit",token:e.flow.token??""}:"oauth2"===e.type&&"password"===e.flow.type?{type:"oauth-password",token:e.flow.token??"",username:e.flow.value??"",password:e.flow.secondValue??"",clientSecret:e.flow.clientSecret??""}:"oauth2"===e.type&&"clientCredentials"===e.flow.type?{type:"oauth-clientCredentials",token:e.flow.token??"",clientSecret:e.flow.clientSecret??""}:"oauth2"===e.type&&"authorizationCode"===e.flow.type?{type:"oauth-authorizationCode",token:e.flow.token??"",clientSecret:e.flow.clientSecret??""}:{type:"apiKey",name:"",value:""})(r)),e}),{});return e[a.uid]={type:"collection",openapi:a.spec?.openapi||"3.1.0",info:a.spec?.info||{title:"OpenAPI Spec",version:"0.0.1"},security:a.spec?.security||[],externalDocs:a.spec?.externalDocs,uid:a.uid,securitySchemes:l,selectedSecuritySchemeUids:[],selectedServerUid:a.selectedServerUid||a.spec?.serverUids?.[0]||"",servers:a.spec?.serverUids||[],requests:[...o],tags:[...i],auth:c,children:a.childUids||[],"x-scalar-icon":"interface-content-folder",watchMode:!1,watchModeStatus:"IDLE"},e}),{}),o=t.cookies??{},i=Object.values(t.environments??{}).reduce(((e,t)=>(e[t.uid]={...t,value:t.raw??""},e)),{}),s=Object.values(t.requests??{}).reduce(((e,t)=>{const n=[...Object.values(t.parameters?.path??{}),...Object.values(t.parameters?.query??{}),...Object.values(t.parameters?.headers??{}),...Object.values(t.parameters?.cookies??{})].filter((e=>e)),a=(t.selectedSecuritySchemeUids||[]).filter((e=>r[t.uid]?.includes(e)));return e[t.uid]={...t,parameters:n,type:"request",method:t.method?.toLowerCase()??"get",examples:t.childUids||[],selectedSecuritySchemeUids:a,selectedServerUid:"",servers:[]},e}),{}),l=Object.values(t.requestExamples??{}).reduce(((e,t)=>(e[t.uid]={...t,type:"requestExample"},e)),{}),c=e=>{const t={refreshUrl:e.refreshUrl||"",selectedScopes:e.selectedScopes||[],scopes:e.scopes||{}};return"implicit"===e.type?{...e,...t,type:"implicit","x-scalar-redirect-uri":("redirectUri"in e?e.redirectUri:"")||""}:"password"===e.type||"clientCredentials"===e.type?{...e,...t,tokenUrl:e.tokenUrl||""}:{...e,...t,"x-usePkce":"no","x-scalar-redirect-uri":("redirectUri"in e?e.redirectUri:"")||"",authorizationUrl:e.authorizationUrl||"",tokenUrl:e.tokenUrl||""}},u=e=>{switch(e?.type){case"apiKey":return`${Go(e.in)}`;case"http":return`${Go(e.scheme)} Authentication`;case"oauth2":return e.flow.type.replace(/([A-Z])/g,(e=>` ${e}`)).replace(/^./,(e=>e.toUpperCase())).trim();case"openIdConnect":return"Open ID Connect";default:return"None"}};return{collections:a,cookies:o,environments:i,requestExamples:l,requests:s,securitySchemes:Object.values(t.securitySchemes??{}).reduce(((e,t)=>(e[t.uid]="oauth2"===t.type?{...t,nameKey:u(t),"x-scalar-client-id":t.clientId||"",flow:c(t.flow)}:{...t,nameKey:u(t)},e)),{}),servers:Object.values(t.servers??{}).reduce(((e,t)=>(e[t.uid]={...t,variables:t.variables??{}},e)),{}),tags:Object.values(t.folders??{}).reduce(((e,t)=>(e[t.uid]={type:"tag",uid:t.uid,name:t.name||"unknownTag",description:t.description,children:t.childUids||[],"x-scalar-children":[]},e)),{}),workspaces:Object.values(t.workspaces??{}).reduce(((e,t)=>(e[t.uid]={...t,description:t.description??"Basic Scalar Workspace",cookies:t.cookieUids||[],collections:t.collectionUids||[],environments:t.environmentUids||[]},e)),{})}})(t)),Zo(e,"2.2.0")&&(t=(e=>{console.info("Performing data migration v-2.1.0 to v-2.2.0");const t=Object.values(e.securitySchemes).reduce(((t,n)=>{const r=Object.values(e.collections).find((e=>e.securitySchemes.includes(n.uid))),a=r?.auth?.[n.uid];if(!a)return t;const o=((e,t)=>{if("apiKey"===e.type&&"apiKey"===t.type)return{...e,value:t.value};if("http"===e.type&&"http"===t.type)return{...e,username:t.username,password:t.password,token:t.token};if("oauth2"===e.type){const{flow:n,...r}=e;if("implicit"===n.type&&"oauth-implicit"===t.type)return{...r,flows:{implicit:{...n,scopes:n.scopes,token:t.token,"x-scalar-client-id":r["x-scalar-client-id"]}}};if("password"===n.type&&"oauth-password"===t.type)return{...r,flows:{password:{...n,username:t.username,password:t.password,token:t.token,clientSecret:t.clientSecret,scopes:n.scopes,"x-scalar-client-id":r["x-scalar-client-id"]}}};if("clientCredentials"===n.type&&"oauth-clientCredentials"===t.type)return{...r,flows:{clientCredentials:{...n,token:t.token,clientSecret:t.clientSecret,scopes:n.scopes,"x-scalar-client-id":r["x-scalar-client-id"]}}};if("authorizationCode"===n.type&&"oauth-authorizationCode"===t.type)return{...r,flows:{authorizationCode:{...n,token:t.token,clientSecret:t.clientSecret,scopes:n.scopes,"x-scalar-client-id":r["x-scalar-client-id"]}}}}return null})(n,a);return o&&(t[n.uid]=o),t}),{}),n=e.servers;return{...e,securitySchemes:t,servers:n}})(t)),Zo(e,"2.3.0")&&(t=(e=>{console.info("Performing data migration v-2.2.0 to v-2.3.0");const t=e.environments,n=Object.values(e.workspaces).reduce(((e,n)=>{const r={};return Object.entries(t).forEach((([e,t])=>{const n="string"==typeof t.value?JSON.parse(t.value):t.value;"default"===e&&Object.assign(r,n)})),e[n.uid]={...n,environments:r},e}),{}),r=Object.values(e.collections).reduce(((e,t)=>(e[t.uid]={...t,"x-scalar-environments":t["x-scalar-environments"]||{}},e)),{});return Object.values(n).forEach((e=>{Object.entries(t).forEach((([t,n])=>{if("default"!==t){const a="string"==typeof n.value?JSON.parse(n.value):n.value,o=n.name;Object.values(r).forEach((n=>{n["x-scalar-environments"]=n["x-scalar-environments"]||{},n["x-scalar-environments"][o]={variables:a},e.activeEnvironmentId===t&&(n["x-scalar-active-environment"]=o??"")}))}}))})),{...e,collections:r,workspaces:n}})(t)),Zo(e,"2.4.0")&&(t=(e=>{console.info("Performing data migration v-2.3.0 to v-2.4.0");const t=Object.values(e.collections).reduce(((t,n)=>("Drafts"===n.info?.title&&(n.servers=[],Object.values(e.requests).forEach((t=>{if(t.selectedServerUid&&n.requests.includes(t.uid)){const n=e.servers[t.selectedServerUid];n&&(t.path=`${n.url}${t.path}`),t.selectedServerUid=""}}))),t[n.uid]=n,t)),{});return{...e,collections:t}})(t)),Zo(e,"2.5.0")&&(t=(e=>{console.info("Performing data migration v-2.4.0 to v-2.5.0");const t=Object.entries(e.cookies||{}).reduce(((e,[t,n])=>(e[t]={...n,uid:n.uid},e)),{}),n=Object.entries(e.collections||{}).reduce(((e,[t,n])=>(e[t]={...n,uid:n.uid,selectedSecuritySchemeUids:n.selectedSecuritySchemeUids,servers:n.servers.map((e=>e)),tags:n.tags.map((e=>e)),requests:n.requests.map((e=>e)),children:n.children.map((e=>e)),selectedServerUid:n.selectedServerUid},e)),{}),r=Object.entries(e.environments||{}).reduce(((e,[t,n])=>(e[t]={...n,uid:n.uid},e)),{}),a=Object.entries(e.requests||{}).reduce(((e,[t,n])=>(e[t]={...n,uid:n.uid,servers:n.servers,selectedServerUid:n.selectedServerUid,examples:n.examples,selectedSecuritySchemeUids:n.selectedSecuritySchemeUids},e)),{}),o=Object.entries(e.requestExamples||{}).reduce(((e,[t,n])=>{const r=n.parameters.headers;return r.some((e=>"accept"===e.key.toLowerCase()))||r.unshift({key:"Accept",value:"*/*",enabled:!0}),e[t]={...n,uid:n.uid,requestUid:n.requestUid,parameters:{...n.parameters,headers:r}},e}),{}),i=Object.entries(e.securitySchemes||{}).reduce(((e,[t,n])=>(e[t]={...n,uid:n.uid},e)),{}),s=Object.entries(e.servers||{}).reduce(((e,[t,n])=>(e[t]={...n,uid:n.uid,variables:Object.entries(n.variables||{}).reduce(((e,[t,n])=>(e[t]={...n,enum:n.enum&&n.enum.length>0?n.enum:void 0,default:n.default??"",description:n.description},e)),{})},e)),{}),l=Object.entries(e.tags||{}).reduce(((e,[t,n])=>(e[t]={...n,uid:n.uid,children:n.children},e)),{}),c=Object.entries(e.workspaces||{}).reduce(((e,[t,n])=>(e[t]={...n,uid:n.uid,collections:n.collections.map((e=>e)),cookies:n.cookies.map((e=>e)),selectedHttpClient:{targetKey:"shell",clientKey:"curl"}},e)),{});return{...e,collections:n,cookies:t,environments:r,requests:a,requestExamples:o,securitySchemes:i,servers:s,tags:l,workspaces:c}})(t)),t={collections:Object.values(t.collections),cookies:Object.values(t.cookies),environments:Object.values(t.environments),requestExamples:Object.values(t.requestExamples),requests:Object.values(t.requests),securitySchemes:Object.values(t.securitySchemes),servers:Object.values(t.servers),tags:Object.values(t.tags),workspaces:Object.values(t.workspaces)},t})();Ko(e,jo.CV,t.rawAdd),Ko(u,Uo.c,n.add),Ko(d,zo.V,r.add),Ko(p,jo.$W,o.rawAdd),Ko(h,jo.TF,i.rawAdd),Ko(f,jo.zT,s.rawAdd),Ko(m,jo.NV,l.rawAdd),Ko(g,jo.c3,a.rawAdd),Ko(v,Wo.N,c.rawAdd),localStorage.setItem(qo,Xo)}catch(e){console.error(e)}})(c)}else if(r&&n.spec){const e=Wo.N.parse({uid:"default",name:"Workspace",isReadOnly:!0,proxyUrl:n.proxyUrl});c.workspaceMutators.rawAdd(e)}else c.workspaceMutators.add({uid:"default",name:"Workspace",proxyUrl:n.proxyUrl}),localStorage.setItem(qo,Xo);const d=(0,rt.Ef)(t);d.use(l),d.provide(St.pd,c),d.provide(Et.K,s),d.provide(_t.UF,u);const{collectionMutators:p,importSpecFile:h,importSpecFromUrl:f,modalState:m,requests:g,securitySchemes:v,securitySchemeMutators:b,servers:y,workspaceMutators:w,requestExampleMutators:O}=c,{activeCollection:k,activeWorkspace:x}=u,S=(t=e)=>{t?d.mount(t):console.error("[@scalar/api-client-modal] Could not create the API client.","Invalid HTML element provided.","Read more: https://github.com/scalar/scalar/tree/main/packages/api-client")};i&&S();const _=async e=>{var t,r;null!=e&&e.url?await f(e.url,(null==(t=x.value)?void 0:t.uid)??"",{...n,setCollectionSecurity:!0}):null!=e&&e.content?await h(null==e?void 0:e.content,(null==(r=x.value)?void 0:r.uid)??"",{...n,setCollectionSecurity:!0}):console.error("[@scalar/api-client-modal] Could not create the API client.","Please provide an OpenAPI document: { spec: { url: '…' } }","Read more: https://github.com/scalar/scalar/tree/main/packages/api-client")};return{app:d,updateSpec:_,updateConfig(e,t=!0){var r;t?Object.assign(n??{},e):(0,Xe.O1)(n??{},e),e.spec&&(c.collectionMutators.reset(),c.requestMutators.reset(),c.requestExampleMutators.reset(),c.securitySchemeMutators.reset(),c.serverMutators.reset(),c.tagMutators.reset(),w.edit(null==(r=x.value)?void 0:r.uid,"collections",[]),_(e.spec))},updateServer:e=>{var t;const n=Object.values(y).find((t=>t.url===e));n&&k.value&&p.edit(null==(t=k.value)?void 0:t.uid,"selectedServerUid",n.uid)},onUpdateServer:e=>{(0,E.wB)((()=>{var e;return null==(e=k.value)?void 0:e.selectedServerUid}),(t=>{const n=Object.values(y).find((e=>e.uid===t));null!=n&&n.url&&e(n.url)}))},updateAuth:({nameKey:e,propertyKey:t,value:n})=>{const r=Object.values(v).find((t=>t.nameKey===e));r&&b.edit(r.uid,t,n)},route:e=>{var t;const{requestUid:n,method:r,path:a,_source:o}=e??{},i=n||(null==(t=Object.values(g).find((e=>!(a&&r&&e.path&&e.method)||e.path===a&&e.method.toUpperCase()===r.toUpperCase())))?void 0:t.uid);i&&l.push({name:"request",query:o?{source:o}:{},params:{workspace:"default",request:i}})},open:e=>{var t;const{path:n,method:r,requestUid:a,_source:o}=e??{},i=a||(null==(t=Object.values(g).find((e=>!(n&&r&&e.path&&e.method)||e.path===n&&e.method.toUpperCase()===r.toUpperCase())))?void 0:t.uid);i&&l.push({name:"request",query:o?{source:o}:{},params:{workspace:"default",request:i}}),m.open=!0},mount:S,modalState:m,store:c,updateExample:(e,t)=>{var n,r,a,o;if(!e||!t)return;const i=Object.values(g).find((({operationId:e,path:n})=>e===t||n===t));if(!i)return;const s=Object.keys((null==(n=i.requestBody)?void 0:n.content)||{})[0]||"",l=null==(o=null==(a=null==(r=i.requestBody)?void 0:r.content)?void 0:a[s])?void 0:o.examples[e];l&&O.edit(i.examples[0],"body.raw.value",(0,Xe.qX)(l.value))}}},Jo=(0,S.KR)(null),ei=()=>({client:Jo,init:async e=>{const t=await(async({el:e=null,configuration:t={},mountOnInitialize:n=!0,store:r})=>{var a,o,i;const s={...t,showSidebar:!1},l=Yo({el:e,appComponent:Fo,configuration:s,persistData:!1,isReadOnly:!0,store:r,mountOnInitialize:n,router:(0,Mo.aE)({history:(0,Mo.sC)(),routes:Qo.Go}),layout:"modal"}),{importSpecFile:c,importSpecFromUrl:u}=l.store;return r||(null!=(a=t.spec)&&a.url?await u(t.spec.url,"default",{proxyUrl:t.proxyUrl,setCollectionSecurity:!0,...t}):null!=(o=t.spec)&&o.content&&await c(null==(i=t.spec)?void 0:i.content,"default",{setCollectionSecurity:!0,...t})),l})(e);return Jo.value=t,t}}),ti=["method"],ni=It((0,E.pM)({__name:"TestRequestButton",props:{operation:{}},setup(e){const{client:t}=ei(),n=ot(),r=(0,E.EW)((()=>!0!==(null==n?void 0:n.hideTestRequestButton))),a=()=>{var n;e.operation&&null!=(n=null==t?void 0:t.value)&&n.open&&t.value.open({requestUid:e.operation.uid})};return(e,t)=>e.operation&&r.value?((0,E.uX)(),(0,E.CE)("button",{key:0,class:"show-api-client-button",method:e.operation.method,type:"button",onClick:(0,rt.D$)(a,["stop"])},[(0,E.bF)((0,S.R1)(Ct.kv),{icon:"Play",size:"sm"}),t[0]||(t[0]=(0,E.Lk)("span",null,"Test Request",-1)),(0,E.bF)(nr,null,{default:(0,E.k6)((()=>[(0,E.eW)("("+(0,_.v_)(e.operation.method)+" "+(0,_.v_)(e.operation.path)+")",1)])),_:1})],8,ti)):((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[(0,E.eW)(" ")],64))}}),[["__scopeId","data-v-ea8189a2"]]);var ri=n(6941);n(3213),n(498);const ai={key:2,class:"empty-state"},oi=It((0,E.pM)({__name:"ExampleResponse",props:{response:{}},setup:e=>(e,t)=>{var n,r,a,o;return null!=(n=e.response)&&n.example?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.z5),{key:0,class:"bg-b-2",content:null==(r=e.response)?void 0:r.example,lang:"json"},null,8,["content"])):null!=(a=e.response)&&a.schema?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.z5),{key:1,class:"bg-b-2",content:(0,S.R1)(ri.M)(null==(o=e.response)?void 0:o.schema,{emptyString:"…",mode:"read"}),lang:"json"},null,8,["content"])):((0,E.uX)(),(0,E.CE)("div",ai," No Body "))}}),[["__scopeId","data-v-4b51e8ca"]]);var ii=n(1972),si=n(805);let li=(0,E.pM)({props:{onFocus:{type:Function,required:!0}},setup(e){let t=(0,S.KR)(!0);return()=>t.value?(0,E.h)(si.j,{as:"button",type:"button",features:si.O.Focusable,onFocus(n){n.preventDefault();let r,a=50;r=requestAnimationFrame((function n(){var o;if(!(a--<=0))return null!=(o=e.onFocus)&&o.call(e)?(t.value=!1,void cancelAnimationFrame(r)):void(r=requestAnimationFrame(n));r&&cancelAnimationFrame(r)}))}}):null}});var ci,ui,di=n(7838),pi=n(6413),hi=n(1417),fi=n(9388),mi=n(3827),gi=n(562),vi=n(3605),bi=((ui=bi||{})[ui.Forwards=0]="Forwards",ui[ui.Backwards=1]="Backwards",ui),yi=((ci=yi||{})[ci.Less=-1]="Less",ci[ci.Equal=0]="Equal",ci[ci.Greater=1]="Greater",ci);let wi=Symbol("TabsContext");function Oi(e){let t=(0,E.WQ)(wi,null);if(null===t){let t=new Error(`<${e} /> is missing a parent <TabGroup /> component.`);throw Error.captureStackTrace&&Error.captureStackTrace(t,Oi),t}return t}let ki=Symbol("TabsSSRContext"),xi=(0,E.pM)({name:"TabGroup",emits:{change:e=>!0},props:{as:{type:[Object,String],default:"template"},selectedIndex:{type:[Number],default:null},defaultIndex:{type:[Number],default:0},vertical:{type:[Boolean],default:!1},manual:{type:[Boolean],default:!1}},inheritAttrs:!1,setup(e,{slots:t,attrs:n,emit:r}){var a;let o=(0,S.KR)(null!=(a=e.selectedIndex)?a:e.defaultIndex),i=(0,S.KR)([]),s=(0,S.KR)([]),l=(0,E.EW)((()=>null!==e.selectedIndex)),c=(0,E.EW)((()=>l.value?e.selectedIndex:o.value));function u(e){var t;let n=(0,hi.wl)(d.tabs.value,pi.t),r=(0,hi.wl)(d.panels.value,pi.t),a=n.filter((e=>{var t;return!(null!=(t=(0,pi.t)(e))&&t.hasAttribute("disabled"))}));if(e<0||e>n.length-1){let t=(0,fi.Y)(null===o.value?0:Math.sign(e-o.value),{[-1]:()=>1,0:()=>(0,fi.Y)(Math.sign(e),{[-1]:()=>0,0:()=>0,1:()=>1}),1:()=>0}),i=(0,fi.Y)(t,{0:()=>n.indexOf(a[0]),1:()=>n.indexOf(a[a.length-1])});-1!==i&&(o.value=i),d.tabs.value=n,d.panels.value=r}else{let i=n.slice(0,e),s=[...n.slice(e),...i].find((e=>a.includes(e)));if(!s)return;let l=null!=(t=n.indexOf(s))?t:d.selectedIndex.value;-1===l&&(l=d.selectedIndex.value),o.value=l,d.tabs.value=n,d.panels.value=r}}let d={selectedIndex:(0,E.EW)((()=>{var t,n;return null!=(n=null!=(t=o.value)?t:e.defaultIndex)?n:null})),orientation:(0,E.EW)((()=>e.vertical?"vertical":"horizontal")),activation:(0,E.EW)((()=>e.manual?"manual":"auto")),tabs:i,panels:s,setSelectedIndex(e){c.value!==e&&r("change",e),l.value||u(e)},registerTab(e){var t;if(i.value.includes(e))return;let n=i.value[o.value];if(i.value.push(e),i.value=(0,hi.wl)(i.value,pi.t),!l.value){let e=null!=(t=i.value.indexOf(n))?t:o.value;-1!==e&&(o.value=e)}},unregisterTab(e){let t=i.value.indexOf(e);-1!==t&&i.value.splice(t,1)},registerPanel(e){s.value.includes(e)||(s.value.push(e),s.value=(0,hi.wl)(s.value,pi.t))},unregisterPanel(e){let t=s.value.indexOf(e);-1!==t&&s.value.splice(t,1)}};(0,E.Gt)(wi,d);let p=(0,S.KR)({tabs:[],panels:[]}),h=(0,S.KR)(!1);(0,E.sV)((()=>{h.value=!0})),(0,E.Gt)(ki,(0,E.EW)((()=>h.value?null:p.value)));let f=(0,E.EW)((()=>e.selectedIndex));return(0,E.sV)((()=>{(0,E.wB)([f],(()=>{var t;return u(null!=(t=e.selectedIndex)?t:e.defaultIndex)}),{immediate:!0})})),(0,E.nT)((()=>{if(!l.value||null==c.value||d.tabs.value.length<=0)return;let e=(0,hi.wl)(d.tabs.value,pi.t);e.some(((e,t)=>(0,pi.t)(d.tabs.value[t])!==(0,pi.t)(e)))&&d.setSelectedIndex(e.findIndex((e=>(0,pi.t)(e)===(0,pi.t)(d.tabs.value[c.value]))))})),()=>{let r={selectedIndex:o.value};return(0,E.h)(E.FK,[i.value.length<=0&&(0,E.h)(li,{onFocus:()=>{for(let e of i.value){let t=(0,pi.t)(e);if(0===(null==t?void 0:t.tabIndex))return t.focus(),!0}return!1}}),(0,vi.XX)({theirProps:{...n,...(0,vi.cJ)(e,["selectedIndex","defaultIndex","manual","vertical","onChange"])},ourProps:{},slot:r,slots:t,attrs:n,name:"TabGroup"})])}}}),Ei=(0,E.pM)({name:"TabList",props:{as:{type:[Object,String],default:"div"}},setup(e,{attrs:t,slots:n}){let r=Oi("TabList");return()=>{let a={selectedIndex:r.selectedIndex.value},o={role:"tablist","aria-orientation":r.orientation.value};return(0,vi.XX)({ourProps:o,theirProps:e,slot:a,attrs:t,slots:n,name:"TabList"})}}}),Si=(0,E.pM)({name:"Tab",props:{as:{type:[Object,String],default:"button"},disabled:{type:[Boolean],default:!1},id:{type:String,default:null}},setup(e,{attrs:t,slots:n,expose:r}){var a;let o=null!=(a=e.id)?a:`headlessui-tabs-tab-${(0,xt.B)()}`,i=Oi("Tab"),s=(0,S.KR)(null);r({el:s,$el:s}),(0,E.sV)((()=>i.registerTab(s))),(0,E.hi)((()=>i.unregisterTab(s)));let l=(0,E.WQ)(ki),c=(0,E.EW)((()=>{if(l.value){let e=l.value.tabs.indexOf(o);return-1===e?l.value.tabs.push(o)-1:e}return-1})),u=(0,E.EW)((()=>{let e=i.tabs.value.indexOf(s);return-1===e?c.value:e})),d=(0,E.EW)((()=>u.value===i.selectedIndex.value));function p(e){var t;let n=e();if(n===hi.Me.Success&&"auto"===i.activation.value){let e=null==(t=(0,gi.T)(s))?void 0:t.activeElement,n=i.tabs.value.findIndex((t=>(0,pi.t)(t)===e));-1!==n&&i.setSelectedIndex(n)}return n}function h(e){let t=i.tabs.value.map((e=>(0,pi.t)(e))).filter(Boolean);if(e.key===di.D.Space||e.key===di.D.Enter)return e.preventDefault(),e.stopPropagation(),void i.setSelectedIndex(u.value);switch(e.key){case di.D.Home:case di.D.PageUp:return e.preventDefault(),e.stopPropagation(),p((()=>(0,hi.CU)(t,hi.BD.First)));case di.D.End:case di.D.PageDown:return e.preventDefault(),e.stopPropagation(),p((()=>(0,hi.CU)(t,hi.BD.Last)))}return p((()=>(0,fi.Y)(i.orientation.value,{vertical:()=>e.key===di.D.ArrowUp?(0,hi.CU)(t,hi.BD.Previous|hi.BD.WrapAround):e.key===di.D.ArrowDown?(0,hi.CU)(t,hi.BD.Next|hi.BD.WrapAround):hi.Me.Error,horizontal:()=>e.key===di.D.ArrowLeft?(0,hi.CU)(t,hi.BD.Previous|hi.BD.WrapAround):e.key===di.D.ArrowRight?(0,hi.CU)(t,hi.BD.Next|hi.BD.WrapAround):hi.Me.Error})))===hi.Me.Success?e.preventDefault():void 0}let f=(0,S.KR)(!1);function m(){var t;f.value||(f.value=!0,!e.disabled&&(null==(t=(0,pi.t)(s))||t.focus({preventScroll:!0}),i.setSelectedIndex(u.value),(0,mi._)((()=>{f.value=!1}))))}function g(e){e.preventDefault()}let v=(0,ii.c)((0,E.EW)((()=>({as:e.as,type:t.type}))),s);return()=>{var r,a;let l={selected:d.value,disabled:null!=(r=e.disabled)&&r},{...c}=e,p={ref:s,onKeydown:h,onMousedown:g,onClick:m,id:o,role:"tab",type:v.value,"aria-controls":null==(a=(0,pi.t)(i.panels.value[u.value]))?void 0:a.id,"aria-selected":d.value,tabIndex:d.value?0:-1,disabled:!!e.disabled||void 0};return(0,vi.XX)({ourProps:p,theirProps:c,slot:l,attrs:t,slots:n,name:"Tab"})}}});(0,E.pM)({name:"TabPanels",props:{as:{type:[Object,String],default:"div"}},setup(e,{slots:t,attrs:n}){let r=Oi("TabPanels");return()=>{let a={selectedIndex:r.selectedIndex.value};return(0,vi.XX)({theirProps:e,ourProps:{},slot:a,attrs:n,slots:t,name:"TabPanels"})}}}),(0,E.pM)({name:"TabPanel",props:{as:{type:[Object,String],default:"div"},static:{type:Boolean,default:!1},unmount:{type:Boolean,default:!0},id:{type:String,default:null},tabIndex:{type:Number,default:0}},setup(e,{attrs:t,slots:n,expose:r}){var a;let o=null!=(a=e.id)?a:`headlessui-tabs-panel-${(0,xt.B)()}`,i=Oi("TabPanel"),s=(0,S.KR)(null);r({el:s,$el:s}),(0,E.sV)((()=>i.registerPanel(s))),(0,E.hi)((()=>i.unregisterPanel(s)));let l=(0,E.WQ)(ki),c=(0,E.EW)((()=>{if(l.value){let e=l.value.panels.indexOf(o);return-1===e?l.value.panels.push(o)-1:e}return-1})),u=(0,E.EW)((()=>{let e=i.panels.value.indexOf(s);return-1===e?c.value:e})),d=(0,E.EW)((()=>u.value===i.selectedIndex.value));return()=>{var r;let a={selected:d.value},{tabIndex:l,...c}=e,p={ref:s,id:o,role:"tabpanel","aria-labelledby":null==(r=(0,pi.t)(i.tabs.value[u.value]))?void 0:r.id,tabIndex:d.value?l:-1};return d.value||!e.unmount||e.static?(0,vi.XX)({ourProps:p,theirProps:c,slot:a,attrs:t,slots:n,features:vi.O5.Static|vi.O5.RenderStrategy,visible:d.value,name:"TabPanel"}):(0,E.h)(si.j,{as:"span","aria-hidden":!0,...p})}}});const _i=It((0,E.pM)({__name:"CardTabHeader",emits:["change"],setup(e,{emit:t}){const n=t,r=e=>{n("change",e)};return(e,t)=>((0,E.uX)(),(0,E.Wv)(Er,{class:"scalar-card-header scalar-card-header-tabs"},{actions:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"actions",{},void 0,!0)])),default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(xi),{onChange:r},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Ei),{class:"tab-list custom-scroll"},{default:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"default",{},void 0,!0)])),_:3})])),_:3})])),_:3}))}}),[["__scopeId","data-v-03fdb072"]]),Ci=It((0,E.pM)({__name:"CardTab",setup:e=>(e,t)=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Si),{as:"template"},{default:(0,E.k6)((({selected:t})=>[(0,E.Lk)("button",{class:(0,_.C4)(["tab",{"tab-selected":t}]),type:"button"},[(0,E.Lk)("span",null,[(0,E.RG)(e.$slots,"default",{},void 0,!0)])],2)])),_:3}))}),[["__scopeId","data-v-d0e07e65"]]),Ti=It((0,E.pM)({__name:"CardFooter",props:{muted:{type:Boolean},contrast:{type:Boolean},frameless:{type:Boolean},transparent:{type:Boolean},borderless:{type:Boolean}},setup(e){const t=e;return(e,n)=>((0,E.uX)(),(0,E.Wv)(wr,(0,_._B)((0,E.Ng)(t)),{default:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"default",{},void 0,!0)])),_:3},16))}}),[["__scopeId","data-v-7621d1be"]]),Ai={class:"text-select-label"},Ri=["aria-controls","tabindex","value"],Pi=["label"],Di=["value"],Mi=["value"],$i=(0,E.pM)({__name:"TextSelect",props:{modelValue:{},options:{},controls:{}},emits:["update:modelValue"],setup(e){const t=e,n=(0,E.EW)((()=>t.options.flatMap((e=>e.options??e)).length));return(e,t)=>((0,E.uX)(),(0,E.CE)("label",{class:(0,_.C4)(["text-select",1===n.value?"text-select--single-option":""])},[(0,E.Lk)("span",Ai,[(0,E.RG)(e.$slots,"default")]),(0,E.Lk)("select",{"aria-controls":e.controls,tabindex:1===e.options.length?-1:0,value:e.modelValue,onInput:t[0]||(t[0]=t=>e.$emit("update:modelValue",t.target.value))},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(e.options,(e=>((0,E.uX)(),(0,E.CE)(E.FK,{key:e.value},[e.options?((0,E.uX)(),(0,E.CE)("optgroup",{key:0,label:e.label},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(e.options,(e=>((0,E.uX)(),(0,E.CE)("option",{key:e.value,value:e.value},(0,_.v_)(e.label),9,Di)))),128))],8,Pi)):((0,E.uX)(),(0,E.CE)("option",{key:e.value,value:e.value},(0,_.v_)(e.label),9,Mi))],64)))),128))],40,Ri)],2))}}),Ni=It((0,E.pM)({__name:"ExamplePicker",props:{examples:{}},emits:["update:modelValue"],setup(e,{emit:t}){const n=e,r=t,a=(0,S.KR)(Object.keys(n.examples)[0]);function o(e){if(!e)return"Select an example";const t=n.examples[e];return(null==t?void 0:t.summary)??e}return(0,E.wB)((()=>n.examples),(()=>{var e;(e=Object.keys(n.examples)[0])&&(a.value=e)}),{immediate:!0}),(0,E.wB)(a,(()=>{a.value&&r("update:modelValue",a.value)}),{immediate:!0}),(e,t)=>((0,E.uX)(),(0,E.Wv)($i,{modelValue:a.value,"onUpdate:modelValue":t[0]||(t[0]=e=>a.value=e),class:"example-selector",options:Object.keys(e.examples).map((e=>({label:o(e),value:e})))},{default:(0,E.k6)((()=>[(0,E.bF)(nr,null,{default:(0,E.k6)((()=>t[1]||(t[1]=[(0,E.eW)("Selected Example Values:")]))),_:1}),(0,E.eW)(" "+(0,_.v_)(o(a.value)),1)])),_:1},8,["modelValue","options"]))}}),[["__scopeId","data-v-a7e22e14"]]),Ii={key:1,class:"scalar-card-checkbox"},Li=["aria-controls"],Bi={class:"scalar-card-container custom-scroll"},Fi={key:1,class:"response-description"},Qi=It((0,E.pM)({__name:"ExampleResponses",props:{responses:{}},setup(e){const t=(0,E.Bi)(),{copyToClipboard:n}=(0,lr.i)(),r=(0,S.KR)(),a=(0,E.EW)((()=>Object.keys(e.responses??{}).sort())),o=(0,E.EW)((()=>!!l.value.examples)),i=(0,S.KR)(0),s=(0,E.EW)((()=>{var t;const n=a.value[i.value];return null==(t=e.responses)?void 0:t[n]})),l=(0,E.EW)((()=>{var e;const t=(0,Xe.LQ)(null==(e=s.value)?void 0:e.content);return(null==t?void 0:t["application/json"])??(null==t?void 0:t["application/xml"])??(null==t?void 0:t["text/plain"])??(null==t?void 0:t["text/html"])??s.value})),c=()=>{if(o.value){if(Array.isArray(l.value.examples))return l.value.examples[0];{const e=Object.keys(l.value.examples)[0];return l.value.examples[e]}}return l.value.example},u=(0,E.EW)((()=>({...l.value,example:o.value&&r.value?l.value.examples[r.value].value??l.value.examples[r.value]:c()}))),d=e=>{i.value=e,r.value=void 0},p=(0,S.KR)(!1);return(e,i)=>a.value.length?((0,E.uX)(),(0,E.Wv)((0,S.R1)(yr),{key:0},{default:(0,E.k6)((()=>{var e;return[(0,E.bF)((0,S.R1)(_i),{muted:"",x:"as",onChange:d},{actions:(0,E.k6)((()=>{var e,r;return[null!=(e=l.value)&&e.example?((0,E.uX)(),(0,E.CE)("button",{key:0,class:"code-copy",type:"button",onClick:i[0]||(i[0]=()=>{var e;return(0,S.R1)(n)(null==(e=l.value)?void 0:e.example)})},[(0,E.bF)((0,S.R1)(Ct.kv),{icon:"Clipboard",width:"12px"})])):(0,E.Q3)("",!0),null!=(r=l.value)&&r.schema?((0,E.uX)(),(0,E.CE)("label",Ii,[i[3]||(i[3]=(0,E.eW)(" Show Schema ")),(0,E.bo)((0,E.Lk)("input",{"onUpdate:modelValue":i[1]||(i[1]=e=>p.value=e),"aria-controls":(0,S.R1)(t),class:"scalar-card-checkbox-input",type:"checkbox"},null,8,Li),[[rt.lH,p.value]]),i[4]||(i[4]=(0,E.Lk)("span",{class:"scalar-card-checkbox-checkmark"},null,-1))])):(0,E.Q3)("",!0)]})),default:(0,E.k6)((()=>[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(a.value,(e=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ci),{key:e,"aria-controls":(0,S.R1)(t)},{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(e),1)])),_:2},1032,["aria-controls"])))),128))])),_:1}),(0,E.Lk)("div",Bi,[(0,E.bF)((0,S.R1)(wr),{muted:""},{default:(0,E.k6)((()=>{var e;return[null!=(e=l.value)&&e.schema?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[p.value&&u.value?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.z5),{key:0,id:(0,S.R1)(t),content:u.value,lang:"json"},null,8,["id","content"])):((0,E.uX)(),(0,E.Wv)(oi,{key:1,id:(0,S.R1)(t),response:u.value},null,8,["id","response"]))],64)):((0,E.uX)(),(0,E.Wv)(oi,{key:1,id:(0,S.R1)(t),response:u.value},null,8,["id","response"]))]})),_:1})]),null!=(e=s.value)&&e.description||o.value?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ti),{key:0,class:"response-card-footer",muted:""},{default:(0,E.k6)((()=>{var e,t;return[o.value?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ni),{key:0,class:"response-example-selector",examples:null==(e=l.value)?void 0:e.examples,"onUpdate:modelValue":i[2]||(i[2]=e=>r.value=e)},null,8,["examples"])):null!=(t=s.value)&&t.description?((0,E.uX)(),(0,E.CE)("div",Fi,[(0,E.bF)((0,S.R1)(Ct.mi),{class:"markdown",value:s.value.description},null,8,["value"])])):(0,E.Q3)("",!0)]})),_:1})):(0,E.Q3)("",!0)]})),_:1})):(0,E.Q3)("",!0)}}),[["__scopeId","data-v-d84b9039"]]);n(1902),n(2453);var Ui=n(3179),zi=(n(3813),n(7525),n(2816));const ji=(0,S.KR)(),Wi=(0,S.KR)(),Xi=()=>({selectedExampleKey:ji,operationId:Wi}),qi=["id"],Vi=["id"],Hi={key:0,class:"request-card-footer-addon"},Zi={class:"request-header"},Gi=It((0,E.pM)({__name:"ExampleRequest",props:{operation:{},server:{},collection:{},fallback:{type:Boolean},transformedOperation:{}},setup(e){const{selectedExampleKey:t,operationId:n}=Xi(),{requestExamples:r,securitySchemes:a}=(0,St.hA)(),o=ot(),{httpClient:i,setHttpClient:s,availableTargets:l,httpTargetTitle:c,httpClientTitle:u}=kt(),d=(0,E.Bi)(),p=(0,E.EW)((()=>{var t;const n=["x-custom-examples","x-codeSamples","x-code-samples"];for(const r of n)if(null!=(t=e.transformedOperation.information)&&t[r])return[...e.transformedOperation.information[r]];return[]})),h=(0,S.KR)(p.value.length?{targetKey:"customExamples",clientKey:0}:{targetKey:i.targetKey,clientKey:i.clientKey});(0,E.wB)(i,(()=>{h.value={targetKey:i.targetKey,clientKey:i.clientKey}}));const f=(0,E.EW)((()=>{var t,n,r,a;return Object.keys((null==(a=null==(r=null==(n=null==(t=e.transformedOperation.information)?void 0:t.requestBody)?void 0:n.content)?void 0:r["application/json"])?void 0:a.examples)??{}).length>1})),m=(0,E.EW)((()=>{try{return(()=>{var t;if("customExamples"===h.value.targetKey)return(null==(t=p.value[h.value.clientKey])?void 0:t.source)??"";const n=i.clientKey,o=i.targetKey,s=r[e.operation.examples[0]];if(!s)return"";const l=(0,zi.z)(e.operation.security||e.collection.security,e.collection.selectedSecuritySchemeUids,a),[c,u]=(0,Ui.y)(o,n,{operation:e.operation,example:s,server:e.server,securitySchemes:l});return c?c.message??"":u})()}catch(e){return console.error("[generateSnippet]",e),""}})),g=(0,E.EW)((()=>{var e,t;const n="customExamples"===(null==(e=h.value)?void 0:e.targetKey)?(null==(t=p.value[h.value.clientKey])?void 0:t.lang)??"plaintext":i.targetKey;return"shell"===n&&m.value.includes("curl")?"curl":"Objective-C"===n?"objc":n})),v=(0,E.EW)((()=>Object.values(a).flatMap((e=>"apiKey"===e.type?e.value:"http"===(null==e?void 0:e.type)?[e.token,e.password,btoa(`${e.username}:${e.password}`)]:"oauth2"===e.type?Object.values(e.flows).map((e=>e.token)):[])))),b=(0,E.EW)((()=>{const e=l.value.map((e=>({value:e.key,label:e.title,options:e.clients.map((t=>({value:JSON.stringify({targetKey:e.key,clientKey:t.client}),label:t.title})))})));return p.value.length&&e.unshift({value:"customExamples",label:"Examples",options:p.value.map(((e,t)=>({value:JSON.stringify({targetKey:"customExamples",clientKey:t}),label:e.label??e.lang??`Example #${t+1}`})))}),e}));function y(e){const t=JSON.parse(e);"customExamples"===t.targetKey?h.value=t:s(t)}return(e,r)=>(0,S.R1)(l).length||p.value.length?((0,E.uX)(),(0,E.Wv)((0,S.R1)(yr),{key:0,class:"dark-mode"},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Er),{muted:""},{actions:(0,E.k6)((()=>[(0,E.bF)($i,{class:"request-client-picker",controls:`${(0,S.R1)(d)}-example`,modelValue:JSON.stringify(h.value),options:b.value,"onUpdate:modelValue":y},{default:(0,E.k6)((()=>["customExamples"===h.value.targetKey?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[(0,E.bF)(nr,null,{default:(0,E.k6)((()=>r[1]||(r[1]=[(0,E.eW)("Selected Example:")]))),_:1}),(0,E.eW)(" "+(0,_.v_)(p.value[h.value.clientKey].label??"Example"),1)],64)):((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[(0,E.bF)(nr,null,{default:(0,E.k6)((()=>r[2]||(r[2]=[(0,E.eW)("Selected HTTP client:")]))),_:1}),(0,E.eW)(" "+(0,_.v_)((0,S.R1)(c))+" "+(0,_.v_)((0,S.R1)(u)),1)],64))])),_:1},8,["controls","modelValue","options"])])),default:(0,E.k6)((()=>[(0,E.Lk)("div",{id:`${(0,S.R1)(d)}-header`,class:"request-header"},[(0,E.bF)((0,S.R1)(qt),{as:"span",class:"request-method",method:e.operation.method},null,8,["method"]),(0,E.RG)(e.$slots,"header",{},void 0,!0)],8,qi)])),_:3}),(0,E.bF)((0,S.R1)(wr),{borderless:"",class:"request-editor-section custom-scroll",frameless:""},{default:(0,E.k6)((()=>[(0,E.Lk)("div",{id:`${(0,S.R1)(d)}-example`,class:"code-snippet"},[(0,E.bF)((0,S.R1)(Ct.z5),{class:"bg-b-2",content:m.value,hideCredentials:v.value,lang:g.value,lineNumbers:""},null,8,["content","hideCredentials","lang"])],8,Vi)])),_:1}),!f.value&&(0,S.R1)(o).hideTestRequestButton||!e.$slots.footer?(0,E.Q3)("",!0):((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ti),{key:0,class:"request-card-footer",contrast:""},{default:(0,E.k6)((()=>{var a,o,i,s;return[f.value?((0,E.uX)(),(0,E.CE)("div",Hi,[(0,E.bF)(Ni,{class:"request-example-selector",examples:(null==(s=null==(i=null==(o=null==(a=e.transformedOperation.information)?void 0:a.requestBody)?void 0:o.content)?void 0:i["application/json"])?void 0:s.examples)??[],"onUpdate:modelValue":r[0]||(r[0]=r=>(t.value=r,n.value=e.operation.operationId))},null,8,["examples"])])):(0,E.Q3)("",!0),(0,E.RG)(e.$slots,"footer",{},void 0,!0)]})),_:3}))])),_:3})):e.fallback?((0,E.uX)(),(0,E.Wv)((0,S.R1)(yr),{key:1,class:"dark-mode"},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(wr),{class:"request-card-simple"},{default:(0,E.k6)((()=>[(0,E.Lk)("div",Zi,[(0,E.bF)((0,S.R1)(qt),{as:"span",class:"request-method",method:e.operation.method},null,8,["method"]),(0,E.RG)(e.$slots,"header",{},void 0,!0)]),(0,E.RG)(e.$slots,"footer",{},void 0,!0)])),_:3})])),_:3})):(0,E.Q3)("",!0)}}),[["__scopeId","data-v-3ff33ef0"]]),Ki={class:"operation-title"},Yi={class:"operation-details"},Ji={class:"endpoint-label"},es={class:"endpoint-label-path"},ts={class:"endpoint-label-name"},ns={class:"endpoint-content"},rs={class:"operation-details-card"},as={class:"operation-details-card-item"},os={class:"operation-details-card-item"},is=It((0,E.pM)({__name:"ClassicLayout",props:{id:{},collection:{},server:{},operation:{},transformedOperation:{}},setup(e){const{copyToClipboard:t}=(0,lr.i)(),n=ot();return(e,r)=>{var a;return(0,E.uX)(),(0,E.Wv)((0,S.R1)(Va),{id:e.id,class:"reference-endpoint",transparent:""},(0,E.eX)({title:(0,E.k6)((()=>[(0,E.Lk)("h3",Ki,[(0,E.Lk)("div",Yi,[(0,E.bF)((0,S.R1)(qt),{class:"endpoint-type",method:e.operation.method,short:""},null,8,["method"]),(0,E.bF)((0,S.R1)(dr),{id:e.id??"",class:"endpoint-anchor"},{default:(0,E.k6)((()=>[(0,E.Lk)("div",Ji,[(0,E.Lk)("div",es,[(0,E.bF)($r,{deprecated:(0,S.R1)(En)(e.transformedOperation),path:e.operation.path},null,8,["deprecated","path"])]),(0,E.Lk)("div",ts,(0,_.v_)(e.transformedOperation.name),1),(0,S.R1)(Sn)(e.transformedOperation)?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Kr),{key:0,class:(0,_.C4)((0,S.R1)(_n)(e.transformedOperation))},{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)((0,S.R1)(Sn)(e.transformedOperation)),1)])),_:1},8,["class"])):(0,E.Q3)("",!0)])])),_:1},8,["id"])])])])),actions:(0,E.k6)((({active:a})=>{var o;return[a?((0,E.uX)(),(0,E.Wv)((0,S.R1)(ni),{key:0,operation:e.operation},null,8,["operation"])):null!=(o=(0,S.R1)(n))&&o.hideTestRequestButton?(0,E.Q3)("",!0):((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.kv),{key:1,class:"endpoint-try-hint",icon:"Play",thickness:"1.75px"})),(0,E.bF)((0,S.R1)(Ct.EX),{class:"endpoint-copy",icon:"Clipboard",label:"Copy endpoint URL",size:"xs",variant:"ghost",onClick:r[0]||(r[0]=(0,rt.D$)((n=>(0,S.R1)(t)(e.operation.path)),["stop"]))})]})),default:(0,E.k6)((()=>[(0,E.Lk)("div",ns,[(0,E.Lk)("div",rs,[(0,E.Lk)("div",as,[(0,E.bF)(Ua,{operation:e.operation},null,8,["operation"])]),(0,E.Lk)("div",os,[(0,E.bF)(za,{collapsableItems:!1,operation:e.transformedOperation},null,8,["operation"])])]),(0,E.bF)((0,S.R1)(Qi),{responses:e.operation.responses},null,8,["responses"]),(0,E.bF)((0,S.R1)(Gi),{collection:e.collection,operation:e.operation,server:e.server,transformedOperation:e.transformedOperation},null,8,["collection","operation","server","transformedOperation"])])])),_:2},[null!=(a=e.operation)&&a.description?{name:"description",fn:(0,E.k6)((()=>{var t;return[(0,E.bF)((0,S.R1)(Ct.mi),{value:null==(t=e.operation)?void 0:t.description,withImages:""},null,8,["value"])]})),key:"0"}:void 0]),1032,["id"])}}}),[["__scopeId","data-v-e104650f"]]),ss={class:"operation-details"},ls={class:"examples"},cs=It((0,E.pM)({__name:"ModernLayout",props:{id:{},collection:{},server:{},operation:{},transformedOperation:{}},setup:e=>(e,t)=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(ar),{id:e.id,label:e.transformedOperation.name},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(hr),null,{default:(0,E.k6)((()=>[(0,S.R1)(Sn)(e.transformedOperation)?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Kr),{key:0,class:(0,_.C4)((0,S.R1)(_n)(e.transformedOperation))},{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)((0,S.R1)(Sn)(e.transformedOperation)),1)])),_:1},8,["class"])):(0,E.Q3)("",!0),(0,E.Lk)("div",{class:(0,_.C4)((0,S.R1)(En)(e.transformedOperation)?"deprecated":"")},[(0,E.bF)((0,S.R1)(sr),{level:3},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(dr),{id:e.id??""},{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(e.transformedOperation.name),1)])),_:1},8,["id"])])),_:1})],2),(0,E.bF)((0,S.R1)(mr),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(vr),null,{default:(0,E.k6)((()=>[(0,E.Lk)("div",ss,[(0,E.bF)((0,S.R1)(Ct.mi),{value:e.operation.description,withImages:""},null,8,["value"]),(0,E.bF)(Ua,{operation:e.operation},null,8,["operation"]),(0,E.bF)(za,{operation:e.transformedOperation},null,8,["operation"])])])),_:1}),(0,E.bF)((0,S.R1)(vr),null,{default:(0,E.k6)((()=>[(0,E.Lk)("div",ls,[(0,E.bF)((0,S.R1)(Ct.XO),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Gi),{collection:e.collection,fallback:"",operation:e.operation,server:e.server,transformedOperation:e.transformedOperation},{header:(0,E.k6)((()=>{var t;return[(0,E.bF)($r,{class:"example-path",deprecated:null==(t=e.transformedOperation.information)?void 0:t.deprecated,path:e.transformedOperation.path},null,8,["deprecated","path"])]})),footer:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(ni),{operation:e.operation},null,8,["operation"])])),_:1},8,["collection","operation","server","transformedOperation"])])),_:1}),(0,E.bF)((0,S.R1)(Ct.XO),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Qi),{responses:e.operation.responses,style:{"margin-top":"12px"}},null,8,["responses"])])),_:1})])])),_:1})])),_:1})])),_:1})])),_:1},8,["id","label"]))}),[["__scopeId","data-v-0916dd9f"]]),us=(0,E.pM)({__name:"Operation",props:{id:{},layout:{default:"modern"},transformedOperation:{},collection:{},server:{}},setup(e){const t=(0,St.hA)(),{operation:n}=function({store:e,location:t}){const n=(0,E.EW)((()=>Object.values((null==e?void 0:e.collections)??{})[0]));return{operation:(0,E.EW)((()=>{if(null==e||!e.collections||!e.requests)return;const r=Object.values(e.requests).filter((e=>{var t;return null==(t=n.value)?void 0:t.requests.includes(e.uid)}));if(!t.startsWith("#/paths/"))throw new Error("Invalid location, try using #/paths/$YOUR_ENDPOINT/$HTTP_METHOD. You can use the getPointer helper to generate a valid location: getPointer(['paths', '/planets/{planetId}', 'get'])");return r.find((e=>{const n=(0,Dr.K)(t.split("/")[2]),r=t.split("/")[3].toLocaleLowerCase();return e.method===r&&e.path===n}))}))}}({store:t,location:Pr(["paths",e.transformedOperation.path,e.transformedOperation.httpVerb.toLowerCase()])});return(e,t)=>e.collection&&(0,S.R1)(n)?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},["classic"===e.layout?((0,E.uX)(),(0,E.Wv)(is,{key:0,id:e.id,collection:e.collection,operation:(0,S.R1)(n),server:e.server,transformedOperation:e.transformedOperation},null,8,["id","collection","operation","server","transformedOperation"])):((0,E.uX)(),(0,E.Wv)(cs,{key:1,id:e.id,collection:e.collection,operation:(0,S.R1)(n),server:e.server,transformedOperation:e.transformedOperation},null,8,["id","collection","operation","server","transformedOperation"]))],64)):(0,E.Q3)("",!0)}}),ds=(0,E.pM)({__name:"Loading",props:{layout:{default:"modern"},parsedSpec:{}},setup(e){const t=e,n=(0,S.KR)(!1),r=(0,S.KR)([]),a=(0,S.KR)([]),{activeCollection:o,activeServer:i}=(0,_t.JC)(),{getModelId:s,getSectionId:l,getTagId:c,hash:u,isIntersectionEnabled:d}=dn(),p=(0,S.KR)("classic"!==t.layout&&u.value);(0,E.wB)((()=>{var e;return null==(e=t.parsedSpec.tags)?void 0:e.length}),(e=>{var o,i;if(!u.value||"number"!=typeof e||!t.parsedSpec.tags)return;const s=l();if(s.startsWith("tag")){let e=0;const a=(null==(o=t.parsedSpec.tags)?void 0:o.findIndex((e=>c(e)===s)))??0,l=u.value.match(/tag\/([^/]+)\/([^/]+)\/(.+)/);if(4===(null==l?void 0:l.length)){const n=l[2],r="/"+l[3];e=null==(i=t.parsedSpec.tags[a])?void 0:i.operations.findIndex((({httpVerb:e,path:t})=>n===e&&r===t))}const d=t.parsedSpec.tags[a];if(!d)return;"default"!==d.name&&(n.value=s!==u.value&&s.startsWith("tag")),r.value.push({...d,lazyOperations:d.operations.slice(e,e+2)})}else if(u.value.startsWith("model")){const e=Object.keys(Rn(t.parsedSpec)??{}),[,n]=u.value.toLowerCase().split("/"),r="models"===u.value?0:e.findIndex((e=>e.toLowerCase()===n));if(-1===r)return;a.value=e.slice(r,r+3)}else typeof window<"u"&&pn(u.value),setTimeout((()=>d.value=!0),1e3)}),{immediate:!0});const h=Tn.on((({id:e})=>{const t=u.value;!t||e!==t||(h(),setTimeout((()=>{typeof window<"u"&&pn(t),p.value=!1,setTimeout((()=>d.value=!0),1e3)}),300))}));return(0,E.sV)((()=>{u.value||setTimeout((()=>d.value=!0),1e3)})),(e,t)=>(0,E.bo)(((0,E.uX)(),(0,E.CE)("div",{class:(0,_.C4)(["references-loading",{"references-loading-hidden-tag":n.value,"references-loading-top-spacer":r.value.length}])},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(r.value,((t,n)=>((0,E.uX)(),(0,E.CE)(E.FK,{key:t.name+n},[t.operations&&t.operations.length>0?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Rr),{key:0,spec:e.parsedSpec,tag:t},{default:(0,E.k6)((()=>[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(t.lazyOperations,(t=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(us),{key:`${t.httpVerb}-${t.operationId}`,collection:(0,S.R1)(o),layout:e.layout,server:(0,S.R1)(i),transformedOperation:t},null,8,["collection","layout","server","transformedOperation"])))),128))])),_:2},1032,["spec","tag"])):(0,E.Q3)("",!0)],64)))),128)),a.value.length?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ar),{key:0},{default:(0,E.k6)((()=>[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(a.value,(t=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(ar),{key:t,label:t},{default:(0,E.k6)((()=>{var n;return[null!=(n=(0,S.R1)(Rn)(e.parsedSpec))&&n[t]?((0,E.uX)(),(0,E.Wv)((0,S.R1)(hr),{key:0},{default:(0,E.k6)((()=>{var n;return[(0,E.bF)((0,S.R1)(sr),{level:2},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(dr),{id:(0,S.R1)(s)({name:t})},{default:(0,E.k6)((()=>{var n;return[(0,E.eW)((0,_.v_)((null==(n=(0,S.R1)(Rn)(e.parsedSpec))?void 0:n[t]).title??t),1)]})),_:2},1032,["id"])])),_:2},1024),(0,E.bF)((0,S.R1)(Xr),{name:t,noncollapsible:"",value:null==(n=(0,S.R1)(Rn)(e.parsedSpec))?void 0:n[t]},null,8,["name","value"])]})),_:2},1024)):(0,E.Q3)("",!0)]})),_:2},1032,["label"])))),128))])),_:1})):(0,E.Q3)("",!0)],2)),[[rt.aG,p.value]])}}),ps={class:"download"},hs=["href"],fs=It((0,E.pM)({__name:"DownloadLink",props:{specTitle:{}},setup(e){const t=e,n=ot(),r=()=>{Ft.emit({id:"",specTitle:t.specTitle})};return(e,t)=>{var a,o,i,s,l;return(0,E.uX)(),(0,E.CE)("div",ps,[null!=(a=(0,S.R1)(n))&&a.hideDownloadButton?(0,E.Q3)("",!0):((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[null!=(i=null==(o=(0,S.R1)(n))?void 0:o.spec)&&i.url?((0,E.uX)(),(0,E.CE)("a",{key:0,class:"download-button",download:"",href:null==(l=null==(s=(0,S.R1)(n))?void 0:s.spec)?void 0:l.url,target:"_blank"}," Download OpenAPI Document ",8,hs)):((0,E.uX)(),(0,E.CE)("button",{key:1,class:"download-button",role:"link",type:"button",onClick:r}," Download OpenAPI Document "))],64))])}}}),[["__scopeId","data-v-1fc94525"]]),ms={key:0,class:"introduction-description"},gs=It((0,E.pM)({__name:"Description",props:{value:{}},setup(e){const t=e,n=(0,E.EW)((()=>{if(!t.value)return[];const e=new Kt;return(0,Dn.sV)(t.value).map((t=>{const[n]=(0,Dn.cf)(t);return{id:n?r({...n,slug:e.slug(n.value)}):void 0,content:t}}))})),{getHeadingId:r,getFullHash:a,isIntersectionEnabled:o,replaceUrlState:i}=dn(),s=new Kt,l=e=>(e.data={hProperties:{id:r({depth:e.depth,value:e.children[0].value,slug:s.slug(e.children[0].value)})}},e);return(e,t)=>e.value?((0,E.uX)(),(0,E.CE)("div",ms,[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(n.value,(e=>((0,E.uX)(),(0,E.CE)(E.FK,{key:e.id},[e.id?((0,E.uX)(),(0,E.Wv)(rr,{key:0,id:(0,S.R1)(a)(e.id),class:"introduction-description-heading",onIntersecting:()=>function(e=""){o.value&&i(e)}(e.id)},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Ct.mi),{transform:l,transformType:"heading",value:e.content},null,8,["value"])])),_:2},1032,["id","onIntersecting"])):((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.mi),{key:1,value:e.content,withImages:""},null,8,["value"]))],64)))),128))])):(0,E.Q3)("",!0)}}),[["__scopeId","data-v-fa3d93c5"]]),vs={class:"badges"},bs={class:"sticky-cards"},ys=It((0,E.pM)({__name:"Introduction",props:{info:{},parsedSpec:{}},setup(e){const t=e,n=(0,E.EW)((()=>{var e,n;return(null==(e=t.parsedSpec)?void 0:e.openapi)??(null==(n=t.parsedSpec)?void 0:n.swagger)??""})),r=new Kt,a=(0,E.EW)((()=>{var e;return r.slug((null==(e=t.info)?void 0:e.title)??"")})),o=(0,E.EW)((()=>{var e,n;return"string"==typeof(null==(e=t.info)?void 0:e.version)?t.info.version.toString().match(/^\d/)?`v${t.info.version}`:t.info.version:"number"==typeof(null==(n=t.info)?void 0:n.version)?`v${t.info.version}`:void 0})),{onLoaded:i}=ot();return(0,E.sV)((()=>null==i?void 0:i())),(e,t)=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ar),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(ar),{class:"introduction-section gap-12"},{default:(0,E.k6)((()=>{var t,r;return[(0,E.bF)((0,S.R1)(hr),{loading:!(null!=(t=e.info)&&t.description||null!=(r=e.info)&&r.title)},{default:(0,E.k6)((()=>[(0,E.Lk)("div",vs,[o.value?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Kr),{key:0},{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(o.value),1)])),_:1})):(0,E.Q3)("",!0),n.value?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Kr),{key:1},{default:(0,E.k6)((()=>[(0,E.eW)("OAS "+(0,_.v_)(n.value),1)])),_:1})):(0,E.Q3)("",!0)]),(0,E.bF)((0,S.R1)(sr),{level:1,loading:!e.info.title,tight:""},{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(e.info.title),1)])),_:1},8,["loading"]),(0,E.bF)(fs,{specTitle:a.value},null,8,["specTitle"]),(0,E.bF)((0,S.R1)(mr),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(vr),null,{default:(0,E.k6)((()=>[(0,E.bF)(gs,{value:e.info.description},null,8,["value"])])),_:1}),e.$slots.aside?((0,E.uX)(),(0,E.Wv)((0,S.R1)(vr),{key:0},{default:(0,E.k6)((()=>[(0,E.Lk)("div",bs,[(0,E.RG)(e.$slots,"aside",{},void 0,!0)])])),_:3})):(0,E.Q3)("",!0)])),_:3})])),_:3},8,["loading"]),(0,E.RG)(e.$slots,"after",{},void 0,!0)]})),_:3})])),_:3}))}}),[["__scopeId","data-v-0336cd4f"]]);var ws=n(9347),Os=n(2549);const ks=["id"],xs=(0,E.pM)({__name:"BaseUrl",setup(e){const{activeCollection:t,activeServer:n}=(0,_t.JC)(),{serverMutators:r}=(0,St.hA)(),a=(0,E.Bi)(),o=(e,t)=>{if(!n.value)return;const a=n.value.variables||{};a[e]={...a[e],default:t},r.edit(n.value.uid,"variables",a)};return(e,r)=>{var i,s,l,c;return(0,E.uX)(),(0,E.CE)(E.FK,null,[r[0]||(r[0]=(0,E.Lk)("label",{class:"bg-b-2 flex font-medium items-center h-8 px-3 py-2.5 text-sm"}," Server ",-1)),(0,E.Lk)("div",{id:(0,S.R1)(a)},[null!=(s=null==(i=(0,S.R1)(t))?void 0:i.servers)&&s.length?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Os.A),{key:0,collection:(0,S.R1)(t),layout:"reference",server:(0,S.R1)(n),target:(0,S.R1)(a)},null,8,["collection","server","target"])):(0,E.Q3)("",!0)],8,ks),(0,E.bF)((0,S.R1)(ws.A),{variables:null==(l=(0,S.R1)(n))?void 0:l.variables,"onUpdate:variable":o},null,8,["variables"]),null!=(c=(0,S.R1)(n))&&c.description?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.mi),{key:0,class:"px-3 py-1.5 text-c-3",value:(0,S.R1)(n).description},null,8,["value"])):(0,E.Q3)("",!0)],64)}}}),Es=["onClick"],Ss={class:"client-libraries-text"},_s=["value"],Cs=["label"],Ts=["aria-label","value"],As={"aria-hidden":"true",class:"client-libraries-icon__more"},Rs={key:1,class:"client-libraries-icon",height:"50",role:"presentation",viewBox:"0 0 50 50",width:"50",xmlns:"http://www.w3.org/2000/svg"},Ps={key:0,"aria-hidden":"true",class:"client-libraries-text"},Ds=It((0,E.pM)({__name:"ClientSelector",setup(e){const{httpClient:t,setHttpClient:n,availableTargets:r,getClientTitle:a,getTargetTitle:o}=kt(),i=(0,S.KR)(),s=[{targetKey:"shell",clientKey:"curl"},{targetKey:"ruby",clientKey:"native"},{targetKey:"node",clientKey:"undici"},{targetKey:"php",clientKey:"guzzle"},{targetKey:"python",clientKey:"python3"}].filter((e=>r.value.find((t=>t.key===e.targetKey&&t.clients.find((t=>t.client===e.clientKey)))))),l=e=>`programming-language-${"js"===e?"javascript":e}`,c=e=>e.targetKey===t.targetKey&&e.clientKey===t.clientKey,u=e=>s.some((t=>t.targetKey===e.targetKey&&t.clientKey===e.clientKey));return(e,d)=>((0,E.uX)(),(0,E.CE)("div",{ref_key:"containerRef",ref:i,class:"client-libraries-content"},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)((0,S.R1)(s),(e=>((0,E.uX)(),(0,E.CE)("button",{key:e.clientKey,"aria-hidden":"true",class:(0,_.C4)(["client-libraries rendered-code-sdks",{"client-libraries__active":c(e)}]),tabindex:"-1",type:"button",onClick:()=>(0,S.R1)(n)(e)},[(0,E.Lk)("div",{class:(0,_.C4)(`client-libraries-icon__${e.targetKey}`)},[(0,E.bF)((0,S.R1)(Ct.kv),{class:"client-libraries-icon",icon:l(e.targetKey)},null,8,["icon"])],2),(0,E.Lk)("span",Ss,(0,_.v_)((0,S.R1)(o)(e)),1)],10,Es)))),128)),(0,E.Lk)("label",{class:(0,_.C4)(["client-libraries client-libraries__select",{"client-libraries__active":(0,S.R1)(t)&&!u((0,S.R1)(t))}])},[d[2]||(d[2]=(0,E.Lk)("span",{class:"sr-only"},"Select Client Library",-1)),(0,E.Lk)("select",{class:"language-select",value:JSON.stringify((0,S.R1)(t)),onInput:d[0]||(d[0]=e=>(0,S.R1)(n)(JSON.parse(e.target.value)))},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)((0,S.R1)(r),(e=>((0,E.uX)(),(0,E.CE)("optgroup",{key:e.key,label:e.title},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(e.clients,(t=>((0,E.uX)(),(0,E.CE)("option",{key:t.client,"aria-label":`${e.title} ${(0,S.R1)(a)({targetKey:e.key,clientKey:t.client})}`,value:JSON.stringify({targetKey:e.key,clientKey:t.client})},(0,_.v_)((0,S.R1)(a)({targetKey:e.key,clientKey:t.client})),9,Ts)))),128))],8,Cs)))),128))],40,_s),(0,E.Lk)("div",As,[(0,S.R1)(t)&&!u((0,S.R1)(t))?((0,E.uX)(),(0,E.CE)("div",{key:0,class:(0,_.C4)(`client-libraries-icon__${(0,S.R1)(t).targetKey}`)},[(0,E.bF)((0,S.R1)(Ct.kv),{class:"client-libraries-icon",icon:l((0,S.R1)(t).targetKey)},null,8,["icon"])],2)):((0,E.uX)(),(0,E.CE)("svg",Rs,d[1]||(d[1]=[(0,E.Lk)("g",{fill:"currentColor","fill-rule":"nonzero"},[(0,E.Lk)("path",{d:"M10.71 25.3a3.87 3.87 0 1 0 7.74 0 3.87 3.87 0 0 0-7.74 0M21.13 25.3a3.87 3.87 0 1 0 7.74 0 3.87 3.87 0 0 0-7.74 0M31.55 25.3a3.87 3.87 0 1 0 7.74 0 3.87 3.87 0 0 0-7.74 0"})],-1)])))]),(0,S.R1)(r).length?((0,E.uX)(),(0,E.CE)("span",Ps," More ")):(0,E.Q3)("",!0)],2)],512))}}),[["__scopeId","data-v-b29373e8"]]),Ms={key:0},$s={class:"selected-client card-footer",muted:""},Ns=It((0,E.pM)({__name:"ClientLibraries",setup(e){const{availableTargets:t,httpTargetTitle:n,httpClientTitle:r}=kt();return(e,a)=>(0,S.R1)(t).length?((0,E.uX)(),(0,E.CE)("div",Ms,[a[0]||(a[0]=(0,E.Lk)("div",{class:"client-libraries-heading"},"Client Libraries",-1)),(0,E.Lk)("div",null,[(0,E.bF)(Ds)]),(0,E.Lk)("div",$s,(0,_.v_)((0,S.R1)(r))+" "+(0,_.v_)((0,S.R1)(n)),1)])):(0,E.Q3)("",!0)}}),[["__scopeId","data-v-35fd541a"]]),Is={class:"section-accordion-wrapper"},Ls={class:"section-accordion-title"},Bs=It((0,E.pM)({__name:"SectionContainerAccordion",setup:e=>(e,t)=>((0,E.uX)(),(0,E.CE)("div",Is,[(0,E.bF)((0,S.R1)(Nr.EN),{as:"div",class:"section-accordion",defaultOpen:""},{default:(0,E.k6)((({open:t})=>[(0,E.bF)((0,S.R1)(Nr._q),{class:"section-accordion-button"},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Ct.kv),{class:"section-accordion-chevron",icon:t?"ChevronDown":"ChevronRight"},null,8,["icon"]),(0,E.Lk)("div",Ls,[(0,E.RG)(e.$slots,"title",{},void 0,!0)])])),_:2},1024),(0,E.bF)((0,S.R1)(Nr.kS),{class:"section-accordion-content"},{default:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"default",{},void 0,!0)])),_:3})])),_:3})]))}),[["__scopeId","data-v-1a30b4ff"]]),Fs=It((0,E.pM)({__name:"TagAccordion",props:{tag:{}},setup(e){const{getTagId:t}=dn();return(e,n)=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Bs),{class:"tag-section"},{title:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(sr),{class:"tag-name",level:2},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(dr),{id:(0,S.R1)(t)(e.tag)},{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(e.tag.name),1)])),_:1},8,["id"])])),_:1}),(0,E.bF)((0,S.R1)(Ct.mi),{class:"tag-description",value:e.tag.description,withImages:""},null,8,["value"])])),default:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"default",{},void 0,!0)])),_:3}))}}),[["__scopeId","data-v-43e1c813"]]),Qs=(0,E.pM)({__name:"Lazy",props:{id:{},isLazy:{type:Boolean,default:!0},lazyTimeout:{default:0}},setup(e){const t=e,n=(0,S.KR)(!t.isLazy);return t.isLazy?((e=()=>{})=>{typeof window>"u"||("requestIdleCallback"in window?setTimeout((()=>window.requestIdleCallback(e)),t.lazyTimeout):setTimeout((()=>(0,E.dY)(e)),t.lazyTimeout??300))})((()=>{n.value=!0,t.id&&(0,E.dY)((()=>Tn.emit({id:t.id})))})):t.id&&(0,E.dY)((()=>Tn.emit({id:t.id}))),(e,t)=>n.value?(0,E.RG)(e.$slots,"default",{key:0}):(0,E.Q3)("",!0)}}),Us=(0,E.pM)({__name:"TagList",props:{tags:{},spec:{},layout:{}},setup(e){const t=e,{getOperationId:n,getTagId:r,hash:a}=dn(),{collapsedSidebarItems:o}=Hn(),{activeCollection:i,activeServer:s}=(0,_t.JC)(),l=(0,E.EW)((()=>"classic"===t.layout?Fs:Rr)),c=(0,E.EW)((()=>t.tags.findIndex((e=>!o[r(e)]))+1)),u=e=>"classic"!==t.layout&&!a.value.startsWith("model")&&e>c.value;return(e,t)=>((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(e.tags,((t,a)=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Qs),{id:(0,S.R1)(r)(t),key:(0,S.R1)(r)(t),isLazy:u(a)},{default:(0,E.k6)((()=>[((0,E.uX)(),(0,E.Wv)((0,E.$y)(l.value),{id:(0,S.R1)(r)(t),spec:e.spec,tag:t},{default:(0,E.k6)((()=>[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(t.operations,((l,c)=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Qs),{id:(0,S.R1)(n)(l,t),key:`${l.httpVerb}-${l.operationId}`,isLazy:u(a)||(0,S.R1)(o)[(0,S.R1)(r)(t)]&&c>0},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Ct.XO),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(us),{id:(0,S.R1)(n)(l,t),collection:(0,S.R1)(i),layout:e.layout,server:(0,S.R1)(s),transformedOperation:l},null,8,["id","collection","layout","server","transformedOperation"])])),_:2},1024)])),_:2},1032,["id","isLazy"])))),128))])),_:2},1032,["id","spec","tag"]))])),_:2},1032,["id","isLazy"])))),128))}}),zs=It((0,E.pM)({__name:"Webhook",props:{webhook:{}},setup:e=>(e,t)=>{var n;return e.webhook?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[(0,E.bF)(Qa,{class:"webhook-request-body",requestBody:null==(n=e.webhook.information)?void 0:n.requestBody},{title:(0,E.k6)((()=>t[0]||(t[0]=[(0,E.eW)("Payload")]))),_:1},8,["requestBody"]),(0,E.bF)(za,{operation:e.webhook},null,8,["operation"])],64)):(0,E.Q3)("",!0)}}),[["__scopeId","data-v-cbaa1001"]]),js=["aria-expanded"],Ws=["id"],Xs=It((0,E.pM)({__name:"CompactSection",props:{id:{},label:{}},setup(e){const t=e,{hash:n}=dn(),r=(0,S.KR)(!1);return(0,E.wB)(n,(async e=>{e===t.id&&!r.value&&(r.value=!0,await(0,E.dY)(),pn(t.id))}),{immediate:!0}),(e,t)=>((0,E.uX)(),(0,E.CE)("div",{"aria-expanded":r.value,class:"collapsible-section"},[(0,E.Lk)("div",{id:e.id,class:(0,_.C4)(["collapsible-section-trigger",{"collapsible-section-trigger-open":r.value}]),onClick:t[0]||(t[0]=e=>r.value=!r.value)},[(0,E.bF)((0,S.R1)(Ct.kv),{icon:r.value?"ChevronDown":"ChevronRight",size:"md",thickness:"1.5"},null,8,["icon"]),(0,E.bF)(dr,{id:e.id,class:"collapsible-section-header"},{default:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"heading",{},void 0,!0)])),_:3},8,["id"])],10,Ws),r.value?((0,E.uX)(),(0,E.Wv)(ar,{key:0,class:"collapsible-section-content",label:e.label},{default:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"default",{},void 0,!0)])),_:3},8,["label"])):(0,E.Q3)("",!0)],8,js))}}),[["__scopeId","data-v-b32ceff4"]]),qs=It((0,E.pM)({__name:"Webhooks",props:{webhooks:{}},setup(e){const t=e,n=(0,E.EW)((()=>Object.keys(t.webhooks??{}))),{getWebhookId:r}=dn(),{collapsedSidebarItems:a}=Hn(),o=(0,E.EW)((()=>n.value.length<=10||a[r()])),i=(0,E.EW)((()=>{const e=n.value;return o.value?e:e.slice(0,10)}));return(e,t)=>n.value.length?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ar),{key:0,id:"webhooks"},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(ar),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(sr),{level:2},{default:(0,E.k6)((()=>t[0]||(t[0]=[(0,E.eW)("Webhooks")]))),_:1}),(0,E.bF)((0,S.R1)(Qs),{id:"webhooks",isLazy:!1}),(0,E.Lk)("div",{class:(0,_.C4)(["webhooks-list",{"webhooks-list-truncated":!o.value}])},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(i.value,(t=>{var n;return(0,E.uX)(),(0,E.CE)(E.FK,{key:t},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(Object.keys(null==(n=e.webhooks)?void 0:n[t]),(n=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Qs),{id:(0,S.R1)(r)({name:t,method:n}),key:n,isLazy:""},{default:(0,E.k6)((()=>{var a;return[null!=(a=e.webhooks)&&a[t]?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Xs),{key:0,id:(0,S.R1)(r)({name:t,method:n}),class:"webhooks-list-item",label:t},{heading:(0,E.k6)((()=>{var r;return[(0,E.eW)((0,_.v_)(null==(r=e.webhooks[t][n])?void 0:r.name),1)]})),default:(0,E.k6)((()=>{var r,a;return[null!=(r=e.webhooks[t][n])&&r.description?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.mi),{key:0,class:"webhook-section-description",value:null==(a=e.webhooks[t][n])?void 0:a.description,withImages:""},null,8,["value"])):(0,E.Q3)("",!0),(0,E.bF)(zs,{webhook:e.webhooks[t][n]},null,8,["webhook"])]})),_:2},1032,["id","label"])):(0,E.Q3)("",!0)]})),_:2},1032,["id"])))),128))],64)})),128))],2),o.value?(0,E.Q3)("",!0):((0,E.uX)(),(0,E.Wv)(er,{key:0,id:"webhooks",class:"show-more-webhooks"}))])),_:1})])),_:1})):(0,E.Q3)("",!0)}}),[["__scopeId","data-v-8cee6fce"]]),Vs={key:0,class:"properties"},Hs={key:1},Zs=It((0,E.pM)({__name:"ModelsAccordion",props:{schemas:{}},setup(e){const t=e,n=(0,E.EW)((()=>t.schemas?Object.entries(t.schemas).map((([e,t])=>({name:e,schema:t}))):[])),{getModelId:r}=dn();return(e,a)=>t.schemas?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Bs),{key:0,class:"reference-models"},{title:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(sr),{level:2},{default:(0,E.k6)((()=>a[0]||(a[0]=[(0,E.eW)("Models")]))),_:1})])),default:(0,E.k6)((()=>[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(n.value,(({name:e,schema:t})=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Va),{id:(0,S.R1)(r)({name:e}),key:e,label:e},{title:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(dr),{id:(0,S.R1)(r)({name:e}),class:"reference-models-anchor"},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(jr),{class:"reference-models-label",name:e,value:t},null,8,["name","value"])])),_:2},1032,["id"])])),default:(0,E.k6)((()=>[null!=t&&t.properties?((0,E.uX)(),(0,E.CE)("div",Vs,[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(Object.entries(t.properties),(([e,n])=>{var r,a,o;return(0,E.uX)(),(0,E.Wv)((0,S.R1)(Ea),{key:e,name:e,required:(null==(r=t.required)?void 0:r.includes(e))||!0===(null==(o=null==(a=t.properties)?void 0:a[e])?void 0:o.required),value:n},null,8,["name","required","value"])})),128))])):((0,E.uX)(),(0,E.CE)("div",Hs,[(0,E.bF)((0,S.R1)(Ea),{value:t},null,8,["value"])]))])),_:2},1032,["id","label"])))),128))])),_:1})):(0,E.Q3)("",!0)}}),[["__scopeId","data-v-4396a256"]]),Gs=It((0,E.pM)({__name:"Models",props:{schemas:{}},setup(e){const t=e,{collapsedSidebarItems:n}=Hn(),{getModelId:r}=dn(),a=(0,E.EW)((()=>Object.keys(t.schemas??{}).length<=10||n[r()])),o=(0,E.EW)((()=>{const e=Object.keys(t.schemas??{});return a.value?e:e.slice(0,10)}));return(e,t)=>e.schemas?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ar),{key:0,id:"models"},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(ar),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(sr),{level:2},{default:(0,E.k6)((()=>t[0]||(t[0]=[(0,E.eW)("Models")]))),_:1}),(0,E.bF)((0,S.R1)(Qs),{id:"models",isLazy:!1}),(0,E.Lk)("div",{class:(0,_.C4)(["models-list",{"models-list-truncated":!a.value}])},[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)(o.value,(t=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Qs),{id:(0,S.R1)(r)({name:t}),key:t,isLazy:""},{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Xs),{id:(0,S.R1)(r)({name:t}),class:"models-list-item",label:t},{heading:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(jr),{name:t,value:e.schemas[t]},null,8,["name","value"])])),default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Ct.XO),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Xr),{hideHeading:!0,noncollapsible:"",value:e.schemas[t]},null,8,["value"])])),_:2},1024)])),_:2},1032,["id","label"])])),_:2},1032,["id"])))),128))],2),a.value?(0,E.Q3)("",!0):((0,E.uX)(),(0,E.Wv)(er,{key:0,id:(0,S.R1)(r)(),class:"show-more-models"},null,8,["id"]))])),_:1})])),_:1})):(0,E.Q3)("",!0)}}),[["__scopeId","data-v-7c1cd4b4"]]),Ks={class:"narrow-references-container"},Ys={key:0,class:"scalar-client introduction-card-item [--scalar-address-bar-height:0px] divide-y text-sm"},Js={key:1,class:"scalar-client introduction-card-item"},el=It((0,E.pM)({__name:"Content",props:{parsedSpec:{},layout:{default:"modern"}},setup(e){const t=e,{hideModels:n}=Hn(),{securitySchemes:r}=(0,St.hA)(),{activeCollection:a,activeEnvVariables:o,activeEnvironment:i,activeServer:s,activeWorkspace:l}=(0,_t.JC)(),c=(0,E.EW)((()=>"classic"===t.layout?"after":"aside"));return(e,t)=>{var u,d,p,h;return(0,E.uX)(),(0,E.CE)(E.FK,null,[t[0]||(t[0]=(0,E.Fv)('<div class="section-flare" data-v-42cfaf8d><div class="section-flare-item" data-v-42cfaf8d></div><div class="section-flare-item" data-v-42cfaf8d></div><div class="section-flare-item" data-v-42cfaf8d></div><div class="section-flare-item" data-v-42cfaf8d></div><div class="section-flare-item" data-v-42cfaf8d></div><div class="section-flare-item" data-v-42cfaf8d></div><div class="section-flare-item" data-v-42cfaf8d></div><div class="section-flare-item" data-v-42cfaf8d></div></div>',1)),(0,E.Lk)("div",Ks,[(0,E.RG)(e.$slots,"start",{},void 0,!0),(0,E.bF)((0,S.R1)(ds),{layout:e.layout,parsedSpec:e.parsedSpec},null,8,["layout","parsedSpec"]),null!=(d=null==(u=e.parsedSpec)?void 0:u.info)&&d.title||null!=(h=null==(p=e.parsedSpec)?void 0:p.info)&&h.description?((0,E.uX)(),(0,E.Wv)((0,S.R1)(ys),{key:0,info:e.parsedSpec.info,parsedSpec:e.parsedSpec},{[c.value]:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Ct.XO),null,{default:(0,E.k6)((()=>{var t,n,c;return[(0,E.Lk)("div",{class:(0,_.C4)(["introduction-card",{"introduction-card-row":"classic"===e.layout}])},[null!=(n=null==(t=(0,S.R1)(a))?void 0:t.servers)&&n.length?((0,E.uX)(),(0,E.CE)("div",Ys,[(0,E.bF)((0,S.R1)(xs))])):(0,E.Q3)("",!0),(0,S.R1)(a)&&(0,S.R1)(l)&&Object.keys((0,S.R1)(r)??{}).length?((0,E.uX)(),(0,E.CE)("div",Js,[(0,E.bF)((0,S.R1)(Jn.A),{collection:(0,S.R1)(a),envVariables:(0,S.R1)(o),environment:(0,S.R1)(i),layout:"reference",selectedSecuritySchemeUids:(null==(c=(0,S.R1)(a))?void 0:c.selectedSecuritySchemeUids)??[],server:(0,S.R1)(s),title:"Authentication",workspace:(0,S.R1)(l)},null,8,["collection","envVariables","environment","selectedSecuritySchemeUids","server","workspace"])])):(0,E.Q3)("",!0),(0,E.bF)((0,S.R1)(Ns),{class:"introduction-card-item"})],2)]})),_:1})])),_:2},1032,["info","parsedSpec"])):(0,E.RG)(e.$slots,"empty-state",{key:1},void 0,!0),e.parsedSpec.tags?((0,E.uX)(),(0,E.CE)(E.FK,{key:2},[e.parsedSpec["x-tagGroups"]?((0,E.uX)(!0),(0,E.CE)(E.FK,{key:0},(0,E.pI)(e.parsedSpec["x-tagGroups"],(t=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Us),{key:t.name,layout:e.layout,spec:e.parsedSpec,tags:t.tags.map((t=>{var n;return null==(n=e.parsedSpec.tags)?void 0:n.find((e=>e.name===t))})).filter((e=>!!e))},null,8,["layout","spec","tags"])))),128)):((0,E.uX)(),(0,E.Wv)((0,S.R1)(Us),{key:1,layout:e.layout,spec:e.parsedSpec,tags:e.parsedSpec.tags},null,8,["layout","spec","tags"]))],64)):(0,E.Q3)("",!0),e.parsedSpec.webhooks?((0,E.uX)(),(0,E.Wv)((0,S.R1)(qs),{key:3,webhooks:e.parsedSpec.webhooks},null,8,["webhooks"])):(0,E.Q3)("",!0),(0,S.R1)(Pn)(e.parsedSpec)&&!(0,S.R1)(n)?((0,E.uX)(),(0,E.CE)(E.FK,{key:4},["classic"===e.layout?((0,E.uX)(),(0,E.Wv)((0,S.R1)(Zs),{key:0,schemas:(0,S.R1)(Rn)(e.parsedSpec)},null,8,["schemas"])):((0,E.uX)(),(0,E.Wv)((0,S.R1)(Gs),{key:1,schemas:(0,S.R1)(Rn)(e.parsedSpec)},null,8,["schemas"]))],64)):(0,E.Q3)("",!0),(0,E.RG)(e.$slots,"end",{},void 0,!0)])],64)}}}),[["__scopeId","data-v-42cfaf8d"]]),tl=(0,E.pM)({__name:"ApiClientModal",props:{configuration:{}},setup(e){const t=(0,S.KR)(null),{client:n,init:r}=ei(),{selectedExampleKey:a,operationId:o}=Xi(),i=(0,St.hA)();return(0,E.sV)((()=>{t.value&&r({el:t.value,configuration:e.configuration,store:i})})),(0,At.Th)((()=>e.configuration),(e=>{var t;return e&&(null==(t=n.value)?void 0:t.updateConfig(e))}),{deep:!0,debounce:300}),(0,E.wB)(a,(e=>{n.value&&e&&o.value&&n.value.updateExample(e,o.value)})),(0,E.xo)((()=>{var e;return null==(e=n.value)?void 0:e.app.unmount()})),(e,n)=>((0,E.uX)(),(0,E.CE)("div",{ref_key:"el",ref:t},null,512))}}),nl=["innerHTML"],rl={class:"references-header"},al={key:0,class:"references-navigation t-doc__sidebar"},ol={class:"references-navigation-list"},il={class:"references-editor"},sl={class:"references-editor-textarea"},ll=["aria-label"],cl={key:0,class:"references-footer"},ul=It((0,E.pM)({inheritAttrs:!1,__name:"ApiReferenceLayout",props:{configuration:{},parsedSpec:{},rawSpec:{}},emits:["changeTheme","updateContent","loadSwaggerFile","linkSwaggerFile","toggleDarkMode"],setup(e){const t=e,{initializeToasts:n,toast:r}=(0,Tt.Yf)();n((e=>r(e)));const a=(0,We.Ubj)("(min-width: 1150px)"),o=(0,S.KR)("100dvh"),i=(0,S.KR)(null);(0,We.wYm)(i,(e=>{o.value=e[0].contentRect.height+"px"}));const s=(0,E.EW)(C.Ro),{breadcrumb:l,collapsedSidebarItems:c,isSidebarOpen:u,setCollapsedSidebarItem:d,hideModels:p,defaultOpenAllTags:h,setParsedSpec:f,scrollToOperation:m}=Hn(),{getReferenceHash:g,getPathRoutingId:v,getSectionId:b,getTagId:y,hash:w,isIntersectionEnabled:O,pathRouting:k,updateHash:x,replaceUrlState:T}=dn();k.value=t.configuration.pathRouting,(0,E.KC)((()=>x()));const A=async e=>{var t;O.value=!1,x(),e?m(e):null==(t=i.value)||t.scrollTo(0,0),await Ht(100),O.value=!0},R=(0,S.KR)(0);(0,E.KC)((()=>(0,Ct.AG)())),(0,E.sV)((()=>{var e,n,r;history.scrollRestoration="manual",Ft.on((({specTitle:e})=>{!function(e,t){const n=(0,Xe.Eo)(e),r=n?new Blob([e],{type:"application/json"}):new Blob([e],{type:"application/x-yaml"}),a=URL.createObjectURL(r),o=n?".json":".yaml",i=t?t+o:"spec"+o,s=document.createElement("a");s.href=a,s.download=i,s.dispatchEvent(new MouseEvent("click",{bubbles:!0,cancelable:!0,view:window})),setTimeout((()=>{window.URL.revokeObjectURL(a),s.remove()}),100)}(t.rawSpec,e)}));const a=null==(n=null==(e=i.value)?void 0:e.parentElement)?void 0:n.getBoundingClientRect(),o=null==(r=i.value)?void 0:r.getBoundingClientRect();if(a&&o){const e=o.top-a.top;R.value=e<2?0:e}window.onhashchange=()=>{A(g())},window.onpopstate=()=>k.value&&A(v(window.location.pathname))}));const P=(0,E.EW)((()=>a.value||!t.configuration.isEditable)),D=(0,At.Q0)((e=>{if((e.target.scrollTop??0)<50){const e=t.configuration.pathRouting?t.configuration.pathRouting.basePath:window.location.pathname;T("",e+window.location.search)}})),M=(0,E.EW)((()=>({breadcrumb:l.value,spec:t.parsedSpec})));(0,E.hi)((()=>{Ft.reset()})),(0,E.wB)((()=>t.parsedSpec),f,{deep:!0}),(0,E.SS)((()=>{var e,n;const r=(0,E.LM)();if(r&&(r.payload||(r.payload={data:(0,Xe.nG)()}),(e=r.payload).data||(e.data=(0,Xe.nG)()),t.configuration.pathRouting)){const e=v(r.url);w.value=e,r.payload.data.hash=e;for(const e in c)Object.hasOwn(c,e)&&delete c[e];if(e)d(b(e),!0);else{const e=null==(n=t.parsedSpec.tags)?void 0:n[0];e&&d(y(e),!0)}r.payload.data["useSidebarContent-collapsedSidebarItems"]=c}})),(0,xt.I)((()=>{var e;const t=(0,E.nI)(),n="scalar-instance-id";if(!t)return n;let r=t.uid;return typeof window>"u"?(t.attrs||(t.attrs={}),t.attrs[n]=r):null!=(e=t.vnode.el)&&e.getAttribute&&(r=t.vnode.el.getAttribute(n)),`${n}-${r}`}));const $=(0,St.Pg)({proxyUrl:t.configuration.proxyUrl||t.configuration.proxy,themeId:t.configuration.theme,useLocalStorage:!1,hideClientButton:t.configuration.hideClientButton,integration:t.configuration._integration});(0,E.wB)((()=>t.rawSpec),(e=>{var n;return e&&$.importSpecFile(e,"default",{shouldLoad:!1,documentUrl:null==(n=t.configuration.spec)?void 0:n.url,setCollectionSecurity:!0,...t.configuration})}),{immediate:!0}),(0,E.Gt)(St.pd,$);const N=(0,_t.T7)($);function I(e,n){(0,E.wB)((()=>t.configuration[e]),(e=>{typeof e<"u"&&n(e)}),{immediate:!0})}(0,E.Gt)(_t.UF,N),(0,E.Gt)(Et.K,"modal"),(0,E.Gt)(at,t.configuration??{});const{setExcludedClients:L,setDefaultHttpClient:B}=kt();var F;I("defaultHttpClient",B),I("hiddenClients",L),p.value=t.configuration.hideModels??!1,h.value=t.configuration.defaultOpenAllTags??!1,F=t.configuration,(0,E.wB)((()=>F),(()=>{F.proxy&&console.warn("[DEPRECATED] You’re using the deprecated 'proxy' attribute, rename it to 'proxyUrl' or update the package."),(F.proxyUrl||F.proxy)===Qt&&(console.warn(`[DEPRECATED] Warning: configuration.proxyUrl points to our old proxy (${Qt}).`),console.warn(`[DEPRECATED] We are overwriting the value and use the new proxy URL (${Ut}) instead.`),delete F.proxy,F.proxyUrl=Ut,console.warn(`[DEPRECATED] Action Required: You should manually update your configuration to use the new URL (${Ut}). Read more: https://github.com/scalar/scalar`))}),{immediate:!0});const Q=(0,E.EW)((()=>`<style>\n  ${(0,C.qL)(t.configuration.theme,{fonts:t.configuration.withDefaultFonts})}</style>`));return(e,t)=>{var n,r;return(0,E.uX)(),(0,E.CE)(E.FK,null,[(0,E.Lk)("div",{innerHTML:Q.value},null,8,nl),(0,E.Lk)("div",{ref_key:"documentEl",ref:i,class:(0,_.C4)(["scalar-app scalar-api-reference references-layout",[{"scalar-scrollbars-obtrusive":s.value,"references-editable":e.configuration.isEditable,"references-sidebar":e.configuration.showSidebar,"references-sidebar-mobile-open":(0,S.R1)(u),"references-classic":"classic"===e.configuration.layout},e.$attrs.class]]),style:(0,_.Tr)({"--scalar-y-offset":`var(--scalar-custom-header-height, ${R.value}px)`}),onScrollPassive:t[4]||(t[4]=(...e)=>(0,S.R1)(D)&&(0,S.R1)(D)(...e))},[(0,E.Lk)("div",rl,[(0,E.RG)(e.$slots,"header",(0,_._B)((0,E.Ng)(M.value)),void 0,!0)]),e.configuration.showSidebar?((0,E.uX)(),(0,E.CE)("aside",al,[(0,E.Lk)("div",ol,[(0,E.bF)((0,S.R1)(Ct.XO),null,{default:(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Yn),{operationsSorter:e.configuration.operationsSorter,parsedSpec:e.parsedSpec,tagsSorter:e.configuration.tagsSorter},{"sidebar-start":(0,E.k6)((()=>[(0,E.RG)(e.$slots,"sidebar-start",(0,_._B)((0,E.Ng)(M.value)),void 0,!0)])),"sidebar-end":(0,E.k6)((()=>[(0,E.RG)(e.$slots,"sidebar-end",(0,_._B)((0,E.Ng)(M.value)),void 0,!0)])),_:3},8,["operationsSorter","parsedSpec","tagsSorter"])])),_:3})])])):(0,E.Q3)("",!0),(0,E.bo)((0,E.Lk)("div",il,[(0,E.Lk)("div",sl,[(0,E.RG)(e.$slots,"editor",(0,_._B)((0,E.Ng)(M.value)),void 0,!0)])],512),[[rt.aG,e.configuration.isEditable]]),P.value?((0,E.uX)(),(0,E.CE)(E.FK,{key:1},[(0,E.Lk)("section",{"aria-label":`Open API Documentation for ${null==(n=e.parsedSpec.info)?void 0:n.title}`,class:"references-rendered"},[(0,E.bF)((0,S.R1)(el),{layout:e.configuration.layout,parsedSpec:e.parsedSpec},(0,E.eX)({start:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"content-start",(0,_._B)((0,E.Ng)(M.value)),void 0,!0)])),end:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"content-end",(0,_._B)((0,E.Ng)(M.value)),void 0,!0)])),_:2},[null!=(r=e.configuration)&&r.isEditable?{name:"empty-state",fn:(0,E.k6)((()=>{var n;return[(0,E.bF)(Lt,{theme:(null==(n=e.configuration)?void 0:n.theme)||"default",onChangeTheme:t[0]||(t[0]=t=>e.$emit("changeTheme",t)),onLinkSwaggerFile:t[1]||(t[1]=t=>e.$emit("linkSwaggerFile")),onLoadSwaggerFile:t[2]||(t[2]=t=>e.$emit("loadSwaggerFile")),onUpdateContent:t[3]||(t[3]=t=>e.$emit("updateContent",t))},null,8,["theme"])]})),key:"0"}:void 0]),1032,["layout","parsedSpec"])],8,ll),e.$slots.footer?((0,E.uX)(),(0,E.CE)("div",cl,[(0,E.RG)(e.$slots,"footer",(0,_._B)((0,E.Ng)(M.value)),void 0,!0)])):(0,E.Q3)("",!0)],64)):(0,E.Q3)("",!0),(0,E.bF)((0,S.R1)(tl),{configuration:e.configuration},null,8,["configuration"])],38),(0,E.bF)((0,S.R1)(Tt.jx))],64)}}}),[["__scopeId","data-v-10065932"]]),dl={class:"references-classic-header-container"},pl={class:"references-classic-header"},hl=It({},[["render",function(e,t){return(0,E.uX)(),(0,E.CE)("div",dl,[(0,E.Lk)("div",pl,[(0,E.RG)(e.$slots,"default",{},void 0,!0),(0,E.RG)(e.$slots,"dark-mode-toggle",{},void 0,!0)])])}],["__scopeId","data-v-a860f78f"]]);var fl=n(5929);function ml({specification:e}){const{getHeadingId:t,getWebhookId:n,getModelId:r,getOperationId:a,getTagId:o}=dn(),{hideModels:i}=Hn(),s=(0,S.KR)([]),l=(0,S.KR)([]),c=(0,S.KR)(0),u=(0,S.KR)(""),d=new fl.A(s.value,{keys:["title","description","body"]}),p=()=>{c.value=0,l.value=d.search(u.value)};(0,E.wB)(u,(e=>{e.length?p():l.value=[]}));const h=(0,E.EW)((()=>0===u.value.length?s.value.slice(0,25).map((e=>({item:e}))):l.value.slice(0,25)));return(0,E.wB)(e.value,(async()=>{var l,c,u,p,h,f,m,g,v;if(s.value=[],!(null!=(c=null==(l=e.value)?void 0:l.tags)&&c.length||null!=(p=null==(u=e.value)?void 0:u.webhooks)&&p.length))return void d.setCollection([]);const b=[],y=Mn((null==(f=null==(h=e.value)?void 0:h.info)?void 0:f.description)??"");y.length&&(y.forEach((e=>{b.push({type:"heading",title:`Info > ${e.value}`,description:"",href:`#${t(e)}`,tag:e.slug,body:""})})),s.value=s.value.concat(b)),null==(g=null==(m=e.value)?void 0:m.tags)||g.forEach((e=>{const t={title:e["x-displayName"]??e.name,href:`#${o(e)}`,description:e.description,type:"tag",tag:e.name,body:""};s.value.push(t),e.operations&&e.operations.forEach((t=>{const{parameterMap:n}=function(e){return{parameterMap:(0,E.EW)((()=>{var t;const n={path:[],query:[],header:[],cookie:[],body:[],formData:[]};e.pathParameters&&e.pathParameters.forEach((e=>{"path"===e.in?n.path.push(e):"query"===e.in?n.query.push(e):"header"===e.in?n.header.push(e):"cookie"===e.in?n.cookie.push(e):"body"===e.in?n.body.push(e):"formData"===e.in&&n.formData.push(e)}));const r=(null==(t=e.information)?void 0:t.parameters)??[];return r&&r.forEach((e=>{"path"===e.in?n.path.push(e):"query"===e.in?n.query.push(e):"header"===e.in?n.header.push(e):"cookie"===e.in?n.cookie.push(e):"body"===e.in?n.body.push(e):"formData"===e.in&&n.formData.push(e)})),n}))}}(t),r=function(e){var t,n,r;try{const a=null==(r=null==(n=null==(t=null==e?void 0:e.information)?void 0:t.requestBody)?void 0:n.content)?void 0:r["application/json"];if(!a)throw new Error("Body not found");return function(e){var t;const n=["Body"],r=null==(t=null==e?void 0:e.schema)?void 0:t.properties;return r&&Object.keys(r).forEach((t=>{if(!e.schema)return;n.push(function(e,t){let n=e;return n+=t.required&&t.required.includes(e)?" REQUIRED ":" optional ",t.properties[e]&&(n+=t.properties[e].type,t.properties[e].description&&(n+=" "+t.properties[e].description)),n}(t,e.schema));const a=r[t];"object"===a.type&&a.properties&&a.properties&&Object.keys(a.properties).forEach((e=>{var t,r;n.push(`${e} ${null==(r=null==(t=a.properties)?void 0:t[e])?void 0:r.type}`)}))})),n}(a)}catch{return!1}}(t)||n.value;let o=null;"boolean"!=typeof r&&(o=r);const i={type:"req",title:t.name??t.path,href:`#${a(t,e)}`,operationId:t.operationId,description:t.description??"",httpVerb:t.httpVerb,path:t.path,tag:e.name,operation:t};o&&(i.body=o),s.value.push(i)}))}));const w=null==(v=e.value)?void 0:v.webhooks,O=[];w&&Object.keys(w).forEach((e=>{Object.keys(w[e]).forEach((t=>{var r;O.push({type:"webhook",title:`Webhook: ${null==(r=w[e][t])?void 0:r.name}`,href:`#${n({name:e,method:t})}`,description:e,httpVerb:t,tag:e,body:""})})),s.value=s.value.concat(O)}));const k=i.value?{}:Rn(e.value),x=[];k&&(Object.keys(k).forEach((e=>{x.push({type:"model",title:"Model",href:`#${r({name:e})}`,description:k[e].title??e,tag:e,body:""})})),s.value=s.value.concat(x)),d.setCollection(s.value)}),{immediate:!0}),{resetSearch:function(){u.value="",c.value=0,l.value=[]},fuseSearch:p,selectedSearchResult:c,searchResultsWithPlaceholderResults:h,searchText:u}}const gl={class:"sr-only"},vl={class:"sr-only"},bl=It((0,E.pM)({__name:"SearchModal",props:{parsedSpec:{},modalState:{}},setup(e){const t=e,n=(0,S.lW)(t,"parsedSpec"),r=((e=21)=>{let t="",n=crypto.getRandomValues(new Uint8Array(e|=0));for(;e--;)t+="useandom-26T198340PX75pxJACKVERYMINDBUSHWOLF_GQZbfghjklqvwyzrict"[63&n[e]];return t})(),a=`${r}-search-result`,o=`${r}-search-instructions`,i=e=>`${r}${e}`,{resetSearch:s,fuseSearch:l,selectedSearchResult:c,searchResultsWithPlaceholderResults:u,searchText:d}=ml({specification:n}),p={heading:"DocsPage",model:"Brackets",req:"Terminal",tag:"CodeFolder",webhook:"Terminal"},h={heading:"Document Heading",model:"",req:"Request",tag:"Tag",webhook:"Webhook"},f=(0,S.KR)(null);(0,E.wB)((()=>t.modalState.open),(e=>{e&&s()}));const{setCollapsedSidebarItem:m}=Hn(),g=/#(tag\/[^/]*)/;function v(e){let n="models";const r=e.item.href.match(g);null!=r&&r.length&&r.length>1&&(n=r[1]),m(n,!0);const a=e.item.href.replace("#","");if(document.getElementById(a))pn(a),t.modalState.hide();else{const e=Tn.on((n=>{n.id===a&&(pn(a),e(),t.modalState.hide())}))}}(0,E.wB)(c,(e=>{var t;const n=u.value[e],r=i(null==n?void 0:n.item.href);null==(t=document.getElementById(r))||t.scrollIntoView({behavior:"smooth",block:"nearest"})}));const b=(0,E.EW)((()=>{const e=c.value,t=u.value[e];return i(null==t?void 0:t.item.href)})),y=e=>{const t="up"===e?-1:1,n=u.value.length,r=(c.value+t+n)%n;c.value=r};function w(e){const t=new URL(window.location.href);return t.hash=e,t.toString()}return(e,t)=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.uT),{state:e.modalState,variant:"search",onKeydown:[t[1]||(t[1]=(0,rt.jR)((0,rt.D$)((e=>y("down")),["stop","prevent"]),["down"])),t[2]||(t[2]=(0,rt.jR)((0,rt.D$)((e=>v((0,S.R1)(u)[(0,S.R1)(c)])),["stop","prevent"]),["enter"])),t[3]||(t[3]=(0,rt.jR)((0,rt.D$)((e=>y("up")),["stop","prevent"]),["up"]))]},{default:(0,E.k6)((()=>[(0,E.Lk)("div",{ref_key:"searchModalRef",ref:f,"aria-label":"Reference Search",class:"ref-search-container",role:"search"},[(0,E.bF)((0,S.R1)(Ct.vW),{modelValue:(0,S.R1)(d),"onUpdate:modelValue":t[0]||(t[0]=e=>(0,S.i9)(d)?d.value=e:null),"aria-activedescendant":b.value,"aria-controls":a,"aria-describedby":o,onInput:(0,S.R1)(l)},null,8,["modelValue","aria-activedescendant","onInput"])],512),(0,E.bF)((0,S.R1)(Ct.oN),{id:a,"aria-describedby":o,"aria-label":"Reference Search Results",class:"ref-search-results custom-scroll",noResults:!(0,S.R1)(u).length},{query:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)((0,S.R1)(d)),1)])),default:(0,E.k6)((()=>[((0,E.uX)(!0),(0,E.CE)(E.FK,null,(0,E.pI)((0,S.R1)(u),((e,n)=>((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.rQ),{id:i(e.item.href),key:e.refIndex,active:(0,S.R1)(c)===n,href:w(e.item.href),icon:p[e.item.type],onClick:t=>v(e),onFocus:e=>c.value=n},(0,E.eX)({default:(0,E.k6)((()=>{var n,r,a,o;return[(0,E.Lk)("span",{class:(0,_.C4)({deprecated:null==(r=null==(n=e.item.operation)?void 0:n.information)?void 0:r.deprecated})},[(0,E.Lk)("span",gl,[(0,E.eW)((0,_.v_)(h[e.item.type])+":  ",1),null!=(o=null==(a=e.item.operation)?void 0:a.information)&&o.deprecated?((0,E.uX)(),(0,E.CE)(E.FK,{key:0},[(0,E.eW)(" (Deprecated)  ")],64)):(0,E.Q3)("",!0)]),(0,E.eW)(" "+(0,_.v_)(e.item.title)+" ",1),t[4]||(t[4]=(0,E.Lk)("span",{class:"sr-only"},",",-1))],2)]})),_:2},[(e.item.httpVerb||e.item.path)&&e.item.path!==e.item.title?{name:"description",fn:(0,E.k6)((()=>[t[5]||(t[5]=(0,E.Lk)("span",{class:"sr-only"},"Path: ",-1)),(0,E.eW)(" "+(0,_.v_)(e.item.path),1)])),key:"0"}:e.item.description?{name:"description",fn:(0,E.k6)((()=>[t[6]||(t[6]=(0,E.Lk)("span",{class:"sr-only"},"Description: ",-1)),(0,E.eW)(" "+(0,_.v_)(e.item.description),1)])),key:"1"}:void 0,"req"===e.item.type?{name:"addon",fn:(0,E.k6)((()=>[(0,E.bF)(Vt,{"aria-hidden":"true",method:e.item.httpVerb??"get"},null,8,["method"]),(0,E.Lk)("span",vl," HTTP Method: "+(0,_.v_)(e.item.httpVerb??"get")+", ",1)])),key:"2"}:void 0]),1032,["id","active","href","icon","onClick","onFocus"])))),128))])),_:1},8,["noResults"]),t[7]||(t[7]=(0,E.Lk)("div",{class:"ref-search-meta"},[(0,E.Lk)("span",null,"↑↓ Navigate"),(0,E.Lk)("span",null,"⏎ Select")],-1)),(0,E.Lk)("div",{id:o,class:"sr-only"}," Press up arrow / down arrow to navigate, enter to select, type to filter results ")])),_:1},8,["state"]))}}),[["__scopeId","data-v-829b4762"]]),yl=()=>typeof navigator<"u"&&/Mac/.test(navigator.platform),wl={class:"sidebar-search-input"},Ol={class:"sidebar-search-shortcut"},kl={class:"sidebar-search-key"},xl=It((0,E.pM)({__name:"SearchButton",props:{spec:{},searchHotKey:{default:"k"}},setup(e){const t=e,n=(0,Ct.hS)(),{client:r}=ei(),a=e=>{var a;(yl()?e.metaKey:e.ctrlKey)&&e.key===t.searchHotKey&&(null==(a=r.value)||!a.modalState.open)&&(e.preventDefault(),e.stopPropagation(),n.open?n.hide():n.show())};return(0,E.sV)((()=>window.addEventListener("keydown",a))),(0,E.xo)((()=>window.removeEventListener("keydown",a))),(e,t)=>((0,E.uX)(),(0,E.CE)(E.FK,null,[(0,E.Lk)("button",{class:(0,_.C4)(["sidebar-search",e.$attrs.class]),type:"button",onClick:t[0]||(t[0]=(...e)=>(0,S.R1)(n).show&&(0,S.R1)(n).show(...e))},[(0,E.bF)((0,S.R1)(Ct.kv),{class:"scalar-search-icon",icon:"Search",size:"sm",thickness:"2.5"}),(0,E.Lk)("div",wl,[t[2]||(t[2]=(0,E.Lk)("span",{class:"sr-only"},"Open ",-1)),t[3]||(t[3]=(0,E.Lk)("span",{class:"sidebar-search-placeholder"},"Search",-1)),(0,E.Lk)("span",Ol,[t[1]||(t[1]=(0,E.Lk)("span",{class:"sr-only"},"(Keyboard Shortcut)",-1)),(0,E.Lk)("kbd",kl,(0,_.v_)((0,S.R1)(yl)()?"⌘":"⌃")+(0,_.v_)(e.searchHotKey),1)])])],2),(0,E.bF)(bl,{modalState:(0,S.R1)(n),parsedSpec:e.spec},null,8,["modalState","parsedSpec"])],64))}}),[["__scopeId","data-v-9f0bd6b4"]]),El=(0,E.pM)({__name:"ClassicLayout",props:{configuration:{},parsedSpec:{},rawSpec:{},isDark:{type:Boolean}},emits:["toggleDarkMode","updateContent"],setup(e){const t=e,n=(0,E.Ht)(),r=(0,E.EW)((()=>({...t.configuration,showSidebar:!1})));return(e,a)=>((0,E.uX)(),(0,E.Wv)(ul,{configuration:r.value,parsedSpec:e.parsedSpec,rawSpec:e.rawSpec,onUpdateContent:a[1]||(a[1]=t=>e.$emit("updateContent",t))},(0,E.eX)({"content-start":(0,E.k6)((({spec:n})=>[(0,E.bF)(hl,null,{"dark-mode-toggle":(0,E.k6)((()=>[t.configuration.hideDarkModeToggle?(0,E.Q3)("",!0):((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.en),{key:0,class:"text-c-2 hover:text-c-1",mode:e.isDark?"dark":"light",style:{transform:"scale(1.4)"},variant:"icon",onClick:a[0]||(a[0]=t=>e.$emit("toggleDarkMode"))},null,8,["mode"]))])),default:(0,E.k6)((()=>[t.configuration.hideSearch?(0,E.Q3)("",!0):((0,E.uX)(),(0,E.Wv)((0,S.R1)(xl),{key:0,class:"t-doc__sidebar",searchHotKey:r.value.searchHotKey,spec:n},null,8,["searchHotKey","spec"]))])),_:2},1024)])),_:2},[(0,E.pI)(n,((t,n)=>({name:n,fn:(0,E.k6)((t=>[(0,E.RG)(e.$slots,n,(0,_._B)((0,E.Ng)(t||{})))]))})))]),1032,["configuration","parsedSpec","rawSpec"]))}});var Sl=n(8207),_l=(n(5336),n(3858),n(4325));const Cl={class:"references-mobile-header t-doc__header"},Tl={class:"references-mobile-breadcrumbs"},Al={class:"references-mobile-header-actions"},Rl=It((0,E.pM)({__name:"MobileHeader",props:{open:{type:Boolean}},emits:["update:open"],setup(e){const{breadcrumb:t}=Hn();return(e,n)=>((0,E.uX)(),(0,E.CE)("div",Cl,[(0,E.bF)((0,S.R1)(Ct.EX),{icon:e.open?"Close":"Menu",label:e.open?"Close Menu":"Open Menu",size:"md",onClick:n[0]||(n[0]=t=>e.$emit("update:open",!e.open))},null,8,["icon","label"]),(0,E.Lk)("span",Tl,(0,_.v_)((0,S.R1)(t)),1),(0,E.Lk)("div",Al,[(0,E.RG)(e.$slots,"actions",{},void 0,!0)])]))}}),[["__scopeId","data-v-b5fb4b8b"]]),Pl={key:0,class:"scalar-api-references-standalone-search"},Dl={key:1},Ml=It((0,E.pM)({__name:"ModernLayout",props:{configuration:{},parsedSpec:{},rawSpec:{},isDark:{type:Boolean}},emits:["toggleDarkMode","updateContent"],setup(e){const t=e,n=(0,E.Ht)(),{mediaQueries:r}=(0,_l.G)(),{isSidebarOpen:a}=Hn();(0,E.wB)(r.lg,((e,t)=>{t&&!e&&(a.value=!1)}));const{hash:o}=dn();return(0,E.wB)(o,((e,t)=>{e&&e!==t&&(a.value=!1)})),(e,r)=>((0,E.uX)(),(0,E.Wv)(ul,{class:(0,_.C4)({"scalar-api-references-standalone-mobile":e.configuration.showSidebar}),configuration:e.configuration,parsedSpec:e.parsedSpec,rawSpec:e.rawSpec,onUpdateContent:r[2]||(r[2]=t=>e.$emit("updateContent",t))},(0,E.eX)({header:(0,E.k6)((()=>[t.configuration.showSidebar?((0,E.uX)(),(0,E.Wv)(Rl,{key:0,open:(0,S.R1)(a),"onUpdate:open":r[0]||(r[0]=e=>(0,S.i9)(a)?a.value=e:null)},null,8,["open"])):(0,E.Q3)("",!0)])),"sidebar-start":(0,E.k6)((({spec:e})=>{var n;return[t.configuration.hideSearch?(0,E.Q3)("",!0):((0,E.uX)(),(0,E.CE)("div",Pl,[(0,E.bF)((0,S.R1)(xl),{searchHotKey:null==(n=t.configuration)?void 0:n.searchHotKey,spec:e},null,8,["searchHotKey","spec"])]))]})),"sidebar-end":(0,E.k6)((()=>[(0,E.bF)((0,S.R1)(Ct.iQ),{class:"darklight-reference"},{toggle:(0,E.k6)((()=>[t.configuration.hideDarkModeToggle?((0,E.uX)(),(0,E.CE)("span",Dl)):((0,E.uX)(),(0,E.Wv)((0,S.R1)(Ct.Jt),{key:0,modelValue:e.isDark,"onUpdate:modelValue":r[1]||(r[1]=t=>e.$emit("toggleDarkMode"))},null,8,["modelValue"]))])),default:(0,E.k6)((()=>{var n;return[t.configuration.hideClientButton?(0,E.Q3)("",!0):((0,E.uX)(),(0,E.Wv)((0,S.R1)(Sl.A),{key:0,buttonSource:"sidebar",integration:e.configuration._integration,isDevelopment:!1,url:null==(n=e.configuration.spec)?void 0:n.url},null,8,["integration","url"]))]})),_:1})])),_:2},[(0,E.pI)(n,((t,n)=>({name:n,fn:(0,E.k6)((t=>[(0,E.RG)(e.$slots,n,(0,_._B)((0,E.Ng)(t||{})),void 0,!0)]))})))]),1032,["class","configuration","parsedSpec","rawSpec"]))}}),[["__scopeId","data-v-3ca2ed96"]]),$l=(0,E.pM)({__name:"Layouts",props:{configuration:{},parsedSpec:{},rawSpec:{},isDark:{type:Boolean}},emits:["toggleDarkMode","updateContent"],setup(e){const t=e,n=(0,E.Ht)(),r={modern:Ml,classic:El};return(e,a)=>((0,E.uX)(),(0,E.Wv)((0,E.$y)(r[e.configuration.layout??"modern"]??Ml),(0,E.v6)(t,{onToggleDarkMode:a[0]||(a[0]=t=>e.$emit("toggleDarkMode")),onUpdateContent:a[1]||(a[1]=t=>e.$emit("updateContent",t))}),(0,E.eX)({_:2},[(0,E.pI)(n,((t,n)=>({name:n,fn:(0,E.k6)((t=>[(0,E.RG)(e.$slots,n,(0,_._B)((0,E.Ng)(t||{})))]))})))]),1040))}}),Nl=(0,E.pM)({__name:"ApiReference",props:{configuration:{}},emits:["updateContent"],setup(e){var t,n,r;const a=e,{toggleColorMode:o,isDarkMode:i}=(0,T.G)({initialColorMode:null!=(t=a.configuration)&&t.darkMode?"dark":void 0,overrideColorMode:null==(n=a.configuration)?void 0:n.forceDarkModeState});(0,E.wB)((()=>{var e;return null==(e=a.configuration)?void 0:e.darkMode}),(e=>i.value=!!e));const s=(0,E.EW)((()=>{var e,t;if(null!=(e=a.configuration)&&e.customCss)return(0,C.kL)(null==(t=a.configuration)?void 0:t.customCss)})),l=(0,E.EW)((()=>{var e;return{spec:{content:void 0,url:void 0,...null==(e=a.configuration)?void 0:e.spec},proxyUrl:void 0,theme:"default",showSidebar:!0,isEditable:!1,...a.configuration,customCss:s.value}}));null!=(r=l.value)&&r.metaData&&function(e,t){const{title:n,titleTemplate:r,...a}=e;je({title:n,titleTemplate:r,_flatMeta:a},{...t,transform(e){const t=G({...e._flatMeta});return delete e._flatMeta,{...e,meta:t}}})}(l.value.metaData);const{parsedSpec:c,rawSpec:u}=nt({proxyUrl:(0,S.lW)((()=>l.value.proxyUrl||l.value.proxy||"")),specConfig:(0,S.lW)((()=>l.value.spec||{}))}),d=(0,E.EW)((()=>l.value.favicon));return(0,We.iwM)(d),(e,t)=>{var n;return(0,E.uX)(),(0,E.CE)(E.FK,null,[null!=(n=l.value)&&n.customCss?((0,E.uX)(),(0,E.Wv)((0,E.$y)("style"),{key:0},{default:(0,E.k6)((()=>[(0,E.eW)((0,_.v_)(l.value.customCss),1)])),_:1})):(0,E.Q3)("",!0),(0,E.bF)((0,S.R1)($l),{configuration:l.value,isDark:(0,S.R1)(i),parsedSpec:(0,S.R1)(c),rawSpec:(0,S.R1)(u),onToggleDarkMode:t[0]||(t[0]=()=>(0,S.R1)(o)()),onUpdateContent:t[1]||(t[1]=t=>e.$emit("updateContent",t))},{footer:(0,E.k6)((()=>[(0,E.RG)(e.$slots,"footer")])),"content-end":(0,E.k6)((()=>[(0,E.RG)(e.$slots,"footer")])),_:3},8,["configuration","isDark","parsedSpec","rawSpec"])],64)}}});function Il(e,t,n=!0){const r=(0,S.Kh)(t),a=(0,rt.Ef)(Nl,{configuration:r}),o=function(e={}){e.domDelayFn=e.domDelayFn||(e=>(0,E.dY)((()=>setTimeout((()=>e()),0))));const t=Ne(e);return t.use(Fe),t.install=function(e){return{install(t){Le&&(t.config.globalProperties.$unhead=e,t.config.globalProperties.$head=e,t.provide(Qe,e))}}.install}(t),t}();function i(t=e){t?a.mount(t):console.warn("Invalid HTML element provided. Cannot mount Scalar References")}return a.use(o),n&&i(),{updateConfig(e,t=!0){t?Object.assign(r,e):(0,Xe.O1)(r,e)},updateSpec(e){r.spec=e},mount:i,unmount:()=>a.unmount()}}!function(){try{if("undefined"==typeof document||document.getElementById("scalar-style-api-reference"))return;setTimeout((()=>{if("true"===getComputedStyle(document.body).getPropertyValue("--scalar-loaded-api-reference"))return;const e=document.createElement("style");e.setAttribute("id","scalar-style-api-reference"),e.appendChild(document.createTextNode('.start[data-v-391b4e37]{padding:24px;display:flex;flex-flow:wrap;justify-content:space-between;position:relative;z-index:0}.swagger-editor .start[data-v-391b4e37]{padding-top:24px}.start-h1[data-v-391b4e37]{font-size:var(--scalar-heading-2);margin-top:0;line-height:1.45;margin-bottom:0;font-weight:var(--scalar-bold);color:var(--scalar-color-1);width:100%;position:relative}.start-h3[data-v-391b4e37]{font-size:var(--scalar-paragraph);margin-top:0;margin-bottom:6px;display:block;line-height:1.45;font-weight:var(--scalar-bold);color:var(--scalar-color-1);width:100%}.start-h1[data-v-391b4e37]:not(:first-of-type){margin-top:24px}.start-p[data-v-391b4e37]{font-size:var(--scalar-paragraph);color:var(--scalar-color-2);line-height:1.5;width:100%;margin-top:12px}.start-ul[data-v-391b4e37]{margin-top:12px;font-size:var(--scalar-paragraph);line-height:1.5;padding-left:0;list-style:initial;display:flex;flex-flow:wrap;gap:24px}.start-ul li[data-v-391b4e37]{margin:0;padding:0;list-style:none;width:calc(50% - 24px);color:var(--scalar-color-2)}.start-ul li[data-v-391b4e37]:first-of-type{margin-top:0}.start-section[data-v-391b4e37]{width:100%;margin-bottom:12px;display:flex;flex-flow:wrap}.start-section[data-v-391b4e37]:last-of-type{margin-bottom:48px}.start-h2[data-v-391b4e37]{background:var(--scalar-background-2);border-top-left-radius:var(--scalar-radius-lg);border-top-right-radius:var(--scalar-radius-lg);border:1px solid var(--scalar-border-color);color:var(--scalar-color-3);font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);padding:9px;width:100%}.start-item[data-v-391b4e37]{align-items:center;background:var(--scalar-background-2);border-right:1px solid var(--scalar-border-color);border-bottom:1px solid var(--scalar-border-color);color:var(--scalar-color-1);cursor:pointer;display:flex;flex:1;font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);padding:9px;text-transform:capitalize;-webkit-user-select:none;-moz-user-select:none;user-select:none}.start-section-integrations .start-item[data-v-391b4e37]:first-of-type{border-bottom-left-radius:var(--scalar-radius-lg);border-left:1px solid var(--scalar-border-color)}.start-section-integrations .start-item[data-v-391b4e37]:last-of-type{border-bottom-right-radius:var(--scalar-radius-lg)}.start-section-colors .start-item[data-v-391b4e37]{min-width:33.33%}.start-section-colors .start-item[data-v-391b4e37]:nth-child(3n+2){border-left:1px solid var(--scalar-border-color)}.start-section-colors .start-item[data-v-391b4e37]:last-of-type,.start-section-colors .start-item-active[data-v-391b4e37]:last-of-type:before{border-radius:0 0 var(--scalar-radius-lg) var(--scalar-radius-lg)}.start-item[data-v-391b4e37]:empty{pointer-events:none}.start-item svg[data-v-391b4e37]{width:14px;height:14px;margin-right:6px}.start-item[data-v-391b4e37]:hover{background:var(--scalar-background-3)}.start-item-active[data-v-391b4e37]{z-index:10;position:relative;color:var(--scalar-color-1)}.start-item-active[data-v-391b4e37]:before{border:1px solid var(--scalar-color-1);content:"";top:-1px;right:-1px;bottom:-1px;left:-1px;pointer-events:none;position:absolute}.start-section-color .start-item[data-v-391b4e37]{text-transform:capitalize}.start-cta[data-v-391b4e37]{display:flex;gap:12px;width:100%;margin-top:24px;margin-bottom:0}.start-row[data-v-391b4e37]{width:100%;margin-top:12px;overflow:hidden}.start-hero-copy[data-v-391b4e37]{background:var(--scalar-background-2);padding:12px;border-radius:var(--scalar-radius-lg)}.start-p-small[data-v-391b4e37]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-2);margin-bottom:12px;line-height:1.4}.start-cta[data-v-391b4e37]{margin-bottom:12px;width:-moz-fit-content;width:fit-content;white-space:nowrap}.start-copy[data-v-391b4e37]{padding:76px 48px 48px;display:flex;justify-content:center;align-items:center;flex-direction:column;text-align:center}.start-logo[data-v-391b4e37]{color:var(--scalar-color-1);margin-bottom:24px;width:72px;aspect-ratio:1;position:relative;box-shadow:var(--scalar-shadow-2);border-radius:50%}.start-logo[data-v-391b4e37]:before{content:"";width:300%;aspect-ratio:1;left:-100%;top:-100%;position:absolute;z-index:-1;border-radius:50%;background-size:24px 24px;box-shadow:inset 0 0 50px var(--scalar-background-1),inset 0 0 50px var(--scalar-background-1);background-image:linear-gradient(to right,var(--scalar-border-color) 1px,transparent 1px),linear-gradient(to bottom,var(--scalar-border-color) 1px,transparent 1px)}.start-logo svg[data-v-391b4e37]{width:100%;height:auto;background:var(--scalar-background-1);padding:3px;border-radius:50%;position:relative}@media screen and (max-width: 600px){.start-section-colors .start-item[data-v-391b4e37],.start-item[data-v-391b4e37]{width:100%;border-radius:0;border-right:none;border-top:1px solid var(--scalar-border-color)}.start-item[data-v-391b4e37]:empty{display:none}.start-h2[data-v-391b4e37]{border-bottom:none}.start li[data-v-391b4e37]{width:100%}.start-copy[data-v-391b4e37]{padding:48px 0 24px}}@media screen and (max-width: 1000px){.start[data-v-391b4e37]{padding:0;overflow:auto}}.sidebar-heading-type[data-v-f51e2805]{display:block;min-width:3.9em;overflow:hidden;line-height:14px;flex-shrink:0;color:#fff;color:color-mix(in srgb,var(--method-color, var(--scalar-color-1)),transparent 0%);text-transform:uppercase;font-size:10px;font-weight:var(--scalar-bold);text-align:right;position:relative;font-family:var(--scalar-font-code);white-space:nowrap;margin-left:3px}.sidebar-heading[data-v-18e6dc3d]{display:flex;gap:6px;color:var(--scalar-sidebar-color-2, var(--scalar-color-2));font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);word-break:break-word;line-height:1.385;max-width:100%;position:relative;cursor:pointer;border-radius:var(--scalar-radius);flex:1;padding-right:9px;-webkit-user-select:none;-moz-user-select:none;user-select:none}.sidebar-heading-link-method[data-v-18e6dc3d]{margin:0}.sidebar-heading.deprecated .sidebar-heading-link-title[data-v-18e6dc3d]{text-decoration:line-through}.sidebar-heading-link-title[data-v-18e6dc3d]{margin:0}.sidebar-heading[data-v-18e6dc3d]:hover{background:var( --scalar-sidebar-item-hover-background, var(--scalar-background-2) )}.sidebar-heading:hover .sidebar-heading-link-title[data-v-18e6dc3d]{color:var(--scalar-sidebar-item-hover-color)}.sidebar-heading-link[data-v-18e6dc3d]:focus-visible{outline:none}.sidebar-heading[data-v-18e6dc3d]:has(>.sidebar-heading-link:focus-visible){z-index:1;outline:1px solid var(--scalar-color-accent)}.active_page.sidebar-heading[data-v-18e6dc3d]:hover,.active_page.sidebar-heading[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-active, var(--scalar-color-accent));background:var( --scalar-sidebar-item-active-background, var(--scalar-background-accent) )}.active_page.sidebar-heading p[data-v-18e6dc3d]{font-weight:var(--scalar-sidebar-font-weight-active, var(--scalar-semibold))}.active_page.sidebar-heading:hover .sidebar-heading-link-title[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-active, var(--scalar-color-accent))}.sidebar-indent-nested .sidebar-indent-nested .sidebar-heading[data-v-18e6dc3d]:before{content:"";position:absolute;top:0;left:calc((var(--scalar-sidebar-level) * 12px));width:1px;height:100%;background:var(--scalar-sidebar-indent-border)}.sidebar-indent-nested .sidebar-indent-nested .sidebar-heading[data-v-18e6dc3d]:hover:before{background:var(--scalar-sidebar-indent-border-hover)}.sidebar-indent-nested .sidebar-indent-nested .active_page.sidebar-heading[data-v-18e6dc3d]:before{background:var(--scalar-sidebar-indent-border-active)}.sidebar-heading-link[data-v-18e6dc3d]{text-decoration:none;color:inherit;padding:6px 0;display:flex;flex:1;justify-content:space-between;gap:2px}.sidebar-heading p[data-v-18e6dc3d]{height:-moz-fit-content;height:fit-content;display:flex;align-items:center;font-weight:var(--scalar-sidebar-font-weight, var(--scalar-semibold))}.sidebar-heading p[data-v-18e6dc3d]:empty{display:none}.link-icon[data-v-18e6dc3d]{position:relative;left:4px}.sidebar-icon[data-v-18e6dc3d]{display:flex;align-items:center;justify-content:center;margin-right:6px;width:13px;height:13px}.sidebar-icon>svg[data-v-18e6dc3d]{width:13px;height:13px}.sidebar-group-item[data-v-18e6dc3d]{position:relative}.sidebar-heading-chevron[data-v-18e6dc3d]{margin:5px -5.5px 5px -9px}.sidebar-heading-chevron .toggle-nested-icon[data-v-18e6dc3d]:focus-visible{outline:none}.sidebar-heading[data-v-18e6dc3d]:has(.sidebar-heading-chevron .toggle-nested-icon:focus-visible){outline:none;box-shadow:inset 0 0 0 1px var(--scalar-color-accent)}.toggle-nested-icon[data-v-18e6dc3d]{color:var(--scalar-color-3);width:20px;height:20px;display:flex;align-items:center;justify-content:center}.active_page .toggle-nested-icon[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-active, var(--scalar-color-accent))}.toggle-nested-icon[data-v-18e6dc3d]:hover,.toggle-nested-icon[data-v-18e6dc3d]:focus-visible{color:currentColor}.action-menu[data-v-18e6dc3d]{position:absolute;top:5px;right:5px;display:flex;gap:6px}.action-menu[data-v-18e6dc3d] .button-wrapper button{opacity:0;width:20px;height:20px;padding:4px}.action-menu[data-v-18e6dc3d]:hover .button-wrapper button,.action-menu[data-v-18e6dc3d] .button-wrapper button:hover,.sidebar-heading:hover~.action-menu[data-v-18e6dc3d] .button-wrapper button,.action-menu[data-v-18e6dc3d] .button-wrapper button[aria-expanded=true]{opacity:1}.sidebar-heading[data-v-18e6dc3d]:has(~.action-menu:hover){color:var(--scalar-sidebar-color-1, var(--scalar-color-1));background:var( --scalar-sidebar-item-hover-background, var(--scalar-background-2) )}.sidebar-group-item__folder[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-1, var(--scalar-color-1));text-transform:var(--scalar-tag-text-transform, initial)}.sidebar-group[data-v-39c84840]{list-style:none;width:100%;margin:0;padding:0}.sidebar-indent-nested[data-v-39c84840] .sidebar-heading{padding-left:calc((var(--scalar-sidebar-level) * var(--scalar-sidebar-indent-base)) + 12px)!important}.sidebar-indent-nested[data-v-39c84840] .sidebar-heading .toggle-nested-icon{left:calc((var(--scalar-sidebar-level) * var(--scalar-sidebar-indent-base)) + 2px)!important}:where(.sidebar-indent-nested[data-v-39c84840]) .sidebar-heading{color:var(--scalar-sidebar-color-1, var(--scalar-color-1))}:where(.sidebar-indent-nested[data-v-39c84840]) :where(.sidebar-indent-nested) .sidebar-heading{color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}.sidebar[data-v-b863366c]{--scalar-sidebar-indent-base: 12px;--scalar-sidebar-font-weight-active: var(--scalar-semibold);--scalar-sidebar-font-weight: var(--scalar-semibold);flex:1;height:100%;display:flex;flex-direction:column;border-right:var(--scalar-border-width) solid var(--scalar-sidebar-border-color, var(--scalar-border-color));background:var(--scalar-sidebar-background-1, var(--scalar-background-1));--scalar-sidebar-level: 0}.sidebar-pages[data-v-b863366c]{flex:1;padding:9px 12px}@media (max-width: 1000px){.sidebar[data-v-b863366c]{min-height:0;border-right:none}.sidebar-pages[data-v-b863366c]{padding-top:12px}}.sidebar-group-title[data-v-b863366c]{color:var(--scalar-sidebar-color-1);font-size:var(--scalar-mini);padding:12px 6px 6px;font-weight:var(--scalar-semibold);text-transform:uppercase;word-break:break-word;line-height:1.385}.sidebar-group-item+.sidebar-group-title[data-v-b863366c]{border-top:var(--scalar-border-width) solid var(--scalar-sidebar-border-color);margin-top:9px}.show-more[data-v-252cd03d]{background:var(--scalar-background-1);-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none;border:var(--scalar-border-width) solid var(--scalar-border-color);margin:auto;padding:8px 12px;border-radius:30px;color:var(--scalar-color-1);font-weight:var(--scalar-semibold);font-size:var(--scalar-small);display:flex;align-items:center;justify-content:center;position:relative;top:-48px}.show-more[data-v-252cd03d]:hover{color:var(--scalar-color-2);cursor:pointer}.show-more-icon[data-v-252cd03d]{width:16px!important;height:16px!important;margin-left:3px}.show-more[data-v-252cd03d]:active{box-shadow:0 0 0 1px var(--scalar-border-color)}@container narrow-references-container (max-width: 900px){.show-more[data-v-252cd03d]{top:-24px}}.screenreader-only[data-v-df2e1026]{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);border:0}.section[data-v-64ccd510]{position:relative;display:flex;flex-direction:column;max-width:var(--refs-content-max-width);margin:auto;padding:90px 0;scroll-margin-top:var(--refs-header-height)}.section[data-v-64ccd510]:has(~div.contents){border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.references-classic .section[data-v-64ccd510]{padding:48px 0;gap:24px}@container narrow-references-container (max-width: 900px){.references-classic .section[data-v-64ccd510],.section[data-v-64ccd510]{padding:48px 24px}}.section[data-v-64ccd510]:not(:last-of-type){border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.loading[data-v-c90b2c46]{background:var(--scalar-background-3);animation:loading-skeleton-c90b2c46 1.5s infinite alternate;border-radius:var(--scalar-radius-lg);min-height:1.6em;margin:.6em 0;max-width:100%}.loading[data-v-c90b2c46]:first-of-type{min-height:3em;margin-bottom:24px;margin-top:0}.loading[data-v-c90b2c46]:last-of-type{width:60%}.loading.single-line[data-v-c90b2c46]{min-height:3em;margin:.6em 0;max-width:80%}@keyframes loading-skeleton-c90b2c46{0%{opacity:1}to{opacity:.33}}.section-header-wrapper[data-v-b6aef53d]{display:grid;grid-template-columns:1fr}@media (min-width: 1200px){.section-header-wrapper[data-v-b6aef53d]{grid-template-columns:repeat(2,1fr)}}.section-header[data-v-b6aef53d]{font-size:var(--font-size, var(--scalar-heading-2));font-weight:var(--font-weight, var(--scalar-bold));color:var(--scalar-color-1);word-wrap:break-word;line-height:1.45;margin-top:0;margin-bottom:12px}.section-header.tight[data-v-b6aef53d]{margin-bottom:6px}.section-header.loading[data-v-b6aef53d]{width:80%}.label[data-v-3cb2dd80]{position:relative;display:inline-block;word-break:break-all}.anchor[data-v-3cb2dd80]{position:relative;display:inline-block;opacity:0}.anchor-copy[data-v-3cb2dd80]{position:absolute;left:0;top:50%;transform:translateY(-50%);cursor:pointer;padding:0 6px;color:var(--scalar-color-3);font-weight:var(--scalar-semibold);font-size:.8em}.anchor-copy[data-v-3cb2dd80]:hover,.anchor-copy[data-v-3cb2dd80]:focus-visible{color:var(--scalar-color-2)}.label:hover .anchor[data-v-3cb2dd80],.label:has(:focus-visible) .anchor[data-v-3cb2dd80]{opacity:1}@container narrow-references-container (max-width: 900px){.section-content--with-columns[data-v-9735459e]{flex-direction:column;gap:24px}}.section-columns[data-v-8b9602bf]{display:flex;gap:48px}@container narrow-references-container (max-width: 900px){.section-columns[data-v-8b9602bf]{flex-direction:column;gap:24px}}.section-column[data-v-699c28e3]{flex:1;min-width:0}@container narrow-references-container (max-width: 900px){.section-column[data-v-699c28e3]:nth-of-type(2){padding-top:0}}.scalar-card[data-v-38801681]{all:unset;font-family:var(--scalar-font);border-radius:var(--scalar-radius-lg);overflow:hidden;border:var(--scalar-border-width) solid var(--scalar-border-color);background:var(--scalar-background-2);display:flex;flex-direction:column;max-height:calc(((var(--full-height) - var(--refs-header-height)) - 60px) / 2);position:relative}.scalar-card-sticky[data-v-38801681]{position:sticky;top:calc(var(--refs-header-height) + 24px)}.scalar-card-content[data-v-dd83e46e]{overflow:auto;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color);display:grid}.scalar-card-content[data-v-dd83e46e] .simple-table .simple-header{display:none}.scalar-card-content[data-v-dd83e46e]:last-of-type,.scalar-card-content.scalar-card--borderless[data-v-dd83e46e]{border-bottom:none}.scalar-card--muted[data-v-dd83e46e]{background:var(--scalar-background-2)}.scalar-card--contrast[data-v-dd83e46e]{background:var(--scalar-background-3)}.scalar-card--frameless[data-v-dd83e46e]{padding:0}.scalar-card--transparent[data-v-dd83e46e]{background:var(--scalar-background-1)}.scalar-card-header[data-v-838dc259]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-3);padding:9px 3px 9px 12px;flex-shrink:0}.scalar-card-header.scalar-card--borderless[data-v-838dc259]+.scalar-card-content{margin-top:-9px}.scalar-card-header-slots[data-v-838dc259]{display:flex;justify-content:space-between;line-height:1.35}.scalar-card-header-title[data-v-838dc259]{text-transform:uppercase;flex:1;min-width:0;text-overflow:ellipsis;overflow:hidden}.scalar-card-header-actions[data-v-838dc259]{display:flex}.endpoints[data-v-cfef28a1]{overflow:auto;background:var(--scalar-background-2);padding:10px 12px}@media (max-width: 580px){.endpoints[data-v-cfef28a1]{max-height:calc(100vh - 150px)}}.endpoints span+span[data-v-cfef28a1]{text-align:left;margin-left:12px;text-transform:initial}.endpoint[data-v-cfef28a1]{display:flex;white-space:nowrap;cursor:pointer;text-decoration:none}.endpoint:hover .endpoint-path[data-v-cfef28a1],.endpoint:focus-visible .endpoint-path[data-v-cfef28a1]{text-decoration:underline}.endpoint span[data-v-cfef28a1]:first-of-type{text-transform:uppercase}.endpoint .post[data-v-cfef28a1],.endpoint .get[data-v-cfef28a1],.endpoint .delete[data-v-cfef28a1],.endpoint .put[data-v-cfef28a1]{white-space:nowrap}.endpoint span[data-v-cfef28a1]{color:var(--scalar-color-1);min-width:62px;display:inline-block;text-align:right;line-height:1.55;font-family:var(--scalar-font-code);font-size:var(--scalar-small);cursor:pointer}.deprecated[data-v-cfef28a1]{text-decoration:line-through}.section-container[data-v-1913755a]{position:relative;padding:0 60px;width:100%;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.section-container[data-v-1913755a]:has(.introduction-section){border-top:none}@container narrow-references-container (max-width: 900px){.section-container[data-v-1913755a]{padding:0}}.section-container[data-v-d47a803a]{border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.section-container[data-v-d47a803a]:has(.show-more){background-color:color-mix(in srgb,var(--scalar-background-2),transparent)}.operation-path[data-v-ec6c8861]{overflow:hidden;word-wrap:break-word;font-weight:var(--scalar-semibold);line-break:anywhere}.deprecated[data-v-ec6c8861]{text-decoration:line-through}.content-type[data-v-4b712394]{display:flex;align-items:center;font-size:var(--scalar-heading-4);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);line-height:1.45;margin-top:24px;padding-bottom:12px;border-bottom:1px solid var(--scalar-border-color);flex-flow:wrap}.content-type-select[data-v-4b712394]{position:relative;padding-left:9px;height:-moz-fit-content;height:fit-content;color:var(--scalar-color-2);font-size:var(--scalar-font-size-3);display:flex;align-items:center}.content-type-no-select.content-type-select[data-v-4b712394]{pointer-events:none}.content-type-no-select.content-type-select[data-v-4b712394]:after{display:none}.content-type-select span[data-v-4b712394]{display:flex;align-items:center}.content-type-select[data-v-4b712394]:after{content:"";width:7px;height:7px;transform:rotate(45deg) translate3d(-2px,-4px,0);display:block;margin-left:7px;box-shadow:1px 1px 0 currentColor}.content-type-select select[data-v-4b712394]{border:none;outline:none;cursor:pointer;background:var(--scalar-background-3);box-shadow:-2px 0 0 0 var(--scalar-background-3);position:absolute;top:0;left:0;width:100%;height:100%;opacity:0;-webkit-appearance:none;-moz-appearance:none;appearance:none}.content-type-select[data-v-4b712394]:hover{color:var(--scalar-color-1)}@media (max-width: 460px){.content-type-select[data-v-4b712394]{margin-left:auto;padding-right:3px}}.schema-type-icon[data-v-2a52eb0b]{color:var(--scalar-color-1);display:none}.schema-type[data-v-2a52eb0b]{font-family:var(--scalar-font-code);color:var(--scalar-color-1)}.error[data-v-f48d0197]{background-color:var(--scalar-color-red)}.schema-card[data-v-f48d0197]{z-index:0;position:relative;font-size:var(--scalar-font-size-4);color:var(--scalar-color-1)}.schema-card-title[data-v-f48d0197]{height:var(--schema-title-height);padding:6px 10px;display:flex;align-items:center;gap:4px;color:var(--scalar-color-2);font-weight:var(--scalar-semibold);font-size:var(--scalar-micro);border-bottom:var(--scalar-border-width) solid transparent}button.schema-card-title[data-v-f48d0197]{cursor:pointer}button.schema-card-title[data-v-f48d0197]:hover{color:var(--scalar-color-1)}.schema-card-title-icon--open[data-v-f48d0197]{transform:rotate(45deg)}.schema-properties-open>.schema-card-title[data-v-f48d0197]{border-bottom-left-radius:0;border-bottom-right-radius:0;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.schema-properties-open>.schema-properties[data-v-f48d0197]{width:-moz-fit-content;width:fit-content}.schema-card-description+.schema-properties[data-v-f48d0197]{width:-moz-fit-content;width:fit-content;margin-top:12px}.schema-properties-open.schema-properties[data-v-f48d0197],.schema-properties-open>.schema-card--open[data-v-f48d0197]{width:100%}.schema-card .property[data-v-f48d0197]:last-of-type{padding-bottom:10px}.schema-properties[data-v-f48d0197]{display:flex;flex-direction:column;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);width:-moz-fit-content;width:fit-content}.schema-properties .schema-properties[data-v-f48d0197]{border-radius:13.5px}.schema-properties .schema-properties.schema-properties-open[data-v-f48d0197]{border-radius:13.5px 13.5px 9px 9px}.schema-properties-open[data-v-f48d0197]{width:100%}.schema-card--compact[data-v-f48d0197]{align-self:flex-start}.schema-card--compact.schema-card--open[data-v-f48d0197]{align-self:initial}.schema-card-title--compact[data-v-f48d0197]{color:var(--scalar-color-2);padding:6px 8px;height:auto;border-bottom:none}.schema-card--compact>.schema-properties[data-v-f48d0197],.schema-card-title--compact[data-v-f48d0197]{border-radius:13.5px}.schema-card-title--compact>.schema-card-title-icon[data-v-f48d0197]{margin:0}.schema-card-title--compact>.schema-card-title-icon--open[data-v-f48d0197]{transform:rotate(45deg)}.schema-properties-open>.schema-card-title--compact[data-v-f48d0197]{position:static}.schema-card--compact.schema-card--level-0>.schema-properties[data-v-f48d0197]{border:none}[data-v-f48d0197] .schema-card-description p{font-size:var(--scalar-mini, var(--scalar-paragraph));color:var(--scalar-color-2);line-height:1.5;display:block;margin-bottom:6px}.schema-card-description[data-v-f48d0197]:first-of-type{padding-top:10px}.children .schema-card-description[data-v-f48d0197]:first-of-type{padding-top:0}.models-list-item .schema-properties[data-v-f48d0197]{margin-bottom:10px}.property-detail[data-v-3ae98a1b]{display:inline-flex}.property-detail+.property-detail[data-v-3ae98a1b]:before{display:block;content:"·";margin:0 .5ch}.property-detail-truncate[data-v-3ae98a1b]{overflow:hidden}.property-detail-truncate>.property-detail-value[data-v-3ae98a1b]{overflow:hidden;text-overflow:ellipsis}.property-detail-prefix[data-v-3ae98a1b]{color:var(--scalar-color-2)}code.property-detail-value[data-v-3ae98a1b]{font-family:var(--scalar-font-code);font-size:var(--scalar-font-size-3);color:var(--scalar-color-2);background:var(--scalar-background-3);padding:0 4px;border:.5px solid var(--scalar-border-color);border-radius:var(--scalar-radius)}.badge[data-v-190574ac]{color:var(--scalar-color-2);font-size:var(--scalar-micro);background:var(--scalar-background-2);padding:2px 6px;border-radius:12px;font-weight:var(--scalar-semibold);display:inline-block;text-transform:uppercase}.property-heading[data-v-dd95f4d7]{display:flex;flex-wrap:wrap;align-items:baseline;row-gap:9px;white-space:nowrap}.property-heading[data-v-dd95f4d7]>*{margin-right:9px}.property-heading[data-v-dd95f4d7]:last-child{margin-right:0}.property-heading>.property-detail[data-v-dd95f4d7]:not(:last-of-type){margin-right:0}.property-name[data-v-dd95f4d7]{font-family:var(--scalar-font-code);font-weight:var(--scalar-semibold);font-size:var(--scalar-font-size-3);display:flex}.property-additional[data-v-dd95f4d7]{font-family:var(--scalar-font-code)}.property-required[data-v-dd95f4d7],.property-optional[data-v-dd95f4d7]{color:var(--scalar-color-2)}.property-required[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-orange)}.property-read-only[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-blue)}.property-write-only[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-green)}.property-detail[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-2);display:flex;align-items:center;min-width:0}.property-const[data-v-dd95f4d7]{color:var(--scalar-color-1)}.property[data-v-f5c211dc]{color:var(--scalar-color-1);padding:10px;font-size:var(--scalar-mini)}.property[data-v-f5c211dc]:last-of-type{padding-bottom:0}.property--compact.property--level-0[data-v-f5c211dc]{padding:12px 0}.property--deprecated[data-v-f5c211dc]{background:repeating-linear-gradient(-45deg,var(--scalar-background-2) 0,var(--scalar-background-2) 2px,transparent 2px,transparent 5px);background-size:100%}.property--deprecated[data-v-f5c211dc]>*{opacity:.75}.property-description[data-v-f5c211dc]{margin-top:6px;line-height:1.4;font-size:var(--scalar-small)}.property-heading:empty+.property-description[data-v-f5c211dc]:last-of-type,.property-description[data-v-f5c211dc]:first-of-type:last-of-type{margin-top:0}[data-v-f5c211dc] .property-description *{color:var(--scalar-color-2)!important}.property[data-v-f5c211dc]:not(:last-of-type){border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.children[data-v-f5c211dc]{display:flex;flex-direction:column;padding-top:8px}.property-example[data-v-f5c211dc]{display:flex;flex-direction:column;margin-top:6px;max-height:calc(((var(--full-height) - var(--refs-header-height))) / 2);font-size:var(--scalar-micro);border:var(--scalar-border-width) solid var(--scalar-border-color);background:var(--scalar-background-2);border-radius:var(--scalar-radius-lg)}.property-example-label[data-v-f5c211dc]{font-weight:var(--scalar-semibold);color:var(--scalar-color-3);padding:6px}.property-example-value[data-v-f5c211dc]{all:unset;font-family:var(--scalar-font-code);padding:6px;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.property-rule[data-v-f5c211dc]{display:flex;flex-direction:column;gap:6px;margin-top:12px;border-radius:var(--scalar-radius-lg)}.property-enum-value[data-v-f5c211dc]{padding:3px 0;color:var(--scalar-color-2);line-height:1.5;word-break:break-word}.property-enum-value[data-v-f5c211dc]:before{content:"⊢";margin-right:6px;color:var(--scalar-color-3)}.property-enum-values[data-v-f5c211dc]{margin-top:8px;list-style:none}.property--compact .property-example[data-v-f5c211dc]{background:transparent;border:none;display:flex;flex-direction:row;gap:8px}.property--compact .property-example-label[data-v-f5c211dc],.property--compact .property-example-value[data-v-f5c211dc]{padding:3px 0 0}.property--compact .property-example-value[data-v-f5c211dc]{background:var(--scalar-background-2);border-top:0;border-radius:var(--scalar-radius);padding:3px 4px}.property-list[data-v-f5c211dc]{border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:var(--scalar-radius);margin-top:10px}.property-list .property[data-v-f5c211dc]:last-of-type{padding-bottom:10px}.property-name[data-v-f5c211dc]{font-family:var(--scalar-font-code);font-weight:var(--scalar-semibold)}.enum-toggle-button[data-v-f5c211dc]{align-items:center;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:13.5px;cursor:pointer;color:var(--scalar-color-2);display:flex;font-weight:var(--scalar-semibold);gap:4px;margin-top:8px;padding:6px 10px;-webkit-user-select:none;-moz-user-select:none;user-select:none;white-space:nowrap}.enum-toggle-button[data-v-f5c211dc]:hover{color:var(--scalar-color-1)}.enum-toggle-button-icon--open[data-v-f5c211dc]{transform:rotate(45deg)}.headers-card[data-v-2f99eae6]{z-index:0;margin-top:12px;margin-bottom:6px;position:relative;font-size:var(--scalar-font-size-4);color:var(--scalar-color-1);align-self:flex-start}.headers-card.headers-card--open[data-v-2f99eae6]{align-self:initial}.headers-card-title[data-v-2f99eae6]{padding:6px 10px;display:flex;align-items:center;gap:4px;color:var(--scalar-color-3);font-weight:var(--scalar-semibold);font-size:var(--scalar-micro);border-radius:13.5px}button.headers-card-title[data-v-2f99eae6]{cursor:pointer}button.headers-card-title[data-v-2f99eae6]:hover{color:var(--scalar-color-1)}.headers-card-title-icon--open[data-v-2f99eae6]{transform:rotate(45deg)}.headers-properties-open>.headers-card-title[data-v-2f99eae6]{border-bottom-left-radius:0;border-bottom-right-radius:0;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.headers-properties-open[data-v-2f99eae6]{width:100%}.headers-properties[data-v-2f99eae6]{display:flex;flex-direction:column;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:13.5px;width:-moz-fit-content;width:fit-content}.headers-card .property[data-v-2f99eae6]:last-of-type{padding-bottom:10px}.headers-card-title>.headers-card-title-icon[data-v-2f99eae6]{width:10px;height:10px;margin:0}.headers-card-title>.headers-card-title-icon--open[data-v-2f99eae6]{transform:rotate(45deg)}.parameter-item[data-v-e1efed5d]{display:flex;flex-direction:column;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.parameter-item:last-of-type .parameter-schema[data-v-e1efed5d]{padding-bottom:0}.parameter-item-container[data-v-e1efed5d]{padding:0}.parameter-item-headers[data-v-e1efed5d]{border:var(--scalar-border-width) solid var(--scalar-border-color)}.parameter-item-name[data-v-e1efed5d]{margin-right:6px;font-weight:var(--scalar-semibold);font-size:var(--scalar-font-size-3);font-family:var(--scalar-font-code);color:var(--scalar-color-1)}.parameter-item-type[data-v-e1efed5d]{font-size:var(--scalar-micro);color:var(--scalar-color-2);margin-right:6px;line-height:1.4;white-space:nowrap;text-overflow:ellipsis;width:100%;overflow:hidden}.parameter-item-trigger-open .parameter-item-type[data-v-e1efed5d]{white-space:normal}.parameter-item-trigger+.parameter-item-container[data-v-e1efed5d] .property--level-0>.property-heading .property-detail-value{font-size:var(--scalar-font-size-3)}.parameter-item-required-optional[data-v-e1efed5d]{color:var(--scalar-color-2);font-weight:var(--scalar-semibold);margin-right:6px;position:relative}.parameter-item--required[data-v-e1efed5d]{text-transform:uppercase;font-size:var(--scalar-micro);font-weight:var(--scalar-semibold);color:var(--scalar-color-orange)}.parameter-item-description[data-v-e1efed5d]{margin-top:3px!important;font-size:var(--scalar-small);color:var(--scalar-color-2);line-height:1.4}.parameter-item-description[data-v-e1efed5d] p{margin-top:4px;font-size:var(--scalar-small);color:var(--scalar-color-2);line-height:1.4}.parameter-schema[data-v-e1efed5d]{padding-bottom:9px;margin-top:3px}.parameter-item-trigger[data-v-e1efed5d]{padding:12px 0;cursor:pointer;outline:none;text-align:left;position:relative;align-items:baseline}.parameter-item-trigger-open[data-v-e1efed5d]{padding-bottom:0}.parameter-item-trigger[data-v-e1efed5d]:after{content:"";position:absolute;height:10px;width:100%;bottom:0}.parameter-item-icon[data-v-e1efed5d]{color:var(--scalar-color-3);height:18px;left:-19px;position:absolute;top:11px;width:18px}.parameter-item-trigger:hover .parameter-item-icon[data-v-e1efed5d],.parameter-item-trigger:focus-visible .parameter-item-icon[data-v-e1efed5d]{color:var(--scalar-color-1)}.parameter-item-trigger:focus-visible .parameter-item-icon[data-v-e1efed5d]{outline:1px solid var(--scalar-color-accent);outline-offset:2px;border-radius:var(--scalar-radius)}.parameter-item-content-type[data-v-e1efed5d]{margin-left:auto;opacity:0;transition:opacity .1s ease-in-out;color:var(--scalar-color-3);font-size:var(--scalar-micro);background:var(--scalar-background-2);padding:2px 6px;border-radius:12px}.parameter-item-trigger:hover .parameter-item-content-type[data-v-e1efed5d]{opacity:1}.parameter-list[data-v-b05cf5ff]{margin-top:24px}.parameter-list-title[data-v-b05cf5ff]{font-size:var(--scalar-font-size-2);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);line-height:1.45;margin-top:12px;margin-bottom:12px}.parameter-list-items[data-v-b05cf5ff]{list-style:none;padding:0;margin:0 0 12px;font-size:var(--scalar-small)}.request-body-title[data-v-dc8f9923]{display:flex;align-items:center;font-size:var(--scalar-font-size-2);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);line-height:1.45;margin-top:24px;padding-bottom:12px;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color);flex-flow:wrap}.request-body-title-select[data-v-dc8f9923]{position:relative;height:-moz-fit-content;height:fit-content;margin-left:auto;font-weight:var(--scalar-regular);display:flex;align-items:center;color:var(--scalar-color-3);font-size:var(--scalar-micro);background:var(--scalar-background-2);padding:2px 6px;border-radius:12px;border:var(--scalar-border-width) solid var(--scalar-border-color)}.request-body-title-no-select.request-body-title-select[data-v-dc8f9923]{pointer-events:none}.request-body-title-no-select[data-v-dc8f9923]{border:none}.request-body-title-no-select.request-body-title-select[data-v-dc8f9923]:after{display:none}.request-body-title-select span[data-v-dc8f9923]{display:flex;align-items:center}.request-body-title-select[data-v-dc8f9923]:after{content:"";width:6px;height:6px;transform:rotate(45deg) translate3d(0,-3px,0);display:block;margin-left:6px;box-shadow:1px 1px 0 currentColor;margin-right:5px}.request-body-title-select select[data-v-dc8f9923]{border:none;outline:none;cursor:pointer;background:var(--scalar-background-3);position:absolute;top:0;left:0;width:100%;height:100%;opacity:0;-webkit-appearance:none;-moz-appearance:none;appearance:none}.request-body-title-select[data-v-dc8f9923]:hover{color:var(--scalar-color-1)}.request-body-description[data-v-dc8f9923]{margin-top:6px;font-size:var(--scalar-small);width:100%}.request-body-description[data-v-dc8f9923] .markdown *{color:var(--scalar-color-2)!important}@media (max-width: 460px){.request-body-title-select[data-v-dc8f9923]{margin-left:auto;padding-right:3px}}.section-wrapper[data-v-07cabcb0]{color:var(--scalar-color-1);padding-top:12px;margin-top:-12px}.section-accordion[data-v-07cabcb0]{display:flex;flex-direction:column;border-radius:var(--scalar-radius-lg);background:var(--scalar-background-2)}.section-accordion-transparent[data-v-07cabcb0]{background:transparent;border:1px solid var(--scalar-border-color)}.section-accordion-button[data-v-07cabcb0]{padding:6px;display:flex;align-items:center;gap:6px;cursor:pointer}.section-accordion-button-content[data-v-07cabcb0]{flex:1;min-width:0}.section-accordion-button-actions[data-v-07cabcb0]{display:flex;align-items:center;gap:6px;color:var(--scalar-color-3)}.section-accordion-chevron[data-v-07cabcb0]{margin-right:4px;width:20px;cursor:pointer;opacity:1;color:var(--scalar-color-3)}.section-accordion-button:hover .section-accordion-chevron[data-v-07cabcb0]{color:var(--scalar-color-1)}.section-accordion-content[data-v-07cabcb0]{border-top:1px solid var(--scalar-border-color);display:flex;flex-direction:column}.section-accordion-description[data-v-07cabcb0]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color--1);padding:10px 12px 0}.section-accordion-content-card[data-v-07cabcb0] .property:last-of-type{padding-bottom:9px}.show-api-client-button[data-v-ea8189a2]{-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none;padding:4px 6px;white-space:nowrap;border-radius:var(--scalar-radius);display:flex;justify-content:center;align-items:center;font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-background-2);font-family:var(--scalar-font);background:var(--scalar-button-1);position:relative;cursor:pointer;box-sizing:border-box;box-shadow:inset 0 0 0 1px #0000001a;outline-offset:2px}.show-api-client-button span[data-v-ea8189a2],.show-api-client-button svg[data-v-ea8189a2]{fill:currentColor;color:var(--scalar-button-1-color);z-index:1}.show-api-client-button[data-v-ea8189a2]:hover{background:var(--scalar-button-1-hover)}.show-api-client-button svg[data-v-ea8189a2]{margin-right:4px}.empty-state[data-v-4b51e8ca]{margin:10px 0 10px 12px;text-align:center;font-size:var(--scalar-micro);min-height:56px;display:flex;align-items:center;justify-content:center;border-radius:var(--scalar-radius-lg);color:var(--scalar-color-2)}.rule-title[data-v-4b51e8ca]{font-family:var(--scalar-font-code);color:var(--scalar-color-1);display:inline-block;margin:12px 0 6px;border-radius:var(--scalar-radius)}.rule[data-v-4b51e8ca]{margin:0 12px;border-radius:var(--scalar-radius-lg)}.rule-items[data-v-4b51e8ca]{counter-reset:list-number;display:flex;flex-direction:column;gap:12px;border-left:1px solid var(--scalar-border-color);padding:12px 0}.rule-item[data-v-4b51e8ca]{counter-increment:list-number;border:1px solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);overflow:hidden;margin-left:24px}.rule-item[data-v-4b51e8ca]:before{border:1px solid var(--scalar-border-color);border-top:0;border-right:0;content:" ";display:block;width:24px;height:6px;border-radius:0 0 0 var(--scalar-radius-lg);margin-top:6px;color:var(--scalar-color-2);transform:translate(-25px);color:var(--scalar-color-1);position:absolute}.tab-list[data-v-03fdb072]{display:flex;gap:6px;position:relative;flex:1;padding:9px 12px;overflow:auto}.scalar-card-header.scalar-card-header-tabs[data-v-03fdb072]{padding:0}.tab[data-v-d0e07e65]{background:none;border:none;font-size:var(--scalar-mini);font-family:var(--scalar-font);color:var(--scalar-color-2);font-weight:var(--scalar-semibold);line-height:calc(var(--scalar-mini) + 2px);white-space:nowrap;cursor:pointer;padding:0;margin-right:3px;text-transform:uppercase;position:relative;line-height:1.35}.tab[data-v-d0e07e65]:before{content:"";position:absolute;z-index:0;left:-6px;top:-6px;width:calc(100% + 12px);height:calc(100% + 12px);border-radius:var(--scalar-radius);background:var(--scalar-background-3);opacity:0}.tab[data-v-d0e07e65]:hover:before,.tab[data-v-d0e07e65]:focus-visible:before{opacity:1}.tab[data-v-d0e07e65]:focus-visible:before{outline:1px solid var(--scalar-color-accent)}.tab span[data-v-d0e07e65]{z-index:1;position:relative}.tab-selected[data-v-d0e07e65]{color:var(--scalar-color-1);text-decoration:underline;text-underline-offset:var(--scalar-micro)}.scalar-card-footer[data-v-7621d1be]{flex-shrink:0}.text-select{position:relative;height:-moz-fit-content;height:fit-content}.text-select:has(:focus-visible) .text-select-label{outline:1px solid var(--scalar-color-accent);outline-offset:4px;border-radius:calc(var(--scalar-radius) / 2)}.text-select--single-option{pointer-events:none}.text-select select{border:none;outline:none;cursor:pointer;background:var(--scalar-background-3);box-shadow:-2px 0 0 0 var(--scalar-background-3);position:absolute;top:0;left:0;width:100%;height:100%;opacity:0;-moz-appearance:none;-webkit-appearance:none;appearance:none}.text-select span{font-size:var(--scalar-mini);color:var(--scalar-color-2);font-weight:var(--scalar-semibold);white-space:nowrap;display:flex;align-items:center;justify-content:center}.text-select:hover span{color:var(--scalar-color-1)}.text-select:not(.text-select--single-option) span:after{content:"";width:7px;height:7px;transform:rotate(45deg) translate3d(-2px,-2px,0);display:block;margin-left:6px;box-shadow:1px 1px 0 currentColor}.text-select span:hover{background:var(--scalar-background-2)}.example-selector[data-v-a7e22e14]{padding:4px}.markdown[data-v-d84b9039] *{margin:0}.code-copy[data-v-d84b9039]{display:flex;align-items:center;justify-content:center;-moz-appearance:none;appearance:none;-webkit-appearance:none;outline:none;background:transparent;cursor:pointer;color:var(--scalar-color-3);border:none;padding:0;margin-right:12px}.code-copy[data-v-d84b9039]:hover{color:var(--scalar-color-1)}.code-copy svg[data-v-d84b9039]{width:13px;height:13px}.response-card-footer[data-v-d84b9039]{display:flex;flex-direction:column;flex-shrink:0;padding:10px 12px;gap:8px;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.response-example-selector[data-v-d84b9039]{align-self:flex-start;margin:-4px}.response-description[data-v-d84b9039]{font-weight:var(--scalar-semibold);font-size:var(--scalar-micro);color:var(--scalar-color--1);display:flex;align-items:center;box-sizing:border-box}.schema-type[data-v-d84b9039]{font-size:var(--scalar-micro);color:var(--scalar-color-2);font-weight:var(--scalar-semibold);background:var(--scalar-background-3);padding:2px 4px;border-radius:4px;margin-right:4px}.schema-example[data-v-d84b9039]{font-size:var(--scalar-micro);color:var(--scalar-color-2);font-weight:var(--scalar-semibold)}.example-response-tab[data-v-d84b9039]{display:block;margin:6px}.scalar-card-container[data-v-d84b9039]{flex:1;background:var(--scalar-background-2);display:grid}.scalar-card-container[data-v-d84b9039] .cm-scroller{overflow-y:hidden}.scalar-card-checkbox[data-v-d84b9039]{display:flex;align-items:center;justify-content:center;position:relative;min-height:17px;cursor:pointer;-webkit-user-select:none;-moz-user-select:none;user-select:none;font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-2);width:-moz-fit-content;width:fit-content;white-space:nowrap;margin-right:9px;gap:6px}.scalar-card-checkbox:has(.scalar-card-checkbox-input:focus-visible) .scalar-card-checkbox-checkmark[data-v-d84b9039]{outline:1px solid var(--scalar-color-accent)}.scalar-card-checkbox[data-v-d84b9039]:hover{color:var(--scalar-color--1)}.scalar-card-checkbox .scalar-card-checkbox-input[data-v-d84b9039]{position:absolute;opacity:0;cursor:pointer;height:0;width:0}.scalar-card-checkbox-checkmark[data-v-d84b9039]{height:17px;width:17px;border-radius:var(--scalar-radius);background-color:transparent;background-color:var(--scalar-background-3);box-shadow:inset 0 0 0 1px var(--scalar-border-color)}.scalar-card-checkbox[data-v-d84b9039]:has(.scalar-card-checkbox-input:checked){color:var(--scalar-color-1)}.scalar-card-checkbox .scalar-card-checkbox-input:checked~.scalar-card-checkbox-checkmark[data-v-d84b9039]{background-color:var(--scalar-button-1);box-shadow:none}.scalar-card-checkbox-checkmark[data-v-d84b9039]:after{content:"";position:absolute;display:none}.scalar-card-checkbox .scalar-card-checkbox-input:checked~.scalar-card-checkbox-checkmark[data-v-d84b9039]:after{display:block}.scalar-card-checkbox .scalar-card-checkbox-checkmark[data-v-d84b9039]:after{right:6px;top:36.5%;width:5px;height:9px;border:solid 1px var(--scalar-button-1-color);border-width:0 1.5px 1.5px 0;transform:rotate(45deg)}.request[data-v-3ff33ef0]{display:flex;flex-wrap:nowrap}.request-header[data-v-3ff33ef0]{display:flex;gap:6px;text-transform:initial}.request-method[data-v-3ff33ef0]{font-family:var(--scalar-font-code);text-transform:uppercase}.request-client-picker[data-v-3ff33ef0]{padding-left:12px;padding-right:9px}.request-card-footer[data-v-3ff33ef0]{display:flex;justify-content:flex-end;padding:6px;flex-shrink:0}.request-card-footer-addon[data-v-3ff33ef0]{display:flex;align-items:center;flex:1;min-width:0}.request-editor-section[data-v-3ff33ef0]{display:flex;flex:1}.request-card-simple[data-v-3ff33ef0]{display:flex;align-items:center;justify-content:space-between;padding:8px 8px 8px 12px;font-size:var(--scalar-small)}.code-snippet[data-v-3ff33ef0]{display:flex;flex-direction:column;width:100%}.operation-title[data-v-e104650f]{display:flex;justify-content:space-between}.operation-details[data-v-e104650f]{display:flex;align-items:center;margin-top:0;gap:9px;min-width:0;flex-shrink:1}.endpoint-type[data-v-e104650f]{display:flex;align-items:center;justify-content:center;gap:6px;position:relative;z-index:0;width:60px;padding:6px;flex-shrink:0;font-size:var(--scalar-small);text-transform:uppercase;font-weight:var(--scalar-bold);font-family:var(--scalar-font)}.endpoint-type[data-v-e104650f]:after{content:"";position:absolute;top:0;right:0;bottom:0;left:0;z-index:-1;background:currentColor;opacity:.15;border-radius:var(--scalar-radius-lg)}.endpoint-anchor[data-v-e104650f]{display:flex;align-items:center;min-width:0;flex-shrink:1;font-size:20px}.endpoint-anchor.label[data-v-e104650f]{display:flex}.endpoint-label[data-v-e104650f]{display:flex;align-items:baseline;gap:9px;min-width:0;flex-shrink:1;color:var(--scalar-color-1)}.endpoint-label-path[data-v-e104650f]{font-family:var(--scalar-font-code);font-size:var(--scalar-mini);text-overflow:ellipsis;white-space:nowrap;overflow:hidden}.endpoint-label-path[data-v-e104650f] em{color:var(--scalar-color-2)}.endpoint-label-name[data-v-e104650f]{color:var(--scalar-color-2);font-size:var(--scalar-small);flex-shrink:1000000000;text-overflow:ellipsis;white-space:nowrap;overflow:hidden}.endpoint-try-hint[data-v-e104650f]{padding:2px;height:24px;width:24px;flex-shrink:0}.endpoint-copy[data-v-e104650f]{color:currentColor;padding:2px}.endpoint-copy[data-v-e104650f] svg{stroke-width:2px}.endpoint-content[data-v-e104650f]{display:grid;grid-auto-columns:1fr;grid-auto-flow:row;gap:9px;padding:9px}@media (min-width: 1000px){.endpoint-content[data-v-e104650f]{grid-auto-flow:column}}@container (max-width: 900px){.endpoint-content[data-v-e104650f]{grid-template-columns:1fr}}.endpoint-content[data-v-e104650f]>*{max-height:unset}.operation-details-card[data-v-e104650f]{display:flex;flex-direction:column;gap:12px}.operation-details-card-item[data-v-e104650f] .parameter-list{border:1px solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);margin-top:0}.operation-details-card-item[data-v-e104650f]{display:flex;flex-direction:column;gap:12px}.operation-details-card-item[data-v-e104650f] .parameter-list-items{margin-bottom:0}.operation-details-card[data-v-e104650f] .parameter-item:last-of-type .parameter-schema{padding-bottom:12px}.operation-details-card[data-v-e104650f] .parameter-list .parameter-list{margin-bottom:12px}.operation-details-card[data-v-e104650f] .parameter-item{margin:0;padding:0 9px}.operation-details-card[data-v-e104650f] .property{padding:9px;margin:0}.operation-details-card[data-v-e104650f] .parameter-list-title,.operation-details-card[data-v-e104650f] .request-body-title{text-transform:uppercase;font-weight:var(--scalar-bold);font-size:var(--scalar-mini);color:var(--scalar-color-2);line-height:1.33;padding:9px;margin:0}.operation-details-card[data-v-e104650f] .request-body-title-select{text-transform:initial;font-weight:initial;margin-left:auto}.examples[data-v-0916dd9f]{position:sticky;top:calc(var(--refs-header-height) + 24px)}.deprecated[data-v-0916dd9f] *{text-decoration:line-through}.example-path[data-v-0916dd9f]{color:var(--scalar-color-2);font-family:var(--scalar-font-code)}.example-path[data-v-0916dd9f] em{color:var(--scalar-color-1);font-style:normal}.references-loading{position:absolute;top:0;left:0;right:0;z-index:1;grid-area:rendered;background:var(--scalar-background-1)}.references-loading-top-spacer{top:-1px}.references-loading-hidden-tag .section-container>.section:first-child{display:none}.download[data-v-1fc94525]{margin-bottom:24px}.download-button[data-v-1fc94525]{color:var(--scalar-link-color, var(--scalar-color-accent));font-weight:var(--scalar-link-font-weight, inherit);-webkit-text-decoration:var(--scalar-text-decoration)!important;text-decoration:var(--scalar-text-decoration)!important;text-decoration-color:var(--scalar-text-decoration-color);font-size:var(--scalar-paragraph);cursor:pointer}.download-button[data-v-1fc94525]:hover{color:var(--scalar-link-color-hover, var(--scalar-color-accent));-webkit-text-decoration:var(--scalar-text-decoration-hover)!important;text-decoration:var(--scalar-text-decoration-hover)!important;text-decoration-color:var(--scalar-text-decoration-color-hover)}.introduction-description-heading[data-v-fa3d93c5]{scroll-margin-top:64px;margin-top:1em;margin-bottom:.5em}.markdown+.markdown[data-v-fa3d93c5]{margin-top:1em}.introduction-description[data-v-fa3d93c5]{display:flex;flex-direction:column}.references-classic .introduction-description[data-v-fa3d93c5] img{max-width:720px}.sticky-cards[data-v-0336cd4f]{display:flex;flex-direction:column;position:sticky;top:calc(var(--refs-header-height) + 24px)}.client-libraries-content[data-v-b29373e8]{container:client-libraries-content / inline-size;display:flex;justify-content:center;overflow:hidden;padding:0 12px;background-color:var(--scalar-background-1);border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.client-libraries[data-v-b29373e8]{display:flex;align-items:center;justify-content:center;width:100%;position:relative;cursor:pointer;white-space:nowrap;padding:8px 2px;gap:6px;color:var(--scalar-color-3);border-bottom:1px solid transparent;-webkit-user-select:none;-moz-user-select:none;user-select:none}.client-libraries[data-v-b29373e8]:first-child{border-radius:var(--scalar-radius) 0 0 0}.client-libraries[data-v-b29373e8]:not(.client-libraries__active):hover:before{content:"";position:absolute;width:calc(100% - 4px);height:calc(100% - 4px);background:var(--scalar-background-2);left:2px;top:2px;z-index:0;border-radius:var(--scalar-radius)}.client-libraries[data-v-b29373e8]:active{color:var(--scalar-color-1)}.client-libraries[data-v-b29373e8]:focus-visible{outline:none;box-shadow:inset 0 0 0 1px var(--scalar-color-accent)}@media screen and (max-width: 450px){.client-libraries[data-v-b29373e8]:nth-of-type(4),.client-libraries[data-v-b29373e8]:nth-of-type(5){display:none}}.client-libraries-icon[data-v-b29373e8]{max-width:14px;max-height:14px;min-width:14px;width:100%;aspect-ratio:1;display:flex;align-items:center;justify-content:center;position:relative;box-sizing:border-box;color:currentColor}.client-libraries-icon__more svg[data-v-b29373e8]{height:initial}@container client-libraries-content (width < 400px){.client-libraries__select[data-v-b29373e8]{width:-moz-fit-content;width:fit-content}.client-libraries__select .client-libraries-icon__more+span[data-v-b29373e8]{display:none}}@container client-libraries-content (width < 380px){.client-libraries[data-v-b29373e8]{width:100%}.client-libraries span[data-v-b29373e8]{display:none}}.client-libraries__active[data-v-b29373e8]{color:var(--scalar-color-1);border-bottom:1px solid var(--scalar-color-1)}@keyframes codeloader-b29373e8{0%{transform:rotate(0)}to{transform:rotate(1turn)}}.client-libraries .client-libraries-text[data-v-b29373e8]:last-of-type{font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);position:relative;display:flex;align-items:center}.client-libraries__active .client-libraries-text[data-v-b29373e8]{color:var(--scalar-color-1)}.client-libraries__select select[data-v-b29373e8]{background:var(--scalar-background-3);color:var(--scalar-color-2);opacity:0;height:100%;width:100%;aspect-ratio:1;position:absolute;top:0;left:0;cursor:pointer;z-index:1;-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none}.client-libraries__select[data-v-b29373e8]:has(select:focus-visible){border-radius:var(--scalar-radius);box-shadow:inset 0 0 0 1px var(--scalar-color-accent)}@media screen and (max-width: 600px){.references-classic .client-libraries[data-v-b29373e8]{flex-direction:column}}.selected-client[data-v-35fd541a]{color:var(--scalar-color-1);font-size:var(--scalar-mini);font-family:var(--scalar-font-code);padding:9px 12px;border-top:none;white-space:nowrap;overflow:hidden;text-overflow:ellipsis;background:var(--scalar-background-1);border-top:var(--scalar-border-width) solid var(--scalar-border-color);min-height:-moz-fit-content;min-height:fit-content}.client-libraries-heading[data-v-35fd541a]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-1);padding:9px 12px;background-color:var(--scalar-background-2);display:flex;align-items:center;max-height:32px}.section-accordion-wrapper[data-v-1a30b4ff]{padding:0 60px}.section-accordion[data-v-1a30b4ff]{position:relative;width:100%;max-width:var(--refs-content-max-width);margin:auto}.section-accordion-content[data-v-1a30b4ff]{display:flex;flex-direction:column;gap:12px;padding-top:12px}.section-accordion-button[data-v-1a30b4ff]{width:100%;display:flex;cursor:pointer;padding:6px 0;margin:-6px 0;border-radius:var(--scalar-radius)}.section-accordion-chevron[data-v-1a30b4ff]{width:24px;position:absolute;left:-24px;height:auto;top:10px;color:var(--scalar-color-3)}.section-accordion-button:hover .section-accordion-chevron[data-v-1a30b4ff]{color:var(--scalar-color-1)}.section-accordion-title[data-v-1a30b4ff]{display:flex;flex-direction:column;align-items:flex-start;flex:1;padding:0 6px}.section-accordion-title[data-v-1a30b4ff] .section-header-wrapper{grid-template-columns:1fr}.section-accordion-title[data-v-1a30b4ff] .section-header{margin-bottom:0}@container narrow-references-container (max-width: 900px){.section-accordion-chevron[data-v-1a30b4ff]{width:16px;left:-16px;top:14px}.section-accordion-wrapper[data-v-1a30b4ff]{padding:calc(var(--refs-header-height)) 24px 0 24px}}.tag-section[data-v-43e1c813]{margin-bottom:48px}.tag-name[data-v-43e1c813]{text-transform:capitalize}.tag-description[data-v-43e1c813]{padding-bottom:4px;text-align:left}.webhook-request-body[data-v-cbaa1001]{margin-top:-18px}.collapsible-section[data-v-b32ceff4]{border-top:var(--scalar-border-width) solid var(--scalar-border-color);position:relative}.collapsible-section-header[data-v-b32ceff4]{color:var(--scalar-color-1)}.collapsible-section .collapsible-section-trigger[data-v-b32ceff4]{display:flex;align-items:center;cursor:pointer;padding:10px 0;font-size:var(--scalar-font-size-3);z-index:1;position:relative}.collapsible-section .collapsible-section-trigger[data-v-b32ceff4]:after{content:"";height:10px;width:100%;position:absolute;bottom:0}.collapsible-section-trigger svg[data-v-b32ceff4]{color:var(--scalar-color-3);position:absolute;left:-19px}.collapsible-section:hover .collapsible-section-trigger svg[data-v-b32ceff4]{color:var(--scalar-color-1)}.collapsible-section .collapsible-section-trigger[data-v-b32ceff4] .anchor-copy{line-height:18.5px}.collapsible-section-content[data-v-b32ceff4]{padding:0;margin:0;scroll-margin-top:140px}.webhooks-list[data-v-8cee6fce]{display:contents}.webhooks-list-truncated .webhooks-list-item[data-v-8cee6fce]:last-child{border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.webhooks-list-item[aria-expanded=true][data-v-8cee6fce]{padding-bottom:48px}.webhook-section-description.webhook-section-description[data-v-8cee6fce]{color:var(--scalar-color-2);font-size:var(--scalar-font-size-3);padding-bottom:12px}.show-more-webhooks[data-v-8cee6fce]{margin-top:32px;top:0}.reference-models[data-v-4396a256]{margin-bottom:48px}.reference-models-anchor[data-v-4396a256]{display:flex;align-items:center;font-size:20px;padding-left:6px;color:var(--scalar-color-1)}.reference-models-label[data-v-4396a256]{font-size:var(--scalar-mini)}.reference-models-label[data-v-4396a256] em{font-weight:var(--scalar-bold)}.models-list[data-v-7c1cd4b4]{display:contents}.models-list-truncated .models-list-item[data-v-7c1cd4b4]:last-child{border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.show-more-models[data-v-7c1cd4b4]{margin-top:32px;top:0}.narrow-references-container{container-name:narrow-references-container;container-type:inline-size}.render-loading[data-v-42cfaf8d]{height:calc(var(--full-height) - var(--refs-header-height));display:flex;align-items:center;justify-content:center}.introduction-card[data-v-42cfaf8d]{display:flex;flex-direction:column;background:var(--scalar-background-1)}.introduction-card-item[data-v-42cfaf8d]{display:flex;overflow:hidden;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);margin-bottom:12px;flex-direction:column;justify-content:start}@container narrow-references-container (max-width: 900px){.introduction-card-item[data-v-42cfaf8d]{border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}}.introduction-card-item[data-v-42cfaf8d]:has(.description) .server-form-container{border-bottom-left-radius:0;border-bottom-right-radius:0}.introduction-card-item[data-v-42cfaf8d] .request-item{border-bottom:0}.introduction-card-title[data-v-42cfaf8d]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-3)}.introduction-card-row[data-v-42cfaf8d]{gap:24px}@media (min-width: 600px){.introduction-card-row[data-v-42cfaf8d]{flex-flow:row wrap}}.introduction-card-row[data-v-42cfaf8d]>*{flex:1}@media (min-width: 600px){.introduction-card-row[data-v-42cfaf8d]>*{min-width:-moz-min-content;min-width:min-content}}@media (max-width: 600px){.introduction-card-row[data-v-42cfaf8d]>*{max-width:100%}}@container (max-width: 900px){.introduction-card-row[data-v-42cfaf8d]{flex-direction:column;align-items:stretch;gap:0px}}.introduction-card[data-v-42cfaf8d] .security-scheme-label{text-transform:uppercase;font-weight:var(--scalar-semibold)}.references-classic .introduction-card-row[data-v-42cfaf8d] .scalar-card:nth-of-type(2) .scalar-card-header{display:none}.references-classic .introduction-card-row[data-v-42cfaf8d] .scalar-card:nth-of-type(2) .scalar-card-header.scalar-card--borderless+.scalar-card-content{margin-top:0}.section-flare[data-v-42cfaf8d]{top:0;right:0;pointer-events:none}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}}@layer scalar-base{:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url(\'data:image/svg+xml,<svg viewBox="0 0 24 24" fill="none" xmlns="http://www.w3.org/2000/svg"><path d="m9 18 6-6-6-6" stroke="currentColor" stroke-width="1.5" stroke-linecap="round" stroke-linejoin="round"/></svg>\');mask-image:url(\'data:image/svg+xml,<svg viewBox="0 0 24 24" fill="none" xmlns="http://www.w3.org/2000/svg"><path d="m9 18 6-6-6-6" stroke="currentColor" stroke-width="1.5" stroke-linecap="round" stroke-linejoin="round"/></svg>\')}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}}@layer scalar-base{:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-api-reference{color:var(--scalar-color-1)}.scalar-api-reference .container{width:100%}@media (min-width: 400px){.scalar-api-reference .container{max-width:400px}}@media (min-width: 600px){.scalar-api-reference .container{max-width:600px}}@media (min-width: 800px){.scalar-api-reference .container{max-width:800px}}@media (min-width: 1000px){.scalar-api-reference .container{max-width:1000px}}@media (min-width: 1200px){.scalar-api-reference .container{max-width:1200px}}.scalar-api-reference .row{display:flex;flex-direction:row}.scalar-api-reference .sr-only{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);white-space:nowrap;border-width:0}.scalar-api-reference .visible{visibility:visible}.scalar-api-reference .collapse{visibility:collapse}.scalar-api-reference .static{position:static}.scalar-api-reference .fixed{position:fixed}.scalar-api-reference .absolute{position:absolute}.scalar-api-reference .relative{position:relative}.scalar-api-reference .sticky{position:sticky}.scalar-api-reference .right-0{right:0}.scalar-api-reference .block{display:block}.scalar-api-reference .flex{display:flex}.scalar-api-reference .grid{display:grid}.scalar-api-reference .contents{display:contents}.scalar-api-reference .hidden{display:none}.scalar-api-reference .h-8{height:32px}.scalar-api-reference .flex-shrink,.scalar-api-reference .shrink{flex-shrink:1}.scalar-api-reference .transform{transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-api-reference .resize{resize:both}.scalar-api-reference .flex-wrap{flex-wrap:wrap}.scalar-api-reference .content-start{align-content:flex-start}.scalar-api-reference .content-end{align-content:flex-end}.scalar-api-reference .items-center{align-items:center}.scalar-api-reference .gap-12{gap:48px}.scalar-api-reference .divide-y>:not([hidden])~:not([hidden]){--tw-divide-y-reverse:0;border-top-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-y-reverse)));border-bottom-width:calc(var(--scalar-border-width) * var(--tw-divide-y-reverse))}.scalar-api-reference .truncate{overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.scalar-api-reference .border{border-width:var(--scalar-border-width)}.scalar-api-reference .bg-b-2{background-color:var(--scalar-background-2)}.scalar-api-reference .px-3{padding-left:12px;padding-right:12px}.scalar-api-reference .py-1{padding-top:4px;padding-bottom:4px}.scalar-api-reference .py-1\\.5{padding-top:6px;padding-bottom:6px}.scalar-api-reference .py-2{padding-top:8px;padding-bottom:8px}.scalar-api-reference .py-2\\.5{padding-top:10px;padding-bottom:10px}.scalar-api-reference .font{font-family:var(--scalar-font)}.scalar-api-reference .text-sm{font-size:var(--scalar-font-size-4)}.scalar-api-reference .font{font-weight:var(--scalar-regular)}.scalar-api-reference .font-medium{font-weight:var(--scalar-semibold)}.scalar-api-reference .uppercase{text-transform:uppercase}.scalar-api-reference .lowercase{text-transform:lowercase}.scalar-api-reference .capitalize{text-transform:capitalize}.scalar-api-reference .leading{line-height:1.5}.scalar-api-reference .text-c-2{color:var(--scalar-color-2)}.scalar-api-reference .text-c-3{color:var(--scalar-color-3)}.scalar-api-reference .text-green{color:var(--scalar-color-green)}.scalar-api-reference .text-orange{color:var(--scalar-color-orange)}.scalar-api-reference .text-red{color:var(--scalar-color-red)}.scalar-api-reference .underline{text-decoration-line:underline}.scalar-api-reference .outline{outline-style:solid}.scalar-api-reference .filter{filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-api-reference .transition{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,-webkit-backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter,-webkit-backdrop-filter;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-api-reference .\\[--scalar-address-bar-height\\:0px\\]{--scalar-address-bar-height:0px}.scalar-api-reference .hover\\:text-c-1:hover{color:var(--scalar-color-1)}.nav-item[data-v-d56f12d9]{padding:0 1rem;cursor:pointer;flex:1;justify-content:center;align-items:center;display:flex;border-radius:var(--scalar-radius-lg);background:var(--scalar-background-3);border:var(--scalar-border-width) solid var(--scalar-background-2);color:var(--scalar-color-3);padding:4.5px;min-width:0;overflow:hidden;position:relative}.dark-mode .nav-item[data-v-d56f12d9]{background:color-mix(in srgb,var(--scalar-background-2),transparent)}.nav-item-icon-copy[data-v-d56f12d9]{max-width:100%;white-space:nowrap;overflow:hidden;-webkit-mask-image:linear-gradient(to left,transparent 0,var(--scalar-background-2) 20px);mask-image:linear-gradient(to left,transparent 0,var(--scalar-background-2) 20px)}.nav-item:hover .nav-item-icon-copy[data-v-d56f12d9]{-webkit-mask-image:linear-gradient(to left,transparent 20px,var(--scalar-background-2) 40px);mask-image:linear-gradient(to left,transparent 20px,var(--scalar-background-2) 40px)}.nav-item-copy[data-v-d56f12d9]{max-width:calc(100% - 20px)}.nav-item[data-v-d56f12d9]:hover{color:var(--scalar-color-1)}.nav-item__active[data-v-d56f12d9]{background-color:var(--scalar-background-1);color:var(--scalar-color-1);border-color:var(--scalar-border-color)}.dark-mode .nav-item__active[data-v-d56f12d9]{background-color:var(--scalar-background-2)}.nav-item-close[data-v-d56f12d9]{position:absolute;right:3px;padding:2px;border-radius:var(--scalar-radius);background:transparent;max-width:20px;stroke-width:1.5px;color:var(--scalar-color-3);margin-left:-20px;opacity:0}.nav-item:hover .nav-item-close[data-v-d56f12d9]{opacity:1}.nav-item-close[data-v-d56f12d9]:hover{background-color:var(--scalar-background-4)}.nav-item__active .nav-item-close[data-v-d56f12d9]:hover{background-color:var(--scalar-background-2)}.download-app-button[data-v-c896207d]{box-shadow:0 0 0 .5px var(--scalar-border-color);background:linear-gradient(#ffffffbf,#00000009)}.dark-mode .download-app-button[data-v-c896207d]{background:linear-gradient(#ffffff1a,#00000026)}.download-app-button[data-v-c896207d]:hover{background:linear-gradient(#00000009,#ffffffbf)}.dark-mode .download-app-button[data-v-c896207d]:hover{background:linear-gradient(#00000026,#ffffff1a)}.http-bg-gradient[data-v-46723816]{background:linear-gradient(#ffffffbf,#00000009)}.http-bg-gradient[data-v-46723816]:hover{background:linear-gradient(#00000009,#ffffffbf)}.dark-mode .http-bg-gradient[data-v-46723816]{background:linear-gradient(#ffffff09,#00000026)}.dark-mode .http-bg-gradient[data-v-46723816]:hover{background:linear-gradient(#00000026,#ffffff09)}.commandmenu[data-v-e9fd9234]{box-shadow:var(--scalar-shadow-2);border-radius:var(--scalar-radius-lg);background-color:var(--scalar-background-1);max-height:60dvh;width:100%;max-width:580px;margin:12px;position:fixed;left:50%;top:150px;opacity:0;transform:translate3d(-50%,10px,0);animation:fadeincommandmenu-e9fd9234 ease-in-out .3s forwards;animation-delay:.1s}.commandmenu-overlay[data-v-e9fd9234]{position:fixed;top:0;right:0;bottom:0;left:0;background:#0003;animation:fadeincommand-e9fd9234 ease-in-out .3s forwards;cursor:pointer}@keyframes fadeincommand-e9fd9234{0%{opacity:0}to{opacity:1}}@keyframes fadeincommandmenu-e9fd9234{0%{opacity:0;transform:translate3d(-50%,10px,0)}to{opacity:1;transform:translate3d(-50%,0,0)}}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url(\'data:image/svg+xml,<svg viewBox="0 0 24 24" fill="none" xmlns="http://www.w3.org/2000/svg"><path d="m9 18 6-6-6-6" stroke="currentColor" stroke-width="1.5" stroke-linecap="round" stroke-linejoin="round"/></svg>\');mask-image:url(\'data:image/svg+xml,<svg viewBox="0 0 24 24" fill="none" xmlns="http://www.w3.org/2000/svg"><path d="m9 18 6-6-6-6" stroke="currentColor" stroke-width="1.5" stroke-linecap="round" stroke-linejoin="round"/></svg>\')}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url(\'data:image/svg+xml,<svg viewBox="0 0 24 24" fill="none" xmlns="http://www.w3.org/2000/svg"><path d="m9 18 6-6-6-6" stroke="currentColor" stroke-width="1.5" stroke-linecap="round" stroke-linejoin="round"/></svg>\');mask-image:url(\'data:image/svg+xml,<svg viewBox="0 0 24 24" fill="none" xmlns="http://www.w3.org/2000/svg"><path d="m9 18 6-6-6-6" stroke="currentColor" stroke-width="1.5" stroke-linecap="round" stroke-linejoin="round"/></svg>\')}@layer scalar-base{*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar .scalar-app-layout[data-v-cc5e3e81]{background:var(--scalar-background-1);height:calc(100% - 120px);max-width:1390px;width:100%;margin:auto;opacity:0;animation:scalarapiclientfadein-cc5e3e81 .35s forwards;position:relative;overflow:hidden;border-radius:8px;border:var(--scalar-border-width) solid var(--scalar-border-color)}@keyframes scalarapiclientfadein-cc5e3e81{0%{opacity:0}to{opacity:1}}.scalar .scalar-app-exit[data-v-cc5e3e81]{position:fixed;top:0;left:0;width:100vw;height:100vh;background:#00000038;transition:all .3s ease-in-out;cursor:pointer;animation:scalardrawerexitfadein-cc5e3e81 .35s forwards}.dark-mode .scalar .scalar-app-exit[data-v-cc5e3e81]{background:#00000073}.scalar .scalar-app-exit[data-v-cc5e3e81]:before{font-family:sans-serif;position:absolute;top:0;font-size:30px;font-weight:100;line-height:50px;right:12px;text-align:center;color:#fff;opacity:.6}.scalar .scalar-app-exit[data-v-cc5e3e81]:hover:before{opacity:1}@keyframes scalardrawerexitfadein-cc5e3e81{0%{opacity:0}to{opacity:1}}.scalar-container[data-v-cc5e3e81]{overflow:hidden;visibility:visible;position:fixed;bottom:0;top:0;left:0;width:100%;height:100%;display:flex;align-items:center;justify-content:center;z-index:10000}.scalar .url-form-input[data-v-cc5e3e81]{min-height:auto!important}.scalar .scalar-container[data-v-cc5e3e81]{line-height:normal}.scalar .scalar-app-header span[data-v-cc5e3e81]{color:var(--scalar-color-3)}.scalar .scalar-app-header a[data-v-cc5e3e81]{color:var(--scalar-color-1)}.scalar .scalar-app-header a[data-v-cc5e3e81]:hover{text-decoration:underline}.scalar-activate[data-v-cc5e3e81]{width:-moz-fit-content;width:fit-content;margin:0px .75rem .75rem auto;line-height:24px;cursor:pointer;font-size:.875rem;font-weight:600;display:flex;align-items:center;gap:6px}.scalar-activate-button[data-v-cc5e3e81]{display:flex;gap:6px;align-items:center;color:var(--scalar-color-blue);-webkit-appearance:none;-moz-appearance:none;appearance:none;outline:none;border:none;background:transparent;padding:0 .5rem}.scalar-activate:hover .scalar-activate-button[data-v-cc5e3e81]{background:var(--scalar-background-3);border-radius:3px}[data-v-4d46a9ea] .cm-editor{height:100%;outline:none;padding:0;background:transparent}[data-v-4d46a9ea] .cm-placeholder{color:var(--scalar-color-3)}[data-v-4d46a9ea] .cm-content{font-family:var(--scalar-font-code);font-size:var(--scalar-small);max-height:20px;padding:8px 0}[data-v-4d46a9ea] .cm-tooltip{background:transparent!important;filter:brightness(var(--scalar-lifted-brightness));border-radius:var(--scalar-radius);box-shadow:var(--scalar-shadow-2);border:none!important;outline:none!important;overflow:hidden!important}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul li{padding:3px 6px!important}[data-v-4d46a9ea] .cm-completionIcon-type:after{color:var(--scalar-color-3)!important}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul li[aria-selected]{background:var(--scalar-background-2)!important;color:var(--scalar-color-1)!important}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul{padding:6px!important;position:relative}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul li:hover{border-radius:3px;color:var(--scalar-color-1)!important;background:var(--scalar-background-3)!important}[data-v-4d46a9ea] .cm-activeLine,[data-v-4d46a9ea] .cm-activeLineGutter{background-color:transparent}[data-v-4d46a9ea] .cm-selectionMatch,[data-v-4d46a9ea] .cm-matchingBracket{border-radius:var(--scalar-radius);background:var(--scalar-background-4)!important}[data-v-4d46a9ea] .cm-css-color-picker-wrapper{display:inline-flex;outline:1px solid var(--scalar-background-3);border-radius:3px;overflow:hidden}[data-v-4d46a9ea] .cm-gutters{background-color:transparent;border-right:none;color:var(--scalar-color-3);font-size:var(--scalar-mini);line-height:1.44;border-radius:0 0 0 3px}[data-v-4d46a9ea] .cm-gutters:before{content:"";position:absolute;top:2px;left:2px;width:calc(100% - 2px);height:calc(100% - 4px);border-radius:var(--scalar-radius) 0 0 var(--scalar-radius);background-color:var(--scalar-background-1)}[data-v-4d46a9ea] .cm-gutterElement{font-family:var(--scalar-font-code)!important;padding:0 6px 0 8px!important;display:flex;align-items:center;justify-content:flex-end;position:relative}[data-v-4d46a9ea] .cm-gutter+.cm-gutter :not(.cm-foldGutter) .cm-gutterElement{padding-left:0!important}[data-v-4d46a9ea] .cm-scroller{overflow:auto}.peer:hover .copy-button[data-v-4d46a9ea],.copy-button[data-v-4d46a9ea]:focus-visible{opacity:100}.scalar-code-copy[data-v-4d46a9ea]{align-items:flex-start;display:flex;top:0;right:0;bottom:0;left:0;justify-content:flex-end;position:sticky}.copy-button[data-v-4d46a9ea]{align-items:center;display:flex;position:relative;background-color:var(--scalar-background-2);border:1px solid var(--scalar-border-color);border-radius:3px;color:var(--scalar-color-3);cursor:pointer;height:30px;margin-bottom:-30px;opacity:0;padding:6px;transition:opacity .15s ease-in-out,color .15s ease-in-out;top:0;right:0}.scalar-code-copy[data-v-4d46a9ea],.copy-button[data-v-4d46a9ea]{background:inherit}.copy-button[data-v-4d46a9ea]:hover{color:var(--scalar-color-1)}.copy-button svg[data-v-4d46a9ea]{stroke-width:1.5}.cm-pill{--tw-bg-base: var(--scalar-color-1);color:var(--tw-bg-base);padding:0 9px;display:inline-block;border-radius:30px;font-size:var(--scalar-mini);background:color-mix(in srgb,var(--tw-bg-base),transparent 94%)!important}.cm-pill.bg-grey{background:var(--scalar-background-3)!important}.dark-mode .cm-pill{background:color-mix(in srgb,var(--tw-bg-base),transparent 80%)!important}.cm-pill:first-of-type{margin-left:0}.cm-editor .cm-widgetBuffer{display:none}.cm-foldPlaceholder:hover{color:var(--scalar-color-1)}.cm-foldGutter .cm-gutterElement{font-size:var(--scalar-heading-4);padding:2px!important}.cm-foldGutter .cm-gutterElement:first-of-type{display:none}.cm-foldGutter .cm-gutterElement .cm-foldMarker{padding:2px}.cm-foldGutter .cm-gutterElement:hover .cm-foldMarker{background:var(--scalar-background-2);border-radius:var(--scalar-radius);color:var(--scalar-color-1)}[data-v-cb0ff5b1] .cm-editor{padding:0}[data-v-cb0ff5b1] .cm-content{align-items:center;background-color:transparent;display:flex;font-family:var(--scalar-font);font-size:var(--scalar-mini);padding:6px 8px}[data-v-cb0ff5b1] .cm-content:has(.cm-pill){padding:4px 3px}[data-v-cb0ff5b1] .cm-content .cm-pill:not(:last-of-type){margin-right:.5px}[data-v-cb0ff5b1] .cm-content .cm-pill:not(:first-of-type){margin-left:.5px}[data-v-cb0ff5b1] .cm-line{padding:0}.required[data-v-cb0ff5b1]:after{content:"Required"}input[data-v-cb0ff5b1]::-moz-placeholder{color:var(--scalar-color-3)}input[data-v-cb0ff5b1]::placeholder{color:var(--scalar-color-3)}.scalar-modal-layout .scalar-button-danger[data-v-b6a60dbc]{background:color-mix(in srgb,var(--scalar-color-red),transparent 95%);color:var(--scalar-color-red)}.scalar-modal-layout .scalar-button-danger[data-v-b6a60dbc]:hover,.scalar-modal-layout .scalar-button-danger[data-v-b6a60dbc]:focus{background:color-mix(in srgb,var(--scalar-color-red),transparent 90%)}.fade-request-section-content[data-v-7906217f]{background:linear-gradient(to left,var(--scalar-background-1) 64%,transparent)}.filter-hover[data-v-7906217f]{height:100%;padding-right:42px;padding-left:24px;position:absolute;right:0;transition:width 0s ease-in-out .2s;overflow:hidden}.filter-hover[data-v-7906217f]:hover,.filter-hover[data-v-7906217f]:has(:focus-visible){width:100%;z-index:10}.filter-hover[data-v-7906217f]:before{content:"";position:absolute;top:0;left:0;width:100%;height:100%;background-color:var(--scalar-background-1);opacity:0;transition:all .3s ease-in-out;pointer-events:none}.filter-hover-item[data-v-7906217f]{opacity:0}.filter-hover-item[data-v-7906217f]:not(:last-of-type){transform:translate3d(0,3px,0)}.filter-hover:hover .filter-hover-item[data-v-7906217f]{transition:opacity .2s ease-in-out,transform .2s ease-in-out}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(1){transition-delay:.05s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(2){transition-delay:.1s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(3){transition-delay:.15s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(4){transition-delay:.2s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(5){transition-delay:.25s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(6){transition-delay:.3s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(7){transition-delay:.35s}.filter-hover:hover .filter-hover-item[data-v-7906217f],.filter-hover:has(:focus-visible) .filter-hover-item[data-v-7906217f]{opacity:1;transform:translateZ(0)}.filter-hover[data-v-7906217f]:hover:before,.filter-hover[data-v-7906217f]:has(:focus-visible):before{opacity:.9;-webkit-backdrop-filter:blur(10px);backdrop-filter:blur(10px)}.context-bar-group:hover .context-bar-group-hover\\:text-c-1[data-v-7906217f],.context-bar-group:has(:focus-visible) .context-bar-group-hover\\:text-c-1[data-v-7906217f]{--tw-text-opacity: 1;color:rgb(var(--scalar-color-1) / var(--tw-text-opacity))}.context-bar-group:hover .context-bar-group-hover\\:hidden[data-v-7906217f],.context-bar-group:has(:focus-visible) .context-bar-group-hover\\:hidden[data-v-7906217f]{display:none}.light-mode .bg-preview[data-v-5f138d26]{background-image:url("data:image/svg+xml,%3Csvg xmlns=\'http://www.w3.org/2000/svg\' width=\'16\' height=\'16\' fill=\'%23000\' fill-opacity=\'10%25\'%3E%3Crect width=\'8\' height=\'8\' /%3E%3Crect x=\'8\' y=\'8\' width=\'8\' height=\'8\' /%3E%3C/svg%3E")}.dark-mode .bg-preview[data-v-5f138d26]{background-image:url("data:image/svg+xml,%3Csvg xmlns=\'http://www.w3.org/2000/svg\' width=\'16\' height=\'16\' fill=\'%23FFF\' fill-opacity=\'10%25\'%3E%3Crect width=\'8\' height=\'8\' /%3E%3Crect x=\'8\' y=\'8\' width=\'8\' height=\'8\' /%3E%3C/svg%3E")}[data-v-99361e9b] .cm-editor{background-color:transparent;font-size:var(--scalar-mini);outline:none}[data-v-99361e9b] .cm-gutters{background-color:var(--scalar-background-1);border-radius:var(--scalar-radius) 0 0 var(--scalar-radius)}.scalar-code-copy[data-v-99361e9b]{align-items:flex-start;display:flex;position:absolute;top:6px;right:6px;z-index:10;pointer-events:none;position:sticky;transform:translate(-6px)}.copy-button[data-v-99361e9b]{align-items:center;display:flex;background-color:var(--scalar-background-1);border:1px solid var(--scalar-border-color);border-radius:3px;color:var(--scalar-color-3);cursor:pointer;height:30px;opacity:0;padding:6px;pointer-events:auto;transition:opacity .15s ease-in-out,color .15s ease-in-out}.body-raw:hover .copy-button[data-v-99361e9b],.copy-button[data-v-99361e9b]:focus-visible{opacity:1}.copy-button[data-v-99361e9b]:hover{color:var(--scalar-color-1)}.scalar-code-block[data-v-cf80aa59] .hljs *{font-size:var(--scalar-mini)}.ascii-art-animate .ascii-art-line[data-v-0018c948]{border-right:1ch solid transparent;animation:typewriter-0018c948 4s steps(1) 1s 1 normal both,blinkTextCursor-0018c948 .5s steps(1) infinite normal}@keyframes typewriter-0018c948{0%{width:0}to{width:100%}}@keyframes blinkTextCursor-0018c948{0%{border-right-color:currentColor}50%{border-right-color:transparent}}.keycap-n[data-v-6c848a44]{background:-webkit-linear-gradient(5deg,transparent 30%,var(--scalar-color-3) 50%);-webkit-background-clip:text;-webkit-text-fill-color:transparent}.keycap-hotkey[data-v-6c848a44]{position:absolute;top:32px;line-height:26px}.scalar-version-number[data-v-04a8f835]{transform:skew(0,13deg);width:76px;height:76px;position:absolute;margin-left:-36px;font-size:8px;font-family:var(--scalar-font-code);line-height:11px;margin-top:-113px;border-radius:9px 9px 16px 12px;box-shadow:inset 2px 0 0 2px var(--scalar-background-2);text-align:center;display:flex;flex-direction:column;align-items:center;justify-content:center;text-transform:initial;text-decoration-color:var(--scalar-color-3)}.scalar-version-number a[data-v-04a8f835]{font-weight:700;background:var(--scalar-background-2);padding:2px 4px;border-radius:3px;text-decoration:none;border:.5px solid var(--scalar-border-color)}.gitbook-show[data-v-04a8f835]{display:none}.v-enter-active[data-v-ef5f83d8]{transition:opacity .5s ease}.v-enter-from[data-v-ef5f83d8]{opacity:0}.response-body-virtual[data-headlessui-state=open],.response-body-virtual[data-headlessui-state=open] .diclosure-panel{display:flex;flex-direction:column;flex-grow:1}.animate-response-heading .response-heading[data-v-ec732740]{animation:push-response-ec732740 .2s ease-in-out forwards;opacity:1}@keyframes push-response-ec732740{0%{opacity:1;transform:translateY(0)}to{opacity:0;transform:translateY(-4px)}}.animate-response-heading .animate-response-children[data-v-ec732740]{animation:response-spans-ec732740 .2s ease-in-out forwards .05s;opacity:0}@keyframes response-spans-ec732740{0%{opacity:0;transform:translateY(4px)}to{opacity:1;transform:translateY(0)}}.schema>span[data-v-30c955c6]:not(:first-child):before{content:"·";display:block;margin:0 .5ch}.schema>span[data-v-30c955c6]{display:flex;white-space:nowrap}[data-v-ccef6266] .cm-editor{padding:0}[data-v-ccef6266] .cm-content{align-items:center;background-color:transparent;display:flex;font-family:var(--scalar-font);font-size:var(--scalar-mini);padding:6px 8px}[data-v-ccef6266] .cm-content:has(.cm-pill){padding:4px 3px}[data-v-ccef6266] .cm-content .cm-pill:not(:last-of-type){margin-right:.5px}[data-v-ccef6266] .cm-content .cm-pill:not(:first-of-type){margin-left:.5px}[data-v-ccef6266] .cm-line{padding:0}.filemask[data-v-ccef6266]{-webkit-mask-image:linear-gradient(to right,transparent 0,var(--scalar-background-2) 20px);mask-image:linear-gradient(to right,transparent 0,var(--scalar-background-2) 20px)}[data-v-3f1cd888] .cm-content{font-size:var(--scalar-mini)}.auth-blend-required[data-v-2b678dcb] .scalar-input-required,.auth-blend-required[data-v-2b678dcb] .required{background:var(--scalar-background-2);--tw-bg-base: var(--scalar-background-2);--tw-shadow: -8px 0 4px var(--scalar-background-2)}.request-example-references-header[data-v-2b678dcb]+tr>td{border-top:0;border-top-left-radius:0;border-top-right-radius:0}.auth-combobox-position[data-v-3569c0b2]{margin-left:120px}.scroll-timeline-x[data-v-3569c0b2]{overflow:auto;scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;scrollbar-width:none;-ms-overflow-style:none}.fade-left[data-v-3569c0b2],.fade-right[data-v-3569c0b2]{position:sticky;content:"";height:100%;animation-name:fadein-3569c0b2;animation-duration:1ms;animation-direction:reverse;animation-timeline:--scroll-timeline;min-height:24px;pointer-events:none}.fade-left[data-v-3569c0b2]{background:linear-gradient(-90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);min-width:3px;left:-1px;animation-direction:normal}.fade-right[data-v-3569c0b2]{background:linear-gradient(90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);margin-left:-20px;min-width:24px;right:-1px;top:0}@keyframes fadein-3569c0b2{0%{opacity:0}15%{opacity:1}}.auth-combobox-position[data-v-3dc33a43]{margin-left:120px}.scroll-timeline-x[data-v-3dc33a43]{overflow:auto;scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;scrollbar-width:none;-ms-overflow-style:none}.fade-left[data-v-3dc33a43],.fade-right[data-v-3dc33a43]{position:sticky;content:"";height:100%;animation-name:fadein-3dc33a43;animation-duration:1ms;animation-direction:reverse;animation-timeline:--scroll-timeline;min-height:24px;pointer-events:none}.fade-left[data-v-3dc33a43]{background:linear-gradient(-90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);min-width:3px;left:-1px;animation-direction:normal}.fade-right[data-v-3dc33a43]{background:linear-gradient(90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);margin-left:-20px;min-width:24px;right:-1px;top:0}@keyframes fadein-3dc33a43{0%{opacity:0}15%{opacity:1}}[data-v-c2f2c62b] code.hljs *{font-size:var(--scalar-mini)}.request-section-content[data-v-5ab558f8]{--scalar-border-width: .5px}.request-section-content-filter[data-v-5ab558f8]{box-shadow:0 -10px 0 10px var(--scalar-background-1)}.request-item:focus-within .request-meta-buttons[data-v-5ab558f8]{opacity:1}.group-hover-input[data-v-5ab558f8]{border-width:var(--scalar-border-width);border-color:transparent}.group:hover .group-hover-input[data-v-5ab558f8]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2));border-color:var(--scalar-border-color)}.group-hover-input[data-v-5ab558f8]:focus{background:transparent!important;border-color:var(--scalar-border-color)!important}@media (min-width: 800px){.has-no-import-url,.has-import-url{max-width:100dvw;overflow-x:hidden;contain:paint}.has-no-import-url{opacity:1;background:var(--scalar-background-1);animation:transform-restore-layout ease-in-out .3s forwards}.has-import-url .scalar-client>main{opacity:0;transform:scale(.85) translate3d(calc(50dvw + 80px),0,0);animation:transform-fade-layout ease-in-out .3s forwards;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:12px;overflow:hidden;z-index:10000}.has-import-url .scalar-client .sidenav{display:none}.has-no-import-url .scalar-app,.has-import-url .scalar-app{background:var(--scalar-background-1)!important}}@keyframes transform-fade-layout{0%{opacity:0;transform:scale(.85) translate3d(calc(50dvw + 80px),10px,0)}to{opacity:1;transform:scale(.85) translate3d(calc(50dvw + 80px),0,0)}}@keyframes transform-restore-layout{0%{opacity:1;transform:scale(.85) translate3d(calc(50dvw + 80px),0,0)}to{opacity:1;transform:scale(1) translateZ(0)}}.openapi-color{color:var(--scalar-color-green)}.section-flare{position:fixed;top:0;right:-50dvw}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}:where(.scalar-app) .container{width:100%}@media (min-width: 400px){:where(.scalar-app) .container{max-width:400px}}@media (min-width: 600px){:where(.scalar-app) .container{max-width:600px}}@media (min-width: 800px){:where(.scalar-app) .container{max-width:800px}}@media (min-width: 1000px){:where(.scalar-app) .container{max-width:1000px}}@media (min-width: 1200px){:where(.scalar-app) .container{max-width:1200px}}:where(.scalar-app) .sr-only{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);white-space:nowrap;border-width:0}:where(.scalar-app) .pointer-events-none{pointer-events:none}:where(.scalar-app) .visible{visibility:visible}:where(.scalar-app) .invisible{visibility:hidden}:where(.scalar-app) .fixed{position:fixed}:where(.scalar-app) .absolute{position:absolute}:where(.scalar-app) .relative{position:relative}:where(.scalar-app) .sticky{position:sticky}:where(.scalar-app) .inset-0{top:0;right:0;bottom:0;left:0}:where(.scalar-app) .left-0{left:0}:where(.scalar-app) .left-2{left:8px}:where(.scalar-app) .left-2\\.5{left:10px}:where(.scalar-app) .left-4{left:16px}:where(.scalar-app) .left-border{left:var(--scalar-border-width)}:where(.scalar-app) .left-px{left:1px}:where(.scalar-app) .right-2{right:8px}:where(.scalar-app) .top-0{top:0}:where(.scalar-app) .top-1\\/2{top:50%}:where(.scalar-app) .top-12{top:48px}:where(.scalar-app) .top-2{top:8px}:where(.scalar-app) .top-22{top:88px}:where(.scalar-app) .top-\\[0\\.5px\\]{top:.5px}:where(.scalar-app) .top-px{top:1px}:where(.scalar-app) .-z-1{z-index:-1}:where(.scalar-app) .z-1{z-index:1}:where(.scalar-app) .z-\\[1001\\]{z-index:1001}:where(.scalar-app) .z-context{z-index:1000}:where(.scalar-app) .\\!m-0{margin:0!important}:where(.scalar-app) .-m-px{margin:-1px}:where(.scalar-app) .m-0{margin:0}:where(.scalar-app) .m-auto{margin:auto}:where(.scalar-app) .-mx-0{margin-left:-0px;margin-right:-0px}:where(.scalar-app) .-mx-0\\.75{margin-left:-3px;margin-right:-3px}:where(.scalar-app) .-mx-px{margin-left:-1px;margin-right:-1px}:where(.scalar-app) .-my-1{margin-top:-4px;margin-bottom:-4px}:where(.scalar-app) .-my-1\\.5{margin-top:-6px;margin-bottom:-6px}:where(.scalar-app) .mx-2{margin-left:8px;margin-right:8px}:where(.scalar-app) .mx-auto{margin-left:auto;margin-right:auto}:where(.scalar-app) .mx-px{margin-left:1px;margin-right:1px}:where(.scalar-app) .my-0{margin-top:0;margin-bottom:0}:where(.scalar-app) .my-0\\.75{margin-top:3px;margin-bottom:3px}:where(.scalar-app) .my-3{margin-top:12px;margin-bottom:12px}:where(.scalar-app) .-mb-1{margin-bottom:-4px}:where(.scalar-app) .-ml-0{margin-left:-0px}:where(.scalar-app) .-ml-0\\.75{margin-left:-3px}:where(.scalar-app) .-mr-0{margin-right:-0px}:where(.scalar-app) .-mr-0\\.25{margin-right:-1px}:where(.scalar-app) .mb-0{margin-bottom:0}:where(.scalar-app) .ml-1{margin-left:4px}:where(.scalar-app) .ml-auto{margin-left:auto}:where(.scalar-app) .mr-2{margin-right:8px}:where(.scalar-app) .mr-3{margin-right:12px}:where(.scalar-app) .mt-0{margin-top:0}:where(.scalar-app) .mt-1{margin-top:4px}:where(.scalar-app) .mt-1\\.5{margin-top:6px}:where(.scalar-app) .mt-2{margin-top:8px}:where(.scalar-app) .mt-20{margin-top:80px}:where(.scalar-app) .line-clamp-4{overflow:hidden;display:-webkit-box;-webkit-box-orient:vertical;-webkit-line-clamp:4}:where(.scalar-app) .block{display:block}:where(.scalar-app) .inline-block{display:inline-block}:where(.scalar-app) .flex{display:flex}:where(.scalar-app) .table{display:table}:where(.scalar-app) .grid{display:grid}:where(.scalar-app) .contents{display:contents}:where(.scalar-app) .hidden{display:none}:where(.scalar-app) .aspect-square{aspect-ratio:1 / 1}:where(.scalar-app) .size-10{width:40px;height:40px}:where(.scalar-app) .size-2{width:8px;height:8px}:where(.scalar-app) .size-3{width:12px;height:12px}:where(.scalar-app) .size-3\\.5{width:14px;height:14px}:where(.scalar-app) .size-4{width:16px;height:16px}:where(.scalar-app) .size-5{width:20px;height:20px}:where(.scalar-app) .size-6{width:24px;height:24px}:where(.scalar-app) .size-8{width:32px;height:32px}:where(.scalar-app) .size-\\[23px\\]{width:23px;height:23px}:where(.scalar-app) .size-full{width:100%;height:100%}:where(.scalar-app) .h-10{height:40px}:where(.scalar-app) .h-3{height:12px}:where(.scalar-app) .h-3\\.5{height:14px}:where(.scalar-app) .h-32{height:128px}:where(.scalar-app) .h-4{height:16px}:where(.scalar-app) .h-5{height:20px}:where(.scalar-app) .h-6{height:24px}:where(.scalar-app) .h-8{height:32px}:where(.scalar-app) .h-\\[100dvh\\]{height:100dvh}:where(.scalar-app) .h-border{height:var(--scalar-border-width)}:where(.scalar-app) .h-fit{height:-moz-fit-content;height:fit-content}:where(.scalar-app) .h-full{height:100%}:where(.scalar-app) .h-screen{height:100vh}:where(.scalar-app) .max-h-\\[440px\\]{max-height:440px}:where(.scalar-app) .max-h-\\[calc\\(100dvh-120px\\)\\]{max-height:calc(100dvh - 120px)}:where(.scalar-app) .max-h-\\[calc\\(100dvh-180px\\)\\]{max-height:calc(100dvh - 180px)}:where(.scalar-app) .max-h-\\[calc\\(100dvh-240px\\)\\]{max-height:calc(100dvh - 240px)}:where(.scalar-app) .max-h-\\[inherit\\]{max-height:inherit}:where(.scalar-app) .max-h-dvh{max-height:100dvh}:where(.scalar-app) .max-h-radix-popper{max-height:calc(var(--radix-popper-available-height) - 8px)}:where(.scalar-app) .min-h-0{min-height:0px}:where(.scalar-app) .min-h-12{min-height:48px}:where(.scalar-app) .min-h-5{min-height:20px}:where(.scalar-app) .min-h-6{min-height:24px}:where(.scalar-app) .min-h-8{min-height:32px}:where(.scalar-app) .min-h-96{min-height:384px}:where(.scalar-app) .min-h-\\[77px\\]{min-height:77px}:where(.scalar-app) .min-h-header{min-height:48px}:where(.scalar-app) .w-12{width:48px}:where(.scalar-app) .w-16{width:64px}:where(.scalar-app) .w-20{width:80px}:where(.scalar-app) .w-24{width:96px}:where(.scalar-app) .w-3{width:12px}:where(.scalar-app) .w-32{width:128px}:where(.scalar-app) .w-4{width:16px}:where(.scalar-app) .w-40{width:160px}:where(.scalar-app) .w-48{width:192px}:where(.scalar-app) .w-5{width:20px}:where(.scalar-app) .w-56{width:224px}:where(.scalar-app) .w-6{width:24px}:where(.scalar-app) .w-64{width:256px}:where(.scalar-app) .w-72{width:288px}:where(.scalar-app) .w-8{width:32px}:where(.scalar-app) .w-\\[100dvw\\]{width:100dvw}:where(.scalar-app) .w-\\[38px\\]{width:38px}:where(.scalar-app) .w-auto{width:auto}:where(.scalar-app) .w-full{width:100%}:where(.scalar-app) .w-min{width:-moz-min-content;width:min-content}:where(.scalar-app) .min-w-0{min-width:0px}:where(.scalar-app) .min-w-5{min-width:20px}:where(.scalar-app) .min-w-6{min-width:24px}:where(.scalar-app) .min-w-8{min-width:32px}:where(.scalar-app) .min-w-max{min-width:-moz-max-content;min-width:max-content}:where(.scalar-app) .min-w-min{min-width:-moz-min-content;min-width:min-content}:where(.scalar-app) .max-w-screen-lg{max-width:800px}:where(.scalar-app) .max-w-screen-md{max-width:640px}:where(.scalar-app) .max-w-screen-sm{max-width:540px}:where(.scalar-app) .max-w-screen-xl{max-width:1000px}:where(.scalar-app) .max-w-screen-xs{max-width:480px}:where(.scalar-app) .max-w-screen-xxs{max-width:360px}:where(.scalar-app) .max-w-xs{max-width:320px}:where(.scalar-app) .flex-1{flex:1 1 0%}:where(.scalar-app) .shrink-0{flex-shrink:0}:where(.scalar-app) .origin-top-left{transform-origin:top left}:where(.scalar-app) .-translate-y-1\\/2{--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .translate-x-2{--tw-translate-x: 8px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .translate-x-2\\.5{--tw-translate-x: 10px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .translate-x-\\[14px\\]{--tw-translate-x: 14px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .rotate-90{--tw-rotate: 90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .transform{transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .cursor-auto{cursor:auto}:where(.scalar-app) .cursor-default{cursor:default}:where(.scalar-app) .cursor-not-allowed{cursor:not-allowed}:where(.scalar-app) .cursor-pointer{cursor:pointer}:where(.scalar-app) .resize-none{resize:none}:where(.scalar-app) .resize{resize:both}:where(.scalar-app) .appearance-none{-webkit-appearance:none;-moz-appearance:none;appearance:none}:where(.scalar-app) .grid-cols-\\[auto_1fr\\]{grid-template-columns:auto 1fr}:where(.scalar-app) .flex-col{flex-direction:column}:where(.scalar-app) .flex-wrap{flex-wrap:wrap}:where(.scalar-app) .items-start{align-items:flex-start}:where(.scalar-app) .items-center{align-items:center}:where(.scalar-app) .justify-start{justify-content:flex-start}:where(.scalar-app) .justify-end{justify-content:flex-end}:where(.scalar-app) .justify-center{justify-content:center}:where(.scalar-app) .justify-between{justify-content:space-between}:where(.scalar-app) .gap-0{gap:0px}:where(.scalar-app) .gap-0\\.5{gap:2px}:where(.scalar-app) .gap-0\\.75{gap:3px}:where(.scalar-app) .gap-1{gap:4px}:where(.scalar-app) .gap-1\\.5{gap:6px}:where(.scalar-app) .gap-16{gap:64px}:where(.scalar-app) .gap-2{gap:8px}:where(.scalar-app) .gap-2\\.5{gap:10px}:where(.scalar-app) .gap-3{gap:12px}:where(.scalar-app) .gap-4{gap:16px}:where(.scalar-app) .gap-x-4{-moz-column-gap:16px;column-gap:16px}:where(.scalar-app) .gap-y-8{row-gap:32px}:where(.scalar-app) .divide-y>:not([hidden])~:not([hidden]){--tw-divide-y-reverse: 0;border-top-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-y-reverse)));border-bottom-width:calc(var(--scalar-border-width) * var(--tw-divide-y-reverse))}:where(.scalar-app) .self-center{align-self:center}:where(.scalar-app) .overflow-auto{overflow:auto}:where(.scalar-app) .overflow-hidden{overflow:hidden}:where(.scalar-app) .overflow-visible{overflow:visible}:where(.scalar-app) .overflow-y-scroll{overflow-y:scroll}:where(.scalar-app) .overscroll-contain{overscroll-behavior:contain}:where(.scalar-app) .truncate{overflow:hidden;text-overflow:ellipsis;white-space:nowrap}:where(.scalar-app) .text-ellipsis{text-overflow:ellipsis}:where(.scalar-app) .whitespace-nowrap{white-space:nowrap}:where(.scalar-app) .text-wrap{text-wrap:wrap}:where(.scalar-app) .text-nowrap{text-wrap:nowrap}:where(.scalar-app) .rounded{border-radius:var(--scalar-radius)}:where(.scalar-app) .rounded-full{border-radius:9999px}:where(.scalar-app) .rounded-lg{border-radius:var(--scalar-radius-lg)}:where(.scalar-app) .rounded-md{border-radius:var(--scalar-radius)}:where(.scalar-app) .rounded-none{border-radius:0}:where(.scalar-app) .rounded-xl{border-radius:var(--scalar-radius-xl)}:where(.scalar-app) .rounded-t-none{border-top-left-radius:0;border-top-right-radius:0}:where(.scalar-app) .border{border-width:var(--scalar-border-width)}:where(.scalar-app) .border-0{border-width:0}:where(.scalar-app) .border-2{border-width:calc(var(--scalar-border-width) * 2)}:where(.scalar-app) .border-\\[1px\\]{border-width:1px}:where(.scalar-app) .border-b{border-bottom-width:var(--scalar-border-width)}:where(.scalar-app) .border-r{border-right-width:var(--scalar-border-width)}:where(.scalar-app) .border-t{border-top-width:var(--scalar-border-width)}:where(.scalar-app) .border-t-1\\/2{border-top-width:calc(var(--scalar-border-width) / 2)}:where(.scalar-app) .border-solid{border-style:solid}:where(.scalar-app) .border-dashed{border-style:dashed}:where(.scalar-app) .border-none{border-style:none}:where(.scalar-app) .border-border{border-color:var(--scalar-border-color)}:where(.scalar-app) .border-c-1{border-color:var(--scalar-color-1)}:where(.scalar-app) .border-c-3{border-color:var(--scalar-color-3)}:where(.scalar-app) .border-red{border-color:var(--scalar-color-red)}:where(.scalar-app) .border-sidebar-border{border-color:var(--scalar-sidebar-border-color, var(--scalar-border-color))}:where(.scalar-app) .bg-b-1{background-color:var(--scalar-background-1)}:where(.scalar-app) .bg-b-2{background-color:var(--scalar-background-2)}:where(.scalar-app) .bg-b-3{background-color:var(--scalar-background-3)}:where(.scalar-app) .bg-b-btn{background-color:var(--scalar-button-1)}:where(.scalar-app) .bg-backdrop{background-color:#00000038}:where(.scalar-app) .bg-blue{background-color:var(--scalar-color-blue)}:where(.scalar-app) .bg-border{background-color:var(--scalar-border-color)}:where(.scalar-app) .bg-c-accent{background-color:var(--scalar-color-accent)}:where(.scalar-app) .bg-c-danger{background-color:var(--scalar-color-danger)}:where(.scalar-app) .bg-red{background-color:var(--scalar-color-red)}:where(.scalar-app) .bg-sidebar-b-1{background-color:var(--scalar-sidebar-background-1, var(--scalar-background-1))}:where(.scalar-app) .bg-transparent{background-color:transparent}:where(.scalar-app) .bg-white{--tw-bg-opacity: 1;background-color:rgb(255 255 255 / var(--tw-bg-opacity))}:where(.scalar-app) .p-0{padding:0}:where(.scalar-app) .p-0\\.5{padding:2px}:where(.scalar-app) .p-0\\.75{padding:3px}:where(.scalar-app) .p-1{padding:4px}:where(.scalar-app) .p-1\\.5{padding:6px}:where(.scalar-app) .p-2{padding:8px}:where(.scalar-app) .p-3{padding:12px}:where(.scalar-app) .p-4{padding:16px}:where(.scalar-app) .p-6{padding:24px}:where(.scalar-app) .px-1{padding-left:4px;padding-right:4px}:where(.scalar-app) .px-1\\.5{padding-left:6px;padding-right:6px}:where(.scalar-app) .px-2{padding-left:8px;padding-right:8px}:where(.scalar-app) .px-2\\.5{padding-left:10px;padding-right:10px}:where(.scalar-app) .px-3{padding-left:12px;padding-right:12px}:where(.scalar-app) .px-4{padding-left:16px;padding-right:16px}:where(.scalar-app) .px-6{padding-left:24px;padding-right:24px}:where(.scalar-app) .px-9{padding-left:36px;padding-right:36px}:where(.scalar-app) .py-1{padding-top:4px;padding-bottom:4px}:where(.scalar-app) .py-1\\.5{padding-top:6px;padding-bottom:6px}:where(.scalar-app) .py-2{padding-top:8px;padding-bottom:8px}:where(.scalar-app) .py-2\\.5{padding-top:10px;padding-bottom:10px}:where(.scalar-app) .py-3{padding-top:12px;padding-bottom:12px}:where(.scalar-app) .py-4{padding-top:16px;padding-bottom:16px}:where(.scalar-app) .pb-0{padding-bottom:0}:where(.scalar-app) .pl-8{padding-left:32px}:where(.scalar-app) .pl-\\[114px\\]{padding-left:114px}:where(.scalar-app) .pl-\\[24px\\]{padding-left:24px}:where(.scalar-app) .pl-\\[42px\\]{padding-left:42px}:where(.scalar-app) .pl-\\[60px\\]{padding-left:60px}:where(.scalar-app) .pl-\\[6px\\]{padding-left:6px}:where(.scalar-app) .pl-\\[78px\\]{padding-left:78px}:where(.scalar-app) .pl-\\[96px\\]{padding-left:96px}:where(.scalar-app) .pr-3{padding-right:12px}:where(.scalar-app) .pt-3{padding-top:12px}:where(.scalar-app) .pt-6{padding-top:24px}:where(.scalar-app) .text-left{text-align:left}:where(.scalar-app) .font-code{font-family:var(--scalar-font-code)}:where(.scalar-app) .text-3xs{font-size:var(--scalar-font-size-7)}:where(.scalar-app) .text-lg{font-size:var(--scalar-font-size-2)}:where(.scalar-app) .text-sm{font-size:var(--scalar-font-size-4)}:where(.scalar-app) .text-xl{font-size:var(--scalar-font-size-1)}:where(.scalar-app) .text-xs{font-size:var(--scalar-font-size-5)}:where(.scalar-app) .text-xxs{font-size:var(--scalar-font-size-6)}:where(.scalar-app) .font-bold{font-weight:var(--scalar-bold)}:where(.scalar-app) .font-medium{font-weight:var(--scalar-semibold)}:where(.scalar-app) .font-normal{font-weight:var(--scalar-regular)}:where(.scalar-app) .leading-3{line-height:var(--scalar-line-height-3)}:where(.scalar-app) .leading-none{line-height:1}:where(.scalar-app) .leading-snug{line-height:1.375}:where(.scalar-app) .text-b-1{color:var(--scalar-background-1)}:where(.scalar-app) .text-blue{color:var(--scalar-color-blue)}:where(.scalar-app) .text-c-1{color:var(--scalar-color-1)}:where(.scalar-app) .text-c-2{color:var(--scalar-color-2)}:where(.scalar-app) .text-c-3{color:var(--scalar-color-3)}:where(.scalar-app) .text-c-accent{color:var(--scalar-color-accent)}:where(.scalar-app) .text-c-btn{color:var(--scalar-button-1-color)}:where(.scalar-app) .text-c-danger{color:var(--scalar-color-danger)}:where(.scalar-app) .text-c-ghost{color:var(--scalar-color-ghost)}:where(.scalar-app) .text-red{color:var(--scalar-color-red)}:where(.scalar-app) .text-sidebar-c-2{color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}:where(.scalar-app) .text-transparent{color:transparent}:where(.scalar-app) .text-white{--tw-text-opacity: 1;color:rgb(255 255 255 / var(--tw-text-opacity))}:where(.scalar-app) .no-underline{text-decoration-line:none}:where(.scalar-app) .opacity-0{opacity:0}:where(.scalar-app) .opacity-40{opacity:.4}:where(.scalar-app) .opacity-50{opacity:.5}:where(.scalar-app) .shadow{--tw-shadow: var(--scalar-shadow-1);--tw-shadow-colored: var(--scalar-shadow-1);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-border{--tw-shadow: inset 0 0 0 1px var(--scalar-border-color);--tw-shadow-colored: inset 0 0 0 1px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-lg{--tw-shadow: var(--scalar-shadow-2);--tw-shadow-colored: var(--scalar-shadow-2);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-none{--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-sm{--tw-shadow: rgba(0, 0, 0, .09) 0px 1px 4px;--tw-shadow-colored: 0px 1px 4px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-border{--tw-shadow-color: var(--scalar-border-color);--tw-shadow: var(--tw-shadow-colored)}:where(.scalar-app) .shadow-current{--tw-shadow-color: currentColor;--tw-shadow: var(--tw-shadow-colored)}:where(.scalar-app) .outline-none{outline:2px solid transparent;outline-offset:2px}:where(.scalar-app) .outline{outline-style:solid}:where(.scalar-app) .-outline-offset-1{outline-offset:-1px}:where(.scalar-app) .outline-offset-1{outline-offset:1px}:where(.scalar-app) .blur{--tw-blur: blur(8px);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .brightness-lifted{--tw-brightness: brightness(var(--scalar-lifted-brightness));filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .filter{filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .backdrop-blur{--tw-backdrop-blur: blur(8px);-webkit-backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia);backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia)}:where(.scalar-app) .transition{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,-webkit-backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter,-webkit-backdrop-filter;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .transition-colors{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .transition-opacity{transition-property:opacity;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .transition-transform{transition-property:transform;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .duration-100{transition-duration:.1s}:where(.scalar-app) .duration-300{transition-duration:.3s}:where(.scalar-app) .ease-in-out{transition-timing-function:cubic-bezier(.4,0,.2,1)}:where(.scalar-app) .bg-mix-transparent{background-color:color-mix(var(--tw-bg-mix-method, in srgb),transparent calc(var(--tw-bg-mix-amount, 0) * 1%),var(--tw-bg-base))}:where(.scalar-app) .bg-mix-amount-25{--tw-bg-mix-amount: 25}:where(.scalar-app) .bg-b-1{--tw-bg-base: var(--scalar-background-1)}:where(.scalar-app) .bg-b-2{--tw-bg-base: var(--scalar-background-2)}:where(.scalar-app) .bg-b-3{--tw-bg-base: var(--scalar-background-3)}:where(.scalar-app) .bg-b-btn{--tw-bg-base: var(--scalar-button-1)}:where(.scalar-app) .bg-backdrop{--tw-bg-base: rgba(0, 0, 0, .22)}:where(.scalar-app) .bg-blue{--tw-bg-base: var(--scalar-color-blue)}:where(.scalar-app) .bg-border{--tw-bg-base: var(--scalar-border-color)}:where(.scalar-app) .bg-c-accent{--tw-bg-base: var(--scalar-color-accent)}:where(.scalar-app) .bg-c-danger{--tw-bg-base: var(--scalar-color-danger)}:where(.scalar-app) .bg-red{--tw-bg-base: var(--scalar-color-red)}:where(.scalar-app) .bg-sidebar-b-1{--tw-bg-base: var(--scalar-sidebar-background-1, var(--scalar-background-1))}:where(.scalar-app) .bg-transparent{--tw-bg-base: transparent}:where(.scalar-app) .bg-white{--tw-bg-opacity: 1;--tw-bg-base: rgb(255 255 255 / var(--tw-bg-opacity))}:where(.scalar-app) .row,:where(.scalar-app) .scalar-row{display:flex;flex-direction:row}:where(.scalar-app) .col{display:flex;flex-direction:column}:where(.scalar-app) .centered-x{left:50%;--tw-translate-x: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .centered{left:50%;--tw-translate-x: -50%;top:50%;--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .\\*\\:flex>*{display:flex}:where(.scalar-app) .\\*\\:flex-1>*{flex:1 1 0%}:where(.scalar-app) .\\*\\:items-center>*{align-items:center}:where(.scalar-app) .\\*\\:gap-px>*{gap:1px}:where(.scalar-app) .hover\\:bg-b-2:hover{background-color:var(--scalar-background-2)}:where(.scalar-app) .hover\\:bg-b-3:hover{background-color:var(--scalar-background-3)}:where(.scalar-app) .hover\\:bg-h-btn:hover{background-color:var(--scalar-button-1-hover)}:where(.scalar-app) .hover\\:text-c-1:hover{color:var(--scalar-color-1)}:where(.scalar-app) .hover\\:underline:hover{text-decoration-line:underline}:where(.scalar-app) .hover\\:bg-b-2:hover{--tw-bg-base: var(--scalar-background-2)}:where(.scalar-app) .hover\\:bg-b-3:hover{--tw-bg-base: var(--scalar-background-3)}:where(.scalar-app) .hover\\:bg-h-btn:hover{--tw-bg-base: var(--scalar-button-1-hover)}:where(.scalar-app) .focus\\:outline-none:focus{outline:2px solid transparent;outline-offset:2px}:where(.scalar-app) .focus-visible\\:border-c-btn:focus-visible{border-color:var(--scalar-button-1-color)}:where(.scalar-app) .focus-visible\\:ring-1:focus-visible{--tw-ring-offset-shadow: var(--tw-ring-inset) 0 0 0 var(--tw-ring-offset-width) var(--tw-ring-offset-color);--tw-ring-shadow: var(--tw-ring-inset) 0 0 0 calc(1px + var(--tw-ring-offset-width)) var(--tw-ring-color);box-shadow:var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow, 0 0 #0000)}:where(.scalar-app) .focus-visible\\:ring-c-accent:focus-visible{--tw-ring-color: var(--scalar-color-accent)}:where(.scalar-app) .active\\:bg-b-btn:active{background-color:var(--scalar-button-1)}:where(.scalar-app) .active\\:text-c-1:active{color:var(--scalar-color-1)}:where(.scalar-app) .active\\:shadow-none:active{--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .active\\:brightness-90:active{--tw-brightness: brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .active\\:bg-b-btn:active{--tw-bg-base: var(--scalar-button-1)}:where(.scalar-app) .group\\/button:hover .group-hover\\/button\\:text-c-1,:where(.scalar-app) .group:hover .group-hover\\:text-c-1{color:var(--scalar-color-1)}:where(.scalar-app) .group\\/item:hover .group-hover\\/item\\:opacity-100{opacity:1}:where(.scalar-app) .group\\/toggle:focus-visible .group-focus-visible\\/toggle\\:outline{outline-style:solid}:where(.scalar-app) .peer:checked~.peer-checked\\:text-c-2{color:var(--scalar-color-2)}:where(.scalar-app) .peer:checked~.peer-checked\\:opacity-100{opacity:1}:where(.scalar-app) .has-\\[\\:focus-visible\\]\\:bg-b-1:has(:focus-visible){background-color:var(--scalar-background-1)}:where(.scalar-app) .has-\\[\\:focus-visible\\]\\:outline:has(:focus-visible){outline-style:solid}:where(.scalar-app) .has-\\[input\\:focus-visible\\]\\:outline:has(input:focus-visible){outline-style:solid}:where(.scalar-app) .has-\\[\\:focus-visible\\]\\:bg-b-1:has(:focus-visible){--tw-bg-base: var(--scalar-background-1)}:where(.scalar-app) .ui-open\\:rotate-180[data-headlessui-state~=open]{--tw-rotate: 180deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) :where([data-headlessui-state~=open]) .ui-open\\:rotate-180{--tw-rotate: 180deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .has-actv-btn\\:border:has(button:active){border-width:var(--scalar-border-width)}:where(.scalar-app) .has-actv-btn\\:border-border:has(button:active){border-color:var(--scalar-border-color)}:where(.scalar-app) .\\!empty\\:flex:not(:empty){display:flex}:where(.scalar-app) .\\!empty\\:w-7:not(:empty){width:28px}:where(.scalar-app) .\\!empty\\:items-center:not(:empty){align-items:center}:where(.scalar-app) .\\!empty\\:pr-3:not(:empty){padding-right:12px}:where(.scalar-app) .hocus\\:text-c-1:hover{color:var(--scalar-color-1)}:where(.scalar-app) .hocus\\:brightness-90:hover{--tw-brightness: brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .hocus\\:text-c-1:focus-visible{color:var(--scalar-color-1)}:where(.scalar-app) .hocus\\:brightness-90:focus-visible{--tw-brightness: brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}@media (min-width: 800px){:where(.scalar-app) .md\\:w-\\[calc\\(100vw-16px\\)\\]{width:calc(100vw - 16px)}:where(.scalar-app) .md\\:row{display:flex;flex-direction:row}}@media (min-width: 1000px){:where(.scalar-app) .lg\\:w-\\[calc\\(100vw-32px\\)\\]{width:calc(100vw - 32px)}:where(.scalar-app) .lg\\:w-full{width:100%}}:where(.scalar-app) .dark\\:bg-b-3:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-3)}:where(.scalar-app) .dark\\:bg-backdropdark:where(.dark-mode,.dark-mode *){background-color:#00000073}:where(.scalar-app) .dark\\:bg-b-3:where(.dark-mode,.dark-mode *){--tw-bg-base: var(--scalar-background-3)}:where(.scalar-app) .dark\\:bg-backdropdark:where(.dark-mode,.dark-mode *){--tw-bg-base: rgba(0, 0, 0, .45)}:where(.scalar-app) .dark\\:hover\\:bg-b-3:hover:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-3);--tw-bg-base: var(--scalar-background-3)}.loader-wrapper[data-v-65265e87]{position:relative;display:flex;align-items:center;justify-content:center;--loader-size: 50%}.svg-loader[data-v-65265e87]{width:var(--loader-size);height:var(--loader-size);top:1rem;right:.9rem;overflow:visible;fill:none;background-color:transparent;stroke:currentColor}.svg-path[data-v-65265e87]{stroke-width:14;fill:none;transition:.3s}.svg-x-mark[data-v-65265e87]{stroke-dasharray:57;stroke-dashoffset:57;transition-delay:0s}.svg-check-mark[data-v-65265e87]{stroke-dasharray:149;stroke-dashoffset:149;transition-delay:0s}.icon-is-invalid .svg-x-mark[data-v-65265e87],.icon-is-valid .svg-check-mark[data-v-65265e87]{stroke-dashoffset:0;transition-delay:.3s}.circular-loader[data-v-65265e87]{animation:rotate-65265e87 .7s linear infinite,fade-in-65265e87 .4s;transform-origin:center center;transform:scale(5);background:transparent}.loader-path[data-v-65265e87]{stroke-dasharray:50,200;stroke-dashoffset:-100;stroke-linecap:round}.loader-path-off[data-v-65265e87]{stroke-dasharray:50,200;stroke-dashoffset:-100;transition:opacity .3s;opacity:0}@keyframes fade-in-65265e87{0%{opacity:0}70%{opacity:0}to{opacity:1}}@keyframes rotate-65265e87{0%{transform:scale(5) rotate(0)}to{transform:scale(5) rotate(360deg)}}.scalar-icon[data-v-f4f53d32],.scalar-icon[data-v-f4f53d32] *{stroke-width:var(--ce5c5c8a)}.scalar-code-block{background:inherit;position:relative;overflow:auto;padding:.75rem}.scalar-code-block:hover .copy-button,.copy-button:focus-visible{opacity:100}.scalar-codeblock-pre{all:unset;margin:0;background:transparent;text-wrap:nowrap;white-space-collapse:preserve;border-radius:0;width:-moz-fit-content;width:fit-content}.scalar-code-copy{display:flex;align-items:flex-start;justify-content:flex-end;position:sticky;top:0;right:0;bottom:0;left:0}.copy-button{position:relative;top:0;right:0;display:flex;align-items:center;background-color:var(--scalar-background-2);border:1px solid var(--scalar-border-color);border-radius:3px;color:var(--scalar-color-3);cursor:pointer;height:30px;margin-bottom:-30px;opacity:0;padding:6px;transition:opacity .15s ease-in-out,color .15s ease-in-out}.scalar-code-copy,.copy-button{background:inherit}.copy-button:hover{color:var(--scalar-color-1)}.copy-button svg{stroke-width:1.5}.toggle-icon-ellipse[data-v-8bd002b1]{width:7px;height:7px;border-radius:50%;box-shadow:inset 0 0 0 1px currentColor;overflow:hidden;position:relative;background:var(--scalar-background-1);display:inline-block;transition:width .3s ease-in-out,height .3s ease-in-out}.toggle-icon-moon-mask[data-v-8bd002b1]{width:100%;height:100%;border:1px solid currentColor;display:block;left:2.5px;bottom:2.5px;position:absolute;border-radius:50%;background:var(--scalar-background-1);transition:transform .3s ease-in-out;transform:translate3d(4px,-4px,0)}.toggle-icon-sun-ray[data-v-8bd002b1]{width:12px;height:1px;border-radius:8px;background:currentColor;position:absolute;transition:transform .3s ease-in-out}.toggle-icon-sun-ray[data-v-8bd002b1]:nth-of-type(2){transform:rotate(90deg)}.toggle-icon-sun-ray[data-v-8bd002b1]:nth-of-type(3){transform:rotate(45deg)}.toggle-icon-sun-ray[data-v-8bd002b1]:nth-of-type(4){transform:rotate(-45deg)}.toggle-icon-dark .toggle-icon-ellipse[data-v-8bd002b1]{width:10px;height:10px;-webkit-mask-image:radial-gradient(circle at bottom left,pink 10px,transparent 12px);mask-image:radial-gradient(circle at bottom left,pink 10px,transparent 12px)}.toggle-icon-dark .toggle-icon-sun-ray[data-v-8bd002b1]{transform:scale(0)}.toggle-icon-dark .toggle-icon-moon-mask[data-v-8bd002b1]{transform:translateZ(0)}.scalar-app code.hljs *{font-size:var(--scalar-small);font-family:var(--scalar-font-code);text-align:left;white-space:pre;word-spacing:normal;word-break:normal;word-wrap:normal;line-height:1.4;-moz-tab-size:4;-o-tab-size:4;tab-size:4}.scalar-app code.hljs{all:unset;font-size:var(--scalar-small);color:var(--scalar-color-2);font-family:var(--scalar-font-code);display:inline-block;counter-reset:linenumber}.scalar-app .hljs{background:transparent;color:var(--scalar-color-2)}.scalar-app .hljs .line:before{color:var(--scalar-color-3);display:inline-block;counter-increment:linenumber;content:counter(linenumber);margin-right:.7em;min-width:calc(var(--line-digits) * 1ch);text-align:right}.scalar-app .hljs-comment,.scalar-app .hljs-quote{color:var(--scalar-color-3);font-style:italic}.scalar-app .hljs-number{color:var(--scalar-color-orange)}.scalar-app .hljs-regexp,.scalar-app .hljs-string,.scalar-app .hljs-built_in{color:var(--scalar-color-blue)}.scalar-app .hljs-title.class_{color:var(--scalar-color-1)}.scalar-app .hljs-keyword{color:var(--scalar-color-purple)}.scalar-app .hljs-title.function_{color:var(--scalar-color-orange)}.scalar-app .hljs-subst,.scalar-app .hljs-name{color:var(--scalar-color-blue)}.scalar-app .hljs-attr,.scalar-app .hljs-attribute{color:var(--scalar-color-1)}.scalar-app .hljs-addition,.scalar-app .hljs-literal,.scalar-app .hljs-selector-tag,.scalar-app .hljs-type{color:var(--scalar-color-green)}.scalar-app .hljs-selector-attr,.scalar-app .hljs-selector-pseudo{color:var(--scalar-color-orange)}.scalar-app .hljs-doctag,.scalar-app .hljs-section,.scalar-app .hljs-title{color:var(--scalar-color-blue)}.scalar-app .hljs-selector-id,.scalar-app .hljs-template-variable,.scalar-app .hljs-variable{color:var(--scalar-color-1)}.scalar-app .hljs-name,.scalar-app .hljs-section,.scalar-app .hljs-strong{font-weight:var(--scalar-semibold)}.scalar-app .hljs-bullet,.scalar-app .hljs-link,.scalar-app .hljs-meta,.scalar-app .hljs-symbol{color:var(--scalar-color-blue)}.scalar-app .hljs-deletion{color:var(--scalar-color-red)}.scalar-app .hljs-formula{background:var(--scalar-color-1)}.scalar-app .hljs-emphasis{font-style:italic}.scalar-app .credentials{font-size:0!important;color:transparent}.scalar-app .credentials:after{content:"·····";font-size:var(--scalar-small);color:var(--scalar-color-3);-webkit-user-select:none;-moz-user-select:none;user-select:none}.hljs.language-html{color:var(--scalar-color-1)}.hljs.language-html .hljs-attr{color:var(--scalar-color-2)}.hljs.language-curl .hljs-keyword{color:var(--scalar-color-orange)}.hljs.language-curl .hljs-string{color:var(--scalar-color-1)}.hljs.language-curl .hljs-literal,.hljs.language-php .hljs-variable{color:var(--scalar-color-blue)}.hljs.language-objectivec .hljs-meta{color:var(--scalar-color-1)}.hljs.language-objectivec .hljs-built_in,.hljs-built_in{color:var(--scalar-color-orange)}.scalar-app .markdown{font-family:var(--scalar-font);color:var(--scalar-color-1);word-break:break-word}.scalar-app .markdown h1,.scalar-app .markdown h2,.scalar-app .markdown h3,.scalar-app .markdown h4,.scalar-app .markdown h5,.scalar-app .markdown h6,.scalar-app .markdown p,.scalar-app .markdown div,.scalar-app .markdown img,.scalar-app .markdown details,.scalar-app .markdown summary,.scalar-app .markdown ul,.scalar-app .markdown ol,.scalar-app .markdown table,.scalar-app .markdown blockquote,.scalar-app .markdown code{margin:12px 0}.scalar-app .markdown details{margin:12px 0;color:var(--scalar-color-1)}.scalar-app .markdown summary{display:block;margin:1em 0;padding-left:20px;position:relative;font-weight:var(--scalar-semibold);cursor:pointer;-webkit-user-select:none;-moz-user-select:none;user-select:none}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url(\'data:image/svg+xml,<svg viewBox="0 0 24 24" fill="none" xmlns="http://www.w3.org/2000/svg"><path d="m9 18 6-6-6-6" stroke="currentColor" stroke-width="1.5" stroke-linecap="round" stroke-linejoin="round"/></svg>\');mask-image:url(\'data:image/svg+xml,<svg viewBox="0 0 24 24" fill="none" xmlns="http://www.w3.org/2000/svg"><path d="m9 18 6-6-6-6" stroke="currentColor" stroke-width="1.5" stroke-linecap="round" stroke-linejoin="round"/></svg>\')}.scalar-app .markdown summary:hover:after{background-color:var(--scalar-color-1)}.scalar-app .markdown details[open] summary:after{transform:rotate(90deg)}.scalar-app .markdown summary::-webkit-details-marker{display:none}.scalar-app .markdown img{overflow:hidden;border-radius:var(--scalar-radius);max-width:100%}.scalar-app .markdown>:first-child{margin-top:0;margin-bottom:0}.scalar-app .markdown h1{--font-size: 1.4em}.scalar-app .markdown h2{--font-size: 1.25em}.scalar-app .markdown h3{--font-size: 1.1em}.scalar-app .markdown h4,.scalar-app .markdown h6{--font-size: 1em}.scalar-app .markdown h1,.scalar-app .markdown h2,.scalar-app .markdown h3,.scalar-app .markdown h4,.scalar-app .markdown h5,.scalar-app .markdown h6{font-size:var(--font-size);margin:18px 0 6px;font-weight:var(--scalar-bold);display:block;line-height:1.45}.scalar-app .markdown b,.scalar-app .markdown strong{font-weight:var(--scalar-bold)}.scalar-app .markdown p{color:inherit;font-weight:var(--font-weight, var(--scalar-regular));line-height:1.5;margin-bottom:0;display:block}.scalar-app .markdown ul,.scalar-app .markdown ol{padding-left:24px;line-height:1.5;margin:12px 0;display:block}.scalar-app .markdown ul{list-style:disc}.scalar-app .markdown ol{list-style:decimal}.scalar-app .markdown ul.contains-task-list{list-style:none;padding-left:0}.scalar-app .markdown li{margin:6px 0;display:list-item}.scalar-app .markdown ul:first-of-type li:first-of-type{margin-top:0}.scalar-app .markdown a{color:var(--scalar-link-color, var(--scalar-color-accent));font-weight:var(--scalar-link-font-weight, inherit);-webkit-text-decoration:var(--scalar-text-decoration);text-decoration:var(--scalar-text-decoration);cursor:pointer}.scalar-app .markdown a:hover{color:var(--scalar-link-color-hover, var(--scalar-color-accent));-webkit-text-decoration:var(--scalar-text-decoration-hover);text-decoration:var(--scalar-text-decoration-hover)}.scalar-app .markdown em{font-style:italic}.scalar-app .markdown sup{font-size:var(--scalar-micro);vertical-align:super;font-weight:450}.scalar-app .markdown sub{font-size:var(--scalar-micro);vertical-align:sub;font-weight:450}.scalar-app .markdown del{text-decoration:line-through}.scalar-app .markdown code{font-family:var(--scalar-font-code);background-color:var(--scalar-background-2);box-shadow:0 0 0 var(--scalar-border-width) var(--scalar-border-color);font-size:var(--scalar-micro);border-radius:2px;padding:0 3px}.scalar-app .markdown .hljs{font-size:var(--scalar-small)}.scalar-app .markdown pre code{display:block;white-space:pre;padding:12px;line-height:1.5;margin:12px 0;-webkit-overflow-scrolling:touch;overflow-x:auto;max-width:100%;min-width:100px}.scalar-app .markdown hr{border:none;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.scalar-app .markdown blockquote{border-left:3px solid var(--scalar-border-color);padding-left:12px;margin:0;display:block}.scalar-app .markdown table{display:block;overflow-x:auto;position:relative;border-collapse:collapse;width:-moz-max-content;width:max-content;max-width:100%;margin:1em 0;box-shadow:0 0 0 var(--scalar-border-width) var(--scalar-border-color);border-radius:var(--scalar-radius-lg)}.scalar-app .markdown tbody{display:table-row-group;vertical-align:middle}.scalar-app .markdown thead{display:table-header-group;vertical-align:middle}.scalar-app .markdown tr{display:table-row;border-color:inherit;vertical-align:inherit}.scalar-app .markdown td,.scalar-app .markdown th{display:table-cell;vertical-align:inherit;min-width:1em;padding:6px 9px;vertical-align:top;line-height:1.5;position:relative;word-break:initial;font-size:var(--scalar-small);color:var(--scalar-color-1);font-weight:var(--font-weight, var(--scalar-regular));border-right:var(--scalar-border-width) solid var(--scalar-border-color);border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.scalar-app .markdown td>*,.scalar-app .markdown th>*{margin-bottom:0}.scalar-app .markdown th:empty{display:none}.scalar-app .markdown td:first-of-type,.scalar-app .markdown th:first-of-type{border-left:none}.scalar-app .markdown td:last-of-type,.scalar-app .markdown th:last-of-type{border-right:none}.scalar-app .markdown tr:last-of-type td{border-bottom:none}.scalar-app .markdown th{font-weight:var(--scalar-semibold)!important;text-align:left;border-left-color:transparent;background:var(--scalar-background-2)}.scalar-app .markdown tr>[align=left]{text-align:left}.scalar-app .markdown tr>[align=right]{text-align:right}.scalar-app .markdown tr>[align=center]{text-align:center}.dark-mode .scalar-dropdown-item[data-v-942bce7e]:hover,.dark-mode .scalar-dropdown-item[data-v-457c1cd3]:hover{filter:brightness(1.1)}.scalar-modal-layout[data-v-0f8e0211]{animation:fadein-layout-0f8e0211 ease-in-out .3s forwards}.scalar-modal[data-v-0f8e0211]{animation:fadein-modal-0f8e0211 ease-in-out .3s forwards;animation-delay:.1s;position:fixed;left:0;top:0;right:0;box-shadow:var(--scalar-shadow-2);transform:translate3d(0,10px,0)}.scalar-modal-layout-full[data-v-0f8e0211]{opacity:1!important;background:transparent!important}.dark-mode .scalar-modal[data-v-0f8e0211]{background-color:color-mix(in srgb,var(--scalar-background-1),black)}.scalar-modal.scalar-modal-search[data-v-0f8e0211]{max-width:540px;max-height:440px;background-color:transparent}.modal-content-search .modal-body[data-v-0f8e0211]{padding:0;overflow:hidden;display:flex;flex-direction:column;max-height:440px}@media (max-width: 1280px){.scalar-modal[data-v-0f8e0211]{max-height:calc(100% - 56px);top:28px}}@keyframes fadein-layout-0f8e0211{0%{opacity:0}to{opacity:1}}@keyframes fadein-modal-0f8e0211{0%{opacity:0;transform:translate3d(0,10px,0)}to{opacity:1;transform:translateZ(0)}}.full-size-styles[data-v-0f8e0211]{transform:translateZ(0);animation:fadein-layout-0f8e0211 ease-in-out .3s forwards;max-height:100%!important;top:0!important;left:0;position:absolute!important;margin:initial;border-radius:0!important;background-color:var(--scalar-background-1)!important;box-shadow:none!important;border-right:var(--scalar-border-width) solid var(--scalar-border-color)}@media (min-width: 800px){.full-size-styles[data-v-0f8e0211]{width:50dvw!important}}.full-size-styles[data-v-0f8e0211]:after{content:"";width:50dvw;height:100dvh;position:absolute;right:-50dvw;top:0}.search-background[data-v-2f2316c6]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2))}.search-background[data-v-2f2316c6]:focus-within{background:transparent}.scalar-input-wrapper-focus .scalar-input-label[data-v-4de7a24e]{color:var(--scalar-color-1)}.scalar-input-wrapper-error .scalar-input-label[data-v-4de7a24e]{color:var(--scalar-color-red)}.scalar-input[data-v-4de7a24e]::-moz-selection{color:var(--scalar-color-1);background:#ffa55859}.scalar-input[data-v-4de7a24e]::selection{color:var(--scalar-color-1);background:#ffa55859}.scalar-input[data-v-4de7a24e]:has(+.scalar-input-label){opacity:0}.scalar-input[data-v-4de7a24e]:not(:-moz-placeholder-shown){opacity:1;-moz-transition:opacity .2s ease-in-out .15s;transition:opacity .2s ease-in-out .15s}.scalar-input[data-v-4de7a24e]:not(:placeholder-shown),.scalar-input-wrapper-focus .scalar-input[data-v-4de7a24e]{opacity:1;transition:opacity .2s ease-in-out .15s}.scalar-input[data-v-4de7a24e]:-webkit-autofill,.scalar-input[data-v-4de7a24e]:-webkit-autofill:hover,.scalar-input[data-v-4de7a24e]:-webkit-autofill:focus,.scalar-input[data-v-4de7a24e]:-webkit-autofill:active,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill:hover,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill:focus,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill:active{-webkit-box-shadow:0 0 0px 1000px var(--scalar-background-1) inset!important;-webkit-text-fill-color:var(--scalar-color-1);color:var(--scalar-color-1);border-radius:var(--scalar-radius)}.scalar-input:not(:-moz-placeholder-shown)+.scalar-input-label[data-v-4de7a24e]{transform:translate3d(0,-20px,0) scale(.8);transform-origin:top left}.scalar-input-wrapper-focus .scalar-input-label[data-v-4de7a24e],.scalar-input:not(:placeholder-shown)+.scalar-input-label[data-v-4de7a24e]{transform:translate3d(0,-20px,0) scale(.8);transform-origin:top left}.scalar-input-wrapper-focus:has(button:active) .scalar-input-label[data-v-4de7a24e]{color:var(--scalar-color-3)!important}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}}:where(.scalar-app){font-family:var(--scalar-font);line-height:1.15;-webkit-text-size-adjust:100%;-moz-tab-size:4;-o-tab-size:4;tab-size:4}:where(.scalar-app) *,:where(.scalar-app) *:before,:where(.scalar-app) *:after{box-sizing:border-box;border-width:0;border-style:solid;border-color:var(--scalar-border-color);outline-width:1px;outline-style:none;outline-color:var(--scalar-color-accent);font-family:inherit;font-feature-settings:inherit;font-variation-settings:inherit;font-size:inherit;font-weight:inherit;line-height:inherit;color:inherit;margin:unset;padding:unset;text-rendering:optimizeLegibility;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale}:where(.scalar-app) :before,:where(.scalar-app) :after{--tw-content: ""}:where(.scalar-app) button,:where(.scalar-app) input,:where(.scalar-app) optgroup,:where(.scalar-app) select,:where(.scalar-app) textarea,:where(.scalar-app) ::file-selector-button{background:transparent}:where(.scalar-app) ol,:where(.scalar-app) ul,:where(.scalar-app) menu{list-style:none}:where(.scalar-app) input:where(:not([type=button],[type=reset],[type=submit])),:where(.scalar-app) select,:where(.scalar-app) textarea{border-radius:var(--scalar-radius);border-width:1px}:where(.scalar-app) input::-moz-placeholder{color:var(--scalar-color-3);font-family:var(--scalar-font)}:where(.scalar-app) input::placeholder{color:var(--scalar-color-3);font-family:var(--scalar-font)}:where(.scalar-app) input[type=search]::-webkit-search-cancel-button,:where(.scalar-app) input[type=search]::-webkit-search-decoration{-webkit-appearance:none;-moz-appearance:none;appearance:none}:where(.scalar-app) input:-webkit-autofill{-webkit-background-clip:text!important;background-clip:text!important}:where(.scalar-app) :focus-visible{outline-style:solid;border-radius:var(--scalar-radius)}:where(.scalar-app) button:focus-visible,:where(.scalar-app) [role=button]:focus-visible{outline-offset:-1px}:where(.scalar-app) button,:where(.scalar-app) [role=button]{cursor:pointer}:where(.scalar-app) :disabled{cursor:default}:where(.scalar-app) img,:where(.scalar-app) svg,:where(.scalar-app) video,:where(.scalar-app) canvas,:where(.scalar-app) audio,:where(.scalar-app) iframe,:where(.scalar-app) embed,:where(.scalar-app) object{display:block;vertical-align:middle}:where(.scalar-app) [hidden]{display:none}.scalar-app .cm-scroller,.scalar-app .custom-scroll{overflow-y:auto;scrollbar-color:transparent transparent;scrollbar-width:thin;-webkit-overflow-scrolling:touch}.scalar-app .custom-scroll-self-contain-overflow{overscroll-behavior:contain}.scalar-app .cm-scroller:hover,.scalar-app .custom-scroll:hover,.scalar-app.scalar-scrollbars-obtrusive .cm-scroller,.scalar-app.scalar-scrollbars-obtrusive .custom-scroll{scrollbar-color:var(--scalar-scrollbar-color, transparent) transparent}.scalar-app .cm-scroller:hover::-webkit-scrollbar-thumb,.scalar-app .custom-scroll:hover::-webkit-scrollbar-thumb{background:var(--scalar-scrollbar-color);background-clip:content-box;border:3px solid transparent}.scalar-app .cm-scroller::-webkit-scrollbar-thumb:active,.scalar-app .custom-scroll::-webkit-scrollbar-thumb:active{background:var(--scalar-scrollbar-color-active);background-clip:content-box;border:3px solid transparent}.scalar-app .cm-scroller::-webkit-scrollbar-corner,.scalar-app .custom-scroll::-webkit-scrollbar-corner{background:transparent}.scalar-app .cm-scroller::-webkit-scrollbar,.scalar-app .custom-scroll::-webkit-scrollbar{height:12px;width:12px}.scalar-app .cm-scroller::-webkit-scrollbar-track,.scalar-app .custom-scroll::-webkit-scrollbar-track{background:transparent}.scalar-app .cm-scroller::-webkit-scrollbar-thumb,.scalar-app .custom-scroll::-webkit-scrollbar-thumb{border-radius:20px;background:transparent;background-clip:content-box;border:3px solid transparent}@media (pointer: coarse){.scalar-app .cm-scroller,.scalar-app .custom-scroll{padding-right:12px}}@layer scalar-base{:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-client{line-height:1.5;color:var(--scalar-color-1)}.scalar-client .container{width:100%}@media (min-width: 400px){.scalar-client .container{max-width:400px}}@media (min-width: 600px){.scalar-client .container{max-width:600px}}@media (min-width: 800px){.scalar-client .container{max-width:800px}}@media (min-width: 1000px){.scalar-client .container{max-width:1000px}}@media (min-width: 1200px){.scalar-client .container{max-width:1200px}}.scalar-client .row{display:flex;flex-direction:row}.scalar-client .row-1{display:flex;flex-direction:row;flex:1 1 0%}.scalar-client .col{display:flex;flex-direction:column}.scalar-client .col-1{display:flex;flex-direction:column;flex:1 1 0%}.scalar-client .flex-center{display:flex;justify-content:center;align-items:center}.scalar-client .centered-x{left:50%;--tw-translate-x: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .centered-y{top:50%;--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .centered{left:50%;--tw-translate-x: -50%;top:50%;--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .floating-bg:before{background-color:var(--scalar-background-2);border-radius:var(--scalar-radius);content:"";height:calc(100% - 4px);left:-4px;opacity:0;position:absolute;top:2.5px;transition:opacity .2s ease-in-out;width:calc(100% + 8px);z-index:1}.scalar-client .floating-bg:hover:before{opacity:1}.scalar-client .sr-only{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);white-space:nowrap;border-width:0}.scalar-client .not-sr-only{position:static;width:auto;height:auto;padding:0;margin:0;overflow:visible;clip:auto;white-space:normal}.scalar-client .pointer-events-none{pointer-events:none}.scalar-client .pointer-events-auto{pointer-events:auto}.scalar-client .visible{visibility:visible}.scalar-client .static{position:static}.scalar-client .fixed{position:fixed}.scalar-client .absolute{position:absolute}.scalar-client .relative{position:relative}.scalar-client .sticky{position:sticky}.scalar-client .inset-0{top:0;right:0;bottom:0;left:0}.scalar-client .-inset-x-96{left:-384px;right:-384px}.scalar-client .inset-x-1{left:4px;right:4px}.scalar-client .-right-\\[30px\\]{right:-30px}.scalar-client .-top-\\[--scalar-address-bar-height\\]{top:calc(var(--scalar-address-bar-height) * -1)}.scalar-client .bottom-0{bottom:0}.scalar-client .bottom-1\\/2{bottom:50%}.scalar-client .bottom-\\[var\\(--scalar-border-width\\)\\]{bottom:var(--scalar-border-width)}.scalar-client .left-0{left:0}.scalar-client .left-1\\/2{left:50%}.scalar-client .right-0{right:0}.scalar-client .right-1{right:4px}.scalar-client .right-1\\/2{right:50%}.scalar-client .right-14{right:56px}.scalar-client .right-16{right:64px}.scalar-client .right-2{right:8px}.scalar-client .right-4{right:16px}.scalar-client .right-7{right:28px}.scalar-client .top-0{top:0}.scalar-client .top-1\\/2{top:50%}.scalar-client .top-12{top:48px}.scalar-client .top-2{top:8px}.scalar-client .-z-1{z-index:-1}.scalar-client .z-0{z-index:0}.scalar-client .z-1{z-index:1}.scalar-client .z-10{z-index:10}.scalar-client .z-20{z-index:20}.scalar-client .z-50{z-index:50}.scalar-client .z-\\[10001\\]{z-index:10001}.scalar-client .z-\\[1\\]{z-index:1}.scalar-client .z-context{z-index:1000}.scalar-client .z-context-plus{z-index:1001}.scalar-client .z-overlay{z-index:10000}.scalar-client .order-last{order:9999}.scalar-client .col-span-full{grid-column:1 / -1}.scalar-client .\\!m-0{margin:0!important}.scalar-client .m-0{margin:0}.scalar-client .m-4{margin:16px}.scalar-client .m-auto{margin:auto}.scalar-client .\\!mx-0{margin-left:0!important;margin-right:0!important}.scalar-client .-mx-1{margin-left:-4px;margin-right:-4px}.scalar-client .-my-1{margin-top:-4px;margin-bottom:-4px}.scalar-client .my-1{margin-top:4px;margin-bottom:4px}.scalar-client .my-1\\.25{margin-top:5px;margin-bottom:5px}.scalar-client .my-12{margin-top:48px;margin-bottom:48px}.scalar-client .\\!mb-0{margin-bottom:0!important}.scalar-client .\\!mr-0{margin-right:0!important}.scalar-client .-mb-0{margin-bottom:-0px}.scalar-client .-mb-0\\.25{margin-bottom:-1px}.scalar-client .-mb-\\[var\\(--scalar-border-width\\)\\]{margin-bottom:calc(var(--scalar-border-width) * -1)}.scalar-client .-ml-0{margin-left:-0px}.scalar-client .-ml-0\\.5{margin-left:-2px}.scalar-client .-ml-12{margin-left:-48px}.scalar-client .-mr-1{margin-right:-4px}.scalar-client .-mr-1\\.5{margin-right:-6px}.scalar-client .-mr-3{margin-right:-12px}.scalar-client .-mt-0{margin-top:-0px}.scalar-client .-mt-0\\.25{margin-top:-1px}.scalar-client .-mt-\\[\\.5px\\]{margin-top:-.5px}.scalar-client .mb-0{margin-bottom:0}.scalar-client .mb-1{margin-bottom:4px}.scalar-client .mb-1\\.5{margin-bottom:6px}.scalar-client .mb-2{margin-bottom:8px}.scalar-client .mb-4{margin-bottom:16px}.scalar-client .mb-\\[\\.5px\\]{margin-bottom:.5px}.scalar-client .ml-0{margin-left:0}.scalar-client .ml-0\\.5{margin-left:2px}.scalar-client .ml-0\\.75{margin-left:3px}.scalar-client .ml-1{margin-left:4px}.scalar-client .ml-2{margin-left:8px}.scalar-client .ml-auto{margin-left:auto}.scalar-client .mr-0{margin-right:0}.scalar-client .mr-0\\.25{margin-right:1px}.scalar-client .mr-0\\.75{margin-right:3px}.scalar-client .mr-1{margin-right:4px}.scalar-client .mr-1\\.5{margin-right:6px}.scalar-client .mr-2{margin-right:8px}.scalar-client .mr-2\\.5{margin-right:10px}.scalar-client .mr-3{margin-right:12px}.scalar-client .mr-auto{margin-right:auto}.scalar-client .mt-1{margin-top:4px}.scalar-client .mt-10{margin-top:40px}.scalar-client .mt-2{margin-top:8px}.scalar-client .mt-3{margin-top:12px}.scalar-client .mt-5{margin-top:20px}.scalar-client .mt-auto{margin-top:auto}.scalar-client .line-clamp-1{overflow:hidden;display:-webkit-box;-webkit-box-orient:vertical;-webkit-line-clamp:1}.scalar-client .block{display:block}.scalar-client .inline-block{display:inline-block}.scalar-client .\\!flex{display:flex!important}.scalar-client .flex{display:flex}.scalar-client .inline-flex{display:inline-flex}.scalar-client .table{display:table}.scalar-client .grid{display:grid}.scalar-client .contents{display:contents}.scalar-client .\\!hidden{display:none!important}.scalar-client .hidden{display:none}.scalar-client .aspect-\\[4\\/3\\]{aspect-ratio:4/3}.scalar-client .aspect-square{aspect-ratio:1 / 1}.scalar-client .size-10{width:40px;height:40px}.scalar-client .size-2{width:8px;height:8px}.scalar-client .size-2\\.5{width:10px;height:10px}.scalar-client .size-3{width:12px;height:12px}.scalar-client .size-3\\.5{width:14px;height:14px}.scalar-client .size-3\\/4{width:75%;height:75%}.scalar-client .size-4{width:16px;height:16px}.scalar-client .size-7{width:28px;height:28px}.scalar-client .h-1{height:4px}.scalar-client .h-1\\.5{height:6px}.scalar-client .h-10{height:40px}.scalar-client .h-12{height:48px}.scalar-client .h-2{height:8px}.scalar-client .h-2\\.5{height:10px}.scalar-client .h-24{height:96px}.scalar-client .h-3{height:12px}.scalar-client .h-3\\.5{height:14px}.scalar-client .h-4{height:16px}.scalar-client .h-5{height:20px}.scalar-client .h-6{height:24px}.scalar-client .h-64{height:256px}.scalar-client .h-7{height:28px}.scalar-client .h-8{height:32px}.scalar-client .h-9{height:36px}.scalar-client .h-\\[--scalar-address-bar-height\\]{height:var(--scalar-address-bar-height)}.scalar-client .h-\\[68px\\]{height:68px}.scalar-client .h-\\[calc\\(100\\%-273\\.5px\\)\\]{height:calc(100% - 273.5px)}.scalar-client .h-\\[calc\\(100\\%_-_50px\\)\\]{height:calc(100% - 50px)}.scalar-client .h-\\[var\\(--scalar-border-width\\)\\]{height:var(--scalar-border-width)}.scalar-client .h-auto{height:auto}.scalar-client .h-fit{height:-moz-fit-content;height:fit-content}.scalar-client .h-full{height:100%}.scalar-client .h-header{height:48px}.scalar-client .h-px{height:1px}.scalar-client .h-screen{height:100vh}.scalar-client .\\!max-h-\\[initial\\]{max-height:initial!important}.scalar-client .max-h-40{max-height:160px}.scalar-client .max-h-8{max-height:32px}.scalar-client .max-h-\\[40dvh\\]{max-height:40dvh}.scalar-client .max-h-\\[50dvh\\]{max-height:50dvh}.scalar-client .max-h-\\[60svh\\]{max-height:60svh}.scalar-client .max-h-\\[calc\\(100\\%-32px\\)\\]{max-height:calc(100% - 32px)}.scalar-client .max-h-\\[inherit\\]{max-height:inherit}.scalar-client .max-h-fit{max-height:-moz-fit-content;max-height:fit-content}.scalar-client .max-h-screen{max-height:100vh}.scalar-client .min-h-0{min-height:0px}.scalar-client .min-h-10{min-height:40px}.scalar-client .min-h-11{min-height:44px}.scalar-client .min-h-12{min-height:48px}.scalar-client .min-h-16{min-height:64px}.scalar-client .min-h-20{min-height:80px}.scalar-client .min-h-8{min-height:32px}.scalar-client .min-h-fit{min-height:-moz-fit-content;min-height:fit-content}.scalar-client .\\!w-fit{width:-moz-fit-content!important;width:fit-content!important}.scalar-client .w-0{width:0px}.scalar-client .w-0\\.5{width:2px}.scalar-client .w-1{width:4px}.scalar-client .w-1\\.5{width:6px}.scalar-client .w-1\\/2{width:50%}.scalar-client .w-10{width:40px}.scalar-client .w-2{width:8px}.scalar-client .w-2\\.5{width:10px}.scalar-client .w-20{width:80px}.scalar-client .w-3{width:12px}.scalar-client .w-3\\.5{width:14px}.scalar-client .w-4{width:16px}.scalar-client .w-5{width:20px}.scalar-client .w-56{width:224px}.scalar-client .w-6{width:24px}.scalar-client .w-64{width:256px}.scalar-client .w-72{width:288px}.scalar-client .w-8{width:32px}.scalar-client .w-\\[60px\\]{width:60px}.scalar-client .w-\\[calc\\(100\\%-10px\\)\\]{width:calc(100% - 10px)}.scalar-client .w-\\[calc\\(100\\%_-_8px\\)\\]{width:calc(100% - 8px)}.scalar-client .w-\\[inherit\\]{width:inherit}.scalar-client .w-auto{width:auto}.scalar-client .w-dvw{width:100dvw}.scalar-client .w-fit{width:-moz-fit-content;width:fit-content}.scalar-client .w-full{width:100%}.scalar-client .w-max{width:-moz-max-content;width:max-content}.scalar-client .min-w-0{min-width:0px}.scalar-client .min-w-2{min-width:8px}.scalar-client .min-w-2\\.5{min-width:10px}.scalar-client .min-w-3{min-width:12px}.scalar-client .min-w-3\\.5{min-width:14px}.scalar-client .min-w-48{min-width:192px}.scalar-client .min-w-8{min-width:32px}.scalar-client .min-w-\\[296px\\]{min-width:296px}.scalar-client .min-w-\\[37px\\]{min-width:37px}.scalar-client .min-w-fit{min-width:-moz-fit-content;min-width:fit-content}.scalar-client .min-w-full{min-width:100%}.scalar-client .max-w-\\[100\\%\\]{max-width:100%}.scalar-client .max-w-\\[14px\\]{max-width:14px}.scalar-client .max-w-\\[150px\\]{max-width:150px}.scalar-client .max-w-\\[320px\\]{max-width:320px}.scalar-client .max-w-\\[37px\\]{max-width:37px}.scalar-client .max-w-\\[380px\\]{max-width:380px}.scalar-client .max-w-\\[420px\\]{max-width:420px}.scalar-client .max-w-\\[720px\\]{max-width:720px}.scalar-client .max-w-\\[calc\\(100dvw-24px\\)\\]{max-width:calc(100dvw - 24px)}.scalar-client .max-w-full{max-width:100%}.scalar-client .flex-1{flex:1 1 0%}.scalar-client .shrink-0{flex-shrink:0}.scalar-client .flex-grow{flex-grow:1}.scalar-client .-translate-x-1\\/2{--tw-translate-x:-50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .-translate-y-1\\/2{--tw-translate-y:-50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .translate-x-0{--tw-translate-x:0px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .translate-x-1\\/2{--tw-translate-x:50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .translate-y-1\\/2{--tw-translate-y:50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .rotate-90{--tw-rotate:90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .scale-75{--tw-scale-x:.75;--tw-scale-y:.75;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .transform{transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .cursor-auto{cursor:auto}.scalar-client .cursor-default{cursor:default}.scalar-client .cursor-help{cursor:help}.scalar-client .cursor-pointer{cursor:pointer}.scalar-client .cursor-text{cursor:text}.scalar-client .select-none{-webkit-user-select:none;-moz-user-select:none;user-select:none}.scalar-client .resize-none{resize:none}.scalar-client .resize{resize:both}.scalar-client .auto-rows-\\[32px\\]{grid-auto-rows:32px}.scalar-client .auto-rows-auto{grid-auto-rows:auto}.scalar-client .grid-cols-2{grid-template-columns:repeat(2,minmax(0,1fr))}.scalar-client .grid-cols-\\[44px\\,1fr\\,repeat\\(3\\,auto\\)\\]{grid-template-columns:44px 1fr repeat(3,auto)}.scalar-client .grid-cols-\\[auto\\,1fr\\]{grid-template-columns:auto 1fr}.scalar-client .grid-cols-\\[repeat\\(auto-fill\\,minmax\\(32px\\,1fr\\)\\)\\]{grid-template-columns:repeat(auto-fill,minmax(32px,1fr))}.scalar-client .flex-row{flex-direction:row}.scalar-client .flex-col{flex-direction:column}.scalar-client .flex-wrap{flex-wrap:wrap}.scalar-client .content-start{align-content:flex-start}.scalar-client .content-between{align-content:space-between}.scalar-client .items-start{align-items:flex-start}.scalar-client .items-end{align-items:flex-end}.scalar-client .items-center{align-items:center}.scalar-client .items-stretch{align-items:stretch}.scalar-client .justify-start{justify-content:flex-start}.scalar-client .justify-end{justify-content:flex-end}.scalar-client .justify-center{justify-content:center}.scalar-client .justify-between{justify-content:space-between}.scalar-client .justify-stretch{justify-content:stretch}.scalar-client .\\!gap-2{gap:8px!important}.scalar-client .gap-0{gap:0px}.scalar-client .gap-0\\.5{gap:2px}.scalar-client .gap-0\\.75{gap:3px}.scalar-client .gap-1{gap:4px}.scalar-client .gap-1\\.5{gap:6px}.scalar-client .gap-1\\/2{gap:.5px}.scalar-client .gap-10{gap:40px}.scalar-client .gap-2{gap:8px}.scalar-client .gap-2\\.5{gap:10px}.scalar-client .gap-3{gap:12px}.scalar-client .gap-4{gap:16px}.scalar-client .gap-6{gap:24px}.scalar-client .gap-8{gap:32px}.scalar-client .gap-\\[1\\.5px\\]{gap:1.5px}.scalar-client .gap-px{gap:1px}.scalar-client .gap-x-2{-moz-column-gap:8px;column-gap:8px}.scalar-client .gap-x-2\\.5{-moz-column-gap:10px;column-gap:10px}.scalar-client .space-x-1>:not([hidden])~:not([hidden]){--tw-space-x-reverse:0;margin-right:calc(4px * var(--tw-space-x-reverse));margin-left:calc(4px * calc(1 - var(--tw-space-x-reverse)))}.scalar-client .divide-x>:not([hidden])~:not([hidden]){--tw-divide-x-reverse:0;border-right-width:calc(var(--scalar-border-width) * var(--tw-divide-x-reverse));border-left-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-x-reverse)))}.scalar-client .divide-y>:not([hidden])~:not([hidden]){--tw-divide-y-reverse:0;border-top-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-y-reverse)));border-bottom-width:calc(var(--scalar-border-width) * var(--tw-divide-y-reverse))}.scalar-client .self-center{align-self:center}.scalar-client .justify-self-end{justify-self:end}.scalar-client .overflow-auto{overflow:auto}.scalar-client .overflow-hidden{overflow:hidden}.scalar-client .overflow-visible{overflow:visible}.scalar-client .overflow-x-auto{overflow-x:auto}.scalar-client .overflow-y-auto{overflow-y:auto}.scalar-client .truncate{overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.scalar-client .text-ellipsis{text-overflow:ellipsis}.scalar-client .whitespace-nowrap{white-space:nowrap}.scalar-client .whitespace-pre{white-space:pre}.scalar-client .text-balance{text-wrap:balance}.scalar-client .text-pretty{text-wrap:pretty}.scalar-client .break-words{overflow-wrap:break-word}.scalar-client .break-all{word-break:break-all}.scalar-client .\\!rounded-none{border-radius:0!important}.scalar-client .rounded{border-radius:var(--scalar-radius)}.scalar-client .rounded-\\[10px\\]{border-radius:10px}.scalar-client .rounded-full{border-radius:9999px}.scalar-client .rounded-lg{border-radius:var(--scalar-radius-lg)}.scalar-client .rounded-md{border-radius:var(--scalar-radius)}.scalar-client .rounded-px{border-radius:1px}.scalar-client .rounded-xl{border-radius:var(--scalar-radius-xl)}.scalar-client .rounded-b{border-bottom-right-radius:var(--scalar-radius);border-bottom-left-radius:var(--scalar-radius)}.scalar-client .rounded-b-lg{border-bottom-right-radius:var(--scalar-radius-lg);border-bottom-left-radius:var(--scalar-radius-lg)}.scalar-client .rounded-t{border-top-left-radius:var(--scalar-radius);border-top-right-radius:var(--scalar-radius)}.scalar-client .rounded-t-lg{border-top-left-radius:var(--scalar-radius-lg);border-top-right-radius:var(--scalar-radius-lg)}.scalar-client .rounded-t-none{border-top-left-radius:0;border-top-right-radius:0}.scalar-client .\\!border-0{border-width:0!important}.scalar-client .border{border-width:var(--scalar-border-width)}.scalar-client .border-0{border-width:0}.scalar-client .border-1\\/2{border-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-\\[1\\.5px\\]{border-width:1.5px}.scalar-client .border-\\[1px\\]{border-width:1px}.scalar-client .\\!border-r-1\\/2{border-right-width:calc(var(--scalar-border-width) / 2)!important}.scalar-client .border-b{border-bottom-width:var(--scalar-border-width)}.scalar-client .border-b-0{border-bottom-width:0}.scalar-client .border-b-1\\/2{border-bottom-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-b-\\[1px\\]{border-bottom-width:1px}.scalar-client .border-l{border-left-width:var(--scalar-border-width)}.scalar-client .border-l-0{border-left-width:0}.scalar-client .border-r{border-right-width:var(--scalar-border-width)}.scalar-client .border-r-0{border-right-width:0}.scalar-client .border-r-1\\/2{border-right-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-t{border-top-width:var(--scalar-border-width)}.scalar-client .border-t-0{border-top-width:0}.scalar-client .border-t-1\\/2{border-top-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-solid{border-style:solid}.scalar-client .border-dashed{border-style:dashed}.scalar-client .border-none{border-style:none}.scalar-client .\\!border-current{border-color:currentColor!important}.scalar-client .border-b-3{border-color:var(--scalar-background-3)}.scalar-client .border-c-3{border-color:var(--scalar-color-3)}.scalar-client .border-transparent{border-color:transparent}.scalar-client .border-r-transparent{border-right-color:transparent}.scalar-client .bg-b-1{background-color:var(--scalar-background-1)}.scalar-client .bg-b-2{background-color:var(--scalar-background-2)}.scalar-client .bg-b-3{background-color:var(--scalar-background-3)}.scalar-client .bg-blue{background-color:var(--scalar-color-blue)}.scalar-client .bg-border{background-color:var(--scalar-border-color)}.scalar-client .bg-c-2{background-color:var(--scalar-color-2)}.scalar-client .bg-c-3{background-color:var(--scalar-color-3)}.scalar-client .bg-c-accent{background-color:var(--scalar-color-accent)}.scalar-client .bg-current{background-color:currentColor}.scalar-client .bg-green{background-color:var(--scalar-color-green)}.scalar-client .bg-grey{background-color:var(--scalar-color-3)}.scalar-client .bg-orange{background-color:var(--scalar-color-orange)}.scalar-client .bg-purple{background-color:var(--scalar-color-purple)}.scalar-client .bg-red{background-color:var(--scalar-color-red)}.scalar-client .bg-sidebar-active-b{background-color:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .bg-yellow{background-color:var(--scalar-color-yellow)}.scalar-client .bg-none{background-image:none}.scalar-client .fill-current{fill:currentColor}.scalar-client .stroke-2{stroke-width:2}.scalar-client .stroke-\\[1\\.5\\]{stroke-width:1.5}.scalar-client .stroke-\\[1\\.75\\]{stroke-width:1.75}.scalar-client .stroke-\\[2\\.25\\]{stroke-width:2.25}.scalar-client .object-contain{-o-object-fit:contain;object-fit:contain}.scalar-client .\\!p-0{padding:0!important}.scalar-client .p-0{padding:0}.scalar-client .p-0\\.5{padding:2px}.scalar-client .p-0\\.75{padding:3px}.scalar-client .p-1{padding:4px}.scalar-client .p-1\\.5{padding:6px}.scalar-client .p-1\\.75{padding:7px}.scalar-client .p-2{padding:8px}.scalar-client .p-3{padding:12px}.scalar-client .p-4{padding:16px}.scalar-client .p-\\[3px\\]{padding:3px}.scalar-client .p-\\[5px\\]{padding:5px}.scalar-client .\\!px-3{padding-left:12px!important;padding-right:12px!important}.scalar-client .\\!py-1{padding-top:4px!important;padding-bottom:4px!important}.scalar-client .\\!py-1\\.5{padding-top:6px!important;padding-bottom:6px!important}.scalar-client .px-0{padding-left:0;padding-right:0}.scalar-client .px-0\\.5{padding-left:2px;padding-right:2px}.scalar-client .px-1{padding-left:4px;padding-right:4px}.scalar-client .px-1\\.5{padding-left:6px;padding-right:6px}.scalar-client .px-2{padding-left:8px;padding-right:8px}.scalar-client .px-2\\.5{padding-left:10px;padding-right:10px}.scalar-client .px-3{padding-left:12px;padding-right:12px}.scalar-client .px-4{padding-left:16px;padding-right:16px}.scalar-client .px-5{padding-left:20px;padding-right:20px}.scalar-client .px-6{padding-left:24px;padding-right:24px}.scalar-client .px-8{padding-left:32px;padding-right:32px}.scalar-client .py-0{padding-top:0;padding-bottom:0}.scalar-client .py-0\\.5{padding-top:2px;padding-bottom:2px}.scalar-client .py-0\\.75{padding-top:3px;padding-bottom:3px}.scalar-client .py-1{padding-top:4px;padding-bottom:4px}.scalar-client .py-1\\.25{padding-top:5px;padding-bottom:5px}.scalar-client .py-1\\.5{padding-top:6px;padding-bottom:6px}.scalar-client .py-2{padding-top:8px;padding-bottom:8px}.scalar-client .py-2\\.5{padding-top:10px;padding-bottom:10px}.scalar-client .py-3{padding-top:12px;padding-bottom:12px}.scalar-client .py-5{padding-top:20px;padding-bottom:20px}.scalar-client .py-8{padding-top:32px;padding-bottom:32px}.scalar-client .py-px{padding-top:1px;padding-bottom:1px}.scalar-client .\\!pl-3{padding-left:12px!important}.scalar-client .\\!pt-0{padding-top:0!important}.scalar-client .pb-0{padding-bottom:0}.scalar-client .pb-14{padding-bottom:56px}.scalar-client .pb-2{padding-bottom:8px}.scalar-client .pb-3{padding-bottom:12px}.scalar-client .pb-5{padding-bottom:20px}.scalar-client .pb-8{padding-bottom:32px}.scalar-client .pb-\\[75px\\]{padding-bottom:75px}.scalar-client .pl-1{padding-left:4px}.scalar-client .pl-1\\.25{padding-left:5px}.scalar-client .pl-1\\.5{padding-left:6px}.scalar-client .pl-2{padding-left:8px}.scalar-client .pl-3{padding-left:12px}.scalar-client .pl-6{padding-left:24px}.scalar-client .pl-8{padding-left:32px}.scalar-client .pl-9{padding-left:36px}.scalar-client .pl-px{padding-left:1px}.scalar-client .pr-0{padding-right:0}.scalar-client .pr-1{padding-right:4px}.scalar-client .pr-1\\.5{padding-right:6px}.scalar-client .pr-2{padding-right:8px}.scalar-client .pr-2\\.5{padding-right:10px}.scalar-client .pr-4{padding-right:16px}.scalar-client .pr-6{padding-right:24px}.scalar-client .pr-\\[26px\\]{padding-right:26px}.scalar-client .pr-px{padding-right:1px}.scalar-client .pt-0{padding-top:0}.scalar-client .pt-2{padding-top:8px}.scalar-client .pt-3{padding-top:12px}.scalar-client .pt-4{padding-top:16px}.scalar-client .pt-px{padding-top:1px}.scalar-client .text-left{text-align:left}.scalar-client .text-center{text-align:center}.scalar-client .text-right{text-align:right}.scalar-client .font-code{font-family:var(--scalar-font-code)}.scalar-client .text-3xs{font-size:var(--scalar-font-size-7)}.scalar-client .text-\\[11px\\]{font-size:11px}.scalar-client .text-\\[21px\\]{font-size:21px}.scalar-client .text-\\[6px\\]{font-size:6px}.scalar-client .text-base{font-size:var(--scalar-font-size-3)}.scalar-client .text-sm{font-size:var(--scalar-font-size-4)}.scalar-client .text-xl{font-size:var(--scalar-font-size-1)}.scalar-client .text-xs{font-size:var(--scalar-font-size-5)}.scalar-client .text-xxs{font-size:var(--scalar-font-size-6)}.scalar-client .\\!font-normal{font-weight:var(--scalar-regular)!important}.scalar-client .font-bold{font-weight:var(--scalar-bold)}.scalar-client .font-medium{font-weight:var(--scalar-semibold)}.scalar-client .font-normal{font-weight:var(--scalar-regular)}.scalar-client .uppercase{text-transform:uppercase}.scalar-client .lowercase{text-transform:lowercase}.scalar-client .capitalize{text-transform:capitalize}.scalar-client .\\!leading-\\[6px\\]{line-height:6px!important}.scalar-client .leading{line-height:1.5}.scalar-client .leading-3{line-height:var(--scalar-line-height-3)}.scalar-client .leading-5{line-height:var(--scalar-line-height-5)}.scalar-client .leading-\\[1\\.44\\]{line-height:1.44}.scalar-client .leading-\\[21px\\]{line-height:21px}.scalar-client .leading-\\[7px\\]{line-height:7px}.scalar-client .leading-\\[normal\\]{line-height:normal}.scalar-client .leading-none{line-height:1}.scalar-client .leading-normal{line-height:1.5}.scalar-client .leading-snug{line-height:1.375}.scalar-client .\\!text-c-1{color:var(--scalar-color-1)!important}.scalar-client .text-b-1{color:var(--scalar-background-1)}.scalar-client .text-blue{color:var(--scalar-color-blue)}.scalar-client .text-border{color:var(--scalar-border-color)}.scalar-client .text-c-1{color:var(--scalar-color-1)}.scalar-client .text-c-2{color:var(--scalar-color-2)}.scalar-client .text-c-3{color:var(--scalar-color-3)}.scalar-client .text-c-btn{color:var(--scalar-button-1-color)}.scalar-client .text-green{color:var(--scalar-color-green)}.scalar-client .text-grey{color:var(--scalar-color-3)}.scalar-client .text-orange{color:var(--scalar-color-orange)}.scalar-client .text-purple{color:var(--scalar-color-purple)}.scalar-client .text-red{color:var(--scalar-color-red)}.scalar-client .text-sidebar-active-c{color:var(--scalar-sidebar-color-active, currentColor)}.scalar-client .text-sidebar-c-2{color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}.scalar-client .text-transparent{color:transparent}.scalar-client .text-yellow{color:var(--scalar-color-yellow)}.scalar-client .underline{text-decoration-line:underline}.scalar-client .no-underline{text-decoration-line:none}.scalar-client .decoration-c-3{text-decoration-color:var(--scalar-color-3)}.scalar-client .underline-offset-2{text-underline-offset:2px}.scalar-client .opacity-0{opacity:0}.scalar-client .opacity-100{opacity:1}.scalar-client .opacity-25{opacity:.25}.scalar-client .shadow{--tw-shadow:var(--scalar-shadow-1);--tw-shadow-colored:var(--scalar-shadow-1);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-\\[-8px_0_4px_var\\(--scalar-background-1\\)\\]{--tw-shadow:-8px 0 4px var(--scalar-background-1);--tw-shadow-colored:-8px 0 4px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-\\[0_-8px_0_8px_var\\(--scalar-background-1\\)\\,0_0_8px_8px_var\\(--scalar-background-1\\)\\]{--tw-shadow:0 -8px 0 8px var(--scalar-background-1),0 0 8px 8px var(--scalar-background-1);--tw-shadow-colored:0 -8px 0 8px var(--tw-shadow-color), 0 0 8px 8px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-border{--tw-shadow:inset 0 0 0 1px var(--scalar-border-color);--tw-shadow-colored:inset 0 0 0 1px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-lg{--tw-shadow:var(--scalar-shadow-2);--tw-shadow-colored:var(--scalar-shadow-2);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-none{--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-border{--tw-shadow-color:var(--scalar-border-color);--tw-shadow:var(--tw-shadow-colored)}.scalar-client .outline-none{outline:2px solid transparent;outline-offset:2px}.scalar-client .outline{outline-style:solid}.scalar-client .-outline-offset-1{outline-offset:-1px}.scalar-client .-outline-offset-2{outline-offset:-2px}.scalar-client .outline-offset-2{outline-offset:2px}.scalar-client .blur{--tw-blur:blur(8px);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .brightness-\\[\\.9\\]{--tw-brightness:brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .brightness-lifted{--tw-brightness:brightness(var(--scalar-lifted-brightness));filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .filter{filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .backdrop-filter{-webkit-backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia);backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia)}.scalar-client .transition{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,-webkit-backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter,-webkit-backdrop-filter;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .transition-colors{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .transition-none{transition-property:none}.scalar-client .transition-opacity{transition-property:opacity;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .transition-transform{transition-property:transform;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .duration-150{transition-duration:.15s}.scalar-client .duration-200{transition-duration:.2s}.scalar-client .duration-300{transition-duration:.3s}.scalar-client .ease-in-out{transition-timing-function:cubic-bezier(.4,0,.2,1)}.scalar-client .bg-mix-transparent{background-color:color-mix(var(--tw-bg-mix-method, in srgb),transparent calc(var(--tw-bg-mix-amount, 0) * 1%),var(--tw-bg-base))}.scalar-client .bg-mix-amount-90{--tw-bg-mix-amount:90}.scalar-client .bg-mix-amount-95{--tw-bg-mix-amount:95}.scalar-client .bg-b-1{--tw-bg-base:var(--scalar-background-1)}.scalar-client .bg-b-2{--tw-bg-base:var(--scalar-background-2)}.scalar-client .bg-b-3{--tw-bg-base:var(--scalar-background-3)}.scalar-client .bg-blue{--tw-bg-base:var(--scalar-color-blue)}.scalar-client .bg-border{--tw-bg-base:var(--scalar-border-color)}.scalar-client .bg-c-2{--tw-bg-base:var(--scalar-color-2)}.scalar-client .bg-c-3{--tw-bg-base:var(--scalar-color-3)}.scalar-client .bg-c-accent{--tw-bg-base:var(--scalar-color-accent)}.scalar-client .bg-current{--tw-bg-base:currentColor}.scalar-client .bg-green{--tw-bg-base:var(--scalar-color-green)}.scalar-client .bg-grey{--tw-bg-base:var(--scalar-color-3)}.scalar-client .bg-orange{--tw-bg-base:var(--scalar-color-orange)}.scalar-client .bg-purple{--tw-bg-base:var(--scalar-color-purple)}.scalar-client .bg-red{--tw-bg-base:var(--scalar-color-red)}.scalar-client .bg-sidebar-active-b{--tw-bg-base:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .bg-yellow{--tw-bg-base:var(--scalar-color-yellow)}.scalar-client .app-drag-region{-webkit-app-region:drag}.scalar-client .app-no-drag-region{-webkit-app-region:no-drag}.scalar-client .\\[--scalar-address-bar-height\\:32px\\]{--scalar-address-bar-height:32px}.scalar-client .\\[--scalar-small\\:--scalar-font-size-4\\]{--scalar-small:var(--scalar-font-size-4)}.scalar-client .\\*\\:flex>*{display:flex}.scalar-client .\\*\\:h-8>*{height:32px}.scalar-client .\\*\\:cursor-pointer>*{cursor:pointer}.scalar-client .\\*\\:items-center>*{align-items:center}.scalar-client .\\*\\:rounded-none>*{border-radius:0}.scalar-client .\\*\\:border-t-1\\/2>*{border-top-width:calc(var(--scalar-border-width) / 2)}.scalar-client .\\*\\:px-1>*{padding-left:4px;padding-right:4px}.scalar-client .\\*\\:px-1\\.5>*{padding-left:6px;padding-right:6px}.scalar-client .\\*\\:pl-4>*{padding-left:16px}.scalar-client .before\\:pointer-events-none:before{content:var(--tw-content);pointer-events:none}.scalar-client .before\\:absolute:before{content:var(--tw-content);position:absolute}.scalar-client .before\\:left-3:before{content:var(--tw-content);left:12px}.scalar-client .before\\:left-\\[calc\\(\\.75rem_\\+_\\.5px\\)\\]:before{content:var(--tw-content);left:calc(.75rem + .5px)}.scalar-client .before\\:right-\\[23px\\]:before{content:var(--tw-content);right:23px}.scalar-client .before\\:top-0:before{content:var(--tw-content);top:0}.scalar-client .before\\:top-0\\.5:before{content:var(--tw-content);top:2px}.scalar-client .before\\:z-1:before{content:var(--tw-content);z-index:1}.scalar-client .before\\:h-\\[calc\\(100\\%_\\+_\\.5px\\)\\]:before{content:var(--tw-content);height:calc(100% + .5px)}.scalar-client .before\\:min-h-\\[calc\\(100\\%-4px\\)\\]:before{content:var(--tw-content);min-height:calc(100% - 4px)}.scalar-client .before\\:w-3:before{content:var(--tw-content);width:12px}.scalar-client .before\\:w-\\[\\.5px\\]:before{content:var(--tw-content);width:.5px}.scalar-client .before\\:bg-border:before{content:var(--tw-content);background-color:var(--scalar-border-color)}.scalar-client .before\\:bg-gradient-to-r:before{content:var(--tw-content);background-image:linear-gradient(to right,var(--tw-gradient-stops))}.scalar-client .before\\:from-transparent:before{content:var(--tw-content);--tw-gradient-from:transparent var(--tw-gradient-from-position);--tw-gradient-to:rgb(0 0 0 / 0) var(--tw-gradient-to-position);--tw-gradient-stops:var(--tw-gradient-from), var(--tw-gradient-to)}.scalar-client .before\\:to-b-1:before{content:var(--tw-content);--tw-gradient-to:var(--scalar-background-1) var(--tw-gradient-to-position)}.scalar-client .before\\:content-\\[\\\'\\\'\\]:before{--tw-content:"";content:var(--tw-content)}.scalar-client .before\\:bg-border:before{content:var(--tw-content);--tw-bg-base:var(--scalar-border-color)}.scalar-client .after\\:content-\\[\\\'\\:\\\'\\]:after{--tw-content:":";content:var(--tw-content)}.scalar-client .first\\:border-t:first-child{border-top-width:var(--scalar-border-width)}.scalar-client .first\\:\\*\\:rounded-l>*:first-child{border-top-left-radius:var(--scalar-radius);border-bottom-left-radius:var(--scalar-radius)}.scalar-client .first\\:\\*\\:border-t-0>*:first-child{border-top-width:0}.scalar-client .last\\:mb-0:last-child{margin-bottom:0}.scalar-client .last\\:border-r-0:last-child{border-right-width:0}.scalar-client .last\\:\\*\\:rounded-r>*:last-child{border-top-right-radius:var(--scalar-radius);border-bottom-right-radius:var(--scalar-radius)}.scalar-client .last\\:before\\:h-full:last-child:before{content:var(--tw-content);height:100%}.scalar-client .focus-within\\:z-20:focus-within{z-index:20}.scalar-client .focus-within\\:border-b-3:focus-within{border-color:var(--scalar-background-3)}.scalar-client .focus-within\\:bg-b-1:focus-within{background-color:var(--scalar-background-1)}.scalar-client .focus-within\\:text-c-1:focus-within{color:var(--scalar-color-1)}.scalar-client .focus-within\\:bg-b-1:focus-within{--tw-bg-base:var(--scalar-background-1)}.scalar-client .hover\\:cursor-default:hover{cursor:default}.scalar-client .hover\\:bg-b-2:hover{background-color:var(--scalar-background-2)}.scalar-client .hover\\:bg-b-3:hover{background-color:var(--scalar-background-3)}.scalar-client .hover\\:bg-inherit:hover{background-color:inherit}.scalar-client .hover\\:bg-sidebar-active-b:hover{background-color:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .hover\\:text-c-1:hover{color:var(--scalar-color-1)}.scalar-client .hover\\:text-c-2:hover{color:var(--scalar-color-2)}.scalar-client .hover\\:bg-b-2:hover{--tw-bg-base:var(--scalar-background-2)}.scalar-client .hover\\:bg-b-3:hover{--tw-bg-base:var(--scalar-background-3)}.scalar-client .hover\\:bg-inherit:hover{--tw-bg-base:inherit}.scalar-client .hover\\:bg-sidebar-active-b:hover{--tw-bg-base:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .focus\\:text-c-1:focus{color:var(--scalar-color-1)}.scalar-client .focus\\:outline-none:focus{outline:2px solid transparent;outline-offset:2px}.scalar-client .focus-visible\\:z-10:focus-visible{z-index:10}.scalar-client .active\\:text-c-1:active{color:var(--scalar-color-1)}.scalar-client .disabled\\:cursor-default:disabled{cursor:default}.scalar-client .disabled\\:text-c-2:disabled{color:var(--scalar-color-2)}.scalar-client .group\\/label:last-child .group-last\\/label\\:rounded-br-lg{border-bottom-right-radius:var(--scalar-radius-lg)}.scalar-client .group:last-child .group-last\\:border-b-transparent{border-bottom-color:transparent}.scalar-client .group\\/upload:hover .group-hover\\/upload\\:block,.scalar-client .group:hover .group-hover\\:block{display:block}.scalar-client .group\\/item:hover .group-hover\\/item\\:flex,.scalar-client .group:hover .group-hover\\:flex{display:flex}.scalar-client .group:hover .group-hover\\:hidden{display:none}.scalar-client .group:hover .group-hover\\:pr-5{padding-right:20px}.scalar-client .group\\/info:hover .group-hover\\/info\\:text-c-1{color:var(--scalar-color-1)}.scalar-client .group\\/scopes-accordion:hover .group-hover\\/scopes-accordion\\:text-c-2{color:var(--scalar-color-2)}.scalar-client .group:hover .group-hover\\:text-c-1{color:var(--scalar-color-1)}.scalar-client .group\\/cell:hover .group-hover\\/cell\\:opacity-100,.scalar-client .group\\/params:hover .group-hover\\/params\\:opacity-100,.scalar-client .group:hover .group-hover\\:opacity-100{opacity:1}.scalar-client .group:hover .group-hover\\:opacity-80{opacity:.8}.scalar-client .group:focus-visible .group-focus-visible\\:opacity-100{opacity:1}.scalar-client .group:focus-visible .group-focus-visible\\:outline{outline-style:solid}.scalar-client .group.alert .group-\\[\\.alert\\]\\:bg-b-alert{background-color:var(--scalar-background-alert)}.scalar-client .group.alert .group-\\[\\.alert\\]\\:bg-transparent{background-color:transparent}.scalar-client .group.alert .group-\\[\\.alert\\]\\:shadow-none{--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .group.alert .group-\\[\\.alert\\]\\:outline-orange{outline-color:var(--scalar-color-orange)}.scalar-client .group.alert .group-\\[\\.alert\\]\\:bg-b-alert{--tw-bg-base:var(--scalar-background-alert)}.scalar-client .group.alert .group-\\[\\.alert\\]\\:bg-transparent{--tw-bg-base:transparent}.scalar-client .group.alert .group-\\[\\.alert\\]\\:before\\:to-b-alert:before{content:var(--tw-content);--tw-gradient-to:var(--scalar-background-alert) var(--tw-gradient-to-position)}.scalar-client .peer:checked~.peer-checked\\:text-c-1{color:var(--scalar-color-1)}.scalar-client .has-\\[\\:focus-visible\\]\\:rounded-\\[4px\\]:has(:focus-visible){border-radius:4px}.scalar-client .has-\\[\\.empty-sidebar-item\\]\\:border-t-1\\/2:has(.empty-sidebar-item){border-top-width:calc(var(--scalar-border-width) / 2)}.scalar-client .has-\\[\\:focus-visible\\]\\:opacity-100:has(:focus-visible){opacity:1}.scalar-client .has-\\[\\:focus-visible\\]\\:outline:has(:focus-visible){outline-style:solid}.scalar-client .group\\/label:has(input) .group-has-\\[input\\]\\/label\\:mr-0{margin-right:0}.scalar-client .group\\/cell:has(:focus-visible) .group-has-\\[\\:focus-visible\\]\\/cell\\:border-c-accent{border-color:var(--scalar-color-accent)}.scalar-client .group\\/cell:has(:focus-visible) .group-has-\\[\\:focus-visible\\]\\/cell\\:opacity-100{opacity:1}.scalar-client .group:has(:focus-visible) .group-has-\\[\\:focus-visible\\]\\:opacity-100{opacity:1}.scalar-client .peer:has(.color-selector)~.peer-has-\\[\\.color-selector\\]\\:hidden{display:none}.scalar-client .peer:has(.cm-focused)~.peer-has-\\[\\.cm-focused\\]\\:opacity-0{opacity:0}.scalar-client .peer:has(:focus-visible)~.peer-has-\\[\\:focus-visible\\]\\:opacity-0{opacity:0}.scalar-client .ui-open\\:rotate-90[data-headlessui-state~=open]{--tw-rotate:90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client :where([data-headlessui-state~=open]) .ui-open\\:rotate-90{--tw-rotate:90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .ui-not-open\\:invisible[data-headlessui-state]:not([data-headlessui-state~=open]){visibility:hidden}.scalar-client .ui-not-open\\:rotate-0[data-headlessui-state]:not([data-headlessui-state~=open]){--tw-rotate:0deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client :where([data-headlessui-state]:not([data-headlessui-state~=open])) .ui-not-open\\:invisible:not([data-headlessui-state]){visibility:hidden}.scalar-client :where([data-headlessui-state]:not([data-headlessui-state~=open])) .ui-not-open\\:rotate-0:not([data-headlessui-state]){--tw-rotate:0deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .ui-checked\\:bg-b-3[data-headlessui-state~=checked]{background-color:var(--scalar-background-3);--tw-bg-base:var(--scalar-background-3)}.scalar-client :where([data-headlessui-state~=checked]) .ui-checked\\:bg-b-3{background-color:var(--scalar-background-3);--tw-bg-base:var(--scalar-background-3)}.scalar-client .ui-active\\:bg-b-2[data-headlessui-state~=active]{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .\\*\\:ui-active\\:bg-b-2[data-headlessui-state~=active]>*{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client :where([data-headlessui-state~=active]) .ui-active\\:bg-b-2{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client :where([data-headlessui-state~=active]) .\\*\\:ui-active\\:bg-b-2>*{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .app-platform-desktop .desktop\\:rounded-lg{border-radius:var(--scalar-radius-lg)}.scalar-client .app-platform-mac .mac\\:hidden{display:none}.scalar-client .app-platform-mac .mac\\:pl-\\[72px\\]{padding-left:72px}@media (min-width: 600px){.scalar-client .sm\\:not-sr-only{position:static;width:auto;height:auto;padding:0;margin:0;overflow:visible;clip:auto;white-space:normal}.scalar-client .sm\\:order-none{order:0}.scalar-client .sm\\:mb-1{margin-bottom:4px}.scalar-client .sm\\:mb-1\\.5{margin-bottom:6px}.scalar-client .sm\\:ml-1{margin-left:4px}.scalar-client .sm\\:ml-1\\.5{margin-left:6px}.scalar-client .sm\\:mr-1{margin-right:4px}.scalar-client .sm\\:mr-1\\.5{margin-right:6px}.scalar-client .sm\\:flex{display:flex}.scalar-client .sm\\:hidden{display:none}.scalar-client .sm\\:min-w-max{min-width:-moz-max-content;min-width:max-content}.scalar-client .sm\\:max-w-max{max-width:-moz-max-content;max-width:max-content}.scalar-client .sm\\:flex-row{flex-direction:row}.scalar-client .sm\\:flex-col{flex-direction:column}.scalar-client .sm\\:justify-between{justify-content:space-between}.scalar-client .sm\\:gap-px{gap:1px}.scalar-client .sm\\:rounded{border-radius:var(--scalar-radius)}.scalar-client .sm\\:rounded-lg{border-radius:var(--scalar-radius-lg)}.scalar-client .sm\\:px-2{padding-left:8px;padding-right:8px}.scalar-client .sm\\:px-3{padding-left:12px;padding-right:12px}.scalar-client .sm\\:py-1{padding-top:4px;padding-bottom:4px}.scalar-client .sm\\:py-1\\.5{padding-top:6px;padding-bottom:6px}.scalar-client .sm\\:\\*\\:rounded-lg>*{border-radius:var(--scalar-radius-lg)}}@media (min-width: 800px){.scalar-client .md\\:bottom-10{bottom:40px}.scalar-client .md\\:right-10{right:40px}.scalar-client .md\\:-ml-1{margin-left:-4px}.scalar-client .md\\:-ml-1\\.25{margin-left:-5px}.scalar-client .md\\:block{display:block}.scalar-client .md\\:w-full{width:100%}.scalar-client .md\\:min-w-fit{min-width:-moz-fit-content;min-width:fit-content}.scalar-client .md\\:flex-none{flex:none}.scalar-client .md\\:translate-x-0{--tw-translate-x:0px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .md\\:translate-y-0{--tw-translate-y:0px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .md\\:flex-row{flex-direction:row}.scalar-client .md\\:border-b-0{border-bottom-width:0}.scalar-client .md\\:border-r-1\\/2{border-right-width:calc(var(--scalar-border-width) / 2)}.scalar-client .md\\:p-1{padding:4px}.scalar-client .md\\:p-1\\.5{padding:6px}.scalar-client .md\\:px-0{padding-left:0;padding-right:0}.scalar-client .md\\:px-1{padding-left:4px;padding-right:4px}.scalar-client .md\\:px-1\\.5{padding-left:6px;padding-right:6px}.scalar-client .md\\:px-2{padding-left:8px;padding-right:8px}.scalar-client .md\\:px-2\\.5{padding-left:10px;padding-right:10px}.scalar-client .md\\:px-4{padding-left:16px;padding-right:16px}.scalar-client .md\\:px-\\[18px\\]{padding-left:18px;padding-right:18px}.scalar-client .md\\:py-2{padding-top:8px;padding-bottom:8px}.scalar-client .md\\:py-2\\.5{padding-top:10px;padding-bottom:10px}.scalar-client .md\\:pb-2{padding-bottom:8px}.scalar-client .md\\:pb-2\\.5{padding-bottom:10px}.scalar-client .md\\:pb-\\[37px\\]{padding-bottom:37px}.scalar-client .md\\:\\*\\:border-t-0>*{border-top-width:0}}@media (min-width: 1000px){.scalar-client .lg\\:order-none{order:0}.scalar-client .lg\\:-mr-1{margin-right:-4px}.scalar-client .lg\\:mb-0{margin-bottom:0}.scalar-client .lg\\:flex{display:flex}.scalar-client .lg\\:min-h-client-header{min-height:48px}.scalar-client .lg\\:w-\\[120px\\]{width:120px}.scalar-client .lg\\:w-auto{width:auto}.scalar-client .lg\\:min-w-\\[580px\\]{min-width:580px}.scalar-client .lg\\:max-w-\\[580px\\]{max-width:580px}.scalar-client .lg\\:flex-1{flex:1 1 0%}.scalar-client .lg\\:p-1{padding:4px}.scalar-client .lg\\:px-1{padding-left:4px;padding-right:4px}.scalar-client .lg\\:px-2{padding-left:8px;padding-right:8px}.scalar-client .lg\\:px-2\\.5{padding-left:10px;padding-right:10px}.scalar-client .lg\\:pr-24{padding-right:96px}.scalar-client .lg\\:pt-1{padding-top:4px}.scalar-client .lg\\:text-sm{font-size:var(--scalar-font-size-4)}}@media (min-width: 1200px){.scalar-client .xl\\:\\!flex{display:flex!important}.scalar-client .xl\\:flex{display:flex}.scalar-client .xl\\:hidden{display:none}.scalar-client .xl\\:h-full{height:100%}.scalar-client .xl\\:min-h-client-header{min-height:48px}.scalar-client .xl\\:min-w-0{min-width:0px}.scalar-client .xl\\:min-w-\\[720px\\]{min-width:720px}.scalar-client .xl\\:max-w-\\[720px\\]{max-width:720px}.scalar-client .xl\\:flex-row{flex-direction:row}.scalar-client .xl\\:overflow-hidden{overflow:hidden}.scalar-client .xl\\:rounded-none{border-radius:0}.scalar-client .xl\\:pl-2{padding-left:8px}.scalar-client .xl\\:pr-0{padding-right:0}.scalar-client .xl\\:pr-0\\.5{padding-right:2px}.scalar-client .xl\\:\\*\\:border-l-1\\/2>*{border-left-width:calc(var(--scalar-border-width) / 2)}.scalar-client .xl\\:\\*\\:border-t-0>*{border-top-width:0}.scalar-client .xl\\:first\\:\\*\\:border-l-0>*:first-child{border-left-width:0}}.scalar-client .dark\\:bg-b-2:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .dark\\:hover\\:bg-b-2:hover:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .\\[\\&\\>a\\]\\:pl-5>a{padding-left:20px}.scalar-client .\\[\\&\\>a\\]\\:pl-\\[1\\.625rem\\]>a{padding-left:1.625rem}html:has(.dark-mode){--scalar-background-1: #0f0f0f}html,body{background-color:var(--scalar-background-1);overscroll-behavior:none}#scalar-client{display:flex;flex-direction:column;height:100dvh;width:100dvw;position:relative;background-color:var(--scalar-background-2)}.dark-mode #scalar-client{background-color:color-mix(in srgb,var(--scalar-background-1) 65%,black)}.open-api-client-button[data-v-2e9f75f1]{cursor:pointer;width:100%;padding:9px 12px;height:31px;display:block;text-align:center;display:flex;align-items:center;justify-content:center;font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);line-height:1.385;text-decoration:none;border-radius:var(--scalar-radius);box-shadow:0 0 0 .5px var(--scalar-border-color);gap:6px;color:var(--scalar-sidebar-color-1)}.open-api-client-button[data-v-2e9f75f1]:hover{background:var( --scalar-sidebar-item-hover-background, var(--scalar-background-2) )}.addressbar-history-button[data-v-a323ca24]:hover{background:var(--scalar-background-3)}.addressbar-history-button[data-v-a323ca24]:focus-within{background:var(--scalar-background-2)}.addressbar-bg-states:has(.cm-focused) .codemirror-bg-switcher[data-v-a323ca24]{--scalar-background-1: var(--scalar-background-1)}.addressbar-bg-states[data-v-a323ca24]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2))}.addressbar-bg-states[data-v-a323ca24]:has(.cm-focused){background:var(--scalar-background-1);border-color:var(--scalar-border-color);outline:1px solid var(--scalar-color-accent)}.description[data-v-8eba7808] .markdown{font-weight:var(--scalar-semibold);color:var(--scalar-color--1);padding:0;display:block}.description[data-v-8eba7808] .markdown>*:first-child{margin-top:0}[data-v-b4d63e6c] .cm-editor{background-color:var(--scalar-background-1);height:100%;outline:none;width:100%}[data-v-b4d63e6c] .cm-content{padding:0;display:flex;align-items:center;font-size:var(--scalar-mini)}.scroll-timeline-x[data-v-b4d63e6c]{scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;-ms-overflow-style:none}.scroll-timeline-x-hidden[data-v-b4d63e6c]{overflow-x:auto}.scroll-timeline-x-hidden[data-v-b4d63e6c] .cm-scroller{scrollbar-width:none;-ms-overflow-style:none;padding-right:20px;overflow:auto}.scroll-timeline-x-hidden[data-v-b4d63e6c]::-webkit-scrollbar{width:0;height:0;display:none}.scroll-timeline-x-hidden[data-v-b4d63e6c] .cm-scroller::-webkit-scrollbar{width:0;height:0;display:none}.scroll-timeline-x-address[data-v-b4d63e6c]{line-height:27px;scrollbar-width:none}.scroll-timeline-x-address[data-v-b4d63e6c]:after{content:"";position:absolute;height:100%;width:24px;right:0;cursor:text}.scroll-timeline-x-address[data-v-b4d63e6c]:empty:before{content:"Enter URL or cURL request";color:var(--scalar-color-3);pointer-events:none}.fade-left[data-v-b4d63e6c],.fade-right[data-v-b4d63e6c]{content:"";position:sticky;height:100%;animation-name:fadein-b4d63e6c;animation-duration:1ms;animation-direction:reverse;animation-timeline:--scroll-timeline;pointer-events:none}.fade-left[data-v-b4d63e6c]{background:linear-gradient(-90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 30%,var(--scalar-background-1) 100%);left:0;min-width:3px;animation-direction:normal}.fade-right[data-v-b4d63e6c]{background:linear-gradient(90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 30%,var(--scalar-background-1) 100%);right:0;min-width:24px}@keyframes fadein-b4d63e6c{0%{opacity:0}1%{opacity:1}}.codemirror-bg-switcher[data-v-b4d63e6c]{--scalar-background-1: color-mix( in srgb, var(--scalar-background-1), var(--scalar-background-2) )}.addressbar-bg-states:has(.cm-focused) .codemirror-bg-switcher[data-v-b4d63e6c]{--scalar-background-1: var(--scalar-background-1)}.addressbar-bg-states[data-v-b4d63e6c]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2))}.addressbar-bg-states[data-v-b4d63e6c]:has(.cm-focused){background:var(--scalar-background-1);border-color:var(--scalar-border-color);outline:1px solid var(--scalar-color-accent)}.sidebar-height[data-v-5397f2a7]{min-height:100%}@media (min-width: 800px){.sidebar-mask[data-v-5397f2a7]{-webkit-mask-image:linear-gradient(0,transparent 0,transparent 0,var(--scalar-background-2) 30px);mask-image:linear-gradient(0,transparent 0,transparent 0,var(--scalar-background-2) 30px)}}.resizer[data-v-5397f2a7]{width:5px;cursor:col-resize;position:absolute;top:0;right:0;bottom:0;border-right:2px solid transparent;transition:border-right-color .3s}.resizer[data-v-5397f2a7]:hover,.dragging .resizer[data-v-5397f2a7]{border-right-color:var(--scalar-background-3)}.dragging[data-v-5397f2a7]{cursor:col-resize}.dragging[data-v-5397f2a7]:before{content:"";display:block;position:absolute;width:100%;height:100%}.scroll-timeline-x[data-v-b77d5bd3]{overflow:auto;scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;-ms-overflow-style:none;scrollbar-width:none}.gitbook-show[data-v-cac03082]{display:none}.app-exit-button[data-v-cac03082]{color:#fff;background:#0000001a}.app-exit-button[data-v-cac03082]:hover{background:#ffffff1a}.ellipsis-position[data-v-daf8ea6e]{transform:translate3d(calc(-100% - 4.5px),0,0)}.dragover-asChild[data-v-3449bf19],.dragover-above[data-v-3449bf19],.dragover-below[data-v-3449bf19]{position:relative}.dragover-above[data-v-3449bf19]:after,.dragover-below[data-v-3449bf19]:after{content:"";position:absolute;top:-1.5px;height:3px;width:100%;display:block;background:color-mix(in srgb,var(--scalar-color-blue),transparent 85%);pointer-events:none;border-radius:var(--scalar-radius)}.dragover-below[data-v-3449bf19]:after{top:initial;bottom:-1.5px}.dragover-asChild[data-v-3449bf19]:after{content:"";position:absolute;top:0;left:0;width:100%;height:100%;display:block;background:color-mix(in srgb,var(--scalar-color-blue),transparent 85%);pointer-events:none;border-radius:var(--scalar-radius)}.indent-border-line-offset[data-v-682d438e]:before{left:var(--0ff55690)}.indent-padding-left[data-v-682d438e]{padding-left:calc(var(--ecb155a4) + 6px)}.sidebar-folderitem[data-v-682d438e] .ellipsis-position{right:6px;transform:none}.search-button-fade[data-v-b7bf7c07]{background:linear-gradient(var(--scalar-background-1) 32px,color-mix(in srgb,var(--scalar-background-1),transparent) 38px,transparent)}.empty-sidebar-item-content[data-v-b7bf7c07]{display:none}.empty-sidebar-item .empty-sidebar-item-content[data-v-b7bf7c07]{display:block}.rabbitjump[data-v-b7bf7c07]{opacity:0}.empty-sidebar-item:hover .rabbitjump[data-v-b7bf7c07]{opacity:1;animation:rabbitAnimation-b7bf7c07 .5s steps(1) infinite}.empty-sidebar-item:hover .rabbitsit[data-v-b7bf7c07]{opacity:0;animation:rabbitAnimation2-b7bf7c07 .5s steps(1) infinite}.empty-sidebar-item:hover .rabbit-ascii[data-v-b7bf7c07]{animation:rabbitRun-b7bf7c07 8s infinite linear}@keyframes rabbitRun-b7bf7c07{0%{transform:translateZ(0)}25%{transform:translate3d(250px,0,0)}25.01%{transform:translate3d(-250px,0,0)}75%{transform:translate3d(250px,0,0)}75.01%{transform:translate3d(-250px,0,0)}to{transform:translateZ(0)}}@keyframes rabbitAnimation-b7bf7c07{0%,to{opacity:1}50%{opacity:0}}@keyframes rabbitAnimation2-b7bf7c07{0%,to{opacity:0}50%{opacity:1;transform:translate3d(0,-8px,0)}}.request-text-color-text[data-v-02ba9b2d]{color:var(--scalar-color-1);background:linear-gradient(var(--scalar-background-1),var(--scalar-background-3));box-shadow:0 0 0 1px var(--scalar-border-color)}@media screen and (max-width: 800px){.sidebar-active-hide-layout[data-v-02ba9b2d]{display:none}.sidebar-active-width[data-v-02ba9b2d]{width:100%}}.active-link[data-v-a6c1ff1c]{background-color:var(--scalar-background-2);color:var(--scalar-color-1);--tw-bg-base: var(--scalar-background-2)}.empty-variable-name[data-v-a6c1ff1c]:empty:before{content:"Untitled";color:var(--scalar-color-3)}.form-group[data-v-c25da223]{margin-bottom:1rem}.modal-actions[data-v-c25da223]{display:flex;justify-content:flex-end;gap:1rem}:root{--scalar-loaded-api-reference: true}@layer scalar-config{.scalar-api-reference[data-v-10065932]{--refs-sidebar-width: var(--scalar-sidebar-width, 0px);--refs-header-height: calc( var(--scalar-y-offset) + var(--scalar-header-height, 0px) );--refs-content-max-width: var(--scalar-content-max-width, 1540px)}.scalar-api-reference.references-classic[data-v-10065932]{--refs-content-max-width: var(--scalar-content-max-width, 1420px);min-height:100dvh;--refs-sidebar-width: 0}}.references-layout[data-v-10065932]{min-height:100dvh;min-width:100%;max-width:100%;flex:1;--full-height: 100dvh;display:grid;grid-template-rows:var(--scalar-header-height, 0px) repeat(2,auto);grid-template-columns:var(--refs-sidebar-width) 1fr;grid-template-areas:"header header" "navigation rendered" "footer footer";background:var(--scalar-background-1)}.references-header[data-v-10065932]{grid-area:header;position:sticky;top:var(--scalar-custom-header-height, 0px);z-index:10;height:var(--scalar-header-height, 0px)}.references-editor[data-v-10065932]{grid-area:editor;display:flex;min-width:0;background:var(--scalar-background-1);z-index:1}.references-navigation[data-v-10065932]{grid-area:navigation}.references-rendered[data-v-10065932]{position:relative;grid-area:rendered;min-width:0;background:var(--scalar-background-1)}.scalar-api-reference.references-classic[data-v-10065932],.references-classic .references-rendered[data-v-10065932]{--full-height: fit-content !important;height:initial!important;max-height:initial!important}.references-navigation-list[data-v-10065932]{position:sticky;top:var(--refs-header-height);height:calc(100dvh - var(--refs-header-height));background:var(--scalar-sidebar-background-1, var(--scalar-background-1));overflow-y:auto;display:flex;flex-direction:column}.references-editor-textarea[data-v-10065932]{position:sticky;top:var(--refs-header-height);height:calc(var(--full-height) - var(--refs-header-height));display:flex;min-width:0;flex:1}.references-editable[data-v-10065932]{grid-template-columns:var(--refs-sidebar-width) 1fr 1fr;grid-template-areas:"header header header" "navigation editor rendered" "footer footer footer"}@layer scalar-config{.references-sidebar[data-v-10065932]{--refs-sidebar-width: var(--scalar-sidebar-width, 280px)}}.references-footer[data-v-10065932]{grid-area:footer}@media (max-width: 1150px){.references-layout[data-v-10065932]{grid-template-columns:var(--refs-sidebar-width) 1fr 0px}}@media (max-width: 1000px){.references-layout[data-v-10065932]{grid-template-columns:auto;grid-template-rows:var(--scalar-header-height, 0px) 0px auto auto;grid-template-areas:"header" "navigation" "rendered" "footer"}.references-editable[data-v-10065932]{grid-template-areas:"header" "navigation" "editor"}.references-navigation[data-v-10065932],.references-rendered[data-v-10065932]{max-height:unset}.references-rendered[data-v-10065932]{position:static}.references-navigation[data-v-10065932]{display:none;z-index:10}.references-sidebar-mobile-open .references-navigation[data-v-10065932]{display:block;top:var(--refs-header-height);height:calc(100dvh - var(--refs-header-height));width:100%;position:sticky}.references-navigation-list[data-v-10065932]{position:absolute;top:-1px;height:calc(var(--full-height) - var(--refs-header-height) + 1px);width:100%;border-top:1px solid var(--scalar-border-color);display:flex;flex-direction:column}}.references-classic-header[data-v-a860f78f]{display:flex;align-items:center;justify-content:space-between;gap:12px;max-width:var(--refs-content-max-width);margin:auto;padding:12px 0}.references-classic-header-container[data-v-a860f78f]{padding:0 60px}@container narrow-references-container (max-width: 900px){.references-classic-header[data-v-a860f78f]{padding:12px 24px}.references-classic-header-container[data-v-a860f78f]{padding:0}}.references-classic-header-icon[data-v-a860f78f]{height:24px;color:var(--scalar-color-1)}a[data-v-829b4762]{text-decoration:none}.ref-search-container[data-v-829b4762]{display:flex;flex-direction:column;padding:12px 12px 0}.ref-search-results[data-v-829b4762]{padding:12px}.ref-search-meta[data-v-829b4762]{background:var(--scalar-background-3);padding:6px 12px;font-size:var(--scalar-font-size-4);color:var(--scalar-color-3);font-weight:var(--scalar-semibold);display:flex;gap:12px}.deprecated[data-v-829b4762]{text-decoration:line-through}.sidebar-search[data-v-9f0bd6b4]{display:flex;align-items:center;position:relative;padding:0 3px 0 9px;min-width:254px;max-width:100%;font-family:var(--scalar-font);background:var( --scalar-sidebar-search-background, var(--scalar-background-1) );color:var(--scalar-sidebar-color-2, var(--scalar-color-2));border-radius:var(--scalar-radius);box-shadow:0 0 0 .5px var(--scalar-sidebar-search-border-color, var(--scalar-border-color));cursor:pointer;-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none}.sidebar-search-input[data-v-9f0bd6b4]{font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);height:31px;-webkit-user-select:none;-moz-user-select:none;user-select:none;z-index:10;position:relative;display:flex;width:100%;justify-content:space-between;align-items:center}.sidebar-search-key[data-v-9f0bd6b4]{text-transform:uppercase;background-color:var(--scalar-background-2);padding:3px 5px;margin:2px;border-radius:var(--scalar-radius);color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}.scalar-search-icon[data-v-9f0bd6b4]{padding:0;margin-right:6px;width:12px}.references-mobile-header[data-v-b5fb4b8b]{display:none;align-items:center;height:100%;width:100%;padding:0 8px;background:var(--scalar-background-1);border-bottom:1px solid var(--scalar-border-color)}.references-mobile-breadcrumbs[data-v-b5fb4b8b]{flex:1;min-width:0;font-size:var(--scalar-small);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);overflow:hidden;white-space:nowrap;text-overflow:ellipsis}.references-mobile-header-actions[data-v-b5fb4b8b]{display:flex;flex-direction:row;gap:4px;height:24px;align-items:center;padding-left:4px}@media (max-width: 1000px){.references-mobile-header[data-v-b5fb4b8b]{display:flex}}@media (max-width: 1000px){.scalar-api-references-standalone-mobile{--scalar-header-height: 50px}}.scalar-api-references-standalone-search[data-v-3ca2ed96]{display:flex;flex-direction:column;padding:12px 12px 6px}.darklight-reference[data-v-3ca2ed96]{width:100%;margin-top:auto}@layer scalar-base{body{margin:0;background-color:var(--scalar-background-1)}}\n')),document.head.appendChild(e),console.warn("Auto-loading the references css through js has been deprecated. Please import the css directly. Visit https://github.com/scalar/scalar for more info.")}),0)}catch(e){console.error(e,"unable to concat style inside the bundled file")}}();var Ll,Bl={exports:{}},Fl={};Bl.exports=function(){if(Ll)return Fl;Ll=1;var e=Symbol.for("react.transitional.element"),t=Symbol.for("react.fragment");function n(t,n,r){var a=null;if(void 0!==r&&(a=""+r),void 0!==n.key&&(a=""+n.key),"key"in n)for(var o in r={},n)"key"!==o&&(r[o]=n[o]);else r=n;return n=r.ref,{$$typeof:e,type:t,key:a,ref:void 0!==n?n:null,props:r}}return Fl.Fragment=t,Fl.jsx=n,Fl.jsxs=n,Fl}();var Ql=Bl.exports;globalThis.__VUE_OPTIONS_API__=!0,globalThis.__VUE_PROD_HYDRATION_MISMATCH_DETAILS__=!0,globalThis.__VUE_PROD_DEVTOOLS__=!1;const Ul=e=>{const n=(0,t.useRef)(null),[r,a]=(0,t.useState)(null);return(0,t.useEffect)((()=>{if(!n.current)return null==r?void 0:r.unmount;const t={_integration:"react",...e.configuration},o=Il(n.current,t);return a(o),o.unmount}),[n]),(0,t.useEffect)((()=>{const t={_integration:"react",...e.configuration};null==r||r.updateConfig(t,!1)}),[e.configuration]),Ql.jsx("div",{ref:n})};class zl{constructor(){this.url="||@ url|safe @||"}}function jl(){const e=new zl;return(0,d.jsx)(Ul,{configuration:{isEditable:!1,spec:{url:e.url},showSidebar:!0,hideModels:!1,hideDownloadButton:!1,hideTestRequestButton:!1,hideSearch:!1,darkMode:!1,forceDarkModeState:"light",hideDarkModeToggle:!0,searchHotKey:"k",withDefaultFonts:!1,defaultOpenAllTags:!1,tagsSorter:"alpha",operationsSorter:"alpha",theme:"none",hideClientButton:!0}})}var Wl=a("outline","brand-linkedin","IconBrandLinkedin",[["path",{d:"M8 11v5",key:"svg-0"}],["path",{d:"M8 8v.01",key:"svg-1"}],["path",{d:"M12 16v-5",key:"svg-2"}],["path",{d:"M16 16v-3a2 2 0 1 0 -4 0",key:"svg-3"}],["path",{d:"M3 7a4 4 0 0 1 4 -4h10a4 4 0 0 1 4 4v10a4 4 0 0 1 -4 4h-10a4 4 0 0 1 -4 -4z",key:"svg-4"}]]),Xl=a("outline","brand-x","IconBrandX",[["path",{d:"M4 4l11.733 16h4.267l-11.733 -16z",key:"svg-0"}],["path",{d:"M4 20l6.768 -6.768m2.46 -2.46l6.772 -6.772",key:"svg-1"}]]),ql=a("outline","brand-medium","IconBrandMedium",[["path",{d:"M4 4m0 2a2 2 0 0 1 2 -2h12a2 2 0 0 1 2 2v12a2 2 0 0 1 -2 2h-12a2 2 0 0 1 -2 -2z",key:"svg-0"}],["path",{d:"M8 9h1l3 3l3 -3h1",key:"svg-1"}],["path",{d:"M8 15l2 0",key:"svg-2"}],["path",{d:"M14 15l2 0",key:"svg-3"}],["path",{d:"M9 9l0 6",key:"svg-4"}],["path",{d:"M15 9l0 6",key:"svg-5"}]]);const Vl=[{url:"https://github.com/vortico/",icon:(0,d.jsx)(s,{})},{url:"https://www.linkedin.com/company/vortico-tech/",icon:(0,d.jsx)(Wl,{})},{url:"https://twitter.com/vorticotech",icon:(0,d.jsx)(Xl,{})},{url:"https://dev.to/vortico",icon:(0,d.jsx)(ql,{})}];function Hl(){return(0,d.jsx)("footer",{children:(0,d.jsx)("div",{className:"border-vortico-500/50 border-t px-4 py-2 sm:px-6 md:px-8",children:(0,d.jsxs)("div",{className:"mx-auto flex max-w-7xl items-center justify-between",children:[(0,d.jsx)("a",{href:"https://vortico.tech",children:(0,d.jsx)(b,{logo:"vortico",color:"vortico",size:"lg"})}),(0,d.jsx)("div",{className:"divide-vortico-500/50 flex h-8 items-center justify-end divide-x",children:Vl.map(((e,t)=>(0,d.jsx)("a",{href:e.url,className:"px-2 first:pl-0 last:pr-0",children:(0,d.jsx)("div",{className:"text-primary-400 hover:text-vortico-500 h-6 w-6 transition-colors duration-200",children:e.icon})},t)))})]})})})}function Zl(){return(0,d.jsxs)("div",{children:[(0,d.jsx)(x,{}),(0,d.jsx)("main",{children:(0,d.jsx)(jl,{})}),(0,d.jsx)(Hl,{})]})}e.createRoot(document.getElementById("app")).render((0,d.jsx)(Zl,{}))})()})();</script></body></html>