<!doctype html><html lang="en"><head><meta charset="UTF-8"/><meta name="viewport" content="width=device-width,initial-scale=1"/><link rel="preconnect" href="https://fonts.googleapis.com"/><link rel="preconnect" href="https://fonts.gstatic.com" crossorigin/><link href="https://fonts.googleapis.com/css2?family=Lato:ital,wght@0,100;0,300;0,400;0,700;0,900;1,100;1,300;1,400;1,700;1,900&family=Montserrat:ital,wght@0,100..900;1,100..900&display=swap" rel="stylesheet"/><title>Internal Server Error - Flama</title><style>:root {
  --scalar-border-width: 1px;

  --scalar-radius: 0px;
  --scalar-radius-lg: 0px;
  --scalar-radius-xl: 0px;

  --scalar-text-decoration: none;
  --scalar-text-decoration-hover: none;

  --scalar-font:
    'Lato', ui-sans-serif, system-ui, sans-serif, 'Apple Color Emoji', 'Segoe UI Emoji', 'Segoe UI Symbol',
    'Noto Color Emoji';
  --scalar-font-code:
    'Fira Mono', ui-monospace, SFMono-Regular, Menlo, Monaco, Consolas, 'Liberation Mono', 'Courier New', monospace;
}

.scalar-app:not(#headlessui-portal-root) {
  margin: calc(3.5rem + 1px) 0 0 0;
  height: calc(100vh - 6.5rem - 2px);
  min-height: calc(100vh - 6.5rem - 2px);
  max-height: calc(100vh - 6.5rem - 2px);
  overflow: auto;
}

:where(.t-doc__sidebar, .references-navigation-list) {
  height: calc(100vh - 6.5rem - 2px) !important;
  width: 18rem;
  min-height: calc(100vh - 6.5rem - 2px);
  max-height: calc(100vh - 6.5rem - 2px);
  overflow: auto !important;
  position: fixed !important;
  left: 0 !important;
  top: calc(3.5rem + 1px) !important;
  z-index: 10;
}

.dark-mode {
  color-scheme: dark;
}

.light-mode {
  color-scheme: light;
}

.light-mode,
.dark-mode {
  --scalar-scrollbar-color: #808290;
  --scalar-scrollbar-color-active: #808290;

  --scalar-shadow-1: 0 1px 3px 0 #aaabb6;
  --scalar-shadow-2: #aaabb6 0px 13px 20px 0px, #aaabb6 0px 3px 8px 0px, #aaabb6 0px 0 0 0.5px;

  /* Basic theme */
  --scalar-background-1: #e5e6e8;
  --scalar-background-2: #f5f5f6;
  --scalar-background-3: #cdced4;
  --scalar-background-accent: #f0ac81;

  --scalar-color-1: #575863;
  --scalar-color-2: #656775;
  --scalar-color-3: #4a4a54;

  --scalar-color-accent: #e25822;
  --scalar-border-color: #aaabb6;

  /* Advanced theme */
  --scalar-color-green: #00976e;
  --scalar-color-red: #9e744f;
  --scalar-color-yellow: #f0ac81;
  --scalar-color-blue: #00bbd5;
  --scalar-color-orange: #e25822;
  --scalar-color-purple: #008080;

  --scalar-button-1: var(--scalar-background-2);
  --scalar-button-1-hover: var(--scalar-background-accent);
  --scalar-button-1-color: var(--scalar-color-1);

  /* Document Sidebar */
  .t-doc__sidebar {
    --scalar-sidebar-background-1: var(--scalar-background-1);
    --scalar-sidebar-color-1: var(--scalar-color-1);
    --scalar-sidebar-color-2: var(--scalar-color-3);
    --scalar-sidebar-border-color: var(--scalar-border-color);

    --scalar-sidebar-item-hover-background: var(--scalar-background-accent);
    --scalar-sidebar-item-hover-color: var(--scalar-color-3);

    --scalar-sidebar-item-active-background: var(--scalar-background-3);
    --scalar-sidebar-color-active: var(--scalar-color-3);

    --scalar-sidebar-indent-border: var(--scalar-sidebar-border-color);
    --scalar-sidebar-indent-border-hover: var(--scalar-color-accent);
    --scalar-sidebar-indent-border-active: var(--scalar-color-3);

    --scalar-sidebar-search-background: var(--scalar-background-2);
    --scalar-sidebar-search-color: var(--scalar-color-2);
    --scalar-sidebar-search-border-color: var(--scalar-border-color);
  }
}

*::selection {
  background-color: color-mix(in srgb, var(--scalar-color-accent), transparent 75%);
}

* {
  border-radius: 0 !important;
}
.start[data-v-391b4e37]{padding:24px;display:flex;flex-flow:wrap;justify-content:space-between;position:relative;z-index:0}.swagger-editor .start[data-v-391b4e37]{padding-top:24px}.start-h1[data-v-391b4e37]{font-size:var(--scalar-heading-2);margin-top:0;line-height:1.45;margin-bottom:0;font-weight:var(--scalar-bold);color:var(--scalar-color-1);width:100%;position:relative}.start-h3[data-v-391b4e37]{font-size:var(--scalar-paragraph);margin-top:0;margin-bottom:6px;display:block;line-height:1.45;font-weight:var(--scalar-bold);color:var(--scalar-color-1);width:100%}.start-h1[data-v-391b4e37]:not(:first-of-type){margin-top:24px}.start-p[data-v-391b4e37]{font-size:var(--scalar-paragraph);color:var(--scalar-color-2);line-height:1.5;width:100%;margin-top:12px}.start-ul[data-v-391b4e37]{margin-top:12px;font-size:var(--scalar-paragraph);line-height:1.5;padding-left:0;list-style:initial;display:flex;flex-flow:wrap;gap:24px}.start-ul li[data-v-391b4e37]{margin:0;padding:0;list-style:none;width:calc(50% - 24px);color:var(--scalar-color-2)}.start-ul li[data-v-391b4e37]:first-of-type{margin-top:0}.start-section[data-v-391b4e37]{width:100%;margin-bottom:12px;display:flex;flex-flow:wrap}.start-section[data-v-391b4e37]:last-of-type{margin-bottom:48px}.start-h2[data-v-391b4e37]{background:var(--scalar-background-2);border-top-left-radius:var(--scalar-radius-lg);border-top-right-radius:var(--scalar-radius-lg);border:1px solid var(--scalar-border-color);color:var(--scalar-color-3);font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);padding:9px;width:100%}.start-item[data-v-391b4e37]{align-items:center;background:var(--scalar-background-2);border-right:1px solid var(--scalar-border-color);border-bottom:1px solid var(--scalar-border-color);color:var(--scalar-color-1);cursor:pointer;display:flex;flex:1;font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);padding:9px;text-transform:capitalize;-webkit-user-select:none;-moz-user-select:none;user-select:none}.start-section-integrations .start-item[data-v-391b4e37]:first-of-type{border-bottom-left-radius:var(--scalar-radius-lg);border-left:1px solid var(--scalar-border-color)}.start-section-integrations .start-item[data-v-391b4e37]:last-of-type{border-bottom-right-radius:var(--scalar-radius-lg)}.start-section-colors .start-item[data-v-391b4e37]{min-width:33.33%}.start-section-colors .start-item[data-v-391b4e37]:nth-child(3n+2){border-left:1px solid var(--scalar-border-color)}.start-section-colors .start-item[data-v-391b4e37]:last-of-type,.start-section-colors .start-item-active[data-v-391b4e37]:last-of-type:before{border-radius:0 0 var(--scalar-radius-lg) var(--scalar-radius-lg)}.start-item[data-v-391b4e37]:empty{pointer-events:none}.start-item svg[data-v-391b4e37]{width:14px;height:14px;margin-right:6px}.start-item[data-v-391b4e37]:hover{background:var(--scalar-background-3)}.start-item-active[data-v-391b4e37]{z-index:10;position:relative;color:var(--scalar-color-1)}.start-item-active[data-v-391b4e37]:before{border:1px solid var(--scalar-color-1);content:"";top:-1px;right:-1px;bottom:-1px;left:-1px;pointer-events:none;position:absolute}.start-section-color .start-item[data-v-391b4e37]{text-transform:capitalize}.start-cta[data-v-391b4e37]{display:flex;gap:12px;width:100%;margin-top:24px;margin-bottom:0}.start-row[data-v-391b4e37]{width:100%;margin-top:12px;overflow:hidden}.start-hero-copy[data-v-391b4e37]{background:var(--scalar-background-2);padding:12px;border-radius:var(--scalar-radius-lg)}.start-p-small[data-v-391b4e37]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-2);margin-bottom:12px;line-height:1.4}.start-cta[data-v-391b4e37]{margin-bottom:12px;width:-moz-fit-content;width:fit-content;white-space:nowrap}.start-copy[data-v-391b4e37]{padding:76px 48px 48px;display:flex;justify-content:center;align-items:center;flex-direction:column;text-align:center}.start-logo[data-v-391b4e37]{color:var(--scalar-color-1);margin-bottom:24px;width:72px;aspect-ratio:1;position:relative;box-shadow:var(--scalar-shadow-2);border-radius:50%}.start-logo[data-v-391b4e37]:before{content:"";width:300%;aspect-ratio:1;left:-100%;top:-100%;position:absolute;z-index:-1;border-radius:50%;background-size:24px 24px;box-shadow:inset 0 0 50px var(--scalar-background-1),inset 0 0 50px var(--scalar-background-1);background-image:linear-gradient(to right,var(--scalar-border-color) 1px,transparent 1px),linear-gradient(to bottom,var(--scalar-border-color) 1px,transparent 1px)}.start-logo svg[data-v-391b4e37]{width:100%;height:auto;background:var(--scalar-background-1);padding:3px;border-radius:50%;position:relative}@media screen and (max-width: 600px){.start-section-colors .start-item[data-v-391b4e37],.start-item[data-v-391b4e37]{width:100%;border-radius:0;border-right:none;border-top:1px solid var(--scalar-border-color)}.start-item[data-v-391b4e37]:empty{display:none}.start-h2[data-v-391b4e37]{border-bottom:none}.start li[data-v-391b4e37]{width:100%}.start-copy[data-v-391b4e37]{padding:48px 0 24px}}@media screen and (max-width: 1000px){.start[data-v-391b4e37]{padding:0;overflow:auto}}.sidebar-heading-type[data-v-f51e2805]{display:block;min-width:3.9em;overflow:hidden;line-height:14px;flex-shrink:0;color:#fff;color:color-mix(in srgb,var(--method-color, var(--scalar-color-1)),transparent 0%);text-transform:uppercase;font-size:10px;font-weight:var(--scalar-bold);text-align:right;position:relative;font-family:var(--scalar-font-code);white-space:nowrap;margin-left:3px}.sidebar-heading[data-v-18e6dc3d]{display:flex;gap:6px;color:var(--scalar-sidebar-color-2, var(--scalar-color-2));font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);word-break:break-word;line-height:1.385;max-width:100%;position:relative;cursor:pointer;border-radius:var(--scalar-radius);flex:1;padding-right:9px;-webkit-user-select:none;-moz-user-select:none;user-select:none}.sidebar-heading-link-method[data-v-18e6dc3d]{margin:0}.sidebar-heading.deprecated .sidebar-heading-link-title[data-v-18e6dc3d]{text-decoration:line-through}.sidebar-heading-link-title[data-v-18e6dc3d]{margin:0}.sidebar-heading[data-v-18e6dc3d]:hover{background:var( --scalar-sidebar-item-hover-background, var(--scalar-background-2) )}.sidebar-heading:hover .sidebar-heading-link-title[data-v-18e6dc3d]{color:var(--scalar-sidebar-item-hover-color)}.sidebar-heading-link[data-v-18e6dc3d]:focus-visible{outline:none}.sidebar-heading[data-v-18e6dc3d]:has(>.sidebar-heading-link:focus-visible){z-index:1;outline:1px solid var(--scalar-color-accent)}.active_page.sidebar-heading[data-v-18e6dc3d]:hover,.active_page.sidebar-heading[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-active, var(--scalar-color-accent));background:var( --scalar-sidebar-item-active-background, var(--scalar-background-accent) )}.active_page.sidebar-heading p[data-v-18e6dc3d]{font-weight:var(--scalar-sidebar-font-weight-active, var(--scalar-semibold))}.active_page.sidebar-heading:hover .sidebar-heading-link-title[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-active, var(--scalar-color-accent))}.sidebar-indent-nested .sidebar-indent-nested .sidebar-heading[data-v-18e6dc3d]:before{content:"";position:absolute;top:0;left:calc((var(--scalar-sidebar-level) * 12px));width:1px;height:100%;background:var(--scalar-sidebar-indent-border)}.sidebar-indent-nested .sidebar-indent-nested .sidebar-heading[data-v-18e6dc3d]:hover:before{background:var(--scalar-sidebar-indent-border-hover)}.sidebar-indent-nested .sidebar-indent-nested .active_page.sidebar-heading[data-v-18e6dc3d]:before{background:var(--scalar-sidebar-indent-border-active)}.sidebar-heading-link[data-v-18e6dc3d]{text-decoration:none;color:inherit;padding:6px 0;display:flex;flex:1;justify-content:space-between;gap:2px}.sidebar-heading p[data-v-18e6dc3d]{height:-moz-fit-content;height:fit-content;display:flex;align-items:center;font-weight:var(--scalar-sidebar-font-weight, var(--scalar-semibold))}.sidebar-heading p[data-v-18e6dc3d]:empty{display:none}.link-icon[data-v-18e6dc3d]{position:relative;left:4px}.sidebar-icon[data-v-18e6dc3d]{display:flex;align-items:center;justify-content:center;margin-right:6px;width:13px;height:13px}.sidebar-icon>svg[data-v-18e6dc3d]{width:13px;height:13px}.sidebar-group-item[data-v-18e6dc3d]{position:relative}.sidebar-heading-chevron[data-v-18e6dc3d]{margin:5px -5.5px 5px -9px}.sidebar-heading-chevron .toggle-nested-icon[data-v-18e6dc3d]:focus-visible{outline:none}.sidebar-heading[data-v-18e6dc3d]:has(.sidebar-heading-chevron .toggle-nested-icon:focus-visible){outline:none;box-shadow:inset 0 0 0 1px var(--scalar-color-accent)}.toggle-nested-icon[data-v-18e6dc3d]{color:var(--scalar-color-3);width:20px;height:20px;display:flex;align-items:center;justify-content:center}.active_page .toggle-nested-icon[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-active, var(--scalar-color-accent))}.toggle-nested-icon[data-v-18e6dc3d]:hover,.toggle-nested-icon[data-v-18e6dc3d]:focus-visible{color:currentColor}.action-menu[data-v-18e6dc3d]{position:absolute;top:5px;right:5px;display:flex;gap:6px}.action-menu[data-v-18e6dc3d] .button-wrapper button{opacity:0;width:20px;height:20px;padding:4px}.action-menu[data-v-18e6dc3d]:hover .button-wrapper button,.action-menu[data-v-18e6dc3d] .button-wrapper button:hover,.sidebar-heading:hover~.action-menu[data-v-18e6dc3d] .button-wrapper button,.action-menu[data-v-18e6dc3d] .button-wrapper button[aria-expanded=true]{opacity:1}.sidebar-heading[data-v-18e6dc3d]:has(~.action-menu:hover){color:var(--scalar-sidebar-color-1, var(--scalar-color-1));background:var( --scalar-sidebar-item-hover-background, var(--scalar-background-2) )}.sidebar-group-item__folder[data-v-18e6dc3d]{color:var(--scalar-sidebar-color-1, var(--scalar-color-1));text-transform:var(--scalar-tag-text-transform, initial)}.sidebar-group[data-v-39c84840]{list-style:none;width:100%;margin:0;padding:0}.sidebar-indent-nested[data-v-39c84840] .sidebar-heading{padding-left:calc((var(--scalar-sidebar-level) * var(--scalar-sidebar-indent-base)) + 12px)!important}.sidebar-indent-nested[data-v-39c84840] .sidebar-heading .toggle-nested-icon{left:calc((var(--scalar-sidebar-level) * var(--scalar-sidebar-indent-base)) + 2px)!important}:where(.sidebar-indent-nested[data-v-39c84840]) .sidebar-heading{color:var(--scalar-sidebar-color-1, var(--scalar-color-1))}:where(.sidebar-indent-nested[data-v-39c84840]) :where(.sidebar-indent-nested) .sidebar-heading{color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}.sidebar[data-v-b863366c]{--scalar-sidebar-indent-base: 12px;--scalar-sidebar-font-weight-active: var(--scalar-semibold);--scalar-sidebar-font-weight: var(--scalar-semibold);flex:1;height:100%;display:flex;flex-direction:column;border-right:var(--scalar-border-width) solid var(--scalar-sidebar-border-color, var(--scalar-border-color));background:var(--scalar-sidebar-background-1, var(--scalar-background-1));--scalar-sidebar-level: 0}.sidebar-pages[data-v-b863366c]{flex:1;padding:9px 12px}@media (max-width: 1000px){.sidebar[data-v-b863366c]{min-height:0;border-right:none}.sidebar-pages[data-v-b863366c]{padding-top:12px}}.sidebar-group-title[data-v-b863366c]{color:var(--scalar-sidebar-color-1);font-size:var(--scalar-mini);padding:12px 6px 6px;font-weight:var(--scalar-semibold);text-transform:uppercase;word-break:break-word;line-height:1.385}.sidebar-group-item+.sidebar-group-title[data-v-b863366c]{border-top:var(--scalar-border-width) solid var(--scalar-sidebar-border-color);margin-top:9px}.show-more[data-v-252cd03d]{background:var(--scalar-background-1);-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none;border:var(--scalar-border-width) solid var(--scalar-border-color);margin:auto;padding:8px 12px;border-radius:30px;color:var(--scalar-color-1);font-weight:var(--scalar-semibold);font-size:var(--scalar-small);display:flex;align-items:center;justify-content:center;position:relative;top:-48px}.show-more[data-v-252cd03d]:hover{color:var(--scalar-color-2);cursor:pointer}.show-more-icon[data-v-252cd03d]{width:16px!important;height:16px!important;margin-left:3px}.show-more[data-v-252cd03d]:active{box-shadow:0 0 0 1px var(--scalar-border-color)}@container narrow-references-container (max-width: 900px){.show-more[data-v-252cd03d]{top:-24px}}.screenreader-only[data-v-df2e1026]{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);border:0}.section[data-v-64ccd510]{position:relative;display:flex;flex-direction:column;max-width:var(--refs-content-max-width);margin:auto;padding:90px 0;scroll-margin-top:var(--refs-header-height)}.section[data-v-64ccd510]:has(~div.contents){border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.references-classic .section[data-v-64ccd510]{padding:48px 0;gap:24px}@container narrow-references-container (max-width: 900px){.references-classic .section[data-v-64ccd510],.section[data-v-64ccd510]{padding:48px 24px}}.section[data-v-64ccd510]:not(:last-of-type){border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.loading[data-v-c90b2c46]{background:var(--scalar-background-3);animation:loading-skeleton-c90b2c46 1.5s infinite alternate;border-radius:var(--scalar-radius-lg);min-height:1.6em;margin:.6em 0;max-width:100%}.loading[data-v-c90b2c46]:first-of-type{min-height:3em;margin-bottom:24px;margin-top:0}.loading[data-v-c90b2c46]:last-of-type{width:60%}.loading.single-line[data-v-c90b2c46]{min-height:3em;margin:.6em 0;max-width:80%}@keyframes loading-skeleton-c90b2c46{0%{opacity:1}to{opacity:.33}}.section-header-wrapper[data-v-b6aef53d]{display:grid;grid-template-columns:1fr}@media (min-width: 1200px){.section-header-wrapper[data-v-b6aef53d]{grid-template-columns:repeat(2,1fr)}}.section-header[data-v-b6aef53d]{font-size:var(--font-size, var(--scalar-heading-2));font-weight:var(--font-weight, var(--scalar-bold));color:var(--scalar-color-1);word-wrap:break-word;line-height:1.45;margin-top:0;margin-bottom:12px}.section-header.tight[data-v-b6aef53d]{margin-bottom:6px}.section-header.loading[data-v-b6aef53d]{width:80%}.label[data-v-3cb2dd80]{position:relative;display:inline-block;word-break:break-all}.anchor[data-v-3cb2dd80]{position:relative;display:inline-block;opacity:0}.anchor-copy[data-v-3cb2dd80]{position:absolute;left:0;top:50%;transform:translateY(-50%);cursor:pointer;padding:0 6px;color:var(--scalar-color-3);font-weight:var(--scalar-semibold);font-size:.8em}.anchor-copy[data-v-3cb2dd80]:hover,.anchor-copy[data-v-3cb2dd80]:focus-visible{color:var(--scalar-color-2)}.label:hover .anchor[data-v-3cb2dd80],.label:has(:focus-visible) .anchor[data-v-3cb2dd80]{opacity:1}@container narrow-references-container (max-width: 900px){.section-content--with-columns[data-v-9735459e]{flex-direction:column;gap:24px}}.section-columns[data-v-8b9602bf]{display:flex;gap:48px}@container narrow-references-container (max-width: 900px){.section-columns[data-v-8b9602bf]{flex-direction:column;gap:24px}}.section-column[data-v-699c28e3]{flex:1;min-width:0}@container narrow-references-container (max-width: 900px){.section-column[data-v-699c28e3]:nth-of-type(2){padding-top:0}}.scalar-card[data-v-38801681]{all:unset;font-family:var(--scalar-font);border-radius:var(--scalar-radius-lg);overflow:hidden;border:var(--scalar-border-width) solid var(--scalar-border-color);background:var(--scalar-background-2);display:flex;flex-direction:column;max-height:calc(((var(--full-height) - var(--refs-header-height)) - 60px) / 2);position:relative}.scalar-card-sticky[data-v-38801681]{position:sticky;top:calc(var(--refs-header-height) + 24px)}.scalar-card-content[data-v-dd83e46e]{overflow:auto;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color);display:grid}.scalar-card-content[data-v-dd83e46e] .simple-table .simple-header{display:none}.scalar-card-content[data-v-dd83e46e]:last-of-type,.scalar-card-content.scalar-card--borderless[data-v-dd83e46e]{border-bottom:none}.scalar-card--muted[data-v-dd83e46e]{background:var(--scalar-background-2)}.scalar-card--contrast[data-v-dd83e46e]{background:var(--scalar-background-3)}.scalar-card--frameless[data-v-dd83e46e]{padding:0}.scalar-card--transparent[data-v-dd83e46e]{background:var(--scalar-background-1)}.scalar-card-header[data-v-838dc259]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-3);padding:9px 3px 9px 12px;flex-shrink:0}.scalar-card-header.scalar-card--borderless[data-v-838dc259]+.scalar-card-content{margin-top:-9px}.scalar-card-header-slots[data-v-838dc259]{display:flex;justify-content:space-between;line-height:1.35}.scalar-card-header-title[data-v-838dc259]{text-transform:uppercase;flex:1;min-width:0;text-overflow:ellipsis;overflow:hidden}.scalar-card-header-actions[data-v-838dc259]{display:flex}.endpoints[data-v-cfef28a1]{overflow:auto;background:var(--scalar-background-2);padding:10px 12px}@media (max-width: 580px){.endpoints[data-v-cfef28a1]{max-height:calc(100vh - 150px)}}.endpoints span+span[data-v-cfef28a1]{text-align:left;margin-left:12px;text-transform:initial}.endpoint[data-v-cfef28a1]{display:flex;white-space:nowrap;cursor:pointer;text-decoration:none}.endpoint:hover .endpoint-path[data-v-cfef28a1],.endpoint:focus-visible .endpoint-path[data-v-cfef28a1]{text-decoration:underline}.endpoint span[data-v-cfef28a1]:first-of-type{text-transform:uppercase}.endpoint .post[data-v-cfef28a1],.endpoint .get[data-v-cfef28a1],.endpoint .delete[data-v-cfef28a1],.endpoint .put[data-v-cfef28a1]{white-space:nowrap}.endpoint span[data-v-cfef28a1]{color:var(--scalar-color-1);min-width:62px;display:inline-block;text-align:right;line-height:1.55;font-family:var(--scalar-font-code);font-size:var(--scalar-small);cursor:pointer}.deprecated[data-v-cfef28a1]{text-decoration:line-through}.section-container[data-v-1913755a]{position:relative;padding:0 60px;width:100%;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.section-container[data-v-1913755a]:has(.introduction-section){border-top:none}@container narrow-references-container (max-width: 900px){.section-container[data-v-1913755a]{padding:0}}.section-container[data-v-d47a803a]{border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.section-container[data-v-d47a803a]:has(.show-more){background-color:color-mix(in srgb,var(--scalar-background-2),transparent)}.operation-path[data-v-ec6c8861]{overflow:hidden;word-wrap:break-word;font-weight:var(--scalar-semibold);line-break:anywhere}.deprecated[data-v-ec6c8861]{text-decoration:line-through}.content-type[data-v-4b712394]{display:flex;align-items:center;font-size:var(--scalar-heading-4);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);line-height:1.45;margin-top:24px;padding-bottom:12px;border-bottom:1px solid var(--scalar-border-color);flex-flow:wrap}.content-type-select[data-v-4b712394]{position:relative;padding-left:9px;height:-moz-fit-content;height:fit-content;color:var(--scalar-color-2);font-size:var(--scalar-font-size-3);display:flex;align-items:center}.content-type-no-select.content-type-select[data-v-4b712394]{pointer-events:none}.content-type-no-select.content-type-select[data-v-4b712394]:after{display:none}.content-type-select span[data-v-4b712394]{display:flex;align-items:center}.content-type-select[data-v-4b712394]:after{content:"";width:7px;height:7px;transform:rotate(45deg) translate3d(-2px,-4px,0);display:block;margin-left:7px;box-shadow:1px 1px 0 currentColor}.content-type-select select[data-v-4b712394]{border:none;outline:none;cursor:pointer;background:var(--scalar-background-3);box-shadow:-2px 0 0 0 var(--scalar-background-3);position:absolute;top:0;left:0;width:100%;height:100%;opacity:0;-webkit-appearance:none;-moz-appearance:none;appearance:none}.content-type-select[data-v-4b712394]:hover{color:var(--scalar-color-1)}@media (max-width: 460px){.content-type-select[data-v-4b712394]{margin-left:auto;padding-right:3px}}.schema-type-icon[data-v-2a52eb0b]{color:var(--scalar-color-1);display:none}.schema-type[data-v-2a52eb0b]{font-family:var(--scalar-font-code);color:var(--scalar-color-1)}.error[data-v-f48d0197]{background-color:var(--scalar-color-red)}.schema-card[data-v-f48d0197]{z-index:0;position:relative;font-size:var(--scalar-font-size-4);color:var(--scalar-color-1)}.schema-card-title[data-v-f48d0197]{height:var(--schema-title-height);padding:6px 10px;display:flex;align-items:center;gap:4px;color:var(--scalar-color-2);font-weight:var(--scalar-semibold);font-size:var(--scalar-micro);border-bottom:var(--scalar-border-width) solid transparent}button.schema-card-title[data-v-f48d0197]{cursor:pointer}button.schema-card-title[data-v-f48d0197]:hover{color:var(--scalar-color-1)}.schema-card-title-icon--open[data-v-f48d0197]{transform:rotate(45deg)}.schema-properties-open>.schema-card-title[data-v-f48d0197]{border-bottom-left-radius:0;border-bottom-right-radius:0;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.schema-properties-open>.schema-properties[data-v-f48d0197]{width:-moz-fit-content;width:fit-content}.schema-card-description+.schema-properties[data-v-f48d0197]{width:-moz-fit-content;width:fit-content;margin-top:12px}.schema-properties-open.schema-properties[data-v-f48d0197],.schema-properties-open>.schema-card--open[data-v-f48d0197]{width:100%}.schema-card .property[data-v-f48d0197]:last-of-type{padding-bottom:10px}.schema-properties[data-v-f48d0197]{display:flex;flex-direction:column;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);width:-moz-fit-content;width:fit-content}.schema-properties .schema-properties[data-v-f48d0197]{border-radius:13.5px}.schema-properties .schema-properties.schema-properties-open[data-v-f48d0197]{border-radius:13.5px 13.5px 9px 9px}.schema-properties-open[data-v-f48d0197]{width:100%}.schema-card--compact[data-v-f48d0197]{align-self:flex-start}.schema-card--compact.schema-card--open[data-v-f48d0197]{align-self:initial}.schema-card-title--compact[data-v-f48d0197]{color:var(--scalar-color-2);padding:6px 8px;height:auto;border-bottom:none}.schema-card--compact>.schema-properties[data-v-f48d0197],.schema-card-title--compact[data-v-f48d0197]{border-radius:13.5px}.schema-card-title--compact>.schema-card-title-icon[data-v-f48d0197]{margin:0}.schema-card-title--compact>.schema-card-title-icon--open[data-v-f48d0197]{transform:rotate(45deg)}.schema-properties-open>.schema-card-title--compact[data-v-f48d0197]{position:static}.schema-card--compact.schema-card--level-0>.schema-properties[data-v-f48d0197]{border:none}[data-v-f48d0197] .schema-card-description p{font-size:var(--scalar-mini, var(--scalar-paragraph));color:var(--scalar-color-2);line-height:1.5;display:block;margin-bottom:6px}.schema-card-description[data-v-f48d0197]:first-of-type{padding-top:10px}.children .schema-card-description[data-v-f48d0197]:first-of-type{padding-top:0}.models-list-item .schema-properties[data-v-f48d0197]{margin-bottom:10px}.property-detail[data-v-3ae98a1b]{display:inline-flex}.property-detail+.property-detail[data-v-3ae98a1b]:before{display:block;content:"·";margin:0 .5ch}.property-detail-truncate[data-v-3ae98a1b]{overflow:hidden}.property-detail-truncate>.property-detail-value[data-v-3ae98a1b]{overflow:hidden;text-overflow:ellipsis}.property-detail-prefix[data-v-3ae98a1b]{color:var(--scalar-color-2)}code.property-detail-value[data-v-3ae98a1b]{font-family:var(--scalar-font-code);font-size:var(--scalar-font-size-3);color:var(--scalar-color-2);background:var(--scalar-background-3);padding:0 4px;border:.5px solid var(--scalar-border-color);border-radius:var(--scalar-radius)}.badge[data-v-190574ac]{color:var(--scalar-color-2);font-size:var(--scalar-micro);background:var(--scalar-background-2);padding:2px 6px;border-radius:12px;font-weight:var(--scalar-semibold);display:inline-block;text-transform:uppercase}.property-heading[data-v-dd95f4d7]{display:flex;flex-wrap:wrap;align-items:baseline;row-gap:9px;white-space:nowrap}.property-heading[data-v-dd95f4d7]>*{margin-right:9px}.property-heading[data-v-dd95f4d7]:last-child{margin-right:0}.property-heading>.property-detail[data-v-dd95f4d7]:not(:last-of-type){margin-right:0}.property-name[data-v-dd95f4d7]{font-family:var(--scalar-font-code);font-weight:var(--scalar-semibold);font-size:var(--scalar-font-size-3);display:flex}.property-additional[data-v-dd95f4d7]{font-family:var(--scalar-font-code)}.property-required[data-v-dd95f4d7],.property-optional[data-v-dd95f4d7]{color:var(--scalar-color-2)}.property-required[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-orange)}.property-read-only[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-blue)}.property-write-only[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-green)}.property-detail[data-v-dd95f4d7]{font-size:var(--scalar-micro);color:var(--scalar-color-2);display:flex;align-items:center;min-width:0}.property-const[data-v-dd95f4d7]{color:var(--scalar-color-1)}.property[data-v-f5c211dc]{color:var(--scalar-color-1);padding:10px;font-size:var(--scalar-mini)}.property[data-v-f5c211dc]:last-of-type{padding-bottom:0}.property--compact.property--level-0[data-v-f5c211dc]{padding:12px 0}.property--deprecated[data-v-f5c211dc]{background:repeating-linear-gradient(-45deg,var(--scalar-background-2) 0,var(--scalar-background-2) 2px,transparent 2px,transparent 5px);background-size:100%}.property--deprecated[data-v-f5c211dc]>*{opacity:.75}.property-description[data-v-f5c211dc]{margin-top:6px;line-height:1.4;font-size:var(--scalar-small)}.property-heading:empty+.property-description[data-v-f5c211dc]:last-of-type,.property-description[data-v-f5c211dc]:first-of-type:last-of-type{margin-top:0}[data-v-f5c211dc] .property-description *{color:var(--scalar-color-2)!important}.property[data-v-f5c211dc]:not(:last-of-type){border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.children[data-v-f5c211dc]{display:flex;flex-direction:column;padding-top:8px}.property-example[data-v-f5c211dc]{display:flex;flex-direction:column;margin-top:6px;max-height:calc(((var(--full-height) - var(--refs-header-height))) / 2);font-size:var(--scalar-micro);border:var(--scalar-border-width) solid var(--scalar-border-color);background:var(--scalar-background-2);border-radius:var(--scalar-radius-lg)}.property-example-label[data-v-f5c211dc]{font-weight:var(--scalar-semibold);color:var(--scalar-color-3);padding:6px}.property-example-value[data-v-f5c211dc]{all:unset;font-family:var(--scalar-font-code);padding:6px;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.property-rule[data-v-f5c211dc]{display:flex;flex-direction:column;gap:6px;margin-top:12px;border-radius:var(--scalar-radius-lg)}.property-enum-value[data-v-f5c211dc]{padding:3px 0;color:var(--scalar-color-2);line-height:1.5;word-break:break-word}.property-enum-value[data-v-f5c211dc]:before{content:"⊢";margin-right:6px;color:var(--scalar-color-3)}.property-enum-values[data-v-f5c211dc]{margin-top:8px;list-style:none}.property--compact .property-example[data-v-f5c211dc]{background:transparent;border:none;display:flex;flex-direction:row;gap:8px}.property--compact .property-example-label[data-v-f5c211dc],.property--compact .property-example-value[data-v-f5c211dc]{padding:3px 0 0}.property--compact .property-example-value[data-v-f5c211dc]{background:var(--scalar-background-2);border-top:0;border-radius:var(--scalar-radius);padding:3px 4px}.property-list[data-v-f5c211dc]{border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:var(--scalar-radius);margin-top:10px}.property-list .property[data-v-f5c211dc]:last-of-type{padding-bottom:10px}.property-name[data-v-f5c211dc]{font-family:var(--scalar-font-code);font-weight:var(--scalar-semibold)}.enum-toggle-button[data-v-f5c211dc]{align-items:center;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:13.5px;cursor:pointer;color:var(--scalar-color-2);display:flex;font-weight:var(--scalar-semibold);gap:4px;margin-top:8px;padding:6px 10px;-webkit-user-select:none;-moz-user-select:none;user-select:none;white-space:nowrap}.enum-toggle-button[data-v-f5c211dc]:hover{color:var(--scalar-color-1)}.enum-toggle-button-icon--open[data-v-f5c211dc]{transform:rotate(45deg)}.headers-card[data-v-2f99eae6]{z-index:0;margin-top:12px;margin-bottom:6px;position:relative;font-size:var(--scalar-font-size-4);color:var(--scalar-color-1);align-self:flex-start}.headers-card.headers-card--open[data-v-2f99eae6]{align-self:initial}.headers-card-title[data-v-2f99eae6]{padding:6px 10px;display:flex;align-items:center;gap:4px;color:var(--scalar-color-3);font-weight:var(--scalar-semibold);font-size:var(--scalar-micro);border-radius:13.5px}button.headers-card-title[data-v-2f99eae6]{cursor:pointer}button.headers-card-title[data-v-2f99eae6]:hover{color:var(--scalar-color-1)}.headers-card-title-icon--open[data-v-2f99eae6]{transform:rotate(45deg)}.headers-properties-open>.headers-card-title[data-v-2f99eae6]{border-bottom-left-radius:0;border-bottom-right-radius:0;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.headers-properties-open[data-v-2f99eae6]{width:100%}.headers-properties[data-v-2f99eae6]{display:flex;flex-direction:column;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:13.5px;width:-moz-fit-content;width:fit-content}.headers-card .property[data-v-2f99eae6]:last-of-type{padding-bottom:10px}.headers-card-title>.headers-card-title-icon[data-v-2f99eae6]{width:10px;height:10px;margin:0}.headers-card-title>.headers-card-title-icon--open[data-v-2f99eae6]{transform:rotate(45deg)}.parameter-item[data-v-e1efed5d]{display:flex;flex-direction:column;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.parameter-item:last-of-type .parameter-schema[data-v-e1efed5d]{padding-bottom:0}.parameter-item-container[data-v-e1efed5d]{padding:0}.parameter-item-headers[data-v-e1efed5d]{border:var(--scalar-border-width) solid var(--scalar-border-color)}.parameter-item-name[data-v-e1efed5d]{margin-right:6px;font-weight:var(--scalar-semibold);font-size:var(--scalar-font-size-3);font-family:var(--scalar-font-code);color:var(--scalar-color-1)}.parameter-item-type[data-v-e1efed5d]{font-size:var(--scalar-micro);color:var(--scalar-color-2);margin-right:6px;line-height:1.4;white-space:nowrap;text-overflow:ellipsis;width:100%;overflow:hidden}.parameter-item-trigger-open .parameter-item-type[data-v-e1efed5d]{white-space:normal}.parameter-item-trigger+.parameter-item-container[data-v-e1efed5d] .property--level-0>.property-heading .property-detail-value{font-size:var(--scalar-font-size-3)}.parameter-item-required-optional[data-v-e1efed5d]{color:var(--scalar-color-2);font-weight:var(--scalar-semibold);margin-right:6px;position:relative}.parameter-item--required[data-v-e1efed5d]{text-transform:uppercase;font-size:var(--scalar-micro);font-weight:var(--scalar-semibold);color:var(--scalar-color-orange)}.parameter-item-description[data-v-e1efed5d]{margin-top:3px!important;font-size:var(--scalar-small);color:var(--scalar-color-2);line-height:1.4}.parameter-item-description[data-v-e1efed5d] p{margin-top:4px;font-size:var(--scalar-small);color:var(--scalar-color-2);line-height:1.4}.parameter-schema[data-v-e1efed5d]{padding-bottom:9px;margin-top:3px}.parameter-item-trigger[data-v-e1efed5d]{padding:12px 0;cursor:pointer;outline:none;text-align:left;position:relative;align-items:baseline}.parameter-item-trigger-open[data-v-e1efed5d]{padding-bottom:0}.parameter-item-trigger[data-v-e1efed5d]:after{content:"";position:absolute;height:10px;width:100%;bottom:0}.parameter-item-icon[data-v-e1efed5d]{color:var(--scalar-color-3);height:18px;left:-19px;position:absolute;top:11px;width:18px}.parameter-item-trigger:hover .parameter-item-icon[data-v-e1efed5d],.parameter-item-trigger:focus-visible .parameter-item-icon[data-v-e1efed5d]{color:var(--scalar-color-1)}.parameter-item-trigger:focus-visible .parameter-item-icon[data-v-e1efed5d]{outline:1px solid var(--scalar-color-accent);outline-offset:2px;border-radius:var(--scalar-radius)}.parameter-item-content-type[data-v-e1efed5d]{margin-left:auto;opacity:0;transition:opacity .1s ease-in-out;color:var(--scalar-color-3);font-size:var(--scalar-micro);background:var(--scalar-background-2);padding:2px 6px;border-radius:12px}.parameter-item-trigger:hover .parameter-item-content-type[data-v-e1efed5d]{opacity:1}.parameter-list[data-v-b05cf5ff]{margin-top:24px}.parameter-list-title[data-v-b05cf5ff]{font-size:var(--scalar-font-size-2);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);line-height:1.45;margin-top:12px;margin-bottom:12px}.parameter-list-items[data-v-b05cf5ff]{list-style:none;padding:0;margin:0 0 12px;font-size:var(--scalar-small)}.request-body-title[data-v-dc8f9923]{display:flex;align-items:center;font-size:var(--scalar-font-size-2);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);line-height:1.45;margin-top:24px;padding-bottom:12px;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color);flex-flow:wrap}.request-body-title-select[data-v-dc8f9923]{position:relative;height:-moz-fit-content;height:fit-content;margin-left:auto;font-weight:var(--scalar-regular);display:flex;align-items:center;color:var(--scalar-color-3);font-size:var(--scalar-micro);background:var(--scalar-background-2);padding:2px 6px;border-radius:12px;border:var(--scalar-border-width) solid var(--scalar-border-color)}.request-body-title-no-select.request-body-title-select[data-v-dc8f9923]{pointer-events:none}.request-body-title-no-select[data-v-dc8f9923]{border:none}.request-body-title-no-select.request-body-title-select[data-v-dc8f9923]:after{display:none}.request-body-title-select span[data-v-dc8f9923]{display:flex;align-items:center}.request-body-title-select[data-v-dc8f9923]:after{content:"";width:6px;height:6px;transform:rotate(45deg) translate3d(0,-3px,0);display:block;margin-left:6px;box-shadow:1px 1px 0 currentColor;margin-right:5px}.request-body-title-select select[data-v-dc8f9923]{border:none;outline:none;cursor:pointer;background:var(--scalar-background-3);position:absolute;top:0;left:0;width:100%;height:100%;opacity:0;-webkit-appearance:none;-moz-appearance:none;appearance:none}.request-body-title-select[data-v-dc8f9923]:hover{color:var(--scalar-color-1)}.request-body-description[data-v-dc8f9923]{margin-top:6px;font-size:var(--scalar-small);width:100%}.request-body-description[data-v-dc8f9923] .markdown *{color:var(--scalar-color-2)!important}@media (max-width: 460px){.request-body-title-select[data-v-dc8f9923]{margin-left:auto;padding-right:3px}}.section-wrapper[data-v-07cabcb0]{color:var(--scalar-color-1);padding-top:12px;margin-top:-12px}.section-accordion[data-v-07cabcb0]{display:flex;flex-direction:column;border-radius:var(--scalar-radius-lg);background:var(--scalar-background-2)}.section-accordion-transparent[data-v-07cabcb0]{background:transparent;border:1px solid var(--scalar-border-color)}.section-accordion-button[data-v-07cabcb0]{padding:6px;display:flex;align-items:center;gap:6px;cursor:pointer}.section-accordion-button-content[data-v-07cabcb0]{flex:1;min-width:0}.section-accordion-button-actions[data-v-07cabcb0]{display:flex;align-items:center;gap:6px;color:var(--scalar-color-3)}.section-accordion-chevron[data-v-07cabcb0]{margin-right:4px;width:20px;cursor:pointer;opacity:1;color:var(--scalar-color-3)}.section-accordion-button:hover .section-accordion-chevron[data-v-07cabcb0]{color:var(--scalar-color-1)}.section-accordion-content[data-v-07cabcb0]{border-top:1px solid var(--scalar-border-color);display:flex;flex-direction:column}.section-accordion-description[data-v-07cabcb0]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color--1);padding:10px 12px 0}.section-accordion-content-card[data-v-07cabcb0] .property:last-of-type{padding-bottom:9px}.show-api-client-button[data-v-ea8189a2]{-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none;padding:4px 6px;white-space:nowrap;border-radius:var(--scalar-radius);display:flex;justify-content:center;align-items:center;font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-background-2);font-family:var(--scalar-font);background:var(--scalar-button-1);position:relative;cursor:pointer;box-sizing:border-box;box-shadow:inset 0 0 0 1px #0000001a;outline-offset:2px}.show-api-client-button span[data-v-ea8189a2],.show-api-client-button svg[data-v-ea8189a2]{fill:currentColor;color:var(--scalar-button-1-color);z-index:1}.show-api-client-button[data-v-ea8189a2]:hover{background:var(--scalar-button-1-hover)}.show-api-client-button svg[data-v-ea8189a2]{margin-right:4px}.empty-state[data-v-4b51e8ca]{margin:10px 0 10px 12px;text-align:center;font-size:var(--scalar-micro);min-height:56px;display:flex;align-items:center;justify-content:center;border-radius:var(--scalar-radius-lg);color:var(--scalar-color-2)}.rule-title[data-v-4b51e8ca]{font-family:var(--scalar-font-code);color:var(--scalar-color-1);display:inline-block;margin:12px 0 6px;border-radius:var(--scalar-radius)}.rule[data-v-4b51e8ca]{margin:0 12px;border-radius:var(--scalar-radius-lg)}.rule-items[data-v-4b51e8ca]{counter-reset:list-number;display:flex;flex-direction:column;gap:12px;border-left:1px solid var(--scalar-border-color);padding:12px 0}.rule-item[data-v-4b51e8ca]{counter-increment:list-number;border:1px solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);overflow:hidden;margin-left:24px}.rule-item[data-v-4b51e8ca]:before{border:1px solid var(--scalar-border-color);border-top:0;border-right:0;content:" ";display:block;width:24px;height:6px;border-radius:0 0 0 var(--scalar-radius-lg);margin-top:6px;color:var(--scalar-color-2);transform:translate(-25px);color:var(--scalar-color-1);position:absolute}.tab-list[data-v-03fdb072]{display:flex;gap:6px;position:relative;flex:1;padding:9px 12px;overflow:auto}.scalar-card-header.scalar-card-header-tabs[data-v-03fdb072]{padding:0}.tab[data-v-d0e07e65]{background:none;border:none;font-size:var(--scalar-mini);font-family:var(--scalar-font);color:var(--scalar-color-2);font-weight:var(--scalar-semibold);line-height:calc(var(--scalar-mini) + 2px);white-space:nowrap;cursor:pointer;padding:0;margin-right:3px;text-transform:uppercase;position:relative;line-height:1.35}.tab[data-v-d0e07e65]:before{content:"";position:absolute;z-index:0;left:-6px;top:-6px;width:calc(100% + 12px);height:calc(100% + 12px);border-radius:var(--scalar-radius);background:var(--scalar-background-3);opacity:0}.tab[data-v-d0e07e65]:hover:before,.tab[data-v-d0e07e65]:focus-visible:before{opacity:1}.tab[data-v-d0e07e65]:focus-visible:before{outline:1px solid var(--scalar-color-accent)}.tab span[data-v-d0e07e65]{z-index:1;position:relative}.tab-selected[data-v-d0e07e65]{color:var(--scalar-color-1);text-decoration:underline;text-underline-offset:var(--scalar-micro)}.scalar-card-footer[data-v-7621d1be]{flex-shrink:0}.text-select{position:relative;height:-moz-fit-content;height:fit-content}.text-select:has(:focus-visible) .text-select-label{outline:1px solid var(--scalar-color-accent);outline-offset:4px;border-radius:calc(var(--scalar-radius) / 2)}.text-select--single-option{pointer-events:none}.text-select select{border:none;outline:none;cursor:pointer;background:var(--scalar-background-3);box-shadow:-2px 0 0 0 var(--scalar-background-3);position:absolute;top:0;left:0;width:100%;height:100%;opacity:0;-moz-appearance:none;-webkit-appearance:none;appearance:none}.text-select span{font-size:var(--scalar-mini);color:var(--scalar-color-2);font-weight:var(--scalar-semibold);white-space:nowrap;display:flex;align-items:center;justify-content:center}.text-select:hover span{color:var(--scalar-color-1)}.text-select:not(.text-select--single-option) span:after{content:"";width:7px;height:7px;transform:rotate(45deg) translate3d(-2px,-2px,0);display:block;margin-left:6px;box-shadow:1px 1px 0 currentColor}.text-select span:hover{background:var(--scalar-background-2)}.example-selector[data-v-a7e22e14]{padding:4px}.markdown[data-v-d84b9039] *{margin:0}.code-copy[data-v-d84b9039]{display:flex;align-items:center;justify-content:center;-moz-appearance:none;appearance:none;-webkit-appearance:none;outline:none;background:transparent;cursor:pointer;color:var(--scalar-color-3);border:none;padding:0;margin-right:12px}.code-copy[data-v-d84b9039]:hover{color:var(--scalar-color-1)}.code-copy svg[data-v-d84b9039]{width:13px;height:13px}.response-card-footer[data-v-d84b9039]{display:flex;flex-direction:column;flex-shrink:0;padding:10px 12px;gap:8px;border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.response-example-selector[data-v-d84b9039]{align-self:flex-start;margin:-4px}.response-description[data-v-d84b9039]{font-weight:var(--scalar-semibold);font-size:var(--scalar-micro);color:var(--scalar-color--1);display:flex;align-items:center;box-sizing:border-box}.schema-type[data-v-d84b9039]{font-size:var(--scalar-micro);color:var(--scalar-color-2);font-weight:var(--scalar-semibold);background:var(--scalar-background-3);padding:2px 4px;border-radius:4px;margin-right:4px}.schema-example[data-v-d84b9039]{font-size:var(--scalar-micro);color:var(--scalar-color-2);font-weight:var(--scalar-semibold)}.example-response-tab[data-v-d84b9039]{display:block;margin:6px}.scalar-card-container[data-v-d84b9039]{flex:1;background:var(--scalar-background-2);display:grid}.scalar-card-container[data-v-d84b9039] .cm-scroller{overflow-y:hidden}.scalar-card-checkbox[data-v-d84b9039]{display:flex;align-items:center;justify-content:center;position:relative;min-height:17px;cursor:pointer;-webkit-user-select:none;-moz-user-select:none;user-select:none;font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-2);width:-moz-fit-content;width:fit-content;white-space:nowrap;margin-right:9px;gap:6px}.scalar-card-checkbox:has(.scalar-card-checkbox-input:focus-visible) .scalar-card-checkbox-checkmark[data-v-d84b9039]{outline:1px solid var(--scalar-color-accent)}.scalar-card-checkbox[data-v-d84b9039]:hover{color:var(--scalar-color--1)}.scalar-card-checkbox .scalar-card-checkbox-input[data-v-d84b9039]{position:absolute;opacity:0;cursor:pointer;height:0;width:0}.scalar-card-checkbox-checkmark[data-v-d84b9039]{height:17px;width:17px;border-radius:var(--scalar-radius);background-color:transparent;background-color:var(--scalar-background-3);box-shadow:inset 0 0 0 1px var(--scalar-border-color)}.scalar-card-checkbox[data-v-d84b9039]:has(.scalar-card-checkbox-input:checked){color:var(--scalar-color-1)}.scalar-card-checkbox .scalar-card-checkbox-input:checked~.scalar-card-checkbox-checkmark[data-v-d84b9039]{background-color:var(--scalar-button-1);box-shadow:none}.scalar-card-checkbox-checkmark[data-v-d84b9039]:after{content:"";position:absolute;display:none}.scalar-card-checkbox .scalar-card-checkbox-input:checked~.scalar-card-checkbox-checkmark[data-v-d84b9039]:after{display:block}.scalar-card-checkbox .scalar-card-checkbox-checkmark[data-v-d84b9039]:after{right:6px;top:36.5%;width:5px;height:9px;border:solid 1px var(--scalar-button-1-color);border-width:0 1.5px 1.5px 0;transform:rotate(45deg)}.request[data-v-3ff33ef0]{display:flex;flex-wrap:nowrap}.request-header[data-v-3ff33ef0]{display:flex;gap:6px;text-transform:initial}.request-method[data-v-3ff33ef0]{font-family:var(--scalar-font-code);text-transform:uppercase}.request-client-picker[data-v-3ff33ef0]{padding-left:12px;padding-right:9px}.request-card-footer[data-v-3ff33ef0]{display:flex;justify-content:flex-end;padding:6px;flex-shrink:0}.request-card-footer-addon[data-v-3ff33ef0]{display:flex;align-items:center;flex:1;min-width:0}.request-editor-section[data-v-3ff33ef0]{display:flex;flex:1}.request-card-simple[data-v-3ff33ef0]{display:flex;align-items:center;justify-content:space-between;padding:8px 8px 8px 12px;font-size:var(--scalar-small)}.code-snippet[data-v-3ff33ef0]{display:flex;flex-direction:column;width:100%}.operation-title[data-v-e104650f]{display:flex;justify-content:space-between}.operation-details[data-v-e104650f]{display:flex;align-items:center;margin-top:0;gap:9px;min-width:0;flex-shrink:1}.endpoint-type[data-v-e104650f]{display:flex;align-items:center;justify-content:center;gap:6px;position:relative;z-index:0;width:60px;padding:6px;flex-shrink:0;font-size:var(--scalar-small);text-transform:uppercase;font-weight:var(--scalar-bold);font-family:var(--scalar-font)}.endpoint-type[data-v-e104650f]:after{content:"";position:absolute;top:0;right:0;bottom:0;left:0;z-index:-1;background:currentColor;opacity:.15;border-radius:var(--scalar-radius-lg)}.endpoint-anchor[data-v-e104650f]{display:flex;align-items:center;min-width:0;flex-shrink:1;font-size:20px}.endpoint-anchor.label[data-v-e104650f]{display:flex}.endpoint-label[data-v-e104650f]{display:flex;align-items:baseline;gap:9px;min-width:0;flex-shrink:1;color:var(--scalar-color-1)}.endpoint-label-path[data-v-e104650f]{font-family:var(--scalar-font-code);font-size:var(--scalar-mini);text-overflow:ellipsis;white-space:nowrap;overflow:hidden}.endpoint-label-path[data-v-e104650f] em{color:var(--scalar-color-2)}.endpoint-label-name[data-v-e104650f]{color:var(--scalar-color-2);font-size:var(--scalar-small);flex-shrink:1000000000;text-overflow:ellipsis;white-space:nowrap;overflow:hidden}.endpoint-try-hint[data-v-e104650f]{padding:2px;height:24px;width:24px;flex-shrink:0}.endpoint-copy[data-v-e104650f]{color:currentColor;padding:2px}.endpoint-copy[data-v-e104650f] svg{stroke-width:2px}.endpoint-content[data-v-e104650f]{display:grid;grid-auto-columns:1fr;grid-auto-flow:row;gap:9px;padding:9px}@media (min-width: 1000px){.endpoint-content[data-v-e104650f]{grid-auto-flow:column}}@container (max-width: 900px){.endpoint-content[data-v-e104650f]{grid-template-columns:1fr}}.endpoint-content[data-v-e104650f]>*{max-height:unset}.operation-details-card[data-v-e104650f]{display:flex;flex-direction:column;gap:12px}.operation-details-card-item[data-v-e104650f] .parameter-list{border:1px solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);margin-top:0}.operation-details-card-item[data-v-e104650f]{display:flex;flex-direction:column;gap:12px}.operation-details-card-item[data-v-e104650f] .parameter-list-items{margin-bottom:0}.operation-details-card[data-v-e104650f] .parameter-item:last-of-type .parameter-schema{padding-bottom:12px}.operation-details-card[data-v-e104650f] .parameter-list .parameter-list{margin-bottom:12px}.operation-details-card[data-v-e104650f] .parameter-item{margin:0;padding:0 9px}.operation-details-card[data-v-e104650f] .property{padding:9px;margin:0}.operation-details-card[data-v-e104650f] .parameter-list-title,.operation-details-card[data-v-e104650f] .request-body-title{text-transform:uppercase;font-weight:var(--scalar-bold);font-size:var(--scalar-mini);color:var(--scalar-color-2);line-height:1.33;padding:9px;margin:0}.operation-details-card[data-v-e104650f] .request-body-title-select{text-transform:initial;font-weight:initial;margin-left:auto}.examples[data-v-0916dd9f]{position:sticky;top:calc(var(--refs-header-height) + 24px)}.deprecated[data-v-0916dd9f] *{text-decoration:line-through}.example-path[data-v-0916dd9f]{color:var(--scalar-color-2);font-family:var(--scalar-font-code)}.example-path[data-v-0916dd9f] em{color:var(--scalar-color-1);font-style:normal}.references-loading{position:absolute;top:0;left:0;right:0;z-index:1;grid-area:rendered;background:var(--scalar-background-1)}.references-loading-top-spacer{top:-1px}.references-loading-hidden-tag .section-container>.section:first-child{display:none}.download[data-v-1fc94525]{margin-bottom:24px}.download-button[data-v-1fc94525]{color:var(--scalar-link-color, var(--scalar-color-accent));font-weight:var(--scalar-link-font-weight, inherit);-webkit-text-decoration:var(--scalar-text-decoration)!important;text-decoration:var(--scalar-text-decoration)!important;text-decoration-color:var(--scalar-text-decoration-color);font-size:var(--scalar-paragraph);cursor:pointer}.download-button[data-v-1fc94525]:hover{color:var(--scalar-link-color-hover, var(--scalar-color-accent));-webkit-text-decoration:var(--scalar-text-decoration-hover)!important;text-decoration:var(--scalar-text-decoration-hover)!important;text-decoration-color:var(--scalar-text-decoration-color-hover)}.introduction-description-heading[data-v-fa3d93c5]{scroll-margin-top:64px;margin-top:1em;margin-bottom:.5em}.markdown+.markdown[data-v-fa3d93c5]{margin-top:1em}.introduction-description[data-v-fa3d93c5]{display:flex;flex-direction:column}.references-classic .introduction-description[data-v-fa3d93c5] img{max-width:720px}.sticky-cards[data-v-0336cd4f]{display:flex;flex-direction:column;position:sticky;top:calc(var(--refs-header-height) + 24px)}.client-libraries-content[data-v-b29373e8]{container:client-libraries-content / inline-size;display:flex;justify-content:center;overflow:hidden;padding:0 12px;background-color:var(--scalar-background-1);border-top:var(--scalar-border-width) solid var(--scalar-border-color)}.client-libraries[data-v-b29373e8]{display:flex;align-items:center;justify-content:center;width:100%;position:relative;cursor:pointer;white-space:nowrap;padding:8px 2px;gap:6px;color:var(--scalar-color-3);border-bottom:1px solid transparent;-webkit-user-select:none;-moz-user-select:none;user-select:none}.client-libraries[data-v-b29373e8]:first-child{border-radius:var(--scalar-radius) 0 0 0}.client-libraries[data-v-b29373e8]:not(.client-libraries__active):hover:before{content:"";position:absolute;width:calc(100% - 4px);height:calc(100% - 4px);background:var(--scalar-background-2);left:2px;top:2px;z-index:0;border-radius:var(--scalar-radius)}.client-libraries[data-v-b29373e8]:active{color:var(--scalar-color-1)}.client-libraries[data-v-b29373e8]:focus-visible{outline:none;box-shadow:inset 0 0 0 1px var(--scalar-color-accent)}@media screen and (max-width: 450px){.client-libraries[data-v-b29373e8]:nth-of-type(4),.client-libraries[data-v-b29373e8]:nth-of-type(5){display:none}}.client-libraries-icon[data-v-b29373e8]{max-width:14px;max-height:14px;min-width:14px;width:100%;aspect-ratio:1;display:flex;align-items:center;justify-content:center;position:relative;box-sizing:border-box;color:currentColor}.client-libraries-icon__more svg[data-v-b29373e8]{height:initial}@container client-libraries-content (width < 400px){.client-libraries__select[data-v-b29373e8]{width:-moz-fit-content;width:fit-content}.client-libraries__select .client-libraries-icon__more+span[data-v-b29373e8]{display:none}}@container client-libraries-content (width < 380px){.client-libraries[data-v-b29373e8]{width:100%}.client-libraries span[data-v-b29373e8]{display:none}}.client-libraries__active[data-v-b29373e8]{color:var(--scalar-color-1);border-bottom:1px solid var(--scalar-color-1)}@keyframes codeloader-b29373e8{0%{transform:rotate(0)}to{transform:rotate(1turn)}}.client-libraries .client-libraries-text[data-v-b29373e8]:last-of-type{font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);position:relative;display:flex;align-items:center}.client-libraries__active .client-libraries-text[data-v-b29373e8]{color:var(--scalar-color-1)}.client-libraries__select select[data-v-b29373e8]{background:var(--scalar-background-3);color:var(--scalar-color-2);opacity:0;height:100%;width:100%;aspect-ratio:1;position:absolute;top:0;left:0;cursor:pointer;z-index:1;-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none}.client-libraries__select[data-v-b29373e8]:has(select:focus-visible){border-radius:var(--scalar-radius);box-shadow:inset 0 0 0 1px var(--scalar-color-accent)}@media screen and (max-width: 600px){.references-classic .client-libraries[data-v-b29373e8]{flex-direction:column}}.selected-client[data-v-35fd541a]{color:var(--scalar-color-1);font-size:var(--scalar-mini);font-family:var(--scalar-font-code);padding:9px 12px;border-top:none;white-space:nowrap;overflow:hidden;text-overflow:ellipsis;background:var(--scalar-background-1);border-top:var(--scalar-border-width) solid var(--scalar-border-color);min-height:-moz-fit-content;min-height:fit-content}.client-libraries-heading[data-v-35fd541a]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-1);padding:9px 12px;background-color:var(--scalar-background-2);display:flex;align-items:center;max-height:32px}.section-accordion-wrapper[data-v-1a30b4ff]{padding:0 60px}.section-accordion[data-v-1a30b4ff]{position:relative;width:100%;max-width:var(--refs-content-max-width);margin:auto}.section-accordion-content[data-v-1a30b4ff]{display:flex;flex-direction:column;gap:12px;padding-top:12px}.section-accordion-button[data-v-1a30b4ff]{width:100%;display:flex;cursor:pointer;padding:6px 0;margin:-6px 0;border-radius:var(--scalar-radius)}.section-accordion-chevron[data-v-1a30b4ff]{width:24px;position:absolute;left:-24px;height:auto;top:10px;color:var(--scalar-color-3)}.section-accordion-button:hover .section-accordion-chevron[data-v-1a30b4ff]{color:var(--scalar-color-1)}.section-accordion-title[data-v-1a30b4ff]{display:flex;flex-direction:column;align-items:flex-start;flex:1;padding:0 6px}.section-accordion-title[data-v-1a30b4ff] .section-header-wrapper{grid-template-columns:1fr}.section-accordion-title[data-v-1a30b4ff] .section-header{margin-bottom:0}@container narrow-references-container (max-width: 900px){.section-accordion-chevron[data-v-1a30b4ff]{width:16px;left:-16px;top:14px}.section-accordion-wrapper[data-v-1a30b4ff]{padding:calc(var(--refs-header-height)) 24px 0 24px}}.tag-section[data-v-43e1c813]{margin-bottom:48px}.tag-name[data-v-43e1c813]{text-transform:capitalize}.tag-description[data-v-43e1c813]{padding-bottom:4px;text-align:left}.webhook-request-body[data-v-cbaa1001]{margin-top:-18px}.collapsible-section[data-v-b32ceff4]{border-top:var(--scalar-border-width) solid var(--scalar-border-color);position:relative}.collapsible-section-header[data-v-b32ceff4]{color:var(--scalar-color-1)}.collapsible-section .collapsible-section-trigger[data-v-b32ceff4]{display:flex;align-items:center;cursor:pointer;padding:10px 0;font-size:var(--scalar-font-size-3);z-index:1;position:relative}.collapsible-section .collapsible-section-trigger[data-v-b32ceff4]:after{content:"";height:10px;width:100%;position:absolute;bottom:0}.collapsible-section-trigger svg[data-v-b32ceff4]{color:var(--scalar-color-3);position:absolute;left:-19px}.collapsible-section:hover .collapsible-section-trigger svg[data-v-b32ceff4]{color:var(--scalar-color-1)}.collapsible-section .collapsible-section-trigger[data-v-b32ceff4] .anchor-copy{line-height:18.5px}.collapsible-section-content[data-v-b32ceff4]{padding:0;margin:0;scroll-margin-top:140px}.webhooks-list[data-v-8cee6fce]{display:contents}.webhooks-list-truncated .webhooks-list-item[data-v-8cee6fce]:last-child{border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.webhooks-list-item[aria-expanded=true][data-v-8cee6fce]{padding-bottom:48px}.webhook-section-description.webhook-section-description[data-v-8cee6fce]{color:var(--scalar-color-2);font-size:var(--scalar-font-size-3);padding-bottom:12px}.show-more-webhooks[data-v-8cee6fce]{margin-top:32px;top:0}.reference-models[data-v-4396a256]{margin-bottom:48px}.reference-models-anchor[data-v-4396a256]{display:flex;align-items:center;font-size:20px;padding-left:6px;color:var(--scalar-color-1)}.reference-models-label[data-v-4396a256]{font-size:var(--scalar-mini)}.reference-models-label[data-v-4396a256] em{font-weight:var(--scalar-bold)}.models-list[data-v-7c1cd4b4]{display:contents}.models-list-truncated .models-list-item[data-v-7c1cd4b4]:last-child{border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.show-more-models[data-v-7c1cd4b4]{margin-top:32px;top:0}.narrow-references-container{container-name:narrow-references-container;container-type:inline-size}.render-loading[data-v-42cfaf8d]{height:calc(var(--full-height) - var(--refs-header-height));display:flex;align-items:center;justify-content:center}.introduction-card[data-v-42cfaf8d]{display:flex;flex-direction:column;background:var(--scalar-background-1)}.introduction-card-item[data-v-42cfaf8d]{display:flex;overflow:hidden;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:var(--scalar-radius-lg);margin-bottom:12px;flex-direction:column;justify-content:start}@container narrow-references-container (max-width: 900px){.introduction-card-item[data-v-42cfaf8d]{border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}}.introduction-card-item[data-v-42cfaf8d]:has(.description) .server-form-container{border-bottom-left-radius:0;border-bottom-right-radius:0}.introduction-card-item[data-v-42cfaf8d] .request-item{border-bottom:0}.introduction-card-title[data-v-42cfaf8d]{font-weight:var(--scalar-semibold);font-size:var(--scalar-mini);color:var(--scalar-color-3)}.introduction-card-row[data-v-42cfaf8d]{gap:24px}@media (min-width: 600px){.introduction-card-row[data-v-42cfaf8d]{flex-flow:row wrap}}.introduction-card-row[data-v-42cfaf8d]>*{flex:1}@media (min-width: 600px){.introduction-card-row[data-v-42cfaf8d]>*{min-width:-moz-min-content;min-width:min-content}}@media (max-width: 600px){.introduction-card-row[data-v-42cfaf8d]>*{max-width:100%}}@container (max-width: 900px){.introduction-card-row[data-v-42cfaf8d]{flex-direction:column;align-items:stretch;gap:0px}}.introduction-card[data-v-42cfaf8d] .security-scheme-label{text-transform:uppercase;font-weight:var(--scalar-semibold)}.references-classic .introduction-card-row[data-v-42cfaf8d] .scalar-card:nth-of-type(2) .scalar-card-header{display:none}.references-classic .introduction-card-row[data-v-42cfaf8d] .scalar-card:nth-of-type(2) .scalar-card-header.scalar-card--borderless+.scalar-card-content{margin-top:0}.section-flare[data-v-42cfaf8d]{top:0;right:0;pointer-events:none}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}}@layer scalar-base{:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>");mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>")}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}}@layer scalar-base{:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-api-reference{color:var(--scalar-color-1)}.scalar-api-reference .container{width:100%}@media (min-width: 400px){.scalar-api-reference .container{max-width:400px}}@media (min-width: 600px){.scalar-api-reference .container{max-width:600px}}@media (min-width: 800px){.scalar-api-reference .container{max-width:800px}}@media (min-width: 1000px){.scalar-api-reference .container{max-width:1000px}}@media (min-width: 1200px){.scalar-api-reference .container{max-width:1200px}}.scalar-api-reference .row{display:flex;flex-direction:row}.scalar-api-reference .sr-only{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);white-space:nowrap;border-width:0}.scalar-api-reference .visible{visibility:visible}.scalar-api-reference .collapse{visibility:collapse}.scalar-api-reference .static{position:static}.scalar-api-reference .fixed{position:fixed}.scalar-api-reference .absolute{position:absolute}.scalar-api-reference .relative{position:relative}.scalar-api-reference .sticky{position:sticky}.scalar-api-reference .right-0{right:0}.scalar-api-reference .block{display:block}.scalar-api-reference .flex{display:flex}.scalar-api-reference .grid{display:grid}.scalar-api-reference .contents{display:contents}.scalar-api-reference .hidden{display:none}.scalar-api-reference .h-8{height:32px}.scalar-api-reference .flex-shrink,.scalar-api-reference .shrink{flex-shrink:1}.scalar-api-reference .transform{transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-api-reference .resize{resize:both}.scalar-api-reference .flex-wrap{flex-wrap:wrap}.scalar-api-reference .content-start{align-content:flex-start}.scalar-api-reference .content-end{align-content:flex-end}.scalar-api-reference .items-center{align-items:center}.scalar-api-reference .gap-12{gap:48px}.scalar-api-reference .divide-y>:not([hidden])~:not([hidden]){--tw-divide-y-reverse:0;border-top-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-y-reverse)));border-bottom-width:calc(var(--scalar-border-width) * var(--tw-divide-y-reverse))}.scalar-api-reference .truncate{overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.scalar-api-reference .border{border-width:var(--scalar-border-width)}.scalar-api-reference .bg-b-2{background-color:var(--scalar-background-2)}.scalar-api-reference .px-3{padding-left:12px;padding-right:12px}.scalar-api-reference .py-1{padding-top:4px;padding-bottom:4px}.scalar-api-reference .py-1\.5{padding-top:6px;padding-bottom:6px}.scalar-api-reference .py-2{padding-top:8px;padding-bottom:8px}.scalar-api-reference .py-2\.5{padding-top:10px;padding-bottom:10px}.scalar-api-reference .font{font-family:var(--scalar-font)}.scalar-api-reference .text-sm{font-size:var(--scalar-font-size-4)}.scalar-api-reference .font{font-weight:var(--scalar-regular)}.scalar-api-reference .font-medium{font-weight:var(--scalar-semibold)}.scalar-api-reference .uppercase{text-transform:uppercase}.scalar-api-reference .lowercase{text-transform:lowercase}.scalar-api-reference .capitalize{text-transform:capitalize}.scalar-api-reference .leading{line-height:1.5}.scalar-api-reference .text-c-2{color:var(--scalar-color-2)}.scalar-api-reference .text-c-3{color:var(--scalar-color-3)}.scalar-api-reference .text-green{color:var(--scalar-color-green)}.scalar-api-reference .text-orange{color:var(--scalar-color-orange)}.scalar-api-reference .text-red{color:var(--scalar-color-red)}.scalar-api-reference .underline{text-decoration-line:underline}.scalar-api-reference .outline{outline-style:solid}.scalar-api-reference .filter{filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-api-reference .transition{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,-webkit-backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter,-webkit-backdrop-filter;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-api-reference .\[--scalar-address-bar-height\:0px\]{--scalar-address-bar-height:0px}.scalar-api-reference .hover\:text-c-1:hover{color:var(--scalar-color-1)}.nav-item[data-v-d56f12d9]{padding:0 1rem;cursor:pointer;flex:1;justify-content:center;align-items:center;display:flex;border-radius:var(--scalar-radius-lg);background:var(--scalar-background-3);border:var(--scalar-border-width) solid var(--scalar-background-2);color:var(--scalar-color-3);padding:4.5px;min-width:0;overflow:hidden;position:relative}.dark-mode .nav-item[data-v-d56f12d9]{background:color-mix(in srgb,var(--scalar-background-2),transparent)}.nav-item-icon-copy[data-v-d56f12d9]{max-width:100%;white-space:nowrap;overflow:hidden;-webkit-mask-image:linear-gradient(to left,transparent 0,var(--scalar-background-2) 20px);mask-image:linear-gradient(to left,transparent 0,var(--scalar-background-2) 20px)}.nav-item:hover .nav-item-icon-copy[data-v-d56f12d9]{-webkit-mask-image:linear-gradient(to left,transparent 20px,var(--scalar-background-2) 40px);mask-image:linear-gradient(to left,transparent 20px,var(--scalar-background-2) 40px)}.nav-item-copy[data-v-d56f12d9]{max-width:calc(100% - 20px)}.nav-item[data-v-d56f12d9]:hover{color:var(--scalar-color-1)}.nav-item__active[data-v-d56f12d9]{background-color:var(--scalar-background-1);color:var(--scalar-color-1);border-color:var(--scalar-border-color)}.dark-mode .nav-item__active[data-v-d56f12d9]{background-color:var(--scalar-background-2)}.nav-item-close[data-v-d56f12d9]{position:absolute;right:3px;padding:2px;border-radius:var(--scalar-radius);background:transparent;max-width:20px;stroke-width:1.5px;color:var(--scalar-color-3);margin-left:-20px;opacity:0}.nav-item:hover .nav-item-close[data-v-d56f12d9]{opacity:1}.nav-item-close[data-v-d56f12d9]:hover{background-color:var(--scalar-background-4)}.nav-item__active .nav-item-close[data-v-d56f12d9]:hover{background-color:var(--scalar-background-2)}.download-app-button[data-v-c896207d]{box-shadow:0 0 0 .5px var(--scalar-border-color);background:linear-gradient(#ffffffbf,#00000009)}.dark-mode .download-app-button[data-v-c896207d]{background:linear-gradient(#ffffff1a,#00000026)}.download-app-button[data-v-c896207d]:hover{background:linear-gradient(#00000009,#ffffffbf)}.dark-mode .download-app-button[data-v-c896207d]:hover{background:linear-gradient(#00000026,#ffffff1a)}.http-bg-gradient[data-v-46723816]{background:linear-gradient(#ffffffbf,#00000009)}.http-bg-gradient[data-v-46723816]:hover{background:linear-gradient(#00000009,#ffffffbf)}.dark-mode .http-bg-gradient[data-v-46723816]{background:linear-gradient(#ffffff09,#00000026)}.dark-mode .http-bg-gradient[data-v-46723816]:hover{background:linear-gradient(#00000026,#ffffff09)}.commandmenu[data-v-e9fd9234]{box-shadow:var(--scalar-shadow-2);border-radius:var(--scalar-radius-lg);background-color:var(--scalar-background-1);max-height:60dvh;width:100%;max-width:580px;margin:12px;position:fixed;left:50%;top:150px;opacity:0;transform:translate3d(-50%,10px,0);animation:fadeincommandmenu-e9fd9234 ease-in-out .3s forwards;animation-delay:.1s}.commandmenu-overlay[data-v-e9fd9234]{position:fixed;top:0;right:0;bottom:0;left:0;background:#0003;animation:fadeincommand-e9fd9234 ease-in-out .3s forwards;cursor:pointer}@keyframes fadeincommand-e9fd9234{0%{opacity:0}to{opacity:1}}@keyframes fadeincommandmenu-e9fd9234{0%{opacity:0;transform:translate3d(-50%,10px,0)}to{opacity:1;transform:translate3d(-50%,0,0)}}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>");mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>")}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>");mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>")}@layer scalar-base{*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar .scalar-app-layout[data-v-cc5e3e81]{background:var(--scalar-background-1);height:calc(100% - 120px);max-width:1390px;width:100%;margin:auto;opacity:0;animation:scalarapiclientfadein-cc5e3e81 .35s forwards;position:relative;overflow:hidden;border-radius:8px;border:var(--scalar-border-width) solid var(--scalar-border-color)}@keyframes scalarapiclientfadein-cc5e3e81{0%{opacity:0}to{opacity:1}}.scalar .scalar-app-exit[data-v-cc5e3e81]{position:fixed;top:0;left:0;width:100vw;height:100vh;background:#00000038;transition:all .3s ease-in-out;cursor:pointer;animation:scalardrawerexitfadein-cc5e3e81 .35s forwards}.dark-mode .scalar .scalar-app-exit[data-v-cc5e3e81]{background:#00000073}.scalar .scalar-app-exit[data-v-cc5e3e81]:before{font-family:sans-serif;position:absolute;top:0;font-size:30px;font-weight:100;line-height:50px;right:12px;text-align:center;color:#fff;opacity:.6}.scalar .scalar-app-exit[data-v-cc5e3e81]:hover:before{opacity:1}@keyframes scalardrawerexitfadein-cc5e3e81{0%{opacity:0}to{opacity:1}}.scalar-container[data-v-cc5e3e81]{overflow:hidden;visibility:visible;position:fixed;bottom:0;top:0;left:0;width:100%;height:100%;display:flex;align-items:center;justify-content:center;z-index:10000}.scalar .url-form-input[data-v-cc5e3e81]{min-height:auto!important}.scalar .scalar-container[data-v-cc5e3e81]{line-height:normal}.scalar .scalar-app-header span[data-v-cc5e3e81]{color:var(--scalar-color-3)}.scalar .scalar-app-header a[data-v-cc5e3e81]{color:var(--scalar-color-1)}.scalar .scalar-app-header a[data-v-cc5e3e81]:hover{text-decoration:underline}.scalar-activate[data-v-cc5e3e81]{width:-moz-fit-content;width:fit-content;margin:0px .75rem .75rem auto;line-height:24px;cursor:pointer;font-size:.875rem;font-weight:600;display:flex;align-items:center;gap:6px}.scalar-activate-button[data-v-cc5e3e81]{display:flex;gap:6px;align-items:center;color:var(--scalar-color-blue);-webkit-appearance:none;-moz-appearance:none;appearance:none;outline:none;border:none;background:transparent;padding:0 .5rem}.scalar-activate:hover .scalar-activate-button[data-v-cc5e3e81]{background:var(--scalar-background-3);border-radius:3px}[data-v-4d46a9ea] .cm-editor{height:100%;outline:none;padding:0;background:transparent}[data-v-4d46a9ea] .cm-placeholder{color:var(--scalar-color-3)}[data-v-4d46a9ea] .cm-content{font-family:var(--scalar-font-code);font-size:var(--scalar-small);max-height:20px;padding:8px 0}[data-v-4d46a9ea] .cm-tooltip{background:transparent!important;filter:brightness(var(--scalar-lifted-brightness));border-radius:var(--scalar-radius);box-shadow:var(--scalar-shadow-2);border:none!important;outline:none!important;overflow:hidden!important}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul li{padding:3px 6px!important}[data-v-4d46a9ea] .cm-completionIcon-type:after{color:var(--scalar-color-3)!important}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul li[aria-selected]{background:var(--scalar-background-2)!important;color:var(--scalar-color-1)!important}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul{padding:6px!important;position:relative}[data-v-4d46a9ea] .cm-tooltip-autocomplete ul li:hover{border-radius:3px;color:var(--scalar-color-1)!important;background:var(--scalar-background-3)!important}[data-v-4d46a9ea] .cm-activeLine,[data-v-4d46a9ea] .cm-activeLineGutter{background-color:transparent}[data-v-4d46a9ea] .cm-selectionMatch,[data-v-4d46a9ea] .cm-matchingBracket{border-radius:var(--scalar-radius);background:var(--scalar-background-4)!important}[data-v-4d46a9ea] .cm-css-color-picker-wrapper{display:inline-flex;outline:1px solid var(--scalar-background-3);border-radius:3px;overflow:hidden}[data-v-4d46a9ea] .cm-gutters{background-color:transparent;border-right:none;color:var(--scalar-color-3);font-size:var(--scalar-mini);line-height:1.44;border-radius:0 0 0 3px}[data-v-4d46a9ea] .cm-gutters:before{content:"";position:absolute;top:2px;left:2px;width:calc(100% - 2px);height:calc(100% - 4px);border-radius:var(--scalar-radius) 0 0 var(--scalar-radius);background-color:var(--scalar-background-1)}[data-v-4d46a9ea] .cm-gutterElement{font-family:var(--scalar-font-code)!important;padding:0 6px 0 8px!important;display:flex;align-items:center;justify-content:flex-end;position:relative}[data-v-4d46a9ea] .cm-gutter+.cm-gutter :not(.cm-foldGutter) .cm-gutterElement{padding-left:0!important}[data-v-4d46a9ea] .cm-scroller{overflow:auto}.peer:hover .copy-button[data-v-4d46a9ea],.copy-button[data-v-4d46a9ea]:focus-visible{opacity:100}.scalar-code-copy[data-v-4d46a9ea]{align-items:flex-start;display:flex;top:0;right:0;bottom:0;left:0;justify-content:flex-end;position:sticky}.copy-button[data-v-4d46a9ea]{align-items:center;display:flex;position:relative;background-color:var(--scalar-background-2);border:1px solid var(--scalar-border-color);border-radius:3px;color:var(--scalar-color-3);cursor:pointer;height:30px;margin-bottom:-30px;opacity:0;padding:6px;transition:opacity .15s ease-in-out,color .15s ease-in-out;top:0;right:0}.scalar-code-copy[data-v-4d46a9ea],.copy-button[data-v-4d46a9ea]{background:inherit}.copy-button[data-v-4d46a9ea]:hover{color:var(--scalar-color-1)}.copy-button svg[data-v-4d46a9ea]{stroke-width:1.5}.cm-pill{--tw-bg-base: var(--scalar-color-1);color:var(--tw-bg-base);padding:0 9px;display:inline-block;border-radius:30px;font-size:var(--scalar-mini);background:color-mix(in srgb,var(--tw-bg-base),transparent 94%)!important}.cm-pill.bg-grey{background:var(--scalar-background-3)!important}.dark-mode .cm-pill{background:color-mix(in srgb,var(--tw-bg-base),transparent 80%)!important}.cm-pill:first-of-type{margin-left:0}.cm-editor .cm-widgetBuffer{display:none}.cm-foldPlaceholder:hover{color:var(--scalar-color-1)}.cm-foldGutter .cm-gutterElement{font-size:var(--scalar-heading-4);padding:2px!important}.cm-foldGutter .cm-gutterElement:first-of-type{display:none}.cm-foldGutter .cm-gutterElement .cm-foldMarker{padding:2px}.cm-foldGutter .cm-gutterElement:hover .cm-foldMarker{background:var(--scalar-background-2);border-radius:var(--scalar-radius);color:var(--scalar-color-1)}[data-v-cb0ff5b1] .cm-editor{padding:0}[data-v-cb0ff5b1] .cm-content{align-items:center;background-color:transparent;display:flex;font-family:var(--scalar-font);font-size:var(--scalar-mini);padding:6px 8px}[data-v-cb0ff5b1] .cm-content:has(.cm-pill){padding:4px 3px}[data-v-cb0ff5b1] .cm-content .cm-pill:not(:last-of-type){margin-right:.5px}[data-v-cb0ff5b1] .cm-content .cm-pill:not(:first-of-type){margin-left:.5px}[data-v-cb0ff5b1] .cm-line{padding:0}.required[data-v-cb0ff5b1]:after{content:"Required"}input[data-v-cb0ff5b1]::-moz-placeholder{color:var(--scalar-color-3)}input[data-v-cb0ff5b1]::placeholder{color:var(--scalar-color-3)}.scalar-modal-layout .scalar-button-danger[data-v-b6a60dbc]{background:color-mix(in srgb,var(--scalar-color-red),transparent 95%);color:var(--scalar-color-red)}.scalar-modal-layout .scalar-button-danger[data-v-b6a60dbc]:hover,.scalar-modal-layout .scalar-button-danger[data-v-b6a60dbc]:focus{background:color-mix(in srgb,var(--scalar-color-red),transparent 90%)}.fade-request-section-content[data-v-7906217f]{background:linear-gradient(to left,var(--scalar-background-1) 64%,transparent)}.filter-hover[data-v-7906217f]{height:100%;padding-right:42px;padding-left:24px;position:absolute;right:0;transition:width 0s ease-in-out .2s;overflow:hidden}.filter-hover[data-v-7906217f]:hover,.filter-hover[data-v-7906217f]:has(:focus-visible){width:100%;z-index:10}.filter-hover[data-v-7906217f]:before{content:"";position:absolute;top:0;left:0;width:100%;height:100%;background-color:var(--scalar-background-1);opacity:0;transition:all .3s ease-in-out;pointer-events:none}.filter-hover-item[data-v-7906217f]{opacity:0}.filter-hover-item[data-v-7906217f]:not(:last-of-type){transform:translate3d(0,3px,0)}.filter-hover:hover .filter-hover-item[data-v-7906217f]{transition:opacity .2s ease-in-out,transform .2s ease-in-out}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(1){transition-delay:.05s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(2){transition-delay:.1s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(3){transition-delay:.15s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(4){transition-delay:.2s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(5){transition-delay:.25s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(6){transition-delay:.3s}.filter-hover:hover .filter-hover-item[data-v-7906217f]:nth-last-of-type(7){transition-delay:.35s}.filter-hover:hover .filter-hover-item[data-v-7906217f],.filter-hover:has(:focus-visible) .filter-hover-item[data-v-7906217f]{opacity:1;transform:translateZ(0)}.filter-hover[data-v-7906217f]:hover:before,.filter-hover[data-v-7906217f]:has(:focus-visible):before{opacity:.9;-webkit-backdrop-filter:blur(10px);backdrop-filter:blur(10px)}.context-bar-group:hover .context-bar-group-hover\:text-c-1[data-v-7906217f],.context-bar-group:has(:focus-visible) .context-bar-group-hover\:text-c-1[data-v-7906217f]{--tw-text-opacity: 1;color:rgb(var(--scalar-color-1) / var(--tw-text-opacity))}.context-bar-group:hover .context-bar-group-hover\:hidden[data-v-7906217f],.context-bar-group:has(:focus-visible) .context-bar-group-hover\:hidden[data-v-7906217f]{display:none}.light-mode .bg-preview[data-v-5f138d26]{background-image:url("data:image/svg+xml,%3Csvg xmlns=%27http://www.w3.org/2000/svg%27 width=%2716%27 height=%2716%27 fill=%27%23000%27 fill-opacity=%2710%25%27%3E%3Crect width=%278%27 height=%278%27 /%3E%3Crect x=%278%27 y=%278%27 width=%278%27 height=%278%27 /%3E%3C/svg%3E")}.dark-mode .bg-preview[data-v-5f138d26]{background-image:url("data:image/svg+xml,%3Csvg xmlns=%27http://www.w3.org/2000/svg%27 width=%2716%27 height=%2716%27 fill=%27%23FFF%27 fill-opacity=%2710%25%27%3E%3Crect width=%278%27 height=%278%27 /%3E%3Crect x=%278%27 y=%278%27 width=%278%27 height=%278%27 /%3E%3C/svg%3E")}[data-v-99361e9b] .cm-editor{background-color:transparent;font-size:var(--scalar-mini);outline:none}[data-v-99361e9b] .cm-gutters{background-color:var(--scalar-background-1);border-radius:var(--scalar-radius) 0 0 var(--scalar-radius)}.scalar-code-copy[data-v-99361e9b]{align-items:flex-start;display:flex;position:absolute;top:6px;right:6px;z-index:10;pointer-events:none;position:sticky;transform:translate(-6px)}.copy-button[data-v-99361e9b]{align-items:center;display:flex;background-color:var(--scalar-background-1);border:1px solid var(--scalar-border-color);border-radius:3px;color:var(--scalar-color-3);cursor:pointer;height:30px;opacity:0;padding:6px;pointer-events:auto;transition:opacity .15s ease-in-out,color .15s ease-in-out}.body-raw:hover .copy-button[data-v-99361e9b],.copy-button[data-v-99361e9b]:focus-visible{opacity:1}.copy-button[data-v-99361e9b]:hover{color:var(--scalar-color-1)}.scalar-code-block[data-v-cf80aa59] .hljs *{font-size:var(--scalar-mini)}.ascii-art-animate .ascii-art-line[data-v-0018c948]{border-right:1ch solid transparent;animation:typewriter-0018c948 4s steps(1) 1s 1 normal both,blinkTextCursor-0018c948 .5s steps(1) infinite normal}@keyframes typewriter-0018c948{0%{width:0}to{width:100%}}@keyframes blinkTextCursor-0018c948{0%{border-right-color:currentColor}50%{border-right-color:transparent}}.keycap-n[data-v-6c848a44]{background:-webkit-linear-gradient(5deg,transparent 30%,var(--scalar-color-3) 50%);-webkit-background-clip:text;-webkit-text-fill-color:transparent}.keycap-hotkey[data-v-6c848a44]{position:absolute;top:32px;line-height:26px}.scalar-version-number[data-v-04a8f835]{transform:skew(0,13deg);width:76px;height:76px;position:absolute;margin-left:-36px;font-size:8px;font-family:var(--scalar-font-code);line-height:11px;margin-top:-113px;border-radius:9px 9px 16px 12px;box-shadow:inset 2px 0 0 2px var(--scalar-background-2);text-align:center;display:flex;flex-direction:column;align-items:center;justify-content:center;text-transform:initial;text-decoration-color:var(--scalar-color-3)}.scalar-version-number a[data-v-04a8f835]{font-weight:700;background:var(--scalar-background-2);padding:2px 4px;border-radius:3px;text-decoration:none;border:.5px solid var(--scalar-border-color)}.gitbook-show[data-v-04a8f835]{display:none}.v-enter-active[data-v-ef5f83d8]{transition:opacity .5s ease}.v-enter-from[data-v-ef5f83d8]{opacity:0}.response-body-virtual[data-headlessui-state=open],.response-body-virtual[data-headlessui-state=open] .diclosure-panel{display:flex;flex-direction:column;flex-grow:1}.animate-response-heading .response-heading[data-v-ec732740]{animation:push-response-ec732740 .2s ease-in-out forwards;opacity:1}@keyframes push-response-ec732740{0%{opacity:1;transform:translateY(0)}to{opacity:0;transform:translateY(-4px)}}.animate-response-heading .animate-response-children[data-v-ec732740]{animation:response-spans-ec732740 .2s ease-in-out forwards .05s;opacity:0}@keyframes response-spans-ec732740{0%{opacity:0;transform:translateY(4px)}to{opacity:1;transform:translateY(0)}}.schema>span[data-v-30c955c6]:not(:first-child):before{content:"·";display:block;margin:0 .5ch}.schema>span[data-v-30c955c6]{display:flex;white-space:nowrap}[data-v-ccef6266] .cm-editor{padding:0}[data-v-ccef6266] .cm-content{align-items:center;background-color:transparent;display:flex;font-family:var(--scalar-font);font-size:var(--scalar-mini);padding:6px 8px}[data-v-ccef6266] .cm-content:has(.cm-pill){padding:4px 3px}[data-v-ccef6266] .cm-content .cm-pill:not(:last-of-type){margin-right:.5px}[data-v-ccef6266] .cm-content .cm-pill:not(:first-of-type){margin-left:.5px}[data-v-ccef6266] .cm-line{padding:0}.filemask[data-v-ccef6266]{-webkit-mask-image:linear-gradient(to right,transparent 0,var(--scalar-background-2) 20px);mask-image:linear-gradient(to right,transparent 0,var(--scalar-background-2) 20px)}[data-v-3f1cd888] .cm-content{font-size:var(--scalar-mini)}.auth-blend-required[data-v-2b678dcb] .scalar-input-required,.auth-blend-required[data-v-2b678dcb] .required{background:var(--scalar-background-2);--tw-bg-base: var(--scalar-background-2);--tw-shadow: -8px 0 4px var(--scalar-background-2)}.request-example-references-header[data-v-2b678dcb]+tr>td{border-top:0;border-top-left-radius:0;border-top-right-radius:0}.auth-combobox-position[data-v-3569c0b2]{margin-left:120px}.scroll-timeline-x[data-v-3569c0b2]{overflow:auto;scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;scrollbar-width:none;-ms-overflow-style:none}.fade-left[data-v-3569c0b2],.fade-right[data-v-3569c0b2]{position:sticky;content:"";height:100%;animation-name:fadein-3569c0b2;animation-duration:1ms;animation-direction:reverse;animation-timeline:--scroll-timeline;min-height:24px;pointer-events:none}.fade-left[data-v-3569c0b2]{background:linear-gradient(-90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);min-width:3px;left:-1px;animation-direction:normal}.fade-right[data-v-3569c0b2]{background:linear-gradient(90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);margin-left:-20px;min-width:24px;right:-1px;top:0}@keyframes fadein-3569c0b2{0%{opacity:0}15%{opacity:1}}.auth-combobox-position[data-v-3dc33a43]{margin-left:120px}.scroll-timeline-x[data-v-3dc33a43]{overflow:auto;scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;scrollbar-width:none;-ms-overflow-style:none}.fade-left[data-v-3dc33a43],.fade-right[data-v-3dc33a43]{position:sticky;content:"";height:100%;animation-name:fadein-3dc33a43;animation-duration:1ms;animation-direction:reverse;animation-timeline:--scroll-timeline;min-height:24px;pointer-events:none}.fade-left[data-v-3dc33a43]{background:linear-gradient(-90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);min-width:3px;left:-1px;animation-direction:normal}.fade-right[data-v-3dc33a43]{background:linear-gradient(90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 60%,var(--scalar-background-1) 100%);margin-left:-20px;min-width:24px;right:-1px;top:0}@keyframes fadein-3dc33a43{0%{opacity:0}15%{opacity:1}}[data-v-c2f2c62b] code.hljs *{font-size:var(--scalar-mini)}.request-section-content[data-v-5ab558f8]{--scalar-border-width: .5px}.request-section-content-filter[data-v-5ab558f8]{box-shadow:0 -10px 0 10px var(--scalar-background-1)}.request-item:focus-within .request-meta-buttons[data-v-5ab558f8]{opacity:1}.group-hover-input[data-v-5ab558f8]{border-width:var(--scalar-border-width);border-color:transparent}.group:hover .group-hover-input[data-v-5ab558f8]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2));border-color:var(--scalar-border-color)}.group-hover-input[data-v-5ab558f8]:focus{background:transparent!important;border-color:var(--scalar-border-color)!important}@media (min-width: 800px){.has-no-import-url,.has-import-url{max-width:100dvw;overflow-x:hidden;contain:paint}.has-no-import-url{opacity:1;background:var(--scalar-background-1);animation:transform-restore-layout ease-in-out .3s forwards}.has-import-url .scalar-client>main{opacity:0;transform:scale(.85) translate3d(calc(50dvw + 80px),0,0);animation:transform-fade-layout ease-in-out .3s forwards;border:var(--scalar-border-width) solid var(--scalar-border-color);border-radius:12px;overflow:hidden;z-index:10000}.has-import-url .scalar-client .sidenav{display:none}.has-no-import-url .scalar-app,.has-import-url .scalar-app{background:var(--scalar-background-1)!important}}@keyframes transform-fade-layout{0%{opacity:0;transform:scale(.85) translate3d(calc(50dvw + 80px),10px,0)}to{opacity:1;transform:scale(.85) translate3d(calc(50dvw + 80px),0,0)}}@keyframes transform-restore-layout{0%{opacity:1;transform:scale(.85) translate3d(calc(50dvw + 80px),0,0)}to{opacity:1;transform:scale(1) translateZ(0)}}.openapi-color{color:var(--scalar-color-green)}.section-flare{position:fixed;top:0;right:-50dvw}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x: 0;--tw-border-spacing-y: 0;--tw-translate-x: 0;--tw-translate-y: 0;--tw-rotate: 0;--tw-skew-x: 0;--tw-skew-y: 0;--tw-scale-x: 1;--tw-scale-y: 1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness: proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width: 0px;--tw-ring-offset-color: #fff;--tw-ring-color: rgb(59 130 246 / .5);--tw-ring-offset-shadow: 0 0 #0000;--tw-ring-shadow: 0 0 #0000;--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}:where(.scalar-app) .container{width:100%}@media (min-width: 400px){:where(.scalar-app) .container{max-width:400px}}@media (min-width: 600px){:where(.scalar-app) .container{max-width:600px}}@media (min-width: 800px){:where(.scalar-app) .container{max-width:800px}}@media (min-width: 1000px){:where(.scalar-app) .container{max-width:1000px}}@media (min-width: 1200px){:where(.scalar-app) .container{max-width:1200px}}:where(.scalar-app) .sr-only{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);white-space:nowrap;border-width:0}:where(.scalar-app) .pointer-events-none{pointer-events:none}:where(.scalar-app) .visible{visibility:visible}:where(.scalar-app) .invisible{visibility:hidden}:where(.scalar-app) .fixed{position:fixed}:where(.scalar-app) .absolute{position:absolute}:where(.scalar-app) .relative{position:relative}:where(.scalar-app) .sticky{position:sticky}:where(.scalar-app) .inset-0{top:0;right:0;bottom:0;left:0}:where(.scalar-app) .left-0{left:0}:where(.scalar-app) .left-2{left:8px}:where(.scalar-app) .left-2\.5{left:10px}:where(.scalar-app) .left-4{left:16px}:where(.scalar-app) .left-border{left:var(--scalar-border-width)}:where(.scalar-app) .left-px{left:1px}:where(.scalar-app) .right-2{right:8px}:where(.scalar-app) .top-0{top:0}:where(.scalar-app) .top-1\/2{top:50%}:where(.scalar-app) .top-12{top:48px}:where(.scalar-app) .top-2{top:8px}:where(.scalar-app) .top-22{top:88px}:where(.scalar-app) .top-\[0\.5px\]{top:.5px}:where(.scalar-app) .top-px{top:1px}:where(.scalar-app) .-z-1{z-index:-1}:where(.scalar-app) .z-1{z-index:1}:where(.scalar-app) .z-\[1001\]{z-index:1001}:where(.scalar-app) .z-context{z-index:1000}:where(.scalar-app) .\!m-0{margin:0!important}:where(.scalar-app) .-m-px{margin:-1px}:where(.scalar-app) .m-0{margin:0}:where(.scalar-app) .m-auto{margin:auto}:where(.scalar-app) .-mx-0{margin-left:-0px;margin-right:-0px}:where(.scalar-app) .-mx-0\.75{margin-left:-3px;margin-right:-3px}:where(.scalar-app) .-mx-px{margin-left:-1px;margin-right:-1px}:where(.scalar-app) .-my-1{margin-top:-4px;margin-bottom:-4px}:where(.scalar-app) .-my-1\.5{margin-top:-6px;margin-bottom:-6px}:where(.scalar-app) .mx-2{margin-left:8px;margin-right:8px}:where(.scalar-app) .mx-auto{margin-left:auto;margin-right:auto}:where(.scalar-app) .mx-px{margin-left:1px;margin-right:1px}:where(.scalar-app) .my-0{margin-top:0;margin-bottom:0}:where(.scalar-app) .my-0\.75{margin-top:3px;margin-bottom:3px}:where(.scalar-app) .my-3{margin-top:12px;margin-bottom:12px}:where(.scalar-app) .-mb-1{margin-bottom:-4px}:where(.scalar-app) .-ml-0{margin-left:-0px}:where(.scalar-app) .-ml-0\.75{margin-left:-3px}:where(.scalar-app) .-mr-0{margin-right:-0px}:where(.scalar-app) .-mr-0\.25{margin-right:-1px}:where(.scalar-app) .mb-0{margin-bottom:0}:where(.scalar-app) .ml-1{margin-left:4px}:where(.scalar-app) .ml-auto{margin-left:auto}:where(.scalar-app) .mr-2{margin-right:8px}:where(.scalar-app) .mr-3{margin-right:12px}:where(.scalar-app) .mt-0{margin-top:0}:where(.scalar-app) .mt-1{margin-top:4px}:where(.scalar-app) .mt-1\.5{margin-top:6px}:where(.scalar-app) .mt-2{margin-top:8px}:where(.scalar-app) .mt-20{margin-top:80px}:where(.scalar-app) .line-clamp-4{overflow:hidden;display:-webkit-box;-webkit-box-orient:vertical;-webkit-line-clamp:4}:where(.scalar-app) .block{display:block}:where(.scalar-app) .inline-block{display:inline-block}:where(.scalar-app) .flex{display:flex}:where(.scalar-app) .table{display:table}:where(.scalar-app) .grid{display:grid}:where(.scalar-app) .contents{display:contents}:where(.scalar-app) .hidden{display:none}:where(.scalar-app) .aspect-square{aspect-ratio:1 / 1}:where(.scalar-app) .size-10{width:40px;height:40px}:where(.scalar-app) .size-2{width:8px;height:8px}:where(.scalar-app) .size-3{width:12px;height:12px}:where(.scalar-app) .size-3\.5{width:14px;height:14px}:where(.scalar-app) .size-4{width:16px;height:16px}:where(.scalar-app) .size-5{width:20px;height:20px}:where(.scalar-app) .size-6{width:24px;height:24px}:where(.scalar-app) .size-8{width:32px;height:32px}:where(.scalar-app) .size-\[23px\]{width:23px;height:23px}:where(.scalar-app) .size-full{width:100%;height:100%}:where(.scalar-app) .h-10{height:40px}:where(.scalar-app) .h-3{height:12px}:where(.scalar-app) .h-3\.5{height:14px}:where(.scalar-app) .h-32{height:128px}:where(.scalar-app) .h-4{height:16px}:where(.scalar-app) .h-5{height:20px}:where(.scalar-app) .h-6{height:24px}:where(.scalar-app) .h-8{height:32px}:where(.scalar-app) .h-\[100dvh\]{height:100dvh}:where(.scalar-app) .h-border{height:var(--scalar-border-width)}:where(.scalar-app) .h-fit{height:-moz-fit-content;height:fit-content}:where(.scalar-app) .h-full{height:100%}:where(.scalar-app) .h-screen{height:100vh}:where(.scalar-app) .max-h-\[440px\]{max-height:440px}:where(.scalar-app) .max-h-\[calc\(100dvh-120px\)\]{max-height:calc(100dvh - 120px)}:where(.scalar-app) .max-h-\[calc\(100dvh-180px\)\]{max-height:calc(100dvh - 180px)}:where(.scalar-app) .max-h-\[calc\(100dvh-240px\)\]{max-height:calc(100dvh - 240px)}:where(.scalar-app) .max-h-\[inherit\]{max-height:inherit}:where(.scalar-app) .max-h-dvh{max-height:100dvh}:where(.scalar-app) .max-h-radix-popper{max-height:calc(var(--radix-popper-available-height) - 8px)}:where(.scalar-app) .min-h-0{min-height:0px}:where(.scalar-app) .min-h-12{min-height:48px}:where(.scalar-app) .min-h-5{min-height:20px}:where(.scalar-app) .min-h-6{min-height:24px}:where(.scalar-app) .min-h-8{min-height:32px}:where(.scalar-app) .min-h-96{min-height:384px}:where(.scalar-app) .min-h-\[77px\]{min-height:77px}:where(.scalar-app) .min-h-header{min-height:48px}:where(.scalar-app) .w-12{width:48px}:where(.scalar-app) .w-16{width:64px}:where(.scalar-app) .w-20{width:80px}:where(.scalar-app) .w-24{width:96px}:where(.scalar-app) .w-3{width:12px}:where(.scalar-app) .w-32{width:128px}:where(.scalar-app) .w-4{width:16px}:where(.scalar-app) .w-40{width:160px}:where(.scalar-app) .w-48{width:192px}:where(.scalar-app) .w-5{width:20px}:where(.scalar-app) .w-56{width:224px}:where(.scalar-app) .w-6{width:24px}:where(.scalar-app) .w-64{width:256px}:where(.scalar-app) .w-72{width:288px}:where(.scalar-app) .w-8{width:32px}:where(.scalar-app) .w-\[100dvw\]{width:100dvw}:where(.scalar-app) .w-\[38px\]{width:38px}:where(.scalar-app) .w-auto{width:auto}:where(.scalar-app) .w-full{width:100%}:where(.scalar-app) .w-min{width:-moz-min-content;width:min-content}:where(.scalar-app) .min-w-0{min-width:0px}:where(.scalar-app) .min-w-5{min-width:20px}:where(.scalar-app) .min-w-6{min-width:24px}:where(.scalar-app) .min-w-8{min-width:32px}:where(.scalar-app) .min-w-max{min-width:-moz-max-content;min-width:max-content}:where(.scalar-app) .min-w-min{min-width:-moz-min-content;min-width:min-content}:where(.scalar-app) .max-w-screen-lg{max-width:800px}:where(.scalar-app) .max-w-screen-md{max-width:640px}:where(.scalar-app) .max-w-screen-sm{max-width:540px}:where(.scalar-app) .max-w-screen-xl{max-width:1000px}:where(.scalar-app) .max-w-screen-xs{max-width:480px}:where(.scalar-app) .max-w-screen-xxs{max-width:360px}:where(.scalar-app) .max-w-xs{max-width:320px}:where(.scalar-app) .flex-1{flex:1 1 0%}:where(.scalar-app) .shrink-0{flex-shrink:0}:where(.scalar-app) .origin-top-left{transform-origin:top left}:where(.scalar-app) .-translate-y-1\/2{--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .translate-x-2{--tw-translate-x: 8px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .translate-x-2\.5{--tw-translate-x: 10px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .translate-x-\[14px\]{--tw-translate-x: 14px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .rotate-90{--tw-rotate: 90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .transform{transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .cursor-auto{cursor:auto}:where(.scalar-app) .cursor-default{cursor:default}:where(.scalar-app) .cursor-not-allowed{cursor:not-allowed}:where(.scalar-app) .cursor-pointer{cursor:pointer}:where(.scalar-app) .resize-none{resize:none}:where(.scalar-app) .resize{resize:both}:where(.scalar-app) .appearance-none{-webkit-appearance:none;-moz-appearance:none;appearance:none}:where(.scalar-app) .grid-cols-\[auto_1fr\]{grid-template-columns:auto 1fr}:where(.scalar-app) .flex-col{flex-direction:column}:where(.scalar-app) .flex-wrap{flex-wrap:wrap}:where(.scalar-app) .items-start{align-items:flex-start}:where(.scalar-app) .items-center{align-items:center}:where(.scalar-app) .justify-start{justify-content:flex-start}:where(.scalar-app) .justify-end{justify-content:flex-end}:where(.scalar-app) .justify-center{justify-content:center}:where(.scalar-app) .justify-between{justify-content:space-between}:where(.scalar-app) .gap-0{gap:0px}:where(.scalar-app) .gap-0\.5{gap:2px}:where(.scalar-app) .gap-0\.75{gap:3px}:where(.scalar-app) .gap-1{gap:4px}:where(.scalar-app) .gap-1\.5{gap:6px}:where(.scalar-app) .gap-16{gap:64px}:where(.scalar-app) .gap-2{gap:8px}:where(.scalar-app) .gap-2\.5{gap:10px}:where(.scalar-app) .gap-3{gap:12px}:where(.scalar-app) .gap-4{gap:16px}:where(.scalar-app) .gap-x-4{-moz-column-gap:16px;column-gap:16px}:where(.scalar-app) .gap-y-8{row-gap:32px}:where(.scalar-app) .divide-y>:not([hidden])~:not([hidden]){--tw-divide-y-reverse: 0;border-top-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-y-reverse)));border-bottom-width:calc(var(--scalar-border-width) * var(--tw-divide-y-reverse))}:where(.scalar-app) .self-center{align-self:center}:where(.scalar-app) .overflow-auto{overflow:auto}:where(.scalar-app) .overflow-hidden{overflow:hidden}:where(.scalar-app) .overflow-visible{overflow:visible}:where(.scalar-app) .overflow-y-scroll{overflow-y:scroll}:where(.scalar-app) .overscroll-contain{overscroll-behavior:contain}:where(.scalar-app) .truncate{overflow:hidden;text-overflow:ellipsis;white-space:nowrap}:where(.scalar-app) .text-ellipsis{text-overflow:ellipsis}:where(.scalar-app) .whitespace-nowrap{white-space:nowrap}:where(.scalar-app) .text-wrap{text-wrap:wrap}:where(.scalar-app) .text-nowrap{text-wrap:nowrap}:where(.scalar-app) .rounded{border-radius:var(--scalar-radius)}:where(.scalar-app) .rounded-full{border-radius:9999px}:where(.scalar-app) .rounded-lg{border-radius:var(--scalar-radius-lg)}:where(.scalar-app) .rounded-md{border-radius:var(--scalar-radius)}:where(.scalar-app) .rounded-none{border-radius:0}:where(.scalar-app) .rounded-xl{border-radius:var(--scalar-radius-xl)}:where(.scalar-app) .rounded-t-none{border-top-left-radius:0;border-top-right-radius:0}:where(.scalar-app) .border{border-width:var(--scalar-border-width)}:where(.scalar-app) .border-0{border-width:0}:where(.scalar-app) .border-2{border-width:calc(var(--scalar-border-width) * 2)}:where(.scalar-app) .border-\[1px\]{border-width:1px}:where(.scalar-app) .border-b{border-bottom-width:var(--scalar-border-width)}:where(.scalar-app) .border-r{border-right-width:var(--scalar-border-width)}:where(.scalar-app) .border-t{border-top-width:var(--scalar-border-width)}:where(.scalar-app) .border-t-1\/2{border-top-width:calc(var(--scalar-border-width) / 2)}:where(.scalar-app) .border-solid{border-style:solid}:where(.scalar-app) .border-dashed{border-style:dashed}:where(.scalar-app) .border-none{border-style:none}:where(.scalar-app) .border-border{border-color:var(--scalar-border-color)}:where(.scalar-app) .border-c-1{border-color:var(--scalar-color-1)}:where(.scalar-app) .border-c-3{border-color:var(--scalar-color-3)}:where(.scalar-app) .border-red{border-color:var(--scalar-color-red)}:where(.scalar-app) .border-sidebar-border{border-color:var(--scalar-sidebar-border-color, var(--scalar-border-color))}:where(.scalar-app) .bg-b-1{background-color:var(--scalar-background-1)}:where(.scalar-app) .bg-b-2{background-color:var(--scalar-background-2)}:where(.scalar-app) .bg-b-3{background-color:var(--scalar-background-3)}:where(.scalar-app) .bg-b-btn{background-color:var(--scalar-button-1)}:where(.scalar-app) .bg-backdrop{background-color:#00000038}:where(.scalar-app) .bg-blue{background-color:var(--scalar-color-blue)}:where(.scalar-app) .bg-border{background-color:var(--scalar-border-color)}:where(.scalar-app) .bg-c-accent{background-color:var(--scalar-color-accent)}:where(.scalar-app) .bg-c-danger{background-color:var(--scalar-color-danger)}:where(.scalar-app) .bg-red{background-color:var(--scalar-color-red)}:where(.scalar-app) .bg-sidebar-b-1{background-color:var(--scalar-sidebar-background-1, var(--scalar-background-1))}:where(.scalar-app) .bg-transparent{background-color:transparent}:where(.scalar-app) .bg-white{--tw-bg-opacity: 1;background-color:rgb(255 255 255 / var(--tw-bg-opacity))}:where(.scalar-app) .p-0{padding:0}:where(.scalar-app) .p-0\.5{padding:2px}:where(.scalar-app) .p-0\.75{padding:3px}:where(.scalar-app) .p-1{padding:4px}:where(.scalar-app) .p-1\.5{padding:6px}:where(.scalar-app) .p-2{padding:8px}:where(.scalar-app) .p-3{padding:12px}:where(.scalar-app) .p-4{padding:16px}:where(.scalar-app) .p-6{padding:24px}:where(.scalar-app) .px-1{padding-left:4px;padding-right:4px}:where(.scalar-app) .px-1\.5{padding-left:6px;padding-right:6px}:where(.scalar-app) .px-2{padding-left:8px;padding-right:8px}:where(.scalar-app) .px-2\.5{padding-left:10px;padding-right:10px}:where(.scalar-app) .px-3{padding-left:12px;padding-right:12px}:where(.scalar-app) .px-4{padding-left:16px;padding-right:16px}:where(.scalar-app) .px-6{padding-left:24px;padding-right:24px}:where(.scalar-app) .px-9{padding-left:36px;padding-right:36px}:where(.scalar-app) .py-1{padding-top:4px;padding-bottom:4px}:where(.scalar-app) .py-1\.5{padding-top:6px;padding-bottom:6px}:where(.scalar-app) .py-2{padding-top:8px;padding-bottom:8px}:where(.scalar-app) .py-2\.5{padding-top:10px;padding-bottom:10px}:where(.scalar-app) .py-3{padding-top:12px;padding-bottom:12px}:where(.scalar-app) .py-4{padding-top:16px;padding-bottom:16px}:where(.scalar-app) .pb-0{padding-bottom:0}:where(.scalar-app) .pl-8{padding-left:32px}:where(.scalar-app) .pl-\[114px\]{padding-left:114px}:where(.scalar-app) .pl-\[24px\]{padding-left:24px}:where(.scalar-app) .pl-\[42px\]{padding-left:42px}:where(.scalar-app) .pl-\[60px\]{padding-left:60px}:where(.scalar-app) .pl-\[6px\]{padding-left:6px}:where(.scalar-app) .pl-\[78px\]{padding-left:78px}:where(.scalar-app) .pl-\[96px\]{padding-left:96px}:where(.scalar-app) .pr-3{padding-right:12px}:where(.scalar-app) .pt-3{padding-top:12px}:where(.scalar-app) .pt-6{padding-top:24px}:where(.scalar-app) .text-left{text-align:left}:where(.scalar-app) .font-code{font-family:var(--scalar-font-code)}:where(.scalar-app) .text-3xs{font-size:var(--scalar-font-size-7)}:where(.scalar-app) .text-lg{font-size:var(--scalar-font-size-2)}:where(.scalar-app) .text-sm{font-size:var(--scalar-font-size-4)}:where(.scalar-app) .text-xl{font-size:var(--scalar-font-size-1)}:where(.scalar-app) .text-xs{font-size:var(--scalar-font-size-5)}:where(.scalar-app) .text-xxs{font-size:var(--scalar-font-size-6)}:where(.scalar-app) .font-bold{font-weight:var(--scalar-bold)}:where(.scalar-app) .font-medium{font-weight:var(--scalar-semibold)}:where(.scalar-app) .font-normal{font-weight:var(--scalar-regular)}:where(.scalar-app) .leading-3{line-height:var(--scalar-line-height-3)}:where(.scalar-app) .leading-none{line-height:1}:where(.scalar-app) .leading-snug{line-height:1.375}:where(.scalar-app) .text-b-1{color:var(--scalar-background-1)}:where(.scalar-app) .text-blue{color:var(--scalar-color-blue)}:where(.scalar-app) .text-c-1{color:var(--scalar-color-1)}:where(.scalar-app) .text-c-2{color:var(--scalar-color-2)}:where(.scalar-app) .text-c-3{color:var(--scalar-color-3)}:where(.scalar-app) .text-c-accent{color:var(--scalar-color-accent)}:where(.scalar-app) .text-c-btn{color:var(--scalar-button-1-color)}:where(.scalar-app) .text-c-danger{color:var(--scalar-color-danger)}:where(.scalar-app) .text-c-ghost{color:var(--scalar-color-ghost)}:where(.scalar-app) .text-red{color:var(--scalar-color-red)}:where(.scalar-app) .text-sidebar-c-2{color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}:where(.scalar-app) .text-transparent{color:transparent}:where(.scalar-app) .text-white{--tw-text-opacity: 1;color:rgb(255 255 255 / var(--tw-text-opacity))}:where(.scalar-app) .no-underline{text-decoration-line:none}:where(.scalar-app) .opacity-0{opacity:0}:where(.scalar-app) .opacity-40{opacity:.4}:where(.scalar-app) .opacity-50{opacity:.5}:where(.scalar-app) .shadow{--tw-shadow: var(--scalar-shadow-1);--tw-shadow-colored: var(--scalar-shadow-1);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-border{--tw-shadow: inset 0 0 0 1px var(--scalar-border-color);--tw-shadow-colored: inset 0 0 0 1px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-lg{--tw-shadow: var(--scalar-shadow-2);--tw-shadow-colored: var(--scalar-shadow-2);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-none{--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-sm{--tw-shadow: rgba(0, 0, 0, .09) 0px 1px 4px;--tw-shadow-colored: 0px 1px 4px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .shadow-border{--tw-shadow-color: var(--scalar-border-color);--tw-shadow: var(--tw-shadow-colored)}:where(.scalar-app) .shadow-current{--tw-shadow-color: currentColor;--tw-shadow: var(--tw-shadow-colored)}:where(.scalar-app) .outline-none{outline:2px solid transparent;outline-offset:2px}:where(.scalar-app) .outline{outline-style:solid}:where(.scalar-app) .-outline-offset-1{outline-offset:-1px}:where(.scalar-app) .outline-offset-1{outline-offset:1px}:where(.scalar-app) .blur{--tw-blur: blur(8px);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .brightness-lifted{--tw-brightness: brightness(var(--scalar-lifted-brightness));filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .filter{filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .backdrop-blur{--tw-backdrop-blur: blur(8px);-webkit-backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia);backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia)}:where(.scalar-app) .transition{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,-webkit-backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter,-webkit-backdrop-filter;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .transition-colors{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .transition-opacity{transition-property:opacity;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .transition-transform{transition-property:transform;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}:where(.scalar-app) .duration-100{transition-duration:.1s}:where(.scalar-app) .duration-300{transition-duration:.3s}:where(.scalar-app) .ease-in-out{transition-timing-function:cubic-bezier(.4,0,.2,1)}:where(.scalar-app) .bg-mix-transparent{background-color:color-mix(var(--tw-bg-mix-method, in srgb),transparent calc(var(--tw-bg-mix-amount, 0) * 1%),var(--tw-bg-base))}:where(.scalar-app) .bg-mix-amount-25{--tw-bg-mix-amount: 25}:where(.scalar-app) .bg-b-1{--tw-bg-base: var(--scalar-background-1)}:where(.scalar-app) .bg-b-2{--tw-bg-base: var(--scalar-background-2)}:where(.scalar-app) .bg-b-3{--tw-bg-base: var(--scalar-background-3)}:where(.scalar-app) .bg-b-btn{--tw-bg-base: var(--scalar-button-1)}:where(.scalar-app) .bg-backdrop{--tw-bg-base: rgba(0, 0, 0, .22)}:where(.scalar-app) .bg-blue{--tw-bg-base: var(--scalar-color-blue)}:where(.scalar-app) .bg-border{--tw-bg-base: var(--scalar-border-color)}:where(.scalar-app) .bg-c-accent{--tw-bg-base: var(--scalar-color-accent)}:where(.scalar-app) .bg-c-danger{--tw-bg-base: var(--scalar-color-danger)}:where(.scalar-app) .bg-red{--tw-bg-base: var(--scalar-color-red)}:where(.scalar-app) .bg-sidebar-b-1{--tw-bg-base: var(--scalar-sidebar-background-1, var(--scalar-background-1))}:where(.scalar-app) .bg-transparent{--tw-bg-base: transparent}:where(.scalar-app) .bg-white{--tw-bg-opacity: 1;--tw-bg-base: rgb(255 255 255 / var(--tw-bg-opacity))}:where(.scalar-app) .row,:where(.scalar-app) .scalar-row{display:flex;flex-direction:row}:where(.scalar-app) .col{display:flex;flex-direction:column}:where(.scalar-app) .centered-x{left:50%;--tw-translate-x: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .centered{left:50%;--tw-translate-x: -50%;top:50%;--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .\*\:flex>*{display:flex}:where(.scalar-app) .\*\:flex-1>*{flex:1 1 0%}:where(.scalar-app) .\*\:items-center>*{align-items:center}:where(.scalar-app) .\*\:gap-px>*{gap:1px}:where(.scalar-app) .hover\:bg-b-2:hover{background-color:var(--scalar-background-2)}:where(.scalar-app) .hover\:bg-b-3:hover{background-color:var(--scalar-background-3)}:where(.scalar-app) .hover\:bg-h-btn:hover{background-color:var(--scalar-button-1-hover)}:where(.scalar-app) .hover\:text-c-1:hover{color:var(--scalar-color-1)}:where(.scalar-app) .hover\:underline:hover{text-decoration-line:underline}:where(.scalar-app) .hover\:bg-b-2:hover{--tw-bg-base: var(--scalar-background-2)}:where(.scalar-app) .hover\:bg-b-3:hover{--tw-bg-base: var(--scalar-background-3)}:where(.scalar-app) .hover\:bg-h-btn:hover{--tw-bg-base: var(--scalar-button-1-hover)}:where(.scalar-app) .focus\:outline-none:focus{outline:2px solid transparent;outline-offset:2px}:where(.scalar-app) .focus-visible\:border-c-btn:focus-visible{border-color:var(--scalar-button-1-color)}:where(.scalar-app) .focus-visible\:ring-1:focus-visible{--tw-ring-offset-shadow: var(--tw-ring-inset) 0 0 0 var(--tw-ring-offset-width) var(--tw-ring-offset-color);--tw-ring-shadow: var(--tw-ring-inset) 0 0 0 calc(1px + var(--tw-ring-offset-width)) var(--tw-ring-color);box-shadow:var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow, 0 0 #0000)}:where(.scalar-app) .focus-visible\:ring-c-accent:focus-visible{--tw-ring-color: var(--scalar-color-accent)}:where(.scalar-app) .active\:bg-b-btn:active{background-color:var(--scalar-button-1)}:where(.scalar-app) .active\:text-c-1:active{color:var(--scalar-color-1)}:where(.scalar-app) .active\:shadow-none:active{--tw-shadow: 0 0 #0000;--tw-shadow-colored: 0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}:where(.scalar-app) .active\:brightness-90:active{--tw-brightness: brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .active\:bg-b-btn:active{--tw-bg-base: var(--scalar-button-1)}:where(.scalar-app) .group\/button:hover .group-hover\/button\:text-c-1,:where(.scalar-app) .group:hover .group-hover\:text-c-1{color:var(--scalar-color-1)}:where(.scalar-app) .group\/item:hover .group-hover\/item\:opacity-100{opacity:1}:where(.scalar-app) .group\/toggle:focus-visible .group-focus-visible\/toggle\:outline{outline-style:solid}:where(.scalar-app) .peer:checked~.peer-checked\:text-c-2{color:var(--scalar-color-2)}:where(.scalar-app) .peer:checked~.peer-checked\:opacity-100{opacity:1}:where(.scalar-app) .has-\[\:focus-visible\]\:bg-b-1:has(:focus-visible){background-color:var(--scalar-background-1)}:where(.scalar-app) .has-\[\:focus-visible\]\:outline:has(:focus-visible){outline-style:solid}:where(.scalar-app) .has-\[input\:focus-visible\]\:outline:has(input:focus-visible){outline-style:solid}:where(.scalar-app) .has-\[\:focus-visible\]\:bg-b-1:has(:focus-visible){--tw-bg-base: var(--scalar-background-1)}:where(.scalar-app) .ui-open\:rotate-180[data-headlessui-state~=open]{--tw-rotate: 180deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) :where([data-headlessui-state~=open]) .ui-open\:rotate-180{--tw-rotate: 180deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}:where(.scalar-app) .has-actv-btn\:border:has(button:active){border-width:var(--scalar-border-width)}:where(.scalar-app) .has-actv-btn\:border-border:has(button:active){border-color:var(--scalar-border-color)}:where(.scalar-app) .\!empty\:flex:not(:empty){display:flex}:where(.scalar-app) .\!empty\:w-7:not(:empty){width:28px}:where(.scalar-app) .\!empty\:items-center:not(:empty){align-items:center}:where(.scalar-app) .\!empty\:pr-3:not(:empty){padding-right:12px}:where(.scalar-app) .hocus\:text-c-1:hover{color:var(--scalar-color-1)}:where(.scalar-app) .hocus\:brightness-90:hover{--tw-brightness: brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}:where(.scalar-app) .hocus\:text-c-1:focus-visible{color:var(--scalar-color-1)}:where(.scalar-app) .hocus\:brightness-90:focus-visible{--tw-brightness: brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}@media (min-width: 800px){:where(.scalar-app) .md\:w-\[calc\(100vw-16px\)\]{width:calc(100vw - 16px)}:where(.scalar-app) .md\:row{display:flex;flex-direction:row}}@media (min-width: 1000px){:where(.scalar-app) .lg\:w-\[calc\(100vw-32px\)\]{width:calc(100vw - 32px)}:where(.scalar-app) .lg\:w-full{width:100%}}:where(.scalar-app) .dark\:bg-b-3:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-3)}:where(.scalar-app) .dark\:bg-backdropdark:where(.dark-mode,.dark-mode *){background-color:#00000073}:where(.scalar-app) .dark\:bg-b-3:where(.dark-mode,.dark-mode *){--tw-bg-base: var(--scalar-background-3)}:where(.scalar-app) .dark\:bg-backdropdark:where(.dark-mode,.dark-mode *){--tw-bg-base: rgba(0, 0, 0, .45)}:where(.scalar-app) .dark\:hover\:bg-b-3:hover:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-3);--tw-bg-base: var(--scalar-background-3)}.loader-wrapper[data-v-65265e87]{position:relative;display:flex;align-items:center;justify-content:center;--loader-size: 50%}.svg-loader[data-v-65265e87]{width:var(--loader-size);height:var(--loader-size);top:1rem;right:.9rem;overflow:visible;fill:none;background-color:transparent;stroke:currentColor}.svg-path[data-v-65265e87]{stroke-width:14;fill:none;transition:.3s}.svg-x-mark[data-v-65265e87]{stroke-dasharray:57;stroke-dashoffset:57;transition-delay:0s}.svg-check-mark[data-v-65265e87]{stroke-dasharray:149;stroke-dashoffset:149;transition-delay:0s}.icon-is-invalid .svg-x-mark[data-v-65265e87],.icon-is-valid .svg-check-mark[data-v-65265e87]{stroke-dashoffset:0;transition-delay:.3s}.circular-loader[data-v-65265e87]{animation:rotate-65265e87 .7s linear infinite,fade-in-65265e87 .4s;transform-origin:center center;transform:scale(5);background:transparent}.loader-path[data-v-65265e87]{stroke-dasharray:50,200;stroke-dashoffset:-100;stroke-linecap:round}.loader-path-off[data-v-65265e87]{stroke-dasharray:50,200;stroke-dashoffset:-100;transition:opacity .3s;opacity:0}@keyframes fade-in-65265e87{0%{opacity:0}70%{opacity:0}to{opacity:1}}@keyframes rotate-65265e87{0%{transform:scale(5) rotate(0)}to{transform:scale(5) rotate(360deg)}}.scalar-icon[data-v-f4f53d32],.scalar-icon[data-v-f4f53d32] *{stroke-width:var(--ce5c5c8a)}.scalar-code-block{background:inherit;position:relative;overflow:auto;padding:.75rem}.scalar-code-block:hover .copy-button,.copy-button:focus-visible{opacity:100}.scalar-codeblock-pre{all:unset;margin:0;background:transparent;text-wrap:nowrap;white-space-collapse:preserve;border-radius:0;width:-moz-fit-content;width:fit-content}.scalar-code-copy{display:flex;align-items:flex-start;justify-content:flex-end;position:sticky;top:0;right:0;bottom:0;left:0}.copy-button{position:relative;top:0;right:0;display:flex;align-items:center;background-color:var(--scalar-background-2);border:1px solid var(--scalar-border-color);border-radius:3px;color:var(--scalar-color-3);cursor:pointer;height:30px;margin-bottom:-30px;opacity:0;padding:6px;transition:opacity .15s ease-in-out,color .15s ease-in-out}.scalar-code-copy,.copy-button{background:inherit}.copy-button:hover{color:var(--scalar-color-1)}.copy-button svg{stroke-width:1.5}.toggle-icon-ellipse[data-v-8bd002b1]{width:7px;height:7px;border-radius:50%;box-shadow:inset 0 0 0 1px currentColor;overflow:hidden;position:relative;background:var(--scalar-background-1);display:inline-block;transition:width .3s ease-in-out,height .3s ease-in-out}.toggle-icon-moon-mask[data-v-8bd002b1]{width:100%;height:100%;border:1px solid currentColor;display:block;left:2.5px;bottom:2.5px;position:absolute;border-radius:50%;background:var(--scalar-background-1);transition:transform .3s ease-in-out;transform:translate3d(4px,-4px,0)}.toggle-icon-sun-ray[data-v-8bd002b1]{width:12px;height:1px;border-radius:8px;background:currentColor;position:absolute;transition:transform .3s ease-in-out}.toggle-icon-sun-ray[data-v-8bd002b1]:nth-of-type(2){transform:rotate(90deg)}.toggle-icon-sun-ray[data-v-8bd002b1]:nth-of-type(3){transform:rotate(45deg)}.toggle-icon-sun-ray[data-v-8bd002b1]:nth-of-type(4){transform:rotate(-45deg)}.toggle-icon-dark .toggle-icon-ellipse[data-v-8bd002b1]{width:10px;height:10px;-webkit-mask-image:radial-gradient(circle at bottom left,pink 10px,transparent 12px);mask-image:radial-gradient(circle at bottom left,pink 10px,transparent 12px)}.toggle-icon-dark .toggle-icon-sun-ray[data-v-8bd002b1]{transform:scale(0)}.toggle-icon-dark .toggle-icon-moon-mask[data-v-8bd002b1]{transform:translateZ(0)}.scalar-app code.hljs *{font-size:var(--scalar-small);font-family:var(--scalar-font-code);text-align:left;white-space:pre;word-spacing:normal;word-break:normal;word-wrap:normal;line-height:1.4;-moz-tab-size:4;-o-tab-size:4;tab-size:4}.scalar-app code.hljs{all:unset;font-size:var(--scalar-small);color:var(--scalar-color-2);font-family:var(--scalar-font-code);display:inline-block;counter-reset:linenumber}.scalar-app .hljs{background:transparent;color:var(--scalar-color-2)}.scalar-app .hljs .line:before{color:var(--scalar-color-3);display:inline-block;counter-increment:linenumber;content:counter(linenumber);margin-right:.7em;min-width:calc(var(--line-digits) * 1ch);text-align:right}.scalar-app .hljs-comment,.scalar-app .hljs-quote{color:var(--scalar-color-3);font-style:italic}.scalar-app .hljs-number{color:var(--scalar-color-orange)}.scalar-app .hljs-regexp,.scalar-app .hljs-string,.scalar-app .hljs-built_in{color:var(--scalar-color-blue)}.scalar-app .hljs-title.class_{color:var(--scalar-color-1)}.scalar-app .hljs-keyword{color:var(--scalar-color-purple)}.scalar-app .hljs-title.function_{color:var(--scalar-color-orange)}.scalar-app .hljs-subst,.scalar-app .hljs-name{color:var(--scalar-color-blue)}.scalar-app .hljs-attr,.scalar-app .hljs-attribute{color:var(--scalar-color-1)}.scalar-app .hljs-addition,.scalar-app .hljs-literal,.scalar-app .hljs-selector-tag,.scalar-app .hljs-type{color:var(--scalar-color-green)}.scalar-app .hljs-selector-attr,.scalar-app .hljs-selector-pseudo{color:var(--scalar-color-orange)}.scalar-app .hljs-doctag,.scalar-app .hljs-section,.scalar-app .hljs-title{color:var(--scalar-color-blue)}.scalar-app .hljs-selector-id,.scalar-app .hljs-template-variable,.scalar-app .hljs-variable{color:var(--scalar-color-1)}.scalar-app .hljs-name,.scalar-app .hljs-section,.scalar-app .hljs-strong{font-weight:var(--scalar-semibold)}.scalar-app .hljs-bullet,.scalar-app .hljs-link,.scalar-app .hljs-meta,.scalar-app .hljs-symbol{color:var(--scalar-color-blue)}.scalar-app .hljs-deletion{color:var(--scalar-color-red)}.scalar-app .hljs-formula{background:var(--scalar-color-1)}.scalar-app .hljs-emphasis{font-style:italic}.scalar-app .credentials{font-size:0!important;color:transparent}.scalar-app .credentials:after{content:"·····";font-size:var(--scalar-small);color:var(--scalar-color-3);-webkit-user-select:none;-moz-user-select:none;user-select:none}.hljs.language-html{color:var(--scalar-color-1)}.hljs.language-html .hljs-attr{color:var(--scalar-color-2)}.hljs.language-curl .hljs-keyword{color:var(--scalar-color-orange)}.hljs.language-curl .hljs-string{color:var(--scalar-color-1)}.hljs.language-curl .hljs-literal,.hljs.language-php .hljs-variable{color:var(--scalar-color-blue)}.hljs.language-objectivec .hljs-meta{color:var(--scalar-color-1)}.hljs.language-objectivec .hljs-built_in,.hljs-built_in{color:var(--scalar-color-orange)}.scalar-app .markdown{font-family:var(--scalar-font);color:var(--scalar-color-1);word-break:break-word}.scalar-app .markdown h1,.scalar-app .markdown h2,.scalar-app .markdown h3,.scalar-app .markdown h4,.scalar-app .markdown h5,.scalar-app .markdown h6,.scalar-app .markdown p,.scalar-app .markdown div,.scalar-app .markdown img,.scalar-app .markdown details,.scalar-app .markdown summary,.scalar-app .markdown ul,.scalar-app .markdown ol,.scalar-app .markdown table,.scalar-app .markdown blockquote,.scalar-app .markdown code{margin:12px 0}.scalar-app .markdown details{margin:12px 0;color:var(--scalar-color-1)}.scalar-app .markdown summary{display:block;margin:1em 0;padding-left:20px;position:relative;font-weight:var(--scalar-semibold);cursor:pointer;-webkit-user-select:none;-moz-user-select:none;user-select:none}.scalar-app .markdown summary:after{display:block;content:"";position:absolute;top:-1px;left:-1px;width:20px;height:20px;background-color:var(--scalar-color-3);-webkit-mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>");mask-image:url("data:image/svg+xml,<svg viewBox=\"0 0 24 24\" fill=\"none\" xmlns=\"http://www.w3.org/2000/svg\"><path d=\"m9 18 6-6-6-6\" stroke=\"currentColor\" stroke-width=\"1.5\" stroke-linecap=\"round\" stroke-linejoin=\"round\"/></svg>")}.scalar-app .markdown summary:hover:after{background-color:var(--scalar-color-1)}.scalar-app .markdown details[open] summary:after{transform:rotate(90deg)}.scalar-app .markdown summary::-webkit-details-marker{display:none}.scalar-app .markdown img{overflow:hidden;border-radius:var(--scalar-radius);max-width:100%}.scalar-app .markdown>:first-child{margin-top:0;margin-bottom:0}.scalar-app .markdown h1{--font-size: 1.4em}.scalar-app .markdown h2{--font-size: 1.25em}.scalar-app .markdown h3{--font-size: 1.1em}.scalar-app .markdown h4,.scalar-app .markdown h6{--font-size: 1em}.scalar-app .markdown h1,.scalar-app .markdown h2,.scalar-app .markdown h3,.scalar-app .markdown h4,.scalar-app .markdown h5,.scalar-app .markdown h6{font-size:var(--font-size);margin:18px 0 6px;font-weight:var(--scalar-bold);display:block;line-height:1.45}.scalar-app .markdown b,.scalar-app .markdown strong{font-weight:var(--scalar-bold)}.scalar-app .markdown p{color:inherit;font-weight:var(--font-weight, var(--scalar-regular));line-height:1.5;margin-bottom:0;display:block}.scalar-app .markdown ul,.scalar-app .markdown ol{padding-left:24px;line-height:1.5;margin:12px 0;display:block}.scalar-app .markdown ul{list-style:disc}.scalar-app .markdown ol{list-style:decimal}.scalar-app .markdown ul.contains-task-list{list-style:none;padding-left:0}.scalar-app .markdown li{margin:6px 0;display:list-item}.scalar-app .markdown ul:first-of-type li:first-of-type{margin-top:0}.scalar-app .markdown a{color:var(--scalar-link-color, var(--scalar-color-accent));font-weight:var(--scalar-link-font-weight, inherit);-webkit-text-decoration:var(--scalar-text-decoration);text-decoration:var(--scalar-text-decoration);cursor:pointer}.scalar-app .markdown a:hover{color:var(--scalar-link-color-hover, var(--scalar-color-accent));-webkit-text-decoration:var(--scalar-text-decoration-hover);text-decoration:var(--scalar-text-decoration-hover)}.scalar-app .markdown em{font-style:italic}.scalar-app .markdown sup{font-size:var(--scalar-micro);vertical-align:super;font-weight:450}.scalar-app .markdown sub{font-size:var(--scalar-micro);vertical-align:sub;font-weight:450}.scalar-app .markdown del{text-decoration:line-through}.scalar-app .markdown code{font-family:var(--scalar-font-code);background-color:var(--scalar-background-2);box-shadow:0 0 0 var(--scalar-border-width) var(--scalar-border-color);font-size:var(--scalar-micro);border-radius:2px;padding:0 3px}.scalar-app .markdown .hljs{font-size:var(--scalar-small)}.scalar-app .markdown pre code{display:block;white-space:pre;padding:12px;line-height:1.5;margin:12px 0;-webkit-overflow-scrolling:touch;overflow-x:auto;max-width:100%;min-width:100px}.scalar-app .markdown hr{border:none;border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.scalar-app .markdown blockquote{border-left:3px solid var(--scalar-border-color);padding-left:12px;margin:0;display:block}.scalar-app .markdown table{display:block;overflow-x:auto;position:relative;border-collapse:collapse;width:-moz-max-content;width:max-content;max-width:100%;margin:1em 0;box-shadow:0 0 0 var(--scalar-border-width) var(--scalar-border-color);border-radius:var(--scalar-radius-lg)}.scalar-app .markdown tbody{display:table-row-group;vertical-align:middle}.scalar-app .markdown thead{display:table-header-group;vertical-align:middle}.scalar-app .markdown tr{display:table-row;border-color:inherit;vertical-align:inherit}.scalar-app .markdown td,.scalar-app .markdown th{display:table-cell;vertical-align:inherit;min-width:1em;padding:6px 9px;vertical-align:top;line-height:1.5;position:relative;word-break:initial;font-size:var(--scalar-small);color:var(--scalar-color-1);font-weight:var(--font-weight, var(--scalar-regular));border-right:var(--scalar-border-width) solid var(--scalar-border-color);border-bottom:var(--scalar-border-width) solid var(--scalar-border-color)}.scalar-app .markdown td>*,.scalar-app .markdown th>*{margin-bottom:0}.scalar-app .markdown th:empty{display:none}.scalar-app .markdown td:first-of-type,.scalar-app .markdown th:first-of-type{border-left:none}.scalar-app .markdown td:last-of-type,.scalar-app .markdown th:last-of-type{border-right:none}.scalar-app .markdown tr:last-of-type td{border-bottom:none}.scalar-app .markdown th{font-weight:var(--scalar-semibold)!important;text-align:left;border-left-color:transparent;background:var(--scalar-background-2)}.scalar-app .markdown tr>[align=left]{text-align:left}.scalar-app .markdown tr>[align=right]{text-align:right}.scalar-app .markdown tr>[align=center]{text-align:center}.dark-mode .scalar-dropdown-item[data-v-942bce7e]:hover,.dark-mode .scalar-dropdown-item[data-v-457c1cd3]:hover{filter:brightness(1.1)}.scalar-modal-layout[data-v-0f8e0211]{animation:fadein-layout-0f8e0211 ease-in-out .3s forwards}.scalar-modal[data-v-0f8e0211]{animation:fadein-modal-0f8e0211 ease-in-out .3s forwards;animation-delay:.1s;position:fixed;left:0;top:0;right:0;box-shadow:var(--scalar-shadow-2);transform:translate3d(0,10px,0)}.scalar-modal-layout-full[data-v-0f8e0211]{opacity:1!important;background:transparent!important}.dark-mode .scalar-modal[data-v-0f8e0211]{background-color:color-mix(in srgb,var(--scalar-background-1),black)}.scalar-modal.scalar-modal-search[data-v-0f8e0211]{max-width:540px;max-height:440px;background-color:transparent}.modal-content-search .modal-body[data-v-0f8e0211]{padding:0;overflow:hidden;display:flex;flex-direction:column;max-height:440px}@media (max-width: 1280px){.scalar-modal[data-v-0f8e0211]{max-height:calc(100% - 56px);top:28px}}@keyframes fadein-layout-0f8e0211{0%{opacity:0}to{opacity:1}}@keyframes fadein-modal-0f8e0211{0%{opacity:0;transform:translate3d(0,10px,0)}to{opacity:1;transform:translateZ(0)}}.full-size-styles[data-v-0f8e0211]{transform:translateZ(0);animation:fadein-layout-0f8e0211 ease-in-out .3s forwards;max-height:100%!important;top:0!important;left:0;position:absolute!important;margin:initial;border-radius:0!important;background-color:var(--scalar-background-1)!important;box-shadow:none!important;border-right:var(--scalar-border-width) solid var(--scalar-border-color)}@media (min-width: 800px){.full-size-styles[data-v-0f8e0211]{width:50dvw!important}}.full-size-styles[data-v-0f8e0211]:after{content:"";width:50dvw;height:100dvh;position:absolute;right:-50dvw;top:0}.search-background[data-v-2f2316c6]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2))}.search-background[data-v-2f2316c6]:focus-within{background:transparent}.scalar-input-wrapper-focus .scalar-input-label[data-v-4de7a24e]{color:var(--scalar-color-1)}.scalar-input-wrapper-error .scalar-input-label[data-v-4de7a24e]{color:var(--scalar-color-red)}.scalar-input[data-v-4de7a24e]::-moz-selection{color:var(--scalar-color-1);background:#ffa55859}.scalar-input[data-v-4de7a24e]::selection{color:var(--scalar-color-1);background:#ffa55859}.scalar-input[data-v-4de7a24e]:has(+.scalar-input-label){opacity:0}.scalar-input[data-v-4de7a24e]:not(:-moz-placeholder-shown){opacity:1;-moz-transition:opacity .2s ease-in-out .15s;transition:opacity .2s ease-in-out .15s}.scalar-input[data-v-4de7a24e]:not(:placeholder-shown),.scalar-input-wrapper-focus .scalar-input[data-v-4de7a24e]{opacity:1;transition:opacity .2s ease-in-out .15s}.scalar-input[data-v-4de7a24e]:-webkit-autofill,.scalar-input[data-v-4de7a24e]:-webkit-autofill:hover,.scalar-input[data-v-4de7a24e]:-webkit-autofill:focus,.scalar-input[data-v-4de7a24e]:-webkit-autofill:active,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill:hover,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill:focus,.scalar-input[data-v-4de7a24e]:focus-within:-webkit-autofill:active{-webkit-box-shadow:0 0 0px 1000px var(--scalar-background-1) inset!important;-webkit-text-fill-color:var(--scalar-color-1);color:var(--scalar-color-1);border-radius:var(--scalar-radius)}.scalar-input:not(:-moz-placeholder-shown)+.scalar-input-label[data-v-4de7a24e]{transform:translate3d(0,-20px,0) scale(.8);transform-origin:top left}.scalar-input-wrapper-focus .scalar-input-label[data-v-4de7a24e],.scalar-input:not(:placeholder-shown)+.scalar-input-label[data-v-4de7a24e]{transform:translate3d(0,-20px,0) scale(.8);transform-origin:top left}.scalar-input-wrapper-focus:has(button:active) .scalar-input-label[data-v-4de7a24e]{color:var(--scalar-color-3)!important}@layer scalar-base,scalar-theme;@layer scalar-base{body{margin:0;line-height:inherit}}:where(.scalar-app){font-family:var(--scalar-font);line-height:1.15;-webkit-text-size-adjust:100%;-moz-tab-size:4;-o-tab-size:4;tab-size:4}:where(.scalar-app) *,:where(.scalar-app) *:before,:where(.scalar-app) *:after{box-sizing:border-box;border-width:0;border-style:solid;border-color:var(--scalar-border-color);outline-width:1px;outline-style:none;outline-color:var(--scalar-color-accent);font-family:inherit;font-feature-settings:inherit;font-variation-settings:inherit;font-size:inherit;font-weight:inherit;line-height:inherit;color:inherit;margin:unset;padding:unset;text-rendering:optimizeLegibility;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale}:where(.scalar-app) :before,:where(.scalar-app) :after{--tw-content: ""}:where(.scalar-app) button,:where(.scalar-app) input,:where(.scalar-app) optgroup,:where(.scalar-app) select,:where(.scalar-app) textarea,:where(.scalar-app) ::file-selector-button{background:transparent}:where(.scalar-app) ol,:where(.scalar-app) ul,:where(.scalar-app) menu{list-style:none}:where(.scalar-app) input:where(:not([type=button],[type=reset],[type=submit])),:where(.scalar-app) select,:where(.scalar-app) textarea{border-radius:var(--scalar-radius);border-width:1px}:where(.scalar-app) input::-moz-placeholder{color:var(--scalar-color-3);font-family:var(--scalar-font)}:where(.scalar-app) input::placeholder{color:var(--scalar-color-3);font-family:var(--scalar-font)}:where(.scalar-app) input[type=search]::-webkit-search-cancel-button,:where(.scalar-app) input[type=search]::-webkit-search-decoration{-webkit-appearance:none;-moz-appearance:none;appearance:none}:where(.scalar-app) input:-webkit-autofill{-webkit-background-clip:text!important;background-clip:text!important}:where(.scalar-app) :focus-visible{outline-style:solid;border-radius:var(--scalar-radius)}:where(.scalar-app) button:focus-visible,:where(.scalar-app) [role=button]:focus-visible{outline-offset:-1px}:where(.scalar-app) button,:where(.scalar-app) [role=button]{cursor:pointer}:where(.scalar-app) :disabled{cursor:default}:where(.scalar-app) img,:where(.scalar-app) svg,:where(.scalar-app) video,:where(.scalar-app) canvas,:where(.scalar-app) audio,:where(.scalar-app) iframe,:where(.scalar-app) embed,:where(.scalar-app) object{display:block;vertical-align:middle}:where(.scalar-app) [hidden]{display:none}.scalar-app .cm-scroller,.scalar-app .custom-scroll{overflow-y:auto;scrollbar-color:transparent transparent;scrollbar-width:thin;-webkit-overflow-scrolling:touch}.scalar-app .custom-scroll-self-contain-overflow{overscroll-behavior:contain}.scalar-app .cm-scroller:hover,.scalar-app .custom-scroll:hover,.scalar-app.scalar-scrollbars-obtrusive .cm-scroller,.scalar-app.scalar-scrollbars-obtrusive .custom-scroll{scrollbar-color:var(--scalar-scrollbar-color, transparent) transparent}.scalar-app .cm-scroller:hover::-webkit-scrollbar-thumb,.scalar-app .custom-scroll:hover::-webkit-scrollbar-thumb{background:var(--scalar-scrollbar-color);background-clip:content-box;border:3px solid transparent}.scalar-app .cm-scroller::-webkit-scrollbar-thumb:active,.scalar-app .custom-scroll::-webkit-scrollbar-thumb:active{background:var(--scalar-scrollbar-color-active);background-clip:content-box;border:3px solid transparent}.scalar-app .cm-scroller::-webkit-scrollbar-corner,.scalar-app .custom-scroll::-webkit-scrollbar-corner{background:transparent}.scalar-app .cm-scroller::-webkit-scrollbar,.scalar-app .custom-scroll::-webkit-scrollbar{height:12px;width:12px}.scalar-app .cm-scroller::-webkit-scrollbar-track,.scalar-app .custom-scroll::-webkit-scrollbar-track{background:transparent}.scalar-app .cm-scroller::-webkit-scrollbar-thumb,.scalar-app .custom-scroll::-webkit-scrollbar-thumb{border-radius:20px;background:transparent;background-clip:content-box;border:3px solid transparent}@media (pointer: coarse){.scalar-app .cm-scroller,.scalar-app .custom-scroll{padding-right:12px}}@layer scalar-base{:root{--scalar-border-width: .5px;--scalar-radius: 3px;--scalar-radius-lg: 6px;--scalar-radius-xl: 8px;--scalar-font: "Inter", -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;--scalar-font-code: "JetBrains Mono", ui-monospace, Menlo, Monaco, "Cascadia Mono", "Segoe UI Mono", "Roboto Mono", "Oxygen Mono", "Ubuntu Monospace", "Source Code Pro", "Fira Mono", "Droid Sans Mono", "Courier New", monospace;--scalar-heading-1: 40px;--scalar-page-description: 24px;--scalar-heading-2: 24px;--scalar-heading-3: 20px;--scalar-heading-4: 16px;--scalar-heading-5: 16px;--scalar-heading-6: 16px;--scalar-paragraph: 16px;--scalar-small: 14px;--scalar-mini: 13px;--scalar-micro: 12px;--scalar-bold: 600;--scalar-semibold: 500;--scalar-regular: 400;--scalar-font-size-1: 24px;--scalar-font-size-2: 16px;--scalar-font-size-3: 14px;--scalar-font-size-4: 13px;--scalar-font-size-5: 12px;--scalar-font-size-6: 12px;--scalar-font-size-7: 10px;--scalar-line-height-1: 32px;--scalar-line-height-2: 24px;--scalar-line-height-3: 20px;--scalar-line-height-4: 18px;--scalar-line-height-5: 16px;--scalar-font-medium: 500;--scalar-font-bold: 700;--scalar-text-decoration: none;--scalar-text-decoration-hover: underline;--scalar-link-font-weight: inherit}.dark-mode{color-scheme:dark;--scalar-scrollbar-color: rgba(255, 255, 255, .18);--scalar-scrollbar-color-active: rgba(255, 255, 255, .36);--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-shadow-1: 0 1px 3px 0 rgb(0, 0, 0, .1);--scalar-shadow-2: 0 0 0 .5px var(--scalar-border-color), rgba(15, 15, 15, .2) 0px 3px 6px, rgba(15, 15, 15, .4) 0px 9px 24px;--scalar-lifted-brightness: 1.45;--scalar-backdrop-brightness: .5;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode{color-scheme:light;--scalar-scrollbar-color-active: rgba(0, 0, 0, .36);--scalar-scrollbar-color: rgba(0, 0, 0, .18);--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-shadow-1: 0 1px 3px 0 rgba(0, 0, 0, .11);--scalar-shadow-2: rgba(0, 0, 0, .08) 0px 13px 20px 0px, rgba(0, 0, 0, .08) 0px 3px 8px 0px, #eeeeed 0px 0 0 .5px;--scalar-lifted-brightness: 1;--scalar-backdrop-brightness: 1;--scalar-sidebar-indent-border: transparent;--scalar-sidebar-indent-border-hover: transparent;--scalar-sidebar-indent-border-active: transparent;--scalar-link-color: var(--scalar-color-accent);--scalar-link-color-hover: var(--scalar-color-accent);--scalar-text-decoration-color: currentColor;--scalar-text-decoration-color-hover: currentColor}.light-mode .dark-mode{color-scheme:dark!important}@media (max-width: 460px){:root{--scalar-font-size-1: 22px;--scalar-font-size-2: 14px;--scalar-font-size-3: 12px}}@media (max-width: 720px){:root{--scalar-heading-1: 24px;--scalar-page-description: 20px}}.light-mode{--scalar-background-1: #fff;--scalar-background-2: #f6f6f6;--scalar-background-3: #e7e7e7;--scalar-background-accent: #8ab4f81f;--scalar-color-1: #2a2f45;--scalar-color-2: #757575;--scalar-color-3: #8e8e8e;--scalar-color-accent: #0099ff;--scalar-border-color: #dfdfdf}.dark-mode{--scalar-background-1: #0f0f0f;--scalar-background-2: #1a1a1a;--scalar-background-3: #272727;--scalar-color-1: #e7e7e7;--scalar-color-2: #a4a4a4;--scalar-color-3: #797979;--scalar-color-accent: #3ea6ff;--scalar-background-accent: #3ea6ff1f;--scalar-border-color: #2d2d2d}.light-mode .t-doc__sidebar,.dark-mode .t-doc__sidebar{--scalar-sidebar-background-1: var(--scalar-background-1);--scalar-sidebar-color-1: var(--scalar-color-1);--scalar-sidebar-color-2: var(--scalar-color-2);--scalar-sidebar-border-color: var(--scalar-border-color);--scalar-sidebar-item-hover-background: var(--scalar-background-2);--scalar-sidebar-item-hover-color: currentColor;--scalar-sidebar-item-active-background: var(--scalar-background-2);--scalar-sidebar-color-active: var(--scalar-color-1);--scalar-sidebar-search-background: transparent;--scalar-sidebar-search-color: var(--scalar-color-3);--scalar-sidebar-search-border-color: var(--scalar-border-color)}.light-mode{--scalar-color-green: #069061;--scalar-color-red: #ef0006;--scalar-color-yellow: #edbe20;--scalar-color-blue: #0082d0;--scalar-color-orange: #fb892c;--scalar-color-purple: #5203d1;--scalar-button-1: rgba(0, 0, 0, 1);--scalar-button-1-hover: rgba(0, 0, 0, .8);--scalar-button-1-color: rgba(255, 255, 255, .9);--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-color-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}.dark-mode{--scalar-color-green: #00b648;--scalar-color-red: #dc1b19;--scalar-color-yellow: #ffc90d;--scalar-color-blue: #4eb3ec;--scalar-color-orange: #ff8d4d;--scalar-color-purple: #b191f9;--scalar-button-1: rgba(255, 255, 255, 1);--scalar-button-1-hover: rgba(255, 255, 255, .9);--scalar-button-1-color: black;--scalar-color-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 20%);--scalar-background-alert: color-mix(in srgb, var(--scalar-color-orange), var(--scalar-background-1) 95%);--scalar-background-danger: color-mix(in srgb, var(--scalar-color-red), var(--scalar-background-1) 95%)}*,:before,:after{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }::backdrop{--tw-border-spacing-x:0;--tw-border-spacing-y:0;--tw-translate-x:0;--tw-translate-y:0;--tw-rotate:0;--tw-skew-x:0;--tw-skew-y:0;--tw-scale-x:1;--tw-scale-y:1;--tw-pan-x: ;--tw-pan-y: ;--tw-pinch-zoom: ;--tw-scroll-snap-strictness:proximity;--tw-gradient-from-position: ;--tw-gradient-via-position: ;--tw-gradient-to-position: ;--tw-ordinal: ;--tw-slashed-zero: ;--tw-numeric-figure: ;--tw-numeric-spacing: ;--tw-numeric-fraction: ;--tw-ring-inset: ;--tw-ring-offset-width:0px;--tw-ring-offset-color:#fff;--tw-ring-color:rgb(59 130 246 / .5);--tw-ring-offset-shadow:0 0 #0000;--tw-ring-shadow:0 0 #0000;--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 #0000;--tw-blur: ;--tw-brightness: ;--tw-contrast: ;--tw-grayscale: ;--tw-hue-rotate: ;--tw-invert: ;--tw-saturate: ;--tw-sepia: ;--tw-drop-shadow: ;--tw-backdrop-blur: ;--tw-backdrop-brightness: ;--tw-backdrop-contrast: ;--tw-backdrop-grayscale: ;--tw-backdrop-hue-rotate: ;--tw-backdrop-invert: ;--tw-backdrop-opacity: ;--tw-backdrop-saturate: ;--tw-backdrop-sepia: ;--tw-contain-size: ;--tw-contain-layout: ;--tw-contain-paint: ;--tw-contain-style: }}.scalar-client{line-height:1.5;color:var(--scalar-color-1)}.scalar-client .container{width:100%}@media (min-width: 400px){.scalar-client .container{max-width:400px}}@media (min-width: 600px){.scalar-client .container{max-width:600px}}@media (min-width: 800px){.scalar-client .container{max-width:800px}}@media (min-width: 1000px){.scalar-client .container{max-width:1000px}}@media (min-width: 1200px){.scalar-client .container{max-width:1200px}}.scalar-client .row{display:flex;flex-direction:row}.scalar-client .row-1{display:flex;flex-direction:row;flex:1 1 0%}.scalar-client .col{display:flex;flex-direction:column}.scalar-client .col-1{display:flex;flex-direction:column;flex:1 1 0%}.scalar-client .flex-center{display:flex;justify-content:center;align-items:center}.scalar-client .centered-x{left:50%;--tw-translate-x: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .centered-y{top:50%;--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .centered{left:50%;--tw-translate-x: -50%;top:50%;--tw-translate-y: -50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .floating-bg:before{background-color:var(--scalar-background-2);border-radius:var(--scalar-radius);content:"";height:calc(100% - 4px);left:-4px;opacity:0;position:absolute;top:2.5px;transition:opacity .2s ease-in-out;width:calc(100% + 8px);z-index:1}.scalar-client .floating-bg:hover:before{opacity:1}.scalar-client .sr-only{position:absolute;width:1px;height:1px;padding:0;margin:-1px;overflow:hidden;clip:rect(0,0,0,0);white-space:nowrap;border-width:0}.scalar-client .not-sr-only{position:static;width:auto;height:auto;padding:0;margin:0;overflow:visible;clip:auto;white-space:normal}.scalar-client .pointer-events-none{pointer-events:none}.scalar-client .pointer-events-auto{pointer-events:auto}.scalar-client .visible{visibility:visible}.scalar-client .static{position:static}.scalar-client .fixed{position:fixed}.scalar-client .absolute{position:absolute}.scalar-client .relative{position:relative}.scalar-client .sticky{position:sticky}.scalar-client .inset-0{top:0;right:0;bottom:0;left:0}.scalar-client .-inset-x-96{left:-384px;right:-384px}.scalar-client .inset-x-1{left:4px;right:4px}.scalar-client .-right-\[30px\]{right:-30px}.scalar-client .-top-\[--scalar-address-bar-height\]{top:calc(var(--scalar-address-bar-height) * -1)}.scalar-client .bottom-0{bottom:0}.scalar-client .bottom-1\/2{bottom:50%}.scalar-client .bottom-\[var\(--scalar-border-width\)\]{bottom:var(--scalar-border-width)}.scalar-client .left-0{left:0}.scalar-client .left-1\/2{left:50%}.scalar-client .right-0{right:0}.scalar-client .right-1{right:4px}.scalar-client .right-1\/2{right:50%}.scalar-client .right-14{right:56px}.scalar-client .right-16{right:64px}.scalar-client .right-2{right:8px}.scalar-client .right-4{right:16px}.scalar-client .right-7{right:28px}.scalar-client .top-0{top:0}.scalar-client .top-1\/2{top:50%}.scalar-client .top-12{top:48px}.scalar-client .top-2{top:8px}.scalar-client .-z-1{z-index:-1}.scalar-client .z-0{z-index:0}.scalar-client .z-1{z-index:1}.scalar-client .z-10{z-index:10}.scalar-client .z-20{z-index:20}.scalar-client .z-50{z-index:50}.scalar-client .z-\[10001\]{z-index:10001}.scalar-client .z-\[1\]{z-index:1}.scalar-client .z-context{z-index:1000}.scalar-client .z-context-plus{z-index:1001}.scalar-client .z-overlay{z-index:10000}.scalar-client .order-last{order:9999}.scalar-client .col-span-full{grid-column:1 / -1}.scalar-client .\!m-0{margin:0!important}.scalar-client .m-0{margin:0}.scalar-client .m-4{margin:16px}.scalar-client .m-auto{margin:auto}.scalar-client .\!mx-0{margin-left:0!important;margin-right:0!important}.scalar-client .-mx-1{margin-left:-4px;margin-right:-4px}.scalar-client .-my-1{margin-top:-4px;margin-bottom:-4px}.scalar-client .my-1{margin-top:4px;margin-bottom:4px}.scalar-client .my-1\.25{margin-top:5px;margin-bottom:5px}.scalar-client .my-12{margin-top:48px;margin-bottom:48px}.scalar-client .\!mb-0{margin-bottom:0!important}.scalar-client .\!mr-0{margin-right:0!important}.scalar-client .-mb-0{margin-bottom:-0px}.scalar-client .-mb-0\.25{margin-bottom:-1px}.scalar-client .-mb-\[var\(--scalar-border-width\)\]{margin-bottom:calc(var(--scalar-border-width) * -1)}.scalar-client .-ml-0{margin-left:-0px}.scalar-client .-ml-0\.5{margin-left:-2px}.scalar-client .-ml-12{margin-left:-48px}.scalar-client .-mr-1{margin-right:-4px}.scalar-client .-mr-1\.5{margin-right:-6px}.scalar-client .-mr-3{margin-right:-12px}.scalar-client .-mt-0{margin-top:-0px}.scalar-client .-mt-0\.25{margin-top:-1px}.scalar-client .-mt-\[\.5px\]{margin-top:-.5px}.scalar-client .mb-0{margin-bottom:0}.scalar-client .mb-1{margin-bottom:4px}.scalar-client .mb-1\.5{margin-bottom:6px}.scalar-client .mb-2{margin-bottom:8px}.scalar-client .mb-4{margin-bottom:16px}.scalar-client .mb-\[\.5px\]{margin-bottom:.5px}.scalar-client .ml-0{margin-left:0}.scalar-client .ml-0\.5{margin-left:2px}.scalar-client .ml-0\.75{margin-left:3px}.scalar-client .ml-1{margin-left:4px}.scalar-client .ml-2{margin-left:8px}.scalar-client .ml-auto{margin-left:auto}.scalar-client .mr-0{margin-right:0}.scalar-client .mr-0\.25{margin-right:1px}.scalar-client .mr-0\.75{margin-right:3px}.scalar-client .mr-1{margin-right:4px}.scalar-client .mr-1\.5{margin-right:6px}.scalar-client .mr-2{margin-right:8px}.scalar-client .mr-2\.5{margin-right:10px}.scalar-client .mr-3{margin-right:12px}.scalar-client .mr-auto{margin-right:auto}.scalar-client .mt-1{margin-top:4px}.scalar-client .mt-10{margin-top:40px}.scalar-client .mt-2{margin-top:8px}.scalar-client .mt-3{margin-top:12px}.scalar-client .mt-5{margin-top:20px}.scalar-client .mt-auto{margin-top:auto}.scalar-client .line-clamp-1{overflow:hidden;display:-webkit-box;-webkit-box-orient:vertical;-webkit-line-clamp:1}.scalar-client .block{display:block}.scalar-client .inline-block{display:inline-block}.scalar-client .\!flex{display:flex!important}.scalar-client .flex{display:flex}.scalar-client .inline-flex{display:inline-flex}.scalar-client .table{display:table}.scalar-client .grid{display:grid}.scalar-client .contents{display:contents}.scalar-client .\!hidden{display:none!important}.scalar-client .hidden{display:none}.scalar-client .aspect-\[4\/3\]{aspect-ratio:4/3}.scalar-client .aspect-square{aspect-ratio:1 / 1}.scalar-client .size-10{width:40px;height:40px}.scalar-client .size-2{width:8px;height:8px}.scalar-client .size-2\.5{width:10px;height:10px}.scalar-client .size-3{width:12px;height:12px}.scalar-client .size-3\.5{width:14px;height:14px}.scalar-client .size-3\/4{width:75%;height:75%}.scalar-client .size-4{width:16px;height:16px}.scalar-client .size-7{width:28px;height:28px}.scalar-client .h-1{height:4px}.scalar-client .h-1\.5{height:6px}.scalar-client .h-10{height:40px}.scalar-client .h-12{height:48px}.scalar-client .h-2{height:8px}.scalar-client .h-2\.5{height:10px}.scalar-client .h-24{height:96px}.scalar-client .h-3{height:12px}.scalar-client .h-3\.5{height:14px}.scalar-client .h-4{height:16px}.scalar-client .h-5{height:20px}.scalar-client .h-6{height:24px}.scalar-client .h-64{height:256px}.scalar-client .h-7{height:28px}.scalar-client .h-8{height:32px}.scalar-client .h-9{height:36px}.scalar-client .h-\[--scalar-address-bar-height\]{height:var(--scalar-address-bar-height)}.scalar-client .h-\[68px\]{height:68px}.scalar-client .h-\[calc\(100\%-273\.5px\)\]{height:calc(100% - 273.5px)}.scalar-client .h-\[calc\(100\%_-_50px\)\]{height:calc(100% - 50px)}.scalar-client .h-\[var\(--scalar-border-width\)\]{height:var(--scalar-border-width)}.scalar-client .h-auto{height:auto}.scalar-client .h-fit{height:-moz-fit-content;height:fit-content}.scalar-client .h-full{height:100%}.scalar-client .h-header{height:48px}.scalar-client .h-px{height:1px}.scalar-client .h-screen{height:100vh}.scalar-client .\!max-h-\[initial\]{max-height:initial!important}.scalar-client .max-h-40{max-height:160px}.scalar-client .max-h-8{max-height:32px}.scalar-client .max-h-\[40dvh\]{max-height:40dvh}.scalar-client .max-h-\[50dvh\]{max-height:50dvh}.scalar-client .max-h-\[60svh\]{max-height:60svh}.scalar-client .max-h-\[calc\(100\%-32px\)\]{max-height:calc(100% - 32px)}.scalar-client .max-h-\[inherit\]{max-height:inherit}.scalar-client .max-h-fit{max-height:-moz-fit-content;max-height:fit-content}.scalar-client .max-h-screen{max-height:100vh}.scalar-client .min-h-0{min-height:0px}.scalar-client .min-h-10{min-height:40px}.scalar-client .min-h-11{min-height:44px}.scalar-client .min-h-12{min-height:48px}.scalar-client .min-h-16{min-height:64px}.scalar-client .min-h-20{min-height:80px}.scalar-client .min-h-8{min-height:32px}.scalar-client .min-h-fit{min-height:-moz-fit-content;min-height:fit-content}.scalar-client .\!w-fit{width:-moz-fit-content!important;width:fit-content!important}.scalar-client .w-0{width:0px}.scalar-client .w-0\.5{width:2px}.scalar-client .w-1{width:4px}.scalar-client .w-1\.5{width:6px}.scalar-client .w-1\/2{width:50%}.scalar-client .w-10{width:40px}.scalar-client .w-2{width:8px}.scalar-client .w-2\.5{width:10px}.scalar-client .w-20{width:80px}.scalar-client .w-3{width:12px}.scalar-client .w-3\.5{width:14px}.scalar-client .w-4{width:16px}.scalar-client .w-5{width:20px}.scalar-client .w-56{width:224px}.scalar-client .w-6{width:24px}.scalar-client .w-64{width:256px}.scalar-client .w-72{width:288px}.scalar-client .w-8{width:32px}.scalar-client .w-\[60px\]{width:60px}.scalar-client .w-\[calc\(100\%-10px\)\]{width:calc(100% - 10px)}.scalar-client .w-\[calc\(100\%_-_8px\)\]{width:calc(100% - 8px)}.scalar-client .w-\[inherit\]{width:inherit}.scalar-client .w-auto{width:auto}.scalar-client .w-dvw{width:100dvw}.scalar-client .w-fit{width:-moz-fit-content;width:fit-content}.scalar-client .w-full{width:100%}.scalar-client .w-max{width:-moz-max-content;width:max-content}.scalar-client .min-w-0{min-width:0px}.scalar-client .min-w-2{min-width:8px}.scalar-client .min-w-2\.5{min-width:10px}.scalar-client .min-w-3{min-width:12px}.scalar-client .min-w-3\.5{min-width:14px}.scalar-client .min-w-48{min-width:192px}.scalar-client .min-w-8{min-width:32px}.scalar-client .min-w-\[296px\]{min-width:296px}.scalar-client .min-w-\[37px\]{min-width:37px}.scalar-client .min-w-fit{min-width:-moz-fit-content;min-width:fit-content}.scalar-client .min-w-full{min-width:100%}.scalar-client .max-w-\[100\%\]{max-width:100%}.scalar-client .max-w-\[14px\]{max-width:14px}.scalar-client .max-w-\[150px\]{max-width:150px}.scalar-client .max-w-\[320px\]{max-width:320px}.scalar-client .max-w-\[37px\]{max-width:37px}.scalar-client .max-w-\[380px\]{max-width:380px}.scalar-client .max-w-\[420px\]{max-width:420px}.scalar-client .max-w-\[720px\]{max-width:720px}.scalar-client .max-w-\[calc\(100dvw-24px\)\]{max-width:calc(100dvw - 24px)}.scalar-client .max-w-full{max-width:100%}.scalar-client .flex-1{flex:1 1 0%}.scalar-client .shrink-0{flex-shrink:0}.scalar-client .flex-grow{flex-grow:1}.scalar-client .-translate-x-1\/2{--tw-translate-x:-50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .-translate-y-1\/2{--tw-translate-y:-50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .translate-x-0{--tw-translate-x:0px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .translate-x-1\/2{--tw-translate-x:50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .translate-y-1\/2{--tw-translate-y:50%;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .rotate-90{--tw-rotate:90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .scale-75{--tw-scale-x:.75;--tw-scale-y:.75;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .transform{transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .cursor-auto{cursor:auto}.scalar-client .cursor-default{cursor:default}.scalar-client .cursor-help{cursor:help}.scalar-client .cursor-pointer{cursor:pointer}.scalar-client .cursor-text{cursor:text}.scalar-client .select-none{-webkit-user-select:none;-moz-user-select:none;user-select:none}.scalar-client .resize-none{resize:none}.scalar-client .resize{resize:both}.scalar-client .auto-rows-\[32px\]{grid-auto-rows:32px}.scalar-client .auto-rows-auto{grid-auto-rows:auto}.scalar-client .grid-cols-2{grid-template-columns:repeat(2,minmax(0,1fr))}.scalar-client .grid-cols-\[44px\,1fr\,repeat\(3\,auto\)\]{grid-template-columns:44px 1fr repeat(3,auto)}.scalar-client .grid-cols-\[auto\,1fr\]{grid-template-columns:auto 1fr}.scalar-client .grid-cols-\[repeat\(auto-fill\,minmax\(32px\,1fr\)\)\]{grid-template-columns:repeat(auto-fill,minmax(32px,1fr))}.scalar-client .flex-row{flex-direction:row}.scalar-client .flex-col{flex-direction:column}.scalar-client .flex-wrap{flex-wrap:wrap}.scalar-client .content-start{align-content:flex-start}.scalar-client .content-between{align-content:space-between}.scalar-client .items-start{align-items:flex-start}.scalar-client .items-end{align-items:flex-end}.scalar-client .items-center{align-items:center}.scalar-client .items-stretch{align-items:stretch}.scalar-client .justify-start{justify-content:flex-start}.scalar-client .justify-end{justify-content:flex-end}.scalar-client .justify-center{justify-content:center}.scalar-client .justify-between{justify-content:space-between}.scalar-client .justify-stretch{justify-content:stretch}.scalar-client .\!gap-2{gap:8px!important}.scalar-client .gap-0{gap:0px}.scalar-client .gap-0\.5{gap:2px}.scalar-client .gap-0\.75{gap:3px}.scalar-client .gap-1{gap:4px}.scalar-client .gap-1\.5{gap:6px}.scalar-client .gap-1\/2{gap:.5px}.scalar-client .gap-10{gap:40px}.scalar-client .gap-2{gap:8px}.scalar-client .gap-2\.5{gap:10px}.scalar-client .gap-3{gap:12px}.scalar-client .gap-4{gap:16px}.scalar-client .gap-6{gap:24px}.scalar-client .gap-8{gap:32px}.scalar-client .gap-\[1\.5px\]{gap:1.5px}.scalar-client .gap-px{gap:1px}.scalar-client .gap-x-2{-moz-column-gap:8px;column-gap:8px}.scalar-client .gap-x-2\.5{-moz-column-gap:10px;column-gap:10px}.scalar-client .space-x-1>:not([hidden])~:not([hidden]){--tw-space-x-reverse:0;margin-right:calc(4px * var(--tw-space-x-reverse));margin-left:calc(4px * calc(1 - var(--tw-space-x-reverse)))}.scalar-client .divide-x>:not([hidden])~:not([hidden]){--tw-divide-x-reverse:0;border-right-width:calc(var(--scalar-border-width) * var(--tw-divide-x-reverse));border-left-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-x-reverse)))}.scalar-client .divide-y>:not([hidden])~:not([hidden]){--tw-divide-y-reverse:0;border-top-width:calc(var(--scalar-border-width) * calc(1 - var(--tw-divide-y-reverse)));border-bottom-width:calc(var(--scalar-border-width) * var(--tw-divide-y-reverse))}.scalar-client .self-center{align-self:center}.scalar-client .justify-self-end{justify-self:end}.scalar-client .overflow-auto{overflow:auto}.scalar-client .overflow-hidden{overflow:hidden}.scalar-client .overflow-visible{overflow:visible}.scalar-client .overflow-x-auto{overflow-x:auto}.scalar-client .overflow-y-auto{overflow-y:auto}.scalar-client .truncate{overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.scalar-client .text-ellipsis{text-overflow:ellipsis}.scalar-client .whitespace-nowrap{white-space:nowrap}.scalar-client .whitespace-pre{white-space:pre}.scalar-client .text-balance{text-wrap:balance}.scalar-client .text-pretty{text-wrap:pretty}.scalar-client .break-words{overflow-wrap:break-word}.scalar-client .break-all{word-break:break-all}.scalar-client .\!rounded-none{border-radius:0!important}.scalar-client .rounded{border-radius:var(--scalar-radius)}.scalar-client .rounded-\[10px\]{border-radius:10px}.scalar-client .rounded-full{border-radius:9999px}.scalar-client .rounded-lg{border-radius:var(--scalar-radius-lg)}.scalar-client .rounded-md{border-radius:var(--scalar-radius)}.scalar-client .rounded-px{border-radius:1px}.scalar-client .rounded-xl{border-radius:var(--scalar-radius-xl)}.scalar-client .rounded-b{border-bottom-right-radius:var(--scalar-radius);border-bottom-left-radius:var(--scalar-radius)}.scalar-client .rounded-b-lg{border-bottom-right-radius:var(--scalar-radius-lg);border-bottom-left-radius:var(--scalar-radius-lg)}.scalar-client .rounded-t{border-top-left-radius:var(--scalar-radius);border-top-right-radius:var(--scalar-radius)}.scalar-client .rounded-t-lg{border-top-left-radius:var(--scalar-radius-lg);border-top-right-radius:var(--scalar-radius-lg)}.scalar-client .rounded-t-none{border-top-left-radius:0;border-top-right-radius:0}.scalar-client .\!border-0{border-width:0!important}.scalar-client .border{border-width:var(--scalar-border-width)}.scalar-client .border-0{border-width:0}.scalar-client .border-1\/2{border-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-\[1\.5px\]{border-width:1.5px}.scalar-client .border-\[1px\]{border-width:1px}.scalar-client .\!border-r-1\/2{border-right-width:calc(var(--scalar-border-width) / 2)!important}.scalar-client .border-b{border-bottom-width:var(--scalar-border-width)}.scalar-client .border-b-0{border-bottom-width:0}.scalar-client .border-b-1\/2{border-bottom-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-b-\[1px\]{border-bottom-width:1px}.scalar-client .border-l{border-left-width:var(--scalar-border-width)}.scalar-client .border-l-0{border-left-width:0}.scalar-client .border-r{border-right-width:var(--scalar-border-width)}.scalar-client .border-r-0{border-right-width:0}.scalar-client .border-r-1\/2{border-right-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-t{border-top-width:var(--scalar-border-width)}.scalar-client .border-t-0{border-top-width:0}.scalar-client .border-t-1\/2{border-top-width:calc(var(--scalar-border-width) / 2)}.scalar-client .border-solid{border-style:solid}.scalar-client .border-dashed{border-style:dashed}.scalar-client .border-none{border-style:none}.scalar-client .\!border-current{border-color:currentColor!important}.scalar-client .border-b-3{border-color:var(--scalar-background-3)}.scalar-client .border-c-3{border-color:var(--scalar-color-3)}.scalar-client .border-transparent{border-color:transparent}.scalar-client .border-r-transparent{border-right-color:transparent}.scalar-client .bg-b-1{background-color:var(--scalar-background-1)}.scalar-client .bg-b-2{background-color:var(--scalar-background-2)}.scalar-client .bg-b-3{background-color:var(--scalar-background-3)}.scalar-client .bg-blue{background-color:var(--scalar-color-blue)}.scalar-client .bg-border{background-color:var(--scalar-border-color)}.scalar-client .bg-c-2{background-color:var(--scalar-color-2)}.scalar-client .bg-c-3{background-color:var(--scalar-color-3)}.scalar-client .bg-c-accent{background-color:var(--scalar-color-accent)}.scalar-client .bg-current{background-color:currentColor}.scalar-client .bg-green{background-color:var(--scalar-color-green)}.scalar-client .bg-grey{background-color:var(--scalar-color-3)}.scalar-client .bg-orange{background-color:var(--scalar-color-orange)}.scalar-client .bg-purple{background-color:var(--scalar-color-purple)}.scalar-client .bg-red{background-color:var(--scalar-color-red)}.scalar-client .bg-sidebar-active-b{background-color:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .bg-yellow{background-color:var(--scalar-color-yellow)}.scalar-client .bg-none{background-image:none}.scalar-client .fill-current{fill:currentColor}.scalar-client .stroke-2{stroke-width:2}.scalar-client .stroke-\[1\.5\]{stroke-width:1.5}.scalar-client .stroke-\[1\.75\]{stroke-width:1.75}.scalar-client .stroke-\[2\.25\]{stroke-width:2.25}.scalar-client .object-contain{-o-object-fit:contain;object-fit:contain}.scalar-client .\!p-0{padding:0!important}.scalar-client .p-0{padding:0}.scalar-client .p-0\.5{padding:2px}.scalar-client .p-0\.75{padding:3px}.scalar-client .p-1{padding:4px}.scalar-client .p-1\.5{padding:6px}.scalar-client .p-1\.75{padding:7px}.scalar-client .p-2{padding:8px}.scalar-client .p-3{padding:12px}.scalar-client .p-4{padding:16px}.scalar-client .p-\[3px\]{padding:3px}.scalar-client .p-\[5px\]{padding:5px}.scalar-client .\!px-3{padding-left:12px!important;padding-right:12px!important}.scalar-client .\!py-1{padding-top:4px!important;padding-bottom:4px!important}.scalar-client .\!py-1\.5{padding-top:6px!important;padding-bottom:6px!important}.scalar-client .px-0{padding-left:0;padding-right:0}.scalar-client .px-0\.5{padding-left:2px;padding-right:2px}.scalar-client .px-1{padding-left:4px;padding-right:4px}.scalar-client .px-1\.5{padding-left:6px;padding-right:6px}.scalar-client .px-2{padding-left:8px;padding-right:8px}.scalar-client .px-2\.5{padding-left:10px;padding-right:10px}.scalar-client .px-3{padding-left:12px;padding-right:12px}.scalar-client .px-4{padding-left:16px;padding-right:16px}.scalar-client .px-5{padding-left:20px;padding-right:20px}.scalar-client .px-6{padding-left:24px;padding-right:24px}.scalar-client .px-8{padding-left:32px;padding-right:32px}.scalar-client .py-0{padding-top:0;padding-bottom:0}.scalar-client .py-0\.5{padding-top:2px;padding-bottom:2px}.scalar-client .py-0\.75{padding-top:3px;padding-bottom:3px}.scalar-client .py-1{padding-top:4px;padding-bottom:4px}.scalar-client .py-1\.25{padding-top:5px;padding-bottom:5px}.scalar-client .py-1\.5{padding-top:6px;padding-bottom:6px}.scalar-client .py-2{padding-top:8px;padding-bottom:8px}.scalar-client .py-2\.5{padding-top:10px;padding-bottom:10px}.scalar-client .py-3{padding-top:12px;padding-bottom:12px}.scalar-client .py-5{padding-top:20px;padding-bottom:20px}.scalar-client .py-8{padding-top:32px;padding-bottom:32px}.scalar-client .py-px{padding-top:1px;padding-bottom:1px}.scalar-client .\!pl-3{padding-left:12px!important}.scalar-client .\!pt-0{padding-top:0!important}.scalar-client .pb-0{padding-bottom:0}.scalar-client .pb-14{padding-bottom:56px}.scalar-client .pb-2{padding-bottom:8px}.scalar-client .pb-3{padding-bottom:12px}.scalar-client .pb-5{padding-bottom:20px}.scalar-client .pb-8{padding-bottom:32px}.scalar-client .pb-\[75px\]{padding-bottom:75px}.scalar-client .pl-1{padding-left:4px}.scalar-client .pl-1\.25{padding-left:5px}.scalar-client .pl-1\.5{padding-left:6px}.scalar-client .pl-2{padding-left:8px}.scalar-client .pl-3{padding-left:12px}.scalar-client .pl-6{padding-left:24px}.scalar-client .pl-8{padding-left:32px}.scalar-client .pl-9{padding-left:36px}.scalar-client .pl-px{padding-left:1px}.scalar-client .pr-0{padding-right:0}.scalar-client .pr-1{padding-right:4px}.scalar-client .pr-1\.5{padding-right:6px}.scalar-client .pr-2{padding-right:8px}.scalar-client .pr-2\.5{padding-right:10px}.scalar-client .pr-4{padding-right:16px}.scalar-client .pr-6{padding-right:24px}.scalar-client .pr-\[26px\]{padding-right:26px}.scalar-client .pr-px{padding-right:1px}.scalar-client .pt-0{padding-top:0}.scalar-client .pt-2{padding-top:8px}.scalar-client .pt-3{padding-top:12px}.scalar-client .pt-4{padding-top:16px}.scalar-client .pt-px{padding-top:1px}.scalar-client .text-left{text-align:left}.scalar-client .text-center{text-align:center}.scalar-client .text-right{text-align:right}.scalar-client .font-code{font-family:var(--scalar-font-code)}.scalar-client .text-3xs{font-size:var(--scalar-font-size-7)}.scalar-client .text-\[11px\]{font-size:11px}.scalar-client .text-\[21px\]{font-size:21px}.scalar-client .text-\[6px\]{font-size:6px}.scalar-client .text-base{font-size:var(--scalar-font-size-3)}.scalar-client .text-sm{font-size:var(--scalar-font-size-4)}.scalar-client .text-xl{font-size:var(--scalar-font-size-1)}.scalar-client .text-xs{font-size:var(--scalar-font-size-5)}.scalar-client .text-xxs{font-size:var(--scalar-font-size-6)}.scalar-client .\!font-normal{font-weight:var(--scalar-regular)!important}.scalar-client .font-bold{font-weight:var(--scalar-bold)}.scalar-client .font-medium{font-weight:var(--scalar-semibold)}.scalar-client .font-normal{font-weight:var(--scalar-regular)}.scalar-client .uppercase{text-transform:uppercase}.scalar-client .lowercase{text-transform:lowercase}.scalar-client .capitalize{text-transform:capitalize}.scalar-client .\!leading-\[6px\]{line-height:6px!important}.scalar-client .leading{line-height:1.5}.scalar-client .leading-3{line-height:var(--scalar-line-height-3)}.scalar-client .leading-5{line-height:var(--scalar-line-height-5)}.scalar-client .leading-\[1\.44\]{line-height:1.44}.scalar-client .leading-\[21px\]{line-height:21px}.scalar-client .leading-\[7px\]{line-height:7px}.scalar-client .leading-\[normal\]{line-height:normal}.scalar-client .leading-none{line-height:1}.scalar-client .leading-normal{line-height:1.5}.scalar-client .leading-snug{line-height:1.375}.scalar-client .\!text-c-1{color:var(--scalar-color-1)!important}.scalar-client .text-b-1{color:var(--scalar-background-1)}.scalar-client .text-blue{color:var(--scalar-color-blue)}.scalar-client .text-border{color:var(--scalar-border-color)}.scalar-client .text-c-1{color:var(--scalar-color-1)}.scalar-client .text-c-2{color:var(--scalar-color-2)}.scalar-client .text-c-3{color:var(--scalar-color-3)}.scalar-client .text-c-btn{color:var(--scalar-button-1-color)}.scalar-client .text-green{color:var(--scalar-color-green)}.scalar-client .text-grey{color:var(--scalar-color-3)}.scalar-client .text-orange{color:var(--scalar-color-orange)}.scalar-client .text-purple{color:var(--scalar-color-purple)}.scalar-client .text-red{color:var(--scalar-color-red)}.scalar-client .text-sidebar-active-c{color:var(--scalar-sidebar-color-active, currentColor)}.scalar-client .text-sidebar-c-2{color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}.scalar-client .text-transparent{color:transparent}.scalar-client .text-yellow{color:var(--scalar-color-yellow)}.scalar-client .underline{text-decoration-line:underline}.scalar-client .no-underline{text-decoration-line:none}.scalar-client .decoration-c-3{text-decoration-color:var(--scalar-color-3)}.scalar-client .underline-offset-2{text-underline-offset:2px}.scalar-client .opacity-0{opacity:0}.scalar-client .opacity-100{opacity:1}.scalar-client .opacity-25{opacity:.25}.scalar-client .shadow{--tw-shadow:var(--scalar-shadow-1);--tw-shadow-colored:var(--scalar-shadow-1);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-\[-8px_0_4px_var\(--scalar-background-1\)\]{--tw-shadow:-8px 0 4px var(--scalar-background-1);--tw-shadow-colored:-8px 0 4px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-\[0_-8px_0_8px_var\(--scalar-background-1\)\,0_0_8px_8px_var\(--scalar-background-1\)\]{--tw-shadow:0 -8px 0 8px var(--scalar-background-1),0 0 8px 8px var(--scalar-background-1);--tw-shadow-colored:0 -8px 0 8px var(--tw-shadow-color), 0 0 8px 8px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-border{--tw-shadow:inset 0 0 0 1px var(--scalar-border-color);--tw-shadow-colored:inset 0 0 0 1px var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-lg{--tw-shadow:var(--scalar-shadow-2);--tw-shadow-colored:var(--scalar-shadow-2);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-none{--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .shadow-border{--tw-shadow-color:var(--scalar-border-color);--tw-shadow:var(--tw-shadow-colored)}.scalar-client .outline-none{outline:2px solid transparent;outline-offset:2px}.scalar-client .outline{outline-style:solid}.scalar-client .-outline-offset-1{outline-offset:-1px}.scalar-client .-outline-offset-2{outline-offset:-2px}.scalar-client .outline-offset-2{outline-offset:2px}.scalar-client .blur{--tw-blur:blur(8px);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .brightness-\[\.9\]{--tw-brightness:brightness(.9);filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .brightness-lifted{--tw-brightness:brightness(var(--scalar-lifted-brightness));filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .filter{filter:var(--tw-blur) var(--tw-brightness) var(--tw-contrast) var(--tw-grayscale) var(--tw-hue-rotate) var(--tw-invert) var(--tw-saturate) var(--tw-sepia) var(--tw-drop-shadow)}.scalar-client .backdrop-filter{-webkit-backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia);backdrop-filter:var(--tw-backdrop-blur) var(--tw-backdrop-brightness) var(--tw-backdrop-contrast) var(--tw-backdrop-grayscale) var(--tw-backdrop-hue-rotate) var(--tw-backdrop-invert) var(--tw-backdrop-opacity) var(--tw-backdrop-saturate) var(--tw-backdrop-sepia)}.scalar-client .transition{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,-webkit-backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter;transition-property:color,background-color,border-color,text-decoration-color,fill,stroke,opacity,box-shadow,transform,filter,backdrop-filter,-webkit-backdrop-filter;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .transition-colors{transition-property:color,background-color,border-color,text-decoration-color,fill,stroke;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .transition-none{transition-property:none}.scalar-client .transition-opacity{transition-property:opacity;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .transition-transform{transition-property:transform;transition-timing-function:cubic-bezier(.4,0,.2,1);transition-duration:.15s}.scalar-client .duration-150{transition-duration:.15s}.scalar-client .duration-200{transition-duration:.2s}.scalar-client .duration-300{transition-duration:.3s}.scalar-client .ease-in-out{transition-timing-function:cubic-bezier(.4,0,.2,1)}.scalar-client .bg-mix-transparent{background-color:color-mix(var(--tw-bg-mix-method, in srgb),transparent calc(var(--tw-bg-mix-amount, 0) * 1%),var(--tw-bg-base))}.scalar-client .bg-mix-amount-90{--tw-bg-mix-amount:90}.scalar-client .bg-mix-amount-95{--tw-bg-mix-amount:95}.scalar-client .bg-b-1{--tw-bg-base:var(--scalar-background-1)}.scalar-client .bg-b-2{--tw-bg-base:var(--scalar-background-2)}.scalar-client .bg-b-3{--tw-bg-base:var(--scalar-background-3)}.scalar-client .bg-blue{--tw-bg-base:var(--scalar-color-blue)}.scalar-client .bg-border{--tw-bg-base:var(--scalar-border-color)}.scalar-client .bg-c-2{--tw-bg-base:var(--scalar-color-2)}.scalar-client .bg-c-3{--tw-bg-base:var(--scalar-color-3)}.scalar-client .bg-c-accent{--tw-bg-base:var(--scalar-color-accent)}.scalar-client .bg-current{--tw-bg-base:currentColor}.scalar-client .bg-green{--tw-bg-base:var(--scalar-color-green)}.scalar-client .bg-grey{--tw-bg-base:var(--scalar-color-3)}.scalar-client .bg-orange{--tw-bg-base:var(--scalar-color-orange)}.scalar-client .bg-purple{--tw-bg-base:var(--scalar-color-purple)}.scalar-client .bg-red{--tw-bg-base:var(--scalar-color-red)}.scalar-client .bg-sidebar-active-b{--tw-bg-base:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .bg-yellow{--tw-bg-base:var(--scalar-color-yellow)}.scalar-client .app-drag-region{-webkit-app-region:drag}.scalar-client .app-no-drag-region{-webkit-app-region:no-drag}.scalar-client .\[--scalar-address-bar-height\:32px\]{--scalar-address-bar-height:32px}.scalar-client .\[--scalar-small\:--scalar-font-size-4\]{--scalar-small:var(--scalar-font-size-4)}.scalar-client .\*\:flex>*{display:flex}.scalar-client .\*\:h-8>*{height:32px}.scalar-client .\*\:cursor-pointer>*{cursor:pointer}.scalar-client .\*\:items-center>*{align-items:center}.scalar-client .\*\:rounded-none>*{border-radius:0}.scalar-client .\*\:border-t-1\/2>*{border-top-width:calc(var(--scalar-border-width) / 2)}.scalar-client .\*\:px-1>*{padding-left:4px;padding-right:4px}.scalar-client .\*\:px-1\.5>*{padding-left:6px;padding-right:6px}.scalar-client .\*\:pl-4>*{padding-left:16px}.scalar-client .before\:pointer-events-none:before{content:var(--tw-content);pointer-events:none}.scalar-client .before\:absolute:before{content:var(--tw-content);position:absolute}.scalar-client .before\:left-3:before{content:var(--tw-content);left:12px}.scalar-client .before\:left-\[calc\(\.75rem_\+_\.5px\)\]:before{content:var(--tw-content);left:calc(.75rem + .5px)}.scalar-client .before\:right-\[23px\]:before{content:var(--tw-content);right:23px}.scalar-client .before\:top-0:before{content:var(--tw-content);top:0}.scalar-client .before\:top-0\.5:before{content:var(--tw-content);top:2px}.scalar-client .before\:z-1:before{content:var(--tw-content);z-index:1}.scalar-client .before\:h-\[calc\(100\%_\+_\.5px\)\]:before{content:var(--tw-content);height:calc(100% + .5px)}.scalar-client .before\:min-h-\[calc\(100\%-4px\)\]:before{content:var(--tw-content);min-height:calc(100% - 4px)}.scalar-client .before\:w-3:before{content:var(--tw-content);width:12px}.scalar-client .before\:w-\[\.5px\]:before{content:var(--tw-content);width:.5px}.scalar-client .before\:bg-border:before{content:var(--tw-content);background-color:var(--scalar-border-color)}.scalar-client .before\:bg-gradient-to-r:before{content:var(--tw-content);background-image:linear-gradient(to right,var(--tw-gradient-stops))}.scalar-client .before\:from-transparent:before{content:var(--tw-content);--tw-gradient-from:transparent var(--tw-gradient-from-position);--tw-gradient-to:rgb(0 0 0 / 0) var(--tw-gradient-to-position);--tw-gradient-stops:var(--tw-gradient-from), var(--tw-gradient-to)}.scalar-client .before\:to-b-1:before{content:var(--tw-content);--tw-gradient-to:var(--scalar-background-1) var(--tw-gradient-to-position)}.scalar-client .before\:content-\[\'\'\]:before{--tw-content:"";content:var(--tw-content)}.scalar-client .before\:bg-border:before{content:var(--tw-content);--tw-bg-base:var(--scalar-border-color)}.scalar-client .after\:content-\[\'\:\'\]:after{--tw-content:":";content:var(--tw-content)}.scalar-client .first\:border-t:first-child{border-top-width:var(--scalar-border-width)}.scalar-client .first\:\*\:rounded-l>*:first-child{border-top-left-radius:var(--scalar-radius);border-bottom-left-radius:var(--scalar-radius)}.scalar-client .first\:\*\:border-t-0>*:first-child{border-top-width:0}.scalar-client .last\:mb-0:last-child{margin-bottom:0}.scalar-client .last\:border-r-0:last-child{border-right-width:0}.scalar-client .last\:\*\:rounded-r>*:last-child{border-top-right-radius:var(--scalar-radius);border-bottom-right-radius:var(--scalar-radius)}.scalar-client .last\:before\:h-full:last-child:before{content:var(--tw-content);height:100%}.scalar-client .focus-within\:z-20:focus-within{z-index:20}.scalar-client .focus-within\:border-b-3:focus-within{border-color:var(--scalar-background-3)}.scalar-client .focus-within\:bg-b-1:focus-within{background-color:var(--scalar-background-1)}.scalar-client .focus-within\:text-c-1:focus-within{color:var(--scalar-color-1)}.scalar-client .focus-within\:bg-b-1:focus-within{--tw-bg-base:var(--scalar-background-1)}.scalar-client .hover\:cursor-default:hover{cursor:default}.scalar-client .hover\:bg-b-2:hover{background-color:var(--scalar-background-2)}.scalar-client .hover\:bg-b-3:hover{background-color:var(--scalar-background-3)}.scalar-client .hover\:bg-inherit:hover{background-color:inherit}.scalar-client .hover\:bg-sidebar-active-b:hover{background-color:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .hover\:text-c-1:hover{color:var(--scalar-color-1)}.scalar-client .hover\:text-c-2:hover{color:var(--scalar-color-2)}.scalar-client .hover\:bg-b-2:hover{--tw-bg-base:var(--scalar-background-2)}.scalar-client .hover\:bg-b-3:hover{--tw-bg-base:var(--scalar-background-3)}.scalar-client .hover\:bg-inherit:hover{--tw-bg-base:inherit}.scalar-client .hover\:bg-sidebar-active-b:hover{--tw-bg-base:var(--scalar-sidebar-item-active-background, var(--scalar-background-2))}.scalar-client .focus\:text-c-1:focus{color:var(--scalar-color-1)}.scalar-client .focus\:outline-none:focus{outline:2px solid transparent;outline-offset:2px}.scalar-client .focus-visible\:z-10:focus-visible{z-index:10}.scalar-client .active\:text-c-1:active{color:var(--scalar-color-1)}.scalar-client .disabled\:cursor-default:disabled{cursor:default}.scalar-client .disabled\:text-c-2:disabled{color:var(--scalar-color-2)}.scalar-client .group\/label:last-child .group-last\/label\:rounded-br-lg{border-bottom-right-radius:var(--scalar-radius-lg)}.scalar-client .group:last-child .group-last\:border-b-transparent{border-bottom-color:transparent}.scalar-client .group\/upload:hover .group-hover\/upload\:block,.scalar-client .group:hover .group-hover\:block{display:block}.scalar-client .group\/item:hover .group-hover\/item\:flex,.scalar-client .group:hover .group-hover\:flex{display:flex}.scalar-client .group:hover .group-hover\:hidden{display:none}.scalar-client .group:hover .group-hover\:pr-5{padding-right:20px}.scalar-client .group\/info:hover .group-hover\/info\:text-c-1{color:var(--scalar-color-1)}.scalar-client .group\/scopes-accordion:hover .group-hover\/scopes-accordion\:text-c-2{color:var(--scalar-color-2)}.scalar-client .group:hover .group-hover\:text-c-1{color:var(--scalar-color-1)}.scalar-client .group\/cell:hover .group-hover\/cell\:opacity-100,.scalar-client .group\/params:hover .group-hover\/params\:opacity-100,.scalar-client .group:hover .group-hover\:opacity-100{opacity:1}.scalar-client .group:hover .group-hover\:opacity-80{opacity:.8}.scalar-client .group:focus-visible .group-focus-visible\:opacity-100{opacity:1}.scalar-client .group:focus-visible .group-focus-visible\:outline{outline-style:solid}.scalar-client .group.alert .group-\[\.alert\]\:bg-b-alert{background-color:var(--scalar-background-alert)}.scalar-client .group.alert .group-\[\.alert\]\:bg-transparent{background-color:transparent}.scalar-client .group.alert .group-\[\.alert\]\:shadow-none{--tw-shadow:0 0 #0000;--tw-shadow-colored:0 0 var(--tw-shadow-color);box-shadow:var(--tw-ring-offset-shadow, 0 0 #0000),var(--tw-ring-shadow, 0 0 #0000),var(--tw-shadow)}.scalar-client .group.alert .group-\[\.alert\]\:outline-orange{outline-color:var(--scalar-color-orange)}.scalar-client .group.alert .group-\[\.alert\]\:bg-b-alert{--tw-bg-base:var(--scalar-background-alert)}.scalar-client .group.alert .group-\[\.alert\]\:bg-transparent{--tw-bg-base:transparent}.scalar-client .group.alert .group-\[\.alert\]\:before\:to-b-alert:before{content:var(--tw-content);--tw-gradient-to:var(--scalar-background-alert) var(--tw-gradient-to-position)}.scalar-client .peer:checked~.peer-checked\:text-c-1{color:var(--scalar-color-1)}.scalar-client .has-\[\:focus-visible\]\:rounded-\[4px\]:has(:focus-visible){border-radius:4px}.scalar-client .has-\[\.empty-sidebar-item\]\:border-t-1\/2:has(.empty-sidebar-item){border-top-width:calc(var(--scalar-border-width) / 2)}.scalar-client .has-\[\:focus-visible\]\:opacity-100:has(:focus-visible){opacity:1}.scalar-client .has-\[\:focus-visible\]\:outline:has(:focus-visible){outline-style:solid}.scalar-client .group\/label:has(input) .group-has-\[input\]\/label\:mr-0{margin-right:0}.scalar-client .group\/cell:has(:focus-visible) .group-has-\[\:focus-visible\]\/cell\:border-c-accent{border-color:var(--scalar-color-accent)}.scalar-client .group\/cell:has(:focus-visible) .group-has-\[\:focus-visible\]\/cell\:opacity-100{opacity:1}.scalar-client .group:has(:focus-visible) .group-has-\[\:focus-visible\]\:opacity-100{opacity:1}.scalar-client .peer:has(.color-selector)~.peer-has-\[\.color-selector\]\:hidden{display:none}.scalar-client .peer:has(.cm-focused)~.peer-has-\[\.cm-focused\]\:opacity-0{opacity:0}.scalar-client .peer:has(:focus-visible)~.peer-has-\[\:focus-visible\]\:opacity-0{opacity:0}.scalar-client .ui-open\:rotate-90[data-headlessui-state~=open]{--tw-rotate:90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client :where([data-headlessui-state~=open]) .ui-open\:rotate-90{--tw-rotate:90deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .ui-not-open\:invisible[data-headlessui-state]:not([data-headlessui-state~=open]){visibility:hidden}.scalar-client .ui-not-open\:rotate-0[data-headlessui-state]:not([data-headlessui-state~=open]){--tw-rotate:0deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client :where([data-headlessui-state]:not([data-headlessui-state~=open])) .ui-not-open\:invisible:not([data-headlessui-state]){visibility:hidden}.scalar-client :where([data-headlessui-state]:not([data-headlessui-state~=open])) .ui-not-open\:rotate-0:not([data-headlessui-state]){--tw-rotate:0deg;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .ui-checked\:bg-b-3[data-headlessui-state~=checked]{background-color:var(--scalar-background-3);--tw-bg-base:var(--scalar-background-3)}.scalar-client :where([data-headlessui-state~=checked]) .ui-checked\:bg-b-3{background-color:var(--scalar-background-3);--tw-bg-base:var(--scalar-background-3)}.scalar-client .ui-active\:bg-b-2[data-headlessui-state~=active]{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .\*\:ui-active\:bg-b-2[data-headlessui-state~=active]>*{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client :where([data-headlessui-state~=active]) .ui-active\:bg-b-2{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client :where([data-headlessui-state~=active]) .\*\:ui-active\:bg-b-2>*{background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .app-platform-desktop .desktop\:rounded-lg{border-radius:var(--scalar-radius-lg)}.scalar-client .app-platform-mac .mac\:hidden{display:none}.scalar-client .app-platform-mac .mac\:pl-\[72px\]{padding-left:72px}@media (min-width: 600px){.scalar-client .sm\:not-sr-only{position:static;width:auto;height:auto;padding:0;margin:0;overflow:visible;clip:auto;white-space:normal}.scalar-client .sm\:order-none{order:0}.scalar-client .sm\:mb-1{margin-bottom:4px}.scalar-client .sm\:mb-1\.5{margin-bottom:6px}.scalar-client .sm\:ml-1{margin-left:4px}.scalar-client .sm\:ml-1\.5{margin-left:6px}.scalar-client .sm\:mr-1{margin-right:4px}.scalar-client .sm\:mr-1\.5{margin-right:6px}.scalar-client .sm\:flex{display:flex}.scalar-client .sm\:hidden{display:none}.scalar-client .sm\:min-w-max{min-width:-moz-max-content;min-width:max-content}.scalar-client .sm\:max-w-max{max-width:-moz-max-content;max-width:max-content}.scalar-client .sm\:flex-row{flex-direction:row}.scalar-client .sm\:flex-col{flex-direction:column}.scalar-client .sm\:justify-between{justify-content:space-between}.scalar-client .sm\:gap-px{gap:1px}.scalar-client .sm\:rounded{border-radius:var(--scalar-radius)}.scalar-client .sm\:rounded-lg{border-radius:var(--scalar-radius-lg)}.scalar-client .sm\:px-2{padding-left:8px;padding-right:8px}.scalar-client .sm\:px-3{padding-left:12px;padding-right:12px}.scalar-client .sm\:py-1{padding-top:4px;padding-bottom:4px}.scalar-client .sm\:py-1\.5{padding-top:6px;padding-bottom:6px}.scalar-client .sm\:\*\:rounded-lg>*{border-radius:var(--scalar-radius-lg)}}@media (min-width: 800px){.scalar-client .md\:bottom-10{bottom:40px}.scalar-client .md\:right-10{right:40px}.scalar-client .md\:-ml-1{margin-left:-4px}.scalar-client .md\:-ml-1\.25{margin-left:-5px}.scalar-client .md\:block{display:block}.scalar-client .md\:w-full{width:100%}.scalar-client .md\:min-w-fit{min-width:-moz-fit-content;min-width:fit-content}.scalar-client .md\:flex-none{flex:none}.scalar-client .md\:translate-x-0{--tw-translate-x:0px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .md\:translate-y-0{--tw-translate-y:0px;transform:translate(var(--tw-translate-x),var(--tw-translate-y)) rotate(var(--tw-rotate)) skew(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))}.scalar-client .md\:flex-row{flex-direction:row}.scalar-client .md\:border-b-0{border-bottom-width:0}.scalar-client .md\:border-r-1\/2{border-right-width:calc(var(--scalar-border-width) / 2)}.scalar-client .md\:p-1{padding:4px}.scalar-client .md\:p-1\.5{padding:6px}.scalar-client .md\:px-0{padding-left:0;padding-right:0}.scalar-client .md\:px-1{padding-left:4px;padding-right:4px}.scalar-client .md\:px-1\.5{padding-left:6px;padding-right:6px}.scalar-client .md\:px-2{padding-left:8px;padding-right:8px}.scalar-client .md\:px-2\.5{padding-left:10px;padding-right:10px}.scalar-client .md\:px-4{padding-left:16px;padding-right:16px}.scalar-client .md\:px-\[18px\]{padding-left:18px;padding-right:18px}.scalar-client .md\:py-2{padding-top:8px;padding-bottom:8px}.scalar-client .md\:py-2\.5{padding-top:10px;padding-bottom:10px}.scalar-client .md\:pb-2{padding-bottom:8px}.scalar-client .md\:pb-2\.5{padding-bottom:10px}.scalar-client .md\:pb-\[37px\]{padding-bottom:37px}.scalar-client .md\:\*\:border-t-0>*{border-top-width:0}}@media (min-width: 1000px){.scalar-client .lg\:order-none{order:0}.scalar-client .lg\:-mr-1{margin-right:-4px}.scalar-client .lg\:mb-0{margin-bottom:0}.scalar-client .lg\:flex{display:flex}.scalar-client .lg\:min-h-client-header{min-height:48px}.scalar-client .lg\:w-\[120px\]{width:120px}.scalar-client .lg\:w-auto{width:auto}.scalar-client .lg\:min-w-\[580px\]{min-width:580px}.scalar-client .lg\:max-w-\[580px\]{max-width:580px}.scalar-client .lg\:flex-1{flex:1 1 0%}.scalar-client .lg\:p-1{padding:4px}.scalar-client .lg\:px-1{padding-left:4px;padding-right:4px}.scalar-client .lg\:px-2{padding-left:8px;padding-right:8px}.scalar-client .lg\:px-2\.5{padding-left:10px;padding-right:10px}.scalar-client .lg\:pr-24{padding-right:96px}.scalar-client .lg\:pt-1{padding-top:4px}.scalar-client .lg\:text-sm{font-size:var(--scalar-font-size-4)}}@media (min-width: 1200px){.scalar-client .xl\:\!flex{display:flex!important}.scalar-client .xl\:flex{display:flex}.scalar-client .xl\:hidden{display:none}.scalar-client .xl\:h-full{height:100%}.scalar-client .xl\:min-h-client-header{min-height:48px}.scalar-client .xl\:min-w-0{min-width:0px}.scalar-client .xl\:min-w-\[720px\]{min-width:720px}.scalar-client .xl\:max-w-\[720px\]{max-width:720px}.scalar-client .xl\:flex-row{flex-direction:row}.scalar-client .xl\:overflow-hidden{overflow:hidden}.scalar-client .xl\:rounded-none{border-radius:0}.scalar-client .xl\:pl-2{padding-left:8px}.scalar-client .xl\:pr-0{padding-right:0}.scalar-client .xl\:pr-0\.5{padding-right:2px}.scalar-client .xl\:\*\:border-l-1\/2>*{border-left-width:calc(var(--scalar-border-width) / 2)}.scalar-client .xl\:\*\:border-t-0>*{border-top-width:0}.scalar-client .xl\:first\:\*\:border-l-0>*:first-child{border-left-width:0}}.scalar-client .dark\:bg-b-2:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .dark\:hover\:bg-b-2:hover:where(.dark-mode,.dark-mode *){background-color:var(--scalar-background-2);--tw-bg-base:var(--scalar-background-2)}.scalar-client .\[\&\>a\]\:pl-5>a{padding-left:20px}.scalar-client .\[\&\>a\]\:pl-\[1\.625rem\]>a{padding-left:1.625rem}html:has(.dark-mode){--scalar-background-1: #0f0f0f}html,body{background-color:var(--scalar-background-1);overscroll-behavior:none}#scalar-client{display:flex;flex-direction:column;height:100dvh;width:100dvw;position:relative;background-color:var(--scalar-background-2)}.dark-mode #scalar-client{background-color:color-mix(in srgb,var(--scalar-background-1) 65%,black)}.open-api-client-button[data-v-2e9f75f1]{cursor:pointer;width:100%;padding:9px 12px;height:31px;display:block;text-align:center;display:flex;align-items:center;justify-content:center;font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);line-height:1.385;text-decoration:none;border-radius:var(--scalar-radius);box-shadow:0 0 0 .5px var(--scalar-border-color);gap:6px;color:var(--scalar-sidebar-color-1)}.open-api-client-button[data-v-2e9f75f1]:hover{background:var( --scalar-sidebar-item-hover-background, var(--scalar-background-2) )}.addressbar-history-button[data-v-a323ca24]:hover{background:var(--scalar-background-3)}.addressbar-history-button[data-v-a323ca24]:focus-within{background:var(--scalar-background-2)}.addressbar-bg-states:has(.cm-focused) .codemirror-bg-switcher[data-v-a323ca24]{--scalar-background-1: var(--scalar-background-1)}.addressbar-bg-states[data-v-a323ca24]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2))}.addressbar-bg-states[data-v-a323ca24]:has(.cm-focused){background:var(--scalar-background-1);border-color:var(--scalar-border-color);outline:1px solid var(--scalar-color-accent)}.description[data-v-8eba7808] .markdown{font-weight:var(--scalar-semibold);color:var(--scalar-color--1);padding:0;display:block}.description[data-v-8eba7808] .markdown>*:first-child{margin-top:0}[data-v-b4d63e6c] .cm-editor{background-color:var(--scalar-background-1);height:100%;outline:none;width:100%}[data-v-b4d63e6c] .cm-content{padding:0;display:flex;align-items:center;font-size:var(--scalar-mini)}.scroll-timeline-x[data-v-b4d63e6c]{scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;-ms-overflow-style:none}.scroll-timeline-x-hidden[data-v-b4d63e6c]{overflow-x:auto}.scroll-timeline-x-hidden[data-v-b4d63e6c] .cm-scroller{scrollbar-width:none;-ms-overflow-style:none;padding-right:20px;overflow:auto}.scroll-timeline-x-hidden[data-v-b4d63e6c]::-webkit-scrollbar{width:0;height:0;display:none}.scroll-timeline-x-hidden[data-v-b4d63e6c] .cm-scroller::-webkit-scrollbar{width:0;height:0;display:none}.scroll-timeline-x-address[data-v-b4d63e6c]{line-height:27px;scrollbar-width:none}.scroll-timeline-x-address[data-v-b4d63e6c]:after{content:"";position:absolute;height:100%;width:24px;right:0;cursor:text}.scroll-timeline-x-address[data-v-b4d63e6c]:empty:before{content:"Enter URL or cURL request";color:var(--scalar-color-3);pointer-events:none}.fade-left[data-v-b4d63e6c],.fade-right[data-v-b4d63e6c]{content:"";position:sticky;height:100%;animation-name:fadein-b4d63e6c;animation-duration:1ms;animation-direction:reverse;animation-timeline:--scroll-timeline;pointer-events:none}.fade-left[data-v-b4d63e6c]{background:linear-gradient(-90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 30%,var(--scalar-background-1) 100%);left:0;min-width:3px;animation-direction:normal}.fade-right[data-v-b4d63e6c]{background:linear-gradient(90deg,color-mix(in srgb,var(--scalar-background-1),transparent 100%) 0%,color-mix(in srgb,var(--scalar-background-1),transparent 20%) 30%,var(--scalar-background-1) 100%);right:0;min-width:24px}@keyframes fadein-b4d63e6c{0%{opacity:0}1%{opacity:1}}.codemirror-bg-switcher[data-v-b4d63e6c]{--scalar-background-1: color-mix( in srgb, var(--scalar-background-1), var(--scalar-background-2) )}.addressbar-bg-states:has(.cm-focused) .codemirror-bg-switcher[data-v-b4d63e6c]{--scalar-background-1: var(--scalar-background-1)}.addressbar-bg-states[data-v-b4d63e6c]{background:color-mix(in srgb,var(--scalar-background-1),var(--scalar-background-2))}.addressbar-bg-states[data-v-b4d63e6c]:has(.cm-focused){background:var(--scalar-background-1);border-color:var(--scalar-border-color);outline:1px solid var(--scalar-color-accent)}.sidebar-height[data-v-5397f2a7]{min-height:100%}@media (min-width: 800px){.sidebar-mask[data-v-5397f2a7]{-webkit-mask-image:linear-gradient(0,transparent 0,transparent 0,var(--scalar-background-2) 30px);mask-image:linear-gradient(0,transparent 0,transparent 0,var(--scalar-background-2) 30px)}}.resizer[data-v-5397f2a7]{width:5px;cursor:col-resize;position:absolute;top:0;right:0;bottom:0;border-right:2px solid transparent;transition:border-right-color .3s}.resizer[data-v-5397f2a7]:hover,.dragging .resizer[data-v-5397f2a7]{border-right-color:var(--scalar-background-3)}.dragging[data-v-5397f2a7]{cursor:col-resize}.dragging[data-v-5397f2a7]:before{content:"";display:block;position:absolute;width:100%;height:100%}.scroll-timeline-x[data-v-b77d5bd3]{overflow:auto;scroll-timeline:--scroll-timeline x;scroll-timeline:--scroll-timeline horizontal;-ms-overflow-style:none;scrollbar-width:none}.gitbook-show[data-v-cac03082]{display:none}.app-exit-button[data-v-cac03082]{color:#fff;background:#0000001a}.app-exit-button[data-v-cac03082]:hover{background:#ffffff1a}.ellipsis-position[data-v-daf8ea6e]{transform:translate3d(calc(-100% - 4.5px),0,0)}.dragover-asChild[data-v-3449bf19],.dragover-above[data-v-3449bf19],.dragover-below[data-v-3449bf19]{position:relative}.dragover-above[data-v-3449bf19]:after,.dragover-below[data-v-3449bf19]:after{content:"";position:absolute;top:-1.5px;height:3px;width:100%;display:block;background:color-mix(in srgb,var(--scalar-color-blue),transparent 85%);pointer-events:none;border-radius:var(--scalar-radius)}.dragover-below[data-v-3449bf19]:after{top:initial;bottom:-1.5px}.dragover-asChild[data-v-3449bf19]:after{content:"";position:absolute;top:0;left:0;width:100%;height:100%;display:block;background:color-mix(in srgb,var(--scalar-color-blue),transparent 85%);pointer-events:none;border-radius:var(--scalar-radius)}.indent-border-line-offset[data-v-682d438e]:before{left:var(--0ff55690)}.indent-padding-left[data-v-682d438e]{padding-left:calc(var(--ecb155a4) + 6px)}.sidebar-folderitem[data-v-682d438e] .ellipsis-position{right:6px;transform:none}.search-button-fade[data-v-b7bf7c07]{background:linear-gradient(var(--scalar-background-1) 32px,color-mix(in srgb,var(--scalar-background-1),transparent) 38px,transparent)}.empty-sidebar-item-content[data-v-b7bf7c07]{display:none}.empty-sidebar-item .empty-sidebar-item-content[data-v-b7bf7c07]{display:block}.rabbitjump[data-v-b7bf7c07]{opacity:0}.empty-sidebar-item:hover .rabbitjump[data-v-b7bf7c07]{opacity:1;animation:rabbitAnimation-b7bf7c07 .5s steps(1) infinite}.empty-sidebar-item:hover .rabbitsit[data-v-b7bf7c07]{opacity:0;animation:rabbitAnimation2-b7bf7c07 .5s steps(1) infinite}.empty-sidebar-item:hover .rabbit-ascii[data-v-b7bf7c07]{animation:rabbitRun-b7bf7c07 8s infinite linear}@keyframes rabbitRun-b7bf7c07{0%{transform:translateZ(0)}25%{transform:translate3d(250px,0,0)}25.01%{transform:translate3d(-250px,0,0)}75%{transform:translate3d(250px,0,0)}75.01%{transform:translate3d(-250px,0,0)}to{transform:translateZ(0)}}@keyframes rabbitAnimation-b7bf7c07{0%,to{opacity:1}50%{opacity:0}}@keyframes rabbitAnimation2-b7bf7c07{0%,to{opacity:0}50%{opacity:1;transform:translate3d(0,-8px,0)}}.request-text-color-text[data-v-02ba9b2d]{color:var(--scalar-color-1);background:linear-gradient(var(--scalar-background-1),var(--scalar-background-3));box-shadow:0 0 0 1px var(--scalar-border-color)}@media screen and (max-width: 800px){.sidebar-active-hide-layout[data-v-02ba9b2d]{display:none}.sidebar-active-width[data-v-02ba9b2d]{width:100%}}.active-link[data-v-a6c1ff1c]{background-color:var(--scalar-background-2);color:var(--scalar-color-1);--tw-bg-base: var(--scalar-background-2)}.empty-variable-name[data-v-a6c1ff1c]:empty:before{content:"Untitled";color:var(--scalar-color-3)}.form-group[data-v-c25da223]{margin-bottom:1rem}.modal-actions[data-v-c25da223]{display:flex;justify-content:flex-end;gap:1rem}:root{--scalar-loaded-api-reference: true}@layer scalar-config{.scalar-api-reference[data-v-10065932]{--refs-sidebar-width: var(--scalar-sidebar-width, 0px);--refs-header-height: calc( var(--scalar-y-offset) + var(--scalar-header-height, 0px) );--refs-content-max-width: var(--scalar-content-max-width, 1540px)}.scalar-api-reference.references-classic[data-v-10065932]{--refs-content-max-width: var(--scalar-content-max-width, 1420px);min-height:100dvh;--refs-sidebar-width: 0}}.references-layout[data-v-10065932]{min-height:100dvh;min-width:100%;max-width:100%;flex:1;--full-height: 100dvh;display:grid;grid-template-rows:var(--scalar-header-height, 0px) repeat(2,auto);grid-template-columns:var(--refs-sidebar-width) 1fr;grid-template-areas:"header header" "navigation rendered" "footer footer";background:var(--scalar-background-1)}.references-header[data-v-10065932]{grid-area:header;position:sticky;top:var(--scalar-custom-header-height, 0px);z-index:10;height:var(--scalar-header-height, 0px)}.references-editor[data-v-10065932]{grid-area:editor;display:flex;min-width:0;background:var(--scalar-background-1);z-index:1}.references-navigation[data-v-10065932]{grid-area:navigation}.references-rendered[data-v-10065932]{position:relative;grid-area:rendered;min-width:0;background:var(--scalar-background-1)}.scalar-api-reference.references-classic[data-v-10065932],.references-classic .references-rendered[data-v-10065932]{--full-height: fit-content !important;height:initial!important;max-height:initial!important}.references-navigation-list[data-v-10065932]{position:sticky;top:var(--refs-header-height);height:calc(100dvh - var(--refs-header-height));background:var(--scalar-sidebar-background-1, var(--scalar-background-1));overflow-y:auto;display:flex;flex-direction:column}.references-editor-textarea[data-v-10065932]{position:sticky;top:var(--refs-header-height);height:calc(var(--full-height) - var(--refs-header-height));display:flex;min-width:0;flex:1}.references-editable[data-v-10065932]{grid-template-columns:var(--refs-sidebar-width) 1fr 1fr;grid-template-areas:"header header header" "navigation editor rendered" "footer footer footer"}@layer scalar-config{.references-sidebar[data-v-10065932]{--refs-sidebar-width: var(--scalar-sidebar-width, 280px)}}.references-footer[data-v-10065932]{grid-area:footer}@media (max-width: 1150px){.references-layout[data-v-10065932]{grid-template-columns:var(--refs-sidebar-width) 1fr 0px}}@media (max-width: 1000px){.references-layout[data-v-10065932]{grid-template-columns:auto;grid-template-rows:var(--scalar-header-height, 0px) 0px auto auto;grid-template-areas:"header" "navigation" "rendered" "footer"}.references-editable[data-v-10065932]{grid-template-areas:"header" "navigation" "editor"}.references-navigation[data-v-10065932],.references-rendered[data-v-10065932]{max-height:unset}.references-rendered[data-v-10065932]{position:static}.references-navigation[data-v-10065932]{display:none;z-index:10}.references-sidebar-mobile-open .references-navigation[data-v-10065932]{display:block;top:var(--refs-header-height);height:calc(100dvh - var(--refs-header-height));width:100%;position:sticky}.references-navigation-list[data-v-10065932]{position:absolute;top:-1px;height:calc(var(--full-height) - var(--refs-header-height) + 1px);width:100%;border-top:1px solid var(--scalar-border-color);display:flex;flex-direction:column}}.references-classic-header[data-v-a860f78f]{display:flex;align-items:center;justify-content:space-between;gap:12px;max-width:var(--refs-content-max-width);margin:auto;padding:12px 0}.references-classic-header-container[data-v-a860f78f]{padding:0 60px}@container narrow-references-container (max-width: 900px){.references-classic-header[data-v-a860f78f]{padding:12px 24px}.references-classic-header-container[data-v-a860f78f]{padding:0}}.references-classic-header-icon[data-v-a860f78f]{height:24px;color:var(--scalar-color-1)}a[data-v-829b4762]{text-decoration:none}.ref-search-container[data-v-829b4762]{display:flex;flex-direction:column;padding:12px 12px 0}.ref-search-results[data-v-829b4762]{padding:12px}.ref-search-meta[data-v-829b4762]{background:var(--scalar-background-3);padding:6px 12px;font-size:var(--scalar-font-size-4);color:var(--scalar-color-3);font-weight:var(--scalar-semibold);display:flex;gap:12px}.deprecated[data-v-829b4762]{text-decoration:line-through}.sidebar-search[data-v-9f0bd6b4]{display:flex;align-items:center;position:relative;padding:0 3px 0 9px;min-width:254px;max-width:100%;font-family:var(--scalar-font);background:var( --scalar-sidebar-search-background, var(--scalar-background-1) );color:var(--scalar-sidebar-color-2, var(--scalar-color-2));border-radius:var(--scalar-radius);box-shadow:0 0 0 .5px var(--scalar-sidebar-search-border-color, var(--scalar-border-color));cursor:pointer;-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none}.sidebar-search-input[data-v-9f0bd6b4]{font-size:var(--scalar-mini);font-weight:var(--scalar-semibold);height:31px;-webkit-user-select:none;-moz-user-select:none;user-select:none;z-index:10;position:relative;display:flex;width:100%;justify-content:space-between;align-items:center}.sidebar-search-key[data-v-9f0bd6b4]{text-transform:uppercase;background-color:var(--scalar-background-2);padding:3px 5px;margin:2px;border-radius:var(--scalar-radius);color:var(--scalar-sidebar-color-2, var(--scalar-color-2))}.scalar-search-icon[data-v-9f0bd6b4]{padding:0;margin-right:6px;width:12px}.references-mobile-header[data-v-b5fb4b8b]{display:none;align-items:center;height:100%;width:100%;padding:0 8px;background:var(--scalar-background-1);border-bottom:1px solid var(--scalar-border-color)}.references-mobile-breadcrumbs[data-v-b5fb4b8b]{flex:1;min-width:0;font-size:var(--scalar-small);font-weight:var(--scalar-semibold);color:var(--scalar-color-1);overflow:hidden;white-space:nowrap;text-overflow:ellipsis}.references-mobile-header-actions[data-v-b5fb4b8b]{display:flex;flex-direction:row;gap:4px;height:24px;align-items:center;padding-left:4px}@media (max-width: 1000px){.references-mobile-header[data-v-b5fb4b8b]{display:flex}}@media (max-width: 1000px){.scalar-api-references-standalone-mobile{--scalar-header-height: 50px}}.scalar-api-references-standalone-search[data-v-3ca2ed96]{display:flex;flex-direction:column;padding:12px 12px 6px}.darklight-reference[data-v-3ca2ed96]{width:100%;margin-top:auto}@layer scalar-base{body{margin:0;background-color:var(--scalar-background-1)}}
/*! tailwindcss v4.0.7 | MIT License | https://tailwindcss.com */
@layer theme{:root,:host{--font-sans:"Lato",ui-sans-serif,system-ui,sans-serif,"Apple Color Emoji","Segoe UI Emoji","Segoe UI Symbol","Noto Color Emoji";--font-serif:"Lato",ui-serif,Georgia,Cambria,"Times New Roman",Times,serif;--font-mono:"Fira Mono",ui-monospace,SFMono-Regular,Menlo,Monaco,Consolas,"Liberation Mono","Courier New",monospace;--spacing:.25rem;--breakpoint-sm:40rem;--breakpoint-md:48rem;--breakpoint-lg:64rem;--breakpoint-xl:80rem;--breakpoint-2xl:96rem;--container-3xs:16rem;--container-2xs:18rem;--container-xs:20rem;--container-sm:24rem;--container-md:28rem;--container-lg:32rem;--container-xl:36rem;--container-2xl:42rem;--container-3xl:48rem;--container-4xl:56rem;--container-5xl:64rem;--container-6xl:72rem;--container-7xl:80rem;--text-xs:.75rem;--text-xs--line-height:calc(1/.75);--text-sm:.875rem;--text-sm--line-height:calc(1.25/.875);--text-base:1rem;--text-base--line-height:calc(1.5/1);--text-lg:1.125rem;--text-lg--line-height:calc(1.75/1.125);--text-xl:1.25rem;--text-xl--line-height:calc(1.75/1.25);--text-2xl:1.5rem;--text-2xl--line-height:calc(2/1.5);--text-3xl:1.875rem;--text-3xl--line-height:calc(2.25/1.875);--text-4xl:2.25rem;--text-4xl--line-height:calc(2.5/2.25);--text-5xl:3rem;--text-5xl--line-height:1;--text-6xl:3.75rem;--text-6xl--line-height:1;--text-7xl:4.5rem;--text-7xl--line-height:1;--text-8xl:6rem;--text-8xl--line-height:1;--text-9xl:8rem;--text-9xl--line-height:1;--font-weight-thin:100;--font-weight-extralight:200;--font-weight-light:300;--font-weight-normal:400;--font-weight-medium:500;--font-weight-semibold:600;--font-weight-bold:700;--font-weight-extrabold:800;--font-weight-black:900;--tracking-tighter:-.05em;--tracking-tight:-.025em;--tracking-normal:0em;--tracking-wide:.025em;--tracking-wider:.05em;--tracking-widest:.1em;--leading-tight:1.25;--leading-snug:1.375;--leading-normal:1.5;--leading-relaxed:1.625;--leading-loose:2;--radius-xs:.125rem;--radius-sm:.25rem;--radius-md:.375rem;--radius-lg:.5rem;--radius-xl:.75rem;--radius-2xl:1rem;--radius-3xl:1.5rem;--radius-4xl:2rem;--shadow-2xs:0 1px #0000000d;--shadow-xs:0 1px 2px 0 #0000000d;--shadow-sm:0 1px 3px 0 #0000001a,0 1px 2px -1px #0000001a;--shadow-md:0 4px 6px -1px #0000001a,0 2px 4px -2px #0000001a;--shadow-lg:0 10px 15px -3px #0000001a,0 4px 6px -4px #0000001a;--shadow-xl:0 20px 25px -5px #0000001a,0 8px 10px -6px #0000001a;--shadow-2xl:0 25px 50px -12px #00000040;--inset-shadow-2xs:inset 0 1px #0000000d;--inset-shadow-xs:inset 0 1px 1px #0000000d;--inset-shadow-sm:inset 0 2px 4px #0000000d;--drop-shadow-xs:0 1px 1px #0000000d;--drop-shadow-sm:0 1px 2px #00000026;--drop-shadow-md:0 3px 3px #0000001f;--drop-shadow-lg:0 4px 4px #00000026;--drop-shadow-xl:0 9px 7px #0000001a;--drop-shadow-2xl:0 25px 25px #00000026;--ease-in:cubic-bezier(.4,0,1,1);--ease-out:cubic-bezier(0,0,.2,1);--ease-in-out:cubic-bezier(.4,0,.2,1);--animate-spin:spin 1s linear infinite;--animate-ping:ping 1s cubic-bezier(0,0,.2,1)infinite;--animate-pulse:pulse 2s cubic-bezier(.4,0,.6,1)infinite;--animate-bounce:bounce 1s infinite;--blur-xs:4px;--blur-sm:8px;--blur-md:12px;--blur-lg:16px;--blur-xl:24px;--blur-2xl:40px;--blur-3xl:64px;--perspective-dramatic:100px;--perspective-near:300px;--perspective-normal:500px;--perspective-midrange:800px;--perspective-distant:1200px;--aspect-video:16/9;--default-transition-duration:.15s;--default-transition-timing-function:cubic-bezier(.4,0,.2,1);--default-font-family:var(--font-sans);--default-font-feature-settings:var(--font-sans--font-feature-settings);--default-font-variation-settings:var(--font-sans--font-variation-settings);--default-mono-font-family:var(--font-mono);--default-mono-font-feature-settings:var(--font-mono--font-feature-settings);--default-mono-font-variation-settings:var(--font-mono--font-variation-settings);--font-alternative:"Montserrat",ui-sans-serif,system-ui,sans-serif,"Apple Color Emoji","Segoe UI Emoji","Segoe UI Symbol","Noto Color Emoji";--color-vortico:teal;--color-vortico-50:#00dcdc;--color-vortico-100:#00d2d2;--color-vortico-200:#00bdbd;--color-vortico-300:#00a9a9;--color-vortico-400:#009494;--color-vortico-500:teal;--color-vortico-600:#007171;--color-vortico-700:#006161;--color-vortico-800:#005252;--color-vortico-900:#004343;--color-vortico-950:#003b3b;--color-bruma:#00bbd5;--color-bruma-50:#ebfeff;--color-bruma-100:#cefcff;--color-bruma-200:#a2f7ff;--color-bruma-300:#63effd;--color-bruma-400:#1cdcf4;--color-bruma-500:#00bbd5;--color-bruma-600:#0398b7;--color-bruma-700:#0a7994;--color-bruma-800:#126178;--color-bruma-900:#145165;--color-bruma-950:#063646;--color-bosque:#9e744f;--color-bosque-50:#f7f4ef;--color-bosque-100:#ebe3d6;--color-bosque-200:#d9c8af;--color-bosque-300:#c2a682;--color-bosque-400:#b0895f;--color-bosque-500:#9e744f;--color-bosque-600:#8a5f44;--color-bosque-700:#6f4939;--color-bosque-800:#5f3f34;--color-bosque-900:#523731;--color-bosque-950:#2f1d19;--color-ciclon:#00976e;--color-ciclon-50:#ebfef6;--color-ciclon-100:#d0fbe7;--color-ciclon-200:#a4f6d3;--color-ciclon-300:#6aebbd;--color-ciclon-400:#2fd8a1;--color-ciclon-500:#0abf8a;--color-ciclon-600:#00976e;--color-ciclon-700:#007c5e;--color-ciclon-800:#03624b;--color-ciclon-900:#04503f;--color-ciclon-950:#012d25;--color-flama:#e25822;--color-flama-50:#fdf5ef;--color-flama-100:#fbe8d9;--color-flama-200:#f6cdb2;--color-flama-300:#f0ac81;--color-flama-400:#e9804e;--color-flama-500:#e25822;--color-flama-600:#d54821;--color-flama-700:#b1351d;--color-flama-800:#8d2d1f;--color-flama-900:#72271c;--color-flama-950:#3d110d;--color-primary:#656775;--color-primary-50:#f5f5f6;--color-primary-100:#e5e6e8;--color-primary-200:#cdced4;--color-primary-300:#aaabb6;--color-primary-400:#808290;--color-primary-500:#656775;--color-primary-600:#575863;--color-primary-700:#4a4a54;--color-primary-800:#3f3f46;--color-primary-900:#3a3a3f;--color-primary-950:#242428;--breakpoint-3xl:120rem;--container-8xl:96rem;--container-9xl:120rem}}@layer base{*,:after,:before,::backdrop{box-sizing:border-box;border:0 solid;margin:0;padding:0}::file-selector-button{box-sizing:border-box;border:0 solid;margin:0;padding:0}html,:host{-webkit-text-size-adjust:100%;tab-size:4;line-height:1.5;font-family:var(--default-font-family,ui-sans-serif,system-ui,sans-serif,"Apple Color Emoji","Segoe UI Emoji","Segoe UI Symbol","Noto Color Emoji");font-feature-settings:var(--default-font-feature-settings,normal);font-variation-settings:var(--default-font-variation-settings,normal);-webkit-tap-highlight-color:transparent}body{line-height:inherit}hr{height:0;color:inherit;border-top-width:1px}abbr:where([title]){-webkit-text-decoration:underline dotted;text-decoration:underline dotted}h1,h2,h3,h4,h5,h6{font-size:inherit;font-weight:inherit}a{color:inherit;-webkit-text-decoration:inherit;-webkit-text-decoration:inherit;-webkit-text-decoration:inherit;text-decoration:inherit}b,strong{font-weight:bolder}code,kbd,samp,pre{font-family:var(--default-mono-font-family,ui-monospace,SFMono-Regular,Menlo,Monaco,Consolas,"Liberation Mono","Courier New",monospace);font-feature-settings:var(--default-mono-font-feature-settings,normal);font-variation-settings:var(--default-mono-font-variation-settings,normal);font-size:1em}small{font-size:80%}sub,sup{vertical-align:baseline;font-size:75%;line-height:0;position:relative}sub{bottom:-.25em}sup{top:-.5em}table{text-indent:0;border-color:inherit;border-collapse:collapse}:-moz-focusring{outline:auto}progress{vertical-align:baseline}summary{display:list-item}ol,ul,menu{list-style:none}img,svg,video,canvas,audio,iframe,embed,object{vertical-align:middle;display:block}img,video{max-width:100%;height:auto}button,input,select,optgroup,textarea{font:inherit;font-feature-settings:inherit;font-variation-settings:inherit;letter-spacing:inherit;color:inherit;opacity:1;background-color:#0000;border-radius:0}::file-selector-button{font:inherit;font-feature-settings:inherit;font-variation-settings:inherit;letter-spacing:inherit;color:inherit;opacity:1;background-color:#0000;border-radius:0}:where(select:is([multiple],[size])) optgroup{font-weight:bolder}:where(select:is([multiple],[size])) optgroup option{padding-inline-start:20px}::file-selector-button{margin-inline-end:4px}::placeholder{opacity:1;color:color-mix(in oklab,currentColor 50%,transparent)}textarea{resize:vertical}::-webkit-search-decoration{-webkit-appearance:none}::-webkit-date-and-time-value{min-height:1lh;text-align:inherit}::-webkit-datetime-edit{display:inline-flex}::-webkit-datetime-edit-fields-wrapper{padding:0}::-webkit-datetime-edit{padding-block:0}::-webkit-datetime-edit-year-field{padding-block:0}::-webkit-datetime-edit-month-field{padding-block:0}::-webkit-datetime-edit-day-field{padding-block:0}::-webkit-datetime-edit-hour-field{padding-block:0}::-webkit-datetime-edit-minute-field{padding-block:0}::-webkit-datetime-edit-second-field{padding-block:0}::-webkit-datetime-edit-millisecond-field{padding-block:0}::-webkit-datetime-edit-meridiem-field{padding-block:0}:-moz-ui-invalid{box-shadow:none}button,input:where([type=button],[type=reset],[type=submit]){appearance:button}::file-selector-button{appearance:button}::-webkit-inner-spin-button{height:auto}::-webkit-outer-spin-button{height:auto}[hidden]:where(:not([hidden=until-found])){display:none!important}}@layer components;@layer utilities{.absolute{position:absolute}.fixed{position:fixed}.relative{position:relative}.inset-0{inset:calc(var(--spacing)*0)}.inset-x-0{inset-inline:calc(var(--spacing)*0)}.inset-x-4{inset-inline:calc(var(--spacing)*4)}.inset-y-\[5vh\]{inset-block:5vh}.top-0{top:calc(var(--spacing)*0)}.top-4{top:calc(var(--spacing)*4)}.right-4{right:calc(var(--spacing)*4)}.-bottom-2{bottom:calc(var(--spacing)*-2)}.left-1\/2{left:50%}.z-10{z-index:10}.z-20{z-index:20}.z-50{z-index:50}.mx-auto{margin-inline:auto}.my-14{margin-block:calc(var(--spacing)*14)}.-mt-px{margin-top:-1px}.mt-\[calc\(3\.5rem\+1px\)\]{margin-top:calc(3.5rem + 1px)}.mt-px{margin-top:1px}.mb-px{margin-bottom:1px}.-ml-px{margin-left:-1px}.ml-1{margin-left:calc(var(--spacing)*1)}.block{display:block}.flex{display:flex}.hidden{display:none}.inline{display:inline}.table{display:table}.h-2{height:calc(var(--spacing)*2)}.h-4{height:calc(var(--spacing)*4)}.h-5{height:calc(var(--spacing)*5)}.h-6{height:calc(var(--spacing)*6)}.h-7{height:calc(var(--spacing)*7)}.h-8{height:calc(var(--spacing)*8)}.h-9{height:calc(var(--spacing)*9)}.h-12{height:calc(var(--spacing)*12)}.h-14{height:calc(var(--spacing)*14)}.h-\[672px\]{height:672px}.h-\[calc\(100vh-6\.5rem-2px\)\]{height:calc(100vh - 6.5rem - 2px)}.h-fit{height:fit-content}.h-full{height:100%}.h-screen{height:100vh}.max-h-0{max-height:calc(var(--spacing)*0)}.max-h-\[32px\]{max-height:32px}.max-h-\[calc\(85vh\)\]{max-height:85vh}.max-h-\[calc\(85vh\+2rem\+3px\)\]{max-height:calc(85vh + 2rem + 3px)}.max-h-\[calc\(100\%-2rem-1px\)\]{max-height:calc(100% - 2rem - 1px)}.max-h-full{max-height:100%}.min-h-full{min-height:100%}.w-2{width:calc(var(--spacing)*2)}.w-4{width:calc(var(--spacing)*4)}.w-5{width:calc(var(--spacing)*5)}.w-6{width:calc(var(--spacing)*6)}.w-7{width:calc(var(--spacing)*7)}.w-8{width:calc(var(--spacing)*8)}.w-9{width:calc(var(--spacing)*9)}.w-40{width:calc(var(--spacing)*40)}.w-fit{width:fit-content}.w-full{width:100%}.w-screen{width:100vw}.max-w-7xl{max-width:var(--container-7xl)}.max-w-8xl{max-width:var(--container-8xl)}.flex-auto{flex:auto}.flex-none{flex:none}.table-fixed{table-layout:fixed}.border-collapse{border-collapse:collapse}.-translate-x-1\/2{--tw-translate-x:calc(calc(1/2*100%)*-1);translate:var(--tw-translate-x)var(--tw-translate-y)}.translate-y-full{--tw-translate-y:100%;translate:var(--tw-translate-x)var(--tw-translate-y)}.rotate-90{rotate:90deg}.cursor-pointer{cursor:pointer}.flex-col{flex-direction:column}.items-center{align-items:center}.items-start{align-items:flex-start}.justify-between{justify-content:space-between}.justify-center{justify-content:center}.justify-end{justify-content:flex-end}.justify-start{justify-content:flex-start}.gap-1{gap:calc(var(--spacing)*1)}.gap-2{gap:calc(var(--spacing)*2)}.gap-8{gap:calc(var(--spacing)*8)}.gap-x-2{column-gap:calc(var(--spacing)*2)}:where(.divide-x>:not(:last-child)){--tw-divide-x-reverse:0;border-inline-style:var(--tw-border-style);border-inline-start-width:calc(1px*var(--tw-divide-x-reverse));border-inline-end-width:calc(1px*calc(1 - var(--tw-divide-x-reverse)))}:where(.divide-y>:not(:last-child)){--tw-divide-y-reverse:0;border-bottom-style:var(--tw-border-style);border-top-style:var(--tw-border-style);border-top-width:calc(1px*var(--tw-divide-y-reverse));border-bottom-width:calc(1px*calc(1 - var(--tw-divide-y-reverse)))}:where(.divide-flama-500\/50>:not(:last-child)){border-color:color-mix(in oklab,var(--color-flama-500)50%,transparent)}:where(.divide-vortico-500\/50>:not(:last-child)){border-color:color-mix(in oklab,var(--color-vortico-500)50%,transparent)}.truncate{text-overflow:ellipsis;white-space:nowrap;overflow:hidden}.overflow-auto{overflow:auto}.overflow-hidden{overflow:hidden}.overflow-visible{overflow:visible}.overflow-y-scroll{overflow-y:scroll}.rounded{border-radius:.25rem}.rounded-lg{border-radius:var(--radius-lg)}.border{border-style:var(--tw-border-style);border-width:1px}.border-t{border-top-style:var(--tw-border-style);border-top-width:1px}.border-r{border-right-style:var(--tw-border-style);border-right-width:1px}.border-b{border-bottom-style:var(--tw-border-style);border-bottom-width:1px}.border-l-4{border-left-style:var(--tw-border-style);border-left-width:4px}.border-current{border-color:currentColor}.border-flama-500{border-color:var(--color-flama-500)}.border-flama-500\/50{border-color:color-mix(in oklab,var(--color-flama-500)50%,transparent)}.border-primary-100{border-color:var(--color-primary-100)}.border-primary-300{border-color:var(--color-primary-300)}.border-primary-500{border-color:var(--color-primary-500)}.border-transparent{border-color:#0000}.border-vortico-500\/50{border-color:color-mix(in oklab,var(--color-vortico-500)50%,transparent)}.bg-flama-500\/25{background-color:color-mix(in oklab,var(--color-flama-500)25%,transparent)}.bg-flama-700{background-color:var(--color-flama-700)}.bg-primary-100{background-color:var(--color-primary-100)}.bg-primary-200{background-color:var(--color-primary-200)}.bg-primary-500\/30{background-color:color-mix(in oklab,var(--color-primary-500)30%,transparent)}.bg-primary-800{background-color:var(--color-primary-800)}.bg-primary-900{background-color:var(--color-primary-900)}.bg-primary-950\/30{background-color:color-mix(in oklab,var(--color-primary-950)30%,transparent)}.bg-linear-to-b{--tw-gradient-position:to bottom in oklab;background-image:linear-gradient(var(--tw-gradient-stops))}.from-flama-500\/10{--tw-gradient-from:color-mix(in oklab,var(--color-flama-500)10%,transparent);--tw-gradient-stops:var(--tw-gradient-via-stops,var(--tw-gradient-position),var(--tw-gradient-from)var(--tw-gradient-from-position),var(--tw-gradient-to)var(--tw-gradient-to-position))}.p-2{padding:calc(var(--spacing)*2)}.p-3{padding:calc(var(--spacing)*3)}.px-2{padding-inline:calc(var(--spacing)*2)}.px-4{padding-inline:calc(var(--spacing)*4)}.py-1{padding-block:calc(var(--spacing)*1)}.py-2{padding-block:calc(var(--spacing)*2)}.py-6{padding-block:calc(var(--spacing)*6)}.py-8{padding-block:calc(var(--spacing)*8)}.pr-3{padding-right:calc(var(--spacing)*3)}.pl-1{padding-left:calc(var(--spacing)*1)}.pl-2{padding-left:calc(var(--spacing)*2)}.pl-4{padding-left:calc(var(--spacing)*4)}.text-left{text-align:left}.text-right{text-align:right}.font-alternative{font-family:var(--font-alternative)}.font-mono{font-family:var(--font-mono)}.text-2xl{font-size:var(--text-2xl);line-height:var(--tw-leading,var(--text-2xl--line-height))}.text-base{font-size:var(--text-base);line-height:var(--tw-leading,var(--text-base--line-height))}.text-lg{font-size:var(--text-lg);line-height:var(--tw-leading,var(--text-lg--line-height))}.text-sm{font-size:var(--text-sm);line-height:var(--tw-leading,var(--text-sm--line-height))}.text-xl{font-size:var(--text-xl);line-height:var(--tw-leading,var(--text-xl--line-height))}.text-xs{font-size:var(--text-xs);line-height:var(--tw-leading,var(--text-xs--line-height))}.leading-4{--tw-leading:calc(var(--spacing)*4);line-height:calc(var(--spacing)*4)}.leading-5{--tw-leading:calc(var(--spacing)*5);line-height:calc(var(--spacing)*5)}.leading-6{--tw-leading:calc(var(--spacing)*6);line-height:calc(var(--spacing)*6)}.leading-7{--tw-leading:calc(var(--spacing)*7);line-height:calc(var(--spacing)*7)}.leading-8{--tw-leading:calc(var(--spacing)*8);line-height:calc(var(--spacing)*8)}.leading-9{--tw-leading:calc(var(--spacing)*9);line-height:calc(var(--spacing)*9)}.font-bold{--tw-font-weight:var(--font-weight-bold);font-weight:var(--font-weight-bold)}.font-medium{--tw-font-weight:var(--font-weight-medium);font-weight:var(--font-weight-medium)}.font-semibold{--tw-font-weight:var(--font-weight-semibold);font-weight:var(--font-weight-semibold)}.text-wrap{text-wrap:wrap}.whitespace-pre{white-space:pre}.text-bosque-200{color:var(--color-bosque-200)}.text-bosque-500{color:var(--color-bosque-500)}.text-bosque-800{color:var(--color-bosque-800)}.text-bruma-200{color:var(--color-bruma-200)}.text-bruma-500{color:var(--color-bruma-500)}.text-bruma-800{color:var(--color-bruma-800)}.text-ciclon-200{color:var(--color-ciclon-200)}.text-ciclon-500{color:var(--color-ciclon-500)}.text-ciclon-800{color:var(--color-ciclon-800)}.text-flama{color:var(--color-flama)}.text-flama-200{color:var(--color-flama-200)}.text-flama-400{color:var(--color-flama-400)}.text-flama-500{color:var(--color-flama-500)}.text-flama-800{color:var(--color-flama-800)}.text-primary-200{color:var(--color-primary-200)}.text-primary-300{color:var(--color-primary-300)}.text-primary-400{color:var(--color-primary-400)}.text-primary-500{color:var(--color-primary-500)}.text-primary-600{color:var(--color-primary-600)}.text-primary-700{color:var(--color-primary-700)}.text-primary-800{color:var(--color-primary-800)}.text-vortico-200{color:var(--color-vortico-200)}.text-vortico-500{color:var(--color-vortico-500)}.text-vortico-800{color:var(--color-vortico-800)}.capitalize{text-transform:capitalize}.italic{font-style:italic}.opacity-0{opacity:0}.opacity-100{opacity:1}.ring-1{--tw-ring-shadow:var(--tw-ring-inset,)0 0 0 calc(1px + var(--tw-ring-offset-width))var(--tw-ring-color,currentColor);box-shadow:var(--tw-inset-shadow),var(--tw-inset-ring-shadow),var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow)}.shadow-md{--tw-shadow:0 4px 6px -1px var(--tw-shadow-color,#0000001a),0 2px 4px -2px var(--tw-shadow-color,#0000001a);box-shadow:var(--tw-inset-shadow),var(--tw-inset-ring-shadow),var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow)}.shadow-primary-300{--tw-shadow-color:var(--color-primary-300)}.ring-primary-500\/50{--tw-ring-color:color-mix(in oklab,var(--color-primary-500)50%,transparent)}.filter{filter:var(--tw-blur,)var(--tw-brightness,)var(--tw-contrast,)var(--tw-grayscale,)var(--tw-hue-rotate,)var(--tw-invert,)var(--tw-saturate,)var(--tw-sepia,)var(--tw-drop-shadow,)}.backdrop-blur{--tw-backdrop-blur:blur(8px);-webkit-backdrop-filter:var(--tw-backdrop-blur,)var(--tw-backdrop-brightness,)var(--tw-backdrop-contrast,)var(--tw-backdrop-grayscale,)var(--tw-backdrop-hue-rotate,)var(--tw-backdrop-invert,)var(--tw-backdrop-opacity,)var(--tw-backdrop-saturate,)var(--tw-backdrop-sepia,);backdrop-filter:var(--tw-backdrop-blur,)var(--tw-backdrop-brightness,)var(--tw-backdrop-contrast,)var(--tw-backdrop-grayscale,)var(--tw-backdrop-hue-rotate,)var(--tw-backdrop-invert,)var(--tw-backdrop-opacity,)var(--tw-backdrop-saturate,)var(--tw-backdrop-sepia,)}.backdrop-blur-sm{--tw-backdrop-blur:blur(var(--blur-sm));-webkit-backdrop-filter:var(--tw-backdrop-blur,)var(--tw-backdrop-brightness,)var(--tw-backdrop-contrast,)var(--tw-backdrop-grayscale,)var(--tw-backdrop-hue-rotate,)var(--tw-backdrop-invert,)var(--tw-backdrop-opacity,)var(--tw-backdrop-saturate,)var(--tw-backdrop-sepia,);backdrop-filter:var(--tw-backdrop-blur,)var(--tw-backdrop-brightness,)var(--tw-backdrop-contrast,)var(--tw-backdrop-grayscale,)var(--tw-backdrop-hue-rotate,)var(--tw-backdrop-invert,)var(--tw-backdrop-opacity,)var(--tw-backdrop-saturate,)var(--tw-backdrop-sepia,)}.transition{transition-property:color,background-color,border-color,outline-color,text-decoration-color,fill,stroke,--tw-gradient-from,--tw-gradient-via,--tw-gradient-to,opacity,box-shadow,transform,translate,scale,rotate,filter,-webkit-backdrop-filter,backdrop-filter;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.transition-all{transition-property:all;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.transition-colors{transition-property:color,background-color,border-color,outline-color,text-decoration-color,fill,stroke,--tw-gradient-from,--tw-gradient-via,--tw-gradient-to;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.transition-opacity{transition-property:opacity;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.transition-transform{transition-property:transform,translate,scale,rotate;transition-timing-function:var(--tw-ease,var(--default-transition-timing-function));transition-duration:var(--tw-duration,var(--default-transition-duration))}.duration-200{--tw-duration:.2s;transition-duration:.2s}.duration-500{--tw-duration:.5s;transition-duration:.5s}.ring-inset{--tw-ring-inset:inset}.select-none{-webkit-user-select:none;user-select:none}@media (hover:hover){.group-hover\:block:is(:where(.group):hover *){display:block}.group-hover\:opacity-100:is(:where(.group):hover *){opacity:1}}.after\:absolute:after{content:var(--tw-content);position:absolute}.after\:bottom-\[100\%\]:after{content:var(--tw-content);bottom:100%}.after\:left-1\/2:after{content:var(--tw-content);left:50%}.after\:-translate-x-1\/2:after{content:var(--tw-content);--tw-translate-x:calc(calc(1/2*100%)*-1);translate:var(--tw-translate-x)var(--tw-translate-y)}.after\:border-8:after{content:var(--tw-content);border-style:var(--tw-border-style);border-width:8px}.after\:border-x-transparent:after{content:var(--tw-content);border-inline-color:#0000}.after\:border-t-transparent:after{content:var(--tw-content);border-top-color:#0000}.after\:border-b-primary-800:after{content:var(--tw-content);border-bottom-color:var(--color-primary-800)}.after\:content-\[\'\'\]:after{content:var(--tw-content);--tw-content:"";content:var(--tw-content)}.first\:pl-0:first-child{padding-left:calc(var(--spacing)*0)}.last\:pr-0:last-child{padding-right:calc(var(--spacing)*0)}@media (hover:hover){.hover\:border-l-4:hover{border-left-style:var(--tw-border-style);border-left-width:4px}.hover\:border-flama-500:hover{border-color:var(--color-flama-500)}.hover\:bg-flama-300:hover{background-color:var(--color-flama-300)}.hover\:pl-0:hover{padding-left:calc(var(--spacing)*0)}.hover\:text-bosque:hover{color:var(--color-bosque)}.hover\:text-ciclon:hover{color:var(--color-ciclon)}.hover\:text-flama:hover{color:var(--color-flama)}.hover\:text-flama-500:hover{color:var(--color-flama-500)}.hover\:text-primary-800:hover{color:var(--color-primary-800)}.hover\:text-vortico-500:hover{color:var(--color-vortico-500)}.hover\:shadow-lg:hover{--tw-shadow:0 10px 15px -3px var(--tw-shadow-color,#0000001a),0 4px 6px -4px var(--tw-shadow-color,#0000001a);box-shadow:var(--tw-inset-shadow),var(--tw-inset-ring-shadow),var(--tw-ring-offset-shadow),var(--tw-ring-shadow),var(--tw-shadow)}.hover\:shadow-flama-300:hover{--tw-shadow-color:var(--color-flama-300)}}@media (width>=40rem){.sm\:inset-x-6{inset-inline:calc(var(--spacing)*6)}.sm\:px-6{padding-inline:calc(var(--spacing)*6)}.sm\:py-10{padding-block:calc(var(--spacing)*10)}}@media (width>=48rem){.md\:sticky{position:sticky}.md\:inset-x-8{inset-inline:calc(var(--spacing)*8)}.md\:left-0{left:calc(var(--spacing)*0)}.md\:block{display:block}.md\:flex{display:flex}.md\:hidden{display:none}.md\:h-5{height:calc(var(--spacing)*5)}.md\:h-6{height:calc(var(--spacing)*6)}.md\:h-7{height:calc(var(--spacing)*7)}.md\:h-8{height:calc(var(--spacing)*8)}.md\:h-9{height:calc(var(--spacing)*9)}.md\:h-10{height:calc(var(--spacing)*10)}.md\:w-5{width:calc(var(--spacing)*5)}.md\:w-6{width:calc(var(--spacing)*6)}.md\:w-7{width:calc(var(--spacing)*7)}.md\:w-8{width:calc(var(--spacing)*8)}.md\:w-9{width:calc(var(--spacing)*9)}.md\:w-10{width:calc(var(--spacing)*10)}.md\:px-8{padding-inline:calc(var(--spacing)*8)}.md\:py-12{padding-block:calc(var(--spacing)*12)}.md\:text-2xl{font-size:var(--text-2xl);line-height:var(--tw-leading,var(--text-2xl--line-height))}.md\:text-3xl{font-size:var(--text-3xl);line-height:var(--tw-leading,var(--text-3xl--line-height))}.md\:text-base{font-size:var(--text-base);line-height:var(--tw-leading,var(--text-base--line-height))}.md\:text-lg{font-size:var(--text-lg);line-height:var(--tw-leading,var(--text-lg--line-height))}.md\:text-sm{font-size:var(--text-sm);line-height:var(--tw-leading,var(--text-sm--line-height))}.md\:text-xl{font-size:var(--text-xl);line-height:var(--tw-leading,var(--text-xl--line-height))}.md\:leading-5{--tw-leading:calc(var(--spacing)*5);line-height:calc(var(--spacing)*5)}.md\:leading-6{--tw-leading:calc(var(--spacing)*6);line-height:calc(var(--spacing)*6)}.md\:leading-7{--tw-leading:calc(var(--spacing)*7);line-height:calc(var(--spacing)*7)}.md\:leading-8{--tw-leading:calc(var(--spacing)*8);line-height:calc(var(--spacing)*8)}.md\:leading-9{--tw-leading:calc(var(--spacing)*9);line-height:calc(var(--spacing)*9)}.md\:leading-10{--tw-leading:calc(var(--spacing)*10);line-height:calc(var(--spacing)*10)}}@media (width>=64rem){.lg\:w-xs{width:var(--container-xs)}.lg\:flex-row{flex-direction:row}}}:root{background-color:var(--color-primary-100);color:var(--color-primary-600);scrollbar-color:var(--color-primary-400)transparent;z-index:0}body{height:100%}@keyframes spin{to{transform:rotate(360deg)}}@keyframes ping{75%,to{opacity:0;transform:scale(2)}}@keyframes pulse{50%{opacity:.5}}@keyframes bounce{0%,to{animation-timing-function:cubic-bezier(.8,0,1,1);transform:translateY(-25%)}50%{animation-timing-function:cubic-bezier(0,0,.2,1);transform:none}}@property --tw-translate-x{syntax:"*";inherits:false;initial-value:0}@property --tw-translate-y{syntax:"*";inherits:false;initial-value:0}@property --tw-translate-z{syntax:"*";inherits:false;initial-value:0}@property --tw-divide-x-reverse{syntax:"*";inherits:false;initial-value:0}@property --tw-border-style{syntax:"*";inherits:false;initial-value:solid}@property --tw-divide-y-reverse{syntax:"*";inherits:false;initial-value:0}@property --tw-gradient-position{syntax:"*";inherits:false}@property --tw-gradient-from{syntax:"<color>";inherits:false;initial-value:#0000}@property --tw-gradient-via{syntax:"<color>";inherits:false;initial-value:#0000}@property --tw-gradient-to{syntax:"<color>";inherits:false;initial-value:#0000}@property --tw-gradient-stops{syntax:"*";inherits:false}@property --tw-gradient-via-stops{syntax:"*";inherits:false}@property --tw-gradient-from-position{syntax:"<length-percentage>";inherits:false;initial-value:0%}@property --tw-gradient-via-position{syntax:"<length-percentage>";inherits:false;initial-value:50%}@property --tw-gradient-to-position{syntax:"<length-percentage>";inherits:false;initial-value:100%}@property --tw-leading{syntax:"*";inherits:false}@property --tw-font-weight{syntax:"*";inherits:false}@property --tw-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-shadow-color{syntax:"*";inherits:false}@property --tw-inset-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-inset-shadow-color{syntax:"*";inherits:false}@property --tw-ring-color{syntax:"*";inherits:false}@property --tw-ring-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-inset-ring-color{syntax:"*";inherits:false}@property --tw-inset-ring-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-ring-inset{syntax:"*";inherits:false}@property --tw-ring-offset-width{syntax:"<length>";inherits:false;initial-value:0}@property --tw-ring-offset-color{syntax:"*";inherits:false;initial-value:#fff}@property --tw-ring-offset-shadow{syntax:"*";inherits:false;initial-value:0 0 #0000}@property --tw-blur{syntax:"*";inherits:false}@property --tw-brightness{syntax:"*";inherits:false}@property --tw-contrast{syntax:"*";inherits:false}@property --tw-grayscale{syntax:"*";inherits:false}@property --tw-hue-rotate{syntax:"*";inherits:false}@property --tw-invert{syntax:"*";inherits:false}@property --tw-opacity{syntax:"*";inherits:false}@property --tw-saturate{syntax:"*";inherits:false}@property --tw-sepia{syntax:"*";inherits:false}@property --tw-drop-shadow{syntax:"*";inherits:false}@property --tw-backdrop-blur{syntax:"*";inherits:false}@property --tw-backdrop-brightness{syntax:"*";inherits:false}@property --tw-backdrop-contrast{syntax:"*";inherits:false}@property --tw-backdrop-grayscale{syntax:"*";inherits:false}@property --tw-backdrop-hue-rotate{syntax:"*";inherits:false}@property --tw-backdrop-invert{syntax:"*";inherits:false}@property --tw-backdrop-opacity{syntax:"*";inherits:false}@property --tw-backdrop-saturate{syntax:"*";inherits:false}@property --tw-backdrop-sepia{syntax:"*";inherits:false}@property --tw-duration{syntax:"*";inherits:false}@property --tw-content{syntax:"*";inherits:false;initial-value:""}
</style></head><body><div id="app"/><script>(()=>{"use strict";var e={34:(e,t,n)=>{var r=n(4901);e.exports=function(e){return"object"==typeof e?null!==e:r(e)}},81:(e,t,n)=>{var r=n(9565),a=n(9306),o=n(8551),i=n(6823),l=n(851),s=TypeError;e.exports=function(e,t){var n=arguments.length<2?l(e):t;if(a(n))return o(r(n,e));throw new s(i(e)+" is not iterable")}},230:(e,t,n)=>{var r=n(6518),a=n(9306),o=n(6194),i=n(6223),l=TypeError;r({target:"Map",proto:!0,real:!0,forced:!0},{reduce:function(e){var t=o(this),n=arguments.length<2,r=n?void 0:arguments[1];if(a(e),i(t,(function(a,o){n?(n=!1,r=a):r=e(r,a,o,t)})),n)throw new l("Reduce of empty map with no initial value");return r}})},283:(e,t,n)=>{var r=n(9504),a=n(9039),o=n(4901),i=n(9297),l=n(3724),s=n(350).CONFIGURABLE,u=n(3706),c=n(1181),p=c.enforce,d=c.get,f=String,h=Object.defineProperty,g=r("".slice),m=r("".replace),y=r([].join),b=l&&!a((function(){return 8!==h((function(){}),"length",{value:8}).length})),v=String(String).split("String"),w=e.exports=function(e,t,n){"Symbol("===g(f(t),0,7)&&(t="["+m(f(t),/^Symbol\(([^)]*)\).*$/,"$1")+"]"),n&&n.getter&&(t="get "+t),n&&n.setter&&(t="set "+t),(!i(e,"name")||s&&e.name!==t)&&(l?h(e,"name",{value:t,configurable:!0}):e.name=t),b&&n&&i(n,"arity")&&e.length!==n.arity&&h(e,"length",{value:n.arity});try{n&&i(n,"constructor")&&n.constructor?l&&h(e,"prototype",{writable:!1}):e.prototype&&(e.prototype=void 0)}catch(e){}var r=p(e);return i(r,"source")||(r.source=y(v,"string"==typeof t?t:"")),e};Function.prototype.toString=w((function(){return o(this)&&d(this).source||u(this)}),"toString")},350:(e,t,n)=>{var r=n(3724),a=n(9297),o=Function.prototype,i=r&&Object.getOwnPropertyDescriptor,l=a(o,"name"),s=l&&"something"===function(){}.name,u=l&&(!r||r&&i(o,"name").configurable);e.exports={EXISTS:l,PROPER:s,CONFIGURABLE:u}},421:e=>{e.exports={}},507:(e,t,n)=>{var r=n(9565);e.exports=function(e,t,n){for(var a,o,i=n?e:e.iterator,l=e.next;!(a=r(l,i)).done;)if(void 0!==(o=t(a.value)))return o}},616:(e,t,n)=>{var r=n(9039);e.exports=!r((function(){var e=function(){}.bind();return"function"!=typeof e||e.hasOwnProperty("prototype")}))},741:e=>{var t=Math.ceil,n=Math.floor;e.exports=Math.trunc||function(e){var r=+e;return(r>0?n:t)(r)}},757:(e,t,n)=>{var r=n(7751),a=n(4901),o=n(1625),i=n(7040),l=Object;e.exports=i?function(e){return"symbol"==typeof e}:function(e){var t=r("Symbol");return a(t)&&o(t.prototype,l(e))}},851:(e,t,n)=>{var r=n(6955),a=n(5966),o=n(4117),i=n(6269),l=n(8227)("iterator");e.exports=function(e){if(!o(e))return a(e,l)||a(e,"@@iterator")||i[r(e)]}},961:(e,t,n)=>{!function e(){if("undefined"!=typeof __REACT_DEVTOOLS_GLOBAL_HOOK__&&"function"==typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE)try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(e){console.error(e)}}(),e.exports=n(6221)},1181:(e,t,n)=>{var r,a,o,i=n(8622),l=n(4576),s=n(34),u=n(6699),c=n(9297),p=n(7629),d=n(6119),f=n(421),h="Object already initialized",g=l.TypeError,m=l.WeakMap;if(i||p.state){var y=p.state||(p.state=new m);y.get=y.get,y.has=y.has,y.set=y.set,r=function(e,t){if(y.has(e))throw new g(h);return t.facade=e,y.set(e,t),t},a=function(e){return y.get(e)||{}},o=function(e){return y.has(e)}}else{var b=d("state");f[b]=!0,r=function(e,t){if(c(e,b))throw new g(h);return t.facade=e,u(e,b,t),t},a=function(e){return c(e,b)?e[b]:{}},o=function(e){return c(e,b)}}e.exports={set:r,get:a,has:o,enforce:function(e){return o(e)?a(e):r(e,{})},getterFor:function(e){return function(t){var n;if(!s(t)||(n=a(t)).type!==e)throw new g("Incompatible receiver, "+e+" required");return n}}}},1247:(e,t,n)=>{var r=n(9982),a=n(6540),o=n(961);function i(e){var t="https://react.dev/errors/"+e;if(1<arguments.length){t+="?args[]="+encodeURIComponent(arguments[1]);for(var n=2;n<arguments.length;n++)t+="&args[]="+encodeURIComponent(arguments[n])}return"Minified React error #"+e+"; visit "+t+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function l(e){return!(!e||1!==e.nodeType&&9!==e.nodeType&&11!==e.nodeType)}var s=Symbol.for("react.element"),u=Symbol.for("react.transitional.element"),c=Symbol.for("react.portal"),p=Symbol.for("react.fragment"),d=Symbol.for("react.strict_mode"),f=Symbol.for("react.profiler"),h=Symbol.for("react.provider"),g=Symbol.for("react.consumer"),m=Symbol.for("react.context"),y=Symbol.for("react.forward_ref"),b=Symbol.for("react.suspense"),v=Symbol.for("react.suspense_list"),w=Symbol.for("react.memo"),x=Symbol.for("react.lazy");Symbol.for("react.scope"),Symbol.for("react.debug_trace_mode");var k=Symbol.for("react.offscreen");Symbol.for("react.legacy_hidden"),Symbol.for("react.tracing_marker");var S=Symbol.for("react.memo_cache_sentinel"),C=Symbol.iterator;function _(e){return null===e||"object"!=typeof e?null:"function"==typeof(e=C&&e[C]||e["@@iterator"])?e:null}var E=Symbol.for("react.client.reference");function N(e){if(null==e)return null;if("function"==typeof e)return e.$$typeof===E?null:e.displayName||e.name||null;if("string"==typeof e)return e;switch(e){case p:return"Fragment";case c:return"Portal";case f:return"Profiler";case d:return"StrictMode";case b:return"Suspense";case v:return"SuspenseList"}if("object"==typeof e)switch(e.$$typeof){case m:return(e.displayName||"Context")+".Provider";case g:return(e._context.displayName||"Context")+".Consumer";case y:var t=e.render;return(e=e.displayName)||(e=""!==(e=t.displayName||t.name||"")?"ForwardRef("+e+")":"ForwardRef"),e;case w:return null!==(t=e.displayName||null)?t:N(e.type)||"Memo";case x:t=e._payload,e=e._init;try{return N(e(t))}catch(e){}}return null}var P,A,L=a.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,T=Object.assign;function R(e){if(void 0===P)try{throw Error()}catch(e){var t=e.stack.trim().match(/\n( *(at )?)/);P=t&&t[1]||"",A=-1<e.stack.indexOf("\n    at")?" (<anonymous>)":-1<e.stack.indexOf("@")?"@unknown:0:0":""}return"\n"+P+e+A}var I=!1;function M(e,t){if(!e||I)return"";I=!0;var n=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{var r={DetermineComponentFrameRoot:function(){try{if(t){var n=function(){throw Error()};if(Object.defineProperty(n.prototype,"props",{set:function(){throw Error()}}),"object"==typeof Reflect&&Reflect.construct){try{Reflect.construct(n,[])}catch(e){var r=e}Reflect.construct(e,[],n)}else{try{n.call()}catch(e){r=e}e.call(n.prototype)}}else{try{throw Error()}catch(e){r=e}(n=e())&&"function"==typeof n.catch&&n.catch((function(){}))}}catch(e){if(e&&r&&"string"==typeof e.stack)return[e.stack,r.stack]}return[null,null]}};r.DetermineComponentFrameRoot.displayName="DetermineComponentFrameRoot";var a=Object.getOwnPropertyDescriptor(r.DetermineComponentFrameRoot,"name");a&&a.configurable&&Object.defineProperty(r.DetermineComponentFrameRoot,"name",{value:"DetermineComponentFrameRoot"});var o=r.DetermineComponentFrameRoot(),i=o[0],l=o[1];if(i&&l){var s=i.split("\n"),u=l.split("\n");for(a=r=0;r<s.length&&!s[r].includes("DetermineComponentFrameRoot");)r++;for(;a<u.length&&!u[a].includes("DetermineComponentFrameRoot");)a++;if(r===s.length||a===u.length)for(r=s.length-1,a=u.length-1;1<=r&&0<=a&&s[r]!==u[a];)a--;for(;1<=r&&0<=a;r--,a--)if(s[r]!==u[a]){if(1!==r||1!==a)do{if(r--,0>--a||s[r]!==u[a]){var c="\n"+s[r].replace(" at new "," at ");return e.displayName&&c.includes("<anonymous>")&&(c=c.replace("<anonymous>",e.displayName)),c}}while(1<=r&&0<=a);break}}}finally{I=!1,Error.prepareStackTrace=n}return(n=e?e.displayName||e.name:"")?R(n):""}function O(e){switch(e.tag){case 26:case 27:case 5:return R(e.type);case 16:return R("Lazy");case 13:return R("Suspense");case 19:return R("SuspenseList");case 0:case 15:return M(e.type,!1);case 11:return M(e.type.render,!1);case 1:return M(e.type,!0);default:return""}}function j(e){try{var t="";do{t+=O(e),e=e.return}while(e);return t}catch(e){return"\nError generating stack: "+e.message+"\n"+e.stack}}function z(e){var t=e,n=e;if(e.alternate)for(;t.return;)t=t.return;else{e=t;do{!!(4098&(t=e).flags)&&(n=t.return),e=t.return}while(e)}return 3===t.tag?n:null}function $(e){if(13===e.tag){var t=e.memoizedState;if(null===t&&null!==(e=e.alternate)&&(t=e.memoizedState),null!==t)return t.dehydrated}return null}function F(e){if(z(e)!==e)throw Error(i(188))}function D(e){var t=e.tag;if(5===t||26===t||27===t||6===t)return e;for(e=e.child;null!==e;){if(null!==(t=D(e)))return t;e=e.sibling}return null}var B=Array.isArray,U=o.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,q={pending:!1,data:null,method:null,action:null},G=[],W=-1;function H(e){return{current:e}}function V(e){0>W||(e.current=G[W],G[W]=null,W--)}function Q(e,t){W++,G[W]=e.current,e.current=t}var X=H(null),K=H(null),Y=H(null),Z=H(null);function J(e,t){switch(Q(Y,t),Q(K,e),Q(X,null),e=t.nodeType){case 9:case 11:t=(t=t.documentElement)&&(t=t.namespaceURI)?Yc(t):0;break;default:if(t=(e=8===e?t.parentNode:t).tagName,e=e.namespaceURI)t=Zc(e=Yc(e),t);else switch(t){case"svg":t=1;break;case"math":t=2;break;default:t=0}}V(X),Q(X,t)}function ee(){V(X),V(K),V(Y)}function te(e){null!==e.memoizedState&&Q(Z,e);var t=X.current,n=Zc(t,e.type);t!==n&&(Q(K,e),Q(X,n))}function ne(e){K.current===e&&(V(X),V(K)),Z.current===e&&(V(Z),zp._currentValue=q)}var re=Object.prototype.hasOwnProperty,ae=r.unstable_scheduleCallback,oe=r.unstable_cancelCallback,ie=r.unstable_shouldYield,le=r.unstable_requestPaint,se=r.unstable_now,ue=r.unstable_getCurrentPriorityLevel,ce=r.unstable_ImmediatePriority,pe=r.unstable_UserBlockingPriority,de=r.unstable_NormalPriority,fe=r.unstable_LowPriority,he=r.unstable_IdlePriority,ge=r.log,me=r.unstable_setDisableYieldValue,ye=null,be=null;function ve(e){if("function"==typeof ge&&me(e),be&&"function"==typeof be.setStrictMode)try{be.setStrictMode(ye,e)}catch(e){}}var we=Math.clz32?Math.clz32:function(e){return 0==(e>>>=0)?32:31-(xe(e)/ke|0)|0},xe=Math.log,ke=Math.LN2,Se=128,Ce=4194304;function _e(e){var t=42&e;if(0!==t)return t;switch(e&-e){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:return 64;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return 4194176&e;case 4194304:case 8388608:case 16777216:case 33554432:return 62914560&e;case 67108864:return 67108864;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 0;default:return e}}function Ee(e,t){var n=e.pendingLanes;if(0===n)return 0;var r=0,a=e.suspendedLanes,o=e.pingedLanes,i=e.warmLanes;e=0!==e.finishedLanes;var l=134217727&n;return 0!==l?0!=(n=l&~a)?r=_e(n):0!=(o&=l)?r=_e(o):e||0!=(i=l&~i)&&(r=_e(i)):0!=(l=n&~a)?r=_e(l):0!==o?r=_e(o):e||0!=(i=n&~i)&&(r=_e(i)),0===r?0:0===t||t===r||t&a||!((a=r&-r)>=(i=t&-t)||32===a&&4194176&i)?r:t}function Ne(e,t){return!(e.pendingLanes&~(e.suspendedLanes&~e.pingedLanes)&t)}function Pe(e,t){switch(e){case 1:case 2:case 4:case 8:return t+250;case 16:case 32:case 64:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return t+5e3;default:return-1}}function Ae(){var e=Se;return!(4194176&(Se<<=1))&&(Se=128),e}function Le(){var e=Ce;return!(62914560&(Ce<<=1))&&(Ce=4194304),e}function Te(e){for(var t=[],n=0;31>n;n++)t.push(e);return t}function Re(e,t){e.pendingLanes|=t,268435456!==t&&(e.suspendedLanes=0,e.pingedLanes=0,e.warmLanes=0)}function Ie(e,t,n){e.pendingLanes|=t,e.suspendedLanes&=~t;var r=31-we(t);e.entangledLanes|=t,e.entanglements[r]=1073741824|e.entanglements[r]|4194218&n}function Me(e,t){var n=e.entangledLanes|=t;for(e=e.entanglements;n;){var r=31-we(n),a=1<<r;a&t|e[r]&t&&(e[r]|=t),n&=~a}}function Oe(e){return 2<(e&=-e)?8<e?134217727&e?32:268435456:8:2}function je(){var e=U.p;return 0!==e?e:void 0===(e=window.event)?32:Zp(e.type)}var ze=Math.random().toString(36).slice(2),$e="__reactFiber$"+ze,Fe="__reactProps$"+ze,De="__reactContainer$"+ze,Be="__reactEvents$"+ze,Ue="__reactListeners$"+ze,qe="__reactHandles$"+ze,Ge="__reactResources$"+ze,We="__reactMarker$"+ze;function He(e){delete e[$e],delete e[Fe],delete e[Be],delete e[Ue],delete e[qe]}function Ve(e){var t=e[$e];if(t)return t;for(var n=e.parentNode;n;){if(t=n[De]||n[$e]){if(n=t.alternate,null!==t.child||null!==n&&null!==n.child)for(e=up(e);null!==e;){if(n=e[$e])return n;e=up(e)}return t}n=(e=n).parentNode}return null}function Qe(e){if(e=e[$e]||e[De]){var t=e.tag;if(5===t||6===t||13===t||26===t||27===t||3===t)return e}return null}function Xe(e){var t=e.tag;if(5===t||26===t||27===t||6===t)return e.stateNode;throw Error(i(33))}function Ke(e){var t=e[Ge];return t||(t=e[Ge]={hoistableStyles:new Map,hoistableScripts:new Map}),t}function Ye(e){e[We]=!0}var Ze=new Set,Je={};function et(e,t){tt(e,t),tt(e+"Capture",t)}function tt(e,t){for(Je[e]=t,e=0;e<t.length;e++)Ze.add(t[e])}var nt=!("undefined"==typeof window||void 0===window.document||void 0===window.document.createElement),rt=RegExp("^[:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD][:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD\\-.0-9\\u00B7\\u0300-\\u036F\\u203F-\\u2040]*$"),at={},ot={};function it(e,t,n){if(a=t,re.call(ot,a)||!re.call(at,a)&&(rt.test(a)?ot[a]=!0:(at[a]=!0,0)))if(null===n)e.removeAttribute(t);else{switch(typeof n){case"undefined":case"function":case"symbol":return void e.removeAttribute(t);case"boolean":var r=t.toLowerCase().slice(0,5);if("data-"!==r&&"aria-"!==r)return void e.removeAttribute(t)}e.setAttribute(t,""+n)}var a}function lt(e,t,n){if(null===n)e.removeAttribute(t);else{switch(typeof n){case"undefined":case"function":case"symbol":case"boolean":return void e.removeAttribute(t)}e.setAttribute(t,""+n)}}function st(e,t,n,r){if(null===r)e.removeAttribute(n);else{switch(typeof r){case"undefined":case"function":case"symbol":case"boolean":return void e.removeAttribute(n)}e.setAttributeNS(t,n,""+r)}}function ut(e){switch(typeof e){case"bigint":case"boolean":case"number":case"string":case"undefined":case"object":return e;default:return""}}function ct(e){var t=e.type;return(e=e.nodeName)&&"input"===e.toLowerCase()&&("checkbox"===t||"radio"===t)}function pt(e){e._valueTracker||(e._valueTracker=function(e){var t=ct(e)?"checked":"value",n=Object.getOwnPropertyDescriptor(e.constructor.prototype,t),r=""+e[t];if(!e.hasOwnProperty(t)&&void 0!==n&&"function"==typeof n.get&&"function"==typeof n.set){var a=n.get,o=n.set;return Object.defineProperty(e,t,{configurable:!0,get:function(){return a.call(this)},set:function(e){r=""+e,o.call(this,e)}}),Object.defineProperty(e,t,{enumerable:n.enumerable}),{getValue:function(){return r},setValue:function(e){r=""+e},stopTracking:function(){e._valueTracker=null,delete e[t]}}}}(e))}function dt(e){if(!e)return!1;var t=e._valueTracker;if(!t)return!0;var n=t.getValue(),r="";return e&&(r=ct(e)?e.checked?"true":"false":e.value),(e=r)!==n&&(t.setValue(e),!0)}function ft(e){if(void 0===(e=e||("undefined"!=typeof document?document:void 0)))return null;try{return e.activeElement||e.body}catch(t){return e.body}}var ht=/[\n"\\]/g;function gt(e){return e.replace(ht,(function(e){return"\\"+e.charCodeAt(0).toString(16)+" "}))}function mt(e,t,n,r,a,o,i,l){e.name="",null!=i&&"function"!=typeof i&&"symbol"!=typeof i&&"boolean"!=typeof i?e.type=i:e.removeAttribute("type"),null!=t?"number"===i?(0===t&&""===e.value||e.value!=t)&&(e.value=""+ut(t)):e.value!==""+ut(t)&&(e.value=""+ut(t)):"submit"!==i&&"reset"!==i||e.removeAttribute("value"),null!=t?bt(e,i,ut(t)):null!=n?bt(e,i,ut(n)):null!=r&&e.removeAttribute("value"),null==a&&null!=o&&(e.defaultChecked=!!o),null!=a&&(e.checked=a&&"function"!=typeof a&&"symbol"!=typeof a),null!=l&&"function"!=typeof l&&"symbol"!=typeof l&&"boolean"!=typeof l?e.name=""+ut(l):e.removeAttribute("name")}function yt(e,t,n,r,a,o,i,l){if(null!=o&&"function"!=typeof o&&"symbol"!=typeof o&&"boolean"!=typeof o&&(e.type=o),null!=t||null!=n){if(("submit"===o||"reset"===o)&&null==t)return;n=null!=n?""+ut(n):"",t=null!=t?""+ut(t):n,l||t===e.value||(e.value=t),e.defaultValue=t}r="function"!=typeof(r=null!=r?r:a)&&"symbol"!=typeof r&&!!r,e.checked=l?e.checked:!!r,e.defaultChecked=!!r,null!=i&&"function"!=typeof i&&"symbol"!=typeof i&&"boolean"!=typeof i&&(e.name=i)}function bt(e,t,n){"number"===t&&ft(e.ownerDocument)===e||e.defaultValue===""+n||(e.defaultValue=""+n)}function vt(e,t,n,r){if(e=e.options,t){t={};for(var a=0;a<n.length;a++)t["$"+n[a]]=!0;for(n=0;n<e.length;n++)a=t.hasOwnProperty("$"+e[n].value),e[n].selected!==a&&(e[n].selected=a),a&&r&&(e[n].defaultSelected=!0)}else{for(n=""+ut(n),t=null,a=0;a<e.length;a++){if(e[a].value===n)return e[a].selected=!0,void(r&&(e[a].defaultSelected=!0));null!==t||e[a].disabled||(t=e[a])}null!==t&&(t.selected=!0)}}function wt(e,t,n){null==t||((t=""+ut(t))!==e.value&&(e.value=t),null!=n)?e.defaultValue=null!=n?""+ut(n):"":e.defaultValue!==t&&(e.defaultValue=t)}function xt(e,t,n,r){if(null==t){if(null!=r){if(null!=n)throw Error(i(92));if(B(r)){if(1<r.length)throw Error(i(93));r=r[0]}n=r}null==n&&(n=""),t=n}n=ut(t),e.defaultValue=n,(r=e.textContent)===n&&""!==r&&null!==r&&(e.value=r)}function kt(e,t){if(t){var n=e.firstChild;if(n&&n===e.lastChild&&3===n.nodeType)return void(n.nodeValue=t)}e.textContent=t}var St=new Set("animationIterationCount aspectRatio borderImageOutset borderImageSlice borderImageWidth boxFlex boxFlexGroup boxOrdinalGroup columnCount columns flex flexGrow flexPositive flexShrink flexNegative flexOrder gridArea gridRow gridRowEnd gridRowSpan gridRowStart gridColumn gridColumnEnd gridColumnSpan gridColumnStart fontWeight lineClamp lineHeight opacity order orphans scale tabSize widows zIndex zoom fillOpacity floodOpacity stopOpacity strokeDasharray strokeDashoffset strokeMiterlimit strokeOpacity strokeWidth MozAnimationIterationCount MozBoxFlex MozBoxFlexGroup MozLineClamp msAnimationIterationCount msFlex msZoom msFlexGrow msFlexNegative msFlexOrder msFlexPositive msFlexShrink msGridColumn msGridColumnSpan msGridRow msGridRowSpan WebkitAnimationIterationCount WebkitBoxFlex WebKitBoxFlexGroup WebkitBoxOrdinalGroup WebkitColumnCount WebkitColumns WebkitFlex WebkitFlexGrow WebkitFlexPositive WebkitFlexShrink WebkitLineClamp".split(" "));function Ct(e,t,n){var r=0===t.indexOf("--");null==n||"boolean"==typeof n||""===n?r?e.setProperty(t,""):"float"===t?e.cssFloat="":e[t]="":r?e.setProperty(t,n):"number"!=typeof n||0===n||St.has(t)?"float"===t?e.cssFloat=n:e[t]=(""+n).trim():e[t]=n+"px"}function _t(e,t,n){if(null!=t&&"object"!=typeof t)throw Error(i(62));if(e=e.style,null!=n){for(var r in n)!n.hasOwnProperty(r)||null!=t&&t.hasOwnProperty(r)||(0===r.indexOf("--")?e.setProperty(r,""):"float"===r?e.cssFloat="":e[r]="");for(var a in t)r=t[a],t.hasOwnProperty(a)&&n[a]!==r&&Ct(e,a,r)}else for(var o in t)t.hasOwnProperty(o)&&Ct(e,o,t[o])}function Et(e){if(-1===e.indexOf("-"))return!1;switch(e){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}var Nt=new Map([["acceptCharset","accept-charset"],["htmlFor","for"],["httpEquiv","http-equiv"],["crossOrigin","crossorigin"],["accentHeight","accent-height"],["alignmentBaseline","alignment-baseline"],["arabicForm","arabic-form"],["baselineShift","baseline-shift"],["capHeight","cap-height"],["clipPath","clip-path"],["clipRule","clip-rule"],["colorInterpolation","color-interpolation"],["colorInterpolationFilters","color-interpolation-filters"],["colorProfile","color-profile"],["colorRendering","color-rendering"],["dominantBaseline","dominant-baseline"],["enableBackground","enable-background"],["fillOpacity","fill-opacity"],["fillRule","fill-rule"],["floodColor","flood-color"],["floodOpacity","flood-opacity"],["fontFamily","font-family"],["fontSize","font-size"],["fontSizeAdjust","font-size-adjust"],["fontStretch","font-stretch"],["fontStyle","font-style"],["fontVariant","font-variant"],["fontWeight","font-weight"],["glyphName","glyph-name"],["glyphOrientationHorizontal","glyph-orientation-horizontal"],["glyphOrientationVertical","glyph-orientation-vertical"],["horizAdvX","horiz-adv-x"],["horizOriginX","horiz-origin-x"],["imageRendering","image-rendering"],["letterSpacing","letter-spacing"],["lightingColor","lighting-color"],["markerEnd","marker-end"],["markerMid","marker-mid"],["markerStart","marker-start"],["overlinePosition","overline-position"],["overlineThickness","overline-thickness"],["paintOrder","paint-order"],["panose-1","panose-1"],["pointerEvents","pointer-events"],["renderingIntent","rendering-intent"],["shapeRendering","shape-rendering"],["stopColor","stop-color"],["stopOpacity","stop-opacity"],["strikethroughPosition","strikethrough-position"],["strikethroughThickness","strikethrough-thickness"],["strokeDasharray","stroke-dasharray"],["strokeDashoffset","stroke-dashoffset"],["strokeLinecap","stroke-linecap"],["strokeLinejoin","stroke-linejoin"],["strokeMiterlimit","stroke-miterlimit"],["strokeOpacity","stroke-opacity"],["strokeWidth","stroke-width"],["textAnchor","text-anchor"],["textDecoration","text-decoration"],["textRendering","text-rendering"],["transformOrigin","transform-origin"],["underlinePosition","underline-position"],["underlineThickness","underline-thickness"],["unicodeBidi","unicode-bidi"],["unicodeRange","unicode-range"],["unitsPerEm","units-per-em"],["vAlphabetic","v-alphabetic"],["vHanging","v-hanging"],["vIdeographic","v-ideographic"],["vMathematical","v-mathematical"],["vectorEffect","vector-effect"],["vertAdvY","vert-adv-y"],["vertOriginX","vert-origin-x"],["vertOriginY","vert-origin-y"],["wordSpacing","word-spacing"],["writingMode","writing-mode"],["xmlnsXlink","xmlns:xlink"],["xHeight","x-height"]]),Pt=/^[\u0000-\u001F ]*j[\r\n\t]*a[\r\n\t]*v[\r\n\t]*a[\r\n\t]*s[\r\n\t]*c[\r\n\t]*r[\r\n\t]*i[\r\n\t]*p[\r\n\t]*t[\r\n\t]*:/i;function At(e){return Pt.test(""+e)?"javascript:throw new Error('React has blocked a javascript: URL as a security precaution.')":e}var Lt=null;function Tt(e){return(e=e.target||e.srcElement||window).correspondingUseElement&&(e=e.correspondingUseElement),3===e.nodeType?e.parentNode:e}var Rt=null,It=null;function Mt(e){var t=Qe(e);if(t&&(e=t.stateNode)){var n=e[Fe]||null;e:switch(e=t.stateNode,t.type){case"input":if(mt(e,n.value,n.defaultValue,n.defaultValue,n.checked,n.defaultChecked,n.type,n.name),t=n.name,"radio"===n.type&&null!=t){for(n=e;n.parentNode;)n=n.parentNode;for(n=n.querySelectorAll('input[name="'+gt(""+t)+'"][type="radio"]'),t=0;t<n.length;t++){var r=n[t];if(r!==e&&r.form===e.form){var a=r[Fe]||null;if(!a)throw Error(i(90));mt(r,a.value,a.defaultValue,a.defaultValue,a.checked,a.defaultChecked,a.type,a.name)}}for(t=0;t<n.length;t++)(r=n[t]).form===e.form&&dt(r)}break e;case"textarea":wt(e,n.value,n.defaultValue);break e;case"select":null!=(t=n.value)&&vt(e,!!n.multiple,t,!1)}}}var Ot=!1;function jt(e,t,n){if(Ot)return e(t,n);Ot=!0;try{return e(t)}finally{if(Ot=!1,(null!==Rt||null!==It)&&(Fu(),Rt&&(t=Rt,e=It,It=Rt=null,Mt(t),e)))for(t=0;t<e.length;t++)Mt(e[t])}}function zt(e,t){var n=e.stateNode;if(null===n)return null;var r=n[Fe]||null;if(null===r)return null;n=r[t];e:switch(t){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":case"onMouseEnter":(r=!r.disabled)||(r=!("button"===(e=e.type)||"input"===e||"select"===e||"textarea"===e)),e=!r;break e;default:e=!1}if(e)return null;if(n&&"function"!=typeof n)throw Error(i(231,t,typeof n));return n}var $t=!1;if(nt)try{var Ft={};Object.defineProperty(Ft,"passive",{get:function(){$t=!0}}),window.addEventListener("test",Ft,Ft),window.removeEventListener("test",Ft,Ft)}catch(e){$t=!1}var Dt=null,Bt=null,Ut=null;function qt(){if(Ut)return Ut;var e,t,n=Bt,r=n.length,a="value"in Dt?Dt.value:Dt.textContent,o=a.length;for(e=0;e<r&&n[e]===a[e];e++);var i=r-e;for(t=1;t<=i&&n[r-t]===a[o-t];t++);return Ut=a.slice(e,1<t?1-t:void 0)}function Gt(e){var t=e.keyCode;return"charCode"in e?0===(e=e.charCode)&&13===t&&(e=13):e=t,10===e&&(e=13),32<=e||13===e?e:0}function Wt(){return!0}function Ht(){return!1}function Vt(e){function t(t,n,r,a,o){for(var i in this._reactName=t,this._targetInst=r,this.type=n,this.nativeEvent=a,this.target=o,this.currentTarget=null,e)e.hasOwnProperty(i)&&(t=e[i],this[i]=t?t(a):a[i]);return this.isDefaultPrevented=(null!=a.defaultPrevented?a.defaultPrevented:!1===a.returnValue)?Wt:Ht,this.isPropagationStopped=Ht,this}return T(t.prototype,{preventDefault:function(){this.defaultPrevented=!0;var e=this.nativeEvent;e&&(e.preventDefault?e.preventDefault():"unknown"!=typeof e.returnValue&&(e.returnValue=!1),this.isDefaultPrevented=Wt)},stopPropagation:function(){var e=this.nativeEvent;e&&(e.stopPropagation?e.stopPropagation():"unknown"!=typeof e.cancelBubble&&(e.cancelBubble=!0),this.isPropagationStopped=Wt)},persist:function(){},isPersistent:Wt}),t}var Qt,Xt,Kt,Yt={eventPhase:0,bubbles:0,cancelable:0,timeStamp:function(e){return e.timeStamp||Date.now()},defaultPrevented:0,isTrusted:0},Zt=Vt(Yt),Jt=T({},Yt,{view:0,detail:0}),en=Vt(Jt),tn=T({},Jt,{screenX:0,screenY:0,clientX:0,clientY:0,pageX:0,pageY:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,getModifierState:fn,button:0,buttons:0,relatedTarget:function(e){return void 0===e.relatedTarget?e.fromElement===e.srcElement?e.toElement:e.fromElement:e.relatedTarget},movementX:function(e){return"movementX"in e?e.movementX:(e!==Kt&&(Kt&&"mousemove"===e.type?(Qt=e.screenX-Kt.screenX,Xt=e.screenY-Kt.screenY):Xt=Qt=0,Kt=e),Qt)},movementY:function(e){return"movementY"in e?e.movementY:Xt}}),nn=Vt(tn),rn=Vt(T({},tn,{dataTransfer:0})),an=Vt(T({},Jt,{relatedTarget:0})),on=Vt(T({},Yt,{animationName:0,elapsedTime:0,pseudoElement:0})),ln=Vt(T({},Yt,{clipboardData:function(e){return"clipboardData"in e?e.clipboardData:window.clipboardData}})),sn=Vt(T({},Yt,{data:0})),un={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},cn={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},pn={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function dn(e){var t=this.nativeEvent;return t.getModifierState?t.getModifierState(e):!!(e=pn[e])&&!!t[e]}function fn(){return dn}var hn=T({},Jt,{key:function(e){if(e.key){var t=un[e.key]||e.key;if("Unidentified"!==t)return t}return"keypress"===e.type?13===(e=Gt(e))?"Enter":String.fromCharCode(e):"keydown"===e.type||"keyup"===e.type?cn[e.keyCode]||"Unidentified":""},code:0,location:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,repeat:0,locale:0,getModifierState:fn,charCode:function(e){return"keypress"===e.type?Gt(e):0},keyCode:function(e){return"keydown"===e.type||"keyup"===e.type?e.keyCode:0},which:function(e){return"keypress"===e.type?Gt(e):"keydown"===e.type||"keyup"===e.type?e.keyCode:0}}),gn=Vt(hn),mn=Vt(T({},tn,{pointerId:0,width:0,height:0,pressure:0,tangentialPressure:0,tiltX:0,tiltY:0,twist:0,pointerType:0,isPrimary:0})),yn=Vt(T({},Jt,{touches:0,targetTouches:0,changedTouches:0,altKey:0,metaKey:0,ctrlKey:0,shiftKey:0,getModifierState:fn})),bn=Vt(T({},Yt,{propertyName:0,elapsedTime:0,pseudoElement:0})),vn=Vt(T({},tn,{deltaX:function(e){return"deltaX"in e?e.deltaX:"wheelDeltaX"in e?-e.wheelDeltaX:0},deltaY:function(e){return"deltaY"in e?e.deltaY:"wheelDeltaY"in e?-e.wheelDeltaY:"wheelDelta"in e?-e.wheelDelta:0},deltaZ:0,deltaMode:0})),wn=Vt(T({},Yt,{newState:0,oldState:0})),xn=[9,13,27,32],kn=nt&&"CompositionEvent"in window,Sn=null;nt&&"documentMode"in document&&(Sn=document.documentMode);var Cn=nt&&"TextEvent"in window&&!Sn,_n=nt&&(!kn||Sn&&8<Sn&&11>=Sn),En=String.fromCharCode(32),Nn=!1;function Pn(e,t){switch(e){case"keyup":return-1!==xn.indexOf(t.keyCode);case"keydown":return 229!==t.keyCode;case"keypress":case"mousedown":case"focusout":return!0;default:return!1}}function An(e){return"object"==typeof(e=e.detail)&&"data"in e?e.data:null}var Ln=!1,Tn={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function Rn(e){var t=e&&e.nodeName&&e.nodeName.toLowerCase();return"input"===t?!!Tn[e.type]:"textarea"===t}function In(e,t,n,r){Rt?It?It.push(r):It=[r]:Rt=r,0<(t=zc(t,"onChange")).length&&(n=new Zt("onChange","change",null,n,r),e.push({event:n,listeners:t}))}var Mn=null,On=null;function jn(e){Ac(e,0)}function zn(e){if(dt(Xe(e)))return e}function $n(e,t){if("change"===e)return t}var Fn=!1;if(nt){var Dn;if(nt){var Bn="oninput"in document;if(!Bn){var Un=document.createElement("div");Un.setAttribute("oninput","return;"),Bn="function"==typeof Un.oninput}Dn=Bn}else Dn=!1;Fn=Dn&&(!document.documentMode||9<document.documentMode)}function qn(){Mn&&(Mn.detachEvent("onpropertychange",Gn),On=Mn=null)}function Gn(e){if("value"===e.propertyName&&zn(On)){var t=[];In(t,On,e,Tt(e)),jt(jn,t)}}function Wn(e,t,n){"focusin"===e?(qn(),On=n,(Mn=t).attachEvent("onpropertychange",Gn)):"focusout"===e&&qn()}function Hn(e){if("selectionchange"===e||"keyup"===e||"keydown"===e)return zn(On)}function Vn(e,t){if("click"===e)return zn(t)}function Qn(e,t){if("input"===e||"change"===e)return zn(t)}var Xn="function"==typeof Object.is?Object.is:function(e,t){return e===t&&(0!==e||1/e==1/t)||e!=e&&t!=t};function Kn(e,t){if(Xn(e,t))return!0;if("object"!=typeof e||null===e||"object"!=typeof t||null===t)return!1;var n=Object.keys(e),r=Object.keys(t);if(n.length!==r.length)return!1;for(r=0;r<n.length;r++){var a=n[r];if(!re.call(t,a)||!Xn(e[a],t[a]))return!1}return!0}function Yn(e){for(;e&&e.firstChild;)e=e.firstChild;return e}function Zn(e,t){var n,r=Yn(e);for(e=0;r;){if(3===r.nodeType){if(n=e+r.textContent.length,e<=t&&n>=t)return{node:r,offset:t-e};e=n}e:{for(;r;){if(r.nextSibling){r=r.nextSibling;break e}r=r.parentNode}r=void 0}r=Yn(r)}}function Jn(e,t){return!(!e||!t)&&(e===t||(!e||3!==e.nodeType)&&(t&&3===t.nodeType?Jn(e,t.parentNode):"contains"in e?e.contains(t):!!e.compareDocumentPosition&&!!(16&e.compareDocumentPosition(t))))}function er(e){for(var t=ft((e=null!=e&&null!=e.ownerDocument&&null!=e.ownerDocument.defaultView?e.ownerDocument.defaultView:window).document);t instanceof e.HTMLIFrameElement;){try{var n="string"==typeof t.contentWindow.location.href}catch(e){n=!1}if(!n)break;t=ft((e=t.contentWindow).document)}return t}function tr(e){var t=e&&e.nodeName&&e.nodeName.toLowerCase();return t&&("input"===t&&("text"===e.type||"search"===e.type||"tel"===e.type||"url"===e.type||"password"===e.type)||"textarea"===t||"true"===e.contentEditable)}function nr(e,t){var n=er(t);t=e.focusedElem;var r=e.selectionRange;if(n!==t&&t&&t.ownerDocument&&Jn(t.ownerDocument.documentElement,t)){if(null!==r&&tr(t))if(e=r.start,void 0===(n=r.end)&&(n=e),"selectionStart"in t)t.selectionStart=e,t.selectionEnd=Math.min(n,t.value.length);else if((n=(e=t.ownerDocument||document)&&e.defaultView||window).getSelection){n=n.getSelection();var a=t.textContent.length,o=Math.min(r.start,a);r=void 0===r.end?o:Math.min(r.end,a),!n.extend&&o>r&&(a=r,r=o,o=a),a=Zn(t,o);var i=Zn(t,r);a&&i&&(1!==n.rangeCount||n.anchorNode!==a.node||n.anchorOffset!==a.offset||n.focusNode!==i.node||n.focusOffset!==i.offset)&&((e=e.createRange()).setStart(a.node,a.offset),n.removeAllRanges(),o>r?(n.addRange(e),n.extend(i.node,i.offset)):(e.setEnd(i.node,i.offset),n.addRange(e)))}for(e=[],n=t;n=n.parentNode;)1===n.nodeType&&e.push({element:n,left:n.scrollLeft,top:n.scrollTop});for("function"==typeof t.focus&&t.focus(),t=0;t<e.length;t++)(n=e[t]).element.scrollLeft=n.left,n.element.scrollTop=n.top}}var rr=nt&&"documentMode"in document&&11>=document.documentMode,ar=null,or=null,ir=null,lr=!1;function sr(e,t,n){var r=n.window===n?n.document:9===n.nodeType?n:n.ownerDocument;lr||null==ar||ar!==ft(r)||(r="selectionStart"in(r=ar)&&tr(r)?{start:r.selectionStart,end:r.selectionEnd}:{anchorNode:(r=(r.ownerDocument&&r.ownerDocument.defaultView||window).getSelection()).anchorNode,anchorOffset:r.anchorOffset,focusNode:r.focusNode,focusOffset:r.focusOffset},ir&&Kn(ir,r)||(ir=r,0<(r=zc(or,"onSelect")).length&&(t=new Zt("onSelect","select",null,t,n),e.push({event:t,listeners:r}),t.target=ar)))}function ur(e,t){var n={};return n[e.toLowerCase()]=t.toLowerCase(),n["Webkit"+e]="webkit"+t,n["Moz"+e]="moz"+t,n}var cr={animationend:ur("Animation","AnimationEnd"),animationiteration:ur("Animation","AnimationIteration"),animationstart:ur("Animation","AnimationStart"),transitionrun:ur("Transition","TransitionRun"),transitionstart:ur("Transition","TransitionStart"),transitioncancel:ur("Transition","TransitionCancel"),transitionend:ur("Transition","TransitionEnd")},pr={},dr={};function fr(e){if(pr[e])return pr[e];if(!cr[e])return e;var t,n=cr[e];for(t in n)if(n.hasOwnProperty(t)&&t in dr)return pr[e]=n[t];return e}nt&&(dr=document.createElement("div").style,"AnimationEvent"in window||(delete cr.animationend.animation,delete cr.animationiteration.animation,delete cr.animationstart.animation),"TransitionEvent"in window||delete cr.transitionend.transition);var hr=fr("animationend"),gr=fr("animationiteration"),mr=fr("animationstart"),yr=fr("transitionrun"),br=fr("transitionstart"),vr=fr("transitioncancel"),wr=fr("transitionend"),xr=new Map,kr="abort auxClick beforeToggle cancel canPlay canPlayThrough click close contextMenu copy cut drag dragEnd dragEnter dragExit dragLeave dragOver dragStart drop durationChange emptied encrypted ended error gotPointerCapture input invalid keyDown keyPress keyUp load loadedData loadedMetadata loadStart lostPointerCapture mouseDown mouseMove mouseOut mouseOver mouseUp paste pause play playing pointerCancel pointerDown pointerMove pointerOut pointerOver pointerUp progress rateChange reset resize seeked seeking stalled submit suspend timeUpdate touchCancel touchEnd touchStart volumeChange scroll scrollEnd toggle touchMove waiting wheel".split(" ");function Sr(e,t){xr.set(e,t),et(t,[e])}var Cr=[],_r=0,Er=0;function Nr(){for(var e=_r,t=Er=_r=0;t<e;){var n=Cr[t];Cr[t++]=null;var r=Cr[t];Cr[t++]=null;var a=Cr[t];Cr[t++]=null;var o=Cr[t];if(Cr[t++]=null,null!==r&&null!==a){var i=r.pending;null===i?a.next=a:(a.next=i.next,i.next=a),r.pending=a}0!==o&&Tr(n,a,o)}}function Pr(e,t,n,r){Cr[_r++]=e,Cr[_r++]=t,Cr[_r++]=n,Cr[_r++]=r,Er|=r,e.lanes|=r,null!==(e=e.alternate)&&(e.lanes|=r)}function Ar(e,t,n,r){return Pr(e,t,n,r),Rr(e)}function Lr(e,t){return Pr(e,null,null,t),Rr(e)}function Tr(e,t,n){e.lanes|=n;var r=e.alternate;null!==r&&(r.lanes|=n);for(var a=!1,o=e.return;null!==o;)o.childLanes|=n,null!==(r=o.alternate)&&(r.childLanes|=n),22===o.tag&&(null===(e=o.stateNode)||1&e._visibility||(a=!0)),e=o,o=o.return;a&&null!==t&&3===e.tag&&(o=e.stateNode,a=31-we(n),null===(e=(o=o.hiddenUpdates)[a])?o[a]=[t]:e.push(t),t.lane=536870912|n)}function Rr(e){if(50<Au)throw Au=0,Lu=null,Error(i(185));for(var t=e.return;null!==t;)t=(e=t).return;return 3===e.tag?e.stateNode:null}var Ir={},Mr=new WeakMap;function Or(e,t){if("object"==typeof e&&null!==e){var n=Mr.get(e);return void 0!==n?n:(t={value:e,source:t,stack:j(t)},Mr.set(e,t),t)}return{value:e,source:t,stack:j(t)}}var jr=[],zr=0,$r=null,Fr=0,Dr=[],Br=0,Ur=null,qr=1,Gr="";function Wr(e,t){jr[zr++]=Fr,jr[zr++]=$r,$r=e,Fr=t}function Hr(e,t,n){Dr[Br++]=qr,Dr[Br++]=Gr,Dr[Br++]=Ur,Ur=e;var r=qr;e=Gr;var a=32-we(r)-1;r&=~(1<<a),n+=1;var o=32-we(t)+a;if(30<o){var i=a-a%5;o=(r&(1<<i)-1).toString(32),r>>=i,a-=i,qr=1<<32-we(t)+a|n<<a|r,Gr=o+e}else qr=1<<o|n<<a|r,Gr=e}function Vr(e){null!==e.return&&(Wr(e,1),Hr(e,1,0))}function Qr(e){for(;e===$r;)$r=jr[--zr],jr[zr]=null,Fr=jr[--zr],jr[zr]=null;for(;e===Ur;)Ur=Dr[--Br],Dr[Br]=null,Gr=Dr[--Br],Dr[Br]=null,qr=Dr[--Br],Dr[Br]=null}var Xr=null,Kr=null,Yr=!1,Zr=null,Jr=!1,ea=Error(i(519));function ta(e){throw ia(Or(Error(i(418,"")),e)),ea}function na(e){var t=e.stateNode,n=e.type,r=e.memoizedProps;switch(t[$e]=e,t[Fe]=r,n){case"dialog":Lc("cancel",t),Lc("close",t);break;case"iframe":case"object":case"embed":Lc("load",t);break;case"video":case"audio":for(n=0;n<Nc.length;n++)Lc(Nc[n],t);break;case"source":Lc("error",t);break;case"img":case"image":case"link":Lc("error",t),Lc("load",t);break;case"details":Lc("toggle",t);break;case"input":Lc("invalid",t),yt(t,r.value,r.defaultValue,r.checked,r.defaultChecked,r.type,r.name,!0),pt(t);break;case"select":Lc("invalid",t);break;case"textarea":Lc("invalid",t),xt(t,r.value,r.defaultValue,r.children),pt(t)}"string"!=typeof(n=r.children)&&"number"!=typeof n&&"bigint"!=typeof n||t.textContent===""+n||!0===r.suppressHydrationWarning||qc(t.textContent,n)?(null!=r.popover&&(Lc("beforetoggle",t),Lc("toggle",t)),null!=r.onScroll&&Lc("scroll",t),null!=r.onScrollEnd&&Lc("scrollend",t),null!=r.onClick&&(t.onclick=Gc),t=!0):t=!1,t||ta(e)}function ra(e){for(Xr=e.return;Xr;)switch(Xr.tag){case 3:case 27:return void(Jr=!0);case 5:case 13:return void(Jr=!1);default:Xr=Xr.return}}function aa(e){if(e!==Xr)return!1;if(!Yr)return ra(e),Yr=!0,!1;var t,n=!1;if((t=3!==e.tag&&27!==e.tag)&&((t=5===e.tag)&&(t=!("form"!==(t=e.type)&&"button"!==t)||Jc(e.type,e.memoizedProps)),t=!t),t&&(n=!0),n&&Kr&&ta(e),ra(e),13===e.tag){if(!(e=null!==(e=e.memoizedState)?e.dehydrated:null))throw Error(i(317));e:{for(e=e.nextSibling,n=0;e;){if(8===e.nodeType)if("/$"===(t=e.data)){if(0===n){Kr=sp(e.nextSibling);break e}n--}else"$"!==t&&"$!"!==t&&"$?"!==t||n++;e=e.nextSibling}Kr=null}}else Kr=Xr?sp(e.stateNode.nextSibling):null;return!0}function oa(){Kr=Xr=null,Yr=!1}function ia(e){null===Zr?Zr=[e]:Zr.push(e)}var la=Error(i(460)),sa=Error(i(474)),ua={then:function(){}};function ca(e){return"fulfilled"===(e=e.status)||"rejected"===e}function pa(){}function da(e,t,n){switch(void 0===(n=e[n])?e.push(t):n!==t&&(t.then(pa,pa),t=n),t.status){case"fulfilled":return t.value;case"rejected":if((e=t.reason)===la)throw Error(i(483));throw e;default:if("string"==typeof t.status)t.then(pa,pa);else{if(null!==(e=nu)&&100<e.shellSuspendCounter)throw Error(i(482));(e=t).status="pending",e.then((function(e){if("pending"===t.status){var n=t;n.status="fulfilled",n.value=e}}),(function(e){if("pending"===t.status){var n=t;n.status="rejected",n.reason=e}}))}switch(t.status){case"fulfilled":return t.value;case"rejected":if((e=t.reason)===la)throw Error(i(483));throw e}throw fa=t,la}}var fa=null;function ha(){if(null===fa)throw Error(i(459));var e=fa;return fa=null,e}var ga=null,ma=0;function ya(e){var t=ma;return ma+=1,null===ga&&(ga=[]),da(ga,e,t)}function ba(e,t){t=t.props.ref,e.ref=void 0!==t?t:null}function va(e,t){if(t.$$typeof===s)throw Error(i(525));throw e=Object.prototype.toString.call(t),Error(i(31,"[object Object]"===e?"object with keys {"+Object.keys(t).join(", ")+"}":e))}function wa(e){return(0,e._init)(e._payload)}function xa(e){function t(t,n){if(e){var r=t.deletions;null===r?(t.deletions=[n],t.flags|=16):r.push(n)}}function n(n,r){if(!e)return null;for(;null!==r;)t(n,r),r=r.sibling;return null}function r(e){for(var t=new Map;null!==e;)null!==e.key?t.set(e.key,e):t.set(e.index,e),e=e.sibling;return t}function a(e,t){return(e=$s(e,t)).index=0,e.sibling=null,e}function o(t,n,r){return t.index=r,e?null!==(r=t.alternate)?(r=r.index)<n?(t.flags|=33554434,n):r:(t.flags|=33554434,n):(t.flags|=1048576,n)}function l(t){return e&&null===t.alternate&&(t.flags|=33554434),t}function s(e,t,n,r){return null===t||6!==t.tag?((t=qs(n,e.mode,r)).return=e,t):((t=a(t,n)).return=e,t)}function d(e,t,n,r){var o=n.type;return o===p?h(e,t,n.props.children,r,n.key):null!==t&&(t.elementType===o||"object"==typeof o&&null!==o&&o.$$typeof===x&&wa(o)===t.type)?(ba(t=a(t,n.props),n),t.return=e,t):(ba(t=Ds(n.type,n.key,n.props,null,e.mode,r),n),t.return=e,t)}function f(e,t,n,r){return null===t||4!==t.tag||t.stateNode.containerInfo!==n.containerInfo||t.stateNode.implementation!==n.implementation?((t=Gs(n,e.mode,r)).return=e,t):((t=a(t,n.children||[])).return=e,t)}function h(e,t,n,r,o){return null===t||7!==t.tag?((t=Bs(n,e.mode,r,o)).return=e,t):((t=a(t,n)).return=e,t)}function g(e,t,n){if("string"==typeof t&&""!==t||"number"==typeof t||"bigint"==typeof t)return(t=qs(""+t,e.mode,n)).return=e,t;if("object"==typeof t&&null!==t){switch(t.$$typeof){case u:return ba(n=Ds(t.type,t.key,t.props,null,e.mode,n),t),n.return=e,n;case c:return(t=Gs(t,e.mode,n)).return=e,t;case x:return g(e,t=(0,t._init)(t._payload),n)}if(B(t)||_(t))return(t=Bs(t,e.mode,n,null)).return=e,t;if("function"==typeof t.then)return g(e,ya(t),n);if(t.$$typeof===m)return g(e,Nl(e,t),n);va(e,t)}return null}function y(e,t,n,r){var a=null!==t?t.key:null;if("string"==typeof n&&""!==n||"number"==typeof n||"bigint"==typeof n)return null!==a?null:s(e,t,""+n,r);if("object"==typeof n&&null!==n){switch(n.$$typeof){case u:return n.key===a?d(e,t,n,r):null;case c:return n.key===a?f(e,t,n,r):null;case x:return y(e,t,n=(a=n._init)(n._payload),r)}if(B(n)||_(n))return null!==a?null:h(e,t,n,r,null);if("function"==typeof n.then)return y(e,t,ya(n),r);if(n.$$typeof===m)return y(e,t,Nl(e,n),r);va(e,n)}return null}function b(e,t,n,r,a){if("string"==typeof r&&""!==r||"number"==typeof r||"bigint"==typeof r)return s(t,e=e.get(n)||null,""+r,a);if("object"==typeof r&&null!==r){switch(r.$$typeof){case u:return d(t,e=e.get(null===r.key?n:r.key)||null,r,a);case c:return f(t,e=e.get(null===r.key?n:r.key)||null,r,a);case x:return b(e,t,n,r=(0,r._init)(r._payload),a)}if(B(r)||_(r))return h(t,e=e.get(n)||null,r,a,null);if("function"==typeof r.then)return b(e,t,n,ya(r),a);if(r.$$typeof===m)return b(e,t,n,Nl(t,r),a);va(t,r)}return null}function v(s,d,f,h){if("object"==typeof f&&null!==f&&f.type===p&&null===f.key&&(f=f.props.children),"object"==typeof f&&null!==f){switch(f.$$typeof){case u:e:{for(var w=f.key;null!==d;){if(d.key===w){if((w=f.type)===p){if(7===d.tag){n(s,d.sibling),(h=a(d,f.props.children)).return=s,s=h;break e}}else if(d.elementType===w||"object"==typeof w&&null!==w&&w.$$typeof===x&&wa(w)===d.type){n(s,d.sibling),ba(h=a(d,f.props),f),h.return=s,s=h;break e}n(s,d);break}t(s,d),d=d.sibling}f.type===p?((h=Bs(f.props.children,s.mode,h,f.key)).return=s,s=h):(ba(h=Ds(f.type,f.key,f.props,null,s.mode,h),f),h.return=s,s=h)}return l(s);case c:e:{for(w=f.key;null!==d;){if(d.key===w){if(4===d.tag&&d.stateNode.containerInfo===f.containerInfo&&d.stateNode.implementation===f.implementation){n(s,d.sibling),(h=a(d,f.children||[])).return=s,s=h;break e}n(s,d);break}t(s,d),d=d.sibling}(h=Gs(f,s.mode,h)).return=s,s=h}return l(s);case x:return v(s,d,f=(w=f._init)(f._payload),h)}if(B(f))return function(a,i,l,s){for(var u=null,c=null,p=i,d=i=0,f=null;null!==p&&d<l.length;d++){p.index>d?(f=p,p=null):f=p.sibling;var h=y(a,p,l[d],s);if(null===h){null===p&&(p=f);break}e&&p&&null===h.alternate&&t(a,p),i=o(h,i,d),null===c?u=h:c.sibling=h,c=h,p=f}if(d===l.length)return n(a,p),Yr&&Wr(a,d),u;if(null===p){for(;d<l.length;d++)null!==(p=g(a,l[d],s))&&(i=o(p,i,d),null===c?u=p:c.sibling=p,c=p);return Yr&&Wr(a,d),u}for(p=r(p);d<l.length;d++)null!==(f=b(p,a,d,l[d],s))&&(e&&null!==f.alternate&&p.delete(null===f.key?d:f.key),i=o(f,i,d),null===c?u=f:c.sibling=f,c=f);return e&&p.forEach((function(e){return t(a,e)})),Yr&&Wr(a,d),u}(s,d,f,h);if(_(f)){if("function"!=typeof(w=_(f)))throw Error(i(150));return function(a,l,s,u){if(null==s)throw Error(i(151));for(var c=null,p=null,d=l,f=l=0,h=null,m=s.next();null!==d&&!m.done;f++,m=s.next()){d.index>f?(h=d,d=null):h=d.sibling;var v=y(a,d,m.value,u);if(null===v){null===d&&(d=h);break}e&&d&&null===v.alternate&&t(a,d),l=o(v,l,f),null===p?c=v:p.sibling=v,p=v,d=h}if(m.done)return n(a,d),Yr&&Wr(a,f),c;if(null===d){for(;!m.done;f++,m=s.next())null!==(m=g(a,m.value,u))&&(l=o(m,l,f),null===p?c=m:p.sibling=m,p=m);return Yr&&Wr(a,f),c}for(d=r(d);!m.done;f++,m=s.next())null!==(m=b(d,a,f,m.value,u))&&(e&&null!==m.alternate&&d.delete(null===m.key?f:m.key),l=o(m,l,f),null===p?c=m:p.sibling=m,p=m);return e&&d.forEach((function(e){return t(a,e)})),Yr&&Wr(a,f),c}(s,d,f=w.call(f),h)}if("function"==typeof f.then)return v(s,d,ya(f),h);if(f.$$typeof===m)return v(s,d,Nl(s,f),h);va(s,f)}return"string"==typeof f&&""!==f||"number"==typeof f||"bigint"==typeof f?(f=""+f,null!==d&&6===d.tag?(n(s,d.sibling),(h=a(d,f)).return=s,s=h):(n(s,d),(h=qs(f,s.mode,h)).return=s,s=h),l(s)):n(s,d)}return function(e,t,n,r){try{ma=0;var a=v(e,t,n,r);return ga=null,a}catch(t){if(t===la)throw t;var o=js(29,t,null,e.mode);return o.lanes=r,o.return=e,o}}}var ka=xa(!0),Sa=xa(!1),Ca=H(null),_a=H(0);function Ea(e,t){Q(_a,e=cu),Q(Ca,t),cu=e|t.baseLanes}function Na(){Q(_a,cu),Q(Ca,Ca.current)}function Pa(){cu=_a.current,V(Ca),V(_a)}var Aa=H(null),La=null;function Ta(e){var t=e.alternate;Q(Oa,1&Oa.current),Q(Aa,e),null===La&&(null===t||null!==Ca.current||null!==t.memoizedState)&&(La=e)}function Ra(e){if(22===e.tag){if(Q(Oa,Oa.current),Q(Aa,e),null===La){var t=e.alternate;null!==t&&null!==t.memoizedState&&(La=e)}}else Ia()}function Ia(){Q(Oa,Oa.current),Q(Aa,Aa.current)}function Ma(e){V(Aa),La===e&&(La=null),V(Oa)}var Oa=H(0);function ja(e){for(var t=e;null!==t;){if(13===t.tag){var n=t.memoizedState;if(null!==n&&(null===(n=n.dehydrated)||"$?"===n.data||"$!"===n.data))return t}else if(19===t.tag&&void 0!==t.memoizedProps.revealOrder){if(128&t.flags)return t}else if(null!==t.child){t.child.return=t,t=t.child;continue}if(t===e)break;for(;null===t.sibling;){if(null===t.return||t.return===e)return null;t=t.return}t.sibling.return=t.return,t=t.sibling}return null}var za="undefined"!=typeof AbortController?AbortController:function(){var e=[],t=this.signal={aborted:!1,addEventListener:function(t,n){e.push(n)}};this.abort=function(){t.aborted=!0,e.forEach((function(e){return e()}))}},$a=r.unstable_scheduleCallback,Fa=r.unstable_NormalPriority,Da={$$typeof:m,Consumer:null,Provider:null,_currentValue:null,_currentValue2:null,_threadCount:0};function Ba(){return{controller:new za,data:new Map,refCount:0}}function Ua(e){e.refCount--,0===e.refCount&&$a(Fa,(function(){e.controller.abort()}))}var qa=null,Ga=0,Wa=0,Ha=null;function Va(){if(0==--Ga&&null!==qa){null!==Ha&&(Ha.status="fulfilled");var e=qa;qa=null,Wa=0,Ha=null;for(var t=0;t<e.length;t++)(0,e[t])()}}var Qa=L.S;L.S=function(e,t){"object"==typeof t&&null!==t&&"function"==typeof t.then&&function(e,t){if(null===qa){var n=qa=[];Ga=0,Wa=kc(),Ha={status:"pending",value:void 0,then:function(e){n.push(e)}}}Ga++,t.then(Va,Va)}(0,t),null!==Qa&&Qa(e,t)};var Xa=H(null);function Ka(){var e=Xa.current;return null!==e?e:nu.pooledCache}function Ya(e,t){Q(Xa,null===t?Xa.current:t.pool)}function Za(){var e=Ka();return null===e?null:{parent:Da._currentValue,pool:e}}var Ja=0,eo=null,to=null,no=null,ro=!1,ao=!1,oo=!1,io=0,lo=0,so=null,uo=0;function co(){throw Error(i(321))}function po(e,t){if(null===t)return!1;for(var n=0;n<t.length&&n<e.length;n++)if(!Xn(e[n],t[n]))return!1;return!0}function fo(e,t,n,r,a,o){return Ja=o,eo=t,t.memoizedState=null,t.updateQueue=null,t.lanes=0,L.H=null===e||null===e.memoizedState?Ni:Pi,oo=!1,o=n(r,a),oo=!1,ao&&(o=go(t,n,r,a)),ho(e),o}function ho(e){L.H=Ei;var t=null!==to&&null!==to.next;if(Ja=0,no=to=eo=null,ro=!1,lo=0,so=null,t)throw Error(i(300));null===e||Wi||null!==(e=e.dependencies)&&Cl(e)&&(Wi=!0)}function go(e,t,n,r){eo=e;var a=0;do{if(ao&&(so=null),lo=0,ao=!1,25<=a)throw Error(i(301));if(a+=1,no=to=null,null!=e.updateQueue){var o=e.updateQueue;o.lastEffect=null,o.events=null,o.stores=null,null!=o.memoCache&&(o.memoCache.index=0)}L.H=Ai,o=t(n,r)}while(ao);return o}function mo(){var e=L.H,t=e.useState()[0];return t="function"==typeof t.then?ko(t):t,e=e.useState()[0],(null!==to?to.memoizedState:null)!==e&&(eo.flags|=1024),t}function yo(){var e=0!==io;return io=0,e}function bo(e,t,n){t.updateQueue=e.updateQueue,t.flags&=-2053,e.lanes&=~n}function vo(e){if(ro){for(e=e.memoizedState;null!==e;){var t=e.queue;null!==t&&(t.pending=null),e=e.next}ro=!1}Ja=0,no=to=eo=null,ao=!1,lo=io=0,so=null}function wo(){var e={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return null===no?eo.memoizedState=no=e:no=no.next=e,no}function xo(){if(null===to){var e=eo.alternate;e=null!==e?e.memoizedState:null}else e=to.next;var t=null===no?eo.memoizedState:no.next;if(null!==t)no=t,to=e;else{if(null===e){if(null===eo.alternate)throw Error(i(467));throw Error(i(310))}e={memoizedState:(to=e).memoizedState,baseState:to.baseState,baseQueue:to.baseQueue,queue:to.queue,next:null},null===no?eo.memoizedState=no=e:no=no.next=e}return no}function ko(e){var t=lo;return lo+=1,null===so&&(so=[]),e=da(so,e,t),t=eo,null===(null===no?t.memoizedState:no.next)&&(t=t.alternate,L.H=null===t||null===t.memoizedState?Ni:Pi),e}function So(e){if(null!==e&&"object"==typeof e){if("function"==typeof e.then)return ko(e);if(e.$$typeof===m)return El(e)}throw Error(i(438,String(e)))}function Co(e){var t=null,n=eo.updateQueue;if(null!==n&&(t=n.memoCache),null==t){var r=eo.alternate;null!==r&&null!==(r=r.updateQueue)&&null!=(r=r.memoCache)&&(t={data:r.data.map((function(e){return e.slice()})),index:0})}if(null==t&&(t={data:[],index:0}),null===n&&(n={lastEffect:null,events:null,stores:null,memoCache:null},eo.updateQueue=n),n.memoCache=t,void 0===(n=t.data[t.index]))for(n=t.data[t.index]=Array(e),r=0;r<e;r++)n[r]=S;return t.index++,n}function _o(e,t){return"function"==typeof t?t(e):t}function Eo(e){return No(xo(),to,e)}function No(e,t,n){var r=e.queue;if(null===r)throw Error(i(311));r.lastRenderedReducer=n;var a=e.baseQueue,o=r.pending;if(null!==o){if(null!==a){var l=a.next;a.next=o.next,o.next=l}t.baseQueue=a=o,r.pending=null}if(o=e.baseState,null===a)e.memoizedState=o;else{var s=l=null,u=null,c=t=a.next,p=!1;do{var d=-536870913&c.lane;if(d!==c.lane?(au&d)===d:(Ja&d)===d){var f=c.revertLane;if(0===f)null!==u&&(u=u.next={lane:0,revertLane:0,action:c.action,hasEagerState:c.hasEagerState,eagerState:c.eagerState,next:null}),d===Wa&&(p=!0);else{if((Ja&f)===f){c=c.next,f===Wa&&(p=!0);continue}d={lane:0,revertLane:c.revertLane,action:c.action,hasEagerState:c.hasEagerState,eagerState:c.eagerState,next:null},null===u?(s=u=d,l=o):u=u.next=d,eo.lanes|=f,du|=f}d=c.action,oo&&n(o,d),o=c.hasEagerState?c.eagerState:n(o,d)}else f={lane:d,revertLane:c.revertLane,action:c.action,hasEagerState:c.hasEagerState,eagerState:c.eagerState,next:null},null===u?(s=u=f,l=o):u=u.next=f,eo.lanes|=d,du|=d;c=c.next}while(null!==c&&c!==t);if(null===u?l=o:u.next=s,!Xn(o,e.memoizedState)&&(Wi=!0,p&&null!==(n=Ha)))throw n;e.memoizedState=o,e.baseState=l,e.baseQueue=u,r.lastRenderedState=o}return null===a&&(r.lanes=0),[e.memoizedState,r.dispatch]}function Po(e){var t=xo(),n=t.queue;if(null===n)throw Error(i(311));n.lastRenderedReducer=e;var r=n.dispatch,a=n.pending,o=t.memoizedState;if(null!==a){n.pending=null;var l=a=a.next;do{o=e(o,l.action),l=l.next}while(l!==a);Xn(o,t.memoizedState)||(Wi=!0),t.memoizedState=o,null===t.baseQueue&&(t.baseState=o),n.lastRenderedState=o}return[o,r]}function Ao(e,t,n){var r=eo,a=xo(),o=Yr;if(o){if(void 0===n)throw Error(i(407));n=n()}else n=t();var l=!Xn((to||a).memoizedState,n);if(l&&(a.memoizedState=n,Wi=!0),a=a.queue,ei(Ro.bind(null,r,a,e),[e]),a.getSnapshot!==t||l||null!==no&&1&no.memoizedState.tag){if(r.flags|=2048,Xo(9,To.bind(null,r,a,n,t),{destroy:void 0},null),null===nu)throw Error(i(349));o||60&Ja||Lo(r,t,n)}return n}function Lo(e,t,n){e.flags|=16384,e={getSnapshot:t,value:n},null===(t=eo.updateQueue)?(t={lastEffect:null,events:null,stores:null,memoCache:null},eo.updateQueue=t,t.stores=[e]):null===(n=t.stores)?t.stores=[e]:n.push(e)}function To(e,t,n,r){t.value=n,t.getSnapshot=r,Io(t)&&Mo(e)}function Ro(e,t,n){return n((function(){Io(t)&&Mo(e)}))}function Io(e){var t=e.getSnapshot;e=e.value;try{var n=t();return!Xn(e,n)}catch(e){return!0}}function Mo(e){var t=Lr(e,2);null!==t&&Iu(t,0,2)}function Oo(e){var t=wo();if("function"==typeof e){var n=e;if(e=n(),oo){ve(!0);try{n()}finally{ve(!1)}}}return t.memoizedState=t.baseState=e,t.queue={pending:null,lanes:0,dispatch:null,lastRenderedReducer:_o,lastRenderedState:e},t}function jo(e,t,n,r){return e.baseState=n,No(e,to,"function"==typeof r?r:_o)}function zo(e,t,n,r,a){if(Si(e))throw Error(i(485));if(null!==(e=t.action)){var o={payload:a,action:e,next:null,isTransition:!0,status:"pending",value:null,reason:null,listeners:[],then:function(e){o.listeners.push(e)}};null!==L.T?n(!0):o.isTransition=!1,r(o),null===(n=t.pending)?(o.next=t.pending=o,$o(t,o)):(o.next=n.next,t.pending=n.next=o)}}function $o(e,t){var n=t.action,r=t.payload,a=e.state;if(t.isTransition){var o=L.T,i={};L.T=i;try{var l=n(a,r),s=L.S;null!==s&&s(i,l),Fo(e,t,l)}catch(n){Bo(e,t,n)}finally{L.T=o}}else try{Fo(e,t,o=n(a,r))}catch(n){Bo(e,t,n)}}function Fo(e,t,n){null!==n&&"object"==typeof n&&"function"==typeof n.then?n.then((function(n){Do(e,t,n)}),(function(n){return Bo(e,t,n)})):Do(e,t,n)}function Do(e,t,n){t.status="fulfilled",t.value=n,Uo(t),e.state=n,null!==(t=e.pending)&&((n=t.next)===t?e.pending=null:(n=n.next,t.next=n,$o(e,n)))}function Bo(e,t,n){var r=e.pending;if(e.pending=null,null!==r){r=r.next;do{t.status="rejected",t.reason=n,Uo(t),t=t.next}while(t!==r)}e.action=null}function Uo(e){e=e.listeners;for(var t=0;t<e.length;t++)(0,e[t])()}function qo(e,t){return t}function Go(e,t){if(Yr){var n=nu.formState;if(null!==n){e:{var r=eo;if(Yr){if(Kr){t:{for(var a=Kr,o=Jr;8!==a.nodeType;){if(!o){a=null;break t}if(null===(a=sp(a.nextSibling))){a=null;break t}}a="F!"===(o=a.data)||"F"===o?a:null}if(a){Kr=sp(a.nextSibling),r="F!"===a.data;break e}}ta(r)}r=!1}r&&(t=n[0])}}return(n=wo()).memoizedState=n.baseState=t,r={pending:null,lanes:0,dispatch:null,lastRenderedReducer:qo,lastRenderedState:t},n.queue=r,n=wi.bind(null,eo,r),r.dispatch=n,r=Oo(!1),o=ki.bind(null,eo,!1,r.queue),a={state:t,dispatch:null,action:e,pending:null},(r=wo()).queue=a,n=zo.bind(null,eo,a,o,n),a.dispatch=n,r.memoizedState=e,[t,n,!1]}function Wo(e){return Ho(xo(),to,e)}function Ho(e,t,n){t=No(e,t,qo)[0],e=Eo(_o)[0],t="object"==typeof t&&null!==t&&"function"==typeof t.then?ko(t):t;var r=xo(),a=r.queue,o=a.dispatch;return n!==r.memoizedState&&(eo.flags|=2048,Xo(9,Vo.bind(null,a,n),{destroy:void 0},null)),[t,o,e]}function Vo(e,t){e.action=t}function Qo(e){var t=xo(),n=to;if(null!==n)return Ho(t,n,e);xo(),t=t.memoizedState;var r=(n=xo()).queue.dispatch;return n.memoizedState=e,[t,r,!1]}function Xo(e,t,n,r){return e={tag:e,create:t,inst:n,deps:r,next:null},null===(t=eo.updateQueue)&&(t={lastEffect:null,events:null,stores:null,memoCache:null},eo.updateQueue=t),null===(n=t.lastEffect)?t.lastEffect=e.next=e:(r=n.next,n.next=e,e.next=r,t.lastEffect=e),e}function Ko(){return xo().memoizedState}function Yo(e,t,n,r){var a=wo();eo.flags|=e,a.memoizedState=Xo(1|t,n,{destroy:void 0},void 0===r?null:r)}function Zo(e,t,n,r){var a=xo();r=void 0===r?null:r;var o=a.memoizedState.inst;null!==to&&null!==r&&po(r,to.memoizedState.deps)?a.memoizedState=Xo(t,n,o,r):(eo.flags|=e,a.memoizedState=Xo(1|t,n,o,r))}function Jo(e,t){Yo(8390656,8,e,t)}function ei(e,t){Zo(2048,8,e,t)}function ti(e,t){return Zo(4,2,e,t)}function ni(e,t){return Zo(4,4,e,t)}function ri(e,t){if("function"==typeof t){e=e();var n=t(e);return function(){"function"==typeof n?n():t(null)}}if(null!=t)return e=e(),t.current=e,function(){t.current=null}}function ai(e,t,n){n=null!=n?n.concat([e]):null,Zo(4,4,ri.bind(null,t,e),n)}function oi(){}function ii(e,t){var n=xo();t=void 0===t?null:t;var r=n.memoizedState;return null!==t&&po(t,r[1])?r[0]:(n.memoizedState=[e,t],e)}function li(e,t){var n=xo();t=void 0===t?null:t;var r=n.memoizedState;if(null!==t&&po(t,r[1]))return r[0];if(r=e(),oo){ve(!0);try{e()}finally{ve(!1)}}return n.memoizedState=[r,t],r}function si(e,t,n){return void 0===n||1073741824&Ja?e.memoizedState=t:(e.memoizedState=n,e=Ru(),eo.lanes|=e,du|=e,n)}function ui(e,t,n,r){return Xn(n,t)?n:null!==Ca.current?(e=si(e,n,r),Xn(e,t)||(Wi=!0),e):42&Ja?(e=Ru(),eo.lanes|=e,du|=e,t):(Wi=!0,e.memoizedState=n)}function ci(e,t,n,r,a){var o=U.p;U.p=0!==o&&8>o?o:8;var i,l,s,u=L.T,c={};L.T=c,ki(e,!1,t,n);try{var p=a(),d=L.S;null!==d&&d(c,p),null!==p&&"object"==typeof p&&"function"==typeof p.then?xi(e,t,(i=r,l=[],s={status:"pending",value:null,reason:null,then:function(e){l.push(e)}},p.then((function(){s.status="fulfilled",s.value=i;for(var e=0;e<l.length;e++)(0,l[e])(i)}),(function(e){for(s.status="rejected",s.reason=e,e=0;e<l.length;e++)(0,l[e])(void 0)})),s),Tu()):xi(e,t,r,Tu())}catch(n){xi(e,t,{then:function(){},status:"rejected",reason:n},Tu())}finally{U.p=o,L.T=u}}function pi(){}function di(e,t,n,r){if(5!==e.tag)throw Error(i(476));var a=fi(e).queue;ci(e,a,t,q,null===n?pi:function(){return hi(e),n(r)})}function fi(e){var t=e.memoizedState;if(null!==t)return t;var n={};return(t={memoizedState:q,baseState:q,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:_o,lastRenderedState:q},next:null}).next={memoizedState:n,baseState:n,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:_o,lastRenderedState:n},next:null},e.memoizedState=t,null!==(e=e.alternate)&&(e.memoizedState=t),t}function hi(e){xi(e,fi(e).next.queue,{},Tu())}function gi(){return El(zp)}function mi(){return xo().memoizedState}function yi(){return xo().memoizedState}function bi(e){for(var t=e.return;null!==t;){switch(t.tag){case 24:case 3:var n=Tu(),r=Il(t,e=Rl(n),n);return null!==r&&(Iu(r,0,n),Ml(r,t,n)),t={cache:Ba()},void(e.payload=t)}t=t.return}}function vi(e,t,n){var r=Tu();n={lane:r,revertLane:0,action:n,hasEagerState:!1,eagerState:null,next:null},Si(e)?Ci(t,n):null!==(n=Ar(e,t,n,r))&&(Iu(n,0,r),_i(n,t,r))}function wi(e,t,n){xi(e,t,n,Tu())}function xi(e,t,n,r){var a={lane:r,revertLane:0,action:n,hasEagerState:!1,eagerState:null,next:null};if(Si(e))Ci(t,a);else{var o=e.alternate;if(0===e.lanes&&(null===o||0===o.lanes)&&null!==(o=t.lastRenderedReducer))try{var i=t.lastRenderedState,l=o(i,n);if(a.hasEagerState=!0,a.eagerState=l,Xn(l,i))return Pr(e,t,a,0),null===nu&&Nr(),!1}catch(e){}if(null!==(n=Ar(e,t,a,r)))return Iu(n,0,r),_i(n,t,r),!0}return!1}function ki(e,t,n,r){if(r={lane:2,revertLane:kc(),action:r,hasEagerState:!1,eagerState:null,next:null},Si(e)){if(t)throw Error(i(479))}else null!==(t=Ar(e,n,r,2))&&Iu(t,0,2)}function Si(e){var t=e.alternate;return e===eo||null!==t&&t===eo}function Ci(e,t){ao=ro=!0;var n=e.pending;null===n?t.next=t:(t.next=n.next,n.next=t),e.pending=t}function _i(e,t,n){if(4194176&n){var r=t.lanes;n|=r&=e.pendingLanes,t.lanes=n,Me(e,n)}}var Ei={readContext:El,use:So,useCallback:co,useContext:co,useEffect:co,useImperativeHandle:co,useLayoutEffect:co,useInsertionEffect:co,useMemo:co,useReducer:co,useRef:co,useState:co,useDebugValue:co,useDeferredValue:co,useTransition:co,useSyncExternalStore:co,useId:co};Ei.useCacheRefresh=co,Ei.useMemoCache=co,Ei.useHostTransitionStatus=co,Ei.useFormState=co,Ei.useActionState=co,Ei.useOptimistic=co;var Ni={readContext:El,use:So,useCallback:function(e,t){return wo().memoizedState=[e,void 0===t?null:t],e},useContext:El,useEffect:Jo,useImperativeHandle:function(e,t,n){n=null!=n?n.concat([e]):null,Yo(4194308,4,ri.bind(null,t,e),n)},useLayoutEffect:function(e,t){return Yo(4194308,4,e,t)},useInsertionEffect:function(e,t){Yo(4,2,e,t)},useMemo:function(e,t){var n=wo();t=void 0===t?null:t;var r=e();if(oo){ve(!0);try{e()}finally{ve(!1)}}return n.memoizedState=[r,t],r},useReducer:function(e,t,n){var r=wo();if(void 0!==n){var a=n(t);if(oo){ve(!0);try{n(t)}finally{ve(!1)}}}else a=t;return r.memoizedState=r.baseState=a,e={pending:null,lanes:0,dispatch:null,lastRenderedReducer:e,lastRenderedState:a},r.queue=e,e=e.dispatch=vi.bind(null,eo,e),[r.memoizedState,e]},useRef:function(e){return e={current:e},wo().memoizedState=e},useState:function(e){var t=(e=Oo(e)).queue,n=wi.bind(null,eo,t);return t.dispatch=n,[e.memoizedState,n]},useDebugValue:oi,useDeferredValue:function(e,t){return si(wo(),e,t)},useTransition:function(){var e=Oo(!1);return e=ci.bind(null,eo,e.queue,!0,!1),wo().memoizedState=e,[!1,e]},useSyncExternalStore:function(e,t,n){var r=eo,a=wo();if(Yr){if(void 0===n)throw Error(i(407));n=n()}else{if(n=t(),null===nu)throw Error(i(349));60&au||Lo(r,t,n)}a.memoizedState=n;var o={value:n,getSnapshot:t};return a.queue=o,Jo(Ro.bind(null,r,o,e),[e]),r.flags|=2048,Xo(9,To.bind(null,r,o,n,t),{destroy:void 0},null),n},useId:function(){var e=wo(),t=nu.identifierPrefix;if(Yr){var n=Gr;t=":"+t+"R"+(n=(qr&~(1<<32-we(qr)-1)).toString(32)+n),0<(n=io++)&&(t+="H"+n.toString(32)),t+=":"}else t=":"+t+"r"+(n=uo++).toString(32)+":";return e.memoizedState=t},useCacheRefresh:function(){return wo().memoizedState=bi.bind(null,eo)}};Ni.useMemoCache=Co,Ni.useHostTransitionStatus=gi,Ni.useFormState=Go,Ni.useActionState=Go,Ni.useOptimistic=function(e){var t=wo();t.memoizedState=t.baseState=e;var n={pending:null,lanes:0,dispatch:null,lastRenderedReducer:null,lastRenderedState:null};return t.queue=n,t=ki.bind(null,eo,!0,n),n.dispatch=t,[e,t]};var Pi={readContext:El,use:So,useCallback:ii,useContext:El,useEffect:ei,useImperativeHandle:ai,useInsertionEffect:ti,useLayoutEffect:ni,useMemo:li,useReducer:Eo,useRef:Ko,useState:function(){return Eo(_o)},useDebugValue:oi,useDeferredValue:function(e,t){return ui(xo(),to.memoizedState,e,t)},useTransition:function(){var e=Eo(_o)[0],t=xo().memoizedState;return["boolean"==typeof e?e:ko(e),t]},useSyncExternalStore:Ao,useId:mi};Pi.useCacheRefresh=yi,Pi.useMemoCache=Co,Pi.useHostTransitionStatus=gi,Pi.useFormState=Wo,Pi.useActionState=Wo,Pi.useOptimistic=function(e,t){return jo(xo(),0,e,t)};var Ai={readContext:El,use:So,useCallback:ii,useContext:El,useEffect:ei,useImperativeHandle:ai,useInsertionEffect:ti,useLayoutEffect:ni,useMemo:li,useReducer:Po,useRef:Ko,useState:function(){return Po(_o)},useDebugValue:oi,useDeferredValue:function(e,t){var n=xo();return null===to?si(n,e,t):ui(n,to.memoizedState,e,t)},useTransition:function(){var e=Po(_o)[0],t=xo().memoizedState;return["boolean"==typeof e?e:ko(e),t]},useSyncExternalStore:Ao,useId:mi};function Li(e,t,n,r){n=null==(n=n(r,t=e.memoizedState))?t:T({},t,n),e.memoizedState=n,0===e.lanes&&(e.updateQueue.baseState=n)}Ai.useCacheRefresh=yi,Ai.useMemoCache=Co,Ai.useHostTransitionStatus=gi,Ai.useFormState=Qo,Ai.useActionState=Qo,Ai.useOptimistic=function(e,t){var n=xo();return null!==to?jo(n,0,e,t):(n.baseState=e,[e,n.queue.dispatch])};var Ti={isMounted:function(e){return!!(e=e._reactInternals)&&z(e)===e},enqueueSetState:function(e,t,n){e=e._reactInternals;var r=Tu(),a=Rl(r);a.payload=t,null!=n&&(a.callback=n),null!==(t=Il(e,a,r))&&(Iu(t,0,r),Ml(t,e,r))},enqueueReplaceState:function(e,t,n){e=e._reactInternals;var r=Tu(),a=Rl(r);a.tag=1,a.payload=t,null!=n&&(a.callback=n),null!==(t=Il(e,a,r))&&(Iu(t,0,r),Ml(t,e,r))},enqueueForceUpdate:function(e,t){e=e._reactInternals;var n=Tu(),r=Rl(n);r.tag=2,null!=t&&(r.callback=t),null!==(t=Il(e,r,n))&&(Iu(t,0,n),Ml(t,e,n))}};function Ri(e,t,n,r,a,o,i){return"function"==typeof(e=e.stateNode).shouldComponentUpdate?e.shouldComponentUpdate(r,o,i):!(t.prototype&&t.prototype.isPureReactComponent&&Kn(n,r)&&Kn(a,o))}function Ii(e,t,n,r){e=t.state,"function"==typeof t.componentWillReceiveProps&&t.componentWillReceiveProps(n,r),"function"==typeof t.UNSAFE_componentWillReceiveProps&&t.UNSAFE_componentWillReceiveProps(n,r),t.state!==e&&Ti.enqueueReplaceState(t,t.state,null)}function Mi(e,t){var n=t;if("ref"in t)for(var r in n={},t)"ref"!==r&&(n[r]=t[r]);if(e=e.defaultProps)for(var a in n===t&&(n=T({},n)),e)void 0===n[a]&&(n[a]=e[a]);return n}var Oi="function"==typeof reportError?reportError:function(e){if("object"==typeof window&&"function"==typeof window.ErrorEvent){var t=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:"object"==typeof e&&null!==e&&"string"==typeof e.message?String(e.message):String(e),error:e});if(!window.dispatchEvent(t))return}else if("object"==typeof process&&"function"==typeof process.emit)return void process.emit("uncaughtException",e);console.error(e)};function ji(e){Oi(e)}function zi(e){console.error(e)}function $i(e){Oi(e)}function Fi(e,t){try{(0,e.onUncaughtError)(t.value,{componentStack:t.stack})}catch(e){setTimeout((function(){throw e}))}}function Di(e,t,n){try{(0,e.onCaughtError)(n.value,{componentStack:n.stack,errorBoundary:1===t.tag?t.stateNode:null})}catch(e){setTimeout((function(){throw e}))}}function Bi(e,t,n){return(n=Rl(n)).tag=3,n.payload={element:null},n.callback=function(){Fi(e,t)},n}function Ui(e){return(e=Rl(e)).tag=3,e}function qi(e,t,n,r){var a=n.type.getDerivedStateFromError;if("function"==typeof a){var o=r.value;e.payload=function(){return a(o)},e.callback=function(){Di(t,n,r)}}var i=n.stateNode;null!==i&&"function"==typeof i.componentDidCatch&&(e.callback=function(){Di(t,n,r),"function"!=typeof a&&(null===Su?Su=new Set([this]):Su.add(this));var e=r.stack;this.componentDidCatch(r.value,{componentStack:null!==e?e:""})})}var Gi=Error(i(461)),Wi=!1;function Hi(e,t,n,r){t.child=null===e?Sa(t,null,n,r):ka(t,e.child,n,r)}function Vi(e,t,n,r,a){n=n.render;var o=t.ref;if("ref"in r){var i={};for(var l in r)"ref"!==l&&(i[l]=r[l])}else i=r;return _l(t),r=fo(e,t,n,i,o,a),l=yo(),null===e||Wi?(Yr&&l&&Vr(t),t.flags|=1,Hi(e,t,r,a),t.child):(bo(e,t,a),fl(e,t,a))}function Qi(e,t,n,r,a){if(null===e){var o=n.type;return"function"!=typeof o||zs(o)||void 0!==o.defaultProps||null!==n.compare?((e=Ds(n.type,null,r,t,t.mode,a)).ref=t.ref,e.return=t,t.child=e):(t.tag=15,t.type=o,Xi(e,t,o,r,a))}if(o=e.child,!hl(e,a)){var i=o.memoizedProps;if((n=null!==(n=n.compare)?n:Kn)(i,r)&&e.ref===t.ref)return fl(e,t,a)}return t.flags|=1,(e=$s(o,r)).ref=t.ref,e.return=t,t.child=e}function Xi(e,t,n,r,a){if(null!==e){var o=e.memoizedProps;if(Kn(o,r)&&e.ref===t.ref){if(Wi=!1,t.pendingProps=r=o,!hl(e,a))return t.lanes=e.lanes,fl(e,t,a);131072&e.flags&&(Wi=!0)}}return Ji(e,t,n,r,a)}function Ki(e,t,n){var r=t.pendingProps,a=r.children,o=!!(2&t.stateNode._pendingVisibility),i=null!==e?e.memoizedState:null;if(Zi(e,t),"hidden"===r.mode||o){if(128&t.flags){if(r=null!==i?i.baseLanes|n:n,null!==e){for(a=t.child=e.child,o=0;null!==a;)o=o|a.lanes|a.childLanes,a=a.sibling;t.childLanes=o&~r}else t.childLanes=0,t.child=null;return Yi(e,t,r,n)}if(!(536870912&n))return t.lanes=t.childLanes=536870912,Yi(e,t,null!==i?i.baseLanes|n:n,n);t.memoizedState={baseLanes:0,cachePool:null},null!==e&&Ya(0,null!==i?i.cachePool:null),null!==i?Ea(t,i):Na(),Ra(t)}else null!==i?(Ya(0,i.cachePool),Ea(t,i),Ia(),t.memoizedState=null):(null!==e&&Ya(0,null),Na(),Ia());return Hi(e,t,a,n),t.child}function Yi(e,t,n,r){var a=Ka();return a=null===a?null:{parent:Da._currentValue,pool:a},t.memoizedState={baseLanes:n,cachePool:a},null!==e&&Ya(0,null),Na(),Ra(t),null!==e&&Sl(e,t,r,!0),null}function Zi(e,t){var n=t.ref;if(null===n)null!==e&&null!==e.ref&&(t.flags|=2097664);else{if("function"!=typeof n&&"object"!=typeof n)throw Error(i(284));null!==e&&e.ref===n||(t.flags|=2097664)}}function Ji(e,t,n,r,a){return _l(t),n=fo(e,t,n,r,void 0,a),r=yo(),null===e||Wi?(Yr&&r&&Vr(t),t.flags|=1,Hi(e,t,n,a),t.child):(bo(e,t,a),fl(e,t,a))}function el(e,t,n,r,a,o){return _l(t),t.updateQueue=null,n=go(t,r,n,a),ho(e),r=yo(),null===e||Wi?(Yr&&r&&Vr(t),t.flags|=1,Hi(e,t,n,o),t.child):(bo(e,t,o),fl(e,t,o))}function tl(e,t,n,r,a){if(_l(t),null===t.stateNode){var o=Ir,i=n.contextType;"object"==typeof i&&null!==i&&(o=El(i)),o=new n(r,o),t.memoizedState=null!==o.state&&void 0!==o.state?o.state:null,o.updater=Ti,t.stateNode=o,o._reactInternals=t,(o=t.stateNode).props=r,o.state=t.memoizedState,o.refs={},Ll(t),i=n.contextType,o.context="object"==typeof i&&null!==i?El(i):Ir,o.state=t.memoizedState,"function"==typeof(i=n.getDerivedStateFromProps)&&(Li(t,n,i,r),o.state=t.memoizedState),"function"==typeof n.getDerivedStateFromProps||"function"==typeof o.getSnapshotBeforeUpdate||"function"!=typeof o.UNSAFE_componentWillMount&&"function"!=typeof o.componentWillMount||(i=o.state,"function"==typeof o.componentWillMount&&o.componentWillMount(),"function"==typeof o.UNSAFE_componentWillMount&&o.UNSAFE_componentWillMount(),i!==o.state&&Ti.enqueueReplaceState(o,o.state,null),$l(t,r,o,a),zl(),o.state=t.memoizedState),"function"==typeof o.componentDidMount&&(t.flags|=4194308),r=!0}else if(null===e){o=t.stateNode;var l=t.memoizedProps,s=Mi(n,l);o.props=s;var u=o.context,c=n.contextType;i=Ir,"object"==typeof c&&null!==c&&(i=El(c));var p=n.getDerivedStateFromProps;c="function"==typeof p||"function"==typeof o.getSnapshotBeforeUpdate,l=t.pendingProps!==l,c||"function"!=typeof o.UNSAFE_componentWillReceiveProps&&"function"!=typeof o.componentWillReceiveProps||(l||u!==i)&&Ii(t,o,r,i),Al=!1;var d=t.memoizedState;o.state=d,$l(t,r,o,a),zl(),u=t.memoizedState,l||d!==u||Al?("function"==typeof p&&(Li(t,n,p,r),u=t.memoizedState),(s=Al||Ri(t,n,s,r,d,u,i))?(c||"function"!=typeof o.UNSAFE_componentWillMount&&"function"!=typeof o.componentWillMount||("function"==typeof o.componentWillMount&&o.componentWillMount(),"function"==typeof o.UNSAFE_componentWillMount&&o.UNSAFE_componentWillMount()),"function"==typeof o.componentDidMount&&(t.flags|=4194308)):("function"==typeof o.componentDidMount&&(t.flags|=4194308),t.memoizedProps=r,t.memoizedState=u),o.props=r,o.state=u,o.context=i,r=s):("function"==typeof o.componentDidMount&&(t.flags|=4194308),r=!1)}else{o=t.stateNode,Tl(e,t),c=Mi(n,i=t.memoizedProps),o.props=c,p=t.pendingProps,d=o.context,u=n.contextType,s=Ir,"object"==typeof u&&null!==u&&(s=El(u)),(u="function"==typeof(l=n.getDerivedStateFromProps)||"function"==typeof o.getSnapshotBeforeUpdate)||"function"!=typeof o.UNSAFE_componentWillReceiveProps&&"function"!=typeof o.componentWillReceiveProps||(i!==p||d!==s)&&Ii(t,o,r,s),Al=!1,d=t.memoizedState,o.state=d,$l(t,r,o,a),zl();var f=t.memoizedState;i!==p||d!==f||Al||null!==e&&null!==e.dependencies&&Cl(e.dependencies)?("function"==typeof l&&(Li(t,n,l,r),f=t.memoizedState),(c=Al||Ri(t,n,c,r,d,f,s)||null!==e&&null!==e.dependencies&&Cl(e.dependencies))?(u||"function"!=typeof o.UNSAFE_componentWillUpdate&&"function"!=typeof o.componentWillUpdate||("function"==typeof o.componentWillUpdate&&o.componentWillUpdate(r,f,s),"function"==typeof o.UNSAFE_componentWillUpdate&&o.UNSAFE_componentWillUpdate(r,f,s)),"function"==typeof o.componentDidUpdate&&(t.flags|=4),"function"==typeof o.getSnapshotBeforeUpdate&&(t.flags|=1024)):("function"!=typeof o.componentDidUpdate||i===e.memoizedProps&&d===e.memoizedState||(t.flags|=4),"function"!=typeof o.getSnapshotBeforeUpdate||i===e.memoizedProps&&d===e.memoizedState||(t.flags|=1024),t.memoizedProps=r,t.memoizedState=f),o.props=r,o.state=f,o.context=s,r=c):("function"!=typeof o.componentDidUpdate||i===e.memoizedProps&&d===e.memoizedState||(t.flags|=4),"function"!=typeof o.getSnapshotBeforeUpdate||i===e.memoizedProps&&d===e.memoizedState||(t.flags|=1024),r=!1)}return o=r,Zi(e,t),r=!!(128&t.flags),o||r?(o=t.stateNode,n=r&&"function"!=typeof n.getDerivedStateFromError?null:o.render(),t.flags|=1,null!==e&&r?(t.child=ka(t,e.child,null,a),t.child=ka(t,null,n,a)):Hi(e,t,n,a),t.memoizedState=o.state,e=t.child):e=fl(e,t,a),e}function nl(e,t,n,r){return oa(),t.flags|=256,Hi(e,t,n,r),t.child}var rl={dehydrated:null,treeContext:null,retryLane:0};function al(e){return{baseLanes:e,cachePool:Za()}}function ol(e,t,n){return e=null!==e?e.childLanes&~n:0,t&&(e|=gu),e}function il(e,t,n){var r,a=t.pendingProps,o=!1,l=!!(128&t.flags);if((r=l)||(r=(null===e||null!==e.memoizedState)&&!!(2&Oa.current)),r&&(o=!0,t.flags&=-129),r=!!(32&t.flags),t.flags&=-33,null===e){if(Yr){if(o?Ta(t):Ia(),Yr){var s,u=Kr;if(s=u){e:{for(s=u,u=Jr;8!==s.nodeType;){if(!u){u=null;break e}if(null===(s=sp(s.nextSibling))){u=null;break e}}u=s}null!==u?(t.memoizedState={dehydrated:u,treeContext:null!==Ur?{id:qr,overflow:Gr}:null,retryLane:536870912},(s=js(18,null,null,0)).stateNode=u,s.return=t,t.child=s,Xr=t,Kr=null,s=!0):s=!1}s||ta(t)}if(null!==(u=t.memoizedState)&&null!==(u=u.dehydrated))return"$!"===u.data?t.lanes=16:t.lanes=536870912,null;Ma(t)}return u=a.children,a=a.fallback,o?(Ia(),u=sl({mode:"hidden",children:u},o=t.mode),a=Bs(a,o,n,null),u.return=t,a.return=t,u.sibling=a,t.child=u,(o=t.child).memoizedState=al(n),o.childLanes=ol(e,r,n),t.memoizedState=rl,a):(Ta(t),ll(t,u))}if(null!==(s=e.memoizedState)&&null!==(u=s.dehydrated)){if(l)256&t.flags?(Ta(t),t.flags&=-257,t=ul(e,t,n)):null!==t.memoizedState?(Ia(),t.child=e.child,t.flags|=128,t=null):(Ia(),o=a.fallback,u=t.mode,a=sl({mode:"visible",children:a.children},u),(o=Bs(o,u,n,null)).flags|=2,a.return=t,o.return=t,a.sibling=o,t.child=a,ka(t,e.child,null,n),(a=t.child).memoizedState=al(n),a.childLanes=ol(e,r,n),t.memoizedState=rl,t=o);else if(Ta(t),"$!"===u.data){if(r=u.nextSibling&&u.nextSibling.dataset)var c=r.dgst;r=c,(a=Error(i(419))).stack="",a.digest=r,ia({value:a,source:null,stack:null}),t=ul(e,t,n)}else if(Wi||Sl(e,t,n,!1),r=!!(n&e.childLanes),Wi||r){if(null!==(r=nu)){if(42&(a=n&-n))a=1;else switch(a){case 2:a=1;break;case 8:a=4;break;case 32:a=16;break;case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:a=64;break;case 268435456:a=134217728;break;default:a=0}if(0!==(a=a&(r.suspendedLanes|n)?0:a)&&a!==s.retryLane)throw s.retryLane=a,Lr(e,a),Iu(r,0,a),Gi}"$?"===u.data||Wu(),t=ul(e,t,n)}else"$?"===u.data?(t.flags|=128,t.child=e.child,t=sc.bind(null,e),u._reactRetry=t,t=null):(e=s.treeContext,Kr=sp(u.nextSibling),Xr=t,Yr=!0,Zr=null,Jr=!1,null!==e&&(Dr[Br++]=qr,Dr[Br++]=Gr,Dr[Br++]=Ur,qr=e.id,Gr=e.overflow,Ur=t),(t=ll(t,a.children)).flags|=4096);return t}return o?(Ia(),o=a.fallback,u=t.mode,c=(s=e.child).sibling,(a=$s(s,{mode:"hidden",children:a.children})).subtreeFlags=31457280&s.subtreeFlags,null!==c?o=$s(c,o):(o=Bs(o,u,n,null)).flags|=2,o.return=t,a.return=t,a.sibling=o,t.child=a,a=o,o=t.child,null===(u=e.child.memoizedState)?u=al(n):(null!==(s=u.cachePool)?(c=Da._currentValue,s=s.parent!==c?{parent:c,pool:c}:s):s=Za(),u={baseLanes:u.baseLanes|n,cachePool:s}),o.memoizedState=u,o.childLanes=ol(e,r,n),t.memoizedState=rl,a):(Ta(t),e=(n=e.child).sibling,(n=$s(n,{mode:"visible",children:a.children})).return=t,n.sibling=null,null!==e&&(null===(r=t.deletions)?(t.deletions=[e],t.flags|=16):r.push(e)),t.child=n,t.memoizedState=null,n)}function ll(e,t){return(t=sl({mode:"visible",children:t},e.mode)).return=e,e.child=t}function sl(e,t){return Us(e,t,0,null)}function ul(e,t,n){return ka(t,e.child,null,n),(e=ll(t,t.pendingProps.children)).flags|=2,t.memoizedState=null,e}function cl(e,t,n){e.lanes|=t;var r=e.alternate;null!==r&&(r.lanes|=t),xl(e.return,t,n)}function pl(e,t,n,r,a){var o=e.memoizedState;null===o?e.memoizedState={isBackwards:t,rendering:null,renderingStartTime:0,last:r,tail:n,tailMode:a}:(o.isBackwards=t,o.rendering=null,o.renderingStartTime=0,o.last=r,o.tail=n,o.tailMode=a)}function dl(e,t,n){var r=t.pendingProps,a=r.revealOrder,o=r.tail;if(Hi(e,t,r.children,n),2&(r=Oa.current))r=1&r|2,t.flags|=128;else{if(null!==e&&128&e.flags)e:for(e=t.child;null!==e;){if(13===e.tag)null!==e.memoizedState&&cl(e,n,t);else if(19===e.tag)cl(e,n,t);else if(null!==e.child){e.child.return=e,e=e.child;continue}if(e===t)break e;for(;null===e.sibling;){if(null===e.return||e.return===t)break e;e=e.return}e.sibling.return=e.return,e=e.sibling}r&=1}switch(Q(Oa,r),a){case"forwards":for(n=t.child,a=null;null!==n;)null!==(e=n.alternate)&&null===ja(e)&&(a=n),n=n.sibling;null===(n=a)?(a=t.child,t.child=null):(a=n.sibling,n.sibling=null),pl(t,!1,a,n,o);break;case"backwards":for(n=null,a=t.child,t.child=null;null!==a;){if(null!==(e=a.alternate)&&null===ja(e)){t.child=a;break}e=a.sibling,a.sibling=n,n=a,a=e}pl(t,!0,n,null,o);break;case"together":pl(t,!1,null,null,void 0);break;default:t.memoizedState=null}return t.child}function fl(e,t,n){if(null!==e&&(t.dependencies=e.dependencies),du|=t.lanes,!(n&t.childLanes)){if(null===e)return null;if(Sl(e,t,n,!1),!(n&t.childLanes))return null}if(null!==e&&t.child!==e.child)throw Error(i(153));if(null!==t.child){for(n=$s(e=t.child,e.pendingProps),t.child=n,n.return=t;null!==e.sibling;)e=e.sibling,(n=n.sibling=$s(e,e.pendingProps)).return=t;n.sibling=null}return t.child}function hl(e,t){return!!(e.lanes&t)||!(null===(e=e.dependencies)||!Cl(e))}function gl(e,t,n){if(null!==e)if(e.memoizedProps!==t.pendingProps)Wi=!0;else{if(!(hl(e,n)||128&t.flags))return Wi=!1,function(e,t,n){switch(t.tag){case 3:J(t,t.stateNode.containerInfo),vl(0,Da,e.memoizedState.cache),oa();break;case 27:case 5:te(t);break;case 4:J(t,t.stateNode.containerInfo);break;case 10:vl(0,t.type,t.memoizedProps.value);break;case 13:var r=t.memoizedState;if(null!==r)return null!==r.dehydrated?(Ta(t),t.flags|=128,null):n&t.child.childLanes?il(e,t,n):(Ta(t),null!==(e=fl(e,t,n))?e.sibling:null);Ta(t);break;case 19:var a=!!(128&e.flags);if((r=!!(n&t.childLanes))||(Sl(e,t,n,!1),r=!!(n&t.childLanes)),a){if(r)return dl(e,t,n);t.flags|=128}if(null!==(a=t.memoizedState)&&(a.rendering=null,a.tail=null,a.lastEffect=null),Q(Oa,Oa.current),r)break;return null;case 22:case 23:return t.lanes=0,Ki(e,t,n);case 24:vl(0,Da,e.memoizedState.cache)}return fl(e,t,n)}(e,t,n);Wi=!!(131072&e.flags)}else Wi=!1,Yr&&1048576&t.flags&&Hr(t,Fr,t.index);switch(t.lanes=0,t.tag){case 16:e:{e=t.pendingProps;var r=t.elementType,a=r._init;if(r=a(r._payload),t.type=r,"function"!=typeof r){if(null!=r){if((a=r.$$typeof)===y){t.tag=11,t=Vi(null,t,r,e,n);break e}if(a===w){t.tag=14,t=Qi(null,t,r,e,n);break e}}throw t=N(r)||r,Error(i(306,t,""))}zs(r)?(e=Mi(r,e),t.tag=1,t=tl(null,t,r,e,n)):(t.tag=0,t=Ji(null,t,r,e,n))}return t;case 0:return Ji(e,t,t.type,t.pendingProps,n);case 1:return tl(e,t,r=t.type,a=Mi(r,t.pendingProps),n);case 3:e:{if(J(t,t.stateNode.containerInfo),null===e)throw Error(i(387));var o=t.pendingProps;r=(a=t.memoizedState).element,Tl(e,t),$l(t,o,null,n);var l=t.memoizedState;if(o=l.cache,vl(0,Da,o),o!==a.cache&&kl(t,[Da],n,!0),zl(),o=l.element,a.isDehydrated){if(a={element:o,isDehydrated:!1,cache:l.cache},t.updateQueue.baseState=a,t.memoizedState=a,256&t.flags){t=nl(e,t,o,n);break e}if(o!==r){ia(r=Or(Error(i(424)),t)),t=nl(e,t,o,n);break e}for(Kr=sp(t.stateNode.containerInfo.firstChild),Xr=t,Yr=!0,Zr=null,Jr=!0,n=Sa(t,null,o,n),t.child=n;n;)n.flags=-3&n.flags|4096,n=n.sibling}else{if(oa(),o===r){t=fl(e,t,n);break e}Hi(e,t,o,n)}t=t.child}return t;case 26:return Zi(e,t),null===e?(n=yp(t.type,null,t.pendingProps,null))?t.memoizedState=n:Yr||(n=t.type,e=t.pendingProps,(r=Kc(Y.current).createElement(n))[$e]=t,r[Fe]=e,Vc(r,n,e),Ye(r),t.stateNode=r):t.memoizedState=yp(t.type,e.memoizedProps,t.pendingProps,e.memoizedState),null;case 27:return te(t),null===e&&Yr&&(r=t.stateNode=cp(t.type,t.pendingProps,Y.current),Xr=t,Jr=!0,Kr=sp(r.firstChild)),r=t.pendingProps.children,null!==e||Yr?Hi(e,t,r,n):t.child=ka(t,null,r,n),Zi(e,t),t.child;case 5:return null===e&&Yr&&((a=r=Kr)&&(null!==(r=function(e,t,n,r){for(;1===e.nodeType;){var a=n;if(e.nodeName.toLowerCase()!==t.toLowerCase()){if(!r&&("INPUT"!==e.nodeName||"hidden"!==e.type))break}else if(r){if(!e[We])switch(t){case"meta":if(!e.hasAttribute("itemprop"))break;return e;case"link":if("stylesheet"===(o=e.getAttribute("rel"))&&e.hasAttribute("data-precedence"))break;if(o!==a.rel||e.getAttribute("href")!==(null==a.href?null:a.href)||e.getAttribute("crossorigin")!==(null==a.crossOrigin?null:a.crossOrigin)||e.getAttribute("title")!==(null==a.title?null:a.title))break;return e;case"style":if(e.hasAttribute("data-precedence"))break;return e;case"script":if(((o=e.getAttribute("src"))!==(null==a.src?null:a.src)||e.getAttribute("type")!==(null==a.type?null:a.type)||e.getAttribute("crossorigin")!==(null==a.crossOrigin?null:a.crossOrigin))&&o&&e.hasAttribute("async")&&!e.hasAttribute("itemprop"))break;return e;default:return e}}else{if("input"!==t||"hidden"!==e.type)return e;var o=null==a.name?null:""+a.name;if("hidden"===a.type&&e.getAttribute("name")===o)return e}if(null===(e=sp(e.nextSibling)))break}return null}(r,t.type,t.pendingProps,Jr))?(t.stateNode=r,Xr=t,Kr=sp(r.firstChild),Jr=!1,a=!0):a=!1),a||ta(t)),te(t),a=t.type,o=t.pendingProps,l=null!==e?e.memoizedProps:null,r=o.children,Jc(a,o)?r=null:null!==l&&Jc(a,l)&&(t.flags|=32),null!==t.memoizedState&&(a=fo(e,t,mo,null,null,n),zp._currentValue=a),Zi(e,t),Hi(e,t,r,n),t.child;case 6:return null===e&&Yr&&((e=n=Kr)&&(null!==(n=function(e,t,n){if(""===t)return null;for(;3!==e.nodeType;){if((1!==e.nodeType||"INPUT"!==e.nodeName||"hidden"!==e.type)&&!n)return null;if(null===(e=sp(e.nextSibling)))return null}return e}(n,t.pendingProps,Jr))?(t.stateNode=n,Xr=t,Kr=null,e=!0):e=!1),e||ta(t)),null;case 13:return il(e,t,n);case 4:return J(t,t.stateNode.containerInfo),r=t.pendingProps,null===e?t.child=ka(t,null,r,n):Hi(e,t,r,n),t.child;case 11:return Vi(e,t,t.type,t.pendingProps,n);case 7:return Hi(e,t,t.pendingProps,n),t.child;case 8:case 12:return Hi(e,t,t.pendingProps.children,n),t.child;case 10:return r=t.pendingProps,vl(0,t.type,r.value),Hi(e,t,r.children,n),t.child;case 9:return a=t.type._context,r=t.pendingProps.children,_l(t),r=r(a=El(a)),t.flags|=1,Hi(e,t,r,n),t.child;case 14:return Qi(e,t,t.type,t.pendingProps,n);case 15:return Xi(e,t,t.type,t.pendingProps,n);case 19:return dl(e,t,n);case 22:return Ki(e,t,n);case 24:return _l(t),r=El(Da),null===e?(null===(a=Ka())&&(a=nu,o=Ba(),a.pooledCache=o,o.refCount++,null!==o&&(a.pooledCacheLanes|=n),a=o),t.memoizedState={parent:r,cache:a},Ll(t),vl(0,Da,a)):(!!(e.lanes&n)&&(Tl(e,t),$l(t,null,null,n),zl()),a=e.memoizedState,o=t.memoizedState,a.parent!==r?(a={parent:r,cache:r},t.memoizedState=a,0===t.lanes&&(t.memoizedState=t.updateQueue.baseState=a),vl(0,Da,r)):(r=o.cache,vl(0,Da,r),r!==a.cache&&kl(t,[Da],n,!0))),Hi(e,t,t.pendingProps.children,n),t.child;case 29:throw t.pendingProps}throw Error(i(156,t.tag))}var ml=H(null),yl=null,bl=null;function vl(e,t,n){Q(ml,t._currentValue),t._currentValue=n}function wl(e){e._currentValue=ml.current,V(ml)}function xl(e,t,n){for(;null!==e;){var r=e.alternate;if((e.childLanes&t)!==t?(e.childLanes|=t,null!==r&&(r.childLanes|=t)):null!==r&&(r.childLanes&t)!==t&&(r.childLanes|=t),e===n)break;e=e.return}}function kl(e,t,n,r){var a=e.child;for(null!==a&&(a.return=e);null!==a;){var o=a.dependencies;if(null!==o){var l=a.child;o=o.firstContext;e:for(;null!==o;){var s=o;o=a;for(var u=0;u<t.length;u++)if(s.context===t[u]){o.lanes|=n,null!==(s=o.alternate)&&(s.lanes|=n),xl(o.return,n,e),r||(l=null);break e}o=s.next}}else if(18===a.tag){if(null===(l=a.return))throw Error(i(341));l.lanes|=n,null!==(o=l.alternate)&&(o.lanes|=n),xl(l,n,e),l=null}else l=a.child;if(null!==l)l.return=a;else for(l=a;null!==l;){if(l===e){l=null;break}if(null!==(a=l.sibling)){a.return=l.return,l=a;break}l=l.return}a=l}}function Sl(e,t,n,r){e=null;for(var a=t,o=!1;null!==a;){if(!o)if(524288&a.flags)o=!0;else if(262144&a.flags)break;if(10===a.tag){var l=a.alternate;if(null===l)throw Error(i(387));if(null!==(l=l.memoizedProps)){var s=a.type;Xn(a.pendingProps.value,l.value)||(null!==e?e.push(s):e=[s])}}else if(a===Z.current){if(null===(l=a.alternate))throw Error(i(387));l.memoizedState.memoizedState!==a.memoizedState.memoizedState&&(null!==e?e.push(zp):e=[zp])}a=a.return}null!==e&&kl(t,e,n,r),t.flags|=262144}function Cl(e){for(e=e.firstContext;null!==e;){if(!Xn(e.context._currentValue,e.memoizedValue))return!0;e=e.next}return!1}function _l(e){yl=e,bl=null,null!==(e=e.dependencies)&&(e.firstContext=null)}function El(e){return Pl(yl,e)}function Nl(e,t){return null===yl&&_l(e),Pl(e,t)}function Pl(e,t){var n=t._currentValue;if(t={context:t,memoizedValue:n,next:null},null===bl){if(null===e)throw Error(i(308));bl=t,e.dependencies={lanes:0,firstContext:t},e.flags|=524288}else bl=bl.next=t;return n}var Al=!1;function Ll(e){e.updateQueue={baseState:e.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,lanes:0,hiddenCallbacks:null},callbacks:null}}function Tl(e,t){e=e.updateQueue,t.updateQueue===e&&(t.updateQueue={baseState:e.baseState,firstBaseUpdate:e.firstBaseUpdate,lastBaseUpdate:e.lastBaseUpdate,shared:e.shared,callbacks:null})}function Rl(e){return{lane:e,tag:0,payload:null,callback:null,next:null}}function Il(e,t,n){var r=e.updateQueue;if(null===r)return null;if(r=r.shared,2&tu){var a=r.pending;return null===a?t.next=t:(t.next=a.next,a.next=t),r.pending=t,t=Rr(e),Tr(e,null,n),t}return Pr(e,r,t,n),Rr(e)}function Ml(e,t,n){if(null!==(t=t.updateQueue)&&(t=t.shared,4194176&n)){var r=t.lanes;n|=r&=e.pendingLanes,t.lanes=n,Me(e,n)}}function Ol(e,t){var n=e.updateQueue,r=e.alternate;if(null!==r&&n===(r=r.updateQueue)){var a=null,o=null;if(null!==(n=n.firstBaseUpdate)){do{var i={lane:n.lane,tag:n.tag,payload:n.payload,callback:null,next:null};null===o?a=o=i:o=o.next=i,n=n.next}while(null!==n);null===o?a=o=t:o=o.next=t}else a=o=t;return n={baseState:r.baseState,firstBaseUpdate:a,lastBaseUpdate:o,shared:r.shared,callbacks:r.callbacks},void(e.updateQueue=n)}null===(e=n.lastBaseUpdate)?n.firstBaseUpdate=t:e.next=t,n.lastBaseUpdate=t}var jl=!1;function zl(){if(jl&&null!==Ha)throw Ha}function $l(e,t,n,r){jl=!1;var a=e.updateQueue;Al=!1;var o=a.firstBaseUpdate,i=a.lastBaseUpdate,l=a.shared.pending;if(null!==l){a.shared.pending=null;var s=l,u=s.next;s.next=null,null===i?o=u:i.next=u,i=s;var c=e.alternate;null!==c&&(l=(c=c.updateQueue).lastBaseUpdate)!==i&&(null===l?c.firstBaseUpdate=u:l.next=u,c.lastBaseUpdate=s)}if(null!==o){var p=a.baseState;for(i=0,c=u=s=null,l=o;;){var d=-536870913&l.lane,f=d!==l.lane;if(f?(au&d)===d:(r&d)===d){0!==d&&d===Wa&&(jl=!0),null!==c&&(c=c.next={lane:0,tag:l.tag,payload:l.payload,callback:null,next:null});e:{var h=e,g=l;d=t;var m=n;switch(g.tag){case 1:if("function"==typeof(h=g.payload)){p=h.call(m,p,d);break e}p=h;break e;case 3:h.flags=-65537&h.flags|128;case 0:if(null==(d="function"==typeof(h=g.payload)?h.call(m,p,d):h))break e;p=T({},p,d);break e;case 2:Al=!0}}null!==(d=l.callback)&&(e.flags|=64,f&&(e.flags|=8192),null===(f=a.callbacks)?a.callbacks=[d]:f.push(d))}else f={lane:d,tag:l.tag,payload:l.payload,callback:l.callback,next:null},null===c?(u=c=f,s=p):c=c.next=f,i|=d;if(null===(l=l.next)){if(null===(l=a.shared.pending))break;l=(f=l).next,f.next=null,a.lastBaseUpdate=f,a.shared.pending=null}}null===c&&(s=p),a.baseState=s,a.firstBaseUpdate=u,a.lastBaseUpdate=c,null===o&&(a.shared.lanes=0),du|=i,e.lanes=i,e.memoizedState=p}}function Fl(e,t){if("function"!=typeof e)throw Error(i(191,e));e.call(t)}function Dl(e,t){var n=e.callbacks;if(null!==n)for(e.callbacks=null,e=0;e<n.length;e++)Fl(n[e],t)}function Bl(e,t){try{var n=t.updateQueue,r=null!==n?n.lastEffect:null;if(null!==r){var a=r.next;n=a;do{if((n.tag&e)===e){r=void 0;var o=n.create,i=n.inst;r=o(),i.destroy=r}n=n.next}while(n!==a)}}catch(e){ac(t,t.return,e)}}function Ul(e,t,n){try{var r=t.updateQueue,a=null!==r?r.lastEffect:null;if(null!==a){var o=a.next;r=o;do{if((r.tag&e)===e){var i=r.inst,l=i.destroy;if(void 0!==l){i.destroy=void 0,a=t;var s=n;try{l()}catch(e){ac(a,s,e)}}}r=r.next}while(r!==o)}}catch(e){ac(t,t.return,e)}}function ql(e){var t=e.updateQueue;if(null!==t){var n=e.stateNode;try{Dl(t,n)}catch(t){ac(e,e.return,t)}}}function Gl(e,t,n){n.props=Mi(e.type,e.memoizedProps),n.state=e.memoizedState;try{n.componentWillUnmount()}catch(n){ac(e,t,n)}}function Wl(e,t){try{var n=e.ref;if(null!==n){var r=e.stateNode;switch(e.tag){case 26:case 27:case 5:var a=r;break;default:a=r}"function"==typeof n?e.refCleanup=n(a):n.current=a}}catch(n){ac(e,t,n)}}function Hl(e,t){var n=e.ref,r=e.refCleanup;if(null!==n)if("function"==typeof r)try{r()}catch(n){ac(e,t,n)}finally{e.refCleanup=null,null!=(e=e.alternate)&&(e.refCleanup=null)}else if("function"==typeof n)try{n(null)}catch(n){ac(e,t,n)}else n.current=null}function Vl(e){var t=e.type,n=e.memoizedProps,r=e.stateNode;try{e:switch(t){case"button":case"input":case"select":case"textarea":n.autoFocus&&r.focus();break e;case"img":n.src?r.src=n.src:n.srcSet&&(r.srcset=n.srcSet)}}catch(t){ac(e,e.return,t)}}function Ql(e,t,n){try{var r=e.stateNode;!function(e,t,n,r){switch(t){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"input":var a=null,o=null,l=null,s=null,u=null,c=null,p=null;for(h in n){var d=n[h];if(n.hasOwnProperty(h)&&null!=d)switch(h){case"checked":case"value":break;case"defaultValue":u=d;default:r.hasOwnProperty(h)||Wc(e,t,h,null,r,d)}}for(var f in r){var h=r[f];if(d=n[f],r.hasOwnProperty(f)&&(null!=h||null!=d))switch(f){case"type":o=h;break;case"name":a=h;break;case"checked":c=h;break;case"defaultChecked":p=h;break;case"value":l=h;break;case"defaultValue":s=h;break;case"children":case"dangerouslySetInnerHTML":if(null!=h)throw Error(i(137,t));break;default:h!==d&&Wc(e,t,f,h,r,d)}}return void mt(e,l,s,u,c,p,o,a);case"select":for(o in h=l=s=f=null,n)if(u=n[o],n.hasOwnProperty(o)&&null!=u)switch(o){case"value":break;case"multiple":h=u;default:r.hasOwnProperty(o)||Wc(e,t,o,null,r,u)}for(a in r)if(o=r[a],u=n[a],r.hasOwnProperty(a)&&(null!=o||null!=u))switch(a){case"value":f=o;break;case"defaultValue":s=o;break;case"multiple":l=o;default:o!==u&&Wc(e,t,a,o,r,u)}return t=s,n=l,r=h,void(null!=f?vt(e,!!n,f,!1):!!r!=!!n&&(null!=t?vt(e,!!n,t,!0):vt(e,!!n,n?[]:"",!1)));case"textarea":for(s in h=f=null,n)if(a=n[s],n.hasOwnProperty(s)&&null!=a&&!r.hasOwnProperty(s))switch(s){case"value":case"children":break;default:Wc(e,t,s,null,r,a)}for(l in r)if(a=r[l],o=n[l],r.hasOwnProperty(l)&&(null!=a||null!=o))switch(l){case"value":f=a;break;case"defaultValue":h=a;break;case"children":break;case"dangerouslySetInnerHTML":if(null!=a)throw Error(i(91));break;default:a!==o&&Wc(e,t,l,a,r,o)}return void wt(e,f,h);case"option":for(var g in n)f=n[g],n.hasOwnProperty(g)&&null!=f&&!r.hasOwnProperty(g)&&("selected"===g?e.selected=!1:Wc(e,t,g,null,r,f));for(u in r)f=r[u],h=n[u],!r.hasOwnProperty(u)||f===h||null==f&&null==h||("selected"===u?e.selected=f&&"function"!=typeof f&&"symbol"!=typeof f:Wc(e,t,u,f,r,h));return;case"img":case"link":case"area":case"base":case"br":case"col":case"embed":case"hr":case"keygen":case"meta":case"param":case"source":case"track":case"wbr":case"menuitem":for(var m in n)f=n[m],n.hasOwnProperty(m)&&null!=f&&!r.hasOwnProperty(m)&&Wc(e,t,m,null,r,f);for(c in r)if(f=r[c],h=n[c],r.hasOwnProperty(c)&&f!==h&&(null!=f||null!=h))switch(c){case"children":case"dangerouslySetInnerHTML":if(null!=f)throw Error(i(137,t));break;default:Wc(e,t,c,f,r,h)}return;default:if(Et(t)){for(var y in n)f=n[y],n.hasOwnProperty(y)&&void 0!==f&&!r.hasOwnProperty(y)&&Hc(e,t,y,void 0,r,f);for(p in r)f=r[p],h=n[p],!r.hasOwnProperty(p)||f===h||void 0===f&&void 0===h||Hc(e,t,p,f,r,h);return}}for(var b in n)f=n[b],n.hasOwnProperty(b)&&null!=f&&!r.hasOwnProperty(b)&&Wc(e,t,b,null,r,f);for(d in r)f=r[d],h=n[d],!r.hasOwnProperty(d)||f===h||null==f&&null==h||Wc(e,t,d,f,r,h)}(r,e.type,n,t),r[Fe]=t}catch(t){ac(e,e.return,t)}}function Xl(e){return 5===e.tag||3===e.tag||26===e.tag||27===e.tag||4===e.tag}function Kl(e){e:for(;;){for(;null===e.sibling;){if(null===e.return||Xl(e.return))return null;e=e.return}for(e.sibling.return=e.return,e=e.sibling;5!==e.tag&&6!==e.tag&&27!==e.tag&&18!==e.tag;){if(2&e.flags)continue e;if(null===e.child||4===e.tag)continue e;e.child.return=e,e=e.child}if(!(2&e.flags))return e.stateNode}}function Yl(e,t,n){var r=e.tag;if(5===r||6===r)e=e.stateNode,t?8===n.nodeType?n.parentNode.insertBefore(e,t):n.insertBefore(e,t):(8===n.nodeType?(t=n.parentNode).insertBefore(e,n):(t=n).appendChild(e),null!=(n=n._reactRootContainer)||null!==t.onclick||(t.onclick=Gc));else if(4!==r&&27!==r&&null!==(e=e.child))for(Yl(e,t,n),e=e.sibling;null!==e;)Yl(e,t,n),e=e.sibling}function Zl(e,t,n){var r=e.tag;if(5===r||6===r)e=e.stateNode,t?n.insertBefore(e,t):n.appendChild(e);else if(4!==r&&27!==r&&null!==(e=e.child))for(Zl(e,t,n),e=e.sibling;null!==e;)Zl(e,t,n),e=e.sibling}var Jl=!1,es=!1,ts=!1,ns="function"==typeof WeakSet?WeakSet:Set,rs=null,as=!1;function os(e,t,n){var r=n.flags;switch(n.tag){case 0:case 11:case 15:bs(e,n),4&r&&Bl(5,n);break;case 1:if(bs(e,n),4&r)if(e=n.stateNode,null===t)try{e.componentDidMount()}catch(e){ac(n,n.return,e)}else{var a=Mi(n.type,t.memoizedProps);t=t.memoizedState;try{e.componentDidUpdate(a,t,e.__reactInternalSnapshotBeforeUpdate)}catch(e){ac(n,n.return,e)}}64&r&&ql(n),512&r&&Wl(n,n.return);break;case 3:if(bs(e,n),64&r&&null!==(r=n.updateQueue)){if(e=null,null!==n.child)switch(n.child.tag){case 27:case 5:case 1:e=n.child.stateNode}try{Dl(r,e)}catch(e){ac(n,n.return,e)}}break;case 26:bs(e,n),512&r&&Wl(n,n.return);break;case 27:case 5:bs(e,n),null===t&&4&r&&Vl(n),512&r&&Wl(n,n.return);break;case 12:default:bs(e,n);break;case 13:bs(e,n),4&r&&ps(e,n);break;case 22:if(!(a=null!==n.memoizedState||Jl)){t=null!==t&&null!==t.memoizedState||es;var o=Jl,i=es;Jl=a,(es=t)&&!i?ws(e,n,!!(8772&n.subtreeFlags)):bs(e,n),Jl=o,es=i}512&r&&("manual"===n.memoizedProps.mode?Wl(n,n.return):Hl(n,n.return))}}function is(e){var t=e.alternate;null!==t&&(e.alternate=null,is(t)),e.child=null,e.deletions=null,e.sibling=null,5===e.tag&&null!==(t=e.stateNode)&&He(t),e.stateNode=null,e.return=null,e.dependencies=null,e.memoizedProps=null,e.memoizedState=null,e.pendingProps=null,e.stateNode=null,e.updateQueue=null}var ls=null,ss=!1;function us(e,t,n){for(n=n.child;null!==n;)cs(e,t,n),n=n.sibling}function cs(e,t,n){if(be&&"function"==typeof be.onCommitFiberUnmount)try{be.onCommitFiberUnmount(ye,n)}catch(e){}switch(n.tag){case 26:es||Hl(n,t),us(e,t,n),n.memoizedState?n.memoizedState.count--:n.stateNode&&(n=n.stateNode).parentNode.removeChild(n);break;case 27:es||Hl(n,t);var r=ls,a=ss;for(ls=n.stateNode,us(e,t,n),t=(n=n.stateNode).attributes;t.length;)n.removeAttributeNode(t[0]);He(n),ls=r,ss=a;break;case 5:es||Hl(n,t);case 6:a=ls;var o=ss;if(ls=null,us(e,t,n),ss=o,null!==(ls=a))if(ss)try{e=ls,r=n.stateNode,8===e.nodeType?e.parentNode.removeChild(r):e.removeChild(r)}catch(e){ac(n,t,e)}else try{ls.removeChild(n.stateNode)}catch(e){ac(n,t,e)}break;case 18:null!==ls&&(ss?(t=ls,n=n.stateNode,8===t.nodeType?ip(t.parentNode,n):1===t.nodeType&&ip(t,n),md(t)):ip(ls,n.stateNode));break;case 4:r=ls,a=ss,ls=n.stateNode.containerInfo,ss=!0,us(e,t,n),ls=r,ss=a;break;case 0:case 11:case 14:case 15:es||Ul(2,n,t),es||Ul(4,n,t),us(e,t,n);break;case 1:es||(Hl(n,t),"function"==typeof(r=n.stateNode).componentWillUnmount&&Gl(n,t,r)),us(e,t,n);break;case 21:us(e,t,n);break;case 22:es||Hl(n,t),es=(r=es)||null!==n.memoizedState,us(e,t,n),es=r;break;default:us(e,t,n)}}function ps(e,t){if(null===t.memoizedState&&null!==(e=t.alternate)&&null!==(e=e.memoizedState)&&null!==(e=e.dehydrated))try{md(e)}catch(e){ac(t,t.return,e)}}function ds(e,t){var n=function(e){switch(e.tag){case 13:case 19:var t=e.stateNode;return null===t&&(t=e.stateNode=new ns),t;case 22:return null===(t=(e=e.stateNode)._retryCache)&&(t=e._retryCache=new ns),t;default:throw Error(i(435,e.tag))}}(e);t.forEach((function(t){var r=uc.bind(null,e,t);n.has(t)||(n.add(t),t.then(r,r))}))}function fs(e,t){var n=t.deletions;if(null!==n)for(var r=0;r<n.length;r++){var a=n[r],o=e,l=t,s=l;e:for(;null!==s;){switch(s.tag){case 27:case 5:ls=s.stateNode,ss=!1;break e;case 3:case 4:ls=s.stateNode.containerInfo,ss=!0;break e}s=s.return}if(null===ls)throw Error(i(160));cs(o,l,a),ls=null,ss=!1,null!==(o=a.alternate)&&(o.return=null),a.return=null}if(13878&t.subtreeFlags)for(t=t.child;null!==t;)gs(t,e),t=t.sibling}var hs=null;function gs(e,t){var n=e.alternate,r=e.flags;switch(e.tag){case 0:case 11:case 14:case 15:fs(t,e),ms(e),4&r&&(Ul(3,e,e.return),Bl(3,e),Ul(5,e,e.return));break;case 1:fs(t,e),ms(e),512&r&&(es||null===n||Hl(n,n.return)),64&r&&Jl&&null!==(e=e.updateQueue)&&null!==(r=e.callbacks)&&(n=e.shared.hiddenCallbacks,e.shared.hiddenCallbacks=null===n?r:n.concat(r));break;case 26:var a=hs;if(fs(t,e),ms(e),512&r&&(es||null===n||Hl(n,n.return)),4&r){var o=null!==n?n.memoizedState:null;if(r=e.memoizedState,null===n)if(null===r)if(null===e.stateNode){e:{r=e.type,n=e.memoizedProps,a=a.ownerDocument||a;t:switch(r){case"title":(!(o=a.getElementsByTagName("title")[0])||o[We]||o[$e]||"http://www.w3.org/2000/svg"===o.namespaceURI||o.hasAttribute("itemprop"))&&(o=a.createElement(r),a.head.insertBefore(o,a.querySelector("head > title"))),Vc(o,r,n),o[$e]=e,Ye(o),r=o;break e;case"link":var l=Pp("link","href",a).get(r+(n.href||""));if(l)for(var s=0;s<l.length;s++)if((o=l[s]).getAttribute("href")===(null==n.href?null:n.href)&&o.getAttribute("rel")===(null==n.rel?null:n.rel)&&o.getAttribute("title")===(null==n.title?null:n.title)&&o.getAttribute("crossorigin")===(null==n.crossOrigin?null:n.crossOrigin)){l.splice(s,1);break t}Vc(o=a.createElement(r),r,n),a.head.appendChild(o);break;case"meta":if(l=Pp("meta","content",a).get(r+(n.content||"")))for(s=0;s<l.length;s++)if((o=l[s]).getAttribute("content")===(null==n.content?null:""+n.content)&&o.getAttribute("name")===(null==n.name?null:n.name)&&o.getAttribute("property")===(null==n.property?null:n.property)&&o.getAttribute("http-equiv")===(null==n.httpEquiv?null:n.httpEquiv)&&o.getAttribute("charset")===(null==n.charSet?null:n.charSet)){l.splice(s,1);break t}Vc(o=a.createElement(r),r,n),a.head.appendChild(o);break;default:throw Error(i(468,r))}o[$e]=e,Ye(o),r=o}e.stateNode=r}else Ap(a,e.type,e.stateNode);else e.stateNode=Sp(a,r,e.memoizedProps);else o!==r?(null===o?null!==n.stateNode&&(n=n.stateNode).parentNode.removeChild(n):o.count--,null===r?Ap(a,e.type,e.stateNode):Sp(a,r,e.memoizedProps)):null===r&&null!==e.stateNode&&Ql(e,e.memoizedProps,n.memoizedProps)}break;case 27:if(4&r&&null===e.alternate){a=e.stateNode,o=e.memoizedProps;try{for(var u=a.firstChild;u;){var c=u.nextSibling,p=u.nodeName;u[We]||"HEAD"===p||"BODY"===p||"SCRIPT"===p||"STYLE"===p||"LINK"===p&&"stylesheet"===u.rel.toLowerCase()||a.removeChild(u),u=c}for(var d=e.type,f=a.attributes;f.length;)a.removeAttributeNode(f[0]);Vc(a,d,o),a[$e]=e,a[Fe]=o}catch(t){ac(e,e.return,t)}}case 5:if(fs(t,e),ms(e),512&r&&(es||null===n||Hl(n,n.return)),32&e.flags){a=e.stateNode;try{kt(a,"")}catch(t){ac(e,e.return,t)}}4&r&&null!=e.stateNode&&Ql(e,a=e.memoizedProps,null!==n?n.memoizedProps:a),1024&r&&(ts=!0);break;case 6:if(fs(t,e),ms(e),4&r){if(null===e.stateNode)throw Error(i(162));r=e.memoizedProps,n=e.stateNode;try{n.nodeValue=r}catch(t){ac(e,e.return,t)}}break;case 3:if(Np=null,a=hs,hs=fp(t.containerInfo),fs(t,e),hs=a,ms(e),4&r&&null!==n&&n.memoizedState.isDehydrated)try{md(t.containerInfo)}catch(t){ac(e,e.return,t)}ts&&(ts=!1,ys(e));break;case 4:r=hs,hs=fp(e.stateNode.containerInfo),fs(t,e),ms(e),hs=r;break;case 12:fs(t,e),ms(e);break;case 13:fs(t,e),ms(e),8192&e.child.flags&&null!==e.memoizedState!=(null!==n&&null!==n.memoizedState)&&(wu=se()),4&r&&null!==(r=e.updateQueue)&&(e.updateQueue=null,ds(e,r));break;case 22:if(512&r&&(es||null===n||Hl(n,n.return)),u=null!==e.memoizedState,c=null!==n&&null!==n.memoizedState,Jl=(p=Jl)||u,es=(d=es)||c,fs(t,e),es=d,Jl=p,ms(e),(t=e.stateNode)._current=e,t._visibility&=-3,t._visibility|=2&t._pendingVisibility,8192&r&&(t._visibility=u?-2&t._visibility:1|t._visibility,u&&(t=Jl||es,null===n||c||t||vs(e)),null===e.memoizedProps||"manual"!==e.memoizedProps.mode))e:for(n=null,t=e;;){if(5===t.tag||26===t.tag||27===t.tag){if(null===n){c=n=t;try{if(a=c.stateNode,u)"function"==typeof(o=a.style).setProperty?o.setProperty("display","none","important"):o.display="none";else{l=c.stateNode;var h=null!=(s=c.memoizedProps.style)&&s.hasOwnProperty("display")?s.display:null;l.style.display=null==h||"boolean"==typeof h?"":(""+h).trim()}}catch(e){ac(c,c.return,e)}}}else if(6===t.tag){if(null===n){c=t;try{c.stateNode.nodeValue=u?"":c.memoizedProps}catch(e){ac(c,c.return,e)}}}else if((22!==t.tag&&23!==t.tag||null===t.memoizedState||t===e)&&null!==t.child){t.child.return=t,t=t.child;continue}if(t===e)break e;for(;null===t.sibling;){if(null===t.return||t.return===e)break e;n===t&&(n=null),t=t.return}n===t&&(n=null),t.sibling.return=t.return,t=t.sibling}4&r&&null!==(r=e.updateQueue)&&null!==(n=r.retryQueue)&&(r.retryQueue=null,ds(e,n));break;case 19:fs(t,e),ms(e),4&r&&null!==(r=e.updateQueue)&&(e.updateQueue=null,ds(e,r));break;case 21:break;default:fs(t,e),ms(e)}}function ms(e){var t=e.flags;if(2&t){try{if(27!==e.tag){e:{for(var n=e.return;null!==n;){if(Xl(n)){var r=n;break e}n=n.return}throw Error(i(160))}switch(r.tag){case 27:var a=r.stateNode;Zl(e,Kl(e),a);break;case 5:var o=r.stateNode;32&r.flags&&(kt(o,""),r.flags&=-33),Zl(e,Kl(e),o);break;case 3:case 4:var l=r.stateNode.containerInfo;Yl(e,Kl(e),l);break;default:throw Error(i(161))}}}catch(t){ac(e,e.return,t)}e.flags&=-3}4096&t&&(e.flags&=-4097)}function ys(e){if(1024&e.subtreeFlags)for(e=e.child;null!==e;){var t=e;ys(t),5===t.tag&&1024&t.flags&&t.stateNode.reset(),e=e.sibling}}function bs(e,t){if(8772&t.subtreeFlags)for(t=t.child;null!==t;)os(e,t.alternate,t),t=t.sibling}function vs(e){for(e=e.child;null!==e;){var t=e;switch(t.tag){case 0:case 11:case 14:case 15:Ul(4,t,t.return),vs(t);break;case 1:Hl(t,t.return);var n=t.stateNode;"function"==typeof n.componentWillUnmount&&Gl(t,t.return,n),vs(t);break;case 26:case 27:case 5:Hl(t,t.return),vs(t);break;case 22:Hl(t,t.return),null===t.memoizedState&&vs(t);break;default:vs(t)}e=e.sibling}}function ws(e,t,n){for(n=n&&!!(8772&t.subtreeFlags),t=t.child;null!==t;){var r=t.alternate,a=e,o=t,i=o.flags;switch(o.tag){case 0:case 11:case 15:ws(a,o,n),Bl(4,o);break;case 1:if(ws(a,o,n),"function"==typeof(a=(r=o).stateNode).componentDidMount)try{a.componentDidMount()}catch(e){ac(r,r.return,e)}if(null!==(a=(r=o).updateQueue)){var l=r.stateNode;try{var s=a.shared.hiddenCallbacks;if(null!==s)for(a.shared.hiddenCallbacks=null,a=0;a<s.length;a++)Fl(s[a],l)}catch(e){ac(r,r.return,e)}}n&&64&i&&ql(o),Wl(o,o.return);break;case 26:case 27:case 5:ws(a,o,n),n&&null===r&&4&i&&Vl(o),Wl(o,o.return);break;case 12:default:ws(a,o,n);break;case 13:ws(a,o,n),n&&4&i&&ps(a,o);break;case 22:null===o.memoizedState&&ws(a,o,n),Wl(o,o.return)}t=t.sibling}}function xs(e,t){var n=null;null!==e&&null!==e.memoizedState&&null!==e.memoizedState.cachePool&&(n=e.memoizedState.cachePool.pool),e=null,null!==t.memoizedState&&null!==t.memoizedState.cachePool&&(e=t.memoizedState.cachePool.pool),e!==n&&(null!=e&&e.refCount++,null!=n&&Ua(n))}function ks(e,t){e=null,null!==t.alternate&&(e=t.alternate.memoizedState.cache),(t=t.memoizedState.cache)!==e&&(t.refCount++,null!=e&&Ua(e))}function Ss(e,t,n,r){if(10256&t.subtreeFlags)for(t=t.child;null!==t;)Cs(e,t,n,r),t=t.sibling}function Cs(e,t,n,r){var a=t.flags;switch(t.tag){case 0:case 11:case 15:Ss(e,t,n,r),2048&a&&Bl(9,t);break;case 3:Ss(e,t,n,r),2048&a&&(e=null,null!==t.alternate&&(e=t.alternate.memoizedState.cache),(t=t.memoizedState.cache)!==e&&(t.refCount++,null!=e&&Ua(e)));break;case 12:if(2048&a){Ss(e,t,n,r),e=t.stateNode;try{var o=t.memoizedProps,i=o.id,l=o.onPostCommit;"function"==typeof l&&l(i,null===t.alternate?"mount":"update",e.passiveEffectDuration,-0)}catch(e){ac(t,t.return,e)}}else Ss(e,t,n,r);break;case 23:break;case 22:o=t.stateNode,null!==t.memoizedState?4&o._visibility?Ss(e,t,n,r):Es(e,t):4&o._visibility?Ss(e,t,n,r):(o._visibility|=4,_s(e,t,n,r,!!(10256&t.subtreeFlags))),2048&a&&xs(t.alternate,t);break;case 24:Ss(e,t,n,r),2048&a&&ks(t.alternate,t);break;default:Ss(e,t,n,r)}}function _s(e,t,n,r,a){for(a=a&&!!(10256&t.subtreeFlags),t=t.child;null!==t;){var o=e,i=t,l=n,s=r,u=i.flags;switch(i.tag){case 0:case 11:case 15:_s(o,i,l,s,a),Bl(8,i);break;case 23:break;case 22:var c=i.stateNode;null!==i.memoizedState?4&c._visibility?_s(o,i,l,s,a):Es(o,i):(c._visibility|=4,_s(o,i,l,s,a)),a&&2048&u&&xs(i.alternate,i);break;case 24:_s(o,i,l,s,a),a&&2048&u&&ks(i.alternate,i);break;default:_s(o,i,l,s,a)}t=t.sibling}}function Es(e,t){if(10256&t.subtreeFlags)for(t=t.child;null!==t;){var n=e,r=t,a=r.flags;switch(r.tag){case 22:Es(n,r),2048&a&&xs(r.alternate,r);break;case 24:Es(n,r),2048&a&&ks(r.alternate,r);break;default:Es(n,r)}t=t.sibling}}var Ns=8192;function Ps(e){if(e.subtreeFlags&Ns)for(e=e.child;null!==e;)As(e),e=e.sibling}function As(e){switch(e.tag){case 26:Ps(e),e.flags&Ns&&null!==e.memoizedState&&function(e,t,n){if(null===Tp)throw Error(i(475));var r=Tp;if(!("stylesheet"!==t.type||"string"==typeof n.media&&!1===matchMedia(n.media).matches||4&t.state.loading)){if(null===t.instance){var a=bp(n.href),o=e.querySelector(vp(a));if(o)return null!==(e=o._p)&&"object"==typeof e&&"function"==typeof e.then&&(r.count++,r=Ip.bind(r),e.then(r,r)),t.state.loading|=4,t.instance=o,void Ye(o);o=e.ownerDocument||e,n=wp(n),(a=pp.get(a))&&_p(n,a),Ye(o=o.createElement("link"));var l=o;l._p=new Promise((function(e,t){l.onload=e,l.onerror=t})),Vc(o,"link",n),t.instance=o}null===r.stylesheets&&(r.stylesheets=new Map),r.stylesheets.set(t,e),(e=t.state.preload)&&!(3&t.state.loading)&&(r.count++,t=Ip.bind(r),e.addEventListener("load",t),e.addEventListener("error",t))}}(hs,e.memoizedState,e.memoizedProps);break;case 5:default:Ps(e);break;case 3:case 4:var t=hs;hs=fp(e.stateNode.containerInfo),Ps(e),hs=t;break;case 22:null===e.memoizedState&&(null!==(t=e.alternate)&&null!==t.memoizedState?(t=Ns,Ns=16777216,Ps(e),Ns=t):Ps(e))}}function Ls(e){var t=e.alternate;if(null!==t&&null!==(e=t.child)){t.child=null;do{t=e.sibling,e.sibling=null,e=t}while(null!==e)}}function Ts(e){var t=e.deletions;if(16&e.flags){if(null!==t)for(var n=0;n<t.length;n++){var r=t[n];rs=r,Ms(r,e)}Ls(e)}if(10256&e.subtreeFlags)for(e=e.child;null!==e;)Rs(e),e=e.sibling}function Rs(e){switch(e.tag){case 0:case 11:case 15:Ts(e),2048&e.flags&&Ul(9,e,e.return);break;case 3:case 12:default:Ts(e);break;case 22:var t=e.stateNode;null!==e.memoizedState&&4&t._visibility&&(null===e.return||13!==e.return.tag)?(t._visibility&=-5,Is(e)):Ts(e)}}function Is(e){var t=e.deletions;if(16&e.flags){if(null!==t)for(var n=0;n<t.length;n++){var r=t[n];rs=r,Ms(r,e)}Ls(e)}for(e=e.child;null!==e;){switch((t=e).tag){case 0:case 11:case 15:Ul(8,t,t.return),Is(t);break;case 22:4&(n=t.stateNode)._visibility&&(n._visibility&=-5,Is(t));break;default:Is(t)}e=e.sibling}}function Ms(e,t){for(;null!==rs;){var n=rs;switch(n.tag){case 0:case 11:case 15:Ul(8,n,t);break;case 23:case 22:if(null!==n.memoizedState&&null!==n.memoizedState.cachePool){var r=n.memoizedState.cachePool.pool;null!=r&&r.refCount++}break;case 24:Ua(n.memoizedState.cache)}if(null!==(r=n.child))r.return=n,rs=r;else e:for(n=e;null!==rs;){var a=(r=rs).sibling,o=r.return;if(is(r),r===n){rs=null;break e}if(null!==a){a.return=o,rs=a;break e}rs=o}}}function Os(e,t,n,r){this.tag=e,this.key=n,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.refCleanup=this.ref=null,this.pendingProps=t,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=r,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function js(e,t,n,r){return new Os(e,t,n,r)}function zs(e){return!(!(e=e.prototype)||!e.isReactComponent)}function $s(e,t){var n=e.alternate;return null===n?((n=js(e.tag,t,e.key,e.mode)).elementType=e.elementType,n.type=e.type,n.stateNode=e.stateNode,n.alternate=e,e.alternate=n):(n.pendingProps=t,n.type=e.type,n.flags=0,n.subtreeFlags=0,n.deletions=null),n.flags=31457280&e.flags,n.childLanes=e.childLanes,n.lanes=e.lanes,n.child=e.child,n.memoizedProps=e.memoizedProps,n.memoizedState=e.memoizedState,n.updateQueue=e.updateQueue,t=e.dependencies,n.dependencies=null===t?null:{lanes:t.lanes,firstContext:t.firstContext},n.sibling=e.sibling,n.index=e.index,n.ref=e.ref,n.refCleanup=e.refCleanup,n}function Fs(e,t){e.flags&=31457282;var n=e.alternate;return null===n?(e.childLanes=0,e.lanes=t,e.child=null,e.subtreeFlags=0,e.memoizedProps=null,e.memoizedState=null,e.updateQueue=null,e.dependencies=null,e.stateNode=null):(e.childLanes=n.childLanes,e.lanes=n.lanes,e.child=n.child,e.subtreeFlags=0,e.deletions=null,e.memoizedProps=n.memoizedProps,e.memoizedState=n.memoizedState,e.updateQueue=n.updateQueue,e.type=n.type,t=n.dependencies,e.dependencies=null===t?null:{lanes:t.lanes,firstContext:t.firstContext}),e}function Ds(e,t,n,r,a,o){var l=0;if(r=e,"function"==typeof e)zs(e)&&(l=1);else if("string"==typeof e)l=function(e,t,n){if(1===n||null!=t.itemProp)return!1;switch(e){case"meta":case"title":return!0;case"style":if("string"!=typeof t.precedence||"string"!=typeof t.href||""===t.href)break;return!0;case"link":if("string"!=typeof t.rel||"string"!=typeof t.href||""===t.href||t.onLoad||t.onError)break;return"stylesheet"!==t.rel||(e=t.disabled,"string"==typeof t.precedence&&null==e);case"script":if(t.async&&"function"!=typeof t.async&&"symbol"!=typeof t.async&&!t.onLoad&&!t.onError&&t.src&&"string"==typeof t.src)return!0}return!1}(e,n,X.current)?26:"html"===e||"head"===e||"body"===e?27:5;else e:switch(e){case p:return Bs(n.children,a,o,t);case d:l=8,a|=24;break;case f:return(e=js(12,n,t,2|a)).elementType=f,e.lanes=o,e;case b:return(e=js(13,n,t,a)).elementType=b,e.lanes=o,e;case v:return(e=js(19,n,t,a)).elementType=v,e.lanes=o,e;case k:return Us(n,a,o,t);default:if("object"==typeof e&&null!==e)switch(e.$$typeof){case h:case m:l=10;break e;case g:l=9;break e;case y:l=11;break e;case w:l=14;break e;case x:l=16,r=null;break e}l=29,n=Error(i(130,null===e?"null":typeof e,"")),r=null}return(t=js(l,n,t,a)).elementType=e,t.type=r,t.lanes=o,t}function Bs(e,t,n,r){return(e=js(7,e,r,t)).lanes=n,e}function Us(e,t,n,r){(e=js(22,e,r,t)).elementType=k,e.lanes=n;var a={_visibility:1,_pendingVisibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null,_current:null,detach:function(){var e=a._current;if(null===e)throw Error(i(456));if(!(2&a._pendingVisibility)){var t=Lr(e,2);null!==t&&(a._pendingVisibility|=2,Iu(t,0,2))}},attach:function(){var e=a._current;if(null===e)throw Error(i(456));if(2&a._pendingVisibility){var t=Lr(e,2);null!==t&&(a._pendingVisibility&=-3,Iu(t,0,2))}}};return e.stateNode=a,e}function qs(e,t,n){return(e=js(6,e,null,t)).lanes=n,e}function Gs(e,t,n){return(t=js(4,null!==e.children?e.children:[],e.key,t)).lanes=n,t.stateNode={containerInfo:e.containerInfo,pendingChildren:null,implementation:e.implementation},t}function Ws(e){e.flags|=4}function Hs(e,t){if("stylesheet"!==t.type||4&t.state.loading)e.flags&=-16777217;else if(e.flags|=16777216,!Lp(t)){if(null!==(t=Aa.current)&&((4194176&au)===au?null!==La:(62914560&au)!==au&&!(536870912&au)||t!==La))throw fa=ua,sa;e.flags|=8192}}function Vs(e,t){null!==t&&(e.flags|=4),16384&e.flags&&(t=22!==e.tag?Le():536870912,e.lanes|=t,mu|=t)}function Qs(e,t){if(!Yr)switch(e.tailMode){case"hidden":t=e.tail;for(var n=null;null!==t;)null!==t.alternate&&(n=t),t=t.sibling;null===n?e.tail=null:n.sibling=null;break;case"collapsed":n=e.tail;for(var r=null;null!==n;)null!==n.alternate&&(r=n),n=n.sibling;null===r?t||null===e.tail?e.tail=null:e.tail.sibling=null:r.sibling=null}}function Xs(e){var t=null!==e.alternate&&e.alternate.child===e.child,n=0,r=0;if(t)for(var a=e.child;null!==a;)n|=a.lanes|a.childLanes,r|=31457280&a.subtreeFlags,r|=31457280&a.flags,a.return=e,a=a.sibling;else for(a=e.child;null!==a;)n|=a.lanes|a.childLanes,r|=a.subtreeFlags,r|=a.flags,a.return=e,a=a.sibling;return e.subtreeFlags|=r,e.childLanes=n,t}function Ks(e,t,n){var r=t.pendingProps;switch(Qr(t),t.tag){case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:case 1:return Xs(t),null;case 3:return n=t.stateNode,r=null,null!==e&&(r=e.memoizedState.cache),t.memoizedState.cache!==r&&(t.flags|=2048),wl(Da),ee(),n.pendingContext&&(n.context=n.pendingContext,n.pendingContext=null),null!==e&&null!==e.child||(aa(t)?Ws(t):null===e||e.memoizedState.isDehydrated&&!(256&t.flags)||(t.flags|=1024,null!==Zr&&(Ou(Zr),Zr=null))),Xs(t),null;case 26:return n=t.memoizedState,null===e?(Ws(t),null!==n?(Xs(t),Hs(t,n)):(Xs(t),t.flags&=-16777217)):n?n!==e.memoizedState?(Ws(t),Xs(t),Hs(t,n)):(Xs(t),t.flags&=-16777217):(e.memoizedProps!==r&&Ws(t),Xs(t),t.flags&=-16777217),null;case 27:ne(t),n=Y.current;var a=t.type;if(null!==e&&null!=t.stateNode)e.memoizedProps!==r&&Ws(t);else{if(!r){if(null===t.stateNode)throw Error(i(166));return Xs(t),null}e=X.current,aa(t)?na(t):(e=cp(a,r,n),t.stateNode=e,Ws(t))}return Xs(t),null;case 5:if(ne(t),n=t.type,null!==e&&null!=t.stateNode)e.memoizedProps!==r&&Ws(t);else{if(!r){if(null===t.stateNode)throw Error(i(166));return Xs(t),null}if(e=X.current,aa(t))na(t);else{switch(a=Kc(Y.current),e){case 1:e=a.createElementNS("http://www.w3.org/2000/svg",n);break;case 2:e=a.createElementNS("http://www.w3.org/1998/Math/MathML",n);break;default:switch(n){case"svg":e=a.createElementNS("http://www.w3.org/2000/svg",n);break;case"math":e=a.createElementNS("http://www.w3.org/1998/Math/MathML",n);break;case"script":(e=a.createElement("div")).innerHTML="<script><\/script>",e=e.removeChild(e.firstChild);break;case"select":e="string"==typeof r.is?a.createElement("select",{is:r.is}):a.createElement("select"),r.multiple?e.multiple=!0:r.size&&(e.size=r.size);break;default:e="string"==typeof r.is?a.createElement(n,{is:r.is}):a.createElement(n)}}e[$e]=t,e[Fe]=r;e:for(a=t.child;null!==a;){if(5===a.tag||6===a.tag)e.appendChild(a.stateNode);else if(4!==a.tag&&27!==a.tag&&null!==a.child){a.child.return=a,a=a.child;continue}if(a===t)break e;for(;null===a.sibling;){if(null===a.return||a.return===t)break e;a=a.return}a.sibling.return=a.return,a=a.sibling}t.stateNode=e;e:switch(Vc(e,n,r),n){case"button":case"input":case"select":case"textarea":e=!!r.autoFocus;break e;case"img":e=!0;break e;default:e=!1}e&&Ws(t)}}return Xs(t),t.flags&=-16777217,null;case 6:if(e&&null!=t.stateNode)e.memoizedProps!==r&&Ws(t);else{if("string"!=typeof r&&null===t.stateNode)throw Error(i(166));if(e=Y.current,aa(t)){if(e=t.stateNode,n=t.memoizedProps,r=null,null!==(a=Xr))switch(a.tag){case 27:case 5:r=a.memoizedProps}e[$e]=t,(e=!!(e.nodeValue===n||null!==r&&!0===r.suppressHydrationWarning||qc(e.nodeValue,n)))||ta(t)}else(e=Kc(e).createTextNode(r))[$e]=t,t.stateNode=e}return Xs(t),null;case 13:if(r=t.memoizedState,null===e||null!==e.memoizedState&&null!==e.memoizedState.dehydrated){if(a=aa(t),null!==r&&null!==r.dehydrated){if(null===e){if(!a)throw Error(i(318));if(!(a=null!==(a=t.memoizedState)?a.dehydrated:null))throw Error(i(317));a[$e]=t}else oa(),!(128&t.flags)&&(t.memoizedState=null),t.flags|=4;Xs(t),a=!1}else null!==Zr&&(Ou(Zr),Zr=null),a=!0;if(!a)return 256&t.flags?(Ma(t),t):(Ma(t),null)}if(Ma(t),128&t.flags)return t.lanes=n,t;if(n=null!==r,e=null!==e&&null!==e.memoizedState,n){a=null,null!==(r=t.child).alternate&&null!==r.alternate.memoizedState&&null!==r.alternate.memoizedState.cachePool&&(a=r.alternate.memoizedState.cachePool.pool);var o=null;null!==r.memoizedState&&null!==r.memoizedState.cachePool&&(o=r.memoizedState.cachePool.pool),o!==a&&(r.flags|=2048)}return n!==e&&n&&(t.child.flags|=8192),Vs(t,t.updateQueue),Xs(t),null;case 4:return ee(),null===e&&Ic(t.stateNode.containerInfo),Xs(t),null;case 10:return wl(t.type),Xs(t),null;case 19:if(V(Oa),null===(a=t.memoizedState))return Xs(t),null;if(r=!!(128&t.flags),null===(o=a.rendering))if(r)Qs(a,!1);else{if(0!==pu||null!==e&&128&e.flags)for(e=t.child;null!==e;){if(null!==(o=ja(e))){for(t.flags|=128,Qs(a,!1),e=o.updateQueue,t.updateQueue=e,Vs(t,e),t.subtreeFlags=0,e=n,n=t.child;null!==n;)Fs(n,e),n=n.sibling;return Q(Oa,1&Oa.current|2),t.child}e=e.sibling}null!==a.tail&&se()>xu&&(t.flags|=128,r=!0,Qs(a,!1),t.lanes=4194304)}else{if(!r)if(null!==(e=ja(o))){if(t.flags|=128,r=!0,e=e.updateQueue,t.updateQueue=e,Vs(t,e),Qs(a,!0),null===a.tail&&"hidden"===a.tailMode&&!o.alternate&&!Yr)return Xs(t),null}else 2*se()-a.renderingStartTime>xu&&536870912!==n&&(t.flags|=128,r=!0,Qs(a,!1),t.lanes=4194304);a.isBackwards?(o.sibling=t.child,t.child=o):(null!==(e=a.last)?e.sibling=o:t.child=o,a.last=o)}return null!==a.tail?(t=a.tail,a.rendering=t,a.tail=t.sibling,a.renderingStartTime=se(),t.sibling=null,e=Oa.current,Q(Oa,r?1&e|2:1&e),t):(Xs(t),null);case 22:case 23:return Ma(t),Pa(),r=null!==t.memoizedState,null!==e?null!==e.memoizedState!==r&&(t.flags|=8192):r&&(t.flags|=8192),r?!!(536870912&n)&&!(128&t.flags)&&(Xs(t),6&t.subtreeFlags&&(t.flags|=8192)):Xs(t),null!==(n=t.updateQueue)&&Vs(t,n.retryQueue),n=null,null!==e&&null!==e.memoizedState&&null!==e.memoizedState.cachePool&&(n=e.memoizedState.cachePool.pool),r=null,null!==t.memoizedState&&null!==t.memoizedState.cachePool&&(r=t.memoizedState.cachePool.pool),r!==n&&(t.flags|=2048),null!==e&&V(Xa),null;case 24:return n=null,null!==e&&(n=e.memoizedState.cache),t.memoizedState.cache!==n&&(t.flags|=2048),wl(Da),Xs(t),null;case 25:return null}throw Error(i(156,t.tag))}function Ys(e,t){switch(Qr(t),t.tag){case 1:return 65536&(e=t.flags)?(t.flags=-65537&e|128,t):null;case 3:return wl(Da),ee(),65536&(e=t.flags)&&!(128&e)?(t.flags=-65537&e|128,t):null;case 26:case 27:case 5:return ne(t),null;case 13:if(Ma(t),null!==(e=t.memoizedState)&&null!==e.dehydrated){if(null===t.alternate)throw Error(i(340));oa()}return 65536&(e=t.flags)?(t.flags=-65537&e|128,t):null;case 19:return V(Oa),null;case 4:return ee(),null;case 10:return wl(t.type),null;case 22:case 23:return Ma(t),Pa(),null!==e&&V(Xa),65536&(e=t.flags)?(t.flags=-65537&e|128,t):null;case 24:return wl(Da),null;default:return null}}function Zs(e,t){switch(Qr(t),t.tag){case 3:wl(Da),ee();break;case 26:case 27:case 5:ne(t);break;case 4:ee();break;case 13:Ma(t);break;case 19:V(Oa);break;case 10:wl(t.type);break;case 22:case 23:Ma(t),Pa(),null!==e&&V(Xa);break;case 24:wl(Da)}}var Js={getCacheForType:function(e){var t=El(Da),n=t.data.get(e);return void 0===n&&(n=e(),t.data.set(e,n)),n}},eu="function"==typeof WeakMap?WeakMap:Map,tu=0,nu=null,ru=null,au=0,ou=0,iu=null,lu=!1,su=!1,uu=!1,cu=0,pu=0,du=0,fu=0,hu=0,gu=0,mu=0,yu=null,bu=null,vu=!1,wu=0,xu=1/0,ku=null,Su=null,Cu=!1,_u=null,Eu=0,Nu=0,Pu=null,Au=0,Lu=null;function Tu(){return 2&tu&&0!==au?au&-au:null!==L.T?0!==Wa?Wa:kc():je()}function Ru(){0===gu&&(gu=536870912&au&&!Yr?536870912:Ae());var e=Aa.current;return null!==e&&(e.flags|=32),gu}function Iu(e,t,n){(e===nu&&2===ou||null!==e.cancelPendingCommit)&&(Bu(e,0),$u(e,au,gu,!1)),Re(e,n),2&tu&&e===nu||(e===nu&&(!(2&tu)&&(fu|=n),4===pu&&$u(e,au,gu,!1)),mc(e))}function Mu(e,t,n){if(6&tu)throw Error(i(327));for(var r=!n&&!(60&t)&&!(t&e.expiredLanes)||Ne(e,t),a=r?function(e,t){var n=tu;tu|=2;var r=qu(),a=Gu();nu!==e||au!==t?(ku=null,xu=se()+500,Bu(e,t)):su=Ne(e,t);e:for(;;)try{if(0!==ou&&null!==ru){t=ru;var o=iu;t:switch(ou){case 1:ou=0,iu=null,Yu(e,t,o,1);break;case 2:if(ca(o)){ou=0,iu=null,Ku(t);break}t=function(){2===ou&&nu===e&&(ou=7),mc(e)},o.then(t,t);break e;case 3:ou=7;break e;case 4:ou=5;break e;case 7:ca(o)?(ou=0,iu=null,Ku(t)):(ou=0,iu=null,Yu(e,t,o,7));break;case 5:var l=null;switch(ru.tag){case 26:l=ru.memoizedState;case 5:case 27:var s=ru;if(!l||Lp(l)){ou=0,iu=null;var u=s.sibling;if(null!==u)ru=u;else{var c=s.return;null!==c?(ru=c,Zu(c)):ru=null}break t}}ou=0,iu=null,Yu(e,t,o,5);break;case 6:ou=0,iu=null,Yu(e,t,o,6);break;case 8:Du(),pu=6;break e;default:throw Error(i(462))}}Qu();break}catch(t){Uu(e,t)}return bl=yl=null,L.H=r,L.A=a,tu=n,null!==ru?0:(nu=null,au=0,Nr(),pu)}(e,t):Hu(e,t,!0),o=r;;){if(0===a){su&&!r&&$u(e,t,0,!1);break}if(6===a)$u(e,t,0,!lu);else{if(n=e.current.alternate,o&&!zu(n)){a=Hu(e,t,!1),o=!1;continue}if(2===a){if(o=t,e.errorRecoveryDisabledLanes&o)var l=0;else l=0!=(l=-536870913&e.pendingLanes)?l:536870912&l?536870912:0;if(0!==l){t=l;e:{var s=e;a=yu;var u=s.current.memoizedState.isDehydrated;if(u&&(Bu(s,l).flags|=256),2!==(l=Hu(s,l,!1))){if(uu&&!u){s.errorRecoveryDisabledLanes|=o,fu|=o,a=4;break e}o=bu,bu=a,null!==o&&Ou(o)}a=l}if(o=!1,2!==a)continue}}if(1===a){Bu(e,0),$u(e,t,0,!0);break}e:{switch(r=e,a){case 0:case 1:throw Error(i(345));case 4:if((4194176&t)===t){$u(r,t,gu,!lu);break e}break;case 2:bu=null;break;case 3:case 5:break;default:throw Error(i(329))}if(r.finishedWork=n,r.finishedLanes=t,(62914560&t)===t&&10<(o=wu+300-se())){if($u(r,t,gu,!lu),0!==Ee(r,0))break e;r.timeoutHandle=tp(ju.bind(null,r,n,bu,ku,vu,t,gu,fu,mu,lu,2,-0,0),o)}else ju(r,n,bu,ku,vu,t,gu,fu,mu,lu,0,-0,0)}}break}mc(e)}function Ou(e){null===bu?bu=e:bu.push.apply(bu,e)}function ju(e,t,n,r,a,o,l,s,u,c,p,d,f){var h=t.subtreeFlags;if((8192&h||!(16785408&~h))&&(Tp={stylesheets:null,count:0,unsuspend:Rp},As(t),null!==(t=function(){if(null===Tp)throw Error(i(475));var e=Tp;return e.stylesheets&&0===e.count&&Op(e,e.stylesheets),0<e.count?function(t){var n=setTimeout((function(){if(e.stylesheets&&Op(e,e.stylesheets),e.unsuspend){var t=e.unsuspend;e.unsuspend=null,t()}}),6e4);return e.unsuspend=t,function(){e.unsuspend=null,clearTimeout(n)}}:null}())))return e.cancelPendingCommit=t(ec.bind(null,e,n,r,a,l,s,u,1,d,f)),void $u(e,o,l,!c);ec(e,n,r,a,l,s,u)}function zu(e){for(var t=e;;){var n=t.tag;if((0===n||11===n||15===n)&&16384&t.flags&&null!==(n=t.updateQueue)&&null!==(n=n.stores))for(var r=0;r<n.length;r++){var a=n[r],o=a.getSnapshot;a=a.value;try{if(!Xn(o(),a))return!1}catch(e){return!1}}if(n=t.child,16384&t.subtreeFlags&&null!==n)n.return=t,t=n;else{if(t===e)break;for(;null===t.sibling;){if(null===t.return||t.return===e)return!0;t=t.return}t.sibling.return=t.return,t=t.sibling}}return!0}function $u(e,t,n,r){t&=~hu,t&=~fu,e.suspendedLanes|=t,e.pingedLanes&=~t,r&&(e.warmLanes|=t),r=e.expirationTimes;for(var a=t;0<a;){var o=31-we(a),i=1<<o;r[o]=-1,a&=~i}0!==n&&Ie(e,n,t)}function Fu(){return!!(6&tu)||(yc(0,!1),!1)}function Du(){if(null!==ru){if(0===ou)var e=ru.return;else bl=yl=null,vo(e=ru),ga=null,ma=0,e=ru;for(;null!==e;)Zs(e.alternate,e),e=e.return;ru=null}}function Bu(e,t){e.finishedWork=null,e.finishedLanes=0;var n=e.timeoutHandle;-1!==n&&(e.timeoutHandle=-1,np(n)),null!==(n=e.cancelPendingCommit)&&(e.cancelPendingCommit=null,n()),Du(),nu=e,ru=n=$s(e.current,null),au=t,ou=0,iu=null,lu=!1,su=Ne(e,t),uu=!1,mu=gu=hu=fu=du=pu=0,bu=yu=null,vu=!1,8&t&&(t|=32&t);var r=e.entangledLanes;if(0!==r)for(e=e.entanglements,r&=t;0<r;){var a=31-we(r),o=1<<a;t|=e[a],r&=~o}return cu=t,Nr(),n}function Uu(e,t){eo=null,L.H=Ei,t===la?(t=ha(),ou=3):t===sa?(t=ha(),ou=4):ou=t===Gi?8:null!==t&&"object"==typeof t&&"function"==typeof t.then?6:1,iu=t,null===ru&&(pu=1,Fi(e,Or(t,e.current)))}function qu(){var e=L.H;return L.H=Ei,null===e?Ei:e}function Gu(){var e=L.A;return L.A=Js,e}function Wu(){pu=4,lu||(4194176&au)!==au&&null!==Aa.current||(su=!0),!(134217727&du)&&!(134217727&fu)||null===nu||$u(nu,au,gu,!1)}function Hu(e,t,n){var r=tu;tu|=2;var a=qu(),o=Gu();nu===e&&au===t||(ku=null,Bu(e,t)),t=!1;var i=pu;e:for(;;)try{if(0!==ou&&null!==ru){var l=ru,s=iu;switch(ou){case 8:Du(),i=6;break e;case 3:case 2:case 6:null===Aa.current&&(t=!0);var u=ou;if(ou=0,iu=null,Yu(e,l,s,u),n&&su){i=0;break e}break;default:u=ou,ou=0,iu=null,Yu(e,l,s,u)}}Vu(),i=pu;break}catch(t){Uu(e,t)}return t&&e.shellSuspendCounter++,bl=yl=null,tu=r,L.H=a,L.A=o,null===ru&&(nu=null,au=0,Nr()),i}function Vu(){for(;null!==ru;)Xu(ru)}function Qu(){for(;null!==ru&&!ie();)Xu(ru)}function Xu(e){var t=gl(e.alternate,e,cu);e.memoizedProps=e.pendingProps,null===t?Zu(e):ru=t}function Ku(e){var t=e,n=t.alternate;switch(t.tag){case 15:case 0:t=el(n,t,t.pendingProps,t.type,void 0,au);break;case 11:t=el(n,t,t.pendingProps,t.type.render,t.ref,au);break;case 5:vo(t);default:Zs(n,t),t=gl(n,t=ru=Fs(t,cu),cu)}e.memoizedProps=e.pendingProps,null===t?Zu(e):ru=t}function Yu(e,t,n,r){bl=yl=null,vo(t),ga=null,ma=0;var a=t.return;try{if(function(e,t,n,r,a){if(n.flags|=32768,null!==r&&"object"==typeof r&&"function"==typeof r.then){if(null!==(t=n.alternate)&&Sl(t,n,a,!0),null!==(n=Aa.current)){switch(n.tag){case 13:return null===La?Wu():null===n.alternate&&0===pu&&(pu=3),n.flags&=-257,n.flags|=65536,n.lanes=a,r===ua?n.flags|=16384:(null===(t=n.updateQueue)?n.updateQueue=new Set([r]):t.add(r),oc(e,r,a)),!1;case 22:return n.flags|=65536,r===ua?n.flags|=16384:(null===(t=n.updateQueue)?(t={transitions:null,markerInstances:null,retryQueue:new Set([r])},n.updateQueue=t):null===(n=t.retryQueue)?t.retryQueue=new Set([r]):n.add(r),oc(e,r,a)),!1}throw Error(i(435,n.tag))}return oc(e,r,a),Wu(),!1}if(Yr)return null!==(t=Aa.current)?(!(65536&t.flags)&&(t.flags|=256),t.flags|=65536,t.lanes=a,r!==ea&&ia(Or(e=Error(i(422),{cause:r}),n))):(r!==ea&&ia(Or(t=Error(i(423),{cause:r}),n)),(e=e.current.alternate).flags|=65536,a&=-a,e.lanes|=a,r=Or(r,n),Ol(e,a=Bi(e.stateNode,r,a)),4!==pu&&(pu=2)),!1;var o=Error(i(520),{cause:r});if(o=Or(o,n),null===yu?yu=[o]:yu.push(o),4!==pu&&(pu=2),null===t)return!0;r=Or(r,n),n=t;do{switch(n.tag){case 3:return n.flags|=65536,e=a&-a,n.lanes|=e,Ol(n,e=Bi(n.stateNode,r,e)),!1;case 1:if(t=n.type,o=n.stateNode,!(128&n.flags||"function"!=typeof t.getDerivedStateFromError&&(null===o||"function"!=typeof o.componentDidCatch||null!==Su&&Su.has(o))))return n.flags|=65536,a&=-a,n.lanes|=a,qi(a=Ui(a),e,n,r),Ol(n,a),!1}n=n.return}while(null!==n);return!1}(e,a,t,n,au))return pu=1,Fi(e,Or(n,e.current)),void(ru=null)}catch(t){if(null!==a)throw ru=a,t;return pu=1,Fi(e,Or(n,e.current)),void(ru=null)}32768&t.flags?(Yr||1===r?e=!0:su||536870912&au?e=!1:(lu=e=!0,(2===r||3===r||6===r)&&null!==(r=Aa.current)&&13===r.tag&&(r.flags|=16384)),Ju(t,e)):Zu(t)}function Zu(e){var t=e;do{if(32768&t.flags)return void Ju(t,lu);e=t.return;var n=Ks(t.alternate,t,cu);if(null!==n)return void(ru=n);if(null!==(t=t.sibling))return void(ru=t);ru=t=e}while(null!==t);0===pu&&(pu=5)}function Ju(e,t){do{var n=Ys(e.alternate,e);if(null!==n)return n.flags&=32767,void(ru=n);if(null!==(n=e.return)&&(n.flags|=32768,n.subtreeFlags=0,n.deletions=null),!t&&null!==(e=e.sibling))return void(ru=e);ru=e=n}while(null!==e);pu=6,ru=null}function ec(e,t,n,r,a,o,l,s,u,c){var p=L.T,d=U.p;try{U.p=2,L.T=null,function(e,t,n,r,a,o,l,s){do{nc()}while(null!==_u);if(6&tu)throw Error(i(327));var u=e.finishedWork;if(r=e.finishedLanes,null===u)return null;if(e.finishedWork=null,e.finishedLanes=0,u===e.current)throw Error(i(177));e.callbackNode=null,e.callbackPriority=0,e.cancelPendingCommit=null;var c=u.lanes|u.childLanes;if(function(e,t,n,r,a,o){var i=e.pendingLanes;e.pendingLanes=n,e.suspendedLanes=0,e.pingedLanes=0,e.warmLanes=0,e.expiredLanes&=n,e.entangledLanes&=n,e.errorRecoveryDisabledLanes&=n,e.shellSuspendCounter=0;var l=e.entanglements,s=e.expirationTimes,u=e.hiddenUpdates;for(n=i&~n;0<n;){var c=31-we(n),p=1<<c;l[c]=0,s[c]=-1;var d=u[c];if(null!==d)for(u[c]=null,c=0;c<d.length;c++){var f=d[c];null!==f&&(f.lane&=-536870913)}n&=~p}0!==r&&Ie(e,r,0),0!==o&&0===a&&0!==e.tag&&(e.suspendedLanes|=o&~(i&~t))}(e,r,c|=Er,o,l,s),e===nu&&(ru=nu=null,au=0),!(10256&u.subtreeFlags)&&!(10256&u.flags)||Cu||(Cu=!0,Nu=c,Pu=n,ae(de,(function(){return nc(),null}))),n=!!(15990&u.flags),15990&u.subtreeFlags||n?(n=L.T,L.T=null,o=U.p,U.p=2,l=tu,tu|=4,function(e,t){if(e=e.containerInfo,Qc=Wp,tr(e=er(e))){if("selectionStart"in e)var n={start:e.selectionStart,end:e.selectionEnd};else e:{var r=(n=(n=e.ownerDocument)&&n.defaultView||window).getSelection&&n.getSelection();if(r&&0!==r.rangeCount){n=r.anchorNode;var a=r.anchorOffset,o=r.focusNode;r=r.focusOffset;try{n.nodeType,o.nodeType}catch(e){n=null;break e}var l=0,s=-1,u=-1,c=0,p=0,d=e,f=null;t:for(;;){for(var h;d!==n||0!==a&&3!==d.nodeType||(s=l+a),d!==o||0!==r&&3!==d.nodeType||(u=l+r),3===d.nodeType&&(l+=d.nodeValue.length),null!==(h=d.firstChild);)f=d,d=h;for(;;){if(d===e)break t;if(f===n&&++c===a&&(s=l),f===o&&++p===r&&(u=l),null!==(h=d.nextSibling))break;f=(d=f).parentNode}d=h}n=-1===s||-1===u?null:{start:s,end:u}}else n=null}n=n||{start:0,end:0}}else n=null;for(Xc={focusedElem:e,selectionRange:n},Wp=!1,rs=t;null!==rs;)if(e=(t=rs).child,1028&t.subtreeFlags&&null!==e)e.return=t,rs=e;else for(;null!==rs;){switch(o=(t=rs).alternate,e=t.flags,t.tag){case 0:case 11:case 15:case 5:case 26:case 27:case 6:case 4:case 17:break;case 1:if(1024&e&&null!==o){e=void 0,n=t,a=o.memoizedProps,o=o.memoizedState,r=n.stateNode;try{var g=Mi(n.type,a,(n.elementType,n.type));e=r.getSnapshotBeforeUpdate(g,o),r.__reactInternalSnapshotBeforeUpdate=e}catch(e){ac(n,n.return,e)}}break;case 3:if(1024&e)if(9===(n=(e=t.stateNode.containerInfo).nodeType))lp(e);else if(1===n)switch(e.nodeName){case"HEAD":case"HTML":case"BODY":lp(e);break;default:e.textContent=""}break;default:if(1024&e)throw Error(i(163))}if(null!==(e=t.sibling)){e.return=t.return,rs=e;break}rs=t.return}g=as,as=!1}(e,u),gs(u,e),nr(Xc,e.containerInfo),Wp=!!Qc,Xc=Qc=null,e.current=u,os(e,u.alternate,u),le(),tu=l,U.p=o,L.T=n):e.current=u,Cu?(Cu=!1,_u=e,Eu=r):tc(e,c),0===(c=e.pendingLanes)&&(Su=null),function(e){if(be&&"function"==typeof be.onCommitFiberRoot)try{be.onCommitFiberRoot(ye,e,void 0,!(128&~e.current.flags))}catch(e){}}(u.stateNode),mc(e),null!==t)for(a=e.onRecoverableError,u=0;u<t.length;u++)a((c=t[u]).value,{componentStack:c.stack});!!(3&Eu)&&nc(),c=e.pendingLanes,4194218&r&&42&c?e===Lu?Au++:(Au=0,Lu=e):Au=0,yc(0,!1)}(e,t,n,r,d,a,o,l)}finally{L.T=p,U.p=d}}function tc(e,t){0==(e.pooledCacheLanes&=t)&&null!=(t=e.pooledCache)&&(e.pooledCache=null,Ua(t))}function nc(){if(null!==_u){var e=_u,t=Nu;Nu=0;var n=Oe(Eu),r=L.T,a=U.p;try{if(U.p=32>n?32:n,L.T=null,null===_u)var o=!1;else{n=Pu,Pu=null;var l=_u,s=Eu;if(_u=null,Eu=0,6&tu)throw Error(i(331));var u=tu;if(tu|=4,Rs(l.current),Cs(l,l.current,s,n),tu=u,yc(0,!1),be&&"function"==typeof be.onPostCommitFiberRoot)try{be.onPostCommitFiberRoot(ye,l)}catch(e){}o=!0}return o}finally{U.p=a,L.T=r,tc(e,t)}}return!1}function rc(e,t,n){t=Or(n,t),null!==(e=Il(e,t=Bi(e.stateNode,t,2),2))&&(Re(e,2),mc(e))}function ac(e,t,n){if(3===e.tag)rc(e,e,n);else for(;null!==t;){if(3===t.tag){rc(t,e,n);break}if(1===t.tag){var r=t.stateNode;if("function"==typeof t.type.getDerivedStateFromError||"function"==typeof r.componentDidCatch&&(null===Su||!Su.has(r))){e=Or(n,e),null!==(r=Il(t,n=Ui(2),2))&&(qi(n,r,t,e),Re(r,2),mc(r));break}}t=t.return}}function oc(e,t,n){var r=e.pingCache;if(null===r){r=e.pingCache=new eu;var a=new Set;r.set(t,a)}else void 0===(a=r.get(t))&&(a=new Set,r.set(t,a));a.has(n)||(uu=!0,a.add(n),e=ic.bind(null,e,t,n),t.then(e,e))}function ic(e,t,n){var r=e.pingCache;null!==r&&r.delete(t),e.pingedLanes|=e.suspendedLanes&n,e.warmLanes&=~n,nu===e&&(au&n)===n&&(4===pu||3===pu&&(62914560&au)===au&&300>se()-wu?!(2&tu)&&Bu(e,0):hu|=n,mu===au&&(mu=0)),mc(e)}function lc(e,t){0===t&&(t=Le()),null!==(e=Lr(e,t))&&(Re(e,t),mc(e))}function sc(e){var t=e.memoizedState,n=0;null!==t&&(n=t.retryLane),lc(e,n)}function uc(e,t){var n=0;switch(e.tag){case 13:var r=e.stateNode,a=e.memoizedState;null!==a&&(n=a.retryLane);break;case 19:r=e.stateNode;break;case 22:r=e.stateNode._retryCache;break;default:throw Error(i(314))}null!==r&&r.delete(t),lc(e,n)}var cc=null,pc=null,dc=!1,fc=!1,hc=!1,gc=0;function mc(e){var t;e!==pc&&null===e.next&&(null===pc?cc=pc=e:pc=pc.next=e),fc=!0,dc||(dc=!0,t=bc,ap((function(){6&tu?ae(ce,t):t()})))}function yc(e,t){if(!hc&&fc){hc=!0;do{for(var n=!1,r=cc;null!==r;){if(!t)if(0!==e){var a=r.pendingLanes;if(0===a)var o=0;else{var i=r.suspendedLanes,l=r.pingedLanes;o=(1<<31-we(42|e)+1)-1,o=201326677&(o&=a&~(i&~l))?201326677&o|1:o?2|o:0}0!==o&&(n=!0,xc(r,o))}else o=au,!(3&(o=Ee(r,r===nu?o:0)))||Ne(r,o)||(n=!0,xc(r,o));r=r.next}}while(n);hc=!1}}function bc(){fc=dc=!1;var e,t=0;0!==gc&&(((e=window.event)&&"popstate"===e.type?e!==ep&&(ep=e,!0):(ep=null,!1))&&(t=gc),gc=0);for(var n=se(),r=null,a=cc;null!==a;){var o=a.next,i=vc(a,n);0===i?(a.next=null,null===r?cc=o:r.next=o,null===o&&(pc=r)):(r=a,(0!==t||3&i)&&(fc=!0)),a=o}yc(t,!1)}function vc(e,t){for(var n=e.suspendedLanes,r=e.pingedLanes,a=e.expirationTimes,o=-62914561&e.pendingLanes;0<o;){var i=31-we(o),l=1<<i,s=a[i];-1===s?l&n&&!(l&r)||(a[i]=Pe(l,t)):s<=t&&(e.expiredLanes|=l),o&=~l}if(n=au,n=Ee(e,e===(t=nu)?n:0),r=e.callbackNode,0===n||e===t&&2===ou||null!==e.cancelPendingCommit)return null!==r&&null!==r&&oe(r),e.callbackNode=null,e.callbackPriority=0;if(!(3&n)||Ne(e,n)){if((t=n&-n)===e.callbackPriority)return t;switch(null!==r&&oe(r),Oe(n)){case 2:case 8:n=pe;break;case 32:default:n=de;break;case 268435456:n=he}return r=wc.bind(null,e),n=ae(n,r),e.callbackPriority=t,e.callbackNode=n,t}return null!==r&&null!==r&&oe(r),e.callbackPriority=2,e.callbackNode=null,2}function wc(e,t){var n=e.callbackNode;if(nc()&&e.callbackNode!==n)return null;var r=au;return 0===(r=Ee(e,e===nu?r:0))?null:(Mu(e,r,t),vc(e,se()),null!=e.callbackNode&&e.callbackNode===n?wc.bind(null,e):null)}function xc(e,t){if(nc())return null;Mu(e,t,!0)}function kc(){return 0===gc&&(gc=Ae()),gc}function Sc(e){return null==e||"symbol"==typeof e||"boolean"==typeof e?null:"function"==typeof e?e:At(""+e)}function Cc(e,t){var n=t.ownerDocument.createElement("input");return n.name=t.name,n.value=t.value,e.id&&n.setAttribute("form",e.id),t.parentNode.insertBefore(n,t),e=new FormData(e),n.parentNode.removeChild(n),e}for(var _c=0;_c<kr.length;_c++){var Ec=kr[_c];Sr(Ec.toLowerCase(),"on"+(Ec[0].toUpperCase()+Ec.slice(1)))}Sr(hr,"onAnimationEnd"),Sr(gr,"onAnimationIteration"),Sr(mr,"onAnimationStart"),Sr("dblclick","onDoubleClick"),Sr("focusin","onFocus"),Sr("focusout","onBlur"),Sr(yr,"onTransitionRun"),Sr(br,"onTransitionStart"),Sr(vr,"onTransitionCancel"),Sr(wr,"onTransitionEnd"),tt("onMouseEnter",["mouseout","mouseover"]),tt("onMouseLeave",["mouseout","mouseover"]),tt("onPointerEnter",["pointerout","pointerover"]),tt("onPointerLeave",["pointerout","pointerover"]),et("onChange","change click focusin focusout input keydown keyup selectionchange".split(" ")),et("onSelect","focusout contextmenu dragend focusin keydown keyup mousedown mouseup selectionchange".split(" ")),et("onBeforeInput",["compositionend","keypress","textInput","paste"]),et("onCompositionEnd","compositionend focusout keydown keypress keyup mousedown".split(" ")),et("onCompositionStart","compositionstart focusout keydown keypress keyup mousedown".split(" ")),et("onCompositionUpdate","compositionupdate focusout keydown keypress keyup mousedown".split(" "));var Nc="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange resize seeked seeking stalled suspend timeupdate volumechange waiting".split(" "),Pc=new Set("beforetoggle cancel close invalid load scroll scrollend toggle".split(" ").concat(Nc));function Ac(e,t){t=!!(4&t);for(var n=0;n<e.length;n++){var r=e[n],a=r.event;r=r.listeners;e:{var o=void 0;if(t)for(var i=r.length-1;0<=i;i--){var l=r[i],s=l.instance,u=l.currentTarget;if(l=l.listener,s!==o&&a.isPropagationStopped())break e;o=l,a.currentTarget=u;try{o(a)}catch(e){Oi(e)}a.currentTarget=null,o=s}else for(i=0;i<r.length;i++){if(s=(l=r[i]).instance,u=l.currentTarget,l=l.listener,s!==o&&a.isPropagationStopped())break e;o=l,a.currentTarget=u;try{o(a)}catch(e){Oi(e)}a.currentTarget=null,o=s}}}}function Lc(e,t){var n=t[Be];void 0===n&&(n=t[Be]=new Set);var r=e+"__bubble";n.has(r)||(Mc(t,e,2,!1),n.add(r))}function Tc(e,t,n){var r=0;t&&(r|=4),Mc(n,e,r,t)}var Rc="_reactListening"+Math.random().toString(36).slice(2);function Ic(e){if(!e[Rc]){e[Rc]=!0,Ze.forEach((function(t){"selectionchange"!==t&&(Pc.has(t)||Tc(t,!1,e),Tc(t,!0,e))}));var t=9===e.nodeType?e:e.ownerDocument;null===t||t[Rc]||(t[Rc]=!0,Tc("selectionchange",!1,t))}}function Mc(e,t,n,r){switch(Zp(t)){case 2:var a=Hp;break;case 8:a=Vp;break;default:a=Qp}n=a.bind(null,t,n,e),a=void 0,!$t||"touchstart"!==t&&"touchmove"!==t&&"wheel"!==t||(a=!0),r?void 0!==a?e.addEventListener(t,n,{capture:!0,passive:a}):e.addEventListener(t,n,!0):void 0!==a?e.addEventListener(t,n,{passive:a}):e.addEventListener(t,n,!1)}function Oc(e,t,n,r,a){var o=r;if(!(1&t||2&t||null===r))e:for(;;){if(null===r)return;var i=r.tag;if(3===i||4===i){var l=r.stateNode.containerInfo;if(l===a||8===l.nodeType&&l.parentNode===a)break;if(4===i)for(i=r.return;null!==i;){var s=i.tag;if((3===s||4===s)&&((s=i.stateNode.containerInfo)===a||8===s.nodeType&&s.parentNode===a))return;i=i.return}for(;null!==l;){if(null===(i=Ve(l)))return;if(5===(s=i.tag)||6===s||26===s||27===s){r=o=i;continue e}l=l.parentNode}}r=r.return}jt((function(){var r=o,a=Tt(n),i=[];e:{var l=xr.get(e);if(void 0!==l){var s=Zt,u=e;switch(e){case"keypress":if(0===Gt(n))break e;case"keydown":case"keyup":s=gn;break;case"focusin":u="focus",s=an;break;case"focusout":u="blur",s=an;break;case"beforeblur":case"afterblur":s=an;break;case"click":if(2===n.button)break e;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":s=nn;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":s=rn;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":s=yn;break;case hr:case gr:case mr:s=on;break;case wr:s=bn;break;case"scroll":case"scrollend":s=en;break;case"wheel":s=vn;break;case"copy":case"cut":case"paste":s=ln;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":s=mn;break;case"toggle":case"beforetoggle":s=wn}var c=!!(4&t),p=!c&&("scroll"===e||"scrollend"===e),d=c?null!==l?l+"Capture":null:l;c=[];for(var f,h=r;null!==h;){var g=h;if(f=g.stateNode,5!==(g=g.tag)&&26!==g&&27!==g||null===f||null===d||null!=(g=zt(h,d))&&c.push(jc(h,g,f)),p)break;h=h.return}0<c.length&&(l=new s(l,u,null,n,a),i.push({event:l,listeners:c}))}}if(!(7&t)){if(s="mouseout"===e||"pointerout"===e,(!(l="mouseover"===e||"pointerover"===e)||n===Lt||!(u=n.relatedTarget||n.fromElement)||!Ve(u)&&!u[De])&&(s||l)&&(l=a.window===a?a:(l=a.ownerDocument)?l.defaultView||l.parentWindow:window,s?(s=r,null!==(u=(u=n.relatedTarget||n.toElement)?Ve(u):null)&&(p=z(u),c=u.tag,u!==p||5!==c&&27!==c&&6!==c)&&(u=null)):(s=null,u=r),s!==u)){if(c=nn,g="onMouseLeave",d="onMouseEnter",h="mouse","pointerout"!==e&&"pointerover"!==e||(c=mn,g="onPointerLeave",d="onPointerEnter",h="pointer"),p=null==s?l:Xe(s),f=null==u?l:Xe(u),(l=new c(g,h+"leave",s,n,a)).target=p,l.relatedTarget=f,g=null,Ve(a)===r&&((c=new c(d,h+"enter",u,n,a)).target=f,c.relatedTarget=p,g=c),p=g,s&&u)e:{for(d=u,h=0,f=c=s;f;f=$c(f))h++;for(f=0,g=d;g;g=$c(g))f++;for(;0<h-f;)c=$c(c),h--;for(;0<f-h;)d=$c(d),f--;for(;h--;){if(c===d||null!==d&&c===d.alternate)break e;c=$c(c),d=$c(d)}c=null}else c=null;null!==s&&Fc(i,l,s,c,!1),null!==u&&null!==p&&Fc(i,p,u,c,!0)}if("select"===(s=(l=r?Xe(r):window).nodeName&&l.nodeName.toLowerCase())||"input"===s&&"file"===l.type)var m=$n;else if(Rn(l))if(Fn)m=Qn;else{m=Hn;var y=Wn}else!(s=l.nodeName)||"input"!==s.toLowerCase()||"checkbox"!==l.type&&"radio"!==l.type?r&&Et(r.elementType)&&(m=$n):m=Vn;switch(m&&(m=m(e,r))?In(i,m,n,a):(y&&y(e,l,r),"focusout"===e&&r&&"number"===l.type&&null!=r.memoizedProps.value&&bt(l,"number",l.value)),y=r?Xe(r):window,e){case"focusin":(Rn(y)||"true"===y.contentEditable)&&(ar=y,or=r,ir=null);break;case"focusout":ir=or=ar=null;break;case"mousedown":lr=!0;break;case"contextmenu":case"mouseup":case"dragend":lr=!1,sr(i,n,a);break;case"selectionchange":if(rr)break;case"keydown":case"keyup":sr(i,n,a)}var b;if(kn)e:{switch(e){case"compositionstart":var v="onCompositionStart";break e;case"compositionend":v="onCompositionEnd";break e;case"compositionupdate":v="onCompositionUpdate";break e}v=void 0}else Ln?Pn(e,n)&&(v="onCompositionEnd"):"keydown"===e&&229===n.keyCode&&(v="onCompositionStart");v&&(_n&&"ko"!==n.locale&&(Ln||"onCompositionStart"!==v?"onCompositionEnd"===v&&Ln&&(b=qt()):(Bt="value"in(Dt=a)?Dt.value:Dt.textContent,Ln=!0)),0<(y=zc(r,v)).length&&(v=new sn(v,e,null,n,a),i.push({event:v,listeners:y}),(b||null!==(b=An(n)))&&(v.data=b))),(b=Cn?function(e,t){switch(e){case"compositionend":return An(t);case"keypress":return 32!==t.which?null:(Nn=!0,En);case"textInput":return(e=t.data)===En&&Nn?null:e;default:return null}}(e,n):function(e,t){if(Ln)return"compositionend"===e||!kn&&Pn(e,t)?(e=qt(),Ut=Bt=Dt=null,Ln=!1,e):null;switch(e){case"paste":default:return null;case"keypress":if(!(t.ctrlKey||t.altKey||t.metaKey)||t.ctrlKey&&t.altKey){if(t.char&&1<t.char.length)return t.char;if(t.which)return String.fromCharCode(t.which)}return null;case"compositionend":return _n&&"ko"!==t.locale?null:t.data}}(e,n))&&0<(v=zc(r,"onBeforeInput")).length&&(y=new sn("onBeforeInput","beforeinput",null,n,a),i.push({event:y,listeners:v}),y.data=b),function(e,t,n,r,a){if("submit"===t&&n&&n.stateNode===a){var o=Sc((a[Fe]||null).action),i=r.submitter;i&&null!==(t=(t=i[Fe]||null)?Sc(t.formAction):i.getAttribute("formAction"))&&(o=t,i=null);var l=new Zt("action","action",null,r,a);e.push({event:l,listeners:[{instance:null,listener:function(){if(r.defaultPrevented){if(0!==gc){var e=i?Cc(a,i):new FormData(a);di(n,{pending:!0,data:e,method:a.method,action:o},null,e)}}else"function"==typeof o&&(l.preventDefault(),e=i?Cc(a,i):new FormData(a),di(n,{pending:!0,data:e,method:a.method,action:o},o,e))},currentTarget:a}]})}}(i,e,r,n,a)}Ac(i,t)}))}function jc(e,t,n){return{instance:e,listener:t,currentTarget:n}}function zc(e,t){for(var n=t+"Capture",r=[];null!==e;){var a=e,o=a.stateNode;5!==(a=a.tag)&&26!==a&&27!==a||null===o||(null!=(a=zt(e,n))&&r.unshift(jc(e,a,o)),null!=(a=zt(e,t))&&r.push(jc(e,a,o))),e=e.return}return r}function $c(e){if(null===e)return null;do{e=e.return}while(e&&5!==e.tag&&27!==e.tag);return e||null}function Fc(e,t,n,r,a){for(var o=t._reactName,i=[];null!==n&&n!==r;){var l=n,s=l.alternate,u=l.stateNode;if(l=l.tag,null!==s&&s===r)break;5!==l&&26!==l&&27!==l||null===u||(s=u,a?null!=(u=zt(n,o))&&i.unshift(jc(n,u,s)):a||null!=(u=zt(n,o))&&i.push(jc(n,u,s))),n=n.return}0!==i.length&&e.push({event:t,listeners:i})}var Dc=/\r\n?/g,Bc=/\u0000|\uFFFD/g;function Uc(e){return("string"==typeof e?e:""+e).replace(Dc,"\n").replace(Bc,"")}function qc(e,t){return t=Uc(t),Uc(e)===t}function Gc(){}function Wc(e,t,n,r,a,o){switch(n){case"children":"string"==typeof r?"body"===t||"textarea"===t&&""===r||kt(e,r):("number"==typeof r||"bigint"==typeof r)&&"body"!==t&&kt(e,""+r);break;case"className":lt(e,"class",r);break;case"tabIndex":lt(e,"tabindex",r);break;case"dir":case"role":case"viewBox":case"width":case"height":lt(e,n,r);break;case"style":_t(e,r,o);break;case"data":if("object"!==t){lt(e,"data",r);break}case"src":case"href":if(""===r&&("a"!==t||"href"!==n)){e.removeAttribute(n);break}if(null==r||"function"==typeof r||"symbol"==typeof r||"boolean"==typeof r){e.removeAttribute(n);break}r=At(""+r),e.setAttribute(n,r);break;case"action":case"formAction":if("function"==typeof r){e.setAttribute(n,"javascript:throw new Error('A React form was unexpectedly submitted. If you called form.submit() manually, consider using form.requestSubmit() instead. If you\\'re trying to use event.stopPropagation() in a submit event handler, consider also calling event.preventDefault().')");break}if("function"==typeof o&&("formAction"===n?("input"!==t&&Wc(e,t,"name",a.name,a,null),Wc(e,t,"formEncType",a.formEncType,a,null),Wc(e,t,"formMethod",a.formMethod,a,null),Wc(e,t,"formTarget",a.formTarget,a,null)):(Wc(e,t,"encType",a.encType,a,null),Wc(e,t,"method",a.method,a,null),Wc(e,t,"target",a.target,a,null))),null==r||"symbol"==typeof r||"boolean"==typeof r){e.removeAttribute(n);break}r=At(""+r),e.setAttribute(n,r);break;case"onClick":null!=r&&(e.onclick=Gc);break;case"onScroll":null!=r&&Lc("scroll",e);break;case"onScrollEnd":null!=r&&Lc("scrollend",e);break;case"dangerouslySetInnerHTML":if(null!=r){if("object"!=typeof r||!("__html"in r))throw Error(i(61));if(null!=(n=r.__html)){if(null!=a.children)throw Error(i(60));e.innerHTML=n}}break;case"multiple":e.multiple=r&&"function"!=typeof r&&"symbol"!=typeof r;break;case"muted":e.muted=r&&"function"!=typeof r&&"symbol"!=typeof r;break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"defaultValue":case"defaultChecked":case"innerHTML":case"ref":case"autoFocus":break;case"xlinkHref":if(null==r||"function"==typeof r||"boolean"==typeof r||"symbol"==typeof r){e.removeAttribute("xlink:href");break}n=At(""+r),e.setAttributeNS("http://www.w3.org/1999/xlink","xlink:href",n);break;case"contentEditable":case"spellCheck":case"draggable":case"value":case"autoReverse":case"externalResourcesRequired":case"focusable":case"preserveAlpha":null!=r&&"function"!=typeof r&&"symbol"!=typeof r?e.setAttribute(n,""+r):e.removeAttribute(n);break;case"inert":case"allowFullScreen":case"async":case"autoPlay":case"controls":case"default":case"defer":case"disabled":case"disablePictureInPicture":case"disableRemotePlayback":case"formNoValidate":case"hidden":case"loop":case"noModule":case"noValidate":case"open":case"playsInline":case"readOnly":case"required":case"reversed":case"scoped":case"seamless":case"itemScope":r&&"function"!=typeof r&&"symbol"!=typeof r?e.setAttribute(n,""):e.removeAttribute(n);break;case"capture":case"download":!0===r?e.setAttribute(n,""):!1!==r&&null!=r&&"function"!=typeof r&&"symbol"!=typeof r?e.setAttribute(n,r):e.removeAttribute(n);break;case"cols":case"rows":case"size":case"span":null!=r&&"function"!=typeof r&&"symbol"!=typeof r&&!isNaN(r)&&1<=r?e.setAttribute(n,r):e.removeAttribute(n);break;case"rowSpan":case"start":null==r||"function"==typeof r||"symbol"==typeof r||isNaN(r)?e.removeAttribute(n):e.setAttribute(n,r);break;case"popover":Lc("beforetoggle",e),Lc("toggle",e),it(e,"popover",r);break;case"xlinkActuate":st(e,"http://www.w3.org/1999/xlink","xlink:actuate",r);break;case"xlinkArcrole":st(e,"http://www.w3.org/1999/xlink","xlink:arcrole",r);break;case"xlinkRole":st(e,"http://www.w3.org/1999/xlink","xlink:role",r);break;case"xlinkShow":st(e,"http://www.w3.org/1999/xlink","xlink:show",r);break;case"xlinkTitle":st(e,"http://www.w3.org/1999/xlink","xlink:title",r);break;case"xlinkType":st(e,"http://www.w3.org/1999/xlink","xlink:type",r);break;case"xmlBase":st(e,"http://www.w3.org/XML/1998/namespace","xml:base",r);break;case"xmlLang":st(e,"http://www.w3.org/XML/1998/namespace","xml:lang",r);break;case"xmlSpace":st(e,"http://www.w3.org/XML/1998/namespace","xml:space",r);break;case"is":it(e,"is",r);break;case"innerText":case"textContent":break;default:(!(2<n.length)||"o"!==n[0]&&"O"!==n[0]||"n"!==n[1]&&"N"!==n[1])&&it(e,n=Nt.get(n)||n,r)}}function Hc(e,t,n,r,a,o){switch(n){case"style":_t(e,r,o);break;case"dangerouslySetInnerHTML":if(null!=r){if("object"!=typeof r||!("__html"in r))throw Error(i(61));if(null!=(n=r.__html)){if(null!=a.children)throw Error(i(60));e.innerHTML=n}}break;case"children":"string"==typeof r?kt(e,r):("number"==typeof r||"bigint"==typeof r)&&kt(e,""+r);break;case"onScroll":null!=r&&Lc("scroll",e);break;case"onScrollEnd":null!=r&&Lc("scrollend",e);break;case"onClick":null!=r&&(e.onclick=Gc);break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"innerHTML":case"ref":case"innerText":case"textContent":break;default:Je.hasOwnProperty(n)||("o"!==n[0]||"n"!==n[1]||(a=n.endsWith("Capture"),t=n.slice(2,a?n.length-7:void 0),"function"==typeof(o=null!=(o=e[Fe]||null)?o[n]:null)&&e.removeEventListener(t,o,a),"function"!=typeof r)?n in e?e[n]=r:!0===r?e.setAttribute(n,""):it(e,n,r):("function"!=typeof o&&null!==o&&(n in e?e[n]=null:e.hasAttribute(n)&&e.removeAttribute(n)),e.addEventListener(t,r,a)))}}function Vc(e,t,n){switch(t){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"img":Lc("error",e),Lc("load",e);var r,a=!1,o=!1;for(r in n)if(n.hasOwnProperty(r)){var l=n[r];if(null!=l)switch(r){case"src":a=!0;break;case"srcSet":o=!0;break;case"children":case"dangerouslySetInnerHTML":throw Error(i(137,t));default:Wc(e,t,r,l,n,null)}}return o&&Wc(e,t,"srcSet",n.srcSet,n,null),void(a&&Wc(e,t,"src",n.src,n,null));case"input":Lc("invalid",e);var s=r=l=o=null,u=null,c=null;for(a in n)if(n.hasOwnProperty(a)){var p=n[a];if(null!=p)switch(a){case"name":o=p;break;case"type":l=p;break;case"checked":u=p;break;case"defaultChecked":c=p;break;case"value":r=p;break;case"defaultValue":s=p;break;case"children":case"dangerouslySetInnerHTML":if(null!=p)throw Error(i(137,t));break;default:Wc(e,t,a,p,n,null)}}return yt(e,r,s,u,c,l,o,!1),void pt(e);case"select":for(o in Lc("invalid",e),a=l=r=null,n)if(n.hasOwnProperty(o)&&null!=(s=n[o]))switch(o){case"value":r=s;break;case"defaultValue":l=s;break;case"multiple":a=s;default:Wc(e,t,o,s,n,null)}return t=r,n=l,e.multiple=!!a,void(null!=t?vt(e,!!a,t,!1):null!=n&&vt(e,!!a,n,!0));case"textarea":for(l in Lc("invalid",e),r=o=a=null,n)if(n.hasOwnProperty(l)&&null!=(s=n[l]))switch(l){case"value":a=s;break;case"defaultValue":o=s;break;case"children":r=s;break;case"dangerouslySetInnerHTML":if(null!=s)throw Error(i(91));break;default:Wc(e,t,l,s,n,null)}return xt(e,a,o,r),void pt(e);case"option":for(u in n)n.hasOwnProperty(u)&&null!=(a=n[u])&&("selected"===u?e.selected=a&&"function"!=typeof a&&"symbol"!=typeof a:Wc(e,t,u,a,n,null));return;case"dialog":Lc("cancel",e),Lc("close",e);break;case"iframe":case"object":Lc("load",e);break;case"video":case"audio":for(a=0;a<Nc.length;a++)Lc(Nc[a],e);break;case"image":Lc("error",e),Lc("load",e);break;case"details":Lc("toggle",e);break;case"embed":case"source":case"link":Lc("error",e),Lc("load",e);case"area":case"base":case"br":case"col":case"hr":case"keygen":case"meta":case"param":case"track":case"wbr":case"menuitem":for(c in n)if(n.hasOwnProperty(c)&&null!=(a=n[c]))switch(c){case"children":case"dangerouslySetInnerHTML":throw Error(i(137,t));default:Wc(e,t,c,a,n,null)}return;default:if(Et(t)){for(p in n)n.hasOwnProperty(p)&&void 0!==(a=n[p])&&Hc(e,t,p,a,n,void 0);return}}for(s in n)n.hasOwnProperty(s)&&null!=(a=n[s])&&Wc(e,t,s,a,n,null)}var Qc=null,Xc=null;function Kc(e){return 9===e.nodeType?e:e.ownerDocument}function Yc(e){switch(e){case"http://www.w3.org/2000/svg":return 1;case"http://www.w3.org/1998/Math/MathML":return 2;default:return 0}}function Zc(e,t){if(0===e)switch(t){case"svg":return 1;case"math":return 2;default:return 0}return 1===e&&"foreignObject"===t?0:e}function Jc(e,t){return"textarea"===e||"noscript"===e||"string"==typeof t.children||"number"==typeof t.children||"bigint"==typeof t.children||"object"==typeof t.dangerouslySetInnerHTML&&null!==t.dangerouslySetInnerHTML&&null!=t.dangerouslySetInnerHTML.__html}var ep=null,tp="function"==typeof setTimeout?setTimeout:void 0,np="function"==typeof clearTimeout?clearTimeout:void 0,rp="function"==typeof Promise?Promise:void 0,ap="function"==typeof queueMicrotask?queueMicrotask:void 0!==rp?function(e){return rp.resolve(null).then(e).catch(op)}:tp;function op(e){setTimeout((function(){throw e}))}function ip(e,t){var n=t,r=0;do{var a=n.nextSibling;if(e.removeChild(n),a&&8===a.nodeType)if("/$"===(n=a.data)){if(0===r)return e.removeChild(a),void md(t);r--}else"$"!==n&&"$?"!==n&&"$!"!==n||r++;n=a}while(n);md(t)}function lp(e){var t=e.firstChild;for(t&&10===t.nodeType&&(t=t.nextSibling);t;){var n=t;switch(t=t.nextSibling,n.nodeName){case"HTML":case"HEAD":case"BODY":lp(n),He(n);continue;case"SCRIPT":case"STYLE":continue;case"LINK":if("stylesheet"===n.rel.toLowerCase())continue}e.removeChild(n)}}function sp(e){for(;null!=e;e=e.nextSibling){var t=e.nodeType;if(1===t||3===t)break;if(8===t){if("$"===(t=e.data)||"$!"===t||"$?"===t||"F!"===t||"F"===t)break;if("/$"===t)return null}}return e}function up(e){e=e.previousSibling;for(var t=0;e;){if(8===e.nodeType){var n=e.data;if("$"===n||"$!"===n||"$?"===n){if(0===t)return e;t--}else"/$"===n&&t++}e=e.previousSibling}return null}function cp(e,t,n){switch(t=Kc(n),e){case"html":if(!(e=t.documentElement))throw Error(i(452));return e;case"head":if(!(e=t.head))throw Error(i(453));return e;case"body":if(!(e=t.body))throw Error(i(454));return e;default:throw Error(i(451))}}var pp=new Map,dp=new Set;function fp(e){return"function"==typeof e.getRootNode?e.getRootNode():e.ownerDocument}var hp=U.d;U.d={f:function(){var e=hp.f(),t=Fu();return e||t},r:function(e){var t=Qe(e);null!==t&&5===t.tag&&"form"===t.type?hi(t):hp.r(e)},D:function(e){hp.D(e),mp("dns-prefetch",e,null)},C:function(e,t){hp.C(e,t),mp("preconnect",e,t)},L:function(e,t,n){hp.L(e,t,n);var r=gp;if(r&&e&&t){var a='link[rel="preload"][as="'+gt(t)+'"]';"image"===t&&n&&n.imageSrcSet?(a+='[imagesrcset="'+gt(n.imageSrcSet)+'"]',"string"==typeof n.imageSizes&&(a+='[imagesizes="'+gt(n.imageSizes)+'"]')):a+='[href="'+gt(e)+'"]';var o=a;switch(t){case"style":o=bp(e);break;case"script":o=xp(e)}pp.has(o)||(e=T({rel:"preload",href:"image"===t&&n&&n.imageSrcSet?void 0:e,as:t},n),pp.set(o,e),null!==r.querySelector(a)||"style"===t&&r.querySelector(vp(o))||"script"===t&&r.querySelector(kp(o))||(Vc(t=r.createElement("link"),"link",e),Ye(t),r.head.appendChild(t)))}},m:function(e,t){hp.m(e,t);var n=gp;if(n&&e){var r=t&&"string"==typeof t.as?t.as:"script",a='link[rel="modulepreload"][as="'+gt(r)+'"][href="'+gt(e)+'"]',o=a;switch(r){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":o=xp(e)}if(!pp.has(o)&&(e=T({rel:"modulepreload",href:e},t),pp.set(o,e),null===n.querySelector(a))){switch(r){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":if(n.querySelector(kp(o)))return}Vc(r=n.createElement("link"),"link",e),Ye(r),n.head.appendChild(r)}}},X:function(e,t){hp.X(e,t);var n=gp;if(n&&e){var r=Ke(n).hoistableScripts,a=xp(e),o=r.get(a);o||((o=n.querySelector(kp(a)))||(e=T({src:e,async:!0},t),(t=pp.get(a))&&Ep(e,t),Ye(o=n.createElement("script")),Vc(o,"link",e),n.head.appendChild(o)),o={type:"script",instance:o,count:1,state:null},r.set(a,o))}},S:function(e,t,n){hp.S(e,t,n);var r=gp;if(r&&e){var a=Ke(r).hoistableStyles,o=bp(e);t=t||"default";var i=a.get(o);if(!i){var l={loading:0,preload:null};if(i=r.querySelector(vp(o)))l.loading=5;else{e=T({rel:"stylesheet",href:e,"data-precedence":t},n),(n=pp.get(o))&&_p(e,n);var s=i=r.createElement("link");Ye(s),Vc(s,"link",e),s._p=new Promise((function(e,t){s.onload=e,s.onerror=t})),s.addEventListener("load",(function(){l.loading|=1})),s.addEventListener("error",(function(){l.loading|=2})),l.loading|=4,Cp(i,t,r)}i={type:"stylesheet",instance:i,count:1,state:l},a.set(o,i)}}},M:function(e,t){hp.M(e,t);var n=gp;if(n&&e){var r=Ke(n).hoistableScripts,a=xp(e),o=r.get(a);o||((o=n.querySelector(kp(a)))||(e=T({src:e,async:!0,type:"module"},t),(t=pp.get(a))&&Ep(e,t),Ye(o=n.createElement("script")),Vc(o,"link",e),n.head.appendChild(o)),o={type:"script",instance:o,count:1,state:null},r.set(a,o))}}};var gp="undefined"==typeof document?null:document;function mp(e,t,n){var r=gp;if(r&&"string"==typeof t&&t){var a=gt(t);a='link[rel="'+e+'"][href="'+a+'"]',"string"==typeof n&&(a+='[crossorigin="'+n+'"]'),dp.has(a)||(dp.add(a),e={rel:e,crossOrigin:n,href:t},null===r.querySelector(a)&&(Vc(t=r.createElement("link"),"link",e),Ye(t),r.head.appendChild(t)))}}function yp(e,t,n,r){var a=(a=Y.current)?fp(a):null;if(!a)throw Error(i(446));switch(e){case"meta":case"title":return null;case"style":return"string"==typeof n.precedence&&"string"==typeof n.href?(t=bp(n.href),(r=(n=Ke(a).hoistableStyles).get(t))||(r={type:"style",instance:null,count:0,state:null},n.set(t,r)),r):{type:"void",instance:null,count:0,state:null};case"link":if("stylesheet"===n.rel&&"string"==typeof n.href&&"string"==typeof n.precedence){e=bp(n.href);var o=Ke(a).hoistableStyles,l=o.get(e);if(l||(a=a.ownerDocument||a,l={type:"stylesheet",instance:null,count:0,state:{loading:0,preload:null}},o.set(e,l),(o=a.querySelector(vp(e)))&&!o._p&&(l.instance=o,l.state.loading=5),pp.has(e)||(n={rel:"preload",as:"style",href:n.href,crossOrigin:n.crossOrigin,integrity:n.integrity,media:n.media,hrefLang:n.hrefLang,referrerPolicy:n.referrerPolicy},pp.set(e,n),o||function(e,t,n,r){e.querySelector('link[rel="preload"][as="style"]['+t+"]")?r.loading=1:(t=e.createElement("link"),r.preload=t,t.addEventListener("load",(function(){return r.loading|=1})),t.addEventListener("error",(function(){return r.loading|=2})),Vc(t,"link",n),Ye(t),e.head.appendChild(t))}(a,e,n,l.state))),t&&null===r)throw Error(i(528,""));return l}if(t&&null!==r)throw Error(i(529,""));return null;case"script":return t=n.async,"string"==typeof(n=n.src)&&t&&"function"!=typeof t&&"symbol"!=typeof t?(t=xp(n),(r=(n=Ke(a).hoistableScripts).get(t))||(r={type:"script",instance:null,count:0,state:null},n.set(t,r)),r):{type:"void",instance:null,count:0,state:null};default:throw Error(i(444,e))}}function bp(e){return'href="'+gt(e)+'"'}function vp(e){return'link[rel="stylesheet"]['+e+"]"}function wp(e){return T({},e,{"data-precedence":e.precedence,precedence:null})}function xp(e){return'[src="'+gt(e)+'"]'}function kp(e){return"script[async]"+e}function Sp(e,t,n){if(t.count++,null===t.instance)switch(t.type){case"style":var r=e.querySelector('style[data-href~="'+gt(n.href)+'"]');if(r)return t.instance=r,Ye(r),r;var a=T({},n,{"data-href":n.href,"data-precedence":n.precedence,href:null,precedence:null});return Ye(r=(e.ownerDocument||e).createElement("style")),Vc(r,"style",a),Cp(r,n.precedence,e),t.instance=r;case"stylesheet":a=bp(n.href);var o=e.querySelector(vp(a));if(o)return t.state.loading|=4,t.instance=o,Ye(o),o;r=wp(n),(a=pp.get(a))&&_p(r,a),Ye(o=(e.ownerDocument||e).createElement("link"));var l=o;return l._p=new Promise((function(e,t){l.onload=e,l.onerror=t})),Vc(o,"link",r),t.state.loading|=4,Cp(o,n.precedence,e),t.instance=o;case"script":return o=xp(n.src),(a=e.querySelector(kp(o)))?(t.instance=a,Ye(a),a):(r=n,(a=pp.get(o))&&Ep(r=T({},n),a),Ye(a=(e=e.ownerDocument||e).createElement("script")),Vc(a,"link",r),e.head.appendChild(a),t.instance=a);case"void":return null;default:throw Error(i(443,t.type))}else"stylesheet"===t.type&&!(4&t.state.loading)&&(r=t.instance,t.state.loading|=4,Cp(r,n.precedence,e));return t.instance}function Cp(e,t,n){for(var r=n.querySelectorAll('link[rel="stylesheet"][data-precedence],style[data-precedence]'),a=r.length?r[r.length-1]:null,o=a,i=0;i<r.length;i++){var l=r[i];if(l.dataset.precedence===t)o=l;else if(o!==a)break}o?o.parentNode.insertBefore(e,o.nextSibling):(t=9===n.nodeType?n.head:n).insertBefore(e,t.firstChild)}function _p(e,t){null==e.crossOrigin&&(e.crossOrigin=t.crossOrigin),null==e.referrerPolicy&&(e.referrerPolicy=t.referrerPolicy),null==e.title&&(e.title=t.title)}function Ep(e,t){null==e.crossOrigin&&(e.crossOrigin=t.crossOrigin),null==e.referrerPolicy&&(e.referrerPolicy=t.referrerPolicy),null==e.integrity&&(e.integrity=t.integrity)}var Np=null;function Pp(e,t,n){if(null===Np){var r=new Map,a=Np=new Map;a.set(n,r)}else(r=(a=Np).get(n))||(r=new Map,a.set(n,r));if(r.has(e))return r;for(r.set(e,null),n=n.getElementsByTagName(e),a=0;a<n.length;a++){var o=n[a];if(!(o[We]||o[$e]||"link"===e&&"stylesheet"===o.getAttribute("rel"))&&"http://www.w3.org/2000/svg"!==o.namespaceURI){var i=o.getAttribute(t)||"";i=e+i;var l=r.get(i);l?l.push(o):r.set(i,[o])}}return r}function Ap(e,t,n){(e=e.ownerDocument||e).head.insertBefore(n,"title"===t?e.querySelector("head > title"):null)}function Lp(e){return!!("stylesheet"!==e.type||3&e.state.loading)}var Tp=null;function Rp(){}function Ip(){if(this.count--,0===this.count)if(this.stylesheets)Op(this,this.stylesheets);else if(this.unsuspend){var e=this.unsuspend;this.unsuspend=null,e()}}var Mp=null;function Op(e,t){e.stylesheets=null,null!==e.unsuspend&&(e.count++,Mp=new Map,t.forEach(jp,e),Mp=null,Ip.call(e))}function jp(e,t){if(!(4&t.state.loading)){var n=Mp.get(e);if(n)var r=n.get(null);else{n=new Map,Mp.set(e,n);for(var a=e.querySelectorAll("link[data-precedence],style[data-precedence]"),o=0;o<a.length;o++){var i=a[o];"LINK"!==i.nodeName&&"not all"===i.getAttribute("media")||(n.set(i.dataset.precedence,i),r=i)}r&&n.set(null,r)}i=(a=t.instance).getAttribute("data-precedence"),(o=n.get(i)||r)===r&&n.set(null,a),n.set(i,a),this.count++,r=Ip.bind(this),a.addEventListener("load",r),a.addEventListener("error",r),o?o.parentNode.insertBefore(a,o.nextSibling):(e=9===e.nodeType?e.head:e).insertBefore(a,e.firstChild),t.state.loading|=4}}var zp={$$typeof:m,Provider:null,Consumer:null,_currentValue:q,_currentValue2:q,_threadCount:0};function $p(e,t,n,r,a,o,i,l){this.tag=1,this.containerInfo=e,this.finishedWork=this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=-1,this.callbackNode=this.next=this.pendingContext=this.context=this.cancelPendingCommit=null,this.callbackPriority=0,this.expirationTimes=Te(-1),this.entangledLanes=this.shellSuspendCounter=this.errorRecoveryDisabledLanes=this.finishedLanes=this.expiredLanes=this.warmLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=Te(0),this.hiddenUpdates=Te(null),this.identifierPrefix=r,this.onUncaughtError=a,this.onCaughtError=o,this.onRecoverableError=i,this.pooledCache=null,this.pooledCacheLanes=0,this.formState=l,this.incompleteTransitions=new Map}function Fp(e,t,n,r,a,o,i,l,s,u,c,p){return e=new $p(e,t,n,i,l,s,u,p),t=1,!0===o&&(t|=24),o=js(3,null,null,t),e.current=o,o.stateNode=e,(t=Ba()).refCount++,e.pooledCache=t,t.refCount++,o.memoizedState={element:r,isDehydrated:n,cache:t},Ll(o),e}function Dp(e){return e?e=Ir:Ir}function Bp(e,t,n,r,a,o){a=Dp(a),null===r.context?r.context=a:r.pendingContext=a,(r=Rl(t)).payload={element:n},null!==(o=void 0===o?null:o)&&(r.callback=o),null!==(n=Il(e,r,t))&&(Iu(n,0,t),Ml(n,e,t))}function Up(e,t){if(null!==(e=e.memoizedState)&&null!==e.dehydrated){var n=e.retryLane;e.retryLane=0!==n&&n<t?n:t}}function qp(e,t){Up(e,t),(e=e.alternate)&&Up(e,t)}function Gp(e){if(13===e.tag){var t=Lr(e,67108864);null!==t&&Iu(t,0,67108864),qp(e,67108864)}}var Wp=!0;function Hp(e,t,n,r){var a=L.T;L.T=null;var o=U.p;try{U.p=2,Qp(e,t,n,r)}finally{U.p=o,L.T=a}}function Vp(e,t,n,r){var a=L.T;L.T=null;var o=U.p;try{U.p=8,Qp(e,t,n,r)}finally{U.p=o,L.T=a}}function Qp(e,t,n,r){if(Wp){var a=Xp(r);if(null===a)Oc(e,t,r,Kp,n),ld(e,r);else if(function(e,t,n,r,a){switch(t){case"focusin":return ed=sd(ed,e,t,n,r,a),!0;case"dragenter":return td=sd(td,e,t,n,r,a),!0;case"mouseover":return nd=sd(nd,e,t,n,r,a),!0;case"pointerover":var o=a.pointerId;return rd.set(o,sd(rd.get(o)||null,e,t,n,r,a)),!0;case"gotpointercapture":return o=a.pointerId,ad.set(o,sd(ad.get(o)||null,e,t,n,r,a)),!0}return!1}(a,e,t,n,r))r.stopPropagation();else if(ld(e,r),4&t&&-1<id.indexOf(e)){for(;null!==a;){var o=Qe(a);if(null!==o)switch(o.tag){case 3:if((o=o.stateNode).current.memoizedState.isDehydrated){var i=_e(o.pendingLanes);if(0!==i){var l=o;for(l.pendingLanes|=2,l.entangledLanes|=2;i;){var s=1<<31-we(i);l.entanglements[1]|=s,i&=~s}mc(o),!(6&tu)&&(xu=se()+500,yc(0,!1))}}break;case 13:null!==(l=Lr(o,2))&&Iu(l,0,2),Fu(),qp(o,2)}if(null===(o=Xp(r))&&Oc(e,t,r,Kp,n),o===a)break;a=o}null!==a&&r.stopPropagation()}else Oc(e,t,r,null,n)}}function Xp(e){return Yp(e=Tt(e))}var Kp=null;function Yp(e){if(Kp=null,null!==(e=Ve(e))){var t=z(e);if(null===t)e=null;else{var n=t.tag;if(13===n){if(null!==(e=$(t)))return e;e=null}else if(3===n){if(t.stateNode.current.memoizedState.isDehydrated)return 3===t.tag?t.stateNode.containerInfo:null;e=null}else t!==e&&(e=null)}}return Kp=e,null}function Zp(e){switch(e){case"beforetoggle":case"cancel":case"click":case"close":case"contextmenu":case"copy":case"cut":case"auxclick":case"dblclick":case"dragend":case"dragstart":case"drop":case"focusin":case"focusout":case"input":case"invalid":case"keydown":case"keypress":case"keyup":case"mousedown":case"mouseup":case"paste":case"pause":case"play":case"pointercancel":case"pointerdown":case"pointerup":case"ratechange":case"reset":case"resize":case"seeked":case"submit":case"toggle":case"touchcancel":case"touchend":case"touchstart":case"volumechange":case"change":case"selectionchange":case"textInput":case"compositionstart":case"compositionend":case"compositionupdate":case"beforeblur":case"afterblur":case"beforeinput":case"blur":case"fullscreenchange":case"focus":case"hashchange":case"popstate":case"select":case"selectstart":return 2;case"drag":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"mousemove":case"mouseout":case"mouseover":case"pointermove":case"pointerout":case"pointerover":case"scroll":case"touchmove":case"wheel":case"mouseenter":case"mouseleave":case"pointerenter":case"pointerleave":return 8;case"message":switch(ue()){case ce:return 2;case pe:return 8;case de:case fe:return 32;case he:return 268435456;default:return 32}default:return 32}}var Jp=!1,ed=null,td=null,nd=null,rd=new Map,ad=new Map,od=[],id="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput copy cut paste click change contextmenu reset".split(" ");function ld(e,t){switch(e){case"focusin":case"focusout":ed=null;break;case"dragenter":case"dragleave":td=null;break;case"mouseover":case"mouseout":nd=null;break;case"pointerover":case"pointerout":rd.delete(t.pointerId);break;case"gotpointercapture":case"lostpointercapture":ad.delete(t.pointerId)}}function sd(e,t,n,r,a,o){return null===e||e.nativeEvent!==o?(e={blockedOn:t,domEventName:n,eventSystemFlags:r,nativeEvent:o,targetContainers:[a]},null!==t&&null!==(t=Qe(t))&&Gp(t),e):(e.eventSystemFlags|=r,t=e.targetContainers,null!==a&&-1===t.indexOf(a)&&t.push(a),e)}function ud(e){var t=Ve(e.target);if(null!==t){var n=z(t);if(null!==n)if(13===(t=n.tag)){if(null!==(t=$(n)))return e.blockedOn=t,void function(e){var t=U.p;try{return U.p=e,function(){if(13===n.tag){var e=Tu(),t=Lr(n,e);null!==t&&Iu(t,0,e),qp(n,e)}}()}finally{U.p=t}}(e.priority)}else if(3===t&&n.stateNode.current.memoizedState.isDehydrated)return void(e.blockedOn=3===n.tag?n.stateNode.containerInfo:null)}e.blockedOn=null}function cd(e){if(null!==e.blockedOn)return!1;for(var t=e.targetContainers;0<t.length;){var n=Xp(e.nativeEvent);if(null!==n)return null!==(t=Qe(n))&&Gp(t),e.blockedOn=n,!1;var r=new(n=e.nativeEvent).constructor(n.type,n);Lt=r,n.target.dispatchEvent(r),Lt=null,t.shift()}return!0}function pd(e,t,n){cd(e)&&n.delete(t)}function dd(){Jp=!1,null!==ed&&cd(ed)&&(ed=null),null!==td&&cd(td)&&(td=null),null!==nd&&cd(nd)&&(nd=null),rd.forEach(pd),ad.forEach(pd)}function fd(e,t){e.blockedOn===t&&(e.blockedOn=null,Jp||(Jp=!0,r.unstable_scheduleCallback(r.unstable_NormalPriority,dd)))}var hd=null;function gd(e){hd!==e&&(hd=e,r.unstable_scheduleCallback(r.unstable_NormalPriority,(function(){hd===e&&(hd=null);for(var t=0;t<e.length;t+=3){var n=e[t],r=e[t+1],a=e[t+2];if("function"!=typeof r){if(null===Yp(r||n))continue;break}var o=Qe(n);null!==o&&(e.splice(t,3),t-=3,di(o,{pending:!0,data:a,method:n.method,action:r},r,a))}})))}function md(e){function t(t){return fd(t,e)}null!==ed&&fd(ed,e),null!==td&&fd(td,e),null!==nd&&fd(nd,e),rd.forEach(t),ad.forEach(t);for(var n=0;n<od.length;n++){var r=od[n];r.blockedOn===e&&(r.blockedOn=null)}for(;0<od.length&&null===(n=od[0]).blockedOn;)ud(n),null===n.blockedOn&&od.shift();if(null!=(n=(e.ownerDocument||e).$$reactFormReplay))for(r=0;r<n.length;r+=3){var a=n[r],o=n[r+1],i=a[Fe]||null;if("function"==typeof o)i||gd(n);else if(i){var l=null;if(o&&o.hasAttribute("formAction")){if(a=o,i=o[Fe]||null)l=i.formAction;else if(null!==Yp(a))continue}else l=i.action;"function"==typeof l?n[r+1]=l:(n.splice(r,3),r-=3),gd(n)}}}function yd(e){this._internalRoot=e}function bd(e){this._internalRoot=e}bd.prototype.render=yd.prototype.render=function(e){var t=this._internalRoot;if(null===t)throw Error(i(409));Bp(t.current,Tu(),e,t,null,null)},bd.prototype.unmount=yd.prototype.unmount=function(){var e=this._internalRoot;if(null!==e){this._internalRoot=null;var t=e.containerInfo;0===e.tag&&nc(),Bp(e.current,2,null,e,null,null),Fu(),t[De]=null}},bd.prototype.unstable_scheduleHydration=function(e){if(e){var t=je();e={blockedOn:null,target:e,priority:t};for(var n=0;n<od.length&&0!==t&&t<od[n].priority;n++);od.splice(n,0,e),0===n&&ud(e)}};var vd=a.version;if("19.0.0"!==vd)throw Error(i(527,vd,"19.0.0"));U.findDOMNode=function(e){var t=e._reactInternals;if(void 0===t){if("function"==typeof e.render)throw Error(i(188));throw e=Object.keys(e).join(","),Error(i(268,e))}return e=function(e){var t=e.alternate;if(!t){if(null===(t=z(e)))throw Error(i(188));return t!==e?null:e}for(var n=e,r=t;;){var a=n.return;if(null===a)break;var o=a.alternate;if(null===o){if(null!==(r=a.return)){n=r;continue}break}if(a.child===o.child){for(o=a.child;o;){if(o===n)return F(a),e;if(o===r)return F(a),t;o=o.sibling}throw Error(i(188))}if(n.return!==r.return)n=a,r=o;else{for(var l=!1,s=a.child;s;){if(s===n){l=!0,n=a,r=o;break}if(s===r){l=!0,r=a,n=o;break}s=s.sibling}if(!l){for(s=o.child;s;){if(s===n){l=!0,n=o,r=a;break}if(s===r){l=!0,r=o,n=a;break}s=s.sibling}if(!l)throw Error(i(189))}}if(n.alternate!==r)throw Error(i(190))}if(3!==n.tag)throw Error(i(188));return n.stateNode.current===n?e:t}(t),null===(e=null!==e?D(e):null)?null:e.stateNode};var wd={bundleType:0,version:"19.0.0",rendererPackageName:"react-dom",currentDispatcherRef:L,findFiberByHostInstance:Ve,reconcilerVersion:"19.0.0"};if("undefined"!=typeof __REACT_DEVTOOLS_GLOBAL_HOOK__){var xd=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(!xd.isDisabled&&xd.supportsFiber)try{ye=xd.inject(wd),be=xd}catch(e){}}t.createRoot=function(e,t){if(!l(e))throw Error(i(299));var n=!1,r="",a=ji,o=zi,s=$i;return null!=t&&(!0===t.unstable_strictMode&&(n=!0),void 0!==t.identifierPrefix&&(r=t.identifierPrefix),void 0!==t.onUncaughtError&&(a=t.onUncaughtError),void 0!==t.onCaughtError&&(o=t.onCaughtError),void 0!==t.onRecoverableError&&(s=t.onRecoverableError),void 0!==t.unstable_transitionCallbacks&&t.unstable_transitionCallbacks),t=Fp(e,1,!1,null,0,n,r,a,o,s,0,null),e[De]=t.current,Ic(8===e.nodeType?e.parentNode:e),new yd(t)},t.hydrateRoot=function(e,t,n){if(!l(e))throw Error(i(299));var r=!1,a="",o=ji,s=zi,u=$i,c=null;return null!=n&&(!0===n.unstable_strictMode&&(r=!0),void 0!==n.identifierPrefix&&(a=n.identifierPrefix),void 0!==n.onUncaughtError&&(o=n.onUncaughtError),void 0!==n.onCaughtError&&(s=n.onCaughtError),void 0!==n.onRecoverableError&&(u=n.onRecoverableError),void 0!==n.unstable_transitionCallbacks&&n.unstable_transitionCallbacks,void 0!==n.formState&&(c=n.formState)),(t=Fp(e,1,!0,t,0,r,a,o,s,u,0,c)).context=Dp(null),n=t.current,(a=Rl(r=Tu())).callback=null,Il(n,a,r),t.current.lanes=r,Re(t,r),mc(t),e[De]=t.current,Ic(e),new bd(t)},t.version="19.0.0"},1291:(e,t,n)=>{var r=n(741);e.exports=function(e){var t=+e;return t!=t||0===t?0:r(t)}},1517:(e,t,n)=>{var r=n(6518),a=n(6194),o=n(2248).remove;r({target:"Map",proto:!0,real:!0,forced:!0},{deleteAll:function(){for(var e,t=a(this),n=!0,r=0,i=arguments.length;r<i;r++)e=o(t,arguments[r]),n=n&&e;return!!n}})},1625:(e,t,n)=>{var r=n(9504);e.exports=r({}.isPrototypeOf)},1828:(e,t,n)=>{var r=n(9504),a=n(9297),o=n(5397),i=n(9617).indexOf,l=n(421),s=r([].push);e.exports=function(e,t){var n,r=o(e),u=0,c=[];for(n in r)!a(l,n)&&a(r,n)&&s(c,n);for(;t.length>u;)a(r,n=t[u++])&&(~i(c,n)||s(c,n));return c}},2140:(e,t,n)=>{var r={};r[n(8227)("toStringTag")]="z",e.exports="[object z]"===String(r)},2195:(e,t,n)=>{var r=n(9504),a=r({}.toString),o=r("".slice);e.exports=function(e){return o(a(e),8,-1)}},2248:(e,t,n)=>{var r=n(9504),a=Map.prototype;e.exports={Map,set:r(a.set),get:r(a.get),has:r(a.has),remove:r(a.delete),proto:a}},2359:(e,t,n)=>{var r=n(6518),a=n(6080),o=n(6194),i=n(6223);r({target:"Map",proto:!0,real:!0,forced:!0},{find:function(e){var t=o(this),n=a(e,arguments.length>1?arguments[1]:void 0),r=i(t,(function(e,r){if(n(e,r,t))return{value:e}}),!0);return r&&r.value}})},2652:(e,t,n)=>{var r=n(6080),a=n(9565),o=n(8551),i=n(6823),l=n(4209),s=n(6198),u=n(1625),c=n(81),p=n(851),d=n(9539),f=TypeError,h=function(e,t){this.stopped=e,this.result=t},g=h.prototype;e.exports=function(e,t,n){var m,y,b,v,w,x,k,S=n&&n.that,C=!(!n||!n.AS_ENTRIES),_=!(!n||!n.IS_RECORD),E=!(!n||!n.IS_ITERATOR),N=!(!n||!n.INTERRUPTED),P=r(t,S),A=function(e){return m&&d(m,"normal",e),new h(!0,e)},L=function(e){return C?(o(e),N?P(e[0],e[1],A):P(e[0],e[1])):N?P(e,A):P(e)};if(_)m=e.iterator;else if(E)m=e;else{if(!(y=p(e)))throw new f(i(e)+" is not iterable");if(l(y)){for(b=0,v=s(e);v>b;b++)if((w=L(e[b]))&&u(g,w))return w;return new h(!1)}m=c(e,y)}for(x=_?e.next:m.next;!(k=a(x,m)).done;){try{w=L(k.value)}catch(e){d(m,"throw",e)}if("object"==typeof w&&w&&u(g,w))return w}return new h(!1)}},2777:(e,t,n)=>{var r=n(9565),a=n(34),o=n(757),i=n(5966),l=n(4270),s=n(8227),u=TypeError,c=s("toPrimitive");e.exports=function(e,t){if(!a(e)||o(e))return e;var n,s=i(e,c);if(s){if(void 0===t&&(t="default"),n=r(s,e,t),!a(n)||o(n))return n;throw new u("Can't convert object to primitive value")}return void 0===t&&(t="number"),l(e,t)}},2796:(e,t,n)=>{var r=n(9039),a=n(4901),o=/#|\.prototype\./,i=function(e,t){var n=s[l(e)];return n===c||n!==u&&(a(t)?r(t):!!t)},l=i.normalize=function(e){return String(e).replace(o,".").toLowerCase()},s=i.data={},u=i.NATIVE="N",c=i.POLYFILL="P";e.exports=i},2839:(e,t,n)=>{var r=n(4576).navigator,a=r&&r.userAgent;e.exports=a?String(a):""},3317:e=>{e.exports=function(e,t){return e===t||e!=e&&t!=t}},3392:(e,t,n)=>{var r=n(9504),a=0,o=Math.random(),i=r(1..toString);e.exports=function(e){return"Symbol("+(void 0===e?"":e)+")_"+i(++a+o,36)}},3706:(e,t,n)=>{var r=n(9504),a=n(4901),o=n(7629),i=r(Function.toString);a(o.inspectSource)||(o.inspectSource=function(e){return i(e)}),e.exports=o.inspectSource},3717:(e,t)=>{t.f=Object.getOwnPropertySymbols},3724:(e,t,n)=>{var r=n(9039);e.exports=!r((function(){return 7!==Object.defineProperty({},1,{get:function(){return 7}})[1]}))},3777:(e,t,n)=>{var r=n(6518),a=n(6080),o=n(6194),i=n(6223);r({target:"Map",proto:!0,real:!0,forced:!0},{every:function(e){var t=o(this),n=a(e,arguments.length>1?arguments[1]:void 0);return!1!==i(t,(function(e,r){if(!n(e,r,t))return!1}),!0)}})},4055:(e,t,n)=>{var r=n(4576),a=n(34),o=r.document,i=a(o)&&a(o.createElement);e.exports=function(e){return i?o.createElement(e):{}}},4117:e=>{e.exports=function(e){return null==e}},4190:(e,t,n)=>{var r=n(6518),a=n(6080),o=n(6194),i=n(2248),l=n(6223),s=i.Map,u=i.set;r({target:"Map",proto:!0,real:!0,forced:!0},{filter:function(e){var t=o(this),n=a(e,arguments.length>1?arguments[1]:void 0),r=new s;return l(t,(function(e,a){n(e,a,t)&&u(r,a,e)})),r}})},4209:(e,t,n)=>{var r=n(8227),a=n(6269),o=r("iterator"),i=Array.prototype;e.exports=function(e){return void 0!==e&&(a.Array===e||i[o]===e)}},4270:(e,t,n)=>{var r=n(9565),a=n(4901),o=n(34),i=TypeError;e.exports=function(e,t){var n,l;if("string"===t&&a(n=e.toString)&&!o(l=r(n,e)))return l;if(a(n=e.valueOf)&&!o(l=r(n,e)))return l;if("string"!==t&&a(n=e.toString)&&!o(l=r(n,e)))return l;throw new i("Can't convert object to primitive value")}},4477:(e,t)=>{function n(e,t){var n=e.length;e.push(t);e:for(;0<n;){var r=n-1>>>1,a=e[r];if(!(0<o(a,t)))break e;e[r]=t,e[n]=a,n=r}}function r(e){return 0===e.length?null:e[0]}function a(e){if(0===e.length)return null;var t=e[0],n=e.pop();if(n!==t){e[0]=n;e:for(var r=0,a=e.length,i=a>>>1;r<i;){var l=2*(r+1)-1,s=e[l],u=l+1,c=e[u];if(0>o(s,n))u<a&&0>o(c,s)?(e[r]=c,e[u]=n,r=u):(e[r]=s,e[l]=n,r=l);else{if(!(u<a&&0>o(c,n)))break e;e[r]=c,e[u]=n,r=u}}}return t}function o(e,t){var n=e.sortIndex-t.sortIndex;return 0!==n?n:e.id-t.id}if(t.unstable_now=void 0,"object"==typeof performance&&"function"==typeof performance.now){var i=performance;t.unstable_now=function(){return i.now()}}else{var l=Date,s=l.now();t.unstable_now=function(){return l.now()-s}}var u=[],c=[],p=1,d=null,f=3,h=!1,g=!1,m=!1,y="function"==typeof setTimeout?setTimeout:null,b="function"==typeof clearTimeout?clearTimeout:null,v="undefined"!=typeof setImmediate?setImmediate:null;function w(e){for(var t=r(c);null!==t;){if(null===t.callback)a(c);else{if(!(t.startTime<=e))break;a(c),t.sortIndex=t.expirationTime,n(u,t)}t=r(c)}}function x(e){if(m=!1,w(e),!g)if(null!==r(u))g=!0,T();else{var t=r(c);null!==t&&R(x,t.startTime-e)}}var k,S=!1,C=-1,_=5,E=-1;function N(){return!(t.unstable_now()-E<_)}function P(){if(S){var e=t.unstable_now();E=e;var n=!0;try{e:{g=!1,m&&(m=!1,b(C),C=-1),h=!0;var o=f;try{t:{for(w(e),d=r(u);null!==d&&!(d.expirationTime>e&&N());){var i=d.callback;if("function"==typeof i){d.callback=null,f=d.priorityLevel;var l=i(d.expirationTime<=e);if(e=t.unstable_now(),"function"==typeof l){d.callback=l,w(e),n=!0;break t}d===r(u)&&a(u),w(e)}else a(u);d=r(u)}if(null!==d)n=!0;else{var s=r(c);null!==s&&R(x,s.startTime-e),n=!1}}break e}finally{d=null,f=o,h=!1}n=void 0}}finally{n?k():S=!1}}}if("function"==typeof v)k=function(){v(P)};else if("undefined"!=typeof MessageChannel){var A=new MessageChannel,L=A.port2;A.port1.onmessage=P,k=function(){L.postMessage(null)}}else k=function(){y(P,0)};function T(){S||(S=!0,k())}function R(e,n){C=y((function(){e(t.unstable_now())}),n)}t.unstable_IdlePriority=5,t.unstable_ImmediatePriority=1,t.unstable_LowPriority=4,t.unstable_NormalPriority=3,t.unstable_Profiling=null,t.unstable_UserBlockingPriority=2,t.unstable_cancelCallback=function(e){e.callback=null},t.unstable_continueExecution=function(){g||h||(g=!0,T())},t.unstable_forceFrameRate=function(e){0>e||125<e?console.error("forceFrameRate takes a positive int between 0 and 125, forcing frame rates higher than 125 fps is not supported"):_=0<e?Math.floor(1e3/e):5},t.unstable_getCurrentPriorityLevel=function(){return f},t.unstable_getFirstCallbackNode=function(){return r(u)},t.unstable_next=function(e){switch(f){case 1:case 2:case 3:var t=3;break;default:t=f}var n=f;f=t;try{return e()}finally{f=n}},t.unstable_pauseExecution=function(){},t.unstable_requestPaint=function(){},t.unstable_runWithPriority=function(e,t){switch(e){case 1:case 2:case 3:case 4:case 5:break;default:e=3}var n=f;f=e;try{return t()}finally{f=n}},t.unstable_scheduleCallback=function(e,a,o){var i=t.unstable_now();switch(o="object"==typeof o&&null!==o&&"number"==typeof(o=o.delay)&&0<o?i+o:i,e){case 1:var l=-1;break;case 2:l=250;break;case 5:l=1073741823;break;case 4:l=1e4;break;default:l=5e3}return e={id:p++,callback:a,priorityLevel:e,startTime:o,expirationTime:l=o+l,sortIndex:-1},o>i?(e.sortIndex=o,n(c,e),null===r(u)&&e===r(c)&&(m?(b(C),C=-1):m=!0,R(x,o-i))):(e.sortIndex=l,n(u,e),g||h||(g=!0,T())),e},t.unstable_shouldYield=N,t.unstable_wrapCallback=function(e){var t=f;return function(){var n=f;f=t;try{return e.apply(this,arguments)}finally{f=n}}}},4495:(e,t,n)=>{var r=n(9519),a=n(9039),o=n(4576).String;e.exports=!!Object.getOwnPropertySymbols&&!a((function(){var e=Symbol("symbol detection");return!o(e)||!(Object(e)instanceof Symbol)||!Symbol.sham&&r&&r<41}))},4576:function(e,t,n){var r=function(e){return e&&e.Math===Math&&e};e.exports=r("object"==typeof globalThis&&globalThis)||r("object"==typeof window&&window)||r("object"==typeof self&&self)||r("object"==typeof n.g&&n.g)||r("object"==typeof this&&this)||function(){return this}()||Function("return this")()},4848:(e,t,n)=>{e.exports=n(9698)},4901:e=>{var t="object"==typeof document&&document.all;e.exports=void 0===t&&void 0!==t?function(e){return"function"==typeof e||e===t}:function(e){return"function"==typeof e}},4913:(e,t,n)=>{var r=n(3724),a=n(5917),o=n(8686),i=n(8551),l=n(6969),s=TypeError,u=Object.defineProperty,c=Object.getOwnPropertyDescriptor,p="enumerable",d="configurable",f="writable";t.f=r?o?function(e,t,n){if(i(e),t=l(t),i(n),"function"==typeof e&&"prototype"===t&&"value"in n&&f in n&&!n[f]){var r=c(e,t);r&&r[f]&&(e[t]=n.value,n={configurable:d in n?n[d]:r[d],enumerable:p in n?n[p]:r[p],writable:!1})}return u(e,t,n)}:u:function(e,t,n){if(i(e),t=l(t),i(n),a)try{return u(e,t,n)}catch(e){}if("get"in n||"set"in n)throw new s("Accessors not supported");return"value"in n&&(e[t]=n.value),e}},5031:(e,t,n)=>{var r=n(7751),a=n(9504),o=n(8480),i=n(3717),l=n(8551),s=a([].concat);e.exports=r("Reflect","ownKeys")||function(e){var t=o.f(l(e)),n=i.f;return n?s(t,n(e)):t}},5122:(e,t,n)=>{var r=n(6518),a=n(6194),o=n(2652),i=n(2248).set;r({target:"Map",proto:!0,real:!0,arity:1,forced:!0},{merge:function(e){for(var t=a(this),n=arguments.length,r=0;r<n;)o(arguments[r++],(function(e,n){i(t,e,n)}),{AS_ENTRIES:!0});return t}})},5338:(e,t,n)=>{!function e(){if("undefined"!=typeof __REACT_DEVTOOLS_GLOBAL_HOOK__&&"function"==typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE)try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(e){console.error(e)}}(),e.exports=n(1247)},5397:(e,t,n)=>{var r=n(7055),a=n(7750);e.exports=function(e){return r(a(e))}},5610:(e,t,n)=>{var r=n(1291),a=Math.max,o=Math.min;e.exports=function(e,t){var n=r(e);return n<0?a(n+t,0):o(n,t)}},5745:(e,t,n)=>{var r=n(7629);e.exports=function(e,t){return r[e]||(r[e]=t||{})}},5917:(e,t,n)=>{var r=n(3724),a=n(9039),o=n(4055);e.exports=!r&&!a((function(){return 7!==Object.defineProperty(o("div"),"a",{get:function(){return 7}}).a}))},5966:(e,t,n)=>{var r=n(9306),a=n(4117);e.exports=function(e,t){var n=e[t];return a(n)?void 0:r(n)}},6080:(e,t,n)=>{var r=n(7476),a=n(9306),o=n(616),i=r(r.bind);e.exports=function(e,t){return a(e),void 0===t?e:o?i(e,t):function(){return e.apply(t,arguments)}}},6097:(e,t,n)=>{var r=n(6518),a=n(6080),o=n(6194),i=n(6223);r({target:"Map",proto:!0,real:!0,forced:!0},{findKey:function(e){var t=o(this),n=a(e,arguments.length>1?arguments[1]:void 0),r=i(t,(function(e,r){if(n(e,r,t))return{key:r}}),!0);return r&&r.key}})},6119:(e,t,n)=>{var r=n(5745),a=n(3392),o=r("keys");e.exports=function(e){return o[e]||(o[e]=a(e))}},6194:(e,t,n)=>{var r=n(2248).has;e.exports=function(e){return r(e),e}},6198:(e,t,n)=>{var r=n(8014);e.exports=function(e){return r(e.length)}},6221:(e,t,n)=>{var r=n(6540);function a(e){var t="https://react.dev/errors/"+e;if(1<arguments.length){t+="?args[]="+encodeURIComponent(arguments[1]);for(var n=2;n<arguments.length;n++)t+="&args[]="+encodeURIComponent(arguments[n])}return"Minified React error #"+e+"; visit "+t+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function o(){}var i={d:{f:o,r:function(){throw Error(a(522))},D:o,C:o,L:o,m:o,X:o,S:o,M:o},p:0,findDOMNode:null},l=Symbol.for("react.portal"),s=r.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE;function u(e,t){return"font"===e?"":"string"==typeof t?"use-credentials"===t?t:"":void 0}t.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=i,t.createPortal=function(e,t){var n=2<arguments.length&&void 0!==arguments[2]?arguments[2]:null;if(!t||1!==t.nodeType&&9!==t.nodeType&&11!==t.nodeType)throw Error(a(299));return function(e,t,n){var r=3<arguments.length&&void 0!==arguments[3]?arguments[3]:null;return{$$typeof:l,key:null==r?null:""+r,children:e,containerInfo:t,implementation:n}}(e,t,null,n)},t.flushSync=function(e){var t=s.T,n=i.p;try{if(s.T=null,i.p=2,e)return e()}finally{s.T=t,i.p=n,i.d.f()}},t.preconnect=function(e,t){"string"==typeof e&&(t=t?"string"==typeof(t=t.crossOrigin)?"use-credentials"===t?t:"":void 0:null,i.d.C(e,t))},t.prefetchDNS=function(e){"string"==typeof e&&i.d.D(e)},t.preinit=function(e,t){if("string"==typeof e&&t&&"string"==typeof t.as){var n=t.as,r=u(n,t.crossOrigin),a="string"==typeof t.integrity?t.integrity:void 0,o="string"==typeof t.fetchPriority?t.fetchPriority:void 0;"style"===n?i.d.S(e,"string"==typeof t.precedence?t.precedence:void 0,{crossOrigin:r,integrity:a,fetchPriority:o}):"script"===n&&i.d.X(e,{crossOrigin:r,integrity:a,fetchPriority:o,nonce:"string"==typeof t.nonce?t.nonce:void 0})}},t.preinitModule=function(e,t){if("string"==typeof e)if("object"==typeof t&&null!==t){if(null==t.as||"script"===t.as){var n=u(t.as,t.crossOrigin);i.d.M(e,{crossOrigin:n,integrity:"string"==typeof t.integrity?t.integrity:void 0,nonce:"string"==typeof t.nonce?t.nonce:void 0})}}else null==t&&i.d.M(e)},t.preload=function(e,t){if("string"==typeof e&&"object"==typeof t&&null!==t&&"string"==typeof t.as){var n=t.as,r=u(n,t.crossOrigin);i.d.L(e,n,{crossOrigin:r,integrity:"string"==typeof t.integrity?t.integrity:void 0,nonce:"string"==typeof t.nonce?t.nonce:void 0,type:"string"==typeof t.type?t.type:void 0,fetchPriority:"string"==typeof t.fetchPriority?t.fetchPriority:void 0,referrerPolicy:"string"==typeof t.referrerPolicy?t.referrerPolicy:void 0,imageSrcSet:"string"==typeof t.imageSrcSet?t.imageSrcSet:void 0,imageSizes:"string"==typeof t.imageSizes?t.imageSizes:void 0,media:"string"==typeof t.media?t.media:void 0})}},t.preloadModule=function(e,t){if("string"==typeof e)if(t){var n=u(t.as,t.crossOrigin);i.d.m(e,{as:"string"==typeof t.as&&"script"!==t.as?t.as:void 0,crossOrigin:n,integrity:"string"==typeof t.integrity?t.integrity:void 0})}else i.d.m(e)},t.requestFormReset=function(e){i.d.r(e)},t.unstable_batchedUpdates=function(e,t){return e(t)},t.useFormState=function(e,t,n){return s.H.useFormState(e,t,n)},t.useFormStatus=function(){return s.H.useHostTransitionStatus()},t.version="19.0.0"},6223:(e,t,n)=>{var r=n(9504),a=n(507),o=n(2248),i=o.Map,l=o.proto,s=r(l.forEach),u=r(l.entries),c=u(new i).next;e.exports=function(e,t,n){return n?a({iterator:u(e),next:c},(function(e){return t(e[1],e[0])})):s(e,t)}},6269:e=>{e.exports={}},6395:e=>{e.exports=!1},6518:(e,t,n)=>{var r=n(4576),a=n(7347).f,o=n(6699),i=n(6840),l=n(9433),s=n(7740),u=n(2796);e.exports=function(e,t){var n,c,p,d,f,h=e.target,g=e.global,m=e.stat;if(n=g?r:m?r[h]||l(h,{}):r[h]&&r[h].prototype)for(c in t){if(d=t[c],p=e.dontCallGetSet?(f=a(n,c))&&f.value:n[c],!u(g?c:h+(m?".":"#")+c,e.forced)&&void 0!==p){if(typeof d==typeof p)continue;s(d,p)}(e.sham||p&&p.sham)&&o(d,"sham",!0),i(n,c,d,e)}}},6540:(e,t,n)=>{e.exports=n(9869)},6699:(e,t,n)=>{var r=n(3724),a=n(4913),o=n(6980);e.exports=r?function(e,t,n){return a.f(e,t,o(1,n))}:function(e,t,n){return e[t]=n,e}},6823:e=>{var t=String;e.exports=function(e){try{return t(e)}catch(e){return"Object"}}},6840:(e,t,n)=>{var r=n(4901),a=n(4913),o=n(283),i=n(9433);e.exports=function(e,t,n,l){l||(l={});var s=l.enumerable,u=void 0!==l.name?l.name:t;if(r(n)&&o(n,u,l),l.global)s?e[t]=n:i(t,n);else{try{l.unsafe?e[t]&&(s=!0):delete e[t]}catch(e){}s?e[t]=n:a.f(e,t,{value:n,enumerable:!1,configurable:!l.nonConfigurable,writable:!l.nonWritable})}return e}},6955:(e,t,n)=>{var r=n(2140),a=n(4901),o=n(2195),i=n(8227)("toStringTag"),l=Object,s="Arguments"===o(function(){return arguments}());e.exports=r?o:function(e){var t,n,r;return void 0===e?"Undefined":null===e?"Null":"string"==typeof(n=function(e,t){try{return e[t]}catch(e){}}(t=l(e),i))?n:s?o(t):"Object"===(r=o(t))&&a(t.callee)?"Arguments":r}},6969:(e,t,n)=>{var r=n(2777),a=n(757);e.exports=function(e){var t=r(e,"string");return a(t)?t:t+""}},6980:e=>{e.exports=function(e,t){return{enumerable:!(1&e),configurable:!(2&e),writable:!(4&e),value:t}}},7040:(e,t,n)=>{var r=n(4495);e.exports=r&&!Symbol.sham&&"symbol"==typeof Symbol.iterator},7055:(e,t,n)=>{var r=n(9504),a=n(9039),o=n(2195),i=Object,l=r("".split);e.exports=a((function(){return!i("z").propertyIsEnumerable(0)}))?function(e){return"String"===o(e)?l(e,""):i(e)}:i},7268:(e,t,n)=>{var r=n(6518),a=n(6080),o=n(6194),i=n(6223);r({target:"Map",proto:!0,real:!0,forced:!0},{some:function(e){var t=o(this),n=a(e,arguments.length>1?arguments[1]:void 0);return!0===i(t,(function(e,r){if(n(e,r,t))return!0}),!0)}})},7273:(e,t,n)=>{var r=n(6518),a=n(3317),o=n(6194),i=n(6223);r({target:"Map",proto:!0,real:!0,forced:!0},{includes:function(e){return!0===i(o(this),(function(t){if(a(t,e))return!0}),!0)}})},7347:(e,t,n)=>{var r=n(3724),a=n(9565),o=n(8773),i=n(6980),l=n(5397),s=n(6969),u=n(9297),c=n(5917),p=Object.getOwnPropertyDescriptor;t.f=r?p:function(e,t){if(e=l(e),t=s(t),c)try{return p(e,t)}catch(e){}if(u(e,t))return i(!a(o.f,e,t),e[t])}},7415:(e,t,n)=>{var r=n(6518),a=n(6194),o=n(6223);r({target:"Map",proto:!0,real:!0,forced:!0},{keyOf:function(e){var t=o(a(this),(function(t,n){if(t===e)return{key:n}}),!0);return t&&t.key}})},7476:(e,t,n)=>{var r=n(2195),a=n(9504);e.exports=function(e){if("Function"===r(e))return a(e)}},7583:(e,t,n)=>{var r=n(6518),a=n(6080),o=n(6194),i=n(2248),l=n(6223),s=i.Map,u=i.set;r({target:"Map",proto:!0,real:!0,forced:!0},{mapValues:function(e){var t=o(this),n=a(e,arguments.length>1?arguments[1]:void 0),r=new s;return l(t,(function(e,a){u(r,a,n(e,a,t))})),r}})},7629:(e,t,n)=>{var r=n(6395),a=n(4576),o=n(9433),i="__core-js_shared__",l=e.exports=a[i]||o(i,{});(l.versions||(l.versions=[])).push({version:"3.40.0",mode:r?"pure":"global",copyright:"© 2014-2025 Denis Pushkarev (zloirock.ru)",license:"https://github.com/zloirock/core-js/blob/v3.40.0/LICENSE",source:"https://github.com/zloirock/core-js"})},7740:(e,t,n)=>{var r=n(9297),a=n(5031),o=n(7347),i=n(4913);e.exports=function(e,t,n){for(var l=a(t),s=i.f,u=o.f,c=0;c<l.length;c++){var p=l[c];r(e,p)||n&&r(n,p)||s(e,p,u(t,p))}}},7750:(e,t,n)=>{var r=n(4117),a=TypeError;e.exports=function(e){if(r(e))throw new a("Can't call method on "+e);return e}},7751:(e,t,n)=>{var r=n(4576),a=n(4901);e.exports=function(e,t){return arguments.length<2?(n=r[e],a(n)?n:void 0):r[e]&&r[e][t];var n}},8014:(e,t,n)=>{var r=n(1291),a=Math.min;e.exports=function(e){var t=r(e);return t>0?a(t,9007199254740991):0}},8227:(e,t,n)=>{var r=n(4576),a=n(5745),o=n(9297),i=n(3392),l=n(4495),s=n(7040),u=r.Symbol,c=a("wks"),p=s?u.for||u:u&&u.withoutSetter||i;e.exports=function(e){return o(c,e)||(c[e]=l&&o(u,e)?u[e]:p("Symbol."+e)),c[e]}},8480:(e,t,n)=>{var r=n(1828),a=n(8727).concat("length","prototype");t.f=Object.getOwnPropertyNames||function(e){return r(e,a)}},8551:(e,t,n)=>{var r=n(34),a=String,o=TypeError;e.exports=function(e){if(r(e))return e;throw new o(a(e)+" is not an object")}},8622:(e,t,n)=>{var r=n(4576),a=n(4901),o=r.WeakMap;e.exports=a(o)&&/native code/.test(String(o))},8686:(e,t,n)=>{var r=n(3724),a=n(9039);e.exports=r&&a((function(){return 42!==Object.defineProperty((function(){}),"prototype",{value:42,writable:!1}).prototype}))},8727:e=>{e.exports=["constructor","hasOwnProperty","isPrototypeOf","propertyIsEnumerable","toLocaleString","toString","valueOf"]},8773:(e,t)=>{var n={}.propertyIsEnumerable,r=Object.getOwnPropertyDescriptor,a=r&&!n.call({1:2},1);t.f=a?function(e){var t=r(this,e);return!!t&&t.enumerable}:n},8981:(e,t,n)=>{var r=n(7750),a=Object;e.exports=function(e){return a(r(e))}},9039:e=>{e.exports=function(e){try{return!!e()}catch(e){return!0}}},9297:(e,t,n)=>{var r=n(9504),a=n(8981),o=r({}.hasOwnProperty);e.exports=Object.hasOwn||function(e,t){return o(a(e),t)}},9306:(e,t,n)=>{var r=n(4901),a=n(6823),o=TypeError;e.exports=function(e){if(r(e))return e;throw new o(a(e)+" is not a function")}},9433:(e,t,n)=>{var r=n(4576),a=Object.defineProperty;e.exports=function(e,t){try{a(r,e,{value:t,configurable:!0,writable:!0})}catch(n){r[e]=t}return t}},9504:(e,t,n)=>{var r=n(616),a=Function.prototype,o=a.call,i=r&&a.bind.bind(o,o);e.exports=r?i:function(e){return function(){return o.apply(e,arguments)}}},9519:(e,t,n)=>{var r,a,o=n(4576),i=n(2839),l=o.process,s=o.Deno,u=l&&l.versions||s&&s.version,c=u&&u.v8;c&&(a=(r=c.split("."))[0]>0&&r[0]<4?1:+(r[0]+r[1])),!a&&i&&(!(r=i.match(/Edge\/(\d+)/))||r[1]>=74)&&(r=i.match(/Chrome\/(\d+)/))&&(a=+r[1]),e.exports=a},9539:(e,t,n)=>{var r=n(9565),a=n(8551),o=n(5966);e.exports=function(e,t,n){var i,l;a(e);try{if(!(i=o(e,"return"))){if("throw"===t)throw n;return n}i=r(i,e)}catch(e){l=!0,i=e}if("throw"===t)throw n;if(l)throw i;return a(i),n}},9565:(e,t,n)=>{var r=n(616),a=Function.prototype.call;e.exports=r?a.bind(a):function(){return a.apply(a,arguments)}},9617:(e,t,n)=>{var r=n(5397),a=n(5610),o=n(6198),i=function(e){return function(t,n,i){var l=r(t),s=o(l);if(0===s)return!e&&-1;var u,c=a(i,s);if(e&&n!=n){for(;s>c;)if((u=l[c++])!=u)return!0}else for(;s>c;c++)if((e||c in l)&&l[c]===n)return e||c||0;return!e&&-1}};e.exports={includes:i(!0),indexOf:i(!1)}},9698:(e,t)=>{var n=Symbol.for("react.transitional.element"),r=Symbol.for("react.fragment");function a(e,t,r){var a=null;if(void 0!==r&&(a=""+r),void 0!==t.key&&(a=""+t.key),"key"in t)for(var o in r={},t)"key"!==o&&(r[o]=t[o]);else r=t;return t=r.ref,{$$typeof:n,type:e,key:a,ref:void 0!==t?t:null,props:r}}t.Fragment=r,t.jsx=a,t.jsxs=a},9733:(e,t,n)=>{var r=n(6518),a=n(9306),o=n(6194),i=n(2248),l=TypeError,s=i.get,u=i.has,c=i.set;r({target:"Map",proto:!0,real:!0,forced:!0},{update:function(e,t){var n=o(this),r=arguments.length;a(t);var i=u(n,e);if(!i&&r<3)throw new l("Updating absent value");var p=i?s(n,e):a(r>2?arguments[2]:void 0)(e,n);return c(n,e,t(p,e,n)),n}})},9869:(e,t)=>{var n=Symbol.for("react.transitional.element"),r=Symbol.for("react.portal"),a=Symbol.for("react.fragment"),o=Symbol.for("react.strict_mode"),i=Symbol.for("react.profiler"),l=Symbol.for("react.consumer"),s=Symbol.for("react.context"),u=Symbol.for("react.forward_ref"),c=Symbol.for("react.suspense"),p=Symbol.for("react.memo"),d=Symbol.for("react.lazy"),f=Symbol.iterator,h={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},g=Object.assign,m={};function y(e,t,n){this.props=e,this.context=t,this.refs=m,this.updater=n||h}function b(){}function v(e,t,n){this.props=e,this.context=t,this.refs=m,this.updater=n||h}y.prototype.isReactComponent={},y.prototype.setState=function(e,t){if("object"!=typeof e&&"function"!=typeof e&&null!=e)throw Error("takes an object of state variables to update or a function which returns an object of state variables.");this.updater.enqueueSetState(this,e,t,"setState")},y.prototype.forceUpdate=function(e){this.updater.enqueueForceUpdate(this,e,"forceUpdate")},b.prototype=y.prototype;var w=v.prototype=new b;w.constructor=v,g(w,y.prototype),w.isPureReactComponent=!0;var x=Array.isArray,k={H:null,A:null,T:null,S:null},S=Object.prototype.hasOwnProperty;function C(e,t,r,a,o,i){return r=i.ref,{$$typeof:n,type:e,key:t,ref:void 0!==r?r:null,props:i}}function _(e){return"object"==typeof e&&null!==e&&e.$$typeof===n}var E=/\/+/g;function N(e,t){return"object"==typeof e&&null!==e&&null!=e.key?function(e){var t={"=":"=0",":":"=2"};return"$"+e.replace(/[=:]/g,(function(e){return t[e]}))}(""+e.key):t.toString(36)}function P(){}function A(e,t,a,o,i){var l=typeof e;"undefined"!==l&&"boolean"!==l||(e=null);var s,u,c=!1;if(null===e)c=!0;else switch(l){case"bigint":case"string":case"number":c=!0;break;case"object":switch(e.$$typeof){case n:case r:c=!0;break;case d:return A((c=e._init)(e._payload),t,a,o,i)}}if(c)return i=i(e),c=""===o?"."+N(e,0):o,x(i)?(a="",null!=c&&(a=c.replace(E,"$&/")+"/"),A(i,t,a,"",(function(e){return e}))):null!=i&&(_(i)&&(s=i,u=a+(null==i.key||e&&e.key===i.key?"":(""+i.key).replace(E,"$&/")+"/")+c,i=C(s.type,u,void 0,0,0,s.props)),t.push(i)),1;c=0;var p,h=""===o?".":o+":";if(x(e))for(var g=0;g<e.length;g++)c+=A(o=e[g],t,a,l=h+N(o,g),i);else if("function"==typeof(g=null===(p=e)||"object"!=typeof p?null:"function"==typeof(p=f&&p[f]||p["@@iterator"])?p:null))for(e=g.call(e),g=0;!(o=e.next()).done;)c+=A(o=o.value,t,a,l=h+N(o,g++),i);else if("object"===l){if("function"==typeof e.then)return A(function(e){switch(e.status){case"fulfilled":return e.value;case"rejected":throw e.reason;default:switch("string"==typeof e.status?e.then(P,P):(e.status="pending",e.then((function(t){"pending"===e.status&&(e.status="fulfilled",e.value=t)}),(function(t){"pending"===e.status&&(e.status="rejected",e.reason=t)}))),e.status){case"fulfilled":return e.value;case"rejected":throw e.reason}}throw e}(e),t,a,o,i);throw t=String(e),Error("Objects are not valid as a React child (found: "+("[object Object]"===t?"object with keys {"+Object.keys(e).join(", ")+"}":t)+"). If you meant to render a collection of children, use an array instead.")}return c}function L(e,t,n){if(null==e)return e;var r=[],a=0;return A(e,r,"","",(function(e){return t.call(n,e,a++)})),r}function T(e){if(-1===e._status){var t=e._result;(t=t()).then((function(t){0!==e._status&&-1!==e._status||(e._status=1,e._result=t)}),(function(t){0!==e._status&&-1!==e._status||(e._status=2,e._result=t)})),-1===e._status&&(e._status=0,e._result=t)}if(1===e._status)return e._result.default;throw e._result}var R="function"==typeof reportError?reportError:function(e){if("object"==typeof window&&"function"==typeof window.ErrorEvent){var t=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:"object"==typeof e&&null!==e&&"string"==typeof e.message?String(e.message):String(e),error:e});if(!window.dispatchEvent(t))return}else if("object"==typeof process&&"function"==typeof process.emit)return void process.emit("uncaughtException",e);console.error(e)};function I(){}t.Children={map:L,forEach:function(e,t,n){L(e,(function(){t.apply(this,arguments)}),n)},count:function(e){var t=0;return L(e,(function(){t++})),t},toArray:function(e){return L(e,(function(e){return e}))||[]},only:function(e){if(!_(e))throw Error("React.Children.only expected to receive a single React element child.");return e}},t.Component=y,t.Fragment=a,t.Profiler=i,t.PureComponent=v,t.StrictMode=o,t.Suspense=c,t.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=k,t.act=function(){throw Error("act(...) is not supported in production builds of React.")},t.cache=function(e){return function(){return e.apply(null,arguments)}},t.cloneElement=function(e,t,n){if(null==e)throw Error("The argument must be a React element, but you passed "+e+".");var r=g({},e.props),a=e.key;if(null!=t)for(o in t.ref,void 0!==t.key&&(a=""+t.key),t)!S.call(t,o)||"key"===o||"__self"===o||"__source"===o||"ref"===o&&void 0===t.ref||(r[o]=t[o]);var o=arguments.length-2;if(1===o)r.children=n;else if(1<o){for(var i=Array(o),l=0;l<o;l++)i[l]=arguments[l+2];r.children=i}return C(e.type,a,void 0,0,0,r)},t.createContext=function(e){return(e={$$typeof:s,_currentValue:e,_currentValue2:e,_threadCount:0,Provider:null,Consumer:null}).Provider=e,e.Consumer={$$typeof:l,_context:e},e},t.createElement=function(e,t,n){var r,a={},o=null;if(null!=t)for(r in void 0!==t.key&&(o=""+t.key),t)S.call(t,r)&&"key"!==r&&"__self"!==r&&"__source"!==r&&(a[r]=t[r]);var i=arguments.length-2;if(1===i)a.children=n;else if(1<i){for(var l=Array(i),s=0;s<i;s++)l[s]=arguments[s+2];a.children=l}if(e&&e.defaultProps)for(r in i=e.defaultProps)void 0===a[r]&&(a[r]=i[r]);return C(e,o,void 0,0,0,a)},t.createRef=function(){return{current:null}},t.forwardRef=function(e){return{$$typeof:u,render:e}},t.isValidElement=_,t.lazy=function(e){return{$$typeof:d,_payload:{_status:-1,_result:e},_init:T}},t.memo=function(e,t){return{$$typeof:p,type:e,compare:void 0===t?null:t}},t.startTransition=function(e){var t=k.T,n={};k.T=n;try{var r=e(),a=k.S;null!==a&&a(n,r),"object"==typeof r&&null!==r&&"function"==typeof r.then&&r.then(I,R)}catch(e){R(e)}finally{k.T=t}},t.unstable_useCacheRefresh=function(){return k.H.useCacheRefresh()},t.use=function(e){return k.H.use(e)},t.useActionState=function(e,t,n){return k.H.useActionState(e,t,n)},t.useCallback=function(e,t){return k.H.useCallback(e,t)},t.useContext=function(e){return k.H.useContext(e)},t.useDebugValue=function(){},t.useDeferredValue=function(e,t){return k.H.useDeferredValue(e,t)},t.useEffect=function(e,t){return k.H.useEffect(e,t)},t.useId=function(){return k.H.useId()},t.useImperativeHandle=function(e,t,n){return k.H.useImperativeHandle(e,t,n)},t.useInsertionEffect=function(e,t){return k.H.useInsertionEffect(e,t)},t.useLayoutEffect=function(e,t){return k.H.useLayoutEffect(e,t)},t.useMemo=function(e,t){return k.H.useMemo(e,t)},t.useOptimistic=function(e,t){return k.H.useOptimistic(e,t)},t.useReducer=function(e,t,n){return k.H.useReducer(e,t,n)},t.useRef=function(e){return k.H.useRef(e)},t.useState=function(e){return k.H.useState(e)},t.useSyncExternalStore=function(e,t,n){return k.H.useSyncExternalStore(e,t,n)},t.useTransition=function(){return k.H.useTransition()},t.version="19.0.0"},9929:(e,t,n)=>{var r=n(6518),a=n(6080),o=n(6194),i=n(2248),l=n(6223),s=i.Map,u=i.set;r({target:"Map",proto:!0,real:!0,forced:!0},{mapKeys:function(e){var t=o(this),n=a(e,arguments.length>1?arguments[1]:void 0),r=new s;return l(t,(function(e,a){u(r,n(e,a,t),e)})),r}})},9982:(e,t,n)=>{e.exports=n(4477)}},t={};function n(r){var a=t[r];if(void 0!==a)return a.exports;var o=t[r]={exports:{}};return e[r].call(o.exports,o,o.exports,n),o.exports}n.d=(e,t)=>{for(var r in t)n.o(t,r)&&!n.o(e,r)&&Object.defineProperty(e,r,{enumerable:!0,get:t[r]})},n.g=function(){if("object"==typeof globalThis)return globalThis;try{return this||new Function("return this")()}catch(e){if("object"==typeof window)return window}}(),n.o=(e,t)=>Object.prototype.hasOwnProperty.call(e,t),n.r=e=>{"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(e,"__esModule",{value:!0})};var r={};n.r(r),n.d(r,{boolean:()=>ft,booleanish:()=>ht,commaOrSpaceSeparated:()=>vt,commaSeparated:()=>bt,number:()=>mt,overloadedBoolean:()=>gt,spaceSeparated:()=>yt});var a=n(5338);class o{constructor(){this.error="||@ error.error @||",this.description="||@ error.description @||",this.traceback=JSON.parse("||@ error.traceback|safe_json @||")}}n(1517),n(3777),n(4190),n(2359),n(6097),n(7273),n(7415),n(9929),n(7583),n(5122),n(230),n(7268),n(9733);class i{constructor(){this.path="||@ request.path @||",this.method="||@ request.method @||",this.clientHost="||@ request.client.host @||",this.clientPort=parseInt("||@ request.client.port @||"),this.pathParams=new Map(Object.entries(JSON.parse("||@ request.params.path|safe_json @||"))),this.queryParams=new Map(Object.entries(JSON.parse("||@ request.params.query|safe_json @||"))),this.headers=new Map(Object.entries(JSON.parse("||@ request.headers|safe_json @||"))),this.cookies=new Map(Object.entries(JSON.parse("||@ request.cookies|safe_json @||")))}}var l=n(6540),s={outline:{xmlns:"http://www.w3.org/2000/svg",width:24,height:24,viewBox:"0 0 24 24",fill:"none",stroke:"currentColor",strokeWidth:2,strokeLinecap:"round",strokeLinejoin:"round"},filled:{xmlns:"http://www.w3.org/2000/svg",width:24,height:24,viewBox:"0 0 24 24",fill:"currentColor",stroke:"none"}};const u=(e,t,n,r)=>{const a=(0,l.forwardRef)((({color:n="currentColor",size:a=24,stroke:o=2,title:i,className:u,children:c,...p},d)=>(0,l.createElement)("svg",{ref:d,...s[e],width:a,height:a,className:["tabler-icon",`tabler-icon-${t}`,u].join(" "),..."filled"===e?{fill:n}:{strokeWidth:o,stroke:n},...p},[i&&(0,l.createElement)("title",{key:"svg-title"},i),...r.map((([e,t])=>(0,l.createElement)(e,t))),...Array.isArray(c)?c:[c]])));return a.displayName=`${n}`,a};var c=u("outline","book","IconBook",[["path",{d:"M3 19a9 9 0 0 1 9 0a9 9 0 0 1 9 0",key:"svg-0"}],["path",{d:"M3 6a9 9 0 0 1 9 0a9 9 0 0 1 9 0",key:"svg-1"}],["path",{d:"M3 6l0 13",key:"svg-2"}],["path",{d:"M12 6l0 13",key:"svg-3"}],["path",{d:"M21 6l0 13",key:"svg-4"}]]),p=u("outline","script","IconScript",[["path",{d:"M17 20h-11a3 3 0 0 1 0 -6h11a3 3 0 0 0 0 6h1a3 3 0 0 0 3 -3v-11a2 2 0 0 0 -2 -2h-10a2 2 0 0 0 -2 2v8",key:"svg-0"}]]),d=u("outline","brand-github","IconBrandGithub",[["path",{d:"M9 19c-4.3 1.4 -4.3 -2.5 -6 -3m12 5v-3.5c0 -1 .1 -1.4 -.5 -2c2.8 -.3 5.5 -1.4 5.5 -6a4.6 4.6 0 0 0 -1.3 -3.2a4.2 4.2 0 0 0 -.1 -3.2s-1.1 -.3 -3.5 1.3a12.3 12.3 0 0 0 -6.2 0c-2.4 -1.6 -3.5 -1.3 -3.5 -1.3a4.2 4.2 0 0 0 -.1 3.2a4.6 4.6 0 0 0 -1.3 3.2c0 4.6 2.7 5.7 5.5 6c-.6 .6 -.6 1.2 -.5 2v3.5",key:"svg-0"}]]),f=u("outline","x","IconX",[["path",{d:"M18 6l-12 12",key:"svg-0"}],["path",{d:"M6 6l12 12",key:"svg-1"}]]),h=u("outline","menu-2","IconMenu2",[["path",{d:"M4 6l16 0",key:"svg-0"}],["path",{d:"M4 12l16 0",key:"svg-1"}],["path",{d:"M4 18l16 0",key:"svg-2"}]]),g=n(961),m=n(4848);function y(e){let{...t}=e;return(0,m.jsx)("svg",{viewBox:"0 0 1024 1024",fill:"currentColor",xmlns:"http://www.w3.org/2000/svg",...t,children:(0,m.jsx)("path",{d:"M734.61,441.99c-7.05-15.75-9.96-32.93-8.26-50.83,1.72-18.08,13.06-35.59,26.3-52.84,.66-1.73,8.25-11.78-7.07-5.84-3.79,.74-16.01,5.48-19.74,6.61-32.26,9.73-43.32,29.76-52.49,66.59-3.39,13.6-3.19,28.47-4.11,42.8-1.09,17.02-1.47,34.1-2.49,51.13-.4,6.65-.96,13.51-2.72,19.8-9.42,33.62-43.08,55.79-72.92,48.45-24.22-5.95-39.46-26.85-40.25-55.39-.36-13.17,1.9-25.66,8.56-36.6,4.8-7.89,9.78-15.64,14.95-23.2,14.1-20.64,26-42.69,33.76-67.61,4.88-15.67,5.79-31.99,6.1-48.32,.25-12.84,.78-25.97-1.13-38.5-5.01-32.85-16.97-62.77-33.87-89.52-19.85-31.43-45.02-56.86-72.02-79.83-40.98-32.52-5.03,50.15-4.46,81.96,.35,19.63,.31,39.17-3.85,58.36-6.6,30.4-20.91,55.67-39.07,78.36-20.7,25.85-44.14,48.14-67.7,70.28-27.96,26.28-56.18,52.2-81.3,82.27-27.42,32.82-48.89,69.9-61.12,113.69-5.24,18.77-10.29,37.67-11.06,57.49-.77,19.99-.76,40.03-.77,60.04,0,7.1,.32,14.35,1.55,21.29,10.52,59.38,34.5,109.72,75.2,148.62,29.12,27.83,52.9,41.17,85.72,53.58-10.19-10.79-16.94-20.99-26.23-35.86-35.68-57.1-48.33-121.18-40.92-190.73,3.28-30.8,11.37-60.07,22.68-88.23,14.75-36.72,19.81-50.3,45.56-81.01,21.79-31.69,23.03-10.04,22.49-6.18-3.86,27.43-6.75,54.98-4.67,82.8,2.13,28.45,12.22,52.83,29.27,72.9,21.31,25.08,47.84,30.81,76.54,23.2,19.23-5.1,38.46-8.62,58.18-7.81,8.4,.35,16.84,.07,24.9,4.1,27.05,13.52,45.64,36.41,55.25,68.43,8.76,29.2,7.75,59.16,1.55,88.8-4.91,23.47-19.9,51.68-36.69,69.87,9.99-7.51,18.59-17.66,29.05-26,11.97-8.06,46.48-42.15,72.77-72.74,117.87-120.71,14.8-391.43,4.53-414.37Z",fill:"#e05922"})})}function b(e){let{...t}=e;return(0,m.jsxs)("svg",{viewBox:"0 0 1024 1024",xmlns:"http://www.w3.org/2000/svg",...t,children:[(0,m.jsx)("path",{d:"M341.49,413.75c-4.13,3.29-7.85,7.08-11.69,10.63-3.67,3.77-7.31,7.52-10.71,11.48-13.54,15.88-25.22,33.01-34.21,51.46-18.33,36.64-27.95,76.95-29.36,116.91-.64,19.99,.85,39.91,4.29,59.3,3.76,19.34,9.4,38.15,17.34,55.83,7.83,17.71,18.02,34.17,30.16,49.5,12.02,15.37,25.79,29.57,40.97,42.3,15.05,12.88,31.44,24.38,48.76,34.48,17.22,10.29,35.46,19.01,54.23,26.65,18.76,7.69,38.19,13.94,57.97,19.23,19.85,5.05,40.06,9.06,60.53,12.06,20.48,3,41.23,4.61,62.11,5.66,20.89,.87,41.91,.7,63.02-.11,21.08-1.09,42.29-2.61,63.49-5.17,21.26-2.26,125.44-27.98,33.77,3.27-10.2,3.64-20.42,7.38-30.83,10.52-20.71,6.81-41.99,11.73-63.46,16.19-21.55,3.99-43.37,7.12-65.41,8.72-5.5,.51-11.03,.67-16.56,.93-5.53,.27-11.07,.48-16.62,.43-5.55,0-11.1,.12-16.66,.01l-16.68-.66c-44.48-2.39-89.13-10.67-131.81-26.02-21.35-7.61-42.13-17.1-62.13-28.2-19.9-11.28-38.93-24.35-56.6-39.25-17.54-15.05-33.78-31.87-47.88-50.57-13.9-18.74-26.19-39.3-34.7-61.32-8.61-21.93-14.47-44.79-17.21-67.89-2.73-23.1-2.51-46.35,.21-69.14,5.74-45.52,21.81-89.49,47.76-126.49,19.27-27.67,45.15-50.46,73.59-67.06,15.58-9.09,12.19-.88,4.33,6.3Z",fill:"#e05922"}),(0,m.jsx)("path",{d:"M459.32,516.03c4.44-3.84,10.57,1.79,7.52,6.81-4.99,8.18-7.54,17.19-6.95,25.72,.51,8.41,3.67,15.98,8.7,22.02,4.93,5.69,11.76,10.38,19.26,13.32,15.06,6.17,31.86,5.47,44.44-.91,12.66-6.28,23.08-18.15,28.39-31.95,5.51-13.78,5.92-28.56,1.92-41.61-1.95-6.6-5.4-12.76-9.53-18.72l-1.51-2.25-1.65-1.92-1.6-1.95-1.88-1.88c-1.19-1.28-2.47-2.53-3.91-3.69-1.37-1.2-2.7-2.43-4.26-3.48-11.78-9.12-26.26-15.18-40.74-17.29-7.29-.85-14.59-1.1-21.9-.32-1.83,.23-3.64,.65-5.47,.93l-2.74,.44-2.71,.76-5.19,1.38c-.17,.05-.34,.1-.51,.16l-5.12,1.95c-14.45,5.37-27.44,13.84-37.15,25.22-10.01,11.1-17.29,24.81-21.02,39.89-3.75,15.11-3.76,31.41-.7,47.65,1.48,8.14,4.12,16.18,7.32,24.12,2.98,7.83,6.67,14.57,11.45,21.26,18.93,26.56,53.25,45,90.6,51.63,18.72,3.52,38.35,4.07,57.9,1.5,9.88-1.38,19.8-3.54,29.53-6.75,5.52-1.82,9.38,5.57,4.65,8.96-8.94,6.4-18.81,11.76-29.12,16-20.39,8.5-42.63,12.95-65.24,13.53-22.6,.48-45.76-2.69-68.01-10.85-11.16-3.97-22-9.31-32.32-15.87-10.29-6.58-20.16-14.38-28.69-23.73-8.53-9.18-16.21-19.97-21.97-30.92-5.74-10.74-10.56-22.15-14.02-34.12-7.19-23.87-8.8-49.99-4.32-75.53,4.59-25.55,16.38-50.09,33.24-70.47,17-20.42,39.62-35.63,63.87-44.89l9.02-3.2c.14-.05,.28-.09,.42-.13l9.38-2.44,4.81-1.18,4.93-.81c3.3-.5,6.57-1.09,9.89-1.45,13.33-1.25,26.85-.88,40.07,1.34,26.3,5.06,51,16.33,70.94,33.99,2.56,2.1,4.94,4.49,7.31,6.88,2.41,2.31,4.69,4.83,6.88,7.48l3.18,3.77c.08,.09,.15,.19,.22,.29l2.99,4.1,2.92,4.11c.08,.11,.15,.22,.22,.34l2.5,4.17c6.99,11.47,12.28,24.45,14.99,38.08,2.75,13.59,3.16,27.69,.94,41.21-2.22,13.53-6.8,26.43-13.32,38.13-6.47,11.72-15.09,22.19-25.47,30.81-10.17,8.62-22.47,15.42-35.6,19.14-13.12,3.76-26.95,4.55-39.89,2.27-12.95-2.27-25.1-7.13-35.66-14.31-10.47-7.25-19.72-16.64-25.64-28.65-5.59-11.7-7.51-25.35-4.7-37.54,1.25-6.13,3.74-11.79,6.69-16.94,3.23-5.02,6.85-9.61,11.16-13.36,.09-.07,.17-.15,.26-.22Z",fill:"#018080"}),(0,m.jsx)("path",{d:"M98.42,646.77l.07-16.47,.52-16.51c1.03-22.04,3.8-44.08,8.11-65.99,8.71-43.73,24.57-87.01,49.25-125.97,24.47-38.99,58.12-73.12,97.84-98.05,19.84-12.47,41.14-22.6,63.18-30.52,22.13-7.67,44.99-13.18,68.18-16.14,46.35-5.78,94-2.55,139.16,12.63,11.27,3.81,22.39,8.57,33.11,14.01,10.64,5.57,20.7,11.96,30.36,18.8,9.61,6.92,18.64,14.52,27.17,22.58,8.41,8.19,16.39,16.76,23.76,25.83,29.1,36.36,48.92,80.72,52.58,126.94,1.85,23.01-.5,46.21-7.07,67.84-6.51,21.64-22.14,38.76-19.52,28.34,3.07-10.23,4.93-20.71,6.18-31.1,1.03-10.45,1.32-20.85,.77-31.14-.86-10.29-1.86-20.48-4.35-30.41-9.05-39.83-30.33-75.81-58.1-104.69-27.76-29.04-62.14-50.98-99.26-61.61-37.74-10.7-78.7-12.91-118.28-7.67-39.65,5.17-78.28,17.85-112.81,37.81-34.75,19.7-64.98,47.12-89.77,79.46-24.82,32.41-43.74,69.97-58.02,109.77-7.22,19.9-13.15,40.44-18.35,61.32l-3.68,15.75-3.28,15.89-3.14,15.99c-2.34,13.37-4.27-5.72-4.62-16.7Z",fill:"#1ab7d1"}),(0,m.jsx)("path",{d:"M922.46,330.55c.86,10.35,1.65,20.72,2.14,31.12,1.23,20.81,1.19,41.7,.44,62.66-1.75,41.88-7.37,84.19-19.75,125.55-12.21,41.26-31.57,81.7-59.3,116.49-13.72,17.46-29.51,33.28-46.44,47.55-17.03,14.18-35.39,26.53-54.36,37.43-19.06,10.8-38.88,19.76-58.94,27.65-10.03,4-20.23,7.31-30.39,10.74l-15.32,4.71c-5.03,1.57-10.71,3.08-16.05,4.29-21.66,5.07-43.48,8.01-65.46,8.76-21.96,.74-44.01-.54-65.72-4.41-43.33-7.44-85.59-25.6-118.6-54.31-16.49-14.26-30.84-30.66-42.15-48.7-2.71-4.6-5.55-9.06-8.08-13.7-2.4-4.72-4.98-9.3-7.18-14.06-2.05-4.85-4.2-9.62-6.12-14.46-1.81-4.9,.67-6.16,3.5-1.87,2.98,4.2,6.16,8.26,9.2,12.35,3.1,4.05,6.42,7.92,9.59,11.86,3.28,3.86,6.83,7.48,10.19,11.2,13.95,14.48,29.35,27.1,45.95,37.33,33.09,20.71,70.7,31.76,108.92,34.74,38.17,3.02,77.34-1.07,113.89-11.4,4.64-1.31,8.86-2.66,13.62-4.25l14.06-4.78c9.24-3.45,18.56-6.75,27.54-10.66,18.15-7.41,35.87-15.62,52.57-25.27,33.62-18.89,63.84-42.56,88.67-71.32,12.47-14.33,23.53-29.95,33.31-46.54,9.68-16.64,18.18-34.2,25.49-52.44,14.8-36.45,24.81-75.52,32.69-115.35,7.99-39.88,21.11-101.25,22.1-90.9Z",fill:"#9d744f"}),(0,m.jsx)("path",{d:"M733.6,626.97c-2.93,5.29-10.97,2.03-9.45-3.82,14.58-55.99,21.84-112.28,17.4-167.2-4.75-66.45-28.39-129.68-68.89-179.51-40.28-50.26-98.17-84.83-162.02-104.09-31.99-9.66-65.52-16.11-99.82-19.24-8.56-1.08-17.19-1.22-25.85-1.97-8.67-.88-17.35-.84-26.1-1.18-8.74-.25-17.55-.77-26.32-.41-8.81-.04-12.77-4.21-5.81-6.22,4.67-1.35,9.39-2.55,14.14-3.62,14.27-3.21,28.63-6.77,43.25-8.46,8.91-1.08,17.84-2.5,26.84-3.13,35.99-3.21,72.78-1.35,109.11,5.38,36.3,6.8,72.1,19.32,105.08,37.84,4.24,2.1,8.22,4.67,12.17,7.29l11.92,7.76c7.78,5.31,15.07,11.35,22.61,17.04,.11,.08,.22,.17,.32,.26,7.01,6.34,14.07,12.66,20.79,19.32l9.73,10.5c3.25,3.5,6.54,6.96,9.37,10.81,5.76,7.61,11.94,14.92,16.9,23.09l7.81,11.98c2.49,4.06,4.67,8.3,6.94,12.49,2.19,4.23,4.6,8.35,6.58,12.66l5.73,13.05c4.09,8.59,6.86,17.68,9.95,26.62,11.69,36,16.61,73.75,15.42,110.81-1.86,61.23-19.21,120.32-47.78,171.95Z",fill:"#00966f"})]})}const v={xs:"h-4 w-4 md:h-5 md:w-5",sm:"h-5 w-5 md:h-6 md:w-6",md:"h-6 w-6 md:h-7 md:w-7",lg:"h-7 w-7 md:h-8 md:w-8",xl:"h-8 w-8 md:h-9 md:w-9","2xl":"h-9 w-9 md:h-10 md:w-10"},w={xs:"text-xs leading-4 md:text-sm md:leading-5",sm:"text-sm leading-5 md:text-base md:leading-6",md:"text-base leading-6 md:text-lg md:leading-7",lg:"text-lg leading-7 md:text-xl md:leading-8",xl:"text-xl leading-8 md:text-2xl md:leading-9","2xl":"text-2xl leading-9 md:text-3xl md:leading-10"},x={vortico:"text-vortico-500",bosque:"text-bosque-500",bruma:"text-bruma-500",ciclon:"text-ciclon-500",flama:"text-flama-500",primary:"text-primary-500","vortico-light":"text-vortico-200","bosque-light":"text-bosque-200","bruma-light":"text-bruma-200","ciclon-light":"text-ciclon-200","flama-light":"text-flama-200","primary-light":"text-primary-200","vortico-dark":"text-vortico-800","bosque-dark":"text-bosque-800","bruma-dark":"text-bruma-800","ciclon-dark":"text-ciclon-800","flama-dark":"text-flama-800","primary-dark":"text-primary-800"},k={flama:{text:"Flama",icon:(0,m.jsx)(y,{})},vortico:{text:"Vortico",icon:(0,m.jsx)(b,{})}};function S(e){let{logo:t,color:n,size:r="md"}=e;const{text:a,icon:o}=k[t],i=n?x[n]:"";return(0,m.jsxs)("div",{className:"flex items-center justify-center gap-1","aria-label":`${a} logo`,children:[(0,m.jsx)("div",{className:`${v[r]} ${i}`,children:o}),(0,m.jsx)("span",{className:`text-brand ${w[r]} ${i}`,children:a})]})}function C(e){let{onClick:t,...n}=e;const r=[{href:"https://flama.dev/docs/",title:"Docs",icon:(0,m.jsx)(c,{}),label:"Flama docs"},{href:"https://flama.dev/blog/",title:"Blog",icon:(0,m.jsx)(p,{}),label:"Flama blog"}];return(0,m.jsx)("nav",{children:(0,m.jsx)("ul",{...n,children:r.map(((e,n)=>{let{href:r,title:a,icon:o,label:i}=e;return(0,m.jsx)("li",{onClick:t,children:(0,m.jsx)("a",{href:r,"aria-label":i,children:(0,m.jsxs)("div",{className:"flex h-12 cursor-pointer items-center justify-start gap-2",children:[(0,m.jsx)("div",{className:"text-primary-400 hover:text-flama-500 h-6 w-6 transition-colors duration-200 md:hidden",children:o}),(0,m.jsx)("div",{className:"text-primary-400 hover:text-flama-500 text-md font-medium transition-colors duration-200",children:a})]})})},n)}))})})}function _(e){let{...t}=e;const n=[{href:"https://github.com/vortico/flama",title:"GitHub",icon:(0,m.jsx)(d,{}),label:"Flama on GitHub"}];return(0,m.jsx)("div",{children:(0,m.jsx)("ul",{...t,children:n.map(((e,t)=>{let{href:n,title:r,icon:a,label:o}=e;return(0,m.jsx)("li",{children:(0,m.jsx)("a",{href:n,"aria-label":o,children:(0,m.jsxs)("div",{className:"flex h-12 cursor-pointer items-center justify-start gap-2",children:[(0,m.jsx)("div",{className:"text-primary-400 hover:text-flama-500 h-6 w-6 transition-colors duration-200",children:a}),(0,m.jsx)("div",{className:"text-primary-400 hover:text-flama-500 text-md font-medium transition-colors duration-200 md:hidden",children:r})]})})},t)}))})})}function E(e){let{onClose:t}=e;return(0,m.jsxs)("div",{className:"bg-primary-100 fixed inset-x-0 top-0 z-50 h-screen w-screen overflow-auto","aria-modal":"true",role:"dialog",children:[(0,m.jsxs)("div",{className:"border-flama-500 flex h-14 w-full items-center justify-between border-b px-4 sm:px-6 md:px-8",children:[(0,m.jsx)("a",{href:"https://flama.dev",className:"block cursor-pointer","aria-label":"Flama website",children:(0,m.jsx)(S,{logo:"flama",color:"flama",size:"lg"})}),(0,m.jsx)("button",{className:"h-6 w-6",onClick:t,"aria-label":"Close menu",children:(0,m.jsx)(f,{className:"text-primary-400 hover:text-flama-500 h-full w-full transition-colors duration-200"})})]}),(0,m.jsxs)("div",{className:"w-full px-4 sm:px-6 md:px-8",children:[(0,m.jsx)(C,{className:"flex flex-col items-start justify-start",onClick:t}),(0,m.jsx)(_,{className:"flex flex-col items-start justify-start gap-8"})]})]})}function N(e){let{onOpen:t}=e;return(0,m.jsx)(m.Fragment,{children:(0,m.jsxs)("div",{className:"flex h-14 w-full items-center justify-between",children:[(0,m.jsx)("a",{href:"https://flama.dev",className:"block cursor-pointer","aria-label":"Flama website",children:(0,m.jsx)(S,{logo:"flama",color:"flama",size:"lg"})}),(0,m.jsxs)("div",{className:"hidden items-center justify-between gap-8 md:flex",children:[(0,m.jsx)(C,{className:"flex items-center justify-start gap-8"}),(0,m.jsx)(_,{className:"flex items-center justify-start gap-8"})]}),(0,m.jsx)("button",{className:"block h-6 w-6 cursor-pointer md:hidden",onClick:t,"aria-label":"Open menu",children:(0,m.jsx)(h,{className:"h-full w-full"})})]})})}function P(){const[e,t]=(0,l.useState)(!1),n=(0,l.useCallback)((()=>t(!0)),[t]),r=(0,l.useCallback)((()=>t(!1)),[t]);return(0,m.jsxs)("div",{className:"border-flama-500 bg-primary-100 fixed inset-x-0 top-0 z-10 mx-auto border-b",children:[(0,m.jsx)("div",{className:"divide-flama-500/50 mx-auto w-full divide-y px-4 sm:px-6 md:px-8",children:(0,m.jsx)(N,{onOpen:n})}),e&&(0,g.createPortal)((0,m.jsx)(E,{onClose:r}),document.body)]})}function A(e){let{title:t,border:n=!0,gradient:r=!0,className:a="",children:o,...i}=e;return(0,m.jsxs)("section",{className:`my-14 ${n?"border-flama-500/50 border-t":""} ${a}`,...i,children:[(0,m.jsx)("div",{className:r?"from-flama-500/10 bg-linear-to-b":"",children:t&&(0,m.jsx)("div",{className:"max-w-8xl mx-auto px-4 py-8 sm:px-6 sm:py-10 md:px-8 md:py-12",children:(0,m.jsx)("h2",{className:"text-primary-700 text-2xl font-semibold",children:t})})}),(0,m.jsx)("div",{className:"max-w-8xl mx-auto px-4 sm:px-6 md:px-8",children:o})]})}function L(e){let{error:t,method:n,path:r,description:a}=e;return(0,m.jsxs)(m.Fragment,{children:[(0,m.jsxs)("h1",{className:"text-2xl",children:[(0,m.jsx)("span",{className:"text-flama-500 font-bold",children:t}),(0,m.jsx)("span",{className:"pl-2",children:"raised at"}),(0,m.jsx)("span",{className:"pl-2 font-bold",children:n}),(0,m.jsx)("span",{className:"pl-2 font-mono",children:r})]}),a&&(0,m.jsx)("h2",{className:"text-xl font-medium",children:a})]})}var T=u("filled","circle-filled","IconCircleFilled",[["path",{d:"M7 3.34a10 10 0 1 1 -4.995 8.984l-.005 -.324l.005 -.324a10 10 0 0 1 4.995 -8.336z",key:"svg-0"}]]),R=u("filled","circle-dot-filled","IconCircleDotFilled",[["path",{d:"M17 3.34a10 10 0 1 1 -14.995 8.984l-.005 -.324l.005 -.324a10 10 0 0 1 14.995 -8.336zm-5 6.66a2 2 0 0 0 -1.977 1.697l-.018 .154l-.005 .149l.005 .15a2 2 0 1 0 1.995 -2.15z",key:"svg-0"}]]),I=u("filled","circle-plus-filled","IconCirclePlusFilled",[["path",{d:"M4.929 4.929a10 10 0 1 1 14.141 14.141a10 10 0 0 1 -14.14 -14.14zm8.071 4.071a1 1 0 1 0 -2 0v2h-2a1 1 0 1 0 0 2h2v2a1 1 0 1 0 2 0v-2h2a1 1 0 1 0 0 -2h-2v-2z",key:"svg-0"}]]),M=u("filled","circle-x-filled","IconCircleXFilled",[["path",{d:"M17 3.34a10 10 0 1 1 -14.995 8.984l-.005 -.324l.005 -.324a10 10 0 0 1 14.995 -8.336zm-6.489 5.8a1 1 0 0 0 -1.218 1.567l1.292 1.293l-1.292 1.293l-.083 .094a1 1 0 0 0 1.497 1.32l1.293 -1.292l1.293 1.292l.094 .083a1 1 0 0 0 1.32 -1.497l-1.292 -1.293l1.292 -1.293l.083 -.094a1 1 0 0 0 -1.497 -1.32l-1.293 1.292l-1.293 -1.292l-.094 -.083z",key:"svg-0"}]]);function O(e){let{title:t,autoScroll:n,className:r,children:a}=e;const o=(0,l.useRef)(null),[i,s]=(0,l.useState)("open"),u=(0,l.useCallback)((()=>s("open"===i?"closed":"open")),[i,s]),c=(0,l.useCallback)((()=>s("full")),[s]),p=(0,l.useCallback)((()=>s("closed")),[s]);return(0,l.useEffect)((()=>{o.current&&n?.current&&o.current.scrollTo({top:n.current.offsetTop-o.current.clientHeight/2,left:0,behavior:"smooth"})}),[n,o]),(0,m.jsxs)("div",{className:"full"===i?"fixed inset-0 z-50 h-screen w-screen overflow-hidden":"relative h-full w-full p-2",..."full"===i&&{"aria-modal":!0,role:"dialog"},children:["full"===i&&(0,m.jsx)("div",{className:"bg-primary-950/30 fixed inset-0 backdrop-blur-sm","aria-hidden":!0,onClick:u}),(0,m.jsxs)("div",{className:`bg-primary-900 shadow-primary-300 border-primary-300 h-full overflow-hidden border shadow-md ${"closed"===i?"relative max-h-[32px]":"full"===i?"fixed inset-x-4 inset-y-[5vh] max-h-[calc(85vh+2rem+3px)] sm:inset-x-6 md:inset-x-8":"open"===i?"relative max-h-full":""} ${r||""}`,children:[(0,m.jsxs)("div",{className:"border-primary-300 bg-primary-900 mb-px flex h-8 w-full flex-none items-center justify-between border-b px-4",children:[(0,m.jsx)("span",{className:"text-primary-200 font-alternative truncate text-sm leading-8 font-semibold",children:t}),(0,m.jsxs)("div",{className:"flex items-center justify-end gap-x-2",children:[(0,m.jsx)("button",{className:"h-4 w-4 cursor-pointer",onClick:u,"aria-label":"Minimize Window",children:(0,m.jsx)(R,{className:"text-primary-300 hover:text-ciclon h-full w-full transition-colors duration-200"})}),(0,m.jsx)("button",{className:"h-4 w-4 cursor-pointer",onClick:c,"aria-label":"Maximize Window",children:(0,m.jsx)(I,{className:"text-primary-300 hover:text-bosque h-full w-full transition-colors duration-200"})}),(0,m.jsx)("button",{className:"h-4 w-4 cursor-pointer",onClick:p,"aria-label":"Close Window",children:(0,m.jsx)(M,{className:"text-primary-300 hover:text-flama h-full w-full transition-colors duration-200"})})]})]}),(0,m.jsx)("div",{ref:o,className:"bg-primary-900 h-full w-full "+("closed"===i?"max-h-0":"full"===i||"open"===i?"max-h-[calc(100%-2rem-1px)]":""),children:a})]})]})}var j=u("outline","check","IconCheck",[["path",{d:"M5 12l5 5l10 -10",key:"svg-0"}]]),z=u("outline","clipboard-copy","IconClipboardCopy",[["path",{d:"M9 5h-2a2 2 0 0 0 -2 2v12a2 2 0 0 0 2 2h3m9 -9v-5a2 2 0 0 0 -2 -2h-2",key:"svg-0"}],["path",{d:"M13 17v-1a1 1 0 0 1 1 -1h1m3 0h1a1 1 0 0 1 1 1v1m0 3v1a1 1 0 0 1 -1 1h-1m-3 0h-1a1 1 0 0 1 -1 -1v-1",key:"svg-1"}],["path",{d:"M9 3m0 2a2 2 0 0 1 2 -2h2a2 2 0 0 1 2 2v0a2 2 0 0 1 -2 2h-2a2 2 0 0 1 -2 -2z",key:"svg-2"}]]);function $(e){let{code:t}=e;const[n,r]=(0,l.useState)(!1),a=(0,l.useCallback)((async()=>{await navigator.clipboard.writeText(t),r(!0),setTimeout((()=>r(!1)),3e3)}),[t,r]);return(0,m.jsx)("div",{className:"bg-primary-500/30 absolute top-4 right-4 flex h-8 w-8 items-center justify-center rounded opacity-0 ring-1 backdrop-blur transition-opacity duration-500 ring-inset group-hover:opacity-100 "+(n?"ring-brand-500/50":"ring-primary-500/50"),children:(0,m.jsx)("button",{onClick:a,"aria-label":"Copy to Clipboard",children:n?(0,m.jsx)(j,{className:"text-brand-500 h-6 w-6"}):(0,m.jsx)(z,{className:"text-primary-500 h-6 w-6"})})})}function F(e){let{type:t,lines:n,token:r,selectedLine:a}=e;return(0,m.jsx)("div",{className:"border-primary-500 bg-primary-900 hidden h-fit min-h-full w-fit flex-none border-r py-1 select-none md:sticky md:left-0 md:block","aria-hidden":"true",children:Array.from(Array(n).keys()).map((e=>(0,m.jsx)("div",{className:"px-2 text-right font-mono text-xs leading-5 "+(e+1===a?"text-flama-500 font-semibold":"text-primary-500"),children:"token"===t?r:e+1},`line-number-${e+1}`)))})}function D(e){let{lines:t={type:"number"},copyButton:n=!0,code:r,selectedLine:a,className:o,children:i}=e;const s=(0,l.useRef)(null);return(0,l.useEffect)((()=>{s.current&&a&&s.current.scrollTo({top:20*a-.5*s.current.clientHeight,left:0,behavior:"smooth"})}),[s,a]),(0,m.jsxs)("div",{className:`bg-primary-900 relative flex h-full min-h-full w-full overflow-auto whitespace-pre ${o||""}`,ref:s,children:[t&&(0,m.jsx)(F,{lines:r.split("\n").length-1,type:t.type,token:t.token,selectedLine:a}),(0,m.jsx)("pre",{className:"bg-primary-900 h-fit min-h-full w-fit flex-auto py-1",children:i}),n&&(0,m.jsx)($,{code:r})]})}function B(e){let{className:t,children:n}=e;return(0,m.jsx)("span",{className:t,children:n})}const U=[Object.freeze(JSON.parse('{"displayName":"Python","name":"python","patterns":[{"include":"#statement"},{"include":"#expression"}],"repository":{"annotated-parameter":{"begin":"\\\\b([[:alpha:]_]\\\\w*)\\\\s*(:)","beginCaptures":{"1":{"name":"variable.parameter.function.language.python"},"2":{"name":"punctuation.separator.annotation.python"}},"end":"(,)|(?=\\\\))","endCaptures":{"1":{"name":"punctuation.separator.parameters.python"}},"patterns":[{"include":"#expression"},{"match":"=(?!=)","name":"keyword.operator.assignment.python"}]},"assignment-operator":{"match":"<<=|>>=|//=|\\\\*\\\\*=|\\\\+=|-=|/=|@=|\\\\*=|%=|~=|\\\\^=|&=|\\\\|=|=(?!=)","name":"keyword.operator.assignment.python"},"backticks":{"begin":"\\\\`","end":"(?:\\\\`|(?<!\\\\\\\\)(\\\\n))","name":"invalid.deprecated.backtick.python","patterns":[{"include":"#expression"}]},"builtin-callables":{"patterns":[{"include":"#illegal-names"},{"include":"#illegal-object-name"},{"include":"#builtin-exceptions"},{"include":"#builtin-functions"},{"include":"#builtin-types"}]},"builtin-exceptions":{"match":"(?<!\\\\.)\\\\b((Arithmetic|Assertion|Attribute|Buffer|BlockingIO|BrokenPipe|ChildProcess|(Connection(Aborted|Refused|Reset)?)|EOF|Environment|FileExists|FileNotFound|FloatingPoint|IO|Import|Indentation|Index|Interrupted|IsADirectory|NotADirectory|Permission|ProcessLookup|Timeout|Key|Lookup|Memory|Name|NotImplemented|OS|Overflow|Reference|Runtime|Recursion|Syntax|System|Tab|Type|UnboundLocal|Unicode(Encode|Decode|Translate)?|Value|Windows|ZeroDivision|ModuleNotFound)Error|((Pending)?Deprecation|Runtime|Syntax|User|Future|Import|Unicode|Bytes|Resource)?Warning|SystemExit|Stop(Async)?Iteration|KeyboardInterrupt|GeneratorExit|(Base)?Exception)\\\\b","name":"support.type.exception.python"},"builtin-functions":{"patterns":[{"match":"(?<!\\\\.)\\\\b(__import__|abs|aiter|all|any|anext|ascii|bin|breakpoint|callable|chr|compile|copyright|credits|delattr|dir|divmod|enumerate|eval|exec|exit|filter|format|getattr|globals|hasattr|hash|help|hex|id|input|isinstance|issubclass|iter|len|license|locals|map|max|memoryview|min|next|oct|open|ord|pow|print|quit|range|reload|repr|reversed|round|setattr|sorted|sum|vars|zip)\\\\b","name":"support.function.builtin.python"},{"match":"(?<!\\\\.)\\\\b(file|reduce|intern|raw_input|unicode|cmp|basestring|execfile|long|xrange)\\\\b","name":"variable.legacy.builtin.python"}]},"builtin-possible-callables":{"patterns":[{"include":"#builtin-callables"},{"include":"#magic-names"}]},"builtin-types":{"match":"(?<!\\\\.)\\\\b(bool|bytearray|bytes|classmethod|complex|dict|float|frozenset|int|list|object|property|set|slice|staticmethod|str|tuple|type|super)\\\\b","name":"support.type.python"},"call-wrapper-inheritance":{"begin":"\\\\b(?=([[:alpha:]_]\\\\w*)\\\\s*(\\\\())","end":"(\\\\))","endCaptures":{"1":{"name":"punctuation.definition.arguments.end.python"}},"name":"meta.function-call.python","patterns":[{"include":"#inheritance-name"},{"include":"#function-arguments"}]},"class-declaration":{"patterns":[{"begin":"\\\\s*(class)\\\\s+(?=[[:alpha:]_]\\\\w*\\\\s*(:|\\\\())","beginCaptures":{"1":{"name":"storage.type.class.python"}},"end":"(:)","endCaptures":{"1":{"name":"punctuation.section.class.begin.python"}},"name":"meta.class.python","patterns":[{"include":"#class-name"},{"include":"#class-inheritance"}]}]},"class-inheritance":{"begin":"(\\\\()","beginCaptures":{"1":{"name":"punctuation.definition.inheritance.begin.python"}},"end":"(\\\\))","endCaptures":{"1":{"name":"punctuation.definition.inheritance.end.python"}},"name":"meta.class.inheritance.python","patterns":[{"match":"(\\\\*\\\\*|\\\\*)","name":"keyword.operator.unpacking.arguments.python"},{"match":",","name":"punctuation.separator.inheritance.python"},{"match":"=(?!=)","name":"keyword.operator.assignment.python"},{"match":"\\\\bmetaclass\\\\b","name":"support.type.metaclass.python"},{"include":"#illegal-names"},{"include":"#class-kwarg"},{"include":"#call-wrapper-inheritance"},{"include":"#expression-base"},{"include":"#member-access-class"},{"include":"#inheritance-identifier"}]},"class-kwarg":{"captures":{"1":{"name":"entity.other.inherited-class.python variable.parameter.class.python"},"2":{"name":"keyword.operator.assignment.python"}},"match":"\\\\b([[:alpha:]_]\\\\w*)\\\\s*(=)(?!=)"},"class-name":{"patterns":[{"include":"#illegal-object-name"},{"include":"#builtin-possible-callables"},{"match":"\\\\b([[:alpha:]_]\\\\w*)\\\\b","name":"entity.name.type.class.python"}]},"codetags":{"captures":{"1":{"name":"keyword.codetag.notation.python"}},"match":"(?:\\\\b(NOTE|XXX|HACK|FIXME|BUG|TODO)\\\\b)"},"comments":{"patterns":[{"begin":"(?:\\\\#\\\\s*(type:)\\\\s*+(?!$|\\\\#))","beginCaptures":{"0":{"name":"meta.typehint.comment.python"},"1":{"name":"comment.typehint.directive.notation.python"}},"contentName":"meta.typehint.comment.python","end":"(?:$|(?=\\\\#))","name":"comment.line.number-sign.python","patterns":[{"match":"\\\\Gignore(?=\\\\s*(?:$|\\\\#))","name":"comment.typehint.ignore.notation.python"},{"match":"(?<!\\\\.)\\\\b(bool|bytes|float|int|object|str|List|Dict|Iterable|Sequence|Set|FrozenSet|Callable|Union|Tuple|Any|None)\\\\b","name":"comment.typehint.type.notation.python"},{"match":"([\\\\[\\\\]\\\\(\\\\),\\\\.\\\\=\\\\*]|(->))","name":"comment.typehint.punctuation.notation.python"},{"match":"([[:alpha:]_]\\\\w*)","name":"comment.typehint.variable.notation.python"}]},{"include":"#comments-base"}]},"comments-base":{"begin":"(\\\\#)","beginCaptures":{"1":{"name":"punctuation.definition.comment.python"}},"end":"($)","name":"comment.line.number-sign.python","patterns":[{"include":"#codetags"}]},"comments-string-double-three":{"begin":"(\\\\#)","beginCaptures":{"1":{"name":"punctuation.definition.comment.python"}},"end":"($|(?=\\"\\"\\"))","name":"comment.line.number-sign.python","patterns":[{"include":"#codetags"}]},"comments-string-single-three":{"begin":"(\\\\#)","beginCaptures":{"1":{"name":"punctuation.definition.comment.python"}},"end":"($|(?=\'\'\'))","name":"comment.line.number-sign.python","patterns":[{"include":"#codetags"}]},"curly-braces":{"begin":"\\\\{","beginCaptures":{"0":{"name":"punctuation.definition.dict.begin.python"}},"end":"\\\\}","endCaptures":{"0":{"name":"punctuation.definition.dict.end.python"}},"patterns":[{"match":":","name":"punctuation.separator.dict.python"},{"include":"#expression"}]},"decorator":{"begin":"^\\\\s*((@))\\\\s*(?=[[:alpha:]_]\\\\w*)","beginCaptures":{"1":{"name":"entity.name.function.decorator.python"},"2":{"name":"punctuation.definition.decorator.python"}},"end":"(\\\\))(?:(.*?)(?=\\\\s*(?:\\\\#|$)))|(?=\\\\n|\\\\#)","endCaptures":{"1":{"name":"punctuation.definition.arguments.end.python"},"2":{"name":"invalid.illegal.decorator.python"}},"name":"meta.function.decorator.python","patterns":[{"include":"#decorator-name"},{"include":"#function-arguments"}]},"decorator-name":{"patterns":[{"include":"#builtin-callables"},{"include":"#illegal-object-name"},{"captures":{"2":{"name":"punctuation.separator.period.python"}},"match":"([[:alpha:]_]\\\\w*)|(\\\\.)","name":"entity.name.function.decorator.python"},{"include":"#line-continuation"},{"captures":{"1":{"name":"invalid.illegal.decorator.python"}},"match":"\\\\s*([^([:alpha:]\\\\s_\\\\.#\\\\\\\\].*?)(?=\\\\#|$)","name":"invalid.illegal.decorator.python"}]},"docstring":{"patterns":[{"begin":"(\\\\\'\\\\\'\\\\\'|\\\\\\"\\\\\\"\\\\\\")","beginCaptures":{"1":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\1)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"}},"name":"string.quoted.docstring.multi.python","patterns":[{"include":"#docstring-prompt"},{"include":"#codetags"},{"include":"#docstring-guts-unicode"}]},{"begin":"([rR])(\\\\\'\\\\\'\\\\\'|\\\\\\"\\\\\\"\\\\\\")","beginCaptures":{"1":{"name":"storage.type.string.python"},"2":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\2)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"}},"name":"string.quoted.docstring.raw.multi.python","patterns":[{"include":"#string-consume-escape"},{"include":"#docstring-prompt"},{"include":"#codetags"}]},{"begin":"(\\\\\'|\\\\\\")","beginCaptures":{"1":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\1)|(\\\\n)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.quoted.docstring.single.python","patterns":[{"include":"#codetags"},{"include":"#docstring-guts-unicode"}]},{"begin":"([rR])(\\\\\'|\\\\\\")","beginCaptures":{"1":{"name":"storage.type.string.python"},"2":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\2)|(\\\\n)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.quoted.docstring.raw.single.python","patterns":[{"include":"#string-consume-escape"},{"include":"#codetags"}]}]},"docstring-guts-unicode":{"patterns":[{"include":"#escape-sequence-unicode"},{"include":"#escape-sequence"},{"include":"#string-line-continuation"}]},"docstring-prompt":{"captures":{"1":{"name":"keyword.control.flow.python"}},"match":"(?:(?:^|\\\\G)\\\\s*((?:>>>|\\\\.\\\\.\\\\.)\\\\s)(?=\\\\s*\\\\S))"},"docstring-statement":{"begin":"^(?=\\\\s*[rR]?(\\\\\'\\\\\'\\\\\'|\\\\\\"\\\\\\"\\\\\\"|\\\\\'|\\\\\\"))","end":"((?<=\\\\1)|^)(?!\\\\s*[rR]?(\\\\\'\\\\\'\\\\\'|\\\\\\"\\\\\\"\\\\\\"|\\\\\'|\\\\\\"))","patterns":[{"include":"#docstring"}]},"double-one-regexp-character-set":{"patterns":[{"match":"\\\\[\\\\^?\\\\](?!.*?\\\\])"},{"begin":"(\\\\[)(\\\\^)?(\\\\])?","beginCaptures":{"1":{"name":"punctuation.character.set.begin.regexp constant.other.set.regexp"},"2":{"name":"keyword.operator.negation.regexp"},"3":{"name":"constant.character.set.regexp"}},"end":"(\\\\]|(?=\\"))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"punctuation.character.set.end.regexp constant.other.set.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.character.set.regexp","patterns":[{"include":"#regexp-charecter-set-escapes"},{"match":"[^\\\\n]","name":"constant.character.set.regexp"}]}]},"double-one-regexp-comments":{"begin":"\\\\(\\\\?#","beginCaptures":{"0":{"name":"punctuation.comment.begin.regexp"}},"end":"(\\\\)|(?=\\"))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"punctuation.comment.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"name":"comment.regexp","patterns":[{"include":"#codetags"}]},"double-one-regexp-conditional":{"begin":"(\\\\()\\\\?\\\\((\\\\w+(?:\\\\s+[[:alnum:]]+)?|\\\\d+)\\\\)","beginCaptures":{"0":{"name":"keyword.operator.conditional.regexp"},"1":{"name":"punctuation.parenthesis.conditional.begin.regexp"}},"end":"(\\\\)|(?=\\"))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"keyword.operator.conditional.negative.regexp punctuation.parenthesis.conditional.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-one-regexp-expression"}]},"double-one-regexp-expression":{"patterns":[{"include":"#regexp-base-expression"},{"include":"#double-one-regexp-character-set"},{"include":"#double-one-regexp-comments"},{"include":"#regexp-flags"},{"include":"#double-one-regexp-named-group"},{"include":"#regexp-backreference"},{"include":"#double-one-regexp-lookahead"},{"include":"#double-one-regexp-lookahead-negative"},{"include":"#double-one-regexp-lookbehind"},{"include":"#double-one-regexp-lookbehind-negative"},{"include":"#double-one-regexp-conditional"},{"include":"#double-one-regexp-parentheses-non-capturing"},{"include":"#double-one-regexp-parentheses"}]},"double-one-regexp-lookahead":{"begin":"(\\\\()\\\\?=","beginCaptures":{"0":{"name":"keyword.operator.lookahead.regexp"},"1":{"name":"punctuation.parenthesis.lookahead.begin.regexp"}},"end":"(\\\\)|(?=\\"))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"keyword.operator.lookahead.regexp punctuation.parenthesis.lookahead.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-one-regexp-expression"}]},"double-one-regexp-lookahead-negative":{"begin":"(\\\\()\\\\?!","beginCaptures":{"0":{"name":"keyword.operator.lookahead.negative.regexp"},"1":{"name":"punctuation.parenthesis.lookahead.begin.regexp"}},"end":"(\\\\)|(?=\\"))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"keyword.operator.lookahead.negative.regexp punctuation.parenthesis.lookahead.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-one-regexp-expression"}]},"double-one-regexp-lookbehind":{"begin":"(\\\\()\\\\?<=","beginCaptures":{"0":{"name":"keyword.operator.lookbehind.regexp"},"1":{"name":"punctuation.parenthesis.lookbehind.begin.regexp"}},"end":"(\\\\)|(?=\\"))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"keyword.operator.lookbehind.regexp punctuation.parenthesis.lookbehind.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-one-regexp-expression"}]},"double-one-regexp-lookbehind-negative":{"begin":"(\\\\()\\\\?<!","beginCaptures":{"0":{"name":"keyword.operator.lookbehind.negative.regexp"},"1":{"name":"punctuation.parenthesis.lookbehind.begin.regexp"}},"end":"(\\\\)|(?=\\"))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"keyword.operator.lookbehind.negative.regexp punctuation.parenthesis.lookbehind.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-one-regexp-expression"}]},"double-one-regexp-named-group":{"begin":"(\\\\()(\\\\?P<\\\\w+(?:\\\\s+[[:alnum:]]+)?>)","beginCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.named.begin.regexp"},"2":{"name":"entity.name.tag.named.group.regexp"}},"end":"(\\\\)|(?=\\"))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.named.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.named.regexp","patterns":[{"include":"#double-one-regexp-expression"}]},"double-one-regexp-parentheses":{"begin":"\\\\(","beginCaptures":{"0":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.begin.regexp"}},"end":"(\\\\)|(?=\\"))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-one-regexp-expression"}]},"double-one-regexp-parentheses-non-capturing":{"begin":"\\\\(\\\\?:","beginCaptures":{"0":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.non-capturing.begin.regexp"}},"end":"(\\\\)|(?=\\"))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.non-capturing.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-one-regexp-expression"}]},"double-three-regexp-character-set":{"patterns":[{"match":"\\\\[\\\\^?\\\\](?!.*?\\\\])"},{"begin":"(\\\\[)(\\\\^)?(\\\\])?","beginCaptures":{"1":{"name":"punctuation.character.set.begin.regexp constant.other.set.regexp"},"2":{"name":"keyword.operator.negation.regexp"},"3":{"name":"constant.character.set.regexp"}},"end":"(\\\\]|(?=\\"\\"\\"))","endCaptures":{"1":{"name":"punctuation.character.set.end.regexp constant.other.set.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.character.set.regexp","patterns":[{"include":"#regexp-charecter-set-escapes"},{"match":"[^\\\\n]","name":"constant.character.set.regexp"}]}]},"double-three-regexp-comments":{"begin":"\\\\(\\\\?#","beginCaptures":{"0":{"name":"punctuation.comment.begin.regexp"}},"end":"(\\\\)|(?=\\"\\"\\"))","endCaptures":{"1":{"name":"punctuation.comment.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"name":"comment.regexp","patterns":[{"include":"#codetags"}]},"double-three-regexp-conditional":{"begin":"(\\\\()\\\\?\\\\((\\\\w+(?:\\\\s+[[:alnum:]]+)?|\\\\d+)\\\\)","beginCaptures":{"0":{"name":"keyword.operator.conditional.regexp"},"1":{"name":"punctuation.parenthesis.conditional.begin.regexp"}},"end":"(\\\\)|(?=\\"\\"\\"))","endCaptures":{"1":{"name":"keyword.operator.conditional.negative.regexp punctuation.parenthesis.conditional.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-three-regexp-expression"},{"include":"#comments-string-double-three"}]},"double-three-regexp-expression":{"patterns":[{"include":"#regexp-base-expression"},{"include":"#double-three-regexp-character-set"},{"include":"#double-three-regexp-comments"},{"include":"#regexp-flags"},{"include":"#double-three-regexp-named-group"},{"include":"#regexp-backreference"},{"include":"#double-three-regexp-lookahead"},{"include":"#double-three-regexp-lookahead-negative"},{"include":"#double-three-regexp-lookbehind"},{"include":"#double-three-regexp-lookbehind-negative"},{"include":"#double-three-regexp-conditional"},{"include":"#double-three-regexp-parentheses-non-capturing"},{"include":"#double-three-regexp-parentheses"},{"include":"#comments-string-double-three"}]},"double-three-regexp-lookahead":{"begin":"(\\\\()\\\\?=","beginCaptures":{"0":{"name":"keyword.operator.lookahead.regexp"},"1":{"name":"punctuation.parenthesis.lookahead.begin.regexp"}},"end":"(\\\\)|(?=\\"\\"\\"))","endCaptures":{"1":{"name":"keyword.operator.lookahead.regexp punctuation.parenthesis.lookahead.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-three-regexp-expression"},{"include":"#comments-string-double-three"}]},"double-three-regexp-lookahead-negative":{"begin":"(\\\\()\\\\?!","beginCaptures":{"0":{"name":"keyword.operator.lookahead.negative.regexp"},"1":{"name":"punctuation.parenthesis.lookahead.begin.regexp"}},"end":"(\\\\)|(?=\\"\\"\\"))","endCaptures":{"1":{"name":"keyword.operator.lookahead.negative.regexp punctuation.parenthesis.lookahead.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-three-regexp-expression"},{"include":"#comments-string-double-three"}]},"double-three-regexp-lookbehind":{"begin":"(\\\\()\\\\?<=","beginCaptures":{"0":{"name":"keyword.operator.lookbehind.regexp"},"1":{"name":"punctuation.parenthesis.lookbehind.begin.regexp"}},"end":"(\\\\)|(?=\\"\\"\\"))","endCaptures":{"1":{"name":"keyword.operator.lookbehind.regexp punctuation.parenthesis.lookbehind.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-three-regexp-expression"},{"include":"#comments-string-double-three"}]},"double-three-regexp-lookbehind-negative":{"begin":"(\\\\()\\\\?<!","beginCaptures":{"0":{"name":"keyword.operator.lookbehind.negative.regexp"},"1":{"name":"punctuation.parenthesis.lookbehind.begin.regexp"}},"end":"(\\\\)|(?=\\"\\"\\"))","endCaptures":{"1":{"name":"keyword.operator.lookbehind.negative.regexp punctuation.parenthesis.lookbehind.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-three-regexp-expression"},{"include":"#comments-string-double-three"}]},"double-three-regexp-named-group":{"begin":"(\\\\()(\\\\?P<\\\\w+(?:\\\\s+[[:alnum:]]+)?>)","beginCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.named.begin.regexp"},"2":{"name":"entity.name.tag.named.group.regexp"}},"end":"(\\\\)|(?=\\"\\"\\"))","endCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.named.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.named.regexp","patterns":[{"include":"#double-three-regexp-expression"},{"include":"#comments-string-double-three"}]},"double-three-regexp-parentheses":{"begin":"\\\\(","beginCaptures":{"0":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.begin.regexp"}},"end":"(\\\\)|(?=\\"\\"\\"))","endCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-three-regexp-expression"},{"include":"#comments-string-double-three"}]},"double-three-regexp-parentheses-non-capturing":{"begin":"\\\\(\\\\?:","beginCaptures":{"0":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.non-capturing.begin.regexp"}},"end":"(\\\\)|(?=\\"\\"\\"))","endCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.non-capturing.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#double-three-regexp-expression"},{"include":"#comments-string-double-three"}]},"ellipsis":{"match":"\\\\.\\\\.\\\\.","name":"constant.other.ellipsis.python"},"escape-sequence":{"match":"\\\\\\\\(x[0-9A-Fa-f]{2}|[0-7]{1,3}|[\\\\\\\\\\"\'abfnrtv])","name":"constant.character.escape.python"},"escape-sequence-unicode":{"patterns":[{"match":"\\\\\\\\(u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8}|N\\\\{[\\\\w\\\\s]+?\\\\})","name":"constant.character.escape.python"}]},"expression":{"patterns":[{"include":"#expression-base"},{"include":"#member-access"},{"match":"\\\\b([[:alpha:]_]\\\\w*)\\\\b"}]},"expression-bare":{"patterns":[{"include":"#backticks"},{"include":"#illegal-anno"},{"include":"#literal"},{"include":"#regexp"},{"include":"#string"},{"include":"#lambda"},{"include":"#generator"},{"include":"#illegal-operator"},{"include":"#operator"},{"include":"#curly-braces"},{"include":"#item-access"},{"include":"#list"},{"include":"#odd-function-call"},{"include":"#round-braces"},{"include":"#function-call"},{"include":"#builtin-functions"},{"include":"#builtin-types"},{"include":"#builtin-exceptions"},{"include":"#magic-names"},{"include":"#special-names"},{"include":"#illegal-names"},{"include":"#special-variables"},{"include":"#ellipsis"},{"include":"#punctuation"},{"include":"#line-continuation"}]},"expression-base":{"patterns":[{"include":"#comments"},{"include":"#expression-bare"},{"include":"#line-continuation"}]},"f-expression":{"patterns":[{"include":"#expression-bare"},{"include":"#member-access"},{"match":"\\\\b([[:alpha:]_]\\\\w*)\\\\b"}]},"fregexp-base-expression":{"patterns":[{"include":"#fregexp-quantifier"},{"include":"#fstring-formatting-braces"},{"match":"\\\\{.*?\\\\}"},{"include":"#regexp-base-common"}]},"fregexp-quantifier":{"match":"\\\\{\\\\{(\\\\d+|\\\\d+,(\\\\d+)?|,\\\\d+)\\\\}\\\\}","name":"keyword.operator.quantifier.regexp"},"fstring-fnorm-quoted-multi-line":{"begin":"(\\\\b[fF])([bBuU])?(\'\'\'|\\"\\"\\")","beginCaptures":{"1":{"name":"string.interpolated.python string.quoted.multi.python storage.type.string.python"},"2":{"name":"invalid.illegal.prefix.python"},"3":{"name":"punctuation.definition.string.begin.python string.interpolated.python string.quoted.multi.python"}},"end":"(\\\\3)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python string.interpolated.python string.quoted.multi.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.fstring.python","patterns":[{"include":"#fstring-guts"},{"include":"#fstring-illegal-multi-brace"},{"include":"#fstring-multi-brace"},{"include":"#fstring-multi-core"}]},"fstring-fnorm-quoted-single-line":{"begin":"(\\\\b[fF])([bBuU])?(([\'\\"]))","beginCaptures":{"1":{"name":"string.interpolated.python string.quoted.single.python storage.type.string.python"},"2":{"name":"invalid.illegal.prefix.python"},"3":{"name":"punctuation.definition.string.begin.python string.interpolated.python string.quoted.single.python"}},"end":"(\\\\3)|((?<!\\\\\\\\)\\\\n)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python string.interpolated.python string.quoted.single.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.fstring.python","patterns":[{"include":"#fstring-guts"},{"include":"#fstring-illegal-single-brace"},{"include":"#fstring-single-brace"},{"include":"#fstring-single-core"}]},"fstring-formatting":{"patterns":[{"include":"#fstring-formatting-braces"},{"include":"#fstring-formatting-singe-brace"}]},"fstring-formatting-braces":{"patterns":[{"captures":{"1":{"name":"constant.character.format.placeholder.other.python"},"2":{"name":"invalid.illegal.brace.python"},"3":{"name":"constant.character.format.placeholder.other.python"}},"match":"({)(\\\\s*?)(})"},{"match":"({{|}})","name":"constant.character.escape.python"}]},"fstring-formatting-singe-brace":{"match":"(}(?!}))","name":"invalid.illegal.brace.python"},"fstring-guts":{"patterns":[{"include":"#escape-sequence-unicode"},{"include":"#escape-sequence"},{"include":"#string-line-continuation"},{"include":"#fstring-formatting"}]},"fstring-illegal-multi-brace":{"patterns":[{"include":"#impossible"}]},"fstring-illegal-single-brace":{"begin":"(\\\\{)(?=[^\\\\n}]*$\\\\n?)","beginCaptures":{"1":{"name":"constant.character.format.placeholder.other.python"}},"end":"(\\\\})|(?=\\\\n)","endCaptures":{"1":{"name":"constant.character.format.placeholder.other.python"}},"patterns":[{"include":"#fstring-terminator-single"},{"include":"#f-expression"}]},"fstring-multi-brace":{"begin":"(\\\\{)","beginCaptures":{"1":{"name":"constant.character.format.placeholder.other.python"}},"end":"(\\\\})","endCaptures":{"1":{"name":"constant.character.format.placeholder.other.python"}},"patterns":[{"include":"#fstring-terminator-multi"},{"include":"#f-expression"}]},"fstring-multi-core":{"match":"(.+?)(($\\\\n?)|(?=[\\\\\\\\\\\\}\\\\{]|\'\'\'|\\"\\"\\"))|\\\\n","name":"string.interpolated.python string.quoted.multi.python"},"fstring-normf-quoted-multi-line":{"begin":"(\\\\b[bBuU])([fF])(\'\'\'|\\"\\"\\")","beginCaptures":{"1":{"name":"invalid.illegal.prefix.python"},"2":{"name":"string.interpolated.python string.quoted.multi.python storage.type.string.python"},"3":{"name":"punctuation.definition.string.begin.python string.quoted.multi.python"}},"end":"(\\\\3)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python string.interpolated.python string.quoted.multi.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.fstring.python","patterns":[{"include":"#fstring-guts"},{"include":"#fstring-illegal-multi-brace"},{"include":"#fstring-multi-brace"},{"include":"#fstring-multi-core"}]},"fstring-normf-quoted-single-line":{"begin":"(\\\\b[bBuU])([fF])(([\'\\"]))","beginCaptures":{"1":{"name":"invalid.illegal.prefix.python"},"2":{"name":"string.interpolated.python string.quoted.single.python storage.type.string.python"},"3":{"name":"punctuation.definition.string.begin.python string.quoted.single.python"}},"end":"(\\\\3)|((?<!\\\\\\\\)\\\\n)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python string.interpolated.python string.quoted.single.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.fstring.python","patterns":[{"include":"#fstring-guts"},{"include":"#fstring-illegal-single-brace"},{"include":"#fstring-single-brace"},{"include":"#fstring-single-core"}]},"fstring-raw-guts":{"patterns":[{"include":"#string-consume-escape"},{"include":"#fstring-formatting"}]},"fstring-raw-multi-core":{"match":"(.+?)(($\\\\n?)|(?=[\\\\\\\\\\\\}\\\\{]|\'\'\'|\\"\\"\\"))|\\\\n","name":"string.interpolated.python string.quoted.raw.multi.python"},"fstring-raw-quoted-multi-line":{"begin":"(\\\\b(?:[rR][fF]|[fF][rR]))(\'\'\'|\\"\\"\\")","beginCaptures":{"1":{"name":"string.interpolated.python string.quoted.raw.multi.python storage.type.string.python"},"2":{"name":"punctuation.definition.string.begin.python string.quoted.raw.multi.python"}},"end":"(\\\\2)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python string.interpolated.python string.quoted.raw.multi.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.fstring.python","patterns":[{"include":"#fstring-raw-guts"},{"include":"#fstring-illegal-multi-brace"},{"include":"#fstring-multi-brace"},{"include":"#fstring-raw-multi-core"}]},"fstring-raw-quoted-single-line":{"begin":"(\\\\b(?:[rR][fF]|[fF][rR]))(([\'\\"]))","beginCaptures":{"1":{"name":"string.interpolated.python string.quoted.raw.single.python storage.type.string.python"},"2":{"name":"punctuation.definition.string.begin.python string.quoted.raw.single.python"}},"end":"(\\\\2)|((?<!\\\\\\\\)\\\\n)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python string.interpolated.python string.quoted.raw.single.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.fstring.python","patterns":[{"include":"#fstring-raw-guts"},{"include":"#fstring-illegal-single-brace"},{"include":"#fstring-single-brace"},{"include":"#fstring-raw-single-core"}]},"fstring-raw-single-core":{"match":"(.+?)(($\\\\n?)|(?=[\\\\\\\\\\\\}\\\\{]|([\'\\"])|((?<!\\\\\\\\)\\\\n)))|\\\\n","name":"string.interpolated.python string.quoted.raw.single.python"},"fstring-single-brace":{"begin":"(\\\\{)","beginCaptures":{"1":{"name":"constant.character.format.placeholder.other.python"}},"end":"(\\\\})|(?=\\\\n)","endCaptures":{"1":{"name":"constant.character.format.placeholder.other.python"}},"patterns":[{"include":"#fstring-terminator-single"},{"include":"#f-expression"}]},"fstring-single-core":{"match":"(.+?)(($\\\\n?)|(?=[\\\\\\\\\\\\}\\\\{]|([\'\\"])|((?<!\\\\\\\\)\\\\n)))|\\\\n","name":"string.interpolated.python string.quoted.single.python"},"fstring-terminator-multi":{"patterns":[{"match":"(=(![rsa])?)(?=})","name":"storage.type.format.python"},{"match":"(=?![rsa])(?=})","name":"storage.type.format.python"},{"captures":{"1":{"name":"storage.type.format.python"},"2":{"name":"storage.type.format.python"}},"match":"((?:=?)(?:![rsa])?)(:\\\\w?[<>=^]?[-+ ]?\\\\#?\\\\d*,?(\\\\.\\\\d+)?[bcdeEfFgGnosxX%]?)(?=})"},{"include":"#fstring-terminator-multi-tail"}]},"fstring-terminator-multi-tail":{"begin":"((?:=?)(?:![rsa])?)(:)(?=.*?{)","beginCaptures":{"1":{"name":"storage.type.format.python"},"2":{"name":"storage.type.format.python"}},"end":"(?=})","patterns":[{"include":"#fstring-illegal-multi-brace"},{"include":"#fstring-multi-brace"},{"match":"([bcdeEfFgGnosxX%])(?=})","name":"storage.type.format.python"},{"match":"(\\\\.\\\\d+)","name":"storage.type.format.python"},{"match":"(,)","name":"storage.type.format.python"},{"match":"(\\\\d+)","name":"storage.type.format.python"},{"match":"(\\\\#)","name":"storage.type.format.python"},{"match":"([-+ ])","name":"storage.type.format.python"},{"match":"([<>=^])","name":"storage.type.format.python"},{"match":"(\\\\w)","name":"storage.type.format.python"}]},"fstring-terminator-single":{"patterns":[{"match":"(=(![rsa])?)(?=})","name":"storage.type.format.python"},{"match":"(=?![rsa])(?=})","name":"storage.type.format.python"},{"captures":{"1":{"name":"storage.type.format.python"},"2":{"name":"storage.type.format.python"}},"match":"((?:=?)(?:![rsa])?)(:\\\\w?[<>=^]?[-+ ]?\\\\#?\\\\d*,?(\\\\.\\\\d+)?[bcdeEfFgGnosxX%]?)(?=})"},{"include":"#fstring-terminator-single-tail"}]},"fstring-terminator-single-tail":{"begin":"((?:=?)(?:![rsa])?)(:)(?=.*?{)","beginCaptures":{"1":{"name":"storage.type.format.python"},"2":{"name":"storage.type.format.python"}},"end":"(?=})|(?=\\\\n)","patterns":[{"include":"#fstring-illegal-single-brace"},{"include":"#fstring-single-brace"},{"match":"([bcdeEfFgGnosxX%])(?=})","name":"storage.type.format.python"},{"match":"(\\\\.\\\\d+)","name":"storage.type.format.python"},{"match":"(,)","name":"storage.type.format.python"},{"match":"(\\\\d+)","name":"storage.type.format.python"},{"match":"(\\\\#)","name":"storage.type.format.python"},{"match":"([-+ ])","name":"storage.type.format.python"},{"match":"([<>=^])","name":"storage.type.format.python"},{"match":"(\\\\w)","name":"storage.type.format.python"}]},"function-arguments":{"begin":"(\\\\()","beginCaptures":{"1":{"name":"punctuation.definition.arguments.begin.python"}},"contentName":"meta.function-call.arguments.python","end":"(?=\\\\))(?!\\\\)\\\\s*\\\\()","patterns":[{"match":"(,)","name":"punctuation.separator.arguments.python"},{"captures":{"1":{"name":"keyword.operator.unpacking.arguments.python"}},"match":"(?:(?<=[,(])|^)\\\\s*(\\\\*{1,2})"},{"include":"#lambda-incomplete"},{"include":"#illegal-names"},{"captures":{"1":{"name":"variable.parameter.function-call.python"},"2":{"name":"keyword.operator.assignment.python"}},"match":"\\\\b([[:alpha:]_]\\\\w*)\\\\s*(=)(?!=)"},{"match":"=(?!=)","name":"keyword.operator.assignment.python"},{"include":"#expression"},{"captures":{"1":{"name":"punctuation.definition.arguments.end.python"},"2":{"name":"punctuation.definition.arguments.begin.python"}},"match":"\\\\s*(\\\\))\\\\s*(\\\\()"}]},"function-call":{"begin":"\\\\b(?=([[:alpha:]_]\\\\w*)\\\\s*(\\\\())","end":"(\\\\))","endCaptures":{"1":{"name":"punctuation.definition.arguments.end.python"}},"name":"meta.function-call.python","patterns":[{"include":"#special-variables"},{"include":"#function-name"},{"include":"#function-arguments"}]},"function-declaration":{"begin":"\\\\s*(?:\\\\b(async)\\\\s+)?\\\\b(def)\\\\s+(?=[[:alpha:]_][[:word:]]*\\\\s*\\\\()","beginCaptures":{"1":{"name":"storage.type.function.async.python"},"2":{"name":"storage.type.function.python"}},"end":"(:|(?=[#\'\\"\\\\n]))","endCaptures":{"1":{"name":"punctuation.section.function.begin.python"}},"name":"meta.function.python","patterns":[{"include":"#function-def-name"},{"include":"#parameters"},{"include":"#line-continuation"},{"include":"#return-annotation"}]},"function-def-name":{"patterns":[{"include":"#illegal-object-name"},{"include":"#builtin-possible-callables"},{"match":"\\\\b([[:alpha:]_]\\\\w*)\\\\b","name":"entity.name.function.python"}]},"function-name":{"patterns":[{"include":"#builtin-possible-callables"},{"match":"\\\\b([[:alpha:]_]\\\\w*)\\\\b","name":"meta.function-call.generic.python"}]},"generator":{"begin":"\\\\bfor\\\\b","beginCaptures":{"0":{"name":"keyword.control.flow.python"}},"end":"\\\\bin\\\\b","endCaptures":{"0":{"name":"keyword.control.flow.python"}},"patterns":[{"include":"#expression"}]},"illegal-anno":{"match":"->","name":"invalid.illegal.annotation.python"},"illegal-names":{"captures":{"1":{"name":"keyword.control.flow.python"},"2":{"name":"keyword.control.import.python"}},"match":"\\\\b(?:(and|assert|async|await|break|class|continue|def|del|elif|else|except|finally|for|from|global|if|in|is|(?<=\\\\.)lambda|lambda(?=\\\\s*[\\\\.=])|nonlocal|not|or|pass|raise|return|try|while|with|yield)|(as|import))\\\\b"},"illegal-object-name":{"match":"\\\\b(True|False|None)\\\\b","name":"keyword.illegal.name.python"},"illegal-operator":{"patterns":[{"match":"&&|\\\\|\\\\||--|\\\\+\\\\+","name":"invalid.illegal.operator.python"},{"match":"[?$]","name":"invalid.illegal.operator.python"},{"match":"!\\\\b","name":"invalid.illegal.operator.python"}]},"import":{"patterns":[{"begin":"\\\\b(?<!\\\\.)(from)\\\\b(?=.+import)","beginCaptures":{"1":{"name":"keyword.control.import.python"}},"end":"$|(?=import)","patterns":[{"match":"\\\\.+","name":"punctuation.separator.period.python"},{"include":"#expression"}]},{"begin":"\\\\b(?<!\\\\.)(import)\\\\b","beginCaptures":{"1":{"name":"keyword.control.import.python"}},"end":"$","patterns":[{"match":"\\\\b(?<!\\\\.)as\\\\b","name":"keyword.control.import.python"},{"include":"#expression"}]}]},"impossible":{"match":"$.^"},"inheritance-identifier":{"captures":{"1":{"name":"entity.other.inherited-class.python"}},"match":"\\\\b([[:alpha:]_]\\\\w*)\\\\b"},"inheritance-name":{"patterns":[{"include":"#lambda-incomplete"},{"include":"#builtin-possible-callables"},{"include":"#inheritance-identifier"}]},"item-access":{"patterns":[{"begin":"\\\\b(?=[[:alpha:]_]\\\\w*\\\\s*\\\\[)","end":"(\\\\])","endCaptures":{"1":{"name":"punctuation.definition.arguments.end.python"}},"name":"meta.item-access.python","patterns":[{"include":"#item-name"},{"include":"#item-index"},{"include":"#expression"}]}]},"item-index":{"begin":"(\\\\[)","beginCaptures":{"1":{"name":"punctuation.definition.arguments.begin.python"}},"contentName":"meta.item-access.arguments.python","end":"(?=\\\\])","patterns":[{"match":":","name":"punctuation.separator.slice.python"},{"include":"#expression"}]},"item-name":{"patterns":[{"include":"#special-variables"},{"include":"#builtin-functions"},{"include":"#special-names"},{"match":"\\\\b([[:alpha:]_]\\\\w*)\\\\b","name":"meta.indexed-name.python"}]},"lambda":{"patterns":[{"captures":{"1":{"name":"keyword.control.flow.python"}},"match":"((?<=\\\\.)lambda|lambda(?=\\\\s*[\\\\.=]))"},{"captures":{"1":{"name":"storage.type.function.lambda.python"}},"match":"\\\\b(lambda)\\\\s*?(?=[,\\\\n]|$)"},{"begin":"\\\\b(lambda)\\\\b","beginCaptures":{"1":{"name":"storage.type.function.lambda.python"}},"contentName":"meta.function.lambda.parameters.python","end":"(:)|(\\\\n)","endCaptures":{"1":{"name":"punctuation.section.function.lambda.begin.python"}},"name":"meta.lambda-function.python","patterns":[{"match":"/","name":"keyword.operator.positional.parameter.python"},{"match":"(\\\\*\\\\*|\\\\*)","name":"keyword.operator.unpacking.parameter.python"},{"include":"#lambda-nested-incomplete"},{"include":"#illegal-names"},{"captures":{"1":{"name":"variable.parameter.function.language.python"},"2":{"name":"punctuation.separator.parameters.python"}},"match":"([[:alpha:]_]\\\\w*)\\\\s*(?:(,)|(?=:|$))"},{"include":"#comments"},{"include":"#backticks"},{"include":"#illegal-anno"},{"include":"#lambda-parameter-with-default"},{"include":"#line-continuation"},{"include":"#illegal-operator"}]}]},"lambda-incomplete":{"match":"\\\\blambda(?=\\\\s*[,)])","name":"storage.type.function.lambda.python"},"lambda-nested-incomplete":{"match":"\\\\blambda(?=\\\\s*[:,)])","name":"storage.type.function.lambda.python"},"lambda-parameter-with-default":{"begin":"\\\\b([[:alpha:]_]\\\\w*)\\\\s*(=)","beginCaptures":{"1":{"name":"variable.parameter.function.language.python"},"2":{"name":"keyword.operator.python"}},"end":"(,)|(?=:|$)","endCaptures":{"1":{"name":"punctuation.separator.parameters.python"}},"patterns":[{"include":"#expression"}]},"line-continuation":{"patterns":[{"captures":{"1":{"name":"punctuation.separator.continuation.line.python"},"2":{"name":"invalid.illegal.line.continuation.python"}},"match":"(\\\\\\\\)\\\\s*(\\\\S.*$\\\\n?)"},{"begin":"(\\\\\\\\)\\\\s*$\\\\n?","beginCaptures":{"1":{"name":"punctuation.separator.continuation.line.python"}},"end":"(?=^\\\\s*$)|(?!(\\\\s*[rR]?(\\\\\'\\\\\'\\\\\'|\\\\\\"\\\\\\"\\\\\\"|\\\\\'|\\\\\\"))|(\\\\G$))","patterns":[{"include":"#regexp"},{"include":"#string"}]}]},"list":{"begin":"\\\\[","beginCaptures":{"0":{"name":"punctuation.definition.list.begin.python"}},"end":"\\\\]","endCaptures":{"0":{"name":"punctuation.definition.list.end.python"}},"patterns":[{"include":"#expression"}]},"literal":{"patterns":[{"match":"\\\\b(True|False|None|NotImplemented|Ellipsis)\\\\b","name":"constant.language.python"},{"include":"#number"}]},"loose-default":{"begin":"(=)","beginCaptures":{"1":{"name":"keyword.operator.python"}},"end":"(,)|(?=\\\\))","endCaptures":{"1":{"name":"punctuation.separator.parameters.python"}},"patterns":[{"include":"#expression"}]},"magic-function-names":{"captures":{"1":{"name":"support.function.magic.python"}},"match":"\\\\b(__(?:abs|add|aenter|aexit|aiter|and|anext|await|bool|call|ceil|class_getitem|cmp|coerce|complex|contains|copy|deepcopy|del|delattr|delete|delitem|delslice|dir|div|divmod|enter|eq|exit|float|floor|floordiv|format|ge|get|getattr|getattribute|getinitargs|getitem|getnewargs|getslice|getstate|gt|hash|hex|iadd|iand|idiv|ifloordiv||ilshift|imod|imul|index|init|instancecheck|int|invert|ior|ipow|irshift|isub|iter|itruediv|ixor|le|len|long|lshift|lt|missing|mod|mul|ne|neg|new|next|nonzero|oct|or|pos|pow|radd|rand|rdiv|rdivmod|reduce|reduce_ex|repr|reversed|rfloordiv||rlshift|rmod|rmul|ror|round|rpow|rrshift|rshift|rsub|rtruediv|rxor|set|setattr|setitem|set_name|setslice|setstate|sizeof|str|sub|subclasscheck|truediv|trunc|unicode|xor|matmul|rmatmul|imatmul|init_subclass|set_name|fspath|bytes|prepare|length_hint)__)\\\\b"},"magic-names":{"patterns":[{"include":"#magic-function-names"},{"include":"#magic-variable-names"}]},"magic-variable-names":{"captures":{"1":{"name":"support.variable.magic.python"}},"match":"\\\\b(__(?:all|annotations|bases|builtins|class|closure|code|debug|defaults|dict|doc|file|func|globals|kwdefaults|match_args|members|metaclass|methods|module|mro|mro_entries|name|qualname|post_init|self|signature|slots|subclasses|version|weakref|wrapped|classcell|spec|path|package|future|traceback)__)\\\\b"},"member-access":{"begin":"(\\\\.)\\\\s*(?!\\\\.)","beginCaptures":{"1":{"name":"punctuation.separator.period.python"}},"end":"(?<=\\\\S)(?=\\\\W)|(^|(?<=\\\\s))(?=[^\\\\\\\\\\\\w\\\\s])|$","name":"meta.member.access.python","patterns":[{"include":"#function-call"},{"include":"#member-access-base"},{"include":"#member-access-attribute"}]},"member-access-attribute":{"match":"\\\\b([[:alpha:]_]\\\\w*)\\\\b","name":"meta.attribute.python"},"member-access-base":{"patterns":[{"include":"#magic-names"},{"include":"#illegal-names"},{"include":"#illegal-object-name"},{"include":"#special-names"},{"include":"#line-continuation"},{"include":"#item-access"}]},"member-access-class":{"begin":"(\\\\.)\\\\s*(?!\\\\.)","beginCaptures":{"1":{"name":"punctuation.separator.period.python"}},"end":"(?<=\\\\S)(?=\\\\W)|$","name":"meta.member.access.python","patterns":[{"include":"#call-wrapper-inheritance"},{"include":"#member-access-base"},{"include":"#inheritance-identifier"}]},"number":{"name":"constant.numeric.python","patterns":[{"include":"#number-float"},{"include":"#number-dec"},{"include":"#number-hex"},{"include":"#number-oct"},{"include":"#number-bin"},{"include":"#number-long"},{"match":"\\\\b[0-9]+\\\\w+","name":"invalid.illegal.name.python"}]},"number-bin":{"captures":{"1":{"name":"storage.type.number.python"}},"match":"(?<![\\\\w\\\\.])(0[bB])(_?[01])+\\\\b","name":"constant.numeric.bin.python"},"number-dec":{"captures":{"1":{"name":"storage.type.imaginary.number.python"},"2":{"name":"invalid.illegal.dec.python"}},"match":"(?<![\\\\w\\\\.])(?:[1-9](?:_?[0-9])*|0+|[0-9](?:_?[0-9])*([jJ])|0([0-9]+)(?![eE\\\\.]))\\\\b","name":"constant.numeric.dec.python"},"number-float":{"captures":{"1":{"name":"storage.type.imaginary.number.python"}},"match":"(?<!\\\\w)(?:(?:\\\\.[0-9](?:_?[0-9])*|[0-9](?:_?[0-9])*\\\\.[0-9](?:_?[0-9])*|[0-9](?:_?[0-9])*\\\\.)(?:[eE][+-]?[0-9](?:_?[0-9])*)?|[0-9](?:_?[0-9])*(?:[eE][+-]?[0-9](?:_?[0-9])*))([jJ])?\\\\b","name":"constant.numeric.float.python"},"number-hex":{"captures":{"1":{"name":"storage.type.number.python"}},"match":"(?<![\\\\w\\\\.])(0[xX])(_?[0-9a-fA-F])+\\\\b","name":"constant.numeric.hex.python"},"number-long":{"captures":{"2":{"name":"storage.type.number.python"}},"match":"(?<![\\\\w\\\\.])([1-9][0-9]*|0)([lL])\\\\b","name":"constant.numeric.bin.python"},"number-oct":{"captures":{"1":{"name":"storage.type.number.python"}},"match":"(?<![\\\\w\\\\.])(0[oO])(_?[0-7])+\\\\b","name":"constant.numeric.oct.python"},"odd-function-call":{"begin":"(?<=\\\\]|\\\\))\\\\s*(?=\\\\()","end":"(\\\\))","endCaptures":{"1":{"name":"punctuation.definition.arguments.end.python"}},"patterns":[{"include":"#function-arguments"}]},"operator":{"captures":{"1":{"name":"keyword.operator.logical.python"},"2":{"name":"keyword.control.flow.python"},"3":{"name":"keyword.operator.bitwise.python"},"4":{"name":"keyword.operator.arithmetic.python"},"5":{"name":"keyword.operator.comparison.python"},"6":{"name":"keyword.operator.assignment.python"}},"match":"\\\\b(?<!\\\\.)(?:(and|or|not|in|is)|(for|if|else|await|(?:yield(?:\\\\s+from)?)))(?!\\\\s*:)\\\\b|(<<|>>|&|\\\\||\\\\^|~)|(\\\\*\\\\*|\\\\*|\\\\+|-|%|//|/|@)|(!=|==|>=|<=|<|>)|(:=)"},"parameter-special":{"captures":{"1":{"name":"variable.parameter.function.language.python"},"2":{"name":"variable.parameter.function.language.special.self.python"},"3":{"name":"variable.parameter.function.language.special.cls.python"},"4":{"name":"punctuation.separator.parameters.python"}},"match":"\\\\b((self)|(cls))\\\\b\\\\s*(?:(,)|(?=\\\\)))"},"parameters":{"begin":"(\\\\()","beginCaptures":{"1":{"name":"punctuation.definition.parameters.begin.python"}},"end":"(\\\\))","endCaptures":{"1":{"name":"punctuation.definition.parameters.end.python"}},"name":"meta.function.parameters.python","patterns":[{"match":"/","name":"keyword.operator.positional.parameter.python"},{"match":"(\\\\*\\\\*|\\\\*)","name":"keyword.operator.unpacking.parameter.python"},{"include":"#lambda-incomplete"},{"include":"#illegal-names"},{"include":"#illegal-object-name"},{"include":"#parameter-special"},{"captures":{"1":{"name":"variable.parameter.function.language.python"},"2":{"name":"punctuation.separator.parameters.python"}},"match":"([[:alpha:]_]\\\\w*)\\\\s*(?:(,)|(?=[)#\\\\n=]))"},{"include":"#comments"},{"include":"#loose-default"},{"include":"#annotated-parameter"}]},"punctuation":{"patterns":[{"match":":","name":"punctuation.separator.colon.python"},{"match":",","name":"punctuation.separator.element.python"}]},"regexp":{"patterns":[{"include":"#regexp-single-three-line"},{"include":"#regexp-double-three-line"},{"include":"#regexp-single-one-line"},{"include":"#regexp-double-one-line"}]},"regexp-backreference":{"captures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.backreference.named.begin.regexp"},"2":{"name":"entity.name.tag.named.backreference.regexp"},"3":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.backreference.named.end.regexp"}},"match":"(\\\\()(\\\\?P=\\\\w+(?:\\\\s+[[:alnum:]]+)?)(\\\\))","name":"meta.backreference.named.regexp"},"regexp-backreference-number":{"captures":{"1":{"name":"entity.name.tag.backreference.regexp"}},"match":"(\\\\\\\\[1-9]\\\\d?)","name":"meta.backreference.regexp"},"regexp-base-common":{"patterns":[{"match":"\\\\.","name":"support.other.match.any.regexp"},{"match":"\\\\^","name":"support.other.match.begin.regexp"},{"match":"\\\\$","name":"support.other.match.end.regexp"},{"match":"[+*?]\\\\??","name":"keyword.operator.quantifier.regexp"},{"match":"\\\\|","name":"keyword.operator.disjunction.regexp"},{"include":"#regexp-escape-sequence"}]},"regexp-base-expression":{"patterns":[{"include":"#regexp-quantifier"},{"include":"#regexp-base-common"}]},"regexp-charecter-set-escapes":{"patterns":[{"match":"\\\\\\\\[abfnrtv\\\\\\\\]","name":"constant.character.escape.regexp"},{"include":"#regexp-escape-special"},{"match":"\\\\\\\\([0-7]{1,3})","name":"constant.character.escape.regexp"},{"include":"#regexp-escape-character"},{"include":"#regexp-escape-unicode"},{"include":"#regexp-escape-catchall"}]},"regexp-double-one-line":{"begin":"\\\\b(([uU]r)|([bB]r)|(r[bB]?))(\\")","beginCaptures":{"2":{"name":"invalid.deprecated.prefix.python"},"3":{"name":"storage.type.string.python"},"4":{"name":"storage.type.string.python"},"5":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\")|(?<!\\\\\\\\)(\\\\n)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.regexp.quoted.single.python","patterns":[{"include":"#double-one-regexp-expression"}]},"regexp-double-three-line":{"begin":"\\\\b(([uU]r)|([bB]r)|(r[bB]?))(\\"\\"\\")","beginCaptures":{"2":{"name":"invalid.deprecated.prefix.python"},"3":{"name":"storage.type.string.python"},"4":{"name":"storage.type.string.python"},"5":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\"\\"\\")","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.regexp.quoted.multi.python","patterns":[{"include":"#double-three-regexp-expression"}]},"regexp-escape-catchall":{"match":"\\\\\\\\(.|\\\\n)","name":"constant.character.escape.regexp"},"regexp-escape-character":{"match":"\\\\\\\\(x[0-9A-Fa-f]{2}|0[0-7]{1,2}|[0-7]{3})","name":"constant.character.escape.regexp"},"regexp-escape-sequence":{"patterns":[{"include":"#regexp-escape-special"},{"include":"#regexp-escape-character"},{"include":"#regexp-escape-unicode"},{"include":"#regexp-backreference-number"},{"include":"#regexp-escape-catchall"}]},"regexp-escape-special":{"match":"\\\\\\\\([AbBdDsSwWZ])","name":"support.other.escape.special.regexp"},"regexp-escape-unicode":{"match":"\\\\\\\\(u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})","name":"constant.character.unicode.regexp"},"regexp-flags":{"match":"\\\\(\\\\?[aiLmsux]+\\\\)","name":"storage.modifier.flag.regexp"},"regexp-quantifier":{"match":"\\\\{(\\\\d+|\\\\d+,(\\\\d+)?|,\\\\d+)\\\\}","name":"keyword.operator.quantifier.regexp"},"regexp-single-one-line":{"begin":"\\\\b(([uU]r)|([bB]r)|(r[bB]?))(\\\\\')","beginCaptures":{"2":{"name":"invalid.deprecated.prefix.python"},"3":{"name":"storage.type.string.python"},"4":{"name":"storage.type.string.python"},"5":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\\')|(?<!\\\\\\\\)(\\\\n)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.regexp.quoted.single.python","patterns":[{"include":"#single-one-regexp-expression"}]},"regexp-single-three-line":{"begin":"\\\\b(([uU]r)|([bB]r)|(r[bB]?))(\\\\\'\\\\\'\\\\\')","beginCaptures":{"2":{"name":"invalid.deprecated.prefix.python"},"3":{"name":"storage.type.string.python"},"4":{"name":"storage.type.string.python"},"5":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\\'\\\\\'\\\\\')","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.regexp.quoted.multi.python","patterns":[{"include":"#single-three-regexp-expression"}]},"return-annotation":{"begin":"(->)","beginCaptures":{"1":{"name":"punctuation.separator.annotation.result.python"}},"end":"(?=:)","patterns":[{"include":"#expression"}]},"round-braces":{"begin":"\\\\(","beginCaptures":{"0":{"name":"punctuation.parenthesis.begin.python"}},"end":"\\\\)","endCaptures":{"0":{"name":"punctuation.parenthesis.end.python"}},"patterns":[{"include":"#expression"}]},"semicolon":{"patterns":[{"match":"\\\\;$","name":"invalid.deprecated.semicolon.python"}]},"single-one-regexp-character-set":{"patterns":[{"match":"\\\\[\\\\^?\\\\](?!.*?\\\\])"},{"begin":"(\\\\[)(\\\\^)?(\\\\])?","beginCaptures":{"1":{"name":"punctuation.character.set.begin.regexp constant.other.set.regexp"},"2":{"name":"keyword.operator.negation.regexp"},"3":{"name":"constant.character.set.regexp"}},"end":"(\\\\]|(?=\\\\\'))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"punctuation.character.set.end.regexp constant.other.set.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.character.set.regexp","patterns":[{"include":"#regexp-charecter-set-escapes"},{"match":"[^\\\\n]","name":"constant.character.set.regexp"}]}]},"single-one-regexp-comments":{"begin":"\\\\(\\\\?#","beginCaptures":{"0":{"name":"punctuation.comment.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"punctuation.comment.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"name":"comment.regexp","patterns":[{"include":"#codetags"}]},"single-one-regexp-conditional":{"begin":"(\\\\()\\\\?\\\\((\\\\w+(?:\\\\s+[[:alnum:]]+)?|\\\\d+)\\\\)","beginCaptures":{"0":{"name":"keyword.operator.conditional.regexp"},"1":{"name":"punctuation.parenthesis.conditional.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"keyword.operator.conditional.negative.regexp punctuation.parenthesis.conditional.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-one-regexp-expression"}]},"single-one-regexp-expression":{"patterns":[{"include":"#regexp-base-expression"},{"include":"#single-one-regexp-character-set"},{"include":"#single-one-regexp-comments"},{"include":"#regexp-flags"},{"include":"#single-one-regexp-named-group"},{"include":"#regexp-backreference"},{"include":"#single-one-regexp-lookahead"},{"include":"#single-one-regexp-lookahead-negative"},{"include":"#single-one-regexp-lookbehind"},{"include":"#single-one-regexp-lookbehind-negative"},{"include":"#single-one-regexp-conditional"},{"include":"#single-one-regexp-parentheses-non-capturing"},{"include":"#single-one-regexp-parentheses"}]},"single-one-regexp-lookahead":{"begin":"(\\\\()\\\\?=","beginCaptures":{"0":{"name":"keyword.operator.lookahead.regexp"},"1":{"name":"punctuation.parenthesis.lookahead.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"keyword.operator.lookahead.regexp punctuation.parenthesis.lookahead.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-one-regexp-expression"}]},"single-one-regexp-lookahead-negative":{"begin":"(\\\\()\\\\?!","beginCaptures":{"0":{"name":"keyword.operator.lookahead.negative.regexp"},"1":{"name":"punctuation.parenthesis.lookahead.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"keyword.operator.lookahead.negative.regexp punctuation.parenthesis.lookahead.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-one-regexp-expression"}]},"single-one-regexp-lookbehind":{"begin":"(\\\\()\\\\?<=","beginCaptures":{"0":{"name":"keyword.operator.lookbehind.regexp"},"1":{"name":"punctuation.parenthesis.lookbehind.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"keyword.operator.lookbehind.regexp punctuation.parenthesis.lookbehind.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-one-regexp-expression"}]},"single-one-regexp-lookbehind-negative":{"begin":"(\\\\()\\\\?<!","beginCaptures":{"0":{"name":"keyword.operator.lookbehind.negative.regexp"},"1":{"name":"punctuation.parenthesis.lookbehind.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"keyword.operator.lookbehind.negative.regexp punctuation.parenthesis.lookbehind.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-one-regexp-expression"}]},"single-one-regexp-named-group":{"begin":"(\\\\()(\\\\?P<\\\\w+(?:\\\\s+[[:alnum:]]+)?>)","beginCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.named.begin.regexp"},"2":{"name":"entity.name.tag.named.group.regexp"}},"end":"(\\\\)|(?=\\\\\'))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.named.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.named.regexp","patterns":[{"include":"#single-one-regexp-expression"}]},"single-one-regexp-parentheses":{"begin":"\\\\(","beginCaptures":{"0":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-one-regexp-expression"}]},"single-one-regexp-parentheses-non-capturing":{"begin":"\\\\(\\\\?:","beginCaptures":{"0":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.non-capturing.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'))|((?=(?<!\\\\\\\\)\\\\n))","endCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.non-capturing.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-one-regexp-expression"}]},"single-three-regexp-character-set":{"patterns":[{"match":"\\\\[\\\\^?\\\\](?!.*?\\\\])"},{"begin":"(\\\\[)(\\\\^)?(\\\\])?","beginCaptures":{"1":{"name":"punctuation.character.set.begin.regexp constant.other.set.regexp"},"2":{"name":"keyword.operator.negation.regexp"},"3":{"name":"constant.character.set.regexp"}},"end":"(\\\\]|(?=\\\\\'\\\\\'\\\\\'))","endCaptures":{"1":{"name":"punctuation.character.set.end.regexp constant.other.set.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.character.set.regexp","patterns":[{"include":"#regexp-charecter-set-escapes"},{"match":"[^\\\\n]","name":"constant.character.set.regexp"}]}]},"single-three-regexp-comments":{"begin":"\\\\(\\\\?#","beginCaptures":{"0":{"name":"punctuation.comment.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'\\\\\'\\\\\'))","endCaptures":{"1":{"name":"punctuation.comment.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"name":"comment.regexp","patterns":[{"include":"#codetags"}]},"single-three-regexp-conditional":{"begin":"(\\\\()\\\\?\\\\((\\\\w+(?:\\\\s+[[:alnum:]]+)?|\\\\d+)\\\\)","beginCaptures":{"0":{"name":"keyword.operator.conditional.regexp"},"1":{"name":"punctuation.parenthesis.conditional.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'\\\\\'\\\\\'))","endCaptures":{"1":{"name":"keyword.operator.conditional.negative.regexp punctuation.parenthesis.conditional.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-three-regexp-expression"},{"include":"#comments-string-single-three"}]},"single-three-regexp-expression":{"patterns":[{"include":"#regexp-base-expression"},{"include":"#single-three-regexp-character-set"},{"include":"#single-three-regexp-comments"},{"include":"#regexp-flags"},{"include":"#single-three-regexp-named-group"},{"include":"#regexp-backreference"},{"include":"#single-three-regexp-lookahead"},{"include":"#single-three-regexp-lookahead-negative"},{"include":"#single-three-regexp-lookbehind"},{"include":"#single-three-regexp-lookbehind-negative"},{"include":"#single-three-regexp-conditional"},{"include":"#single-three-regexp-parentheses-non-capturing"},{"include":"#single-three-regexp-parentheses"},{"include":"#comments-string-single-three"}]},"single-three-regexp-lookahead":{"begin":"(\\\\()\\\\?=","beginCaptures":{"0":{"name":"keyword.operator.lookahead.regexp"},"1":{"name":"punctuation.parenthesis.lookahead.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'\\\\\'\\\\\'))","endCaptures":{"1":{"name":"keyword.operator.lookahead.regexp punctuation.parenthesis.lookahead.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-three-regexp-expression"},{"include":"#comments-string-single-three"}]},"single-three-regexp-lookahead-negative":{"begin":"(\\\\()\\\\?!","beginCaptures":{"0":{"name":"keyword.operator.lookahead.negative.regexp"},"1":{"name":"punctuation.parenthesis.lookahead.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'\\\\\'\\\\\'))","endCaptures":{"1":{"name":"keyword.operator.lookahead.negative.regexp punctuation.parenthesis.lookahead.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-three-regexp-expression"},{"include":"#comments-string-single-three"}]},"single-three-regexp-lookbehind":{"begin":"(\\\\()\\\\?<=","beginCaptures":{"0":{"name":"keyword.operator.lookbehind.regexp"},"1":{"name":"punctuation.parenthesis.lookbehind.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'\\\\\'\\\\\'))","endCaptures":{"1":{"name":"keyword.operator.lookbehind.regexp punctuation.parenthesis.lookbehind.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-three-regexp-expression"},{"include":"#comments-string-single-three"}]},"single-three-regexp-lookbehind-negative":{"begin":"(\\\\()\\\\?<!","beginCaptures":{"0":{"name":"keyword.operator.lookbehind.negative.regexp"},"1":{"name":"punctuation.parenthesis.lookbehind.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'\\\\\'\\\\\'))","endCaptures":{"1":{"name":"keyword.operator.lookbehind.negative.regexp punctuation.parenthesis.lookbehind.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-three-regexp-expression"},{"include":"#comments-string-single-three"}]},"single-three-regexp-named-group":{"begin":"(\\\\()(\\\\?P<\\\\w+(?:\\\\s+[[:alnum:]]+)?>)","beginCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.named.begin.regexp"},"2":{"name":"entity.name.tag.named.group.regexp"}},"end":"(\\\\)|(?=\\\\\'\\\\\'\\\\\'))","endCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.named.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"name":"meta.named.regexp","patterns":[{"include":"#single-three-regexp-expression"},{"include":"#comments-string-single-three"}]},"single-three-regexp-parentheses":{"begin":"\\\\(","beginCaptures":{"0":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'\\\\\'\\\\\'))","endCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-three-regexp-expression"},{"include":"#comments-string-single-three"}]},"single-three-regexp-parentheses-non-capturing":{"begin":"\\\\(\\\\?:","beginCaptures":{"0":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.non-capturing.begin.regexp"}},"end":"(\\\\)|(?=\\\\\'\\\\\'\\\\\'))","endCaptures":{"1":{"name":"support.other.parenthesis.regexp punctuation.parenthesis.non-capturing.end.regexp"},"2":{"name":"invalid.illegal.newline.python"}},"patterns":[{"include":"#single-three-regexp-expression"},{"include":"#comments-string-single-three"}]},"special-names":{"match":"\\\\b(_*[[:upper:]][_\\\\d]*[[:upper:]])[[:upper:]\\\\d]*(_\\\\w*)?\\\\b","name":"constant.other.caps.python"},"special-variables":{"captures":{"1":{"name":"variable.language.special.self.python"},"2":{"name":"variable.language.special.cls.python"}},"match":"\\\\b(?<!\\\\.)(?:(self)|(cls))\\\\b"},"statement":{"patterns":[{"include":"#import"},{"include":"#class-declaration"},{"include":"#function-declaration"},{"include":"#generator"},{"include":"#statement-keyword"},{"include":"#assignment-operator"},{"include":"#decorator"},{"include":"#docstring-statement"},{"include":"#semicolon"}]},"statement-keyword":{"patterns":[{"match":"\\\\b((async\\\\s+)?\\\\s*def)\\\\b","name":"storage.type.function.python"},{"match":"\\\\b(?<!\\\\.)as\\\\b(?=.*[:\\\\\\\\])","name":"keyword.control.flow.python"},{"match":"\\\\b(?<!\\\\.)as\\\\b","name":"keyword.control.import.python"},{"match":"\\\\b(?<!\\\\.)(async|continue|del|assert|break|finally|for|from|elif|else|if|except|pass|raise|return|try|while|with)\\\\b","name":"keyword.control.flow.python"},{"match":"\\\\b(?<!\\\\.)(global|nonlocal)\\\\b","name":"storage.modifier.declaration.python"},{"match":"\\\\b(?<!\\\\.)(class)\\\\b","name":"storage.type.class.python"},{"captures":{"1":{"name":"keyword.control.flow.python"}},"match":"^\\\\s*(case|match)(?=\\\\s*([-+\\\\w\\\\d(\\\\[{\'\\":#]|$))\\\\b"}]},"string":{"patterns":[{"include":"#string-quoted-multi-line"},{"include":"#string-quoted-single-line"},{"include":"#string-bin-quoted-multi-line"},{"include":"#string-bin-quoted-single-line"},{"include":"#string-raw-quoted-multi-line"},{"include":"#string-raw-quoted-single-line"},{"include":"#string-raw-bin-quoted-multi-line"},{"include":"#string-raw-bin-quoted-single-line"},{"include":"#fstring-fnorm-quoted-multi-line"},{"include":"#fstring-fnorm-quoted-single-line"},{"include":"#fstring-normf-quoted-multi-line"},{"include":"#fstring-normf-quoted-single-line"},{"include":"#fstring-raw-quoted-multi-line"},{"include":"#fstring-raw-quoted-single-line"}]},"string-bin-quoted-multi-line":{"begin":"(\\\\b[bB])(\'\'\'|\\"\\"\\")","beginCaptures":{"1":{"name":"storage.type.string.python"},"2":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\2)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.quoted.binary.multi.python","patterns":[{"include":"#string-entity"}]},"string-bin-quoted-single-line":{"begin":"(\\\\b[bB])(([\'\\"]))","beginCaptures":{"1":{"name":"storage.type.string.python"},"2":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\2)|((?<!\\\\\\\\)\\\\n)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.quoted.binary.single.python","patterns":[{"include":"#string-entity"}]},"string-brace-formatting":{"patterns":[{"captures":{"1":{"name":"constant.character.format.placeholder.other.python"},"3":{"name":"storage.type.format.python"},"4":{"name":"storage.type.format.python"}},"match":"({{|}}|(?:{\\\\w*(\\\\.[[:alpha:]_]\\\\w*|\\\\[[^\\\\]\'\\"]+\\\\])*(![rsa])?(:\\\\w?[<>=^]?[-+ ]?\\\\#?\\\\d*,?(\\\\.\\\\d+)?[bcdeEfFgGnosxX%]?)?}))","name":"meta.format.brace.python"},{"captures":{"1":{"name":"constant.character.format.placeholder.other.python"},"3":{"name":"storage.type.format.python"},"4":{"name":"storage.type.format.python"}},"match":"({\\\\w*(\\\\.[[:alpha:]_]\\\\w*|\\\\[[^\\\\]\'\\"]+\\\\])*(![rsa])?(:)[^\'\\"{}\\\\n]*(?:\\\\{[^\'\\"}\\\\n]*?\\\\}[^\'\\"{}\\\\n]*)*})","name":"meta.format.brace.python"}]},"string-consume-escape":{"match":"\\\\\\\\[\'\\"\\\\n\\\\\\\\]"},"string-entity":{"patterns":[{"include":"#escape-sequence"},{"include":"#string-line-continuation"},{"include":"#string-formatting"}]},"string-formatting":{"captures":{"1":{"name":"constant.character.format.placeholder.other.python"}},"match":"(%(\\\\([\\\\w\\\\s]*\\\\))?[-+#0 ]*(\\\\d+|\\\\*)?(\\\\.(\\\\d+|\\\\*))?([hlL])?[diouxXeEfFgGcrsab%])","name":"meta.format.percent.python"},"string-line-continuation":{"match":"\\\\\\\\$","name":"constant.language.python"},"string-multi-bad-brace1-formatting-raw":{"begin":"(?=\\\\{%(.*?(?!\'\'\'|\\"\\"\\"))%\\\\})","end":"(?=\'\'\'|\\"\\"\\")","patterns":[{"include":"#string-consume-escape"}]},"string-multi-bad-brace1-formatting-unicode":{"begin":"(?=\\\\{%(.*?(?!\'\'\'|\\"\\"\\"))%\\\\})","end":"(?=\'\'\'|\\"\\"\\")","patterns":[{"include":"#escape-sequence-unicode"},{"include":"#escape-sequence"},{"include":"#string-line-continuation"}]},"string-multi-bad-brace2-formatting-raw":{"begin":"(?!\\\\{\\\\{)(?=\\\\{(\\\\w*?(?!\'\'\'|\\"\\"\\")[^!:\\\\.\\\\[}\\\\w]).*?(?!\'\'\'|\\"\\"\\")\\\\})","end":"(?=\'\'\'|\\"\\"\\")","patterns":[{"include":"#string-consume-escape"},{"include":"#string-formatting"}]},"string-multi-bad-brace2-formatting-unicode":{"begin":"(?!\\\\{\\\\{)(?=\\\\{(\\\\w*?(?!\'\'\'|\\"\\"\\")[^!:\\\\.\\\\[}\\\\w]).*?(?!\'\'\'|\\"\\"\\")\\\\})","end":"(?=\'\'\'|\\"\\"\\")","patterns":[{"include":"#escape-sequence-unicode"},{"include":"#string-entity"}]},"string-quoted-multi-line":{"begin":"(?:\\\\b([rR])(?=[uU]))?([uU])?(\'\'\'|\\"\\"\\")","beginCaptures":{"1":{"name":"invalid.illegal.prefix.python"},"2":{"name":"storage.type.string.python"},"3":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\3)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.quoted.multi.python","patterns":[{"include":"#string-multi-bad-brace1-formatting-unicode"},{"include":"#string-multi-bad-brace2-formatting-unicode"},{"include":"#string-unicode-guts"}]},"string-quoted-single-line":{"begin":"(?:\\\\b([rR])(?=[uU]))?([uU])?(([\'\\"]))","beginCaptures":{"1":{"name":"invalid.illegal.prefix.python"},"2":{"name":"storage.type.string.python"},"3":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\3)|((?<!\\\\\\\\)\\\\n)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.quoted.single.python","patterns":[{"include":"#string-single-bad-brace1-formatting-unicode"},{"include":"#string-single-bad-brace2-formatting-unicode"},{"include":"#string-unicode-guts"}]},"string-raw-bin-guts":{"patterns":[{"include":"#string-consume-escape"},{"include":"#string-formatting"}]},"string-raw-bin-quoted-multi-line":{"begin":"(\\\\b(?:R[bB]|[bB]R))(\'\'\'|\\"\\"\\")","beginCaptures":{"1":{"name":"storage.type.string.python"},"2":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\2)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.quoted.raw.binary.multi.python","patterns":[{"include":"#string-raw-bin-guts"}]},"string-raw-bin-quoted-single-line":{"begin":"(\\\\b(?:R[bB]|[bB]R))(([\'\\"]))","beginCaptures":{"1":{"name":"storage.type.string.python"},"2":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\2)|((?<!\\\\\\\\)\\\\n)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.quoted.raw.binary.single.python","patterns":[{"include":"#string-raw-bin-guts"}]},"string-raw-guts":{"patterns":[{"include":"#string-consume-escape"},{"include":"#string-formatting"},{"include":"#string-brace-formatting"}]},"string-raw-quoted-multi-line":{"begin":"\\\\b(([uU]R)|(R))(\'\'\'|\\"\\"\\")","beginCaptures":{"2":{"name":"invalid.deprecated.prefix.python"},"3":{"name":"storage.type.string.python"},"4":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\4)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.quoted.raw.multi.python","patterns":[{"include":"#string-multi-bad-brace1-formatting-raw"},{"include":"#string-multi-bad-brace2-formatting-raw"},{"include":"#string-raw-guts"}]},"string-raw-quoted-single-line":{"begin":"\\\\b(([uU]R)|(R))(([\'\\"]))","beginCaptures":{"2":{"name":"invalid.deprecated.prefix.python"},"3":{"name":"storage.type.string.python"},"4":{"name":"punctuation.definition.string.begin.python"}},"end":"(\\\\4)|((?<!\\\\\\\\)\\\\n)","endCaptures":{"1":{"name":"punctuation.definition.string.end.python"},"2":{"name":"invalid.illegal.newline.python"}},"name":"string.quoted.raw.single.python","patterns":[{"include":"#string-single-bad-brace1-formatting-raw"},{"include":"#string-single-bad-brace2-formatting-raw"},{"include":"#string-raw-guts"}]},"string-single-bad-brace1-formatting-raw":{"begin":"(?=\\\\{%(.*?(?!([\'\\"])|((?<!\\\\\\\\)\\\\n)))%\\\\})","end":"(?=([\'\\"])|((?<!\\\\\\\\)\\\\n))","patterns":[{"include":"#string-consume-escape"}]},"string-single-bad-brace1-formatting-unicode":{"begin":"(?=\\\\{%(.*?(?!([\'\\"])|((?<!\\\\\\\\)\\\\n)))%\\\\})","end":"(?=([\'\\"])|((?<!\\\\\\\\)\\\\n))","patterns":[{"include":"#escape-sequence-unicode"},{"include":"#escape-sequence"},{"include":"#string-line-continuation"}]},"string-single-bad-brace2-formatting-raw":{"begin":"(?!\\\\{\\\\{)(?=\\\\{(\\\\w*?(?!([\'\\"])|((?<!\\\\\\\\)\\\\n))[^!:\\\\.\\\\[}\\\\w]).*?(?!([\'\\"])|((?<!\\\\\\\\)\\\\n))\\\\})","end":"(?=([\'\\"])|((?<!\\\\\\\\)\\\\n))","patterns":[{"include":"#string-consume-escape"},{"include":"#string-formatting"}]},"string-single-bad-brace2-formatting-unicode":{"begin":"(?!\\\\{\\\\{)(?=\\\\{(\\\\w*?(?!([\'\\"])|((?<!\\\\\\\\)\\\\n))[^!:\\\\.\\\\[}\\\\w]).*?(?!([\'\\"])|((?<!\\\\\\\\)\\\\n))\\\\})","end":"(?=([\'\\"])|((?<!\\\\\\\\)\\\\n))","patterns":[{"include":"#escape-sequence-unicode"},{"include":"#string-entity"}]},"string-unicode-guts":{"patterns":[{"include":"#escape-sequence-unicode"},{"include":"#string-entity"},{"include":"#string-brace-formatting"}]}},"scopeName":"source.python","aliases":["py"]}'))];class q extends Error{constructor(e){super(e),this.name="ShikiError"}}function G(e){return Array.isArray(e)?function(e){let t=[];for(let n=0,r=e.length;n<r;n++)t[n]=G(e[n]);return t}(e):e instanceof RegExp?e:"object"==typeof e?function(e){let t={};for(let n in e)t[n]=G(e[n]);return t}(e):e}function W(e,...t){return t.forEach((t=>{for(let n in t)e[n]=t[n]})),e}function H(e){const t=~e.lastIndexOf("/")||~e.lastIndexOf("\\");return 0===t?e:~t==e.length-1?H(e.substring(0,e.length-1)):e.substr(1+~t)}var V=/\$(\d+)|\${(\d+):\/(downcase|upcase)}/g,Q=class{static hasCaptures(e){return null!==e&&(V.lastIndex=0,V.test(e))}static replaceCaptures(e,t,n){return e.replace(V,((e,r,a,o)=>{let i=n[parseInt(r||a,10)];if(!i)return e;{let e=t.substring(i.start,i.end);for(;"."===e[0];)e=e.substring(1);switch(o){case"downcase":return e.toLowerCase();case"upcase":return e.toUpperCase();default:return e}}}))}};function X(e,t){return e<t?-1:e>t?1:0}function K(e,t){if(null===e&&null===t)return 0;if(!e)return-1;if(!t)return 1;let n=e.length,r=t.length;if(n===r){for(let r=0;r<n;r++){let n=X(e[r],t[r]);if(0!==n)return n}return 0}return n-r}function Y(e){return!!(/^#[0-9a-f]{6}$/i.test(e)||/^#[0-9a-f]{8}$/i.test(e)||/^#[0-9a-f]{3}$/i.test(e)||/^#[0-9a-f]{4}$/i.test(e))}function Z(e){return e.replace(/[\-\\\{\}\*\+\?\|\^\$\.\,\[\]\(\)\#\s]/g,"\\$&")}var J=class{constructor(e){this.fn=e}cache=new Map;get(e){if(this.cache.has(e))return this.cache.get(e);const t=this.fn(e);return this.cache.set(e,t),t}},ee=class{constructor(e,t,n){this._colorMap=e,this._defaults=t,this._root=n}static createFromRawTheme(e,t){return this.createFromParsedTheme(function(e){if(!e)return[];if(!e.settings||!Array.isArray(e.settings))return[];let t=e.settings,n=[],r=0;for(let e=0,a=t.length;e<a;e++){let a,o=t[e];if(!o.settings)continue;if("string"==typeof o.scope){let e=o.scope;e=e.replace(/^[,]+/,""),e=e.replace(/[,]+$/,""),a=e.split(",")}else a=Array.isArray(o.scope)?o.scope:[""];let i=-1;if("string"==typeof o.settings.fontStyle){i=0;let e=o.settings.fontStyle.split(" ");for(let t=0,n=e.length;t<n;t++)switch(e[t]){case"italic":i|=1;break;case"bold":i|=2;break;case"underline":i|=4;break;case"strikethrough":i|=8}}let l=null;"string"==typeof o.settings.foreground&&Y(o.settings.foreground)&&(l=o.settings.foreground);let s=null;"string"==typeof o.settings.background&&Y(o.settings.background)&&(s=o.settings.background);for(let t=0,o=a.length;t<o;t++){let o=a[t].trim().split(" "),u=o[o.length-1],c=null;o.length>1&&(c=o.slice(0,o.length-1),c.reverse()),n[r++]=new ae(u,c,e,i,l,s)}}return n}(e),t)}static createFromParsedTheme(e,t){return function(e,t){e.sort(((e,t)=>{let n=X(e.scope,t.scope);return 0!==n?n:(n=K(e.parentScopes,t.parentScopes),0!==n?n:e.index-t.index)}));let n=0,r="#000000",a="#ffffff";for(;e.length>=1&&""===e[0].scope;){let t=e.shift();-1!==t.fontStyle&&(n=t.fontStyle),null!==t.foreground&&(r=t.foreground),null!==t.background&&(a=t.background)}let o=new ie(t),i=new re(n,o.getId(r),o.getId(a)),l=new ue(new se(0,null,-1,0,0),[]);for(let t=0,n=e.length;t<n;t++){let n=e[t];l.insert(0,n.scope,n.parentScopes,n.fontStyle,o.getId(n.foreground),o.getId(n.background))}return new ee(o,i,l)}(e,t)}_cachedMatchRoot=new J((e=>this._root.match(e)));getColorMap(){return this._colorMap.getColorMap()}getDefaults(){return this._defaults}match(e){if(null===e)return this._defaults;const t=e.scopeName,n=this._cachedMatchRoot.get(t).find((t=>function(e,t){if(0===t.length)return!0;for(let n=0;n<t.length;n++){let r=t[n],a=!1;if(">"===r){if(n===t.length-1)return!1;r=t[++n],a=!0}for(;e&&!ne(e.scopeName,r);){if(a)return!1;e=e.parent}if(!e)return!1;e=e.parent}return!0}(e.parent,t.parentScopes)));return n?new re(n.fontStyle,n.foreground,n.background):null}},te=class e{constructor(e,t){this.parent=e,this.scopeName=t}static push(t,n){for(const r of n)t=new e(t,r);return t}static from(...t){let n=null;for(let r=0;r<t.length;r++)n=new e(n,t[r]);return n}push(t){return new e(this,t)}getSegments(){let e=this;const t=[];for(;e;)t.push(e.scopeName),e=e.parent;return t.reverse(),t}toString(){return this.getSegments().join(" ")}extends(e){return this===e||null!==this.parent&&this.parent.extends(e)}getExtensionIfDefined(e){const t=[];let n=this;for(;n&&n!==e;)t.push(n.scopeName),n=n.parent;return n===e?t.reverse():void 0}};function ne(e,t){return t===e||e.startsWith(t)&&"."===e[t.length]}var re=class{constructor(e,t,n){this.fontStyle=e,this.foregroundId=t,this.backgroundId=n}},ae=class{constructor(e,t,n,r,a,o){this.scope=e,this.parentScopes=t,this.index=n,this.fontStyle=r,this.foreground=a,this.background=o}},oe=(e=>(e[e.NotSet=-1]="NotSet",e[e.None=0]="None",e[e.Italic=1]="Italic",e[e.Bold=2]="Bold",e[e.Underline=4]="Underline",e[e.Strikethrough=8]="Strikethrough",e))(oe||{}),ie=class{_isFrozen;_lastColorId;_id2color;_color2id;constructor(e){if(this._lastColorId=0,this._id2color=[],this._color2id=Object.create(null),Array.isArray(e)){this._isFrozen=!0;for(let t=0,n=e.length;t<n;t++)this._color2id[e[t]]=t,this._id2color[t]=e[t]}else this._isFrozen=!1}getId(e){if(null===e)return 0;e=e.toUpperCase();let t=this._color2id[e];if(t)return t;if(this._isFrozen)throw new Error(`Missing color in color map - ${e}`);return t=++this._lastColorId,this._color2id[e]=t,this._id2color[t]=e,t}getColorMap(){return this._id2color.slice(0)}},le=Object.freeze([]),se=class e{scopeDepth;parentScopes;fontStyle;foreground;background;constructor(e,t,n,r,a){this.scopeDepth=e,this.parentScopes=t||le,this.fontStyle=n,this.foreground=r,this.background=a}clone(){return new e(this.scopeDepth,this.parentScopes,this.fontStyle,this.foreground,this.background)}static cloneArr(e){let t=[];for(let n=0,r=e.length;n<r;n++)t[n]=e[n].clone();return t}acceptOverwrite(e,t,n,r){this.scopeDepth>e?console.log("how did this happen?"):this.scopeDepth=e,-1!==t&&(this.fontStyle=t),0!==n&&(this.foreground=n),0!==r&&(this.background=r)}},ue=class e{constructor(e,t=[],n={}){this._mainRule=e,this._children=n,this._rulesWithParentScopes=t}_rulesWithParentScopes;static _cmpBySpecificity(e,t){if(e.scopeDepth!==t.scopeDepth)return t.scopeDepth-e.scopeDepth;let n=0,r=0;for(;">"===e.parentScopes[n]&&n++,">"===t.parentScopes[r]&&r++,!(n>=e.parentScopes.length||r>=t.parentScopes.length);){const a=t.parentScopes[r].length-e.parentScopes[n].length;if(0!==a)return a;n++,r++}return t.parentScopes.length-e.parentScopes.length}match(t){if(""!==t){let e,n,r=t.indexOf(".");if(-1===r?(e=t,n=""):(e=t.substring(0,r),n=t.substring(r+1)),this._children.hasOwnProperty(e))return this._children[e].match(n)}const n=this._rulesWithParentScopes.concat(this._mainRule);return n.sort(e._cmpBySpecificity),n}insert(t,n,r,a,o,i){if(""===n)return void this._doInsertHere(t,r,a,o,i);let l,s,u,c=n.indexOf(".");-1===c?(l=n,s=""):(l=n.substring(0,c),s=n.substring(c+1)),this._children.hasOwnProperty(l)?u=this._children[l]:(u=new e(this._mainRule.clone(),se.cloneArr(this._rulesWithParentScopes)),this._children[l]=u),u.insert(t+1,s,r,a,o,i)}_doInsertHere(e,t,n,r,a){if(null!==t){for(let o=0,i=this._rulesWithParentScopes.length;o<i;o++){let i=this._rulesWithParentScopes[o];if(0===K(i.parentScopes,t))return void i.acceptOverwrite(e,n,r,a)}-1===n&&(n=this._mainRule.fontStyle),0===r&&(r=this._mainRule.foreground),0===a&&(a=this._mainRule.background),this._rulesWithParentScopes.push(new se(e,t,n,r,a))}else this._mainRule.acceptOverwrite(e,n,r,a)}},ce=class e{static toBinaryStr(e){return e.toString(2).padStart(32,"0")}static print(t){const n=e.getLanguageId(t),r=e.getTokenType(t),a=e.getFontStyle(t),o=e.getForeground(t),i=e.getBackground(t);console.log({languageId:n,tokenType:r,fontStyle:a,foreground:o,background:i})}static getLanguageId(e){return(255&e)>>>0}static getTokenType(e){return(768&e)>>>8}static containsBalancedBrackets(e){return!!(1024&e)}static getFontStyle(e){return(30720&e)>>>11}static getForeground(e){return(16744448&e)>>>15}static getBackground(e){return(4278190080&e)>>>24}static set(t,n,r,a,o,i,l){let s=e.getLanguageId(t),u=e.getTokenType(t),c=e.containsBalancedBrackets(t)?1:0,p=e.getFontStyle(t),d=e.getForeground(t),f=e.getBackground(t);return 0!==n&&(s=n),8!==r&&(u=r),null!==a&&(c=a?1:0),-1!==o&&(p=o),0!==i&&(d=i),0!==l&&(f=l),(s|u<<8|c<<10|p<<11|d<<15|f<<24)>>>0}};function pe(e,t){const n=[],r=function(e){let t=/([LR]:|[\w\.:][\w\.:\-]*|[\,\|\-\(\)])/g,n=t.exec(e);return{next:()=>{if(!n)return null;const r=n[0];return n=t.exec(e),r}}}(e);let a=r.next();for(;null!==a;){let e=0;if(2===a.length&&":"===a.charAt(1)){switch(a.charAt(0)){case"R":e=1;break;case"L":e=-1;break;default:console.log(`Unknown priority ${a} in scope selector`)}a=r.next()}let t=i();if(n.push({matcher:t,priority:e}),","!==a)break;a=r.next()}return n;function o(){if("-"===a){a=r.next();const e=o();return t=>!!e&&!e(t)}if("("===a){a=r.next();const e=function(){const e=[];let t=i();for(;t&&(e.push(t),"|"===a||","===a);){do{a=r.next()}while("|"===a||","===a);t=i()}return t=>e.some((e=>e(t)))}();return")"===a&&(a=r.next()),e}if(de(a)){const e=[];do{e.push(a),a=r.next()}while(de(a));return n=>t(e,n)}return null}function i(){const e=[];let t=o();for(;t;)e.push(t),t=o();return t=>e.every((e=>e(t)))}}function de(e){return!!e&&!!e.match(/[\w\.:]+/)}function fe(e){"function"==typeof e.dispose&&e.dispose()}var he=class{constructor(e){this.scopeName=e}toKey(){return this.scopeName}},ge=class{constructor(e,t){this.scopeName=e,this.ruleName=t}toKey(){return`${this.scopeName}#${this.ruleName}`}},me=class{_references=[];_seenReferenceKeys=new Set;get references(){return this._references}visitedRule=new Set;add(e){const t=e.toKey();this._seenReferenceKeys.has(t)||(this._seenReferenceKeys.add(t),this._references.push(e))}},ye=class{constructor(e,t){this.repo=e,this.initialScopeName=t,this.seenFullScopeRequests.add(this.initialScopeName),this.Q=[new he(this.initialScopeName)]}seenFullScopeRequests=new Set;seenPartialScopeRequests=new Set;Q;processQueue(){const e=this.Q;this.Q=[];const t=new me;for(const n of e)be(n,this.initialScopeName,this.repo,t);for(const e of t.references)if(e instanceof he){if(this.seenFullScopeRequests.has(e.scopeName))continue;this.seenFullScopeRequests.add(e.scopeName),this.Q.push(e)}else{if(this.seenFullScopeRequests.has(e.scopeName))continue;if(this.seenPartialScopeRequests.has(e.toKey()))continue;this.seenPartialScopeRequests.add(e.toKey()),this.Q.push(e)}}};function be(e,t,n,r){const a=n.lookup(e.scopeName);if(!a){if(e.scopeName===t)throw new Error(`No grammar provided for <${t}>`);return}const o=n.lookup(t);e instanceof he?we({baseGrammar:o,selfGrammar:a},r):ve(e.ruleName,{baseGrammar:o,selfGrammar:a,repository:a.repository},r);const i=n.injections(e.scopeName);if(i)for(const e of i)r.add(new he(e))}function ve(e,t,n){t.repository&&t.repository[e]&&xe([t.repository[e]],t,n)}function we(e,t){e.selfGrammar.patterns&&Array.isArray(e.selfGrammar.patterns)&&xe(e.selfGrammar.patterns,{...e,repository:e.selfGrammar.repository},t),e.selfGrammar.injections&&xe(Object.values(e.selfGrammar.injections),{...e,repository:e.selfGrammar.repository},t)}function xe(e,t,n){for(const r of e){if(n.visitedRule.has(r))continue;n.visitedRule.add(r);const e=r.repository?W({},t.repository,r.repository):t.repository;Array.isArray(r.patterns)&&xe(r.patterns,{...t,repository:e},n);const a=r.include;if(!a)continue;const o=Ne(a);switch(o.kind){case 0:we({...t,selfGrammar:t.baseGrammar},n);break;case 1:we(t,n);break;case 2:ve(o.ruleName,{...t,repository:e},n);break;case 3:case 4:const r=o.scopeName===t.selfGrammar.scopeName?t.selfGrammar:o.scopeName===t.baseGrammar.scopeName?t.baseGrammar:void 0;if(r){const a={baseGrammar:t.baseGrammar,selfGrammar:r,repository:e};4===o.kind?ve(o.ruleName,a,n):we(a,n)}else 4===o.kind?n.add(new ge(o.scopeName,o.ruleName)):n.add(new he(o.scopeName))}}}var ke=class{kind=0},Se=class{kind=1},Ce=class{constructor(e){this.ruleName=e}kind=2},_e=class{constructor(e){this.scopeName=e}kind=3},Ee=class{constructor(e,t){this.scopeName=e,this.ruleName=t}kind=4};function Ne(e){if("$base"===e)return new ke;if("$self"===e)return new Se;const t=e.indexOf("#");if(-1===t)return new _e(e);if(0===t)return new Ce(e.substring(1));{const n=e.substring(0,t),r=e.substring(t+1);return new Ee(n,r)}}var Pe=/\\(\d+)/,Ae=/\\(\d+)/g;Symbol("RuleId");var Le=class{$location;id;_nameIsCapturing;_name;_contentNameIsCapturing;_contentName;constructor(e,t,n,r){this.$location=e,this.id=t,this._name=n||null,this._nameIsCapturing=Q.hasCaptures(this._name),this._contentName=r||null,this._contentNameIsCapturing=Q.hasCaptures(this._contentName)}get debugName(){const e=this.$location?`${H(this.$location.filename)}:${this.$location.line}`:"unknown";return`${this.constructor.name}#${this.id} @ ${e}`}getName(e,t){return this._nameIsCapturing&&null!==this._name&&null!==e&&null!==t?Q.replaceCaptures(this._name,e,t):this._name}getContentName(e,t){return this._contentNameIsCapturing&&null!==this._contentName?Q.replaceCaptures(this._contentName,e,t):this._contentName}},Te=class extends Le{retokenizeCapturedWithRuleId;constructor(e,t,n,r,a){super(e,t,n,r),this.retokenizeCapturedWithRuleId=a}dispose(){}collectPatterns(e,t){throw new Error("Not supported!")}compile(e,t){throw new Error("Not supported!")}compileAG(e,t,n,r){throw new Error("Not supported!")}},Re=class extends Le{_match;captures;_cachedCompiledPatterns;constructor(e,t,n,r,a){super(e,t,n,null),this._match=new ze(r,this.id),this.captures=a,this._cachedCompiledPatterns=null}dispose(){this._cachedCompiledPatterns&&(this._cachedCompiledPatterns.dispose(),this._cachedCompiledPatterns=null)}get debugMatchRegExp(){return`${this._match.source}`}collectPatterns(e,t){t.push(this._match)}compile(e,t){return this._getCachedCompiledPatterns(e).compile(e)}compileAG(e,t,n,r){return this._getCachedCompiledPatterns(e).compileAG(e,n,r)}_getCachedCompiledPatterns(e){return this._cachedCompiledPatterns||(this._cachedCompiledPatterns=new $e,this.collectPatterns(e,this._cachedCompiledPatterns)),this._cachedCompiledPatterns}},Ie=class extends Le{hasMissingPatterns;patterns;_cachedCompiledPatterns;constructor(e,t,n,r,a){super(e,t,n,r),this.patterns=a.patterns,this.hasMissingPatterns=a.hasMissingPatterns,this._cachedCompiledPatterns=null}dispose(){this._cachedCompiledPatterns&&(this._cachedCompiledPatterns.dispose(),this._cachedCompiledPatterns=null)}collectPatterns(e,t){for(const n of this.patterns)e.getRule(n).collectPatterns(e,t)}compile(e,t){return this._getCachedCompiledPatterns(e).compile(e)}compileAG(e,t,n,r){return this._getCachedCompiledPatterns(e).compileAG(e,n,r)}_getCachedCompiledPatterns(e){return this._cachedCompiledPatterns||(this._cachedCompiledPatterns=new $e,this.collectPatterns(e,this._cachedCompiledPatterns)),this._cachedCompiledPatterns}},Me=class extends Le{_begin;beginCaptures;_end;endHasBackReferences;endCaptures;applyEndPatternLast;hasMissingPatterns;patterns;_cachedCompiledPatterns;constructor(e,t,n,r,a,o,i,l,s,u){super(e,t,n,r),this._begin=new ze(a,this.id),this.beginCaptures=o,this._end=new ze(i||"",-1),this.endHasBackReferences=this._end.hasBackReferences,this.endCaptures=l,this.applyEndPatternLast=s||!1,this.patterns=u.patterns,this.hasMissingPatterns=u.hasMissingPatterns,this._cachedCompiledPatterns=null}dispose(){this._cachedCompiledPatterns&&(this._cachedCompiledPatterns.dispose(),this._cachedCompiledPatterns=null)}get debugBeginRegExp(){return`${this._begin.source}`}get debugEndRegExp(){return`${this._end.source}`}getEndWithResolvedBackReferences(e,t){return this._end.resolveBackReferences(e,t)}collectPatterns(e,t){t.push(this._begin)}compile(e,t){return this._getCachedCompiledPatterns(e,t).compile(e)}compileAG(e,t,n,r){return this._getCachedCompiledPatterns(e,t).compileAG(e,n,r)}_getCachedCompiledPatterns(e,t){if(!this._cachedCompiledPatterns){this._cachedCompiledPatterns=new $e;for(const t of this.patterns)e.getRule(t).collectPatterns(e,this._cachedCompiledPatterns);this.applyEndPatternLast?this._cachedCompiledPatterns.push(this._end.hasBackReferences?this._end.clone():this._end):this._cachedCompiledPatterns.unshift(this._end.hasBackReferences?this._end.clone():this._end)}return this._end.hasBackReferences&&(this.applyEndPatternLast?this._cachedCompiledPatterns.setSource(this._cachedCompiledPatterns.length()-1,t):this._cachedCompiledPatterns.setSource(0,t)),this._cachedCompiledPatterns}},Oe=class extends Le{_begin;beginCaptures;whileCaptures;_while;whileHasBackReferences;hasMissingPatterns;patterns;_cachedCompiledPatterns;_cachedCompiledWhilePatterns;constructor(e,t,n,r,a,o,i,l,s){super(e,t,n,r),this._begin=new ze(a,this.id),this.beginCaptures=o,this.whileCaptures=l,this._while=new ze(i,-2),this.whileHasBackReferences=this._while.hasBackReferences,this.patterns=s.patterns,this.hasMissingPatterns=s.hasMissingPatterns,this._cachedCompiledPatterns=null,this._cachedCompiledWhilePatterns=null}dispose(){this._cachedCompiledPatterns&&(this._cachedCompiledPatterns.dispose(),this._cachedCompiledPatterns=null),this._cachedCompiledWhilePatterns&&(this._cachedCompiledWhilePatterns.dispose(),this._cachedCompiledWhilePatterns=null)}get debugBeginRegExp(){return`${this._begin.source}`}get debugWhileRegExp(){return`${this._while.source}`}getWhileWithResolvedBackReferences(e,t){return this._while.resolveBackReferences(e,t)}collectPatterns(e,t){t.push(this._begin)}compile(e,t){return this._getCachedCompiledPatterns(e).compile(e)}compileAG(e,t,n,r){return this._getCachedCompiledPatterns(e).compileAG(e,n,r)}_getCachedCompiledPatterns(e){if(!this._cachedCompiledPatterns){this._cachedCompiledPatterns=new $e;for(const t of this.patterns)e.getRule(t).collectPatterns(e,this._cachedCompiledPatterns)}return this._cachedCompiledPatterns}compileWhile(e,t){return this._getCachedCompiledWhilePatterns(e,t).compile(e)}compileWhileAG(e,t,n,r){return this._getCachedCompiledWhilePatterns(e,t).compileAG(e,n,r)}_getCachedCompiledWhilePatterns(e,t){return this._cachedCompiledWhilePatterns||(this._cachedCompiledWhilePatterns=new $e,this._cachedCompiledWhilePatterns.push(this._while.hasBackReferences?this._while.clone():this._while)),this._while.hasBackReferences&&this._cachedCompiledWhilePatterns.setSource(0,t||""),this._cachedCompiledWhilePatterns}},je=class e{static createCaptureRule(e,t,n,r,a){return e.registerRule((e=>new Te(t,e,n,r,a)))}static getCompiledRuleId(t,n,r){return t.id||n.registerRule((a=>{if(t.id=a,t.match)return new Re(t.$vscodeTextmateLocation,t.id,t.name,t.match,e._compileCaptures(t.captures,n,r));if(void 0===t.begin){t.repository&&(r=W({},r,t.repository));let a=t.patterns;return void 0===a&&t.include&&(a=[{include:t.include}]),new Ie(t.$vscodeTextmateLocation,t.id,t.name,t.contentName,e._compilePatterns(a,n,r))}return t.while?new Oe(t.$vscodeTextmateLocation,t.id,t.name,t.contentName,t.begin,e._compileCaptures(t.beginCaptures||t.captures,n,r),t.while,e._compileCaptures(t.whileCaptures||t.captures,n,r),e._compilePatterns(t.patterns,n,r)):new Me(t.$vscodeTextmateLocation,t.id,t.name,t.contentName,t.begin,e._compileCaptures(t.beginCaptures||t.captures,n,r),t.end,e._compileCaptures(t.endCaptures||t.captures,n,r),t.applyEndPatternLast,e._compilePatterns(t.patterns,n,r))})),t.id}static _compileCaptures(t,n,r){let a=[];if(t){let o=0;for(const e in t){if("$vscodeTextmateLocation"===e)continue;const t=parseInt(e,10);t>o&&(o=t)}for(let e=0;e<=o;e++)a[e]=null;for(const o in t){if("$vscodeTextmateLocation"===o)continue;const i=parseInt(o,10);let l=0;t[o].patterns&&(l=e.getCompiledRuleId(t[o],n,r)),a[i]=e.createCaptureRule(n,t[o].$vscodeTextmateLocation,t[o].name,t[o].contentName,l)}}return a}static _compilePatterns(t,n,r){let a=[];if(t)for(let o=0,i=t.length;o<i;o++){const i=t[o];let l=-1;if(i.include){const t=Ne(i.include);switch(t.kind){case 0:case 1:l=e.getCompiledRuleId(r[i.include],n,r);break;case 2:let a=r[t.ruleName];a&&(l=e.getCompiledRuleId(a,n,r));break;case 3:case 4:const o=t.scopeName,s=4===t.kind?t.ruleName:null,u=n.getExternalGrammar(o,r);if(u)if(s){let t=u.repository[s];t&&(l=e.getCompiledRuleId(t,n,u.repository))}else l=e.getCompiledRuleId(u.repository.$self,n,u.repository)}}else l=e.getCompiledRuleId(i,n,r);if(-1!==l){const e=n.getRule(l);let t=!1;if((e instanceof Ie||e instanceof Me||e instanceof Oe)&&e.hasMissingPatterns&&0===e.patterns.length&&(t=!0),t)continue;a.push(l)}}return{patterns:a,hasMissingPatterns:(t?t.length:0)!==a.length}}},ze=class e{source;ruleId;hasAnchor;hasBackReferences;_anchorCache;constructor(e,t){if(e&&"string"==typeof e){const t=e.length;let n=0,r=[],a=!1;for(let o=0;o<t;o++)if("\\"===e.charAt(o)&&o+1<t){const t=e.charAt(o+1);"z"===t?(r.push(e.substring(n,o)),r.push("$(?!\\n)(?<!\\n)"),n=o+2):"A"!==t&&"G"!==t||(a=!0),o++}this.hasAnchor=a,0===n?this.source=e:(r.push(e.substring(n,t)),this.source=r.join(""))}else this.hasAnchor=!1,this.source=e;this.hasAnchor?this._anchorCache=this._buildAnchorCache():this._anchorCache=null,this.ruleId=t,"string"==typeof this.source?this.hasBackReferences=Pe.test(this.source):this.hasBackReferences=!1}clone(){return new e(this.source,this.ruleId)}setSource(e){this.source!==e&&(this.source=e,this.hasAnchor&&(this._anchorCache=this._buildAnchorCache()))}resolveBackReferences(e,t){if("string"!=typeof this.source)throw new Error("This method should only be called if the source is a string");let n=t.map((t=>e.substring(t.start,t.end)));return Ae.lastIndex=0,this.source.replace(Ae,((e,t)=>Z(n[parseInt(t,10)]||"")))}_buildAnchorCache(){if("string"!=typeof this.source)throw new Error("This method should only be called if the source is a string");let e,t,n,r,a=[],o=[],i=[],l=[];for(e=0,t=this.source.length;e<t;e++)n=this.source.charAt(e),a[e]=n,o[e]=n,i[e]=n,l[e]=n,"\\"===n&&e+1<t&&(r=this.source.charAt(e+1),"A"===r?(a[e+1]="",o[e+1]="",i[e+1]="A",l[e+1]="A"):"G"===r?(a[e+1]="",o[e+1]="G",i[e+1]="",l[e+1]="G"):(a[e+1]=r,o[e+1]=r,i[e+1]=r,l[e+1]=r),e++);return{A0_G0:a.join(""),A0_G1:o.join(""),A1_G0:i.join(""),A1_G1:l.join("")}}resolveAnchors(e,t){return this.hasAnchor&&this._anchorCache&&"string"==typeof this.source?e?t?this._anchorCache.A1_G1:this._anchorCache.A1_G0:t?this._anchorCache.A0_G1:this._anchorCache.A0_G0:this.source}},$e=class{_items;_hasAnchors;_cached;_anchorCache;constructor(){this._items=[],this._hasAnchors=!1,this._cached=null,this._anchorCache={A0_G0:null,A0_G1:null,A1_G0:null,A1_G1:null}}dispose(){this._disposeCaches()}_disposeCaches(){this._cached&&(this._cached.dispose(),this._cached=null),this._anchorCache.A0_G0&&(this._anchorCache.A0_G0.dispose(),this._anchorCache.A0_G0=null),this._anchorCache.A0_G1&&(this._anchorCache.A0_G1.dispose(),this._anchorCache.A0_G1=null),this._anchorCache.A1_G0&&(this._anchorCache.A1_G0.dispose(),this._anchorCache.A1_G0=null),this._anchorCache.A1_G1&&(this._anchorCache.A1_G1.dispose(),this._anchorCache.A1_G1=null)}push(e){this._items.push(e),this._hasAnchors=this._hasAnchors||e.hasAnchor}unshift(e){this._items.unshift(e),this._hasAnchors=this._hasAnchors||e.hasAnchor}length(){return this._items.length}setSource(e,t){this._items[e].source!==t&&(this._disposeCaches(),this._items[e].setSource(t))}compile(e){if(!this._cached){let t=this._items.map((e=>e.source));this._cached=new Fe(e,t,this._items.map((e=>e.ruleId)))}return this._cached}compileAG(e,t,n){return this._hasAnchors?t?n?(this._anchorCache.A1_G1||(this._anchorCache.A1_G1=this._resolveAnchors(e,t,n)),this._anchorCache.A1_G1):(this._anchorCache.A1_G0||(this._anchorCache.A1_G0=this._resolveAnchors(e,t,n)),this._anchorCache.A1_G0):n?(this._anchorCache.A0_G1||(this._anchorCache.A0_G1=this._resolveAnchors(e,t,n)),this._anchorCache.A0_G1):(this._anchorCache.A0_G0||(this._anchorCache.A0_G0=this._resolveAnchors(e,t,n)),this._anchorCache.A0_G0):this.compile(e)}_resolveAnchors(e,t,n){let r=this._items.map((e=>e.resolveAnchors(t,n)));return new Fe(e,r,this._items.map((e=>e.ruleId)))}},Fe=class{constructor(e,t,n){this.regExps=t,this.rules=n,this.scanner=e.createOnigScanner(t)}scanner;dispose(){"function"==typeof this.scanner.dispose&&this.scanner.dispose()}toString(){const e=[];for(let t=0,n=this.rules.length;t<n;t++)e.push("   - "+this.rules[t]+": "+this.regExps[t]);return e.join("\n")}findNextMatchSync(e,t,n){const r=this.scanner.findNextMatchSync(e,t,n);return r?{ruleId:this.rules[r.index],captureIndices:r.captureIndices}:null}},De=class{constructor(e,t){this.languageId=e,this.tokenType=t}},Be=class e{_defaultAttributes;_embeddedLanguagesMatcher;constructor(e,t){this._defaultAttributes=new De(e,8),this._embeddedLanguagesMatcher=new Ue(Object.entries(t||{}))}getDefaultAttributes(){return this._defaultAttributes}getBasicScopeAttributes(t){return null===t?e._NULL_SCOPE_METADATA:this._getBasicScopeAttributes.get(t)}static _NULL_SCOPE_METADATA=new De(0,0);_getBasicScopeAttributes=new J((e=>{const t=this._scopeToLanguage(e),n=this._toStandardTokenType(e);return new De(t,n)}));_scopeToLanguage(e){return this._embeddedLanguagesMatcher.match(e)||0}_toStandardTokenType(t){const n=t.match(e.STANDARD_TOKEN_TYPE_REGEXP);if(!n)return 8;switch(n[1]){case"comment":return 1;case"string":return 2;case"regex":return 3;case"meta.embedded":return 0}throw new Error("Unexpected match for standard token type!")}static STANDARD_TOKEN_TYPE_REGEXP=/\b(comment|string|regex|meta\.embedded)\b/},Ue=class{values;scopesRegExp;constructor(e){if(0===e.length)this.values=null,this.scopesRegExp=null;else{this.values=new Map(e);const t=e.map((([e,t])=>Z(e)));t.sort(),t.reverse(),this.scopesRegExp=new RegExp(`^((${t.join(")|(")}))($|\\.)`,"")}}match(e){if(!this.scopesRegExp)return;const t=e.match(this.scopesRegExp);return t?this.values.get(t[1]):void 0}},qe=("undefined"!=typeof process&&process.env.VSCODE_TEXTMATE_DEBUG,class{constructor(e,t){this.stack=e,this.stoppedEarly=t}});function Ge(e,t,n,r,a,o,i,l){const s=t.content.length;let u=!1,c=-1;if(i){const i=function(e,t,n,r,a,o){let i=a.beginRuleCapturedEOL?0:-1;const l=[];for(let t=a;t;t=t.pop()){const n=t.getRule(e);n instanceof Oe&&l.push({rule:n,stack:t})}for(let s=l.pop();s;s=l.pop()){const{ruleScanner:l,findOptions:u}=He(s.rule,e,s.stack.endRule,n,r===i),c=l.findNextMatchSync(t,r,u);if(!c){a=s.stack.pop();break}if(-2!==c.ruleId){a=s.stack.pop();break}c.captureIndices&&c.captureIndices.length&&(o.produce(s.stack,c.captureIndices[0].start),Ve(e,t,n,s.stack,o,s.rule.whileCaptures,c.captureIndices),o.produce(s.stack,c.captureIndices[0].end),i=c.captureIndices[0].end,c.captureIndices[0].end>r&&(r=c.captureIndices[0].end,n=!1))}return{stack:a,linePos:r,anchorPosition:i,isFirstLine:n}}(e,t,n,r,a,o);a=i.stack,r=i.linePos,n=i.isFirstLine,c=i.anchorPosition}const p=Date.now();for(;!u;){if(0!==l&&Date.now()-p>l)return new qe(a,!0);d()}return new qe(a,!1);function d(){const i=function(e,t,n,r,a,o){const i=function(e,t,n,r,a,o){const i=a.getRule(e),{ruleScanner:l,findOptions:s}=We(i,e,a.endRule,n,r===o),u=l.findNextMatchSync(t,r,s);return u?{captureIndices:u.captureIndices,matchedRuleId:u.ruleId}:null}(e,t,n,r,a,o),l=e.getInjections();if(0===l.length)return i;const s=function(e,t,n,r,a,o,i){let l,s=Number.MAX_VALUE,u=null,c=0;const p=o.contentNameScopesList.getScopeNames();for(let o=0,d=e.length;o<d;o++){const d=e[o];if(!d.matcher(p))continue;const f=t.getRule(d.ruleId),{ruleScanner:h,findOptions:g}=We(f,t,null,r,a===i),m=h.findNextMatchSync(n,a,g);if(!m)continue;const y=m.captureIndices[0].start;if(!(y>=s)&&(s=y,u=m.captureIndices,l=m.ruleId,c=d.priority,s===a))break}return u?{priorityMatch:-1===c,captureIndices:u,matchedRuleId:l}:null}(l,e,t,n,r,a,o);if(!s)return i;if(!i)return s;const u=i.captureIndices[0].start,c=s.captureIndices[0].start;return c<u||s.priorityMatch&&c===u?s:i}(e,t,n,r,a,c);if(!i)return o.produce(a,s),void(u=!0);const l=i.captureIndices,p=i.matchedRuleId,d=!!(l&&l.length>0)&&l[0].end>r;if(-1===p){const i=a.getRule(e);o.produce(a,l[0].start),a=a.withContentNameScopesList(a.nameScopesList),Ve(e,t,n,a,o,i.endCaptures,l),o.produce(a,l[0].end);const p=a;if(a=a.parent,c=p.getAnchorPos(),!d&&p.getEnterPos()===r)return a=p,o.produce(a,s),void(u=!0)}else{const i=e.getRule(p);o.produce(a,l[0].start);const f=a,h=i.getName(t.content,l),g=a.contentNameScopesList.pushAttributed(h,e);if(a=a.push(p,r,c,l[0].end===s,null,g,g),i instanceof Me){const r=i;Ve(e,t,n,a,o,r.beginCaptures,l),o.produce(a,l[0].end),c=l[0].end;const p=r.getContentName(t.content,l),h=g.pushAttributed(p,e);if(a=a.withContentNameScopesList(h),r.endHasBackReferences&&(a=a.withEndRule(r.getEndWithResolvedBackReferences(t.content,l))),!d&&f.hasSameRuleAs(a))return a=a.pop(),o.produce(a,s),void(u=!0)}else if(i instanceof Oe){const r=i;Ve(e,t,n,a,o,r.beginCaptures,l),o.produce(a,l[0].end),c=l[0].end;const p=r.getContentName(t.content,l),h=g.pushAttributed(p,e);if(a=a.withContentNameScopesList(h),r.whileHasBackReferences&&(a=a.withEndRule(r.getWhileWithResolvedBackReferences(t.content,l))),!d&&f.hasSameRuleAs(a))return a=a.pop(),o.produce(a,s),void(u=!0)}else if(Ve(e,t,n,a,o,i.captures,l),o.produce(a,l[0].end),a=a.pop(),!d)return a=a.safePop(),o.produce(a,s),void(u=!0)}l[0].end>r&&(r=l[0].end,n=!1)}}function We(e,t,n,r,a){return{ruleScanner:e.compileAG(t,n,r,a),findOptions:0}}function He(e,t,n,r,a){return{ruleScanner:e.compileWhileAG(t,n,r,a),findOptions:0}}function Ve(e,t,n,r,a,o,i){if(0===o.length)return;const l=t.content,s=Math.min(o.length,i.length),u=[],c=i[0].end;for(let t=0;t<s;t++){const s=o[t];if(null===s)continue;const p=i[t];if(0===p.length)continue;if(p.start>c)break;for(;u.length>0&&u[u.length-1].endPos<=p.start;)a.produceFromScopes(u[u.length-1].scopes,u[u.length-1].endPos),u.pop();if(u.length>0?a.produceFromScopes(u[u.length-1].scopes,p.start):a.produce(r,p.start),s.retokenizeCapturedWithRuleId){const t=s.getName(l,i),o=r.contentNameScopesList.pushAttributed(t,e),u=s.getContentName(l,i),c=o.pushAttributed(u,e),d=r.push(s.retokenizeCapturedWithRuleId,p.start,-1,!1,null,o,c),f=e.createOnigString(l.substring(0,p.end));Ge(e,f,n&&0===p.start,p.start,d,a,!1,0),fe(f);continue}const d=s.getName(l,i);if(null!==d){const t=(u.length>0?u[u.length-1].scopes:r.contentNameScopesList).pushAttributed(d,e);u.push(new Qe(t,p.end))}}for(;u.length>0;)a.produceFromScopes(u[u.length-1].scopes,u[u.length-1].endPos),u.pop()}var Qe=class{scopes;endPos;constructor(e,t){this.scopes=e,this.endPos=t}};function Xe(e,t,n,r,a){const o=pe(t,Ke),i=je.getCompiledRuleId(n,r,a.repository);for(const n of o)e.push({debugSelector:t,matcher:n.matcher,ruleId:i,grammar:a,priority:n.priority})}function Ke(e,t){if(t.length<e.length)return!1;let n=0;return e.every((e=>{for(let r=n;r<t.length;r++)if(Ye(t[r],e))return n=r+1,!0;return!1}))}function Ye(e,t){if(!e)return!1;if(e===t)return!0;const n=t.length;return e.length>n&&e.substr(0,n)===t&&"."===e[n]}var Ze=class{constructor(e,t,n,r,a,o,i,l){if(this._rootScopeName=e,this.balancedBracketSelectors=o,this._onigLib=l,this._basicScopeAttributesProvider=new Be(n,r),this._rootId=-1,this._lastRuleId=0,this._ruleId2desc=[null],this._includedGrammars={},this._grammarRepository=i,this._grammar=Je(t,null),this._injections=null,this._tokenTypeMatchers=[],a)for(const e of Object.keys(a)){const t=pe(e,Ke);for(const n of t)this._tokenTypeMatchers.push({matcher:n.matcher,type:a[e]})}}_rootId;_lastRuleId;_ruleId2desc;_includedGrammars;_grammarRepository;_grammar;_injections;_basicScopeAttributesProvider;_tokenTypeMatchers;get themeProvider(){return this._grammarRepository}dispose(){for(const e of this._ruleId2desc)e&&e.dispose()}createOnigScanner(e){return this._onigLib.createOnigScanner(e)}createOnigString(e){return this._onigLib.createOnigString(e)}getMetadataForScope(e){return this._basicScopeAttributesProvider.getBasicScopeAttributes(e)}_collectInjections(){const e=[],t=this._rootScopeName,n=(e=>e===this._rootScopeName?this._grammar:this.getExternalGrammar(e))(t);if(n){const r=n.injections;if(r)for(let t in r)Xe(e,t,r[t],this,n);const a=this._grammarRepository.injections(t);a&&a.forEach((t=>{const n=this.getExternalGrammar(t);if(n){const t=n.injectionSelector;t&&Xe(e,t,n,this,n)}}))}return e.sort(((e,t)=>e.priority-t.priority)),e}getInjections(){return null===this._injections&&(this._injections=this._collectInjections()),this._injections}registerRule(e){const t=++this._lastRuleId,n=e(t);return this._ruleId2desc[t]=n,n}getRule(e){return this._ruleId2desc[e]}getExternalGrammar(e,t){if(this._includedGrammars[e])return this._includedGrammars[e];if(this._grammarRepository){const n=this._grammarRepository.lookup(e);if(n)return this._includedGrammars[e]=Je(n,t&&t.$base),this._includedGrammars[e]}}tokenizeLine(e,t,n=0){const r=this._tokenize(e,t,!1,n);return{tokens:r.lineTokens.getResult(r.ruleStack,r.lineLength),ruleStack:r.ruleStack,stoppedEarly:r.stoppedEarly}}tokenizeLine2(e,t,n=0){const r=this._tokenize(e,t,!0,n);return{tokens:r.lineTokens.getBinaryResult(r.ruleStack,r.lineLength),ruleStack:r.ruleStack,stoppedEarly:r.stoppedEarly}}_tokenize(e,t,n,r){let a;if(-1===this._rootId&&(this._rootId=je.getCompiledRuleId(this._grammar.repository.$self,this,this._grammar.repository),this.getInjections()),t&&t!==tt.NULL)a=!1,t.reset();else{a=!0;const e=this._basicScopeAttributesProvider.getDefaultAttributes(),n=this.themeProvider.getDefaults(),r=ce.set(0,e.languageId,e.tokenType,null,n.fontStyle,n.foregroundId,n.backgroundId),o=this.getRule(this._rootId).getName(null,null);let i;i=o?et.createRootAndLookUpScopeName(o,r,this):et.createRoot("unknown",r),t=new tt(null,this._rootId,-1,-1,!1,null,i,i)}e+="\n";const o=this.createOnigString(e),i=o.content.length,l=new rt(n,e,this._tokenTypeMatchers,this.balancedBracketSelectors),s=Ge(this,o,a,0,t,l,!0,r);return fe(o),{lineLength:i,lineTokens:l,ruleStack:s.stack,stoppedEarly:s.stoppedEarly}}};function Je(e,t){return(e=G(e)).repository=e.repository||{},e.repository.$self={$vscodeTextmateLocation:e.$vscodeTextmateLocation,patterns:e.patterns,name:e.scopeName},e.repository.$base=t||e.repository.$self,e}var et=class e{constructor(e,t,n){this.parent=e,this.scopePath=t,this.tokenAttributes=n}static fromExtension(t,n){let r=t,a=t?.scopePath??null;for(const t of n)a=te.push(a,t.scopeNames),r=new e(r,a,t.encodedTokenAttributes);return r}static createRoot(t,n){return new e(null,new te(null,t),n)}static createRootAndLookUpScopeName(t,n,r){const a=r.getMetadataForScope(t),o=new te(null,t),i=r.themeProvider.themeMatch(o),l=e.mergeAttributes(n,a,i);return new e(null,o,l)}get scopeName(){return this.scopePath.scopeName}toString(){return this.getScopeNames().join(" ")}equals(t){return e.equals(this,t)}static equals(e,t){for(;;){if(e===t)return!0;if(!e&&!t)return!0;if(!e||!t)return!1;if(e.scopeName!==t.scopeName||e.tokenAttributes!==t.tokenAttributes)return!1;e=e.parent,t=t.parent}}static mergeAttributes(e,t,n){let r=-1,a=0,o=0;return null!==n&&(r=n.fontStyle,a=n.foregroundId,o=n.backgroundId),ce.set(e,t.languageId,t.tokenType,null,r,a,o)}pushAttributed(t,n){if(null===t)return this;if(-1===t.indexOf(" "))return e._pushAttributed(this,t,n);const r=t.split(/ /g);let a=this;for(const t of r)a=e._pushAttributed(a,t,n);return a}static _pushAttributed(t,n,r){const a=r.getMetadataForScope(n),o=t.scopePath.push(n),i=r.themeProvider.themeMatch(o),l=e.mergeAttributes(t.tokenAttributes,a,i);return new e(t,o,l)}getScopeNames(){return this.scopePath.getSegments()}getExtensionIfDefined(e){const t=[];let n=this;for(;n&&n!==e;)t.push({encodedTokenAttributes:n.tokenAttributes,scopeNames:n.scopePath.getExtensionIfDefined(n.parent?.scopePath??null)}),n=n.parent;return n===e?t.reverse():void 0}},tt=class e{constructor(e,t,n,r,a,o,i,l){this.parent=e,this.ruleId=t,this.beginRuleCapturedEOL=a,this.endRule=o,this.nameScopesList=i,this.contentNameScopesList=l,this.depth=this.parent?this.parent.depth+1:1,this._enterPos=n,this._anchorPos=r}_stackElementBrand=void 0;static NULL=new e(null,0,0,0,!1,null,null,null);_enterPos;_anchorPos;depth;equals(t){return null!==t&&e._equals(this,t)}static _equals(e,t){return e===t||!!this._structuralEquals(e,t)&&et.equals(e.contentNameScopesList,t.contentNameScopesList)}static _structuralEquals(e,t){for(;;){if(e===t)return!0;if(!e&&!t)return!0;if(!e||!t)return!1;if(e.depth!==t.depth||e.ruleId!==t.ruleId||e.endRule!==t.endRule)return!1;e=e.parent,t=t.parent}}clone(){return this}static _reset(e){for(;e;)e._enterPos=-1,e._anchorPos=-1,e=e.parent}reset(){e._reset(this)}pop(){return this.parent}safePop(){return this.parent?this.parent:this}push(t,n,r,a,o,i,l){return new e(this,t,n,r,a,o,i,l)}getEnterPos(){return this._enterPos}getAnchorPos(){return this._anchorPos}getRule(e){return e.getRule(this.ruleId)}toString(){const e=[];return this._writeString(e,0),"["+e.join(",")+"]"}_writeString(e,t){return this.parent&&(t=this.parent._writeString(e,t)),e[t++]=`(${this.ruleId}, ${this.nameScopesList?.toString()}, ${this.contentNameScopesList?.toString()})`,t}withContentNameScopesList(e){return this.contentNameScopesList===e?this:this.parent.push(this.ruleId,this._enterPos,this._anchorPos,this.beginRuleCapturedEOL,this.endRule,this.nameScopesList,e)}withEndRule(t){return this.endRule===t?this:new e(this.parent,this.ruleId,this._enterPos,this._anchorPos,this.beginRuleCapturedEOL,t,this.nameScopesList,this.contentNameScopesList)}hasSameRuleAs(e){let t=this;for(;t&&t._enterPos===e._enterPos;){if(t.ruleId===e.ruleId)return!0;t=t.parent}return!1}toStateStackFrame(){return{ruleId:this.ruleId,beginRuleCapturedEOL:this.beginRuleCapturedEOL,endRule:this.endRule,nameScopesList:this.nameScopesList?.getExtensionIfDefined(this.parent?.nameScopesList??null)??[],contentNameScopesList:this.contentNameScopesList?.getExtensionIfDefined(this.nameScopesList)??[]}}static pushFrame(t,n){const r=et.fromExtension(t?.nameScopesList??null,n.nameScopesList);return new e(t,n.ruleId,n.enterPos??-1,n.anchorPos??-1,n.beginRuleCapturedEOL,n.endRule,r,et.fromExtension(r,n.contentNameScopesList))}},nt=class{balancedBracketScopes;unbalancedBracketScopes;allowAny=!1;constructor(e,t){this.balancedBracketScopes=e.flatMap((e=>"*"===e?(this.allowAny=!0,[]):pe(e,Ke).map((e=>e.matcher)))),this.unbalancedBracketScopes=t.flatMap((e=>pe(e,Ke).map((e=>e.matcher))))}get matchesAlways(){return this.allowAny&&0===this.unbalancedBracketScopes.length}get matchesNever(){return 0===this.balancedBracketScopes.length&&!this.allowAny}match(e){for(const t of this.unbalancedBracketScopes)if(t(e))return!1;for(const t of this.balancedBracketScopes)if(t(e))return!0;return this.allowAny}},rt=class{constructor(e,t,n,r){this.balancedBracketSelectors=r,this._emitBinaryTokens=e,this._tokenTypeOverrides=n,this._lineText=null,this._tokens=[],this._binaryTokens=[],this._lastTokenEndIndex=0}_emitBinaryTokens;_lineText;_tokens;_binaryTokens;_lastTokenEndIndex;_tokenTypeOverrides;produce(e,t){this.produceFromScopes(e.contentNameScopesList,t)}produceFromScopes(e,t){if(this._lastTokenEndIndex>=t)return;if(this._emitBinaryTokens){let n=e?.tokenAttributes??0,r=!1;if(this.balancedBracketSelectors?.matchesAlways&&(r=!0),this._tokenTypeOverrides.length>0||this.balancedBracketSelectors&&!this.balancedBracketSelectors.matchesAlways&&!this.balancedBracketSelectors.matchesNever){const t=e?.getScopeNames()??[];for(const e of this._tokenTypeOverrides)e.matcher(t)&&(n=ce.set(n,0,e.type,null,-1,0,0));this.balancedBracketSelectors&&(r=this.balancedBracketSelectors.match(t))}return r&&(n=ce.set(n,0,8,r,-1,0,0)),this._binaryTokens.length>0&&this._binaryTokens[this._binaryTokens.length-1]===n||(this._binaryTokens.push(this._lastTokenEndIndex),this._binaryTokens.push(n)),void(this._lastTokenEndIndex=t)}const n=e?.getScopeNames()??[];this._tokens.push({startIndex:this._lastTokenEndIndex,endIndex:t,scopes:n}),this._lastTokenEndIndex=t}getResult(e,t){return this._tokens.length>0&&this._tokens[this._tokens.length-1].startIndex===t-1&&this._tokens.pop(),0===this._tokens.length&&(this._lastTokenEndIndex=-1,this.produce(e,t),this._tokens[this._tokens.length-1].startIndex=0),this._tokens}getBinaryResult(e,t){this._binaryTokens.length>0&&this._binaryTokens[this._binaryTokens.length-2]===t-1&&(this._binaryTokens.pop(),this._binaryTokens.pop()),0===this._binaryTokens.length&&(this._lastTokenEndIndex=-1,this.produce(e,t),this._binaryTokens[this._binaryTokens.length-2]=0);const n=new Uint32Array(this._binaryTokens.length);for(let e=0,t=this._binaryTokens.length;e<t;e++)n[e]=this._binaryTokens[e];return n}},at=class{constructor(e,t){this._onigLib=t,this._theme=e}_grammars=new Map;_rawGrammars=new Map;_injectionGrammars=new Map;_theme;dispose(){for(const e of this._grammars.values())e.dispose()}setTheme(e){this._theme=e}getColorMap(){return this._theme.getColorMap()}addGrammar(e,t){this._rawGrammars.set(e.scopeName,e),t&&this._injectionGrammars.set(e.scopeName,t)}lookup(e){return this._rawGrammars.get(e)}injections(e){return this._injectionGrammars.get(e)}getDefaults(){return this._theme.getDefaults()}themeMatch(e){return this._theme.match(e)}grammarForScopeName(e,t,n,r,a){if(!this._grammars.has(e)){let o=this._rawGrammars.get(e);if(!o)return null;this._grammars.set(e,function(e,t,n,r,a,o,i,l){return new Ze(e,t,n,r,a,o,i,l)}(e,o,t,n,r,a,this,this._onigLib))}return this._grammars.get(e)}},ot=class{_options;_syncRegistry;_ensureGrammarCache;constructor(e){this._options=e,this._syncRegistry=new at(ee.createFromRawTheme(e.theme,e.colorMap),e.onigLib),this._ensureGrammarCache=new Map}dispose(){this._syncRegistry.dispose()}setTheme(e,t){this._syncRegistry.setTheme(ee.createFromRawTheme(e,t))}getColorMap(){return this._syncRegistry.getColorMap()}loadGrammarWithEmbeddedLanguages(e,t,n){return this.loadGrammarWithConfiguration(e,t,{embeddedLanguages:n})}loadGrammarWithConfiguration(e,t,n){return this._loadGrammar(e,t,n.embeddedLanguages,n.tokenTypes,new nt(n.balancedBracketSelectors||[],n.unbalancedBracketSelectors||[]))}loadGrammar(e){return this._loadGrammar(e,0,null,null,null)}_loadGrammar(e,t,n,r,a){const o=new ye(this._syncRegistry,e);for(;o.Q.length>0;)o.Q.map((e=>this._loadSingleGrammar(e.scopeName))),o.processQueue();return this._grammarForScopeName(e,t,n,r,a)}_loadSingleGrammar(e){this._ensureGrammarCache.has(e)||(this._doLoadSingleGrammar(e),this._ensureGrammarCache.set(e,!0))}_doLoadSingleGrammar(e){const t=this._options.loadGrammar(e);if(t){const n="function"==typeof this._options.getInjections?this._options.getInjections(e):void 0;this._syncRegistry.addGrammar(t,n)}}addGrammar(e,t=[],n=0,r=null){return this._syncRegistry.addGrammar(e,t),this._grammarForScopeName(e.scopeName,n,r)}_grammarForScopeName(e,t=0,n=null,r=null,a=null){return this._syncRegistry.grammarForScopeName(e,t,n,r,a)}},it=tt.NULL;const lt=["area","base","basefont","bgsound","br","col","command","embed","frame","hr","image","img","input","keygen","link","meta","param","source","track","wbr"];class st{constructor(e,t,n){this.normal=t,this.property=e,n&&(this.space=n)}}function ut(e,t){const n={},r={};for(const t of e)Object.assign(n,t.property),Object.assign(r,t.normal);return new st(n,r,t)}function ct(e){return e.toLowerCase()}st.prototype.normal={},st.prototype.property={},st.prototype.space=void 0;class pt{constructor(e,t){this.attribute=t,this.property=e}}pt.prototype.attribute="",pt.prototype.booleanish=!1,pt.prototype.boolean=!1,pt.prototype.commaOrSpaceSeparated=!1,pt.prototype.commaSeparated=!1,pt.prototype.defined=!1,pt.prototype.mustUseProperty=!1,pt.prototype.number=!1,pt.prototype.overloadedBoolean=!1,pt.prototype.property="",pt.prototype.spaceSeparated=!1,pt.prototype.space=void 0;let dt=0;const ft=wt(),ht=wt(),gt=wt(),mt=wt(),yt=wt(),bt=wt(),vt=wt();function wt(){return 2**++dt}const xt=Object.keys(r);class kt extends pt{constructor(e,t,n,a){let o=-1;if(super(e,t),St(this,"space",a),"number"==typeof n)for(;++o<xt.length;){const e=xt[o];St(this,xt[o],(n&r[e])===r[e])}}}function St(e,t,n){n&&(e[t]=n)}function Ct(e){const t={},n={};for(const[r,a]of Object.entries(e.properties)){const o=new kt(r,e.transform(e.attributes||{},r),a,e.space);e.mustUseProperty&&e.mustUseProperty.includes(r)&&(o.mustUseProperty=!0),t[r]=o,n[ct(r)]=r,n[ct(o.attribute)]=r}return new st(t,n,e.space)}kt.prototype.defined=!0;const _t=Ct({properties:{ariaActiveDescendant:null,ariaAtomic:ht,ariaAutoComplete:null,ariaBusy:ht,ariaChecked:ht,ariaColCount:mt,ariaColIndex:mt,ariaColSpan:mt,ariaControls:yt,ariaCurrent:null,ariaDescribedBy:yt,ariaDetails:null,ariaDisabled:ht,ariaDropEffect:yt,ariaErrorMessage:null,ariaExpanded:ht,ariaFlowTo:yt,ariaGrabbed:ht,ariaHasPopup:null,ariaHidden:ht,ariaInvalid:null,ariaKeyShortcuts:null,ariaLabel:null,ariaLabelledBy:yt,ariaLevel:mt,ariaLive:null,ariaModal:ht,ariaMultiLine:ht,ariaMultiSelectable:ht,ariaOrientation:null,ariaOwns:yt,ariaPlaceholder:null,ariaPosInSet:mt,ariaPressed:ht,ariaReadOnly:ht,ariaRelevant:null,ariaRequired:ht,ariaRoleDescription:yt,ariaRowCount:mt,ariaRowIndex:mt,ariaRowSpan:mt,ariaSelected:ht,ariaSetSize:mt,ariaSort:null,ariaValueMax:mt,ariaValueMin:mt,ariaValueNow:mt,ariaValueText:null,role:null},transform:(e,t)=>"role"===t?t:"aria-"+t.slice(4).toLowerCase()});function Et(e,t){return t in e?e[t]:t}function Nt(e,t){return Et(e,t.toLowerCase())}const Pt=Ct({attributes:{acceptcharset:"accept-charset",classname:"class",htmlfor:"for",httpequiv:"http-equiv"},mustUseProperty:["checked","multiple","muted","selected"],properties:{abbr:null,accept:bt,acceptCharset:yt,accessKey:yt,action:null,allow:null,allowFullScreen:ft,allowPaymentRequest:ft,allowUserMedia:ft,alt:null,as:null,async:ft,autoCapitalize:null,autoComplete:yt,autoFocus:ft,autoPlay:ft,blocking:yt,capture:null,charSet:null,checked:ft,cite:null,className:yt,cols:mt,colSpan:null,content:null,contentEditable:ht,controls:ft,controlsList:yt,coords:mt|bt,crossOrigin:null,data:null,dateTime:null,decoding:null,default:ft,defer:ft,dir:null,dirName:null,disabled:ft,download:gt,draggable:ht,encType:null,enterKeyHint:null,fetchPriority:null,form:null,formAction:null,formEncType:null,formMethod:null,formNoValidate:ft,formTarget:null,headers:yt,height:mt,hidden:ft,high:mt,href:null,hrefLang:null,htmlFor:yt,httpEquiv:yt,id:null,imageSizes:null,imageSrcSet:null,inert:ft,inputMode:null,integrity:null,is:null,isMap:ft,itemId:null,itemProp:yt,itemRef:yt,itemScope:ft,itemType:yt,kind:null,label:null,lang:null,language:null,list:null,loading:null,loop:ft,low:mt,manifest:null,max:null,maxLength:mt,media:null,method:null,min:null,minLength:mt,multiple:ft,muted:ft,name:null,nonce:null,noModule:ft,noValidate:ft,onAbort:null,onAfterPrint:null,onAuxClick:null,onBeforeMatch:null,onBeforePrint:null,onBeforeToggle:null,onBeforeUnload:null,onBlur:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onContextLost:null,onContextMenu:null,onContextRestored:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnded:null,onError:null,onFocus:null,onFormData:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLanguageChange:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadEnd:null,onLoadStart:null,onMessage:null,onMessageError:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRejectionHandled:null,onReset:null,onResize:null,onScroll:null,onScrollEnd:null,onSecurityPolicyViolation:null,onSeeked:null,onSeeking:null,onSelect:null,onSlotChange:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnhandledRejection:null,onUnload:null,onVolumeChange:null,onWaiting:null,onWheel:null,open:ft,optimum:mt,pattern:null,ping:yt,placeholder:null,playsInline:ft,popover:null,popoverTarget:null,popoverTargetAction:null,poster:null,preload:null,readOnly:ft,referrerPolicy:null,rel:yt,required:ft,reversed:ft,rows:mt,rowSpan:mt,sandbox:yt,scope:null,scoped:ft,seamless:ft,selected:ft,shadowRootClonable:ft,shadowRootDelegatesFocus:ft,shadowRootMode:null,shape:null,size:mt,sizes:null,slot:null,span:mt,spellCheck:ht,src:null,srcDoc:null,srcLang:null,srcSet:null,start:mt,step:null,style:null,tabIndex:mt,target:null,title:null,translate:null,type:null,typeMustMatch:ft,useMap:null,value:ht,width:mt,wrap:null,writingSuggestions:null,align:null,aLink:null,archive:yt,axis:null,background:null,bgColor:null,border:mt,borderColor:null,bottomMargin:mt,cellPadding:null,cellSpacing:null,char:null,charOff:null,classId:null,clear:null,code:null,codeBase:null,codeType:null,color:null,compact:ft,declare:ft,event:null,face:null,frame:null,frameBorder:null,hSpace:mt,leftMargin:mt,link:null,longDesc:null,lowSrc:null,marginHeight:mt,marginWidth:mt,noResize:ft,noHref:ft,noShade:ft,noWrap:ft,object:null,profile:null,prompt:null,rev:null,rightMargin:mt,rules:null,scheme:null,scrolling:ht,standby:null,summary:null,text:null,topMargin:mt,valueType:null,version:null,vAlign:null,vLink:null,vSpace:mt,allowTransparency:null,autoCorrect:null,autoSave:null,disablePictureInPicture:ft,disableRemotePlayback:ft,prefix:null,property:null,results:mt,security:null,unselectable:null},space:"html",transform:Nt}),At=Ct({attributes:{accentHeight:"accent-height",alignmentBaseline:"alignment-baseline",arabicForm:"arabic-form",baselineShift:"baseline-shift",capHeight:"cap-height",className:"class",clipPath:"clip-path",clipRule:"clip-rule",colorInterpolation:"color-interpolation",colorInterpolationFilters:"color-interpolation-filters",colorProfile:"color-profile",colorRendering:"color-rendering",crossOrigin:"crossorigin",dataType:"datatype",dominantBaseline:"dominant-baseline",enableBackground:"enable-background",fillOpacity:"fill-opacity",fillRule:"fill-rule",floodColor:"flood-color",floodOpacity:"flood-opacity",fontFamily:"font-family",fontSize:"font-size",fontSizeAdjust:"font-size-adjust",fontStretch:"font-stretch",fontStyle:"font-style",fontVariant:"font-variant",fontWeight:"font-weight",glyphName:"glyph-name",glyphOrientationHorizontal:"glyph-orientation-horizontal",glyphOrientationVertical:"glyph-orientation-vertical",hrefLang:"hreflang",horizAdvX:"horiz-adv-x",horizOriginX:"horiz-origin-x",horizOriginY:"horiz-origin-y",imageRendering:"image-rendering",letterSpacing:"letter-spacing",lightingColor:"lighting-color",markerEnd:"marker-end",markerMid:"marker-mid",markerStart:"marker-start",navDown:"nav-down",navDownLeft:"nav-down-left",navDownRight:"nav-down-right",navLeft:"nav-left",navNext:"nav-next",navPrev:"nav-prev",navRight:"nav-right",navUp:"nav-up",navUpLeft:"nav-up-left",navUpRight:"nav-up-right",onAbort:"onabort",onActivate:"onactivate",onAfterPrint:"onafterprint",onBeforePrint:"onbeforeprint",onBegin:"onbegin",onCancel:"oncancel",onCanPlay:"oncanplay",onCanPlayThrough:"oncanplaythrough",onChange:"onchange",onClick:"onclick",onClose:"onclose",onCopy:"oncopy",onCueChange:"oncuechange",onCut:"oncut",onDblClick:"ondblclick",onDrag:"ondrag",onDragEnd:"ondragend",onDragEnter:"ondragenter",onDragExit:"ondragexit",onDragLeave:"ondragleave",onDragOver:"ondragover",onDragStart:"ondragstart",onDrop:"ondrop",onDurationChange:"ondurationchange",onEmptied:"onemptied",onEnd:"onend",onEnded:"onended",onError:"onerror",onFocus:"onfocus",onFocusIn:"onfocusin",onFocusOut:"onfocusout",onHashChange:"onhashchange",onInput:"oninput",onInvalid:"oninvalid",onKeyDown:"onkeydown",onKeyPress:"onkeypress",onKeyUp:"onkeyup",onLoad:"onload",onLoadedData:"onloadeddata",onLoadedMetadata:"onloadedmetadata",onLoadStart:"onloadstart",onMessage:"onmessage",onMouseDown:"onmousedown",onMouseEnter:"onmouseenter",onMouseLeave:"onmouseleave",onMouseMove:"onmousemove",onMouseOut:"onmouseout",onMouseOver:"onmouseover",onMouseUp:"onmouseup",onMouseWheel:"onmousewheel",onOffline:"onoffline",onOnline:"ononline",onPageHide:"onpagehide",onPageShow:"onpageshow",onPaste:"onpaste",onPause:"onpause",onPlay:"onplay",onPlaying:"onplaying",onPopState:"onpopstate",onProgress:"onprogress",onRateChange:"onratechange",onRepeat:"onrepeat",onReset:"onreset",onResize:"onresize",onScroll:"onscroll",onSeeked:"onseeked",onSeeking:"onseeking",onSelect:"onselect",onShow:"onshow",onStalled:"onstalled",onStorage:"onstorage",onSubmit:"onsubmit",onSuspend:"onsuspend",onTimeUpdate:"ontimeupdate",onToggle:"ontoggle",onUnload:"onunload",onVolumeChange:"onvolumechange",onWaiting:"onwaiting",onZoom:"onzoom",overlinePosition:"overline-position",overlineThickness:"overline-thickness",paintOrder:"paint-order",panose1:"panose-1",pointerEvents:"pointer-events",referrerPolicy:"referrerpolicy",renderingIntent:"rendering-intent",shapeRendering:"shape-rendering",stopColor:"stop-color",stopOpacity:"stop-opacity",strikethroughPosition:"strikethrough-position",strikethroughThickness:"strikethrough-thickness",strokeDashArray:"stroke-dasharray",strokeDashOffset:"stroke-dashoffset",strokeLineCap:"stroke-linecap",strokeLineJoin:"stroke-linejoin",strokeMiterLimit:"stroke-miterlimit",strokeOpacity:"stroke-opacity",strokeWidth:"stroke-width",tabIndex:"tabindex",textAnchor:"text-anchor",textDecoration:"text-decoration",textRendering:"text-rendering",transformOrigin:"transform-origin",typeOf:"typeof",underlinePosition:"underline-position",underlineThickness:"underline-thickness",unicodeBidi:"unicode-bidi",unicodeRange:"unicode-range",unitsPerEm:"units-per-em",vAlphabetic:"v-alphabetic",vHanging:"v-hanging",vIdeographic:"v-ideographic",vMathematical:"v-mathematical",vectorEffect:"vector-effect",vertAdvY:"vert-adv-y",vertOriginX:"vert-origin-x",vertOriginY:"vert-origin-y",wordSpacing:"word-spacing",writingMode:"writing-mode",xHeight:"x-height",playbackOrder:"playbackorder",timelineBegin:"timelinebegin"},properties:{about:vt,accentHeight:mt,accumulate:null,additive:null,alignmentBaseline:null,alphabetic:mt,amplitude:mt,arabicForm:null,ascent:mt,attributeName:null,attributeType:null,azimuth:mt,bandwidth:null,baselineShift:null,baseFrequency:null,baseProfile:null,bbox:null,begin:null,bias:mt,by:null,calcMode:null,capHeight:mt,className:yt,clip:null,clipPath:null,clipPathUnits:null,clipRule:null,color:null,colorInterpolation:null,colorInterpolationFilters:null,colorProfile:null,colorRendering:null,content:null,contentScriptType:null,contentStyleType:null,crossOrigin:null,cursor:null,cx:null,cy:null,d:null,dataType:null,defaultAction:null,descent:mt,diffuseConstant:mt,direction:null,display:null,dur:null,divisor:mt,dominantBaseline:null,download:ft,dx:null,dy:null,edgeMode:null,editable:null,elevation:mt,enableBackground:null,end:null,event:null,exponent:mt,externalResourcesRequired:null,fill:null,fillOpacity:mt,fillRule:null,filter:null,filterRes:null,filterUnits:null,floodColor:null,floodOpacity:null,focusable:null,focusHighlight:null,fontFamily:null,fontSize:null,fontSizeAdjust:null,fontStretch:null,fontStyle:null,fontVariant:null,fontWeight:null,format:null,fr:null,from:null,fx:null,fy:null,g1:bt,g2:bt,glyphName:bt,glyphOrientationHorizontal:null,glyphOrientationVertical:null,glyphRef:null,gradientTransform:null,gradientUnits:null,handler:null,hanging:mt,hatchContentUnits:null,hatchUnits:null,height:null,href:null,hrefLang:null,horizAdvX:mt,horizOriginX:mt,horizOriginY:mt,id:null,ideographic:mt,imageRendering:null,initialVisibility:null,in:null,in2:null,intercept:mt,k:mt,k1:mt,k2:mt,k3:mt,k4:mt,kernelMatrix:vt,kernelUnitLength:null,keyPoints:null,keySplines:null,keyTimes:null,kerning:null,lang:null,lengthAdjust:null,letterSpacing:null,lightingColor:null,limitingConeAngle:mt,local:null,markerEnd:null,markerMid:null,markerStart:null,markerHeight:null,markerUnits:null,markerWidth:null,mask:null,maskContentUnits:null,maskUnits:null,mathematical:null,max:null,media:null,mediaCharacterEncoding:null,mediaContentEncodings:null,mediaSize:mt,mediaTime:null,method:null,min:null,mode:null,name:null,navDown:null,navDownLeft:null,navDownRight:null,navLeft:null,navNext:null,navPrev:null,navRight:null,navUp:null,navUpLeft:null,navUpRight:null,numOctaves:null,observer:null,offset:null,onAbort:null,onActivate:null,onAfterPrint:null,onBeforePrint:null,onBegin:null,onCancel:null,onCanPlay:null,onCanPlayThrough:null,onChange:null,onClick:null,onClose:null,onCopy:null,onCueChange:null,onCut:null,onDblClick:null,onDrag:null,onDragEnd:null,onDragEnter:null,onDragExit:null,onDragLeave:null,onDragOver:null,onDragStart:null,onDrop:null,onDurationChange:null,onEmptied:null,onEnd:null,onEnded:null,onError:null,onFocus:null,onFocusIn:null,onFocusOut:null,onHashChange:null,onInput:null,onInvalid:null,onKeyDown:null,onKeyPress:null,onKeyUp:null,onLoad:null,onLoadedData:null,onLoadedMetadata:null,onLoadStart:null,onMessage:null,onMouseDown:null,onMouseEnter:null,onMouseLeave:null,onMouseMove:null,onMouseOut:null,onMouseOver:null,onMouseUp:null,onMouseWheel:null,onOffline:null,onOnline:null,onPageHide:null,onPageShow:null,onPaste:null,onPause:null,onPlay:null,onPlaying:null,onPopState:null,onProgress:null,onRateChange:null,onRepeat:null,onReset:null,onResize:null,onScroll:null,onSeeked:null,onSeeking:null,onSelect:null,onShow:null,onStalled:null,onStorage:null,onSubmit:null,onSuspend:null,onTimeUpdate:null,onToggle:null,onUnload:null,onVolumeChange:null,onWaiting:null,onZoom:null,opacity:null,operator:null,order:null,orient:null,orientation:null,origin:null,overflow:null,overlay:null,overlinePosition:mt,overlineThickness:mt,paintOrder:null,panose1:null,path:null,pathLength:mt,patternContentUnits:null,patternTransform:null,patternUnits:null,phase:null,ping:yt,pitch:null,playbackOrder:null,pointerEvents:null,points:null,pointsAtX:mt,pointsAtY:mt,pointsAtZ:mt,preserveAlpha:null,preserveAspectRatio:null,primitiveUnits:null,propagate:null,property:vt,r:null,radius:null,referrerPolicy:null,refX:null,refY:null,rel:vt,rev:vt,renderingIntent:null,repeatCount:null,repeatDur:null,requiredExtensions:vt,requiredFeatures:vt,requiredFonts:vt,requiredFormats:vt,resource:null,restart:null,result:null,rotate:null,rx:null,ry:null,scale:null,seed:null,shapeRendering:null,side:null,slope:null,snapshotTime:null,specularConstant:mt,specularExponent:mt,spreadMethod:null,spacing:null,startOffset:null,stdDeviation:null,stemh:null,stemv:null,stitchTiles:null,stopColor:null,stopOpacity:null,strikethroughPosition:mt,strikethroughThickness:mt,string:null,stroke:null,strokeDashArray:vt,strokeDashOffset:null,strokeLineCap:null,strokeLineJoin:null,strokeMiterLimit:mt,strokeOpacity:mt,strokeWidth:null,style:null,surfaceScale:mt,syncBehavior:null,syncBehaviorDefault:null,syncMaster:null,syncTolerance:null,syncToleranceDefault:null,systemLanguage:vt,tabIndex:mt,tableValues:null,target:null,targetX:mt,targetY:mt,textAnchor:null,textDecoration:null,textRendering:null,textLength:null,timelineBegin:null,title:null,transformBehavior:null,type:null,typeOf:vt,to:null,transform:null,transformOrigin:null,u1:null,u2:null,underlinePosition:mt,underlineThickness:mt,unicode:null,unicodeBidi:null,unicodeRange:null,unitsPerEm:mt,values:null,vAlphabetic:mt,vMathematical:mt,vectorEffect:null,vHanging:mt,vIdeographic:mt,version:null,vertAdvY:mt,vertOriginX:mt,vertOriginY:mt,viewBox:null,viewTarget:null,visibility:null,width:null,widths:null,wordSpacing:null,writingMode:null,x:null,x1:null,x2:null,xChannelSelector:null,xHeight:mt,y:null,y1:null,y2:null,yChannelSelector:null,z:null,zoomAndPan:null},space:"svg",transform:Et}),Lt=Ct({properties:{xLinkActuate:null,xLinkArcRole:null,xLinkHref:null,xLinkRole:null,xLinkShow:null,xLinkTitle:null,xLinkType:null},space:"xlink",transform:(e,t)=>"xlink:"+t.slice(5).toLowerCase()}),Tt=Ct({attributes:{xmlnsxlink:"xmlns:xlink"},properties:{xmlnsXLink:null,xmlns:null},space:"xmlns",transform:Nt}),Rt=Ct({properties:{xmlBase:null,xmlLang:null,xmlSpace:null},space:"xml",transform:(e,t)=>"xml:"+t.slice(3).toLowerCase()}),It=ut([_t,Pt,Lt,Tt,Rt],"html"),Mt=ut([_t,At,Lt,Tt,Rt],"svg"),Ot={}.hasOwnProperty,jt=/["&'<>`]/g,zt=/[\uD800-\uDBFF][\uDC00-\uDFFF]/g,$t=/[\x01-\t\v\f\x0E-\x1F\x7F\x81\x8D\x8F\x90\x9D\xA0-\uFFFF]/g,Ft=/[|\\{}()[\]^$+*?.]/g,Dt=new WeakMap;const Bt=/[\dA-Fa-f]/,Ut=/\d/,qt=["AElig","AMP","Aacute","Acirc","Agrave","Aring","Atilde","Auml","COPY","Ccedil","ETH","Eacute","Ecirc","Egrave","Euml","GT","Iacute","Icirc","Igrave","Iuml","LT","Ntilde","Oacute","Ocirc","Ograve","Oslash","Otilde","Ouml","QUOT","REG","THORN","Uacute","Ucirc","Ugrave","Uuml","Yacute","aacute","acirc","acute","aelig","agrave","amp","aring","atilde","auml","brvbar","ccedil","cedil","cent","copy","curren","deg","divide","eacute","ecirc","egrave","eth","euml","frac12","frac14","frac34","gt","iacute","icirc","iexcl","igrave","iquest","iuml","laquo","lt","macr","micro","middot","nbsp","not","ntilde","oacute","ocirc","ograve","ordf","ordm","oslash","otilde","ouml","para","plusmn","pound","quot","raquo","reg","sect","shy","sup1","sup2","sup3","szlig","thorn","times","uacute","ucirc","ugrave","uml","uuml","yacute","yen","yuml"],Gt={nbsp:" ",iexcl:"¡",cent:"¢",pound:"£",curren:"¤",yen:"¥",brvbar:"¦",sect:"§",uml:"¨",copy:"©",ordf:"ª",laquo:"«",not:"¬",shy:"­",reg:"®",macr:"¯",deg:"°",plusmn:"±",sup2:"²",sup3:"³",acute:"´",micro:"µ",para:"¶",middot:"·",cedil:"¸",sup1:"¹",ordm:"º",raquo:"»",frac14:"¼",frac12:"½",frac34:"¾",iquest:"¿",Agrave:"À",Aacute:"Á",Acirc:"Â",Atilde:"Ã",Auml:"Ä",Aring:"Å",AElig:"Æ",Ccedil:"Ç",Egrave:"È",Eacute:"É",Ecirc:"Ê",Euml:"Ë",Igrave:"Ì",Iacute:"Í",Icirc:"Î",Iuml:"Ï",ETH:"Ð",Ntilde:"Ñ",Ograve:"Ò",Oacute:"Ó",Ocirc:"Ô",Otilde:"Õ",Ouml:"Ö",times:"×",Oslash:"Ø",Ugrave:"Ù",Uacute:"Ú",Ucirc:"Û",Uuml:"Ü",Yacute:"Ý",THORN:"Þ",szlig:"ß",agrave:"à",aacute:"á",acirc:"â",atilde:"ã",auml:"ä",aring:"å",aelig:"æ",ccedil:"ç",egrave:"è",eacute:"é",ecirc:"ê",euml:"ë",igrave:"ì",iacute:"í",icirc:"î",iuml:"ï",eth:"ð",ntilde:"ñ",ograve:"ò",oacute:"ó",ocirc:"ô",otilde:"õ",ouml:"ö",divide:"÷",oslash:"ø",ugrave:"ù",uacute:"ú",ucirc:"û",uuml:"ü",yacute:"ý",thorn:"þ",yuml:"ÿ",fnof:"ƒ",Alpha:"Α",Beta:"Β",Gamma:"Γ",Delta:"Δ",Epsilon:"Ε",Zeta:"Ζ",Eta:"Η",Theta:"Θ",Iota:"Ι",Kappa:"Κ",Lambda:"Λ",Mu:"Μ",Nu:"Ν",Xi:"Ξ",Omicron:"Ο",Pi:"Π",Rho:"Ρ",Sigma:"Σ",Tau:"Τ",Upsilon:"Υ",Phi:"Φ",Chi:"Χ",Psi:"Ψ",Omega:"Ω",alpha:"α",beta:"β",gamma:"γ",delta:"δ",epsilon:"ε",zeta:"ζ",eta:"η",theta:"θ",iota:"ι",kappa:"κ",lambda:"λ",mu:"μ",nu:"ν",xi:"ξ",omicron:"ο",pi:"π",rho:"ρ",sigmaf:"ς",sigma:"σ",tau:"τ",upsilon:"υ",phi:"φ",chi:"χ",psi:"ψ",omega:"ω",thetasym:"ϑ",upsih:"ϒ",piv:"ϖ",bull:"•",hellip:"…",prime:"′",Prime:"″",oline:"‾",frasl:"⁄",weierp:"℘",image:"ℑ",real:"ℜ",trade:"™",alefsym:"ℵ",larr:"←",uarr:"↑",rarr:"→",darr:"↓",harr:"↔",crarr:"↵",lArr:"⇐",uArr:"⇑",rArr:"⇒",dArr:"⇓",hArr:"⇔",forall:"∀",part:"∂",exist:"∃",empty:"∅",nabla:"∇",isin:"∈",notin:"∉",ni:"∋",prod:"∏",sum:"∑",minus:"−",lowast:"∗",radic:"√",prop:"∝",infin:"∞",ang:"∠",and:"∧",or:"∨",cap:"∩",cup:"∪",int:"∫",there4:"∴",sim:"∼",cong:"≅",asymp:"≈",ne:"≠",equiv:"≡",le:"≤",ge:"≥",sub:"⊂",sup:"⊃",nsub:"⊄",sube:"⊆",supe:"⊇",oplus:"⊕",otimes:"⊗",perp:"⊥",sdot:"⋅",lceil:"⌈",rceil:"⌉",lfloor:"⌊",rfloor:"⌋",lang:"〈",rang:"〉",loz:"◊",spades:"♠",clubs:"♣",hearts:"♥",diams:"♦",quot:'"',amp:"&",lt:"<",gt:">",OElig:"Œ",oelig:"œ",Scaron:"Š",scaron:"š",Yuml:"Ÿ",circ:"ˆ",tilde:"˜",ensp:" ",emsp:" ",thinsp:" ",zwnj:"‌",zwj:"‍",lrm:"‎",rlm:"‏",ndash:"–",mdash:"—",lsquo:"‘",rsquo:"’",sbquo:"‚",ldquo:"“",rdquo:"”",bdquo:"„",dagger:"†",Dagger:"‡",permil:"‰",lsaquo:"‹",rsaquo:"›",euro:"€"},Wt=["cent","copy","divide","gt","lt","not","para","times"],Ht={}.hasOwnProperty,Vt={};let Qt;for(Qt in Gt)Ht.call(Gt,Qt)&&(Vt[Gt[Qt]]=Qt);const Xt=/[^\dA-Za-z]/;function Kt(e,t,n){let r,a=function(e,t,n){const r="&#x"+e.toString(16).toUpperCase();return n&&t&&!Bt.test(String.fromCharCode(t))?r:r+";"}(e,t,n.omitOptionalSemicolons);if((n.useNamedReferences||n.useShortestReferences)&&(r=function(e,t,n,r){const a=String.fromCharCode(e);if(Ht.call(Vt,a)){const e=Vt[a],o="&"+e;return n&&qt.includes(e)&&!Wt.includes(e)&&(!r||t&&61!==t&&Xt.test(String.fromCharCode(t)))?o:o+";"}return""}(e,t,n.omitOptionalSemicolons,n.attribute)),(n.useShortestReferences||!r)&&n.useShortestReferences){const r=function(e,t,n){const r="&#"+String(e);return n&&t&&!Ut.test(String.fromCharCode(t))?r:r+";"}(e,t,n.omitOptionalSemicolons);r.length<a.length&&(a=r)}return r&&(!n.useShortestReferences||r.length<a.length)?r:a}function Yt(e,t){return function(e,t){return e=e.replace(t.subset?function(e){let t=Dt.get(e);return t||(t=function(e){const t=[];let n=-1;for(;++n<e.length;)t.push(e[n].replace(Ft,"\\$&"));return new RegExp("(?:"+t.join("|")+")","g")}(e),Dt.set(e,t)),t}(t.subset):jt,n),t.subset||t.escapeOnly?e:e.replace(zt,(function(e,n,r){return t.format(1024*(e.charCodeAt(0)-55296)+e.charCodeAt(1)-56320+65536,r.charCodeAt(n+2),t)})).replace($t,n);function n(e,n,r){return t.format(e.charCodeAt(0),r.charCodeAt(n+1),t)}}(e,Object.assign({format:Kt},t))}const Zt=/^>|^->|<!--|-->|--!>|<!-$/g,Jt=[">"],en=["<",">"];function tn(e,t){const n=String(e);if("string"!=typeof t)throw new TypeError("Expected character");let r=0,a=n.indexOf(t);for(;-1!==a;)r++,a=n.indexOf(t,a+t.length);return r}function nn(e,t){const n=t||{};return(""===e[e.length-1]?[...e,""]:e).join((n.padRight?" ":"")+","+(!1===n.padLeft?"":" ")).trim()}const rn=/[A-Z]/g,an=/-[a-z]/g,on=/^data[-\w.:]+$/i;function ln(e){return"-"+e.toLowerCase()}function sn(e){return e.charAt(1).toUpperCase()}function un(e){return e.join(" ").trim()}const cn=/[ \t\n\f\r]/g;function pn(e){return"object"==typeof e?"text"===e.type&&dn(e.value):dn(e)}function dn(e){return""===e.replace(cn,"")}const fn=mn(1),hn=mn(-1),gn=[];function mn(e){return function(t,n,r){const a=t?t.children:gn;let o=(n||0)+e,i=a[o];if(!r)for(;i&&pn(i);)o+=e,i=a[o];return i}}const yn={}.hasOwnProperty;function bn(e){return function(t,n,r){return yn.call(e,t.tagName)&&e[t.tagName](t,n,r)}}const vn=bn({body:function(e,t,n){const r=fn(n,t);return!r||"comment"!==r.type},caption:wn,colgroup:wn,dd:function(e,t,n){const r=fn(n,t);return!r||"element"===r.type&&("dt"===r.tagName||"dd"===r.tagName)},dt:function(e,t,n){const r=fn(n,t);return Boolean(r&&"element"===r.type&&("dt"===r.tagName||"dd"===r.tagName))},head:wn,html:function(e,t,n){const r=fn(n,t);return!r||"comment"!==r.type},li:function(e,t,n){const r=fn(n,t);return!r||"element"===r.type&&"li"===r.tagName},optgroup:function(e,t,n){const r=fn(n,t);return!r||"element"===r.type&&"optgroup"===r.tagName},option:function(e,t,n){const r=fn(n,t);return!r||"element"===r.type&&("option"===r.tagName||"optgroup"===r.tagName)},p:function(e,t,n){const r=fn(n,t);return r?"element"===r.type&&("address"===r.tagName||"article"===r.tagName||"aside"===r.tagName||"blockquote"===r.tagName||"details"===r.tagName||"div"===r.tagName||"dl"===r.tagName||"fieldset"===r.tagName||"figcaption"===r.tagName||"figure"===r.tagName||"footer"===r.tagName||"form"===r.tagName||"h1"===r.tagName||"h2"===r.tagName||"h3"===r.tagName||"h4"===r.tagName||"h5"===r.tagName||"h6"===r.tagName||"header"===r.tagName||"hgroup"===r.tagName||"hr"===r.tagName||"main"===r.tagName||"menu"===r.tagName||"nav"===r.tagName||"ol"===r.tagName||"p"===r.tagName||"pre"===r.tagName||"section"===r.tagName||"table"===r.tagName||"ul"===r.tagName):!n||!("element"===n.type&&("a"===n.tagName||"audio"===n.tagName||"del"===n.tagName||"ins"===n.tagName||"map"===n.tagName||"noscript"===n.tagName||"video"===n.tagName))},rp:xn,rt:xn,tbody:function(e,t,n){const r=fn(n,t);return!r||"element"===r.type&&("tbody"===r.tagName||"tfoot"===r.tagName)},td:kn,tfoot:function(e,t,n){return!fn(n,t)},th:kn,thead:function(e,t,n){const r=fn(n,t);return Boolean(r&&"element"===r.type&&("tbody"===r.tagName||"tfoot"===r.tagName))},tr:function(e,t,n){const r=fn(n,t);return!r||"element"===r.type&&"tr"===r.tagName}});function wn(e,t,n){const r=fn(n,t,!0);return!r||"comment"!==r.type&&!("text"===r.type&&pn(r.value.charAt(0)))}function xn(e,t,n){const r=fn(n,t);return!r||"element"===r.type&&("rp"===r.tagName||"rt"===r.tagName)}function kn(e,t,n){const r=fn(n,t);return!r||"element"===r.type&&("td"===r.tagName||"th"===r.tagName)}const Sn=bn({body:function(e){const t=fn(e,-1,!0);return!(t&&("comment"===t.type||"text"===t.type&&pn(t.value.charAt(0))||"element"===t.type&&("meta"===t.tagName||"link"===t.tagName||"script"===t.tagName||"style"===t.tagName||"template"===t.tagName)))},colgroup:function(e,t,n){const r=hn(n,t),a=fn(e,-1,!0);return!(n&&r&&"element"===r.type&&"colgroup"===r.tagName&&vn(r,n.children.indexOf(r),n))&&Boolean(a&&"element"===a.type&&"col"===a.tagName)},head:function(e){const t=new Set;for(const n of e.children)if("element"===n.type&&("base"===n.tagName||"title"===n.tagName)){if(t.has(n.tagName))return!1;t.add(n.tagName)}const n=e.children[0];return!n||"element"===n.type},html:function(e){const t=fn(e,-1);return!t||"comment"!==t.type},tbody:function(e,t,n){const r=hn(n,t),a=fn(e,-1);return(!n||!r||"element"!==r.type||"thead"!==r.tagName&&"tbody"!==r.tagName||!vn(r,n.children.indexOf(r),n))&&Boolean(a&&"element"===a.type&&"tr"===a.tagName)}}),Cn={name:[["\t\n\f\r &/=>".split(""),"\t\n\f\r \"&'/=>`".split("")],["\0\t\n\f\r \"&'/<=>".split(""),"\0\t\n\f\r \"&'/<=>`".split("")]],unquoted:[["\t\n\f\r &>".split(""),"\0\t\n\f\r \"&'<=>`".split("")],["\0\t\n\f\r \"&'<=>`".split(""),"\0\t\n\f\r \"&'<=>`".split("")]],single:[["&'".split(""),"\"&'`".split("")],["\0&'".split(""),"\0\"&'`".split("")]],double:[['"&'.split(""),"\"&'`".split("")],['\0"&'.split(""),"\0\"&'`".split("")]]};function _n(e,t,n){const r=function(e,t){const n=ct(t);let r=t,a=pt;if(n in e.normal)return e.property[e.normal[n]];if(n.length>4&&"data"===n.slice(0,4)&&on.test(t)){if("-"===t.charAt(4)){const e=t.slice(5).replace(an,sn);r="data"+e.charAt(0).toUpperCase()+e.slice(1)}else{const e=t.slice(4);if(!an.test(e)){let n=e.replace(rn,ln);"-"!==n.charAt(0)&&(n="-"+n),t="data"+n}}a=kt}return new a(r,t)}(e.schema,t),a=e.settings.allowParseErrors&&"html"===e.schema.space?0:1,o=e.settings.allowDangerousCharacters?0:1;let i,l=e.quote;if(!r.overloadedBoolean||n!==r.attribute&&""!==n?!r.boolean&&!r.overloadedBoolean||"string"==typeof n&&n!==r.attribute&&""!==n||(n=Boolean(n)):n=!0,null==n||!1===n||"number"==typeof n&&Number.isNaN(n))return"";const s=Yt(r.attribute,Object.assign({},e.settings.characterReferences,{subset:Cn.name[a][o]}));return!0===n?s:(n=Array.isArray(n)?(r.commaSeparated?nn:un)(n,{padLeft:!e.settings.tightCommaSeparatedLists}):String(n),e.settings.collapseEmptyAttributes&&!n?s:(e.settings.preferUnquoted&&(i=Yt(n,Object.assign({},e.settings.characterReferences,{attribute:!0,subset:Cn.unquoted[a][o]}))),i!==n&&(e.settings.quoteSmart&&tn(n,l)>tn(n,e.alternative)&&(l=e.alternative),i=l+Yt(n,Object.assign({},e.settings.characterReferences,{subset:("'"===l?Cn.single:Cn.double)[a][o],attribute:!0}))+l),s+(i?"="+i:i)))}const En=["<","&"];function Nn(e,t,n,r){return!n||"element"!==n.type||"script"!==n.tagName&&"style"!==n.tagName?Yt(e.value,Object.assign({},r.settings.characterReferences,{subset:En})):e.value}const Pn=function(e,t){const n=t||{};function r(t,...n){let a=r.invalid;const o=r.handlers;if(t&&Ot.call(t,e)){const n=String(t[e]);a=Ot.call(o,n)?o[n]:r.unknown}if(a)return a.call(this,t,...n)}return r.handlers=n.handlers||{},r.invalid=n.invalid,r.unknown=n.unknown,r}("type",{invalid:function(e){throw new Error("Expected node, not `"+e+"`")},unknown:function(e){throw new Error("Cannot compile unknown node `"+e.type+"`")},handlers:{comment:function(e,t,n,r){return r.settings.bogusComments?"<?"+Yt(e.value,Object.assign({},r.settings.characterReferences,{subset:Jt}))+">":"\x3c!--"+e.value.replace(Zt,(function(e){return Yt(e,Object.assign({},r.settings.characterReferences,{subset:en}))}))+"--\x3e"},doctype:function(e,t,n,r){return"<!"+(r.settings.upperDoctype?"DOCTYPE":"doctype")+(r.settings.tightDoctype?"":" ")+"html>"},element:function(e,t,n,r){const a=r.schema,o="svg"!==a.space&&r.settings.omitOptionalTags;let i="svg"===a.space?r.settings.closeEmptyElements:r.settings.voids.includes(e.tagName.toLowerCase());const l=[];let s;"html"===a.space&&"svg"===e.tagName&&(r.schema=Mt);const u=function(e,t){const n=[];let r,a=-1;if(t)for(r in t)if(null!==t[r]&&void 0!==t[r]){const a=_n(e,r,t[r]);a&&n.push(a)}for(;++a<n.length;){const t=e.settings.tightAttributes?n[a].charAt(n[a].length-1):void 0;a!==n.length-1&&'"'!==t&&"'"!==t&&(n[a]+=" ")}return n.join("")}(r,e.properties),c=r.all("html"===a.space&&"template"===e.tagName?e.content:e);return r.schema=a,c&&(i=!1),!u&&o&&Sn(e,t,n)||(l.push("<",e.tagName,u?" "+u:""),i&&("svg"===a.space||r.settings.closeSelfClosing)&&(s=u.charAt(u.length-1),(!r.settings.tightSelfClosing||"/"===s||s&&'"'!==s&&"'"!==s)&&l.push(" "),l.push("/")),l.push(">")),l.push(c),i||o&&vn(e,t,n)||l.push("</"+e.tagName+">"),l.join("")},raw:function(e,t,n,r){return r.settings.allowDangerousHtml?e.value:Nn(e,0,n,r)},root:function(e,t,n,r){return r.all(e)},text:Nn}}),An={},Ln={},Tn=[];function Rn(e,t,n){return Pn(e,t,n,this)}function In(e){const t=[],n=e&&e.children||Tn;let r=-1;for(;++r<n.length;)t[r]=this.one(n[r],r,e);return t.join("")}function Mn(e,t){const n="string"==typeof e?{}:{...e.colorReplacements},r="string"==typeof e?e:e.name;for(const[e,a]of Object.entries(t?.colorReplacements||{}))"string"==typeof a?n[e]=a:e===r&&Object.assign(n,a);return n}function On(e,t){return e&&t?.[e?.toLowerCase()]||e}async function jn(e){return Promise.resolve("function"==typeof e?e():e).then((e=>e.default||e))}function zn(e){return!e||["plaintext","txt","text","plain"].includes(e)}function $n(e){return"none"===e}function Fn(e,t){if(!t)return e;e.properties||={},e.properties.class||=[],"string"==typeof e.properties.class&&(e.properties.class=e.properties.class.split(/\s+/g)),Array.isArray(e.properties.class)||(e.properties.class=[]);const n=Array.isArray(t)?t:t.split(/\s+/g);for(const t of n)t&&!e.properties.class.includes(t)&&e.properties.class.push(t);return e}function Dn(e,t=!1){const n=e.split(/(\r?\n)/g);let r=0;const a=[];for(let e=0;e<n.length;e+=2){const o=t?n[e]+(n[e+1]||""):n[e];a.push([o,r]),r+=n[e].length,r+=n[e+1]?.length||0}return a}function Bn(e){const t={};return e.color&&(t.color=e.color),e.bgColor&&(t["background-color"]=e.bgColor),e.fontStyle&&(e.fontStyle&oe.Italic&&(t["font-style"]="italic"),e.fontStyle&oe.Bold&&(t["font-weight"]="bold"),e.fontStyle&oe.Underline&&(t["text-decoration"]="underline")),t}function Un(e){return"string"==typeof e?e:Object.entries(e).map((([e,t])=>`${e}:${t}`)).join(";")}const qn=new WeakMap;function Gn(e,t){qn.set(e,t)}function Wn(e){return qn.get(e)}class Hn{_stacks={};lang;get themes(){return Object.keys(this._stacks)}get theme(){return this.themes[0]}get _stack(){return this._stacks[this.theme]}static initial(e,t){return new Hn(Object.fromEntries((n=t,Array.isArray(n)?n:[n]).map((e=>[e,it]))),e);var n}constructor(...e){if(2===e.length){const[t,n]=e;this.lang=n,this._stacks=t}else{const[t,n,r]=e;this.lang=n,this._stacks={[r]:t}}}getInternalStack(e=this.theme){return this._stacks[e]}getScopes(e=this.theme){return function(e){const t=[],n=new Set;return function e(r){if(n.has(r))return;n.add(r);const a=r?.nameScopesList?.scopeName;a&&t.push(a),r.parent&&e(r.parent)}(e),t}(this._stacks[e])}toJSON(){return{lang:this.lang,theme:this.theme,themes:this.themes,scopes:this.getScopes()}}}function Vn(){const e=new WeakMap;function t(t){if(!e.has(t.meta)){let n=function(e){if("number"==typeof e){if(e<0||e>t.source.length)throw new q(`Invalid decoration offset: ${e}. Code length: ${t.source.length}`);return{...r.indexToPos(e),offset:e}}{const t=r.lines[e.line];if(void 0===t)throw new q(`Invalid decoration position ${JSON.stringify(e)}. Lines length: ${r.lines.length}`);if(e.character<0||e.character>t.length)throw new q(`Invalid decoration position ${JSON.stringify(e)}. Line ${e.line} length: ${t.length}`);return{...e,offset:r.posToIndex(e.line,e.character)}}};const r=function(e){const t=Dn(e,!0).map((([e])=>e));return{lines:t,indexToPos:function(n){if(n===e.length)return{line:t.length-1,character:t[t.length-1].length};let r=n,a=0;for(const e of t){if(r<e.length)break;r-=e.length,a++}return{line:a,character:r}},posToIndex:function(e,n){let r=0;for(let n=0;n<e;n++)r+=t[n].length;return r+=n,r}}}(t.source),a=(t.options.decorations||[]).map((e=>({...e,start:n(e.start),end:n(e.end)})));!function(e){for(let t=0;t<e.length;t++){const n=e[t];if(n.start.offset>n.end.offset)throw new q(`Invalid decoration range: ${JSON.stringify(n.start)} - ${JSON.stringify(n.end)}`);for(let r=t+1;r<e.length;r++){const t=e[r],a=n.start.offset<=t.start.offset&&t.start.offset<n.end.offset,o=n.start.offset<t.end.offset&&t.end.offset<=n.end.offset,i=t.start.offset<=n.start.offset&&n.start.offset<t.end.offset,l=t.start.offset<n.end.offset&&n.end.offset<=t.end.offset;if(a||o||i||l){if(a&&o)continue;if(i&&l)continue;throw new q(`Decorations ${JSON.stringify(n.start)} and ${JSON.stringify(t.start)} intersect.`)}}}}(a),e.set(t.meta,{decorations:a,converter:r,source:t.source})}return e.get(t.meta)}return{name:"shiki:decorations",tokens(e){if(!this.options.decorations?.length)return;const n=function(e,t){const n=Array.from(t instanceof Set?t:new Set(t)).sort(((e,t)=>e-t));return n.length?e.map((e=>e.flatMap((e=>{const t=n.filter((t=>e.offset<t&&t<e.offset+e.content.length)).map((t=>t-e.offset)).sort(((e,t)=>e-t));return t.length?function(e,t){let n=0;const r=[];for(const a of t)a>n&&r.push({...e,content:e.content.slice(n,a),offset:e.offset+n}),n=a;return n<e.content.length&&r.push({...e,content:e.content.slice(n),offset:e.offset+n}),r}(e,t):e})))):e}(e,t(this).decorations.flatMap((e=>[e.start.offset,e.end.offset])));return n},code(e){if(!this.options.decorations?.length)return;const n=t(this),r=Array.from(e.children).filter((e=>"element"===e.type&&"span"===e.tagName));if(r.length!==n.converter.lines.length)throw new q(`Number of lines in code element (${r.length}) does not match the number of lines in the source (${n.converter.lines.length}). Failed to apply decorations.`);function a(e,t,n,a){const o=r[e];let l="",s=-1,u=-1;if(0===t&&(s=0),0===n&&(u=0),n===Number.POSITIVE_INFINITY&&(u=o.children.length),-1===s||-1===u)for(let e=0;e<o.children.length;e++)l+=Qn(o.children[e]),-1===s&&l.length===t&&(s=e+1),-1===u&&l.length===n&&(u=e+1);if(-1===s)throw new q(`Failed to find start index for decoration ${JSON.stringify(a.start)}`);if(-1===u)throw new q(`Failed to find end index for decoration ${JSON.stringify(a.end)}`);const c=o.children.slice(s,u);if(a.alwaysWrap||c.length!==o.children.length)if(a.alwaysWrap||1!==c.length||"element"!==c[0].type){const e={type:"element",tagName:"span",properties:{},children:c};i(e,a,"wrapper"),o.children.splice(s,c.length,e)}else i(c[0],a,"token");else i(o,a,"line")}function o(e,t){r[e]=i(r[e],t,"line")}function i(e,t,n){const r=t.properties||{},a=t.transform||(e=>e);return e.tagName=t.tagName||"span",e.properties={...e.properties,...r,class:e.properties.class},t.properties?.class&&Fn(e,t.properties.class),a(e,n)||e}const l=[],s=n.decorations.sort(((e,t)=>t.start.offset-e.start.offset||e.end.offset-t.end.offset));for(const e of s){const{start:t,end:n}=e;if(t.line===n.line)a(t.line,t.character,n.character,e);else if(t.line<n.line){a(t.line,t.character,Number.POSITIVE_INFINITY,e);for(let r=t.line+1;r<n.line;r++)l.unshift((()=>o(r,e)));a(n.line,0,n.character,e)}}l.forEach((e=>e()))}}}function Qn(e){return"text"===e.type?e.value:"element"===e.type?e.children.map(Qn).join(""):""}const Xn=[Vn()];function Kn(e){return[...e.transformers||[],...Xn]}var Yn=["black","red","green","yellow","blue","magenta","cyan","white","brightBlack","brightRed","brightGreen","brightYellow","brightBlue","brightMagenta","brightCyan","brightWhite"],Zn={1:"bold",2:"dim",3:"italic",4:"underline",7:"reverse",8:"hidden",9:"strikethrough"};function Jn(e,t){const n=e.indexOf("",t);if(-1!==n&&"["===e[n+1]){const t=e.indexOf("m",n);if(-1!==t)return{sequence:e.substring(n+2,t).split(";"),startPosition:n,position:t+1}}return{position:e.length}}function er(e){const t=e.shift();if("2"===t){const t=e.splice(0,3).map((e=>Number.parseInt(e)));if(3!==t.length||t.some((e=>Number.isNaN(e))))return;return{type:"rgb",rgb:t}}if("5"===t){const t=e.shift();if(t)return{type:"table",index:Number(t)}}}function tr(e){const t=[];for(;e.length>0;){const n=e.shift();if(!n)continue;const r=Number.parseInt(n);if(!Number.isNaN(r))if(0===r)t.push({type:"resetAll"});else if(r<=9)Zn[r]&&t.push({type:"setDecoration",value:Zn[r]});else if(r<=29){const e=Zn[r-20];e&&(t.push({type:"resetDecoration",value:e}),"dim"===e&&t.push({type:"resetDecoration",value:"bold"}))}else if(r<=37)t.push({type:"setForegroundColor",value:{type:"named",name:Yn[r-30]}});else if(38===r){const n=er(e);n&&t.push({type:"setForegroundColor",value:n})}else if(39===r)t.push({type:"resetForegroundColor"});else if(r<=47)t.push({type:"setBackgroundColor",value:{type:"named",name:Yn[r-40]}});else if(48===r){const n=er(e);n&&t.push({type:"setBackgroundColor",value:n})}else 49===r?t.push({type:"resetBackgroundColor"}):53===r?t.push({type:"setDecoration",value:"overline"}):55===r?t.push({type:"resetDecoration",value:"overline"}):r>=90&&r<=97?t.push({type:"setForegroundColor",value:{type:"named",name:Yn[r-90+8]}}):r>=100&&r<=107&&t.push({type:"setBackgroundColor",value:{type:"named",name:Yn[r-100+8]}})}return t}var nr={black:"#000000",red:"#bb0000",green:"#00bb00",yellow:"#bbbb00",blue:"#0000bb",magenta:"#ff00ff",cyan:"#00bbbb",white:"#eeeeee",brightBlack:"#555555",brightRed:"#ff5555",brightGreen:"#00ff00",brightYellow:"#ffff55",brightBlue:"#5555ff",brightMagenta:"#ff55ff",brightCyan:"#55ffff",brightWhite:"#ffffff"};function rr(e=nr){function t(t){return e[t]}function n(e){return`#${e.map((e=>Math.max(0,Math.min(e,255)).toString(16).padStart(2,"0"))).join("")}`}let r;return{value:function(e){switch(e.type){case"named":return t(e.name);case"rgb":return n(e.rgb);case"table":return a=e.index,function(){if(r)return r;r=[];for(let e=0;e<Yn.length;e++)r.push(t(Yn[e]));let e=[0,95,135,175,215,255];for(let t=0;t<6;t++)for(let a=0;a<6;a++)for(let o=0;o<6;o++)r.push(n([e[t],e[a],e[o]]));let a=8;for(let e=0;e<24;e++,a+=10)r.push(n([a,a,a]));return r}()[a]}var a}}}function ar(e,t,n={}){const{lang:r="text",theme:a=e.getLoadedThemes()[0]}=n;if(zn(r)||$n(a))return Dn(t).map((e=>[{content:e[0],offset:e[1]}]));const{theme:o,colorMap:i}=e.setTheme(a);if("ansi"===r)return function(e,t,n){const r=Mn(e,n),a=Dn(t),o=rr(Object.fromEntries(Yn.map((t=>[t,e.colors?.[`terminal.ansi${t[0].toUpperCase()}${t.substring(1)}`]])))),i=function(){let e=null,t=null,n=new Set;return{parse(r){const a=[];let o=0;do{const i=Jn(r,o),l=i.sequence?r.substring(o,i.startPosition):r.substring(o);if(l.length>0&&a.push({value:l,foreground:e,background:t,decorations:new Set(n)}),i.sequence){const r=tr(i.sequence);for(const a of r)"resetAll"===a.type?(e=null,t=null,n.clear()):"resetForegroundColor"===a.type?e=null:"resetBackgroundColor"===a.type?t=null:"resetDecoration"===a.type&&n.delete(a.value);for(const a of r)"setForegroundColor"===a.type?e=a.value:"setBackgroundColor"===a.type?t=a.value:"setDecoration"===a.type&&n.add(a.value)}o=i.position}while(o<r.length);return a}}}();return a.map((t=>i.parse(t[0]).map((n=>{let a,i;n.decorations.has("reverse")?(a=n.background?o.value(n.background):e.bg,i=n.foreground?o.value(n.foreground):e.fg):(a=n.foreground?o.value(n.foreground):e.fg,i=n.background?o.value(n.background):void 0),a=On(a,r),i=On(i,r),n.decorations.has("dim")&&(a=function(e){const t=e.match(/#([0-9a-f]{3})([0-9a-f]{3})?([0-9a-f]{2})?/);if(t){if(t[3]){const e=Math.round(Number.parseInt(t[3],16)/2).toString(16).padStart(2,"0");return`#${t[1]}${t[2]}${e}`}return t[2]?`#${t[1]}${t[2]}80`:`#${Array.from(t[1]).map((e=>`${e}${e}`)).join("")}80`}const n=e.match(/var\((--[\w-]+-ansi-[\w-]+)\)/);return n?`var(${n[1]}-dim)`:e}(a));let l=oe.None;return n.decorations.has("bold")&&(l|=oe.Bold),n.decorations.has("italic")&&(l|=oe.Italic),n.decorations.has("underline")&&(l|=oe.Underline),{content:n.value,offset:t[1],color:a,bgColor:i,fontStyle:l}}))))}(o,t,n);const l=e.getLanguage(r);if(n.grammarState){if(n.grammarState.lang!==l.name)throw new q(`Grammar state language "${n.grammarState.lang}" does not match highlight language "${l.name}"`);if(!n.grammarState.themes.includes(o.name))throw new q(`Grammar state themes "${n.grammarState.themes}" do not contain highlight theme "${o.name}"`)}return function(e,t,n,r,a){const o=or(e,t,n,r,a),i=new Hn(or(e,t,n,r,a).stateStack,t.name,n.name);return Gn(o.tokens,i),o.tokens}(t,l,o,i,n)}function or(e,t,n,r,a){const o=Mn(n,a),{tokenizeMaxLineLength:i=0,tokenizeTimeLimit:l=500}=a,s=Dn(e);let u=a.grammarState?function(e,t){if(!(e instanceof Hn))throw new q("Invalid grammar state");return e.getInternalStack(t)}(a.grammarState,n.name)??it:null!=a.grammarContextCode?or(a.grammarContextCode,t,n,r,{...a,grammarState:void 0,grammarContextCode:void 0}).stateStack:it,c=[];const p=[];for(let e=0,f=s.length;e<f;e++){const[f,h]=s[e];if(""===f){c=[],p.push([]);continue}if(i>0&&f.length>=i){c=[],p.push([{content:f,offset:h,color:"",fontStyle:0}]);continue}let g,m,y;a.includeExplanation&&(g=t.tokenizeLine(f,u,l),m=g.tokens,y=0);const b=t.tokenizeLine2(f,u,l),v=b.tokens.length/2;for(let e=0;e<v;e++){const t=b.tokens[2*e],i=e+1<v?b.tokens[2*e+2]:f.length;if(t===i)continue;const l=b.tokens[2*e+1],s=On(r[ce.getForeground(l)],o),u=ce.getFontStyle(l),p={content:f.substring(t,i),offset:h+t,color:s,fontStyle:u};if(a.includeExplanation){const e=[];if("scopeName"!==a.includeExplanation)for(const t of n.settings){let n;switch(typeof t.scope){case"string":n=t.scope.split(/,/).map((e=>e.trim()));break;case"object":n=t.scope;break;default:continue}e.push({settings:t,selectors:n.map((e=>e.split(/ /)))})}p.explanation=[];let r=0;for(;t+r<i;){const t=m[y],n=f.substring(t.startIndex,t.endIndex);r+=n.length,p.explanation.push({content:n,scopes:"scopeName"===a.includeExplanation?(d=t.scopes,d.map((e=>({scopeName:e})))):ir(e,t.scopes)}),y+=1}}c.push(p)}p.push(c),c=[],u=b.ruleStack}var d;return{tokens:p,stateStack:u}}function ir(e,t){const n=[];for(let r=0,a=t.length;r<a;r++){const a=t[r];n[r]={scopeName:a,themeMatches:ur(e,a,t.slice(0,r))}}return n}function lr(e,t){return e===t||t.substring(0,e.length)===e&&"."===t[e.length]}function sr(e,t,n){if(!lr(e[e.length-1],t))return!1;let r=e.length-2,a=n.length-1;for(;r>=0&&a>=0;)lr(e[r],n[a])&&(r-=1),a-=1;return-1===r}function ur(e,t,n){const r=[];for(const{selectors:a,settings:o}of e)for(const e of a)if(sr(e,t,n)){r.push(o);break}return r}function cr(e,t,n){const r=Object.entries(n.themes).filter((e=>e[1])).map((e=>({color:e[0],theme:e[1]}))),a=r.map((r=>{const a=ar(e,t,{...n,theme:r.theme});return{tokens:a,state:Wn(a),theme:"string"==typeof r.theme?r.theme:r.theme.name}})),o=function(...e){const t=e.map((()=>[])),n=e.length;for(let r=0;r<e[0].length;r++){const a=e.map((e=>e[r])),o=t.map((()=>[]));t.forEach(((e,t)=>e.push(o[t])));const i=a.map((()=>0)),l=a.map((e=>e[0]));for(;l.every((e=>e));){const e=Math.min(...l.map((e=>e.content.length)));for(let t=0;t<n;t++){const n=l[t];n.content.length===e?(o[t].push(n),i[t]+=1,l[t]=a[t][i[t]]):(o[t].push({...n,content:n.content.slice(0,e)}),l[t]={...n,content:n.content.slice(e),offset:n.offset+e})}}}return t}(...a.map((e=>e.tokens))),i=o[0].map(((e,t)=>e.map(((e,a)=>{const i={content:e.content,variants:{},offset:e.offset};return"includeExplanation"in n&&n.includeExplanation&&(i.explanation=e.explanation),o.forEach(((e,n)=>{const{content:o,explanation:l,offset:s,...u}=e[t][a];i.variants[r[n].color]=u})),i})))),l=a[0].state?new Hn(Object.fromEntries(a.map((e=>[e.theme,e.state?.getInternalStack(e.theme)]))),a[0].state.lang):void 0;return l&&Gn(i,l),i}function pr(e,t,n){let r,a,o,i,l,s;if("themes"in n){const{defaultColor:u="light",cssVariablePrefix:c="--shiki-"}=n,p=Object.entries(n.themes).filter((e=>e[1])).map((e=>({color:e[0],theme:e[1]}))).sort(((e,t)=>e.color===u?-1:t.color===u?1:0));if(0===p.length)throw new q("`themes` option must not be empty");const d=cr(e,t,n);if(s=Wn(d),u&&!p.find((e=>e.color===u)))throw new q(`\`themes\` option must contain the defaultColor key \`${u}\``);const f=p.map((t=>e.getTheme(t.theme))),h=p.map((e=>e.color));o=d.map((e=>e.map((e=>function(e,t,n,r){const a={content:e.content,explanation:e.explanation,offset:e.offset},o=t.map((t=>Bn(e.variants[t]))),i=new Set(o.flatMap((e=>Object.keys(e)))),l={};return o.forEach(((e,a)=>{for(const o of i){const i=e[o]||"inherit";if(0===a&&r)l[o]=i;else{const e="color"===o?"":"background-color"===o?"-bg":`-${o}`,r=n+t[a]+("color"===o?"":e);l[r]=i}}})),a.htmlStyle=l,a}(e,h,c,u))))),s&&Gn(o,s);const g=p.map((e=>Mn(e.theme,n)));a=p.map(((e,t)=>(0===t&&u?"":`${c+e.color}:`)+(On(f[t].fg,g[t])||"inherit"))).join(";"),r=p.map(((e,t)=>(0===t&&u?"":`${c+e.color}-bg:`)+(On(f[t].bg,g[t])||"inherit"))).join(";"),i=`shiki-themes ${f.map((e=>e.name)).join(" ")}`,l=u?void 0:[a,r].join(";")}else{if(!("theme"in n))throw new q("Invalid options, either `theme` or `themes` must be provided");{const l=Mn(n.theme,n);o=ar(e,t,n);const u=e.getTheme(n.theme);r=On(u.bg,l),a=On(u.fg,l),i=u.name,s=Wn(o)}}return{tokens:o,fg:a,bg:r,themeName:i,rootStyle:l,grammarState:s}}function dr(e,t,n,r={meta:{},options:n,codeToHast:(t,n)=>dr(e,t,n),codeToTokens:(t,n)=>pr(e,t,n)}){let a=t;for(const e of Kn(n))a=e.preprocess?.call(r,a,n)||a;let{tokens:o,fg:i,bg:l,themeName:s,rootStyle:u,grammarState:c}=pr(e,a,n);const{mergeWhitespaces:p=!0}=n;!0===p?o=function(e){return e.map((e=>{const t=[];let n="",r=0;return e.forEach(((a,o)=>{const i=!(a.fontStyle&&a.fontStyle&oe.Underline);i&&a.content.match(/^\s+$/)&&e[o+1]?(r||(r=a.offset),n+=a.content):n?(i?t.push({...a,offset:r,content:n+a.content}):t.push({content:n,offset:r},a),r=0,n=""):t.push(a)})),t}))}(o):"never"===p&&(o=function(e){return e.map((e=>e.flatMap((e=>{if(e.content.match(/^\s+$/))return e;const t=e.content.match(/^(\s*)(.*?)(\s*)$/);if(!t)return e;const[,n,r,a]=t;if(!n&&!a)return e;const o=[{...e,offset:e.offset+n.length,content:r}];return n&&o.unshift({content:n,offset:e.offset}),a&&o.push({content:a,offset:e.offset+n.length+r.length}),o}))))}(o));const d={...r,get source(){return a}};for(const e of Kn(n))o=e.tokens?.call(d,o)||o;return function(e,t,n,r=Wn(e)){const a=Kn(t),o=[],i={type:"root",children:[]},{structure:l="classic",tabindex:s="0"}=t;let u={type:"element",tagName:"pre",properties:{class:`shiki ${t.themeName||""}`,style:t.rootStyle||`background-color:${t.bg};color:${t.fg}`,...!1!==s&&null!=s?{tabindex:s.toString()}:{},...Object.fromEntries(Array.from(Object.entries(t.meta||{})).filter((([e])=>!e.startsWith("_"))))},children:[]},c={type:"element",tagName:"code",properties:{},children:o};const p=[],d={...n,structure:l,addClassToHast:Fn,get source(){return n.source},get tokens(){return e},get options(){return t},get root(){return i},get pre(){return u},get code(){return c},get lines(){return p}};if(e.forEach(((e,t)=>{t&&("inline"===l?i.children.push({type:"element",tagName:"br",properties:{},children:[]}):"classic"===l&&o.push({type:"text",value:"\n"}));let n={type:"element",tagName:"span",properties:{class:"line"},children:[]},r=0;for(const o of e){let e={type:"element",tagName:"span",properties:{...o.htmlAttrs},children:[{type:"text",value:o.content}]};const s=Un(o.htmlStyle||Bn(o));s&&(e.properties.style=s);for(const i of a)e=i?.span?.call(d,e,t+1,r,n,o)||e;"inline"===l?i.children.push(e):"classic"===l&&n.children.push(e),r+=o.content.length}if("classic"===l){for(const e of a)n=e?.line?.call(d,n,t+1)||n;p.push(n),o.push(n)}})),"classic"===l){for(const e of a)c=e?.code?.call(d,c)||c;u.children.push(c);for(const e of a)u=e?.pre?.call(d,u)||u;i.children.push(u)}let f=i;for(const e of a)f=e?.root?.call(d,f)||f;return r&&Gn(f,r),f}(o,{...n,fg:i,bg:l,themeName:s,rootStyle:u},d,c)}function fr(e,t,n){const r={meta:{},options:n,codeToHast:(t,n)=>dr(e,t,n),codeToTokens:(t,n)=>pr(e,t,n)};let a=function(e,t){const n=t||An,r=n.quote||'"',a='"'===r?"'":'"';if('"'!==r&&"'"!==r)throw new Error("Invalid quote `"+r+"`, expected `'` or `\"`");return{one:Rn,all:In,settings:{omitOptionalTags:n.omitOptionalTags||!1,allowParseErrors:n.allowParseErrors||!1,allowDangerousCharacters:n.allowDangerousCharacters||!1,quoteSmart:n.quoteSmart||!1,preferUnquoted:n.preferUnquoted||!1,tightAttributes:n.tightAttributes||!1,upperDoctype:n.upperDoctype||!1,tightDoctype:n.tightDoctype||!1,bogusComments:n.bogusComments||!1,tightCommaSeparatedLists:n.tightCommaSeparatedLists||!1,tightSelfClosing:n.tightSelfClosing||!1,collapseEmptyAttributes:n.collapseEmptyAttributes||!1,allowDangerousHtml:n.allowDangerousHtml||!1,voids:n.voids||lt,characterReferences:n.characterReferences||Ln,closeSelfClosing:n.closeSelfClosing||!1,closeEmptyElements:n.closeEmptyElements||!1},schema:"svg"===n.space?Mt:It,quote:r,alternative:a}.one(Array.isArray(e)?{type:"root",children:e}:e,void 0,void 0)}(dr(e,t,n,r));for(const e of Kn(n))a=e.postprocess?.call(r,a,n)||a;return a}const hr="__shiki_resolved";function gr(e){if(e?.[hr])return e;const t={...e};t.tokenColors&&!t.settings&&(t.settings=t.tokenColors,delete t.tokenColors),t.type||="dark",t.colorReplacements={...t.colorReplacements},t.settings||=[];let{bg:n,fg:r}=t;if(!n||!r){const e=t.settings?t.settings.find((e=>!e.name&&!e.scope)):void 0;e?.settings?.foreground&&(r=e.settings.foreground),e?.settings?.background&&(n=e.settings.background),!r&&t?.colors?.["editor.foreground"]&&(r=t.colors["editor.foreground"]),!n&&t?.colors?.["editor.background"]&&(n=t.colors["editor.background"]),r||(r="light"===t.type?"#333333":"#bbbbbb"),n||(n="light"===t.type?"#fffffe":"#1e1e1e"),t.fg=r,t.bg=n}t.settings[0]&&t.settings[0].settings&&!t.settings[0].scope||t.settings.unshift({settings:{foreground:t.fg,background:t.bg}});let a=0;const o=new Map;function i(e){if(o.has(e))return o.get(e);a+=1;const n=`#${a.toString(16).padStart(8,"0").toLowerCase()}`;return t.colorReplacements?.[`#${n}`]?i(e):(o.set(e,n),n)}t.settings=t.settings.map((e=>{const n=e.settings?.foreground&&!e.settings.foreground.startsWith("#"),r=e.settings?.background&&!e.settings.background.startsWith("#");if(!n&&!r)return e;const a={...e,settings:{...e.settings}};if(n){const n=i(e.settings.foreground);t.colorReplacements[n]=e.settings.foreground,a.settings.foreground=n}if(r){const n=i(e.settings.background);t.colorReplacements[n]=e.settings.background,a.settings.background=n}return a}));for(const e of Object.keys(t.colors||{}))if(("editor.foreground"===e||"editor.background"===e||e.startsWith("terminal.ansi"))&&!t.colors[e]?.startsWith("#")){const n=i(t.colors[e]);t.colorReplacements[n]=t.colors[e],t.colors[e]=n}return Object.defineProperty(t,hr,{enumerable:!1,writable:!1,value:!0}),t}async function mr(e){return Array.from(new Set((await Promise.all(e.filter((e=>{return!("ansi"===(t=e)||zn(t));var t})).map((async e=>await jn(e).then((e=>Array.isArray(e)?e:[e])))))).flat()))}async function yr(e){const t=await Promise.all(e.map((async e=>function(e){return $n(e)}(e)?null:gr(await jn(e)))));return t.filter((e=>!!e))}class br extends Error{constructor(e){super(e),this.name="ShikiError"}}class vr extends ot{constructor(e,t,n,r={}){super(e),this._resolver=e,this._themes=t,this._langs=n,this._alias=r,this._themes.map((e=>this.loadTheme(e))),this.loadLanguages(this._langs)}_resolvedThemes=new Map;_resolvedGrammars=new Map;_langMap=new Map;_langGraph=new Map;_textmateThemeCache=new WeakMap;_loadedThemesCache=null;_loadedLanguagesCache=null;getTheme(e){return"string"==typeof e?this._resolvedThemes.get(e):this.loadTheme(e)}loadTheme(e){const t=gr(e);return t.name&&(this._resolvedThemes.set(t.name,t),this._loadedThemesCache=null),t}getLoadedThemes(){return this._loadedThemesCache||(this._loadedThemesCache=[...this._resolvedThemes.keys()]),this._loadedThemesCache}setTheme(e){let t=this._textmateThemeCache.get(e);t||(t=ee.createFromRawTheme(e),this._textmateThemeCache.set(e,t)),this._syncRegistry.setTheme(t)}getGrammar(e){if(this._alias[e]){const t=new Set([e]);for(;this._alias[e];){if(e=this._alias[e],t.has(e))throw new br(`Circular alias \`${Array.from(t).join(" -> ")} -> ${e}\``);t.add(e)}}return this._resolvedGrammars.get(e)}loadLanguage(e){if(this.getGrammar(e.name))return;const t=new Set([...this._langMap.values()].filter((t=>t.embeddedLangsLazy?.includes(e.name))));this._resolver.addLanguage(e);const n={balancedBracketSelectors:e.balancedBracketSelectors||["*"],unbalancedBracketSelectors:e.unbalancedBracketSelectors||[]};this._syncRegistry._rawGrammars.set(e.scopeName,e);const r=this.loadGrammarWithConfiguration(e.scopeName,1,n);if(r.name=e.name,this._resolvedGrammars.set(e.name,r),e.aliases&&e.aliases.forEach((t=>{this._alias[t]=e.name})),this._loadedLanguagesCache=null,t.size)for(const e of t)this._resolvedGrammars.delete(e.name),this._loadedLanguagesCache=null,this._syncRegistry?._injectionGrammars?.delete(e.scopeName),this._syncRegistry?._grammars?.delete(e.scopeName),this.loadLanguage(this._langMap.get(e.name))}dispose(){super.dispose(),this._resolvedThemes.clear(),this._resolvedGrammars.clear(),this._langMap.clear(),this._langGraph.clear(),this._loadedThemesCache=null}loadLanguages(e){for(const t of e)this.resolveEmbeddedLanguages(t);const t=Array.from(this._langGraph.entries()),n=t.filter((([e,t])=>!t));if(n.length){const e=t.filter((([e,t])=>t&&t.embeddedLangs?.some((e=>n.map((([e])=>e)).includes(e))))).filter((e=>!n.includes(e)));throw new br(`Missing languages ${n.map((([e])=>`\`${e}\``)).join(", ")}, required by ${e.map((([e])=>`\`${e}\``)).join(", ")}`)}for(const[e,n]of t)this._resolver.addLanguage(n);for(const[e,n]of t)this.loadLanguage(n)}getLoadedLanguages(){return this._loadedLanguagesCache||(this._loadedLanguagesCache=[...new Set([...this._resolvedGrammars.keys(),...Object.keys(this._alias)])]),this._loadedLanguagesCache}resolveEmbeddedLanguages(e){if(this._langMap.set(e.name,e),this._langGraph.set(e.name,e),e.embeddedLangs)for(const t of e.embeddedLangs)this._langGraph.set(t,this._langMap.get(t))}}class wr{_langs=new Map;_scopeToLang=new Map;_injections=new Map;_onigLib;constructor(e,t){this._onigLib={createOnigScanner:t=>e.createScanner(t),createOnigString:t=>e.createString(t)},t.forEach((e=>this.addLanguage(e)))}get onigLib(){return this._onigLib}getLangRegistration(e){return this._langs.get(e)}loadGrammar(e){return this._scopeToLang.get(e)}addLanguage(e){this._langs.set(e.name,e),e.aliases&&e.aliases.forEach((t=>{this._langs.set(t,e)})),this._scopeToLang.set(e.scopeName,e),e.injectTo&&e.injectTo.forEach((t=>{this._injections.get(t)||this._injections.set(t,[]),this._injections.get(t).push(e.scopeName)}))}getInjections(e){const t=e.split(".");let n=[];for(let e=1;e<=t.length;e++){const r=t.slice(0,e).join(".");n=[...n,...this._injections.get(r)||[]]}return n}}let xr=0;const kr=String.raw,Sr=kr`(?:\p{Emoji}\uFE0F\u20E3?|\p{Emoji_Modifier_Base}\p{Emoji_Modifier}?|\p{Emoji_Presentation})`,Cr=kr`\u{E0061}-\u{E007A}`,_r=String.raw`\(\?(?:[:=!>A-Za-z\-]|<[=!]|\(DEFINE\))`;function Er(e,t){for(let n=0;n<e.length;n++)e[n]>=t&&e[n]++}const Nr=Object.freeze({DEFAULT:"DEFAULT",CHAR_CLASS:"CHAR_CLASS"});function Pr(e,t,n,r){const a=new RegExp(String.raw`${t}|(?<$skip>\[\^?|\\?.)`,"gsu"),o=[!1];let i=0,l="";for(const t of e.matchAll(a)){const{0:e,groups:{$skip:a}}=t;a||r&&r===Nr.DEFAULT!=!i?("["===e[0]?(i++,o.push("^"===e[1])):"]"===e&&i&&(i--,o.pop()),l+=e):l+=n instanceof Function?n(t,{context:i?Nr.CHAR_CLASS:Nr.DEFAULT,negated:o[o.length-1]}):n}return l}function Ar(e,t,n,r){Pr(e,t,n,r)}function Lr(e,t,n){return!!function(e,t,n=0,r){if(!new RegExp(t,"su").test(e))return null;const a=new RegExp(`${t}|(?<$skip>\\\\?.)`,"gsu");a.lastIndex=n;let o,i=0;for(;o=a.exec(e);){const{0:e,groups:{$skip:t}}=o;if(!(t||r&&r===Nr.DEFAULT!=!i))return o;"["===e?i++:"]"===e&&i&&i--,a.lastIndex==o.index&&a.lastIndex++}return null}(e,t,0,n)}function Tr(e,t){const n=/\\?./gsu;n.lastIndex=t;let r,a=e.length,o=0,i=1;for(;r=n.exec(e);){const[e]=r;if("["===e)o++;else if(o)"]"===e&&o--;else if("("===e)i++;else if(")"===e&&(i--,!i)){a=r.index;break}}return e.slice(t,a)}const Rr=new RegExp(String.raw`(?<noncapturingStart>${_r})|(?<capturingStart>\((?:\?<[^>]+>)?)|\\?.`,"gsu"),Ir=String.raw`(?:[?*+]|\{\d+(?:,\d*)?\})`,Mr=new RegExp(String.raw`
\\(?: \d+
  | c[A-Za-z]
  | [gk]<[^>]+>
  | [pPu]\{[^\}]+\}
  | u[A-Fa-f\d]{4}
  | x[A-Fa-f\d]{2}
  )
| \((?: \? (?: [:=!>]
  | <(?:[=!]|[^>]+>)
  | [A-Za-z\-]+:
  | \(DEFINE\)
  ))?
| (?<qBase>${Ir})(?<qMod>[?+]?)(?<invalidQ>[?*+\{]?)
| \\?.
`.replace(/\s+/g,""),"gsu");RegExp;const Or=String.raw,jr=Or`\\g<(?<gRNameOrNum>[^>&]+)&R=(?<gRDepth>[^>]+)>`,zr=Or`\(\?R=(?<rDepth>[^\)]+)\)|${jr}`,$r=Or`\(\?<(?![=!])(?<captureName>[^>]+)>`,Fr=Or`${$r}|(?<unnamed>\()(?!\?)`,Dr=new RegExp(Or`${$r}|${zr}|\(\?|\\?.`,"gsu"),Br="Cannot use multiple overlapping recursions";function Ur(e){const t=`Max depth must be integer between 2 and 100; used ${e}`;if(!/^[1-9]\d*$/.test(e))throw new Error(t);if((e=+e)<2||e>100)throw new Error(t)}function qr(e,t,n,r,a,o,i){const l=new Set;r&&Ar(e+t,$r,(({groups:{captureName:e}})=>{l.add(e)}),Nr.DEFAULT);const s=[n,r?l:null,a,o,i];return`${e}${Gr(`(?:${e}`,"forward",...s)}(?:)${Gr(`${t})`,"backward",...s)}${t}`}function Gr(e,t,n,r,a,o,i){const l=e=>"forward"===t?e+2:n-e+2-1;let s="";for(let t=0;t<n;t++){const n=l(t);s+=Pr(e,Or`${Fr}|\\k<(?<backref>[^>]+)>`,(({0:e,groups:{captureName:t,unnamed:l,backref:s}})=>{if(s&&r&&!r.has(s))return e;const u=`_$${n}`;if(l||t){const n=i+o.length+1;return o.push(n),Wr(a,n),l?e:`(?<${t}${u}>`}return Or`\k<${s}${u}>`}),Nr.DEFAULT)}return s}function Wr(e,t){for(let n=0;n<e.length;n++)e[n]>=t&&e[n]++}function Hr(e,t,n,r,a,o){if(e.size&&r){let i=0;Ar(t,Fr,(()=>i++),Nr.DEFAULT);const l=o-i+a,s=new Map;return e.forEach(((e,t)=>{const a=(r-i*n)/n,o=i*n,u=t>l+i?t+r:t,c=[];for(const t of e)if(t<=l)c.push(t);else if(t>l+i+a)c.push(t+r);else if(t<=l+i)for(let e=0;e<=n;e++)c.push(t+i*e);else for(let e=0;e<=n;e++)c.push(t+o+a*e);s.set(u,c)})),s}return e}var Vr=String.fromCodePoint,Qr=String.raw,Xr={flagGroups:(()=>{try{new RegExp("(?i:)")}catch{return!1}return!0})(),unicodeSets:(()=>{try{new RegExp("","v")}catch{return!1}return!0})()};function Kr(e,{enable:t,disable:n}){return{dotAll:!(n?.dotAll||!t?.dotAll&&!e.dotAll),ignoreCase:!(n?.ignoreCase||!t?.ignoreCase&&!e.ignoreCase)}}function Yr(e,t,n){return e.has(t)||e.set(t,n),e.get(t)}function Zr(e,t){return ea[e]>=ea[t]}function Jr(e,t){if(!e)throw new Error(t??"Value expected");return e}Xr.literalHyphenIncorrectlyCreatesRange=(()=>{if(!Xr.unicodeSets)return!1;try{new RegExp(Qr`[\d\-a]`,"v")}catch{return!0}return!1})();var ea={ES2025:2025,ES2024:2024,ES2018:2018},ta={auto:"auto",ES2025:"ES2025",ES2024:"ES2024",ES2018:"ES2018"};function na(e){if(void 0!==e?.target&&!ta[e.target])throw new Error(`Unexpected target "${e.target}"`);const t={accuracy:"default",avoidSubclass:!1,flags:"",global:!1,hasIndices:!1,lazyCompileLength:1/0,target:"auto",verbose:!1,...e,rules:{allowOrphanBackrefs:!1,asciiWordBoundaries:!1,captureGroup:!1,recursionLimit:20,singleline:!1,...e?.rules}};return"auto"===t.target&&(t.target=Xr.flagGroups?"ES2025":Xr.unicodeSets?"ES2024":"ES2018"),t}var ra=new Set([Vr(304),Vr(305)]);function aa(e){if(ra.has(e))return[e];const t=new Set,n=e.toLowerCase(),r=n.toUpperCase(),a=pa.get(n),o=ua.get(n),i=ca.get(n);return 1===[...r].length&&t.add(r),i&&t.add(i),a&&t.add(a),t.add(n),o&&t.add(o),[...t]}var oa=new Set("C Other\nCc Control cntrl\nCf Format\nCn Unassigned\nCo Private_Use\nCs Surrogate\nL Letter\nLC Cased_Letter\nLl Lowercase_Letter\nLm Modifier_Letter\nLo Other_Letter\nLt Titlecase_Letter\nLu Uppercase_Letter\nM Mark Combining_Mark\nMc Spacing_Mark\nMe Enclosing_Mark\nMn Nonspacing_Mark\nN Number\nNd Decimal_Number digit\nNl Letter_Number\nNo Other_Number\nP Punctuation punct\nPc Connector_Punctuation\nPd Dash_Punctuation\nPe Close_Punctuation\nPf Final_Punctuation\nPi Initial_Punctuation\nPo Other_Punctuation\nPs Open_Punctuation\nS Symbol\nSc Currency_Symbol\nSk Modifier_Symbol\nSm Math_Symbol\nSo Other_Symbol\nZ Separator\nZl Line_Separator\nZp Paragraph_Separator\nZs Space_Separator\nASCII\nASCII_Hex_Digit AHex\nAlphabetic Alpha\nAny\nAssigned\nBidi_Control Bidi_C\nBidi_Mirrored Bidi_M\nCase_Ignorable CI\nCased\nChanges_When_Casefolded CWCF\nChanges_When_Casemapped CWCM\nChanges_When_Lowercased CWL\nChanges_When_NFKC_Casefolded CWKCF\nChanges_When_Titlecased CWT\nChanges_When_Uppercased CWU\nDash\nDefault_Ignorable_Code_Point DI\nDeprecated Dep\nDiacritic Dia\nEmoji\nEmoji_Component EComp\nEmoji_Modifier EMod\nEmoji_Modifier_Base EBase\nEmoji_Presentation EPres\nExtended_Pictographic ExtPict\nExtender Ext\nGrapheme_Base Gr_Base\nGrapheme_Extend Gr_Ext\nHex_Digit Hex\nIDS_Binary_Operator IDSB\nIDS_Trinary_Operator IDST\nID_Continue IDC\nID_Start IDS\nIdeographic Ideo\nJoin_Control Join_C\nLogical_Order_Exception LOE\nLowercase Lower\nMath\nNoncharacter_Code_Point NChar\nPattern_Syntax Pat_Syn\nPattern_White_Space Pat_WS\nQuotation_Mark QMark\nRadical\nRegional_Indicator RI\nSentence_Terminal STerm\nSoft_Dotted SD\nTerminal_Punctuation Term\nUnified_Ideograph UIdeo\nUppercase Upper\nVariation_Selector VS\nWhite_Space space\nXID_Continue XIDC\nXID_Start XIDS".split(/\s/)),ia=new Map;for(const e of oa)ia.set(ha(e),e);var la=new Set(["Basic_Emoji","Emoji_Keycap_Sequence","RGI_Emoji","RGI_Emoji_Flag_Sequence","RGI_Emoji_Modifier_Sequence","RGI_Emoji_Tag_Sequence","RGI_Emoji_ZWJ_Sequence"]),sa=new Map;for(const e of la)sa.set(ha(e),e);var ua=new Map([["s",Vr(383)],[Vr(383),"s"]]),ca=new Map([[Vr(223),Vr(7838)],[Vr(107),Vr(8490)],[Vr(229),Vr(8491)],[Vr(969),Vr(8486)]]),pa=new Map([ga(453),ga(456),ga(459),ga(498),...ma(8072,8079),...ma(8088,8095),...ma(8104,8111),ga(8124),ga(8140),ga(8188)]),da=new Map([["alnum",Qr`[\p{Alpha}\p{Nd}]`],["alpha",Qr`\p{Alpha}`],["ascii",Qr`\p{ASCII}`],["blank",Qr`[\p{Zs}\t]`],["cntrl",Qr`\p{cntrl}`],["digit",Qr`\p{Nd}`],["graph",Qr`[\P{space}&&\P{cntrl}&&\P{Cn}&&\P{Cs}]`],["lower",Qr`\p{Lower}`],["print",Qr`[[\P{space}&&\P{cntrl}&&\P{Cn}&&\P{Cs}]\p{Zs}]`],["punct",Qr`[\p{P}\p{S}]`],["space",Qr`\p{space}`],["upper",Qr`\p{Upper}`],["word",Qr`[\p{Alpha}\p{M}\p{Nd}\p{Pc}]`],["xdigit",Qr`\p{AHex}`]]),fa=new Set(["alnum","blank","graph","print","word","xdigit"]);function ha(e){return e.replace(/[- _]+/g,"").toLowerCase()}function ga(e){const t=Vr(e);return[t.toLowerCase(),t]}function ma(e,t){return function(e,t){const n=[];for(let r=e;r<=t;r++)n.push(r);return n}(e,t).map((e=>ga(e)))}var ya=new Set(["Lower","Lowercase","Upper","Uppercase","Ll","Lowercase_Letter","Lt","Titlecase_Letter","Lu","Uppercase_Letter"]),ba="Alternator",va="Assertion",wa="Backreference",xa="Character",ka="CharacterClassClose",Sa="CharacterClassHyphen",Ca="CharacterClassIntersector",_a="CharacterClassOpen",Ea="CharacterSet",Na="Directive",Pa="GroupClose",Aa="GroupOpen",La="Subroutine",Ta="Quantifier",Ra="VariableLengthCharacterSet",Ia="EscapedNumber",Ma={any:"any",digit:"digit",dot:"dot",hex:"hex",non_newline:"non_newline",posix:"posix",property:"property",space:"space",word:"word"},Oa={flags:"flags",keep:"keep"},ja="absent_repeater",za="atomic",$a="capturing",Fa="group",Da="lookahead",Ba="lookbehind",Ua=new Map([["a",7],["b",8],["e",27],["f",12],["n",10],["r",13],["t",9],["v",11]]),qa=Qr`\[\^?`,Ga=`c.? | C(?:-.?)?|${Qr`[pP]\{(?:\^?[-\x20_]*[A-Za-z][-\x20\w]*\})?`}|${Qr`x[89A-Fa-f]\p{AHex}(?:\\x[89A-Fa-f]\p{AHex})*`}|${Qr`u(?:\p{AHex}{4})? | x\{[^\}]*\}? | x\p{AHex}{0,2}`}|${Qr`o\{[^\}]*\}?`}|${Qr`\d{1,3}`}`,Wa=/[?*+][?+]?|\{(?:\d+(?:,\d*)?|,\d+)\}\??/,Ha=new RegExp(Qr`
  \\ (?:
    ${Ga}
    | [gk]<[^>]*>?
    | [gk]'[^']*'?
    | .
  )
  | \( (?:
    \? (?:
      [:=!>({]
      | <[=!]
      | <[^>]*>
      | '[^']*'
      | ~\|?
      | #(?:[^)\\]|\\.?)*
      | [^:)]*[:)]
    )?
    | \*
  )?
  | ${Wa.source}
  | ${qa}
  | .
`.replace(/\s+/g,""),"gsu"),Va=new RegExp(Qr`
  \\ (?:
    ${Ga}
    | .
  )
  | \[:(?:\^?\p{Alpha}+|\^):\]
  | ${qa}
  | &&
  | .
`.replace(/\s+/g,""),"gsu");function Qa(e,t="",n){if(n={captureGroup:!1,singleline:!1,...n},"string"!=typeof e)throw new Error("String expected as pattern");if(!/^[imxDSW]*$/.test(t))throw new Error(`Flags "${t}" includes unsupported value`);const r=t.includes("x"),a=[r],o={captureGroup:n.captureGroup,getCurrentModX:()=>a.at(-1),numOpenGroups:0,popModX(){a.pop()},pushModX(e){a.push(e)},replaceCurrentModX(e){a[a.length-1]=e},singleline:n.singleline};let i,l=[];for(Ha.lastIndex=0;i=Ha.exec(e);){const t=Xa(o,e,i[0],Ha.lastIndex);t.tokens?l.push(...t.tokens):t.token&&l.push(t.token),void 0!==t.lastIndex&&(Ha.lastIndex=t.lastIndex)}const s=[];let u=0;l.forEach((e=>{e.type===Aa&&(e.kind===$a?e.number=++u:"("===e.raw&&s.push(e))})),u||s.forEach(((e,t)=>{e.kind=$a,e.number=t+1}));const c=u||s.length;return l=l.map((e=>e.type===Ia?function(e,t){const{raw:n,inCharClass:r}=e,a=n.slice(1);if(!r&&("0"!==a&&1===a.length||"0"!==a[0]&&+a<=t))return[Za(wa,n)];const o=[],i=a.match(/^[0-7]+|\d/g);for(let e=0;e<i.length;e++){const t=i[e];let r;if(0===e&&"8"!==t&&"9"!==t){if(r=parseInt(t,8),r>127)throw new Error(Qr`Octal encoded byte above 177 unsupported "${n}"`)}else r=t.codePointAt(0);o.push(Za(xa,(0===e?"\\":"")+t,{value:r}))}return o}(e,c):e)).flat(),{tokens:l,flags:{ignoreCase:t.includes("i"),dotAll:t.includes("m"),extended:r,digitIsAscii:t.includes("D"),spaceIsAscii:t.includes("S"),wordIsAscii:t.includes("W")},rules:n}}function Xa(e,t,n,r){const[a,o]=n;if("["===a){const e=function(e,t,n){const r=[Za(_a,t,{negate:"^"===t[1]})];let a,o=1;for(Va.lastIndex=n;a=Va.exec(e);){const e=a[0];if("["===e[0]&&":"!==e[1])o++,r.push(Za(_a,e,{negate:"^"===e[1]}));else if("]"===e){if(r.at(-1).type===_a)r.push(Za(xa,e,{value:93}));else if(o--,r.push(Za(ka,e)),!o)break}else{const t=Ka(e);Array.isArray(t)?r.push(...t):r.push(t)}}return{tokens:r,lastIndex:Va.lastIndex||e.length}}(t,n,r);return{tokens:e.tokens,lastIndex:e.lastIndex}}if("\\"===a){if("AbBGzZ".includes(o))return{token:Za(va,n,{kind:n})};if(/^\\g[<']/.test(n)){if(!/^\\g(?:<[^>]+>|'[^']+')$/.test(n))throw new Error(`Invalid group name "${n}"`);return{token:Za(La,n)}}if(/^\\k[<']/.test(n)){if(!/^\\k(?:<[^>]+>|'[^']+')$/.test(n))throw new Error(`Invalid group name "${n}"`);return{token:Za(wa,n)}}if("K"===o)return{token:Za(Na,n,{kind:Oa.keep})};if("N"===o)return{token:Za(Ea,n,{kind:Ma.non_newline})};if("O"===o)return{token:Za(Ea,n,{kind:Ma.any})};if("RX".includes(o))return{token:Za(Ra,n,{kind:n})};if("yY".includes(o))throw new Error(`Unsupported grapheme boundary "${n}"`);const e=Ya(n,{inCharClass:!1});return Array.isArray(e)?{tokens:e}:{token:e}}if("("===a){if("(*"===n)throw new Error(`Unsupported named callout "${n}"`);if("(?{"===n)throw new Error(`Unsupported callout "${n}"`);if("(?#"===n){if(")"!==t[r])throw new Error('Unclosed comment group "(?#"');return{lastIndex:r+1}}if(/^\(\?[-imx]+[:)]$/.test(n))return{token:Ja(n,e)};if(e.pushModX(e.getCurrentModX()),e.numOpenGroups++,"("===n&&!e.captureGroup||"(?:"===n)return{token:Za(Aa,n,{kind:Fa})};if("(?>"===n)return{token:Za(Aa,n,{kind:za})};if("(?="===n||"(?!"===n||"(?<="===n||"(?<!"===n)return{token:Za(Aa,n,{kind:"<"===n[2]?Ba:Da,negate:n.endsWith("!")})};if("("===n&&e.captureGroup||n.startsWith("(?<")&&n.endsWith(">")||n.startsWith("(?'")&&n.endsWith("'")){const e=Za(Aa,n,{kind:$a});return"("!==n&&(e.name=n.slice(3,-1)),{token:e}}if(n.startsWith("(?~")){if("(?~|"===n)throw new Error(`Unsupported absent function kind "${n}"`);return{token:Za(Aa,n,{kind:ja})}}if("(?("===n)throw new Error(`Unsupported conditional "${n}"`);throw new Error(`Invalid or unsupported group option "${n}"`)}if(")"===n){if(e.popModX(),e.numOpenGroups--,e.numOpenGroups<0)throw new Error('Unmatched ")"');return{token:Za(Pa,n)}}if("#"===n&&e.getCurrentModX()){const e=t.indexOf("\n",r);return{lastIndex:-1===e?t.length:e}}if(/^\s$/.test(n)&&e.getCurrentModX()){const e=/\s+/y;return e.lastIndex=r,{lastIndex:e.exec(t)?e.lastIndex:r}}if("."===n)return{token:Za(Ea,n,{kind:Ma.dot})};if("^"===n||"$"===n){const t=e.singleline?{"^":Qr`\A`,$:Qr`\Z`}[n]:n;return{token:Za(va,n,{kind:t})}}return"|"===n?{token:Za(ba,n)}:Wa.test(n)?{token:eo(n)}:(ro(n),{token:Za(xa,n,{value:n.codePointAt(0)})})}function Ka(e){if("\\"===e[0])return Ya(e,{inCharClass:!0});if("["===e[0]){const t=/\[:(?<negate>\^?)(?<name>[a-z]+):\]/.exec(e);if(!t||!da.get(t.groups.name))throw new Error(`Invalid POSIX class "${e}"`);return Za(Ea,e,{kind:Ma.posix,negate:!!t.groups.negate,value:t.groups.name})}return"-"===e?Za(Sa,e):"&&"===e?Za(Ca,e):(ro(e),Za(xa,e,{value:e.codePointAt(0)}))}function Ya(e,{inCharClass:t}){const n=e[1];if("c"===n||"C"===n)return function(e){const t="c"===e[1]?e[2]:e[3];if(!t||!/[A-Za-z]/.test(t))throw new Error(`Unsupported control character "${e}"`);return Za(xa,e,{value:t.toUpperCase().codePointAt(0)-64})}(e);if("dDhHsSwW".includes(n))return function(e){const t=e[1].toLowerCase();return Za(Ea,e,{kind:{d:Ma.digit,h:Ma.hex,s:Ma.space,w:Ma.word}[t],negate:e[1]!==t})}(e);if(e.startsWith(Qr`\o{`))throw new Error(`Incomplete, invalid, or unsupported octal code point "${e}"`);if(/^\\[pP]\{/.test(e)){if(3===e.length)throw new Error(`Incomplete or invalid Unicode property "${e}"`);return function(e){const{p:t,neg:n,value:r}=/^\\(?<p>[pP])\{(?<neg>\^?)(?<value>[^}]+)/.exec(e).groups;return Za(Ea,e,{kind:Ma.property,negate:"P"===t&&!n||"p"===t&&!!n,value:r})}(e)}if(/^\\x[89A-Fa-f]\p{AHex}/u.test(e))try{const t=e.split(/\\x/).slice(1).map((e=>parseInt(e,16))),n=new TextDecoder("utf-8",{ignoreBOM:!0,fatal:!0}).decode(new Uint8Array(t)),r=new TextEncoder;return[...n].map((e=>{const t=[...r.encode(e)].map((e=>`\\x${e.toString(16)}`)).join("");return Za(xa,t,{value:e.codePointAt(0)})}))}catch{throw new Error(`Multibyte code "${e}" incomplete or invalid in Oniguruma`)}if("u"===n||"x"===n)return Za(xa,e,{value:no(e)});if(Ua.has(n))return Za(xa,e,{value:Ua.get(n)});if(/\d/.test(n))return Za(Ia,e,{inCharClass:t});if("\\"===e)throw new Error(Qr`Incomplete escape "\"`);if("M"===n)throw new Error(`Unsupported meta "${e}"`);if(2===[...e].length)return Za(xa,e,{value:e.codePointAt(1)});throw new Error(`Unexpected escape "${e}"`)}function Za(e,t,n){return{type:e,raw:t,...n}}function Ja(e,t){let{on:n,off:r}=/^\(\?(?<on>[imx]*)(?:-(?<off>[-imx]*))?/.exec(e).groups;r??="";const a=(t.getCurrentModX()||n.includes("x"))&&!r.includes("x"),o=to(n),i=to(r),l={};if(o&&(l.enable=o),i&&(l.disable=i),e.endsWith(")"))return t.replaceCurrentModX(a),Za(Na,e,{kind:Oa.flags,flags:l});if(e.endsWith(":")){t.pushModX(a),t.numOpenGroups++;const n=Za(Aa,e,{kind:Fa});return(o||i)&&(n.flags=l),n}throw new Error(`Unexpected flag modifier "${e}"`)}function eo(e){const t={};if("{"===e[0]){const{min:n,max:r}=/^\{(?<min>\d*)(?:,(?<max>\d*))?/.exec(e).groups,a=1e5;if(+n>a||+r>a)throw new Error("Quantifier value unsupported in Oniguruma");t.min=+n,t.max=void 0===r?+n:""===r?1/0:+r,t.greedy=!e.endsWith("?"),t.possessive=!1}else t.min="+"===e[0]?1:0,t.max="?"===e[0]?1:1/0,t.greedy="?"!==e[1],t.possessive="+"===e[1];return Za(Ta,e,t)}function to(e){const t={};return e.includes("i")&&(t.ignoreCase=!0),e.includes("m")&&(t.dotAll=!0),e.includes("x")&&(t.extended=!0),Object.keys(t).length?t:null}function no(e){if(/^(?:\\u(?!\p{AHex}{4})|\\x(?!\p{AHex}{1,2}|\{\p{AHex}{1,8}\}))/u.test(e))throw new Error(`Incomplete or invalid escape "${e}"`);const t="{"===e[2]?/^\\x\{\s*(?<hex>\p{AHex}+)/u.exec(e).groups.hex:e.slice(2);return parseInt(t,16)}function ro(e){if(1!==[...e].length)throw new Error(`Expected "${e}" to be a single code point`)}function ao({alternatives:e},t){return 1===e.length&&1===e[0].elements.length&&(!t||t(e[0].elements[0]))}function oo({type:e}){return e===uo.CapturingGroup||e===uo.Group}function io({type:e,kind:t}){return e===uo.Assertion&&(t===po.lookahead||t===po.lookbehind)}function lo(e,t,n){let r=e.node;for(;r.parent;)r=r.parent;function a(e,t){for(let n=0;n<e.length;n++){const r=o(e[n],t,n,e);n=Math.max(-1,n+r)}}function o(e,i=null,l=null,s=null){let u=0,c=!1;const p={node:e,parent:i,key:l,container:s,ast:r,remove(){Jr(s,"Container expected").splice(Math.max(0,l+u),1),u-=1},removeAllNextSiblings:()=>Jr(s,"Container expected").splice(l+1),removeAllPrevSiblings(){const e=l+u;return u-=e,Jr(s,"Container expected").splice(0,Math.max(0,e))},replaceWith(e){!function(e,t){"parent"in t&&(e.parent=t)}(e,i),s?s[Math.max(0,l+u)]=e:i[l]=e},skip(){c=!0}},d=function(e){const t=[so.AnyNode];return(oo(e)||io(e))&&t.push(so.AnyGroup),t.push(e.type),t}(e).find((e=>!!n[e])),f=d&&n[d],h="function"==typeof f?f:f?.enter,g=f?.exit;if(h?.(p,t),!c)switch(e.type){case uo.Regex:o(e.pattern,e,"pattern"),o(e.flags,e,"flags");break;case uo.Alternative:case uo.CharacterClass:a(e.elements,e);break;case uo.Assertion:io(e)&&a(e.alternatives,e);break;case uo.Backreference:case uo.Character:case uo.CharacterSet:case uo.Directive:case uo.Flags:case uo.Recursion:case uo.Subroutine:case uo.VariableLengthCharacterSet:break;case uo.AbsentFunction:case uo.CapturingGroup:case uo.Group:case uo.Pattern:a(e.alternatives,e);break;case uo.CharacterClassIntersection:a(e.classes,e);break;case uo.CharacterClassRange:o(e.min,e,"min"),o(e.max,e,"max");break;case uo.Quantifier:o(e.element,e,"element");break;default:throw new Error(`Unexpected node type "${e.type}"`)}return g?.(p,t),u}o(e.node,e.parent,e.key,e.container)}var so={AnyGroup:"AnyGroup",AnyNode:"AnyNode"},uo={AbsentFunction:"AbsentFunction",Alternative:"Alternative",Assertion:"Assertion",Backreference:"Backreference",CapturingGroup:"CapturingGroup",Character:"Character",CharacterClass:"CharacterClass",CharacterClassIntersection:"CharacterClassIntersection",CharacterClassRange:"CharacterClassRange",CharacterSet:"CharacterSet",Directive:"Directive",Flags:"Flags",Group:"Group",Pattern:"Pattern",Quantifier:"Quantifier",Regex:"Regex",Subroutine:"Subroutine",VariableLengthCharacterSet:"VariableLengthCharacterSet",Recursion:"Recursion"},co="repeater",po={line_end:"line_end",line_start:"line_start",lookahead:"lookahead",lookbehind:"lookbehind",search_start:"search_start",string_end:"string_end",string_end_newline:"string_end_newline",string_start:"string_start",word_boundary:"word_boundary"},fo=Ma,ho=Oa,go="grapheme",mo="newline";function yo({tokens:e,flags:t,rules:n},r){const a={skipBackrefValidation:!1,skipLookbehindValidation:!1,skipPropertyNameValidation:!1,verbose:!1,...r},o={capturingGroups:[],current:0,hasNumberedRef:!1,namedGroupsByName:new Map,parent:null,skipBackrefValidation:a.skipBackrefValidation,skipLookbehindValidation:a.skipLookbehindValidation,skipPropertyNameValidation:a.skipPropertyNameValidation,subroutines:[],token:null,tokens:e,verbose:a.verbose,walk:i};function i(t,n){const r=e[o.current];switch(o.parent=t,o.token=r,o.current++,r.type){case ba:return bo();case va:return function({kind:e}){return vo(Jr({"^":po.line_start,$:po.line_end,"\\A":po.string_start,"\\b":po.word_boundary,"\\B":po.word_boundary,"\\G":po.search_start,"\\z":po.string_end,"\\Z":po.string_end_newline}[e],`Unexpected assertion kind "${e}"`),{negate:e===Qr`\B`})}(r);case wa:return function(e){const{raw:t}=e.token,n=/^\\k[<']/.test(t),r=n?t.slice(3,-1):t.slice(1),a=(n,r=!1)=>{const a=e.capturingGroups.length;let o=!1;if(n>a){if(!e.skipBackrefValidation)throw new Error(`Not enough capturing groups defined to the left "${t}"`);o=!0}return e.hasNumberedRef=!0,wo(r?a+1-n:n,{orphan:o})};if(n){const n=/^(?<sign>-?)0*(?<num>[1-9]\d*)$/.exec(r);if(n)return a(+n.groups.num,!!n.groups.sign);if(/[-+]/.test(r))throw new Error(`Invalid backref name "${t}"`);if(!e.namedGroupsByName.has(r))throw new Error(`Group name not defined to the left "${t}"`);return wo(r)}return a(+r)}(o);case xa:return ko(r.value,{useLastValid:!!n.isCheckingRangeEnd});case Sa:return function(e,t){const{parent:n,tokens:r,walk:a}=e,o=n.elements.at(-1),i=r[e.current];if(!t.isCheckingRangeEnd&&o&&o.type!==uo.CharacterClass&&o.type!==uo.CharacterClassRange&&i&&i.type!==_a&&i.type!==ka&&i.type!==Ca){const e=a(n,{...t,isCheckingRangeEnd:!0});if(o.type===uo.Character&&e.type===uo.Character)return n.elements.pop(),function(e,t){if(t.value<e.value)throw new Error("Character class range out of order");return{type:uo.CharacterClassRange,min:e,max:t}}(o,e);throw new Error("Invalid character class range")}return ko(45)}(o,n);case _a:return function(e,t){const{token:n,tokens:r,verbose:a,walk:o}=e,i=r[e.current];let l=So({negate:n.negate});const s=l.elements[0];let u=Lo(i);for(;u.type!==ka;){if(u.type===Ca)s.classes.push(So({negate:!1,baseOnly:!0})),e.current++;else{const e=s.classes.at(-1);e.elements.push(o(e,t))}u=Lo(r[e.current],i)}if(a||function(e){for(let t=0;t<e.classes.length;t++){const n=e.classes[t],r=n.elements[0];1===n.elements.length&&r.type===uo.CharacterClass&&(e.classes[t]=r,r.negate=n.negate!==r.negate)}}(s),1===s.classes.length){const e=s.classes[0];e.negate=l.negate!==e.negate,l=e}return e.current++,l}(o,n);case Ea:return function({token:e,skipPropertyNameValidation:t}){let{kind:n,negate:r,value:a}=e;if(n===Ma.property){const e=ha(a);if(!fa.has(e))return Po(a,{negate:r,skipPropertyNameValidation:t});n=Ma.posix,a=e}return n===Ma.posix?{type:uo.CharacterSet,kind:fo.posix,negate:r,value:a}:Co(n,{negate:r})}(o);case Na:return function({kind:e,flags:t}){const n={type:uo.Directive,kind:Jr(ho[e],`Unexpected directive kind "${e}"`)};return e===Oa.flags&&(n.flags=t),n}(r);case Aa:return function(e,t){const{token:n,tokens:r,capturingGroups:a,namedGroupsByName:o,skipLookbehindValidation:i,verbose:l,walk:s}=e;let u=function({flags:e,kind:t,name:n,negate:r,number:a}){switch(t){case ja:return function(e){return{type:uo.AbsentFunction,kind:e,alternatives:[bo()]}}(co);case za:return _o({atomic:!0});case $a:return xo(a,n);case Fa:return _o({flags:e});case Da:case Ba:return Eo({behind:t===Ba,negate:r});default:throw new Error(`Unexpected group kind "${t}"`)}}(n);const c=u.type===uo.AbsentFunction,p=u.kind===po.lookbehind,d=p&&u.negate;if(u.type===uo.CapturingGroup&&(a.push(u),u.name&&Yr(o,u.name,[]).push(u)),c&&t.isInAbsentFunction)throw new Error("Nested absent function not supported by Oniguruma");let f=To(r[e.current]);for(;f.type!==Pa;){if(f.type===ba)u.alternatives.push(bo()),e.current++;else{const e=u.alternatives.at(-1),n=s(e,{...t,isInAbsentFunction:t.isInAbsentFunction||c,isInLookbehind:t.isInLookbehind||p,isInNegLookbehind:t.isInNegLookbehind||d});if(e.elements.push(n),(p||t.isInLookbehind)&&!i){const e="Lookbehind includes a pattern not allowed by Oniguruma";if(d||t.isInNegLookbehind){if(n.kind===po.lookahead||n.type===uo.CapturingGroup)throw new Error(e)}else if(n.kind===po.lookahead||n.kind===po.lookbehind&&n.negate)throw new Error(e)}}f=To(r[e.current])}return l||(u=function(e){const t=e.alternatives[0].elements[0];return e.type!==uo.Group||!ao(e,(e=>e.type===uo.Group))||e.atomic&&t.flags||e.flags&&(t.atomic||t.flags)?e:(e.atomic?t.atomic=!0:e.flags&&(t.flags=e.flags),t)}(u)),e.current++,u}(o,n);case Ta:return function({token:e,parent:t}){const{min:n,max:r,greedy:a,possessive:o}=e,i=t.elements.at(-1);if(!i||i.type===uo.Assertion||i.type===uo.Directive)throw new Error("Quantifier requires a repeatable token");const l=No(i,n,r,a,o);return t.elements.pop(),l}(o);case La:return function(e){const{token:t,capturingGroups:n,subroutines:r}=e;let a=t.raw.slice(3,-1);const o=/^(?<sign>[-+]?)0*(?<num>[1-9]\d*)$/.exec(a);if(o){const t=+o.groups.num,r=n.length;if(e.hasNumberedRef=!0,a={"":t,"+":r+t,"-":r+1-t}[o.groups.sign],a<1)throw new Error("Invalid subroutine number")}else"0"===a&&(a=0);const i=function(e){return{type:uo.Subroutine,ref:e}}(a);return r.push(i),i}(o);case Ra:return a=r.kind,{type:uo.VariableLengthCharacterSet,kind:Jr({"\\R":mo,"\\X":go}[a],`Unexpected varcharset kind "${a}"`)};default:throw new Error(`Unexpected token type "${r.type}"`)}var a}const l=function(e,t){return{type:uo.Regex,pattern:e,flags:t}}({type:uo.Pattern,alternatives:[bo()]},function({ignoreCase:e,dotAll:t,extended:n,digitIsAscii:r,spaceIsAscii:a,wordIsAscii:o}){return{type:uo.Flags,ignoreCase:e,dotAll:t,extended:n,digitIsAscii:r,spaceIsAscii:a,wordIsAscii:o}}(t));let s=l.pattern.alternatives[0];for(;o.current<e.length;){const e=i(s,{});e.type===uo.Alternative?(l.pattern.alternatives.push(e),s=e):s.elements.push(e)}const{capturingGroups:u,hasNumberedRef:c,namedGroupsByName:p,subroutines:d}=o;if(c&&p.size&&!n.captureGroup)throw new Error("Numbered backref/subroutine not allowed when using named capture");for(const{ref:e}of d)if("number"==typeof e){if(e>u.length)throw new Error("Subroutine uses a group number that's not defined")}else{if(!p.has(e))throw new Error(Qr`Subroutine uses a group name that's not defined "\g<${e}>"`);if(p.get(e).length>1)throw new Error(Qr`Subroutine uses a duplicate group name "\g<${e}>"`)}return lo({node:l},null,{AnyNode({node:e,parent:t}){e.parent=t}}),l}function bo(){return{type:uo.Alternative,elements:[]}}function vo(e,t){const n=!!t?.negate;return{type:uo.Assertion,kind:e,...e===po.word_boundary&&{negate:n}}}function wo(e,t){const n=!!t?.orphan;return{type:uo.Backreference,...n&&{orphan:n},ref:e}}function xo(e,t){const n=void 0!==t;if(n&&!function(e){return/^[\p{Alpha}\p{Pc}][^)]*$/u.test(e)}(t))throw new Error(`Group name "${t}" invalid in Oniguruma`);return{type:uo.CapturingGroup,number:e,...n&&{name:t},alternatives:[bo()]}}function ko(e,t){const n={useLastValid:!1,...t};if(e>1114111){const t=e.toString(16);if(!n.useLastValid)throw e>1310719?new Error(`Invalid code point out of range "\\x{${t}}"`):new Error(`Invalid code point out of range in JS "\\x{${t}}"`);e=1114111}return{type:uo.Character,value:e}}function So(e){const t={baseOnly:!1,negate:!1,...e};return{type:uo.CharacterClass,negate:t.negate,elements:t.baseOnly?[]:[{type:uo.CharacterClassIntersection,classes:[So({negate:!1,baseOnly:!0})]}]}}function Co(e,{negate:t}){const n={type:uo.CharacterSet,kind:Jr(fo[e],`Unexpected character set kind "${e}"`)};return e!==Ma.digit&&e!==Ma.hex&&e!==Ma.space&&e!==Ma.word||(n.negate=t),n}function _o(e){const t=e?.atomic,n=e?.flags;return{type:uo.Group,...t&&{atomic:t},...n&&{flags:n},alternatives:[bo()]}}function Eo(e){const t={behind:!1,negate:!1,...e};return{type:uo.Assertion,kind:t.behind?po.lookbehind:po.lookahead,negate:t.negate,alternatives:[bo()]}}function No(e,t,n,r=!0,a=!1){const o={type:uo.Quantifier,min:t,max:n,greedy:r,possessive:a,element:e};return n<t?{...o,min:n,max:t,possessive:!0}:o}function Po(e,t){const n={negate:!1,skipPropertyNameValidation:!1,...t};return{type:uo.CharacterSet,kind:fo.property,value:n.skipPropertyNameValidation?e:Ao(e),negate:n.negate}}function Ao(e){const t=ha(e);if(sa.has(t))throw new Error(Qr`Unicode property "\p{${e}}" unsupported in Oniguruma`);return ia.get(t)||e.trim().replace(/[- _]+/g,"_").replace(/[A-Z][a-z]+(?=[A-Z])/g,"$&_").replace(/[A-Za-z]+/g,(e=>e[0].toUpperCase()+e.slice(1).toLowerCase()))}function Lo(e,t){return Jr(e,(93===t?.value?"Empty":"Unclosed")+" character class")}function To(e){return Jr(e,"Unclosed group")}var Ro={AbsentFunction({node:e,replaceWith:t}){const n=Xo(_o(),[zo(Eo({negate:!0}),e.alternatives),Po("Any")]),r=No(n,0,1/0);n.parent=r,t(Xo(_o(),[r]))},Alternative:{enter({node:e,parent:t,key:n},{flagDirectivesByAlt:r}){const a=e.elements.filter((e=>e.kind===ho.flags));for(let e=n+1;e<t.alternatives.length;e++)Yr(r,t.alternatives[e],[]).push(...a)},exit({node:e},{flagDirectivesByAlt:t}){if(t.get(e)?.length){const n=Uo(t.get(e));if(n){const t=Xo(_o({flags:n}),e.elements);t.parent=e,e.elements=[t]}}}},Assertion({node:e,key:t,container:n,ast:r,remove:a,replaceWith:o},i){const{kind:l,negate:s}=e,{asciiWordBoundaries:u,avoidSubclass:c,supportedGNodes:p,wordIsAscii:d}=i;if(l===po.line_end)o(Qo(Qr`(?=\z|\n)`));else if(l===po.line_start)o(Qo(Qr`(?<=\A|\n(?!\z))`,{skipLookbehindValidation:!0}));else if(l===po.lookbehind)i.passedLookbehind=!0;else if(l===po.search_start)if(p.has(e))r.flags.sticky=!0,a();else{const e=n[t-1];if(e&&function(e){const t=[uo.Character,uo.CharacterClass,uo.CharacterSet];return t.includes(e.type)||e.type===uo.Quantifier&&e.min&&t.includes(e.element.type)}(e))o(Xo(Eo({negate:!0})));else{if(c)throw new Error(Qr`Uses "\G" in a way that requires a subclass`);o(vo(po.string_start)),i.strategy="clip_search"}}else l===po.string_end_newline?o(Qo(Qr`(?=\n?\z)`)):l!==po.word_boundary||d||u||o(Qo(s?`(?:(?<=${jo})(?=${jo})|(?<!${jo})(?!${jo}))`:`(?:(?<=${jo})(?!${jo})|(?<!${jo})(?=${jo}))`))},Backreference({node:e},{jsGroupNameMap:t}){let{ref:n}=e;"string"!=typeof n||Vo(n)||(n=Do(n,t),e.ref=n)},CapturingGroup({node:e},{jsGroupNameMap:t,subroutineRefMap:n}){let{name:r}=e;r&&!Vo(r)&&(r=Do(r,t),e.name=r),n.set(e.number,e),r&&n.set(r,e)},CharacterSet({node:e,replaceWith:t},{accuracy:n,minTargetEs2024:r,digitIsAscii:a,spaceIsAscii:o,wordIsAscii:i}){const{kind:l,negate:s,value:u}=e;if(!a||l!==fo.digit&&"digit"!==u)if(!o||l!==fo.space&&"space"!==u)if(!i||l!==fo.word&&"word"!==u){if(l===fo.any)t(Po("Any"));else if(l===fo.digit)t(Po("Nd",{negate:s}));else if(l===fo.hex)t(Po("AHex",{negate:s}));else if(l===fo.non_newline)t(Qo(Qr`[^\n]`));else if(l===fo.space)t(Po("space",{negate:s}));else if(l===fo.word)t(Ko(Qo(jo),s));else if(l===fo.property)oa.has(u)||(e.key="sc");else if(l===fo.posix)if(r||"graph"!==u&&"print"!==u)t(Ko(Qo(da.get(u)),s));else{if("strict"===n)throw new Error(`POSIX class "${u}" requires min target ES2024 or non-strict accuracy`);let e={graph:"!-~",print:" -~"}[u];s&&(e=`\0-${Vr(e.codePointAt(0)-1)}${Vr(e.codePointAt(2)+1)}-􏿿`),t(Qo(`[${e}]`))}}else t(Co(fo.word,{negate:s}));else t(Ko(Qo(Oo),s));else t(Co(fo.digit,{negate:s}))},Directive(e,t){const{node:n,parent:r,ast:a,remove:o,replaceWith:i,removeAllPrevSiblings:l,removeAllNextSiblings:s}=e,{kind:u,flags:c}=n;if(u===ho.flags)if(c.enable||c.disable){const n=Xo(_o({flags:c}),s());i(n),Yo(n,e,t,Ro)}else o();else if(u===ho.keep){const e=a.pattern.alternatives[0].elements[0],t=ao(a.pattern,(e=>e.type===uo.Group))&&1===e.alternatives.length?e:a.pattern;if(r.parent!==t||t.alternatives.length>1)throw new Error(Qr`Uses "\K" in a way that's unsupported`);i(Xo(Eo({behind:!0}),l()))}},Flags({node:e,parent:t}){["digitIsAscii","extended","spaceIsAscii","wordIsAscii"].forEach((t=>delete e[t])),Object.assign(e,{global:!1,hasIndices:!1,multiline:!1,sticky:e.sticky??!1}),t.options={disable:{x:!0,n:!0},force:{v:!0}}},Group({node:e}){if(!e.flags)return;const{enable:t,disable:n}=e.flags;t?.extended&&delete t.extended,n?.extended&&delete n.extended,t?.dotAll&&n?.dotAll&&delete t.dotAll,t?.ignoreCase&&n?.ignoreCase&&delete t.ignoreCase,t&&!Object.keys(t).length&&delete e.flags.enable,n&&!Object.keys(n).length&&delete e.flags.disable,!e.flags.enable&&!e.flags.disable&&delete e.flags},Pattern:{enter({node:e},{supportedGNodes:t}){const n=[];let r=!1,a=!1;for(const t of e.alternatives)if(1===t.elements.length&&t.elements[0].kind===po.search_start)t.elements.pop();else{const e=Wo(t.elements);e?(r=!0,Array.isArray(e)?n.push(...e):n.push(e)):a=!0}r&&!a&&n.forEach((e=>t.add(e)))},exit(e,{accuracy:t,passedLookbehind:n,strategy:r}){if("strict"===t&&n&&r)throw new Error(Qr`Uses "\G" in a way that requires non-strict accuracy`)}},Quantifier({node:e}){if(e.element.type===uo.Quantifier){const t=Xo(_o(),[e.element]);t.parent=e,e.element=t}},Subroutine({node:e},{jsGroupNameMap:t}){let{ref:n}=e;"string"!=typeof n||Vo(n)||(n=Do(n,t),e.ref=n)},VariableLengthCharacterSet({node:e,replaceWith:t},{accuracy:n,minTargetEs2024:r}){const{kind:a}=e;if(a===mo)t(Qo("(?>\r\n?|[\n\v\f\u2028\u2029])"));else{if(a!==go)throw new Error(`Unexpected varcharset kind "${a}"`);{if("strict"===n)throw new Error(Qr`Use of "\X" requires non-strict accuracy`);const e=r?Qr`\p{RGI_Emoji}`:new RegExp(kr`[\u{1F1E6}-\u{1F1FF}]{2}|\u{1F3F4}[${Cr}]{2}[\u{E0030}-\u{E0039}${Cr}]{1,3}\u{E007F}|${Sr}(?:\u200D${Sr})*`,"gu").source.replace(/\\u\{/g,"\\x{");t(Qo(Qr`(?>\r\n|${e}|\P{M}\p{M}*)`,{skipPropertyNameValidation:!0}))}}}},Io={Backreference({node:e},{multiplexCapturesToLeftByRef:t,reffedNodesByReferencer:n}){const{orphan:r,ref:a}=e;r||n.set(e,[...t.get(a).map((({node:e})=>e))])},CapturingGroup:{enter({node:e,replaceWith:t,skip:n},{groupOriginByCopy:r,groupsByName:a,multiplexCapturesToLeftByRef:o,openRefs:i,reffedNodesByReferencer:l}){const s=r.get(e);if(s&&i.has(e.number)){const r=Fo(e.number);return l.set(r,i.get(e.number)),t(r),void n()}i.set(e.number,e),o.set(e.number,[]),e.name&&Yr(o,e.name,[]);const u=o.get(e.name??e.number);for(let t=0;t<u.length;t++){const n=u[t];if(s===n.node||s&&s===n.origin||e===n.origin){u.splice(t,1);break}}if(o.get(e.number).push({node:e,origin:s}),e.name&&o.get(e.name).push({node:e,origin:s}),e.name){const t=Yr(a,e.name,new Map);let n=!1;if(s)n=!0;else for(const e of t.values())if(!e.hasDuplicateNameToRemove){n=!0;break}a.get(e.name).set(e,{node:e,hasDuplicateNameToRemove:n})}},exit({node:e},{openRefs:t}){t.delete(e.number)}},Group:{enter({node:e},t){t.prevFlags=t.currentFlags,e.flags&&(t.currentFlags=Kr(t.currentFlags,e.flags))},exit(e,t){t.currentFlags=t.prevFlags}},Recursion({node:e,parent:t},{reffedNodesByReferencer:n}){const{ref:r}=e;let a=t;for(;(a=a.parent)&&(a.type!==uo.CapturingGroup||a.name!==r&&a.number!==r););n.set(e,a)},Subroutine(e,t){const{node:n,replaceWith:r}=e,{ref:a}=n,o=t.subroutineRefMap.get(a),i=0===a,l=i?Fo(0):$o(o,t.groupOriginByCopy,null);let s=l;if(!i){const e=Uo(function(e,t){const n=[];for(;e=e.parent;)t&&!t(e)||n.push(e);return n}(o,(e=>e.type===uo.Group&&!!e.flags))),n=e?Kr(t.globalFlags,e):t.globalFlags;u=n,c=t.currentFlags,(u.dotAll!==c.dotAll||u.ignoreCase!==c.ignoreCase)&&(s=Xo(_o({flags:qo(n)}),[l]))}var u,c;r(s),i||Yo(s,e,t,Io)}},Mo={Backreference({node:e,replaceWith:t},n){if(e.orphan)return void(n.highestOrphanBackref=Math.max(n.highestOrphanBackref,e.ref));const r=n.reffedNodesByReferencer.get(e).filter((t=>function(e,t){let n=t;do{if(n.type===uo.Pattern)return!1;if(n.type===uo.Alternative)continue;if(n===e)return!1;const t=Go(n.parent);for(const r of t){if(r===n)break;if(r===e)return!0;if(Ho(r,e))return!0}}while(n=n.parent);throw new Error("Unexpected path")}(t,e)));if(r.length)if(r.length>1){const e=r.map((e=>zo(bo(),[wo(e.number)])));t(zo(_o(),e))}else e.ref=r[0].number;else t(Xo(Eo({negate:!0})))},CapturingGroup({node:e},t){e.number=++t.numCapturesToLeft,e.name&&t.groupsByName.get(e.name).get(e).hasDuplicateNameToRemove&&delete e.name},Recursion({node:e},t){0!==e.ref&&(e.ref=t.reffedNodesByReferencer.get(e).number)},Regex:{exit({node:e},t){const n=Math.max(t.highestOrphanBackref-t.numCapturesToLeft,0);for(let t=0;t<n;t++){const t=xo();e.pattern.alternatives.at(-1).elements.push(t)}}}},Oo="[\t-\r ]",jo=Qr`[\p{L}\p{M}\p{N}\p{Pc}]`;function zo(e,t){return t.forEach((t=>t.parent=e)),e[Bo(e)]=t,e}function $o(e,t,n,r){const a=Array.isArray(e)?[]:{};for(const[o,i]of Object.entries(e))"parent"===o?a.parent=Array.isArray(n)?r:n:i&&"object"==typeof i?a[o]=$o(i,t,a,n):("type"===o&&i===uo.CapturingGroup&&t.set(a,t.get(e)??e),a[o]=i);return a}function Fo(e){return{type:uo.Recursion,ref:e}}function Do(e,t){if(t.has(e))return t.get(e);const n=`$${t.size}_${e.replace(/^[^$_\p{IDS}]|[^$\u200C\u200D\p{IDC}]/gu,"_")}`;return t.set(e,n),n}function Bo(e){for(const t of["alternatives","classes","elements"])if(e[t])return t;return null}function Uo(e){const t=["dotAll","ignoreCase"],n={enable:{},disable:{}};return e.forEach((({flags:e})=>{t.forEach((t=>{e.enable?.[t]&&(delete n.disable[t],n.enable[t]=!0),e.disable?.[t]&&(n.disable[t]=!0)}))})),Object.keys(n.enable).length||delete n.enable,Object.keys(n.disable).length||delete n.disable,n.enable||n.disable?n:null}function qo({dotAll:e,ignoreCase:t}){const n={};return(e||t)&&(n.enable={},e&&(n.enable.dotAll=!0),t&&(n.enable.ignoreCase=!0)),e&&t||(n.disable={},!e&&(n.disable.dotAll=!0),!t&&(n.disable.ignoreCase=!0)),n}function Go(e){if(!e)throw new Error("Node expected");if(e.type===uo.Quantifier)return[e.element];const t=Bo(e);return t&&e[t]}function Wo(e){const t=e.find((e=>e.kind===po.search_start||function(e){const t={negate:null,negate:!1};return io(e)&&(null===t.negate||e.negate===t.negate)&&ao(e,(e=>e.kind===po.search_start))}(e)||!function({type:e}){return e===uo.Assertion||e===uo.Directive}(e)));if(!t)return null;if(t.kind===po.search_start)return t;if(io(t))return t.alternatives[0].elements[0];if(oo(t)){const e=[];for(const n of t.alternatives){const t=Wo(n.elements);if(!t)return null;Array.isArray(t)?e.push(...t):e.push(t)}return e}return null}function Ho(e,t){const n=Go(e)??[];for(const e of n)if(e===t||Ho(e,t))return!0;return!1}function Vo(e){return/^[$_\p{IDS}][$\u200C\u200D\p{IDC}]*$/u.test(e)}function Qo(e,t){const n=yo(Qa(e),t).pattern.alternatives;return n.length>1||n[0].elements.length>1?zo(_o(),n):n[0].elements[0]}function Xo(e,t){const n=Bo(e);return e[n][0].parent=e,t&&zo(e[n][0],t),e}function Ko(e,t){return e.negate=t,e}function Yo(e,{parent:t,key:n,container:r},a,o){lo({node:e,parent:t,key:n,container:r},a,o)}var Zo={AnyGroup:{enter({node:e},t){const n=t.getCurrentModI();t.pushModI(e.flags?Kr({ignoreCase:n},e.flags).ignoreCase:n)},exit(e,t){t.popModI()}},Backreference(e,t){t.setHasCasedChar()},Character({node:e},t){ai(Vr(e.value))&&t.setHasCasedChar()},CharacterClassRange({node:e,skip:t},n){t(),oi(e,{firstOnly:!0}).length&&n.setHasCasedChar()},CharacterSet({node:e},t){e.kind===fo.property&&ya.has(e.value)&&t.setHasCasedChar()}},Jo=new Set(["$","(",")","*","+",".","?","[","\\","]","^","{","|","}"]),ei=new Set(["-","\\","]","^","["]),ti=new Set(["(",")","-","/","[","\\","]","^","{","|","}","!","#","$","%","&","*","+",",",".",":",";","<","=",">","?","@","`","~"]),ni=new Map([[9,Qr`\t`],[10,Qr`\n`],[11,Qr`\v`],[12,Qr`\f`],[13,Qr`\r`],[8232,Qr`\u2028`],[8233,Qr`\u2029`],[65279,Qr`\uFEFF`]]),ri=/^\p{Cased}$/u;function ai(e){return ri.test(e)}function oi(e,t){const n=!!t?.firstOnly,r=e.min.value,a=e.max.value,o=[];if(r<65&&(65535===a||a>=131071)||65536===r&&a>=131071)return o;for(let e=r;e<=a;e++){const t=Vr(e);if(!ai(t))continue;const i=aa(t).filter((e=>{const t=e.codePointAt(0);return t<r||t>a}));if(i.length&&(o.push(...i),n))break}return o}function ii(e,{isAfterBackref:t,inCharClass:n,useFlagV:r}){if(ni.has(e))return ni.get(e);if(e<32||e>126&&e<160||e>262143||t&&(a=e)>47&&a<58)return e>255?`\\u{${e.toString(16).toUpperCase()}}`:`\\x${e.toString(16).toUpperCase().padStart(2,"0")}`;var a;const o=n?r?ti:ei:Jo,i=Vr(e);return(o.has(i)?"\\":"")+i}function li({type:e,value:t}){return e===uo.Character&&45===t}var si=class e extends RegExp{#e=new Map;#t=null;#n;#r=null;#a=null;rawOptions={};get source(){return this.#n||"(?:)"}constructor(t,n,r){const a=!!r?.lazyCompile;if(t instanceof RegExp){if(r)throw new Error("Cannot provide options when copying a regexp");const a=t;super(a,n),this.#n=a.source,a instanceof e&&(this.#e=a.#e,this.#r=a.#r,this.#a=a.#a,this.rawOptions=a.rawOptions)}else{const e={hiddenCaptures:[],strategy:null,transfers:[],...r};super(a?"":t,n),this.#n=t,this.#e=function(e,t){const n=new Map;for(const t of e)n.set(t,{hidden:!0});for(const[e,r]of t)for(const t of r)Yr(n,t,{}).transferTo=e;return n}(e.hiddenCaptures,e.transfers),this.#a=e.strategy,this.rawOptions=r??{}}a||(this.#t=this)}exec(t){if(!this.#t){const{lazyCompile:t,...n}=this.rawOptions;this.#t=new e(this.#n,this.flags,n)}const n=this.global||this.sticky,r=this.lastIndex;if("clip_search"===this.#a&&n&&r){this.lastIndex=0;const e=this.#o(t.slice(r));return e&&(function(e,t,n,r){if(e.index+=t,e.input=n,r){const n=e.indices;for(let e=0;e<n.length;e++){const r=n[e];r&&(n[e]=[r[0]+t,r[1]+t])}const r=n.groups;r&&Object.keys(r).forEach((e=>{const n=r[e];n&&(r[e]=[n[0]+t,n[1]+t])}))}}(e,r,t,this.hasIndices),this.lastIndex+=r),e}return this.#o(t)}#o(e){this.#t.lastIndex=this.lastIndex;const t=super.exec.call(this.#t,e);if(this.lastIndex=this.#t.lastIndex,!t||!this.#e.size)return t;const n=[...t];let r;t.length=1,this.hasIndices&&(r=[...t.indices],t.indices.length=1);const a=[0];for(let e=1;e<n.length;e++){const{hidden:o,transferTo:i}=this.#e.get(e)??{};if(o?a.push(null):(a.push(t.length),t.push(n[e]),this.hasIndices&&t.indices.push(r[e])),i&&void 0!==n[e]){const o=Jr(a[i]);if(t[o]=n[e],this.hasIndices&&(t.indices[o]=r[e]),t.groups){this.#r||(this.#r=ui(this.source));const a=this.#r.get(i);a&&(t.groups[a]=n[e],this.hasIndices&&(t.indices.groups[a]=r[e]))}}}return t}};function ui(e){const t=/(?<capture>\((?:\?<(?![=!])(?<name>[^>]+)>|(?!\?)))|\\?./gsu,n=new Map;let r,a=0,o=0;for(;r=t.exec(e);){const{0:e,groups:{capture:t,name:i}}=r;"["===e?a++:a?"]"===e&&a--:t&&(o++,i&&n.set(o,i))}return n}function ci(e,t){const n=function(e,t){const n=na(t),r=function(e,t){const n={accuracy:"default",asciiWordBoundaries:!1,avoidSubclass:!1,bestEffortTarget:"ES2025",...t},r={accuracy:n.accuracy,asciiWordBoundaries:n.asciiWordBoundaries,avoidSubclass:n.avoidSubclass,flagDirectivesByAlt:new Map,jsGroupNameMap:new Map,minTargetEs2024:Zr(n.bestEffortTarget,"ES2024"),passedLookbehind:!1,strategy:null,subroutineRefMap:new Map,supportedGNodes:new Set,digitIsAscii:e.flags.digitIsAscii,spaceIsAscii:e.flags.spaceIsAscii,wordIsAscii:e.flags.wordIsAscii};lo({node:e},r,Ro);const a={dotAll:e.flags.dotAll,ignoreCase:e.flags.ignoreCase},o={currentFlags:a,prevFlags:null,globalFlags:a,groupOriginByCopy:new Map,groupsByName:new Map,multiplexCapturesToLeftByRef:new Map,openRefs:new Map,reffedNodesByReferencer:new Map,subroutineRefMap:r.subroutineRefMap};return lo({node:e},o,Io),lo({node:e},{groupsByName:o.groupsByName,highestOrphanBackref:0,numCapturesToLeft:0,reffedNodesByReferencer:o.reffedNodesByReferencer},Mo),e._originMap=o.groupOriginByCopy,e._strategy=r.strategy,e}(yo(Qa(e,n.flags,{captureGroup:n.rules.captureGroup,singleline:n.rules.singleline}),{skipBackrefValidation:n.rules.allowOrphanBackrefs,verbose:n.verbose}),{accuracy:n.accuracy,asciiWordBoundaries:n.rules.asciiWordBoundaries,avoidSubclass:n.avoidSubclass,bestEffortTarget:n.target}),a=function(e,t){const n=na(t),r=Zr(n.target,"ES2024"),a=Zr(n.target,"ES2025"),o=n.rules.recursionLimit;if(!Number.isInteger(o)||o<2||o>20)throw new Error("Invalid recursionLimit; use 2-20");let i=null,l=null;if(!a){const t=[e.flags.ignoreCase];lo({node:e},{getCurrentModI:()=>t.at(-1),popModI(){t.pop()},pushModI(e){t.push(e)},setHasCasedChar(){t.at(-1)?i=!0:l=!0}},Zo)}const s={dotAll:e.flags.dotAll,ignoreCase:!(!e.flags.ignoreCase&&!i||l)};let u=null;const c={accuracy:n.accuracy,appliedGlobalFlags:s,captureMap:new Map,currentFlags:{dotAll:e.flags.dotAll,ignoreCase:e.flags.ignoreCase},inCharClass:!1,lastNode:u,originMap:e._originMap,recursionLimit:o,useAppliedIgnoreCase:!(a||!i||!l),useFlagMods:a,useFlagV:r,verbose:n.verbose},p=function e(t){switch(c.lastNode=u,u=t,t.type){case uo.Regex:return{pattern:e(t.pattern),flags:e(t.flags),options:{...t.options}};case uo.Alternative:return t.elements.map(e).join("");case uo.Assertion:return function(e,t,n){const{kind:r,negate:a,alternatives:o}=e;if(io(e))return`(?${r===po.lookahead?"":"<"}${a?"!":"="}${o.map(n).join("|")})`;if(r===po.string_end)return"$";if(r===po.string_start)return"^";if(r===po.word_boundary)return a?Qr`\B`:Qr`\b`;throw new Error(`Unexpected assertion kind "${r}"`)}(t,0,e);case uo.Backreference:return function({ref:e},t){if("number"!=typeof e)throw new Error("Unexpected named backref in transformed AST");if(!t.useFlagMods&&"strict"===t.accuracy&&t.currentFlags.ignoreCase&&!t.captureMap.get(e).ignoreCase)throw new Error("Use of case-insensitive backref to case-sensitive group requires target ES2025 or non-strict accuracy");return"\\"+e}(t,c);case uo.CapturingGroup:return function(e,t,n){const{name:r,number:a,alternatives:o}=e,i={ignoreCase:t.currentFlags.ignoreCase},l=t.originMap.get(e);return l&&(i.hidden=!0,a>l.number&&(i.transferTo=l.number)),t.captureMap.set(a,i),`(${r?`?<${r}>`:""}${o.map(n).join("|")})`}(t,c,e);case uo.Character:return function({value:e},t){const n=Vr(e),r=ii(e,{isAfterBackref:t.lastNode.type===uo.Backreference,inCharClass:t.inCharClass,useFlagV:t.useFlagV});if(r!==n)return r;if(t.useAppliedIgnoreCase&&t.currentFlags.ignoreCase&&ai(n)){const e=aa(n);return t.inCharClass?e.join(""):e.length>1?`[${e.join("")}]`:e[0]}return n}(t,c);case uo.CharacterClass:return function({negate:e,parent:t,elements:n},r,a){Xr.literalHyphenIncorrectlyCreatesRange&&r.useFlagV&&n.some(li)&&(n=n.filter((e=>!li(e))),n.push(ko(45)));const o=()=>`[${e?"^":""}${n.map(a).join("")}]`;if(!r.inCharClass){r.inCharClass=!0;const e=o();return r.inCharClass=!1,e}const i=n[0]?.type;if(!e&&i&&((!r.useFlagV||!r.verbose)&&t.type===uo.CharacterClass&&i!==uo.CharacterClassIntersection&&(!Xr.literalHyphenIncorrectlyCreatesRange||!r.useFlagV)||!r.verbose&&t.type===uo.CharacterClassIntersection&&1===n.length&&i!==uo.CharacterClass&&i!==uo.CharacterClassRange))return n.map(a).join("");if(!r.useFlagV&&t.type===uo.CharacterClass)throw new Error("Use of nested character class requires min target ES2024");return o()}(t,c,e);case uo.CharacterClassIntersection:if(!c.useFlagV)throw new Error("Use of class intersection requires min target ES2024");return t.classes.map(e).join("&&");case uo.CharacterClassRange:return function(e,t){const n=e.min.value,r=e.max.value,a={isAfterBackref:!1,inCharClass:!0,useFlagV:t.useFlagV},o=ii(n,a),i=ii(r,a),l=new Set;return t.useAppliedIgnoreCase&&t.currentFlags.ignoreCase&&function(e){const t=e.map((e=>e.codePointAt(0))).sort(((e,t)=>e-t)),n=[];let r=null;for(let e=0;e<t.length;e++)t[e+1]===t[e]+1?r??=t[e]:null===r?n.push(t[e]):(n.push([r,t[e]]),r=null);return n}(oi(e)).forEach((e=>{l.add(Array.isArray(e)?`${ii(e[0],a)}-${ii(e[1],a)}`:ii(e,a))})),`${o}-${i}${[...l].join("")}`}(t,c);case uo.CharacterSet:return function({kind:e,negate:t,value:n,key:r},a){if(e===fo.dot)return a.currentFlags.dotAll?a.appliedGlobalFlags.dotAll||a.useFlagMods?".":"[^]":Qr`[^\n]`;if(e===fo.digit)return t?Qr`\D`:Qr`\d`;if(e===fo.property){if(a.useAppliedIgnoreCase&&a.currentFlags.ignoreCase&&ya.has(n))throw new Error(`Unicode property "${n}" can't be case-insensitive when other chars have specific case`);return`${t?Qr`\P`:Qr`\p`}{${r?`${r}=`:""}${n}}`}if(e===fo.word)return t?Qr`\W`:Qr`\w`;throw new Error(`Unexpected character set kind "${e}"`)}(t,c);case uo.Flags:return function(e,t){return(t.appliedGlobalFlags.ignoreCase?"i":"")+(e.dotAll?"s":"")+(e.sticky?"y":"")}(t,c);case uo.Group:return function({atomic:e,flags:t,parent:n,alternatives:r},a,o){const i=a.currentFlags;t&&(a.currentFlags=Kr(i,t));const l=r.map(o).join("|"),s=a.verbose||1!==r.length||n.type===uo.Quantifier||e||a.useFlagMods&&t?`(?${function(e,t,n){if(e)return">";let r="";if(t&&n){const{enable:e,disable:n}=t;r=(e?.ignoreCase?"i":"")+(e?.dotAll?"s":"")+(n?"-":"")+(n?.ignoreCase?"i":"")+(n?.dotAll?"s":"")}return`${r}:`}(e,t,a.useFlagMods)}${l})`:l;return a.currentFlags=i,s}(t,c,e);case uo.Pattern:return t.alternatives.map(e).join("|");case uo.Quantifier:return e(t.element)+function({min:e,max:t,greedy:n,possessive:r}){let a;return a=e||1!==t?e||t!==1/0?1===e&&t===1/0?"+":e===t?`{${e}}`:`{${e},${t===1/0?"":t}}`:"*":"?",a+(r?"+":n?"":"?")}(t);case uo.Recursion:return function({ref:e},t){const n=t.recursionLimit;return 0===e?`(?R=${n})`:Qr`\g<${e}&R=${n}>`}(t,c);default:throw new Error(`Unexpected node type "${t.type}"`)}}(e);return r||(delete p.options.force.v,p.options.disable.v=!0,p.options.unicodeSetsPlugin=null),p._captureTransfers=new Map,p._hiddenCaptures=[],c.captureMap.forEach(((e,t)=>{e.hidden&&p._hiddenCaptures.push(t),e.transferTo&&Yr(p._captureTransfers,e.transferTo,[]).push(t)})),p}(r,n),o=function(e,t){const{hiddenCaptures:n,mode:r}={hiddenCaptures:[],mode:"plugin",...t};let a=t?.captureTransfers??new Map;if(!new RegExp(zr,"su").test(e))return{pattern:e,captureTransfers:a,hiddenCaptures:n};if("plugin"===r&&Lr(e,Or`\(\?\(DEFINE\)`,Nr.DEFAULT))throw new Error("DEFINE groups cannot be used with recursion");const o=[],i=Lr(e,Or`\\[1-9]`,Nr.DEFAULT),l=new Map,s=[];let u,c=!1,p=0,d=0;for(Dr.lastIndex=0;u=Dr.exec(e);){const{0:t,groups:{captureName:f,rDepth:h,gRNameOrNum:g,gRDepth:m}}=u;if("["===t)p++;else if(p)"]"===t&&p--;else{if(h){if(Ur(h),c)throw new Error(Br);if(i)throw new Error(("external"===r?"Backrefs":"Numbered backrefs")+" cannot be used with global recursion");const t=e.slice(0,u.index),l=e.slice(Dr.lastIndex);if(Lr(l,zr,Nr.DEFAULT))throw new Error(Br);const s=+h-1;e=qr(t,l,s,!1,n,o,d),a=Hr(a,t,s,o.length,0,d);break}if(g){Ur(m);let p=!1;for(const e of s)if(e.name===g||e.num===+g){if(p=!0,e.hasRecursedWithin)throw new Error(Br);break}if(!p)throw new Error(Or`Recursive \g cannot be used outside the referenced group "${"external"===r?g:Or`\g<${g}&R=${m}>`}"`);const f=l.get(g),h=Tr(e,f);if(i&&Lr(h,Or`${$r}|\((?!\?)`,Nr.DEFAULT))throw new Error(("external"===r?"Backrefs":"Numbered backrefs")+" cannot be used with recursion of capturing groups");const y=e.slice(f,u.index),b=h.slice(y.length+t.length),v=o.length,w=+m-1,x=qr(y,b,w,!0,n,o,d);a=Hr(a,y,w,o.length-v,v,d);const k=e.slice(0,f),S=e.slice(f+h.length);e=`${k}${x}${S}`,Dr.lastIndex+=x.length-t.length-y.length-b.length,s.forEach((e=>e.hasRecursedWithin=!0)),c=!0}else if(f)d++,l.set(String(d),Dr.lastIndex),l.set(f,Dr.lastIndex),s.push({num:d,name:f});else if("("===t[0]){const e="("===t;e&&(d++,l.set(String(d),Dr.lastIndex)),s.push(e?{num:d}:{})}else")"===t&&s.pop()}}return n.push(...o),{pattern:e,captureTransfers:a,hiddenCaptures:n}}(a.pattern,{captureTransfers:a._captureTransfers,hiddenCaptures:a._hiddenCaptures,mode:"external"}),i=function(e,t){const n=t?.hiddenCaptures??[];let r=t?.captureTransfers??new Map;if(!/\(\?>/.test(e))return{pattern:e,captureTransfers:r,hiddenCaptures:n};const a=[0],o=[];let i,l=0,s=0,u=NaN;do{i=!1;let t,c=0,p=0,d=!1;for(Rr.lastIndex=Number.isNaN(u)?0:u+7;t=Rr.exec(e);){const{0:f,index:h,groups:{capturingStart:g,noncapturingStart:m}}=t;if("["===f)c++;else if(c)"]"===f&&c--;else if("(?>"!==f||d){if(d&&m)p++;else if(g)d?p++:(l++,a.push(l+s));else if(")"===f&&d){if(!p){s++;const t=l+s;if(e=`${e.slice(0,u)}(?:(?=(${e.slice(u+3,h)}))<$$${t}>)${e.slice(h+1)}`,i=!0,o.push(t),Er(n,t),r.size){const e=new Map;r.forEach(((n,r)=>{e.set(r>=t?r+1:r,n.map((e=>e>=t?e+1:e)))})),r=e}break}p--}}else u=h,d=!0}}while(i);return n.push(...o),{pattern:e=Pr(e,String.raw`\\(?<backrefNum>[1-9]\d*)|<\$\$(?<wrappedBackrefNum>\d+)>`,(({0:e,groups:{backrefNum:t,wrappedBackrefNum:n}})=>{if(t){const n=+t;if(n>a.length-1)throw new Error(`Backref "${e}" greater than number of captures`);return`\\${a[n]}`}return`\\${n}`}),Nr.DEFAULT),captureTransfers:r,hiddenCaptures:n}}(function(e){if(!new RegExp(`${Ir}\\+`).test(e))return{pattern:e};const t=[];let n,r=null,a=null,o="",i=0;for(Mr.lastIndex=0;n=Mr.exec(e);){const{0:c,index:p,groups:{qBase:d,qMod:f,invalidQ:h}}=n;if("["===c)i||(a=p),i++;else if("]"===c)i?i--:a=null;else if(!i)if("+"===f&&o&&!o.startsWith("(")){if(h)throw new Error(`Invalid quantifier "${c}"`);let t=-1;if(/^\{\d+\}$/.test(d))l=e,s=p+d.length,u=f,e=l.slice(0,s)+""+l.slice(s+u.length);else{if(")"===o||"]"===o){const t=")"===o?r:a;if(null===t)throw new Error(`Invalid unmatched "${o}"`);e=`${e.slice(0,t)}(?>${e.slice(t,p)}${d})${e.slice(p+c.length)}`}else e=`${e.slice(0,p-o.length)}(?>${o}${d})${e.slice(p+c.length)}`;t+=4}Mr.lastIndex+=t}else"("===c[0]?t.push(p):")"===c&&(r=t.length?t.pop():null);o=c}var l,s,u;return{pattern:e}}(o.pattern).pattern,{captureTransfers:o.captureTransfers,hiddenCaptures:o.hiddenCaptures}),l={pattern:i.pattern,flags:`${n.hasIndices?"d":""}${n.global?"g":""}${a.flags}${a.options.disable.v?"u":"v"}`};if(n.avoidSubclass){if(n.lazyCompileLength!==1/0)throw new Error("Lazy compilation requires subclass")}else{const e=i.hiddenCaptures.sort(((e,t)=>e-t)),t=Array.from(i.captureTransfers),a=r._strategy,o=l.pattern.length>=n.lazyCompileLength;(e.length||t.length||a||o)&&(l.options={...e.length&&{hiddenCaptures:e},...t.length&&{transfers:t},...a&&{strategy:a},...o&&{lazyCompile:o}})}return l}(e,t);return n.options?new si(n.pattern,n.flags,n.options):new RegExp(n.pattern,n.flags)}const pi=4294967295;class di{constructor(e,t={}){this.patterns=e,this.options=t;const{forgiving:n=!1,cache:r,regexConstructor:a}=t;if(!a)throw new Error("Option `regexConstructor` is not provided");this.regexps=e.map((e=>{if("string"!=typeof e)return e;const t=r?.get(e);if(t){if(t instanceof RegExp)return t;if(n)return null;throw t}try{const t=a(e);return r?.set(e,t),t}catch(t){if(r?.set(e,t),n)return null;throw t}}))}regexps;findNextMatchSync(e,t,n){const r="string"==typeof e?e:e.content,a=[];function o(e,t,n=0){return{index:e,captureIndices:t.indices.map((e=>null==e?{start:pi,end:pi,length:0}:{start:e[0]+n,end:e[1]+n,length:e[1]-e[0]}))}}for(let e=0;e<this.regexps.length;e++){const n=this.regexps[e];if(n)try{n.lastIndex=t;const i=n.exec(r);if(!i)continue;if(i.index===t)return o(e,i,0);a.push([e,i,0])}catch(e){if(this.options.forgiving)continue;throw e}}if(a.length){const e=Math.min(...a.map((e=>e[1].index)));for(const[t,n,r]of a)if(n.index===e)return o(t,n,r)}return null}}function fi(e={}){const t=Object.assign({target:"auto",cache:new Map},e);return t.regexConstructor||=e=>function(e,t){return ci(e,{global:!0,hasIndices:!0,lazyCompileLength:3e3,rules:{allowOrphanBackrefs:!0,asciiWordBoundaries:!0,captureGroup:!0,recursionLimit:5,singleline:!0},...t})}(e,{target:t.target}),{createScanner:e=>new di(e,t),createString:e=>({content:e})}}const hi={name:"vortico",settings:[{scope:["markup.changed","markup.deleted","markup.inserted","punctuation","variable","constant","source"],settings:{foreground:"#cdced4"}},{scope:["comment","punctuation.definition.comment"],settings:{foreground:"#575863"}},{scope:["constant.character.character-class.regexp","constant.character.set.regexp","constant.language","constant.numeric","constant.other.character-class.regexp","constant.other.character-class.set.regexp","constant.other.color.rgb-value","constant.other.rgb-value","constant.regexp","constant.sha.git-rebase","invalid","meta.class","meta.function-call.generic","support.function.git-rebase","token.error-token"],settings:{foreground:"#9e744f"}},{scope:["keyword","meta.structure.dictionary.key.python","support.class","support.function","support.type","support.type.vendored.property-name","support.type.property-name","storage"],settings:{foreground:"#e25822"}},{scope:["meta.attribute","meta.definition","meta.indexed-name","punctuation.character.set.begin.regexp","punctuation.character.set.end.regexp","punctuation.definition.character-class.regexp","punctuation.definition.group.regexp","punctuation.definition.group.assertion.regexp","punctuation.definition.template-expression.begin","punctuation.definition.template-expression.end","support.other.parenthesis.regexp"],settings:{foreground:"#00bbd5"}},{scope:["string","string.tag","string.value","string.regexp","string.quoted","punctuation.definition.string"],settings:{foreground:"#00976e"}},{scope:["header","variable.language","variable.other.enummember","variable.parameter","entity.name.function"],settings:{foreground:"#008080"}},{scope:["emphasis","markup.bold","markup.changed","markup.deleted","markup.heading","markup.inserted","markup.italic","markup.strikethrough","meta.template.expression","punctuation.definition.quote.begin.markdown","punctuation.definition.list.begin.markdown"],settings:{fontStyle:"italic"}},{scope:["strong"],settings:{fontStyle:"bold"}}]};function gi(e){return Bn(e)}let mi;function yi(){return void 0===mi&&(mi=function(e){const t=function(e){xr+=1,!1!==e.warnings&&xr>=10&&xr%10==0&&console.warn(`[Shiki] ${xr} instances have been created. Shiki is supposed to be used as a singleton, consider refactoring your code to cache your highlighter instance; Or call \`highlighter.dispose()\` to release unused instances.`);let t=!1;if(!e.engine)throw new br("`engine` option is required for synchronous mode");const n=(e.langs||[]).flat(1),r=(e.themes||[]).flat(1).map(gr),a=new wr(e.engine,n),o=new vr(a,r,n,e.langAlias);let i;function l(e){if("none"===e)return{bg:"",fg:"",name:"none",settings:[],type:"dark"};c();const t=o.getTheme(e);if(!t)throw new br(`Theme \`${e}\` not found, you may need to load it first`);return t}function s(...e){c(),o.loadLanguages(e.flat(1))}function u(...e){c();for(const t of e.flat(1))o.loadTheme(t)}function c(){if(t)throw new br("Shiki instance has been disposed")}function p(){t||(t=!0,o.dispose(),xr-=1)}return{setTheme:function(e){c();const t=l(e);return i!==e&&(o.setTheme(t),i=e),{theme:t,colorMap:o.getColorMap()}},getTheme:l,getLanguage:function(e){c();const t=o.getGrammar("string"==typeof e?e:e.name);if(!t)throw new br(`Language \`${e}\` not found, you may need to load it first`);return t},getLoadedThemes:function(){return c(),o.getLoadedThemes()},getLoadedLanguages:function(){return c(),o.getLoadedLanguages()},loadLanguage:async function(...e){return s(await mr(e))},loadLanguageSync:s,loadTheme:async function(...e){return c(),u(await yr(e))},loadThemeSync:u,dispose:p,[Symbol.dispose]:p}}(e);return{getLastGrammarState:(...e)=>function(...e){if(2===e.length)return Wn(e[1]);const[t,n,r={}]=e,{lang:a="text",theme:o=t.getLoadedThemes()[0]}=r;if(zn(a)||$n(o))throw new q("Plain language does not have grammar state");if("ansi"===a)throw new q("ANSI language does not have grammar state");const{theme:i,colorMap:l}=t.setTheme(o),s=t.getLanguage(a);return new Hn(or(n,s,i,l,r).stateStack,s.name,i.name)}(t,...e),codeToTokensBase:(e,n)=>ar(t,e,n),codeToTokensWithThemes:(e,n)=>cr(t,e,n),codeToTokens:(e,n)=>pr(t,e,n),codeToHast:(e,n)=>dr(t,e,n),codeToHtml:(e,n)=>fr(t,e,n),getBundledLanguages:()=>({}),getBundledThemes:()=>({}),...t,getInternalContext:()=>t}}({themes:[hi],langs:[U],engine:fi()})),mi}function bi(e){let{code:t,language:n,selectedLine:r}=e;const a=yi().codeToTokens(t,{lang:n,theme:hi}),o=1===a?.tokens.length;return(0,m.jsx)("code",{children:a?.tokens.map(((e,t)=>(0,m.jsx)("span",{className:`${o?"inline":"block w-full px-2 text-sm"} ${r===t+1?"bg-flama-500/25":""}`,children:0===e.length?(0,m.jsx)("br",{}):e.map(((e,t)=>(0,m.jsx)("span",{style:gi(e),children:e.content},t)))},t)))})}function vi(e){let{code:t,selectedLine:n}=e;const r=t.split("\n"),a=1===r.length;return(0,m.jsx)("code",{children:r.map(((e,t)=>(0,m.jsx)("span",{className:`line ${a?"inline":"block w-full px-2"} ${n===t+1?"bg-flama-700":""}`,children:0===e.length?(0,m.jsx)("br",{}):e.split(" ").map(((e,t)=>(0,m.jsx)("span",{className:"token",children:e},t)))},t)))})}function wi(e){let{code:t,language:n,lines:r,copyButton:a=!0,selectedLine:o,className:i}=e;const l=void 0===r&&1===t.split("\n").length,s=n?(0,m.jsx)(bi,{code:t,language:n,selectedLine:o}):(0,m.jsx)(vi,{code:t,selectedLine:o});return l?(0,m.jsx)(B,{className:i,children:s}):(0,m.jsx)(D,{code:t,lines:r,copyButton:a,className:i,selectedLine:o,children:s})}const xi=function(e){return(new DOMParser).parseFromString(e,"text/html").body.textContent||""};function ki(e){let{frame:t,isActive:n}=e;const r=t.filename.length>35&&t.filename.split("/").length>2?t.filename.split("/",1).concat(["...",t.filename.split("/").slice(-1)[0]]):t.filename.split("/");return(0,m.jsx)("div",{className:"border-primary-300 shadow-primary-300 -mt-px border shadow-md transition duration-200 "+(n?"bg-primary-200":"bg-primary-100 hover:bg-flama-300 hover:shadow-flama-300 hover:shadow-lg"),children:(0,m.jsxs)("div",{className:"flex h-12 w-full cursor-pointer items-center justify-start transition duration-200 "+(n?"border-primary-300 border-l-4":"hover:border-flama-500 border-primary-100 pl-1 hover:border-l-4 hover:pl-0"),children:[(0,m.jsx)("div",{className:"flex h-8 w-8 items-center justify-center pl-1 "+(t.vendor?"text-flama-400":"text-primary-500"),children:(0,m.jsx)(T,{className:"h-2 w-2"})}),(0,m.jsxs)("div",{className:"text-primary-600 overflow-hidden font-mono text-xs "+(n?"border-current font-semibold":"hover:text-primary-800"),children:[(0,m.jsxs)("div",{children:[r.map(((e,n)=>(0,m.jsxs)(l.Fragment,{children:[n>0&&(0,m.jsx)("span",{children:"/"}),(0,m.jsx)("span",{className:0===n&&t.vendor?"text-flama-500":"",children:e})]},`filename-${n}`))),(0,m.jsx)("span",{children:":"}),(0,m.jsx)("span",{children:t.line})]}),(0,m.jsx)("div",{className:"truncate",children:(0,m.jsx)("span",{className:"italic",children:t.function})})]})]})})}function Si(e){let{traceback:t,selected:n,setSelected:r}=e;const a=(0,l.useRef)(null),o=(0,l.useRef)(null);return(0,l.useEffect)((()=>{a.current&&o.current&&o.current.scrollTo({top:a.current.offsetTop-o.current.clientHeight/2,left:0,behavior:"smooth"})}),[a,o]),(0,m.jsx)("ul",{ref:o,className:"mt-px h-full w-full overflow-y-scroll py-6 pr-3 pl-2",children:t.map(((e,t)=>(0,m.jsx)("li",{ref:n===t?a:void 0,onClick:()=>r(t),children:(0,m.jsx)(ki,{frame:e,isActive:n===t})},`traceback-frame-${t}`)))})}function Ci(){const{traceback:e}=new o,[t,n]=(0,l.useState)(e.length-1),r=(0,l.useMemo)((()=>xi(e[t].code)),[t,e]);return(0,m.jsxs)("div",{className:"flex flex-col gap-8 lg:flex-row",children:[(0,m.jsx)("div",{className:"h-[672px] max-h-[calc(85vh)] w-full flex-none overflow-hidden lg:w-xs",children:(0,m.jsx)(Si,{traceback:e,selected:t,setSelected:n})}),(0,m.jsx)("div",{className:"h-[672px] max-h-[calc(85vh)] flex-auto overflow-hidden",children:(0,m.jsx)(O,{title:e[t].filename,children:(0,m.jsx)(wi,{code:r,language:"python",lines:{type:"number"},selectedLine:e[t].line})})})]})}function _i(e){let{children:t}=e;return(0,m.jsx)("table",{className:"border-primary-300 w-full table-fixed border-collapse overflow-hidden border",children:(0,m.jsx)("tbody",{className:"text-left text-wrap",children:t})})}function Ei(e){let{header:t,children:n,...r}=e;return(0,m.jsxs)("tr",{...r,children:[(0,m.jsx)("th",{className:"border-primary-300 w-40 border px-4 py-1",children:(0,m.jsx)("span",{className:"text-flama",children:t})}),(0,m.jsx)("td",{className:"border-primary-300 border px-4 py-1",children:n})]})}function Ni(e){let{value:t}=e;return(0,m.jsx)("span",{children:xi(String(t))})}function Pi(e){let{map:t}=e;return(0,m.jsx)("ul",{children:Array.from(t).map(((e,t)=>{let[n,r]=e;return(0,m.jsxs)("li",{children:[(0,m.jsx)("span",{className:"font-semibold",children:`${n}:`}),(0,m.jsx)("span",{className:"ml-1",children:(0,m.jsx)(Ni,{value:r})})]},t)}))})}function Ai(){const{path:e,method:t,pathParams:n,queryParams:r,headers:a,cookies:o,clientHost:l,clientPort:s}=new i;return(0,m.jsxs)(_i,{children:[(0,m.jsx)(Ei,{header:"Path",children:(0,m.jsx)(Ni,{value:e})}),(0,m.jsx)(Ei,{header:"Method",children:(0,m.jsx)(Ni,{value:t})}),(0,m.jsx)(Ei,{header:"Query parameters",children:(0,m.jsx)(Pi,{map:r})}),(0,m.jsx)(Ei,{header:"Path parameters",children:(0,m.jsx)(Pi,{map:n})}),(0,m.jsx)(Ei,{header:"Headers",children:(0,m.jsx)(Pi,{map:a})}),(0,m.jsx)(Ei,{header:"Cookies",children:(0,m.jsx)(Pi,{map:o})}),l&&(0,m.jsx)(Ei,{header:"Client host",children:(0,m.jsx)(Ni,{value:l})}),s&&(0,m.jsx)(Ei,{header:"Client port",children:(0,m.jsx)(Ni,{value:String(s)})})]})}class Li{constructor(){this.pythonVersion="||@ environment.python_version|safe @||",this.python="||@ environment.python|safe @||",this.platform="||@ environment.platform|safe @||",this.path=JSON.parse("||@ environment.path|safe_json @||")}}function Ti(e){let{array:t}=e;return(0,m.jsx)("ul",{children:t.map(((e,t)=>(0,m.jsx)("li",{children:(0,m.jsx)(Ni,{value:e})},t)))})}function Ri(){const{pythonVersion:e,python:t,platform:n,path:r}=new Li;return(0,m.jsxs)(_i,{children:[(0,m.jsx)(Ei,{header:"Python version",children:(0,m.jsx)(Ni,{value:e})}),(0,m.jsx)(Ei,{header:"Python",children:(0,m.jsx)(Ni,{value:t})}),(0,m.jsx)(Ei,{header:"Platform",children:(0,m.jsx)(Ni,{value:n})}),(0,m.jsx)(Ei,{header:"Path",children:(0,m.jsx)(Ti,{array:r})})]})}var Ii=u("outline","brand-linkedin","IconBrandLinkedin",[["path",{d:"M8 11v5",key:"svg-0"}],["path",{d:"M8 8v.01",key:"svg-1"}],["path",{d:"M12 16v-5",key:"svg-2"}],["path",{d:"M16 16v-3a2 2 0 1 0 -4 0",key:"svg-3"}],["path",{d:"M3 7a4 4 0 0 1 4 -4h10a4 4 0 0 1 4 4v10a4 4 0 0 1 -4 4h-10a4 4 0 0 1 -4 -4z",key:"svg-4"}]]),Mi=u("outline","brand-x","IconBrandX",[["path",{d:"M4 4l11.733 16h4.267l-11.733 -16z",key:"svg-0"}],["path",{d:"M4 20l6.768 -6.768m2.46 -2.46l6.772 -6.772",key:"svg-1"}]]),Oi=u("outline","brand-medium","IconBrandMedium",[["path",{d:"M4 4m0 2a2 2 0 0 1 2 -2h12a2 2 0 0 1 2 2v12a2 2 0 0 1 -2 2h-12a2 2 0 0 1 -2 -2z",key:"svg-0"}],["path",{d:"M8 9h1l3 3l3 -3h1",key:"svg-1"}],["path",{d:"M8 15l2 0",key:"svg-2"}],["path",{d:"M14 15l2 0",key:"svg-3"}],["path",{d:"M9 9l0 6",key:"svg-4"}],["path",{d:"M15 9l0 6",key:"svg-5"}]]);const ji=[{url:"https://github.com/vortico/",icon:(0,m.jsx)(d,{})},{url:"https://www.linkedin.com/company/vortico-tech/",icon:(0,m.jsx)(Ii,{})},{url:"https://twitter.com/vorticotech",icon:(0,m.jsx)(Mi,{})},{url:"https://dev.to/vortico",icon:(0,m.jsx)(Oi,{})}];function zi(){return(0,m.jsx)("footer",{children:(0,m.jsx)("div",{className:"border-vortico-500/50 border-t px-4 py-2 sm:px-6 md:px-8",children:(0,m.jsxs)("div",{className:"mx-auto flex max-w-7xl items-center justify-between",children:[(0,m.jsx)("a",{href:"https://vortico.tech",children:(0,m.jsx)(S,{logo:"vortico",color:"vortico",size:"lg"})}),(0,m.jsx)("div",{className:"divide-vortico-500/50 flex h-8 items-center justify-end divide-x",children:ji.map(((e,t)=>(0,m.jsx)("a",{href:e.url,className:"px-2 first:pl-0 last:pr-0",children:(0,m.jsx)("div",{className:"text-primary-400 hover:text-vortico-500 h-6 w-6 transition-colors duration-200",children:e.icon})},t)))})]})})})}function $i(){const e=new o,t=new i;return(0,m.jsxs)("div",{children:[(0,m.jsx)(P,{}),(0,m.jsxs)("div",{className:"mt-[calc(3.5rem+1px)] h-[calc(100vh-6.5rem-2px)] overflow-auto",children:[(0,m.jsx)("header",{children:(0,m.jsx)(A,{id:"error",border:!1,children:(0,m.jsx)(L,{error:e.error,path:t.path,method:t.method,description:e.description})})}),(0,m.jsxs)("main",{children:[(0,m.jsx)(A,{id:"traceback",title:"Traceback",children:(0,m.jsx)(Ci,{})}),(0,m.jsx)(A,{id:"request",title:"Request",children:(0,m.jsx)(Ai,{})}),(0,m.jsx)(A,{id:"environment",title:"Environment",children:(0,m.jsx)(Ri,{})})]})]}),(0,m.jsx)(zi,{})]})}a.createRoot(document.getElementById("app")).render((0,m.jsx)($i,{}))})();</script></body></html>